﻿/** @file cl_frame.cpp  Frame Reception.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/cl_frame.h"
#include "client/cl_mobj.h"
#include "client/cl_player.h"
#include "client/cl_sound.h"
#include "client/cl_world.h"
#include "network/net_main.h"
#include "network/net_buf.h"
#include "network/net_msg.h"
#if 0
#define SET_HISTORY_SIZE    50
#define RESEND_HISTORY_SIZE 50
#endif
/** @file cl_infine.cpp  Clientside InFine.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/cl_infine.h"
#include <de/memory.h>
#include "network/net_main.h"
#include "network/net_msg.h"
#include "ui/infine/finaleinterpreter.h"
/** @file cl_main.cpp  Network client.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/cl_def.h"
#include "api_client.h"
#include "client/cl_frame.h"
#include "client/cl_infine.h"
#include "client/cl_player.h"
#include "client/cl_sound.h"
#include "client/cl_world.h"
#include "network/net_main.h"
#include "network/net_buf.h"
#include "network/net_demo.h"
#include "world/map.h"
#include "world/p_players.h"
#include <doomsday/console/exec.h>
#include <doomsday/filesys/fs_main.h>
#include <de/CommandLine>
#include <de/Config>
#include <de/timer.h>
#include <cstdio>
#include <cstdlib>
#include <ctime>
/** @file cl_mobj.cpp  Client map objects.
 * @ingroup client
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#define DENG_NO_API_MACROS_CLIENT
#include "de_base.h"
#include "client/cl_mobj.h"
#include "api_client.h"
#include "client/cl_player.h"
#include "client/cl_world.h"
#include "network/net_main.h"
#include "network/protocol.h"
#include "world/map.h"
#include "world/p_players.h"
#include "api_sound.h"
#include <de/timer.h>
#include <de/vector1.h>
#include <de/LogBuffer>
#include <cmath>
/** @file cl_player.cpp  Clientside player management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/cl_player.h"
#include "api_client.h"
#include "network/net_main.h"
#include "network/protocol.h"
#include "world/map.h"
#include "world/p_players.h"
#include "BspLeaf"
#include "Sector"
#include <de/LogBuffer>
#include <de/Vector>
/** @file cl_sound.cpp  Clientside sounds.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/cl_sound.h"
#include "api_client.h"
#include "client/cl_player.h"
#include "api_sound.h"
#include "network/net_msg.h"
#include "world/map.h"
#include "world/p_players.h"
#include "Sector"
#include <de/LogBuffer>
/** @file cl_world.cpp  Clientside world management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "client/cl_world.h"
#include <QVector>
#include <de/stringarray.h>
#include <doomsday/world/MaterialArchive>
#include "client/cl_player.h"
#include "api_map.h"
#include "world/map.h"
#include "Sector"
#include "Surface"
#include "network/net_msg.h"
#include "network/protocol.h"
/** @file cledgeloop.cpp  Client-side world map subsector boundary loop.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "client/cledgeloop.h"
#include "world/maputil.h"
#include "world/surface.h"
#include "world/convexsubspace.h"
#include "misc/face.h"
#include <doomsday/world/Material>
#include <doomsday/world/Materials>
#include <de/Log>
/** @file clientsubsector.cpp  Client-side world map subsector.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "client/clientsubsector.h"
#include "world/map.h"
#include "world/blockmap.h"
#include "world/convexsubspace.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/surface.h"
#include "client/clskyplane.h"
#include "client/cledgeloop.h"
#include "render/rend_main.h" // Rend_SkyLightColor(), useBias
#include "LightDecoration"
#include "MaterialAnimator"
#include "WallEdge"
#include "misc/face.h"
#include "dd_main.h"  // verbose
#include <de/LogBuffer>
#include <QtAlgorithms>
#include <QHash>
#include <QMap>
#include <QRect>
#include <QSet>
/** @file clplanemover.cpp  Clientside plane mover (thinker).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/clplanemover.h"
#include "client/cl_def.h"
#include "client/cl_player.h"
#include "world/map.h"
#include "world/thinkers.h"
#include "world/p_players.h"
#include "Sector"
#include <de/LogBuffer>
/** @file clpolymover.cpp  Clientside polyobj mover (thinker).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "client/clpolymover.h"
#include "client/cl_def.h"
#include "client/cl_player.h"
#include "world/map.h"
#include "world/p_players.h"
#include "world/thinkers.h"
#include "world/polyobjdata.h"
#include <de/LogBuffer>
/** @file clskyplane.cpp  Client-side world map sky plane.
 *
 * @authors Copyright © 2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "client/clskyplane.h"
using namespace world;
using namespace de;
extern int gotFrame;
dd_bool gotFirstFrame;
static float frameGameTime = 0;
#if 0
static uint     latestSetOrdinal;
static byte     latestSet;
static uint     setOrdinalBase;
static short    setHistory[SET_HISTORY_SIZE];
static int      historyIdx;
static byte     resendHistory[RESEND_HISTORY_SIZE];
static int      resendHistoryIdx;
#endif
/**
 * Clear the history of received set numbers.
 */
void Cl_InitFrame(void) {
    Cl_ResetFrame();
#if 0
    memset(setHistory, -1, sizeof(setHistory));
    historyIdx = 0;
    memset(resendHistory, 0, sizeof(resendHistory));
    resendHistoryIdx = 0;
    latestSet = 0;
    latestSetOrdinal = 0;
    setOrdinalBase = 256;
#endif
}
/**
 * Called when the map changes.
 */
void Cl_ResetFrame() {
    gotFrame = false;
    gotFirstFrame = false;
}
float Cl_FrameGameTime() {
    return frameGameTime;
}
void Cl_Frame2Received(int packetType) {
    frameGameTime = Reader_ReadFloat(msgReader);
    if (packetType == PSV_FIRST_FRAME2) {
        gotFirstFrame = true;
    }
    else if (!gotFirstFrame) {
        return;
    }
    while (!Reader_AtEnd(msgReader)) {
        byte const deltaType = Reader_ReadByte(msgReader);
        switch (deltaType) {
        case DT_CREATE_MOBJ:
            ClMobj_ReadDelta();
            break;
        case DT_MOBJ:
            ClMobj_ReadDelta();
            break;
        case DT_NULL_MOBJ:
            ClMobj_ReadNullDelta();
            break;
        case DT_PLAYER:
            ClPlayer_ReadDelta();
            break;
        case DT_SECTOR:
            Cl_ReadSectorDelta(deltaType);
            break;
        case DT_SIDE:
            Cl_ReadSideDelta(deltaType);
            break;
        case DT_POLY:
            Cl_ReadPolyDelta();
            break;
        case DT_SOUND:
        case DT_MOBJ_SOUND:
        case DT_SECTOR_SOUND:
        case DT_SIDE_SOUND:
        case DT_POLY_SOUND:
            Cl_ReadSoundDelta((deltatype_t) deltaType);
            break;
        default:
            LOG_NET_ERROR("Received unknown delta type %i (message size: %i bytes)")
                    << deltaType << netBuffer.length;
            return;
        }
    }
    if (!gotFrame) {
        LOGDEV_NET_NOTE("First frame received");
    }
    gotFrame = true;
}
static finaleid_t currentFinale;
static finaleid_t remoteFinale;
finaleid_t Cl_CurrentFinale() {
    return currentFinale;
}
void Cl_Finale(Reader1 *msg) {
    LOG_AS("Cl_Finale");
    byte *script = 0;
    int const flags           = Reader_ReadByte(msg);
    finaleid_t const finaleId = Reader_ReadUInt32(msg);
    if (flags & FINF_SCRIPT) {
        int len = Reader_ReadUInt32(msg);
        script = (byte *) M_Malloc(len + 1);
        Reader_Read(msg, script, len);
        script[len] = 0;
    }
    if ((flags & FINF_SCRIPT) && (flags & FINF_BEGIN)) {
        currentFinale = FI_Execute((const char*)script, FF_LOCAL);
        remoteFinale = finaleId;
        LOGDEV_NET_MSG("Started finale %i (remote id %i)") << currentFinale << remoteFinale;
    }
    if ((flags & FINF_END) && currentFinale) {
        FI_ScriptTerminate(currentFinale);
        currentFinale = 0;
        remoteFinale = 0;
    }
    if ((flags & FINF_SKIP) && currentFinale) {
        FI_ScriptRequestSkip(currentFinale);
    }
    if (script) M_Free(script);
}
void Cl_RequestFinaleSkip() {
    Msg_Begin(PCL_FINALE_REQUEST);
    Writer_WriteUInt32(msgWriter, remoteFinale);
    Writer_WriteUInt16(msgWriter, 1); // skip
    Msg_End();
    LOGDEV_NET_MSG("Requesting skip on finale %i") << remoteFinale;
    Net_SendBuffer(0, 0);
}
static String const VAR_NETWORK_ID("network.ID");
ident_t clientID;
bool handshakeReceived;
int gameReady;
int serverTime;
bool netLoggedIn; // Logged in to the server.
int clientPaused; // Set by the server.
void Cl_InitID() {
    if (auto arg = CommandLine::get().check("-id", 1)) {
        bool ok;
        auto newId = arg.params.at(0).toUInt(&ok, 0);
        if (ok) {
            clientID = newId;
            LOG_NET_NOTE("Using custom client ID: 0x%08x") << clientID;
            return;
        }
        else {
            LOG_NET_WARNING("Option '-id' was given invalid argument: %s")
                    << arg.params.at(0);
        }
    }
    auto &config = Config::get();
    if (config.has(VAR_NETWORK_ID)) {
        clientID = config.gets(VAR_NETWORK_ID).toUInt(nullptr, 16);
        return;
    }
    clientID = ident_t(Timer_RealMilliseconds() * rand() +
                       (rand() & 0xfff) +
                       ((rand() & 0xfff) << 12) +
                       ((rand() & 0xff) << 24));
    config.set(VAR_NETWORK_ID, String::format("%x", clientID));
}
int Cl_GameReady() {
    return (handshakeReceived && gameReady);
}
void Cl_CleanUp() {
    LOG_NET_MSG("Cleaning up client state");
    clientPaused = false;
    handshakeReceived = false;
    App_AudioSystem().aboutToUnloadMap();
    App_World().reset();
    Cl_ResetTransTables();
    GL_ResetViewEffects();
    N_ClearMessages();
}
void Cl_SendHello() {
    LOG_AS("Cl_SendHello");
    Msg_Begin(PCL_HELLO2);
    Writer_WriteUInt32(msgWriter, clientID);
    char buf[256]; zap(buf);
    strncpy(buf, App_CurrentGame().id().toUtf8().constData(), sizeof(buf) - 1);
    LOGDEV_NET_VERBOSE("game mode = %s") << buf;
    Writer_Write(msgWriter, buf, 16);
    Msg_End();
    Net_SendBuffer(0, 0);
}
void Cl_AnswerHandshake() {
    LOG_AS("Cl_AnswerHandshake");
    byte remoteVersion   = Reader_ReadByte(msgReader);
    byte myConsole       = Reader_ReadByte(msgReader);
    uint playersInGame   = Reader_ReadUInt32(msgReader);
    float remoteGameTime = Reader_ReadFloat(msgReader);
    Msg_Begin(PCL_ACK_SHAKE);
    Msg_End();
    Net_SendBuffer(0, 0);
    if (remoteVersion != SV_VERSION) {
        LOG_NET_ERROR("Version conflict! (you:%i, server:%i)")
                << SV_VERSION << remoteVersion;
        Con_Execute(CMDS_DDAY, "net disconnect", false, false);
        Demo_StopPlayback();
        Con_Open(true);
        return;
    }
    gameTime = remoteGameTime;
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        DD_Player(i)->publicData().flags &= ~DDPF_LOCAL;
        DD_Player(i)->publicData().inGame = (playersInGame & (1 << i)) != 0;
    }
    consolePlayer = displayPlayer = myConsole;
    DD_Player(consolePlayer)->viewConsole = consolePlayer;
    DD_Player(consolePlayer)->publicData().flags |= DDPF_LOCAL;
    Smoother_Clear(DD_Player(consolePlayer)->smoother());
    DD_Player(consolePlayer)->publicData().flags &= ~DDPF_USE_VIEW_FILTER;
    isClient = true;
    isServer = false;
    netLoggedIn = false;
    clientPaused = false;
    if (handshakeReceived)
        return;
    handshakeReceived = true;
    gameReady = false;
    Cl_InitFrame();
    LOGDEV_NET_MSG("Answering handshake: myConsole:%i, remoteGameTime:%.2f")
            << myConsole << remoteGameTime;
    /*
     * Tell the game that we have arrived. The map will be changed when the
     * game's handshake arrives (handled in the game).
     */
    gx.NetPlayerEvent(consolePlayer, DDPE_ARRIVAL, 0);
    Cl_InitPlayers();
    Cl_InitTransTables();
    DD_ResetTimer();
    Con_Executef(CMDS_DDAY, true, "setcon %i", consolePlayer);
}
void Cl_HandlePlayerInfo() {
    byte console = Reader_ReadByte(msgReader);
    size_t len = Reader_ReadUInt16(msgReader);
    len = MIN_OF(PLAYERNAMELEN - 1, len);
    char name[PLAYERNAMELEN]; zap(name);
    Reader_Read(msgReader, name, len);
    LOG_NET_VERBOSE("Player %i named \"%s\"") << console << name;
    if (console >= DDMAXPLAYERS)
        return;
    player_t *plr = DD_Player(console);
    bool present = plr->publicData().inGame;
    plr->publicData().inGame = true;
    strcpy(plr->name, name);
    if (!present) {
        gx.NetPlayerEvent(console, DDPE_ARRIVAL, 0);
        Smoother_Clear(plr->smoother());
    }
}
void Cl_PlayerLeaves(int plrNum) {
    LOG_NET_NOTE("Player %i has left the game") << plrNum;
    DD_Player(plrNum)->publicData().inGame = false;
    gx.NetPlayerEvent(plrNum, DDPE_EXIT, 0);
}
void Cl_GetPackets() {
    while (Net_GetPacket()) {
        Msg_BeginRead();
        if (Cl_GameReady()) {
            switch (netBuffer.msg.type) {
            case PSV_FIRST_FRAME2:
            case PSV_FRAME2:
                Cl_Frame2Received(netBuffer.msg.type);
                Msg_EndRead();
                continue; // Get the next packet.
            case PSV_SOUND:
                Cl_Sound();
                Msg_EndRead();
                continue; // Get the next packet.
            default: break;
            }
        }
        switch (netBuffer.msg.type) {
        case PSV_PLAYER_FIX:
            ClPlayer_HandleFix();
            break;
        case PKT_DEMOCAM:
        case PKT_DEMOCAM_RESUME:
            Demo_ReadLocalCamera();
            break;
        case PKT_PING:
            Net_PingResponse();
            break;
        case PSV_SYNC:
            gameTime = Reader_ReadFloat(msgReader);
            LOGDEV_NET_VERBOSE("PSV_SYNC: gameTime=%.3f") << gameTime;
            DD_ResetTimer();
            break;
        case PSV_HANDSHAKE:
            Cl_AnswerHandshake();
            break;
        case PSV_MATERIAL_ARCHIVE:
            Cl_ReadServerMaterials();
            break;
        case PSV_MOBJ_TYPE_ID_LIST:
            Cl_ReadServerMobjTypeIDs();
            break;
        case PSV_MOBJ_STATE_ID_LIST:
            Cl_ReadServerMobjStateIDs();
            break;
        case PKT_PLAYER_INFO:
            Cl_HandlePlayerInfo();
            break;
        case PSV_PLAYER_EXIT:
            Cl_PlayerLeaves(Reader_ReadByte(msgReader));
            break;
        case PKT_CHAT: {
            int msgfrom = Reader_ReadByte(msgReader);
            int mask = Reader_ReadUInt32(msgReader);
            DENG2_UNUSED(mask);
            size_t len = Reader_ReadUInt16(msgReader);
            char *msg = (char *) M_Malloc(len + 1);
            Reader_Read(msgReader, msg, len);
            msg[len] = 0;
            Net_ShowChatMessage(msgfrom, msg);
            gx.NetPlayerEvent(msgfrom, DDPE_CHAT_MESSAGE, msg);
            M_Free(msg);
            break; }
        case PSV_SERVER_CLOSE:  // We should quit?
            netLoggedIn = false;
            Con_Execute(CMDS_DDAY, "net disconnect", true, false);
            break;
        case PSV_CONSOLE_TEXT: {
            uint32_t conFlags = Reader_ReadUInt32(msgReader);
            uint16_t textLen = Reader_ReadUInt16(msgReader);
            char *text = (char *) M_Malloc(textLen + 1);
            Reader_Read(msgReader, text, textLen);
            text[textLen] = 0;
            DENG_UNUSED(conFlags);
            LOG_NOTE("%s") << text;
            M_Free(text);
            break; }
        case PKT_LOGIN:
            netLoggedIn = CPP_BOOL(Reader_ReadByte(msgReader));
            break;
        case PSV_FINALE:
            Cl_Finale(msgReader);
            break;
        case PSV_FRAME2:
        case PSV_FIRST_FRAME2:
        case PSV_SOUND:
            LOGDEV_NET_WARNING("Packet type %i was discarded (client not ready)") << netBuffer.msg.type;
            break;
        default:
            if (netBuffer.msg.type >= PKT_GAME_MARKER) {
                gx.HandlePacket(netBuffer.player, netBuffer.msg.type,
                                netBuffer.msg.data, netBuffer.length);
            }
            else {
                LOG_NET_WARNING("Packet was discarded (unknown type %i)") << netBuffer.msg.type;
            }
        }
        Msg_EndRead();
    }
}
#ifdef DENG_DEBUG
/**
 * Check the state of the client on engineside. A debugging utility.
 */
static void assertPlayerIsValid(int plrNum) {
    LOG_AS("Client.assertPlayerIsValid");
    if (!isClient || !Cl_GameReady() || clientPaused) return;
    if (plrNum < 0 || plrNum >= DDMAXPLAYERS) return;
    player_t *plr = DD_Player(plrNum);
    clplayerstate_t *s = ClPlayer_State(plrNum);
    if (!s->clMobjId || !plr->publicData().mo)
        return;
    mobj_t *clmo = ClMobj_Find(s->clMobjId);
    if (!clmo) {
        LOGDEV_NET_NOTE("Player %i does not have a clmobj yet [%i]") << plrNum << s->clMobjId;
        return;
    }
    mobj_t *mo = plr->publicData().mo;
    /*
    ("Assert: client %i, clmo %i (flags 0x%x)", plrNum, clmo->thinker.id, clmo->ddFlags);
    */
    if (mo->ddFlags & DDMF_REMOTE) {
        LOGDEV_NET_NOTE("Player %i's mobj should not be remote") << plrNum;
    }
    if (clmo->ddFlags & DDMF_SOLID) {
        LOGDEV_NET_NOTE("Player %i's clmobj should not be solid (when player is alive)") << plrNum;
    }
}
#endif // DENG_DEBUG
void Cl_Ticker(timespan_t ticLength) {
    if (!isClient || !Cl_GameReady() || clientPaused)
        return;
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        if (!DD_Player(i)->publicData().inGame) continue;
        if (i != consolePlayer) {
            if (DD_Player(i)->publicData().mo) {
                Smoother_AddPos(DD_Player(i)->smoother(),
                                Cl_FrameGameTime(),
                                DD_Player(i)->publicData().mo->origin[VX],
                                DD_Player(i)->publicData().mo->origin[VY],
                                DD_Player(i)->publicData().mo->origin[VZ],
                                false);
            }
            Smoother_Advance(DD_Player(i)->smoother(), ticLength);
        }
        ClPlayer_ApplyPendingFixes(i);
        ClPlayer_UpdateOrigin(i);
#ifdef DENG_DEBUG
        assertPlayerIsValid(i);
#endif
    }
    if (App_World().hasMap()) {
        App_World().map().expireClMobjs();
    }
}
/**
 * Clients use this to establish a remote connection to the server.
 */
D_CMD(Login) {
    DENG2_UNUSED(src);
    if (!isClient)
        return false;
    Msg_Begin(PKT_LOGIN);
    if (argc == 1) {
        Writer_WriteByte(msgWriter, 0); // No password given!
    }
    else if (strlen(argv[1]) <= 255) {
        Writer_WriteByte(msgWriter, strlen(argv[1]));
        Writer_Write(msgWriter, argv[1], strlen(argv[1]));
    }
    else {
        Msg_End();
        return false;
    }
    Msg_End();
    Net_SendBuffer(0, 0);
    return true;
}
#define UNFIXED8_8(x)   (((x) << 16) / 256)
#define UNFIXED10_6(x)  (((x) << 16) / 64)
#if 0
ClMobjInfo::ClMobjInfo()
    : startMagic(CLM_MAGIC1)
    , next      (0)
    , prev      (0)
    , flags     (0)
    , time      (Timer_RealMilliseconds())
    , sound     (0)
    , volume    (0)
    , endMagic  (CLM_MAGIC2)
{}
mobj_t *ClMobjInfo::mobj() {
    DENG2_ASSERT(startMagic == CLM_MAGIC1);
    DENG2_ASSERT(endMagic == CLM_MAGIC2);
    return reinterpret_cast<mobj_t *>((char *)this + sizeof(ClMobjInfo));
}
#endif
void ClMobj_Link(mobj_t *mo) {
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mo);
    CL_ASSERT_CLMOBJ(mo);
    if ((info->flags & (CLMF_HIDDEN | CLMF_UNPREDICTABLE)) || mo->dPlayer) {
        return;
    }
    LOG_NET_XVERBOSE("ClMobj_Link: id %i, x%f Y%f, solid:%b",
                     mo->thinker.id << mo->origin[VX] << mo->origin[VY]
                     << (mo->ddFlags & DDMF_SOLID));
    Mobj_Link(mo, (mo->ddFlags & DDMF_DONTDRAW ? 0 : MLF_SECTOR) |
                  (mo->ddFlags & DDMF_SOLID ? MLF_BLOCKMAP : 0));
}
#undef ClMobj_EnableLocalActions
void ClMobj_EnableLocalActions(mobj_t *mo, dd_bool enable) {
    LOG_AS("ClMobj_EnableLocalActions");
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mo);
    if (!isClient || !info) return;
    if (enable) {
        LOGDEV_NET_VERBOSE("Enabled for clmobj %i") << mo->thinker.id;
        info->flags |= CLMF_LOCAL_ACTIONS;
    }
    else {
        LOGDEV_NET_VERBOSE("Disabled for clmobj %i") << mo->thinker.id;
        info->flags &= ~CLMF_LOCAL_ACTIONS;
    }
}
#undef ClMobj_LocalActionsEnabled
dd_bool ClMobj_LocalActionsEnabled(mobj_t *mo) {
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mo);
    if (!isClient || !info) return true;
    return (info->flags & CLMF_LOCAL_ACTIONS) != 0;
}
void ClMobj_SetState(mobj_t *mo, int stnum) {
    if (stnum < 0) return;
    do {
        Mobj_SetState(mo, stnum);
        stnum = runtimeDefs.states[stnum].nextState;
    } while (!mo->tics && stnum > 0);
}
void Cl_UpdateRealPlayerMobj(mobj_t *localMobj, mobj_t *remoteClientMobj,
                             int flags, dd_bool onFloor) {
    if (!localMobj || !remoteClientMobj) {
        LOGDEV_MAP_VERBOSE("Cl_UpdateRealPlayerMobj: mo=%p clmo=%p") << localMobj << remoteClientMobj;
        return;
    }
    localMobj->radius = Mobj_Radius(*remoteClientMobj);
    if (flags & MDF_MOM_X) localMobj->mom[MX] = remoteClientMobj->mom[MX];
    if (flags & MDF_MOM_Y) localMobj->mom[MY] = remoteClientMobj->mom[MY];
    if (flags & MDF_MOM_Z) localMobj->mom[MZ] = remoteClientMobj->mom[MZ];
    if (flags & MDF_ANGLE) {
        localMobj->angle = remoteClientMobj->angle;
        LOGDEV_MAP_XVERBOSE_DEBUGONLY("Cl_UpdateRealPlayerMobj: localMobj=%p angle=%x",
                                     localMobj << localMobj->angle);
    }
    localMobj->sprite = remoteClientMobj->sprite;
    localMobj->frame = remoteClientMobj->frame;
    localMobj->tics = remoteClientMobj->tics;
    localMobj->state = remoteClientMobj->state;
#define DDMF_KEEP_MASK (DDMF_REMOTE | DDMF_SOLID)
    localMobj->ddFlags = (localMobj->ddFlags & DDMF_KEEP_MASK) | (remoteClientMobj->ddFlags & ~DDMF_KEEP_MASK);
    if (flags & MDF_FLAGS) {
        localMobj->flags = (localMobj->flags & ~0x1c000000) |
                           (remoteClientMobj->flags & 0x1c000000); // color translation flags (MF_TRANSLATION)
    }
    localMobj->height = remoteClientMobj->height;
    localMobj->selector &= ~DDMOBJ_SELECTOR_MASK;
    localMobj->selector |= remoteClientMobj->selector & DDMOBJ_SELECTOR_MASK;
    localMobj->visAngle = remoteClientMobj->angle >> 16;
    if (flags & (MDF_ORIGIN_X | MDF_ORIGIN_Y)) {
        /*
        P_MobjUnlink(localMobj);
        localMobj->pos[VX] = remoteClientMobj->pos[VX];
        localMobj->pos[VY] = remoteClientMobj->pos[VY];
        P_MobjLink(localMobj, MLF_SECTOR | MLF_BLOCKMAP);
        */
        if (gx.MobjTryMoveXYZ) {
            if (gx.MobjTryMoveXYZ(localMobj,
                                 remoteClientMobj->origin[VX],
                                 remoteClientMobj->origin[VY],
                                 (flags & MDF_ORIGIN_Z)? remoteClientMobj->origin[VZ] : localMobj->origin[VZ])) {
                if ((flags & MDF_ORIGIN_Z) && onFloor) {
                    localMobj->origin[VZ] = remoteClientMobj->origin[VZ] = localMobj->floorZ;
                }
            }
        }
    }
    if (flags & MDF_ORIGIN_Z) {
        if (!onFloor) {
            localMobj->floorZ = remoteClientMobj->floorZ;
        }
        localMobj->ceilingZ = remoteClientMobj->ceilingZ;
        localMobj->origin[VZ] = remoteClientMobj->origin[VZ];
        if (localMobj->origin[VZ] < localMobj->floorZ)
            localMobj->origin[VZ] = localMobj->floorZ;
    }
}
dd_bool Cl_IsClientMobj(mobj_t const *mob) {
    DENG2_ASSERT(mob);
    if (ClientMobjThinkerData *data = THINKER_DATA_MAYBE(mob->thinker, ClientMobjThinkerData)) {
        return data->hasRemoteSync();
    }
    return false;
}
#undef ClMobj_IsValid
dd_bool ClMobj_IsValid(mobj_t *mob) {
    if (!Cl_IsClientMobj(mob)) return true;
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mob);
    if (info->flags & (CLMF_HIDDEN | CLMF_UNPREDICTABLE)) {
        return false;
    }
    if (!mob->info) {
        return false;
    }
    return true;
}
ClientMobjThinkerData::RemoteSync *ClMobj_GetInfo(mobj_t *mob) {
    if (!mob) return nullptr;
    if (ClientMobjThinkerData *data = THINKER_DATA_MAYBE(mob->thinker, ClientMobjThinkerData)) {
        if (!data->hasRemoteSync()) return nullptr;
        return &data->remoteSync();
    }
    return nullptr;
}
dd_bool ClMobj_Reveal(mobj_t *mob) {
    LOG_AS("ClMobj_Reveal");
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mob);
    CL_ASSERT_CLMOBJ(mob);
    if (mob->dPlayer != &DD_Player(consolePlayer)->publicData() &&
       (!(info->flags & CLMF_KNOWN_X) ||
        !(info->flags & CLMF_KNOWN_Y) ||
        !(info->flags & CLMF_KNOWN_STATE))) {
        return false;
    }
    LOGDEV_MAP_XVERBOSE("clmobj %i 'Hidden' status lifted (z=%f)", mob->thinker.id << mob->origin[VZ]);
    info->flags &= ~CLMF_HIDDEN;
    if (info->flags & CLMF_SOUND) {
        info->flags &= ~CLMF_SOUND;
        S_StartSoundAtVolume(info->sound, mob, info->volume);
    }
    LOGDEV_MAP_XVERBOSE("Revealing id %i, state %p (%i)",
                        mob->thinker.id << mob->state <<
                        ::runtimeDefs.states.indexOf(mob->state));
    return true;
}
/**
 * Determines whether @a mo happens to reside inside one of the local players.
 * In normal gameplay solid mobjs cannot enter inside each other.
 *
 * @param mo  Client mobj (must be solid).
 */
static dd_bool ClMobj_IsStuckInsideLocalPlayer(mobj_t *mo) {
    if (!(mo->ddFlags & DDMF_SOLID) || mo->dPlayer)
        return false;
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        if (!DD_Player(i)->publicData().inGame) continue;
        if (P_ConsoleToLocal(i) < 0) continue; // Not a local player.
        mobj_t *plmo = DD_Player(i)->publicData().mo;
        if (!plmo) continue;
        float blockRadius = Mobj_Radius(*mo) + Mobj_Radius(*plmo);
        if (fabs(mo->origin[VX] - plmo->origin[VX]) >= blockRadius ||
           fabs(mo->origin[VY] - plmo->origin[VY]) >= blockRadius)
            continue; // Too far.
        if (mo->origin[VZ] > plmo->origin[VZ] + plmo->height)
            continue; // Above.
        if (plmo->origin[VZ] > mo->origin[VZ] + mo->height)
            continue; // Under.
        return true;
    }
    return false; // Not stuck.
}
void ClMobj_ReadDelta() {
    world::Map &map = App_World().map();
    thid_t const id = Reader_ReadUInt16(msgReader); // Read the ID.
    dint const df   = Reader_ReadUInt16(msgReader); // Flags.
    byte moreFlags = 0, fastMom = false;
    if (df & MDF_MORE_FLAGS) {
        moreFlags = Reader_ReadByte(msgReader);
        if (moreFlags & MDFE_FAST_MOM)
            fastMom = true;
    }
    LOGDEV_NET_XVERBOSE("Reading mobj delta for %i (df:0x%x edf:0x%x)",
                        id << df << moreFlags);
    mobj_t *mo = map.clMobjFor(id);
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mo);
    bool needsLinking = false, justCreated = false;
    if (!mo) {
        LOG_NET_XVERBOSE("Creating new clmobj %i (hidden)", id);
        mo = map.clMobjFor(id, true /* create it now */);
        info = ClMobj_GetInfo(mo);
        justCreated = true;
        needsLinking = true;
        info->flags |= CLMF_HIDDEN;
    }
    if (!(info->flags & CLMF_NULLED)) {
        info->flags &= ~CLMF_UNPREDICTABLE;
        info->time = Timer_RealMilliseconds();
    }
    mobj_t *d = mo;
    /*if (d->dPlayer && d->dPlayer == DD_Player(consolePlayer))
    {
        cmo->flags |= CLMF_KNOWN;
    }*/
    if (df & (MDF_ORIGIN_X | MDF_ORIGIN_Y | MDF_ORIGIN_Z | MDF_FLAGS) &&
       !justCreated && !d->dPlayer) {
        needsLinking = true;
        Mobj_Unlink(mo);
    }
    MobjThinker oldState(*mo);
    bool onFloor = false;
    if (df & MDF_ORIGIN_X) {
        d->origin[VX] = FIX2FLT((Reader_ReadInt16(msgReader) << FRACBITS) | (Reader_ReadByte(msgReader) << 8));
        if (info)
            info->flags |= CLMF_KNOWN_X;
    }
    if (df & MDF_ORIGIN_Y) {
        d->origin[VY] = FIX2FLT((Reader_ReadInt16(msgReader) << FRACBITS) | (Reader_ReadByte(msgReader) << 8));
        if (info)
            info->flags |= CLMF_KNOWN_Y;
    }
    if (df & MDF_ORIGIN_Z) {
        if (!(moreFlags & MDFE_Z_FLOOR)) {
            d->origin[VZ] = FIX2FLT((Reader_ReadInt16(msgReader) << FRACBITS) | (Reader_ReadByte(msgReader) << 8));
            if (info) {
                info->flags |= CLMF_KNOWN_Z;
                info->flags &= ~(CLMF_STICK_FLOOR | CLMF_STICK_CEILING);
            }
            d->floorZ = Reader_ReadFloat(msgReader);
        }
        else {
            onFloor = true;
            Reader_ReadInt16(msgReader);
            Reader_ReadByte(msgReader);
            Reader_ReadFloat(msgReader);
            info->flags |= CLMF_KNOWN_Z;
        }
        d->ceilingZ = Reader_ReadFloat(msgReader);
    }
    if (df & MDF_MOM_X) {
        short mom = Reader_ReadInt16(msgReader);
        d->mom[MX] = FIX2FLT(fastMom? UNFIXED10_6(mom) : UNFIXED8_8(mom));
    }
    if (df & MDF_MOM_Y) {
        short mom = Reader_ReadInt16(msgReader);
        d->mom[MY] = FIX2FLT(fastMom ? UNFIXED10_6(mom) : UNFIXED8_8(mom));
    }
    if (df & MDF_MOM_Z) {
        short mom = Reader_ReadInt16(msgReader);
        d->mom[MZ] = FIX2FLT(fastMom ? UNFIXED10_6(mom) : UNFIXED8_8(mom));
    }
    if (df & MDF_ANGLE)
        d->angle = Reader_ReadInt16(msgReader) << 16;
    if (df & MDF_SELECTOR)
        d->selector = Reader_ReadPackedUInt16(msgReader);
    if (df & MDF_SELSPEC)
        d->selector |= Reader_ReadByte(msgReader) << 24;
    if (df & MDF_STATE) {
        int stateIdx = Reader_ReadPackedUInt16(msgReader);
        stateIdx = Cl_LocalMobjState(stateIdx);
        if (!(info->flags & CLMF_LOCAL_ACTIONS)) {
            ClMobj_SetState(d, stateIdx);
            info->flags |= CLMF_KNOWN_STATE;
        }
    }
    if (df & MDF_FLAGS) {
        d->ddFlags &= ~DDMF_PACK_MASK;
        d->ddFlags |= DDMF_REMOTE | (Reader_ReadUInt32(msgReader) & DDMF_PACK_MASK);
        d->flags  = Reader_ReadUInt32(msgReader);
        d->flags2 = Reader_ReadUInt32(msgReader);
        d->flags3 = Reader_ReadUInt32(msgReader);
    }
    if (df & MDF_HEALTH)
        d->health = Reader_ReadInt32(msgReader);
    if (df & MDF_RADIUS)
        d->radius = Reader_ReadFloat(msgReader);
    if (df & MDF_HEIGHT)
        d->height = Reader_ReadFloat(msgReader);
    if (df & MDF_FLOORCLIP)
        d->floorClip = Reader_ReadFloat(msgReader);
    if (moreFlags & MDFE_TRANSLUCENCY)
        d->translucency = Reader_ReadByte(msgReader);
    if (moreFlags & MDFE_FADETARGET)
        d->visTarget = ((short)Reader_ReadByte(msgReader)) - 1;
    if (moreFlags & MDFE_TYPE) {
        d->type = Cl_LocalMobjType(Reader_ReadInt32(msgReader));
        d->info = &runtimeDefs.mobjInfo[d->type];
    }
    if (info->flags & CLMF_HIDDEN) {
        if (ClMobj_Reveal(d)) {
            needsLinking = true;
        }
    }
    if (!d->dPlayer && onFloor && gx.MobjCheckPositionXYZ) {
        if (coord_t *floorZ = (coord_t *) gx.GetPointer(DD_TM_FLOOR_Z)) {
            gx.MobjCheckPositionXYZ(d, d->origin[VX], d->origin[VY], DDMAXFLOAT);
            d->origin[VZ] = d->floorZ = *floorZ;
        }
    }
    if (!(info->flags & (CLMF_HIDDEN | CLMF_NULLED))) {
        if (needsLinking && !d->dPlayer) {
            ClMobj_Link(mo);
            if (ClMobj_IsStuckInsideLocalPlayer(mo)) {
                Mobj_Unlink(mo);
                V3d_Copy(mo->origin, oldState->origin);
                mo->floorZ = oldState->floorZ;
                mo->ceilingZ = oldState->ceilingZ;
                ClMobj_Link(mo);
            }
        }
        if (d->dPlayer) {
            LOG_NET_XVERBOSE("Updating player %i local mobj with new clmobj state {%f, %f, %f}",
                             P_GetDDPlayerIdx(d->dPlayer) <<
                             d->origin[VX] << d->origin[VY] << d->origin[VZ]);
            Cl_UpdateRealPlayerMobj(d->dPlayer->mo, d, df, onFloor);
        }
    }
}
void ClMobj_ReadNullDelta() {
    LOG_AS("ClMobj_ReadNullDelta");
    Map &map = App_World().map();
    thid_t id = Reader_ReadUInt16(msgReader);
    LOGDEV_NET_XVERBOSE("Null %i", id);
    mobj_t *mo = map.clMobjFor(id);
    if (!mo) {
        LOGDEV_NET_MSG("Request to remove id %i that doesn't exist here") << id;
        return;
    }
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mo);
    if (!mo->dPlayer) {
        Mobj_Unlink(mo);
    }
    else {
        LOGDEV_NET_MSG("clmobj of player %i deleted") << P_GetDDPlayerIdx(mo->dPlayer);
        ClPlayer_State(P_GetDDPlayerIdx(mo->dPlayer))->clMobjId = 0;
    }
    info->time = Timer_RealMilliseconds();
    info->flags |= CLMF_UNPREDICTABLE | CLMF_NULLED;
}
#undef ClMobj_Find
mobj_t *ClMobj_Find(thid_t id) {
    if (!App_World().hasMap()) return nullptr;
    return App_World().map().clMobjFor(id);
}
#undef ClPlayer_ClMobj
DENG_EXTERN_C mobj_t *ClPlayer_ClMobj(int plrNum);
DENG_DECLARE_API(Client) = {
    { DE_API_CLIENT },
    ClMobj_Find,
    ClMobj_EnableLocalActions,
    ClMobj_LocalActionsEnabled,
    ClMobj_IsValid,
    ClPlayer_ClMobj
};
#define TOP_PSPY            (32)
#define BOTTOM_PSPY         (128)
float pspMoveSpeed = 6;
float cplrThrustMul = 1;
static float pspY;
static float cpMom[3][LOCALCAM_WRITE_TICS];
void Cl_InitPlayers() {
    DoomsdayApp::players().forAll([] (Player &plr) {
        zap(plr.as<ClientPlayer>().clPlayerState());
        return LoopContinue;
    });
    pspY = 0;
    de::zap(cpMom);
}
clplayerstate_t *ClPlayer_State(int plrNum) {
    return &DD_Player(plrNum)->clPlayerState();
}
#undef ClPlayer_ClMobj
DENG_EXTERN_C struct mobj_s *ClPlayer_ClMobj(int plrNum) {
    if (plrNum < 0 || plrNum >= DDMAXPLAYERS) return 0;
    return ClMobj_Find(ClPlayer_State(plrNum)->clMobjId);
}
void ClPlayer_UpdateOrigin(int plrNum) {
    DENG2_ASSERT(plrNum >= 0 && plrNum < DDMAXPLAYERS);
    player_t *plr = DD_Player(plrNum);
    clplayerstate_t *s = ClPlayer_State(plrNum);
    if (!s->clMobjId || !plr->publicData().mo)
        return;                 // Must have a mobj!
    mobj_t *remoteClientMobj = ClMobj_Find(s->clMobjId);
    mobj_t *localMobj = plr->publicData().mo;
    remoteClientMobj->ddFlags &= ~DDMF_SOLID;
    remoteClientMobj->angle = localMobj->angle;
    std::memcpy(remoteClientMobj->origin, localMobj->origin, sizeof(localMobj->origin));
    Mobj_Link(remoteClientMobj, 0); // Update bspLeaf pointer.
    remoteClientMobj->floorZ = localMobj->floorZ;
    remoteClientMobj->ceilingZ = localMobj->ceilingZ;
    remoteClientMobj->mom[MX] = localMobj->mom[MX];
    remoteClientMobj->mom[MY] = localMobj->mom[MY];
    remoteClientMobj->mom[MZ] = localMobj->mom[MZ];
}
void ClPlayer_ApplyPendingFixes(int plrNum) {
    LOG_AS("ClPlayer_ApplyPendingFixes");
    clplayerstate_t *state = ClPlayer_State(plrNum);
    player_t *plr = DD_Player(plrNum);
    mobj_t *clmo = ClPlayer_ClMobj(plrNum);
    ddplayer_t *ddpl = &plr->publicData();
    mobj_t *mo = ddpl->mo;
    bool sendAck = false;
    if (!mo || !clmo) return;
    if (clmo->thinker.id != state->pendingFixTargetClMobjId)
        return;
    DENG_ASSERT(clmo->thinker.id == state->clMobjId);
    if (state->pendingFixes & DDPF_FIXANGLES) {
        state->pendingFixes &= ~DDPF_FIXANGLES;
        ddpl->fixAcked.angles = ddpl->fixCounter.angles;
        sendAck = true;
        LOGDEV_NET_MSG("Applying angle %x to mobj %p and clmo %i")
                << state->pendingAngleFix << mo << clmo->thinker.id;
        clmo->angle = mo->angle = state->pendingAngleFix;
        ddpl->lookDir = state->pendingLookDirFix;
    }
    if (state->pendingFixes & DDPF_FIXORIGIN) {
        state->pendingFixes &= ~DDPF_FIXORIGIN;
        ddpl->fixAcked.origin = ddpl->fixCounter.origin;
        sendAck = true;
        LOGDEV_NET_MSG("Applying pos %s to mobj %p and clmo %i")
                << Vector3d(state->pendingOriginFix).asText()
                << mo << clmo->thinker.id;
        Mobj_SetOrigin(mo, state->pendingOriginFix[VX], state->pendingOriginFix[VY], state->pendingOriginFix[VZ]);
        mo->reactionTime = 18;
        ddpl->flags &= ~DDPF_UNDEFINED_ORIGIN;
        Smoother_Clear(DD_Player(plrNum)->smoother());
        ClPlayer_UpdateOrigin(plrNum);
    }
    if (state->pendingFixes & DDPF_FIXMOM) {
        state->pendingFixes &= ~DDPF_FIXMOM;
        ddpl->fixAcked.mom = ddpl->fixCounter.mom;
        sendAck = true;
        LOGDEV_NET_MSG("Applying mom %s to mobj %p and clmo %i")
                << Vector3d(state->pendingMomFix).asText()
                << mo << clmo->thinker.id;
        mo->mom[MX] = clmo->mom[VX] = state->pendingMomFix[VX];
        mo->mom[MY] = clmo->mom[VY] = state->pendingMomFix[VY];
        mo->mom[MZ] = clmo->mom[VZ] = state->pendingMomFix[VZ];
    }
    if (sendAck && plrNum == consolePlayer) {
        Msg_Begin(PCL_ACK_PLAYER_FIX);
        Writer_WriteInt32(msgWriter, ddpl->fixAcked.angles);
        Writer_WriteInt32(msgWriter, ddpl->fixAcked.origin);
        Writer_WriteInt32(msgWriter, ddpl->fixAcked.mom);
        Msg_End();
        Net_SendBuffer(0, 0);
    }
}
void ClPlayer_HandleFix() {
    LOG_AS("Cl_HandlePlayerFix");
    int plrNum = Reader_ReadByte(msgReader);
    player_t *plr = DD_Player(plrNum);
    ddplayer_t *ddpl = &plr->publicData();
    clplayerstate_t *state = ClPlayer_State(plrNum);
    int fixes = Reader_ReadUInt32(msgReader);
    state->pendingFixTargetClMobjId = Reader_ReadUInt16(msgReader);
    LOGDEV_NET_MSG("Fixing player %i") << plrNum;
    if (fixes & 1) // fix angles?
    {
        ddpl->fixCounter.angles = Reader_ReadInt32(msgReader);
        state->pendingAngleFix = Reader_ReadUInt32(msgReader);
        state->pendingLookDirFix = Reader_ReadFloat(msgReader);
        state->pendingFixes |= DDPF_FIXANGLES;
        LOGDEV_NET_VERBOSE("Pending fix angles %i: angle=%x, lookdir=%f")
                << ddpl->fixAcked.angles << state->pendingAngleFix << state->pendingLookDirFix;
    }
    if (fixes & 2) // fix pos?
    {
        ddpl->fixCounter.origin = Reader_ReadInt32(msgReader);
        state->pendingOriginFix[VX] = Reader_ReadFloat(msgReader);
        state->pendingOriginFix[VY] = Reader_ReadFloat(msgReader);
        state->pendingOriginFix[VZ] = Reader_ReadFloat(msgReader);
        state->pendingFixes |= DDPF_FIXORIGIN;
        LOGDEV_NET_VERBOSE("Pending fix pos %i: %s")
                << ddpl->fixAcked.origin << Vector3d(state->pendingOriginFix).asText();
    }
    if (fixes & 4) // fix momentum?
    {
        ddpl->fixCounter.mom = Reader_ReadInt32(msgReader);
        state->pendingMomFix[VX] = Reader_ReadFloat(msgReader);
        state->pendingMomFix[VY] = Reader_ReadFloat(msgReader);
        state->pendingMomFix[VZ] = Reader_ReadFloat(msgReader);
        state->pendingFixes |= DDPF_FIXMOM;
        LOGDEV_NET_VERBOSE("Pending fix momentum %i: %s")
                << ddpl->fixAcked.mom << Vector3d(state->pendingMomFix).asText();
    }
    ClPlayer_ApplyPendingFixes(plrNum);
}
void ClPlayer_MoveLocal(coord_t dx, coord_t dy, coord_t z, bool onground) {
    player_t *plr = DD_Player(consolePlayer);
    ddplayer_t *ddpl = &plr->publicData();
    mobj_t *mo = ddpl->mo;
    if (!mo) return;
    cpMom[MX][SECONDS_TO_TICKS(gameTime) % LOCALCAM_WRITE_TICS] = dx;
    cpMom[MY][SECONDS_TO_TICKS(gameTime) % LOCALCAM_WRITE_TICS] = dy;
    Vector2d mom;
    for (int i = 0; i < LOCALCAM_WRITE_TICS; ++i) {
        mom += Vector2d(cpMom[MX][i], cpMom[MY][i]);
    }
    mom /= LOCALCAM_WRITE_TICS;
    mo->mom[MX] = mom.x;
    mo->mom[MY] = mom.y;
    if (dx != 0 || dy != 0) {
        Mobj_Unlink(mo);
        mo->origin[VX] += dx;
        mo->origin[VY] += dy;
        Mobj_Link(mo, MLF_SECTOR | MLF_BLOCKMAP);
    }
    mo->_bspLeaf = &Mobj_Map(*mo).bspLeafAt_FixedPrecision(Mobj_Origin(*mo));
    mo->floorZ   = Mobj_Sector(mo)->floor().height();
    mo->ceilingZ = Mobj_Sector(mo)->ceiling().height();
    if (onground) {
        mo->origin[VZ] = z - 1;
    }
    else {
        mo->origin[VZ] = z;
    }
    ClPlayer_UpdateOrigin(consolePlayer);
}
void ClPlayer_ReadDelta() {
    LOG_AS("ClPlayer_ReadDelta2");
    world::Map &map = App_World().map();
    dint df = 0;
    ushort num;
    num = Reader_ReadByte(msgReader);
    df = (num & 0xf0) << 8;
    df |= Reader_ReadByte(msgReader); // Second byte is just flags.
    num &= 0xf; // Clear the upper bits of the number.
    clplayerstate_t *s = ClPlayer_State(num);
    ddplayer_t *ddpl = &DD_Player(num)->publicData();
    if (df & PDF_MOBJ) {
        mobj_t *old  = map.clMobjFor(s->clMobjId);
        ushort newId = Reader_ReadUInt16(msgReader);
        if (newId != s->clMobjId) {
            mobj_t *clmo = 0;
            bool justCreated = false;
            s->clMobjId = newId;
            clmo = map.clMobjFor(s->clMobjId);
            if (!clmo) {
                LOGDEV_NET_NOTE("Player %i's new clmobj is %i, but we haven't received it yet")
                        << num << newId;
                clmo = map.clMobjFor(s->clMobjId, true/*create*/);
                /*
                if (num == consolePlayer) {
                }*/
                justCreated = true;
            }
            else {
                Mobj_Unlink(clmo);
            }
            clmo->dPlayer = ddpl;
            if (old) {
                old->dPlayer = NULL;
                ClMobj_Link(old);
            }
            if (!justCreated) // && num != consolePlayer)
            {
                LOGDEV_NET_XVERBOSE("Copying clmo %i state to real player %i mobj %p",
                                    newId << num << ddpl->mo);
                Cl_UpdateRealPlayerMobj(ddpl->mo, clmo, 0xffffffff, true);
            }
            LOGDEV_NET_VERBOSE("Player %i: mobj=%i old=%p x=%.1f y=%.1f z=%.1f Fz=%.1f Cz=%.1f")
                    << num << s->clMobjId << old
                    << clmo->origin[VX] << clmo->origin[VY] << clmo->origin[VZ]
                    << clmo->floorZ << clmo->ceilingZ;
            LOGDEV_NET_VERBOSE("Player %i using mobj id %i") << num << s->clMobjId;
        }
    }
    if (df & PDF_FORWARDMOVE) {
        s->forwardMove = (char) Reader_ReadByte(msgReader) * 2048;
    }
    if (df & PDF_SIDEMOVE) {
        s->sideMove = (char) Reader_ReadByte(msgReader) * 2048;
    }
    if (df & PDF_ANGLE) {
        DENG_UNUSED(Reader_ReadByte(msgReader));
    }
    if (df & PDF_TURNDELTA) {
        s->turnDelta = ((char) Reader_ReadByte(msgReader) << 24) / 16;
    }
    if (df & PDF_FRICTION) {
        s->friction = Reader_ReadByte(msgReader) << 8;
    }
    if (df & PDF_EXTRALIGHT) {
        int val = Reader_ReadByte(msgReader);
        ddpl->fixedColorMap = val & 7;
        ddpl->extraLight    = val & 0xf8;
    }
    if (df & PDF_FILTER) {
        uint filter = Reader_ReadUInt32(msgReader);
        ddpl->filterColor[CR] = (filter & 0xff) / 255.f;
        ddpl->filterColor[CG] = ((filter >> 8) & 0xff) / 255.f;
        ddpl->filterColor[CB] = ((filter >> 16) & 0xff) / 255.f;
        ddpl->filterColor[CA] = ((filter >> 24) & 0xff) / 255.f;
        if (ddpl->filterColor[CA] > 0) {
            ddpl->flags |= DDPF_REMOTE_VIEW_FILTER;
        }
        else {
            ddpl->flags &= ~DDPF_REMOTE_VIEW_FILTER;
        }
        LOG_NET_XVERBOSE("View filter color set remotely to %s",
                         Vector4f(ddpl->filterColor).asText());
    }
    if (df & PDF_PSPRITES) {
        for (int i = 0; i < 2; ++i) {
            int psdf = Reader_ReadByte(msgReader);
            ddpsprite_t *psp = ddpl->pSprites + i;
            if (psdf & PSDF_STATEPTR) {
                int idx = Reader_ReadPackedUInt16(msgReader);
                if (!idx) {
                    psp->statePtr = 0;
                }
                else if (idx < runtimeDefs.states.size()) {
                    psp->statePtr = &runtimeDefs.states[idx - 1];
                    psp->tics = psp->statePtr->tics;
                }
            }
            /*if (psdf & PSDF_LIGHT)
            {
                psp->light = Reader_ReadByte(msgReader) / 255.0f;
            }*/
            if (psdf & PSDF_ALPHA) {
                psp->alpha = Reader_ReadByte(msgReader) / 255.0f;
            }
            if (psdf & PSDF_STATE) {
                psp->state = Reader_ReadByte(msgReader);
            }
            if (psdf & PSDF_OFFSET) {
                psp->offset[VX] = (char) Reader_ReadByte(msgReader) * 2;
                psp->offset[VY] = (char) Reader_ReadByte(msgReader) * 2;
            }
        }
    }
}
mobj_t *ClPlayer_LocalGameMobj(int plrNum) {
    return DD_Player(plrNum)->publicData().mo;
}
bool ClPlayer_IsFreeToMove(int plrNum) {
    mobj_t *mo = ClPlayer_LocalGameMobj(plrNum);
    if (!mo) return false;
    return (mo->origin[VZ] >= mo->floorZ &&
            mo->origin[VZ] + mo->height <= mo->ceilingZ);
}
void Cl_ReadSoundDelta(deltatype_t type) {
    LOG_AS("Cl_ReadSoundDelta");
    world::Map &map = App_World().map();
    dint sound = 0, soundFlags = 0;
    mobj_t *cmo = 0;
    thid_t mobjId = 0;
    Sector *sector = 0;
    Polyobj *poly = 0;
    LineSide *side = 0;
    mobj_t *emitter = 0;
    duint16 const deltaId = Reader_ReadUInt16(::msgReader);
    byte const flags      = Reader_ReadByte(::msgReader);
    bool skip = false;
    if (type == DT_SOUND) {
        sound = deltaId;
    }
    else if (type == DT_MOBJ_SOUND) // Mobj as emitter
    {
        mobjId = deltaId;
        if ((cmo = ClMobj_Find(mobjId)) != nullptr) {
            ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(cmo);
            if (info->flags & CLMF_HIDDEN) {
                cmo = nullptr;
                LOG_NET_VERBOSE("Can't find sound emitter ") << mobjId;
            }
            else {
                emitter = cmo;
            }
        }
    }
    else if (type == DT_SECTOR_SOUND) // Plane as emitter
    {
        dint index = deltaId;
        if (!(sector = map.sectorPtr(index))) {
            LOG_NET_WARNING("Received sound delta has invalid sector index %i") << index;
            skip = true;
        }
    }
    else if (type == DT_SIDE_SOUND) // Side section as emitter
    {
        dint index = deltaId;
        if (!(side = map.sidePtr(index))) {
            LOG_NET_WARNING("Received sound delta has invalid side index %i") << index;
            skip = true;
        }
    }
    else // DT_POLY_SOUND
    {
        dint index = deltaId;
        LOG_NET_XVERBOSE("DT_POLY_SOUND: poly=%d", index);
        if (!(emitter = (mobj_t *) (poly = map.polyobjPtr(index)))) {
            LOG_NET_WARNING("Received sound delta has invalid polyobj index %i") << index;
            skip = true;
        }
    }
    if (type != DT_SOUND) {
        sound = Reader_ReadUInt16(::msgReader);
    }
    if (type == DT_SECTOR_SOUND && !skip) {
        if (flags & SNDDF_PLANE_FLOOR) {
            emitter = (mobj_t *) &sector->floor().soundEmitter();
        }
        else if (flags & SNDDF_PLANE_CEILING) {
            emitter = (mobj_t *) &sector->ceiling().soundEmitter();
        }
        else {
            emitter = (mobj_t *) &sector->soundEmitter();
        }
    }
    if (type == DT_SIDE_SOUND && !skip) {
        if (flags & SNDDF_SIDE_MIDDLE)
            emitter = (mobj_t *) &side->middleSoundEmitter();
        else if (flags & SNDDF_SIDE_TOP)
            emitter = (mobj_t *) &side->topSoundEmitter();
        else if (flags & SNDDF_SIDE_BOTTOM)
            emitter = (mobj_t *) &side->bottomSoundEmitter();
    }
    dfloat volume = 1;
    if (flags & SNDDF_VOLUME) {
        byte b = Reader_ReadByte(::msgReader);
        if (b == 255) {
            soundFlags |= DDSF_NO_ATTENUATION;
        }
        else {
            volume = b / 127.0f;
        }
    }
    if (flags & SNDDF_REPEAT) {
        soundFlags |= DDSF_REPEAT;
    }
    if (skip) return;
    if (volume > 0 && sound > 0) {
        if (type == DT_MOBJ_SOUND && !cmo) {
            cmo = map.clMobjFor(mobjId, true/*create*/);
            ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(cmo);
            info->flags |= CLMF_HIDDEN | CLMF_SOUND;
            info->sound  = sound;
            info->volume = volume;
            return;
        }
        if (type != DT_SOUND && !emitter) {
            LOG_NET_VERBOSE("Cl_ReadSoundDelta2(%i): Insufficient data, snd=%i") << type << sound;
            return;
        }
        if (cmo && cmo->thinker.id == ClPlayer_State(consolePlayer)->clMobjId) {
            emitter = DD_Player(consolePlayer)->publicData().mo;
        }
        if (type != DT_SOUND && emitter) {
            S_StopSound(0, emitter);
        }
        S_LocalSoundAtVolume(sound | soundFlags, emitter, volume);
    }
    else if (sound >= 0) {
        if (emitter) {
            S_StopSound(sound, emitter);
        }
    }
}
void Cl_Sound() {
    LOG_AS("Cl_Sound");
    world::Map &map = App_World().map();
    byte const flags = Reader_ReadByte(::msgReader);
    dint sound;
    if (flags & SNDF_SHORT_SOUND_ID) {
        sound = Reader_ReadUInt16(::msgReader);
    }
    else {
        sound = Reader_ReadByte(::msgReader);
    }
    if (sound < 1 || sound >= DED_Definitions()->sounds.size()) {
        LOGDEV_NET_WARNING("Invalid sound ID %i") << sound;
        return;
    }
    LOGDEV_NET_XVERBOSE("id %i", sound);
    dint volume = 127;
    if (flags & SNDF_VOLUME) {
        volume = Reader_ReadByte(::msgReader);
        if (volume > 127) {
            volume = 127;
            sound |= DDSF_NO_ATTENUATION;
        }
    }
    if (flags & SNDF_ID) {
        thid_t sourceId = Reader_ReadUInt16(::msgReader);
        if (mobj_t *cmob = ClMobj_Find(sourceId)) {
            S_LocalSoundAtVolume(sound, cmob, volume / 127.0f);
        }
    }
    else if (flags & SNDF_SECTOR) {
        auto num = dint( Reader_ReadPackedUInt16(::msgReader) );
        if (num >= map.sectorCount()) {
            LOG_NET_WARNING("Invalid sector number %i") << num;
            return;
        }
        auto *mob = (mobj_t *) &map.sector(num).soundEmitter();
        S_LocalSoundAtVolume(sound, mob, volume / 127.0f);
    }
    else if (flags & SNDF_ORIGIN) {
        coord_t pos[3];
        pos[0] = Reader_ReadInt16(::msgReader);
        pos[1] = Reader_ReadInt16(::msgReader);
        pos[2] = Reader_ReadInt16(::msgReader);
        S_LocalSoundAtVolumeFrom(sound, nullptr, pos, volume / 127.0f);
    }
    else if (flags & SNDF_PLAYER) {
        dint const player = (flags & 0xf0) >> 4;
        DENG2_ASSERT(player >= 0 && player < DDMAXPLAYERS);
        S_LocalSoundAtVolume(sound, DD_Player(player)->publicData().mo, volume / 127.0f);
    }
    else {
        LOGDEV_NET_VERBOSE("Unspecified origin for sound %i") << sound;
        S_LocalSoundAtVolume(sound, nullptr, volume / 127.0f);
    }
}
static world::MaterialArchive *serverMaterials;
typedef QVector<int> IndexTransTable;
static IndexTransTable xlatMobjType;
static IndexTransTable xlatMobjState;
void Cl_InitTransTables() {
    serverMaterials = 0;
    xlatMobjType.clear();
    xlatMobjState.clear();
}
void Cl_ResetTransTables() {
    if (serverMaterials) {
        delete serverMaterials;
        serverMaterials = 0;
    }
    xlatMobjType.clear();
    xlatMobjState.clear();
}
void Cl_ReadServerMaterials() {
    LOG_AS("Cl_ReadServerMaterials");
    if (!serverMaterials) {
        serverMaterials = new world::MaterialArchive(false /*no segment check*/, false /*empty*/);
    }
    serverMaterials->read(*msgReader, -1 /*no forced version*/);
    LOGDEV_NET_VERBOSE("Received %i materials") << serverMaterials->count();
}
void Cl_ReadServerMobjTypeIDs() {
    LOG_AS("Cl_ReadServerMobjTypeIDs");
    StringArray *ar = StringArray_New();
    StringArray_Read(ar, msgReader);
    LOGDEV_NET_VERBOSE("Received %i mobj type IDs") << StringArray_Size(ar);
    xlatMobjType.resize(StringArray_Size(ar));
    for (dint i = 0; i < StringArray_Size(ar); ++i) {
        xlatMobjType[i] = DED_Definitions()->getMobjNum(StringArray_At(ar, i));
        if (xlatMobjType[i] < 0) {
            LOG_NET_WARNING("Could not find '%s' in local thing definitions")
                    << StringArray_At(ar, i);
        }
    }
    StringArray_Delete(ar);
}
void Cl_ReadServerMobjStateIDs() {
    LOG_AS("Cl_ReadServerMobjStateIDs");
    StringArray *ar = StringArray_New();
    StringArray_Read(ar, msgReader);
    LOGDEV_NET_VERBOSE("Received %i mobj state IDs") << StringArray_Size(ar);
    xlatMobjState.resize(StringArray_Size(ar));
    for (dint i = 0; i < StringArray_Size(ar); ++i) {
        xlatMobjState[i] = DED_Definitions()->getStateNum(StringArray_At(ar, i));
        if (xlatMobjState[i] < 0) {
            LOG_NET_WARNING("Could not find '%s' in local state definitions")
                    << StringArray_At(ar, i);
        }
    }
    StringArray_Delete(ar);
}
world::Material *Cl_LocalMaterial(materialarchive_serialid_t archId) {
    if (!serverMaterials) {
        LOGDEV_NET_WARNING("Cannot translate serial id %i, server has not sent its materials!") << archId;
        return 0;
    }
    return serverMaterials->find(archId, 0);
}
int Cl_LocalMobjType(int serverMobjType) {
    if (serverMobjType < 0 || serverMobjType >= xlatMobjType.size())
        return 0; // Invalid type.
    return xlatMobjType[serverMobjType];
}
int Cl_LocalMobjState(int serverMobjState) {
    if (serverMobjState < 0 || serverMobjState >= xlatMobjState.size())
        return 0; // Invalid state.
    return xlatMobjState[serverMobjState];
}
void Cl_ReadSectorDelta(dint /*deltaType*/)
{
    world::Map &map = App_World().map();
#define PLN_FLOOR   0
#define PLN_CEILING 1
    dfloat height[2] = { 0, 0 };
    dfloat target[2] = { 0, 0 };
    dfloat speed[2]  = { 0, 0 };
    Sector *sec = map.sectorPtr(Reader_ReadUInt16(msgReader));
    DENG2_ASSERT(sec);
    dint df = Reader_ReadPackedUInt32(msgReader);
    if (df & SDF_FLOOR_MATERIAL) {
        P_SetPtrp(sec, DMU_FLOOR_OF_SECTOR | DMU_MATERIAL,
                  Cl_LocalMaterial(Reader_ReadPackedUInt16(msgReader)));
    }
    if (df & SDF_CEILING_MATERIAL) {
        P_SetPtrp(sec, DMU_CEILING_OF_SECTOR | DMU_MATERIAL,
                  Cl_LocalMaterial(Reader_ReadPackedUInt16(msgReader)));
    }
    if (df & SDF_LIGHT)
        P_SetFloatp(sec, DMU_LIGHT_LEVEL, Reader_ReadByte(msgReader) / 255.0f);
    if (df & SDF_FLOOR_HEIGHT)
        height[PLN_FLOOR] = FIX2FLT(Reader_ReadInt16(msgReader) << 16);
    if (df & SDF_CEILING_HEIGHT)
        height[PLN_CEILING] = FIX2FLT(Reader_ReadInt16(msgReader) << 16);
    if (df & SDF_FLOOR_TARGET)
        target[PLN_FLOOR] = FIX2FLT(Reader_ReadInt16(msgReader) << 16);
    if (df & SDF_FLOOR_SPEED)
        speed[PLN_FLOOR] = FIX2FLT(Reader_ReadByte(msgReader) << (df & SDF_FLOOR_SPEED_44 ? 12 : 15));
    if (df & SDF_CEILING_TARGET)
        target[PLN_CEILING] = FIX2FLT(Reader_ReadInt16(msgReader) << 16);
    if (df & SDF_CEILING_SPEED)
        speed[PLN_CEILING] = FIX2FLT(Reader_ReadByte(msgReader) << (df & SDF_CEILING_SPEED_44 ? 12 : 15));
    if (df & (SDF_COLOR_RED | SDF_COLOR_GREEN | SDF_COLOR_BLUE)) {
        Vector3f newColor = sec->lightColor();
        if (df & SDF_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        sec->setLightColor(newColor);
    }
    if (df & (SDF_FLOOR_COLOR_RED | SDF_FLOOR_COLOR_GREEN | SDF_FLOOR_COLOR_BLUE)) {
        Vector3f newColor = sec->floor().surface().color();
        if (df & SDF_FLOOR_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_FLOOR_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_FLOOR_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        sec->floor().surface().setColor(newColor);
    }
    if (df & (SDF_CEIL_COLOR_RED | SDF_CEIL_COLOR_GREEN | SDF_CEIL_COLOR_BLUE)) {
        Vector3f newColor = sec->ceiling().surface().color();
        if (df & SDF_CEIL_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_CEIL_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SDF_CEIL_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        sec->ceiling().surface().setColor(newColor);
    }
    if (df & SDF_FLOOR_HEIGHT) {
        ClPlaneMover::newThinker(sec->floor(), height[PLN_FLOOR], 0);
    }
    else if (df & (SDF_FLOOR_TARGET | SDF_FLOOR_SPEED)) {
        ClPlaneMover::newThinker(sec->floor(), target[PLN_FLOOR], speed[PLN_FLOOR]);
    }
    if (df & SDF_CEILING_HEIGHT) {
        ClPlaneMover::newThinker(sec->ceiling(), height[PLN_CEILING], 0);
    }
    else if (df & (SDF_CEILING_TARGET | SDF_CEILING_SPEED)) {
        ClPlaneMover::newThinker(sec->ceiling(), target[PLN_CEILING], speed[PLN_CEILING]);
    }
#undef PLN_CEILING
#undef PLN_FLOOR
}
void Cl_ReadSideDelta(dint /*deltaType*/)
{
    world::Map &map = App_World().map();
    dint const index = Reader_ReadUInt16(msgReader);
    dint const df    = Reader_ReadPackedUInt32(msgReader); // Flags.
    LineSide *side = map.sidePtr(index);
    DENG2_ASSERT(side != 0);
    if (df & SIDF_TOP_MATERIAL) {
        dint matIndex = Reader_ReadPackedUInt16(msgReader);
        side->top().setMaterial(Cl_LocalMaterial(matIndex));
    }
    if (df & SIDF_MID_MATERIAL) {
        dint matIndex = Reader_ReadPackedUInt16(msgReader);
        side->middle().setMaterial(Cl_LocalMaterial(matIndex));
    }
    if (df & SIDF_BOTTOM_MATERIAL) {
        dint matIndex = Reader_ReadPackedUInt16(msgReader);
        side->bottom().setMaterial(Cl_LocalMaterial(matIndex));
    }
    if (df & SIDF_LINE_FLAGS) {
        dint lineFlags = Reader_ReadByte(msgReader);
        Line &line = side->line();
        line.setFlags((line.flags() & ~0xff) | lineFlags, de::ReplaceFlags);
    }
    if (df & (SIDF_TOP_COLOR_RED | SIDF_TOP_COLOR_GREEN | SIDF_TOP_COLOR_BLUE)) {
        Vector3f newColor = side->top().color();
        if (df & SIDF_TOP_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_TOP_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_TOP_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        side->top().setColor(newColor);
    }
    if (df & (SIDF_MID_COLOR_RED | SIDF_MID_COLOR_GREEN | SIDF_MID_COLOR_BLUE)) {
        Vector3f newColor = side->middle().color();
        if (df & SIDF_MID_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_MID_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_MID_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        side->middle().setColor(newColor);
    }
    if (df & SIDF_MID_COLOR_ALPHA) {
        side->middle().setOpacity(Reader_ReadByte(msgReader) / 255.f);
    }
    if (df & (SIDF_BOTTOM_COLOR_RED | SIDF_BOTTOM_COLOR_GREEN | SIDF_BOTTOM_COLOR_BLUE)) {
        Vector3f newColor = side->bottom().color();
        if (df & SIDF_BOTTOM_COLOR_RED)
            newColor.x = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_BOTTOM_COLOR_GREEN)
            newColor.y = Reader_ReadByte(msgReader) / 255.f;
        if (df & SIDF_BOTTOM_COLOR_BLUE)
            newColor.z = Reader_ReadByte(msgReader) / 255.f;
        side->bottom().setColor(newColor);
    }
    if (df & SIDF_MID_BLENDMODE) {
        side->middle().setBlendMode(blendmode_t(Reader_ReadInt32(msgReader)));
    }
    if (df & SIDF_FLAGS) {
        dint sideFlags = Reader_ReadByte(msgReader);
        side->setFlags((side->flags() & ~0xff) | sideFlags, de::ReplaceFlags);
    }
}
void Cl_ReadPolyDelta() {
    world::Map &map = App_World().map();
    Polyobj &pob    = map.polyobj(Reader_ReadPackedUInt16(msgReader));
    dint const df = Reader_ReadByte(msgReader); // Flags.
    if (df & PODF_DEST_X) {
        pob.dest[VX] = Reader_ReadFloat(msgReader);
    }
    if (df & PODF_DEST_Y) {
        pob.dest[VY] = Reader_ReadFloat(msgReader);
    }
    if (df & PODF_SPEED) {
        pob.speed = Reader_ReadFloat(msgReader);
    }
    if (df & PODF_DEST_ANGLE) {
        pob.destAngle = ((angle_t)Reader_ReadInt16(msgReader)) << 16;
    }
    if (df & PODF_ANGSPEED) {
        pob.angleSpeed = ((angle_t)Reader_ReadInt16(msgReader)) << 16;
    }
    if (df & PODF_PERPETUAL_ROTATE) {
        pob.destAngle = -1;
    }
    ClPolyMover::newThinker(pob,
            /* move: */   CPP_BOOL(df & (PODF_DEST_X | PODF_DEST_Y | PODF_SPEED)),
            /* rotate: */ CPP_BOOL(df & (PODF_DEST_ANGLE | PODF_ANGSPEED | PODF_PERPETUAL_ROTATE)));
}
namespace world {
DENG2_PIMPL(ClEdgeLoop) {
    ClientSubsector &owner;
    HEdge *firstHEdge = nullptr;
    dint loopId = 0;
    Impl(Public *i, ClientSubsector &owner)
        : Base(i)
        , owner(owner)
    {}
    /**
     * Look at the neighboring surfaces and pick the best choice of
     * material used on those surfaces to be applied to "this" surface.
     *
     * Material on back neighbor plane has priority. Non-animated materials are preferred.
     * Sky-masked materials are ignored.
     */
    Material *chooseFixMaterial(HEdge &hedge, dint section) const {
        Material *choice1 = nullptr, *choice2 = nullptr;
        if (self().hasBackSubsector()) {
            ClientSubsector &backSubsec = self().backSubsector();
            switch (section) {
            case LineSide::Bottom:
                if (owner.visFloor().height() < backSubsec.visFloor().height()) {
                    choice1 = backSubsec.visFloor().surface().materialPtr();
                }
                break;
            case LineSide::Top:
                if (owner.visCeiling().height() > backSubsec.visCeiling().height()) {
                    choice1 = backSubsec.visCeiling().surface().materialPtr();
                }
                break;
            default: break;
            }
            if (choice1 && choice1->isSkyMasked()) {
                return choice1;
            }
        }
        else {
            Sector *frontSec = &owner.sector();
            LineSide &side = hedge.mapElementAs<LineSideSegment>().lineSide();
            Line *other = R_FindLineNeighbor(side.line(), *side.line().vertexOwner(side.sideId()),
                                             Clockwise, frontSec);
            if (!other) {
                other = R_FindLineNeighbor(side.line(), *side.line().vertexOwner(side.sideId()^1),
                                           Anticlockwise, frontSec);
            }
            if (other) {
                if (!other->back().hasSector()) {
                    choice1 = other->front().middle().materialPtr();
                }
                else {
                    LineSide &otherSide = other->side(&other->front().sector() == frontSec ? Line::Front : Line::Back);
                    Sector &otherSec    = other->side(&other->front().sector() == frontSec ? Line::Back  : Line::Front).sector();
                    if (otherSec.ceiling().height() <= frontSec->floor().height())
                        choice1 = otherSide.top().materialPtr();
                    else if (otherSec.floor().height() >= frontSec->ceiling().height())
                        choice1 = otherSide.bottom().materialPtr();
                    else if (otherSec.ceiling().height() < frontSec->ceiling().height())
                        choice1 = otherSide.top().materialPtr();
                    else if (otherSec.floor().height() > frontSec->floor().height())
                        choice1 = otherSide.bottom().materialPtr();
                }
            }
        }
        choice2 = owner.visPlane(section == LineSide::Bottom ? Sector::Floor : Sector::Ceiling)
                           .surface().materialPtr();
        if (choice1 && !choice1->hasAnimatedTextureLayers() && !choice1->isSkyMasked())
            return choice1;
        if (choice2 && !choice2->hasAnimatedTextureLayers() && !choice2->isSkyMasked())
            return choice2;
        if (choice1 && !choice1->isSkyMasked())
            return choice1;
        if (choice2 && !choice2->isSkyMasked())
            return choice2;
        if (choice1) return choice1;
        if (choice2) return choice2;
        return &Materials::get().material(de::Uri("System", Path("missing")));
    }
    void fixMissingMaterial(HEdge &hedge, dint section) const {
        LineSide &side = hedge.mapElementAs<LineSideSegment>().lineSide();
        if (!side.hasSections()) return;
        if (side.line().isSelfReferencing()) return;
        if (!side.back().hasSections() && side.back().hasSector()) return;
        Surface &surface = side.surface(section);
        if (!surface.hasMaterial() || surface.hasFixMaterial()) {
            Material *oldMaterial = surface.materialPtr();
            surface.setMaterial(chooseFixMaterial(hedge, section), true/* is missing fix */);
            if (surface.materialPtr() != oldMaterial) {
                owner.markReverbDirty();
            }
        }
    }
};
ClEdgeLoop::ClEdgeLoop(ClientSubsector &owner, HEdge &first, dint loopId)
    : d(new Impl(this, owner)) {
    d->firstHEdge = &first;
    d->loopId     = loopId;
}
ClientSubsector &ClEdgeLoop::owner() const {
    return d->owner;
}
String ClEdgeLoop::description() const {
    auto desc = String(    _E(l) "Loop: "       _E(.)_E(i) "%1" _E(.)
                       " " _E(l) "Half-edge: "  _E(.)_E(i) "%2" _E(.))
                  .arg(ClientSubsector::edgeLoopIdAsText(loopId()).upperFirstChar())
                  .arg(String("[0x%1]").arg(de::dintptr(d->firstHEdge), 0, 16));
    DENG2_DEBUG_ONLY(
        desc.prepend(String(_E(b) "ClEdgeLoop " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
    )
    return desc;
}
dint ClEdgeLoop::loopId() const {
    return d->loopId;
}
bool ClEdgeLoop::isInner() const {
    return d->loopId == ClientSubsector::InnerLoop;
}
bool ClEdgeLoop::isOuter() const {
    return d->loopId == ClientSubsector::OuterLoop;
}
HEdge &ClEdgeLoop::first() const {
    DENG2_ASSERT(d->firstHEdge);
    return *d->firstHEdge;
}
bool ClEdgeLoop::isSelfReferencing() const {
    DENG2_ASSERT(d->firstHEdge);
    return d->firstHEdge->mapElementAs<LineSideSegment>().line().isSelfReferencing();
}
bool ClEdgeLoop::hasBackSubsector() const {
    DENG2_ASSERT(d->firstHEdge);
    return d->firstHEdge->hasTwin()
        && d->firstHEdge->twin().hasFace()
        && d->firstHEdge->twin().face().mapElementAs<ConvexSubspace>().hasSubsector();
}
ClientSubsector &ClEdgeLoop::backSubsector() const {
    DENG2_ASSERT(d->firstHEdge);
    return d->firstHEdge->twin().face().mapElementAs<ConvexSubspace>().subsector().as<ClientSubsector>();
}
void ClEdgeLoop::fixSurfacesMissingMaterials() {
    DENG2_ASSERT(d->firstHEdge);
    SubsectorCirculator it(d->firstHEdge);
    do {
        if (it->hasMapElement()) // BSP errors may fool the circulator wrt interior edges -ds
        {
            LineSide &lineSide = it->mapElementAs<LineSideSegment>().lineSide();
            if (lineSide.hasSections()) // Not a "one-way window" -ds
            {
                if (hasBackSubsector()) {
                    auto const &backSubsec = backSubsector().as<ClientSubsector>();
                    if (!d->owner.hasSkyFloor() && !backSubsec.hasSkyFloor()) {
                        if (d->owner.visFloor().height() < backSubsec.visFloor().height()) {
                            d->fixMissingMaterial(*it, LineSide::Bottom);
                        }
                        else if (lineSide.bottom().hasFixMaterial()) {
                            lineSide.bottom().setMaterial(nullptr);
                        }
                    }
                    if (!d->owner.hasSkyCeiling() && !backSubsec.hasSkyCeiling()) {
                        if (d->owner.visCeiling().height() > backSubsec.visCeiling().height()) {
                            d->fixMissingMaterial(*it, LineSide::Top);
                        }
                        else if (lineSide.top().hasFixMaterial()) {
                            lineSide.top().setMaterial(nullptr);
                        }
                    }
                }
                else if (!lineSide.back().hasSector()) {
                    d->fixMissingMaterial(*it, LineSide::Middle);
                }
            }
        }
    } while (&it.next() != d->firstHEdge);
}
} // namespace world
namespace world {
enum SubsectorFlag {
    NeverMapped      = 0x01,
    AllMissingBottom = 0x02,
    AllMissingTop    = 0x04,
    AllSelfRef       = 0x08,
    PartSelfRef      = 0x10
};
Q_DECLARE_FLAGS(SubsectorFlags, SubsectorFlag)
Q_DECLARE_OPERATORS_FOR_FLAGS(SubsectorFlags)
#ifdef DENG2_DEBUG
/**
 * Returns a textual, map-relative path for the given @a surface.
 */
static DotPath composeSurfacePath(Surface const &surface) {
    DENG2_ASSERT(surface.hasParent());
    MapElement const &owner = surface.parent();
    switch (owner.type()) {
    case DMU_PLANE:
        return String("sector#%1.%2")
                 .arg(owner.as<Plane>().sector().indexInMap())
                 .arg(Sector::planeIdAsText(owner.as<Plane>().indexInSector()));
    case DMU_SIDE:
        return String("line#%1.%2.%3")
                 .arg(owner.as<LineSide>().line().indexInMap())
                 .arg(Line::sideIdAsText(owner.as<LineSide>().sideId()))
                 .arg(LineSide::sectionIdAsText(  &surface == &owner.as<LineSide>().middle() ? LineSide::Middle
                                                : &surface == &owner.as<LineSide>().bottom() ? LineSide::Bottom
                                                : LineSide::Top));
    default: return "";
    }
}
#endif // DENG2_DEBUG
/**
 * @todo optimize: Translation of decorations on the world up axis would be a trivial
 * operation to perform, which, would not require plotting decorations again. This
 * frequent case should be designed for. -ds
 */
DENG2_PIMPL(ClientSubsector)
, DENG2_OBSERVES(Subsector, Deletion)
, DENG2_OBSERVES(Plane,     Deletion)
, DENG2_OBSERVES(Line,    FlagsChange)
, DENG2_OBSERVES(Plane,   HeightChange)
, DENG2_OBSERVES(Plane,   HeightSmoothedChange)
, DENG2_OBSERVES(Surface, MaterialChange)
, DENG2_OBSERVES(Surface, OriginChange)
, DENG2_OBSERVES(Surface, OriginSmoothedChange)
, DENG2_OBSERVES(Material,         DimensionsChange)
, DENG2_OBSERVES(MaterialAnimator, DecorationStageChange) {
    struct BoundaryData {
        struct EdgeLoops : public QList<ClEdgeLoop *> {
            ~EdgeLoops() { clear(); }
            void clear() {
                qDeleteAll(*this);
                QList<ClEdgeLoop *>::clear();
            }
        };
        std::unique_ptr<ClEdgeLoop> outerLoop;
        EdgeLoops innerLoops;
        void clear() {
            outerLoop.release();
            innerLoops.clear();
        }
        /**
         * Add @a loop to the data set (ownership is given).
         */
        BoundaryData &addEdgeLoop(ClEdgeLoop *loop) {
            if (loop) {
                if (loop->isOuter()) outerLoop.reset(loop);
                else                 innerLoops.append(loop);
            }
            return *this;
        }
    };
    struct GeometryData {
        MapElement *mapElement;
        dint geomId;
#if 0
        std::unique_ptr<Shard> shard;
#endif
        GeometryData(MapElement *mapElement, dint geomId)
            : mapElement(mapElement)
            , geomId(geomId)
        {}
    };
    typedef QMap<dint, GeometryData *> Shards;
    struct GeometryGroups : public QMap<MapElement *, Shards> {
        ~GeometryGroups() { DENG2_FOR_EACH(GeometryGroups, g, *this) qDeleteAll(*g); }
    };
    struct DecoratedSurface : public Surface::IDecorationState {
        QVector<Decoration *> decorations;
        bool needUpdate = true;
        DecoratedSurface() {}
        ~DecoratedSurface() {
            qDeleteAll(decorations);
        }
        void markForUpdate(bool yes = true) {
            if (::ddMapSetup) return;
            needUpdate = yes;
        }
        void clear() {
            markForUpdate(false);
            qDeleteAll(decorations);
            decorations.clear();
        }
    };
    dint validFrame;
    bool hasWorldVolumeInValidFrame;
    bool hasInvisibleTop = false;
    bool           needClassify = true; ///< @c true= (Re)classification is necessary.
    SubsectorFlags flags        = 0;
    struct VisPlaneLink {
        VisPlaneLinkMode linkMode = LinkAlways;
        ClientSubsector *target   = nullptr;
        ClientSubsector *current  = nullptr; // current link in effect, perhaps nullptr
    };
    VisPlaneLink visPlaneLinks[2];
    std::unique_ptr<BoundaryData> boundaryData;
#if 0
    GeometryGroups geomGroups;
    QHash<Shard *, GeometryData *> shardToGeomData; ///< Reverse lookup.
#endif
    typedef QSet<ConvexSubspace *> ReverbSubspaces;
    ReverbSubspaces reverbSubspaces;
    AudioEnvironment reverb;
    bool needReverbUpdate = true;
    QSet<Surface *> decorSurfaces;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() override {
        unlinkVisPlane(Sector::Floor);
        unlinkVisPlane(Sector::Ceiling);
    }
#if 0
    inline bool floorIsMapped() {
        return visPlaneLinks[Sector::Floor].current &&
               visPlaneLinks[Sector::Floor].current != thisPublic;
    }
    inline bool ceilingIsMapped() {
        return visPlaneLinks[Sector::Ceiling].current &&
               visPlaneLinks[Sector::Ceiling].current != thisPublic;
    }
    inline bool needRemapVisPlanes() {
        return visPlaneLinks[Sector::Floor].target == 0 || mappedVisCeiling == 0;
    }
    ClientSubsector **mappedSubsectorAdr(dint planeIdx) {
        if (planeIdx == Sector::Floor)   return &mappedVisFloor;
        if (planeIdx == Sector::Ceiling) return &mappedVisCeiling;
        return nullptr;
    }
#endif
    Plane *linkedPlane(dint planeIdx) {
        DE_ASSERT(planeIdx >= Sector::Floor && planeIdx <= Sector::Ceiling);
        const auto &visp = visPlaneLinks[planeIdx];
        if (visp.current) {
            return &visp.current->sector().plane(planeIdx);
        }
        return nullptr;
    }
    void observeMaterial(Material *material, bool yes = true) {
        if (!material) return;
        MaterialAnimator &materialAnimator = material->as<ClientMaterial>().getAnimator(Rend_MapSurfaceMaterialSpec());
        if (yes) {
            material->audienceForDimensionsChange() += this;
            materialAnimator.audienceForDecorationStageChange += this;
        }
        else {
            materialAnimator.audienceForDecorationStageChange -= this;
            material->audienceForDimensionsChange() -= this;
        }
    }
    static bool hasDecoratedMaterial(const Surface &surface) {
        return surface.hasMaterial() && surface.material().as<ClientMaterial>().hasDecorations();
    }
    void observeSurface(Surface *surface, bool yes = true) {
        if (!surface) return;
        if (yes) {
            surface->audienceForMaterialChange      () += this;
            surface->audienceForOriginChange        () += this;
            surface->audienceForOriginSmoothedChange() += this;
            if (hasDecoratedMaterial(*surface)) {
                allocDecorationState(*surface);
            }
        }
        else {
            surface->audienceForOriginSmoothedChange() -= this;
            surface->audienceForOriginChange        () -= this;
            surface->audienceForMaterialChange      () -= this;
            decorSurfaces.remove(surface);
        }
    }
    void observePlane(Plane *plane, bool yes = true, bool observeHeight = true) {
        if (!plane) return;
        if (yes) {
            plane->audienceForDeletion() += this;
            if (observeHeight) {
                plane->audienceForHeightChange        () += this;
                plane->audienceForHeightSmoothedChange() += this;
            }
        }
        else {
            plane->audienceForHeightSmoothedChange() -= this;
            plane->audienceForHeightChange        () -= this;
            plane->audienceForDeletion            () -= this;
        }
    }
    void observeSubsector(ClientSubsector *subsec, bool yes = true) {
        if (!subsec || subsec == thisPublic)
            return;
        if (yes) subsec->audienceForDeletion += this;
        else     subsec->audienceForDeletion -= this;
    }
    void linkVisPlane(dint planeIdx, ClientSubsector *newSubsector) {
        auto *visp = &visPlaneLinks[planeIdx];
        if (visp->current == newSubsector) {
            return;
        }
        if (visp->current != thisPublic) {
            if (Plane *oldPlane = linkedPlane(planeIdx)) {
                observeMaterial(oldPlane->surface().materialPtr(), false);
                observeSurface(&oldPlane->surface(), false);
                observePlane(oldPlane, false);
            }
        }
        observeSubsector(visp->current, false);
        visp->current = newSubsector;
        observeSubsector(visp->current);
        if (visp->current != thisPublic) {
            if (Plane *newPlane = linkedPlane(planeIdx)) {
                observePlane(newPlane, true, true); // !permanent);
                observeSurface(&newPlane->surface(), true);
                observeMaterial(newPlane->surface().materialPtr(), true);
            }
        }
    }
    void unlinkVisPlane(dint planeIdx) {
        linkVisPlane(planeIdx, nullptr);
    }
    void relinkVisPlanes() {
        for (int planeIdx = 0; planeIdx < 2; ++planeIdx) {
            auto *visp = &visPlaneLinks[planeIdx];
            if (!visp->target) continue; // Not linked to anything.
            const coord_t planeZ  = self().sector().plane(planeIdx).heightSmoothed();
            const coord_t targetZ = visp->target->visPlane(planeIdx).heightSmoothed();
            if ( visp->linkMode == LinkAlways ||
                (visp->linkMode == LinkWhenDifferentThanTarget && !fequal(planeZ, targetZ)) ||
                (visp->linkMode == LinkWhenLowerThanTarget     && planeZ < targetZ) ||
                (visp->linkMode == LinkWhenHigherThanTarget    && planeZ > targetZ)) {
                linkVisPlane(planeIdx, visp->target);
            }
            else {
                unlinkVisPlane(planeIdx);
            }
        }
    }
#if 0
    /*
     * To be called when a plane moves to possibly invalidate mapped planes so
     * that they will be re-evaluated later.
     */
    void maybeInvalidateMapping(dint planeIdx) {
        if (classification() & NeverMapped)
            return;
        ClientSubsector **subsecAdr = mappedSubsectorAdr(planeIdx);
        if (!subsecAdr || *subsecAdr == thisPublic)
            return;
        unlinkVisPlane(planeIdx);
        if (classification() & (AllMissingBottom|AllMissingTop)) {
            needClassify = true;
        }
    }
#endif
    /**
     * Returns a copy of the classification flags for the subsector, performing
     * classification of the subsector if necessary.
     */
    SubsectorFlags classification() {
        if (needClassify) {
            needClassify = false;
            flags &= ~(NeverMapped | PartSelfRef);
            flags |= AllSelfRef | AllMissingBottom | AllMissingTop;
            self().forAllSubspaces([this] (ConvexSubspace &subspace) {
                HEdge const *base  = subspace.poly().hedge();
                HEdge const *hedge = base;
                do {
                    if (!hedge->hasMapElement())
                        continue;
                    if (!hedge->twin().hasFace()) {
                        flags |= NeverMapped;
                        flags &= ~(PartSelfRef | AllSelfRef | AllMissingBottom | AllMissingTop);
                        return LoopAbort;
                    }
                    if (!hedge->twin().face().hasMapElement())
                        continue;
                    auto const &backSpace = hedge->twin().face().mapElementAs<ConvexSubspace>();
                    if (&backSpace.subsector() == thisPublic)
                        continue;
                    LineSide const &frontSide = hedge->mapElementAs<LineSideSegment>().lineSide();
                    LineSide const &backSide  = hedge->twin().mapElementAs<LineSideSegment>().lineSide();
                    if (!frontSide.hasSections() || !backSide.hasSections()) {
                        flags |= NeverMapped;
                        flags &= ~(PartSelfRef | AllSelfRef | AllMissingBottom | AllMissingTop);
                        return LoopAbort;
                    }
                    if (frontSide.line().isSelfReferencing()) {
                        flags |= PartSelfRef;
                        continue;
                    }
                    flags &= ~AllSelfRef;
                    if (frontSide.bottom().hasDrawableNonFixMaterial()) {
                        flags &= ~AllMissingBottom;
                    }
                    if (frontSide.top().hasDrawableNonFixMaterial()) {
                        flags &= ~AllMissingTop;
                    }
                    if (backSpace.subsector().sector().floor().height() < self().sector().floor().height()
                        && backSide.bottom().hasDrawableNonFixMaterial()) {
                        flags &= ~AllMissingBottom;
                    }
                    if (backSpace.subsector().sector().ceiling().height() > self().sector().ceiling().height()
                        && backSide.top().hasDrawableNonFixMaterial()) {
                        flags &= ~AllMissingTop;
                    }
                } while ((hedge = &hedge->next()) != base);
                return LoopContinue;
            });
        }
        return flags;
    }
    void initBoundaryDataIfNeeded() {
        if (boundaryData) return;
        boundaryData.reset(new BoundaryData);
        QList<HEdge *> neighbors = self().listUniqueBoundaryEdges();
        if (neighbors.count() == 1) {
            boundaryData->addEdgeLoop(new ClEdgeLoop(self(), *neighbors.first()));
        }
        else {
            DENG2_ASSERT(!neighbors.isEmpty());
            QList<QRectF> boundaries;
            for (HEdge *base : neighbors) {
                QRectF bounds;
                SubsectorCirculator it(base);
                do {
                    bounds |= QRectF(QPointF(it->origin().x, it->origin().y),
                                     QPointF(it->twin().origin().x, it->twin().origin().y))
                        .normalized();
                } while (&it.next() != base);
                boundaries << bounds;
            }
            QRectF const *largest = nullptr;
            for (QRectF const &boundary : boundaries) {
                if (!largest || boundary.contains(*largest))
                    largest = &boundary;
            }
            for (dint i = 0; i < neighbors.count(); ++i) {
                QRectF &boundary = boundaries[i];
                HEdge *hedge     = neighbors[i];
                boundaryData->addEdgeLoop(new ClEdgeLoop(self(), *hedge, !(&boundary == largest || (largest && boundary == *largest))));
            }
        }
    }
#if 0
    void remapVisPlanes() {
        if (!floorIsMapped())   map(Sector::Floor,   thisPublic);
        if (!ceilingIsMapped()) map(Sector::Ceiling, thisPublic);
        if (classification() & NeverMapped)
            return;
        if (classification() & (AllSelfRef | PartSelfRef)) {
            self().forAllEdgeLoops([this] (ClEdgeLoop const &loop) {
                auto &backSubsec = loop.backSubsector().as<ClientSubsector>();
                if (loop.isSelfReferencing()
                    && (classification() & AllSelfRef)
                    && !(backSubsec.d->classification() & AllSelfRef)
                    && backSubsec.d->mappedVisFloor != thisPublic) {
                    map(Sector::Floor  , &backSubsec, true /*permanently*/);
                    map(Sector::Ceiling, &backSubsec, true /*permanently*/);
                    return LoopAbort;
                }
                else {
                    return LoopContinue;
                }
            });
            if (floorIsMapped()) {
                for (ClEdgeLoop *loop : boundaryData->innerLoops) {
                    if (loop->isSelfReferencing()) {
                        auto &backSubsec = loop->backSubsector().as<ClientSubsector>();
                        if ((classification() & AllSelfRef)
                            && !(backSubsec.d->classification() & AllSelfRef)) {
                            if (backSubsec.d->mappedVisFloor == thisPublic) {
                                backSubsec.d->unlinkVisPlane(Sector::Floor);
                            }
                            if (backSubsec.d->mappedVisCeiling == thisPublic) {
                                backSubsec.d->unlinkVisPlane(Sector::Ceiling);
                            }
                        }
                    }
                }
                return;
            }
        }
        if (classification() & AllSelfRef)
            return;
        if (self().sector().ceiling().height() <= self().sector().floor().height())
            return;
        bool doFloor   =   !floorIsMapped() && classification().testFlag(AllMissingBottom);
        bool doCeiling = !ceilingIsMapped() && classification().testFlag(AllMissingTop);
        if (!doFloor && !doCeiling)
            return;
        initBoundaryDataIfNeeded();
        DENG2_ASSERT(boundaryData->outerLoop); {
            ClEdgeLoop const &loop = *boundaryData->outerLoop;
            if (loop.hasBackSubsector()) {
                auto &backSubsec = loop.backSubsector().as<ClientSubsector>();
                SubsectorCirculator it(&loop.first());
                do {
                    if (it->hasMapElement()) // BSP errors may fool the circulator wrt interior edges -ds
                    {
                        if (doFloor && !floorIsMapped()
                            && !backSubsec.hasSkyFloor()
                            && backSubsec.visFloor().height() > self().sector().floor().height()) {
                            map(Sector::Floor, &backSubsec);
                            if (!doCeiling) break;
                        }
                        if (doCeiling && !ceilingIsMapped()
                            && !backSubsec.hasSkyCeiling()
                            && backSubsec.visCeiling().height() < self().sector().ceiling().height()) {
                            map(Sector::Ceiling, &backSubsec);
                            if (!doFloor) break;
                        }
                    }
                } while (&it.next() != &loop.first());
            }
        }
        if (!floorIsMapped() && !ceilingIsMapped())
            return;
        for (ClEdgeLoop const *loop : boundaryData->innerLoops) {
            if (loop->hasBackSubsector()) {
                auto &backSubsec = loop->backSubsector().as<ClientSubsector>();
                SubsectorCirculator it(&loop->first());
                do {
                    if (it->hasMapElement() // BSP errors may fool the circulator wrt interior edges -ds
                        && !(backSubsec.d->classification() & NeverMapped)) {
                        if (doFloor && floorIsMapped()
                            && backSubsec.visFloor().height() >= self().sector().floor().height()) {
                            backSubsec.d->unlinkVisPlane(Sector::Floor);
                        }
                        if (doCeiling && ceilingIsMapped()
                            && backSubsec.visCeiling().height() <= self().sector().ceiling().height()) {
                            backSubsec.d->unlinkVisPlane(Sector::Ceiling);
                        }
                    }
                } while (&it.next() != &loop->first());
            }
        }
    }
#endif
#if 0
    void updateBiasForWallSectionsAfterGeometryMove(HEdge *hedge) {
        if (!hedge) return;
        if (!hedge->hasMapElement()) return;
        MapElement *mapElement = &hedge->mapElement();
        if (Shard *shard = self().findShard(*mapElement, LineSide::Middle)) {
            shard->updateBiasAfterMove();
        }
        if (Shard *shard = self().findShard(*mapElement, LineSide::Bottom)) {
            shard->updateBiasAfterMove();
        }
        if (Shard *shard = self().findShard(*mapElement, LineSide::Top)) {
            shard->updateBiasAfterMove();
        }
    }
#endif
#if 0
    /**
     * Find the GeometryData for a MapElement by the element-unique @a group identifier.
     *
     * @param geomId    Geometry identifier.
     *
     * @param canAlloc  @c true= to allocate if no data exists. Note that the number of
     * vertices in the fan geometry must be known at this time.
     */
    GeometryData *geomData(MapElement &mapElement, dint geomId, bool canAlloc = false) {
        GeometryGroups::iterator foundGroup = geomGroups.find(&mapElement);
        if (foundGroup != geomGroups.end()) {
            Shards &shards = *foundGroup;
            Shards::iterator found = shards.find(geomId);
            if (found != shards.end()) {
                return *found;
            }
        }
        if (!canAlloc) return nullptr;
        if (foundGroup == geomGroups.end()) {
            foundGroup = geomGroups.insert(&mapElement, Shards());
        }
        return *foundGroup->insert(geomId, new GeometryData(&mapElement, geomId));
    }
    /**
     * Find the GeometryData for the given @a shard.
     */
    GeometryData *geomDataForShard(Shard *shard) {
        if (shard && shard->subsector() == thisPublic) {
            auto found = shardToGeomData.find(shard);
            if (found != shardToGeomData.end()) return *found;
        }
        return nullptr;
    }
#endif
    void addReverbSubspace(ConvexSubspace *subspace) {
        if (!subspace) return;
        reverbSubspaces.insert(subspace);
    }
    /**
     * Perform environmental audio (reverb) initialization.
     *
     * Determines the subspaces which contribute to the environmental audio
     * characteristics. Given that subspaces do not change shape (on the XY plane,
     * that is), they do not move and are not created/destroyed once the map has
     * been loaded; this step can be pre-processed.
     *
     * @pre The Map's BSP leaf blockmap must be ready for use.
     */
    void findReverbSubspaces() {
        Map const &map = self().sector().map();
        AABoxd box = self().bounds();
        box.minX -= 128;
        box.minY -= 128;
        box.maxX += 128;
        box.maxY += 128;
        dint const localValidCount = ++validCount;
        map.subspaceBlockmap().forAllInBox(box, [this, &box, &localValidCount] (void *object) {
            auto &sub = *(ConvexSubspace *)object;
            if (sub.validCount() != localValidCount) // not yet processed
            {
                sub.setValidCount(localValidCount);
                AABoxd const &polyBounds = sub.poly().bounds();
                if (!(   polyBounds.maxX < box.minX
                      || polyBounds.minX > box.maxX
                      || polyBounds.minY > box.maxY
                      || polyBounds.maxY < box.minY)) {
                    addReverbSubspace(&sub);
                }
            }
            return LoopContinue;
        });
    }
    /**
     * Recalculate environmental audio (reverb) for the sector.
     */
    void updateReverb() {
        if (reverbSubspaces.isEmpty()) {
            findReverbSubspaces();
        }
        needReverbUpdate = false;
        duint spaceVolume = dint((self().visCeiling().height() - self().visFloor().height())
                          * self().roughArea());
        reverb.reset();
        for (ConvexSubspace *subspace : reverbSubspaces) {
            if (subspace->updateAudioEnvironment()) {
                auto const &aenv = subspace->audioEnvironment();
                reverb.space   += aenv.space;
                reverb.volume  += aenv.volume  / 255.0f * aenv.space;
                reverb.decay   += aenv.decay   / 255.0f * aenv.space;
                reverb.damping += aenv.damping / 255.0f * aenv.space;
            }
        }
        dfloat spaceScatter;
        if (reverb.space) {
            spaceScatter = spaceVolume / reverb.space;
            reverb.volume  /= reverb.space;
            reverb.decay   /= reverb.space;
            reverb.damping /= reverb.space;
        }
        else {
            spaceScatter = 0;
            reverb.volume  = .2f;
            reverb.decay   = .4f;
            reverb.damping = 1;
        }
        reverb.space /= (spaceScatter > .8 ? 10 : spaceScatter > .6 ? 4 : 1);
        reverb.space /= 120e6;
        if (reverb.space > .99)
            reverb.space = .99f;
        if (self().hasSkyPlane()) {
            if (reverb.space > .5)
                reverb.volume = 1;    // Full volume.
            else
                reverb.volume = .5f;  // Small, but still open.
            reverb.space = 1;
        }
        else {
            reverb.volume += reverb.space / 4;
        }
        if (reverb.volume > 1)
            reverb.volume = 1;
    }
    bool prepareGeometry(Surface &surface, Vector3d &topLeft, Vector3d &bottomRight,
                         Vector2f &materialOrigin) const {
        if (surface.parent().type() == DMU_SIDE) {
            auto &side = surface.parent().as<LineSide>();
            dint section = &side.middle() == &surface ? LineSide::Middle
                         : &side.bottom() == &surface ? LineSide::Bottom
                         :                              LineSide::Top;
            if (!side.hasSections()) return false;
            HEdge *leftHEdge = side.leftHEdge();
            HEdge *rightHEdge = side.rightHEdge();
            if (!leftHEdge || !rightHEdge) return false;
            WallSpec const wallSpec = WallSpec::fromMapSide(side, section);
            WallEdge leftEdge(wallSpec, *leftHEdge, Line::From);
            WallEdge rightEdge(wallSpec, *rightHEdge, Line::To);
            if (!leftEdge.isValid() || !rightEdge.isValid()
                || de::fequal(leftEdge.bottom().z(), rightEdge.top().z()))
                return false;
            topLeft = leftEdge.top().origin();
            bottomRight = rightEdge.bottom().origin();
            materialOrigin = -leftEdge.materialOrigin();
            return true;
        }
        if (surface.parent().type() == DMU_PLANE) {
            auto &plane = surface.parent().as<Plane>();
            AABoxd const &sectorBounds = plane.sector().bounds();
            topLeft = Vector3d(sectorBounds.minX,
                               plane.isSectorFloor() ? sectorBounds.maxY : sectorBounds.minY,
                               plane.heightSmoothed());
            bottomRight = Vector3d(sectorBounds.maxX,
                                   plane.isSectorFloor() ? sectorBounds.minY : sectorBounds.maxY,
                                   plane.heightSmoothed());
            materialOrigin = Vector2f(-fmod(sectorBounds.minX, 64), -fmod(sectorBounds.minY, 64))
                           - surface.originSmoothed();
            return true;
        }
        return false;
    }
    void projectDecorations(Surface &suf, MaterialAnimator &matAnimator,
                            Vector2f const &materialOrigin, Vector3d const &topLeft,
                            Vector3d const &bottomRight) {
        Vector3d delta = bottomRight - topLeft;
        if (de::fequal(delta.length(), 0)) return;
        ClientMaterial &material = matAnimator.material();
        dint const axis = suf.normal().maxAxis();
        Vector2d sufDimensions;
        if (axis == 0 || axis == 1) {
            sufDimensions.x = std::sqrt(de::squared(delta.x) + de::squared(delta.y));
            sufDimensions.y = delta.z;
        }
        else {
            sufDimensions.x = std::sqrt(de::squared(delta.x));
            sufDimensions.y = delta.y;
        }
        if (sufDimensions.x < 0) sufDimensions.x = -sufDimensions.x;
        if (sufDimensions.y < 0) sufDimensions.y = -sufDimensions.y;
        dint decorIndex = 0;
        material.forAllDecorations([this, &suf, &matAnimator, &materialOrigin
                                   , &topLeft, &bottomRight
                                   , &delta, &axis, &sufDimensions, &decorIndex]
                                   (MaterialDecoration &decor) {
            Vector2ui const &matDimensions = matAnimator.material().dimensions();
            MaterialAnimator::Decoration const &decorSS = matAnimator.decoration(decorIndex);
            Vector2i skip = Vector2i(decor.patternSkip().x + 1, decor.patternSkip().y + 1)
                .max(Vector2i(1, 1));
            Vector2f repeat = skip.toVector2ui() * matDimensions;
            if (repeat == Vector2f(0, 0))
                return LoopAbort;
            Vector3d origin = topLeft + suf.normal() * decorSS.elevation();
            dfloat s = de::wrap(decorSS.origin().x - matDimensions.x * decor.patternOffset().x + materialOrigin.x,
                                0.f, repeat.x);
            for (; s < sufDimensions.x; s += repeat.x) {
                dfloat t = de::wrap(decorSS.origin().y - matDimensions.y * decor.patternOffset().y + materialOrigin.y,
                                    0.f, repeat.y);
                for (; t < sufDimensions.y; t += repeat.y) {
                    auto const offset = Vector2f(s, t) / sufDimensions;
                    Vector3d patternOffset(offset.x,
                                           axis == 2 ? offset.y : offset.x,
                                           axis == 2 ? offset.x : offset.y);
                    Vector3d decorOrigin = origin + delta * patternOffset;
                    if (suf.map().subsectorAt(decorOrigin) == thisPublic) {
                        std::unique_ptr<LightDecoration> decor(new LightDecoration(decorSS, decorOrigin));
                        decor->setSurface(&suf);
                        if (self().sector().hasMap()) decor->setMap(&self().sector().map());
                        static_cast<DecoratedSurface *>(suf.decorationState())->
                                decorations.append(decor.get()); // take ownership.
                        decor.release();
                    }
                }
            }
            decorIndex += 1;
            return LoopContinue;
        });
    }
    DecoratedSurface &allocDecorationState(Surface &surface) {
        auto *ds = static_cast<DecoratedSurface *>(surface.decorationState());
        if (!ds) {
            surface.setDecorationState(ds = new DecoratedSurface);
            decorSurfaces.insert(&surface);
        }
        return *ds;
    }
    void decorate(Surface &surface) {
        if (!hasDecoratedMaterial(surface)) {
            if (surface.decorationState()) {
                static_cast<DecoratedSurface *>(surface.decorationState())->clear();
            }
            return;
        }
        auto &ds = allocDecorationState(surface);
        if (!ds.needUpdate) return;
        LOGDEV_MAP_XVERBOSE_DEBUGONLY("  decorating %s%s"
            , composeSurfacePath(surface)
                << (surface.parent().type() == DMU_PLANE
                    && &surface.parent() == linkedPlane(surface.parent().as<Plane>().indexInSector()) ? " (mapped)" : "")
        );
        ds.clear();
        Vector2f materialOrigin;
        Vector3d bottomRight, topLeft;
        if (prepareGeometry(surface, topLeft, bottomRight, materialOrigin)) {
            MaterialAnimator &animator = *surface.materialAnimator();
            projectDecorations(surface, animator, materialOrigin, topLeft, bottomRight);
        }
    }
    void markDependentSurfacesForRedecoration(Plane &plane, bool yes = true) {
        if (::ddMapSetup) return;
        LOGDEV_MAP_XVERBOSE_DEBUGONLY("Marking [%p] (sector: %i) for redecoration..."
            , thisPublic << self().sector().indexInMap()
        );
        self().forAllEdgeLoops([this, &plane, &yes] (ClEdgeLoop const &loop) {
            SubsectorCirculator it(&loop.first());
            do {
                if (it->hasMapElement()) // BSP errors may fool the circulator wrt interior edges -ds
                {
                    if (    &plane == &self().visPlane(plane.indexInSector())
                        || (&plane == (it->hasTwin() && it->twin().hasFace()
                                       ? &it->twin().face().mapElementAs<ConvexSubspace>()
                                                  .subsector().as<ClientSubsector>().visPlane(plane.indexInSector())
                                       : nullptr))) {
                        LineSide &side = it->mapElementAs<LineSideSegment>().lineSide();
                        side.forAllSurfaces([this, &yes] (Surface &surface) {
                            LOGDEV_MAP_XVERBOSE_DEBUGONLY("  ", composeSurfacePath(surface));
                            if (auto *decor = surface.decorationState()) {
                                static_cast<DecoratedSurface *>(decor)->markForUpdate(yes);
                            }
                            return LoopContinue;
                        });
                    }
                }
            } while (&it.next() != &loop.first());
            return LoopContinue;
        });
        if (&plane == &self().visPlane(plane.indexInSector())) {
            LOGDEV_MAP_XVERBOSE_DEBUGONLY("  ", composeSurfacePath(plane.surface()));
            if (auto *decor = plane.surface().decorationState()) {
                static_cast<DecoratedSurface *>(decor)->markForUpdate(yes);
            }
        }
    }
    void markDependentSurfacesForRedecoration(Material &material, bool yes = true) {
        if (::ddMapSetup) return;
        LOGDEV_MAP_XVERBOSE_DEBUGONLY("Marking [%p] (sector: %i) for redecoration..."
            , thisPublic << self().sector().indexInMap()
        );
        self().forAllEdgeLoops([this, &material, &yes] (ClEdgeLoop const &loop) {
            SubsectorCirculator it(&loop.first());
            do {
                if (it->hasMapElement()) // BSP errors may fool the circulator wrt interior edges -ds
                {
                    LineSide &side = it->mapElementAs<LineSideSegment>().lineSide();
                    side.forAllSurfaces([this, &material, &yes] (Surface &surface) {
                        if (surface.materialPtr() == &material) {
                            LOGDEV_MAP_XVERBOSE_DEBUGONLY("  ", composeSurfacePath(surface));
                            if (auto *decor = surface.decorationState()) {
                                static_cast<DecoratedSurface *>(decor)->markForUpdate(yes);
                            }
                        }
                        return LoopContinue;
                    });
                }
            } while (&it.next() != &loop.first());
            return LoopContinue;
        });
        Plane &floor = self().visFloor();
        if (floor.surface().materialPtr() == &material) {
            LOGDEV_MAP_XVERBOSE_DEBUGONLY("  ", composeSurfacePath(floor.surface()));
            if (auto *decor = floor.surface().decorationState()) {
                static_cast<DecoratedSurface *>(decor)->markForUpdate(yes);
            }
        }
        Plane &ceiling = self().visCeiling();
        if (ceiling.surface().materialPtr() == &material) {
            LOGDEV_MAP_XVERBOSE_DEBUGONLY("  ", composeSurfacePath(ceiling.surface()));
            if (auto *decor = ceiling.surface().decorationState()) {
                static_cast<DecoratedSurface *>(decor)->markForUpdate(yes);
            }
        }
    }
    void lineFlagsChanged(Line &line, dint oldFlags) override {
        LOG_AS("ClientSubsector");
        line.forAllSides([this, &oldFlags] (LineSide &side) {
            if (side.sectorPtr() == &self().sector()) {
                if ((side.line().flags() & DDLF_DONTPEGTOP) != (oldFlags & DDLF_DONTPEGTOP)) {
                    if (auto *decor = side.top().decorationState()) {
                        static_cast<DecoratedSurface *>(decor)->markForUpdate();
                    }
                }
                if ((side.line().flags() & DDLF_DONTPEGBOTTOM) != (oldFlags & DDLF_DONTPEGBOTTOM)) {
                    if (auto *decor = side.bottom().decorationState()) {
                        static_cast<DecoratedSurface *>(decor)->markForUpdate();
                    }
                }
            }
            return LoopContinue;
        });
    }
    void materialDimensionsChanged(Material &material) override {
        LOG_AS("ClientSubsector");
        markDependentSurfacesForRedecoration(material);
    }
    void materialAnimatorDecorationStageChanged(MaterialAnimator &animator) override {
        LOG_AS("ClientSubsector");
        markDependentSurfacesForRedecoration(animator.material());
    }
    void planeBeingDeleted(Plane const &plane) override {
        LOG_AS("ClientSubsector");
        if (&plane == linkedPlane(plane.indexInSector())) {
            unlinkVisPlane(plane.indexInSector());
            decorSurfaces.remove(plane.surfacePtr());
        }
    }
    void planeHeightChanged(Plane &plane) override {
        LOG_AS("ClientSubsector");
        relinkVisPlanes();
        /*
        self().forAllEdgeLoops([] (ClEdgeLoop &loop) {
            loop.fixSurfacesMissingMaterials();
            return LoopContinue;
        });
        */
        markDependentSurfacesForRedecoration(plane);
        bool const planeIsInterior = (&plane == &self().visPlane(plane.indexInSector()));
        if (planeIsInterior) {
            needReverbUpdate = true;
            DoomsdayApp::players().forAll([this] (Player &plr) {
                ddplayer_t const &ddpl = plr.publicData();
                if (plr.isInGame()
                    && (ddpl.flags & DDPF_CAMERA)
                    && Mobj_SubsectorPtr(*ddpl.mo) == thisPublic
                    && (   ddpl.mo->origin[2] > self().visCeiling().height() - 4
                        || ddpl.mo->origin[2] < self().visFloor  ().height())) {
                    plr.as<ClientPlayer>().inVoid = true;
                }
                return LoopContinue;
            });
#if 0
            if (!::ddMapSetup && ::useBias) {
                self().forAllSubspaces([this, &plane] (ConvexSubspace &subspace) {
                    if (Shard *shard = self().findShard(subspace, plane.indexInSector())) {
                        shard->updateBiasAfterMove();
                    }
                    HEdge *hedge = subspace.poly().hedge();
                    do {
                        updateBiasForWallSectionsAfterGeometryMove(hedge);
                    } while ((hedge = &hedge->next()) != subspace.poly().hedge());
                    return subspace.forAllExtraMeshes([this] (Mesh &mesh) {
                        for (HEdge *hedge : mesh.hedges()) {
                            updateBiasForWallSectionsAfterGeometryMove(hedge);
                        }
                        return LoopContinue;
                    });
                });
            }
#endif
        }
    }
    void planeHeightSmoothedChanged(Plane &plane) override {
        LOG_AS("ClientSubsector");
        relinkVisPlanes();
        markDependentSurfacesForRedecoration(plane);
    }
#if 0
    void sectorLightLevelChanged(Sector &DENG2_DEBUG_ONLY(changed)) {
        DENG2_ASSERT(&changed == &self().sector());
        LOG_AS("ClientSubsector");
        if (self().sector().map().hasLightGrid()) {
            self().sector().map().lightGrid().blockLightSourceChanged(thisPublic);
        }
    }
    void sectorLightColorChanged(Sector &DENG2_DEBUG_ONLY(changed)) {
        DENG2_ASSERT(&changed == &self().sector());
        LOG_AS("ClientSubsector");
        if (self().sector().map().hasLightGrid()) {
            self().sector().map().lightGrid().blockLightSourceChanged(thisPublic);
        }
    }
#endif
    void surfaceMaterialChanged(Surface &surface) override {
        LOG_AS("ClientSubsector");
        if (auto *ds = static_cast<DecoratedSurface *>(surface.decorationState())) {
            ds->clear();
        }
        if (hasDecoratedMaterial(surface)) {
            auto &ds = allocDecorationState(surface);
            ds.markForUpdate();
        }
        observeMaterial(surface.materialPtr());
    }
    void surfaceOriginChanged(Surface &surface) override {
        LOG_AS("ClientSubsector");
        if (surface.hasMaterial()) {
            allocDecorationState(surface).markForUpdate();
        }
    }
    void surfaceOriginSmoothedChanged(Surface &surface) override {
        LOG_AS("ClientSubsector");
        if (surface.hasMaterial()) {
            allocDecorationState(surface).markForUpdate();
        }
    }
    void subsectorBeingDeleted(Subsector const &subsec) override {
        if (visPlaneLinks[Sector::Floor  ].current == &subsec) unlinkVisPlane(Sector::Floor);
        if (visPlaneLinks[Sector::Ceiling].current == &subsec) unlinkVisPlane(Sector::Ceiling);
    }
};
ClientSubsector::ClientSubsector(QVector<ConvexSubspace *> const &subspaces)
    : Subsector(subspaces)
    , d(new Impl(this)) {
    forAllSubspaces([this] (ConvexSubspace &subspace) {
        HEdge *hedge = subspace.poly().hedge();
        do {
            if (hedge->hasMapElement()) {
                LineSide &front = hedge->mapElementAs<LineSideSegment>().lineSide();
                front.line().audienceForFlagsChange += d;
                front.forAllSurfaces([this] (Surface &surface) {
                    d->observeSurface(&surface);
                    d->observeMaterial(surface.materialPtr());
                    return LoopContinue;
                });
                if (front.back().hasSector()) {
                    Sector &backsec = front.back().sector();
                    d->observePlane(&backsec.floor());
                    d->observePlane(&backsec.ceiling());
                }
            }
        } while ((hedge = &hedge->next()) != subspace.poly().hedge());
        return LoopContinue;
    });
    Plane *floor = &sector().floor();
    d->observePlane(floor);
    d->observeSurface(&floor->surface());
    d->observeMaterial(floor->surface().materialPtr());
    Plane *ceiling = &sector().ceiling();
    d->observePlane(ceiling);
    d->observeSurface(&ceiling->surface());
    d->observeMaterial(ceiling->surface().materialPtr());
#if 0
    sector().audienceForLightLevelChange() += d;
    sector().audienceForLightColorChange() += d;
#endif
}
String ClientSubsector::description() const {
    auto desc = String(    _E(l) "%1: " _E(.) _E(i) "Sector %2%3" _E(.)
                       " " _E(l) "%4: " _E(.) _E(i) "Sector %5%6" _E(.))
                    .arg(Sector::planeIdAsText(Sector::Floor  ).upperFirstChar())
                    .arg(visFloor  ().sector().indexInMap())
                    .arg(&visFloor  () != &sector().floor  () ? " (mapped)" : "")
                    .arg(Sector::planeIdAsText(Sector::Ceiling).upperFirstChar())
                    .arg(visCeiling().sector().indexInMap())
                    .arg(&visCeiling() != &sector().ceiling() ? " (mapped)" : "");
    if (d->boundaryData) {
        desc += String(_E(D) "\nEdge loops (%1):" _E(.)).arg(edgeLoopCount());
        dint index = 0;
        forAllEdgeLoops([&desc, &index] (ClEdgeLoop const &loop) {
            desc += String("\n[%1]: ").arg(index) + _E(>) + loop.description() + _E(<);
            index += 1;
            return LoopContinue;
        });
    }
    if (hasDecorations()) {
        desc += String(_E(D) "\nDecorations:" _E(.));
        dint decorIndex = 0;
        for (Surface *surface : d->decorSurfaces) {
            for (Decoration *decor : static_cast<Impl::DecoratedSurface *>
                                     (surface->decorationState())->decorations) {
                desc += String("\n[%1]: ").arg(decorIndex) + _E(>) + decor->description() + _E(<);
                decorIndex += 1;
            }
        }
    }
    DENG2_DEBUG_ONLY(
        desc.prepend(String(_E(b) "ClientSubsector " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
    )
    return Subsector::description() + "\n" + desc;
}
String ClientSubsector::edgeLoopIdAsText(dint loopId) // static
{
    switch (loopId) {
    case OuterLoop: return "outer";
    case InnerLoop: return "inner";
    default:
        DE_ASSERT(nullptr != "ClientSubsector::edgeLoopIdAsText: Invalid loopId");
        throw Error("ClientSubsector::edgeLoopIdAsText", "Unknown loop ID " + QString::number(loopId));
    }
}
dint ClientSubsector::edgeLoopCount() const {
    d->initBoundaryDataIfNeeded();
    return (bool(d->boundaryData->outerLoop) ? 1 : 0) + d->boundaryData->innerLoops.count();
}
LoopResult ClientSubsector::forAllEdgeLoops(const std::function<LoopResult (ClEdgeLoop &)> &func) {
    d->initBoundaryDataIfNeeded();
    DENG2_ASSERT(bool(d->boundaryData->outerLoop)); {
        if (auto result = func(*d->boundaryData->outerLoop))
            return result;
    }
    for (ClEdgeLoop *loop : d->boundaryData->innerLoops) {
        if (auto result = func(*loop))
            return result;
    }
    return LoopContinue;
}
LoopResult ClientSubsector::forAllEdgeLoops(const std::function<LoopResult (ClEdgeLoop const &)> &func) const {
    d->initBoundaryDataIfNeeded();
    DENG2_ASSERT(bool(d->boundaryData->outerLoop)); {
        if (auto result = func(*d->boundaryData->outerLoop))
            return result;
    }
    for (ClEdgeLoop *loop : d->boundaryData->innerLoops) {
        if (auto result = func(*loop))
            return result;
    }
    return LoopContinue;
}
bool ClientSubsector::hasSkyPlane(dint planeIndex) const {
    if (planeIndex < 0) {
        for (dint i = 0; i < sector().planeCount(); ++i) {
            if (visPlane(i).surface().hasSkyMaskedMaterial())
                return true;
        }
        return false;
    }
    else {
        return visPlane(planeIndex).surface().hasSkyMaskedMaterial();
    }
}
bool ClientSubsector::hasSkyFloor() const {
    return hasSkyPlane(Sector::Floor);
}
bool ClientSubsector::hasSkyCeiling() const {
    return hasSkyPlane(Sector::Ceiling);
}
void ClientSubsector::linkVisPlane(int planeIndex, ClientSubsector &target, VisPlaneLinkMode linkMode) {
    DE_ASSERT(planeIndex >= Sector::Floor && planeIndex <= Sector::Ceiling);
    d->visPlaneLinks[planeIndex].target   = &target;
    d->visPlaneLinks[planeIndex].linkMode = linkMode;
    d->relinkVisPlanes();
}
dint ClientSubsector::visPlaneCount() const {
    return sector().planeCount();
}
Plane &ClientSubsector::visPlane(dint planeIndex) {
    return const_cast<Plane &>(const_cast<ClientSubsector const *>(this)->visPlane(planeIndex));
}
Plane const &ClientSubsector::visPlane(dint planeIndex) const {
    if (planeIndex >= Sector::Floor && planeIndex <= Sector::Ceiling) {
#if 0
        if (d->needRemapVisPlanes()) {
        }
#endif
        ClientSubsector *mapping = d->visPlaneLinks[planeIndex].current;
        if (mapping && mapping != this) {
            return mapping->visPlane(planeIndex);
        }
    }
    return sector().plane(planeIndex);
}
LoopResult ClientSubsector::forAllVisPlanes(std::function<LoopResult (Plane &)> func) {
    for (dint i = 0; i < visPlaneCount(); ++i) {
        if (auto result = func(visPlane(i)))
            return result;
    }
    return LoopContinue;
}
LoopResult ClientSubsector::forAllVisPlanes(std::function<LoopResult (Plane const &)> func) const {
    for (dint i = 0; i < visPlaneCount(); ++i) {
        if (auto result = func(visPlane(i)))
            return result;
    }
    return LoopContinue;
}
bool ClientSubsector::isHeightInVoid(ddouble height) const {
    if (visCeiling().surface().hasSkyMaskedMaterial()) {
        ClSkyPlane const &skyCeil = sector().map().skyCeiling();
        if (skyCeil.height() < DDMAXFLOAT && height > skyCeil.height())
            return true;
    }
    else if (height > visCeiling().heightSmoothed()) {
        return true;
    }
    if (visFloor().surface().hasSkyMaskedMaterial()) {
        ClSkyPlane const &skyFloor = sector().map().skyFloor();
        if (skyFloor.height() > DDMINFLOAT && height < skyFloor.height())
            return true;
    }
    else if (height < visFloor().heightSmoothed()) {
        return true;
    }
    return false;  // Not in the void.
}
bool ClientSubsector::hasWorldVolume(bool useSmoothedHeights) const {
    auto const currentFrame = R_FrameCount();
    if (d->validFrame != currentFrame) {
        d->validFrame = currentFrame;
        if (useSmoothedHeights) {
            d->hasWorldVolumeInValidFrame = visCeiling().heightSmoothed() - visFloor().heightSmoothed() > 0;
        }
        else {
            d->hasWorldVolumeInValidFrame = sector().ceiling().height() - sector().floor().height() > 0;
        }
    }
    return d->hasWorldVolumeInValidFrame;
}
void ClientSubsector::markReverbDirty(bool yes) {
    d->needReverbUpdate = yes;
}
ClientSubsector::AudioEnvironment const &ClientSubsector::reverb() const {
    if (d->needReverbUpdate) {
        d->updateReverb();
    }
    return d->reverb;
}
#if 0
void ClientSubsector::markVisPlanesDirty() {
    d->maybeInvalidateMapping(Sector::Floor);
    d->maybeInvalidateMapping(Sector::Ceiling);
}
#endif
ClientSubsector::LightId ClientSubsector::lightSourceId() const {
    return LightId(sector().indexInMap());
}
Vector3f ClientSubsector::lightSourceColorf() const {
    if (Rend_SkyLightIsEnabled() && hasSkyPlane()) {
        return Rend_SkyLightColor();
    }
    return sector().lightColor();
}
dfloat ClientSubsector::lightSourceIntensity(Vector3d const &/*viewPoint*/) const
{
    return sector().lightLevel();
}
dint ClientSubsector::blockLightSourceZBias() {
    dint height      = dint(visCeiling().height() - visFloor().height());
    bool hasSkyFloor = visFloor().surface().hasSkyMaskedMaterial();
    bool hasSkyCeil  = visCeiling().surface().hasSkyMaskedMaterial();
    if (hasSkyFloor && !hasSkyCeil) {
        return -height / 6;
    }
    if (!hasSkyFloor && hasSkyCeil) {
        return height / 6;
    }
    if (height > 100) {
        return (height - 100) / 2;
    }
    return 0;
}
#if 0
void ClientSubsector::applyBiasChanges(QBitArray &allChanges) {
    DENG2_FOR_EACH(Impl::GeometryGroups, g, d->geomGroups) {
        for (Impl::GeometryData *gdata : g.value()) {
            DENG2_ASSERT(bool(gdata->shard));
            gdata->shard->biasTracker().applyChanges(allChanges);
        }
    }
}
#endif
#if 0
static dint countIlluminationPoints(MapElement &mapElement, dint DENG2_DEBUG_ONLY(group)) {
    switch (mapElement.type()) {
    case DMU_SUBSPACE: {
        auto &space = mapElement.as<ConvexSubspace>();
        DENG2_ASSERT(group >= 0 && group < space.subsector().sector().planeCount()); // sanity check
        return space.fanVertexCount(); }
    case DMU_SEGMENT:
        DENG2_ASSERT(group >= 0 && group <= LineSide::Top); // sanity check
        return 4;
    default:
        throw Error("ClientSubsector::countIlluminationPoints", "Invalid MapElement type");
    }
    return 0;
}
Shard &ClientSubsector::shard(MapElement &mapElement, dint geomId) {
    auto *gdata = d->geomData(mapElement, geomId, true /*create*/);
    if (!gdata->shard) {
        gdata->shard.reset(new Shard(/*countIlluminationPoints(mapElement, geomId),*/ this));
    }
    return *gdata->shard;
}
Shard *ClientSubsector::findShard(MapElement &mapElement, dint geomId) {
    if (auto *gdata = d->geomData(mapElement, geomId)) {
        return gdata->shard.get();
    }
    return nullptr;
}
/**
 * @todo This could be enhanced so that only the lights on the right side of the
 * surface are taken into consideration.
 */
bool ClientSubsector::updateBiasContributors(Shard *shard) {
    if (Impl::GeometryData *gdata = d->geomDataForShard(shard)) {
        Map const &map = sector().map();
        BiasTracker &tracker = shard->biasTracker();
        tracker.clearContributors();
        switch (gdata->mapElement->type()) {
        case DMU_SUBSPACE: {
            auto &subspace         = gdata->mapElement->as<ConvexSubspace>();
            Plane const &plane     = visPlane(gdata->geomId);
            Surface const &surface = plane.surface();
            Vector3d const surfacePoint(subspace.poly().center(), plane.heightSmoothed());
            map.forAllBiasSources([&tracker, &subspace, &surface, &surfacePoint] (BiasSource &source) {
                if (source.intensity() <= 0)
                    return LoopContinue;
                Vector3d sourceToSurface = (source.origin() - surfacePoint).normalize();
                ddouble distance = 0;
                HEdge *baseNode = subspace.poly().hedge();
                HEdge *node = baseNode;
                do {
                    ddouble len = (Vector2d(source.origin()) - node->origin()).length();
                    if (node == baseNode || len < distance)
                        distance = len;
                } while ((node = &node->next()) != baseNode);
                if (sourceToSurface.dot(surface.normal()) < 0)
                    return LoopContinue;
                tracker.addContributor(&source, source.evaluateIntensity() / de::max(distance, 1.0));
                return LoopContinue;
            });
            break; }
        case DMU_SEGMENT: {
            auto &seg              = gdata->mapElement->as<LineSideSegment>();
            Surface const &surface = seg.lineSide().middle();
            Vector2d const &from   = seg.hedge().origin();
            Vector2d const &to     = seg.hedge().twin().origin();
            Vector2d const center  = (from + to) / 2;
            map.forAllBiasSources([&tracker, &surface, &from, &to, &center] (BiasSource &source) {
                if (source.intensity() <= 0)
                    return LoopContinue;
                Vector3d sourceToSurface = (source.origin() - center).normalize();
                ddouble distance = 0;
                for (dint k = 0; k < 2; ++k) {
                    ddouble len = (Vector2d(source.origin()) - (!k? from : to)).length();
                    if (k == 0 || len < distance)
                        distance = len;
                }
                if (sourceToSurface.dot(surface.normal()) < 0)
                    return LoopContinue;
                tracker.addContributor(&source, source.evaluateIntensity() / de::max(distance, 1.0));
                return LoopContinue;
            });
            break; }
        default:
            throw Error("ClientSubsector::updateBiasContributors", "Invalid MapElement type");
        }
        return true;
    }
    return false;
}
duint ClientSubsector::biasLastChangeOnFrame() const {
    return sector().map().biasLastChangeOnFrame();
}
#endif
void ClientSubsector::decorate() {
    LOG_AS("ClientSubsector::decorate");
    if (!hasDecorations()) return;
    forAllEdgeLoops([this/*, &decorateFunc*/] (ClEdgeLoop const &loop)
    {
        SubsectorCirculator it(&loop.first());
        do {
            if (it->hasMapElement()) // BSP errors may fool the circulator wrt interior edges -ds
            {
                LineSide &side = it->mapElementAs<LineSideSegment>().lineSide();
                if (side.hasSections()) {
                    for (dint i = LineSide::Middle; i <= LineSide::Top; ++i) {
                        d->decorate(side.surface(i));
                    }
                }
            }
        }
        while (&it.next() != &loop.first());
        return LoopContinue;
    });
    d->decorate(visFloor  ().surface());
    d->decorate(visCeiling().surface());
}
bool ClientSubsector::hasDecorations() const {
    return !d.getConst()->decorSurfaces.isEmpty();
    /*
    for (Surface *surface : d.getConst()->decorSurfaces) {
        if (!static_cast<Impl::DecoratedSurface const *>
                (surface->decorationState())->decorations.isEmpty()) {
            return true;
        }
    }
    return false;*/
}
void ClientSubsector::generateLumobjs() {
    world::Map &map = sector().map();
    for (Surface *surface : d.getConst()->decorSurfaces) {
        for (Decoration *decor : static_cast<Impl::DecoratedSurface *>(surface->decorationState())->decorations) {
            if (auto const *lightDecor = maybeAs<LightDecoration>(decor)) {
                if (Lumobj *lum = lightDecor->generateLumobj()) {
                    map.addLumobj(lum);
                }
            }
        }
    }
}
void ClientSubsector::markForDecorationUpdate(bool yes) {
    for (Surface *surface : d.getConst()->decorSurfaces) {
        if (auto *decor = static_cast<Impl::DecoratedSurface *>(surface->decorationState())) {
            decor->markForUpdate(yes);
        }
    }
}
} // namespace world
thinker_s *ClPlaneMover::newThinker(Plane &plane, coord_t dest, float speed) // static
{
    Thinker th(Thinker::AllocateMemoryZone);
    th.setData(new ClPlaneMover(plane, dest, speed));
    thinker_s *ptr = th.take();
    plane.map().thinkers().add(*ptr, false /* not public */);
    LOGDEV_MAP_XVERBOSE("New mover %p", ptr);
    if (fequal(speed, 0)) {
        THINKER_DATA(*ptr, ClPlaneMover).think();
    }
    return ptr;
}
ClPlaneMover::ClPlaneMover(Plane &plane, coord_t dest, float speed)
    : _plane       (&plane)
    , _destination ( dest )
    , _speed       ( speed) {
    if (_destination < P_GetDoublep(_plane, DMU_HEIGHT)) {
        _speed = -_speed;
    }
    P_SetDoublep(_plane, DMU_TARGET_HEIGHT, _destination);
    P_SetFloatp(_plane, DMU_SPEED, _speed);
    _plane->addMover(*this);
}
ClPlaneMover::~ClPlaneMover() {
    _plane->removeMover(*this);
}
void ClPlaneMover::think() {
    LOG_AS("ClPlaneMover::think");
    if (!Cl_GameReady()) return;
    DENG2_ASSERT(_plane != 0);
    bool const freeMove = ClPlayer_IsFreeToMove(consolePlayer);
    bool remove = false;
    coord_t const original = P_GetDoublep(_plane, DMU_HEIGHT);
    if (de::abs(_speed) > 0 && de::abs(_destination - original) > de::abs(_speed)) {
        P_SetDoublep(_plane, DMU_HEIGHT, original + _speed);
    }
    else {
        P_SetDoublep(_plane, DMU_HEIGHT, _destination);
        remove = true;
    }
    LOGDEV_MAP_XVERBOSE_DEBUGONLY("plane height %f in sector #%i",
            P_GetDoublep(_plane, DMU_HEIGHT)
            << _plane->sector().indexInMap());
    if (gx.SectorHeightChangeNotification) {
        gx.SectorHeightChangeNotification(_plane->sector().indexInMap());
    }
    if (freeMove != ClPlayer_IsFreeToMove(consolePlayer)) {
        LOG_MAP_VERBOSE("move blocked in sector #%i, undoing move")
                << _plane->sector().indexInMap();
        P_SetDoublep(_plane, DMU_HEIGHT, original);
        if (gx.SectorHeightChangeNotification) {
            gx.SectorHeightChangeNotification(_plane->sector().indexInMap());
        }
    }
    else {
        if (remove) {
            LOG_MAP_VERBOSE("finished in sector #%i")
                    << _plane->sector().indexInMap();
            P_SetDoublep(_plane, DMU_SPEED, 0);
            _plane->map().thinkers().remove(thinker()); // we get deleted
        }
    }
}
thinker_s *ClPolyMover::newThinker(Polyobj &polyobj, bool moving, bool rotating) // static
{
    if (ClPolyMover *mover = polyobj.data().mover()) {
        mover->_move   = moving;
        mover->_rotate = rotating;
        return &mover->thinker();
    }
    Thinker th(Thinker::AllocateMemoryZone);
    th.setData(new ClPolyMover(polyobj, moving, rotating));
    thinker_s *ptr = th.take();
    polyobj.map().thinkers().add(*ptr, false /*not public*/);
    LOGDEV_MAP_XVERBOSE("New polymover %p for polyobj #%i.", ptr << polyobj.indexInMap());
    return ptr;
}
ClPolyMover::ClPolyMover(Polyobj &pobj, bool moving, bool rotating)
    : _polyobj (&pobj    )
    , _move    ( moving  )
    , _rotate  ( rotating) {
    _polyobj->data().addMover(*this);
}
ClPolyMover::~ClPolyMover() {
    _polyobj->data().removeMover(*this);
}
void ClPolyMover::think() {
    LOG_AS("ClPolyMover::think");
    Polyobj *po = _polyobj;
    if (_move) {
        Vector2d delta = Vector2d(po->dest) - Vector2d(po->origin);
        ddouble dist = M_ApproxDistance(delta.x, delta.y);
        if (dist <= po->speed || de::fequal(po->speed, 0)) {
            _move = false;
        }
        else {
            delta = (delta / dist) * po->speed;
        }
        po->move(delta);
    }
    if (_rotate) {
        int dist = po->destAngle - po->angle;
        int speed = po->angleSpeed;
        if (!po->angleSpeed || ABS(dist >> 2) <= ABS(speed >> 2)) {
            LOGDEV_MAP_XVERBOSE("Mover %p reached end of turn, destAngle=%i",
                                &thinker() << po->destAngle);
            _rotate = false;
        }
        else {
            dist = /*FIX2FLT((int)*/ po->angleSpeed;
        }
        po->rotate(dist);
    }
    if (!_move && !_rotate) {
        _polyobj->map().thinkers().remove(thinker()); // we get deleted
    }
}
namespace world {
DENG2_PIMPL_NOREF(ClSkyPlane) {
    bool isCeiling = false; ///< @c true if this is the ceiling; otherwise the floor.
    ddouble height = 0;
    Impl(bool ceiling, ddouble defaultHeight)
        : isCeiling(ceiling), height(defaultHeight)
    {}
    DENG2_PIMPL_AUDIENCE(HeightChange)
};
DENG2_AUDIENCE_METHOD(ClSkyPlane, HeightChange)
ClSkyPlane::ClSkyPlane(bool isCeiling, ddouble defaultHeight)
    : d(new Impl(isCeiling, defaultHeight))
{}
bool ClSkyPlane::isCeiling() const {
    return d->isCeiling;
}
bool ClSkyPlane::isFloor() const {
    return !d->isCeiling;
}
ddouble ClSkyPlane::height() const {
    return d->height;
}
void ClSkyPlane::setHeight(ddouble newHeight) {
    if (d->height != newHeight) {
        d->height = newHeight;
        DENG2_FOR_AUDIENCE2(HeightChange, i) i->clSkyPlaneHeightChanged(*this);
    }
}
} // namespace world
