﻿/** @file net_demo.cpp  Handling of demo recording and playback.
 *
 * Opening of, writing to, reading from and closing of demo files.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "network/net_demo.h"
#include <doomsday/doomsdayapp.h>
#include <doomsday/console/cmd.h>
#include <doomsday/filesys/fs_util.h>
#include "client/cl_player.h"
#include "api_filesys.h"
#include "api_player.h"
#include "network/net_main.h"
#include "network/net_buf.h"
#include "render/rend_main.h"
#include "render/viewports.h"
#include "world/p_object.h"
#include "world/p_players.h"
/** @file serverlink.cpp  Network connection to a server.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "network/serverlink.h"
#include "network/masterserver.h"
#include "network/net_main.h"
#include "network/net_buf.h"
#include "network/net_demo.h"
#include "network/net_event.h"
#include "network/protocol.h"
#include "client/cl_def.h"
#include "ui/clientwindow.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/dialogs/filedownloaddialog.h"
#include "dd_def.h"
#include "dd_main.h"
#include "clientapp.h"
#include <de/Async>
#include <de/BlockPacket>
#include <de/ByteRefArray>
#include <de/ByteSubArray>
#include <de/GuiApp>
#include <de/Message>
#include <de/MessageDialog>
#include <de/PackageLoader>
#include <de/RecordValue>
#include <de/Socket>
#include <de/data/json.h>
#include <de/memory.h>
#include <de/shell/ServerFinder>
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <QElapsedTimer>
#include <QTimer>
/** @file sys_network.cpp  Low-level network socket routines (deprecated).
 *
 * @todo Remove this source file entirely once dependent code is revised.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "network/sys_network.h"
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include "clientapp.h"
#include "network/net_buf.h"
using namespace de;
#define DEMOTIC SECONDS_TO_TICKS(demoTime)
#define LCAMF_ONGROUND      0x1
#define LCAMF_FOV           0x2  ///< FOV has changed (short).
#define LCAMF_CAMERA        0x4  ///< Camera mode.
#pragma pack(1)
struct demopacket_header_t {
    dushort length;
};
#pragma pack()
extern dfloat netConnectTime;
static char const *demoPath = "/home/demo/";
#if 0
LZFILE *playdemo;
#endif
dint playback;
dint viewangleDelta;
dfloat lookdirDelta;
dfloat posDelta[3];
dfloat demoFrameZ, demoZ;
dd_bool demoOnGround;
static DemoTimer readInfo;
static dfloat startFOV;
static dint demoStartTic;
void Demo_WriteLocalCamera(dint plrNum);
void Demo_Init() {
}
/**
 * Open a demo file and begin recording.
 * Returns @c false if the recording can't be begun.
 */
dd_bool Demo_BeginRecording(char const * /*fileName*/, dint /*plrNum*/)
{
    return false;
#if 0
    client_t *cl  = &::clients[plrNum];
    player_t *plr = &DD_Player(plrNum);
    if(cl->recording || ::playback || (::isDedicated && !plrNum) || !plr->shared.inGame)
        return false;
    ddstring_t buf; Str_Appendf(Str_InitStd(&buf), "%s%s", ::demoPath, fileName);
    F_ExpandBasePath(&buf, &buf);
    F_ToNativeSlashes(&buf, &buf);
    cl->demo = lzOpen(Str_Text(&buf), "wp");
    Str_Free(&buf);
    if(!cl->demo) {
        return false;  // Couldn't open it!
    }
    cl->recording    = true;
    cl->recordPaused = false;
    ::writeInfo[plrNum].first       = true;
    ::writeInfo[plrNum].canwrite    = false;
    ::writeInfo[plrNum].cameratimer = 0;
    ::writeInfo[plrNum].fov         = -1;  // Must be written in the first packet.
    if(::isServer) {
        ::clients[0].lastTransmit = -1;
        ::allowSending = false;
        Sv_Handshake(plrNum, false);
        ::allowSending = true;
    }
    else {
        Cl_SendHello();
    }
    return true;
#endif
}
void Demo_PauseRecording(dint playerNum) {
    DENG2_ASSERT(playerNum >= 0 && playerNum < DDMAXPLAYERS);
    auto &cl = *DD_Player(playerNum);
    if(!cl.recording || cl.recordPaused)
        return;
    cl.demoTimer().pausetime = SECONDS_TO_TICKS(demoTime);
    cl.recordPaused = true;
}
/**
 * Resumes a paused recording.
 */
void Demo_ResumeRecording(dint playerNum) {
    DENG2_ASSERT(playerNum >= 0 && playerNum < DDMAXPLAYERS);
    auto &cl = *DD_Player(playerNum);
    if(!cl.recording || !cl.recordPaused)
        return;
    Demo_WriteLocalCamera(playerNum);
    cl.recordPaused = false;
    cl.demoTimer().begintime += DEMOTIC - cl.demoTimer().pausetime;
}
/**
 * Stop recording a demo.
 */
void Demo_StopRecording(dint playerNum) {
    DENG2_ASSERT(playerNum >= 0 && playerNum < DDMAXPLAYERS);
    auto &cl = *DD_Player(playerNum);
    if(!cl.recording) return;
    cl.recording = false;
}
void Demo_WritePacket(dint playerNum) {
    DENG_UNUSED(playerNum);
#if 0
    if(playerNum < 0) {
        Demo_BroadcastPacket();
        return;
    }
    DENG2_ASSERT(playerNum >= 0 && playerNum < DDMAXPLAYERS);
    auto &cl = *DD_Player(playerNum);
    DemoTimer &inf = cl.demoTimer();
    if(!cl.recording)
        return;
    if(!inf.canwrite) {
        if(::netBuffer.msg.type != PSV_HANDSHAKE)
            return;
        inf.canwrite = true;
    }
    if(cl.recordPaused) {
        if(::netBuffer.msg.type == PSV_SOUND ||
           ::netBuffer.msg.type == DDPT_MESSAGE)
            return;
    }
    LZFILE *file = cl.demo;
    if(!file) App_Error("Demo_WritePacket: No demo file!\n");
    byte ptime;
    if(!inf.first) {
        ptime = (cl.recordPaused ? inf.pausetime : DEMOTIC)
              - inf.begintime;
    }
    else {
        ptime = 0;
        inf.first     = false;
        inf.begintime = DEMOTIC;
    }
    lzWrite(&ptime, 1, file);
    demopacket_header_t hdr;  // The header.
    if(::netBuffer.length >= sizeof(hdr.length))
        App_Error("Demo_WritePacket: Write buffer too large!\n");
    hdr.length = (dushort) 1 + ::netBuffer.length;
    lzWrite(&hdr, sizeof(hdr), file);
    lzPutC(::netBuffer.msg.type, file);
    lzWrite(::netBuffer.msg.data, (long) netBuffer.length, file);
#endif
}
void Demo_BroadcastPacket() {
    for(dint i = 0; i < DDMAXPLAYERS; ++i) {
        Demo_WritePacket(i);
    }
}
dd_bool Demo_BeginPlayback(char const *fileName) {
    if(::playback) return false;
    if(::netGame || ::isClient) return false;
    for(dint i = 0; i < DDMAXPLAYERS; ++i) {
        if(DD_Player(i)->recording)
            return false;
    }
    ddstring_t buf; Str_Set(Str_InitStd(&buf), fileName);
    if(!F_IsAbsolute(&buf)) {
        Str_Prepend(&buf, demoPath);
    }
    F_ExpandBasePath(&buf, &buf);
    F_ToNativeSlashes(&buf, &buf);
#if 0
    playdemo = lzOpen(Str_Text(&buf), "rp");
#endif
    Str_Free(&buf);
#if 0
    if(!playdemo)
#endif
        return false;
    ::playback       = true;
    ::isServer       = false;
    ::isClient       = true;
    ::readInfo.first = true;
    ::viewangleDelta = 0;
    ::lookdirDelta   = 0;
    ::demoFrameZ     = 1;
    ::demoZ          = 0;
    ::startFOV       = 95; //Rend_FieldOfView();
    ::demoStartTic   = DEMOTIC;
    std::memset(::posDelta, 0, sizeof(::posDelta));
    /*if(ArgCheck("-timedemo"))
    {
        ::r_framecounter = 0;
    }*/
    return true;
}
void Demo_StopPlayback() {
    if(!::playback) return;
#if 0
    LOG_MSG("Demo was %.2f seconds (%i tics) long.")
        << ((DEMOTIC - ::demoStartTic) / dfloat( TICSPERSEC ))
        << (DEMOTIC - ::demoStartTic);
    ::playback = false;
    lzClose(::playdemo); playdemo = nullptr;
    Net_StopGame();
    /*if(ArgCheck("-timedemo"))
    {
        diff = Sys_GetSeconds() - ::netConnectTime;
        if(!diff) diff = 1;
        LOG_MSG("Timedemo results: %i game tics in %.1f seconds", ::r_framecounter, diff);
        LOG_MSG("%f FPS", ::r_framecounter / diff);
        Sys_Quit();
    }*/
    if(CommandLine_Check("-playdemo"))
        Sys_Quit();
#endif
}
dd_bool Demo_ReadPacket() {
    return false;
#if 0
    static byte ptime;
    dint nowtime = DEMOTIC;
    if(!playback)
        return false;
    if(lzEOF(::playdemo)) {
        Demo_StopPlayback();
        DoomsdayApp::plugins().callAllHooks(HOOK_DEMO_STOP);
        return false;
    }
    if(::readInfo.first) {
        ::readInfo.first = false;
        ::readInfo.begintime = nowtime;
        ptime = lzGetC(::playdemo);
    }
    if(Net_TimeDelta(nowtime - ::readInfo.begintime, ptime) < 0)
        return false;  // Can't read yet.
    demopacket_header_t hdr;
    lzRead(&hdr, sizeof(hdr), ::playdemo);
    ::netBuffer.length = hdr.length - 1;
    ::netBuffer.player = 0; // From the server.
    ::netBuffer.msg.type = lzGetC(::playdemo);
    lzRead(::netBuffer.msg.data, (long) ::netBuffer.length, ::playdemo);
    ptime = lzGetC(::playdemo);
    return true;
#endif
}
/**
 * Writes a view angle and coords packet. Doesn't send the packet outside.
 */
void Demo_WriteLocalCamera(dint plrNum) {
    DENG2_ASSERT(plrNum >= 0 && plrNum <= DDMAXPLAYERS);
    player_t *plr    = DD_Player(plrNum);
    ddplayer_t *ddpl = &plr->publicData();
    mobj_t *mob      = ddpl->mo;
    if(!mob) return;
    Msg_Begin(plr->recordPaused ? PKT_DEMOCAM_RESUME : PKT_DEMOCAM);
    byte flags = (mob->origin[VZ] <= mob->floorZ ? LCAMF_ONGROUND : 0)  // On ground?
             /*| (::writeInfo[plrNum].fov != ::fieldOfView ? LCAMF_FOV : 0)*/;
    if(ddpl->flags & DDPF_CAMERA) {
        flags &= ~LCAMF_ONGROUND;
        flags |= LCAMF_CAMERA;
    }
    Writer_WriteByte(::msgWriter, flags);
    fixed_t x = FLT2FIX(mob->origin[VX]);
    fixed_t y = FLT2FIX(mob->origin[VY]);
    Writer_WriteInt16(::msgWriter, x >> 16);
    Writer_WriteByte(::msgWriter, x >> 8);
    Writer_WriteInt16(::msgWriter, y >> 16);
    Writer_WriteByte(::msgWriter, y >> 8);
    fixed_t z = FLT2FIX(mob->origin[VZ] + plr->viewport().current.origin.z);
    Writer_WriteInt16(::msgWriter, z >> 16);
    Writer_WriteByte(::msgWriter, z >> 8);
    Writer_WriteInt16(msgWriter, mob->angle /*ddpl->clAngle*/ >> 16); /* $unifiedangles */
    Writer_WriteInt16(msgWriter, ddpl->lookDir / 110 * DDMAXSHORT /* $unifiedangles */);
    /*if(incfov)
    {
        Writer_WriteInt16(::msgWriter, ::fieldOfView / 180 * DDMAXSHORT);
        ::writeInfo[plrNum].fov = ::fieldOfView;
    }*/
    Msg_End();
    Net_SendBuffer(plrNum, SPF_DONT_SEND);
}
/**
 * Read a view angle and coords packet. NOTE: The Z coordinate of the camera is the
 * real eye Z coordinate, not the player mobj's Z coord.
 */
void Demo_ReadLocalCamera() {
    DENG2_ASSERT(::consolePlayer >= 0 && consolePlayer < DDMAXPLAYERS);
    ddplayer_t *pl = &DD_Player(::consolePlayer)->publicData();
    mobj_t *mob    = pl->mo;
    if(!mob) return;
    dint intertics = LOCALCAM_WRITE_TICS;
    if(::netBuffer.msg.type == PKT_DEMOCAM_RESUME)
        intertics = 1;
    ::demoFrameZ += demoZ;
    dint flags = Reader_ReadByte(::msgReader);
    ::demoOnGround = (flags & LCAMF_ONGROUND) != 0;
    if(flags & LCAMF_CAMERA)
        pl->flags |= DDPF_CAMERA;
    else
        pl->flags &= ~DDPF_CAMERA;
    ::posDelta[VX] =
        (FIX2FLT((Reader_ReadInt16(::msgReader) << 16) + (Reader_ReadByte(::msgReader) << 8)) - mob->origin[VX]) / intertics;
    ::posDelta[VY] =
        (FIX2FLT((Reader_ReadInt16(::msgReader) << 16) + (Reader_ReadByte(::msgReader) << 8)) - mob->origin[VY]) / intertics;
    dfloat z = FIX2FLT((Reader_ReadInt16(::msgReader) << 16) + (Reader_ReadByte(::msgReader) << 8));
    ::posDelta[VZ] = (z - ::demoFrameZ) / LOCALCAM_WRITE_TICS;
    dint dang    = Reader_ReadInt16(::msgReader) << 16;
    dfloat dlook = Reader_ReadInt16(::msgReader) * 110.0f / DDMAXSHORT;
    /*
    if(flags & LCAMF_FOV)
        ::fieldOfView = Reader_ReadInt16(::msgReader) * 180.0f / DDMAXSHORT;
    */
    if(intertics == 1 || demoFrameZ == 1) {
        /*pl->clAngle = dang;
        pl->clLookDir = dlook;*/
        pl->mo->angle = dang;
        pl->lookDir = dlook;
        /* $unifiedangles */
        ::viewangleDelta = 0;
        ::lookdirDelta   = 0;
    }
    else {
        ::viewangleDelta = (dang  - pl->mo->angle) / intertics;
        ::lookdirDelta   = (dlook - pl->lookDir)   / intertics;
        /* $unifiedangles */
    }
    if(::demoFrameZ == 1) {
        ::demoFrameZ = z;
        ::posDelta[VZ] = 0;
    }
    ::demoZ = 0;
    if(intertics == 1) {
        R_ResetViewer();
        ::demoFrameZ = z;
        ClPlayer_MoveLocal(::posDelta[VX], ::posDelta[VY], z, ::demoOnGround);
        de::zap(::posDelta);
    }
}
/**
 * Called once per tic.
 */
void Demo_Ticker(timespan_t /*time*/)
{
    if(!DD_IsSharpTick()) return;
    if(::playback) {
        DENG2_ASSERT(::consolePlayer >= 0 && ::consolePlayer < DDMAXPLAYERS);
        player_t   *plr  = DD_Player(::consolePlayer);
        ddplayer_t *ddpl = &plr->publicData();
        ddpl->mo->angle += ::viewangleDelta;
        ddpl->lookDir += ::lookdirDelta;
        /* $unifiedangles */
        ClPlayer_MoveLocal(::posDelta[VX], ::posDelta[VY], ::demoFrameZ + ::demoZ, ::demoOnGround);
        ::demoZ += ::posDelta[VZ];
    }
    else {
        for(dint i = 0; i < DDMAXPLAYERS; ++i) {
            player_t   &plr  = *DD_Player(i);
            ddplayer_t &ddpl = plr.publicData();
            if(ddpl.inGame && plr.recording && !plr.recordPaused &&
               ++plr.demoTimer().cameratimer >= LOCALCAM_WRITE_TICS) {
                plr.demoTimer().cameratimer = 0;
                Demo_WriteLocalCamera(i);
            }
        }
    }
}
D_CMD(PlayDemo) {
    DENG2_UNUSED2(src, argc);
    LOG_MSG("Playing demo \"%s\"...") << argv[1];
    return Demo_BeginPlayback(argv[1]);
}
D_CMD(RecordDemo) {
    DENG2_UNUSED(src);
    if(argc == 3 && ::isClient) {
        LOG_ERROR("Clients can only record the consolePlayer");
        return true;
    }
    if(::isClient && argc != 2) {
        LOG_SCR_NOTE("Usage: %s (fileName)") << argv[0];
        return true;
    }
    if(::isServer && (argc < 2 || argc > 3)) {
        LOG_SCR_NOTE("Usage: %s (fileName) (plnum)") << argv[0];
        LOG_SCR_MSG("(plnum) is the player which will be recorded.");
        return true;
    }
    dint plnum = ::consolePlayer;
    if(argc == 3) {
        plnum = String(argv[2]).toInt();
    }
    LOG_MSG("Recording demo of player %i to \"%s\"") << plnum << argv[1];
    return Demo_BeginRecording(argv[1], plnum);
}
D_CMD(PauseDemo) {
    DENG2_UNUSED(src);
    dint plnum = ::consolePlayer;
    if(argc >= 2) {
        plnum = String(argv[1]).toInt();
    }
    if(plnum < 0 || plnum >= DDMAXPLAYERS) {
        LOG_SCR_ERROR("Invalid player #%i") << plnum;
        return false;
    }
    if(!DD_Player(plnum)->recording) {
        LOG_SCR_ERROR("Not recording for player %i") << plnum;
        return false;
    }
    if(DD_Player(plnum)->recordPaused) {
        Demo_ResumeRecording(plnum);
        LOG_SCR_MSG("Demo recording of player %i resumed") << plnum;
    }
    else {
        Demo_PauseRecording(plnum);
        LOG_SCR_MSG("Demo recording of player %i paused") << plnum;
    }
    return true;
}
D_CMD(StopDemo) {
    DENG2_UNUSED(src);
    if(argc > 2) {
        LOG_SCR_NOTE("Usage: stopdemo (plrnum)");
        return true;
    }
    dint plnum = ::consolePlayer;
    if(argc == 2) {
        plnum = String(argv[1]).toInt();
    }
    if(plnum < 0 || plnum >= DDMAXPLAYERS) {
        LOG_SCR_ERROR("Invalid player #%i") << plnum;
        return false;
    }
    if(!::playback && !DD_Player(plnum)->recording)
        return true;
    LOG_SCR_MSG("Demo %s of player %i stopped")
        << (DD_Player(plnum)->recording ? "recording" : "playback") << plnum;
    if(::playback) {
        Demo_StopPlayback();
        DoomsdayApp::plugins().callAllHooks(HOOK_DEMO_STOP, true);
    }
    else {
        Demo_StopRecording(plnum);
    }
    return true;
}
#if 0
/**
 * Make a demo lump.
 *
 * @todo Why does this exist? -ds
 */
D_CMD(DemoLump) {
    DENG2_UNUSED2(src, argc);
    char buf[64]; de::zap(buf);
    strncpy(buf, argv[1], 64);
    return M_WriteFile(argv[2], buf, 64);
}
#endif
void Demo_Register() {
    C_CMD_FLAGS("pausedemo",    nullptr,    PauseDemo,  CMDF_NO_NULLGAME);
    C_CMD_FLAGS("playdemo",     "s",        PlayDemo,   CMDF_NO_NULLGAME);
    C_CMD_FLAGS("recorddemo",   nullptr,    RecordDemo, CMDF_NO_NULLGAME);
    C_CMD_FLAGS("stopdemo",     nullptr,    StopDemo,   CMDF_NO_NULLGAME);
}
enum LinkState {
    None,
    Discovering,
    Pinging,
    WaitingForPong,
    QueryingMapOutline,
    WaitingForInfoResponse,
    Joining,
    WaitingForJoinResponse,
    InGame
};
static int const NUM_PINGS = 5;
DENG2_PIMPL(ServerLink) {
    std::unique_ptr<shell::ServerFinder> finder; ///< Finding local servers.
    LinkState state;
    bool fetching;
    typedef QMap<Address, shell::ServerInfo> Servers;
    Servers discovered;
    Servers fromMaster;
    QElapsedTimer pingTimer;
    QList<TimeSpan> pings;
    int pingCounter;
    std::unique_ptr<GameProfile> serverProfile; ///< Profile used when joining.
    std::function<void (GameProfile const *)> profileResultCallback;
    std::function<void (Address, GameProfile const *)> profileResultCallbackWithAddress;
    shell::PackageDownloader downloader;
    LoopCallback deferred; // for deferred actions
    Impl(Public *i, Flags flags)
        : Base(i)
        , state(None)
        , fetching(false) {
        if (flags & DiscoverLocalServers) {
            finder.reset(new shell::ServerFinder);
        }
    }
    void notifyDiscoveryUpdate() {
        DENG2_FOR_PUBLIC_AUDIENCE2(DiscoveryUpdate, i) i->linkDiscoveryUpdate(self());
        emit self().serversDiscovered();
    }
    bool handleInfoResponse(Block const &reply) {
        DENG2_ASSERT(state == WaitingForInfoResponse);
        Address svAddress = self().address();
        if (svAddress.isLocal()) svAddress.setHost(QHostAddress::LocalHost);
        self().disconnect();
        if (reply.size() >= 5 && reply.startsWith("Info\n")) {
            try {
                QVariant const response = parseJSON(String::fromUtf8(reply.mid(5)));
                std::unique_ptr<Value> rec(Value::constructFrom(response.toMap()));
                if (!is<RecordValue>(*rec)) {
                    throw Error("ServerLink::handleInfoResponse", "Failed to parse response contents");
                }
                shell::ServerInfo svInfo(*rec->as<RecordValue>().record());
                LOG_NET_VERBOSE("Discovered server at ") << svAddress;
                svAddress = Address(svAddress.host(), svInfo.port());
                svInfo.setAddress(svAddress);
                svInfo.printToLog(0, true);
                discovered.insert(svAddress, svInfo);
                LOG_NET_NOTE("%i server%s been found")
                        << discovered.size()
                        << (discovered.size() != 1 ? "s have" : " has");
                notifyDiscoveryUpdate();
                if (profileResultCallback ||
                    profileResultCallbackWithAddress) {
                    if (prepareServerProfile(svAddress)) {
                        LOG_NET_MSG("Received server's game profile from ") << svAddress;
                        deferred.enqueue([this, svAddress] () {
                            if (profileResultCallback) {
                                profileResultCallback(serverProfile.get());
                                profileResultCallback = nullptr;
                            }
                            if (profileResultCallbackWithAddress) {
                                profileResultCallbackWithAddress(svAddress, serverProfile.get());
                                profileResultCallbackWithAddress = nullptr;
                            }
                        });
                    }
                }
                return true;
            }
            catch (Error const &er) {
                LOG_NET_WARNING("Info reply from %s was invalid: %s") << svAddress << er.asText();
            }
        }
        else if (reply.size() >= 11 && reply.startsWith("MapOutline\n")) {
            try {
                shell::MapOutlinePacket outline; {
                    Block const data = Block(ByteSubArray(reply, 11)).decompressed();
                    Reader src(data);
                    src.withHeader() >> outline;
                }
                DENG2_FOR_PUBLIC_AUDIENCE2(MapOutline, i) {
                    i->mapOutlineReceived(svAddress, outline);
                }
            }
            catch (Error const &er) {
                LOG_NET_WARNING("MapOutline reply from %s was invalid: %s") << svAddress << er.asText();
            }
        }
        else {
            LOG_NET_WARNING("Reply from %s was invalid") << svAddress;
        }
        return false;
    }
    /**
     * Handles the server's response to a client's join request.
     * @return @c false to stop processing further messages.
     */
    bool handleJoinResponse(Block const &reply) {
        if (reply.size() < 5 || reply != "Enter") {
            LOG_NET_WARNING("Server refused connection");
            LOGDEV_NET_WARNING("Received %i bytes instead of \"Enter\")")
                    << reply.size();
            self().disconnect();
            return false;
        }
        state = InGame;
        handshakeReceived = false;
        allowSending = true;
        netGame = true;             // Allow sending/receiving of packets.
        isServer = false;
        isClient = true;
        gx.NetConnect(false);
        DENG2_FOR_PUBLIC_AUDIENCE2(Join, i) i->networkGameJoined();
        Cl_SendHello();
        return true;
    }
    void fetchFromMaster() {
        if (fetching) return;
        fetching = true;
        N_MAPost(MAC_REQUEST);
        N_MAPost(MAC_WAIT);
        deferred.enqueue([this] () { checkMasterReply(); });
    }
    void checkMasterReply() {
        DENG2_ASSERT(fetching);
        if (N_MADone()) {
            fetching = false;
            fromMaster.clear();
            int const count = N_MasterGet(0, 0);
            for (int i = 0; i < count; i++) {
                shell::ServerInfo info;
                N_MasterGet(i, &info);
                fromMaster.insert(info.address(), info);
            }
            notifyDiscoveryUpdate();
        }
        else {
            deferred.enqueue([this] () { checkMasterReply(); });
        }
    }
    bool prepareServerProfile(Address const &host) {
        serverProfile.reset();
        shell::ServerInfo info;
        if (!self().foundServerInfo(host, info)) {
            return false;
        }
        if (info.gameId().isEmpty() || info.packages().isEmpty()) {
            return false;
        }
        serverProfile.reset(new GameProfile(info.name()));
        serverProfile->setGame(info.gameId());
        serverProfile->setUseGameRequirements(false); // server lists all packages (that affect gameplay)
        serverProfile->setPackages(info.packages());
        return true; // profile available
    }
    Servers allFound(FoundMask const &mask) const {
        Servers all;
        if (finder && mask.testFlag(LocalNetwork)) {
            foreach (Address const &sv, finder->foundServers()) {
                all.insert(sv, finder->messageFromServer(sv));
            }
        }
        if (mask.testFlag(MasterServer)) {
            for (auto i = fromMaster.constBegin(); i != fromMaster.constEnd(); ++i) {
                all.insert(i.key(), i.value());
            }
        }
        if (mask.testFlag(Direct)) {
            for (auto i = discovered.constBegin(); i != discovered.constEnd(); ++i) {
                all.insert(i.key(), i.value());
            }
        }
        return all;
    }
    void reportError(String const &msg) {
        MessageDialog *dlg = new MessageDialog;
        dlg->setDeleteAfterDismissed(true);
        dlg->title().setText(tr("Cannot Join Game"));
        dlg->message().setText(msg);
        dlg->buttons() << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept);
        ClientWindow::main().root().addOnTop(dlg);
        dlg->open(MessageDialog::Modal);
    }
    DENG2_PIMPL_AUDIENCE(DiscoveryUpdate)
    DENG2_PIMPL_AUDIENCE(PingResponse)
    DENG2_PIMPL_AUDIENCE(MapOutline)
    DENG2_PIMPL_AUDIENCE(Join)
    DENG2_PIMPL_AUDIENCE(Leave)
};
DENG2_AUDIENCE_METHOD(ServerLink, DiscoveryUpdate)
DENG2_AUDIENCE_METHOD(ServerLink, PingResponse)
DENG2_AUDIENCE_METHOD(ServerLink, MapOutline)
DENG2_AUDIENCE_METHOD(ServerLink, Join)
DENG2_AUDIENCE_METHOD(ServerLink, Leave)
ServerLink::ServerLink(Flags flags) : d(new Impl(this, flags)) {
    if (d->finder) {
        connect(d->finder.get(), SIGNAL(updated()), this, SLOT(localServersFound()));
    }
    connect(this, SIGNAL(packetsReady()), this, SLOT(handleIncomingPackets()));
    connect(this, SIGNAL(disconnected()), this, SLOT(linkDisconnected()));
}
shell::PackageDownloader &ServerLink::packageDownloader() {
    return d->downloader;
}
void ServerLink::clear() {
    d->finder->clear();
}
void ServerLink::connectToServerAndChangeGameAsync(shell::ServerInfo info) {
    if (netGame && isClient) {
        disconnect();
    }
    acquireServerProfileAsync(info.address(),
                              [this, info] (GameProfile const *serverProfile) {
        auto &win = ClientWindow::main();
        win.glActivate();
        if (!serverProfile) {
            String const errorMsg = QString("Not enough information known about server %1")
                    .arg(info.address().asText());
            LOG_NET_ERROR("Failed to join: ") << errorMsg;
            d->reportError(errorMsg);
            return;
        }
        LOG_NET_MSG("Received server's game profile");
        LOG_NET_VERBOSE(serverProfile->toInfoSource());
        GameProfile const *joinProfile = serverProfile;
        auto const localPackages = serverProfile->game().localMultiplayerPackages();
        if (localPackages.count()) {
            for (String const &pkg : localPackages) {
                if (!PackageLoader::get().select(pkg)) {
                    String const errorMsg = tr("The configured local multiplayer "
                                               "package %1 is unavailable.")
                            .arg(Package::splitToHumanReadable(pkg));
                    LOG_NET_ERROR("Failed to join %s: ") << info.address() << errorMsg;
                    d->reportError(errorMsg);
                    return;
                }
            }
            auto &adhoc = DoomsdayApp::app().adhocProfile();
            adhoc = *serverProfile;
            adhoc.setPackages(serverProfile->packages() + localPackages);
            joinProfile = &adhoc;
        }
        d->downloader.mountServerRepository(info, [this, joinProfile, info] (filesys::Link const *) {
            LOG_RES_MSG("Received metadata about server files");
            StringList const neededPackages = joinProfile->unavailablePackages();
            LOG_RES_MSG("Packages needed to join: ")
                    << String::join(neededPackages, " ");
            auto *dlPopup = new FileDownloadDialog(d->downloader);
            dlPopup->setDeleteAfterDismissed(true);
            ClientWindow::main().root().addOnTop(dlPopup);
            dlPopup->open(DialogWidget::Modal);
            d->downloader.download(neededPackages,
                                   [this, dlPopup, joinProfile, info] () {
                auto &win = ClientWindow::main();
                win.glActivate();
                dlPopup->close();
                if (d->downloader.isCancelled()) {
                    d->downloader.unmountServerRepository();
                    return;
                }
                if (!joinProfile->isPlayable()) {
                    String const errorMsg = tr("Server's game \"%1\" is not playable on this system. "
                                               "The following packages are unavailable:\n\n%2")
                            .arg(info.gameId())
                            .arg(String::join(de::map(joinProfile->unavailablePackages(),
                                                      Package::splitToHumanReadable), "\n"));
                    LOG_NET_ERROR("Failed to join %s: ") << info.address() << errorMsg;
                    d->downloader.unmountServerRepository();
                    d->reportError(errorMsg);
                    return;
                }
                BusyMode_FreezeGameForBusyMode();
                win.taskBar().close();
                DoomsdayApp::app().changeGame(*joinProfile, DD_ActivateGameWorker);
                connectHost(info.address());
                win.glDone();
            });
        });
    });
}
void ServerLink::acquireServerProfileAsync(Address const &address,
                                           std::function<void (GameProfile const *)> resultHandler) {
    if (d->prepareServerProfile(address)) {
        d->deferred.enqueue([this, resultHandler] () {
            DENG2_ASSERT(d->serverProfile.get() != nullptr);
            resultHandler(d->serverProfile.get());
        });
    }
    else {
        AbstractLink::connectHost(address);
        d->profileResultCallback = resultHandler;
        d->state = Discovering;
        LOG_NET_MSG("Querying server %s for full status") << address;
    }
}
void ServerLink::acquireServerProfileAsync(String const &domain,
                                           std::function<void (Address, GameProfile const *)> resultHandler) {
    d->profileResultCallbackWithAddress = resultHandler;
    discover(domain);
    LOG_NET_MSG("Querying server %s for full status") << domain;
}
void ServerLink::requestMapOutline(Address const &address) {
    AbstractLink::connectHost(address);
    d->state = QueryingMapOutline;
    LOG_NET_VERBOSE("Querying server %s for map outline") << address;
}
void ServerLink::ping(Address const &address) {
    if (d->state != Pinging &&
        d->state != WaitingForPong) {
        AbstractLink::connectHost(address);
        d->state = Pinging;
    }
}
void ServerLink::connectDomain(String const &domain, TimeSpan const &timeout) {
    LOG_AS("ServerLink::connectDomain");
    AbstractLink::connectDomain(domain, timeout);
    d->state = Joining;
}
void ServerLink::connectHost(Address const &address) {
    LOG_AS("ServerLink::connectHost");
    AbstractLink::connectHost(address);
    d->state = Joining;
}
void ServerLink::linkDisconnected() {
    LOG_AS("ServerLink");
    if (d->state != None) {
        LOG_NET_NOTE("Connection to server was disconnected");
        disconnect();
    }
}
void ServerLink::disconnect() {
    if (d->state == None) return;
    LOG_AS("ServerLink::disconnect");
    if (d->state == InGame) {
        d->state = None;
        if (gx.NetDisconnect)
            gx.NetDisconnect(true);
        DENG2_FOR_AUDIENCE2(Leave, i) i->networkGameLeft();
        LOG_NET_NOTE("Link to server %s disconnected") << address();
        d->downloader.unmountServerRepository();
        AbstractLink::disconnect();
        Net_StopGame();
        if (gx.NetDisconnect)
            gx.NetDisconnect(false);
    }
    else {
        d->state = None;
        LOG_NET_NOTE("Connection attempts aborted");
        AbstractLink::disconnect();
    }
}
void ServerLink::discover(String const &domain) {
    AbstractLink::connectDomain(domain, 5 /*timeout*/);
    d->discovered.clear();
    d->state = Discovering;
}
void ServerLink::discoverUsingMaster() {
    d->fetchFromMaster();
}
bool ServerLink::isDiscovering() const {
    return (d->state == Discovering ||
            d->state == WaitingForInfoResponse ||
            d->fetching);
}
int ServerLink::foundServerCount(FoundMask mask) const {
    return d->allFound(mask).size();
}
QList<Address> ServerLink::foundServers(FoundMask mask) const {
    return d->allFound(mask).keys();
}
bool ServerLink::isFound(Address const &host, FoundMask mask) const {
    Address const addr = shell::checkPort(host);
    return d->allFound(mask).contains(addr);
}
bool ServerLink::foundServerInfo(int index, shell::ServerInfo &info, FoundMask mask) const {
    Impl::Servers const all = d->allFound(mask);
    QList<Address> const listed = all.keys();
    if (index < 0 || index >= listed.size()) return false;
    info = all[listed[index]];
    return true;
}
bool ServerLink::isServerOnLocalNetwork(Address const &host) const {
    if (!d->finder) return host.isLocal(); // Best guess...
    Address const addr = shell::checkPort(host);
    return d->finder->foundServers().contains(addr);
}
bool ServerLink::foundServerInfo(de::Address const &host, shell::ServerInfo &info, FoundMask mask) const {
    Address const addr = shell::checkPort(host);
    Impl::Servers const all = d->allFound(mask);
    if (!all.contains(addr)) return false;
    info = all[addr];
    return true;
}
Packet *ServerLink::interpret(Message const &msg) {
    return new BlockPacket(msg);
}
void ServerLink::initiateCommunications() {
    switch (d->state) {
    case Discovering:
        *this << ByteRefArray("Info?", 5);
        d->state = WaitingForInfoResponse;
        break;
    case Pinging:
        *this << ByteRefArray("Ping?", 5);
        d->state = WaitingForPong;
        d->pingCounter = NUM_PINGS;
        d->pings.clear();
        d->pingTimer.start();
        break;
    case QueryingMapOutline:
        *this << ByteRefArray("MapOutline?", 11);
        d->state = WaitingForInfoResponse;
        break;
    case Joining: {
        ClientWindow::main().glActivate();
        Demo_StopPlayback();
        BusyMode_FreezeGameForBusyMode();
        gx.NetConnect(true);
        String pName = (playerName? playerName : "");
        if (pName.isEmpty()) {
            pName = "Player";
        }
        String req = String("Join %1 %2").arg(SV_VERSION, 4, 16, QChar('0')).arg(pName);
        *this << req.toUtf8();
        d->state = WaitingForJoinResponse;
        ClientWindow::main().glDone();
        break; }
    default:
        DENG2_ASSERT(false);
        break;
    }
}
void ServerLink::localServersFound() {
    d->notifyDiscoveryUpdate();
}
void ServerLink::handleIncomingPackets() {
    if (d->state == Discovering)
        return;
    LOG_AS("ServerLink");
    forever {
        QScopedPointer<BlockPacket> packet(static_cast<BlockPacket *>(nextPacket()));
        if (packet.isNull()) break;
        Block const &packetData = packet->block();
        switch (d->state) {
        case WaitingForInfoResponse:
            if (!d->handleInfoResponse(packetData)) return;
            break;
        case WaitingForJoinResponse:
            if (!d->handleJoinResponse(packetData)) return;
            break;
        case WaitingForPong:
            if (packetData.size() == 4 && packetData == "Pong" &&
                d->pingCounter-- > 0) {
                d->pings.append(TimeSpan::fromMilliSeconds(d->pingTimer.elapsed()));
                *this << ByteRefArray("Ping?", 5);
                d->pingTimer.restart();
            }
            else {
                Address const svAddress = address();
                disconnect();
                if (d->pings.count()) {
                    TimeSpan average = 0;
                    for (TimeSpan i : d->pings) average += i;
                    average /= d->pings.count();
                    DENG2_FOR_AUDIENCE2(PingResponse, i) {
                        i->pingResponse(svAddress, average);
                    }
                }
            }
            break;
        case InGame: {
            netmessage_t *msg = reinterpret_cast<netmessage_t *>(M_Calloc(sizeof(netmessage_t)));
            msg->sender = 0; // the server
            msg->data = new byte[packetData.size()];
            memcpy(msg->data, packetData.data(), packetData.size());
            msg->size = packetData.size();
            msg->handle = msg->data; // needs delete[]
            N_PostMessage(msg);
            break; }
        default:
            break;
        }
    }
}
ServerLink &ServerLink::get() // static
{
    return ClientApp::serverLink();
}
#ifdef _DEBUG
D_CMD(NetFreqs);
#endif
void N_Register(void) {
#ifdef _DEBUG
    C_CMD("netfreq", NULL, NetFreqs);
#endif
}
ServerLink &Net_ServerLink(void) {
    return ClientApp::app().serverLink();
}
/**
 * Called from "net info" (client-side).
 */
void N_PrintNetworkStatus(void) {
    if (isClient) {
        LOG_NET_NOTE(_E(b) "CLIENT: " _E(.) "Connected to server at %s") << Net_ServerLink().address();
    }
    else {
        LOG_NET_NOTE(_E(b) "OFFLINE: " _E(.) "Single-player mode");
    }
    N_PrintBufferInfo();
}
