﻿/** @file aboutdialog.cpp Information about the Doomsday Client.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/aboutdialog.h"
#include "gl/sys_opengl.h"
#include "audio/audiodriver.h"
#include "clientapp.h"
#include "dd_main.h"
#include "dd_def.h"
#include <de/Version>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/LabelWidget>
#include <de/DocumentPopupWidget>
#include <de/Style>
/** @file alertdialog.cpp  Dialog for listing recent alerts.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/alertdialog.h"
#include "ui/dialogs/logsettingsdialog.h"
#include "ui/clientwindow.h"
#include "ui/clientwindowsystem.h"
#include "clientapp.h"
#include <de/App>
#include <de/ChoiceWidget>
#include <de/Config>
#include <de/DialogContentStylist>
#include <de/FIFO>
#include <de/NotificationAreaWidget>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/ui/ActionItem>
#include <de/ui/ListData>
#include <QTimer>
/** @file audiosettingsdialog.cpp  Dialog for audio settings.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/audiosettingsdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include "ui/widgets/cvartogglewidget.h"
#include "ui/widgets/cvarchoicewidget.h"
#include "ui/widgets/cvarnativepathwidget.h"
#include "audio/audiosystem.h"
#include "clientapp.h"
#include "ConfigProfiles"
#include <de/FoldPanelWidget>
#include <de/GridPopupWidget>
#include <de/ScriptSystem>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/TextValue>
#include <de/VariableChoiceWidget>
#include <de/VariableSliderWidget>
#include <de/VariableToggleWidget>
/** @file coloradjustmentdialog.cpp  Dialog for gamma, etc. adjustments.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/coloradjustmentdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include <de/GridLayout>
#include <de/SignalAction>
#include "api_console.h"
/** @file createprofiledialog.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/createprofiledialog.h"
#include "ui/widgets/packagesbuttonwidget.h"
#include "ui/widgets/packageswidget.h"
#include "ui/dialogs/datafilesettingsdialog.h"
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <doomsday/GameProfiles>
#include <doomsday/DataBundle>
#include <de/AuxButtonWidget>
#include <de/CallbackAction>
#include <de/ChoiceWidget>
#include <de/GridLayout>
#include <de/DialogContentStylist>
#include <de/FoldPanelWidget>
#include <de/PackageLoader>
#include <de/PersistentState>
#include <de/ScriptedInfo>
#include <de/SliderWidget>
#include <de/ToggleWidget>
/** @file datafilesettings.cpp  Data file settings.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/datafilesettingsdialog.h"
#include "ui/clientwindow.h"
#include "ui/widgets/consolewidget.h"
#include "ui/widgets/packageswidget.h"
#include "ui/widgets/taskbarwidget.h"
#include <doomsday/DoomsdayApp>
#include <de/Config>
/**
 * @file downloaddialog.h
 * Dialog for downloads. @ingroup ui
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/downloaddialog.h"
#include <de/CallbackAction>
/** @file filedownloaddialog.cpp  Dialog for monitoring file download progress.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/filedownloaddialog.h"
#include "network/serverlink.h"
#include <de/charsymbols.h>
/** @file inputsettingsdialog.cpp Dialog for input settings.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/inputsettingsdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include "ui/widgets/cvartogglewidget.h"
#include "ui/widgets/keygrabberwidget.h"
#include "ui/axisinputcontrol.h"
#include "ui/controllerpresets.h"
#include "ui/inputsystem.h"
#include "clientapp.h"
#include "api_console.h"
#include <de/VariableToggleWidget>
#include <de/ChoiceWidget>
#include <de/GridPopupWidget>
#include <de/SignalAction>
/** @file logsettingsdialog.cpp  Dialog for Log settings.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/logsettingsdialog.h"
#include "clientapp.h"
#include "ConfigProfiles"
#include <de/Config>
#include <de/FoldPanelWidget>
#include <de/LogFilter>
#include <de/SignalAction>
#include <de/VariableChoiceWidget>
#include <de/VariableToggleWidget>
/** @file manualconnectiondialog.cpp  Dialog for connecting to a server.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/manualconnectiondialog.h"
#include "ui/widgets/multiplayerservermenuwidget.h"
#include "ui/home/multiplayerpanelbuttonwidget.h"
#include "network/serverlink.h"
#include "clientapp.h"
#include <de/SignalAction>
#include <de/FoldPanelWidget>
#include <de/PersistentState>
#include <de/ProgressWidget>
/** @file networksettingsdialog.cpp  Dialog for network settings.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/networksettingsdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include "ui/widgets/cvartogglewidget.h"
#include "ui/widgets/cvarchoicewidget.h"
#include "ConfigProfiles"
#include "clientapp.h"
#include <de/SignalAction>
#include <de/GridPopupWidget>
#include <de/VariableLineEditWidget>
#include <de/VariableToggleWidget>
/** @file packagecompatibilitydialog.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/packagecompatibilitydialog.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/widgets/packageswidget.h"
#include <doomsday/DoomsdayApp>
#include <doomsday/DataBundle>
#include <de/CallbackAction>
#include <de/Package>
#include <de/PackageLoader>
#include <de/ui/SubwidgetItem>
#include <de/ToggleWidget>
#include <de/Loop>
#include "dd_main.h"
/** @file packagepopupwidget.cpp  Popup showing information about a package.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/packageinfodialog.h"
#include "ui/widgets/packagecontentoptionswidget.h"
#include "resource/idtech1image.h"
#include "dd_main.h"
#include <doomsday/DataBundle>
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <doomsday/LumpCatalog>
#include <de/App>
#include <de/ArchiveEntryFile>
#include <de/CallbackAction>
#include <de/DirectoryFeed>
#include <de/DocumentWidget>
#include <de/Folder>
#include <de/ImageFile>
#include <de/LabelWidget>
#include <de/Loop>
#include <de/NativeFile>
#include <de/PackageLoader>
#include <de/PopupMenuWidget>
#include <de/ProgressWidget>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <QDesktopServices>
#include <QRegularExpression>
#include <sstream>
/** @file packagesdialog.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/packagesdialog.h"
#include "ui/widgets/packageswidget.h"
#include "ui/widgets/homeitemwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/dialogs/datafilesettingsdialog.h"
#include "resource/idtech1image.h"
#include "ui/clientwindow.h"
#include "ui/clientstyle.h"
#include "clientapp.h"
#include <doomsday/Games>
#include <doomsday/LumpCatalog>
#include <doomsday/res/Bundles>
#include <de/charsymbols.h>
#include <de/CallbackAction>
#include <de/ChildWidgetOrganizer>
#include <de/FileSystem>
#include <de/DocumentPopupWidget>
#include <de/MenuWidget>
#include <de/NativeFile>
#include <de/PackageLoader>
#include <de/PopupButtonWidget>
#include <de/PopupMenuWidget>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/ui/SubwidgetItem>
#include <de/ui/VariantActionItem>
/** @file renderersettingsdialog.cpp  Settings for the renderer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/renderersettingsdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include "ui/widgets/cvartogglewidget.h"
#include "ui/widgets/cvarchoicewidget.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/editors/rendererappearanceeditor.h"
#include "ui/widgets/profilepickerwidget.h"
#include "ui/clientwindow.h"
#include "render/rendersystem.h"
#include "gl/gl_texmanager.h"
#include "clientapp.h"
#include <doomsday/console/exec.h>
#include <de/GridPopupWidget>
#include <de/GridLayout>
#include <de/SignalAction>
#include <de/DialogContentStylist>
#include <de/InputDialog>
#include <de/VariableSliderWidget>
#include <de/VariableToggleWidget>
/** @file repositorybrowserdialog.cpp  Remote package repository browser.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/repositorybrowserdialog.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/widgets/homeitemwidget.h"
#include <de/Async>
#include <de/ChoiceWidget>
#include <de/Config>
#include <de/DictionaryValue>
#include <de/DocumentWidget>
#include <de/FileSystem>
#include <de/Folder>
#include <de/LineEditWidget>
#include <de/ProgressWidget>
#include <de/RemoteFeedRelay>
#include <de/SequentialLayout>
#include <de/ToggleWidget>
#include <de/ui/FilteredData>
/** @file serverinfodialog.cpp  Information about a multiplayer server.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/serverinfodialog.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/widgets/packageswidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/widgets/mapoutlinewidget.h"
#include "ui/widgets/packagesbuttonwidget.h"
#include "network/serverlink.h"
#include <doomsday/Games>
#include <de/charsymbols.h>
#include <de/ButtonWidget>
#include <de/CallbackAction>
#include <de/Config>
#include <de/DictionaryValue>
#include <de/PackageLoader>
#include <de/SequentialLayout>
#include <de/ui/SubwidgetItem>
#include <de/ProgressWidget>
#include <QTimer>
/** @file uisettingsdialog.cpp  User interface settings.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/uisettingsdialog.h"
#include "clientapp.h"
#include "ConfigProfiles"
#include <de/CallbackAction>
#include <de/Config>
#include <de/GridLayout>
#include <de/VariableChoiceWidget>
#include <de/VariableToggleWidget>
/** @file videosettingsdialog.cpp  Dialog for video settings.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/videosettingsdialog.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/widgets/cvarchoicewidget.h"
#include "ui/clientwindow.h"
#include "ui/clientwindowsystem.h"
#include "CommandAction"
#include "ConfigProfiles"
#include "clientapp.h"
#include "dd_main.h"
#include <doomsday/console/exec.h>
#include <de/Config>
#include <de/VariableToggleWidget>
#include <de/VariableSliderWidget>
#include <de/ChoiceWidget>
#include <de/SequentialLayout>
#include <de/GridLayout>
#include <de/SignalAction>
#include <de/DisplayMode>
#include <QPoint>
/** @file vrsettingsdialog.cpp  Settings for virtual reality.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/dialogs/vrsettingsdialog.h"
#include "ui/widgets/cvarsliderwidget.h"
#include "ui/widgets/cvartogglewidget.h"
#include "ui/widgets/cvarchoicewidget.h"
#include "render/vr.h"
#include "clientapp.h"
#include "api_console.h"
#include <doomsday/console/exec.h>
#include <de/VariableSliderWidget>
#include <de/SignalAction>
#include <de/Config>
#include "CommandAction"
using namespace de::ui;
using namespace de;
DENG2_PIMPL(AboutDialog) {
    DocumentPopupWidget *glPopup;
    DocumentPopupWidget *audioPopup;
    Impl(Public *i) : Base(i) {
        glPopup = new DocumentPopupWidget;
        glPopup->document().setText(Sys_GLDescription());
        self().add(glPopup);
        audioPopup = new DocumentPopupWidget;
        audioPopup->document().setText(App_AudioSystem().description());
        self().add(audioPopup);
    }
};
AboutDialog::AboutDialog() : DialogWidget("about"), d(new Impl(this)) {
    /*
     * Construct the widgets.
     */
    LabelWidget *logo = new LabelWidget;
    logo->setImage(style().images().image("logo.px256"));
    logo->setSizePolicy(ui::Fixed, ui::Expand);
    de::Version const version = de::Version::currentBuild();
    LabelWidget *title = LabelWidget::newWithText(String("%1 %2.%3")
                                                  .arg(DOOMSDAY_NICENAME)
                                                  .arg(version.major)
                                                  .arg(version.minor));
    title->margins().set("");
    title->setFont("title");
    title->setTextColor("accent");
    title->setSizePolicy(ui::Fixed, ui::Expand);
    LabelWidget *info = new LabelWidget;
    String txt = String(_E(b) "%4 build %5" _E(.) "\n%6%8\n\n%1 (%2-%7)%3")
            .arg(version.operatingSystem() == "windows"? tr("Windows") :
                 version.operatingSystem() == "macx"? tr("macOS") :
                 version.operatingSystem() == "ios"? tr("iOS") : tr("Unix"))
            .arg(version.cpuBits())
            .arg(version.isDebugBuild()? tr(" Debug") : "")
            .arg(DOOMSDAY_RELEASE_TYPE)
            .arg(version.build)
            .arg(Time::fromText(__DATE__ " " __TIME__, Time::CompilerDateTime)
                 .asDateTime().toString(Qt::SystemLocaleShortDate))
            .arg(tr("bit"))
            .arg(version.gitDescription.isEmpty()? "" : ("\n" _E(s)_E(F) + version.gitDescription + _E(.)_E(.)));
    info->setText(txt);
    info->setSizePolicy(ui::Fixed, ui::Expand);
    area().add(logo);
    area().add(title);
    area().add(info);
    RuleRectangle const &cont = area().contentRule();
    SequentialLayout layout(cont.left(), cont.top());
    layout.setOverrideWidth(rule("dialog.about.width"));
    layout << *logo << *title << *info;
    area().setContentSize(layout);
    buttons()
            << new DialogButtonItem(DialogWidget::Accept | DialogWidget::Default, tr("Close"))
            << new DialogButtonItem(DialogWidget::ActionPopup, tr("GL"))
            << new DialogButtonItem(DialogWidget::ActionPopup, tr("Audio"))
            << new DialogButtonItem(DialogWidget::Action, tr("Homepage..."),
                                    new SignalAction(&ClientApp::app(), SLOT(openHomepageInBrowser())));
    popupButtonWidget(tr("GL")).setPopup(*d->glPopup);
    popupButtonWidget(tr("Audio")).setPopup(*d->audioPopup);
}
static String const VAR_AUTOHIDE = "alert.autoHide";
DENG_GUI_PIMPL(AlertDialog)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate)
, DENG2_OBSERVES(Variable, Change) {
    class AlertItem : public ui::ActionItem {
    public:
        AlertItem(String const &msg, Level level)
            : ui::ActionItem(ShownAsLabel, msg)
            , _level(level)
        {}
        Level level() const { return _level; }
    private:
        Level _level;
    };
    struct TextStyling : public Font::RichFormat::IStyle {
        Color richStyleColor(int /*index*/) const
        {
            return Color(255, 255, 255, 255); // not used by LabelWidget
        }
        void richStyleFormat(int contentStyle, float &sizeFactor,
                             Font::RichFormat::Weight &fontWeight,
                             Font::RichFormat::Style &fontStyle,
                             int &colorIndex) const {
            ClientApp::windowSystem().style()
                    .richStyleFormat(contentStyle, sizeFactor, fontWeight, fontStyle, colorIndex);
            if (contentStyle == Font::RichFormat::MajorStyle ||
               contentStyle == Font::RichFormat::MajorMetaStyle) {
                fontWeight = Font::RichFormat::Normal;
            }
        }
    };
    UniqueWidgetPtr<PopupButtonWidget> notification;
    MenuWidget *alerts;
    bool clearOnDismiss;
    TextStyling styling;
    QTimer hideTimer; ///< Automatically hides the notification.
    ChoiceWidget *autohideTimes;
    DialogContentStylist stylist;
    dsize maxCount;
    typedef FIFO<AlertItem> Pending;
    Pending pending;
    Impl(Public *i)
        : Base(i)
        , clearOnDismiss(false)
        , maxCount(100) {
        notification.reset(new PopupButtonWidget);
        notification->setBehavior(Focusable, false);
        notification->setSizePolicy(ui::Expand, ui::Expand);
        notification->setImage(style().images().image("alert"));
        notification->setOverrideImageSize(style().fonts().font("default").height());
        notification->setOpener([this] (PopupWidget *) {
            self().showListOfAlerts();
        });
        ScrollAreaWidget &area = self().area();
        alerts = new MenuWidget;
        alerts->enableScrolling(false);
        alerts->setGridSize(1, ui::Expand, 0, ui::Expand);
        alerts->rule().setLeftTop(area.contentRule().left(), area.contentRule().top());
        area.setContentSize(alerts->rule());
        area.add(alerts);
        area.enableIndicatorDraw(true);
        alerts->organizer().audienceForWidgetCreation() += this;
        alerts->organizer().audienceForWidgetUpdate() += this;
        QObject::connect(&hideTimer, SIGNAL(timeout()), thisPublic, SLOT(hideNotification()));
        hideTimer.setSingleShot(true);
        App::config(VAR_AUTOHIDE).audienceForChange() += this;
    }
    NotificationAreaWidget &notifs() {
        return ClientWindow::main().notifications();
    }
    int autoHideAfterSeconds() const {
        return App::config().geti(VAR_AUTOHIDE, 3 * 60);
    }
    void variableValueChanged(Variable &, Value const &) {
        if (!autoHideAfterSeconds()) {
            hideTimer.stop();
        }
        else {
            hideTimer.setInterval(autoHideAfterSeconds() * 1000);
            if (!hideTimer.isActive()) hideTimer.start();
        }
    }
    void queueAlert(String const &msg, Level level) {
        pending.put(new AlertItem(msg, level));
    }
    bool addPendingAlerts() {
        bool changed = false;
        while (AlertItem *alert = pending.take()) {
            add(alert);
            changed = true;
        }
        while (alerts->items().size() > maxCount) {
            alerts->items().remove(alerts->items().size() - 1);
            changed = true;
        }
        return changed;
    }
    void add(AlertItem *alert) {
        DENG2_ASSERT_IN_MAIN_THREAD();
        for (ui::Data::Pos i = 0; i < alerts->items().size(); ++i) {
            if (!alerts->items().at(i).label().compareWithoutCase(alert->label()))
                return;
        }
        alerts->items().insert(0, alert);
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &item) {
        widget.set(Background());
        LabelWidget &label = widget.as<LabelWidget>();
        label.setBehavior(ContentClipping); // allow clip-culling
        label.setTextStyle(&styling);
        label.setMaximumTextWidth(rule("alerts.width").valuei());
        label.setSizePolicy(ui::Expand, ui::Expand);
        label.setAppearanceAnimation(LabelWidget::AppearGrowVertically, 0.5);
        label.setAlignment(ui::AlignBottom);
        label.setImage(style().images().image("alert"));
        label.setOverrideImageSize(style().fonts().font("default").height());
        label.setImageAlignment(ui::AlignTop);
        label.setTextAlignment(ui::AlignRight);
        label.margins()
                .setLeft("")
                .setRight("")
                .setBottom("");
        AlertItem const &alert = item.as<AlertItem>();
        switch (alert.level()) {
        case Minor:
            label.setImageColor(Vector4f(style().colors().colorf("text"), .5f));
            break;
        case Normal:
            label.setImageColor(style().colors().colorf("text"));
            break;
        case Major:
            label.setTextStyle(0); // use default styling with bold weights
            label.setImageColor(style().colors().colorf("accent"));
            break;
        }
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) {
        DENG2_UNUSED(widget);
        DENG2_UNUSED(item);
    }
    void showNotification() {
        notification->setImageColor(style().colors().colorf("accent"));
        notifs().showOrHide(*notification, true);
        if (autoHideAfterSeconds() > 0) {
            hideTimer.start(autoHideAfterSeconds() * 1000);
        }
    }
    void hideNotification() {
        notifs().hideChild(*notification);
    }
    /**
     * @returns @c true, if the notification was hidden due to the alerts list being
     * empty.
     */
    bool hideIfEmpty() {
        if (alerts->items().isEmpty()) {
            hideNotification();
            return true; // was hidden
        }
        return false;
    }
    void updateAutohideTimeSelection() {
        int const time = autoHideAfterSeconds();
        ui::DataPos pos = autohideTimes->items().findData(time);
        if (pos != ui::Data::InvalidPos) {
            autohideTimes->setSelected(pos);
        }
        else {
            autohideTimes->setSelected(autohideTimes->items().findData(0));
        }
    }
};
AlertDialog::AlertDialog(String const &/*name*/) : d(new Impl(this))
{
    setOutlineColor("transparent");
    d->notification->setPopup(*this, ui::Down);
    buttons() << new DialogButtonItem(DialogWidget::Accept | DialogWidget::Default,
                                      tr("Clear All"))
              << new DialogButtonItem(DialogWidget::ActionPopup | DialogWidget::Id1,
                                      style().images().image("gear"));
    d->stylist.setContainer(*this);
    auto &gearButton = *popupButtonWidget(DialogWidget::Id1);
    gearButton.setPopup([] (PopupButtonWidget const &) {
        return new LogSettingsDialog;
    }, ui::Left);
    gearButton.setOpener([this] (PopupWidget *pop) {
        auto &dlg = pop->as<LogSettingsDialog>();
        connect(this, SIGNAL(closed()), pop, SLOT(close()));
        dlg.orphan();
        dlg.exec(root());
    });
    auto *lab = LabelWidget::newWithText(tr("Hide After:"), this);
    add(d->autohideTimes = new ChoiceWidget);
    d->autohideTimes->items()
            << new ChoiceItem(tr("1 min"),   60)
            << new ChoiceItem(tr("3 mins"),  3 * 60)
            << new ChoiceItem(tr("5 mins"),  5 * 60)
            << new ChoiceItem(tr("10 mins"), 10 * 60)
            << new ChoiceItem(tr("Never"),   0);
    d->updateAutohideTimeSelection();
    lab->rule()
        .setInput(Rule::Left, gearButton.rule().right())
        .setMidAnchorY(gearButton.rule().midY());
    d->autohideTimes->rule()
        .setInput(Rule::Left, lab->rule().right())
        .setInput(Rule::Top,  lab->rule().top());
    setMinimumContentWidth(extraButtonsMenu().rule().width() +
                           buttonsMenu().rule().width() +
                           lab->rule().width() +
                           d->autohideTimes->rule().width());
    connect(d->autohideTimes, SIGNAL(selectionChangedByUser(uint)), this, SLOT(autohideTimeChanged()));
}
void AlertDialog::newAlert(String const &message, Level level) {
    d->queueAlert(message, level);
}
void AlertDialog::update() {
    DialogWidget::update();
    if (d->addPendingAlerts()) {
        d->showNotification();
    }
}
void AlertDialog::showListOfAlerts() {
    if (isOpen() || d->hideIfEmpty()) return;
    d->notification->setImageColor(style().colors().colorf("text"));
    area().scrollToTop(0);
    open();
}
void AlertDialog::hideNotification() {
    d->hideNotification();
}
void AlertDialog::autohideTimeChanged() {
    App::config().set(VAR_AUTOHIDE, d->autohideTimes->selectedItem().data().toInt());
}
void AlertDialog::finish(int result) {
    DialogWidget::finish(result);
    d->clearOnDismiss = (result != 0);
}
void AlertDialog::panelDismissed() {
    if (d->clearOnDismiss) {
        d->clearOnDismiss = false;
        d->alerts->items().clear();
        d->hideIfEmpty();
    }
    DialogWidget::panelDismissed();
}
DENG_GUI_PIMPL(AudioSettingsDialog) {
    CVarSliderWidget *sfxVolume;
    CVarSliderWidget *musicVolume;
    CVarSliderWidget *reverbVolume;
    CVarToggleWidget *sound3D;
    CVarToggleWidget *overlapStop;
    CVarChoiceWidget *musicSource;
    CVarNativePathWidget *musicSoundfont;
    VariableToggleWidget *pauseOnFocus;
    CVarToggleWidget     *soundInfo;
    GridPopupWidget      *devPopup;
    FoldPanelWidget      *backendFold;
    GuiWidget            *backendBase;
    VariableSliderWidget *sfxChannels;
    VariableChoiceWidget *audioOutput;
    VariableChoiceWidget *soundPlugin;
    VariableChoiceWidget *musicPlugin;
#if defined (WIN32)
    VariableChoiceWidget *cdPlugin;
#endif
    bool needAudioReinit = false;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        area.enableIndicatorDraw(true);
        if (DoomsdayApp::isGameLoaded()) {
            area.add(sfxVolume      = new CVarSliderWidget    ("sound-volume"));
            area.add(musicVolume    = new CVarSliderWidget    ("music-volume"));
            area.add(reverbVolume   = new CVarSliderWidget    ("sound-reverb-volume"));
            area.add(sound3D        = new CVarToggleWidget    ("sound-3d"));
            area.add(overlapStop    = new CVarToggleWidget    ("sound-overlap-stop"));
            area.add(musicSource    = new CVarChoiceWidget    ("music-source"));
            area.add(musicSoundfont = new CVarNativePathWidget("music-soundfont"));
            musicSoundfont->setBlankText(tr("GeneralUser GS"));
            musicSoundfont->setFilters(StringList()
                                       << "SF2 soundfonts (*.sf2)"
                                       << "DLS soundfonts (*.dls)"
                                       << "All files (*)");
            area.add(pauseOnFocus = new VariableToggleWidget("Pause on Focus Lost",
                                                             App::config("audio.pauseOnFocus"),
                                                             "pause-on-focus"));
            sfxVolume  ->setDisplayFactor(100.0 / 255.0);
            musicVolume->setDisplayFactor(100.0 / 255.0);
            sfxVolume  ->setStep(1.0 / sfxVolume->displayFactor());
            musicVolume->setStep(1.0 / musicVolume->displayFactor());
            self().add(devPopup = new GridPopupWidget);
            soundInfo = new CVarToggleWidget("sound-info", tr("Sound Channel Status"));
            *devPopup << soundInfo;
            devPopup->commit();
        }
        backendFold = FoldPanelWidget::makeOptionsGroup("audio-backend", "Audio Backend", &area);
        backendBase = new GuiWidget("fold-base");
        backendFold->setContent(backendBase);
        backendBase->add(audioOutput  = new VariableChoiceWidget(App::config("audio.output"), VariableChoiceWidget::Number));
        backendBase->add(sfxChannels  = new VariableSliderWidget(App::config("audio.channels"), Ranged(1, 64), 1.0));
        backendBase->add(soundPlugin  = new VariableChoiceWidget(App::config("audio.soundPlugin"), VariableChoiceWidget::Text));
        backendBase->add(musicPlugin  = new VariableChoiceWidget(App::config("audio.musicPlugin"), VariableChoiceWidget::Text));
#if defined (WIN32)
        backendBase->add(cdPlugin = new VariableChoiceWidget(App::config("audio.cdPlugin"), VariableChoiceWidget::Text));
#endif
        {
            GridLayout layout(backendBase->rule().left(), backendBase->rule().top());
            layout.setGridSize(2, 0);
            layout.setColumnAlignment(0, ui::AlignRight);
            layout << *LabelWidget::newWithText("SFX Plugin:", backendBase) << *soundPlugin
                   << *LabelWidget::newWithText("Music Plugin:", backendBase) << *musicPlugin
#if defined (WIN32)
                   << *LabelWidget::newWithText("CD Plugin:", backendBase) << *cdPlugin
#endif
                   << *LabelWidget::newWithText("Output:", backendBase) << *audioOutput
                   << *LabelWidget::newWithText("SFX Channels:", backendBase) << *sfxChannels;
            backendBase->rule().setSize(layout);
        }
        enumerateAudioOutputs();
        /*
        fmodSpeakerMode->items()
                << new ChoiceItem("Mono", "mono")
                << new ChoiceItem("Stereo", "")
                << new ChoiceItem("Quad", "quad")
                << new ChoiceItem("Surround", "surround")
                << new ChoiceItem("5.1", "5.1")
                << new ChoiceItem("7.1", "7.1")
                << new ChoiceItem("Raw", "raw");
         */
        soundPlugin->items()
                << new ChoiceItem(tr("FMOD"), "fmod")
           #if !defined (DENG_DISABLE_SDLMIXER)
                << new ChoiceItem(tr("SDL_mixer"), "sdlmixer")
           #endif
                << new ChoiceItem(tr("OpenAL"), "openal")
           #if defined (WIN32)
                << new ChoiceItem(tr("DirectSound"), "dsound")
           #endif
                << new ChoiceItem(tr("Disabled"), "dummy");
        musicPlugin->items()
                << new ChoiceItem(tr("Fluidsynth"), "fluidsynth")
                << new ChoiceItem(tr("FMOD"), "fmod")
           #if !defined (DENG_DISABLE_SDLMIXER)
                << new ChoiceItem(tr("SDL_mixer"), "sdlmixer")
           #endif
           #if defined (WIN32)
                << new ChoiceItem(tr("Windows Multimedia"), "winmm")
           #endif
                << new ChoiceItem(tr("Disabled"), "dummy");
#if defined (WIN32)
        cdPlugin->items()
                << new ChoiceItem(tr("Windows Multimedia"), "winmm")
                << new ChoiceItem(tr("Disabled"), "dummy");
        cdPlugin->updateFromVariable();
#endif
        soundPlugin    ->updateFromVariable();
        musicPlugin    ->updateFromVariable();
        auto changeFunc = [this](uint) {
            needAudioReinit = true;
            self().buttonWidget(Id2)->setText(_E(b) "Apply");
        };
        QObject::connect(soundPlugin,     &ChoiceWidget::selectionChangedByUser, changeFunc);
        QObject::connect(musicPlugin,     &ChoiceWidget::selectionChangedByUser, changeFunc);
#if defined (WIN32)
        QObject::connect(cdPlugin,        &ChoiceWidget::selectionChangedByUser, changeFunc);
#endif
        QObject::connect(sfxChannels, &SliderWidget::valueChangedByUser, changeFunc);
    }
    void enumerateAudioOutputs() {
        audioOutput->items().clear();
        const auto &outputs =
            ScriptSystem::get()["Audio"]["outputs"].value<DictionaryValue>();
        const TextValue key("fmod");
        if (outputs.contains(key)) {
            const auto &names = outputs.element(key).as<ArrayValue>();
            for (unsigned i = 0; i < names.size(); ++i) {
                audioOutput->items() << new ChoiceItem(names.at(i).asText(), i);
            }
        }
    }
    void fetch() {
        if (!DoomsdayApp::isGameLoaded()) return;
        foreach (GuiWidget *w, self().area().childWidgets() + devPopup->content().childWidgets()) {
            if (ICVarWidget *cv = maybeAs<ICVarWidget>(w)) {
                cv->updateFromCVar();
            }
        }
    }
};
AudioSettingsDialog::AudioSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading)
    , d(new Impl(this)) {
    bool const gameLoaded = DoomsdayApp::isGameLoaded();
    heading().setText(tr("Audio Settings"));
    heading().setImage(style().images().image("audio"));
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(2, 0);
    layout.setColumnAlignment(0, ui::AlignRight);
    if (gameLoaded) {
        auto *sfxVolLabel   = LabelWidget::newWithText(tr("SFX Volume:"     ), &area());
        auto *musicVolLabel = LabelWidget::newWithText(tr("Music Volume:"   ), &area());
        auto *rvbVolLabel   = LabelWidget::newWithText(tr("Reverb Strength:"), &area());
        d->sound3D    ->setText(tr("3D Effects & Reverb"  ));
        d->overlapStop->setText(tr("One Sound per Emitter"));
        /*auto *rateLabel = LabelWidget::newWithText(tr("Resampling:"), &area());
        d->sampleRate->items()
                << new ChoiceItem(tr("1x @ 11025 Hz"), 11025)
                << new ChoiceItem(tr("2x @ 22050 Hz"), 22050)
                << new ChoiceItem(tr("4x @ 44100 Hz"), 44100);*/
        auto *musSrcLabel = LabelWidget::newWithText(tr("Preferred Music:"), &area());
        d->musicSource->items()
                << new ChoiceItem(tr("MUS lumps"),      AudioSystem::MUSP_MUS)
                << new ChoiceItem(tr("External files"), AudioSystem::MUSP_EXT)
                << new ChoiceItem(tr("CD"),             AudioSystem::MUSP_CD);
        auto *sfLabel = LabelWidget::newWithText(tr("MIDI Sound Font:"), &area());
        LabelWidget::appendSeparatorWithText("Sound Effects", &area(), &layout);
        layout << *sfxVolLabel      << *d->sfxVolume
               << Const(0)          << *d->overlapStop
               << Const(0)          << *d->sound3D
               << *rvbVolLabel      << *d->reverbVolume;
        LabelWidget::appendSeparatorWithText("Music", &area(), &layout);
        layout << *musicVolLabel    << *d->musicVolume
               << *musSrcLabel      << *d->musicSource
               << *sfLabel          << *d->musicSoundfont
               << Const(0)          << *d->pauseOnFocus;
    }
    else {
        d->backendFold->open();
    }
    SequentialLayout layout2(area().contentRule().left(),
                             area().contentRule().top() + layout.height());
    layout2.setOverrideWidth(d->backendBase->rule().width());
    layout2 << d->backendFold->title()
            << *d->backendFold;
    area().setContentSize(OperatorRule::maximum(layout.width(), layout2.width()),
                          layout.height() + layout2.height());
    d->backendFold->title().rule().setInput(Rule::Width, area().contentRule().width());
    buttons()
            << new DialogButtonItem(Default | Accept | Id2, tr("Close"))
            << new DialogButtonItem(Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())));
    if (gameLoaded) {
        buttons() << new DialogButtonItem(ActionPopup | Id1,
                                          style().images().image("gauge"));
        popupButtonWidget(Id1)->setPopup(*d->devPopup);
    }
    d->fetch();
}
void AudioSettingsDialog::resetToDefaults() {
    ClientApp::audioSettings().resetToDefaults();
    d->fetch();
    d->needAudioReinit = true;
}
void AudioSettingsDialog::finish(int result) {
    DialogWidget::finish(result);
    if (result) {
        if (d->needAudioReinit) {
            AudioSystem::get().reinitialize();
        }
    }
}
DENG_GUI_PIMPL(ColorAdjustmentDialog) {
    CVarSliderWidget *gamma;
    CVarSliderWidget *contrast;
    CVarSliderWidget *brightness;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        LabelWidget *gammaLabel      = LabelWidget::newWithText(tr("Gamma:"), &area);
        LabelWidget *contrastLabel   = LabelWidget::newWithText(tr("Contrast:"), &area);
        LabelWidget *brightnessLabel = LabelWidget::newWithText(tr("Brightness:"), &area);
        area.add(gamma = new CVarSliderWidget("vid-gamma"));
        area.add(contrast = new CVarSliderWidget("vid-contrast"));
        area.add(brightness = new CVarSliderWidget("vid-bright"));
        Rule const &sliderWidth = rule("coloradjustment.slider");
        gamma->rule().setInput(Rule::Width, sliderWidth);
        contrast->rule().setInput(Rule::Width, sliderWidth);
        brightness->rule().setInput(Rule::Width, sliderWidth);
        GridLayout layout(area.contentRule().left(), area.contentRule().top());
        layout.setGridSize(2, 3);
        layout.setColumnAlignment(0, ui::AlignRight);
        layout << *gammaLabel      << *gamma
               << *contrastLabel   << *contrast
               << *brightnessLabel << *brightness;
        auto *msg = LabelWidget::newWithText("These only affect the 3D game view.", &area); {
            msg->margins().setTop("unit").setBottom("");
            msg->setFont("separator.annotation");
            msg->setTextColor("altaccent");
        }
        layout << Const(0) << *msg;
        area.setContentSize(layout);
    }
    void fetch() {
        gamma->updateFromCVar();
        contrast->updateFromCVar();
        brightness->updateFromCVar();
    }
};
ColorAdjustmentDialog::ColorAdjustmentDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Color Adjustments"));
    buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())));
}
void ColorAdjustmentDialog::prepare() {
    DialogWidget::prepare();
    d->fetch();
}
void ColorAdjustmentDialog::resetToDefaults() {
    Con_SetFloat("vid-gamma", 1);
    Con_SetFloat("vid-contrast", 1);
    Con_SetFloat("vid-bright", 0);
    d->gamma->updateFromCVar();
    d->contrast->updateFromCVar();
    d->brightness->updateFromCVar();
}
DENG_GUI_PIMPL(CreateProfileDialog)
, DENG2_OBSERVES(ToggleWidget, Toggle)
, DENG2_OBSERVES(SliderWidget, Change) {
    GuiWidget *form;
    ChoiceWidget *gameChoice;
    PackagesButtonWidget *packages;
    ChoiceWidget *autoStartMap;
    ChoiceWidget *autoStartSkill;
    AuxButtonWidget *customDataFileName;
    ui::ListData customDataFileActions;
    SafeWidgetPtr<PackagesWidget> customPicker;
    FoldPanelWidget *optionsFold;
    GuiWidget *optionsBase;
    DialogContentStylist stylist;
    bool editing = false;
    String oldName;
    std::unique_ptr<GameProfile> tempProfile; // Used only internally.
    Impl(Public *i) : Base(i) {}
    void populateOptions() {
        optionsBase->clearTree();
        GridLayout layout(optionsBase->rule().left(), optionsBase->rule().top());
        layout.setGridSize(2, 0);
        layout.setColumnAlignment(0, ui::AlignRight);
        const auto &opts = tempProfile->game()[Game::DEF_OPTIONS].valueAsRecord().members();
        if (opts.isEmpty()) {
            return;
        }
        StringList keys = opts.keys();
        std::sort(keys.begin(), keys.end(), [&opts](const String &k1, const String &k2) {
            return opts[k1]->valueAsRecord().gets("label").compareWithoutCase(
                       opts[k2]->valueAsRecord().gets("label")) < 0;
        });
        for (const auto &key : keys) {
            const Record &optDef   = opts[key]->valueAsRecord();
            const String  optType  = optDef["type"];
            const String  optLabel = optDef["label"];
            if (optType == "boolean") {
                auto *tog = new ToggleWidget;
                optionsBase->add(tog);
                tog->setText(optLabel);
                tog->setActive(ScriptedInfo::isTrue(tempProfile->optionValue(key)));
                tog->objectNamespace().set("option", key);
                tog->audienceForToggle() += this;
                layout << Const(0) << *tog;
            }
            else if (optType == "number") {
                auto *slider = new SliderWidget;
                optionsBase->add(slider);
                slider->setRange(Ranged(optDef.getd("min"), optDef.getd("max")));
                const double step = optDef.getd("step", 1.0);
                slider->setStep(step);
                slider->setValue(tempProfile->optionValue(key).asNumber());
                slider->setPrecision(step < 1 ? 1 : 0);
                slider->objectNamespace().set("option", key);
                slider->audienceForChange() += this;
                layout << *LabelWidget::newWithText(optLabel + ":", optionsBase)
                       << *slider;
            }
        }
        optionsBase->rule().setSize(layout);
    }
    void toggleStateChanged(ToggleWidget &widget) override {
        tempProfile->setOptionValue(widget["option"], NumberValue(widget.isActive()));
    }
    void sliderValueChanged(SliderWidget &widget) override {
        tempProfile->setOptionValue(widget["option"], NumberValue(widget.value()));
    }
    void checkValidProfileName() {
        bool valid = false;
        String const entry = self().profileName();
        if (!entry.isEmpty()) {
            if (editing && !oldName.compareWithoutCase(entry)) {
                valid = true;
            }
            else {
                valid = DoomsdayApp::gameProfiles().forAll([&entry] (const GameProfile &prof) {
                    if (!entry.compareWithoutCase(prof.name())) {
                        return LoopAbort;
                    }
                    return LoopContinue;
                }) == LoopContinue;
            }
        }
        if (!gameChoice->isValidSelection()) {
            valid = false;
        }
        if (gameChoice->isValidSelection() &&
            gameChoice->selectedItem().data().toString().isEmpty()) {
            valid = false;
        }
        self().buttonWidget(Id1)->enable(valid);
    }
    void gameChanged() {
        if (gameChoice->isValidSelection()) {
            tempProfile->setGame(gameChoice->selectedItem().data().toString());
            updateMapList();
        }
    }
    void updateMapList() {
        auto &mapItems = autoStartMap->items();
        String oldChoice;
        if (!mapItems.isEmpty()) {
            oldChoice = autoStartMap->selectedItem().data().toString();
        }
        mapItems.clear();
        mapItems << new ChoiceItem(tr("Title screen"), "");
        StringList packageIds;
        if (gameChoice->isValidSelection()) {
            packageIds += tempProfile->allRequiredPackages();
        }
        for (int i = packageIds.size() - 1; i >= 0; --i) {
            String const pkgId = packageIds.at(i);
            if (File const *pkgFile = PackageLoader::get().select(pkgId)) {
                DataBundle const *bundle = maybeAs<DataBundle>(pkgFile->target());
                if (!bundle || !bundle->lumpDirectory()) {
                    continue;
                }
                auto const maps = bundle->lumpDirectory()->findMapLumpNames();
                if (!maps.isEmpty()) {
                    mapItems << new ui::Item(ui::Item::Separator);
                    String const wadName = Package::metadata(*pkgFile).gets(Package::VAR_TITLE);
                    foreach (String mapId, maps) {
                        if (mapItems.findData(mapId) == ui::Data::InvalidPos) {
                            mapItems << new ChoiceItem(String("%1  " _E(s)_E(C) "%2").arg(mapId).arg(wadName),
                                                       mapId);
                        }
                    }
                }
            }
        }
        auto const pos = mapItems.findData(oldChoice);
        autoStartMap->setSelected(pos != ui::Data::InvalidPos ? pos : 0);
    }
    void updateDataFile() {
        if (tempProfile->customDataFile().isEmpty()) {
            customDataFileName->setText(_E(l) "Default game data");
        }
        else {
            if (const auto *pkgFile = PackageLoader::get().select(tempProfile->customDataFile())) {
                const auto path = pkgFile->correspondingNativePath();
                customDataFileName->setText(path.pretty());
            }
            else {
                customDataFileName->setText(_E(b) "Not found");
            }
        }
        updateMapList();
    }
};
CreateProfileDialog::CreateProfileDialog(String const &gameFamily)
    : InputDialog("create-profile")
    , d(new Impl(this)) {
    title()  .setText(tr("New Profile"));
    message().setText(tr("Enter a name for the new game profile.")); {
        auto *form = new GuiWidget;
        d->form = form;
        d->stylist.setContainer(*form);
        area().add(form);
        form->add(d->gameChoice = new ChoiceWidget);
        DoomsdayApp::games().forAll([this, &gameFamily] (Game &game) {
            if (game.family() == gameFamily) {
                const char *labelColor = (!game.isPlayable() ? _E(F) : "");
                d->gameChoice->items() << new ChoiceItem(labelColor + game.title(), game.id());
            }
            return LoopContinue;
        });
        d->gameChoice->items().sort();
        if (d->gameChoice->items().isEmpty()) {
            d->gameChoice->items() << new ChoiceItem(tr("No playable games"), "");
        } {
            form->add(d->customDataFileName = new AuxButtonWidget);
            d->customDataFileName->setMaximumTextWidth(
                rule().right() - d->customDataFileName->rule().left() -
                        d->customDataFileName->margins().width());
            d->customDataFileName->setTextLineAlignment(ui::AlignLeft);
            d->customDataFileName->auxiliary().setText("Reset");
            d->customDataFileActions
                    << new ui::ActionItem("Select", new CallbackAction([this]() {
                if (d->customPicker) {
                    d->tempProfile->setCustomDataFile(d->customPicker->actionPackage());
                    d->customPicker->ancestorOfType<MessageDialog>()->accept();
                    d->updateDataFile();
                }
            }));
            d->customDataFileName->setActionFn([this]() {
                auto *dlg = new MessageDialog;
                dlg->buttons() << new DialogButtonItem(Reject, "Cancel")
                               << new DialogButtonItem(Action | Id1, style().images().image("gear"),
                                                       "Data Files", new CallbackAction([dlg]() {
                    auto *dfsDlg = new DataFileSettingsDialog;
                    dfsDlg->setAnchorAndOpeningDirection(dlg->buttonWidget(Id1)->rule(), ui::Up);
                    dfsDlg->setDeleteAfterDismissed(true);
                    dlg->add(dfsDlg);
                    dfsDlg->open();
                }));
                dlg->setDeleteAfterDismissed(true);
                dlg->title().setText("Game Data File");
                dlg->message().setText(
                    "Select the main data file for this profile. When using a manually chosen main "
                    "data file, the default data files of the game will not be automatically loaded.");
                dlg->area().enableIndicatorDraw(true);
                d->customPicker.reset(new PackagesWidget);
                dlg->area().add(d->customPicker);
                d->customPicker->setHiddenTags({"hidden", "core"});
                d->customPicker->searchTermsEditor().setText("gamedata");
                d->customPicker->setAllowPackageInfoActions(false);
                d->customPicker->setActionItems(d->customDataFileActions);
                dlg->setAnchorAndOpeningDirection(d->customDataFileName->rule(), ui::Right);
                dlg->updateLayout();
                root().addOnTop(dlg);
                dlg->open(Modal);
            });
            d->customDataFileName->auxiliary().setActionFn([this]() {
                d->tempProfile->setCustomDataFile("");
                d->updateDataFile();
            });
        }
        form->add(d->packages = new PackagesButtonWidget);
        d->packages->setNoneLabel(tr("None"));
        d->tempProfile.reset(new GameProfile);
        d->packages->setGameProfile(*d->tempProfile);
        GridLayout layout(form->rule().left(), form->rule().top() + rule("dialog.gap"));
        layout.setGridSize(2, 0);
        layout.setColumnAlignment(0, ui::AlignRight);
        layout << *LabelWidget::newWithText(tr("Game:"), form)
               << *d->gameChoice
               << *LabelWidget::newWithText("Data File:", form)
               << *d->customDataFileName
               << *LabelWidget::newWithText(tr("Mods:"), form)
               << *d->packages;
        form->rule().setSize(layout);
    } {
        auto *launchOptions =
            FoldPanelWidget::makeOptionsGroup("launch-options", "Launch Options", &area());
        launchOptions->title().margins().setTop(rule("gap") + d->form->margins().bottom());
        auto *form = new GuiWidget;
        launchOptions->setContent(form);
        form->add(d->autoStartMap = new ChoiceWidget);
        d->autoStartMap->popup().menu().enableIndicatorDraw(true);
        d->autoStartMap->items() << new ChoiceItem(tr("Title screen"), "");
        form->add(d->autoStartSkill = new ChoiceWidget);
        d->autoStartSkill->items()
                << new ChoiceItem(tr("Novice"),    1)
                << new ChoiceItem(tr("Easy"),      2)
                << new ChoiceItem(tr("Normal"),    3)
                << new ChoiceItem(tr("Hard"),      4)
                << new ChoiceItem(tr("Nightmare"), 5);
        d->autoStartSkill->disable();
        d->autoStartSkill->setSelected(2);
        GridLayout layout(form->rule().left(), form->rule().top());
        layout.setGridSize(2, 0);
        layout.setColumnAlignment(0, ui::AlignRight);
        layout << *LabelWidget::newWithText(tr("Starts in:"), form) << *d->autoStartMap
               << *LabelWidget::newWithText(tr("Skill:"), form) << *d->autoStartSkill;
        form->rule().setSize(layout);
    } {
        d->optionsFold =
            FoldPanelWidget::makeOptionsGroup("gameplay-options", "Gameplay Options", &area());
        d->optionsBase = new GuiWidget;
        d->optionsFold->setContent(d->optionsBase);
    }
    buttons().clear()
            << new DialogButtonItem(Id1 | Default | Accept, tr("Create"))
            << new DialogButtonItem(Reject);
    buttonWidget(Id1)->disable();
    updateLayout(IncludeHidden);
    d->updateDataFile();
    d->gameChanged();
    d->populateOptions();
    connect(d->gameChoice, &ChoiceWidget::selectionChanged, [this] () { d->gameChanged(); });
    connect(d->packages, &PackagesButtonWidget::packageSelectionChanged,
            [this] (QStringList) { d->updateMapList(); });
    connect(d->autoStartMap, &ChoiceWidget::selectionChanged, [this] () {
        d->autoStartSkill->disable(d->autoStartMap->items().isEmpty() ||
                                   d->autoStartMap->selectedItem().data().toString().isEmpty());
    });
    connect(&editor(), &LineEditWidget::editorContentChanged,
            [this] () { d->checkValidProfileName(); });
}
GameProfile *CreateProfileDialog::makeProfile() const {
    auto *prof = new GameProfile(profileName());
    prof->setUserCreated(true);
    applyTo(*prof);
    if (!prof->saveLocationId()) {
        prof->createSaveLocation();
    }
    return prof;
}
void CreateProfileDialog::fetchFrom(GameProfile const &profile) {
    editor().setText(profile.name());
    d->gameChoice->setSelected(d->gameChoice->items().findData(profile.gameId()));
    d->tempProfile->setCustomDataFile(profile.customDataFile());
    d->packages->setPackages(profile.packages());
    d->updateDataFile();
    d->autoStartMap->setSelected(d->autoStartMap->items().findData(profile.autoStartMap()));
    d->autoStartSkill->setSelected(d->autoStartSkill->items().findData(profile.autoStartSkill()));
    d->tempProfile->setSaveLocationId(profile.saveLocationId());
    d->tempProfile->objectNamespace() = profile.objectNamespace();
    d->populateOptions();
}
void CreateProfileDialog::applyTo(GameProfile &profile) const {
    profile.setName(profileName());
    if (d->gameChoice->isValidSelection()) {
        profile.setGame(d->gameChoice->selectedItem().data().toString());
    }
    profile.setCustomDataFile(d->tempProfile->customDataFile());
    profile.setUseGameRequirements(true);
    profile.setPackages(d->packages->packages());
    profile.setAutoStartMap(d->autoStartMap->selectedItem().data().toString());
    profile.setAutoStartSkill(d->autoStartSkill->selectedItem().data().toInt());
    profile.setSaveLocationId(profile.saveLocationId());
    profile.objectNamespace() = d->tempProfile->objectNamespace();
}
String CreateProfileDialog::profileName() const {
    return editor().text().strip();
}
void CreateProfileDialog::operator>>(PersistentState &toState) const {
    auto &ps = toState.objectNamespace();
    auto &launch = find("launch-options")->as<FoldPanelWidget>();
    ps.set(name().concatenateMember(launch.name()) + ".open", launch.isOpen());
    ps.set(name().concatenateMember(d->optionsFold->name()) + ".open",
           d->optionsFold->isOpen());
}
void CreateProfileDialog::operator<<(const PersistentState &fromState) {
    auto &ps = fromState.objectNamespace();
    auto &launch = find("launch-options")->as<FoldPanelWidget>();
    if (ps.getb(name().concatenateMember(launch.name()) + ".open", true)) {
        launch.open();
    }
    else {
        launch.close(0.0);
    }
    if (ps.getb(name().concatenateMember(d->optionsFold->name()) + ".open", false)) {
        d->optionsFold->open();
    }
    else {
        d->optionsFold->close(0.0);
    }
}
CreateProfileDialog *CreateProfileDialog::editProfile(String const &gameFamily,
                                                      GameProfile &profile) {
    auto *dlg = new CreateProfileDialog(gameFamily);
    dlg->d->editing = true;
    dlg->d->oldName = profile.name();
    dlg->title().setText(tr("Edit Profile"));
    delete &dlg->message();
    dlg->buttonWidget(Id1)->setText(_E(b) + tr("OK"));
    dlg->fetchFrom(profile);
    dlg->updateLayout(IncludeHidden);
    return dlg;
}
DENG2_PIMPL_NOREF(DataFileSettingsDialog) {
    Variable &pkgFolders  = Config::get("resource.packageFolder");
    Id searchGroup;
    bool modified = false;
};
DataFileSettingsDialog::DataFileSettingsDialog(String const &name)
    : DirectoryListDialog(name)
    , d(new Impl) {
    buttons().remove(1); // remove the Cancel
    buttons().at(0).setLabel("Apply");
    title().setFont("heading");
    title().setText(tr("Data Files"));
    title().setStyleImage("package.icon", "heading");
    message().hide();
    d->searchGroup = addGroup(
        tr("Search Folders"),
        tr("The following folders are searched for game IWAD files and mods like PWADs, PK3s, and "
           "Doomsday packages. Toggle the " _E(b) "Subdirs" _E(.)
           " option to include all subfolders as well."));
    setValue(d->searchGroup, d->pkgFolders.value());
    connect(this, &DirectoryListDialog::arrayChanged, [this] () {
        d->modified = true;
    });
    updateLayout();
}
void DataFileSettingsDialog::finish(int result) {
    DirectoryListDialog::finish(result);
    if (d->modified) {
        d->pkgFolders.set(value(d->searchGroup));
        ClientWindow::main().taskBar().close();
        DoomsdayApp::app().initPackageFolders();
    }
}
DENG_GUI_PIMPL(DownloadDialog) {
    ProgressWidget *progress;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        progress = new ProgressWidget;
        area.add(progress);
        progress->setAlignment(ui::AlignLeft);
        progress->setSizePolicy(ui::Fixed, ui::Expand);
        progress->setRange(Rangei(0, 100));
        progress->rule()
                .setLeftTop(area.contentRule().left(), area.contentRule().top())
                .setInput(Rule::Width, rule("dialog.download.width"));
        area.setContentSize(progress->rule());
        self().buttons() << new DialogButtonItem(DialogWidget::Reject,
                                                 "Cancel Download",
                                                 new CallbackAction([this] () { self().cancel(); }));
        updateStyle();
    }
    void updateStyle() {
        progress->setImageScale(.4f); //pointsToPixels(.4f));
    }
};
DownloadDialog::DownloadDialog(String const &name)
    : DialogWidget(name)
    , d(new Impl(this))
{}
DownloadDialog::~DownloadDialog()
{}
ProgressWidget &DownloadDialog::progressIndicator() {
    return *d->progress;
}
void DownloadDialog::updateStyle() {
    DialogWidget::updateStyle();
    d->updateStyle();
}
DENG_GUI_PIMPL(FileDownloadDialog)
, DENG2_OBSERVES(shell::PackageDownloader, Status) {
    shell::PackageDownloader &downloader;
    String message = tr("Downloading data files...");
    Impl(Public *i, shell::PackageDownloader &downloader)
        : Base(i)
        , downloader(downloader) {
        downloader.audienceForStatus() += this;
        self().progressIndicator().setText(tr("%1\n" _E(l)_E(F) "%2"
                                              DENG2_CHAR_MDASH " files / "
                                              DENG2_CHAR_MDASH " MB")
                                           .arg(message)
                                           .arg(_E(l)));
    }
    void downloadStatusUpdate(Rangei64 const &bytes, Rangei const &files) {
        if (!bytes.end) return;
        auto &indicator = self().progressIndicator();
        indicator.setProgress(round<int>(100.0 * double(bytes.size())/double(bytes.end)));
        indicator.setText(tr("%1\n" _E(l)_E(F) "%2 file%3 / %4 MB")
                          .arg(message)
                          .arg(files.start)
                          .arg(DENG2_PLURAL_S(files.start))
                          .arg(bytes.start/1.0e6, 0, 'f', 1));
    }
};
FileDownloadDialog::FileDownloadDialog(shell::PackageDownloader &downloader)
    : d(new Impl(this, downloader))
{}
void FileDownloadDialog::cancel() {
    d->downloader.cancel();
}
void FileDownloadDialog::finish(int result) {
    if (!result) cancel();
    DownloadDialog::finish(result);
}
DENG_GUI_PIMPL(InputSettingsDialog) {
    ChoiceWidget *gamepad;
    ButtonWidget *applyGamepad;
    VariableToggleWidget *syncMouse;
    CVarToggleWidget *syncInput;
    CVarSliderWidget *mouseSensiX;
    CVarSliderWidget *mouseSensiY;
    ToggleWidget *mouseDisableX;
    ToggleWidget *mouseDisableY;
    ToggleWidget *mouseInvertX;
    ToggleWidget *mouseInvertY;
    ToggleWidget *mouseFilterX;
    ToggleWidget *mouseFilterY;
    CVarToggleWidget *joyEnable;
    GridPopupWidget *devPopup;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        area.add(joyEnable     = new CVarToggleWidget("input-joy"));
        area.add(gamepad       = new ChoiceWidget);
        area.add(applyGamepad  = new ButtonWidget);
        area.add(syncMouse     = new VariableToggleWidget(App::config("input.mouse.syncSensitivity")));
        area.add(mouseSensiX   = new CVarSliderWidget("input-mouse-x-factor"));
        area.add(mouseInvertX  = new ToggleWidget);
        area.add(mouseFilterX  = new ToggleWidget);
        area.add(mouseDisableX = new ToggleWidget);
        area.add(mouseSensiY   = new CVarSliderWidget("input-mouse-y-factor"));
        area.add(mouseInvertY  = new ToggleWidget);
        area.add(mouseFilterY  = new ToggleWidget);
        area.add(mouseDisableY = new ToggleWidget);
        gamepad->items() << new ChoiceItem(tr("None"), "");
        QStringList ids = ClientApp::inputSystem().gameControllerPresets().ids();
        ids.sort(Qt::CaseInsensitive);
        foreach (QString id, ids) {
            gamepad->items() << new ChoiceItem(id, id);
        }
        syncInput = new CVarToggleWidget("input-sharp");
        self().add(devPopup = new GridPopupWidget);
        devPopup->addSpanning(syncInput);
        *devPopup << LabelWidget::newWithText(tr("Key Grabber:"))
                  << new KeyGrabberWidget;
        devPopup->commit();
    }
    void fetch() {
        syncInput->updateFromCVar();
        mouseSensiX->updateFromCVar();
        mouseSensiY->updateFromCVar();
        joyEnable->updateFromCVar();
        mouseDisableX->setActive(Con_GetInteger("input-mouse-x-flags") & IDA_DISABLED);
        mouseDisableY->setActive(Con_GetInteger("input-mouse-y-flags") & IDA_DISABLED);
        mouseInvertX->setActive(Con_GetInteger("input-mouse-x-flags") & IDA_INVERT);
        mouseInvertY->setActive(Con_GetInteger("input-mouse-y-flags") & IDA_INVERT);
        mouseFilterX->setInactive(Con_GetInteger("input-mouse-x-flags") & IDA_RAW);
        mouseFilterY->setInactive(Con_GetInteger("input-mouse-y-flags") & IDA_RAW);
        enableOrDisable();
        gamepad->setSelected(gamepad->items().findData(
                                 ClientApp::inputSystem().gameControllerPresets().currentPreset()));
    }
    void enableOrDisable() {
        mouseSensiX->disable(mouseDisableX->isActive());
        mouseSensiY->disable(mouseDisableY->isActive());
        mouseInvertX->disable(mouseDisableX->isActive());
        mouseInvertY->disable(mouseDisableY->isActive());
        mouseFilterX->disable(mouseDisableX->isActive());
        mouseFilterY->disable(mouseDisableY->isActive());
    }
    void updateMouseFlags() {
        Con_SetInteger("input-mouse-x-flags",
                       (mouseDisableX->isActive()?  IDA_DISABLED : 0) |
                       (mouseInvertX->isActive()?   IDA_INVERT   : 0) |
                       (mouseFilterX->isInactive()? IDA_RAW      : 0));
        Con_SetInteger("input-mouse-y-flags",
                       (mouseDisableY->isActive()?  IDA_DISABLED : 0) |
                       (mouseInvertY->isActive()?   IDA_INVERT   : 0) |
                       (mouseFilterY->isInactive()? IDA_RAW      : 0));
        enableOrDisable();
    }
};
InputSettingsDialog::InputSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Input Settings"));
    heading().setImage(style().images().image("input"));
    d->syncInput->setText(tr("Vanilla 35Hz Input Rate"));
    d->syncMouse->setText(tr("Sync Axis Sensitivities"));
    d->applyGamepad->setText(tr("Apply"));
    connect(d->applyGamepad, SIGNAL(pressed()), this, SLOT(applyControllerPreset()));
    LabelWidget *mouseXLabel = LabelWidget::appendSeparatorWithText("Mouse: Horizontal", &area());
    LabelWidget *mouseYLabel = LabelWidget::appendSeparatorWithText("Mouse: Vertical", &area());
    LabelWidget *applyNote = LabelWidget::newWithText(tr("Clicking " _E(b) "Apply" _E(.) " will remove all "
                                                         "existing game controller bindings and apply "
                                                         "the selected preset."), &area());
    applyNote->margins().setTop("");
    applyNote->setFont("separator.annotation");
    applyNote->setTextColor("altaccent");
    applyNote->setTextLineAlignment(ui::AlignLeft);
    d->mouseSensiX->setRange(Rangef(.5f, 75.f));
    d->mouseSensiX->setDisplayFactor(0.1);
    d->mouseSensiY->setRange(Rangef(.5f, 75.f));
    d->mouseSensiY->setDisplayFactor(0.1);
    connect(d->mouseSensiX, SIGNAL(valueChangedByUser(double)), this, SLOT(mouseSensitivityChanged(double)));
    connect(d->mouseSensiY, SIGNAL(valueChangedByUser(double)), this, SLOT(mouseSensitivityChanged(double)));
    d->mouseInvertX->setText(tr("Invert X Axis"));
    d->mouseDisableX->setText(tr("Disable X Axis"));
    d->mouseFilterX->setText(tr("Filter X Axis"));
    d->mouseInvertY->setText(tr("Invert Y Axis"));
    d->mouseDisableY->setText(tr("Disable Y Axis"));
    d->mouseFilterY->setText(tr("Filter Y Axis"));
    connect(d->mouseInvertX, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    connect(d->mouseInvertY, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    connect(d->mouseDisableX, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    connect(d->mouseDisableY, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    connect(d->mouseFilterX, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    connect(d->mouseFilterY, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)), this, SLOT(mouseTogglesChanged()));
    d->joyEnable->setText(tr("Game Controllers Enabled"));
    d->syncMouse->margins().setBottom("gap");
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(2, 0);
    layout.append(*d->joyEnable, 2);
    layout << *LabelWidget::newWithText(tr("Game Controller Preset:"), &area())
           << *d->gamepad;
    d->applyGamepad->setSizePolicy(ui::Expand, ui::Expand);
    d->applyGamepad->rule()
            .setInput(Rule::Left, d->gamepad->rule().right())
            .setMidAnchorY(d->gamepad->rule().midY());
    layout.append(*applyNote, 2);
    GridLayout layout2(area().contentRule().left(), applyNote->rule().bottom());
    layout2.setGridSize(2, 0);
    layout2 << *mouseXLabel      << *mouseYLabel
            << *d->mouseSensiX   << *d->mouseSensiY;
    layout2.append(*d->syncMouse, 2);
    layout2 << *d->mouseInvertX  << *d->mouseInvertY
            << *d->mouseFilterX  << *d->mouseFilterY
            << *d->mouseDisableX << *d->mouseDisableY;
    applyNote->setMaximumTextWidth(layout2.width() - rule("dialog.gap"));
    area().setContentSize(OperatorRule::maximum(layout.width(), layout2.width()),
                          layout.height() + layout2.height());
    buttons()
            << new DialogButtonItem(Default | Accept, tr("Close"))
            << new DialogButtonItem(Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())))
            << new DialogButtonItem(ActionPopup | Id1,
                                    style().images().image("gauge"));
    popupButtonWidget(Id1)->setPopup(*d->devPopup);
    d->fetch();
}
void InputSettingsDialog::resetToDefaults() {
    ClientApp::inputSystem().settings().resetToDefaults();
    d->fetch();
}
void InputSettingsDialog::mouseTogglesChanged() {
    d->updateMouseFlags();
}
void InputSettingsDialog::mouseSensitivityChanged(double value) {
    if (d->syncMouse->isActive()) {
        if (sender() == d->mouseSensiX) {
            d->mouseSensiY->setValue(value);
            d->mouseSensiY->setCVarValueFromWidget();
        }
        else {
            d->mouseSensiX->setValue(value);
            d->mouseSensiX->setCVarValueFromWidget();
        }
    }
}
void InputSettingsDialog::applyControllerPreset() {
    String const presetId = d->gamepad->selectedItem().data().toString();
    ClientApp::inputSystem().gameControllerPresets().applyPreset(presetId);
}
struct DomainText {
    char const *name;
    char const *label;
};
static DomainText const domainText[] = {
    { "generic",    "Minimum Level" },
    { "resource",   "Resources" },
    { "map",        "Map"       },
    { "script",     "Scripting" },
    { "gl",         "Graphics"  },
    { "audio",      "Audio"     },
    { "input",      "Input"     },
    { "network",    "Network"   }
};
#define NUM_DOMAINS (sizeof(domainText)/sizeof(domainText[0]))
DENG2_PIMPL(LogSettingsDialog)
, DENG2_OBSERVES(ToggleWidget, Toggle) {
    ui::ListData levels;
    VariableToggleWidget *separately;
    FoldPanelWidget *fold;
    GridLayout foldLayout;
    IndirectRule *columnWidth; ///< Sync column width in and out of the fold.
    struct DomainWidgets {
        LabelWidget *label;
        VariableChoiceWidget *level;
        VariableToggleWidget *dev;
        VariableToggleWidget *alert;
    };
    DomainWidgets domWidgets[NUM_DOMAINS];
    Impl(Public *i) : Base(i) {
        try {
            zap(domWidgets);
            columnWidth = new IndirectRule;
            separately = new VariableToggleWidget(tr("Filter by Subsystem"),
                                                  App::config("log.filterBySubsystem"));
            levels << new ChoiceItem(        tr("1 - X.Verbose"), LogEntry::XVerbose)
                   << new ChoiceItem(        tr("2 - Verbose"),   LogEntry::Verbose )
                   << new ChoiceItem(        tr("3 - Message"),   LogEntry::Message )
                   << new ChoiceItem(        tr("4 - Note"),      LogEntry::Note    )
                   << new ChoiceItem(_E(D) + tr("5 - Warning"),   LogEntry::Warning )
                   << new ChoiceItem(_E(D) + tr("6 - Error"),     LogEntry::Error   )
                   << new ChoiceItem(_E(D) + tr("7 - Critical"),  LogEntry::Critical);
            fold = new FoldPanelWidget;
            fold->setContent(new GuiWidget);
            foldLayout.setLeftTop(fold->content().rule().left(),
                                  fold->content().rule().top());
            foldLayout.setGridSize(4, 0);
            foldLayout.setColumnFixedWidth(1, *columnWidth);
            foldLayout.setColumnAlignment(0, ui::AlignRight);
            for (uint i = 0; i < NUM_DOMAINS; ++i) {
                initDomain(domainText[i],
                           domWidgets[i],
                           i == 0? &self().area() : &fold->content());
            }
            self().area().add(separately);
            self().area().add(fold);
            columnWidth->setSource(domWidgets[0].level->maximumWidth());
            separately->audienceForToggle() += this;
        }
        catch (Error const &er) {
            LOGDEV_ERROR("") << er.asText();
            deinit();
            throw;
        }
    }
    ~Impl() {
        deinit();
        releaseRef(columnWidth);
    }
    void initDomain(DomainText const &dom, DomainWidgets &wgt, GuiWidget *parent) {
        wgt.label = LabelWidget::newWithText(tr(dom.label) + ":", parent);
        parent->add(wgt.level = new VariableChoiceWidget(
                        Config::get(String("log.filter.%1.minLevel").arg(dom.name)),
                        VariableChoiceWidget::Number));
        wgt.level->setItems(levels);
        wgt.level->updateFromVariable();
        QObject::connect(wgt.level, SIGNAL(selectionChangedByUser(uint)),
                         thisPublic, SLOT(updateLogFilter()));
        parent->add(wgt.dev =
                new VariableToggleWidget(tr("Dev"), Config::get(String("log.filter.%1.allowDev").arg(dom.name))));
        QObject::connect(wgt.dev, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)),
                         thisPublic, SLOT(updateLogFilter()));
        parent->add(wgt.alert =
                new VariableToggleWidget(tr("Alerts"), Config::get(String("alert.") + dom.name)));
        wgt.alert->setActiveValue(LogEntry::Warning);
        wgt.alert->setInactiveValue(LogEntry::HighestLogLevel + 1);
        if (parent == &fold->content()) {
            foldLayout << *wgt.label << *wgt.level << *wgt.dev << *wgt.alert;
        }
    }
    void deinit() {
        for (uint i = 0; i < NUM_DOMAINS; ++i) {
            if (domWidgets[i].level) {
                domWidgets[i].level->useDefaultItems();
            }
        }
    }
    void overrideWithGeneric() {
        Config &cfg = App::config();
        LogFilter &logf = App::logFilter();
        LogEntry::Level minLevel = LogEntry::Level(domWidgets[0].level->selectedItem().data().toInt());
        bool allowDev = domWidgets[0].dev->isActive();
        bool alerts = domWidgets[0].alert->isActive();
        logf.setMinLevel(LogEntry::AllDomains, minLevel);
        logf.setAllowDev(LogEntry::AllDomains, allowDev);
        logf.write(cfg.objectNamespace().subrecord("log.filter"));
        for (uint i = 0; i < NUM_DOMAINS; ++i) {
            char const *name = domainText[i].name;
            cfg.set(String("alert.") + name, int(alerts? LogEntry::Warning : (LogEntry::HighestLogLevel + 1)));
        }
    }
    void toggleStateChanged(ToggleWidget &toggle) {
        if (toggle.isActive()) {
            overrideWithGeneric();
            fold->open();
        }
        else {
            fold->close();
            overrideWithGeneric();
        }
    }
    void updateLogFilter() {
        if (separately->isInactive()) {
            overrideWithGeneric();
        }
        applyFilterFromConfig();
    }
    void applyFilterFromConfig() {
        App::logFilter().read(App::config().subrecord("log.filter"));
    }
};
LogSettingsDialog::LogSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Log Filter & Alerts"));
    heading().setImage(style().images().image("log"));
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(4, 0);
    layout.setColumnFixedWidth(1, *d->columnWidth);
    layout.setColumnAlignment(0, ui::AlignRight);
    layout << *d->domWidgets[0].label
           << *d->domWidgets[0].level
           << *d->domWidgets[0].dev
           << *d->domWidgets[0].alert
           << Const(0);
    layout.append(*d->separately, 3);
    layout.append(*d->fold, 4)
           << Const(0);
    d->fold->content().rule().setSize(d->foldLayout.width(), d->foldLayout.height());
    area().setContentSize(layout);
    buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())));
    if (d->separately->isActive()) {
        d->fold->open();
    }
}
void LogSettingsDialog::resetToDefaults() {
    ClientApp::logSettings().resetToDefaults();
    d->applyFilterFromConfig();
}
void LogSettingsDialog::updateLogFilter() {
    d->updateLogFilter();
}
DENG2_PIMPL(ManualConnectionDialog)
, DENG2_OBSERVES(ServerLink, DiscoveryUpdate)
, DENG2_OBSERVES(MultiplayerServerMenuWidget, AboutToJoin) {
    String usedAddress;
    FoldPanelWidget *fold;
    MultiplayerServerMenuWidget *servers;
    ProgressWidget *progress;
    bool querying;
    bool joinWhenEnterPressed;
    bool autoJoin;
    Impl(Public *i)
        : Base(i)
        , querying(false)
        , joinWhenEnterPressed(false)
        , autoJoin(true) {
        ClientApp::serverLink().audienceForDiscoveryUpdate() += this;
    }
    void linkDiscoveryUpdate(ServerLink const &link) override {
        if (querying) {
            querying = false;
            progress->setRotationSpeed(0);
            self().editor().enable();
            self().validate();
            if (link.foundServerCount(ServerLink::Direct) > 0) {
                progress->hide();
                fold->open();
                if (link.foundServerCount(ServerLink::Direct) == 1) {
                    joinWhenEnterPressed = true;
                }
            }
            else {
                fold->close(0);
                progress->setRotationSpeed(0);
                progress->setText(_E(l) + tr("No response"));
                progress->setOpacity(0, 4, 2);
            }
        }
    }
    ButtonWidget &connectButton() {
        return self().buttonWidget(tr("Connect"));
    }
    void aboutToJoinMultiplayerGame(shell::ServerInfo const &) override {
        self().accept();
    }
};
ManualConnectionDialog::ManualConnectionDialog(String const &name)
    : InputDialog(name), d(new Impl(this)) {
    area().enableIndicatorDraw(true);
    add(d->progress = new ProgressWidget);
    d->progress->useMiniStyle("altaccent");
    d->progress->setWidthPolicy(ui::Expand);
    d->progress->setTextAlignment(ui::AlignLeft);
    d->progress->hide();
    d->fold = new FoldPanelWidget;
    d->servers = new MultiplayerServerMenuWidget(MultiplayerServerMenuWidget::DirectDiscoveryOnly);
    d->servers->audienceForAboutToJoin() += d;
    d->servers->margins().setLeft("dialog.gap");
    /*connect(d->servers, SIGNAL(sessionSelected(de::ui::Item const *)),
            this,     SIGNAL(selected(de::ui::Item const *)));
    connect(d->servers, SIGNAL(sessionSelected(de::ui::Item const *)),
            this,     SLOT  (serverSelected(de::ui::Item const *)));*/
    d->servers->rule().setInput(Rule::Width, rule().width() - margins().width());
    d->fold->setContent(d->servers);
    area().add(d->fold);
    title().setText(tr("Connect to Server"));
    message().setText(tr("Enter the IP address or domain name of the multiplayer server you want to connect to. "
                         "Optionally, you may include a TCP port number, for example "
                         _E(b) "10.0.1.1:13209" _E(.) "."));
    buttons().clear()
            << new DialogButtonItem(Default, tr("Connect"),
                                    new SignalAction(this, SLOT(queryOrConnect())))
            << new DialogButtonItem(Reject);
    d->connectButton().disable();
    d->progress->rule()
            .setInput(Rule::Top,    buttonsMenu().rule().top())
            .setInput(Rule::Right,  buttonsMenu().rule().left())
            .setInput(Rule::Height, buttonsMenu().rule().height() - margins().bottom());
    disconnect(&editor(), SIGNAL(enterPressed(QString)), this, SLOT(accept()));
    connect(&editor(), SIGNAL(enterPressed(QString)), this, SLOT(queryOrConnect()));
    connect(&editor(), SIGNAL(editorContentChanged()), this, SLOT(validate()));
    connect(&editor(), SIGNAL(editorContentChanged()), this, SLOT(contentChanged()));
    updateLayout(IncludeHidden); // fold widgets are hidden while closed
}
void ManualConnectionDialog::enableJoinWhenSelected(bool joinWhenSelected) {
    d->autoJoin = joinWhenSelected;
}
void ManualConnectionDialog::operator>>(PersistentState &toState) const {
    toState.objectNamespace().set(name() + ".address", d->usedAddress);
}
void ManualConnectionDialog::operator<<(PersistentState const &fromState) {
    d->usedAddress = fromState[name() + ".address"];
    editor().setText(d->usedAddress);
    validate();
}
void ManualConnectionDialog::queryOrConnect() {
    if (d->connectButton().isDisabled()) {
        return;
    }
    if (!d->querying) {
        if (d->joinWhenEnterPressed) {
            d->servers->childWidgets().first()->as<MultiplayerPanelButtonWidget>()
                    .joinButton().trigger();
            return;
        }
        d->joinWhenEnterPressed = false;
        d->querying = true;
        d->progress->setText(_E(l) + tr("Looking for host..."));
        d->progress->setRotationSpeed(40);
        d->progress->show();
        d->progress->setOpacity(1);
        editor().disable();
        validate();
        ClientApp::serverLink().discover(editor().text());
    }
}
void ManualConnectionDialog::contentChanged() {
    d->joinWhenEnterPressed = false;
}
void ManualConnectionDialog::validate() {
    bool valid = true;
    if (d->querying) {
        valid = false;
    }
    if (editor().text().isEmpty()     ||
        editor().text().contains(';') ||
        editor().text().endsWith(":") ||
        (editor().text().startsWith(":") && !editor().text().startsWith("::"))) {
        valid = false;
    }
    d->connectButton().enable(valid);
}
/*
void ManualConnectionDialog::serverSelected(ui::Item const *item) {
    if (d->autoJoin) {
        setAcceptanceAction(makeAction(*item));
    }
    accept();
}*/
void ManualConnectionDialog::finish(int result) {
    if (result) {
        d->usedAddress = editor().text();
    }
    InputDialog::finish(result);
}
DENG_GUI_PIMPL(NetworkSettingsDialog) {
    VariableToggleWidget *localPackages;
    VariableLineEditWidget *webApiUrl;
    GridPopupWidget *devPopup;
    CVarToggleWidget *devInfo;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        area.add(localPackages = new VariableToggleWidget(tr("Local Multiplayer Mods"),
                                                          App::config("resource.localPackages")));
        self().add(devPopup = new GridPopupWidget);
        *devPopup << LabelWidget::newWithText(tr("Web API:"))
                  << (webApiUrl = new VariableLineEditWidget(App::config("apiUrl")))
                  << Const(0)
                  << (devInfo = new CVarToggleWidget("net-dev"));
        devPopup->commit();
    }
    void fetch() {
        foreach (GuiWidget *w, self().area().childWidgets() + devPopup->content().childWidgets()) {
            if (ICVarWidget *cv = maybeAs<ICVarWidget>(w)) {
                cv->updateFromCVar();
            }
        }
    }
};
NetworkSettingsDialog::NetworkSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Network Settings"));
    heading().setStyleImage("network");
    d->devInfo->setText(tr("Developer Info"));
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(1, 0);
    layout << *d->localPackages;
    auto *caution = LabelWidget::newWithText(
                tr("Caution: Loading additional add-ons or mods may cause gameplay bugs "
                   "or client instability in multiplayer games."), &area());
    caution->margins().setTop("");
    caution->setTextLineAlignment(ui::AlignLeft);
    caution->setMaximumTextWidth(area().rule().width() - area().margins().width());
    caution->setFont("separator.annotation");
    caution->setTextColor("altaccent");
    layout << *caution;
    area().setContentSize(layout);
    buttons()
            << new DialogButtonItem(Default | Accept, tr("Close"))
            << new DialogButtonItem(Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())))
            << new DialogButtonItem(ActionPopup | Id1, style().images().image("gauge"));
    popupButtonWidget(Id1)->setPopup(*d->devPopup);
    d->fetch();
}
void NetworkSettingsDialog::resetToDefaults() {
    ClientApp::networkSettings().resetToDefaults();
    d->fetch();
}
DENG2_PIMPL(PackageCompatibilityDialog) {
    String          message;
    StringList      wanted;
    bool            conflicted    = false;
    PackagesWidget *list          = nullptr;
    ToggleWidget *  ignoreToggle  = nullptr;
    LabelWidget *   ignoreWarning = nullptr;
    ui::ListData    actions;
    ProgressWidget *updating;
    bool            ignoreCheck = false;
    LoopCallback    mainCall;
    Impl(Public *i) : Base(i) {
        self().add(updating = new ProgressWidget);
        self().area().enableIndicatorDraw(true);
        updating->setSizePolicy(ui::Expand, ui::Expand);
        updating->useMiniStyle("altaccent");
        updating->setText(tr("Updating..."));
        updating->setTextAlignment(ui::AlignLeft);
        updating->setMode(ProgressWidget::Indefinite);
        updating->setOpacity(0);
        updating->rule()
                .setInput(Rule::Top,    self().buttonsMenu().rule().top())
                .setInput(Rule::Right,  self().buttonsMenu().rule().left())
                .setInput(Rule::Height, self().buttonsMenu().rule().height() - self().margins().bottom());
    }
    String defaultButtonLabel() const {
        if (!list) return "";
        if (ignoreCheck) {
            return _E(b)_E(D) + tr("Ignore and Continue");
        }
        if (list->itemCount() > 0) {
            return _E(b) + tr("Load Mods");
        }
        return _E(b) + tr("Unload Mods");
    }
    void enableIgnore(bool yes) {
        ignoreCheck = yes;
        if (auto *button = self().buttonWidget(Id1)) {
            button->setText(defaultButtonLabel());
        }
        if (auto *button = self().buttonWidget(Id2)) {
            button->enable(ignoreCheck);
        }
    }
    void update() {
        if (list) {
            delete list;
            list = nullptr;
        }
        if (ignoreWarning) {
            delete ignoreWarning;
            ignoreWarning = nullptr;
        }
        if (ignoreToggle) {
            delete ignoreToggle;
            ignoreToggle = nullptr;
        }
        self().buttons().clear();
        try {
            actions << new ui::SubwidgetItem(tr("..."), ui::Up, [this]() -> PopupWidget * {
                return new PackageInfoDialog(list->actionPackage(),
                                             PackageInfoDialog::InformationOnly);
            });
            self().area().add(ignoreToggle = new ToggleWidget);
            ignoreToggle->setText("Ignore incompatible/missing mods");
            ignoreToggle->setAlignment(ui::AlignLeft);
            ignoreToggle->setActionFn([this]() {
                enableIgnore(ignoreToggle->isActive());
            });
            self().area().add(ignoreWarning = new LabelWidget);
            ignoreWarning->setText(_E(b) "Caution: " _E(.) "Playing without the right resources "
                                                           "may cause a crash.");
            ignoreWarning->setFont("separator.annotation");
            ignoreWarning->setTextColor("altaccent");
            ignoreWarning->setAlignment(ui::AlignLeft);
            ignoreWarning->setTextLineAlignment(ui::AlignLeft);
            ignoreWarning->margins().setTop(ConstantRule::zero());
            StringList wantedUnavailable;
            StringList wantedAvailable;
            QList<std::pair<String, Version>> wantedDifferentVersionAvailable; {
                auto &pkgLoader = PackageLoader::get();
                for (const auto &id : wanted) {
                    if (pkgLoader.isAvailable(id)) {
                        wantedAvailable << id;
                    }
                    else {
                        auto id_ver = Package::split(id);
                        if (const auto *p = pkgLoader.select(id_ver.first)) {
                            wantedDifferentVersionAvailable.push_back(
                                {Package::versionedIdentifierForFile(*p), id_ver.second});
                        }
                        else {
                            wantedUnavailable << id;
                        }
                    }
                }
            }
            String unavailNote;
            if (!wantedDifferentVersionAvailable.empty()) {
                unavailNote +=
                    _E(b)_E(D) "Caution:" _E(.)_E(.) " There is a different version of "
                                                     "the following mods available:\n";
                for (const auto &avail_expected : wantedDifferentVersionAvailable) {
                    unavailNote +=
                        " - " _E(>) + Package::splitToHumanReadable(avail_expected.first) +
                        " " _E(l)_E(F) "\n(expected " +
                        avail_expected.second.fullNumber() + _E(<) ")\n" _E(w)_E(A);
                }
                unavailNote += "\n";
            }
            if (!wantedUnavailable.empty()) {
                unavailNote += "The following mods are missing:\n";
                for (const auto &id : wantedUnavailable) {
                    unavailNote += " - " + Package::splitToHumanReadable(id) + "\n";
                }
                unavailNote += "\n";
            }
            StringList wantedDifferent =
                map<StringList>(wantedDifferentVersionAvailable,
                                [](const std::pair<String, Version> &sv) { return sv.first; });
            self().area().add(list = new PackagesWidget(wantedAvailable + wantedDifferent));
            list->setAllowPackageInfoActions(false);
            list->setHiddenTags({"core", "gamedata"});
            list->setActionItems(actions);
            list->setActionsAlwaysShown(true);
            list->setFilterEditorMinimumY(self().area().rule().top());
            const StringList loaded = DoomsdayApp::loadedPackagesAffectingGameplay();
            if (!GameProfiles::arePackageListsCompatible(loaded, wanted)) {
                conflicted = true;
                if (list->itemCount() == 0) {
                    list->hide();
                }
                if (!wantedUnavailable.empty()) {
                    self().message().setText(message + "\n\n" + unavailNote +
                                             "Please locate the missing mods before continuing. "
                                             "You may need to add more folders in your Data Files "
                                             "settings so the mods can be found.");
                    if (list->itemCount() > 0) {
                        self().message().setText(
                            self().message().text() +
                            "\n\nThe mods listed below are required and available, "
                            "and should be loaded now (the highlighted ones are already loaded).");
                    }
                    self().buttons() << new DialogButtonItem(Default | Accept | Id2,
                                                             _E(b) _E(D) "Ignore and Continue");
                    self().buttonWidget(Id2)->disable();
                }
                else {
                    if (list->itemCount() > 0) {
                        self().message().setText(message + "\n\n" + unavailNote +
                                                 tr("All the mods listed below should be loaded."));
                        self().buttons()
                                << new DialogButtonItem(Default | Accept | Id1, defaultButtonLabel(),
                                                        new CallbackAction([this] () { resolvePackages(); }));
                    }
                    else {
                        self().message().setText(message + "\n\n" + unavailNote +
                                                 tr("All additional mods should be unloaded."));
                        self().buttons()
                                   << new DialogButtonItem(Default | Accept | Id1, defaultButtonLabel(),
                                                           new CallbackAction([this] () { resolvePackages(); }));
                    }
                }
                self().buttons()
                        << new DialogButtonItem(Reject, tr("Cancel"));
            }
        }
        catch (PackagesWidget::UnavailableError const &er) {
            conflicted = true;
            self().message().setText(message + "\n\n" + er.asText());
            self().buttons() << new DialogButtonItem(Default | Reject);
        }
        self().updateLayout();
    }
    static bool containsIdentifier(String const &identifier, StringList const &ids) {
        for (String i : ids) {
            if (Package::equals(i, identifier)) return true;
        }
        return false;
    }
    void resolvePackages() {
        if (ignoreCheck) {
            LOG_RES_NOTE("Ignoring package compatibility check due to user request!");
            self().accept();
            return;
        }
        LOG_RES_MSG("Resolving packages...");
        auto &pkgLoader = PackageLoader::get();
        StringList loaded = DoomsdayApp::loadedPackagesAffectingGameplay();
        int goodUntil = -1;
        for (int i = 0; i < loaded.size() && i < wanted.size(); ++i) {
            if (Package::equals(loaded.at(i), wanted.at(i))) {
                goodUntil = i;
            }
            else {
                break;
            }
        }
        LOG_RES_MSG("Good until %s") << goodUntil;
        for (int i = loaded.size() - 1; i > goodUntil; --i) {
            LOG_RES_MSG("Unloading excess ") << loaded.at(i);
            pkgLoader.unload(loaded.at(i));
            loaded.removeAt(i);
        }
        for (int i = goodUntil + 1; i < wanted.size(); ++i) {
            const auto &pkgId = wanted.at(i);
            if (pkgLoader.isAvailable(pkgId)) {
                LOG_RES_MSG("Loading wanted ") << wanted.at(i);
                pkgLoader.load(wanted.at(i));
            }
            else {
                auto id_ver = Package::split(pkgId);
                LOG_RES_MSG("Loading latest version of wanted ") << id_ver.first;
                pkgLoader.load(id_ver.first);
            }
        }
        LOG_RES_MSG("Packages affecting gameplay:\n")
            << String::join(DoomsdayApp::loadedPackagesAffectingGameplay(), "\n");
        self().buttonsMenu().disable();
        updating->setOpacity(1, 0.3);
        self().accept();
    }
};
PackageCompatibilityDialog::PackageCompatibilityDialog(String const &name)
    : MessageDialog(name)
    , d(new Impl(this)) {
    title().setText(tr("Incompatible Mods"));
}
void PackageCompatibilityDialog::setMessage(String const &msg) {
    d->message = msg;
}
void PackageCompatibilityDialog::setWantedPackages(StringList packages) {
    DENG2_ASSERT(!packages.isEmpty());
    d->wanted = packages;
    d->update();
}
bool PackageCompatibilityDialog::isCompatible() const {
    return !d->conflicted;
}
bool PackageCompatibilityDialog::handleEvent(Event const &event) {
    if (event.isKey()) {
        auto const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_LALT || key.ddKey() == DDKEY_RALT) {
            d->enableIgnore(key.type() != Event::KeyRelease);
        }
    }
    return MessageDialog::handleEvent(event);
}
DENG_GUI_PIMPL(PackageInfoDialog) {
    Mode                           mode;
    LabelWidget *                  title;
    LabelWidget *                  path;
    DocumentWidget *               description;
    LabelWidget *                  icon;
    LabelWidget *                  metaInfo;
    IndirectRule *                 targetHeight;
    IndirectRule *                 descriptionWidth;
    IndirectRule *                 minContentHeight;
    String                         packageId;
    String                         compatibleGame; // guessed
    NativePath                     nativePath;
    const DataBundle *             bundle = nullptr;
    SafeWidgetPtr<PopupWidget>     configurePopup;
    SafeWidgetPtr<PopupMenuWidget> profileMenu;
    enum MenuMode { AddToProfile, PlayInProfile };
    MenuMode menuMode;
    Impl(Public * i, Mode mode)
        : Base(i)
        , mode(mode) {
        targetHeight     = new IndirectRule;
        descriptionWidth = new IndirectRule;
        minContentHeight = new IndirectRule;
        self().buttons() << new DialogButtonItem(Default | Accept, tr("Close"));
        createWidgets();
    }
    ~Impl() {
        releaseRef(targetHeight);
        releaseRef(descriptionWidth);
        releaseRef(minContentHeight);
    }
    void createWidgets() {
        auto &area = self().area();
        title = LabelWidget::newWithText("", &area);
        title->setFont("title");
        title->setSizePolicy(ui::Filled, ui::Expand);
        title->setTextColor("accent");
        title->setTextLineAlignment(ui::AlignLeft);
        title->margins().setBottom("");
        path = LabelWidget::newWithText("", &area);
        path->setSizePolicy(ui::Filled, ui::Expand);
        path->setTextColor("text");
        path->setTextLineAlignment(ui::AlignLeft);
        path->margins().setTop("unit");
        AutoRef<Rule> contentHeight(OperatorRule::maximum(*minContentHeight, *targetHeight));
        description = new DocumentWidget;
        description->setFont("small");
        description->setStyleColor(Font::RichFormat::NormalColor, "label.dimmed");
        description->setStyleColor(Font::RichFormat::HighlightColor, "label.highlight");
        description->setStyleColor(Font::RichFormat::DimmedColor, "label.dimmed");
        description->setStyleColor(Font::RichFormat::AccentColor, "label.accent");
        description->setStyleColor(Font::RichFormat::DimAccentColor, "label.dimaccent");
        description->progress().setColor("progress.light.wheel");
        description->progress().setShadowColor("progress.light.shadow");
        description->setWidthPolicy(ui::Fixed);
        description->rule().setInput(
            Rule::Height, contentHeight - title->rule().height() - path->rule().height());
        area.add(description);
        SequentialLayout layout(area.contentRule().left(), area.contentRule().top(), ui::Down);
        layout.setOverrideWidth(*descriptionWidth);
        layout << *title << *path << *description;
        icon = LabelWidget::newWithText("", &area);
        icon->setSizePolicy(ui::Filled, ui::Expand);
        metaInfo = LabelWidget::newWithText("", &area);
        metaInfo->setSizePolicy(ui::Filled, ui::Expand);
        metaInfo->setTextLineAlignment(ui::AlignLeft);
        metaInfo->setFont("small");
        metaInfo->setTextColor("altaccent");
        SequentialLayout rightLayout(title->rule().right(), title->rule().top(), ui::Down);
        rightLayout.setOverrideWidth(rule("dialog.packageinfo.metadata.width"));
        rightLayout << *icon << *metaInfo;
        targetHeight->setSource(rightLayout.height());
        area.setContentSize(layout.width() + rightLayout.width(), contentHeight);
    }
    void setPackageIcon(Image const &iconImage) {
        icon->setImage(iconImage);
        icon->setImageColor(Vector4f(1, 1, 1, 1));
        icon->setImageFit(ui::FitToWidth | ui::OriginalAspectRatio);
        icon->setImageScale(1);
        icon->setBehavior(ContentClipping, true);
    }
    bool useGameTitlePicture() {
        DENG_ASSERT(bundle != nullptr);
        auto *lumpDir = bundle->lumpDirectory();
        if (!lumpDir || (!lumpDir->has("TITLEPIC") && !lumpDir->has("TITLE"))) {
            return false;
        }
        Game const &game = Games::get()[compatibleGame];
        res::LumpCatalog catalog;
        catalog.setPackages(game.requiredPackages() + StringList({packageId}));
        Image img = IdTech1Image::makeGameLogo(
            game, catalog, IdTech1Image::NullImageIfFails | IdTech1Image::UnmodifiedAppearance);
        if (!img.isNull()) {
            setPackageIcon(img);
            return true;
        }
        return false;
    }
    void useDefaultIcon() {
        icon->setStyleImage("package.large");
        icon->setImageColor(style().colors().colorf("altaccent"));
        icon->setImageFit(ui::FitToSize | ui::OriginalAspectRatio);
        icon->setOpacity(.5f);
        icon->setBehavior(ContentClipping, false);
    }
    bool useIconFromPackage(const String &packagePath) {
        try {
            for (const auto *ext : {".jpg", ".jpeg", ".png"}) {
                String const imgPath = packagePath / "icon" + ext;
                if (ImageFile const *img = FS::tryLocate<ImageFile const>(imgPath)) {
                    Image iconImage = img->image();
                    if (iconImage.width() > 512 || iconImage.height() > 512) {
                        throw Error("PackageInfoDialog::useIconFile",
                                    "Icon file " + img->description() +
                                        " is too large (max 512x512)");
                    }
                    setPackageIcon(iconImage);
                    return true;
                }
            }
        }
        catch (Error const &er) {
            LOG_RES_WARNING("Failed to use package icon image: %s") << er.asText();
        }
        return false;
    }
    void useIconFile(String const &packagePath) {
        if (!useIconFromPackage(packagePath)) {
            useDefaultIcon();
        }
    }
    bool setup(File const *file) {
        self().setOutlineColor("popup.outline");
        if (!file) return false; // Not a package?
        Record const &names = file->objectNamespace();
        if (!names.has(Package::VAR_PACKAGE)) {
            return false;
        }
        Record const &meta = names.subrecord(Package::VAR_PACKAGE);
        packageId       = Package::versionedIdentifierForFile(*file);
        nativePath      = file->correspondingNativePath();
        String fileDesc = file->source()->description();
        String lumpDirCrc32;
        String format;
        bundle = maybeAs<DataBundle>(file->target());
        if (bundle) {
            format         = bundle->formatAsText().upperFirstChar();
            compatibleGame = bundle->guessCompatibleGame();
            if (!useIconFromPackage(file->path())) {
                if (!useGameTitlePicture()) {
                    useDefaultIcon();
                }
            }
            if (bundle->format() == DataBundle::Collection) {
                fileDesc = file->target().description();
            }
            else if (bundle->format() == DataBundle::Iwad || bundle->format() == DataBundle::Pwad) {
                lumpDirCrc32 = String::format("%08x", bundle->lumpDirectory()->crc32());
            }
        }
        else {
            format = tr("Doomsday 2 Package");
            useIconFile(file->path());
        }
        if (is<ArchiveEntryFile>(file->source())) {
            fileDesc = file->description();
        }
        title->setText(meta.gets(Package::VAR_TITLE));
        path->setText(String(_E(b) "%1" _E(.) "\n%2").arg(format).arg(fileDesc.upperFirstChar()));
        String metaMsg =
            String(_E(Ta) _E(l) "Version: " _E(.) _E(Tb) "%1\n" _E(Ta) _E(l) "Tags: " _E(.)
                       _E(Tb) "%2\n" _E(Ta) _E(l) "License: " _E(.) _E(Tb) "%3")
                .arg(meta.gets("version"))
                .arg(meta.gets("tags"))
                .arg(meta.gets("license"));
        if (meta.has("author")) {
            metaMsg +=
                String("\n" _E(Ta) _E(l) "Author: " _E(.) _E(Tb) "%1").arg(meta.gets("author"));
        }
        if (meta.has("contact")) {
            metaMsg +=
                String("\n" _E(Ta) _E(l) "Contact: " _E(.) _E(Tb) "%1").arg(meta.gets("contact"));
        }
        metaInfo->setText(metaMsg);
        String msg;
        if (compatibleGame.isEmpty()) {
            QRegularExpression reg(DataBundle::anyGameTagPattern());
            if (!reg.match(meta.gets("tags")).hasMatch()) {
                msg += "Not enough information to determine which game this package is for.";
            }
        }
        else {
            msg += String("This package is likely meant for " _E(b) "%1" _E(.) ".")
                       .arg(compatibleGame);
        }
        String moreMsg; {
            if (bundle && bundle->lumpDirectory() &&
                bundle->lumpDirectory()->mapType() != res::LumpDirectory::None) {
                int const mapCount = bundle->lumpDirectory()->findMaps().count();
                moreMsg +=
                    QString("Contains %1 map%2: ").arg(mapCount).arg(DENG2_PLURAL_S(mapCount)) +
                    String::join(bundle->lumpDirectory()->mapsInContiguousRangesAsText(), ", ");
                moreMsg += "\n";
                if (bundle->lumpDirectory()->has("DEHACKED")) {
                    moreMsg += "Contains a DeHackEd patch\n";
                }
            }
            if (lumpDirCrc32) {
                moreMsg += "WAD directory CRC32: " _E(m) + lumpDirCrc32 + _E(.) "\n";
            }
            moreMsg += "Package ID: " _E(>) + meta.gets(Package::VAR_ID) + _E(<) "\n";
        }
        msg += "\n\n" + moreMsg.strip();
        if (meta.has("notes")) {
            String notesText = meta.gets("notes");
            notesText.remove('\r'); // maybe old MS-DOS text
            notesText.replace("\t", "        ");
            msg += "\n\n" + notesText + _E(r);
        }
        if (!bundle) {
            if (meta.has("requires")) {
                msg += "\n\nRequires:";
                ArrayValue const &reqs = meta.geta("requires");
                for (Value const *val : reqs.elements()) {
                    msg += "\n - " _E(>) + val->asText() + _E(<);
                }
            }
            if (meta.has("dataFiles") && meta.geta("dataFiles").size() > 0) {
                msg += "\n\nData files:";
                ArrayValue const &files = meta.geta("dataFiles");
                for (Value const *val : files.elements()) {
                    msg += "\n - " _E(>) + val->asText() + _E(<);
                }
            }
        }
        description->setText(msg.trimmed());
        if (mode == EnableActions) {
            self().buttons()
                << new DialogButtonItem(Action | Id2,
                                        style().images().image("play"),
                                        tr("Play in..."),
                                        new SignalAction(thisPublic, SLOT(playInGame())))
                << new DialogButtonItem(Action | Id3,
                                        style().images().image("create"),
                                        tr("Add to..."),
                                        new SignalAction(thisPublic, SLOT(addToProfile())));
        }
        if (!nativePath.isEmpty()) {
            self().buttons() << new DialogButtonItem(
                Action, tr("Show File"), new SignalAction(thisPublic, SLOT(showFile())));
        }
        if (mode == EnableActions && Package::hasOptionalContent(*file)) {
            self().buttons() << new DialogButtonItem(
                Action | Id1,
                style().images().image("gear"),
                tr("Options"),
                new SignalAction(thisPublic, SLOT(configure())));
        }
        return true;
    }
    static String visibleFamily(String const &family) {
        if (family.isEmpty()) return tr("Other");
        return family.upperFirstChar();
    }
    /**
     * Opens a popup menu listing the available game profiles. The compatible game
     * is highlighted, as well as profiles that already have this package added.
     *
     * @param anchor        Popup menu anchor.
     * @param playableOnly  Only show profiles that can be started at the moment.
     */
    void openProfileMenu(RuleRectangle const &anchor, bool playableOnly) {
        if (profileMenu) return;
        profileMenu.reset(new PopupMenuWidget);
        profileMenu->setDeleteAfterDismissed(true);
        profileMenu->setAnchorAndOpeningDirection(anchor, ui::Left);
        QList<GameProfile *> profs = DoomsdayApp::gameProfiles().profilesSortedByFamily();
        auto & items = profileMenu->items();
        String lastFamily;
        for (GameProfile *prof : profs) {
            if (playableOnly && !prof->isPlayable()) continue;
            if (lastFamily != prof->game().family()) {
                if (!items.isEmpty()) {
                    items << new ui::Item(ui::Item::Separator);
                }
                items << new ui::Item(ui::Item::ShownAsLabel | ui::Item::Separator,
                                      visibleFamily(prof->game().family()));
                lastFamily = prof->game().family();
            }
            String label = prof->name();
            String color;
            if (prof->packages().contains(packageId)) {
                color = _E(C);
                label += " " _E(s) _E(b) _E(D) + tr("ADDED");
            }
            if (!compatibleGame.isEmpty() && prof->gameId() == compatibleGame) {
                label = _E(b) + label;
            }
            if (prof->isUserCreated()) {
                color = _E(F);
            }
            label = color + label;
            items << new ui::ActionItem(
                label, new CallbackAction([this, prof]() { profileSelectedFromMenu(*prof); }));
        }
        self().add(profileMenu);
        profileMenu->open();
    }
    void profileSelectedFromMenu(GameProfile & profile) {
        switch (menuMode) {
            case AddToProfile: profile.appendPackage(packageId); break;
            case PlayInProfile: {
                auto &prof = DoomsdayApp::app().adhocProfile();
                prof       = profile;
                prof.appendPackage(packageId);
                if (bundle && bundle->format() == DataBundle::Pwad) {
                    const auto mapIds = bundle->lumpDirectory()->findMapLumpNames();
                    if (mapIds.size() == 1) {
                        using namespace std;
                        ostringstream os;
                        os << "# Episode definition for Ad-hoc profile (autogenerated)\n"
                              "Episode {\n"
                              "  ID = \"adhoc_play_in\"\n"
                              "  Title = \""
                           << bundle->packageMetadata()
                                  .gets(Package::VAR_TITLE)
                                  .escaped()
                                  .toStdString()
                           << "\"\n"
                              "  Start Map = \""
                           << mapIds.front().toStdString()
                           << "\"\n"
                              "  Map {\n"
                              "    ID = \""
                           << mapIds.front().toStdString()
                           << "\"\n"
                              "    Warp Number = 1\n"
                              "  }\n"
                              "}\n";
                        File &f = App::homeFolder().replaceFile("AdhocPlayInEpisode_1.0.ded");
                        f << Block(os.str().c_str());
                        f.reinterpret();
                        App::homeFolder().populate();
                        prof.appendPackage("file.ded.adhocplayinepisode_1.0.0");
                    }
                }
                Loop::timer(0.1, [] {
                    GLWindow::main().glActivate();
                    BusyMode_FreezeGameForBusyMode();
                    DoomsdayApp::app().changeGame(DoomsdayApp::app().adhocProfile(),
                                                  DD_ActivateGameWorker);
                });
                self().accept();
                break;
            }
        }
    }
};
PackageInfoDialog::PackageInfoDialog(String const &packageId, Mode mode)
    : DialogWidget("packagepopup")
    , d(new Impl(this, mode)) {
    if (!d->setup(App::packageLoader().select(packageId))) {
    }
}
PackageInfoDialog::PackageInfoDialog(File const *packageFile, Mode mode)
    : DialogWidget("packagepopup")
    , d(new Impl(this, mode)) {
    if (!d->setup(packageFile)) {
    }
}
void PackageInfoDialog::prepare() {
    DialogWidget::prepare();
    bool const useWideLayout = d->description->text().contains(_E(m)) &&
                               d->description->text().size() > 200; // reasonably long?
    d->descriptionWidth->setSource(OperatorRule::minimum(
        useWideLayout ? rule("dialog.packageinfo.description.wide")
                      : rule("dialog.packageinfo.description.normal"),
        root().viewWidth() - margins().width() - d->metaInfo->rule().width()));
    d->minContentHeight->setSource(useWideLayout ? rule("dialog.packageinfo.content.minheight")
                                                 : ConstantRule::zero());
}
void PackageInfoDialog::playInGame() {
    d->menuMode = Impl::PlayInProfile;
    d->openProfileMenu(buttonWidget(Id2)->rule(), true);
}
void PackageInfoDialog::addToProfile() {
    d->menuMode = Impl::AddToProfile;
    d->openProfileMenu(buttonWidget(Id3)->rule(), false);
}
void PackageInfoDialog::configure() {
    if (d->configurePopup) return; // Let it close itself.
    PopupWidget *pop = PackageContentOptionsWidget::makePopup(
        d->packageId, rule("dialog.packages.left.width"), root().viewHeight());
    d->configurePopup.reset(pop);
    pop->setAnchorAndOpeningDirection(buttonWidget(Id1)->rule(), ui::Left);
    pop->closeButton().setActionFn([pop]() { pop->close(); });
    add(pop);
    pop->open();
}
void PackageInfoDialog::showFile() {
    if (!d->nativePath.isEmpty()) {
        QDesktopServices::openUrl(QUrl::fromLocalFile(
            d->nativePath.isDirectory() ? d->nativePath : d->nativePath.fileNamePath()));
    }
}
DENG_GUI_PIMPL(PackagesDialog)
, public ChildWidgetOrganizer::IWidgetFactory
, public PackagesWidget::IPackageStatus
, DENG2_OBSERVES(Widget, ChildAddition) {
    StringList selectedPackages;
    LabelWidget *nothingSelected;
    ui::ListData actions;
    HomeMenuWidget *menu;
    PackagesWidget *browser;
    LabelWidget *gameTitle;
    LabelWidget *gameDataFiles;
    const GameProfile *gameProfile = nullptr;
    res::LumpCatalog catalog;
    /**
     * Information about a selected package. If the package file is not found, only
     * the ID is known.
     */
    class SelectedPackageItem
        : public ui::Item
        , DENG2_OBSERVES(res::Bundles, Identify) {
    public:
        SelectedPackageItem(String const &packageId) {
            setData(packageId);
            updatePackageInfo();
            DoomsdayApp::bundles().audienceForIdentify() += this;
        }
        void updatePackageInfo() {
            if ((_file = App::packageLoader().select(packageId())) != nullptr) {
                _info = &_file->objectNamespace().subrecord(Package::VAR_PACKAGE);
            }
            else {
                _info = nullptr;
            }
        }
        String packageId() const {
            return data().toString();
        }
        Record const *info() const {
            return _info;
        }
        File const *packageFile() const {
            return _file;
        }
        void dataBundlesIdentified() override {
            Loop::mainCall([this]() {
                updatePackageInfo();
                notifyChange();
            });
        }
    private:
        File   const *_file = nullptr;
        Record const *_info = nullptr;
    };
    /**
     * Widget showing information about a selected package, with a button for dragging
     * the item up and down.
     */
    class SelectedPackageWidget : public HomeItemWidget {
    public:
        SelectedPackageWidget(SelectedPackageItem const &item,
                              PackagesDialog &owner)
            : _owner(owner)
            , _item(&item) {
            useColorTheme(Normal, Normal);
            _removeButton = new ButtonWidget;
            _removeButton->setStyleImage("close.ringless", "small");
            _removeButton->margins().setTopBottom(RuleBank::UNIT);
            _removeButton->setActionFn([this] () {
                PackagesDialog::Impl *d = _owner.d;
                d->removePackage(packageId());
                d->browser->updateItems();
            });
            addButton(_removeButton);
            setKeepButtonsVisible(true);
            icon().set(Background());
            icon().setImageFit(ui::FitToSize | ui::OriginalAspectRatio);
            icon().setStyleImage("package.icon", "default");
            icon().margins().set("dialog.gap");
            Rule const &height = style().fonts().font("default").height();
            icon().rule().setInput(Rule::Width, height + rule("dialog.gap")*2);
        }
        void updateContents() {
            if (_item->info()) {
                label().setText(_item->info()->gets("title"));
            }
            else {
                label().setText(Package::splitToHumanReadable(_item->packageId()) +
                                " " _E(D) DENG2_CHAR_MDASH " Missing");
            }
        }
        String packageId() const {
            return _item->packageId();
        }
        PopupWidget *makeInfoPopup() const {
            return new PackageInfoDialog(_item->packageFile(), PackageInfoDialog::EnableActions);
        }
    private:
        PackagesDialog &_owner;
        SelectedPackageItem const *_item;
        ButtonWidget *_removeButton;
    };
    Impl(Public *i) : Base(i) {
        self().leftArea().add(gameTitle = new LabelWidget);
        gameTitle->add(gameDataFiles = new LabelWidget);
        nothingSelected = new LabelWidget;
        nothingSelected->setText(tr("No Mods Selected"));
        style().as<ClientStyle>().emptyMenuLabelStylist().applyStyle(*nothingSelected);
        nothingSelected->rule()
                .setRect(self().leftArea().rule())
                .setInput(Rule::Top, gameTitle->rule().bottom());
        self().leftArea().add(nothingSelected);
        gameTitle->setSizePolicy(ui::Filled, ui::Expand);
        gameTitle->setImageFit(ui::FitToWidth | ui::OriginalAspectRatio | ui::CoverArea);
        gameTitle->margins().setZero();
        gameDataFiles->setFont("small");
        gameDataFiles->setSizePolicy(ui::Fixed, ui::Expand);
        gameDataFiles->set(Background(style().colors().colorf("background")));
        gameDataFiles->setTextLineAlignment(ui::AlignLeft);
        gameDataFiles->setAlignment(ui::AlignLeft);
        gameTitle->rule()
                .setInput(Rule::Left,  self().leftArea().contentRule().left())
                .setInput(Rule::Top,   self().leftArea().contentRule().top())
                .setInput(Rule::Width, rule("dialog.packages.left.width"));
        gameDataFiles->rule()
                .setRect(gameTitle->rule())
                .clearInput(Rule::Top);
        self().leftArea().add(menu = new HomeMenuWidget);
        menu->layout().setRowPadding(Const(0));
        menu->rule()
                .setInput(Rule::Left,  self().leftArea().contentRule().left())
                .setInput(Rule::Top,   gameTitle->rule().bottom())
                .setInput(Rule::Width, rule("dialog.packages.left.width"));
        menu->organizer().setWidgetFactory(*this);
        menu->audienceForChildAddition() += this;
        self().leftArea().enableIndicatorDraw(true);
        QObject::connect(menu, &HomeMenuWidget::itemClicked, [this] (int index) {
            if (index >= 0) {
                browser->scrollToPackage(menu->items().at(index)
                                         .as<SelectedPackageItem>().packageId());
            }
        });
        self().rightArea().add(browser = new PackagesWidget(PackagesWidget::PopulationDisabled,
                                                          self().name() + ".filter"));
        browser->setActionsAlwaysShown(true);
        browser->setRightClickToOpenContextMenu(true);
        browser->setPackageStatus(*this);
        actions << new ui::SubwidgetItem(tr("..."), ui::Up, [this] () -> PopupWidget * {
            String const id = browser->actionPackage();
            return new PackageInfoDialog(id, PackageInfoDialog::EnableActions);
        });
        actions << new ui::VariantActionItem("create",
                                             "close.ringless",
                                             String(),
                                             String(),
                                             new CallbackAction([this] () {
            String const packageId = browser->actionPackage();
            if (!selectedPackages.contains(packageId)) {
                selectedPackages.append(packageId);
                menu->items() << new SelectedPackageItem(packageId);
                updateNothingIndicator();
                updateGameTitle();
            }
            else {
                removePackage(packageId);
            }
            browser->actionItem()->notifyChange();
        }));
        browser->setActionItems(actions);
        browser->rule()
                .setInput(Rule::Left,  self().rightArea().contentRule().left())
                .setInput(Rule::Top,   self().rightArea().contentRule().top())
                .setInput(Rule::Width, rule("dialog.packages.right.width"));
        self().rightArea().enableIndicatorDraw(true);
        browser->setFilterEditorMinimumY(self().rightArea().rule().top());
    }
    void populate() {
        menu->items().clear();
        for (String packageId : selectedPackages) {
            menu->items() << new SelectedPackageItem(packageId);
        }
        updateNothingIndicator();
    }
    void updateNothingIndicator() {
        nothingSelected->setOpacity(menu->items().isEmpty()? .5f : 0.f, 0.4);
    }
    void updateGameTitle() {
        if (gameProfile && catalog.setPackages(gameProfile->allRequiredPackages() + selectedPackages)) {
            gameTitle->setImage(IdTech1Image::makeGameLogo(gameProfile->game(), catalog,
                                                           IdTech1Image::UnmodifiedAppearance |
                                                           IdTech1Image::AlwaysTryLoad));
            StringList dataFiles;
                for (String packageId : gameProfile->allRequiredPackages()) {
                    if (const File *file = PackageLoader::get().select(packageId)) {
                        if (Package::matchTags(*file, QStringLiteral("\\bgamedata\\b"))) {
                            dataFiles << file->source()->description(0);
                        }
                    }
                }
            if (!dataFiles.isEmpty()) {
                gameDataFiles->setText(_E(l) + String::format("Game data file%s: ", dataFiles.size() != 1? "s" : "") +
                                       _E(.) + String::join(dataFiles, _E(l) " and " _E(.)));
            }
            else {
                gameDataFiles->setText(_E(D) + tr("Locate data file in Data Files settings"));
            }
        }
    }
    GuiWidget *makeItemWidget(ui::Item const &item, GuiWidget const *) override {
        return new SelectedPackageWidget(item.as<SelectedPackageItem>(), self());
    }
    void updateItemWidget(GuiWidget &widget, ui::Item const &) override {
        widget.as<SelectedPackageWidget>().updateContents();
    }
    bool isPackageHighlighted(String const &packageId) const override {
        return selectedPackages.contains(packageId);
    }
    void removePackage(String const &packageId) {
        selectedPackages.removeOne(packageId);
        auto pos = menu->items().findData(packageId);
        DENG2_ASSERT(pos != ui::Data::InvalidPos);
        menu->items().remove(pos);
        updateNothingIndicator();
        updateGameTitle();
    }
    void widgetChildAdded(Widget &child) override {
        ui::DataPos pos = menu->findItem(child.as<GuiWidget>());
        Loop::get().timer(0.3, [this, pos] () {
            menu->setSelectedIndex(pos);
        });
    }
};
PackagesDialog::PackagesDialog(String const &titleText)
    : DialogWidget("packagesdialog", WithHeading)
    , d(new Impl(this)) {
    if (titleText.isEmpty()) {
        heading().setText(tr("Mods"));
    }
    else {
        heading().setText(titleText);
    }
    heading().setStyleImage("package.icon");
    buttons()
            << new DialogButtonItem(Default | Accept, tr("OK"))
            << new DialogButtonItem(Reject, tr("Cancel"))
            << new DialogButtonItem(Action, style().images().image("refresh"),
                                    new CallbackAction([]() { FS::get().refreshAsync(); }))
            << new DialogButtonItem(Action | Id1, style().images().image("gear"),
                                    "Data Files",
                                    new CallbackAction([this]() {
        auto *dfsDlg = new DataFileSettingsDialog;
        dfsDlg->setAnchorAndOpeningDirection(buttonWidget(Id1)->rule(), ui::Up);
        dfsDlg->setDeleteAfterDismissed(true);
        add(dfsDlg);
        dfsDlg->open();
    }));
    leftArea() .setContentSize(d->menu->rule().width(),
                               OperatorRule::maximum(d->menu->rule().height(),
                                                     d->nothingSelected->rule().height(),
                                                     rule("dialog.packages.left.minheight")) +
                               d->gameTitle->rule().height());
    rightArea().setContentSize(d->browser->rule());
    d->browser->progress().rule().setRect(rightArea().rule());
    setMaximumContentHeight(rule().width() * 0.9f);
    d->browser->setPopulationEnabled(true);
}
void PackagesDialog::setProfile(const GameProfile &profile) {
    d->gameProfile = &profile;
    d->updateGameTitle();
}
void PackagesDialog::setSelectedPackages(StringList packages) {
    d->selectedPackages = packages;
    d->browser->populate();
    d->updateGameTitle();
}
StringList PackagesDialog::selectedPackages() const {
    return d->selectedPackages;
}
void PackagesDialog::preparePanelForOpening() {
    DialogWidget::preparePanelForOpening();
    d->populate();
    root().setFocus(&d->browser->searchTermsEditor());
}
DENG_GUI_PIMPL(RendererSettingsDialog) {
    ProfilePickerWidget *appear;
    CVarSliderWidget *fov;
    VariableToggleWidget *enableExtWithPWADs;
    VariableToggleWidget *disableExtTextures;
    VariableToggleWidget *disableExtPatches;
    CVarToggleWidget *precacheModels;
    CVarToggleWidget *precacheSprites;
    GridPopupWidget *devPopup;
    bool texSettingsToggled = false;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().area();
        area.add(appear = new ProfilePickerWidget(ClientApp::renderSystem().appearanceSettings(),
                                                  tr("appearance"), "profile-picker"));
        appear->setOpeningDirection(ui::Down);
        area.add(fov = new CVarSliderWidget("rend-camera-fov"));
        fov->setPrecision(0);
        fov->setRange(Ranged(30, 160));
        area.add(enableExtWithPWADs = new VariableToggleWidget(tr("Use with PWADs"),   App::config("resource.highResWithPWAD")));
        area.add(disableExtTextures = new VariableToggleWidget(tr("Disable for Textures"), App::config("resource.noHighResTex")));
        area.add(disableExtPatches  = new VariableToggleWidget(tr("Disable for Patches"),  App::config("resource.noHighResPatches")));
        self().add(devPopup = new GridPopupWidget);
        CVarChoiceWidget *rendTex = new CVarChoiceWidget("rend-tex");
        rendTex->items()
                << new ChoiceItem(tr("Materials"),   1)
                << new ChoiceItem(tr("Plain white"), 0)
                << new ChoiceItem(tr("Plain gray"),  2);
        CVarChoiceWidget *wireframe = new CVarChoiceWidget("rend-dev-wireframe");
        wireframe->items()
                << new ChoiceItem(tr("Nothing"), 0)
                << new ChoiceItem(tr("Game world"), 1)
                << new ChoiceItem(tr("Game world and UI"), 2);
        precacheModels  = new CVarToggleWidget("rend-model-precache",       tr("3D Models"));
        precacheSprites = new CVarToggleWidget("rend-sprite-precache",      tr("Sprites"));
        devPopup->addSeparatorLabel(tr("Behavior"));
        *devPopup << LabelWidget::newWithText(tr("Precaching:")) << precacheModels
                  << Const(0) << precacheSprites;
        devPopup->addSeparatorLabel(tr("Diagnosis"));
        *devPopup << LabelWidget::newWithText(tr("Surface Texturing:"))
                  << rendTex
                  << LabelWidget::newWithText(tr("Draw as Wireframe:"))
                  << wireframe
                  << LabelWidget::newWithText(tr("Bounds:"))
                  << new CVarToggleWidget("rend-dev-mobj-bbox", tr("Mobj Bounding Boxes"))
                  << Const(0)
                  << new CVarToggleWidget("rend-dev-polyobj-bbox", tr("Polyobj Bounding Boxes"))
                  << LabelWidget::newWithText(tr("Identifiers:"))
                  << new CVarToggleWidget("rend-dev-thinker-ids", tr("Thinker IDs"))
                  << Const(0)
                  << new CVarToggleWidget("rend-dev-sector-show-indices", tr("Sector Indices"))
                  << Const(0)
                  << new CVarToggleWidget("rend-dev-vertex-show-indices", tr("Vertex Indices"))
                  << Const(0)
                  << new CVarToggleWidget("rend-dev-generator-show-indices", tr("Particle Generator Indices"));
        devPopup->commit();
    }
    void fetch() {
        foreach (GuiWidget *child, self().area().childWidgets() + devPopup->content().childWidgets()) {
            if (ICVarWidget *w = maybeAs<ICVarWidget>(child)) {
                w->updateFromCVar();
            }
        }
    }
    void apply() {
        if (texSettingsToggled) {
            GL_TexReset();
        }
    }
};
RendererSettingsDialog::RendererSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Renderer Settings"));
    heading().setStyleImage("renderer");
    LabelWidget *appearLabel = LabelWidget::newWithText(tr("Appearance:"), &area());
    appearLabel->setName("appearance-label"); // for lookup from tutorial
    LabelWidget *fovLabel = LabelWidget::newWithText(tr("Field of View:"), &area());
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(2, 0);
    layout.setColumnAlignment(0, ui::AlignRight);
    layout << *appearLabel;
    layout.append(*d->appear, d->appear->rule().width() + d->appear->button().rule().width());
    layout << *fovLabel << *d->fov; {
        auto *pd = new VariableSliderWidget(App::config("render.pixelDensity"), Ranged(0, 1), .05);
        pd->setPrecision(2);
        area().add(pd);
        layout << *LabelWidget::newWithText(tr("Pixel Density:"), &area()) << *pd;
    }
    LabelWidget::appendSeparatorWithText("Textures", &area(), &layout);
    layout << *LabelWidget::newWithText(tr("External Images:"), &area()) << *d->enableExtWithPWADs
           << Const(0) << *d->disableExtTextures
           << Const(0) << *d->disableExtPatches;
    area().setContentSize(layout);
    buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())))
            << new DialogButtonItem(DialogWidget::ActionPopup | Id1,
                                    style().images().image("gauge"));
    popupButtonWidget(Id1)->setPopup(*d->devPopup);
    connect(this, SIGNAL(closed()), d->devPopup, SLOT(close()));
    connect(d->appear, SIGNAL(profileEditorRequested()), this, SLOT(editProfile()));
    d->fetch();
    auto toggledFunc = [this] (ToggleWidget::ToggleState) {
        d->texSettingsToggled = true;
    };
    connect(d->enableExtWithPWADs, &ToggleWidget::stateChangedByUser, toggledFunc);
    connect(d->disableExtTextures, &ToggleWidget::stateChangedByUser, toggledFunc);
    connect(d->disableExtPatches,  &ToggleWidget::stateChangedByUser, toggledFunc);
}
void RendererSettingsDialog::resetToDefaults() {
    ClientApp::renderSystem().settings().resetToDefaults();
    d->fetch();
}
void RendererSettingsDialog::showDeveloperPopup() {
    d->devPopup->open();
}
void RendererSettingsDialog::editProfile() {
    RendererAppearanceEditor *editor = new RendererAppearanceEditor;
    editor->open();
    ClientWindow::main().taskBar().closeConfigMenu();
}
void RendererSettingsDialog::finish(int result) {
    DialogWidget::finish(result);
    d->apply();
}
static String const VAR_RESOURCE_BROWSER_REPOSITORY("resource.browserRepository");
static String const ALL_CATEGORIES(QObject::tr("All Categories"));
DENG_GUI_PIMPL(RepositoryBrowserDialog)
, DENG2_OBSERVES(filesys::RemoteFeedRelay, Status)
, public ChildWidgetOrganizer::IWidgetFactory
, public AsyncScope {
    using RFRelay = filesys::RemoteFeedRelay;
    ui::ListData categoryData;
    std::shared_ptr<ui::ListData> data;
    std::unique_ptr<ui::FilteredData> shownData;
    std::unique_ptr<AsyncScope> populating;
    Lockable linkBusy;
    ProgressWidget *refreshProgress;
    ChoiceWidget *repo;
    LineEditWidget *search;
    MenuWidget *category;
    LabelWidget *statusText;
    HomeMenuWidget *nameList;
    DocumentWidget *description;
    String connectedRepository;
    String mountPath;
    QSet<String> filterTerms;
    Impl(Public *i) : Base(i) {
        auto &area = self().area();
        area.enableScrolling(false);
        area.enableIndicatorDraw(false);
        area.enablePageKeys(false);
        self().add(refreshProgress = new ProgressWidget);
        refreshProgress->useMiniStyle();
        refreshProgress->setOpacity(0);
        refreshProgress->setColor("altaccent");
        refreshProgress->setTextColor("altaccent");
        refreshProgress->setText(tr("Loading..."));
        refreshProgress->setTextAlignment(ui::AlignLeft);
        refreshProgress->setSizePolicy(ui::Expand, ui::Expand);
        area.add(statusText  = new LabelWidget);
        area.add(repo        = new ChoiceWidget);
        area.add(search      = new LineEditWidget);
        area.add(category    = new MenuWidget);
        area.add(nameList    = new HomeMenuWidget);
        area.add(description = new DocumentWidget);
        statusText->setTextColor("altaccent");
        try {
            auto &repos = Config::get().getdt("resource.repositories");
            for (auto i = repos.elements().begin(); i != repos.elements().end(); ++i) {
                repo->items()
                        << new ChoiceItem(i->first.value->asText(), i->second->asText());
            }
            repo->setSelected(repo->items().findLabel(App::config().gets(VAR_RESOURCE_BROWSER_REPOSITORY, "")));
        }
        catch (Error const &er) {
            LOG_MSG("Remote repositories not listed in configuration; "
                    "set Config.resource.repositories: %s")
                    << er.asText();
        }
        category->setItems(categoryData);
        category->setGridSize(0, ui::Expand, 1, ui::Expand, GridLayout::RowFirst);
        category->organizer().setWidgetFactory(*this);
        nameList->setVirtualizationEnabled(true, style().fonts().font("default").height().valuei() +
                                           rule("unit").valuei() * 2);
        nameList->organizer().setWidgetFactory(*this);
        nameList->enableScrolling(true);
        nameList->enablePageKeys(true);
        nameList->enableIndicatorDraw(true);
        nameList->setGridSize(1, ui::Filled, 0, ui::Fixed);
        nameList->layout().setRowPadding(Const(0));
        nameList->setBehavior(ChildVisibilityClipping);
        QObject::connect(repo, &ChoiceWidget::selectionChangedByUser, [this] (uint) {
            updateSelectedRepository();
        });
        QObject::connect(search, &LineEditWidget::editorContentChanged, [this] () {
            updateFilter();
        });
        RFRelay::get().audienceForStatus() += this;
        updateSelectedRepository();
    }
    ~Impl() {
        if (populating) populating->waitForFinished();
        disconnect();
    }
    bool filterItem(ui::Item const &item) const {
        if (filterTerms.isEmpty()) return true;
        DotPath const path(item.label());
        for (String const &term : filterTerms) {
            bool matched = false;
            for (int i = 0; i < path.segmentCount(); ++i) {
                if (path.segment(i).toStringRef().contains(term)) {
                    matched = true;
                    break;
                }
            }
            if (!matched) return false;
        }
        return true;
    }
    void updateFilter() {
        const auto oldTerms = filterTerms;
        filterTerms.clear();
        foreach (String term, search->text().split(' ')) {
            if (auto cleaned = term.strip()) {
                filterTerms.insert(cleaned.toLower());
            }
        }
        if (oldTerms != filterTerms && shownData) {
            shownData->refilter();
            shownData->stableSort([] (ui::Item const &a, ui::Item const &b) {
                return a.label().compare(b.label()) < 0;
            });
            updateStatusText();
        }
    }
    GuiWidget *makeItemWidget(ui::Item const &item, GuiWidget const *parent) override {
        if (parent == category) {
            auto *toggle = new ToggleWidget(ToggleWidget::WithoutIndicator);
            QObject::connect(toggle, &ToggleWidget::stateChanged,
                             [toggle, &item] (ToggleWidget::ToggleState state) {
                toggle->setColorTheme(state == ToggleWidget::Active? Inverted : Normal);
            });
            if (item.label() == ALL_CATEGORIES) {
                toggle->setToggleState(ToggleWidget::Active);
            }
            return toggle;
        }
        return new HomeItemWidget(HomeItemWidget::NonAnimatedHeight |
                                  HomeItemWidget::WithoutIcon);
    }
    void updateItemWidget(GuiWidget &widget, ui::Item const &item) override {
        if (widget.parentGuiWidget() == category) {
            auto &catButton = widget.as<ButtonWidget>();
            catButton.setText(item.label());
            catButton.setSizePolicy(ui::Expand, ui::Expand);
            catButton.margins().set("dialog.gap");
        }
        else {
            auto &w = widget.as<HomeItemWidget>();
            w.useColorTheme(Normal, Inverted);
            w.label().margins().set("unit");
            w.label().setText(item.label().fileName('.'));
        }
    }
    void updateSelectedRepository() {
        if (repo->isValidSelection()) {
            auto const &selItem = repo->selectedItem();
            App::config().set(VAR_RESOURCE_BROWSER_REPOSITORY, selItem.label());
            connect(selItem.data().toString());
        }
    }
    void connect(String address) {
        refreshProgress->setOpacity(1, 0.5);
        repo->disable();
        *this += async([this] () {
            disconnect();
            return 0;
        },
        [this, address] (int) {
            QUrl const url(address);
            RFRelay::get().addRepository(address, "/remote" / url.host());
            connectedRepository = address;
        });
    }
    void remoteRepositoryStatusChanged(String const &repository, RFRelay::Status status) override {
        if (repository == connectedRepository) {
            repo->enable();
            refreshProgress->setOpacity(0, 0.5);
            if (status == RFRelay::Connected) {
                populateAsync();
            }
            else {
                disconnect();
            }
        }
    }
    void disconnect() {
        populating.reset();
        if (connectedRepository) {
            DENG2_GUARD(linkBusy);
            mountPath.clear();
            connectedRepository.clear();
            RFRelay::get().removeRepository(connectedRepository);
        }
    }
    filesys::Link &link() {
        auto *link = RFRelay::get().repository(connectedRepository);
        DENG2_ASSERT(link);
        return *link;
    }
    void populateAsync() {
        populating.reset(new AsyncScope);
        *populating += async([this] () {
            DENG2_GUARD(linkBusy);
            std::shared_ptr<ui::ListData> pkgs(new ui::ListData);
            link().forPackageIds([&pkgs] (String const &id) {
                pkgs->append(new ui::Item(ui::Item::DefaultSemantics, id));
                return LoopContinue;
            });
            return pkgs;
        },
        [this] (std::shared_ptr<ui::ListData> pkgs) {
            setData(pkgs);
        });
    }
    void setData(std::shared_ptr<ui::ListData> newData) {
        DENG2_ASSERT_IN_MAIN_THREAD();
        nameList->useDefaultItems();
        shownData.reset(new ui::FilteredData(*newData));
        shownData->setFilter([this] (ui::Item const &i) { return filterItem(i); });
        data = newData;
        shownData->sort();
        nameList->setItems(*shownData);
        categoryData.clear();
        categoryData.append(new ui::Item(ui::Item::ShownAsButton, ALL_CATEGORIES));
        StringList tags = link().categoryTags();
        qSort(tags);
        foreach (String category, tags) {
            categoryData.append(new ui::Item(ui::Item::ShownAsButton, category));
        }
        updateStatusText();
        repo->enable();
        refreshProgress->setOpacity(0, 0.5);
    }
    void updateStatusText() {
        statusText->setText(tr("showing %1 out of %2 mods").arg(shownData->size()).arg(data->size()));
    }
};
RepositoryBrowserDialog::RepositoryBrowserDialog()
    : DialogWidget("repository-browser", WithHeading)
    , d(new Impl(this)) {
    heading().setText(tr("Install Mods"));
    heading().setStyleImage("package.icon", heading().fontId());
    AutoRef<Rule> nameListWidth(new ConstantRule(2*175));
    AutoRef<Rule> descriptionWidth(new ConstantRule(2*525));
    AutoRef<Rule> listHeight(new ConstantRule(2*325));
    auto &acRule = area().contentRule();
    auto *searchLabel = LabelWidget::newWithText(tr("Search:"), &area());
    auto *repoLabel   = LabelWidget::newWithText(tr("Repository:"), &area()); {
        SequentialLayout layout(acRule.left(),
                                acRule.top(),
                                ui::Right);
        layout << *searchLabel << *d->search << *repoLabel << *d->repo;
        d->search->rule().setInput(Rule::Width,
                                   acRule.width() -
                                   searchLabel->rule().width() -
                                   repoLabel->rule().width() -
                                   d->repo->rule().width());
    }
    d->category->rule()
            .setInput(Rule::Left, acRule.left())
            .setInput(Rule::Top, d->search->rule().bottom());
    d->statusText->rule()
            .setInput(Rule::Right, acRule.right())
            .setInput(Rule::Top, d->category->rule().top())
            .setInput(Rule::Height, d->category->rule().height());
    d->nameList->rule()
            .setSize(nameListWidth, listHeight)
            .setInput(Rule::Left, acRule.left())
            .setInput(Rule::Top, d->category->rule().bottom());
    d->description->rule()
            .setSize(descriptionWidth, listHeight)
            .setInput(Rule::Left, d->nameList->rule().right())
            .setInput(Rule::Top, d->nameList->rule().top());
    area().setContentSize(nameListWidth + descriptionWidth,
                          d->search->rule().height() + d->category->rule().height() +
                          listHeight);
    d->refreshProgress->rule()
            .setInput(Rule::Right, rule().right() - area().margins().right())
            .setInput(Rule::Top,   rule().top() + area().margins().top());
    buttons() << new DialogButtonItem(Default | Accept, tr("Close"))
              << new DialogButtonItem(Action  | Id1,    tr("Download & Install"))
              << new DialogButtonItem(Action  | Id2,    tr("Try in..."));
    buttonWidget(Id1)->disable();
    buttonWidget(Id2)->disable();
    extraButtonsMenu().margins().setLeft(area().margins().left() + nameListWidth);
}
void RepositoryBrowserDialog::finish(int result) {
    DialogWidget::finish(result);
    d->disconnect();
}
static DialogWidget::RoleFlags const ID_SV_PACKAGES = DialogWidget::Id1;
static DialogWidget::RoleFlags const ID_JOIN        = DialogWidget::Id2;
static DialogWidget::RoleFlags const ID_PING        = DialogWidget::Id3;
DENG_GUI_PIMPL(ServerInfoDialog)
, DENG2_OBSERVES(ServerLink, MapOutline)
, DENG2_OBSERVES(ServerLink, PingResponse)
, public PackagesWidget::IPackageStatus {
    Address host;
    String domainName;
    GameProfile profile;
    shell::ServerInfo serverInfo;
    TimeSpan ping = -1;
    ServerLink link; // querying details from the server
    QTimer queryTimer; // allow the dialog to open nicely before starting network queries
    enum Query {
        QueryNone,
        QueryStatus,
        QueryPing,
        QueryMapOutline,
    };
    Query pendingQuery = QueryNone;
    LabelWidget *title;
    LabelWidget *subtitle;
    LabelWidget *description;
    PopupWidget *serverPopup;
    PackagesWidget *serverPackages;
    PackagesButtonWidget *localPackages;
    MapOutlineWidget *mapOutline;
    LabelWidget *gameState;
    ui::ListData serverPackageActions;
    Impl(Public *i, shell::ServerInfo const &sv)
        : Base(i)
        , serverInfo(sv)
        , link(ServerLink::ManualConnectionOnly) {
        link.audienceForMapOutline()   += this;
        link.audienceForPingResponse() += this;
        connect(&queryTimer, &QTimer::timeout, [this] () { beginPendingQuery(); });
        self().useInfoStyle();
        self().buttons()
                << new DialogButtonItem(Default | Accept, tr("Close"))
                << new DialogButtonItem(Action | ID_JOIN, tr("Join Game"), new CallbackAction([this] () {
                    self().accept();
                    emit self().joinGame();
                }))
                << new DialogButtonItem(Action | ID_PING,
                                        style().images().image("refresh"), tr("Ping"),
                                        new CallbackAction([this] () {
                    ping = -1;
                    updateContent();
                    startQuery(QueryPing);
                }))
                << new DialogButtonItem(ActionPopup | ID_SV_PACKAGES,
                                        style().images().image("package.icon"), tr("Server"));
        createWidgets();
        self().buttonWidget(ID_JOIN)->disable();
        self().buttonWidget(ID_PING)->disable();
    }
    bool isPackageHighlighted(String const &) const {
        return false;
    }
    void createWidgets() {
        auto &area = self().area();
        title = LabelWidget::newWithText("", &area);
        title->setFont("title");
        title->setSizePolicy(ui::Filled, ui::Expand);
        title->setTextColor("inverted.accent");
        title->setTextLineAlignment(ui::AlignLeft);
        title->margins().setBottom("");
        subtitle = LabelWidget::newWithText("", &area);
        subtitle->setSizePolicy(ui::Filled, ui::Expand);
        subtitle->setTextColor("inverted.altaccent");
        subtitle->setTextLineAlignment(ui::AlignLeft);
        subtitle->margins().setTop("unit");
        description = LabelWidget::newWithText("", &area);
        description->setSizePolicy(ui::Filled, ui::Expand);
        description->setTextColor("inverted.text");
        description->setTextLineAlignment(ui::AlignLeft);
        LabelWidget *bg = new LabelWidget;
        bg->set(Background(Vector4f(style().colors().colorf("inverted.altaccent"), .1f),
                           Background::GradientFrameWithRoundedFill,
                           Vector4f(), 8));
        area.add(bg);
        mapOutline = new MapOutlineWidget;
        area.add(mapOutline);
        mapOutline->rule().setInput(Rule::Width, rule("dialog.serverinfo.mapoutline.width"));
        mapOutline->margins().set(rule("gap") * 2).setBottom("gap");
        gameState = LabelWidget::newWithText("", &area);
        gameState->setSizePolicy(ui::Filled, ui::Expand);
        gameState->setTextColor("inverted.altaccent");
        gameState->margins().setBottom(mapOutline->margins().top());
        bg->rule().setRect(mapOutline->rule())
                  .setInput(Rule::Bottom, gameState->rule().bottom());
        serverPackageActions << new ui::SubwidgetItem(
            tr("..."), ui::Right, [this]() -> PopupWidget * {
                return new PackageInfoDialog(serverPackages->actionPackage(),
                                             PackageInfoDialog::EnableActions);
            });
        serverPopup = new PopupWidget;
        self().add(serverPopup);
        serverPackages = new PackagesWidget(PackagesWidget::PopulationDisabled);
        serverPackages->margins().set("gap");
        serverPackages->setHiddenTags({"hidden"}); // show everything (top level)
        serverPackages->setActionItems(serverPackageActions);
        serverPackages->setActionsAlwaysShown(true);
        serverPackages->setPackageStatus(*this);
        serverPackages->searchTermsEditor().setEmptyContentHint(tr("Filter Server Mods"));
        serverPackages->rule().setInput(Rule::Width, rule("dialog.serverinfo.popup.width"));
        serverPopup->setContent(serverPackages);
        auto *svBut = self().popupButtonWidget(ID_SV_PACKAGES);
        svBut->setPopup(*serverPopup);
        svBut->setText(tr("Server"));
        svBut->setTextAlignment(ui::AlignLeft);
        svBut->disable();
        localPackages = new PackagesButtonWidget;
        localPackages->setColorTheme(Inverted);
        localPackages->setLabelPrefix(tr("Local: "));
        localPackages->setNoneLabel(tr("Local Mods..."));
        localPackages->setGameProfile(profile);
        localPackages->disable();
        localPackages->rule().setLeftTop(svBut->rule().right(), svBut->rule().top());
        self().add(localPackages);
        QObject::connect(localPackages, &PackagesButtonWidget::packageSelectionChanged,
                         [this] (QStringList packages) {
            Game::setLocalMultiplayerPackages(profile.gameId(), toStringList(packages));
        });
        updateLayout();
    }
    void updateLayout() {
        auto &area = self().area();
        SequentialLayout layout(area.contentRule().left(),
                                area.contentRule().top(),
                                ui::Down);
        layout.setOverrideWidth(rule("dialog.serverinfo.description.width"));
        layout << *title << *subtitle << *description;
        AutoRef<Rule> height(OperatorRule::maximum
                             (layout.height(), rule("dialog.serverinfo.content.minheight")));
        mapOutline->rule()
                .setInput(Rule::Height, height - gameState->rule().height())
                .setLeftTop(title->rule().right(), title->rule().top());
        gameState->rule()
                .setInput(Rule::Width,  mapOutline->rule().width())
                .setInput(Rule::Left,   mapOutline->rule().left())
                .setInput(Rule::Bottom, area.contentRule().bottom());
        area.setContentSize(layout.width() + mapOutline->rule().width(), height);
    }
    void updateContent() {
        title->setText(serverInfo.name()); {
            StringList lines;
            if (!domainName.isEmpty()) {
                lines << _E(b) + domainName + _E(.) + String(" (%1)").arg(host.asText());
            }
            else {
                lines << _E(b) + host.asText() + _E(.);
            }
            if (!serverInfo.description().isEmpty()) {
                lines << "\n" _E(A) + serverInfo.description() + _E(.);
            }
            subtitle->setText(String::join(lines, "\n"));
        } {
            auto const players = serverInfo.players();
            String plrDesc;
            if (players.isEmpty()) {
                plrDesc = DENG2_CHAR_MDASH;
            }
            else {
                plrDesc = String("%1 " DENG2_CHAR_MDASH " %2")
                        .arg(players.count())
                        .arg(String::join(players, ", "));
            }
            String msg = String(_E(Ta)_E(l) "%1:" _E(.)_E(Tb) " %2\n"
                                _E(Ta)_E(l) "%3:" _E(.)_E(Tb) " %4\n"
                                _E(Ta)_E(l) "%5:" _E(.)_E(Tb) " %6\n"
                                _E(Ta)_E(l) "%7:" _E(.)_E(Tb) " %8")
                    .arg(tr("Rules"))  .arg(serverInfo.gameConfig())
                    .arg(tr("Players")).arg(plrDesc)
                    .arg(tr("Version")).arg(serverInfo.version().asHumanReadableText())
                    .arg(tr("Ping"))   .arg(ping < 0.0? String(DENG2_CHAR_MDASH)
                                                      : String("%1 ms").arg(ping.asMilliSeconds()));
            description->setText(msg);
        }
        String const gameId = serverInfo.gameId();
        String gameTitle = gameId;
        if (Games::get().contains(gameId)) {
            gameTitle = Games::get()[gameId].title();
        } {
            String msg = String(_E(b) "%1" _E(.)_E(s) "\n%2 " DENG2_CHAR_MDASH " %3")
                        .arg(serverInfo.map())
                        .arg(serverInfo.gameConfig().containsWord("coop")? tr("Co-op")
                                                                         : tr("Deathmatch"))
                        .arg(gameTitle);
            gameState->setText(msg);
        }
        self().buttonWidget(ID_JOIN)->enable(serverInfo.flags().testFlag(shell::ServerInfo::AllowJoin)); {
            localPackages->setDialogTitle(tr("Local Mods for %1 Multiplayer").arg(gameTitle));
            localPackages->setGameProfile(profile);
            localPackages->setPackages(Game::localMultiplayerPackages(gameId));
        }
        if (!serverInfo.packages().isEmpty()) {
            StringList available;
            StringList missing;
            foreach (String pkgId, serverInfo.packages()) {
                if (PackageLoader::get().select(pkgId)) {
                    available << pkgId;
                }
                else {
                    auto const id_ver = Package::split(pkgId);
                    pkgId = Package::splitToHumanReadable(pkgId);
                    Version localVersion;
                    if (id_ver.second.isValid()) {
                        if (auto *pkgFile = PackageLoader::get().select(id_ver.first)) {
                            localVersion = Package::versionForFile(*pkgFile);
                            missing << String("%1 " _E(s) "(you have: %2)" _E(.))
                                       .arg(pkgId).arg(localVersion.fullNumber());
                            continue;
                        }
                    }
                    missing << pkgId;
                }
            }
            if (missing.size() > 0) {
                description->setText(description->text() +
                                     _E(T`) "\n\n" _E(b) "Missing packages:" _E(.) "\n- " _E(>) +
                                     String::join(missing, _E(<) "\n- " _E(>)));
            }
            serverPackages->setPopulationEnabled(true);
            serverPackages->setManualPackageIds(available);
            self().buttonWidget(ID_SV_PACKAGES)->enable();
            self().buttonWidget(ID_SV_PACKAGES)->setText(tr("Server: %1").arg(serverInfo.packages().size()));
        }
    }
    void openServerPackagesPopup() {
        serverPopup->openOrClose();
    }
    void startQuery(Query query) {
        pendingQuery = query;
        queryTimer.stop();
        queryTimer.setInterval(500);
        queryTimer.setSingleShot(true);
        queryTimer.start();
    }
    void beginPendingQuery() {
        Query const handling = pendingQuery;
        pendingQuery = QueryNone;
        switch (handling) {
        case QueryStatus:
            if (!domainName.isEmpty()) {
                link.acquireServerProfileAsync(domainName, [this] (Address resolvedAddress,
                                                              GameProfile const *svProfile) {
                    host = resolvedAddress;
                    statusReceived(*svProfile);
                });
            }
            else {
                link.acquireServerProfileAsync(host, [this] (GameProfile const *svProfile) {
                    statusReceived(*svProfile);
                });
            }
            break;
        case QueryPing:
            link.ping(host);
            self().buttonWidget(ID_PING)->disable();
            break;
        case QueryMapOutline:
            link.requestMapOutline(host);
            break;
        case QueryNone:
            break;
        }
    }
    void statusReceived(GameProfile const &svProfile) {
        link.foundServerInfo(0, serverInfo);
        profile = svProfile;
        if (Game::isLocalPackagesEnabled()) {
            localPackages->enable();
        }
        updateContent();
        startQuery(QueryMapOutline);
    }
    void mapOutlineReceived(Address const &, shell::MapOutlinePacket const &packet) {
        mapOutline->setOutline(packet);
        startQuery(QueryPing);
    }
    void pingResponse(Address const &, TimeSpan pingTime) {
        ping = pingTime;
        updateContent();
        self().buttonWidget(ID_PING)->enable();
    }
};
ServerInfoDialog::ServerInfoDialog(shell::ServerInfo const &serverInfo)
    : d(new Impl(this, serverInfo)) {
    d->domainName = serverInfo.domainName();
    d->host       = serverInfo.address();
    d->updateContent();
    d->startQuery(Impl::QueryStatus);
}
DENG2_PIMPL(UISettingsDialog) {
    VariableChoiceWidget *uiScale;
    VariableToggleWidget *uiTranslucency;
    VariableToggleWidget *showAnnotations;
    VariableToggleWidget *showDoom;
    VariableToggleWidget *showHeretic;
    VariableToggleWidget *showHexen;
    VariableToggleWidget *showOther;
    VariableToggleWidget *showMultiplayer;
    Impl(Public *i) : Base(i) {
        auto &area = self().area();
        area.add(uiScale               = new VariableChoiceWidget(Config::get("ui.scaleFactor"), VariableChoiceWidget::Number));
        area.add(uiTranslucency        = new VariableToggleWidget("Background Translucency", Config::get("ui.translucency")));
        area.add(showAnnotations       = new VariableToggleWidget("Menu Annotations",  Config::get("ui.showAnnotations")));
        area.add(showDoom              = new VariableToggleWidget("Doom",              Config::get("home.columns.doom")));
        area.add(showHeretic           = new VariableToggleWidget("Heretic",           Config::get("home.columns.heretic")));
        area.add(showHexen             = new VariableToggleWidget("Hexen",             Config::get("home.columns.hexen")));
        area.add(showOther             = new VariableToggleWidget("Other Games",       Config::get("home.columns.otherGames")));
        area.add(showMultiplayer       = new VariableToggleWidget("Multiplayer",       Config::get("home.columns.multiplayer")));
        uiScale->items() << new ChoiceItem("Double (200%)", 2.0) << new ChoiceItem("175%", 1.75)
                         << new ChoiceItem("150%", 1.5) << new ChoiceItem("125%", 1.25)
                         << new ChoiceItem("110%", 1.1) << new ChoiceItem("Normal (100%)", 1.0)
                         << new ChoiceItem("90%", 0.9) << new ChoiceItem("75%", 0.75)
                         << new ChoiceItem("Half (50%)", 0.5);
        uiScale->updateFromVariable();
    }
    void resetToDefaults() {
        ClientApp::uiSettings().resetToDefaults();
    }
};
UISettingsDialog::UISettingsDialog(String const &name)
    : DialogWidget(name, WithHeading)
    , d(new Impl(this)) {
    heading().setText(tr("UI Settings"));
    heading().setImage(style().images().image("home.icon"));
    d->showAnnotations->margins().setBottom(RuleBank::UNIT);
    auto *annots = LabelWidget::newWithText(tr("Annotations briefly describe menu functions."), &area());
    annots->margins().setTop("");
    annots->setFont("separator.annotation");
    annots->setTextColor("altaccent");
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(2, 0);
    layout.setColumnAlignment(0, ui::AlignRight);
    layout << *LabelWidget::newWithText(tr("Scale:"), &area()) << *d->uiScale
           << Const(0) << *d->uiTranslucency
           << Const(0) << *d->showAnnotations
           << Const(0) << *annots;
    auto *library = LabelWidget::appendSeparatorWithText("Game Library", &area(), &layout);
    auto *showLabel = LabelWidget::newWithText(tr("Enabled Tabs:"), &area());
    showLabel->rule().setLeftTop(library->rule().left(), library->rule().bottom());
    GridLayout showLayout(showLabel->rule().right(), showLabel->rule().top(),
                          GridLayout::RowFirst);
    showLayout.setGridSize(2, 3);
    showLayout << *d->showDoom
               << *d->showHeretic
               << *d->showHexen
               << *d->showOther
               << *d->showMultiplayer;
    area().setContentSize(OperatorRule::maximum(layout.width(),
                                                showLabel->rule().width() + showLayout.width()),
                          layout.height() + showLayout.height());
    buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new CallbackAction([this] () { d->resetToDefaults(); }));
}
#if !defined (MACOSX)
#  define USE_REFRESH_RATE_CHOICE
#  define USE_COLOR_DEPTH_CHOICE
#endif
DENG2_PIMPL(VideoSettingsDialog)
#if !defined (DENG_MOBILE)
, DENG2_OBSERVES(PersistentGLWindow, AttributeChange)
#endif
{
    ClientWindow &win;
    VariableToggleWidget *showFps;
    ToggleWidget *fullscreen;
    ToggleWidget *maximized;
    ToggleWidget *centered;
    VariableToggleWidget *fsaa;
    VariableToggleWidget *vsync;
    ToggleWidget *fpsLimiter = nullptr;
    SliderWidget *fpsMax = nullptr;
    ChoiceWidget *modes = nullptr;
    ButtonWidget *windowButton = nullptr;
#ifdef USE_REFRESH_RATE_CHOICE
    ChoiceWidget *refreshRates;
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
    ChoiceWidget *depths;
#endif
    ListData stretchChoices;
    CVarChoiceWidget *finaleAspect = nullptr;
    CVarChoiceWidget *hudAspect    = nullptr;
    CVarChoiceWidget *inludeAspect = nullptr;
    CVarChoiceWidget *menuAspect   = nullptr;
    Impl(Public *i)
        : Base(i)
        , win(ClientWindow::main()) {
        ScrollAreaWidget &area = self().area();
        area.add(showFps      = new VariableToggleWidget(App::config("window.main.showFps")));
        area.add(fullscreen   = new ToggleWidget);
        area.add(maximized    = new ToggleWidget);
        area.add(centered     = new ToggleWidget);
        area.add(fsaa         = new VariableToggleWidget(App::config("window.main.fsaa")));
        area.add(vsync        = new VariableToggleWidget(App::config("window.main.vsync")));
        if (gotDisplayMode()) {
            area.add(modes        = new ChoiceWidget("modes"));
            area.add(windowButton = new ButtonWidget);
        }
#ifdef USE_REFRESH_RATE_CHOICE
        area.add(refreshRates = new ChoiceWidget("refresh-rate"));
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
        area.add(depths       = new ChoiceWidget("depths"));
#endif
#if !defined (DENG_MOBILE)
        win.audienceForAttributeChange() += this;
#endif
        area.add(fpsLimiter = new ToggleWidget);
        fpsLimiter->margins().setTop("dialog.separator");
        fpsLimiter->setText(tr("FPS Limiter"));
        QObject::connect(fpsLimiter, &ToggleWidget::stateChangedByUser, [this] (ToggleWidget::ToggleState st) {
            fpsMax->enable(st == ToggleWidget::Active);
            applyFpsMax();
        });
        area.add(fpsMax = new SliderWidget);
        fpsMax->margins().setTop("dialog.separator");
        fpsMax->setRange(Ranged(35, 60));
        fpsMax->setPrecision(0);
        fpsMax->rule().setInput(Rule::Width, centered->rule().width());
        QObject::connect(fpsMax, &SliderWidget::valueChangedByUser, [this] (double) { applyFpsMax(); });
        if (App_GameLoaded()) {
            stretchChoices
                << new ChoiceItem(tr("Smart"),        SCALEMODE_SMART_STRETCH)
                << new ChoiceItem(tr("Original 1:1"), SCALEMODE_NO_STRETCH)
                << new ChoiceItem(tr("Stretched"),    SCALEMODE_STRETCH);
            area.add(finaleAspect = new CVarChoiceWidget("rend-finale-stretch"));
            area.add(hudAspect    = new CVarChoiceWidget("rend-hud-stretch"));
            area.add(inludeAspect = new CVarChoiceWidget("inlude-stretch"));
            area.add(menuAspect   = new CVarChoiceWidget("menu-stretch"));
            finaleAspect->setItems(stretchChoices);
            hudAspect   ->setItems(stretchChoices);
            inludeAspect->setItems(stretchChoices);
            menuAspect  ->setItems(stretchChoices);
        }
    }
    ~Impl() {
        if (finaleAspect) {
            finaleAspect->useDefaultItems();
            hudAspect->useDefaultItems();
            inludeAspect->useDefaultItems();
            menuAspect->useDefaultItems();
        }
    }
    /**
     * Updates the widgets with the actual current state.
     */
    void fetch() {
#if !defined (DENG_MOBILE)
        fullscreen->setActive(win.isFullScreen());
        maximized->setActive(win.isMaximized());
        centered->setActive(win.isCentered());
        if (windowButton) {
            windowButton->enable(!win.isFullScreen() && !win.isMaximized());
        }
        if (gotDisplayMode()) {
            GLWindow::Size current = win.fullscreenSize();
            ui::Data::Pos closest = ui::Data::InvalidPos;
            int delta = 0;
            for (ui::Data::Pos i = 0; i < modes->items().size(); ++i) {
                QPoint const res = modes->items().at(i).data().toPoint();
                int dx = res.x() - current.x;
                int dy = res.y() - current.y;
                int d = dx*dx + dy*dy;
                if (closest == ui::Data::InvalidPos || d < delta) {
                    closest = i;
                    delta = d;
                }
            }
            modes->setSelected(closest);
        }
#ifdef USE_REFRESH_RATE_CHOICE
        refreshRates->setSelected(refreshRates->items().findData(int(win.refreshRate() * 10)));
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
        depths->setSelected(depths->items().findData(win.colorDepthBits()));
#endif
#endif // !DENG_MOBILE
        if (fpsMax) {
            const int max = Config::get().geti("window.main.maxFps", 0);
            fpsLimiter->setActive(max != 0);
            fpsMax->enable(max != 0);
            fpsMax->setValue(!max ? 35 : max);
        }
        foreach (GuiWidget *child, self().area().childWidgets()) {
            if (ICVarWidget *cw = maybeAs<ICVarWidget>(child))
                cw->updateFromCVar();
        }
    }
#if !defined (DENG_MOBILE)
    void windowAttributesChanged(PersistentGLWindow &) {
        fetch();
    }
#endif
    void applyFpsMax() {
        if (fpsMax) {
            Config::get().set("window.main.maxFps",
                              fpsLimiter->isActive() ? int(fpsMax->value()) : 0);
        }
    }
    bool gotDisplayMode() const {
        return DisplayMode_Count() > 0;
    }
};
VideoSettingsDialog::VideoSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("Video Settings"));
    heading().setImage(style().images().image("display"));
    d->fullscreen->setText(tr("Fullscreen"));
    d->fullscreen->setAction(new CommandAction("togglefullscreen"));
    d->maximized->setText(tr("Maximized"));
    d->maximized->setAction(new CommandAction("togglemaximized"));
    d->centered->setText(tr("Center Window"));
    d->centered->setAction(new CommandAction("togglecentered"));
    d->showFps->setText(tr("Show FPS"));
    d->fsaa->setText(tr("Antialias"));
    d->vsync->setText(tr("VSync"));
#ifdef USE_REFRESH_RATE_CHOICE
    LabelWidget *refreshLabel = nullptr;
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
    LabelWidget *colorLabel = nullptr;
#endif
    if (d->gotDisplayMode()) {
        d->modes->setOpeningDirection(ui::Up);
        if (DisplayMode_Count() > 10) {
            d->modes->popup().menu().setGridSize(2, ui::Expand, 0, ui::Expand);
        }
        for (int i = 0; i < DisplayMode_Count(); ++i) {
            DisplayMode const *m = DisplayMode_ByIndex(i);
            QPoint const res(m->width, m->height);
            if (d->modes->items().findData(res) != ui::Data::InvalidPos) {
                continue;
            }
            String desc = String("%1 x %2 (%3:%4)")
                    .arg(m->width).arg(m->height)
                    .arg(m->ratioX).arg(m->ratioY);
            d->modes->items() << new ChoiceItem(desc, res);
        }
#ifdef USE_REFRESH_RATE_CHOICE
        {
            refreshLabel = LabelWidget::newWithText(tr("Monitor Refresh:"), &area());
            QSet<int> rates;
            rates.insert(0);
            for (int i = 0; i < DisplayMode_Count(); ++i) {
                rates.insert(int(DisplayMode_ByIndex(i)->refreshRate * 10));
            }
            foreach (int rate, rates) {
                if (rate == 0) {
                    d->refreshRates->items() << new ChoiceItem(tr("Default"), 0);
                }
                else {
                    d->refreshRates->items()
                        << new ChoiceItem(tr("%1 Hz").arg(float(rate) / 10.f, 0, 'f', 1), rate);
                }
            }
            d->refreshRates->items().sort([] (ui::Item const &a, ui::Item const &b) {
                int const i = a.data().toInt();
                int const j = b.data().toInt();
                if (!i) return true;
                if (!j) return false;
                return i < j;
            });
        }
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
        {
            colorLabel = LabelWidget::newWithText(tr("Colors:"), &area());
            d->depths->items()
                << new ChoiceItem(tr("32-bit"), 32)
                << new ChoiceItem(tr("24-bit"), 24)
                << new ChoiceItem(tr("16-bit"), 16);
        }
#endif
    }
    buttons()
            << new DialogButtonItem(DialogWidget::Accept | DialogWidget::Default, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())));
    if (d->windowButton) {
        d->windowButton->setImage(style().images().image("window.icon"));
        d->windowButton->setOverrideImageSize(style().fonts().font("default").height());
        d->windowButton->setAction(new SignalAction(this, SLOT(showWindowMenu())));
    }
    Rule const &gap = rule("dialog.gap");
    GridLayout layout(area().contentRule().left(),
                      area().contentRule().top(), GridLayout::RowFirst);
    layout.setGridSize(2, d->fpsLimiter? 4 : 3);
    layout.setColumnPadding(rule(RuleBank::UNIT));
    layout << *d->fsaa
           << *d->vsync
           << *d->showFps;
    if (d->fpsLimiter) layout << *d->fpsLimiter;
    layout << *d->fullscreen
           << *d->maximized
           << *d->centered;
    if (d->fpsMax) layout << *d->fpsMax;
    GridLayout modeLayout(d->vsync->rule().left(),
                          (d->fpsLimiter? d->fpsLimiter : d->showFps)->rule().bottom() + gap);
    modeLayout.setGridSize(2, 0);
    modeLayout.setColumnAlignment(0, ui::AlignRight);
    if (d->gotDisplayMode()) {
        modeLayout << *LabelWidget::newWithText(tr("Resolution:"), &area());
        modeLayout.append(*d->modes, d->modes->rule().width() + d->windowButton->rule().width());
        d->windowButton->rule()
                .setInput(Rule::Top,  d->modes->rule().top())
                .setInput(Rule::Left, d->modes->rule().right());
#ifdef USE_REFRESH_RATE_CHOICE
        modeLayout << *refreshLabel << *d->refreshRates;
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
        modeLayout << *colorLabel << *d->depths;
#endif
    } {
        auto *adjustButton = new ButtonWidget;
        adjustButton->setText(tr("Color Adjustments..."));
        adjustButton->setAction(new SignalAction(this, SLOT(showColorAdjustments())));
        area().add(adjustButton);
        modeLayout << Const(0) << *adjustButton;
    }
    if (d->inludeAspect) {
        LabelWidget::appendSeparatorWithText("Aspect Ratios", &area(), &modeLayout);
        modeLayout
                << *LabelWidget::newWithText(tr("Player Weapons:"), &area()) << *d->hudAspect
                << *LabelWidget::newWithText(tr("Intermissions:"), &area()) << *d->inludeAspect
                << *LabelWidget::newWithText(tr("Finales:"), &area()) << *d->finaleAspect
                << *LabelWidget::newWithText(tr("Menus:"), &area()) << *d->menuAspect;
    }
    area().setContentSize(OperatorRule::maximum(layout.width(), modeLayout.width()),
                          layout.height() + gap + modeLayout.height());
    d->fetch();
    if (d->gotDisplayMode()) {
        connect(d->modes, SIGNAL(selectionChangedByUser(uint)), this, SLOT(changeMode(uint)));
    }
#ifdef USE_REFRESH_RATE_CHOICE
    connect(d->refreshRates, SIGNAL(selectionChangedByUser(uint)), this, SLOT(changeRefreshRate(uint)));
#endif
#ifdef USE_COLOR_DEPTH_CHOICE
    connect(d->depths, SIGNAL(selectionChangedByUser(uint)), this, SLOT(changeColorDepth(uint)));
#endif
}
void VideoSettingsDialog::resetToDefaults() {
    ClientApp::windowSystem().settings().resetToDefaults();
    d->fetch();
}
#if !defined (DENG_MOBILE)
void VideoSettingsDialog::changeMode(uint selected) {
    DENG2_ASSERT(d->modes);
    QPoint const res = d->modes->items().at(selected).data().toPoint();
    int const attribs[] = {
        ClientWindow::FullscreenWidth,  int(res.x()),
        ClientWindow::FullscreenHeight, int(res.y()),
        ClientWindow::End
    };
    d->win.changeAttributes(attribs);
}
void VideoSettingsDialog::changeColorDepth(uint selected) {
#ifdef USE_COLOR_DEPTH_CHOICE
    Con_Executef(CMDS_DDAY, true, "setcolordepth %i",
                 d->depths->items().at(selected).data().toInt());
#else
    DENG2_UNUSED(selected);
#endif
}
void VideoSettingsDialog::changeRefreshRate(uint selected) {
#ifdef USE_REFRESH_RATE_CHOICE
    float const rate = float(d->refreshRates->items().at(selected).data().toInt()) / 10.f;
    int const attribs[] = {
        ClientWindow::RefreshRate, int(rate * 1000), // milli-Hz
        ClientWindow::End
    };
    d->win.changeAttributes(attribs);
#else
    DENG2_UNUSED(selected);
#endif
}
void VideoSettingsDialog::showColorAdjustments() {
    d->win.showColorAdjustments();
    d->win.taskBar().closeConfigMenu();
}
void VideoSettingsDialog::showWindowMenu() {
    DENG2_ASSERT(d->windowButton);
    PopupMenuWidget *menu = new PopupMenuWidget;
    menu->setDeleteAfterDismissed(true);
    add(menu);
    menu->setAnchorAndOpeningDirection(d->windowButton->rule(), ui::Up);
    menu->items()
            << new ActionItem(tr("Apply to Window"),
                              new SignalAction(this, SLOT(applyModeToWindow())));
    menu->open();
}
void VideoSettingsDialog::applyModeToWindow() {
    QPoint const res = d->modes->selectedItem().data().toPoint();
    int attribs[] = {
        ClientWindow::Width,  res.x(),
        ClientWindow::Height, res.y(),
        ClientWindow::End
    };
    d->win.changeAttributes(attribs);
}
#endif
DENG_GUI_PIMPL(VRSettingsDialog) {
    CVarChoiceWidget *mode;
    CVarToggleWidget *swapEyes;
    CVarSliderWidget *dominantEye;
    CVarSliderWidget *humanHeight;
    CVarSliderWidget *ipd;
    VariableSliderWidget *riftDensity;
    CVarSliderWidget *riftSamples;
    ButtonWidget *riftReset;
    ButtonWidget *riftSetup;
    ButtonWidget *desktopSetup;
    Impl(Public *i)
        : Base(i)
        , riftReset(0)
        , riftSetup(0) {
        ScrollAreaWidget &area = self().area();
        area.add(mode = new CVarChoiceWidget("rend-vr-mode"));
        mode->items()
                << new ChoiceItem(tr("No stereo"),                VRConfig::Mono)
                << new ChoiceItem(tr("Anaglyph (green/magenta)"), VRConfig::GreenMagenta)
                << new ChoiceItem(tr("Anaglyph (red/cyan)"),      VRConfig::RedCyan)
                << new ChoiceItem(tr("Left eye only"),            VRConfig::LeftOnly)
                << new ChoiceItem(tr("Right eye only"),           VRConfig::RightOnly)
                << new ChoiceItem(tr("Top/bottom"),               VRConfig::TopBottom)
                << new ChoiceItem(tr("Side-by-side"),             VRConfig::SideBySide)
                << new ChoiceItem(tr("Parallel"),                 VRConfig::Parallel)
                << new ChoiceItem(tr("Cross-eye"),                VRConfig::CrossEye)
                << new ChoiceItem(tr("Hardware stereo"),          VRConfig::QuadBuffered)
                << new ChoiceItem(tr("Row interleaved"),          VRConfig::RowInterleaved)
                ;
        if (vrCfg().oculusRift().isEnabled()) {
            mode->items() << new ChoiceItem(tr("Oculus Rift"), VRConfig::OculusRift);
        }
        area.add(swapEyes    = new CVarToggleWidget("rend-vr-swap-eyes", tr("Swap Eyes")));
        area.add(dominantEye = new CVarSliderWidget("rend-vr-dominant-eye"));
        area.add(humanHeight = new CVarSliderWidget("rend-vr-player-height"));
        area.add(riftSamples = new CVarSliderWidget("rend-vr-rift-samples"));
        area.add(ipd = new CVarSliderWidget("rend-vr-ipd"));
        ipd->setDisplayFactor(1000);
        ipd->setPrecision(1);
        if (vrCfg().oculusRift().isReady()) {
            area.add(riftDensity = new VariableSliderWidget(App::config("vr.oculusRift.pixelDensity"),
                     Ranged(0.5, 1.0), .01));
            riftDensity->setPrecision(2);
            area.add(riftReset = new ButtonWidget);
            riftReset->setText(tr("Recenter Tracking"));
            riftReset->setAction(new CommandAction("resetriftpose"));
            area.add(riftSetup = new ButtonWidget);
            riftSetup->setText(tr("Apply Rift Settings"));
            riftSetup->setAction(new SignalAction(thisPublic, SLOT(autoConfigForOculusRift())));
        }
        area.add(desktopSetup = new ButtonWidget);
        desktopSetup->setText(tr("Apply Desktop Settings"));
        desktopSetup->setAction(new SignalAction(thisPublic, SLOT(autoConfigForDesktop())));
    }
    void fetch() {
        foreach (GuiWidget *child, self().area().childWidgets()) {
            if (ICVarWidget *w = maybeAs<ICVarWidget>(child)) {
                w->updateFromCVar();
            }
        }
    }
};
VRSettingsDialog::VRSettingsDialog(String const &name)
    : DialogWidget(name, WithHeading), d(new Impl(this)) {
    heading().setText(tr("3D & VR Settings"));
    heading().setImage(style().images().image("vr"));
    LabelWidget *modeLabel     = LabelWidget::newWithText(tr("Stereo Mode:"), &area());
    LabelWidget *heightLabel   = LabelWidget::newWithText(tr("Height (m):"), &area());
    LabelWidget *ipdLabel      = LabelWidget::newWithText(tr("IPD (mm):"), &area());
    LabelWidget *dominantLabel = LabelWidget::newWithText(tr("Dominant Eye:"), &area());
    GridLayout layout(area().contentRule().left(), area().contentRule().top());
    layout.setGridSize(2, 0);
    layout.setColumnAlignment(0, ui::AlignRight);
    layout << *modeLabel     << *d->mode
           << *heightLabel   << *d->humanHeight
           << *ipdLabel      << *d->ipd
           << *dominantLabel << *d->dominantEye
           << Const(0)       << *d->swapEyes;
#ifdef DENG_HAVE_OCULUS_API
    LabelWidget *ovrLabel    = LabelWidget::newWithText(_E(D) + tr("Oculus Rift"), &area());
    LabelWidget *sampleLabel = LabelWidget::newWithText(tr("Multisampling:"), &area());
    ovrLabel->setFont("separator.label");
    ovrLabel->margins().setTop("gap");
    sampleLabel->setTextLineAlignment(ui::AlignRight);
    layout.setCellAlignment(Vector2i(0, 5), ui::AlignLeft);
    layout.append(*ovrLabel, 2);
    LabelWidget *utilLabel = LabelWidget::newWithText(tr("Utilities:"), &area());
    if (vrCfg().oculusRift().isReady()) {
        layout << *sampleLabel << *d->riftSamples
               << *LabelWidget::newWithText(tr("Pixel Density:"), &area())
               << *d->riftDensity;
        layout << *utilLabel << *d->riftReset
               << Const(0)   << *d->riftSetup
               << Const(0)   << *d->desktopSetup;
    }
    else {
        layout << *utilLabel << *d->desktopSetup;
    }
#endif
    area().setContentSize(layout);
    buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Accept, tr("Close"))
            << new DialogButtonItem(DialogWidget::Action, tr("Reset to Defaults"),
                                    new SignalAction(this, SLOT(resetToDefaults())));
    d->fetch();
}
void VRSettingsDialog::resetToDefaults() {
    Con_SetInteger("rend-vr-mode",          VRConfig::Mono);
    Con_SetInteger("rend-vr-swap-eyes",     0);
    Con_SetFloat  ("rend-vr-dominant-eye",  0);
    Con_SetFloat  ("rend-vr-player-height", 1.75f);
    Con_SetFloat  ("rend-vr-ipd",           0.064f);
    Con_SetFloat  ("rend-vr-rift-latency",  0.030f);
    Con_SetInteger("rend-vr-rift-samples",  2);
    d->fetch();
}
void VRSettingsDialog::autoConfigForOculusRift() {
    Con_Execute(CMDS_DDAY, "bindcontrol lookpitch head-pitch", false, false);
    Con_Execute(CMDS_DDAY, "bindcontrol yawbody head-yaw", false, false);
    Con_SetInteger("rend-vr-mode", VRConfig::OculusRift);
    App::config().set("window.main.fsaa", false);
    Con_SetFloat  ("vid-gamma", 1.176f);
    Con_SetFloat  ("vid-contrast", 1.186f);
    Con_SetFloat  ("vid-bright", .034f);
    Con_SetFloat  ("view-bob-height", .2f);
    Con_SetFloat  ("msg-scale", 1);
    Con_SetFloat  ("hud-scale", 1);
    d->fetch();
    ClientApp::vr().oculusRift().moveWindowToScreen(OculusRift::HMDScreen);
}
void VRSettingsDialog::autoConfigForDesktop() {
    Con_SetInteger("rend-vr-mode", VRConfig::Mono);
    Con_SetFloat  ("vid-gamma", 1);
    Con_SetFloat  ("vid-contrast", 1);
    Con_SetFloat  ("vid-bright", 0);
    Con_SetFloat  ("view-bob-height", 1);
    Con_SetFloat  ("msg-scale", .8f);
    Con_SetFloat  ("hud-scale", .6f);
    d->fetch();
    ClientApp::vr().oculusRift().moveWindowToScreen(OculusRift::DefaultScreen);
}
