﻿/** @file columnwidget.cpp  Home column.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/columnwidget.h"
#include <de/App>
#include <de/GLProgram>
#include <de/LabelWidget>
#include <de/Range>
#include <de/StyleProceduralImage>
#include <de/math.h>
#include <QColor>
/** @file gamecolumnwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/gamecolumnwidget.h"
#include "ui/home/headerwidget.h"
#include "ui/home/gamepanelbuttonwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/dialogs/createprofiledialog.h"
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <doomsday/GameProfiles>
#include <de/CallbackAction>
#include <de/ChildWidgetOrganizer>
#include <de/Config>
#include <de/DirectoryListDialog>
#include <de/FileSystem>
#include <de/GridPopupWidget>
#include <de/Loop>
#include <de/MenuWidget>
#include <de/PersistentState>
#include <de/PopupMenuWidget>
#include <de/SignalAction>
#include <de/VariableChoiceWidget>
#include <de/VariableToggleWidget>
#include <QDesktopServices>
/** @file gamepanelbuttonwidget.cpp  Panel button for games.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/gamepanelbuttonwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/home/savelistwidget.h"
#include "ui/home/gamecolumnwidget.h"
#include "ui/savelistdata.h"
#include "ui/dialogs/packagesdialog.h"
#include "ui/widgets/packagesbuttonwidget.h"
#include "resource/idtech1image.h"
#include "dd_main.h"
#include <doomsday/console/exec.h>
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <doomsday/LumpCatalog>
#include <doomsday/LumpDirectory>
#include <doomsday/res/Bundles>
#include <de/App>
#include <de/Async>
#include <de/CallbackAction>
#include <de/ChildWidgetOrganizer>
#include <de/Config>
#include <de/FileSystem>
#include <de/Loop>
#include <de/PopupMenuWidget>
#include <de/ui/FilteredData>
/** @file headerwidget.cpp  Home column header.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/headerwidget.h"
#include <de/Config>
#include <de/LabelWidget>
#include <de/ButtonWidget>
#include <de/PanelWidget>
#include <de/CallbackAction>
/** @file homewidget.cpp  Root widget for the Home UI.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/homewidget.h"
#include "ui/home/columnwidget.h"
#include "ui/home/nogamescolumnwidget.h"
#include "ui/home/gamecolumnwidget.h"
#include "ui/home/multiplayercolumnwidget.h"
#include "ui/home/packagescolumnwidget.h"
#include "ui/widgets/busywidget.h"
#include "ui/savelistdata.h"
#include "ui/clientwindow.h"
#include "ui/widgets/taskbarwidget.h"
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <de/App>
#include <de/FadeToBlackWidget>
#include <de/LabelWidget>
#include <de/Loop>
#include <de/PackageLoader>
#include <de/PersistentState>
#include <de/PopupMenuWidget>
#include <de/SequentialLayout>
#include <de/StyleProceduralImage>
#include <de/TabWidget>
#include <QTimer>
/** @file multiplayercolumnwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/multiplayercolumnwidget.h"
#include "ui/widgets/multiplayerservermenuwidget.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/home/headerwidget.h"
#include "ui/clientwindow.h"
#include "ui/clientstyle.h"
#include "network/serverlink.h"
#include <doomsday/Games>
#include <de/CallbackAction>
#include <de/PopupButtonWidget>
#include <de/PopupMenuWidget>
#include <de/SignalAction>
#include <de/ui/SubwidgetItem>
/** @file multiplayerpanelbuttonwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/multiplayerpanelbuttonwidget.h"
#include "ui/dialogs/serverinfodialog.h"
#include "ui/clientwindow.h"
#include "network/net_main.h"
#include "network/serverlink.h"
#include "resource/idtech1image.h"
#include "clientapp.h"
#include "dd_main.h"
#include <doomsday/doomsdayapp.h>
#include <doomsday/games.h>
#include <doomsday/console/exec.h>
#include <doomsday/LumpCatalog>
#include <doomsday/Games>
#include <de/Async>
#include <de/charsymbols.h>
#include <de/CallbackAction>
#include <de/MessageDialog>
#include <de/PopupButtonWidget>
#include <de/PopupMenuWidget>
#include <QRegExp>
/** @file nogamescolumnwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/nogamescolumnwidget.h"
#include "ui/clientwindow.h"
#include "ui/widgets/taskbarwidget.h"
#include "clientapp.h"
#include <de/ButtonWidget>
#include <de/Config>
#include <de/SignalAction>
#include <de/TextValue>
#include <QFileDialog>
/** @file packagescolumnwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/packagescolumnwidget.h"
#include "ui/widgets/packageswidget.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/dialogs/datafilesettingsdialog.h"
#include "ui/dialogs/repositorybrowserdialog.h"
#include "ui/widgets/homeitemwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include <de/CallbackAction>
#include <de/Config>
#include <de/DirectoryListDialog>
#include <de/FileSystem>
#include <de/Loop>
#include <de/Package>
#include <de/PopupMenuWidget>
#include <de/ui/ActionItem>
#include <de/ui/SubwidgetItem>
#include <doomsday/DoomsdayApp>
#include <doomsday/Games>
#include <doomsday/resource/databundle.h>
/** @file savelistwidget.cpp  List showing the available saves of a game.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/home/savelistwidget.h"
#include "ui/home/gamepanelbuttonwidget.h"
#include "ui/savelistdata.h"
#include <doomsday/game.h>
#include <doomsday/GameStateFolder>
#include <de/CallbackAction>
#include <de/DocumentPopupWidget>
#include <de/FileSystem>
using namespace de;
DENG_GUI_PIMPL(ColumnWidget) {
    /**
     * Procedural image for drawing the background of a column.
     */
    struct BackgroundImage : public StyleProceduralImage {
        AnimationVector3 colorAnim { Animation::Linear };
        bool needUpdate = false;
        BackgroundImage(DotPath const &styleImageId, ColumnWidget &owner)
            : StyleProceduralImage(styleImageId, owner)
        {}
        void setColor(Color const &color) {
            StyleProceduralImage::setColor(color);
            colorAnim.setValue(color, 0.5);
        }
        bool update() override {
            StyleProceduralImage::update();
            const Size newPtSize = GuiWidget::pixelsToPoints(owner().rule().size());
            if (newPtSize != pointSize()) {
                setPointSize(newPtSize);
                return true;
            }
            bool update = !colorAnim.done() || needUpdate;
            if (!colorAnim.done()) needUpdate = true;
            return update;
        }
        void glMakeGeometry(GuiVertexBuilder &verts, Rectanglef const &rect) override {
            if (!allocId().isNone()) {
                Rectanglef uv = root().atlas().imageRectf(allocId());
                Vector2f const reduction(uv.width() / 40, uv.height() / 40);
                uv = uv.adjusted(reduction, -reduction);
                Rectanglef const norm = owner().normalizedRect();
                verts.makeQuad(rect,
                               Vector4f(colorAnim.value(), 1.f),
                               Rectanglef(uv.topLeft + norm.topLeft     * uv.size(),
                                          uv.topLeft + norm.bottomRight * uv.size()));
                int const edgeWidth = GuiWidget::pointsToPixels(1);
                auto const edgeUv = owner().root().atlas().imageRectf(owner().root().solidWhitePixel());
                verts.makeQuad(Rectanglef(rect.left(), rect.top(),
                                          edgeWidth, rect.height()),
                               Vector4f(0, 0, 0, 1), edgeUv);
                verts.makeQuad(Rectanglef(rect.right() - edgeWidth, rect.top(),
                                          edgeWidth, rect.height()),
                               Vector4f(0, 0, 0, 1), edgeUv);
            }
        }
    };
    bool highlighted = false;
    LabelWidget *back;
    ScrollAreaWidget *scrollArea;
    HeaderWidget *header;
    Rule const *maxContentWidth = nullptr;
    Vector4f backTintColor;
    Animation backSaturation { 0.f, Animation::Linear };
    Impl(Public *i) : Base(i) {
        back = new LabelWidget;
        back->margins().setZero();
        scrollArea = new ScrollAreaWidget;
        scrollArea->setBehavior(ChildVisibilityClipping, UnsetFlags);
        scrollArea->enableIndicatorDraw(true);
        header = new HeaderWidget;
        scrollArea->add(header);
        header->rule()
                .setInput(Rule::Left,  scrollArea->contentRule().left())
                .setInput(Rule::Top,   scrollArea->contentRule().top())
                .setInput(Rule::Width, scrollArea->contentRule().width());
    }
    ~Impl() {
        releaseRef(maxContentWidth);
    }
};
ColumnWidget::ColumnWidget(String const &name)
    : GuiWidget(name)
    , d(new Impl(this)) {
    changeRef(d->maxContentWidth, rule("home.column.content.width"));
    AutoRef<Rule> contentMargin = (rule().width() - *d->maxContentWidth) / 2;
    d->scrollArea->margins()
            .setLeft(contentMargin)
            .setRight(contentMargin);
    d->back->rule().setRect(rule());
    d->scrollArea->rule().setRect(rule());
    add(d->back);
    add(d->scrollArea);
    updateStyle();
    setBackgroundImage("home.background.column");
    setBehavior(ChildVisibilityClipping);
}
void ColumnWidget::setBackgroundImage(DotPath const &imageId) {
    auto *img = new Impl::BackgroundImage(imageId, *this);
    img->setColor(d->backTintColor);
    d->back->setImage(img);
}
ScrollAreaWidget &ColumnWidget::scrollArea() {
    return *d->scrollArea;
}
HeaderWidget &ColumnWidget::header() {
    return *d->header;
}
Rule const &ColumnWidget::maximumContentWidth() const {
    return *d->maxContentWidth;
}
String ColumnWidget::configVariableName() const {
    return ""; // Defaults to none.
}
Variable *ColumnWidget::configVariable() const {
    String name = configVariableName();
    if (name.isEmpty()) return nullptr;
    return &App::config(name);
}
String ColumnWidget::tabShortcut() const {
    return String();
}
void ColumnWidget::setHighlighted(bool highlighted) {
    if (d->highlighted != highlighted) {
        d->highlighted = highlighted;
        auto &img = d->back->image()->as<Impl::BackgroundImage>();
        img.setColor(highlighted? Vector4f(1, 1, 1, 1) : d->backTintColor);
        d->backSaturation.setValue(highlighted? 1.f : 0.f, 0.5);
    }
}
bool ColumnWidget::isHighlighted() const {
    return d->highlighted;
}
void ColumnWidget::update() {
    GuiWidget::update();
    d->back->setSaturation(d->backSaturation);
}
void ColumnWidget::updateStyle() {
    GuiWidget::updateStyle();
    d->backTintColor = Vector4f(style().colors().colorf("home.background.tint"), 1.f);
}
bool ColumnWidget::dispatchEvent(Event const &event, bool (Widget::*memberFunc)(Event const &)) {
    if (event.type() == Event::MouseButton ||
        event.type() == Event::MouseWheel) {
        MouseEvent const &mouse = event.as<MouseEvent>();
        if ((mouse.motion() == MouseEvent::Wheel || mouse.state() == MouseEvent::Pressed) &&
            rule().recti().contains(mouse.pos())) {
            emit mouseActivity(this);
        }
    }
    return GuiWidget::dispatchEvent(event, memberFunc);
}
const String GameColumnWidget::SORT_GAME_ID("game");
const String GameColumnWidget::SORT_MODS("mods");
const String GameColumnWidget::SORT_RECENTLY_PLAYED("recent");
const String GameColumnWidget::SORT_RELEASE_DATE("release");
const String GameColumnWidget::SORT_TITLE("title");
static const String VAR_SORT_BY("home.sortBy");
static const String VAR_SORT_ASCENDING("home.sortAscending");
static const String VAR_SORT_CUSTOM_SEPARATELY("home.sortCustomSeparately");
DENG_GUI_PIMPL(GameColumnWidget)
, DENG2_OBSERVES(Games, Readiness)
, DENG2_OBSERVES(Profiles, Addition)
, DENG2_OBSERVES(Variable, Change)
, DENG2_OBSERVES(ButtonWidget, StateChange)
, DENG2_OBSERVES(Profiles::AbstractProfile, Change)
, public ChildWidgetOrganizer::IWidgetFactory {
    /**
     * Menu item for a game profile.
     */
    struct ProfileItem
            : public ui::Item
            , DENG2_OBSERVES(Deletable, Deletion) // profile deletion
    {
        GameColumnWidget::Impl *d;
        GameProfile *profile;
        ProfileItem(GameColumnWidget::Impl *d, GameProfile &gameProfile)
            : d(d)
            , profile(&gameProfile) {
            profile->audienceForDeletion += this;
        }
        Game const &game() const {
            DENG2_ASSERT(profile != nullptr);
            return profile->game();
        }
        void update() const { notifyChange(); }
        void objectWasDeleted(Deletable *obj) {
            DENG2_ASSERT(static_cast<GameProfile *>(obj) == profile);
            DENG2_UNUSED(obj);
            profile = nullptr;
            d->addOrRemoveSubheading();
            auto &items = d->menu->items();
            items.remove(items.find(*this)); // item deleted
        }
        DENG2_CAST_METHODS()
    };
    LoopCallback mainCall;
    String gameFamily;
    SaveListData const &savedItems;
    HomeMenuWidget *menu;
    ButtonWidget *newProfileButton;
    int restoredSelected = -1;
    bool gotSubheading = false;
    Impl(Public *i,
         String const &gameFamily,
         SaveListData const &savedItems)
        : Base(i)
        , gameFamily(gameFamily)
        , savedItems(savedItems) {
        ScrollAreaWidget &area = self().scrollArea();
        area.add(menu = new HomeMenuWidget);
        menu->organizer().setWidgetFactory(*this);
        menu->rule()
                .setInput(Rule::Width, area.contentRule().width())
                .setInput(Rule::Left,  area.contentRule().left())
                .setInput(Rule::Top,   self().header().rule().bottom());
        menu->margins().setBottom("");
        area.add(newProfileButton = new ButtonWidget);
        newProfileButton->audienceForStateChange() += this;
        newProfileButton->setText(tr("New Profile..."));
        newProfileButton->setStyleImage("create");
        newProfileButton->setOverrideImageSize(style().fonts().font("default").height() * 1.5f);
        newProfileButton->set(Background());
        newProfileButton->setSizePolicy(ui::Filled, ui::Expand);
        newProfileButton->setTextAlignment(ui::AlignRight);
        newProfileButton->setOpacity(actionOpacity());
        newProfileButton->setActionFn([this] () {
            auto *dlg = new CreateProfileDialog(this->gameFamily);
            dlg->setDeleteAfterDismissed(true);
            dlg->setAnchorAndOpeningDirection(newProfileButton->rule(), ui::Up);
            if (dlg->exec(root())) {
                auto *added = dlg->makeProfile();
                DoomsdayApp::gameProfiles().add(added);
            }
        });
        newProfileButton->rule()
                .setInput(Rule::Left,  area.contentRule().left())
                .setInput(Rule::Width, area.contentRule().width())
                .setInput(Rule::Top,   menu->rule().bottom());
        DoomsdayApp::games().audienceForReadiness() += this;
        Config::get("home.showUnplayableGames").audienceForChange() += this;
        Config::get(VAR_SORT_BY).audienceForChange() += this;
        Config::get(VAR_SORT_ASCENDING).audienceForChange() += this;
        Config::get(VAR_SORT_CUSTOM_SEPARATELY).audienceForChange() += this;
    }
    ui::Item *findProfileItem(GameProfile const &profile) const {
        for (dsize i = 0; i < menu->items().size(); ++i) {
            ui::Item &item = menu->items().at(i);
            if (!item.semantics().testFlag(ui::Item::Separator)) {
                if (item.as<ProfileItem>().profile == &profile) return &item;
            }
        }
        return nullptr;
    }
    GamePanelButtonWidget &widgetForItem(ui::Item const &item) const {
        DENG2_ASSERT(menu->items().find(item) != ui::Data::InvalidPos);
        return menu->itemWidget<GamePanelButtonWidget>(item);
    }
    int userProfileCount() const {
        int count = 0;
        menu->items().forAll([this, &count] (ui::Item const &item) {
            if (!item.semantics().testFlag(ui::Item::Separator)) {
                auto const *profile = item.as<ProfileItem>().profile;
                if (profile && profile->isUserCreated()) ++count;
            }
            return LoopContinue;
        });
        return count;
    }
    bool addItemForProfile(GameProfile &profile) {
        auto const &games = DoomsdayApp::games();
        if (games.contains(profile.gameId())) {
            if (profile.game().family() == gameFamily) {
                DENG2_ASSERT(!findProfileItem(profile));
                profile.upgradePackages();
                menu->items() << new ProfileItem(this, profile);
                addOrRemoveSubheading();
                profile.audienceForChange() += this;
                return true;
            }
        }
        return false;
    }
    void profileAdded(Profiles::AbstractProfile &prof) override {
        mainCall.enqueue([this, &prof] () {
            if (addItemForProfile(prof.as<GameProfile>())) {
                sortItems();
                auto const *newItem = findProfileItem(prof.as<GameProfile>());
                DENG2_ASSERT(newItem);
                menu->setSelectedIndex(menu->items().find(*newItem));
            }
        });
    }
    void profileChanged(Profiles::AbstractProfile &) override {
        Loop::timer(0.1, [this]() { sortItems(); });
    }
    bool isSubheadingVisibleWithSortOptions() const {
        return Config::get().getb(VAR_SORT_CUSTOM_SEPARATELY, true);
    }
    void addOrRemoveSubheading() {
        const int  userCount         = userProfileCount();
        const bool subheadingVisible = isSubheadingVisibleWithSortOptions() && userCount > 0;
        if (subheadingVisible && !gotSubheading) {
            gotSubheading = true;
            menu->items() << new ui::Item(ui::Item::Separator, tr("Custom Profiles"));
        }
        else if (!subheadingVisible && gotSubheading) {
            for (dsize pos = 0; pos < menu->items().size(); ++pos) {
                if (menu->items().at(pos).semantics().testFlag(ui::Item::Separator)) {
                    menu->items().remove(pos);
                    break;
                }
            }
            gotSubheading = false;
        }
    }
    /**
     * Populates the game items using the currently available game profiles.
     */
    void populateItems() {
        QSet<GameProfile *> profiles;
        foreach (GameProfile *prof,
                 DoomsdayApp::gameProfiles().profilesInFamily(gameFamily)) {
            profiles.insert(prof);
        }
        QSet<GameProfile *> toAdd = profiles;
        for (ui::DataPos i = 0; i < menu->items().size(); ++i) {
            ui::Item const &item = menu->items().at(i);
            if (item.semantics() & ui::Item::Separator) {
                continue;
            }
            auto const &profItem = item.as<ProfileItem>();
            if (profiles.contains(profItem.profile)) {
                toAdd.remove(profItem.profile);
                profItem.update();
            }
            else {
                menu->items().remove(i--);
            }
        }
        foreach (GameProfile *newProf, toAdd) {
            addItemForProfile(*newProf);
        }
        addOrRemoveSubheading();
        sortItems();
    }
    enum Section { BuiltIn, Subheading, Custom };
    static Section itemSection(ui::Item const &item) {
        if (item.semantics().testFlag(ui::Item::Separator)) return Subheading;
        return item.as<ProfileItem>().profile->isUserCreated()? Custom : BuiltIn;
    }
    void sortItems() {
        const String sortBy               = Config::get().gets(VAR_SORT_BY, SORT_RELEASE_DATE);
        const bool   sortAscending        = Config::get().getb(VAR_SORT_ASCENDING, true);
        const bool   sortCustomSeparately = Config::get().getb(VAR_SORT_CUSTOM_SEPARATELY, true);
        const auto *oldSelectedItem = menu->selectedItem();
        menu->items().sort([&sortBy, sortAscending, sortCustomSeparately](const ui::Item &a,
                                                                          const ui::Item &b) {
            Section const section1 = itemSection(a);
            Section const section2 = itemSection(b);
            if (sortCustomSeparately) {
                if (section1 < section2) {
                    return true;
                }
                if (section1 > section2) {
                    return false;
                }
            }
            const GameProfile &prof1 = *a.as<ProfileItem>().profile;
            const GameProfile &prof2 = *b.as<ProfileItem>().profile;
            int cmp = 0;
            if (sortBy == SORT_RELEASE_DATE) {
                cmp = a.as<ProfileItem>().game().releaseDate().year() -
                      b.as<ProfileItem>().game().releaseDate().year();
            }
            else if (sortBy == SORT_GAME_ID) {
                cmp = a.as<ProfileItem>().game().id().compare(b.as<ProfileItem>().game().id());
            }
            else if (sortBy == SORT_TITLE) {
                cmp = prof1.name().compareWithoutCase(prof2.name());
            }
            else if (sortBy == SORT_MODS) {
                cmp = de::cmp(prof1.packages().size(), prof2.packages().size());
            }
            else if (sortBy == SORT_RECENTLY_PLAYED) {
                if (prof1.lastPlayedAt().isValid() && prof2.lastPlayedAt().isValid()) {
                    cmp = -de::cmp(prof1.lastPlayedAt(), prof2.lastPlayedAt());
                }
                else if (prof1.lastPlayedAt().isValid()) {
                    cmp = -1;
                }
                else if (prof2.lastPlayedAt().isValid()) {
                    cmp = +1;
                }
            }
            if (cmp == 0) {
                if (prof1.isPlayable() && !prof2.isPlayable()) return sortAscending;
                if (!prof1.isPlayable() && prof2.isPlayable()) return !sortAscending;
                cmp = prof1.name().compareWithoutCase(prof2.name());
                if (cmp == 0) {
                    cmp = prof1.gameId().compareWithoutCase(prof2.gameId());
                }
            }
            if (!sortAscending) {
                cmp = -cmp;
            }
            return cmp < 0;
        });
        if (oldSelectedItem) {
            menu->setSelectedIndex(menu->items().find(*oldSelectedItem));
        }
    }
    void updateItems() {
        menu->items().forAll([] (ui::Item const &item) {
            if (!item.semantics().testFlag(ui::Item::Separator)) {
                item.as<ProfileItem>().update();
            }
            return LoopContinue;
        });
        menu->updateLayout();
    }
    void gameReadinessUpdated() override {
        populateItems();
        if (restoredSelected >= 0) {
            menu->setSelectedIndex(restoredSelected);
            restoredSelected = -1;
        }
        DoomsdayApp::gameProfiles().audienceForAddition() += this;
    }
    void variableValueChanged(Variable &var, Value const &) override {
        if (var.name().beginsWith("sort")) {
            addOrRemoveSubheading();
            sortItems();
        }
        else {
            updateItems();
        }
    }
    GuiWidget *makeItemWidget(ui::Item const &item, GuiWidget const *) override {
        if (item.semantics().testFlag(ui::Item::Separator)) {
            auto *heading = LabelWidget::newWithText(tr("Custom Profiles"));
            heading->setSizePolicy(ui::Filled, ui::Expand);
            heading->setFont("heading");
            heading->setAlignment(ui::AlignLeft);
            heading->margins().setLeftRight("");
            heading->setOpacity(.75f);
            return heading;
        }
        auto const *profileItem = &item.as<ProfileItem>();
        auto *button = new GamePanelButtonWidget(*profileItem->profile, savedItems);
        QObject::connect(button, &HomeItemWidget::openContextMenu, [this, button, profileItem] () {
            auto *popup = new PopupMenuWidget;
            button->add(popup);
            popup->setDeleteAfterDismissed(true);
            popup->setAnchorAndOpeningDirection(button->label().rule(), ui::Down);
            bool const isUserProfile = profileItem->profile->isUserCreated();
            if (isUserProfile) {
                popup->items()
                    << new ui::ActionItem(tr("Edit..."), new CallbackAction([this, button, profileItem] () {
                        auto *dlg = CreateProfileDialog::editProfile(gameFamily, *profileItem->profile);
                        dlg->setAnchorAndOpeningDirection(button->label().rule(), ui::Up);
                        dlg->setAnchorX(button->rule().midX()); // keep centered in column
                        dlg->setDeleteAfterDismissed(true);
                        if (dlg->exec(root())) {
                            dlg->applyTo(*profileItem->profile);
                            profileItem->update();
                        }
                    }));
            }
            popup->items()
                << new ui::ActionItem(tr("Select Mods..."),
                                      new SignalAction(button, SLOT(selectPackages())))
                << new ui::ActionItem(tr("Clear Mods"), new CallbackAction([this, button]() {
                                          button->clearPackages();
                                      }))
                << new ui::ActionItem(
                       tr("Duplicate"), new CallbackAction([profileItem]() {
                           GameProfile *dup = new GameProfile(*profileItem->profile);
                           dup->setUserCreated(true);
                           dup->createSaveLocation();
                           for (int attempt = 1;; ++attempt) {
                               String newName;
                               if (attempt > 1)
                                   newName = String("%1 (Copy %2)").arg(dup->name()).arg(attempt);
                               else
                                   newName = String("%1 (Copy)").arg(dup->name());
                               if (!DoomsdayApp::gameProfiles().tryFind(newName)) {
                                   dup->setName(newName);
                                   DoomsdayApp::gameProfiles().add(dup);
                                   break;
                               }
                           }
                       }));
            if (const auto *loc = FS::tryLocate<const Folder>(profileItem->profile->savePath())) {
                popup->items() << new ui::Item(ui::Item::Separator)
                               << new ui::ActionItem(
                                      "Show Save Folder", new CallbackAction([loc]() {
                                          QDesktopServices::openUrl(
                                              QUrl::fromLocalFile(loc->correspondingNativePath()));
                                      }));
            }
            if (isUserProfile && !profileItem->profile->saveLocationId()) {
                popup->items()
                    << new ui::ActionItem(
                           "Create New Save Folder", new CallbackAction([this, profileItem]() {
                               profileItem->profile->createSaveLocation();
                               auto *msg = new MessageDialog;
                               msg->setDeleteAfterDismissed(true);
                               msg->title().setText("Save Folder Created");
                               msg->message().setText(
                                   "Save files of the profile " _E(b) +
                                   profileItem->profile->name() + _E(.) " will be written to " +
                                   FS::locate<const Folder>(profileItem->profile->savePath())
                                       .description() + ".");
                               msg->buttons() << new DialogButtonItem(DialogWidget::Accept |
                                                                      DialogWidget::Default);
                               msg->exec(root());
                           }));
            }
            if (isUserProfile) {
                auto *deleteSub = new ui::SubmenuItem(style().images().image("close.ring"),
                                                      tr("Delete"), ui::Left);
                deleteSub->items()
                    << new ui::Item(ui::Item::Separator, tr("Are you sure?"))
                    << new ui::ActionItem(
                           tr("Delete Profile"),
                           new CallbackAction([this, button, profileItem, popup]() {
                               if (profileItem->profile->saveLocationId()) {
                                   const Folder *saveFolder =
                                       FS::tryLocate<const Folder>(profileItem->profile->savePath());
                                   if (saveFolder && !profileItem->profile->isSaveLocationEmpty()) {
                                       auto *question = new MessageDialog;
                                       question->setDeleteAfterDismissed(true);
                                       question->title().setText("Delete Saved Games?");
                                       question->title().setStyleImage("alert");
                                       question->message().setText(
                                           "The profile " _E(b) + profileItem->profile->name() +
                                           _E(.) " that is being deleted has saved games. "
                                                 "Do you wish to delete the save files as well?");
                                       const NativePath savePath =
                                           saveFolder->correspondingNativePath();
                                       question->buttons()
                                           << new DialogButtonItem(DialogWidget::Accept,
                                                                   "Delete All")
                                           << new DialogButtonItem(DialogWidget::Reject |
                                                                       DialogWidget::Default,
                                                                   "Cancel")
                                           << new DialogButtonItem(
                                                  DialogWidget::Action,
                                                  "Show Folder",
                                                  new CallbackAction([savePath]() {
                                                      QDesktopServices::openUrl(
                                                          QUrl::fromLocalFile(savePath));
                                                  }));
                                       if (!question->exec(root())) {
                                           return;
                                       }
                                   }
                                   if (saveFolder) {
                                       profileItem->profile->destroySaveLocation();
                                   }
                               }
                               const TimeSpan SPAN = 0.2;
                               button->setOpacity(0, SPAN);
                               popup->detachAnchor();
                               popup->close();
                               Loop::get().timer(SPAN,
                                                 [profileItem]() { delete profileItem->profile; });
                           }))
                    << new ui::ActionItem(tr("Cancel"), new Action);
                popup->items()
                    << new ui::Item(ui::Item::Separator)
                    << deleteSub;
            }
            popup->open();
        });
        return button;
    }
    void updateItemWidget(GuiWidget &widget, ui::Item const &item) override {
        if (item.semantics().testFlag(ui::Item::Separator)) return; // Ignore.
        auto &drawer = widget.as<GamePanelButtonWidget>();
        drawer.updateContent();
        if (!Config::get().getb("home.showUnplayableGames")) {
            drawer.show(item.as<ProfileItem>().game().isPlayableWithDefaultPackages());
        }
        else {
            drawer.show();
        }
    }
    float actionOpacity() const {
        return self().isHighlighted()? .4f : 0.f;
    }
    void buttonStateChanged(ButtonWidget &button, ButtonWidget::State state) override {
        TimeSpan const SPAN = 0.25;
        switch (state) {
        case ButtonWidget::Up:
            button.setOpacity(actionOpacity(), SPAN);
            break;
        case ButtonWidget::Hover:
            button.setOpacity(.8f, SPAN);
            break;
        case ButtonWidget::Down:
            button.setOpacity(1);
            break;
        }
    }
    void showActions(bool show) {
        newProfileButton->setOpacity(show? actionOpacity() : 0, 0.6);
    }
};
GameColumnWidget::GameColumnWidget(String const &gameFamily,
                                   SaveListData const &savedItems)
    : ColumnWidget(gameFamily.isEmpty()? "other-column"
                                       : (gameFamily.toLower() + "-column"))
    , d(new Impl(this, gameFamily.toLower(), savedItems)) {
    scrollArea().setContentSize(maximumContentWidth(),
                                header().rule().height() +
                                rule("gap") +
                                d->menu->rule().height() +
                                d->newProfileButton->rule().height());
    header().title().setText(String(_E(s)_E(C) "%1\n" _E(.)_E(.)_E(w) "%2")
                             .arg( gameFamily == "DOOM"? "id Software" :
                                  !gameFamily.isEmpty()? "Raven Software" : "")
                             .arg(!gameFamily.isEmpty()? QString(gameFamily)
                                                       : tr("Other Games")));
    if (!gameFamily.isEmpty()) {
        header().setLogoImage("logo.game." + gameFamily.toLower());
        header().setLogoBackground("home.background." + d->gameFamily);
        setBackgroundImage("home.background." + d->gameFamily);
    }
    else {
        setBackgroundImage("home.background.other");
    } {
        header().menuButton().setPopup(
            [](const PopupButtonWidget &) -> PopupWidget * {
                auto *pop = new GridPopupWidget;
                auto *sortBy =
                    new VariableChoiceWidget(Config::get(VAR_SORT_BY), VariableChoiceWidget::Text);
                sortBy->items() << new ChoiceItem("Recently played", SORT_RECENTLY_PLAYED)
                                << new ChoiceItem("Release date", SORT_RELEASE_DATE)
                                << new ChoiceItem("Title", SORT_TITLE)
                                << new ChoiceItem("Mods", SORT_MODS)
                                << new ChoiceItem("Game ID", SORT_GAME_ID);
                sortBy->updateFromVariable();
                *pop << LabelWidget::newWithText("Show:")
                     << new VariableToggleWidget("Descriptions",
                                                 Config::get("home.showColumnDescription"))
                     << Const(0)
                     << new VariableToggleWidget("Unplayable Games",
                                                 Config::get("home.showUnplayableGames"))
                     << LabelWidget::newWithText("Sort By:") << sortBy << Const(0)
                     << new VariableToggleWidget("Ascending", Config::get(VAR_SORT_ASCENDING))
                     << Const(0)
                     << new VariableToggleWidget("Separate Custom",
                                                 Config::get(VAR_SORT_CUSTOM_SEPARATELY));
                pop->commit();
                return pop;
            },
            ui::Down);
    } {
    if (name() == "doom-column") {
        header().info().setText("id Software released DOOM for MS-DOS in 1993. "
                                "It soon became a massive hit and is regarded as "
                                "the game that popularized the first-person shooter "
                                "genre. Since then the franchise has been continued "
                                "in several sequels, starting with DOOM II: Hell on "
                                "Earth in 1994. DOOM and many of its follow-ups "
                                "have been ported to numerous other platforms, and "
                                "to this day remains a favorite among gamers.");
    }
    else if (name() == "heretic-column") {
        header().info().setText("Raven Software released Heretic in 1994. It used "
                                "a modified version of id Software's DOOM engine. "
                                "The game featured such enhancements as inventory "
                                "management and the ability to look up and down. "
                                "Ambient sound effects were used to improve the "
                                "atmosphere of the game world.");
    }
    else if (name() == "hexen-column") {
        header().info().setText("Raven Software released Hexen in 1996. The "
                                "company had continued making heavy modifications "
                                "to the DOOM engine, and Hexen introduced such "
                                "sophisticated features as a scripting language "
                                "for game events. The maps were well-designed and "
                                "interconnected with each other, resulting in a "
                                "more intriguing game world and more complex "
                                "puzzles to solve.");
    }
    else {
        header().info().setText("Thanks to its excellent modding support, DOOM has "
                                "been used as a basis for many games and community "
                                "projects.");
    }
    }
}
String GameColumnWidget::tabHeading() const {
    if (d->gameFamily.isEmpty()) return tr("Other");
    return d->gameFamily.at(0).toUpper() + d->gameFamily.mid(1);
}
String GameColumnWidget::tabShortcut() const {
    if (name() == "hexen-column") return QStringLiteral("x");
    return String();
}
String GameColumnWidget::configVariableName() const {
    return "home.columns." + (!d->gameFamily.isEmpty()? d->gameFamily
                                                      : String("otherGames"));
}
void GameColumnWidget::setHighlighted(bool highlighted) {
    ColumnWidget::setHighlighted(highlighted);
    if (highlighted) {
        d->menu->restorePreviousSelection();
    }
    else {
        root().setFocus(nullptr);
        d->menu->unselectAll();
    }
    d->showActions(highlighted);
}
void GameColumnWidget::operator>>(PersistentState &toState) const {
    Record &rec = toState.objectNamespace();
    rec.set(name().concatenateMember("selected"), d->menu->selectedIndex());
}
void GameColumnWidget::operator<<(PersistentState const &fromState) {
    Record const &rec = fromState.objectNamespace();
    d->restoredSelected = rec.geti(name().concatenateMember("selected"), -1);
}
DENG_GUI_PIMPL(GamePanelButtonWidget)
, DENG2_OBSERVES(Profiles::AbstractProfile, Change)
, DENG2_OBSERVES(res::Bundles, Identify)
, DENG2_OBSERVES(Variable, Change)
, DENG2_OBSERVES(ButtonWidget, StateChange)
, public AsyncScope {
    GameProfile &gameProfile;
    ui::FilteredDataT<SaveListData::SaveItem> savedItems;
    SaveListWidget *saves;
    PackagesButtonWidget *packagesButton;
    ButtonWidget *playButton;
    ButtonWidget *deleteSaveButton;
    LabelWidget *problemIcon;
    LabelWidget *packagesCounter;
    res::LumpCatalog catalog;
    bool playHovering = false;
    Impl(Public *i, GameProfile &profile, SaveListData const &allSavedItems)
        : Base(i)
        , gameProfile(profile)
        , savedItems(allSavedItems) {
        savedItems.setFilter([this] (ui::Item const &it) {
            auto const &item = it.as<SaveListData::SaveItem>();
            if (item.gameId() != gameProfile.gameId()) {
                return false;
            }
            if (!gameProfile.isPlayable()) {
                return false;
            }
            if (item.savePath().fileNamePath().compareWithoutCase(gameProfile.savePath())) {
                return false;
            }
            StringList const savePacks = item.loadedPackages();
            if (savePacks.isEmpty()) {
                return !gameProfile.isUserCreated();
            }
            if (!gameProfile.isCompatibleWithPackages(savePacks)) {
                return false;
            }
            return true;
        });
        packagesButton = new PackagesButtonWidget;
        packagesButton->setGameProfile(gameProfile);
        packagesButton->setDialogTitle(tr("Mods for %1").arg(profile.name()));
        packagesButton->setSetupCallback([this] (PackagesDialog &dialog) {
            dialog.buttons()
                    << new DialogButtonItem(DialogWidget::Action,
                                            style().images().image("play"),
                                            tr("Play"),
                                            new CallbackAction([this, &dialog] () {
                                                dialog.accept();
                                                playButtonPressed();
                                            }));
        });
        self().addButton(packagesButton);
        QObject::connect(packagesButton,
                         &PackagesButtonWidget::packageSelectionChanged,
                         [this] (QStringList ids) {
            gameProfile.setPackages(toStringList(ids));
            self().updateContent();
        });
        playButton = new ButtonWidget;
        playButton->useInfoStyle();
        playButton->setStyleImage("play", "default");
        playButton->setImageColor(style().colors().colorf("inverted.text"));
        playButton->setActionFn([this] () { playButtonPressed(); });
        playButton->audienceForStateChange() += this;
        self().addButton(playButton);
        saves = new SaveListWidget(self());
        saves->rule().setInput(Rule::Width, self().rule().width());
        saves->margins().setZero().setLeft(self().icon().rule().width());
        saves->setItems(savedItems);
        deleteSaveButton = new ButtonWidget;
        deleteSaveButton->setStyleImage("close.ring", "default");
        deleteSaveButton->setSizePolicy(ui::Expand, ui::Expand);
        deleteSaveButton->set(Background());
        deleteSaveButton->hide();
        deleteSaveButton->setActionFn([this] () { deleteButtonPressed(); });
        self().panel().add(deleteSaveButton);
        problemIcon = new LabelWidget;
        problemIcon->setStyleImage("alert", "default");
        problemIcon->setImageColor(style().colors().colorf("accent"));
        problemIcon->rule().setRect(self().icon().rule());
        problemIcon->hide();
        self().icon().add(problemIcon);
        packagesCounter = new LabelWidget;
        packagesCounter->setOpacity(.5f);
        packagesCounter->setSizePolicy(ui::Expand, ui::Expand);
        packagesCounter->set(GuiWidget::Background());
        packagesCounter->setStyleImage("package.icon", "small");
        packagesCounter->setFont("small");
        packagesCounter->margins().setLeft("");
        packagesCounter->setTextAlignment(ui::AlignLeft);
        packagesCounter->rule()
                .setInput(Rule::Right, self().label().rule().right())
                .setMidAnchorY(self().label().rule().midY());
        packagesCounter->hide();
        self().label().add(packagesCounter);
        self().label().setMinimumHeight(style().fonts().font("default").lineSpacing() * 3 +
                                      self().label().margins().height());
        self().panel().setContent(saves);
        saves->updateLayout();
        self().panel().open();
        DoomsdayApp::bundles().audienceForIdentify() += this;
        Config::get("home.sortBy").audienceForChange() += this;
    }
    Game const &game() const {
        return gameProfile.game();
    }
    void updatePackagesIndicator() {
        int const  count = gameProfile.packages().size();
        bool const shown = !isMissingPackages() && count > 0 && !self().isSelected();
        packagesCounter->setText(String::number(count));
        packagesCounter->show(shown);
        if (shown) {
            self().setLabelMinimumRightMargin(packagesCounter->rule().width());
        }
        else {
            self().setLabelMinimumRightMargin(Const(0));
        }
    }
    void buttonStateChanged(ButtonWidget &/*playButton*/, ButtonWidget::State state) override
    {
        playHovering = (state != ButtonWidget::Up);
        self().updateContent();
    }
    void playButtonPressed() {
        if (playButton->isEnabled()) {
            BusyMode_FreezeGameForBusyMode();
            gameProfile.setLastPlayedAt();
            DoomsdayApp::app().changeGame(gameProfile, DD_ActivateGameWorker);
            if (saves->selectedPos() != ui::Data::InvalidPos) {
                auto const &saveItem = savedItems.at(saves->selectedPos());
                Con_Execute(CMDS_DDAY, ("loadgame " + saveItem.name() + " confirm").toLatin1(),
                            false, false);
            }
        }
    }
    void deleteButtonPressed() {
        DENG2_ASSERT(saves->selectedPos() != ui::Data::InvalidPos);
        PopupMenuWidget *pop = new PopupMenuWidget;
        pop->setDeleteAfterDismissed(true);
        pop->setAnchorAndOpeningDirection(deleteSaveButton->rule(), ui::Down);
        pop->items()
                << new ui::Item(ui::Item::Separator, tr("Are you sure?"))
                << new ui::ActionItem(tr("Delete Savegame"),
                                      new CallbackAction([this, pop] () {
                    pop->detachAnchor();
                    String const path = savedItems.at(saves->selectedPos()).savePath();
                    self().unselectSave();
                    App::rootFolder().destroyFile(path);
                    FS::get().refreshAsync();
                }))
                << new ui::ActionItem(tr("Cancel"), new Action /* nop */);
        self().add(pop);
        pop->open();
    }
    void updateGameTitleImage() {
        *this += async([this]() { return IdTech1Image::makeGameLogo(game(), catalog); },
                       [this](const Image &gameLogo) { self().icon().setImage(gameLogo); });
    }
    void profileChanged(Profiles::AbstractProfile &) override {
        self().updateContent();
    }
    void variableValueChanged(Variable &, Value const &) override {
        self().updateContent();
    }
    void dataBundlesIdentified() override {
        Loop::get().mainCall([this]() {
            gameProfile.upgradePackages();
            catalog.clear();
            self().updateContent();
        });
    }
    bool isMissingPackages() const {
        return !gameProfile.isPlayable() && game().isPlayableWithDefaultPackages();
    }
    String defaultSubtitle() const {
        if (gameProfile.isUserCreated()) {
            return game().title();
        }
        if (Config::get("home.sortBy") != GameColumnWidget::SORT_RECENTLY_PLAYED) {
            return String::number(game().releaseDate().year());
        }
        return {};
    }
};
GamePanelButtonWidget::GamePanelButtonWidget(GameProfile &game, SaveListData const &savedItems)
    : d(new Impl(this, game, savedItems)) {
    connect(d->saves, SIGNAL(selectionChanged(de::ui::DataPos)), this, SLOT(saveSelected(de::ui::DataPos)));
    connect(d->saves, SIGNAL(doubleClicked(de::ui::DataPos)), this, SLOT(saveDoubleClicked(de::ui::DataPos)));
    connect(this, SIGNAL(doubleClicked()), this, SLOT(play()));
    game.audienceForChange() += d;
}
void GamePanelButtonWidget::setSelected(bool selected) {
    if ((isSelected() && !selected) || (!isSelected() && selected)) {
        PanelButtonWidget::setSelected(selected);
        d->playButton->enable(selected);
        if (!selected) {
            unselectSave();
        }
        updateContent();
    }
}
void GamePanelButtonWidget::updateContent() {
    bool const isPlayable     = d->gameProfile.isPlayable();
    bool const isGamePlayable = d->game().isPlayableWithDefaultPackages();
    playButton().enable(isPlayable);
    playButton().show(isPlayable);
    updateButtonLayout();
    d->problemIcon->show(!isGamePlayable);
    d->updatePackagesIndicator();
    const String sortBy = Config::get("home.sortBy");
    String meta;
    if (sortBy == GameColumnWidget::SORT_RECENTLY_PLAYED) {
        meta = d->gameProfile.lastPlayedAt().isValid()
                   ? "Last played " + d->gameProfile.lastPlayedAt().asText(Time::FriendlyFormat)
                   : "";
    }
    else if (sortBy == GameColumnWidget::SORT_GAME_ID) {
        meta = d->game().id();
    }
    if (meta.isEmpty()) {
        meta = d->defaultSubtitle();
    }
    if (isSelected()) {
        if (!isGamePlayable) {
            meta = _E(D) + tr("Missing data files") + _E(.);
        }
    }
    if (d->playHovering) {
        if (d->saves->selectedPos() != ui::Data::InvalidPos) {
            meta = tr("Restore saved game");
        }
        else {
            meta = tr("Start game");
        }
    }
    if (d->isMissingPackages()) {
        meta = _E(D) "Missing mods" _E(.);
        d->packagesButton->setOverrideLabel("Fix...");
        setKeepButtonsVisible(true);
    }
    else {
        d->packagesButton->setOverrideLabel("");
        setKeepButtonsVisible(false);
    }
    label().setText(String(_E(b) "%1\n" _E(l)_E(C) "%2")
                    .arg(d->gameProfile.name())
                    .arg(meta));
    d->packagesButton->setPackages(d->gameProfile.packages());
    d->updatePackagesIndicator();
    if (d->catalog.setPackages(d->gameProfile.allRequiredPackages())) {
        d->updateGameTitleImage();
    }
    d->savedItems.refilter();
}
void GamePanelButtonWidget::unselectSave() {
    d->saves->clearSelection();
}
ButtonWidget &GamePanelButtonWidget::playButton() {
    return *d->playButton;
}
bool GamePanelButtonWidget::handleEvent(Event const &event) {
    if (hasFocus() && event.isKey()) {
        auto const &key = event.as<KeyEvent>();
        if (event.isKeyDown() &&
            (key.ddKey() == DDKEY_ENTER || key.ddKey() == DDKEY_RETURN || key.ddKey() == ' ')) {
            if (playButton().isEnabled()) {
                root().setFocus(&playButton());
            }
            else {
                root().setFocus(d->packagesButton);
            }
        }
    }
    return PanelButtonWidget::handleEvent(event);
}
void GamePanelButtonWidget::play() {
    d->playButtonPressed();
}
void GamePanelButtonWidget::selectPackages() {
    d->packagesButton->trigger();
}
void GamePanelButtonWidget::clearPackages() {
    d->gameProfile.setPackages(StringList());
    updateContent();
}
void GamePanelButtonWidget::saveSelected(de::ui::DataPos savePos) {
    if (savePos != ui::Data::InvalidPos) {
        DENG2_ASSERT(parentMenu() != nullptr);
        parentMenu()->setSelectedIndex(parentMenu()->findItem(*this));
        GuiWidget &widget = d->saves->itemWidget<GuiWidget>(d->savedItems.at(savePos));
        d->deleteSaveButton->rule()
                .setMidAnchorY(widget.rule().midY())
                .setInput(Rule::Right, widget.rule().left());
        d->deleteSaveButton->show();
    }
    else {
        d->deleteSaveButton->hide();
    }
    updateContent();
}
void GamePanelButtonWidget::saveDoubleClicked(de::ui::DataPos savePos) {
    d->saves->setSelectedPos(savePos);
    play();
}
DENG_GUI_PIMPL(HeaderWidget)
, DENG2_OBSERVES(Variable, Change) {
    LabelWidget *logo;
    LabelWidget *logoBg; /// @todo Backgrounds should support ProceduralImages. -jk
    LabelWidget *title;
    PanelWidget *infoPanel;
    LabelWidget *info;
    PopupButtonWidget *menuButton;
    Impl(Public *i) : Base(i) {
        self().add(logoBg    = new LabelWidget);
        self().add(logo      = new LabelWidget);
        self().add(title     = new LabelWidget);
        self().add(infoPanel = new PanelWidget);
        info  = new LabelWidget;
        infoPanel->setContent(info);
        if (showDescriptionVar().value().isTrue()) {
            infoPanel->open();
        }
        self().add(menuButton = new PopupButtonWidget);
        showDescriptionVar().audienceForChange() += this;
    }
    void variableValueChanged(Variable &, Value const &newValue) {
        if (newValue.isTrue()) {
            infoPanel->open();
        }
        else {
            infoPanel->close();
        }
    }
    static Variable const &showDescriptionVar() {
        return Config::get("home.showColumnDescription");
    }
};
HeaderWidget::HeaderWidget()
    : d(new Impl(this)) {
    const Rule &logoHeight = rule("home.header.logo.height");
    margins().setTop("gap");
    d->title->setAlignment(ui::AlignLeft);
    d->title->setTextLineAlignment(ui::AlignLeft);
    d->title->setFont("title");
    d->title->setSizePolicy(ui::Fixed, ui::Expand);
    d->title->margins()
            .setLeft("")
            .setBottom(style().fonts().font("title").descent());
    d->logo->setSizePolicy(ui::Filled, ui::Filled);
    d->logo->setImageFit(ui::FitToSize | ui::OriginalAspectRatio);
    d->logoBg->setSizePolicy(ui::Filled, ui::Filled);
    d->logoBg->setImageFit(ui::FitToSize);
    d->logoBg->margins().setZero();
    d->info->setTextColor("label.dimmed");
    d->info->setAlignment(ui::AlignLeft);
    d->info->setTextLineAlignment(ui::AlignLeft);
    d->info->setMaximumTextWidth(rule().width());
    d->info->setSizePolicy(ui::Fixed, ui::Expand);
    d->info->margins().setZero().setTop("gap");
    d->menuButton->setSizePolicy(ui::Expand, ui::Expand);
    d->menuButton->setFont("small");
    d->menuButton->setText("...");
    d->menuButton->margins().setTopBottom(RuleBank::UNIT);
    /*d->menuButton->setAction(new CallbackAction([this] ()
    {
        if (d->infoPanel->isOpen()) {
            d->infoPanel->close(0);
        }
        else {
            d->infoPanel->open();
        }
    }));*/
    d->logoBg->rule().setRect(d->logo->rule());
    d->logo->rule()
            .setInput(Rule::Height, logoHeight)
            .setInput(Rule::Width,  Const(0))
            .setInput(Rule::Top,    rule().top() + margins().top())
            .setInput(Rule::Left,   rule().left());
    d->menuButton->rule()
            .setInput(Rule::Bottom, d->logo->rule().bottom())
            .setInput(Rule::Left,   d->logo->rule().right() +
                                    d->title->margins().left());
    d->title->rule()
            .setInput(Rule::Right,  rule().right())
            .setInput(Rule::Bottom, d->menuButton->rule().top())
            .setInput(Rule::Left,   d->logo->rule().right());
    d->infoPanel->rule()
            .setInput(Rule::Top,    d->logo->rule().bottom())
            .setInput(Rule::Left,   rule().left());
    d->info->rule()
            .setInput(Rule::Width,  rule().width());
    rule().setInput(Rule::Height, logoHeight + d->infoPanel->rule().height() +
                    margins().height());
}
LabelWidget &HeaderWidget::logo() {
    return *d->logo;
}
LabelWidget &HeaderWidget::title() {
    return *d->title;
}
LabelWidget &HeaderWidget::info() {
    return *d->info;
}
PanelWidget &HeaderWidget::infoPanel() {
    return *d->infoPanel;
}
PopupButtonWidget &HeaderWidget::menuButton() {
    return *d->menuButton;
}
void HeaderWidget::setLogoImage(DotPath const &imageId) {
    d->logo->setStyleImage(imageId);
    d->logo->rule().setInput(Rule::Width, rule("home.header.logo.width"));
    d->title->margins().setLeft("gap");
}
void HeaderWidget::setLogoBackground(DotPath const &imageId) {
    d->logoBg->setStyleImage(imageId);
}
static TimeSpan const SCROLL_SPAN = .5;
static TimeSpan const DISMISS_SPAN = 1.5;
DENG_GUI_PIMPL(HomeWidget)
, DENG2_OBSERVES(App,          StartupComplete)
, DENG2_OBSERVES(Games,        Readiness)
, DENG2_OBSERVES(DoomsdayApp,  GameLoad)
, DENG2_OBSERVES(DoomsdayApp,  GameChange)
, DENG2_OBSERVES(Variable,     Change)
, DENG2_OBSERVES(ButtonWidget, StateChange) {
    struct Column {
        ColumnWidget *widget;
        SafePtr<Variable> configVar;
        Column(ColumnWidget *w, Variable *var) : widget(w), configVar(var) {}
    };
    LoopCallback mainCall;
    SaveListData savedItems; ///< All the available save games as items.
    dsize visibleColumnCount = 2;
    QList<Column> allColumns; // not owned
    QList<ColumnWidget *> columns; // Only the visible ones (not owned).
    IndirectRule *columnWidth;
    LabelWidget *tabsBackground;
    TabWidget *tabs;
    SafeWidgetPtr<FadeToBlackWidget> blanker;
    int currentOffsetTab = 0;
    AnimationRule *scrollOffset;
    AnimationRule *dismissOffset;
    ButtonWidget *moveLeft;
    ButtonWidget *moveRight;
    QTimer moveShowTimer;
    ButtonWidget *taskBarHintButton;
    bool dismissing = false;
    bool havePackages = false;
    int restoredOffsetTab = -1;
    int restoredActiveTab = -1;
    Impl(Public *i) : Base(i) {
        App::app().audienceForStartupComplete()     += this;
        DoomsdayApp::games().audienceForReadiness() += this;
        DoomsdayApp::app().audienceForGameChange()  += this;
        DoomsdayApp::app().audienceForGameLoad()    += this;
        columnWidth   = new IndirectRule;
        scrollOffset  = new AnimationRule(0, Animation::EaseOut);
        dismissOffset = new AnimationRule(0, Animation::EaseBoth);
        tabs = new TabWidget;
        tabsBackground = new LabelWidget;
        if (style().isBlurringAllowed()) {
            tabsBackground->set(Background(Vector4f(1, 1, 1, 1), Background::Blurred));
        }
        else {
            tabsBackground->set(Background(Vector4f(0, 0, 0, 1)));
        }
        moveLeft  = new ButtonWidget;
        moveRight = new ButtonWidget;
        moveLeft ->setActionFn([this] () { tabs->setCurrent(visibleTabRange().start - 1); });
        moveRight->setActionFn([this] () { tabs->setCurrent(visibleTabRange().end); });
        configureEdgeNavigationButton(*moveLeft);
        configureEdgeNavigationButton(*moveRight);
        taskBarHintButton = new ButtonWidget;
        taskBarHintButton->setBackgroundColor("");
        taskBarHintButton->setSizePolicy(ui::Expand, ui::Expand);
        taskBarHintButton->margins().set("dialog.gap");
        taskBarHintButton->setText(_E(b) "ESC" _E(.) + tr(" Task Bar"));
        taskBarHintButton->setFont("small");
        taskBarHintButton->setOpacity(.66f);
        taskBarHintButton->rule()
                .setInput(Rule::Right,  self().rule().right()  - rule("dialog.gap"))
                .setInput(Rule::Bottom, self().rule().bottom() - rule("dialog.gap") + *dismissOffset);
        taskBarHintButton->setActionFn([this] () {
            ClientWindow::main().taskBar().open();
        });
        mainCall.enqueue([this] () {
            QObject::connect(&ClientWindow::main().taskBar(), &TaskBarWidget::opened, [this] () {
                taskBarHintButton->disable();
                taskBarHintButton->setOpacity(0, 0.25);
            });
            QObject::connect(&ClientWindow::main().taskBar(), &TaskBarWidget::closed, [this] () {
                taskBarHintButton->enable();
                taskBarHintButton->setOpacity(.66f, 0.5);
            });
        });
        moveShowTimer.setSingleShot(true);
        moveShowTimer.setInterval(SCROLL_SPAN.asMilliSeconds());
        QObject::connect(&moveShowTimer, &QTimer::timeout, [this] () {
            moveLeft ->show(tabs->current() != 0);
            moveRight->show(tabs->current() < tabs->items().size() - 1);
        });
    }
    ~Impl() {
        releaseRef(columnWidth);
        releaseRef(scrollOffset);
        releaseRef(dismissOffset);
    }
    void configureEdgeNavigationButton(ButtonWidget &button) {
        button.set(Background(style().colors().colorf("text")));
        button.setOpacity(0);
        button.setBehavior(Widget::Focusable, UnsetFlags); // only for the mouse
        button.setAttribute(EatAllMouseEvents, SetFlags);
        button.audienceForStateChange() += this;
        button.rule()
                .setInput(Rule::Width,  rule(RuleBank::UNIT))
                .setInput(Rule::Bottom, self().rule().bottom())
                .setInput(Rule::Top,    tabs->rule().bottom());
    }
    void buttonStateChanged(ButtonWidget &button, ButtonWidget::State state) {
        if (state == ButtonWidget::Down) {
            button.setOpacity(.3f);
        }
        else {
            button.setOpacity(state == ButtonWidget::Up? 0 : .1f, 0.25);
        }
    }
    void updateVisibleColumnsAndTabHeadings() {
        bool const gotGames = DoomsdayApp::games().numPlayable() > 0;
        allColumns.at(0).widget->show(!gotGames);
        for (int i = 1; i < allColumns.size(); ++i) {
            Column const &col = allColumns.at(i);
            if (col.configVar) {
                col.widget->show(gotGames && col.configVar->value().isTrue());
            }
            if (is<PackagesColumnWidget>(col.widget)) {
                col.widget->show(gotGames || havePackages);
            }
        }
        struct TabSpec {
            Column const *col;
            int visibleIndex;
        };
        QList<TabSpec> specs;
        int visibleIndex = 0;
        for (int index = 0; index < allColumns.size(); ++index) {
            Column const &col = allColumns.at(index);
            if (!col.widget->behavior().testFlag(Widget::Hidden)) {
                TabSpec ts;
                ts.col = &col;
                ts.visibleIndex = visibleIndex++;
                specs << ts;
            }
        }
        if (tabs->items().size() == dsize(specs.size())) {
            bool differenceFound = false;
            for (dsize pos = 0; pos < tabs->items().size(); ++pos) {
                auto const &item = tabs->items().at(pos);
                if (item.label() != specs.at(pos).col->widget->tabHeading() ||
                    item.data().toInt() != specs.at(pos).visibleIndex) {
                    differenceFound = true;
                    break;
                }
            }
            if (!differenceFound) return;
        }
        tabs->clearItems();
        for (TabSpec const &ts : specs) {
            auto *tabItem = new TabItem(ts.col->widget->tabHeading(), ts.visibleIndex);
            tabItem->setShortcutKey(ts.col->widget->tabShortcut());
            tabs->items() << tabItem;
        }
    }
    void updateVisibleTabsAndLayout() {
        updateVisibleColumnsAndTabHeadings();
        calculateColumnCount();
        updateLayout();
    }
    void appStartupCompleted() {
        blanker->start(0.25);
    }
    void gameReadinessUpdated() {
        self().root().window().glActivate();
        updateVisibleTabsAndLayout();
        if (restoredActiveTab >= 0) {
            currentOffsetTab = restoredOffsetTab;
            setScrollOffset(currentOffsetTab, 0.0);
            tabs->setCurrent(ui::DataPos(restoredActiveTab));
            restoredActiveTab = -1;
            restoredOffsetTab = -1;
        }
    }
    void aboutToLoadGame(Game const &gameBeingLoaded) {
        self().root().clearFocusStack();
        self().root().setFocus(nullptr);
        if (gameBeingLoaded.isNull()) {
            moveOnscreen();
        }
        else {
            TimeSpan span = DISMISS_SPAN;
            auto &win = self().root().window().as<ClientWindow>();
            if (win.isGameMinimized()) {
                win.busy().clearTransitionFrameToBlack();
                span = 1.0;
            }
            moveOffscreen();
        }
    }
    void currentGameChanged(Game const &newGame) {
        if (!newGame.isNull()) {
            ClientWindow::main().fadeContent(ClientWindow::FadeFromBlack, 1.0);
            self().root().clearFocusStack();
            self().root().setFocus(nullptr);
        }
    }
    void variableValueChanged(Variable &, Value const &) {
        updateVisibleTabsAndLayout();
    }
    void moveOffscreen(TimeSpan span = DISMISS_SPAN) {
        self().disable();
        self().setBehavior(DisableEventDispatchToChildren);
        self().root().clearFocusStack();
        self().root().setFocus(nullptr);
        ClientWindow::main().setGameMinimized(false);
        if (fequal(dismissOffset->animation().target(), 0.f)) {
            dismissOffset->set(-self().rule().height(), span);
            dismissing = true;
        }
    }
    void moveOnscreen(TimeSpan span = DISMISS_SPAN) {
        if (!fequal(dismissOffset->animation().target(), 0.f)) {
            self().show();
            dismissOffset->set(0, span);
            self().enable();
            self().setBehavior(DisableEventDispatchToChildren, false);
        }
    }
    void checkDismissHiding() {
        if (dismissing && dismissOffset->animation().done()) {
            self().hide();
            dismissing = false;
        }
    }
    void addColumn(ColumnWidget *col) {
        QObject::connect(col, SIGNAL(mouseActivity(QObject const *)),
                         thisPublic, SLOT(mouseActivityInColumn(QObject const *)));
        col->scrollArea().margins().setTop(tabsBackground->rule().height());
        col->rule()
                .setInput(Rule::Width,  *columnWidth)
                .setInput(Rule::Height, self().rule().height());
        self().add(col);
        Variable *conf = col->configVariable();
        if (conf) {
            conf->audienceForChange() += this;
        }
        allColumns << Column(col, conf);
    }
    void calculateColumnCount() {
        visibleColumnCount = de::min(de::max(dsize(1), dsize(self().rule().width().valuei() /
                                                             rule("home.column.width").valuei())),
                                     tabs->items().size());
    }
    void updateLayout() {
        columns.clear();
        columnWidth->setSource(self().rule().width() / visibleColumnCount);
        SequentialLayout layout(self().rule().left() - *scrollOffset,
                                self().rule().top()  + *dismissOffset,
                                ui::Right);
        for (GuiWidget *widget : self().childWidgets()) {
            if (!widget->behavior().testFlag(Widget::Hidden)) {
                if (ColumnWidget *column = maybeAs<ColumnWidget>(widget)) {
                    layout << *column;
                    columns << column;
                }
            }
        }
        updateHighlightedTab();
        if (!visibleTabRange().contains(int(tabs->current()))) {
            currentOffsetTab = int(tabs->current());
            setScrollOffset(currentOffsetTab, 0.0);
        }
        if (visibleTabRange().end >= columns.size()) {
            currentOffsetTab = columns.size() - int(visibleColumnCount);
            setScrollOffset(currentOffsetTab, 0.0);
        }
    }
    Rangei visibleTabRange() const {
        return Rangei(currentOffsetTab, currentOffsetTab + visibleColumnCount);
    }
    void switchTab(int dir, bool scrollInstantly = false) {
        if (scrollInstantly && dir) {
            Rangei const vis = visibleTabRange();
            tabs->setCurrent(dir < 0? vis.start : vis.end - 1);
        }
        auto pos = tabs->current();
        if (dir < 0 && pos > 0) {
            tabs->setCurrent(pos - 1);
        }
        else if (dir > 0 && pos < tabs->items().size() - 1) {
            tabs->setCurrent(pos + 1);
        }
    }
    /**
     * Scrolls the view so that column @a pos is visible.
     * @param pos   Column/tab index.
     * @param span  Animation duration.
     */
    void scrollToTab(int pos, TimeSpan span) {
        pos = de::clamp(0, pos, columns.size() - 1);
        if (currentOffsetTab + int(visibleColumnCount) > columns.size()) {
            currentOffsetTab = columns.size() - visibleColumnCount;
            span = 0;
        }
        else if (visibleTabRange().contains(pos)) {
            return;
        }
        if (pos < currentOffsetTab) {
            currentOffsetTab = pos;
        }
        else {
            currentOffsetTab = pos - visibleColumnCount + 1;
        }
        setScrollOffset(currentOffsetTab, span);
    }
    void setScrollOffset(int tab, TimeSpan const &span) {
        scrollOffset->set(*columnWidth * tab, span);
    }
    void updateHighlightedTab() {
        if (columns.isEmpty()) return;
        if (tabs->current() >= tabs->items().size()) {
            tabs->setCurrent(tabs->items().size() - 1); // calls back here via observer
            return;
        }
        const int newHighlightPos = tabs->currentItem().data().toInt();
        for (int pos = 0; pos < columns.size(); ++pos) {
            if (pos != newHighlightPos && columns[pos]->isHighlighted()) {
                columns[pos]->setHighlighted(false);
            }
        }
        columns[newHighlightPos]->setHighlighted(true);
        moveShowTimer.start();
    }
    int highlightedTab() const {
        for (int pos = 0; pos < columns.size(); ++pos) {
            if (columns[pos]->isHighlighted()) {
                return pos;
            }
        }
        return -1;
    }
};
HomeWidget::HomeWidget()
    : GuiWidget("home")
    , d(new Impl(this)) {
    setAttribute(ManualOpacity);
    ColumnWidget *column;
    column = new NoGamesColumnWidget();
    d->addColumn(column);
    column = new GameColumnWidget("DOOM", d->savedItems);
    d->addColumn(column);
    column = new GameColumnWidget("Heretic", d->savedItems);
    d->addColumn(column);
    column = new GameColumnWidget("Hexen", d->savedItems);
    d->addColumn(column);
    column = new GameColumnWidget("", d->savedItems);
    d->addColumn(column);
    column = new MultiplayerColumnWidget();
    d->addColumn(column); {
        auto *packagesColumn = new PackagesColumnWidget;
        d->addColumn(packagesColumn);
        connect(packagesColumn,
                &PackagesColumnWidget::availablePackageCountChanged,
                [this] (int count) {
            d->havePackages = count > 0;
            d->updateVisibleTabsAndLayout();
        });
    }
    d->updateVisibleColumnsAndTabHeadings();
    d->tabs->setCurrent(0);
    add(d->tabsBackground);
    add(d->tabs);
    add(d->moveLeft);
    add(d->moveRight);
    d->moveLeft ->rule().setInput(Rule::Left,  rule().left());
    d->moveRight->rule().setInput(Rule::Right, rule().right());
    add(d->taskBarHintButton);
    d->blanker.reset(new FadeToBlackWidget);
    d->blanker->rule().setRect(rule());
    d->blanker->initFadeFromBlack(0.75);
    add(d->blanker);
    Rule const &gap = rule("gap");
    d->tabsBackground->rule()
            .setInput(Rule::Left,   rule().left())
            .setInput(Rule::Top,    rule().top() + *d->dismissOffset)
            .setInput(Rule::Width,  rule().width())
            .setInput(Rule::Height, d->tabs->rule().height() + gap*2);
    d->tabs->rule()
            .setInput(Rule::Width,  rule().width())
            .setInput(Rule::Top,    rule().top() + gap + *d->dismissOffset)
            .setInput(Rule::Left,   rule().left());
    d->updateLayout();
    connect(d->tabs, SIGNAL(currentTabChanged()), this, SLOT(tabChanged()));
}
void HomeWidget::viewResized() {
    d->calculateColumnCount();
    d->updateLayout();
}
bool HomeWidget::handleEvent(Event const &event) {
    if (event.isKeyDown()) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.qtKey() == Qt::Key_Left) {
            d->switchTab(-1);
            return true;
        }
        else if (key.qtKey() == Qt::Key_Right) {
            d->switchTab(+1);
            return true;
        }
        else if (key.qtKey() == Qt::Key_Home) {
            d->tabs->setCurrent(0);
            return true;
        }
        else if (key.qtKey() == Qt::Key_End) {
            d->tabs->setCurrent(d->tabs->items().size() - 1);
            return true;
        }
    }
    return false;
}
void HomeWidget::moveOnscreen(TimeSpan span) {
    d->moveOnscreen(span);
}
void HomeWidget::moveOffscreen(TimeSpan span) {
    d->moveOffscreen(span);
}
void HomeWidget::update() {
    GuiWidget::update();
    d->checkDismissHiding();
    if (d->blanker) {
        d->blanker->disposeIfDone();
    }
}
PopupWidget *HomeWidget::makeSettingsPopup() {
    PopupMenuWidget *menu = new PopupMenuWidget;
    menu->items()
            << new ui::Item(ui::Item::Separator, tr("Library"))
            << new ui::VariableToggleItem(tr("Show Descriptions"), App::config("home.showColumnDescription"))
            << new ui::VariableToggleItem(tr("Show Unplayable"),   App::config("home.showUnplayableGames"))
            << new ui::Item(ui::Item::Separator)
            << new ui::Item(ui::Item::Separator, tr("Columns"))
            << new ui::VariableToggleItem(tr("Doom"),            App::config("home.columns.doom"))
            << new ui::VariableToggleItem(tr("Heretic"),         App::config("home.columns.heretic"))
            << new ui::VariableToggleItem(tr("Hexen"),           App::config("home.columns.hexen"))
            << new ui::VariableToggleItem(tr("Other Games"),     App::config("home.columns.otherGames"))
            << new ui::VariableToggleItem(tr("Multiplayer"),     App::config("home.columns.multiplayer"));
    return menu;
}
bool HomeWidget::dispatchEvent(Event const &event, bool (Widget::*memberFunc)(const Event &)) {
    if (event.type() == Event::MouseWheel) {
        if (event.isMouse()) {
#if 0
            MouseEvent const &mouse = event.as<MouseEvent>();
            if (event.type() == Event::MouseWheel &&
               mouse.wheelMotion() == MouseEvent::Step) {
                if (abs(mouse.wheel().x) > abs(mouse.wheel().y)) {
                    d->switchTab(-de::sign(mouse.wheel().x), true /*instant*/);
                }
            }
#endif
        }
    }
    return GuiWidget::dispatchEvent(event, memberFunc);
}
void HomeWidget::tabChanged() {
    d->scrollToTab(d->tabs->currentItem().data().toInt(), SCROLL_SPAN);
    d->updateHighlightedTab();
}
void HomeWidget::mouseActivityInColumn(QObject const *columnWidget) {
    for (int i = 0; i < d->columns.size(); ++i) {
        if (d->columns.at(i) == columnWidget) {
            d->tabs->setCurrent(i);
        }
    }
}
void HomeWidget::updateStyle() {
    GuiWidget::updateStyle();
    viewResized();
}
void HomeWidget::operator >> (PersistentState &toState) const {
    Record &rec = toState.objectNamespace();
    rec.set(name().concatenateMember("firstTab"),  d->currentOffsetTab);
    rec.set(name().concatenateMember("activeTab"), d->highlightedTab());
}
void HomeWidget::operator << (PersistentState const &fromState) {
    Record const &rec = fromState.objectNamespace();
    d->restoredOffsetTab = rec.geti(name().concatenateMember("firstTab"), -1);
    d->restoredActiveTab = rec.geti(name().concatenateMember("activeTab"), -1);
}
DENG_GUI_PIMPL(MultiplayerColumnWidget)
, DENG2_OBSERVES(ui::Data, Addition)
, DENG2_OBSERVES(ui::Data, Removal) {
    MultiplayerServerMenuWidget *menu;
    LabelWidget *noServers;
    Impl(Public *i) : Base(i) {
        ScrollAreaWidget &area = self().scrollArea();
        area.add(menu = new MultiplayerServerMenuWidget);
        self().header().menuButton().setPopup([] (PopupButtonWidget const &) -> PopupWidget * {
            auto *menu = new PopupMenuWidget;
            menu->items()
                    << new ui::ActionItem(tr("Connect to Server..."),
                                          new SignalAction(&ClientWindow::main().taskBar(),
                                                           SLOT(connectToServerManually())))
                    << new ui::ActionItem(tr("Refresh List"), new CallbackAction([] () {
                            ServerLink::get().discoverUsingMaster(); }));
            return menu;
        }, ui::Down);
        menu->rule()
                .setInput(Rule::Width, area.contentRule().width())
                .setInput(Rule::Left,  area.contentRule().left())
                .setInput(Rule::Top,   self().header().rule().bottom());
        noServers = new LabelWidget;
        noServers->setText(tr("No Servers Found"));
        style().as<ClientStyle>().emptyMenuLabelStylist().applyStyle(*noServers);
        noServers->rule().setRect(self().rule());
        self().add(noServers);
        menu->items().audienceForAddition() += this;
        menu->items().audienceForRemoval() += this;
    }
    void dataItemAdded(ui::DataPos, ui::Item const &) override {
        noServers->hide();
    }
    void dataItemRemoved(ui::DataPos, ui::Item &) override {
        if (menu->items().isEmpty()) {
            noServers->show();
        }
    }
};
MultiplayerColumnWidget::MultiplayerColumnWidget()
    : ColumnWidget("multiplayer-column")
    , d(new Impl(this)) {
    scrollArea().setContentSize(maximumContentWidth(),
                                header().rule().height() +
                                rule("gap") +
                                d->menu->rule().height());
    header().title().setText(_E(s)_E(C) "dengine.net\n" _E(.)_E(.)_E(w) + tr("Multiplayer Games"));
    header().info().setText(tr("Multiplayer servers are discovered via the dengine.net "
                               "master server and by broadcasting on the local network."));
}
String MultiplayerColumnWidget::tabHeading() const {
    return tr("Multiplayer");
}
String MultiplayerColumnWidget::configVariableName() const {
    return "home.columns.multiplayer";
}
void MultiplayerColumnWidget::setHighlighted(bool highlighted) {
    ColumnWidget::setHighlighted(highlighted);
    if (highlighted) {
        d->menu->restorePreviousSelection();
    }
    else {
        root().setFocus(nullptr);
        d->menu->unselectAll();
    }
}
DENG_GUI_PIMPL(MultiplayerPanelButtonWidget)
, DENG2_OBSERVES(Games, Readiness)
, public AsyncScope {
    shell::ServerInfo serverInfo;
    ButtonWidget *joinButton;
    String gameConfig;
    LabelWidget *info;
    PopupButtonWidget *extra;
    PopupMenuWidget *extraMenu;
    res::LumpCatalog catalog;
    Impl(Public *i) : Base(i) {
        DoomsdayApp::games().audienceForReadiness() += this;
        joinButton = new ButtonWidget;
        joinButton->setAttribute(AutomaticOpacity);
        joinButton->disable();
        joinButton->setText(tr("Join"));
        joinButton->useInfoStyle();
        joinButton->setSizePolicy(ui::Expand, ui::Expand);
        joinButton->setActionFn([this] () { joinButtonPressed(); });
        self().addButton(joinButton);
        info = new LabelWidget;
        info->setSizePolicy(ui::Fixed, ui::Expand);
        info->setAlignment(ui::AlignLeft);
        info->setTextLineAlignment(ui::AlignLeft);
        info->rule().setInput(Rule::Width, self().rule().width());
        info->margins().setLeft(self().icon().rule().width());
        extra = new PopupButtonWidget;
        extra->hide();
        extra->setSizePolicy(ui::Expand, ui::Expand);
        extra->setText("...");
        extra->setFont("small");
        extra->margins().setTopBottom(RuleBank::UNIT);
        extra->rule()
                .setInput(Rule::Bottom, info->rule().bottom() - info->margins().bottom())
                .setMidAnchorX(info->rule().left() + self().icon().rule().width()/2);
        info->add(extra);
        self().panel().setContent(info);
        self().panel().open();
        extra->setPopup([this] (PopupButtonWidget const &) -> PopupWidget * {
            auto *dlg = new ServerInfoDialog(serverInfo);
            QObject::connect(dlg, SIGNAL(joinGame()), thisPublic, SLOT(joinGame()));
            return dlg;
        }, ui::Right);
    }
    void joinButtonPressed() const {
        self().root().setFocus(nullptr);
        DENG2_FOR_PUBLIC_AUDIENCE2(AboutToJoin, i) {
            i->aboutToJoinMultiplayerGame(serverInfo);
        }
        ClientApp::serverLink().connectToServerAndChangeGameAsync(serverInfo);
    }
    bool hasConfig(String const &token) const {
        return gameConfig.containsWord(token);
    }
    void gameReadinessUpdated() {
        catalog.clear();
        self().updateContent(serverInfo);
    }
    DENG2_PIMPL_AUDIENCE(AboutToJoin)
};
DENG2_AUDIENCE_METHOD(MultiplayerPanelButtonWidget, AboutToJoin)
MultiplayerPanelButtonWidget::MultiplayerPanelButtonWidget()
    : d(new Impl(this))
{}
ButtonWidget &MultiplayerPanelButtonWidget::joinButton() {
    return *d->joinButton;
}
void MultiplayerPanelButtonWidget::setSelected(bool selected) {
    PanelButtonWidget::setSelected(selected);
    d->extra->show(selected);
}
void MultiplayerPanelButtonWidget::itemRightClicked() {
    PanelButtonWidget::itemRightClicked();
    d->extra->trigger();
}
void MultiplayerPanelButtonWidget::updateContent(shell::ServerInfo const &info) {
    d->serverInfo = info;
    d->gameConfig = info.gameConfig();
#define DIM_MDASH   _E(C) DENG2_CHAR_MDASH _E(.)
    String meta;
    int const playerCount = info.playerCount();
    if (playerCount > 0) {
        meta = String("%1 player%2 " DIM_MDASH " ")
                .arg(playerCount)
                .arg(DENG2_PLURAL_S(playerCount));
    }
    meta += String("%1").arg(tr(d->hasConfig("coop")? "Co-op" :
                                d->hasConfig("dm2")?  "Deathmatch II" :
                                                      "Deathmatch"));
    if (ClientApp::serverLink().isServerOnLocalNetwork(info.address())) {
        meta = "LAN " DIM_MDASH " " + meta;
    }
    label().setText(String(_E(b) "%1\n" _E(l) "%2")
                    .arg(info.name())
                    .arg(meta));
    String infoText = String(info.map()) + " " DIM_MDASH " ";
    if (DoomsdayApp::games().contains(info.gameId())) {
        auto const &game = DoomsdayApp::games()[info.gameId()];
        infoText += game.title();
        d->joinButton->enable();
        if (d->catalog.setPackages(game.requiredPackages())) {
            res::LumpCatalog catalog{d->catalog};
            *d += async([&game, catalog]() { return IdTech1Image::makeGameLogo(game, catalog); },
                        [this](const Image &logo) { icon().setImage(logo); });
        }
    }
    else {
        infoText += tr("Unknown game");
        d->joinButton->disable();
        icon().setImage(nullptr);
    }
    if (!info.flags().testFlag(shell::ServerInfo::AllowJoin)) {
        d->joinButton->disable();
    }
    infoText += "\n" _E(C) + String(info.description()) + _E(.);
    int const localCount = Game::localMultiplayerPackages(info.gameId()).size();
    if (localCount) {
        infoText += "\n" _E(D)_E(b) + String("%1 local package%2").arg(localCount)
                .arg(DENG2_PLURAL_S(localCount));
    }
    d->info->setFont("small");
    d->info->setText(infoText);
}
void MultiplayerPanelButtonWidget::joinGame() {
    d->joinButtonPressed();
}
NoGamesColumnWidget::NoGamesColumnWidget()
    : ColumnWidget("nogames-column") {
    header().hide();
    LabelWidget *notice = LabelWidget::newWithText(
                _E(b) + tr("No playable games were found.\n") + _E(.) +
                tr("Please select the folder where you have one or more game WAD files."),
                this);
    notice->setMaximumTextWidth(maximumContentWidth());
    notice->setTextColor("text");
    notice->setSizePolicy(ui::Expand, ui::Expand);
    notice->rule()
            .setMidAnchorX(rule().midX())
            .setInput(Rule::Bottom, rule().midY());
    ButtonWidget *chooseIwad = new ButtonWidget;
    chooseIwad->setText(tr("Select WAD Folder..."));
    chooseIwad->setSizePolicy(ui::Expand, ui::Expand);
    chooseIwad->rule()
            .setMidAnchorX(rule().midX())
            .setInput(Rule::Top, notice->rule().bottom());
    chooseIwad->setAction(new SignalAction(this, SLOT(browseForDataFiles())));
    add(chooseIwad);
}
String NoGamesColumnWidget::tabHeading() const {
    return tr("Data Files?");
}
void NoGamesColumnWidget::browseForDataFiles() {
    bool reload = false;
#if !defined (DENG_MOBILE)
    ClientApp::app().beginNativeUIMode();
    const auto folders = Config::get().getStringList("resource.packageFolder");
    String lastDir;
    if (!folders.isEmpty()) {
        lastDir = folders.back();
    }
    QFileDialog dlg(nullptr, "Select IWAD Folder", lastDir);
    dlg.setFileMode(QFileDialog::Directory);
    dlg.setReadOnly(true);
    dlg.setLabelText(QFileDialog::Accept, tr("Select"));
    if (dlg.exec()) {
        Variable &      var = Config::get("resource.packageFolder");
        const TextValue selDir{dlg.selectedFiles().at(0)};
        if (is<ArrayValue>(var.value())) {
            auto &array = var.value<ArrayValue>();
            int found = array.indexOf(TextValue(selDir));
            if (found >= 0) {
                array.remove(found);
            }
            array.add(selDir);
        }
        else {
            var.set(selDir);
        }
        reload = true;
    }
    ClientApp::app().endNativeUIMode();
#endif
    if (reload) {
        ClientWindow::main().console().closeLogAndUnfocusCommandLine();
        DoomsdayApp::app().initPackageFolders();
    }
}
DENG_GUI_PIMPL(PackagesColumnWidget)
, DENG2_OBSERVES(DoomsdayApp, GameChange) {
    PackagesWidget *packages;
    LabelWidget *countLabel;
    ButtonWidget *folderOptionsButton;
    ui::ListData actions;
    LoopCallback mainCall;
    int totalPackageCount = 0;
    Impl(Public *i) : Base(i) {
        DoomsdayApp::app().audienceForGameChange() += this;
        actions << new ui::SubwidgetItem(tr("..."), ui::Left, [this]() -> PopupWidget * {
            return new PackageInfoDialog(packages->actionPackage(),
                                         PackageInfoDialog::EnableActions);
        });
        countLabel = new LabelWidget;
        ScrollAreaWidget &area = self().scrollArea();
        area.add(packages = new PackagesWidget(PackagesWidget::PopulationEnabled, "home-packages"));
        packages->setActionItems(actions);
        packages->setRightClickToOpenContextMenu(true);
        packages->margins().setLeft("").setRight("");
        packages->rule()
                .setInput(Rule::Width, area.contentRule().width())
                .setInput(Rule::Top,   self().header().rule().bottom() + rule("gap"))
                .setInput(Rule::Left,  area.contentRule().left());
        QObject::connect(packages, &PackagesWidget::itemCountChanged,
                         [this] (duint shown, duint total) {
            if (shown == total) {
                countLabel->setText(tr("%1 available").arg(total));
            }
            else {
                countLabel->setText(tr("%1 shown out of %2 available").arg(shown).arg(total));
            }
            totalPackageCount = total;
            emit self().availablePackageCountChanged(total);
        });
        area.add(folderOptionsButton = new ButtonWidget);
        folderOptionsButton->setStyleImage("gear", "default");
        folderOptionsButton->setText(tr("Configure Data Files"));
        folderOptionsButton->setTextAlignment(ui::AlignRight);
        folderOptionsButton->setSizePolicy(ui::Fixed, ui::Expand);
        folderOptionsButton->rule()
                .setInput(Rule::Width, area.contentRule().width())
                .setInput(Rule::Left,  area.contentRule().left())
                .setInput(Rule::Top,   packages->rule().bottom());
        folderOptionsButton->setAction(new CallbackAction([this] () {
            auto *dlg = new DataFileSettingsDialog;
            dlg->setDeleteAfterDismissed(true);
            dlg->setAnchorAndOpeningDirection(folderOptionsButton->rule(), ui::Left);
            self().root().addOnTop(dlg);
            dlg->open();
        }));
        self().header().menuButton().setPopup([this] (PopupButtonWidget const &) -> PopupWidget * {
            auto *menu = new PopupMenuWidget;
            menu->items()
                    << new ui::SubwidgetItem(style().images().image("gear"),
                                             ui::Item::ShownAsButton | ui::Item::ClosesParentPopup,
                                             tr("Settings"), ui::Left, makePopup<DataFileSettingsDialog>)
                    << new ui::Item(ui::Item::Separator)
                    << new ui::ActionItem("Show Recognized IWADs",
                                          new CallbackAction([this]() { packages->searchTermsEditor().setText("gamedata"); }))
                    << new ui::ActionItem("Show Box Contents",
                                          new CallbackAction([this]() { packages->searchTermsEditor().setText("hidden"); }))
                    << new ui::ActionItem("Show Core Packages",
                                          new CallbackAction([this]() { packages->searchTermsEditor().setText("core"); }))
                    << new ui::Item(ui::Item::Separator)
                    << new ui::ActionItem(tr("Refresh List"),
                                          new CallbackAction([]() { FS::get().refreshAsync(); }));
                return menu;
        }, ui::Down);
    }
    void currentGameChanged(Game const &game) override {
        folderOptionsButton->show(game.isNull());
    }
    void openRepositoryBrowser() {
        auto *dlg = new RepositoryBrowserDialog;
        dlg->setDeleteAfterDismissed(true);
        dlg->exec(root());
    }
};
PackagesColumnWidget::PackagesColumnWidget()
    : ColumnWidget("packages-column")
    , d(new Impl(this)) {
    header().title().setText(_E(s) "\n" _E(.) + tr("Mods"));
    header().info().setText(tr("Browse available mods/add-ons and install new ones."));
    header().infoPanel().close(0);
    d->countLabel->setFont("small");
    d->countLabel->setSizePolicy(ui::Expand, ui::Fixed);
    d->countLabel->rule()
            .setInput(Rule::Left,   header().menuButton().rule().right())
            .setInput(Rule::Height, header().menuButton().rule().height())
            .setInput(Rule::Top,    header().menuButton().rule().top());
    header().add(d->countLabel);
    scrollArea().setContentSize(maximumContentWidth(),
                                header().rule().height() +
                                rule("gap") +
                                d->packages->rule().height() +
                                d->folderOptionsButton->rule().height()*2);
    d->packages->setFilterEditorMinimumY(scrollArea().margins().top());
    d->packages->progress().rule().setRect(scrollArea().rule());
}
int PackagesColumnWidget::availablePackageCount() const {
    return d->totalPackageCount;
}
String PackagesColumnWidget::tabHeading() const {
    return tr("Mods");
}
String PackagesColumnWidget::tabShortcut() const {
    return QStringLiteral("s");
}
void PackagesColumnWidget::setHighlighted(bool highlighted) {
    ColumnWidget::setHighlighted(highlighted);
    if (highlighted) {
        root().setFocus(&d->packages->searchTermsEditor());
    }
}
DENG_GUI_PIMPL(SaveListWidget)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate) {
    /**
     * Handles mouse button doubleclicks on the save items.
     */
    struct ClickHandler : public GuiWidget::IEventHandler {
        SaveListWidget::Impl *d;
        ClickHandler(SaveListWidget::Impl *d) : d(d) {}
        bool handleEvent(GuiWidget &button, Event const &event) {
            if (event.type() == Event::MouseButton) {
                MouseEvent const &mouse = event.as<MouseEvent>();
                if (mouse.button() == MouseEvent::Right) {
                    switch (button.handleMouseClick(event, MouseEvent::Right)) {
                    case MouseClickStarted:
                        return true;
                    case MouseClickAborted:
                        return true;
                    case MouseClickFinished:
                        saveRightClicked(button);
                        return true;
                    default:
                        return false; // Ignore.
                    }
                }
                /*if (button.hitTest(mouse) && mouse.state() == MouseEvent::DoubleClick)
                {
                    emit d->self().doubleClicked(d->self().items().find(
                            *d->self().organizer().findItemForWidget(button)));
                    return true;
                }*/
            }
            return false;
        }
        void saveRightClicked(GuiWidget &saveButton) {
            auto const &saveItem = d->self().organizer()
                    .findItemForWidget(saveButton)->as<SaveListData::SaveItem>();
            if (GameStateFolder const *saved = FS::tryLocate<GameStateFolder>(saveItem.savePath())) {
                auto *docPop = new DocumentPopupWidget;
                docPop->setDeleteAfterDismissed(true);
                docPop->setCloseButtonVisible(true);
                docPop->setAnchorAndOpeningDirection(saveButton.rule(), ui::Right);
                docPop->document().setText(saved->metadata().asStyledText());
                saveButton.add(docPop);
                docPop->open();
            }
        }
    };
    GamePanelButtonWidget &owner;
    ui::DataPos selected = ui::Data::InvalidPos;
    Impl(Public *i, GamePanelButtonWidget &owner) : Base(i), owner(owner) {
        self().organizer().audienceForWidgetUpdate() += this;
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) {
        auto &button = widget.as<ButtonWidget>();
        button.setTextAlignment(ui::AlignRight);
        button.setAlignment(ui::AlignLeft);
        button.setTextLineAlignment(ui::AlignLeft);
        button.setSizePolicy(ui::Filled, ui::Expand);
        button.setText(item.label());
        button.margins().set("dialog.gap");
        button.set(Background(Vector4f()));
        button.setActionFn([this, &button] () {
            toggleSelectedItem(button);
            emit owner.mouseActivity();
        });
        button.addEventHandler(new ClickHandler(this));
        auto const &saveItem = item.as<SaveListData::SaveItem>();
        button.setImage(style().images().image(Game::logoImageForId(saveItem.gameId())));
        button.setOverrideImageSize(style().fonts().font("default").height() * 1.4f);
    }
    void toggleSelectedItem(ButtonWidget &button) {
        auto const buttonItemPos = self().items().find(*self().organizer().findItemForWidget(button));
        if (selected == buttonItemPos) {
            selected = ui::Data::InvalidPos;
            updateItemHighlights(nullptr);
        }
        else {
            selected = buttonItemPos;
            updateItemHighlights(&button);
        }
        emit self().selectionChanged(selected);
        self().root().setFocus(&self());
    }
    void updateItemHighlights(ButtonWidget *selectedButton) {
        for (auto *w : self().childWidgets()) {
            if (auto *bw = maybeAs<ButtonWidget>(w)) {
                if (selectedButton == bw) {
                    bw->useInfoStyle();
                    bw->set(Background(style().colors().colorf("inverted.background")));
                }
                else {
                    bw->useNormalStyle();
                    bw->set(Background());
                }
            }
        }
    }
};
SaveListWidget::SaveListWidget(GamePanelButtonWidget &owner)
    : d(new Impl(this, owner)) {
    setGridSize(1, ui::Filled, 0, ui::Expand);
    enableScrolling(false);
    enablePageKeys(false);
}
ui::DataPos SaveListWidget::selectedPos() const {
    return d->selected;
}
void SaveListWidget::setSelectedPos(ui::DataPos pos) {
    if (d->selected != pos) {
        d->selected = pos;
        d->updateItemHighlights(&itemWidget<ButtonWidget>(items().at(pos)));
        emit selectionChanged(d->selected);
    }
}
void SaveListWidget::clearSelection() {
    if (d->selected != ui::Data::InvalidPos) {
        d->selected = ui::Data::InvalidPos;
        d->updateItemHighlights(nullptr);
        emit selectionChanged(d->selected);
    }
}
