﻿/** @file finale.cpp  InFine animation system, Finale script.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include <de/LogBuffer>
#include "ui/infine/finale.h"
#include "BindContext"
#include "ui/infine/finaleinterpreter.h"
#include "network/net_main.h"
#ifdef __SERVER__
#  include "server/sv_infine.h"
#endif
/** @file finaleanimwidget.cpp  InFine animation system, FinaleAnimWidget.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <de/vector1.h>
#include <doomsday/res/Textures>
#include "ui/infine/finaleanimwidget.h"
#include "dd_main.h"   // App_Resources()
#include "api_sound.h"
#ifdef __CLIENT__
#  include "gl/gl_main.h"
#  include "gl/gl_texmanager.h" // GL_PrepareRawTexture()
#  include "render/r_draw.h"    // Rend_PatchTextureSpec()
#  include "render/rend_main.h" // filterUI
#  include "MaterialAnimator"
#  include <de/GLInfo>
#endif
/** @file finaleinterpreter.cpp  InFine animation system Finale script interpreter.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "ui/infine/finaleinterpreter.h"
#include <QList>
#include <de/memory.h>
#include <de/timer.h>
#include <de/LogBuffer>
#include <doomsday/doomsdayapp.h>
#include <doomsday/console/cmd.h>
#include <doomsday/console/exec.h>
#include <doomsday/filesys/fs_main.h>
#include <doomsday/world/Materials>
#include <doomsday/Game>
#include "dd_def.h"
#include "def_main.h"  // ::defs
#include "api_material.h"
#include "api_render.h"
#include "api_resource.h"
#include "network/net_main.h"
#include "ui/infine/finalewidget.h"
#include "ui/infine/finaleanimwidget.h"
#include "ui/infine/finalepagewidget.h"
#include "ui/infine/finaletextwidget.h"
#ifdef __CLIENT__
#  include "api_fontrender.h"
#  include "client/cl_infine.h"
#  include "gl/gl_main.h"
#  include "gl/gl_texmanager.h"
#  include "gl/texturecontent.h"
#  include "gl/sys_opengl.h" // TODO: get rid of this
#endif
#ifdef __SERVER__
#  include "server/sv_infine.h"
#endif
#define MAX_TOKEN_LENGTH        (8192)
#define FRACSECS_TO_TICKS(sec)  (int(sec * TICSPERSEC + 0.5))
/** @file finalepagewidget.cpp  InFine animation system, FinalePageWidget.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/infine/finalepagewidget.h"
#include <doomsday/world/Material>
#include <de/vector1.h>
#include "dd_main.h" // App_Resources()
#ifdef __CLIENT__
#  include <de/GLInfo>
#  include <de/GLState>
#  include "gl/gl_draw.h"
#  include "gl/gl_main.h"
#  include "render/rend_main.h" // renderWireframe
#  include "MaterialAnimator"
#endif
/** @file finaletextwidget.cpp  InFine animation system, FinaleTextWidget.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "ui/infine/finaletextwidget.h"
#include <cstring> // memcpy, memmove
#include <de/concurrency.h>
#include <de/memoryzone.h>
#include <de/timer.h>
#include "api_fontrender.h"
#include "ui/infine/finalepagewidget.h"
#ifdef __CLIENT__
#  include "gl/gl_main.h"
#  include <de/GLInfo>
#endif
/** @file finalewidget.cpp  InFine animation system, FinaleWidget.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/infine/finalewidget.h"
#include "ui/infine/finalepagewidget.h"
/** @file infinesystem.cpp  Interactive animation sequence system.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_INFINE
#include <QtAlgorithms>
#include <de/Log>
#include <doomsday/console/var.h>
#include "de_base.h"
#include "ui/infine/infinesystem.h"
#include "BindContext"
#include "ui/infine/finale.h"
#include "ui/infine/finaleinterpreter.h"
#ifdef __CLIENT__
#  include "ui/inputsystem.h"
#  include "clientapp.h"
#endif
using namespace de;
DENG2_PIMPL(Finale) {
    bool active;
    int flags;  ///< @ref finaleFlags
    finaleid_t id;
    FinaleInterpreter interpreter;
    Impl(Public *i, int flags, finaleid_t id)
        : Base(i)
        , active     (false)
        , flags      (flags)
        , id         (id)
        , interpreter(id)
    {}
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) i->finaleBeingDeleted(self());
    }
    void loadScript(String const &script) {
        if (script.isEmpty()) return;
        LOGDEV_SCR_MSG("Begin Finale - id:%i '%.30s'") << id << script;
        Block const scriptAsUtf8 = script.toUtf8();
        interpreter.loadScript(scriptAsUtf8.constData());
#ifdef __SERVER__
        if (!(flags & FF_LOCAL) && ::isServer) {
            Sv_Finale(id, FINF_BEGIN | FINF_SCRIPT, scriptAsUtf8.constData());
        }
#endif
        active = true;
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
};
DENG2_AUDIENCE_METHOD(Finale, Deletion)
Finale::Finale(int flags, finaleid_t id, String const &script)
    : d(new Impl(this, flags, id)) {
    d->loadScript(script);
}
int Finale::flags() const {
    return d->flags;
}
finaleid_t Finale::id() const {
    return d->id;
}
bool Finale::isActive() const {
    return d->active;
}
bool Finale::isSuspended() const {
    return d->interpreter.isSuspended();
}
void Finale::resume() {
    d->active = true;
    d->interpreter.resume();
}
void Finale::suspend() {
    d->active = false;
    d->interpreter.suspend();
}
bool Finale::terminate() {
    if (!d->active) return false;
    LOGDEV_SCR_VERBOSE("Terminating finaleid %i") << d->id;
    d->active = false;
    d->interpreter.terminate();
    return true;
}
bool Finale::runTicks(timespan_t timeDelta) {
    if (d->interpreter.runTicks(timeDelta, d->active && DD_IsSharpTick())) {
        terminate();
        return true;
    }
    return false;
}
int Finale::handleEvent(ddevent_t const &ev) {
    if (!d->active) return false;
    return d->interpreter.handleEvent(ev);
}
bool Finale::requestSkip() {
    if (!d->active) return false;
    return d->interpreter.skip();
}
bool Finale::isMenuTrigger() const {
    if (!d->active) return false;
    LOG_SCR_XVERBOSE("IsMenuTrigger: %i", d->interpreter.isMenuTrigger());
    return d->interpreter.isMenuTrigger();
}
FinaleInterpreter const &Finale::interpreter() const {
    return d->interpreter;
}
FinaleAnimWidget::Frame::Frame()
    : tics (0)
    , type (PFT_MATERIAL)
    , sound(0) {
    de::zap(flags);
    de::zap(texRef);
}
FinaleAnimWidget::Frame::~Frame() {
#ifdef __CLIENT__
    if (type == PFT_XIMAGE) {
        DGL_DeleteTextures(1, (DGLuint *)&texRef.tex);
    }
#endif
}
DENG2_PIMPL_NOREF(FinaleAnimWidget) {
    bool animComplete = true;
    bool animLooping  = false;  ///< @c true= loop back to the start when the end is reached.
    int tics          = 0;
    int curFrame      = 0;
    Frames frames;
    animatorvector4_t color;
    animatorvector4_t otherColor;
    animatorvector4_t edgeColor;
    animatorvector4_t otherEdgeColor;
    Impl() {
        AnimatorVector4_Init(color,          1, 1, 1, 1);
        AnimatorVector4_Init(otherColor,     0, 0, 0, 0);
        AnimatorVector4_Init(edgeColor,      0, 0, 0, 0);
        AnimatorVector4_Init(otherEdgeColor, 0, 0, 0, 0);
    }
    static Frame *makeFrame(Frame::Type type, int tics, void *texRef, short sound, bool flagFlipH) {
        Frame *f = new Frame;
        f->flags.flip = flagFlipH;
        f->type  = type;
        f->tics  = tics;
        f->sound = sound;
        switch (f->type) {
        case Frame::PFT_MATERIAL:  f->texRef.material = ((world::Material *)texRef); break;
        case Frame::PFT_PATCH:     f->texRef.patch    = *((patchid_t *)texRef); break;
        case Frame::PFT_RAW:       f->texRef.lumpNum  = *((lumpnum_t *)texRef); break;
        case Frame::PFT_XIMAGE:    f->texRef.tex      = *((DGLuint *)texRef);   break;
        default: throw Error("FinaleAnimWidget::makeFrame", "Unknown frame type #" + String::number(type));
        }
        return f;
    }
};
FinaleAnimWidget::FinaleAnimWidget(String const &name)
    : FinaleWidget(name)
    , d(new Impl)
{}
FinaleAnimWidget::~FinaleAnimWidget() {
    clearAllFrames();
}
bool FinaleAnimWidget::animationComplete() const {
    return d->animComplete;
}
FinaleAnimWidget &FinaleAnimWidget::setLooping(bool yes) {
    d->animLooping = yes;
    return *this;
}
#ifdef __CLIENT__
static void useColor(animator_t const *color, int components) {
    if (components == 3) {
        DGL_Color3f(color[0].value, color[1].value, color[2].value);
    }
    else if (components == 4) {
        DGL_Color4f(color[0].value, color[1].value, color[2].value, color[3].value);
    }
}
static int buildGeometry(float const /*dimensions*/[3], dd_bool flipTextureS,
    Vector4f const &bottomColor, Vector4f const &topColor, Vector3f **posCoords,
    Vector4f **colorCoords, Vector2f **texCoords) {
    static Vector3f posCoordBuf[4];
    static Vector4f colorCoordBuf[4];
    static Vector2f texCoordBuf[4];
    posCoordBuf[0] = Vector3f(0, 0, 0);
    posCoordBuf[1] = Vector3f(1, 0, 0);
    posCoordBuf[2] = Vector3f(0, 1, 0);
    posCoordBuf[3] = Vector3f(1, 1, 0);
    texCoordBuf[0] = Vector2f((flipTextureS? 1:0), 0);
    texCoordBuf[1] = Vector2f((flipTextureS? 0:1), 0);
    texCoordBuf[2] = Vector2f((flipTextureS? 1:0), 1);
    texCoordBuf[3] = Vector2f((flipTextureS? 0:1), 1);
    colorCoordBuf[0] = bottomColor;
    colorCoordBuf[1] = bottomColor;
    colorCoordBuf[2] = topColor;
    colorCoordBuf[3] = topColor;
    *posCoords   = posCoordBuf;
    *texCoords   = texCoordBuf;
    *colorCoords = colorCoordBuf;
    return 4;
}
static void drawGeometry(int numVerts, Vector3f const *posCoords,
    Vector4f const *colorCoords, Vector2f const *texCoords) {
    DGL_Begin(DGL_TRIANGLE_STRIP);
    Vector3f const *posIt   = posCoords;
    Vector4f const *colorIt = colorCoords;
    Vector2f const *texIt   = texCoords;
    for (int i = 0; i < numVerts; ++i, posIt++, colorIt++, texIt++) {
        if (texCoords)
            DGL_TexCoord2f(0, texIt->x, texIt->y);
        if (colorCoords)
            DGL_Color4f(colorIt->x, colorIt->y, colorIt->z, colorIt->w);
        DGL_Vertex3f(posIt->x, posIt->y, posIt->z);
    }
    DGL_End();
}
static inline MaterialVariantSpec const &uiMaterialSpec_FinaleAnim() {
    return App_Resources().materialSpec(UiContext, 0, 0, 0, 0,
                                             GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE,
                                             0, -3, 0, false, false, false, false);
}
static void drawPicFrame(FinaleAnimWidget *p, uint frame, float const _origin[3],
    float /*const*/ scale[3], float const rgba[4], float const rgba2[4], float angle,
    Vector3f const &worldOffset) {
    vec3f_t offset = { 0, 0, 0 }, dimensions, origin, originOffset, center;
    vec2f_t texScale = { 1, 1 };
    vec2f_t rotateCenter = { .5f, .5f };
    dd_bool showEdges = true, flipTextureS = false;
    dd_bool mustPopTextureMatrix = false;
    dd_bool textureEnabled = false;
    int numVerts;
    Vector3f *posCoords;
    Vector4f *colorCoords;
    Vector2f *texCoords;
    if (p->frameCount()) {
        FinaleAnimWidget::Frame *f = p->allFrames().at(frame);
        flipTextureS = (f->flags.flip != 0);
        showEdges = false;
        switch (f->type) {
        case FinaleAnimWidget::Frame::PFT_RAW: {
            rawtex_t *rawTex = ClientResources::get().declareRawTexture(f->texRef.lumpNum);
            if (rawTex) {
                DGLuint glName = GL_PrepareRawTexture(*rawTex);
                V3f_Set(offset, 0, 0, 0);
                V3f_Set(dimensions, 320 /*rawTex->width*/, 200 /*rawTex->height*/, 0);
                V2f_Set(rotateCenter, 160, 100);
                GL_BindTextureUnmanaged(glName, gl::ClampToEdge, gl::ClampToEdge,
                                        (filterUI ? gl::Linear : gl::Nearest));
                if (glName) {
                    DGL_Enable(DGL_TEXTURE_2D);
                    textureEnabled = true;
                }
            }
            break; }
        case FinaleAnimWidget::Frame::PFT_XIMAGE:
            V3f_Set(offset, 0, 0, 0);
            V3f_Set(dimensions, 1, 1, 0);
            V2f_Set(rotateCenter, .5f, .5f);
            GL_BindTextureUnmanaged(f->texRef.tex, gl::ClampToEdge, gl::ClampToEdge,
                                    (filterUI ? gl::Linear : gl::Nearest));
            if (f->texRef.tex) {
                DGL_Enable(DGL_TEXTURE_2D);
                textureEnabled = true;
            }
            break;
        case FinaleAnimWidget::Frame::PFT_MATERIAL:
            if (ClientMaterial *mat = static_cast<ClientMaterial *>(f->texRef.material)) {
                MaterialAnimator &matAnimator      = mat->getAnimator(uiMaterialSpec_FinaleAnim());
                matAnimator.prepare();
                Vector2ui const &matDimensions = matAnimator.dimensions();
                TextureVariant *tex            = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
                int const texBorder            = tex->spec().variant.border;
                GL_BindTexture(tex);
                DGL_Enable(DGL_TEXTURE_2D);
                textureEnabled = true;
                V3f_Set(dimensions, matDimensions.x + texBorder * 2, matDimensions.y + texBorder * 2, 0);
                V2f_Set(rotateCenter, dimensions[VX] / 2, dimensions[VY] / 2);
                tex->glCoords(&texScale[VX], &texScale[VY]);
                bool const texIsSprite = !tex->base().manifest().scheme().name().compareWithoutCase("Sprites");
                if (texIsSprite) {
                    V3f_Set(offset, tex->base().origin().x, tex->base().origin().y, 0);
                }
                else {
                    V3f_Set(offset, 0, 0, 0);
                }
            }
            break;
        case FinaleAnimWidget::Frame::PFT_PATCH: {
            res::TextureManifest &manifest = res::Textures::get().textureScheme("Patches")
                                            .findByUniqueId(f->texRef.patch);
            if (manifest.hasTexture()) {
                res::Texture &tex = manifest.texture();
                TextureVariantSpec const &texSpec =
                    Rend_PatchTextureSpec(0 | (tex.isFlagged(res::Texture::Monochrome)        ? TSF_MONOCHROME : 0)
                                            | (tex.isFlagged(res::Texture::UpscaleAndSharpen) ? TSF_UPSCALE_AND_SHARPEN : 0));
                GL_BindTexture(static_cast<ClientTexture &>(tex).prepareVariant(texSpec));
                DGL_Enable(DGL_TEXTURE_2D);
                textureEnabled = true;
                V3f_Set(offset, tex.origin().x, tex.origin().y, 0);
                V3f_Set(dimensions, tex.width(), tex.height(), 0);
                V2f_Set(rotateCenter, dimensions[VX]/2, dimensions[VY]/2);
            }
            break; }
        default:
            App_Error("drawPicFrame: Invalid FI_PIC frame type %i.", int(f->type));
        }
    }
    if (!textureEnabled) {
        V3f_Copy(dimensions, scale);
        V3f_Set(scale, 1, 1, 1);
        V2f_Set(rotateCenter, dimensions[VX] / 2, dimensions[VY] / 2);
    }
    V3f_Set(center, dimensions[VX] / 2, dimensions[VY] / 2, dimensions[VZ] / 2);
    V3f_Sum(origin, _origin, center);
    V3f_Subtract(origin, origin, offset);
    for (int i = 0; i < 3; ++i) { origin[i] += worldOffset[i]; }
    V3f_Subtract(originOffset, offset, center);
    offset[VX] *= scale[VX]; offset[VY] *= scale[VY]; offset[VZ] *= scale[VZ];
    V3f_Sum(originOffset, originOffset, offset);
    numVerts = buildGeometry(dimensions, flipTextureS, rgba, rgba2, &posCoords, &colorCoords, &texCoords);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(origin[VX], origin[VY], origin[VZ]);
    DGL_Translatef(originOffset[VX], originOffset[VY], originOffset[VZ]);
    DGL_Scalef(scale[VX], scale[VY], scale[VZ]);
    if (angle != 0) {
        DGL_Translatef(rotateCenter[VX], rotateCenter[VY], 0);
        DGL_Scalef(1, 200.0f / 240.0f, 1);
        DGL_Rotatef(angle, 0, 0, 1);
        DGL_Scalef(1, 240.0f / 200.0f, 1);
        DGL_Translatef(-rotateCenter[VX], -rotateCenter[VY], 0);
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Scalef(dimensions[VX], dimensions[VY], dimensions[VZ]);
    if (texScale[0] != 1 || texScale[1] != 1) {
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_PushMatrix();
        DGL_Scalef(texScale[0], texScale[1], 1);
        mustPopTextureMatrix = true;
    }
    drawGeometry(numVerts, posCoords, colorCoords, texCoords);
    GL_SetNoTexture();
    if (mustPopTextureMatrix) {
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_PopMatrix();
    }
    if (showEdges) {
        DGL_Begin(DGL_LINES);
            useColor(p->edgeColor(), 4);
            DGL_Vertex2f(0, 0);
            DGL_Vertex2f(1, 0);
            DGL_Vertex2f(1, 0);
            useColor(p->otherEdgeColor(), 4);
            DGL_Vertex2f(1, 1);
            DGL_Vertex2f(1, 1);
            DGL_Vertex2f(0, 1);
            DGL_Vertex2f(0, 1);
            useColor(p->edgeColor(), 4);
            DGL_Vertex2f(0, 0);
        DGL_End();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void FinaleAnimWidget::draw(Vector3f const &offset) {
    if (!(d->color[3].value > 0)) return;
    vec3f_t _scale, _origin;
    V3f_Set(_origin, origin()[VX].value, origin()[VY].value, origin()[VZ].value);
    V3f_Set(_scale, scale()[VX].value, scale()[VY].value, scale()[VZ].value);
    vec4f_t rgba, rgba2;
    V4f_Set(rgba, d->color[0].value, d->color[1].value, d->color[2].value, d->color[3].value);
    if (!frameCount()) {
        V4f_Set(rgba2, d->otherColor[0].value, d->otherColor[1].value, d->otherColor[2].value, d->otherColor[3].value);
    }
    else {
        V4f_Set(rgba2, 0, 0, 0, 0);
    }
    drawPicFrame(this, d->curFrame, _origin, _scale, rgba, (!frameCount()? rgba2 : rgba), angle().value, offset);
}
#endif
void FinaleAnimWidget::runTicks(/*timespan_t timeDelta*/)
{
    FinaleWidget::runTicks(/*timeDelta*/);
    AnimatorVector4_Think(d->color);
    AnimatorVector4_Think(d->otherColor);
    AnimatorVector4_Think(d->edgeColor);
    AnimatorVector4_Think(d->otherEdgeColor);
    if (!(d->frames.count() > 1)) return;
    if (d->frames.at(d->curFrame)->tics > 0) {
        if (--d->tics <= 0) {
            Frame *f;
            uint next = d->curFrame + 1;
            if (next == uint(d->frames.count())) {
                d->animComplete = true;
                if (d->animLooping) {
                    next = 0; // Rewind back to beginning.
                }
                else // Yes.
                {
                    d->frames.at(next = d->curFrame)->tics = 0;
                }
            }
            f = d->frames.at(d->curFrame = next);
            d->tics = f->tics;
            if (f->sound > 0)
                S_LocalSound(f->sound, 0);
        }
    }
}
int FinaleAnimWidget::newFrame(Frame::Type type, int tics, void *texRef, short sound, bool flagFlipH) {
    d->frames.append(d->makeFrame(type, tics, texRef, sound, flagFlipH));
    d->animComplete = false;
    return d->frames.count();
}
FinaleAnimWidget::Frames const &FinaleAnimWidget::allFrames() const {
    return d->frames;
}
FinaleAnimWidget &FinaleAnimWidget::clearAllFrames() {
    qDeleteAll(d->frames); d->frames.clear();
    d->curFrame     = 0;
    d->animComplete = true;  // Nothing to animate.
    d->animLooping  = false; // Yeah?
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::resetAllColors() {
    AnimatorVector4_Init(d->color,      1, 1, 1, 1);
    AnimatorVector4_Init(d->otherColor, 1, 1, 1, 1);
    AnimatorVector4_Init(d->edgeColor,      1, 1, 1, 0);
    AnimatorVector4_Init(d->otherEdgeColor, 1, 1, 1, 0);
    return *this;
}
animator_t const *FinaleAnimWidget::color() const {
    return d->color;
}
FinaleAnimWidget &FinaleAnimWidget::setColor(Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->color, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setAlpha(float newAlpha, int steps) {
    Animator_Set(&d->color[3], newAlpha, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->color, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
animator_t const *FinaleAnimWidget::edgeColor() const {
    return d->edgeColor;
}
FinaleAnimWidget &FinaleAnimWidget::setEdgeColor(Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->edgeColor, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setEdgeAlpha(float newAlpha, int steps) {
    Animator_Set(&d->edgeColor[3], newAlpha, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setEdgeColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->edgeColor, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
animator_t const *FinaleAnimWidget::otherColor() const {
    return d->otherColor;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherColor(de::Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->otherColor, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherAlpha(float newAlpha, int steps) {
    Animator_Set(&d->otherColor[3], newAlpha, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->otherColor, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
animator_t const *FinaleAnimWidget::otherEdgeColor() const {
    return d->otherEdgeColor;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherEdgeColor(de::Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->otherEdgeColor, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherEdgeAlpha(float newAlpha, int steps) {
    Animator_Set(&d->otherEdgeColor[3], newAlpha, steps);
    return *this;
}
FinaleAnimWidget &FinaleAnimWidget::setOtherEdgeColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->otherEdgeColor, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
enum fi_operand_type_t {
    FVT_INT,
    FVT_FLOAT,
    FVT_STRING,
    FVT_URI
};
static fi_operand_type_t operandTypeForCharCode(char code) {
    switch (code) {
    case 'i': return FVT_INT;
    case 'f': return FVT_FLOAT;
    case 's': return FVT_STRING;
    case 'u': return FVT_URI;
    default: throw Error("operandTypeForCharCode", String("Unknown char-code %1").arg(code));
    }
}
#define OP_INT(n)           (ops[n].data.integer)
#define OP_FLOAT(n)         (ops[n].data.flt)
#define OP_CSTRING(n)       (ops[n].data.cstring)
#define OP_URI(n)           (ops[n].data.uri)
struct fi_operand_t {
    fi_operand_type_t type;
    union {
        int         integer;
        float       flt;
        char const *cstring;
        uri_s      *uri;
    } data;
};
#define DEFFC(name) void FIC_##name(command_t const &cmd, fi_operand_t const *ops, FinaleInterpreter &fi)
/**
 * @defgroup finaleInterpreterCommandDirective Finale Interpreter Command Directive
 * @ingroup infine
 */
/*@{*/
#define FID_NORMAL          0
#define FID_ONLOAD          0x1
/*@}*/
struct command_t {
    char const *token;
    char const *operands;
    typedef void (*Func) (command_t const &, fi_operand_t const *, FinaleInterpreter &);
    Func func;
    struct command_flags_s {
        char when_skipping:1;
        char when_condition_skipping:1; // Skipping because condition failed.
    } flags;
    int excludeDirectives;
};
DEFFC(Do);
DEFFC(End);
DEFFC(BGMaterial);
DEFFC(NoBGMaterial);
DEFFC(Wait);
DEFFC(WaitText);
DEFFC(WaitAnim);
DEFFC(Tic);
DEFFC(InTime);
DEFFC(Color);
DEFFC(ColorAlpha);
DEFFC(Pause);
DEFFC(CanSkip);
DEFFC(NoSkip);
DEFFC(SkipHere);
DEFFC(Events);
DEFFC(NoEvents);
DEFFC(OnKey);
DEFFC(UnsetKey);
DEFFC(If);
DEFFC(IfNot);
DEFFC(Else);
DEFFC(GoTo);
DEFFC(Marker);
DEFFC(Image);
DEFFC(ImageAt);
DEFFC(XImage);
DEFFC(Delete);
DEFFC(Patch);
DEFFC(SetPatch);
DEFFC(Anim);
DEFFC(AnimImage);
DEFFC(StateAnim);
DEFFC(Repeat);
DEFFC(ClearAnim);
DEFFC(PicSound);
DEFFC(ObjectOffX);
DEFFC(ObjectOffY);
DEFFC(ObjectOffZ);
DEFFC(ObjectScaleX);
DEFFC(ObjectScaleY);
DEFFC(ObjectScaleZ);
DEFFC(ObjectScale);
DEFFC(ObjectScaleXY);
DEFFC(ObjectScaleXYZ);
DEFFC(ObjectRGB);
DEFFC(ObjectAlpha);
DEFFC(ObjectAngle);
DEFFC(Rect);
DEFFC(FillColor);
DEFFC(EdgeColor);
DEFFC(OffsetX);
DEFFC(OffsetY);
DEFFC(Sound);
DEFFC(SoundAt);
DEFFC(SeeSound);
DEFFC(DieSound);
DEFFC(Music);
DEFFC(MusicOnce);
DEFFC(Filter);
DEFFC(Text);
DEFFC(TextFromDef);
DEFFC(TextFromLump);
DEFFC(SetText);
DEFFC(SetTextDef);
DEFFC(DeleteText);
DEFFC(Font);
DEFFC(FontA);
DEFFC(FontB);
DEFFC(PredefinedColor);
DEFFC(PredefinedFont);
DEFFC(TextRGB);
DEFFC(TextAlpha);
DEFFC(TextOffX);
DEFFC(TextOffY);
DEFFC(TextCenter);
DEFFC(TextNoCenter);
DEFFC(TextScroll);
DEFFC(TextPos);
DEFFC(TextRate);
DEFFC(TextLineHeight);
DEFFC(NoMusic);
DEFFC(TextScaleX);
DEFFC(TextScaleY);
DEFFC(TextScale);
DEFFC(PlayDemo);
DEFFC(Command);
DEFFC(ShowMenu);
DEFFC(NoShowMenu);
/**
 * Time is measured in seconds.
 * Colors are floating point and [0..1].
 *
 * @todo This data should be pre-processed (i.e., parsed) during module init
 *       and any symbolic references or other indirections resolved once
 *       rather than repeatedly during script interpretation.
 *
 *       At this time the command names could also be hashed and chained to
 *       improve performance. -ds
 */
static command_t const *findCommand(char const *name) {
    static command_t const commands[] = {
        { "DO",         "", FIC_Do, true, true },
        { "END",        "", FIC_End },
        { "IF",         "s", FIC_If }, // if (value-id)
        { "IFNOT",      "s", FIC_IfNot }, // ifnot (value-id)
        { "ELSE",       "", FIC_Else },
        { "GOTO",       "s", FIC_GoTo }, // goto (marker)
        { "MARKER",     "s", FIC_Marker, true },
        { "in",         "f", FIC_InTime }, // in (time)
        { "pause",      "", FIC_Pause },
        { "tic",        "", FIC_Tic },
        { "wait",       "f", FIC_Wait }, // wait (time)
        { "waittext",   "s", FIC_WaitText }, // waittext (id)
        { "waitanim",   "s", FIC_WaitAnim }, // waitanim (id)
        { "canskip",    "", FIC_CanSkip },
        { "noskip",     "", FIC_NoSkip },
        { "skiphere",   "", FIC_SkipHere, true },
        { "events",     "", FIC_Events },
        { "noevents",   "", FIC_NoEvents },
        { "onkey",      "ss", FIC_OnKey }, // onkey (keyname) (marker)
        { "unsetkey",   "s", FIC_UnsetKey }, // unsetkey (keyname)
        { "color",      "fff", FIC_Color }, // color (red) (green) (blue)
        { "coloralpha", "ffff", FIC_ColorAlpha }, // coloralpha (r) (g) (b) (a)
        { "flat",       "u(flats:)", FIC_BGMaterial }, // flat (flat-id)
        { "texture",    "u(textures:)", FIC_BGMaterial }, // texture (texture-id)
        { "noflat",     "", FIC_NoBGMaterial },
        { "notexture",  "", FIC_NoBGMaterial },
        { "offx",       "f", FIC_OffsetX }, // offx (x)
        { "offy",       "f", FIC_OffsetY }, // offy (y)
        { "filter",     "ffff", FIC_Filter }, // filter (r) (g) (b) (a)
        { "sound",      "s", FIC_Sound }, // sound (snd)
        { "soundat",    "sf", FIC_SoundAt }, // soundat (snd) (vol:0-1)
        { "seesound",   "s", FIC_SeeSound }, // seesound (mobjtype)
        { "diesound",   "s", FIC_DieSound }, // diesound (mobjtype)
        { "music",      "s", FIC_Music }, // music (musicname)
        { "musiconce",  "s", FIC_MusicOnce }, // musiconce (musicname)
        { "nomusic",    "", FIC_NoMusic },
        { "del",        "s", FIC_Delete }, // del (wi)
        { "x",          "sf", FIC_ObjectOffX }, // x (wi) (x)
        { "y",          "sf", FIC_ObjectOffY }, // y (wi) (y)
        { "z",          "sf", FIC_ObjectOffZ }, // z (wi) (z)
        { "sx",         "sf", FIC_ObjectScaleX }, // sx (wi) (x)
        { "sy",         "sf", FIC_ObjectScaleY }, // sy (wi) (y)
        { "sz",         "sf", FIC_ObjectScaleZ }, // sz (wi) (z)
        { "scale",      "sf", FIC_ObjectScale }, // scale (wi) (factor)
        { "scalexy",    "sff", FIC_ObjectScaleXY }, // scalexy (wi) (x) (y)
        { "scalexyz",   "sfff", FIC_ObjectScaleXYZ }, // scalexyz (wi) (x) (y) (z)
        { "rgb",        "sfff", FIC_ObjectRGB }, // rgb (wi) (r) (g) (b)
        { "alpha",      "sf", FIC_ObjectAlpha }, // alpha (wi) (alpha)
        { "angle",      "sf", FIC_ObjectAngle }, // angle (wi) (degrees)
        { "rect",       "sffff", FIC_Rect }, // rect (hndl) (x) (y) (w) (h)
        { "fillcolor",  "ssffff", FIC_FillColor }, // fillcolor (wi) (top/bottom/both) (r) (g) (b) (a)
        { "edgecolor",  "ssffff", FIC_EdgeColor }, // edgecolor (wi) (top/bottom/both) (r) (g) (b) (a)
        { "image",      "ss", FIC_Image }, // image (id) (raw-image-lump)
        { "imageat",    "sffs", FIC_ImageAt }, // imageat (id) (x) (y) (raw)
        { "ximage",     "ss", FIC_XImage }, // ximage (id) (ext-gfx-filename)
        { "patch",      "sffs", FIC_Patch }, // patch (id) (x) (y) (patch)
        { "set",        "ss", FIC_SetPatch }, // set (id) (lump)
        { "clranim",    "s", FIC_ClearAnim }, // clranim (wi)
        { "anim",       "ssf", FIC_Anim }, // anim (id) (patch) (time)
        { "imageanim",  "ssf", FIC_AnimImage }, // imageanim (id) (raw-img) (time)
        { "picsound",   "ss", FIC_PicSound }, // picsound (id) (sound)
        { "repeat",     "s", FIC_Repeat }, // repeat (id)
        { "states",     "ssi", FIC_StateAnim }, // states (id) (state) (count)
        { "text",       "sffs", FIC_Text }, // text (hndl) (x) (y) (string)
        { "textdef",    "sffs", FIC_TextFromDef }, // textdef (hndl) (x) (y) (txt-id)
        { "textlump",   "sffs", FIC_TextFromLump }, // textlump (hndl) (x) (y) (lump)
        { "settext",    "ss", FIC_SetText }, // settext (id) (newtext)
        { "settextdef", "ss", FIC_SetTextDef }, // settextdef (id) (txt-id)
        { "center",     "s", FIC_TextCenter }, // center (id)
        { "nocenter",   "s", FIC_TextNoCenter }, // nocenter (id)
        { "scroll",     "sf", FIC_TextScroll }, // scroll (id) (speed)
        { "pos",        "si", FIC_TextPos }, // pos (id) (pos)
        { "rate",       "si", FIC_TextRate }, // rate (id) (rate)
        { "font",       "su", FIC_Font }, // font (id) (font)
        { "linehgt",    "sf", FIC_TextLineHeight }, // linehgt (hndl) (hgt)
        { "playdemo",   "s", FIC_PlayDemo }, // playdemo (filename)
        { "cmd",        "s", FIC_Command }, // cmd (console command)
        { "trigger",    "", FIC_ShowMenu },
        { "notrigger",  "", FIC_NoShowMenu },
        { "precolor",   "ifff", FIC_PredefinedColor }, // precolor (num) (r) (g) (b)
        { "prefont",    "iu", FIC_PredefinedFont }, // prefont (num) (font)
        { "fonta",      "s", FIC_FontA }, // fonta (id)
        { "fontb",      "s", FIC_FontB }, // fontb (id)
        { "delpic",     "s", FIC_Delete }, // delpic (wi)
        { "deltext",    "s", FIC_DeleteText }, // deltext (wi)
        { "textrgb",    "sfff", FIC_TextRGB }, // textrgb (id) (r) (g) (b)
        { "textalpha",  "sf", FIC_TextAlpha }, // textalpha (id) (alpha)
        { "tx",         "sf", FIC_TextOffX }, // tx (id) (x)
        { "ty",         "sf", FIC_TextOffY }, // ty (id) (y)
        { "tsx",        "sf", FIC_TextScaleX }, // tsx (id) (x)
        { "tsy",        "sf", FIC_TextScaleY }, // tsy (id) (y)
        { "textscale",  "sf", FIC_TextScale }, // textscale (id) (x) (y)
        { nullptr, 0, nullptr } // Terminate.
    };
    for (size_t i = 0; commands[i].token; ++i) {
        command_t const *cmd = &commands[i];
        if (!qstricmp(cmd->token, name)) {
            return cmd;
        }
    }
    return nullptr; // Not found.
}
DENG2_PIMPL(FinaleInterpreter) {
    struct Flags {
        char stopped:1;
        char suspended:1;
        char paused:1;
        char can_skip:1;
        char eat_events:1; /// Script will eat all input events.
        char show_menu:1;
    } flags;
    finaleid_t id = 0;                  ///< Unique identifier.
    ddstring_t *script      = nullptr;  ///< The script to be interpreted.
    char const *scriptBegin = nullptr;  ///< Beginning of the script (after any directive blocks).
    char const *cp          = nullptr;  ///< Current position in the script.
    char token[MAX_TOKEN_LENGTH];       ///< Script token read/parse buffer.
    std::unique_ptr<FinalePageWidget> pages[2];
    bool cmdExecuted = false;  ///< Set to true after first command is executed.
    bool skipping    = false;
    bool lastSkipped = false;
    bool skipNext    = false;
    bool gotoEnd     = false;
    bool gotoSkip    = false;
    String gotoTarget;
    int doLevel = 0;  ///< Level of DO-skipping.
    uint timer  = 0;
    int wait    = 0;
    int inTime  = 0;
    FinaleAnimWidget *waitAnim = nullptr;
    FinaleTextWidget *waitText = nullptr;
#ifdef __CLIENT__
    struct EventHandler {
        ddevent_t ev; // Template.
        String gotoMarker;
        explicit EventHandler(ddevent_t const *evTemplate = nullptr,
                              String const &gotoMarker    = "")
            : gotoMarker(gotoMarker) {
            std::memcpy(&ev, &evTemplate, sizeof(ev));
        }
        EventHandler(EventHandler const &other)
            : gotoMarker(other.gotoMarker) {
            std::memcpy(&ev, &other.ev, sizeof(ev));
        }
    };
    typedef QList<EventHandler> EventHandlers;
    EventHandlers eventHandlers;
#endif // __CLIENT__
    Impl(Public *i, finaleid_t id) : Base(i), id(id) {
        de::zap(flags);
        de::zap(token);
    }
    ~Impl() {
        stop();
        releaseScript();
    }
    void initDefaultState() {
        flags.suspended = false;
        flags.paused    = false;
        flags.show_menu = true; // Unhandled events will show a menu.
        flags.can_skip  = true; // By default skipping is enabled.
        cmdExecuted = false; // Nothing is drawn until a cmd has been executed.
        skipping    = false;
        wait        = 0;     // Not waiting for anything.
        inTime      = 0;     // Interpolation is off.
        timer       = 0;
        gotoSkip    = false;
        gotoEnd     = false;
        skipNext    = false;
        waitText    = nullptr;
        waitAnim    = nullptr;
        gotoTarget.clear();
#ifdef __CLIENT__
        eventHandlers.clear();
#endif
    }
    void releaseScript() {
        Str_Delete(script); script = nullptr;
        scriptBegin = nullptr;
        cp          = nullptr;
    }
    void stop() {
        if (flags.stopped) return;
        flags.stopped = true;
        LOGDEV_SCR_MSG("Finale End - id:%i '%.30s'") << id << scriptBegin;
#ifdef __SERVER__
        if (::isServer && !(FI_ScriptFlags(id) & FF_LOCAL)) {
            Sv_Finale(id, FINF_END, 0);
        }
#endif
        DoomsdayApp::plugins().callAllHooks(HOOK_FINALE_SCRIPT_STOP, id);
    }
    bool atEnd() const {
        DENG2_ASSERT(script);
        return (cp - Str_Text(script)) >= Str_Length(script);
    }
    void findBegin() {
        char const *tok;
        while (!gotoEnd && 0 != (tok = nextToken()) && qstricmp(tok, "{")) {}
    }
    void findEnd() {
        char const *tok;
        while (!gotoEnd && 0 != (tok = nextToken()) && qstricmp(tok, "}")) {}
    }
    char const *nextToken() {
        while (!atEnd() && isspace(*cp)) { cp++; }
        if (atEnd()) return nullptr;
        char *out = token;
        if (*cp == '"') // A string?
        {
            for (cp++; !atEnd(); cp++) {
                if (*cp == '"') {
                    cp++;
                    if (*cp != '"') break;
                }
                *out++ = *cp;
            }
        }
        else {
            while (!isspace(*cp) && !atEnd()) { *out++ = *cp++; }
        }
        *out++ = 0;
        return token;
    }
    bool executeNextCommand() {
        if (char const *tok = nextToken()) {
            executeCommand(tok, FID_NORMAL);
            if (cmdExecuted) {
                pages[Anims]->makeVisible();
                pages[Texts]->makeVisible();
            }
            return false;
        }
        return true;
    }
    static inline char const *findDefaultValueEnd(char const *str) {
        char const *defaultValueEnd;
        for (defaultValueEnd = str; defaultValueEnd && *defaultValueEnd != ')'; defaultValueEnd++)
        {}
        DENG2_ASSERT(defaultValueEnd < str + qstrlen(str));
        return defaultValueEnd;
    }
    static char const *nextOperand(char const *operands) {
        if (operands && operands[0]) {
            int len = qstrlen(operands);
            if (len > 1 && operands[1] == '(') {
                return findDefaultValueEnd(operands + 2) + 1;
            }
            return operands + 1;
        }
        return nullptr; // No more operands.
    }
    static int countCommandOperands(char const *operands) {
        int count = 0;
        while (operands && operands[0]) {
            count += 1;
            operands = nextOperand(operands);
        }
        return count;
    }
    /**
     * Prepare the command operands from the script. If successfull, a ptr to a new
     * vector of @c fi_operand_t objects is returned. Ownership of the vector is
     * given to the caller.
     *
     * @return  Array of @c fi_operand_t else @c nullptr. Must be free'd with M_Free().
     */
    fi_operand_t *prepareCommandOperands(command_t const *cmd, int *count) {
        DENG2_ASSERT(cmd);
        char const *origCursorPos = cp;
        int const operandCount    = countCommandOperands(cmd->operands);
        if (operandCount <= 0) return nullptr;
        fi_operand_t *operands = (fi_operand_t *) M_Malloc(sizeof(*operands) * operandCount);
        char const *opRover    = cmd->operands;
        for (fi_operand_t *op = operands; opRover && opRover[0]; opRover = nextOperand(opRover), op++) {
            char const charCode = *opRover;
            op->type = operandTypeForCharCode(charCode);
            bool const opHasDefaultValue = (opRover < cmd->operands + (qstrlen(cmd->operands) - 2) && opRover[1] == '(');
            bool const haveValue         = !!nextToken();
            if (!haveValue && !opHasDefaultValue) {
                cp = origCursorPos;
                if (operands) M_Free(operands);
                if (count) *count = 0;
                App_Error("prepareCommandOperands: Too few operands for command '%s'.\n", cmd->token);
                return 0; // Unreachable.
            }
            switch (op->type) {
            case FVT_INT: {
                char const *valueStr = haveValue? token : nullptr;
                if (!valueStr) {
                    int const defaultValueLen = (findDefaultValueEnd(opRover + 2) - opRover) - 1;
                    AutoStr *defaultValue     = Str_PartAppend(AutoStr_NewStd(), opRover + 2, 0, defaultValueLen);
                    valueStr = Str_Text(defaultValue);
                }
                op->data.integer = strtol(valueStr, nullptr, 0);
                break; }
            case FVT_FLOAT: {
                char const *valueStr = haveValue? token : nullptr;
                if (!valueStr) {
                    int const defaultValueLen = (findDefaultValueEnd(opRover + 2) - opRover) - 1;
                    AutoStr *defaultValue     = Str_PartAppend(AutoStr_NewStd(), opRover + 2, 0, defaultValueLen);
                    valueStr = Str_Text(defaultValue);
                }
                op->data.flt = strtod(valueStr, nullptr);
                break; }
            case FVT_STRING: {
                char const *valueStr = haveValue? token : nullptr;
                int valueLen         = haveValue? qstrlen(token) : 0;
                if (!valueStr) {
                    int const defaultValueLen = (findDefaultValueEnd(opRover + 2) - opRover) - 1;
                    AutoStr *defaultValue     = Str_PartAppend(AutoStr_NewStd(), opRover + 2, 0, defaultValueLen);
                    valueStr = Str_Text(defaultValue);
                    valueLen = defaultValueLen;
                }
                op->data.cstring = (char *)M_Malloc(valueLen + 1);
                qstrcpy((char *)op->data.cstring, valueStr);
                break; }
            case FVT_URI: {
                uri_s *uri = Uri_New();
                if (opHasDefaultValue) {
                    int const defaultValueLen = (findDefaultValueEnd(opRover + 2) - opRover) - 1;
                    AutoStr *defaultValue     = Str_PartAppend(AutoStr_NewStd(), opRover + 2, 0, defaultValueLen);
                    Uri_SetUri2(uri, Str_Text(defaultValue), RC_NULL);
                }
                if (haveValue) {
                    Uri_SetUri2(uri, token, RC_NULL);
                }
                op->data.uri = uri;
                break; }
            default: break; // Unreachable.
            }
        }
        if (count) *count = operandCount;
        return operands;
    }
    bool skippingCommand(command_t const *cmd) {
        DENG2_ASSERT(cmd);
        if ((skipNext && !cmd->flags.when_condition_skipping) ||
           ((skipping || gotoSkip) && !cmd->flags.when_skipping)) {
            if (!doLevel) {
                if (skipNext)
                    lastSkipped = true;
                skipNext = false;
            }
            return true;
        }
        return false;
    }
    /**
     * Execute one (the next) command, advance script cursor.
     */
    bool executeCommand(char const *commandString, int directive) {
        DENG2_ASSERT(commandString);
        bool didSkip = false;
        if (!qstrcmp(commandString, ";")) {
            if (doLevel > 0) {
                if (--doLevel == 0) {
                    skipNext    = false;
                    lastSkipped = true;
                }
            }
            return true; // Success!
        }
        cmdExecuted = true;
        if (command_t const *cmd = findCommand(commandString)) {
            bool const requiredOperands = (cmd->operands && cmd->operands[0]);
            if (directive != 0 && cmd->excludeDirectives != 0 &&
               (cmd->excludeDirectives & directive) == 0)
                App_Error("executeCommand: Command \"%s\" is not supported for directive %i.",
                          cmd->token, directive);
            int numOps        = 0;
            fi_operand_t *ops = nullptr;
            if (!requiredOperands || (ops = prepareCommandOperands(cmd, &numOps))) {
                if (!(didSkip = skippingCommand(cmd))) {
                    cmd->func(*cmd, ops, self());
                }
            }
            if (!didSkip) {
                if (gotoEnd) {
                    wait = 1;
                }
                else {
                    lastSkipped = false;
                }
            }
            if (ops) {
                for (int i = 0; i < numOps; ++i) {
                    fi_operand_t *op = &ops[i];
                    switch (op->type) {
                    case FVT_STRING: M_Free((char *)op->data.cstring); break;
                    case FVT_URI:    Uri_Delete(op->data.uri);         break;
                    default: break;
                    }
                }
                M_Free(ops);
            }
        }
        return !didSkip;
    }
    static inline PageIndex choosePageFor(FinaleWidget &widget) {
        return is<FinaleAnimWidget>(widget)? Anims : Texts;
    }
    static inline PageIndex choosePageFor(fi_obtype_e type) {
        return type == FI_ANIM? Anims : Texts;
    }
    FinaleWidget *locateWidget(fi_obtype_e type, String const &name) {
        if (!name.isEmpty()) {
            FinalePageWidget::Children const &children = pages[choosePageFor(type)]->children();
            for (FinaleWidget *widget : children) {
                if (!widget->name().compareWithoutCase(name)) {
                    return widget;
                }
            }
        }
        return nullptr; // Not found.
    }
    FinaleWidget *makeWidget(fi_obtype_e type, String const &name) {
        if (type == FI_ANIM) {
            return new FinaleAnimWidget(name);
        }
        if (type == FI_TEXT) {
            auto *wi = new FinaleTextWidget(name);
            wi->setPageFont(1)
               .setPageColor(1);
            return wi;
        }
        return nullptr;
    }
#if __CLIENT__
    EventHandler *findEventHandler(ddevent_t const &ev) const {
        for (EventHandler const &eh : eventHandlers) {
            if (eh.ev.device != ev.device && eh.ev.type != ev.type)
                continue;
            switch (eh.ev.type) {
            case E_TOGGLE:
                if (eh.ev.toggle.id != ev.toggle.id)
                    continue;
                break;
            case E_AXIS:
                if (eh.ev.axis.id != ev.axis.id)
                    continue;
                break;
            case E_ANGLE:
                if (eh.ev.angle.id != ev.angle.id)
                    continue;
                break;
            default:
                App_Error("Internal error: Invalid event template (type=%i) in finale event handler.", int(eh.ev.type));
            }
            return &const_cast<EventHandler &>(eh);
        }
        return nullptr;
    }
#endif // __CLIENT__
};
FinaleInterpreter::FinaleInterpreter(finaleid_t id) : d(new Impl(this, id))
{}
finaleid_t FinaleInterpreter::id() const {
    return d->id;
}
void FinaleInterpreter::loadScript(char const *script) {
    DENG2_ASSERT(script && script[0]);
    d->pages[Anims].reset(new FinalePageWidget);
    d->pages[Texts].reset(new FinalePageWidget);
    d->pages[Anims]->makeVisible(false);
    d->pages[Texts]->makeVisible(false);
    d->script      = Str_Set(Str_NewStd(), script);
    d->scriptBegin = Str_Text(d->script);
    d->cp          = Str_Text(d->script); // Init cursor.
    d->initDefaultState();
    if (d->nextToken()) {
        if (!qstricmp(d->token, "OnLoad")) {
            d->findBegin();
            forever {
                d->nextToken();
                if (!qstricmp(d->token, "}"))
                    goto end_read;
                if (!d->executeCommand(d->token, FID_ONLOAD))
                    App_Error("FinaleInterpreter::LoadScript: Unknown error"
                              "occured executing directive \"OnLoad\".");
            }
            d->findEnd();
end_read:
            while (*d->cp && isspace(*d->cp)) { d->cp++; }
            d->scriptBegin = Str_Text(d->script) + (d->cp - Str_Text(d->script));
            d->cp          = d->scriptBegin;
            d->initDefaultState();
        }
    }
    DoomsdayApp::plugins().callAllHooks(HOOK_FINALE_SCRIPT_BEGIN, d->id);
}
void FinaleInterpreter::resume() {
    if (!d->flags.suspended) return;
    d->flags.suspended = false;
    d->pages[Anims]->pause(false);
    d->pages[Texts]->pause(false);
    if (d->cmdExecuted) {
        d->pages[Anims]->makeVisible();
        d->pages[Texts]->makeVisible();
    }
}
void FinaleInterpreter::suspend() {
    LOG_AS("FinaleInterpreter");
    if (d->flags.suspended) return;
    d->flags.suspended = true;
    d->pages[Anims]->pause();
    d->pages[Anims]->makeVisible(false);
    d->pages[Texts]->pause();
    d->pages[Texts]->makeVisible(false);
}
void FinaleInterpreter::terminate() {
    d->stop();
#ifdef __CLIENT__
    d->eventHandlers.clear();
#endif
    d->releaseScript();
}
bool FinaleInterpreter::isMenuTrigger() const {
    if (d->flags.paused || d->flags.can_skip) {
        return false;
    }
    return (d->flags.show_menu != 0);
}
bool FinaleInterpreter::isSuspended() const {
    return (d->flags.suspended != 0);
}
void FinaleInterpreter::allowSkip(bool yes) {
    d->flags.can_skip = yes;
}
bool FinaleInterpreter::canSkip() const {
    return (d->flags.can_skip != 0);
}
bool FinaleInterpreter::commandExecuted() const {
    return d->cmdExecuted;
}
static bool runOneTick(FinaleInterpreter &fi) {
    ddhook_finale_script_ticker_paramaters_t parm;
    de::zap(parm);
    parm.runTick = true;
    parm.canSkip = fi.canSkip();
    DoomsdayApp::plugins().callAllHooks(HOOK_FINALE_SCRIPT_TICKER, fi.id(), &parm);
    return parm.runTick;
}
bool FinaleInterpreter::runTicks(timespan_t timeDelta, bool processCommands) {
    LOG_AS("FinaleInterpreter");
    page(Anims).runTicks(timeDelta);
    page(Texts).runTicks(timeDelta);
    if (!processCommands)   return false;
    if (d->flags.stopped)   return false;
    if (d->flags.suspended) return false;
    d->timer++;
    if (!runOneTick(*this))
        return false;
    if (d->wait && --d->wait)
        return false;
    if (d->flags.paused)
        return false;
    if (d->waitText) {
        if (!d->waitText->animationComplete())
            return false;
        d->waitText = nullptr;
    }
    if (d->waitAnim) {
        if (!d->waitAnim->animationComplete())
            return false;
        d->waitAnim = nullptr;
    }
    bool foundEnd = false;
    while (!d->gotoEnd && !d->wait && !d->waitText && !d->waitAnim && !foundEnd) {
        foundEnd = d->executeNextCommand();
    }
    return (d->gotoEnd || (foundEnd && d->flags.can_skip));
}
bool FinaleInterpreter::skip() {
    LOG_AS("FinaleInterpreter");
    if (d->waitText && d->flags.can_skip && !d->flags.paused) {
        d->waitText->accelerate();
        return true;
    }
    d->waitText = nullptr;
    d->waitAnim = nullptr;
    if (d->flags.paused) {
        d->flags.paused = false;
        d->wait = 0;
        return true;
    }
    if (d->flags.can_skip) {
        d->skipping = true; // Start skipping ahead.
        d->wait     = 0;
        return true;
    }
    return (d->flags.eat_events != 0);
}
bool FinaleInterpreter::skipToMarker(String const &marker) {
    LOG_AS("FinaleInterpreter");
    if (marker.isEmpty()) return false;
    d->gotoTarget = marker;
    d->gotoSkip   = true;    // Start skipping until the marker is found.
    d->wait       = 0;       // Stop any waiting.
    d->waitText   = nullptr;
    d->waitAnim   = nullptr;
    d->cp = d->scriptBegin;
    return true;
}
bool FinaleInterpreter::skipInProgress() const {
    return d->skipNext;
}
bool FinaleInterpreter::lastSkipped() const {
    return d->lastSkipped;
}
int FinaleInterpreter::handleEvent(ddevent_t const &ev) {
    LOG_AS("FinaleInterpreter");
    if (d->flags.suspended)
        return false;
    if (d->timer < 20)
        return false;
    if (!::isClient) {
#ifdef __CLIENT__
        if (IS_KEY_DOWN(&ev)) {
            if (Impl::EventHandler *eh = d->findEventHandler(ev)) {
                skipToMarker(eh->gotoMarker);
                if (IS_TOGGLE_UP(&ev)) return false;
                return (d->flags.eat_events != 0);
            }
        }
#endif
    }
    if (!d->flags.can_skip && !d->flags.paused)
        return false;
    if (!IS_TOGGLE_DOWN(&ev))
        return false;
#ifdef __CLIENT__
    if (::isClient) {
        Cl_RequestFinaleSkip();
        return true;
    }
#endif
#ifdef __SERVER__
    Sv_Finale(d->id, FINF_SKIP, 0);
#endif
    return skip();
}
#ifdef __CLIENT__
void FinaleInterpreter::addEventHandler(ddevent_t const &evTemplate, String const &gotoMarker) {
    if (d->findEventHandler(evTemplate)) return;
    d->eventHandlers.append(Impl::EventHandler(&evTemplate, gotoMarker));
}
void FinaleInterpreter::removeEventHandler(ddevent_t const &evTemplate) {
    if (Impl::EventHandler *eh = d->findEventHandler(evTemplate)) {
        int index = 0;
        while (&d->eventHandlers.at(index) != eh) { index++; }
        d->eventHandlers.removeAt(index);
    }
}
#endif // __CLIENT__
FinalePageWidget &FinaleInterpreter::page(PageIndex index) {
    if (index >= Anims && index <= Texts) {
        DENG2_ASSERT(d->pages[index]);
        return *d->pages[index];
    }
    throw MissingPageError("FinaleInterpreter::page", "Unknown page #" + String::number(int(index)));
}
FinalePageWidget const &FinaleInterpreter::page(PageIndex index) const {
    return const_cast<FinalePageWidget const &>(const_cast<FinaleInterpreter *>(this)->page(index));
}
FinaleWidget *FinaleInterpreter::tryFindWidget(String const &name) {
    if (FinaleWidget *found = d->locateWidget(FI_ANIM, name)) {
        return found;
    }
    if (FinaleWidget *found = d->locateWidget(FI_TEXT, name)) {
        return found;
    }
    return nullptr;
}
FinaleWidget &FinaleInterpreter::findWidget(fi_obtype_e type, String const &name) {
    if (FinaleWidget *foundWidget = d->locateWidget(type, name)) {
        return *foundWidget;
    }
    throw MissingWidgetError("FinaleInterpeter::findWidget", "Failed locating widget for name:'" + name + "'");
}
FinaleWidget &FinaleInterpreter::findOrCreateWidget(fi_obtype_e type, String const &name) {
    DENG2_ASSERT(type >= FI_ANIM && type <= FI_TEXT);
    DENG2_ASSERT(!name.isEmpty());
    if (FinaleWidget *foundWidget = d->locateWidget(type, name)) {
        return *foundWidget;
    }
    FinaleWidget *newWidget = d->makeWidget(type, name);
    if (!newWidget) throw Error("FinaleInterpreter::findOrCreateWidget", "Failed making widget for type:" + String::number(int(type)));
    return *page(d->choosePageFor(*newWidget)).addChild(newWidget);
}
void FinaleInterpreter::beginDoSkipMode() {
    if (!skipInProgress()) return;
    d->doLevel++;
}
void FinaleInterpreter::gotoEnd() {
    d->gotoEnd = true;
}
void FinaleInterpreter::pause() {
    d->flags.paused = true;
    wait(1);
}
void FinaleInterpreter::wait(int ticksToWait) {
    d->wait = ticksToWait;
}
void FinaleInterpreter::foundSkipHere() {
    d->skipping = false;
}
void FinaleInterpreter::foundSkipMarker(String const &marker) {
    if (!d->gotoTarget.compareWithoutCase(marker)) {
        d->gotoSkip = false;
    }
}
int FinaleInterpreter::inTime() const {
    return d->inTime;
}
void FinaleInterpreter::setInTime(int seconds) {
    d->inTime = seconds;
}
void FinaleInterpreter::setHandleEvents(bool yes) {
    d->flags.eat_events = yes;
}
void FinaleInterpreter::setShowMenu(bool yes) {
    d->flags.show_menu = yes;
}
void FinaleInterpreter::setSkip(bool allowed) {
    d->flags.can_skip = allowed;
}
void FinaleInterpreter::setSkipNext(bool yes) {
   d->skipNext = yes;
}
void FinaleInterpreter::setWaitAnim(FinaleAnimWidget *newWaitAnim) {
    d->waitAnim = newWaitAnim;
}
void FinaleInterpreter::setWaitText(FinaleTextWidget *newWaitText) {
    d->waitText = newWaitText;
}
DEFFC(Do) {
    DENG2_UNUSED2(cmd, ops);
    fi.beginDoSkipMode();
}
DEFFC(End) {
    DENG2_UNUSED2(cmd, ops);
    fi.gotoEnd();
}
static void changePageBackground(FinalePageWidget &page, world::Material *newMaterial) {
    if (newMaterial && !page.backgroundMaterial()) {
        page.setBackgroundTopColorAndAlpha   (Vector4f(1, 1, 1, 1))
            .setBackgroundBottomColorAndAlpha(Vector4f(1, 1, 1, 1));
    }
    page.setBackgroundMaterial(newMaterial);
}
DEFFC(BGMaterial) {
    DENG2_UNUSED(cmd);
    world::Material *material = nullptr;
    try {
        if (ded_value_t *value = DED_Definitions()->getValueByUri(*reinterpret_cast<de::Uri const *>(OP_URI(0)))) {
            material = &world::Materials::get().material(de::makeUri(value->text));
        }
        else {
            material = &world::Materials::get().material(*reinterpret_cast<de::Uri const *>(OP_URI(0)));
        }
    }
    catch (world::MaterialManifest::MissingMaterialError const &)
    {} // Ignore this error.
    catch (Resources::MissingResourceManifestError const &)
    {} // Ignore this error.
    changePageBackground(fi.page(FinaleInterpreter::Anims), material);
}
DEFFC(NoBGMaterial) {
    DENG2_UNUSED2(cmd, ops);
    changePageBackground(fi.page(FinaleInterpreter::Anims), 0);
}
DEFFC(InTime) {
    DENG2_UNUSED(cmd);
    fi.setInTime(FRACSECS_TO_TICKS(OP_FLOAT(0)));
}
DEFFC(Tic) {
    DENG2_UNUSED2(cmd, ops);
    fi.wait();
}
DEFFC(Wait) {
    DENG2_UNUSED(cmd);
    fi.wait(FRACSECS_TO_TICKS(OP_FLOAT(0)));
}
DEFFC(WaitText) {
    DENG2_UNUSED(cmd);
    fi.setWaitText(&fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>());
}
DEFFC(WaitAnim) {
    DENG2_UNUSED(cmd);
    fi.setWaitAnim(&fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>());
}
DEFFC(Color) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Anims)
            .setBackgroundTopColor   (Vector3f(OP_FLOAT(0), OP_FLOAT(1), OP_FLOAT(2)), fi.inTime())
            .setBackgroundBottomColor(Vector3f(OP_FLOAT(0), OP_FLOAT(1), OP_FLOAT(2)), fi.inTime());
}
DEFFC(ColorAlpha) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Anims)
            .setBackgroundTopColorAndAlpha   (Vector4f(OP_FLOAT(0), OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime())
            .setBackgroundBottomColorAndAlpha(Vector4f(OP_FLOAT(0), OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
}
DEFFC(Pause) {
    DENG2_UNUSED2(cmd, ops);
    fi.pause();
}
DEFFC(CanSkip) {
    DENG2_UNUSED2(cmd, ops);
    fi.setSkip(true);
}
DEFFC(NoSkip) {
    DENG2_UNUSED2(cmd, ops);
    fi.setSkip(false);
}
DEFFC(SkipHere) {
    DENG2_UNUSED2(cmd, ops);
    fi.foundSkipHere();
}
DEFFC(Events) {
    DENG2_UNUSED2(cmd, ops);
    fi.setHandleEvents();
}
DEFFC(NoEvents) {
    DENG2_UNUSED2(cmd, ops);
    fi.setHandleEvents(false);
}
DEFFC(OnKey) {
#ifdef __CLIENT__
    DENG2_UNUSED(cmd);
    ddevent_t ev; de::zap(ev);
    ev.device = IDEV_KEYBOARD;
    ev.type   = E_TOGGLE;
    ev.toggle.id    = DD_GetKeyCode(OP_CSTRING(0));
    ev.toggle.state = ETOG_DOWN;
    fi.addEventHandler(ev, OP_CSTRING(1));
#else
    DENG2_UNUSED3(cmd, ops, fi);
#endif
}
DEFFC(UnsetKey) {
#ifdef __CLIENT__
    DENG2_UNUSED(cmd);
    ddevent_t ev; de::zap(ev);
    ev.device = IDEV_KEYBOARD;
    ev.type   = E_TOGGLE;
    ev.toggle.id    = DD_GetKeyCode(OP_CSTRING(0));
    ev.toggle.state = ETOG_DOWN;
    fi.removeEventHandler(ev);
#else
    DENG2_UNUSED3(cmd, ops, fi);
#endif
}
DEFFC(If) {
    DENG2_UNUSED2(cmd, ops);
    LOG_AS("FIC_If");
    char const *token = OP_CSTRING(0);
    bool val          = false;
    if (!qstricmp(token, "netgame")) {
        val = netGame;
    }
    else if (!qstrnicmp(token, "mode:", 5)) {
        if (App_GameLoaded())
            val = !String(token + 5).compareWithoutCase(App_CurrentGame().id());
        else
            val = 0;
    }
    else if (Plug_CheckForHook(HOOK_FINALE_EVAL_IF)) {
        ddhook_finale_script_evalif_paramaters_t p; de::zap(p);
        p.token     = token;
        p.returnVal = 0;
        if (DoomsdayApp::plugins().callAllHooks(HOOK_FINALE_EVAL_IF, fi.id(), (void *) &p)) {
            val = p.returnVal;
            LOG_SCR_XVERBOSE("HOOK_FINALE_EVAL_IF: %s => %i", token << val);
        }
        else {
            LOG_SCR_XVERBOSE("HOOK_FINALE_EVAL_IF: no hook (for %s)", token);
        }
    }
    else {
        LOG_SCR_WARNING("Unknown condition '%s'") << token;
    }
    fi.setSkipNext(!val);
}
DEFFC(IfNot) {
    FIC_If(cmd, ops, fi);
    fi.setSkipNext(!fi.skipInProgress());
}
DEFFC(Else) {
    DENG2_UNUSED2(cmd, ops);
    fi.setSkipNext(!fi.lastSkipped());
}
DEFFC(GoTo) {
    DENG2_UNUSED(cmd);
    fi.skipToMarker(OP_CSTRING(0));
}
DEFFC(Marker) {
    DENG2_UNUSED(cmd);
    fi.foundSkipMarker(OP_CSTRING(0));
}
DEFFC(Delete) {
    DENG2_UNUSED(cmd);
    delete fi.tryFindWidget(OP_CSTRING(0));
}
DEFFC(Image) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_Image");
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    anim.clearAllFrames();
#ifdef __CLIENT__
    char const *name  = OP_CSTRING(1);
    lumpnum_t lumpNum = App_FileSystem().lumpNumForName(name);
    if (rawtex_t *rawTex = ClientResources::get().declareRawTexture(lumpNum)) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_RAW, -1, &rawTex->lumpNum, 0, false);
        return;
    }
    LOG_SCR_WARNING("Missing lump '%s'") << name;
#endif
}
DEFFC(ImageAt) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_ImageAt");
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    float x = OP_FLOAT(1);
    float y = OP_FLOAT(2);
    anim.clearAllFrames()
        .setOrigin(Vector2f(x, y));
#ifdef __CLIENT__
    char const *name  = OP_CSTRING(3);
    lumpnum_t lumpNum = App_FileSystem().lumpNumForName(name);
    if (rawtex_t *rawTex = App_Resources().declareRawTexture(lumpNum)) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_RAW, -1, &rawTex->lumpNum, 0, false);
        return;
    }
    LOG_SCR_WARNING("Missing lump '%s'") << name;
#endif
}
#ifdef __CLIENT__
static DGLuint loadAndPrepareExtTexture(char const *fileName) {
    image_t image;
    DGLuint glTexName = 0;
    if (GL_LoadExtImage(image, fileName, LGM_NORMAL)) {
        glTexName = GL_NewTextureWithParams(
            ( image.pixelSize == 2 ? DGL_LUMINANCE_PLUS_A8 :
              image.pixelSize == 3 ? DGL_RGB :
              image.pixelSize == 4 ? DGL_RGBA : DGL_LUMINANCE ),
            image.size.x, image.size.y, image.pixels,
            (image.size.x < 128 && image.size.y < 128? TXCF_NO_COMPRESSION : 0),
            0, GL_LINEAR, GL_LINEAR, 0 /*no anisotropy*/,
            GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
        Image_ClearPixelData(image);
    }
    return glTexName;
}
#endif // __CLIENT__
DEFFC(XImage) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_XImage");
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
#ifdef __CLIENT__
    char const *fileName  = OP_CSTRING(1);
#endif
    anim.clearAllFrames();
#ifdef __CLIENT__
    if (DGLuint tex = loadAndPrepareExtTexture(fileName)) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_XIMAGE, -1, &tex, 0, false);
    }
    else {
        LOG_SCR_WARNING("Missing graphic '%s'") << fileName;
    }
#endif // __CLIENT__
}
DEFFC(Patch) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_Patch");
    FinaleAnimWidget &anim  = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    char const *encodedName = OP_CSTRING(3);
    anim.setOrigin(Vector2f(OP_FLOAT(1), OP_FLOAT(2)));
    anim.clearAllFrames();
    patchid_t patchId = R_DeclarePatch(encodedName);
    if (patchId) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_PATCH, -1, (void *)&patchId, 0, 0);
    }
    else {
        LOG_SCR_WARNING("Missing Patch '%s'") << encodedName;
    }
}
DEFFC(SetPatch) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_SetPatch");
    FinaleAnimWidget &anim  = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    char const *encodedName = OP_CSTRING(1);
    patchid_t patchId = R_DeclarePatch(encodedName);
    if (patchId == 0) {
        LOG_SCR_WARNING("Missing Patch '%s'") << encodedName;
        return;
    }
    if (!anim.frameCount()) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_PATCH, -1, (void *)&patchId, 0, false);
        return;
    }
    FinaleAnimWidget::Frame *f = anim.allFrames().first();
    f->type  = FinaleAnimWidget::Frame::PFT_PATCH;
    f->texRef.patch = patchId;
    f->tics  = -1;
    f->sound = 0;
}
DEFFC(ClearAnim) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        if (FinaleAnimWidget *anim = maybeAs<FinaleAnimWidget>(wi)) {
            anim->clearAllFrames();
        }
    }
}
DEFFC(Anim) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_Anim");
    FinaleAnimWidget &anim  = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    char const *encodedName = OP_CSTRING(1);
    int const tics          = FRACSECS_TO_TICKS(OP_FLOAT(2));
    patchid_t patchId = R_DeclarePatch(encodedName);
    if (!patchId) {
        LOG_SCR_WARNING("Patch '%s' not found") << encodedName;
        return;
    }
    anim.newFrame(FinaleAnimWidget::Frame::PFT_PATCH, tics, (void *)&patchId, 0, false);
}
DEFFC(AnimImage) {
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_AnimImage");
    FinaleAnimWidget &anim  = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
#ifdef __CLIENT__
    char const *encodedName = OP_CSTRING(1);
    int const tics          = FRACSECS_TO_TICKS(OP_FLOAT(2));
    lumpnum_t lumpNum       = App_FileSystem().lumpNumForName(encodedName);
    if (rawtex_t *rawTex = App_Resources().declareRawTexture(lumpNum)) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_RAW, tics, &rawTex->lumpNum, 0, false);
        return;
    }
    LOG_SCR_WARNING("Lump '%s' not found") << encodedName;
#else
    DENG2_UNUSED(anim);
#endif
}
DEFFC(Repeat) {
    DENG2_UNUSED(cmd);
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    anim.setLooping();
}
DEFFC(StateAnim) {
    DENG2_UNUSED(cmd);
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
#if !defined(__CLIENT__)
    DENG2_UNUSED(anim);
#endif
    dint stateId = DED_Definitions()->getStateNum(OP_CSTRING(1));
    for (dint count = OP_INT(2); count > 0 && stateId > 0; count--) {
        state_t *st = &runtimeDefs.states[stateId];
#ifdef __CLIENT__
        spriteinfo_t sinf;
        R_GetSpriteInfo(st->sprite, st->frame & 0x7fff, &sinf);
        anim.newFrame(FinaleAnimWidget::Frame::PFT_MATERIAL, (st->tics <= 0? 1 : st->tics), sinf.material, 0, sinf.flip);
#endif
        stateId = st->nextState;
    }
}
DEFFC(PicSound) {
    DENG2_UNUSED(cmd);
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    int const sound        = DED_Definitions()->getSoundNum(OP_CSTRING(1));
    if (!anim.frameCount()) {
        anim.newFrame(FinaleAnimWidget::Frame::PFT_MATERIAL, -1, 0, sound, false);
        return;
    }
    anim.allFrames().at(anim.frameCount() - 1)->sound = sound;
}
DEFFC(ObjectOffX) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setOriginX(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectOffY) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setOriginY(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectOffZ) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setOriginZ(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectRGB) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        Vector3f const color(OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3));
        if (FinaleTextWidget *text = maybeAs<FinaleTextWidget>(wi)) {
            text->setColor(color, fi.inTime());
        }
        if (FinaleAnimWidget *anim = maybeAs<FinaleAnimWidget>(wi)) {
            anim->setColor         (color, fi.inTime())
                 .setEdgeColor     (color, fi.inTime())
                 .setOtherColor    (color, fi.inTime())
                 .setOtherEdgeColor(color, fi.inTime());
        }
    }
}
DEFFC(ObjectAlpha) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        float const alpha = OP_FLOAT(1);
        if (FinaleTextWidget *text = maybeAs<FinaleTextWidget>(wi)) {
            text->setAlpha(alpha, fi.inTime());
        }
        if (FinaleAnimWidget *anim = maybeAs<FinaleAnimWidget>(wi)) {
            anim->setAlpha     (alpha, fi.inTime())
                 .setOtherAlpha(alpha, fi.inTime());
        }
    }
}
DEFFC(ObjectScaleX) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScaleX(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectScaleY) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScaleY(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectScaleZ) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScaleZ(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectScale) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScaleX(OP_FLOAT(1), fi.inTime())
           .setScaleY(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(ObjectScaleXY) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScaleX(OP_FLOAT(1), fi.inTime())
           .setScaleY(OP_FLOAT(2), fi.inTime());
    }
}
DEFFC(ObjectScaleXYZ) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setScale(Vector3f(OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
    }
}
DEFFC(ObjectAngle) {
    DENG2_UNUSED(cmd);
    if (FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0))) {
        wi->setAngle(OP_FLOAT(1), fi.inTime());
    }
}
DEFFC(Rect) {
    DENG2_UNUSED(cmd);
    FinaleAnimWidget &anim = fi.findOrCreateWidget(FI_ANIM, OP_CSTRING(0)).as<FinaleAnimWidget>();
    anim.clearAllFrames()
        .resetAllColors()
        .setLooping(false) // Yeah?
        .setOrigin(Vector3f(OP_FLOAT(1), OP_FLOAT(2), 0))
        .setScale(Vector3f(OP_FLOAT(3), OP_FLOAT(4), 1));
}
DEFFC(FillColor) {
    DENG2_UNUSED(cmd);
    FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0));
    if (!wi || !is<FinaleAnimWidget>(wi)) return;
    FinaleAnimWidget &anim = wi->as<FinaleAnimWidget>();
    int which = 0;
    if (!qstricmp(OP_CSTRING(1), "top"))         which |= 1;
    else if (!qstricmp(OP_CSTRING(1), "bottom")) which |= 2;
    else                                         which = 3;
    Vector4f color;
    for (int i = 0; i < 4; ++i) {
        color[i] = OP_FLOAT(2 + i);
    }
    if (which & 1)
        anim.setColorAndAlpha(color, fi.inTime());
    if (which & 2)
        anim.setOtherColorAndAlpha(color, fi.inTime());
}
DEFFC(EdgeColor) {
    DENG2_UNUSED(cmd);
    FinaleWidget *wi = fi.tryFindWidget(OP_CSTRING(0));
    if (!wi || !is<FinaleAnimWidget>(wi)) return;
    FinaleAnimWidget &anim = wi->as<FinaleAnimWidget>();
    int which = 0;
    if (!qstricmp(OP_CSTRING(1), "top"))         which |= 1;
    else if (!qstricmp(OP_CSTRING(1), "bottom")) which |= 2;
    else                                        which = 3;
    Vector4f color;
    for (int i = 0; i < 4; ++i) {
        color[i] = OP_FLOAT(2 + i);
    }
    if (which & 1)
        anim.setEdgeColorAndAlpha(color, fi.inTime());
    if (which & 2)
        anim.setOtherEdgeColorAndAlpha(color, fi.inTime());
}
DEFFC(OffsetX) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Anims).setOffsetX(OP_FLOAT(0), fi.inTime());
}
DEFFC(OffsetY) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Anims).setOffsetY(OP_FLOAT(0), fi.inTime());
}
DEFFC(Sound) {
    DENG2_UNUSED2(cmd, fi);
    S_LocalSound(DED_Definitions()->getSoundNum(OP_CSTRING(0)), nullptr);
}
DEFFC(SoundAt) {
    DENG2_UNUSED2(cmd, fi);
    dint const soundId = DED_Definitions()->getSoundNum(OP_CSTRING(0));
    dfloat const vol   = de::min(OP_FLOAT(1), 1.f);
    S_LocalSoundAtVolume(soundId, nullptr, vol);
}
DEFFC(SeeSound) {
    DENG2_UNUSED2(cmd, fi);
    dint num = DED_Definitions()->getMobjNum(OP_CSTRING(0));
    if (num >= 0 && ::runtimeDefs.mobjInfo[num].seeSound > 0) {
        S_LocalSound(runtimeDefs.mobjInfo[num].seeSound, nullptr);
    }
}
DEFFC(DieSound) {
    DENG2_UNUSED2(cmd, fi);
    dint num = DED_Definitions()->getMobjNum(OP_CSTRING(0));
    if (num >= 0 && ::runtimeDefs.mobjInfo[num].deathSound > 0) {
        S_LocalSound(runtimeDefs.mobjInfo[num].deathSound, nullptr);
    }
}
DEFFC(Music) {
    DENG2_UNUSED3(cmd, ops, fi);
    S_StartMusic(OP_CSTRING(0), true);
}
DEFFC(MusicOnce) {
    DENG2_UNUSED3(cmd, ops, fi);
    S_StartMusic(OP_CSTRING(0), false);
}
DEFFC(Filter) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Texts).setFilterColorAndAlpha(Vector4f(OP_FLOAT(0), OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
}
DEFFC(Text) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setText(OP_CSTRING(3))
        .setCursorPos(0) // Restart the text.
        .setOrigin(Vector3f(OP_FLOAT(1), OP_FLOAT(2), 0));
}
DEFFC(TextFromDef) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    int textIdx          = DED_Definitions()->getTextNum((char *)OP_CSTRING(3));
    text.setText(textIdx >= 0? DED_Definitions()->text[textIdx].text : "(undefined)")
        .setCursorPos(0) // Restart the type-in animation (if any).
        .setOrigin(Vector3f(OP_FLOAT(1), OP_FLOAT(2), 0));
}
DEFFC(TextFromLump) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setOrigin(Vector3f(OP_FLOAT(1), OP_FLOAT(2), 0));
    lumpnum_t lumpNum = App_FileSystem().lumpNumForName(OP_CSTRING(3));
    if (lumpNum >= 0) {
        File1 &lump            = App_FileSystem().lump(lumpNum);
        uint8_t const *rawStr = lump.cache();
        AutoStr *str = AutoStr_NewStd();
        Str_Reserve(str, lump.size() * 2);
        char *out = Str_Text(str);
        for (size_t i = 0; i < lump.size(); ++i) {
            char ch = (char)(rawStr[i]);
            if (ch == '\r') continue;
            if (ch == '\n') {
                *out++ = '\\';
                *out++ = 'n';
            }
            else {
                *out++ = ch;
            }
        }
        lump.unlock();
        text.setText(Str_Text(str));
    }
    else {
        text.setText("(not found)");
    }
    text.setCursorPos(0); // Restart.
}
DEFFC(SetText) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setText(OP_CSTRING(1));
}
DEFFC(SetTextDef) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    int textIdx            = DED_Definitions()->getTextNum((char *)OP_CSTRING(1));
    text.setText(textIdx >= 0? DED_Definitions()->text[textIdx].text : "(undefined)");
}
DEFFC(DeleteText) {
    DENG2_UNUSED(cmd);
    delete fi.tryFindWidget(OP_CSTRING(0));
}
DEFFC(PredefinedColor) {
    DENG2_UNUSED(cmd);
    fi.page(FinaleInterpreter::Texts)
            .setPredefinedColor(de::clamp(1, OP_INT(0), FIPAGE_NUM_PREDEFINED_COLORS) - 1,
                                Vector3f(OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
    fi.page(FinaleInterpreter::Anims)
            .setPredefinedColor(de::clamp(1, OP_INT(0), FIPAGE_NUM_PREDEFINED_COLORS) - 1,
                                Vector3f(OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
}
DEFFC(PredefinedFont) {
#ifdef __CLIENT__
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_PredefinedFont");
    fontid_t const fontNum = Fonts_ResolveUri(OP_URI(1));
    if (fontNum) {
        int const idx = de::clamp(1, OP_INT(0), FIPAGE_NUM_PREDEFINED_FONTS) - 1;
        fi.page(FinaleInterpreter::Anims).setPredefinedFont(idx, fontNum);
        fi.page(FinaleInterpreter::Texts).setPredefinedFont(idx, fontNum);
        return;
    }
    AutoStr *fontPath = Uri_ToString(OP_URI(1));
    LOG_SCR_WARNING("Unknown font '%s'") << Str_Text(fontPath);
#else
    DENG2_UNUSED3(cmd, ops, fi);
#endif
}
DEFFC(TextRGB) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setColor(Vector3f(OP_FLOAT(1), OP_FLOAT(2), OP_FLOAT(3)), fi.inTime());
}
DEFFC(TextAlpha) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setAlpha(OP_FLOAT(1), fi.inTime());
}
DEFFC(TextOffX) {
    DENG2_UNUSED(cmd);
    FinaleWidget &wi = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0));
    wi.setOriginX(OP_FLOAT(1), fi.inTime());
}
DEFFC(TextOffY) {
    DENG2_UNUSED(cmd);
    FinaleWidget &wi = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0));
    wi.setOriginY(OP_FLOAT(1), fi.inTime());
}
DEFFC(TextCenter) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setAlignment(text.alignment() & ~(ALIGN_LEFT | ALIGN_RIGHT));
}
DEFFC(TextNoCenter) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setAlignment(text.alignment() | ALIGN_LEFT);
}
DEFFC(TextScroll) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setScrollRate(OP_INT(1));
}
DEFFC(TextPos) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setCursorPos(OP_INT(1));
}
DEFFC(TextRate) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setTypeInRate(OP_INT(1));
}
DEFFC(TextLineHeight) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setLineHeight(OP_FLOAT(1));
}
DEFFC(Font) {
#ifdef __CLIENT__
    DENG2_UNUSED(cmd);
    LOG_AS("FIC_Font");
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    fontid_t fontNum       = Fonts_ResolveUri(OP_URI(1));
    if (fontNum) {
        text.setFont(fontNum);
        return;
    }
    AutoStr *fontPath = Uri_ToString(OP_URI(1));
    LOG_SCR_WARNING("Unknown font '%s'") << Str_Text(fontPath);
#else
    DENG2_UNUSED3(cmd, ops, fi);
#endif
}
DEFFC(FontA) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setFont(fi.page(FinaleInterpreter::Texts).predefinedFont(0));
}
DEFFC(FontB) {
    DENG2_UNUSED(cmd);
    FinaleTextWidget &text = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0)).as<FinaleTextWidget>();
    text.setFont(fi.page(FinaleInterpreter::Texts).predefinedFont(1));
}
DEFFC(NoMusic) {
    DENG2_UNUSED3(cmd, ops, fi);
    S_StopMusic();
}
DEFFC(TextScaleX) {
    DENG2_UNUSED(cmd);
    FinaleWidget &wi = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0));
    wi.setScaleX(OP_FLOAT(1), fi.inTime());
}
DEFFC(TextScaleY) {
    DENG2_UNUSED(cmd);
    FinaleWidget &wi = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0));
    wi.setScaleY(OP_FLOAT(1), fi.inTime());
}
DEFFC(TextScale) {
    DENG2_UNUSED(cmd);
    FinaleWidget &wi = fi.findOrCreateWidget(FI_TEXT, OP_CSTRING(0));
    wi.setScaleX(OP_FLOAT(1), fi.inTime())
      .setScaleY(OP_FLOAT(2), fi.inTime());
}
DEFFC(PlayDemo) {
#if 0
    fi.suspend();
    if (!Con_Executef(CMDS_DDAY, true, "playdemo \"%s\"", OP_CSTRING(0))) {
        fi.resume();
    }
#else
    DENG2_UNUSED3(cmd, ops, fi);
#endif
}
DEFFC(Command) {
    DENG2_UNUSED2(cmd, fi);
    Con_Executef(CMDS_SCRIPT, false, "%s", OP_CSTRING(0));
}
DEFFC(ShowMenu) {
    DENG2_UNUSED2(cmd, ops);
    fi.setShowMenu();
}
DEFFC(NoShowMenu) {
    DENG2_UNUSED2(cmd, ops);
    fi.setShowMenu(false);
}
DENG2_PIMPL_NOREF(FinalePageWidget)
, DENG2_OBSERVES(FinaleWidget, Deletion) {
    struct Flags {
        char hidden:1;         ///< Not drawn.
        char paused:1;         ///< Does not tick.
        char showBackground:1;
    } flags;
    Children children;         ///< Child widgets (owned).
    uint timer = 0;
    animatorvector3_t offset;  ///< Offset the world origin.
    animatorvector4_t filter;
    animatorvector3_t preColor[FIPAGE_NUM_PREDEFINED_COLORS];
    fontid_t preFont[FIPAGE_NUM_PREDEFINED_FONTS];
    struct Background {
        world::Material *material;
        animatorvector4_t topColor;
        animatorvector4_t bottomColor;
        Background() : material(nullptr) {
            AnimatorVector4_Init(topColor,    1, 1, 1, 0);
            AnimatorVector4_Init(bottomColor, 1, 1, 1, 0);
        }
    } bg;
    Impl() {
        de::zap(flags);
        flags.showBackground = true; /// Draw background by default.
        AnimatorVector3_Init(offset, 0, 0, 0);
        AnimatorVector4_Init(filter, 0, 0, 0, 0);
        for (int i = 0; i < FIPAGE_NUM_PREDEFINED_COLORS; ++i) {
            AnimatorVector3_Init(preColor[i], 1, 1, 1);
        }
        de::zap(preFont);
    }
    ~Impl() {
        qDeleteAll(children);
        DENG2_ASSERT(children.isEmpty());
    }
    void finaleWidgetBeingDeleted(FinaleWidget const &widget) {
        DENG2_ASSERT(children.contains(&const_cast<FinaleWidget &>(widget)));
        children.removeOne(&const_cast<FinaleWidget &>(widget));
    }
};
FinalePageWidget::FinalePageWidget() : d(new Impl)
{}
FinalePageWidget::~FinalePageWidget()
{}
#ifdef __CLIENT__
static inline MaterialVariantSpec const &uiMaterialSpec() {
    return App_Resources().materialSpec(UiContext, 0, 0, 0, 0,
                                             GL_REPEAT, GL_REPEAT, 0, 1, 0,
                                             false, false, false, false);
}
void FinalePageWidget::draw() const {
    if (d->flags.hidden) return;
    if (d->flags.showBackground) {
        vec3f_t topColor; V3f_Set(topColor, d->bg.topColor[0].value, d->bg.topColor[1].value, d->bg.topColor[2].value);
        float topAlpha = d->bg.topColor[3].value;
        vec3f_t bottomColor; V3f_Set(bottomColor, d->bg.bottomColor[0].value, d->bg.bottomColor[1].value, d->bg.bottomColor[2].value);
        float bottomAlpha = d->bg.bottomColor[3].value;
        if (topAlpha > 0 && bottomAlpha > 0) {
            if (world::Material *material = d->bg.material) {
                MaterialAnimator &matAnimator = material->as<ClientMaterial>().getAnimator(uiMaterialSpec());
                matAnimator.prepare();
                GL_BindTexture(matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture);
                DGL_Enable(DGL_TEXTURE_2D);
            }
            if (d->bg.material || topAlpha < 1.0 || bottomAlpha < 1.0) {
                GL_BlendMode(BM_NORMAL);
            }
            else {
                DGL_Disable(DGL_BLEND);
            }
            GL_DrawRectf2TextureColor(0, 0, SCREENWIDTH, SCREENHEIGHT, 64, 64,
                                      topColor, topAlpha, bottomColor, bottomAlpha);
            GL_SetNoTexture();
            DGL_Enable(DGL_BLEND);
        }
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    LIBGUI_GL.glClear(GL_DEPTH_BUFFER_BIT);
#if defined (DENG_OPENGL)
    if (renderWireframe > 1) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    }
#endif
    DGL_PushState();
    DGL_Enable(DGL_ALPHA_TEST);
    Vector3f worldOrigin(/*-SCREENWIDTH/2*/ - d->offset[VX].value,
                         /*-SCREENHEIGHT/2*/ - d->offset[VY].value,
                         0/*.05f - d->offset[VZ].value*/);
    for (FinaleWidget *widget : d->children) {
        widget->draw(worldOrigin);
    }
    DGL_PopState();
#if defined (DENG_OPENGL)
    if (renderWireframe > 1) {
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
#endif
    if (d->filter[3].value > 0) {
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_LoadIdentity();
        DGL_MatrixMode(DGL_PROJECTION);
        DGL_PushMatrix();
        DGL_LoadIdentity();
        DGL_Ortho(0, 0, 1, 1, 0, 1);
        DGL_PushState();
        DGL_Disable(DGL_SCISSOR_TEST); // set by bordered projection mode
        GL_DrawRectf2Color(0,
                           0,
                           1,
                           1,
                           d->filter[0].value,
                           d->filter[1].value,
                           d->filter[2].value,
                           d->filter[3].value);
        DGL_PopState();
        DGL_MatrixMode(DGL_PROJECTION);
        DGL_PopMatrix();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
#endif
void FinalePageWidget::makeVisible(bool yes) {
    d->flags.hidden = !yes;
}
void FinalePageWidget::pause(bool yes) {
    d->flags.paused = yes;
}
void FinalePageWidget::runTicks(timespan_t /*timeDelta*/)
{
    if (!DD_IsSharpTick()) return;
    d->timer++;
    for (FinaleWidget *widget : d->children) {
        widget->runTicks(/*timeDelta*/);
    }
    AnimatorVector3_Think(d->offset);
    AnimatorVector4_Think(d->bg.topColor);
    AnimatorVector4_Think(d->bg.bottomColor);
    AnimatorVector4_Think(d->filter);
    for (int i = 0; i < FIPAGE_NUM_PREDEFINED_COLORS; ++i) {
        AnimatorVector3_Think(d->preColor[i]);
    }
}
bool FinalePageWidget::hasWidget(FinaleWidget *widget) {
    if (!widget) return false;
    return d->children.contains(widget);
}
FinaleWidget *FinalePageWidget::addChild(FinaleWidget *widgetToAdd) {
    if (!hasWidget(widgetToAdd)) {
        d->children.append(widgetToAdd);
        widgetToAdd->setPage(this);
        widgetToAdd->audienceForDeletion += d;
    }
    return widgetToAdd;
}
FinaleWidget *FinalePageWidget::removeChild(FinaleWidget *widgetToRemove) {
    if (hasWidget(widgetToRemove)) {
        widgetToRemove->audienceForDeletion -= d;
        widgetToRemove->setPage(nullptr);
        d->children.removeOne(widgetToRemove);
    }
    return widgetToRemove;
}
FinalePageWidget::Children const &FinalePageWidget::children() const {
    return d->children;
}
world::Material *FinalePageWidget::backgroundMaterial() const {
    return d->bg.material;
}
FinalePageWidget &FinalePageWidget::setBackgroundMaterial(world::Material *newMaterial) {
    d->bg.material = newMaterial;
    return *this;
}
FinalePageWidget &FinalePageWidget::setBackgroundTopColor(Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->bg.topColor, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setBackgroundTopColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->bg.topColor, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setBackgroundBottomColor(Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(d->bg.bottomColor, newColor.x, newColor.y, newColor.z, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setBackgroundBottomColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->bg.bottomColor, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setOffset(Vector3f const &newOffset, int steps) {
    AnimatorVector3_Set(d->offset, newOffset.x, newOffset.y, newOffset.z, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setOffsetX(float x, int steps) {
    Animator_Set(&d->offset[VX], x, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setOffsetY(float y, int steps) {
    Animator_Set(&d->offset[VY], y, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setOffsetZ(float y, int steps) {
    Animator_Set(&d->offset[VZ], y, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setFilterColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->filter, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    return *this;
}
FinalePageWidget &FinalePageWidget::setPredefinedColor(uint idx, Vector3f const &newColor, int steps) {
    if (VALID_FIPAGE_PREDEFINED_COLOR(idx)) {
        AnimatorVector3_Set(d->preColor[idx], newColor.x, newColor.y, newColor.z, steps);
    }
    else {
        throw InvalidColorError("FinalePageWidget::setPredefinedColor", "Invalid color #" + String::number(idx));
    }
    return *this;
}
animatorvector3_t const *FinalePageWidget::predefinedColor(uint idx) {
    if (VALID_FIPAGE_PREDEFINED_COLOR(idx)) {
        return &d->preColor[idx];
    }
    throw InvalidColorError("FinalePageWidget::predefinedColor", "Invalid color #" + String::number(idx));
}
FinalePageWidget &FinalePageWidget::setPredefinedFont(uint idx, fontid_t fontNum) {
    if (VALID_FIPAGE_PREDEFINED_FONT(idx)) {
        d->preFont[idx] = fontNum;
    }
    else {
        throw InvalidFontError("FinalePageWidget::setPredefinedFont", "Invalid font #" + String::number(idx));
    }
    return *this;
}
fontid_t FinalePageWidget::predefinedFont(uint idx) {
    if (VALID_FIPAGE_PREDEFINED_FONT(idx)) {
        return d->preFont[idx];
    }
    throw InvalidFontError("FinalePageWidget::predefinedFont", "Invalid font #" + String::number(idx));
}
DENG2_PIMPL_NOREF(FinaleTextWidget) {
    animatorvector4_t color;
    uint pageColor    = 0;                ///< 1-based page color index. @c 0= Use our own color.
    uint pageFont     = 0;                ///< 1-based page font index. @c 0= Use our own font.
    int alignFlags    = ALIGN_TOPLEFT;    ///< @ref alignmentFlags
    short textFlags   = DTF_ONLY_SHADOW;  ///< @ref drawTextFlags
    int scrollWait    = 0;
    int scrollTimer   = 0;                ///< Automatic scrolling upwards.
    int cursorPos     = 0;
    int wait          = 3;
    int timer         = 0;
    float lineHeight  = 11.f / 7 - 1;
    fontid_t fontNum  = 0;
    char *text        = nullptr;
    bool animComplete = true;
    Impl()  { AnimatorVector4_Init(color, 1, 1, 1, 1); }
    ~Impl() { Z_Free(text); }
#ifdef __CLIENT__
    static int textLineWidth(char const *text) {
        int width = 0;
        for (; *text; text++) {
            if (*text == '\n') break;
            if (*text == '\\') {
                if (!*++text)     break;
                if (*text == 'n') break;
                if (*text >= '0' && *text <= '9')
                    continue;
                if (*text == 'w' || *text == 'W' || *text == 'p' || *text == 'P')
                    continue;
            }
            width += FR_CharWidth(*text);
        }
        return width;
    }
#endif
};
FinaleTextWidget::FinaleTextWidget(String const &name)
    : FinaleWidget(name)
    , d(new Impl)
{}
FinaleTextWidget::~FinaleTextWidget()
{}
void FinaleTextWidget::accelerate() {
    d->wait = -10;
}
FinaleTextWidget &FinaleTextWidget::setCursorPos(int newPos) {
    d->cursorPos = de::max(0, newPos);
    return *this;
}
void FinaleTextWidget::runTicks(/*timespan_t timeDelta*/)
{
    FinaleWidget::runTicks(/*timeDelta*/);
    AnimatorVector4_Think(d->color);
    if (d->wait) {
        if (--d->timer <= 0) {
            if (d->wait > 0) {
                d->cursorPos++;
                d->timer = d->wait;
            }
            else {
                d->cursorPos += ABS(d->wait);
                d->timer = 1;
            }
        }
    }
    if (d->scrollWait) {
        if (--d->scrollTimer <= 0) {
            d->scrollTimer = d->scrollWait;
            setOriginY(origin()[1].target - 1, d->scrollWait);
        }
    }
    d->animComplete = (!d->wait || d->cursorPos >= visLength());
}
#ifdef __CLIENT__
void FinaleTextWidget::draw(Vector3f const &offset) {
    if (!d->text) return;
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(origin()[0].value + offset.x, origin()[1].value + offset.y, origin()[2].value + offset.z);
    if (angle().value != 0) {
        DGL_Scalef(1, 200.0f / 240.0f, 1);
        DGL_Rotatef(angle().value, 0, 0, 1);
        DGL_Scalef(1, 240.0f / 200.0f, 1);
    }
    DGL_Scalef(scale()[0].value, scale()[1].value, scale()[2].value);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(d->pageFont? page()->predefinedFont(d->pageFont - 1) : d->fontNum);
    animatorvector3_t const *color;
    if (d->pageColor == 0)
        color = (animatorvector3_t const *)&d->color;
    else
        color = page()->predefinedColor(d->pageColor - 1);
    FR_SetColor((*color)[CR].value, (*color)[CG].value, (*color)[CB].value);
    FR_SetAlpha(d->color[CA].value);
    int x = 0, y = 0, ch, linew = -1;
    char *ptr = d->text;
    for (int cnt = 0; *ptr && (!d->wait || cnt < d->cursorPos); ptr++) {
        if (linew < 0)
            linew = d->textLineWidth(ptr);
        ch = *ptr;
        if (*ptr == '\n') // Newline? (see below for unescaped \n)
        {
            x = 0;
            y += FR_CharHeight('A') * (1 + d->lineHeight);
            linew = -1;
            cnt++; // Include newlines in the wait count.
            continue;
        }
        if (*ptr == '\\') // Escape?
        {
            if (!*++ptr)
                break;
            if (*ptr >= '0' && *ptr <= '9') {
                uint colorIdx = *ptr - '0';
                if (colorIdx == 0)
                    color = (animatorvector3_t const *)&d->color;
                else
                    color = page()->predefinedColor(colorIdx - 1);
                FR_SetColor((*color)[CR].value, (*color)[CG].value, (*color)[CB].value);
                FR_SetAlpha(d->color[CA].value);
                continue;
            }
            if (*ptr == 'w' || *ptr == 'W') // Wait?
            {
                if (d->wait)
                    cnt += int(float(TICRATE) / d->wait / (*ptr == 'w' ? 2 : 1));
                continue;
            }
            if (*ptr == 'p' || *ptr == 'P') // Longer pause?
            {
                if (d->wait)
                    cnt += int(float(TICRATE) / d->wait * (*ptr == 'p' ? 5 : 10));
                continue;
            }
            if (*ptr == 'n' || *ptr == 'N') // Newline?
            {
                x = 0;
                y += FR_CharHeight('A') * (1 + d->lineHeight);
                linew = -1;
                cnt++; // Include newlines in the wait count.
                continue;
            }
            if (*ptr == '_')
                ch = ' ';
        }
        if (scale()[1].value * y + origin()[1].value >= -scale()[1].value * d->lineHeight &&
           scale()[1].value * y + origin()[1].value < SCREENHEIGHT) {
            FR_DrawCharXY(ch, (d->alignFlags & ALIGN_LEFT) ? x : x - linew / 2, y);
            x += FR_CharWidth(ch);
        }
        ++cnt;
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
#endif
bool FinaleTextWidget::animationComplete() const {
    return d->animComplete;
}
int FinaleTextWidget::visLength() {
    int count = 0;
    if (d->text) {
        float const secondLen = (d->wait? TICRATE / d->wait : 0);
        for (char const *ptr = d->text; *ptr; ptr++) {
            if (*ptr == '\\') // Escape?
            {
                if (!*++ptr) break;
                switch (*ptr) {
                case 'w':   count += secondLen / 2;   break;
                case 'W':   count += secondLen;       break;
                case 'p':   count += 5 * secondLen;   break;
                case 'P':   count += 10 * secondLen;  break;
                default:
                    if ((*ptr >= '0' && *ptr <= '9') || *ptr == 'n' || *ptr == 'N')
                        continue;
                }
            }
            count++;
        }
    }
    return count;
}
char const *FinaleTextWidget::text() const {
    return d->text;
}
FinaleTextWidget &FinaleTextWidget::setText(char const *newText) {
    Z_Free(d->text); d->text = nullptr;
    if (newText && newText[0]) {
        int len = (int)qstrlen(newText) + 1;
        d->text = (char *) Z_Malloc(len, PU_APPSTATIC, 0);
        std::memcpy(d->text, newText, len);
    }
    int const visLen = visLength();
    if (d->cursorPos > visLen) {
        d->cursorPos = visLen;
    }
    return *this;
}
fontid_t FinaleTextWidget::font() const {
    return d->fontNum;
}
FinaleTextWidget &FinaleTextWidget::setFont(fontid_t newFont) {
    d->fontNum  = newFont;
    d->pageFont = 0;
    return *this;
}
int FinaleTextWidget::alignment() const {
    return d->alignFlags;
}
FinaleTextWidget &FinaleTextWidget::setAlignment(int newAlignment) {
    d->alignFlags = newAlignment;
    return *this;
}
float FinaleTextWidget::lineHeight() const {
    return d->lineHeight;
}
FinaleTextWidget &FinaleTextWidget::setLineHeight(float newLineHeight) {
    d->lineHeight = newLineHeight;
    return *this;
}
int FinaleTextWidget::scrollRate() const {
    return d->scrollWait;
}
FinaleTextWidget &FinaleTextWidget::setScrollRate(int newRateInTics) {
    d->scrollWait  = newRateInTics;
    d->scrollTimer = 0;
    return *this;
}
int FinaleTextWidget::typeInRate() const {
    return d->wait;
}
FinaleTextWidget &FinaleTextWidget::setTypeInRate(int newRateInTics) {
    d->wait = newRateInTics;
    return *this;
}
FinaleTextWidget &FinaleTextWidget::setColor(Vector3f const &newColor, int steps) {
    AnimatorVector3_Set(*((animatorvector3_t *)d->color), newColor.x, newColor.y, newColor.z, steps);
    d->pageColor = 0;
    return *this;
}
FinaleTextWidget &FinaleTextWidget::setAlpha(float alpha, int steps) {
    Animator_Set(&d->color[CA], alpha, steps);
    return *this;
}
FinaleTextWidget &FinaleTextWidget::setColorAndAlpha(Vector4f const &newColorAndAlpha, int steps) {
    AnimatorVector4_Set(d->color, newColorAndAlpha.x, newColorAndAlpha.y, newColorAndAlpha.z, newColorAndAlpha.w, steps);
    d->pageColor = 0;
    return *this;
}
FinaleTextWidget &FinaleTextWidget::setPageColor(uint id) {
    d->pageColor = id;
    return *this;
}
FinaleTextWidget &FinaleTextWidget::setPageFont(uint id) {
    d->pageFont = id;
    return *this;
}
DENG2_PIMPL_NOREF(FinaleWidget) {
    Id id;
    String name;
    animatorvector3_t pos;
    animator_t angle;
    animatorvector3_t scale;
    FinalePageWidget *page = nullptr;
    Impl() {
        AnimatorVector3_Init(pos, 0, 0, 0);
        Animator_Init(&angle, 0);
        AnimatorVector3_Init(scale, 1, 1, 1);
    }
};
FinaleWidget::FinaleWidget(de::String const &name) : d(new Impl) {
    setName(name);
}
FinaleWidget::~FinaleWidget() {
    DENG2_FOR_AUDIENCE(Deletion, i) i->finaleWidgetBeingDeleted(*this);
}
Id FinaleWidget::id() const {
    return d->id;
}
String FinaleWidget::name() const {
    return d->name;
}
FinaleWidget &FinaleWidget::setName(String const &newName) {
    d->name = newName;
    return *this;
}
animatorvector3_t const &FinaleWidget::origin() const {
    return d->pos;
}
FinaleWidget &FinaleWidget::setOrigin(Vector3f const &newPos, int steps) {
    AnimatorVector3_Set(d->pos, newPos.x, newPos.y, newPos.z, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setOriginX(float newX, int steps) {
    Animator_Set(&d->pos[0], newX, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setOriginY(float newY, int steps) {
    Animator_Set(&d->pos[1], newY, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setOriginZ(float newZ, int steps) {
    Animator_Set(&d->pos[2], newZ, steps);
    return *this;
}
animator_t const &FinaleWidget::angle() const {
    return d->angle;
}
FinaleWidget &FinaleWidget::setAngle(float newAngle, int steps) {
    Animator_Set(&d->angle, newAngle, steps);
    return *this;
}
animatorvector3_t const &FinaleWidget::scale() const {
    return d->scale;
}
FinaleWidget &FinaleWidget::setScale(Vector3f const &newScale, int steps) {
    AnimatorVector3_Set(d->scale, newScale.x, newScale.y, newScale.z, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setScaleX(float newXScale, int steps) {
    Animator_Set(&d->scale[0], newXScale, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setScaleY(float newYScale, int steps) {
    Animator_Set(&d->scale[1], newYScale, steps);
    return *this;
}
FinaleWidget &FinaleWidget::setScaleZ(float newZScale, int steps) {
    Animator_Set(&d->scale[2], newZScale, steps);
    return *this;
}
FinalePageWidget *FinaleWidget::page() const {
    return d->page;
}
FinaleWidget &FinaleWidget::setPage(FinalePageWidget *newPage) {
    d->page = newPage;
    return *this;
}
void FinaleWidget::runTicks(/*timespan_t timeDelta*/)
{
    AnimatorVector3_Think(d->pos);
    AnimatorVector3_Think(d->scale);
    Animator_Think(&d->angle);
}
DENG2_PIMPL_NOREF(InFineSystem)
, DENG2_OBSERVES(Finale, Deletion) {
    Finales finales;
    ~Impl() { qDeleteAll(finales); }
    Finale *finaleForId(finaleid_t id) {
        if (id != 0) {
            for (Finale const *f : finales) {
                if (f->id() == id) return const_cast<Finale *>(f);
            }
        }
        return nullptr;
    }
    finaleid_t nextUnusedId() {
        finaleid_t id = 0;
        while (finaleForId(++id)) {}
        return id;
    }
    void finaleBeingDeleted(Finale const &finale) {
        finales.removeOne(const_cast<Finale *>(&finale));
    }
};
InFineSystem::InFineSystem() : d(new Impl)
{}
void InFineSystem::reset() {
    LOG_AS("InFineSystem");
    while (!d->finales.isEmpty()) {
        std::unique_ptr<Finale> finale(d->finales.takeFirst());
        finale->terminate();
    }
}
bool InFineSystem::finaleInProgess() const {
    for (Finale *finale : d->finales) {
        if (finale->isActive() || finale->isSuspended())
            return true;
    }
    return false;
}
void InFineSystem::runTicks(timespan_t timeDelta) {
    LOG_AS("InFineSystem");
    for (int i = 0; i < d->finales.count(); ++i) {
        Finale *finale = d->finales[i];
        if (finale->runTicks(timeDelta)) {
            delete finale;
        }
    }
}
Finale &InFineSystem::newFinale(int flags, String script, String const &setupCmds) {
    LOG_AS("InFineSystem");
    if (!setupCmds.isEmpty()) {
        script.prepend("OnLoad {\n" + setupCmds + "}\n");
    }
    d->finales << new Finale(flags, d->nextUnusedId(), script);
    auto *finale = d->finales.last();
    finale->audienceForDeletion() += d;
    return *finale;
}
bool InFineSystem::hasFinale(finaleid_t id) const {
    return d->finaleForId(id) != nullptr;
}
Finale &InFineSystem::finale(finaleid_t id) {
    Finale *finale = d->finaleForId(id);
    if (finale) return *finale;
    throw MissingFinaleError("finale", "No Finale known by id:" + String::number(id));
}
InFineSystem::Finales const &InFineSystem::finales() const {
    return d->finales;
}
#ifdef __CLIENT__
static bool inited;
void InFineSystem::initBindingContext() // static
{
    if (inited) return;
    inited = true;
    BindContext &context = ClientApp::inputSystem().context("finale");
    context.setDDFallbackResponder(de::function_cast<int (*)(ddevent_t const *)>(gx.FinaleResponder));
    context.activate(); // always on
}
void InFineSystem::deinitBindingContext() // static
{
    if (!inited) return;
    BindContext &context = ClientApp::inputSystem().context("finale");
    context.setDDFallbackResponder(nullptr);
    context.deactivate();
    inited = false;
}
#endif // __CLIENT__
namespace {
byte scaleMode = SCALEMODE_SMART_STRETCH;
}
void InFineSystem::consoleRegister() // static
{
    C_VAR_BYTE("rend-finale-stretch", &scaleMode, 0, SCALEMODE_FIRST, SCALEMODE_LAST);
}
#undef FI_Execute2
finaleid_t FI_Execute2(char const *script, int flags, char const *setupCmds) {
    LOG_AS("InFine.Execute");
    if (!script || !script[0]) {
        LOGDEV_SCR_WARNING("Attempted to play an empty script");
        return 0;
    }
    if ((flags & FF_LOCAL) && isDedicated) {
        LOGDEV_SCR_WARNING("No local finales in dedicated mode");
        return 0;
    }
    return App_InFineSystem().newFinale(flags, script, setupCmds).id();
}
#undef FI_Execute
finaleid_t FI_Execute(char const *script, int flags) {
    return FI_Execute2(script, flags, 0);
}
#undef FI_ScriptTerminate
void FI_ScriptTerminate(finaleid_t id) {
    LOG_AS("InFine.ScriptTerminate");
    if (App_InFineSystem().hasFinale(id)) {
        Finale &finale = App_InFineSystem().finale(id);
        if (finale.terminate()) {
            delete &finale;
        }
        return;
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
}
#undef FI_ScriptActive
dd_bool FI_ScriptActive(finaleid_t id) {
    LOG_AS("InFine.ScriptActive");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).isActive();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
#undef FI_ScriptSuspend
void FI_ScriptSuspend(finaleid_t id) {
    LOG_AS("InFine.ScriptSuspend");
    if (App_InFineSystem().hasFinale(id)) {
        App_InFineSystem().finale(id).suspend();
        return;
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
}
#undef FI_ScriptResume
void FI_ScriptResume(finaleid_t id) {
    LOG_AS("InFine.ScriptResume");
    if (App_InFineSystem().hasFinale(id)) {
        App_InFineSystem().finale(id).resume();
        return;
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
}
#undef FI_ScriptSuspended
dd_bool FI_ScriptSuspended(finaleid_t id) {
    LOG_AS("InFine.ScriptSuspended");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).interpreter().isSuspended();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
#undef FI_ScriptFlags
int FI_ScriptFlags(finaleid_t id) {
    LOG_AS("InFine.ScriptFlags");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).flags();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return 0;
}
#undef FI_ScriptResponder
int FI_ScriptResponder(finaleid_t id, void const *ev) {
    DENG2_ASSERT(ev);
    LOG_AS("InFine.ScriptResponder");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).handleEvent(*static_cast<ddevent_t const *>(ev));
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
#undef FI_ScriptCmdExecuted
dd_bool FI_ScriptCmdExecuted(finaleid_t id) {
    LOG_AS("InFine.CmdExecuted");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).interpreter().commandExecuted();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
#undef FI_ScriptRequestSkip
dd_bool FI_ScriptRequestSkip(finaleid_t id) {
    LOG_AS("InFine.ScriptRequestSkip");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).requestSkip();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
#undef FI_ScriptIsMenuTrigger
dd_bool FI_ScriptIsMenuTrigger(finaleid_t id) {
    LOG_AS("InFine.ScriptIsMenuTrigger");
    if (App_InFineSystem().hasFinale(id)) {
        return App_InFineSystem().finale(id).isMenuTrigger();
    }
    LOGDEV_SCR_WARNING("Unknown finaleid %i") << id;
    return false;
}
DENG_DECLARE_API(Infine) = {
    { DE_API_INFINE },
    FI_Execute2,
    FI_Execute,
    FI_ScriptActive,
    FI_ScriptFlags,
    FI_ScriptTerminate,
    FI_ScriptSuspend,
    FI_ScriptResume,
    FI_ScriptSuspended,
    FI_ScriptRequestSkip,
    FI_ScriptCmdExecuted,
    FI_ScriptIsMenuTrigger,
    FI_ScriptResponder
};
