﻿/** @file busywidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "busyrunner.h"
#include "ui/widgets/busywidget.h"
#include "ui/busyvisual.h"
#include "ui/ui_main.h"
#include "ui/clientwindow.h"
#include "gl/gl_main.h"
#include "render/r_main.h"
#include "sys_system.h"
#include <de/concurrency.h>
#include <de/Drawable>
#include <de/GLTextureFramebuffer>
#include <de/GuiRootWidget>
#include <de/ProgressWidget>
/** @file consolecommandwidget.h  Text editor with a history buffer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/consolecommandwidget.h"
#include <de/App>
#include <de/KeyEvent>
#include <doomsday/console/knownword.h>
#include <doomsday/console/exec.h>
#include "ui/inputsystem.h"
#include "clientapp.h"
#include "dd_main.h"
#include "BindContext"
/** @file consolewidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/consolewidget.h"
#include "CommandAction"
#include "ui/widgets/consolecommandwidget.h"
#include "ui/widgets/inputbindingwidget.h"
#include "ui/dialogs/logsettingsdialog.h"
#include "ui/clientwindow.h"
#include "ui/styledlogsinkformatter.h"
#include <de/AnimationRule>
#include <de/App>
#include <de/Config>
#include <de/FileSystem>
#include <de/KeyEvent>
#include <de/LogBuffer>
#include <de/LogWidget>
#include <de/MouseEvent>
#include <de/NativeFile>
#include <de/PersistentState>
#include <de/PopupButtonWidget>
#include <de/PopupMenuWidget>
#include <de/ScriptCommandWidget>
#include <de/SignalAction>
#include <de/ToggleWidget>
#include <de/ui/SubwidgetItem>
#include <de/ui/VariableToggleItem>
#include <doomsday/doomsdayapp.h>
#if !defined (DENG_MOBILE)
#  include <QApplication>
#  include <QCursor>
#  include <QClipboard>
#endif
/** @file cvarchoicewidget.cpp  Console variable choice.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/cvarchoicewidget.h"
#include <doomsday/console/var.h>
/** @file cvarlineeditwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/cvarlineeditwidget.h"
#include <doomsday/console/var.h>
/** @file cvarnativepathwidget.cpp  Console variable with a native path.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/cvarnativepathwidget.h"
#include "ui/clientwindow.h"
#include "clientapp.h"
#include <doomsday/console/var.h>
#include <de/PopupMenuWidget>
#include <de/SignalAction>
#include <QFileDialog>
/** @file cvarsliderwidget.cpp  Slider for adjusting a cvar.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/cvarsliderwidget.h"
#include <doomsday/console/var.h>
/** @file cvartogglewidget.cpp Console variable toggle.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/cvartogglewidget.h"
#include <doomsday/console/var.h>
/** @file gamewidget.cpp
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2014-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h" // must be included first
#include "api_render.h"
#include "ui/widgets/gamewidget.h"
#include "clientapp.h"
#include "ui/ddevent.h"
#include "ui/ui_main.h"
#include "ui/sys_input.h"
#include "ui/busyvisual.h"
#include "ui/clientwindowsystem.h"
#include "ui/viewcompositor.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/widgets/busywidget.h"
#include "dd_def.h"
#include "dd_main.h"
#include "dd_loop.h"
#include "sys_system.h"
#include "world/map.h"
#include "world/p_players.h"
#include "network/net_main.h"
#include "client/cl_def.h" // clientPaused
#include "render/r_main.h"
#include "render/rend_main.h"
#include "render/cameralensfx.h"
#include "gl/gl_main.h"
#include "gl/sys_opengl.h"
#include "gl/gl_defer.h"
#include <doomsday/console/exec.h>
#include <de/FileSystem>
#include <de/GLState>
#include <de/GLTextureFramebuffer>
#include <de/LogBuffer>
#include <de/VRConfig>
#include <QPainter>
#include <QImage>
#include <QBuffer>
/**
 * Maximum number of milliseconds spent uploading textures at the beginning
 * of a frame. Note that non-uploaded textures will appear as pure white
 * until their content gets uploaded (you should precache them).
 */
#define FRAME_DEFERRED_UPLOAD_TIMEOUT 20
/** @file homeitemwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/homeitemwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/home/columnwidget.h"
#include "resource/idtech1image.h"
#include <doomsday/LumpCatalog>
#include <doomsday/Game>
#include <de/SequentialLayout>
#include <QTimer>
/** @file homemenuwidget.cpp  Menu for the Home.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/homemenuwidget.h"
#include "ui/widgets/homeitemwidget.h"
#include "ui/home/columnwidget.h"
#include <de/FocusWidget>
/** @file inputbindingwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/inputbindingwidget.h"
#include <de/charsymbols.h>
#include <de/AuxButtonWidget>
#include "clientapp.h"
#include "BindContext"
#include "CommandBinding"
#include "ui/inputsystem.h"
#include "ui/b_util.h"
/** @file keygrabberwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/keygrabberwidget.h"
#include "clientapp.h"
#include "ui/ddevent.h"
#include "ui/b_util.h"
#include "ui/inputsystem.h"
#include <de/KeyEvent>
#include <de/ddstring.h>
/** @file mapoutlinewidget.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/mapoutlinewidget.h"
#include <de/ProgressWidget>
#include <de/Drawable>
/** @file multiplayerservermenuwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/multiplayerservermenuwidget.h"
#include "ui/home/multiplayerpanelbuttonwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "network/serverlink.h"
#include "clientapp.h"
#include <doomsday/Games>
#include <de/MenuWidget>
#include <de/Address>
/** @file multiplayerstatuswidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/multiplayerstatuswidget.h"
#include "network/serverlink.h"
#include "clientapp.h"
#include "CommandAction"
#include <de/ui/ActionItem>
#include <de/ui/SubwidgetItem>
#include <QTimer>
/** @file nativepathwidget.cpp  Widget for selecting a native path.
 *
 * @authors Copyright (c) 2014-2018 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/nativepathwidget.h"
#include "ui/clientwindow.h"
#include "clientapp.h"
#include <de/PopupMenuWidget>
#include <de/SignalAction>
#include <QFileDialog>
/** @file packagecontentoptionswidget.cpp  Widget for package content options.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/packagecontentoptionswidget.h"
#include <de/ButtonWidget>
#include <de/Config>
#include <de/DictionaryValue>
#include <de/LabelWidget>
#include <de/MenuWidget>
#include <de/PackageLoader>
#include <de/PopupWidget>
#include <de/SequentialLayout>
#include <de/ToggleWidget>
#include <de/TextValue>
/** @file packagesbuttonwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/packagesbuttonwidget.h"
#include "ui/dialogs/packagesdialog.h"
#include <de/CallbackAction>
/** @file packagessidebarwidget.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/packagessidebarwidget.h"
#include "ui/widgets/packageswidget.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/widgets/homeitemwidget.h"
#include <de/CallbackAction>
#include <de/FileSystem>
#include <de/ui/ActionItem>
/** @file packageswidget.cpp  Widget for searching and examining packages.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/packageswidget.h"
#include "ui/widgets/homeitemwidget.h"
#include "ui/widgets/homemenuwidget.h"
#include "ui/widgets/panelbuttonwidget.h"
#include "ui/dialogs/packageinfodialog.h"
#include "ui/widgets/packagecontentoptionswidget.h"
#include "ui/clientrootwidget.h"
#include "clientapp.h"
#include <de/AtlasProceduralImage>
#include <de/CallbackAction>
#include <de/ChildWidgetOrganizer>
#include <de/DocumentPopupWidget>
#include <de/FileSystem>
#include <de/LineEditWidget>
#include <de/Loop>
#include <de/MenuWidget>
#include <de/NativeFile>
#include <de/PackageIconBank>
#include <de/PackageLoader>
#include <de/PopupButtonWidget>
#include <de/ProgressWidget>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/TaskPool>
#include <de/ui/FilteredData>
#include <de/ui/VariantActionItem>
#include <doomsday/DoomsdayApp>
#include <doomsday/resource/bundles.h>
#include <QTimer>
#include <set>
/** @file panelbuttonwidget.cpp  Button with an extensible drawer.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/panelbuttonwidget.h"
/** @file privilegedlogwidget.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/privilegedlogwidget.h"
#include "ui/styledlogsinkformatter.h"
#include <de/LogWidget>
#include <de/ButtonWidget>
/** @file profilepickerwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/profilepickerwidget.h"
#include "ui/clientwindow.h"
#include <de/SignalAction>
#include <de/PopupMenuWidget>
#include <de/InputDialog>
/** @file sidebarwidget.cpp  Base class for sidebar widgets.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/sidebarwidget.h"
#include "ui/editors/variablegroupeditor.h"
#include "ui/clientwindow.h"
#include <doomsday/doomsdayapp.h>
#include <de/DialogContentStylist>
#include <de/SequentialLayout>
#include <de/SignalAction>
/** @file taskbarwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/taskbarwidget.h"
#include "CommandAction"
#include "clientapp.h"
#include "dd_main.h"
#include "network/serverlink.h"
#include "ui/clientrootwidget.h"
#include "ui/clientwindow.h"
#include "ui/dialogs/aboutdialog.h"
#include "ui/dialogs/audiosettingsdialog.h"
#include "ui/dialogs/datafilesettingsdialog.h"
#include "ui/dialogs/inputsettingsdialog.h"
#include "ui/dialogs/manualconnectiondialog.h"
#include "ui/dialogs/networksettingsdialog.h"
#include "ui/dialogs/renderersettingsdialog.h"
#include "ui/dialogs/uisettingsdialog.h"
#include "ui/dialogs/videosettingsdialog.h"
#include "ui/dialogs/vrsettingsdialog.h"
#include "ui/home/homewidget.h"
#include "ui/inputsystem.h"
#include "ui/progress.h"
#include "ui/ui_main.h"
#include "ui/widgets/consolecommandwidget.h"
#include "ui/widgets/multiplayerstatuswidget.h"
#include "ui/widgets/packagessidebarwidget.h"
#include "ui/widgets/tutorialwidget.h"
#include "updater/updatersettingsdialog.h"
#include <doomsday/filesys/fs_main.h>
#include <doomsday/console/exec.h>
#include <de/AnimationRule>
#include <de/BlurWidget>
#include <de/ButtonWidget>
#include <de/CallbackAction>
#include <de/Config>
#include <de/DirectoryListDialog>
#include <de/Drawable>
#include <de/GLBuffer>
#include <de/KeyEvent>
#include <de/Painter>
#include <de/PopupMenuWidget>
#include <de/SequentialLayout>
#include <de/SignalAction>
#include <de/ui/SubwidgetItem>
#include <QFileDialog>
/** @file tutorialwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/widgets/tutorialwidget.h"
#include "ui/clientwindow.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/widgets/inputbindingwidget.h"
#include "dd_version.h"
#include "dd_main.h"
#include <doomsday/doomsdayapp.h>
#include <de/MessageDialog>
#include <de/SignalAction>
#include <de/Untrapper>
#include <de/PopupMenuWidget>
#include <de/ProgressWidget>
#include <de/NotificationAreaWidget>
#include <de/StyleProceduralImage>
#include <QTimer>
using namespace de::ui;
using namespace de;
DENG_GUI_PIMPL(BusyWidget) {
    typedef DefaultVertexBuf VertexBuf;
    ProgressWidget *progress;
    GameWidget *gameWidget = nullptr;
    Time frameDrawnAt;
    GLTextureFramebuffer transitionFrame;
    Drawable drawable;
    GLUniform uTex       { "uTex",       GLUniform::Sampler2D };
    GLUniform uMvpMatrix { "uMvpMatrix", GLUniform::Mat4      };
    Impl(Public *i) : Base(i) {
        progress = new ProgressWidget;
        progress->setAlignment(ui::AlignCenter, LabelWidget::AlignOnlyByImage);
        progress->setRange(Rangei(0, 200));
        progress->setImageScale(.2f);
        progress->rule().setRect(self().rule());
        self().add(progress);
    }
    void glInit() {
        VertexBuf *buf = new VertexBuf;
        VertexBuf::Builder verts;
        verts.makeQuad(Rectanglef(0, 0, 1, 1), Vector4f(1, 1, 1, 1), Rectanglef(0, 0, 1, 1));
        buf->setVertices(gl::TriangleStrip, verts, gl::Static);
        drawable.addBuffer(buf);
        shaders().build(drawable.program(), "generic.textured.color")
                << uMvpMatrix << uTex;
    }
    void glDeinit() {
        drawable.clear();
        transitionFrame.glDeinit();
    }
    bool haveTransitionFrame() const {
        return transitionFrame.isReady();
    }
};
BusyWidget::BusyWidget(String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    requestGeometry(false);
}
ProgressWidget &BusyWidget::progress() {
    return *d->progress;
}
void BusyWidget::setGameWidget(GameWidget &gameWidget) {
    d->gameWidget = &gameWidget;
}
/*void BusyWidget::viewResized()
{
    GuiWidget::viewResized();
}*/
void BusyWidget::update() {
    GuiWidget::update();
    if (BusyMode_Active()) {
        BusyMode_Loop();
    }
}
void BusyWidget::drawContent() {
    if (!BusyMode_Active()) {
        d->progress->hide();
        if (Con_TransitionInProgress()) {
            root().painter().flush();
            GLState::push()
                    .setViewport(Rectangleui::fromSize(GLState::current().target().size()));
            Con_DrawTransition();
            GLState::pop();
        }
        else {
            hide();
            releaseTransitionFrame();
        }
        return;
    }
    if (d->haveTransitionFrame()) {
        root().painter().flush();
        GLState::push()
                .setAlphaTest(false)
                .setBlend(false);
        Rectanglei pos = rule().recti();
        d->uMvpMatrix = Matrix4f::scale(Vector3f(1, -1, 1)) *
                root().projMatrix2D() *
                Matrix4f::scaleThenTranslate(pos.size(), pos.topLeft);
        d->drawable.draw();
        GLState::pop();
    }
}
bool BusyWidget::handleEvent(Event const &) {
    return true;
}
void BusyWidget::renderTransitionFrame() {
    LOG_AS("BusyWidget");
    if (d->haveTransitionFrame()) {
        LOGDEV_GL_VERBOSE("Skipping rendering of transition frame (got one already)");
        return;
    }
    d->frameDrawnAt = Time();
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DENG2_ASSERT(d->gameWidget);
    Rectanglei grabRect = Rectanglei::fromSize(rule().recti().size());
    LOGDEV_GL_VERBOSE("Rendering transition frame, size %s pixels") << grabRect.size().asText();
    d->transitionFrame.resize(grabRect.size());
    if (!d->transitionFrame.isReady()) {
        d->transitionFrame.glInit();
    }
    GLState::push()
            .setTarget(d->transitionFrame)
            .setViewport(Rectangleui::fromSize(d->transitionFrame.size()));
    d->gameWidget->drawComposited();
    GLState::pop();
    d->transitionFrame.resolveSamples();
    d->uTex = d->transitionFrame.colorTexture();
}
void BusyWidget::releaseTransitionFrame() {
    if (d->haveTransitionFrame()) {
        LOGDEV_GL_VERBOSE("Releasing transition frame");
        d->transitionFrame.glDeinit();
    }
}
void BusyWidget::clearTransitionFrameToBlack() {
    if (d->haveTransitionFrame()) {
        d->transitionFrame.clear(GLFramebuffer::Color);
    }
}
GLTexture const *BusyWidget::transitionFrame() const {
    if (d->haveTransitionFrame()) {
        return &d->transitionFrame.colorTexture();
    }
    return nullptr;
}
void BusyWidget::glInit() {
    d->glInit();
}
void BusyWidget::glDeinit() {
    d->glDeinit();
}
DENG_GUI_PIMPL(ConsoleCommandWidget),
DENG2_OBSERVES(App, StartupComplete),
DENG2_OBSERVES(DoomsdayApp, GameChange) {
    Impl(Public *i) : Base(i) {
        App::app().audienceForStartupComplete() += this;
        DoomsdayApp::app().audienceForGameChange() += this;
    }
    void appStartupCompleted() {
        updateLexicon();
    }
    void currentGameChanged(Game const &) {
        updateLexicon();
    }
    void updateLexicon() {
        self().setLexicon(Con_Lexicon());
    }
};
ConsoleCommandWidget::ConsoleCommandWidget(String const &name)
    : CommandWidget(name), d(new Impl(this)) {
    d->updateLexicon();
}
void ConsoleCommandWidget::focusGained() {
    CommandWidget::focusGained();
    ClientApp::inputSystem().context("console").activate();
}
void ConsoleCommandWidget::focusLost() {
    CommandWidget::focusLost();
    ClientApp::inputSystem().context("console").deactivate();
}
bool ConsoleCommandWidget::handleEvent(Event const &event) {
    if (isDisabled()) return false;
    if (hasFocus()) {
        if (ClientApp::inputSystem().tryEvent(event, "console")) {
            return true;
        }
    }
    return CommandWidget::handleEvent(event);
}
bool ConsoleCommandWidget::isAcceptedAsCommand(String const &) {
    return true;
}
void ConsoleCommandWidget::executeCommand(String const &text) {
    LOG_SCR_NOTE(_E(1) "> ") << text;
    Con_Execute(CMDS_CONSOLE, text.toUtf8(), false, false);
}
void ConsoleCommandWidget::autoCompletionBegan(String const &) {
    QStringList const compls = suggestedCompletions();
    if (compls.size() > 1) {
        showAutocompletionPopup(Con_AnnotatedConsoleTerms(compls));
    }
}
static TimeSpan const LOG_OPEN_CLOSE_SPAN = 0.2;
DENG_GUI_PIMPL(ConsoleWidget)
, DENG2_OBSERVES(Variable, Change)
, DENG2_OBSERVES(DoomsdayApp, GameChange) {
    GuiWidget *buttons = nullptr;
    PopupButtonWidget *button = nullptr;
    PopupButtonWidget *promptButton = nullptr;
    PopupMenuWidget *logMenu = nullptr;
    PopupMenuWidget *promptMenu = nullptr;
    ConsoleCommandWidget *cmdLine = nullptr;
    ScriptCommandWidget *scriptCmd = nullptr;
    LogWidget *log = nullptr;
    AnimationRule *horizShift;
    AnimationRule *height;
    AnimationRule *width;
    StyledLogSinkFormatter formatter;
    bool opened = true;
    bool scriptMode = false;
    int grabWidth = 0;
    enum GrabEdge {
        NotGrabbed = 0,
        RightEdge,
        TopEdge
    };
    GrabEdge grabHover = NotGrabbed;
    GrabEdge grabbed = NotGrabbed;
    Impl(Public *i) : Base(i) {
        horizShift = new AnimationRule(0);
        width      = new AnimationRule(rule("console.width").valuei());
        height     = new AnimationRule(0);
        grabWidth  = rule("gap").valuei();
        App::config("console.script").audienceForChange() += this;
        DoomsdayApp::app().audienceForGameChange() += this;
    }
    ~Impl() {
        releaseRef(horizShift);
        releaseRef(width);
        releaseRef(height);
    }
    void currentGameChanged(Game const &) {
        scriptCmd->updateCompletion();
    }
    void expandLog(int delta, bool useOffsetAnimation) {
        if (grabbed == TopEdge) return;
        Animation::Style style = useOffsetAnimation? Animation::EaseOut : Animation::Linear;
        if (height->animation().target() == 0) {
            delta += log->margins().height().valuei();
        }
        height->set(height->animation().target() + delta, .25f);
        height->setStyle(style);
        if (useOffsetAnimation && opened) {
            log->setContentYOffset(Animation::range(style,
                                                    log->contentYOffset().value() + delta, 0,
                                                    height->animation().remainingTime()));
        }
    }
    bool handleMousePositionEvent(MouseEvent const &mouse) {
        if (grabbed == RightEdge) {
            width->set(mouse.pos().x - self().rule().left().valuei(), .1f);
            return true;
        }
        else if (grabbed == TopEdge) {
            if (mouse.pos().y < self().rule().bottom().valuei()) {
                height->set(self().rule().bottom().valuei() - mouse.pos().y, .1f);
                log->enablePageKeys(false);
            }
            return true;
        }
        Rectanglei pos = self().rule().recti();
        pos.topLeft.x = pos.bottomRight.x - grabWidth;
        if (pos.contains(mouse.pos())) {
            if (grabHover != RightEdge) {
                grabHover = RightEdge;
#if !defined (DENG_MOBILE)
                self().root().window().setCursor(Qt::SizeHorCursor);
#endif
            }
        }
        else {
            pos = self().rule().recti();
            pos.bottomRight.y = pos.topLeft.y + grabWidth;
            if (pos.contains(mouse.pos())) {
                if (grabHover != TopEdge) {
                    grabHover = TopEdge;
#if !defined (DENG_MOBILE)
                    self().root().window().setCursor(Qt::SizeVerCursor);
#endif
                }
            }
            else if (grabHover != NotGrabbed) {
                grabHover = NotGrabbed;
#if !defined (DENG_MOBILE)
                self().root().window().setCursor(Qt::ArrowCursor);
#endif
            }
        }
        return false;
    }
    void variableValueChanged(Variable &, Value const &newValue) {
        setScriptMode(newValue.isTrue());
    }
    void setScriptMode(bool yes) {
        CommandWidget *current;
        CommandWidget *next;
        if (scriptMode) {
            current = scriptCmd;
        }
        else {
            current = cmdLine;
        }
        if (yes) {
            next = scriptCmd;
        }
        else {
            next = cmdLine;
        }
        promptButton->setText(yes? _E(b)_E(F) "$" : ">");
        self().rule().setInput(Rule::Bottom, next->rule().top() - rule(RuleBank::UNIT));
        scriptCmd->setAttribute(AnimateOpacityWhenEnabledOrDisabled, UnsetFlags);
        cmdLine  ->setAttribute(AnimateOpacityWhenEnabledOrDisabled, UnsetFlags);
        scriptCmd->show(yes);
        scriptCmd->enable(yes);
        cmdLine->show(!yes);
        cmdLine->disable(yes);
        if (current->hasFocus()) {
            root().setFocus(next);
        }
        if (scriptMode != yes) {
            scriptMode = yes;
            emit self().commandModeChanged();
        }
    }
    struct RightClick : public GuiWidget::IEventHandler {
        Impl *d;
        RightClick(Impl *i) : d(i) {}
        bool handleEvent(GuiWidget &widget, Event const &event) {
            switch (widget.handleMouseClick(event, MouseEvent::Right)) {
            case MouseClickFinished:
                App::config().set("console.script", !d->scriptMode);
                return true;
            case MouseClickUnrelated:
                break; // not consumed
            default:
                return true;
            }
            return false;
        }
    };
};
static PopupWidget *consoleShortcutPopup() {
    auto *pop = new PopupWidget;
    pop->setOutlineColor("popup.outline");
    auto *padding = new GuiWidget;
    padding->margins().set("dialog.gap");
    auto *bind = InputBindingWidget::newTaskBarShortcut();
    bind->setSizePolicy(ui::Expand, ui::Expand);
    padding->add(bind);
    bind->rule()
            .setLeftTop(padding->rule().left() + padding->margins().left(),
                        padding->rule().top()  + padding->margins().top());
    padding->rule()
            .setInput(Rule::Width,  bind->rule().width()  + padding->margins().width())
            .setInput(Rule::Height, bind->rule().height() + padding->margins().height());
    pop->setContent(padding);
    return pop;
}
ConsoleWidget::ConsoleWidget() : GuiWidget("console"), d(new Impl(this)) {
    d->buttons = new GuiWidget;
    add(d->buttons);
    d->button = new PopupButtonWidget;
    d->button->setSizePolicy(ui::Expand, ui::Expand);
    d->button->setImage(style().images().image("log"));
    d->button->setOverrideImageSize(style().fonts().font("default").height());
    d->buttons->add(d->button);
    d->button->rule()
            .setInput(Rule::Top,    d->buttons->rule().top())
            .setInput(Rule::Left,   d->buttons->rule().left())
            .setInput(Rule::Height, d->buttons->rule().height())
            .setInput(Rule::Width,  d->button->rule().height()); // square
    d->promptButton = new PopupButtonWidget;
    d->promptButton->addEventHandler(new Impl::RightClick(d));
    d->promptButton->rule()
            .setInput(Rule::Width,  d->promptButton->rule().height())
            .setInput(Rule::Height, d->buttons->rule().height())
            .setInput(Rule::Left,   d->button->rule().right())
            .setInput(Rule::Top,    d->buttons->rule().top());
    d->buttons->add(d->promptButton);
    d->buttons->rule().setInput(Rule::Width, d->button->rule().width() +
                                             d->promptButton->rule().width());
    d->cmdLine = new ConsoleCommandWidget("commandline");
    d->cmdLine->setEmptyContentHint(tr("Enter commands here") /*  " _E(r)_E(l)_E(t) "SHIFT-ESC" */);
    add(d->cmdLine);
    d->scriptCmd = new ScriptCommandWidget("script");
    d->scriptCmd->setEmptyContentHint(tr("Enter scripts here"));
    add(d->scriptCmd);
    d->buttons->setOpacity(.75f);
    d->cmdLine->setOpacity(.75f);
    d->scriptCmd->setOpacity(.75f);
    d->log = new LogWidget("log");
    d->log->setLogFormatter(d->formatter);
    d->log->rule()
            .setInput(Rule::Left,   rule().left())
            .setInput(Rule::Right,  rule().right())
            .setInput(Rule::Bottom, rule().bottom())
            .setInput(Rule::Top,    OperatorRule::maximum(rule().top(), Const(0)));
    add(d->log);
    enableBlur();
    rule()
        .setInput(Rule::Height, *d->height)
        .setInput(Rule::Width, OperatorRule::minimum(ClientWindow::main().root().viewWidth(),
                                                     OperatorRule::maximum(*d->width, Const(320))));
    closeLog();
    d->logMenu = new PopupMenuWidget;
    d->button->setPopup(*d->logMenu);
    d->logMenu->items()
            << new ui::ActionItem(tr("Show Full Log"), new SignalAction(this, SLOT(showFullLog())))
            << new ui::ActionItem(tr("Close Log"), new SignalAction(this, SLOT(closeLogAndUnfocusCommandLine())))
            << new ui::ActionItem(tr("Go to Latest"), new SignalAction(d->log, SLOT(scrollToBottom())))
            << new ui::ActionItem(tr("Copy Path to Clipboard"),
                                  new SignalAction(this, SLOT(copyLogPathToClipboard())))
            << new ui::Item(ui::Item::Annotation,
                            tr("Copies the location of the log output file to the OS clipboard."))
            << new ui::Item(ui::Item::Separator)
            << new ui::ActionItem(style().images().image("close.ring"), tr("Clear Log"),
                                  new CommandAction("clear"))
            << new ui::Item(ui::Item::Separator)
            << new ui::VariableToggleItem(tr("Snap to Latest Entry"), App::config("console.snap"))
            << new ui::Item(ui::Item::Annotation, tr("Running a command or script causes the log to scroll down to the latest entry."))
            << new ui::VariableToggleItem(tr("Entry Metadata"), App::config("log.showMetadata"))
            << new ui::Item(ui::Item::Annotation, tr("Time and subsystem of each new entry is printed."))
            << new ui::Item(ui::Item::Separator)
            << new ui::SubwidgetItem(tr("Log Filter & Alerts"), ui::Right, makePopup<LogSettingsDialog>);
    add(d->logMenu);
    d->promptMenu = new PopupMenuWidget;
    d->promptButton->setPopup(*d->promptMenu);
    d->promptMenu->items()
            << new ui::SubwidgetItem(tr("Shortcut Key"), ui::Right, consoleShortcutPopup)
            << new ui::Item(ui::Item::Separator)
            << new ui::VariableToggleItem(tr("Doomsday Script"), App::config("console.script"))
            << new ui::Item(ui::Item::Annotation,
                            tr("The command prompt becomes an interactive script "
                               "process with access to all the runtime DS modules."));
    add(d->promptMenu);
    d->setScriptMode(App::config().getb("console.script"));
    connect(d->log, SIGNAL(contentHeightIncreased(int)), this, SLOT(logContentHeightIncreased(int)));
    connect(d->cmdLine, SIGNAL(gotFocus()), this, SLOT(commandLineFocusGained()));
    connect(d->cmdLine, SIGNAL(lostFocus()), this, SLOT(commandLineFocusLost()));
    connect(d->cmdLine, SIGNAL(commandEntered(de::String)), this, SLOT(commandWasEntered(de::String)));
    connect(d->scriptCmd, SIGNAL(gotFocus()), this, SLOT(commandLineFocusGained()));
    connect(d->scriptCmd, SIGNAL(lostFocus()), this, SLOT(commandLineFocusLost()));
    connect(d->scriptCmd, SIGNAL(commandEntered(de::String)), this, SLOT(commandWasEntered(de::String)));
}
GuiWidget &ConsoleWidget::buttons() {
    return *d->buttons;
}
CommandWidget &ConsoleWidget::commandLine() {
    if (d->scriptMode) return *d->scriptCmd;
    return *d->cmdLine;
}
LogWidget &ConsoleWidget::log() {
    return *d->log;
}
Rule const &ConsoleWidget::shift() {
    return *d->horizShift;
}
bool ConsoleWidget::isLogOpen() const {
    return d->opened;
}
void ConsoleWidget::enableBlur(bool yes) {
    Background logBg = d->log->background();
    if (yes) {
        logBg.type = Background::SharedBlur;
        logBg.blur = &ClientWindow::main().taskBarBlur();
    }
    else {
        logBg.type = Background::None;
    }
    d->log->set(logBg);
}
void ConsoleWidget::viewResized() {
    GuiWidget::viewResized();
    if (!d->opened) {
        d->horizShift->set(-rule().width().valuei() - 1);
    }
}
void ConsoleWidget::update() {
    GuiWidget::update();
    if (rule().top().valuei() <= 0) {
        d->log->enablePageKeys(true);
    }
}
bool ConsoleWidget::handleEvent(Event const &event) {
    if (event.type() == Event::MousePosition) {
        if (d->handleMousePositionEvent(event.as<MouseEvent>())) {
            return true;
        }
    }
    if (d->grabHover != Impl::NotGrabbed && event.type() == Event::MouseButton) {
        switch (handleMouseClick(event)) {
        case MouseClickStarted:
            d->grabbed = d->grabHover;
            return true;
        case MouseClickAborted:
        case MouseClickFinished:
            d->grabbed = Impl::NotGrabbed;
            return true;
        default:
            break;
        }
    }
    if (event.type() == Event::MouseButton && hitTest(event)) {
        return true;
    }
    if (event.type() == Event::KeyPress) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.qtKey() == Qt::Key_PageUp ||
           key.qtKey() == Qt::Key_PageDown) {
            if (!isLogOpen()) openLog();
            showFullLog();
            return true;
        }
        if (key.qtKey() == Qt::Key_F5) // Clear history.
        {
            clearLog();
            return true;
        }
    }
    return false;
}
void ConsoleWidget::operator >> (PersistentState &toState) const {
    toState.objectNamespace().set("console.width", d->width->value());
}
void ConsoleWidget::operator << (PersistentState const &fromState) {
    d->width->set(fromState["console.width"]);
    if (!d->opened) {
        d->horizShift->set(-rule().width().valuei() - 1);
    }
}
void ConsoleWidget::openLog() {
    if (d->opened) return;
    d->opened = true;
    d->horizShift->set(0, LOG_OPEN_CLOSE_SPAN);
    d->log->unsetBehavior(DisableEventDispatch);
}
void ConsoleWidget::closeLog() {
    if (!d->opened) return;
    d->opened = false;
    d->horizShift->set(-rule().width().valuei() - 1, LOG_OPEN_CLOSE_SPAN);
    d->log->setBehavior(DisableEventDispatch);
}
void ConsoleWidget::closeLogAndUnfocusCommandLine() {
    closeLog();
    root().setFocus(nullptr);
}
void ConsoleWidget::clearLog() {
    zeroLogHeight();
    d->log->clear();
}
void ConsoleWidget::zeroLogHeight() {
    d->height->set(0);
    d->log->scrollToBottom();
    d->log->enablePageKeys(false);
}
void ConsoleWidget::showFullLog() {
    openLog();
    d->log->enablePageKeys(true);
    d->expandLog(rule().top().valuei(), false);
}
void ConsoleWidget::logContentHeightIncreased(int delta) {
    d->expandLog(delta, true);
}
void ConsoleWidget::setFullyOpaque() {
    d->scriptCmd->setAttribute(AnimateOpacityWhenEnabledOrDisabled);
    d->cmdLine->setAttribute(AnimateOpacityWhenEnabledOrDisabled);
    d->buttons->setOpacity(1, .25f);
    d->cmdLine->setOpacity(1, .25f);
    d->scriptCmd->setOpacity(1, .25f);
}
void ConsoleWidget::commandLineFocusGained() {
    setFullyOpaque();
    openLog();
    emit commandLineGotFocus();
}
void ConsoleWidget::commandLineFocusLost() {
    d->scriptCmd->setAttribute(AnimateOpacityWhenEnabledOrDisabled);
    d->cmdLine->setAttribute(AnimateOpacityWhenEnabledOrDisabled);
    d->buttons->setOpacity(.75f, .25f);
    d->cmdLine->setOpacity(.75f, .25f);
    d->scriptCmd->setOpacity(.75f, .25f);
    closeLog();
}
void ConsoleWidget::focusOnCommandLine() {
    root().setFocus(&commandLine());
}
void ConsoleWidget::closeMenu() {
    d->logMenu->close();
    d->promptMenu->close();
}
void ConsoleWidget::commandWasEntered(String const &) {
    if (App::config().getb("console.snap") && !d->log->isAtBottom()) {
        d->log->scrollToBottom();
    }
}
#if !defined (DENG_MOBILE)
void ConsoleWidget::copyLogPathToClipboard() {
    if (NativeFile *native = App::rootFolder().tryLocate<NativeFile>(LogBuffer::get().outputFile())) {
        qApp->clipboard()->setText(native->nativePath());
    }
}
#endif
DENG2_PIMPL_NOREF(CVarChoiceWidget) {
    char const *cvar;
    cvar_t *var() const {
        cvar_t *cv = Con_FindVariable(cvar);
        DENG2_ASSERT(cv != 0);
        return cv;
    }
};
CVarChoiceWidget::CVarChoiceWidget(char const *cvarPath)
    : ChoiceWidget(String::format("cvar-%s", cvarPath))
    , d(new Impl) {
    d->cvar = cvarPath;
    updateFromCVar();
    connect(this, SIGNAL(selectionChangedByUser(uint)),
            this, SLOT(setCVarValueFromWidget()));
}
char const *CVarChoiceWidget::cvarPath() const {
    return d->cvar;
}
void CVarChoiceWidget::updateFromCVar() {
    setSelected(items().findData(CVar_Integer(d->var())));
}
void CVarChoiceWidget::setCVarValueFromWidget() {
    CVar_SetInteger(d->var(), selectedItem().data().toInt());
}
DENG2_PIMPL_NOREF(CVarLineEditWidget) {
    char const *cvar;
    cvar_t *var() const {
        cvar_t *cv = Con_FindVariable(cvar);
        DENG2_ASSERT(cv != 0);
        return cv;
    }
};
CVarLineEditWidget::CVarLineEditWidget(char const *cvarPath)
    : d(new Impl) {
    setSignalOnEnter(true);
    connect(this, SIGNAL(enterPressed(QString)), this, SLOT(endEditing()));
    d->cvar = cvarPath;
    updateFromCVar();
}
void CVarLineEditWidget::contentChanged() {
    LineEditWidget::contentChanged();
    if (String(CVar_String(d->var())) != text()) {
        setCVarValueFromWidget();
    }
}
char const *CVarLineEditWidget::cvarPath() const {
    return d->cvar;
}
void CVarLineEditWidget::updateFromCVar() {
    setText(CVar_String(d->var()));
}
void CVarLineEditWidget::endEditing() {
    root().setFocus(0);
}
void CVarLineEditWidget::setCVarValueFromWidget() {
    CVar_SetString(d->var(), text().toUtf8());
}
DENG2_PIMPL(CVarNativePathWidget)
, DENG2_OBSERVES(NativePathWidget, UserChange) {
    char const *cvar;
    Impl(Public *i) : Base(i)
    {}
    cvar_t *var() const {
        cvar_t *cv = Con_FindVariable(cvar);
        DENG2_ASSERT(cv != 0);
        return cv;
    }
    void pathChangedByUser(NativePathWidget &) {
        self().setCVarValueFromWidget();
    }
};
CVarNativePathWidget::CVarNativePathWidget(char const *cvarPath)
    : d(new Impl(this)) {
    d->cvar = cvarPath;
    updateFromCVar();
    setPrompt(QString("Select File for \"%1\"").arg(d->cvar));
    audienceForUserChange() += d;
}
char const *CVarNativePathWidget::cvarPath() const {
    return d->cvar;
}
void CVarNativePathWidget::updateFromCVar() {
    setPath(CVar_String(d->var()));
}
void CVarNativePathWidget::setCVarValueFromWidget() {
    CVar_SetString(d->var(), path().toString().toUtf8());
}
DENG2_PIMPL_NOREF(CVarSliderWidget) {
    char const *cvar;
    cvar_t *var() const {
        cvar_t *cv = Con_FindVariable(cvar);
        DENG2_ASSERT(cv != 0);
        return cv;
    }
};
CVarSliderWidget::CVarSliderWidget(char const *cvarPath) : d(new Impl) {
    d->cvar = cvarPath;
    if (d->var()->type == CVT_FLOAT) {
        if (!(d->var()->flags & (CVF_NO_MIN | CVF_NO_MAX))) {
            setRange(Rangef(d->var()->min, d->var()->max));
        }
        setPrecision(2);
    }
    else {
        if (!(d->var()->flags & (CVF_NO_MIN | CVF_NO_MAX))) {
            setRange(Rangei(d->var()->min, d->var()->max));
        }
    }
    updateFromCVar();
    connect(this, SIGNAL(valueChangedByUser(double)), this, SLOT(setCVarValueFromWidget()));
}
char const *CVarSliderWidget::cvarPath() const {
    return d->cvar;
}
void CVarSliderWidget::updateFromCVar() {
    cvar_t *var = d->var();
    if (var->type == CVT_FLOAT) {
        setValue(CVar_Float(var));
    }
    else {
        setValue(CVar_Integer(var));
    }
}
void CVarSliderWidget::setCVarValueFromWidget() {
    cvar_t *var = d->var();
    if (var->type == CVT_FLOAT) {
        CVar_SetFloat(d->var(), value());
    }
    else {
        CVar_SetInteger(d->var(), round<int>(value()));
    }
}
DENG2_PIMPL_NOREF(CVarToggleWidget) {
    char const *cvar;
    cvar_t *var() const {
        cvar_t *cv = Con_FindVariable(cvar);
        DENG2_ASSERT(cv != 0);
        return cv;
    }
};
CVarToggleWidget::CVarToggleWidget(char const *cvarPath, String const &labelText)
    : d(new Impl) {
    setText(labelText);
    d->cvar = cvarPath;
    updateFromCVar();
    connect(this, SIGNAL(stateChangedByUser(ToggleWidget::ToggleState)),
            this, SLOT(setCVarValueFromWidget()));
}
char const *CVarToggleWidget::cvarPath() const {
    return d->cvar;
}
void CVarToggleWidget::updateFromCVar() {
    setActive(CVar_Integer(d->var()) != 0);
}
void CVarToggleWidget::setCVarValueFromWidget() {
    CVar_SetInteger(d->var(), isActive()? 1 : 0);
}
DENG2_PIMPL(GameWidget) {
    bool needFrames = true; // Rendering a new frame is necessary.
    VRConfig::Eye lastFrameEye = VRConfig::NeitherEye;
    Impl(Public *i) : Base(i) {}
    /**
     * Render the 3D game view of each local player. The results are stored in each
     * players' ViewCompositor as a texture. This is a slow operation and should only
     * be done once after each time game tics have been run.
     */
    void renderGameViews() {
        lastFrameEye = ClientApp::vr().currentEye();
        ClientApp::forLocalPlayers([] (ClientPlayer &player) {
            player.viewCompositor().renderGameView([] (int playerNum) {
                R_RenderViewPort(playerNum);
            });
            return LoopContinue;
        });
    }
    void renderPlayerViewToFramebuffer(int playerNum, GLFramebuffer &dest) {
        GLState::push()
                .setTarget(dest)
                .setViewport(Rectangleui::fromSize(dest.size()));
        dest.clear(GLFramebuffer::ColorDepthStencil);
        R_RenderViewPort(playerNum);
        GLState::pop();
    }
    /**
     * Draw the game widget's contents by compositing the various layers: game view,
     * border, HUD, finale, intermission, and engine/debug overlays. This is generally
     * a quick operation and can be done multiple times per window paint.
     */
    void drawComposited() {
        int numLocal = 0;
        ClientApp::forLocalPlayers([&numLocal](ClientPlayer &player) {
            player.viewCompositor().drawCompositedLayers();
            ++numLocal;
            return LoopContinue;
        });
        if (numLocal == 0) {
            ClientApp::player(0).viewCompositor().drawCompositedLayers();
        }
    }
    void draw() {
        if (ClientApp::vr().currentEye() != lastFrameEye) {
            needFrames = true;
        }
        if (needFrames) {
            App_World().beginFrame(CPP_BOOL(R_NextViewer()));
            renderGameViews();
            DGL_End();
            DGL_Flush();
            App_World().endFrame();
            needFrames = false;
        }
        drawComposited();
    }
    void updateSize() {
        LOG_AS("GameWidget");
        LOG_GL_XVERBOSE("View resized to ", self().rule().recti().size().asText());
        R_SetViewGrid(0, 0);
        /*if (!App_GameLoaded())
        {
            R_UseViewPort(0);
        }*/
    }
};
GameWidget::GameWidget(String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    requestGeometry(false);
}
void GameWidget::pause() {
    if (App_GameLoaded() && !clientPaused) {
        Con_Execute(CMDS_DDAY, "pause", true, false);
    }
}
void GameWidget::drawComposited() {
    d->drawComposited();
}
void GameWidget::renderCubeMap(uint size, String const &outputImagePath) {
    const int player = consolePlayer;
    Vec2ui fbSize(size, size);
    LOG_GL_MSG("Rendering %ix%i cube map...") << 6 * fbSize.x << fbSize.y;
    const int old_devNoCulling = devNoCulling;
    devNoCulling = 1;
    ClientPlayer &plr = ClientApp::player(player);
    const auto oldPlrFlags = plr.publicData().flags;
    plr.publicData().flags |= DDPF_CAMERA;
    App_World().beginFrame(CPP_BOOL(R_NextViewer()));
    QImage composited(QSize(6 * size, size), QImage::Format_RGB32);
    QPainter painter(&composited);
    int const baseYaw = 180;
    GLTextureFramebuffer destFb(Image::RGB_888, fbSize, 1);
    destFb.glInit();
    for (int i = 0; i < 6; ++i) {
        if (i < 4) {
            Rend_SetFixedView(player, baseYaw + 90 + i * -90, 0, 90, fbSize);
        }
        else {
            Rend_SetFixedView(player, baseYaw, i == 4? -90 : 90, 90, fbSize);
        }
        d->renderPlayerViewToFramebuffer(player, destFb);
        painter.drawImage(i * size, 0, destFb.toImage());
        GLState::considerNativeStateUndefined();
    }
    destFb.glDeinit();
    App_World().endFrame(); {
        QBuffer buf;
        buf.open(QBuffer::WriteOnly);
        composited.save(&buf, outputImagePath.fileNameExtension().mid(1).toLatin1());
        int counter = 0;
        String uniquePath = outputImagePath;
        while (FS::tryLocate<File const>(uniquePath)) {
            uniquePath = outputImagePath.fileNameAndPathWithoutExtension() +
                    String::format("-%03i", counter++) +
                    outputImagePath.fileNameExtension();
        }
        File &outFile = FS::get().root().replaceFile(uniquePath);
        outFile << Block(buf.data());
        outFile.flush();
        LOG_GL_MSG("Cube map saved to \"%s\"") << outFile.correspondingNativePath();
    }
    Rend_UnsetFixedView();
    devNoCulling = old_devNoCulling;
    plr.publicData().flags = oldPlrFlags;
}
void GameWidget::viewResized() {
    GuiWidget::viewResized();
    if (!BusyMode_Active() && !isDisabled() && !Sys_IsShuttingDown() &&
        ClientWindow::mainExists()) {
        d->updateSize();
    }
}
void GameWidget::update() {
    GuiWidget::update();
    if (isDisabled() || BusyMode_Active()) return;
    ClientWindow::main().glActivate();
    Loop_RunTics();
    d->needFrames = true;
    if (Sys_IsShuttingDown()) return;
    GL_ProcessDeferredTasks(FRAME_DEFERRED_UPLOAD_TIMEOUT);
    if (!Con_TransitionInProgress()) {
        ClientWindow::main().busy().releaseTransitionFrame();
    }
}
void GameWidget::drawContent() {
    if (isDisabled() || !GL_IsFullyInited() || !App_GameLoaded())
        return;
    root().painter().flush();
    GLState::push();
    Rectanglei pos;
    if (hasChangedPlace(pos)) {
        d->updateSize();
    }
    d->draw();
    GLState::considerNativeStateUndefined();
    GLState::pop();
}
bool GameWidget::handleEvent(Event const &event) {
    /**
     * @todo Event processing should occur here, not during Loop_RunTics().
     * However, care must be taken to reproduce the vanilla behavior of
     * controls with regard to response times.
     *
     * @todo Input drivers need to support Unicode text; for now we have to
     * submit as Latin1.
     */
    ClientWindow &window = root().window().as<ClientWindow>();
    if (event.type() == Event::MouseButton && !root().window().eventHandler().isMouseTrapped() &&
        rule().recti().contains(event.as<MouseEvent>().pos())) {
        if (!window.hasSidebar() && !window.isGameMinimized()) {
            return true;
        }
        switch (handleMouseClick(event)) {
        case MouseClickFinished:
            window.eventHandler().trapMouse();
            window.taskBar().close();
            root().setFocus(0);         // Allow input to reach here.
            root().clearFocusStack();   // Ensure no popups steal focus away.
            break;
        default:
            return true;
        }
    }
    if (event.type() == Event::KeyPress ||
        event.type() == Event::KeyRepeat ||
        event.type() == Event::KeyRelease) {
        KeyEvent const &ev = event.as<KeyEvent>();
        Keyboard_Submit(ev.state() == KeyEvent::Pressed? IKE_DOWN :
                        ev.state() == KeyEvent::Repeat?  IKE_REPEAT :
                                                         IKE_UP,
                        ev.ddKey(), ev.nativeCode(), ev.text().toLatin1());
    }
    return false;
}
void GameWidget::glDeinit() {
    GuiWidget::glDeinit();
}
D_CMD(CubeShot) {
    DENG2_UNUSED2(src, argc);
    int size = String(argv[1]).toInt();
    if (size < 8) return false;
    ClientWindow::main().game().renderCubeMap(size, "/home/cubeshot.png");
    return true;
}
DENG_GUI_PIMPL(HomeItemWidget)
, DENG2_OBSERVES(MenuWidget, ItemTriggered) {
    struct ClickHandler : public GuiWidget::IEventHandler {
        Public &owner;
        ClickHandler(Public &owner)
            : owner(owner) {}
        void acquireFocus() {
            owner.acquireFocus();
        }
        bool handleEvent(GuiWidget &widget, Event const &event) {
            if (widget.isDisabled()) return false;
            if (event.type() == Event::MouseButton) {
                MouseEvent const &mouse = event.as<MouseEvent>();
                if (owner.hitTest(event)) {
                    if (mouse.button() == MouseEvent::Right) {
                        switch (widget.handleMouseClick(event, MouseEvent::Right)) {
                        case MouseClickStarted:
                            acquireFocus();
                            return true;
                        case MouseClickAborted:
                            return true;
                        case MouseClickFinished:
                            owner.itemRightClicked();
                            emit owner.openContextMenu();
                            return true;
                        default:
                            return false; // Ignore.
                        }
                    }
                    if (mouse.state() == MouseEvent::Pressed ||
                        mouse.state() == MouseEvent::DoubleClick) {
                        acquireFocus();
                    }
                    if (mouse.state()  == MouseEvent::DoubleClick &&
                        mouse.button() == MouseEvent::Left) {
                        emit owner.doubleClicked();
                        return true;
                    }
                }
            }
            return false;
        }
    };
    Flags flags;
    AssetGroup assets;
    LabelWidget *background;
    LabelWidget *icon { nullptr };
    LabelWidget *label;
    QList<GuiWidget *> buttons;
    AnimationRule *labelRightMargin;
    IndirectRule *labelMinRightMargin = new IndirectRule;
    Rule const *buttonsWidth = nullptr;
    bool selected = false;
    bool keepButtonsVisible = false;
    bool buttonsShown = false;
    DotPath bgColor           { "transparent" };
    DotPath selectedBgColor   { "background" };
    DotPath textColor         { "text" };
    DotPath selectedTextColor { "text" };
    QTimer buttonHideTimer;
    Impl(Public *i, Flags flags) : Base(i), flags(flags) {
        labelRightMargin = new AnimationRule(0);
        self().add(background = new LabelWidget);
        if (!(flags & WithoutIcon)) {
            self().add(icon = new LabelWidget);
        }
        self().add(label = new LabelWidget);
        assets += *background;
        assets += *label;
        if (icon) {
            assets += *icon;
            icon->setBehavior(ContentClipping);
            icon->setImageFit(ui::CoverArea | ui::OriginalAspectRatio);
            icon->setSizePolicy(ui::Filled, ui::Filled);
            icon->margins().setZero();
        }
        label->setSizePolicy(ui::Filled, ui::Expand);
        label->setTextLineAlignment(ui::AlignLeft);
        label->setAlignment(ui::AlignLeft);
        label->setBehavior(ChildVisibilityClipping);
        buttonHideTimer.setSingleShot(true);
        QObject::connect(&buttonHideTimer, &QTimer::timeout, [this] () {
            for (auto *button : buttons) {
                button->setAttribute(DontDrawContent);
            }
        });
    }
    ~Impl() {
        releaseRef(labelRightMargin);
        releaseRef(labelMinRightMargin);
        releaseRef(buttonsWidth);
    }
    void updateButtonLayout() {
        SequentialLayout layout(label->rule().right() - *labelRightMargin,
                                label->rule().top(), ui::Right);
        for (auto *button : buttons) {
            if (!button->behavior().testFlag(Hidden)) {
                layout << *button;
            }
            button->rule().setMidAnchorY(label->rule().midY());
        }
        changeRef(buttonsWidth, layout.width() + rule("gap"));
        if (buttonsShown) {
            labelRightMargin->set(*buttonsWidth,
                                  labelRightMargin->animation().done()? TimeSpan(0.4) :
                                  labelRightMargin->animation().remainingTime());
        }
    }
    void showButtons(bool show) {
        buttonsShown = show;
        if (!buttonsWidth) return;
        if (show) {
            buttonHideTimer.stop();
            for (auto *button : buttons) {
                button->setAttribute(DontDrawContent, false);
                button->setBehavior(Focusable);
            }
        }
        else {
            for (auto *button : buttons) {
                button->setBehavior(Focusable, false);
            }
        }
        TimeSpan const SPAN = (self().hasBeenUpdated()? 0.4 : 0.0);
        if (show) {
            labelRightMargin->set(*buttonsWidth, SPAN/2);
        }
        else {
            labelRightMargin->set(-rule("halfunit"), SPAN);
            buttonHideTimer.setInterval(SPAN.asMilliSeconds());
            buttonHideTimer.start();
        }
    }
    void menuItemTriggered(ui::Item const &actionItem) override {
        self().parentMenu()->setInteractedItem(self().parentMenu()->organizer()
                                             .findItemForWidget(self()),
                                             &actionItem);
    }
    void updateColors() {
        auto bg = Background(style().colors().colorf(selected? selectedBgColor : bgColor));
        background->set(bg);
        label->setTextColor(selected? selectedTextColor : textColor);
        if (icon) {
            icon->setImageColor(label->textColorf());
            icon->set(bg);
        }
    }
    /**
     * Determines if this item is inside a Home column. This is not true if the item
     * is used in a standalone dialog, for example.
     */
    bool hasColumnAncestor() const {
        for (Widget *i = self().parentWidget(); i; i = i->parent()) {
            if (is<ColumnWidget>(i)) return true;
        }
        return false;
    }
};
HomeItemWidget::HomeItemWidget(Flags flags, String const &name)
    : GuiWidget(name)
    , d(new Impl(this, flags)) {
    setBehavior(Focusable | ContentClipping);
    setAttribute(AutomaticOpacity);
    addEventHandler(new Impl::ClickHandler(*this));
    AutoRef<Rule> height;
    if (flags.testFlag(AnimatedHeight)) {
        height.reset(new AnimationRule(d->label->rule().height(), 0.3));
    }
    else {
        height.reset(d->label->rule().height());
    }
    d->background->rule()
            .setInput(Rule::Top,    rule().top())
            .setInput(Rule::Height, height)
            .setInput(Rule::Left,   d->icon? d->icon->rule().right() : rule().left())
            .setInput(Rule::Right,  rule().right());
    if (d->icon) {
        d->icon->rule()
                .setSize(d->label->margins().height() +
                         style().fonts().font("default").height() +
                         style().fonts().font("default").lineSpacing(),
                         height)
                .setInput(Rule::Left, rule().left())
                .setInput(Rule::Top,  rule().top());
        d->icon->set(Background(Background::BorderGlow,
                                style().colors().colorf("home.icon.shadow"), 20));
    }
    d->label->rule()
            .setInput(Rule::Top,   rule().top())
            .setInput(Rule::Left,  d->icon? d->icon->rule().right() : rule().left())
            .setInput(Rule::Right, rule().right());
    d->label->margins().setRight(OperatorRule::maximum(*d->labelMinRightMargin,
                                                       *d->labelRightMargin) + rule("gap"));
    rule().setInput(Rule::Height, height);
}
AssetGroup &HomeItemWidget::assets() {
    return d->assets;
}
LabelWidget &HomeItemWidget::icon() {
    DENG2_ASSERT(d->icon);
    return *d->icon;
}
LabelWidget &HomeItemWidget::label() {
    return *d->label;
}
LabelWidget const &HomeItemWidget::label() const {
    return *d->label;
}
void HomeItemWidget::setSelected(bool selected) {
    if (d->selected != selected) {
        d->selected = selected;
        if (selected) {
            d->showButtons(true);
        }
        else if (!d->keepButtonsVisible) {
            d->showButtons(false);
        }
        d->updateColors();
    }
}
bool HomeItemWidget::isSelected() const {
    return d->selected;
}
void HomeItemWidget::useNormalStyle() {
    useColorTheme(Normal);
}
void HomeItemWidget::useInvertedStyle() {
    useColorTheme(Inverted);
}
void HomeItemWidget::useColorTheme(ColorTheme style) {
    useColorTheme(style, style);
}
void HomeItemWidget::useColorTheme(ColorTheme unselected, ColorTheme selected) {
    if (unselected == Inverted) {
        d->bgColor   = "inverted.background";
        d->textColor = "inverted.text";
    }
    else {
        d->bgColor   = "transparent";
        d->textColor = "text";
    }
    if (selected == Inverted) {
        d->selectedBgColor   = "home.item.background.selected.inverted";
        d->selectedTextColor = "inverted.text";
    }
    else {
        d->selectedBgColor   = "background";
        d->selectedTextColor = "text";
    }
    d->updateColors();
}
DotPath const &HomeItemWidget::textColorId() const {
    return d->textColor;
}
void HomeItemWidget::acquireFocus() {
    root().setFocus(this);
}
HomeMenuWidget *HomeItemWidget::parentMenu() {
    return maybeAs<HomeMenuWidget>(parentWidget());
}
bool HomeItemWidget::handleEvent(Event const &event) {
    if (hasFocus() && event.isKey()) {
        auto const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_LEFTARROW || key.ddKey() == DDKEY_RIGHTARROW ||
            key.ddKey() == DDKEY_UPARROW   || key.ddKey() == DDKEY_DOWNARROW) {
            if ( ! ((key.ddKey() == DDKEY_UPARROW    && isFirstChild()) ||
                    (key.ddKey() == DDKEY_DOWNARROW  && isLastChild())  ||
                    (key.ddKey() == DDKEY_LEFTARROW  && !d->hasColumnAncestor()) ||
                    (key.ddKey() == DDKEY_RIGHTARROW && !d->hasColumnAncestor())) ) {
                return false;
            }
        }
    }
    return GuiWidget::handleEvent(event);
}
void HomeItemWidget::focusGained() {
    setSelected(true);
    emit selected();
    emit mouseActivity();
}
void HomeItemWidget::focusLost() {
}
void HomeItemWidget::itemRightClicked()
{}
void HomeItemWidget::addButton(GuiWidget *widget) {
    if (auto *label = maybeAs<LabelWidget>(widget)) {
        label->setSizePolicy(ui::Expand, ui::Expand);
    }
    if (auto *menu = maybeAs<MenuWidget>(widget)) {
        menu->audienceForItemTriggered() += d;
    }
    d->buttons << widget;
    d->label->add(widget);
    widget->setAttribute(DontDrawContent);
    widget->setBehavior(Focusable, false);
    d->updateButtonLayout();
}
GuiWidget &HomeItemWidget::buttonWidget(int index) const {
    return *d->buttons.at(index);
}
void HomeItemWidget::setKeepButtonsVisible(bool yes) {
    d->keepButtonsVisible = yes;
    if (yes) {
        d->showButtons(true);
    }
    else if (!d->selected) {
        d->showButtons(false);
    }
}
void HomeItemWidget::setLabelMinimumRightMargin(Rule const &rule) {
    d->labelMinRightMargin->setSource(rule);
}
void HomeItemWidget::updateButtonLayout() {
    d->updateButtonLayout();
}
DENG_GUI_PIMPL(HomeMenuWidget)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(Asset, StateChange) {
    ui::DataPos selectedIndex = ui::Data::InvalidPos;
    ui::DataPos previousSelectedIndex = 0;
    ui::Item const *interacted = nullptr;
    ui::Item const *interactedAction = nullptr;
    Impl(Public *i) : Base(i) {
        self().organizer().audienceForWidgetCreation() += this;
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &) {
        if (is<HomeItemWidget>(widget)) {
            QObject::connect(&widget, SIGNAL(mouseActivity()),
                             thisPublic, SLOT(mouseActivityInItem()));
            QObject::connect(&widget, SIGNAL(selected()),
                             thisPublic, SLOT(itemSelectionChanged()));
        }
    }
    void assetStateChanged(Asset &asset) {
        if (asset.state() == Asset::Ready) {
            self().assets().audienceForStateChange() -= this; // only scroll once
            scrollToSelected();
        }
    }
    void scrollToSelected() {
        if (self().hasRoot()) {
            if (auto *widget = self().itemWidget<GuiWidget>(selectedIndex)) {
                self().findTopmostScrollable().scrollToWidget(*widget);
            }
        }
    }
};
HomeMenuWidget::HomeMenuWidget(String const &name)
    : MenuWidget(name)
    , d(new Impl(this)) {
    enableScrolling(false);
    enablePageKeys(false);
    setGridSize(1, ui::Filled, 0, ui::Expand);
    margins().setLeftRight("");
    layout().setRowPadding(rule("gap"));
}
void HomeMenuWidget::unselectAll() {
    if (d->selectedIndex < items().size()) {
        d->selectedIndex = ui::Data::InvalidPos;
        for (auto *w : childWidgets()) {
            if (auto *item = maybeAs<HomeItemWidget>(w)) {
                if (root().focus() != item) {
                    item->setSelected(false);
                }
            }
        }
    }
}
void HomeMenuWidget::restorePreviousSelection() {
    setSelectedIndex(d->previousSelectedIndex);
}
ui::DataPos HomeMenuWidget::selectedIndex() const {
    return d->selectedIndex;
}
const ui::Item *HomeMenuWidget::selectedItem() const {
    if (selectedIndex() < items().size()) {
        return &items().at(selectedIndex());
    }
    return nullptr;
}
void HomeMenuWidget::setSelectedIndex(ui::DataPos index) {
    DENG2_ASSERT(hasRoot());
    if (auto *widget = itemWidget<HomeItemWidget>(index)) {
        if (d->selectedIndex != index) {
            root().setFocus(nullptr);
            unselectAll();
            widget->acquireFocus();
        }
        if (root().focusIndicator().isKeyboardFocusActive()) {
            if (assets().isReady()) {
                d->scrollToSelected();
            }
            else {
                assets().audienceForStateChange() += d;
            }
        }
    }
}
void HomeMenuWidget::setInteractedItem(ui::Item const *menuItem, ui::Item const *actionItem) {
    d->interacted       = menuItem;
    d->interactedAction = actionItem;
}
ColumnWidget *HomeMenuWidget::parentColumn() const {
    for (Widget *i = parentWidget(); i; i = i->parent()) {
        if (ColumnWidget *column = maybeAs<ColumnWidget>(i)) {
            return column;
        }
    }
    return nullptr;
}
ui::Item const *HomeMenuWidget::interactedItem() const {
    return d->interacted;
}
ui::Item const *HomeMenuWidget::actionItem() const {
    return d->interactedAction;
}
void HomeMenuWidget::mouseActivityInItem() {
    if (auto *clickedWidget = dynamic_cast<HomeItemWidget *>(sender())) {
        emit itemClicked(findItem(*clickedWidget));
    }
    if (auto *column = parentColumn()) {
        emit column->mouseActivity(column);
    }
}
void HomeMenuWidget::itemSelectionChanged() {
    if (auto *clickedItem = dynamic_cast<HomeItemWidget *>(sender())) {
        ui::DataPos const newSelection = findItem(*clickedItem);
        if (d->selectedIndex != newSelection) {
            if (auto *item = itemWidget<HomeItemWidget>(d->selectedIndex)) {
                item->setSelected(false);
            }
            d->selectedIndex = d->previousSelectedIndex = newSelection;
            emit selectedIndexChanged(d->selectedIndex);
        }
    }
}
#ifdef MACOSX
#  define CONTROL_MOD   KeyEvent::Meta
#  define CONTROL_CHAR  DENG2_CHAR_MAC_CONTROL_KEY
#else
#  define CONTROL_MOD   KeyEvent::Control
#  define CONTROL_CHAR  DENG2_CHAR_CONTROL_KEY
#endif
DENG_GUI_PIMPL(InputBindingWidget)
, DENG2_OBSERVES(ButtonWidget, Press) {
    String defaultEvent;
    String command;
    QStringList contexts;
    int device = IDEV_KEYBOARD;
    bool useModifiers = false;
    Impl(Public *i) : Base(i) {
        self().setSizePolicy(ui::Fixed, ui::Expand);
        self().auxiliary().setText(_E(l) + tr("Reset"));
        self().auxiliary().audienceForPress() += this;
    }
    String prettyKey(String const &eventDesc) {
        if (!eventDesc.startsWith("key-")) {
            return eventDesc;
        }
        String name = eventDesc.substr(Rangei(4, eventDesc.indexOf("-", 4)));
        name = name.left(1).toUpper() + name.mid(1).toLower();
        int idx = eventDesc.indexOf("+");
        if (idx > 0) {
            String const conds = eventDesc.mid(idx + 1);
            if (conds.contains("key-alt-down")) {
                name = String(DENG2_CHAR_ALT_KEY) + name;
            }
            if (conds.contains("key-ctrl-down") || conds.contains("key-control-down")) {
                name = String(CONTROL_CHAR) + name;
            }
            if (conds.contains("key-shift-down")) {
                name = String(DENG2_CHAR_SHIFT_KEY) + name;
            }
        }
        return name;
    }
    void updateLabel() {
        String text = _E(l) + tr("(not bound)");
        foreach (QString bcName, contexts) {
            if (!InputSystem::get().hasContext(bcName)) continue;
            BindContext const &context = InputSystem::get().context(bcName);
            if (Record const *rec = context.findCommandBinding(command.toLatin1(), device)) {
                CommandBinding bind(*rec);
                text = prettyKey(bind.composeDescriptor());
                break;
            }
        }
        self().setText(_E(b) + text);
    }
    void bindCommand(String const &eventDesc) {
        Block const cmd = command.toLatin1();
        InputSystem::get().forAllContexts([&cmd] (BindContext &context) {
            while (Record *bind = context.findCommandBinding(cmd.constData())) {
                context.deleteBinding(bind->geti("id"));
            }
            return LoopContinue;
        });
        foreach (QString bcName, contexts) {
            String ev = String("%1:%2").arg(bcName, eventDesc);
            InputSystem::get().bindCommand(ev.toLatin1(), command.toLatin1());
        }
    }
    void buttonPressed(ButtonWidget &) {
        /*if (&button == thisPublic)
        {
            if (!self().hasFocus()) {
                focus();
            }
            else {
                unfocus();
            }
        }
        else*/
        {
            bindCommand(defaultEvent);
            updateLabel();
        }
    }
};
InputBindingWidget::InputBindingWidget() : d(new Impl(this)) {
    setBehavior(Focusable);
    auxiliary().hide();
}
void InputBindingWidget::setDefaultBinding(String const &eventDesc) {
    d->defaultEvent = eventDesc;
    auxiliary().show();
}
void InputBindingWidget::setCommand(String const &command) {
    d->command = command;
    d->updateLabel();
}
void InputBindingWidget::enableModifiers(bool mods) {
    d->useModifiers = mods;
}
void InputBindingWidget::setContexts(QStringList const &contexts) {
    d->contexts = contexts;
    d->updateLabel();
}
void InputBindingWidget::focusGained() {
    AuxButtonWidget::focusGained();
    auxiliary().disable();
    invertStyle();
}
void InputBindingWidget::focusLost() {
    AuxButtonWidget::focusLost();
    auxiliary().enable();
    invertStyle();
}
bool InputBindingWidget::handleEvent(Event const &event) {
    if (hasFocus()) {
        if (KeyEvent const *key = maybeAs<KeyEvent>(event)) {
            if (key->state() != KeyEvent::Pressed) return false;
            if (d->useModifiers && key->isModifier()) {
                return false;
            }
            if (key->ddKey() == DDKEY_ESCAPE) {
                root().popFocus();
                return true;
            }
            ddevent_t ev;
            InputSystem::convertEvent(event, ev);
            String desc = B_EventToString(ev);
            if (d->useModifiers) {
                if (key->modifiers().testFlag(KeyEvent::Shift)) {
                    desc += " + key-shift-down";
                }
                else {
                    desc += " + key-shift-up";
                }
                if (key->modifiers().testFlag(KeyEvent::Alt)) {
                    desc += " + key-alt-down";
                }
                else {
                    desc += " + key-alt-up";
                }
                if (key->modifiers().testFlag(CONTROL_MOD)) {
                    desc += " + key-ctrl-down";
                }
                else {
                    desc += " + key-ctrl-up";
                }
            }
            d->bindCommand(desc);
            d->updateLabel();
            root().popFocus();
            return true;
        }
        /*if (MouseEvent const *mouse = maybeAs<MouseEvent>(event))
        {
            if (mouse->type()  == Event::MouseButton &&
                mouse->state() == MouseEvent::Released &&
                !hitTest(event)) {
                d->unfocus();
                return true;
            }
        }*/
    }
    return AuxButtonWidget::handleEvent(event);
}
InputBindingWidget *InputBindingWidget::newTaskBarShortcut() {
    InputBindingWidget *bind = new InputBindingWidget;
    bind->setCommand("taskbar");
    bind->setDefaultBinding("key-tilde-down + key-shift-up");
    bind->enableModifiers(true);
    bind->setContexts(QStringList() << "global" << "console");
    return bind;
}
DENG_GUI_PIMPL(KeyGrabberWidget) {
    Impl(Public *i) : Base(i)
    {}
    void focus() {
        root().setFocus(thisPublic);
        self().set(Background(Background::GradientFrame, style().colors().colorf("accent"), 6));
        self().setText(tr("Waiting for a key..."));
    }
    void unfocus() {
        root().setFocus(0);
        self().set(Background());
        self().setText(tr("Click to focus"));
    }
};
KeyGrabberWidget::KeyGrabberWidget(String const &name)
    : LabelWidget(name), d(new Impl(this)) {
    setBehavior(Focusable);
    setTextLineAlignment(AlignLeft);
    setText(tr("Click to focus"));
}
bool KeyGrabberWidget::handleEvent(Event const &event) {
    if (!hasFocus()) {
        switch (handleMouseClick(event)) {
        case MouseClickFinished:
            d->focus();
            return true;
        default:
            return false;
        }
    }
    else {
        if (KeyEvent const *key = maybeAs<KeyEvent>(event)) {
            if (key->ddKey() == DDKEY_ESCAPE) {
                d->unfocus();
                return true;
            }
            ddevent_t ev;
            InputSystem::convertEvent(event, ev);
            String info = String("DD:%1 Qt:0x%2 Native:0x%3\n" _E(m) "%4")
                              .arg(key->ddKey())
                              .arg(key->qtKey(), 0, 16)
                              .arg(key->nativeCode(), 0, 16)
                              .arg(B_EventToString(ev));
            setText(info);
            return true;
        }
        if (MouseEvent const *mouse = maybeAs<MouseEvent>(event)) {
            if (mouse->type() == Event::MouseButton &&
               mouse->state() == MouseEvent::Released &&
               !hitTest(event)) {
                d->unfocus();
                return true;
            }
        }
    }
    return false;
}
DENG_GUI_PIMPL(MapOutlineWidget) {
    ProgressWidget *progress; // shown initially, before outline received
    Rectanglei mapBounds;
    DefaultVertexBuf *vbuf = nullptr;
    Drawable drawable;
    GLUniform uMvpMatrix { "uMvpMatrix", GLUniform::Mat4 };
    GLUniform uColor     { "uColor",     GLUniform::Vec4 };
    Animation mapOpacity { 0, Animation::Linear };
    Impl(Public *i) : Base(i) {
        self().add(progress = new ProgressWidget);
        progress->setMode(ProgressWidget::Indefinite);
        progress->setColor("progress.dark.wheel");
        progress->setShadowColor("progress.dark.shadow");
        progress->rule().setRect(self().rule());
    }
    void glInit() {
        vbuf = new DefaultVertexBuf;
        drawable.addBuffer(vbuf);
        shaders().build(drawable.program(), "generic.color_ucolor")
                << uMvpMatrix << uColor;
    }
    void glDeinit() {
        drawable.clear();
        vbuf = nullptr;
    }
    void makeOutline(shell::MapOutlinePacket const &mapOutline) {
        if (!vbuf) return;
        root().window().glActivate();
        progress->setOpacity(0, 0.5);
        mapOpacity.setValue(1, 0.5);
        mapBounds = Rectanglei();
        Vector4f const oneSidedColor = style().colors().colorf("inverted.altaccent");
        Vector4f const twoSidedColor = style().colors().colorf("altaccent");
        DefaultVertexBuf::Builder verts;
        DefaultVertexBuf::Type vtx;
        for (int i = 0; i < mapOutline.lineCount(); ++i) {
            auto const &line = mapOutline.line(i);
            vtx.rgba = (line.type == shell::MapOutlinePacket::OneSidedLine? oneSidedColor : twoSidedColor);
            vtx.pos = line.start; verts << vtx;
            vtx.pos = line.end;   verts << vtx;
            if (i > 0) {
                mapBounds.include(line.start);
            }
            else {
                mapBounds = Rectanglei(line.start, line.start);
            }
            mapBounds.include(line.end);
        }
        vbuf->setVertices(gl::Lines, verts, gl::Static);
        root().window().glDone();
    }
    Matrix4f modelMatrix() const {
        DENG2_ASSERT(vbuf);
        if (mapBounds.isNull()) return Matrix4f();
        Rectanglef const rect = self().contentRect();
        float const scale = de::min(rect.width()  / mapBounds.width(),
                                    rect.height() / mapBounds.height());
        return Matrix4f::translate(rect.middle()) *
               Matrix4f::scale    (Vector3f(scale, -scale, 1)) *
               Matrix4f::translate(Vector2f(-mapBounds.middle()));
    }
};
MapOutlineWidget::MapOutlineWidget(String const &name)
    : GuiWidget(name)
    , d(new Impl(this))
{}
void MapOutlineWidget::setOutline(shell::MapOutlinePacket const &mapOutline) {
    d->makeOutline(mapOutline);
}
void MapOutlineWidget::drawContent() {
    GuiWidget::drawContent();
    if (d->vbuf && d->vbuf->count()) {
        auto &painter = root().painter();
        painter.flush();
        GLState::push().setNormalizedScissor(painter.normalizedScissor());
        d->uMvpMatrix = root().projMatrix2D() * d->modelMatrix();
        d->uColor = Vector4f(1, 1, 1, d->mapOpacity * visibleOpacity());
        d->drawable.draw();
        GLState::pop();
    }
}
void MapOutlineWidget::glInit() {
    GuiWidget::glInit();
    d->glInit();
}
void MapOutlineWidget::glDeinit() {
    GuiWidget::glDeinit();
    d->glDeinit();
}
DENG2_PIMPL(MultiplayerServerMenuWidget)
, DENG2_OBSERVES(DoomsdayApp, GameChange)
, DENG2_OBSERVES(Games, Readiness)
, DENG2_OBSERVES(ServerLink, DiscoveryUpdate)
, DENG2_OBSERVES(MultiplayerPanelButtonWidget, AboutToJoin)
, public ChildWidgetOrganizer::IWidgetFactory {
    static ServerLink &link() { return ClientApp::serverLink(); }
    static String hostId(shell::ServerInfo const &sv) {
        if (sv.serverId()) {
            return String::format("%x", sv.serverId());
        }
        return sv.address().asText();
    }
    /**
     * Data item with information about a found server.
     */
    class ServerListItem : public ui::Item {
    public:
        ServerListItem(shell::ServerInfo const &serverInfo, bool isLocal)
            : _lan(isLocal) {
            setData(hostId(serverInfo));
            _info = serverInfo;
        }
        bool isLocal() const {
            return _lan;
        }
        void setLocal(bool isLocal) {
            _lan = isLocal;
        }
        shell::ServerInfo const &info() const {
            return _info;
        }
        void setInfo(shell::ServerInfo const &serverInfo) {
            _info = serverInfo;
            notifyChange();
        }
        String title() const {
            return _info.name();
        }
        String gameId() const {
            return _info.gameId();
        }
    private:
        shell::ServerInfo _info;
        bool _lan;
    };
    DiscoveryMode                mode = NoDiscovery;
    ServerLink::FoundMask        mask = ServerLink::Any;
    Impl(Public *i) : Base(i) {
        DoomsdayApp::app().audienceForGameChange()  += this;
        DoomsdayApp::games().audienceForReadiness() += this;
        link().audienceForDiscoveryUpdate() += this;
        self().organizer().setWidgetFactory(*this);
    }
    void linkDiscoveryUpdate(ServerLink const &link) override {
        ui::Data &items = self().items();
        QSet<String> foundHosts;
        foreach (Address const &host, link.foundServers(mask)) {
            shell::ServerInfo info;
            if (link.foundServerInfo(host, info, mask)) {
                foundHosts.insert(hostId(info));
            }
        }
        for (ui::Data::Pos idx = 0; idx < items.size(); ++idx) {
            String const id = items.at(idx).data().toString();
            if (!foundHosts.contains(id)) {
                items.remove(idx--);
            }
        }
        foreach (Address const &host, link.foundServers(mask)) {
            shell::ServerInfo info;
            if (!link.foundServerInfo(host, info, mask)) continue;
            ui::Data::Pos found   = items.findData(hostId(info));
            const bool    isLocal = link.isServerOnLocalNetwork(info.address());
            if (found == ui::Data::InvalidPos) {
                items.append(new ServerListItem(info, isLocal));
            }
            else {
                auto &it = items.at(found).as<ServerListItem>();
                if (!it.isLocal() || isLocal) {
                    it.setInfo(info);
                    it.setLocal(isLocal);
                }
            }
        }
        items.stableSort([] (ui::Item const &a, ui::Item const &b) {
            auto const &first  = a.as<ServerListItem>();
            auto const &second = b.as<ServerListItem>();
            if (first.isLocal() == second.isLocal()) {
                if (first.info().playerCount() == second.info().playerCount()) {
                    int cmp = first.info().gameId().compareWithCase(second.info().gameId());
                    if (!cmp) {
                        return first.info().name().compareWithoutCase(second.info().name()) < 0;
                    }
                    return cmp < 0;
                }
                return first.info().playerCount() - second.info().playerCount() > 0;
            }
            return first.isLocal();
        });
    }
    void currentGameChanged(Game const &newGame) override {
        if (newGame.isNull() && mode == DiscoverUsingMaster) {
            link().discoverUsingMaster();
        }
    }
    void gameReadinessUpdated() override {
        foreach (GuiWidget *w, self().childWidgets()) {
            updateAvailability(*w);
        }
    }
    void updateAvailability(GuiWidget &menuItemWidget) {
        auto const &item = self().organizer().findItemForWidget(menuItemWidget)->as<ServerListItem>();
        bool playable = false;
        String gameId = item.gameId();
        if (DoomsdayApp::games().contains(gameId)) {
            playable = DoomsdayApp::games()[gameId].isPlayable();
        }
        menuItemWidget.enable(playable);
    }
    void aboutToJoinMultiplayerGame(shell::ServerInfo const &sv) override {
        DENG2_FOR_PUBLIC_AUDIENCE2(AboutToJoin, i) i->aboutToJoinMultiplayerGame(sv);
    }
    GuiWidget *makeItemWidget(ui::Item const &, GuiWidget const *) override {
        auto *b = new MultiplayerPanelButtonWidget;
        b->audienceForAboutToJoin() += this;
        return b;
    }
    void updateItemWidget(GuiWidget &widget, ui::Item const &item) override {
        auto const &serverItem = item.as<ServerListItem>();
        widget.as<MultiplayerPanelButtonWidget>()
                .updateContent(serverItem.info());
        updateAvailability(widget);
    }
    DENG2_PIMPL_AUDIENCE(AboutToJoin)
};
DENG2_AUDIENCE_METHOD(MultiplayerServerMenuWidget, AboutToJoin);
MultiplayerServerMenuWidget::MultiplayerServerMenuWidget(DiscoveryMode discovery,
                                                         String const &name)
    : HomeMenuWidget(name)
    , d(new Impl(this)) {
    d->mode = discovery;
    switch (d->mode) {
    case DiscoverUsingMaster:
        d->mask = ServerLink::LocalNetwork | ServerLink::MasterServer;
        d->link().discoverUsingMaster();
        break;
    case DirectDiscoveryOnly:
        d->mask = ServerLink::Direct;
        break;
    case NoDiscovery:
        break;
    }
}
enum {
    POS_STATUS = 1
};
DENG_GUI_PIMPL(MultiplayerStatusWidget)
, DENG2_OBSERVES(ServerLink, Join)
, DENG2_OBSERVES(ServerLink, Leave) {
    QTimer timer;
    Impl(Public *i) : Base(i) {
        timer.setInterval(1000);
        link().audienceForJoin() += this;
        link().audienceForLeave() += this;
    }
    void networkGameJoined() {
        /*self().menu().organizer().itemWidget(POS_SERVER_ADDRESS)->as<LabelWidget>()
                .setText(_E(l) + tr("Server:") + _E(.) +
                         " " + link().address().asText());*/
    }
    void networkGameLeft() {
    }
    static ServerLink &link() {
        return ClientApp::serverLink();
    }
};
MultiplayerStatusWidget::MultiplayerStatusWidget()
    : PopupMenuWidget("multiplayer-menu"), d(new Impl(this)) {
    connect(&d->timer, SIGNAL(timeout()), this, SLOT(updateElapsedTime()));
    items()
            << new ui::ActionItem(tr("Disconnect"), new CommandAction("net disconnect"))
            << new ui::Item(ui::Item::ShownAsLabel, ""); // sv address & time
}
void MultiplayerStatusWidget::updateElapsedTime() {
    if (d->link().status() != ServerLink::Connected)
        return;
    TimeSpan const elapsed = d->link().connectedAt().since();
    items().at(POS_STATUS).setLabel(
                _E(s)_E(l) + tr("Server:") + _E(.) " " + d->link().address().asText() + "\n"
                _E(l) + tr("Connected:") + _E(.) +
                String(" %1:%2:%3")
                .arg(int(elapsed.asHours()))
                .arg(int(elapsed.asMinutes()) % 60, 2, 10, QLatin1Char('0'))
                .arg(int(elapsed) % 60, 2, 10, QLatin1Char('0')));
}
void MultiplayerStatusWidget::preparePanelForOpening() {
    d->timer.start();
    updateElapsedTime();
    PopupMenuWidget::preparePanelForOpening();
}
void MultiplayerStatusWidget::panelClosing() {
    d->timer.stop();
    PopupMenuWidget::panelClosing();
}
DENG2_PIMPL(NativePathWidget) {
    NativePath       path;
    QStringList      filters;
    PopupMenuWidget *menu;
    String           blankText = "(not set)";
    String           prompt = "Select File";
    Impl(Public *i)
        : Base(i)
    {}
    String labelText() const {
        if (path.isEmpty()) {
            return String(_E(l)) + blankText + _E(.);
        }
        return path.fileName();
    }
    void notifyChange() {
        DENG2_FOR_PUBLIC_AUDIENCE2(UserChange, i) {
            i->pathChangedByUser(self());
        }
    }
    DENG2_PIMPL_AUDIENCE(UserChange)
};
DENG2_AUDIENCE_METHOD(NativePathWidget, UserChange)
NativePathWidget::NativePathWidget()
    : d(new Impl(this)) {
    add(d->menu = new PopupMenuWidget);
    d->menu->setAnchorAndOpeningDirection(rule(), ui::Up);
    d->menu->items()
            << new ui::ActionItem(tr("Browse..."),
                                  new SignalAction(this, SLOT(chooseUsingNativeFileDialog())))
            << new ui::ActionItem(style().images().image("close.ring"), tr("Reset"),
                                  new SignalAction(this, SLOT(clearPath())));
    auxiliary().setText(tr("Browse"));
    connect(&auxiliary(), SIGNAL(pressed()), this, SLOT(chooseUsingNativeFileDialog()));
    connect(this, SIGNAL(pressed()), this, SLOT(showActionsPopup()));
}
void NativePathWidget::setFilters(StringList const &filters) {
    d->filters.clear();
    for (auto const &f : filters) {
        d->filters << f;
    }
}
void NativePathWidget::setBlankText(String const &text) {
    d->blankText = text;
    setText(d->labelText());
}
void NativePathWidget::setPrompt(const String &promptText) {
    d->prompt = promptText;
}
NativePath NativePathWidget::path() const {
    return d->path;
}
void NativePathWidget::setPath(const NativePath &path) {
    if (d->path != path) {
        d->path = path;
        setText(d->labelText());
        d->notifyChange();
    }
}
void NativePathWidget::chooseUsingNativeFileDialog() {
    ClientApp::app().beginNativeUIMode();
    QDir dir(d->path);
    if (d->path.isEmpty()) dir = QDir::home();
    QFileDialog dlg(nullptr, d->prompt, dir.absolutePath());
    if (!d->filters.isEmpty()) {
        dlg.setNameFilters(d->filters);
    }
    dlg.setFileMode(QFileDialog::ExistingFile);
    dlg.setOption(QFileDialog::ReadOnly, true);
    dlg.setLabelText(QFileDialog::Accept, tr("Select"));
    if (dlg.exec()) {
        setPath(dlg.selectedFiles().at(0));
    }
    ClientApp::app().endNativeUIMode();
}
void NativePathWidget::clearPath() {
    setPath({});
}
void NativePathWidget::showActionsPopup() {
    if (!d->menu->isOpen()) {
        d->menu->open();
    }
    else {
        d->menu->close(0);
    }
}
DENG_GUI_PIMPL(PackageContentOptionsWidget)
, public ChildWidgetOrganizer::IWidgetFactory {
    /**
     * Item representing a contained package.
     */
    struct Item : public ui::Item {
        bool selectedByDefault;
        String containerPackageId;
        String category;
        Item(String const &packageId, bool selectedByDefault, String const &containerPackageId)
            : selectedByDefault(selectedByDefault)
            , containerPackageId(containerPackageId) {
            setData(packageId);
            if (File const *file = PackageLoader::get().select(packageId)) {
                Record const &meta = file->objectNamespace();
                setLabel(meta.gets(Package::VAR_PACKAGE_TITLE));
                category = meta.gets(QStringLiteral("package.category"), "");
            }
            else {
                setLabel(packageId);
            }
            if (!selectedByDefault) {
                setLabel(label() + " " _E(s)_E(b)_E(D) "ALT");
            }
        }
        String packageId() const {
            return data().toString();
        }
        DictionaryValue &conf() {
            return Config::get("fs.selectedPackages").value<DictionaryValue>();
        }
        DictionaryValue const &conf() const {
            return Config::get("fs.selectedPackages").value<DictionaryValue>();
        }
        bool isSelected() const {
            if (conf().contains(TextValue(containerPackageId))) {
                DictionaryValue const &sel = conf().element(TextValue(containerPackageId))
                                             .as<DictionaryValue>();
                if (sel.contains(TextValue(packageId()))) {
                    return sel.element(TextValue(packageId())).isTrue();
                }
            }
            return selectedByDefault;
        }
        void setSelected(bool selected) {
            if (!conf().contains(TextValue(containerPackageId))) {
                conf().add(new TextValue(containerPackageId),
                           new DictionaryValue);
            }
            DictionaryValue &sel = conf().element(TextValue(containerPackageId))
                                   .as<DictionaryValue>();
            sel.add(new TextValue(packageId()), new NumberValue(selected));
            notifyChange();
        }
        void reset() {
            conf().remove(TextValue(containerPackageId));
            notifyChange();
        }
    };
    String packageId;
    LabelWidget *summary;
    MenuWidget *contents;
    Impl(Public *i, String const &packageId, Rule const &maxHeight)
        : Base(i)
        , packageId(packageId) {
        self().add(summary = new LabelWidget);
        summary->setSizePolicy(ui::Fixed, ui::Expand);
        summary->setFont("small");
        summary->setTextColor("altaccent");
        summary->setAlignment(ui::AlignLeft);
        auto *label = LabelWidget::newWithText(tr("Select:"), thisPublic);
        label->setSizePolicy(ui::Expand, ui::Expand);
        label->setFont("small");
        label->setAlignment(ui::AlignLeft);
        auto *allButton = new ButtonWidget;
        allButton->setSizePolicy(ui::Expand, ui::Expand);
        allButton->setText(tr("All"));
        allButton->setFont("small");
        self().add(allButton);
        auto *noneButton = new ButtonWidget;
        noneButton->setSizePolicy(ui::Expand, ui::Expand);
        noneButton->setText(tr("None"));
        noneButton->setFont("small");
        self().add(noneButton);
        auto *defaultsButton = new ButtonWidget;
        defaultsButton->setSizePolicy(ui::Expand, ui::Expand);
        defaultsButton->setText(tr("Defaults"));
        defaultsButton->setFont("small");
        self().add(defaultsButton);
        contents = new MenuWidget;
        contents->enableIndicatorDraw(true);
        contents->setBehavior(ChildVisibilityClipping);
        contents->setVirtualizationEnabled(true, rule(RuleBank::UNIT).value()*2 +
                                           style().fonts().font("default").height().value());
        contents->layout().setRowPadding(Const(0));
        contents->organizer().setWidgetFactory(*this);
        contents->setGridSize(1, ui::Filled, 0, ui::Fixed);
        self().add(contents);
        auto &rect = self().rule();
        SequentialLayout layout(rect.left() + self().margins().left(),
                                rect.top()  + self().margins().top());
        layout << *label << *summary << *contents;
        summary->rule()
                .setInput(Rule::Width,  rect.width() - self().margins().width());
        contents->rule()
                .setInput(Rule::Left,   rect.left())
                .setInput(Rule::Width,  summary->rule().width() + self().margins().left())
                .setInput(Rule::Height, OperatorRule::minimum(contents->contentHeight(),
                                                              maxHeight - self().margins().height() -
                                                              label->rule().height() -
                                                              summary->rule().height()));
        SequentialLayout(label->rule().right(), label->rule().top(), ui::Right)
               << *defaultsButton << *noneButton << *allButton;
        self().rule().setInput(Rule::Height, layout.height() + self().margins().bottom());
        for (GuiWidget *w : self().childWidgets()) {
            w->margins().set("dialog.gap");
        }
        contents->margins().setLeftRight("gap");
        defaultsButton->setActionFn([this] () {
            contents->items().forAll([] (ui::Item &item) {
                if (auto *i = maybeAs<Item>(item)) i->reset();
                return LoopContinue;
            });
        });
        noneButton->setActionFn([this] () {
            contents->items().forAll([] (ui::Item &item) {
                if (auto *i = maybeAs<Item>(item)) i->setSelected(false);
                return LoopContinue;
            });
        });
        allButton->setActionFn([this] () {
            contents->items().forAll([] (ui::Item &item) {
                if (auto *i = maybeAs<Item>(item)) i->setSelected(true);
                return LoopContinue;
            });
        });
    }
    void populate() {
        contents->items().clear();
        File const *file = PackageLoader::get().select(packageId);
        if (!file) return;
        Record const &meta = file->objectNamespace().subrecord(Package::VAR_PACKAGE);
        ArrayValue const &requires   = meta.geta("requires");
        ArrayValue const &recommends = meta.geta("recommends");
        ArrayValue const &extras     = meta.geta("extras");
        auto const totalCount    = requires.size() + recommends.size() + extras.size();
        auto const optionalCount = recommends.size() + extras.size();
        summary->setText(tr("%1 package%2 (%3 optional)")
                         .arg(totalCount)
                         .arg(DENG2_PLURAL_S(totalCount))
                         .arg(optionalCount));
        makeItems(recommends, true);
        makeItems(extras,     false);
        QSet<QString> categories;
        contents->items().forAll([&categories] (ui::Item const &i) {
            String const cat = i.as<Item>().category;
            if (!cat.isEmpty()) categories.insert(cat);
            return LoopContinue;
        });
        for (QString cat : categories) {
            contents->items() << new ui::Item(ui::Item::Separator, cat);
        }
        contents->items().sort([] (ui::Item const &s, ui::Item const &t) {
            if (!s.isSeparator() && !t.isSeparator()) {
                Item const &a = s.as<Item>();
                Item const &b = t.as<Item>();
                int const catComp = a.category.compareWithoutCase(b.category);
                if (!catComp) {
                    int const nameComp = a.label().compareWithoutCase(b.label());
                    return nameComp < 0;
                }
                return catComp < 0;
            }
            if (s.isSeparator() && !t.isSeparator()) {
                return s.label().compareWithoutCase(t.as<Item>().category) <= 0;
            }
            if (!s.isSeparator() && t.isSeparator()) {
                return s.as<Item>().category.compareWithoutCase(t.label()) < 0;
            }
            return s.label().compareWithoutCase(t.label()) < 0;
        });
    }
    void makeItems(ArrayValue const &ids, bool recommended) {
        for (Value const *value : ids.elements()) {
            contents->items() << new Item(value->asText(), recommended, packageId);
        }
    }
    GuiWidget *makeItemWidget(ui::Item const &item, GuiWidget const *) override {
        if (item.semantics() & ui::Item::Separator) {
            auto *label = new LabelWidget;
            label->setFont("separator.label");
            label->setTextColor("accent");
            label->margins().setBottom(RuleBank::UNIT);
            label->setSizePolicy(ui::Fixed, ui::Expand);
            label->setAlignment(ui::AlignLeft);
            return label;
        }
        auto *toggle = new ToggleWidget;
        toggle->setSizePolicy(ui::Fixed, ui::Expand);
        toggle->setAlignment(ui::AlignLeft);
        toggle->set(Background());
        toggle->margins().setTopBottom(RuleBank::UNIT);
        QObject::connect(toggle, &ToggleWidget::stateChangedByUser,
                         [&item] (ToggleWidget::ToggleState active) {
            const_cast<ui::Item &>(item).as<Item>()
                    .setSelected(active == ToggleWidget::Active);
        });
        return toggle;
    }
    void updateItemWidget(GuiWidget &widget, ui::Item const &item) override {
        LabelWidget &label = widget.as<LabelWidget>();
        label.setText(item.label());
        if (!(item.semantics() & ui::Item::Separator)) {
            widget.as<ToggleWidget>().setActive(item.as<Item>().isSelected());
        }
    }
};
PackageContentOptionsWidget::PackageContentOptionsWidget(String const &packageId,
                                                         Rule   const &maxHeight,
                                                         String const &name)
    : GuiWidget(name)
    , d(new Impl(this, packageId, maxHeight)) {
    d->populate();
}
PopupWidget *PackageContentOptionsWidget::makePopup(String const &packageId,
                                                    Rule const &width,
                                                    Rule const &maxHeight) {
    PopupWidget *pop = new PopupWidget;
    pop->setOutlineColor("popup.outline");
    pop->setDeleteAfterDismissed(true);
    /*pop->closeButton().setActionFn([this] ()
    {
        root().setFocus(this);
        _optionsPopup->close();
    });*/
    auto *opts = new PackageContentOptionsWidget(packageId, maxHeight);
    opts->rule().setInput(Rule::Width, width);
    pop->setContent(opts);
    return pop;
}
DENG_GUI_PIMPL(PackagesButtonWidget) {
    StringList packages;
    String dialogTitle;
    DotPath dialogIcon { "package.icon" };
    String labelPrefix;
    String noneLabel;
    GameProfile const *profile = nullptr;
    std::function<void (PackagesDialog &)> setupFunc;
    String overrideLabel;
    Impl(Public *i) : Base(i)
    {}
    void updateLabel() {
        if (overrideLabel) {
            self().setImage(nullptr);
            self().setText(overrideLabel);
            self().setTextColor(self().colorTheme() == Normal? "accent" : "inverted.accent");
        }
        else {
            self().setStyleImage("package.icon");
            if (packages.isEmpty()) {
                self().setText(noneLabel);
                self().setTextColor(self().colorTheme() == Normal? "text" : "inverted.text");
                if (!noneLabel.isEmpty()) self().setImage(nullptr);
            }
            else {
                self().setText(labelPrefix + String::format("%i", packages.count()));
                self().setTextColor(self().colorTheme() == Normal? "accent" : "inverted.accent");
            }
            self().setImageColor(self().textColorf());
        }
    }
    void pressed() {
        auto *dlg = new PackagesDialog(dialogTitle);
        dlg->heading().setStyleImage(dialogIcon);
        if (profile) {
            dlg->setProfile(*profile);
        }
        dlg->setDeleteAfterDismissed(true);
        dlg->setSelectedPackages(packages);
        dlg->setAcceptanceAction(new CallbackAction([this, dlg] () {
            packages = dlg->selectedPackages();
            updateLabel();
            QStringList ids;
            for (auto const &p : packages) ids << p;
            emit self().packageSelectionChanged(ids);
        }));
        root().addOnTop(dlg);
        if (setupFunc) setupFunc(*dlg);
        dlg->open();
    }
};
PackagesButtonWidget::PackagesButtonWidget()
    : d(new Impl(this)) {
    setOverrideImageSize(style().fonts().font("default").height());
    setSizePolicy(ui::Expand, ui::Expand);
    setTextAlignment(ui::AlignLeft);
    connect(this, &ButtonWidget::pressed, [this] () { d->pressed(); });
    d->updateLabel();
}
void PackagesButtonWidget::setGameProfile(GameProfile const &profile) {
    d->profile = &profile;
}
void PackagesButtonWidget::setSetupCallback(std::function<void (PackagesDialog &)> func) {
    d->setupFunc = func;
}
void PackagesButtonWidget::setLabelPrefix(String const &labelPrefix) {
    d->labelPrefix = labelPrefix;
    d->updateLabel();
}
void PackagesButtonWidget::setNoneLabel(String const &noneLabel) {
    d->noneLabel = noneLabel;
    d->updateLabel();
}
void PackagesButtonWidget::setOverrideLabel(const String &overrideLabel) {
    d->overrideLabel = overrideLabel;
    d->updateLabel();
}
void PackagesButtonWidget::setDialogTitle(String const &title) {
    d->dialogTitle = title;
}
void PackagesButtonWidget::setDialogIcon(DotPath const &imageId) {
    d->dialogIcon = imageId;
}
void PackagesButtonWidget::setPackages(StringList packageIds) {
    d->packages = packageIds;
    d->updateLabel();
}
StringList PackagesButtonWidget::packages() const {
    return d->packages;
}
DENG_GUI_PIMPL(PackagesSidebarWidget) {
    PackagesWidget *browser;
    Impl(Public *i) : Base(i) {
        GuiWidget *container = &self().containerWidget();
        container->add(browser = new PackagesWidget);
        browser->setRightClickToOpenContextMenu(true);
        browser->rule().setInput(Rule::Width, rule("sidebar.width"));
        browser->actionItems().insert(
            0,
            new ui::ActionItem(
                tr("..."), new CallbackAction([this]() {
                    auto *pop = new PackageInfoDialog(browser->actionPackage(),
                                                      PackageInfoDialog::EnableActions);
                    root().addOnTop(pop);
                    pop->setDeleteAfterDismissed(true);
                    pop->setAnchorAndOpeningDirection(
                        browser->actionWidget()->as<HomeItemWidget>().buttonWidget(0).rule(),
                        ui::Up);
                    pop->open();
                })));
    }
};
PackagesSidebarWidget::PackagesSidebarWidget()
    : SidebarWidget("Mods", "packages-sidebar")
    , d(new Impl(this)) {
    auto *refreshButton = new ButtonWidget;
    closeButton().parentGuiWidget()->add(refreshButton);
    refreshButton->setSizePolicy(ui::Expand, ui::Fixed);
    refreshButton->rule()
            .setInput(Rule::Right,  closeButton().rule().left())
            .setInput(Rule::Top,    closeButton().rule().top())
            .setInput(Rule::Height, closeButton().rule().height());
    refreshButton->setStyleImage("refresh", "default");
    refreshButton->setActionFn([]() { FS::get().refreshAsync(); });
    d->browser->setFilterEditorMinimumY(closeButton().rule().bottom());
    layout() << *d->browser;
    updateSidebarLayout();
    d->browser->progress().rule().setRect(rule());
}
static String const VAR_TITLE("title");
static String const VAR_TAGS("tags");
static String const TAG_HIDDEN("hidden");
static String const TAG_LOADED("loaded");
static String const TAG_CACHED("cached");
static TimeSpan const REFILTER_DELAY(0.2);
struct PackageLoadStatus : public PackagesWidget::IPackageStatus {
    bool isPackageHighlighted(String const &packageId) const {
        return App::packageLoader().isLoaded(packageId);
    }
};
static PackageLoadStatus isPackageLoaded;
PackagesWidget::IPackageStatus::~IPackageStatus() {}
DENG_GUI_PIMPL(PackagesWidget)
, DENG2_OBSERVES(FileSystem, Busy)
, public ChildWidgetOrganizer::IWidgetFactory {
    /**
     * Information about an available package.
     */
    struct PackageItem
        : public ui::Item
        , DENG2_OBSERVES(File, Deletion) {
        SafePtr<const File> file;
        const Record *      info = nullptr;
        NativePath          nativePath;
        PackageItem(File const &packFile) {
            setFile(packFile);
        }
        void setFile(File const &packFile) {
            if (file) {
                file->audienceForDeletion() -= this;
            }
            DENG2_GUARD(packFile);
            packFile.audienceForDeletion() += this;
            file.reset(&packFile);
            if (const auto *nat = maybeAs<NativeFile>(packFile.source())) {
                nativePath = nat->nativePath();
            }
            else {
                nativePath.clear();
            }
            info = &file->objectNamespace().subrecord(Package::VAR_PACKAGE);
            setData(Package::versionedIdentifierForFile(packFile));
            setLabel(info->gets(Package::VAR_TITLE));
            notifyChange();
        }
        void fileBeingDeleted(File const &) {
            info = nullptr;
            nativePath.clear();
        }
        bool isLoaded() const {
            DENG2_GUARD(file);
            if (!file) return false;
            return PackageLoader::get().isLoaded(*file);
        }
    };
    /**
     * Widget showing information about a package and containing buttons for manipulating
     * the package.
     */
    class PackageListItemWidget
        : public HomeItemWidget
        , DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation) // actions
        , DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate)   // actions
        , DENG2_OBSERVES(Bank, Load)                           // package icons
    {
    public:
        PackageListItemWidget(PackageItem const &item, PackagesWidget &owner)
            : HomeItemWidget(NonAnimatedHeight) // virtualized, so don't make things difficult
            , _owner(owner)
            , _item(&item) {
            icon().setImageFit(ui::FitToWidth | ui::OriginalAspectRatio);
            icon().margins().set("gap").setRight("");
            icon().rule().setInput(Rule::Width, style().fonts().font("default").height() * 3);
            _actions = new MenuWidget;
            _actions->enablePageKeys(false);
            _actions->enableScrolling(false);
            _actions->margins().setZero();
            _actions->organizer().audienceForWidgetCreation() += this;
            _actions->organizer().audienceForWidgetUpdate() += this;
            _actions->setGridSize(0, ui::Expand, 1, ui::Expand);
            _actions->setItems(*owner.d->actionItems);
            connect(this, &HomeItemWidget::doubleClicked, [this]() { triggerAction(); });
            addButton(_actions);
            setKeepButtonsVisible(!_owner.d->actionOnlyForSelection);
            createTagButtons();
            fetchIcon();
        }
        void setItem(PackageItem const &item) {
            if (_item != &item) {
                iconBank().audienceForLoad() -= this;
                setSelected(false);
                _packagePath.clear();
                _item = &item;
                destroyTagButtons();
                createTagButtons();
                fetchIcon();
            }
        }
        void createTagButtons() {
            if (!_item->file) return;
            SequentialLayout layout(label().rule().left() + label().margins().left(),
                                    label().rule().bottom() - label().margins().bottom(),
                                    ui::Right);
            for (QString tag : Package::tags(*_item->file)) {
                auto *btn = new ButtonWidget;
                btn->setText(_E(l) + tag.toLower());
                btn->setActionFn([this, tag]() {
                    String terms = _owner.d->search->text();
                    if (!terms.isEmpty() && !terms.last().isSpace()) terms += " ";
                    terms += tag.toLower();
                    _owner.d->search->setText(terms);
                });
                updateTagButtonStyle(btn, "accent");
                btn->setSizePolicy(ui::Expand, ui::Expand);
                btn->margins()
                    .setTop(RuleBank::UNIT)
                    .setBottom(RuleBank::UNIT)
                    .setLeft("gap")
                    .setRight("gap");
                add(btn);
                layout << *btn;
                _tags.append(btn);
            }
            if (!_tags.isEmpty()) {
                label().margins().setBottom(rule(RuleBank::UNIT) * 2 +
                                            style().fonts().font("small").height() + rule("gap"));
            }
            else {
                label().margins().setBottom(rule("gap"));
            }
        }
        void destroyTagButtons() {
            foreach (ButtonWidget *button, _tags) {
                GuiWidget::destroy(button);
            }
            _tags.clear();
        }
        void updateTagButtonStyle(ButtonWidget *tag, String const &color) {
            tag->setFont("small");
            tag->setTextColor(color);
            tag->set(Background(Background::Rounded, style().colors().colorf(color), 6));
        }
        PackageIconBank &iconBank() {
            return _owner.root().as<ClientRootWidget>().packageIconBank();
        }
        void fetchIcon() {
            _iconId = Id::None;
            if (_item && _item->file && iconBank().packageContainsIcon(*_item->file)) {
                _packagePath = _item->file->path();
                if (Id pkgIcon = iconBank().packageIcon(*_item->file)) {
                    setPackageIcon(pkgIcon);
                }
                else {
                    iconBank().audienceForLoad() += this;
                }
            }
        }
        void widgetCreatedForItem(GuiWidget &widget, ui::Item const &) override {
            LabelWidget &label = widget.as<LabelWidget>();
            label.setSizePolicy(ui::Expand, ui::Expand);
        }
        void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) override {
            LabelWidget &label = widget.as<LabelWidget>();
            if (ui::VariantActionItem const *varItem = maybeAs<ui::VariantActionItem>(item)) {
                label.setText(varItem->label(_actions->variantItemsEnabled()));
                label.setStyleImage(varItem->styleImageId(_actions->variantItemsEnabled()),
                                    label.fontId());
            }
            else {
                label.setText(item.label());
                if (ui::ImageItem const *imgItem = maybeAs<ui::ImageItem>(item)) {
                    if (imgItem->styleImageId().isEmpty()) {
                        label.setImage(imgItem->image());
                    }
                    else {
                        label.setStyleImage(imgItem->styleImageId(), label.fontId());
                    }
                }
            }
        }
        void updateContents() {
            bool isFile   = false;
            auto pkgIdVer = Package::split(packageId());
            if (pkgIdVer.first.startsWith("file.")) {
                isFile = true;
                if (!_iconId) {
                    icon().setStyleImage("file", "default");
                    icon().setImageScale(.5f);
                }
                if (!_item->nativePath.isEmpty()) {
                    pkgIdVer.first = _item->nativePath.pretty();
                }
            }
            else if (!_iconId) {
                icon().setStyleImage("package.icon", "default");
                icon().setImageScale(.5f);
            }
            String labelText =
                String(_E(b) "%1\n" _E(l) _E(s) "%2").arg(_item->label()).arg(pkgIdVer.first);
            if (!isFile && pkgIdVer.second.isValid()) {
                labelText += String(_E(C) " %1" _E(.)).arg(pkgIdVer.second.compactNumber());
            }
            label().setText(labelText);
            bool const highlight = _owner.d->packageStatus->isPackageHighlighted(packageId());
            _actions->setVariantItemsEnabled(highlight);
            for (GuiWidget *w : _actions->childWidgets()) {
                if (ButtonWidget *button = maybeAs<ButtonWidget>(w)) {
                    button->setImageColor(
                        style().colors().colorf(highlight ? "text" : "inverted.text"));
                    button->setColorTheme(highlight ? invertColorTheme(_owner.d->selectedItemHilit)
                                                    : invertColorTheme(_owner.d->selectedItem));
                }
            }
            String auxColor = "accent";
            if (highlight) {
                useColorTheme(_owner.d->unselectedItemHilit, _owner.d->selectedItemHilit);
                auxColor = "background";
            }
            else {
                useColorTheme(_owner.d->unselectedItem, _owner.d->selectedItem);
            }
            for (ButtonWidget *b : _tags) {
                updateTagButtonStyle(b, auxColor);
            }
            checkIconColor();
        }
        void setSelected(bool selected) override {
            HomeItemWidget::setSelected(selected);
            checkIconColor();
        }
        void checkIconColor() {
            if (_iconId) {
                icon().setImageColor(Vector4f(1, 1, 1, 1));
            }
        }
        String packageId() const { return _item->data().toString(); }
        PopupWidget *makeInfoPopup() const {
            return new PackageInfoDialog(_item->file, _owner.d->packageInfoMode);
        }
        float estimatedHeight() const override {
            float const fontHeight = style().fonts().font("default").height().value();
            float       estimate   = fontHeight * 2 + label().margins().height().value();
            return estimate;
        }
        void triggerAction(bool defaultAction = true) {
            if (!_actions->items().isEmpty()) {
                auto *w = (defaultAction ? _actions->childWidgets().last()
                                         : _actions->childWidgets().first());
                if (auto *button = maybeAs<ButtonWidget>(w)) {
                    button->trigger();
                }
            }
        }
        void itemRightClicked() override {
            if (_owner.d->rightClickToOpenContextMenu) {
                triggerAction(false);
            }
        }
        void bankLoaded(DotPath const &path) override {
            if (_iconId) return;
            DENG2_ASSERT_IN_MAIN_THREAD();
            auto &bank = iconBank();
            if (_packagePath == path) {
                setPackageIcon(bank.texture(path));
                bank.audienceForLoad() -= this;
            }
        }
        void setPackageIcon(Id const &id) {
            _iconId = id;
            if (_iconId) {
                auto *img = new AtlasProceduralImage(_owner);
                img->setPreallocatedImage(_iconId);
                icon().setImage(img);
                icon().setImageScale(1);
                checkIconColor();
            }
        }
    private:
        PackagesWidget &      _owner;
        PackageItem const *   _item;
        Path                  _packagePath;
        QList<ButtonWidget *> _tags;
        MenuWidget *          _actions = nullptr;
        Id                    _iconId;
    };
    using Strings = std::set<String>;
    LoopCallback mainCall;
    LoopCallback mainCallForIdentify;
    LineEditWidget *search;
    Rule const *    searchMinY = nullptr;
    ButtonWidget *  clearSearch;
    Animation       searchBackgroundOpacity{0.f, Animation::Linear};
    QStringList     filterTerms;
    Strings         hiddenTagsInEffect;
    QTimer          refilterTimer;
    ProgressWidget *refreshProgress;
    StringList                 hiddenTags; // Packages with these not shown unless filterTerms has the tag.
    StringList                 manualPackagePaths; // if empty, all available packages used
    HomeMenuWidget *           menu;
    ui::ListDataT<PackageItem> allPackages;
    ui::FilteredData           filteredPackages{allPackages};
    ui::ListData               defaultActionItems;
    ui::Data const *           actionItems                 = &defaultActionItems;
    bool                       populateEnabled             = true;
    bool                       showOnlyLoaded              = false;
    bool                       actionOnlyForSelection      = true;
    bool                       rightClickToOpenContextMenu = false;
    PackageInfoDialog::Mode    packageInfoMode             = PackageInfoDialog::EnableActions;
    IPackageStatus const *packageStatus = &isPackageLoaded;
    GuiWidget::ColorTheme unselectedItem      = GuiWidget::Normal;
    GuiWidget::ColorTheme selectedItem        = GuiWidget::Normal;
    GuiWidget::ColorTheme unselectedItemHilit = GuiWidget::Inverted;
    GuiWidget::ColorTheme selectedItemHilit   = GuiWidget::Inverted;
    /**
     * Initializes the PackagesWidget private implementation.
     * @param i  Public instance.
     */
    Impl(Public * i)
        : Base(i)
        , hiddenTags({"hidden", "core", "gamedata"}) {
        defaultActionItems << new ui::VariantActionItem(
            tr("Load"), tr("Unload"), new CallbackAction([this]() {
                DENG2_ASSERT(menu->interactedItem());
                String const packageId =
                    menu->interactedItem()->as<PackageItem>().data().toString();
                auto &loader = App::packageLoader();
                if (loader.isLoaded(packageId)) {
                    loader.unload(packageId);
                }
                else {
                    try {
                        loader.load(packageId);
                    }
                    catch (Error const &er) {
                        LOG_RES_ERROR("Package \"" + packageId +
                                      "\" could not be loaded: " + er.asText());
                    }
                }
                menu->interactedItem()->notifyChange();
            }));
        self().add(menu = new HomeMenuWidget);
        self().add(search = new LineEditWidget);
        self().add(clearSearch = new ButtonWidget);
        search->rule()
            .setInput(Rule::Left, self().rule().left() + self().margins().left())
            .setInput(Rule::Right, self().rule().right() - self().margins().right())
            .setInput(Rule::Top, self().rule().top() + self().margins().top());
        search->setEmptyContentHint("Enter search terms");
        search->setSignalOnEnter(true);
        search->margins().setRight(style().fonts().font("default").height() + rule("gap"));
        clearSearch->set(Background());
        clearSearch->setStyleImage("close.ring", "default");
        clearSearch->setSizePolicy(ui::Filled, ui::Filled);
        clearSearch->margins().set("");
        clearSearch->rule()
            .setInput(Rule::Right, search->rule().right() - rule("gap"))
            .setInput(Rule::Height, search->font().lineSpacing())
            .setInput(Rule::Width, search->font().lineSpacing())
            .setMidAnchorY(search->rule().midY());
        clearSearch->setActionFn([this]() {
            search->setText("");
            root().setFocus(search);
        });
        filteredPackages.setFilter([this](ui::Item const &it) {
            auto &item = it.as<PackageItem>();
            if (!item.info) {
                return false;
            }
            if (showOnlyLoaded && !item.isLoaded()) {
                return false;
            } {
                bool isHidden = false;
                for (const auto &h : hiddenTagsInEffect) {
                    if (Package::matchTags(*item.file, h)) {
                        isHidden = true;
                        break;
                    }
                }
                if (isHidden) {
                    return false;
                }
            }
            return filterTerms.isEmpty() || checkTerms({item.data().toString(),      // ID
                                                        item.file->source()->name(), // file name
                                                        item.info->gets(VAR_TITLE),
                                                        item.info->gets(VAR_TAGS)});
        });
        menu->setItems(filteredPackages);
        menu->setBehavior(ChildVisibilityClipping);
        if (!self().name().isEmpty()) {
            menu->setOpacity(0); // initially
        }
        menu->layout().setRowPadding(Const(0));
        menu->rule()
            .setInput(Rule::Left, self().rule().left() + self().margins().left())
            .setInput(Rule::Right, self().rule().right() - self().margins().right())
            .setInput(Rule::Top,
                      self().rule().top() + self().margins().top() + search->rule().height());
        menu->organizer().setWidgetFactory(*this);
        menu->setVirtualizationEnabled(
            true,
            rule("gap").valuei() * 2 + rule(RuleBank::UNIT).valuei() +
                int(style().fonts().font("default").height().value() * 3));
        menu->organizer().setRecyclingEnabled(true); // homogeneous widgets
        QObject::connect(
            search, &LineEditWidget::editorContentChanged, [this]() { updateFilterTerms(); });
        QObject::connect(
            search, &LineEditWidget::enterPressed, [this]() { focusFirstListedPackage(); });
        refilterTimer.setSingleShot(true);
        refilterTimer.setInterval(int(REFILTER_DELAY.asMilliSeconds()));
        QObject::connect(&refilterTimer, &QTimer::timeout, [this]() { updateFilterTerms(true); });
        refreshProgress = new ProgressWidget;
        refreshProgress->setMode(ProgressWidget::Indefinite);
        refreshProgress->setImageScale(.3f);
        self().add(refreshProgress);
        showProgressIndicator(true);
        self().rule().setInput(Rule::Height,
                               search->rule().height() + menu->rule().height() +
                                   self().margins().height());
    }
    ~Impl() override {
        releaseRef(searchMinY);
        menu->useDefaultItems();
    }
    void showProgressIndicator(bool show) {
        if (show) {
            refreshProgress->setOpacity(1);
            menu->hide();
            search->hide();
            clearSearch->hide();
        }
        else {
            refreshProgress->setOpacity(0, 0.3);
            menu->show();
            search->show();
            clearSearch->show();
        }
    }
    void setManualPackages(StringList ids) {
        auto &loader = PackageLoader::get();
        manualPackagePaths.clear();
        for (String id : ids) {
            if (File const *file = loader.select(id)) {
                manualPackagePaths << file->path();
            }
            else {
                throw UnavailableError("PackagesWidget::setManualPackages",
                                       "Package \"" + id + "\" not found");
            }
        }
    }
    void populate() {
        if (!populateEnabled) return;
        showProgressIndicator(false);
        StringList packages = (manualPackagePaths.isEmpty() ? App::packageLoader().findAllPackages()
                                                            : manualPackagePaths);
        for (ui::DataPos i = 0; i < allPackages.size(); ++i) {
            auto &pkgItem = allPackages.at(i);
            if (!pkgItem.info || !packages.contains(pkgItem.data().toString())) {
                allPackages.remove(i--);
            }
        }
        for (String const &path : packages) {
            const File &pack = App::rootFolder().locate<File>(path);
            const ui::DataPos pos = allPackages.findData(Package::versionedIdentifierForFile(pack));
            if (pos != ui::Data::InvalidPos) {
                allPackages.at(pos).setFile(pack);
            }
            else {
                allPackages << new PackageItem(pack);
            }
        }
        allPackages.sort();
        emit self().itemCountChanged(filteredPackages.size(), allPackages.size());
    }
    void updateItems() {
        filteredPackages.forAll([](ui::Item &item) {
            item.as<PackageItem>().notifyChange();
            return LoopContinue;
        });
    }
    void updateFilterTerms(bool immediately = false) {
        if (!immediately) {
            if (!refilterTimer.isActive()) {
                menu->setOpacity(0.f, REFILTER_DELAY);
            }
            refilterTimer.start();
        }
        else {
            mainCall.enqueue([this]() {
                setFilterTerms(
                    search->text().strip().split(QRegExp("\\s"), QString::SkipEmptyParts));
                menu->setOpacity(1.f, REFILTER_DELAY);
            });
        }
    }
    void setFilterTerms(QStringList terms) {
        filterTerms = terms;
        hiddenTagsInEffect.clear();
        for (const auto &t : hiddenTags) {
            if (!terms.contains(t)) {
                hiddenTagsInEffect.insert("\\b" + t + "\\b"); // tag regexp
            }
        }
        clearSearch->show(!terms.isEmpty());
        if ((showOnlyLoaded = filterTerms.contains(TAG_LOADED)) != false) {
            filterTerms.removeAll(TAG_LOADED);
        }
        filteredPackages.refilter();
        emit self().itemCountChanged(filteredPackages.size(), allPackages.size());
    }
    void focusFirstListedPackage() {
        if (filteredPackages.size() > 0) {
            menu->scrollY(0);
            root().setFocus(menu->childWidgets().first());
        }
    }
    void fileSystemBusyStatusChanged(FS::BusyStatus bs) override {
        if (bs == FS::Busy) {
            showProgressIndicator(true);
        }
        else if (DoomsdayApp::bundles().isEverythingIdentified()) {
            if (!mainCallForIdentify) {
                mainCallForIdentify.enqueue([this]() {
                    root().window().glActivate();
                    populateEnabled = true;
                    self().populate();
                });
            }
        }
    }
    /**
     * Checks whether the filter terms can be found in the provided text strings.
     * All terms must be found in at least one string, but all terms need not be found in
     * every provided string.
     *
     * @param texts  Text strings.
     *
     * @return @c true, if all filter terms found.
     */
    bool checkTerms(StringList texts) const {
        for (QString const &filterTerm : filterTerms) {
            bool found = false;
            for (String const &text : texts) {
                if (text.contains(filterTerm, Qt::CaseInsensitive)) {
                    found = true;
                    break;
                }
            }
            if (!found) return false;
        }
        return true;
    }
    GuiWidget *makeItemWidget(ui::Item const &item, GuiWidget const *) override {
        return new PackageListItemWidget(item.as<PackageItem>(), self());
    }
    void updateItemWidget(GuiWidget & widget, ui::Item const &item) override {
        auto &w = widget.as<PackageListItemWidget>();
        w.setItem(item.as<PackageItem>());
        DENG2_ASSERT_IN_MAIN_THREAD();
        w.updateContents();
    }
};
PackagesWidget::PackagesWidget(PopulateBehavior initBehavior, String const &name)
    : GuiWidget(name)
    , d(new Impl(this)) {
    d->populateEnabled = (initBehavior == PopulationEnabled);
    FS::get().audienceForBusy() += d;
    populate();
}
PackagesWidget::PackagesWidget(StringList manualPackageIds, String const &name)
    : GuiWidget(name)
    , d(new Impl(this)) {
    margins().set(ConstantRule::zero());
    setManualPackageIds(manualPackageIds);
}
HomeMenuWidget &PackagesWidget::menu() {
    return *d->menu;
}
ProgressWidget &PackagesWidget::progress() {
    return *d->refreshProgress;
}
void PackagesWidget::setManualPackageIds(StringList manualPackageIds) {
    d->setManualPackages(manualPackageIds);
    populate();
}
void PackagesWidget::setAllowPackageInfoActions(bool allowActions) {
    d->packageInfoMode =
        allowActions ? PackageInfoDialog::EnableActions : PackageInfoDialog::InformationOnly;
}
void PackagesWidget::setRightClickToOpenContextMenu(bool enable) {
    d->rightClickToOpenContextMenu = enable;
}
void PackagesWidget::setHiddenTags(StringList hiddenTags) {
    d->hiddenTags = hiddenTags;
    populate();
}
void PackagesWidget::setPopulationEnabled(bool enable) {
    d->populateEnabled = enable;
}
void PackagesWidget::setFilterEditorMinimumY(Rule const &minY) {
    d->search->rule().setInput(Rule::Top,
                               OperatorRule::maximum(minY, rule().top() + margins().top()));
    changeRef(d->searchMinY, minY);
}
void PackagesWidget::setPackageStatus(IPackageStatus const &packageStatus) {
    d->packageStatus = &packageStatus;
}
void PackagesWidget::showProgressIndicator() {
    d->showProgressIndicator(true);
}
void PackagesWidget::setActionItems(ui::Data const &actionItems) {
    d->actionItems = &actionItems;
}
ui::Data &PackagesWidget::actionItems() {
    return *const_cast<ui::Data *>(d->actionItems);
}
void PackagesWidget::setActionsAlwaysShown(bool showActions) {
    d->actionOnlyForSelection = !showActions;
    for (auto *w : d->menu->childWidgets()) {
        if (HomeItemWidget *item = maybeAs<HomeItemWidget>(w)) {
            item->setKeepButtonsVisible(showActions);
        }
    }
}
void PackagesWidget::setColorTheme(ColorTheme unselectedItem,
                                   ColorTheme selectedItem,
                                   ColorTheme unselectedItemHilit,
                                   ColorTheme selectedItemHilit,
                                   ColorTheme search) {
    d->unselectedItem      = unselectedItem;
    d->selectedItem        = selectedItem;
    d->unselectedItemHilit = unselectedItemHilit;
    d->selectedItemHilit   = selectedItemHilit;
    d->search->setColorTheme(search);
    d->clearSearch->setImageColor(d->search->textColorf());
    d->populate();
}
void PackagesWidget::populate() {
    d->populate();
}
void PackagesWidget::updateItems() {
    d->updateItems();
}
dsize PackagesWidget::itemCount() const {
    return d->allPackages.size();
}
ui::Item const *PackagesWidget::itemForPackage(String const &packageId) const {
    ui::DataPos found = d->filteredPackages.findData(packageId);
    if (found != ui::Data::InvalidPos) {
        return &d->filteredPackages.at(found);
    }
    return nullptr;
}
String PackagesWidget::actionPackage() const {
    if (d->menu->interactedItem()) {
        return d->menu->interactedItem()->as<Impl::PackageItem>().data().toString();
    }
    return String();
}
GuiWidget *PackagesWidget::actionWidget() const {
    if (d->menu->interactedItem()) {
        return d->menu->organizer().itemWidget(*d->menu->interactedItem());
    }
    return nullptr;
}
ui::Item const *PackagesWidget::actionItem() const {
    return d->menu->interactedItem();
}
void PackagesWidget::scrollToPackage(String const &packageId) const {
    if (auto const *item = itemForPackage(packageId)) {
        auto &scrollArea = d->menu->findTopmostScrollable();
        if (auto const *widget = d->menu->organizer().itemWidget(*item)) {
            scrollArea.scrollToWidget(*widget);
        }
        else {
            auto const pos = d->filteredPackages.find(*item);
            scrollArea.scrollY(pos * d->menu->organizer().averageChildHeight() -
                                   scrollArea.rule().height().value() / 2,
                               0.3);
        }
    }
}
LineEditWidget &PackagesWidget::searchTermsEditor() {
    return *d->search;
}
void PackagesWidget::initialize() {
    GuiWidget::initialize();
    d->menu->organizer().setVisibleArea(root().viewTop(), root().viewBottom());
}
void PackagesWidget::update() {
    GuiWidget::update();
    if (d->searchMinY) {
        TimeSpan const SPAN = 0.3;
        if (d->searchBackgroundOpacity.target() < .5f &&
            d->search->rule().top().valuei() == d->searchMinY->valuei()) {
            d->searchBackgroundOpacity.setValue(1.f, SPAN);
        }
        else if (d->searchBackgroundOpacity.target() > .5f &&
                 d->search->rule().top().valuei() > d->searchMinY->valuei()) {
            d->searchBackgroundOpacity.setValue(0.f, SPAN);
        }
        d->search->setUnfocusedBackgroundOpacity(d->searchBackgroundOpacity);
    }
}
void PackagesWidget::operator>>(PersistentState &toState) const {
    if (name().isEmpty()) return;
    Record &rec = toState.objectNamespace();
    rec.set(name().concatenateMember("search"), d->search->text());
}
void PackagesWidget::operator<<(PersistentState const &fromState) {
    if (name().isEmpty()) return;
    Record const &rec = fromState.objectNamespace();
    d->search->setText(rec.gets(name().concatenateMember("search"), ""));
    d->updateFilterTerms(true);
}
DENG_GUI_PIMPL(PanelButtonWidget) {
    PanelWidget *drawer;
    Impl(Public *i) : Base(i) {
        self().add(drawer = new PanelWidget);
        drawer->set(Background(Vector4f(0, 0, 0, .15f)));
        drawer->setEatMouseEvents(false);
    }
};
PanelButtonWidget::PanelButtonWidget(Flags flags, String const &name)
    : HomeItemWidget(flags, name)
    , d(new Impl(this)) {
    setBehavior(Focusable);
    d->drawer->rule()
            .setInput(Rule::Top,  icon().rule().bottom())
            .setInput(Rule::Left, rule().left());
    if (flags & AnimatedHeight) {
        rule().setInput(Rule::Height, new AnimationRule(label().rule().height() +
                                                        d->drawer->rule().height(), 0.3));
    }
    else {
        rule().setInput(Rule::Height, label().rule().height() + d->drawer->rule().height());
    }
}
PanelWidget &PanelButtonWidget::panel() {
    return *d->drawer;
}
void PanelButtonWidget::setSelected(bool selected) {
    HomeItemWidget::setSelected(selected);
    if (selected) {
        d->drawer->set(Background(Vector4f(0, 0, 0, .4f)));
    }
    else {
        d->drawer->set(Background(Vector4f(0, 0, 0, .15f)));
    }
}
DENG_GUI_PIMPL(PrivilegedLogWidget) {
    StyledLogSinkFormatter formatter;
    LogWidget *log = nullptr;
    ButtonWidget *closeButton = nullptr;
    Impl(Public *i) : Base(i) {
        log = new LogWidget("privlog");
        log->setLogFormatter(formatter);
        log->setPrivilegedEntries(true);
        self().add(log);
        closeButton = new ButtonWidget;
        closeButton->setSizePolicy(ui::Expand, ui::Expand);
        self().add(closeButton);
        updateStyle();
    }
    void updateStyle() {
        closeButton->setImageScale(.25f);
        closeButton->setImage(style().images().image("close.ringless"));
        closeButton->setImageColor(style().colors().colorf("altaccent"));
    }
};
PrivilegedLogWidget::PrivilegedLogWidget()
    : d(new Impl(this)) {
    auto const &gap = rule("gap");
    hide();
    connect(d->log, SIGNAL(contentHeightIncreased(int)), this, SLOT(showLog()));
    connect(d->closeButton, SIGNAL(pressed()), this, SLOT(hideLog()));
    d->log->rule()
            .setInput(Rule::Left, rule().left() + gap)
            .setInput(Rule::Top,  rule().top() + gap)
            .setInput(Rule::Right, (rule().left() + rule().midX()) / 2)
            .setInput(Rule::Bottom, rule().midY() - gap);
    d->closeButton->rule()
            .setInput(Rule::Top, d->log->rule().top())
            .setInput(Rule::Right, d->log->rule().right());
}
void PrivilegedLogWidget::showLog() {
    show();
}
void PrivilegedLogWidget::hideLog() {
    hide();
    d->log->clear();
}
void PrivilegedLogWidget::updateStyle() {
    GuiWidget::updateStyle();
    d->updateStyle();
}
static int const MAX_VISIBLE_PROFILE_NAME = 50;
static int const MAX_PROFILE_NAME = 100;
DENG_GUI_PIMPL(ProfilePickerWidget) {
    ConfigProfiles &settings;
    String description;
    PopupButtonWidget *button;
    bool invertedPopups = false;
    Impl(Public *i, ConfigProfiles& reg)
        : Base(i)
        , settings(reg)
        , button(0) {
        self().add(button = new PopupButtonWidget);
        button->setOpener([this] (PopupWidget *) {
            self().openMenu();
        });
        updateStyle();
    }
    void updateStyle() {
        button->setSizePolicy(ui::Expand, ui::Expand);
        button->setImage(style().images().image("gear"));
        button->setOverrideImageSize(style().fonts().font("default").height());
    }
    void populate() {
        self().items().clear();
        foreach (String prof, settings.profiles()) {
            self().items() << new ChoiceItem(prof.left(MAX_VISIBLE_PROFILE_NAME), prof);
        }
        self().items().sort();
        self().setSelected(self().items().findData(settings.currentProfile()));
    }
    String currentProfile() const {
        return self().selectedItem().data().toString();
    }
};
ProfilePickerWidget::ProfilePickerWidget(ConfigProfiles &settings, String const &description, String const &name)
    : ChoiceWidget(name), d(new Impl(this, settings)) {
    d->description = description;
    d->populate();
    d->button->margins().setAll(margins());
    d->button->rule()
            .setInput(Rule::Left, rule().right())
            .setInput(Rule::Top,  rule().top());
    connect(this, SIGNAL(selectionChangedByUser(uint)), this, SLOT(applySelectedProfile()));
}
ButtonWidget &ProfilePickerWidget::button() {
    return *d->button;
}
void ProfilePickerWidget::useInvertedStyleForPopups() {
    d->invertedPopups = true;
    popup().useInfoStyle();
}
void ProfilePickerWidget::updateStyle() {
    ChoiceWidget::updateStyle();
    d->updateStyle();
}
void ProfilePickerWidget::openMenu() {
    ConfigProfiles &reg = d->settings;
    auto *menu = new PopupMenuWidget;
    menu->setAllowDirectionFlip(false);
    menu->setDeleteAfterDismissed(true);
    if (d->invertedPopups) menu->useInfoStyle();
    d->button->setPopup(*menu, ui::Down);
    menu->items()
            << new ActionItem(tr("Edit"), new SignalAction(this, SLOT(edit())))
            << new ActionItem(tr("Rename..."), new SignalAction(this, SLOT(rename())))
            << new ui::Item(Item::Separator)
            << new ActionItem(tr("Duplicate..."), new SignalAction(this, SLOT(duplicate())))
            << new ui::Item(Item::Separator)
            << new ActionItem(tr("Reset to Defaults..."), new SignalAction(this, SLOT(reset())))
            << new ActionItem(style().images().image("close.ring"), tr("Delete..."),
                              new SignalAction(this, SLOT(remove())));
    add(menu);
    ChildWidgetOrganizer const &org = menu->menu().organizer();
    String selProf = selectedItem().data().toString();
    if (reg.find(selProf).isReadOnly()) {
        menu->items().at(0).setLabel("View");
        org.itemWidget(1)->disable();
        org.itemWidget(5)->disable();
        org.itemWidget(6)->disable();
    }
    if (reg.count() == 1) {
        org.itemWidget(6)->disable();
    }
    if (root().window().as<ClientWindow>().hasSidebar()) {
        org.itemWidget(0)->disable();
    }
    menu->open();
}
void ProfilePickerWidget::edit() {
    emit profileEditorRequested();
}
void ProfilePickerWidget::rename() {
    InputDialog *dlg = new InputDialog;
    dlg->setDeleteAfterDismissed(true);
    dlg->title().setText(tr("Renaming \"%1\"").arg(d->currentProfile()));
    dlg->message().setText(tr("Enter a new name for the %1 profile:").arg(d->description));
    dlg->defaultActionItem()->setLabel(tr("Rename Profile"));
    dlg->editor().setText(d->currentProfile());
    if (dlg->exec(root())) {
        String newName = dlg->editor().text().trimmed().left(MAX_PROFILE_NAME);
        if (!newName.isEmpty() && d->currentProfile() != newName) {
            if (d->settings.rename(newName)) {
                ui::Item &item = items().at(selected());
                item.setLabel(newName.left(MAX_VISIBLE_PROFILE_NAME));
                item.setData(newName);
                items().sort();
                setSelected(items().findData(newName));
            }
            else {
                LOG_WARNING("Failed to rename profile to \"%s\"") << newName;
            }
        }
    }
}
void ProfilePickerWidget::duplicate() {
    InputDialog *dlg = new InputDialog;
    dlg->setDeleteAfterDismissed(true);
    dlg->title().setText(tr("Duplicating \"%1\"").arg(d->currentProfile()));
    dlg->message().setText(tr("Enter a name for the new %1 profile:").arg(d->description));
    dlg->defaultActionItem()->setLabel(tr("Duplicate Profile"));
    if (dlg->exec(root())) {
        String newName = dlg->editor().text().trimmed().left(MAX_PROFILE_NAME);
        if (!newName.isEmpty()) {
            if (d->settings.saveAsProfile(newName)) {
                d->settings.setProfile(newName);
                items().append(new ChoiceItem(newName.left(MAX_VISIBLE_PROFILE_NAME), newName)).sort();
                setSelected(items().findData(newName));
                applySelectedProfile();
            }
            else {
                LOG_WARNING("Failed to duplicate current profile to create \"%s\"")
                        << newName;
            }
        }
    }
}
void ProfilePickerWidget::reset() {
    MessageDialog *dlg = new MessageDialog;
    dlg->setDeleteAfterDismissed(true);
    dlg->title().setText(tr("Reset?"));
    dlg->message().setText(tr("Are you sure you want to reset the %1 profile %2 to the default values?")
                           .arg(d->description)
                           .arg(_E(b) + d->currentProfile() + _E(.)));
    dlg->buttons()
            << new DialogButtonItem(DialogWidget::Default | DialogWidget::Reject)
            << new DialogButtonItem(DialogWidget::Accept, tr("Reset Profile"));
    if (dlg->exec(root())) {
        d->settings.resetToDefaults();
    }
}
void ProfilePickerWidget::remove() {
    MessageDialog *dlg = new MessageDialog;
    dlg->setDeleteAfterDismissed(true);
    dlg->title().setText(tr("Delete?"));
    dlg->message().setText(
                tr("Are you sure you want to delete the %1 profile %2? This cannot be undone.")
                .arg(d->description)
                .arg(_E(b) + d->currentProfile() + _E(.)));
    dlg->buttons()
               << new DialogButtonItem(DialogWidget::Default | DialogWidget::Reject)
               << new DialogButtonItem(DialogWidget::Accept, tr("Delete Profile"));
    if (!dlg->exec(root())) return;
    String const profToDelete = d->currentProfile();
    items().remove(selected());
    applySelectedProfile();
    d->settings.deleteProfile(profToDelete);
}
void ProfilePickerWidget::applySelectedProfile() {
    d->settings.setProfile(d->currentProfile());
}
DENG_GUI_PIMPL(SidebarWidget)
, DENG2_OBSERVES(DoomsdayApp, GameChange) {
    DialogContentStylist stylist;
    GuiWidget *container;
    ScrollAreaWidget *sidebarContent;
    IndirectRule *firstColumnWidth; ///< Shared by all groups.
    LabelWidget *title;
    ButtonWidget *close;
    std::unique_ptr<SequentialLayout> layout;
    Impl(Public *i)
        : Base(i)
        , firstColumnWidth(new IndirectRule) {
        DoomsdayApp::app().audienceForGameChange() += this;
        container = new GuiWidget;
        sidebarContent = new ScrollAreaWidget;
        sidebarContent->enableIndicatorDraw(true);
        stylist.setContainer(*sidebarContent);
        container->add(sidebarContent);
        container->add(title = new LabelWidget);
        title->margins().set(rule("dialog.gap"));
        title->margins().setLeft("gap");
        title->setFont("title");
        title->setTextColor("accent");
        title->setSizePolicy(ui::Expand, ui::Expand);
        container->add(close = new ButtonWidget);
        close->margins().set(rule("dialog.gap"));
        close->setImage(style().images().image("close.ringless"));
        close->setImageColor(title->textColorf());
        close->setOverrideImageSize(title->font().height());
        close->setAction(new SignalAction(thisPublic, SLOT(close())));
        close->setSizePolicy(ui::Expand, ui::Expand);
    }
    ~Impl() {
        releaseRef(firstColumnWidth);
    }
    void currentGameChanged(Game const &newGame) {
        if (newGame.isNull()) {
            self().close();
        }
    }
};
SidebarWidget::SidebarWidget(String const &titleText, String const &name)
    : PanelWidget(name)
    , d(new Impl(this)) {
    setSizePolicy(Fixed);
    setWaitForContentReady(false);
    setOpeningDirection(Left);
    setAnimationStyle(Smooth);
    set(Background(style().colors().colorf("background")).withSolidFillOpacity(1));
    d->title->setText(titleText);
    d->title->rule()
            .setInput(Rule::Top,  d->container->rule().top())
            .setInput(Rule::Left, d->container->rule().left());
    d->close->rule()
            .setInput(Rule::Right,  d->container->rule().right())
            .setInput(Rule::Bottom, d->title->rule().bottom());
    d->sidebarContent->rule()
        .setInput(Rule::Left,   d->container->rule().left())
        .setInput(Rule::Width,  d->sidebarContent->contentRule().width() +
                                d->sidebarContent->margins().width())
        .setInput(Rule::Top,    d->title->rule().bottom())
        .setInput(Rule::Bottom, rule().bottom());
    d->container->rule().setSize(d->sidebarContent->rule().width(),
                                 rule().height());
    setContent(d->container);
    RuleRectangle const &area = d->sidebarContent->contentRule();
    d->layout.reset(new SequentialLayout(area.left(), area.top(), Down));
    ClientWindow::main().setSidebar(ClientWindow::RightEdge, this);
}
SequentialLayout &SidebarWidget::layout() {
    return *d->layout;
}
LabelWidget &SidebarWidget::title() {
    return *d->title;
}
Rule const &SidebarWidget::maximumOfAllGroupFirstColumns() const {
    Rule const *max = nullptr;
    foreach (GuiWidget *child, d->sidebarContent->childWidgets()) {
        if (auto *g = maybeAs<VariableGroupEditor>(child)) {
            changeRef(max, OperatorRule::maximum(g->firstColumnWidth(), max));
        }
    }
    if (!max) {
        return ConstantRule::zero();
    }
    return *refless(max);
}
IndirectRule &SidebarWidget::firstColumnWidth() {
    return *d->firstColumnWidth;
}
ScrollAreaWidget &SidebarWidget::containerWidget() {
    return *d->sidebarContent;
}
ButtonWidget &SidebarWidget::closeButton() {
    return *d->close;
}
void SidebarWidget::preparePanelForOpening() {
    PanelWidget::preparePanelForOpening();
}
void SidebarWidget::panelDismissed() {
    PanelWidget::panelDismissed();
    ClientWindow::main().unsetSidebar(ClientWindow::RightEdge);
}
void SidebarWidget::updateSidebarLayout(Rule const &minWidth,
                                        Rule const &extraHeight) {
    d->firstColumnWidth->setSource(maximumOfAllGroupFirstColumns());
    d->sidebarContent->setContentSize(
            OperatorRule::maximum(minWidth, d->layout->width(), rule("sidebar.width")),
            d->layout->height() + extraHeight);
}
static TimeSpan OPEN_CLOSE_SPAN = 0.2;
enum MenuItemPositions {
    POS_HOME              = 1,
    POS_CONNECT           = 2,
    POS_GAMES_SEPARATOR   = 3,
    POS_UNLOAD            = 4,
    POS_PACKAGES          = 7,
    POS_PACKAGES_NOTE     = 8,
    POS_RENDERER_SETTINGS = 0,
    POS_3D_VR_SETTINGS    = 1,
    POS_CONFIG_SEPARATOR  = 2,
    POS_AUDIO_SETTINGS    = 4,
    POS_INPUT_SETTINGS    = 5,
    POS_DATA_FILES        = 8,
    POS_UI_SETTINGS       = 9,
};
DENG_GUI_PIMPL(TaskBarWidget)
, DENG2_OBSERVES(DoomsdayApp, GameChange)
, DENG2_OBSERVES(ServerLink, Join)
, DENG2_OBSERVES(ServerLink, Leave)
, DENG2_OBSERVES(PanelWidget, AboutToOpen) // update menu
{
    typedef DefaultVertexBuf VertexBuf;
    enum LayoutMode {
        NormalLayout,           ///< Taskbar widgets are full-sized.
        CompressedLayout,       ///< Cull some redundant information.
        ExtraCompressedLayout   ///< Hide current game indicator.
    };
    LayoutMode layoutMode;
    bool opened;
    GuiWidget *backBlur;
    ConsoleWidget *console;
    PopupButtonWidget *logo;
    PopupButtonWidget *conf;
    PopupButtonWidget *multi;
    LabelWidget *status;
    PopupMenuWidget *mainMenu;
    PopupMenuWidget *configMenu;
    MultiplayerStatusWidget *multiMenu;
    AnimationRule *vertShift;
    bool mouseWasTrappedWhenOpening;
    int minSpace;
    int maxSpace;
    Impl(Public *i)
        : Base(i)
        , layoutMode(NormalLayout)
        , opened(true)
        , logo(0)
        , conf(0)
        , status(0)
        , mainMenu(0)
        , configMenu(0)
        , multiMenu(0)
        , mouseWasTrappedWhenOpening(false) {
        self().set(Background(style().colors().colorf("background")));
        vertShift = new AnimationRule(0);
        DoomsdayApp::app().audienceForGameChange() += this;
        ClientApp::serverLink().audienceForJoin()  += this;
        ClientApp::serverLink().audienceForLeave() += this;
        updateStyle();
    }
    ~Impl() {
        releaseRef(vertShift);
    }
    void updateStyle() {
        minSpace = rule("console.commandline.width.min").valuei();
        maxSpace = rule("console.commandline.width.max").valuei();
    }
    void updateLayoutMode() {
        LayoutMode wanted = layoutMode;
        if (console->commandLine().rule().width().valuei() < minSpace) {
            wanted = (layoutMode == NormalLayout?     CompressedLayout :
                      layoutMode == CompressedLayout? ExtraCompressedLayout :
                                                      layoutMode);
        }
        else if (console->commandLine().rule().width().valuei() > maxSpace) {
            wanted = (layoutMode == CompressedLayout?      NormalLayout :
                      layoutMode == ExtraCompressedLayout? CompressedLayout :
                                                           layoutMode);
        }
        if (layoutMode != wanted) {
            layoutMode = wanted;
            updateLogoButtonText();
            switch (layoutMode) {
            case NormalLayout:
            case CompressedLayout:
                status->show();
                break;
            case ExtraCompressedLayout:
                status->hide();
                break;
            }
            self().updateCommandLineLayout();
            self().requestGeometry();
            console->commandLine().requestGeometry();
        }
    }
    void glInit() {
        /*shaders().build(drawable.program(), "generic.color_ucolor")
                << uMvpMatrix
                << uColor;
        updateProjection();*/
    }
    void glDeinit() {
    }
    void updateLogoButtonText() {
        String text;
        if (layoutMode == NormalLayout) {
            Version const currentVersion = Version::currentBuild();
            if (String(DOOMSDAY_RELEASE_TYPE) == "Stable") {
                text = _E(b) + currentVersion.compactNumber();
            }
            else {
                text = _E(b) + currentVersion.compactNumber() + " " +
                       _E(l) + String("#%1").arg(currentVersion.build);
            }
        }
        else {
        }
        logo->setText(text);
    }
    /*void updateProjection()
    {
        uMvpMatrix = root().projMatrix2D();
    }*/
    void updateGeometry() {
        /*
        Rectanglei pos;
        if (self().hasChangedPlace(pos) || self().geometryRequested()) {
            self().requestGeometry(false);
            verts.clear();
            self().glMakeGeometry(verts);
        }*/
    }
    GuiWidget &itemWidget(PopupMenuWidget *menu, uint pos) const {
        return *menu->menu().organizer().itemWidget(pos);
    }
    void showOrHideMenuItems() {
        Game const &game = App_CurrentGame();
        itemWidget(mainMenu, POS_UNLOAD)           .show(!game.isNull());
        itemWidget(mainMenu, POS_GAMES_SEPARATOR)  .show(!game.isNull());
        itemWidget(mainMenu, POS_PACKAGES)         .show(!game.isNull());
        itemWidget(mainMenu, POS_PACKAGES_NOTE)    .show(!game.isNull());
        itemWidget(mainMenu, POS_HOME)             .show(!game.isNull());
        itemWidget(configMenu, POS_RENDERER_SETTINGS).show(!game.isNull());
        itemWidget(configMenu, POS_3D_VR_SETTINGS)   .show(!game.isNull());
        itemWidget(configMenu, POS_CONFIG_SEPARATOR) .show(!game.isNull());
        itemWidget(configMenu, POS_INPUT_SETTINGS)   .show(!game.isNull());
        itemWidget(configMenu, POS_DATA_FILES)       .show(game.isNull());
        if (self().hasRoot()) {
            configMenu->menu().updateLayout();
            mainMenu->menu().updateLayout(); // Include/exclude shown/hidden menu items.
        }
    }
    void currentGameChanged(Game const &) {
        updateStatus();
        showOrHideMenuItems();
        backBlur->show(style().isBlurringAllowed());
    }
    void networkGameJoined() {
        multi->show();
        self().updateCommandLineLayout();
    }
    void networkGameLeft() {
        multi->hide();
        self().updateCommandLineLayout();
    }
    void updateStatus() {
        if (auto const *prof = DoomsdayApp::currentGameProfile()) {
            status->setText(prof->name().truncateWithEllipsis(30));
        }
        else if (App_GameLoaded()) {
            status->setText(App_CurrentGame().id());
        }
        else {
            status->setText(tr("No game loaded"));
        }
    }
    void panelAboutToOpen(PanelWidget &) {
        if (self().root().window().as<ClientWindow>().isGameMinimized()) {
            mainMenu->items().at(POS_HOME).setLabel(tr("Hide Home"));
        }
        else {
            mainMenu->items().at(POS_HOME).setLabel(tr("Show Home"));
        }
    }
};
#if defined (DENG_HAVE_UPDATER)
static PopupWidget *makeUpdaterSettings() {
    return new UpdaterSettingsDialog(UpdaterSettingsDialog::WithApplyAndCheckButton);
}
#endif
TaskBarWidget::TaskBarWidget() : GuiWidget("taskbar"), d(new Impl(this)) {
    Background bg(style().colors().colorf("background"));
    Rule const &gap = rule("gap");
    d->backBlur = new LabelWidget;
    d->backBlur->rule()
            .setInput(Rule::Left,   rule().left())
            .setInput(Rule::Bottom, rule().bottom())
            .setInput(Rule::Right,  rule().right())
            .setInput(Rule::Top,    rule().top());
    if (style().isBlurringAllowed()) {
        d->backBlur->set(Background(ClientWindow::main().taskBarBlur(), Vector4f(1, 1, 1, 1)));
    }
    else {
        d->backBlur->set(Background(Vector4f(0, 0, 0, 1)));
    }
    add(d->backBlur);
    d->console = new ConsoleWidget;
    d->console->rule()
            .setInput(Rule::Left, rule().left() + d->console->shift());
    add(d->console);
    d->console->buttons().rule()
            .setInput(Rule::Left,   rule().left())
            .setInput(Rule::Bottom, rule().bottom())
            .setInput(Rule::Height, rule().height());
    d->logo = new PopupButtonWidget("de-button");
    d->logo->setImage(style().images().image("logo.px128"));
    d->logo->setImageScale(.475f);
    d->logo->setImageFit(FitToHeight | OriginalAspectRatio);
    d->updateLogoButtonText();
    d->logo->setWidthPolicy(ui::Expand);
    d->logo->setTextAlignment(AlignLeft);
    d->logo->rule().setInput(Rule::Height, rule().height());
    add(d->logo);
    d->conf = new PopupButtonWidget("conf-button");
    d->conf->setImage(style().images().image("gear"));
    d->conf->setSizePolicy(ui::Expand, ui::Filled);
    d->conf->rule().setInput(Rule::Height, rule().height());
    add(d->conf);
    d->status = new LabelWidget;
    d->status->set(bg);
    d->status->setWidthPolicy(ui::Expand);
    d->status->rule().setInput(Rule::Height, rule().height());
    add(d->status);
    d->updateStatus();
    d->multi = new PopupButtonWidget;
    d->multi->hide(); // hidden when not connected
    d->multi->setImage(style().images().image("network"));
    d->multi->setTextAlignment(ui::AlignRight);
    d->multi->setText(tr("MP"));
    d->multi->setSizePolicy(ui::Expand, ui::Filled);
    d->multi->rule().setInput(Rule::Height, rule().height());
    add(d->multi);
    rule().setInput(Rule::Height, style().fonts().font("default").height() + gap * 2);
    add(d->configMenu = new PopupMenuWidget("conf-menu"));
    d->conf->setPopup(*d->configMenu);
    add(d->multiMenu = new MultiplayerStatusWidget);
    d->multi->setPopup(*d->multiMenu);
    add(d->mainMenu = new PopupMenuWidget("de-menu"));
    d->mainMenu->audienceForAboutToOpen() += d;
    d->logo->setPopup(*d->mainMenu);
    auto *unloadMenu = new ui::SubmenuItem(style().images().image("close.ring"),
                                           tr("Unload Game"), ui::Left);
    unloadMenu->items()
            << new ui::Item(ui::Item::Separator, tr("Really unload the game?"))
            << new ui::ActionItem(tr("Unload") + " " _E(b) + tr("(discard progress)"), new SignalAction(this, SLOT(unloadGame())))
            << new ui::ActionItem(tr("Cancel"), new SignalAction(&d->mainMenu->menu(), SLOT(dismissPopups())));
    /*
     * Set up items for the config and DE menus. Some of these are shown/hidden
     * depending on whether a game is loaded.
     */
    d->configMenu->items()
            << new ui::SubwidgetItem(style().images().image("renderer"),  tr("Renderer"),       ui::Left, makePopup<RendererSettingsDialog>)
            << new ui::SubwidgetItem(style().images().image("vr"),        tr("3D & VR"),        ui::Left, makePopup<VRSettingsDialog>)
            << new ui::Item(ui::Item::Separator)
            << new ui::SubwidgetItem(style().images().image("display"),   tr("Video"),          ui::Left, makePopup<VideoSettingsDialog>)
            << new ui::SubwidgetItem(style().images().image("audio"),     tr("Audio"),          ui::Left, makePopup<AudioSettingsDialog>)
            << new ui::SubwidgetItem(style().images().image("input"),     tr("Input"),          ui::Left, makePopup<InputSettingsDialog>)
            << new ui::SubwidgetItem(style().images().image("network"),   tr("Network"),        ui::Left, makePopup<NetworkSettingsDialog>)
            << new ui::Item(ui::Item::Separator)
            << new ui::SubwidgetItem(style().images().image("package.icon"), tr("Data Files"),     ui::Left, makePopup<DataFileSettingsDialog>)
            << new ui::SubwidgetItem(style().images().image("home.icon"), tr("User Interface"), ui::Left, makePopup<UISettingsDialog>);
#if defined (DENG_HAVE_UPDATER)
    d->configMenu->items()
            << new ui::SubwidgetItem(style().images().image("updater"),   tr("Updater"),        ui::Left, makeUpdaterSettings);
#endif
    auto *helpMenu = new ui::SubmenuItem(tr("Help"), ui::Left);
    helpMenu->items()
            << new ui::ActionItem(tr("Doomsday Manual..."),
                              new CallbackAction([]() {
                                  ClientApp::app().openInBrowser(QUrl("https://manual.dengine.net/"));
                              }))
            << new ui::ActionItem(tr("Show Tutorial"), new SignalAction(this, SLOT(showTutorial())));
    d->mainMenu->items()
            << new ui::Item(ui::Item::Separator, tr("Games"))
            << new ui::ActionItem(style().images().image("home.icon"), "",
                                  new SignalAction(this, SLOT(showOrHideHome())))
            << new ui::ActionItem(tr("Connect to Server..."), new SignalAction(this, SLOT(connectToServerManually())))
            << new ui::Item(ui::Item::Separator)
            << unloadMenu                           // hidden with null-game
            << new ui::Item(ui::Item::Separator)
            << new ui::Item(ui::Item::Separator, tr("Resources"))
            << new ui::ActionItem(tr("Browse Mods..."), new SignalAction(this, SLOT(openPackagesSidebar())))
            << new ui::Item(ui::Item::Annotation,
                            tr("Load/unload data files and view package information."))
            << new ui::ActionItem(tr("Clear Cache"), new CallbackAction([] () { DoomsdayApp::app().clearCache(); }))
            << new ui::Item(ui::Item::Annotation,
                            tr("Forces a refresh of resource file metadata."))
            << new ui::Item(ui::Item::Separator)
            << new ui::Item(ui::Item::Separator, tr("Doomsday"))
#if defined (DENG_HAVE_UPDATER)
            << new ui::ActionItem(tr("Check for Updates"), new CommandAction("updateandnotify"))
#endif
            << new ui::ActionItem(tr("About Doomsday"), new SignalAction(this, SLOT(showAbout())))
            << helpMenu
#if !defined (DENG_MOBILE)
            << new ui::Item(ui::Item::Separator)
            << new ui::ActionItem(tr("Quit Doomsday"), new CommandAction("quit!"))
#endif
            ;
    d->showOrHideMenuItems();
    updateCommandLineLayout();
    connect(d->console, SIGNAL(commandModeChanged()), this, SLOT(updateCommandLineLayout()));
    connect(d->console, SIGNAL(commandLineGotFocus()), this, SLOT(closeMainMenu()));
    connect(d->console, SIGNAL(commandLineGotFocus()), this, SLOT(closeConfigMenu()));
    close();
}
ConsoleWidget &TaskBarWidget::console() {
    return *d->console;
}
CommandWidget &TaskBarWidget::commandLine() {
    return d->console->commandLine();
}
ButtonWidget &TaskBarWidget::logoButton() {
    return *d->logo;
}
bool TaskBarWidget::isOpen() const {
    return d->opened;
}
Rule const &TaskBarWidget::shift() {
    return *d->vertShift;
}
void TaskBarWidget::glInit() {
    LOG_AS("TaskBarWidget");
    d->glInit();
}
void TaskBarWidget::glDeinit() {
    d->glDeinit();
}
/*void TaskBarWidget::viewResized()
{
    GuiWidget::viewResized();
}*/
void TaskBarWidget::update() {
    GuiWidget::update();
    d->updateLayoutMode();
}
void TaskBarWidget::drawContent() {
    d->updateGeometry();
}
bool TaskBarWidget::handleEvent(Event const &event) {
    ClientWindow &window = root().window().as<ClientWindow>();
    if (!window.eventHandler().isMouseTrapped()
            && event.type() == Event::MouseButton
            && !window.hasSidebar()
            && !window.isGameMinimized()) {
        MouseEvent const &mouse = event.as<MouseEvent>();
        if (mouse.state() == MouseEvent::Released && !hitTest(mouse.pos())) {
            /*if (root().focus())
            {
                root().setFocus(0);
                return true;
            }*/
            if (App_GameLoaded()) {
                window.eventHandler().trapMouse();
            }
            window.taskBar().close();
            return true;
        }
    }
    if (event.type() == Event::MouseButton) {
        if (hitTest(event)) {
            return true;
        }
    }
    if (isOpen() && event.isKey() && event.as<KeyEvent>().isModifier()) {
        ClientApp::inputSystem().trackEvent(event);
        return true;
    }
    if (event.type() == Event::KeyPress) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_ESCAPE) {
            if (isOpen()) {
                if (d->console->isLogOpen() && !key.modifiers().testFlag(KeyEvent::Shift)) {
                    d->console->commandLine().setText("");
                    d->console->closeLog();
                    root().setFocus(0);
                    return true;
                }
                close();
                return true;
            }
            else {
                if (key.modifiers().testFlag(KeyEvent::Shift) ||
                   !App_GameLoaded()) {
                    if (!window.hasSidebar() && !window.isGameMinimized()) {
                        root().setFocus(&d->console->commandLine());
                    }
                    open();
                    return true;
                }
            }
            return false;
        }
    }
    return false;
}
void TaskBarWidget::open() {
    if (!d->opened) {
        d->opened = true;
        unsetBehavior(DisableEventDispatchToChildren);
        d->console->zeroLogHeight();
        d->vertShift->set(0, OPEN_CLOSE_SPAN);
        setOpacity(1, OPEN_CLOSE_SPAN);
        emit opened();
    }
    if (hasRoot()) {
        auto &handler = root().window().eventHandler();
        d->mouseWasTrappedWhenOpening = handler.isMouseTrapped();
        if (handler.isMouseTrapped()) {
            handler.trapMouse(false);
        }
        if (!App_GameLoaded()) {
            root().setFocus(&d->console->commandLine());
        }
    }
}
void TaskBarWidget::openAndPauseGame() {
    root().window().as<ClientWindow>().game().pause();
    open();
}
void TaskBarWidget::close() {
    if (d->opened) {
        d->opened = false;
        setBehavior(DisableEventDispatchToChildren);
        d->vertShift->set(rule().height().valuei() +
                          rule(RuleBank::UNIT).valuei(), OPEN_CLOSE_SPAN);
        setOpacity(0, OPEN_CLOSE_SPAN);
        d->console->closeLog();
        d->console->closeMenu();
        d->console->commandLine().dismissContentToHistory();
        closeMainMenu();
        closeConfigMenu();
        if (hasRoot()) root().setFocus(0);
        emit closed();
        if (hasRoot()) {
            auto &window = root().window().as<ClientWindow>();
            if (App_GameLoaded() && !window.hasSidebar() && !window.isGameMinimized()) {
                if (d->mouseWasTrappedWhenOpening) {
                    window.eventHandler().trapMouse();
                }
            }
        }
    }
}
void TaskBarWidget::openConfigMenu() {
    d->mainMenu->close(0);
    d->configMenu->open();
}
void TaskBarWidget::closeConfigMenu() {
    d->configMenu->close();
}
void TaskBarWidget::openMainMenu() {
    d->configMenu->close(0);
    d->mainMenu->open();
}
void TaskBarWidget::closeMainMenu() {
    d->mainMenu->close();
}
void TaskBarWidget::openMultiplayerMenu() {
    d->multiMenu->open();
}
void TaskBarWidget::unloadGame() {
    Con_Execute(CMDS_DDAY, "unload", false, false);
    d->mainMenu->close();
}
void TaskBarWidget::showAbout() {
    root().window().glActivate();
    AboutDialog *about = new AboutDialog;
    about->setDeleteAfterDismissed(true);
    root().addOnTop(about);
    about->open();
    root().window().glDone();
}
#if defined (DENG_HAVE_UPDATER)
void TaskBarWidget::showUpdaterSettings() {
    UpdaterSettingsDialog *dlg = new UpdaterSettingsDialog(UpdaterSettingsDialog::WithApplyAndCheckButton);
    dlg->setDeleteAfterDismissed(true);
    root().addOnTop(dlg);
    dlg->open();
}
#endif
void TaskBarWidget::showOrHideHome() {
    DENG2_ASSERT(App_GameLoaded());
    auto &win = ClientWindow::main();
    if (!win.isGameMinimized()) {
        win.game().pause();
        win.setGameMinimized(true);
        win.home().moveOnscreen(1.0);
        close();
    }
    else {
        win.home().moveOffscreen(1.0);
    }
}
void TaskBarWidget::connectToServerManually() {
    auto *dlg = new ManualConnectionDialog;
    dlg->setDeleteAfterDismissed(true);
    dlg->exec(root());
}
void TaskBarWidget::showTutorial() {
    if (BusyMode_Active()) return;
    d->mainMenu->close();
    TutorialWidget *tutorial = new TutorialWidget;
    root().addOnTop(tutorial);
    tutorial->rule().setRect(root().viewRule());
    tutorial->start();
}
void TaskBarWidget::openPackagesSidebar() {
    auto *sidebar = new PackagesSidebarWidget; // ClientWindow gets ownership
    sidebar->open();
}
void TaskBarWidget::updateCommandLineLayout() {
    SequentialLayout layout(rule().right(), rule().top(), ui::Left);
    layout << *d->logo << *d->conf;
    if (!d->multi->behavior().testFlag(Hidden)) {
        layout << *d->multi;
    }
    if (!d->status->behavior().testFlag(Hidden)) {
        layout << *d->status;
    }
    RuleRectangle &cmdRule = d->console->commandLine().rule();
    cmdRule.setInput(Rule::Left,   d->console->buttons().rule().right())
           .setInput(Rule::Bottom, rule().bottom())
           .setInput(Rule::Right,  layout.widgets().last()->rule().left());
    d->console->commandLine().set(Background(style().colors().colorf("background")));
}
static TimeSpan const FLASH_SPAN = 0.6;
DENG_GUI_PIMPL(TutorialWidget) {
    enum Step {
        Welcome,
        HomeScreen,
        Notifications,
        TaskBar,
        DEMenu,
        ConfigMenus,
        RendererAppearance,
        ConsoleKey,
        Finish
    };
    Step current = Welcome;
    MessageDialog *dlg = nullptr;
    LabelWidget *highlight = nullptr;
    NotificationAreaWidget *notifs = nullptr; ///< Fake notifications just for an example.
    UniqueWidgetPtr<LabelWidget> exampleAlert;
    QTimer flashing;
    bool taskBarInitiallyOpen;
    Untrapper untrapper;
    Impl(Public *i)
        : Base(i)
        , taskBarInitiallyOpen(ClientWindow::main().taskBar().isOpen())
        , untrapper(ClientWindow::main()) {
        exampleAlert.reset(new LabelWidget);
        exampleAlert->setSizePolicy(ui::Expand, ui::Expand);
        exampleAlert->setImage(style().images().image("alert"));
        exampleAlert->setOverrideImageSize(style().fonts().font("default").height());
        exampleAlert->setImageColor(style().colors().colorf("accent"));
        self().add(highlight = new LabelWidget);
        highlight->set(Background(Background::GradientFrame,
                                  style().colors().colorf("accent"), 6));
        highlight->setOpacity(0);
        flashing.setSingleShot(false);
        flashing.setInterval(FLASH_SPAN.asMilliSeconds());
    }
    void startHighlight(GuiWidget const &w) {
        highlight->rule().setRect(w.rule());
        highlight->setOpacity(0);
        highlight->show();
        flashing.start();
        flash();
    }
    /**
     * Animates the highlight flash rectangle. Called periodically.
     */
    void flash() {
        if (highlight->opacity().target() == 0) {
            highlight->setOpacity(.8f, FLASH_SPAN + .1, .1);
        }
        else if (highlight->opacity().target() > .5f) {
            highlight->setOpacity(.2f, FLASH_SPAN);
        }
        else {
            highlight->setOpacity(.8f, FLASH_SPAN);
        }
    }
    void stopHighlight() {
        highlight->hide();
        flashing.stop();
    }
    /**
     * Counts the total number of steps currently available.
     */
    int stepCount() const {
        int count = 0;
        for (Step s = Welcome; s != Finish; s = advanceStep(s)) count++;
        return count;
    }
    int stepOrdinal(Step s) const {
        int ord = stepCount() - 1;
        for (; s != Finish; s = advanceStep(s)) ord--;
        return ord;
    }
    /**
     * Determines which step follows step @a s.
     * @param s  Step.
     * @return The next step.
     */
    Step advanceStep(Step s) const {
        s = Impl::Step(s + 1);
        validateStep(s);
        return s;
    }
    Step previousStep(Step s) const {
        if (s == Welcome) return s;
        Step prev = Welcome;
        while (prev != Finish) {
            Step following = advanceStep(prev);
            if (following == s) break;
            prev = following;
        }
        return prev;
    }
    /**
     * Checks if step @a s is valid for the current engine state and if not,
     * skips to the next valid state.
     *
     * @param s  Current step.
     */
    void validateStep(Step &s) const {
        forever {
            bool skip = false;
            if (!App_GameLoaded()) // in Home
            {
                if (s == RendererAppearance) skip = true;
            }
            else // A game is loaded.
            {
                if (s == HomeScreen) skip = true;
            }
            if (!skip) break;
            s = Step(s + 1);
        }
    }
    void initStep(Step s) {
        deinitStep();
        validateStep(s);
        if (s == Finish) {
            self().stop();
            return;
        }
        current = s;
        bool const isFirstStep = (current == Welcome);
        bool const isLastStep  = (current == Finish - 1);
        dlg = new MessageDialog;
        dlg->useInfoStyle();
        dlg->setDeleteAfterDismissed(true);
        dlg->setClickToClose(false);
        QObject::connect(dlg, SIGNAL(accepted(int)), thisPublic, SLOT(continueToNextStep()));
        QObject::connect(dlg, SIGNAL(rejected(int)), thisPublic, SLOT(stop()));
        dlg->buttons() << new DialogButtonItem(DialogWidget::Accept | DialogWidget::Default | DialogWidget::Id1,
                                               isLastStep? tr("Done") : tr("Next"));
        if (!isFirstStep) {
            dlg->buttons() << new DialogButtonItem(DialogWidget::Action | DialogWidget::Id2, "",
                                                   new SignalAction(thisPublic, SLOT(backToPreviousStep())));
            auto &prevBtn = *dlg->buttonWidget(DialogWidget::Id2);
            prevBtn.setImage(new StyleProceduralImage("fold", prevBtn, 90));
            prevBtn.setImageColor(style().colors().colorf("inverted.text"));
        }
        if (!isLastStep) {
            dlg->buttons() << new DialogButtonItem(DialogWidget::Reject | DialogWidget::Action,
                                                   tr("Close"));
            auto &nextBtn = *dlg->buttonWidget(DialogWidget::Id1);
            nextBtn.setImage(new StyleProceduralImage("fold", nextBtn, -90));
            nextBtn.setImageColor(style().colors().colorf("inverted.text"));
            nextBtn.setTextAlignment(ui::AlignLeft);
        }
        ClientWindow &win = ClientWindow::main();
        switch (current) {
        case Welcome:
            dlg->title().setText(tr("Welcome to Doomsday"));
            dlg->message().setText(tr("This tutorial will give you a brief walkthrough of the "
                                      "major features of Doomsday's UI. You will also get a "
                                      "chance to pick a shortcut key for opening the console.\n\n"
                                      "The tutorial can be restarted later via the application menu."));
            dlg->setAnchor(self().rule().midX(), self().rule().top());
            dlg->setOpeningDirection(ui::Down);
            break;
        case HomeScreen:
            dlg->title().setText(tr("Game Library"));
            dlg->message().setText(tr("Here you can browse the library of available games "
                                      "and configure engine settings. You can also join ongoing "
                                      "multiplayer games and manage your mods and resource packages. "
                                      "You can unload the current game at "
                                      "any time to get back to the Game Library."));
            startHighlight(*root().guiFind("home"));
            break;
        case Notifications:
            notifs = new NotificationAreaWidget("tutorial-notifications");
            notifs->useDefaultPlacement(ClientWindow::main().game().rule(), Const(0));
            root().addOnTop(notifs);
            notifs->showChild(*exampleAlert);
            dlg->title().setText(tr("Notifications"));
            dlg->message().setText(tr("The notification area shows the current notifications. "
                                      "For example, this one here is an example of a warning or an error "
                                      "that has occurred. You can click on the notification icons to "
                                      "get more information.\n\nOther possible notifications include the current "
                                      "FPS, ongoing downloads, and available updates."));
            dlg->setAnchorAndOpeningDirection(exampleAlert->rule(), ui::Down);
            startHighlight(*exampleAlert);
            break;
        case TaskBar:
            dlg->title().setText(tr("Task Bar"));
            dlg->message().setText(tr("The task bar is where you find all the important functionality: loading "
                                      "and switching games, joining a multiplayer game, "
                                      "configuration settings, "
                                      "and a console command line for advanced users.\n\n"
                                      "Press %1 to access the task bar at any time.")
                                   .arg(_E(b) "Shift-Esc" _E(.)));
            win.taskBar().open();
            win.taskBar().closeMainMenu();
            win.taskBar().closeConfigMenu();
            dlg->setAnchor(self().rule().midX(), win.taskBar().rule().top());
            dlg->setOpeningDirection(ui::Up);
            startHighlight(win.taskBar());
            break;
        case DEMenu:
            dlg->title().setText(tr("Application Menu"));
            dlg->message().setText(tr("Click the DE icon in the bottom right corner to open "
                                      "the application menu. "
                                      "You can check for available updates, switch games, or look for "
                                      "ongoing multiplayer games. You can also unload the current game "
                                      "and return to Doomsday's Game Library."));
            win.taskBar().openMainMenu();
            dlg->setAnchorAndOpeningDirection(root().guiFind("de-menu")->rule(), ui::Left);
            startHighlight(*root().guiFind("de-button"));
            break;
        case ConfigMenus:
            dlg->title().setText(tr("Settings"));
            dlg->message().setText(tr("Configuration menus are found under buttons with a gear icon. "
                                      "The task bar's configuration button has the settings for "
                                      "all of Doomsday's subsystems."));
            win.taskBar().openConfigMenu();
            dlg->setAnchorAndOpeningDirection(root().guiFind("conf-menu")->rule(), ui::Left);
            startHighlight(*root().guiFind("conf-button"));
            break;
        case RendererAppearance:
            dlg->title().setText(tr("Appearance"));
            dlg->message().setText(tr("By default Doomsday applies many visual "
                                      "embellishments to how the game world appears. These "
                                      "can be configured individually in the Renderer "
                                      "Appearance editor, or you can use one of the built-in "
                                      "default profiles: %1, %2, or %3.")
                                   .arg(_E(b) "Defaults" _E(.))
                                   .arg(_E(b) "Vanilla" _E(.))
                                   .arg(_E(b) "Amplified" _E(.)));
            win.taskBar().openConfigMenu();
            win.root().guiFind("conf-menu")->as<PopupMenuWidget>().menu()
                    .organizer().itemWidget(tr("Renderer"))->as<ButtonWidget>().trigger();
            dlg->setAnchorAndOpeningDirection(
                        win.root().guiFind("renderersettings")->guiFind("appearance-label")->rule(), ui::Left);
            startHighlight(*root().guiFind("profile-picker"));
            break;
        case ConsoleKey: {
            dlg->title().setText(tr("Console"));
            String msg = tr("The console is a \"Quake style\" command line prompt where "
                            "you enter commands and change variable values. To get started, "
                            "try typing %1 in the console.").arg(_E(b) "help" _E(.));
            if (App_GameLoaded()) {
                msg += tr("\n\nBelow you can see the current keyboard shortcut for accessing the console quickly. "
                          "To change it, click in the box and then press the key or key combination you "
                          "want to assign as the shortcut.");
                InputBindingWidget *bind = InputBindingWidget::newTaskBarShortcut();
                bind->invertStyle();
                dlg->area().add(bind);
            }
            dlg->message().setText(msg);
            dlg->setAnchor(win.taskBar().console().commandLine().rule().left() +
                           rule("gap"),
                           win.taskBar().rule().top());
            dlg->setOpeningDirection(ui::Up);
            dlg->updateLayout();
            startHighlight(win.taskBar().console().commandLine());
            break; }
        default:
            break;
        }
        auto *progress = new ProgressWidget;
        progress->setColor("inverted.text");
        progress->setRange(Rangei(0, stepCount()));
        progress->setProgress(stepOrdinal(current) + 1, 0);
        progress->setMode(ProgressWidget::Dots);
        progress->rule()
                .setInput(Rule::Top,    dlg->buttonsMenu().rule().top())
                .setInput(Rule::Bottom, dlg->buttonsMenu().rule().bottom() -
                                        dlg->buttonsMenu().margins().bottom())
                .setInput(Rule::Left,   dlg->rule().left())
                .setInput(Rule::Right,  dlg->rule().right());
        dlg->add(progress);
        GuiRootWidget &root = self().root();
        root.moveToTop(self());
        root.addOnTop(dlg);
        dlg->open();
    }
    /**
     * Cleans up after a tutorial step is done.
     */
    void deinitStep() {
        if (dlg) {
            dlg->close(0);
            dlg = 0;
        }
        stopHighlight();
        ClientWindow &win = ClientWindow::main();
        switch (current) {
        case Notifications:
            notifs->hideChild(*exampleAlert);
            QTimer::singleShot(500, notifs, SLOT(guiDeleteLater()));
            notifs = 0;
            break;
        case DEMenu:
            win.taskBar().closeMainMenu();
            break;
        case ConfigMenus:
            win.taskBar().closeConfigMenu();
            break;
        case RendererAppearance:
            win.taskBar().closeConfigMenu();
            break;
        default:
            break;
        }
    }
};
TutorialWidget::TutorialWidget()
    : GuiWidget("tutorial"), d(new Impl(this)) {
    connect(&d->flashing, SIGNAL(timeout()), this, SLOT(flashHighlight()));
}
void TutorialWidget::start() {
    ClientWindow::main().fadeInTaskBarBlur(.5);
    d->initStep(Impl::Welcome);
}
void TutorialWidget::stop() {
    if (!d->taskBarInitiallyOpen) {
        ClientWindow::main().taskBar().close();
    }
    d->deinitStep();
    ClientWindow::main().fadeOutTaskBarBlur(.5);
    QTimer::singleShot(500, this, SLOT(dismiss()));
}
void TutorialWidget::dismiss() {
    hide();
    guiDeleteLater();
}
void TutorialWidget::flashHighlight() {
    d->flash();
}
bool TutorialWidget::handleEvent(Event const &event) {
    GuiWidget::handleEvent(event);
    return true;
}
void TutorialWidget::continueToNextStep() {
    d->initStep(d->advanceStep(d->current));
}
void TutorialWidget::backToPreviousStep() {
    d->initStep(d->previousStep(d->current));
}
