﻿/** @file convexsubspaceproxy.cpp  BSP builder convex subspace proxy.
 *
 * @authors Copyright © 2013-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "world/bsp/convexsubspaceproxy.h"
#include "Face"
#include "HEdge"
#include "Mesh"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Line"
#include "Sector"
#include "world/bsp/linesegment.h"
#include "world/clientserverworld.h" /// validCount @todo Remove me
#include <de/Log>
#include <QHash>
#include <QSet>
#include <QVarLengthArray>
#include <QtAlgorithms>
/** @file hplane.cpp  World map BSP builder half-plane.
 *
 * Originally based on glBSP 2.24 (in turn, based on BSP 2.3)
 * @see http://sourceforge.net/projects/glbsp/
 *
 * @authors Copyright © 2007-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bsp/hplane.h"
#include <memory>
#include <QtAlgorithms>
#include <de/mathutil.h> // M_InverseAngle
#include <de/vector1.h>  // remove me
#include <de/Error>
#include <de/Log>
#include "Line"
#include "Sector"
#include "Vertex"
#include "world/bsp/edgetip.h"
#include "world/bsp/linesegment.h"
#include "world/bsp/partitioner.h"
/** @file linesegment.cpp  BSP Builder Line Segment.
 *
 * Originally based on glBSP 2.24 (in turn, based on BSP 2.3)
 * @see http://sourceforge.net/projects/glbsp/
 *
 * @authors Copyright © 2007-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bsp/linesegment.h"
#include <de/vector1.h> /// @todo remove me
#include <de/aabox.h> // M_BoxOnLineSide2
#include <de/Observers>
#include <de/Vector>
#include "world/bsp/convexsubspaceproxy.h"
#include "world/bsp/superblockmap.h"
/** @file partitioner.cpp  World map, binary space partitioner.
 *
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 * @authors Copyright © 1997-1998 Raphael.Quinet <raphael.quinet@eed.ericsson.se>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bsp/partitioner.h"
#include <algorithm>
#include <QHash>
#include <QList>
#include <QtAlgorithms>
#include <de/vector1.h>
#include <de/LogBuffer>
#include <doomsday/BspNode>
#include "BspLeaf"
#include "Line"
#include "Sector"
#include "Vertex"
#include "world/bsp/convexsubspaceproxy.h"
#include "world/bsp/edgetip.h"
#include "world/bsp/hplane.h"
#include "world/bsp/linesegment.h"
#include "world/bsp/partitionevaluator.h"
#include "world/bsp/superblockmap.h"
/** @file partitionevaluator.cpp  Evaluator for a would-be BSP.
 *
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bsp/partitionevaluator.h"
#include <QList>
#include <de/Log>
#include <de/String>
#include <de/Task>
#include <de/TaskPool>
#include "world/bsp/partitioner.h"
#include "world/clientserverworld.h" // validCount
/** @file superblockmap.cpp  BSP line segment blockmap block.
 *
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bsp/superblockmap.h"
#include <de/vector1.h>
using namespace de;
namespace world {
namespace bsp {
typedef QList<LineSegmentSide *> SegmentList;
/**
 * Represents a clockwise ordering of a subset of the line segments and implements logic
 * for partitioning the subset into @em contiguous ranges, for geometry construction.
 */
struct Continuity {
    typedef QList<OrderedSegment *> OrderedSegmentList;
    Sector *sector = nullptr;  ///< Front sector uniformly referenced by all line segments.
    double coverage = 0;       ///< Coverage metric.
    int discordSegments = 0;   ///< Number of discordant (i.e., non-contiguous) line segments.
    int norm = 0;
    int part = 0;
    int self = 0;
    OrderedSegmentList orderedSegs;  ///< Ordered line segments (not owned).
    OrderedSegmentList discordSegs;  ///< The discordant line segment subset (not owned).
    Continuity(Sector *frontSector) {
        sector = frontSector;
    }
    /**
     * Perform heuristic comparison between two continuities to determine a
     * preference order for BSP sector attribution. The algorithm used weights
     * the two choices according to the number and "type" of the referencing
     * line segments and the "coverage" metric.
     *
     * @return  @c true if "this" choice is rated better than @a other.
     *
     * @todo Remove when heuristic sector selection is no longer necessary.
     */
    bool operator < (Continuity const &other) const {
        if(norm == other.norm) {
            return !(coverage < other.coverage);
        }
        return norm > other.norm;
    }
    /**
     * Assumes that segments are added in clockwise order.
     */
    void addOneSegment(OrderedSegment const &oseg) {
        DENG2_ASSERT(oseg.segment->sectorPtr() == sector);
        OrderedSegmentList *list = &orderedSegs;
        for(OrderedSegment const *other : orderedSegs) {
            if(oseg == *other) {
                list = &discordSegs;
                break;
            }
        }
        list->append(const_cast<OrderedSegment *>(&oseg));
        LineSegmentSide const &seg = *oseg.segment;
        if(!seg.hasMapSide()) {
            part += 1;
        }
        else if(seg.mapSide().line().isSelfReferencing()) {
            self += 1;
        }
        else {
            norm += 1;
        }
        if(oseg.fromAngle > oseg.toAngle)
            coverage += oseg.fromAngle - oseg.toAngle;
        else
            coverage += oseg.fromAngle + (360.0 - oseg.toAngle);
    }
    void evaluate() {
        discordSegments = 0;
        for(int i = 0; i < orderedSegs.count() - 1; ++i) {
            LineSegmentSide const &segA = *orderedSegs[i  ]->segment;
            LineSegmentSide const &segB = *orderedSegs[i+1]->segment;
            if(segB.from().origin() != segA.to().origin())
                discordSegments += 1;
        }
        if(orderedSegs.count() > 1) {
            LineSegmentSide const &segB = *orderedSegs.last()->segment;
            LineSegmentSide const &segA = *orderedSegs.first()->segment;
            if(segB.to().origin() != segA.from().origin())
                discordSegments += 1;
        }
    }
#ifdef DENG2_DEBUG
    void debugPrint() const {
        LOGDEV_MAP_MSG("Continuity %p (sector:%i, coverage:%f, discord:%i)")
                << this
                << (sector? sector->indexInArchive() : -1)
                << coverage
                << discordSegments;
        for(OrderedSegment const *oseg : orderedSegs) {
            oseg->debugPrint();
        }
        for(OrderedSegment const *oseg : discordSegs) {
            oseg->debugPrint();
        }
    }
#endif
};
DENG2_PIMPL_NOREF(ConvexSubspaceProxy) {
    typedef QSet<LineSegmentSide *> Segments;
    Segments segments;                ///< All line segments.
    OrderedSegments orderedSegments;  ///< All line segments in clockwise order, with angle info.
    bool needRebuildOrderedSegments;  ///< @c true= the ordered segment list needs to be rebuilt.
    BspLeaf *bspLeaf;                 ///< BSP leaf attributed to the subspace (if any).
    Impl()
        : needRebuildOrderedSegments(false)
        , bspLeaf                   (nullptr)
    {}
    Impl(Impl const &other)
        : de::IPrivate()
        , segments                  (other.segments)
        , orderedSegments           (other.orderedSegments)
        , needRebuildOrderedSegments(other.needRebuildOrderedSegments)
        , bspLeaf                   (other.bspLeaf)
    {}
    /**
     * Returns @c true iff at least one line segment in the set is derived
     * from a map line.
     */
    bool haveMapLineSegment() const {
        for(LineSegmentSide const *seg : segments) {
            if(seg->hasMapSide())
                return true;
        }
        return false;
    }
    Vector2d findCenter() const {
        Vector2d center;
        int numPoints = 0;
        for(LineSegmentSide const *seg : segments) {
            center += seg->from().origin();
            center += seg->to().origin();
            numPoints += 2;
        }
        if(numPoints) {
            center /= numPoints;
        }
        return center;
    }
    /**
     * Builds the ordered list of line segments, which, is sorted firstly in
     * a clockwise order (i.e., descending angles) according to the origin of
     * their 'from' vertex relative to @a point. A secondary ordering is also
     * applied such that line segments with the same origin coordinates are
     * sorted by descending 'to' angle.
     */
    void buildOrderedSegments(Vector2d const &point) {
        needRebuildOrderedSegments = false;
        orderedSegments.clear();
        for(LineSegmentSide *seg : segments) {
            Vector2d fromDist = seg->from().origin() - point;
            Vector2d toDist   = seg->to().origin() - point;
            OrderedSegment oseg;
            oseg.segment   = seg;
            oseg.fromAngle = M_DirectionToAngleXY(fromDist.x, fromDist.y);
            oseg.toAngle   = M_DirectionToAngleXY(toDist.x, toDist.y);
            orderedSegments.append(oseg);
        }
        int const numSegments = orderedSegments.count();
        for(int pass = 0; pass < numSegments - 1; ++pass) {
            bool swappedAny = false;
            for(int i = 0; i < numSegments - 1; ++i) {
                OrderedSegment const &a = orderedSegments.at(i);
                OrderedSegment const &b = orderedSegments.at(i+1);
                if(a.fromAngle < b.fromAngle) {
                    orderedSegments.swap(i, i + 1);
                    swappedAny = true;
                }
            }
            if(!swappedAny) break;
        }
        for(int pass = 0; pass < numSegments - 1; ++pass) {
            bool swappedAny = false;
            for(int i = 0; i < numSegments - 1; ++i) {
                OrderedSegment const &a = orderedSegments.at(i);
                OrderedSegment const &b = orderedSegments.at(i+1);
                if(a.fromAngle == b.fromAngle) {
                    if(b.segment->length() > a.segment->length()) {
                        orderedSegments.swap(i, i + 1);
                        swappedAny = true;
                    }
                }
            }
            if(!swappedAny) break;
        }
    }
    Impl &operator = (Impl const &) = delete; // no assignment
};
ConvexSubspaceProxy::ConvexSubspaceProxy()
    : d(new Impl)
{}
ConvexSubspaceProxy::ConvexSubspaceProxy(QList<LineSegmentSide *> const &segments)
    : d(new Impl) {
    addSegments(segments);
}
ConvexSubspaceProxy::ConvexSubspaceProxy(ConvexSubspaceProxy const &other)
    : d(new Impl(*other.d))
{}
ConvexSubspaceProxy &ConvexSubspaceProxy::operator = (ConvexSubspaceProxy const &other) {
    d.reset(new Impl(*other.d));
    return *this;
}
void ConvexSubspaceProxy::addSegments(QList<LineSegmentSide *> const &newSegments) {
    int sizeBefore = d->segments.size();
    d->segments.unite(QSet<LineSegmentSide *>::fromList(newSegments));
    if(d->segments.size() != sizeBefore) {
        d->needRebuildOrderedSegments = true;
    }
#ifdef DENG2_DEBUG
    int numSegmentsAdded = d->segments.size() - sizeBefore;
    if(numSegmentsAdded < newSegments.size()) {
        LOG_DEBUG("ConvexSubspaceProxy pruned %i duplicate segments")
                << (newSegments.size() - numSegmentsAdded);
    }
#endif
}
void ConvexSubspaceProxy::addOneSegment(LineSegmentSide const &newSegment) {
    int sizeBefore = d->segments.size();
    d->segments.insert(const_cast<LineSegmentSide *>(&newSegment));
    if(d->segments.size() != sizeBefore) {
        d->needRebuildOrderedSegments = true;
    }
    else {
        LOG_DEBUG("ConvexSubspaceProxy pruned one duplicate segment");
    }
}
void ConvexSubspaceProxy::buildGeometry(BspLeaf &leaf, Mesh &mesh) const {
    LOG_AS("ConvexSubspaceProxy::buildGeometry");
    if(segmentCount() >= 3 && !d->haveMapLineSegment())
        throw Error("ConvexSubspaceProxy::buildGeometry", "No map line segment");
    if(d->needRebuildOrderedSegments) {
        d->buildOrderedSegments(d->findCenter());
    }
    /*
     * Build the line segment -> sector continuity map.
     */
    typedef QList<Continuity> Continuities;
    Continuities continuities;
    typedef QHash<Sector *, Continuity *> SectorContinuityMap;
    SectorContinuityMap scMap;
    for(OrderedSegment const &oseg : d->orderedSegments) {
        Sector *frontSector = oseg.segment->sectorPtr();
        SectorContinuityMap::iterator found = scMap.find(frontSector);
        if(found == scMap.end()) {
            continuities.append(Continuity(frontSector));
            found = scMap.insert(frontSector, &continuities.last());
        }
        Continuity *conty = found.value();
        conty->addOneSegment(oseg);
    }
    QVarLengthArray<Mesh *, 2> extraMeshes;
    int extraMeshSegments = 0;
    for(int i = 0; i < continuities.count(); ++i) {
        Continuity &conty = continuities[i];
        conty.evaluate();
        if(!conty.discordSegs.isEmpty()) {
            Mesh *extraMesh = nullptr;
            Face *face      = nullptr;
            for(OrderedSegment const *oseg : conty.discordSegs) {
                LineSegmentSide *lineSeg = oseg->segment;
                LineSide *mapSide        = lineSeg->mapSidePtr();
                if(!mapSide) continue;
                if(!extraMesh) {
                    extraMesh = new Mesh;
                    face = extraMesh->newFace();
                }
                HEdge *hedge = extraMesh->newHEdge(lineSeg->from());
                LineSideSegment *seg = mapSide->addSegment(*hedge);
                extraMeshSegments += 1;
#ifdef __CLIENT__
                seg->setLineSideOffset(Vector2d(mapSide->from().origin() - lineSeg->from().origin()).length());
                seg->setLength(Vector2d(lineSeg->to().origin() - lineSeg->from().origin()).length());
#else
                DENG2_UNUSED(seg);
#endif
                hedge->setNext(face->hedge());
                face->setHEdge(hedge);
                if(lineSeg->back().hasHEdge()) {
                    lineSeg->back().hedge().setTwin(hedge);
                    hedge->setTwin(lineSeg->back().hedgePtr());
                }
                lineSeg->setHEdge(hedge);
            }
            if(extraMesh) {
                HEdge *hedge = face->hedge();
                forever {
                    face->_hedgeCount += 1;
                    hedge->setFace(face);
                    if(hedge->hasNext()) {
                        hedge->next().setPrev(hedge);
                        hedge = &hedge->next();
                    }
                    else {
                        hedge->setNext(face->hedge());
                        hedge->next().setPrev(hedge);
                        break;
                    }
                }
                face->updateBounds();
                face->updateCenter();
                extraMeshes.append(extraMesh);
            }
        }
    }
    qSort(continuities.begin(), continuities.end());
    leaf.setSector(continuities.first().sector);
/*#ifdef DENG_DEBUG
    LOG_INFO("ConvexSubspace %s BSP sector:%i (%i continuities)")
            << d->findCenter().asText()
            << (leaf.sectorPtr()? leaf.sectorPtr()->indexInArchive() : -1)
            << continuities.count();
    for(Continuity const &conty : continuities) {
        conty.debugPrint();
    }
#endif*/
    if(segmentCount() - extraMeshSegments >= 3) {
        Face *face = mesh.newFace();
        for(int i = d->orderedSegments.size(); i-- > 0; ) {
            LineSegmentSide *lineSeg = d->orderedSegments[i].segment;
            if(lineSeg->hasHEdge())
                continue;
            HEdge *hedge = mesh.newHEdge(lineSeg->from());
            if(LineSide *mapSide = lineSeg->mapSidePtr()) {
                LineSideSegment *seg = mapSide->addSegment(*hedge);
#ifdef __CLIENT__
                seg->setLineSideOffset(Vector2d(mapSide->from().origin() - lineSeg->from().origin()).length());
                seg->setLength(Vector2d(lineSeg->to().origin() - lineSeg->from().origin()).length());
#else
                DENG2_UNUSED(seg);
#endif
            }
            hedge->setNext(face->hedge());
            face->setHEdge(hedge);
            if(lineSeg->back().hasHEdge()) {
                lineSeg->back().hedge().setTwin(hedge);
                hedge->setTwin(lineSeg->back().hedgePtr());
            }
            lineSeg->setHEdge(hedge);
        }
        HEdge *hedge = face->hedge();
        forever {
            face->_hedgeCount += 1;
            hedge->setFace(face);
            if(hedge->hasNext()) {
                hedge->next().setPrev(hedge);
                hedge = &hedge->next();
            }
            else {
                hedge->setNext(face->hedge());
                hedge->next().setPrev(hedge);
                break;
            }
        }
        face->updateBounds();
        face->updateCenter();
        leaf.setSubspace(ConvexSubspace::newFromConvexPoly(*face));
        for(int i = 0; i < extraMeshes.count(); ++i) {
            leaf.subspace().assignExtraMesh(*extraMeshes.at(i));
        }
    }
    /*else
    {
        qDeleteAll(extraMeshes);
    }*/
}
int ConvexSubspaceProxy::segmentCount() const {
    return d->segments.count();
}
OrderedSegments const &ConvexSubspaceProxy::segments() const {
    if(d->needRebuildOrderedSegments) {
        d->buildOrderedSegments(d->findCenter());
    }
    return d->orderedSegments;
}
BspLeaf *ConvexSubspaceProxy::bspLeaf() const {
    return d->bspLeaf;
}
void ConvexSubspaceProxy::setBspLeaf(BspLeaf *newBspLeaf) {
    d->bspLeaf = newBspLeaf;
}
}  // namespace bsp
}  // namespace world
namespace world {
namespace bsp {
HPlane::Intercept::Intercept(ddouble distance, LineSegmentSide &lineSeg, dint edge) {
    _distance = distance;
    _lineSeg  = &lineSeg;
    _edge     = edge;
}
LineSegmentSide &HPlane::Intercept::lineSegment() const {
    DENG2_ASSERT(_lineSeg);
    return *_lineSeg;
}
dint HPlane::Intercept::lineSegmentEdge() const {
    return _edge;
}
Sector *HPlane::Intercept::before() const {
    return _before? _before->sectorPtr() : nullptr;
}
Sector *HPlane::Intercept::after() const {
    return _after? _after->sectorPtr() : nullptr;
}
LineSegmentSide *HPlane::Intercept::beforeLineSegment() const {
    return _before;
}
LineSegmentSide *HPlane::Intercept::afterLineSegment() const {
    return _after;
}
#ifdef DENG2_DEBUG
void HPlane::Intercept::debugPrint() const {
    LOGDEV_MAP_MSG("Vertex #%i %s beforeSector: #%d afterSector: #%d %s")
        << vertex().indexInMap()
        << vertex().origin().asText()
        << (_before && _before->hasSector() ? _before->sector().indexInArchive() : -1)
        << (_after  && _after->hasSector()  ? _after-> sector().indexInArchive() : -1);
}
#endif
DENG2_PIMPL(HPlane) {
    Partition partition;          ///< The partition line.
    ddouble length;               ///< Direction vector length.
    ddouble angle;                ///< Cartesian world angle.
    slopetype_t slopeType;        ///< Logical world angle classification.
    ddouble perp;                 ///< Perpendicular scale factor.
    ddouble para;                 ///< Parallel scale factor.
    LineSegmentSide *lineSegment = nullptr;  ///< Source of the partition (if any, not owned).
    Intercepts intercepts;                   ///< Points along the half-plane.
    bool needSortIntercepts = false;         ///< @c true= @var intercepts requires sorting.
    Impl(Public *i, Partition const &partition)
        : Base(i)
        , partition  (partition)
        , length     (partition.direction.length())
        , angle      (M_DirectionToAngleXY(partition.direction.x, partition.direction.y))
        , slopeType  (M_SlopeTypeXY(partition.direction.x, partition.direction.y))
        , perp       ( partition.origin.y * partition.direction.x - partition.origin.x * partition.direction.y)
        , para       (-partition.origin.x * partition.direction.x - partition.origin.y * partition.direction.y)
        , needSortIntercepts(false)
    {}
    /**
     * Find an intercept by @a vertex.
     */
    Intercept *interceptByVertex(Vertex const &vertex) {
        for(Intercept const &icpt : intercepts) {
            if(&icpt.vertex() == &vertex)
                return const_cast<Intercept *>(&icpt);
        }
        return nullptr;
    }
    /**
     * Merges @a next into @a cur.
     */
    static void mergeIntercepts(Intercept &cur, Intercept const &next) {
        /*
        LOG_AS("HPlane::mergeIntercepts");
        cur.debugPrint();
        next.debugPrint();
        */
        if(&cur.lineSegment().line() == &next.lineSegment().line())
            return;
        if(cur.lineSegmentIsSelfReferencing() &&
           !next.lineSegmentIsSelfReferencing()) {
            if(cur.before() && next.before())
                cur._before = next._before;
            if(cur.after() && next.after())
                cur._after = next._after;
        }
        if(!cur.before() && next.before())
            cur._before = next._before;
        if(!cur.after() && next.after())
            cur._after = next._after;
        /*
        LOG_TRACE("Result:");
        cur.debugPrint();
        */
    }
};
HPlane::HPlane(Partition const &partition) : d(new Impl(this, partition))
{}
void HPlane::clearIntercepts() {
    d->intercepts.clear();
    d->needSortIntercepts = false;
}
void HPlane::configure(LineSegmentSide const &newBaseSeg) {
    DENG2_ASSERT(newBaseSeg.hasMapSide());
    LOG_AS("HPlane::configure");
    clearIntercepts();
    LineSide &mapSide = newBaseSeg.mapSide();
    d->partition.direction = mapSide.to().origin() - mapSide.from().origin();
    d->partition.origin    = mapSide.from().origin();
    d->lineSegment = const_cast<LineSegmentSide *>(&newBaseSeg);
    d->length    = d->partition.direction.length();
    d->angle     = M_DirectionToAngleXY(d->partition.direction.x, d->partition.direction.y);
    d->slopeType = M_SlopeTypeXY(d->partition.direction.x, d->partition.direction.y);
    d->perp = d->partition.origin.y * d->partition.direction.x
            - d->partition.origin.x * d->partition.direction.y;
    d->para = -d->partition.origin.x * d->partition.direction.x
            -  d->partition.origin.y * d->partition.direction.y;
}
/**
 * Determines whether a conceptual line oriented at @a vtx and "pointing"
 * at the specified world @a angle enters an "open" sector (which is to say
 * that said line does not enter void space and does not intercept with any
 * existing map or partition line segment in the plane, thus "closed").
 *
 * @return  The "open" sector at this angle; otherwise @c 0 (closed).
 */
static LineSegmentSide *lineSegAtAngle(EdgeTips const &tips, coord_t angle) {
    if(tips.at(angle))
        return nullptr; // Closed.
    if(EdgeTip const *tip = tips.after(angle)) {
        return tip->hasFront()? &tip->front() : nullptr;
    }
    if(EdgeTip const *tip = tips.largest()) {
        return tip->hasBack()? &tip->back() : nullptr;
    }
    return nullptr; // No edge tips.
}
double HPlane::intersect(LineSegmentSide const &lineSeg, int edge) {
    Vertex &vertex = lineSeg.vertex(edge);
    coord_t pointV1[2]     = { vertex.origin().x, vertex.origin().y };
    coord_t directionV1[2] = { d->partition.direction.x, d->partition.direction.y };
    return V2d_PointLineParaDistance(pointV1, directionV1, d->para, d->length);
}
HPlane::Intercept *HPlane::intercept(LineSegmentSide const &lineSeg, int edge,
    EdgeTips const &edgeTips) {
    bool const selfRef = (lineSeg.hasMapSide() && lineSeg.mapLine().isSelfReferencing());
    Intercept *icpt;
    if((icpt = d->interceptByVertex(lineSeg.vertex(edge)))) {
        if(!(icpt->lineSegmentIsSelfReferencing() && !selfRef)) {
            return icpt;
        }
    }
    else {
        d->intercepts.append(
            Intercept(intersect(lineSeg, edge),
                      const_cast<LineSegmentSide &>(lineSeg), edge));
        icpt = &d->intercepts.last();
        d->needSortIntercepts = true;
    }
    icpt->_lineSeg = const_cast<LineSegmentSide *>(&lineSeg);
    icpt->_edge    = edge;
    icpt->_before  = lineSegAtAngle(edgeTips, inverseAngle());
    icpt->_after   = lineSegAtAngle(edgeTips, angle());
    return icpt;
}
void HPlane::sortAndMergeIntercepts() {
    if(!d->needSortIntercepts) return;
    qSort(d->intercepts.begin(), d->intercepts.end());
    for(int i = 0; i < d->intercepts.count() - 1; ++i) {
        Intercept &cur  = d->intercepts[i];
        Intercept &next = d->intercepts[i+1];
        ddouble distance = next.distance() - cur.distance();
        if(distance < -0.1) {
            throw Error("HPlane::sortAndMergeIntercepts",
                        String("Invalid intercept order - %1 > %2")
                            .arg(cur.distance(),  0, 'f', 3)
                            .arg(next.distance(), 0, 'f', 3));
        }
        if(distance <= HPLANE_INTERCEPT_MERGE_DISTANCE_EPSILON) {
            d->mergeIntercepts(cur, next);
            d->intercepts.removeAt(i + 1);
            i -= 1;
        }
    }
    d->needSortIntercepts = false;
}
Partition const &HPlane::partition() const {
    return d->partition;
}
coord_t HPlane::angle() const {
    return d->angle;
}
coord_t HPlane::inverseAngle() const {
    return M_InverseAngle(angle());
}
slopetype_t HPlane::slopeType() const {
    return d->slopeType;
}
LineSegmentSide *HPlane::lineSegment() const {
    return d->lineSegment;
}
void HPlane::distance(LineSegmentSide const &lineSeg, coord_t *fromDist, coord_t *toDist) const {
    if(!fromDist && !toDist) return;
    if(d->lineSegment && &d->lineSegment->mapSide().line() == lineSeg.partitionMapLine()) {
        if(fromDist) *fromDist = 0;
        if(toDist)   *toDist   = 0;
        return;
    }
    coord_t toSegDirectionV1[2] = { d->partition.direction.x, d->partition.direction.y } ;
    if(fromDist) {
        coord_t fromV1[2] = { lineSeg.from().origin().x, lineSeg.from().origin().y };
        *fromDist = V2d_PointLinePerpDistance(fromV1, toSegDirectionV1, d->perp, d->length);
    }
    if(toDist) {
        coord_t toV1[2]   = { lineSeg.to().origin().x, lineSeg.to().origin().y };
        *toDist = V2d_PointLinePerpDistance(toV1, toSegDirectionV1, d->perp, d->length);
    }
}
LineRelationship HPlane::relationship(LineSegmentSide const &lineSeg,
    coord_t *retFromDist, coord_t *retToDist) const {
    coord_t fromDist, toDist;
    distance(lineSeg, &fromDist, &toDist);
    LineRelationship rel = lineRelationship(fromDist, toDist);
    if(retFromDist) *retFromDist = fromDist;
    if(retToDist)   *retToDist   = toDist;
    return rel;
}
HPlane::Intercepts const &HPlane::intercepts() const {
    return d->intercepts;
}
#ifdef DENG2_DEBUG
void HPlane::printIntercepts() const {
    uint index = 0;
    for(Intercept const &icpt : d->intercepts) {
        LOG_DEBUG(" %u: >%1.2f") << (index++) << icpt.distance();
        icpt.debugPrint();
    }
}
#endif
}  // namespace bsp
}  // namespace world
namespace world {
namespace bsp {
DENG2_PIMPL_NOREF(LineSegment::Side) {
    LineSegment *line = nullptr;
    Vector2d direction;
    LineSide *mapSide = nullptr;
    Line *partitionMapLine = nullptr;
    LineSegment::Side *rightNeighbor = nullptr;
    LineSegment::Side *leftNeighbor  = nullptr;
    LineSegmentBlockTreeNode *blockTreeNode = nullptr;
    ConvexSubspaceProxy *convexSubspace = nullptr;
    Sector *sector = nullptr;
    coord_t pLength = 0;
    coord_t pAngle  = 0;
    coord_t pPara   = 0;
    coord_t pPerp   = 0;
    slopetype_t pSlopeType = ST_VERTICAL;
    HEdge *hedge = nullptr;
    inline LineSegment::Side **neighborAdr(int edge) {
        return edge ? &rightNeighbor : &leftNeighbor;
    }
};
LineSegment::Side::Side(LineSegment &line) : d(new Impl) {
    d->line = &line;
}
void LineSegment::Side::updateCache() {
    d->direction  = to().origin() - from().origin();
    d->pLength    = d->direction.length();
    DENG2_ASSERT(d->pLength > 0);
    d->pAngle     = M_DirectionToAngleXY(d->direction.x, d->direction.y);
    d->pSlopeType = M_SlopeTypeXY(d->direction.x, d->direction.y);
    d->pPerp      =  from().origin().y * d->direction.x - from().origin().x * d->direction.y;
    d->pPara      = -from().origin().x * d->direction.x - from().origin().y * d->direction.y;
}
LineSegment &LineSegment::Side::line() const {
    return *d->line;
}
int LineSegment::Side::lineSideId() const {
    return &d->line->front() == this? LineSegment::Front : LineSegment::Back;
}
Vector2d const &LineSegment::Side::direction() const {
    return d->direction;
}
slopetype_t LineSegment::Side::slopeType() const {
    return d->pSlopeType;
}
coord_t LineSegment::Side::angle() const {
    return d->pAngle;
}
bool LineSegment::Side::hasHEdge() const {
    return d->hedge != nullptr;
}
HEdge &LineSegment::Side::hedge() const {
    if(d->hedge) {
        return *d->hedge;
    }
    throw MissingHEdgeError("LineSegment::Side::hedge", "No half-edge is associated");
}
void LineSegment::Side::setHEdge(HEdge *newHEdge) {
    d->hedge = newHEdge;
}
ConvexSubspaceProxy *LineSegment::Side::convexSubspace() const {
    return d->convexSubspace;
}
void LineSegment::Side::setConvexSubspace(ConvexSubspaceProxy *newConvexSubspace) {
    d->convexSubspace = newConvexSubspace;
}
bool LineSegment::Side::hasMapSide() const {
    return d->mapSide != nullptr;
}
LineSide &LineSegment::Side::mapSide() const {
    if(d->mapSide) {
        return *d->mapSide;
    }
    throw MissingMapSideError("LineSegment::Side::mapSide", "No map line side is attributed");
}
void LineSegment::Side::setMapSide(LineSide *newMapSide) {
    d->mapSide = newMapSide;
}
Line *LineSegment::Side::partitionMapLine() const {
    return d->partitionMapLine;
}
void LineSegment::Side::setPartitionMapLine(Line *newMapLine) {
    d->partitionMapLine = newMapLine;
}
bool LineSegment::Side::hasNeighbor(int edge) const {
    return (*d->neighborAdr(edge)) != nullptr;
}
LineSegment::Side &LineSegment::Side::neighbor(int edge) const {
    LineSegment::Side **neighborAdr = d->neighborAdr(edge);
    if(*neighborAdr) {
        return **neighborAdr;
    }
    throw MissingNeighborError("LineSegment::Side::neighbor", QString("No %1 neighbor is attributed").arg(edge? "Right" : "Left"));
}
void LineSegment::Side::setNeighbor(int edge, LineSegment::Side *newNeighbor) {
    *d->neighborAdr(edge) = newNeighbor;
}
/*LineSegmentBlockTreeNode*/ void *LineSegment::Side::blockTreeNodePtr() const
{
    return d->blockTreeNode;
}
void LineSegment::Side::setBlockTreeNode(/*LineSegmentBlockTreeNode*/ void *newNode)
{
    d->blockTreeNode = reinterpret_cast<LineSegmentBlockTreeNode *>(newNode);
}
bool LineSegment::Side::hasSector() const {
    return d->sector != nullptr;
}
Sector &LineSegment::Side::sector() const {
    if(d->sector) {
        return *d->sector;
    }
    throw LineSegment::MissingSectorError("LineSegment::Side::sector", "No sector is attributed");
}
void LineSegment::Side::setSector(Sector *newSector) {
    d->sector = newSector;
}
coord_t LineSegment::Side::length() const {
    return d->pLength;
}
coord_t LineSegment::Side::distance(Vector2d point) const {
    coord_t const pointV1[2]     = { point.x, point.y };
    coord_t const directionV1[2] = { d->direction.x, d->direction.y };
    return V2d_PointLineParaDistance(pointV1, directionV1, d->pPara, d->pLength);
}
void LineSegment::Side::distance(LineSegment::Side const &other, coord_t *fromDist, coord_t *toDist) const {
    if(!fromDist && !toDist) return;
    if(d->partitionMapLine && d->partitionMapLine == other.partitionMapLine()) {
        if(fromDist) *fromDist = 0;
        if(toDist)   *toDist   = 0;
        return;
    }
    coord_t toSegDirectionV1[2] = { other.direction().x, other.direction().y } ;
    if(fromDist) {
        coord_t fromV1[2] = { from().origin().x, from().origin().y };
        *fromDist = V2d_PointLinePerpDistance(fromV1, toSegDirectionV1, other.d->pPerp, other.d->pLength);
    }
    if(toDist) {
        coord_t toV1[2]   = { to().origin().x, to().origin().y };
        *toDist = V2d_PointLinePerpDistance(toV1, toSegDirectionV1, other.d->pPerp, other.d->pLength);
    }
}
LineRelationship lineRelationship(coord_t fromDist, coord_t toDist) {
    static coord_t const distEpsilon = LINESEGMENT_INCIDENT_DISTANCE_EPSILON;
    if(de::abs(fromDist) <= distEpsilon && de::abs(toDist) <= distEpsilon) {
        return Collinear;
    }
    if(fromDist > -distEpsilon && toDist > -distEpsilon) {
        if(fromDist < distEpsilon || toDist < distEpsilon) return RightIntercept;
        return Right;
    }
    if(fromDist < distEpsilon && toDist < distEpsilon) {
        if(fromDist > -distEpsilon || toDist > -distEpsilon) return LeftIntercept;
        return Left;
    }
    return Intersects;
}
LineRelationship LineSegment::Side::relationship(LineSegment::Side const &other,
    coord_t *retFromDist, coord_t *retToDist) const {
    coord_t fromDist, toDist;
    distance(other, &fromDist, &toDist);
    LineRelationship rel = lineRelationship(fromDist, toDist);
    if(retFromDist) *retFromDist = fromDist;
    if(retToDist)   *retToDist   = toDist;
    return rel;
}
int LineSegment::Side::boxOnSide(AABoxd const &box) const {
    coord_t const fromV1[2]      = { from().origin().x, from().origin().y };
    coord_t const directionV1[2] = { d->direction.x, d->direction.y } ;
    return M_BoxOnLineSide2(&box, fromV1, directionV1, d->pPerp, d->pLength,
                            LINESEGMENT_INCIDENT_DISTANCE_EPSILON);
}
DENG2_PIMPL(LineSegment)
, DENG2_OBSERVES(Vertex, OriginChange) {
    Vertex *from;
    Vertex *to;
    LineSegment::Side front;
    LineSegment::Side back;
    Impl(Public *i, Vertex &from_, Vertex &to_)
        : Base (i)
        , from (&from_)
        , to   (&to_)
        , front(*i)
        , back (*i) {
        from->audienceForOriginChange += this;
        to->audienceForOriginChange   += this;
    }
    inline Vertex **vertexAdr(int edge) {
        return edge? &to : &from;
    }
    void replaceVertex(int edge, Vertex &newVertex) {
        Vertex **adr = vertexAdr(edge);
        if(*adr && *adr == &newVertex) return;
        if(*adr) (*adr)->audienceForOriginChange -= this;
        *adr = &newVertex;
        if(*adr) (*adr)->audienceForOriginChange += this;
        front.updateCache();
        back.updateCache();
    }
    void vertexOriginChanged(Vertex & /*vertex*/)
    {
        front.updateCache();
        back.updateCache();
    }
};
LineSegment::LineSegment(Vertex &from, Vertex &to) : d(new Impl(this, from, to)) {
    d->front.updateCache();
    d->back.updateCache();
}
LineSegment::Side &LineSegment::side(int back) {
    return back? d->back : d->front;
}
LineSegment::Side const &LineSegment::side(int back) const {
    return back? d->back : d->front;
}
Vertex &LineSegment::vertex(int to) const {
    DENG2_ASSERT((to? d->to : d->from) != nullptr);
    return to? *d->to : *d->from;
}
AABoxd LineSegment::bounds() const {
    AABoxd bounds;
    Vector2d min = d->from->origin().min(d->to->origin());
    Vector2d max = d->from->origin().max(d->to->origin());
    V2d_Set(bounds.min, min.x, min.y);
    V2d_Set(bounds.max, max.x, max.y);
    return bounds;
}
void LineSegment::replaceVertex(int to, Vertex &newVertex) {
    d->replaceVertex(to, newVertex);
}
}  // namespace bsp
}  // namespace world
namespace world {
using namespace bsp;
typedef QList<Line *>              Lines;
typedef QList<LineSegment *>       LineSegments;
typedef QList<LineSegmentSide *>   LineSegmentSides;
typedef QList<ConvexSubspaceProxy> SubspaceProxys;
typedef QHash<Vertex *, EdgeTips>  EdgeTipSetMap;
DENG2_PIMPL(Partitioner) {
    int splitCostFactor = 7;     ///< Cost of splitting a line segment.
    Lines lines;                 ///< Set of map lines to build from (in index order, not owned).
    Mesh *mesh = nullptr;        ///< Provider of map geometries (cf. Factory).
    int segmentCount = 0;        ///< Running total of segments built.
    int vertexCount  = 0;        ///< Running total of vertexes built.
    LineSegments lineSegments;   ///< Line segments in the plane.
    SubspaceProxys subspaces;    ///< Proxy subspaces in the plane.
    EdgeTipSetMap edgeTipSets;   ///< One set for each vertex.
    BspTree *bspRoot = nullptr;  ///< The BSP tree under construction.
    HPlane hplane;               ///< Current space half-plane (partitioner state).
    struct LineSegmentBlockTree {
        LineSegmentBlockTreeNode *rootNode;
        LineSegmentBlockTree(AABox const &bounds)
            : rootNode(new LineSegmentBlockTreeNode(new LineSegmentBlock(bounds)))
        {}
        ~LineSegmentBlockTree() {
            clear();
        }
        inline operator LineSegmentBlockTreeNode /*const*/ &()
        {
            return *rootNode;
        }
    private:
        void clear() {
            delete rootNode;
            rootNode = nullptr;
        }
    };
    Impl(Public *i) : Base(i) {}
    ~Impl() { clear(); }
    static int clearBspElementWorker(BspTree &subtree, void *) {
        delete subtree.userData();
        subtree.setUserData(nullptr);
        return 0; // Continue iteration.
    }
    void clearBspTree() {
        if(!bspRoot) return;
        bspRoot->traversePostOrder(clearBspElementWorker);
        delete bspRoot; bspRoot = nullptr;
    }
    void clear() {
        lines.clear();
        mesh = nullptr;
        qDeleteAll(lineSegments);
        lineSegments.clear();
        subspaces.clear();
        edgeTipSets.clear();
        hplane.clearIntercepts();
        segmentCount = vertexCount = 0;
    }
    /**
     * Returns a newly allocated Vertex at the given map space @a origin from the
     * map geometry mesh (ownership is @em not given to the caller).
     */
    Vertex *makeVertex(Vector2d const &origin) {
        Vertex *vtx = mesh->newVertex(origin);
        vertexCount += 1; // We built another one.
        return vtx;
    }
    /**
     * @return The new line segment (front is from @a start to @a end).
     */
    LineSegment *makeLineSegment(Vertex &start, Vertex &end, Sector *frontSec,
        Sector *backSec, LineSide *frontSide, Line *partitionLine = nullptr) {
        LineSegment *newSeg = new LineSegment(start, end);
        lineSegments << newSeg;
        LineSegmentSide &front = newSeg->front();
        front.setMapSide(frontSide);
        front.setPartitionMapLine(partitionLine);
        front.setSector(frontSec);
        LineSegmentSide &back = newSeg->back();
        back.setMapSide(frontSide? &frontSide->back() : nullptr);
        back.setPartitionMapLine(partitionLine);
        back.setSector(backSec);
        return newSeg;
    }
    /**
     * Link @a seg into the line segment block tree.
     *
     * Performs k-d tree subdivision of the 2D coordinate space, splitting the node
     * tree as necessary, however new nodes are created only when they need to be
     * populated (i.e., a split does not generate two nodes at the same time).
     */
    void linkLineSegmentInBlockTree(LineSegmentBlockTreeNode &node_, LineSegmentSide &seg) {
        for(LineSegmentBlockTreeNode *node = &node_; ;) {
            LineSegmentBlock &block = *node->userData();
            AABox const &bounds     = block.bounds();
            block.addRef(seg);
            Vector2i dimensions(Vector2i(bounds.max) - Vector2i(bounds.min));
            if(dimensions.x <= 256 && dimensions.y <= 256) {
                block.link(seg);
                seg.setBlockTreeNode(node);
                return;
            }
            int const splitAxis = (dimensions.x < dimensions.y); // x=0, y=1
            int const midOnAxis = (bounds.min[splitAxis] + bounds.max[splitAxis]) / 2;
            LineSegmentBlockTreeNode::ChildId fromSide = LineSegmentBlockTreeNode::ChildId(seg.from().origin()[splitAxis] >= midOnAxis);
            LineSegmentBlockTreeNode::ChildId toSide   = LineSegmentBlockTreeNode::ChildId(seg.to  ().origin()[splitAxis] >= midOnAxis);
            if(fromSide != toSide) {
                block.link(seg);
                seg.setBlockTreeNode(node);
                return;
            }
            if(!node->hasChild(fromSide)) {
                bool const toLeft = (fromSide == LineSegmentBlockTreeNode::Left);
                AABox childBounds;
                if(splitAxis) {
                    int division = bounds.minY + 0.5 + (bounds.maxY - bounds.minY) / 2;
                    childBounds.minX = bounds.minX;
                    childBounds.minY = (toLeft? division : bounds.minY);
                    childBounds.maxX = bounds.maxX;
                    childBounds.maxY = (toLeft? bounds.maxY : division);
                }
                else {
                    int division = bounds.minX + 0.5 + (bounds.maxX - bounds.minX) / 2;
                    childBounds.minX = (toLeft? division : bounds.minX);
                    childBounds.minY = bounds.minY;
                    childBounds.maxX = (toLeft? bounds.maxX : division);
                    childBounds.maxY = bounds.maxY;
                }
                LineSegmentBlock *childBlock = new LineSegmentBlock(childBounds);
                node->setChild(fromSide, new LineSegmentBlockTreeNode(childBlock, node));
            }
            node = node->childPtr(fromSide);
        }
    }
    /**
     * Returns the EdgeTips set associated with @a vertex.
     */
    EdgeTips &edgeTipSet(Vertex const &vertex) {
        EdgeTipSetMap::iterator found = edgeTipSets.find(const_cast<Vertex *>(&vertex));
        if(found == edgeTipSets.end()) {
            found = edgeTipSets.insert(const_cast<Vertex *>(&vertex), EdgeTips());
        }
        return found.value();
    }
    /**
     * Create all initial line segments. We can be certain there are no zero-length
     * lines as these are screened earlier.
     */
    void createInitialLineSegments(LineSegmentBlockTreeNode &rootNode) {
        for(Line *line : lines) {
            Sector *frontSec = line->front().sectorPtr();
            Sector *backSec  = line->back().sectorPtr();
            if(!backSec && line->_bspWindowSector) {
                backSec = line->_bspWindowSector;
            }
            LineSegment *seg = makeLineSegment(line->from(), line->to(),
                                               frontSec, backSec, &line->front());
            if(seg->front().hasSector()) {
                linkLineSegmentInBlockTree(rootNode, seg->front());
            }
            if(seg->back().hasSector()) {
                linkLineSegmentInBlockTree(rootNode, seg->back());
            }
            edgeTipSet(line->from()) << EdgeTip(seg->front());
            edgeTipSet(line->to())   << EdgeTip(seg->back());
        }
    }
    /**
     * Splits the given line segment at the point (x,y). The new line segment
     * is returned. The old line segment is shortened (the original start vertex
     * is unchanged), the new line segment becomes the cut-off tail (keeping
     * the original end vertex).
     *
     * @note If the line segment has a twin it is also split.
     */
    LineSegmentSide &splitLineSegment(LineSegmentSide &frontLeft,
        Vector2d const &point, bool updateEdgeTips = true) {
        DENG2_ASSERT(point != frontLeft.from().origin() &&
                     point != frontLeft.to().origin());
        Vertex *newVert = makeVertex(point);
        LineSegment &oldSeg = frontLeft.line();
        LineSegment &newSeg = *makeLineSegment(oldSeg.from(), oldSeg.to(),
                                               oldSeg.front().sectorPtr(),
                                               oldSeg.back().sectorPtr(),
                                               oldSeg.front().mapSidePtr(),
                                               oldSeg.front().partitionMapLine());
        LineSegmentSide &frontRight = newSeg.side(frontLeft.lineSideId());
        oldSeg.replaceVertex(frontLeft.lineSideId() ^ LineSegment::To, *newVert);
        newSeg.replaceVertex(frontLeft.lineSideId(),                   *newVert);
        LineSegmentSide &backRight = frontLeft.back();
        LineSegmentSide &backLeft  = frontRight.back();
        if(ConvexSubspaceProxy *convexSet = frontLeft.convexSubspace()) {
            *convexSet << frontRight;
            frontRight.setConvexSubspace(convexSet);
        }
        frontLeft.setRight(&frontRight);
        frontRight.setLeft(&frontLeft);
        if(ConvexSubspaceProxy *convexSet = backRight.convexSubspace()) {
            *convexSet << backLeft;
            backLeft.setConvexSubspace(convexSet);
        }
        backLeft.setRight(&backRight);
        backRight.setLeft(&backLeft);
        if(updateEdgeTips) {
            edgeTipSet(oldSeg.from()).clearByLineSegment(oldSeg);
            edgeTipSet(oldSeg.to()  ).clearByLineSegment(oldSeg);
            edgeTipSet(newSeg.from()).clearByLineSegment(newSeg);
            edgeTipSet(newSeg.to()  ).clearByLineSegment(newSeg);
            edgeTipSet(oldSeg.from()) << EdgeTip(oldSeg.front());
            edgeTipSet(oldSeg.to())   << EdgeTip(oldSeg.back());
            edgeTipSet(newSeg.from()) << EdgeTip(newSeg.front());
            edgeTipSet(newSeg.to())   << EdgeTip(newSeg.back());
        }
        return frontRight;
    }
    /**
     * Find the intersection point between a line segment and the current
     * partition plane. Takes advantage of some common situations like
     * horizontal and vertical lines to choose a 'nicer' intersection point.
     */
    Vector2d intersectPartition(LineSegmentSide const &seg, coord_t fromDist,
                                coord_t toDist) const {
        if(hplane.slopeType() == ST_HORIZONTAL && seg.slopeType() == ST_VERTICAL) {
            return Vector2d(seg.from().origin().x, hplane.partition().origin.y);
        }
        if(hplane.slopeType() == ST_VERTICAL && seg.slopeType() == ST_HORIZONTAL) {
            return Vector2d(hplane.partition().origin.x, seg.from().origin().y);
        }
        coord_t ds = fromDist / (fromDist - toDist);
        Vector2d point = seg.from().origin();
        if(seg.slopeType() != ST_VERTICAL)
            point.x += seg.direction().x * ds;
        if(seg.slopeType() != ST_HORIZONTAL)
            point.y += seg.direction().y * ds;
        return point;
    }
    inline void interceptPartition(LineSegmentSide &seg, int edge) {
        hplane.intercept(seg, edge, edgeTipSet(seg.vertex(edge)));
    }
    /**
     * Take the given line segment @a lineSeg, compare it with the partition
     * plane and determine into which of the two sets it should be. If the
     * line segment is found to intersect the partition, the intercept point
     * is determined and the line segment then split in two at this point.
     * Each piece of the line segment is then added to the relevant set.
     *
     * If the line segment is collinear with, or intersects the partition then
     * a new intercept is added to the partitioning half-plane.
     *
     * @note Any existing @em twin of @a lineSeg is so too handled uniformly.
     *
     * @param seg     Line segment to be "partitioned".
     * @param rights  Set of line segments on the right side of the partition.
     * @param lefts   Set of line segments on the left side of the partition.
     */
    void divideOneSegment(LineSegmentSide &seg, LineSegmentBlockTreeNode &rights,
                          LineSegmentBlockTreeNode &lefts) {
        coord_t fromDist, toDist;
        LineRelationship rel = hplane.relationship(seg, &fromDist, &toDist);
        switch(rel) {
        case Collinear: {
            interceptPartition(seg, LineSegment::From);
            interceptPartition(seg, LineSegment::To);
            if(seg.direction().dot(hplane.partition().direction) < 0) {
                linkLineSegmentInBlockTree(lefts, seg);
            }
            else {
                linkLineSegmentInBlockTree(rights, seg);
            }
            break; }
        case Right:
        case RightIntercept:
            if(rel == RightIntercept) {
                interceptPartition(seg, (fromDist < DIST_EPSILON? LineSegment::From : LineSegment::To));
            }
            linkLineSegmentInBlockTree(rights, seg);
            break;
        case Left:
        case LeftIntercept:
            if(rel == LeftIntercept) {
                interceptPartition(seg, (fromDist > -DIST_EPSILON? LineSegment::From : LineSegment::To));
            }
            linkLineSegmentInBlockTree(lefts, seg);
            break;
        case Intersects: {
            Vector2d point = intersectPartition(seg, fromDist, toDist);
            LineSegmentSide &newFrontRight = splitLineSegment(seg, point);
            if(LineSegmentBlockTreeNode *backLeftBlock = (LineSegmentBlockTreeNode *)seg.back().blockTreeNodePtr()) {
                linkLineSegmentInBlockTree(*backLeftBlock, newFrontRight.back());
            }
            interceptPartition(seg, LineSegment::To);
            if(fromDist < 0) {
                linkLineSegmentInBlockTree(rights, newFrontRight);
                linkLineSegmentInBlockTree(lefts,  seg);
            }
            else {
                linkLineSegmentInBlockTree(rights, seg);
                linkLineSegmentInBlockTree(lefts,  newFrontRight);
            }
            break; }
        }
    }
    /**
     * Remove all the line segments from the list, partitioning them into the
     * left or right sets according to their position relative to partition line.
     * Adds any intersections onto the intersection list as it goes.
     *
     * @param node    Block tree node containing the line segments to be partitioned.
     * @param rights  Set of line segments on the right side of the partition.
     * @param lefts   Set of line segments on the left side of the partition.
     */
    void divideSegments(LineSegmentBlockTreeNode &node, LineSegmentBlockTreeNode &rights,
                        LineSegmentBlockTreeNode &lefts) {
        /**
         * @todo Revise this algorithm so that @var segments is not modified
         * during the partitioning process.
         */
        int const totalSegs = node.userData()->totalCount();
        DENG2_ASSERT(totalSegs != 0);
        DENG2_UNUSED(totalSegs);
        LineSegmentBlockTreeNode *cur  = &node;
        LineSegmentBlockTreeNode *prev = nullptr;
        while(cur) {
            while(cur) {
                LineSegmentBlock &segs = *cur->userData();
                LineSegmentSide *seg;
                while((seg = segs.pop())) {
                    seg->setBlockTreeNode(nullptr);
                    divideOneSegment(*seg, rights, lefts);
                }
                if(prev == cur->parentPtr()) {
                    prev = cur;
                    if(cur->hasRight()) cur = cur->rightPtr();
                    else                cur = cur->leftPtr();
                }
                else if(prev == cur->rightPtr()) {
                    prev = cur;
                    cur = cur->leftPtr();
                }
                else if(prev == cur->leftPtr()) {
                    prev = cur;
                    cur = cur->parentPtr();
                }
            }
            if(prev) {
                cur = prev->parentPtr();
            }
        }
        DENG2_ASSERT(rights.userData()->totalCount());
        DENG2_ASSERT(lefts.userData ()->totalCount());
        DENG2_ASSERT((  rights.userData()->totalCount()
                      + lefts.userData ()->totalCount()) >= totalSegs);
    }
    /**
     * Analyze the half-plane intercepts, building new line segments to cap
     * any gaps (new segments are added onto the end of the appropriate list
     * (rights to @a rightList and lefts to @a leftList)).
     *
     * @param rights  Set of line segments on the right of the partition.
     * @param lefts   Set of line segments on the left of the partition.
     */
    void addPartitionLineSegments(LineSegmentBlockTreeNode &rights,
                                  LineSegmentBlockTreeNode &lefts) {
        LOG_TRACE("Building line segments along partition %s",
                  hplane.partition().asText());
        hplane.sortAndMergeIntercepts();
        LineSegmentSide *partSeg = hplane.lineSegment();
        double nearDist = 0, farDist = 0;
        if(partSeg) {
            nearDist = hplane.intersect(*partSeg, LineSegment::From);
            farDist  = hplane.intersect(*partSeg, LineSegment::To);
        }
        for(int i = 0; i < hplane.interceptCount() - 1; ++i) {
            HPlaneIntercept const &cur  = hplane.intercepts()[i];
            HPlaneIntercept const &next = hplane.intercepts()[i+1];
            if(partSeg && cur.distance() >= nearDist && next.distance() <= farDist)
                continue;
            if(!cur.after() && !next.before())
                continue;
            if(cur.after() && !next.before()) {
                if(!cur.lineSegmentIsSelfReferencing()) {
                    Vector2d nearPoint = (cur.vertex().origin() + next.vertex().origin()) / 2;
                    notifyUnclosedSectorFound(*cur.after(), nearPoint);
                }
                continue;
            }
            if(!cur.after() && next.before()) {
                if(!next.lineSegmentIsSelfReferencing()) {
                    Vector2d nearPoint = (cur.vertex().origin() + next.vertex().origin()) / 2;
                    notifyUnclosedSectorFound(*next.before(), nearPoint);
                }
                continue;
            }
            /*
             * This is definitely open space.
             */
            Vertex &fromVertex = cur.vertex();
            Vertex &toVertex   = next.vertex();
            Sector *sector = cur.after();
            if(!cur.before() && next.before() == next.after()) {
                sector = next.before();
            }
            else if(cur.after() != next.before()) {
                if(!cur.lineSegmentIsSelfReferencing() &&
                   !next.lineSegmentIsSelfReferencing()) {
                    LOG_DEBUG("Sector mismatch #%d %s != #%d %s")
                            << cur.after()->indexInMap()
                            << cur.vertex().origin().asText()
                            << next.before()->indexInMap()
                            << next.vertex().origin().asText();
                }
                LineSegmentSide *afterSeg = cur.afterLineSegment();
                if(afterSeg->hasMapLine() && afterSeg->mapLine().isSelfReferencing()) {
                    LineSegmentSide *beforeSeg = next.beforeLineSegment();
                    if(beforeSeg->hasMapLine() && !beforeSeg->mapLine().isSelfReferencing()) {
                        sector = next.before();
                    }
                }
            }
            DENG2_ASSERT(sector);
            LineSegment &newSeg = *makeLineSegment(fromVertex, toVertex,
                                                   sector, sector, nullptr /*no map line*/,
                                                   partSeg? &partSeg->mapLine() : nullptr);
            edgeTipSet(newSeg.from()) << EdgeTip(newSeg.front());
            edgeTipSet(newSeg.to())   << EdgeTip(newSeg.back());
            linkLineSegmentInBlockTree(rights, newSeg.front());
            linkLineSegmentInBlockTree(lefts,  newSeg.back());
            /*
            LOG_DEBUG("Built line segment from %s to %s (sector #%i)")
                    << fromVertex.origin().asText()
                    << toVertex.origin().asText()
                    << sector->indexInArchive();
            */
        }
    }
    /**
     * Collate (unlink) all line segments at or beneath @a node to a new list.
     */
    static LineSegmentSides collectAllSegments(LineSegmentBlockTreeNode &node) {
        LineSegmentSides allSegs;
#ifdef DENG2_QT_4_7_OR_NEWER
        allSegs.reserve(node.userData()->totalCount());
#endif
        LineSegmentBlockTreeNode *cur  = &node;
        LineSegmentBlockTreeNode *prev = nullptr;
        while(cur) {
            while(cur) {
                LineSegmentBlock &segs = *cur->userData();
                LineSegmentSide *seg;
                while((seg = segs.pop())) {
                    allSegs << seg;
                }
                if(prev == cur->parentPtr()) {
                    prev = cur;
                    if(cur->hasRight()) cur = cur->rightPtr();
                    else                cur = cur->leftPtr();
                }
                else if(prev == cur->rightPtr()) {
                    prev = cur;
                    cur = cur->leftPtr();
                }
                else if(prev == cur->leftPtr()) {
                    prev = cur;
                    cur = cur->parentPtr();
                }
            }
            if(prev) {
                cur = prev->parentPtr();
            }
        }
        return allSegs;
    }
    /**
     * Determine the axis-aligned bounding box containing the vertex coordinates
     * from all segments.
     */
    static AABoxd segmentBounds(LineSegmentSides const &allSegments) {
        bool initialized = false;
        AABoxd bounds;
        for(LineSegmentSide *seg : allSegments) {
            if(initialized) {
                V2d_UniteBox(bounds.arvec2, seg->bounds().arvec2);
            }
            else {
                V2d_CopyBox(bounds.arvec2, seg->bounds().arvec2);
                initialized = true;
            }
        }
        return bounds;
    }
    /**
     * Determine the axis-aligned bounding box containing the vertices of all
     * line segments at or beneath @a node in the block tree.
     *
     * @return  Determined AABox (might be empty (i.e., min > max) if no segments).
     */
    static AABoxd segmentBounds(LineSegmentBlockTreeNode const &node) {
        bool initialized = false;
        AABoxd bounds;
        LineSegmentBlockTreeNode const *cur  = &node;
        LineSegmentBlockTreeNode const *prev = nullptr;
        while(cur) {
            while(cur) {
                LineSegmentBlock const &block = *cur->userData();
                if(block.totalCount()) {
                    AABoxd boundsAtNode = segmentBounds(block.all());
                    if(initialized) {
                        V2d_AddToBox(bounds.arvec2, boundsAtNode.min);
                    }
                    else {
                        V2d_InitBox(bounds.arvec2, boundsAtNode.min);
                        initialized = true;
                    }
                    V2d_AddToBox(bounds.arvec2, boundsAtNode.max);
                }
                if(prev == cur->parentPtr()) {
                    prev = cur;
                    if(cur->hasRight()) cur = cur->rightPtr();
                    else                cur = cur->leftPtr();
                }
                else if(prev == cur->rightPtr()) {
                    prev = cur;
                    cur = cur->leftPtr();
                }
                else if(prev == cur->leftPtr()) {
                    prev = cur;
                    cur = cur->parentPtr();
                }
            }
            if(prev) {
                cur = prev->parentPtr();
            }
        }
        return bounds;
    }
    LineSegmentSide *choosePartition(LineSegmentBlockTreeNode &candidateSet) {
        return PartitionEvaluator(splitCostFactor).choose(candidateSet);
    }
    /**
     * Takes the line segment list and determines if it is convex, possibly
     * converting it into a BSP leaf. Otherwise, the list is divided into two
     * halves and recursion will continue on the new sub list.
     *
     * This is done by scanning all of the line segments and finding the one
     * that does the least splitting and has the least difference in numbers
     * of line segments on either side (why is this valued? -ds).
     *
     * If the line segments on the left side are convex create another leaf
     * else put the line segments into the left list.
     *
     * If the line segments on the right side are convex create another leaf
     * else put the line segments into the right list.
     *
     * @param node  Tree node for the block containing the line segments to
     *              be partitioned.
     *
     * @return  Newly created BSP subtree; otherwise @c nullptr (degenerate).
     */
    BspTree *partitionSpace(LineSegmentBlockTreeNode &node) {
        LOG_AS("Partitioner::partitionSpace");
        BspElement *bspElement = nullptr; ///< Built BSP map element at this node.
        BspTree *rightBspTree  = nullptr;
        BspTree *leftBspTree   = nullptr;
        if(LineSegmentSide *partSeg = choosePartition(node)) {
            hplane.configure(*partSeg);
            /*
            LOG_TRACE("%s, segment side %p %i (segment #%i) %s %s")
                    << hplane.partition().asText()
                    << partSeg
                    << partSeg->lineSideId()
                    << lineSegments.indexOf(&partSeg->line())
                    << partSeg->from().origin().asText()
                    << partSeg->to().origin().asText();
            */
            Partition partition(hplane.partition());
            LineSegmentBlockTree rightTree(node.userData()->bounds());
            LineSegmentBlockTree leftTree(node.userData()->bounds());
            divideSegments(node, rightTree, leftTree);
            node.clear();
            addPartitionLineSegments(rightTree, leftTree);
            rightBspTree = partitionSpace(rightTree);
            leftBspTree  = partitionSpace(leftTree);
            if(!rightBspTree || !leftBspTree)
                return rightBspTree? rightBspTree : leftBspTree;
            bspElement = new BspNode(partition);
        }
        else {
            LineSegmentSides segments = collectAllSegments(node);
            node.clear();
            subspaces.append(ConvexSubspaceProxy());
            ConvexSubspaceProxy &convexSet = subspaces.last();
            convexSet.addSegments(segments);
            for(LineSegmentSide *seg : segments) {
                seg->setConvexSubspace(&convexSet);
                seg->setBlockTreeNode(nullptr);
            }
            auto *leaf = new BspLeaf;
            convexSet.setBspLeaf(leaf);
            bspElement = leaf;
        }
        auto *subtree = new BspTree(bspElement, nullptr/*no parent*/, rightBspTree, leftBspTree);
        if(rightBspTree) rightBspTree->setParent(subtree);
        if(leftBspTree)  leftBspTree->setParent(subtree);
        return subtree;
    }
    /**
     * Split any overlapping line segments in the convex subspaces, creating new
     * line segments (and vertices) as required. A subspace may well include such
     * overlapping segments as if they do not break the convexity rule they won't
     * have been split during the partitioning process.
     *
     * @todo Perform the split in divideSpace()
     */
    void splitOverlappingSegments() {
        for(ConvexSubspaceProxy const &subspace : subspaces) {
            /*
             * The subspace provides a specially ordered list of the segments to
             * simplify this task. The primary clockwise ordering (decreasing angle
             * relative to the center of the subspace) places overlapping segments
             * adjacently. The secondary anticlockwise ordering sorts the overlapping
             * segments enabling the use of single pass algorithm here.
             */
            OrderedSegments convexSet = subspace.segments();
            int const numSegments = convexSet.count();
            for(int i = 0; i < numSegments - 1; ++i) {
                int k = i;
                while(de::fequal(convexSet[k + 1].fromAngle, convexSet[i].fromAngle) &&
                      ++k < numSegments - 1)
                {}
                for(int l = i; l < k; ++l) {
                    OrderedSegment &a = convexSet[l];
                    for(int m = l + 1; m <= k; ++m) {
                        OrderedSegment &b = convexSet[m];
                        if(de::fequal(b.segment->length(), a.segment->length()))
                            continue;
                        Vector2d const &point = b.segment->to().origin();
                        if(point == a.segment->from().origin() ||
                           point == a.segment->to().origin())
                            continue;
                        splitLineSegment(*a.segment, point, false /*don't update edge tips*/);
                    }
                }
                i = k;
            }
        }
    }
    void buildSubspaceGeometries() {
        for(ConvexSubspaceProxy const &subspace : subspaces) {
            BspLeaf &bspLeaf = *subspace.bspLeaf();
            subspace.buildGeometry(bspLeaf, *mesh);
            for(OrderedSegment const &oseg : subspace.segments()) {
                if(oseg.segment->hasHEdge()) {
                    segmentCount += 1;
                }
            }
        }
        /*
         * Finalize the built geometry by adding a twin half-edge for any
         * which don't yet have one.
         */
        for(ConvexSubspaceProxy const &convexSet : subspaces)
        for(OrderedSegment const &oseg : convexSet.segments()) {
            LineSegmentSide *seg = oseg.segment;
            if(seg->hasHEdge() && !seg->back().hasHEdge()) {
                HEdge *hedge = &seg->hedge();
                DENG2_ASSERT(!hedge->hasTwin());
                hedge->setTwin(hedge->mesh().newHEdge(seg->back().from()));
                hedge->twin().setTwin(hedge);
            }
        }
    }
    /**
     * Notify interested parties of an unclosed sector in the map.
     *
     * @param sector    The problem sector.
     * @param nearPoint  Coordinates near to where the problem was found.
     */
    void notifyUnclosedSectorFound(Sector &sector, Vector2d const &nearPoint) {
        DENG2_FOR_PUBLIC_AUDIENCE(UnclosedSectorFound, i) {
            i->unclosedSectorFound(sector, nearPoint);
        }
    }
#ifdef DENG2_DEBUG
    void printSegments(LineSegmentSides const &allSegs) {
        for(LineSegmentSide const *seg : allSegs) {
            LOG_DEBUG("Build: %s line segment %p sector: %d %s -> %s")
                    << (seg->hasMapSide()? "map" : "part")
                    << seg
                    << (seg->hasSector()? seg->sector().indexInMap() : -1)
                    << seg->from().origin().asText() << seg->to().origin().asText();
        }
    }
#endif
};
Partitioner::Partitioner(int splitCostFactor) : d(new Impl(this)) {
    setSplitCostFactor(splitCostFactor);
}
void Partitioner::setSplitCostFactor(int newFactor) {
    d->splitCostFactor = newFactor;
}
static AABox blockmapBounds(AABoxd const &mapBounds) {
    AABox mapBoundsi;
    mapBoundsi.minX = int( de::floor(mapBounds.minX) );
    mapBoundsi.minY = int( de::floor(mapBounds.minY) );
    mapBoundsi.maxX = int(  de::ceil(mapBounds.maxX) );
    mapBoundsi.maxY = int(  de::ceil(mapBounds.maxY) );
    AABox blockBounds;
    blockBounds.minX = mapBoundsi.minX - (mapBoundsi.minX & 0x7);
    blockBounds.minY = mapBoundsi.minY - (mapBoundsi.minY & 0x7);
    int bw = ((mapBoundsi.maxX - blockBounds.minX) / 128) + 1;
    int bh = ((mapBoundsi.maxY - blockBounds.minY) / 128) + 1;
    blockBounds.maxX = blockBounds.minX + 128 * M_CeilPow2(bw);
    blockBounds.maxY = blockBounds.minY + 128 * M_CeilPow2(bh);
    return blockBounds;
}
static bool lineIndexLessThan(Line const *a, Line const *b) {
     return a->indexInMap() < b->indexInMap();
}
BspTree *Partitioner::makeBspTree(QSet<Line *> const &lines, Mesh &mesh) {
    d->clear();
    d->lines = lines.toList();
    qSort(d->lines.begin(), d->lines.end(), lineIndexLessThan);
    d->mesh = &mesh;
    d->edgeTipSets.reserve(d->lines.count() * 2);
    AABoxd bounds;
    bool isFirst = true;
    for(Line *line : d->lines) {
        if(isFirst) {
            V2d_CopyBox(bounds.arvec2, line->bounds().arvec2);
            isFirst = false;
        }
        else {
            V2d_UniteBox(bounds.arvec2, line->bounds().arvec2);
        }
    }
    Impl::LineSegmentBlockTree blockTree(blockmapBounds(bounds));
    d->createInitialLineSegments(blockTree);
    d->bspRoot = d->partitionSpace(blockTree);
    d->splitOverlappingSegments();
    d->buildSubspaceGeometries();
    return d->bspRoot;
}
int Partitioner::segmentCount() {
    return d->segmentCount;
}
int Partitioner::vertexCount() {
    return d->vertexCount;
}
}  // namespace world
namespace world {
namespace bsp {
namespace internal {
    struct PartitionCost {
        int total     = 0;
        int splits    = 0;
        int iffy      = 0;
        int nearMiss  = 0;
        int mapRight  = 0;
        int mapLeft   = 0;
        int partRight = 0;
        int partLeft  = 0;
        inline PartitionCost &addSegmentRight(LineSegmentSide const &seg) {
            if(seg.hasMapSide()) mapRight  += 1;
            else                 partRight += 1;
            return *this;
        }
        inline PartitionCost &addSegmentLeft(LineSegmentSide const &seg) {
            if(seg.hasMapSide()) mapLeft  += 1;
            else                 partLeft += 1;
            return *this;
        }
        PartitionCost &operator += (PartitionCost const &other) {
            total     += other.total;
            splits    += other.splits;
            iffy      += other.iffy;
            nearMiss  += other.nearMiss;
            mapLeft   += other.mapLeft;
            mapRight  += other.mapRight;
            partLeft  += other.partLeft;
            partRight += other.partRight;
            return *this;
        }
        PartitionCost &operator = (PartitionCost const &other) {
            total     = other.total;
            splits    = other.splits;
            iffy      = other.iffy;
            nearMiss  = other.nearMiss;
            mapLeft   = other.mapLeft;
            mapRight  = other.mapRight;
            partLeft  = other.partLeft;
            partRight = other.partRight;
            return *this;
        }
        bool operator < (PartitionCost const &rhs) const {
            return total < rhs.total;
        }
        String asText() const {
            return String("PartitionCost(Total= %1.%2; splits:%3, iffy:%4, near:%5, left:%6+%7, right:%8+%9)")
                       .arg(total / 100).arg(total % 100, 2, QChar('0'))
                       .arg(splits).arg(iffy).arg(nearMiss)
                       .arg(mapLeft).arg(partLeft)
                       .arg(mapRight).arg(partRight);
        }
    };
    /**
     * "Near miss" predicate.
     */
    static bool nearMiss(LineRelationship rel, coord_t fromDist, coord_t toDist,
                         coord_t *distance) {
        if(rel == Right &&
           !((fromDist >= SHORT_HEDGE_EPSILON && toDist >= SHORT_HEDGE_EPSILON) ||
             (fromDist <= DIST_EPSILON        && toDist >= SHORT_HEDGE_EPSILON) ||
             (toDist <= DIST_EPSILON && fromDist >= SHORT_HEDGE_EPSILON))) {
            if(distance) {
                if(fromDist <= DIST_EPSILON || toDist <= DIST_EPSILON) {
                    *distance = SHORT_HEDGE_EPSILON / de::max(fromDist, toDist);
                }
                else {
                    *distance = SHORT_HEDGE_EPSILON / de::min(fromDist, toDist);
                }
            }
            return true;
        }
        if(rel == Left &&
           !((fromDist <= -SHORT_HEDGE_EPSILON && toDist <= -SHORT_HEDGE_EPSILON) ||
             (fromDist >= -DIST_EPSILON        && toDist <= -SHORT_HEDGE_EPSILON) ||
             (toDist >= -DIST_EPSILON && fromDist <= -SHORT_HEDGE_EPSILON))) {
            if(distance) {
                if(fromDist >= -DIST_EPSILON || toDist >= -DIST_EPSILON) {
                    *distance = SHORT_HEDGE_EPSILON / -de::min(fromDist, toDist);
                }
                else {
                    *distance = SHORT_HEDGE_EPSILON / -de::max(fromDist, toDist);
                }
            }
            return true;
        }
        return false;
    }
    /**
     * "Near edge" predicate. Assumes intersecting line segment relationship.
     */
    static bool nearEdge(coord_t fromDist, coord_t toDist, coord_t *distance) {
        if(de::abs(fromDist) < SHORT_HEDGE_EPSILON || de::abs(toDist) < SHORT_HEDGE_EPSILON) {
            if(distance) {
                *distance = SHORT_HEDGE_EPSILON / de::min(de::abs(fromDist), de::abs(toDist));
            }
            return true;
        }
        return false;
    }
}
using namespace internal;
DENG2_PIMPL_NOREF(PartitionEvaluator) {
    int splitCostFactor = 7;
    LineSegmentBlockTreeNode *rootNode = nullptr; ///< Current block tree root node.
    struct PartitionCandidate {
        LineSegmentSide *line;  ///< Candidate partition line.
        PartitionCost cost;     ///< Running cost metric total.
        PartitionCandidate(LineSegmentSide &partition) : line(&partition)
        {}
    };
    typedef QList<PartitionCandidate *> Candidates;
    Candidates candidates;
    PartitionCandidate *nextCandidate() {
        DENG2_ASSERT(costTaskPool.isDone());
        if(candidates.isEmpty()) return nullptr;
        return candidates.takeFirst();
    }
    class CostTask : public Task {
    public:
        Impl &evaluator;
        PartitionCandidate &candidate;
        CostTask(Impl &evaluator, PartitionCandidate &candidate)
            : evaluator(evaluator), candidate(candidate)
        {}
        /**
         * Evaluate the cost of the partition candidate.
         *
         * If the candidate is not suitable (or a better choice has already been
         * determined) then @var partition is zeroed. Otherwise the candidate is
         * suitable and @var cost contains valid costing metrics.
         */
        void runTask() {
            LineSegmentSide **partition = &candidate.line;
            PartitionCost &cost         = candidate.cost;
            costForBlock(*evaluator.rootNode);
            if(!cost.mapLeft || !cost.mapRight) {
                *partition = nullptr;
                return;
            }
            cost.total += 100 * de::abs(cost.mapLeft - cost.mapRight);
            cost.total += 50 * de::abs(cost.partLeft - cost.partRight);
            if((*partition)->slopeType() != ST_HORIZONTAL &&
               (*partition)->slopeType() != ST_VERTICAL) {
                cost.total += 25;
            }
        }
    private:
        void costForSegment(LineSegmentSide const &seg) {
            LineSegmentSide **partition = &candidate.line;
            PartitionCost &cost         = candidate.cost;
            int const splitCostFactor   = evaluator.splitCostFactor;
            coord_t fromDist, toDist;
            LineRelationship rel = seg.relationship(**partition, &fromDist, &toDist);
            switch(rel) {
            case Collinear: {
                if(seg.direction().dot((*partition)->direction()) < 0) {
                    cost.addSegmentLeft(seg);
                }
                else {
                    cost.addSegmentRight(seg);
                }
                break; }
            case Right:
            case RightIntercept: {
                cost.addSegmentRight(seg);
                /*
                 * Near misses are bad, as they have the potential to result in
                 * really short line segments being produced later on.
                 *
                 * The closer the near miss, the higher the cost.
                 */
                coord_t nearDist;
                if(nearMiss(rel, fromDist, toDist, &nearDist)) {
                    cost.nearMiss += 1;
                    cost.total += int( 100 * splitCostFactor * (nearDist * nearDist - 1.0) );
                }
                break; }
            case Left:
            case LeftIntercept: {
                cost.addSegmentLeft(seg);
                coord_t nearDist;
                if(nearMiss(rel, fromDist, toDist, &nearDist)) {
                    cost.nearMiss += 1;
                    cost.total += int( 70 * splitCostFactor * (nearDist * nearDist - 1.0) );
                }
                break; }
            case Intersects: {
                cost.splits += 1;
                cost.total  += 100 * splitCostFactor;
                /*
                 * If the split point is very close to one end, which is quite an
                 * undesirable situation (producing really short edges), thus a
                 * rather hefty surcharge.
                 *
                 * The closer to the edge, the higher the cost.
                 */
                coord_t nearDist;
                if(nearEdge(fromDist, toDist, &nearDist)) {
                    cost.iffy += 1;
                    cost.total += int( 140 * splitCostFactor * (nearDist * nearDist - 1.0) );
                }
                break; }
            }
        }
        /**
         * Test the whole block against the partition line to quickly handle all the
         * line segments within it at once. Only when the partition line intercepts
         * the block do we need to go deeper into it.
         */
        void costForBlock(LineSegmentBlockTreeNode const &node) {
            LineSegmentBlock const &block    = *node.userData();
            LineSegmentSide const *partition = candidate.line;
            PartitionCost &cost              = candidate.cost;
            AABoxd bounds(coord_t( block.bounds().minX ) - SHORT_HEDGE_EPSILON * 1.5,
                          coord_t( block.bounds().minY ) - SHORT_HEDGE_EPSILON * 1.5,
                          coord_t( block.bounds().maxX ) + SHORT_HEDGE_EPSILON * 1.5,
                          coord_t( block.bounds().maxY ) + SHORT_HEDGE_EPSILON * 1.5);
            int side = partition->boxOnSide(bounds);
            if(side > 0) {
                cost.mapRight  += block.mapCount();
                cost.partRight += block.partCount();
                return;
            }
            if(side < 0) {
                cost.mapLeft  += block.mapCount();
                cost.partLeft += block.partCount();
                return;
            }
            for(LineSegmentSide *otherSeg : block.all()) {
                costForSegment(*otherSeg);
            }
            if(node.hasRight()) {
                costForBlock(*node.rightPtr());
            }
            if(node.hasLeft()) {
                costForBlock(*node.leftPtr());
            }
        }
    };
    TaskPool costTaskPool;
    /**
     * @param line  Partition line to evaluate.
     */
    void beginPartitionCosting(LineSegmentSide *line) {
        DENG2_ASSERT(line && line->hasMapSide());
        PartitionCandidate *newCandidate = new PartitionCandidate(*line);
        candidates << newCandidate;
        costTaskPool.start(new CostTask(*this, *newCandidate));
    }
};
PartitionEvaluator::PartitionEvaluator(int splitCostFactor) : d(new Impl) {
    d->splitCostFactor = splitCostFactor;
}
LineSegmentSide *PartitionEvaluator::choose(LineSegmentBlockTreeNode &node) {
    LOG_AS("PartitionEvaluator");
    d->rootNode = &node;
    validCount++;
    LineSegmentBlockTreeNode const *cur  = d->rootNode;
    LineSegmentBlockTreeNode const *prev = nullptr;
    while(cur) {
        while(cur) {
            LineSegmentBlock const &segs = *cur->userData();
            for(LineSegmentSide *candidate : segs.all()) {
                if(!candidate->hasMapSide())
                    continue;
                if(candidate->mapLine().validCount() == validCount)
                    continue; // Skip this.
                candidate->mapLine().setValidCount(validCount);
                d->beginPartitionCosting(candidate);
            }
            if(prev == cur->parentPtr()) {
                prev = cur;
                if(cur->hasRight()) cur = cur->rightPtr();
                else                cur = cur->leftPtr();
            }
            else if(prev == cur->rightPtr()) {
                prev = cur;
                cur = cur->leftPtr();
            }
            else if(prev == cur->leftPtr()) {
                prev = cur;
                cur = cur->parentPtr();
            }
        }
        if(prev) {
            cur = prev->parentPtr();
        }
    }
    LineSegmentSide *best = nullptr;
    if(!d->candidates.isEmpty()) {
        d->costTaskPool.waitForDone();
        PartitionCost bestCost;
        while(Impl::PartitionCandidate *candidate = d->nextCandidate()) {
            if(candidate->line && (!best || candidate->cost < bestCost)) {
                best     = candidate->line;
                bestCost = candidate->cost;
            }
            delete candidate;
        }
    }
    return best;
}
}  // namespace bsp
}  // namespace world
namespace world {
namespace bsp {
DENG2_PIMPL_NOREF(LineSegmentBlock) {
    AABox bounds;       ///< Block bounds at the node.
    All segments;       ///< Line segments contained by the node (not owned).
    int mapCount  = 0;  ///< Running total of map-line segments at/under this node.
    int partCount = 0;  ///< Running total of partition-line segments at/under this node.
};
LineSegmentBlock::LineSegmentBlock(AABox const &bounds) : d(new Impl) {
    d->bounds = bounds;
}
AABox const &LineSegmentBlock::bounds() const {
    return d->bounds;
}
void LineSegmentBlock::link(LineSegmentSide &seg) {
    d->segments.prepend(&seg);
}
void LineSegmentBlock::addRef(LineSegmentSide const &seg) {
    if(seg.hasMapSide()) d->mapCount++;
    else                 d->partCount++;
}
void LineSegmentBlock::decRef(LineSegmentSide const &seg) {
    if(seg.hasMapSide()) d->mapCount--;
    else                 d->partCount--;
}
LineSegmentSide *LineSegmentBlock::pop() {
    if(!d->segments.isEmpty()) {
        LineSegmentSide *seg = d->segments.takeFirst();
        decRef(*seg);
        return seg;
    }
    return 0;
}
int LineSegmentBlock::mapCount() const {
    return d->mapCount;
}
int LineSegmentBlock::partCount() const {
    return d->partCount;
}
int LineSegmentBlock::totalCount() const {
    return d->mapCount + d->partCount;
}
LineSegmentBlock::All const &LineSegmentBlock::all() const {
    return d->segments;
}
LineSegmentBlockTreeNode::LineSegmentBlockTreeNode(LineSegmentBlock *lsb,
                                                   LineSegmentBlockTreeNode *parent)
    : BinaryTree<LineSegmentBlock *>(lsb, parent)
{}
LineSegmentBlockTreeNode::~LineSegmentBlockTreeNode() {
    delete userData();
}
LineSegmentBlockTreeNode *LineSegmentBlockTreeNode::childPtr(ChildId side) const {
    return static_cast<LineSegmentBlockTreeNode *>
            (BinaryTree<LineSegmentBlock *>::childPtr(side));
}
LineSegmentBlockTreeNode *LineSegmentBlockTreeNode::parentPtr() const {
    return static_cast<LineSegmentBlockTreeNode *>
            (BinaryTree<LineSegmentBlock *>::parentPtr());
}
}  // namespace bsp
}  // namespace world
