﻿/** @file
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/console/alias.h"
#include "doomsday/console/knownword.h"
#include <de/memory.h>
/** @file
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/console/cmd.h"
#include "doomsday/console/alias.h"
#include "doomsday/console/knownword.h"
#include "doomsday/help.h"
#include <de/memoryblockset.h>
#include <de/memoryzone.h>
#include <de/c_wrapper.h>
#include <de/App>
#include <de/Log>
#include <de/Config>
#include <de/TextValue>
#include <QMap>
/** @file console.cpp  Console subsystem.
 *
 * @todo The Console subsystem should be rewritten to be a de::System and it
 * should use Doomsday Script as the underlying engine; everything should be
 * mapped to Doomsday Script processes, functions, variables, etc., making the
 * Console a mere convenience layer. -jk
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/console/exec.h"
#include <cctype>
#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <QFile>
#include <QTextStream>
#include <de/memory.h>
#include <de/charsymbols.h>
#include <de/App>
#include <de/Binder>
#include <de/Log>
#include <de/LogBuffer>
#include <de/NativeFile>
#include <de/Process>
#include <de/Script>
#include <de/ScriptSystem>
#include <de/Time>
#include <de/TextValue>
#include "doomsday/doomsdayapp.h"
#include "doomsday/game.h"
#include "doomsday/console/knownword.h"
#include "doomsday/console/cmd.h"
#include "doomsday/console/var.h"
#include "doomsday/console/alias.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/uri.h"
#include "dd_share.h"
/** @file knownword.cpp
 * @ingroup console
 *
 * @todo Rewrite this whole thing with sensible C++ data structures and de::String! -jk
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/console/knownword.h"
#include "doomsday/console/exec.h"
#include "doomsday/console/var.h"
#include "doomsday/console/alias.h"
#include "doomsday/console/cmd.h"
#include "doomsday/help.h"
#include "doomsday/game.h"
#include <de/memory.h>
#include <de/c_wrapper.h>
#include <de/strutil.h>
#include <QList>
#include <QRegExp>
/** @file
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_URI
#include "doomsday/console/var.h"
#include "doomsday/console/exec.h"
#include "doomsday/console/knownword.h"
#include "doomsday/uri.h"
#include <de/c_wrapper.h>
#include <de/memory.h>
#include <de/Function>
#include <de/PathTree>
#include <de/TextValue>
using namespace de;
static uint numCAliases;
static calias_t **caliases;
void Con_InitAliases() {
    numCAliases = 0;
    caliases = 0;
}
void Con_ClearAliases() {
    if (caliases) {
        calias_t** cal = caliases;
        for (uint i = 0; i < numCAliases; ++i, ++cal) {
            M_Free((*cal)->name);
            M_Free((*cal)->command);
            M_Free(*cal);
        }
        M_Free(caliases);
    }
    caliases = 0;
    numCAliases = 0;
}
calias_t *Con_FindAlias(char const *name) {
    uint bottomIdx, topIdx, pivot;
    calias_t* cal;
    dd_bool isDone;
    int result;
    if (numCAliases == 0) return 0;
    if (!name || !name[0]) return 0;
    bottomIdx = 0;
    topIdx = numCAliases-1;
    cal = NULL;
    isDone = false;
    while (bottomIdx <= topIdx && !isDone) {
        pivot = bottomIdx + (topIdx - bottomIdx)/2;
        result = qstricmp(caliases[pivot]->name, name);
        if (result == 0) {
            cal = caliases[pivot];
            isDone = true;
        }
        else {
            if (result > 0) {
                if (pivot == 0) {
                    isDone = true;
                }
                else
                    topIdx = pivot - 1;
            }
            else
                bottomIdx = pivot + 1;
        }
    }
    return cal;
}
calias_t* Con_AddAlias(char const* name, char const* command) {
    if (!name || !name[0] || !command || !command[0]) return 0;
    caliases = (calias_t**) M_Realloc(caliases, sizeof(*caliases) * ++numCAliases);
    uint idx;
    for (idx = 0; idx < numCAliases-1; ++idx) {
        if (qstricmp(caliases[idx]->name, name) > 0)
            break;
    }
    if (idx != numCAliases-1)
        memmove(caliases + idx + 1, caliases + idx, sizeof(*caliases) * (numCAliases - 1 - idx));
    calias_t* newAlias = caliases[idx] = (calias_t*) M_Malloc(sizeof(*newAlias));
    newAlias->name = (char*) M_Malloc(strlen(name) + 1);
    strcpy(newAlias->name, name);
    newAlias->command = (char*) M_Malloc(strlen(command) + 1);
    strcpy(newAlias->command, command);
    Con_UpdateKnownWords();
    return newAlias;
}
void Con_DeleteAlias(calias_t* cal) {
    DENG_ASSERT(cal);
    uint idx;
    for (idx = 0; idx < numCAliases; ++idx) {
        if (caliases[idx] == cal)
            break;
    }
    if (idx == numCAliases) return;
    Con_UpdateKnownWords();
    M_Free(cal->name);
    M_Free(cal->command);
    M_Free(cal);
    if (idx < numCAliases - 1) {
        memmove(caliases + idx, caliases + idx + 1, sizeof(*caliases) * (numCAliases - idx - 1));
    }
    --numCAliases;
}
String Con_AliasAsStyledText(calias_t *alias) {
    QString str;
    QTextStream os(&str);
    os << _E(b) << alias->name << _E(.) " == " _E(>) << alias->command << _E(<);
    return str;
}
void Con_AddKnownWordsForAliases() {
    calias_t** cal = caliases;
    for (uint i = 0; i < numCAliases; ++i, ++cal) {
        Con_AddKnownWord(WT_CALIAS, *cal);
    }
}
static ccmd_t *ccmdListHead;
static blockset_t *ccmdBlockSet;
static uint numUniqueNamedCCmds;
static QMap<String, String> mappedConfigVariables;
void Con_InitCommands() {
    ccmdListHead = 0;
    ccmdBlockSet = 0;
    numUniqueNamedCCmds = 0;
}
void Con_ClearCommands(void) {
    if (ccmdBlockSet) {
        BlockSet_Delete(ccmdBlockSet);
    }
    ccmdBlockSet = 0;
    ccmdListHead = 0;
    numUniqueNamedCCmds = 0;
    mappedConfigVariables.clear();
}
void Con_AddKnownWordsForCommands() {
    for (ccmd_t* ccmd = ccmdListHead; ccmd; ccmd = ccmd->next) {
        if (ccmd->prevOverload) continue;
        Con_AddKnownWord(WT_CCMD, ccmd);
    }
}
void Con_AddCommand(ccmdtemplate_t const* ccmd) {
    int minArgs, maxArgs;
    cvartype_t args[DENG_MAX_ARGS];
    ccmd_t* newCCmd, *overloaded = 0;
    if (!ccmd) return;
    DENG_ASSERT(ccmd->name != 0);
    if (ccmd->argTemplate != 0) {
        size_t l, len;
        cvartype_t type = CVT_NULL;
        dd_bool unlimitedArgs;
        char c;
        len = strlen(ccmd->argTemplate);
        minArgs = 0;
        unlimitedArgs = false;
        for (l = 0; l < len; ++l) {
            c = ccmd->argTemplate[l];
            switch (c) {
            case 'b': type = CVT_BYTE;     break;
            case 'i': type = CVT_INT;      break;
            case 'f': type = CVT_FLOAT;    break;
            case 's': type = CVT_CHARPTR;  break;
            case '*':
                if (l != len-1)
                    App_FatalError("Con_AddCommand: CCmd '%s': '*' character "
                                   "not last in argument template: \"%s\".",
                                   ccmd->name, ccmd->argTemplate);
                unlimitedArgs = true;
                type = CVT_NULL; // Not a real argument.
                break;
            default:
                App_FatalError("Con_AddCommand: CCmd '%s': Invalid character "
                               "'%c' in argument template: \"%s\".", ccmd->name, c,
                               ccmd->argTemplate);
            }
            if (type != CVT_NULL) {
                if (minArgs >= DENG_MAX_ARGS)
                    App_FatalError("Con_AddCommand: CCmd '%s': Too many arguments. "
                                   "Limit is %i.", ccmd->name, DENG_MAX_ARGS);
                args[minArgs++] = type;
            }
        }
        if (unlimitedArgs) {
            maxArgs = -1;
            if (minArgs == 0)
                minArgs = -1;
        }
        else {
            maxArgs = minArgs;
        }
    }
    else // It's usage is NOT validated by Doomsday.
    {
        minArgs = maxArgs = -1;
    }
    { ccmd_t* other;
    if ((other = Con_FindCommand(ccmd->name)) != 0) {
        dd_bool unique = true;
        if (minArgs == -1 && maxArgs == -1)
            unique = false;
        if (unique) {
            ccmd_t* variant = other;
            do {
                if (variant->minArgs == -1 && variant->maxArgs == -1)
                    unique = false;
                else if (variant->minArgs < minArgs && variant->maxArgs == -1)
                    unique = false;
                else if (variant->minArgs > minArgs && maxArgs == -1)
                    unique = false;
                else if (variant->minArgs == minArgs) {
                    unique = false;
                }
                if (!unique && variant->execFunc == ccmd->execFunc)
                    App_FatalError("Con_AddCommand: A CCmd by the name '%s' is already registered and the callback funcs are "
                                   "the same, is this really what you wanted?", ccmd->name);
            } while ((variant = variant->nextOverload) != 0);
        }
        if (!unique)
            App_FatalError("Con_AddCommand: A CCmd by the name '%s' is already registered. Their parameter lists would be ambiguant.", ccmd->name);
        overloaded = other;
    }}
    if (!ccmdBlockSet)
        ccmdBlockSet = BlockSet_New(sizeof(ccmd_t), 32);
    newCCmd = (ccmd_t*) BlockSet_Allocate(ccmdBlockSet);
    char* nameCopy = (char*) Z_Malloc(strlen(ccmd->name) + 1, PU_APPSTATIC, NULL);
    if (!nameCopy) App_FatalError("Con_AddCommand: Failed on allocation of %lu bytes for command name.", (unsigned long) (strlen(ccmd->name) + 1));
    strcpy(nameCopy, ccmd->name);
    newCCmd->name = nameCopy;
    newCCmd->execFunc = ccmd->execFunc;
    newCCmd->flags = ccmd->flags;
    newCCmd->nextOverload = newCCmd->prevOverload = 0;
    newCCmd->minArgs = minArgs;
    newCCmd->maxArgs = maxArgs;
    memcpy(newCCmd->args, &args, sizeof(newCCmd->args));
    newCCmd->next = ccmdListHead;
    ccmdListHead = newCCmd;
    if (!overloaded) {
        ++numUniqueNamedCCmds;
        Con_UpdateKnownWords();
        return;
    }
    newCCmd->nextOverload = overloaded;
    overloaded->prevOverload = newCCmd;
}
void Con_AddCommandList(ccmdtemplate_t const* cmdList) {
    if (!cmdList) return;
    for (; cmdList->name; ++cmdList) {
        Con_AddCommand(cmdList);
    }
}
ccmd_t *Con_FindCommand(char const *name) {
    if (name && name[0]) {
        for (ccmd_t *ccmd = ccmdListHead; ccmd; ccmd = ccmd->next) {
            if (qstricmp(name, ccmd->name)) continue;
            while (ccmd->prevOverload) { ccmd = ccmd->prevOverload; }
            return ccmd;
        }
    }
    return 0;
}
/**
 * Outputs the usage information for the given ccmd to the console.
 *
 * @param ccmd          The ccmd to print the usage info for.
 * @param allOverloads  @c true= print usage info for all overloaded variants.
 *                      Otherwise only the info for @a ccmd.
 */
void Con_PrintCommandUsage(ccmd_t const *ccmd, bool allOverloads) {
    if (!ccmd) return;
    if (allOverloads) {
        while (ccmd->prevOverload) { ccmd = ccmd->prevOverload; }
    }
    LOG_SCR_NOTE(_E(b) "Usage:" _E(.) "\n  " _E(>) + Con_CmdUsageAsStyledText(ccmd));
    if (allOverloads) {
        while ((ccmd = ccmd->nextOverload)) {
            LOG_SCR_MSG("  " _E(>) + Con_CmdUsageAsStyledText(ccmd));
        }
    }
}
ccmd_t *Con_FindCommandMatchArgs(cmdargs_t *args) {
    if (!args) return 0;
    if (ccmd_t *ccmd = Con_FindCommand(args->argv[0])) {
        ccmd_t *variant = ccmd;
        do {
            dd_bool invalidArgs = false;
            if (!(variant->minArgs == -1 && variant->maxArgs == -1)) {
                if (args->argc-1 < variant->minArgs) {
                    invalidArgs = true;
                }
                else if (variant->maxArgs != -1 && args->argc-1 > variant->maxArgs) {
                    invalidArgs = true;
                }
                else {
                    for (int i = 0; i < variant->minArgs && !invalidArgs; ++i) {
                        switch (variant->args[i]) {
                        case CVT_BYTE:
                            invalidArgs = !M_IsStringValidByte(args->argv[i+1]);
                            break;
                        case CVT_INT:
                            invalidArgs = !M_IsStringValidInt(args->argv[i+1]);
                            break;
                        case CVT_FLOAT:
                            invalidArgs = !M_IsStringValidFloat(args->argv[i+1]);
                            break;
                        default:
                            break;
                        }
                    }
                }
            }
            if (!invalidArgs) {
                return variant; // This is the one!
            }
        } while ((variant = variant->nextOverload) != 0);
        Con_PrintCommandUsage(ccmd);
    }
    return 0;
}
dd_bool Con_IsValidCommand(char const* name) {
    if (!name || !name[0])
        return false;
    if (Con_FindCommand(name) != 0)
        return true;
    return (Con_FindAlias(name) != 0);
}
String Con_CmdUsageAsStyledText(ccmd_t const *ccmd) {
    DENG2_ASSERT(ccmd != 0);
    if (ccmd->minArgs == -1 && ccmd->maxArgs == -1)
        return String();
    String argText;
    for (int i = 0; i < ccmd->minArgs; ++i) {
        switch (ccmd->args[i]) {
        case CVT_BYTE:    argText += " (byte)";   break;
        case CVT_INT:     argText += " (int)";    break;
        case CVT_FLOAT:   argText += " (float)";  break;
        case CVT_CHARPTR: argText += " (string)"; break;
        default: break;
        }
    }
    if (ccmd->maxArgs == -1) {
        argText += " ...";
    }
    return _E(b) + String(ccmd->name) + _E(.) _E(l) + argText + _E(.);
}
String Con_CmdAsStyledText(ccmd_t *cmd) {
    char const *str;
    if ((str = DH_GetString(DH_Find(cmd->name), HST_DESCRIPTION))) {
        return String(_E(b) "%1 " _E(.) _E(>) _E(2) "%2" _E(.) _E(<)).arg(cmd->name).arg(str);
    }
    else {
        return String(_E(b) "%1" _E(.)).arg(cmd->name);
    }
}
D_CMD(MappedConfigVariable) {
    DENG_UNUSED(src);
    auto const found = mappedConfigVariables.constFind(argv[0]);
    DENG2_ASSERT(found != mappedConfigVariables.constEnd()); // mapping must be defined
    Variable &var = Config::get(found.value());
    if (argc == 1) {
        LOG_SCR_MSG(_E(b) "%s" _E(.) " = " _E(>) "%s " _E(l)_E(C) "[Config.%s]")
                << argv[0]
                << var.value().asText()
                << found.value();
    }
    else if (argc > 1) {
        if (maybeAs<TextValue>(var.value())) {
            var.set(new TextValue(argv[1]));
        }
        else {
            var.set(new NumberValue(String(argv[1]).toDouble()));
        }
    }
    return true;
}
void Con_AddMappedConfigVariable(char const *consoleName, char const *opts, String const &configVariable) {
    DENG2_ASSERT(!mappedConfigVariables.contains(consoleName)); // redefining not handled
    mappedConfigVariables.insert(consoleName, configVariable);
    C_CMD(consoleName, "",   MappedConfigVariable);
    C_CMD(consoleName, opts, MappedConfigVariable);
}
#define SC_EMPTY_QUOTE  -1
#define PRBUFF_SIZE 0x2000
enum {
    IF_EQUAL,
    IF_NOT_EQUAL,
    IF_GREATER,
    IF_LESS,
    IF_GEQUAL,
    IF_LEQUAL
};
#define CMDTYPESTR(src) \
         (src == CMDS_DDAY? "a direct call" \
        : src == CMDS_GAME? "a game library call" \
        : src == CMDS_CONSOLE? "the console" \
        : src == CMDS_BIND? "a binding" \
        : src == CMDS_CONFIG? "a cfg file" \
        : src == CMDS_PROFILE? "a player profile" \
        : src == CMDS_CMDLINE? "the command line" \
        : src == CMDS_SCRIPT? "an action command" : "???")
typedef struct execbuff_s {
    dd_bool used;               // Is this in use?
    timespan_t when;            // System time when to execute the command.
    byte    source;             // Where the command came from
    dd_bool isNetCmd;           // Command was sent over the net to us.
    char    subCmd[1024];       // A single command w/args.
} execbuff_t;
D_CMD(AddSub);
D_CMD(IncDec);
D_CMD(Alias);
D_CMD(Echo);
D_CMD(Help);
D_CMD(If);
D_CMD(Parse);
D_CMD(Quit);
D_CMD(Repeat);
D_CMD(Toggle);
D_CMD(Wait);
D_CMD(InspectMobj);
D_CMD(DebugCrash);
D_CMD(DebugError);
D_CMD(DoomsdayScript);
void initVariableBindings(Binder &);
static int executeSubCmd(const char *subCmd, byte src, dd_bool isNetCmd);
static void Con_SplitIntoSubCommands(char const *command,
                                     timespan_t markerOffset, byte src,
                                     dd_bool isNetCmd);
static void Con_ClearExecBuffer(void);
byte    ConsoleSilent = false;
static dd_bool      ConsoleInited;   // Has Con_Init() been called?
static Binder       consoleBinder;
static execbuff_t * exBuff;
static int          exBuffSize;
static execbuff_t * curExec;
static bool         consoleChanged = false;
void Con_Register(void) {
    C_CMD("add",            NULL,   AddSub);
    C_CMD("after",          "is",   Wait);
    C_CMD("alias",          NULL,   Alias);
    C_CMD("dec",            NULL,   IncDec);
    C_CMD("echo",           "s*",   Echo);
    C_CMD("print",          "s*",   Echo);
    C_CMD("exec",           "s*",   Parse);
    C_CMD("if",             NULL,   If);
    C_CMD("inc",            NULL,   IncDec);
    C_CMD("repeat",         "ifs",  Repeat);
    C_CMD("sub",            NULL,   AddSub);
    C_CMD("toggle",         "s",    Toggle);
#ifdef _DEBUG
    C_CMD("crash",          NULL,   DebugCrash);
#endif
    C_CMD("ds",             "s*",   DoomsdayScript);
    Con_DataRegister();
}
static void PrepareCmdArgs(cmdargs_t *cargs, const char *lpCmdLine) {
#define IS_ESC_CHAR(x)  ((x) == '"' || (x) == '\\' || (x) == '{' || (x) == '}')
    size_t              i, len = strlen(lpCmdLine);
    memset(cargs, 0, sizeof(cmdargs_t));
    strcpy(cargs->cmdLine, lpCmdLine);
    for (i = 0; i < len; ++i) {
        if (DENG_ISSPACE(cargs->cmdLine[i]))
            cargs->cmdLine[i] = 0;
        if (cargs->cmdLine[i] == '\\' && IS_ESC_CHAR(cargs->cmdLine[i + 1]))
        {   // Escape sequence.
            memmove(cargs->cmdLine + i, cargs->cmdLine + i + 1,
                    sizeof(cargs->cmdLine) - i - 1);
            len--;
            continue;
        }
        if (cargs->cmdLine[i] == '"')
        {   // Find the end.
            size_t              start = i;
            cargs->cmdLine[i] = 0;
            for (++i; i < len && cargs->cmdLine[i] != '"'; ++i) {
                if (cargs->cmdLine[i] == '\\' && IS_ESC_CHAR(cargs->cmdLine[i + 1])) // Escape sequence?
                {
                    memmove(cargs->cmdLine + i, cargs->cmdLine + i + 1,
                            sizeof(cargs->cmdLine) - i - 1);
                    len--;
                    continue;
                }
            }
            if (i == len)
                break;
            if (i == start + 1)
                cargs->cmdLine[i] = SC_EMPTY_QUOTE;
            else
                cargs->cmdLine[i] = 0;
        }
        if (cargs->cmdLine[i] == '{')
        {   // Find matching end, braces are another notation for quotes.
            int             level = 0;
            size_t          start = i;
            cargs->cmdLine[i] = 0;
            for (++i; i < len; ++i) {
                if (cargs->cmdLine[i] == '\\' && IS_ESC_CHAR(cargs->cmdLine[i + 1])) // Escape sequence?
                {
                    memmove(cargs->cmdLine + i, cargs->cmdLine + i + 1,
                            sizeof(cargs->cmdLine) - i - 1);
                    len--;
                    i++;
                    continue;
                }
                if (cargs->cmdLine[i] == '}') {
                    if (!level)
                        break;
                    level--;
                }
                if (cargs->cmdLine[i] == '{')
                    level++;
            }
            if (i == len)
                break;
            if (i == start + 1)
                cargs->cmdLine[i] = SC_EMPTY_QUOTE;
            else
                cargs->cmdLine[i] = 0;
        }
    }
    cargs->argc = 0;
    for (i = 0; i < len; ++i) {
        if (!cargs->cmdLine[i])
            continue;
        if (cargs->cmdLine[i] == char(SC_EMPTY_QUOTE))
            cargs->cmdLine[i] = 0;  // Just an empty string.
        cargs->argv[cargs->argc++] = cargs->cmdLine + i;
        i += strlen(cargs->cmdLine + i);
    }
#undef IS_ESC_CHAR
}
static Value *Function_Console_ListVars(Context &, Function::ArgumentValues const &args) {
    StringList vars;
    Con_TermsRegex(vars, args.at(0)->asText(), WT_CVAR);
    std::unique_ptr<ArrayValue> result(new ArrayValue);
    for (String v : vars) {
        *result << new TextValue(v);
    }
    return result.release();
}
dd_bool Con_Init(void) {
    if (ConsoleInited) return true;
    LOG_SCR_VERBOSE("Initializing the console...");
    consoleBinder.initNew();
    initVariableBindings(consoleBinder);
    consoleBinder << DENG2_FUNC(Console_ListVars, "listVars", "pattern");
    App::scriptSystem().addNativeModule("Console", consoleBinder.module());
    exBuff = NULL;
    exBuffSize = 0;
    ConsoleInited = true;
    return true;
}
void Con_Shutdown(void) {
    if (!ConsoleInited) return;
    LOG_SCR_VERBOSE("Shutting down the console...");
    Con_ClearExecBuffer();
    Con_ShutdownDatabases();
    consoleBinder.deinit();
    ConsoleInited = false;
}
void Con_MarkAsChanged(dd_bool changed) {
    consoleChanged = changed;
}
dd_bool Con_IsChanged() {
    return consoleChanged;
}
#if 0 // should use libshell!
#ifdef __CLIENT__
/**
 * Send a console command to the server.
 * This shouldn't be called unless we're logged in with the right password.
 */
static void Con_Send(const char *command, byte src, int silent) {
    ushort len = (ushort) strlen(command);
    LOG_AS("Con_Send");
    if (len >= 0x8000) {
        LOGDEV_NET_ERROR("Command is too long, length=%i") << len;
        return;
    }
    Msg_Begin(PKT_COMMAND2);
    Writer_WriteUInt16(msgWriter, len | (silent ? 0x8000 : 0));
    Writer_WriteUInt16(msgWriter, 0); // flags. Unused at present.
    Writer_WriteByte(msgWriter, src);
    Writer_Write(msgWriter, command, len);
    Msg_End();
    Net_SendBuffer(0, 0);
}
#endif // __CLIENT__
#endif
static void Con_QueueCmd(const char *singleCmd, timespan_t atSecond,
                         byte source, dd_bool isNetCmd) {
    execbuff_t *ptr = NULL;
    int     i;
    for (i = 0; i < exBuffSize; ++i)
        if (!exBuff[i].used) {
            ptr = exBuff + i;
            break;
        }
    if (ptr == NULL) {
        exBuff = (execbuff_t *) M_Realloc(exBuff, sizeof(execbuff_t) * ++exBuffSize);
        ptr = exBuff + exBuffSize - 1;
    }
    ptr->used = true;
    strcpy(ptr->subCmd, singleCmd);
    ptr->when = atSecond;
    ptr->source = source;
    ptr->isNetCmd = isNetCmd;
}
static void Con_ClearExecBuffer(void) {
    M_Free(exBuff);
    exBuff = NULL;
    exBuffSize = 0;
}
/**
 * The execbuffer is used to schedule commands for later.
 *
 * @return          @c false, if an executed command fails.
 */
static dd_bool Con_CheckExecBuffer(void) {
#define BUFFSIZE 1024 /// @todo Rewrite all of this; use de::String. -jk
    dd_bool ret = true;
    int     i;
    char    storage[BUFFSIZE];
    storage[255] = 0;
    TimeSpan const now = TimeSpan::sinceStartOfProcess();
    for (i = 0; i < exBuffSize; ++i) {
        execbuff_t *ptr = exBuff + i;
        if (!ptr->used || ptr->when > now)
            continue;
        curExec = ptr;
        ptr->used = false;
        strncpy(storage, ptr->subCmd, BUFFSIZE - 1);
        bool const isInteractive =
                (ptr->source == CMDS_CONSOLE ||
                 ptr->source == CMDS_CMDLINE);
        if (isInteractive) {
            LOG().beginInteractive();
        }
        if (!executeSubCmd(storage, ptr->source, ptr->isNetCmd)) {
            ret = false;
        }
        if (isInteractive) {
            LOG().endInteractive();
        }
    }
    return ret;
#undef BUFFSIZE
}
void Con_Ticker(timespan_t /*time*/)
{
    Con_CheckExecBuffer();
}
/**
 * expCommand gets reallocated in the expansion process.
 * This could be bit more clever.
 */
static void expandWithArguments(char **expCommand, cmdargs_t *args) {
    int             p;
    char           *text = *expCommand;
    size_t          i, off, size = strlen(text) + 1;
    for (i = 0; text[i]; ++i) {
        if (text[i] == '%' && (text[i + 1] >= '1' && text[i + 1] <= '9')) {
            char           *substr;
            int             aidx = text[i + 1] - '1' + 1;
            if (aidx > args->argc - 1)
                substr = (char *) "";
            else
                substr = args->argv[aidx];
            memmove(text + i, text + i + 2, size - i - 2);
            off = strlen(substr);
            text = *expCommand = (char *) M_Realloc(*expCommand, size += off - 2);
            if (off) {
                memmove(text + i + off, text + i, size - i - off);
                memcpy(text + i, substr, off);
            }
            i += off - 1;
        }
        else if (text[i] == '%' && text[i + 1] == '0') {
            memmove(text + i, text + i + 2, size - i - 2);
            text = *expCommand = (char *) M_Realloc(*expCommand, size -= 2);
            for (p = args->argc - 1; p > 0; p--) {
                off = strlen(args->argv[p]) + 1;
                text = *expCommand = (char *) M_Realloc(*expCommand, size += off);
                memmove(text + i + off, text + i, size - i - off);
                text[i] = ' ';
                memcpy(text + i + 1, args->argv[p], off - 1);
            }
        }
    }
}
/**
 * The command is executed forthwith!!
 */
static int executeSubCmd(const char *subCmd, byte src, dd_bool isNetCmd) {
    cmdargs_t args;
    ccmd_t *ccmd;
    cvar_t *cvar;
    calias_t *cal;
    PrepareCmdArgs(&args, subCmd);
    if (!args.argc)
        return true;
    ccmd = Con_FindCommandMatchArgs(&args);
    if (ccmd != NULL) {
        dd_bool canExecute = true;
        if ((ccmd->flags & CMDF_NO_NULLGAME) && DoomsdayApp::game().isNull()) {
            LOG_SCR_ERROR("Execution of command '%s' is only allowed when a game is loaded") << ccmd->name;
            return true;
        }
#if 0
        if (isDedicated && (ccmd->flags & CMDF_NO_DEDICATED)) {
            LOG_SCR_ERROR("Execution of command '%s' not possible in dedicated mode") << ccmd->name;
            return true;
        }
        if (isServer && isNetCmd) {
            if (ccmd->flags & CMDF_CLIENT) {
                LOG_NET_ERROR("Execution of command '%s' blocked (client attempted invocation);"
                              "this command is not permitted for use by clients") << ccmd->name;
                return true;
            }
            switch (src) {
            case CMDS_UNKNOWN:
            case CMDS_CONFIG:
            case CMDS_PROFILE:
            case CMDS_CMDLINE:
            case CMDS_SCRIPT:
                LOG_NET_ERROR("Execution of command '%s' blocked (client attempted invocation via %s); "
                              "this method is not permitted by clients") << ccmd->name << CMDTYPESTR(src);
                return true;
            default:
                break;
            }
        }
#endif
        switch (src) {
        case CMDS_UNKNOWN:
            canExecute = false;
            break;
        case CMDS_DDAY:
            if (ccmd->flags & CMDF_DDAY)
                canExecute = false;
            break;
        case CMDS_GAME:
            if (ccmd->flags & CMDF_GAME)
                canExecute = false;
            break;
        case CMDS_CONSOLE:
            if (ccmd->flags & CMDF_CONSOLE)
                canExecute = false;
            break;
        case CMDS_BIND:
            if (ccmd->flags & CMDF_BIND)
                canExecute = false;
            break;
        case CMDS_CONFIG:
            if (ccmd->flags & CMDF_CONFIG)
                canExecute = false;
            break;
        case CMDS_PROFILE:
            if (ccmd->flags & CMDF_PROFILE)
                canExecute = false;
            break;
        case CMDS_CMDLINE:
            if (ccmd->flags & CMDF_CMDLINE)
                canExecute = false;
            break;
        case CMDS_SCRIPT:
            if (ccmd->flags & CMDF_DED)
                canExecute = false;
            break;
        default:
            return true;
        }
        if (!canExecute) {
            LOG_SCR_ERROR("'%s' cannot be executed via %s") << ccmd->name << CMDTYPESTR(src);
            return true;
        }
        if (canExecute) {
            /**
             * Execute the command!
             * \note Console command execution may invoke a full update of the
             * console databases; thus the @c ccmd pointer may be invalid after
             * this call.
             */
            int result;
            if ((result = ccmd->execFunc(src, args.argc, args.argv)) == false) {
                LOG_SCR_ERROR("'%s' failed") << args.argv[0];
            }
            return result;
        }
        return true;
    }
    cvar = Con_FindVariable(args.argv[0]);
    if (cvar != NULL) {
        dd_bool outOfRange = false, setting = false, hasCallback;
        /**
         * \note Change notification callback execution may invoke
         * a full update of the console databases; thus the @c cvar
         * pointer may be invalid once a callback executes.
         */
        hasCallback = (cvar->notifyChanged != 0);
        if (args.argc == 2 ||
           (args.argc == 3 && !qstricmp(args.argv[1], "force"))) {
            char* argptr = args.argv[args.argc - 1];
            dd_bool forced = args.argc == 3;
            setting = true;
            if (cvar->flags & CVF_READ_ONLY) {
                CVar_PrintReadOnlyWarning(cvar);
            }
            else if ((cvar->flags & CVF_PROTECTED) && !forced) {
                AutoStr* name = CVar_ComposePath(cvar);
                LOG_SCR_NOTE("%s is protected; you shouldn't change its value -- "
                             "use the command: " _E(b) "'%s force %s'" _E(.) " to modify it anyway")
                        << Str_Text(name) << Str_Text(name) << argptr;
            }
            else {
                Con_MarkAsChanged(true);
                switch (cvar->type) {
                case CVT_BYTE: {
                    byte val = (byte) strtol(argptr, NULL, 0);
                    if (!forced &&
                       ((!(cvar->flags & CVF_NO_MIN) && val < cvar->min) ||
                        (!(cvar->flags & CVF_NO_MAX) && val > cvar->max)))
                        outOfRange = true;
                    else
                        CVar_SetInteger(cvar, val);
                    break;
                  }
                case CVT_INT: {
                    int val = strtol(argptr, NULL, 0);
                    if (!forced &&
                       ((!(cvar->flags & CVF_NO_MIN) && val < cvar->min) ||
                        (!(cvar->flags & CVF_NO_MAX) && val > cvar->max)))
                        outOfRange = true;
                    else
                        CVar_SetInteger(cvar, val);
                    break;
                  }
                case CVT_FLOAT: {
                    float val = strtod(argptr, NULL);
                    if (!forced &&
                       ((!(cvar->flags & CVF_NO_MIN) && val < cvar->min) ||
                        (!(cvar->flags & CVF_NO_MAX) && val > cvar->max)))
                        outOfRange = true;
                    else
                        CVar_SetFloat(cvar, val);
                    break;
                  }
                case CVT_CHARPTR:
                    CVar_SetString(cvar, argptr);
                    break;
                case CVT_URIPTR:
                    CVar_SetUri(cvar, de::makeUri(argptr));
                    break;
                default: break;
                }
            }
        }
        if (outOfRange) {
            AutoStr* name = CVar_ComposePath(cvar);
            if (!(cvar->flags & (CVF_NO_MIN | CVF_NO_MAX))) {
                char temp[20];
                strcpy(temp, M_TrimmedFloat(cvar->min));
                LOG_SCR_ERROR("%s <= %s <= %s") << temp << Str_Text(name) << M_TrimmedFloat(cvar->max);
            }
            else if (cvar->flags & CVF_NO_MAX) {
                LOG_SCR_ERROR("%s >= %s") << Str_Text(name) << M_TrimmedFloat(cvar->min);
            }
            else {
                LOG_SCR_ERROR("%s <= %s") << Str_Text(name) << M_TrimmedFloat(cvar->max);
            }
        }
        else if (!setting) // Show the value.
        {
            if (setting && hasCallback) {
                cvar = Con_FindVariable(args.argv[0]);
            }
            if (cvar) {
                Con_PrintCVar(cvar, "");
            }
        }
        return true;
    }
    cal = Con_FindAlias(args.argv[0]);
    if (cal != NULL) {
        char* expCommand;
        expCommand = (char *) M_Malloc(strlen(cal->command) + 1);
        strcpy(expCommand, cal->command);
        expandWithArguments(&expCommand, &args);
        Con_SplitIntoSubCommands(expCommand, 0, src, isNetCmd);
        M_Free(expCommand);
        return true;
    }
    if (Con_FindCommand(args.argv[0])) {
        LOG_SCR_WARNING("%s: command arguments invalid") << args.argv[0];
        Con_Executef(CMDS_DDAY, false, "help %s", args.argv[0]);
    }
    else {
        LOG_SCR_MSG("%s: unknown identifier") << args.argv[0];
    }
    return false;
}
/**
 * Splits the command into subcommands and queues them into the
 * execution buffer.
 */
static void Con_SplitIntoSubCommands(const char *command,
                                     timespan_t markerOffset, byte src,
                                     dd_bool isNetCmd) {
#define BUFFSIZE        2048
    char            subCmd[BUFFSIZE];
    int             inQuotes = false, escape = false;
    size_t          gPos = 0, scPos = 0, len;
    if (!command || command[0] == 0)
        return;
    len = strlen(command);
    while (gPos < len && command[gPos] == ';' && command[gPos] != 0)
        gPos++;
    subCmd[0] = subCmd[BUFFSIZE-1] = 0;
    for (; command[gPos];) {
        escape = false;
        if (inQuotes && command[gPos] == '\\')   // Escape sequence?
        {
            subCmd[scPos++] = command[gPos++];
            escape = true;
        }
        if (command[gPos] == '"' && !escape)
            inQuotes = !inQuotes;
        subCmd[scPos++] = command[gPos++];
        if (subCmd[0] == ' ')
            scPos = 0;          // No spaces in the beginning.
        if ((command[gPos] == ';' && !inQuotes) || command[gPos] == 0) {
            while (gPos < len && command[gPos] == ';' && command[gPos] != 0)
                gPos++;
            subCmd[scPos] = 0;
        }
        else {
            continue;
        }
        Con_QueueCmd(subCmd, TimeSpan::sinceStartOfProcess() + markerOffset, src, isNetCmd);
        scPos = 0;
    }
#undef BUFFSIZE
}
int Con_Execute(byte src, const char *command, int silent, dd_bool netCmd) {
    if (silent) {
        ConsoleSilent = true;
    }
    Con_SplitIntoSubCommands(command, 0, src, netCmd);
    int ret = Con_CheckExecBuffer();
    if (silent) {
        ConsoleSilent = false;
    }
    return ret;
}
int Con_Executef(byte src, int silent, const char *command, ...) {
    va_list         argptr;
    char            buffer[4096];
    va_start(argptr, command);
    dd_vsnprintf(buffer, sizeof(buffer), command, argptr);
    va_end(argptr);
    return Con_Execute(src, buffer, silent, false);
}
bool Con_Parse(File const &file, bool silently) {
    Block utf8;
    file >> utf8;
    String contents = String::fromUtf8(utf8);
    QTextStream in(&contents);
    int currentLine = 1;
    while (!in.atEnd()) {
        String const line = String(in.readLine()).leftStrip();
        if (!line.isEmpty() && line.first() != '#') {
            if (!Con_Execute(CMDS_CONFIG, line.toUtf8(), silently, false)) {
                if (!silently) {
                    LOG_SCR_WARNING("%s (line %i): error executing command \"%s\"")
                            << file.description()
                            << currentLine
                            << line;
                }
            }
        }
        currentLine += 1;
    }
    Con_MarkAsChanged(false);
    return true;
}
/**
 * Create an alias.
 */
static void makeAlias(char *aName, char *command) {
    calias_t *cal = Con_FindAlias(aName);
    dd_bool remove = false;
    if (command == NULL)
        remove = true;
    else if (command[0] == 0)
        remove = true;
    if (cal && remove) // This alias will be removed.
    {
        Con_DeleteAlias(cal);
        return; // We're done.
    }
    if (cal) {
        cal->command = (char *) M_Realloc(cal->command, strlen(command) + 1);
        strcpy(cal->command, command);
        return;
    }
    Con_AddAlias(aName, command);
}
D_CMD(Alias) {
    DENG2_UNUSED(src);
    if (argc != 3 && argc != 2) {
        LOG_SCR_NOTE("Usage: %s (alias) (cmd)") << argv[0];
        LOG_SCR_MSG("Example: alias bigfont \"font size 3\"");
        LOG_SCR_MSG("Use %%1-%%9 to pass the alias arguments to the command.");
        return true;
    }
    makeAlias(argv[1], argc == 3 ? argv[2] : NULL);
    if (argc != 3) {
        LOG_SCR_MSG("Alias '%s' deleted") << argv[1];
    }
    return true;
}
D_CMD(Parse) {
    DENG2_UNUSED(src);
    int     i;
    for (i = 1; i < argc; ++i) {
        try {
            LOG_SCR_MSG("Parsing \"%s\"") << argv[i];
            std::unique_ptr<NativeFile> file(
                        NativeFile::newStandalone(App::app().nativeHomePath() / NativePath(argv[i])));
            Con_Parse(*file, false /*not silent*/);
        }
        catch (Error const &er) {
            LOG_SCR_ERROR("Failed to parse \"%s\": %s")
                    << argv[i] << er.asText();
        }
    }
    return true;
}
D_CMD(Wait) {
    DENG2_UNUSED2(src, argc);
    timespan_t offset;
    offset = strtod(argv[1], NULL) / 35;    // Offset in seconds.
    if (offset < 0)
        offset = 0;
    Con_SplitIntoSubCommands(argv[2], offset, CMDS_CONSOLE, false);
    return true;
}
D_CMD(Repeat) {
    DENG2_UNUSED2(src, argc);
    int     count;
    timespan_t interval, offset;
    count = atoi(argv[1]);
    interval = strtod(argv[2], NULL) / 35;  // In seconds.
    offset = 0;
    while (count-- > 0) {
        offset += interval;
        Con_SplitIntoSubCommands(argv[3], offset, CMDS_CONSOLE, false);
    }
    return true;
}
D_CMD(Echo) {
    DENG2_UNUSED(src);
    int     i;
    for (i = 1; i < argc; ++i) {
        LOG_MSG("%s") << argv[i];
    }
    return true;
}
static dd_bool cvarAddSub(const char* name, float delta, dd_bool force) {
    cvar_t* cvar = Con_FindVariable(name);
    float val;
    if (!cvar) {
        if (name && name[0])
            LOG_SCR_ERROR("%s is not a known cvar") << name;
        return false;
    }
    if (cvar->flags & CVF_READ_ONLY) {
        CVar_PrintReadOnlyWarning(cvar);
        return false;
    }
    val = CVar_Float(cvar) + delta;
    if (!force) {
        if (!(cvar->flags & CVF_NO_MAX) && val > cvar->max)
            val = cvar->max;
        if (!(cvar->flags & CVF_NO_MIN) && val < cvar->min)
            val = cvar->min;
    }
    CVar_SetFloat(cvar, val);
    return true;
}
/**
 * Rather messy, wouldn't you say?
 */
D_CMD(AddSub) {
    DENG2_UNUSED(src);
    dd_bool             force = false;
    float               delta = 0;
    if (argc <= 2) {
        LOG_SCR_NOTE("Usage: %s (cvar) (val) (force)") << argv[0];
        LOG_SCR_MSG("Use force to make cvars go off limits.");
        return true;
    }
    if (argc >= 4) {
        force = !qstricmp(argv[3], "force");
    }
    delta = strtod(argv[2], NULL);
    if (!qstricmp(argv[0], "sub"))
        delta = -delta;
    return cvarAddSub(argv[1], delta, force);
}
/**
 * Rather messy, wouldn't you say?
 */
D_CMD(IncDec) {
    DENG2_UNUSED(src);
    dd_bool force = false;
    cvar_t* cvar;
    float val;
    if (argc == 1) {
        LOG_SCR_NOTE("Usage: %s (cvar) (force)") << argv[0];
        LOG_SCR_MSG("Use force to make cvars go off limits.");
        return true;
    }
    if (argc >= 3) {
        force = !qstricmp(argv[2], "force");
    }
    cvar = Con_FindVariable(argv[1]);
    if (!cvar)
        return false;
    if (cvar->flags & CVF_READ_ONLY) {
        LOG_SCR_ERROR("%s (cvar) is read-only, it cannot be changed (even with force)") << argv[1];
        return false;
    }
    val = CVar_Float(cvar);
    val += !qstricmp(argv[0], "inc")? 1 : -1;
    if (!force) {
        if (!(cvar->flags & CVF_NO_MAX) && val > cvar->max)
            val = cvar->max;
        if (!(cvar->flags & CVF_NO_MIN) && val < cvar->min)
            val = cvar->min;
    }
    CVar_SetFloat(cvar, val);
    return true;
}
/**
 * Toggle the value of a variable between zero and nonzero.
 */
D_CMD(Toggle) {
    DENG2_UNUSED2(src, argc);
    cvar_t *cvar = Con_FindVariable(argv[1]);
    if (!cvar) return false;
    CVar_SetInteger(cvar, CVar_Integer(cvar)? 0 : 1);
    return true;
}
/**
 * Execute a command if the condition passes.
 */
D_CMD(If) {
    struct {
        const char *opstr;
        uint    op;
    } operators[] = {
        {"not", IF_NOT_EQUAL},
        {"=",   IF_EQUAL},
        {">",   IF_GREATER},
        {"<",   IF_LESS},
        {">=",  IF_GEQUAL},
        {"<=",  IF_LEQUAL},
        {NULL,  0}
    };
    uint        i, oper = IF_EQUAL;
    cvar_t     *var;
    dd_bool     isTrue = false;
    if (argc != 5 && argc != 6) {
        LOG_SCR_NOTE("Usage: %s (cvar) (operator) (value) (cmd) (else-cmd)") << argv[0];
        LOG_SCR_MSG("Operator must be one of: not, =, >, <, >=, <=");
        LOG_SCR_MSG("The (else-cmd) can be omitted.");
        return true;
    }
    var = Con_FindVariable(argv[1]);
    if (!var)
        return false;
    for (i = 0; operators[i].opstr; ++i)
        if (!qstricmp(operators[i].opstr, argv[2])) {
            oper = operators[i].op;
            break;
        }
    if (!operators[i].opstr)
        return false;           // Bad operator.
    switch (var->type) {
    case CVT_BYTE:
    case CVT_INT: {
        int value = (var->type == CVT_INT ? CV_INT(var) : CV_BYTE(var));
        int test = strtol(argv[3], 0, 0);
        isTrue = (oper == IF_EQUAL     ? value == test :
                  oper == IF_NOT_EQUAL ? value != test :
                  oper == IF_GREATER   ? value >  test :
                  oper == IF_LESS      ? value <  test :
                  oper == IF_GEQUAL    ? value >= test :
                                         value <= test);
        break;
        }
    case CVT_FLOAT: {
        float value = CV_FLOAT(var);
        float test = strtod(argv[3], 0);
        isTrue = (oper == IF_EQUAL     ? value == test :
                  oper == IF_NOT_EQUAL ? value != test :
                  oper == IF_GREATER   ? value >  test :
                  oper == IF_LESS      ? value <  test :
                  oper == IF_GEQUAL    ? value >= test :
                                         value <= test);
        break;
        }
    case CVT_CHARPTR: {
        int comp = qstricmp(CV_CHARPTR(var), argv[3]);
        isTrue = (oper == IF_EQUAL     ? comp == 0 :
                  oper == IF_NOT_EQUAL ? comp != 0 :
                  oper == IF_GREATER   ? comp >  0 :
                  oper == IF_LESS      ? comp <  0 :
                  oper == IF_GEQUAL    ? comp >= 0 :
                                         comp <= 0);
        }
        break;
    default:
        DENG_ASSERT(!"CCmdIf: Invalid cvar type");
        return false;
    }
    if (isTrue) {
        Con_Execute(src, argv[4], ConsoleSilent, false);
    }
    else if (argc == 6) {
        Con_Execute(src, argv[5], ConsoleSilent, false);
    }
    return true;
}
D_CMD(DebugCrash) {
    DENG2_UNUSED3(src, argv, argc);
    int* ptr = (int*) 0x123;
    *ptr = 0;
    return true;
}
D_CMD(HelpWhat);
D_CMD(HelpApropos);
D_CMD(ListAliases);
D_CMD(ListCmds);
D_CMD(ListVars);
#if _DEBUG
D_CMD(PrintVarStats);
#endif
static bool inited;
D_CMD(ListCmds);
D_CMD(HelpApropos);
void Con_DataRegister() {
    C_CMD("apropos",        "s",    HelpApropos);
    C_CMD("listaliases",    NULL,   ListAliases);
    C_CMD("listcmds",       NULL,   ListCmds);
    C_CMD("listvars",       NULL,   ListVars);
#ifdef DENG_DEBUG
    C_CMD("varstats",       NULL,   PrintVarStats);
#endif
}
void Con_InitDatabases(void) {
    if (inited) return;
    Con_InitVariableDirectory();
    Con_InitCommands();
    Con_InitAliases();
    Con_ClearKnownWords();
    inited = true;
}
void Con_ClearDatabases(void) {
    if (!inited) return;
    Con_ClearKnownWords();
    Con_ClearAliases();
    Con_ClearCommands();
    Con_ClearVariables();
}
void Con_ShutdownDatabases(void) {
    if (!inited) return;
    Con_ClearDatabases();
    Con_DeinitVariableDirectory();
    inited = false;
}
String Con_GameAsStyledText(Game const *game) {
    DENG2_ASSERT(game != 0);
    return String(_E(1)) + game->id() + _E(.);
}
static int printKnownWordWorker(knownword_t const *word, void *parameters) {
    DENG_ASSERT(word);
    uint *numPrinted = (uint *) parameters;
    switch (word->type) {
    case WT_CCMD: {
        ccmd_t *ccmd = (ccmd_t *) word->data;
        if (ccmd->prevOverload) {
            return 0; // Skip overloaded variants.
        }
        LOG_SCR_MSG("%s") << Con_CmdAsStyledText(ccmd);
        break; }
    case WT_CVAR: {
        cvar_t *cvar = (cvar_t *) word->data;
        if (cvar->flags & CVF_HIDE) {
            return 0; // Skip hidden variables.
        }
        Con_PrintCVar(cvar, "");
        break; }
    case WT_CALIAS:
        LOG_SCR_MSG("%s") << Con_AliasAsStyledText((calias_t *) word->data);
        break;
    case WT_GAME:
        LOG_SCR_MSG("%s") << Con_GameAsStyledText((Game const *) word->data);
        break;
    default:
        DENG_ASSERT(false);
        break;
    }
    if (numPrinted) ++(*numPrinted);
    return 0; // Continue iteration.
}
D_CMD(ListVars) {
    DENG_UNUSED(src);
    uint numPrinted = 0;
    LOG_SCR_MSG(_E(b) "Console variables:");
    Con_IterateKnownWords(argc > 1? argv[1] : 0, WT_CVAR, printKnownWordWorker, &numPrinted);
    LOG_SCR_MSG("Found %i console variables") << numPrinted;
    return true;
}
D_CMD(ListCmds) {
    DENG_UNUSED(src);
    LOG_SCR_MSG(_E(b) "Console commands:");
    uint numPrinted = 0;
    Con_IterateKnownWords(argc > 1? argv[1] : 0, WT_CCMD, printKnownWordWorker, &numPrinted);
    LOG_SCR_MSG("Found %i console commands") << numPrinted;
    return true;
}
D_CMD(ListAliases) {
    DENG_UNUSED(src);
    LOG_SCR_MSG(_E(b) "Aliases:");
    uint numPrinted = 0;
    Con_IterateKnownWords(argc > 1? argv[1] : 0, WT_CALIAS, printKnownWordWorker, &numPrinted);
    LOG_SCR_MSG("Found %i aliases") << numPrinted;
    return true;
}
D_CMD(DoomsdayScript) {
    DENG_UNUSED(src);
    DENG_UNUSED(argc);
    String source;
    for (int i = 1; i < argc; ++i) {
        if (source) source += " ";
        source += String(argv[i]);
    }
    Script script(source);
    Process proc(script);
    proc.execute();
    return true;
}
static QList<knownword_t> knownWords;
static dd_bool knownWordsNeedUpdate;
static void (*appWordsCallback)();
void Con_ClearKnownWords(void) {
    knownWords.clear();
    knownWordsNeedUpdate = false;
}
static bool compareKnownWordByName(knownword_t const &a, knownword_t const &b) {
    knownword_t const *wA = &a;
    knownword_t const *wB = &b;
    AutoStr *textA = 0, *textB = 0;
    switch (wA->type) {
    case WT_CALIAS:   textA = AutoStr_FromTextStd(((calias_t *)wA->data)->name); break;
    case WT_CCMD:     textA = AutoStr_FromTextStd(((ccmd_t *)wA->data)->name); break;
    case WT_CVAR:     textA = CVar_ComposePath((cvar_t *)wA->data); break;
    case WT_GAME:     textA = AutoStr_FromTextStd(reinterpret_cast<Game const *>(wA->data)->id().toUtf8().constData()); break;
    default:
        DENG_ASSERT(!"compareKnownWordByName: Invalid type for word A");
        return false;
    }
    switch (wB->type) {
    case WT_CALIAS:   textB = AutoStr_FromTextStd(((calias_t *)wB->data)->name); break;
    case WT_CCMD:     textB = AutoStr_FromTextStd(((ccmd_t *)wB->data)->name); break;
    case WT_CVAR:     textB = CVar_ComposePath((cvar_t *)wB->data); break;
    case WT_GAME:     textB = AutoStr_FromTextStd(reinterpret_cast<Game const *>(wB->data)->id().toUtf8().constData()); break;
    default:
        DENG_ASSERT(!"compareKnownWordByName: Invalid type for word B");
        return false;
    }
    return Str_CompareIgnoreCase(textA, Str_Text(textB)) < 0;
}
/**
 * @return New AutoStr with the text of the known word. Caller gets ownership.
 */
static AutoStr *textForKnownWord(knownword_t const *word) {
    AutoStr *text = 0;
    switch (word->type) {
    case WT_CALIAS:   text = AutoStr_FromTextStd(((calias_t *)word->data)->name); break;
    case WT_CCMD:     text = AutoStr_FromTextStd(((ccmd_t *)word->data)->name); break;
    case WT_CVAR:     text = CVar_ComposePath((cvar_t *)word->data); break;
    case WT_GAME:     text = AutoStr_FromTextStd(reinterpret_cast<Game const *>(word->data)->id().toUtf8().constData()); break;
    default:
        DENG_ASSERT(!"textForKnownWord: Invalid type for word");
        text = AutoStr_FromTextStd("");
    }
    return text;
}
#if 0
static dd_bool removeFromKnownWords(knownwordtype_t type, void* data) {
    DENG_ASSERT(VALID_KNOWNWORDTYPE(type) && data != 0);
    for (int i = 0; i < knownWords.size(); ++i) {
        knownword_t const &word = knownWords.at(i);
        if (word.type != type || word.data != data)
            continue;
        knownWords.removeAt(i);
        return true;
    }
    return false;
}
#endif
void Con_AddKnownWord(knownwordtype_t type, void *ptr) {
    knownword_t word;
    word.type = type;
    word.data = ptr;
    knownWords << word;
}
void Con_UpdateKnownWords() {
    knownWordsNeedUpdate = true;
}
/**
 * Collate all known words and sort them alphabetically.
 * Commands, variables (except those hidden) and aliases are known words.
 */
static void updateKnownWords(void) {
    if (!knownWordsNeedUpdate) return;
    /*
    countvariableparams_t countCVarParams;
    countCVarParams.count = 0;
    countCVarParams.type = cvartype_t(-1);
    countCVarParams.hidden = false;
    countCVarParams.ignoreHidden = true;
    if (cvarDirectory) {
        cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash, countVariable, &countCVarParams);
    }*/
    /*numKnownWords =
            numUniqueNamedCCmds +
            countCVarParams.count +
            numCAliases +
            App_Games().count();
    size_t len = sizeof(knownword_t) * numKnownWords;
    knownWords = (knownword_t*) M_Realloc(knownWords, len);
    memset(knownWords, 0, len);
*/
    knownWords.clear();
    Con_AddKnownWordsForCommands();
    Con_AddKnownWordsForVariables();
    Con_AddKnownWordsForAliases();
    if (appWordsCallback) {
        appWordsCallback();
    }
    qSort(knownWords.begin(), knownWords.end(), compareKnownWordByName);
    knownWordsNeedUpdate = false;
}
AutoStr *Con_KnownWordToString(knownword_t const *word) {
    return textForKnownWord(word);
}
int Con_IterateKnownWords(char const *pattern, knownwordtype_t type,
                          int (*callback)(knownword_t const *word, void *parameters),
                          void *parameters) {
    return Con_IterateKnownWords(KnownWordStartsWith, pattern, type, callback, parameters);
}
int Con_IterateKnownWords(KnownWordMatchMode matchMode,
                          char const* pattern, knownwordtype_t type,
                          int (*callback)(knownword_t const* word, void* parameters),
                          void* parameters) {
    DENG_ASSERT(callback);
    knownwordtype_t matchType = (VALID_KNOWNWORDTYPE(type)? type : WT_ANY);
    size_t patternLength = (pattern? strlen(pattern) : 0);
    int result = 0;
    QRegExp const regex(matchMode == KnownWordRegex? pattern : "", Qt::CaseInsensitive);
    updateKnownWords();
    for (int i = 0; i < knownWords.size(); ++i) {
        knownword_t const *word = &knownWords.at(i);
        if (matchType != WT_ANY && word->type != type) continue;
        if (patternLength) {
            AutoStr* textString = textForKnownWord(word);
            if (matchMode == KnownWordStartsWith) {
                if (qstrnicmp(Str_Text(textString), pattern, patternLength))
                    continue; // Didn't match.
            }
            else if (matchMode == KnownWordExactMatch) {
                if (strcasecmp(Str_Text(textString), pattern))
                    continue; // Didn't match.
            }
            else if (matchMode == KnownWordRegex) {
                if (!regex.exactMatch(Str_Text(textString)))
                    continue; // Didn't match.
            }
        }
        result = callback(word, parameters);
        if (result) break;
    }
    return result;
}
static int countMatchedWordWorker(knownword_t const* /*word*/, void* parameters)
{
    DENG_ASSERT(parameters);
    uint* count = (uint*) parameters;
    ++(*count);
    return 0; // Continue iteration.
}
typedef struct {
    knownword_t const** matches; /// Matched word array.
    uint index; /// Current position in the collection.
} collectmatchedwordworker_paramaters_t;
static int collectMatchedWordWorker(knownword_t const* word, void* parameters) {
    DENG_ASSERT(word && parameters);
    collectmatchedwordworker_paramaters_t* p = (collectmatchedwordworker_paramaters_t*) parameters;
    p->matches[p->index++] = word;
    return 0; // Continue iteration.
}
knownword_t const** Con_CollectKnownWordsMatchingWord(char const* word,
    knownwordtype_t type, uint* count) {
    uint localCount = 0;
    Con_IterateKnownWords(word, type, countMatchedWordWorker, &localCount);
    if (count) *count = localCount;
    if (localCount != 0) {
        collectmatchedwordworker_paramaters_t p;
        p.matches = (knownword_t const**) M_Malloc(sizeof(*p.matches) * (localCount + 1));
        p.index = 0;
        Con_IterateKnownWords(word, type, collectMatchedWordWorker, &p);
        p.matches[localCount] = 0; // Terminate.
        return p.matches;
    }
    return 0; // No matches.
}
static int aproposPrinter(knownword_t const *word, void *matching) {
    AutoStr *text = textForKnownWord(word);
    if (strcasestr(Str_Text(text), (const char*)matching)) {
        char const* wType[KNOWNWORDTYPE_COUNT] = {
            "cmd ", "var ", "alias ", "game "
        };
        String str;
        QTextStream os(&str);
        os << _E(l) << wType[word->type]
           << _E(0) << _E(b) << Str_Text(text) << " " << _E(2) << _E(>);
        String tmp;
        if (word->type == WT_CCMD || word->type == WT_CVAR) {
            char const *desc = DH_GetString(DH_Find(Str_Text(text)), HST_DESCRIPTION);
            if (desc) {
                tmp = desc;
            }
        }
        else if (word->type == WT_GAME) {
            tmp = reinterpret_cast<Game const *>(word->data)->title();
        }
        os << tmp;
        LOG_SCR_MSG("%s") << str;
    }
    return 0;
}
static void printApropos(char const *matching) {
    Con_IterateKnownWords(0, WT_ANY, aproposPrinter, (void *)matching);
}
D_CMD(HelpApropos) {
    DENG2_UNUSED2(argc, src);
    printApropos(argv[1]);
    return true;
}
struct AnnotationWork {
    QSet<QString> terms;
    de::String result;
};
static int annotateMatchedWordCallback(knownword_t const *word, void *parameters) {
    AnnotationWork *work = reinterpret_cast<AnnotationWork *>(parameters);
    AutoStr *name = Con_KnownWordToString(word);
    de::String found;
    if (!work->terms.contains(Str_Text(name)))
        return false; // keep going
    switch (word->type) {
    case WT_CVAR:
        if (!(((cvar_t *)word->data)->flags & CVF_HIDE)) {
            found = Con_VarAsStyledText((cvar_t *) word->data, "");
        }
        break;
    case WT_CCMD:
        if (!((ccmd_t *)word->data)->prevOverload) {
            found = Con_CmdAsStyledText((ccmd_t *) word->data);
        }
        break;
    case WT_CALIAS:
        found = Con_AliasAsStyledText((calias_t *) word->data);
        break;
    case WT_GAME:
        found = Con_GameAsStyledText(reinterpret_cast<Game const *>(word->data));
        break;
    default:
        break;
    }
    if (!found.isEmpty()) {
        if (!work->result.isEmpty()) work->result.append("\n");
        work->result.append(found);
    }
    return false; // don't stop
}
de::String Con_AnnotatedConsoleTerms(QStringList terms) {
    AnnotationWork work;
    foreach (QString term, terms) {
        work.terms.insert(term);
    }
    Con_IterateKnownWords(NULL, WT_ANY, annotateMatchedWordCallback, &work);
    return work.result;
}
static int addToTerms(knownword_t const *word, void *parameters) {
    shell::Lexicon *lexi = reinterpret_cast<shell::Lexicon *>(parameters);
    lexi->addTerm(Str_Text(Con_KnownWordToString(word)));
    return 0;
}
shell::Lexicon Con_Lexicon() {
    shell::Lexicon lexi;
    Con_IterateKnownWords(0, WT_ANY, addToTerms, &lexi);
    lexi.setAdditionalWordChars("-_.");
    return lexi;
}
void Con_SetApplicationKnownWordCallback(void (*callback)()) {
    appWordsCallback = callback;
}
static int addToStringList(knownword_t const *word, void *parameters) {
    StringList *terms = reinterpret_cast<StringList *>(parameters);
    terms->append(Str_Text(Con_KnownWordToString(word)));
    return 0;
}
void Con_TermsRegex(StringList &terms, String const &pattern, knownwordtype_t wordType) {
    terms.clear();
    Con_IterateKnownWords(KnownWordRegex, pattern.toUtf8(), wordType,
                          addToStringList, &terms);
}
#define CVARDIRECTORY_DELIMITER         '-'
typedef UserDataPathTree CVarDirectory;
static CVarDirectory *cvarDirectory;
static ddstring_s *emptyStr;
static de::Uri *emptyUri;
void Con_InitVariableDirectory() {
    cvarDirectory = new CVarDirectory;
    emptyStr = Str_NewStd();
    emptyUri = new de::Uri;
}
void Con_DeinitVariableDirectory() {
    delete cvarDirectory; cvarDirectory = 0;
    Str_Delete(emptyStr); emptyStr = 0;
    delete emptyUri; emptyUri = 0;
}
static int markVariableUserDataFreed(CVarDirectory::Node &node, void *context) {
    DENG_ASSERT(context);
    cvar_t *var = reinterpret_cast<cvar_t *>(node.userPointer());
    void **ptr = (void **) context;
    if (var)
    switch (CVar_Type(var)) {
    case CVT_CHARPTR:
        if (*ptr == CV_CHARPTR(var)) var->flags &= ~CVF_CAN_FREE;
        break;
    case CVT_URIPTR:
        if (*ptr == CV_URIPTR(var)) var->flags &= ~CVF_CAN_FREE;
        break;
    default: break;
    }
    return 0; // Continue iteration.
}
static int clearVariable(CVarDirectory::Node& node, void * /*context*/)
{
    cvar_t *var = reinterpret_cast<cvar_t*>(node.userPointer());
    if (var) {
        node.setUserPointer(0);
        if (CVar_Flags(var) & CVF_CAN_FREE) {
            void** ptr = NULL;
            switch (CVar_Type(var)) {
            case CVT_CHARPTR:
                if (!CV_CHARPTR(var)) break;
                ptr = (void**)var->ptr;
                cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash, markVariableUserDataFreed, ptr);
                M_Free(*ptr); *ptr = Str_Text(emptyStr);
                break;
            case CVT_URIPTR:
                if (!CV_URIPTR(var)) break;
                ptr = (void**)var->ptr;
                cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash, markVariableUserDataFreed, ptr);
                delete reinterpret_cast<de::Uri *>(*ptr); *ptr = emptyUri;
                break;
            default:
                LOGDEV_SCR_WARNING("Attempt to free user data for non-pointer type variable %s [%p]")
                        << Str_Text(CVar_ComposePath(var)) << var;
                break;
            }
        }
        M_Free(var);
    }
    return 0; // Continue iteration.
}
void Con_ClearVariables() {
#if _DEBUG
    PathTree::ComparisonFlags flags;
#else
    PathTree::ComparisonFlags flags = PathTree::NoBranch;
#endif
    if (!cvarDirectory) return;
    cvarDirectory->traverse(flags, NULL, CVarDirectory::no_hash, clearVariable);
    cvarDirectory->clear();
}
static cvar_t* addVariable(cvartemplate_t const& tpl) {
    Path path(tpl.path, CVARDIRECTORY_DELIMITER);
    CVarDirectory::Node* node = &cvarDirectory->insert(path);
    cvar_t* newVar;
    DENG_ASSERT(!node->userPointer());
    if (node->userPointer()) {
        throw Error("Con_AddVariable", "A variable with path '" + String(tpl.path) + "' is already known!");
    }
    newVar = (cvar_t*) M_Malloc(sizeof *newVar);
    newVar->flags = tpl.flags;
    newVar->type = tpl.type;
    newVar->ptr = tpl.ptr;
    newVar->min = tpl.min;
    newVar->max = tpl.max;
    newVar->notifyChanged = tpl.notifyChanged;
    newVar->directoryNode = node;
    node->setUserPointer(newVar);
    Con_UpdateKnownWords();
    return newVar;
}
String CVar_TypeAsText(cvar_t const *var) {
    DENG_ASSERT(var);
    switch (var->type) {
    case CVT_BYTE:
        return "byte";
    case CVT_CHARPTR:
        return "text";
    case CVT_FLOAT:
        return "float";
    case CVT_INT:
        return "integer";
    case CVT_NULL:
        return "null";
    case CVT_URIPTR:
        return "uri";
    default:
        DENG_ASSERT(!"Con_VarTypeAsText: Unknown variable type");
        break;
    }
    return "";
}
template <typename ValueType>
void printTypeWarning(cvar_t const *var, String const &attemptedType, ValueType value) {
    AutoStr* path = CVar_ComposePath(var);
    LOG_SCR_WARNING("Variable %s (of type '%s') is incompatible with %s ")
            << Str_Text(path) << CVar_TypeAsText(var)
            << attemptedType << value;
}
void CVar_PrintReadOnlyWarning(cvar_t const *var) {
    AutoStr* path = CVar_ComposePath(var);
    LOG_SCR_WARNING("%s (%s cvar) is read-only; it cannot be changed (even with force)")
            << CVar_TypeAsText(var) << Str_Text(path);
}
ddstring_t const* CVar_TypeName(cvartype_t type) {
    static de::Str const names[CVARTYPE_COUNT] = {
        "invalid",
        "CVT_BYTE",
        "CVT_INT",
        "CVT_FLOAT",
        "CVT_CHARPTR",
        "CVT_URIPTR"
    };
    return names[(VALID_CVARTYPE(type)? type : 0)];
}
cvartype_t CVar_Type(cvar_t const* var) {
    DENG_ASSERT(var);
    return var->type;
}
int CVar_Flags(const cvar_t* var) {
    DENG_ASSERT(var);
    return var->flags;
}
AutoStr *CVar_ComposePath(cvar_t const *var) {
    DENG_ASSERT(var != 0);
    CVarDirectory::Node &node = *reinterpret_cast<CVarDirectory::Node *>(var->directoryNode);
    QByteArray path = node.path(CVARDIRECTORY_DELIMITER).toUtf8();
    return AutoStr_FromTextStd(path.constData());
}
void CVar_SetUri2(cvar_t *var, de::Uri const &uri, int svFlags) {
    DENG_ASSERT(var);
    de::Uri *newUri;
    bool changed = false;
    if ((var->flags & CVF_READ_ONLY) && !(svFlags & SVF_WRITE_OVERRIDE)) {
        CVar_PrintReadOnlyWarning(var);
        return;
    }
    if (var->type != CVT_URIPTR) {
        App_FatalError("CVar::SetUri: Not of type %s.", Str_Text(CVar_TypeName(CVT_URIPTR)));
        return; // Unreachable.
    }
    /*
    if (!CV_URIPTR(var) && !uri) {
        return;
    }
    */
    newUri = new de::Uri(uri);
    if (!CV_URIPTR(var) || *CV_URIPTR(var) != *newUri) {
        changed = true;
    }
    if ((var->flags & CVF_CAN_FREE) && CV_URIPTR(var)) {
        delete CV_URIPTR(var);
    }
    var->flags |= CVF_CAN_FREE;
    CV_URIPTR(var) = newUri;
    if (var->notifyChanged && changed) {
        var->notifyChanged();
    }
}
void CVar_SetUri(cvar_t *var, de::Uri const &uri) {
    CVar_SetUri2(var, uri, 0);
}
void CVar_SetString2(cvar_t *var, char const *text, int svFlags) {
    DENG_ASSERT(var != 0);
    bool changed = false;
    size_t oldLen, newLen;
    if ((var->flags & CVF_READ_ONLY) && !(svFlags & SVF_WRITE_OVERRIDE)) {
        CVar_PrintReadOnlyWarning(var);
        return;
    }
    if (var->type != CVT_CHARPTR) {
        printTypeWarning(var, "text", text);
        return;
    }
    oldLen = (!CV_CHARPTR(var)? 0 : strlen(CV_CHARPTR(var)));
    newLen = (!text           ? 0 : strlen(text));
    if (oldLen == 0 && newLen == 0)
        return;
    if (oldLen != newLen || qstricmp(text, CV_CHARPTR(var)))
        changed = true;
    if ((var->flags & CVF_CAN_FREE) && CV_CHARPTR(var))
        free(CV_CHARPTR(var));
    var->flags |= CVF_CAN_FREE;
    CV_CHARPTR(var) = (char*) M_Malloc(newLen + 1);
    qstrcpy(CV_CHARPTR(var), text);
    if (var->notifyChanged != NULL && changed)
        var->notifyChanged();
}
void CVar_SetString(cvar_t* var, char const* text) {
    CVar_SetString2(var, text, 0);
}
void CVar_SetInteger2(cvar_t* var, int value, int svFlags) {
    DENG_ASSERT(var);
    bool changed = false;
    if ((var->flags & CVF_READ_ONLY) && !(svFlags & SVF_WRITE_OVERRIDE)) {
        CVar_PrintReadOnlyWarning(var);
        return;
    }
    switch (var->type) {
    case CVT_INT:
        if (CV_INT(var) != value)
            changed = true;
        CV_INT(var) = value;
        break;
    case CVT_BYTE:
        if (CV_BYTE(var) != (byte) value)
            changed = true;
        CV_BYTE(var) = (byte) value;
        break;
    case CVT_FLOAT:
        if (CV_FLOAT(var) != (float) value)
            changed = true;
        CV_FLOAT(var) = (float) value;
        break;
    default:
        printTypeWarning(var, "integer", value);
        return;
    }
    if (var->notifyChanged != 0 && changed)
        var->notifyChanged();
}
void CVar_SetInteger(cvar_t* var, int value) {
    CVar_SetInteger2(var, value, 0);
}
void CVar_SetFloat2(cvar_t* var, float value, int svFlags) {
    DENG_ASSERT(var);
    bool changed = false;
    LOG_AS("CVar_SetFloat2");
    if ((var->flags & CVF_READ_ONLY) && !(svFlags & SVF_WRITE_OVERRIDE)) {
        CVar_PrintReadOnlyWarning(var);
        return;
    }
    switch (var->type) {
    case CVT_INT:
        if (CV_INT(var) != (int) value)
            changed = true;
        CV_INT(var) = (int) value;
        break;
    case CVT_BYTE:
        if (CV_BYTE(var) != (byte) value)
            changed = true;
        CV_BYTE(var) = (byte) value;
        break;
    case CVT_FLOAT:
        if (CV_FLOAT(var) != value)
            changed = true;
        CV_FLOAT(var) = value;
        break;
    default:
        printTypeWarning(var, "float", value);
        return;
    }
    if (var->notifyChanged != 0 && changed)
        var->notifyChanged();
}
void CVar_SetFloat(cvar_t* var, float value) {
    CVar_SetFloat2(var, value, 0);
}
static void printConversionWarning(cvar_t const *var) {
    AutoStr* path = CVar_ComposePath(var);
    LOGDEV_SCR_WARNING("Incompatible variable %s [%p type:%s]")
            << Str_Text(path) << var << Str_Text(CVar_TypeName(CVar_Type(var)));
}
int CVar_Integer(cvar_t const* var) {
    DENG_ASSERT(var);
    switch (var->type) {
    case CVT_BYTE:      return CV_BYTE(var);
    case CVT_INT:       return CV_INT(var);
    case CVT_FLOAT:     return CV_FLOAT(var);
    case CVT_CHARPTR:   return strtol(CV_CHARPTR(var), 0, 0);
    default: {
        LOG_AS("CVar_Integer");
        printConversionWarning(var);
        return 0; }
    }
}
float CVar_Float(cvar_t const* var) {
    DENG_ASSERT(var);
    switch (var->type) {
    case CVT_BYTE:      return CV_BYTE(var);
    case CVT_INT:       return CV_INT(var);
    case CVT_FLOAT:     return CV_FLOAT(var);
    case CVT_CHARPTR:   return strtod(CV_CHARPTR(var), 0);
    default: {
        LOG_AS("CVar_Float");
        printConversionWarning(var);
        return 0; }
    }
}
byte CVar_Byte(cvar_t const* var) {
    DENG_ASSERT(var);
    switch (var->type) {
    case CVT_BYTE:      return CV_BYTE(var);
    case CVT_INT:       return CV_INT(var);
    case CVT_FLOAT:     return CV_FLOAT(var);
    case CVT_CHARPTR:   return strtol(CV_CHARPTR(var), 0, 0);
    default: {
        LOG_AS("CVar_Byte");
        printConversionWarning(var);
        return 0; }
    }
}
char const* CVar_String(cvar_t const* var) {
    DENG_ASSERT(var);
    switch (var->type) {
    case CVT_CHARPTR:   return CV_CHARPTR(var);
    default: {
        LOG_AS("CVar_String");
        printConversionWarning(var);
        return Str_Text(emptyStr); }
    }
}
de::Uri const &CVar_Uri(cvar_t const *var) {
    if (!var) return *emptyUri;
    switch (var->type) {
    case CVT_URIPTR:   return *CV_URIPTR(var);
    default: {
        LOG_AS("CVar_Uri");
        printConversionWarning(var);
        return *emptyUri; }
    }
}
void Con_AddVariable(cvartemplate_t const *tpl) {
    LOG_AS("Con_AddVariable");
    if (!tpl) return;
    if (CVT_NULL == tpl->type) {
        LOGDEV_SCR_WARNING("Ignored attempt to register variable '%s' as type %s")
            << tpl->path << Str_Text(CVar_TypeName(CVT_NULL));
        return;
    }
    addVariable(*tpl);
}
void Con_AddVariableList(cvartemplate_t const *tplList) {
    if (!tplList) return;
    for (; tplList->path; ++tplList) {
        if (Con_FindVariable(tplList->path)) {
            App_FatalError("Console variable with the name '%s' is already registered", tplList->path);
        }
        addVariable(*tplList);
    }
}
cvar_t *Con_FindVariable(Path const &path) {
    if (CVarDirectory::Node const *node =
            cvarDirectory->tryFind(path, PathTree::NoBranch | PathTree::MatchFull)) {
        return reinterpret_cast<cvar_t *>(node->userPointer());
    }
    return nullptr;
}
cvar_t *Con_FindVariable(char const *path) {
    return Con_FindVariable(Path(path, CVARDIRECTORY_DELIMITER));
}
String Con_VarAsStyledText(cvar_t *var, char const *prefix) {
    if (!var) return "";
    char equals = '=';
    if ((var->flags & CVF_PROTECTED) || (var->flags & CVF_READ_ONLY))
        equals = ':';
    String str;
    QTextStream os(&str);
    if (prefix) os << prefix;
    AutoStr* path = CVar_ComposePath(var);
    os << _E(b) << Str_Text(path) << _E(.) << " " << equals << " " << _E(>);
    switch (var->type) {
    case CVT_BYTE:      os << CV_BYTE(var); break;
    case CVT_INT:       os << CV_INT(var); break;
    case CVT_FLOAT:     os << CV_FLOAT(var); break;
    case CVT_CHARPTR:   os << "\"" << CV_CHARPTR(var) << "\""; break;
    case CVT_URIPTR: {
        os << "\"" << (CV_URIPTR(var)? CV_URIPTR(var)->asText() : "") << "\"";
        break; }
    default:
        DENG_ASSERT(false);
        break;
    }
    os << _E(<);
    return str;
}
void Con_PrintCVar(cvar_t* var, char const *prefix) {
    LOG_SCR_MSG("%s") << Con_VarAsStyledText(var, prefix);
}
static int addVariableToKnownWords(CVarDirectory::Node& node, void* /*parameters*/)
{
    cvar_t* var = reinterpret_cast<cvar_t*>( node.userPointer() );
    if (var && !(var->flags & CVF_HIDE)) {
        Con_AddKnownWord(WT_CVAR, var);
    }
    return 0; // Continue iteration.
}
void Con_AddKnownWordsForVariables() {
    if (!cvarDirectory) return;
    cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash,
                            addVariableToKnownWords);
}
void Con_SetVariable(Path const &varPath, int value, int svFlags) {
    if (cvar_t *var = Con_FindVariable(varPath)) {
        CVar_SetInteger2(var, value, svFlags);
    }
    else {
        DENG2_ASSERT(!"Con_SetVariable: Invalid console variable path");
    }
}
int Con_GetVariableInteger(Path const &varPath) {
    if (cvar_t *var = Con_FindVariable(varPath)) {
        return CVar_Integer(var);
    }
    return 0;
}
static Value *Function_Console_Get(Context &, Function::ArgumentValues const &args) {
    String const name = args.at(0)->asText();
    cvar_t *var = Con_FindVariable(name.toUtf8());
    if (!var) {
        throw Error("Function_Console_Get",
                    QString("Unknown console variable: %1").arg(name));
    }
    switch (var->type) {
    case CVT_BYTE:
        return new NumberValue(CVar_Byte(var));
    case CVT_INT:
        return new NumberValue(CVar_Integer(var));
    case CVT_FLOAT:
        return new NumberValue(CVar_Float(var));
    case CVT_CHARPTR:  ///< ptr points to a char*, which points to the string.
        return new TextValue(CVar_String(var));
    case CVT_URIPTR:
        return new TextValue(CVar_Uri(var).asText());
    default:
        break;
    }
    return nullptr;
}
static Value *Function_Console_Set(Context &, Function::ArgumentValues const &args) {
    String const name = args.at(0)->asText();
    cvar_t *var = Con_FindVariable(name.toUtf8());
    if (!var) {
        throw Error("Function_Console_Set",
                    QString("Unknown console variable: %1").arg(name));
    }
    Value const &value = *args.at(1);
    switch (var->type) {
    case CVT_BYTE:
    case CVT_INT:
        CVar_SetInteger(var, value.asInt());
        break;
    case CVT_FLOAT:
        CVar_SetFloat(var, value.asNumber());
        break;
    case CVT_CHARPTR:  ///< ptr points to a char*, which points to the string.
        CVar_SetString(var, value.asText().toUtf8());
        break;
    case CVT_URIPTR:
        CVar_SetUri(var, de::Uri(value.asText()));
        break;
    default:
        break;
    }
    return nullptr;
}
void initVariableBindings(Binder &binder) {
    binder
        << DENG2_FUNC(Console_Get, "get", "name")
        << DENG2_FUNC(Console_Set, "set", "name" << "value");
}
#ifdef DENG_DEBUG
typedef struct {
    uint count;
    cvartype_t type;
    dd_bool hidden;
    dd_bool ignoreHidden;
} countvariableparams_t;
static int countVariable(CVarDirectory::Node& node, void* parameters) {
    DENG_ASSERT(parameters);
    countvariableparams_t* p = (countvariableparams_t*) parameters;
    cvar_t* var = reinterpret_cast<cvar_t*>( node.userPointer() );
    if (!(p->ignoreHidden && (var->flags & CVF_HIDE))) {
        if (!VALID_CVARTYPE(p->type) && !p->hidden) {
            if (!p->ignoreHidden || !(var->flags & CVF_HIDE))
                ++(p->count);
        }
        else if ((p->hidden && (var->flags & CVF_HIDE)) ||
                (VALID_CVARTYPE(p->type) && p->type == CVar_Type(var))) {
            ++(p->count);
        }
    }
    return 0; // Continue iteration.
}
D_CMD(PrintVarStats) {
    DENG2_UNUSED3(src, argc, argv);
    uint numCVars = 0, numCVarsHidden = 0;
    LOG_SCR_MSG(_E(b) "Console Variable Statistics:");
    if (cvarDirectory) {
        countvariableparams_t p;
        p.hidden = false;
        p.ignoreHidden = false;
        for (uint i = uint(CVT_BYTE); i < uint(CVARTYPE_COUNT); ++i) {
            p.count = 0;
            p.type = cvartype_t(i);
            cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash, countVariable, &p);
            LOGDEV_SCR_MSG("%12s: %i") << Str_Text(CVar_TypeName(p.type)) << p.count;
        }
        p.count = 0;
        p.type = cvartype_t(-1);
        p.hidden = true;
        cvarDirectory->traverse(PathTree::NoBranch, NULL, CVarDirectory::no_hash, countVariable, &p);
        numCVars = cvarDirectory->size();
        numCVarsHidden = p.count;
    }
    LOG_SCR_MSG("       Total: %i\n      Hidden: %i") << numCVars << numCVarsHidden;
    if (cvarDirectory) {
        cvarDirectory->debugPrintHashDistribution();
        cvarDirectory->debugPrint(CVARDIRECTORY_DELIMITER);
    }
    return true;
}
#endif
