﻿/** @file decoration.cpp  Decoration definition accessor.
 *
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/decoration.h"
#include <de/Record>
#include <de/RecordValue>
#include "doomsday/defs/ded.h"
#include "doomsday/defs/material.h"
/** @file ded.cpp  Doomsday Engine Definition database.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/ded.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <de/memory.h>
#include <de/strutil.h>
#include <de/ArrayValue>
#include <de/NumberValue>
#include <de/RecordValue>
#include "doomsday/defs/decoration.h"
#include "doomsday/defs/episode.h"
#include "doomsday/defs/thing.h"
#include "doomsday/defs/state.h"
#include "doomsday/defs/finale.h"
#include "doomsday/defs/mapinfo.h"
#include "doomsday/defs/material.h"
#include "doomsday/defs/model.h"
#include "doomsday/defs/music.h"
#include "doomsday/defs/sky.h"
/** @file dedfile.cpp
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/dedfile.h"
#include <de/App>
#include <de/Folder>
#include <de/LogBuffer>
#include "doomsday/defs/dedparser.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include <de/c_wrapper.h>
/** @file dedparser.cpp  Doomsday Engine Definition File Reader.
 *
 * A GHASTLY MESS!!! This should be rewritten.
 *
 * You have been warned!
 *
 * At the moment the idea is that a lot of macros are used to read a more
 * or less fixed structure of definitions, and if an error occurs then
 * "goto out_of_here;". It leads to a lot more code than an elegant parser
 * would require.
 *
 * The current implementation of the reader is a "structural" approach:
 * the definition file is parsed based on the structure implied by
 * the read tokens. A true parser would have syntax definitions for
 * a bunch of tokens, and the same parsing rules would be applied for
 * everything.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2010-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#define DENG_NO_API_MACROS_URI
#include "doomsday/defs/dedparser.h"
#include "doomsday/doomsdayapp.h"
#include "doomsday/game.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <de/c_wrapper.h>
#include <de/memory.h>
#include <de/vector1.h>
#include <de/App>
#include <de/ArrayValue>
#include <de/NativePath>
#include <de/RecordValue>
#include <de/TextValue>
#include "doomsday/defs/decoration.h"
#include "doomsday/defs/ded.h"
#include "doomsday/defs/dedfile.h"
#include "doomsday/defs/episode.h"
#include "doomsday/defs/finale.h"
#include "doomsday/defs/mapgraphnode.h"
#include "doomsday/defs/mapinfo.h"
#include "doomsday/defs/material.h"
#include "doomsday/defs/model.h"
#include "doomsday/defs/music.h"
#include "doomsday/defs/sky.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/uri.h"
#include "doomsday/world/xgclass.h"
#ifdef WIN32
#  define stricmp _stricmp
#endif
/** @file dedregister.h  Register of definitions.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/dedregister.h"
#include "doomsday/defs/definition.h"
#include <de/ArrayValue>
#include <de/DictionaryValue>
#include <de/TextValue>
#include <de/RecordValue>
#include <QSet>
#include <QMap>
/** @file definition.cpp  Base class for definition record accessors.
 *
 * @authors Copyright © 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/definition.h"
#include <de/Record>
/** @file episode.cpp  Episode definition accessor.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/episode.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file finale.cpp  Finale definition accessor.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/finale.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file mapgraphnode.cpp  MapGraphNode definition accessor.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/mapgraphnode.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file defs/mapinfo.cpp  MapInfo definition accessor.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/mapinfo.h"
#include "doomsday/defs/sky.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file material.cpp  Material definition accessor.
 *
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/material.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file defs/model.cpp  Model definition accessor.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/model.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file defs/music.cpp  Music definition accessor.
 *
 * @authors Copyright © 2014-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/music.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file defs/sky.cpp  Sky definition accessor.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/sky.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file sprite.cpp  Sprite definition accessor.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/defs/sprite.h"
#include "doomsday/defs/ded.h"
#include "doomsday/UriValue"
#include <de/Record>
#include <de/RecordValue>
#include <de/DictionaryValue>
/** @file state.cpp  Mobj state definition.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/state.h"
#include "doomsday/defs/ded.h"
#include <de/Record>
#include <de/RecordValue>
/** @file thing.cpp  Thing definition.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/defs/thing.h"
#include "doomsday/defs/ded.h"
#include <de/TextValue>
using namespace de;
namespace defn {
static String const VAR_TEXTURE = "texture";
static String const VAR_FLAGS   = "flags";
static String const VAR_LIGHT   = "light";
void Decoration::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_TEXTURE, "");  // URI. Unknown.
    def().addNumber(VAR_FLAGS, 0);
    def().addArray (VAR_LIGHT, new ArrayValue);
}
Record &Decoration::addLight() {
    auto *decor = new Record;
    MaterialDecoration(*decor).resetToDefaults();
    def()[VAR_LIGHT].array().add(new RecordValue(decor, RecordValue::OwnsRecord));
    return *decor;
}
int Decoration::lightCount() const {
    return int(geta(VAR_LIGHT).size());
}
bool Decoration::hasLight(int index) const {
    return index >= 0 && index < lightCount();
}
Record &Decoration::light(int index) {
    return *def().geta(VAR_LIGHT)[index].as<RecordValue>().record();
}
Record const &Decoration::light(int index) const {
    return *geta(VAR_LIGHT)[index].as<RecordValue>().record();
}
} // namespace defn
float ded_ptcstage_t::particleRadius(int ptcIDX) const {
    if (radiusVariance != 0.f) {
        static float const rnd[16] = { .875f, .125f, .3125f, .75f, .5f, .375f,
            .5625f, .0625f, 1, .6875f, .625f, .4375f, .8125f, .1875f,
            .9375f, .25f
        };
        return (rnd[ptcIDX & 0xf] * radiusVariance + (1 - radiusVariance)) * radius;
    }
    return radius;
}
ded_s::ded_s()
    : flags      (names.addSubrecord("flags"))
    , episodes   (names.addSubrecord("episodes"))
    , things     (names.addSubrecord("things"))
    , states     (names.addSubrecord("states"))
    , materials  (names.addSubrecord("materials"))
    , models     (names.addSubrecord("models"))
    , skies      (names.addSubrecord("skies"))
    , musics     (names.addSubrecord("musics"))
    , mapInfos   (names.addSubrecord("mapInfos"))
    , finales    (names.addSubrecord("finales"))
    , decorations(names.addSubrecord("decorations")) {
    decorations.addLookupKey("texture");
    episodes.addLookupKey(defn::Definition::VAR_ID);
    things.addLookupKey(defn::Definition::VAR_ID, DEDRegister::OnlyFirst);
    things.addLookupKey("name");
    states.addLookupKey(defn::Definition::VAR_ID, DEDRegister::OnlyFirst);
    finales.addLookupKey(defn::Definition::VAR_ID);
    finales.addLookupKey("before");
    finales.addLookupKey("after");
    flags.addLookupKey(defn::Definition::VAR_ID);
    mapInfos.addLookupKey(defn::Definition::VAR_ID);
    materials.addLookupKey(defn::Definition::VAR_ID);
    models.addLookupKey(defn::Definition::VAR_ID, DEDRegister::OnlyFirst);
    models.addLookupKey("state");
    musics.addLookupKey(defn::Definition::VAR_ID, DEDRegister::OnlyFirst);
    skies.addLookupKey(defn::Definition::VAR_ID);
    clear();
}
void ded_s::clear() {
    release();
    version = DED_VERSION;
    modelFlags = 0;
    modelScale = 0;
    modelOffset = 0;
}
int ded_s::addFlag(String const &id, int value) {
    Record &def = flags.append();
    def.addText(defn::Definition::VAR_ID, id);
    def.addNumber("value", value);
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addEpisode() {
    Record &def = episodes.append();
    defn::Episode(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addThing(String const &id) {
    Record &def = things.append();
    defn::Thing(def).resetToDefaults();
    def.set(defn::Definition::VAR_ID, id);
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addState(String const &id) {
    Record &def = states.append();
    defn::State(def).resetToDefaults();
    def.set(defn::Definition::VAR_ID, id);
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addDecoration() {
    Record &def = decorations.append();
    defn::Decoration(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addFinale() {
    Record &def = finales.append();
    defn::Finale(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addMapInfo() {
    Record &def = mapInfos.append();
    defn::MapInfo(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addMaterial() {
    Record &def = materials.append();
    defn::Material(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addModel() {
    Record &def = models.append();
    defn::Model(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addMusic() {
    Record &def = musics.append();
    defn::Music(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
int ded_s::addSky() {
    Record &def = skies.append();
    defn::Sky(def).resetToDefaults();
    return def.geti(defn::Definition::VAR_ORDER);
}
void ded_s::release() {
    flags.clear();
    episodes.clear();
    things.clear();
    states.clear();
    sprites.clear();
    lights.clear();
    models.clear();
    sounds.clear();
    musics.clear();
    mapInfos.clear();
    skies.clear();
    details.clear();
    materials.clear();
    text.clear();
    textureEnv.clear();
    compositeFonts.clear();
    values.clear();
    decorations.clear();
    reflections.clear();
    groups.clear();
    sectorTypes.clear();
    lineTypes.clear();
    ptcGens.clear();
    finales.clear();
}
/*
int DED_AddMobj(ded_t* ded, char const* idstr) {
    ded_mobj_t *mo = ded->mobjs.append();
    strcpy(mo->id, idstr);
    return ded->mobjs.indexOf(mo);
}
 */
/*
int DED_AddState(ded_t* ded, char const* id) {
    ded_state_t *st = ded->states.append();
    strcpy(st->id, id);
    return ded->states.indexOf(st);
}
 */
int DED_AddSprite(ded_t* ded, char const* name) {
    ded_sprid_t *sp = ded->sprites.append();
    strcpy(sp->id, name);
    return ded->sprites.indexOf(sp);
}
int DED_AddLight(ded_t* ded, char const* stateid) {
    ded_light_t* light = ded->lights.append();
    strcpy(light->state, stateid);
    return ded->lights.indexOf(light);
}
int DED_AddSound(ded_t* ded, char const* id) {
    ded_sound_t* snd = ded->sounds.append();
    strcpy(snd->id, id);
    return ded->sounds.indexOf(snd);
}
int DED_AddText(ded_t* ded, char const* id) {
    ded_text_t* txt = ded->text.append();
    strcpy(txt->id, id);
    return ded->text.indexOf(txt);
}
int DED_AddTextureEnv(ded_t* ded, char const* id) {
    ded_tenviron_t* env = ded->textureEnv.append();
    strcpy(env->id, id);
    return ded->textureEnv.indexOf(env);
}
int DED_AddCompositeFont(ded_t* ded, char const *uri) {
    ded_compositefont_t* cfont = ded->compositeFonts.append();
    if (uri) cfont->uri = new de::Uri(uri, RC_NULL);
    return ded->compositeFonts.indexOf(cfont);
}
int DED_AddValue(ded_t* ded, char const* id) {
    ded_value_t* val = ded->values.append();
    if (id) {
        val->id = (char *) M_Malloc(strlen(id) + 1);
        strcpy(val->id, id);
    }
    return ded->values.indexOf(val);
}
int DED_AddDetail(ded_t *ded, char const *lumpname) {
    ded_detailtexture_t *dtl = ded->details.append();
    dtl->flags = DTLF_PWAD|DTLF_EXTERNAL;
    if (lumpname && lumpname[0]) {
        dtl->stage.texture = new de::Uri(lumpname, RC_NULL);
    }
    dtl->stage.scale = 1;
    dtl->stage.strength = 1;
    return ded->details.indexOf(dtl);
}
int DED_AddPtcGen(ded_t* ded, char const* state) {
    ded_ptcgen_t* gen = ded->ptcGens.append();
    strcpy(gen->state, state);
    gen->subModel = -1;
    return ded->ptcGens.indexOf(gen);
}
int DED_AddPtcGenStage(ded_ptcgen_t* gen) {
    ded_ptcstage_t* stage = gen->stages.append();
    stage->model = -1;
    stage->sound.volume = 1;
    stage->hitSound.volume = 1;
    return gen->stages.indexOf(stage);
}
int DED_AddReflection(ded_t *ded) {
    ded_reflection_t *ref = ded->reflections.append();
    ref->flags = REFF_PWAD|REFF_EXTERNAL;
    ref->stage.shininess  = 1.0f;
    ref->stage.blendMode  = BM_ADD;
    ref->stage.maskWidth  = 1.0f;
    ref->stage.maskHeight = 1.0f;
    return ded->reflections.indexOf(ref);
}
int DED_AddGroup(ded_t* ded) {
    ded_group_t* group = ded->groups.append();
    return ded->groups.indexOf(group);
}
int DED_AddGroupMember(ded_group_t* grp) {
    ded_group_member_t* memb = grp->members.append();
    return grp->members.indexOf(memb);
}
int DED_AddSectorType(ded_t* ded, int id) {
    ded_sectortype_t* sec = ded->sectorTypes.append();
    sec->id = id;
    return ded->sectorTypes.indexOf(sec);
}
int DED_AddLineType(ded_t* ded, int id) {
    ded_linetype_t* li = ded->lineTypes.append();
    li->id = id;
    return ded->lineTypes.indexOf(li);
}
int ded_s::getMobjNum(String const &id) const {
    if (Record const *def = things.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    /*
    for (i = 0; i < mobjs.size(); ++i)
        if (!qstricmp(mobjs[i].id, id))
            return i;*/
    return -1;
}
int ded_s::getMobjNumForName(const char *name) const {
    if (!name || !name[0])
        return -1;
    /*
    for (int i = mobjs.size() - 1; i >= 0; --i)
        if (!qstricmp(mobjs[i].name, name))
            return i;*/
    if (Record const *def = things.tryFind("name", name)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
}
String ded_s::getMobjName(int num) const {
    if (num < 0) return "(<0)";
    if (num >= things.size()) return "(>mobjtypes)";
    return things[num].gets(defn::Definition::VAR_ID);
}
int ded_s::getStateNum(String const &id) const {
    if (Record const *def = states.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
}
int ded_s::getStateNum(char const *id) const {
    return getStateNum(String(id));
}
dint ded_s::evalFlags(char const *ptr) const {
    LOG_AS("Defs::evalFlags");
    dint value = 0;
    while (*ptr) {
        ptr = M_SkipWhite(const_cast<char *>(ptr));
        dint flagNameLength = M_FindWhite(const_cast<char *>(ptr)) - ptr;
        String flagName(ptr, flagNameLength);
        ptr += flagNameLength;
        if (Record const *flag = flags.tryFind(defn::Definition::VAR_ID, flagName.toLower())) {
            value |= flag->geti("value");
        }
        else {
            LOG_RES_WARNING("Flag '%s' is not defined (or used out of context)") << flagName;
        }
    }
    return value;
}
int ded_s::getEpisodeNum(String const &id) const {
    if (Record const *def = episodes.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
}
int ded_s::getMapInfoNum(de::Uri const &uri) const {
    if (Record const *def = mapInfos.tryFind(defn::Definition::VAR_ID, uri.compose())) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;  // Not found.
}
int ded_s::getMaterialNum(de::Uri const &uri) const {
    if (uri.isEmpty()) return -1;  // Not found.
    if (uri.scheme().isEmpty()) {
        de::Uri temp(uri);
        temp.setScheme("Sprites");
        int idx = getMaterialNum(temp);
        if (idx >= 0) return idx;
        temp.setScheme("Textures");
        idx = getMaterialNum(temp);
        if (idx >= 0) return idx;
        temp.setScheme("Flats");
        idx = getMaterialNum(temp);
        /*if (idx >= 0)*/ return idx;
    }
    if (Record const *def = materials.tryFind(defn::Definition::VAR_ID, uri.compose())) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;  // Not found.
}
int ded_s::getModelNum(const char *id) const {
    if (Record const *def = models.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
/*    int idx = -1;
    if (id && id[0] && !models.empty()) {
        int i = 0;
        do {
            if (!qstricmp(models[i].id, id)) idx = i;
        } while (idx == -1 && ++i < (int)models.size());
    }
    return idx;*/
}
int ded_s::getSkyNum(char const *id) const {
    if (Record const *def = skies.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
    /*if (!id || !id[0]) return -1;
    for (int i = skies.size() - 1; i >= 0; i--) {
        if (!qstricmp(skies[i].id, id))
            return i;
    }
    return -1;*/
}
int ded_s::getSoundNum(String const &id) const {
    return getSoundNum(id.toUtf8());
}
int ded_s::getSoundNum(const char *id) const {
    int idx = -1;
    if (id && id[0] && sounds.size()) {
        int i = 0;
        do {
            if (!qstricmp(sounds[i].id, id)) idx = i;
        } while (idx == -1 && ++i < sounds.size());
    }
    return idx;
}
int ded_s::getSoundNumForName(const char *name) const {
    if (!name || !name[0])
        return -1;
    for (int i = 0; i < sounds.size(); ++i)
        if (!qstricmp(sounds[i].name, name))
            return i;
    return 0;
}
int ded_s::getSpriteNum(String const &id) const {
    return getSpriteNum(id.toLatin1());
}
int ded_s::getSpriteNum(char const *id) const {
    if (id && id[0]) {
        for (dint i = 0; i < sprites.size(); ++i) {
            if (!qstricmp(sprites[i].id, id))
                return i;
        }
    }
    return -1;  // Not found.
}
int ded_s::getMusicNum(char const *id) const {
    if (Record const *def = musics.tryFind(defn::Definition::VAR_ID, id)) {
        return def->geti(defn::Definition::VAR_ORDER);
    }
    return -1;
    /*int idx = -1;
    if (id && id[0] && musics.size()) {
        int i = 0;
        do {
            if (!qstricmp(musics[i].id, id)) idx = i;
        } while (idx == -1 && ++i < musics.size());
    }
    return idx;*/
}
int ded_s::getValueNum(char const *id) const {
    if (id && id[0]) {
        for (dint i = values.size() - 1; i >= 0; i--) {
            if (!qstricmp(values[i].id, id))
                return i;
        }
    }
    return -1;  // Not found.
}
int ded_s::getValueNum(String const &id) const {
    return getValueNum(id.toLatin1());
}
ded_value_t *ded_s::getValueById(char const *id) const {
    if (!id || !id[0]) return nullptr;
    for (dint i = values.size() - 1; i >= 0; i--) {
        if (!qstricmp(values[i].id, id))
            return &values[i];
    }
    return nullptr;
}
ded_value_t *ded_s::getValueById(String const &id) const {
    return getValueById(id.toLatin1());
}
ded_value_t *ded_s::getValueByUri(de::Uri const &uri) const {
    if (!uri.scheme().compareWithoutCase("Values")) {
        return getValueById(uri.pathCStr());
    }
    return nullptr;
}
ded_compositefont_t *ded_s::findCompositeFontDef(de::Uri const &uri) const {
    for (dint i = compositeFonts.size() - 1; i >= 0; i--) {
        ded_compositefont_t *def = &compositeFonts[i];
        if (def->uri && uri == *def->uri) {
            return def;
        }
    }
    return nullptr;
}
ded_compositefont_t *ded_s::getCompositeFont(char const *uriCString) const {
    ded_compositefont_t *def = nullptr;
    if (uriCString && uriCString[0]) {
        de::Uri uri(uriCString, RC_NULL);
        if (uri.scheme().isEmpty()) {
            de::Uri temp(uri);
            temp.setScheme("Game");
            def = findCompositeFontDef(temp);
            if (!def) {
                temp.setScheme("System");
                def = findCompositeFontDef(temp);
            }
        }
        if (!def) {
            def = findCompositeFontDef(uri);
        }
    }
    return def;
}
String ded_s::findEpisode(String const &mapId) const {
    de::Uri mapUri(mapId, RC_NULL);
    if (mapUri.scheme().isEmpty()) mapUri.setScheme("Maps");
    for (int i = 0; i < episodes.size(); ++i) {
        defn::Episode episode(episodes[i]);
        if (episode.tryFindMapGraphNode(mapUri.compose())) {
            return episode.gets(defn::Definition::VAR_ID);
        }
    }
    return String();
}
int ded_s::getTextNum(char const *id) const {
    if (id && id[0]) {
        for (int i = text.size() - 1; i >= 0; i--) {
            if (!qstricmp(text[i].id, id)) return i;
        }
    }
    return -1; // Not found.
}
static ded_t *s_defs = nullptr;
void DED_DestroyDefinitions() {
    delete s_defs;
    s_defs = nullptr;
}
ded_t *DED_Definitions() {
    if (!s_defs) {
        s_defs = new ded_t;
    }
    return s_defs;
}
static char dedReadError[512];
void DED_SetError(String const &message) {
    String msg = "Error: " + message + ".";
    strncpy(dedReadError, msg.toUtf8().constData(), sizeof(dedReadError));
}
void Def_ReadProcessDED(ded_t *defs, String sourcePath) {
     LOG_AS("Def_ReadProcessDED");
     if (sourcePath.isEmpty()) return;
     try {
         Block text;
         App::rootFolder().locate<File const>(sourcePath) >> text;
         if (!DED_ReadData(defs, text, sourcePath, true/*consider it custom; there is no way to check...*/))
         {
             App_FatalError("Def_ReadProcessDED: %s\n", dedReadError);
         }
        return; // Done!
    }
    catch (...) {
    }
    de::Uri const uri(sourcePath, RC_NULL);
    if (!App_FileSystem().accessFile(uri)) {
        LOG_RES_WARNING("\"%s\" not found!") << NativePath(uri.asText()).pretty();
        return;
    }
    if (!App_FileSystem().checkFileId(uri)) {
        LOG_RES_XVERBOSE("\"%s\" has already been read", NativePath(uri.asText()).pretty());
        return;
    }
    if (!DED_Read(defs, sourcePath)) {
        App_FatalError("Def_ReadProcessDED: %s\n", dedReadError);
    }
}
int DED_ReadLump(ded_t *ded, lumpnum_t lumpNum) {
    try {
        File1 &lump = App_FileSystem().lump(lumpNum);
        if (lump.size() > 0) {
            uint8_t const *data = lump.cache();
            String sourcePath = lump.container().composePath();
            bool custom       = (lump.isContained()? lump.container().hasCustom() : lump.hasCustom());
            const int result  = DED_ReadData(ded, (char const *)data, sourcePath, custom);
            lump.unlock();
            return result;
        }
        return true;
    }
    catch (LumpIndex::NotFoundError const&)
    {} // Ignore error.
    DED_SetError("Bad lump number");
    return 0;
}
int DED_Read(ded_t *ded, String path) {
    try {
        String fullPath = (NativePath::workPath() / NativePath(path).expand()).withSeparators('/');
        QScopedPointer<FileHandle> hndl(&App_FileSystem().openFile(fullPath, "rb"));
        hndl->seek(0, SeekEnd);
        size_t bufferedDefSize = hndl->tell();
        hndl->rewind();
        char *bufferedDef = (char *) M_Calloc(bufferedDefSize + 1);
        File1 &file = hndl->file();
        bool const isCustom = (file.isContained()? file.container().hasCustom() : file.hasCustom());
        hndl->read((uint8_t *)bufferedDef, bufferedDefSize);
        int result = DED_ReadData(ded, bufferedDef, path, isCustom);
        App_FileSystem().releaseFile(file);
        M_Free(bufferedDef);
        return result;
    }
    catch (FS1::NotFoundError const &)
    {} // Ignore.
    DED_SetError("File could not be opened for reading");
    return false;
}
int DED_ReadData(ded_t *ded, char const *buffer, String sourceFile, bool sourceIsCustom) {
    return DEDParser(ded).parse(buffer, sourceFile, sourceIsCustom);
}
char const *DED_Error() {
    return dedReadError;
}
#define MAX_RECUR_DEPTH     30
#define MAX_TOKEN_LEN       128
#define STOPCHAR(x) (isspace(x) || x == ';' || x == '#' || x == '{' \
                    || x == '}' || x == '=' || x == '"' || x == '*' \
                    || x == '|')
#define ISTOKEN(X)  (!stricmp(token, X))
#define READSTR(X)  if (!ReadString(X, sizeof(X))) { \
                    setError("Syntax error in string value"); \
                    retVal = false; goto ded_end_read; }
#define READURI(X, SHM) if (!ReadUri(X, SHM)) { \
    setError("Syntax error parsing resource path"); \
    retVal = false; goto ded_end_read; }
#define MISSING_SC_ERROR    setError("Missing semicolon"); \
                            retVal = false; goto ded_end_read;
#define CHECKSC     if (source->version <= 5) { ReadToken(); if (!ISTOKEN(";")) { MISSING_SC_ERROR; } }
#define SKIPSC      { ReadToken(); if (!ISTOKEN(";")) { UnreadToken(token); }}
#define FINDBEGIN   while (!ISTOKEN("{") && !source->atEnd) ReadToken();
#define FINDEND     while (!ISTOKEN("}") && !source->atEnd) ReadToken();
#define ISLABEL(X)  (!stricmp(label, X))
#define READLABEL   if (!ReadLabel(label)) { retVal = false; goto ded_end_read; } \
                    if (ISLABEL("}")) break;
#define READLABEL_NOBREAK   if (!ReadLabel(label)) { retVal = false; goto ded_end_read; }
#define FAILURE     retVal = false; goto ded_end_read;
#define READBYTE(X) if (!ReadByte(&X)) { FAILURE }
#define READINT(X)  if (!ReadInt(&X, false)) { FAILURE }
#define READUINT(X) if (!ReadInt(&X, true)) { FAILURE }
#define READFLT(X)  if (!ReadFloat(&X)) { FAILURE }
#define READNBYTEVEC(X,N) if (!ReadNByteVector(X, N)) { FAILURE }
#define READFLAGS(X,P) if (!ReadFlags(&X, P)) { FAILURE }
#define READBLENDMODE(X) if (!ReadBlendmode(&X)) { FAILURE }
#define READSTRING(S,I) if (!ReadString(S, sizeof(S))) { I = strtol(token,0,0); }
#define READVECTOR(X,N) if (!ReadVector(X, N)) { FAILURE }
#define RV_BYTE(lab, X) if (ISLABEL(lab)) { READBYTE(X); } else
#define RV_INT(lab, X)  if (ISLABEL(lab)) { READINT(X); } else
#define RV_UINT(lab, X) if (ISLABEL(lab)) { READUINT(X); } else
#define RV_FLT(lab, X)  if (ISLABEL(lab)) { READFLT(X); } else
#define RV_VEC(lab, X, N)   if (ISLABEL(lab)) { int b; FINDBEGIN; \
                        for (b=0; b<N; ++b) {READFLT(X[b])} ReadToken(); } else
#define RV_VEC_VAR(lab, X, N) if (ISLABEL(lab)) { READVECTOR(X, N); } else
#define RV_IVEC(lab, X, N)  if (ISLABEL(lab)) { int b; FINDBEGIN; \
                        for (b=0; b<N; ++b) {READINT(X[b])} ReadToken(); } else
#define RV_NBVEC(lab, X, N) if (ISLABEL(lab)) { READNBYTEVEC(X,N); } else
#define RV_INT_ELEM(lab, ARRAY, ELEM) if (ISLABEL(lab)) { \
    int _value; READINT(_value); \
    (ARRAY).array().setElement(ELEM, _value); } else
#define RV_FLT_ELEM(lab, ARRAY, ELEM) if (ISLABEL(lab)) { \
    float _value; READFLT(_value); \
    (ARRAY).array().setElement(ELEM, _value); } else
#define RV_STR_ELEM(lab, ARRAY, ELEM) if (ISLABEL(lab)) { \
    String _value; if (!ReadString(_value)) { FAILURE } \
    (ARRAY).array().setElement(ELEM, _value); } else
#define RV_STR(lab, X)  if (ISLABEL(lab)) { READSTR(X); } else
#define RV_STR_INT(lab, S, I)   if (ISLABEL(lab)) { if (!ReadString(S,sizeof(S))) \
                                I = strtol(token,0,0); } else
#define RV_URI(lab, X, RN)  if (ISLABEL(lab)) { READURI(X, RN); } else
#define RV_FLAGS(lab, X, P) if (ISLABEL(lab)) { READFLAGS(X, P); } else
#define RV_FLAGS_ELEM(lab, X, ELEM, P) if (ISLABEL(lab)) { if (!ReadFlags(&X, P, ELEM)) { FAILURE } } else
#define RV_BLENDMODE(lab, X) if (ISLABEL(lab)) { READBLENDMODE(X); } else
#define RV_ANYSTR(lab, X)   if (ISLABEL(lab)) { if (!ReadAnyString(&X)) { FAILURE } } else
#define RV_END          { setError("Unknown label '" + String(label) + "'"); retVal = false; goto ded_end_read; }
static struct xgclass_s *xgClassLinks;
template <typename T>
struct Dummy : public T {
    Dummy() { zap(*this); }
    ~Dummy() { T::release(); }
    void clear() { T::release(); zap(*this); }
};
void DED_SetXGClassLinks(struct xgclass_s *links) {
    xgClassLinks = links;
}
DENG2_PIMPL(DEDParser) {
    ded_t *ded;
    struct dedsource_s {
        char const *buffer;
        char const *pos;
        dd_bool     atEnd;
        int         lineNumber;
        String      fileName;
        int         version;  ///< v6 does not require semicolons.
        bool        custom;   ///< @c true= source is a user supplied add-on.
    };
    typedef dedsource_s dedsource_t;
    dedsource_t sourceStack[MAX_RECUR_DEPTH];
    dedsource_t *source; // Points to the current source.
    char token[MAX_TOKEN_LEN+1];
    char unreadToken[MAX_TOKEN_LEN+1];
    Impl(Public *i) : Base(i), ded(0), source(0) {
        zap(token);
        zap(unreadToken);
    }
    void DED_InitReader(char const *buffer, String fileName, bool sourceIsCustom) {
        if (source && source - sourceStack >= MAX_RECUR_DEPTH) {
            App_FatalError("DED_InitReader: Include recursion is too deep.\n");
        }
        if (!source) {
            source = sourceStack;
        }
        else {
            source++;
        }
        source->pos = source->buffer = buffer;
        source->atEnd      = false;
        source->lineNumber = 1;
        source->fileName   = fileName;
        source->version    = DED_VERSION;
        source->custom     = sourceIsCustom;
    }
    void DED_CloseReader() {
        if (source == sourceStack) {
            source = 0;
        }
        else {
            source--;
        }
    }
    String readPosAsText() {
        return "\"" + (source? source->fileName : "[buffered-data]") + "\""
               " on line #" + String::number(source? source->lineNumber : 0);
    }
    void setError(String const &message) {
        DED_SetError("In " + readPosAsText() + "\n  " + message);
    }
    /**
     * Reads a single character from the input file. Increments the line
     * number counter if necessary.
     */
    int FGetC(void) {
        int ch = (unsigned char) *source->pos;
        if (ch)
            source->pos++;
        else
            source->atEnd = true;
        if (ch == '\n')
            source->lineNumber++;
        if (ch == '\r')
            return FGetC();
        return ch;
    }
    /**
     * Undoes an FGetC.
     */
    int FUngetC(int ch) {
        if (source->atEnd)
            return 0;
        if (ch == '\n')
            source->lineNumber--;
        if (source->pos > source->buffer)
            source->pos--;
        return ch;
    }
    /**
     * Reads stuff until a newline is found.
     */
    void SkipComment(void) {
        int                 ch = FGetC();
        dd_bool             seq = false;
        if (ch == '\n')
            return; // Comment ends right away.
        if (ch != '>') // Single-line comment?
        {
            while (FGetC() != '\n' && !source->atEnd) {}
        }
        else // Multiline comment?
        {
            while (!source->atEnd) {
                ch = FGetC();
                if (seq) {
                    if (ch == '#')
                        break;
                    seq = false;
                }
                if (ch == '<')
                    seq = true;
            }
        }
    }
    int ReadToken(void) {
        int                 ch;
        char*               out = token;
        if (unreadToken[0]) {
            strcpy(token, unreadToken);
            strcpy(unreadToken, "");
            return true;
        }
        ch = FGetC();
        if (source->atEnd)
            return false;
        while ((ch == '#' || isspace(ch))) {
            if (ch == '#')
                SkipComment();
            ch = FGetC();
            if (source->atEnd)
                return false;
        }
        *out++ = ch;
        if (STOPCHAR(ch)) {
            *out = 0;
            return true;
        }
        while (!STOPCHAR(ch) && !source->atEnd) {
            ch = FGetC();
            *out++ = ch;
        }
        *(out - 1) = 0; // End token.
        FUngetC(ch);
        return true;
    }
    void UnreadToken(const char* token) {
        strcpy(unreadToken, token);
    }
    /**
     * Current pos in the file is at the first ".
     * Does not expand escape sequences, only checks for \".
     *
     * @return  @c true if successful.
     */
    int ReadString(String &dest, bool inside = false, bool doubleq = false) {
        if (!inside) {
            ReadToken();
            if (!ISTOKEN("\""))
                return false;
        }
        bool esc = false, newl = false;
        int ch = FGetC();
        while (esc || ch != '"') // The string-end-character.
        {
            if (source->atEnd)
                return false;
            if (newl) {
                if (isspace(ch)) {
                    ch = FGetC();
                    continue;
                }
                else {
                    newl = false;
                }
            }
            if (!esc && ch == '\\') {
                esc = true;
            }
            else {
                if (esc && ch != '"' && ch != '\\')
                    dest += '\\';
                esc = false;
            }
            if (ch == '\n')
                newl = true;
            if (!esc && !newl) {
                dest += char(ch);
                if (doubleq && ch == '"')
                    dest += '"';
            }
            ch = FGetC();
        }
        return true;
    }
    int ReadString(char *dest, int maxLen) {
        DENG_ASSERT(dest != 0);
        String buffer;
        if (!ReadString(buffer)) return false;
        qstrncpy(dest, buffer.toUtf8().constData(), maxLen);
        return true;
    }
    int ReadString(Variable &var, int) {
        String buffer;
        if (!ReadString(buffer)) return false;
        var.set(TextValue(buffer));
        return true;
    }
    /**
     * Read a string of (pretty much) any length.
     */
    int ReadAnyString(char **dest) {
        String buffer;
        if (!ReadString(buffer))
            return false;
        if (*dest)
            M_Free(*dest);
        QByteArray bufferUtf8 = buffer.toUtf8();
        *dest = (char *) M_Malloc(bufferUtf8.length() + 1);
        qstrcpy(*dest, bufferUtf8.constData());
        return true;
    }
    int ReadUri(de::Uri **dest, char const *defaultScheme) {
        String buffer;
        if (!ReadString(buffer))
            return false;
        buffer = Path::normalizeString(buffer);
        if (!*dest)
            *dest = new de::Uri(buffer, RC_NULL);
        else
            (*dest)->setUri(buffer, RC_NULL);
        if (defaultScheme && defaultScheme[0] && (*dest)->scheme().isEmpty())
            (*dest)->setScheme(defaultScheme);
        return true;
    }
    int ReadUri(Variable &var, char const *defaultScheme) {
        de::Uri *uri = 0;
        if (!ReadUri(&uri, defaultScheme)) return false;
        var.set(TextValue(uri->compose()));
        delete uri;
        return true;
    }
    int ReadNByteVector(Variable &var, int count) {
        FINDBEGIN;
        for (int i = 0; i < count; ++i) {
            ReadToken();
            if (ISTOKEN("}"))
                return true;
            var.array().setElement(i, strtoul(token, 0, 0));
        }
        FINDEND;
        return true;
    }
    /**
     * @return              @c true, if successful.
     */
    int ReadByte(unsigned char* dest) {
        ReadToken();
        if (ISTOKEN(";")) {
            setError("Missing integer value");
            return false;
        }
        *dest = strtoul(token, 0, 0);
        return true;
    }
    /**
     * @return              @c true, if successful.
     */
    int ReadInt(int* dest, int unsign) {
        ReadToken();
        if (ISTOKEN(";")) {
            setError("Missing integer value");
            return false;
        }
        *dest = unsign? (int)strtoul(token, 0, 0) : strtol(token, 0, 0);
        return true;
    }
    int ReadInt(Variable *var, int unsign) {
        int value = 0;
        if (ReadInt(&value, unsign)) {
            var->set(NumberValue(value));
            return true;
        }
        return false;
    }
    /**
     * @return              @c true, if successful.
     */
    int ReadFloat(float* dest) {
        ReadToken();
        if (ISTOKEN(";")) {
            setError("Missing float value");
            return false;
        }
        *dest = (float) strtod(token, 0);
        return true;
    }
    int ReadFloat(Variable *var) {
        float value = 0;
        if (ReadFloat(&value)) {
            var->set(NumberValue(value));
            return true;
        }
        return false;
    }
    int ReadVector(Variable &var, int componentCount) {
        FINDBEGIN;
        for (int b = 0; b < componentCount; ++b) {
            float value = 0;
            if (!ReadFloat(&value)) return false;
            var.array().setElement(b, value);
        }
        ReadToken();
        return true;
    }
    int ReadFlags(int *dest, char const *prefix) {
        DENG2_ASSERT(dest);
        *dest = 0;
        ReadToken();
        if (ISTOKEN(";")) {
            setError("Missing flags value");
            return false;
        }
        if (ISTOKEN("0")) {
            return true;
        }
        UnreadToken(token);
        String flag;
        if (ISTOKEN("\"")) {
            if (!ReadString(flag))
                return false;
            flag.strip();
            if (!flag.isEmpty()) {
                *dest = ded->evalFlags(flag.toUtf8().constData());
            }
            return true;
        }
        forever {
            ReadToken();
            if (prefix) {
                flag = String(prefix) + String(token);
            }
            else {
                flag = String(token);
            }
            flag.strip();
            if (!flag.isEmpty()) {
                *dest |= ded->evalFlags(flag.toUtf8().constData());
            }
            if (!ReadToken())
                break;
            if (!ISTOKEN("|")) // | is required for multiple flags.
            {
                UnreadToken(token);
                break;
            }
        }
        return true;
    }
    int ReadFlags(Variable *dest, char const *prefix, int elementIndex = -1) {
        int value = 0;
        if (ReadFlags(&value, prefix)) {
            std::unique_ptr<NumberValue> flagsValue(new NumberValue(value, NumberValue::Hex));
            if (elementIndex < 0) {
                dest->set(flagsValue.release());
            }
            else {
                dest->array().setElement(NumberValue(elementIndex), flagsValue.release());
            }
            return true;
        }
        return false;
    }
    int ReadBlendmode(blendmode_t *dest) {
        LOG_AS("ReadBlendmode");
        String flag;
        blendmode_t bm;
        ReadToken();
        UnreadToken(token);
        if (ISTOKEN("\"")) {
            if (!ReadString(flag)) return false;
            bm = blendmode_t(ded->evalFlags(flag.toUtf8().constData()));
        }
        else {
            ReadToken();
            flag = String("bm_") + String(token);
            bm = blendmode_t(ded->evalFlags(flag.toUtf8().constData()));
        }
        if (bm != BM_NORMAL) {
            *dest = bm;
        }
        else {
            LOG_RES_WARNING("Unknown BlendMode '%s' in \"%s\" on line #%i")
                << flag << (source ? source->fileName : "?") << (source ? source->lineNumber : 0);
        }
        return true;
    }
    int ReadBlendmode(Variable *var) {
        blendmode_t mode = BM_NORMAL;
        if (!ReadBlendmode(&mode)) return false;
        var->set(NumberValue(int(mode)));
        return true;
    }
    /**
     * @return @c true, if successful.
     */
    int ReadLabel(char* label) {
        strcpy(label, "");
        for (;;) {
            ReadToken();
            if (source->atEnd) {
                setError("Unexpected end of file");
                return false;
            }
            if (ISTOKEN("}")) // End block.
            {
                strcpy(label, token);
                return true;
            }
            if (ISTOKEN(";")) {
                if (source->version <= 5) {
                    setError("Label without value");
                    return false;
                }
                continue; // Semicolons are optional in v6.
            }
            if (ISTOKEN("=") || ISTOKEN("{"))
                break;
            if (label[0])
                strcat(label, " ");
            strcat(label, token);
        }
        return true;
    }
    /**
     * @param cond          Condition token. Can be a command line option
     *                      or a game mode.
     * @return              @c true if the condition passes.
     */
    dd_bool DED_CheckCondition(char const *cond, dd_bool expected) {
        dd_bool value = false;
        if (cond[0] == '-') {
            value = (CommandLine_Check(token) != 0);
        }
        else if (isalnum(cond[0]) && !DoomsdayApp::game().isNull()) {
            value = !String(cond).compareWithoutCase(DoomsdayApp::game().id());
        }
        return value == expected;
    }
    int readData(char const *buffer, String sourceFile, bool sourceIsCustom) {
        const String &VAR_ID = defn::Definition::VAR_ID;
        char  dummy[128], label[128], tmp[256];
        int   dummyInt, idx, retVal = true;
        int   prevEpisodeDefIdx = -1; // For "Copy".
        int   prevMobjDefIdx = -1; // For "Copy".
        int   prevStateDefIdx = -1; // For "Copy"
        int   prevLightDefIdx = -1; // For "Copy".
        int   prevMaterialDefIdx = -1; // For "Copy".
        int   prevModelDefIdx = -1; // For "Copy".
        int   prevMapInfoDefIdx = -1; // For "Copy".
        int   prevMusicDefIdx = -1; // For "Copy".
        int   prevSkyDefIdx = -1; // For "Copy".
        int   prevDetailDefIdx = -1; // For "Copy".
        int   prevGenDefIdx = -1; // For "Copy".
        int   prevDecorDefIdx = -1; // For "Copy".
        int   prevRefDefIdx = -1; // For "Copy".
        int   prevLineTypeDefIdx = -1; // For "Copy".
        int   prevSectorTypeDefIdx = -1; // For "Copy".
        int   depth;
        char *rootStr = 0, *ptr;
        int   bCopyNext = 0;
        DED_InitReader(buffer, sourceFile, sourceIsCustom);
        String sourceFileDir = sourceFile.fileNamePath();
        if (sourceFileDir.isEmpty()) {
            sourceFileDir = NativePath::workPath();
        }
        while (ReadToken()) {
            if (ISTOKEN("Copy") || ISTOKEN("*")) {
                bCopyNext = true;
                continue; // Read the next token.
            }
            if (ISTOKEN(";")) {
                continue;
            }
            if (ISTOKEN("SkipIf")) {
                dd_bool expected = true;
                ReadToken();
                if (ISTOKEN("Not")) {
                    expected = false;
                    ReadToken();
                }
                if (DED_CheckCondition(token, expected)) {
                    goto ded_end_read;
                }
                CHECKSC;
            }
            if (ISTOKEN("Include")) {
                READSTR(tmp);
                CHECKSC;
                DED_Include(tmp, sourceFileDir);
                strcpy(label, "");
            }
            if (ISTOKEN("IncludeIf")) // An optional include.
            {
                dd_bool expected = true;
                ReadToken();
                if (ISTOKEN("Not")) {
                    expected = false;
                    ReadToken();
                }
                if (DED_CheckCondition(token, expected)) {
                    READSTR(tmp);
                    CHECKSC;
                    DED_Include(tmp, sourceFileDir);
                    strcpy(label, "");
                }
                else {
                    READSTR(tmp);
                    CHECKSC;
                }
            }
            if (ISTOKEN("ModelPath")) {
                READSTR(label);
                CHECKSC;
                de::Uri newSearchPath = de::Uri::fromNativeDirPath(NativePath(label));
                FS1::Scheme& scheme = App_FileSystem().scheme(ResourceClass::classForId(RC_MODEL).defaultScheme());
                scheme.addSearchPath(reinterpret_cast<de::Uri const&>(newSearchPath), FS1::ExtraPaths);
            }
            if (ISTOKEN("Header")) {
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    if (ISLABEL("Version")) {
                        READINT(ded->version);
                        source->version = ded->version;
                    }
                    else RV_STR("Thing prefix", dummy)
                    RV_STR("State prefix", dummy)
                    RV_STR("Sprite prefix", dummy)
                    RV_STR("Sfx prefix", dummy)
                    RV_STR("Mus prefix", dummy)
                    RV_STR("Text prefix", dummy)
                    RV_STR("Model path", dummy)
                    RV_FLAGS("Common model flags", ded->modelFlags, "df_")
                    RV_FLT("Default model scale", ded->modelScale)
                    RV_FLT("Default model offset", ded->modelOffset)
                    RV_END
                    CHECKSC;
                }
            }
            if (ISTOKEN("Flag")) {
                ded_stringid_t id;
                int value = 0;
                char dummyStr[2];
                de::zap(id);
                FINDBEGIN;
                forever {
                    READLABEL;
                    RV_STR("ID", id)
                    RV_UINT("Value", value)
                    RV_STR("Info", dummyStr) // ignored
                    RV_END
                    CHECKSC;
                }
                if (qstrlen(id)) {
                    ded->addFlag(id, value);
                    Record &flag = ded->flags.find(VAR_ID, id);
                    DENG2_ASSERT(flag.geti("value") == value); // Sanity check.
                    if (source->custom) flag.set("custom", true);
                }
            }
            if (ISTOKEN("Episode")) {
                bool bModify = false;
                Record dummyEpsd;
                Record *epsd = 0;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addEpisode();
                    epsd = &ded->episodes[idx];
                }
                else if (!bCopyNext) {
                    ded_stringid_t otherEpisodeId;
                    READSTR(otherEpisodeId);
                    ReadToken();
                    idx = ded->getEpisodeNum(otherEpisodeId);
                    if (idx >= 0) {
                        epsd = &ded->episodes[idx];
                        bModify = true;
                    }
                    else {
                        LOG_RES_WARNING("Ignoring unknown Episode \"%s\" in %s on line #%i")
                                << otherEpisodeId << (source? source->fileName : "?")
                                << (source? source->lineNumber : 0);
                        defn::Episode(dummyEpsd).resetToDefaults();
                        epsd = &dummyEpsd;
                    }
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                DENG2_ASSERT(epsd != 0);
                if (prevEpisodeDefIdx >= 0 && bCopyNext) {
                    ded->episodes.copy(prevEpisodeDefIdx, *epsd);
                }
                if (source->custom) epsd->set("custom", true);
                defn::Episode mainDef(*epsd);
                int hub = 0;
                int notHubMap = 0;
                FINDBEGIN;
                forever {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READSTR((*epsd)[VAR_ID]);
                    }
                    else RV_URI("Start Map", (*epsd)["startMap"], "Maps")
                    RV_STR("Title", (*epsd)["title"])
                    RV_STR("Menu Help Info", (*epsd)["menuHelpInfo"])
                    RV_URI("Menu Image", (*epsd)["menuImage"], "Patches")
                    RV_STR("Menu Shortcut", (*epsd)["menuShortcut"])
                    if (ISLABEL("Hub")) {
                        Record *hubRec;
                        if (hub >= mainDef.hubCount()) {
                            mainDef.addHub();
                        }
                        DENG_ASSERT(hub < mainDef.hubCount());
                        hubRec = &mainDef.hub(hub);
                        if (source->custom) hubRec->set("custom", true);
                        int map = 0;
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_STR("ID", (*hubRec)[VAR_ID])
                            if (ISLABEL("Map")) {
                                if (map >= int(hubRec->geta("map").size())) {
                                    QScopedPointer<Record> map(new Record);
                                    defn::MapGraphNode(*map).resetToDefaults();
                                    (*hubRec)["map"].array()
                                            .add(new RecordValue(map.take(), RecordValue::OwnsRecord));
                                }
                                DENG_ASSERT(map < int(hubRec->geta("map").size()));
                                Record &mapRec = *hubRec->geta("map")[map].as<RecordValue>().record();
                                if (source->custom) mapRec.set("custom", true);
                                int exit = 0;
                                FINDBEGIN;
                                forever {
                                    READLABEL;
                                    RV_URI("ID", mapRec[VAR_ID], "Maps")
                                    RV_INT("Warp Number", mapRec["warpNumber"])
                                    if (ISLABEL("Exit")) {
                                        defn::MapGraphNode mgNodeDef(mapRec);
                                        if (exit >= mgNodeDef.exitCount()) {
                                            mgNodeDef.addExit();
                                        }
                                        DENG_ASSERT(exit < mgNodeDef.exitCount());
                                        Record &exitRec = mgNodeDef.exit(exit);
                                        if (source->custom) exitRec.set("custom", true);
                                        FINDBEGIN;
                                        forever {
                                            READLABEL;
                                            RV_STR("ID", exitRec[VAR_ID])
                                            RV_URI("Target Map", exitRec["targetMap"], "Maps")
                                            RV_END
                                            CHECKSC;
                                        }
                                        exit++;
                                    }
                                    else RV_END
                                    CHECKSC;
                                }
                                map++;
                            }
                            else RV_END
                            CHECKSC;
                        }
                        hub++;
                    }
                    else if (ISLABEL("Map")) {
                        if (notHubMap >= int(epsd->geta("map").size())) {
                            QScopedPointer<Record> map(new Record);
                            defn::MapGraphNode(*map).resetToDefaults();
                            (*epsd)["map"].array()
                                    .add(new RecordValue(map.take(), RecordValue::OwnsRecord));
                        }
                        DENG_ASSERT(notHubMap < int(epsd->geta("map").size()));
                        Record &mapRec = *epsd->geta("map")[notHubMap].as<RecordValue>().record();
                        if (source->custom) mapRec.set("custom", true);
                        int exit = 0;
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_URI("ID", mapRec[VAR_ID], "Maps")
                            RV_INT("Warp Number", mapRec["warpNumber"])
                            if (ISLABEL("Exit")) {
                                defn::MapGraphNode mgNodeDef(mapRec);
                                if (exit >= mgNodeDef.exitCount()) {
                                    mgNodeDef.addExit();
                                }
                                DENG_ASSERT(exit < mgNodeDef.exitCount());
                                Record &exitRec = mgNodeDef.exit(exit);
                                if (source->custom) exitRec.set("custom", true);
                                FINDBEGIN;
                                forever {
                                    READLABEL;
                                    RV_STR("ID", exitRec[VAR_ID])
                                    RV_URI("Target Map", exitRec["targetMap"], "Maps")
                                    RV_END
                                    CHECKSC;
                                }
                                exit++;
                            }
                            else RV_END
                            SKIPSC;
                        }
                        notHubMap++;
                    }
                    else RV_END
                    SKIPSC;
                }
                if (idx > 0) {
                    prevEpisodeDefIdx = idx;
                }
            }
            if (ISTOKEN("Mobj") || ISTOKEN("Thing")) {
                dd_bool bModify = false;
                Record* mo;
                Record dummyMo;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addThing("");
                    mo = &ded->things[idx];
                }
                else if (!bCopyNext) {
                    String otherMobjId;
                    if (!ReadString(otherMobjId)) { FAILURE }
                    ReadToken();
                    idx = ded->getMobjNum(otherMobjId);
                    if (idx < 0) {
                        LOG_RES_WARNING("Ignoring unknown Mobj \"%s\" in %s on line #%i")
                                << otherMobjId << (source? source->fileName : "?")
                                << (source? source->lineNumber : 0);
                        dummyMo.clear();
                        mo = &dummyMo;
                    }
                    else {
                        mo = &ded->things[idx];
                        bModify = true;
                    }
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                if (prevMobjDefIdx >= 0 && bCopyNext) {
                    ded->things.copy(prevMobjDefIdx, *mo);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READSTR((*mo)[VAR_ID]);
                    }
                    else RV_INT("DoomEd number", (*mo)["doomEdNum"])
                    RV_STR("Name", (*mo)["name"])
                    RV_STR_ELEM("Spawn state",   (*mo)["states"], SN_SPAWN)
                    RV_STR_ELEM("See state",     (*mo)["states"], SN_SEE)
                    RV_STR_ELEM("Pain state",    (*mo)["states"], SN_PAIN)
                    RV_STR_ELEM("Melee state",   (*mo)["states"], SN_MELEE)
                    RV_STR_ELEM("Missile state", (*mo)["states"], SN_MISSILE)
                    RV_STR_ELEM("Crash state",   (*mo)["states"], SN_CRASH)
                    RV_STR_ELEM("Death state",   (*mo)["states"], SN_DEATH)
                    RV_STR_ELEM("Xdeath state",  (*mo)["states"], SN_XDEATH)
                    RV_STR_ELEM("Raise state",   (*mo)["states"], SN_RAISE)
                    RV_STR_ELEM("See sound",    (*mo)["sounds"], SDN_SEE)
                    RV_STR_ELEM("Attack sound", (*mo)["sounds"], SDN_ATTACK)
                    RV_STR_ELEM("Pain sound",   (*mo)["sounds"], SDN_PAIN)
                    RV_STR_ELEM("Death sound",  (*mo)["sounds"], SDN_DEATH)
                    RV_STR_ELEM("Active sound", (*mo)["sounds"], SDN_ACTIVE)
                    RV_INT("Reaction time", (*mo)["reactionTime"])
                    RV_INT("Pain chance", (*mo)["painChance"])
                    RV_INT("Spawn health", (*mo)["spawnHealth"])
                    RV_FLT("Speed", (*mo)["speed"])
                    RV_FLT("Radius", (*mo)["radius"])
                    RV_FLT("Height", (*mo)["height"])
                    RV_INT("Mass", (*mo)["mass"])
                    RV_INT("Damage", (*mo)["damage"])
                    RV_FLAGS_ELEM("Flags", (*mo)["flags"], 0, "mf_")
                    RV_FLAGS_ELEM("Flags2", (*mo)["flags"], 1, "mf2_")
                    RV_FLAGS_ELEM("Flags3", (*mo)["flags"], 2, "mf3_")
                    RV_INT_ELEM("Misc1", (*mo)["misc"], 0)
                    RV_INT_ELEM("Misc2", (*mo)["misc"], 1)
                    RV_INT_ELEM("Misc3", (*mo)["misc"], 2)
                    RV_INT_ELEM("Misc4", (*mo)["misc"], 3)
                    RV_STR("On touch", (*mo)["onTouch"]) // script function (MF_SPECIAL)
                    RV_STR("On death", (*mo)["onDeath"]) // script function
                    RV_END
                    CHECKSC;
                }
                if (idx > 0) {
                    prevMobjDefIdx = idx;
                }
            }
            if (ISTOKEN("State")) {
                dd_bool bModify = false;
                Record *st;
                Record dummyState;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addState("");
                    st = &ded->states[idx];
                }
                else if (!bCopyNext) {
                    ded_stateid_t otherStateId;
                    READSTR(otherStateId);
                    ReadToken();
                    idx = ded->getStateNum(otherStateId);
                    if (idx < 0) {
                        LOG_RES_WARNING("Ignoring unknown State \"%s\" in %s on line #%i")
                                << otherStateId << (source? source->fileName : "?")
                                << (source? source->lineNumber : 0);
                        dummyState.clear();
                        st = &dummyState;
                    }
                    else {
                        st = &ded->states[idx];
                        bModify = true;
                    }
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                if (prevStateDefIdx >= 0 && bCopyNext) {
                    ded->states.copy(prevStateDefIdx, *st);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READSTR((*st)[VAR_ID]);
                    }
                    else if (ISLABEL("Frame")) {
    #define FF_FULLBRIGHT               0x8000
    #define FF_FRAMEMASK                0x7fff
                        int frame = 0;
                        READINT(frame);
                        if (frame & FF_FULLBRIGHT) {
                            frame &= FF_FRAMEMASK;
                            st->set("flags", st->geti("flags") | STF_FULLBRIGHT);
                        }
                        st->set("frame", frame);
    #undef FF_FRAMEMASK
    #undef FF_FULLBRIGHT
                    }
                    else RV_FLAGS("Flags", (*st)["flags"], "statef_")
                    RV_STR("Sprite",     (*st)["sprite"])
                    RV_INT("Tics",       (*st)["tics"])
                    RV_STR("Action",     (*st)["action"])
                    RV_STR("Next state", (*st)["nextState"])
                    RV_INT_ELEM("Misc1", (*st)["misc"], 0)
                    RV_INT_ELEM("Misc2", (*st)["misc"], 1)
                    RV_INT_ELEM("Misc3", (*st)["misc"], 2)
                    RV_STR("Execute",    (*st)["execute"])
                    RV_END
                    CHECKSC;
                }
                if (idx > 0) {
                    prevStateDefIdx = idx;
                }
            }
            if (ISTOKEN("Sprite")) {
                idx = DED_AddSprite(ded, "");
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("ID", ded->sprites[idx].id)
                    RV_END
                    CHECKSC;
                }
            }
            if (ISTOKEN("Light")) {
                ded_light_t* lig;
                idx = DED_AddLight(ded, "");
                lig = &ded->lights[idx];
                if (prevLightDefIdx >= 0 && bCopyNext) {
                    ded->lights.copyTo(lig, prevLightDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("State", lig->state)
                    RV_STR("Map", lig->uniqueMapID)
                    RV_FLT("X Offset", lig->offset[VX])
                    RV_FLT("Y Offset", lig->offset[VY])
                    RV_VEC("Origin", lig->offset, 3)
                    RV_FLT("Size", lig->size)
                    RV_FLT("Intensity", lig->size)
                    RV_FLT("Red", lig->color[0])
                    RV_FLT("Green", lig->color[1])
                    RV_FLT("Blue", lig->color[2])
                    RV_VEC("Color", lig->color, 3)
                    if (ISLABEL("Sector levels")) {
                        int     b;
                        FINDBEGIN;
                        for (b = 0; b < 2; ++b) {
                            READFLT(lig->lightLevel[b])
                            lig->lightLevel[b] /= 255.0f;
                            if (lig->lightLevel[b] < 0)
                                lig->lightLevel[b] = 0;
                            else if (lig->lightLevel[b] > 1)
                                lig->lightLevel[b] = 1;
                        }
                        ReadToken();
                    }
                    else
                    RV_FLAGS("Flags", lig->flags, "lgf_")
                    RV_URI("Top map", &lig->up, "LightMaps")
                    RV_URI("Bottom map", &lig->down, "LightMaps")
                    RV_URI("Side map", &lig->sides, "LightMaps")
                    RV_URI("Flare map", &lig->flare, "LightMaps")
                    RV_FLT("Halo radius", lig->haloRadius)
                    RV_END
                    CHECKSC;
                }
                prevLightDefIdx = idx;
            }
            if (ISTOKEN("Material")) {
                Record dummyMat;
                Record *mat  = nullptr;
                bool bModify = false;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addMaterial();
                    mat = &ded->materials[idx];
                }
                else if (!bCopyNext) {
                    de::Uri *otherMat = nullptr;
                    READURI(&otherMat, nullptr);
                    ReadToken();
                    idx = ded->getMaterialNum(*otherMat);
                    if (idx >= 0) {
                        mat     = &ded->materials[idx];
                        bModify = true;
                    }
                    else {
                        LOG_RES_WARNING("Ignoring unknown Material \"%s\" in %s on line #%i")
                                << otherMat->asText()
                                << (source? source->fileName : "?")
                                << (source? source->lineNumber : 0);
                        defn::Material(dummyMat).resetToDefaults();
                        mat = &dummyMat;
                    }
                    delete otherMat;
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                DENG2_ASSERT(mat);
                if (prevMaterialDefIdx >= 0 && bCopyNext) {
                    ded->materials.copy(prevMaterialDefIdx, *mat);
                }
                defn::Material mainDef(*mat);
                int decor = 0;
                int layer = 0;
                FINDBEGIN;
                forever {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READURI((*mat)[VAR_ID], nullptr);
                    }
                    else
                    RV_FLAGS("Flags", (*mat)["flags"], "matf_")
                    RV_INT_ELEM("Width", (*mat)["dimensions"], 0)
                    RV_INT_ELEM("Height", (*mat)["dimensions"], 1)
                    if (ISLABEL("Layer")) {
                        if (layer >= DED_MAX_MATERIAL_LAYERS) {
                            setError("Too many Material.Layers");
                            retVal = false;
                            goto ded_end_read;
                        }
                        if (layer >= mainDef.layerCount()) mainDef.addLayer();
                        defn::MaterialLayer layerDef(mainDef.layer(layer));
                        int stage = 0;
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            if (ISLABEL("Stage")) {
                                if (stage >= layerDef.stageCount()) {
                                    layerDef.addStage();
                                    if (mainDef.getb("autoGenerated") && stage > 0) {
                                        layerDef.stage(stage).copyMembersFrom(layerDef.stage(stage - 1));
                                    }
                                }
                                Record &st = layerDef.stage(stage);
                                FINDBEGIN;
                                forever {
                                    READLABEL;
                                    RV_URI("Texture", st["texture"], 0) // Default to "any" scheme.
                                    RV_INT("Tics", st["tics"])
                                    RV_FLT("Rnd", st["variance"])
                                    RV_VEC_VAR("Offset", st["texOrigin"], 2)
                                    RV_FLT("Glow Rnd", st["glowStrengthVariance"])
                                    RV_FLT("Glow", st["glowStrength"])
                                    RV_END
                                    CHECKSC;
                                }
                                stage++;
                            }
                            else RV_END
                            CHECKSC;
                        }
                        layer++;
                    }
                    else if (ISLABEL("Light")) {
                        if (decor >= DED_MAX_MATERIAL_DECORATIONS) {
                            setError("Too many Material.Lights");
                            retVal = false;
                            goto ded_end_read;
                        }
                        if (decor >= mainDef.decorationCount()) mainDef.addDecoration();
                        defn::MaterialDecoration decorDef(mainDef.decoration(decor));
                        int stage = 0;
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_VEC_VAR("Pattern offset", decorDef.def()["patternOffset"], 2)
                            RV_VEC_VAR("Pattern skip", decorDef.def()["patternSkip"], 2)
                            if (ISLABEL("Stage")) {
                                if (stage >= decorDef.stageCount()) {
                                    decorDef.addStage();
                                }
                                Record &st = decorDef.stage(stage);
                                FINDBEGIN;
                                forever {
                                    READLABEL;
                                    RV_INT("Tics", st["tics"])
                                    RV_FLT("Rnd", st["variance"])
                                    RV_VEC_VAR("Offset", st["origin"], 2)
                                    RV_FLT("Distance", st["elevation"])
                                    RV_VEC_VAR("Color", st["color"], 3)
                                    RV_FLT("Radius", st["radius"])
                                    RV_FLT("Halo radius", st["haloRadius"])
                                    if (ISLABEL("Levels")) {
                                        FINDBEGIN;
                                        Vector2f levels;
                                        for (int b = 0; b < 2; ++b) {
                                            float val;
                                            READFLT(val)
                                            levels[b] = de::clamp(0.f, val / 255.0f, 1.f);
                                        }
                                        ReadToken();
                                        st["lightLevels"] = new ArrayValue(levels);
                                    }
                                    else
                                    RV_INT("Flare texture", st["haloTextureIndex"])
                                    RV_URI("Flare map", st["haloTexture"], "LightMaps")
                                    RV_URI("Top map", st["lightmapUp"], "LightMaps")
                                    RV_URI("Bottom map", st["lightmapDown"], "LightMaps")
                                    RV_URI("Side map", st["lightmapSide"], "LightMaps")
                                    RV_END
                                    CHECKSC;
                                }
                                stage++;
                            }
                            else RV_END
                            CHECKSC;
                        }
                        decor++;
                    }
                    else RV_END
                    CHECKSC;
                }
                if (idx > 0) {
                    prevMaterialDefIdx = idx;
                }
            }
            if (ISTOKEN("Model")) {
                Record *prevModel = 0;
                int sub = 0;
                idx = ded->addModel();
                Record &mdl = ded->models[idx];
                if (prevModelDefIdx >= 0) {
                    prevModel = &ded->models[prevModelDefIdx];
                    if (bCopyNext) ded->models.copy(prevModelDefIdx, mdl);
                }
                if (source->custom) mdl.set("custom", true);
                FINDBEGIN;
                forever {
                    READLABEL;
                    RV_STR("ID", mdl[VAR_ID])
                    RV_STR("State", mdl["state"])
                    RV_INT("Off", mdl["off"])
                    RV_STR("Sprite", mdl["sprite"])
                    RV_INT("Sprite frame", mdl["spriteFrame"])
                    RV_FLAGS("Group", mdl["group"], "mg_")
                    RV_INT("Selector", mdl["selector"])
                    RV_FLAGS("Flags", mdl["flags"], "df_")
                    RV_FLT("Inter", mdl["interMark"])
                    RV_INT("Skin tics", mdl["skinTics"])
                    RV_FLT("Resize", mdl["resize"])
                    if (ISLABEL("Scale")) {
                        float scale; READFLT(scale);
                        mdl["scale"] = new ArrayValue(Vector3f(scale, scale, scale));
                    }
                    else
                    RV_VEC_VAR("Scale XYZ", mdl["scale"], 3)
                    RV_FLT_ELEM("Offset", mdl["offset"], 1)
                    RV_VEC_VAR("Offset XYZ", mdl["offset"], 3)
                    RV_VEC_VAR("Interpolate", mdl["interRange"], 2)
                    RV_FLT("Shadow radius", mdl["shadowRadius"])
                    if (ISLABEL("Md2") || ISLABEL("Sub")) {
                        defn::Model mainDef(mdl);
                        if (sub >= mainDef.subCount()) {
                            mainDef.addSub();
                        }
                        DENG_ASSERT(sub < mainDef.subCount());
                        Record &subDef = mainDef.sub(sub);
                        FINDBEGIN;
                        for (;;) {
                            READLABEL;
                            RV_URI("File", subDef["filename"], "Models")
                            RV_STR("Frame", subDef["frame"])
                            RV_INT("Frame range", subDef["frameRange"])
                            RV_BLENDMODE("Blending mode", subDef["blendMode"])
                            RV_INT("Skin", subDef["skin"])
                            RV_URI("Skin file", subDef["skinFilename"], "Models")
                            RV_INT("Skin range", subDef["skinRange"])
                            RV_VEC_VAR("Offset XYZ", subDef["offset"], 3)
                            RV_FLAGS("Flags", subDef["flags"], "df_")
                            RV_FLT("Transparent", subDef["alpha"])
                            RV_FLT("Parm", subDef["parm"])
                            RV_INT("Selskin mask", subDef["selSkinMask"])
                            RV_INT("Selskin shift", subDef["selSkinShift"])
                            RV_NBVEC("Selskins", subDef["selSkins"], 8)
                            RV_URI("Shiny skin", subDef["shinySkin"], "Models")
                            RV_FLT("Shiny", subDef["shiny"])
                            RV_VEC_VAR("Shiny color", subDef["shinyColor"], 3)
                            RV_FLT("Shiny reaction", subDef["shinyReact"])
                            RV_END
                            CHECKSC;
                        }
                        sub++;
                    }
                    else RV_END
                    CHECKSC;
                }
                if (prevModel) {
                    defn::Model(mdl).cleanupAfterParsing(*prevModel);
                }
                prevModelDefIdx = idx;
            }
            if (ISTOKEN("Sound"))
            {   // A new sound.
                ded_sound_t*        snd;
                idx = DED_AddSound(ded, "");
                snd = &ded->sounds[idx];
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("ID", snd->id)
                    RV_STR("Lump", snd->lumpName)
                    RV_STR("Name", snd->name)
                    RV_STR("Link", snd->link)
                    RV_INT("Link pitch", snd->linkPitch)
                    RV_INT("Link volume", snd->linkVolume)
                    RV_INT("Priority", snd->priority)
                    RV_INT("Max channels", snd->channels)
                    RV_INT("Group", snd->group)
                    RV_FLAGS("Flags", snd->flags, "sf_")
                    RV_URI("Ext", &snd->ext, "Sfx")
                    RV_URI("File", &snd->ext, "Sfx")
                    RV_URI("File name", &snd->ext, "Sfx")
                    RV_END
                    CHECKSC;
                }
            }
            if (ISTOKEN("Music")) {
                bool bModify = false;
                Record dummyMusic;
                Record *music = 0;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addMusic();
                    music = &ded->musics[idx];
                }
                else if (!bCopyNext) {
                    ded_stringid_t otherMusicId;
                    READSTR(otherMusicId);
                    ReadToken();
                    idx = ded->getMusicNum(otherMusicId);
                    if (idx >= 0) {
                        music = &ded->musics[idx];
                        bModify = true;
                    }
                    else {
                        if (!source || source->fileName != "[TranslatedMapInfos]") {
                            LOG_RES_WARNING("Ignoring unknown Music \"%s\" in %s on line #%i")
                            << otherMusicId << (source? source->fileName : "?")
                            << (source? source->lineNumber : 0);
                        }
                        defn::Music(dummyMusic).resetToDefaults();
                        music = &dummyMusic;
                    }
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                DENG2_ASSERT(music != 0);
                if (prevMusicDefIdx >= 0 && bCopyNext) {
                    ded->musics.copy(prevMusicDefIdx, *music);
                }
                if (source->custom) music->set("custom", true);
                FINDBEGIN;
                forever {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READSTR((*music)[VAR_ID]);
                    }
                    else RV_STR("Name", (*music)["title"])
                    RV_STR("Lump", (*music)["lumpName"])
                    RV_URI("File name", (*music)["path"], "Music")
                    RV_URI("File", (*music)["path"], "Music")
                    RV_URI("Ext", (*music)["path"], "Music") // Both work.
                    RV_INT("CD track", (*music)["cdTrack"])
                    RV_END
                    CHECKSC;
                }
                if (idx > 0) {
                    prevMusicDefIdx = idx;
                }
            }
            if (ISTOKEN("Sky")) {
                int model = 0;
                idx = ded->addSky();
                Record &sky = ded->skies[idx];
                if (prevSkyDefIdx >= 0 && bCopyNext) {
                    ded->skies.copy(prevSkyDefIdx, sky);
                }
                if (source->custom) sky.set("custom", true);
                FINDBEGIN;
                forever {
                    READLABEL;
                    RV_STR("ID", sky[VAR_ID])
                    RV_FLAGS("Flags", sky["flags"], "sif_")
                    RV_FLT("Height", sky["height"])
                    RV_FLT("Horizon offset", sky["horizonOffset"])
                    RV_VEC_VAR("Light color", sky["color"], 3)
                    if (ISLABEL("Layer 1") || ISLABEL("Layer 2")) {
                        defn::Sky mainDef(sky);
                        Record &layerDef = mainDef.layer(atoi(label+6) - 1);
                        if (source->custom) layerDef.set("custom", true);
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_URI("Material", layerDef["material"], 0)
                            RV_URI("Texture", layerDef["material"], "Textures" )
                            RV_FLAGS("Flags", layerDef["flags"], "slf_")
                            RV_FLT("Offset", layerDef["offset"])
                            RV_FLT("Offset speed", layerDef["offsetSpeed"])
                            RV_FLT("Color limit", layerDef["colorLimit"])
                            RV_END
                            CHECKSC;
                        }
                    }
                    else if (ISLABEL("Model")) {
                        defn::Sky mainDef(sky);
                        if (model == 32/*MAX_SKY_MODELS*/)
                        {
                            setError("Too many Sky models");
                            retVal = false;
                            goto ded_end_read;
                        }
                        if (model >= mainDef.modelCount()) {
                            mainDef.addModel();
                        }
                        DENG_ASSERT(model < mainDef.modelCount());
                        Record &mdlDef = mainDef.model(model);
                        if (source->custom) mdlDef.set("custom", true);
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_STR("ID", mdlDef[VAR_ID])
                            RV_INT("Layer", mdlDef["layer"])
                            RV_FLT("Frame interval", mdlDef["frameInterval"])
                            RV_FLT("Yaw", mdlDef["yaw"])
                            RV_FLT("Yaw speed", mdlDef["yawSpeed"])
                            RV_VEC_VAR("Rotate", mdlDef["rotate"], 2)
                            RV_VEC_VAR("Offset factor", mdlDef["originOffset"], 3)
                            RV_VEC_VAR("Color", mdlDef["color"], 4)
                            RV_STR("Execute", mdlDef["execute"])
                            RV_END
                            CHECKSC;
                        }
                        model++;
                    }
                    else RV_END
                    CHECKSC;
                }
                prevSkyDefIdx = idx;
            }
            if (ISTOKEN("Map")) // Info
            {
                ReadToken();
                if (!ISTOKEN("Info")) {
                    setError("Unknown token 'Map" + String(token) + "'");
                    retVal = false;
                    goto ded_end_read;
                }
                bool bModify = false;
                Record dummyMi;
                Record *mi = nullptr;
                int model = 0;
                ReadToken();
                if (!ISTOKEN("Mods")) {
                    idx = ded->addMapInfo();
                    mi = &ded->mapInfos[idx];
                }
                else if (!bCopyNext) {
                    de::Uri *otherMap = nullptr;
                    READURI(&otherMap, "Maps");
                    ReadToken();
                    idx = ded->getMapInfoNum(*otherMap);
                    if (idx >= 0) {
                        mi = &ded->mapInfos[idx];
                        bModify = true;
                    }
                    else {
                        LOG_RES_WARNING("Ignoring unknown Map \"%s\" in %s on line #%i")
                                << otherMap->asText()
                                << (source? source->fileName : "?")
                                << (source? source->lineNumber : 0);
                    }
                    delete otherMap;
                    if (mi && ISTOKEN("if")) {
                        bool negate = false;
                        bool testCustom = false;
                        forever {
                            ReadToken();
                            if (ISTOKEN("{")) {
                                break;
                            }
                            if (!testCustom) {
                                if (ISTOKEN("not")) {
                                    negate = !negate;
                                }
                                else if (ISTOKEN("custom")) {
                                    testCustom = true;
                                }
                                else RV_END
                            }
                            else RV_END
                        }
                        if (testCustom) {
                            if (mi->getb("custom") != !negate) {
                                mi = nullptr; // skip
                            }
                        }
                        else {
                            setError("Expected condition expression to follow 'if'");
                            retVal = false;
                            goto ded_end_read;
                        }
                    }
                    if (!mi) {
                        defn::MapInfo(dummyMi).resetToDefaults();
                        mi = &dummyMi;
                    }
                }
                else {
                    setError("Cannot both Copy(Previous) and Modify");
                    retVal = false;
                    goto ded_end_read;
                }
                DENG2_ASSERT(mi != 0);
                if (prevMapInfoDefIdx >= 0 && bCopyNext) {
                    ded->mapInfos.copy(prevMapInfoDefIdx, *mi);
                }
                if (source->custom) mi->set("custom", true);
                Record &sky = mi->subrecord("sky");
                if (source->custom) sky.set("custom", true);
                FINDBEGIN;
                forever {
                    READLABEL;
                    if (!bModify && ISLABEL("ID")) {
                        READURI((*mi)[VAR_ID], "Maps");
                    }
                    else RV_STR("Title", (*mi)["title"])
                    RV_STR("Name", (*mi)["title"]) // Alias
                    RV_URI("Title image", (*mi)["titleImage"], "Patches")
                    RV_STR("Author", (*mi)["author"])
                    RV_FLAGS("Flags", (*mi)["flags"], "mif_")
                    RV_STR("Music", (*mi)["music"])
                    RV_FLT("Par time", (*mi)["parTime"])
                    RV_FLT_ELEM("Fog color R", (*mi)["fogColor"], 0)
                    RV_FLT_ELEM("Fog color G", (*mi)["fogColor"], 1)
                    RV_FLT_ELEM("Fog color B", (*mi)["fogColor"], 2)
                    RV_FLT("Fog start", (*mi)["fogStart"])
                    RV_FLT("Fog end", (*mi)["fogEnd"])
                    RV_FLT("Fog density", (*mi)["fogDensity"])
                    RV_STR("Fade Table", (*mi)["fadeTable"])
                    RV_FLT("Ambient light", (*mi)["ambient"])
                    RV_FLT("Gravity", (*mi)["gravity"])
                    RV_STR("Execute", (*mi)["execute"])
                    RV_STR("On setup", (*mi)["onSetup"]) // script function
                    RV_STR("Sky", (*mi)["skyId"])
                    RV_FLT("Sky height", sky["height"])
                    RV_FLT("Horizon offset", sky["horizonOffset"])
                    RV_STR("Intermission background", (*mi)["intermissionBg"])
                    RV_VEC_VAR("Sky light color", sky["color"], 3)
                    if (ISLABEL("Sky Layer 1") || ISLABEL("Sky Layer 2")) {
                        defn::Sky skyDef(sky);
                        Record &layerDef = skyDef.layer(atoi(label+10) - 1);
                        if (source->custom) layerDef.set("custom", true);
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_URI("Material", layerDef["material"], 0)
                            RV_URI("Texture", layerDef["material"], "Textures" )
                            RV_FLAGS("Flags", layerDef["flags"], "slf_")
                            RV_FLT("Offset", layerDef["offset"])
                            RV_FLT("Offset speed", layerDef["offsetSpeed"])
                            RV_FLT("Color limit", layerDef["colorLimit"])
                            RV_END
                            CHECKSC;
                        }
                    }
                    else if (ISLABEL("Sky Model")) {
                        defn::Sky skyDef(sky);
                        if (model == 32/*MAX_SKY_MODELS*/)
                        {
                            setError("Too many Sky models");
                            retVal = false;
                            goto ded_end_read;
                        }
                        if (model >= skyDef.modelCount()) {
                            skyDef.addModel();
                        }
                        DENG_ASSERT(model < skyDef.modelCount());
                        Record &mdlDef = skyDef.model(model);
                        if (source->custom) mdlDef.set("custom", true);
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_STR("ID", mdlDef[VAR_ID])
                            RV_INT("Layer", mdlDef["layer"])
                            RV_FLT("Frame interval", mdlDef["frameInterval"])
                            RV_FLT("Yaw", mdlDef["yaw"])
                            RV_FLT("Yaw speed", mdlDef["yawSpeed"])
                            RV_VEC_VAR("Rotate", mdlDef["rotate"], 2)
                            RV_VEC_VAR("Offset factor", mdlDef["originOffset"], 3)
                            RV_VEC_VAR("Color", mdlDef["color"], 4)
                            RV_STR("Execute", mdlDef["execute"])
                            RV_END
                            CHECKSC;
                        }
                        model++;
                    }
                    else RV_END
                    SKIPSC;
                }
                if (idx > 0) {
                    prevMapInfoDefIdx = idx;
                }
            }
            if (ISTOKEN("Text")) {
                idx = DED_AddText(ded, "");
                ded_text_t *txt = &ded->text[idx];
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("ID", txt->id)
                    if (ISLABEL("Text")) {
                        String buffer;
                        if (ReadString(buffer)) {
                            buffer.replace("\\n", "\n"); // newline escapes are allowed
                            QByteArray bufferUtf8 = buffer.toUtf8();
                            txt->text = (char *) M_Realloc(txt->text, bufferUtf8.length() + 1);
                            qstrcpy(txt->text, bufferUtf8.constData());
                        }
                        else {
                            setError("Syntax error in Text value");
                            retVal = false;
                            goto ded_end_read;
                        }
                    }
                    else RV_END
                    CHECKSC;
                }
            }
            if (ISTOKEN("Texture")) {
                ReadToken();
                if (ISTOKEN("Environment"))
                {   // A new environment.
                    ded_tenviron_t*     tenv;
                    idx = DED_AddTextureEnv(ded, "");
                    tenv = &ded->textureEnv[idx];
                    FINDBEGIN;
                    for (;;) {
                        ded_uri_t *mn;
                        READLABEL;
                        RV_STR("ID", tenv->id)
                        if (ISLABEL("Material") || ISLABEL("Texture") || ISLABEL("Flat")) {
                            ddstring_t schemeName; Str_Init(&schemeName);
                            Str_Set(&schemeName, ISLABEL("Material")? "" : ISLABEL("Texture")? "Textures" : "Flats");
                            mn = tenv->materials.append();
                            FINDBEGIN;
                            for (;;) {
                                READLABEL;
                                RV_URI("ID", &mn->uri, Str_Text(&schemeName))
                                RV_END
                                CHECKSC;
                            }
                            Str_Free(&schemeName);
                        }
                        else RV_END
                        CHECKSC;
                    }
                }
            }
            if (ISTOKEN("Composite")) {
                ReadToken();
                if (ISTOKEN("BitmapFont")) {
                    ded_compositefont_t* cfont;
                    idx = DED_AddCompositeFont(ded, NULL);
                    cfont = &ded->compositeFonts[idx];
                    FINDBEGIN;
                    for (;;) {
                        READLABEL;
                        if (ISLABEL("ID")) {
                            READURI(&cfont->uri, "Game")
                            CHECKSC;
                        }
                        else if (M_IsStringValidInt(label)) {
                            ded_compositefont_mappedcharacter_t* mc = 0;
                            int ascii = atoi(label);
                            if (ascii < 0 || ascii > 255) {
                                setError("Invalid ascii code");
                                retVal = false;
                                goto ded_end_read;
                            }
                            for (int i = 0; i < cfont->charMap.size(); ++i) {
                                if (cfont->charMap[i].ch == (unsigned char) ascii)
                                    mc = &cfont->charMap[i];
                            }
                            if (mc == 0) {
                                mc = cfont->charMap.append();
                                mc->ch = ascii;
                            }
                            FINDBEGIN;
                            for (;;) {
                                READLABEL;
                                RV_URI("Texture", &mc->path, "Patches")
                                RV_END
                                CHECKSC;
                            }
                        }
                        else
                        RV_END
                    }
                }
            }
            if (ISTOKEN("Values")) // String Values
            {
                depth = 0;
                rootStr = (char*) M_Calloc(1); // A null string.
                FINDBEGIN;
                for (;;) {
                    READLABEL_NOBREAK;
                    if (strchr(label, '|')) {
                        setError("Value labels can not include '|' characters (ASCII 124)");
                        retVal = false;
                        goto ded_end_read;
                    }
                    if (ISTOKEN("=")) {
                        String buffer;
                        if (ReadString(buffer)) {
                            idx = DED_AddValue(ded, 0);
                            ded_value_t *val = &ded->values[idx];
                            QByteArray bufferUtf8 = buffer.toUtf8();
                            val->text = (char *) M_Malloc(bufferUtf8.length() + 1);
                            qstrcpy(val->text, bufferUtf8.constData());
                            val->id = (char *) M_Malloc(strlen(rootStr) + strlen(label) + 1);
                            strcpy(val->id, rootStr);
                            strcat(val->id, label);
                        }
                        else {
                            setError("Syntax error in Value string");
                            retVal = false;
                            goto ded_end_read;
                        }
                    }
                    else if (ISTOKEN("{")) {
                        rootStr = (char*) M_Realloc(rootStr, strlen(rootStr) + strlen(label) + 2);
                        strcat(rootStr, label);
                        strcat(rootStr, "|");   // The separator.
                        depth++;
                        continue;
                    }
                    else if (ISTOKEN("}")) {
                        size_t len;
                        if (!depth)
                            break;   // End root depth.
                        depth--;
                        len = strlen(rootStr);
                        rootStr[len-1] = 0; // Remove last |.
                        ptr = strrchr(rootStr, '|');
                        if (ptr) {
                            ptr[1] = 0;
                            rootStr = (char*) M_Realloc(rootStr, strlen(rootStr) + 1);
                        }
                        else {
                            rootStr = (char*) M_Realloc(rootStr, 1);
                            *rootStr = 0;
                        }
                    }
                    else {
                        setError("Illegal token");
                        retVal = false;
                        goto ded_end_read;
                    }
                    CHECKSC;
                }
                M_Free(rootStr);
                rootStr = 0;
            }
            if (ISTOKEN("Detail")) // Detail Texture
            {
                idx = DED_AddDetail(ded, "");
                ded_detailtexture_t *dtl = &ded->details[idx];
                if (prevDetailDefIdx >= 0 && bCopyNext) {
                    ded->details.copyTo(dtl, prevDetailDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_FLAGS("Flags", dtl->flags, "dtf_")
                    if (ISLABEL("Texture")) {
                        READURI(&dtl->material1, "Textures")
                    }
                    else if (ISLABEL("Wall")) // Alias
                    {
                        READURI(&dtl->material1, "Textures")
                    }
                    else if (ISLABEL("Flat")) {
                        READURI(&dtl->material2, "Flats")
                    }
                    else if (ISLABEL("Lump")) {
                        READURI(&dtl->stage.texture, "Lumps")
                    }
                    else if (ISLABEL("File")) {
                        READURI(&dtl->stage.texture, 0)
                    }
                    else
                    RV_FLT("Scale", dtl->stage.scale)
                    RV_FLT("Strength", dtl->stage.strength)
                    RV_FLT("Distance", dtl->stage.maxDistance)
                    RV_END
                    CHECKSC;
                }
                prevDetailDefIdx = idx;
            }
            if (ISTOKEN("Reflection")) // Surface reflection
            {
                ded_reflection_t* ref = 0;
                idx = DED_AddReflection(ded);
                ref = &ded->reflections[idx];
                if (prevRefDefIdx >= 0 && bCopyNext) {
                    ded->reflections.copyTo(ref, prevRefDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_FLAGS("Flags", ref->flags, "rff_")
                    RV_FLT("Shininess", ref->stage.shininess)
                    RV_VEC("Min color", ref->stage.minColor, 3)
                    RV_BLENDMODE("Blending mode", ref->stage.blendMode)
                    RV_URI("Shiny map", &ref->stage.texture, "LightMaps")
                    RV_URI("Mask map", &ref->stage.maskTexture, "LightMaps")
                    RV_FLT("Mask width", ref->stage.maskWidth)
                    RV_FLT("Mask height", ref->stage.maskHeight)
                    if (ISLABEL("Material")) {
                        READURI(&ref->material, 0)
                    }
                    else if (ISLABEL("Texture")) {
                        READURI(&ref->material, "Textures")
                    }
                    else if (ISLABEL("Flat")) {
                        READURI(&ref->material, "Flats")
                    }
                    else RV_END
                    CHECKSC;
                }
                prevRefDefIdx = idx;
            }
            if (ISTOKEN("Generator")) // Particle Generator
            {
                ded_ptcgen_t* gen;
                int sub = 0;
                idx = DED_AddPtcGen(ded, "");
                gen = &ded->ptcGens[idx];
                if (prevGenDefIdx >= 0 && bCopyNext) {
                    ded->ptcGens.copyTo(gen, prevGenDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("State", gen->state)
                    if (ISLABEL("Material")) {
                        READURI(&gen->material, 0)
                    }
                    else if (ISLABEL("Flat")) {
                        READURI(&gen->material, "Flats")
                    }
                    else if (ISLABEL("Texture")) {
                        READURI(&gen->material, "Textures")
                    }
                    else
                    RV_STR("Mobj", gen->type)
                    RV_STR("Alt mobj", gen->type2)
                    RV_STR("Damage mobj", gen->damage)
                    RV_URI("Map", &gen->map, "Maps")
                    RV_FLAGS("Flags", gen->flags, "gnf_")
                    RV_FLT("Speed", gen->speed)
                    RV_FLT("Speed Rnd", gen->speedVariance)
                    RV_VEC("Vector", gen->vector, 3)
                    RV_FLT("Vector Rnd", gen->vectorVariance)
                    RV_FLT("Init vector Rnd", gen->initVectorVariance)
                    RV_VEC("Center", gen->center, 3)
                    RV_INT("Submodel", gen->subModel)
                    RV_FLT("Spawn radius", gen->spawnRadius)
                    RV_FLT("Min spawn radius", gen->spawnRadiusMin)
                    RV_FLT("Distance", gen->maxDist)
                    RV_INT("Spawn age", gen->spawnAge)
                    RV_INT("Max age", gen->maxAge)
                    RV_INT("Particles", gen->particles)
                    RV_FLT("Spawn rate", gen->spawnRate)
                    RV_FLT("Spawn Rnd", gen->spawnRateVariance)
                    RV_INT("Presim", gen->preSim)
                    RV_INT("Alt start", gen->altStart)
                    RV_FLT("Alt Rnd", gen->altStartVariance)
                    RV_VEC("Force axis", gen->forceAxis, 3)
                    RV_FLT("Force radius", gen->forceRadius)
                    RV_FLT("Force", gen->force)
                    RV_VEC("Force origin", gen->forceOrigin, 3)
                    if (ISLABEL("Stage")) {
                        ded_ptcstage_t *st = NULL;
                        if (sub >= gen->stages.size()) {
                            sub = DED_AddPtcGenStage(gen);
                        }
                        st = &gen->stages[sub];
                        FINDBEGIN;
                        for (;;) {
                            READLABEL;
                            RV_FLAGS("Type", st->type, "pt_")
                            RV_INT("Tics", st->tics)
                            RV_FLT("Rnd", st->variance)
                            RV_VEC("Color", st->color, 4)
                            RV_FLT("Radius", st->radius)
                            RV_FLT("Radius rnd", st->radiusVariance)
                            RV_FLAGS("Flags", st->flags, "ptf_")
                            RV_FLT("Bounce", st->bounce)
                            RV_FLT("Gravity", st->gravity)
                            RV_FLT("Resistance", st->resistance)
                            RV_STR("Frame", st->frameName)
                            RV_STR("End frame", st->endFrameName)
                            RV_VEC("Spin", st->spin, 2)
                            RV_VEC("Spin resistance", st->spinResistance, 2)
                            RV_STR("Sound", st->sound.name)
                            RV_FLT("Volume", st->sound.volume)
                            RV_STR("Hit sound", st->hitSound.name)
                            RV_FLT("Hit volume", st->hitSound.volume)
                            RV_VEC("Force", st->vectorForce, 3)
                            RV_END
                            CHECKSC;
                        }
                        sub++;
                    }
                    else RV_END
                    CHECKSC;
                }
                prevGenDefIdx = idx;
            }
            if (ISTOKEN("Finale") || ISTOKEN("InFine")) {
                idx = ded->addFinale();
                Record &fin = ded->finales[idx];
                if (source->custom) fin.set("custom", true);
                FINDBEGIN;
                forever {
                    READLABEL;
                    RV_STR("ID", fin[VAR_ID])
                    RV_URI("Before", fin["before"], "Maps")
                    RV_URI("After", fin["after"], "Maps")
                    RV_INT("Game", dummyInt)
                    if (ISLABEL("Script")) {
                        String buffer; buffer.reserve(1600);
                        FINDBEGIN;
                        ReadToken();
                        while (!ISTOKEN("}") && !source->atEnd) {
                            if (!buffer.isEmpty())
                                buffer += ' ';
                            buffer += String(token);
                            if (ISTOKEN("\"")) {
                                ReadString(buffer, true, true);
                                buffer += '"';
                            }
                            ReadToken();
                        }
                        buffer.squeeze();
                        fin.set("script", buffer);
                    }
                    else RV_END
                    CHECKSC;
                }
            }
            if (ISTOKEN("Decoration")) {
                idx = ded->addDecoration();
                Record &decor = ded->decorations[idx];
                if (prevDecorDefIdx >= 0 && bCopyNext) {
                    ded->decorations.copy(prevDecorDefIdx, decor);
                }
                defn::Decoration mainDef(decor);
                int light = 0;
                FINDBEGIN;
                forever {
                    READLABEL;
                    RV_FLAGS("Flags", decor["flags"], "dcf_")
                    if (ISLABEL("Material")) {
                        READURI(decor["texture"], 0)
                    }
                    else if (ISLABEL("Texture")) {
                        READURI(decor["texture"], "Textures")
                    }
                    else if (ISLABEL("Flat")) {
                        READURI(decor["texture"], "Flats")
                    }
                    else if (ISLABEL("Light")) {
                        if (light == DED_MAX_MATERIAL_DECORATIONS) {
                            setError("Too many Decoration.Lights");
                            retVal = false;
                            goto ded_end_read;
                        }
                        if (light >= mainDef.lightCount()) mainDef.addLight();
                        defn::MaterialDecoration lightDef(mainDef.light(light));
                        Record &st = (lightDef.stageCount()? lightDef.stage(0) : lightDef.addStage());
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_VEC_VAR("Offset", st["origin"], 2)
                            RV_FLT("Distance", st["elevation"])
                            RV_VEC_VAR("Color", st["color"], 3)
                            RV_FLT("Radius", st["radius"])
                            RV_FLT("Halo radius", st["haloRadius"])
                            RV_VEC_VAR("Pattern offset", lightDef.def()["patternOffset"], 2)
                            RV_VEC_VAR("Pattern skip", lightDef.def()["patternSkip"], 2)
                            if (ISLABEL("Levels")) {
                                FINDBEGIN;
                                Vector2f levels;
                                for (int b = 0; b < 2; ++b) {
                                    float val;
                                    READFLT(val)
                                    levels[b] = de::clamp(0.f, val / 255.0f, 1.f);
                                }
                                ReadToken();
                                st["lightLevels"] = new ArrayValue(levels);
                            }
                            else
                            RV_INT("Flare texture", st["haloTextureIndex"])
                            RV_URI("Flare map", st["haloTexture"], "LightMaps")
                            RV_URI("Top map", st["lightmapUp"], "LightMaps")
                            RV_URI("Bottom map", st["lightmapDown"], "LightMaps")
                            RV_URI("Side map", st["lightmapSide"], "LightMaps")
                            RV_END
                            CHECKSC;
                        }
                        light++;
                    }
                    else RV_END
                    CHECKSC;
                }
                prevDecorDefIdx = idx;
            }
            if (ISTOKEN("Group")) {
                idx = DED_AddGroup(ded);
                ded_group_t *grp = &ded->groups[idx];
                int sub = 0;
                FINDBEGIN;
                forever {
                    READLABEL;
                    if (ISLABEL("Texture") || ISLABEL("Flat")) {
                        bool const haveTexture = ISLABEL("Texture");
                        if (sub >= grp->members.size()) {
                            sub = DED_AddGroupMember(grp);
                        }
                        ded_group_member_t *memb = &grp->members[sub];
                        FINDBEGIN;
                        forever {
                            READLABEL;
                            RV_URI("ID", &memb->material, (haveTexture? "Textures" : "Flats"))
                            if (ISLABEL("Tics")) {
                                READINT(memb->tics);
                                if (memb->tics < 0) {
                                    LOG_RES_WARNING("Invalid Group.%s.Tics: %i (< min: 0) in \"%s\" on line #%i"
                                                    "\nWill ignore this Group if used for Material animation")
                                            << (haveTexture ? "Texture" : "Flat")
                                            << memb->tics
                                            << (source ? source->fileName : "?") << (source ? source->lineNumber : 0);
                                }
                            }
                            else
                            RV_INT("Random", memb->randomTics)
                            RV_END
                            CHECKSC;
                        }
                        ++sub;
                    }
                    else RV_FLAGS("Flags", grp->flags, "tgf_")
                    RV_END
                    CHECKSC;
                }
            }
            /*if (ISTOKEN("XGClass"))     // XG Class
            {
                idx = DED_AddXGClass(ded);
                xgc = ded->xgClasses + idx;
                sub = 0;
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_STR("ID", xgc->id)
                    RV_STR("Name", xgc->name)
                    if (ISLABEL("Property")) {
                        ded_xgclass_property_t *xgcp = NULL;
                        if (sub >= xgc->properties_count.num) {
                            sub = DED_AddXGClassProperty(xgc);
                        }
                        xgcp = &xgc->properties[sub];
                        FINDBEGIN;
                        for (;;) {
                            READLABEL;
                            RV_FLAGS("ID", xgcp->id, "xgcp_")
                            RV_FLAGS("Flags", xgcp->flags, "xgcpf_")
                            RV_STR("Name", xgcp->name)
                            RV_STR("Flag Group", xgcp->flagprefix)
                            RV_END
                            CHECKSC;
                        }
                        sub++;
                    }
                    else RV_END
                    CHECKSC;
                }
            }*/
            if (ISTOKEN("Line")) // Line Type
            {
                ded_linetype_t* l;
                idx = DED_AddLineType(ded, 0);
                l = &ded->lineTypes[idx];
                if (prevLineTypeDefIdx >= 0 && bCopyNext) {
                    ded->lineTypes.copyTo(l, prevLineTypeDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_INT("ID", l->id)
                    RV_STR("Comment", l->comment)
                    RV_FLAGS("Flags", l->flags[0], "ltf_")
                    RV_FLAGS("Flags2", l->flags[1], "ltf2_")
                    RV_FLAGS("Flags3", l->flags[2], "ltf3_")
                    RV_FLAGS("Class", l->lineClass, "ltc_")
                    RV_FLAGS("Type", l->actType, "lat_")
                    RV_INT("Count", l->actCount)
                    RV_FLT("Time", l->actTime)
                    RV_INT("Act tag", l->actTag)
                    RV_INT("Ap0", l->aparm[0])
                    RV_INT("Ap1", l->aparm[1])
                    RV_INT("Ap2", l->aparm[2])
                    RV_INT("Ap3", l->aparm[3])
                    RV_FLAGS("Ap4", l->aparm[4], "lref_")
                    RV_INT("Ap5", l->aparm[5])
                    RV_FLAGS("Ap6", l->aparm[6], "lref_")
                    RV_INT("Ap7", l->aparm[7])
                    RV_INT("Ap8", l->aparm[8])
                    RV_STR("Ap9", l->aparm9)
                    RV_INT("Health above", l->aparm[0])
                    RV_INT("Health below", l->aparm[1])
                    RV_INT("Power above", l->aparm[2])
                    RV_INT("Power below", l->aparm[3])
                    RV_FLAGS("Line act lref", l->aparm[4], "lref_")
                    RV_INT("Line act lrefd", l->aparm[5])
                    RV_FLAGS("Line inact lref", l->aparm[6], "lref_")
                    RV_INT("Line inact lrefd", l->aparm[7])
                    RV_INT("Color", l->aparm[8])
                    RV_STR("Thing type", l->aparm9)
                    RV_FLT("Ticker start time", l->tickerStart)
                    RV_FLT("Ticker end time", l->tickerEnd)
                    RV_INT("Ticker tics", l->tickerInterval)
                    RV_STR("Act sound", l->actSound)
                    RV_STR("Deact sound", l->deactSound)
                    RV_INT("Event chain", l->evChain)
                    RV_INT("Act chain", l->actChain)
                    RV_INT("Deact chain", l->deactChain)
                    RV_FLAGS("Wall section", l->wallSection, "lws_")
                    if (ISLABEL("Act material")) {
                        READURI(&l->actMaterial, 0)
                    }
                    else if (ISLABEL("Act texture")) // Alias
                    {
                        READURI(&l->actMaterial, "Textures")
                    }
                    else if (ISLABEL("Deact material")) {
                        READURI(&l->deactMaterial, 0)
                    }
                    else if (ISLABEL("Deact texture")) // Alias
                    {
                        READURI(&l->deactMaterial, "Textures")
                    }
                    else
                    RV_INT("Act type", l->actLineType)
                    RV_INT("Deact type", l->deactLineType)
                    RV_STR("Act message", l->actMsg)
                    RV_STR("Deact message", l->deactMsg)
                    RV_FLT("Texmove angle", l->materialMoveAngle)
                    RV_FLT("Materialmove angle", l->materialMoveAngle) // Alias
                    RV_FLT("Texmove speed", l->materialMoveSpeed)
                    RV_FLT("Materialmove speed", l->materialMoveSpeed) // Alias
                    RV_STR_INT("Ip0", l->iparmStr[0], l->iparm[0])
                    RV_STR_INT("Ip1", l->iparmStr[1], l->iparm[1])
                    RV_STR_INT("Ip2", l->iparmStr[2], l->iparm[2])
                    RV_STR_INT("Ip3", l->iparmStr[3], l->iparm[3])
                    RV_STR_INT("Ip4", l->iparmStr[4], l->iparm[4])
                    RV_STR_INT("Ip5", l->iparmStr[5], l->iparm[5])
                    RV_STR_INT("Ip6", l->iparmStr[6], l->iparm[6])
                    RV_STR_INT("Ip7", l->iparmStr[7], l->iparm[7])
                    RV_STR_INT("Ip8", l->iparmStr[8], l->iparm[8])
                    RV_STR_INT("Ip9", l->iparmStr[9], l->iparm[9])
                    RV_STR_INT("Ip10", l->iparmStr[10], l->iparm[10])
                    RV_STR_INT("Ip11", l->iparmStr[11], l->iparm[11])
                    RV_STR_INT("Ip12", l->iparmStr[12], l->iparm[12])
                    RV_STR_INT("Ip13", l->iparmStr[13], l->iparm[13])
                    RV_STR_INT("Ip14", l->iparmStr[14], l->iparm[14])
                    RV_STR_INT("Ip15", l->iparmStr[15], l->iparm[15])
                    RV_STR_INT("Ip16", l->iparmStr[16], l->iparm[16])
                    RV_STR_INT("Ip17", l->iparmStr[17], l->iparm[17])
                    RV_STR_INT("Ip18", l->iparmStr[18], l->iparm[18])
                    RV_STR_INT("Ip19", l->iparmStr[19], l->iparm[19])
                    RV_FLT("Fp0", l->fparm[0])
                    RV_FLT("Fp1", l->fparm[1])
                    RV_FLT("Fp2", l->fparm[2])
                    RV_FLT("Fp3", l->fparm[3])
                    RV_FLT("Fp4", l->fparm[4])
                    RV_FLT("Fp5", l->fparm[5])
                    RV_FLT("Fp6", l->fparm[6])
                    RV_FLT("Fp7", l->fparm[7])
                    RV_FLT("Fp8", l->fparm[8])
                    RV_FLT("Fp9", l->fparm[9])
                    RV_FLT("Fp10", l->fparm[10])
                    RV_FLT("Fp11", l->fparm[11])
                    RV_FLT("Fp12", l->fparm[12])
                    RV_FLT("Fp13", l->fparm[13])
                    RV_FLT("Fp14", l->fparm[14])
                    RV_FLT("Fp15", l->fparm[15])
                    RV_FLT("Fp16", l->fparm[16])
                    RV_FLT("Fp17", l->fparm[17])
                    RV_FLT("Fp18", l->fparm[18])
                    RV_FLT("Fp19", l->fparm[19])
                    RV_STR("Sp0", l->sparm[0])
                    RV_STR("Sp1", l->sparm[1])
                    RV_STR("Sp2", l->sparm[2])
                    RV_STR("Sp3", l->sparm[3])
                    RV_STR("Sp4", l->sparm[4])
                    if (l->lineClass) {
                        int i;
                        for (i = 0; i < 20; ++i) {
                            xgclassparm_t const& iParm = xgClassLinks[l->lineClass].iparm[i];
                            if (!iParm.name[0]) continue;
                            if (!ISLABEL(iParm.name)) continue;
                            if (iParm.flagPrefix[0]) {
                                READFLAGS(l->iparm[i], iParm.flagPrefix)
                            }
                            else {
                                READSTRING(l->iparmStr[i], l->iparm[i])
                            }
                            break;
                        }
                        if (i == 20) RV_END
                    } else
                    RV_END
                    CHECKSC;
                }
                prevLineTypeDefIdx = idx;
            }
            if (ISTOKEN("Sector")) // Sector Type
            {
                ded_sectortype_t*   sec;
                idx = DED_AddSectorType(ded, 0);
                sec = &ded->sectorTypes[idx];
                if (prevSectorTypeDefIdx >= 0 && bCopyNext) {
                    ded->sectorTypes.copyTo(sec, prevSectorTypeDefIdx);
                }
                FINDBEGIN;
                for (;;) {
                    READLABEL;
                    RV_INT("ID", sec->id)
                    RV_STR("Comment", sec->comment)
                    RV_FLAGS("Flags", sec->flags, "stf_")
                    RV_INT("Act tag", sec->actTag)
                    RV_INT("Floor chain", sec->chain[0])
                    RV_INT("Ceiling chain", sec->chain[1])
                    RV_INT("Inside chain", sec->chain[2])
                    RV_INT("Ticker chain", sec->chain[3])
                    RV_FLAGS("Floor chain flags", sec->chainFlags[0], "scef_")
                    RV_FLAGS("Ceiling chain flags", sec->chainFlags[1], "scef_")
                    RV_FLAGS("Inside chain flags", sec->chainFlags[2], "scef_")
                    RV_FLAGS("Ticker chain flags", sec->chainFlags[3], "scef_")
                    RV_FLT("Floor chain start time", sec->start[0])
                    RV_FLT("Ceiling chain start time", sec->start[1])
                    RV_FLT("Inside chain start time", sec->start[2])
                    RV_FLT("Ticker chain start time", sec->start[3])
                    RV_FLT("Floor chain end time", sec->end[0])
                    RV_FLT("Ceiling chain end time", sec->end[1])
                    RV_FLT("Inside chain end time", sec->end[2])
                    RV_FLT("Ticker chain end time", sec->end[3])
                    RV_FLT("Floor chain min interval", sec->interval[0][0])
                    RV_FLT("Ceiling chain min interval", sec->interval[1][0])
                    RV_FLT("Inside chain min interval", sec->interval[2][0])
                    RV_FLT("Ticker chain min interval", sec->interval[3][0])
                    RV_FLT("Floor chain max interval", sec->interval[0][1])
                    RV_FLT("Ceiling chain max interval", sec->interval[1][1])
                    RV_FLT("Inside chain max interval", sec->interval[2][1])
                    RV_FLT("Ticker chain max interval", sec->interval[3][1])
                    RV_INT("Floor chain count", sec->count[0])
                    RV_INT("Ceiling chain count", sec->count[1])
                    RV_INT("Inside chain count", sec->count[2])
                    RV_INT("Ticker chain count", sec->count[3])
                    RV_STR("Ambient sound", sec->ambientSound)
                    RV_FLT("Ambient min interval", sec->soundInterval[0])
                    RV_FLT("Ambient max interval", sec->soundInterval[1])
                    RV_FLT("Floor texmove angle", sec->materialMoveAngle[0])
                    RV_FLT("Floor materialmove angle", sec->materialMoveAngle[0]) // Alias
                    RV_FLT("Ceiling texmove angle", sec->materialMoveAngle[1])
                    RV_FLT("Ceiling materialmove angle", sec->materialMoveAngle[1]) // Alias
                    RV_FLT("Floor texmove speed", sec->materialMoveSpeed[0])
                    RV_FLT("Floor materialmove speed", sec->materialMoveSpeed[0]) // Alias
                    RV_FLT("Ceiling texmove speed", sec->materialMoveSpeed[1])
                    RV_FLT("Ceiling materialmove speed", sec->materialMoveSpeed[1]) // Alias
                    RV_FLT("Wind angle", sec->windAngle)
                    RV_FLT("Wind speed", sec->windSpeed)
                    RV_FLT("Vertical wind", sec->verticalWind)
                    RV_FLT("Gravity", sec->gravity)
                    RV_FLT("Friction", sec->friction)
                    RV_STR("Light fn", sec->lightFunc)
                    RV_INT("Light fn min tics", sec->lightInterval[0])
                    RV_INT("Light fn max tics", sec->lightInterval[1])
                    RV_STR("Red fn", sec->colFunc[0])
                    RV_STR("Green fn", sec->colFunc[1])
                    RV_STR("Blue fn", sec->colFunc[2])
                    RV_INT("Red fn min tics", sec->colInterval[0][0])
                    RV_INT("Red fn max tics", sec->colInterval[0][1])
                    RV_INT("Green fn min tics", sec->colInterval[1][0])
                    RV_INT("Green fn max tics", sec->colInterval[1][1])
                    RV_INT("Blue fn min tics", sec->colInterval[2][0])
                    RV_INT("Blue fn max tics", sec->colInterval[2][1])
                    RV_STR("Floor fn", sec->floorFunc)
                    RV_FLT("Floor fn scale", sec->floorMul)
                    RV_FLT("Floor fn offset", sec->floorOff)
                    RV_INT("Floor fn min tics", sec->floorInterval[0])
                    RV_INT("Floor fn max tics", sec->floorInterval[1])
                    RV_STR("Ceiling fn", sec->ceilFunc)
                    RV_FLT("Ceiling fn scale", sec->ceilMul)
                    RV_FLT("Ceiling fn offset", sec->ceilOff)
                    RV_INT("Ceiling fn min tics", sec->ceilInterval[0])
                    RV_INT("Ceiling fn max tics", sec->ceilInterval[1])
                    RV_END
                    CHECKSC;
                }
                prevSectorTypeDefIdx = idx;
            }
            bCopyNext = false;
        }
    ded_end_read:
        M_Free(rootStr);
        DED_CloseReader();
        return retVal;
    }
    void DED_Include(char const *fileName, String parentDirectory) {
        ddstring_t tmp;
        Str_InitStd(&tmp); Str_Set(&tmp, fileName);
        F_FixSlashes(&tmp, &tmp);
        F_ExpandBasePath(&tmp, &tmp);
        if (!F_IsAbsolute(&tmp)) {
            Str_PrependChar(&tmp, '/');
            Str_Prepend(&tmp, parentDirectory.toUtf8().constData());
        }
        Def_ReadProcessDED(ded, String(Str_Text(&tmp)));
        Str_Free(&tmp);
        strncpy(token, "", MAX_TOKEN_LEN);
    }
};
DEDParser::DEDParser(ded_t *ded) : d(new Impl(this)) {
    d->ded = ded;
}
int DEDParser::parse(char const *buffer, String sourceFile, bool sourceIsCustom) {
    return d->readData(buffer, sourceFile, sourceIsCustom);
}
static String const VAR_ORDER = "order";
DENG2_PIMPL(DEDRegister)
, DENG2_OBSERVES(Record, Deletion)
, DENG2_OBSERVES(Record, Addition)
, DENG2_OBSERVES(Record, Removal)
, DENG2_OBSERVES(Variable, ChangeFrom) {
    Record *names;
    ArrayValue *orderArray;
    struct Key {
        LookupFlags flags;
        Key(LookupFlags const &f = DefaultLookup) : flags(f) {}
    };
    typedef QMap<String, Key> Keys;
    Keys keys;
    QMap<Variable *, Record *> parents;
    Impl(Public *i, Record &rec) : Base(i), names(&rec) {
        names->audienceForDeletion() += this;
        orderArray = &names->addArray(VAR_ORDER).array();
    }
    void recordBeingDeleted(Record &DENG2_DEBUG_ONLY(record)) {
        DENG2_ASSERT(names == &record);
        names = nullptr;
        orderArray = nullptr;
    }
    void clear() {
        order().clear();
#ifdef DENG2_DEBUG
        DENG2_ASSERT(parents.isEmpty());
        foreach (String const &key, keys.keys()) {
            DENG2_ASSERT(lookup(key).size() == 0);
        }
#endif
    }
    void addKey(String const &name, LookupFlags const &flags) {
        keys.insert(name, Key(flags));
        names->addDictionary(name + "Lookup");
    }
    ArrayValue &order() {
        DENG2_ASSERT(orderArray != nullptr);
        return *orderArray;
    }
    ArrayValue const &order() const {
        DENG2_ASSERT(orderArray != nullptr);
        return *orderArray;
    }
    DictionaryValue &lookup(String const &keyName) {
        return (*names)[keyName + "Lookup"].value<DictionaryValue>();
    }
    DictionaryValue const &lookup(String const &keyName) const {
        return (*names)[keyName + "Lookup"].value<DictionaryValue>();
    }
    template <typename Type>
    Type lookupOperation(String const &key, String value,
                         std::function<Type (DictionaryValue const &, String)> operation) const {
        auto foundKey = keys.constFind(key);
        if (foundKey == keys.constEnd()) return 0;
        if (!foundKey.value().flags.testFlag(CaseSensitive)) {
            value = value.lower();
        }
        return operation(lookup(key), value);
    }
    Record const *tryFind(String const &key, String const &value) const {
        return lookupOperation<Record const *>(key, value,
            [] (DictionaryValue const &lut, String v) -> Record const * {
                TextValue const val(v);
                auto i = lut.elements().find(DictionaryValue::ValueRef(&val));
                if (i == lut.elements().end()) return 0; // Value not in dictionary.
                return i->second->as<RecordValue>().record();
            });
    }
    bool has(String const &key, String const &value) const {
        return lookupOperation<bool>(key, value, [] (DictionaryValue const &lut, String v) {
            return lut.contains(TextValue(v));
        });
    }
    Record &append() {
        Record *sub = new Record;
        sub->set(defn::Definition::VAR_ORDER, int(order().size())).setReadOnly();
        sub->audienceForAddition() += this;
        sub->audienceForRemoval()  += this;
        order().add(new RecordValue(sub, RecordValue::OwnsRecord));
        return *sub;
    }
    bool isEmptyKeyValue(Value const &value) const {
        return is<TextValue>(value) && value.asText().isEmpty();
    }
    bool isValidKeyValue(Value const &value) const {
        if (isEmptyKeyValue(value)) return false;
        return true;
    }
    bool addToLookup(String const &key, Value const &value, Record &def) {
        if (!isValidKeyValue(value))
            return false;
        String valText = value.asText();
        DENG2_ASSERT(!valText.isEmpty());
        DENG2_ASSERT(keys.contains(key));
        if (!keys[key].flags.testFlag(CaseSensitive)) {
            valText = valText.lower();
        }
        DictionaryValue &dict = lookup(key);
        if (keys[key].flags.testFlag(OnlyFirst)) {
            if (dict.contains(TextValue(valText))) return false;
        }
        dict.add(new TextValue(valText), new RecordValue(&def));
        return true;
    }
    bool removeFromLookup(String const &key, Value const &value, Record &def) {
        if (!isValidKeyValue(value))
            return false;
        String valText = value.asText();
        DENG2_ASSERT(!valText.isEmpty());
        DENG2_ASSERT(keys.contains(key));
        if (!keys[key].flags.testFlag(CaseSensitive)) {
            valText = valText.lower();
        }
        DictionaryValue &dict = lookup(key);
        if (dict.contains(TextValue(valText))) {
            Value const &indexed = dict.element(TextValue(valText));
            Record const *indexedDef = indexed.as<RecordValue>().record();
            if (!indexedDef || indexedDef == &def) {
                dict.remove(TextValue(valText));
                return true;
            }
        }
        return false;
    }
    void recordMemberAdded(Record &def, Variable &key) {
        if (keys.contains(key.name())) {
            if (addToLookup(key.name(), key.value(), def) ||
               isEmptyKeyValue(key.value())) {
                parents.insert(&key, &def);
                key.audienceForChangeFrom() += this;
            }
        }
    }
    void recordMemberRemoved(Record &def, Variable &key) {
        if (keys.contains(key.name())) {
            key.audienceForChangeFrom() -= this;
            parents.remove(&key);
            removeFromLookup(key.name(), key.value(), def);
        }
    }
    void variableValueChangedFrom(Variable &key, Value const &oldValue, Value const &newValue) {
        DENG2_ASSERT(parents.contains(&key));
        removeFromLookup(key.name(), oldValue, *parents[&key]);
        addToLookup(key.name(), newValue, *parents[&key]);
    }
};
DEDRegister::DEDRegister(Record &names) : d(new Impl(this, names))
{}
void DEDRegister::addLookupKey(String const &variableName, LookupFlags const &flags) {
    d->addKey(variableName, flags);
}
void DEDRegister::clear() {
    d->clear();
}
Record &DEDRegister::append() {
    return d->append();
}
Record &DEDRegister::appendCopy(int index) {
    return copy(index, append());
}
Record &DEDRegister::copy(int fromIndex, Record &to) {
    QStringList omitted;
    omitted << "__.*"; // double-underscore
    DENG2_FOR_EACH_CONST(Impl::Keys, i, d->keys) {
        if (i.value().flags.testFlag(AllowCopy)) continue;
        omitted << i.key();
    }
    return to.assign((*this)[fromIndex], QRegExp(omitted.join("|")));
}
int DEDRegister::size() const {
    return d->order().size();
}
bool DEDRegister::has(String const &key, String const &value) const {
    return d->has(key, value);
}
Record &DEDRegister::operator [] (int index) {
    return *d->order().at(index).as<RecordValue>().record();
}
Record const &DEDRegister::operator [] (int index) const {
    return *d->order().at(index).as<RecordValue>().record();
}
Record *DEDRegister::tryFind(String const &key, String const &value) {
    return const_cast<Record *>(d->tryFind(key, value));
}
Record const *DEDRegister::tryFind(String const &key, String value) const {
    return d->tryFind(key, value);
}
Record &DEDRegister::find(String const &key, String const &value) {
    return const_cast<Record &>(const_cast<DEDRegister const *>(this)->find(key, value));
}
Record const &DEDRegister::find(String const &key, String const &value) const {
    if (!d->keys.contains(key)) {
        throw UndefinedKeyError("DEDRegister::find", "Key '" + key + "' not defined");
    }
    Record const *rec = tryFind(key, value);
    if (!rec) {
        throw NotFoundError("DEDRegister::find", key + " '" + value + "' not found");
    }
    return *rec;
}
DictionaryValue const &DEDRegister::lookup(String const &key) const {
    if (!d->keys.contains(key)) {
        throw UndefinedKeyError("DEDRegister::lookup", "Key '" + key + "' not defined");
    }
    return d->lookup(key);
}
namespace defn {
const String Definition::VAR_ID("id");
const String Definition::VAR_ORDER("__order__");
Definition::~Definition()
{}
Record &Definition::def() {
    return const_cast<Record &>(accessedRecord());
}
Record const &Definition::def() const {
    return accessedRecord();
}
String Definition::id() const {
    if (!accessedRecordPtr()) return {};
    return gets(VAR_ID);
}
int Definition::order() const {
    if (!accessedRecordPtr()) return -1;
    return geti(VAR_ORDER);
}
Definition::operator bool() const {
    return accessedRecordPtr() != 0;
}
void Definition::resetToDefaults() {
    def().addBoolean("custom", false);
}
} // namespace defn
namespace defn {
void Episode::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText(VAR_ID, "");
    def().addText("startMap", "Maps:"); // URI. Unknown.
    def().addText("title", "Untitled");
    def().addText("menuHelpInfo", "");  // None.
    def().addText("menuImage", "");     // URI. None.
    def().addText("menuShortcut", "");  // Key name. None.
    def().addArray("hub", new ArrayValue);
    def().addArray("map", new ArrayValue);
}
Record &Episode::addHub() {
    Record *hub = new Record;
    hub->addBoolean("custom", false);
    hub->addText (VAR_ID, "");
    hub->addArray("map", new ArrayValue);
    def()["hub"].array().add(new RecordValue(hub, RecordValue::OwnsRecord));
    return *hub;
}
int Episode::hubCount() const {
    return int(geta("hub").size());
}
bool Episode::hasHub(int index) const {
    return index >= 0 && index < hubCount();
}
Record &Episode::hub(int index) {
    return *def().geta("hub")[index].as<RecordValue>().record();
}
Record const &Episode::hub(int index) const {
    return *geta("hub")[index].as<RecordValue>().record();
}
Record *Episode::tryFindHubByMapId(String const &mapId) {
    de::Uri const mapUri(mapId, RC_NULL);
    if (!mapUri.path().isEmpty()) {
        for (int i = 0; i < hubCount(); ++i) {
            Record &hubRec = hub(i);
            foreach (Value *mapIt, hubRec.geta("map").elements()) {
                Record &mgNodeDef = mapIt->as<RecordValue>().dereference();
                if (mapUri == de::makeUri(mgNodeDef.gets(VAR_ID))) {
                    return &hubRec;
                }
            }
        }
    }
    return 0; // Not found.
}
Record *Episode::tryFindMapGraphNode(String const &mapId) {
    de::Uri const mapUri(mapId, RC_NULL);
    if (!mapUri.path().isEmpty()) {
        for (int i = 0; i < hubCount(); ++i) {
            Record const &hubRec = hub(i);
            foreach (Value *mapIt, hubRec.geta("map").elements()) {
                Record &mgNodeDef = mapIt->as<RecordValue>().dereference();
                if (mapUri == de::makeUri(mgNodeDef.gets(VAR_ID))) {
                    return &mgNodeDef;
                }
            }
        }
        foreach (Value *mapIt, geta("map").elements()) {
            Record &mgNodeDef = mapIt->as<RecordValue>().dereference();
            if (mapUri == de::makeUri(mgNodeDef.gets(VAR_ID))) {
                return &mgNodeDef;
            }
        }
    }
    return 0; // Not found.
}
de::Record *Episode::tryFindMapGraphNodeByWarpNumber(int warpNumber) {
    if (warpNumber > 0) {
        for (int i = 0; i < hubCount(); ++i) {
            Record const &hubRec = hub(i);
            foreach (Value *mapIt, hubRec.geta("map").elements()) {
                Record &mgNodeDef = mapIt->as<RecordValue>().dereference();
                if (mgNodeDef.geti("warpNumber") == warpNumber) {
                    return &mgNodeDef;
                }
            }
        }
        foreach (Value *mapIt, geta("map").elements()) {
            Record &mgNodeDef = mapIt->as<RecordValue>().dereference();
            if (mgNodeDef.geti("warpNumber") == warpNumber) {
                return &mgNodeDef;
            }
        }
    }
    return 0; // Not found.
}
} // namespace defn
namespace defn {
void Finale::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText(VAR_ID, "");
    def().addText("before", "");
    def().addText("after", "");
    def().addText("script", "");
}
} // namespace defn
namespace defn {
void MapGraphNode::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addNumber("warpNumber", 0);
    def().addArray ("exit", new ArrayValue);
}
Record &MapGraphNode::addExit() {
    Record *exit = new Record;
    exit->addBoolean("custom", false);
    exit->addText(VAR_ID, "");
    exit->addText("targetMap", "");
    def()["exit"].array().add(new RecordValue(exit, RecordValue::OwnsRecord));
    return *exit;
}
int MapGraphNode::exitCount() const {
    return int(geta("exit").size());
}
bool MapGraphNode::hasExit(int index) const {
    return index >= 0 && index < exitCount();
}
Record &MapGraphNode::exit(int index) {
    return *def().geta("exit")[index].as<RecordValue>().record();
}
Record const &MapGraphNode::exit(int index) const {
    return *geta("exit")[index].as<RecordValue>().record();
}
} // namespace defn
namespace defn {
void MapInfo::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addText  ("title", "Untitled");
    def().addText  ("titleImage", "");
    def().addText  ("author", "Unknown");
    def().addNumber("flags", 0);
    def().addText  ("music", "");
    def().addNumber("parTime", -1); // unknown
    def().addArray ("fogColor", new ArrayValue(Vector3f(DEFAULT_FOG_COLOR_RED, DEFAULT_FOG_COLOR_GREEN, DEFAULT_FOG_COLOR_BLUE)));
    def().addNumber("fogStart", DEFAULT_FOG_START);
    def().addNumber("fogEnd", DEFAULT_FOG_END);
    def().addNumber("fogDensity", DEFAULT_FOG_DENSITY);
    def().addText  ("fadeTable", "");
    def().addNumber("ambient", 0);
    def().addNumber("gravity", 1);
    def().addText  ("skyId", "");
    def().addText  ("execute", "");
    def().addText  ("onSetup", "");
    def().addText  (QStringLiteral("intermissionBg"), "");
    QScopedPointer<Record> sky(new Record);
    Sky(*sky).resetToDefaults();
    def().add      ("sky", sky.take());
}
} // namespace defn
namespace defn {
void MaterialDecoration::resetToDefaults() {
    Definition::resetToDefaults();
    def().addArray("patternOffset", new ArrayValue(Vector2i()));
    def().addArray("patternSkip", new ArrayValue(Vector2i()));
    def().addArray("stage", new ArrayValue);
}
Record &MaterialDecoration::addStage() {
    auto *stage = new Record;
    stage->addNumber("tics", 0);
    stage->addNumber("variance", 0);                              // Time.
    stage->addArray ("origin", new ArrayValue(Vector2f()));       // Surface-relative offset.
    stage->addNumber("elevation", 1);                             // Distance from the surface.
    stage->addArray ("color", new ArrayValue(Vector3f()));        // Light color. (0,0,0) means not visible during this stage.
    stage->addNumber("radius", 1);                                // Dynamic light radius (-1 = no light).
    stage->addArray ("lightLevels", new ArrayValue(Vector2f()));  // Fade by sector lightlevel.
    stage->addText  ("lightmapUp", "");                           // Uri. None.
    stage->addText  ("lightmapDown", "");                         // Uri. None.
    stage->addText  ("lightmapSide", "");                         // Uri. None.
    stage->addNumber("haloRadius", 0);                            // Halo radius (zero = no halo).
    stage->addText  ("haloTexture", "");                          // Uri. None.
    stage->addNumber("haloTextureIndex", 0);                      // Overrides haloTexture
    def()["stage"].array().add(new RecordValue(stage, RecordValue::OwnsRecord));
    return *stage;
}
int MaterialDecoration::stageCount() const {
    return int(geta("stage").size());
}
bool MaterialDecoration::hasStage(int index) const {
    return index >= 0 && index < stageCount();
}
Record &MaterialDecoration::stage(int index) {
    return *def().geta("stage")[index].as<RecordValue>().record();
}
Record const &MaterialDecoration::stage(int index) const {
    return *geta("stage")[index].as<RecordValue>().record();
}
void MaterialLayer::resetToDefaults() {
    Definition::resetToDefaults();
    def().addArray("stage", new ArrayValue);
}
Record &MaterialLayer::addStage() {
    auto *stage = new Record;
    stage->addText  ("texture", "");  // Uri. None.
    stage->addNumber("tics", 0);
    stage->addNumber("variance", 0);  // Time.
    stage->addNumber("glowStrength", 0);
    stage->addNumber("glowStrengthVariance", 0);
    stage->addArray ("texOrigin", new ArrayValue(Vector2f()));
    def()["stage"].array().add(new RecordValue(stage, RecordValue::OwnsRecord));
    return *stage;
}
int MaterialLayer::stageCount() const {
    return int(geta("stage").size());
}
bool MaterialLayer::hasStage(int index) const {
    return index >= 0 && index < stageCount();
}
Record &MaterialLayer::stage(int index) {
    return *def().geta("stage")[index].as<RecordValue>().record();
}
Record const &MaterialLayer::stage(int index) const {
    return *geta("stage")[index].as<RecordValue>().record();
}
void Material::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText   (VAR_ID, "");  // URI. Unknown.
    def().addBoolean("autoGenerated", false);
    def().addNumber ("flags", 0);
    def().addArray  ("dimensions", new ArrayValue(Vector2i()));
    def().addArray  ("decoration", new ArrayValue);
    def().addArray  ("layer", new ArrayValue);
}
Record &Material::addDecoration() {
    auto *decor = new Record;
    MaterialDecoration(*decor).resetToDefaults();
    def()["decoration"].array().add(new RecordValue(decor, RecordValue::OwnsRecord));
    return *decor;
}
int Material::decorationCount() const {
    return int(geta("decoration").size());
}
bool Material::hasDecoration(int index) const {
    return index >= 0 && index < decorationCount();
}
Record &Material::decoration(int index) {
    return *def().geta("decoration")[index].as<RecordValue>().record();
}
Record const &Material::decoration(int index) const {
    return *geta("decoration")[index].as<RecordValue>().record();
}
Record &Material::addLayer() {
    auto *layer = new Record;
    MaterialLayer(*layer).resetToDefaults();
    def()["layer"].array().add(new RecordValue(layer, RecordValue::OwnsRecord));
    return *layer;
}
int Material::layerCount() const {
    return int(geta("layer").size());
}
bool Material::hasLayer(int index) const {
    return index >= 0 && index < layerCount();
}
Record &Material::layer(int index) {
    return *def().geta("layer")[index].as<RecordValue>().record();
}
Record const &Material::layer(int index) const {
    return *geta("layer")[index].as<RecordValue>().record();
}
} // namespace defn
namespace defn {
void Model::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addText  ("state", "");
    def().addNumber("off", 0);
    def().addText  ("sprite", "");
    def().addNumber("spriteFrame", 0);
    def().addNumber("group", 0);
    def().addNumber("selector", 0);
    def().addNumber("flags", 0);
    def().addNumber("interMark", 0);
    def().addArray ("interRange", new ArrayValue(Vector2i(0, 1)));
    def().addNumber("skinTics", 0);
    def().addArray ("scale", new ArrayValue(Vector3i(1, 1, 1)));
    def().addNumber("resize", 0);
    def().addArray ("offset", new ArrayValue(Vector3f()));
    def().addNumber("shadowRadius", 0);
    def().addArray ("sub", new ArrayValue);
}
Record &Model::addSub() {
    Record *sub = new Record;
    sub->addBoolean("custom", false);
    sub->addText  ("filename", "");
    sub->addText  ("skinFilename", "");
    sub->addText  ("frame", "");
    sub->addNumber("frameRange", 0);
    sub->addNumber("flags", 0);
    sub->addNumber("skin", 0);
    sub->addNumber("skinRange", 0);
    sub->addArray ("offset", new ArrayValue(Vector3f()));
    sub->addNumber("alpha", 0);
    sub->addNumber("parm", 0);
    sub->addNumber("selSkinMask", 0);
    sub->addNumber("selSkinShift", 0);
    ArrayValue *skins = new ArrayValue;
    for (int i = 0; i < 8; ++i) *skins << NumberValue(0);
    sub->addArray ("selSkins", skins);
    sub->addText  ("shinySkin", "");
    sub->addNumber("shiny", 0);
    sub->addArray ("shinyColor", new ArrayValue(Vector3f(1, 1, 1)));
    sub->addNumber("shinyReact", 1);
    sub->addNumber("blendMode", BM_NORMAL);
    def()["sub"].array().add(new RecordValue(sub, RecordValue::OwnsRecord));
    return *sub;
}
int Model::subCount() const {
    return int(geta("sub").size());
}
bool Model::hasSub(int index) const {
    return index >= 0 && index < subCount();
}
Record &Model::sub(int index) {
    return *def().geta("sub")[index].as<RecordValue>().record();
}
Record const &Model::sub(int index) const {
    return *geta("sub")[index].as<RecordValue>().record();
}
void Model::cleanupAfterParsing(Record const &prev) {
    if (gets("state") == "-") {
        def().set("state", prev.gets("state"));
    }
    if (gets("sprite") == "-") {
        def().set("sprite", prev.gets("sprite"));
    }
    for (int i = 0; i < subCount(); ++i) {
        Record &subDef = sub(i);
        if (subDef.gets("filename") == "-")     subDef.set("filename", "");
        if (subDef.gets("skinFilename") == "-") subDef.set("skinFilename", "");
        if (subDef.gets("shinySkin") == "-")    subDef.set("shinySkin", "");
        if (subDef.gets("frame") == "-")        subDef.set("frame", "");
    }
}
} // namespace defn
namespace defn {
void Music::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addText  ("lumpName", "");
    def().addText  ("path", "");
    def().addNumber("cdTrack", 0);
}
dint Music::cdTrack() const {
    if (dint track = geti("cdTrack"))
        return track;
    String const path = gets("path");
    if (!path.compareWithoutCase("cd")) {
        bool ok;
        dint track = path.toInt(&ok);
        if (ok) return track;
    }
    return 0;
}
}  // namespace defn
namespace defn {
void Sky::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addNumber("flags", 0);
    def().addNumber("height", DEFAULT_SKY_HEIGHT);
    def().addNumber("horizonOffset", DEFAULT_SKY_HORIZON_OFFSET);
    def().addArray ("color", new ArrayValue(Vector3f()));
    def().addArray ("layer", new ArrayValue);
    def().addArray ("model", new ArrayValue);
    addLayer();
    addLayer();
}
Record &Sky::addLayer() {
    Record *layer = new Record;
    layer->addBoolean("custom", false);
    layer->addNumber("flags", 0);
    layer->addText  ("material", "");
    layer->addNumber("offset", DEFAULT_SKY_SPHERE_XOFFSET);
    layer->addNumber("offsetSpeed", 0);
    layer->addNumber("colorLimit", DEFAULT_SKY_SPHERE_FADEOUT_LIMIT);
    def()["layer"].array().add(new RecordValue(layer, RecordValue::OwnsRecord));
    return *layer;
}
int Sky::layerCount() const {
    return int(geta("layer").size());
}
bool Sky::hasLayer(int index) const {
    return index >= 0 && index < layerCount();
}
Record &Sky::layer(int index) {
    return *def().geta("layer")[index].as<RecordValue>().record();
}
Record const &Sky::layer(int index) const {
    return *geta("layer")[index].as<RecordValue>().record();
}
Record &Sky::addModel() {
    Record *model = new Record;
    model->addBoolean("custom", false);
    model->addText  (VAR_ID, "");
    model->addNumber("layer", -1);
    model->addNumber("frameInterval", 1);
    model->addNumber("yaw", 0);
    model->addNumber("yawSpeed", 0);
    model->addArray ("originOffset", new ArrayValue(Vector3f()));
    model->addArray ("rotate", new ArrayValue(Vector2f()));
    model->addText  ("execute", "");
    model->addArray ("color", new ArrayValue(Vector4f(1, 1, 1, 1)));
    def()["model"].array().add(new RecordValue(model, RecordValue::OwnsRecord));
    return *model;
}
int Sky::modelCount() const {
    return int(geta("model").size());
}
bool Sky::hasModel(int index) const {
    return index >= 0 && index < modelCount();
}
Record &Sky::model(int index) {
    return *def().geta("model")[index].as<RecordValue>().record();
}
Record const &Sky::model(int index) const {
    return *geta("model")[index].as<RecordValue>().record();
}
} // namespace defn
namespace defn {
static QString const VAR_VIEWS     ("views");
static QString const VAR_FRONT_ONLY("frontOnly");
static QString const VAR_MATERIAL  ("material"); // UriValue
static QString const VAR_MIRROR_X  ("mirrorX");
CompiledSprite::CompiledSprite()
{}
CompiledSprite::CompiledSprite(Record const &spriteDef) {
    frontOnly = spriteDef.getb(VAR_FRONT_ONLY);
    auto const &viewsDict = spriteDef.getdt(VAR_VIEWS).elements();
    for (auto iter = viewsDict.begin(); iter != viewsDict.end(); ++iter) {
        ++viewCount;
        int angle = iter->first.value->asInt();
        if (views.size() <= angle) views.resize(angle + 1);
        Record const &viewDef = iter->second->as<RecordValue>().dereference();
        auto &view = views[angle];
        view.uri     = viewDef.get(VAR_MATERIAL).as<UriValue>().uri();
        view.mirrorX = viewDef.getb(VAR_MIRROR_X);
    }
}
CompiledSpriteRecord &Sprite::def() {
    return static_cast<CompiledSpriteRecord &>(Definition::def());
}
CompiledSpriteRecord const &Sprite::def() const {
    return static_cast<CompiledSpriteRecord const &>(Definition::def());
}
void Sprite::resetToDefaults() {
    Definition::resetToDefaults();
    def().resetCompiled();
    def().addBoolean(VAR_FRONT_ONLY, true);        ///< @c true= only use the front View.
    def().addDictionary(VAR_VIEWS);
}
DictionaryValue &Sprite::viewsDict() {
    return def()[VAR_VIEWS].value<DictionaryValue>();
}
/*DictionaryValue const &Sprite::viewsDict() const
{
    return def().getdt(VAR_VIEWS);
}*/
Record &Sprite::addView(String material, dint angle, bool mirrorX) {
    def().resetCompiled();
    if (angle <= 0) {
        def().addDictionary(VAR_VIEWS);
    }
    def().set(VAR_FRONT_ONLY, angle <= 0);
    auto *view = new Record;
    view->add(VAR_MATERIAL).set(new UriValue(de::makeUri(material)));
    view->addBoolean(VAR_MIRROR_X, mirrorX);
    viewsDict().add(new NumberValue(de::max(0, angle - 1)), new RecordValue(view, RecordValue::OwnsRecord));
    return *view;
}
dint Sprite::viewCount() const {
    return def().compiled().viewCount;
}
bool Sprite::hasView(dint angle) const {
    auto const &cmpl = def().compiled();
    if (cmpl.frontOnly) angle = 0;
    return (angle < cmpl.views.size() && !cmpl.views.at(angle).uri.isEmpty());
    /*for (Value const *val : geta("views").elements())
    {
        Record const &view = val->as<RecordValue>().dereference();
        if (view.geti("angle") == angle)
            return true;
    }
    return false;*/
}
#if 0
Record &Sprite::findView(dint angle) {
    if (angle && getb(VAR_FRONT_ONLY)) angle = 0;
    return viewsDict().element(NumberValue(angle)).as<RecordValue>().dereference();
    /*for (Value *val : def().geta("views").elements())
    {
        Record &view = val->as<RecordValue>().dereference();
        if (view.geti("angle") == angle)
            return view;
    }
    throw MissingViewError("Sprite::view", "Unknown view:" + String::number(angle));*/
}
Record const *Sprite::tryFindView(dint angle) const {
    if (angle && getb(VAR_FRONT_ONLY)) angle = 0;
    NumberValue const ang(angle);
    auto const &elems = viewsDict().elements();
    auto found = elems.find(&ang);
    if (found != elems.end()) {
        return found->second->as<RecordValue>().record();
    }
    return nullptr;
}
#endif
static de::Uri nullUri;
Sprite::View Sprite::view(de::dint angle) const {
    auto const &cmpl = def().compiled();
    if (cmpl.frontOnly) angle = 0;
    View v;
    if (angle < cmpl.views.size()) {
        v.material = &cmpl.views.at(angle).uri;
        v.mirrorX  = cmpl.views.at(angle).mirrorX;
    }
    else {
        v.material = &nullUri;
        v.mirrorX  = false;
    }
    return v;
}
de::Uri const &Sprite::viewMaterial(de::dint angle) const {
    auto const &cmpl = def().compiled();
    if (angle < cmpl.views.size()) {
        return cmpl.views.at(angle).uri;
    }
    return nullUri;
}
Sprite::View Sprite::nearestView(angle_t mobjAngle, angle_t angleToEye, bool noRotation) const {
    dint angle = 0;  // Use the front view (default).
    if (!noRotation) {
        angle = ((angleToEye - mobjAngle + (unsigned) (ANG45 / 2) * 9) - (unsigned) (ANGLE_180 / 16)) >> 28;
    }
    return view(angle);
}
}  // namespace defn
namespace defn {
void State::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addText  ("sprite", "");
    def().addNumber("flags", 0);
    def().addNumber("frame", 0);
    def().addNumber("tics", 0);
    def().addText  ("action", "");
    def().addText  ("nextState", "");
    def().addText  ("execute", "");
    def().addArray ("misc").array().addMany(NUM_STATE_MISC, 0);
}
dint32 State::misc(dint index) const {
    return def().geta("misc")[index].asInt();
}
void State::setMisc(dint index, dint32 value) {
    def()["misc"].array().setElement(index, value);
}
} // namespace defn
namespace defn {
void Thing::resetToDefaults() {
    Definition::resetToDefaults();
    def().addText  (VAR_ID, "");
    def().addNumber("doomEdNum", 0);
    def().addText  ("name", "");
    def().addArray ("states").array().addMany(STATENAMES_COUNT, "");
    def().addArray ("sounds").array().addMany(SOUNDNAMES_COUNT, "");
    def().addNumber("reactionTime", 0);
    def().addNumber("painChance", 0);
    def().addNumber("spawnHealth", 0);
    def().addNumber("speed", 0);
    def().addNumber("radius", 0);
    def().addNumber("height", 0);
    def().addNumber("mass", 0);
    def().addNumber("damage", 0);
    def().addText  ("onTouch", ""); // script function to call when touching a special thing
    def().addText  ("onDeath", ""); // script function to call when thing is killed
    def().addArray ("flags").array().addMany(NUM_MOBJ_FLAGS, 0);
    def().addArray ("misc").array().addMany(NUM_MOBJ_MISC, 0);
}
void Thing::setSound(int soundId, String const &sound) {
    def()["sounds"].array().setElement(soundId, sound);
}
String Thing::sound(int soundId) const {
    return geta("sounds")[soundId].asText();
}
int Thing::flags(dint index) const {
    return geta("flags")[index].asInt();
}
void Thing::setFlags(dint index, int flags) {
    def()["flags"].array()
        .setElement(NumberValue(index), new NumberValue(flags, NumberValue::Hex));
}
} // namespace defn
