﻿/** @file datafile.cpp  Classic data files: WAD, LMP, DED, DEH.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/filesys/datafile.h"
#include <de/Reader>
/** @file datafolder.cpp  Classic data files: PK3.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/filesys/datafolder.h"
#include <de/ArchiveFeed>
#include <de/ZipArchive>
/** @file file.cpp
 *
 * Abstract base for all classes which represent loaded files.
 * @ingroup fs
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/file.h"
#include "doomsday/filesys/fs_main.h"
#include <de/NativePath>
/** @file filehandle.cpp
 *
 * Reference/handle to a unique file in the engine's virtual file system.
 *
 * @ingroup fs
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/file.h"
#include <cctype>
#include <ctime>
#include <sys/stat.h>
#include <de/memory.h>
#include <de/memoryblockset.h>
#include <de/LogBuffer>
#include <de/NativePath>
/** @file fileid.cpp
 *
 * Implements a file identifier in terms of a MD5 hash of its absolute path.
 *
 * @ingroup types
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QDir>
#include <QCryptographicHash>
#include <de/App>
#include <de/Log>
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fileid.h"
/** @file
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/filesys/filetype.h"
/** @file fs_main.cpp
 * @ingroup fs
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2006-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright &copy; 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright &copy; 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/console/exec.h"
#include "doomsday/console/cmd.h"
#include "doomsday/filesys/file.h"
#include "doomsday/filesys/fileid.h"
#include "doomsday/filesys/fileinfo.h"
#include "doomsday/filesys/lumpindex.h"
#include "doomsday/filesys/wad.h"
#include "doomsday/filesys/zip.h"
#include <ctime>
#include <QDir>
#include <QList>
#include <QtAlgorithms>
#include <de/App>
#include <de/Log>
#include <de/NativePath>
#include <de/LogBuffer>
#include <de/memory.h>
#include <de/findfile.h>
#include <de/FileSystem>
/** @file fs_scheme.cpp File System Subspace Scheme.
 * @ingroup fs
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QDir>
#include <cstring>
#include <de/App>
#include <de/Log>
#include <de/NativePath>
#include <de/PathTree>
#include <de/findfile.h>
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/searchpath.h"
/** @file fs_util.cpp
 *
 * Miscellaneous file system utility routines.
 *
 * @ingroup fs
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#define DENG_NO_API_MACROS_FILESYS
#ifdef WIN32
#  include <direct.h>
#  include <io.h>
#endif
#ifdef UNIX
#  include <pwd.h>
#  include <limits.h>
#  include <unistd.h>
#  include <errno.h>
#endif
#include <cctype>
#include <sys/types.h>
#include <sys/stat.h>
#include "doomsday/DoomsdayApp"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/filesys/lumpindex.h"
#include <de/App>
#include <de/FileSystem>
#include <de/Folder>
#include <de/Log>
#include <de/NativeFile>
#include <de/findfile.h>
/** @file idgameslink.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/filesys/idgameslink.h"
#include "doomsday/filesys/idgamespackageinfofile.h"
#include "doomsday/DataBundle"
#include <de/Async>
#include <de/FileSystem>
#include <de/RemoteFile>
#include <de/data/gzip.h>
#include <QRegularExpression>
#include <QUrl>
/** @file idgamespackageinfofile.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/idgamespackageinfofile.h"
#include "doomsday/res/Bundles"
#include "doomsday/DoomsdayApp"
#include <de/ArrayValue>
#include <de/FileSystem>
#include <de/Folder>
#include <de/Package>
#include <de/TextValue>
#include <QRegularExpression>
/** @file lumpcache.cpp  Provides a data cache tailored to storing lumps (i.e., files).
 *
 * @author Copyright © 2013-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/lumpcache.h"
#include <de/memory.h>
#include <de/memoryzone.h>
#include <de/Error>
#include <de/Log>
/** @file lumpindex.cpp  Index of lumps.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 1999-2006 by Colin Phipps, Florian Schulze, Neil Stevens, Andrey Budko (PrBoom 2.2.6)
 * @authors Copyright © 1999-2001 by Jess Haas, Nicolas Kalkhof (PrBoom 2.2.6)
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/lumpindex.h"
#include <QBitArray>
#include <QHash>
#include <QVector>
#include <de/LogBuffer>
/** @file readfile.cpp  Legacy file reading utility routines.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/filesys/readfile.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#if defined(WIN32)
#  include <direct.h>
#  include <io.h>
#  include <conio.h>
#endif
#if defined(UNIX)
#  include <unistd.h>
#  include <string.h>
#endif
#ifndef O_BINARY
#  define O_BINARY 0
#endif
#include "doomsday/filesys/fs_main.h"
#include "lzss.h"
#include <de/App>
#include <de/str.h>
#include <de/memoryzone.h>
#include <cstdlib>
#include <cctype>
#if defined(WIN32)
#  define open _open
#endif
/** @file searchpath.cpp SearchPath
 *
 * @authors Copyright &copy; 2010-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright &copy; 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/searchpath.h"
/** @file virtualmappings.cpp  Maps WAD lumps and native files to virtual FS1 files.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/filesys/lumpindex.h"
#include "doomsday/filesys/virtualmappings.h"
#include "doomsday/doomsdayapp.h"
#include <de/memory.h>
#include <de/c_wrapper.h>
#ifdef WIN32
#  define strupr _strupr
#endif
/** @file wad.cpp  WAD Archive (file).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/wad.h"
#include "doomsday/DoomsdayApp"
#include "doomsday/filesys/lumpcache.h"
#include <de/ByteOrder>
#include <de/NativePath>
#include <de/LogBuffer>
#include <de/memoryzone.h>
#include <cstring> // memcpy
/** @file zip.cpp  ZIP Archive (file).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/filesys/zip.h"
#include "doomsday/filesys/lumpcache.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/game.h"
#include "doomsday/doomsdayapp.h"
#include <zlib.h>
#include <vector>
#include <de/App>
#include <de/ByteOrder>
#include <de/NativePath>
#include <de/LogBuffer>
#include <de/memory.h>
#include <de/memoryzone.h>
#include <cstring> // memcpy
#ifdef MSVC
#  undef min
#  undef max
#endif
using namespace de;
DataFile::DataFile(Format format, File &sourceFile)
    : ByteArrayFile(sourceFile.name())
    , DataBundle(format, sourceFile) {
    setSource(&sourceFile);
    if (format == Iwad || format == Pwad) {
        Block fileType;
        fileType.resize(4);
        get(0, fileType.data(), fileType.size());
        setFormat(fileType == "IWAD"? Iwad : Pwad);
    }
}
DataFile::~DataFile() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
String DataFile::describe() const {
    return DataBundle::description();
}
void DataFile::get(Offset at, Byte *values, Size count) const {
    DataBundle::get(at, values, count);
}
void DataFile::set(Offset at, Byte const *values, Size count) {
    DataBundle::set(at, values, count);
}
DataFolder::DataFolder(Format format, File &sourceFile)
    : Folder(sourceFile.name())
    , DataBundle(format, sourceFile) {
    setSource(&sourceFile);
    if (ZipArchive::recognize(sourceFile)) {
        attach(new ArchiveFeed(sourceFile));
    }
    /*else if (sourceFile.originFeed())
    {
        attach(sourceFile.originFeed()->newSubFeed(sourceFile.name()));
    }
    else {
        DENG2_ASSERT(!"DataFolder doesn't know how to access the source file");
    }*/
}
DataFolder::~DataFolder() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
String DataFolder::describe() const {
    String desc = DataBundle::description();
    String const feedDesc = describeFeeds();
    if (!feedDesc.isEmpty()) {
        desc += String(" (%1)").arg(feedDesc);
    }
    return desc;
}
IIStream const &DataFolder::operator >> (IByteArray &bytes) const {
    if (source()) {
        *source() >> bytes;
    }
    return *this;
}
namespace de {
File1::File1(FileHandle *hndl, String _path, FileInfo const &_info, File1 *_container)
    : handle_(hndl)
    , info_(_info)
    , container_(_container)
    , flags(DefaultFlags)
    , path_(_path)
    , name_(_path.fileName()) {
    static uint fileCounter = 0;
    order = fileCounter++;
}
File1::~File1() {
    App_FileSystem().releaseFile(*this);
    if (handle_) delete handle_;
}
FileInfo const &File1::info() const {
    return info_;
}
bool File1::isContained() const {
    return !!container_;
}
File1 &File1::container() const {
    if (!container_) throw NotContainedError("File1::container", "File \"" + NativePath(composePath()).pretty() + " is not contained");
    return *container_;
}
FileHandle &File1::handle() {
    return *handle_;
}
Uri File1::composeUri(QChar delimiter) const {
    return Uri(path_, RC_NULL, delimiter);
}
uint File1::loadOrderIndex() const {
    return order;
}
bool File1::hasStartup() const {
    return flags.testFlag(Startup);
}
File1 &File1::setStartup(bool yes) {
    if (yes) flags |= Startup;
    else    flags &= ~Startup;
    return *this;
}
bool File1::hasCustom() const {
    return flags.testFlag(Custom);
}
File1 &File1::setCustom(bool yes) {
    if (yes) flags |= Custom;
    else    flags &= ~Custom;
    return *this;
}
String const &File1::name() const {
    return name_;
}
size_t File1::read(uint8_t* /*buffer*/, bool /*tryCache*/)
{
    throw Error("File1::read", "Not yet implemented");
}
size_t File1::read(uint8_t* /*buffer*/, size_t /*startOffset*/, size_t /*length*/,
                   bool /*tryCache*/)
{
    throw Error("File1::read", "Not yet implemented");
}
uint8_t const *File1::cache() {
    throw Error("File1::cache", "Not yet implemented");
}
File1& File1::unlock() {
    throw Error("File1::unlock", "Not yet implemented");
}
File1 &File1::clearCache(bool* /*retCleared*/)
{
    throw Error("File1::clearCache", "Not yet implemented");
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(FileHandle) {
    File1 *file;
    void *list;
    struct dfile_flags_s {
        uint open:1;       ///< Presently open.
        uint eof:1;        ///< Reader has reached the end of the stream.
        uint reference:1;  ///< This handle is a reference to another dfile instance.
    } flags;
    size_t baseOffset;
    FILE *hndl;
    size_t size;
    uint8_t *data;
    uint8_t *pos;
    Impl() : file(0), list(0), baseOffset(0), hndl(0), size(0), data(0), pos(0) {
        flags.eof  = false;
        flags.open = false;
        flags.reference = false;
    }
};
static void errorIfNotValid(FileHandle const &file, char const * /*callerName*/)
{
    DENG2_ASSERT(file.isValid());
    if (!file.isValid()) exit(1);
}
FileHandle::FileHandle()
    : d(new Impl)
{}
FileHandle::~FileHandle() {
    close();
    if (d->data) {
        M_Free(d->data); d->data = 0;
    }
}
FileHandle &FileHandle::close() {
    if (!d->flags.open) return *this;
    if (d->hndl) {
        fclose(d->hndl); d->hndl = 0;
    }
    if (d->data) {
        M_Free(d->data); d->data = 0;
    }
    d->pos        = 0;
    d->flags.open = false;
    return *this;
}
bool FileHandle::isValid() const {
    return true;
}
FileList *FileHandle::list() {
    errorIfNotValid(*this, "FileHandle::list");
    return (FileList *)d->list;
}
FileHandle &FileHandle::setList(FileList *list) {
    d->list = list;
    return *this;
}
bool FileHandle::hasFile() const {
    errorIfNotValid(*this, "FileHandle::file");
    return !!d->file;
}
File1 &FileHandle::file() {
    errorIfNotValid(*this, "FileHandle::file");
    return *d->file;
}
File1 &FileHandle::file() const {
    errorIfNotValid(*this, "FileHandle::file const");
    return *d->file;
}
size_t FileHandle::baseOffset() const {
    if (d->flags.reference) {
        return d->file->handle().baseOffset();
    }
    return d->baseOffset;
}
size_t FileHandle::length() {
    errorIfNotValid(*this, "FileHandle::Length");
    if (d->flags.reference) {
        return d->file->handle().length();
    }
    else {
        size_t const currentPosition = seek(0, SeekEnd);
        size_t const length = tell();
        seek(currentPosition, SeekSet);
        return length;
    }
}
size_t FileHandle::read(uint8_t *buffer, size_t count) {
    errorIfNotValid(*this, "FileHandle::read");
    if (d->flags.reference) {
        return d->file->handle().read(buffer, count);
    }
    else {
        if (d->hndl) {
            count = fread(buffer, 1, count, d->hndl);
            if (feof(d->hndl)) {
                d->flags.eof = true;
            }
            return count;
        }
        size_t bytesleft = d->size - (d->pos - d->data);
        if (count > bytesleft) {
            count = bytesleft;
            d->flags.eof = true;
        }
        if (count) {
            memcpy(buffer, d->pos, count);
            d->pos += count;
        }
        return count;
    }
}
bool FileHandle::atEnd() {
    errorIfNotValid(*this, "FileHandle::atEnd");
    if (d->flags.reference) {
        return d->file->handle().atEnd();
    }
    return (d->flags.eof != 0);
}
unsigned char FileHandle::getC() {
    errorIfNotValid(*this, "FileHandle::getC");
    unsigned char ch = 0;
    read((uint8_t *)&ch, 1);
    return ch;
}
size_t FileHandle::tell() {
    errorIfNotValid(*this, "FileHandle::tell");
    if (d->flags.reference) {
        return d->file->handle().tell();
    }
    else {
        if (d->hndl) {
            return (size_t) ftell(d->hndl);
        }
        return d->pos - d->data;
    }
}
size_t FileHandle::seek(size_t offset, SeekMethod whence) {
    if (d->flags.reference) {
        return d->file->handle().seek(offset, whence);
    }
    else {
        size_t oldpos = tell();
        d->flags.eof = false;
        if (d->hndl) {
            int fwhence = whence == SeekSet? SEEK_SET :
                          whence == SeekCur? SEEK_CUR : SEEK_END;
            fseek(d->hndl, (long) (d->baseOffset + offset), fwhence);
        }
        else {
            if (whence == SeekSet)
                d->pos = d->data + offset;
            else if (whence == SeekEnd)
                d->pos = d->data + (d->size + offset);
            else if (whence == SeekCur)
                d->pos += offset;
        }
        return oldpos;
    }
}
FileHandle &FileHandle::rewind() {
    seek(0, SeekSet);
    return *this;
}
FileHandle *FileHandle::fromFile(File1 &file) // static
{
    FileHandle *hndl = new FileHandle();
    hndl->d->file            = &file;
    hndl->d->flags.open      = true;
    hndl->d->flags.reference = true;
    return hndl;
}
FileHandle *FileHandle::fromNativeFile(FILE &file, size_t baseOffset) // static
{
    FileHandle *hndl = new FileHandle();
    hndl->d->flags.open = true;
    hndl->d->hndl       = &file;
    hndl->d->baseOffset = baseOffset;
    return hndl;
}
FileHandle *FileHandle::fromLump(File1 &lump, bool dontBuffer) // static
{
    LOG_AS("FileHandle::fromLump");
    FileHandle *hndl = new FileHandle();
    hndl->d->file       = &lump;
    hndl->d->flags.open = true;
    if (!dontBuffer) {
        hndl->d->size = lump.size();
        hndl->d->pos  = hndl->d->data = (uint8_t *) M_Malloc(hndl->d->size);
        LOGDEV_RES_XVERBOSE_DEBUGONLY("[%p] Buffering \"%s:%s\"...", dintptr(hndl)
                                     << NativePath(lump.container().composePath()).pretty()
                                     << NativePath(lump.composePath()).pretty());
        lump.read((uint8_t *)hndl->d->data, 0, lump.size());
    }
    return hndl;
}
} // namespace de
FileId::FileId(Md5Hash _md5)
    : md5_(_md5.left(16))
#ifdef DENG_DEBUG
    , path_("unknown-path")
#endif
{}
FileId::FileId(FileId const &other)
    : LogEntry::Arg::Base()
    , md5_(other.md5())
#ifdef DENG_DEBUG
    , path_(other.path())
#endif
{}
FileId &FileId::operator = (FileId other) {
    swap(*this, other);
    return *this;
}
bool FileId::operator < (FileId const &other) const {
    return md5_ < other.md5_;
}
bool FileId::operator == (FileId const &other) const {
    return md5_ == other.md5_;
}
bool FileId::operator != (FileId const &other) const {
    return md5_ != other.md5_;
}
String FileId::asText() const {
    String txt;
    txt.reserve(32);
    for (int i = 0; i < 16; ++i) {
        txt += String("%1").arg(String::number((byte)md5_.at(i), 16), 2, '0');
    }
    return txt;
}
FileId FileId::fromPath(String path) {
    FileId fileId = FileId(hash(path));
#ifdef DENG_DEBUG
    fileId.setPath(path);
#endif
    return fileId;
}
FileId::Md5Hash FileId::hash(String path) {
    if (QDir::isRelativePath(path)) {
        path = App_BasePath() / path;
    }
#if defined(WIN32) || defined(MACOSX)
    path = path.toUpper();
#endif
    return QCryptographicHash::hash(path.toUtf8(), QCryptographicHash::Md5);
}
static NullFileType nullFileType;
static FileTypes fileTypeMap;
void DD_AddFileType(FileType const &ftype) {
    fileTypeMap.insert(ftype.name().toLower(), &ftype);
}
FileType const &DD_FileTypeByName(String name) {
    if (!name.isEmpty()) {
        FileTypes::const_iterator found = fileTypeMap.constFind(name.toLower());
        if (found != fileTypeMap.constEnd()) return **found;
    }
    return nullFileType; // Not found.
}
FileType const &DD_GuessFileTypeFromFileName(String path) {
    if (!path.isEmpty()) {
        DENG2_FOR_EACH_CONST(FileTypes, i, fileTypeMap) {
            FileType const &ftype = **i;
            if (ftype.fileNameIsKnown(path))
                return ftype;
        }
    }
    return nullFileType;
}
FileTypes &DD_FileTypes() {
    return fileTypeMap;
}
extern uint F_GetLastModified(char const *path);
static FS1 *fileSystem;
typedef QList<FileId> FileIds;
/**
 * Virtual (file) path => Lump name mapping.
 *
 * @todo We can't presently use a Map or Hash for these. Although the paths are
 *       unique, several of the existing algorithms which match using patterns
 *       assume they are sorted in a quasi load ordering.
 */
typedef QPair<QString, QString> LumpMapping;
typedef QList<LumpMapping> LumpMappings;
/**
 * Virtual file-directory mapping.
 * Maps one (absolute) path in the virtual file system to another.
 *
 * @todo We can't presently use a Map or Hash for these. Although the paths are
 *       unique, several of the existing algorithms which match using patterns
 *       assume they are sorted in a quasi load ordering.
 */
typedef QPair<QString, QString> PathMapping;
typedef QList<PathMapping> PathMappings;
/**
 * @note Performance is O(n).
 * @return @c iterator pointing to list->end() if not found.
 */
static FS1::FileList::iterator findListFile(FS1::FileList &list, File1 &file) {
    if (list.empty()) return list.end();
    FS1::FileList::iterator i;
    for (i = list.begin(); i != list.end(); ++i) {
        if (&file == &(*i)->file()) {
            break; // This is the node we are looking for.
        }
    }
    return i;
}
/**
 * @note Performance is O(n).
 * @return @c iterator pointing to list->end() if not found.
 */
static FS1::FileList::iterator findListFileByPath(FS1::FileList &list, String path) {
    if (list.empty()) return list.end();
    if (path.isEmpty()) return list.end();
    FS1::FileList::iterator i;
    for (i = list.begin(); i != list.end(); ++i) {
        File1 &file = (*i)->file();
        if (!file.composePath().compare(path, Qt::CaseInsensitive)) {
            break; // This is the node we are looking for.
        }
    }
    return i;
}
static bool applyPathMapping(ddstring_t *path, PathMapping const &vdm);
/**
 * Performs a case-insensitive pattern match. The pattern can contain
 * wildcards.
 *
 * @param filePath  Path to match.
 * @param pattern   Pattern with * and ? as wildcards.
 *
 * @return  @c true, if @a filePath matches the pattern.
 */
static bool matchFileName(String const &string, String const &pattern) {
    static QChar const ASTERISK('*');
    static QChar const QUESTION_MARK('?');
    QChar const *in = string.constData(), *st = pattern.constData();
    while (!in->isNull()) {
        if (*st == ASTERISK) {
            st++;
            continue;
        }
        if (*st != QUESTION_MARK && st->toLower() != in->toLower()) {
            while (st >= pattern && *st != ASTERISK) { st--; }
            if (st < pattern)
                return false; // No match!
        }
        st++;
        in++;
    }
    while (*st == ASTERISK) { st++; }
    return st->isNull();
}
DENG2_PIMPL(FS1) {
    bool loadingForStartup;     ///< @c true= Flag newly opened files as "startup".
    FileList openFiles;         ///< List of currently opened files.
    FileList loadedFiles;       ///< List of all loaded files present in the system.
    uint loadedFilesCRC;
    FileIds fileIds;            ///< Database of unique identifiers for all loaded/opened files.
    LumpIndex primaryIndex;     ///< Primary index of all files in the system.
    LumpIndex zipFileIndex;     ///< Type-specific index for ZipFiles.
    LumpMappings lumpMappings;  ///< Virtual (file) path => Lump name mapping.
    PathMappings pathMappings;  ///< Virtual file-directory mapping.
    Schemes schemes;            ///< File subsets.
    Impl(Public *i)
        : Base(i)
        , loadingForStartup(true)
        , loadedFilesCRC   (0)
        , zipFileIndex     (true/*paths are unique*/)
    {}
    ~Impl() {
        clearLoadedFiles();
        clearOpenFiles();
        clearIndexes();
        fileIds.clear(); // Should be NOP, if bookkeeping is correct.
        pathMappings.clear();
        lumpMappings.clear();
        clearAllSchemes();
    }
    void clearAllSchemes() {
        DENG2_FOR_EACH(Schemes, i, schemes) {
            delete *i;
        }
        schemes.clear();
    }
    bool releaseFileId(String path) {
        if (!path.isEmpty()) {
            FileId fileId = FileId::fromPath(path);
            FileIds::iterator place = qLowerBound(fileIds.begin(), fileIds.end(), fileId);
            if (place != fileIds.end() && *place == fileId) {
                LOGDEV_RES_XVERBOSE_DEBUGONLY("Released FileId %s - \"%s\"", *place << fileId.path());
                fileIds.erase(place);
                return true;
            }
        }
        return false;
    }
    void clearLoadedFiles(LumpIndex *index = 0) {
        loadedFilesCRC = 0;
        for (int i = loadedFiles.size() - 1; i >= 0; i--) {
            File1 &file = loadedFiles[i]->file();
            if (!index || index->catalogues(file)) {
                self().deindex(file);
                delete &file;
            }
        }
    }
    void clearOpenFiles() {
        while (!openFiles.isEmpty()){ delete openFiles.takeLast(); }
    }
    void clearIndexes() {
        primaryIndex.clear();
        zipFileIndex.clear();
    }
    String findPath(de::Uri const &search) {
        try {
            Scheme &scheme = self().scheme(search.scheme());
            LOG_RES_XVERBOSE("Using scheme '%s'...", scheme.name());
            scheme.rebuild();
            String name = search.path().lastSegment().toString().fileNameWithoutExtension();
            Scheme::FoundNodes foundNodes;
            if (scheme.findAll(name, foundNodes)) {
                DENG2_FOR_EACH_CONST(Scheme::FoundNodes, i, foundNodes) {
                    PathTree::Node &node = **i;
                    if (!node.comparePath(search.path(), PathTree::NoBranch)) {
                        return node.path();
                    }
                }
            }
        }
        catch (UnknownSchemeError const &)
        {} // Ignore this error.
        QScopedPointer<File1> file(openFile(search.path(), "rb", 0, true /* allow duplicates */));
        if (!file.isNull()) {
            return file->composePath();
        }
        return ""; // Not found.
    }
    File1 *findLump(String path, String const & /*mode*/)
    {
        if (path.isEmpty()) return 0;
        if (QDir::isRelativePath(path)) {
            path = App_BasePath() / path;
        }
        lumpnum_t lumpNum = zipFileIndex.findLast(path);
        if (lumpNum >= 0) {
            return &zipFileIndex[lumpNum];
        }
        if (!lumpMappings.empty()) {
            DENG2_FOR_EACH_CONST(LumpMappings, i, lumpMappings) {
                LumpMapping const &mapping = *i;
                if (mapping.first.compare(path)) continue;
                lumpnum_t lumpNum = self().lumpNumForName(mapping.second);
                if (lumpNum < 0) continue;
                return &self().lump(lumpNum);
            }
        }
        return 0;
    }
    FILE *findAndOpenNativeFile(String path, String const &mymode, String &foundPath) {
        DENG2_ASSERT(!path.isEmpty());
        path = NativePath::workPath().withSeparators('/') / path;
        char mode[8] = "";
        if (mymode.contains('r'))      strcat(mode, "r");
        else if (mymode.contains('w')) strcat(mode, "w");
        if (mymode.contains('b'))      strcat(mode, "b");
        else if (mymode.contains('t')) strcat(mode, "t");
        NativePath nativePath = NativePath(path);
        FILE *nativeFile = fopen(nativePath.toUtf8().constData(), mode);
        if (nativeFile) {
            foundPath = nativePath.expand().withSeparators('/');
            return nativeFile;
        }
        if (!pathMappings.empty()) {
            QByteArray pathUtf8 = path.toUtf8();
            AutoStr *mapped = AutoStr_NewStd();
            DENG2_FOR_EACH_CONST(PathMappings, i, pathMappings) {
                Str_Set(mapped, pathUtf8.constData());
                if (!applyPathMapping(mapped, *i)) continue;
                nativePath = NativePath(Str_Text(mapped));
                nativeFile = fopen(nativePath.toUtf8().constData(), mode);
                if (nativeFile) {
                    foundPath = nativePath.expand().withSeparators('/');
                    return nativeFile;
                }
            }
        }
        return 0;
    }
    File1 *openFile(String path, String const& mode, size_t baseOffset,
                    bool allowDuplicate) {
        if (path.isEmpty()) return 0;
        LOG_AS("FS1::openFile");
        path = App_BasePath() / path;
        LOG_RES_XVERBOSE("Trying \"%s\"...", NativePath(path).pretty());
        bool const reqNativeFile = mode.contains('f');
        FileHandle *hndl = 0;
        FileInfo info; // The temporary info descriptor.
        if (!reqNativeFile) {
            if (File1 *found = findLump(path, mode)) {
                if (!allowDuplicate && !self().checkFileId(found->composeUri())) return 0;
                hndl = FileHandle::fromLump(*found);
                info = found->info();
            }
        }
        if (!hndl) {
            String foundPath;
            if (FILE *found = findAndOpenNativeFile(path, mode, foundPath)) {
                if (!allowDuplicate && !self().checkFileId(de::makeUri(foundPath))) {
                    fclose(found);
                    return 0;
                }
                hndl = FileHandle::fromNativeFile(*found, baseOffset);
                info = FileInfo(F_GetLastModified(foundPath.toUtf8().constData()));
            }
        }
        if (!hndl) return 0;
        File1 &file = self().interpret(*hndl, path, info);
        if (loadingForStartup) {
            file.setStartup(true);
        }
        return &file;
    }
};
FS1::FS1() : d(new Impl(this))
{}
FS1::Scheme &FS1::createScheme(String name, Scheme::Flags flags) {
    DENG2_ASSERT(name.length() >= Scheme::min_name_length);
    if (knownScheme(name)) return scheme(name);
    Scheme *newScheme = new Scheme(name, flags);
    d->schemes.insert(name.toLower(), newScheme);
    return *newScheme;
}
void FS1::index(File1 &file) {
#ifdef DENG_DEBUG
    FileList::const_iterator found = findListFile(d->loadedFiles, file);
    if (found != d->loadedFiles.end())
        throw Error("FS1::index", "File \"" + NativePath(file.composePath()).pretty() + "\" has already been indexed");
#endif
    if (Zip *zip = maybeAs<Zip>(file)) {
        if (!zip->isEmpty()) {
            for (int i = 0; i < zip->lumpCount(); ++i) {
                File1 &lump = zip->lump(i);
                d->primaryIndex.catalogLump(lump);
                d->zipFileIndex.catalogLump(lump);
            }
        }
    }
    else if (Wad *wad = maybeAs<Wad>(file)) {
        if (!wad->isEmpty()) {
            for (int i = 0; i < wad->lumpCount(); ++i) {
                d->primaryIndex.catalogLump(wad->lump(i));
            }
        }
    }
    FileHandle *hndl = FileHandle::fromFile(file);
    d->loadedFiles.push_back(hndl); hndl->setList(reinterpret_cast<de::FileList *>(&d->loadedFiles));
    d->loadedFilesCRC = 0;
}
void FS1::deindex(File1 &file) {
    FileList::iterator found = findListFile(d->loadedFiles, file);
    if (found == d->loadedFiles.end()) return; // Most peculiar..
    FileHandle *fileHandle = *found;
    d->releaseFileId(file.composePath());
    d->zipFileIndex.pruneByFile(file);
    d->primaryIndex.pruneByFile(file);
    d->loadedFiles.erase(found);
    d->loadedFilesCRC = 0;
    delete fileHandle;
}
File1 &FS1::find(de::Uri const &search) {
    LOG_AS("FS1::find");
    if (!search.isEmpty()) {
        try {
            String searchPath = search.resolved();
            if (!QDir::isAbsolutePath(searchPath)) {
                searchPath = App_BasePath() / searchPath;
            }
            FileList::iterator found = findListFileByPath(d->loadedFiles, searchPath);
            if (found != d->loadedFiles.end()) {
                DENG_ASSERT((*found)->hasFile());
                return (*found)->file();
            }
        }
        catch (de::Uri::ResolveError const &er) {
            LOGDEV_RES_VERBOSE(er.asText());
        }
    }
    throw NotFoundError("FS1::find", "No files found matching '" + search.compose() + "'");
}
String FS1::findPath(de::Uri const &search, int flags, ResourceClass &rclass) {
    LOG_AS("FS1::findPath");
    if (!search.isEmpty()) {
        try {
            String searchPath = search.resolved();
            String ext = searchPath.fileNameExtension();
            if (!ext.isEmpty() && ext.compare(".*")) {
                String found = d->findPath(de::Uri(search.scheme(), searchPath));
                if (!found.isEmpty()) return found;
                if (flags & RLF_MATCH_EXTENSION) return "";
            }
            if (isNullResourceClass(rclass) || !rclass.fileTypeCount()) return "";
            /*
             * Try each expected file type name extension for this resource class.
             */
            String searchPathWithoutFileNameExtension = searchPath.fileNamePath() / searchPath.fileNameWithoutExtension();
            DENG2_FOR_EACH_CONST(ResourceClass::FileTypes, typeIt, rclass.fileTypes()) {
                DENG2_FOR_EACH_CONST(QStringList, i, (*typeIt)->knownFileNameExtensions()) {
                    String const &ext = *i;
                    String found = d->findPath(de::Uri(search.scheme(), searchPathWithoutFileNameExtension + ext));
                    if (!found.isEmpty()) return found;
                }
            };
        }
        catch (de::Uri::ResolveError const &er) {
            LOGDEV_RES_VERBOSE(er.asText());
        }
    }
    throw NotFoundError("FS1::findPath", "No paths found matching '" + search.compose() + "'");
}
String FS1::findPath(de::Uri const &search, int flags) {
    return findPath(search, flags, ResourceClass::classForId(RC_NULL));
}
#ifdef DENG_DEBUG
static void printFileIds(FileIds const &fileIds) {
    uint idx = 0;
    DENG2_FOR_EACH_CONST(FileIds, i, fileIds) {
        LOGDEV_RES_MSG("  %u - %s : \"%s\"") << idx << *i << i->path();
        ++idx;
    }
}
#endif
#ifdef DENG_DEBUG
static void printFileList(FS1::FileList &list) {
    uint idx = 0;
    DENG2_FOR_EACH_CONST(FS1::FileList, i, list) {
        FileHandle &hndl = **i;
        File1 &file   = hndl.file();
        FileId fileId = FileId::fromPath(file.composePath());
        LOGDEV_RES_VERBOSE(" %c%d: %s - \"%s\" (handle: %p)")
                << (file.hasStartup()? '*' : ' ') << idx
                << fileId << fileId.path() << &hndl;
        ++idx;
    }
}
#endif
int FS1::unloadAllNonStartupFiles() {
#ifdef DENG_DEBUG
    if (LogBuffer::get().isEnabled(LogEntry::Generic | LogEntry::Verbose)) {
        LOGDEV_RES_VERBOSE("Open files at reset:");
        printFileList(d->openFiles);
        LOGDEV_RES_VERBOSE("End\n");
    }
#endif
    int numUnloadedFiles = 0;
    for (int i = d->loadedFiles.size() - 1; i >= 0; i--) {
        File1 &file = d->loadedFiles[i]->file();
        if (file.hasStartup()) continue;
        deindex(file);
        delete &file;
        numUnloadedFiles += 1;
    }
#ifdef DENG_DEBUG
    if (!d->fileIds.empty()) {
        LOGDEV_RES_MSG("Orphan FileIds:");
        printFileIds(d->fileIds);
    }
#endif
    return numUnloadedFiles;
}
bool FS1::checkFileId(de::Uri const &path) {
    if (!accessFile(path)) return false;
    FileId fileId = FileId::fromPath(path.compose());
    FileIds::iterator place = qLowerBound(d->fileIds.begin(), d->fileIds.end(), fileId);
    if (place != d->fileIds.end() && *place == fileId) return false;
    LOGDEV_RES_XVERBOSE_DEBUGONLY("checkFileId \"%s\" => %s", fileId.path() << fileId); /* path() is debug-only */
    d->fileIds.insert(place, fileId);
    return true;
}
void FS1::resetFileIds() {
    d->fileIds.clear();
}
void FS1::endStartup() {
    d->loadingForStartup = false;
}
LumpIndex const &FS1::nameIndex() const {
    return d->primaryIndex;
}
lumpnum_t FS1::lumpNumForName(String name) {
    LOG_AS("FS1::lumpNumForName");
    if (name.isEmpty()) return -1;
    if (name.fileNameExtension().isEmpty()) {
        name += ".lmp";
    }
    return d->primaryIndex.findLast(Path(name));
}
void FS1::releaseFile(File1 &file) {
    for (int i = d->openFiles.size() - 1; i >= 0; i--) {
        FileHandle &hndl = *(d->openFiles[i]);
        if (&hndl.file() == &file) {
            d->openFiles.removeAt(i);
        }
    }
}
static Wad *findFirstWadFile(FS1::FileList &list, bool custom) {
    if (list.empty()) return 0;
    DENG2_FOR_EACH(FS1::FileList, i, list) {
        File1 &file = (*i)->file();
        if (custom != file.hasCustom()) continue;
        if (Wad *wad = maybeAs<Wad>(file)) {
            return wad;
        }
    }
    return 0;
}
uint FS1::loadedFilesCRC() {
    if (!d->loadedFilesCRC) {
        /**
         * We define the CRC as that of the lump directory of the first loaded IWAD.
         * @todo Really kludgy...
         */
        Wad *iwad = findFirstWadFile(d->loadedFiles, false/*not-custom*/);
        if (!iwad) return 0;
        d->loadedFilesCRC = iwad->calculateCRC();
    }
    return d->loadedFilesCRC;
}
FS1::FileList const &FS1::loadedFiles() const {
    return d->loadedFiles;
}
int FS1::findAll(bool (*predicate)(File1 &file, void *parameters), void *parameters,
    FS1::FileList &found) const {
    int numFound = 0;
    DENG2_FOR_EACH_CONST(FS1::FileList, i, d->loadedFiles) {
        if (predicate && !predicate((*i)->file(), parameters)) continue; // Nope.
        found.push_back(*i);
        numFound += 1;
    }
    return numFound;
}
int FS1::findAllPaths(Path searchPattern, int flags, FS1::PathList &found) {
    int const numFoundSoFar = found.count();
    if (!QDir::isAbsolutePath(searchPattern)) {
        searchPattern = App_BasePath() / searchPattern;
    }
    /*
     * Check the Zip directory.
     */
    DENG2_FOR_EACH_CONST(LumpIndex::Lumps, i, d->zipFileIndex.allLumps()) {
        File1 const &lump = **i;
        PathTree::Node const &node = lump.directoryNode();
        String filePath;
        bool patternMatched;
        if (!(flags & SearchPath::NoDescend)) {
            filePath = lump.composePath();
            patternMatched = matchFileName(filePath, searchPattern);
        }
        else {
            patternMatched = !node.comparePath(searchPattern, PathTree::MatchFull);
        }
        if (!patternMatched) continue;
        if (filePath.isEmpty()) {
            filePath = lump.composePath();
        }
        found.push_back(PathListItem(filePath, !node.isLeaf()? A_SUBDIR : 0));
    }
    /*
     * Check the dir/WAD direcs.
     */
    if (!d->lumpMappings.empty()) {
        DENG2_FOR_EACH_CONST(LumpMappings, i, d->lumpMappings) {
            if (!matchFileName(i->first, searchPattern)) continue;
            found.push_back(PathListItem(i->first, 0 /*only filepaths (i.e., leaves) can be mapped to lumps*/));
        }
    }
    /*
     * Check native paths.
     */
    String searchDirectory = searchPattern.toString().fileNamePath();
    if (!searchDirectory.isEmpty()) {
        QByteArray searchDirectoryUtf8 = searchDirectory.toUtf8();
        PathList nativeFilePaths;
        AutoStr *wildPath = AutoStr_NewStd();
        Str_Reserve(wildPath, searchDirectory.length() + 2 + 16); // Conservative estimate.
        for (int i = -1; i < (int)d->pathMappings.count(); ++i) {
            Str_Clear(wildPath);
            Str_Appendf(wildPath, "%s/", searchDirectoryUtf8.constData());
            if (i > -1) {
                if (!applyPathMapping(wildPath, d->pathMappings[i])) continue;
            }
            Str_AppendChar(wildPath, '*');
            FindData fd;
            if (!FindFile_FindFirst(&fd, Str_Text(wildPath))) {
                do {
                    if (Str_Compare(&fd.name, ".") && Str_Compare(&fd.name, "..")) {
                        String foundPath = searchDirectory / NativePath(Str_Text(&fd.name)).withSeparators('/');
                        if (!matchFileName(foundPath, searchPattern)) continue;
                        nativeFilePaths.push_back(PathListItem(foundPath, fd.attrib));
                    }
                } while (!FindFile_FindNext(&fd));
            }
            FindFile_Finish(&fd);
        }
        qSort(nativeFilePaths.begin(), nativeFilePaths.end());
        found.append(nativeFilePaths);
    }
    return found.count() - numFoundSoFar;
}
File1 &FS1::interpret(FileHandle &hndl, String filePath, FileInfo const &info) {
    DENG2_ASSERT(!filePath.isEmpty());
    File1 *interpretedFile = 0;
    FileType const &ftypeGuess = DD_GuessFileTypeFromFileName(filePath);
    if (NativeFileType const* fileType = dynamic_cast<NativeFileType const *>(&ftypeGuess)) {
        interpretedFile = fileType->interpret(hndl, filePath, info);
    }
    if (!interpretedFile) {
        FileTypes const &fileTypes = DD_FileTypes();
        DENG2_FOR_EACH_CONST(FileTypes, i, fileTypes) {
            if (NativeFileType const *fileType = dynamic_cast<NativeFileType const *>(*i)) {
                if (fileType == &ftypeGuess) continue;
                interpretedFile = fileType->interpret(hndl, filePath, info);
                if (interpretedFile) break;
            }
        }
    }
    if (!interpretedFile) {
        File1 *container = (hndl.hasFile() && hndl.file().isContained())? &hndl.file().container() : 0;
        interpretedFile = new File1(&hndl, filePath, info, container);
    }
    DENG2_ASSERT(interpretedFile);
    return *interpretedFile;
}
FileHandle &FS1::openFile(String const &path, String const &mode, size_t baseOffset, bool allowDuplicate) {
#ifdef DENG_DEBUG
    for (int i = 0; i < mode.length(); ++i) {
        if (mode[i] != 'r' && mode[i] != 't' && mode[i] != 'b' && mode[i] != 'f')
            throw Error("FS1::openFile", "Unknown argument in mode string '" + mode + "'");
    }
#endif
    File1 *file = d->openFile(path, mode, baseOffset, allowDuplicate);
    if (!file) throw NotFoundError("FS1::openFile", "No files found matching '" + path + "'");
    FileHandle &hndl = *FileHandle::fromFile(*file);
    d->openFiles.push_back(&hndl); hndl.setList(reinterpret_cast<de::FileList *>(&d->openFiles));
    return hndl;
}
FileHandle &FS1::openLump(File1 &lump) {
    FileHandle &hndl = *FileHandle::fromLump(lump);
    d->openFiles.push_back(&hndl); hndl.setList(reinterpret_cast<de::FileList *>(&d->openFiles));
    return hndl;
}
bool FS1::accessFile(de::Uri const &search) {
    try {
        QScopedPointer<File1> file(d->openFile(search.resolved(), "rb", 0, true /* allow duplicates */));
        return !file.isNull();
    }
    catch (de::Uri::ResolveError const &er) {
        LOGDEV_RES_VERBOSE(er.asText());
    }
    return false;
}
void FS1::addPathLumpMapping(String lumpName, String destination) {
    if (lumpName.isEmpty() || destination.isEmpty()) return;
    if (QDir::isRelativePath(destination)) {
        String workPath = DENG2_APP->currentWorkPath().withSeparators('/');
        destination = workPath / destination;
    }
    LumpMappings::iterator found = d->lumpMappings.begin();
    for (; found != d->lumpMappings.end(); ++found) {
        LumpMapping const &ldm = *found;
        if (!ldm.first.compare(destination, Qt::CaseInsensitive))
            break;
    }
    LumpMapping *ldm;
    if (found == d->lumpMappings.end()) {
        d->lumpMappings.push_back(LumpMapping(destination, lumpName));
        ldm = &d->lumpMappings.back();
    }
    else {
        ldm = &*found;
        ldm->second = lumpName;
    }
    LOG_RES_MSG("Path \"%s\" now mapped to lump \"%s\"") << NativePath(ldm->first).pretty() << ldm->second;
}
void FS1::clearPathLumpMappings() {
    d->lumpMappings.clear();
}
static bool applyPathMapping(ddstring_t *path, PathMapping const &pm) {
    if (!path) return false;
    QByteArray destUtf8 = pm.first.toUtf8();
    AutoStr *dest = AutoStr_FromTextStd(destUtf8.constData());
    if (qstrnicmp(Str_Text(path), Str_Text(dest), Str_Length(dest))) return false;
    QByteArray sourceUtf8 = pm.second.toUtf8();
    AutoStr *temp = AutoStr_FromTextStd(sourceUtf8.constData());
    Str_PartAppend(temp, Str_Text(path), pm.first.length(), Str_Length(path) - pm.first.length());
    Str_Copy(path, temp);
    return true;
}
void FS1::addPathMapping(String source, String destination) {
    if (source.isEmpty() || destination.isEmpty()) return;
    PathMappings::iterator found = d->pathMappings.begin();
    for (; found != d->pathMappings.end(); ++found) {
        PathMapping const &pm = *found;
        if (!pm.second.compare(source, Qt::CaseInsensitive))
            break;
    }
    PathMapping* pm;
    if (found == d->pathMappings.end()) {
        d->pathMappings.push_back(PathMapping(destination, source));
        pm = &d->pathMappings.back();
    }
    else {
        pm = &*found;
        pm->first = destination;
    }
    LOG_RES_MSG("Path \"%s\" now mapped to \"%s\"")
            << NativePath(pm->second).pretty() << NativePath(pm->first).pretty();
}
void FS1::clearPathMappings() {
    d->pathMappings.clear();
}
void FS1::printDirectory(Path path) {
    LOG_RES_MSG(_E(b) "Directory: %s") << NativePath(path).pretty();
    path = path / "*";
    PathList found;
    if (findAllPaths(path, 0, found)) {
        qSort(found.begin(), found.end());
        DENG2_FOR_EACH_CONST(PathList, i, found) {
            LOG_RES_MSG("  %s") << NativePath(i->path).pretty();
        }
    }
}
bool FS1::knownScheme(String name) {
    if (!name.isEmpty()) {
        Schemes::iterator found = d->schemes.find(name.toLower());
        if (found != d->schemes.end()) return true;
    }
    return false;
}
FS1::Scheme &FS1::scheme(String name) {
    if (!name.isEmpty()) {
        Schemes::iterator found = d->schemes.find(name.toLower());
        if (found != d->schemes.end()) return **found;
    }
    throw UnknownSchemeError("FS1::scheme", "No scheme found matching '" + name + "'");
}
FS1::Schemes const &FS1::allSchemes() {
    return d->schemes;
}
D_CMD(Dir) {
    DENG2_UNUSED(src);
    if (argc > 1) {
        for (int i = 1; i < argc; ++i) {
            String path = NativePath(argv[i]).expand().withSeparators('/');
            App_FileSystem().printDirectory(path);
        }
    }
    else {
        App_FileSystem().printDirectory(String("/"));
    }
    return true;
}
D_CMD(DumpLump) {
    DENG2_UNUSED2(src, argc);
    if (fileSystem) {
        lumpnum_t lumpNum = App_FileSystem().lumpNumForName(argv[1]);
        if (lumpNum >= 0) {
            return F_DumpFile(App_FileSystem().lump(lumpNum), 0);
        }
        LOG_RES_ERROR("No such lump");
        return false;
    }
    return false;
}
D_CMD(ListLumps) {
    DENG2_UNUSED3(src, argc, argv);
    if (!fileSystem) return false;
    LumpIndex const &lumpIndex = App_FileSystem().nameIndex();
    int const numRecords       = lumpIndex.size();
    int const numIndexDigits   = de::max(3, M_NumDigits(numRecords));
    LOG_RES_MSG("LumpIndex %p (%i records):") << &lumpIndex << numRecords;
    int idx = 0;
    DENG2_FOR_EACH_CONST(LumpIndex::Lumps, i, lumpIndex.allLumps()) {
        File1 const &lump = **i;
        String containerPath  = NativePath(lump.container().composePath()).pretty();
        String lumpPath       = NativePath(lump.composePath()).pretty();
        LOG_RES_MSG(String("%1 - \"%2:%3\" (size: %4 bytes%5)")
                        .arg(idx++, numIndexDigits, 10, QChar('0'))
                        .arg(containerPath)
                        .arg(lumpPath)
                        .arg(lump.info().size)
                        .arg(lump.info().isCompressed()? " compressed" : ""));
    }
    LOG_RES_MSG("---End of lumps---");
    return true;
}
D_CMD(ListFiles) {
    DENG2_UNUSED3(src, argc, argv);
    LOG_RES_MSG(_E(b) "Loaded Files " _E(l) "(in load order)" _E(w) ":");
    int totalFiles = 0;
    int totalPackages = 0;
    if (fileSystem) {
        FS1::FileList const &allLoadedFiles = App_FileSystem().loadedFiles();
        DENG2_FOR_EACH_CONST(FS1::FileList, i, allLoadedFiles) {
            File1 &file = (*i)->file();
            uint crc = 0;
            int fileCount = 1;
            if (de::Zip *zip = maybeAs<de::Zip>(file)) {
                fileCount = zip->lumpCount();
            }
            else if (de::Wad *wad = maybeAs<de::Wad>(file)) {
                fileCount = wad->lumpCount();
                crc = (!file.hasCustom()? wad->calculateCRC() : 0);
            }
            LOG_RES_MSG(" %s " _E(2)_E(>) "(%i %s%s)%s")
                    << NativePath(file.composePath()).pretty()
                    << fileCount << (fileCount != 1 ? "files" : "file")
                    << (file.hasStartup()? ", startup" : "")
                    << (crc? QString(" [%1]").arg(crc, 0, 16) : "");
            totalFiles += fileCount;
            ++totalPackages;
        }
    }
    LOG_RES_MSG(_E(b)"Total: " _E(.) "%i files in %i packages")
            << totalFiles << totalPackages;
    if (auto *svFiles = FS::get().tryLocate<Folder const>("/sys/server/public")) {
        LOG_RES_MSG("Server files:\n" _E(m) "%s") << svFiles->contentsAsText();
    }
    return true;
}
void FS1::consoleRegister() {
    C_CMD("dir", "",   Dir);
    C_CMD("ls",  "",   Dir); // Alias
    C_CMD("dir", "s*", Dir);
    C_CMD("ls",  "s*", Dir); // Alias
    C_CMD("dump",      "s", DumpLump);
    C_CMD("listfiles", "",  ListFiles);
    C_CMD("listlumps", "",  ListLumps);
}
FS1 &App_FileSystem() {
    if (!fileSystem) throw Error("App_FileSystem", "File system not yet initialized");
    return *fileSystem;
}
String App_BasePath() {
    return App::app().nativeBasePath().withSeparators('/');
}
void F_Init() {
    DENG2_ASSERT(!fileSystem);
    fileSystem = new FS1();
}
void F_Shutdown() {
    if (!fileSystem) return;
    delete fileSystem; fileSystem = 0;
}
void const *F_LumpIndex() {
    return &App_FileSystem().nameIndex();
}
namespace de {
/**
 * Reference to a file in the virtual file system.
 */
class FileRef {
public:
    FileRef(PathTree::Node &_directoryNode)
        : directoryNode_(&_directoryNode)
    {}
    PathTree::Node &directoryNode() const {
        return *directoryNode_;
    }
    FileRef &setDirectoryNode(PathTree::Node &newDirectoryNode) {
        directoryNode_ = &newDirectoryNode;
        return *this;
    }
#if _DEBUG
    String const &name() const {
        return name_;
    }
    FileRef &setName(String const &newName) {
        name_ = newName;
        return *this;
    }
#endif
private:
    PathTree::Node *directoryNode_;
#if _DEBUG
    String name_;
#endif
};
/**
 * Name search hash.
 */
struct NameHash {
public:
    typedef ushort hash_type;
    static hash_type const hash_range = 512;
    struct Node {
        Node *next;
        FileRef fileRef;
        Node(PathTree::Node &resourceNode)
            : next(0), fileRef(resourceNode)
        {}
    };
    struct Bucket {
        Node *first;
        Node *last;
    };
public:
    Bucket buckets[hash_range];
public:
    NameHash() {
        std::memset(buckets, 0, sizeof(buckets));
    }
    ~NameHash() {
        clear();
    }
    void clear() {
        for (hash_type hashKey = 0; hashKey < hash_range; ++hashKey) {
            while (buckets[hashKey].first) {
                NameHash::Node *nextNode = buckets[hashKey].first->next;
                delete buckets[hashKey].first;
                buckets[hashKey].first = nextNode;
            }
            buckets[hashKey].last = 0;
        }
    }
    Node *findDirectoryNode(hash_type hashKey, PathTree::Node const &directoryNode) {
        Node *node = buckets[hashKey].first;
        while (node && &node->fileRef.directoryNode() != &directoryNode) {
            node = node->next;
        }
        return node;
    }
    static hash_type hashName(String const &str) {
        hash_type hashKey = 0;
        int op = 0;
        for (int i = 0; i < str.length(); ++i) {
            ushort unicode = str.at(i).toLower().unicode();
            switch (op) {
            case 0: hashKey ^= unicode; ++op;   break;
            case 1: hashKey *= unicode; ++op;   break;
            case 2: hashKey -= unicode;   op=0; break;
            }
        }
        return hashKey % hash_range;
    }
};
struct FS1::Scheme::Impl {
    FS1::Scheme &self;
    String name;
    FS1::Scheme::Flags flags;
    UserDataPathTree directory;
    UserDataNode *rootNode;
    NameHash nameHash;
    byte nameHashIsDirty;
    FS1::Scheme::SearchPaths searchPaths;
    Impl(FS1::Scheme &d, String _name, FS1::Scheme::Flags _flags)
        : self(d), name(_name), flags(_flags), directory(), rootNode(0),
          nameHash(), nameHashIsDirty(true)
    {}
    /**
     * Add files to this scheme by resolving search path, searching the
     * file system and populating our internal directory with the results.
     * Duplicates are automatically pruned.
     *
     * @param searchPath  The path to resolve and search.
     */
    void addFromSearchPath(SearchPath const &searchPath) {
        try {
            addDirectoryPathAndMaybeDescendBranch(true/*do descend*/, searchPath.resolved(),
                                                  true/*is-directory*/, searchPath.flags());
        }
        catch (de::Uri::ResolveError const &er) {
            LOGDEV_RES_VERBOSE(er.asText());
        }
    }
    /**
     * Add files to this scheme by resolving each search path in @a group.
     *
     * @param group  Group of paths to search.
     */
    void addFromSearchPaths(FS1::PathGroup group) {
        for (FS1::Scheme::SearchPaths::const_iterator i = searchPaths.find(group);
             i != searchPaths.end() && i.key() == group; ++i) {
            addFromSearchPath(*i);
        }
    }
    UserDataNode *addDirectoryPath(String path) {
        if (path.isEmpty()) return 0;
        if (QDir::isAbsolutePath(path)) {
            String const basePath = App_BasePath();
            if (path.beginsWith(basePath)) {
                path = path.mid(basePath.length() + 1);
            }
        }
        if (path.isEmpty()) {
            if (!rootNode) {
                rootNode = &directory.insert(Path("./"));
            }
            return rootNode;
        }
        return &directory.insert(path);
    }
    void addDirectoryChildren(PathTree::Node &node, int flags) {
        if (node.isLeaf()) return;
        Path searchPattern = node.path() / "*";
        FS1::PathList found;
        App_FileSystem().findAllPaths(searchPattern, flags, found);
        DENG2_FOR_EACH_CONST(FS1::PathList, i, found) {
            addDirectoryPathAndMaybeDescendBranch(!(flags & SearchPath::NoDescend),
                                                  i->path, !!(i->attrib & A_SUBDIR),
                                                  flags);
        }
    }
    /**
     * @param descendBranch     @c true = descend this branch (if it is a branch).
     * @param filePath          Possibly-relative path to an element in the virtual file system.
     * param isFolder           @c true = @a filePath is a folder in the virtual file system.
     * @param flags             @ref searchPathFlags
     */
    void addDirectoryPathAndMaybeDescendBranch(bool descendBranch,
        String filePath, bool /*isFolder*/, int flags)
    {
        UserDataNode *node = addDirectoryPath(filePath);
        if (!node) return;
        if (!node->isLeaf()) {
            if (descendBranch) {
                if (node->userValue()) {
                    DENG2_FOR_EACH_CONST(PathTree::Nodes, i, directory.leafNodes()) {
                        PathTree::Node &sibling = **i;
                        if (&sibling.parent() == node) {
                            self.add(sibling);
                        }
                    }
                }
                else {
                    addDirectoryChildren(*node, flags);
                    node->setUserValue(true);
                }
            }
        }
        else {
            self.add(*node);
            node->setUserValue(true);
        }
    }
};
FS1::Scheme::Scheme(String symbolicName, Flags flags) {
    d = new Impl(*this, symbolicName, flags);
}
FS1::Scheme::~Scheme() {
    delete d;
}
String const &FS1::Scheme::name() const {
    return d->name;
}
void FS1::Scheme::clear() {
    d->nameHash.clear();
    d->nameHashIsDirty = true;
    d->directory.clear();
    d->rootNode = 0;
}
void FS1::Scheme::rebuild() {
    if (!d->nameHashIsDirty) return;
    LOG_AS("Scheme::rebuild");
    LOGDEV_RES_MSG("Rebuilding '%s'...") << d->name;
    Time begunAt;
    clear();
    d->addFromSearchPaths(FS1::OverridePaths);
    d->addFromSearchPaths(FS1::ExtraPaths);
    d->addFromSearchPaths(FS1::DefaultPaths);
    d->addFromSearchPaths(FS1::FallbackPaths);
    d->nameHashIsDirty = false;
    LOGDEV_RES_VERBOSE("Completed in %.2f seconds") << begunAt.since();
/*#if _DEBUG
    PathTree::debugPrint(d->directory);
    debugPrint();
#endif*/
}
static inline String composeSchemeName(String const &filePath) {
    return filePath.fileNameWithoutExtension();
}
bool FS1::Scheme::add(PathTree::Node &resourceNode) {
    if (!resourceNode.isLeaf()) return false;
    String name = composeSchemeName(resourceNode.name());
    NameHash::hash_type hashKey = NameHash::hashName(name);
    bool isNewNode = false;
    NameHash::Node *hashNode = d->nameHash.findDirectoryNode(hashKey, resourceNode);
    if (!hashNode) {
        isNewNode = true;
        hashNode = new NameHash::Node(resourceNode);
#if _DEBUG
        hashNode->fileRef.setName(name);
#endif
        NameHash::Bucket &bucket = d->nameHash.buckets[hashKey];
        if (bucket.last) bucket.last->next = hashNode;
        bucket.last = hashNode;
        if (!bucket.first) bucket.first = hashNode;
        d->nameHashIsDirty = true;
    }
    FileRef *fileRef = &hashNode->fileRef;
    fileRef->setDirectoryNode(resourceNode);
    return isNewNode;
}
static String const &nameForPathGroup(FS1::PathGroup group) {
    static String const names[] = {
        "Override",
        "Extra",
        "Default",
        "Fallback"
    };
    DENG_ASSERT(int(group) >= FS1::OverridePaths && int(group) <= FS1::FallbackPaths);
    return names[int(group)];
}
bool FS1::Scheme::addSearchPath(SearchPath const &search, FS1::PathGroup group) {
    LOG_AS("Scheme::addSearchPath");
    if (search.isEmpty() ||
        !search.path().toString().compareWithoutCase("/") ||
        !search.path().toString().endsWith("/"))
        return false;
    d->nameHashIsDirty = true;
    DENG2_FOR_EACH(SearchPaths, i, d->searchPaths) {
        if (!i->asText().compareWithoutCase(search.asText())) {
            i->setFlags(search.flags());
            return true;
        }
    }
    d->searchPaths.insert(group, search);
    LOGDEV_RES_MSG("\"%s\" added to scheme '%s' (group:%s)")
        << search << name() << nameForPathGroup(group);
    return true;
}
void FS1::Scheme::clearSearchPathGroup(FS1::PathGroup group) {
    d->searchPaths.remove(group);
}
void FS1::Scheme::clearAllSearchPaths() {
    d->searchPaths.clear();
}
FS1::Scheme::SearchPaths const &FS1::Scheme::allSearchPaths() const {
    return d->searchPaths;
}
int FS1::Scheme::findAll(String name, FoundNodes &found) {
    int numFoundSoFar = found.count();
    NameHash::hash_type fromKey, toKey;
    if (!name.isEmpty()) {
        fromKey = NameHash::hashName(name);
        toKey   = fromKey;
    }
    else {
        fromKey = 0;
        toKey   = NameHash::hash_range - 1;
    }
    for (NameHash::hash_type key = fromKey; key < toKey + 1; ++key) {
        NameHash::Bucket &bucket = d->nameHash.buckets[key];
        for (NameHash::Node *hashNode = bucket.first; hashNode; hashNode = hashNode->next) {
            FileRef &fileRef = hashNode->fileRef;
            PathTree::Node &node = fileRef.directoryNode();
            if (!name.isEmpty() && !node.name().beginsWith(name, String::CaseInsensitive)) continue;
            found.push_back(&node);
        }
    }
    return found.count() - numFoundSoFar;
}
bool FS1::Scheme::mapPath(String &path) const {
    if (path.isEmpty()) return false;
    if (!(d->flags & MappedInPackages)) return false;
    if (path.length() <= name().length()) return false;
    if (path.at(name().length()) != '/')  return false;
    if (!path.beginsWith(name(), String::CaseInsensitive)) return false;
    path = String("$(App.DataPath)/$(GamePlugin.Name)") / path;
    return true;
}
#ifdef DENG_DEBUG
void FS1::Scheme::debugPrint() const {
    LOG_AS("Scheme::debugPrint");
    LOGDEV_RES_MSG("%p:") << this;
    uint schemeIdx = 0;
    for (NameHash::hash_type key = 0; key < NameHash::hash_range; ++key) {
        NameHash::Bucket &bucket = d->nameHash.buckets[key];
        for (NameHash::Node *node = bucket.first; node; node = node->next) {
            FileRef const &fileRef = node->fileRef;
            LOGDEV_RES_MSG("  %u - %u:\"%s\" => %s")
                << schemeIdx << key << fileRef.name()
                << NativePath(fileRef.directoryNode().path()).pretty();
            ++schemeIdx;
        }
    }
    LOGDEV_RES_MSG("  %u %s in scheme.") << schemeIdx << (schemeIdx == 1? "file" : "files");
}
#endif
} // namespace de
int F_Access(char const *nativePath) {
    de::Uri path = de::Uri::fromNativePath(nativePath);
    return App_FileSystem().accessFile(path)? 1 : 0;
}
int F_FileExists(char const *path) {
    int result = -1;
    if (path && path[0]) {
        ddstring_t buf;
        Str_Init(&buf); Str_Set(&buf, path);
        Str_Strip(&buf);
        F_ExpandBasePath(&buf, &buf);
        F_ToNativeSlashes(&buf, &buf);
        result = !access(Str_Text(&buf), 4); // Read permission?
        Str_Free(&buf);
    }
    return result;
}
uint F_GetLastModified(char const *path) {
#ifdef UNIX
    struct stat s;
    stat(path, &s);
    return s.st_mtime;
#endif
#ifdef WIN32
    struct _stat s;
    _stat(path, &s);
    return s.st_mtime;
#endif
}
dd_bool F_MakePath(char const *path) {
    try {
        App::fileSystem().makeFolder(path);
        return true;
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to create path \"%s\": %s") << path << er.asText();
        return false;
    }
    /*
#if !defined(WIN32) && !defined(UNIX)
#  error F_MakePath has no implementation for this platform.
#endif
    ddstring_t full, buf;
    char* ptr, *endptr;
    dd_bool result;
    Str_Init(&full); Str_Set(&full, path);
    Str_Strip(&full);
    F_ToNativeSlashes(&full, &full);
    if (0 == access(Str_Text(&full), 0)) {
        Str_Free(&full);
        return true;
    }
    ptr = Str_Text(&full);
    Str_Init(&buf);
    do {
        endptr = strchr(ptr, DENG_DIR_SEP_CHAR);
        if (!endptr)
            Str_Append(&buf, ptr);
        else
            Str_PartAppend(&buf, ptr, 0, endptr - ptr);
        if (0 != access(Str_Text(&buf), 0)) {
#ifdef WIN32
            mkdir(Str_Text(&buf));
#elif defined(UNIX)
            mkdir(Str_Text(&buf), 0775);
#endif
        }
        Str_AppendChar(&buf, DENG_DIR_SEP_CHAR);
        ptr = endptr + 1;
    } while (NULL != endptr);
    result = (0 == access(Str_Text(&full), 0));
    Str_Free(&buf);
    Str_Free(&full);
    return result;*/
}
dd_bool F_FixSlashes(ddstring_t* dstStr, const ddstring_t* srcStr) {
    dd_bool result = false;
    assert(dstStr && srcStr);
    if (!Str_IsEmpty(srcStr)) {
        char* dst = Str_Text(dstStr);
        const char* src = Str_Text(srcStr);
        size_t i;
        if (dstStr != srcStr) {
            Str_Clear(dstStr);
            Str_Reserve(dstStr, Str_Length(srcStr));
        }
        for (i = 0; src[i]; ++i) {
            if (src[i] != '\\') {
                if (dstStr != srcStr)
                    Str_AppendChar(dstStr, src[i]);
                continue;
            }
            if (dstStr != srcStr)
                Str_AppendChar(dstStr, '/');
            else
                dst[i] = '/';
            result = true;
        }
    }
    return result;
}
/**
 * Appends a slash at the end of @a pathStr if there isn't one.
 * @return @c true if a slash was appended, @c false otherwise.
 */
#ifdef UNIX
static bool F_AppendMissingSlash(ddstring_t *pathStr) {
    if (Str_RAt(pathStr, 0) != '/') {
        Str_AppendChar(pathStr, '/');
        return true;
    }
    return false;
}
#endif
dd_bool F_AppendMissingSlashCString(char* path, size_t maxLen) {
    if (path[strlen(path) - 1] != '/') {
        M_StrCat(path, "/", maxLen);
        return true;
    }
    return false;
}
dd_bool F_ToNativeSlashes(ddstring_t* dstStr, const ddstring_t* srcStr) {
    dd_bool result = false;
    assert(dstStr && srcStr);
    if (!Str_IsEmpty(srcStr)) {
        char* dst = Str_Text(dstStr);
        const char* src = Str_Text(srcStr);
        size_t i;
        if (dstStr != srcStr) {
            Str_Clear(dstStr);
            Str_Reserve(dstStr, Str_Length(srcStr));
        }
        for (i = 0; src[i]; ++i) {
            if (src[i] != DENG_DIR_WRONG_SEP_CHAR) {
                if (dstStr != srcStr)
                    Str_AppendChar(dstStr, src[i]);
                continue;
            }
            if (dstStr != srcStr)
                Str_AppendChar(dstStr, DENG_DIR_SEP_CHAR);
            else
                dst[i] = DENG_DIR_SEP_CHAR;
            result = true;
        }
    }
    return result;
}
/**
 * @return  @c true iff the path can be made into a relative path.
 */
static bool F_IsRelativeToBase(char const *path, char const *base) {
    DENG_ASSERT(path != 0 && base != 0);
    return !qstrnicmp(path, base, strlen(base));
}
/**
 * Attempt to remove the base path if found at the beginning of the path.
 *
 * @param dst  Potential base-relative path written here.
 * @param src  Possibly absolute path.
 *
 * @return  @c true iff the base path was found and removed.
 */
static dd_bool F_RemoveBasePath(ddstring_t *dst, ddstring_t const *absPath) {
    DENG_ASSERT(dst != 0 && absPath != 0);
    ddstring_t basePath;
    Str_InitStatic(&basePath, DoomsdayApp::app().doomsdayBasePath().c_str());
    if (F_IsRelativeToBase(Str_Text(absPath), Str_Text(&basePath))) {
        dd_bool mustCopy = (dst == absPath);
        if (mustCopy) {
            ddstring_t buf;
            Str_Init(&buf);
            Str_PartAppend(&buf, Str_Text(absPath), Str_Length(&basePath), Str_Length(absPath) - Str_Length(&basePath));
            Str_Set(dst, Str_Text(&buf));
            Str_Free(&buf);
            return true;
        }
        Str_Clear(dst);
        Str_PartAppend(dst, Str_Text(absPath), Str_Length(&basePath), Str_Length(absPath) - Str_Length(&basePath));
        return true;
    }
    if (dst != absPath) {
        Str_Set(dst, Str_Text(absPath));
    }
    return false;
}
dd_bool F_IsAbsolute(const ddstring_t* str) {
    if (!str)
        return false;
    if (Str_At(str, 0) == DENG_DIR_SEP_CHAR || Str_At(str, 0) == DENG_DIR_WRONG_SEP_CHAR || Str_At(str, 1) == ':')
        return true;
#ifdef UNIX
    if (Str_At(str, 0) == '~')
        return true;
#endif
    return false;
}
dd_bool F_ExpandBasePath(ddstring_t* dst, const ddstring_t* src) {
    assert(dst && src);
    if (Str_At(src, 0) == '>' || Str_At(src, 0) == '}') {
        dd_bool mustCopy = (dst == src);
        if (mustCopy) {
            ddstring_t buf;
            Str_Init(&buf); Str_Set(&buf, DoomsdayApp::app().doomsdayBasePath().c_str());
            Str_PartAppend(&buf, Str_Text(src), 1, Str_Length(src)-1);
            Str_Set(dst, Str_Text(&buf));
            Str_Free(&buf);
            return true;
        }
        Str_Set(dst, DoomsdayApp::app().doomsdayBasePath().c_str());
        Str_PartAppend(dst, Str_Text(src), 1, Str_Length(src)-1);
        return true;
    }
#ifdef UNIX
    else if (Str_At(src, 0) == '~') {
        if (Str_At(src, 1) == '/' && getenv("HOME"))
        {   // Replace it with the HOME environment variable.
            ddstring_t buf;
            ddstring_t homeStr;
            Str_Init(&buf);
            Str_Init(&homeStr);
            Str_Set(&homeStr, getenv("HOME"));
            F_FixSlashes(&buf, &homeStr);
            F_AppendMissingSlash(&buf);
            Str_PartAppend(&buf, Str_Text(src), 2, Str_Length(src)-2);
            Str_Set(dst, Str_Text(&buf));
            Str_Free(&buf);
            Str_Free(&homeStr);
            return true;
        }
        { dd_bool result = false;
        ddstring_t userName;
        const char* p = Str_Text(src)+2;
        Str_Init(&userName);
        if ((p = Str_CopyDelim2(&userName, p, '/', CDF_OMIT_DELIMITER))) {
            ddstring_t buf;
            struct passwd* pw;
            Str_Init(&buf);
            if ((pw = getpwnam(Str_Text(&userName))) != NULL) {
                ddstring_t pwStr;
                Str_Init(&pwStr);
                Str_Set(&pwStr, pw->pw_dir);
                F_FixSlashes(&buf, &pwStr);
                F_AppendMissingSlash(&buf);
                result = true;
                Str_Free(&pwStr);
            }
            Str_Append(&buf, Str_Text(src) + 1);
            Str_Set(dst, Str_Text(&buf));
            Str_Free(&buf);
        }
        Str_Free(&userName);
        if (result)
            return result;
        }
    }
#endif
    if (dst != src)
        Str_Set(dst, Str_Text(src));
    return false;
}
static dd_bool pathHasDirective(const char* path) {
    if (NULL != path && path[0]) {
#ifdef UNIX
        if ('~' == path[0]) return true;
#endif
        if ('}' == path[0] || '>' == path[0]) return true;
    }
    return false;
}
const char* F_PrettyPath(const char* path) {
#define NUM_BUFS            8
    static ddstring_t buffers[NUM_BUFS]; // @todo: never free'd!
    static uint index = 0;
    ddstring_t* buf = NULL;
    int len;
    if (!path || 0 == (len = (int)strlen(path)))
        return path;
    if (len > 1 && pathHasDirective(path)) {
        buf = &buffers[index++ % NUM_BUFS];
        Str_Clear(buf);
        Str_PartAppend(buf, path, 1, len-1);
        path = Str_Text(buf);
    }
    if (F_IsRelativeToBase(path, DoomsdayApp::app().doomsdayBasePath().c_str())) {
        if (!buf) {
            buf = &buffers[index++ % NUM_BUFS];
            Str_Set(buf, path);
        }
        F_RemoveBasePath(buf, buf);
        path = Str_Text(buf);
    }
    if (strchr(path, DENG_DIR_WRONG_SEP_CHAR)) {
        int i;
        if (!buf) {
            buf = &buffers[index++ % NUM_BUFS];
            Str_Set(buf, path);
        }
        for (i = 0; i < len; ++i) {
            if (buf->str[i] == DENG_DIR_WRONG_SEP_CHAR)
               buf->str[i] = DENG_DIR_SEP_CHAR;
        }
        path = Str_Text(buf);
    }
    return path;
#undef NUM_BUFS
}
/*
dd_bool F_Dump(void const *data, size_t size, char const *path) {
    DENG2_ASSERT(data != 0 && path != 0);
    if (!size) return false;
    AutoStr *nativePath = AutoStr_NewStd();
    Str_Set(nativePath, path);
    F_ToNativeSlashes(nativePath, nativePath);
    FILE *outFile = fopen(Str_Text(nativePath), "wb");
    if (!outFile) {
        LOG_RES_WARNING("Failed to open \"%s\" for writing: %s")
                << F_PrettyPath(Str_Text(nativePath)) << strerror(errno);
        return false;
    }
    fwrite(data, 1, size, outFile);
    fclose(outFile);
    return true;
}*/
dd_bool F_DumpFile(File1 &file, char const *outputPath) {
    String dumpPath = "/home" / ((!outputPath || !outputPath[0])? file.name() : String(outputPath));
    try {
        File &out = App::rootFolder().replaceFile(dumpPath);
        out << Block(file.cache(), file.info().size);
        out.flush();
        file.unlock();
        LOG_RES_VERBOSE("%s dumped to %s") << file.name() << out.description();
        return true;
    }
    catch (Error const &er) {
        LOG_RES_ERROR("Failed to write to \"%s\": %s") << dumpPath << er.asText();
        return false;
    }
}
bool F_DumpNativeFile(Block const &data, NativePath const &filePath) {
    try {
        std::unique_ptr<File> file(NativeFile::newStandalone(filePath));
        file->setMode(File::Write);
        *file << data;
        return true;
    }
    catch (Error const &) {
        return false;
    }
}
static String const DOMAIN_IDGAMES("idgames");
static String const CATEGORY_LEVELS("levels");
static String const CATEGORY_MUSIC ("music");
static String const CATEGORY_SOUNDS("sounds");
static String const CATEGORY_THEMES("themes");
DENG2_PIMPL(IdgamesLink) {
    struct PackageIndexEntry : public PathTree::Node {
        FileEntry const *file = nullptr;
        Version version;
        PackageIndexEntry(PathTree::NodeArgs const &args) : Node(args) {}
        String descriptionPath() const {
            return file->path().toString().fileNameAndPathWithoutExtension() + ".txt";
        }
    };
    PathTreeT<PackageIndexEntry> packageIndex;
    String localRootPath;
    Impl(Public *i) : Base(i) {
        localRootPath = "/remote/" + QUrl(self().address()).host();
    }
    String packageIdentifierForFileEntry(FileEntry const &entry) const {
        if (entry.name().fileNameExtension() == QStringLiteral(".zip")) {
            Path const path = entry.path();
            String id = QString("%1_%2")
                    .arg(DataBundle::cleanIdentifier(path.fileName().fileNameWithoutExtension()))
                    .arg(DataBundle::versionFromTimestamp(entry.modTime));
            id.truncate(id.size() - 5);
            if (path.segment(1) == CATEGORY_MUSIC ||
                path.segment(1) == CATEGORY_SOUNDS ||
                path.segment(1) == CATEGORY_THEMES) {
                return String("%1.%2.%3")
                        .arg(DOMAIN_IDGAMES)
                        .arg(path.segment(1))
                        .arg(id);
            }
            if (path.segment(1) == CATEGORY_LEVELS) {
                String subset;
                if      (path.segment(3) == QStringLiteral("deathmatch")) subset = QStringLiteral("deathmatch.");
                else if (path.segment(3) == QStringLiteral("megawads"))   subset = QStringLiteral("megawads.");
                return String("%1.%2.%3.%4%5")
                        .arg(DOMAIN_IDGAMES)
                        .arg(CATEGORY_LEVELS)
                        .arg(path.segment(2))
                        .arg(subset)
                        .arg(id);
            }
            return String("%1.%2").arg(DOMAIN_IDGAMES).arg(id);
        }
        else {
            return String();
        }
    }
    void buildPackageIndex() {
        packageIndex.clear();
        PathTreeIterator<FileTree> iter(self().fileTree().leafNodes());
        while (iter.hasNext()) {
            auto const &fileEntry = iter.next();
            if (String pkg = packageIdentifierForFileEntry(fileEntry)) {
                auto const id_ver = Package::split(pkg);
                auto &pkgEntry = packageIndex.insert(DotPath(id_ver.first));
                pkgEntry.file    = &fileEntry;
                pkgEntry.version = id_ver.second;
            }
        }
        qDebug() << "idgames package index has" << packageIndex.size() << "entries";
    }
    PackageIndexEntry const *findPackage(String const &packageId) const {
        auto const id_ver = Package::split(packageId);
        if (auto *found = packageIndex.tryFind(DotPath(id_ver.first),
                                               PathTree::MatchFull | PathTree::NoBranch)) {
            if (!id_ver.second.isValid() || found->version == id_ver.second) {
                return found;
            }
        }
        return nullptr;
    }
    RemoteFile &makeRemoteFile(Folder &folder, String const &remotePath, Block const &remoteMetaId) {
        auto *file = new RemoteFile(remotePath.fileName(),
                                    remotePath,
                                    remoteMetaId,
                                    self().address());
        folder.add(file);
        FS::get().index(*file);
        return *file;
    }
};
IdgamesLink::IdgamesLink(String const &address)
    : filesys::WebHostedLink(address, "ls-laR.gz")
    , d(new Impl(this))
{}
void IdgamesLink::parseRepositoryIndex(QByteArray data) {
    scope() += async([this, data] () -> String {
        Block const listing = gDecompress(data);
        QTextStream is(listing, QIODevice::ReadOnly);
        is.setCodec("UTF-8");
        QRegularExpression const reDir("^\\.?(.*):$");
        QRegularExpression const reTotal("^total\\s+\\d+$");
        QRegularExpression const reFile("^(-|d)[-rwxs]+\\s+\\d+\\s+\\w+\\s+\\w+\\s+"
                                        "(\\d+)\\s+(\\w+\\s+\\d+\\s+[0-9:]+)\\s+(.*)$",
                                        QRegularExpression::CaseInsensitiveOption);
        String currentPath;
        bool ignore = false;
        QRegularExpression const reIncludedPaths("^/(levels|music|sounds|themes)");
        std::unique_ptr<FileTree> tree(new FileTree);
        while (!is.atEnd()) {
            if (String const line = is.readLine().trimmed()) {
                if (!currentPath) {
                    auto match = reDir.match(line);
                    if (match.hasMatch()) {
                        currentPath = match.captured(1);
                        ignore = !reIncludedPaths.match(currentPath).hasMatch();
                    }
                }
                else if (!ignore && reTotal.match(line).hasMatch()) {
                }
                else if (!ignore) {
                    auto match = reFile.match(line);
                    if (match.hasMatch()) {
                        bool const isFolder = (match.captured(1) == QStringLiteral("d"));
                        if (!isFolder) {
                            String const name = match.captured(4);
                            if (name.startsWith(QChar('.')) || name.contains(" -> "))
                                continue;
                            auto &entry = tree->insert((currentPath / name).toLower());
                            entry.size = match.captured(2).toULongLong(nullptr, 10);;
                            entry.modTime = Time::fromText(match.captured(3), Time::UnixLsStyleDateTime);
                        }
                    }
                }
            }
            else {
                currentPath.clear();
            }
        }
        qDebug() << "idgames file tree contains" << tree->size() << "entries";
        setFileTree(tree.release());
        return String();
    },
    [this] (String const &errorMessage) {
        if (!errorMessage) {
            wasConnected();
        }
        else {
            handleError("Failed to parse directory listing: " + errorMessage);
            wasDisconnected();
        }
});
}
StringList IdgamesLink::categoryTags() const {
    return StringList({ CATEGORY_LEVELS,
                        CATEGORY_MUSIC,
                        CATEGORY_SOUNDS,
                        CATEGORY_THEMES });
}
LoopResult IdgamesLink::forPackageIds(std::function<LoopResult (String const &)> func) const {
    PathTreeIterator<Impl::PackageIndexEntry> iter(d->packageIndex.leafNodes());
    while (iter.hasNext()) {
        if (auto result = func(iter.next().path('.'))) {
            return result;
        }
    }
    return LoopContinue;
}
String IdgamesLink::findPackagePath(String const &packageId) const {
    if (auto *found = d->findPackage(packageId)) {
        return found->file->path();
    }
    return String();
}
filesys::Link *IdgamesLink::construct(String const &address) {
    if ((address.startsWith("http:") || address.startsWith("https:")) &&
        !address.contains("dengine.net")) {
        return new IdgamesLink(address);
    }
    return nullptr;
}
File *IdgamesLink::populateRemotePath(String const &packageId,
                                      filesys::RepositoryPath const &path) const {
    DENG2_ASSERT(path.link == this);
    auto *pkgEntry = d->findPackage(packageId);
    if (!pkgEntry) {
        DENG2_ASSERT(pkgEntry);
        return nullptr;
    }
    auto &pkgFolder = FS::get().makeFolder(path.localPath, FS::DontInheritFeeds);
    auto &dataFile = d->makeRemoteFile(pkgFolder,
                                       pkgEntry->file->path(),
                                       pkgEntry->file->metaId(*this));
    dataFile.setStatus(File::Status(pkgEntry->file->size, pkgEntry->file->modTime));
    auto &txtFile = d->makeRemoteFile(pkgFolder,
                                      pkgEntry->descriptionPath(),
                                      md5Hash(address(), pkgEntry->descriptionPath(), pkgEntry->file->modTime));
    if (auto const *txtEntry = findFile(pkgEntry->descriptionPath())) {
        txtFile.setStatus(File::Status(txtEntry->size, txtEntry->modTime));
    }
    auto *infoFile = new IdgamesPackageInfoFile("info.dei");
    infoFile->setSourceFiles(dataFile, txtFile);
    pkgFolder.add(infoFile);
    FS::get().index(*infoFile);
    return &pkgFolder;
}
void IdgamesLink::setFileTree(FileTree *tree) {
    WebHostedLink::setFileTree(tree);
    d->buildPackageIndex();
}
DENG2_PIMPL(IdgamesPackageInfoFile)
, DENG2_OBSERVES(Asset, StateChange) {
    Asset packageAsset;
    AssetGroup assets;
    SafePtr<RemoteFile const> dataFile;
    SafePtr<RemoteFile const> descriptionFile;
    Block serializedContent;
    Impl(Public *i) : Base(i) {
        assets.audienceForStateChange() += this;
    }
    String cacheFolderPath() const {
        String const hex = dataFile->metaId().asHexadecimalText();
        return RemoteFile::CACHE_PATH / hex.right(1) / hex;
    }
    void assetStateChanged(Asset &) {
        if (!assets.isEmpty() && assets.isReady()) {
            qDebug() << "[IdgamesPackageInfoFile] Time to unzip and analyze!";
            DENG2_ASSERT(dataFile->isReady());
            DENG2_ASSERT(descriptionFile->isReady());
            auto &fs = FileSystem::get();
            Folder &packFolder = *self().parent();
            DotPath const packageId { packFolder.name() };
            /*Folder &cacheFolder = */fs.makeFolder(cacheFolderPath());
            FS::copySerialized(descriptionFile->path(),
                               cacheFolderPath()/descriptionFile->name());
            StringList dataFiles;
            Record meta;
            meta.set("title",   dataFile->name().fileNameWithoutExtension().toUpper() + " (idgames)");
            meta.set("version", DataBundle::versionFromTimestamp(dataFile->target().status().modifiedAt));
            if (Folder *zip = const_cast<Folder *>(maybeAs<Folder>(dataFile->target()))) {
                zip->populate();
                zip->forContents([this, &dataFiles] (String name, File &file) {
                    String const ext = name.fileNameExtension().toLower();
                    if (ext == ".wad" || ext == ".deh" || ext == ".lmp" || ext == ".pk3") {
                        File &copied = FS::copySerialized
                                (file.path(), cacheFolderPath()/file.name());
                        dataFiles << copied.path();
                    }
                    return LoopContinue;
                });
                FS::waitForIdle();
                StringList components;
                foreach (String path, dataFiles) {
                    if (DataBundle const *bundle = FS::tryLocate<DataBundle const>(path)) {
                        components << bundle->asFile().path();
                        Record const &compMeta = bundle->packageMetadata();
                        if (compMeta.has("notes")) {
                            meta.set("notes", compMeta.gets("notes"));
                            if (compMeta.has("title")) {
                                meta.set("title", compMeta.gets("title"));
                            }
                            if (compMeta.has("license")) {
                                meta.set("license", compMeta.gets("license"));
                            }
                            if (compMeta.has("author")) {
                                meta.set("author", compMeta.gets("author"));
                            }
                            if (compMeta.has("tags")) {
                                meta.appendMultipleUniqueWords("tags", compMeta.gets("tags"));
                            }
                        }
                    }
                }
                std::unique_ptr<ArrayValue> comps { new ArrayValue };
                foreach (String comp, components) {
                    comps->add(new TextValue(comp));
                }
                meta.addArray("dataFiles", comps.release());
                meta.set("tags", meta.gets("tags", "")
                         .removed(QRegularExpression("\\b(hidden|cached)\\b"))
                         .normalizeWhitespace());
                qDebug() << "idgames package will contain:" << components;
                qDebug() << meta.asText().toUtf8().constData();
                if (packageId.segment(1) == "levels") {
                    meta.set("tags", meta.gets("tags", "")
                             .removed(QRegularExpression(DataBundle::anyGameTagPattern()))
                             .normalizeWhitespace());
                    meta.appendUniqueWord("tags", packageId.segment(2).toString());
                }
                auto &pkgMeta = Package::initializeMetadata(packFolder, packageId);
                pkgMeta.copyMembersFrom(meta); {
                    Record info { pkgMeta, Record::IgnoreDoubleUnderscoreMembers };
                    delete info.remove("ID");
                    delete info.remove("path");
                    String const cachePath = cacheFolderPath();
                    ArrayValue *pkgData = new ArrayValue;
                    foreach (String path, info.getStringList("dataFiles")) {
                        if (path.startsWith(cachePath)) {
                            path.remove(0, cachePath.size() + 1);
                        }
                        pkgData->add(new TextValue(path));
                    }
                    info.set("dataFiles", pkgData);
                    serializedContent = info.asInfo().toUtf8();
                }
            }
            DENG2_FOR_PUBLIC_AUDIENCE(Download, i) {
                i->downloadProgress(self(), 0);
            }
            packageAsset.setState(Asset::Ready);
        }
    }
};
IdgamesPackageInfoFile::IdgamesPackageInfoFile(String const &name)
    : File(name)
    , d(new Impl(this))
{}
void IdgamesPackageInfoFile::setSourceFiles(RemoteFile const &dataFile,
                                            RemoteFile const &descriptionFile) {
    d->dataFile       .reset(&dataFile);
    d->descriptionFile.reset(&descriptionFile);
    d->assets += dataFile.asset();
    d->assets += descriptionFile.asset();
}
Asset &IdgamesPackageInfoFile::asset() {
    return d->packageAsset;
}
Asset const &IdgamesPackageInfoFile::asset() const {
    return d->packageAsset;
}
dsize IdgamesPackageInfoFile::downloadSize() const {
    return 0;
}
void IdgamesPackageInfoFile::download()
{}
void IdgamesPackageInfoFile::cancelDownload()
{}
IIStream const &IdgamesPackageInfoFile::operator >> (IByteArray &bytes) const {
    bytes.set(0, d->serializedContent.dataConst(), d->serializedContent.size());
    return *this;
}
LumpCache::Data::Data(uint8_t *data)
    : data_(data)
{}
LumpCache::Data::~Data() {
    clearData();
}
uint8_t *LumpCache::Data::data() const {
    if (data_ && Z_GetTag(data_) == PU_PURGELEVEL) {
        Z_ChangeTag2(data_, PU_APPSTATIC);
        Z_ChangeUser(data_, (void*)&data_);
    }
    return data_;
}
uint8_t const *LumpCache::Data::replaceData(uint8_t *newData) {
    clearData();
    data_ = newData;
    if (data_) {
        Z_ChangeUser(data_, &data_);
    }
    return newData;
}
LumpCache::Data &LumpCache::Data::clearData(bool *retCleared) {
    bool hasData = !!data_;
    if (hasData) {
        if (Z_GetTag(data_) != PU_PURGELEVEL) {
            Z_ChangeTag2(data_, PU_PURGELEVEL);
        }
        Z_ChangeUser(data_, (void *) 0x2);
    }
    if (retCleared) *retCleared = hasData;
    return *this;
}
LumpCache::Data &LumpCache::Data::lock() {
    return *this;
}
LumpCache::Data &LumpCache::Data::unlock() {
    if (data_) {
        Z_ChangeTag2(data_, PU_PURGELEVEL);
    }
    return *this;
}
LumpCache::LumpCache(uint size) : _size(size), _dataCache(0)
{}
LumpCache::~LumpCache() {
    if (_dataCache) delete _dataCache;
}
uint LumpCache::size() const {
    return _size;
}
bool LumpCache::isValidIndex(uint idx) const {
    return idx < _size;
}
uint8_t const *LumpCache::data(uint lumpIdx) const {
    LOG_AS("LumpCache::data");
    Data const* record = cacheRecord(lumpIdx);
    return record? record->data() : 0;
}
LumpCache &LumpCache::insert(uint lumpIdx, uint8_t *data) {
    LOG_AS("LumpCache::insert");
    if (!isValidIndex(lumpIdx)) throw Error("LumpCache::insert", QString("Invalid index %1").arg(lumpIdx));
    if (!_dataCache) {
        _dataCache = new DataCache(_size);
    }
    Data *record = cacheRecord(lumpIdx);
    record->replaceData(data);
    return *this;
}
LumpCache &LumpCache::insertAndLock(uint lumpIdx, uint8_t *data) {
    return insert(lumpIdx, data).lock(lumpIdx);
}
LumpCache &LumpCache::lock(uint lumpIdx) {
    LOG_AS("LumpCache::lock");
    if (!isValidIndex(lumpIdx)) throw Error("LumpCache::lock", QString("Invalid index %1").arg(lumpIdx));
    Data* record = cacheRecord(lumpIdx);
    record->lock();
    return *this;
}
LumpCache &LumpCache::unlock(uint lumpIdx) {
    LOG_AS("LumpCache::unlock");
    if (!isValidIndex(lumpIdx)) throw Error("LumpCache::unlock", QString("Invalid index %1").arg(lumpIdx));
    Data* record = cacheRecord(lumpIdx);
    record->unlock();
    return *this;
}
LumpCache &LumpCache::remove(uint lumpIdx, bool *retRemoved) {
    Data *record = cacheRecord(lumpIdx);
    if (record) {
        record->clearData(retRemoved);
    }
    else if (retRemoved) {
        *retRemoved = false;
    }
    return *this;
}
LumpCache &LumpCache::clear() {
    if (_dataCache) {
        DENG2_FOR_EACH(DataCache, i, *_dataCache) {
            i->clearData();
        }
    }
    return *this;
}
LumpCache::Data *LumpCache::cacheRecord(uint lumpIdx) {
    if (!isValidIndex(lumpIdx)) return 0;
    return _dataCache? &(*_dataCache)[lumpIdx] : 0;
}
LumpCache::Data const *LumpCache::cacheRecord(uint lumpIdx) const {
    if (!isValidIndex(lumpIdx)) return 0;
    return _dataCache? &(*_dataCache)[lumpIdx] : 0;
}
namespace de {
namespace internal {
    struct LumpSortInfo {
        File1 const *lump;
        String path;
        int origIndex;
    };
    static int lumpSorter(void const *a, void const *b) {
        LumpSortInfo const *infoA = (LumpSortInfo const *)a;
        LumpSortInfo const *infoB = (LumpSortInfo const *)b;
        if (int delta = infoA->path.compare(infoB->path, Qt::CaseInsensitive))
            return delta;
        if (int delta = (infoA->lump->container().loadOrderIndex() -
                         infoB->lump->container().loadOrderIndex()))
            return delta;
        return (infoB->origIndex - infoA->origIndex);
    }
} // namespace internal
using namespace internal;
DENG2_PIMPL_NOREF(LumpIndex::Id1MapRecognizer) {
    lumpnum_t lastLump = -1;
    Lumps lumps;
    String id;
    Format format = UnknownFormat;
};
LumpIndex::Id1MapRecognizer::Id1MapRecognizer(LumpIndex const &lumpIndex, lumpnum_t lumpIndexOffset)
    : d(new Impl) {
    LOG_AS("LumpIndex::Id1MapRecognizer");
    LOG_RES_XVERBOSE("Locating data lumps...", "");
    dint const numLumps = lumpIndex.size();
    String sourceFile;
    for (d->lastLump = de::max(lumpIndexOffset, 0); d->lastLump < numLumps; ++d->lastLump) {
        File1 &lump       = lumpIndex[d->lastLump];
        DataType dataType = typeForLumpName(lump.name());
        if (d->lumps.isEmpty()) {
            if (dataType == UnknownData) continue;
            if (d->lastLump == 0) return;
            if (dataType == UDMFTextmapData) {
                d->format = UniversalFormat;
            }
            d->id = lumpIndex[d->lastLump - 1].name().fileNameAndPathWithoutExtension();
            sourceFile = lump.container().composePath();
        }
        else {
            if (d->format == UniversalFormat) {
                if (dataType == UDMFEndmapData) break;
            }
            else {
                if (dataType == UnknownData) break;
            }
            if (sourceFile.compareWithoutCase(lump.container().composePath()))
                break;
        }
        d->lumps.insert(dataType, &lump);
    }
    if (d->lumps.isEmpty()) return;
    if (d->format == UnknownFormat) {
        if (d->lumps.contains(BehaviorData)) {
            d->format = HexenFormat;
        }
        else if (d->lumps.contains(MacroData) ||
                 d->lumps.contains(TintColorData) ||
                 d->lumps.contains(LeafData)) {
            d->format = Doom64Format;
        }
        else {
            d->format = DoomFormat;
        }
        duint numVertexes = 0, numThings = 0, numLines = 0, numSides = 0, numSectors = 0, numLights = 0;
        DENG2_FOR_EACH_CONST(Lumps, i, d->lumps) {
            DataType const dataType = i.key();
            File1 const &lump       = *i.value();
            duint *elemCountAddr = 0;
            dsize const elemSize = elementSizeForDataType(d->format, dataType);
            switch (dataType) {
            default: break;
            case VertexData:    elemCountAddr = &numVertexes; break;
            case ThingData:     elemCountAddr = &numThings;   break;
            case LineDefData:   elemCountAddr = &numLines;    break;
            case SideDefData:   elemCountAddr = &numSides;    break;
            case SectorDefData: elemCountAddr = &numSectors;  break;
            case TintColorData: elemCountAddr = &numLights;   break;
            }
            if (elemCountAddr) {
                if (lump.size() % elemSize != 0) {
                    d->format = UnknownFormat;
                    d->id.clear();
                    return;
                }
                *elemCountAddr += lump.size() / elemSize;
            }
        }
        if (!numVertexes || !numLines || !numSides || !numSectors) {
            d->format = UnknownFormat;
            d->id.clear();
            return;
        }
    }
    LOG_RES_VERBOSE("Recognized %s format map") << formatName(d->format);
}
String const &LumpIndex::Id1MapRecognizer::id() const {
    return d->id;
}
LumpIndex::Id1MapRecognizer::Format LumpIndex::Id1MapRecognizer::format() const {
    return d->format;
}
LumpIndex::Id1MapRecognizer::Lumps const &LumpIndex::Id1MapRecognizer::lumps() const {
    return d->lumps;
}
File1 *LumpIndex::Id1MapRecognizer::sourceFile() const {
    if (d->lumps.isEmpty()) return nullptr;
    return &lumps().first()->container();
}
lumpnum_t LumpIndex::Id1MapRecognizer::lastLump() const {
    return d->lastLump;
}
String const &LumpIndex::Id1MapRecognizer::formatName(Format id) // static
{
    static String const names[1 + KnownFormatCount] = {
        "Unknown",
        "id Tech 1 (Doom)",
        "id Tech 1 (Hexen)",
        "id Tech 1 (Doom64)",
        "id Tech 1 (UDMF)"
    };
    if (id >= DoomFormat && id < KnownFormatCount) {
        return names[1 + id];
    }
    return names[0];
}
LumpIndex::Id1MapRecognizer::DataType LumpIndex::Id1MapRecognizer::typeForLumpName(String name) // static
{
    static QHash<String, DataType> const lumpTypeInfo {
        std::make_pair(String("THINGS"),   ThingData      ),
        std::make_pair(String("LINEDEFS"), LineDefData    ),
        std::make_pair(String("SIDEDEFS"), SideDefData    ),
        std::make_pair(String("VERTEXES"), VertexData     ),
        std::make_pair(String("SEGS"),     SegData        ),
        std::make_pair(String("SSECTORS"), SubsectorData  ),
        std::make_pair(String("NODES"),    NodeData       ),
        std::make_pair(String("SECTORS"),  SectorDefData  ),
        std::make_pair(String("REJECT"),   RejectData     ),
        std::make_pair(String("BLOCKMAP"), BlockmapData   ),
        std::make_pair(String("BEHAVIOR"), BehaviorData   ),
        std::make_pair(String("SCRIPTS"),  ScriptData     ),
        std::make_pair(String("LIGHTS"),   TintColorData  ),
        std::make_pair(String("MACROS"),   MacroData      ),
        std::make_pair(String("LEAFS"),    LeafData       ),
        std::make_pair(String("GL_VERT"),  GLVertexData   ),
        std::make_pair(String("GL_SEGS"),  GLSegData      ),
        std::make_pair(String("GL_SSECT"), GLSubsectorData),
        std::make_pair(String("GL_NODES"), GLNodeData     ),
        std::make_pair(String("GL_PVS"),   GLPVSData      ),
        std::make_pair(String("TEXTMAP"),  UDMFTextmapData),
        std::make_pair(String("ENDMAP"),   UDMFEndmapData ),
    };
    auto found = lumpTypeInfo.constFind(name.fileNameWithoutExtension().toUpper());
    if (found != lumpTypeInfo.constEnd()) {
        return found.value();
    }
    return UnknownData;
}
dsize LumpIndex::Id1MapRecognizer::elementSizeForDataType(Format mapFormat, DataType dataType) // static
{
    dsize const SIZEOF_64VERTEX  = (4 * 2);
    dsize const SIZEOF_VERTEX    = (2 * 2);
    dsize const SIZEOF_SIDEDEF   = (2 * 3 + 8 * 3);
    dsize const SIZEOF_64SIDEDEF = (2 * 6);
    dsize const SIZEOF_LINEDEF   = (2 * 7);
    dsize const SIZEOF_64LINEDEF = (2 * 6 + 1 * 4);
    dsize const SIZEOF_XLINEDEF  = (2 * 5 + 1 * 6);
    dsize const SIZEOF_SECTOR    = (2 * 5 + 8 * 2);
    dsize const SIZEOF_64SECTOR  = (2 * 12);
    dsize const SIZEOF_THING     = (2 * 5);
    dsize const SIZEOF_64THING   = (2 * 7);
    dsize const SIZEOF_XTHING    = (2 * 7 + 1 * 6);
    dsize const SIZEOF_LIGHT     = (1 * 6);
    switch (dataType) {
    default: return 0;
    case VertexData:
        return (mapFormat == Doom64Format? SIZEOF_64VERTEX : SIZEOF_VERTEX);
    case LineDefData:
        return (mapFormat == Doom64Format? SIZEOF_64LINEDEF :
                mapFormat == HexenFormat ? SIZEOF_XLINEDEF  : SIZEOF_LINEDEF);
    case SideDefData:
        return (mapFormat == Doom64Format? SIZEOF_64SIDEDEF : SIZEOF_SIDEDEF);
    case SectorDefData:
        return (mapFormat == Doom64Format? SIZEOF_64SECTOR : SIZEOF_SECTOR);
    case ThingData:
        return (mapFormat == Doom64Format? SIZEOF_64THING :
                mapFormat == HexenFormat ? SIZEOF_XTHING  : SIZEOF_THING);
    case TintColorData: return SIZEOF_LIGHT;
    }
}
DENG2_PIMPL(LumpIndex) {
    bool pathsAreUnique;
    Lumps lumps;
    bool needPruneDuplicateLumps;
    struct PathHashRecord {
        lumpnum_t head, nextInLoadOrder;
    };
    typedef QVector<PathHashRecord> PathHash;
    QScopedPointer<PathHash> lumpsByPath;
    Impl(Public *i)
        : Base(i)
        , pathsAreUnique         (false)
        , needPruneDuplicateLumps(false)
    {}
    ~Impl() { self().clear(); }
    void buildLumpsByPathIfNeeded() {
        if (!lumpsByPath.isNull()) return;
        int const numElements = lumps.size();
        lumpsByPath.reset(new PathHash(numElements));
        DENG2_FOR_EACH(PathHash, i, *lumpsByPath) {
            i->head = -1;
        }
        for (int i = 0; i < numElements; ++i) {
            File1 const &lump          = *(lumps[i]);
            PathTree::Node const &node = lump.directoryNode();
            ushort k = node.hash() % (unsigned)numElements;
            (*lumpsByPath)[i].nextInLoadOrder = (*lumpsByPath)[k].head;
            (*lumpsByPath)[k].head = i;
        }
        LOG_RES_XVERBOSE("Rebuilt hashMap for LumpIndex %p", thisPublic);
    }
    /**
     * @param pruneFlags  Passed by reference to avoid deep copy on value-write.
     * @param file        Flag only those lumps contained by this file.
     *
     * @return Number of lumps newly flagged during this op.
     */
    int flagContainedLumps(QBitArray &pruneFlags, File1 &file) {
        DENG2_ASSERT(pruneFlags.size() == lumps.size());
        int const numRecords = lumps.size();
        int numFlagged = 0;
        for (int i = 0; i < numRecords; ++i) {
            File1 *lump = lumps[i];
            if (pruneFlags.testBit(i)) continue;
            if (!lump->isContained() || &lump->container() != &file) continue;
            pruneFlags.setBit(i, true);
            numFlagged += 1;
        }
        return numFlagged;
    }
    /**
     * @param pruneFlags  Passed by reference to avoid deep copy on value-write.
     * @return Number of lumps newly flagged during this op.
     */
    int flagDuplicateLumps(QBitArray &pruneFlags) {
        DENG2_ASSERT(pruneFlags.size() == lumps.size());
        if (!pathsAreUnique) return 0;
        if (!needPruneDuplicateLumps) return 0;
        int const numRecords = lumps.size();
        if (numRecords <= 1) return 0;
        LumpSortInfo *sortInfos = new LumpSortInfo[numRecords];
        for (int i = 0; i < numRecords; ++i) {
            LumpSortInfo &sortInfo = sortInfos[i];
            File1 const *lump      = lumps[i];
            sortInfo.lump      = lump;
            sortInfo.path      = lump->composePath();
            sortInfo.origIndex = i;
        }
        qsort(sortInfos, numRecords, sizeof(*sortInfos), lumpSorter);
        int numFlagged = 0;
        for (int i = 1; i < numRecords; ++i) {
            if (pruneFlags.testBit(i)) continue;
            if (sortInfos[i - 1].path.compare(sortInfos[i].path, Qt::CaseInsensitive)) continue;
            pruneFlags.setBit(sortInfos[i].origIndex, true);
            numFlagged += 1;
        }
        delete[] sortInfos;
        return numFlagged;
    }
    int pruneFlaggedLumps(QBitArray flaggedLumps) {
        DENG2_ASSERT(flaggedLumps.size() == lumps.size());
        int const numFlaggedForPrune = flaggedLumps.count(true);
        if (numFlaggedForPrune) {
            lumpsByPath.reset();
            int numRecords = lumps.size();
            if (numRecords == numFlaggedForPrune) {
                lumps.clear();
            }
            else {
                for (int i = 0, newIdx = 0; i < numRecords; ++i) {
                    if (!flaggedLumps.testBit(i)) {
                        ++newIdx;
                        continue;
                    }
                    lumps.move(newIdx, lumps.size() - 1);
                }
                int firstPruned = lumps.size() - numFlaggedForPrune;
                lumps.erase(lumps.begin() + firstPruned, lumps.end());
            }
        }
        return numFlaggedForPrune;
    }
    void pruneDuplicatesIfNeeded() {
        if (!needPruneDuplicateLumps) return;
        needPruneDuplicateLumps = false;
        int const numRecords = lumps.size();
        if (numRecords <= 1) return;
        QBitArray pruneFlags(numRecords);
        flagDuplicateLumps(pruneFlags);
        pruneFlaggedLumps(pruneFlags);
    }
};
LumpIndex::LumpIndex(bool pathsAreUnique) : d(new Impl(this)) {
    d->pathsAreUnique = pathsAreUnique;
}
LumpIndex::~LumpIndex()
{}
bool LumpIndex::hasLump(lumpnum_t lumpNum) const {
    d->pruneDuplicatesIfNeeded();
    return (lumpNum >= 0 && lumpNum < d->lumps.size());
}
static String LumpIndex_invalidIndexMessage(int invalidIdx, int lastValidIdx) {
    String msg = String("Invalid lump index %1").arg(invalidIdx);
    if (lastValidIdx < 0) msg += " (file is empty)";
    else                 msg += String(", valid range: [0..%2)").arg(lastValidIdx);
    return msg;
}
File1 &LumpIndex::lump(lumpnum_t lumpNum) const {
    if (!hasLump(lumpNum)) throw NotFoundError("LumpIndex::lump", LumpIndex_invalidIndexMessage(lumpNum, size() - 1));
    return *d->lumps[lumpNum];
}
LumpIndex::Lumps const &LumpIndex::allLumps() const {
    d->pruneDuplicatesIfNeeded();
    return d->lumps;
}
int LumpIndex::size() const {
    d->pruneDuplicatesIfNeeded();
    return d->lumps.size();
}
int LumpIndex::lastIndex() const {
    return d->lumps.size() - 1;
}
int LumpIndex::pruneByFile(File1 &file) {
    if (d->lumps.empty()) return 0;
    int const numRecords = d->lumps.size();
    QBitArray pruneFlags(numRecords);
    d->flagDuplicateLumps(pruneFlags);
    int numFlaggedForFile = d->flagContainedLumps(pruneFlags, file);
    d->pruneFlaggedLumps(pruneFlags);
    d->needPruneDuplicateLumps = false;
    return numFlaggedForFile;
}
bool LumpIndex::pruneLump(File1 &lump) {
    if (d->lumps.empty()) return 0;
    d->pruneDuplicatesIfNeeded();
    if (!d->lumps.removeOne(&lump)) return false;
    d->lumpsByPath.reset();
    return true;
}
void LumpIndex::catalogLump(File1 &lump) {
    d->lumps.push_back(&lump);
    d->lumpsByPath.reset();    // We'll need to rebuild the path hash chains.
    if (d->pathsAreUnique) {
        d->needPruneDuplicateLumps = true;
    }
}
void LumpIndex::clear() {
    d->lumps.clear();
    d->lumpsByPath.reset();
    d->needPruneDuplicateLumps = false;
}
bool LumpIndex::catalogues(File1 &file) {
    d->pruneDuplicatesIfNeeded();
    DENG2_FOR_EACH(Lumps, i, d->lumps) {
        File1 const &lump = **i;
        if (&lump.container() == &file)
            return true;
    }
    return false;
}
bool LumpIndex::contains(Path const &path) const {
    return findFirst(path) >= 0;
}
int LumpIndex::findAll(Path const &path, FoundIndices &found) const {
    LOG_AS("LumpIndex::findAll");
    found.clear();
    if (path.isEmpty() || d->lumps.empty()) return 0;
    d->pruneDuplicatesIfNeeded();
    d->buildLumpsByPathIfNeeded();
    DENG2_ASSERT(!d->lumpsByPath.isNull());
    ushort hash = path.lastSegment().hash() % d->lumpsByPath->size();
    for (int idx = (*d->lumpsByPath)[hash].head; idx != -1;
        idx = (*d->lumpsByPath)[idx].nextInLoadOrder) {
        File1 const &lump          = *d->lumps[idx];
        PathTree::Node const &node = lump.directoryNode();
        if (!node.comparePath(path, 0)) {
            found.push_front(idx);
        }
    }
    return int(found.size());
}
lumpnum_t LumpIndex::findLast(Path const &path) const {
    if (path.isEmpty() || d->lumps.empty()) return -1;
    d->pruneDuplicatesIfNeeded();
    d->buildLumpsByPathIfNeeded();
    DENG2_ASSERT(!d->lumpsByPath.isNull());
    ushort hash = path.lastSegment().hash() % d->lumpsByPath->size();
    for (int idx = (*d->lumpsByPath)[hash].head; idx != -1;
        idx = (*d->lumpsByPath)[idx].nextInLoadOrder) {
        File1 const &lump          = *d->lumps[idx];
        PathTree::Node const &node = lump.directoryNode();
        if (!node.comparePath(path, 0)) {
            return idx; // This is the lump we are looking for.
        }
    }
    return -1; // Not found.
}
lumpnum_t LumpIndex::findFirst(Path const &path) const {
    if (path.isEmpty() || d->lumps.empty()) return -1;
    d->pruneDuplicatesIfNeeded();
    d->buildLumpsByPathIfNeeded();
    lumpnum_t earliest = -1; // Not found.
    DENG2_ASSERT(!d->lumpsByPath.isNull());
    ushort hash = path.lastSegment().hash() % d->lumpsByPath->size();
    for (int idx = (*d->lumpsByPath)[hash].head; idx != -1;
        idx = (*d->lumpsByPath)[idx].nextInLoadOrder) {
        File1 const &lump          = *d->lumps[idx];
        PathTree::Node const &node = lump.directoryNode();
        if (!node.comparePath(path, 0)) {
            earliest = idx; // This is now the first lump loaded.
        }
    }
    return earliest;
}
Uri LumpIndex::composeResourceUrn(lumpnum_t lumpNum) // static
{
    return Uri("LumpIndex", Path(String("%1").arg(lumpNum)));
}
} // namespace de
static size_t FileReader(char const* name, char** buffer);
size_t M_ReadFile(const char* name, char** buffer) {
    return FileReader(name, buffer);
}
AutoStr *M_ReadFileIntoString(ddstring_t const *path, dd_bool *isCustom) {
    if (isCustom) *isCustom = false;
    if (Str_StartsWith(path, "LumpIndex:")) {
        bool isNumber;
        lumpnum_t const lumpNum    = String(Str_Text(path) + 10).toInt(&isNumber);
        LumpIndex const &lumpIndex = App_FileSystem().nameIndex();
        if (isNumber && lumpIndex.hasLump(lumpNum)) {
            File1 &lump = lumpIndex.lump(lumpNum);
            if (isCustom) {
                *isCustom = (lump.isContained()? lump.container().hasCustom() : lump.hasCustom());
            }
            if (!lump.size()) return 0;
            AutoStr *string = Str_PartAppend(AutoStr_NewStd(), (char const *)lump.cache(), 0, lump.size());
            lump.unlock();
            if (Str_IsEmpty(string))
                return 0;
            return string;
        }
        return 0;
    }
    if (Str_StartsWith(path, "Lumps:")) {
        char const *lumpName       = Str_Text(path) + 6;
        LumpIndex const &lumpIndex = App_FileSystem().nameIndex();
        if (!lumpIndex.contains(String(lumpName) + ".lmp"))
            return 0;
        File1 &lump = lumpIndex[lumpIndex.findLast(String(lumpName) + ".lmp")];
        if (isCustom) {
            *isCustom = (lump.isContained()? lump.container().hasCustom() : lump.hasCustom());
        }
        if (!lump.size()) return 0;
        AutoStr *string = Str_PartAppend(AutoStr_NewStd(), (char const *)lump.cache(), 0, lump.size());
        lump.unlock();
        if (Str_IsEmpty(string))
            return 0;
        return string;
    }
    try {
        QScopedPointer<FileHandle> hndl(&App_FileSystem().openFile(Str_Text(path), "rb"));
        if (isCustom) {
            File1 &file = hndl->file();
            *isCustom = (file.isContained()? file.container().hasCustom() : file.hasCustom());
        }
        AutoStr *string = nullptr;
        if (size_t lumpLength = hndl->length()) {
            Block buffer;
            buffer.resize(lumpLength);
            hndl->read((uint8_t *)buffer.data(), lumpLength);
            string = Str_PartAppend(AutoStr_NewStd(), buffer.constData(), 0, lumpLength);
        }
        App_FileSystem().releaseFile(hndl->file());
        if (!string || Str_IsEmpty(string))
            return 0;
        return string;
    }
    catch (FS1::NotFoundError const &)
    {} // Ignore this error.
    char *readBuf = 0;
    if (size_t bytesRead = M_ReadFile(Str_Text(path), &readBuf)) {
        AutoStr *string = Str_PartAppend(AutoStr_New(), readBuf, 0, int(bytesRead));
        Z_Free(readBuf);
        if (Str_IsEmpty(string))
            return 0;
        return string;
    }
    return 0;
}
#if defined(WIN32)
#  define close _close
#  define read  _read
#  define write _write
#endif
static size_t FileReader(const char* name, char** buffer) {
    struct stat fileinfo;
    char* buf = NULL;
    size_t length = 0;
    int handle;
    LOG_AS("FileReader");
    LZFILE *file = lzOpen((char*) name, "rp");
    if (NULL != file) {
#define BSIZE 1024
        char readBuf[BSIZE];
        while (!lzEOF(file)) {
            size_t bytesRead = lzRead(readBuf, BSIZE, file);
            char* newBuf;
            newBuf = (char*) Z_Malloc(length + bytesRead, PU_APPSTATIC, 0);
            if (buf != NULL) {
                memcpy(newBuf, buf, length);
                Z_Free(buf);
            }
            buf = newBuf;
            memcpy(buf + length, readBuf, bytesRead);
            length += bytesRead;
        }
        lzClose(file);
        *buffer = (char*)buf;
        return length;
#undef BSIZE
    }
    handle = open(name, O_RDONLY | O_BINARY, 0666);
    if (handle == -1) {
        LOG_RES_WARNING("Failed opening \"%s\" for reading") << name;
        return length;
    }
    if (-1 == fstat(handle, &fileinfo)) {
        LOG_RES_ERROR("Couldn't read file \"%s\"") << name;
        return 0;
    }
    length = fileinfo.st_size;
    if (!length) {
        *buffer = 0;
        return 0;
    }
    buf = (char *) Z_Malloc(length, PU_APPSTATIC, 0);
    DENG_ASSERT(buf != 0);
    size_t bytesRead = read(handle, buf, length);
    close(handle);
    if (bytesRead < length) {
        LOG_RES_ERROR("Couldn't read file \"%s\"") << name;
    }
    *buffer = buf;
    return length;
}
#if defined(WIN32)
#  undef close
#  undef read
#  undef write
#  undef open
#endif
namespace de {
SearchPath::SearchPath(de::Uri const &_uri, SearchPath::Flags _flags)
    : Uri(_uri), flags_(_flags)
{}
SearchPath::SearchPath(SearchPath const &other)
    : Uri(other), flags_(other.flags_)
{}
SearchPath::Flags SearchPath::flags() const {
    return flags_;
}
SearchPath &SearchPath::setFlags(Flags newFlags) {
    flags_ = newFlags;
    return *this;
}
} // namespace de
void FS_InitVirtualPathMappings() {
    App_FileSystem().clearPathMappings();
    if (DoomsdayApp::app().isShuttingDown()) return;
    dint argC = CommandLine_Count();
    for (dint i = 0; i < argC; ++i) {
        if (qstrnicmp("-vdmap", CommandLine_At(i), 6)) {
            continue;
        }
        if (i < argC - 1 && !CommandLine_IsOption(i + 1) && !CommandLine_IsOption(i + 2)) {
            String source      = NativePath(CommandLine_PathAt(i + 1)).expand().withSeparators('/');
            String destination = NativePath(CommandLine_PathAt(i + 2)).expand().withSeparators('/');
            App_FileSystem().addPathMapping(source, destination);
            i += 2;
        }
    }
}
static inline char const *skipSpace(char const *ptr) {
    DENG2_ASSERT(ptr != 0);
    while (*ptr && *ptr != '\n' && isspace(*ptr))
    { ptr++; }
    return ptr;
}
static bool parsePathLumpMapping(char lumpName[9/*LUMPNAME_T_MAXLEN*/], ddstring_t *path, char const *buffer)
{
    DENG2_ASSERT(lumpName != 0 && path != 0);
    char const *ptr = skipSpace(buffer);
    if (!*ptr || *ptr == '\n') return false;
    char const *end = (char const *)M_FindWhite((char *)ptr);
    if (!*end || *end == '\n') return false;
    size_t len = end - ptr;
    if (len > 8) return false;
    memset(lumpName, 0, 9/*LUMPNAME_T_MAXLEN*/);
    strncpy(lumpName, ptr, len);
    strupr(lumpName);
    ptr = skipSpace(end);
    if (!*ptr || *ptr == '\n') return false; // Missing file path.
    Str_Set(path, ptr);
    Str_StripRight(path);
    F_FixSlashes(path, path);
    return true;
}
/**
 * <pre> LUMPNAM0 \\Path\\In\\The\\Base.ext
 * LUMPNAM1 Path\\In\\The\\RuntimeDir.ext
 *  :</pre>
 */
static bool parsePathLumpMappings(char const *buffer) {
    DENG2_ASSERT(buffer != 0);
    bool successful = false;
    ddstring_t path; Str_Init(&path);
    ddstring_t line; Str_Init(&line);
    char const *ch = buffer;
    char lumpName[9/*LUMPNAME_T_MAXLEN*/];
    do {
        ch = Str_GetLine(&line, ch);
        if (!parsePathLumpMapping(lumpName, &path, Str_Text(&line))) {
        }
        else {
            String destination = NativePath(Str_Text(&path)).expand().withSeparators('/');
            App_FileSystem().addPathLumpMapping(lumpName, destination);
        }
    } while (*ch);
    successful = true;
    Str_Free(&line);
    Str_Free(&path);
    return successful;
}
void FS_InitPathLumpMappings() {
    App_FileSystem().clearPathLumpMappings();
    if (DoomsdayApp::app().isShuttingDown()) return;
    size_t bufSize = 0;
    uint8_t *buf = 0;
    LumpIndex const &lumpIndex = App_FileSystem().nameIndex();
    LumpIndex::FoundIndices foundDirecs;
    lumpIndex.findAll("DD_DIREC.lmp", foundDirecs);
    DENG2_FOR_EACH_CONST(LumpIndex::FoundIndices, i, foundDirecs) // in load order
    {
        File1 &lump = lumpIndex[*i];
        FileInfo const &lumpInfo = lump.info();
        if (!buf || bufSize < lumpInfo.size + 1) {
            bufSize = lumpInfo.size + 1;
            buf = (uint8_t *) M_Realloc(buf, bufSize);
        }
        lump.read(buf, 0, lumpInfo.size);
        buf[lumpInfo.size] = 0;
        parsePathLumpMappings(reinterpret_cast<char const *>(buf));
    }
    M_Free(buf);
}
namespace de {
namespace internal {
struct FileHeader {
    DENG2_ERROR(ReadError);
    Block identification; // 4 bytes
    dint32 lumpRecordsCount;
    dint32 lumpRecordsOffset;
    void operator << (FileHandle &from) {
        uint8_t buf[12];
        dsize readBytes = from.read(buf, 12);
        if (readBytes != 12) throw ReadError("FileHeader::operator << (FileHandle &)", "Source file is truncated");
        identification    = Block(buf, 4);
        lumpRecordsCount  = littleEndianByteOrder.toHost(*reinterpret_cast<dint32 *>(buf + 4));
        lumpRecordsOffset = littleEndianByteOrder.toHost(*reinterpret_cast<dint32 *>(buf + 8));
    }
};
struct IndexEntry {
    DENG2_ERROR(ReadError);
    dint32 offset;
    dint32 size;
    Block name; // 8 bytes
    void operator << (FileHandle &from) {
        uint8_t buf[16];
        dsize readBytes = from.read(buf, 16);
        if (readBytes != 16) throw ReadError("IndexEntry::operator << (FileHandle &)", "Source file is truncated");
        name   = Block(buf + 8, 8);
        dint32 const *off = reinterpret_cast<dint32 const *>(buf);
        offset = littleEndianByteOrder.toHost(*off);
        size   = littleEndianByteOrder.toHost(*reinterpret_cast<dint32 const *>(buf + 4));
    }
    String nameNormalized() const {
        String normName;
        int nameLen = 0;
        while (nameLen < 8 && name[nameLen]) { nameLen++; }
        for (int i = 0; i < nameLen; ++i) {
            char ch = name[i] & 0x7f;
            normName += ch;
        }
        if (!normName.isEmpty()) {
            normName = QString(normName.toLatin1().toPercentEncoding());
        }
        else {
            normName = "________";
        }
        if (normName.fileNameExtension().isEmpty()) {
            normName += !normName.compareWithoutCase("DEHACKED")? ".deh" : ".lmp";
        }
        return normName;
    }
};
static QString Wad_invalidIndexMessage(int invalidIdx, int lastValidIdx) {
    QString msg = QString("Invalid lump index %1 ").arg(invalidIdx);
    if (lastValidIdx < 0) msg += "(file is empty)";
    else                 msg += QString("(valid range: [0..%2])").arg(lastValidIdx);
    return msg;
}
} // namespace internal
using namespace internal;
Wad::LumpFile::LumpFile(Entry &entry, FileHandle *hndl, String path,
    FileInfo const &info, File1 *container)
    : File1(hndl, path, info, container)
    , entry(entry)
{}
String const &Wad::LumpFile::name() const {
    return directoryNode().name();
}
Uri Wad::LumpFile::composeUri(QChar delimiter) const {
    return directoryNode().path(delimiter);
}
PathTree::Node &Wad::LumpFile::directoryNode() const {
    return entry;
}
size_t Wad::LumpFile::read(uint8_t *buffer, bool tryCache) {
    return wad().readLump(info_.lumpIdx, buffer, tryCache);
}
size_t Wad::LumpFile::read(uint8_t *buffer, size_t startOffset, size_t length, bool tryCache) {
    return wad().readLump(info_.lumpIdx, buffer, startOffset, length, tryCache);
}
uint8_t const *Wad::LumpFile::cache() {
    return wad().cacheLump(info_.lumpIdx);
}
Wad::LumpFile &Wad::LumpFile::unlock() {
    wad().unlockLump(info_.lumpIdx);
    return *this;
}
Wad &Wad::LumpFile::wad() const {
    return container().as<Wad>();
}
DENG2_PIMPL_NOREF(Wad) {
    LumpTree entries;                     ///< Directory structure and entry records for all lumps.
    QScopedPointer<LumpCache> dataCache;  ///< Data payload cache.
    Impl() : entries(PathTree::MultiLeaf) {}
};
Wad::Wad(FileHandle &hndl, String path, FileInfo const &info, File1 *container)
    : File1(&hndl, path, info, container)
    , LumpIndex()
    , d(new Impl) {
    LOG_AS("Wad");
    handle_->seek(0, SeekSet);
    FileHeader hdr;
    hdr << *handle_;
    if (hdr.lumpRecordsCount <= 0) return;
    handle_->seek(hdr.lumpRecordsOffset, SeekSet);
    for (int i = 0; i < hdr.lumpRecordsCount; ++i) {
        IndexEntry lump;
        lump << *handle_;
        String absPath = String::fromStdString(DoomsdayApp::app().doomsdayBasePath()) / lump.nameNormalized();
        Entry &entry = d->entries.insert(absPath);
        entry.offset = lump.offset;
        entry.size   = lump.size;
        LumpFile *lumpFile = new LumpFile(entry, nullptr, entry.path(),
                                          FileInfo(lastModified(), // Inherited from the container (note recursion).
                                                   i, entry.offset, entry.size, entry.size),
                                          this);
        entry.lumpFile.reset(lumpFile); // takes ownership
        catalogLump(*lumpFile);
    }
}
Wad::~Wad()
{}
void Wad::clearCachedLump(int lumpIndex, bool *retCleared) {
    LOG_AS("Wad::clearCachedLump");
    if (retCleared) *retCleared = false;
    if (hasLump(lumpIndex)) {
        if (!d->dataCache.isNull()) {
            d->dataCache->remove(lumpIndex, retCleared);
        }
    }
    else {
        LOGDEV_RES_WARNING(Wad_invalidIndexMessage(lumpIndex, lastIndex()));
    }
}
void Wad::clearLumpCache() {
    LOG_AS("Wad::clearLumpCache");
    if (!d->dataCache.isNull()) {
        d->dataCache->clear();
    }
}
uint8_t const *Wad::cacheLump(int lumpIndex) {
    LOG_AS("Wad::cacheLump");
    LumpFile const &lumpFile = static_cast<LumpFile &>(lump(lumpIndex));
    LOGDEV_RES_XVERBOSE("\"%s:%s\" (%u bytes%s)",
               NativePath(composePath()).pretty()
            << NativePath(lumpFile.composePath()).pretty()
            << (unsigned long) lumpFile.info().size
            << (lumpFile.info().isCompressed()? ", compressed" : ""));
    if (d->dataCache.isNull()) {
        d->dataCache.reset(new LumpCache(LumpIndex::size()));
    }
    uint8_t const *data = d->dataCache->data(lumpIndex);
    if (data) return data;
    uint8_t *region = (uint8_t *) Z_Malloc(lumpFile.info().size, PU_APPSTATIC, 0);
    if (!region) throw Error("Wad::cacheLump", QString("Failed on allocation of %1 bytes for cache copy of lump #%2").arg(lumpFile.info().size).arg(lumpIndex));
    readLump(lumpIndex, region, false);
    d->dataCache->insert(lumpIndex, region);
    return region;
}
void Wad::unlockLump(int lumpIndex) {
    LOG_AS("Wad::unlockLump");
    LOGDEV_RES_XVERBOSE("\"%s:%s\"",
                        NativePath(composePath()).pretty() <<
                        NativePath(lump(lumpIndex).composePath()).pretty());
    if (hasLump(lumpIndex)) {
        if (!d->dataCache.isNull()) {
            d->dataCache->unlock(lumpIndex);
        }
    }
    else {
        LOGDEV_RES_WARNING(Wad_invalidIndexMessage(lumpIndex, lastIndex()));
    }
}
size_t Wad::readLump(int lumpIndex, uint8_t *buffer, bool tryCache) {
    LOG_AS("Wad::readLump");
    return readLump(lumpIndex, buffer, 0, lump(lumpIndex).size(), tryCache);
}
size_t Wad::readLump(int lumpIndex, uint8_t *buffer, size_t startOffset,
    size_t length, bool tryCache) {
    LOG_AS("Wad::readLump");
    LumpFile const &lumpFile = static_cast<LumpFile &>(lump(lumpIndex));
    LOGDEV_RES_XVERBOSE("\"%s:%s\" (%u bytes%s) [%u +%u]",
                        NativePath(composePath()).pretty()
                        << NativePath(lumpFile.composePath()).pretty()
                        << (unsigned long) lumpFile.size()
                        << (lumpFile.isCompressed()? ", compressed" : "")
                        << startOffset
                        << length);
    if (tryCache) {
        uint8_t const *data = (!d->dataCache.isNull() ? d->dataCache->data(lumpIndex) : 0);
        LOGDEV_RES_XVERBOSE("Cache %s on #%i", (data? "hit" : "miss") << lumpIndex);
        if (data) {
            size_t readBytes = de::min(size_t(lumpFile.size()), length);
            std::memcpy(buffer, data + startOffset, readBytes);
            return readBytes;
        }
    }
    handle_->seek(lumpFile.info().baseOffset + startOffset, SeekSet);
    size_t readBytes = handle_->read(buffer, length);
    if (readBytes < length)
        throw Error("Wad::readLumpSection", QString("Only read %1 of %2 bytes of lump #%3").arg(readBytes).arg(length).arg(lumpIndex));
    return readBytes;
}
uint Wad::calculateCRC() {
    uint crc = 0;
    foreach (File1 *file, allLumps()) {
        Entry &entry = static_cast<Entry &>(file->as<LumpFile>().directoryNode());
        entry.update();
        crc += entry.crc;
    }
    return crc;
}
bool Wad::recognise(FileHandle &file) {
    size_t initPos = file.tell();
    file.seek(0, SeekSet);
    bool readOk = false;
    FileHeader hdr;
    try {
        hdr << file;
        readOk = true;
    }
    catch (FileHeader::ReadError const &)
    {} // Ignore
    file.seek(initPos, SeekSet);
    if (!readOk) return false;
    return (hdr.identification == "IWAD" || hdr.identification == "PWAD");
}
Wad::LumpTree const &Wad::lumpTree() const {
    return d->entries;
}
Wad::LumpFile &Wad::Entry::file() const {
    DENG2_ASSERT(!lumpFile.isNull());
    return *lumpFile;
}
void Wad::Entry::update() {
    crc = uint(file().size());
    String const lumpName = Node::name();
    int const nameLen = lumpName.length();
    for (int i = 0; i < nameLen; ++i) {
        crc += lumpName.at(i).unicode();
    }
}
} // namespace de
namespace de {
namespace internal {
#define SIG_LOCAL_FILE_HEADER   0x04034b50
#define SIG_CENTRAL_FILE_HEADER 0x02014b50
#define SIG_END_OF_CENTRAL_DIR  0x06054b50
#define MAXIMUM_COMMENT_SIZE    2048
#define CENTRAL_END_SIZE        22
#define ZFH_ENCRYPTED           0x1
#define ZFH_COMPRESSION_OPTS    0x6
#define ZFH_DESCRIPTOR          0x8
#define ZFH_COMPRESS_PATCHED    0x20 ///< Not supported.
enum {
    ZFC_NO_COMPRESSION = 0,     ///< Supported format.
    ZFC_SHRUNK,
    ZFC_REDUCED_1,
    ZFC_REDUCED_2,
    ZFC_REDUCED_3,
    ZFC_REDUCED_4,
    ZFC_IMPLODED,
    ZFC_DEFLATED = 8,           ///< The only supported compression (via zlib).
    ZFC_DEFLATED_64,
    ZFC_PKWARE_DCL_IMPLODED
};
#pragma pack(1)
typedef struct localfileheader_s {
    uint32_t      signature;
    uint16_t      requiredVersion;
    uint16_t      flags;
    uint16_t      compression;
    uint16_t      lastModTime;
    uint16_t      lastModDate;
    uint32_t      crc32;
    uint32_t      compressedSize;
    uint32_t      size;
    uint16_t      fileNameSize;
    uint16_t      extraFieldSize;
} localfileheader_t;
typedef struct descriptor_s {
    uint32_t      crc32;
    uint32_t      compressedSize;
    uint32_t      size;
} descriptor_t;
typedef struct centralfileheader_s {
    uint32_t      signature;
    uint16_t      version;
    uint16_t      requiredVersion;
    uint16_t      flags;
    uint16_t      compression;
    uint16_t      lastModTime;
    uint16_t      lastModDate;
    uint32_t      crc32;
    uint32_t      compressedSize;
    uint32_t      size;
    uint16_t      fileNameSize;
    uint16_t      extraFieldSize;
    uint16_t      commentSize;
    uint16_t      diskStart;
    uint16_t      internalAttrib;
    uint32_t      externalAttrib;
    uint32_t      relOffset;
    /*
     * file name (variable size)
     * extra field (variable size)
     * file comment (variable size)
     */
} centralfileheader_t;
static bool readArchiveHeader(FileHandle &file, localfileheader_t &hdr) {
    size_t readBytes, initPos = file.tell();
    file.seek(0, SeekSet);
    readBytes = file.read((uint8_t *)&hdr, sizeof(localfileheader_t));
    file.seek(initPos, SeekSet);
    if (!(readBytes < sizeof(localfileheader_t))) {
        hdr.signature       = littleEndianByteOrder.toHost(hdr.signature);
        hdr.requiredVersion = littleEndianByteOrder.toHost(hdr.requiredVersion);
        hdr.flags           = littleEndianByteOrder.toHost(hdr.flags);
        hdr.compression     = littleEndianByteOrder.toHost(hdr.compression);
        hdr.lastModTime     = littleEndianByteOrder.toHost(hdr.lastModTime);
        hdr.lastModDate     = littleEndianByteOrder.toHost(hdr.lastModDate);
        hdr.crc32           = littleEndianByteOrder.toHost(hdr.crc32);
        hdr.compressedSize  = littleEndianByteOrder.toHost(hdr.compressedSize);
        hdr.size            = littleEndianByteOrder.toHost(hdr.size);
        hdr.fileNameSize    = littleEndianByteOrder.toHost(hdr.fileNameSize);
        hdr.extraFieldSize  = littleEndianByteOrder.toHost(hdr.extraFieldSize);
        return true;
    }
    return false;
}
typedef struct centralend_s {
    uint16_t      disk;
    uint16_t      centralStartDisk;
    uint16_t      diskEntryCount;
    uint16_t      totalEntryCount;
    uint32_t      size;
    uint32_t      offset;
    uint16_t      commentSize;
} centralend_t;
#pragma pack()
static bool readCentralEnd(FileHandle &file, centralend_t &end) {
    size_t readBytes = file.read((uint8_t *)&end, sizeof(centralend_t));
    if (!(readBytes < sizeof(centralend_t))) {
        end.disk            = littleEndianByteOrder.toHost(end.disk);
        end.centralStartDisk= littleEndianByteOrder.toHost(end.centralStartDisk);
        end.diskEntryCount  = littleEndianByteOrder.toHost(end.diskEntryCount);
        end.totalEntryCount = littleEndianByteOrder.toHost(end.totalEntryCount);
        end.size            = littleEndianByteOrder.toHost(end.size);
        end.offset          = littleEndianByteOrder.toHost(end.offset);
        end.commentSize     = littleEndianByteOrder.toHost(end.commentSize);
        return true;
    }
    return false;
}
static String Zip_invalidIndexMessage(int invalidIdx, int lastValidIdx) {
    String msg = String("Invalid lump index %1").arg(invalidIdx);
    if (lastValidIdx < 0) msg += " (file is empty)";
    else                 msg += String(", valid range: [0..%2)").arg(lastValidIdx);
    return msg;
}
/**
 * The path inside the zip might be mapped to another virtual location.
 *
 * @return  @c true= iff @a path was mapped to another location.
 *
 * @todo This is clearly implemented in the wrong place. Path mapping
 *       should be done at a higher level.
 */
static bool applyGamePathMappings(String &path) {
    if (path.beginsWith('@')) {
        path.remove(0, 1);
        if (path.at(0) == '/') path.remove(0, 1);
        path = String("$(App.DefsPath)/$(GamePlugin.Name)/auto") / path;
        return true;
    }
    if (path.beginsWith('#')) {
        path.remove(0, 1);
        if (path.at(0) == '/') path.remove(0, 1);
        if (int slash = path.lastIndexOf('/')) {
            if (slash >= 2) {
                if (path.at(slash - 2) == '.' && path.at(slash - 1) >= '1' && path.at(slash - 1) <= '9')
                    path.remove(slash - 2, 2);
            }
        }
        path = String("$(App.DataPath)/$(GamePlugin.Name)/auto") / path;
        return true;
    }
    if (!path.contains('/')) {
        FileType const &ftype = DD_GuessFileTypeFromFileName(path.fileName());
        switch (ftype.defaultClass()) {
        case RC_PACKAGE:
            path = String("$(App.DataPath)/$(GamePlugin.Name)/auto") / path;
            return true;
        case RC_DEFINITION:
            path = String("$(App.DefsPath)/$(GamePlugin.Name)/auto") / path;
            return true;
        default:
            return false;
        }
    }
    FS1::Schemes const &schemes = App_FileSystem().allSchemes();
    DENG2_FOR_EACH_CONST(FS1::Schemes, i, schemes) {
        if ((*i)->mapPath(path)) {
            return true;
        }
    }
    return false;
}
} // namespace internal
using namespace internal;
Zip::LumpFile::LumpFile(Entry &entry, FileHandle *hndl, String path,
    FileInfo const &info, File1 *container)
    : File1(hndl, path, info, container)
    , entry(entry)
{}
String const &Zip::LumpFile::name() const {
    return directoryNode().name();
}
Uri Zip::LumpFile::composeUri(QChar delimiter) const {
    return directoryNode().path(delimiter);
}
PathTree::Node &Zip::LumpFile::directoryNode() const {
    return entry;
}
size_t Zip::LumpFile::read(uint8_t *buffer, bool tryCache) {
    return zip().readLump(info_.lumpIdx, buffer, tryCache);
}
size_t Zip::LumpFile::read(uint8_t *buffer, size_t startOffset, size_t length, bool tryCache) {
    return zip().readLump(info_.lumpIdx, buffer, startOffset, length, tryCache);
}
uint8_t const *Zip::LumpFile::cache() {
    return zip().cacheLump(info_.lumpIdx);
}
Zip::LumpFile &Zip::LumpFile::unlock() {
    zip().unlockLump(info_.lumpIdx);
    return *this;
}
Zip &Zip::LumpFile::zip() const {
    return container().as<Zip>();
}
DENG2_PIMPL(Zip) {
    LumpTree entries;                     ///< Directory structure and entry records for all lumps.
    QScopedPointer<LumpCache> dataCache;  ///< Data payload cache.
    Impl(Public *i) : Base(i)
    {}
    /**
     * @param lump      Lump/file to be buffered.
     * @param buffer    Must be large enough to hold the entire uncompressed data lump.
     */
    size_t bufferLump(LumpFile const &lump, uint8_t *buffer) {
        DENG2_ASSERT(buffer);
        LOG_AS("Zip");
        FileInfo const &lumpInfo = lump.info();
        self().handle_->seek(lumpInfo.baseOffset, SeekSet);
        if (lumpInfo.isCompressed()) {
            bool result;
            uint8_t *compressedData = (uint8_t *) M_Malloc(lumpInfo.compressedSize);
            if (!compressedData) throw Error("Zip::bufferLump", QString("Failed on allocation of %1 bytes for decompression buffer").arg(lumpInfo.compressedSize));
            self().handle_->read(compressedData, lumpInfo.compressedSize);
            result = uncompressRaw(compressedData, lumpInfo.compressedSize, buffer, lumpInfo.size);
            M_Free(compressedData);
            if (!result) return 0; // Inflate failed.
        }
        else {
            self().handle_->read(buffer, lumpInfo.size);
        }
        return lumpInfo.size;
    }
};
Zip::Zip(FileHandle &hndl, String path, FileInfo const &info, File1 *container)
    : File1(&hndl, path, info, container)
    , LumpIndex(true/*paths are unique*/)
    , d(new Impl(this)) {
    bool foundCentralDirectory = false; {
        int pos = CENTRAL_END_SIZE; // Offset from the end.
        uint32_t signature;
        while (!foundCentralDirectory && pos < MAXIMUM_COMMENT_SIZE) {
            handle_->seek(-pos, SeekEnd);
            handle_->read((uint8_t *)&signature, sizeof(signature));
            if (littleEndianByteOrder.toHost(signature) == SIG_END_OF_CENTRAL_DIR) {
                foundCentralDirectory = true; // Yes, this is it.
            }
            else {
                pos++;
            }
        }
    }
    if (!foundCentralDirectory)
        throw FormatError("Zip", "Central directory in \"" + NativePath(composePath()).pretty() + "\" not found");
    centralend_t summary;
    readCentralEnd(*handle_, summary);
    if (summary.diskEntryCount != summary.totalEntryCount)
        throw FormatError("Zip", "Multipart zip file \"" + NativePath(composePath()).pretty() + "\" not supported");
    void *centralDirectory = M_Malloc(summary.size);
    if (!centralDirectory) throw FormatError("Zip", String("Failed on allocation of %1 bytes for temporary copy of the central centralDirectory").arg(summary.size));
    handle_->seek(summary.offset, SeekSet);
    handle_->read((uint8_t *)centralDirectory, summary.size);
    /**
     * Pass 1: Validate support and count the number of lump records we need.
     * Pass 2: Read all zip entries and populate the lump directory.
     */
    char *pos;
    int entryCount = 0;
    for (int pass = 0; pass < 2; ++pass) {
        if (pass == 1) {
            if (entryCount == 0) break;
        }
        pos = (char *)centralDirectory;
        uint lumpIdx = 0;
        for (int index = 0; index < summary.totalEntryCount; ++index, pos += sizeof(centralfileheader_t)) {
            centralfileheader_t const *header = (centralfileheader_t *) pos;
            char const *nameStart = pos + sizeof(centralfileheader_t);
            localfileheader_t localHeader;
            pos += DD_USHORT(header->fileNameSize) + DD_USHORT(header->extraFieldSize) + DD_USHORT(header->commentSize);
            String filePath = NativePath(nameStart, DD_USHORT(header->fileNameSize)).withSeparators('/');
            if (DD_ULONG(header->size) == 0 && filePath.last() == '/') continue;
            if (DD_USHORT(header->compression) != ZFC_NO_COMPRESSION &&
               DD_USHORT(header->compression) != ZFC_DEFLATED) {
                if (pass != 0) continue;
                LOG_RES_WARNING("Zip %s:'%s' uses an unsupported compression algorithm")
                        << NativePath(composePath()).pretty() << NativePath(filePath).pretty();
            }
            if (DD_USHORT(header->flags) & ZFH_ENCRYPTED) {
                if (pass != 0) continue;
                LOG_RES_WARNING("Zip %s:'%s' is encrypted; encryption is not supported")
                        << NativePath(composePath()).pretty() << NativePath(filePath).pretty();
            }
            if (pass == 0) {
                ++entryCount;
                continue;
            }
            handle_->seek(DD_ULONG(header->relOffset), SeekSet);
            handle_->read((uint8_t *)&localHeader, sizeof(localHeader));
            size_t baseOffset = DD_ULONG(header->relOffset) + sizeof(localfileheader_t)
                              + DD_USHORT(header->fileNameSize) + DD_USHORT(localHeader.extraFieldSize);
            size_t compressedSize;
            if (DD_USHORT(header->compression) == ZFC_DEFLATED) {
                compressedSize = DD_ULONG(header->compressedSize);
            }
            else // No compression.
            {
                compressedSize = DD_ULONG(header->size);
            }
            if (!DoomsdayApp::game().isNull()) {
                String filePathCopy = filePath;
                if (applyGamePathMappings(filePathCopy)) {
                    try {
                        filePath = Uri(filePathCopy, RC_NULL).resolved();
                    }
                    catch (de::Uri::ResolveError const& er) {
                        LOG_RES_WARNING(er.asText());
                    }
                }
            }
            filePath = App_BasePath() / filePath;
            Entry &entry = d->entries.insert(Path(filePath));
            entry.offset         = baseOffset;
            entry.size           = DD_ULONG(header->size);
            entry.compressedSize = compressedSize;
            LumpFile *lumpFile = new LumpFile(entry, nullptr, entry.path(),
                                              FileInfo(lastModified(), // Inherited from the file (note recursion).
                                              lumpIdx, entry.offset, entry.size, entry.compressedSize),
                                              this);
            entry.lumpFile.reset(lumpFile); // takes ownership
            catalogLump(*lumpFile);
            lumpIdx++;
        }
    }
    M_Free(centralDirectory);
}
Zip::~Zip()
{}
void Zip::clearCachedLump(int lumpIndex, bool *retCleared) {
    LOG_AS("Zip::clearCachedLump");
    if (retCleared) *retCleared = false;
    if (hasLump(lumpIndex)) {
        if (!d->dataCache.isNull()) {
            d->dataCache->remove(lumpIndex, retCleared);
        }
    }
    else {
        LOGDEV_RES_WARNING(Zip_invalidIndexMessage(lumpIndex, lastIndex()));
    }
}
void Zip::clearLumpCache() {
    LOG_AS("Zip::clearLumpCache");
    if (!d->dataCache.isNull()) {
        d->dataCache->clear();
    }
}
uint8_t const *Zip::cacheLump(int lumpIndex) {
    LOG_AS("Zip::cacheLump");
    LumpFile &lumpFile = static_cast<LumpFile &>(lump(lumpIndex));
    LOGDEV_RES_XVERBOSE("\"%s:%s\" (%u bytes%s)",
                        NativePath(composePath()).pretty()
                        << NativePath(lumpFile.composePath()).pretty()
                        << (unsigned long) lumpFile.info().size
                        << (lumpFile.info().isCompressed()? ", compressed" : ""));
    if (d->dataCache.isNull()) {
        d->dataCache.reset(new LumpCache(lumpCount()));
    }
    uint8_t const *data = d->dataCache->data(lumpIndex);
    if (data) return data;
    uint8_t *region = (uint8_t *) Z_Malloc(lumpFile.info().size, PU_APPSTATIC, 0);
    if (!region) throw Error("Zip::cacheLump", QString("Failed on allocation of %1 bytes for cache copy of lump #%2").arg(lumpFile.info().size).arg(lumpIndex));
    readLump(lumpIndex, region, false);
    d->dataCache->insert(lumpIndex, region);
    return region;
}
void Zip::unlockLump(int lumpIndex) {
    LOG_AS("Zip::unlockLump");
    LOGDEV_RES_XVERBOSE("\"%s:%s\"", NativePath(composePath()).pretty()
                        << NativePath(lump(lumpIndex).composePath()).pretty());
    if (hasLump(lumpIndex)) {
        if (!d->dataCache.isNull()) {
            d->dataCache->unlock(lumpIndex);
        }
    }
    else {
        LOGDEV_RES_WARNING(Zip_invalidIndexMessage(lumpIndex, lastIndex()));
    }
}
size_t Zip::readLump(int lumpIndex, uint8_t *buffer, bool tryCache) {
    LOG_AS("Zip::readLump");
    return readLump(lumpIndex, buffer, 0, lump(lumpIndex).size(), tryCache);
}
size_t Zip::readLump(int lumpIndex, uint8_t *buffer, size_t startOffset,
    size_t length, bool tryCache) {
    LOG_AS("Zip::readLump");
    LumpFile const &lumpFile = static_cast<LumpFile &>(lump(lumpIndex));
    LOGDEV_RES_XVERBOSE("\"%s:%s\" (%u bytes%s) [%u +%u]",
               NativePath(composePath()).pretty()
            << NativePath(lumpFile.composePath()).pretty()
            << (unsigned long) lumpFile.size()
            << (lumpFile.isCompressed()? ", compressed" : "")
            << startOffset
            << length);
    if (tryCache) {
        uint8_t const *data = (!d->dataCache.isNull() ? d->dataCache->data(lumpIndex) : 0);
        LOGDEV_RES_XVERBOSE("Cache %s on #%i", (data? "hit" : "miss") << lumpIndex);
        if (data) {
            size_t readBytes = de::min(size_t(lumpFile.size()), length);
            std::memcpy(buffer, data + startOffset, readBytes);
            return readBytes;
        }
    }
    size_t readBytes = 0;
    if (!startOffset && length == lumpFile.size()) {
        readBytes = d->bufferLump(lumpFile, buffer);
    }
    else {
        uint8_t *readBuf = (uint8_t *) M_Malloc(lumpFile.size());
        if (!readBuf) throw Error("Zip::readLump", QString("Failed on allocation of %1 bytes for work buffer").arg(lumpFile.size()));
        if (d->bufferLump(lumpFile, readBuf)) {
            readBytes = de::min(size_t(lumpFile.size()), length);
            std::memcpy(buffer, readBuf + startOffset, readBytes);
        }
        M_Free(readBuf);
    }
    if (readBytes < de::min(size_t(lumpFile.size()), length))
        throw Error("Zip::readLump", QString("Only read %1 of %2 bytes of lump #%3").arg(readBytes).arg(length).arg(lumpIndex));
    return readBytes;
}
bool Zip::recognise(FileHandle &file) {
    localfileheader_t hdr;
    if (!readArchiveHeader(file, hdr)) return false;
    return hdr.signature == SIG_LOCAL_FILE_HEADER;
}
uint8_t *Zip::compress(uint8_t *in, size_t inSize, size_t *outSize) {
    return compressAtLevel(in, inSize, outSize, Z_DEFAULT_COMPRESSION);
}
uint8_t *Zip::compressAtLevel(uint8_t *in, size_t inSize, size_t *outSize, int level) {
#define CHUNK_SIZE 32768
    LOG_AS("Zip::compressAtLevel");
    z_stream stream;
    uint8_t chunk[CHUNK_SIZE];
    size_t allocSize = CHUNK_SIZE;
    uint8_t *output  = (uint8_t *) M_Malloc(allocSize); // some initial space
    int result;
    int have;
    DENG2_ASSERT(outSize);
    *outSize = 0;
    std::memset(&stream, 0, sizeof(stream));
    stream.next_in  = (Bytef*) in;
    stream.avail_in = (uInt) inSize;
    stream.zalloc   = Z_NULL;
    stream.zfree    = Z_NULL;
    stream.opaque   = Z_NULL;
    if (level < Z_NO_COMPRESSION) {
        level = Z_NO_COMPRESSION;
    }
    if (level > Z_BEST_COMPRESSION) {
        level = Z_BEST_COMPRESSION;
    }
    result = deflateInit(&stream, level);
    if (result != Z_OK) {
        M_Free(output);
        return 0;
    }
    do {
        stream.next_out  = chunk;
        stream.avail_out = CHUNK_SIZE;
        result = deflate(&stream, Z_FINISH);
        if (result == Z_STREAM_ERROR) {
            M_Free(output);
            *outSize = 0;
            return 0;
        }
        have = CHUNK_SIZE - stream.avail_out;
        if (have) {
            if (*outSize + have > allocSize) {
                allocSize *= 2;
                output = (uint8_t *) M_Realloc(output, allocSize);
            }
            std::memcpy(output + *outSize, chunk, have);
            *outSize += have;
        }
    } while (!stream.avail_out); // output chunk full, more data may follow
    DENG2_ASSERT(result == Z_STREAM_END);
    DENG2_ASSERT(stream.total_out == *outSize);
    deflateEnd(&stream);
    return output;
#undef CHUNK_SIZE
}
uint8_t *Zip::uncompress(uint8_t *in, size_t inSize, size_t *outSize) {
#define INF_CHUNK_SIZE 4096 // Uncompress in 4KB chunks.
    LOG_AS("Zip::uncompress");
    z_stream stream;
    uint8_t chunk[INF_CHUNK_SIZE];
    size_t allocSize = INF_CHUNK_SIZE;
    uint8_t *output  = (uint8_t *) M_Malloc(allocSize); // some initial space
    int result;
    int have;
    DENG2_ASSERT(outSize);
    *outSize = 0;
    std::memset(&stream, 0, sizeof(stream));
    stream.next_in  = (Bytef *) in;
    stream.avail_in = (uInt) inSize;
    result = inflateInit(&stream);
    if (result != Z_OK) {
        M_Free(output);
        return 0;
    }
    do {
        stream.next_out  = chunk;
        stream.avail_out = INF_CHUNK_SIZE;
        result = inflate(&stream, Z_FINISH);
        if (result == Z_STREAM_ERROR) {
            M_Free(output);
            *outSize = 0;
            return 0;
        }
        have = INF_CHUNK_SIZE - stream.avail_out;
        if (have) {
            if (*outSize + have > allocSize) {
                allocSize *= 2;
                output = (uint8_t *) M_Realloc(output, allocSize);
            }
            std::memcpy(output + *outSize, chunk, have);
            *outSize += have;
        }
    } while (!stream.avail_out); // output chunk full, more data may follow
    DENG2_ASSERT(result == Z_STREAM_END);
    inflateEnd(&stream);
    return output;
#undef INF_CHUNK_SIZE
}
bool Zip::uncompressRaw(uint8_t *in, size_t inSize, uint8_t *out, size_t outSize) {
    LOG_AS("Zip::uncompressRaw");
    z_stream stream;
    int result;
    std::memset(&stream, 0, sizeof(stream));
    stream.next_in   = (Bytef *) in;
    stream.avail_in  = (uInt) inSize;
    stream.zalloc    = Z_NULL;
    stream.zfree     = Z_NULL;
    stream.next_out  = (Bytef *) out;
    stream.avail_out = (uInt) outSize;
    if (inflateInit2(&stream, -MAX_WBITS) != Z_OK)
        return false;
    result = inflate(&stream, Z_FINISH);
    if (stream.total_out != outSize) {
        inflateEnd(&stream);
        LOG_RES_WARNING("Failure due to %s (result code: %i)")
                << (result == Z_DATA_ERROR ? "corrupt data" : "zlib error")
                << result;
        return false;
    }
    inflateEnd(&stream);
    return true;
}
Zip::LumpTree const &Zip::lumpTree() const {
    return d->entries;
}
Zip::LumpFile &Zip::Entry::file() const {
    DENG2_ASSERT(!lumpFile.isNull());
    return *lumpFile;
}
} // namespace de
