﻿/** @file world/actions.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/world/actions.h"
#include "doomsday/gameapi.h"
#include "doomsday/doomsdayapp.h"
#include "doomsday/players.h"
#include "doomsday/world/mobjthinkerdata.h"
#include "doomsday/defs/ded.h"
#include <de/DictionaryValue>
#include <de/NativePointerValue>
#include <de/Process>
#include <de/String>
#include <de/TextValue>
#include <QMap>
/** @file bspnode.cpp  Map BSP node.
 *
 * @authors Copyright © 2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/bspnode.h"
/** @file materialdetaillayer.cpp  Logical material, detail-texture layer.
 *
 * @authors Copyright © 2011-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/detailtexturemateriallayer.h"
#include <doomsday/res/TextureScheme>
#include <doomsday/res/Textures>
/** @file dmuargs.cpp Doomsday Map Update (DMU) API arguments.
 *
 * @authors Copyright © 2006-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/mapelement.h"
#include "dd_share.h"
#include <de/Log>
#include <de/c_wrapper.h>
/** @file entitydatabase.cpp World map entity property value database.
 * @ingroup world
 *
 * @authors Copyright &copy; 2007-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/entitydatabase.h"
#include <de/Log>
#include <map>
/** @file entitydef.cpp  World playsim data structures.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is M_Free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/entitydef.h"
#include "doomsday/world/world.h"
#include "doomsday/world/map.h"
#include "doomsday/world/propertyvalue.h"
#include "doomsday/EntityDatabase"
#include <cmath>
#include <map>
#include <de/memory.h>
#include <de/Error>
#include <de/Log>
#include <de/String>
#include <de/StringPool>
/** @file world.cpp  Base for world maps.
 *
 * @authors Copyright � 2014-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/map.h"
#include "doomsday/EntityDatabase"
/** @file mapelement.cpp  Base class for all map elements.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/mapelement.h"
#include "doomsday/world/map.h"
/** @file material.cpp  World material.
 *
 * @authors Copyright © 2009-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/material.h"
#include "doomsday/console/cmd.h"
#include "doomsday/res/Textures"
#include "doomsday/resource/resources.h"
#include "doomsday/world/materials.h"
#include "doomsday/world/materialmanifest.h"
#include "doomsday/world/texturemateriallayer.h"
#include "doomsday/world/detailtexturemateriallayer.h"
#include "doomsday/world/shinetexturemateriallayer.h"
#include <QFlag>
#include <QtAlgorithms>
#include <de/Log>
/** @file materialarchive.cpp  Material Archive.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/materialarchive.h"
#include "doomsday/world/materials.h"
#include "doomsday/world/MaterialManifest"
#include "doomsday/resource/resources.h"
#include "doomsday/uri.h"
#include <de/StringPool>
#include <de/reader.h>
#include <de/writer.h>
#define MATERIALARCHIVE_VERSION     4
#define ASEG_MATERIAL_ARCHIVE       112
#define UNKNOWN_MATERIALNAME        "DD_BADTX"
/** @file materialmanifest.cpp  Description of a material.
 *
 * @authors Copyright © 2011-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/materialmanifest.h"
#include "doomsday/world/MaterialScheme"
#include "doomsday/world/materials.h"
/** @file materials.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/world/materials.h"
#include "doomsday/world/MaterialScheme"
#include "doomsday/world/MaterialManifest"
#include "doomsday/world/world.h"
#include "doomsday/resource/resources.h"
#include <de/memory.h>
#include <unordered_set>
/** @file materialscheme.cpp  Material system subspace scheme.
 *
 * @authors Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/materialscheme.h"
#include "doomsday/world/MaterialManifest"
/** @file mobjthinkerdata.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/mobjthinkerdata.h"
#include <de/ScriptSystem>
#include <de/RecordValue>
/** @file propertyvalue.cpp  Data types for representing property values.
 * @ingroup data
 *
 * @authors Copyright &copy; 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/propertyvalue.h"
#include <de/Error>
/** @file materialshinelayer.cpp  Logical material, shine/reflection layer.
 *
 * @authors Copyright © 2011-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/shinetexturemateriallayer.h"
#include "doomsday/res/TextureScheme"
#include "doomsday/res/Textures"
/** @file texturemateriallayer.cpp   Logical material, texture layer.
 *
 * @authors Copyright © 2011-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/texturemateriallayer.h"
#include "doomsday/defs/material.h"
/** @file thinker.cpp  Base for all thinkers.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/thinker.h"
#include <de/math.h>
#include <de/memory.h>
#include <de/memoryzone.h>
/** @file thinkerdata.cpp  Base class for thinker private data.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/world/thinkerdata.h"
#include "doomsday/world/map.h"
#include <QMultiHash>
/** @file world.cpp  World base class.
 *
 * @authors Copyright © 2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/world/world.h"
#include "doomsday/world/materials.h"
#include "doomsday/world/map.h"
#include "doomsday/DoomsdayApp"
#include "doomsday/players.h"
#include "api_player.h"
#include <de/App>
#include <de/Context>
/** @file world/xg.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/world/xg.h"
#include "doomsday/defs/dedparser.h"
#include "doomsday/doomsdayapp.h"
using namespace de;
using namespace world;
static QMap<String, acfnptr_t> s_actions; ///< name => native function pointer.
static String s_currentAction;
static void C_DECL A_DoomsdayScript(void *actor) {
    struct mobj_s *mobj = reinterpret_cast<struct mobj_s *>(actor);
    int plrNum = -1;
    auto &plrs = DoomsdayApp::players(); {
        for (int i = 0; i < DDMAXPLAYERS; ++i) {
            if (&plrs.at(i).publicData() ==
                *reinterpret_cast<const ddplayer_t * const *>(actor)) {
                mobj = plrs.at(i).publicData().mo;
                plrNum = i;
            }
        }
    }
    LOG_AS("A_DoomsdayScript");
    try {
        const ThinkerData &data = THINKER_DATA(mobj->thinker, ThinkerData);
        Record ns;
        if (plrNum >= 0) {
            ns.add(new Variable(QStringLiteral("player"),
                                new RecordValue(plrs.at(plrNum).objectNamespace())));
        }
        ns.add(new Variable(QStringLiteral("self"), new RecordValue(data.objectNamespace())));
        Process proc(&ns);
        const Script script(s_currentAction);
        proc.run(script);
        proc.execute();
    }
    catch (const Error &er) {
        LOG_SCR_ERROR(er.asText());
    }
}
static bool isScriptAction(const String &name) {
    return !name.beginsWith("A_");
}
void P_GetGameActions() {
    s_actions.clear();
    if (auto getVar = DoomsdayApp::plugins().gameExports().GetPointer) {
        auto const *links = (actionlink_t const *) getVar(DD_ACTION_LINK);
        for (actionlink_t const *link = links; link && link->name; link++) {
            s_actions.insert(String(link->name).toLower(), link->func);
        }
    }
}
void P_SetCurrentAction(const String &name) {
    s_currentAction = name;
}
void P_SetCurrentActionState(int state) {
    P_SetCurrentAction(DED_Definitions()->states[state].gets(QStringLiteral("action")));
}
acfnptr_t P_GetAction(const String &name) {
    if (!name.isEmpty()) {
        if (isScriptAction(name)) {
            return de::function_cast<acfnptr_t>(A_DoomsdayScript);
        }
        auto found = s_actions.find(name.toLower());
        if (found != s_actions.end()) return found.value();
    }
    return nullptr;  // Not found.
}
acfnptr_t P_GetAction(const char *name) {
    return P_GetAction(String(name));
}
namespace world {
BspNode::BspNode(Partition const &partition) : Partition(partition)
{}
}  // namespace world
namespace world {
static de::Uri findTextureForDetailStage(ded_detail_stage_t const &def) {
    try {
        if (def.texture) {
            return res::Textures::get().textureScheme("Details")
                       .findByResourceUri(*def.texture)
                           .composeUri();
        }
    }
    catch (res::TextureScheme::NotFoundError const &)
    {} // Ignore this error.
    return de::Uri();
}
DetailTextureMaterialLayer::AnimationStage::AnimationStage(de::Uri const &texture, int tics,
    float variance, float scale, float strength, float maxDistance)
    : TextureMaterialLayer::AnimationStage(texture, tics, variance)
    , scale(scale)
    , strength(strength)
    , maxDistance(maxDistance) {
}
DetailTextureMaterialLayer::AnimationStage::AnimationStage(AnimationStage const &other)
    : TextureMaterialLayer::AnimationStage(other)
{}
DetailTextureMaterialLayer::AnimationStage::~AnimationStage()
{}
void DetailTextureMaterialLayer::AnimationStage::resetToDefaults() {
    TextureMaterialLayer::AnimationStage::resetToDefaults();
    scale = 1;
    strength = 1;
    maxDistance = 0;
}
DetailTextureMaterialLayer::AnimationStage *
DetailTextureMaterialLayer::AnimationStage::fromDef(ded_detail_stage_t const &def) {
    de::Uri const texture = findTextureForDetailStage(def);
    return new AnimationStage(texture, def.tics, def.variance,
                              def.scale, def.strength, def.maxDistance);
}
DetailTextureMaterialLayer *DetailTextureMaterialLayer::fromDef(ded_detailtexture_t const &layerDef) {
    auto *layer = new DetailTextureMaterialLayer();
    layer->_stages.append(AnimationStage::fromDef(layerDef.stage));
    return layer;
}
int DetailTextureMaterialLayer::addStage(DetailTextureMaterialLayer::AnimationStage const &stageToCopy) {
    _stages.append(new AnimationStage(stageToCopy));
    return _stages.count() - 1;
}
String DetailTextureMaterialLayer::describe() const {
    return "Detail layer";
}
} // namespace world
namespace world {
static DmuArgs::PointerToIndexFunc ptrToIndexFunc;
/**
 * Convert propertyType enum constant into a string for error/debug messages.
 */
static char const *value_Str(dint val) {
    static char valStr[40];
    struct val_s {
        dint val;
        char const *str;
    } valuetypes[] = {
        { DDVT_BOOL,        "DDVT_BOOL" },
        { DDVT_BYTE,        "DDVT_BYTE" },
        { DDVT_SHORT,       "DDVT_SHORT" },
        { DDVT_INT,         "DDVT_INT" },
        { DDVT_UINT,        "DDVT_UINT" },
        { DDVT_FIXED,       "DDVT_FIXED" },
        { DDVT_ANGLE,       "DDVT_ANGLE" },
        { DDVT_FLOAT,       "DDVT_FLOAT" },
        { DDVT_DOUBLE,      "DDVT_DOUBLE" },
        { DDVT_LONG,        "DDVT_LONG" },
        { DDVT_ULONG,       "DDVT_ULONG" },
        { DDVT_PTR,         "DDVT_PTR" },
        { DDVT_BLENDMODE,   "DDVT_BLENDMODE" },
        { 0, nullptr }
    };
    for (duint i = 0; valuetypes[i].str; ++i) {
        if (valuetypes[i].val == val)
            return valuetypes[i].str;
    }
    sprintf(valStr, "(unnamed %i)", val);
    return valStr;
}
DmuArgs::DmuArgs(int type, uint prop)
    : type          (type)
    , prop          (prop & ~DMU_FLAG_MASK)
    , modifiers     (prop & DMU_FLAG_MASK)
    , valueType     (DDVT_NONE)
    , booleanValues (0)
    , byteValues    (0)
    , intValues     (0)
    , fixedValues   (0)
    , floatValues   (0)
    , doubleValues  (0)
    , angleValues   (0)
    , ptrValues     (0) {
    DENG_ASSERT(VALID_DMU_ELEMENT_TYPE_ID(type));
}
void DmuArgs::value(valuetype_t dstValueType, void *dst, uint index) const {
#define DMUARGS_LOG_ERROR(retval) { *d = retval; LOG_AS("DmuArgs::value"); \
    LOGDEV_MAP_ERROR("%s incompatible with value type %s") \
            << value_Str(dstValueType) << value_Str(valueType); }
    if (dstValueType == DDVT_FIXED) {
        fixed_t *d = (fixed_t *)dst;
        switch (valueType) {
        case DDVT_BYTE:
            *d = (byteValues[index] << FRACBITS);
            break;
        case DDVT_INT:
            *d = (intValues[index] << FRACBITS);
            break;
        case DDVT_FIXED:
            *d = fixedValues[index];
            break;
        case DDVT_FLOAT:
            *d = FLT2FIX(floatValues[index]);
            break;
        case DDVT_DOUBLE:
            *d = FLT2FIX(doubleValues[index]);
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_FLOAT) {
        float *d = (float *)dst;
        switch (valueType) {
        case DDVT_BYTE:
            *d = byteValues[index];
            break;
        case DDVT_INT:
            *d = intValues[index];
            break;
        case DDVT_FIXED:
            *d = FIX2FLT(fixedValues[index]);
            break;
        case DDVT_FLOAT:
            *d = floatValues[index];
            break;
        case DDVT_DOUBLE:
            *d = (float)doubleValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_DOUBLE) {
        double *d = (double *)dst;
        switch (valueType) {
        case DDVT_BYTE:
            *d = byteValues[index];
            break;
        case DDVT_INT:
            *d = intValues[index];
            break;
        case DDVT_FIXED:
            *d = FIX2FLT(fixedValues[index]);
            break;
        case DDVT_FLOAT:
            *d = floatValues[index];
            break;
        case DDVT_DOUBLE:
            *d = doubleValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_BOOL) {
        dd_bool *d = (dd_bool *)dst;
        switch (valueType) {
        case DDVT_BOOL:
            *d = booleanValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(false);
            break;
        }
    }
    else if (dstValueType == DDVT_BYTE) {
        byte *d = (byte *)dst;
        switch (valueType) {
        case DDVT_BOOL:
            *d = booleanValues[index];
            break;
        case DDVT_BYTE:
            *d = byteValues[index];
            break;
        case DDVT_INT:
            *d = intValues[index];
            break;
        case DDVT_FLOAT:
            *d = (byte) floatValues[index];
            break;
        case DDVT_DOUBLE:
            *d = (byte) doubleValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_INT) {
        int *d = (int *)dst;
        switch (valueType) {
        case DDVT_BOOL:
            *d = booleanValues[index];
            break;
        case DDVT_BYTE:
            *d = byteValues[index];
            break;
        case DDVT_INT:
            *d = intValues[index];
            break;
        case DDVT_FLOAT:
            *d = floatValues[index];
            break;
        case DDVT_DOUBLE:
            *d = doubleValues[index];
            break;
        case DDVT_FIXED:
            *d = (fixedValues[index] >> FRACBITS);
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_SHORT) {
        short *d = (short *)dst;
        switch (valueType) {
        case DDVT_BOOL:
            *d = booleanValues[index];
            break;
        case DDVT_BYTE:
            *d = byteValues[index];
            break;
        case DDVT_INT:
            *d = intValues[index];
            break;
        case DDVT_FLOAT:
            *d = floatValues[index];
            break;
        case DDVT_DOUBLE:
            *d = doubleValues[index];
            break;
        case DDVT_FIXED:
            *d = (fixedValues[index] >> FRACBITS);
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_ANGLE) {
        angle_t *d = (angle_t *)dst;
        switch (valueType) {
        case DDVT_ANGLE:
            *d = angleValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(0);
            break;
        }
    }
    else if (dstValueType == DDVT_BLENDMODE) {
        blendmode_t *d = (blendmode_t *)dst;
        switch (valueType) {
        case DDVT_INT:
            if (intValues[index] >= 0 && intValues[index] < DDNUM_BLENDMODES) {
                *d = blendmode_t(intValues[index]);
            }
            else {
                LOG_AS("DmuArgs::setValue");
                LOGDEV_MAP_ERROR("%i is not a valid value for DDVT_BLENDMODE") << intValues[index];
            }
            break;
        default:
            DMUARGS_LOG_ERROR(BM_NORMAL);
            break;
        }
    }
    else if (dstValueType == DDVT_PTR) {
        void **d = (void **)dst;
        switch (valueType) {
        case DDVT_PTR:
            *d = ptrValues[index];
            break;
        default:
            DMUARGS_LOG_ERROR(nullptr);
            break;
        }
    }
    else {
        LOG_AS("DmuArgs::value");
        LOGDEV_MAP_ERROR("Unknown value type %i") << dstValueType;
    }
#undef DMUARGS_LOG_ERROR
}
void DmuArgs::setValue(valuetype_t srcValueType, void const *src, uint index) {
#define DMUARGS_LOG_ERROR() { LOG_AS("DmuArgs::setValue"); \
    LOGDEV_MAP_ERROR("%s incompatible with value type %s") \
            << value_Str(srcValueType) << value_Str(valueType); }
    if (srcValueType == DDVT_FIXED) {
        fixed_t const *s = (fixed_t const *)src;
        switch (valueType) {
        case DDVT_BYTE:
            byteValues[index] = (*s >> FRACBITS);
            break;
        case DDVT_INT:
            intValues[index] = (*s >> FRACBITS);
            break;
        case DDVT_FIXED:
            fixedValues[index] = *s;
            break;
        case DDVT_FLOAT:
            floatValues[index] = FIX2FLT(*s);
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = FIX2FLT(*s);
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_FLOAT) {
        float const *s = (float const *)src;
        switch (valueType) {
        case DDVT_BYTE:
            byteValues[index] = *s;
            break;
        case DDVT_INT:
            intValues[index] = (int) *s;
            break;
        case DDVT_FIXED:
            fixedValues[index] = FLT2FIX(*s);
            break;
        case DDVT_FLOAT:
            floatValues[index] = *s;
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = (double)*s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_DOUBLE) {
        double const *s = (double const *)src;
        switch (valueType) {
        case DDVT_BYTE:
            byteValues[index] = (byte)*s;
            break;
        case DDVT_INT:
            intValues[index] = (int) *s;
            break;
        case DDVT_FIXED:
            fixedValues[index] = FLT2FIX(*s);
            break;
        case DDVT_FLOAT:
            floatValues[index] = (float)*s;
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_BOOL) {
        dd_bool const *s = (dd_bool const *)src;
        switch (valueType) {
        case DDVT_BOOL:
            booleanValues[index] = *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_BYTE) {
        byte const *s = (byte const *)src;
        switch (valueType) {
        case DDVT_BOOL:
            booleanValues[index] = *s;
            break;
        case DDVT_BYTE:
            byteValues[index] = *s;
            break;
        case DDVT_INT:
            intValues[index] = *s;
            break;
        case DDVT_FLOAT:
            floatValues[index] = *s;
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_INT) {
        int const *s = (int const *)src;
        switch (valueType) {
        case DDVT_BOOL:
            booleanValues[index] = *s;
            break;
        case DDVT_BYTE:
            byteValues[index] = *s;
            break;
        case DDVT_INT:
            intValues[index] = *s;
            break;
        case DDVT_FLOAT:
            floatValues[index] = *s;
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = *s;
            break;
        case DDVT_FIXED:
            fixedValues[index] = (*s << FRACBITS);
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_SHORT) {
        short const *s = (short const *)src;
        switch (valueType) {
        case DDVT_BOOL:
            booleanValues[index] = *s;
            break;
        case DDVT_BYTE:
            byteValues[index] = *s;
            break;
        case DDVT_INT:
            intValues[index] = *s;
            break;
        case DDVT_FLOAT:
            floatValues[index] = *s;
            break;
        case DDVT_DOUBLE:
            doubleValues[index] = *s;
            break;
        case DDVT_FIXED:
            fixedValues[index] = (*s << FRACBITS);
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_ANGLE) {
        angle_t const *s = (angle_t const *)src;
        switch (valueType) {
        case DDVT_ANGLE:
            angleValues[index] = *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_BLENDMODE) {
        blendmode_t const *s = (blendmode_t const *)src;
        switch (valueType) {
        case DDVT_INT:
            intValues[index] = *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else if (srcValueType == DDVT_PTR) {
        void const *const *s = (void const *const *)src;
        switch (valueType) {
        case DDVT_INT:
            DENG_ASSERT(ptrToIndexFunc);
            intValues[index] = ptrToIndexFunc(*s);
            break;
        case DDVT_PTR:
            ptrValues[index] = (void *) *s;
            break;
        default:
            DMUARGS_LOG_ERROR();
            break;
        }
    }
    else {
        LOG_AS("DmuArgs::setValue");
        LOGDEV_MAP_ERROR("Unknown value type %i") << srcValueType;
    }
#undef DMUARGS_LOG_ERROR
}
void DmuArgs::setPointerToIndexFunc(PointerToIndexFunc func) {
    ptrToIndexFunc = func;
}
} // namespace world
typedef std::map<int, PropertyValue *> Entity;
typedef std::map<int, Entity> Entities;
typedef std::map<int, Entities> EntitySet;
DENG2_PIMPL(EntityDatabase) {
    EntitySet entitySets;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        DENG2_FOR_EACH(EntitySet, setIt, entitySets)
        DENG2_FOR_EACH(Entities, entityIt, setIt->second)
        DENG2_FOR_EACH(Entity, propIt, entityIt->second) {
            delete propIt->second;
        }
    }
    /**
     * Lookup the set in which entities with the unique identifier
     * @a entityId are stored.
     */
    Entities *entities(int entityId) {
        std::pair<EntitySet::iterator, bool> result;
        result = entitySets.insert(std::pair<int, Entities>(entityId, Entities()));
        return &result.first->second;
    }
    /**
     * Lookup an entity in @a set by its unique @a elementIndex.
     */
    Entity *entityByElementIndex(Entities& set, int elementIndex, bool canCreate) {
        Entities::iterator found = set.find(elementIndex);
        if (found != set.end()) return &found->second;
        if (!canCreate) return 0;
        std::pair<Entities::iterator, bool> result;
        result = set.insert(std::pair<int, Entity>(elementIndex, Entity()));
        return &result.first->second;
    }
    PropertyValue const *tryFindPropertyValue(MapEntityPropertyDef const &def,
                                              int elementIndex) {
         Entities *set = entities(def.entity->id);
         Entity const *entity = entityByElementIndex(*set, elementIndex, false /*do not create*/);
         if (!entity) {
             throw Error("EntityDatabase::property", QString("There is no element %1 of type %2")
                                                         .arg(elementIndex)
                                                         .arg(Str_Text(P_NameForMapEntityDef(def.entity))));
         }
         Entity::const_iterator found = entity->find(def.id);
         if (found == entity->end()) {
             return nullptr;
         }
         return found->second;
    }
};
EntityDatabase::EntityDatabase() : d(new Impl(this))
{}
uint EntityDatabase::entityCount(MapEntityDef const *entityDef) const {
    DENG2_ASSERT(entityDef);
    Entities *set = d->entities(entityDef->id);
    return set->size();
}
bool EntityDatabase::hasEntity(MapEntityDef const *entityDef, int elementIndex) const {
    DENG2_ASSERT(entityDef);
    Entities *set = d->entities(entityDef->id);
    return d->entityByElementIndex(*set, elementIndex, false /*do not create*/) != 0;
}
PropertyValue const &EntityDatabase::property(MapEntityPropertyDef const *def,
                                              int elementIndex) const {
    DENG2_ASSERT(def);
    if (auto const *propValue = d->tryFindPropertyValue(*def, elementIndex)) {
        return *propValue;
    }
    throw Error("EntityDatabase::property",
                QString("Element %1 of type %2 has no value for property %3")
                .arg(elementIndex)
                .arg(Str_Text(P_NameForMapEntityDef(def->entity)))
                .arg(def->id));
}
bool EntityDatabase::hasPropertyValue(MapEntityPropertyDef const *def, int elementIndex) const {
    DENG2_ASSERT(def);
    return d->tryFindPropertyValue(*def, elementIndex) != nullptr;
}
void EntityDatabase::setProperty(MapEntityPropertyDef const *def, int elementIndex,
    PropertyValue *value) {
    DENG2_ASSERT(def);
    Entities *set = d->entities(def->entity->id);
    Entity *entity = d->entityByElementIndex(*set, elementIndex, true);
    if (!entity)
        throw Error("EntityDatabase::setProperty", "Failed adding new entity element record");
    Entity::iterator found = entity->find(def->id);
    if (found != entity->end()) {
        PropertyValue **adr = &found->second;
        delete *adr;
        *adr = value;
        return;
    }
    entity->insert(std::pair<int, PropertyValue *>(def->id, value));
}
static StringPool *entityDefs;
typedef std::map<int, StringPool::Id> EntityDefIdMap;
static EntityDefIdMap entityDefIdMap;
static void clearEntityDefs() {
    if (!::entityDefs) return;
    ::entityDefs->forAll([] (StringPool::Id id) {
        auto *def = static_cast<MapEntityDef *>( ::entityDefs->userPointer(id) );
        DENG2_ASSERT(def);
        for (duint i = 0; i < def->numProps; ++i) {
            M_Free(def->props[i].name);
        }
        M_Free(def->props);
        delete def;
        return LoopContinue;
    });
    delete ::entityDefs; ::entityDefs = nullptr;
    ::entityDefIdMap.clear();
}
MapEntityDef *P_MapEntityDef(int id) {
    EntityDefIdMap::iterator i = entityDefIdMap.find(id);
    if (i != entityDefIdMap.end()) {
        StringPool::Id id = i->second;
        return static_cast<MapEntityDef *>( entityDefs->userPointer(id) );
    }
    return nullptr;  // Not found.
}
MapEntityDef *P_MapEntityDefByName(char const *name) {
    if (name && entityDefs) {
        StringPool::Id id = entityDefs->isInterned(String(name));
        return static_cast<MapEntityDef *>( entityDefs->userPointer(id) );
    }
    return nullptr;  // Not found.
}
AutoStr *P_NameForMapEntityDef(MapEntityDef const *def) {
    String name;  // Not found.
    if (def) {
        ::entityDefs->forAll([&def, &name] (StringPool::Id id) {
            if (::entityDefs->userPointer(id) == def) {
                name = ::entityDefs->string(id);
                return LoopAbort;
            }
            return LoopContinue;
        });
    }
    return AutoStr_FromText(name.toUtf8().constData());
}
int MapEntityDef_Property(MapEntityDef *def, int propertyId,
                          MapEntityPropertyDef **retDef = 0) {
    DENG2_ASSERT(def);
    MapEntityPropertyDef *found = 0;
    for (uint i = 0; i < def->numProps; ++i) {
        MapEntityPropertyDef *prop = def->props + i;
        if (prop->id == propertyId) {
            found = prop;
            break;
        }
    }
    if (retDef) *retDef = found;
    return found? found - def->props : -1/* not found */;
}
int MapEntityDef_PropertyByName(MapEntityDef *def, char const *propertyName,
                                MapEntityPropertyDef **retDef) {
    DENG2_ASSERT(def);
    MapEntityPropertyDef *found = 0;
    if (propertyName && propertyName[0]) {
        for (uint i = 0; i < def->numProps; ++i) {
            MapEntityPropertyDef *prop = def->props + i;
            if (!qstricmp(prop->name, propertyName)) {
                found = prop;
                break;
            }
        }
    }
    if (retDef) *retDef = found;
    return found? found - def->props : -1/* not found */;
}
void MapEntityDef_AddProperty(MapEntityDef* def, int propertyId, const char* propertyName,
                              valuetype_t type) {
    DENG2_ASSERT(def);
    if (propertyId == 0) // Not a valid identifier.
        throw Error("MapEntityDef_AddProperty", "0 is not a valid propertyId");
    if (!propertyName || !propertyName[0]) // Must have a name.
        throw Error("MapEntityDef_AddProperty", "Invalid propertyName (zero-length string)");
    switch (type) {
    case DDVT_BYTE:
    case DDVT_SHORT:
    case DDVT_INT:
    case DDVT_FIXED:
    case DDVT_ANGLE:
    case DDVT_FLOAT:
    case DDVT_DOUBLE:
        break;
    default:
        throw Error("MapEntityDef_AddProperty", QString("Unknown/not supported value type %1").arg(type));
    }
    if (MapEntityDef_Property(def, propertyId) >= 0)
        throw Error("MapEntityDef_AddProperty", QString("propertyId %1 not unique for %2")
                                                    .arg(propertyId).arg(Str_Text(P_NameForMapEntityDef(def))));
    if (MapEntityDef_PropertyByName(def, propertyName) >= 0)
        throw Error("MapEntityDef_AddProperty", QString("propertyName \"%1\" not unique for %2")
                                                    .arg(propertyName).arg(Str_Text(P_NameForMapEntityDef(def))));
    def->props = (MapEntityPropertyDef*) M_Realloc(def->props, ++def->numProps * sizeof(*def->props));
    if (!def->props)
        throw Error("MapEntityDef_AddProperty",
                        QString("Failed on (re)allocation of %1 bytes for new MapEntityPropertyDef array")
                            .arg((unsigned long) sizeof(*def->props)));
    MapEntityPropertyDef* prop = &def->props[def->numProps - 1];
    prop->id = propertyId;
    int len = (int)strlen(propertyName);
    prop->name = (char *) M_Malloc(sizeof(*prop->name) * (len + 1));
    if (!prop->name)
        throw Error("MapEntityDef_AddProperty",
                        QString("Failed on allocation of %1 bytes for property name")
                            .arg((unsigned long) (sizeof(*prop->name) * (len + 1))));
    strncpy(prop->name, propertyName, len);
    prop->name[len] = '\0';
    prop->type = type;
    prop->entity = def;
}
/**
 * Look up a mapobj definition.
 *
 * @param identifer   If objName is @c NULL, compare using this unique identifier.
 * @param entityName  If not @c NULL, compare using this unique name.
 * @param canCreate   @c true= create a new definition if not found.
 */
static MapEntityDef *findMapEntityDef(int identifier, char const *entityName,
                                      bool canCreate) {
    if (identifier == 0 && (!entityName || !entityName[0])) return 0;
    if (entityName && entityName[0]) {
        MapEntityDef *found = P_MapEntityDefByName(entityName);
        if (found) return found;
    }
    else {
        MapEntityDef *found = P_MapEntityDef(identifier);
        if (found) return found;
    }
    if (!canCreate) return 0;
    if (P_MapEntityDefByName(entityName)) return 0;
    if (P_MapEntityDef(identifier)) return 0;
    if (!entityDefs) {
        entityDefs = new StringPool;
    }
    StringPool::Id id = entityDefs->intern(String(entityName));
    MapEntityDef *def = new MapEntityDef(identifier);
    entityDefs->setUserPointer(id, def);
    entityDefIdMap.insert(std::pair<int, StringPool::Id>(identifier, id));
    return def;
}
DENG_EXTERN_C dd_bool P_RegisterMapObj(int identifier, char const *name) {
    return findMapEntityDef(identifier, name, true /*do create*/) != 0;
}
DENG_EXTERN_C dd_bool P_RegisterMapObjProperty(int entityId, int propertyId,
                                               char const *propertyName, valuetype_t type) {
    try {
        MapEntityDef *def = findMapEntityDef(entityId, 0, false /*do not create*/);
        if (!def) throw Error("P_RegisterMapObjProperty", QString("Unknown entityId %1").arg(entityId));
        MapEntityDef_AddProperty(def, propertyId, propertyName, type);
        return true; // Success!
    }
    catch (Error const &er) {
        LOG_WARNING("%s. Ignoring.") << er.asText();
    }
    return false;
}
void P_InitMapEntityDefs() {
    clearEntityDefs();
}
void P_ShutdownMapEntityDefs() {
    clearEntityDefs();
}
static MapEntityPropertyDef *entityPropertyDef(int entityId, int propertyId) {
    MapEntityDef *entity = P_MapEntityDef(entityId);
    if (!entity) throw Error("entityPropertyDef", QString("Unknown entity definition id %1").arg(entityId));
    MapEntityPropertyDef *property;
    if (MapEntityDef_Property(entity, propertyId, &property) < 0)
        throw Error("entityPropertyDef", QString("Entity definition %1 has no property with id %2")
                                                 .arg(Str_Text(P_NameForMapEntityDef(entity)))
                                                 .arg(propertyId));
    return property; // Found it.
}
static void setValue(void *dst, valuetype_t dstType, PropertyValue const &pvalue) {
    switch (dstType) {
    case DDVT_FIXED:  *((fixed_t *) dst) = pvalue.asFixed();  break;
    case DDVT_FLOAT:  *(  (float *) dst) = pvalue.asFloat();  break;
    case DDVT_DOUBLE: *( (double *) dst) = pvalue.asDouble(); break;
    case DDVT_BYTE:   *(   (byte *) dst) = pvalue.asByte();   break;
    case DDVT_INT:    *(    (int *) dst) = pvalue.asInt32();  break;
    case DDVT_SHORT:  *(  (short *) dst) = pvalue.asInt16();  break;
    case DDVT_ANGLE:  *((angle_t *) dst) = pvalue.asAngle();  break;
    default:
        throw Error("setValue", QString("Unknown value type %d").arg(dstType));
    }
}
dd_bool P_GMOPropertyIsSet(int entityId, int elementIndex, int propertyId) {
    if (World::get().hasMap()) {
        World::get().map().entityDatabase()
                .hasPropertyValue(entityPropertyDef(entityId, propertyId), elementIndex);
    }
    return false;
}
template <typename Type, valuetype_t returnValueType>
Type getEntityValue(int entityId, int elementIndex, int propertyId) {
    try {
        Type returnVal = 0;
        if (World::get().hasMap()) {
            EntityDatabase const &db = World::get().map().entityDatabase();
            MapEntityPropertyDef const *propDef = entityPropertyDef(entityId, propertyId);
            setValue(&returnVal, returnValueType, db.property(propDef, elementIndex));
        }
        return returnVal;
    }
    catch (Error const &er) {
        LOG_WARNING("%s. Returning 0.") << er.asText();
        return 0;
    }
}
DENG_EXTERN_C byte P_GetGMOByte(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<byte, DDVT_BYTE>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C short P_GetGMOShort(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<short, DDVT_SHORT>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C int P_GetGMOInt(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<int, DDVT_INT>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C fixed_t P_GetGMOFixed(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<fixed_t, DDVT_FIXED>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C angle_t P_GetGMOAngle(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<angle_t, DDVT_ANGLE>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C float P_GetGMOFloat(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<float, DDVT_FLOAT>(entityId, elementIndex, propertyId);
}
DENG_EXTERN_C double P_GetGMODouble(int entityId, int elementIndex, int propertyId) {
    return getEntityValue<double, DDVT_DOUBLE>(entityId, elementIndex, propertyId);
}
namespace world {
DENG2_PIMPL(BaseMap)
, DENG2_OBSERVES(Record, Deletion) {
    EntityDatabase entityDatabase;
    res::MapManifest *manifest = nullptr;  ///< Not owned, may be @c nullptr.
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) i->mapBeingDeleted(self());
    }
    void recordBeingDeleted(Record &record) {
        if (manifest == &record) {
            manifest = nullptr;
        }
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
};
DENG2_AUDIENCE_METHOD(BaseMap, Deletion)
BaseMap::BaseMap(res::MapManifest *manifest) : d(new Impl(this)) {
    setManifest(manifest);
}
BaseMap::~BaseMap()
{}
String BaseMap::id() const {
    if (!hasManifest()) return "";
    return manifest().gets("id");
}
bool BaseMap::hasManifest() const {
    return d->manifest != nullptr;
}
res::MapManifest &BaseMap::manifest() const {
    if (hasManifest()) {
        DENG2_ASSERT(d->manifest != nullptr);
        return *d->manifest;
    }
    throw MissingResourceManifestError("world::Map", "No associated resource manifest");
}
void BaseMap::setManifest(res::MapManifest *newManifest) {
    if (d->manifest) d->manifest->audienceForDeletion() -= d;
    d->manifest = newManifest;
    if (d->manifest) d->manifest->audienceForDeletion() += d;
}
EntityDatabase &BaseMap::entityDatabase() const {
    return d->entityDatabase;
}
void BaseMap::serializeInternalState(Writer &) const
{}
void BaseMap::deserializeInternalState(Reader &, IThinkerMapping const &)
{}
}  // namespace world
namespace world {
DENG2_PIMPL_NOREF(MapElement) {
    dint type;
    Map *map            = nullptr;
    dint indexInMap     = NoIndex;
    dint indexInArchive = NoIndex;
    Impl(dint type) : type(type) {}
};
MapElement::MapElement(dint type, MapElement *parent)
    : d(new Impl(type)) {
    setParent(parent);
}
MapElement::~MapElement()
{}
dint MapElement::type() const {
    return d->type;
}
void MapElement::setParent(MapElement *newParent) {
    if (newParent == this) {
        throw InvalidParentError("MapElement::setParent", "Cannot attribute 'this' map element as a parent of itself");
    }
    _parent = newParent;
}
bool MapElement::hasMap() const {
    if (_parent) {
        return _parent->hasMap();
    }
    return d->map != nullptr;
}
Map &MapElement::map() const {
    if (_parent) {
        return _parent->map();
    }
    if (d->map) {
        return *d->map;
    }
    throw MissingMapError("MapElement::map", "No map is attributed");
}
void MapElement::setMap(Map *newMap) {
    if (!_parent) {
        d->map = newMap;
        return;
    }
    throw WritePropertyError("MapElement::setMap", "The 'map' property has been delegated");
}
dint MapElement::indexInMap() const {
    return d->indexInMap;
}
void MapElement::setIndexInMap(dint newIndex) {
    d->indexInMap = newIndex;
}
dint MapElement::indexInArchive() const {
    return d->indexInArchive;
}
void MapElement::setIndexInArchive(dint newIndex) {
    d->indexInArchive = newIndex;
}
dint MapElement::property(DmuArgs &args) const {
    switch (args.prop) {
    case DMU_ARCHIVE_INDEX:
        args.setValue(DMT_ARCHIVE_INDEX, &d->indexInArchive, 0);
        break;
    default:
        throw UnknownPropertyError(QString("%1::property").arg(DMU_Str(d->type)),
                                   QString("'%1' is unknown/not readable").arg(DMU_Str(args.prop)));
    }
    return false; // Continue iteration.
}
dint MapElement::setProperty(DmuArgs const &args) {
    throw WritePropertyError(QString("%1::setProperty").arg(DMU_Str(d->type)),
                             QString("'%1' is unknown/not writable").arg(DMU_Str(args.prop)));
}
}  // namespace world
char const *DMU_Str(uint prop) {
    static char propStr[40];
    struct prop_s {
        uint prop;
        char const *str;
    } props[] = {
        { DMU_NONE,              "(invalid)" },
        { DMU_VERTEX,            "DMU_VERTEX" },
        { DMU_SEGMENT,           "DMU_SEGMENT" },
        { DMU_LINE,              "DMU_LINE" },
        { DMU_SIDE,              "DMU_SIDE" },
        { DMU_SUBSPACE,          "DMU_SUBSPACE" },
        { DMU_SECTOR,            "DMU_SECTOR" },
        { DMU_PLANE,             "DMU_PLANE" },
        { DMU_SURFACE,           "DMU_SURFACE" },
        { DMU_MATERIAL,          "DMU_MATERIAL" },
        { DMU_SKY,               "DMU_SKY" },
        { DMU_LINE_BY_TAG,       "DMU_LINE_BY_TAG" },
        { DMU_SECTOR_BY_TAG,     "DMU_SECTOR_BY_TAG" },
        { DMU_LINE_BY_ACT_TAG,   "DMU_LINE_BY_ACT_TAG" },
        { DMU_SECTOR_BY_ACT_TAG, "DMU_SECTOR_BY_ACT_TAG" },
        { DMU_ARCHIVE_INDEX,     "DMU_ARCHIVE_INDEX" },
        { DMU_X,                 "DMU_X" },
        { DMU_Y,                 "DMU_Y" },
        { DMU_XY,                "DMU_XY" },
        { DMU_TANGENT_X,         "DMU_TANGENT_X" },
        { DMU_TANGENT_Y,         "DMU_TANGENT_Y" },
        { DMU_TANGENT_Z,         "DMU_TANGENT_Z" },
        { DMU_TANGENT_XYZ,       "DMU_TANGENT_XYZ" },
        { DMU_BITANGENT_X,       "DMU_BITANGENT_X" },
        { DMU_BITANGENT_Y,       "DMU_BITANGENT_Y" },
        { DMU_BITANGENT_Z,       "DMU_BITANGENT_Z" },
        { DMU_BITANGENT_XYZ,     "DMU_BITANGENT_XYZ" },
        { DMU_NORMAL_X,          "DMU_NORMAL_X" },
        { DMU_NORMAL_Y,          "DMU_NORMAL_Y" },
        { DMU_NORMAL_Z,          "DMU_NORMAL_Z" },
        { DMU_NORMAL_XYZ,        "DMU_NORMAL_XYZ" },
        { DMU_VERTEX0,           "DMU_VERTEX0" },
        { DMU_VERTEX1,           "DMU_VERTEX1" },
        { DMU_FRONT,             "DMU_FRONT" },
        { DMU_BACK,              "DMU_BACK" },
        { DMU_FLAGS,             "DMU_FLAGS" },
        { DMU_DX,                "DMU_DX" },
        { DMU_DY,                "DMU_DY" },
        { DMU_DXY,               "DMU_DXY" },
        { DMU_LENGTH,            "DMU_LENGTH" },
        { DMU_SLOPETYPE,         "DMU_SLOPETYPE" },
        { DMU_ANGLE,             "DMU_ANGLE" },
        { DMU_OFFSET,            "DMU_OFFSET" },
        { DMU_OFFSET_X,          "DMU_OFFSET_X" },
        { DMU_OFFSET_Y,          "DMU_OFFSET_Y" },
        { DMU_OFFSET_XY,         "DMU_OFFSET_XY" },
        { DMU_BLENDMODE,         "DMU_BLENDMODE" },
        { DMU_VALID_COUNT,       "DMU_VALID_COUNT" },
        { DMU_COLOR,             "DMU_COLOR" },
        { DMU_COLOR_RED,         "DMU_COLOR_RED" },
        { DMU_COLOR_GREEN,       "DMU_COLOR_GREEN" },
        { DMU_COLOR_BLUE,        "DMU_COLOR_BLUE" },
        { DMU_ALPHA,             "DMU_ALPHA" },
        { DMU_LIGHT_LEVEL,       "DMU_LIGHT_LEVEL" },
        { DMT_MOBJS,             "DMT_MOBJS" },
        { DMU_BOUNDING_BOX,      "DMU_BOUNDING_BOX" },
        { DMU_EMITTER,           "DMU_EMITTER" },
        { DMU_WIDTH,             "DMU_WIDTH" },
        { DMU_HEIGHT,            "DMU_HEIGHT" },
        { DMU_TARGET_HEIGHT,     "DMU_TARGET_HEIGHT" },
        { DMU_SPEED,             "DMU_SPEED" },
        { DMU_FLOOR_PLANE,       "DMU_FLOOR_PLANE" },
        { DMU_CEILING_PLANE,     "DMU_CEILING_PLANE" },
        { 0, NULL }
    };
    for (uint i = 0; props[i].str; ++i) {
        if (props[i].prop == prop)
            return props[i].str;
    }
    dd_snprintf(propStr, 40, "(unnamed %i)", prop);
    return propStr;
}
namespace world {
Material::Layer::~Layer() {
    qDeleteAll(_stages);
}
Material::Layer::Stage &Material::Layer::stage(int index) const {
    if (!stageCount()) {
        throw MissingStageError("Material::Layer::stage", "Layer has no stages");
    }
    return *_stages[de::wrap(index, 0, _stages.count())];
}
int Material::Layer::nextStageIndex(int index) const {
    if (!stageCount()) return -1;
    return de::wrap(index + 1, 0, _stages.count());
}
String Material::Layer::describe() const {
    return "abstract Layer";
}
String Material::Layer::description() const {
    int const numStages = stageCount();
    String str = _E(b) + describe() + _E(.) + " (" + String::number(numStages) + " stage" + DENG2_PLURAL_S(numStages) + "):";
    for (int i = 0; i < numStages; ++i) {
        str += String("\n  [%1] ").arg(i, 2) + _E(>) + stage(i).description() + _E(<);
    }
    return str;
}
DENG2_PIMPL(Material)
, DENG2_OBSERVES(res::Texture, Deletion)
, DENG2_OBSERVES(res::Texture, DimensionsChange) {
    MaterialManifest *manifest = nullptr;  ///< Source manifest (always valid, not owned).
    Vector2ui dimensions;                  ///< World dimensions in map coordinate space units.
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        qDeleteAll(self()._layers);
    }
    inline bool haveValidDimensions() const {
        return dimensions.x > 0 && dimensions.y > 0;
    }
    TextureMaterialLayer *firstTextureLayer() const {
        for (Layer *layer : self()._layers) {
            if (is<DetailTextureMaterialLayer>(layer)) continue;
            if (is<ShineTextureMaterialLayer> (layer)) continue;
            if (auto *texLayer = maybeAs<TextureMaterialLayer>(layer)) {
                return texLayer;
            }
        }
        return nullptr;
    }
    /**
     * Determines which texture we would be interested in obtaining our world dimensions
     * from if our own dimensions are undefined.
     */
    res::Texture *inheritDimensionsTexture() const {
        if (auto const *texLayer = firstTextureLayer()) {
            if (texLayer->stageCount() >= 1) {
                try {
                    return &res::Textures::get().texture(texLayer->stage(0).texture);
                }
                catch (res::TextureManifest::MissingTextureError &)
                {}
                catch (Resources::MissingResourceManifestError &)
                {}
            }
        }
        return nullptr;
    }
    /**
     * Determines whether the world dimensions are now defined and if so cancels future
     * notifications about changes to texture dimensions.
     */
    void maybeCancelTextureDimensionsChangeNotification() {
        if (haveValidDimensions()) return;
        res::Texture *inheritanceTexture = inheritDimensionsTexture();
        if (!inheritanceTexture) return;
        inheritanceTexture->audienceForDimensionsChange -= this;
        inheritanceTexture->audienceForDeletion -= this;
    }
    void textureDimensionsChanged(res::Texture const &texture) {
        DENG2_ASSERT(!haveValidDimensions()); // Sanity check.
        self().setDimensions(texture.dimensions());
    }
    void textureBeingDeleted(res::Texture const &texture) {
        DENG2_ASSERT(!haveValidDimensions()); // Sanity check.
        DENG2_ASSERT(inheritDimensionsTexture() == &texture); // Sanity check.
        firstTextureLayer()->stage(0).texture = de::Uri();
#if !defined(DENG2_DEBUG)
        DENG2_UNUSED(texture);
#endif
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(DimensionsChange)
};
DENG2_AUDIENCE_METHOD(Material, Deletion)
DENG2_AUDIENCE_METHOD(Material, DimensionsChange)
Material::Material(MaterialManifest &manifest)
    : MapElement(DMU_MATERIAL)
    , d(new Impl(this)) {
    d->manifest = &manifest;
}
Material::~Material() {
    d->maybeCancelTextureDimensionsChangeNotification();
    DENG2_FOR_AUDIENCE2(Deletion, i) i->materialBeingDeleted(*this);
}
MaterialManifest &Material::manifest() const {
    DENG2_ASSERT(d->manifest);
    return *d->manifest;
}
Vector2ui const &Material::dimensions() const {
    return d->dimensions;
}
void Material::setDimensions(Vector2ui const &newDimensions) {
    if (d->dimensions != newDimensions) {
        d->dimensions = newDimensions;
        d->maybeCancelTextureDimensionsChangeNotification();
        DENG2_FOR_AUDIENCE2(DimensionsChange, i) i->materialDimensionsChanged(*this);
    }
}
void Material::setHeight(int newHeight) {
    setDimensions(Vector2ui(width(), newHeight));
}
bool Material::isAnimated() const {
    return hasAnimatedTextureLayers();
}
void Material::setWidth(int newWidth) {
    setDimensions(Vector2ui(newWidth, height()));
}
void Material::markDontDraw(bool yes) {
    applyFlagOperation(_flags, DontDraw, yes);
}
void Material::markSkyMasked(bool yes) {
    applyFlagOperation(_flags, SkyMasked, yes);
}
void Material::markValid(bool yes) {
    applyFlagOperation(_flags, Valid, yes);
}
void Material::clearAllLayers() {
    d->maybeCancelTextureDimensionsChangeNotification();
    qDeleteAll(_layers);
    _layers.clear();
}
bool Material::hasAnimatedTextureLayers() const {
    for (Layer const *layer : _layers) {
        if (   !is<DetailTextureMaterialLayer>(layer)
            && !is<ShineTextureMaterialLayer>(layer)) {
            if (layer->isAnimated())
                return true;
        }
    }
    return false;
}
void Material::addLayerAt(Layer *layer, int position) {
    if (!layer) return;
    if (_layers.contains(layer)) return;
    position = de::clamp(0, position, layerCount());
    d->maybeCancelTextureDimensionsChangeNotification();
    _layers.insert(position, layer);
    if (!d->haveValidDimensions()) {
        if (res::Texture *tex = d->inheritDimensionsTexture()) {
            tex->audienceForDeletion         += d;
            tex->audienceForDimensionsChange += d;
        }
    }
}
/*Material::Layer &Material::layer(int index) const
{
    throw MissingLayerError("Material::layer", "Unknown layer #" + String::number(index));
}*/
String Material::describe() const {
    return "Material \"" + manifest().composeUri().asText() + "\"";
}
String Material::description() const {
    String str = String(_E(l) "Dimensions: ") + _E(.) + (d->haveValidDimensions()? dimensions().asText() : "unknown (not yet prepared)")
               + _E(l) + " Source: "     + _E(.) + manifest().sourceDescription()
               + _E(l) + "\nDrawable: "  + _E(.) + DENG2_BOOL_YESNO(isDrawable())
               + _E(l) + " SkyMasked: "  + _E(.) + DENG2_BOOL_YESNO(isSkyMasked());
    for (Layer const *layer : _layers) {
        str += "\n" + layer->description();
    }
    return str;
}
int Material::property(DmuArgs &args) const {
    switch (args.prop) {
    case DMU_FLAGS: {
        short f = _flags;
        args.setValue(DMT_MATERIAL_FLAGS, &f, 0);
        break; }
    case DMU_HEIGHT: {
        int h = d->dimensions.y;
        args.setValue(DMT_MATERIAL_HEIGHT, &h, 0);
        break; }
    case DMU_WIDTH: {
        int w = d->dimensions.x;
        args.setValue(DMT_MATERIAL_WIDTH, &w, 0);
        break; }
    default:
        return MapElement::property(args);
    }
    return false; // Continue iteration.
}
D_CMD(InspectMaterial) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1);
    if (!search.scheme().isEmpty() &&
        !world::Materials::get().isKnownMaterialScheme(search.scheme())) {
        LOG_SCR_WARNING("Unknown scheme \"%s\"") << search.scheme();
        return false;
    }
    try {
        MaterialManifest &manifest = world::Materials::get().materialManifest(search);
        if (Material *material = manifest.materialPtr()) {
            LOG_SCR_MSG(_E(D)_E(b) "%s\n" _E(.)_E(.)) << material->describe() << material->description();
        }
        else {
            LOG_SCR_MSG(manifest.description());
        }
        return true;
    }
    catch (Resources::MissingResourceManifestError const &er) {
        LOG_SCR_WARNING("%s") << er.asText();
    }
    return false;
}
void Material::consoleRegister() // static
{
    C_CMD("inspectmaterial",    "ss",   InspectMaterial)
    C_CMD("inspectmaterial",    "s",    InspectMaterial)
}
} // namespace world
namespace world {
static String readArchivedPath(reader_s &reader) {
    char _path[9];
    Reader_Read(&reader, _path, 8); _path[8] = 0;
    return QString(QByteArray(_path, qstrlen(_path)).toPercentEncoding());
}
static void readArchivedUri(de::Uri &uri, int version, reader_s &reader) {
    if (version >= 4) {
        uri.readUri(&reader);
    }
    else if (version == 3) {
        ddstring_t *_uri = Str_NewFromReader(&reader);
        uri.setUri(Str_Text(_uri), RC_NULL);
        Str_Delete(_uri);
    }
    else if (version == 2) {
        ddstring_t *_uri = Str_NewFromReader(&reader);
        uri.setUri(QString(QByteArray(Str_Text(_uri), Str_Length(_uri)).toPercentEncoding()), RC_NULL);
        Str_Delete(_uri);
    }
    else // ver 1
    {
        uri.setPath(readArchivedPath(reader));
        int oldSchemeId = Reader_ReadByte(&reader);
        switch (oldSchemeId) {
        case 0: uri.setScheme("Textures"); break;
        case 1: uri.setScheme("Flats");    break;
        case 2: uri.setScheme("Sprites");  break;
        case 3: uri.setScheme("System");   break;
        default:
            throw Error("readArchiveUri", QString("Unknown old-scheme id #%1, expected [0..3)").arg(oldSchemeId));
        }
    }
}
/**
 * Mappings between URI and Material.
 * The pointer user value holds a pointer to the resolved Material (if found).
 * The integer user value tracks whether a material has yet been looked up.
 */
typedef StringPool Records;
typedef StringPool::Id SerialId;
static Material *findRecordMaterial(Records &records, SerialId id) {
    if (!records.userValue(id)) {
        Material *material = nullptr;
        try {
            material = &world::Materials::get().material(de::makeUri(records.stringRef(id)));
        }
        catch (Resources::MissingResourceManifestError const &)
        {}  // Ignore this error.
        records.setUserPointer(id, material);
        records.setUserValue(id, true);
        return material;
    }
    return reinterpret_cast<Material *>(records.userPointer(id));
}
DENG2_PIMPL(MaterialArchive) {
    int version = MATERIALARCHIVE_VERSION;
    bool useSegments = false;   ///< Segment id assertion (Hexen saves).
    Records records;            ///< Mappings between URI and Material.
    int numFlats = 0;           ///< Used with older versions.
    Impl(Public *i) : Base(i)
    {}
    inline SerialId insertRecord(de::Uri const &uri) {
        return records.intern(uri.compose());
    }
    void beginSegment(int seg, writer_s &writer) {
        if (!useSegments) return;
        Writer_WriteUInt32(&writer, seg);
    }
    void assertSegment(int seg, reader_s &reader) {
        if (!useSegments) return;
        int i = Reader_ReadUInt32(&reader);
        if (i != seg) {
            throw MaterialArchive::ReadError("MaterialArchive::assertSegment",
                                             QString("Expected ASEG_MATERIAL_ARCHIVE (%1), but got %2")
                                                 .arg(ASEG_MATERIAL_ARCHIVE).arg(i));
        }
    }
    void writeHeader(writer_s &writer) {
        beginSegment(ASEG_MATERIAL_ARCHIVE, writer);
        Writer_WriteByte(&writer, version);
    }
    void readHeader(reader_s &reader) {
        assertSegment(ASEG_MATERIAL_ARCHIVE, reader);
        version = Reader_ReadByte(&reader);
    }
    void readGroup(reader_s &reader) {
        DENG_ASSERT(version >= 1);
        int num = Reader_ReadUInt16(&reader);
        de::Uri uri;
        for (int i = 0; i < num; ++i) {
            readArchivedUri(uri, version, reader);
            insertRecord(uri);
        }
    }
    void writeGroup(writer_s &writer) {
        Writer_WriteUInt16(&writer, records.size());
        de::Uri uri;
        for (int i = 1; i < int(records.size()) + 1; ++i) {
            uri.setUri(records.stringRef(i), RC_NULL);
            uri.writeUri(&writer);
        }
    }
};
MaterialArchive::MaterialArchive(int useSegments, bool recordSymbolicMaterials)
    : d(new Impl(this)) {
    d->useSegments = useSegments;
    if (recordSymbolicMaterials) {
        d->insertRecord(de::makeUri(UNKNOWN_MATERIALNAME));
    }
}
void MaterialArchive::addWorldMaterials() {
    world::Materials::get().forAllMaterials([this] (Material &material) {
        addRecord(material);
        return de::LoopContinue;
    });
}
materialarchive_serialid_t MaterialArchive::findUniqueSerialId(Material *material) const {
    materialarchive_serialid_t serialId = 0;  // Invalid.
    if (material) {
        LoopResult found = d->records.forAll([this, &material, &serialId] (SerialId id) {
            if (findRecordMaterial(d->records, id) == material) {
                serialId = id;
                return LoopAbort;
            }
            return LoopContinue;
        });
        if (!found) {
            serialId = d->records.size() + 1;
        }
    }
    return serialId;
}
world::Material *MaterialArchive::find(materialarchive_serialid_t serialId, int group) const {
    if (serialId <= 0 || serialId > d->records.size() + 1) return 0; // Invalid.
    if (d->version < 1 && group == 1) {
        serialId += d->numFlats;
    }
    if (d->version <= 1) {
        de::Uri uri(d->records.stringRef(serialId), RC_NULL);
        if (!uri.path().toStringRef().compareWithoutCase(UNKNOWN_MATERIALNAME))
            return 0;
    }
    return findRecordMaterial(d->records, serialId);
}
materialarchive_serialid_t MaterialArchive::addRecord(Material const &material) {
    SerialId id = d->insertRecord(material.manifest().composeUri());
    d->records.setUserPointer(id, const_cast<Material *>(&material));
    d->records.setUserValue(id, true);
    return materialarchive_serialid_t(id);
}
int MaterialArchive::count() const {
    return d->records.size();
}
void MaterialArchive::write(writer_s &writer) const {
    d->writeHeader(writer);
    d->writeGroup(writer);
}
void MaterialArchive::read(reader_s &reader, int forcedVersion) {
    d->records.clear();
    d->readHeader(reader);
    if (forcedVersion >= 0) {
        d->version = forcedVersion;
    }
    if (d->version >= 1) {
        d->readGroup(reader);
        return;
    } {
        de::Uri uri("Flats", "");
        d->numFlats = Reader_ReadUInt16(&reader);
        for (int i = 0; i < d->numFlats; ++i) {
            uri.setPath(readArchivedPath(reader));
            d->insertRecord(uri);
        }
    } {
        de::Uri uri("Textures", "");
        int num = Reader_ReadUInt16(&reader);
        for (int i = 0; i < num; ++i) {
            uri.setPath(readArchivedPath(reader));
            d->insertRecord(uri);
        }
    }
}
} // namespace world
#if 0
/*
 * C Wrapper API:
 */
#define DENG_NO_API_MACROS_MATERIAL_ARCHIVE
#include "api_materialarchive.h"
#define TOINTERNAL(inst) \
    reinterpret_cast<de::MaterialArchive *>(inst)
#define TOINTERNAL_CONST(inst) \
    reinterpret_cast<de::MaterialArchive const*>(inst)
#define SELF(inst) \
    DENG2_ASSERT(inst); \
    de::MaterialArchive *self = TOINTERNAL(inst)
#define SELF_CONST(inst) \
    DENG2_ASSERT(inst); \
    de::MaterialArchive const *self = TOINTERNAL_CONST(inst)
#undef MaterialArchive_New
MaterialArchive *MaterialArchive_New(int useSegments) {
    auto *archive = new world::MaterialArchive(useSegments);
    App_Resources().forAllMaterials([&archive] (Material &material) {
        archive->addRecord(material);
        return de::LoopContinue;
    });
    return reinterpret_cast<MaterialArchive *>(archive);
}
#undef MaterialArchive_NewEmpty
MaterialArchive *MaterialArchive_NewEmpty(int useSegments) {
    return reinterpret_cast<MaterialArchive *>(new world::MaterialArchive(useSegments, false /*don't populate*/));
}
#undef MaterialArchive_Delete
void MaterialArchive_Delete(MaterialArchive *arc) {
    if (arc) {
        SELF(arc);
        delete self;
    }
}
#undef MaterialArchive_FindUniqueSerialId
materialarchive_serialid_t MaterialArchive_FindUniqueSerialId(MaterialArchive const *arc, Material *mat) {
    SELF_CONST(arc);
    return self->findUniqueSerialId(mat);
}
#undef MaterialArchive_Find
Material *MaterialArchive_Find(MaterialArchive const *arc, materialarchive_serialid_t serialId, int group) {
    SELF_CONST(arc);
    return self->find(serialId, group);
}
#undef MaterialArchive_Count
int MaterialArchive_Count(MaterialArchive const *arc) {
    SELF_CONST(arc);
    return self->count();
}
#undef MaterialArchive_Write
void MaterialArchive_Write(MaterialArchive const *arc, Writer *writer) {
    SELF_CONST(arc);
    self->write(*writer);
}
#undef MaterialArchive_Read
void MaterialArchive_Read(MaterialArchive *arc, Reader *reader, int forcedVersion) {
    SELF(arc);
    self->read(*reader, forcedVersion);
}
#endif
namespace world {
static MaterialManifest::MaterialConstructor materialConstructor;
DENG2_PIMPL_NOREF(MaterialManifest)
, DENG2_OBSERVES(Material, Deletion) {
    MaterialScheme *ownerScheme = nullptr;
    Flags flags;                         ///< Classification flags.
    materialid_t id = 0;                 ///< Globally unique identifier.
    std::unique_ptr<Material> material;  ///< Associated resource (if any).
    void materialBeingDeleted(Material const &) {
        material.release();
    }
};
MaterialManifest::MaterialManifest(PathTree::NodeArgs const &args)
    : Node(args)
    , d(new Impl)
{}
MaterialManifest::~MaterialManifest() {
    DENG2_FOR_AUDIENCE(Deletion, i) i->materialManifestBeingDeleted(*this);
}
Material *MaterialManifest::derive() {
    if (!hasMaterial()) {
        DENG_ASSERT(materialConstructor);
        setMaterial(materialConstructor(*this));
        DENG2_FOR_AUDIENCE(MaterialDerived, i) i->materialManifestMaterialDerived(*this, material());
    }
    return &material();
}
void MaterialManifest::setScheme(MaterialScheme &scheme) {
    d->ownerScheme = &scheme;
}
materialid_t MaterialManifest::id() const {
    return d->id;
}
void MaterialManifest::setId(materialid_t id) {
    d->id = id;
}
MaterialScheme &MaterialManifest::scheme() const {
    DENG_ASSERT(d->ownerScheme);
    return *d->ownerScheme;
}
String const &MaterialManifest::schemeName() const {
    return scheme().name();
}
String MaterialManifest::description(de::Uri::ComposeAsTextFlags uriCompositionFlags) const {
    String info = String("%1 %2")
                      .arg(composeUri().compose(uriCompositionFlags | de::Uri::DecodePath),
                           ( uriCompositionFlags.testFlag(de::Uri::OmitScheme)? -14 : -22 ) )
                      .arg(sourceDescription(), -7);
#ifdef __CLIENT__
    info += String("x%1").arg(!hasMaterial()? 0 : material().animatorCount());
#endif
    return info;
}
String MaterialManifest::sourceDescription() const {
    if (!isCustom())       return "game";
    if (isAutoGenerated()) return "add-on";  // Unintuitive but correct.
    return "def";
}
MaterialManifest::Flags MaterialManifest::flags() const {
    return d->flags;
}
void MaterialManifest::setFlags(MaterialManifest::Flags flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
}
bool MaterialManifest::hasMaterial() const {
    return bool(d->material);
}
Material &MaterialManifest::material() const {
    if (hasMaterial()) return *d->material;
    throw MissingMaterialError("MaterialManifest::material", "Missing required material");
}
Material *MaterialManifest::materialPtr() const {
    return d->material.get();
}
void MaterialManifest::setMaterial(Material *newMaterial) {
    if (d->material.get() != newMaterial) {
        if (d->material) {
            d->material->audienceForDeletion() -= d;
        }
        d->material.reset(newMaterial);
        if (d->material) {
            d->material->audienceForDeletion() += d;
        }
    }
}
void MaterialManifest::setMaterialConstructor(MaterialConstructor func) // static
{
    materialConstructor = func;
}
} // namespace world
namespace world {
struct SchemeHashKey {
    String scheme;
    SchemeHashKey(String const &s) : scheme(s) {}
    bool operator == (SchemeHashKey const &other) const {
        return !scheme.compare(other.scheme, Qt::CaseInsensitive);
    }
};
uint qHash(SchemeHashKey const &key) {
    return key.scheme.at(1).toLower().unicode();
}
DENG2_PIMPL(Materials)
, DENG2_OBSERVES(MaterialScheme,   ManifestDefined)
, DENG2_OBSERVES(MaterialManifest, MaterialDerived)
, DENG2_OBSERVES(MaterialManifest, Deletion)
, DENG2_OBSERVES(Material,         Deletion) {
    QHash<SchemeHashKey, MaterialScheme *> materialSchemes;
    QList<MaterialScheme *> materialSchemeCreationOrder;
    QList<Material *> materials;       ///< From all schemes.
    int materialManifestCount = 0;     ///< Total number of material manifests (in all schemes).
    std::unordered_set<Material *> animatedMaterialsSubset; ///< Subset of materials (not owned) that need to animate.
    MaterialManifestGroups materialGroups;
    uint materialManifestIdMapSize = 0;
    MaterialManifest **materialManifestIdMap = nullptr;  ///< Index with materialid_t-1
    Impl(Public *i) : Base(i) {
        createMaterialScheme("Sprites");
        createMaterialScheme("Textures");
        createMaterialScheme("Flats");
        createMaterialScheme("System");
    }
    ~Impl() {
        self().clearAllMaterialGroups();
        self().clearAllMaterialSchemes();
        clearMaterialManifests();
    }
    void clearMaterialManifests() {
        qDeleteAll(materialSchemes);
        materialSchemes.clear();
        materialSchemeCreationOrder.clear();
        if (materialManifestIdMap) {
            M_Free(materialManifestIdMap);
            materialManifestIdMap = 0;
            materialManifestIdMapSize = 0;
        }
        materialManifestCount = 0;
    }
    void createMaterialScheme(String name) {
        DENG2_ASSERT(name.length() >= MaterialScheme::min_name_length);
        MaterialScheme *newScheme = new MaterialScheme(name);
        materialSchemes.insert(name, newScheme);
        materialSchemeCreationOrder.append(newScheme);
        newScheme->audienceForManifestDefined += this;
    }
    void materialSchemeManifestDefined(MaterialScheme & /*scheme*/, MaterialManifest &manifest)
    {
        int const MANIFESTIDMAP_BLOCK_ALLOC = 32;
        manifest.audienceForMaterialDerived += this;
        manifest.audienceForDeletion += this;
        materialid_t const id = materialid_t(++materialManifestCount); // 1-based.
        manifest.setId(id);
        if (materialManifestCount > int(materialManifestIdMapSize)) {
            materialManifestIdMapSize += MANIFESTIDMAP_BLOCK_ALLOC;
            materialManifestIdMap = (MaterialManifest **) M_Realloc(materialManifestIdMap, sizeof(*materialManifestIdMap) * materialManifestIdMapSize);
        }
        materialManifestIdMap[materialManifestCount - 1] = &manifest;
    }
    void materialManifestMaterialDerived(MaterialManifest & /*manifest*/, Material &material)
    {
        materials.append(&material);
        material.audienceForDeletion() += this;
    }
    void materialManifestBeingDeleted(MaterialManifest const &manifest) {
        foreach (MaterialManifestGroup *group, materialGroups) {
            group->remove(const_cast<MaterialManifest *>(&manifest));
        }
        materialManifestIdMap[manifest.id() - 1 /*1-based*/] = 0;
        materialManifestCount -= 1;
    }
    void materialBeingDeleted(Material const &material) {
        Material *pMat = const_cast<Material *>(&material);
        materials.removeOne(pMat);
        animatedMaterialsSubset.erase(pMat);
    }
};
Materials::Materials()
    : d(new Impl(this))
{}
MaterialScheme &Materials::materialScheme(String name) const {
    if (!name.isEmpty()) {
        auto found = d->materialSchemes.find(name);
        if (found != d->materialSchemes.end()) return **found;
    }
    throw Resources::UnknownSchemeError("Materials::materialScheme",
                                        "No scheme found matching '" + name + "'");
}
bool Materials::isKnownMaterialScheme(String name) const {
    if (!name.isEmpty()) {
        return d->materialSchemes.contains(name);
    }
    return false;
}
int Materials::materialSchemeCount() const {
    return d->materialSchemes.count();
}
LoopResult Materials::forAllMaterialSchemes(std::function<LoopResult (MaterialScheme &)> func) const {
    for (MaterialScheme *scheme : d->materialSchemes) {
        if (auto result = func(*scheme)) return result;
    }
    return LoopContinue;
}
MaterialManifest &Materials::toMaterialManifest(materialid_t id) const {
    duint32 idx = id - 1; // 1-based index.
    if (idx < duint32(d->materialManifestCount)) {
        if (d->materialManifestIdMap[idx]) {
            return *d->materialManifestIdMap[idx];
        }
        DENG2_ASSERT(false);
    }
    throw UnknownMaterialIdError("Materials::toMaterialManifest",
                                 "Invalid material ID " + String::number(id) +
                                 ", valid range " +
                                 Rangei(1, d->materialManifestCount + 1).asText());
}
Material *Materials::materialPtr(de::Uri const &path) {
    if (auto *manifest = materialManifestPtr(path)) return manifest->materialPtr();
    return nullptr;
}
bool Materials::hasMaterialManifest(de::Uri const &path) const {
    return materialManifestPtr(path) != nullptr;
}
MaterialManifest &Materials::materialManifest(de::Uri const &uri) const {
    if (auto *mm = materialManifestPtr(uri)) {
        return *mm;
    }
    throw Resources::MissingResourceManifestError("Materials::materialManifest",
                                                  "Failed to locate a manifest matching \"" + uri.asText() + "\"");
}
MaterialManifest *Materials::materialManifestPtr(de::Uri const &uri) const {
    if (!uri.scheme().isEmpty()) {
        MaterialScheme &specifiedScheme = materialScheme(uri.scheme());
        return specifiedScheme.tryFind(uri.path());
    }
    else {
        foreach (MaterialScheme *scheme, d->materialSchemeCreationOrder) {
            if (auto *manifest = scheme->tryFind(uri.path())) {
                return manifest;
            }
        }
    }
    return nullptr;
}
dint Materials::materialCount() const {
    return d->materials.count();
}
LoopResult Materials::forAllMaterials(const std::function<LoopResult (Material &)> &func) const {
    for (Material *mat : d.getConst()->materials) {
        if (auto result = func(*mat)) {
            return result;
        }
    }
    return LoopContinue;
}
LoopResult Materials::forAnimatedMaterials(const std::function<LoopResult (Material &)> &func) const {
    for (Material *mat : d.getConst()->animatedMaterialsSubset) {
        if (auto result = func(*mat)) {
            return result;
        }
    }
    return LoopContinue;
}
void Materials::updateLookup() {
    d->animatedMaterialsSubset.clear();
    for (auto *mat : d->materials) {
        if (mat->isAnimated()) {
            d->animatedMaterialsSubset.insert(mat);
        }
    }
}
Materials::MaterialManifestGroup &Materials::newMaterialGroup() {
    d->materialGroups.append(new MaterialManifestGroup());
    return *d->materialGroups.back();
}
Materials::MaterialManifestGroup &Materials::materialGroup(dint groupIdx) const {
    groupIdx -= 1; // 1-based index.
    if (groupIdx >= 0 && groupIdx < d->materialGroups.count()) {
        return *d->materialGroups[groupIdx];
    }
    throw UnknownMaterialGroupError("Materials::materialGroup",
                                    "Invalid group #" + String::number(groupIdx+1) + ", valid range " +
                                    Rangeui(1, d->materialGroups.count() + 1).asText());
}
Materials::MaterialManifestGroups const &Materials::allMaterialGroups() const {
    return d->materialGroups;
}
void Materials::clearAllMaterialGroups() {
    qDeleteAll(d->materialGroups);
    d->materialGroups.clear();
}
void Materials::clearAllMaterialSchemes() {
    forAllMaterialSchemes([] (MaterialScheme &scheme) {
        scheme.clear();
        return LoopContinue;
    });
    DENG2_ASSERT(materialCount() == 0); // sanity check
}
Materials &Materials::get() // static
{
    return World::get().materials();
}
} // namespace world
namespace world {
DENG2_PIMPL(MaterialScheme) {
    String name;
    MaterialScheme::Index index;
    Impl(Public *i, String symbolicName) : Base(i), name(symbolicName)
    {}
    ~Impl() {
        self().clear();
        DENG2_ASSERT(index.isEmpty());
    }
};
MaterialScheme::MaterialScheme(String symbolicName)
    : d(new Impl(this, symbolicName))
{}
void MaterialScheme::clear() {
    d->index.clear();
}
String const &MaterialScheme::name() const {
    return d->name;
}
MaterialManifest &MaterialScheme::declare(Path const &path) {
    LOG_AS("MaterialScheme::declare");
    if (path.isEmpty()) {
        throw InvalidPathError("MaterialScheme::declare", "Missing/zero-length path was supplied");
    }
    int const sizeBefore = d->index.size();
    Manifest *newManifest = &d->index.insert(path);
    DENG2_ASSERT(newManifest);
    newManifest->setScheme(*this);
    if (d->index.size() != sizeBefore) {
        DENG2_FOR_AUDIENCE(ManifestDefined, i) i->materialSchemeManifestDefined(*this, *newManifest);
    }
    return *newManifest;
}
bool MaterialScheme::has(Path const &path) const {
    return d->index.has(path, Index::NoBranch | Index::MatchFull);
}
MaterialManifest const &MaterialScheme::find(Path const &path) const {
    if (has(path)) {
        return d->index.find(path, Index::NoBranch | Index::MatchFull);
    }
    throw NotFoundError("MaterialScheme::find", "Failed to locate a manifest matching \"" + path.asText() + "\"");
}
MaterialManifest &MaterialScheme::find(Path const &path) {
    Index::Node const &found = const_cast<MaterialScheme const *>(this)->find(path);
    return const_cast<Index::Node &>(found);
}
MaterialManifest *MaterialScheme::tryFind(Path const & path) const {
    return d->index.tryFind(path, Index::NoBranch | Index::MatchFull);
}
MaterialScheme::Index const &MaterialScheme::index() const {
    return d->index;
}
} // namespace world
static String const VAR_ID("__id__");
DENG2_PIMPL_NOREF(MobjThinkerData)
{};
MobjThinkerData::MobjThinkerData(de::Id const &id)
    : ThinkerData(id)
    , d(new Impl)
{}
MobjThinkerData::MobjThinkerData(MobjThinkerData const &other)
    : ThinkerData(other)
    , d(new Impl)
{}
void MobjThinkerData::think() {
    auto &mo = *mobj();
    coord_t lastOrigin[3];
    memcpy(lastOrigin, mo.origin, sizeof(lastOrigin));
    mo.ddFlags &= ~DDMF_MOVEBLOCKED;
    ThinkerData::think();
    for (int axis = 0; axis < 3; ++axis) {
        if (fequal(lastOrigin[axis], mo.origin[axis])) {
            mo.ddFlags |= DDMF_MOVEBLOCKEDX << axis;
        }
    }
}
Thinker::IData *MobjThinkerData::duplicate() const {
    return new MobjThinkerData(*this);
}
mobj_t *MobjThinkerData::mobj() {
    return reinterpret_cast<mobj_t *>(&thinker());
}
mobj_t const *MobjThinkerData::mobj() const {
    return reinterpret_cast<mobj_t const *>(&thinker());
}
void MobjThinkerData::initBindings() {
    ThinkerData::initBindings();
    objectNamespace().addSuperRecord(ScriptSystem::builtInClass(
            QStringLiteral("World"), QStringLiteral("Thing")));
    objectNamespace().addNumber(VAR_ID, mobj()->thinker.id).setReadOnly();
}
void MobjThinkerData::stateChanged(state_t const *) {
}
void MobjThinkerData::damageReceived(int, mobj_t const *)
{}
void MobjThinkerData::operator << (Reader &from) {
    ThinkerData::operator << (from);
    initBindings();
}
PropertyValue *BuildPropertyValue(valuetype_t type, void *valueAdr) {
    DENG2_ASSERT(valueAdr != 0);
    switch (type) {
    case DDVT_BYTE:     return new PropertyByteValue  (*(   (byte *) valueAdr));
    case DDVT_SHORT:    return new PropertyInt16Value (*(  (short *) valueAdr));
    case DDVT_INT:      return new PropertyInt32Value (*(    (int *) valueAdr));
    case DDVT_FIXED:    return new PropertyFixedValue (*((fixed_t *) valueAdr));
    case DDVT_ANGLE:    return new PropertyAngleValue (*((angle_t *) valueAdr));
    case DDVT_FLOAT:    return new PropertyFloatValue (*(  (float *) valueAdr));
    case DDVT_DOUBLE:   return new PropertyDoubleValue(*( (double *) valueAdr));
    default:
        throw Error("BuildPropertyValue", QString("Unknown/not-supported value type %1").arg(type));
    }
}
namespace world {
static de::Uri findTextureForShineStage(ded_shine_stage_t const &def, bool findMask) {
    if (de::Uri *resourceUri = (findMask? def.maskTexture : def.texture)) {
        try {
            return res::Textures::get()
                       .textureScheme(findMask? "Masks" : "Reflections")
                           .findByResourceUri(*resourceUri)
                               .composeUri();
        }
        catch (res::TextureScheme::NotFoundError const &)
        {} // Ignore this error.
    }
    return de::Uri();
}
ShineTextureMaterialLayer::AnimationStage::AnimationStage(de::Uri const &texture, int tics,
    float variance, de::Uri const &maskTexture, blendmode_t blendMode, float opacity,
    Vector3f const &minColor, Vector2f const &maskDimensions)
    : TextureMaterialLayer::AnimationStage(texture, tics, variance, 0, 0, Vector2f(0, 0),
                                           maskTexture, maskDimensions, blendMode, opacity)
    , minColor(minColor) {
}
ShineTextureMaterialLayer::AnimationStage::AnimationStage(AnimationStage const &other)
    : TextureMaterialLayer::AnimationStage(other)
{}
ShineTextureMaterialLayer::AnimationStage::~AnimationStage()
{}
void ShineTextureMaterialLayer::AnimationStage::resetToDefaults() {
    TextureMaterialLayer::AnimationStage::resetToDefaults();
    minColor = Vector3f();
}
ShineTextureMaterialLayer::AnimationStage *
ShineTextureMaterialLayer::AnimationStage::fromDef(ded_shine_stage_t const &def) {
    de::Uri const texture     = findTextureForShineStage(def, false/*not mask*/);
    de::Uri const maskTexture = findTextureForShineStage(def, true/*mask*/);
    return new AnimationStage(texture, def.tics, def.variance, maskTexture,
                              def.blendMode, def.shininess, Vector3f(def.minColor),
                              Vector2f(def.maskWidth, def.maskHeight));
}
ShineTextureMaterialLayer::ShineTextureMaterialLayer()
    : TextureMaterialLayer()
{}
ShineTextureMaterialLayer::~ShineTextureMaterialLayer()
{}
ShineTextureMaterialLayer *ShineTextureMaterialLayer::fromDef(ded_reflection_t const &layerDef) {
    auto *layer = new ShineTextureMaterialLayer();
    layer->_stages.append(AnimationStage::fromDef(layerDef.stage));
    return layer;
}
int ShineTextureMaterialLayer::addStage(ShineTextureMaterialLayer::AnimationStage const &stageToCopy) {
    _stages.append(new AnimationStage(stageToCopy));
    return _stages.count() - 1;
}
String ShineTextureMaterialLayer::describe() const {
    return "Shine layer";
}
} // namespace world
namespace world {
TextureMaterialLayer::AnimationStage::AnimationStage(de::Uri const &texture, int tics,
    float variance, float glowStrength, float glowStrengthVariance, Vector2f const origin,
    de::Uri const &maskTexture, Vector2f const &maskDimensions, blendmode_t blendMode, float opacity)
    : /*Record()
    , */ Stage(tics, variance)
    , texture(texture)
    , glowStrength(glowStrength)
    , glowStrengthVariance(glowStrengthVariance)
    , origin(origin)
    , maskTexture(maskTexture)
    , maskDimensions(maskDimensions)
    , blendMode(blendMode)
    , opacity(opacity) {
    /*set("origin", new ArrayValue(origin));
    set("texture", texture.compose());
    set("maskTexture", maskTexture.compose());
    set("maskDimensions", new ArrayValue(maskDimensions));
    set("blendMode", blendMode);
    set("opacity", opacity);
    set("glowStrength", glowStrength);
    set("glowStrengthVariance", glowStrengthVariance);*/
}
TextureMaterialLayer::AnimationStage::AnimationStage(AnimationStage const &other)
    : /*Record(other)
    , */Stage(other)
{}
TextureMaterialLayer::AnimationStage::~AnimationStage()
{}
void TextureMaterialLayer::AnimationStage::resetToDefaults() {
    origin = Vector2f();
    texture = de::Uri();
    maskTexture = de::Uri();
    maskDimensions = Vector2f();
    blendMode = BM_NORMAL;
    opacity = 1;
    glowStrength = 0;
    glowStrengthVariance = 0;
    /*addArray ("origin", new ArrayValue(Vector2f(0, 0)));
    addText  ("texture", "");
    addText  ("maskTexture", "");
    addArray ("maskDimensions", new ArrayValue(Vector2f(0, 0)));
    addNumber("blendMode", BM_NORMAL);
    addNumber("opacity", 1);
    addNumber("glowStrength", 0);
    addNumber("glowStrengthVariance", 0);*/
}
TextureMaterialLayer::AnimationStage *
TextureMaterialLayer::AnimationStage::fromDef(Record const &stageDef) {
    return new AnimationStage(de::makeUri(stageDef.gets("texture")),
                              stageDef.geti("tics"),
                              stageDef.getf("variance"),
                              stageDef.getf("glowStrength"),
                              stageDef.getf("glowStrengthVariance"),
                              Vector2f(stageDef.geta("texOrigin")));
}
String TextureMaterialLayer::AnimationStage::description() const {
    return "AnimationStage";
}
TextureMaterialLayer *TextureMaterialLayer::fromDef(Record const &definition) {
    defn::MaterialLayer layerDef(definition);
    auto *layer = new TextureMaterialLayer();
    for (int i = 0; i < layerDef.stageCount(); ++i) {
        layer->_stages.append(AnimationStage::fromDef(layerDef.stage(i)));
    }
    return layer;
}
int TextureMaterialLayer::addStage(TextureMaterialLayer::AnimationStage const &stageToCopy) {
    _stages.append(new AnimationStage(stageToCopy));
    return _stages.count() - 1;
}
TextureMaterialLayer::AnimationStage const &TextureMaterialLayer::stage(int index) const {
    return static_cast<AnimationStage const &>(Layer::stage(index));
}
TextureMaterialLayer::AnimationStage &TextureMaterialLayer::stage(int index) {
    return static_cast<AnimationStage &>(Layer::stage(index));
}
bool TextureMaterialLayer::hasGlow() const {
    for (int i = 0; i < stageCount(); ++i) {
        if (stage(i).glowStrength > .0001f)
            return true;
    }
    return false;
}
String TextureMaterialLayer::describe() const {
    return "TextureMaterialLayer";
}
} // namespace world
DENG2_PIMPL_NOREF(Thinker) {
    dsize size;
    thinker_s *base;    // owned
    IData *data;        // owned, optional
    Impl(AllocMethod alloc, dsize sizeInBytes, IData *data_)
        : size(max<dsize>(sizeInBytes, sizeof(thinker_s)))
        , base(0)
        , data(data_) {
        if (alloc == AllocateStandard) {
            base = reinterpret_cast<thinker_s *>(M_Calloc(size));
            base->_flags = THINKF_STD_MALLOC;
        }
        else // using memory zone
        {
            base = reinterpret_cast<thinker_s *>(Z_Calloc(size, PU_MAP, NULL));
        }
        if (data) data->setThinker(base);
    }
    Impl(Impl const &other)
        : size(other.size)
        , base(reinterpret_cast<thinker_s *>(other.base->_flags & THINKF_STD_MALLOC?
                                                 M_MemDup(other.base, size) :
                                                 Z_MemDup(other.base, size)))
        , data(other.data? other.data->duplicate() : 0) {
        base->d = data;
        if (data) data->setThinker(base);
    }
    Impl(thinker_s *podThinkerToTake, dsize sizeInBytes)
        : size(sizeInBytes)
        , base(podThinkerToTake)
        , data(reinterpret_cast<IData *>(podThinkerToTake->d)) // also take ownership of the private data
    {}
    ~Impl() {
        release();
    }
    void release() {
        if (base) {
            if (isStandardAllocated()) {
                M_Free(base);
            }
            else {
                Z_Free(base);
            }
        }
        delete data;
    }
    bool isStandardAllocated() const {
        return base && (base->_flags & THINKF_STD_MALLOC);
    }
    void relinquish() {
        base = 0;
        data = 0;
        size = 0;
    }
    static void clearBaseToZero(thinker_s *base, dsize size) {
        bool const stdAlloc = CPP_BOOL(base->_flags & THINKF_STD_MALLOC);
        memset(base, 0, size);
        if (stdAlloc) base->_flags |= THINKF_STD_MALLOC;
    }
};
#define STRUCT_MEMBER_ACCESSORS() \
      prev    (*this, offsetof(thinker_s, prev    )) \
    , next    (*this, offsetof(thinker_s, next    )) \
    , function(*this, offsetof(thinker_s, function)) \
    , id      (*this, offsetof(thinker_s, id      ))
Thinker::Thinker(dsize sizeInBytes, IData *data)
    : d(new Impl(AllocateStandard, sizeInBytes, data))
    , STRUCT_MEMBER_ACCESSORS() {
    function = Thinker_NoOperation;
}
Thinker::Thinker(AllocMethod alloc, dsize sizeInBytes, Thinker::IData *data)
    : d(new Impl(alloc, sizeInBytes, data))
    , STRUCT_MEMBER_ACCESSORS() {
    function = Thinker_NoOperation;
}
Thinker::Thinker(Thinker const &other)
    : d(new Impl(*other.d))
    , STRUCT_MEMBER_ACCESSORS()
{}
Thinker::Thinker(thinker_s const &podThinker, dsize sizeInBytes, AllocMethod alloc)
    : d(new Impl(alloc, sizeInBytes, 0))
    , STRUCT_MEMBER_ACCESSORS() {
    DENG2_ASSERT(d->size == sizeInBytes);
    memcpy(d->base, &podThinker, sizeInBytes);
    d->base->_flags &= ~THINKF_STD_MALLOC;
    if (alloc == AllocateStandard) d->base->_flags |= THINKF_STD_MALLOC;
    if (podThinker.d) {
        setData(reinterpret_cast<IData *>(podThinker.d)->duplicate());
    }
}
Thinker::Thinker(thinker_s *podThinkerToTake, de::dsize sizeInBytes)
    : d(new Impl(podThinkerToTake, sizeInBytes))
    , STRUCT_MEMBER_ACCESSORS()
{}
Thinker &Thinker::operator = (Thinker const &other) {
    d.reset(new Impl(*other.d));
    return *this;
}
void Thinker::enable(bool yes) {
    applyFlagOperation(d->base->_flags, duint32(THINKF_DISABLED), yes? SetFlags : UnsetFlags);
}
void Thinker::zap() {
    delete d->data;
    d->data = 0;
    Impl::clearBaseToZero(d->base, d->size);
}
bool Thinker::isDisabled() const {
    return (d->base->_flags & THINKF_DISABLED) != 0;
}
thinker_s &Thinker::base() {
    return *d->base;
}
thinker_s const &Thinker::base() const {
    return *d->base;
}
bool Thinker::hasData() const {
    return d->data != 0;
}
Thinker::IData &Thinker::data() {
    DENG2_ASSERT(hasData());
    return *d->data;
}
Thinker::IData const &Thinker::data() const {
    DENG2_ASSERT(hasData());
    return *d->data;
}
dsize Thinker::sizeInBytes() const {
    return d->size;
}
thinker_s *Thinker::take() {
    DENG2_ASSERT(d->base->d == d->data);
    thinker_s *th = d->base;
    d->relinquish();
    return th;
}
void Thinker::destroy(thinker_s *thinkerBase) {
    DENG2_ASSERT(thinkerBase != 0);
    release(*thinkerBase);
    if (thinkerBase->_flags & THINKF_STD_MALLOC) {
        M_Free(thinkerBase);
    }
    else {
        Z_Free(thinkerBase);
    }
}
void Thinker::release(thinker_s &thinkerBase) {
    delete reinterpret_cast<IData *>(thinkerBase.d);
    thinkerBase.d = 0;
}
void Thinker::zap(thinker_s &thinkerBase, dsize sizeInBytes) {
    delete reinterpret_cast<IData *>(thinkerBase.d);
    Impl::clearBaseToZero(&thinkerBase, sizeInBytes);
}
void Thinker::setData(Thinker::IData *data) {
    if (d->data) delete d->data;
    d->data    = data;
    d->base->d = data;
    if (data) {
        data->setThinker(*this);
    }
}
dd_bool Thinker_InStasis(thinker_s const *thinker) {
    if (!thinker) return false;
    return (thinker->_flags & THINKF_DISABLED) != 0;
}
void Thinker_SetStasis(thinker_t *thinker, dd_bool on) {
    if (thinker) {
        applyFlagOperation(thinker->_flags, duint32(THINKF_DISABLED),
                           on? SetFlags : UnsetFlags);
    }
}
void Thinker_NoOperation(void *) {
}
static QMultiHash<Id::Type, ThinkerData *> thinkerLookup;
DENG2_PIMPL(ThinkerData) {
    thinker_s *think;
    Id id; ///< Internal unique ID.
    Record names;
    Impl(Public *i, Id const &id)
        : Base(i)
        , think(0)
        , id(id)
    {}
    Impl(Public *i, Impl const &other)
        : Base(i)
        , think(other.think)
        , id(other.id)
        , names(other.names)
    {}
    ~Impl() {
        thinkerLookup.remove(id, &self());
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) {
            i->thinkerBeingDeleted(*think);
        }
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
};
DENG2_AUDIENCE_METHOD(ThinkerData, Deletion)
ThinkerData::ThinkerData(Id const &id)
    : d(new Impl(this, id)) {
    if (d->id) {
        thinkerLookup.insert(d->id, this);
    }
}
ThinkerData::ThinkerData(ThinkerData const &other) : d(new Impl(this, *other.d)) {
    if (d->id) {
        thinkerLookup.insert(d->id, this);
    }
}
Id const &ThinkerData::id() const {
    return d->id;
}
void ThinkerData::setId(Id const &id) {
    thinkerLookup.remove(d->id, this);
    thinkerLookup.insert(id, this);
    d->id = id;
}
void ThinkerData::setThinker(thinker_s *thinker) {
    d->think = thinker;
}
void ThinkerData::think() {
}
Thinker::IData *ThinkerData::duplicate() const {
    return new ThinkerData(*this);
}
thinker_s &ThinkerData::thinker() {
    DENG2_ASSERT(d->think != 0);
    return *d->think;
}
thinker_s const &ThinkerData::thinker() const {
    DENG2_ASSERT(d->think != 0);
    return *d->think;
}
Record &ThinkerData::objectNamespace() {
    return d->names;
}
Record const &ThinkerData::objectNamespace() const {
    return d->names;
}
void ThinkerData::initBindings()
{}
void ThinkerData::operator >> (Writer &to) const {
    to << world::InternalSerialId(world::THINKER_DATA)
       << d->id
       << Record(d->names, Record::IgnoreDoubleUnderscoreMembers);
}
void ThinkerData::operator << (Reader &from) {
    thinkerLookup.remove(d->id, this);
    world::InternalSerialId sid;
    from >> sid;
    switch (sid) {
    case world::THINKER_DATA:
        from >> d->id >> d->names;
        break;
    default:
        throw DeserializationError("ThinkerData::operator <<",
                                   "Invalid serial identifier " +
                                   String::number(sid));
    }
    thinkerLookup.insert(d->id, this);
}
ThinkerData *ThinkerData::find(Id const &id) {
    auto found = thinkerLookup.constFind(id);
    if (found != thinkerLookup.constEnd()) {
        return found.value();
    }
    return nullptr;
}
#ifdef DENG2_DEBUG
duint32 ThinkerData::DebugCounter::total = 0;
ThinkerData::DebugValidator ensureAllPrivateDataIsReleased;
#endif
static World *theWorld = nullptr;
DENG2_PIMPL(World) {
    BaseMap *map = nullptr;
    world::Materials materials;
    Impl(Public *i) : Base(i) {
        theWorld = thisPublic;
    }
    ~Impl() {
        theWorld = nullptr;
    }
    DENG2_PIMPL_AUDIENCE(MapChange)
};
DENG2_AUDIENCE_METHOD(World, MapChange)
World::World() : d(new Impl(this)) {
    DoomsdayApp::players().forAll([this] (Player &plr) {
        plr.setWorld(this);
        return LoopContinue;
    });
}
void World::reset() {
    DoomsdayApp::players().forAll([] (Player &plr) {
        ddplayer_t &ddpl = plr.publicData();
        ddpl.mo            = nullptr;
        ddpl.extraLight    = 0;
        ddpl.fixedColorMap = 0;
        ddpl.flags         &= ~DDPF_CAMERA;
        for (ddpsprite_t &pspr : ddpl.pSprites) {
            pspr.statePtr = nullptr;
        }
        return LoopContinue;
    });
}
void World::timeChanged(Clock const &) {
}
void World::setMap(BaseMap *map) {
    d->map = map;
}
bool World::hasMap() const {
    return d->map != nullptr;
}
BaseMap &World::map() const {
    DENG2_ASSERT(hasMap());
    return *d->map;
}
Materials &World::materials() {
    return d->materials;
}
Materials const &World::materials() const {
    return d->materials;
}
World &World::get() {
    DENG2_ASSERT(theWorld);
    return *theWorld;
}
void World::notifyMapChange() {
    DENG2_FOR_AUDIENCE2(MapChange, i) i->worldMapChanged();
}
static xgclass_t nullXgClassLinks;  ///< Used when none defined.
static xgclass_t *xgClassLinks;
void XG_GetGameClasses() {
    ::xgClassLinks = nullptr;
    if (auto getVar = DoomsdayApp::plugins().gameExports().GetPointer) {
        ::xgClassLinks = (xgclass_t *) getVar(DD_XGFUNC_LINK);
    }
    if (!::xgClassLinks) {
        ::xgClassLinks = &::nullXgClassLinks;
    }
    DED_SetXGClassLinks(::xgClassLinks);
}
xgclass_t *XG_Class(int number) {
    return &::xgClassLinks[number];
}
