﻿/** @file buttonwidget.cpp  Button widget.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/buttonwidget.h"
#include "g_defs.h"
#include "hu_menu.h"   // menu sounds
#include "hu_stuff.h"
#include "menu/page.h" // mnRendState
/** @file coloreditwidget.cpp  UI widget for editing a color.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/coloreditwidget.h"
#include "hu_menu.h"   // menu sounds
#include "hu_stuff.h"  // borderPatches
#include "menu/page.h" // mnRendState
/** @file cvarcoloreditwidget.cpp  UI widget for editing a color in a cvar.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvarcoloreditwidget.h"
#include "hu_menu.h" // Hu_MenuDefaultFocusAction
#include "menu/page.h" // mnRendState
/** @file cvarinlinelistwidget.cpp  UI widget for a selectable, inline list of items.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvarinlinelistwidget.h"
#include "hu_menu.h" // Hu_MenuDefaultFocusAction
/** @file cvarlineeditwidget.cpp  UI widget for an editable line of text in a cvar.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvarlineeditwidget.h"
#include "hu_menu.h" // Hu_MenuDefaultFocusAction
/** @file cvarsliderwidget.cpp  UI widget for manipulating a cvar with a graphical slider.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvarsliderwidget.h"
#include "hu_menu.h" // Hu_MenuDefaultFocusAction
/** @file cvartextualsliderwidget.cpp  UI widget for a textual slider.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvartextualsliderwidget.h"
#include "hu_menu.h" // Hu_MenuMergeEffectWithDrawTextFlags
#include "menu/page.h" // mnRendState
/** @file cvartogglewidget.cpp  Button widget for toggling cvars.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/cvartogglewidget.h"
#include "hu_menu.h" // Hu_MenuDefaultFocusAction
/** @file inlinelistwidget.cpp  UI widget for a selectable, inline-list of items.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QtAlgorithms>
#include "common.h"
#include "menu/widgets/inlinelistwidget.h"
#include "hu_menu.h" // menu sounds
#include "menu/page.h" // mnRendState
/** @file inputbindingwidget.cpp  UI widget for manipulating input bindings.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/inputbindingwidget.h"
#include "hu_menu.h"   // menu sounds
#include "hu_stuff.h"
#include "m_ctrl.h"    // controlconfig_t, etc..
#include "menu/page.h" // mnRendState
#include <map>
#include <tuple>
/** @file labelwidget.cpp  Text label widget.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/labelwidget.h"
#include "hu_menu.h"   // Hu_MenuMergeEffectWithDrawTextFlags
#include "hu_stuff.h"
#include "menu/page.h" // mnRendState
/** @file lineeditwidget.cpp  UI widget for an editable line of text.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/lineeditwidget.h"
#include "hu_menu.h"   // menu sounds
#include "hu_stuff.h"  // shiftXForm
#include "menu/page.h" // mnRendState
/** @file listwidget.cpp  UI widget for a selectable list of items.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QtAlgorithms>
#include "common.h"
#include "menu/widgets/listwidget.h"
#include "hu_menu.h" // menu sounds
#include "menu/page.h"
/** @file mobjpreviewwidget.cpp  UI widget for previewing a mobj.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/mobjpreviewwidget.h"
#include "hu_menu.h" // menuTime
#include "menu/page.h" // mnRendState
/** @file rectwidget.cpp  Simple rectangular widget with a background image.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/rectwidget.h"
#include "menu/page.h" // mnRendState
/** @file sliderwidget.cpp  UI widget for a graphical slider.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "menu/widgets/sliderwidget.h"
#include "hu_menu.h" // menu sounds
#include "hu_stuff.h" // M_DrawGlowBar, etc..
#include "menu/page.h" // mnRendState
/** @file widget.cpp  Base class for widgets.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QMap>
#include "common.h"
#include "menu/widgets/widget.h"
#include "menu/page.h"
#include "g_defs.h"
#include "hu_menu.h" // menu sounds
using namespace de;
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(ButtonWidget) {
    String    text;           ///< Label text.
    patchid_t patch     = -1; ///< Used when drawing this instead of text, if set.
    bool      noAltText = false;
    bool      silent    = false;
};
ButtonWidget::ButtonWidget(String const &text, patchid_t patch)
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT2);
    setColor(MENU_COLOR1);
    setText(text);
    setPatch(patch);
}
ButtonWidget::~ButtonWidget()
{}
void ButtonWidget::draw() const {
    fontid_t const fontId     = mnRendState->textFonts[font()];
    Vector4f const &textColor = mnRendState->textColors[color()];
    const Vector4f color = selectionFlashColor(textColor);
    const float fadeout = scrollingFadeout();
    if (fadeout < .001f) return;
    FR_SetFont(fontId);
    FR_SetColorAndAlpha(color.x, color.y, color.z, color.w * fadeout);
    DGL_Color4f(1, 1, 1, color.w * fadeout);
    if(d->patch >= 0) {
        String replacement;
        if(!d->noAltText) {
            replacement = Hu_ChoosePatchReplacement(patchreplacemode_t(cfg.common.menuPatchReplaceMode), d->patch, d->text);
        }
        DGL_Enable(DGL_TEXTURE_2D);
        WI_DrawPatch(d->patch,
                     replacement,
                     geometry().topLeft,
                     ALIGN_TOPLEFT,
                     0,
                     Hu_MenuMergeEffectWithDrawTextFlags(0));
        DGL_Disable(DGL_TEXTURE_2D);
        return;
    }
    DGL_Enable(DGL_TEXTURE_2D);
    FR_DrawTextXY3(d->text.toUtf8().constData(), geometry().topLeft.x, geometry().topLeft.y,
                   ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
    DGL_Disable(DGL_TEXTURE_2D);
}
int ButtonWidget::handleCommand(menucommand_e cmd) {
    if (cmd == MCMD_SELECT) {
        if (!isActive()) {
            setFlags(Active);
            execAction(Activated);
        }
        if (!d->silent) {
            S_LocalSound(SFX_MENU_ACCEPT, NULL);
        }
        setFlags(Active, UnsetFlags);
        execAction(Deactivated);
        return true;
    }
    return false; // Not eaten.
}
void ButtonWidget::setSilent(bool silent) {
    d->silent = silent;
}
void ButtonWidget::updateGeometry() {
    String useText = d->text;
    if(d->patch >= 0) {
        if(!d->noAltText) {
            useText = Hu_ChoosePatchReplacement(patchreplacemode_t(cfg.common.menuPatchReplaceMode), d->patch, d->text);
        }
        if(useText.isEmpty()) {
            patchinfo_t info;
            R_GetPatchInfo(d->patch, &info);
            geometry().setSize(Vector2ui(info.geometry.size.width, info.geometry.size.height));
            return;
        }
    }
    FR_PushAttrib();
    Size2Raw size;
    FR_SetFont(page().predefinedFont(mn_page_fontid_t(font())));
    FR_TextSize(&size, useText.toUtf8().constData());
    geometry().setSize(Vector2ui(size.width, size.height));
    FR_PopAttrib();
}
String ButtonWidget::text() const {
    return d->text;
}
ButtonWidget &ButtonWidget::setText(String const &newText) {
    d->text = labelText(newText);
    return *this;
}
patchid_t ButtonWidget::patch() const {
    return d->patch;
}
ButtonWidget &ButtonWidget::setPatch(patchid_t newPatch) {
    d->patch = newPatch;
    return *this;
}
bool ButtonWidget::noAltText() const {
    return d->noAltText;
}
ButtonWidget &ButtonWidget::setNoAltText(bool yes) {
    d->noAltText = yes;
    return *this;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL(ColorEditWidget) {
    bool rgbaMode       = false;
    Vector4f color      = Vector4f(0, 0, 0, 1.f);
    Vector2i dimensions = Vector2i(MNDATA_COLORBOX_WIDTH, MNDATA_COLORBOX_HEIGHT);  ///< Inner dimensions in fixed 320x200 space.
    Impl(Public *i) : Base(i) {}
    bool setRed(float red, int flags) {
        float const oldRed = color.x;
        color.x = red;
        if(color.x != oldRed) {
            if(!(flags & MNCOLORBOX_SCF_NO_ACTION)) {
                self().execAction(Modified);
            }
            return true;
        }
        return false;
    }
    bool setGreen(float green, int flags) {
        float const oldGreen = color.y;
        color.y = green;
        if(color.y != oldGreen) {
            if(!(flags & MNCOLORBOX_SCF_NO_ACTION)) {
                self().execAction(Modified);
            }
            return true;
        }
        return false;
    }
    bool setBlue(float blue, int flags) {
        float const oldBlue = color.z;
        color.z = blue;
        if(color.z != oldBlue) {
            if(!(flags & MNCOLORBOX_SCF_NO_ACTION)) {
                self().execAction(Modified);
            }
            return true;
        }
        return false;
    }
    bool setAlpha(float alpha, int flags) {
        if(rgbaMode) {
            float const oldAlpha = color.w;
            color.w = alpha;
            if(color.w != oldAlpha) {
                if(!(flags & MNCOLORBOX_SCF_NO_ACTION)) {
                    self().execAction(Modified);
                }
                return true;
            }
        }
        return false;
    }
};
ColorEditWidget::ColorEditWidget(Vector4f const &color, bool rgbaMode)
    : Widget()
    , d(new Impl(this)) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
    d->rgbaMode = rgbaMode;
    d->color    = color;
    if(!d->rgbaMode) d->color.w = 1.f;
}
ColorEditWidget::~ColorEditWidget()
{}
void ColorEditWidget::draw() const {
    patchinfo_t t, b, l, r, tl, tr, br, bl;
    int const up = 1;
    R_GetPatchInfo(borderPatches[0], &t);
    R_GetPatchInfo(borderPatches[2], &b);
    R_GetPatchInfo(borderPatches[3], &l);
    R_GetPatchInfo(borderPatches[1], &r);
    R_GetPatchInfo(borderPatches[4], &tl);
    R_GetPatchInfo(borderPatches[5], &tr);
    R_GetPatchInfo(borderPatches[6], &br);
    R_GetPatchInfo(borderPatches[7], &bl);
    int x = geometry().topLeft.x;
    int y = geometry().topLeft.y;
    int w = d->dimensions.x;
    int h = d->dimensions.y;
    if(t.id || tl.id || tr.id) {
        int height = 0;
        if(t.id)  height = t.geometry.size.height;
        if(tl.id) height = de::max(height, tl.geometry.size.height);
        if(tr.id) height = de::max(height, tr.geometry.size.height);
        y += height;
    }
    if(l.id || tl.id || bl.id) {
        int width = 0;
        if(l.id)  width = l.geometry.size.width;
        if(tl.id) width = de::max(width, tl.geometry.size.width);
        if(bl.id) width = de::max(width, bl.geometry.size.width);
        x += width;
    }
    const float fadeout = scrollingFadeout();
    DGL_Color4f(1, 1, 1, mnRendState->pageAlpha * fadeout);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_SetMaterialUI((world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUriCString(borderGraphics[0])), DGL_REPEAT, DGL_REPEAT);
    DGL_DrawRectf2Tiled(x, y, w, h, 64, 64);
    if(t.id) {
        DGL_SetPatch(t.id, DGL_REPEAT, DGL_REPEAT);
        DGL_DrawRectf2Tiled(x, y - t.geometry.size.height, w, t.geometry.size.height, up * t.geometry.size.width, up * t.geometry.size.height);
    }
    if(b.id) {
        DGL_SetPatch(b.id, DGL_REPEAT, DGL_REPEAT);
        DGL_DrawRectf2Tiled(x, y + h, w, b.geometry.size.height, up * b.geometry.size.width, up * b.geometry.size.height);
    }
    if(l.id) {
        DGL_SetPatch(l.id, DGL_REPEAT, DGL_REPEAT);
        DGL_DrawRectf2Tiled(x - l.geometry.size.width, y, l.geometry.size.width, h, up * l.geometry.size.width, up * l.geometry.size.height);
    }
    if(r.id) {
        DGL_SetPatch(r.id, DGL_REPEAT, DGL_REPEAT);
        DGL_DrawRectf2Tiled(x + w, y, r.geometry.size.width, h, up * r.geometry.size.width, up * r.geometry.size.height);
    }
    if(tl.id) {
        DGL_SetPatch(tl.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(x - tl.geometry.size.width, y - tl.geometry.size.height, tl.geometry.size.width, tl.geometry.size.height);
    }
    if(tr.id) {
        DGL_SetPatch(tr.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(x + w, y - tr.geometry.size.height, tr.geometry.size.width, tr.geometry.size.height);
    }
    if(br.id) {
        DGL_SetPatch(br.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(x + w, y + h, br.geometry.size.width, br.geometry.size.height);
    }
    if(bl.id) {
        DGL_SetPatch(bl.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(x - bl.geometry.size.width, y + h, bl.geometry.size.width, bl.geometry.size.height);
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_SetNoMaterial();
    DGL_DrawRectf2Color(x,
                        y,
                        w,
                        h,
                        d->color.x,
                        d->color.y,
                        d->color.z,
                        d->color.w * mnRendState->pageAlpha * fadeout);
}
int ColorEditWidget::handleCommand(menucommand_e cmd) {
    if(cmd == MCMD_SELECT) {
        S_LocalSound(SFX_MENU_CYCLE, NULL);
        if(!isActive()) {
            setFlags(Active);
            execAction(Activated);
        }
        else {
            setFlags(Active, UnsetFlags);
            execAction(Deactivated);
        }
        return true;
    }
    return false; // Not eaten.
}
void ColorEditWidget::updateGeometry() {
    patchinfo_t info;
    geometry().setSize(d->dimensions.toVector2ui());
    if(R_GetPatchInfo(borderPatches[2], &info)) {
        info.geometry.size.width = d->dimensions.x;
        info.geometry.origin.y   = d->dimensions.y;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[1], &info)) {
        info.geometry.size.height = d->dimensions.y;
        info.geometry.origin.x    = d->dimensions.x;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[0], &info)) {
        info.geometry.size.width = d->dimensions.x;
        info.geometry.origin.y   = -info.geometry.size.height;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[3], &info)) {
        info.geometry.size.height = d->dimensions.y;
        info.geometry.origin.x    = -info.geometry.size.width;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[4], &info)) {
        info.geometry.origin.x = -info.geometry.size.width;
        info.geometry.origin.y = -info.geometry.size.height;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[5], &info)) {
        info.geometry.origin.x = d->dimensions.x;
        info.geometry.origin.y = -info.geometry.size.height;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[6], &info)) {
        info.geometry.origin.x = d->dimensions.x;
        info.geometry.origin.y = d->dimensions.y;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(borderPatches[7], &info)) {
        info.geometry.origin.x = -info.geometry.size.width;
        info.geometry.origin.y = d->dimensions.y;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
}
ColorEditWidget &ColorEditWidget::setPreviewDimensions(Vector2i const &newDimensions) {
    d->dimensions = newDimensions;
    return *this;
}
Vector2i ColorEditWidget::previewDimensions() const {
    return d->dimensions;
}
bool ColorEditWidget::rgbaMode() const {
    return d->rgbaMode;
}
ColorEditWidget &ColorEditWidget::setColor(Vector4f const &newColor, int flags) {
    int setComps = 0;
    int const setCompFlags = (flags | MNCOLORBOX_SCF_NO_ACTION);
    if(d->setRed  (newColor.x, setCompFlags)) setComps |= 0x1;
    if(d->setGreen(newColor.y, setCompFlags)) setComps |= 0x2;
    if(d->setBlue (newColor.z, setCompFlags)) setComps |= 0x4;
    if(d->setAlpha(newColor.w, setCompFlags)) setComps |= 0x8;
    if(setComps && !(flags & MNCOLORBOX_SCF_NO_ACTION)) {
        execAction(Modified);
    }
    return *this;
}
Vector4f ColorEditWidget::color() const {
    if(!d->rgbaMode) {
        return Vector4f(d->color, 1.0f);
    }
    return d->color;
}
ColorEditWidget &ColorEditWidget::setRed(float newRed, int flags) {
    d->setRed(newRed, flags);
    return *this;
}
ColorEditWidget &ColorEditWidget::setGreen(float newGreen, int flags) {
    d->setGreen(newGreen, flags);
    return *this;
}
ColorEditWidget &ColorEditWidget::setBlue(float newBlue, int flags) {
    d->setBlue(newBlue, flags);
    return *this;
}
ColorEditWidget &ColorEditWidget::setAlpha(float newAlpha, int flags) {
    d->setAlpha(newAlpha, flags);
    return *this;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
CVarColorEditWidget::CVarColorEditWidget(char const *redCVarPath, char const *greenCVarPath,
    char const *blueCVarPath, char const *alphaCVarPath, Vector4f const &color, bool rgbaMode)
    : ColorEditWidget(color, rgbaMode) {
    setAction(Modified,    CVarColorEditWidget_UpdateCVar);
    setAction(FocusGained, Hu_MenuDefaultFocusAction);
    _cvarPaths[0] = redCVarPath;
    _cvarPaths[1] = greenCVarPath;
    _cvarPaths[2] = blueCVarPath;
    _cvarPaths[3] = alphaCVarPath;
}
CVarColorEditWidget::~CVarColorEditWidget()
{}
char const *CVarColorEditWidget::cvarPath(int component) const {
    if(component >= 0 && component < 4) {
        return _cvarPaths[component];
    }
    return 0;
}
void CVarColorEditWidget_UpdateCVar(Widget &wi, Widget::Action action) {
    CVarColorEditWidget *cbox = &wi.as<CVarColorEditWidget>();
    if(action != Widget::Modified) return;
    Con_SetFloat2(cbox->redCVarPath(),   cbox->red(),   SVF_WRITE_OVERRIDE);
    Con_SetFloat2(cbox->greenCVarPath(), cbox->green(), SVF_WRITE_OVERRIDE);
    Con_SetFloat2(cbox->blueCVarPath(),  cbox->blue(),  SVF_WRITE_OVERRIDE);
    if(cbox->rgbaMode()) {
        Con_SetFloat2(cbox->alphaCVarPath(), cbox->alpha(), SVF_WRITE_OVERRIDE);
    }
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
CVarInlineListWidget::CVarInlineListWidget(char const *cvarPath, int cvarValueMask)
    : InlineListWidget()
    , _cvarPath(cvarPath)
    , _cvarValueMask(cvarValueMask) {
    setColor(MENU_COLOR3);
    setAction(Modified,    CVarInlineListWidget_UpdateCVar);
    setAction(FocusGained, Hu_MenuDefaultFocusAction);
}
CVarInlineListWidget::~CVarInlineListWidget()
{}
char const *CVarInlineListWidget::cvarPath() const {
    return _cvarPath;
}
int CVarInlineListWidget::cvarValueMask() const {
    return _cvarValueMask;
}
void CVarInlineListWidget_UpdateCVar(Widget &wi, Widget::Action action) {
    CVarInlineListWidget const *list = &wi.as<CVarInlineListWidget>();
    if(action != Widget::Modified) return;
    if(list->selection() < 0) return; // Hmm?
    cvartype_t varType = Con_GetVariableType(list->cvarPath());
    if(CVT_NULL == varType) return;
    ListWidget::Item const *item = list->items()[list->selection()];
    int value;
    if(list->cvarValueMask()) {
        value = Con_GetInteger(list->cvarPath());
        value = (value & ~list->cvarValueMask()) | (item->userValue() & list->cvarValueMask());
    }
    else {
        value = item->userValue();
    }
    switch(varType) {
    case CVT_INT:
        Con_SetInteger2(list->cvarPath(), value, SVF_WRITE_OVERRIDE);
        break;
    case CVT_BYTE:
        Con_SetInteger2(list->cvarPath(), (byte) value, SVF_WRITE_OVERRIDE);
        break;
    default: Con_Error("CVarInlineListWidget_UpdateCVar: Unsupported variable type %i", (int)varType);
    }
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
CVarLineEditWidget::CVarLineEditWidget(char const *cvarPath)
    : LineEditWidget()
    , _cvarPath(cvarPath) {
    setAction(Modified,    CVarLineEditWidget_UpdateCVar);
    setAction(FocusGained, Hu_MenuDefaultFocusAction);
}
CVarLineEditWidget::~CVarLineEditWidget()
{}
char const *CVarLineEditWidget::cvarPath() const {
    return _cvarPath;
}
void CVarLineEditWidget_UpdateCVar(Widget &wi, Widget::Action action) {
    CVarLineEditWidget const &edit = wi.as<CVarLineEditWidget>();
    cvartype_t varType = Con_GetVariableType(edit.cvarPath());
    if(action != Widget::Modified) return;
    switch(varType) {
    case CVT_CHARPTR:
        Con_SetString2(edit.cvarPath(), edit.text().toUtf8().constData(), SVF_WRITE_OVERRIDE);
        break;
    case CVT_URIPTR: {
        de::Uri uri(edit.text(), RC_NULL);
        Con_SetUri2(edit.cvarPath(), reinterpret_cast<uri_s *>(&uri), SVF_WRITE_OVERRIDE);
        break; }
    default: break;
    }
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
CVarSliderWidget::CVarSliderWidget(char const *cvarPath, float min, float max, float step, bool floatMode)
    : SliderWidget(min, max, step, floatMode)
    , _cvarPath(cvarPath) {
    setAction(Modified,    CVarSliderWidget_UpdateCVar);
    setAction(FocusGained, Hu_MenuDefaultFocusAction);
}
CVarSliderWidget::~CVarSliderWidget()
{}
char const *CVarSliderWidget::cvarPath() const {
    return _cvarPath;
}
void CVarSliderWidget_UpdateCVar(Widget &wi, Widget::Action action) {
    if(action != Widget::Modified) return;
    CVarSliderWidget &sldr = wi.as<CVarSliderWidget>();
    cvartype_t varType = Con_GetVariableType(sldr.cvarPath());
    if(CVT_NULL == varType) return;
    float value = sldr.value();
    switch(varType) {
    case CVT_FLOAT:
        if(sldr.step() >= .01f) {
            Con_SetFloat2(sldr.cvarPath(), (int) (100 * value) / 100.0f, SVF_WRITE_OVERRIDE);
        }
        else {
            Con_SetFloat2(sldr.cvarPath(), value, SVF_WRITE_OVERRIDE);
        }
        break;
    case CVT_INT:
        Con_SetInteger2(sldr.cvarPath(), (int) value, SVF_WRITE_OVERRIDE);
        break;
    case CVT_BYTE:
        Con_SetInteger2(sldr.cvarPath(), (byte) value, SVF_WRITE_OVERRIDE);
        break;
    default: break;
    }
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL(CVarTextualSliderWidget) {
    String onethSuffix;
    String nthSuffix;
    String emptyText;
    Impl(Public *i) : Base(i) {}
    inline bool valueIsOne(float value) {
        if(self().floatMode()) {
            return INRANGE_OF(1, value, .0001f);
        }
        return (value > 0 && 1 == int(value + .5f));
    }
    String chooseSuffix(float value) {
        if(!onethSuffix.isEmpty() && valueIsOne(value)) {
            return onethSuffix;
        }
        if(!nthSuffix.isEmpty()) {
            return nthSuffix;
        }
        return "";
    }
    String composeTextualValue(float value, int precision = 0) {
        if(self().floatMode() && !valueIsOne(value)) {
            return String::number(value, 'f', de::max(0, precision));
        }
        else {
            return String::number(int(value));
        }
    }
    String valueAsText(float emptyValue = 0, int precision = 0) {
        float const value = de::clamp(self().min(), self().value(), self().max()); /// @todo clamp necessary?
        if(!emptyText.isEmpty() && INRANGE_OF(value, emptyValue, .0001f)) {
            return emptyText;
        }
        String const suffix      = chooseSuffix(value);
        String const valueAsText = composeTextualValue(value, precision);
#if 0
        if(!templateString.isEmpty()) {
            AutoStr *compStr = AutoStr_NewStd();
            Str_Reserve(compStr, bufSize);
            char const *c;
            char *beginSubstring = templateString;
            for(c = beginSubstring; *c; c++) {
                if(c[0] == '%' && c[1] == '1') {
                    Str_PartAppend(compStr, beginSubstring, 0, c - beginSubstring);
                    Str_Appendf(compStr, "%s%s", textualValue, suffix);
                    beginSubstring = c + 2;
                    c += 1;
                }
            }
            if(beginSubstring != c)
                Str_Append(compStr, beginSubstring);
            return String(Str_Text(compStr));
        }
        else
#endif
        {
            return valueAsText + suffix;
        }
    }
};
CVarTextualSliderWidget::CVarTextualSliderWidget(char const *cvarPath, float min, float max,
                                                 float step, bool floatMode)
    : CVarSliderWidget(cvarPath, min, max, step, floatMode)
    , d(new Impl(this)) {
    setColor(MENU_COLOR3);
}
CVarTextualSliderWidget::~CVarTextualSliderWidget()
{}
void CVarTextualSliderWidget::draw() const {
    Vector2i const &origin    = geometry().topLeft;
    String const valueAsText  = d->valueAsText();
    Vector4f const &textColor = mnRendState->textColors[color()];
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(origin.x, origin.y, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(mnRendState->textFonts[font()]);
    FR_SetColorAndAlpha(textColor.x, textColor.y, textColor.z, textColor.w * scrollingFadeout());
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), 0, 0, ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(-origin.x, -origin.y, 0);
}
void CVarTextualSliderWidget::updateGeometry() {
    String const valueAsText = d->valueAsText();
    FR_PushAttrib();
    FR_SetFont(page().predefinedFont(mn_page_fontid_t(font())));
    Size2Raw size; FR_TextSize(&size, valueAsText.toUtf8().constData());
    geometry().setSize(Vector2ui(size.width, size.height));
    FR_PopAttrib();
}
CVarTextualSliderWidget &CVarTextualSliderWidget::setEmptyText(String const &newEmptyText) {
    d->emptyText = newEmptyText;
    return *this;
}
String CVarTextualSliderWidget::emptyText() const {
    return d->emptyText;
}
CVarTextualSliderWidget &CVarTextualSliderWidget::setOnethSuffix(String const &newOnethSuffix) {
    d->onethSuffix = newOnethSuffix;
    return *this;
}
String CVarTextualSliderWidget::onethSuffix() const {
    return d->onethSuffix;
}
CVarTextualSliderWidget &CVarTextualSliderWidget::setNthSuffix(String const &newNthSuffix) {
    d->nthSuffix = newNthSuffix;
    return *this;
}
String CVarTextualSliderWidget::nthSuffix() const {
    return d->nthSuffix;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(CVarToggleWidget) {
    State       state         = Up;
    const char *cvarPath      = nullptr;
    int         cvarValueMask = 0;
    String      downText;
    String      upText;
    std::function<void(State)> stateChangeCallback;
};
CVarToggleWidget::CVarToggleWidget(char const *cvarPath, int cvarValueMask,
                                   String const &downText, String const &upText)
    : ButtonWidget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR3);
    setAction(Modified,    CVarToggleWidget_UpdateCVar);
    setAction(FocusGained, Hu_MenuDefaultFocusAction);
    d->cvarPath      = cvarPath;
    d->cvarValueMask = cvarValueMask;
    setDownText(downText);
    setUpText(upText);
}
CVarToggleWidget::~CVarToggleWidget()
{}
int CVarToggleWidget::handleCommand(menucommand_e cmd) {
    if(cmd == MCMD_SELECT) {
        bool justActivated = false;
        if(!isActive()) {
            justActivated = true;
            S_LocalSound(SFX_MENU_CYCLE, NULL);
            setFlags(Active);
            execAction(Activated);
        }
        if(!justActivated) {
            setFlags(Active, isActive()? UnsetFlags : SetFlags);
        }
        setState(isActive()? Down : Up);
        execAction(Modified);
        if(!justActivated && !isActive()) {
            S_LocalSound(SFX_MENU_CYCLE, NULL);
            execAction(Deactivated);
        }
        return true;
    }
    return false; // Not eaten.
}
void CVarToggleWidget_UpdateCVar(Widget &wi, Widget::Action action) {
    CVarToggleWidget *tog = &wi.as<CVarToggleWidget>();
    if(action != Widget::Modified) return;
    tog->setText(tog->isDown()? tog->downText() : tog->upText());
    if(Con_GetVariableType(tog->cvarPath()) == CVT_NULL) {
        return;
    }
    int value;
    if(int const valueMask = tog->cvarValueMask()) {
        value = Con_GetInteger(tog->cvarPath());
        if(tog->isDown()) {
            value |= valueMask;
        }
        else {
            value &= ~valueMask;
        }
    }
    else {
        value = int(tog->state());
    }
    Con_SetInteger2(tog->cvarPath(), value, SVF_WRITE_OVERRIDE);
}
void CVarToggleWidget::setState(State newState) {
    if (newState != d->state) {
        d->state = newState;
        if (d->stateChangeCallback) {
            d->stateChangeCallback(newState);
        }
    }
}
CVarToggleWidget::State CVarToggleWidget::state() const {
    return d->state;
}
char const *CVarToggleWidget::cvarPath() const {
    return d->cvarPath;
}
int CVarToggleWidget::cvarValueMask() const {
    return d->cvarValueMask;
}
void CVarToggleWidget::setDownText(String const &newDownText) {
    d->downText = newDownText;
}
String CVarToggleWidget::downText() const {
    return d->downText;
}
void CVarToggleWidget::setUpText(String const &newUpText) {
    d->upText = newUpText;
}
String CVarToggleWidget::upText() const {
    return d->upText;
}
void CVarToggleWidget::setStateChangeCallback(const std::function<void (State)> &stateChanged) {
    d->stateChangeCallback = stateChanged;
}
void CVarToggleWidget::pageActivated() {
    ButtonWidget::pageActivated();
    setFlags(Active, isDown()? SetFlags : UnsetFlags);
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
InlineListWidget::InlineListWidget()
    : ListWidget()
{}
InlineListWidget::~InlineListWidget()
{}
void InlineListWidget::draw() const {
    const Item *   item  = items()[selection()];
    const Vector4f color = selectionFlashColor(mnRendState->textColors[this->color()]);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(mnRendState->textFonts[font()]);
    FR_SetColorAndAlpha(color.x, color.y, color.z, color.w * scrollingFadeout());
    FR_DrawTextXY3(item->text().toUtf8().constData(), geometry().topLeft.x, geometry().topLeft.y,
                   ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
    DGL_Disable(DGL_TEXTURE_2D);
}
int InlineListWidget::handleCommand(menucommand_e cmd) {
    switch(cmd) {
    case MCMD_SELECT: // Treat as @c MCMD_NAV_RIGHT
    case MCMD_NAV_LEFT:
    case MCMD_NAV_RIGHT: {
        int oldSelection = selection();
        if(MCMD_NAV_LEFT == cmd) {
            if(selection() > 0)
                selectItem(selection() - 1);
            else
                selectItem(itemCount() - 1);
        }
        else {
            if(selection() < itemCount() - 1)
                selectItem(selection() + 1);
            else
                selectItem(0);
        }
        updateVisibleSelection();
        if(oldSelection != selection()) {
            S_LocalSound(SFX_MENU_SLIDER_MOVE, NULL);
            execAction(Modified);
        }
        return true;
      }
    default:
        return false; // Not eaten.
    }
}
void InlineListWidget::updateGeometry() {
    FR_PushAttrib();
    FR_SetFont(page().predefinedFont(mn_page_fontid_t(font())));
    Size2Raw maxSize{};
    Size2Raw size;
    for (int i = 0; i < items().size(); ++i) {
        FR_TextSize(&size, items().at(i)->text().toUtf8().constData());
        maxSize.width = de::max(maxSize.width, size.width);
        maxSize.height = de::max(maxSize.height, size.height);
    }
    geometry().setSize(Vector2ui(maxSize.width, maxSize.height));
    FR_PopAttrib();
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
#define SMALL_SCALE .75f
#define MIBF_IGNORE_REPEATS     0x1
enum bindingitertype_t {
    MIBT_KEY,
    MIBT_MOUSE,
    MIBT_JOY
};
static char const *findInString(char const *str, char const *token, int n) {
    int tokenLen = strlen(token);
    char const *at = strstr(str, token);
    if(!at) return 0;
    if(at - str <= n - tokenLen) {
        return at;
    }
    return 0;
}
static void drawSmallText(char const *string, int x, int y, float alpha) {
    int height = FR_TextHeight(string);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(x, y + height/2, 0);
    DGL_Scalef(SMALL_SCALE, SMALL_SCALE, 1);
    DGL_Translatef(-x, -y - height/2, 0);
    FR_SetColorAndAlpha(1, 1, 1, alpha);
    FR_DrawTextXY3(string, x, y, ALIGN_TOPLEFT, DTF_NO_EFFECTS);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
DENG2_PIMPL(InputBindingWidget) {
    bool needGeometry = true; ///< Recalculate geometry based on bindings.
    const int maxWidth = SCREENWIDTH * 55 / 100;
    Impl(Public *i) : Base(i)
    {}
    void iterateBindings(
        int flags,
        const std::function<
            void(bindingitertype_t type, int bid, char const *ev, dd_bool isInverse)>
            &callback) const {
        std::map<bindingitertype_t, std::list<std::tuple<int, std::string, bool>>> bindings;
        const controlconfig_t *binds = self().binds;
        DENG2_ASSERT(binds != nullptr);
        char bindingsBuf[1024];
        if (binds->controlName) {
            B_BindingsForControl(
                0, binds->controlName, BFCI_BOTH, bindingsBuf, sizeof(bindingsBuf));
        }
        else {
            B_BindingsForCommand(binds->command, bindingsBuf, sizeof(bindingsBuf));
        }
        const char *ptr = strchr(bindingsBuf, ':');
        const char *begin, *end, *end2, *k, *bindingStart, *bindingEnd;
        char buf[80], *b;
        dd_bool isInverse;
        int bid;
        std::memset(buf, 0, sizeof(buf));
        while (ptr) {
            for (k = ptr; k > bindingsBuf && *k != '@'; --k) {;}
            if (*k == '@') {
                for (begin = k - 1; begin > bindingsBuf && isdigit(*(begin - 1)); --begin) {;}
                bid = strtol(begin, NULL, 10);
            }
            else {
                bid = 0;
            }
            bindingStart = k + 1;
            bindingEnd   = strchr(bindingStart, '@');
            if (!bindingEnd) {
                bindingEnd = strchr(k + 1, 0);
            }
            ptr++;
            end = strchr(ptr, '-');
            if (!end) return;
            end++;
            b = buf;
            while (*end && *end != ' ' && *end != '-' && *end != '+') {
                *b++ = *end++;
            }
            *b = 0;
            end2 = strchr(end, ' ');
            if (!end2) {
                end = end + strlen(end); // Then point to the end.
            }
            else {
                end = end2;
            }
            if (!findInString(bindingStart, "modifier-1-down", bindingEnd - bindingStart) &&
                (!(flags & MIBF_IGNORE_REPEATS) || !findInString(ptr, "-repeat", end - ptr))) {
                isInverse = (findInString(ptr, "-inverse", end - ptr) != NULL);
                if (!strncmp(ptr, "key", 3) || !strncmp(ptr, "joy-button", 10) ||
                    !strncmp(ptr, "mouse-left", 10) || !strncmp(ptr, "mouse-middle", 12) ||
                    !strncmp(ptr, "mouse-right", 11)) {
                    if (((binds->flags & CCF_INVERSE) && isInverse) ||
                        ((binds->flags & CCF_NON_INVERSE) && !isInverse) ||
                        !(binds->flags & (CCF_INVERSE | CCF_NON_INVERSE))) {
                        bindings[!strncmp(ptr, "key", 3)
                                     ? MIBT_KEY
                                     : !strncmp(ptr, "mouse", 5) ? MIBT_MOUSE : MIBT_JOY]
                            .emplace_back(bid, buf, isInverse);
                    }
                }
                else {
                    if (!(binds->flags & (CCF_INVERSE | CCF_NON_INVERSE)) ||
                        (binds->flags & CCF_INVERSE)) {
                        isInverse = !isInverse;
                    }
                    if (!strncmp(ptr, "joy", 3)) {
                        bindings[MIBT_JOY].emplace_back(bid, buf, isInverse);
                    }
                    else if (!strncmp(ptr, "mouse", 5)) {
                        bindings[MIBT_MOUSE].emplace_back(bid, buf, isInverse);
                    }
                }
            }
            ptr = end;
            while (*ptr == ' ') {
                ptr++;
            }
            ptr = strchr(ptr, ':');
        }
        for (auto iterType = bindings.begin(); iterType != bindings.end(); ++iterType) {
            for (const auto &bind : iterType->second) {
                callback(iterType->first,
                         std::get<0>(bind),
                         std::get<1>(bind).c_str(),
                         std::get<2>(bind));
            }
        }
    }
    Vector2ui measureAndDraw(bool drawing = true) const {
        const auto widgetTopLeft = self().geometry().topLeft;
        struct {
            bool     drawing;
            float    alpha;
            Vector2i widgetTopLeft;
            Vector2i origin;
            Vector2i size;
        } ctx = {drawing,
                 mnRendState->pageAlpha * self().scrollingFadeout(),
                 widgetTopLeft,
                 {0, 0},
                 {0, 0}};
        if (drawing) {
            if (ctx.alpha < .001f) return {};
        }
        iterateBindings(
            MIBF_IGNORE_REPEATS,
            [this, &ctx](bindingitertype_t type, int /*bid*/, const char *name, dd_bool isInverse) {
                static const int BIND_GAP = 2;
#if __JHERETIC__
                static float const bgRGB[] = {0, .5f, 0};
#elif __JHEXEN__
                static float const bgRGB[] = {.5f, 0, 0};
#else
                static float const bgRGB[] = {0, 0, 0};
#endif
                FR_SetFont(FID(GF_FONTA));
                const int lineHeight = FR_TextHeight("W");
                if (type == MIBT_KEY) {
                    const int width = FR_TextWidth(name) * SMALL_SCALE;
                    if (ctx.origin.x + width > maxWidth) {
                        ctx.origin.x = 0;
                        ctx.origin.y += lineHeight + 1;
                    }
                    if (ctx.drawing) {
                        DGL_SetNoMaterial();
                        DGL_DrawRectf2Color(ctx.widgetTopLeft.x + ctx.origin.x,
                                            ctx.widgetTopLeft.y + ctx.origin.y,
                                            width + 2,
                                            lineHeight,
                                            bgRGB[0],
                                            bgRGB[1],
                                            bgRGB[2],
                                            ctx.alpha * .6f);
                        DGL_Enable(DGL_TEXTURE_2D);
                        drawSmallText(name,
                                      ctx.widgetTopLeft.x + ctx.origin.x + 1,
                                      ctx.widgetTopLeft.y + ctx.origin.y,
                                      ctx.alpha);
                        DGL_Disable(DGL_TEXTURE_2D);
                    }
                    ctx.origin.x += width + 2 + BIND_GAP;
                }
                else {
                    char buf[256];
                    sprintf(buf,
                            "%s%c%s",
                            type == MIBT_MOUSE ? "mouse" : "joy",
                            isInverse ? '-' : '+',
                            name);
                    int const width = FR_TextWidth(buf) * SMALL_SCALE;
                    if (ctx.origin.x + width > maxWidth) {
                        ctx.origin.x = 0;
                        ctx.origin.y += lineHeight + 1;
                    }
                    if (ctx.drawing) {
                        DGL_Enable(DGL_TEXTURE_2D);
                        drawSmallText(buf,
                                      ctx.widgetTopLeft.x + ctx.origin.x,
                                      ctx.widgetTopLeft.y + ctx.origin.y,
                                      ctx.alpha);
                        DGL_Disable(DGL_TEXTURE_2D);
                    }
                    ctx.origin.x += width + BIND_GAP;
                }
                ctx.size = ctx.size.max({ctx.origin.x, ctx.origin.y + lineHeight});
            });
        return (ctx.size).toVector2ui();
    }
};
InputBindingWidget::InputBindingWidget()
    : Widget()
    , binds(nullptr)
    , d(new Impl(this)) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
}
InputBindingWidget::~InputBindingWidget()
{}
void InputBindingWidget::draw() const {
    d->measureAndDraw(true);
}
int InputBindingWidget::handleCommand(menucommand_e cmd) {
    switch(cmd) {
    case MCMD_DELETE:
        S_LocalSound(SFX_MENU_CANCEL, NULL);
        d->iterateBindings(0, [](bindingitertype_t, int bid, const char *, dd_bool) {
            DD_Executef(true, "delbind %i", bid);
        });
        d->needGeometry = true;
        if (binds->command && !strcmp(binds->command, "menuselect")) {
            DD_Execute(true, "bindevent menu:key-return menuselect");
        }
        return true;
    case MCMD_SELECT:
        S_LocalSound(SFX_MENU_CYCLE, NULL);
        setFlags(Active);
        if(hasAction(Activated)) {
            execAction(Activated);
            return true;
        }
        break;
    default: break;
    }
    return false; // Not eaten.
}
void InputBindingWidget::updateGeometry() {
    if (d->needGeometry) {
        d->needGeometry = false;
        geometry().setSize(d->measureAndDraw(false /* just measure */));
        geometry().setWidth(d->maxWidth);
    }
}
/**
 * Read the symbolic descriptor from the given @a event.
 */
static String symbolicDescriptor(event_t const &event) {
    if(event.type == EV_SYMBOLIC) {
#ifndef __64BIT__
        String symbol = (char const *) event.data1;
#else
        String symbol = (char const *) event.data_u64;
#endif
        if(symbol.beginsWith("echo-")) {
            return symbol.substr(5);
        }
    }
    return ""; // No symbolic descriptor.
}
int InputBindingWidget::handleEvent_Privileged(event_t const &event) {
    LOG_AS("InputBindingWidget");
    if (!isActive()) return false;
    String symbol = symbolicDescriptor(event);
    if (symbol.isEmpty()) return false;
    if((symbol.beginsWith("key-")         ||
        symbol.beginsWith("joy-button")   ||
        symbol.beginsWith("mouse-left")   ||
        symbol.beginsWith("mouse-middle") ||
        symbol.beginsWith("mouse-right")) && !symbol.endsWith("-down")) {
        return false;
    }
    String const context = bindContext();
    if((!context.compareWithCase("menu") || !context.compareWithCase("shortcut")) &&
       symbol.beginsWith("key-delete-down")) {
        return false;
    }
    String cmd;
    if (binds->command) {
        cmd = String("bindevent {%1:%2%3} {%4}")
                  .arg(context)
                  .arg(symbol)
                  .arg(binds->flags & CCF_MULTIPLAYER? " + multiplayer" : "")
                  .arg(binds->command);
        if ((binds->flags & CCF_REPEAT) && symbol.endsWith("-down")) {
            cmd += String("; bindevent {%1:%2-repeat} {%3}")
                       .arg(context)
                       .arg(symbol.left(symbol.length() - 5))
                       .arg(binds->command);
        }
    }
    else if (binds->controlName) {
        String stateFlags;
        int const endOfName = symbol.indexOf('-', symbol.indexOf('-') + 1);
        if (endOfName < 0) {
            throw Error("InputBindingWidget::handleEvent_Privileged", "Invalid symbol:" + symbol);
        }
        String const name = symbol.left(endOfName);
        if (binds->flags & CCF_STAGED) {
            if (name.beginsWith("key-") || name.beginsWith("joy-button") ||
                name.beginsWith("mouse-left") || name.beginsWith("mouse-middle") ||
                name.beginsWith("mouse-right")) {
                stateFlags += "-staged";
            }
        }
        bool inv = (binds->flags & CCF_INVERSE) != 0;
        if (symbol.substr(endOfName).beginsWith("-neg")) {
            inv = !inv;
        }
        if (inv) {
            stateFlags += "-inverse";
        }
        cmd = String("bindcontrol {%1} {%2%3%4}")
                  .arg(binds->controlName)
                  .arg(name)
                  .arg(stateFlags)
                  .arg((binds->flags & CCF_SIDESTEP_MODIFIER)? " + modifier-1-up" : "");
        if(binds->flags & CCF_SIDESTEP_MODIFIER) {
            cmd += String("; bindcontrol sidestep {%1%2 + modifier-1-down}")
                       .arg(name)
                       .arg(stateFlags);
        }
    }
    LOGDEV_INPUT_MSG("PrivilegedResponder: ") << cmd;
    DD_Execute(true, cmd.toUtf8().constData());
    setFlags(Active, UnsetFlags);
    DD_SetInteger(DD_SYMBOLIC_ECHO, false);
    S_LocalSound(SFX_MENU_ACCEPT, nullptr);
    d->needGeometry = true;
    return true;
}
char const *InputBindingWidget::controlName() const {
    DENG2_ASSERT(binds);
    if (PTR2INT(binds->text) > 0 && PTR2INT(binds->text) < NUMTEXT) {
        return GET_TXT(PTR2INT(binds->text));
    }
    return binds->text;
}
String InputBindingWidget::bindContext() const {
    DENG2_ASSERT(binds);
    return (binds->bindContext ? binds->bindContext : "game");
}
void InputBindingWidget::pageActivated() {
    d->needGeometry = true;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(LabelWidget) {
    String text;
    patchid_t *patch = nullptr;  ///< Used instead of text if Patch Replacement is in use.
    int flags = 0;               ///< @ref mnTextFlags
};
LabelWidget::LabelWidget(String const &text, patchid_t *patch)
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
    setFlags(NoFocus); // never focusable.
    setText(text);
    setPatch(patch);
}
LabelWidget::~LabelWidget()
{}
void LabelWidget::draw() const {
    fontid_t fontId           = mnRendState->textFonts[font()];
    Vector4f const &textColor = mnRendState->textColors[color()];
    const Vector4f color = selectionFlashColor(textColor);
    const float fadeout = scrollingFadeout();
    if (fadeout < .001f) return;
    DGL_Color4f(1, 1, 1, color.w * fadeout);
    FR_SetFont(fontId);
    FR_SetColorAndAlpha(color.x, color.y, color.z, color.w * fadeout);
    if (d->patch) {
        String replacement;
        if(!(d->flags & MNTEXT_NO_ALTTEXT)) {
            replacement = Hu_ChoosePatchReplacement(patchreplacemode_t(cfg.common.menuPatchReplaceMode), *d->patch, d->text);
        }
        DGL_Enable(DGL_TEXTURE_2D);
        WI_DrawPatch(*d->patch, replacement, geometry().topLeft, ALIGN_TOPLEFT, 0, Hu_MenuMergeEffectWithDrawTextFlags(0));
        DGL_Disable(DGL_TEXTURE_2D);
        return;
    }
    DGL_Enable(DGL_TEXTURE_2D);
    FR_DrawTextXY3(d->text.toUtf8().constData(), geometry().topLeft.x, geometry().topLeft.y, ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
    DGL_Disable(DGL_TEXTURE_2D);
}
void LabelWidget::updateGeometry() {
    if(d->patch) {
        patchinfo_t info;
        R_GetPatchInfo(*d->patch, &info);
        geometry().setSize(Vector2ui(info.geometry.size.width, info.geometry.size.height));
        return;
    }
    FR_PushAttrib();
    Size2Raw size;
    FR_SetFont(page().predefinedFont(mn_page_fontid_t(font())));
    FR_TextSize(&size, d->text.toUtf8().constData());
    geometry().setSize(Vector2ui(size.width, size.height));
    FR_PopAttrib();
}
LabelWidget &LabelWidget::setPatch(patchid_t *newPatch) {
    d->patch = newPatch;
    return *this;
}
LabelWidget &LabelWidget::setText(const String &text) {
    d->text = labelText(text);
    return *this;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
static patchid_t pEditLeft;
static patchid_t pEditRight;
static patchid_t pEditMiddle;
DENG2_PIMPL_NOREF(LineEditWidget) {
    String text;
    String oldText;    ///< For restoring a canceled edit.
    String emptyText;  ///< Used when value is empty.
    int maxLength       = 0;
    int maxVisibleChars = 0;
};
LineEditWidget::LineEditWidget()
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
}
LineEditWidget::~LineEditWidget()
{}
void LineEditWidget::loadResources() // static
{
#if defined(MNDATA_EDIT_BACKGROUND_PATCH_LEFT)
    pEditLeft   = R_DeclarePatch(MNDATA_EDIT_BACKGROUND_PATCH_LEFT);
#else
    pEditLeft   = 0;
#endif
#if defined(MNDATA_EDIT_BACKGROUND_PATCH_RIGHT)
    pEditRight  = R_DeclarePatch(MNDATA_EDIT_BACKGROUND_PATCH_RIGHT);
#else
    pEditRight  = 0;
#endif
    pEditMiddle = R_DeclarePatch(MNDATA_EDIT_BACKGROUND_PATCH_MIDDLE);
}
static void drawEditBackground(Vector2i const &origin, int width, float alpha) {
    DGL_Color4f(1, 1, 1, alpha);
    int leftOffset = 0;
    patchinfo_t leftInfo;
    if(R_GetPatchInfo(pEditLeft, &leftInfo)) {
        DGL_SetPatch(pEditLeft, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(origin.x, origin.y, leftInfo.geometry.size.width, leftInfo.geometry.size.height);
        leftOffset = leftInfo.geometry.size.width;
    }
    int rightOffset = 0;
    patchinfo_t rightInfo;
    if(R_GetPatchInfo(pEditRight, &rightInfo)) {
        DGL_SetPatch(pEditRight, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_DrawRectf2(origin.x + width - rightInfo.geometry.size.width, origin.y, rightInfo.geometry.size.width, rightInfo.geometry.size.height);
        rightOffset = rightInfo.geometry.size.width;
    }
    patchinfo_t middleInfo;
    if(R_GetPatchInfo(pEditMiddle, &middleInfo)) {
        if (!pEditLeft && !pEditRight) {
            DGL_SetPatch(pEditMiddle, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
            DGL_DrawRectf2(origin.x, origin.y,
                           width, middleInfo.geometry.size.height);
        }
        else {
            DGL_SetPatch(pEditMiddle, DGL_REPEAT, DGL_REPEAT);
            DGL_DrawRectf2Tiled(origin.x + leftOffset, origin.y, width - leftOffset - rightOffset, middleInfo.geometry.size.height, middleInfo.geometry.size.width, middleInfo.geometry.size.height);
        }
    }
}
void LineEditWidget::draw() const {
    fontid_t fontId = mnRendState->textFonts[font()];
    Vector2i origin = geometry().topLeft + Vector2i(MNDATA_EDIT_OFFSET_X, MNDATA_EDIT_OFFSET_Y);
    String useText;
    float light = 1, textOpacity = mnRendState->pageAlpha;
    if(!d->text.isEmpty()) {
        useText = d->text;
    }
    else if(!(isActive() && isFocused())) {
        useText = d->emptyText;
        light *= .5f;
        textOpacity = mnRendState->pageAlpha * .75f;
    }
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(fontId);
    const float fadeout = scrollingFadeout();
    drawEditBackground(origin + Vector2i(MNDATA_EDIT_BACKGROUND_OFFSET_X, MNDATA_EDIT_BACKGROUND_OFFSET_Y),
                       geometry().width(), mnRendState->pageAlpha * fadeout); {
        Vector4f color = Vector4f(Vector3f(cfg.common.menuTextColors[MNDATA_EDIT_TEXT_COLORIDX]), 1.f);
        if (!isActive()) /* && isFocused() && cfg.common.menuTextFlashSpeed > 0)
        {
            float const speed = cfg.common.menuTextFlashSpeed / 2.f;
            t = (1 + sin(page().timer() / (float)TICSPERSEC * speed * DD_PI)) / 2;
        }*/
        {
            color = selectionFlashColor(color);
        }
        color *= light;
        color.w = textOpacity;
        FR_SetColorAndAlpha(color.x, color.y, color.z, color.w * fadeout);
        FR_DrawTextXY3(useText.toUtf8().constData(), origin.x, origin.y, ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
        if(isActive() && isFocused() && (menuTime & 8) &&
           (!d->maxLength || d->text.length() < d->maxLength)) {
            origin.x += FR_TextWidth(useText.toUtf8().constData());
            FR_DrawCharXY3('_', origin.x, origin.y, ALIGN_TOPLEFT,  Hu_MenuMergeEffectWithDrawTextFlags(0));
        }
    }
    DGL_Disable(DGL_TEXTURE_2D);
}
int LineEditWidget::maxLength() const {
    return d->maxLength;
}
LineEditWidget &LineEditWidget::setMaxLength(int newMaxLength) {
    newMaxLength = de::max(newMaxLength, 0);
    if(d->maxLength != newMaxLength) {
        if(newMaxLength < d->maxLength) {
            d->text.truncate(newMaxLength);
            d->oldText.truncate(newMaxLength);
        }
        d->maxLength = newMaxLength;
    }
    return *this;
}
String LineEditWidget::text() const {
    return d->text;
}
LineEditWidget &LineEditWidget::setText(String const &newText, int flags) {
    d->text = newText;
    if(d->maxLength) d->text.truncate(d->maxLength);
    if(flags & MNEDIT_STF_REPLACEOLD) {
        d->oldText = d->text;
    }
    if(!(flags & MNEDIT_STF_NO_ACTION)) {
        execAction(Modified);
    }
    return *this;
}
LineEditWidget &LineEditWidget::setEmptyText(String const &newEmptyText) {
    d->emptyText = newEmptyText;
    return *this;
}
String LineEditWidget::emptyText() const {
    return d->emptyText;
}
/**
 * Responds to alphanumeric input for edit fields.
 */
int LineEditWidget::handleEvent(event_t const &ev) {
    if(!isActive() || ev.type != EV_KEY)
        return false;
    if(ev.data1 == DDKEY_RSHIFT) {
        shiftdown = (ev.state == EVS_DOWN || ev.state == EVS_REPEAT);
        return true;
    }
    if(!(ev.state == EVS_DOWN || ev.state == EVS_REPEAT))
        return false;
    if(ev.data1 == DDKEY_BACKSPACE) {
        if(!d->text.isEmpty()) {
            d->text.truncate(d->text.length() - 1);
            execAction(Modified);
        }
        return true;
    }
    if(ev.data1 >= ' ' && ev.data1 <= 'z') {
        char ch = char(ev.data1);
        if(shiftdown) {
            ch = shiftXForm[int(ch)];
        }
        if(ch == '%')
            return true;
        if(!d->maxLength || d->text.length() < d->maxLength) {
            d->text += ch;
            execAction(Modified);
        }
        return true;
    }
    return false;
}
int LineEditWidget::handleCommand(menucommand_e cmd) {
    if(cmd == MCMD_SELECT) {
        if(!isActive()) {
            S_LocalSound(SFX_MENU_CYCLE, NULL);
            setFlags(Active);
            d->oldText = d->text;
            execAction(Activated);
        }
        else {
            S_LocalSound(SFX_MENU_ACCEPT, NULL);
            d->oldText = d->text;
            setFlags(Active, UnsetFlags);
            execAction(Deactivated);
        }
        return true;
    }
    if(isActive()) {
        switch(cmd) {
        case MCMD_NAV_OUT:
            d->text = d->oldText;
            setFlags(Active, UnsetFlags);
            execAction(Closed);
            return true;
        case MCMD_NAV_LEFT:
        case MCMD_NAV_RIGHT:
        case MCMD_NAV_DOWN:
        case MCMD_NAV_UP:
        case MCMD_NAV_PAGEDOWN:
        case MCMD_NAV_PAGEUP:
            return true;
        default: break;
        }
    }
    return false; // Not eaten.
}
void LineEditWidget::updateGeometry() {
    FR_SetFont(mnRendState->textFonts[font()]);
    geometry().setSize(Vector2ui(FR_CharWidth('w') * d->maxLength - MNDATA_EDIT_BACKGROUND_OFFSET_X*2, 14));
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
ListWidget::Item::Item(String const &text, int userValue) {
    setText(text);
    setUserValue(userValue);
}
void ListWidget::Item::setText(String const &newText) {
    _text = newText;
}
String ListWidget::Item::text() const {
    return _text;
}
void ListWidget::Item::setUserValue(int newUserValue) {
    _userValue = newUserValue;
}
int ListWidget::Item::userValue() const {
    return _userValue;
}
DENG2_PIMPL_NOREF(ListWidget) {
    Items items;
    int   selection      = 0; ///< Selected item (-1 if none).
    int   first          = 0; ///< First visible item.
    int   numvis         = 0;
    bool  reorderEnabled = false;
    ~Impl() { qDeleteAll(items); }
};
ListWidget::ListWidget()
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
}
ListWidget::~ListWidget()
{}
ListWidget &ListWidget::addItem(Item *item) {
    if(item) d->items << item;
    return *this;
}
ListWidget &ListWidget::addItems(Items const &itemsToAdd) {
    for(Item *item : itemsToAdd) addItem(item);
    return *this;
}
ListWidget::Items const &ListWidget::items() const {
    return d->items;
}
void ListWidget::updateGeometry() {
    geometry().setSize(Vector2ui(0, 0));
    FR_PushAttrib();
    FR_SetFont(page().predefinedFont(mn_page_fontid_t(font())));
    RectRaw itemGeometry{};
    for(int i = 0; i < itemCount(); ++i) {
        Item *item = d->items[i];
        FR_TextSize(&itemGeometry.size, item->text().toUtf8().constData());
        if(i != itemCount() - 1) {
            itemGeometry.size.height *= 1 + MNDATA_LIST_LEADING;
        }
        geometry() |= Rectanglei::fromSize(Vector2i(itemGeometry.origin.xy), Vector2ui(itemGeometry.size.width, itemGeometry.size.height));
        itemGeometry.origin.y += itemGeometry.size.height;
    }
    FR_PopAttrib();
}
void ListWidget::draw() const {
    bool const flashSelection = (isActive() && selectionIsVisible());
    Vector4f const &textColor = mnRendState->textColors[color()];
    float t = flashSelection? 1 : 0;
    Vector4f flashColor = textColor;
    if (flashSelection) /* && cfg.common.menuTextFlashSpeed > 0)
    {
        float const speed = cfg.common.menuTextFlashSpeed / 2.f;
        t = (1 + sin(page().timer() / (float)TICSPERSEC * speed * DD_PI)) / 2;
    }*/
    {
        flashColor = selectionFlashColor(flashColor);
    }
    Vector4f const dimColor   = Vector4f(Vector3f(textColor) * MNDATA_LIST_NONSELECTION_LIGHT, textColor.w);
    if(d->first < d->items.count() && d->numvis > 0) {
        DGL_Enable(DGL_TEXTURE_2D);
        FR_SetFont(mnRendState->textFonts[font()]);
        Vector2i origin = geometry().topLeft;
        int itemIdx = d->first;
        do {
            const Item *item = d->items[itemIdx];
            const Vector4f &color =
                d->selection == itemIdx ? (flashSelection ? flashColor : textColor) : dimColor;
            const int itemHeight =
                FR_TextHeight(item->text().toUtf8().constData()) * (1 + MNDATA_LIST_LEADING);
            FR_SetColorAndAlpha(color.x,
                                color.y,
                                color.z,
                                color.w * scrollingFadeout(origin.y, origin.y + itemHeight));
            FR_DrawTextXY3(item->text().toUtf8().constData(), origin.x, origin.y, ALIGN_TOPLEFT, Hu_MenuMergeEffectWithDrawTextFlags(0));
            origin.y += itemHeight;
        } while (++itemIdx < d->items.count() && itemIdx < d->first + d->numvis);
        DGL_Disable(DGL_TEXTURE_2D);
    }
}
int ListWidget::handleCommand(menucommand_e cmd) {
    switch (cmd) {
    case MCMD_NAV_DOWN:
    case MCMD_NAV_UP:
        if (isActive()) {
            int oldSelection = d->selection;
            if(MCMD_NAV_DOWN == cmd) {
                if(d->selection < itemCount() - 1)
                    selectItem(d->selection + 1);
            }
            else {
                if(d->selection > 0)
                    selectItem(d->selection - 1);
            }
            if(d->selection != oldSelection) {
                S_LocalSound(cmd == MCMD_NAV_DOWN? SFX_MENU_NAV_DOWN : SFX_MENU_NAV_UP, NULL);
                execAction(Modified);
            }
            return true;
        }
        return false; // Not eaten.
    case MCMD_NAV_OUT:
        if (isActive()) {
            S_LocalSound(SFX_MENU_CANCEL, NULL);
            setFlags(Active, UnsetFlags);
            execAction(Closed);
            return true;
        }
        return false; // Not eaten.
    case MCMD_SELECT:
        if (!isActive()) {
            S_LocalSound(SFX_MENU_ACCEPT, NULL);
            setFlags(Active);
            execAction(Activated);
        }
        else {
            S_LocalSound(SFX_MENU_ACCEPT, NULL);
            setFlags(Active, UnsetFlags);
            execAction(Deactivated);
        }
        return true;
    case MCMD_NAV_LEFT:
    case MCMD_NAV_RIGHT:
        if (d->reorderEnabled && isActive()) {
            if (reorder(selection(), cmd == MCMD_NAV_LEFT ? -1 : +1)) {
                S_LocalSound(SFX_MENU_SLIDER_MOVE, NULL);
                execAction(Modified);
            }
        }
        return true;
    default: return false; // Not eaten.
    }
}
int ListWidget::selection() const {
    return d->selection;
}
int ListWidget::first() const {
    return d->first;
}
bool ListWidget::selectionIsVisible() const {
    return (d->selection >= d->first && d->selection < d->first + d->numvis);
}
void ListWidget::updateVisibleSelection() {
    d->numvis = itemCount();
    if(d->selection >= 0) {
        if(d->selection < d->first)
            d->first = d->selection;
        if(d->selection > d->first + d->numvis - 1)
            d->first = d->selection - d->numvis + 1;
    }
}
void ListWidget::pageActivated() {
    Widget::pageActivated();
    updateVisibleSelection();
}
int ListWidget::itemData(int index) const {
    if (index >= 0 && index < itemCount()) {
        return d->items[index]->userValue();
    }
    return 0;
}
int ListWidget::findItem(int userValue) const {
    for(int i = 0; i < d->items.count(); ++i) {
        Item *item = d->items[i];
        if(item->userValue() == userValue) {
            return i;
        }
    }
    return -1;
}
bool ListWidget::selectItem(int itemIndex, int flags) {
    if(itemIndex >= 0 && itemIndex < itemCount()) {
        if(d->selection != itemIndex) {
            d->selection = itemIndex;
            if(!(flags & MNLIST_SIF_NO_ACTION)) {
                execAction(Modified);
            }
            return true;
        }
    }
    return false;
}
bool ListWidget::selectItemByValue(int userValue, int flags) {
    return selectItem(findItem(userValue), flags);
}
bool ListWidget::reorder(int itemIndex, int indexOffset) {
    if (itemIndex + indexOffset < 0 || itemIndex + indexOffset >= d->items.size()) {
        return false; // Would go out of bounds.
    }
    if (d->selection == itemIndex) {
        d->selection += indexOffset;
    }
    while (indexOffset < 0) {
        std::swap(d->items[itemIndex - 1], d->items[itemIndex]);
        --itemIndex;
        ++indexOffset;
    }
    while (indexOffset > 0) {
        std::swap(d->items[itemIndex + 1], d->items[itemIndex]);
        ++itemIndex;
        --indexOffset;
    }
    return true;
}
ListWidget &ListWidget::setReorderingEnabled(bool reorderEnabled) {
    d->reorderEnabled = reorderEnabled;
    return *this;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(MobjPreviewWidget) {
    int mobjType    = 0;
    int playerClass = 0;
    int xlatClass   = 0;  ///< Color translation class.
    int xlatMap     = 0;  ///< Color translation map.
};
MobjPreviewWidget::MobjPreviewWidget()
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
    setFlags(NoFocus); // never focusable.
}
MobjPreviewWidget::~MobjPreviewWidget()
{}
static void findSpriteForMobjType(int mobjType, spritetype_e *sprite, int *frame) {
    DENG2_ASSERT(mobjType >= MT_FIRST && mobjType < NUMMOBJTYPES && sprite && frame);
    mobjinfo_t *info = &MOBJINFO[mobjType];
    int stateNum = info->states[SN_SPAWN];
    *sprite = spritetype_e(STATES[stateNum].sprite);
    *frame  = ((menuTime >> 3) & 3);
}
void MobjPreviewWidget::setMobjType(int newMobjType) {
    d->mobjType = newMobjType;
}
void MobjPreviewWidget::setPlayerClass(int newPlayerClass) {
    d->playerClass = newPlayerClass;
}
void MobjPreviewWidget::setTranslationClass(int newTranslationClass) {
    d->xlatClass = newTranslationClass;
}
void MobjPreviewWidget::setTranslationMap(int newTranslationMap) {
    d->xlatMap = newTranslationMap;
}
void MobjPreviewWidget::draw() const {
    if(MT_NONE == d->mobjType) return;
    spritetype_e sprite;
    int spriteFrame;
    findSpriteForMobjType(d->mobjType, &sprite, &spriteFrame);
    spriteinfo_t info;
    if(!R_GetSpriteInfo(sprite, spriteFrame, &info)) return;
    Point2Raw origin = {{{info.geometry.origin.x, info.geometry.origin.y}}};
    Size2Raw size = {{{info.geometry.size.width, info.geometry.size.height}}};
    float scale = (size.height > size.width? (float)MNDATA_MOBJPREVIEW_HEIGHT / size.height :
                                             (float)MNDATA_MOBJPREVIEW_WIDTH  / size.width);
    float s = info.texCoord[0];
    float t = info.texCoord[1];
    int tClassCycled = d->xlatClass;
    int tMapCycled   = d->xlatMap;
    if(tMapCycled == NUMPLAYERCOLORS) {
        tMapCycled = menuTime / 5 % NUMPLAYERCOLORS;
    }
#if __JHEXEN__
    if(gameMode == hexen_v10) {
        if(d->xlatMap == NUMPLAYERCOLORS) tMapCycled = menuTime / 5 % 4;
    }
    if(d->playerClass >= PCLASS_FIGHTER) {
        R_GetTranslation(d->playerClass, tMapCycled, &tClassCycled, &tMapCycled);
    }
#endif
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(geometry().topLeft.x, geometry().topLeft.y, 0);
    DGL_Scalef(scale, scale, 1);
    DGL_Translatef(-origin.x, -origin.y, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_SetPSprite2(info.material, tClassCycled, tMapCycled);
    DGL_Color4f(1, 1, 1, mnRendState->pageAlpha);
    DGL_Begin(DGL_QUADS);
        DGL_TexCoord2f(0, 0 * s, 0);
        DGL_Vertex2f(0, 0);
        DGL_TexCoord2f(0, 1 * s, 0);
        DGL_Vertex2f(size.width, 0);
        DGL_TexCoord2f(0, 1 * s, t);
        DGL_Vertex2f(size.width, size.height);
        DGL_TexCoord2f(0, 0 * s, t);
        DGL_Vertex2f(0, size.height);
    DGL_End();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
    DGL_Disable(DGL_TEXTURE_2D);
}
void MobjPreviewWidget::updateGeometry() {
    geometry().setSize(Vector2ui(MNDATA_MOBJPREVIEW_WIDTH, MNDATA_MOBJPREVIEW_HEIGHT));
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(RectWidget) {
    Vector2ui dimensions;  ///< Dimensions of the rectangle.
    patchid_t patch = 0;   ///< Background patch.
};
RectWidget::RectWidget(patchid_t backgroundPatch)
    : Widget()
    , d(new Impl) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
    setBackgroundPatch(backgroundPatch);
}
RectWidget::~RectWidget()
{}
void RectWidget::draw() const {
    Vector2i const origin = geometry().topLeft;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(origin.x, origin.y, 0);
    if(d->patch) {
        DGL_SetPatch(d->patch, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
        DGL_Enable(DGL_TEXTURE_2D);
    }
    DGL_Color4f(1, 1, 1, mnRendState->pageAlpha);
    DGL_DrawRect2(0, 0, d->dimensions.x, d->dimensions.y);
    if(d->patch) {
        DGL_Disable(DGL_TEXTURE_2D);
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(-origin.x, -origin.y, 0);
}
void RectWidget::updateGeometry() {
    if(d->dimensions == Vector2ui(0, 0)) {
        patchinfo_t info;
        if(R_GetPatchInfo(d->patch, &info)) {
            d->dimensions = Vector2ui(info.geometry.size.width, info.geometry.size.height);
        }
    }
    geometry().setSize(d->dimensions);
}
void RectWidget::setBackgroundPatch(patchid_t newBackgroundPatch) {
    d->patch = newBackgroundPatch;
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
static patchid_t pSliderLeft;
static patchid_t pSliderRight;
static patchid_t pSliderMiddle;
static patchid_t pSliderHandle;
DENG2_PIMPL(SliderWidget) {
    float min      = 0.0f;
    float max      = 1.0f;
    float value    = 0.0f;
    float step     = 0.1f;  ///< Button step.
    bool floatMode = true;  ///< @c false= only integers are allowed.
    Impl(Public *i) : Base(i) {}
    int thumbPos() const {
#define WIDTH           (middleInfo.geometry.size.width)
        patchinfo_t middleInfo;
        if(!R_GetPatchInfo(pSliderMiddle, &middleInfo)) return 0;
        float range = max - min;
        if(!range) range = 1; // Should never happen...
        float useVal = self().value() - min;
        return useVal / range * MNDATA_SLIDER_SLOTS * WIDTH;
#undef WIDTH
    }
};
SliderWidget::SliderWidget(float min, float max, float step, bool floatMode)
    : Widget()
    , d(new Impl(this)) {
    setFont(MENU_FONT1);
    setColor(MENU_COLOR1);
    d->min       = min;
    d->max       = max;
    d->step      = step;
    d->floatMode = floatMode;
}
SliderWidget::~SliderWidget()
{}
void SliderWidget::loadResources() // static
{
    pSliderLeft   = R_DeclarePatch(MNDATA_SLIDER_PATCH_LEFT);
    pSliderRight  = R_DeclarePatch(MNDATA_SLIDER_PATCH_RIGHT);
    pSliderMiddle = R_DeclarePatch(MNDATA_SLIDER_PATCH_MIDDLE);
    pSliderHandle = R_DeclarePatch(MNDATA_SLIDER_PATCH_HANDLE);
}
SliderWidget &SliderWidget::setValue(float newValue, int /*flags*/)
{
    if(d->floatMode) d->value = newValue;
    else             d->value = int( newValue + (newValue > 0? + .5f : -.5f) );
    return *this;
}
float SliderWidget::value() const {
    if(d->floatMode) {
        return d->value;
    }
    return int( d->value + (d->value > 0? .5f : -.5f) );
}
SliderWidget &SliderWidget::setRange(float newMin, float newMax, float newStep) {
    d->min  = newMin;
    d->max  = newMax;
    d->step = newStep;
    return *this;
}
float SliderWidget::min() const {
    return d->min;
}
float SliderWidget::max() const {
    return d->max;
}
float SliderWidget::step() const {
    return d->step;
}
SliderWidget &SliderWidget::setFloatMode(bool yes) {
    d->floatMode = yes;
    return *this;
}
bool SliderWidget::floatMode() const {
    return d->floatMode;
}
void SliderWidget::draw() const {
#define WIDTH                   (middleInfo.geometry.size.width)
#define HEIGHT                  (middleInfo.geometry.size.height)
    patchinfo_t middleInfo, leftInfo;
    if(!R_GetPatchInfo(pSliderMiddle, &middleInfo)) return;
    if(!R_GetPatchInfo(pSliderLeft, &leftInfo)) return;
    if(WIDTH <= 0 || HEIGHT <= 0) return;
    const float fadeout = scrollingFadeout();
    if (fadeout < .001f) return;
    Vector2f origin = geometry().topLeft;
    origin += Vector2f(MNDATA_SLIDER_OFFSET_X + leftInfo.geometry.size.width, MNDATA_SLIDER_OFFSET_Y) * MNDATA_SLIDER_SCALE;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(origin.x, origin.y, 0);
    DGL_Scalef(MNDATA_SLIDER_SCALE, MNDATA_SLIDER_SCALE, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    if(cfg.common.menuShadow > 0) {
        float const from[] = { 2, float(1 + HEIGHT / 2) };
        float const to[]   = { float(MNDATA_SLIDER_SLOTS * WIDTH - 2), float(1 + HEIGHT / 2) };
        M_DrawGlowBar(from, to, HEIGHT * 1.1f, true, true, true, 0, 0, 0, mnRendState->pageAlpha * mnRendState->textShadow);
    }
    DGL_Color4f(1, 1, 1, mnRendState->pageAlpha * fadeout);
    GL_DrawPatch(pSliderLeft,  Vector2i(0, 0), ALIGN_TOPRIGHT, DPF_NO_OFFSETX);
    GL_DrawPatch(pSliderRight, Vector2i(MNDATA_SLIDER_SLOTS * WIDTH, 0));
    DGL_SetPatch(pSliderMiddle, DGL_REPEAT, DGL_REPEAT);
    DGL_DrawRectf2Tiled(0, middleInfo.geometry.origin.y, MNDATA_SLIDER_SLOTS * WIDTH, HEIGHT, middleInfo.geometry.size.width, middleInfo.geometry.size.height);
    DGL_Color4f(1, 1, 1, mnRendState->pageAlpha * fadeout);
    GL_DrawPatch(pSliderHandle, Vector2i(d->thumbPos(), 1), ALIGN_TOP, DPF_NO_OFFSET);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef HEIGHT
#undef WIDTH
}
int SliderWidget::handleCommand(menucommand_e cmd) {
    switch(cmd) {
    case MCMD_NAV_LEFT:
    case MCMD_NAV_RIGHT: {
        float const oldvalue = d->value;
        if(MCMD_NAV_LEFT == cmd) {
            d->value -= d->step;
            if(d->value < d->min)
                d->value = d->min;
        }
        else {
            d->value += d->step;
            if(d->value > d->max)
                d->value = d->max;
        }
        if(oldvalue != d->value) {
            S_LocalSound(SFX_MENU_SLIDER_MOVE, NULL);
            execAction(Modified);
        }
        return true; }
    default: break;
    }
    return false; // Not eaten.
}
void SliderWidget::updateGeometry() {
    patchinfo_t info;
    if(!R_GetPatchInfo(pSliderMiddle, &info)) return;
    int middleWidth = info.geometry.size.width * MNDATA_SLIDER_SLOTS;
    geometry().setSize(Vector2ui(middleWidth, info.geometry.size.height));
    if(R_GetPatchInfo(pSliderLeft, &info)) {
        info.geometry.origin.x = -info.geometry.size.width;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    if(R_GetPatchInfo(pSliderRight, &info)) {
        info.geometry.origin.x += middleWidth;
        geometry() |= Rectanglei::fromSize(Vector2i(info.geometry.origin.xy), Vector2ui(info.geometry.size.width, info.geometry.size.height));
    }
    geometry().setSize(Vector2ui(.5f + geometry().width()  * MNDATA_SLIDER_SCALE,
                                 .5f + geometry().height() * MNDATA_SLIDER_SCALE));
}
} // namespace menu
} // namespace common
namespace common {
namespace menu {
DENG2_PIMPL_NOREF(Widget) {
    Flags flags;
    int group = 0;             ///< Object group identifier.
    Vector2i origin;           ///< Used with the fixed layout (in the page coordinate space).
    Rectanglei geometry;       ///< "Physical" geometry.
    Page *page = nullptr;      ///< Page which owns this object (if any).
    String helpInfo;           ///< Additional help information displayed when the widget has focus.
    int shortcut     = 0;      ///< DDKEY used to switch focus (@c 0= none).
    int pageFontIdx  = 0;      ///< Index of the predefined page font to use when drawing this.
    int pageColorIdx = 0;      ///< Index of the predefined page color to use when drawing this.
    typedef QMap<Action, ActionCallback> Actions;
    Actions actions;
    OnTickCallback onTickCallback = nullptr;
    CommandResponder cmdResponder = nullptr;
    QVariant userValue;
    QVariant userValue2;
};
Widget::Widget() : d(new Impl)
{}
Widget::~Widget()
{}
int Widget::handleEvent(event_t const &) {
    return 0; // Not handled.
}
int Widget::handleEvent_Privileged(event_t const &) {
    return 0; // Not handled.
}
Widget &Widget::setCommandResponder(CommandResponder newResponder) {
    d->cmdResponder = newResponder;
    return *this;
}
int Widget::cmdResponder(menucommand_e command) {
    if(d->cmdResponder) {
        if(int result = d->cmdResponder(*this, command))
            return result;
    }
    else {
        if(int result = handleCommand(command))
            return result;
    }
    return false; // Not handled.
}
void Widget::tick() {
    if(isHidden()) return;
    if(isPaused()) return;
    if(d->onTickCallback) {
        d->onTickCallback(*this);
    }
}
Widget &Widget::setOnTickCallback(OnTickCallback newCallback) {
    d->onTickCallback = newCallback;
    return *this;
}
bool Widget::hasPage() const {
    return d->page != 0;
}
Widget &Widget::setPage(Page *newPage) {
    d->page = newPage;
    return *this;
}
Page &Widget::page() const {
    if(d->page) {
        return *d->page;
    }
    throw MissingPageError("Widget::page", "No page is attributed");
}
Rectanglei &Widget::geometry() {
    return d->geometry;
}
Rectanglei const &Widget::geometry() const {
    return d->geometry;
}
Widget &Widget::setFixedOrigin(Vector2i const &newOrigin) {
    d->origin = newOrigin;
    return *this;
}
Vector2i Widget::fixedOrigin() const {
    return d->origin;
}
Widget &Widget::setFixedX(int newX) {
    d->origin.x = newX;
    return *this;
}
Widget &Widget::setFixedY(int newY) {
    d->origin.y = newY;
    return *this;
}
Widget &Widget::setFlags(Flags flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
    return *this;
}
Widget::Flags Widget::flags() const {
    return d->flags;
}
Widget &Widget::setGroup(int newGroup) {
    d->group = newGroup;
    return *this;
}
int Widget::group() const {
    return d->group;
}
Widget &Widget::setShortcut(int ddkey) {
    if(isalnum(ddkey)) {
        d->shortcut = tolower(ddkey);
    }
    return *this;
}
int Widget::shortcut() {
    return d->shortcut;
}
Widget &Widget::setFont(int newPageFont) {
    d->pageFontIdx = newPageFont;
    return *this;
}
int Widget::font() const {
    return d->pageFontIdx;
}
Widget &Widget::setColor(int newPageColor) {
    d->pageColorIdx = newPageColor;
    return *this;
}
int Widget::color() const {
    return d->pageColorIdx;
}
Widget &Widget::setHelpInfo(String newHelpInfo) {
    d->helpInfo = newHelpInfo;
    return *this;
}
String const &Widget::helpInfo() const {
    return d->helpInfo;
}
int Widget::handleCommand(menucommand_e cmd) {
    if(MCMD_SELECT == cmd && isFocused() && !isDisabled()) {
        S_LocalSound(SFX_MENU_ACCEPT, NULL);
        if(!isActive()) {
            setFlags(Active);
            execAction(Activated);
        }
        setFlags(Active, UnsetFlags);
        execAction(Deactivated);
        return true;
    }
    return false; // Not eaten.
}
bool Widget::hasAction(Action id) const {
    return d->actions.contains(id);
}
Widget &Widget::setAction(Action id, ActionCallback callback) {
    if(callback) {
        d->actions.insert(id, callback);
    }
    else {
        d->actions.remove(id);
    }
    return *this;
}
void Widget::execAction(Action id) {
    if(hasAction(id)) {
        d->actions[id](*this, id);
    }
}
Widget &Widget::setUserValue(QVariant const &newValue) {
    d->userValue = newValue;
    return *this;
}
QVariant const &Widget::userValue() const {
    return d->userValue;
}
Widget &Widget::setUserValue2(QVariant const &newValue) {
    d->userValue2 = newValue;
    return *this;
}
QVariant const &Widget::userValue2() const {
    return d->userValue2;
}
float Widget::scrollingFadeout() const {
    const auto geom = geometry();
    return scrollingFadeout(geom.topLeft.y, geom.bottomRight.y);
}
float Widget::scrollingFadeout(int yTop, int yBottom) const {
    if (page().flags() & Page::NoScroll) {
        return 1.f;
    }
    const float FADEOUT_LEN = 20;
    const auto  viewRegion  = page().viewRegion();
    if (yBottom < viewRegion.topLeft.y) {
        return de::max(0.f, 1.f - (viewRegion.topLeft.y - yBottom) / FADEOUT_LEN);
    }
    if (yTop > viewRegion.bottomRight.y) {
        return de::max(0.f, 1.f - (yTop - viewRegion.bottomRight.y) / FADEOUT_LEN);
    }
    return 1.f;
}
Vector4f Widget::selectionFlashColor(const Vector4f &noFlashColor) const {
    if (isFocused() && cfg.common.menuTextFlashSpeed > 0) {
        const float speed = cfg.common.menuTextFlashSpeed / 2.f;
        const float t = (1 + sin(page().timer() / (float)TICSPERSEC * speed * DD_PI)) / 2;
        return lerp(noFlashColor, Vector4f(Vector3f(cfg.common.menuTextFlashColor), 1.f), t);
    }
    return noFlashColor;
}
void Widget::pageActivated()
{}
String Widget::labelText(const String &text, const String &context) // static
{
    if (const auto *repl = Defs().getValueById(context + "|" + text)) {
        return repl->text;
    }
    return text;
}
} // namespace menu
} // namespace common
