﻿/** @file ammowidget.cpp  GUI widget for visualizing player ammo ownership.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/ammowidget.h"
#include "player.h"
/** @file armoriconwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/armoriconwidget.h"
#include "player.h"
/** @file ammowidget.h  GUI widget for visualizing high-level player status.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/facewidget.h"
#include "gl_drawpatch.h"
#include "player.h"
/** @file healthiconwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/healthiconwidget.h"
#include "common.h"
#include "p_actor.h"
/** @file maxammowidget.cpp  GUI widget for visualizing max player ammo ownership.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/maxammowidget.h"
#include "player.h"
/** @file weaponslotwidget.cpp  GUI widget for visualizing player weapon ownership.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/weaponslotwidget.h"
#include <de/Vector>
#include "hu_stuff.h"
#include "player.h"
using namespace de;
static void AmmoWidget_UpdateGeometry(guidata_ammo_t *ammo) {
    DENG2_ASSERT(ammo);
    ammo->updateGeometry();
}
static void AmmoWidget_Draw(guidata_ammo_t *wi, Point2Raw const *offset) {
    DENG2_ASSERT(wi);
    wi->draw(offset? Vector2i(offset->xy) : Vector2i());
}
guidata_ammo_t::guidata_ammo_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(AmmoWidget_UpdateGeometry),
                function_cast<DrawFunc>(AmmoWidget_Draw),
                player)
{}
guidata_ammo_t::~guidata_ammo_t()
{}
void guidata_ammo_t::reset() {
    _value = 1994;
}
guidata_ammo_t &guidata_ammo_t::setAmmoType(ammotype_t newAmmoType) {
    _ammotype = newAmmoType;
    return *this;
}
void guidata_ammo_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    _value = plr->ammo[_ammotype].owned;
}
void guidata_ammo_t::draw(Vector2i const &offset) const {
    static Vector2i const origin(-ST_WIDTH / 2, -ST_HEIGHT);
    static Vector2i const offsets[NUM_AMMO_TYPES] = {
        { 288, 5 + 6 * 0 }, { 288, 5 + 6 * 1 }, { 288, 5 + 6 * 3 }, { 288, 5 + 6 * 2 },
    };
    dint const activeHud     = ST_ActiveHud(player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(_value == 1994) return;
    Vector2i const pos     = origin + offsets[dint( _ammotype )];
    auto const valueAsText = String::number(_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(font());
    FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), pos.x, pos.y, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void guidata_ammo_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    FR_SetFont(font());
    Rect_SetWidthHeight(&geometry(), (FR_CharWidth('0') * 3) * ::cfg.common.statusbarScale,
                                      FR_CharHeight('0') * ::cfg.common.statusbarScale);
}
static void ArmorIcon_Draw(guidata_armoricon_t *icon, Point2Raw const *offset) {
    DENG2_ASSERT(icon);
    icon->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void ArmorIcon_UpdateGeometry(guidata_armoricon_t *icon) {
    DENG2_ASSERT(icon);
    icon->updateGeometry();
}
guidata_armoricon_t::guidata_armoricon_t(dint player, int sprite1, int sprite2)
    : HudWidget(function_cast<UpdateGeometryFunc>(ArmorIcon_UpdateGeometry),
                function_cast<DrawFunc>(ArmorIcon_Draw),
                player) {
    this->armorSprite1 = sprite1;
    this->armorSprite2 = sprite2;
}
guidata_armoricon_t::~guidata_armoricon_t()
{}
void guidata_armoricon_t::reset() {
    currentSprite = -1;
}
void guidata_armoricon_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &players[player()];
    currentSprite = (plr->armorType == 2 ? this->armorSprite2 : this->armorSprite1);
}
void guidata_armoricon_t::draw(Vector2i const &offset) const {
    dfloat const iconOpacity = uiRendState->pageAlpha * cfg.common.hudIconAlpha;
    if(!cfg.hudShown[HUD_ARMOR]) return;
    if(ST_AutomapIsOpen(player()) && cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(currentSprite < 0) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(cfg.common.hudScale, cfg.common.hudScale, 1);
    GUI_DrawSprite(currentSprite, 0, 0, HOT_TLEFT, 1, iconOpacity, false, nullptr, nullptr);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void guidata_armoricon_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!cfg.hudShown[HUD_ARMOR]) return;
    if(ST_AutomapIsOpen(player()) && cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(currentSprite < 0) return;
    Size2Raw iconSize;
    GUI_SpriteSize(currentSprite, 1, &iconSize.width, &iconSize.height);
    iconSize.width  *= cfg.common.hudScale;
    iconSize.height *= cfg.common.hudScale;
    Rect_SetWidthHeight(&geometry(), iconSize.width, iconSize.height);
}
#define FACE_STRAIGHT_COUNT         ( 3 )
#define FACE_TURN_COUNT             ( 2 )
#define FACE_SPECIAL_COUNT          ( 3 )
#define FACE_PAIN_COUNT             ( 5 )
#define FACE_EXTRA_COUNT            ( 2 )
#define FACE_STRIDE                 ( FACE_STRAIGHT_COUNT + FACE_TURN_COUNT + FACE_SPECIAL_COUNT )
#define FACE_COUNT                  ( FACE_STRIDE * FACE_PAIN_COUNT + FACE_EXTRA_COUNT )
#define FACE_TURN_FIRST             ( FACE_STRAIGHT_COUNT )
#define FACE_OUCH_FIRST             ( FACE_TURN_FIRST + FACE_TURN_COUNT )
#define FACE_GRIN_FIRST             ( FACE_OUCH_FIRST + 1 )
#define FACE_RAMPAGE_FIRST          ( FACE_GRIN_FIRST + 1 )
#define FACE_GOD_FIRST              ( FACE_PAIN_COUNT * FACE_STRIDE )
#define FACE_DEAD_FIRST             ( FACE_GOD_FIRST + 1 )
#define FACE_STRAIGHT_TICS          ( TICRATE / 2 )
#define FACE_TURN_TICS              ( 1 * TICRATE )
#define FACE_OUCH_TICS              ( 1 * TICRATE )
#define FACE_GRIN_TICS              ( 2 * TICRATE )
#define FACE_RAMPAGE_TICS           ( 2 * TICRATE )
#define FACE_PAIN_THRESHOLD         ( 20 )
static patchid_t pFaces[FACE_COUNT];
static patchid_t pBackground[NUMTEAMS];
DENG2_PIMPL_NOREF(guidata_face_t) {
    dint faceTicks = 0;  ///< Count until face changes.
    dint faceIndex = 0;  ///< Current face index, used by wFaces.
    dint priority  = 0;
    dint oldHealth = 0;
    bool oldWeaponsOwned[NUM_WEAPON_TYPES];
    dint lastAttackDown = 0;
    Impl() { de::zap(oldWeaponsOwned); }
    dint painOffset(dint player) const {
        player_t const *plr = &::players[player];
        return FACE_STRIDE * (((100 - de::min(100, plr->health)) * FACE_PAIN_COUNT) / 101);
    }
};
guidata_face_t::guidata_face_t(void (*updateGeometry) (HudWidget *wi),
                               void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                               dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
    , d(new Impl)
{}
guidata_face_t::~guidata_face_t()
{}
void guidata_face_t::reset() {
    player_t *plr = &::players[player()];
    d->faceTicks      = 0;
    d->faceIndex      = 0;
    d->priority       = 0;
    d->lastAttackDown = -1;
    d->oldHealth      = -1;
    for(dint i = 0; i < NUM_WEAPON_TYPES; ++i) {
        d->oldWeaponsOwned[i] = CPP_BOOL(plr->weapons[i].owned);
    }
}
/**
 * This not-very-pretty routine handles face animation states and timing thereof.
 * The precedence of expressions is:
 *
 * dead > evil grin > turned head > straight ahead
 */
void guidata_face_t::tick(timespan_t /*elapsed*/)
{
    player_t const *plr = &::players[player()];
    if(Pause_IsPaused() || !DD_IsSharpTick())
        return;
    if(d->priority < 10) {
        if(!plr->health) {
            d->priority  = 9;
            d->faceIndex = FACE_DEAD_FIRST;
            d->faceTicks = 1;
        }
    }
    if(d->priority < 9) {
        if(plr->bonusCount) {
            bool doGrin = false;
            for(dint i = 0; i < NUM_WEAPON_TYPES; ++i) {
                if(d->oldWeaponsOwned[i] != CPP_BOOL(plr->weapons[i].owned)) {
                    doGrin = true;
                    d->oldWeaponsOwned[i] = CPP_BOOL(plr->weapons[i].owned);
                }
            }
            if(doGrin) {
                d->priority  = 8;
                d->faceTicks = FACE_GRIN_TICS;
                d->faceIndex = FACE_GRIN_FIRST + d->painOffset(player());
            }
        }
    }
    if(d->priority < 8) {
        if(plr->damageCount && plr->attacker && plr->attacker != plr->plr->mo) {
            d->priority = 7;
            if((::cfg.fixOuchFace?
               (d->oldHealth - plr->health) :
               (plr->health - d->oldHealth)) > FACE_PAIN_THRESHOLD) {
                d->faceTicks = FACE_TURN_TICS;
                d->faceIndex = FACE_OUCH_FIRST + d->painOffset(player());
                if(::cfg.fixOuchFace) {
                    d->priority = 8; // Added to fix 1 tic issue.
                }
            }
            else {
                angle_t badGuyAngle = M_PointToAngle2(plr->plr->mo->origin, plr->attacker->origin);
                angle_t diffAng;
                dint i;
                if(badGuyAngle > plr->plr->mo->angle) {
                    diffAng = badGuyAngle - plr->plr->mo->angle;
                    i = diffAng > ANG180;
                }
                else {
                    diffAng = plr->plr->mo->angle - badGuyAngle;
                    i = diffAng <= ANG180;
                }
                d->faceTicks = FACE_TURN_TICS;
                d->faceIndex = d->painOffset(player());
                if(diffAng < ANG45) {
                    d->faceIndex += FACE_RAMPAGE_FIRST;
                }
                else if(i) {
                    d->faceIndex += FACE_TURN_FIRST;
                }
                else {
                    d->faceIndex += FACE_TURN_FIRST + 1;
                }
            }
        }
    }
    if(d->priority < 7) {
        if(plr->damageCount) {
            if((::cfg.fixOuchFace?
               (d->oldHealth - plr->health) :
               (plr->health - d->oldHealth)) > FACE_PAIN_THRESHOLD) {
                d->priority  = 7;
                d->faceTicks = FACE_TURN_TICS;
                d->faceIndex = FACE_OUCH_FIRST + d->painOffset(player());
            }
            else {
                d->priority  = 6;
                d->faceTicks = FACE_TURN_TICS;
                d->faceIndex = FACE_RAMPAGE_FIRST + d->painOffset(player());
            }
        }
    }
    if(d->priority < 6) {
        if(plr->attackDown) {
            if(d->lastAttackDown == -1) {
                d->lastAttackDown = FACE_RAMPAGE_TICS;
            }
            else if(!--d->lastAttackDown) {
                d->lastAttackDown = 1;
                d->priority  = 5;
                d->faceTicks = 1;
                d->faceIndex = FACE_RAMPAGE_FIRST + d->painOffset(player());
            }
        }
        else {
            d->lastAttackDown = -1;
        }
    }
    if(d->priority < 5) {
        if((P_GetPlayerCheats(plr) & CF_GODMODE) || plr->powers[PT_INVULNERABILITY]) {
            d->priority  = 4;
            d->faceTicks = 1;
            d->faceIndex = FACE_GOD_FIRST;
        }
    }
    if(!d->faceTicks) {
        d->priority  = 0;
        d->faceTicks = FACE_STRAIGHT_TICS;
        d->faceIndex = d->painOffset(player()) + (M_Random() % 3);
    }
    d->oldHealth = plr->health;
    d->faceTicks -= 1;
}
void Face_Drawer(guidata_face_t *face, Point2Raw const *offset) {
#define X_OFFSET                ( 143 )
#define Y_OFFSET                (   0 )
#define SCALE                   ( 0.7 )
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_FACE]) return;
    if(ST_AutomapIsOpen(face->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[face->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchid_t pFace = ::pFaces[face->d->faceIndex % FACE_COUNT];
    if(!pFace) return;
    dint x = -(SCREENWIDTH / 2 - X_OFFSET);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(SCALE * ::cfg.common.hudScale, SCALE * ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    patchinfo_t bgInfo;
    if(R_GetPatchInfo(::pBackground[::cfg.playerColor[face->player()]], &bgInfo)) {
        if(IS_NETGAME) {
            GL_DrawPatch(bgInfo.id);
        }
        x += bgInfo.geometry.size.width/2;
    }
    GL_DrawPatch(pFace, Vector2i(x, -1));
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef SCALE
#undef Y_OFFSET
#undef X_OFFSET
}
void SBarFace_Drawer(guidata_face_t *face, Point2Raw const *offset) {
#define X_OFFSET                ( 143 )
#define Y_OFFSET                (   0 )
#define SCALE                   ( 1.0 )
    Vector2i const origin(-ST_WIDTH / 2, -ST_HEIGHT);
    dint const activeHud     = ST_ActiveHud(face->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(face->player()));
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(face->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[face->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(face->d->faceIndex < 0) return;
    patchid_t patchId = ::pFaces[face->d->faceIndex];
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(SCALE * ::cfg.common.statusbarScale, SCALE * ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(patchId, origin + Vector2i(X_OFFSET, Y_OFFSET), ALIGN_TOPLEFT);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef SCALE
#undef Y_OFFSET
#undef X_OFFSET
}
void Face_UpdateGeometry(guidata_face_t *face) {
#define SCALE                   ( 0.7 )
    Rect_SetWidthHeight(&face->geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_FACE]) return;
    if(ST_AutomapIsOpen(face->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[face->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t info;
    patchid_t pFace = ::pFaces[face->d->faceIndex % FACE_COUNT];
    if(!pFace) return;
    if(!(R_GetPatchInfo(::pBackground[::cfg.playerColor[face->player()]], &info) ||
         R_GetPatchInfo(pFace, &info))) return;
    Rect_SetWidthHeight(&face->geometry(), info.geometry.size.width  * SCALE * ::cfg.common.hudScale,
                                           info.geometry.size.height * SCALE * ::cfg.common.hudScale);
#undef SCALE
}
void SBarFace_UpdateGeometry(guidata_face_t *face) {
#define SCALE                   ( 1 )
    Rect_SetWidthHeight(&face->geometry(), 0, 0);
    if(ST_AutomapIsOpen(face->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[face->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t info;
    patchid_t pFace = ::pFaces[face->d->faceIndex % FACE_COUNT];
    if(!pFace || !R_GetPatchInfo(pFace, &info)) return;
    Rect_SetWidthHeight(&face->geometry(), info.geometry.size.width  * SCALE * ::cfg.common.hudScale,
                                           info.geometry.size.height * SCALE * ::cfg.common.hudScale);
#undef SCALE
}
void guidata_face_t::prepareAssets()  // static
{
    for(dint i = 0; i < NUMTEAMS; ++i) {
        ::pBackground[i] = R_DeclarePatch(String("STFB%1").arg(i).toUtf8().constData());
    }
    dint idx = 0;
    for(dint i = 0; i < FACE_PAIN_COUNT; ++i) {
        for(dint k = 0; k < FACE_STRAIGHT_COUNT; ++k) {
            ::pFaces[idx++] = R_DeclarePatch(String("STFST%1%2").arg(i).arg(k).toUtf8().constData());
        }
        ::pFaces[idx++] = R_DeclarePatch(String("STFTR%1"  ).arg(i * 10, 2, 'g', 2, '0').toUtf8().constData());  // Turn right.
        ::pFaces[idx++] = R_DeclarePatch(String("STFTL%1"  ).arg(i * 10, 2, 'g', 2, '0').toUtf8().constData());  // Turn left.
        ::pFaces[idx++] = R_DeclarePatch(String("STFOUCH%1").arg(i).toUtf8().constData());  // Ouch.
        ::pFaces[idx++] = R_DeclarePatch(String("STFEVL%1" ).arg(i).toUtf8().constData());  // Evil grin.
        ::pFaces[idx++] = R_DeclarePatch(String("STFKILL%1").arg(i).toUtf8().constData());  // Pissed off.
    }
    ::pFaces[idx++] = R_DeclarePatch("STFGOD0");
    ::pFaces[idx++] = R_DeclarePatch("STFDEAD0");
    DENG_ASSERT(idx == FACE_COUNT);
}
static void HealthIconWidget_Draw(guidata_healthicon_t *icon, Point2Raw const *offset) {
    DENG2_ASSERT(icon);
    icon->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void HealthIconWidget_UpdateGeometry(guidata_healthicon_t *icon) {
    DENG2_ASSERT(icon);
    icon->updateGeometry();
}
guidata_healthicon_t::guidata_healthicon_t(dint player, int sprite)
    : HudWidget(function_cast<UpdateGeometryFunc>(HealthIconWidget_UpdateGeometry),
                function_cast<DrawFunc>(HealthIconWidget_Draw),
                player) {
    this->iconSpriteId = sprite;
}
guidata_healthicon_t::~guidata_healthicon_t()
{}
void guidata_healthicon_t::tick(timespan_t /*elapsed*/)
{
}
void guidata_healthicon_t::draw(Vector2i const &offset) const {
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_HEALTH]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    GUI_DrawSprite(this->iconSpriteId, 0, 0, HOT_TLEFT, 1, iconOpacity, false, nullptr, nullptr);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void guidata_healthicon_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_HEALTH]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    Size2Raw iconSize;
    GUI_SpriteSize(this->iconSpriteId, 1, &iconSize.width, &iconSize.height);
    iconSize.width  *= ::cfg.common.hudScale;
    iconSize.height *= ::cfg.common.hudScale;
    Rect_SetWidthHeight(&geometry(), iconSize.width, iconSize.height);
}
static void MaxAmmoWidget_UpdateGeometry(guidata_maxammo_t *wi) {
    DENG2_ASSERT(wi);
    wi->updateGeometry();
}
static void MaxAmmoWidget_Draw(guidata_maxammo_t *wi, Point2Raw const *offset) {
    DENG2_ASSERT(wi);
    wi->draw(offset? Vector2i(offset->xy) : Vector2i());
}
guidata_maxammo_t::guidata_maxammo_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(MaxAmmoWidget_UpdateGeometry),
                function_cast<DrawFunc>(MaxAmmoWidget_Draw),
                player)
{}
guidata_maxammo_t::~guidata_maxammo_t()
{}
void guidata_maxammo_t::reset() {
    _value = 1994;
}
guidata_maxammo_t &guidata_maxammo_t::setAmmoType(ammotype_t newAmmoType) {
    _ammotype = newAmmoType;
    return *this;
}
void guidata_maxammo_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    _value = plr->ammo[_ammotype].max;
}
void guidata_maxammo_t::draw(Vector2i const &offset) const {
    static Vector2i const origin(-ST_WIDTH / 2, -ST_HEIGHT);
    static Vector2i const offsets[NUM_AMMO_TYPES] = {
        { 314, 5 + 6 * 0 }, { 314, 5 + 6 * 1 }, { 314, 5 + 6 * 3 }, { 314, 5 + 6 * 2 }
    };
    dint const activeHud     = ST_ActiveHud(player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(_value == 1994) return;
    Vector2i const pos     = origin + offsets[dint( _ammotype )];
    auto const valueAsText = String::number(_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(font());
    FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), pos.x, pos.y, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void guidata_maxammo_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    FR_SetFont(font());
    Rect_SetWidthHeight(&geometry(), (FR_CharWidth('0') * 3) * ::cfg.common.statusbarScale,
                                      FR_CharHeight('0') * ::cfg.common.statusbarScale);
}
#define ST_ARMSX                ( 111 )
#define ST_ARMSY                (   4 )
#define ST_ARMSXSPACE           (  12 )
#define ST_ARMSYSPACE           (  10 )
static patchid_t pArms[6][2];
static void WeaponSlotWidget_Draw(guidata_weaponslot_t *ws, Point2Raw const *offset) {
    DENG2_ASSERT(ws);
    ws->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void WeaponSlotWidget_UpdateGeometry(guidata_weaponslot_t *ws) {
    DENG2_ASSERT(ws);
    ws->updateGeometry();
}
struct countownedweaponsinslot_params_t {
    int player;
    int numOwned;
};
static int countOwnedWeaponsInSlot(weapontype_t type, void *context) {
    auto &p = *static_cast<countownedweaponsinslot_params_t *>(context);
    player_t const *plr = &players[p.player];
    if(plr->weapons[type].owned) {
        p.numOwned += 1;
    }
    return 1;  // Continue iteration.
}
guidata_weaponslot_t::guidata_weaponslot_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(WeaponSlotWidget_UpdateGeometry),
                function_cast<DrawFunc>(WeaponSlotWidget_Draw),
                player)
{}
guidata_weaponslot_t::~guidata_weaponslot_t()
{}
void guidata_weaponslot_t::reset() {
    _patchId = 0;
}
guidata_weaponslot_t &guidata_weaponslot_t::setSlot(dint newSlotNum) {
    _slot = newSlotNum + 1;  // 1-based
    return *this;
}
void guidata_weaponslot_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &players[player()];
    bool used = false;
    if(cfg.fixStatusbarOwnedWeapons) {
        countownedweaponsinslot_params_t parm;
        parm.player   = player();
        parm.numOwned = 0;
        P_IterateWeaponsBySlot(_slot, false, countOwnedWeaponsInSlot, &parm);
        used = (parm.numOwned > 0);
    }
    else {
        used = CPP_BOOL(plr->weapons[_slot].owned);
    }
    _patchId = pArms[_slot - 1][dint( used )];
}
void guidata_weaponslot_t::draw(Vector2i const &offset) const {
#define ORIGINX (-ST_WIDTH/2)
#define ORIGINY (-ST_HEIGHT)
    static Vector2i const elements[] = {
        Vector2i( ORIGINX+ST_ARMSX,                     ORIGINY+ST_ARMSY ),
        Vector2i( ORIGINX+ST_ARMSX + ST_ARMSXSPACE,     ORIGINY+ST_ARMSY ),
        Vector2i( ORIGINX+ST_ARMSX + ST_ARMSXSPACE*2,   ORIGINY+ST_ARMSY ),
        Vector2i( ORIGINX+ST_ARMSX,                     ORIGINY+ST_ARMSY + ST_ARMSYSPACE ),
        Vector2i( ORIGINX+ST_ARMSX + ST_ARMSXSPACE,     ORIGINY+ST_ARMSY + ST_ARMSYSPACE ),
        Vector2i( ORIGINX+ST_ARMSX + ST_ARMSXSPACE*2,   ORIGINY+ST_ARMSY + ST_ARMSYSPACE ),
    };
    Vector2i const &element = elements[_slot - 1];
    int const activeHud     = ST_ActiveHud(player());
    int const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
    float const textOpacity = (activeHud == 0? 1 : uiRendState->pageAlpha * cfg.common.statusbarCounterAlpha);
    if(gfw_Rule(deathmatch)) return;
    if(ST_AutomapIsOpen(player()) && cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(cfg.common.statusbarScale, cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, textOpacity);
    FR_SetFont(font());
    if(gameMode == doom_chex) {
        FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    }
    else {
        FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    }
    WI_DrawPatch(_patchId, Hu_ChoosePatchReplacement(patchreplacemode_t(cfg.common.hudPatchReplaceMode), _patchId),
                 element, ALIGN_TOPLEFT, 0, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef ORIGINY
#undef ORIGINX
}
void guidata_weaponslot_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(gfw_Rule(deathmatch)) return;
    if(ST_AutomapIsOpen(player()) && cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t info;
    String const text = Hu_ChoosePatchReplacement(patchreplacemode_t(cfg.common.hudPatchReplaceMode), _patchId);
    if(text.isEmpty() && !R_GetPatchInfo(_patchId, &info)) return;
    if(!text.isEmpty()) {
        FR_SetFont(font());
        Size2Raw textSize; FR_TextSize(&textSize, text.toUtf8().constData());
        Rect_SetWidthHeight(&geometry(), textSize.width  * cfg.common.statusbarScale,
                                         textSize.height * cfg.common.statusbarScale);
        return;
    }
    R_GetPatchInfo(_patchId, &info);
    Rect_SetWidthHeight(&geometry(), info.geometry.size.width  * cfg.common.statusbarScale,
                                     info.geometry.size.height * cfg.common.statusbarScale);
}
void guidata_weaponslot_t::prepareAssets()  // static
{
    char nameBuf[9];
    for(int i = 0; i < 6; ++i) {
        sprintf(nameBuf, "STGNUM%d", i + 2);
        pArms[i][0] = R_DeclarePatch(nameBuf);
        sprintf(nameBuf, "STYSNUM%d", i + 2);
        pArms[i][1] = R_DeclarePatch(nameBuf);
    }
}
