﻿/** @file g_defs.cpp  Game definition lookup utilities.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "g_defs.h"
#include <de/RecordValue>
#include <doomsday/defs/episode.h>
/** @file g_eventsequence.cpp
 * Input (keyboard) event sequences. @ingroup libcommon
 *
 * @author Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @author Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 * @author Copyright &copy; 1999 Activision
 * @author Copyright &copy; 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <cstring>
#include <vector>
#include "common.h"
#include "g_eventsequence.h"
/** @file g_game.c  Top-level (common) game routines.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Activision
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "g_common.h"
#include <cctype>
#include <cstring>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <de/App>
#include <de/CommandLine>
#include <de/LogBuffer>
#include <de/NativePath>
#include <de/RecordValue>
#include <doomsday/DoomsdayApp>
#include <doomsday/defs/episode.h>
#include <doomsday/defs/mapinfo.h>
#include <doomsday/busymode.h>
#include <doomsday/uri.h>
#include "acs/system.h"
#include "animdefs.h"
#include "d_net.h"
#include "d_netcl.h"
#include "d_netsv.h"
#include "dmu_lib.h"
#include "fi_lib.h"
#include "g_controls.h"
#include "g_defs.h"
#include "g_eventsequence.h"
#include "g_update.h"
#include "gamesession.h"
#include "hu_lib.h"
#include "hu_inventory.h"
#include "hu_menu.h"
#include "hu_msg.h"
#include "hu_pspr.h"
#include "hu_stuff.h"
#include "p_actor.h"
#include "p_inventory.h"
#include "p_map.h"
#include "p_mapsetup.h"
#include "p_mapspec.h"
#include "p_savedef.h"
#include "p_saveio.h"
#include "p_saveg.h"
#include "p_sound.h"
#include "p_start.h"
#include "p_tick.h"
#include "p_user.h"
#include "player.h"
#include "r_common.h"
#include "r_special.h"
#include "saveslots.h"
#include "x_hair.h"
#include "menu/widgets/widget.h"
/** @file g_update.cpp  Engine reset => game update logic.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "g_update.h"
#include <cctype>
#include "animdefs.h"
#include "g_common.h"
#include "gamesession.h"
#include "hu_menu.h"
#include "hu_pspr.h"
#include "hu_stuff.h"
#include "p_inventory.h"
#include "p_sound.h"
#include "p_start.h"
#include "r_common.h"
#include "saveslots.h"
/** @file gamerules.cpp  Game rule set.
 *
 * @authors Copyright © 2003-2018 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "gamerules.h"
#include "gamesession.h"
/** @file gamesession.cpp  Logical game session and saved session marshalling.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "gamesession.h"
#include <de/App>
#include <de/CommandLine>
#include <de/ArrayValue>
#include <de/NumberValue>
#include <de/RecordValue>
#include <de/PackageLoader>
#include <de/Time>
#include <de/TextValue>
#include <de/ZipArchive>
#include <doomsday/DoomsdayApp>
#include <doomsday/GameStateFolder>
#include <doomsday/SaveGames>
#include <doomsday/defs/episode.h>
#include "acs/system.h"
#include "api_gl.h"
#include "d_netsv.h"
#include "g_common.h"
#include "g_defs.h"
#include "g_game.h"
#include "hu_menu.h"
#include "hu_inventory.h"
#include "mapstatewriter.h"
#include "p_inventory.h"
#include "p_map.h"
#include "p_mapsetup.h"
#include "p_savedef.h"
#include "p_saveg.h"
#include "p_saveio.h"
#include "p_sound.h"
#include "p_tick.h"
#include "r_common.h"
#if __JDOOM__
#  include "doomv9mapstatereader.h"
#endif
#if __JHERETIC__
#  include "hereticv13mapstatereader.h"
#endif
/** @file saveslots.cpp  Map of logical saved game session slots.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "saveslots.h"
#include "g_common.h"
#include "gamesession.h"
#include "hu_menu.h"
#include "menu/page.h"
#include "menu/widgets/lineeditwidget.h"
#include <doomsday/SaveGames>
#include <de/App>
#include <de/Folder>
#include <de/LogBuffer>
#include <de/Observers>
#include <de/Writer>
#include <de/Loop>
#include <map>
#include <utility>
using namespace de;
using namespace common::menu;
using namespace common;
ded_t &Defs() {
    return *static_cast<ded_t *>(DD_GetVariable(DD_DEFS));
}
dint PlayableEpisodeCount() {
    dint count = 0;
    DictionaryValue::Elements const &episodesById = Defs().episodes.lookup("id").elements();
    for(auto const &pair : episodesById) {
        Record const &episodeDef = *pair.second->as<RecordValue>().record();
        de::Uri startMap(episodeDef.gets("startMap"), RC_NULL);
        if(P_MapExists(startMap.compose().toUtf8().constData())) {
            count += 1;
        }
    }
    return count;
}
String FirstPlayableEpisodeId() {
    DictionaryValue::Elements const &episodesById = Defs().episodes.lookup("id").elements();
    for(auto const &pair : episodesById) {
        Record const &episodeDef = *pair.second->as<RecordValue>().record();
        de::Uri startMap(episodeDef.gets("startMap"), RC_NULL);
        if(P_MapExists(startMap.compose().toUtf8().constData())) {
            return episodeDef.gets("id");
        }
    }
    return "";  // Not found.
}
de::Uri TranslateMapWarpNumber(String const &episodeId, dint warpNumber) {
    if(Record const *rec = Defs().episodes.tryFind("id", episodeId)) {
        defn::Episode episodeDef(*rec);
        if(Record const *mgNodeRec = episodeDef.tryFindMapGraphNodeByWarpNumber(warpNumber)) {
            return de::makeUri(mgNodeRec->gets("id"));
        }
    }
    return de::makeUri("Maps:");  // Not found.
}
class ISequenceCompleteHandler {
public:
    virtual ~ISequenceCompleteHandler() {}
    virtual void invoke(int player, EventSequenceArg* args, int numArgs) = 0;
};
class SequenceCompleteHandler : public ISequenceCompleteHandler {
public:
    SequenceCompleteHandler(eventsequencehandler_t _callback) : callback(_callback) {}
    virtual void invoke(int player, EventSequenceArg* args, int numArgs) {
        callback(player, args, numArgs);
    }
private:
    eventsequencehandler_t callback;
};
class SequenceCompleteCommandHandler : public ISequenceCompleteHandler {
public:
    SequenceCompleteCommandHandler(const char* _commandTemplate) {
        Str_Set(Str_InitStd(&commandTemplate), _commandTemplate);
    }
    virtual ~SequenceCompleteCommandHandler() {
        Str_Free(&commandTemplate);
    }
    virtual void invoke(int player, EventSequenceArg* args, int numArgs) {
        if(!strchr(Str_Text(&commandTemplate), '%')) {
            DD_Execute(true/*silent*/, Str_Text(&commandTemplate));
            return;
        }
        AutoStr* cmd = AutoStr_NewStd();
        Str_Reserve(cmd, Str_Length(&commandTemplate) + numArgs + 1);
        int len = Str_Length(&commandTemplate);
        const char* start = Str_Text(&commandTemplate);
        const char* ch = start, *substart = start;
        while(ch + 1 < start + len) {
            if(ch[0] == '%' && ch[1] && ch[1] != '%') {
                Str_PartAppend(cmd, substart, 0, ch - substart);
                if(ch[1] == 'p') {
                    Str_AppendChar(cmd, '0' + player);
                }
                else {
                    int arg = ch[1] - '0' - 1;
                    DENG_ASSERT(arg >= 0 && arg < 9);
                    Str_AppendChar(cmd, char(args[arg]));
                }
                ch += 2;
                substart = ch;
            }
            else {
                ch++;
            }
        }
        Str_Append(cmd, substart);
        DD_Execute(true/*silent*/, Str_Text(cmd));
    }
private:
    ::Str commandTemplate;
};
class EventSequence {
public:
    EventSequence(const char* _sequence, ISequenceCompleteHandler& _handler)
        : sequence(), handler(_handler), pos(0), numArgs(0), args(0) {
        int len = strlen(_sequence);
        if(strchr(_sequence, '%')) {
            const char* ch = _sequence;
            while(ch + 1 < _sequence + len) {
                if(ch[0] == '%' && ch[1] && ch[1] != '%') {
                    int arg = ch[1] - '0';
                    if(arg < 1 || arg > 9) {
                        App_Log(DE2_DEV_WARNING, "EventSequence: Sequence %s truncated due to bad suffix %c",
                                _sequence, ch[1]);
                        len = ch - _sequence;
                        break;
                    }
                    numArgs += 1;
                    ch += 2;
                }
                else {
                    ch++;
                }
            }
        }
        Str_PartAppend(Str_Init(&sequence), _sequence, 0, len);
        if(numArgs) {
            args = new EventSequenceArg[numArgs];
            for(int i = 0; i < numArgs; ++i) {
                args[i] = 0;
            }
        }
    }
    ~EventSequence() {
        Str_Free(&sequence);
        if(args) delete[] args;
        delete &handler;
    }
    /**
     * Rewind the sequence and forget any current progress (and arguments).
     * @return
     */
    EventSequence& rewind() {
        pos = 0;
        return *this;
    }
    bool complete(event_t* ev, int player, bool* eat) {
        DENG_ASSERT(ev && ev->type == EV_KEY && ev->state == EVS_DOWN);
        const int key = ev->data1;
        if(Str_At(&sequence, pos  ) == '%' && pos+1 < Str_Length(&sequence) &&
           Str_At(&sequence, pos+1) >= '0' && Str_At(&sequence, pos+1) <= '9') {
            const int arg = Str_At(&sequence, pos+1) - '0' - 1;
            DENG_ASSERT(arg >= 0 && arg < numArgs);
            EventSequenceArg& currentArg = args[arg];
            currentArg = key;
            pos += 2;
            if(eat) *eat = true;
        }
        else if(key == Str_At(&sequence, pos)) {
            pos++;
            if(eat) *eat = false;
        }
        else {
            rewind();
        }
        if(pos < Str_Length(&sequence)) return false;
        handler.invoke(player, args, numArgs);
        rewind();
        return true;
    }
private:
    ::Str sequence;
    ISequenceCompleteHandler& handler;
    int pos;
    int numArgs;
    EventSequenceArg* args;
};
typedef std::vector<EventSequence*> EventSequences;
static bool inited = false;
static EventSequences sequences;
static void clearSequences() {
    for(EventSequences::iterator i = sequences.begin(); i != sequences.end(); ++i) {
        delete *i;
    }
    sequences.clear();
}
void G_InitEventSequences(void) {
    if(inited) {
        clearSequences();
    }
    inited = true;
}
void G_ShutdownEventSequences(void) {
    if(!inited) return;
    clearSequences();
    inited = false;
}
int G_EventSequenceResponder(event_t* ev) {
    DENG_ASSERT(inited);
    if(!inited) return false;
    if(!ev || ev->type != EV_KEY || ev->state != EVS_DOWN) return false;
    const int player = CONSOLEPLAYER; /// @todo Should be defined by the event...
    bool eventWasEaten = false;
    for(EventSequences::iterator i = sequences.begin(); i != sequences.end(); ++i) {
        EventSequence* seq = *i;
        if(seq->complete(ev, player, &eventWasEaten)) return true;
    }
    return eventWasEaten;
}
void G_AddEventSequence(const char* sequence, eventsequencehandler_t callback) {
    if(!inited) Con_Error("G_AddEventSequence: Subsystem not presently initialized.");
    if(!sequence || !sequence[0] || !callback) Con_Error("G_AddEventSequence: Invalid argument(s).");
    SequenceCompleteHandler* handler = new SequenceCompleteHandler(callback);
    sequences.push_back(new EventSequence(sequence, *handler));
}
void G_AddEventSequenceCommand(const char* sequence, const char* commandTemplate) {
    if(!inited) Con_Error("G_AddEventSequenceCommand: Subsystem not presently initialized.");
    if(!sequence || !sequence[0] || !commandTemplate || !commandTemplate[0]) Con_Error("G_AddEventSequenceCommand: Invalid argument(s).");
    SequenceCompleteCommandHandler* handler = new SequenceCompleteCommandHandler(commandTemplate);
    sequences.push_back(new EventSequence(sequence, *handler));
}
void R_LoadVectorGraphics();
int Hook_DemoStop(int hookType, int val, void *parm);
game_config_t cfg; // The global cfg.
#if __JDOOM__ || __JDOOM64__
#define BODYQUEUESIZE       (32)
mobj_t *bodyQueue[BODYQUEUESIZE];
int bodyQueueSlot;
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
bool secretExit;
#endif
dd_bool monsterInfight;
player_t players[MAXPLAYERS];
int totalKills, totalItems, totalSecret; // For intermission.
dd_bool singledemo; // Quit after playing a demo from cmdline.
dd_bool briefDisabled;
dd_bool precache = true; // If @c true, load all graphics at start.
dd_bool customPal; // If @c true, a non-IWAD palette is in use.
wbstartstruct_t wmInfo; // Intermission parameters.
de::Uri nextMapUri;
uint nextMapEntryPoint;
static bool quitInProgress;
static gamestate_t gameState = GS_STARTUP;
static SaveSlots *sslots;
static gameaction_t gameAction;
static GameRules &G_NewSessionRules() {
    static GameRules gaNewSessionRules;
    return gaNewSessionRules;
}
static String  gaNewSessionEpisodeId;
static de::Uri gaNewSessionMapUri;
static uint    gaNewSessionMapEntrance;
static String gaSaveSessionSlot;
static bool   gaSaveSessionGenerateDescription = true;
static String gaSaveSessionUserDescription;
static String gaLoadSessionSlot;
dd_bool G_QuitInProgress() {
    return ::quitInProgress;
}
void G_SetGameAction(gameaction_t newAction) {
    if (G_QuitInProgress()) return;
    if (::gameAction != newAction) {
        ::gameAction = newAction;
    }
}
void G_SetGameActionNewSession(GameRules const &rules, String episodeId, de::Uri const &mapUri,
                               uint mapEntrance) {
    G_NewSessionRules()       = rules;
    ::gaNewSessionEpisodeId   = episodeId;
    ::gaNewSessionMapUri      = mapUri;
    ::gaNewSessionMapEntrance = mapEntrance;
    G_SetGameAction(GA_NEWSESSION);
}
bool G_SetGameActionSaveSession(String slotId, String *userDescription) {
    if (!gfw_Session()->isSavingPossible()) return false;
    if (!G_SaveSlots().has(slotId)) return false;
    ::gaSaveSessionSlot = slotId;
    if (userDescription && !userDescription->isEmpty()) {
        ::gaSaveSessionGenerateDescription = false;
        ::gaSaveSessionUserDescription = *userDescription;
    }
    else {
        ::gaSaveSessionGenerateDescription = (userDescription && userDescription->isEmpty());
        ::gaSaveSessionUserDescription.clear();
    }
    G_SetGameAction(GA_SAVESESSION);
    return true;
}
bool G_SetGameActionLoadSession(String slotId) {
    if (!gfw_Session()->isLoadingPossible()) return false;
    auto scheduleLoad = [slotId] () {
        if (G_SaveSlots()[slotId].isLoadable()) {
            gaLoadSessionSlot = slotId;
            G_SetGameAction(GA_LOADSESSION);
        }
        else {
            LOG_RES_ERROR("Cannot load from save slot '%s': not in use") << slotId;
        }
    };
    try {
        auto const &slot = G_SaveSlots()[slotId];
        GameStateFolder const &save = App::rootFolder().locate<GameStateFolder const>(slot.savePath());
        Record const &meta = save.metadata();
        if (meta.has("packages")) {
            DoomsdayApp::app().checkPackageCompatibility(
                meta.getStringList("packages"),
                String::format("The savegame " _E(b) "%s" _E(.) " was created with "
                               "mods that are different than the ones currently in use.",
                               meta.gets("userDescription").toUtf8().constData()),
                scheduleLoad);
        }
        else {
            scheduleLoad();
        }
    }
    catch (SaveSlots::MissingSlotError const &er) {
        LOG_RES_WARNING("Save slot '%s' not found: %s") << slotId << er.asText();
        return false;
    }
    return true;
}
void G_SetGameActionMapCompleted(de::Uri const &nextMapUri, uint nextMapEntryPoint, bool secretExit) {
#if __JHEXEN__
    DENG2_UNUSED(secretExit);
#else
    DENG2_UNUSED(nextMapEntryPoint);
#endif
    if (IS_CLIENT) return;
    if (::cyclingMaps && ::mapCycleNoExit) return;
#if __JHEXEN__
    if ((::gameMode == hexen_betademo || ::gameMode == hexen_demo) &&
       !(nextMapUri.path() == "MAP01" ||
         nextMapUri.path() == "MAP02" ||
         nextMapUri.path() == "MAP03" ||
         nextMapUri.path() == "MAP04")) {
        P_SetMessage(&players[CONSOLEPLAYER], "PORTAL INACTIVE -- DEMO");
        return;
    }
#endif
    ::nextMapUri        = nextMapUri;
#if __JHEXEN__
    ::nextMapEntryPoint = nextMapEntryPoint;
#else
    ::secretExit        = secretExit;
# if __JDOOM__
    if (::secretExit && (::gameModeBits & GM_ANY_DOOM2)) {
        if (!P_MapExists(de::makeUri("Maps:MAP31").compose().toUtf8().constData())) {
            ::secretExit = false;
        }
    }
# endif
#endif
    G_SetGameAction(GA_MAPCOMPLETED);
}
void G_SetGameActionMapCompletedAndSetNextMap() {
    G_SetGameActionMapCompleted(gfw_Session()->mapUriForNamedExit("next"));
}
static void initSaveSlots() {
    delete sslots;
    sslots = new SaveSlots;
    menu::Widget::Flag const gameMenuSaveSlotWidgetIds[NUMSAVESLOTS] = {
        menu::Widget::Id0, menu::Widget::Id1, menu::Widget::Id2,
        menu::Widget::Id3, menu::Widget::Id4, menu::Widget::Id5,
#if !__JHEXEN__
        menu::Widget::Id6, menu::Widget::Id7
#endif
    };
    for (int i = 0; i < NUMSAVESLOTS; ++i) {
        sslots->add(String::number(i), true, String(SAVEGAMENAME"%1").arg(i),
                    int(gameMenuSaveSlotWidgetIds[i]));
    }
}
/**
 * Common Pre Game Initialization routine.
 * Game-specfic pre init actions should be placed in eg D_PreInit() (for jDoom)
 */
void G_CommonPreInit() {
    ::quitInProgress = false;
    cfg.common.pushableMomentumLimitedToPusher = true;
    gfw_Session()->applyNewRules(gfw_DefaultGameRules() = GameRules());
    Plug_AddHook(HOOK_DEMO_STOP, Hook_DemoStop);
    for (int i = 0; i < MAXPLAYERS; ++i) {
        player_t *pl = &players[i];
        pl->plr = DD_GetPlayer(i);
        pl->plr->extraData = (void *) pl;
        for (int k = 0; k < NUMPSPRITES; ++k) {
            pl->pSprites[k].state = nullptr;
            pl->plr->pSprites[k].statePtr = nullptr;
        }
    }
    G_RegisterBindClasses();
    P_RegisterMapObjs();
    R_LoadVectorGraphics();
    R_LoadColorPalettes();
    P_InitPicAnims();
    G_ConsoleRegistration();      // Main command list.
    acs::System::consoleRegister();
    D_NetConsoleRegister();
    G_ConsoleRegister();
    Pause_Register();
    G_ControlRegister();
    SaveSlots::consoleRegister();
    Hu_MenuConsoleRegister();
    GUI_Register();
    Hu_MsgRegister();
    ST_Register();                // For the hud/statusbar.
    IN_ConsoleRegister();         // For the interlude/intermission.
    X_Register();                 // For the crosshair.
    FI_StackRegister();           // For the InFine lib.
    R_SpecialFilterRegister();
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    XG_Register();
#endif
    Con_SetString2("map-author", "Unknown", SVF_WRITE_OVERRIDE);
    Con_SetString2("map-name",   "Unknown", SVF_WRITE_OVERRIDE);
}
#if __JHEXEN__
/**
 * @todo all this swapping colors around is rather silly, why not simply
 * reorder the translation tables at load time?
 */
void R_GetTranslation(int plrClass, int plrColor, int *tclass, int *tmap) {
    int mapped;
    if (plrClass == PCLASS_PIG) {
        *tclass = *tmap = 0;
        return;
    }
    if (gameMode == hexen_v10) {
        int const mapping[3][4] = {
            /* Fighter */ { 1, 2, 0, 3 },
            /* Cleric */  { 1, 0, 2, 3 },
            /* Mage */    { 1, 0, 2, 3 }
        };
        DENG2_ASSERT(plrClass >= 0 && plrClass < 3);
        DENG2_ASSERT(plrColor >= 0 && plrColor < 4);
        mapped = mapping[plrClass][plrColor];
    }
    else {
        int const mapping[3][8] = {
            /* Fighter */ { 1, 2, 0, 3, 4, 5, 6, 7 },
            /* Cleric */  { 1, 0, 2, 3, 4, 5, 6, 7 },
            /* Mage */    { 1, 0, 2, 3, 4, 5, 6, 7 }
        };
        DENG2_ASSERT(plrClass >= 0 && plrClass < 3);
        DENG2_ASSERT(plrColor >= 0 && plrColor < 8);
        mapped = mapping[plrClass][plrColor];
    }
    *tclass = (mapped? plrClass : 0);
    *tmap   = mapped;
}
void Mobj_UpdateTranslationClassAndMap(mobj_t *mo) {
    DENG2_ASSERT(mo);
    if (mo->player) {
        int plrColor = (mo->flags & MF_TRANSLATION) >> MF_TRANSSHIFT;
        R_GetTranslation(mo->player->class_, plrColor, &mo->tclass, &mo->tmap);
    }
    else if (mo->flags & MF_TRANSLATION) {
        mo->tclass = mo->special1;
        mo->tmap   = (mo->flags & MF_TRANSLATION) >> MF_TRANSSHIFT;
    }
    else {
        mo->tmap = mo->tclass = 0;
    }
}
#endif // __JHEXEN__
void R_LoadColorPalettes() {
#define PALLUMPNAME         "PLAYPAL"
#define PALENTRIES          (256)
#define PALID               (0)
    File1 &playpal = CentralLumpIndex()[CentralLumpIndex().findLast(String(PALLUMPNAME) + ".lmp")];
    customPal = playpal.hasCustom(); // Remember whether we are using a custom palette.
    uint8_t colors[PALENTRIES * 3];
    playpal.read(colors, 0 + PALID * (PALENTRIES * 3), PALENTRIES * 3);
    colorpaletteid_t palId = R_CreateColorPalette("R8G8B8", PALLUMPNAME, colors, PALENTRIES);
    ddstring_s xlatId; Str_InitStd(&xlatId);
#if __JHEXEN__
    {
        int const numPerClass = (gameMode == hexen_v10? 3 : 7);
        int xlatNum = 0;
        for (int cl = 0; cl < 3; ++cl)
        for (int i = 0; i < 7; ++i) {
            if (i == numPerClass) break; // Not present.
            String lumpName;
            if (xlatNum < 10) {
                lumpName = String("TRANTBL%1").arg(xlatNum);
            }
            else {
                lumpName = String::format("TRANTBL%c", 'A' + (xlatNum - 10));
            }
            xlatNum++;
            LOG_AS("R_LoadColorPalettes")
            LOG_RES_XVERBOSE("Reading translation table '%s' as tclass=%i tmap=%i",
                             lumpName << cl << i);
            lumpName += ".lmp";
            if (CentralLumpIndex().contains(lumpName)) {
                File1 &lump = CentralLumpIndex()[CentralLumpIndex().findLast(lumpName)];
                uint8_t const *mappings = lump.cache();
                Str_Appendf(Str_Clear(&xlatId), "%i", 7 * cl + i);
                R_CreateColorPaletteTranslation(palId, &xlatId, mappings);
                lump.unlock();
            }
        }
    }
#else
    {
        uint8_t xlat[PALENTRIES];
        for (int xlatNum = 0; xlatNum < 3; ++xlatNum) {
            for (int palIdx = 0; palIdx < 256; ++palIdx) {
#  if __JHERETIC__
                if (palIdx >= 225 && palIdx <= 240) {
                    xlat[palIdx] = xlatNum == 0? 114 + (palIdx - 225) /*yellow*/ :
                                   xlatNum == 1? 145 + (palIdx - 225) /*red*/ :
                                                 190 + (palIdx - 225) /*blue*/;
                }
#  else
                if (palIdx >= 0x70 && palIdx <= 0x7f) {
                    xlat[palIdx] = xlatNum == 0? 0x60 + (palIdx & 0xf) :
                                   xlatNum == 1? 0x40 + (palIdx & 0xf) :
                                                 0x20 + (palIdx & 0xf);
                }
#  endif
                else {
                    xlat[palIdx] = palIdx;
                }
            }
            Str_Appendf(Str_Clear(&xlatId), "%i", xlatNum);
            R_CreateColorPaletteTranslation(palId, &xlatId, xlat);
        }
    }
#endif
    Str_Free(&xlatId);
#undef PALID
#undef PALENTRIES
#undef PALLUMPNAME
}
/**
 * @todo Read this information from a definition (ideally with more user
 *       friendly mnemonics...).
 */
void R_LoadVectorGraphics() {
#define R           (1.0f)
#define NUMITEMS(x) (sizeof(x)/sizeof((x)[0]))
#define Pt           Point2Rawf
    Point2Rawf const keyPoints[] = {
        Pt(-3 * R / 4, 0), Pt(-3 * R / 4, -R / 4), // Mid tooth.
        Pt(    0,      0), Pt(   -R,      0), Pt(   -R, -R / 2), // Shaft and end tooth.
        Pt(    0,      0), Pt(R / 4, -R / 2), // Bow.
        Pt(R / 2, -R / 2), Pt(R / 2,  R / 2),
        Pt(R / 4,  R / 2), Pt(    0,      0),
    };
    def_svgline_t const key[] = {
        { 2, &keyPoints[ 0] },
        { 3, &keyPoints[ 2] },
        { 6, &keyPoints[ 5] }
    };
    Point2Rawf const thintrianglePoints[] = {
        Pt(-R / 2,  R - R / 2),
        Pt(     R,          0), // `
        Pt(-R / 2, -R + R / 2), // /
        Pt(-R / 2,  R - R / 2) // |>
    };
    def_svgline_t const thintriangle[] = {
        { 4, thintrianglePoints },
    };
#if __JDOOM__ || __JDOOM64__
    Point2Rawf const arrowPoints[] = {
        Pt(    -R + R / 8, 0 ), Pt(             R, 0), // -----
        Pt( R - R / 2, -R / 4), Pt(             R, 0), Pt( R - R / 2,  R / 4), // ----->
        Pt(-R - R / 8, -R / 4), Pt(    -R + R / 8, 0), Pt(-R - R / 8,  R / 4), // >---->
        Pt(-R + R / 8, -R / 4), Pt(-R + 3 * R / 8, 0), Pt(-R + R / 8,  R / 4), // >>--->
    };
    def_svgline_t const arrow[] = {
        { 2, &arrowPoints[ 0] },
        { 3, &arrowPoints[ 2] },
        { 3, &arrowPoints[ 5] },
        { 3, &arrowPoints[ 8] }
    };
#elif __JHERETIC__ || __JHEXEN__
    Point2Rawf const arrowPoints[] = {
        Pt(-R + R / 4,      0), Pt(         0,      0), // center line.
        Pt(-R + R / 4,  R / 8), Pt(         R,      0), Pt(-R + R / 4, -R / 8), // blade
        Pt(-R + R / 8, -R / 4), Pt(-R + R / 4, -R / 4), // guard
        Pt(-R + R / 4,  R / 4), Pt(-R + R / 8,  R / 4),
        Pt(-R + R / 8, -R / 4),
        Pt(-R + R / 8, -R / 8), Pt(-R - R / 4, -R / 8), // hilt
        Pt(-R - R / 4,  R / 8), Pt(-R + R / 8,  R / 8),
    };
    def_svgline_t const arrow[] = {
        { 2, &arrowPoints[ 0] },
        { 3, &arrowPoints[ 2] },
        { 5, &arrowPoints[ 5] },
        { 4, &arrowPoints[10] }
    };
#endif
#if __JDOOM__
    Point2Rawf const cheatarrowPoints[] = {
        Pt(    -R + R / 8, 0 ), Pt(             R, 0), // -----
        Pt( R - R / 2, -R / 4), Pt(             R, 0), Pt( R - R / 2,  R / 4), // ----->
        Pt(-R - R / 8, -R / 4), Pt(    -R + R / 8, 0), Pt(-R - R / 8,  R / 4), // >---->
        Pt(-R + R / 8, -R / 4), Pt(-R + 3 * R / 8, 0), Pt(-R + R / 8,  R / 4), // >>--->
        Pt(        -R / 2,      0), Pt(        -R / 2, -R / 6), // >>-d--->
        Pt(-R / 2 + R / 6, -R / 6), Pt(-R / 2 + R / 6,  R / 4),
        Pt(        -R / 6,      0), Pt(        -R / 6, -R / 6), // >>-dd-->
        Pt(             0, -R / 6), Pt(             0,  R / 4),
        Pt(         R / 6,  R / 4), Pt(         R / 6, -R / 7), // >>-ddt->
        Pt(R / 6 + R / 32, -R / 7 - R / 32), Pt(R / 6 + R / 10, -R / 7)
    };
    def_svgline_t const cheatarrow[] = {
        { 2, &cheatarrowPoints[ 0] },
        { 3, &cheatarrowPoints[ 2] },
        { 3, &cheatarrowPoints[ 5] },
        { 3, &cheatarrowPoints[ 8] },
        { 4, &cheatarrowPoints[11] },
        { 4, &cheatarrowPoints[15] },
        { 4, &cheatarrowPoints[19] }
    };
#endif
    Point2Rawf const crossPoints[] = { // + (open center)
        Pt(-R,  0), Pt(-R / 5 * 2,          0),
        Pt( 0, -R), Pt(         0, -R / 5 * 2),
        Pt( R,  0), Pt( R / 5 * 2,          0),
        Pt( 0,  R), Pt(         0,  R / 5 * 2)
    };
    def_svgline_t const cross[] = {
        { 2, &crossPoints[0] },
        { 2, &crossPoints[2] },
        { 2, &crossPoints[4] },
        { 2, &crossPoints[6] }
    };
    Point2Rawf const twinanglesPoints[] = { // > <
        Pt(-R, -R * 10 / 14), Pt(-(R - (R * 10 / 14)), 0), Pt(-R,  R * 10 / 14), // >
        Pt( R, -R * 10 / 14), Pt(  R - (R * 10 / 14) , 0), Pt( R,  R * 10 / 14), // <
    };
    def_svgline_t const twinangles[] = {
        { 3, &twinanglesPoints[0] },
        { 3, &twinanglesPoints[3] }
    };
    Point2Rawf const squarePoints[] = { // square
        Pt(-R, -R), Pt(-R,  R),
        Pt( R,  R), Pt( R, -R),
        Pt(-R, -R)
    };
    def_svgline_t const square[] = {
        { 5, squarePoints },
    };
    Point2Rawf const squarecornersPoints[] = { // square (open center)
        Pt(   -R, -R / 2), Pt(-R, -R), Pt(-R / 2,      -R), // topleft
        Pt(R / 2,     -R), Pt( R, -R), Pt(     R,  -R / 2), // topright
        Pt(   -R,  R / 2), Pt(-R,  R), Pt(-R / 2,       R), // bottomleft
        Pt(R / 2,      R), Pt( R,  R), Pt(     R,   R / 2), // bottomright
    };
    def_svgline_t const squarecorners[] = {
        { 3, &squarecornersPoints[ 0] },
        { 3, &squarecornersPoints[ 3] },
        { 3, &squarecornersPoints[ 6] },
        { 3, &squarecornersPoints[ 9] }
    };
    Point2Rawf const anglePoints[] = { // v
        Pt(-R, -R), Pt(0,  0), Pt(R, -R)
    };
    def_svgline_t const angle[] = {
        { 3, anglePoints }
    };
    if (IS_DEDICATED) return;
    R_NewSvg(VG_KEY, key, NUMITEMS(key));
    R_NewSvg(VG_TRIANGLE, thintriangle, NUMITEMS(thintriangle));
    R_NewSvg(VG_ARROW, arrow, NUMITEMS(arrow));
#if __JDOOM__
    R_NewSvg(VG_CHEATARROW, cheatarrow, NUMITEMS(cheatarrow));
#endif
    R_NewSvg(VG_XHAIR1, cross, NUMITEMS(cross));
    R_NewSvg(VG_XHAIR2, twinangles, NUMITEMS(twinangles));
    R_NewSvg(VG_XHAIR3, square, NUMITEMS(square));
    R_NewSvg(VG_XHAIR4, squarecorners, NUMITEMS(squarecorners));
    R_NewSvg(VG_XHAIR5, angle, NUMITEMS(angle));
#undef P
#undef R
#undef NUMITEMS
}
/**
 * @param name  Name of the font to lookup.
 * @return  Unique id of the found font.
 */
fontid_t R_MustFindFontForName(char const *name) {
    uri_s *uri = Uri_NewWithPath2(name, RC_NULL);
    fontid_t fontId = Fonts_ResolveUri(uri);
    Uri_Delete(uri);
    if (fontId) return fontId;
    Con_Error("Failed loading font \"%s\".", name);
    exit(1); // Unreachable.
}
void R_InitRefresh() {
    if (IS_DEDICATED) return;
    LOG_RES_VERBOSE("Loading data for refresh...");
    cfg.common.screenBlocks = cfg.common.setBlocks; {
        uri_s *paths[9];
        for (int i = 0; i < 9; ++i) {
            paths[i] = ((borderGraphics[i] && borderGraphics[i][0])? Uri_NewWithPath2(borderGraphics[i], RC_NULL) : 0);
        }
        R_SetBorderGfx((uri_s const **)paths);
        for (int i = 0; i < 9; ++i) {
            if (paths[i]) {
                Uri_Delete(paths[i]);
            }
        }
    }
    R_ResizeViewWindow(RWF_FORCE|RWF_NO_LERP);
    fonts[GF_FONTA]    = R_MustFindFontForName("a");
    fonts[GF_FONTB]    = R_MustFindFontForName("b");
    fonts[GF_STATUS]   = R_MustFindFontForName("status");
#if __JDOOM__
    fonts[GF_INDEX]    = R_MustFindFontForName("index");
#endif
#if __JDOOM__ || __JDOOM64__
    fonts[GF_SMALL]    = R_MustFindFontForName("small");
#endif
#if __JHERETIC__ || __JHEXEN__
    fonts[GF_SMALLIN]  = R_MustFindFontForName("smallin");
#endif
    fonts[GF_MAPPOINT] = R_MustFindFontForName("mappoint");
    float mul = 1.4f;
    DD_SetVariable(DD_PSPRITE_LIGHTLEVEL_MULTIPLIER, &mul);
}
void R_InitHud() {
    Hu_LoadData();
#if __JHERETIC__ || __JHEXEN__
    LOG_VERBOSE("Initializing inventory...");
    Hu_InventoryInit();
#endif
    LOG_VERBOSE("Initializing statusbar...");
    ST_Init();
    LOG_VERBOSE("Initializing menu...");
    Hu_MenuInit();
    LOG_VERBOSE("Initializing status-message/question system...");
    Hu_MsgInit();
}
SaveSlots &G_SaveSlots() {
    DENG2_ASSERT(sslots != 0);
    return *sslots;
}
/**
 * @attention Game-specific post init actions should be placed in the game-appropriate
 * post init routine (e.g., D_PostInit() for libdoom) and NOT here.
 */
void G_CommonPostInit() {
    R_InitRefresh();
    FI_StackInit();
    GUI_Init();
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    XG_ReadTypes();
#endif
    LOG_VERBOSE("Initializing playsim...");
    P_Init();
    LOG_VERBOSE("Initializing head-up displays...");
    R_InitHud();
    initSaveSlots();
    G_InitEventSequences();
#if __JDOOM__ || __JHERETIC__ || __JHEXEN__
    G_RegisterCheats();
#endif
    {
        auto &cmdLine = CommandLine::get();
        turboMul = float(gfw_GameProfile()->optionValue("turbo").asNumber());
        if (int arg = cmdLine.check("-turbo")) {
            int scale = 200; // Default to 2x without a numeric value.
            if (arg + 1 < cmdLine.count() && !cmdLine.isOption(arg + 1)) {
                scale = cmdLine.at(arg + 1).toInt();
            }
            scale = de::clamp(10, scale, 400);
            turboMul = scale / 100.f;
            LOG_NOTE("Turbo speed: %i%%") << scale;
        }
    }
    DD_Execute(true, "activatebcontext shortcut");
    DD_Execute(true, "listmaps");
}
void G_AutoStartOrBeginTitleLoop() {
    CommandLine &cmdLine = DENG2_APP->commandLine();
    String startEpisodeId;
    de::Uri startMapUri;
    if (int arg = cmdLine.check("-episode", 1)) {
        String episodeId = cmdLine.at(arg + 1);
        if (Record const *episodeDef = Defs().episodes.tryFind("id", episodeId)) {
            de::Uri startMap(episodeDef->gets("startMap"), RC_NULL);
            if (P_MapExists(startMap.compose().toUtf8().constData())) {
                startEpisodeId = episodeId;
            }
        }
    }
    if (int arg = cmdLine.check("-warp", 1)) {
        bool haveEpisode = (arg + 2 < cmdLine.count() && !cmdLine.isOption(arg + 2));
        if (haveEpisode) {
            if (Record const *episodeDef = Defs().episodes.tryFind("id", cmdLine.at(arg + 1))) {
                de::Uri startMap(episodeDef->gets("startMap"), RC_NULL);
                if (P_MapExists(startMap.compose().toUtf8().constData())) {
                    startEpisodeId = episodeDef->gets("id");
                }
            }
        }
        bool isNumber;
        int mapWarpNumber = cmdLine.at(arg + (haveEpisode? 2 : 1)).toInt(&isNumber);
        if (!isNumber) {
            Block rawMapUri = cmdLine.at(arg + (haveEpisode? 2 : 1)).toUtf8();
            char *args[1] = { const_cast<char *>(rawMapUri.constData()) };
            startMapUri = de::Uri::fromUserInput(args, 1);
            if (startMapUri.scheme().isEmpty()) startMapUri.setScheme("Maps");
        }
        else {
            if (startEpisodeId.isEmpty()) {
                startEpisodeId = FirstPlayableEpisodeId();
            }
            startMapUri = TranslateMapWarpNumber(startEpisodeId, mapWarpNumber);
        }
    }
    bool autoStart = (IS_NETGAME || !startEpisodeId.isEmpty() || !startMapUri.isEmpty());
    if (autoStart) {
        if (startEpisodeId.isEmpty()) {
            startEpisodeId = FirstPlayableEpisodeId();
        }
        if (!P_MapExists(startMapUri.compose().toUtf8().constData())) {
            startMapUri.clear();
            if (Record const *episodeDef = Defs().episodes.tryFind("id", startEpisodeId)) {
                de::Uri startMap(episodeDef->gets("startMap"), RC_NULL);
                if (P_MapExists(startMap.compose().toUtf8().constData())) {
                    startMapUri = startMap;
                }
            }
        }
    }
    if (!startEpisodeId.isEmpty() && !startMapUri.isEmpty()) {
        LOG_NOTE("Auto-starting episode '%s', map \"%s\", skill %i")
                << startEpisodeId
                << startMapUri
                << gfw_DefaultRule(skill);
        ::briefDisabled = true;
        G_SetGameActionNewSession(gfw_DefaultGameRules(), startEpisodeId, startMapUri);
    }
    else {
        gfw_Session()->endAndBeginTitle(); // Start up intro loop.
    }
}
/**
 * Common game shutdown routine.
 * @note Game-specific actions should be placed in G_Shutdown rather than here.
 */
void G_CommonShutdown() {
    gfw_Session()->end();
    Plug_RemoveHook(HOOK_DEMO_STOP, Hook_DemoStop);
    Hu_MsgShutdown();
    Hu_UnloadData();
    D_NetClearBuffer();
    P_Shutdown();
    G_ShutdownEventSequences();
    FI_StackShutdown();
    Hu_MenuShutdown();
    ST_Shutdown();
    GUI_Shutdown();
    delete sslots; sslots = 0;
}
gamestate_t G_GameState() {
    return gameState;
}
static char const *getGameStateStr(gamestate_t state) {
    struct statename_s {
        gamestate_t state;
        char const *name;
    } stateNames[] = {
        { GS_MAP,          "GS_MAP" },
        { GS_INTERMISSION, "GS_INTERMISSION" },
        { GS_FINALE,       "GS_FINALE" },
        { GS_STARTUP,      "GS_STARTUP" },
        { GS_WAITING,      "GS_WAITING" },
        { GS_INFINE,       "GS_INFINE" },
        { gamestate_t(-1), 0 }
    };
    for (uint i = 0; stateNames[i].name; ++i) {
        if (stateNames[i].state == state)
            return stateNames[i].name;
    }
    return 0;
}
/**
 * Called when the gameui binding context is active. Triggers the menu.
 */
int G_UIResponder(event_t *ev) {
    if (Hu_MsgResponder(ev))
        return true;
    if (ev->state != EVS_DOWN)
        return false;
    if (!(ev->type == EV_KEY || ev->type == EV_MOUSE_BUTTON || ev->type == EV_JOY_BUTTON))
        return false;
    if (!Hu_MenuIsActive() && !DD_GetInteger(DD_SHIFT_DOWN)) {
        if ((gameAction == GA_NONE && !singledemo && Get(DD_PLAYBACK)) ||
           (G_GameState() == GS_INFINE && FI_IsMenuTrigger())) {
            Hu_MenuCommand(MCMD_OPEN);
            return true;
        }
    }
    return false;
}
void G_ChangeGameState(gamestate_t state) {
    if (G_QuitInProgress()) return;
    if (state < 0 || state >= NUM_GAME_STATES) {
        DENG2_ASSERT(!"G_ChangeGameState: Invalid state");
        return;
    }
    if (gameState != state) {
        gameState = state;
        LOGDEV_NOTE("Game state changed to %s") << getGameStateStr(state);
    }
    bool gameUIActive = false;
    bool gameActive   = true;
    switch (gameState) {
    case GS_FINALE:
    case GS_STARTUP:
    case GS_WAITING:
    case GS_INFINE:
        gameActive = false;
    case GS_INTERMISSION:
        gameUIActive = true;
        break;
    default: break;
    }
    if (!IS_DEDICATED) {
        if (gameUIActive) {
            DD_Execute(true, "activatebcontext gameui");
            B_SetContextFallback("gameui", G_UIResponder);
        }
        DD_Executef(true, "%sactivatebcontext game", gameActive? "" : "de");
    }
}
dd_bool G_StartFinale(char const *script, int flags, finale_mode_t mode, char const *defId) {
    if (!script || !script[0])
        return false;
    for (int i = 0; i < MAXPLAYERS; ++i) {
        ST_LogEmpty(i);               // Clear the message queue for all local players.
        ST_CloseAll(i, true/*fast*/); // Close the HUDs left open for all local players.
    }
    G_SetGameAction(GA_NONE);
    FI_StackExecuteWithId(script, flags, mode, defId);
    return true;
}
void G_StartHelp() {
    if (G_QuitInProgress()) return;
    if (IS_CLIENT) {
        return;
    }
    char const *scriptId = "help";
    if (Record const *finale = Defs().finales.tryFind("id", scriptId)) {
        Hu_MenuCommand(MCMD_CLOSEFAST);
        G_StartFinale(finale->gets("script").toUtf8().constData(), FF_LOCAL, FIMODE_NORMAL, scriptId);
        return;
    }
    LOG_SCR_WARNING("InFine script '%s' not defined") << scriptId;
}
void G_BeginMap() {
    G_ChangeGameState(GS_MAP);
    if (!IS_DEDICATED) {
        R_SetViewPortPlayer(CONSOLEPLAYER, CONSOLEPLAYER); // View the guy you are playing.
        R_ResizeViewWindow(RWF_FORCE|RWF_NO_LERP);
    }
    G_ControlReset();
    mapTime = actualMapTime = 0;
    S_PauseMusic(false);
    LOG_MSG(DE2_ESC(R));
    LOG_NOTE("%s") << G_MapDescription(gfw_Session()->episodeId(),
                                       gfw_Session()->mapUri());
    LOG_MSG(DE2_ESC(R));
}
int G_Responder(event_t *ev) {
    DENG2_ASSERT(ev);
    if (G_QuitInProgress()) return true;
    if (G_GameState() == GS_MAP) {
        Pause_Responder(ev);
        if (!Hu_MenuIsActive() && !Hu_IsMessageActive()) {
            if (ST_Responder(ev))
                return true;
            if (G_EventSequenceResponder(ev))
                return true;
        }
    }
    return Hu_MenuResponder(ev);
}
int G_PrivilegedResponder(event_t *ev) {
    DENG2_ASSERT(ev);
    if (G_QuitInProgress()) return false;
    if (Hu_MenuPrivilegedResponder(ev))
        return true;
    if (ev->type == EV_KEY && ev->data1 == DDKEY_F1) {
        if (CommandLine_Check("-devparm")) {
            if (ev->state == EVS_DOWN) {
                G_SetGameAction(GA_SCREENSHOT);
            }
            return true; // All F1 events are eaten.
        }
    }
    return false; // Not eaten.
}
static sfxenum_t randomQuitSound() {
#if __JDOOM__ || __JDOOM64__
    if (cfg.menuQuitSound) {
# if __JDOOM64__
        static sfxenum_t quitSounds[] = {
            SFX_VILACT,
            SFX_GETPOW,
            SFX_PEPAIN,
            SFX_SLOP,
            SFX_SKESWG,
            SFX_KNTDTH,
            SFX_BSPACT,
            SFX_SGTATK
        };
        static int numQuitSounds = sizeof(quitSounds) / sizeof(quitSounds[0]);
# else
        static sfxenum_t quitSounds[] = {
            SFX_PLDETH,
            SFX_DMPAIN,
            SFX_POPAIN,
            SFX_SLOP,
            SFX_TELEPT,
            SFX_POSIT1,
            SFX_POSIT3,
            SFX_SGTATK
        };
        static int numQuitSounds = sizeof(quitSounds) / sizeof(quitSounds[0]);
        static sfxenum_t quitSounds2[] = {
            SFX_VILACT,
            SFX_GETPOW,
            SFX_BOSCUB,
            SFX_SLOP,
            SFX_SKESWG,
            SFX_KNTDTH,
            SFX_BSPACT,
            SFX_SGTATK
        };
        static int numQuitSounds2 = sizeof(quitSounds2) / sizeof(quitSounds2[0]);
# endif
        sfxenum_t *sndTable = quitSounds;
        int sndTableSize    = numQuitSounds;
# if !__JDOOM64__
        if (gameModeBits & GM_ANY_DOOM2) {
            sndTable     = quitSounds2;
            sndTableSize = numQuitSounds2;
        }
# endif
        if (sndTable && sndTableSize > 0) {
            return sndTable[P_Random() & (sndTableSize - 1)];
        }
    }
#endif
    return SFX_NONE;
}
/**
 * Determines whether an intermission is enabled and will be scheduled when the players
 * leave the  @em current map.
 */
static bool intermissionEnabled() {
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    if (gfw_MapInfoFlags() & MIF_NO_INTERMISSION) {
        return false;
    }
#elif __JHEXEN__
    if (!gfw_Rule(deathmatch)) {
        return false;
    }
#endif
    return true;
}
/**
 * Returns the unique identifier of the music to play during the intermission.
 */
static String intermissionMusic() {
#if __JDOOM64__
    return "dm2int";
#elif __JDOOM__
    return (::gameModeBits & GM_ANY_DOOM2)? "dm2int" : "inter";
#elif __JHERETIC__
    return "intr";
#elif __JHEXEN__
    return "hub";
#endif
}
#if __JDOOM__ || __JDOOM64__
void G_PrepareWIData() {
    wbstartstruct_t *info = &::wmInfo;
    info->maxFrags = 0;
    float parTime = gfw_Session()->mapInfo().getf("parTime");
    info->parTime = (parTime > 0? TICRATE * int(parTime) : -1 /*N/A*/);
    info->pNum = CONSOLEPLAYER;
    for (int i = 0; i < MAXPLAYERS; ++i) {
        player_t const *p        = &players[i];
        wbplayerstruct_t *pStats = &info->plyr[i];
        pStats->inGame = p->plr->inGame;
        pStats->kills  = p->killCount;
        pStats->items  = p->itemCount;
        pStats->secret = p->secretCount;
        pStats->time   = mapTime;
        std::memcpy(pStats->frags, p->frags, sizeof(pStats->frags));
    }
}
#endif
static int prepareIntermission(void * /*context*/)
{
    ::wmInfo.nextMap           = ::nextMapUri;
#if __JHEXEN__
    ::wmInfo.nextMapEntryPoint = ::nextMapEntryPoint;
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    ::wmInfo.currentMap        = gfw_Session()->mapUri();
    ::wmInfo.didSecret         = ::players[CONSOLEPLAYER].didSecret;
# if __JDOOM__ || __JDOOM64__
    ::wmInfo.maxKills          = de::max(1, ::totalKills);
    ::wmInfo.maxItems          = de::max(1, ::totalItems);
    ::wmInfo.maxSecret         = de::max(1, ::totalSecret);
    G_PrepareWIData();
# endif
#endif
    IN_Begin(::wmInfo);
    G_ChangeGameState(GS_INTERMISSION);
    return 0;
}
static void runGameAction() {
#define QUITWAIT_MILLISECONDS 1500
    static uint quitTime = 0;
    static bool unloadTriggered = false;
    if (::quitInProgress) {
        if (Timer_RealMilliseconds() > quitTime + QUITWAIT_MILLISECONDS) {
            if (!unloadTriggered) {
                unloadTriggered = true;
                if (CommandLine_Exists("-game")) {
                    App_Timer(1, Sys_Quit);
                }
                else {
                    DD_Execute(true, "after 1 unload");
                }
            }
        }
        else {
            ::quitDarkenOpacity = de::cubed((Timer_RealMilliseconds() - quitTime) / (float) QUITWAIT_MILLISECONDS);
        }
        return;
    }
    gameaction_t currentAction;
    while ((currentAction = ::gameAction) != GA_NONE) {
        BusyMode_FreezeGameForBusyMode();
        G_SetGameAction(GA_NONE);
        switch (currentAction) {
        case GA_NEWSESSION:
            gfw_Session()->end();
            gfw_Session()->begin(G_NewSessionRules(),
                                 ::gaNewSessionEpisodeId,
                                 ::gaNewSessionMapUri,
                                 ::gaNewSessionMapEntrance);
            break;
        case GA_LOADSESSION:
            gfw_Session()->end();
            try {
                SaveSlot const &sslot = G_SaveSlots()[::gaLoadSessionSlot];
                gfw_Session()->load(sslot.saveName());
                Con_SetInteger2("game-save-last-slot", sslot.id().toInt(), SVF_WRITE_OVERRIDE);
            }
            catch (Error const &er) {
                LOG_RES_WARNING("Error loading from save slot #%s:\n")
                        << ::gaLoadSessionSlot << er.asText();
            }
            if (!gfw_Session()->hasBegun()) {
                gfw_Session()->endAndBeginTitle();
            }
            break;
        case GA_SAVESESSION:
            try {
                SaveSlot const &sslot = G_SaveSlots()[::gaSaveSessionSlot];
                gfw_Session()->save(sslot.saveName(), ::gaSaveSessionUserDescription);
                Con_SetInteger2("game-save-last-slot", sslot.id().toInt(), SVF_WRITE_OVERRIDE);
            }
            catch (Error const &er) {
                LOG_RES_WARNING("Error saving to save slot #%s:\n")
                        << ::gaSaveSessionSlot << er.asText();
            }
            break;
        case GA_QUIT:
            ::quitInProgress = true;
            unloadTriggered  = false;
            quitTime         = Timer_RealMilliseconds();
            Hu_MenuCommand(MCMD_CLOSEFAST);
            if (!IS_NETGAME) {
                S_LocalSound(randomQuitSound(), 0);
                DD_Executef(true, "activatebcontext deui");
            }
            break;
        case GA_LEAVEMAP:
            if (!P_MapExists(::nextMapUri.compose().toUtf8().constData())) {
                ::nextMapUri = de::makeUri(gfw_Session()->episodeDef()->gets("startMap"));
            }
            gfw_Session()->leaveMap(::nextMapUri, ::nextMapEntryPoint);
            break;
        case GA_RESTARTMAP:
            gfw_Session()->reloadMap();
            break;
        case GA_MAPCOMPLETED: {
            dd_bool newHub = true;
#if __JHEXEN__
            if (Record const *episodeDef = gfw_Session()->episodeDef()) {
                defn::Episode epsd(*episodeDef);
                Record const *currentHub = epsd.tryFindHubByMapId(gfw_Session()->mapUri().compose());
                newHub = (!currentHub || currentHub != epsd.tryFindHubByMapId(::nextMapUri.compose()));
            }
#endif
            for (int i = 0; i < MAXPLAYERS; ++i) {
                ST_CloseAll(i, true/*fast*/);         // hide any HUDs left open
                Player_LeaveMap(&players[i], newHub); // take away cards and stuff
            }
#if __JHEXEN__
            SN_StopAllSequences();
#endif
            if (!IS_DEDICATED) {
                G_ResetViewEffects();
            }
            if (intermissionEnabled()) {
                S_StartMusic(intermissionMusic().toUtf8().constData(), true);
                S_PauseMusic(true);
                BusyMode_RunNewTask(BUSYF_TRANSITION, prepareIntermission, nullptr);
#if __JHERETIC__
                NetSv_SendGameState(0, DDSP_ALL_PLAYERS); // @todo necessary at this time?
#endif
                NetSv_Intermission(IMF_BEGIN, 0, 0);
                S_PauseMusic(false);
            }
            else {
                G_IntermissionDone();
            }
            break; }
        case GA_ENDDEBRIEFING:
            ::briefDisabled = true;
            G_IntermissionDone();
            break;
        case GA_SCREENSHOT: {
            String fileName = gfw_GameId() + "-";
            int const numPos = fileName.length();
            for (int i = 0; i < 1e6; ++i) // Stop eventually...
            {
                fileName += String("%1.png").arg(i, 3, 10, QChar('0'));
                if (!M_ScreenShot(fileName.toUtf8(), DD_SCREENSHOT_CHECK_EXISTS)) break; // Check only.
                fileName.truncate(numPos);
            }
            if (M_ScreenShot(fileName.toUtf8(), 0)) {
                String msg = "Saved screenshot: " + NativePath(fileName).pretty();
                P_SetMessageWithFlags(&::players[CONSOLEPLAYER], msg.toLatin1().constData(), LMF_NO_HIDE);
            }
            else {
                LOG_RES_WARNING("Failed taking screenshot \"%s\"")
                        << NativePath(fileName).pretty();
            }
            break; }
        default: break;
        }
    }
#undef QUITWAIT_MILLISECONDS
}
static int rebornLoadConfirmed(msgresponse_t response, int, void *) {
    if (response == MSG_YES) {
        G_SetGameAction(GA_RESTARTMAP);
    }
    else {
        player_t *plr    = players;
        plr->rebornWait  = PLAYER_REBORN_TICS;
        plr->playerState = PST_DEAD;
    }
    return true;
}
/**
 * Do needed reborns for any fallen players.
 */
static void rebornPlayers() {
    if (!gfw_Session()->hasBegun()) return;
    if (G_GameState() != GS_MAP) return;
    if (!IS_NETGAME && P_CountPlayersInGame(LocalOnly) == 1) {
        if (Player_WaitingForReborn(&players[0])) {
            if (Hu_IsMessageActiveWithCallback(rebornLoadConfirmed))
                return;
            if (gfw_Session()->progressRestoredOnReload() && cfg.common.confirmRebornLoad) {
                S_LocalSound(SFX_REBORNLOAD_CONFIRM, NULL);
                AutoStr *msg = Str_Appendf(AutoStr_NewStd(), REBORNLOAD_CONFIRM, gfw_Session()->userDescription().toUtf8().constData());
                Hu_MsgStart(MSG_YESNO, Str_Text(msg), rebornLoadConfirmed, 0, 0);
                return;
            }
            rebornLoadConfirmed(MSG_YES, 0, 0);
        }
        return;
    }
    for (int i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = &players[i];
        if (Player_WaitingForReborn(plr)) {
            P_RebornPlayerInMultiplayer(i);
        }
        if ((int)plr->playerState == PST_GONE) {
            plr->playerState  = PST_REBORN;
            ddplayer_t *ddplr = plr->plr;
            if (mobj_t *plmo = ddplr->mo) {
                if (!IS_CLIENT) {
                    P_SpawnTeleFog(plmo->origin[VX], plmo->origin[VY], plmo->angle + ANG180);
                }
                LOGDEV_MAP_MSG("rebornPlayers: Removing player %i's mobj") << i;
                P_MobjRemove(plmo, true);
                ddplr->mo = 0;
            }
        }
    }
}
/**
 * The core of the timing loop. Game state, game actions etc occur here.
 *
 * @param ticLength  How long this tick is, in seconds.
 */
void G_Ticker(timespan_t ticLength) {
    static gamestate_t oldGameState = gamestate_t(-1);
    Hu_FogEffectTicker(ticLength);
    Hu_MenuTicker(ticLength);
    Hu_MsgTicker();
    if (IS_CLIENT && !Get(DD_GAME_READY)) return;
    runGameAction();
    if (!G_QuitInProgress()) {
        rebornPlayers();
        if (!IS_CLIENT) {
            Set(DD_SERVER_ALLOW_FRAMES, G_GameState() == GS_MAP);
            Set(DD_CLIENT_PAUSED, Pause_IsPaused());
        }
        P_RunPlayers(ticLength);
    }
    else {
        if (!IS_CLIENT) {
            Set(DD_SERVER_ALLOW_FRAMES, false);
        }
    }
    if (G_GameState() == GS_MAP && !IS_DEDICATED) {
        ST_Ticker(ticLength);
    }
    R_ResizeViewWindow(0);
    if (DD_IsSharpTick()) {
        switch (G_GameState()) {
        case GS_MAP:
            if (oldGameState != GS_MAP) {
                Con_SetInteger2("game-state-map", 1, SVF_WRITE_OVERRIDE);
            }
            P_DoTick();
            HU_UpdatePsprites();
            briefDisabled = false;
            if (IS_DEDICATED)
                break;
            Hu_Ticker();
            break;
        case GS_INTERMISSION:
            IN_Ticker();
            break;
        default:
            if (oldGameState != G_GameState()) {
                Con_SetInteger2("game-state-map", 0,         SVF_WRITE_OVERRIDE);
                Con_SetString2 ("map-author",     "Unknown", SVF_WRITE_OVERRIDE);
                Con_SetString2 ("map-name",       "Unknown", SVF_WRITE_OVERRIDE);
                Con_SetInteger2("map-music",      -1,        SVF_WRITE_OVERRIDE);
            }
            break;
        }
        for (int i = 0; i < MAXPLAYERS; ++i) {
            Player_PostTick(&players[i]);
        }
        if (!IS_CLIENT) {
            NetSv_Ticker();
        }
    }
    oldGameState = gameState;
}
/**
 * Safely clears the player data structures.
 */
static void clearPlayer(player_t *p) {
    DENG2_ASSERT(p);
    player_t playerCopy;
    ddplayer_t ddPlayerCopy;
    std::memcpy(&playerCopy, p, sizeof(*p));
    std::memcpy(&ddPlayerCopy, p->plr, sizeof(*p->plr));
    de::zapPtr(p->plr);
    de::zapPtr(p);
    p->plr = playerCopy.plr;
#if __JHERETIC__ || __JHEXEN__ || __JDOOM64__
    P_InventoryEmpty(p - players);
    P_InventorySetReadyItem(p - players, IIT_NONE);
#endif
    p->plr->extraData = p;
    p->plr->inGame = ddPlayerCopy.inGame;
    p->plr->flags  = ddPlayerCopy.flags & ~(DDPF_INTERYAW | DDPF_INTERPITCH);
    p->startSpot = playerCopy.startSpot;
    std::memcpy(&p->plr->fixCounter, &ddPlayerCopy.fixCounter, sizeof(ddPlayerCopy.fixCounter));
    std::memcpy(&p->plr->fixAcked,   &ddPlayerCopy.fixAcked,   sizeof(ddPlayerCopy.fixAcked));
    p->plr->fixCounter.angles++;
    p->plr->fixCounter.origin++;
    p->plr->fixCounter.mom++;
}
/**
 * Called after a player dies (almost everything is cleared and then
 * re-initialized).
 */
void G_PlayerReborn(int player) {
    if (player < 0 || player >= MAXPLAYERS)
        return; // Wha?
    LOGDEV_MAP_NOTE("G_PlayerReborn: reseting player %i") << player;
    player_t *p = &players[player];
    int frags[MAXPLAYERS];
    DENG2_ASSERT(sizeof(p->frags) == sizeof(frags));
    std::memcpy(frags, p->frags, sizeof(frags));
    int killcount    = p->killCount;
    int itemcount    = p->itemCount;
    int secretcount  = p->secretCount;
#if __JHEXEN__
    uint worldTimer  = p->worldTimer;
#endif
#if __JHERETIC__
    dd_bool secret = p->didSecret;
    int spot       = p->startSpot;
#endif
    clearPlayer(p);
#if __JHERETIC__
    p->startSpot = spot;
#endif
    std::memcpy(p->frags, frags, sizeof(p->frags));
    p->killCount   = killcount;
    p->itemCount   = itemcount;
    p->secretCount = secretcount;
#if __JHEXEN__
    p->worldTimer  = worldTimer;
#endif
    p->colorMap    = cfg.playerColor[player];
    p->class_      = P_ClassForPlayerWhenRespawning(player, false);
#if __JHEXEN__
    if (p->class_ == PCLASS_FIGHTER && !IS_NETGAME) {
        p->colorMap = 2;
    }
#endif
    p->useDown      = p->attackDown = true; // Don't do anything immediately.
    p->playerState  = PST_LIVE;
    p->health       = maxHealth;
    p->brain.changeWeapon = WT_NOCHANGE;
#if __JDOOM__ || __JDOOM64__
    p->readyWeapon  = p->pendingWeapon = WT_SECOND;
    p->weapons[WT_FIRST ].owned = true;
    p->weapons[WT_SECOND].owned = true;
    de::zap(p->ammo);
    p->ammo[AT_CLIP].owned = 50;
    P_InitPlayerValues(p);
#elif __JHERETIC__
    p->readyWeapon = p->pendingWeapon = WT_SECOND;
    p->weapons[WT_FIRST].owned = true;
    p->weapons[WT_SECOND].owned = true;
    p->ammo[AT_CRYSTAL].owned = 50;
    de::Uri const mapUri = gfw_Session()->mapUri();
    if (secret ||
       (mapUri.path() == "E1M9" ||
        mapUri.path() == "E2M9" ||
        mapUri.path() == "E3M9" ||
        mapUri.path() == "E4M9" ||
        mapUri.path() == "E5M9")) {
        p->didSecret = true;
    }
#ifdef DENG2_DEBUG
    for (int i = 0; i < NUM_WEAPON_TYPES; ++i) {
        LOGDEV_MAP_MSG("Player %i owns wpn %i: %i") << player << i << p->weapons[i].owned;
    }
#endif
#else
    p->readyWeapon = p->pendingWeapon = WT_FIRST;
    p->weapons[WT_FIRST].owned = true;
#endif
#if defined(HAVE_EARTHQUAKE)
    localQuakeHappening[player] = false;
    localQuakeTimeout[player] = 0;
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    for (int i = 0; i < NUM_AMMO_TYPES; ++i) {
        p->ammo[i].max = maxAmmo[i];
    }
#endif
    p->viewHeight      = cfg.common.plrViewHeight;
    p->viewHeightDelta = 0;
#if __JHERETIC__
    p->update |= PSF_VIEW_HEIGHT |
        PSF_STATE | PSF_HEALTH | PSF_ARMOR_TYPE | PSF_ARMOR_POINTS |
        PSF_INVENTORY | PSF_POWERS | PSF_KEYS | PSF_OWNED_WEAPONS | PSF_AMMO |
        PSF_MAX_AMMO | PSF_PENDING_WEAPON | PSF_READY_WEAPON | PSF_MORPH_TIME;
#else
    p->update |= PSF_REBORN;
#endif
    p->plr->flags &= ~DDPF_DEAD;
}
#if __JDOOM__ || __JDOOM64__
void G_QueueBody(mobj_t *mo) {
    if (!mo) return;
    if (bodyQueueSlot >= BODYQUEUESIZE) {
        P_MobjRemove(bodyQueue[bodyQueueSlot % BODYQUEUESIZE], false);
    }
    bodyQueue[bodyQueueSlot % BODYQUEUESIZE] = mo;
    bodyQueueSlot++;
}
#endif
/**
 * Lookup the debriefing Finale for the current episode and map (if any).
 */
static Record const *finaleDebriefing() {
    if (::briefDisabled) return 0;
#if __JHEXEN__
    if (::cfg.overrideHubMsg && G_GameState() == GS_MAP) {
        defn::Episode epsd(*gfw_Session()->episodeDef());
        Record const *currentHub = epsd.tryFindHubByMapId(gfw_Session()->mapUri().compose());
        if (!currentHub || currentHub != epsd.tryFindHubByMapId(::nextMapUri.compose())) {
            return 0;
        }
    }
#endif
    if (IS_CLIENT || Get(DD_PLAYBACK)) return 0;
    if (G_GameState() == GS_INFINE) return 0;
    return Defs().finales.tryFind("after", gfw_Session()->mapUri().compose());
}
void G_IntermissionDone() {
    if (Record const *finale = finaleDebriefing()) {
        if (G_StartFinale(finale->gets("script").toUtf8().constData(), 0, FIMODE_AFTER, 0)) {
            return;
        }
    }
    ::briefDisabled = false;
    FI_StackClear();
    if (::nextMapUri.isEmpty()) {
        G_SetGameAction(GA_VICTORY);
        return;
    }
    G_SetGameAction(GA_LEAVEMAP);
}
String G_DefaultGameStateFolderUserDescription(String const &saveName, bool autogenerate) {
    if (!saveName.isEmpty()) {
        String const existing = gfw_Session()->savedUserDescription(saveName);
        if (!existing.isEmpty()) return existing;
    }
    if (!autogenerate) return "";
    String description;
    de::Uri const mapUri = gfw_Session()->mapUri();
    String mapUriAsText  = mapUri.compose();
    if (P_MapIsCustom(mapUriAsText.toUtf8().constData())) {
        String const mapSourcePath(Str_Text(P_MapSourceFile(mapUriAsText.toUtf8().constData())));
        description += mapSourcePath.fileNameWithoutExtension() + ":";
    }
    String mapTitle = G_MapTitle(mapUri);
    if (mapTitle.isEmpty() || mapTitle.at(0) == ' ') {
        mapTitle = mapUri.path();
    }
    description += mapTitle;
    int time = mapTime / TICRATE;
    int const hours   = time / 3600; time -= hours * 3600;
    int const minutes = time / 60;   time -= minutes * 60;
    int const seconds = time;
    description += String(" %1:%2:%3")
                       .arg(hours,   2, 10, QChar('0'))
                       .arg(minutes, 2, 10, QChar('0'))
                       .arg(seconds, 2, 10, QChar('0'));
    return description;
}
String G_EpisodeTitle(String episodeId) {
    String title;
    if (Record const *episodeDef = Defs().episodes.tryFind("id", episodeId)) {
        title = episodeDef->gets("title");
        int textIdx = Defs().getTextNum(title.toUtf8().constData());
        if (textIdx >= 0) {
            title = Defs().text[textIdx].text; // Yes, use the resolved text string.
        }
    }
    return title;
}
uint G_MapNumberFor(de::Uri const &mapUri) {
    String path = mapUri.path();
    if (!path.isEmpty()) {
#if __JDOOM__ || __JHERETIC__
# if __JDOOM__
        if (gameModeBits & (GM_ANY_DOOM | ~GM_DOOM_CHEX))
# endif
        {
            if (path.at(0).toLower() == 'e' && path.at(2).toLower() == 'm') {
                return path.substr(3).toInt() - 1;
            }
        }
#endif
        if (path.beginsWith("map", String::CaseInsensitive)) {
            return path.substr(3).toInt() - 1;
        }
    }
    return 0;
}
AutoStr *G_CurrentMapUriPath() {
    return AutoStr_FromTextStd(gfw_Session()->mapUri().path().toStringRef().toUtf8().constData());
}
de::Uri G_ComposeMapUri(uint episode, uint map) {
    String mapId;
#if __JDOOM64__
    mapId = String("map%1").arg(map+1, 2, 10, QChar('0'));
    DENG2_UNUSED(episode);
#elif __JDOOM__
    if (gameModeBits & GM_ANY_DOOM2)
        mapId = String("map%1").arg(map+1, 2, 10, QChar('0'));
    else
        mapId = String("e%1m%2").arg(episode+1).arg(map+1);
#elif  __JHERETIC__
    mapId = String("e%1m%2").arg(episode+1).arg(map+1);
#else
    mapId = String("map%1").arg(map+1, 2, 10, QChar('0'));
    DENG2_UNUSED(episode);
#endif
    return de::Uri("Maps", mapId);
}
Record &G_MapInfoForMapUri(de::Uri const &mapUri) {
    if (Record *def = Defs().mapInfos.tryFind("id", mapUri.compose())) {
        return *def;
    }
    if (Record *def = Defs().mapInfos.tryFind("id", de::Uri("Maps", Path("*")).compose())) {
        return *def;
    } {
        static Record fallbackDef;
        static bool needInitFallbackDef = true;
        if (needInitFallbackDef) {
            needInitFallbackDef = false;
            defn::MapInfo(fallbackDef).resetToDefaults();
        }
        return fallbackDef;
    }
}
String G_MapTitle(de::Uri const &mapUri) {
    String title = G_MapInfoForMapUri(mapUri).gets("title");
    int textIdx = Defs().getTextNum(title.toUtf8().constData());
    if (textIdx >= 0) {
        title = Defs().text[textIdx].text; // Yes, use the resolved text string.
    }
    int idSuffixAt = title.indexOf(':');
    if (idSuffixAt >= 0) {
        int subStart = idSuffixAt + 1;
        while (subStart < title.length() && title.at(subStart).isSpace()) { subStart++; }
        return title.substr(subStart);
    }
    return title;
}
String G_MapAuthor(de::Uri const &mapUri, bool supressGameAuthor) {
    String author = G_MapInfoForMapUri(mapUri).gets("author");
    if (!author.isEmpty()) {
        GameInfo gameInfo;
        DD_GameInfo(&gameInfo);
        if (supressGameAuthor || P_MapIsCustom(mapUri.compose().toUtf8().constData())) {
            if (!author.compareWithoutCase(Str_Text(gameInfo.author)))
                return "";
        }
    }
    if (cfg.common.hideUnknownAuthor && !author.compareWithoutCase("unknown")) {
        return "";
    }
    return author;
}
de::Uri G_MapTitleImage(de::Uri const &mapUri) {
    return de::makeUri(G_MapInfoForMapUri(mapUri).gets("titleImage"));
}
String G_MapDescription(String episodeId, de::Uri const &mapUri) {
    Block mapUriUtf8 = mapUri.compose().toUtf8();
    if (!P_MapExists(mapUriUtf8.constData())) {
        return String("Unknown map (Episode: ") + episodeId + ", Uri: " + mapUri + ")";
    }
    String desc;
    QTextStream os(&desc);
    String const title = G_MapTitle(mapUri);
    if (!title.isEmpty()) {
        os << "Map: " DE2_ESC(i) DE2_ESC(b) << title << DE2_ESC(.)
           << " (Uri: " << mapUri;
        if (Record const *rec = Defs().episodes.tryFind("id", episodeId)) {
            if (Record const *mgNodeDef = defn::Episode(*rec).tryFindMapGraphNode(mapUri.compose())) {
                os << ", warp: " << String::number(mgNodeDef->geti("warpNumber"));
            }
        }
        os << ")" << DE2_ESC(.);
    }
    String const author = G_MapAuthor(mapUri, P_MapIsCustom(mapUriUtf8.constData()));
    if (!author.isEmpty()) {
        os << "\n - Author: " DE2_ESC(i) << author;
    }
    return desc;
}
/**
 * Stops both playback and a recording. Called at critical points like
 * starting a new game, or ending the game in the menu.
 */
void G_StopDemo() {
    if (!IS_SERVER) {
        DD_Execute(true, "stopdemo");
    }
}
int Hook_DemoStop(int /*hookType*/, int val, void * /*context*/)
{
    bool aborted = val != 0;
    G_ChangeGameState(GS_WAITING);
    if (!aborted && singledemo) {
        G_SetGameAction(GA_QUIT);
        return true;
    }
    G_SetGameAction(GA_NONE);
    if (IS_NETGAME && IS_CLIENT) {
        GameRules newRules(gfw_Session()->rules());
        GameRules_Set(newRules, deathmatch, 0);
        GameRules_Set(newRules, noMonsters, false);
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        GameRules_Set(newRules, respawnMonsters, false);
#endif
#if __JHEXEN__
        GameRules_Set(newRules, randomClasses, false);
#endif
        gfw_Session()->applyNewRules(newRules);
    }
    for (int i = 0; i < MAXPLAYERS; ++i) {
        ST_CloseAll(i, true/*fast*/);
    }
    return true;
}
static int quitGameConfirmed(msgresponse_t response, int /*userValue*/, void * /*userPointer*/)
{
    if (response == MSG_YES) {
        G_SetGameAction(GA_QUIT);
    }
    return true;
}
void G_QuitGame() {
    if (G_QuitInProgress()) return;
    if (Hu_IsMessageActiveWithCallback(quitGameConfirmed)) {
        DD_Execute(true, "quit!");
        return;
    }
    char const *endString;
#if __JDOOM__ || __JDOOM64__
    endString = endmsg[((int) GAMETIC % (NUM_QUITMESSAGES + 1))];
#else
    endString = GET_TXT(TXT_QUITMSG);
#endif
    Con_Open(false);
    Hu_MsgStart(MSG_YESNO, endString, quitGameConfirmed, 0, nullptr);
}
D_CMD(OpenLoadMenu) {
    DENG2_UNUSED3(src, argc, argv);
    if (!gfw_Session()->isLoadingPossible()) return false;
    DD_Execute(true, "menu loadgame");
    return true;
}
D_CMD(OpenSaveMenu) {
    DENG2_UNUSED3(src, argc, argv);
    if (!gfw_Session()->isSavingPossible()) return false;
    DD_Execute(true, "menu savegame");
    return true;
}
static int endSessionConfirmed(msgresponse_t response, int /*userValue*/, void * /*context*/)
{
    if (response == MSG_YES) {
        DD_Execute(true, "endgame confirm");
    }
    return true;
}
D_CMD(EndSession) {
    DENG2_UNUSED3(src, argc, argv);
    if (G_QuitInProgress()) return true;
    if (IS_NETGAME && IS_SERVER) {
        LOG_NET_ERROR("Cannot end a networked game session. Stop the server instead");
        return false;
    }
    if (!gfw_Session()->hasBegun()) {
        if (IS_NETGAME && IS_CLIENT) {
            LOG_NET_ERROR("%s") << ENDNOGAME;
        }
        else {
            Hu_MsgStart(MSG_ANYKEY, ENDNOGAME, nullptr, 0, nullptr);
        }
        return true;
    }
    bool const confirmed = (argc >= 2 && !qstricmp(argv[argc-1], "confirm"));
    if (confirmed || (IS_NETGAME && IS_SERVER)) {
        if (IS_NETGAME && IS_CLIENT) {
            DD_Executef(false, "net disconnect");
        }
        else {
            gfw_Session()->endAndBeginTitle();
        }
    }
    else {
        Hu_MsgStart(MSG_YESNO, IS_CLIENT? GET_TXT(TXT_DISCONNECT) : ENDGAME, endSessionConfirmed, 0, nullptr);
    }
    return true;
}
static int loadSessionConfirmed(msgresponse_t response, int /*userValue*/, void *context)
{
    String *slotId = static_cast<String *>(context);
    DENG2_ASSERT(slotId != 0);
    if (response == MSG_YES) {
        DD_Executef(true, "loadgame %s confirm", slotId->toUtf8().constData());
    }
    delete slotId;
    return true;
}
D_CMD(LoadSession) {
    DENG2_UNUSED(src);
    bool const confirmed = (argc == 3 && !qstricmp(argv[2], "confirm"));
    if (G_QuitInProgress()) return false;
    if (!gfw_Session()->isLoadingPossible()) return false;
    if (IS_NETGAME) {
        S_LocalSound(SFX_QUICKLOAD_PROMPT, nullptr);
        Hu_MsgStart(MSG_ANYKEY, QLOADNET, nullptr, 0, nullptr);
        return false;
    }
    if (SaveSlot *sslot = G_SaveSlots().slotByUserInput(argv[1])) {
        if (sslot->isLoadable()) {
            if (confirmed || !cfg.common.confirmQuickGameSave) {
                S_LocalSound(SFX_MENU_ACCEPT, nullptr);
                return G_SetGameActionLoadSession(sslot->id());
            }
            if (Hu_IsMessageActive()) return false;
            S_LocalSound(SFX_QUICKLOAD_PROMPT, nullptr);
            String const &existingDescription = gfw_Session()->savedUserDescription(sslot->saveName());
            AutoStr *msg = Str_Appendf(AutoStr_NewStd(), QLPROMPT,
                                       sslot->id().toUtf8().constData(),
                                       existingDescription.toUtf8().constData());
            Hu_MsgStart(MSG_YESNO, Str_Text(msg), loadSessionConfirmed, 0, new String(sslot->id()));
            return true;
        }
    }
    if (!qstricmp(argv[1], "quick") || !qstricmp(argv[1], "<quick>")) {
        S_LocalSound(SFX_QUICKLOAD_PROMPT, nullptr);
        Hu_MsgStart(MSG_ANYKEY, QSAVESPOT, nullptr, 0, nullptr);
        return true;
    }
    if (!G_SaveSlots().has(argv[1])) {
        LOG_SCR_WARNING("Failed to determine save slot from \"%s\"") << argv[1];
    }
    if (src == CMDS_CONSOLE) {
        LOG_SCR_MSG("Opening Load Game menu...");
        DD_Execute(true, "menu loadgame");
        return true;
    }
    return false;
}
D_CMD(QuickLoadSession) {
    DENG2_UNUSED3(src, argc, argv);
    return DD_Execute(true, "loadgame quick");
}
struct savesessionconfirmed_params_t {
    String slotId;
    String userDescription;
};
static int saveSessionConfirmed(msgresponse_t response, int /*userValue*/, void *context)
{
    savesessionconfirmed_params_t *p = static_cast<savesessionconfirmed_params_t *>(context);
    DENG2_ASSERT(p != 0);
    if (response == MSG_YES) {
        DD_Executef(true, "savegame %s \"%s\" confirm", p->slotId.toUtf8().constData(), p->userDescription.toUtf8().constData());
    }
    delete p;
    return true;
}
D_CMD(SaveSession) {
    DENG2_UNUSED(src);
    bool const confirmed = (argc >= 3 && !qstricmp(argv[argc-1], "confirm"));
    if (G_QuitInProgress()) return false;
    if (IS_CLIENT || IS_NETWORK_SERVER) {
        LOG_ERROR("Network savegames are not supported at the moment");
        return false;
    }
    player_t *player = &players[CONSOLEPLAYER];
    if (player->playerState == PST_DEAD || Get(DD_PLAYBACK)) {
        S_LocalSound(SFX_QUICKSAVE_PROMPT, nullptr);
        Hu_MsgStart(MSG_ANYKEY, SAVEDEAD, nullptr, 0, nullptr);
        return true;
    }
    if (G_GameState() != GS_MAP) {
        S_LocalSound(SFX_QUICKSAVE_PROMPT, nullptr);
        Hu_MsgStart(MSG_ANYKEY, SAVEOUTMAP, nullptr, 0, nullptr);
        return true;
    }
    if (SaveSlot *sslot = G_SaveSlots().slotByUserInput(argv[1])) {
        if (sslot->isUserWritable()) {
            String userDescription;
            if (argc >= 3 && qstricmp(argv[2], "confirm")) {
                userDescription = argv[2];
            }
            if (sslot->isUnused() || confirmed || !cfg.common.confirmQuickGameSave) {
                S_LocalSound(SFX_MENU_ACCEPT, nullptr);
                return G_SetGameActionSaveSession(sslot->id(), &userDescription);
            }
            if (Hu_IsMessageActive()) return false;
            S_LocalSound(SFX_QUICKSAVE_PROMPT, nullptr);
            String const existingDescription = gfw_Session()->savedUserDescription(sslot->saveName());
            AutoStr *msg = Str_Appendf(AutoStr_NewStd(), QSPROMPT,
                                       sslot->id().toUtf8().constData(),
                                       existingDescription.toUtf8().constData());
            savesessionconfirmed_params_t *parm = new savesessionconfirmed_params_t;
            parm->slotId          = sslot->id();
            parm->userDescription = userDescription;
            Hu_MsgStart(MSG_YESNO, Str_Text(msg), saveSessionConfirmed, 0, parm);
            return true;
        }
        LOG_SCR_ERROR("Save slot '%s' is non-user-writable") << sslot->id();
    }
    if (!qstricmp(argv[1], "quick") || !qstricmp(argv[1], "<quick>")) {
        Hu_MenuCommand(MCMD_OPEN);
        Hu_MenuSetPage("SaveGame");
        menuNominatingQuickSaveSlot = true;
        return true;
    }
    if (!G_SaveSlots().has(argv[1])) {
        LOG_SCR_WARNING("Failed to determine save slot from \"%s\"") << argv[1];
    }
    return false;
}
D_CMD(QuickSaveSession) {
    DENG2_UNUSED3(src, argc, argv);
    return DD_Execute(true, "savegame quick");
}
static int deleteGameStateFolderConfirmed(msgresponse_t response, int /*userValue*/, void *context)
{
    String const *saveName = static_cast<de::String const *>(context);
    DENG2_ASSERT(saveName != 0);
    if (response == MSG_YES) {
        DD_Executef(true, "deletegamesave %s confirm", saveName->toUtf8().constData());
    }
    delete saveName;
    return true;
}
D_CMD(DeleteSaveGame) {
    DENG2_UNUSED(src);
    if (G_QuitInProgress()) return false;
    bool const confirmed = (argc >= 3 && !qstricmp(argv[argc-1], "confirm"));
    if (SaveSlot *sslot = G_SaveSlots().slotByUserInput(argv[1])) {
        if (sslot->isUserWritable()) {
            if (sslot->isUnused()) return false;
            if (confirmed) {
                gfw_Session()->removeSaved(sslot->saveName());
            }
            else {
                if (Hu_IsMessageActive()) return false;
                S_LocalSound(SFX_DELETESAVEGAME_CONFIRM, nullptr);
                String const existingDescription = gfw_Session()->savedUserDescription(sslot->saveName());
                AutoStr *msg = Str_Appendf(AutoStr_NewStd(), DELETESAVEGAME_CONFIRM, existingDescription.toUtf8().constData());
                Hu_MsgStart(MSG_YESNO, Str_Text(msg), deleteGameStateFolderConfirmed, 0, new String(sslot->saveName()));
            }
            return true;
        }
        LOG_SCR_ERROR("Save slot '%s' is non-user-writable") << sslot->id();
    }
    else {
        LOG_SCR_WARNING("Failed to determine save slot from '%s'") << argv[1];
    }
    return false;
}
D_CMD(HelpScreen) {
    DENG2_UNUSED3(src, argc, argv);
    G_StartHelp();
    return true;
}
D_CMD(CycleTextureGamma) {
    DENG2_UNUSED3(src, argc, argv);
    R_CycleGammaLevel();
    return true;
}
D_CMD(LeaveMap) {
    DENG2_UNUSED(src);
    String exitName(argc > 1? argv[1] : "next");
    if (IS_NETGAME && !IS_NETWORK_SERVER)
        return false;
    if (G_GameState() != GS_MAP) {
#if __JHERETIC__ || __JHEXEN__
        S_LocalSound(SFX_CHAT, nullptr);
#else
        S_LocalSound(SFX_OOF, nullptr);
#endif
        LOG_MAP_ERROR("Can only exit a map when in a game!");
        return false;
    }
    G_SetGameActionMapCompleted(gfw_Session()->mapUriForNamedExit(exitName));
    return true;
}
D_CMD(SetDefaultSkill) {
    DENG_UNUSED(src);
    if (argc != 2) {
        LOG_SCR_NOTE("Usage: %s (skill)") << argv[0];
        return true;
    }
    gfw_SetDefaultRule(skill, String(argv[1]).toInt() - 1);
    if (gfw_DefaultRule(skill) < SM_BABY || gfw_DefaultRule(skill) >= NUM_SKILL_MODES) {
        gfw_SetDefaultRule(skill, SM_MEDIUM);
    }
    char const *skillNames[] = {
        "Novice",
        "Easy",
        "Normal",
        "Hard",
        "Nightmare!"
    };
    LOG_SCR_MSG("Default skill level for new games: %s") << skillNames[gfw_DefaultRule(skill)];
    return true;
}
/**
 * Warp behavior is as follows:
 *
 * if a game session is in progress and episode id matches current
 *     continue the session and change map
 *     if Hexen and the targt map is in another hub
 *         force a new session.
 * else
 *     begin a new game session and warp to the specified map.
 *
 * @note In a networked game we must presently force a new game session when a
 * map change outside the normal progression occurs to allow session-level state
 * changes to take effect. In single player this behavior is not necessary.
 *
 * @note "setmap" is an alias of "warp"
 *
 * @todo Clean up the map/episode selection logic... -jk
 */
D_CMD(WarpMap) {
    if (IS_NETGAME && !IS_NETWORK_SERVER) {
        return false;
    }
    if (argc == 1) {
        LOG_SCR_NOTE("Usage: %s (episode) (map)") << argv[0];
        return true;
    }
    String episodeId = gfw_Session()->episodeId();
    if (episodeId.isEmpty() && PlayableEpisodeCount() == 1) {
        episodeId = FirstPlayableEpisodeId();
    }
    bool const haveEpisode = (argc >= 3);
    if (haveEpisode) {
        episodeId = argv[1];
        if (Record const *episodeDef = Defs().episodes.tryFind("id", episodeId)) {
            de::Uri startMap(episodeDef->gets("startMap"), RC_NULL);
            if (!P_MapExists(startMap.compose().toUtf8().constData())) {
                LOG_SCR_NOTE("Failed to locate the start map for episode '%s'."
                             " This episode is not playable") << episodeId;
                return false;
            }
        }
        else {
            LOG_SCR_NOTE("Unknown episode '%s'") << episodeId;
            return false;
        }
    }
    de::Uri mapUri;
    bool isNumber;
    int mapWarpNumber = String(argv[haveEpisode? 2 : 1]).toInt(&isNumber);
    if (!isNumber) {
        if (!haveEpisode) {
            if (String implicitEpisodeId = Defs().findEpisode(argv[1])) {
                episodeId = implicitEpisodeId;
            }
        }
        Block rawMapUri = String(argv[haveEpisode? 2 : 1]).toUtf8();
        char *args[1] = { const_cast<char *>(rawMapUri.constData()) };
        mapUri = de::Uri::fromUserInput(args, 1);
        if (mapUri.scheme().isEmpty()) mapUri.setScheme("Maps");
    }
    else {
        mapUri = TranslateMapWarpNumber(episodeId, mapWarpNumber);
        if (mapUri.isEmpty()) {
            bool isNumber;
            int episodeNum = episodeId.toInt(&isNumber);
            if (isNumber) {
                mapUri = G_ComposeMapUri(episodeNum, mapWarpNumber > 0? mapWarpNumber - 1 : 0);
            }
        }
    }
    if (!P_MapExists(mapUri.compose().toUtf8().constData())) {
        String msg("Unknown map");
        if (argc >= 3) msg += String(" \"%1 %2\"").arg(argv[1]).arg(argv[2]);
        else          msg += String(" \"%1\"").arg(argv[1]);
        P_SetMessageWithFlags(&players[CONSOLEPLAYER], msg.toUtf8().constData(), LMF_NO_HIDE);
        return false;
    }
    bool forceNewSession = (IS_NETGAME != 0);
    if (gfw_Session()->hasBegun()) {
        if (gfw_Session()->episodeId().compareWithoutCase(episodeId)) {
            forceNewSession = true;
        }
    }
#if __JHEXEN__
    if (!forceNewSession && gfw_Session()->mapUri() == mapUri) {
        P_SetMessageWithFlags(&players[CONSOLEPLAYER], "Cannot warp to the current map.", LMF_NO_HIDE);
        return false;
    }
    for (int i = 0; i < MAXPLAYERS; ++i) {
        if (players[i].plr->inGame && players[i].playerState == PST_DEAD) {
            players[i].health = maxHealth; /// @todo: Game sessions vs. hubs needs to be debugged.
        }
    }
#endif
    for (int i = 0; i < MAXPLAYERS; ++i) {
        ST_CloseAll(i, true/*fast*/);
    }
    Hu_MenuCommand(MCMD_CLOSEFAST);
    ::briefDisabled = true;
    if (!forceNewSession && gfw_Session()->hasBegun()) {
#if __JHEXEN__
        ::nextMapUri        = mapUri;
        ::nextMapEntryPoint = 0;
        G_SetGameAction(GA_LEAVEMAP);
#else
        G_SetGameActionNewSession(gfw_Session()->rules(), gfw_Session()->episodeId(), mapUri);
#endif
    }
    else {
        GameRules rules = (gfw_Session()->hasBegun() ? gfw_Session()->rules() : gfw_DefaultGameRules());
        if (IS_DEDICATED) {
            GameRules_Set(rules, skill, cfg.common.netSkill);
        }
        G_SetGameActionNewSession(rules, episodeId, mapUri);
    }
    if (src == CMDS_GAME && !(IS_NETGAME && IS_SERVER)) {
#if __JHEXEN__
        char const *msg = TXT_CHEATWARP;
        int soundId     = SFX_PLATFORM_STOP;
#elif __JHERETIC__
        char const *msg = TXT_CHEATWARP;
        int soundId     = SFX_DORCLS;
#else //__JDOOM__ || __JDOOM64__
        char const *msg = STSTR_CLEV;
        int soundId     = SFX_NONE;
#endif
        P_SetMessageWithFlags(&players[CONSOLEPLAYER], msg, LMF_NO_HIDE);
        S_LocalSound(soundId, nullptr);
    }
    return true;
}
namespace {
char *gsvMapAuthor;// = "Unknown";
int gsvMapMusic = -1;
char *gsvMapTitle;// = "Unknown";
int gsvInMap;
int gsvCurrentMusic;
int gsvArmor;
int gsvHealth;
#if !__JHEXEN__
int gsvKills;
int gsvItems;
int gsvSecrets;
#endif
int gsvCurrentWeapon;
int gsvWeapons[NUM_WEAPON_TYPES];
int gsvKeys[NUM_KEY_TYPES];
int gsvAmmo[NUM_AMMO_TYPES];
#if __JHERETIC__ || __JHEXEN__ || __JDOOM64__
int gsvInvItems[NUM_INVENTORYITEM_TYPES];
#endif
#if __JHEXEN__
int gsvWPieces[WEAPON_FOURTH_PIECE_COUNT + 1];
#endif
}
#define READONLYCVAR        CVF_READ_ONLY|CVF_NO_MAX|CVF_NO_MIN|CVF_NO_ARCHIVE
static void registerGameStatusCVars() {
    cvartemplate_t cvars[] = {
        {"game-music", READONLYCVAR, CVT_INT, &gsvCurrentMusic, 0, 0, 0},
        {"game-state", READONLYCVAR, CVT_INT, &gameState, 0, 0, 0},
        {"game-state-map", READONLYCVAR, CVT_INT, &gsvInMap, 0, 0, 0},
#if !__JHEXEN__
        {"game-stats-kills", READONLYCVAR, CVT_INT, &gsvKills, 0, 0, 0},
        {"game-stats-items", READONLYCVAR, CVT_INT, &gsvItems, 0, 0, 0},
        {"game-stats-secrets", READONLYCVAR, CVT_INT, &gsvSecrets, 0, 0, 0},
#endif
        {"map-author", READONLYCVAR, CVT_CHARPTR, &gsvMapAuthor, 0, 0, 0},
        {"map-music", READONLYCVAR, CVT_INT, &gsvMapMusic, 0, 0, 0},
        {"map-name", READONLYCVAR, CVT_CHARPTR, &gsvMapTitle, 0, 0, 0},
        {"player-health", READONLYCVAR, CVT_INT, &gsvHealth, 0, 0, 0},
        {"player-armor", READONLYCVAR, CVT_INT, &gsvArmor, 0, 0, 0},
        {"player-weapon-current", READONLYCVAR, CVT_INT, &gsvCurrentWeapon, 0, 0, 0},
#if __JDOOM__ || __JDOOM64__
        {"player-ammo-bullets", READONLYCVAR, CVT_INT, &gsvAmmo[AT_CLIP], 0, 0, 0},
        {"player-ammo-shells", READONLYCVAR, CVT_INT, &gsvAmmo[AT_SHELL], 0, 0, 0},
        {"player-ammo-cells", READONLYCVAR, CVT_INT, &gsvAmmo[AT_CELL], 0, 0, 0},
        {"player-ammo-missiles", READONLYCVAR, CVT_INT, &gsvAmmo[AT_MISSILE], 0, 0, 0},
        {"player-weapon-fist", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FIRST], 0, 0, 0},
        {"player-weapon-pistol", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SECOND], 0, 0, 0},
        {"player-weapon-shotgun", READONLYCVAR, CVT_INT, &gsvWeapons[WT_THIRD], 0, 0, 0},
        {"player-weapon-chaingun", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FOURTH], 0, 0, 0},
        {"player-weapon-mlauncher", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FIFTH], 0, 0, 0},
        {"player-weapon-plasmarifle", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SIXTH], 0, 0, 0},
        {"player-weapon-bfg", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SEVENTH], 0, 0, 0},
        {"player-weapon-chainsaw", READONLYCVAR, CVT_INT, &gsvWeapons[WT_EIGHTH], 0, 0, 0},
        {"player-weapon-sshotgun", READONLYCVAR, CVT_INT, &gsvWeapons[WT_NINETH], 0, 0, 0},
    #if __JDOOM64__
        {"player-weapon-unmaker", READONLYCVAR, CVT_INT, &gsvWeapons[WT_TENTH], 0, 0, 0},
    #endif
        {"player-key-blue", READONLYCVAR, CVT_INT, &gsvKeys[KT_BLUECARD], 0, 0, 0},
        {"player-key-yellow", READONLYCVAR, CVT_INT, &gsvKeys[KT_YELLOWCARD], 0, 0, 0},
        {"player-key-red", READONLYCVAR, CVT_INT, &gsvKeys[KT_REDCARD], 0, 0, 0},
        {"player-key-blueskull", READONLYCVAR, CVT_INT, &gsvKeys[KT_BLUESKULL], 0, 0, 0},
        {"player-key-yellowskull", READONLYCVAR, CVT_INT, &gsvKeys[KT_YELLOWSKULL], 0, 0, 0},
        {"player-key-redskull", READONLYCVAR, CVT_INT, &gsvKeys[KT_REDSKULL], 0, 0, 0},
#  if __JDOOM64__
        {"player-artifact-bluedemonkey", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_DEMONKEY1], 0, 0, 0},
        {"player-artifact-yellowdemonkey", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_DEMONKEY2], 0, 0, 0},
        {"player-artifact-reddemonkey", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_DEMONKEY3], 0, 0, 0},
#  endif
#elif __JHERETIC__
        {"player-ammo-goldwand", READONLYCVAR, CVT_INT, &gsvAmmo[AT_CRYSTAL], 0, 0, 0},
        {"player-ammo-crossbow", READONLYCVAR, CVT_INT, &gsvAmmo[AT_ARROW], 0, 0, 0},
        {"player-ammo-dragonclaw", READONLYCVAR, CVT_INT, &gsvAmmo[AT_ORB], 0, 0, 0},
        {"player-ammo-hellstaff", READONLYCVAR, CVT_INT, &gsvAmmo[AT_RUNE], 0, 0, 0},
        {"player-ammo-phoenixrod", READONLYCVAR, CVT_INT, &gsvAmmo[AT_FIREORB], 0, 0, 0},
        {"player-ammo-mace", READONLYCVAR, CVT_INT, &gsvAmmo[AT_MSPHERE], 0, 0, 0},
        {"player-weapon-staff", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FIRST], 0, 0, 0},
        {"player-weapon-goldwand", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SECOND], 0, 0, 0},
        {"player-weapon-crossbow", READONLYCVAR, CVT_INT, &gsvWeapons[WT_THIRD], 0, 0, 0},
        {"player-weapon-dragonclaw", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FOURTH], 0, 0, 0},
        {"player-weapon-hellstaff", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FIFTH], 0, 0, 0},
        {"player-weapon-phoenixrod", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SIXTH], 0, 0, 0},
        {"player-weapon-mace", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SEVENTH], 0, 0, 0},
        {"player-weapon-gauntlets", READONLYCVAR, CVT_INT, &gsvWeapons[WT_EIGHTH], 0, 0, 0},
        {"player-key-yellow", READONLYCVAR, CVT_INT, &gsvKeys[KT_YELLOW], 0, 0, 0},
        {"player-key-green", READONLYCVAR, CVT_INT, &gsvKeys[KT_GREEN], 0, 0, 0},
        {"player-key-blue", READONLYCVAR, CVT_INT, &gsvKeys[KT_BLUE], 0, 0, 0},
        {"player-artifact-ring", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_INVULNERABILITY], 0, 0, 0},
        {"player-artifact-shadowsphere", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_INVISIBILITY], 0, 0, 0},
        {"player-artifact-crystalvial", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_HEALTH], 0, 0, 0},
        {"player-artifact-mysticurn", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_SUPERHEALTH], 0, 0, 0},
        {"player-artifact-tomeofpower", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TOMBOFPOWER], 0, 0, 0},
        {"player-artifact-torch", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TORCH], 0, 0, 0},
        {"player-artifact-firebomb", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_FIREBOMB], 0, 0, 0},
        {"player-artifact-egg", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_EGG], 0, 0, 0},
        {"player-artifact-wings", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_FLY], 0, 0, 0},
        {"player-artifact-chaosdevice", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TELEPORT], 0, 0, 0},
#elif __JHEXEN__
        {"player-ammo-bluemana", READONLYCVAR, CVT_INT, &gsvAmmo[AT_BLUEMANA], 0, 0, 0},
        /* Alias */ {"player-mana-blue", READONLYCVAR, CVT_INT, &gsvAmmo[AT_BLUEMANA], 0, 0, 0},
        {"player-ammo-greenmana", READONLYCVAR, CVT_INT, &gsvAmmo[AT_GREENMANA], 0, 0, 0},
        /* Alias */ {"player-mana-green", READONLYCVAR, CVT_INT, &gsvAmmo[AT_GREENMANA], 0, 0, 0},
        {"player-key-steel", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY1], 0, 0, 0},
        {"player-key-cave", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY2], 0, 0, 0},
        {"player-key-axe", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY3], 0, 0, 0},
        {"player-key-fire", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY4], 0, 0, 0},
        {"player-key-emerald", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY5], 0, 0, 0},
        {"player-key-dungeon", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY6], 0, 0, 0},
        {"player-key-silver", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY7], 0, 0, 0},
        {"player-key-rusted", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY8], 0, 0, 0},
        {"player-key-horn", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEY9], 0, 0, 0},
        {"player-key-swamp", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEYA], 0, 0, 0},
        {"player-key-castle", READONLYCVAR, CVT_INT, &gsvKeys[KT_KEYB], 0, 0, 0},
        {"player-weapon-first", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FIRST], 0, 0, 0},
        {"player-weapon-second", READONLYCVAR, CVT_INT, &gsvWeapons[WT_SECOND], 0, 0, 0},
        {"player-weapon-third", READONLYCVAR, CVT_INT, &gsvWeapons[WT_THIRD], 0, 0, 0},
        {"player-weapon-fourth", READONLYCVAR, CVT_INT, &gsvWeapons[WT_FOURTH], 0, 0, 0},
        {"player-artifact-defender", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_INVULNERABILITY], 0, 0, 0},
        {"player-artifact-quartzflask", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_HEALTH], 0, 0, 0},
        {"player-artifact-mysticurn", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_SUPERHEALTH], 0, 0, 0},
        {"player-artifact-mysticambit", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_HEALINGRADIUS], 0, 0, 0},
        {"player-artifact-darkservant", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_SUMMON], 0, 0, 0},
        {"player-artifact-torch", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TORCH], 0, 0, 0},
        {"player-artifact-porkalator", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_EGG], 0, 0, 0},
        {"player-artifact-wings", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_FLY], 0, 0, 0},
        {"player-artifact-repulsion", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_BLASTRADIUS], 0, 0, 0},
        {"player-artifact-flechette", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_POISONBAG], 0, 0, 0},
        {"player-artifact-banishment", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TELEPORTOTHER], 0, 0, 0},
        {"player-artifact-speed", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_SPEED], 0, 0, 0},
        {"player-artifact-might", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_BOOSTMANA], 0, 0, 0},
        {"player-artifact-bracers", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_BOOSTARMOR], 0, 0, 0},
        {"player-artifact-chaosdevice", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_TELEPORT], 0, 0, 0},
        {"player-artifact-skull", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZSKULL], 0, 0, 0},
        {"player-artifact-heart", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMBIG], 0, 0, 0},
        {"player-artifact-ruby", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMRED], 0, 0, 0},
        {"player-artifact-emerald1", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMGREEN1], 0, 0, 0},
        {"player-artifact-emerald2", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMGREEN2], 0, 0, 0},
        {"player-artifact-sapphire1", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMBLUE1], 0, 0, 0},
        {"player-artifact-sapphire2", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEMBLUE2], 0, 0, 0},
        {"player-artifact-daemoncodex", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZBOOK1], 0, 0, 0},
        {"player-artifact-liberoscura", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZBOOK2], 0, 0, 0},
        {"player-artifact-flamemask", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZSKULL2], 0, 0, 0},
        {"player-artifact-glaiveseal", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZFWEAPON], 0, 0, 0},
        {"player-artifact-holyrelic", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZCWEAPON], 0, 0, 0},
        {"player-artifact-sigilmagus", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZMWEAPON], 0, 0, 0},
        {"player-artifact-gear1", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEAR1], 0, 0, 0},
        {"player-artifact-gear2", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEAR2], 0, 0, 0},
        {"player-artifact-gear3", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEAR3], 0, 0, 0},
        {"player-artifact-gear4", READONLYCVAR, CVT_INT, &gsvInvItems[IIT_PUZZGEAR4], 0, 0, 0},
#endif
        {NULL, 0, CVT_NULL, 0, 0, 0, 0}
    };
    Con_AddVariableList(cvars);
#if __JHEXEN__
    for (int i = 0; i < WEAPON_FOURTH_PIECE_COUNT; ++i) {
        Block const cvarName = String("player-weapon-piece%1").arg(i + 1).toUtf8();
        C_VAR_INT(cvarName.constData(), &gsvWPieces[i], READONLYCVAR, 0, 1);
    }
    C_VAR_INT("player-weapon-allpieces", &gsvWPieces[WEAPON_FOURTH_PIECE_COUNT], READONLYCVAR, 0, 1);
#endif
}
#undef READONLYCVAR
void G_ConsoleRegister() {
    GameSession::consoleRegister();
    C_VAR_BYTE("game-save-confirm",              &cfg.common.confirmQuickGameSave,  0, 0, 1);
    /* Alias */ C_VAR_BYTE("menu-quick-ask",     &cfg.common.confirmQuickGameSave,  0, 0, 1);
    C_VAR_BYTE("game-save-confirm-loadonreborn", &cfg.common.confirmRebornLoad,     0, 0, 1);
    C_VAR_BYTE("game-save-last-loadonreborn",    &cfg.common.loadLastSaveOnReborn,  0, 0, 1);
    C_CMD("deletegamesave",     "ss",       DeleteSaveGame);
    C_CMD("deletegamesave",     "s",        DeleteSaveGame);
    C_CMD("endgame",            "s",        EndSession);
    C_CMD("endgame",            "",         EndSession);
    C_CMD("helpscreen",         "",         HelpScreen);
    C_CMD("leavemap",           "",         LeaveMap);
    C_CMD("leavemap",           "s",        LeaveMap);
    C_CMD("loadgame",           "ss",       LoadSession);
    C_CMD("loadgame",           "s",        LoadSession);
    C_CMD("loadgame",           "",         OpenLoadMenu);
    C_CMD("quickload",          "",         QuickLoadSession);
    C_CMD("quicksave",          "",         QuickSaveSession);
    C_CMD("savegame",           "sss",      SaveSession);
    C_CMD("savegame",           "ss",       SaveSession);
    C_CMD("savegame",           "s",        SaveSession);
    C_CMD("savegame",           "",         OpenSaveMenu);
    C_CMD("togglegamma",        "",         CycleTextureGamma);
    C_CMD("warp",               nullptr,    WarpMap);
    /* Alias */ C_CMD("setmap", nullptr,    WarpMap);
    C_CMD("setdefaultskill",    "i",        SetDefaultSkill);
    registerGameStatusCVars();
}
#define MANGLE_STATE(x)     (INT2PTR(state_t, ((x)? (x)-STATES : -1)))
#define RESTORE_STATE(x)    (PTR2INT(x) < 0? NULL : &STATES[PTR2INT(x)])
static int mangleMobj(thinker_t *th, void * /*context*/)
{
    mobj_t *mo = (mobj_t *) th;
    mo->state = MANGLE_STATE(mo->state);
    mo->info  = INT2PTR(mobjinfo_t, mo->info - MOBJINFO);
    return false; // Continue iteration.
}
static int restoreMobj(thinker_t *th, void * /*context*/)
{
    mobj_t *mo = (mobj_t *) th;
    mo->state = RESTORE_STATE(mo->state);
    mo->info  = &MOBJINFO[PTR2INT(mo->info)];
    return false; // Continue iteration.
}
/**
 * Called before the engine re-inits the definitions. After that all the
 * state, info, etc. pointers will be obsolete.
 */
void G_MangleState() {
    Thinker_Iterate(P_MobjThinker, mangleMobj, NULL);
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = &players[i];
        for(int k = 0; k < NUMPSPRITES; ++k) {
            plr->pSprites[k].state = MANGLE_STATE(plr->pSprites[k].state);
        }
    }
}
void G_RestoreState() {
    Thinker_Iterate(P_MobjThinker, restoreMobj, NULL);
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = &players[i];
        for(int k = 0; k < NUMPSPRITES; ++k) {
            plr->pSprites[k].state = RESTORE_STATE(plr->pSprites[k].state);
        }
    }
    HU_UpdatePsprites();
}
/**
 * Handles engine updates and renderer restarts.
 */
void G_UpdateState(int step) {
    switch(step) {
    case DD_PRE:
        G_MangleState();
        P_InitPicAnims(); // Redefine texture animations.
        break;
    case DD_POST:
        G_RestoreState();
        R_InitRefresh();
        R_LoadColorPalettes();
        P_Update();
#if !__JHEXEN__
        XG_Update();
#endif
#if __JHERETIC__ || __JHEXEN__ || __JDOOM64__
        P_InitInventory();
#endif
        Hu_MenuInit();
        G_SaveSlots().updateAll();
#if __JHEXEN__
        SndInfoParser(AutoStr_FromText("Lumps:SNDINFO"));
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        S_MapMusic(gfw_Session()->mapUri());
#endif
        break;
    case DD_RENDER_RESTART_PRE:
        Hu_UnloadData();
        GUI_ReleaseResources();
        break;
    case DD_RENDER_RESTART_POST:
        Hu_LoadData();
        GUI_LoadResources();
        break;
    }
}
/*
 * These keys are used for serialization, so if changed, the only keys still need to be changed
 * when reading data.
 */
String const GameRules::VAR_skill           = "skill";
String const GameRules::VAR_fast            = "fast";
String const GameRules::VAR_deathmatch      = "deathmatch";
String const GameRules::VAR_noMonsters      = "noMonsters";
String const GameRules::VAR_respawnMonsters = "respawnMonsters";
String const GameRules::VAR_randomClasses   = "randomClasses";
DENG2_PIMPL_NOREF(GameRules) {
    Record rules {
        Record::withMembers(GameRules::VAR_skill,           2, // medium
                            GameRules::VAR_fast,            false,
                            GameRules::VAR_deathmatch,      0,
                            GameRules::VAR_noMonsters,      false,
                            GameRules::VAR_randomClasses,   false,
                            GameRules::VAR_respawnMonsters, false) };
    Impl() {}
    Impl(Impl const &other)
        : rules(other.rules)
    {}
};
GameRules::GameRules()
    : d(new Impl) {
    update();
}
GameRules::GameRules(GameRules const &other)
    : d(new Impl(*other.d)) {
    update();
}
GameRules *GameRules::fromRecord(Record const &record, GameRules const *defaults) // static
{
    GameRules *gr = new GameRules;
    if (defaults) {
        gr->d->rules.copyMembersFrom(defaults->asRecord(), Record::IgnoreDoubleUnderscoreMembers);
    }
    gr->d->rules.copyMembersFrom(record, Record::IgnoreDoubleUnderscoreMembers);
    return gr;
}
Record &GameRules::asRecord() {
    return d->rules;
}
Record const &GameRules::asRecord() const {
    return d->rules;
}
GameRules &GameRules::operator = (GameRules const &other) {
    d->rules = other.d->rules;
    update();
    return *this;
}
String GameRules::description() const {
    if (IS_NETGAME) {
        if(values.deathmatch == 2) return "Deathmatch2";
        if(values.deathmatch)      return "Deathmatch";
        return "Co-op";
    }
    return "Singleplayer";
}
String GameRules::asText() const {
    String str;
    QTextStream os(&str);
    os << "skillmode: " << int(values.skill);
#if defined(__JHEXEN__)
    os << " random player classes: " << (values.randomClasses ? "yes" : "no");
#endif
    os << " monsters: " << (!values.noMonsters     ? "yes" : "no");
#if !defined(__JHEXEN__)
    os << " (fast: "    << (values.fast            ? "yes" : "no");
    os << " respawn: "  << (values.respawnMonsters ? "yes" : "no") << ")";
#endif
    return str;
}
void GameRules::update() {
    Values *vals = const_cast<Values *>(&values);
    vals->skill           = d->rules.geti(VAR_skill);
    vals->fast            = d->rules.getb(VAR_fast);
    vals->deathmatch      = byte(d->rules.geti(VAR_deathmatch));
    vals->noMonsters      = d->rules.getb(VAR_noMonsters);
    vals->respawnMonsters = d->rules.getb(VAR_respawnMonsters);
#if defined(__JHEXEN__)
    vals->randomClasses = d->rules.getb(VAR_randomClasses);
#endif
}
DENG_EXTERN_C int gfw_SessionRule(gfw_gamerule_t rule) {
    switch (rule) {
    case GFW_RULE_skill:            return gfw_Rule(skill);
    case GFW_RULE_fast:             return gfw_Rule(fast);
    case GFW_RULE_deathmatch:       return gfw_Rule(deathmatch);
    case GFW_RULE_noMonsters:       return gfw_Rule(noMonsters);
    case GFW_RULE_respawnMonsters:  return gfw_Rule(respawnMonsters);
#if defined(__JHEXEN__)
    case GFW_RULE_randomClasses:    return gfw_Rule(randomClasses);
#endif
    default: return 0;
    }
}
namespace common {
namespace internal {
    static String composeSaveInfo(GameStateMetadata const &metadata) {
        String info;
        QTextStream os(&info);
        os.setCodec("UTF-8");
        Time now;
        os <<   "# Doomsday Engine game state package.\n#"
           << "\n# Generator: GameSession (libcommon)"
           << "\n# Date: " + now.asDateTime().toString(Qt::SystemLocaleShortDate);
        os << "\n\n" + metadata.asInfo() + "\n";
        return info;
    }
    /**
     * Lookup the briefing Finale for the current episode, map (if any).
     */
    static Record const *finaleBriefing(de::Uri const &mapUri) {
        if (::briefDisabled) return nullptr;
        if (IS_CLIENT || Get(DD_PLAYBACK)) return nullptr;
        if (G_GameState() == GS_INFINE) return nullptr;
        return Defs().finales.tryFind("before", mapUri.compose());
    }
}
using namespace internal;
static String const internalSavePath = "/home/cache/internal.save";
static GameSession theSession;
DENG2_PIMPL(GameSession), public GameStateFolder::IMapStateReaderFactory {
    String episodeId;
    GameRules rules;
    duint mapEntryPoint = 0;  ///< Player entry point, for reborn.
    bool rememberVisitedMaps = false;
    QSet<de::Uri> visitedMaps;
    acs::System acscriptSys;  ///< The One acs::System instance.
    Impl(Public *i) : Base(i)
    {}
    inline String userSavePath(String const &fileName) {
        DENG_ASSERT(DoomsdayApp::currentGameProfile());
        return SaveGames::savePath() / fileName + ".save";
    }
    void cleanupInternalSave() {
        App::fileSystem().makeFolder(internalSavePath.fileNamePath());
        AbstractSession::removeSaved(internalSavePath);
    }
    void resetStateForNewSession() {
        cleanupInternalSave();
        G_StopDemo();
        Hu_MenuCommand(MCMD_CLOSEFAST);
        FI_StackClear();
        G_SetGameAction(GA_NONE);
        if (!IS_CLIENT) {
            for (player_t &plr : players) {
                if (!plr.plr->inGame) continue;
                plr.playerState = PST_REBORN;
#if __JHEXEN__
                plr.worldTimer  = 0;
#else
                plr.didSecret   = false;
#endif
            }
        }
        M_ResetRandom();
    }
    void setEpisode(String const &newEpisodeId) {
        DENG2_ASSERT(!self().hasBegun());
        episodeId = newEpisodeId;
        Con_SetString2("map-episode", episodeId.toUtf8().constData(), SVF_WRITE_OVERRIDE);
    }
    /**
     * Returns GameStateMetadata for the game configuration in progress.
     */
    GameStateMetadata metadata() {
        DENG2_ASSERT(self().hasBegun());
        GameStateMetadata meta;
        meta.set("sessionId",       duint(Timer_RealMilliseconds() + (mapTime << 24)) & DDMAXINT);
        meta.set("gameIdentityKey", gfw_GameId());
        meta.set("episode",         episodeId);
        meta.set("userDescription", "(Unsaved)");
        meta.set("mapUri",          self().mapUri().compose());
        meta.set("mapTime",         mapTime);
        meta.add("gameRules",       new Record(self().rules().asRecord()));
        auto *loadedPackages = new ArrayValue;
        for (String id : PackageLoader::get().loadedPackageIdsInOrder()) {
            if (GameStateFolder::isPackageAffectingGameplay(id)) {
                *loadedPackages << new TextValue(id);
            }
        }
        meta.set("packages", loadedPackages);
        auto *playersArray = new ArrayValue;
        for (player_t &plr : players) {
            *playersArray << new NumberValue(CPP_BOOL(plr.plr->inGame), NumberValue::Boolean);
        }
        meta.set("players", playersArray);  // Takes ownership.
        if (rememberVisitedMaps) {
            ArrayValue *visitedMapsArray = new ArrayValue;
            for (de::Uri const &visitedMap : visitedMaps) {
                *visitedMapsArray << TextValue(visitedMap.compose());
            }
            meta.set("visitedMaps", visitedMapsArray); // Takes ownership.
        }
        return meta;
    }
    /**
     * Write the current map state to a file and notify the application about the change
     * in the game state folder.
     *
     * @param dest            Destination file for the serialized map state.
     * @param saveFolder      Folder containing the save.
     * @param excludePlayers  Should players be excluded from the state?
     */
    void serializeCurrentMapState(File &dest, GameStateFolder &saveFolder, bool excludePlayers = false) {
        Block data;
        SV_OpenFileForWrite(data);
        writer_s *writer = SV_NewWriter();
        MapStateWriter mapStateWriter;
        mapStateWriter.write(writer, excludePlayers);
        Writer_Delete(writer);
        SV_CloseFile();
        dest << data;
        DoomsdayApp::app().gameSessionWasSaved(self(), saveFolder);
    }
    /**
     * Update/create a new GameStateFolder at the specified @a path from the current
     * game state.
     */
    GameStateFolder &updateGameStateFolder(String const &path, GameStateMetadata const &metadata) {
        DENG2_ASSERT(self().hasBegun());
        LOG_AS("GameSession");
        LOG_RES_VERBOSE("Serializing to \"%s\"...") << path;
        auto *saved = App::rootFolder().tryLocate<GameStateFolder>(path);
        if (saved) {
            DENG2_ASSERT(saved->mode().testFlag(File::Write));
            saved->replaceFile("Info") << composeSaveInfo(metadata).toUtf8();
        }
        else {
            File &save = App::rootFolder().replaceFile(path);
            ZipArchive arch;
            arch.add("Info", composeSaveInfo(metadata).toUtf8());
            de::Writer(save) << arch;
            save.flush();
            saved = &save.reinterpret()->as<GameStateFolder>();
            saved->populate();
        }
#if __JHEXEN__
        de::Writer(saved->replaceFile("ACScriptState")).withHeader()
                << acscriptSys.serializeWorldState();
#endif
        Folder &mapsFolder = App::fileSystem().makeFolder(saved->path() / "maps");
        DENG2_ASSERT(mapsFolder.mode().testFlag(File::Write));
        serializeCurrentMapState(mapsFolder.replaceFile(self().mapUri().path() + "State"),
                                 *saved);
        saved->flush();  // No need to populate; FS2 Files already in sync with source data.
        saved->cacheMetadata(metadata);  // Avoid immediately reopening the .save package.
        return *saved;
    }
#if __JDOOM__ || __JDOOM64__
    /**
     * @todo fixme: (Kludge) Assumes the original mobj info tic timing values have
     * not been modified!
     */
    void applyRuleFastMonsters(bool fast) {
        static bool oldFast = false;
        if (fast == oldFast) return;
        oldFast = fast;
        for (dint i = S_SARG_RUN1; i <= S_SARG_RUN8; ++i) {
            STATES[i].tics = fast? 1 : 2;
        }
        for (dint i = S_SARG_ATK1; i <= S_SARG_ATK3; ++i) {
            STATES[i].tics = fast? 4 : 8;
        }
        for (dint i = S_SARG_PAIN; i <= S_SARG_PAIN2; ++i) {
            STATES[i].tics = fast? 1 : 2;
        }
    }
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    /**
     * @todo fixme: (Kludge) Assumes the original mobj info speed values have
     * not been modified!
     */
    void applyRuleFastMissiles(bool fast) {
        struct MissileData { mobjtype_t type; dfloat speed[2]; } const missileData[] = {
#if __JDOOM__ || __JDOOM64__
            { MT_BRUISERSHOT,       { 15, 20 } },
            { MT_HEADSHOT,          { 10, 20 } },
            { MT_TROOPSHOT,         { 10, 20 } },
# if __JDOOM64__
            { MT_BRUISERSHOTRED,    { 15, 20 } },
            { MT_NTROSHOT,          { 20, 40 } },
# endif
#elif __JHERETIC__
            { MT_IMPBALL,           { 10, 20 } },
            { MT_MUMMYFX1,          {  9, 18 } },
            { MT_KNIGHTAXE,         {  9, 18 } },
            { MT_REDAXE,            {  9, 18 } },
            { MT_BEASTBALL,         { 12, 20 } },
            { MT_WIZFX1,            { 18, 24 } },
            { MT_SNAKEPRO_A,        { 14, 20 } },
            { MT_SNAKEPRO_B,        { 14, 20 } },
            { MT_HEADFX1,           { 13, 20 } },
            { MT_HEADFX3,           { 10, 18 } },
            { MT_MNTRFX1,           { 20, 26 } },
            { MT_MNTRFX2,           { 14, 20 } },
            { MT_SRCRFX1,           { 20, 28 } },
            { MT_SOR2FX1,           { 20, 28 } },
#endif
        };
        static bool oldFast = false;
        if (fast == oldFast) return;
        oldFast = fast;
        for (MissileData const &mdata : missileData) {
            MOBJINFO[mdata.type].speed = mdata.speed[dint( fast )];
        }
    }
#endif
    void applyCurrentRules() {
        if (rules.values.skill < SM_NOTHINGS) {
            GameRules_Set(rules, skill, SM_NOTHINGS);
        }
        if (rules.values.skill > NUM_SKILL_MODES - 1) {
            GameRules_Set(rules, skill, skillmode_t(NUM_SKILL_MODES - 1));
        }
        if (!IS_NETGAME) {
#if !__JHEXEN__
            GameRules_Set(rules, deathmatch, 0);
            GameRules_Set(rules,
                          respawnMonsters,
                          CommandLine::get().has("-respawn") ||
                              gfw_GameProfile()->optionValue("respawn").isTrue());
            GameRules_Set(rules,
                          noMonsters,
                          CommandLine::get().has("-nomonsters") ||
                              gfw_GameProfile()->optionValue("noMonsters").isTrue());
#endif
#if __JDOOM__ || __JHERETIC__
            if (rules.values.skill == SM_NIGHTMARE) {
                GameRules_Set(rules, respawnMonsters, cfg.respawnMonstersNightmare);
            }
#endif
        }
        else if (IS_DEDICATED) {
#if !__JHEXEN__
            GameRules_Set(rules, deathmatch, cfg.common.netDeathmatch);
            GameRules_Set(rules, respawnMonsters, cfg.netRespawn);
            GameRules_Set(rules, noMonsters, cfg.common.netNoMonsters);
            /*rules.*/cfg.common.jumpEnabled = cfg.common.netJumping;
#else
            GameRules_Set(rules, randomClasses, cfg.netRandomClass);
#endif
        }
#if __JDOOM__ || __JDOOM64__
        bool fastMonsters = rules.values.fast;
# if __JDOOM__
        if (rules.values.skill == SM_NIGHTMARE) {
            fastMonsters = true;
        }
# endif
        applyRuleFastMonsters(fastMonsters);
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        bool fastMissiles = rules.values.fast;
# if !__JDOOM64__
        if (rules.values.skill == SM_NIGHTMARE) {
            fastMissiles = true;
        }
# endif
        applyRuleFastMissiles(fastMissiles);
#endif
        NetSv_UpdateGameConfigDescription();
        Con_SetInteger2("game-skill", rules.values.skill, SVF_WRITE_OVERRIDE);
    }
    /**
     * Constructs a MapStateReader for serialized map state format interpretation.
     */
    GameStateFolder::MapStateReader *makeMapStateReader(
        GameStateFolder const &session, String const &mapUriAsText) {
        de::Uri const mapUri(mapUriAsText, RC_NULL);
        auto const &mapStateFile = session.locateState<File const>(String("maps") / mapUri.path());
        if (!SV_OpenFileForRead(mapStateFile)) {
            throw Error("GameSession::makeMapStateReader", "Failed to open \"" + mapStateFile.path() + "\" for read");
        }
        std::unique_ptr<GameStateFolder::MapStateReader> p;
        reader_s *reader = SV_NewReader();
        dint const magic = Reader_ReadInt32(reader);
        if (magic == MY_SAVE_MAGIC || magic == MY_CLIENT_SAVE_MAGIC)  // Native format.
        {
            p.reset(new MapStateReader(session));
        }
#if __JDOOM__
        else if (magic == 0x1DEAD600)  // DoomV9
        {
            p.reset(new DoomV9MapStateReader(session));
        }
#endif
#if __JHERETIC__
        else if (magic == 0x7D9A1200)  // HereticV13
        {
            p.reset(new HereticV13MapStateReader(session));
        }
#endif
        SV_CloseFile();
        if (p.get()) {
            return p.release();
        }
        throw Error("GameSession::makeMapStateReader", "Unrecognized map state format");
    }
    void loadSaved(String const &savePath) {
        ::briefDisabled = true;
        G_StopDemo();
        Hu_MenuCommand(MCMD_CLOSEFAST);
        FI_StackClear(); // Stop any running InFine scripts.
        M_ResetRandom();
        if (!IS_CLIENT) {
            for (player_t &plr : players) {
                if (!plr.plr->inGame) continue;
                plr.playerState = PST_REBORN;
#if __JHEXEN__
                plr.worldTimer  = 0;
#else
                plr.didSecret   = false;
#endif
            }
        }
        self().setInProgress(false);
        if (savePath.compareWithoutCase(internalSavePath)) {
            cleanupInternalSave();
            AbstractSession::copySaved(internalSavePath, savePath);
        }
        auto const &saved = App::rootFolder().locate<GameStateFolder>(internalSavePath);
        GameStateMetadata const &metadata = saved.metadata();
        std::unique_ptr<GameRules> newRules;
        try {
            newRules.reset(GameRules::fromRecord(metadata.subrecord("gameRules")));
        }
        catch (Record::NotFoundError const &) {
            LOG_WARNING("Using current game rules as basis for loading savegame \"%s\"."
                        " (The original save format omits this information).")
                    << saved.path();
            newRules.reset(GameRules::fromRecord(metadata.subrecord("gameRules"), &rules));
        }
        rules = *newRules; // make a copy
        applyCurrentRules();
        setEpisode(metadata.gets("episode"));
        visitedMaps.clear();
        rememberVisitedMaps = metadata.has("visitedMaps");
        if (rememberVisitedMaps) {
            ArrayValue const &vistedMapsArray = metadata.geta("visitedMaps");
            for (Value const *value : vistedMapsArray.elements()) {
                visitedMaps << de::makeUri(value->as<TextValue>());
            }
        }
#if __JHEXEN__
        if (File const *state = saved.tryLocateStateFile("ACScript")) {
            acscriptSys.readWorldState(de::Reader(*state).withHeader());
        }
#endif
        self().setInProgress(true);
        setMap(de::makeUri(metadata.gets("mapUri")));
        reloadMap();
#if !__JHEXEN__
        ::mapTime = metadata.geti("mapTime");
#endif
        String const mapUriAsText = self().mapUri().compose();
        std::unique_ptr<GameStateFolder::MapStateReader> mapReader(makeMapStateReader(saved, mapUriAsText));
        self().setThinkerMapping(mapReader.get());
        mapReader->read(mapUriAsText);
        DoomsdayApp::app().gameSessionWasLoaded(self(), saved);
        self().setThinkerMapping(nullptr);
    }
    void setMap(de::Uri const &newMapUri) {
        DENG2_ASSERT(self().hasBegun());
        self().setMapUri(newMapUri);
        de::Uri const mapUri = self().mapUri();
        if (rememberVisitedMaps) {
            visitedMaps << mapUri;
        }
        Con_SetUri2("map-id", reinterpret_cast<uri_s const *>(&mapUri), SVF_WRITE_OVERRIDE);
        String hubId;
        if (Record const *hubRec = defn::Episode(*self().episodeDef()).tryFindHubByMapId(mapUri.compose())) {
            hubId = hubRec->gets("id");
        }
        Con_SetString2("map-hub", hubId.toUtf8().constData(), SVF_WRITE_OVERRIDE);
        String mapAuthor = G_MapAuthor(mapUri);
        if (mapAuthor.isEmpty()) mapAuthor = "Unknown";
        Con_SetString2("map-author", mapAuthor.toUtf8().constData(), SVF_WRITE_OVERRIDE);
        String mapTitle = G_MapTitle(mapUri);
        if (mapTitle.isEmpty()) mapTitle = "Unknown";
        Con_SetString2("map-name", mapTitle.toUtf8().constData(), SVF_WRITE_OVERRIDE);
    }
    inline void setMapAndEntryPoint(de::Uri const &newMapUri, duint newMapEntryPoint) {
        setMap(newMapUri);
        mapEntryPoint = newMapEntryPoint;
    }
    /**
     * Reload the @em current map.
     *
     * @param revisit  @c true= load progress in this map from a previous visit in the current
     * game session. If no saved progress exists then the map will be in the default state.
     */
    void reloadMap(bool revisit = false) {
        DENG2_ASSERT(self().hasBegun());
        Pause_End();
        for (dint i = 0; i < MAXPLAYERS; ++i) {
            ST_CloseAll(i, true/*fast*/);
        }
        if (!IS_SERVER) {
            DD_Executef(true, "texreset raw");
        }
        if (revisit) {
            ::briefDisabled = true;
        }
        Record const *briefing = finaleBriefing(self().mapUri());
        if (!briefing) {
            S_MapMusic(self().mapUri());
            S_PauseMusic(true);
        }
        P_SetupMap(self().mapUri());
        if (revisit) {
#if __JHEXEN__
            targetPlayerAddrs = nullptr; // player mobj redirection...
#endif
            String const mapUriAsText = self().mapUri().compose();
            auto const &saved = App::rootFolder().locate<GameStateFolder>(internalSavePath);
            std::unique_ptr<GameStateFolder::MapStateReader> reader(makeMapStateReader(saved, mapUriAsText));
            self().setThinkerMapping(reader.get());
            reader->read(mapUriAsText);
            DoomsdayApp::app().gameSessionWasLoaded(self(), saved);
            self().setThinkerMapping(nullptr);
        }
        if (!briefing || !G_StartFinale(briefing->gets("script").toUtf8().constData(), 0, FIMODE_BEFORE, 0)) {
            HU_WakeWidgets(-1/* all players */);
            G_BeginMap();
        }
        Z_CheckHeap();
    }
#if __JHEXEN__
    struct playerbackup_t {
        player_t player;
        duint numInventoryItems[NUM_INVENTORYITEM_TYPES];
        inventoryitemtype_t readyItem;
    };
    void backupPlayersInHub(playerbackup_t playerBackup[MAXPLAYERS]) {
        DENG2_ASSERT(playerBackup);
        for (dint i = 0; i < MAXPLAYERS; ++i) {
            playerbackup_t *pb  = &playerBackup[i];
            player_t const *plr = &players[i];
            std::memcpy(&pb->player, plr, sizeof(player_t));
            for (dint k = 0; k < NUM_INVENTORYITEM_TYPES; ++k) {
                pb->numInventoryItems[k] = P_InventoryCount(i, inventoryitemtype_t(k));
            }
            pb->readyItem = P_InventoryReadyItem(i);
        }
    }
    /**
     * @param playerBackup  Player state backup.
     */
    void restorePlayersInHub(playerbackup_t playerBackup[MAXPLAYERS]) {
        DENG2_ASSERT(playerBackup);
        mobj_t *targetPlayerMobj = nullptr;
        for (dint i = 0; i < MAXPLAYERS; ++i) {
            playerbackup_t *pb = &playerBackup[i];
            player_t *plr      = &players[i];
            ddplayer_t *ddplr  = plr->plr;
            dint oldKeys = 0, oldPieces = 0;
            dd_bool oldWeaponOwned[NUM_WEAPON_TYPES];
            if (!ddplr->inGame) continue;
            std::memcpy(plr, &pb->player, sizeof(player_t));
            P_InventoryEmpty(i);
            for (dint k = 0; k < NUM_INVENTORYITEM_TYPES; ++k) {
                if (k == IIT_FLY && plr->playerState == PST_REBORN)
                    continue;
                for (duint m = 0; m < pb->numInventoryItems[k]; ++m) {
                    P_InventoryGive(i, inventoryitemtype_t(k), true);
                }
            }
            P_InventorySetReadyItem(i, pb->readyItem);
            ST_LogEmpty(i);
            plr->attacker = nullptr;
            plr->poisoner = nullptr;
            if (IS_NETGAME || rules.values.deathmatch) {
                if (plr->playerState == PST_DEAD) {
                    plr->playerState = PST_REBORN;
                }
                if (!rules.values.deathmatch) {
                    oldKeys   = plr->keys;
                    oldPieces = plr->pieces;
                    for (dint k = 0; k < NUM_WEAPON_TYPES; ++k) {
                        oldWeaponOwned[k] = plr->weapons[k].owned;
                    }
                }
            }
            bool wasReborn = (plr->playerState == PST_REBORN);
            if (rules.values.deathmatch) {
                de::zap(plr->frags);
                ddplr->mo = nullptr;
                G_DeathMatchSpawnPlayer(i);
            }
            else {
                if (playerstart_t const *start = P_GetPlayerStart(mapEntryPoint, i, false)) {
                    mapspot_t const *spot = &mapSpots[start->spot];
                    P_SpawnPlayer(i, cfg.playerClass[i], spot->origin[0],
                                  spot->origin[1], spot->origin[2], spot->angle,
                                  spot->flags, false, true);
                }
                else {
                    P_SpawnPlayer(i, cfg.playerClass[i], 0, 0, 0, 0,
                                  MSF_Z_FLOOR, true, true);
                }
            }
            if (wasReborn && IS_NETGAME && !rules.values.deathmatch) {
                dint bestWeapon = 0;
                plr->keys   = oldKeys;
                plr->pieces = oldPieces;
                for (dint k = 0; k < NUM_WEAPON_TYPES; ++k) {
                    if (oldWeaponOwned[k]) {
                        bestWeapon = k;
                        plr->weapons[k].owned = true;
                    }
                }
                plr->ammo[AT_BLUEMANA].owned  = 25;  /// @todo values.ded
                plr->ammo[AT_GREENMANA].owned = 25;  /// @todo values.ded
                if (bestWeapon) {
                    plr->pendingWeapon = weapontype_t(bestWeapon);
                }
            }
        }
        for (player_t const &plr : players) {
            if (!plr.plr->inGame) continue;
            if (!targetPlayerMobj) {
                targetPlayerMobj = plr.plr->mo;
            }
        }
        if (targetPlayerAddrs) {
            for (targetplraddress_t *p = targetPlayerAddrs; p; p = p->next) {
                *(p->address) = targetPlayerMobj;
            }
            SV_ClearTargetPlayers();
            /*
             * When XG is available in Hexen, call this after updating target player
             * references (after a load) - ds
            XL_UpdateActivators();
            */
        }
        P_TelefragMobjsTouchingPlayers();
    }
#endif // __JHEXEN__
};
GameSession::GameSession() : d(new Impl(this))
{}
GameSession::~GameSession() {
    LOG_AS("~GameSession");
    d.reset();
}
GameSession &GameSession::gameSession() {
    return theSession;
}
bool GameSession::isLoadingPossible() {
    return !(IS_CLIENT && !Get(DD_PLAYBACK));
}
bool GameSession::isSavingPossible() {
    if (IS_CLIENT || Get(DD_PLAYBACK)) return false;
    if (!hasBegun()) return false;
    if (GS_MAP != G_GameState()) return false;
    player_t *player = &players[CONSOLEPLAYER];
    if (PST_DEAD == player->playerState) return false;
    return true;
}
Record const *GameSession::episodeDef() const {
    if (hasBegun()) {
        return Defs().episodes.tryFind("id", d->episodeId);
    }
    return nullptr;
}
String GameSession::episodeId() const {
    return hasBegun()? d->episodeId : "";
}
Record const *GameSession::mapGraphNodeDef() const {
    if (Record const *episode = episodeDef()) {
        return defn::Episode(*episode).tryFindMapGraphNode(mapUri().compose());
    }
    return nullptr;
}
Record const &GameSession::mapInfo() const {
    return G_MapInfoForMapUri(mapUri());
}
uint GameSession::mapEntryPoint() const {
    return d->mapEntryPoint;
}
GameSession::VisitedMaps GameSession::allVisitedMaps() const {
    if (hasBegun() && d->rememberVisitedMaps) {
        return d->visitedMaps.toList();
    }
    return VisitedMaps();
}
de::Uri GameSession::mapUriForNamedExit(String name) const {
    LOG_AS("GameSession");
    if (Record const *mgNode = mapGraphNodeDef()) {
        QMap<String, Record const *> exits;
        for (Value const *value : mgNode->geta("exit").elements()) {
            Record const &exit = value->as<RecordValue>().dereference();
            String id = exit.gets("id");
            if (!id.isEmpty()) {
                exits.insert(id, &exit);
            }
        }
        Record const *chosenExit = nullptr;
        if (exits.count() > 1) {
            auto found = exits.constFind(name.toLower());
            if (found != exits.constEnd()) {
                chosenExit = found.value();
            }
            else {
                LOG_SCR_WARNING("Episode '%s' map \"%s\" defines no Exit with ID '%s'")
                        << d->episodeId << mapUri() << name;
            }
        }
        else if (exits.count() == 1) {
            chosenExit = exits.values().first();
            String chosenExitId = chosenExit->gets("id");
            if (chosenExitId != name.toLower()) {
                LOGDEV_SCR_NOTE("Exit ID:%s chosen instead of '%s'")
                        << chosenExitId << name;
            }
        }
        if (chosenExit) {
            return de::makeUri(chosenExit->gets("targetMap"));
        }
    }
    return de::Uri();
}
GameRules const &GameSession::rules() const {
    return d->rules;
}
void GameSession::applyNewRules(GameRules const &newRules) {
    LOG_AS("GameSession");
    d->rules = newRules;
    if (hasBegun()) {
        d->applyCurrentRules();
        LOGDEV_WARNING("Applied new rules while in progress!");
    } // Otherwise deferred
}
bool GameSession::progressRestoredOnReload() const {
    if (d->rules.values.deathmatch) return false; // Never.
#if __JHEXEN__
    return true; // Cannot be disabled.
#else
    return cfg.common.loadLastSaveOnReborn;
#endif
}
void GameSession::end() {
    if (!hasBegun()) return;
#if __JHEXEN__
    d->acscriptSys.reset();
#endif
    if (!IS_DEDICATED) {
        G_ResetViewEffects();
    }
    AbstractSession::removeSaved(internalSavePath);
    setInProgress(false);
    LOG_MSG("Game ended");
}
void GameSession::endAndBeginTitle() {
    end();
    if (Record const *finale = Defs().finales.tryFind("id", "title")) {
        G_StartFinale(finale->gets("script").toUtf8().constData(), FF_LOCAL, FIMODE_NORMAL, "title");
        return;
    }
    throw Error("GameSession::endAndBeginTitle", "An InFine 'title' script must be defined");
}
void GameSession::begin(GameRules const &newRules, String const &episodeId,
    de::Uri const &mapUri, uint mapEntryPoint) {
    if (hasBegun()) {
        throw InProgressError("GameSession::begin", "The game session has already begun");
    }
    if (!Defs().episodes.has("id", episodeId)) {
        throw Error("GameSession::begin", "Episode '" + episodeId + "' is not known");
    }
    if (!P_MapExists(mapUri.compose().toUtf8().constData())) {
        throw Error("GameSession::begin", "Map \"" + mapUri.asText() + "\" does not exist");
    }
    LOG_MSG("Game begins...");
    d->resetStateForNewSession();
    d->rules = newRules; // make a copy
    d->applyCurrentRules();
    d->setEpisode(episodeId);
    d->visitedMaps.clear();
    d->rememberVisitedMaps = true;
    setInProgress(true);
    d->setMapAndEntryPoint(mapUri, mapEntryPoint);
    GameStateMetadata metadata = d->metadata();
    LOG_MSG(DE2_ESC(R));
    LOG_NOTE("Episode: " DE2_ESC(i) DE2_ESC(b) "%s" DE2_ESC(.) " (%s)")
            << G_EpisodeTitle(episodeId)
            << d->rules.description();
    LOG_VERBOSE("%s") << metadata.asStyledText();
    LOG_MSG(DE2_ESC(R));
    d->reloadMap();
    d->updateGameStateFolder(internalSavePath, metadata);
}
void GameSession::reloadMap() {
    if (!hasBegun()) {
        throw InProgressError("GameSession::reloadMap", "No game session is in progress");
    }
    if (progressRestoredOnReload()) {
        try {
            d->loadSaved(internalSavePath);
            return;
        }
        catch (Error const &er) {
            LOG_AS("GameSession");
            LOG_WARNING("Error loading current map state:\n") << er.asText();
        }
        endAndBeginTitle();
    }
    else {
        bool oldBriefDisabled = ::briefDisabled;
        ::briefDisabled = true; // We won't brief again.
        end();
        d->resetStateForNewSession();
        setInProgress(true);
        d->reloadMap();
        d->updateGameStateFolder(internalSavePath, d->metadata());
        ::briefDisabled = oldBriefDisabled;
    }
}
void GameSession::leaveMap(de::Uri const &nextMapUri, uint nextMapEntryPoint) {
    if (!hasBegun()) {
        throw InProgressError("GameSession::leaveMap", "No game session is in progress");
    }
    if (!P_MapExists(nextMapUri.compose().toUtf8().constData())) {
        throw Error("GameSession::leaveMap", "Map \"" + nextMapUri.asText() + "\" does not exist");
    }
    FI_StackClear();
#if __JHEXEN__
    Impl::playerbackup_t playerBackup[MAXPLAYERS];
    d->backupPlayersInHub(playerBackup);
    dbyte oldRandomClassesRule = d->rules.values.randomClasses;
    GameRules_Set(d->rules, randomClasses, false);
#endif
    GameStateFolder *saved = nullptr;
    if (!d->rules.values.deathmatch) // Never save in deathmatch.
    {
        saved = &App::rootFolder().locate<GameStateFolder>(internalSavePath);
        auto &mapsFolder = saved->locate<Folder>("maps");
        DENG2_ASSERT(saved->mode().testFlag(File::Write));
        DENG2_ASSERT(mapsFolder.mode().testFlag(File::Write));
#if __JHEXEN__
        defn::Episode epsd(*episodeDef());
        Record const *currentHub = epsd.tryFindHubByMapId(mapUri().compose());
        if (!currentHub || currentHub != epsd.tryFindHubByMapId(nextMapUri.compose()))
#endif
        {
            for (auto name : mapsFolder.contents().keys()) {
                mapsFolder.destroyFile(name);
            }
        }
#if __JHEXEN__
        else {
            File &outFile = mapsFolder.replaceFile(mapUri().path() + "State");
            d->serializeCurrentMapState(outFile, *saved, true /*exclude players*/);
        }
#endif
        saved->flush();
    }
#if __JHEXEN__
    if (!IS_CLIENT) {
        for (player_t &plr : players) {
            if (plr.plr->inGame) {
                plr.playerState = PST_REBORN;
                plr.worldTimer = 0;
            }
        }
    }
    M_ResetRandom();
#endif
    d->setMapAndEntryPoint(nextMapUri, nextMapEntryPoint);
    bool const revisit = saved && saved->hasState(String("maps") / mapUri().path());
    d->reloadMap(revisit);
#if __JHEXEN__
    if (!revisit) {
        for (player_t &plr : players) {
            if (plr.plr->inGame) {
                P_MobjRemove(plr.plr->mo, true);
            }
        }
    }
    d->restorePlayersInHub(playerBackup);
    GameRules_Set(d->rules, randomClasses, oldRandomClassesRule);
    d->acscriptSys.runDeferredTasks(mapUri());
#endif
    if (saved) {
        DENG2_ASSERT(saved->mode().testFlag(File::Write));
        GameStateMetadata metadata = d->metadata();
        saved->replaceFile("Info") << composeSaveInfo(metadata).toUtf8();
#if __JHEXEN__
        de::Writer(saved->replaceFile("ACScriptState")).withHeader()
                << d->acscriptSys.serializeWorldState();
#endif
        auto &mapsFolder = saved->locate<Folder>("maps");
        DENG2_ASSERT(mapsFolder.mode().testFlag(File::Write));
        File &outFile = mapsFolder.replaceFile(mapUri().path() + "State");
        d->serializeCurrentMapState(outFile, *saved);
        saved->flush(); // Write all changes to the package.
        saved->cacheMetadata(metadata); // Avoid immediately reopening the .save package.
    }
}
String GameSession::userDescription() {
    if (!hasBegun()) return "";
    return App::rootFolder().locate<GameStateFolder>(internalSavePath)
                                .metadata().gets("userDescription", "");
}
static String chooseSaveDescription(String const &savePath, String const &userDescription) {
    if (!userDescription.isEmpty()) {
        return userDescription;
    }
    return G_DefaultGameStateFolderUserDescription(savePath.fileNameWithoutExtension());
}
void GameSession::save(String const &saveName, String const &userDescription) {
    if (!hasBegun()) {
        throw InProgressError("GameSession::save", "No game session is in progress");
    }
    String const savePath = d->userSavePath(saveName);
    LOG_MSG("Saving game to \"%s\"...") << savePath;
    try {
        GameStateMetadata metadata = d->metadata();
        metadata.set("userDescription", chooseSaveDescription(savePath, userDescription));
        d->updateGameStateFolder(internalSavePath, metadata);
        NetSv_SaveGame(metadata.getui("sessionId"));
        AbstractSession::copySaved(savePath, internalSavePath);
        P_SetMessage(&players[CONSOLEPLAYER], TXT_GAMESAVED);
        Plug_Notify(DD_NOTIFY_GAME_SAVED, nullptr);
    }
    catch (Error const &er) {
        LOG_RES_WARNING("Error saving game session to '%s':\n")
                << savePath << er.asText();
    }
}
void GameSession::load(String const &saveName) {
    String const savePath = d->userSavePath(saveName);
    LOG_MSG("Loading game from \"%s\"...") << savePath;
    d->loadSaved(savePath);
    P_SetMessage(&players[CONSOLEPLAYER], "Game loaded");
}
void GameSession::copySaved(String const &destName, String const &sourceName) {
    AbstractSession::copySaved(d->userSavePath(destName), d->userSavePath(sourceName));
    LOG_MSG("Copied savegame \"%s\" to \"%s\"") << sourceName << destName;
}
void GameSession::removeSaved(String const &saveName) {
    AbstractSession::removeSaved(d->userSavePath(saveName));
}
String GameSession::savedUserDescription(String const &saveName) {
    String const savePath = d->userSavePath(saveName);
    if (auto const *saved = App::rootFolder().tryLocate<GameStateFolder>(savePath)) {
        return saved->metadata().gets("userDescription", "");
    }
    return ""; // Not found.
}
acs::System &GameSession::acsSystem() {
    return d->acscriptSys;
}
void GameSession::consoleRegister()  // static
{
    static dint        gsvRuleSkill = 0;
    static char const *gsvEpisode = "";
    static uri_s      *gsvMap = nullptr;
    static char const *gsvHub = "";
#define READONLYCVAR  (CVF_READ_ONLY | CVF_NO_MAX | CVF_NO_MIN | CVF_NO_ARCHIVE)
    C_VAR_INT    ("game-skill",     &gsvRuleSkill,  READONLYCVAR, 0, 0);
    C_VAR_CHARPTR("map-episode",    &gsvEpisode,    READONLYCVAR, 0, 0);
    C_VAR_CHARPTR("map-hub",        &gsvHub,        READONLYCVAR, 0, 0);
    C_VAR_URIPTR ("map-id",         &gsvMap,        READONLYCVAR, 0, 0);
#undef READONLYCVAR
}
} // namespace common
DENG_EXTERN_C unsigned int gfw_MapInfoFlags(void) {
    return gfw_Session()->mapInfo().getui(QStringLiteral("flags"));
}
String gfw_GameId() {
    if (auto *gp = DoomsdayApp::currentGameProfile()) {
        return gp->gameId();
    }
    return {};
}
const GameProfile *gfw_GameProfile() {
    return DoomsdayApp::currentGameProfile();
}
DENG2_PIMPL_NOREF(SaveSlots::Slot)
, DENG2_OBSERVES(GameStateFolder, MetadataChange) {
    String id;
    bool userWritable;
    String savePath;
    int menuWidgetId;
    GameStateFolder *session; // Not owned.
    SessionStatus status;
    Impl()
        : userWritable(true)
        , menuWidgetId(0 /*none*/)
        , session     (0)
        , status      (Unused)
    {}
    void updateStatus() {
        LOGDEV_XVERBOSE("Updating SaveSlot '%s' status", id);
        status = Unused;
        if (session) {
            status = Incompatible;
            if (!session->metadata().gets("gameIdentityKey", "").compareWithoutCase(gfw_GameId())) {
                status = Loadable; // It's good!
            }
        }
        updateMenuWidget("LoadGame");
        updateMenuWidget("SaveGame");
    }
    void updateMenuWidget(String const pageName) {
        if (!menuWidgetId) return;
        if (!Hu_MenuHasPage(pageName)) return; // Not initialized yet?
        Page &page = Hu_MenuPage(pageName);
        Widget *wi = page.tryFindWidget(menuWidgetId);
        if (!wi) {
            LOG_DEBUG("Failed locating menu widget with id ") << menuWidgetId;
            return;
        }
        LineEditWidget &edit = wi->as<LineEditWidget>();
        wi->setFlags(Widget::Disabled, pageName == "LoadGame"? SetFlags : UnsetFlags);
        if (status == Loadable) {
            edit.setText(session->metadata().gets("userDescription", ""));
            wi->setFlags(Widget::Disabled, UnsetFlags);
        }
        else {
            edit.setText("");
        }
        if (Hu_MenuIsActive() && Hu_MenuPagePtr() == &page) {
            Hu_MenuSetPage(&page, true);
        }
    }
    void gameStateFolderMetadataChanged(GameStateFolder &changed) {
        DENG2_ASSERT(&changed == session);
        DENG2_UNUSED(changed);
        updateStatus();
    }
};
SaveSlots::Slot::Slot(String id, bool userWritable, String saveName, int menuWidgetId)
    : d(new Impl) {
    d->id           = id;
    d->userWritable = userWritable;
    d->menuWidgetId = menuWidgetId;
    d->savePath     = SaveGames::savePath() / saveName;
    if (d->savePath.fileNameExtension().isEmpty()) {
        d->savePath += ".save";
    }
    setGameStateFolder(App::rootFolder().tryLocate<GameStateFolder>(d->savePath));
}
SaveSlots::Slot::SessionStatus SaveSlots::Slot::sessionStatus() const {
    return d->status;
}
bool SaveSlots::Slot::isUserWritable() const {
    return d->userWritable;
}
String const &SaveSlots::Slot::id() const {
    return d->id;
}
String const &SaveSlots::Slot::savePath() const {
    return d->savePath;
}
void SaveSlots::Slot::bindSaveName(String newName) {
    String newPath = SaveGames::savePath() / newName;
    if (newPath.fileNameExtension().isEmpty()) {
        newPath += ".save";
    }
    if (d->savePath != newPath) {
        d->savePath = newPath;
        setGameStateFolder(App::rootFolder().tryLocate<GameStateFolder>(d->savePath));
    }
}
void SaveSlots::Slot::setGameStateFolder(GameStateFolder *newSession) {
    if (d->session == newSession) return;
    if (d->session) {
        d->session->audienceForMetadataChange() -= d;
    }
    d->session = newSession;
    d->updateStatus();
    if (d->session) {
        d->session->audienceForMetadataChange() += d;
    }
#if !defined DENG_DEBUG // Always
    if (isUserWritable())
#endif
    {
        String statusText;
        if (d->session) {
            statusText = String("associated with \"%1\"").arg(d->session->path());
        }
        else {
            statusText = "unused";
        }
        LOG_VERBOSE("Save slot '%s' now %s") << d->id << statusText;
    }
}
void SaveSlots::Slot::updateStatus() {
    d->updateStatus();
}
DENG2_PIMPL(SaveSlots)
, DENG2_OBSERVES(FileIndex, Addition)
, DENG2_OBSERVES(FileIndex, Removal) {
    typedef std::map<String, Slot *> Slots;
    typedef std::pair<String, Slot *> SlotItem;
    Slots sslots;
    LoopCallback mainCall;
    Impl(Public *i) : Base(i) {
        SaveGames::get().saveIndex().audienceForAddition() += this;
        SaveGames::get().saveIndex().audienceForRemoval()  += this;
    }
    ~Impl() {
        DENG2_FOR_EACH(Slots, i, sslots) { delete i->second; }
    }
    SaveSlot *slotById(String const &id) {
        Slots::const_iterator found = sslots.find(id);
        if (found != sslots.end()) {
            return found->second;
        }
        return 0; // Not found.
    }
    SaveSlot *slotBySavePath(String path) {
        if (!path.isEmpty()) {
            if (path.fileNameExtension().isEmpty()) {
                path += ".save";
            }
            DENG2_FOR_EACH_CONST(Slots, i, sslots) {
                if (!i->second->savePath().compareWithoutCase(path)) {
                    return i->second;
                }
            }
        }
        return nullptr; // Not found.
    }
    void fileAdded(File const &saveFolder, FileIndex const &) {
        mainCall.enqueue([this, &saveFolder] () {
            if (SaveSlot *sslot = slotBySavePath(saveFolder.path())) {
                sslot->setGameStateFolder(const_cast<GameStateFolder *>(&saveFolder.as<GameStateFolder>()));
            }
        });
    }
    void fileRemoved(File const &saveFolder, FileIndex const &) {
        DENG2_FOR_EACH_CONST(Slots, i, sslots) {
            SaveSlot *sslot = i->second;
            if (sslot->savePath() == saveFolder.path()) {
                sslot->setGameStateFolder(nullptr);
            }
        }
    }
    void setAllIndexedSaves() {
        auto const &index = SaveGames::get().saveIndex();
        foreach (File *file, index.files()) {
            fileAdded(*file, index);
        }
    }
};
SaveSlots::SaveSlots() : d(new Impl(this))
{}
void SaveSlots::add(String const &id, bool userWritable, String const &saveName, int menuWidgetId) {
    if (d->slotById(id)) return;
    d->sslots.insert(Impl::SlotItem(id, new Slot(id, userWritable, saveName, menuWidgetId)));
}
int SaveSlots::count() const {
    return int(d->sslots.size());
}
bool SaveSlots::has(String const &id) const {
    return d->slotById(id) != 0;
}
SaveSlots::Slot &SaveSlots::slot(String const &id) const {
    if (SaveSlot *sslot = d->slotById(id)) {
        return *sslot;
    }
    throw MissingSlotError("SaveSlots::slot", "Invalid slot id '" + id + "'");
}
SaveSlots::Slot *SaveSlots::slotBySaveName(String const &name) const {
    return d->slotBySavePath(SaveGames::savePath() / name);
}
SaveSlots::Slot *SaveSlots::slotBySavedUserDescription(String const &description) const {
    if (!description.isEmpty()) {
        DENG2_FOR_EACH_CONST(Impl::Slots, i, d->sslots) {
            if (!gfw_Session()->savedUserDescription(i->second->saveName())
                                      .compareWithoutCase(description)) {
                return i->second;
            }
        }
    }
    return 0; // Not found.
}
SaveSlots::Slot *SaveSlots::slotByUserInput(String const &str) const {
    if (Slot *sslot = slotBySavedUserDescription(str)) {
        return sslot;
    }
    if (Slot *sslot = slotBySaveName(str)) {
        return sslot;
    }
    String id = str;
    if (!id.compareWithoutCase("last") || !id.compareWithoutCase("<last>")) {
        id = String::number(Con_GetInteger("game-save-last-slot"));
    }
    else if (!id.compareWithoutCase("quick") || !id.compareWithoutCase("<quick>")) {
        id = String::number(Con_GetInteger("game-save-quick-slot"));
    }
    return d->slotById(id);
}
void SaveSlots::updateAll() {
    d->setAllIndexedSaves();
    DENG2_FOR_EACH(Impl::Slots, i, d->sslots) {
        i->second->updateStatus();
    }
}
namespace {
int cvarLastSlot;  ///< @c -1= Not yet loaded/saved in this game session.
int cvarQuickSlot; ///< @c -1= Not yet chosen/determined.
}
void SaveSlots::consoleRegister() // static
{
    cvarLastSlot  = -1;
    cvarQuickSlot = -1;
    C_VAR_INT("game-save-last-slot",    &cvarLastSlot,  CVF_NO_MIN|CVF_NO_MAX|CVF_NO_ARCHIVE|CVF_READ_ONLY, -1, 0);
    C_VAR_INT("game-save-quick-slot",   &cvarQuickSlot, CVF_NO_MAX|CVF_NO_ARCHIVE, -1, 0);
}
