﻿/** @file sv_frame.cpp  Frame Generation and Transmission.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "server/sv_frame.h"
#include "def_main.h"
#include "sys_system.h"
#include "network/net_main.h"
#include "server/sv_pool.h"
#include "world/p_players.h"
#include <de/LogBuffer>
#include <cmath>
/** @file sv_infine.cpp  Server-side InFine.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "server/sv_infine.h"
#include "network/net_main.h"
#include "network/net_msg.h"
#include "ui/infine/finale.h"
/** @file sv_main.cpp  Network server.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_SERVER
#include "de_base.h"
#include "server/sv_def.h"
#include <cmath>
#include <de/stringarray.h>
#include <de/timer.h>
#include <de/ArrayValue>
#include <de/NumberValue>
#include <de/LogBuffer>
#include <doomsday/console/exec.h>
#include <doomsday/filesys/fs_main.h>
#include <doomsday/filesys/wad.h>
#include <doomsday/world/MaterialArchive>
#include "dd_main.h"
#include "def_main.h"
#include "network/net_main.h"
#include "network/net_buf.h"
#include "network/net_event.h"
#include "world/map.h"
#include "world/p_players.h"
#include "api_server.h"
#include "serversystem.h"
#include "server/sv_def.h"
#include "server/sv_pool.h"
/** @file sv_missile.cpp  Delta Pool Missile Record.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "server/sv_missile.h"
#include "server/sv_pool.h"
/**
 * Returns a pointer to the referenced hash chain.
 */
/** @file sv_pool.cpp  Delta Pools.
 *
 * Delta Pools use PU_MAP, which means all the memory allocated for them
 * is deallocated when the map changes. Sv_InitPools() is called in
 * R_SetupMap() to clear out all the old data.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "server/sv_pool.h"
#include <cmath>
#include <de/mathutil.h>
#include <de/timer.h>
#include <de/vector1.h>
#include <de/LogBuffer>
#include "def_main.h"  // Def_SameStateSequence
#include "network/net_main.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/thinkers.h"
#include "Sector"
/** @file sv_sound.cpp  Serverside Sound Management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "server/sv_sound.h"
#include <de/LogBuffer>
#include "network/net_main.h"
#include "world/p_players.h"
#include "server/sv_pool.h"
using namespace de;
#define BWR_ADJUST_TICS     (TICSPERSEC / 2)
#define MINIMUM_FRAME_SIZE  1800 // bytes
#define MAX_FIRST_FRAME_SIZE    64000
#define FRAME_SIZE_FACTOR   13
#define FIXED8_8(x)         (((x)*256) >> 16)
#define FIXED10_6(x)        (((x)*64) >> 16)
#define CLAMPED_CHAR(x)     ((x)>127? 127 : (x)<-128? -128 : (x))
#define MOM_FAST_LIMIT      (127)
void Sv_SendFrame(dint playerNumber);
dint allowFrames;
dint frameInterval = 1;  ///< Skip every second frame by default (17.5fps)
#ifdef DENG2_DEBUG
static dint byteCounts[256];
static dint totalFrameCount;
#endif
static dint lastTransmitTic;
/**
 * Send all the relevant information to each client.
 */
void Sv_TransmitFrame() {
    if (!::allowFrames || ::isClient || Sys_IsShuttingDown()) {
        return;
    }
    if (!::netGame) {
        dint i = 0;
        for (; i < DDMAXPLAYERS; ++i) {
            if (Sv_IsFrameTarget(i))
                break;
        }
        if (i == DDMAXPLAYERS) return;  // Nobody is a frame target.
    }
    if (SECONDS_TO_TICKS(::gameTime) == ::lastTransmitTic) {
        return;
    }
    ::lastTransmitTic = SECONDS_TO_TICKS(::gameTime);
    LOG_AS("Sv_TransmitFrame");
    Sv_GenerateFrameDeltas();
    dint const numInGame = Sv_GetNumPlayers();
    dint pCount = 0;
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        auto &plr = *DD_Player(i);
        if (!Sv_IsFrameTarget(i)) {
            continue;
        }
        pCount++;
        dint cTime = SECONDS_TO_TICKS(::gameTime);
        if (::frameInterval > 0 && numInGame > 1) {
            cTime += (pCount * ::frameInterval) / numInGame;
        }
        if (cTime <= plr.lastTransmit + ::frameInterval) {
            continue;
        }
        plr.lastTransmit = cTime;
        if (plr.ready) {
            Sv_SendFrame(i);
        }
        else {
            LOG_NET_XVERBOSE("NOT sending at tic %i to plr %i (ready:%b)",
                             ::lastTransmitTic << i << plr.ready);
        }
    }
}
/**
 * Shutdown routine for the server.
 */
void Sv_Shutdown() {
#ifdef DENG2_DEBUG
    if (::totalFrameCount > 0) {
        for (dint i = 0; i < 256; ++i) {
            LOGDEV_NET_NOTE("Byte %02x: %f")
                << i << (byteCounts[i] / dfloat( totalFrameCount ));
        }
    }
#endif
    Sv_ShutdownPools();
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WriteMobjDelta(const void* deltaPtr) {
    auto const *delta  = reinterpret_cast<mobjdelta_t const *>(deltaPtr);
    dt_mobj_t const *d = &delta->mo;
    dint df            = delta->delta.flags;
    byte moreFlags = 0;
    if (fabs(d->mom[MX]) >= MOM_FAST_LIMIT ||
       fabs(d->mom[MY]) >= MOM_FAST_LIMIT ||
       fabs(d->mom[MZ]) >= MOM_FAST_LIMIT) {
        df |= MDF_MORE_FLAGS;
        moreFlags |= MDFE_FAST_MOM;
    }
    if (df & MDFC_TRANSLUCENCY) {
        df |= MDF_MORE_FLAGS;
        moreFlags |= MDFE_TRANSLUCENCY;
    }
    if (df & MDFC_FADETARGET) {
        df |= MDF_MORE_FLAGS;
        moreFlags |= MDFE_FADETARGET;
    }
    if (df & MDFC_ON_FLOOR) {
        df |= MDF_MORE_FLAGS;
        moreFlags |= MDFE_Z_FLOOR;
    }
    if (df & MDFC_TYPE) {
        df |= MDF_MORE_FLAGS;
        moreFlags |= MDFE_TYPE;
    }
    if (d->selector & ~DDMOBJ_SELECTOR_MASK)
        df |= MDF_SELSPEC;
    if (!d->state) {
        df &= ~MDF_STATE;
    }
    /*
    if (df & MDF_ORIGIN_Z) {
        if (d->pos[VZ] == DDMINFLOAT || d->pos[VZ] == DDMAXFLOAT) {
            df &= ~MDF_ORIGIN_Z;
            df |= MDF_MORE_FLAGS;
            moreFlags |= (d->pos[VZ] == DDMINFLOAT ? MDFE_Z_FLOOR : MDFE_Z_CEILING);
        }
    }
    */
    DENG2_ASSERT(!(df & MDFC_NULL));     // don't write NULL deltas
    DENG2_ASSERT((df & 0xffff) != 0);    // don't write empty deltas
    Writer_WriteUInt16(::msgWriter, delta->delta.id);
    Writer_WriteUInt16(::msgWriter, df & 0xffff);
    if (df & MDF_MORE_FLAGS) {
        Writer_WriteByte(::msgWriter, moreFlags);
    }
    if (df & MDF_ORIGIN_X) {
        fixed_t vx = FLT2FIX(d->origin[VX]);
        Writer_WriteInt16(::msgWriter, vx >> FRACBITS);
        Writer_WriteByte(::msgWriter, vx >> 8);
    }
    if (df & MDF_ORIGIN_Y) {
        fixed_t vy = FLT2FIX(d->origin[VY]);
        Writer_WriteInt16(::msgWriter, vy >> FRACBITS);
        Writer_WriteByte(::msgWriter, vy >> 8);
    }
    if (df & MDF_ORIGIN_Z) {
        fixed_t vz = FLT2FIX(d->origin[VZ]);
        Writer_WriteInt16(::msgWriter, vz >> FRACBITS);
        Writer_WriteByte(::msgWriter, vz >> 8);
        Writer_WriteFloat(::msgWriter, d->floorZ);
        Writer_WriteFloat(::msgWriter, d->ceilingZ);
    }
    if (df & MDF_MOM_X) {
        fixed_t mx = FLT2FIX(d->mom[MX]);
        Writer_WriteInt16(::msgWriter, moreFlags & MDFE_FAST_MOM ? FIXED10_6(mx) : FIXED8_8(mx));
    }
    if (df & MDF_MOM_Y) {
        fixed_t my = FLT2FIX(d->mom[MY]);
        Writer_WriteInt16(::msgWriter, moreFlags & MDFE_FAST_MOM ? FIXED10_6(my) : FIXED8_8(my));
    }
    if (df & MDF_MOM_Z) {
        fixed_t mz = FLT2FIX(d->mom[MZ]);
        Writer_WriteInt16(::msgWriter, moreFlags & MDFE_FAST_MOM ? FIXED10_6(mz) : FIXED8_8(mz));
    }
    if (df & MDF_ANGLE)
        Writer_WriteInt16(::msgWriter, d->angle >> 16);
    if (df & MDF_SELECTOR)
        Writer_WritePackedUInt16(::msgWriter, d->selector);
    if (df & MDF_SELSPEC)
        Writer_WriteByte(::msgWriter, d->selector >> 24);
    if (df & MDF_STATE) {
        DENG2_ASSERT(d->state != 0);
        Writer_WritePackedUInt16(::msgWriter, ::runtimeDefs.states.indexOf(d->state));
    }
    if (df & MDF_FLAGS) {
        Writer_WriteUInt32(::msgWriter, d->ddFlags & DDMF_PACK_MASK);
        Writer_WriteUInt32(::msgWriter, d->flags);
        Writer_WriteUInt32(::msgWriter, d->flags2);
        Writer_WriteUInt32(::msgWriter, d->flags3);
    }
    if (df & MDF_HEALTH)
        Writer_WriteInt32(::msgWriter, d->health);
    if (df & MDF_RADIUS)
        Writer_WriteFloat(::msgWriter, d->radius);
    if (df & MDF_HEIGHT)
        Writer_WriteFloat(::msgWriter, d->height);
    if (df & MDF_FLOORCLIP)
        Writer_WriteFloat(::msgWriter, d->floorClip);
    if (df & MDFC_TRANSLUCENCY)
        Writer_WriteByte(::msgWriter, d->translucency);
    if (df & MDFC_FADETARGET)
        Writer_WriteByte(::msgWriter, byte( d->visTarget + 1 ));
    if (df & MDFC_TYPE)
        Writer_WriteInt32(::msgWriter, d->type);
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WritePlayerDelta(void const *deltaPtr) {
    auto const *delta    = reinterpret_cast<playerdelta_t const *>(deltaPtr);
    dt_player_t const *d = &delta->player;
    dint df              = delta->delta.flags;
    Writer_WriteByte(::msgWriter, delta->delta.id | (df >> 8));
    Writer_WriteByte(::msgWriter, df & 0xff);
    if (df & PDF_MOBJ)
        Writer_WriteUInt16(::msgWriter, d->mobj);
    if (df & PDF_FORWARDMOVE)
        Writer_WriteByte(::msgWriter, d->forwardMove);
    if (df & PDF_SIDEMOVE)
        Writer_WriteByte(::msgWriter, d->sideMove);
    /*if (df & PDF_ANGLE)
        Writer_WriteByte(::msgWriter, d->angle >> 24);*/
    if (df & PDF_TURNDELTA)
        Writer_WriteByte(::msgWriter, (d->turnDelta * 16) >> 24);
    if (df & PDF_FRICTION)
        Writer_WriteByte(::msgWriter, FLT2FIX(d->friction) >> 8);
    if (df & PDF_EXTRALIGHT) {
        dint const cmap = de::clamp(0, d->fixedColorMap, 7);
        Writer_WriteByte(::msgWriter, cmap | (d->extraLight & 0xf8));
    }
    if (df & PDF_FILTER) {
        Writer_WriteUInt32(::msgWriter, d->filter);
        LOGDEV_NET_XVERBOSE_DEBUGONLY("Sv_WritePlayerDelta: Plr %i, filter %08x", delta->delta.id << d->filter);
    }
    if (df & PDF_PSPRITES)       // Only set if there's something to write.
    {
        for (dint i = 0; i < 2; ++i) {
            ddpsprite_t const &psp = d->psp[i];
            dint const flags       = df >> (16 + i * 8);
            Writer_WriteByte(::msgWriter, flags);
            if (flags & PSDF_STATEPTR) {
                Writer_WritePackedUInt16(::msgWriter, psp.statePtr ? (::runtimeDefs.states.indexOf(psp.statePtr) + 1) : 0);
            }
            /*if (flags & PSDF_LIGHT)
            {
                dint const light = de::clamp(0, psp.light * 255, 255);
                Writer_WriteByte(::msgWriter, light);
            }*/
            if (flags & PSDF_ALPHA) {
                dint const alpha = de::clamp(0.f, psp.alpha * 255, 255.f);
                Writer_WriteByte(::msgWriter, alpha);
            }
            if (flags & PSDF_STATE) {
                Writer_WriteByte(::msgWriter, psp.state);
            }
            if (flags & PSDF_OFFSET) {
                Writer_WriteByte(::msgWriter, CLAMPED_CHAR(psp.offset[VX] / 2));
                Writer_WriteByte(::msgWriter, CLAMPED_CHAR(psp.offset[VY] / 2));
            }
        }
    }
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WriteSectorDelta(void const *deltaPtr) {
    auto const *delta    = reinterpret_cast<sectordelta_t const *>(deltaPtr);
    dt_sector_t const *d = &delta->sector;
    dint              df = delta->delta.flags;
    byte floorSpd = 0;
    if (df & SDF_FLOOR_SPEED) {
        dint const spd = FLT2FIX(fabs(d->planes[PLN_FLOOR].speed));
        floorSpd = spd >> 15;
        if (!floorSpd) {
            df |= SDF_FLOOR_SPEED_44;
            floorSpd = spd >> 12;
        }
    }
    byte ceilSpd = 0;
    if (df & SDF_CEILING_SPEED) {
        dint const spd = FLT2FIX(fabs(d->planes[PLN_CEILING].speed));
        ceilSpd = spd >> 15;
        if (!ceilSpd) {
            df |= SDF_CEILING_SPEED_44;
            ceilSpd = spd >> 12;
        }
    }
    Writer_WriteUInt16(::msgWriter, delta->delta.id);
    Writer_WritePackedUInt32(::msgWriter, df);
    if (df & SDF_FLOOR_MATERIAL)
        Writer_WritePackedUInt16(::msgWriter, Sv_IdForMaterial(d->planes[PLN_FLOOR].surface.material));
    if (df & SDF_CEILING_MATERIAL)
        Writer_WritePackedUInt16(::msgWriter, Sv_IdForMaterial(d->planes[PLN_CEILING].surface.material));
    if (df & SDF_LIGHT) {
        auto lightlevel = dint( 255.0f * d->lightLevel );
        lightlevel = (lightlevel < 0 ? 0 : lightlevel > 255 ? 255 : lightlevel);
        Writer_WriteByte(::msgWriter, byte( lightlevel ));
    }
    if (df & SDF_FLOOR_HEIGHT) {
        Writer_WriteInt16(::msgWriter, FLT2FIX(d->planes[PLN_FLOOR].height) >> 16);
    }
    if (df & SDF_CEILING_HEIGHT) {
        LOGDEV_NET_XVERBOSE_DEBUGONLY("Sv_WriteSectorDelta: (%i) Absolute ceiling height=%f",
                                     delta->delta.id << d->planes[PLN_CEILING].height);
        Writer_WriteInt16(::msgWriter, FLT2FIX(d->planes[PLN_CEILING].height) >> 16);
    }
    if (df & SDF_FLOOR_TARGET)
        Writer_WriteInt16(::msgWriter, FLT2FIX(d->planes[PLN_FLOOR].target) >> 16);
    if (df & SDF_FLOOR_SPEED)    // 7.1/4.4 fixed-point
        Writer_WriteByte(::msgWriter, floorSpd);
    if (df & SDF_CEILING_TARGET)
        Writer_WriteInt16(::msgWriter, FLT2FIX(d->planes[PLN_CEILING].target) >> 16);
    if (df & SDF_CEILING_SPEED)  // 7.1/4.4 fixed-point
        Writer_WriteByte(::msgWriter, ceilSpd);
    if (df & SDF_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->rgb[0] ));
    if (df & SDF_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->rgb[1] ));
    if (df & SDF_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->rgb[2] ));
    if (df & SDF_FLOOR_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_FLOOR].surface.rgba[0] ));
    if (df & SDF_FLOOR_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_FLOOR].surface.rgba[1] ));
    if (df & SDF_FLOOR_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_FLOOR].surface.rgba[2] ));
    if (df & SDF_CEIL_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_CEILING].surface.rgba[0] ));
    if (df & SDF_CEIL_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_CEILING].surface.rgba[1] ));
    if (df & SDF_CEIL_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->planes[PLN_CEILING].surface.rgba[2] ));
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WriteSideDelta(void const *deltaPtr) {
    auto const *delta  = (sidedelta_t const *) deltaPtr;
    dt_side_t const *d = &delta->side;
    dint            df = delta->delta.flags;
    Writer_WriteUInt16(::msgWriter, delta->delta.id);
    Writer_WritePackedUInt32(::msgWriter, df);
    if (df & SIDF_TOP_MATERIAL)
        Writer_WritePackedUInt16(::msgWriter, Sv_IdForMaterial(d->top.material));
    if (df & SIDF_MID_MATERIAL)
        Writer_WritePackedUInt16(::msgWriter, Sv_IdForMaterial(d->middle.material));
    if (df & SIDF_BOTTOM_MATERIAL)
        Writer_WritePackedUInt16(::msgWriter, Sv_IdForMaterial(d->bottom.material));
    if (df & SIDF_LINE_FLAGS)
        Writer_WriteByte(::msgWriter, d->lineFlags);
    if (df & SIDF_TOP_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->top.rgba[0] ));
    if (df & SIDF_TOP_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->top.rgba[1] ));
    if (df & SIDF_TOP_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->top.rgba[2] ));
    if (df & SIDF_MID_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->middle.rgba[0] ));
    if (df & SIDF_MID_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->middle.rgba[1] ));
    if (df & SIDF_MID_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->middle.rgba[2] ));
    if (df & SIDF_MID_COLOR_ALPHA)
        Writer_WriteByte(::msgWriter, byte( 255 * d->middle.rgba[3] ));
    if (df & SIDF_BOTTOM_COLOR_RED)
        Writer_WriteByte(::msgWriter, byte( 255 * d->bottom.rgba[0] ));
    if (df & SIDF_BOTTOM_COLOR_GREEN)
        Writer_WriteByte(::msgWriter, byte( 255 * d->bottom.rgba[1] ));
    if (df & SIDF_BOTTOM_COLOR_BLUE)
        Writer_WriteByte(::msgWriter, byte( 255 * d->bottom.rgba[2] ));
    if (df & SIDF_MID_BLENDMODE)
        Writer_WriteInt32(::msgWriter, d->middle.blendMode);
    if (df & SIDF_FLAGS)
        Writer_WriteByte(::msgWriter, d->flags);
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WritePolyDelta(void const *deltaPtr) {
    auto const  *delta = (polydelta_t const *) deltaPtr;
    dt_poly_t const *d = &delta->po;
    dint            df = delta->delta.flags;
    if (d->destAngle == (unsigned) -1) {
        df |= PODF_PERPETUAL_ROTATE;
        df &= ~PODF_DEST_ANGLE;
    }
    Writer_WritePackedUInt16(::msgWriter, delta->delta.id);
    Writer_WriteByte(::msgWriter, df & 0xff);
    if (df & PODF_DEST_X)
        Writer_WriteFloat(::msgWriter, d->dest[VX]);
    if (df & PODF_DEST_Y)
        Writer_WriteFloat(::msgWriter, d->dest[VY]);
    if (df & PODF_SPEED)
        Writer_WriteFloat(::msgWriter, d->speed);
    if (df & PODF_DEST_ANGLE)
        Writer_WriteInt16(::msgWriter, d->destAngle >> 16);
    if (df & PODF_ANGSPEED)
        Writer_WriteInt16(::msgWriter, d->angleSpeed >> 16);
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WriteSoundDelta(void const *deltaPtr) {
    auto const *delta = (sounddelta_t const *) deltaPtr;
    dint           df = delta->delta.flags;
    Writer_WriteUInt16(::msgWriter, delta->delta.id);
    Writer_WriteByte(::msgWriter, df & 0xff);
    switch (delta->delta.type) {
    case DT_MOBJ_SOUND:
    case DT_SECTOR_SOUND:
    case DT_SIDE_SOUND:
    case DT_POLY_SOUND:
        Writer_WriteUInt16(::msgWriter, delta->sound);
        break;
    default: break;
    }
    if (df & SNDDF_VOLUME) {
        if (delta->volume > 1) {
            Writer_WriteByte(::msgWriter, 255);
        }
        else if (delta->volume <= 0) {
            Writer_WriteByte(::msgWriter, 0);
        }
        else {
            Writer_WriteByte(::msgWriter, delta->volume * 127 + 0.5f);
        }
    }
}
/**
 * Write the type and possibly the set number (for Unacked deltas).
 */
void Sv_WriteDeltaHeader(byte type, delta_t const *delta) {
#ifdef DENG2_DEBUG
    if (type >= NUM_DELTA_TYPES) {
        App_Error("Sv_WriteDeltaHeader: Invalid delta type %i.\n", type);
    }
#endif
    DENG2_ASSERT(delta->state != DELTA_UNACKED);
    if (delta->state == DELTA_UNACKED) {
        DENG2_ASSERT(!"Unacked");
        type |= DT_RESENT;
    }
    Writer_WriteByte(::msgWriter, type);
    if (type & DT_RESENT) {
        Writer_WriteByte(::msgWriter, delta->set);
        Writer_WriteByte(::msgWriter, delta->resend);
    }
}
/**
 * The delta is written to the message buffer.
 */
void Sv_WriteDelta(delta_t const *delta) {
    DENG2_ASSERT(delta);
#ifdef _NETDEBUG
    dint const lengthOffset = Msg_Offset();
    Msg_WriteLong(0);
#endif
    if (delta->type == DT_MOBJ) {
        if (delta->flags & MDFC_NULL) {
            Sv_WriteDeltaHeader(DT_NULL_MOBJ, delta);
            Writer_WriteUInt16(::msgWriter, delta->id);
#ifdef _NETDEBUG
            goto writeDeltaLength;
#else
            return;
#endif
        }
    }
    Sv_WriteDeltaHeader(delta->type, delta);
    switch (delta->type) {
    case DT_MOBJ:   Sv_WriteMobjDelta(delta);   break;
    case DT_PLAYER: Sv_WritePlayerDelta(delta); break;
    case DT_SECTOR: Sv_WriteSectorDelta(delta); break;
    case DT_SIDE:   Sv_WriteSideDelta(delta);   break;
    case DT_POLY:   Sv_WritePolyDelta(delta);   break;
    case DT_SOUND:
    case DT_MOBJ_SOUND:
    case DT_SECTOR_SOUND:
    case DT_SIDE_SOUND:
    case DT_POLY_SOUND:
        Sv_WriteSoundDelta(delta);
        break;
    default: App_Error("Sv_WriteDelta: Unknown delta type %i.\n", delta->type);
    }
#ifdef _NETDEBUG
writeDeltaLength:
    dint endOffset = Msg_Offset();
    Msg_SetOffset(lengthOffset);
    Msg_WriteLong(endOffset - lengthOffset);
    Msg_SetOffset(endOffset);
#endif
}
/**
 * Returns an estimate for the maximum frame size appropriate for the client.
 * The bandwidth rating is updated whenever a frame is sent.
 */
dsize Sv_GetMaxFrameSize(dint playerNumber) {
    DENG2_UNUSED(playerNumber);
    DENG2_ASSERT(playerNumber >= 0 && playerNumber < DDMAXPLAYERS);
    dsize size = MINIMUM_FRAME_SIZE + FRAME_SIZE_FACTOR * 40 /* BWR_DEFAULT */;
    if (size > PROTOCOL_MAX_DATAGRAM_SIZE)
        size = PROTOCOL_MAX_DATAGRAM_SIZE;
    return size;
}
/**
 * @return A unique resend ID. Never returns zero.
 */
byte Sv_GetNewResendID(pool_t *pool) {
    DENG2_ASSERT(pool);
    byte id = pool->resendDealer;
    while (!++pool->resendDealer) {}
    return id;
}
/**
 * Send a sv_frame packet to the specified player. The amount of data sent
 * depends on the player's bandwidth rating.
 */
void Sv_SendFrame(dint plrNum) {
    pool_t *pool = Sv_GetPool(plrNum);
    if (!Sv_CheckBandwidth(plrNum)) {
        return;
    }
    Sv_RatePool(pool);
    DENG2_ASSERT(pool);
    pool->setDealer++;
    dsize maxFrameSize = Sv_GetMaxFrameSize(plrNum);
    if (pool->isFirst) {
        maxFrameSize = MAX_FIRST_FRAME_SIZE;
    }
    Msg_Begin(pool->isFirst ? PSV_FIRST_FRAME2 : PSV_FRAME2);
    Writer_WriteFloat(::msgWriter, ::gameTime);
    delta_t *delta;
    size_t lastStart;
    while ((delta = Sv_PoolQueueExtract(pool)) != nullptr &&
          (lastStart = Writer_Size(::msgWriter)) < maxFrameSize) {
        byte const oldResend = pool->resendDealer;
        if (delta->state == DELTA_UNACKED && !delta->resend) {
            delta->resend = Sv_GetNewResendID(pool);
        }
        Sv_WriteDelta(delta);
        if (Writer_Size(::msgWriter) > maxFrameSize) {
            /*
            if (::clients[plrNum].bwrAdjustTime <= 0) {
                ::clients[plrNum].bwrAdjustTime = BWR_ADJUST_TICS;
            }
            */
            Writer_SetPos(::msgWriter, lastStart);
            if (oldResend)
                pool->resendDealer = oldResend;
            break;
        }
        if (delta->state == DELTA_NEW) {
            delta->set       = pool->setDealer;
            delta->timeStamp = Sv_GetTimeStamp();
            delta->state     = DELTA_UNACKED;
        }
    }
    Msg_End();
    Net_SendBuffer(plrNum, 0);
    Sv_AckDeltaSet(plrNum, pool->setDealer, 0);
    pool->isFirst = false;
}
void Sv_Finale(finaleid_t id, dint flags, char const *script) {
    if (::isClient) return;
    dsize scriptLen = 0;
    if (script) {
        flags |= FINF_SCRIPT;
        scriptLen = qstrlen(script);
    }
    Msg_Begin(PSV_FINALE);
    Writer_WriteByte(::msgWriter, flags);
    Writer_WriteUInt32(::msgWriter, id); // serverside Id
    if (script) {
        Writer_WriteUInt32(::msgWriter, scriptLen);
        Writer_Write(::msgWriter, script, scriptLen);
    }
    Msg_End();
    Net_SendBuffer(NSP_BROADCAST, 0);
}
#define MAX_BANDWIDTH_RATING    100
#define WARP_LIMIT              300
void Sv_ClientCoords(dint playerNum);
dint netRemoteUser;  ///< The client who is currently logged in.
char *netPassword = (char *) "";  ///< Remote login password.
dint svMaxPlayers = DDMAXPLAYERS;
static world::MaterialArchive *materialDict;
#if 0
/**
 * @defgroup pathToStringFlags  Path To String Flags
 * @ingroup flags
 */
#define PTSF_QUOTED                     0x1  ///< Add double quotes around the path.
#define PTSF_TRANSFORM_EXCLUDE_PATH     0x2  ///< Exclude the path; e.g., c:/doom/myaddon.wad => myaddon.wad
#define PTSF_TRANSFORM_EXCLUDE_EXT      0x4  ///< Exclude the extension; e.g., c:/doom/myaddon.wad => c:/doom/myaddon
#define DEFAULT_PATHTOSTRINGFLAGS       (PTSF_QUOTED)
/**
 * @param files      List of files from which to compose the path string.
 * @param flags      @ref pathToStringFlags
 * @param delimiter  If not @c nullptr, path fragments in the resultant string
 * will be delimited by this.
 *
 * @return  New string containing a concatenated, possibly delimited set of all
 * file paths in the list.
 */
static String composeFilePathString(FS1::FileList &files, dint flags = DEFAULT_PATHTOSTRINGFLAGS,
    String const &delimiter = ";") {
    String result;
    DENG2_FOR_EACH_CONST(FS1::FileList, i, files) {
        File1 &file = (*i)->file();
        if (flags & PTSF_QUOTED)
            result.append('"');
        if (flags & PTSF_TRANSFORM_EXCLUDE_PATH) {
            if (flags & PTSF_TRANSFORM_EXCLUDE_EXT)
                result.append(file.name().fileNameWithoutExtension());
            else
                result.append(file.name());
        }
        else {
            String path = file.composePath();
            if (flags & PTSF_TRANSFORM_EXCLUDE_EXT) {
                result.append(path.fileNamePath() + '/' + path.fileNameWithoutExtension());
            }
            else {
                result.append(path);
            }
        }
        if (flags & PTSF_QUOTED)
            result.append('"');
        if (*i != files.last())
            result.append(delimiter);
    }
    return result;
}
static bool findCustomFilesPredicate(File1 &file, void * /*parameters*/)
{
    return file.hasCustom();
}
/**
 * Compiles a list of file names, separated by @a delimiter.
 */
static void composePWADFileList(char *outBuf, dsize outBufSize, char const *delimiter) {
    if (!outBuf || 0 == outBufSize) return;
    std::memset(outBuf, 0, outBufSize);
    FS1::FileList foundFiles;
    if (!App_FileSystem().findAll<de::Wad>(findCustomFilesPredicate, 0/*no params*/, foundFiles)) return;
    String str = composeFilePathString(foundFiles, PTSF_TRANSFORM_EXCLUDE_PATH, delimiter);
    QByteArray strUtf8 = str.toUtf8();
    strncpy(outBuf, strUtf8.constData(), outBufSize);
}
#endif
/*Record *Sv_InfoToRecord(serverinfo_t *info)
{
    DENG2_ASSERT(info);
    auto *rec = new Record;
    rec->addNumber ("port",  info->port);
    rec->addText   ("name",  info->name);
    rec->addText   ("info",  info->description);
    rec->addNumber ("ver",   info->version);
    rec->addText   ("game",  info->plugin);
    rec->addText   ("mode",  info->gameIdentityKey);
    rec->addText   ("setup", info->gameConfig);
    rec->addText   ("iwad",  info->iwad);
    rec->addNumber ("wcrc",  info->loadedFilesCRC);
    rec->addText   ("pwads", info->pwads);
    rec->addText   ("map",   info->map);
    rec->addNumber ("nump",  info->numPlayers);
    rec->addNumber ("maxp",  info->maxPlayers);
    rec->addBoolean("open",  info->canJoin);
    rec->addText   ("plrn",  info->clientNames);
    ArrayValue &data = rec->addArray("data").value<ArrayValue>();
    for (duint i = 0; i < sizeof(info->data) / sizeof(info->data[0]); ++i) {
        data << NumberValue(info->data[i]);
    }
    return rec;
}*/
/*
 * @return  Length of the string.
 */
/*dsize Sv_InfoToString(serverinfo_t *info, ddstring_t *msg)
{
    DENG2_ASSERT(info && msg);
    Str_Appendf(msg, "port:%i\n",  info->port);
    Str_Appendf(msg, "name:%s\n",  info->name);
    Str_Appendf(msg, "info:%s\n",  info->description);
    Str_Appendf(msg, "ver:%i\n",   info->version);
    Str_Appendf(msg, "game:%s\n",  info->plugin);
    Str_Appendf(msg, "mode:%s\n",  info->gameIdentityKey);
    Str_Appendf(msg, "setup:%s\n", info->gameConfig);
    Str_Appendf(msg, "iwad:%s\n",  info->iwad);
    Str_Appendf(msg, "wcrc:%i\n",  info->loadedFilesCRC);
    Str_Appendf(msg, "pwads:%s\n", info->pwads);
    Str_Appendf(msg, "map:%s\n",   info->map);
    Str_Appendf(msg, "nump:%i\n",  info->numPlayers);
    Str_Appendf(msg, "maxp:%i\n",  info->maxPlayers);
    Str_Appendf(msg, "open:%i\n",  info->canJoin);
    Str_Appendf(msg, "plrn:%s\n",  info->clientNames);
    for (duint i = 0; i < sizeof(info->data) / sizeof(info->data[0]); ++i) {
        Str_Appendf(msg, "data%i:%x\n", i, info->data[i]);
    }
    return Str_Length(msg);
}*/
/**
 * @return  gametic - cmdtime.
 */
dint Sv_Latency(byte cmdtime) {
    return Net_TimeDelta(SECONDS_TO_TICKS(gameTime), cmdtime);
}
/**
 * For local players.
 * $unifiedangles
 */
/*
void Sv_FixLocalAngles(bool clearFixAnglesFlag) {
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        ddplayer_t &pl = players[i];
        if (!pl.inGame || !(pl.flags & DDPF_LOCAL))
            continue;
        if (::isDedicated && i == 0)
            continue;
        if (pl.flags & DDPF_FIXANGLES) {
            if (clearFixAnglesFlag) {
                pl.flags &= ~DDPF_FIXANGLES;
            }
            else {
                DENG2_ASSERT(pl.mo);
                pl.clAngle   = pl.mo->angle;
                pl.clLookDir = pl.lookDir;
            }
        }
    }
}
*/
void Sv_HandlePlayerInfoFromClient(ServerPlayer *sender) {
    DENG2_ASSERT(sender);
    LOG_AS("Sv_HandlePlayerInfoFromClient");
    DENG2_ASSERT(::netBuffer.player == DoomsdayApp::players().indexOf(sender));
    dint console = Reader_ReadByte(::msgReader); // ignored
    LOG_NET_VERBOSE("from=%i, console=%i") << ::netBuffer.player << console;
    console = ::netBuffer.player;
    char oldName[PLAYERNAMELEN]; strcpy(oldName, sender->name);
    dsize len = Reader_ReadUInt16(msgReader);
    len = de::min(dsize( PLAYERNAMELEN - 1), len);  // There is a maximum size.
    Reader_Read(::msgReader, sender->name, len);
    sender->name[len] = 0;
    LOG_NET_NOTE("Player %s renamed to %s") << oldName << sender->name;
    Net_SendPlayerInfo(console, DDSP_ALL_PLAYERS);
}
/**
 * Handles a server-specific network message. Assumes that Msg_BeginRead()
 * has already been called to begin reading the message.
 */
void Sv_HandlePacket() {
    LOG_AS("Sv_HandlePacket");
    dint const from  = ::netBuffer.player;
    DENG2_ASSERT(from >= 0 && from < DDMAXPLAYERS);
    player_t *sender = DD_Player(from);
    ddplayer_t *ddpl = &sender->publicData();
    switch (netBuffer.msg.type) {
    case PCL_HELLO:
    case PCL_HELLO2: {
        ident_t id = Reader_ReadUInt32(::msgReader);
        LOG_NET_XVERBOSE("Hello from client %i (%08X)", from << id);
        if (!ddpl->inGame && !sender->handshake) {
            dint i = 1;
            for ( ; i < DDMAXPLAYERS; ++i) {
                if (sender->isConnected() && sender->id == id) {
                    LOG_NET_WARNING("New client connection refused: duplicate ID (%08x)") << id;
                    LOGDEV_NET_WARNING("ID conflict from=%i, i=%i") << from << i;
                    N_TerminateClient(from);
                    break;
                }
            }
            if (i < DDMAXPLAYERS) break;  // Can't continue, refused!
        }
        sender->id = id;
        if (netBuffer.msg.type == PCL_HELLO2) {
            char buf[17]; Reader_Read(::msgReader, buf, 16);
            if (strnicmp(buf, App_CurrentGame().id().toUtf8().constData(), 16)) {
                LOG_NET_ERROR("Client's game ID is incompatible: %-.16s") << buf;
                N_TerminateClient(from);
                break;
            }
        }
        if (!ddpl->inGame && !sender->handshake) {
            sender->handshake = true;
            sender->publicData().inGame = true;
            gx.NetPlayerEvent(from, DDPE_ARRIVAL, 0);
            Sv_Handshake(from, true);
            sender->enterTime = Timer_RealSeconds();
        }
        else if (ddpl->inGame) {
            Sv_Handshake(from, false);
        }
        break; }
    case PKT_OK:
        sender->ready = true;
        LOG_NET_VERBOSE("OK (\"ready!\") from client %i (%08X)") << from << sender->id;
        if (sender->handshake) {
            sender->handshake = false;
            Msg_Begin(PSV_SYNC);
            Writer_WriteFloat(::msgWriter, ::gameTime);
            Msg_End();
            Net_SendBuffer(from, 0);
            Sv_SendText(from, SV_CONSOLE_PRINT_FLAGS, SV_WELCOME_STRING "\n");
        }
        break;
    case PKT_CHAT: {
        dint const msgfrom = Reader_ReadByte(::msgReader);
        dint const mask    = Reader_ReadUInt32(::msgReader);
        dsize len = Reader_ReadUInt16(::msgReader);
        auto *msg = (char *) M_Malloc(len + 1);
        Reader_Read(::msgReader, msg, len);
        msg[len] = 0;
        if (mask & 1) {
            Net_ShowChatMessage(msgfrom, msg);
            gx.NetPlayerEvent(msgfrom, DDPE_CHAT_MESSAGE, msg);
        }
        Net_WriteChatMessage(msgfrom, mask, msg);
        for (dint i = 1; i < DDMAXPLAYERS; ++i) {
            if (DD_Player(i)->publicData().inGame && (mask & (1 << i)) && i != from) {
                Net_SendBuffer(i, 0);
            }
        }
        M_Free(msg);
        break; }
    case PCL_FINALE_REQUEST: {
        finaleid_t const fid = Reader_ReadUInt32(::msgReader);
        duint16 const params = Reader_ReadUInt16(::msgReader);
        LOGDEV_NET_MSG("PCL_FINALE_REQUEST: fid=%i params=%i") << fid << params;
        if (params == 1) {
            FI_ScriptRequestSkip(fid);
        }
        break; }
    case PKT_PLAYER_INFO:
        Sv_HandlePlayerInfoFromClient(sender);
        break;
    default:
        LOGDEV_NET_ERROR("Invalid value: netBuffer.msg.type = %i") << ::netBuffer.msg.type;
        break;
    }
}
/**
 * Handles a login packet. If the password is OK and no other client
 * is current logged in, a response is sent.
 */
void Sv_Login(void) {
    char password[300];
    byte passLen = 0;
    if (netRemoteUser) {
        Sv_SendText(netBuffer.player, SV_CONSOLE_PRINT_FLAGS,
                    "Sv_Login: A client is already logged in.\n");
        return;
    }
    LOG_AS("Sv_Login");
    passLen = Reader_ReadByte(msgReader);
    de::zap(password);
    Reader_Read(msgReader, password, passLen);
    if (strcmp(password, netPassword)) {
        Sv_SendText(netBuffer.player, SV_CONSOLE_PRINT_FLAGS,
                    "Sv_Login: Invalid password.\n");
        return;
    }
    netRemoteUser = netBuffer.player;
    LOG_NET_NOTE("%s (client %i) logged in" ) << DD_Player(netRemoteUser)->name << netRemoteUser;
    Msg_Begin(PKT_LOGIN);
    Writer_WriteByte(msgWriter, true);        // Yes, you're logged in.
    Msg_End();
    Net_SendBuffer(netRemoteUser, 0);
}
/**
 * Executes the command in the message buffer.
 * Usually sent by Con_Send.
 */
void Sv_ExecuteCommand(void) {
    int flags;
    byte cmdSource;
    unsigned short len;
    dd_bool silent;
    char *cmd = 0;
    LOG_AS("Sv_ExecuteCommand");
    if (!netRemoteUser) {
        LOGDEV_NET_ERROR("Command received but no one's logged in!");
        return;
    }
    len = Reader_ReadUInt16(msgReader);
    silent = (len & 0x8000) != 0;
    len &= 0x7fff;
    switch (netBuffer.msg.type) {
    /*case PKT_COMMAND:
        cmdSource = CMDS_UNKNOWN; // unknown command source.
        break;*/
    case PKT_COMMAND2:
        flags = Reader_ReadUInt16(msgReader);
        DENG_UNUSED(flags);
        cmdSource = Reader_ReadByte(msgReader);
        break;
    default:
        DENG_ASSERT(!"Sv_ExecuteCommand: Not a command packet!");
        return;
    }
    cmd = (char *) M_Malloc(len + 1);
    Reader_Read(msgReader, cmd, len);
    cmd[len] = 0;
    Con_Execute(cmdSource, cmd, silent, true);
    M_Free(cmd);
}
/**
 * Server's packet handler.
 */
void Sv_GetPackets(void) {
    int netconsole;
    while (Net_GetPacket()) {
        Msg_BeginRead();
        switch (netBuffer.msg.type) {
        case PCL_GOODBYE:
            N_TerminateClient(netBuffer.player);
            break;
        case PKT_COORDS:
            Sv_ClientCoords(netBuffer.player);
            break;
        case PCL_ACK_SHAKE:
            netconsole = netBuffer.player;
            if (netconsole >= 0 && netconsole < DDMAXPLAYERS) {
                ServerPlayer *sender = DD_Player(netconsole);
                sender->shakePing = Timer_RealMilliseconds() - sender->shakePing;
                LOG_NET_MSG("Client %i ping at handshake: %i ms")
                        << netconsole << sender->shakePing;
            }
            break;
        case PCL_ACK_PLAYER_FIX: {
            player_t* plr = DD_Player(netBuffer.player);
            ddplayer_t* ddpl = &plr->publicData();
            fixcounters_t* acked = &ddpl->fixAcked;
            acked->angles = Reader_ReadInt32(msgReader);
            acked->origin = Reader_ReadInt32(msgReader);
            acked->mom = Reader_ReadInt32(msgReader);
            LOGDEV_NET_XVERBOSE_DEBUGONLY("PCL_ACK_PLAYER_FIX: (%i) Angles %i (%i), pos %i (%i), mom %i (%i)",
                        netBuffer.player << acked->angles << ddpl->fixCounter.angles <<
                        acked->origin << ddpl->fixCounter.origin << acked->mom <<
                        ddpl->fixCounter.mom);
            break; }
        case PKT_PING:
            Net_PingResponse();
            break;
        case PCL_HELLO:
        case PCL_HELLO2:
        case PKT_OK:
        case PKT_CHAT:
        case PKT_PLAYER_INFO:
        case PCL_FINALE_REQUEST:
            Sv_HandlePacket();
            break;
        case PKT_LOGIN:
            Sv_Login();
            break;
        case PKT_COMMAND2:
            Sv_ExecuteCommand();
            break;
        default:
            if (netBuffer.msg.type >= PKT_GAME_MARKER) {
                gx.HandlePacket(netBuffer.player, netBuffer.msg.type,
                                netBuffer.msg.data, netBuffer.length);
            }
            break;
        }
        Msg_EndRead();
    }
}
/**
 * Assign a new console to the player. Returns true if successful.
 * Called by N_Update().
 */
dd_bool Sv_PlayerArrives(unsigned int nodeID, char const *name) {
    LOG_AS("Sv_PlayerArrives");
    LOG_NET_NOTE("'%s' has arrived") << name;
    for (int i = 1; i < DDMAXPLAYERS; ++i) {
        player_t *plr = DD_Player(i);
        if (!plr->isConnected()) {
            ddplayer_t *ddpl = &plr->publicData();
            plr->remoteUserId = nodeID;
            plr->lastTransmit = -1;
            plr->ready = false;
            plr->viewConsole = i;
            strncpy(plr->name, name, PLAYERNAMELEN);
            ddpl->fixAcked.angles =
                ddpl->fixAcked.origin =
                ddpl->fixAcked.mom = -1;
            de::zap(ddpl->filterColor);
            ddpl->flags &= ~DDPF_VIEW_FILTER;
            Sv_InitPoolForClient(i);
            Smoother_Clear(plr->smoother());
            LOG_NET_MSG("'%s' assigned to console %i (node:%u)") << plr->name << i << nodeID;
            plr->handshake = false;
            return true;
        }
    }
    return false;
}
/**
 * Remove the specified player from the game. Called by N_Update().
 */
void Sv_PlayerLeaves(unsigned int nodeID) {
    int                 plrNum = N_IdentifyPlayer(nodeID);
    dd_bool             wasInGame;
    player_t           *plr;
    if (plrNum == -1) return; // Bogus?
    LOG_AS("Sv_PlayerLeaves");
    if (netRemoteUser == plrNum)
        netRemoteUser = 0;
    plr = DD_Player(plrNum);
    LOG_NET_NOTE("'%s' (console %i) has left, was connected for %.1f seconds")
            << plr->name << plrNum << (Timer_RealSeconds() - plr->enterTime);
    wasInGame = plr->publicData().inGame;
    plr->publicData().inGame = false;
    plr->remoteUserId = 0;
    plr->ready        = false;
    plr->handshake    = false;
    Sv_PlayerRemoved(plrNum);
    if (wasInGame) {
        gx.NetPlayerEvent(plrNum, DDPE_EXIT, NULL);
        Msg_Begin(PSV_PLAYER_EXIT);
        Writer_WriteByte(msgWriter, plrNum);
        Msg_End();
        Net_SendBuffer(NSP_BROADCAST, 0);
    }
    plr->id = 0;
}
/**
 * Compiles a list of all the defined thing types. Indices in this list match those
 * in the @c mobjInfo array.
 *
 * @return  StringArray instance. Caller gets ownership.
 */
static StringArray *listThingTypeIDs() {
    StringArray *array = StringArray_New();
    for (dint i = 0; i < DED_Definitions()->things.size(); ++i) {
        StringArray_Append(array, DED_Definitions()->things[i].gets("id").toUtf8());
    }
    return array;
}
/**
 * Compiles a list of all the defined mobj states. Indices in this list match those
 * in the @c states array.
 *
 * @return  StringArray instance. Caller gets ownership.
 */
static StringArray *listStateIDs() {
    StringArray *array = StringArray_New();
    for (dint i = 0; i < DED_Definitions()->states.size(); ++i) {
        StringArray_Append(array, DED_Definitions()->states[i].gets("id").toUtf8());
    }
    return array;
}
/**
 * The player will be sent the introductory handshake packets.
 */
void Sv_Handshake(dint plrNum, dd_bool newPlayer) {
    LOG_AS("Sv_Handshake");
    LOG_NET_VERBOSE("Shaking hands with player %i (newPlayer:%b)") << plrNum << newPlayer;
    duint playersInGame = 0;
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        if (DD_Player(i)->isConnected())
            playersInGame |= 1 << i;
    }
    Msg_Begin(PSV_HANDSHAKE);
    Writer_WriteByte(msgWriter, SV_VERSION);
    Writer_WriteByte(msgWriter, plrNum);
    Writer_WriteUInt32(msgWriter, playersInGame);
    Writer_WriteFloat(msgWriter, gameTime);
    Msg_End();
    Net_SendBuffer(plrNum, 0);
    Msg_Begin(PSV_MATERIAL_ARCHIVE);
    materialDict->write(*msgWriter);
    Msg_End();
    Net_SendBuffer(plrNum, 0); {
        StringArray *ar = listThingTypeIDs();
        Msg_Begin(PSV_MOBJ_TYPE_ID_LIST);
        StringArray_Write(ar, msgWriter);
        Msg_End();
        Net_SendBuffer(plrNum, 0);
        StringArray_Delete(ar);
    } {
        StringArray *ar = listStateIDs();
        Msg_Begin(PSV_MOBJ_STATE_ID_LIST);
        StringArray_Write(ar, msgWriter);
        Msg_End();
        Net_SendBuffer(plrNum, 0);
        StringArray_Delete(ar);
    }
    if (newPlayer) {
        DD_Player(plrNum)->shakePing = Timer_RealMilliseconds();
    }
    gx.NetWorldEvent(DDWE_HANDSHAKE, plrNum, (void *) &newPlayer);
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        if (DD_Player(i)->isConnected()) {
            Net_SendPlayerInfo(i, plrNum);
        }
        if (newPlayer && i != 0 && i != plrNum && DD_Player(i)->isConnected()) {
            Net_SendPlayerInfo(plrNum, i);
        }
    }
    if (!newPlayer) {
        Sv_InitPoolForClient(plrNum);
    }
    DD_Player(plrNum)->publicData().flags |= DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM;
}
void Sv_StartNetGame(void) {
    int                 i;
    for (i = 0; i < DDMAXPLAYERS; ++i) {
        player_t   *plr = DD_Player(i);
        ddplayer_t *ddpl = &plr->publicData();
        ddpl->inGame = false;
        ddpl->flags &= ~DDPF_CAMERA;
        plr->remoteUserId = 0;
        plr->lastTransmit = -1;
        plr->ready = false;
        plr->enterTime = 0;
        plr->fov = 90;
        plr->viewConsole = -1;
        de::zap(plr->name);
        Smoother_Clear(plr->smoother());
    }
    gameTime = 0;
    firstNetUpdate = true;
    netRemoteUser = 0;
    consolePlayer = displayPlayer = 0;
    netGame = true;
    isServer = true;
    allowSending = true;
    materialDict = new world::MaterialArchive(false);
    materialDict->addWorldMaterials();
    LOGDEV_NET_XVERBOSE("Prepared material dictionary with %i materials",
                        materialDict->count());
}
void Sv_StopNetGame(void) {
    if (materialDict) {
        delete materialDict;
        materialDict = 0;
    }
}
unsigned int Sv_IdForMaterial(world::Material *mat) {
    DENG_ASSERT(materialDict);
    return materialDict->findUniqueSerialId(mat);
}
void Sv_SendText(int to, int con_flags, const char* text) {
    uint32_t len = MIN_OF(0xffff, strlen(text));
    Msg_Begin(PSV_CONSOLE_TEXT);
    Writer_WriteUInt32(msgWriter, con_flags & ~CPF_TRANSMIT);
    Writer_WriteUInt16(msgWriter, len);
    Writer_Write(msgWriter, text, len);
    Msg_End();
    Net_SendBuffer(to, 0);
}
/**
 * Asks a client to disconnect. Clients will immediately disconnect
 * after receiving the PSV_SERVER_CLOSE message.
 */
void Sv_Kick(int who) {
    if (!DD_Player(who)->isConnected())
        return;
    Sv_SendText(who, SV_CONSOLE_PRINT_FLAGS, "You were kicked out!\n");
    Msg_Begin(PSV_SERVER_CLOSE);
    Msg_End();
    Net_SendBuffer(who, 0);
}
/**
 * Sends player @a plrNum's position, momentum and/or angles override to all
 * clients.
 */
void Sv_SendPlayerFixes(int plrNum) {
    int         fixes = 0;
    player_t   *plr = DD_Player(plrNum);
    ddplayer_t *ddpl = &plr->publicData();
    if (!(ddpl->flags & (DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM))) {
        return;
    }
    LOG_AS("Sv_SendPlayerFixes");
    Msg_Begin(PSV_PLAYER_FIX);
    Writer_WriteByte(msgWriter, plrNum);
    if (ddpl->flags & DDPF_FIXANGLES) fixes |= 1;
    if (ddpl->flags & DDPF_FIXORIGIN) fixes |= 2;
    if (ddpl->flags & DDPF_FIXMOM) fixes |= 4;
    Writer_WriteUInt32(msgWriter, fixes);
    Writer_WriteUInt16(msgWriter, ddpl->mo->thinker.id);
    LOGDEV_NET_MSG("Fixing mobj %i of player %i") << ddpl->mo->thinker.id << plrNum;
    if (ddpl->flags & DDPF_FIXANGLES) {
        Writer_WriteInt32(msgWriter, ++ddpl->fixCounter.angles);
        Writer_WriteUInt32(msgWriter, ddpl->mo->angle);
        Writer_WriteFloat(msgWriter, ddpl->lookDir);
        LOGDEV_NET_MSG("Sent angles (%i): angle=%x lookdir=%.2f")
                << ddpl->fixCounter.angles << ddpl->mo->angle << ddpl->lookDir;
    }
    if (ddpl->flags & DDPF_FIXORIGIN) {
        Writer_WriteInt32(msgWriter, ++ddpl->fixCounter.origin);
        Writer_WriteFloat(msgWriter, ddpl->mo->origin[VX]);
        Writer_WriteFloat(msgWriter, ddpl->mo->origin[VY]);
        Writer_WriteFloat(msgWriter, ddpl->mo->origin[VZ]);
        LOGDEV_NET_MSG("Sent position (%i): %s")
                << ddpl->fixCounter.origin << Vector3d(ddpl->mo->origin).asText();
    }
    if (ddpl->flags & DDPF_FIXMOM) {
        Writer_WriteInt32(msgWriter, ++ddpl->fixCounter.mom);
        Writer_WriteFloat(msgWriter, ddpl->mo->mom[MX]);
        Writer_WriteFloat(msgWriter, ddpl->mo->mom[MY]);
        Writer_WriteFloat(msgWriter, ddpl->mo->mom[MZ]);
        LOGDEV_NET_MSG("Sent momentum (%i): %s")
                << ddpl->fixCounter.mom << Vector3d(ddpl->mo->mom).asText();
    }
    Msg_End();
    Net_SendBuffer(DDSP_ALL_PLAYERS, 0);
    ddpl->flags &= ~(DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM);
    LOGDEV_NET_VERBOSE("Cleared FIX flags of player %i") << plrNum;
    Smoother_Clear(DD_Player(plrNum)->smoother());
}
void Sv_Ticker(timespan_t ticLength) {
    int i;
    DENG_ASSERT(isDedicated);
    for (i = 0; i < DDMAXPLAYERS; ++i) {
        player_t *plr = DD_Player(i);
        if (!plr->publicData().inGame || !plr->publicData().mo)
            continue;
        if (plr->smoother()) {
            Smoother_Advance(plr->smoother(), ticLength);
        }
        if (DD_IsSharpTick()) {
            plr->publicData().lastAngle = plr->publicData().mo->angle;
        }
        Sv_SendPlayerFixes(i);
    }
}
/**
 * @return              The number of players in the game.
 */
int Sv_GetNumPlayers(void) {
    int i, count;
    if (isClient)
        return 1;
    for (i = count = 0; i < DDMAXPLAYERS; ++i) {
        if (DD_Player(i)->isInGame())
            count++;
    }
    return count;
}
/**
 * @return              The number of connected clients.
 */
int Sv_GetNumConnected(void) {
    int                 i, count = 0;
    if (isClient)
        return 1;
    for (i = isDedicated ? 1 : 0; i < DDMAXPLAYERS; ++i)
        if (DD_Player(i)->isConnected())
            count++;
    return count;
}
/**
 * The bandwidth rating is updated according to the status of the
 * player's send queue. Returns true if a new packet may be sent.
 *
 * @todo This functionality needs to be restored: servers can't simply output
 * an arbitrary amount of data to clients with no regard to the available
 * bandwidth.
 */
dd_bool Sv_CheckBandwidth(int /*playerNumber*/)
{
    return true;
    /*
    client_t           *client = &clients[playerNumber];
    uint                qSize = N_GetSendQueueSize(playerNumber);
    uint                limit = 400;
    return true;
    if (qSize > limit) {
        client->bandwidthRating -= 10;
    }
    if (qSize < limit / 20 && client->bwrAdjustTime > 0) {
        client->bandwidthRating++;
        client->bwrAdjustTime = 0;
    }
    if (client->bandwidthRating < 0) {
        client->bandwidthRating = 0;
    }
    if (client->bandwidthRating > MAX_BANDWIDTH_RATING) {
        client->bandwidthRating = MAX_BANDWIDTH_RATING;
    }
    return qSize <= 10 * limit;
    */
}
/**
 * Reads a PKT_COORDS packet from the message buffer. We trust the
 * client's position and change ours to match it. The client better not
 * be cheating.
 */
void Sv_ClientCoords(int plrNum) {
    player_t           *plr = DD_Player(plrNum);
    ddplayer_t         *ddpl = &plr->publicData();
    mobj_t             *mo = ddpl->mo;
    int                 clz;
    float               clientGameTime;
    float               clientPos[3];
    angle_t             clientAngle;
    float               clientLookDir;
    dd_bool             onFloor = false;
    if (!plr->isInGame() || (ddpl->flags & DDPF_DEAD))
        return;
    clientGameTime = Reader_ReadFloat(msgReader);
    clientPos[VX] = Reader_ReadFloat(msgReader);
    clientPos[VY] = Reader_ReadFloat(msgReader);
    clz = Reader_ReadInt32(msgReader);
    if (clz == DDMININT) {
        clientPos[VZ] = mo->floorZ;
        onFloor = true;
    }
    else {
        clientPos[VZ] = FIX2FLT(clz);
    }
    clientAngle = ((angle_t) Reader_ReadUInt16(msgReader)) << 16;
    clientLookDir = P_ShortToLookDir(Reader_ReadInt16(msgReader));
    ddpl->forwardMove = FIX2FLT(Reader_ReadChar(msgReader) << 13);
    ddpl->sideMove = FIX2FLT(Reader_ReadChar(msgReader) << 13);
    if (ddpl->fixCounter.angles == ddpl->fixAcked.angles && !(ddpl->flags & DDPF_FIXANGLES)) {
        LOGDEV_NET_XVERBOSE_DEBUGONLY("Sv_ClientCoords: Setting angles for player %i: %x, %f",
                                     plrNum << clientAngle << clientLookDir);
        mo->angle = clientAngle;
        ddpl->lookDir = clientLookDir;
    }
    LOGDEV_NET_XVERBOSE_DEBUGONLY("Sv_ClientCoords: Received coords for player %i: %f, %f, %f",
                                 plrNum << clientPos[VX] << clientPos[VY] << clientPos[VZ]);
    if (Sv_CanTrustClientPos(plrNum)) {
        LOGDEV_NET_XVERBOSE_DEBUGONLY("Sv_ClientCoords: Setting coords for player %i: %f, %f, %f",
                                     plrNum << clientPos[VX] << clientPos[VY] << clientPos[VZ]);
        Smoother_AddPos(DD_Player(plrNum)->smoother(), clientGameTime,
                        clientPos[VX], clientPos[VY], clientPos[VZ], onFloor);
    }
}
dd_bool Sv_CanTrustClientPos(int plrNum) {
    player_t* plr = DD_Player(plrNum);
    ddplayer_t* ddpl = &plr->publicData();
    if (ddpl->fixCounter.origin == ddpl->fixAcked.origin && !(ddpl->flags & DDPF_FIXORIGIN)) {
        return true;
    }
    return false;
}
/**
 * Console command for terminating a remote console connection.
 */
D_CMD(Logout) {
    DENG2_UNUSED3(src, argc, argv);
    if (!netRemoteUser || !isServer)
        return false;
    Sv_SendText(netRemoteUser, SV_CONSOLE_PRINT_FLAGS, "Goodbye...\n");
    Msg_Begin(PKT_LOGIN);
    Writer_WriteByte(msgWriter, false);       // You're outta here.
    Msg_End();
    Net_SendBuffer(netRemoteUser, 0);
    netRemoteUser = 0;
    return true;
}
DENG_DECLARE_API(Server) = {
    { DE_API_SERVER },
    Sv_CanTrustClientPos
};
mislink_t *Sv_MRHash(pool_t *pool, thid_t id) {
    return &pool->misHash[(unsigned) id % POOL_MISSILE_HASH_SIZE];
}
/**
 * Returns a pointer to the missile record for the specified ID; otherwise @c nullptr.
 */
misrecord_t *Sv_MRFind(pool_t *pool, thid_t id) {
    mislink_t *hash = Sv_MRHash(pool, id);
    misrecord_t *mis;
    for (mis = hash->first; mis; mis = mis->next) {
        if (mis->id == id) {
            return mis;
        }
    }
    return NULL;
}
/**
 * Adds an entry for the mobj into the missile record.
 */
void Sv_MRAdd(pool_t *pool, const mobjdelta_t *delta) {
    thid_t  id = delta->delta.id;
    mislink_t *hash = Sv_MRHash(pool, id);
    misrecord_t *mis;
#ifdef _DEBUG
if (!(delta->mo.ddFlags & DDMF_MISSILE)) {
    App_Error("Sv_MRAdd: Not a missile.\n");
}
#endif
    mis = Sv_MRFind(pool, id);
    if (!mis) {
        mis = (misrecord_t *) Z_Malloc(sizeof(misrecord_t), PU_MAP, 0);
        mis->id = id;
        mis->next = NULL;
        mis->prev = hash->last;
        if (hash->last)
            hash->last->next = mis;
        hash->last = mis;
        if (!hash->first)
            hash->first = mis;
    }
    /*
       mis->momx = delta->mo.momx;
       mis->momy = delta->mo.momy;
       mis->momz = delta->mo.momz;
     */
}
/**
 * Remove the missile from the record. This is called when the missile
 * mobj is destroyed.
 */
void Sv_MRRemove(pool_t *pool, thid_t id) {
    mislink_t *hash = Sv_MRHash(pool, id);
    misrecord_t *mis;
    for (mis = hash->first; mis; mis = mis->next) {
        if (mis->id == id) {
            if (hash->first == mis)
                hash->first = mis->next;
            if (hash->last == mis)
                hash->last = mis->prev;
            if (mis->next)
                mis->next->prev = mis->prev;
            if (mis->prev)
                mis->prev->next = mis->next;
            break;
        }
    }
}
/**
 * @return              The flags that should be excluded from the missile delta.
 */
int Sv_MRCheck(pool_t *pool, const mobjdelta_t *mobj) {
    misrecord_t *mis;
    int         exclude = 0;
#ifdef _DEBUG
if (!(mobj->mo.ddFlags & DDMF_MISSILE)) {
    App_Error("Sv_MRCheck: Not a missile.\n");
}
#endif
    if ((mis = Sv_MRFind(pool, mobj->delta.id)) == NULL) {
        return 0;
    }
    if (!(mobj->delta.flags & MDF_MOM_X))
        exclude |= MDF_ORIGIN_X;
    if (!(mobj->delta.flags & MDF_MOM_Y))
        exclude |= MDF_ORIGIN_Y;
    if (!(mobj->delta.flags & MDF_MOM_Z))
        exclude |= MDF_ORIGIN_Z;
    return exclude;
}
#define DEFAULT_DELTA_BASE_SCORE    ( 10000 )
#define REG_MOBJ_HASH_SIZE          ( 1024 )
#define REG_MOBJ_HASH_FUNCTION_MASK ( 0x3ff )
#define PLANE_SKIP_LIMIT            ( 40 )
struct reg_mobj_t {
    reg_mobj_t *next;  ///< In the register hash.
    reg_mobj_t *prev;  ///< In the register hash.
    dt_mobj_t mo;      ///< The state of the mobj.
};
struct mobjhash_t {
    reg_mobj_t *first, *last;
};
/**
 * One cregister_t holds the state of the entire world.
 */
struct cregister_t {
    dint gametic;       ///< The time the register was last updated.
    dd_bool isInitial;  ///< @c true if *this* register contains a read-only copy of the initial state of the world.
    mobjhash_t mobjs[REG_MOBJ_HASH_SIZE];
    dt_player_t ddPlayers[DDMAXPLAYERS];
    dt_sector_t *sectors;
    dt_side_t *sides;
    dt_poly_t *polyObjs;
};
void Sv_RegisterWorld(cregister_t *reg, dd_bool isInitial);
void Sv_NewDelta(void *deltaPtr, deltatype_t type, duint id);
dd_bool Sv_IsVoidDelta(void const *delta);
void Sv_PoolQueueClear(pool_t *pool);
void Sv_GenerateNewDeltas(cregister_t *reg, dint clientNumber, dd_bool doUpdate);
cregister_t worldRegister;
cregister_t initialRegister;
static dfloat deltaBaseScores[NUM_DELTA_TYPES];
static ThinkerT<dt_mobj_t> dummyZeroMobj;
static inline ClientServerWorld &worldSys() {
    return App_World();
}
/**
 * Called once for each map, from R_SetupMap(). Initialize the world
 * register and drain all pools.
 */
void Sv_InitPools() {
    Time startedAt;
    if (::isClient) return;
    LOG_AS("Sv_InitPools");
    for (dint i = 0; i < NUM_DELTA_TYPES; ++i) {
        ::deltaBaseScores[i] = DEFAULT_DELTA_BASE_SCORE;
    }
    ::deltaBaseScores[DT_MOBJ        ] = 1000;
    ::deltaBaseScores[DT_PLAYER      ] = 1000;
    ::deltaBaseScores[DT_SECTOR      ] = 2000;
    ::deltaBaseScores[DT_SIDE        ] = 800;
    ::deltaBaseScores[DT_POLY        ] = 2000;
    ::deltaBaseScores[DT_LUMP        ] = 0;
    ::deltaBaseScores[DT_SOUND       ] = 2000;
    ::deltaBaseScores[DT_MOBJ_SOUND  ] = 3000;
    ::deltaBaseScores[DT_SECTOR_SOUND] = 5000;
    ::deltaBaseScores[DT_SIDE_SOUND  ] = 5500;
    ::deltaBaseScores[DT_POLY_SOUND  ] = 5000;
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        pool_t &pool = *Sv_GetPool(i);
        pool.owner         = i;
        pool.resendDealer  = 1;
        de::zap(pool.hash);
        de::zap(pool.misHash);
        pool.queueSize     = 0;
        pool.allocatedSize = 0;
        pool.queue         = nullptr;
        pool.isFirst       = true;  // Set to @c false when a frame is sent.
    }
    Sv_RegisterWorld(&::worldRegister, false);
    Sv_RegisterWorld(&::initialRegister, true);
    LOG_MAP_VERBOSE("World registered in %.2f seconds") << startedAt.since();
}
/**
 * Called during server shutdown (when shutting down the engine).
 */
void Sv_ShutdownPools() {
}
/**
 * Called when a client joins the game.
 */
void Sv_InitPoolForClient(duint clientNumber) {
    DENG2_ASSERT(clientNumber < DDMAXPLAYERS);
    Sv_DrainPool(clientNumber);
    Sv_GenerateNewDeltas(&initialRegister, clientNumber, false);
    Sv_GetPool(clientNumber)->isFirst = true;
}
/**
 * Returns a pointer to the delta pool associated with the given console number.
 */
pool_t *Sv_GetPool(duint consoleNumber) {
    return &DD_Player(consoleNumber)->deltaPool();
}
/**
 * The hash function for the register mobj hash.
 */
duint Sv_RegisterHashFunction(thid_t id) {
    return (duint) id & REG_MOBJ_HASH_FUNCTION_MASK;
}
/**
 * Returns a pointer to the register map-object, if it already exists.
 */
reg_mobj_t *Sv_RegisterFindMobj(cregister_t *reg, thid_t id) {
    DENG2_ASSERT(reg);
    mobjhash_t const &hash = reg->mobjs[Sv_RegisterHashFunction(id)];
    for (reg_mobj_t *it = hash.first; it; it = it->next) {
        if (it->mo.thinker.id == id) return it;
    }
    return nullptr;  // Not found.
}
/**
 * Adds a new reg_mobj_t to the register's mobj hash.
 */
reg_mobj_t *Sv_RegisterAddMobj(cregister_t *reg, thid_t id) {
    DENG2_ASSERT(reg);
    mobjhash_t &hash = reg->mobjs[Sv_RegisterHashFunction(id)];
    if (reg_mobj_t *newRegMo = Sv_RegisterFindMobj(reg, id))
        return newRegMo;
    auto *newRegMo = (reg_mobj_t *) Z_Calloc(sizeof(reg_mobj_t), PU_MAP, 0);
    if (hash.last) {
        hash.last->next = newRegMo;
        newRegMo->prev = hash.last;
    }
    hash.last = newRegMo;
    if (!hash.first) {
        hash.first = newRegMo;
    }
    return newRegMo;
}
/**
 * Removes a reg_mobj_t from the register's mobj hash.
 */
void Sv_RegisterRemoveMobj(cregister_t *reg, reg_mobj_t *regMo) {
    DENG2_ASSERT(regMo);
    mobjhash_t &hash = reg->mobjs[Sv_RegisterHashFunction(regMo->mo.thinker.id)];
    if (hash.last == regMo) {
        hash.last = regMo->prev;
    }
    if (hash.first == regMo) {
        hash.first = regMo->next;
    }
    if (regMo->next) {
        regMo->next->prev = regMo->prev;
    }
    if (regMo->prev) {
        regMo->prev->next = regMo->next;
    }
    Z_Free(regMo);
}
/**
 * @return @c DDMINFLOAT= @a mob is on the floor.
 *         @c DDMAXFLOAT= @a mob is touching the ceiling.
 *         Otherwise returns the actual world Z coordinate.
 */
dfloat Sv_GetMaxedMobjZ(mobj_t const *mob) {
    /*if (mob->origin[VZ] == mob->floorZ)
    {
        return DDMINFLOAT;
    }
    if (mob->origin[VZ] + mob->height == mob->ceilingZ) {
        return DDMAXFLOAT;
    }*/
    return mob->origin[VZ];
}
/**
 * Store the state of the mobj into the register map-object.
 * Called at register init and after each delta generation cycle.
 */
void Sv_RegisterMobj(dt_mobj_t *reg, mobj_t const *mob) {
    DENG2_ASSERT(reg && mob);
    reg->thinker.id   = mob->thinker.id;
    reg->type         = mob->type;
    reg->dPlayer      = mob->dPlayer;
    reg->_bspLeaf     = mob->_bspLeaf;
    reg->origin[0]    = mob->origin[0];
    reg->origin[1]    = mob->origin[1];
    reg->origin[2]    = Sv_GetMaxedMobjZ(mob);
    reg->floorZ       = mob->floorZ;
    reg->ceilingZ     = mob->ceilingZ;
    reg->mom[0]       = mob->mom[0];
    reg->mom[1]       = mob->mom[1];
    reg->mom[2]       = mob->mom[2];
    reg->angle        = mob->angle;
    reg->selector     = mob->selector;
    reg->state        = mob->state;
    reg->radius       = mob->radius;
    reg->height       = mob->height;
    reg->ddFlags      = mob->ddFlags;
    reg->flags        = mob->flags;
    reg->flags2       = mob->flags2;
    reg->flags3       = mob->flags3;
    reg->health       = mob->health;
    reg->floorClip    = mob->floorClip;
    reg->translucency = mob->translucency;
    reg->visTarget    = mob->visTarget;
}
/**
 * Reset the data of the registered mobj to reasonable defaults.
 * In effect, forces a resend of the zeroed entries as deltas.
 */
void Sv_RegisterResetMobj(dt_mobj_t *reg) {
    DENG2_ASSERT(reg);
    reg->origin[0]    = DDMINFLOAT;
    reg->origin[1]    = DDMINFLOAT;
    reg->origin[2]    = -1000000;
    reg->angle        = 0;
    reg->type         = -1;
    reg->selector     = 0;
    reg->state        = 0;
    reg->radius       = -1;
    reg->height       = -1;
    reg->ddFlags      = 0;
    reg->flags        = 0;
    reg->flags2       = 0;
    reg->flags3       = 0;
    reg->health       = 0;
    reg->floorClip    = 0;
    reg->translucency = 0;
    reg->visTarget    = 0;
}
/**
 * Store the state of the player into the register-player.
 * Called at register init and after each delta generation cycle.
 */
void Sv_RegisterPlayer(dt_player_t *reg, duint number) {
#define FMAKERGBA(r, g, b, a) \
    ( byte( 0xff * r ) + ( byte( 0xff * g ) << 8 ) + ( byte( 0xff * b ) << 16 ) + ( byte (0xff * a ) << 24 ) )
    DENG2_ASSERT(reg);
    DENG2_ASSERT(number < DDMAXPLAYERS);
    player_t *plr    = DD_Player(number);
    ddplayer_t *ddpl = &plr->publicData();
    reg->mobj          = (ddpl->mo ? ddpl->mo->thinker.id : 0);
    reg->forwardMove   = 0;
    reg->sideMove      = 0;
    reg->angle         = (ddpl->mo ? ddpl->mo->angle : 0);
    reg->turnDelta     = (ddpl->mo ? ddpl->mo->angle - ddpl->lastAngle : 0);
    reg->friction      = ddpl->mo && (gx.MobjFriction ? gx.MobjFriction(ddpl->mo) : DEFAULT_FRICTION);
    reg->extraLight    = ddpl->extraLight;
    reg->fixedColorMap = ddpl->fixedColorMap;
    if (ddpl->flags & DDPF_VIEW_FILTER) {
        reg->filter = FMAKERGBA(ddpl->filterColor[0],
                                ddpl->filterColor[1],
                                ddpl->filterColor[2],
                                ddpl->filterColor[3]);
    }
    else {
        reg->filter = 0;
    }
    reg->clYaw   = (ddpl->mo ? ddpl->mo->angle : 0);
    reg->clPitch = ddpl->lookDir;
    std::memcpy(reg->psp, ddpl->pSprites, sizeof(ddpsprite_t) * 2);
#undef FMAKERGBA
}
/**
 * Store the state of the sector into the register-sector.
 * Called at register init and after each delta generation.
 *
 * @param reg     The sector register to be initialized.
 * @param number  The world sector number to be registered.
 */
void Sv_RegisterSector(dt_sector_t *reg, dint number) {
    DENG2_ASSERT(reg);
    Sector &sector = worldSys().map().sector(number);
    reg->lightLevel = sector.lightLevel();
    for (dint i = 0; i < 3; ++i) {
        reg->rgb[i] = sector.lightColor()[i];
    }
    for (dint i = 0; i < 2; ++i) // number of planes in sector.
    {
        Plane const &plane = sector.plane(i);
        reg->planes[i].height = plane.height();
        reg->planes[i].target = plane.heightTarget();
        reg->planes[i].speed  = plane.speed();
        Surface const &surface = plane.surface();
        Vector3f const &tintColor = surface.color();
        for (dint c = 0; c < 3; ++c) {
            reg->planes[i].surface.rgba[c] = tintColor[c];
        }
        reg->planes[i].surface.rgba[CA] = surface.opacity();
        reg->planes[i].surface.material = surface.materialPtr();
    }
}
/**
 * Store the state of the side into the register-side.
 * Called at register init and after each delta generation.
 */
void Sv_RegisterSide(dt_side_t *reg, dint number) {
    DENG2_ASSERT(reg);
    LineSide *side = worldSys().map().sidePtr(number);
    if (side->hasSections()) {
        reg->top   .material = side->top   ().materialPtr();
        reg->middle.material = side->middle().materialPtr();
        reg->bottom.material = side->bottom().materialPtr();
        for (dint c = 0; c < 3; ++c) {
            reg->middle.rgba[c] = side->middle().color()[c];
            reg->bottom.rgba[c] = side->bottom().color()[c];
            reg->top   .rgba[c] = side->top   ().color()[c];
        }
        reg->middle.rgba[3]   = side->middle().opacity();
        reg->middle.blendMode = side->middle().blendMode();
    }
    reg->lineFlags = side->line().flags() & 0xff;
    reg->flags     = side->flags() & 0xff;
}
/**
 * Store the state of the polyobj into the register-poly.
 * Called at register init and after each delta generation.
 */
void Sv_RegisterPoly(dt_poly_t *reg, duint number) {
    DENG_ASSERT(reg);
    Polyobj const &pob = worldSys().map().polyobj(number);
    reg->dest[0]    = pob.dest[0];
    reg->dest[1]    = pob.dest[1];
    reg->speed      = pob.speed;
    reg->destAngle  = pob.destAngle;
    reg->angleSpeed = pob.angleSpeed;
}
/**
 * Returns @c true if the result is not void.
 */
dd_bool Sv_RegisterCompareMobj(cregister_t *reg, mobj_t const *s, mobjdelta_t *d) {
    dint df;
    dt_mobj_t const *r = ::dummyZeroMobj;
    reg_mobj_t *regMo  = Sv_RegisterFindMobj(reg, s->thinker.id);
    if (regMo) {
        r  = &regMo->mo;
        df = 0;
    }
    else {
        df = MDFC_CREATE | MDF_EVERYTHING | MDFC_TYPE;
    }
    if (r->origin[0] != s->origin[0])
        df |= MDF_ORIGIN_X;
    if (r->origin[1] != s->origin[1])
        df |= MDF_ORIGIN_Y;
    if (r->origin[2] != Sv_GetMaxedMobjZ(s) || r->floorZ != s->floorZ || r->ceilingZ != s->ceilingZ) {
        df |= MDF_ORIGIN_Z;
        if (!(df & MDFC_CREATE) && s->origin[2] <= s->floorZ) {
            df |= MDFC_ON_FLOOR;
        }
    }
    if (r->mom[0] != s->mom[0])
        df |= MDF_MOM_X;
    if (r->mom[1] != s->mom[1])
        df |= MDF_MOM_Y;
    if (r->mom[2] != s->mom[2])
        df |= MDF_MOM_Z;
    if (r->angle != s->angle)
        df |= MDF_ANGLE;
    if (r->selector != s->selector)
        df |= MDF_SELECTOR;
    if (r->translucency != s->translucency)
        df |= MDFC_TRANSLUCENCY;
    if (r->visTarget != s->visTarget)
        df |= MDFC_FADETARGET;
    if (r->type != s->type)
        df |= MDFC_TYPE;
    if (regMo && !Def_SameStateSequence(s->state, r->state)) {
        df |= MDF_STATE;
        if (s->state == nullptr) {
            return false;
        }
    }
    if (r->radius != s->radius)
        df |= MDF_RADIUS;
    if (r->height != s->height)
        df |= MDF_HEIGHT;
    if ((r->ddFlags & DDMF_PACK_MASK) != (s->ddFlags & DDMF_PACK_MASK) ||
       r->flags != s->flags || r->flags2 != s->flags2 || r->flags3 != s->flags3) {
        df |= MDF_FLAGS;
    }
    if (r->health != s->health)
        df |= MDF_HEALTH;
    if (r->floorClip != s->floorClip)
        df |= MDF_FLOORCLIP;
    if (df) {
        Sv_NewDelta(d, DT_MOBJ, s->thinker.id);
        Sv_RegisterMobj(&d->mo, s);
    }
    d->delta.flags = df;
    return !Sv_IsVoidDelta(d);
}
/**
 * Returns @c true if the result is not void.
 */
dd_bool Sv_RegisterComparePlayer(cregister_t *reg, duint number, playerdelta_t *d) {
    DENG2_ASSERT(number < DDMAXPLAYERS);
    dt_player_t const *r = &reg->ddPlayers[number];
    dt_player_t *s       = &d->player;
    dint df = 0;
    Sv_NewDelta(d, DT_PLAYER, number);
    Sv_RegisterPlayer(&d->player, number);
    if (r->mobj != s->mobj)
        df |= PDF_MOBJ;
    if (r->forwardMove != s->forwardMove)
        df |= PDF_FORWARDMOVE;
    if (r->sideMove != s->sideMove)
        df |= PDF_SIDEMOVE;
    if (r->turnDelta != s->turnDelta)
        df |= PDF_TURNDELTA;
    if (r->friction != s->friction)
        df |= PDF_FRICTION;
    if (r->extraLight != s->extraLight || r->fixedColorMap != s->fixedColorMap)
        df |= PDF_EXTRALIGHT;
    if (r->filter != s->filter)
        df |= PDF_FILTER;
    d->delta.flags = df;
    return !Sv_IsVoidDelta(d);
}
/**
 * @return  @c true if the result is not void.
 */
dd_bool Sv_RegisterCompareSector(cregister_t *reg, dint number, sectordelta_t *d, byte doUpdate) {
    DENG2_ASSERT(reg && d);
    dt_sector_t *r  = &reg->sectors[number];
    Sector const &s = worldSys().map().sector(number);
    dint df = 0;
    if (s.floor().surface().materialPtr() != r->planes[PLN_FLOOR].surface.material)
        df |= SDF_FLOOR_MATERIAL;
    if (s.ceiling().surface().materialPtr() != r->planes[PLN_CEILING].surface.material)
       df |= SDF_CEILING_MATERIAL;
    if (r->lightLevel != s.lightLevel())
        df |= SDF_LIGHT;
    if (r->rgb[0] != s.lightColor().x)
        df |= SDF_COLOR_RED;
    if (r->rgb[1] != s.lightColor().y)
        df |= SDF_COLOR_GREEN;
    if (r->rgb[2] != s.lightColor().z)
        df |= SDF_COLOR_BLUE;
    if (r->planes[PLN_FLOOR].surface.rgba[0] != s.floor().surface().color().x)
        df |= SDF_FLOOR_COLOR_RED;
    if (r->planes[PLN_FLOOR].surface.rgba[1] != s.floor().surface().color().y)
        df |= SDF_FLOOR_COLOR_GREEN;
    if (r->planes[PLN_FLOOR].surface.rgba[2] != s.floor().surface().color().z)
        df |= SDF_FLOOR_COLOR_BLUE;
    if (r->planes[PLN_CEILING].surface.rgba[0] != s.ceiling().surface().color().x)
        df |= SDF_CEIL_COLOR_RED;
    if (r->planes[PLN_CEILING].surface.rgba[1] != s.ceiling().surface().color().y)
        df |= SDF_CEIL_COLOR_GREEN;
    if (r->planes[PLN_CEILING].surface.rgba[2] != s.ceiling().surface().color().z)
        df |= SDF_CEIL_COLOR_BLUE;
    if (fequal(r->planes[PLN_FLOOR].speed, 0) && fequal(s.floor().speed(), 0)) {
        if (!fequal(r->planes[PLN_FLOOR].height, s.floor().height()))
            df |= SDF_FLOOR_HEIGHT;
    }
    else {
        if (fabs(r->planes[PLN_FLOOR].height - s.floor().height()) > PLANE_SKIP_LIMIT)
            df |= SDF_FLOOR_HEIGHT;
    }
    if (fequal(r->planes[PLN_CEILING].speed, 0) && fequal(s.ceiling().speed(), 0)) {
        if (!fequal(r->planes[PLN_CEILING].height, s.ceiling().height()))
            df |= SDF_CEILING_HEIGHT;
    }
    else {
        if (fabs(r->planes[PLN_CEILING].height - s.ceiling().height()) > PLANE_SKIP_LIMIT)
            df |= SDF_CEILING_HEIGHT;
    }
    if (!fequal(r->planes[PLN_FLOOR].target, s.floor().heightTarget())) {
        df |= SDF_FLOOR_TARGET | SDF_FLOOR_SPEED;
    }
    if (!fequal(r->planes[PLN_FLOOR].speed, s.floor().speed())) {
        df |= SDF_FLOOR_SPEED | SDF_FLOOR_TARGET;
    }
    if (!fequal(r->planes[PLN_CEILING].target, s.ceiling().heightTarget())) {
        df |= SDF_CEILING_TARGET | SDF_CEILING_SPEED;
    }
    if (!fequal(r->planes[PLN_CEILING].speed, s.ceiling().speed())) {
        df |= SDF_CEILING_SPEED | SDF_CEILING_TARGET;
    }
#ifdef DENG2_DEBUG
    if (df & (SDF_CEILING_HEIGHT | SDF_CEILING_SPEED | SDF_CEILING_TARGET)) {
        LOGDEV_NET_XVERBOSE("Sector %i: ceiling state change noted (target = %f)",
                            number << s.ceiling().heightTarget());
    }
#endif
    if (df) {
        Sv_NewDelta(d, DT_SECTOR, number);
        Sv_RegisterSector(&d->sector, number);
        if (doUpdate) {
            Sv_RegisterSector(r, number);
        }
    }
    if (doUpdate) {
        r->planes[PLN_FLOOR  ].height = s.floor().height();
        r->planes[PLN_CEILING].height = s.ceiling().height();
    }
    d->delta.flags = df;
    return !Sv_IsVoidDelta(d);
}
/**
 * @return  @c true= the result is not void.
 */
dd_bool Sv_RegisterCompareSide(cregister_t *reg, duint number, sidedelta_t *d, byte doUpdate) {
    DENG2_ASSERT(reg/* && d*/);
    LineSide const *side = worldSys().map().sidePtr(number);
    dt_side_t *r         = &reg->sides[number];
    byte lineFlags = side->line().flags() & 0xff;
    byte sideFlags = side->flags() & 0xff;
    dint df = 0;
    if (side->hasSections()) {
        if (!side->top().hasFixMaterial() && r->top.material != side->top().materialPtr()) {
            df |= SIDF_TOP_MATERIAL;
            if (doUpdate)
                r->top.material = side->top().materialPtr();
        }
        if (!side->middle().hasFixMaterial() && r->middle.material != side->middle().materialPtr()) {
            df |= SIDF_MID_MATERIAL;
            if (doUpdate)
                r->middle.material = side->middle().materialPtr();
        }
        if (!side->bottom().hasFixMaterial() && r->bottom.material != side->bottom().materialPtr()) {
            df |= SIDF_BOTTOM_MATERIAL;
            if (doUpdate)
                r->bottom.material = side->bottom().materialPtr();
        }
        if (r->top.rgba[0] != side->top().color().x) {
            df |= SIDF_TOP_COLOR_RED;
            if (doUpdate)
                r->top.rgba[0] = side->top().color().x;
        }
        if (r->top.rgba[1] != side->top().color().y) {
            df |= SIDF_TOP_COLOR_GREEN;
            if (doUpdate)
                r->top.rgba[1] = side->top().color().y;
        }
        if (r->top.rgba[2] != side->top().color().z) {
            df |= SIDF_TOP_COLOR_BLUE;
            if (doUpdate)
                r->top.rgba[3] = side->top().color().z;
        }
        if (r->middle.rgba[0] != side->middle().color().x) {
            df |= SIDF_MID_COLOR_RED;
            if (doUpdate)
                r->middle.rgba[0] = side->middle().color().x;
        }
        if (r->middle.rgba[1] != side->middle().color().y) {
            df |= SIDF_MID_COLOR_GREEN;
            if (doUpdate)
                r->middle.rgba[1] = side->middle().color().y;
        }
        if (r->middle.rgba[2] != side->middle().color().z) {
            df |= SIDF_MID_COLOR_BLUE;
            if (doUpdate)
                r->middle.rgba[3] = side->middle().color().z;
        }
        if (r->middle.rgba[3] != side->middle().opacity()) {
            df |= SIDF_MID_COLOR_ALPHA;
            if (doUpdate)
                r->middle.rgba[3] = side->middle().opacity();
        }
        if (r->bottom.rgba[0] != side->bottom().color().x) {
            df |= SIDF_BOTTOM_COLOR_RED;
            if (doUpdate)
                r->bottom.rgba[0] = side->bottom().color().x;
        }
        if (r->bottom.rgba[1] != side->bottom().color().y) {
            df |= SIDF_BOTTOM_COLOR_GREEN;
            if (doUpdate)
                r->bottom.rgba[1] = side->bottom().color().y;
        }
        if (r->bottom.rgba[2] != side->bottom().color().z) {
            df |= SIDF_BOTTOM_COLOR_BLUE;
            if (doUpdate)
                r->bottom.rgba[3] = side->bottom().color().z;
        }
        if (r->middle.blendMode != side->middle().blendMode()) {
            df |= SIDF_MID_BLENDMODE;
            if (doUpdate)
                r->middle.blendMode = side->middle().blendMode();
        }
    }
    if (r->lineFlags != lineFlags) {
        df |= SIDF_LINE_FLAGS;
        if (doUpdate)
            r->lineFlags = lineFlags;
    }
    if (r->flags != sideFlags) {
        df |= SIDF_FLAGS;
        if (doUpdate)
            r->flags = sideFlags;
    }
    if (df) {
        Sv_NewDelta(d, DT_SIDE, number);
        Sv_RegisterSide(&d->side, number);
    }
    d->delta.flags = df;
    return !Sv_IsVoidDelta(d);
}
/**
 * @return  @c true if the result is not void.
 */
dd_bool Sv_RegisterComparePoly(cregister_t *reg, dint number, polydelta_t *d) {
    DENG2_ASSERT(reg/* && d*/);
    dt_poly_t const *r = &reg->polyObjs[number];
    dt_poly_t *s       = &d->po;
    dint df = 0;
    Sv_NewDelta(d, DT_POLY, number);
    Sv_RegisterPoly(&d->po, number);
    if (r->dest[VX] != s->dest[VX])
        df |= PODF_DEST_X;
    if (r->dest[VY] != s->dest[VY])
        df |= PODF_DEST_Y;
    if (r->speed != s->speed)
        df |= PODF_SPEED;
    if (r->destAngle != s->destAngle)
        df |= PODF_DEST_ANGLE;
    if (r->angleSpeed != s->angleSpeed)
        df |= PODF_ANGSPEED;
    d->delta.flags = df;
    return !Sv_IsVoidDelta(d);
}
/**
 * Returns @c true if the map-object can be excluded from delta processing.
 */
dd_bool Sv_IsMobjIgnored(mobj_t const &mob) {
    return (mob.ddFlags & DDMF_LOCAL) != 0;
}
/**
 * Returns @c true if the player can be excluded from delta processing.
 */
dd_bool Sv_IsPlayerIgnored(dint plrNum) {
    DENG2_ASSERT(plrNum >= 0 && plrNum < DDMAXPLAYERS);
    return !DD_Player(plrNum)->publicData().inGame;
}
/**
 * Initialize the register with the current state of the world.
 *
 * The arrays are allocated and the data is copied, nothing else is done.
 *
 * An initial register doesn't contain any mobjs. When new clients enter, they know
 * nothing about any mobjs. If the mobjs were included in the initial register, clients
 * wouldn't receive much info from mobjs that haven't moved since the beginning.
 */
void Sv_RegisterWorld(cregister_t *reg, dd_bool isInitial) {
    DENG2_ASSERT(reg);
    world::Map &map = worldSys().map();
    de::zapPtr(reg);
    reg->gametic = SECONDS_TO_TICKS(gameTime);
    reg->isInitial = isInitial;
    reg->sectors = (dt_sector_t *) Z_Calloc(sizeof(*reg->sectors) * map.sectorCount(), PU_MAP, 0);
    for (dint i = 0; i < map.sectorCount(); ++i) {
        Sv_RegisterSector(&reg->sectors[i], i);
    }
    reg->sides = (dt_side_t *) Z_Calloc(sizeof(*reg->sides) * map.sideCount(), PU_MAP, 0);
    for (dint i = 0; i < map.sideCount(); ++i) {
        Sv_RegisterSide(&reg->sides[i], i);
    }
    dint numPolyobjs = map.polyobjCount();
    if (numPolyobjs) {
        reg->polyObjs = (dt_poly_t *) Z_Calloc(sizeof(*reg->polyObjs) * map.polyobjCount(), PU_MAP, 0);
        for (dint i = 0; i < numPolyobjs; ++i) {
            Sv_RegisterPoly(&reg->polyObjs[i], i);
        }
    }
    else {
        reg->polyObjs = nullptr;
    }
}
/**
 * Update the pool owner's info.
 */
void Sv_UpdateOwnerInfo(pool_t *pool) {
    DENG2_ASSERT(pool);
    player_t *plr     = DD_Player(pool->owner);
    ownerinfo_t *info = &pool->ownerInfo;
    de::zapPtr(info);
    info->pool = pool;
    if (plr->publicData().mo) {
        mobj_t *mob = plr->publicData().mo;
        V3d_Copy(info->origin, mob->origin);
        info->angle = mob->angle;
        info->speed = M_ApproxDistance(mob->mom[0], mob->mom[1]);
    }
    info->ackThreshold = 0; //Net_GetAckThreshold(pool->owner);
}
/**
 * @return  A timestamp that is used to track how old deltas are.
 */
uint Sv_GetTimeStamp() {
    return Timer_RealMilliseconds();
}
/**
 * Initialize a new delta.
 */
void Sv_NewDelta(void *deltaPtr, deltatype_t type, uint id) {
    if (!deltaPtr) return;
    delta_t *delta = (delta_t *) deltaPtr;
    de::zapPtr(delta);
    delta->id = id;
    delta->type = type;
    delta->state = DELTA_NEW;
    delta->timeStamp = Sv_GetTimeStamp();
}
/**
 * @return  @c true, if the delta contains no information.
 */
dd_bool Sv_IsVoidDelta(void const *delta) {
    return ((delta_t const *) delta)->flags == 0;
}
/**
 * @return  @c true, if the delta is a Sound delta.
 */
dd_bool Sv_IsSoundDelta(void const *delta) {
    delta_t const *d = (delta_t const *) delta;
    return (d->type == DT_SOUND ||
            d->type == DT_MOBJ_SOUND ||
            d->type == DT_SECTOR_SOUND ||
            d->type == DT_SIDE_SOUND ||
            d->type == DT_POLY_SOUND);
}
/**
 * @return  @c true, if the delta is a Start Sound delta.
 */
dd_bool Sv_IsStartSoundDelta(void const *delta) {
    sounddelta_t const *d = (sounddelta_t const *) delta;
    return Sv_IsSoundDelta(delta) &&
        ((d->delta.flags & SNDDF_VOLUME) && d->volume > 0);
}
/**
 * @return  @c true, if the delta is Stop Sound delta.
 */
dd_bool Sv_IsStopSoundDelta(void const *delta) {
    sounddelta_t const *d = (sounddelta_t const *) delta;
    return Sv_IsSoundDelta(delta) &&
        ((d->delta.flags & SNDDF_VOLUME) && d->volume <= 0);
}
/**
 * @return  @c true, if the delta is a Null Mobj delta.
 */
dd_bool Sv_IsNullMobjDelta(void const *delta) {
    return ((delta_t const *) delta)->type == DT_MOBJ &&
        (((delta_t const *) delta)->flags & MDFC_NULL);
}
/**
 * @return  @c true, if the delta is a Create Mobj delta.
 */
dd_bool Sv_IsCreateMobjDelta(void const *delta) {
    return ((delta_t const *) delta)->type == DT_MOBJ &&
        (((delta_t const *) delta)->flags & MDFC_CREATE);
}
/**
 * @return  @c true, if the deltas refer to the same object.
 */
dd_bool Sv_IsSameDelta(void const *delta1, void const *delta2) {
    delta_t const *a = (delta_t const *) delta1, *b = (delta_t const *) delta2;
    return (a->type == b->type) && (a->id == b->id);
}
/**
 * Makes a copy of the delta.
 */
void* Sv_CopyDelta(void* deltaPtr) {
    void*               newDelta;
    delta_t*            delta = (delta_t *) deltaPtr;
    size_t              size =
        ( delta->type == DT_MOBJ ?         sizeof(mobjdelta_t)
        : delta->type == DT_PLAYER ?       sizeof(playerdelta_t)
        : delta->type == DT_SECTOR ?       sizeof(sectordelta_t)
        : delta->type == DT_SIDE ?         sizeof(sidedelta_t)
        : delta->type == DT_POLY ?         sizeof(polydelta_t)
        : delta->type == DT_SOUND ?        sizeof(sounddelta_t)
        : delta->type == DT_MOBJ_SOUND ?   sizeof(sounddelta_t)
        : delta->type == DT_SECTOR_SOUND ? sizeof(sounddelta_t)
        : delta->type == DT_SIDE_SOUND ?   sizeof(sounddelta_t)
        : delta->type == DT_POLY_SOUND ?   sizeof(sounddelta_t)
         /* : delta->type == DT_LUMP?   sizeof(lumpdelta_t) */
        : 0);
    if (size == 0) {
        App_Error("Sv_CopyDelta: Unknown delta type %i.\n", delta->type);
    }
    newDelta = Z_Malloc(size, PU_MAP, 0);
    memcpy(newDelta, deltaPtr, size);
    return newDelta;
}
/**
 * Subtracts the contents of the second delta from the first delta.
 * Subtracting means that if a given flag is defined for both 1 and 2,
 * the flag for 1 is cleared (2 overrides 1). The result is that the
 * deltas can be applied in any order and the result is still correct.
 *
 * 1 and 2 must refer to the same entity!
 */
void Sv_SubtractDelta(void* deltaPtr1, const void* deltaPtr2) {
    delta_t*            delta = (delta_t *) deltaPtr1;
    const delta_t*      sub = (delta_t const *) deltaPtr2;
#ifdef _DEBUG
    if (!Sv_IsSameDelta(delta, sub)) {
        App_Error("Sv_SubtractDelta: Not the same!\n");
    }
#endif
    if (Sv_IsNullMobjDelta(sub)) {
        delta->flags = 0;
    }
    else {
        delta->flags &= ~(delta->flags & sub->flags);
    }
}
/**
 * Applies the data in the source delta to the destination delta.
 * Both must be in the NEW state. Handles all types of deltas.
 */
void Sv_ApplyDeltaData(void *destDelta, void const *srcDelta) {
    delta_t const *src = (delta_t const *) srcDelta;
    delta_t *dest      = (delta_t *) destDelta;
    int sf = src->flags;
    if (src->type == DT_MOBJ) {
        dt_mobj_t const *s = &((mobjdelta_t const *) src)->mo;
        dt_mobj_t *d       = &((mobjdelta_t *) dest)->mo;
        d->dPlayer = s->dPlayer;
        if (sf & (MDF_ORIGIN_X | MDF_ORIGIN_Y))
            d->_bspLeaf = s->_bspLeaf;
        if (sf & MDF_ORIGIN_X)
            d->origin[VX] = s->origin[VX];
        if (sf & MDF_ORIGIN_Y)
            d->origin[VY] = s->origin[VY];
        if (sf & MDF_ORIGIN_Z)
            d->origin[VZ] = s->origin[VZ];
        if (sf & MDF_MOM_X)
            d->mom[MX] = s->mom[MX];
        if (sf & MDF_MOM_Y)
            d->mom[MY] = s->mom[MY];
        if (sf & MDF_MOM_Z)
            d->mom[MZ] = s->mom[MZ];
        if (sf & MDF_ANGLE)
            d->angle = s->angle;
        if (sf & MDF_SELECTOR)
            d->selector = s->selector;
        if (sf & MDF_STATE) {
            d->state = s->state;
            d->tics = (s->state ? s->state->tics : 0);
        }
        if (sf & MDF_RADIUS)
            d->radius = s->radius;
        if (sf & MDF_HEIGHT)
            d->height = s->height;
        if (sf & MDF_FLAGS) {
            d->ddFlags = s->ddFlags;
            d->flags   = s->flags;
            d->flags2  = s->flags2;
            d->flags3  = s->flags3;
        }
        if (sf & MDF_FLOORCLIP)
            d->floorClip = s->floorClip;
        if (sf & MDFC_TRANSLUCENCY)
            d->translucency = s->translucency;
        if (sf & MDFC_FADETARGET)
            d->visTarget = s->visTarget;
    }
    else if (src->type == DT_PLAYER) {
        const dt_player_t*  s = &((const playerdelta_t *) src)->player;
        dt_player_t*        d = &((playerdelta_t *) dest)->player;
        if (sf & PDF_MOBJ)
            d->mobj = s->mobj;
        if (sf & PDF_FORWARDMOVE)
            d->forwardMove = s->forwardMove;
        if (sf & PDF_SIDEMOVE)
            d->sideMove = s->sideMove;
        /*if (sf & PDF_ANGLE)
            d->angle = s->angle;*/
        if (sf & PDF_TURNDELTA)
            d->turnDelta = s->turnDelta;
        if (sf & PDF_FRICTION)
            d->friction = s->friction;
        if (sf & PDF_EXTRALIGHT) {
            d->extraLight = s->extraLight;
            d->fixedColorMap = s->fixedColorMap;
        }
        if (sf & PDF_FILTER)
            d->filter = s->filter;
        if (sf & PDF_PSPRITES) {
            uint                i;
            for (i = 0; i < 2; ++i) {
                int                 off = 16 + i * 8;
                if (sf & (PSDF_STATEPTR << off)) {
                    d->psp[i].statePtr = s->psp[i].statePtr;
                    d->psp[i].tics =
                        (s->psp[i].statePtr ? s->psp[i].statePtr->tics : 0);
                }
                /*if (sf & (PSDF_LIGHT << off))
                    d->psp[i].light = s->psp[i].light;*/
                if (sf & (PSDF_ALPHA << off))
                    d->psp[i].alpha = s->psp[i].alpha;
                if (sf & (PSDF_STATE << off))
                    d->psp[i].state = s->psp[i].state;
                if (sf & (PSDF_OFFSET << off)) {
                    d->psp[i].offset[VX] = s->psp[i].offset[VX];
                    d->psp[i].offset[VY] = s->psp[i].offset[VY];
                }
            }
        }
    }
    else if (src->type == DT_SECTOR) {
        const dt_sector_t*  s = &((const sectordelta_t *) src)->sector;
        dt_sector_t*        d = &((sectordelta_t *) dest)->sector;
        if (sf & SDF_FLOOR_MATERIAL)
            d->planes[PLN_FLOOR].surface.material =
                s->planes[PLN_FLOOR].surface.material;
        if (sf & SDF_CEILING_MATERIAL)
            d->planes[PLN_CEILING].surface.material =
                s->planes[PLN_CEILING].surface.material;
        if (sf & SDF_LIGHT)
            d->lightLevel = s->lightLevel;
        if (sf & SDF_FLOOR_TARGET)
            d->planes[PLN_FLOOR].target = s->planes[PLN_FLOOR].target;
        if (sf & SDF_FLOOR_SPEED)
            d->planes[PLN_FLOOR].speed = s->planes[PLN_FLOOR].speed;
        if (sf & SDF_CEILING_TARGET)
            d->planes[PLN_CEILING].target = s->planes[PLN_CEILING].target;
        if (sf & SDF_CEILING_SPEED)
            d->planes[PLN_CEILING].speed = s->planes[PLN_CEILING].speed;
        if (sf & SDF_FLOOR_HEIGHT)
            d->planes[PLN_FLOOR].height = s->planes[PLN_FLOOR].height;
        if (sf & SDF_CEILING_HEIGHT)
            d->planes[PLN_CEILING].height = s->planes[PLN_CEILING].height;
        if (sf & SDF_COLOR_RED)
            d->rgb[0] = s->rgb[0];
        if (sf & SDF_COLOR_GREEN)
            d->rgb[1] = s->rgb[1];
        if (sf & SDF_COLOR_BLUE)
            d->rgb[2] = s->rgb[2];
        if (sf & SDF_FLOOR_COLOR_RED)
            d->planes[PLN_FLOOR].surface.rgba[0] =
                s->planes[PLN_FLOOR].surface.rgba[0];
        if (sf & SDF_FLOOR_COLOR_GREEN)
            d->planes[PLN_FLOOR].surface.rgba[1] =
                s->planes[PLN_FLOOR].surface.rgba[1];
        if (sf & SDF_FLOOR_COLOR_BLUE)
            d->planes[PLN_FLOOR].surface.rgba[2] =
                s->planes[PLN_FLOOR].surface.rgba[2];
        if (sf & SDF_CEIL_COLOR_RED)
            d->planes[PLN_CEILING].surface.rgba[0] =
                s->planes[PLN_CEILING].surface.rgba[0];
        if (sf & SDF_CEIL_COLOR_GREEN)
            d->planes[PLN_CEILING].surface.rgba[1] =
                s->planes[PLN_CEILING].surface.rgba[1];
        if (sf & SDF_CEIL_COLOR_BLUE)
            d->planes[PLN_CEILING].surface.rgba[2] =
                s->planes[PLN_CEILING].surface.rgba[2];
    }
    else if (src->type == DT_SIDE) {
        const dt_side_t*    s = &((const sidedelta_t *) src)->side;
        dt_side_t*          d = &((sidedelta_t *) dest)->side;
        if (sf & SIDF_TOP_MATERIAL)
            d->top.material = s->top.material;
        if (sf & SIDF_MID_MATERIAL)
            d->middle.material = s->middle.material;
        if (sf & SIDF_BOTTOM_MATERIAL)
            d->bottom.material = s->bottom.material;
        if (sf & SIDF_LINE_FLAGS)
            d->lineFlags = s->lineFlags;
        if (sf & SIDF_TOP_COLOR_RED)
            d->top.rgba[0] = s->top.rgba[0];
        if (sf & SIDF_TOP_COLOR_GREEN)
            d->top.rgba[1] = s->top.rgba[1];
        if (sf & SIDF_TOP_COLOR_BLUE)
            d->top.rgba[2] = s->top.rgba[2];
        if (sf & SIDF_MID_COLOR_RED)
            d->middle.rgba[0] = s->middle.rgba[0];
        if (sf & SIDF_MID_COLOR_GREEN)
            d->middle.rgba[1] = s->middle.rgba[1];
        if (sf & SIDF_MID_COLOR_BLUE)
            d->middle.rgba[2] = s->middle.rgba[2];
        if (sf & SIDF_MID_COLOR_ALPHA)
            d->middle.rgba[3] = s->middle.rgba[3];
        if (sf & SIDF_BOTTOM_COLOR_RED)
            d->bottom.rgba[0] = s->bottom.rgba[0];
        if (sf & SIDF_BOTTOM_COLOR_GREEN)
            d->bottom.rgba[1] = s->bottom.rgba[1];
        if (sf & SIDF_BOTTOM_COLOR_BLUE)
            d->bottom.rgba[2] = s->bottom.rgba[2];
        if (sf & SIDF_MID_BLENDMODE)
            d->middle.blendMode = s->middle.blendMode;
        if (sf & SIDF_FLAGS)
            d->flags = s->flags;
    }
    else if (src->type == DT_POLY) {
        const dt_poly_t*    s = &((const polydelta_t *) src)->po;
        dt_poly_t*          d = &((polydelta_t *) dest)->po;
        if (sf & PODF_DEST_X)
            d->dest[VX] = s->dest[VX];
        if (sf & PODF_DEST_Y)
            d->dest[VY] = s->dest[VY];
        if (sf & PODF_SPEED)
            d->speed = s->speed;
        if (sf & PODF_DEST_ANGLE)
            d->destAngle = s->destAngle;
        if (sf & PODF_ANGSPEED)
            d->angleSpeed = s->angleSpeed;
    }
    else if (Sv_IsSoundDelta(src)) {
        const sounddelta_t* s = (sounddelta_t const *) srcDelta;
        sounddelta_t*       d = (sounddelta_t *) destDelta;
        if (sf & SNDDF_VOLUME)
            d->volume = s->volume;
        d->sound = s->sound;
    }
    else {
        App_Error("Sv_ApplyDeltaData: Unknown delta type %i.\n", src->type);
    }
}
/**
 * Merges the second delta with the first one.
 * The source and destination must refer to the same entity.
 *
 * @return              @c false, if the result of the merge is a void delta.
 */
dd_bool Sv_MergeDelta(void* destDelta, const void* srcDelta) {
    const delta_t *src = (delta_t const *) srcDelta;
    delta_t *dest = (delta_t *) destDelta;
#ifdef _DEBUG
    if (!Sv_IsSameDelta(src, dest)) {
        App_Error("Sv_MergeDelta: Not the same!\n");
    }
    if (dest->state != DELTA_NEW) {
        App_Error("Sv_MergeDelta: Dest is not NEW.\n");
    }
#endif
    if (Sv_IsNullMobjDelta(dest)) {
        return true;
    }
    if (Sv_IsNullMobjDelta(src)) {
        dest->flags = MDFC_NULL;
        return true;
    }
    if (Sv_IsCreateMobjDelta(dest) && Sv_IsNullMobjDelta(src)) {
        dest->flags = 0;
        return false;
    }
    if (Sv_IsStartSoundDelta(src) || Sv_IsStopSoundDelta(src)) {
        sounddelta_t *destSound = (sounddelta_t *) destDelta;
        const sounddelta_t* srcSound = (sounddelta_t const *) srcDelta;
        dest->flags = src->flags;
        destSound->sound = srcSound->sound;
        destSound->mobj = srcSound->mobj;
        destSound->volume = srcSound->volume;
        return true;
    }
    dest->flags |= src->flags;
    /* dest->timeStamp = src->timeStamp; */
    Sv_ApplyDeltaData(dest, src);
    return true;
}
/**
 * @return              The age of the delta, in milliseconds.
 */
uint Sv_DeltaAge(const delta_t* delta) {
    return Sv_GetTimeStamp() - delta->timeStamp;
}
/**
 * Approximate the distance to the given sector. Set 'mayBeGone' to true
 * if the mobj may have been destroyed and should not be processed.
 */
coord_t Sv_MobjDistance(mobj_t const *mo, ownerinfo_t const *info, dd_bool isReal) {
    coord_t z;
    if (isReal && !Mobj_Map(*mo).thinkers().isUsedMobjId(mo->thinker.id)) {
        return DDMAXFLOAT;
    }
    z = mo->origin[VZ];
    if (!isReal) {
        if (z == DDMINFLOAT)
            z = mo->floorZ;
        if (z == DDMAXFLOAT)
            z = mo->ceilingZ - mo->height;
    }
    return M_ApproxDistance3(info->origin[VX] - mo->origin[VX],
                             info->origin[VY] - mo->origin[VY],
                             (info->origin[VZ] - z + mo->height / 2) * 1.2);
}
/**
 * Approximate the distance to the given sector.
 */
coord_t Sv_SectorDistance(int index, ownerinfo_t const *info) {
    DENG2_ASSERT(info);
    Sector const &sector = worldSys().map().sector(index);
    return M_ApproxDistance3(info->origin[0] - sector.soundEmitter().origin[0],
                             info->origin[1] - sector.soundEmitter().origin[1],
                             (info->origin[2] - sector.soundEmitter().origin[2]) * 1.2);
}
coord_t Sv_SideDistance(int index, int deltaFlags, ownerinfo_t const *info) {
    DENG2_ASSERT(info);
    LineSide const *side = worldSys().map().sidePtr(index);
    SoundEmitter const &emitter = (  deltaFlags & SNDDF_SIDE_MIDDLE? side->middleSoundEmitter()
                                   : deltaFlags & SNDDF_SIDE_TOP   ? side->topSoundEmitter()
                                                                   : side->bottomSoundEmitter());
    return M_ApproxDistance3(info->origin[0]  - emitter.origin[0],
                             info->origin[1]  - emitter.origin[1],
                             (info->origin[2] - emitter.origin[2]) * 1.2);
}
/**
 * @return  The distance to the origin of the delta's entity.
 */
coord_t Sv_DeltaDistance(void const *deltaPtr, ownerinfo_t const *info) {
    delta_t const *delta = (delta_t const *) deltaPtr;
    if (delta->type == DT_MOBJ) {
        return Sv_MobjDistance(&((mobjdelta_t *) deltaPtr)->mo, info, false);
    }
    if (delta->type == DT_PLAYER) {
        mobj_t const *mo = DD_Player(delta->id)->publicData().mo;
        if (mo) {
            return Sv_MobjDistance(mo, info, true);
        }
    }
    if (delta->type == DT_SECTOR) {
        return Sv_SectorDistance(delta->id, info);
    }
    if (delta->type == DT_SIDE) {
        LineSide *side = worldSys().map().sidePtr(delta->id);
        Line &line = side->line();
        return M_ApproxDistance(info->origin[0] - line.center().x,
                                info->origin[1] - line.center().y);
    }
    if (delta->type == DT_POLY) {
        Polyobj const &pob = worldSys().map().polyobj(delta->id);
        return M_ApproxDistance(info->origin[0] - pob.origin[0],
                                info->origin[1] - pob.origin[1]);
    }
    if (delta->type == DT_MOBJ_SOUND) {
        sounddelta_t const *sound = (sounddelta_t const *) deltaPtr;
        return Sv_MobjDistance(sound->mobj, info, true);
    }
    if (delta->type == DT_SECTOR_SOUND) {
        return Sv_SectorDistance(delta->id, info);
    }
    if (delta->type == DT_SIDE_SOUND) {
        return Sv_SideDistance(delta->id, delta->flags, info);
    }
    if (delta->type == DT_POLY_SOUND) {
        Polyobj const &pob = worldSys().map().polyobj(delta->id);
        return M_ApproxDistance(info->origin[VX] - pob.origin[VX],
                                info->origin[VY] - pob.origin[VY]);
    }
    return 1;
}
/**
 * The hash function for the pool delta hash.
 */
deltalink_t* Sv_PoolHash(pool_t* pool, int id) {
    return &pool->hash[(uint) id & POOL_HASH_FUNCTION_MASK];
}
/**
 * The delta is removed from the pool's delta hash.
 */
void Sv_RemoveDelta(pool_t* pool, void* deltaPtr) {
    delta_t*            delta = (delta_t *) deltaPtr;
    deltalink_t*        hash = Sv_PoolHash(pool, delta->id);
    if (hash->last == delta) {
        hash->last = delta->prev;
    }
    if (hash->first == delta) {
        hash->first = delta->next;
    }
    if (delta->next) {
        delta->next->prev = delta->prev;
    }
    if (delta->prev) {
        delta->prev->next = delta->next;
    }
    Z_Free(delta);
}
/**
 * Draining the pool means emptying it of all contents. (Doh?)
 */
void Sv_DrainPool(uint clientNumber) {
    pool_t*             pool = Sv_GetPool(clientNumber);
    delta_t*            delta;
    misrecord_t*        mis;
    void*               next = NULL;
    int                 i;
    pool->owner = clientNumber;
    pool->setDealer = 0;
    pool->resendDealer = 0;
    Sv_PoolQueueClear(pool);
    for (i = 0; i < POOL_HASH_SIZE; ++i) {
        for (delta = (delta_t *) pool->hash[i].first; delta; delta = (delta_t *) next) {
            next = delta->next;
            Z_Free(delta);
        }
    }
    for (i = 0; i < POOL_MISSILE_HASH_SIZE; ++i) {
        for (mis = (misrecord_t *) pool->misHash[i].first; mis; mis = (misrecord_t *) next) {
            next = mis->next;
            Z_Free(mis);
        }
    }
    de::zap(pool->hash);
    de::zap(pool->misHash);
}
/**
 * Returns the maximum distance for the sound. If the origin is any farther,
 * the delta will not be sent to the client in question.
 */
dfloat Sv_GetMaxSoundDistance(sounddelta_t const *delta) {
    dfloat volume = 1;
    if (delta->delta.flags & SNDDF_VOLUME) {
        volume = delta->volume;
    }
    if (volume <= 0) {
        return DDMAXFLOAT;
    }
    return volume * ::soundMaxDist;
}
/**
 * @return  The flags that remain after exclusion.
 */
int Sv_ExcludeDelta(pool_t* pool, const void* deltaPtr) {
    const delta_t*      delta = (delta_t const *) deltaPtr;
    player_t*           plr = DD_Player(pool->owner);
    mobj_t*             poolViewer = plr->publicData().mo;
    int                 flags = delta->flags;
    if (delta->type == DT_MOBJ) {
        const mobjdelta_t *mobjDelta = (mobjdelta_t const *) deltaPtr;
        if (poolViewer && poolViewer->thinker.id == delta->id) {
            flags &= ~MDF_CAMERA_EXCLUDE;
            flags &= ~MDF_ORIGIN;
            flags &= ~MDF_MOM;
            flags &= ~MDF_ANGLE;
        }
        if (mobjDelta->mo.ddFlags & DDMF_MISSILE) {
            if (Sv_IsNullMobjDelta(delta)) {
                Sv_MRRemove(pool, delta->id);
            }
            else if (!Sv_IsCreateMobjDelta(delta)) {
                flags &= ~Sv_MRCheck(pool, mobjDelta);
            }
        }
    }
    else if (delta->type == DT_PLAYER) {
        if (pool->owner == delta->id) {
            flags &= ~PDF_CAMERA_EXCLUDE;
            /* $unifiedangles */
            /*
            if (!(player->flags & DDPF_FIXANGLES)) {
                flags &= ~(PDF_CLYAW | PDF_CLPITCH);
            }
             */
        }
        else {
            flags &= ~PDF_NONCAMERA_EXCLUDE;
        }
    }
    else if (Sv_IsSoundDelta(delta)) {
        if (Sv_DeltaDistance(delta, &pool->ownerInfo) >
           Sv_GetMaxSoundDistance((sounddelta_t const *) deltaPtr)) {
            return 0;
        }
    }
    return flags;
}
/**
 * When adding a delta to the pool, it subtracts from the unacked deltas
 * there and is merged with matching new deltas. If a delta becomes void
 * after subtraction, it's removed from the pool. All the processing is
 * done based on the ID number of the delta (and type), so to make things
 * more efficient, a hash table is used (key is ID).
 *
 * Deltas are unique only in the NEW state. There may be multiple UNACKED
 * deltas for the same entity.
 *
 * The contents of the delta must not be modified.
 */
void Sv_AddDelta(pool_t* pool, void* deltaPtr) {
    delta_t*            iter, *next = NULL, *existingNew = NULL;
    delta_t*            delta = (delta_t *) deltaPtr;
    deltalink_t*        hash = Sv_PoolHash(pool, delta->id);
    int                 flags, originalFlags;
    flags = Sv_ExcludeDelta(pool, delta);
    if (!flags) {
        return;
    }
    originalFlags = delta->flags;
    delta->flags = flags;
    for (iter = hash->first; iter; iter = next) {
        next = iter->next;
        if (Sv_IsSameDelta(iter, delta)) {
            if (iter->state == DELTA_NEW) {
                existingNew = iter;
            }
            else if (iter->state == DELTA_UNACKED) {
                Sv_SubtractDelta(iter, delta);
                if (Sv_IsVoidDelta(iter)) {
                    Sv_RemoveDelta(pool, iter);
                    continue;
                }
            }
        }
    }
    if (existingNew) {
        if (!Sv_MergeDelta(existingNew, delta)) {
            Sv_RemoveDelta(pool, existingNew);
        }
    }
    else {
        iter = (delta_t *) Sv_CopyDelta(delta);
        if (hash->last) {
            hash->last->next = iter;
            iter->prev = hash->last;
        }
        hash->last = iter;
        if (!hash->first) {
            hash->first = iter;
        }
    }
    delta->flags = originalFlags;
}
/**
 * Add the delta to all the pools in the NULL-terminated array.
 */
void Sv_AddDeltaToPools(void* deltaPtr, pool_t** targets) {
    for (; *targets; targets++) {
        Sv_AddDelta(*targets, deltaPtr);
    }
}
/**
 * All NEW deltas for the mobj are removed from the pool as obsolete.
 */
void Sv_PoolMobjRemoved(pool_t* pool, thid_t id) {
    deltalink_t*        hash = Sv_PoolHash(pool, id);
    delta_t*            delta, *next = NULL;
    for (delta = hash->first; delta; delta = next) {
        next = delta->next;
        if (delta->state == DELTA_NEW && delta->type == DT_MOBJ &&
           delta->id == id) {
            Sv_RemoveDelta(pool, delta);
        }
    }
    Sv_MRRemove(pool, id);
}
/**
 * This is called when a mobj is removed in a predictable fashion.
 * (Mobj state is NULL when it's destroyed. Assumption: The NULL state
 * is set only when animation reaches its end.) Because the register-mobj
 * is removed, no Null Mobj delta is generated for the mobj.
 */
void Sv_MobjRemoved(thid_t id) {
    uint                i;
    reg_mobj_t*         mo = Sv_RegisterFindMobj(&worldRegister, id);
    if (mo) {
        Sv_RegisterRemoveMobj(&worldRegister, mo);
        for (i = 0; i < DDMAXPLAYERS; ++i) {
            if (DD_Player(i)->isConnected()) {
                Sv_PoolMobjRemoved(Sv_GetPool(i), id);
            }
        }
    }
}
/**
 * When a player leaves the game, his data is removed from the register.
 * Otherwise he'll not get all the data if he reconnects before the map
 * is changed.
 */
void Sv_PlayerRemoved(uint playerNumber) {
    de::zap(worldRegister.ddPlayers[playerNumber]);
}
/**
 * @return              @c true, if the pool is in the targets array.
 */
dd_bool Sv_IsPoolTargeted(pool_t* pool, pool_t** targets) {
    for (; *targets; targets++) {
        if (pool == *targets)
            return true;
    }
    return false;
}
/**
 * Fills the array with pointers to the pools of the connected clients,
 * if specificClient is < 0.
 *
 * @return              The number of pools in the list.
 */
int Sv_GetTargetPools(pool_t** targets, int clientsMask) {
    int i, numTargets = 0;
    for (i = 0; i < DDMAXPLAYERS; ++i) {
        if (clientsMask & (1 << i) && DD_Player(i)->isConnected()) {
            targets[numTargets++] = Sv_GetPool(i);
        }
    }
/*
    if (specificClient & SVSF_ specificClient >= 0) {
        targets[0] = &pools[specificClient];
        targets[1] = NULL;
        return 1;
    }
    for (i = 0; i < DDMAXPLAYERS; ++i) {
        if (clients[i].connected) {
            if (specificClient == SV_TARGET_ALL_POOLS || i != -specificClient)
                targets[numTargets++] = &pools[i];
        }
    }
*/
    targets[numTargets] = NULL;
    return numTargets;
}
/**
 * Null deltas are generated for mobjs that have been destroyed.
 * The register's mobj hash is scanned to see which mobjs no longer exist.
 *
 * When updating, the destroyed mobjs are removed from the register.
 */
void Sv_NewNullDeltas(cregister_t *reg, dd_bool doUpdate, pool_t **targets) {
    int i;
    mobjhash_t *hash;
    reg_mobj_t *obj, *next = 0;
    mobjdelta_t null;
    for (i = 0, hash = reg->mobjs; i < REG_MOBJ_HASH_SIZE; ++i, hash++) {
        for (obj = hash->first; obj; obj = next) {
            next = obj->next;
            if (!worldSys().map().thinkers().isUsedMobjId(obj->mo.thinker.id)) {
                Sv_NewDelta(&null, DT_MOBJ, obj->mo.thinker.id);
                null.delta.flags = MDFC_NULL;
                memcpy(&null.mo, &obj->mo, sizeof(dt_mobj_t));
                Sv_AddDeltaToPools(&null, targets);
                if (doUpdate) {
                    Sv_RegisterRemoveMobj(reg, obj);
                }
            }
        }
    }
}
/**
 * Mobj deltas are generated for all mobjs that have changed.
 */
void Sv_NewMobjDeltas(cregister_t *reg, dd_bool doUpdate, pool_t **targets) {
    worldSys().map().thinkers().forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker),
                                       0x1 /*public*/, [&reg, &doUpdate, &targets] (thinker_t *th)
    {
        auto &mob = *reinterpret_cast<mobj_t *>(th);
        if (!Sv_IsMobjIgnored(mob)) {
            mobjdelta_t delta;
            if (Sv_RegisterCompareMobj(reg, &mob, &delta)) {
                Sv_AddDeltaToPools(&delta, targets);
                if (doUpdate) {
                    Sv_RegisterMobj(&Sv_RegisterAddMobj(reg, mob.thinker.id)->mo, &mob);
                }
            }
        }
        return LoopContinue;
    });
}
/**
 * Player deltas are generated for changed player data.
 */
void Sv_NewPlayerDeltas(cregister_t* reg, dd_bool doUpdate, pool_t** targets) {
    playerdelta_t player;
    uint i;
    for (i = 0; i < DDMAXPLAYERS; ++i) {
        if (Sv_IsPlayerIgnored(i)) continue;
        if (Sv_RegisterComparePlayer(reg, i, &player)) {
            if (doUpdate && (player.delta.flags & PDF_MOBJ)) {
                reg_mobj_t* registered = Sv_RegisterFindMobj(reg, reg->ddPlayers[i].mobj);
                if (registered) {
                    Sv_RegisterResetMobj(&registered->mo);
                }
            }
            Sv_AddDeltaToPools(&player, targets);
        }
        if (doUpdate) {
            Sv_RegisterPlayer(&reg->ddPlayers[i], i);
        }
        if (Sv_IsPoolTargeted(Sv_GetPool(i), targets)) {
#if 0
            if (DD_Player(i).flags & DDPF_FIXANGLES) {
                Sv_NewDelta(&player, DT_PLAYER, i);
                Sv_RegisterPlayer(&player.player, i);
                Sv_AddDelta(&pools[i], &player);
                DD_Player(i).flags &= ~DDPF_FIXANGLES;
            }
            if (DD_Player(i).mo && (DD_Player(i).flags & (DDPF_FIXORIGIN | DDPF_FIXMOM))) {
                const mobj_t *mo = DD_Player(i).mo;
                mobjdelta_t mobj;
                Sv_NewDelta(&mobj, DT_MOBJ, mo->thinker.id);
                Sv_RegisterMobj(&mobj.mo, mo);
                if (DD_Player(i).flags & DDPF_FIXORIGIN) {
                    mobj.delta.flags |= MDF_ORIGIN;
                }
                if (DD_Player(i).flags & DDPF_FIXMOM) {
                    mobj.delta.flags |= MDF_MOM;
                }
                Sv_AddDelta(&pools[i], &mobj);
                DD_Player(i).flags &= ~(DDPF_FIXORIGIN | DDPF_FIXMOM);
            }
#endif
        }
    }
}
/**
 * Sector deltas are generated for changed sectors.
 */
void Sv_NewSectorDeltas(cregister_t *reg, dd_bool doUpdate, pool_t **targets) {
    sectordelta_t delta;
    for (int i = 0; i < worldSys().map().sectorCount(); ++i) {
        if (Sv_RegisterCompareSector(reg, i, &delta, doUpdate)) {
            Sv_AddDeltaToPools(&delta, targets);
        }
    }
}
/**
 * Side deltas are generated for changed sides (and line flags).
 * Changes in sides (textures) are so rare that all sides need not be
 * checked on every tic.
 */
void Sv_NewSideDeltas(cregister_t *reg, dd_bool doUpdate, pool_t **targets) {
    static uint numShifts = 2, shift = 0;
    world::Map &map = worldSys().map();
    uint start, end;
    if (reg->isInitial) {
        start = 0;
        end = map.sideCount();
    }
    else {
        start = shift * map.sideCount() / numShifts;
        end = ++shift * map.sideCount() / numShifts;
        shift %= numShifts;
    }
    sidedelta_t delta;
    for (uint i = start; i < end; ++i) {
        if (Sv_RegisterCompareSide(reg, i, &delta, doUpdate)) {
            Sv_AddDeltaToPools(&delta, targets);
        }
    }
}
/**
 * Poly deltas are generated for changed polyobjs.
 */
void Sv_NewPolyDeltas(cregister_t *reg, dd_bool doUpdate, pool_t **targets) {
    LOG_AS("Sv_NewPolyDeltas");
    polydelta_t delta;
    for (int i = 0; i < worldSys().map().polyobjCount(); ++i) {
        if (Sv_RegisterComparePoly(reg, i, &delta)) {
            LOGDEV_NET_XVERBOSE_DEBUGONLY("Change in poly %i", i);
            Sv_AddDeltaToPools(&delta, targets);
        }
        if (doUpdate) {
            Sv_RegisterPoly(&reg->polyObjs[i], i);
        }
    }
}
void Sv_NewSoundDelta(int soundId, mobj_t const *emitter, Sector *sourceSector,
    Polyobj *sourcePoly, Plane *sourcePlane, Surface *sourceSurface,
    float volume, dd_bool isRepeating, int clientsMask) {
    pool_t *targets[DDMAXPLAYERS + 1];
    sounddelta_t soundDelta;
    int type = DT_SOUND, df = 0;
    int id = soundId;
    Sv_GetTargetPools(targets, clientsMask);
    if (sourceSector) {
        type = DT_SECTOR_SOUND;
        id = sourceSector->indexInMap();
    }
    else if (sourcePoly) {
        type = DT_POLY_SOUND;
        id = sourcePoly->indexInMap();
    }
    else if (sourcePlane) {
        type = DT_SECTOR_SOUND;
        if (emitter && emitter == (mobj_t const *) &sourcePlane->soundEmitter()) {
            if (sourcePlane->isSectorFloor()) {
                df |= SNDDF_PLANE_FLOOR;
            }
            else if (sourcePlane->isSectorCeiling()) {
                df |= SNDDF_PLANE_CEILING;
            }
        }
        id = sourcePlane->sector().indexInMap();
    }
    else if (sourceSurface) {
        DENG_ASSERT(sourceSurface->parent().type() == DMU_SIDE);
        DENG2_ASSERT(emitter == 0); // surface sound emitter rather than a real mobj
        type = DT_SIDE_SOUND;
        LineSide &side = sourceSurface->parent().as<LineSide>();
        if (&side.middle() == sourceSurface) {
            df |= SNDDF_SIDE_MIDDLE;
        }
        else if (&side.bottom() == sourceSurface) {
            df |= SNDDF_SIDE_BOTTOM;
        }
        else if (&side.top() == sourceSurface) {
            df |= SNDDF_SIDE_TOP;
        }
        id = side.indexInMap();
    }
    else if (emitter) {
        type = DT_MOBJ_SOUND;
        id = emitter->thinker.id;
        soundDelta.mobj = emitter;
    }
    Sv_NewDelta(&soundDelta, (deltatype_t) type, id);
    df |= SNDDF_VOLUME;
    soundDelta.volume = volume;
    if (isRepeating)
        df |= SNDDF_REPEAT;
    LOGDEV_NET_XVERBOSE("New sound delta: type=%i id=%i flags=%x", type << id << df);
    soundDelta.sound = soundId;
    soundDelta.delta.flags = df;
    Sv_AddDeltaToPools(&soundDelta, targets);
}
/**
 * Returns @c true if the client should receive frames.
 */
dd_bool Sv_IsFrameTarget(duint plrNum) {
    DENG2_ASSERT(plrNum < DDMAXPLAYERS);
    player_t const &plr = *DD_Player(plrNum);
    return (plr.publicData().inGame && plr.ready);
}
/**
 * Compare the current state of the world with the register and add the
 * deltas to all the pools, or if a specific client number is given, only
 * to its pool (done when a new client enters the game). No deltas will be
 * generated for predictable changes (state changes, linear movement...).
 *
 * @param reg           World state register.
 * @param clientNumber  Client for whom to generate deltas. < 0 = all ingame
 *                      clients should get the deltas.
 * @param doUpdate      Updating the register means that the current state
 *                      of the world is stored in the register after the
 *                      deltas have been generated.
 */
void Sv_GenerateNewDeltas(cregister_t* reg, int clientNumber, dd_bool doUpdate) {
    pool_t* targets[DDMAXPLAYERS + 1], **pool;
    Sv_GetTargetPools(targets, (clientNumber < 0 ? 0xff : (1 << clientNumber)));
    for (pool = targets; *pool; pool++) {
        Sv_UpdateOwnerInfo(*pool);
    }
    Sv_NewNullDeltas(reg, doUpdate, targets);
    Sv_NewMobjDeltas(reg, doUpdate, targets);
    Sv_NewPlayerDeltas(reg, doUpdate, targets);
    Sv_NewSectorDeltas(reg, doUpdate, targets);
    Sv_NewSideDeltas(reg, doUpdate, targets);
    Sv_NewPolyDeltas(reg, doUpdate, targets);
    if (doUpdate) {
        reg->gametic = SECONDS_TO_TICKS(gameTime);
    }
}
/**
 * This is called once for each frame, in Sv_TransmitFrame().
 */
void Sv_GenerateFrameDeltas(void) {
    Sv_GenerateNewDeltas(&worldRegister, -1, true);
}
/**
 * Clears the priority queue of the pool.
 */
void Sv_PoolQueueClear(pool_t* pool) {
    pool->queueSize = 0;
}
/**
 * Exchanges two elements in the queue.
 */
void Sv_PoolQueueExchange(pool_t* pool, int index1, int index2) {
    delta_t *temp = pool->queue[index1];
    pool->queue[index1] = pool->queue[index2];
    pool->queue[index2] = temp;
}
/**
 * Adds the delta to the priority queue. More memory is allocated for the
 * queue if necessary.
 */
void Sv_PoolQueueAdd(pool_t* pool, delta_t* delta) {
    int                 i, parent;
    if (pool->allocatedSize == pool->queueSize) {
        delta_t**           newQueue;
        pool->allocatedSize *= 2;
        if (!pool->allocatedSize) {
            pool->allocatedSize = 8;
        }
        newQueue = (delta_t **) Z_Malloc(pool->allocatedSize * sizeof(delta_t *), PU_MAP, 0);
        if (pool->queue) {
            memcpy(newQueue, pool->queue,
                   sizeof(delta_t *) * pool->queueSize);
            Z_Free(pool->queue);
        }
        pool->queue = newQueue;
    }
    i = pool->queueSize;
    pool->queue[i] = delta;
    ++pool->queueSize;
    while (i > 0) {
        parent = HEAP_PARENT(i);
        if (pool->queue[parent]->score >= delta->score)
            break;
        Sv_PoolQueueExchange(pool, parent, i);
        i = parent;
    }
}
/**
 * Extracts the delta with the highest priority from the queue.
 *
 * @return              @c NULL, if there are no more deltas.
 */
delta_t* Sv_PoolQueueExtract(pool_t* pool) {
    delta_t*            max;
    int                 i, left, right, big;
    dd_bool             isDone;
    if (!pool->queueSize) {
        return NULL;
    }
    max = pool->queue[0];
    pool->queue[0] = pool->queue[--pool->queueSize];
    i = 0;
    isDone = false;
    while (!isDone) {
        left = HEAP_LEFT(i);
        right = HEAP_RIGHT(i);
        big = i;
        if (left < pool->queueSize &&
           pool->queue[left]->score > pool->queue[i]->score) {
            big = left;
        }
        if (right < pool->queueSize &&
           pool->queue[right]->score > pool->queue[big]->score) {
            big = right;
        }
        if (big != i) {
            Sv_PoolQueueExchange(pool, i, big);
            i = big;
        }
        else {
            isDone = true;
        }
    }
    return max;
}
/**
 * Postponed deltas can't be sent yet.
 */
dd_bool Sv_IsPostponedDelta(void* deltaPtr, ownerinfo_t* info) {
    delta_t *delta = (delta_t *) deltaPtr;
    uint age = Sv_DeltaAge(delta);
    if (delta->state == DELTA_UNACKED) {
        return age < info->ackThreshold;
    }
    else if (delta->state == DELTA_NEW) {
        if (Sv_IsStopSoundDelta(delta)) {
            delta_t*            iter;
            for (iter = Sv_PoolHash(info->pool, delta->id)->first; iter;
                iter = iter->next) {
                if (iter->state == DELTA_UNACKED && Sv_IsSameDelta(iter, delta)
                   && Sv_IsStartSoundDelta(iter)) {
                    return true;
                }
            }
        }
    }
    return false;
}
/**
 * Calculate a priority score for the delta. A higher score indicates
 * greater importance.
 *
 * @return              @c true iff the delta should be included in the
 *                      queue.
 */
dd_bool Sv_RateDelta(void* deltaPtr, ownerinfo_t* info) {
    float score, size;
    coord_t distance;
    delta_t *delta = (delta_t *) deltaPtr;
    int df = delta->flags;
    uint age = Sv_DeltaAge(delta);
    float ageScoreDouble = 1.0f;
    if (Sv_IsPostponedDelta(delta, info)) {
        return false;
    }
    distance = Sv_DeltaDistance(delta, info);
    if (distance < 1)
        distance = 1;
    distance = distance * distance; // Power of two.
    score = deltaBaseScores[delta->type] / distance;
    if (Sv_IsSoundDelta(delta)) {
        ageScoreDouble = 1;
    }
    score *= 1 + age / (ageScoreDouble * 1000.0f);
    if (delta->type == DT_MOBJ) {
        const mobj_t* mo = &((mobjdelta_t *) delta)->mo;
        if (df & MDFC_CREATE)
            score *= 1.5f;
        if (df & (MDF_ORIGIN_X | MDF_ORIGIN_Y))
            score *= 1.2f;
        size = MAX_OF(mo->radius, mo->height);
        if (size < 16) {
            if (size < 2)
                size = 2;
            score *= size / 16;
        }
        else if (size > 50) {
            score *= size / 50;
        }
    }
    else if (delta->type == DT_PLAYER) {
        if (df & PDF_MOBJ)
            score *= 2;
    }
    else if (delta->type == DT_SECTOR) {
        if (df & SDF_LIGHT)
            score *= 1.2f;
        if (df &
           (SDF_FLOOR_HEIGHT | SDF_CEILING_HEIGHT | SDF_FLOOR_SPEED |
            SDF_CEILING_SPEED | SDF_FLOOR_TARGET | SDF_CEILING_TARGET)) {
            score *= 3;
        }
    }
    else if (delta->type == DT_POLY) {
        if (df & PODF_SPEED)
            score *= 1.2f;
    }
    delta->score = score;
    return (score > 0);
}
/**
 * Calculate a priority score for each delta and build the priority queue.
 * The most important deltas will be included in a frame packet.
 * A pool is rated after new deltas have been generated.
 */
void Sv_RatePool(pool_t* pool) {
#ifdef _DEBUG
    player_t*           plr = DD_Player(pool->owner);
#endif
    delta_t*            delta;
    int                 i;
#ifdef _DEBUG
    if (!plr->publicData().mo) {
        App_Error("Sv_RatePool: Player %i has no mobj.\n", pool->owner);
    }
#endif
    Sv_PoolQueueClear(pool);
    for (i = 0; i < POOL_HASH_SIZE; ++i) {
        for (delta = pool->hash[i].first; delta; delta = delta->next) {
            if (Sv_RateDelta(delta, &pool->ownerInfo)) {
                Sv_PoolQueueAdd(pool, delta);
            }
        }
    }
}
/**
 * Do special things that need to be done when the delta has been acked.
 */
void Sv_AckDelta(pool_t* pool, delta_t* delta) {
    if (Sv_IsCreateMobjDelta(delta)) {
        mobjdelta_t*        mobjDelta = (mobjdelta_t *) delta;
        if (mobjDelta->mo.ddFlags & DDMF_MISSILE) {
            Sv_MRAdd(pool, mobjDelta);
        }
    }
}
/**
 * Acknowledged deltas are removed from the pool, never to be seen again.
 * Clients ack deltas to tell the server they've received them.
 *
 * @note This is obsolete: deltas no longer need to be acknowledged as
 * they are sent over TCP.
 *
 * @param clientNumber  Client whose deltas to ack.
 * @param set           Delta set number.
 * @param resent        If nonzero, ignore 'set' and ack by resend ID.
 */
void Sv_AckDeltaSet(uint clientNumber, int set, byte resent) {
    pool_t  *pool = Sv_GetPool(clientNumber);
    delta_t *delta, *next = NULL;
    for (int i = 0; i < POOL_HASH_SIZE; ++i) {
        for (delta = pool->hash[i].first; delta; delta = next) {
            next = delta->next;
            if (delta->state == DELTA_UNACKED &&
               ((!resent && delta->set == set) ||
                (resent && delta->resend == resent))) {
                Sv_AckDelta(pool, delta);
                Sv_RemoveDelta(pool, delta);
            }
        }
    }
}
/**
 * Debugging metric.
 */
uint Sv_CountUnackedDeltas(uint clientNumber) {
    uint                i, count;
    pool_t*             pool = Sv_GetPool(clientNumber);
    delta_t*            delta;
    count = 0;
    for (i = 0; i < POOL_HASH_SIZE; ++i) {
        for (delta = pool->hash[i].first; delta; delta = delta->next) {
            if (delta->state == DELTA_UNACKED)
               ++count;
        }
    }
    return count;
}
static inline dd_bool isRealMobj(mobj_t const *base) {
    return base && base->thinker.id != 0;
}
/**
 * Find the map object to whom @a base belongs.
 */
static void identifySoundEmitter(mobj_t const **base, Sector **sector, Polyobj **poly,
                                 Plane **plane, Surface **surface) {
    *sector  = nullptr;
    *poly    = nullptr;
    *plane   = nullptr;
    *surface = nullptr;
    if (!*base || isRealMobj(*base)) return;
    App_World().map().identifySoundEmitter(*reinterpret_cast<SoundEmitter const *>(*base),
                                           sector, poly, plane, surface);
#ifdef DENG2_DEBUG
    if (!*sector && !*poly && !*plane && !*surface) {
        throw Error("Sv_IdentifySoundBase", "Bad sound base");
    }
#endif
    *base = nullptr;
}
void Sv_Sound(dint soundId, mobj_t const *origin, dint toPlr) {
    Sv_SoundAtVolume(soundId, origin, 1, toPlr);
}
void Sv_SoundAtVolume(dint soundIDAndFlags, mobj_t const *origin, dfloat volume, dint toPlr) {
    if (::isClient) return;
    dint soundID = (soundIDAndFlags & ~DDSF_FLAG_MASK);
    if (!soundID) return;
    Sector *sector;
    Polyobj *poly;
    Plane *plane;
    Surface *surface;
    identifySoundEmitter(&origin, &sector, &poly, &plane, &surface);
    dint targetPlayers = 0;
    if (toPlr & SVSF_TO_ALL) {
        targetPlayers = -1;
    }
    else {
        targetPlayers = (1 << (toPlr & 0xf));
    }
    if (toPlr & SVSF_EXCLUDE_ORIGIN) {
        if (origin && origin->dPlayer) {
            targetPlayers &= ~(1 << P_GetDDPlayerIdx(origin->dPlayer));
        }
    }
    LOGDEV_NET_XVERBOSE("Sv_SoundAtVolume: id: #%i volume: %f targets: %x",
                        soundID << volume << targetPlayers);
    Sv_NewSoundDelta(soundID, origin, sector, poly, plane, surface, volume,
                     (soundIDAndFlags & DDSF_REPEAT) != 0, targetPlayers);
}
void Sv_StopSound(dint soundId, mobj_t const *origin) {
    if (::isClient) return;
    Sector *sector;
    Polyobj *poly;
    Plane *plane;
    Surface *surface;
    identifySoundEmitter(&origin, &sector, &poly, &plane, &surface);
    LOGDEV_NET_XVERBOSE("Sv_StopSound: id: #%i origin: %i(%p) sec: %p poly: %p plane: %p surface: %p",
                        soundId << (origin? origin->thinker.id : 0)
                        << origin << sector << poly << plane << surface);
    Sv_NewSoundDelta(soundId, origin, sector, poly, plane, surface, 0/*silence*/,
                     false/*non-repeating*/, -1/*all clients*/);
}
