﻿/** @file async.cpp  Asynchronous utilities.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Async"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Guard"
#include "de/Lockable"
#include "de/ReadWriteLockable"
/*
 * The Doomsday Engine Project -- libcore
 *
 * @authors Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Lockable"
#include "de/error.h"
/** @file readwritelockable.cpp  Read-write lock.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ReadWriteLockable"
#include <QReadWriteLock>
/** @file task.cpp  Concurrent task.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Task"
#include "de/TaskPool"
#include "de/Log"
/** @file taskpool.cpp
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TaskPool"
#include "de/Task"
#include "de/Guard"
#include <QThreadPool>
#include <QSet>
#include <de/Lockable>
#include <de/Loop>
#include <de/Waitable>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Waitable"
#include "de/Time"
using namespace de;
namespace de {
AsyncScope::~AsyncScope() {
    DENG2_GUARD(_tasks);
    foreach (AsyncTask *task, _tasks.value) {
        task->invalidate();
    }
}
AsyncScope &AsyncScope::operator += (AsyncTask *task) {
    if (task) { {
            DENG2_GUARD(_tasks);
            _tasks.value.insert(task);
        }
        QObject::connect(task, &QObject::destroyed, [this] (QObject *obj) {
            DENG2_GUARD(_tasks);
            _tasks.value.remove(static_cast<AsyncTask *>(obj));
        });
    }
    return *this;
}
bool AsyncScope::isAsyncFinished() const {
    DENG2_GUARD(_tasks);
    return _tasks.value.isEmpty();
}
void AsyncScope::waitForFinished(TimeSpan timeout) {
    forever {
        AsyncTask *task = nullptr; {
            DENG2_GUARD(_tasks);
            if (!_tasks.value.isEmpty()) {
                auto i = _tasks.value.begin();
                task = *i;
                _tasks.value.erase(i);
            }
        }
        if (!task) break;
        task->wait(timeout.asMilliSeconds());
    }
}
} // namespace de
namespace de {
} // namespace de
namespace de {
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(ReadWriteLockable) {
    QReadWriteLock lock;
    Impl() : lock(QReadWriteLock::Recursive)
    {}
    ~Impl() {
        lock.lockForWrite();
        lock.unlock();
    }
};
ReadWriteLockable::ReadWriteLockable() : d(new Impl)
{}
ReadWriteLockable::~ReadWriteLockable()
{}
void ReadWriteLockable::lockForRead() const {
    d->lock.lockForRead();
}
void ReadWriteLockable::lockForWrite() const {
    d->lock.lockForWrite();
}
void ReadWriteLockable::unlock() const {
    d->lock.unlock();
}
} // namespace de
namespace de {
Task::Task() : _pool(nullptr)
{}
void Task::run() {
    try {
        runTask();
    }
    catch (Error const &er) {
        LOG_AS("Task");
        LOG_WARNING("Aborted due to exception: ") << er.asText();
    }
    if (_pool) _pool->taskFinishedRunning(*this);
}
} // namespace de
namespace de {
namespace internal {
    class CallbackTask : public Task {
    public:
        CallbackTask(TaskPool::TaskFunction func) : _func(func) {}
        void runTask() override { _func(); }
    private:
        TaskPool::TaskFunction _func;
    };
}
DENG2_PIMPL(TaskPool), public Lockable, public Waitable, public TaskPool::IPool {
    bool deleteWhenDone = false;
    QSet<Task *> tasks;
    Impl(Public *i) : Base(i) {
        post();
    }
    ~Impl() {
        DENG2_ASSERT(tasks.isEmpty());
    }
    void add(Task *t) {
        DENG2_GUARD(this);
        t->_pool = this;
        if (tasks.isEmpty()) {
            wait(); // Semaphore now unavailable.
        }
        tasks.insert(t);
    }
    bool remove(Task *t) {
        DENG2_GUARD(this);
        tasks.remove(t);
        if (tasks.isEmpty()) {
            post();
            return true;
        }
        return false;
    }
    void waitForEmpty() const {
        wait();
        DENG2_GUARD(this);
        DENG2_ASSERT(tasks.isEmpty());
        post(); // When empty, the semaphore is available.
    }
    bool isEmpty() const {
        DENG2_GUARD(this);
        return tasks.isEmpty();
    }
    void taskFinishedRunning(Task &task) {
        lock();
        if (remove(&task)) {
            if (deleteWhenDone) {
                unlock();
                delete this;
                return;
            }
            else {
                try {
                    emit self().allTasksDone();
                    DENG2_FOR_AUDIENCE(Done, i) i->taskPoolDone(self());
                }
                catch (Error const &er) {
                    unlock();
                    throw er;
                }
            }
        }
        unlock();
    }
    DENG2_PIMPL_AUDIENCE(Done)
};
DENG2_AUDIENCE_METHOD(TaskPool, Done)
TaskPool::TaskPool() : d(new Impl(this))
{}
TaskPool::~TaskPool() {
    DENG2_GUARD(d);
    if (!d->isEmpty()) {
        d.release()->deleteWhenDone = true;
    }
}
void TaskPool::start(Task *task, Priority priority) {
    d->add(task);
    QThreadPool::globalInstance()->start(task, int(priority));
}
void TaskPool::start(TaskFunction taskFunction, Priority priority) {
    start(new internal::CallbackTask(taskFunction), priority);
}
void TaskPool::waitForDone() {
    d->waitForEmpty();
}
bool TaskPool::isDone() const {
    return d->isEmpty();
}
} // namespace de
Waitable::Waitable(duint initialValue) : _semaphore(initialValue)
{}
Waitable::~Waitable()
{}
void Waitable::reset() {
    _semaphore.tryAcquire(_semaphore.available());
}
void Waitable::wait() const {
    wait(0.0);
}
void Waitable::wait(TimeSpan const &timeOut) const {
    if (!tryWait(timeOut)) {
        throw WaitError("Waitable::wait", "Timed out");
    }
}
bool Waitable::tryWait(TimeSpan const &timeOut) const {
    if (timeOut <= 0.0) {
        _semaphore.acquire();
    }
    else {
        if (!_semaphore.tryAcquire(1, int(timeOut.asMilliSeconds()))) {
            return false;
        }
    }
    return true;
}
void Waitable::post() const {
    _semaphore.release();
}
