﻿/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Animation"
#include "de/App"
#include "de/ArchiveFeed"
#include "de/ArchiveFolder"
#include "de/ArrayValue"
#include "de/Block"
#include "de/CommandLine"
#include "de/Config"
#include "de/DictionaryValue"
#include "de/DirectoryFeed"
#include "de/EscapeParser"
#include "de/FileLogSink"
#include "de/FileSystem"
#include "de/LibraryFile"
#include "de/Log"
#include "de/LogBuffer"
#include "de/LogFilter"
#include "de/Loop"
#include "de/math.h"
#include "de/MetadataBank"
#include "de/Module"
#include "de/NativeFile"
#include "de/NumberValue"
#include "de/PackageFeed"
#include "de/PackageLoader"
#include "de/Record"
#include "de/RemoteFeedRelay"
#include "de/ScriptSystem"
#include "de/StaticLibraryFeed"
#include "de/TextValue"
#include "de/UnixInfo"
#include "de/Version"
#include "de/Writer"
#include "de/ZipArchive"
#include <QDir>
#include <QThread>
#ifdef UNIX
#  include <locale.h>
#endif
/** @file asset.cpp  Information about the state of an asset (e.g., resource).
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Asset"
#include "de/Waitable"
/**
 * @file callbacktimer.cpp
 * Internal helper class for making callbacks to C code.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "../src/core/callbacktimer.h"
/** @file clock.cpp Time source.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Clock"
#include <atomic>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/CommandLine"
#include "de/String"
#include "de/NativePath"
#include "de/Log"
#include "de/App"
#include <QFile>
#include <QDir>
#include <QDebug>
#include <fstream>
#include <sstream>
#include <cctype>
#include <string.h>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Config"
#include "de/App"
#include "de/Archive"
#include "de/ArrayValue"
#include "de/CommandLine"
#include "de/Folder"
#include "de/Log"
#include "de/NumberValue"
#include "de/Package"
#include "de/Reader"
#include "de/Refuge"
#include "de/Version"
#include "de/Writer"
/** @file debuglogsink.cpp  Log sink that uses QDebug for output.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DebugLogSink"
/** @file filelogsink.cpp  Log sink that uses a File for output.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FileLogSink"
#include "de/FileSystem"
/** @file garbage.cpp  Garbage collector.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Garbage"
#include "de/Lockable"
#include "de/Guard"
#include "de/Log"
#include <QThread>
#include <map>
#include <set>
/** @file highperformancetimer.cpp  Timer for performance-critical use.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/HighPerformanceTimer"
#include <de/Guard>
#include <de/Lockable>
#include <QElapsedTimer>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Id"
#include "de/String"
#include "de/Writer"
#include "de/Reader"
#include <atomic>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/libcore.h"
#include "de/Library"
#include "de/Log"
#include "de/LogBuffer"
#if !defined (DENG_STATIC_LINK)
#  include <QLibrary>
#endif
#include <array>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Log"
#include "de/Time"
#include "de/Date"
#include "de/LogBuffer"
#include "de/Guard"
#include "de/Reader"
#include "de/Writer"
#include "de/FIFO"
#include "../src/core/logtextstyle.h"
#include <QMap>
#include <QTextStream>
#include <QThread>
#include <QThreadStorage>
#include <QStringList>
#include <array>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LogBuffer"
#include "de/App"
#include "de/DebugLogSink"
#include "de/FileLogSink"
#include "de/FixedByteArray"
#include "de/Folder"
#include "de/Guard"
#include "de/LogSink"
#include "de/SimpleLogFilter"
#include "de/TextStreamLogSink"
#include "de/Writer"
#include <stdio.h>
#include <QTextStream>
#include <QCoreApplication>
#include <QList>
#include <QSet>
#include <QTimer>
#include <QDebug>
/** @file logfilter.cpp  Log entry filter.
 *
 * @authors Copyright © 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LogFilter"
/** @file logsink.cpp  Sink where log entries are flushed from the LogBuffer.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LogSink"
/** @file loop.cpp
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Loop"
#include "de/App"
#include "de/Time"
#include "de/Log"
#include "de/math.h"
#include <QCoreApplication>
#include <QTimer>
#include "../src/core/callbacktimer.h"
/** @file memorylogsink.h  Log sink that stores log entries in memory.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/MemoryLogSink"
/** @file monospacelogsinkformatter.cpp  Fixed-width log entry formatter.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/MonospaceLogSinkFormatter"
#include "de/EscapeParser"
#include <de/math.h>
#include <QVector>
#include <de/Vector>
/** @file slope.cpp  Vector/slope conversions.
 *
 * @authors Copyright (c) 2019 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/core/slope.h"
#include <cmath>
/** @file system.cpp  Base class for application subsystems.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/System"
/** @file textapp.h  Application with text-based/console interface.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TextApp"
#include <de/Log>
#include <de/math.h>
#include <de/NativePath>
#include <QDir>
/** @file textstreamlogsink.cpp  Log sink that uses a QTextStream for output.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TextStreamLogSink"
/**
 * @file unixinfo.cpp
 * Unix system-level configuration.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/UnixInfo"
#include "de/Info"
#include "de/App"
#include <QDir>

namespace de {
static App *singletonApp;
static Value *Function_App_Locate(Context &, Function::ArgumentValues const &args) {
    std::unique_ptr<DictionaryValue> result(new DictionaryValue);
    String const packageId = args.first()->asText();
    if (File const *pkg = PackageLoader::get().select(packageId)) {
        result->add(new TextValue(packageId), RecordValue::takeRecord(
                        Record::withMembers(
                            "localPath",   pkg->path(),
                            "description", pkg->description()
                        )));
    }
    auto found = App::remoteFeedRelay().locatePackages(StringList({ packageId }));
    for (auto i = found.begin(); i != found.end(); ++i) {
        result->add(new TextValue(i.key()), RecordValue::takeRecord(
                        Record::withMembers(
                            "repository", i->link->address(),
                            "localPath",  i->localPath,
                            "remotePath", i->remotePath
                        )));
    }
    return result.release();
}
DENG2_PIMPL(App)
, DENG2_OBSERVES(PackageLoader, Activity) {
    QThread *mainThread = nullptr;
    String appName;
    CommandLine cmdLine;
    LogFilter logFilter;
    LogBuffer logBuffer;
    NativePath appPath;
    String unixHomeFolder;
    NativePath cachedBasePath;
    NativePath cachedPluginBinaryPath;
    NativePath cachedHomePath;
    Clock clock;
    QList<System *> systems;
    ScriptSystem scriptSys;
    Record appModule;
    Binder binder;
    FileSystem fs;
    std::unique_ptr<MetadataBank> metaBank;
    std::unique_ptr<NativeFile> basePackFile;
    Archive *persistentData;
    std::unique_ptr<UnixInfo> unixInfo;
    filesys::RemoteFeedRelay remoteFeedRelay;
    Path configPath;
    Config *config;
    StringList packagesToLoadAtInit;
    PackageLoader packageLoader;
    void (*terminateFunc)(char const *);
    std::unique_ptr<FileLogSink> errorSink;
    Impl(Public *a, QStringList args)
        : Base(a)
        , appName("Doomsday Engine")
        , cmdLine(args)
        , unixHomeFolder(".doomsday")
        , persistentData(0)
        , configPath("/packs/net.dengine.stdlib/modules/Config.ds")
        , config(0)
        , terminateFunc(0) {
        #ifdef UNIX
        {
            setenv("LC_ALL", "en_US.UTF-8", 1);
            setenv("LC_NUMERIC", "C", 1);
            setlocale(LC_ALL, "en_US.UTF-8");
            setlocale(LC_NUMERIC, "C");
        }
        #endif
        QLocale::setDefault(QLocale("en_US.UTF-8"));
        packagesToLoadAtInit << "net.dengine.stdlib";
        singletonApp = a;
        mainThread = QThread::currentThread();
        logBuffer.setEntryFilter(&logFilter);
        Clock::setAppClock(&clock);
        Animation::setClock(&clock);
        qsrand(Time().asDateTime().toTime_t());
        systems << &fs << &scriptSys;
        static LibraryFile::Interpreter intrpLibraryFile;
        static ZipArchive::Interpreter  intrpZipArchive;
        fs.addInterpreter(intrpLibraryFile);
        fs.addInterpreter(intrpZipArchive); {
            scriptSys.addNativeModule("App", appModule);
            binder.init(appModule)
                    << DENG2_FUNC(App_Locate, "locate", "packageId");
        }
    }
    ~Impl() {
        metaBank.reset();
        if (errorSink) {
            logBuffer.removeSink(*errorSink);
        }
        if (config) {
            Record *filter = new Record;
            logFilter.write(*filter);
            config->objectNamespace().add("log.filter", filter);
            delete config;
        }
        Clock::setAppClock(0);
        logBuffer.setOutputFile("");
    }
    NativePath defaultNativeModulePath() const {
        #ifdef WIN32
        {
            NativePath appDir = appPath.fileNamePath();
            return appDir / "..\\modules";
        }
        #else
        {
            return self().nativeBasePath() / "modules";
        }
        #endif
    }
    void initFileSystem(bool allowPlugins) {
        Folder::checkDefaultSettings();
        Folder &binFolder = fs.makeFolder("/bin");
        if (ZipArchive::recognize(self().nativeBasePath())) {
            basePackFile.reset(NativeFile::newStandalone(self().nativeBasePath()));
            fs.root().attach(new ArchiveFeed(*basePackFile));
        }
        else {
            #ifdef MACOSX
            {
                NativePath appDir = appPath.fileNamePath();
                binFolder.attach(new DirectoryFeed(appDir));
                fs.makeFolder("/data").attach(new DirectoryFeed(self().nativeBasePath()));
            }
            #elif WIN32
            {
                NativePath appDir = appPath.fileNamePath();
                fs.makeFolder("/data").attach(new DirectoryFeed(appDir / "..\\data"));
            }
            #else // UNIX
            {
                if ((self().nativeBasePath() / "data").exists()) {
                    fs.makeFolder("/data").attach(new DirectoryFeed(self().nativeBasePath() / "data"));
                }
                else {
                    fs.makeFolder("/data").attach(new DirectoryFeed(self().nativeBasePath()));
                }
            }
            #endif
            if (defaultNativeModulePath().exists()) {
                fs.makeFolder("/modules").attach(new DirectoryFeed(defaultNativeModulePath()));
            }
        }
        if (allowPlugins) {
#if !defined (DENG_STATIC_LINK)
            binFolder.attach(new DirectoryFeed(self().nativePluginBinaryPath()));
#else
            binFolder.attach(new StaticLibraryFeed);
#endif
        }
        fs.makeFolder("/home", FS::DontInheritFeeds).attach(new DirectoryFeed(self().nativeHomePath(),
                DirectoryFeed::AllowWrite | DirectoryFeed::CreateIfMissing | DirectoryFeed::DefaultFlags));
        fs.makeFolder("/packs").attach(new PackageFeed(packageLoader));
        fs.makeFolder("/sys").setMode(File::Write);
        metaBank.reset(new MetadataBank);
        fs.root().populate(Folder::PopulateFullTree);
        fs.makeFolder("/home/configs");
        packageLoader.audienceForActivity() += this;
    }
    void setLogLevelAccordingToOptions() {
        if (cmdLine.has("-loglevel") || cmdLine.has("-verbose") || cmdLine.has("-v") ||
            cmdLine.has("-vv") || cmdLine.has("-vvv")) {
            LogEntry::Level level = LogEntry::Message;
            try {
                int pos;
                if ((pos = cmdLine.check("-loglevel", 1)) > 0) {
                    level = LogEntry::textToLevel(cmdLine.at(pos + 1));
                }
            }
            catch (Error const &er) {
                qWarning("%s", er.asText().toLatin1().constData());
            }
            level = LogEntry::Level(level
                                    - cmdLine.has("-verbose")
                                    - cmdLine.has("-v") * (cmdLine.isAliasDefinedFor("-verbose")? 0 : 1)
                                    - cmdLine.has("-vv") * 2
                                    - cmdLine.has("-vvv") * 3);
            if (level < LogEntry::XVerbose) {
                logFilter.setAllowDev(LogEntry::AllDomains, true);
                level = LogEntry::XVerbose;
            }
            logFilter.setMinLevel(LogEntry::AllDomains, level);
        }
        if (cmdLine.has("-devlog")) {
            logFilter.setAllowDev(LogEntry::AllDomains, true);
        }
        if (cmdLine.has("-nodevlog")) {
            logFilter.setAllowDev(LogEntry::AllDomains, false);
        }
    }
    void checkForErrorDumpFile() {
        if (CommandLine::ArgWithParams arg = cmdLine.check("-errors", 1)) {
            File &errors = self().rootFolder().replaceFile(Path("/home") / arg.params.at(0));
            errorSink.reset(new FileLogSink(errors));
            errorSink->setMode(LogSink::OnlyWarningEntries);
            logBuffer.addSink(*errorSink);
        }
    }
    ArchiveFolder &persistPackFolder() {
        return self().homeFolder().locate<ArchiveFolder>("persist.pack");
    }
    void setOfLoadedPackagesChanged() {
        fs.root().locate<Folder>("/packs").populate();
    }
    Record const *findAsset(String const &identifier) const {
        Folder const *pkg = fs.root().tryLocate<Folder>("/packs/asset." + identifier + "/.");
        if (!pkg) return 0;
        String const ns = "package." + identifier;
        if (pkg->objectNamespace().has(ns)) {
            return &(*pkg)[ns].valueAsRecord();
        }
        return nullptr;
    }
    DENG2_PIMPL_AUDIENCE(StartupComplete)
};
DENG2_AUDIENCE_METHOD(App, StartupComplete)
App::App(NativePath const &appFilePath, QStringList args)
    : d(new Impl(this, args)) {
    d->unixInfo.reset(new UnixInfo);
    Animation::setClock(&d->clock);
    LogBuffer::setAppBuffer(d->logBuffer);
    d->logBuffer.enableFlushing(false);
    if (d->cmdLine.has("-stdout")) {
        d->logBuffer.enableStandardOutput(true);
        d->logBuffer.enableFlushing(true);
    }
    d->setLogLevelAccordingToOptions();
    d->appPath = appFilePath;
    LOG_NOTE("Application path: ") << d->appPath;
    LOG_NOTE("Version: ") << Version::currentBuild().asHumanReadableText();
    #if defined (MACOSX)
    {
        if (d->cmdLine.count() >= 2 && d->cmdLine.at(1).beginsWith("-psn")) {
            DirectoryFeed::changeWorkingDir(d->cmdLine.at(0).fileNamePath() + "/..");
        }
    }
    #endif
}
App::~App() {
    LOG_AS("~App");
    d.reset();
    singletonApp = 0;
}
void App::addInitPackage(String const &identifier) {
    d->packagesToLoadAtInit << identifier;
}
void App::setConfigScript(Path const &path) {
    d->configPath = path;
}
void App::setName(String const &appName) {
    d->appName = appName;
}
void App::setUnixHomeFolderName(String const &name) {
    d->unixHomeFolder = name;
    d->unixInfo.reset(new UnixInfo);
}
String App::unixHomeFolderName() const {
    return d->unixHomeFolder;
}
String App::unixEtcFolderName() const {
    if (d->unixHomeFolder.startsWith(".")) {
        return d->unixHomeFolder.mid(1);
    }
    return d->unixHomeFolder;
}
void App::setTerminateFunc(void (*func)(char const *)) {
    d->terminateFunc = func;
}
void App::handleUncaughtException(String message) {
    LOG_CRITICAL(message);
    EscapeParser esc;
    esc.parse(message);
    if (d->terminateFunc) d->terminateFunc(esc.plainText().toUtf8());
}
bool App::processEvent(Event const &ev) {
    foreach (System *sys, d->systems) {
        if (sys->behavior() & System::ReceivesInputEvents) {
            if (sys->processEvent(ev))
                return true;
        }
    }
    return false;
}
void App::timeChanged(Clock const &clock) {
    foreach (System *sys, d->systems) {
        if (sys->behavior() & System::ObservesTime) {
            sys->timeChanged(clock);
        }
    }
}
bool App::inMainThread() {
    if (!App::appExists()) {
        return true;
    }
    return DENG2_APP->d->mainThread == QThread::currentThread();
}
#if !defined (DENG_STATIC_LINK)
NativePath App::nativePluginBinaryPath() {
    if (!d->cachedPluginBinaryPath.isEmpty()) return d->cachedPluginBinaryPath;
    if (auto const opt = d->cmdLine.check("-libdir", 1)) {
        d->cmdLine.makeAbsolutePath(opt.pos + 1);
        return (d->cachedPluginBinaryPath = d->cmdLine.at(opt.pos + 1));
    }
    NativePath path;
    #if defined (WIN32)
    {
        path = d->appPath.fileNamePath() / "plugins";
        return (d->cachedPluginBinaryPath = path);
    }
    #else
    {
        #if defined (MACOSX)
        {
            path = d->appPath.fileNamePath() / "../PlugIns/Doomsday";
        }
        #else
        {
            path = d->appPath.fileNamePath() / DENG_LIBRARY_DIR;
            if (!path.exists()) {
                path = d->appPath.fileNamePath() / "../lib/doomsday";
            }
        }
        #endif
        d->unixInfo->path("libdir", path);
        return (d->cachedPluginBinaryPath = path);
    }
    #endif
}
#endif
NativePath App::nativeHomePath() {
    if (!d->cachedHomePath.isEmpty()) return d->cachedHomePath;
    if (auto const opt = d->cmdLine.check("-userdir", 1)) {
        d->cmdLine.makeAbsolutePath(opt.pos + 1);
        return (d->cachedHomePath = d->cmdLine.at(opt.pos + 1));
    }
    NativePath nativeHome;
    #if defined (DENG_IOS)
    {
        nativeHome = QDir::homePath();
        nativeHome = nativeHome / "Documents/runtime";
    }
    #elif defined (MACOSX)
    {
        nativeHome = QDir::homePath();
        nativeHome = nativeHome / "Library/Application Support" / d->appName / "runtime";
    }
    #elif defined (WIN32)
    {
        nativeHome = appDataPath();
        nativeHome = nativeHome / "runtime";
    }
    #else // UNIX
    {
        nativeHome = QDir::homePath();
        nativeHome = nativeHome / d->unixHomeFolder / "runtime";
    }
    #endif
    return (d->cachedHomePath = nativeHome);
}
Archive const &App::persistentData() {
    Archive const *persist = DENG2_APP->d->persistentData;
    if (!persist) {
        throw PersistentDataNotAvailable("App::persistentData", "Persistent data is disabled");
    }
    return *persist;
}
Archive &App::mutablePersistentData() {
    Archive *persist = DENG2_APP->d->persistentData;
    if (!persist) {
        throw PersistentDataNotAvailable("App::mutablePersistentData", "Persistent data is disabled");
    }
    return *persist;
}
bool App::hasPersistentData() {
    return DENG2_APP->d->persistentData != 0;
}
ArchiveFolder &App::persistPackFolder() {
    return DENG2_APP->d->persistPackFolder();
}
NativePath App::currentWorkPath() {
    return NativePath::workPath();
}
bool App::setCurrentWorkPath(NativePath const &cwd) {
    return NativePath::setWorkPath(cwd);
}
NativePath App::nativeBasePath() {
    if (!d->cachedBasePath.isEmpty()) return d->cachedBasePath;
    int i;
    if ((i = d->cmdLine.check("-basedir", 1))) {
        d->cmdLine.makeAbsolutePath(i + 1);
        return (d->cachedBasePath = d->cmdLine.at(i + 1));
    }
    NativePath path;
    #ifdef WIN32
    {
        path = d->appPath.fileNamePath() / "..";
    }
    #else
    {
        #ifdef MACOSX
        {
            path = d->appPath.fileNamePath() / "../Resources";
            if (!path.exists()) {
                path = d->appPath.fileNamePath() / DENG_BASE_DIR;
            }
        }
        #else
        {
            path = d->appPath.fileNamePath() / DENG_BASE_DIR;
        }
        #endif
        if (!path.exists()) {
            path = d->appPath.fileNamePath();
        }
        d->unixInfo->path("basedir", path);
    }
    #endif
    return (d->cachedBasePath = path);
}
void App::initSubsystems(SubsystemInitFlags flags) {
    bool allowPlugins = !flags.testFlag(DisablePlugins);
    d->initFileSystem(allowPlugins);
    foreach (String pkg, d->packagesToLoadAtInit) {
        d->packageLoader.load(pkg);
    }
    if (!flags.testFlag(DisablePersistentData)) {
        if (!homeFolder().has("persist.pack") || commandLine().has("-reset")) {
            ZipArchive arch;
            arch.add("Info", String(QString("# Package for %1's persistent state.\n").arg(d->appName)).toUtf8());
            File &persistPack = homeFolder().replaceFile("persist.pack");
            Writer(persistPack) << arch;
            persistPack.reinterpret()->as<ArchiveFolder>().populate();
        }
        d->persistentData = &d->persistPackFolder().archive();
    }
    d->config = new Config(d->configPath);
    d->scriptSys.addNativeModule("Config", d->config->objectNamespace());
    if (d->config->read() == Config::DifferentVersion) {
        LOG_RES_NOTE("Clearing cached metadata due to version change");
        d->metaBank->clear();
        Version oldVer = d->config->upgradedFromVersion();
        if (oldVer != Version::currentBuild()) {
            ArrayValue *old = new ArrayValue;
            *old << NumberValue(oldVer.major) << NumberValue(oldVer.minor)
                 << NumberValue(oldVer.patch) << NumberValue(oldVer.build);
            d->scriptSys["Version"].addArray("OLD_VERSION", old).setReadOnly();
        }
    }
    LogBuffer &logBuf = LogBuffer::get();
    logBuf.setMaxEntryCount(d->config->getui("log.bufferSize", 1000));
    try {
        if (CommandLine::ArgWithParams outArg = commandLine().check("-out", 1)) {
            logBuf.setOutputFile(String("/home") / outArg.params.at(0));
        }
        else if (d->config->has("log")) {
            logBuf.setOutputFile(d->config->gets("log.file"));
        }
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to set log output file:\n" + er.asText());
    }
    d->checkForErrorDumpFile();
    try {
        if (d->config->has("log.filter")) {
            d->logFilter.read(d->config->subrecord("log.filter"));
        }
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to apply log filter:\n" + er.asText());
    }
    d->setLogLevelAccordingToOptions();
    LOGDEV_NOTE("Developer log entries enabled");
    logBuf.enableFlushing(true);
    Time::updateCurrentHighPerformanceTime();
    d->clock.setTime(Time::currentHighPerformanceTime());
    d->clock.audienceForTimeChange() += this;
    if (allowPlugins) {
#if 0 // not yet handled by libcore
        loadPlugins();
#endif
    }
    LOG_VERBOSE("Subsystems initialized");
}
void App::addSystem(System &system) {
    d->systems.removeAll(&system);
    d->systems.append(&system);
}
void App::removeSystem(System &system) {
    d->systems.removeAll(&system);
}
void App::notifyStartupComplete() {
    DENG2_FOR_AUDIENCE2(StartupComplete, i) {
        i->appStartupCompleted();
    }
}
bool App::appExists() {
    return singletonApp != 0;
}
App &App::app() {
    DENG2_ASSERT(appExists());
    return *singletonApp;
}
LogFilter &App::logFilter() {
    return DENG2_APP->d->logFilter;
}
CommandLine &App::commandLine() {
    return DENG2_APP->d->cmdLine;
}
NativePath App::executablePath() {
    return DENG2_APP->d->appPath;
}
#ifdef MACOSX
NativePath App::nativeAppContentsPath() {
    return DENG2_APP->d->appPath/"../..";
}
#endif
FileSystem &App::fileSystem() {
    return DENG2_APP->d->fs;
}
MetadataBank &App::metadataBank() {
    DENG2_ASSERT(DENG2_APP->d->metaBank);
    return *DENG2_APP->d->metaBank;
}
PackageLoader &App::packageLoader() {
    return DENG2_APP->d->packageLoader;
}
int App::findInPackages(String const &partialPath, FS::FoundFiles &files) {
    return App::fileSystem().nameIndex().findPartialPathInPackageOrder(partialPath, files);
}
bool App::assetExists(String const &identifier) {
    return DENG2_APP->d->findAsset(identifier) != 0;
}
Package::Asset App::asset(String const &identifier) {
    Record const *info = DENG2_APP->d->findAsset(identifier);
    if (!info) {
        throw AssetNotFoundError("App::asset", "Asset \"" + identifier +
                                 "\" does not exist");
    }
    return *info;
}
ScriptSystem &App::scriptSystem() {
    return DENG2_APP->d->scriptSys;
}
bool App::configExists() {
    return DENG2_APP->d->config != nullptr;
}
Folder &App::rootFolder() {
    return fileSystem().root();
}
Folder &App::homeFolder() {
    return rootFolder().locate<Folder>("home");
}
filesys::RemoteFeedRelay &App::remoteFeedRelay() {
    return DENG2_APP->d->remoteFeedRelay;
}
Config &App::config() {
    DENG2_ASSERT(DENG2_APP->d->config != 0);
    return *DENG2_APP->d->config;
}
Variable &App::config(String const &name) {
    return config()[name];
}
String App::apiUrl() // static
{
    String u = Config::get().gets(QStringLiteral("apiUrl"));
    if (!u.startsWith("http")) u = "http://" + u;
    if (!u.endsWith("/")) u += "/";
    return u;
}
UnixInfo &App::unixInfo() {
    return *DENG2_APP->d->unixInfo;
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Asset) {
    State state;
    Impl(State s) : state(s) {}
    Impl(Impl const &other) : de::IPrivate(), state(other.state) {}
    DENG2_PIMPL_AUDIENCE(StateChange)
    DENG2_PIMPL_AUDIENCE(Deletion)
};
DENG2_AUDIENCE_METHOD(Asset, StateChange)
DENG2_AUDIENCE_METHOD(Asset, Deletion)
Asset::Asset(State initialState) : d(new Impl(initialState))
{}
Asset::Asset(Asset const &other) : d(new Impl(*other.d))
{}
Asset::~Asset() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->assetBeingDeleted(*this);
}
void Asset::setState(State s) {
    State const old = d->state;
    d->state = s;
    if (old != d->state) {
        DENG2_FOR_AUDIENCE2(StateChange, i) i->assetStateChanged(*this);
    }
}
void Asset::setState(bool assetReady) {
    setState(assetReady ? Ready : NotReady);
}
Asset::State Asset::state() const {
    return d->state;
}
bool Asset::isReady() const {
    return d->state == Ready;
}
void Asset::waitForState(State s) const {
    struct Waiter
        : public Waitable
        , DENG2_OBSERVES(Asset, StateChange) {
        State waitingFor;
        Waiter(State wf)
            : waitingFor(wf)
        {}
        void assetStateChanged(Asset &asset) override {
            if (asset.state() == waitingFor) {
                post();
            }
        }
    };
    Waiter w(s);
    audienceForStateChange() += w;
    if (d->state != s) {
        w.wait();
    }
}
DENG2_PIMPL_NOREF(AssetGroup) {
    Members deps;
    /**
     * Determines if all the assets in the group are ready.
     */
    bool allReady() const {
        DENG2_FOR_EACH_CONST(Members, i, deps) {
            switch (i->second) {
            case Required:
                if (!i->first->isReady()) return false;
                break;
            default:
                break;
            }
        }
        return true;
    }
    void update(AssetGroup &self) {
        self.setState(allReady()? Ready : NotReady);
    }
};
AssetGroup::AssetGroup() : d(new Impl) {
    setState(Ready);
}
AssetGroup::~AssetGroup() {
    audienceForStateChange().clear();
    clear();
}
dint AssetGroup::size() const {
    return dint(d->deps.size());
}
void AssetGroup::clear() {
    DENG2_FOR_EACH(Members, i, d->deps) {
        i->first->audienceForDeletion() -= this;
        i->first->audienceForStateChange() -= this;
    }
    d->deps.clear();
    d->update(*this);
}
void AssetGroup::insert(Asset const &asset, Policy policy) {
    d->deps[&asset] = policy;
    asset.audienceForDeletion() += this;
    asset.audienceForStateChange() += this;
    d->update(*this);
}
void AssetGroup::remove(Asset const &asset) {
    asset.audienceForDeletion() -= this;
    asset.audienceForStateChange() -= this;
    d->deps.erase(&asset);
    d->update(*this);
}
bool AssetGroup::has(Asset const &asset) const {
    return d->deps.find(&asset) != d->deps.end();
}
void AssetGroup::setPolicy(Asset const &asset, Policy policy) {
    auto found = d->deps.find(&asset);
    DENG2_ASSERT(found != d->deps.end());
    if (found->second != policy) {
        found->second = policy;
        d->update(*this);
    }
}
AssetGroup::Members const &AssetGroup::all() const {
    return d->deps;
}
void AssetGroup::assetBeingDeleted(Asset &asset) {
    if (has(asset)) {
        remove(asset);
    }
}
void AssetGroup::assetStateChanged(Asset &) {
    DENG2_ASSERT(!d.isNull());
    d->update(*this);
}
IAssetGroup::~IAssetGroup()
{}
IAssetGroup::operator Asset &() {
    return assets();
}
} // namespace de
namespace de {
namespace internal {
CallbackTimer::CallbackTimer(std::function<void ()> func, QObject *parent)
    : QTimer(parent), _func(func) {
    setSingleShot(true);
    connect(this, SIGNAL(timeout()), this, SLOT(callbackAndDeleteLater()));
}
void CallbackTimer::callbackAndDeleteLater() {
    if (_func) _func();
    this->deleteLater();
}
} // namespace internal
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Clock) {
    Time startedAt;
    Time time;
    std::atomic_uint tickCount { 0 };
    DENG2_PIMPL_AUDIENCE(TimeChange)
};
DENG2_AUDIENCE_METHOD(Clock, TimeChange)
Clock *Clock::_appClock = 0;
Clock::Clock() : d(new Impl)
{}
Clock::~Clock()
{}
void Clock::setTime(Time const &currentTime) {
    bool changed = (d->time != currentTime);
    d->time = currentTime;
    if (changed) {
        d->tickCount++;
        DENG2_FOR_EACH_OBSERVER(PriorityTimeChangeAudience, i, audienceForPriorityTimeChange) {
            i->timeChanged(*this);
        }
        DENG2_FOR_AUDIENCE2(TimeChange, i) i->timeChanged(*this);
    }
}
void Clock::advanceTime(TimeSpan const &span) {
    setTime(d->time + span);
}
TimeSpan Clock::elapsed() const {
    return d->time - d->startedAt;
}
Time const &Clock::time() const {
    return d->time;
}
duint32 Clock::tickCount() const {
    return d->tickCount;
}
void Clock::setAppClock(Clock *c) {
    _appClock = c;
}
Clock &Clock::get() {
    DENG2_ASSERT(_appClock != 0);
    return *_appClock;
}
Time const &Clock::appTime() {
    return get().time();
}
} // namespace de
namespace de {
static char *duplicateStringAsUtf8(QString const &s) {
    QByteArray utf = s.toUtf8();
    char *copy = (char *) malloc(utf.size() + 1);
    memcpy(copy, utf.constData(), utf.size());
    copy[utf.size()] = 0;
    return copy;
}
DENG2_PIMPL(CommandLine) {
    QDir initialDir;
    typedef QList<QString> Arguments;
    Arguments arguments;
    typedef std::vector<char *> ArgumentPointers; // UTF-8 representation
    ArgumentPointers pointers;
    typedef std::vector<String> ArgumentStrings;
    typedef std::map<std::string, ArgumentStrings> Aliases;
    Aliases aliases;
    Impl(Public &i) : Base(i) {
        initialDir = QDir::current();
    }
    ~Impl() {
        clear();
    }
    void clear() {
        arguments.clear();
        DENG2_FOR_EACH(ArgumentPointers, i, pointers) free(*i);
        pointers.clear();
        pointers.push_back(0);
    }
    void appendArg(QString const &arg) {
        arguments.append(arg);
        if (pointers.empty()) {
            pointers.push_back(duplicateStringAsUtf8(arg));
            pointers.push_back(0); // Keep null-terminated.
        }
        else {
            pointers.insert(pointers.end() - 1, duplicateStringAsUtf8(arg));
        }
        DENG2_ASSERT(pointers.back() == 0);
    }
    void insert(duint pos, String const &arg) {
        if (pos > (duint) arguments.size()) {
            throw OutOfRangeError("CommandLine::insert", "Index out of range");
        }
        arguments.insert(pos, arg);
        pointers.insert(pointers.begin() + pos, duplicateStringAsUtf8(arg));
        DENG2_ASSERT(pointers.back() == 0);
    }
    void remove(duint pos) {
        if (pos >= (duint) arguments.size()) {
            throw OutOfRangeError("CommandLine::remove", "Index out of range");
        }
        arguments.removeAt(pos);
        free(pointers[pos]);
        pointers.erase(pointers.begin() + pos);
        DENG2_ASSERT(pointers.back() == 0);
    }
};
CommandLine::CommandLine() : d(new Impl(*this))
{}
CommandLine::CommandLine(QStringList args) : d(new Impl(*this)) {
    for (int i = 0; i < args.size(); ++i) {
        if (args.at(i)[0] == '@') {
            parseResponseFile(args.at(i).mid(1));
        }
        else {
            d->appendArg(args.at(i));
        }
    }
}
CommandLine::CommandLine(CommandLine const &other) : d(new Impl(*this)) {
    DENG2_FOR_EACH_CONST(Impl::Arguments, i, other.d->arguments) {
        d->appendArg(*i);
    }
}
NativePath CommandLine::startupPath() {
    return d->initialDir.path();
}
dint CommandLine::count() const {
    return d->arguments.size();
}
void CommandLine::clear() {
    d->clear();
}
void CommandLine::append(String const &arg) {
    d->appendArg(arg);
}
void CommandLine::insert(duint pos, String const &arg) {
    d->insert(pos, arg);
}
void CommandLine::remove(duint pos) {
    d->remove(pos);
}
CommandLine::ArgWithParams CommandLine::check(String const &arg, dint numParams) const {
    Impl::Arguments::const_iterator i = d->arguments.begin();
    for (; i != d->arguments.end() && !matches(arg, *i); ++i) {}
    if (i == d->arguments.end()) {
        return ArgWithParams();
    }
    ArgWithParams found;
    found.arg = arg;
    Impl::Arguments::const_iterator k = i;
    while (numParams-- > 0) {
        if (++k == d->arguments.end() || isOption(*k)) {
            return ArgWithParams();
        }
        found.params.append(*k);
    }
    found.pos = i - d->arguments.begin();
    return found;
}
int CommandLine::forAllParameters(String const &arg,
                                  std::function<void (duint, String const &)> paramHandler) const {
    int total = 0;
    bool inside = false;
    for (Impl::Arguments::const_iterator i = d->arguments.begin();
         i != d->arguments.end(); ++i) {
        if (matches(arg, *i)) {
            inside = true;
        }
        else if (inside) {
            if (isOption(*i)) {
                inside = false;
            }
            else {
                paramHandler(i - d->arguments.begin(), *i);
                ++total;
            }
        }
    }
    return total;
}
bool CommandLine::getParameter(String const &arg, String &param) const {
    dint pos = check(arg, 1);
    if (pos > 0) {
        param = at(pos + 1);
        return true;
    }
    return false;
}
dint CommandLine::has(String const &arg) const {
    dint howMany = 0;
    DENG2_FOR_EACH_CONST(Impl::Arguments, i, d->arguments) {
        if (matches(arg, *i)) {
            howMany++;
        }
    }
    return howMany;
}
bool CommandLine::isOption(duint pos) const {
    if (pos >= (duint) d->arguments.size()) {
        throw OutOfRangeError("CommandLine::isOption", "Index out of range");
    }
    DENG2_ASSERT(!d->arguments[pos].isEmpty());
    return isOption(d->arguments[pos]);
}
bool CommandLine::isOption(String const &arg) {
    return !(arg.empty() || arg[0] != '-');
}
String CommandLine::at(duint pos) const {
    return d->arguments.at(pos);
}
char const *const *CommandLine::argv() const {
    DENG2_ASSERT(*d->pointers.rbegin() == 0); // the list itself must be null-terminated
    return &d->pointers[0];
}
void CommandLine::makeAbsolutePath(duint pos) {
    if (pos >= (duint) d->arguments.size()) {
        throw OutOfRangeError("CommandLine::makeAbsolutePath", "Index out of range");
    }
    QString arg = d->arguments[pos];
    if (!isOption(pos) && !arg.startsWith("}")) {
        bool converted = false;
        QDir dir(NativePath(arg).expand()); // note: strips trailing slash
        if (!QDir::isAbsolutePath(arg)) {
            dir.setPath(d->initialDir.filePath(dir.path()));
            converted = true;
        }
        d->arguments[pos] = NativePath(dir.path());
        QFileInfo info(dir.path());
        if (info.isDir()) {
            d->arguments[pos] += '/';
        }
        free(d->pointers[pos]);
        d->pointers[pos] = duplicateStringAsUtf8(d->arguments[pos]);
        if (converted) {
            LOG_DEBUG("Argument %i converted to absolute path: \"%s\"") << pos << d->pointers[pos];
        }
    }
}
void CommandLine::parseResponseFile(NativePath const &nativePath) {
    QFile response(nativePath.expand());
    if (response.open(QFile::ReadOnly | QFile::Text)) {
        parse(QString::fromUtf8(response.readAll().constData()));
    }
    else {
        qWarning() << "Failed to open response file:" << nativePath;
    }
}
void CommandLine::parse(String const &cmdLine) {
    String::const_iterator i = cmdLine.begin();
    bool isDone = false;
    bool quote = false;
    while (i != cmdLine.end() && !isDone) {
        String::skipSpace(i, cmdLine.end());
        bool isResponse = false;
        if (*i == '@') {
            isResponse = true;
            String::skipSpace(++i, cmdLine.end());
        }
        String word;
        while (i != cmdLine.end() && (quote || !(*i).isSpace())) {
            bool copyChar = true;
            if (!quote) {
                if (*i == '\"') // Quote begins.
                {
                    quote = true;
                    copyChar = false;
                }
            }
            else {
                if (*i == '\"') // Quote ends.
                {
                    if (i + 1 != cmdLine.end() && *(i + 1) == '\"') // Doubles?
                    {
                        i++;
                    }
                    else {
                        quote = false;
                        copyChar = false;
                    }
                }
            }
            if (copyChar) {
                word.push_back(*i);
            }
            i++;
        }
        if (isResponse) // Response file?
        {
            parseResponseFile(word);
        }
        else if (word == "--") // End of arguments.
        {
            isDone = true;
        }
        else if (!word.empty()) // Make sure there *is *a word.
        {
            d->appendArg(word);
        }
    }
}
void CommandLine::alias(String const &full, String const &alias) {
    d->aliases[full.toStdString()].push_back(alias);
}
bool CommandLine::isAliasDefinedFor(String const &full) const {
    auto const &aliases = d.getConst()->aliases;
    return aliases.find(full.toStdString()) != aliases.end();
}
bool CommandLine::matches(String const &full, String const &fullOrAlias) const {
    if (!full.compareWithoutCase(fullOrAlias)) {
        return true;
    }
    Impl::Aliases::const_iterator found = d->aliases.find(full.toStdString());
    if (found != d->aliases.end()) {
        DENG2_FOR_EACH_CONST(Impl::ArgumentStrings, i, found->second) {
            if (!i->compareWithoutCase(fullOrAlias)) {
                return true;
            }
        }
    }
    return false;
}
#if defined (DENG_HAVE_QPROCESS)
bool CommandLine::execute() const {
    LOG_AS("CommandLine");
    if (count() < 1) return false;
    QStringList args;
    for (int i = 1; i < count(); ++i) args << at(i);
    qint64 pid = 0;
    if (!QProcess::startDetached(at(0), args, d->initialDir.path(), &pid)) {
        LOG_ERROR("Failed to start \"%s\"") << at(0);
        return false;
    }
    LOG_DEBUG("Started detached process %i \"%s\"") << pid << at(0);
    return true;
}
bool CommandLine::executeAndWait(String *output) const {
    std::unique_ptr<QProcess> proc(executeProcess());
    if (!proc) {
        return false;
    }
    bool result = proc->waitForFinished();
    if (output) {
        *output = String::fromUtf8(proc->readAll());
    }
    return result;
}
QProcess *CommandLine::executeProcess() const {
    LOG_AS("CommandLine");
    if (count() < 1) return nullptr;
    QStringList args;
    for (int i = 1; i < count(); ++i) args << at(i);
    auto *proc = new QProcess;
    proc->start(at(0), args);
    if (!proc->waitForStarted()) {
        delete proc;
        return nullptr;
    }
    LOG_DEBUG("Started process %i \"%s\"") << proc->pid() << at(0);
    return proc;
}
#endif // DENG_HAVE_QPROCESS
CommandLine &CommandLine::get() {
    return App::commandLine();
}
CommandLine::ArgWithParams::ArgWithParams() : pos(0)
{}
CommandLine::ArgWithParams::operator dint () const {
    return pos;
}
dint CommandLine::ArgWithParams::size() const {
    return params.size();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Config) {
    Path configPath;
    Refuge refuge;
    Process config;
    Version oldVersion;
    Impl(Path const &path)
        : configPath(path)
        , refuge("modules/Config")
        , config(&refuge.objectNamespace())
    {}
    void setOldVersion(Value const &old) {
        try {
            ArrayValue const &vers = old.as<ArrayValue>();
            oldVersion.major = int(vers.at(0).asNumber());
            oldVersion.minor = int(vers.at(1).asNumber());
            oldVersion.patch = int(vers.at(2).asNumber());
            oldVersion.build = int(vers.at(3).asNumber());
        }
        catch (...)
        {}
    }
    void write() {
        if (configPath.isEmpty()) return;
        refuge.write();
    }
};
Config::Config(Path const &path) : RecordAccessor(0), d(new Impl(path)) {
    setAccessedRecord(objectNamespace());
}
Config::ReadStatus Config::read() {
    ReadStatus readStatus = WasNotRead;
    if (d->configPath.isEmpty()) return readStatus;
    LOG_AS("Config::read");
    Version verInfo = Version::currentBuild();
    QScopedPointer<ArrayValue> version(new ArrayValue);
    *version << NumberValue(verInfo.major)
             << NumberValue(verInfo.minor)
             << NumberValue(verInfo.patch)
             << NumberValue(verInfo.build);
    File &scriptFile = App::rootFolder().locate<File>(d->configPath);
    bool shouldRunScript = App::commandLine().has("-reconfig");
    try {
        d->refuge.read();
        readStatus = DifferentVersion;
        LOG_DEBUG("Found serialized Config:\n") << objectNamespace();
        if (objectNamespace().has("__version__")) {
            Value const &oldVersion = objectNamespace()["__version__"].value();
            d->setOldVersion(oldVersion);
            if (oldVersion.compare(*version)) {
                d->config.globals().add(new Variable("__oldversion__", oldVersion.duplicate(),
                                                 Variable::AllowArray | Variable::ReadOnly));
                shouldRunScript = true;
            }
            else {
                readStatus = SameVersion;
                LOG_MSG("") << d->refuge.path() << " matches version " << version->asText();
            }
        }
        else {
            shouldRunScript = true;
        }
        if (scriptFile.status().modifiedAt > d->refuge.lastWrittenAt()) {
            LOG_MSG("%s is newer than %s, rerunning the script")
                    << d->configPath << d->refuge.path();
            shouldRunScript = true;
        }
        if (!shouldRunScript &&
            Package::containerOfFileModifiedAt(scriptFile) > d->refuge.lastWrittenAt()) {
            LOG_MSG("Package '%s' is newer than %s, rerunning the script")
                    << Package::identifierForContainerOfFile(scriptFile)
                    << d->refuge.path();
            shouldRunScript = true;
        }
    }
    catch (Archive::NotFoundError const &) {
        shouldRunScript = true;
    }
    catch (IByteArray::OffsetError const &) {
        shouldRunScript = true;
    }
    catch (Error const &error) {
        LOG_WARNING(error.what());
        shouldRunScript = true;
    }
    d->config.globals().add(new Variable("__version__", version.take(),
                                         Variable::AllowArray | Variable::ReadOnly));
    if (shouldRunScript) {
        Script script(scriptFile);
        d->config.run(script);
        d->config.execute();
    }
    return readStatus;
}
void Config::write() const {
    d->write();
}
void Config::writeIfModified() const {
    try {
        if (d->refuge.hasModifiedVariables()) {
            write();
        }
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to write Config: ") << er.asText();
    }
}
Record &Config::objectNamespace() {
    return d->config.globals();
}
Record const &Config::objectNamespace() const {
    return d->config.globals();
}
Config &Config::get() {
    return App::config();
}
Variable &Config::get(String const &name) // static
{
    return get()[name];
}
bool Config::exists() {
    return App::configExists();
}
Version Config::upgradedFromVersion() const {
    return d->oldVersion;
}
Variable &Config::set(String const &name, bool value) {
    return objectNamespace().set(name, value);
}
Variable &Config::set(String const &name, Value::Number const &value) {
    return objectNamespace().set(name, value);
}
Variable &Config::set(String const &name, dint value) {
    return objectNamespace().set(name, value);
}
Variable &Config::set(String const &name, duint value) {
    return objectNamespace().set(name, value);
}
Variable &Config::set(String const &name, ArrayValue *value) {
    return objectNamespace().set(name, value);
}
Variable &Config::set(String const &name, Value::Text const &value) {
    return objectNamespace().set(name, value);
}
} // namespace de
namespace de {
DebugLogSink::DebugLogSink(QtMsgType msgType) : LogSink(_format), _msgType(msgType)
{}
DebugLogSink::~DebugLogSink()
{}
LogSink &DebugLogSink::operator << (String const &plainText) {
    QByteArray utf8 = plainText.toUtf8();
    if (_msgType == QtWarningMsg)
        qWarning() << utf8.constData();
    else
        qDebug() << utf8.constData();
    return *this;
}
void DebugLogSink::flush()
{}
} // namespace de
namespace de {
FileLogSink::FileLogSink(File &outputFile)
    : LogSink(_format)
    , _file(outputFile)
{}
LogSink &FileLogSink::operator << (String const &plainText) {
    _file << Block((plainText + "\n").toUtf8());
    return *this;
}
void FileLogSink::flush() {
    _file.flush();
}
} // namespace de
namespace de {
namespace internal {
struct Garbage : public Lockable {
    typedef std::map<void *, GarbageDestructor> Allocs; // O(log n) search
    Allocs allocs;
    ~Garbage() {
        recycle();
    }
    bool contains(void const *ptr) const {
        DENG2_GUARD(this);
        Allocs::const_iterator i = allocs.find(const_cast<void *>(ptr));
        return i != allocs.end();
    }
    void recycle(GarbageDestructor condition = nullptr) {
        DENG2_GUARD(this);
        if (allocs.empty()) return;
#ifdef DENG2_DEBUG
#endif
        for (Allocs::iterator i = allocs.begin(); i != allocs.end(); ) {
            Allocs::iterator next = i;
            ++next;
            DENG2_ASSERT(i->second);
            if (!condition || i->second == condition) {
                i->second(i->first);
                if (condition) allocs.erase(i);
            }
            i = next;
        }
        if (!condition) {
            allocs.clear();
        }
    }
    void forgetAndLeak() {
        allocs.clear(); // Oh well...
    }
};
struct Garbages : public std::map<QThread *, Garbage *>, public Lockable {
    /**
     * Recycles all collected garbage and deletes the collectors.
     */
    ~Garbages() {
        clearAll();
    }
    void clearAll() {
        DENG2_GUARD(this);
        for (iterator i = begin(); i != end(); ++i) {
            delete i->second;
        }
        clear();
    }
    void recycleWithDestructor(GarbageDestructor func) {
        DENG2_GUARD(this);
        for (iterator i = begin(); i != end(); ++i) {
            i->second->recycle(func);
        }
    }
    void forgetAndLeak() {
        DENG2_GUARD(this);
        for (iterator i = begin(); i != end(); ++i) {
            i->second->forgetAndLeak();
        }
        clear();
    }
};
} // namespace internal
} // namespace de
using namespace de;
using namespace de::internal;
static Garbages garbages;
static Garbage *garbageForThread(QThread *thread) {
    DENG2_GUARD(garbages);
    Garbage *result;
    Garbages::iterator i = garbages.find(thread);
    if (i != garbages.end()) {
        result = i->second;
    }
    else {
        garbages[thread] = result = new Garbage;
    }
    return result;
}
void Garbage_ClearForThread(void) {
    DENG2_GUARD(garbages);
    Garbages::iterator i = garbages.find(QThread::currentThread());
    if (i != garbages.end()) {
        Garbage *g = i->second;
        delete g;
        garbages.erase(i);
    }
}
void Garbage_TrashMalloc(void *ptr) {
    Garbage_TrashInstance(ptr, free);
}
void Garbage_TrashInstance(void *ptr, GarbageDestructor destructor) {
    if (ptr) {
        Garbage *g = garbageForThread(QThread::currentThread());
        g->allocs[ptr] = destructor;
    }
}
int Garbage_IsTrashed(void const *ptr) {
    Garbage *g = garbageForThread(QThread::currentThread());
    return g->contains(ptr);
}
void Garbage_Untrash(void *ptr) {
    Garbage *g = garbageForThread(QThread::currentThread());
    DENG2_ASSERT(g->contains(ptr));
    g->allocs.erase(ptr);
}
void Garbage_RemoveIfTrashed(void *ptr) {
    Garbage *g = garbageForThread(QThread::currentThread());
    Garbage::Allocs::iterator found = g->allocs.find(ptr);
    if (found != g->allocs.end()) {
        g->allocs.erase(found);
    }
}
void Garbage_Recycle(void) {
    Garbage *g = garbageForThread(QThread::currentThread());
    g->recycle();
}
void Garbage_ForgetAndLeak(void) {
    garbages.forgetAndLeak();
}
void Garbage_RecycleAllWithDestructor(GarbageDestructor destructor) {
    garbages.recycleWithDestructor(destructor);
}
namespace de {
DENG2_PIMPL_NOREF(HighPerformanceTimer) {
    QDateTime origin;
    QElapsedTimer startedAt;
    Impl() {
        origin = QDateTime::currentDateTime();
        startedAt.start();
    }
    duint64 milliSeconds() {
        return duint64(startedAt.elapsed());
    }
};
HighPerformanceTimer::HighPerformanceTimer() : d(new Impl)
{}
TimeSpan HighPerformanceTimer::elapsed() const {
    return TimeSpan(d->milliSeconds() / 1000.0);
}
Time HighPerformanceTimer::startedAt() const {
    return d->origin;
}
} // namespace de
namespace de {
static std::atomic_uint idGenerator { 1 };
Id::Id() : _id(idGenerator++) {
    if (_id == None) {
        ++_id;
    }
}
Id::Id(String const &text) : _id(None) {
    if (text.beginsWith("{") && text.endsWith("}")) {
        _id = text.substr(1, text.size() - 2).toUInt(nullptr, 16);
    }
}
Id::~Id()
{}
Id::operator String () const {
    return asText();
}
Id::operator Value::Number () const {
    return static_cast<Value::Number>(_id);
}
String Id::asText() const {
    return QString("{%1}").arg(_id, 8, 16, QChar('0'));
}
ddouble Id::asDouble() const {
    return _id;
}
dint64 Id::asInt64() const {
    return _id;
}
QTextStream &operator << (QTextStream &os, Id const &id) {
    os << id.asText();
    return os;
}
void Id::operator >> (Writer &to) const {
    to << _id;
}
void Id::operator << (Reader &from) {
    from >> _id;
    resetGenerator(_id); // Try to avoid overlaps.
}
void Id::resetGenerator(Id::Type largestKnownId) {
    if (idGenerator <= largestKnownId) {
        idGenerator = largestKnownId + 1;
    }
}
} // namespace de
namespace de {
#if defined (DENG_STATIC_LINK)
struct ImportedStaticLibrary {
    char const *name;
    StaticSymbolFunc symbolFunc;
};
static std::array<ImportedStaticLibrary, 16> importedLibraries;
ImportedStaticLibrary const *findStaticLibrary(char const *name) {
    for (auto const &imp : importedLibraries) {
        if (!imp.name) break;
        if (!qstrcmp(imp.name, name)) {
            return &imp;
        }
    }
    return nullptr;
}
#endif // DENG_STATIC_LINK
#if defined (DENG_STATIC_LINK)
typedef ImportedStaticLibrary const *Handle;
#else
#  if defined(UNIX) && defined(DENG2_QT_4_7_OR_NEWER) && !defined(DENG2_QT_4_8_OR_NEWER)
#    define DENG2_USE_DLOPEN
#    include <dlfcn.h>
typedef void *Handle;
#  else
typedef QLibrary *Handle;
#  endif
#endif
char const *Library::DEFAULT_TYPE = "library/generic";
DENG2_PIMPL(Library) {
    Handle library;
    typedef QHash<String, void *> Symbols;
    Symbols symbols;
    String type;
    Impl(Public *i) : Base(i), library(0), type(DEFAULT_TYPE)
    {}
#if defined (DENG_STATIC_LINK) || defined(DENG2_USE_DLOPEN)
    NativePath fileName;
    NativePath nativePath() { return fileName; }
    bool isLoaded() const { return library != 0; }
#else
    NativePath nativePath() {
        DENG2_ASSERT(library);
        return library->fileName();
    }
    bool isLoaded() const { return library->isLoaded(); }
#endif
};
Library::Library(NativePath const &nativePath) : d(new Impl(this)) {
    LOG_AS("Library");
    LOG_TRACE("Loading \"%s\"", nativePath.pretty());
#if defined (DENG_STATIC_LINK)
    d->fileName = nativePath;
    d->library = findStaticLibrary(nativePath.toUtf8().constData());
#elif defined (DENG2_USE_DLOPEN)
    d->fileName = nativePath;
    d->library = dlopen(nativePath.toUtf8().constData(), RTLD_NOW);
#else
    d->library = new QLibrary(nativePath);
    d->library->setLoadHints(QLibrary::ResolveAllSymbolsHint);
    d->library->load();
#endif
    if (!d->isLoaded()) {
#if defined (DENG_STATIC_LINK)
        QString msg = "static library has not been imported";
#elif defined (DENG2_USE_DLOPEN)
        QString msg = dlerror();
#else
        QString msg = d->library->errorString();
        delete d->library;
#endif
        d->library = 0;
        throw LoadError("Library::Library", msg);
    }
    if (hasSymbol("deng_LibraryType")) {
        d->type = DENG2_SYMBOL(deng_LibraryType)();
    }
    if (d->type.beginsWith("deng-plugin/") && hasSymbol("deng_InitializePlugin")) {
        DENG2_SYMBOL(deng_InitializePlugin)();
    }
}
Library::~Library() {
    if (d->library) {
        LOG_AS("~Library");
        LOG_TRACE("Unloading \"%s\"", d->nativePath().pretty());
        if (d->type.beginsWith("deng-plugin/") && hasSymbol("deng_ShutdownPlugin")) {
            DENG2_SYMBOL(deng_ShutdownPlugin)();
        }
        LogBuffer::get().clear();
#if defined (DENG_STATIC_LINK)
        d->library = nullptr;
#elif defined (DENG2_USE_DLOPEN)
        dlclose(d->library);
#else
        d->library->unload();
        delete d->library;
#endif
    }
}
String const &Library::type() const {
    return d->type;
}
void *Library::address(String const &name, SymbolLookupMode lookup) {
    if (!d->library) {
        throw SymbolMissingError("Library::symbol", "Library not loaded");
    }
    Impl::Symbols::iterator found = d->symbols.find(name);
    if (found != d->symbols.end()) {
        return found.value();
    }
#if defined (DENG_STATIC_LINK)
    void *ptr = d->library->symbolFunc(name.toLatin1().constData());
    qDebug() << d->fileName.toString() << name << ptr;
#elif defined (DENG2_USE_DLOPEN)
    void *ptr = dlsym(d->library, name.toLatin1().constData());
#else
    void *ptr = de::function_cast<void *>(d->library->resolve(name.toLatin1().constData()));
#endif
    if (!ptr) {
        if (lookup == RequiredSymbol) {
            throw SymbolMissingError("Library::symbol", "Symbol '" + name + "' was not found");
        }
        return 0;
    }
    d->symbols[name] = ptr;
    return ptr;
}
bool Library::hasSymbol(String const &name) const {
    if (d->symbols.find(name) != d->symbols.end()) return true;
#if defined (DENG_STATIC_LINK)
    return d->library->symbolFunc(name.toLatin1().constData()) != 0;
#elif defined (DENG2_USE_DLOPEN)
    return dlsym(d->library, name.toLatin1().constData()) != 0;
#else
    return d->library->resolve(name.toLatin1().constData()) != 0;
#endif
}
#if defined (DENG_STATIC_LINK)
void Library::importStaticLibrary(char const *name, StaticSymbolFunc symbolFunc) // static
{
    for (uint i = 0; i < importedLibraries.size(); ++i) {
        auto &impLib = importedLibraries[i];
        if (!impLib.name) {
            impLib.name = name;
            impLib.symbolFunc = symbolFunc;
            return;
        }
    }
}
StringList Library::staticLibraries() // static
{
    StringList names;
    for (auto const &lib : importedLibraries) {
        if (lib.name) {
            names << lib.name;
        }
    }
    return names;
}
#endif
} // namespace de
namespace de {
char const *TEXT_STYLE_NORMAL        = DENG2_ESC("0"); // normal
char const *TEXT_STYLE_MESSAGE       = DENG2_ESC("0"); // normal
char const *TEXT_STYLE_MAJOR_MESSAGE = DENG2_ESC("1"); // major
char const *TEXT_STYLE_MINOR_MESSAGE = DENG2_ESC("2"); // minor
char const *TEXT_STYLE_SECTION       = DENG2_ESC("3"); // meta
char const *TEXT_STYLE_MAJOR_SECTION = DENG2_ESC("4"); // major meta
char const *TEXT_STYLE_MINOR_SECTION = DENG2_ESC("5"); // minor meta
char const *TEXT_STYLE_LOG_TIME      = DENG2_ESC("6"); // aux meta
char const *TEXT_MARK_INDENT         = DENG2_ESC(">");
char const *MAIN_SECTION = "";
static int const LINE_BREAKING_SECTION_LENGTH = 30;
namespace internal {
#if 0
/**
 * @internal
 * The logs table is lockable so that multiple threads can access their logs at
 * the same time.
 */
class Logs : public Lockable, public QHash<QThread *, Log *> {
public:
    Logs() {}
    ~Logs() {
        DENG2_GUARD_PTR(this);
        for (Log *log : *this) delete log;
    }
};
#endif
} // namespace internal
static FIFO<LogEntry::Arg> argPool;
LogEntry::Arg::Arg() : _type(IntegerArgument) {
    _data.intValue = 0;
}
LogEntry::Arg::~Arg() {
    clear();
}
void LogEntry::Arg::clear() {
    if (_type == StringArgument) {
        delete _data.stringValue;
        _data.stringValue = 0;
        _type = IntegerArgument;
    }
}
void LogEntry::Arg::setValue(dint i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = i;
}
void LogEntry::Arg::setValue(duint i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = i;
}
void LogEntry::Arg::setValue(long int i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = i;
}
void LogEntry::Arg::setValue(long unsigned int i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = i;
}
void LogEntry::Arg::setValue(duint64 i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = dint64(i);
}
void LogEntry::Arg::setValue(dint64 i) {
    clear();
    _type = IntegerArgument;
    _data.intValue = i;
}
void LogEntry::Arg::setValue(ddouble d) {
    clear();
    _type = FloatingPointArgument;
    _data.floatValue = d;
}
void LogEntry::Arg::setValue(void const *p) {
    clear();
    _type = IntegerArgument;
    _data.intValue = dint64(p);
}
void LogEntry::Arg::setValue(char const *s) {
    clear();
    _type = StringArgument;
    _data.stringValue = new String(s);
}
void LogEntry::Arg::setValue(String const &s) {
    clear();
    _type = StringArgument;
    _data.stringValue = new String(s.data(), s.size());
}
void LogEntry::Arg::setValue(std::array<char, 4> const &typecode) {
    setValue(QString::fromLatin1(typecode.data(), 4));
}
void LogEntry::Arg::setValue(LogEntry::Arg::Base const &arg) {
    switch (arg.logEntryArgType()) {
    case IntegerArgument:
        setValue(arg.asInt64());
        break;
    case FloatingPointArgument:
        setValue(arg.asDouble());
        break;
    case StringArgument:
        setValue(arg.asText()); // deep copy
        break;
    }
}
LogEntry::Arg &LogEntry::Arg::operator = (Arg const &other) {
    clear();
    if (other._type == StringArgument) {
        setValue(*other._data.stringValue); // deep copy
    }
    else {
        _type = other._type;
        _data = other._data;
    }
    return *this;
}
ddouble LogEntry::Arg::asNumber() const {
    if (_type == IntegerArgument) {
        return ddouble(_data.intValue);
    }
    else if (_type == FloatingPointArgument) {
        return _data.floatValue;
    }
    throw TypeError("Log::Arg::asNumber", "String argument cannot be used as a number");
}
String LogEntry::Arg::asText() const {
    if (_type == StringArgument) {
        return *_data.stringValue;
    }
    else if (_type == IntegerArgument) {
        return String::number(_data.intValue);
    }
    else if (_type == FloatingPointArgument) {
        return String::number(_data.floatValue);
    }
    throw TypeError("Log::Arg::asText", "Number argument cannot be used a string");
}
void LogEntry::Arg::operator >> (Writer &to) const {
    to << dbyte(_type);
    switch (_type) {
    case IntegerArgument:
        to << _data.intValue;
        break;
    case FloatingPointArgument:
        to << _data.floatValue;
        break;
    case StringArgument:
        to << *_data.stringValue;
        break;
    }
}
void LogEntry::Arg::operator << (Reader &from) {
    if (_type == StringArgument) delete _data.stringValue;
    from.readAs<dbyte>(_type);
    switch (_type) {
    case IntegerArgument:
        from >> _data.intValue;
        break;
    case FloatingPointArgument:
        from >> _data.floatValue;
        break;
    case StringArgument:
        _data.stringValue = new String;
        from >> *_data.stringValue;
        break;
    }
}
LogEntry::Arg *LogEntry::Arg::newFromPool() {
    Arg *arg = argPool.take();
    if (arg) return arg;
    return new LogEntry::Arg;
}
void LogEntry::Arg::returnToPool(Arg *arg) {
    arg->clear();
    argPool.put(arg);
}
LogEntry::LogEntry() : _metadata(0), _sectionDepth(0), _disabled(true)
{}
LogEntry::LogEntry(duint32 metadata, String const &section, int sectionDepth, String const &format, Args args)
    : _metadata(metadata)
    , _section(section)
    , _sectionDepth(sectionDepth)
    , _format(format)
    , _disabled(false)
    , _args(args) {
    if (!LogBuffer::get().isEnabled(metadata)) {
        _disabled = true;
    }
}
LogEntry::LogEntry(LogEntry const &other, Flags extraFlags)
    : Lockable()
    , ISerializable()
    , _when(other._when)
    , _metadata(other._metadata)
    , _section(other._section)
    , _sectionDepth(other._sectionDepth)
    , _format(other._format)
    , _defaultFlags(other._defaultFlags | extraFlags)
    , _disabled(other._disabled) {
    DENG2_FOR_EACH_CONST(Args, i, other._args) {
        Arg *a = Arg::newFromPool();
        *a = **i;
        _args.append(a);
    }
}
LogEntry::~LogEntry() {
    DENG2_GUARD(this);
    for (Args::iterator i = _args.begin(); i != _args.end(); ++i) {
        Arg::returnToPool(*i);
    }
}
LogEntry::Flags LogEntry::flags() const {
    return _defaultFlags;
}
String LogEntry::asText(Flags const &formattingFlags, int shortenSection) const {
    DENG2_GUARD(this);
    Flags flags = formattingFlags;
    QString result;
    QTextStream output(&result);
    if (_defaultFlags & Simple) {
        flags |= Simple;
    }
    if (!flags.testFlag(Simple)) {
        if (flags.testFlag(Styled)) output << TEXT_STYLE_LOG_TIME;
        output << _when.asText(Date::SecondsSinceStart) << " ";
        if (!flags.testFlag(OmitDomain)) {
            QChar dc = (_metadata & Resource? 'R' :
                        _metadata & Map?      'M' :
                        _metadata & Script?   'S' :
                        _metadata & GL?       'G' :
                        _metadata & Audio?    'A' :
                        _metadata & Input?    'I' :
                        _metadata & Network?  'N' : ' ');
            if (_metadata & Dev) {
                if (dc != ' ')
                    dc = dc.toLower();
                else
                    dc = '-'; // Generic developer message
            }
            if (!flags.testFlag(Styled)) {
                output << dc;
            }
            else {
                output << _E(s)_E(F)_E(m) << dc << _E(.) << " ";
            }
        }
        if (!flags.testFlag(OmitLevel)) {
            if (!flags.testFlag(Styled)) {
                char const *levelNames[] = {
                    "", // not used
                    "(vv)",
                    "(v)",
                    "",
                    "(i)",
                    "(WRN)",
                    "(ERR)",
                    "(!!!)"
                };
                output << qSetPadChar(' ') << qSetFieldWidth(5)
                       << levelNames[level()] << qSetFieldWidth(0) << " ";
            }
            else {
                char const *levelNames[] = {
                    "", // not used
                    "XVerbose",
                    "Verbose",
                    "",
                    "Note!",
                    "Warning",
                    "ERROR",
                    "FATAL!"
                };
                output << "\t"
                    << (level() >= LogEntry::Warning? TEXT_STYLE_MAJOR_SECTION :
                        level() <= LogEntry::Verbose? TEXT_STYLE_MINOR_SECTION :
                                                      TEXT_STYLE_SECTION)
                    << levelNames[level()] << "\t";
            }
        }
    }
    if (!flags.testFlag(OmitSection) && !_section.empty()) {
        if (flags.testFlag(Styled)) {
            output << TEXT_MARK_INDENT
                   << (level() >= LogEntry::Warning? TEXT_STYLE_MAJOR_SECTION :
                       level() <= LogEntry::Verbose? TEXT_STYLE_MINOR_SECTION :
                                                     TEXT_STYLE_SECTION);
        }
        QString sect;
        if (flags.testFlag(AbbreviateSection)) {
            /*
             * We'll split the section into parts, and then abbreviate some of
             * the parts, trying not to lose too much information.
             * @a shortenSection controls how much of the section can be
             * abbreviated (num of chars from beginning).
             */
            QStringList parts = _section.split(" > ");
            int len = 0;
            while (!parts.isEmpty()) {
                if (!sect.isEmpty()) {
                    len += 3;
                    sect += " > ";
                }
                if (len + parts.first().size() >= shortenSection) break;
                len += parts.first().size();
                if (sect.isEmpty()) {
                    sect += parts.first();
                }
                else {
                    sect += "..";
                }
                parts.removeFirst();
            }
            sect += _section.mid(len);
        }
        else {
            if (shortenSection < _section.size()) {
                sect = _section.right(_section.size() - shortenSection);
            }
        }
        if (flags.testFlag(SectionSameAsBefore)) {
            int visibleSectLen = (!sect.isEmpty() && shortenSection? sect.size() : 0);
            int fillLen = de::max(shortenSection, _section.size()) - visibleSectLen;
            if (fillLen > LINE_BREAKING_SECTION_LENGTH) fillLen = 2;
            output << String(fillLen, QChar(' '));
            if (visibleSectLen) {
                output << "[" << sect << "] ";
            }
            else {
                output << "   ";
            }
        }
        else {
            char const *separator = (sect.size() > LINE_BREAKING_SECTION_LENGTH? "\n    " : " ");
            output << "[" << sect << "]" << separator;
        }
    }
    if (flags.testFlag(Styled)) {
        output << TEXT_MARK_INDENT
               << (level() >= LogEntry::Warning? TEXT_STYLE_MAJOR_MESSAGE :
                   level() <= LogEntry::Verbose? TEXT_STYLE_MINOR_MESSAGE :
                                                 TEXT_STYLE_MESSAGE);
    }
    if (_args.empty()) {
        output << _format; // Verbatim.
    }
    else {
        String::PatternArgs patArgs;
        DENG2_FOR_EACH_CONST(Args, i, _args) patArgs << *i;
        output << _format % patArgs;
    }
    return result;
}
void LogEntry::operator >> (Writer &to) const {
    to << _when
       << _section
       << _format
       << duint32(_metadata)
       << dbyte(_sectionDepth)
       << duint32(_defaultFlags);
    to.writeObjects(_args);
}
void LogEntry::operator << (Reader &from) {
    foreach (Arg *a, _args) delete a;
    _args.clear();
    from >> _when
         >> _section
         >> _format;
    if (from.version() >= DENG2_PROTOCOL_1_14_0_LogEntry_metadata) {
        from.readAs<duint32>(_metadata);
    }
    else {
        dbyte oldLevel;
        from >> oldLevel;
        _metadata = oldLevel; // lacks audience information
    }
    from.readAs<dbyte>(_sectionDepth)
        .readAs<duint32>(_defaultFlags)
        .readObjects<Arg>(_args);
}
QTextStream &operator << (QTextStream &stream, LogEntry::Arg const &arg) {
    switch (arg.type()) {
    case LogEntry::Arg::IntegerArgument:
        stream << arg.intValue();
        break;
    case LogEntry::Arg::FloatingPointArgument:
        stream << arg.floatValue();
        break;
    case LogEntry::Arg::StringArgument:
        stream << arg.stringValue();
        break;
    }
    return stream;
}
Log::Section::Section(char const *name) : _log(threadLog()), _name(name) {
    _log.beginSection(_name);
}
Log::Section::~Section() {
    _log.endSection(_name);
}
DENG2_PIMPL_NOREF(Log) {
    typedef QVector<char const *> SectionStack;
    SectionStack sectionStack;
    LogEntry *throwawayEntry;
    duint32 currentEntryMedata; ///< Applies to the current entry being staged in the thread.
    int interactive = 0;
    Impl()
        : throwawayEntry(new LogEntry) ///< A disabled LogEntry, so doesn't accept arguments.
        , currentEntryMedata(0) {
        sectionStack.push_back(MAIN_SECTION);
    }
    ~Impl() {
        delete throwawayEntry;
    }
};
Log::Log() : d(new Impl)
{}
Log::~Log() // virtual
{}
void Log::setCurrentEntryMetadata(duint32 metadata) {
    d->currentEntryMedata = metadata;
}
duint32 Log::currentEntryMetadata() const {
    return d->currentEntryMedata;
}
bool Log::isStaging() const {
    return d->currentEntryMedata != 0;
}
void Log::beginSection(char const *name) {
    d->sectionStack.push_back(name);
}
void Log::endSection(char const *DENG2_DEBUG_ONLY(name)) {
    DENG2_ASSERT(d->sectionStack.back() == name);
    d->sectionStack.pop_back();
}
void Log::beginInteractive() {
    d->interactive++;
}
void Log::endInteractive() {
    d->interactive--;
    DENG2_ASSERT(d->interactive >= 0);
}
bool Log::isInteractive() const {
    return d->interactive > 0;
}
LogEntry &Log::enter(String const &format, LogEntry::Args arguments) {
    return enter(LogEntry::Message, format, arguments);
}
LogEntry &Log::enter(duint32 metadata, String const &format, LogEntry::Args arguments) {
    d->currentEntryMedata = 0;
    if (isInteractive()) {
        metadata |= LogEntry::Interactive;
    }
    if (!LogBuffer::get().isEnabled(metadata)) {
        DENG2_ASSERT(arguments.isEmpty());
        return *d->throwawayEntry;
    }
    String context;
    String latest;
    int depth = 0;
    foreach (char const *i, d->sectionStack) {
        if (i == latest) {
            continue;
        }
        if (context.size()) {
            context += " > ";
        }
        latest = i;
        context += i;
        ++depth;
    }
    LogEntry *entry = new LogEntry(metadata, context, depth, format, arguments);
    LogBuffer::get().add(entry);
    return *entry;
}
/*static internal::Logs &theLogs()
{
    if (logsPtr.get()) return *logsPtr;
    static Lockable lock;
    DENG2_GUARD(lock);
    if (!logsPtr.get()) logsPtr.reset(new internal::Logs);
    return *logsPtr;
}*/
static QThreadStorage<Log> theLogs;
Log &Log::threadLog() {
    return theLogs.localData();
    /*if (!theLogs.hasLocalData())
    {
        theLogs.setLocalData(<#de::Log t#>)
    internal::Logs &logs = theLogs();
    DENG2_GUARD(logs);
    auto found = logs.constFind(thread);
    if (found == logs.constEnd()) {
        Log *theLog = new Log;
        logs.insert(thread, theLog);
        return *theLog;
    }
    else {
        return *found.value();
    }*/
}
void Log::disposeThreadLog() {
    /*internal::Logs &logs = theLogs();
    DENG2_GUARD(logs);
    QThread *thread = QThread::currentThread();
    auto found = logs.find(thread);
    if (found != logs.end()) {
        delete found.value();
        logs.erase(found);
    }*/
}
LogEntryStager::LogEntryStager(duint32 metadata, String const &format)
    : _metadata(metadata) {
    if (!LogBuffer::appBufferExists()) {
        _disabled = true;
    }
    else {
        auto &log = LOG();
        if (!(_metadata & LogEntry::DomainMask)) {
            _metadata |= LogEntry::Generic;
        }
        if (log.isInteractive()) {
            _metadata |= LogEntry::Interactive;
        }
        _disabled = !LogBuffer::get().isEnabled(_metadata);
        if (!_disabled) {
            _format = format;
            log.setCurrentEntryMetadata(_metadata);
        }
    }
}
LogEntryStager::~LogEntryStager() {
    if (!_disabled) {
        LOG().enter(_metadata, _format, _args);
    }
}
} // namespace de
namespace de {
TimeSpan const FLUSH_INTERVAL = .2; // seconds
DENG2_PIMPL(LogBuffer) {
    typedef QList<LogEntry *> EntryList;
    typedef QSet<LogSink *> Sinks;
    SimpleLogFilter defaultFilter;
    IFilter const *entryFilter;
    dint maxEntryCount;
    bool useStandardOutput;
    bool flushingEnabled;
    String outputPath;
    FileLogSink *fileLogSink;
#ifndef WIN32
    TextStreamLogSink outSink;
    TextStreamLogSink errSink;
#else
    DebugLogSink outSink;
    DebugLogSink errSink;
#endif
    EntryList entries;
    EntryList toBeFlushed;
    Time lastFlushedAt;
    QTimer *autoFlushTimer;
    Sinks sinks;
    Impl(Public *i, duint maxEntryCount)
        : Base(i)
        , entryFilter(&defaultFilter)
        , maxEntryCount(maxEntryCount)
        , useStandardOutput(true)
        , flushingEnabled(true)
        , fileLogSink(0)
#ifndef WIN32
        , outSink(new QTextStream(stdout))
        , errSink(new QTextStream(stderr))
#else
        , outSink(QtDebugMsg)
        , errSink(QtWarningMsg)
#endif
        , lastFlushedAt(Time::invalidTime())
        , autoFlushTimer(0) {
        outSink.setMode(LogSink::OnlyNormalEntries);
        errSink.setMode(LogSink::OnlyWarningEntries);
        sinks.insert(&outSink);
        sinks.insert(&errSink);
    }
    ~Impl() {
        if (autoFlushTimer) autoFlushTimer->stop();
        delete fileLogSink;
    }
    void enableAutoFlush(bool yes) {
        DENG2_ASSERT(qApp);
        if (yes) {
            if (!autoFlushTimer->isActive()) {
                autoFlushTimer->start(int(FLUSH_INTERVAL.asMilliSeconds()));
            }
        }
        else {
            autoFlushTimer->stop();
        }
    }
    void createFileLogSink(bool truncate) {
        if (!outputPath.isEmpty()) {
            File *outputFile = nullptr;
            if (!truncate) {
                outputFile = App::rootFolder().tryLocate<File>(outputPath);
            }
            if (!outputFile) {
                outputFile = &App::rootFolder().replaceFile(outputPath);
            }
            DENG2_ASSERT(outputFile);
            DENG2_ASSERT(!fileLogSink);
            fileLogSink = new FileLogSink(*outputFile);
            sinks.insert(fileLogSink);
        }
    }
    void disposeFileLogSink() {
        if (fileLogSink) {
            sinks.remove(fileLogSink);
            delete fileLogSink;
            fileLogSink = nullptr;
        }
    }
};
LogBuffer *LogBuffer::_appBuffer = nullptr;
LogBuffer::LogBuffer(duint maxEntryCount)
    : d(new Impl(this, maxEntryCount)) {
    d->autoFlushTimer = new QTimer(this);
    connect(d->autoFlushTimer, SIGNAL(timeout()), this, SLOT(flush()));
}
LogBuffer::~LogBuffer() {
    DENG2_GUARD(this);
    setOutputFile("");
    clear();
    if (_appBuffer == this) _appBuffer = 0;
}
void LogBuffer::clear() {
    DENG2_GUARD(this);
    flush();
    DENG2_FOR_EACH(Impl::EntryList, i, d->entries) {
        delete *i;
    }
    d->entries.clear();
}
dsize LogBuffer::size() const {
    DENG2_GUARD(this);
    return d->entries.size();
}
void LogBuffer::latestEntries(Entries &entries, int count) const {
    DENG2_GUARD(this);
    entries.clear();
    for (int i = d->entries.size() - 1; i >= 0; --i) {
        entries.append(d->entries[i]);
        if (count && entries.size() >= count) {
            return;
        }
    }
}
void LogBuffer::setEntryFilter(IFilter const *entryFilter) {
    if (entryFilter) {
        d->entryFilter = entryFilter;
    }
    else {
        d->entryFilter = &d->defaultFilter;
    }
}
bool LogBuffer::isEnabled(duint32 entryMetadata) const {
    DENG2_ASSERT(d->entryFilter != 0);
    DENG2_ASSERT(entryMetadata & LogEntry::DomainMask); // must have a domain
    if (entryMetadata & LogEntry::Privileged) {
        return true; // always passes
    }
    return d->entryFilter->isLogEntryAllowed(entryMetadata);
}
void LogBuffer::setMaxEntryCount(duint maxEntryCount) {
    d->maxEntryCount = maxEntryCount;
}
void LogBuffer::add(LogEntry *entry) {
    DENG2_GUARD(this);
    if (d->lastFlushedAt.isValid() && d->lastFlushedAt.since() > FLUSH_INTERVAL) {
        flush();
    }
    d->entries.push_back(entry);
    d->toBeFlushed.push_back(entry);
}
void LogBuffer::enableStandardOutput(bool yes) {
    DENG2_GUARD(this);
    d->useStandardOutput = yes;
    d->outSink.setMode(yes? LogSink::OnlyNormalEntries  : LogSink::Disabled);
    d->errSink.setMode(yes? LogSink::OnlyWarningEntries : LogSink::Disabled);
}
void LogBuffer::enableFlushing(bool yes) {
    d->flushingEnabled = yes;
    d->enableAutoFlush(true);
}
void LogBuffer::setAutoFlushInterval(TimeSpan const &interval) {
    enableFlushing();
    d->autoFlushTimer->setInterval(interval.asMilliSeconds());
}
void LogBuffer::setOutputFile(String const &path, OutputChangeBehavior behavior) {
    DENG2_GUARD(this);
    if (behavior == FlushFirstToOldOutputs) {
        flush();
    }
    d->disposeFileLogSink();
    d->outputPath = path;
    d->createFileLogSink(true /* truncated */);
}
String LogBuffer::outputFile() const {
    return d->outputPath;
}
void LogBuffer::addSink(LogSink &sink) {
    DENG2_GUARD(this);
    d->sinks.insert(&sink);
}
void LogBuffer::removeSink(LogSink &sink) {
    DENG2_GUARD(this);
    d->sinks.remove(&sink);
}
void LogBuffer::flush() {
    if (!d->flushingEnabled) return;
    DENG2_GUARD(this);
    if (!d->toBeFlushed.isEmpty()) {
        DENG2_FOR_EACH(Impl::EntryList, i, d->toBeFlushed) {
            DENG2_GUARD_FOR(**i, guardingCurrentLogEntry);
            foreach (LogSink *sink, d->sinks) {
                if (sink->willAccept(**i)) {
                    try {
                        *sink << **i;
                    }
                    catch (Error const &error) {
                        *sink << String("Exception during log flush:\n") +
                                        error.what() + "\n(the entry format is: '" +
                                        (*i)->format() + "')";
                    }
                }
            }
        }
        d->toBeFlushed.clear();
        foreach (LogSink *sink, d->sinks) sink->flush();
    }
    d->lastFlushedAt = Time();
    while (d->entries.size() > d->maxEntryCount) {
        LogEntry *old = d->entries.front();
        d->entries.pop_front();
        delete old;
    }
}
void LogBuffer::setAppBuffer(LogBuffer &appBuffer) {
    _appBuffer = &appBuffer;
}
LogBuffer &LogBuffer::get() {
    DENG2_ASSERT(_appBuffer != 0);
    return *_appBuffer;
}
bool LogBuffer::appBufferExists() {
    return _appBuffer != 0;
}
} // namespace de
namespace de {
namespace internal {
    enum FilterId {
        GenericFilter,
        ResourceFilter,
        MapFilter,
        ScriptFilter,
        GLFilter,
        AudioFilter,
        InputFilter,
        NetworkFilter,
        NUM_FILTERS
    };
    static char const *subRecName[NUM_FILTERS] = { // for Config
        "generic",
        "resource",
        "map",
        "script",
        "gl",
        "audio",
        "input",
        "network"
    };
}
using namespace internal;
/**
 * Filter for determining which log entries will be put in the log buffer.
 */
DENG2_PIMPL_NOREF(LogFilter) {
    struct Filter {
        int domainBit;
        LogEntry::Level minLevel;
        bool allowDev;
        Filter()
            : domainBit(LogEntry::GenericBit)
            , minLevel(LogEntry::Message)
            , allowDev(false)
        {}
        inline bool checkContextBit(duint32 md) const {
            return (md & (1 << domainBit)) != 0;
        }
        void read(Record const &rec) {
            minLevel = LogEntry::Level(dint(rec["minLevel"].value().asNumber()));
            allowDev = rec["allowDev"].value().isTrue();
        }
        void write(Record &rec) const {
            rec.set("minLevel", dint(minLevel));
            rec.set("allowDev", allowDev);
        }
    };
    Filter filterByContext[NUM_FILTERS];
    Impl() {
        for (int i = 0; i < NUM_FILTERS; ++i) {
            filterByContext[i].domainBit = LogEntry::FirstDomainBit + i;
        }
    }
    bool isLogEntryAllowed(duint32 md) const {
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            Filter const &ftr = filterByContext[i];
            if (ftr.checkContextBit(md)) {
                if ((md & LogEntry::Dev) && !ftr.allowDev) continue; // No devs.
                if (ftr.minLevel <= int(md & LogEntry::LevelMask)) {
                    return true;
                }
                if ((md & LogEntry::Interactive) && i == ScriptFilter) {
                    return true;
                }
            }
        }
        return false;
    }
    LogEntry::Level minLevel(duint32 md) const {
        int lev = LogEntry::HighestLogLevel + 1;
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            Filter const &ftr = filterByContext[i];
            if (ftr.checkContextBit(md)) {
                lev = de::min(lev, int(ftr.minLevel));
            }
        }
        return LogEntry::Level(lev);
    }
    bool allowDev(duint32 md) const {
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            Filter const &ftr = filterByContext[i];
            if (ftr.checkContextBit(md)) {
                if (ftr.allowDev) return true;
            }
        }
        return false;
    }
    void setAllowDev(duint32 md, bool allow) {
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            Filter &ftr = filterByContext[i];
            if (ftr.checkContextBit(md)) {
                ftr.allowDev = allow;
            }
        }
    }
    void setMinLevel(duint32 md, LogEntry::Level level) {
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            Filter &ftr = filterByContext[i];
            if (ftr.checkContextBit(md)) {
                ftr.minLevel = level;
            }
        }
    }
    void read(Record const &rec) {
        try {
            for (uint i = 0; i < NUM_FILTERS; ++i) {
                filterByContext[i].read(rec.subrecord(subRecName[i]));
            }
        }
        catch (Error const &er) {
            LOGDEV_WARNING("Failed to read filter from record: %s\nThe record is:\n%s")
                    << er.asText() << rec.asText();
            LOG_WARNING("Log filter reset to defaults");
            *this = Impl(); // Reset.
        }
    }
    void write(Record &rec) const {
        for (uint i = 0; i < NUM_FILTERS; ++i) {
            if (!rec.hasSubrecord(subRecName[i])) {
                rec.add(subRecName[i], new Record);
            }
            filterByContext[i].write(rec.subrecord(subRecName[i]));
        }
    }
};
LogFilter::LogFilter() : d(new Impl)
{}
bool LogFilter::isLogEntryAllowed(duint32 metadata) const {
    DENG2_ASSERT(metadata & LogEntry::DomainMask); // must have a domain
    return d->isLogEntryAllowed(metadata);
}
void LogFilter::setAllowDev(duint32 md, bool allow) {
    d->setAllowDev(md, allow);
}
void LogFilter::setMinLevel(duint32 md, LogEntry::Level level) {
    d->setMinLevel(md, level);
}
bool LogFilter::allowDev(duint32 md) const {
    return d->allowDev(md);
}
LogEntry::Level LogFilter::minLevel(duint32 md) const {
    return d->minLevel(md);
}
void LogFilter::read(Record const &rec) {
    d->read(rec);
}
void LogFilter::write(Record &rec) const {
    d->write(rec);
}
String LogFilter::domainRecordName(LogEntry::Context domain) {
    for (int i = LogEntry::FirstDomainBit; i <= LogEntry::LastDomainBit; ++i) {
        if (domain & (1 << i)) return subRecName[i - LogEntry::FirstDomainBit];
    }
    return "";
}
} // namespace de
namespace de {
LogSink::LogSink()
    : _formatter(0),
      _mode(Enabled)
{}
LogSink::LogSink(LogSink::IFormatter &formatter)
    : _formatter(&formatter),
      _mode(Enabled)
{}
LogSink::~LogSink()
{}
void LogSink::setMode(Mode mode) {
    _mode = mode;
}
LogSink::Mode LogSink::mode() const {
    return _mode;
}
bool LogSink::willAccept(LogEntry const &entry) const {
    switch (_mode) {
    case Enabled:
        return true;
    case Disabled:
        return false;
    case OnlyNormalEntries:
        return entry.level() < LogEntry::Warning;
    case OnlyWarningEntries:
        return entry.level() >= LogEntry::Warning;
    }
    return false;
}
LogSink::IFormatter *LogSink::formatter() {
    return _formatter;
}
LogSink &LogSink::operator << (LogEntry const &entry) {
    DENG2_ASSERT(formatter());
    foreach (String line, formatter()->logEntryToTextLines(entry)) {
        *this << line;
    }
    return *this;
}
} // namespace de
namespace de {
static Loop *loopSingleton = 0;
DENG2_PIMPL(Loop) {
    TimeSpan interval;
    bool running;
    QTimer *timer;
    LoopCallback mainCall;
    Impl(Public *i) : Base(i), interval(0), running(false) {
        DENG2_ASSERT(!loopSingleton);
        loopSingleton = i;
        audienceForIteration.setAdditionAllowedDuringIteration(true);
        timer = new QTimer(thisPublic);
        QObject::connect(timer, SIGNAL(timeout()), thisPublic, SLOT(nextLoopIteration()));
    }
    ~Impl() {
        loopSingleton = 0;
    }
    DENG2_PIMPL_AUDIENCE(Iteration)
};
DENG2_AUDIENCE_METHOD(Loop, Iteration)
Loop::Loop() : d(new Impl(this))
{}
void Loop::setRate(double freqHz) {
    if (fequal(freqHz, 0.0)) {
        freqHz = 1000.0;
    }
    d->interval = 1.0 / freqHz;
    d->timer->setInterval(de::max(1, int(d->interval.asMilliSeconds())));
}
double Loop::rate() const {
    return 1.0 / d->interval;
}
void Loop::start() {
    d->running = true;
    d->timer->start();
}
void Loop::stop() {
    d->running = false;
    d->timer->stop();
}
void Loop::pause() {
    d->timer->stop();
}
void Loop::resume() {
    d->timer->start();
}
void Loop::timer(TimeSpan const &delay, std::function<void ()> func) {
    internal::CallbackTimer *timer = new internal::CallbackTimer(func, qApp);
    timer->start(delay.asMilliSeconds());
}
void Loop::mainCall(std::function<void ()> func) // static
{
    if (App::inMainThread()) {
        func();
    }
    else {
        Loop::get().d->mainCall.enqueue(func);
    }
}
Loop &Loop::get() {
    DENG2_ASSERT(loopSingleton != 0);
    return *loopSingleton;
}
void Loop::nextLoopIteration() {
    try {
        if (d->running) {
            DENG2_FOR_AUDIENCE2(Iteration, i) i->loopIteration();
        }
    }
    catch (Error const &er) {
        LOG_AS("Loop");
        App::app().handleUncaughtException("Uncaught exception during loop iteration:\n" + er.asText());
    }
}
LoopCallback::LoopCallback()
{}
LoopCallback::~LoopCallback()
{}
bool LoopCallback::isEmpty() const {
    DENG2_GUARD(this);
    return _funcs.isEmpty();
}
void LoopCallback::enqueue(Callback func) {
    DENG2_GUARD(this);
    _funcs << func;
    Loop::get().audienceForIteration() += this;
}
void LoopCallback::loopIteration() {
    QList<Callback> funcs; {
        DENG2_GUARD(this);
        Loop::get().audienceForIteration() -= this;
        funcs = _funcs;
        _funcs.clear();
    }
    for (Callback const &cb : funcs) {
        cb();
    }
}
} // namespace de
namespace de {
MemoryLogSink::MemoryLogSink(LogEntry::Level minimumLevel)
    : _minLevel(minimumLevel)
    , _privileged(false)
{}
MemoryLogSink::~MemoryLogSink() {
    DENG2_GUARD(this);
    qDeleteAll(_entries);
}
void MemoryLogSink::setPrivileged(bool onlyPrivileged) {
    _privileged = onlyPrivileged;
}
void MemoryLogSink::clear() {
    DENG2_GUARD(this);
    qDeleteAll(_entries);
    _entries.clear();
}
LogSink &MemoryLogSink::operator << (LogEntry const &entry) {
    if ((!_privileged &&  (entry.context() & LogEntry::Privileged)) ||
        (_privileged && !(entry.context() & LogEntry::Privileged))) {
        return *this;
    }
    if (entry.level() >= _minLevel) {
        DENG2_GUARD(this);
        _entries.append(new LogEntry(entry));
        addedNewEntry(*_entries.back());
    }
    return *this;
}
LogSink &MemoryLogSink::operator << (String const &) {
    return *this;
}
int MemoryLogSink::entryCount() const {
    DENG2_GUARD(this);
    return _entries.size();
}
LogEntry const &MemoryLogSink::entry(int index) const {
    DENG2_GUARD(this);
    DENG2_ASSERT(index >= 0);
    DENG2_ASSERT(index < _entries.size());
    return *_entries.at(index);
}
void MemoryLogSink::remove(int pos, int n) {
    DENG2_GUARD(this);
    while (n-- > 0) {
        delete _entries.takeAt(pos);
    }
}
void MemoryLogSink::addedNewEntry(LogEntry &)
{}
} // namespace de
namespace de {
/**
 * Parses the tab escapes in a styled text and fills them using spaces. All lines of the
 * input text are handled, however they must be separated by newline characters.
 */
struct TabFiller
        : DENG2_OBSERVES(EscapeParser, PlainText)
        , DENG2_OBSERVES(EscapeParser, EscapeSequence) {
    EscapeParser esc;
    QStringList lines;
    String current;
    bool hasTabs;
    TabFiller(String const &text) : hasTabs(false) {
        esc.audienceForPlainText() += this;
        esc.audienceForEscapeSequence() += this;
        esc.parse(text);
        if (!current.isEmpty()) lines << current;
    }
    void handlePlainText(Rangei const &range) {
        for (int i = range.start; i < range.end; ++i) {
            QChar ch = esc.originalText().at(i);
            if (ch == '\n') {
                lines << current;
                current.clear();
                current.reserve(80); // Prepare for per-character appending.
            }
            else {
                current.append(ch);
            }
        }
    }
    void handleEscapeSequence(Rangei const &range) {
        String seq = esc.originalText().substr(range);
        if (seq.at(0) == '\t') {
            current.append("\t+");
            hasTabs = true;
        }
        else if (seq.at(0) == 'T') {
            current.append('\t');
            current.append(seq.at(1));
            hasTabs = true;
        }
    }
    int highestTabStop() const {
        int maxStop = 0;
        foreach (QString const &ln, lines) {
            int stop = 0;
            for (int i = 0; i < ln.size(); ++i) {
                if (ln.at(i) == '\t') {
                    ++i;
                    if (ln.at(i) == '+') {
                        stop++;
                    }
                    else if (ln.at(i) == '`') {
                        stop = 0;
                    }
                    else {
                        stop = ln.at(i).toLatin1() - 'a';
                    }
                    maxStop = max(stop, maxStop);
                }
            }
        }
        return maxStop;
    }
    bool fillTabs(QStringList &fills, int maxStop, int minIndent) const {
        Vector2i resetAt(-1, -1);
        for (int stop = 0; stop <= maxStop; ++stop) {
            int tabWidth = 0;
            for (int idx = 0; idx < fills.size(); ++idx) {
                String const &ln = fills.at(idx);
                int w = (idx > 0? minIndent : 0);
                for (int i = 0; i < ln.size(); ++i) {
                    if (ln.at(i) == '\t') {
                        ++i;
                        if (ln.at(i) == '`') {
                            goto replaceTabs;
                        }
                        if (ln.at(i) == '+' || ln.at(i).toLatin1() - 'a' == stop) {
                            tabWidth = max(tabWidth, w);
                        }
                        continue;
                    }
                    else {
                        ++w;
                    }
                }
            }
replaceTabs:
            for (int idx = 0; idx < fills.size(); ++idx) {
                QString &ln = fills[idx];
                int w = (idx > 0? minIndent : 0);
                for (int i = 0; i < ln.size(); ++i) {
                    if (ln.at(i) == '\t') {
                        ++i;
                        if (ln.at(i) == '`') {
                            resetAt = Vector2i(idx, i - 1);
                            goto nextStop;
                        }
                        if (ln.at(i) == '+' || ln.at(i).toLatin1() - 'a' == stop) {
                            ln.remove(--i, 2);
                            ln.insert(i, String(max(0, tabWidth - w), QChar(' ')));
                        }
                        continue;
                    }
                    else {
                        ++w;
                    }
                }
            }
nextStop:;
        }
        if (resetAt.x >= 0) {
            fills[resetAt.x].remove(resetAt.y, 2);
            return false;
        }
        return true;
    }
    String filled(int minIndent) const {
        if (!hasTabs) return esc.plainText();
        int const maxStop = highestTabStop();
        QStringList fills = lines;
        while (!fillTabs(fills, maxStop, minIndent)) {}
#if 0
#ifdef DENG2_DEBUG
        foreach (QString ln, fills) DENG2_ASSERT(!ln.contains('\t'));
#endif
#endif
        return fills.join("\n");
    }
};
MonospaceLogSinkFormatter::MonospaceLogSinkFormatter()
    : _maxLength(100), _minimumIndent(0), _sectionDepthOfPreviousLine(0) {
#ifdef DENG2_DEBUG
    _maxLength = 125;
    _minimumIndent = 15;
#endif
}
QList<String> MonospaceLogSinkFormatter::logEntryToTextLines(LogEntry const &entry) {
    QList<String> resultLines;
    String const &section = entry.section();
    int cutSection = 0;
#ifndef DENG2_DEBUG
    LogEntry::Flags entryFlags = LogEntry::Simple;
#else
    LogEntry::Flags entryFlags;
#endif
    if (!_sectionOfPreviousLine.isEmpty()
            && entry.sectionDepth() >= 1
            && _sectionDepthOfPreviousLine <= entry.sectionDepth()) {
        if (_sectionOfPreviousLine == section) {
            entryFlags |= LogEntry::SectionSameAsBefore;
        }
        else if (section.startsWith(_sectionOfPreviousLine)) {
            cutSection = _sectionOfPreviousLine.size();
            entryFlags |= LogEntry::SectionSameAsBefore;
        }
        else {
            int prefix = section.commonPrefixLength(_sectionOfPreviousLine);
            if (prefix > 5) {
                entryFlags |= LogEntry::AbbreviateSection;
                cutSection = prefix;
            }
        }
    }
    String message = TabFiller(entry.asText(entryFlags, cutSection)).filled(_minimumIndent);
    _sectionOfPreviousLine      = section;
    _sectionDepthOfPreviousLine = entry.sectionDepth();
    int wrapIndent = -1;
    int nextWrapIndent = -1;
    String::size_type pos = 0;
    while (pos != String::npos) {
        String::size_type next = message.indexOf('\n', pos);
        duint lineLen = (next == String::npos? message.size() - pos : next - pos);
        duint const maxLen = (pos > 0? _maxLength - wrapIndent : _maxLength);
        if (lineLen > maxLen) {
            next = pos + maxLen;
            lineLen = maxLen;
            int checkPos = pos + maxLen;
            while (checkPos > pos) {
                if (message[checkPos].isSpace() ||
                        (message[checkPos].isPunct() && message[checkPos] != '.' &&
                         message[checkPos] != ','    && message[checkPos] != '-' &&
                         message[checkPos] != '\''   && message[checkPos] != '"' &&
                         message[checkPos] != '('    && message[checkPos] != ')' &&
                         message[checkPos] != '['    && message[checkPos] != ']' &&
                         message[checkPos] != '_')) {
                    if (!message[checkPos].isSpace()) {
                        checkPos++;
                    }
                    next = checkPos;
                    lineLen = checkPos - pos;
                    break;
                }
                checkPos--;
            }
        }
        String lineText = message.substr(pos, lineLen);
        if (pos > 0) {
            lineText = QString(wrapIndent, QChar(' ')) + lineText;
        }
        bool const lineStartsWithSpace = lineText.isEmpty() || lineText[0].isSpace();
        int firstNonSpace = -1;
        if (nextWrapIndent < 0 && !lineStartsWithSpace) {
            int w = _minimumIndent;
            int firstBracket = -1;
            for (; w < lineText.size(); ++w) {
                if (w < lineText.size() - 1 && lineText[w + 1].isSpace()) {
                    if (firstBracket == -1 && lineText[w] == ']') firstBracket = w;
                }
                if (firstNonSpace == -1 && !lineText[w].isSpace())
                    firstNonSpace = w;
            }
            if (firstBracket > 0) {
                nextWrapIndent = firstBracket + 2;
            }
            else if (firstNonSpace > 0) {
                nextWrapIndent = firstNonSpace;
            }
            else {
                nextWrapIndent = _minimumIndent;
            }
        }
        lineText.replace(DENG2_ESC("R"), String(maxLen - _minimumIndent, '-'));
        resultLines.append(lineText);
        wrapIndent = nextWrapIndent;
        pos = next;
        if (pos != String::npos && message[pos].isSpace()) {
            if (message[pos] == '\n') {
            }
            pos++; // Skip whitespace.
        }
    }
    return resultLines;
}
void MonospaceLogSinkFormatter::setMaxLength(duint maxLength) {
    _maxLength = de::max(duint(_minimumIndent + 10), maxLength);
}
duint MonospaceLogSinkFormatter::maxLength() const {
    return _maxLength;
}
} // namespace de
namespace de {
Slope::Slope(float angle, float slope)
    : angle(angle)
    , slope(slope)
{}
Vector3f Slope::toUnitVec() const {
    return Vector3f(std::cos(angle), std::sin(angle), slope).normalize();
}
Slope Slope::fromVec(const Vector3f &vector) {
    const auto vec = vector.normalize();
    if (fequal(vec.x, 0.0f) && fequal(vec.y, 0.0f)) {
        return {0.0f, vec.z > 0 ? std::numeric_limits<float>::max() :
                                  std::numeric_limits<float>::min()};
    }
    return {atan2(vec.y, vec.x), float(vec.z / vec.xy().length())};
}
} // namespace de
namespace de {
DENG2_PIMPL(System) {
    Flags behavior;
    Impl(Public *i) : Base(*i)
    {}
};
System::System(Flags const &behavior) : d(new Impl(this)) {
    d->behavior = behavior;
}
void System::setBehavior(Flags const &behavior, FlagOp operation) {
    applyFlagOperation(d->behavior, behavior, operation);
}
System::Flags System::behavior() const {
    return d->behavior;
}
bool System::processEvent(Event const &) {
    return false;
}
void System::timeChanged(Clock const &)
{}
} // namespace de
namespace de {
DENG2_PIMPL(TextApp)
, DENG2_OBSERVES(Loop, Iteration) {
    Loop loop;
    Impl(Public *i) : Base(i) {
        loop.audienceForIteration() += this;
        loop.setRate(35);
    }
    void loopIteration() override {
        Time::updateCurrentHighPerformanceTime();
        Clock::get().setTime(Time());
    }
};
TextApp::TextApp(int &argc, char **argv)
    : QCoreApplication(argc, argv)
    , App(applicationFilePath(), arguments())
    , d(new Impl(this))
{}
void TextApp::setMetadata(String const &orgName, String const &orgDomain,
                          String const &appName, String const &appVersion) {
    setName(appName);
    setOrganizationName  (orgName);
    setOrganizationDomain(orgDomain);
    setApplicationName   (appName);
    setApplicationVersion(appVersion);
}
bool TextApp::notify(QObject *receiver, QEvent *event) {
    try {
        return QCoreApplication::notify(receiver, event);
    }
    catch (std::exception const &error) {
        handleUncaughtException(error.what());
    }
    catch (...) {
        handleUncaughtException("de::TextApp caught exception of unknown type.");
    }
    return false;
}
int TextApp::execLoop() {
    LOGDEV_NOTE("Starting TextApp event loop...");
    d->loop.start();
    int code = QCoreApplication::exec();
    LOGDEV_NOTE("TextApp event loop exited with code %i") << code;
    return code;
}
void TextApp::stopLoop(int code) {
    d->loop.stop();
    return QCoreApplication::exit(code);
}
Loop &TextApp::loop() {
    return d->loop;
}
NativePath TextApp::appDataPath() const {
    return NativePath(QDir::homePath()) / unixHomeFolderName();
}
} // namespace de
namespace de {
TextStreamLogSink::TextStreamLogSink(QTextStream *ts)
    : LogSink(_format), _ts(ts) {
    _ts->setCodec("UTF-8");
}
TextStreamLogSink::~TextStreamLogSink() {
    delete _ts;
}
LogSink &TextStreamLogSink::operator << (String const &plainText) {
    *_ts << plainText + "\n";
    return *this;
}
void TextStreamLogSink::flush() {
    _ts->flush();
}
} // namespace de
namespace de {
namespace internal {
class Infos {
    Info *etcInfo;
    Info *userInfo;
public:
    Infos(String fileName) : etcInfo(0), userInfo(0) {
        String fn = String("/etc") / App::app().unixEtcFolderName() / fileName;
        if (QFile::exists(fn)) {
            etcInfo = new Info;
            etcInfo->parseNativeFile(fn);
        }
        fn = String(QDir::homePath()) / App::app().unixHomeFolderName() / fileName;
        if (QFile::exists(fn)) {
            userInfo = new Info;
            userInfo->parseNativeFile(fn);
        }
    }
    ~Infos() {
        delete userInfo;
        delete etcInfo;
    }
    bool find(String const &key, String &value) const {
        if (userInfo && userInfo->findValueForKey(key, value)) {
            return true;
        }
        if (etcInfo && etcInfo->findValueForKey(key, value)) {
            return true;
        }
        return false;
    }
};
} // namespace internal
using namespace internal;
DENG2_PIMPL_NOREF(UnixInfo) {
    Infos *paths;
    Infos *defaults;
    Impl() : paths(0), defaults(0)
    {}
    ~Impl() {
        delete paths;
        delete defaults;
    }
};
UnixInfo::UnixInfo() : d(new Impl) {
#ifdef UNIX
    /**
     * @note There is only the "paths" and "plugins" config files for now; more
     * could be added for different purposes. There could also be .ds scripts
     * for configuration.
     */
    d->paths    = new Infos("paths");
    d->defaults = new Infos("defaults");
#endif
}
bool UnixInfo::path(String const &key, NativePath &value) const {
    if (d->paths) {
        String s;
        if (d->paths->find(key, s)) {
            value = NativePath(s).expand();
            return true;
        }
    }
    return false;
}
bool UnixInfo::defaults(String const &key, String &value) const {
    if (d->defaults) {
        return d->defaults->find(key, value);
    }
    return false;
}
} // namespace de
