﻿/** @file remote/link.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/Link"
#include "de/RemoteFeedRelay"
#include "de/FileSystem"
#include <de/App>
#include <de/Async>
#include <de/Garbage>
#include <de/charsymbols.h>
/** @file remote/nativelink.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/NativeLink"
#include "de/App"
#include "de/FileSystem"
#include "de/Loop"
#include "de/Message"
#include "de/RemoteFeed"
#include "de/RemoteFeedProtocol"
#include "de/Socket"
/** @file remote/query.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/Query"
/** @file remotefeed.cpp  Feed for remote files.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RemoteFeed"
#include "de/BlockValue"
#include "de/Folder"
#include "de/Path"
#include "de/RecordValue"
#include "de/RemoteFeedProtocol"
#include "de/RemoteFeedRelay"
#include "de/RemoteFile"
#include "de/Socket"
#include "de/TimeValue"
#include "de/Waitable"
#include "de/charsymbols.h"
/** @file remotefeedprotocol.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RemoteFeedProtocol"
#include "de/BlockValue"
#include "de/TextValue"
#include "de/RecordValue"
#include "de/Folder"
/** @file remotefeedrelay.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RemoteFeedRelay"
#include "de/App"
#include "de/Async"
#include "de/Date"
#include "de/DictionaryValue"
#include "de/filesys/Link"
#include "de/filesys/NativeLink"
#include "de/Loop"
#include "de/Message"
#include "de/RemoteFeedProtocol"
#include "de/Version"
#include "de/charsymbols.h"
#include <QNetworkAccessManager>
#include <QNetworkDiskCache>
#include <QNetworkReply>
#include <QRegularExpression>
#include <QStandardPaths>
#include <QFile>
/** @file remotefile.cpp  Remote file.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RemoteFile"
#include "de/App"
#include "de/DirectoryFeed"
#include "de/FileSystem"
#include "de/RecordValue"
#include "de/RemoteFeedRelay"
#include "de/ScriptSystem"
#include "de/TextValue"
#include "de/TimeValue"
/** @file
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/WebHostedLink"
#include "de/Async"
#include "de/Folder"
#include "de/PathTree"
#include "de/RecordValue"
#include "de/RemoteFeedRelay"
#include "de/TextValue"
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QNetworkDiskCache>

namespace de {
namespace filesys {
DENG2_PIMPL(Link), public AsyncScope {
    String localRootPath;
    State state = Initializing;
    String address;
    QueryId nextQueryId = 1;
    QList<Query> deferredQueries;
    QHash<QueryId, Query> pendingQueries;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        cancelAllQueries();
    }
    void cancelAllQueries() {
        for (auto i = deferredQueries.begin(); i != deferredQueries.end(); ++i) {
            i->cancel();
        }
        for (auto i = pendingQueries.begin(); i != pendingQueries.end(); ++i) {
            i.value().cancel();
        }
    }
    void cleanup() {
        QMutableHashIterator<QueryId, Query> iter(pendingQueries);
        while (iter.hasNext()) {
            iter.next();
            if (!iter.value().isValid()) {
                iter.remove();
            }
        }
    }
    void sendQuery(Query query) {
        try {
            query.id = nextQueryId++;
            if (state == Ready) {
                self().transmit(query);
                pendingQueries.insert(query.id, query);
                cleanup();
            }
            else {
                deferredQueries.append(query);
            }
        }
        catch (Error const &er) {
            LOG_NET_ERROR("Error sending file repository query: %s")
                    << er.asText();
        }
    }
    void sendDeferredQueries() {
        foreach (Query const &query, deferredQueries) {
            if (query.isValid()) {
                pendingQueries.insert(query.id, query);
                self().transmit(query);
            }
        }
        deferredQueries.clear();
    }
    void notifyStatus(RemoteFeedRelay::Status status) {
        using Relay = RemoteFeedRelay;
        DENG2_FOR_EACH_OBSERVER(Relay::StatusAudience, i,
                                Relay::get().audienceForStatus()) {
            i->remoteRepositoryStatusChanged(address, status);
        }
    }
};
Link::Link(String const &address)
    : d(new Impl(this)) {
    d->address = address;
}
Link::~Link() {
    d->cancelAllQueries();
    if (d->state != Deinitialized) {
        if (auto *folder = FS::tryLocate<Folder>(d->localRootPath)) {
            delete folder;
        }
    }
}
void Link::setLocalRoot(String const &rootPath) {
    d->localRootPath = rootPath;
    localRoot();
}
Folder &Link::localRoot() const {
    return FS::get().makeFolder(d->localRootPath, FS::DontInheritFeeds);
}
String Link::address() const {
    return d->address;
}
Link::State Link::state() const {
    return d->state;
}
StringList Link::categoryTags() const {
    return StringList();
}
void Link::wasConnected() {
    DENG2_ASSERT_IN_MAIN_THREAD();
    d->state = Ready;
    d->sendDeferredQueries();
    d->notifyStatus(RemoteFeedRelay::Connected);
}
void Link::wasDisconnected() {
    DENG2_ASSERT_IN_MAIN_THREAD();
    d->state = Deinitialized;
    d->cancelAllQueries();
    d->cleanup();
    d->notifyStatus(RemoteFeedRelay::Disconnected);
    if (Folder *root = FS::tryLocate<Folder>(d->localRootPath)) {
        trash(root);
    }
}
void Link::handleError(QString errorMessage) {
    LOG_NET_ERROR("Error accessing remote file repository \"%s\": %s " DENG2_CHAR_MDASH
                  " files from repository may not be available")
            << d->address
            << errorMessage;
}
AsyncScope &Link::scope() {
    return *d;
}
void Link::cancelAllQueries() {
    d->cancelAllQueries();
}
void Link::cleanupQueries() {
    d->cleanup();
}
Query *Link::findQuery(QueryId id) {
    auto found = d->pendingQueries.find(id);
    if (found != d->pendingQueries.end()) {
        return &found.value();
    }
    return nullptr;
}
QueryId Link::sendQuery(Query query) {
    try {
        query.id = d->nextQueryId++;
        if (d->state == Ready) {
            d->pendingQueries.insert(query.id, query);
            transmit(query);
            d->cleanup();
        }
        else {
            d->deferredQueries.append(query);
        }
        return query.id;
    }
    catch (Error const &er) {
        LOG_NET_ERROR("Error sending file repository query: %s") << er.asText();
        return 0;
    }
}
File *Link::populateRemotePath(String const &, RepositoryPath const &path) const {
    return FS::tryLocate<File>(path.localPath);
}
void Link::metadataReceived(QueryId id, DictionaryValue const &metadata) {
    if (auto *query = findQuery(id)) {
        if (query->fileMetadata) {
            query->fileMetadata->call(metadata);
        }
        d->pendingQueries.remove(id);
    }
}
void Link::chunkReceived(QueryId id, duint64 startOffset, Block const &chunk, duint64 fileSize) {
    if (auto *query = findQuery(id)) {
        if (!query->isValid()) {
            d->pendingQueries.remove(id);
            return;
        }
        if (!query->fileSize) {
            query->fileContents->call(0, Block(), fileSize);
        }
        query->fileSize = fileSize;
        query->receivedBytes += chunk.size();
        query->fileContents->call(startOffset, chunk,
                                  fileSize - query->receivedBytes);
        if (fileSize == query->receivedBytes) {
            d->pendingQueries.remove(id);
        }
    }
}
} // namespace filesys
} // namespace de
namespace de {
namespace filesys {
String const NativeLink::URL_SCHEME("doomsday:");
static String const PATH_SERVER_REPOSITORY_ROOT("/sys/server/public"); // serverside folder
DENG2_PIMPL(NativeLink) {
    RemoteFeedProtocol protocol;
    Socket socket;
    Impl(Public *i) : Base(i) {}
    void receiveMessages() {
        while (socket.hasIncoming()) {
            try {
                DENG2_ASSERT_IN_MAIN_THREAD();
                std::unique_ptr<Message> response(socket.receive());
                std::unique_ptr<Packet>  packet  (protocol.interpret(*response));
                if (!packet) continue;
                switch (protocol.recognize(*packet)) {
                case RemoteFeedProtocol::Metadata: {
                    auto const &md = packet->as<RemoteFeedMetadataPacket>();
                    self().metadataReceived(md.id(), md.metadata());
                    break; }
                case RemoteFeedProtocol::FileContents: {
                    auto const &fc = packet->as<RemoteFeedFileContentsPacket>();
                    self().chunkReceived(fc.id(), fc.startOffset(), fc.data(), fc.fileSize());
                    break; }
                default:
                    break;
                }
            }
            catch (Error const &er) {
                LOG_NET_ERROR("Error when handling remote feed response: %s")
                        << er.asText();
            }
        }
        self().cleanupQueries();
    }
};
NativeLink::NativeLink(String const &address)
    : Link(address)
    , d(new Impl(this)) {
    DENG2_ASSERT(address.startsWith(URL_SCHEME));
    QObject::connect(&d->socket, &Socket::connected,     [this] () { wasConnected(); });
    QObject::connect(&d->socket, &Socket::disconnected,  [this] () { wasDisconnected(); });
    QObject::connect(&d->socket, &Socket::error,         [this] (QString msg) { handleError(msg); });
    QObject::connect(&d->socket, &Socket::messagesReady, [this] () { d->receiveMessages(); });
    d->socket.open(address.mid(URL_SCHEME.size()));
}
Link *NativeLink::construct(String const &address) {
    if (address.startsWith(URL_SCHEME)) {
        return new NativeLink(address);
    }
    return nullptr;
}
void NativeLink::setLocalRoot(String const &rootPath) {
    Link::setLocalRoot(rootPath);
    auto &root = localRoot();
    root.attach(new RemoteFeed(address(), PATH_SERVER_REPOSITORY_ROOT));
    root.populate(Folder::PopulateAsyncFullTree);
}
PackagePaths NativeLink::locatePackages(StringList const &packageIds) const {
    PackagePaths remotePaths;
    foreach (String pkg, packageIds) {
        if (File *rem = FS::tryLocate<File>("/remote/server"/pkg)) {
            remotePaths.insert(pkg, RepositoryPath(*this, rem->path(), ""));
        }
    }
    return remotePaths;
}
LoopResult filesys::NativeLink::forPackageIds(std::function<LoopResult (String const &)> func) const {
    return FS::locate<Folder>("/remote/server").forContents([&func] (String name, File &) -> LoopResult {
        if (auto result = func(name)) {
            return result;
        }
        return LoopContinue;
    });
}
void NativeLink::wasConnected() {
    d->socket << ByteRefArray("RemoteFeed", 10);
    Link::wasConnected();
}
void NativeLink::transmit(Query const &query) {
    DENG2_ASSERT(query.isValid());
    RemoteFeedQueryPacket packet;
    packet.setId(query.id);
    packet.setPath(query.path);
    if (query.fileMetadata) {
        packet.setQuery(RemoteFeedQueryPacket::ListFiles);
    }
    else if (query.fileContents) {
        packet.setQuery(RemoteFeedQueryPacket::FileContents);
    }
    d->socket.sendPacket(packet);
}
} // namespace filesys
} // namespace de
namespace de {
namespace filesys {
Query::Query(Request<FileMetadata> req, String path)
    : path(path), fileMetadata(req)
{}
Query::Query(Request<FileContents> req, String path)
    : path(path), fileContents(req)
{}
bool Query::isValid() const {
    if (fileMetadata) return fileMetadata->isValid();
    if (fileContents) return fileContents->isValid();
    return false;
}
void Query::cancel() {
    if (fileMetadata) fileMetadata->cancel();
    if (fileContents) fileContents->cancel();
}
} // namespace filesys
} // namespace de
namespace de {
static TimeSpan const POPULATE_TIMEOUT = 15.0;
DENG2_PIMPL(RemoteFeed)
, DENG2_OBSERVES(filesys::RemoteFeedRelay, Status) {
    String repository;
    Path remotePath;
    std::unique_ptr<DictionaryValue> fileMetadata;
    SafePtr<Folder> pendingPopulation;
    Impl(Public *i) : Base(i)
    {}
    PopulatedFiles populate() {
        PopulatedFiles populated;
        for (auto i : fileMetadata->elements()) {
            String const path = remotePath / i.first.value->asText();
            if (RecordValue const *meta = maybeAs<RecordValue>(i.second)) {
                Record const &md = *meta->record();
                File::Type const fileType = RemoteFeedMetadataPacket::toFileType(md.geti("type", 0));
                dsize      const fileSize = md.getui("size", 0);
                Time       const modTime  = md.getAs<TimeValue>("modifiedAt").time();
                File *file = nullptr;
                if (fileType == File::Type::File) {
                    file = new RemoteFile(path.fileName(), path, md.getAs<BlockValue>("metaId").block());
                }
                else {
                    Folder *subfolder = new Folder(path.fileName());
                    subfolder->attach(new RemoteFeed(*thisPublic, path));
                    file = subfolder;
                }
                if (md.has("package")) {
                    file->objectNamespace().add("package", new Record(md.subrecord("package")));
                }
                file->setStatus(File::Status(fileType, fileSize, modTime));
                file->setOriginFeed(thisPublic);
                populated << file;
            }
        }
        return populated;
    }
    void remoteRepositoryStatusChanged(String const &address, filesys::RemoteFeedRelay::Status status) override {
        if (repository == address && status == filesys::RemoteFeedRelay::Connected) {
            if (pendingPopulation) {
                pendingPopulation->populate(Folder::PopulateAsyncFullTree);
                pendingPopulation.reset();
            }
            filesys::RemoteFeedRelay::get().audienceForStatus() -= this;
        }
    }
};
RemoteFeed::RemoteFeed(String const &repository, String const &remotePath)
    : d(new Impl(this)) {
    d->repository = repository;
    d->remotePath = remotePath;
}
RemoteFeed::RemoteFeed(RemoteFeed const &parentFeed, String const &remotePath)
    : d(new Impl(this)) {
    d->repository = parentFeed.d->repository;
    d->remotePath = remotePath;
}
String RemoteFeed::repository() const {
    return d->repository;
}
String RemoteFeed::description() const {
    return String("remote repository \"%1\"")
            .arg(d->repository / d->remotePath);
}
Feed::PopulatedFiles RemoteFeed::populate(Folder const &folder) {
    LOG_AS("RemoteFeed");
    auto &relay = filesys::RemoteFeedRelay::get();
    PopulatedFiles files;
    if (!relay.isConnected(d->repository)) {
        d->pendingPopulation.reset(const_cast<Folder *>(&folder));
        relay.audienceForStatus() += d;
        return files;
    }
    auto request = relay.fetchFileList
            (d->repository,
             d->remotePath,
             [this, /*&folder,*/ &files]
             (DictionaryValue const &fileMetadata) {
        d->fileMetadata.reset(static_cast<DictionaryValue *>(fileMetadata.duplicate()));
        files = d->populate();
    });
    request->wait(POPULATE_TIMEOUT);
    return files;
}
bool RemoteFeed::prune(File &) const {
    return false;
}
} // namespace de
namespace de {
static Packet::Type const QUERY_PACKET_TYPE = Packet::typeFromString("RFQu");
RemoteFeedQueryPacket::RemoteFeedQueryPacket()
    : IdentifiedPacket(QUERY_PACKET_TYPE)
{}
void RemoteFeedQueryPacket::setQuery(Query query) {
    _query = query;
}
void RemoteFeedQueryPacket::setPath(String const &path) {
    _path = path;
}
RemoteFeedQueryPacket::Query RemoteFeedQueryPacket::query() const {
    return _query;
}
String RemoteFeedQueryPacket::path() const {
    return _path;
}
void RemoteFeedQueryPacket::operator >> (Writer &to) const {
    IdentifiedPacket::operator >> (to);
    to << duint8(_query) << _path;
}
void RemoteFeedQueryPacket::operator << (Reader &from) {
    IdentifiedPacket::operator << (from);
    from.readAs<duint8>(_query) >> _path;
}
Packet *RemoteFeedQueryPacket::fromBlock(Block const &block) {
    return constructFromBlock<RemoteFeedQueryPacket>(block, QUERY_PACKET_TYPE);
}
static Packet::Type const METADATA_PACKET_TYPE = Packet::typeFromString("RFMt");
RemoteFeedMetadataPacket::RemoteFeedMetadataPacket()
    : IdentifiedPacket(METADATA_PACKET_TYPE)
{}
void RemoteFeedMetadataPacket::addFile(File const &file, String const &prefix) {
    auto const &ns = file.target().objectNamespace();
    auto const status = file.target().status();
    std::unique_ptr<Record> fileMeta(new Record);
    fileMeta->addTime  ("modifiedAt", status.modifiedAt);
    fileMeta->addNumber("type",       status.type() == File::Type::File? 0 : 1);
    if (status.type() == File::Type::Folder) {
        fileMeta->addNumber("size", file.target().as<Folder>().contents().size());
    }
    else {
        fileMeta->addNumber("size", status.size);
        fileMeta->addBlock ("metaId").value<BlockValue>().block() = file.metaId();
    }
    if (ns.hasSubrecord("package")) {
        fileMeta->add("package", new Record(ns.getr("package").dereference(),
                      Record::IgnoreDoubleUnderscoreMembers));
    }
    _metadata.add(new TextValue(prefix / file.name()),
                  new RecordValue(fileMeta.release(), RecordValue::OwnsRecord));
}
void RemoteFeedMetadataPacket::addFolder(Folder const &folder, String prefix) {
    folder.forContents([this, prefix] (String, File &file) {
        addFile(file, prefix);
        return LoopContinue;
    });
}
DictionaryValue const &RemoteFeedMetadataPacket::metadata() const {
    return _metadata;
}
File::Type RemoteFeedMetadataPacket::toFileType(int value) {
    return (value == 0? File::Type::File : File::Type::Folder);
}
void RemoteFeedMetadataPacket::operator >> (Writer &to) const {
    IdentifiedPacket::operator >> (to);
    to << _metadata;
}
void RemoteFeedMetadataPacket::operator << (Reader &from) {
    IdentifiedPacket::operator << (from);
    from >> _metadata;
}
Packet *RemoteFeedMetadataPacket::fromBlock(Block const &block) {
    return constructFromBlock<RemoteFeedMetadataPacket>(block, METADATA_PACKET_TYPE);
}
static Packet::Type const FILE_CONTENTS_PACKET_TYPE = Packet::typeFromString("RFCo");
RemoteFeedFileContentsPacket::RemoteFeedFileContentsPacket()
    : IdentifiedPacket(FILE_CONTENTS_PACKET_TYPE)
    , _startOffset(0)
{}
void RemoteFeedFileContentsPacket::setData(Block const &data) {
    _data = data;
}
void RemoteFeedFileContentsPacket::setStartOffset(dsize offset) {
    _startOffset = offset;
}
void RemoteFeedFileContentsPacket::setFileSize(dsize size) {
    _fileSize = size;
}
Block const &RemoteFeedFileContentsPacket::data() const {
    return _data;
}
dsize RemoteFeedFileContentsPacket::startOffset() const {
    return _startOffset;
}
dsize RemoteFeedFileContentsPacket::fileSize() const {
    return _fileSize;
}
void RemoteFeedFileContentsPacket::operator >> (Writer &to) const {
    IdentifiedPacket::operator >> (to);
    to << duint64(_fileSize) << duint64(_startOffset) << _data;
}
void RemoteFeedFileContentsPacket::operator << (Reader &from) {
    IdentifiedPacket::operator << (from);
    from.readAs<duint64>(_fileSize)
        .readAs<duint64>(_startOffset)
        >> _data;
}
Packet *RemoteFeedFileContentsPacket::fromBlock(Block const &block) {
    return constructFromBlock<RemoteFeedFileContentsPacket>(block, FILE_CONTENTS_PACKET_TYPE);
}
RemoteFeedProtocol::RemoteFeedProtocol() {
    define(RemoteFeedQueryPacket::fromBlock);
    define(RemoteFeedMetadataPacket::fromBlock);
    define(RemoteFeedFileContentsPacket::fromBlock);
}
RemoteFeedProtocol::PacketType RemoteFeedProtocol::recognize(Packet const &packet) {
    if (packet.type() == QUERY_PACKET_TYPE) {
        DENG2_ASSERT(is<RemoteFeedQueryPacket>(&packet));
        return Query;
    }
    if (packet.type() == METADATA_PACKET_TYPE) {
        DENG2_ASSERT(is<RemoteFeedMetadataPacket>(&packet));
        return Metadata;
    }
    if (packet.type() == FILE_CONTENTS_PACKET_TYPE) {
        DENG2_ASSERT(is<RemoteFeedFileContentsPacket>(&packet));
        return FileContents;
    }
    return Unknown;
}
} // namespace de
namespace de {
namespace filesys {
DENG2_PIMPL(RemoteFeedRelay) {
    std::unique_ptr<QNetworkAccessManager> network;
    QList<Link::Constructor> linkConstructors;
    QHash<String, filesys::Link *> repositories; // owned
    Impl(Public *i) : Base(i) {
        network.reset(new QNetworkAccessManager);
        auto *cache = new QNetworkDiskCache;
        String const dir = NativePath(QStandardPaths::writableLocation(QStandardPaths::CacheLocation))
                / "RemoteFiles";
        cache->setCacheDirectory(dir);
        network->setCache(cache);
    }
    ~Impl() {
        qDeleteAll(repositories.values());
    }
    DENG2_PIMPL_AUDIENCE(Status)
};
DENG2_AUDIENCE_METHOD(RemoteFeedRelay, Status)
RemoteFeedRelay &RemoteFeedRelay::get() {
    return App::remoteFeedRelay();
}
RemoteFeedRelay::RemoteFeedRelay()
    : d(new Impl(this)) {
    defineLink(NativeLink::construct);
}
void RemoteFeedRelay::defineLink(filesys::Link::Constructor linkConstructor) {
    d->linkConstructors.push_front(linkConstructor);
}
void RemoteFeedRelay::addRepository(String const &address,
                                    String const &localRootPath) {
    for (auto constructor : d->linkConstructors) {
        if (auto *link = constructor(address)) {
            d->repositories.insert(address, link);
            link->setLocalRoot(localRootPath);
            return;
        }
    }
}
void RemoteFeedRelay::removeRepository(String const &address) {
    if (auto *repo = d->repositories.take(address)) {
        delete repo;
    }
}
Link *RemoteFeedRelay::repository(String const &address) const {
    auto found = d->repositories.constFind(address);
    if (found != d->repositories.constEnd()) {
        return found.value();
    }
    return nullptr;
}
StringList RemoteFeedRelay::repositories() const {
    StringList repos;
    foreach (String a, d->repositories.keys()) {
        repos << a;
    }
    return repos;
}
bool RemoteFeedRelay::isConnected(String const &address) const {
    if (auto *repo = repository(address)) {
        return repo->state() == Link::Ready;
    }
    return false;
}
PackagePaths RemoteFeedRelay::locatePackages(StringList const &packageIds) const {
    PackagePaths located;
    foreach (auto *repo, d->repositories) {
        if (repo->state() == Link::Ready) {
            auto const paths = repo->locatePackages(packageIds);
            for (auto i = paths.begin(); i != paths.end(); ++i) {
                if (!located.contains(i.key())) {
                    located.insert(i.key(), i.value());
                }
            }
        }
    }
    return located;
}
Request<FileMetadata>
RemoteFeedRelay::fetchFileList(String const &repository, String folderPath, FileMetadata metadataReceived) {
    DENG2_ASSERT(d->repositories.contains(repository));
    Waitable done;
    Request<FileMetadata> request;
    Loop::mainCall([&] () {
        auto *repo = d->repositories[repository];
        request.reset(new Request<FileMetadata>::element_type(metadataReceived));
        repo->sendQuery(Query(request, folderPath));
        done.post();
    });
    done.wait();
    return request;
}
Request<FileContents>
RemoteFeedRelay::fetchFileContents(String const &repository, String filePath, FileContents contentsReceived) {
    DENG2_ASSERT(d->repositories.contains(repository));
    Waitable done;
    Request<FileContents> request;
    Loop::mainCall([&] () {
        auto *repo = d->repositories[repository];
        request.reset(new Request<FileContents>::element_type(contentsReceived));
        repo->sendQuery(Query(request, filePath));
        done.post();
    });
    done.wait();
    return request;
}
QNetworkAccessManager &RemoteFeedRelay::network() {
    return *d->network;
}
} // namespace filesys
} // namespace de
namespace de {
using namespace de::filesys;
String const RemoteFile::CACHE_PATH = "/home/cache/remote";
DENG2_PIMPL(RemoteFile) {
    String remotePath;
    Block remoteMetaId;
    String repositoryAddress; // If empty, use feed's repository.
    Block buffer;
    Request<FileContents> fetching;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        if (fetching) {
            fetching->cancel();
        }
    }
    String cachePath() const {
        String const hex = remoteMetaId.asHexadecimalText();
        String path = CACHE_PATH / hex.right(1);
        String original = self().objectNamespace().gets("package.path", remotePath);
        return path / hex + "_" + original.fileName();
    }
    void findCachedFile(bool requireExists = true) {
        if (!self().isBroken()) return;
        if (self().state() == NotReady) {
            throw UnfetchedError("RemoteFile::operator >>",
                                 self().description() + " not downloaded");
        }
        if (self().isBroken()) {
            self().setTarget(FS::tryLocate<File const>(cachePath()));
        }
        if (requireExists && self().isBroken()) {
            throw InputError("RemoteFile::operator >>",
                             self().description() + " has no locally cached data");
        }
    }
    void setTarget(File const &cachedFile) {
        self().setTarget(cachedFile);
        auto &ns = self().objectNamespace();
        if (ns.has("package.path")) {
            ns["package.path"] = self().target().path();
        }
    }
    bool checkExistingCache() {
        if (File const *cached = FS::tryLocate<File const>(cachePath())) {
            if (cached->status() == self().status()) {
                LOG_RES_MSG("Using local cached copy of %s") << cached->description();
                setTarget(*cached);
                return true;
            }
        }
        return false;
    }
    String repository() const {
        if (repositoryAddress) {
            return repositoryAddress;
        }
        DENG2_ASSERT(is<RemoteFeed>(self().originFeed()));
        return self().originFeed()->as<RemoteFeed>().repository();
    }
};
RemoteFile::RemoteFile(String const &name, String const &remotePath, Block const &remoteMetaId,
                       String const &repositoryAddress)
    : LinkFile(name)
    , d(new Impl(this)) {
    objectNamespace().addSuperRecord(ScriptSystem::builtInClass(QStringLiteral("RemoteFile")));
    d->repositoryAddress = repositoryAddress;
    d->remotePath        = remotePath;
    d->remoteMetaId      = remoteMetaId;
    qDebug() << "RemoteFile remotePath:" << remotePath;
    setState(NotReady);
}
void RemoteFile::download() {
    if (state() != NotReady) return;
    setState(Recovering);
    if (d->checkExistingCache()) {
        DENG2_FOR_AUDIENCE(Download, i) {
            i->downloadProgress(*this, 0);
        }
        setState(Ready);
        return;
    }
    LOG_NET_MSG("Requesting download of \"%s\"") << name();
    d->fetching = filesys::RemoteFeedRelay::get().fetchFileContents
            (d->repository(),
             d->remotePath,
             [this] (duint64 startOffset, Block const &chunk, duint64 remainingBytes) {
        DENG2_ASSERT_IN_MAIN_THREAD();
        DENG2_FOR_AUDIENCE(Download, i) {
            i->downloadProgress(*this, remainingBytes);
        }
        if (d->buffer.size() < remainingBytes) {
            d->buffer.resize(remainingBytes);
        }
        d->buffer.set(startOffset, chunk.data(), chunk.size());
        if (remainingBytes == 0) {
            LOG_NET_MSG("\"%s\" downloaded (%i bytes)") << name() << d->buffer.size();
            d->fetching = nullptr;
            String const fn = d->cachePath();
            Folder &cacheFolder = FS::get().makeFolder(fn.fileNamePath());
            File &data = cacheFolder.replaceFile(fn);
            data << d->buffer;
            d->buffer.clear();
            data.flush(); {
                auto st = data.status();
                st.modifiedAt = status().modifiedAt;
                data.setStatus(st);
                DirectoryFeed::setFileModifiedTime(data.correspondingNativePath(),
                                                   st.modifiedAt);
            }
            setTarget(data.reinterpret());
            if (objectNamespace().has("package.path")) {
                objectNamespace()["package.path"] = target().path();
            }
            setState(Ready);
        }
    });
}
void RemoteFile::cancelDownload() {
    if (d->fetching) {
        d->fetching->cancel();
        d->fetching = nullptr;
        d->buffer.clear();
        setState(NotReady);
    }
}
void RemoteFile::deleteCache() {
    setState(NotReady);
    FS::get().root().tryDestroyFile(d->cachePath());
}
IIStream const &RemoteFile::operator >> (IByteArray &bytes) const {
    if (state() != Ready) {
        throw UnfetchedError("RemoteFile::operator >>",
                             description() + " not downloaded");
    }
    DENG2_ASSERT(!isBroken());
    return LinkFile::operator >> (bytes);
}
String RemoteFile::describe() const {
    if (isReady()) {
        return String("\"%1\"").arg(name());
    }
    String targetDesc;
    if (!isBroken()) {
        targetDesc = " cached in " + target().description();
    }
    return String("remote file \"%1\" (%2)")
            .arg(name())
            .arg(  state() == NotReady   ? "not ready"
                 : state() == Recovering ? "downloading"
                                         : "ready")
            + targetDesc;
}
Block RemoteFile::metaId() const {
    return d->remoteMetaId;
}
Asset &RemoteFile::asset() {
    return *this;
}
Asset const &RemoteFile::asset() const {
    return *this;
}
dsize RemoteFile::downloadSize() const {
    return size();
}
} // namespace de
namespace de {
namespace filesys {
DENG2_PIMPL(WebHostedLink), public Lockable {
    QSet<QNetworkReply *> pendingRequests;
    std::shared_ptr<FileTree> fileTree;
    Impl(Public *i) : Base(i)
    {}
    ~Impl()
    {}
    Block metaIdForFileEntry(FileEntry const &entry) const {
        if (entry.isBranch()) return Block(); // not applicable
        return md5Hash(self().address(), entry.path(), entry.size, entry.modTime);
    }
    void handleFileListQueryAsync(Query query) {
        QueryId const id = query.id;
        String const queryPath = query.path;
        self().scope() += async([this, queryPath] () -> std::shared_ptr<DictionaryValue> {
            DENG2_GUARD(this);
            if (auto const *dir = fileTree->tryFind
                    (queryPath, FileTree::MatchFull | FileTree::NoLeaf)) {
                std::shared_ptr<DictionaryValue> list(new DictionaryValue);
                static String const VAR_TYPE       ("type");
                static String const VAR_MODIFIED_AT("modifiedAt");
                static String const VAR_SIZE       ("size");
                static String const VAR_META_ID    ("metaId");
                auto addMeta = [this]
                        (DictionaryValue &list, PathTree::Nodes const &nodes) {
                    for (auto i = nodes.begin(); i != nodes.end(); ++i) {
                        auto const &entry = i.value()->as<FileEntry>();
                        list.add(new TextValue(entry.name()),
                                  RecordValue::takeRecord(
                                      Record::withMembers(
                                          VAR_TYPE, entry.isLeaf()? 0 : 1,
                                          VAR_SIZE, entry.size,
                                          VAR_MODIFIED_AT, entry.modTime,
                                          VAR_META_ID, metaIdForFileEntry(entry)
                                      )));
                    }
                };
                addMeta(*list.get(), dir->children().branches);
                addMeta(*list.get(), dir->children().leaves);
                return list;
            }
            return nullptr;
        },
        [this, id] (std::shared_ptr<DictionaryValue> list) {
            self().metadataReceived(id, list? *list : DictionaryValue());
        });
    }
    void receiveFileContents(QueryId id, QNetworkReply *reply) {
        if (reply->error() == QNetworkReply::NoError) {
            dsize const contentLength = reply->header(QNetworkRequest::ContentLengthHeader).toULongLong();
            QByteArray const data = reply->readAll();
            Query const *query = self().findQuery(id);
            self().chunkReceived(id, query->receivedBytes, data,
                                 contentLength? contentLength : dsize(data.size()));
        }
        else {
            LOG_NET_WARNING(reply->errorString());
        }
    }
};
WebHostedLink::WebHostedLink(String const &address, String const &indexPath)
    : Link(address)
    , d(new Impl(this)) { {
        QNetworkRequest req(QUrl(address / indexPath));
        req.setRawHeader("User-Agent", Version::currentBuild().userAgent().toLatin1());
        QNetworkReply *reply = filesys::RemoteFeedRelay::get().network().get(req);
        QObject::connect(reply, &QNetworkReply::finished, [this, reply] () {
            reply->deleteLater();
            if (reply->error() == QNetworkReply::NoError) {
                parseRepositoryIndex(reply->readAll());
            }
            else {
                handleError(reply->errorString());
                wasDisconnected();
            }
        });
    }
}
void WebHostedLink::setFileTree(FileTree *tree) {
    DENG2_GUARD(d);
    d->fileTree.reset(tree);
}
WebHostedLink::FileTree const &WebHostedLink::fileTree() const {
    return *d->fileTree;
}
WebHostedLink::FileEntry const *WebHostedLink::findFile(Path const &path) const {
    DENG2_GUARD(d);
    return d->fileTree->tryFind(path, PathTree::MatchFull);
}
filesys::PackagePaths WebHostedLink::locatePackages(StringList const &packageIds) const {
    PackagePaths remotePaths;
    foreach (String packageId, packageIds) {
        if (String remotePath = findPackagePath(packageId)) {
            remotePaths.insert(packageId,
                               RepositoryPath(*this, localRoot().path()/packageId, remotePath));
        }
    }
    return remotePaths;
}
void WebHostedLink::transmit(Query const &query) {
    if (query.fileMetadata) {
        d->handleFileListQueryAsync(query);
        return;
    }
    DENG2_ASSERT(query.fileContents);
    String url = address();
    QNetworkRequest req(url.concatenateRelativePath(query.path));
    qDebug() << req.url().toString();
    req.setRawHeader("User-Agent", Version::currentBuild().userAgent().toLatin1());
    QNetworkReply *reply = RemoteFeedRelay::get().network().get(req);
    d->pendingRequests.insert(reply);
    auto const id = query.id;
    QObject::connect(reply, &QNetworkReply::readyRead, [this, id, reply] () {
        d->receiveFileContents(id, reply);
    });
    QObject::connect(reply, &QNetworkReply::finished, [this, id, reply] () {
        d->pendingRequests.remove(reply);
        reply->deleteLater();
    });
}
Block WebHostedLink::FileEntry::metaId(Link const &link) const {
    return md5Hash(link.address(), path(), size, modTime);
}
} // namespace filesys
} // namespace de
