﻿/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ArchiveEntryFile"
#include "de/ArchiveFeed"
#include "de/Archive"
#include "de/Block"
#include "de/Guard"
/** @file archivefeed.cpp Archive Feed.
 *
 * @author Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @author Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ArchiveFeed"
#include "de/ArchiveEntryFile"
#include "de/ArchiveFolder"
#include "de/ByteArrayFile"
#include "de/ZipArchive"
#include "de/Writer"
#include "de/Folder"
#include "de/FS"
#include "de/LogBuffer"
/** @file archivefolder.cpp  Folder whose contents represent an archive.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ArchiveFolder"
#include "de/ArchiveFeed"
/** @file assetobserver.cpp  Utility for observing available assets.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/AssetObserver"
#include "de/App"
#include "de/FileSystem"
#include "de/LinkFile"
#include "de/Loop"
#include <regex>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ByteArrayFile"
/** @file directoryfeed.cpp Directory Feed.
 *
 * @author Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @author Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DirectoryFeed"
#include "de/Folder"
#include "de/NativeFile"
#include "de/FS"
#include "de/Date"
#include "de/App"
#include <QDir>
#include <QFileInfo>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Feed"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/File"
#include "de/App"
#include "de/Date"
#include "de/DirectoryFeed"
#include "de/FS"
#include "de/Feed"
#include "de/Folder"
#include "de/Guard"
#include "de/LinkFile"
#include "de/NativeFile"
#include "de/NativePointerValue"
#include "de/NumberValue"
#include "de/RecordValue"
#include "de/ScriptSystem"
/** @file fileindex.cpp  Index for looking up files of specific type.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FileIndex"
#include "de/ReadWriteLockable"
#include "de/PackageLoader"
#include "de/App"
#include "de/LogBuffer"
/** @file filesystem.cpp File System.
 *
 * @author Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @author Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FS"
#include "de/App"
#include "de/ArchiveFeed"
#include "de/ArchiveFolder"
#include "de/DirectoryFeed"
#include "de/DictionaryValue"
#include "de/Guard"
#include "de/LibraryFile"
#include "de/Log"
#include "de/LogBuffer"
#include "de/Loop"
#include "de/NativePath"
#include "de/ScriptSystem"
#include "de/TextValue"
#include "de/ZipArchive"
#include <QHash>
#include <condition_variable>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Folder"
#include "de/App"
#include "de/Async"
#include "de/DirectoryFeed"
#include "de/FS"
#include "de/Feed"
#include "de/Guard"
#include "de/LogBuffer"
#include "de/NumberValue"
#include "de/ScriptedInfo"
#include "de/ScriptSystem"
#include "de/Task"
#include "de/TaskPool"
#include "de/UnixInfo"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LibraryFile"
#include "de/Library"
#include "de/NativeFile"
#include "de/NativePath"
#include "de/LogBuffer"
#include <QLibrary>
/** @file linkfile.cpp  Symbolic link.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LinkFile"
#include "de/Folder"
/** @file metadatabank.cpp  Cache for file metadata.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/MetadataBank"
#include "de/App"
#include "de/FileSystem"
#include "de/Folder"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NativeFile"
#include "de/DirectoryFeed"
#include "de/Guard"
#include "de/math.h"
/**
 * @file nativepath.cpp
 * File paths for the native file system.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NativePath"
#include "de/App"
#include <QDir>
#include <QFile>
/**
 * @def NATIVE_BASE_SYMBOLIC
 *
 * Symbol used in "pretty" paths to represent the base directory (the one set
 * with -basedir).
 */
#define NATIVE_BASE_SYMBOLIC    "(basedir)"
/**
 * @def NATIVE_HOME_SYMBOLIC
 *
 * Symbol used in "pretty" paths to represent the user's native home directory.
 * Note that this is not the same thing as App::nativeHomePath(), which returns
 * the native location of the Doomsday runtime "/home" path where runtime files
 * are stored.
 */
#ifdef WIN32
#  define NATIVE_HOME_SYMBOLIC  "%HOMEPATH%"
#  define DIR_SEPARATOR         QChar('\\')
#else
#  define NATIVE_HOME_SYMBOLIC  "~"
#  define DIR_SEPARATOR         QChar('/')
#  ifdef UNIX
#    include <sys/types.h>
#    include <pwd.h>
#  endif
#endif
/** @file filesys/node.cpp  File system node.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/Node"
/** @file package.cpp  Package containing metadata, data, and/or files.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Package"
#include "de/App"
#include "de/DotPath"
#include "de/LogBuffer"
#include "de/PackageLoader"
#include "de/Process"
#include "de/Script"
#include "de/ScriptSystem"
#include "de/ScriptedInfo"
#include "de/TextValue"
#include "de/TimeValue"
#include <QRegularExpression>
/** @file packagefeed.cpp  Links to loaded packages.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PackageFeed"
#include "de/LinkFile"
#include "de/ArchiveFolder"
#include "de/PackageLoader"
#include "de/FS"
/** @file packageloader.cpp  Loads and unloads packages.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PackageLoader"
#include "de/App"
#include "de/CommandLine"
#include "de/Config"
#include "de/DictionaryValue"
#include "de/FS"
#include "de/Info"
#include "de/LinkFile"
#include "de/LogBuffer"
#include "de/PackageFeed"
#include "de/Parser"
#include "de/TextValue"
#include "de/Version"
#include <QMap>
#include <QRegularExpression>
/** @file staticlibraryfeed.cpp
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/filesys/staticlibraryfeed.h"
#include "de/Folder"
#include "de/Library"
#include "de/LibraryFile"
#include "de/String"
using namespace de;
namespace de {
DENG2_PIMPL_NOREF(ArchiveEntryFile) {
    Archive *archive;
    Path entryPath;
    Block const *readBlock = nullptr;
    Block const &entryData() {
        if (!readBlock) {
#if 0
            {
                static Lockable dbg;
                DENG2_GUARD(dbg);
                qDebug() << "--------\nAEF being read" << entryPath;
                DENG2_PRINT_BACKTRACE();
            }
#endif
            readBlock = &const_cast<Archive const *>(archive)->entryBlock(entryPath);
        }
        return *readBlock;
    }
};
ArchiveEntryFile::ArchiveEntryFile(String const &name, Archive &archive, String const &entryPath)
    : ByteArrayFile(name)
    , d(new Impl) {
    d->archive   = &archive;
    d->entryPath = entryPath;
}
ArchiveEntryFile::~ArchiveEntryFile() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
String ArchiveEntryFile::entryPath() const {
    return d->entryPath;
}
String ArchiveEntryFile::describe() const {
    DENG2_GUARD(this);
    return String("archive entry \"%1\"").arg(d->entryPath);
}
void ArchiveEntryFile::clear() {
    DENG2_GUARD(this);
    verifyWriteAccess();
    File::clear();
    archive().entryBlock(d->entryPath).clear();
    Status st = status();
    st.size = 0;
    st.modifiedAt = Time();
    setStatus(st);
}
void ArchiveEntryFile::flush() {
    ByteArrayFile::flush();
    if (ArchiveFeed *feed = maybeAs<ArchiveFeed>(originFeed())) {
        feed->rewriteFile();
    }
}
Block ArchiveEntryFile::metaId() const {
    Block data = File::metaId() + d->entryPath.toUtf8();
    if (File const *sourceFile = maybeAs<File const>(archive().source())) {
        data += sourceFile->metaId();
    }
    return data.md5Hash();
}
Archive &ArchiveEntryFile::archive() {
    return *d->archive;
}
Archive const &ArchiveEntryFile::archive() const {
    return *d->archive;
}
void ArchiveEntryFile::uncache() const {
    DENG2_GUARD(this);
    if (d->readBlock) {
        archive().uncacheBlock(d->entryPath);
        d->readBlock = nullptr;
    }
}
IByteArray::Size ArchiveEntryFile::size() const {
    DENG2_GUARD(this);
    return archive().entryStatus(d->entryPath).size;
}
void ArchiveEntryFile::get(Offset at, Byte *values, Size count) const {
    DENG2_GUARD(this);
    d->entryData().get(at, values, count);
}
void ArchiveEntryFile::set(Offset at, Byte const *values, Size count) {
    DENG2_GUARD(this);
    verifyWriteAccess();
    Block &entryBlock = archive().entryBlock(d->entryPath);
    entryBlock.set(at, values, count);
    Status st = status();
    st.size = entryBlock.size();
    st.modifiedAt = archive().entryStatus(d->entryPath).modifiedAt;
    setStatus(st);
}
} // namespace de
namespace de {
DENG2_PIMPL(ArchiveFeed)
, DENG2_OBSERVES(File, Deletion) {
    File *file;
    Block serializedArchive;
    Archive *arch;
    String basePath;
    ArchiveFeed *parentFeed;
    bool allowWrite;
    typedef LockableT<PointerSetT<ArchiveEntryFile>> EntryTable;
    EntryTable entries;
    Impl(Public *feed, File &f)
        : Base(feed)
        , file(&f)
        , arch(0)
        , parentFeed(0)
        , allowWrite(f.mode().testFlag(File::Write)) // write access depends on file
    {
        if (auto *bytes = maybeAs<IByteArray>(f)) {
            LOG_RES_XVERBOSE("Source %s is a byte array", f.description());
            arch = new ZipArchive(*bytes, f.metaId());
        }
        else {
            LOG_RES_XVERBOSE("Source %s is a stream", f.description());
            f >> serializedArchive;
            arch = new ZipArchive(serializedArchive);
        }
        file->audienceForDeletion() += this;
    }
    Impl(Public *feed, ArchiveFeed &parentFeed, String const &path)
        : Base(feed)
        , file(parentFeed.d->file)
        , arch(0)
        , basePath(path)
        , parentFeed(&parentFeed)
        , allowWrite(isWriteAllowed()) {
        file->audienceForDeletion() += this;
    }
    ~Impl() {
        if (arch) {
            writeIfModified();
            delete arch;
        }
    }
    bool isWriteAllowed() const {
        if (parentFeed) {
            return parentFeed->d->isWriteAllowed();
        }
        return allowWrite;
    }
    void writeIfModified() {
        if (!file || !arch || !file->mode().testFlag(File::Write)) {
            return;
        }
        if (arch->modified()) {
            LOG_RES_VERBOSE("Updating archive in ") << file->description();
            arch->cache();
            file->clear();
            Writer(*file) << *arch;
            file->flush();
        }
        else {
            LOG_RES_VERBOSE("Not updating archive in %s (not changed)") << file->description();
        }
    }
    void fileBeingDeleted(File const &deleted) {
        if (file == &deleted) {
            writeIfModified();
            file = 0;
        }
        else {
            auto &table = entryTable();
            DENG2_GUARD(table);
            table.value.remove(&deleted.as<ArchiveEntryFile>());
        }
    }
    Archive &archive() {
        if (parentFeed) {
            return parentFeed->archive();
        }
        return *arch;
    }
    EntryTable &entryTable() {
        return parentFeed? parentFeed->d->entries : entries;
    }
    void addToEntryTable(ArchiveEntryFile *file) {
        auto &table = entryTable();
        DENG2_GUARD(table);
        table.value.insert(file);
        file->audienceForDeletion() += this;
    }
    PopulatedFiles populate(Folder const &folder) {
        PopulatedFiles populated;
        Archive::Names names;
        archive().listFiles(names, basePath);
        DENG2_FOR_EACH(Archive::Names, i, names) {
            if (folder.has(*i)) {
                continue;
            }
            String entry = basePath / *i;
            std::unique_ptr<ArchiveEntryFile> archFile(new ArchiveEntryFile(*i, archive(), entry));
            addToEntryTable(archFile.get());
            if (allowWrite) archFile->setMode(File::Write);
            archFile->setStatus(archive().entryStatus(entry));
            File *f = folder.fileSystem().interpret(archFile.release());
            f->setOriginFeed(thisPublic);
            populated << f;
        }
        archive().listFolders(names, basePath);
        for (Archive::Names::iterator i = names.begin(); i != names.end(); ++i) {
            folder.fileSystem().makeFolder(folder.path() / *i, FS::InheritPrimaryFeed);
        }
        return populated;
    }
};
ArchiveFeed::ArchiveFeed(File &archiveFile)
    : d(new Impl(this, archiveFile))
{}
ArchiveFeed::ArchiveFeed(ArchiveFeed &parentFeed, String const &basePath)
    : d(new Impl(this, parentFeed, basePath))
{}
ArchiveFeed::~ArchiveFeed() {
    LOG_AS("~ArchiveFeed");
    d.reset();
}
String ArchiveFeed::description() const {
    return "archive in " + (d->file? d->file->description() : "(deleted file)");
}
Feed::PopulatedFiles ArchiveFeed::populate(Folder const &folder) {
    LOG_AS("ArchiveFeed::populate");
    return d->populate(folder);
}
bool ArchiveFeed::prune(File &file) const {
    LOG_AS("ArchiveFeed::prune");
    ArchiveEntryFile *entryFile = maybeAs<ArchiveEntryFile>(file);
    if (entryFile && &entryFile->archive() == &archive()) {
        if (!archive().hasEntry(entryFile->entryPath())) {
            LOG_RES_VERBOSE("%s removed from archive") << file.description();
            return true; // It's gone...
        }
        if (archive().entryStatus(entryFile->entryPath()).modifiedAt !=
           file.status().modifiedAt) {
            LOG_RES_XVERBOSE("%s has been modified (arch:%s != file:%s)",
                             file.description()
                             << archive().entryStatus(entryFile->entryPath()).modifiedAt.asText()
                             << file.status().modifiedAt.asText());
            return true;
        }
    }
    return false;
}
File *ArchiveFeed::createFile(String const &name) {
    String newEntry = d->basePath / name;
    if (archive().hasEntry(newEntry)) {
        throw AlreadyExistsError("ArchiveFeed::createFile", name + ": already exists");
    }
    archive().add(newEntry, Block());
    auto *file = new ArchiveEntryFile(name, archive(), newEntry);
    d->addToEntryTable(file);
    file->setOriginFeed(this);
    return file;
}
void ArchiveFeed::destroyFile(String const &name) {
    archive().remove(d->basePath / name);
}
Feed *ArchiveFeed::newSubFeed(String const &name) {
    return new ArchiveFeed(*this, d->basePath / name);
}
Archive &ArchiveFeed::archive() {
    return d->archive();
}
Archive const &ArchiveFeed::archive() const {
    return d->archive();
}
String const &ArchiveFeed::basePath() const {
    return d->basePath;
}
File const &ArchiveFeed::archiveSourceFile() const {
    if (d->file) {
        return *d->file;
    }
    throw InvalidSourceError("ArchiveFeed::archiveSourceFile", "Archive source file is missing");
}
void ArchiveFeed::rewriteFile() {
    if (d->parentFeed) {
        DENG2_ASSERT(!d->arch);
        d->parentFeed->rewriteFile();
    }
    else {
        DENG2_ASSERT(d->arch != 0);
        d->writeIfModified();
    }
}
void ArchiveFeed::uncache() {
    auto &table = d->entryTable();
    DENG2_GUARD(table);
    for (auto *entry : table.value) {
        entry->uncache();
    }
}
void ArchiveFeed::uncacheAllEntries(StringList folderTypes) // static
{
    if (Folder::isPopulatingAsync()) return; // Never mind.
    for (String const &folderType : folderTypes) {
        foreach (File *file, FileSystem::get().indexFor(folderType).files()) {
            if (auto *feed = file->as<Folder>().primaryFeedMaybeAs<ArchiveFeed>()) {
                feed->uncache();
            }
        }
    }
}
} // namespace de
namespace de {
ArchiveFolder::ArchiveFolder(File &sourceArchiveFile, String const &name) : Folder(name) {
    attach(new ArchiveFeed(sourceArchiveFile));
}
ArchiveFolder::~ArchiveFolder() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
void ArchiveFolder::flush() {
    Folder::flush();
    primaryFeed()->as<ArchiveFeed>().rewriteFile();
}
String ArchiveFolder::describe() const {
    DENG2_GUARD(this);
    String desc = String("archive \"%1\"").arg(name());
    String const feedDesc = describeFeeds();
    if (!feedDesc.isEmpty()) {
        desc += String(" (%1)").arg(feedDesc);
    }
    return desc;
}
Archive &ArchiveFolder::archive() {
    DENG2_ASSERT(!feeds().empty());
    return primaryFeed()->as<ArchiveFeed>().archive();
}
Archive const &ArchiveFolder::archive() const {
    return const_cast<ArchiveFolder *>(this)->archive();
}
} // namespace de
namespace de {
namespace filesys {
static const std::string PREFIX = "asset";
DENG2_PIMPL(AssetObserver)
, DENG2_OBSERVES(FileIndex, Addition)
, DENG2_OBSERVES(FileIndex, Removal) {
    const std::regex pattern;
    LoopCallback mainCall;
    static FileIndex const &linkIndex() {
        return App::fileSystem().indexFor(DENG2_TYPE_NAME(LinkFile));
    }
    static String assetIdentifier(File const &link) {
        DENG2_ASSERT(link.name().beginsWith(String::fromStdString(PREFIX + ".")));
        return link.name().mid(6);
    }
    Impl(Public *i, String const &regex)
        : Base(i)
        , pattern(PREFIX + "\\." + regex.toStdString(), std::regex::icase) {
        linkIndex().audienceForAddition() += this;
        linkIndex().audienceForRemoval()  += this;
    }
    void fileAdded(File const &link, FileIndex const &) {
        if (!std::regex_match(link.name().toStdString(), pattern)) return;
        String const ident = assetIdentifier(link);
        mainCall.enqueue([this, ident] () {
            DENG2_FOR_PUBLIC_AUDIENCE2(Availability, i) {
                i->assetAvailabilityChanged(ident, Added);
            }
        });
    }
    void fileRemoved(File const &link, FileIndex const &) {
        if (!std::regex_match(link.name().toStdString(), pattern)) return;
        DENG2_FOR_PUBLIC_AUDIENCE2(Availability, i) {
            i->assetAvailabilityChanged(assetIdentifier(link), Removed);
        }
    }
    DENG2_PIMPL_AUDIENCE(Availability)
};
DENG2_AUDIENCE_METHOD(AssetObserver, Availability)
AssetObserver::AssetObserver(String const &regexPattern)
    : d(new Impl(this, regexPattern))
{}
} // namespace filesys
} // namespace de
namespace de {
dsize ByteArrayFile::size() const {
    return File::size();
}
IOStream &ByteArrayFile::operator << (IByteArray const &bytes) {
    Block block(bytes);
    set(File::size(), block.data(), block.size());
    return *this;
}
IIStream &ByteArrayFile::operator >> (IByteArray &) {
    throw InputError("ByteArrayFile::operator >>", "ByteArrayFile is an immutable stream");
}
IIStream const &ByteArrayFile::operator >> (IByteArray &bytes) const {
    Block block(File::size());
    get(0, block.data(), block.size());
    bytes.set(0, block.data(), block.size());
    return *this;
}
} // namespace de
namespace de {
static String const fileStatusSuffix = ".doomsday_file_status";
DENG2_PIMPL_NOREF(DirectoryFeed) {
    NativePath nativePath;
    Flags mode;
    String namePattern;
};
DirectoryFeed::DirectoryFeed(NativePath const &nativePath, Flags const &mode)
    : d(new Impl) {
    d->nativePath = nativePath;
    d->mode = mode;
}
void DirectoryFeed::setNamePattern(const String &namePattern) {
    d->namePattern = namePattern;
}
String DirectoryFeed::description() const {
    String desc;
    if (d->namePattern) {
        desc = "files matching \"" + d->namePattern + "\" in ";
    }
    desc += "directory \"" + d->nativePath.pretty() + "\"";
    return desc;
}
const NativePath &DirectoryFeed::nativePath() const {
    return d->nativePath;
}
Feed::PopulatedFiles DirectoryFeed::populate(Folder const &folder) {
    if (d->mode & AllowWrite) {
        const_cast<Folder &>(folder).setMode(File::Write);
    }
    if (d->mode.testFlag(CreateIfMissing) && !NativePath::exists(d->nativePath)) {
        NativePath::createPath(d->nativePath);
    }
    QDir dir(d->nativePath);
    if (!dir.isReadable()) {
        throw NotFoundError("DirectoryFeed::populate", "Path '" + d->nativePath + "' inaccessible");
    }
    QStringList nameFilters;
    if (d->namePattern) {
        nameFilters << d->namePattern;
    }
    else {
        nameFilters << "*";
    }
    QDir::Filters dirFlags = QDir::Files | QDir::NoDotAndDotDot;
    if (d->mode.testFlag(PopulateNativeSubfolders)) {
        dirFlags |= QDir::Dirs;
    }
    PopulatedFiles populated;
    foreach (QFileInfo entry, dir.entryInfoList(nameFilters, dirFlags)) {
        if (entry.isDir()) {
            populateSubFolder(folder, entry.fileName());
        }
        else {
            if (!entry.fileName().endsWith(fileStatusSuffix)) // ignore meta files
            {
                populateFile(folder, entry.fileName(), populated);
            }
        }
    }
    return populated;
}
void DirectoryFeed::populateSubFolder(Folder const &folder, String const &entryName) {
    LOG_AS("DirectoryFeed::populateSubFolder");
    if (entryName != "." && entryName != "..") {
        Folder *subFolder = nullptr;
        if (!folder.has(entryName)) {
            subFolder = &folder.fileSystem()
                    .makeFolderWithFeed(folder.path() / entryName,
                                        newSubFeed(entryName),
                                        Folder::PopulateFullTree,
                                        FS::DontInheritFeeds);
        }
        else {
            subFolder = &folder.locate<Folder>(entryName);
        }
        if (d->mode & AllowWrite) {
            subFolder->setMode(File::Write);
        }
        else {
            subFolder->setMode(File::ReadOnly);
        }
    }
}
void DirectoryFeed::populateFile(Folder const &folder, String const &entryName,
                                 PopulatedFiles &populated) {
    try {
        if (folder.has(entryName)) {
            return;
        }
        NativePath const entryPath = d->nativePath / entryName;
        std::unique_ptr<NativeFile> nativeFile(new NativeFile(entryName, entryPath));
        nativeFile->setStatus(fileStatus(entryPath));
        if (d->mode & AllowWrite) {
            nativeFile->setMode(File::Write);
        }
        File *file = folder.fileSystem().interpret(nativeFile.release());
        file->setOriginFeed(this);
        populated << file;
    }
    catch (StatusError const &er) {
        LOG_WARNING("Error with \"%s\" in %s: %s")
                << entryName
                << folder.description()
                << er.asText();
    }
}
bool DirectoryFeed::prune(File &file) const {
    LOG_AS("DirectoryFeed::prune");
    if (NativeFile *nativeFile = maybeAs<NativeFile>(file.source())) {
        try {
            if (fileStatus(nativeFile->nativePath()) != nativeFile->status()) {
                LOG_RES_MSG("Pruning \"%s\": status has changed") << nativeFile->nativePath();
                return true;
            }
        }
        catch (StatusError const &) {
            return true;
        }
    }
    if (Folder *subFolder = maybeAs<Folder>(file)) {
        if (subFolder->feeds().size() == 1) {
            DirectoryFeed *dirFeed = maybeAs<DirectoryFeed>(subFolder->feeds().front());
            if (dirFeed && !dirFeed->d->nativePath.exists()) {
                LOG_RES_NOTE("Pruning %s: no longer exists") << dirFeed->description(); //d->nativePath;
                return true;
            }
        }
    }
    return false;
}
File *DirectoryFeed::createFile(String const &name) {
    NativePath newPath = d->nativePath / name;
    /*if (NativePath::exists(newPath))
    {
    }*/
    File *file = new NativeFile(name, newPath);
    file->setOriginFeed(this);
    return file;
}
void DirectoryFeed::destroyFile(String const &name) {
    NativePath path = d->nativePath / name;
    if (!path.exists()) {
        return;
    }
    if (!path.destroy()) {
        throw RemoveError("DirectoryFeed::destroyFile", "Cannot remove \"" + name +
                          "\" in " + description());
    }
}
Feed *DirectoryFeed::newSubFeed(String const &name) {
    NativePath subPath = d->nativePath / name;
    if (d->mode.testFlag(CreateIfMissing) || (subPath.exists() && subPath.isReadable())) {
        return new DirectoryFeed(subPath, d->mode);
    }
    return nullptr;
}
void DirectoryFeed::changeWorkingDir(NativePath const &nativePath) {
    if (!App::setCurrentWorkPath(nativePath)) {
        throw WorkingDirError("DirectoryFeed::changeWorkingDir",
                              "Failed to change to " + nativePath);
    }
}
File::Status DirectoryFeed::fileStatus(NativePath const &nativePath) {
    QFileInfo info(nativePath);
    if (!info.exists()) {
        throw StatusError("DirectoryFeed::fileStatus", nativePath + " inaccessible");
    }
    File::Status st { info.isDir()? File::Type::Folder : File::Type::File,
                      dsize(info.size()),
                      info.lastModified() };
    String const overrideName = nativePath + fileStatusSuffix;
    if (QFileInfo().exists(overrideName)) {
        QFile f(overrideName);
        if (f.open(QFile::ReadOnly)) {
            st.modifiedAt = Time::fromText(String::fromUtf8(f.readAll()), Time::ISOFormat);
        }
    }
    return st;
}
void DirectoryFeed::setFileModifiedTime(NativePath const &nativePath, Time const &modifiedAt) {
    String const overrideName = nativePath + fileStatusSuffix;
    if (!modifiedAt.isValid()) {
        QFile::remove(overrideName);
        return;
    }
    QFile f(overrideName);
    if (f.open(QFile::WriteOnly | QFile::Truncate)) {
        f.write(modifiedAt.asText(Time::ISOFormat).toUtf8());
    }
}
File &DirectoryFeed::manuallyPopulateSingleFile(NativePath const &nativePath,
                                                Folder &parentFolder) // static
{
    const bool isExisting = nativePath.exists();
    Folder *   parent     = &parentFolder;
    File::Status status;
    if (isExisting) {
        status = fileStatus(nativePath);
    }
    else {
        status.modifiedAt = Time(); // file being created now
    }
    if (parentFolder.extension() != ".pack" &&
        nativePath.fileName().fileNameExtension() == ".pack") {
        int const last = nativePath.segmentCount() - 1;
        Rangei packRange(last, last);
        while (packRange.start > 0 &&
               nativePath.segment(packRange.start - 1).toStringRef()
               .endsWith(".pack", Qt::CaseInsensitive)) {
            packRange.start--;
        }
        if (!packRange.isEmpty()) {
            parent = &FS::get().makeFolder(parentFolder.path() /
                                           nativePath.subPath(packRange).withSeparators('/'),
                                           FS::DontInheritFeeds);
        }
    }
    const String newFilePath = parent->path() / nativePath.fileName();
    if (status.type() == File::Type::File) {
        parent->clear();
        parent->clearFeeds();
        auto *feed = new DirectoryFeed(nativePath.fileNamePath());
        feed->setNamePattern(nativePath.fileName());
        parent->attach(feed);
        if (isExisting) {
            parent->populate();
        }
        else {
            parent->replaceFile(nativePath.fileName());
        }
        return FS::locate<File>(newFilePath);
    }
    else {
        return FS::get().makeFolderWithFeed(newFilePath,
                                            new DirectoryFeed(nativePath),
                                            Folder::PopulateFullTree,
                                            FS::DontInheritFeeds | FS::PopulateNewFolder);
    }
}
} // namespace de
namespace de {
Feed::Feed()
{}
Feed::~Feed()
{}
File *Feed::createFile(String const &/*name*/)
{
    return 0;
}
void Feed::destroyFile(String const &/*name*/)
{}
Feed *Feed::newSubFeed(String const &/*name*/)
{
    return 0;
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(File) {
    File *source;
    Feed *originFeed;
    Status status;
    Flags mode;
    Record info;
    Impl() : source(0), originFeed(0) {}
    DENG2_PIMPL_AUDIENCE(Deletion)
};
DENG2_AUDIENCE_METHOD(File, Deletion)
File::File(String const &fileName) : Node(fileName), d(new Impl) {
    d->source = this;
    d->info.add(Record::VAR_NATIVE_SELF).set(new NativePointerValue(this)).setReadOnly();
    d->info.addSuperRecord(ScriptSystem::builtInClass(QStringLiteral("File")));
}
File::~File() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    flush();
    if (d->source != this) {
        delete d->source;
        d->source = 0;
    }
    if (Folder *parentFolder = parent()) {
        parentFolder->remove(this);
    }
    deindex();
}
void File::deindex() {
    fileSystem().deindex(*this);
}
void File::flush()
{}
void File::clear() {
    verifyWriteAccess();
}
FileSystem &File::fileSystem() {
    return DENG2_APP->fileSystem();
}
Folder *File::parent() const {
    return maybeAs<Folder>(Node::parent());
}
String File::description(int verbosity) const {
    DENG2_GUARD(this);
    String desc = describe();
    if (verbosity < 0) {
        Log &log = Log::threadLog();
        if (log.isStaging()) {
            verbosity = 0;
            if (log.currentEntryMetadata() & LogEntry::Dev) {
                verbosity = 2;
            }
            else if ((log.currentEntryMetadata() & LogEntry::LevelMask) <= LogEntry::Verbose) {
                verbosity = 1;
            }
        }
    }
    if (verbosity > 1) {
        if (!mode().testFlag(Write)) {
            desc = "read-only " + desc;
        }
        if (parent()) {
            desc += " (path \"" + path() + "\")";
        }
    }
    if (originFeed() && (verbosity >= 2 || !is<DirectoryFeed>(originFeed()))) {
        String const feedDesc = originFeed()->description();
        if (!desc.contains(feedDesc)) // A bit of a kludge; don't repeat the feed description.
        {
            desc += " from " + feedDesc;
        }
    }
#ifdef DENG2_DEBUG
    if (source() != this) {
        desc += _E(s)_E(D) " {\"" + path() + "\": data sourced from " +
                source()->description(verbosity) + "}" _E(.)_E(.);
    }
#endif
    return desc;
}
String File::describe() const {
    return "abstract File";
}
void File::setOriginFeed(Feed *feed) {
    DENG2_GUARD(this);
    d->originFeed = feed;
}
Feed *File::originFeed() const {
    DENG2_GUARD(this);
    return d->originFeed;
}
void File::setSource(File *source) {
    DENG2_GUARD(this);
    if (d->source != this) {
        delete d->source;
    }
    d->source = source;
}
File const *File::source() const {
    DENG2_GUARD(this);
    if (&target() != this) {
        return target().source();
    }
    if (d->source != this) {
        return d->source->source();
    }
    return d->source;
}
File *File::source() {
    DENG2_GUARD(this);
    if (&target() != this) {
        return target().source();
    }
    if (d->source != this) {
        return d->source->source();
    }
    return d->source;
}
File &File::target() {
    return *this;
}
File const &File::target() const {
    return *this;
}
void File::setStatus(Status const &status) {
    DENG2_GUARD(this);
    if (this != d->source) {
        d->source->setStatus(status);
    }
    else {
        d->status = status;
    }
}
File::Status const &File::status() const {
    DENG2_GUARD(this);
    if (this != d->source) {
        return d->source->status();
    }
    return d->status;
}
void File::setMode(Flags const &newMode) {
    DENG2_GUARD(this);
    if (d->mode.testFlag(Write) && !newMode.testFlag(Write)) {
        flush();
    }
    if (this != d->source) {
        d->source->setMode(newMode);
    }
    else {
        d->mode = newMode;
    }
}
void File::setMode(Flags flags, FlagOpArg op) {
    applyFlagOperation(d->mode, flags, op);
}
Record &File::objectNamespace() {
    return d->info;
}
Record const &File::objectNamespace() const {
    return d->info;
}
File::Flags const &File::mode() const {
    DENG2_GUARD(this);
    if (this != d->source) {
        return d->source->mode();
    }
    return d->mode;
}
void File::verifyWriteAccess() {
    if (!mode().testFlag(Write)) {
        throw ReadOnlyError("File::verifyWriteAccess", path() + " is in read-only mode");
    }
}
File *File::reinterpret() {
    Folder *folder  = parent();
    File *original  = source();
    bool deleteThis = false;
    if (original != this) {
        DENG2_ASSERT(!original->parent());
        d->source = 0; // source is owned, so take it away
        deleteThis = true;
    }
    if (folder) {
        folder->remove(*this);
        deindex();
    }
    original->flush();
    File *result = fileSystem().interpret(original);
    result->setOriginFeed(d->originFeed);
    if (deleteThis) {
        DENG2_ASSERT(result != this);
        delete this;
    }
    if (folder) {
        folder->add(result);
        fileSystem().index(*result);
    }
    return result;
}
NativePath File::correspondingNativePath() const {
    if (NativeFile const *native = maybeAs<NativeFile>(source())) {
        return native->nativePath();
    }
    else if (Folder const *folder = maybeAs<Folder>(target())) {
        if (auto *feed = folder->primaryFeedMaybeAs<DirectoryFeed>()) {
            return feed->nativePath();
        }
    }
    return NativePath();
}
IOStream &File::operator << (IByteArray const &bytes) {
    DENG2_UNUSED(bytes);
    throw OutputError("File::operator <<", description() + " does not accept a byte stream");
}
IIStream &File::operator >> (IByteArray &bytes) {
    DENG2_UNUSED(bytes);
    throw InputError("File::operator >>", description() + " does not produce a byte stream");
}
IIStream const &File::operator >> (IByteArray &bytes) const {
    DENG2_UNUSED(bytes);
    throw InputError("File::operator >>", description() + " does not offer an immutable byte stream");
}
static bool sortByNameAsc(File const *a, File const *b) {
    return a->name().compareWithoutCase(b->name()) < 0;
}
String File::fileListAsText(QList<File const *> files) {
    qSort(files.begin(), files.end(), sortByNameAsc);
    String txt;
    foreach (File const *f, files) {
        if (!txt.isEmpty()) txt += "\n";
        String flags = QString("%1%2%3%4%5")
                .arg(is<Folder>(f)?                'd' : '-')
                .arg(f->mode().testFlag(Write)?    'w' : 'r')
                .arg(f->mode().testFlag(Truncate)? 't' : '-')
                .arg(f->source() != f?             'i' : '-')
                .arg(f->originFeed()?              'f' : '-');
        txt += flags + QString("%1 %2 %3")
                .arg(f->size(), 9)
                .arg(f->status().modifiedAt.asText(), 23)
                .arg(f->name());
        if (LinkFile const *link = maybeAs<LinkFile>(f)) {
            if (!link->isBroken()) {
                txt += QString(" -> %1").arg(link->target().path());
            }
            else {
                txt += " (broken link)";
            }
        }
    }
    return txt;
}
dsize File::size() const {
    return status().size;
}
Block File::metaId() const {
    auto const &st = target().status();
    return md5Hash(path(), duint64(st.size), st.modifiedAt);
}
} // namespace de
namespace de {
DENG2_PIMPL(FileIndex), public ReadWriteLockable {
    IPredicate const *predicate;
    Index index;
    Impl(Public *i)
        : Base(i)
        , predicate(0) {
        audienceForAddition.setAdditionAllowedDuringIteration(true);
        audienceForRemoval .setAdditionAllowedDuringIteration(true);
    }
    static String indexedName(File const &file) {
        String name = file.name().lower();
        if (name.endsWith(".pack")) {
            name = Package::split(name.fileNameWithoutExtension()).first + ".pack";
        }
        return name;
    }
    void add(File const &file) {
        DENG2_GUARD_WRITE(this);
        index.insert(std::pair<String, File *>(indexedName(file), const_cast<File *>(&file)));
    }
    void remove(File const &file) {
        DENG2_GUARD_WRITE(this);
        if (index.empty()) {
            return;
        }
        IndexRange range = index.equal_range(indexedName(file));
        for (Index::iterator i = range.first; i != range.second; ++i) {
            if (i->second == &file) {
                index.erase(i);
                break;
            }
        }
    }
    void findPartialPath(String const &path, FoundFiles &found) const {
        String baseName = path.fileName().lower();
        String dir      = path.fileNamePath().lower();
        if (!dir.empty() && !dir.beginsWith("/")) {
            dir = "/" + dir;
        }
        DENG2_GUARD_READ(this);
        ConstIndexRange range = index.equal_range(baseName);
        for (Index::const_iterator i = range.first; i != range.second; ++i) {
            File *file = i->second;
            if (file->path().fileNamePath().endsWith(dir, String::CaseInsensitive)) {
                found.push_back(file);
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(Addition)
    DENG2_PIMPL_AUDIENCE(Removal)
};
DENG2_AUDIENCE_METHOD(FileIndex, Addition)
DENG2_AUDIENCE_METHOD(FileIndex, Removal)
FileIndex::FileIndex() : d(new Impl(this))
{}
void FileIndex::setPredicate(IPredicate const &predicate) {
    d->predicate = &predicate;
}
bool FileIndex::maybeAdd(File const &file) {
    if (d->predicate && !d->predicate->shouldIncludeInIndex(file)) {
        return false;
    }
    d->add(file);
    DENG2_FOR_AUDIENCE2(Addition, i) {
        i->fileAdded(file, *this);
    }
    return true;
}
void FileIndex::remove(File const &file) {
    d->remove(file);
    DENG2_FOR_AUDIENCE2(Removal, i) {
        i->fileRemoved(file, *this);
    }
}
int FileIndex::size() const {
    DENG2_GUARD_READ(d);
    return int(d->index.size());
}
static bool fileNotInAnyLoadedPackage(File *file) {
    String const identifier = Package::identifierForContainerOfFile(*file);
    return !App::packageLoader().isLoaded(identifier);
}
void FileIndex::findPartialPath(String const &path, FoundFiles &found, Behavior behavior) const {
    d->findPartialPath(path, found);
    if (behavior == FindOnlyInLoadedPackages) {
        found.remove_if(fileNotInAnyLoadedPackage);
    }
}
void FileIndex::findPartialPath(Folder const &rootFolder, String const &path,
                                FoundFiles &found, Behavior behavior) const {
    findPartialPath(path, found, behavior);
    found.remove_if([&rootFolder] (File *file) {
        return !file->hasAncestor(rootFolder);
    });
}
void FileIndex::findPartialPath(String const &packageId, String const &path,
                                FoundFiles &found) const {
    Package const &pkg = App::packageLoader().package(packageId);
    if (is<Folder>(pkg.file())) {
        findPartialPath(pkg.root(), path, found, FindInEntireIndex);
        found.remove_if([&packageId](File *file) {
            return Package::identifierForContainerOfFile(*file) != packageId;
        });
    }
}
int FileIndex::findPartialPathInPackageOrder(String const &path, FoundFiles &found, Behavior behavior) const {
    findPartialPath(path, found, behavior);
    App::packageLoader().sortInPackageOrder(found);
    return int(found.size());
}
FileIndex::Index::const_iterator FileIndex::begin() const {
    return d.getConst()->index.begin();
}
FileIndex::Index::const_iterator FileIndex::end() const {
    return d.getConst()->index.end();
}
void FileIndex::print() const {
    DENG2_GUARD_READ(d);
    for (auto i = begin(); i != end(); ++i) {
        LOG_TRACE("\"%s\": ", i->first << i->second->description());
    }
}
QList<File *> FileIndex::files() const {
    DENG2_GUARD_READ(d);
    QList<File *> list;
    for (auto i = begin(); i != end(); ++i) {
        list.append(i->second);
    }
    return list;
}
} // namespace de
namespace de {
static FileIndex const emptyIndex; // never contains any files
DENG2_PIMPL_NOREF(FileSystem) {
    std::mutex              busyMutex;
    int                     busyLevel = 0;
    std::condition_variable busyFinished;
    Record fsModule;
    QList<filesys::IInterpreter const *> interpreters;
    FileIndex index;
    typedef QHash<String, FileIndex *> TypeIndex; // owned
    LockableT<TypeIndex> typeIndex;
    QSet<FileIndex *> userIndices; // not owned
    std::unique_ptr<Folder> root;
    Impl() {
        root.reset(new Folder);
        ScriptSystem::get().addNativeModule("FS", fsModule);
    }
    ~Impl() {
        root.reset();
        DENG2_GUARD(typeIndex);
        qDeleteAll(typeIndex.value);
        typeIndex.value.clear();
    }
    FileIndex &getTypeIndex(String const &typeName) {
        DENG2_GUARD(typeIndex);
        FileIndex *&idx = typeIndex.value[typeName];
        if (!idx) {
            idx = new FileIndex;
        }
        return *idx;
    }
    DENG2_PIMPL_AUDIENCE(Busy)
};
DENG2_AUDIENCE_METHOD(FileSystem, Busy)
FileSystem::FileSystem() : d(new Impl)
{}
void FileSystem::addInterpreter(filesys::IInterpreter const &interpreter) {
    d->interpreters.prepend(&interpreter);
}
void FileSystem::refreshAsync() {
    Folder::afterPopulation([this] () {
        LOG_AS("FS::refresh");
        d->root->populate(Folder::PopulateAsyncFullTree);
    });
}
Folder &FileSystem::makeFolder(String const &path, FolderCreationBehaviors behavior) {
    LOG_AS("FS::makeFolder");
    Folder *subFolder = d->root->tryLocate<Folder>(path);
    if (!subFolder) {
        Folder &parentFolder = makeFolder(path.fileNamePath(), behavior & ~PopulateNewFolder);
        if (Folder *folder = parentFolder.tryLocate<Folder>(path.fileName())) {
            return *folder;
        }
        subFolder = &interpret(new Folder(path.fileName()))->as<Folder>();
        if (parentFolder.mode() & File::Write) {
            subFolder->setMode(File::Write);
        }
        if (behavior & (InheritPrimaryFeed | InheritAllFeeds)) {
            DENG2_GUARD(parentFolder);
            foreach (Feed *parentFeed, parentFolder.feeds()) {
                Feed *feed = parentFeed->newSubFeed(subFolder->name());
                if (!feed) continue; // Check next one instead.
                LOGDEV_RES_XVERBOSE_DEBUGONLY("Creating subfeed \"%s\" from %s",
                                             subFolder->name() << parentFeed->description());
                subFolder->attach(feed);
                if (!behavior.testFlag(InheritAllFeeds)) break;
            }
        }
        parentFolder.add(subFolder);
        index(*subFolder);
        if (behavior.testFlag(PopulateNewFolder)) {
            subFolder->populate();
        }
    }
    return *subFolder;
}
Folder &FileSystem::makeFolderWithFeed(String const &path, Feed *feed,
                                       Folder::PopulationBehavior populationBehavior,
                                       FolderCreationBehaviors behavior) {
    makeFolder(path.fileNamePath(), behavior);
    Folder &folder = makeFolder(path, DontInheritFeeds /* we have a specific feed to attach */);
    folder.clear();
    folder.clearFeeds();
    folder.attach(feed);
    if (behavior & PopulateNewFolder) {
        folder.populate(populationBehavior);
    }
    return folder;
}
File *FileSystem::interpret(File *sourceData) {
    DENG2_ASSERT(sourceData != nullptr);
    LOG_AS("FS::interpret");
    try {
        for (filesys::IInterpreter const *i : d->interpreters) {
            if (auto *file = i->interpretFile(sourceData)) {
                return file;
            }
        }
    }
    catch (Error const &er) {
        LOG_RES_ERROR("Failed to interpret contents of %s: %s")
                << sourceData->description() << er.asText();
        delete sourceData;
        throw;
    }
    return sourceData;
}
FileIndex const &FileSystem::nameIndex() const {
    return d->index;
}
int FileSystem::findAll(String const &path, FoundFiles &found) const {
    LOG_AS("FS::findAll");
    found.clear();
    d->index.findPartialPath(path, found);
    return int(found.size());
}
LoopResult FileSystem::forAll(String const &partialPath, std::function<LoopResult (File &)> func) {
    FoundFiles files;
    findAll(partialPath, files);
    for (File *f : files) {
        if (auto result = func(*f)) return result;
    }
    return LoopContinue;
}
int FileSystem::findAllOfType(String const &typeIdentifier, String const &path, FoundFiles &found) const {
    LOG_AS("FS::findAllOfType");
    return findAllOfTypes(StringList() << typeIdentifier, path, found);
}
LoopResult FileSystem::forAllOfType(String const &typeIdentifier, String const &path,
                                    std::function<LoopResult (File &)> func) {
    FoundFiles files;
    findAllOfType(typeIdentifier, path, files);
    for (File *f : files) {
        if (auto result = func(*f)) return result;
    }
    return LoopContinue;
}
int FileSystem::findAllOfTypes(StringList typeIdentifiers, String const &path, FoundFiles &found) const {
    LOG_AS("FS::findAllOfTypes");
    found.clear();
    foreach (String const &id, typeIdentifiers) {
        indexFor(id).findPartialPath(path, found);
    }
    return int(found.size());
}
File &FileSystem::find(String const &path) const {
    return find<File>(path);
}
void FileSystem::index(File &file) {
    d->index.maybeAdd(file);
    d->getTypeIndex(DENG2_TYPE_NAME(file)).maybeAdd(file);
    foreach (FileIndex *user, d->userIndices) {
        user->maybeAdd(file);
    }
}
void FileSystem::deindex(File &file) {
    d->index.remove(file);
    d->getTypeIndex(DENG2_TYPE_NAME(file)).remove(file);
    foreach (FileIndex *user, d->userIndices) {
        user->remove(file);
    }
}
File &FileSystem::copySerialized(String const &sourcePath, String const &destinationPath,
                                 CopyBehaviors behavior) // static
{
    auto &fs = get();
    Block contents;
    *fs.root().locate<File const>(sourcePath).source() >> contents;
    File *dest = &fs.root().replaceFile(destinationPath);
    *dest << contents;
    dest->flush();
    if (behavior & ReinterpretDestination) {
        dest = dest->reinterpret();
    }
    if (behavior.testFlag(PopulateDestination) && is<Folder>(dest)) {
        dest->as<Folder>().populate();
    }
    return *dest;
}
void FileSystem::timeChanged(Clock const &) {
}
void FileSystem::changeBusyLevel(int increment) {
    using namespace std;
    bool       notify = false;
    BusyStatus bs     = Idle; {
        lock_guard<mutex> g(d->busyMutex);
        const int oldLevel = d->busyLevel;
        d->busyLevel += increment;
        if (d->busyLevel == 0) {
            notify = true;
            bs     = Idle;
            d->busyFinished.notify_all();
        }
        else if (oldLevel == 0) {
            notify = true;
            bs     = Busy;
        }
    }
    if (notify) {
        Loop::mainCall([this, bs]() {
            lock_guard<mutex> g(d->busyMutex);
            if ((bs == Busy && d->busyLevel > 0) ||
                (bs == Idle && d->busyLevel == 0)) {
                DENG2_FOR_AUDIENCE2(Busy, i) { i->fileSystemBusyStatusChanged(bs); }
            }
        });
    }
}
int FileSystem::busyLevel() const {
    std::lock_guard<std::mutex> lk(d->busyMutex);
    return d->busyLevel;
}
void FileSystem::waitForIdle() // static
{
    using namespace std;
    auto &fs = get();
    unique_lock<mutex> lk(fs.d->busyMutex);
    if (fs.d->busyLevel > 0) {
        LOG_MSG("Waiting until file system is ready");
        fs.d->busyFinished.wait(lk);
    }
}
FileIndex const &FileSystem::indexFor(String const &typeName) const {
    return d->getTypeIndex(typeName);
}
void FileSystem::addUserIndex(FileIndex &userIndex) {
    d->userIndices.insert(&userIndex);
}
void FileSystem::removeUserIndex(FileIndex &userIndex) {
    d->userIndices.remove(&userIndex);
}
void FileSystem::printIndex() {
    if (!LogBuffer::get().isEnabled(LogEntry::Generic | LogEntry::Dev | LogEntry::Verbose))
        return;
    LOG_DEBUG("Main FS index has %i entries") << d->index.size();
    d->index.print();
    DENG2_GUARD_FOR(d->typeIndex, G);
    DENG2_FOR_EACH_CONST(Impl::TypeIndex, i, d->typeIndex.value) {
        LOG_DEBUG("Index for type '%s' has %i entries") << i.key() << i.value()->size();
        LOG_AS_STRING(i.key());
        i.value()->print();
    }
}
String FileSystem::accessNativeLocation(NativePath const &nativePath, File::Flags flags) // static
{
    static const String SYS_NATIVE = "/sys/native";
    static const String VAR_MAPPING = "accessNames";
    auto &fs = get();
    Folder &sysNative = fs.makeFolder(SYS_NATIVE);
    if (!sysNative.objectNamespace().hasMember(VAR_MAPPING)) {
        sysNative.objectNamespace().addDictionary(VAR_MAPPING);
    }
    DictionaryValue &mapping = sysNative[VAR_MAPPING].value<DictionaryValue>();
    const TextValue key(nativePath);
    if (!mapping.contains(key)) {
        String name;
        do {
            name = String("%1").arg(Rangei(0, 65536).random(), 4, 16, QChar('0'));
        } while (sysNative.has(name));
        mapping.setElement(key, new TextValue(name));
    }
    File &f = DirectoryFeed::manuallyPopulateSingleFile(
        nativePath, fs.makeFolder(sysNative.path() / mapping[key].asText()));
    f.setMode(flags);
    return f.path();
    /*    String const path = folders / nativePath.fileNamePath().fileName();
    if (!fs.root().has(path)) {
        DirectoryFeed::Flags feedFlags = DirectoryFeed::OnlyThisFolder;
        if (flags.testFlag(File::Write)) feedFlags |= DirectoryFeed::AllowWrite;
        fs.makeFolderWithFeed(path,
                              new DirectoryFeed(nativePath.fileNamePath(), feedFlags),
                              Folder::PopulateOnlyThisFolder,
                              FS::DontInheritFeeds | FS::PopulateNewFolder)
                .setMode(flags);
    }*/
}
Folder &FileSystem::root() {
    return *d->root;
}
Folder const &FileSystem::root() const {
    return *d->root;
}
Folder &FileSystem::rootFolder() // static
{
    return get().root();
}
FileSystem &FileSystem::get() // static
{
    return App::fileSystem();
}
} // namespace de
namespace de {
FolderPopulationAudience audienceForFolderPopulation; // public
namespace internal {
static TaskPool populateTasks;
static bool     enableBackgroundPopulation = true;
struct PopulationNotifier : DENG2_OBSERVES(TaskPool, Done) {
    PopulationNotifier() { populateTasks.audienceForDone() += this; }
    void taskPoolDone(TaskPool &) { notify(); }
    void notify() { DENG2_FOR_AUDIENCE(FolderPopulation, i) i->folderPopulationFinished(); }
};
static PopulationNotifier populationNotifier;
} // namespace internal
DENG2_PIMPL(Folder) {
    Contents contents;
    Feeds feeds;
    Impl(Public *i) : Base(i) {}
    void add(File *file) {
        contents.insert(file->name().toLower(), file);
        file->setParent(thisPublic);
    }
    void destroy(String path, File *file) {
        Feed *originFeed = file->originFeed();
        delete file;
        if (originFeed) {
            originFeed->destroyFile(path);
        }
    }
    QList<Folder *> subfolders() const {
        DENG2_GUARD_FOR(self(), G);
        QList<Folder *> subs;
        for (Contents::const_iterator i = contents.begin(); i != contents.end(); ++i) {
            if (Folder *folder = maybeAs<Folder>(i.value())) {
                subs << folder;
            }
        }
        return subs;
    }
    static void destroyRecursive(Folder &folder) {
        foreach (Folder *sub, folder.subfolders()) {
            destroyRecursive(*sub);
        }
        folder.destroyAllFiles();
    }
};
Folder::Folder(String const &name) : File(name), d(new Impl(this)) {
    setStatus(Type::Folder);
    objectNamespace().addSuperRecord(ScriptSystem::builtInClass(QStringLiteral("Folder")));
}
Folder::~Folder() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
    clear();
    while (!d->feeds.isEmpty()) {
        delete d->feeds.takeLast();
    }
}
String Folder::describe() const {
    if (auto const *direcFeed = primaryFeedMaybeAs<DirectoryFeed>()) {
        return String("directory \"%1\"").arg(direcFeed->nativePath().pretty());
    }
    String desc;
    if (name().isEmpty()) {
        desc = "root folder";
    }
    else {
        desc = String("folder \"%1\"").arg(name());
    }
    String const feedDesc = describeFeeds();
    if (!feedDesc.isEmpty()) {
        desc += String(" (%1)").arg(feedDesc);
    }
    return desc;
}
String Folder::describeFeeds() const {
    DENG2_GUARD(this);
    String desc;
    if (d->feeds.size() == 1) {
        desc += String("contains %1 file%2 from %3")
                .arg(d->contents.size())
                .arg(DENG2_PLURAL_S(d->contents.size()))
                .arg(d->feeds.front()->description());
    }
    else if (d->feeds.size() > 1) {
        desc += String("contains %1 file%2 from %3 feed%4")
                .arg(d->contents.size())
                .arg(DENG2_PLURAL_S(d->contents.size()))
                .arg(d->feeds.size())
                .arg(DENG2_PLURAL_S(d->feeds.size()));
        int n = 0;
        DENG2_FOR_EACH_CONST(Feeds, i, d->feeds) {
            desc += String("; feed #%2 is %3")
                    .arg(n + 1)
                    .arg((*i)->description());
            ++n;
        }
    }
    return desc;
}
void Folder::clear() {
    DENG2_GUARD(this);
    if (d->contents.empty()) return;
    for (Contents::iterator i = d->contents.begin(); i != d->contents.end(); ++i) {
        i.value()->setParent(0);
        delete i.value();
    }
    d->contents.clear();
}
void Folder::populate(PopulationBehaviors behavior) {
    fileSystem().changeBusyLevel(+1);
    LOG_AS("Folder"); {
        DENG2_GUARD(this);
        QMutableMapIterator<String, File *> iter(d->contents);
        while (iter.hasNext()) {
            iter.next();
            bool mustPrune = false;
            File *file = iter.value();
            if (file->mode() & DontPrune) {
                continue;
            }
            Feed *originFeed = file->originFeed();
            if (originFeed && originFeed->prune(*file)) {
                LOG_RES_XVERBOSE("Pruning \"%s\" due to origin feed %s", file->path() << originFeed->description());
                mustPrune = true;
            }
            else if (!originFeed) {
                for (Feeds::iterator f = d->feeds.begin(); f != d->feeds.end(); ++f) {
                    if ((*f)->prune(*file)) {
                        LOG_RES_XVERBOSE("Pruning %s due to non-origin feed %s", file->path() << (*f)->description());
                        mustPrune = true;
                        break;
                    }
                }
            }
            if (mustPrune) {
                file->setParent(nullptr);
                iter.remove();
                delete file;
            }
        }
    }
    auto populationTask = [this, behavior]() {
        Feed::PopulatedFiles newFiles;
        for (int i = d->feeds.size() - 1; i >= 0; --i) {
            newFiles.append(d->feeds.at(i)->populate(*this));
        } {
            DENG2_GUARD(this);
            for (File *i : newFiles) {
                if (i) {
                    std::unique_ptr<File> file(i);
                    if (!d->contents.contains(i->name().toLower())) {
                        d->add(file.release());
                        fileSystem().index(*i);
                    }
                }
            }
            newFiles.clear();
        }
        if (behavior & PopulateFullTree) {
            for (Folder *folder : d->subfolders()) {
                folder->populate(behavior | PopulateCalledRecursively);
            }
        }
        fileSystem().changeBusyLevel(-1);
    };
    if (internal::enableBackgroundPopulation) {
        if (behavior & PopulateAsync) {
            internal::populateTasks.start(populationTask, TaskPool::MediumPriority);
        }
        else {
            populationTask();
        }
    }
    else {
        populationTask();
        if (!(behavior & PopulateCalledRecursively)) {
            internal::populationNotifier.notify();
        }
    }
}
Folder::Contents Folder::contents() const {
    DENG2_GUARD(this);
    return d->contents;
}
LoopResult Folder::forContents(std::function<LoopResult (String, File &)> func) const {
    DENG2_GUARD(this);
    for (Contents::const_iterator i = d->contents.constBegin(); i != d->contents.constEnd(); ++i) {
        if (auto result = func(i.key(), *i.value())) {
            return result;
        }
    }
    return LoopContinue;
}
QList<Folder *> Folder::subfolders() const {
    return d->subfolders();
}
File &Folder::createFile(String const &newPath, FileCreationBehavior behavior) {
    String path = newPath.fileNamePath();
    if (!path.empty()) {
        return locate<Folder>(path).createFile(newPath.fileName(), behavior);
    }
    verifyWriteAccess();
    if (behavior == ReplaceExisting && has(newPath)) {
        try {
            destroyFile(newPath);
        }
        catch (Feed::RemoveError const &er) {
            LOG_RES_WARNING("Failed to replace %s: existing file could not be removed.\n")
                    << newPath << er.asText();
        }
    }
    for (Feeds::iterator i = d->feeds.begin(); i != d->feeds.end(); ++i) {
        File *file = (*i)->createFile(newPath);
        if (file) {
            file->setMode(Write | DontPrune);
            add(file);
            fileSystem().index(*file);
            return *file;
        }
    }
    throw NewFileError("Folder::createFile", "Unable to create new file '" + newPath +
                       "' in " + description());
}
File &Folder::replaceFile(String const &newPath) {
    return createFile(newPath, ReplaceExisting);
}
void Folder::destroyFile(String const &removePath) {
    DENG2_GUARD(this);
    String path = removePath.fileNamePath();
    if (!path.empty()) {
        return locate<Folder>(path).destroyFile(removePath.fileName());
    }
    verifyWriteAccess();
    d->destroy(removePath, &locate<File>(removePath));
}
bool Folder::tryDestroyFile(String const &removePath) {
    try {
        if (has(removePath)) {
            destroyFile(removePath);
            return true;
        }
    }
    catch (Error const &) {
    }
    return false;
}
void Folder::destroyAllFiles() {
    DENG2_GUARD(this);
    verifyWriteAccess();
    foreach (File *file, d->contents) {
        file->setParent(nullptr);
        d->destroy(file->name(), file);
    }
    d->contents.clear();
}
void Folder::destroyAllFilesRecursively() {
    Impl::destroyRecursive(*this);
}
bool Folder::has(String const &name) const {
    if (name.isEmpty()) return false;
    String path = name.fileNamePath();
    if (!path.empty()) {
        Folder *folder = tryLocate<Folder>(path);
        if (folder) {
            return folder->has(name.fileName());
        }
        return false;
    }
    DENG2_GUARD(this);
    return (d->contents.find(name.lower()) != d->contents.end());
}
File &Folder::add(File *file) {
    DENG2_ASSERT(file != 0);
    if (has(file->name())) {
        throw DuplicateNameError("Folder::add", "Folder cannot contain two files with the same name: '" +
            file->name() + "'");
    }
    DENG2_GUARD(this);
    d->add(file);
    return *file;
}
File *Folder::remove(String name) {
    DENG2_GUARD(this);
    String const key = name.toLower();
    DENG2_ASSERT(d->contents.contains(key));
    File *removed = d->contents.take(key);
    removed->setParent(nullptr);
    return removed;
}
File *Folder::remove(char const *nameUtf8) {
    return remove(String(nameUtf8));
}
File *Folder::remove(File &file) {
    return remove(file.name());
}
filesys::Node const *Folder::tryGetChild(String const &name) const {
    DENG2_GUARD(this);
    Contents::const_iterator found = d->contents.find(name.toLower());
    if (found != d->contents.end()) {
        return found.value();
    }
    return nullptr;
}
Folder &Folder::root() {
    return FS::get().root();
}
void Folder::waitForPopulation(WaitBehavior waitBehavior) {
    if (waitBehavior == OnlyInBackground && App::inMainThread()) {
        DENG2_ASSERT(!App::inMainThread());
        throw Error("Folder::waitForPopulation", "Not allowed to block the main thread");
    }
    Time startedAt; {
        internal::populateTasks.waitForDone();
    }
    const auto elapsed = startedAt.since();
    if (elapsed > .01) {
        LOG_MSG("Waited for %.3f seconds for file system to be ready") << elapsed;
    }
}
AsyncTask *Folder::afterPopulation(std::function<void ()> func) {
    if (!isPopulatingAsync()) {
        func();
        return nullptr;
    }
    return async([] () {
        waitForPopulation();
        return 0;
    },
    [func] (int) {
        func();
    });
}
bool Folder::isPopulatingAsync() {
    return !internal::populateTasks.isDone();
}
void Folder::checkDefaultSettings() {
    String mtEnabled;
    if (App::app().unixInfo().defaults("fs:multithreaded", mtEnabled)) {
        internal::enableBackgroundPopulation = !ScriptedInfo::isFalse(mtEnabled);
    }
}
filesys::Node const *Folder::tryFollowPath(PathRef const &path) const {
    if (path.isAbsolute()) {
        return fileSystem().root().tryFollowPath(path.subPath(Rangei(1, path.segmentCount())));
    }
    return Node::tryFollowPath(path);
}
File *Folder::tryLocateFile(String const &path) const {
    if (filesys::Node const *node = tryFollowPath(Path(path))) {
        return const_cast<File *>(maybeAs<File>(node));
    }
    return nullptr;
}
void Folder::attach(Feed *feed) {
    if (feed) {
        DENG2_GUARD(this);
        d->feeds.push_back(feed);
    }
}
Feed *Folder::detach(Feed &feed) {
    DENG2_GUARD(this);
    d->feeds.removeOne(&feed);
    return &feed;
}
void Folder::setPrimaryFeed(Feed &feed) {
    DENG2_GUARD(this);
    d->feeds.removeOne(&feed);
    d->feeds.push_front(&feed);
}
Feed *Folder::primaryFeed() const {
    DENG2_GUARD(this);
    if (d->feeds.isEmpty()) return nullptr;
    return d->feeds.front();
}
void Folder::clearFeeds() {
    DENG2_GUARD(this);
    while (!d->feeds.empty()) {
        delete detach(*d->feeds.front());
    }
}
Folder::Feeds Folder::feeds() const {
    DENG2_GUARD(this);
    return d->feeds;
}
String Folder::contentsAsText() const {
    QList<File const *> files;
    forContents([&files] (String, File &f) {
        files << &f;
        return LoopContinue;
    });
    return File::fileListAsText(files);
}
} // namespace de
DENG2_PIMPL_NOREF(LibraryFile) {
    Library *library = nullptr;
    NativePath nativePath;
};
LibraryFile::LibraryFile(File *source)
    : File(source->name())
    , d(new Impl) {
    DENG2_ASSERT(source != 0);
    setSource(source); // takes ownership
}
LibraryFile::LibraryFile(NativePath const &nativePath)
    : File(nativePath.fileName())
    , d(new Impl) {
    d->nativePath = nativePath;
}
LibraryFile::~LibraryFile() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
    delete d->library;
}
String LibraryFile::describe() const {
    String desc = "shared library";
    if (loaded()) desc += " [" + library().type() + "]";
    return desc;
}
bool LibraryFile::loaded() const {
    return d->library != 0;
}
Library &LibraryFile::library() {
    if (d->library) {
        return *d->library;
    }
    if (!d->nativePath.isEmpty()) {
        d->library = new Library(d->nativePath);
    }
    else {
        NativeFile *native = maybeAs<NativeFile>(source());
        if (!native) {
            throw UnsupportedSourceError("LibraryFile::library", source()->description() +
                ": can only load from NativeFile");
        }
        d->library = new Library(native->nativePath());
    }
    return *d->library;
}
Library const &LibraryFile::library() const {
    if (d->library) return *d->library;
    throw NotLoadedError("LibraryFile::library", "Library is not loaded: " + description());
}
void LibraryFile::clear() {
    if (d->library) {
        delete d->library;
        d->library = nullptr;
    }
}
bool LibraryFile::hasUnderscoreName(String const &nameAfterUnderscore) const {
    return name().contains("_" + nameAfterUnderscore + ".") ||
           name().endsWith("_" + nameAfterUnderscore);
}
bool LibraryFile::recognize(File const &file) {
    #if defined (DENG_APPLE)
    {
        if (NativeFile const *native = maybeAs<NativeFile>(file)) {
            if (native->nativePath().fileNamePath().toString()
                .endsWith(file.name() + ".bundle/Contents/MacOS")) {
                return true;
            }
        }
    }
    #else // not Apple
    {
        if (QLibrary::isLibrary(file.name())) {
            #if defined(UNIX)
            {
                if (!file.name().endsWith(".so")) // just checks the file name
                {
                    return false;
                }
            }
            #endif
            return true;
        }
    }
    #endif
    return false;
}
File *LibraryFile::Interpreter::interpretFile(File *sourceData) const {
    if (recognize(*sourceData)) {
        LOG_RES_XVERBOSE("Interpreted %s as a shared library", sourceData->description());
        return new LibraryFile(sourceData);
    }
    return nullptr;
}
namespace de {
DENG2_PIMPL(LinkFile) {
    SafePtr<File const> target;
    Impl(Public *i)
        : Base(i)
        , target(i) {}
};
LinkFile::LinkFile(String const &name)
    : File(name)
    , d(new Impl(this))
{}
LinkFile::~LinkFile() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    deindex();
}
File const &LinkFile::target() const {
    DENG2_GUARD(this);
    if (d->target) {
        return *d->target;
    }
    return File::target();
}
File &LinkFile::target() {
    DENG2_GUARD(this);
    if (d->target) {
        return *const_cast<File *>(d->target.get());
    }
    return File::target();
}
Folder const *LinkFile::targetFolder() const {
    return maybeAs<Folder>(target());
}
Folder *LinkFile::targetFolder() {
    return maybeAs<Folder>(target());
}
void LinkFile::setTarget(File const &file) {
    DENG2_GUARD(this);
    d->target.reset(&file);
}
void LinkFile::setTarget(File const *fileOrNull) {
    DENG2_GUARD(this);
    d->target.reset(fileOrNull);
}
bool LinkFile::isBroken() const {
    return &target() == this;
}
String LinkFile::describe() const {
    DENG2_GUARD(this);
    if (!isBroken()) {
        DENG2_GUARD_FOR(target(), G);
        return "link to " + target().description();
    }
    return "broken link";
}
IIStream const &LinkFile::operator >> (IByteArray &bytes) const {
    if (!isBroken()) {
        target() >> bytes;
        return *this;
    }
    else {
        return File::operator >> (bytes);
    }
}
filesys::Node const *LinkFile::tryFollowPath(PathRef const &path) const {
    if (Folder const *folder = targetFolder()) {
        return folder->tryFollowPath(path);
    }
    return nullptr;
}
filesys::Node const *LinkFile::tryGetChild(String const &name) const {
    if (Folder const *folder = targetFolder()) {
        return folder->tryGetChild(name);
    }
    return nullptr;
}
LinkFile *LinkFile::newLinkToFile(File const &file, String linkName) {
    if (linkName.isEmpty()) linkName = file.name();
    LinkFile *link = new LinkFile(linkName);
    link->setTarget(file);
    link->setStatus(file.status());
    return link;
}
} // namespace de
namespace de {
DENG2_PIMPL(MetadataBank), public Lockable {
    struct Source : public ISource {
        Block metaId;
        Source(Block const &id) : metaId(id) {}
    };
    struct Data : public IData {
        Block metadata;
        bool isChanged = false;
        bool shouldBeSerialized() const override {
            return isChanged;
        }
        ISerializable *asSerializable() override {
            return &metadata;
        }
        virtual duint sizeInMemory() const override {
            return duint(metadata.size());
        }
    };
    Impl(Public *i) : Base(i) {}
    static DotPath pathFromId(String const &category, Block const &id) {
        DENG2_ASSERT(!id.isEmpty());
        String const hex = id.asHexadecimalText();
        return String("%1.%2.%3").arg(category).arg(hex.last()).arg(hex);
    }
};
MetadataBank::MetadataBank()
    : Bank("MetadataBank", SingleThread | EnableHotStorage, "/home/cache/metadata")
    , d(new Impl(this))
{}
MetadataBank::~MetadataBank() {
    unloadAll(InHotStorage);
}
MetadataBank &MetadataBank::get() // static
{
    return App::metadataBank();
}
Block MetadataBank::check(String const &category, Block const &id) {
    DENG2_GUARD(d);
    DotPath const path = Impl::pathFromId(category, id);
    if (!has(path)) {
        Bank::add(path, new Impl::Source(id));
    }
    return data(path).as<Impl::Data>().metadata;
}
void MetadataBank::setMetadata(String const &category, Block const &id, Block const &metadata) {
    DENG2_GUARD(d);
    DotPath const path = Impl::pathFromId(category, id);
    if (!has(path)) {
        Bank::add(path, new Impl::Source(id));
    }
    auto &entry = data(path).as<Impl::Data>();
    entry.metadata = metadata;
    entry.isChanged = true;
}
Block MetadataBank::metadata(String const &category, Block const &id) const {
    DENG2_GUARD(d);
    return data(Impl::pathFromId(category, id)).as<Impl::Data>().metadata;
}
void MetadataBank::clear() {
    DENG2_GUARD(d);
    Bank::clear();
    clearHotStorage();
}
Bank::IData *MetadataBank::loadFromSource(ISource &) {
    return newData();
}
Bank::IData *MetadataBank::newData() {
    return new Impl::Data;
}
} // namespace de
namespace de {
DENG2_PIMPL(NativeFile) {
    NativePath nativePath;
    mutable QFile *in;
    QFile *out;
    bool needTruncation;
    Impl(Public *i)
        : Base(i)
        , in(0)
        , out(0)
        , needTruncation(false)
    {}
    ~Impl() {
        DENG2_ASSERT(!in);
        DENG2_ASSERT(!out);
    }
    QFile &getInput() {
        if (!in) {
            in = new QFile(nativePath);
            if (!in->open(QFile::ReadOnly)) {
                delete in;
                in = 0;
                throw InputError("NativeFile::openInput", "Failed to read " + nativePath);
            }
        }
        return *in;
    }
    QFile &getOutput() {
        if (!out) {
            self().verifyWriteAccess();
            QFile::OpenMode fileMode = QFile::ReadWrite;
            if (self().mode() & Truncate) {
                if (needTruncation) {
                    fileMode |= QFile::Truncate;
                    needTruncation = false;
                }
            }
            out = new QFile(nativePath);
            if (!out->open(fileMode)) {
                delete out;
                out = 0;
                throw OutputError("NativeFile::output", "Failed to write " + nativePath +
                                  " (" + strerror(errno) + ")");
            }
            if (self().mode() & Truncate) {
                Status st = self().status();
                st.size = 0;
                st.modifiedAt = Time();
                self().setStatus(st);
            }
        }
        return *out;
    }
    void closeInput() {
        if (in) {
            delete in;
            in = 0;
        }
    }
    void closeOutput() {
        if (out) {
            delete out;
            out = 0;
        }
    }
};
NativeFile::NativeFile(String const &name, NativePath const &nativePath)
    : ByteArrayFile(name), d(new Impl(this)) {
    d->nativePath = nativePath;
}
NativeFile::~NativeFile() {
    DENG2_GUARD(this);
    DENG2_FOR_AUDIENCE2(Deletion, i) i->fileBeingDeleted(*this);
    audienceForDeletion().clear();
    close();
    deindex();
}
String NativeFile::describe() const {
    DENG2_GUARD(this);
    return String("\"%1\"").arg(d->nativePath.pretty());
}
Block NativeFile::metaId() const {
    if (name() == QStringLiteral("persist.pack")) {
        return Block();
    }
    return Block(File::metaId() + d->nativePath.toUtf8()).md5Hash();
}
void NativeFile::close() {
    DENG2_GUARD(this);
    flush();
    DENG2_ASSERT(!d->out);
    d->closeInput();
}
void NativeFile::flush() {
    DENG2_GUARD(this);
    d->closeOutput();
    DENG2_ASSERT(!d->out);
}
NativePath const &NativeFile::nativePath() const {
    DENG2_GUARD(this);
    return d->nativePath;
}
void NativeFile::clear() {
    DENG2_GUARD(this);
    File::clear();
    Flags oldMode = mode();
    setMode(Write | Truncate);
    d->getOutput();
    File::setMode(oldMode);
}
NativeFile::Size NativeFile::size() const {
    DENG2_GUARD(this);
    return status().size;
}
void NativeFile::get(Offset at, Byte *values, Size count) const {
    DENG2_GUARD(this);
    if (at + count > size()) {
        d->closeInput();
        throw OffsetError("NativeFile::get", description() + ": cannot read past end of file " +
                          String("(%1[+%2] > %3)").arg(at).arg(count).arg(size()));
    }
    QFile &in = input();
    if (in.pos() != qint64(at)) in.seek(qint64(at));
    in.read(reinterpret_cast<char *>(values), count);
    if (in.atEnd()) {
        d->closeInput();
    }
}
void NativeFile::set(Offset at, Byte const *values, Size count) {
    DENG2_GUARD(this);
    QFile &out = output();
    if (at > size()) {
        throw OffsetError("NativeFile::set", "Cannot write past end of file");
    }
    out.seek(at);
    out.write(reinterpret_cast<char const *>(values), count);
    if (out.error() != QFile::NoError) {
        throw OutputError("NativeFile::set", "Error writing to file:" +
                          out.errorString());
    }
    Status st = status();
    st.size = max(st.size, at + count);
    st.modifiedAt = Time();
    setStatus(st);
}
NativeFile *NativeFile::newStandalone(NativePath const &nativePath) {
    std::unique_ptr<NativeFile> file(new NativeFile(nativePath.fileName(), nativePath));
    if (nativePath.exists()) {
        file->setStatus(DirectoryFeed::fileStatus(nativePath));
    }
    return file.release();
}
void NativeFile::setMode(Flags const &newMode) {
    DENG2_GUARD(this);
    close();
    File::setMode(newMode);
    if (newMode.testFlag(Truncate)) {
        d->needTruncation = true;
    }
}
QFile &NativeFile::input() const {
    DENG2_GUARD(this);
    return d->getInput();
}
QFile &NativeFile::output() {
    DENG2_GUARD(this);
    return d->getOutput();
}
} // namespace de
namespace de {
static QString toNative(String const &s) {
    return Path::normalizeString(QDir::cleanPath(s), DIR_SEPARATOR);
}
NativePath::NativePath() : Path()
{}
NativePath::NativePath(NativePath const &other)
    : Path(other)
{}
NativePath::NativePath(NativePath &&moved)
    : Path(moved)
{}
NativePath::NativePath(String const &str) : Path(toNative(str), DIR_SEPARATOR)
{}
NativePath::NativePath(QString const &qstr) : Path(toNative(qstr), DIR_SEPARATOR)
{}
NativePath::NativePath(char const *nullTerminatedCStr)
    : Path(toNative(nullTerminatedCStr), DIR_SEPARATOR)
{}
NativePath::NativePath(char const *cStr, dsize length)
    : Path(toNative(String(cStr, length)), DIR_SEPARATOR)
{}
NativePath &NativePath::operator = (String const &str) {
    set(toNative(str), DIR_SEPARATOR);
    return *this;
}
NativePath &NativePath::operator = (QString const &str) {
    set(toNative(str), DIR_SEPARATOR);
    return *this;
}
NativePath &NativePath::operator = (NativePath &&moved) {
    Path::operator = (moved);
    return *this;
}
NativePath &NativePath::operator = (NativePath const &other) {
    Path::operator = (other);
    return *this;
}
NativePath &NativePath::operator = (char const *nullTerminatedCStr) {
    return (*this) = String(nullTerminatedCStr);
}
NativePath NativePath::concatenatePath(NativePath const &nativePath) const {
    if (nativePath.isAbsolute()) return nativePath;
    return toString().concatenatePath(nativePath, QChar(DIR_SEPARATOR));
}
NativePath NativePath::concatenatePath(String const &nativePath) const {
    return concatenatePath(NativePath(nativePath));
}
NativePath NativePath::operator / (NativePath const &nativePath) const {
    return concatenatePath(nativePath);
}
NativePath NativePath::operator / (String const &str) const {
    return *this / NativePath(str);
}
NativePath NativePath::operator / (QString const &str) const {
    return *this / NativePath(str);
}
NativePath NativePath::operator / (char const *nullTerminatedCStr) const {
    return *this / NativePath(nullTerminatedCStr);
}
NativePath NativePath::fileNamePath() const {
    return String(*this).fileNamePath(DIR_SEPARATOR);
}
bool NativePath::isAbsolute() const {
    return QDir::isAbsolutePath(expand());
}
bool NativePath::isDirectory() const {
    QFileInfo info(*this);
    return info.exists() && info.isDir();
}
NativePath NativePath::expand(bool *didExpand) const {
    if (first() == '>' || first() == '}') {
        if (didExpand) *didExpand = true;
        return App::app().nativeBasePath() / toString().mid(1);
    }
#ifdef UNIX
    else if (first() == '~') {
        if (didExpand) *didExpand = true;
        String const path = toString();
        int firstSlash = path.indexOf('/');
        if (firstSlash > 1) {
            QByteArray userName = path.mid(1, firstSlash - 1).toLatin1();
            struct passwd *pw = getpwnam(userName);
            if (!pw) {
                throw UnknownUserError("NativePath::expand",
                                       String("Unknown user '%1'").arg(QLatin1String(userName)));
            }
            return NativePath(pw->pw_dir) / path.mid(firstSlash + 1);
        }
        else {
            return NativePath(QDir::homePath()) / path.mid(2);
        }
    }
#endif
    if (didExpand) *didExpand = false;
    return *this;
}
String NativePath::pretty() const {
    if (isEmpty()) return *this;
    String result = *this;
    if (result.length() > 1 && (result.first() == '}' || result.first() == '>')) {
        return String(NATIVE_BASE_SYMBOLIC) + DIR_SEPARATOR + result.mid(1);
    }
    if (QDir::isAbsolutePath(result)) {
        NativePath basePath = App::app().nativeBasePath();
        if (result.beginsWith(basePath)) {
            result = NATIVE_BASE_SYMBOLIC + result.mid(basePath.length());
        }
        else {
#ifdef MACOSX
            NativePath contentsPath = App::app().nativeAppContentsPath();
            if (result.beginsWith(contentsPath)) {
                return "(app)" + result.mid(contentsPath.length());
            }
#endif
#ifndef WIN32 // Windows users are not familiar with a symbolic home path.
            NativePath homePath = QDir::homePath(); // actual native home dir, not FS2 "/home"
            if (result.beginsWith(homePath)) {
                result = NATIVE_HOME_SYMBOLIC + result.mid(homePath.length());
            }
#endif
        }
    }
    return result;
}
String NativePath::withSeparators(QChar sep) const {
    return Path::withSeparators(sep);
}
bool NativePath::exists() const {
    if (isEmpty()) return false;
    return QFile::exists(toString());
}
bool NativePath::isReadable() const {
    return QFileInfo(toString()).isReadable();
}
static NativePath currentNativeWorkPath;
NativePath NativePath::workPath() {
    if (currentNativeWorkPath.isEmpty()) {
        currentNativeWorkPath = QDir::currentPath();
    }
    return currentNativeWorkPath;
}
bool NativePath::setWorkPath(NativePath const &cwd) {
    if (QDir::setCurrent(cwd)) {
        currentNativeWorkPath = cwd;
        return true;
    }
    return false;
}
bool NativePath::exists(NativePath const &nativePath) {
    return QDir::current().exists(nativePath);
}
void NativePath::createPath(NativePath const &nativePath) {
    NativePath parentPath = nativePath.fileNamePath();
    if (!parentPath.isEmpty() && !exists(parentPath)) {
        parentPath.create();
    }
    QDir::current().mkdir(nativePath);
    if (!nativePath.exists()) {
        throw CreateDirError("NativePath::createPath", "Could not create: " + nativePath);
    }
}
bool NativePath::destroyPath(const NativePath &nativePath) {
    if (!nativePath.isEmpty()) {
        if (nativePath.isDirectory()) {
            return QDir::current().rmdir(nativePath);
        }
        else {
            return QDir::current().remove(nativePath);
        }
    }
    return true;
}
QChar NativePath::separator() {
    return DIR_SEPARATOR;
}
} // namespace de
namespace de {
namespace filesys {
DENG2_PIMPL_NOREF(Node) {
    String name;
    Node *parent;
    Impl(String const &name_) : name(name_), parent(0) {}
};
Node::Node(String const &name) : d(new Impl(name))
{}
Node::~Node()
{}
String Node::name() const {
    return d->name;
}
String Node::extension() const {
    return d->name.fileNameExtension();
}
void Node::setParent(Node *parent) {
    d->parent = parent;
}
Node *Node::parent() const {
    return d->parent;
}
bool Node::hasAncestor(Node const &possibleAncestor) const {
    for (Node const *iter = parent(); iter; iter = iter->parent()) {
        if (iter == &possibleAncestor) return true;
    }
    return false;
}
String Node::path() const {
    Node const *p = parent();
    if (!p) {
        return "/" + name();
    }
    return p->path() / name();
}
Node const *Node::tryFollowPath(PathRef const &path) const {
    static QString const DOT_SINGLE(".");
    static QString const DOT_DOUBLE("..");
    if (path.isEmpty()) {
        return this;
    }
    Path::Segment const &component = path.firstSegment();
    if (path.segmentCount() == 1 && component != DOT_DOUBLE) {
        if (component == DOT_SINGLE) {
            return this;
        }
        return tryGetChild(component);
    }
    PathRef const remainder = path.subPath(Rangei(1, path.segmentCount()));
    if (component == DOT_SINGLE) {
        return tryFollowPath(remainder);
    }
    if (component == DOT_DOUBLE) {
        if (!parent()) {
            return nullptr;
        }
        return parent()->tryFollowPath(remainder);
    }
    if (Node const *child = tryGetChild(component)) {
        return child->tryFollowPath(remainder);
    }
    return nullptr;
}
Node const *Node::tryGetChild(String const &) const {
    return nullptr;
}
} // namespace filesys
} // namespace de
namespace de {
String const Package::VAR_PACKAGE      ("package");
String const Package::VAR_PACKAGE_ID   ("package.ID");
String const Package::VAR_PACKAGE_ALIAS("package.alias");
String const Package::VAR_PACKAGE_TITLE("package.title");
String const Package::VAR_ID           ("ID");
String const Package::VAR_TITLE        ("title");
String const Package::VAR_VERSION      ("version");
static String const PACKAGE_VERSION    ("package.version");
static String const PACKAGE_ORDER      ("package.__order__");
static String const PACKAGE_IMPORT_PATH("package.importPath");
static String const PACKAGE_REQUIRES   ("package.requires");
static String const PACKAGE_RECOMMENDS ("package.recommends");
static String const PACKAGE_EXTRAS     ("package.extras");
static String const PACKAGE_PATH       ("package.path");
static String const PACKAGE_TAGS       ("package.tags");
static String const VAR_ID  ("ID");
static String const VAR_PATH("path");
static String const VAR_TAGS("tags");
Package::Asset::Asset(Record const &rec) : RecordAccessor(rec) {}
Package::Asset::Asset(Record const *rec) : RecordAccessor(rec) {}
String Package::Asset::absolutePath(String const &name) const {
    Record const *context = &accessedRecord().parentRecordForMember(name);
    if (!context->has(ScriptedInfo::VAR_SOURCE)) {
        context = &accessedRecord();
    }
    return ScriptedInfo::absolutePathInContext(*context, gets(name));
}
DENG2_PIMPL(Package) {
    SafePtr<const File> file;
    Version version; // version of the loaded package
    Impl(Public *i, File const *f)
        : Base(i)
        , file(f) {
        if (file) {
            version = split(versionedIdentifierForFile(*file)).second;
            if (!version.isValid()) {
                version = Version(metadata(*file).gets(VAR_VERSION, String()));
            }
        }
    }
    void verifyFile() const {
        if (!file) {
            throw SourceError("Package::verifyFile", "Package's source file missing");
        }
    }
    StringList importPaths() const {
        StringList paths;
        if (self().objectNamespace().has(PACKAGE_IMPORT_PATH)) {
            ArrayValue const &imp = self().objectNamespace().geta(PACKAGE_IMPORT_PATH);
            DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, imp.elements()) {
                Path importPath = (*i)->asText();
                if (!importPath.isAbsolute()) {
                    importPath = self().root().locate<File const >(importPath).path();
                }
                paths << importPath;
            }
        }
        return paths;
    }
    Record &packageInfo() {
        return self().objectNamespace().subrecord(VAR_PACKAGE);
    }
};
Package::Package(File const &file) : d(new Impl(this, &file))
{}
Package::~Package()
{}
File const &Package::file() const {
    d->verifyFile();
    return *d->file;
}
File const &Package::sourceFile() const {
    return FS::locate<const File>(objectNamespace().gets(PACKAGE_PATH));
}
bool Package::sourceFileExists() const {
    return d->file && FS::tryLocate<const File>(objectNamespace().gets(PACKAGE_PATH));
}
Folder const &Package::root() const {
    d->verifyFile();
    if (const Folder *f = maybeAs<Folder>(&d->file->target())) {
        return *f;
    }
    return *sourceFile().parent();
}
Record &Package::objectNamespace() {
    d->verifyFile();
    return const_cast<File *>(d->file.get())->objectNamespace();
}
Record const &Package::objectNamespace() const {
    return const_cast<Package *>(this)->objectNamespace();
}
String Package::identifier() const {
    d->verifyFile();
    return identifierForFile(*d->file);
}
Version Package::version() const {
    return d->version;
}
Package::Assets Package::assets() const {
    return ScriptedInfo::allBlocksOfType(QStringLiteral("asset"), d->packageInfo());
}
bool Package::executeFunction(String const &name) {
    Record &pkgInfo = d->packageInfo();
    if (pkgInfo.has(name)) {
        Process::scriptCall(Process::IgnoreResult, pkgInfo, name);
        return true;
    }
    return false;
}
void Package::setOrder(int ordinal) {
    objectNamespace().set(PACKAGE_ORDER, ordinal);
}
int Package::order() const {
    return objectNamespace().geti(PACKAGE_ORDER);
}
void Package::findPartialPath(String const &path, FileIndex::FoundFiles &found) const {
    App::fileSystem().nameIndex().findPartialPath(identifier(), path, found);
}
void Package::didLoad() {
    foreach (String imp, d->importPaths()) {
        App::scriptSystem().addModuleImportPath(imp);
    }
    executeFunction("onLoad");
}
void Package::aboutToUnload() {
    executeFunction("onUnload");
    foreach (String imp, d->importPaths()) {
        App::scriptSystem().removeModuleImportPath(imp);
    }
    delete objectNamespace().remove(PACKAGE_ORDER);
}
void Package::parseMetadata(File &packageFile) // static
{
    static String const TIMESTAMP("__timestamp__");
    if (Folder *folder = maybeAs<Folder>(packageFile)) {
        File *initializerScript = folder->tryLocateFile(QStringLiteral("__init__.ds"));
        if (!initializerScript) initializerScript = folder->tryLocateFile(QStringLiteral("__init__.de")); // old extension
        File *metadataInfo      = folder->tryLocateFile(QStringLiteral("Info.dei"));
        if (!metadataInfo) metadataInfo = folder->tryLocateFile(QStringLiteral("Info")); // alternate name
        Time parsedAt           = Time::invalidTime();
        bool needParse          = true;
        if (!metadataInfo && !initializerScript) return; // Nothing to do.
        if (packageFile.objectNamespace().has(VAR_PACKAGE)) {
            Record &metadata = packageFile.objectNamespace().subrecord(VAR_PACKAGE);
            if (metadata.has(TIMESTAMP)) {
                needParse = false;
                if (auto const *time = maybeAs<TimeValue>(metadata.get(TIMESTAMP))) {
                    needParse =
                            (metadataInfo      && metadataInfo->status().modifiedAt      > time->time()) ||
                            (initializerScript && initializerScript->status().modifiedAt > time->time());
                }
            }
            if (!needParse) return;
        }
        Record &metadata = initializeMetadata(packageFile);
        if (metadataInfo) {
            ScriptedInfo script(&metadata);
            script.parse(*metadataInfo);
            parsedAt = metadataInfo->status().modifiedAt;
        }
        if (initializerScript) {
            Script script(*initializerScript);
            Process proc(&metadata);
            proc.run(script);
            proc.execute();
            if (parsedAt.isValid() && initializerScript->status().modifiedAt > parsedAt) {
                parsedAt = initializerScript->status().modifiedAt;
            }
        }
        metadata.addTime(TIMESTAMP, parsedAt);
        if (LogBuffer::get().isEnabled(LogEntry::Dev | LogEntry::XVerbose | LogEntry::Resource)) {
            LOGDEV_RES_XVERBOSE("Parsed metadata of '%s':\n" _E(m),
                    identifierForFile(packageFile)
                    << packageFile.objectNamespace().asText());
        }
    }
}
void Package::validateMetadata(Record const &packageInfo) {
    if (!packageInfo.has(VAR_ID)) {
        throw NotPackageError("Package::validateMetadata", "Not a package");
    }
    DotPath const ident(packageInfo.gets(VAR_ID));
    if (ident.segmentCount() <= 1) {
        throw ValidationError("Package::validateMetadata",
                              QString("Identifier of package \"%1\" must specify a domain")
                              .arg(packageInfo.gets("path")));
    }
    String const &topLevelDomain = ident.segment(0).toString();
    if (topLevelDomain == QStringLiteral("feature") ||
        topLevelDomain == QStringLiteral("asset")) {
        throw ValidationError("Package::validateMetadata",
                              QString("Package \"%1\" has an invalid domain: functional top-level "
                                      "domains can only be used as aliases")
                              .arg(packageInfo.gets("path")));
    }
    static String const required[] = { "title", "version", "license", VAR_TAGS };
    for (auto const &req : required) {
        if (!packageInfo.has(req)) {
            throw IncompleteMetadataError("Package::validateMetadata",
                                          QString("Package \"%1\" does not have '%2' in its metadata")
                                          .arg(packageInfo.gets("path"))
                                          .arg(req));
        }
    }
    static QRegularExpression const regexReservedTags("\\b(loaded)\\b");
    auto match = regexReservedTags.match(packageInfo.gets(VAR_TAGS));
    if (match.hasMatch()) {
        throw ValidationError("Package::validateMetadata",
                              QString("Package \"%1\" has a tag that is reserved for internal use (%2)")
                              .arg(packageInfo.gets("path"))
                              .arg(match.captured(1)));
    }
}
Record &Package::initializeMetadata(File &packageFile, String const &id) {
    if (!packageFile.objectNamespace().has(VAR_PACKAGE)) {
        packageFile.objectNamespace().addSubrecord(VAR_PACKAGE);
    }
    Record &metadata = packageFile.objectNamespace().subrecord(VAR_PACKAGE);
    metadata.set(VAR_ID,   id.isEmpty()? identifierForFile(packageFile) : id);
    metadata.set(VAR_PATH, packageFile.path());
    return metadata;
}
Record const &Package::metadata(File const &packageFile) {
    return packageFile.objectNamespace().subrecord(VAR_PACKAGE);
}
QStringList Package::tags(File const &packageFile) {
    return tags(packageFile.objectNamespace().gets(PACKAGE_TAGS));
}
bool Package::matchTags(File const &packageFile, String const &tagRegExp) {
    return QRegExp(tagRegExp).indexIn(packageFile.objectNamespace().gets(PACKAGE_TAGS, "")) >= 0;
}
QStringList Package::tags(String const &tagsString) {
    return tagsString.split(' ', QString::SkipEmptyParts);
}
StringList Package::requires(File const &packageFile) {
    return packageFile.objectNamespace().getStringList(PACKAGE_REQUIRES);
}
void Package::addRequiredPackage(File &packageFile, String const &id) {
    packageFile.objectNamespace().appendToArray(PACKAGE_REQUIRES, new TextValue(id));
}
bool Package::hasOptionalContent(String const &packageId) {
    if (File const *file = PackageLoader::get().select(packageId)) {
        return hasOptionalContent(*file);
    }
    return false;
}
bool Package::hasOptionalContent(File const &packageFile) {
    Record const &meta = packageFile.objectNamespace();
    return meta.has(PACKAGE_RECOMMENDS) || meta.has(PACKAGE_EXTRAS);
}
static String stripAfterFirstUnderscore(String str) {
    int pos = str.indexOf('_');
    if (pos > 0) return str.left(pos);
    return str;
}
static String extractIdentifier(String str) {
    return stripAfterFirstUnderscore(str.fileNameWithoutExtension());
}
std::pair<String, Version> Package::split(String const &identifier_version) {
    std::pair<String, Version> idVer;
    if (identifier_version.contains(QChar('_'))) {
        idVer.first  = stripAfterFirstUnderscore(identifier_version);
        idVer.second = Version(identifier_version.substr(idVer.first.size() + 1));
    }
    else {
        idVer.first  = identifier_version;
    }
    return idVer;
}
String Package::splitToHumanReadable(String const &identifier_version) {
    auto const id_ver = split(identifier_version);
    return QObject::tr("%1 " _E(C) "(%2)" _E(.))
            .arg(id_ver.first)
            .arg(id_ver.second.isValid()? QObject::tr("version %1").arg(id_ver.second.fullNumber())
                                        : QObject::tr("any version"));
}
bool Package::equals(String const &id1, String const &id2) {
    return split(id1).first == split(id2).first;
}
String Package::identifierForFile(File const &file) {
    if (auto const *pkgId = file.objectNamespace().tryFind(VAR_PACKAGE_ID)) {
        return pkgId->value().asText();
    }
    String prefix;
    Folder const *parent = file.parent();
    while (parent && parent->extension() == ".pack") {
        prefix = extractIdentifier(parent->name()) + "." + prefix;
        parent = parent->parent();
    }
    return prefix + extractIdentifier(file.name());
}
String Package::versionedIdentifierForFile(File const &file) {
    String id = identifierForFile(file);
    if (id.isEmpty()) return String();
    auto const id_ver = split(file.name().fileNameWithoutExtension());
    if (id_ver.second.isValid()) {
        return String("%1_%2").arg(id).arg(id_ver.second.fullNumber());
    }
    if (auto const *pkgVer = file.objectNamespace().tryFind(PACKAGE_VERSION)) {
        return String("%1_%2").arg(id).arg(Version(pkgVer->value().asText()).fullNumber());
    }
    return id;
}
Version Package::versionForFile(File const &file) {
    return split(versionedIdentifierForFile(file)).second;
}
File const *Package::containerOfFile(File const &file) {
    File const *i = file.parent();
    while (i && i->extension() != ".pack") {
        i = i->parent();
    }
    return i;
}
String Package::identifierForContainerOfFile(File const &file) {
    File const *c = containerOfFile(file);
    return c? identifierForFile(*c) : "";
}
Time Package::containerOfFileModifiedAt(File const &file) {
    File const *c = containerOfFile(file);
    if (!c) return file.status().modifiedAt;
    return c->status().modifiedAt;
}
} // namespace de
namespace de {
static String const VAR_LINK_PACKAGE_ID("link.package");
DENG2_PIMPL(PackageFeed) {
    PackageLoader &loader;
    LinkMode linkMode;
    Filter filter;
    Impl(Public *i, PackageLoader &ldr, LinkMode lm)
        : Base(i), loader(ldr), linkMode(lm)
    {}
    File *linkToPackage(Package &pkg, String const &linkName, Folder const &folder) {
        if (folder.has(linkName)) return nullptr; // Already there, keep the existing link.
        if (filter && !filter(pkg)) return nullptr;
        String name;
        if (linkMode == LinkIdentifier) {
            name = linkName;
        }
        else {
            name = Package::versionedIdentifierForFile(pkg.file());
        }
        LinkFile *link = LinkFile::newLinkToFile(pkg.file(), name);
        link->setOriginFeed(thisPublic);
        link->objectNamespace().addText(VAR_LINK_PACKAGE_ID, pkg.identifier());
        return link;
    }
    PopulatedFiles populate(Folder const &folder) {
        PopulatedFiles populated;
        DENG2_FOR_EACH_CONST(PackageLoader::LoadedPackages, i, loader.loadedPackages()) {
            Package *pkg = i.value();
            populated << linkToPackage(*pkg, i.key(), folder);
            if (pkg->objectNamespace().has(Package::VAR_PACKAGE_ALIAS)) {
                populated << linkToPackage(*pkg, pkg->objectNamespace()
                                           .gets(Package::VAR_PACKAGE_ALIAS), folder);
            }
            foreach (String ident, pkg->assets()) {
                populated << linkToPackage(*pkg, "asset." + ident, folder);
            }
        }
        return populated;
    }
};
PackageFeed::PackageFeed(PackageLoader &loader, LinkMode linkMode)
    : d(new Impl(this, loader, linkMode))
{}
void PackageFeed::setFilter(Filter filter) {
    d->filter = filter;
}
PackageLoader &PackageFeed::loader() {
    return d->loader;
}
String PackageFeed::description() const {
    return "loaded packages";
}
Feed::PopulatedFiles PackageFeed::populate(Folder const &folder) {
    return d->populate(folder);
}
bool PackageFeed::prune(File &file) const {
    if (LinkFile const *link = maybeAs<LinkFile>(file)) {
        if (!d->loader.isLoaded(link->objectNamespace().gets(VAR_LINK_PACKAGE_ID)))
            return true; {
            if (link->status() != link->target().status()) return true;
        }
    }
    return false; // Don't prune.
}
} // namespace de
namespace de {
static String const VAR_PACKAGE_VERSION("package.version");
DENG2_PIMPL(PackageLoader)
, DENG2_OBSERVES(File, Deletion) // loaded package source file is deleted?
{
    LoadedPackages loaded; ///< Identifiers are unversioned; only one version can be loaded at a time.
    int loadCounter;
    Impl(Public *i) : Base(i), loadCounter(0)
    {}
    ~Impl() {
        qDeleteAll(loaded.values());
    }
    /**
     * Determines if a specific package is currently loaded.
     * @param file  Package root.
     * @return @c true, if loaded.
     */
    Package *tryFindLoaded(File const &file) const {
        #if 0
        {
            qDebug() << "tryFindLoaded:" << Package::identifierForFile(file) << file.description();
            for (auto i = loaded.begin(); i != loaded.end(); ++i) {
                qDebug() << "  currently loaded:" << i.key() << "file:" << i.value()->file().path() << "source:" << i.value()->sourceFile().path();
            }
        }
        #endif
        LoadedPackages::const_iterator found = loaded.constFind(Package::identifierForFile(file));
        if (found != loaded.constEnd()) {
            auto const &pkg = *found.value();
            if (&pkg.file() == &file || &pkg.sourceFile() == &file) {
                return found.value();
            }
        }
        return nullptr;
    }
    static bool shouldIgnoreFile(File const &packageFile) {
        if (Feed const *feed = packageFile.originFeed()) {
            if (is<PackageFeed>(feed)) return true;
        }
        return false;
    }
    int findAllVariants(String const &packageIdVer, FS::FoundFiles &found) const {
        auto idVer = Package::split(packageIdVer);
        String const packageId = idVer.first;
        QStringList const components = packageId.split('.');
        String id;
        for (int i = components.size() - 1; i >= 0; --i) {
            id = components.at(i) + (!id.isEmpty()? "." + id : "");
            FS::FoundFiles files;
            App::fileSystem().findAllOfTypes(StringList()
                                             << DENG2_TYPE_NAME(Folder)
                                             << DENG2_TYPE_NAME(ArchiveFolder)
                                             << DENG2_TYPE_NAME(LinkFile),
                                             id + ".pack", files);
            files.remove_if([&packageId] (File *file) {
                if (shouldIgnoreFile(*file)) return true;
                return Package::identifierForFile(*file) != packageId;
            });
            std::copy(files.begin(), files.end(), std::back_inserter(found));
        }
        return int(found.size());
    }
    /**
     * Parses or updates the metadata of a package, and checks it for validity.
     * A Package::ValidationError is thrown if the package metadata does not
     * comply with the minimum requirements.
     *
     * @param packFile  Package file (".pack" folder).
     */
    static void checkPackage(File &packFile) {
        Package::parseMetadata(packFile);
        if (!packFile.objectNamespace().has(Package::VAR_PACKAGE)) {
            throw Package::NotPackageError("PackageLoader::checkPackage",
                                           packFile.description() + " is not a package");
        }
        Package::validateMetadata(packFile.objectNamespace().subrecord("package"));
    }
    File const *selectPackage(IdentifierList const &idList) const {
        for (auto const &id : idList.ids) {
            if (File const *f = selectPackage(id))
                return f;
        }
        return nullptr;
    }
    /**
     * Given a package identifier, pick one of the available versions of the package
     * based on predefined criteria.
     *
     * @param packageId  Package identifier. This may optionally include the
     *                   required version number.
     *
     * @return Selected package, or @c NULL if a version could not be selected.
     */
    File const *selectPackage(String const &packageId) const {
        LOG_AS("selectPackage");
        FS::FoundFiles found;
        if (!findAllVariants(packageId, found)) {
            return nullptr;
        } {
            FS::FoundFiles checked;
            DENG2_FOR_EACH_CONST(FS::FoundFiles, i, found) {
                try {
                    checkPackage(**i);
                    checked.push_back(*i);
                }
                catch (const Error &) {
                }
            }
            found = std::move(checked);
        }
        bool const mustMatchExactVersion = (packageId.contains('_'));
        if (mustMatchExactVersion) {
            auto const idVer = Package::split(packageId);
            found.remove_if([&idVer] (File *f) {
                Version const pkgVer = f->objectNamespace().gets(VAR_PACKAGE_VERSION);
                return (pkgVer != idVer.second); // Ignore other versions.
            });
            if (found.empty()) return nullptr;
        }
        found.sort([] (File const *a, File const *b) -> bool {
            Version const aVer(a->objectNamespace().gets(VAR_PACKAGE_VERSION));
            Version const bVer(b->objectNamespace().gets(VAR_PACKAGE_VERSION));
            if (aVer == bVer) {
                return de::cmp(a->status().modifiedAt, b->status().modifiedAt) < 0;
            }
            return aVer < bVer;
        });
        LOG_RES_XVERBOSE("Selected '%s': %s", packageId << found.back()->description());
        return found.back();
    }
    Package &load(String const &packageId, File const &source) {
        if (loaded.contains(packageId)) {
            throw AlreadyLoadedError("PackageLoader::load",
                                     "Package '" + packageId + "' is already loaded from \"" +
                                     loaded[packageId]->objectNamespace().gets("path") + "\"");
        }
        loadRequirements(source);
        if (App::configExists()) {
            loadOptionalContent(source);
        }
        Package *pkg = new Package(source);
        loaded.insert(packageId, pkg);
        pkg->setOrder(loadCounter++);
        pkg->didLoad();
        source.audienceForDeletion() += this;
        return *pkg;
    }
    bool unload(String identifier) {
        LoadedPackages::iterator found = loaded.find(identifier);
        if (found == loaded.end()) return false;
        Package *pkg = found.value();
        pkg->aboutToUnload();
        if (pkg->sourceFileExists()) {
            pkg->sourceFile().audienceForDeletion() -= this;
        }
        delete pkg;
        loaded.remove(identifier);
        return true;
    }
    void fileBeingDeleted(const File &file) override {
        const String pkgId = Package::identifierForFile(file);
        LOG_RES_WARNING("Unloading package \"%s\": source file has been modified or deleted")
            << pkgId;
        unload(pkgId);
    }
    void listPackagesInIndex(FileIndex const &index, StringList &list) {
        foreach (File *indexedFile, index.files()) {
            if (shouldIgnoreFile(*indexedFile)) continue;
            String const fileName = indexedFile->name();
            if (fileName.fileNameExtension() == ".pack") {
                try {
                    String path = indexedFile->path();
                    if (path == QStringLiteral("/home/persist.pack")) continue;
                    checkPackage(*indexedFile);
                    list.append(path);
                }
                catch (Package::NotPackageError const &er) {
                    LOG_RES_XVERBOSE("\"%s\": %s", fileName << er.asText());
                }
                catch (Package::ValidationError const &er) {
                    LOG_RES_MSG("\"%s\": Package is invalid: %s") << fileName << er.asText();
                }
                catch (Parser::SyntaxError const &er) {
                    LOG_RES_WARNING("\"%s\": Package has a Doomsday Script syntax error: %s")
                            << fileName << er.asText();
                }
                catch (Info::SyntaxError const &er) {
                    LOG_RES_WARNING("\"%s\": Package has a syntax error: %s")
                            << fileName << er.asText();
                }
                catch (const Error &er) {
                    LOG_RES_WARNING("\"%s\": %s") << fileName << er.asText();
                }
            }
        }
    }
    void loadRequirements(File const &packageFile) {
        for (String const &reqId : Package::requires(packageFile)) {
            if (!self().isLoaded(reqId)) {
                self().load(reqId);
            }
        }
    }
    void forOptionalContent(File const &packageFile,
                            std::function<void (String const &)> callback) const {
        DictionaryValue const &selPkgs = Config::get("fs.selectedPackages")
                                         .value<DictionaryValue>();
        const Record &meta = Package::metadata(packageFile);
        const String idVer = Package::versionedIdentifierForFile(packageFile);
        auto isPackageSelected = [&idVer, &selPkgs] (TextValue const &id, bool byDefault) -> bool {
            TextValue const key(idVer);
            if (selPkgs.contains(key)) {
                auto const &sels = selPkgs.element(key).as<DictionaryValue>();
                if (sels.contains(id)) {
                    return sels.element(id).isTrue();
                }
                else {
                }
            }
            else {
            }
            return byDefault;
        };
        if (meta.has("recommends")) {
            for (auto const *id : meta.geta("recommends").elements()) {
                String const pkgId = id->asText();
                if (isPackageSelected(pkgId, true)) {
                    callback(pkgId);
                }
            }
        }
        if (meta.has("extras")) {
            for (auto const *id : meta.geta("extras").elements()) {
                String const pkgId = id->asText();
                if (isPackageSelected(pkgId, false)) {
                    callback(pkgId);
                }
            }
        }
    }
    void loadOptionalContent(File const &packageFile) {
        forOptionalContent(packageFile, [this] (String const &pkgId) {
            if (!self().isLoaded(pkgId)) {
                self().load(pkgId);
            }
        });
    }
    QList<Package *> loadedInOrder() const {
        QList<Package *> pkgs = loaded.values();
        qSort(pkgs.begin(), pkgs.end(), [] (Package const *a, Package const *b) {
            return a->order() < b->order();
        });
        return pkgs;
    }
    DENG2_PIMPL_AUDIENCE(Activity)
    DENG2_PIMPL_AUDIENCE(Load)
    DENG2_PIMPL_AUDIENCE(Unload)
};
DENG2_AUDIENCE_METHOD(PackageLoader, Activity)
DENG2_AUDIENCE_METHOD(PackageLoader, Load)
DENG2_AUDIENCE_METHOD(PackageLoader, Unload)
PackageLoader::PackageLoader() : d(new Impl(this))
{}
bool PackageLoader::isAvailable(String const &packageId) const {
    return d->selectPackage(IdentifierList(packageId)) != nullptr;
}
File const *PackageLoader::select(String const &packageId) const {
    return d->selectPackage(IdentifierList(packageId));
}
Package const &PackageLoader::load(String const &packageId) {
    LOG_AS("PackageLoader");
    File const *packFile = d->selectPackage(IdentifierList(packageId));
    if (!packFile) {
        throw NotFoundError("PackageLoader::load",
                            "Package \"" + packageId + "\" is not available");
    }
    String const id = Package::identifierForFile(*packFile);
    d->load(id, *packFile);
    try {
        DENG2_FOR_AUDIENCE2(Load, i) {
            i->packageLoaded(id);
        }
        DENG2_FOR_AUDIENCE2(Activity, i) {
            i->setOfLoadedPackagesChanged();
        }
    }
    catch (Error const &er) {
        unload(id);
        throw PostLoadError("PackageLoader::load",
                            "Error during post-load actions of package \"" +
                            id + "\": " + er.asText());
    }
    return package(id);
}
void PackageLoader::unload(String const &packageId) {
    String id = Package::split(packageId).first;
    if (isLoaded(id)) {
        DENG2_FOR_AUDIENCE2(Unload, i) {
            i->aboutToUnloadPackage(id);
        }
        d->unload(id);
        DENG2_FOR_AUDIENCE2(Activity, i) {
            i->setOfLoadedPackagesChanged();
        }
    }
}
void PackageLoader::unloadAll() {
    LOG_AS("PackageLoader");
    LOG_RES_MSG("Unloading %i packages") << d->loaded.size();
    while (!d->loaded.isEmpty()) {
        unload(d->loaded.begin().key());
    }
}
bool PackageLoader::isLoaded(String const &packageId) const {
    auto const id_ver = Package::split(packageId);
    auto found = d->loaded.constFind(id_ver.first);
    if (found == d->loaded.constEnd()) {
        return false;
    }
    return (!id_ver.second.isValid() /* no valid version provided in argumnet */ ||
            id_ver.second == found.value()->version());
}
bool PackageLoader::isLoaded(File const &file) const {
    return d->tryFindLoaded(file) != nullptr;
}
Package const *PackageLoader::tryFindLoaded(File const &file) const {
    return d->tryFindLoaded(file);
}
PackageLoader::LoadedPackages const &PackageLoader::loadedPackages() const {
    return d->loaded;
}
QList<Package *> PackageLoader::loadedPackagesInOrder() const {
    return d->loadedInOrder();
}
FS::FoundFiles PackageLoader::loadedPackagesAsFilesInPackageOrder() const {
    QList<Package *> pkgs = d->loadedInOrder();
    FS::FoundFiles sorted;
    for (auto p : pkgs) {
        sorted.push_back(const_cast<File *>(&p->sourceFile()));
    }
    return sorted;
}
StringList PackageLoader::loadedPackageIdsInOrder(IdentifierType idType) const {
    QList<Package *> pkgs = d->loadedInOrder();
    StringList ids;
    for (auto p : pkgs) {
        Record const &meta = Package::metadata(p->file());
        Version const pkgVersion(meta.gets("version"));
        if (idType == Versioned && pkgVersion.isValid()) // nonzero
        {
            ids << String("%1_%2").arg(meta.gets("ID")).arg(pkgVersion.fullNumber());
        }
        else {
            ids << meta.gets("ID");
        }
    }
    return ids;
}
Package const &PackageLoader::package(String const &packageId) const {
    if (!isLoaded(packageId)) {
        throw NotFoundError("PackageLoader::package", "Package '" + packageId + "' is not loaded");
    }
    return *d->loaded[packageId];
}
void PackageLoader::sortInPackageOrder(FS::FoundFiles &filesToSort) const {
    typedef std::pair<File *, int> FileAndOrder;
    QList<FileAndOrder> all;
    DENG2_FOR_EACH_CONST(FS::FoundFiles, i, filesToSort) {
        Package const *pkg = 0;
        String identifier = Package::identifierForContainerOfFile(**i);
        if (isLoaded(identifier)) {
            pkg = &package(identifier);
        }
        all << FileAndOrder(*i, pkg? pkg->order() : -1);
    }
    std::sort(all.begin(), all.end(), [] (FileAndOrder const &a, FileAndOrder const &b) {
        return a.second < b.second;
    });
    filesToSort.clear();
    foreach (FileAndOrder const &f, all) {
        filesToSort.push_back(f.first);
    }
}
StringList PackageLoader::loadedFromCommandLine() const {
    StringList pkgs;
    CommandLine &args = App::commandLine();
    for (duint p = 0; p < duint(args.count()); ) {
        if (!args.matches("-pkg", args.at(p))) {
            ++p;
            continue;
        }
        while (++p != duint(args.count()) && !args.isOption(p)) {
            pkgs << args.at(p);
        }
    }
    return pkgs;
}
StringList PackageLoader::findAllPackages() const {
    StringList all;
    for (QString typeName : QStringList({ DENG2_TYPE_NAME(Folder),
                                          DENG2_TYPE_NAME(ArchiveFolder),
                                          DENG2_TYPE_NAME(LinkFile) })) {
        d->listPackagesInIndex(App::fileSystem().indexFor(typeName), all);
    }
    return all;
}
StringList PackageLoader::expandDependencies(StringList const &packageIdentifiers) const {
    StringList expanded;
    for (String pkgId : packageIdentifiers) {
        if (File const *file = select(pkgId)) {
            for (String reqId : Package::requires(*file)) {
                expanded << reqId;
            }
            d->forOptionalContent(*file, [&expanded] (String const &pkgId) {
                expanded << pkgId;
            });
        }
        expanded << pkgId;
    }
    return expanded;
}
PackageLoader &PackageLoader::get() {
    return App::packageLoader();
}
PackageLoader::IdentifierList::IdentifierList(String const &spaceSeparatedIds) {
    static const QRegularExpression anySpace("\\s");
    for (auto const &qs : spaceSeparatedIds.split(anySpace, String::SkipEmptyParts)) {
        ids.append(qs);
    }
}
} // namespace de
namespace de {
StaticLibraryFeed::StaticLibraryFeed()
{}
String StaticLibraryFeed::description() const {
    return "imported static libraries";
}
Feed::PopulatedFiles StaticLibraryFeed::populate(Folder const &folder) {
    PopulatedFiles files;
#if defined (DENG_STATIC_LINK)
    for (String name : Library::staticLibraries()) {
        if (!folder.has(name)) {
            files << new LibraryFile(NativePath(name));
        }
    }
#else
    DENG2_UNUSED(folder);
#endif
    return files;
}
bool StaticLibraryFeed::prune(File &) const {
    return false;
}
} // namespace de
