﻿/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Address"
#include "de/String"
#include "../src/net/networkinterfaces.h"
#include <QHostInfo>
#include <QRegularExpression>
/** @file beacon.cpp  Presence service based on UDP broadcasts.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Beacon"
#include "de/Reader"
#include "de/Writer"
#include "de/LogBuffer"
#include <QUdpSocket>
#include <QHostInfo>
#include <QTimer>
#include <QMap>
/** @file blockpacket.cpp  Packet that contains a block of data.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/BlockPacket"
#include "de/Writer"
#include "de/Reader"
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/IdentifiedPacket"
#include "de/Writer"
#include "de/Reader"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ListenSocket"
#include "de/Address"
#include "de/Socket"
#include <QCoreApplication>
#include <QTcpServer>
#include <QThread>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Message"
/** @file networkinterfaces.cpp  Information about network interfaces.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "../src/net/networkinterfaces.h"
#include "de/Task"
#include "de/TaskPool"
#include <QNetworkInterface>
#include <QTimer>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Packet"
#include "de/ByteRefArray"
#include "de/Reader"
#include "de/String"
#include "de/Writer"
#include <QByteArray>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Protocol"
#include "de/RecordPacket"
#include "de/TextValue"
#include "de/Record"
#include "de/Transmitter"
#include "de/Log"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RecordPacket"
#include "de/Record"
#include "de/Writer"
#include "de/Reader"
#include "de/Block"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
/**
 * @page sysNetwork Low-Level Networking
 *
 * On server-side connected clients can be either in "unjoined" mode or
 * "joined" mode. The former is for querying information about the server's
 * status, while the latter one is for clients participating in the on-going
 * game.
 *
 * In unjoined mode, messages are not compressed.
 * @todo Document uncompressed message format.
 *
 * @section netProtocol Network Protocol
 *
 * In joined mode, the network protocol works as follows. All messages are sent
 * over a TCP socket. Every message consists of a header and the message payload.
 * The content of these depends on the compressed message size.
 *
 * @par 1&ndash;127 bytes
 * Very small messages, such as the position updates that a client streams
 * to the server, are encoded with Huffman codes (see huffman.h). If
 * the Huffman coded payload happens to exceed 127 bytes, the message is
 * switched to the medium format (see below). Message structure:
 * - 1 byte: payload size
 * - @em n bytes: payload contents (Huffman)
 *
 * @par 128&ndash;4095 bytes
 * Medium-sized messages are compressed either using a fast zlib deflate level,
 * or Huffman codes if it yields better compression.
 * If the deflated message size exceeds 4095 bytes, the message is switched to
 * the large format (see below). Message structure:
 * - 1 byte: 0x80 | (payload size & 0x7f)
 * - 1 byte: (payload size >> 7) | (0x40 for deflated, otherwise Huffman)
 * - @em n bytes: payload contents (as produced by ZipFile::compressAtLevel()).
 *
 * @par >= 4096 bytes (up to 4MB)
 * Large messages are compressed using the best zlib deflate level.
 * Message structure:
 * - 1 byte: 0x80 | (payload size & 0x7f)
 * - 1 byte: 0x80 | (payload size >> 7) & 0x7f
 * - 1 byte: payload size >> 14
 * - @em n bytes: payload contents (as produced by ZipFile::compressAtLevel()).
 *
 * Messages larger than or equal to 2^22 bytes (about 4MB) must be broken into
 * smaller pieces before sending.
 *
 * @see Protocol_Send()
 * @see Protocol_Receive()
 */
#include "de/Socket"
#include "de/Async"
#include "de/Message"
#include "de/Writer"
#include "de/Reader"
#include "de/data/huffman.h"
#include <QThread>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Transmitter"
#include "de/Block"
#include "de/Packet"
#include "de/Writer"
#include "de/Reader"

namespace de {
DENG2_PIMPL_NOREF(Address) {
    std::shared_ptr<QHostAddress> host;
    duint16                       port = 0;
    String                        textRepr;
    enum Special { Undefined, LocalHost, RemoteHost };
    Special special = Undefined;
    void clearCached() {
        textRepr.clear();
        special = Impl::Undefined;
    }
};
Address::Address() : d(new Impl)
{}
Address::Address(char const *address, duint16 port) : d(new Impl) {
    d->port = port;
    if (QLatin1String(address) == "localhost") // special case
    {
        d->host.reset(new QHostAddress(QHostAddress::LocalHostIPv6));
        d->special = Impl::LocalHost;
    }
    else {
        d->host.reset(new QHostAddress(QHostAddress(address).toIPv6Address()));
    }
}
Address::Address(QHostAddress const &host, duint16 port) : d(new Impl) {
    d->host.reset(new QHostAddress(host.toIPv6Address()));
    d->port = port;
}
Address::Address(Address const &other)
    : d(new Impl(*other.d))
{}
Address &Address::operator=(Address const &other) {
    d->host     = other.d->host;
    d->port     = other.d->port;
    d->textRepr = other.d->textRepr;
    d->special  = other.d->special;
    return *this;
}
bool Address::operator<(Address const &other) const {
    return asText() < other.asText();
}
bool Address::operator==(Address const &other) const {
    return asText() == other.asText();
}
bool Address::isNull() const {
    return d->host->isNull();
}
QHostAddress const &Address::host() const {
    return *d->host;
}
void Address::setHost(QHostAddress const &host) {
    d->clearCached();
    d->host.reset(new QHostAddress(host.toIPv6Address()));
}
bool Address::isLocal() const {
    if (d->special == Impl::Undefined) {
        d->special = isHostLocal(*d->host) ? Impl::LocalHost : Impl::RemoteHost;
    }
    return (d->special == Impl::LocalHost);
}
duint16 Address::port() const {
    return d->port;
}
void Address::setPort(duint16 p) {
    d->clearCached();
    d->port = p;
}
bool Address::matches(Address const &other, duint32 mask) {
    return (d->host->toIPv4Address() & mask) == (other.d->host->toIPv4Address() & mask);
}
String Address::asText() const {
    if (!d->textRepr) {
        d->textRepr = (isLocal()? String("localhost") : d->host->toString());
        if (d->port) {
            d->textRepr += ":" + QString::number(d->port);
        }
    }
    return d->textRepr;
}
Address Address::parse(String const &addressWithOptionalPort, duint16 defaultPort) // static
{
    duint16 port = defaultPort;
    String str = addressWithOptionalPort;
    /*int portPosMin = 1;
    if (str.beginsWith(QStringLiteral("::ffff:"))) {
        portPosMin = 8;
    }
    int pos = str.lastIndexOf(':');
    if (pos >= portPosMin) {
        port = duint16(str.mid(pos + 1).toInt());
        str = str.left(pos);
    }*/
    static QRegularExpression const ipPortRegex
            ("^(localhost|::1|(::ffff:)?[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+|[:A-Fa-f0-9]+[A-Fa-f0-9]):([0-9]+)$");
    auto match = ipPortRegex.match(addressWithOptionalPort);
    if (match.hasMatch()) {
        str  = match.captured(1);
        port = duint16(match.captured(3).toInt());
    }
    else {
    }
    return Address(str.toLatin1(), port);
}
QTextStream &operator<<(QTextStream &os, Address const &address) {
    os << address.asText();
    return os;
}
bool Address::isHostLocal(QHostAddress const &host) // static
{
    if (host.isLoopback()) return true;
    QHostAddress const hostv6(host.toIPv6Address());
    foreach (QHostAddress addr, internal::NetworkInterfaces::get().allAddresses()) {
        if (addr == hostv6)
            return true;
    }
    return false;
}
} // namespace de
namespace de {
/**
 * Maximum number of Beacon UDP ports in simultaneous use at one machine, i.e.,
 * maximum number of servers on one machine.
 */
static duint16 const MAX_LISTEN_RANGE = 16;
static char const *discoveryMessage = "Doomsday Beacon 1.1";
DENG2_PIMPL_NOREF(Beacon) {
    duint16 port;
    duint16 servicePort;
    QUdpSocket *socket;
    Block message;
    QTimer *timer;
    Time discoveryEndsAt;
    QMap<Address, Block> found;
    Impl() : socket(0), timer(0)
    {}
    ~Impl() {
        delete socket;
        delete timer;
    }
};
Beacon::Beacon(duint16 port) : d(new Impl) {
    d->port = port;
}
duint16 Beacon::port() const {
    return d->port;
}
void Beacon::start(duint16 serviceListenPort) {
    DENG2_ASSERT(!d->socket);
    d->servicePort = serviceListenPort;
    d->socket = new QUdpSocket;
    connect(d->socket, SIGNAL(readyRead()), this, SLOT(readIncoming()));
    for (duint16 attempt = 0; attempt < MAX_LISTEN_RANGE; ++attempt) {
        if (d->socket->bind(d->port + attempt, QUdpSocket::DontShareAddress)) {
            d->port = d->port + attempt;
            return;
        }
    }
    throw PortError("Beacon::start", "Could not bind to UDP port " + String::number(d->port));
}
void Beacon::setMessage(IByteArray const &advertisedMessage) {
    d->message.clear();
    Writer(d->message) << d->servicePort;
    d->message += Block(advertisedMessage).compressed();
}
void Beacon::stop() {
    delete d->socket;
    d->socket = 0;
}
void Beacon::discover(TimeSpan const &timeOut, TimeSpan const &interval) {
    if (d->timer) return; // Already discovering.
    DENG2_ASSERT(!d->socket);
    d->socket = new QUdpSocket;
    connect(d->socket, SIGNAL(readyRead()), this, SLOT(readDiscoveryReply()));
    int tries = 10;
    forever {
        if (d->socket->bind(d->port + Rangeui16(1, 0x4000).random(), QUdpSocket::DontShareAddress)) {
            break;
        }
        if (!--tries) {
            throw PortError("Beacon::start", "Could not bind to UDP port " + String::number(d->port));
        }
    }
    d->found.clear();
    if (timeOut > 0.0) {
        d->discoveryEndsAt = Time() + timeOut;
    }
    else {
        d->discoveryEndsAt = Time::invalidTime();
    }
    d->timer = new QTimer;
    connect(d->timer, SIGNAL(timeout()), this, SLOT(continueDiscovery()));
    d->timer->start(interval.asMilliSeconds());
    continueDiscovery();
}
QList<Address> Beacon::foundHosts() const {
    return d->found.keys();
}
Block Beacon::messageFromHost(Address const &host) const {
    if (!d->found.contains(host)) return Block();
    return d->found[host];
}
void Beacon::readIncoming() {
    LOG_AS("Beacon");
    if (!d->socket) return;
    while (d->socket->hasPendingDatagrams()) {
        QHostAddress from;
        quint16 port = 0;
        Block block(d->socket->pendingDatagramSize());
        d->socket->readDatagram(reinterpret_cast<char *>(block.data()),
                                block.size(), &from, &port);
        LOG_NET_XVERBOSE("Received %i bytes from %s port %i", block.size() << from.toString() << port);
        if (block == discoveryMessage) {
            d->socket->writeDatagram(d->message, from, port);
        }
    }
}
void Beacon::readDiscoveryReply() {
    LOG_AS("Beacon");
    if (!d->socket) return;
    while (d->socket->hasPendingDatagrams()) {
        QHostAddress from;
        quint16 port = 0;
        Block block(d->socket->pendingDatagramSize());
        d->socket->readDatagram(reinterpret_cast<char *>(block.data()),
                                block.size(), &from, &port);
        if (block == discoveryMessage)
            continue;
        try {
            duint16 listenPort = 0;
            Reader(block) >> listenPort;
            block.remove(0, 2);
            block = block.decompressed();
            Address const host(from, listenPort);
            d->found.insert(host, block);
            emit found(host, block);
        }
        catch (Error const &) {
        }
    }
}
void Beacon::continueDiscovery() {
    DENG2_ASSERT(d->socket);
    DENG2_ASSERT(d->timer);
    if (d->discoveryEndsAt.isValid() && Time() > d->discoveryEndsAt) {
        d->timer->stop();
        emit finished();
        d->socket->deleteLater();
        d->socket = 0;
        d->timer->deleteLater();
        d->timer = 0;
        return;
    }
    Block block(discoveryMessage);
    LOG_NET_XVERBOSE("Broadcasting %i bytes", block.size());
    for (duint16 range = 0; range < MAX_LISTEN_RANGE; ++range) {
        d->socket->writeDatagram(block,
                                 QHostAddress::Broadcast,
                                 d->port + range);
    }
}
} // namespace de
namespace de {
static Packet::Type const BLOCK_PACKET_TYPE = Packet::typeFromString("BLCK");
BlockPacket::BlockPacket() : Packet(BLOCK_PACKET_TYPE)
{}
BlockPacket::BlockPacket(Block const &block)
    : Packet(BLOCK_PACKET_TYPE), Block(block)
{}
void BlockPacket::operator >> (Writer &to) const {
    Packet::operator >> (to);
    to << *static_cast<Block const *>(this);
}
void BlockPacket::operator << (Reader &from) {
    Packet::operator << (from);
    from >> *static_cast<Block *>(this);
}
Block &BlockPacket::block() {
    return *this;
}
Block const &BlockPacket::block() const {
    return *this;
}
Packet *BlockPacket::fromBlock(Block const &block) {
    return constructFromBlock<BlockPacket>(block, BLOCK_PACKET_TYPE);
}
} // namespace de
namespace de {
static IdentifiedPacket::Id idGen = 0;
IdentifiedPacket::IdentifiedPacket(Type const &type, Id i)
    : Packet(type), _id(i)
{}
void IdentifiedPacket::setId(Id id) {
    _id = id;
}
void IdentifiedPacket::operator >> (Writer &to) const {
    Packet::operator >> (to);
    to << id();
}
void IdentifiedPacket::operator << (Reader &from) {
    Packet::operator << (from);
    from >> _id;
}
IdentifiedPacket::Id IdentifiedPacket::id() const {
    if (!_id) {
        _id = ++idGen;
    }
    return _id;
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(ListenSocket) {
    QTcpServer *socket;
    duint16 port;
    QList<QTcpSocket *> incoming;
    Impl() : socket(0), port(0) {}
    ~Impl() {
        delete socket;
    }
};
ListenSocket::ListenSocket(duint16 port) : d(new Impl) {
    LOG_AS("ListenSocket");
    d->socket = new QTcpServer(this);
    d->port = port;
    if (!d->socket->listen(QHostAddress::Any, d->port)) {
        throw OpenError("ListenSocket", "Port " + QString::number(d->port) + ": " +
                        d->socket->errorString());
    }
    connect(d->socket, SIGNAL(newConnection()), this, SLOT(acceptNewConnection()));
}
void ListenSocket::acceptNewConnection() {
    LOG_AS("ListenSocket::acceptNewConnection");
    d->incoming << d->socket->nextPendingConnection();
    emit incomingConnection();
}
Socket *ListenSocket::accept() {
    if (d->incoming.empty()) {
        return 0;
    }
    QTcpSocket *s = d->incoming.takeFirst();
    LOG_NET_NOTE("Accepted new connection from %s") << Address(s->peerAddress()).asText();
    return new Socket(s);
}
duint16 ListenSocket::port() const {
    return d->port;
}
} // namespace de
namespace de {
Message::Message(IByteArray const &other) : Block(other), _channel(0)
{}
Message::Message(Address const &addr, Channel channel, Size initialSize)
    : Block(initialSize), _address(addr), _channel(channel)
{}
Message::Message(Address const &addr, Channel channel, IByteArray const &other)
    : Block(other), _address(addr), _channel(channel)
{}
Message::Message(Address const &addr, Channel channel, IByteArray const &other, Offset at, Size count)
    : Block(other, at, count), _address(addr), _channel(channel)
{}
} // namespace de
namespace de {
namespace internal {
DENG2_PIMPL(NetworkInterfaces), public Lockable {
    struct AddressQueryTask : public Task {
        NetworkInterfaces::Impl *d;
        AddressQueryTask(NetworkInterfaces::Impl *d) : d(d) {}
        void runTask() override {
            QList<QHostAddress> ipv6;
            foreach (QHostAddress addr, QNetworkInterface::allAddresses()) {
                ipv6 << QHostAddress(addr.toIPv6Address());
            } {
                DENG2_GUARD(d);
                d->addresses = ipv6;
                d->gotAddresses = true;
            }
        }
    };
    QList<QHostAddress> addresses; // lock before access
    TaskPool tasks;
    QTimer updateTimer;
    bool gotAddresses = false;
    Impl(Public *i) : Base(i) {
        updateTimer.setInterval(1000 * 60 * 1);
        updateTimer.setSingleShot(false);
        QObject::connect(&updateTimer, &QTimer::timeout, [this] () {
            tasks.start(new AddressQueryTask(this));
        });
        updateTimer.start();
        tasks.start(new AddressQueryTask(this));
    }
};
NetworkInterfaces::NetworkInterfaces()
    : d(new Impl(this))
{}
QList<QHostAddress> NetworkInterfaces::allAddresses() const {
    if (!d->gotAddresses) {
        d->tasks.waitForDone();
    }
    DENG2_GUARD(d);
    return d->addresses;
}
NetworkInterfaces &NetworkInterfaces::get() // static
{
    static NetworkInterfaces nif;
    return nif;
}
} // namespace internal
} // namespace de
namespace de {
Packet::Type Packet::typeFromString(char const *fourcc) {
    Type type;
    std::memcpy(type.data(), fourcc, 4);
    return type;
}
Packet::Packet(Type const &t) {
    setType(t);
}
void Packet::setType(Type const &t) {
    _type = t;
}
void Packet::operator >> (Writer &to) const {
    to.writeBytes(ByteRefArray(_type.data(), 4));
}
void Packet::operator << (Reader &from) {
    Type ident;
    ByteRefArray ref(ident.data(), 4);
    from.readBytesFixedSize(ref);
    if (_type != ident) {
        throw InvalidTypeError("Packet::operator <<", "Invalid ID");
    }
}
void Packet::execute() const
{}
bool Packet::checkType(Reader &from, Type const &type) {
    from.mark();
    Type ident;
    ByteRefArray ref(ident.data(), 4);
    from.readBytesFixedSize(ref);
    from.rewind();
    return type == ident;
}
} // namespace de
namespace de {
Protocol::Protocol() {
    define(RecordPacket::fromBlock);
}
Protocol::~Protocol()
{}
void Protocol::define(Constructor constructor) {
    _constructors.append(constructor);
}
Packet *Protocol::interpret(Block const &block) const {
    foreach (Constructor constructor, _constructors) {
        if (Packet *p = constructor(block)) {
            return p;
        }
    }
    return nullptr;
}
/*
void Protocol::syncCommand(Transmitter& to, const CommandPacket& command, RecordPacket** response) {
    LOG_AS("Protocol::syncCommand");
    LOG_DEBUG("Sending: '%s' with args:\n%s") << command.command() << command.arguments();
    to << command;
    QScopedPointer<RecordPacket> rep(to.receivePacket<RecordPacket>());
    if (rep->label() == "failure") {
        throw FailureError("Protocol::decree", "Command '" + command.command() +
            "' failed: " + rep->valueAsText("message"));
    }
    else if (rep->label() == "deny") {
        throw DenyError("Protocol::decree", "Command '" + command.command() +
            "' was denied: " + rep->valueAsText("message"));
    }
    LOG_DEBUG("Reply to the command '%s' was '%s':") << command.command() << rep->label();
    LOG_DEBUG("") << rep->record();
    if (response) {
        *response = rep.take();
    }
}
*/
void Protocol::reply(Transmitter &to, Reply type, Record *record) {
    String label;
    switch (type) {
    case OK:
        label = "ok";
        break;
    case FAILURE:
        label = "failure";
        break;
    case DENY:
        label = "deny";
        break;
    }
    RecordPacket packet(label);
    if (record) {
        packet.take(record);
    }
    to << packet;
}
void Protocol::reply(Transmitter &to, Reply type, String const &message) {
    Record* rec = new Record();
    if (!message.empty()) {
        rec->addText("message", message);
    }
    reply(to, type, rec);
}
} // namespace de
namespace de {
static Packet::Type const RECORD_PACKET_TYPE = Packet::typeFromString("RECO");
RecordPacket::RecordPacket(String const &name, Id i)
    : IdentifiedPacket(RECORD_PACKET_TYPE, i), _name(name), _record(0) {
    _record = new Record();
}
RecordPacket::~RecordPacket() {
    delete _record;
}
void RecordPacket::take(Record *newRecord) {
    delete _record;
    _record = newRecord;
}
Record *RecordPacket::give() {
    Record *detached = _record;
    _record = new Record;
    return detached;
}
Variable const &RecordPacket::operator [] (String const &variableName) const {
    return (*_record)[variableName];
}
String RecordPacket::valueAsText(String const &variableName) const {
    return (*_record)[variableName].value().asText();
}
void RecordPacket::operator >> (Writer &to) const {
    IdentifiedPacket::operator >> (to);
    to << _name << *_record;
}
void RecordPacket::operator << (Reader &from) {
    IdentifiedPacket::operator << (from);
    from >> _name >> *_record;
}
Packet *RecordPacket::fromBlock(Block const &block) {
    return constructFromBlock<RecordPacket>(block, RECORD_PACKET_TYPE);
}
} // namespace de
namespace de {
struct Counters {
    duint64 sentUncompressedBytes = 0;
    duint64 sentBytes = 0;
    duint64 sentPeriodBytes = 0;
    double outputBytesPerSecond = 0;
    Time periodStartedAt;
};
static LockableT<Counters> counters;
static TimeSpan const sendPeriodDuration = 5;
static duint const MAX_CHANNELS = 2;
static int const MAX_SIZE_SMALL  = 127; // bytes
static int const MAX_SIZE_MEDIUM = 4095; // bytes
static int const MAX_SIZE_BIG    = 10*MAX_SIZE_MEDIUM;
static int const MAX_SIZE_LARGE  = DENG2_SOCKET_MAX_PAYLOAD_SIZE;
static int const MAX_HUFFMAN_INPUT_SIZE = 4096; // bytes
#define TRMF_CONTINUE           0x80
#define TRMF_DEFLATED           0x40
#define TRMF_SIZE_MASK          0x7f
#define TRMF_SIZE_MASK_MEDIUM   0x3f
#define TRMF_SIZE_SHIFT         7
namespace internal {
/**
 * Network message header.
 */
struct MessageHeader : public ISerializable {
    int size;
    bool isHuffmanCoded;
    bool isDeflated;
    duint channel; /// @todo include in the written header
    MessageHeader() : size(0), isHuffmanCoded(false), isDeflated(false), channel(0)
    {}
    void operator >> (Writer &writer) const {
        if (size <= MAX_SIZE_SMALL && !isDeflated) {
            writer << dbyte(size);
        }
        else if (size <= MAX_SIZE_MEDIUM) {
            writer << dbyte(TRMF_CONTINUE | (size & TRMF_SIZE_MASK));
            writer << dbyte((isDeflated? TRMF_DEFLATED : 0) | (size >> TRMF_SIZE_SHIFT));
        }
        else if (size <= MAX_SIZE_LARGE) {
            DENG2_ASSERT(isDeflated);
            writer << dbyte(TRMF_CONTINUE | (size & TRMF_SIZE_MASK));
            writer << dbyte(TRMF_CONTINUE | ((size >> TRMF_SIZE_SHIFT) & TRMF_SIZE_MASK));
            writer << dbyte(size >> (2*TRMF_SIZE_SHIFT));
        }
        else {
            DENG2_ASSERT(false);
        }
    }
    /**
    * Throws an exception if the header is malformed/incomplete.
    */
    void operator << (Reader &reader) {
        dbyte b;
        reader >> b;
        size = b & TRMF_SIZE_MASK;
        isDeflated = false;
        isHuffmanCoded = true;
        if (b & TRMF_CONTINUE) // More follows...
        {
            reader >> b;
            if (b & TRMF_CONTINUE) // Yet more to come...
            {
                isDeflated = true;
                isHuffmanCoded = false;
                size |= ((b & TRMF_SIZE_MASK) << TRMF_SIZE_SHIFT);
                reader >> b;
                size |= (b << (2*TRMF_SIZE_SHIFT));
            }
            else {
                if (b & TRMF_DEFLATED) {
                    isDeflated = true;
                    isHuffmanCoded = false;
                }
                size |= ((b & TRMF_SIZE_MASK_MEDIUM) << TRMF_SIZE_SHIFT);
            }
        }
    }
};
} // namespace internal
using namespace internal;
DENG2_PIMPL_NOREF(Socket) {
    Address peer;
    bool quiet = false;
    bool retainOrder = true;
    enum ReceptionState { ReceivingHeader, ReceivingPayload };
    ReceptionState receptionState = ReceivingHeader;
    Block receivedBytes;
    MessageHeader incomingHeader;
    duint activeChannel = 0;
    QTcpSocket *socket = nullptr;
    QList<Message *> receivedMessages;
    dint64 bytesToBeWritten = 0;
    dint64 totalBytesWritten = 0;
    ~Impl() {
        foreach (Message *msg, receivedMessages) delete msg;
    }
    void serializeMessage(MessageHeader &header, Block &payload) {
        Block huffData;
        if (payload.size() <= MAX_HUFFMAN_INPUT_SIZE) // Potentially short enough.
        {
            huffData = codec::huffmanEncode(payload);
            if (int(huffData.size()) <= MAX_SIZE_SMALL) {
                header.isHuffmanCoded = true;
                header.size = huffData.size();
                payload = huffData;
            }
        }
        if (!header.size) // Try deflate.
        {
            int const level = 1; //(payload.size() < MAX_SIZE_BIG? 1 /*fast*/ : 9 /*best*/);
            Block const deflated = payload.compressed(level);
            if (!deflated.size()) {
                throw ProtocolError("Socket::send:", "Failed to deflate message payload");
            }
            if (deflated.size() > MAX_SIZE_LARGE) {
                throw ProtocolError("Socket::send",
                                    QString("Compressed payload is too large (%1 bytes)").arg(deflated.size()));
            }
            if (huffData.size() && huffData.size() <= deflated.size() && int(huffData.size()) <= MAX_SIZE_MEDIUM) {
                header.isHuffmanCoded = true;
                header.size = huffData.size();
                payload = huffData;
            }
            else {
                header.isDeflated = true;
                header.size = deflated.size();
                payload = deflated;
            }
        }
    }
    void sendMessage(MessageHeader const &header,
                     Block const &payload) {
        DENG2_ASSERT(socket != nullptr);
        DENG2_ASSERT(QThread::currentThread() == socket->thread());
        Block dest;
        Writer(dest) << header;
        socket->write(dest);
        dsize const total = dest.size() + payload.size();
        bytesToBeWritten  += total;
        totalBytesWritten += total;
        socket->write(payload); {
            DENG2_GUARD(counters);
            counters.value.sentPeriodBytes += total;
            counters.value.sentBytes       += total;
            if (!counters.value.periodStartedAt.isValid()
                || counters.value.periodStartedAt.since() > sendPeriodDuration) {
                counters.value.outputBytesPerSecond = double(counters.value.sentPeriodBytes)
                                                    / sendPeriodDuration;
                counters.value.sentPeriodBytes = 0;
                counters.value.periodStartedAt = Time::currentHighPerformanceTime();
            }
        }
    }
    void serializeAndSendMessage(IByteArray const &packet) {
        Block payload = packet; {
            DENG2_GUARD(counters);
            counters.value.sentUncompressedBytes += payload.size();
        }
        if (!retainOrder && packet.size() >= MAX_SIZE_BIG) {
            async([this, payload] () {
                MessageHeader header;
                Block msgData = payload;
                serializeMessage(header, msgData);
                return std::make_pair(header, msgData);
            },
            [this] (std::pair<MessageHeader, Block> msg) {
                if (socket) {
                    sendMessage(msg.first, msg.second);
                }
            });
        }
        else {
            MessageHeader header;
            serializeMessage(header, payload);
            sendMessage(header, payload);
        }
    }
    /**
     * Checks the incoming bytes and sees if any messages can be formed.
     */
    void deserializeMessages() {
        forever {
            if (receptionState == ReceivingHeader) {
                if (receivedBytes.size() < 2) {
                    return;
                }
                try {
                    Reader reader(receivedBytes);
                    reader >> incomingHeader;
                    receptionState = ReceivingPayload;
                    receivedBytes.remove(0, reader.offset());
                }
                catch (de::Error const &) {
                    return;
                }
            }
            if (receptionState == ReceivingPayload) {
                if (int(receivedBytes.size()) >= incomingHeader.size) {
                    Block payload = receivedBytes.left(incomingHeader.size);
                    receivedBytes.remove(0, incomingHeader.size);
                    if (incomingHeader.isHuffmanCoded) {
                        payload = codec::huffmanDecode(payload);
                        if (!payload.size()) {
                            throw ProtocolError("Socket::Impl::deserializeMessages", "Huffman decoding failed");
                        }
                    }
                    else if (incomingHeader.isDeflated) {
                        payload = qUncompress(payload);
                        if (!payload.size()) {
                            throw ProtocolError("Socket::Impl::deserializeMessages", "Deflate failed");
                        }
                    }
                    receivedMessages << new Message(Address(socket->peerAddress(), socket->peerPort()),
                                                    incomingHeader.channel, payload);
                    receptionState = ReceivingHeader;
                    incomingHeader = MessageHeader();
                }
                else {
                    return;
                }
            }
        }
    }
};
Socket::Socket() : d(new Impl) {
    d->socket = new QTcpSocket;
    initialize();
    QObject::connect(d->socket, SIGNAL(connected()), this, SIGNAL(connected()));
}
Socket::Socket(Address const &address, TimeSpan const &timeOut) : d(new Impl) // blocking
{
    LOG_AS("Socket");
    d->socket = new QTcpSocket;
    initialize();
    d->socket->connectToHost(address.host(), address.port());
    if (!d->socket->waitForConnected(int(timeOut.asMilliSeconds()))) {
        QString msg = d->socket->errorString();
        delete d->socket;
        d.reset();
        throw ConnectionError("Socket", "Opening the connection to " + address.asText() + " failed: " + msg);
    }
    LOG_NET_NOTE("Connection opened to %s") << address.asText();
    d->peer = address;
    DENG2_ASSERT(d->socket->isOpen() && d->socket->isWritable() &&
                 d->socket->state() == QAbstractSocket::ConnectedState);
}
void Socket::open(Address const &address) // non-blocking
{
    DENG2_ASSERT(d->socket);
    DENG2_ASSERT(d->socket->state() == QAbstractSocket::UnconnectedState);
    LOG_AS("Socket");
    if (!d->quiet) LOG_NET_MSG("Opening connection to %s") << address.asText();
    d->socket->connectToHost(address.host(), address.port());
    d->peer = address;
}
void Socket::open(String const &domainNameWithOptionalPort,
                  duint16 defaultPort) // non-blocking
{
    String str = domainNameWithOptionalPort;
    duint16 port = defaultPort;
    if (str.contains(':')) {
        int pos = str.lastIndexOf(':');
        port = duint16(str.mid(pos + 1).toInt());
        if (!port) port = defaultPort;
        str = str.left(pos);
    }
    if (str == "localhost") {
        open(Address(str.toLatin1(), port));
        return;
    }
    QHostAddress host(str);
    if (!host.isNull()) {
        open(Address(str.toLatin1(), port));
        return;
    }
    d->peer.setPort(port);
    QHostInfo::lookupHost(str, this, SLOT(hostResolved(QHostInfo)));
}
void Socket::reconnect() {
    DENG2_ASSERT(!isOpen());
    open(d->peer);
}
Socket::Socket(QTcpSocket *existingSocket) : d(new Impl) {
    d->socket = existingSocket;
    initialize();
    readIncomingBytes();
}
Socket::~Socket() {
    close();
    delete d->socket;
}
void Socket::socketDestroyed() {
    d->socket = 0;
}
void Socket::initialize() {
    d->socket->setSocketOption(QTcpSocket::LowDelayOption, 1); // prefer short buffering
    QObject::connect(d->socket, SIGNAL(bytesWritten(qint64)), this, SLOT(bytesWereWritten(qint64)));
    QObject::connect(d->socket, SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    QObject::connect(d->socket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(socketError(QAbstractSocket::SocketError)), Qt::DirectConnection);
    QObject::connect(d->socket, SIGNAL(readyRead()), this, SLOT(readIncomingBytes()));
    QObject::connect(d->socket, SIGNAL(destroyed()), this, SLOT(socketDestroyed()));
    /*QObject::connect(d->socket, &QAbstractSocket::stateChanged, [] (QAbstractSocket::SocketState state)
    {
        qDebug() << "socket state:" << state;
    });*/
}
void Socket::close() {
    if (!d->socket) return;
    if (d->socket->state() == QAbstractSocket::ConnectedState) {
        d->socket->disconnectFromHost();
    }
    else {
        d->socket->abort();
    }
    if (d->socket->state() != QAbstractSocket::UnconnectedState) {
        d->socket->waitForDisconnected();
    }
    d->socket->close();
}
void Socket::setQuiet(bool noLogOutput) {
    d->quiet = noLogOutput;
}
void Socket::resetCounters() {
    DENG2_GUARD(counters);
    counters.value = Counters();
}
duint64 Socket::sentUncompressedBytes() {
    DENG2_GUARD(counters);
    return counters.value.sentUncompressedBytes;
}
duint64 Socket::sentBytes() {
    DENG2_GUARD(counters);
    return counters.value.sentBytes;
}
double Socket::outputBytesPerSecond() {
    DENG2_GUARD(counters);
    return counters.value.outputBytesPerSecond;
}
duint Socket::channel() const {
    return d->activeChannel;
}
void Socket::setChannel(duint number) {
    DENG2_ASSERT(number < MAX_CHANNELS);
    d->activeChannel = min(number, MAX_CHANNELS - 1);
}
void Socket::setRetainOrder(bool retainOrder) {
    d->retainOrder = retainOrder;
}
void Socket::send(IByteArray const &packet) {
    send(packet, d->activeChannel);
}
Socket &Socket::operator << (IByteArray const &packet) {
    send(packet, d->activeChannel);
    return *this;
}
void Socket::send(IByteArray const &packet, duint /*channel*/)
{
    if (!d->socket) {
        throw DisconnectedError("Socket::send", "Socket is unavailable");
    }
    DENG2_ASSERT(thread() == QThread::currentThread());
    d->serializeAndSendMessage(packet);
}
void Socket::readIncomingBytes() {
    if (!d->socket) return;
    auto available = d->socket->bytesAvailable();
    if (available > 0) {
        d->receivedBytes += d->socket->read(d->socket->bytesAvailable());
    }
    d->deserializeMessages();
    if (!d->receivedMessages.isEmpty()) {
        emit messagesReady();
    }
}
void Socket::hostResolved(QHostInfo const &info) {
    if (info.error() != QHostInfo::NoError || info.addresses().isEmpty()) {
        LOG_NET_ERROR("Could not resolve host: ") << info.errorString();
        emit disconnected();
    }
    else {
        open(Address(info.addresses().first(), d->peer.port()));
        emit addressResolved();
    }
}
Message *Socket::receive() {
    if (d->receivedMessages.isEmpty()) {
        return nullptr;
    }
    return d->receivedMessages.takeFirst();
}
Message *Socket::peek() {
    if (d->receivedMessages.isEmpty()) {
        return nullptr;
    }
    return d->receivedMessages.first();
}
void Socket::flush() {
    if (!d->socket) return;
    d->socket->flush();
    d->socket->waitForBytesWritten();
}
Address Socket::peerAddress() const {
    if (isOpen() && d->socket->state() == QTcpSocket::ConnectedState) {
        return Address(d->socket->peerAddress(), d->socket->peerPort());
    }
    return d->peer;
}
bool Socket::isOpen() const {
    return d->socket && d->socket->state() != QTcpSocket::UnconnectedState;
}
bool Socket::isLocal() const {
    return peerAddress().isLocal();
}
void Socket::socketDisconnected() {
    emit disconnected();
}
void Socket::socketError(QAbstractSocket::SocketError socketError) {
    if (socketError != QAbstractSocket::SocketTimeoutError) {
        LOG_AS("Socket");
        if (!d->quiet) LOG_NET_WARNING(d->socket->errorString());
        emit error(d->socket->errorString());
        emit disconnected();
    }
}
bool Socket::hasIncoming() const {
    return !d->receivedMessages.empty();
}
dsize Socket::bytesBuffered() const {
    return dsize(d->bytesToBeWritten);
}
void Socket::bytesWereWritten(qint64 bytes) {
    d->bytesToBeWritten -= bytes;
    DENG2_ASSERT(d->bytesToBeWritten >= 0);
    if (d->bytesToBeWritten == 0) {
        emit allSent();
    }
}
} // namespace de
namespace de {
Transmitter::~Transmitter()
{}
IOStream &Transmitter::operator << (IByteArray const &data) {
    send(data);
    return *this;
}
Transmitter &Transmitter::operator << (Packet const &packet) {
    sendPacket(packet);
    return *this;
}
void Transmitter::sendPacket(Packet const &packet) {
    Block data;
    Writer(data) << packet;
    send(data);
}
} // namespace de
