﻿/** @file widgets/action.cpp  Abstract base class for UI actions.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Action"
/** @file animation.h Animation function.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Animation"
#include "de/Writer"
#include "de/Reader"
#include "de/math.h"
#include <atomic>
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/AnimationRule"
#include "de/Clock"
#include <QDebug>
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ConstantRule"
#include "de/math.h"
/** @file indirectrule.cpp  Indirect rule.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/IndirectRule"
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/OperatorRule"
#include "de/math.h"
/** @file rootwidget.cpp Widget for managing the root of the UI.
 * @ingroup widget
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RootWidget"
#include "de/ConstantRule"
#include "de/RuleRectangle"
#include "de/math.h"
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Rule"
#include "de/math.h"
#include "de/PointerSet"
/** @file rulebank.cpp  Bank of Rules.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RuleBank"
#include <de/ConstantRule>
#include <de/OperatorRule>
#include <de/ScriptedInfo>
/*
 * The Doomsday Engine Project
 *
 * Copyright © 2011-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RuleRectangle"
#include "de/App"
#include "de/math.h"
/** @file widget.cpp Base class for widgets.
 * @ingroup widget
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Widget"
#include "de/RootWidget"
#include "de/Asset"
#include "de/LogBuffer"
#include <QList>
#include <QMap>

namespace de {
DENG2_PIMPL_NOREF(Action) {
    DENG2_PIMPL_AUDIENCE(Triggered)
};
DENG2_AUDIENCE_METHOD(Action, Triggered)
Action::Action() : d(new Impl)
{}
Action::~Action()
{}
void Action::trigger() {
    DENG2_FOR_AUDIENCE2(Triggered, i) {
        i->actionTriggered(*this);
    }
}
} // namespace de
namespace de {
static float const DEFAULT_SPRING = 3.f;
namespace internal {
static inline float easeOut(TimeSpan t) {
    return t * (2 - t);
}
static inline float easeOutSofter(TimeSpan t) {
    const float a = -std::pow(t - 1.0, 4) + 1;
    return (easeOut(t) + a) / 2.0f;
}
static inline float easeIn(TimeSpan t) {
    return t * t;
}
static inline float easeBoth(TimeSpan t) {
    if (t < .5) {
        return easeIn(t * 2.0) / 2.0;
    }
    else {
        return .5 + easeOut((t - .5) * 2.0) / 2.0;
    }
}
enum AnimationFlag {
    Paused = 0x1,
    Finished = 0x2,
};
Q_DECLARE_FLAGS(AnimationFlags, AnimationFlag)
Q_DECLARE_OPERATORS_FOR_FLAGS(AnimationFlags)
struct AnimationTime : DENG2_OBSERVES(Clock, TimeChange) {
    double now;
    void timeChanged(Clock const &clock) override {
        now = clock.time().highPerformanceTime();
    }
};
static AnimationTime theTime;
} // namespace internal
using namespace internal;
Clock const *Animation::_clock = 0;
DENG2_PIMPL_NOREF(Animation) {
    float value;
    float target;
    TimeSpan startDelay;
    TimeSpan setTime;
    TimeSpan targetTime;
    TimeSpan pauseTime;
    Style style;
    float spring;
    mutable AnimationFlags flags;
    Impl(float val, Style s)
        : value(val)
        , target(val)
        , startDelay(0)
        , setTime   (theTime.now)
        , targetTime(theTime.now)
        , style(s)
        , spring(DEFAULT_SPRING)
    {}
    /**
     * Calculates the value of the animation.
     * @param now  Point of time at which to evaluate.
     * @return Value of the animation when the frame time is @a now.
     */
    float valueAt(TimeSpan const &now) const {
        TimeSpan span = targetTime - setTime;
        float s2 = 0;
        TimeSpan peak = 0;
        TimeSpan peak2 = 0;
        if (style == Bounce || style == FixedBounce) {
            s2 = spring * spring;
            peak = 1.f/3;
            peak2 = 2.f/3;
        }
        if (now >= targetTime || span <= 0) {
            flags |= Finished;
            return target;
        }
        else {
            span -= startDelay;
            TimeSpan const elapsed = now - setTime - startDelay;
            TimeSpan const t = clamp(0.0, elapsed/span, 1.0);
            float const delta = target - value;
            switch (style) {
            case EaseOut:
                return value + easeOut(t) * delta;
            case EaseOutSofter:
                return value + easeOutSofter(t) * delta;
            case EaseIn:
                return value + easeIn(t) * delta;
            case EaseBoth:
                return value + easeBoth(t) * delta;
            case Bounce:
            case FixedBounce: {
                const float bounce1 = (style == Bounce? delta/spring : (delta>=0? spring : -spring));
                const float bounce2 = (style == Bounce? delta/s2 : (delta>=0? spring/2 : -spring/2));
                float peakDelta = (delta + bounce1);
                if (t < peak) {
                    return value + easeOut(t/peak) * peakDelta;
                }
                else if (t < peak2) {
                    return (value + peakDelta) - easeBoth((t - peak)/(peak2 - peak)) * (bounce1 + bounce2);
                }
                else {
                    return (target - bounce2) + easeBoth((t - peak2)/(1 - peak2)) * bounce2;
                }
                break;
            }
            default:
                return value + t * delta;
            }
        }
        return target;
    }
    void checkDone() {
        if (!(flags & Finished) && currentTime() >= targetTime) {
            flags |= Finished;
        }
    }
    double currentTime() const {
        if (flags & Paused) return pauseTime;
        return theTime.now; //Animation::currentTime();
    }
};
Animation::Animation(float val, Style s) : d(new Impl(val, s))
{}
Animation::Animation(Animation const &other)
    : de::ISerializable()
    , d(new Impl(*other.d))
{}
Animation &Animation::operator = (Animation const &other) {
    d.reset(new Impl(*other.d));
    return *this;
}
void Animation::setStyle(Animation::Style s) {
    d->style = s;
}
void Animation::setStyle(Style style, float bounce) {
    d->style = style;
    d->spring = bounce;
    if (!d->spring) {
        d->spring = DEFAULT_SPRING;
    }
}
Animation::Style Animation::style() const {
    return d->style;
}
float Animation::bounce() const {
    return d->spring;
}
void Animation::setValue(float v, TimeSpan transitionSpan, TimeSpan startDelay) {
    resume();
    TimeSpan const now = d->currentTime();
    if (transitionSpan <= 0) {
        d->value = d->target = v;
        d->setTime = d->targetTime = now;
        d->flags |= Finished;
    }
    else {
        d->value = d->valueAt(now);
        d->target = v;
        d->setTime = now;
        d->targetTime = d->setTime + transitionSpan;
        d->flags &= ~Finished;
    }
    d->startDelay = startDelay;
}
void Animation::setValue(int v, TimeSpan transitionSpan, TimeSpan startDelay) {
    setValue(float(v), transitionSpan, startDelay);
}
void Animation::setValueFrom(float fromValue, float toValue, TimeSpan transitionSpan, TimeSpan startDelay) {
    setValue(fromValue);
    setValue(toValue, transitionSpan, startDelay);
}
float Animation::value() const {
    if (d->flags & Paused) {
        return d->valueAt(d->pauseTime);
    }
    if (d->flags & Finished) {
        return d->target;
    }
    return d->valueAt(theTime.now);
}
bool Animation::done() const {
    d->checkDone();
    if (d->flags & Finished) return true;
    return false;
}
float Animation::target() const {
    return d->target;
}
void Animation::adjustTarget(float newTarget) {
    d->target = newTarget;
}
TimeSpan Animation::remainingTime() const {
    TimeSpan const now = d->currentTime();
    if (now >= d->targetTime) {
        return 0;
    }
    return d->targetTime - now;
}
TimeSpan Animation::transitionTime() const {
    return d->targetTime - d->setTime;
}
void Animation::shift(float valueDelta) {
    d->value  += valueDelta;
    d->target += valueDelta;
}
void Animation::pause() {
    if (d->flags.testFlag(Paused) || done()) return;
    d->pauseTime = d->currentTime();
    d->flags |= Paused;
}
void Animation::resume() {
    if (!d->flags.testFlag(Paused)) return;
    d->flags &= ~Paused;
    TimeSpan const delta = d->currentTime() - d->pauseTime;
    d->setTime    += delta;
    d->targetTime += delta;
}
void Animation::finish() {
    setValue(d->target);
}
String Animation::asText() const {
    return String("Animation(%1 -> %2, ETA:%3 s; curr: %4)").arg(d->value).arg(d->target).arg(remainingTime()).arg(value());
}
Clock const &Animation::clock() {
    DENG2_ASSERT(_clock != 0);
    if (!_clock) {
        throw ClockMissingError("Animation::clock", "Animation has no clock");
    }
    return *_clock;
}
void Animation::operator >> (Writer &to) const {
    TimeSpan const now = currentTime();
    to << d->value << d->target;
    to << (d->setTime - now) << (d->targetTime - now);
    to << d->startDelay;
    to << dint32(d->style) << d->spring;
}
void Animation::operator << (Reader &from) {
    TimeSpan const now = currentTime();
    from >> d->value >> d->target;
    TimeSpan relSet, relTarget;
    from >> relSet >> relTarget;
    d->setTime    = now + relSet;
    d->targetTime = now + relTarget;
    from >> d->startDelay;
    dint32 s;
    from >> s;
    d->style = Style(s);
    from >> d->spring;
}
void Animation::setClock(Clock const *clock) {
    if (_clock) _clock->audienceForPriorityTimeChange -= theTime;
    _clock = clock;
    if (clock) clock->audienceForPriorityTimeChange += theTime;
}
TimeSpan Animation::currentTime() // static
{
    DENG2_ASSERT(_clock != 0);
    if (!_clock) {
        throw ClockMissingError("Animation::clock", "Animation has no clock");
    }
    return TimeSpan(theTime.now);
}
Animation Animation::range(Style style, float from, float to, TimeSpan span, TimeSpan delay) {
    Animation anim(from, style);
    anim.setValue(to, span, delay);
    return anim;
}
} // namespace de
namespace de {
AnimationRule::AnimationRule(float initialValue, Animation::Style style)
    : Rule(initialValue)
    , _animation(initialValue, style)
    , _targetRule(0)
    , _behavior(Singleshot)
{}
AnimationRule::AnimationRule(Rule const &target, TimeSpan transition, Animation::Style style)
    : Rule(target.value())
    , _animation(target.value(), style)
    , _targetRule(0)
    , _behavior(RestartWhenTargetChanges | DontAnimateFromZero) {
    set(target, transition);
}
AnimationRule::~AnimationRule() {
    independentOf(_targetRule);
}
void AnimationRule::set(float target, TimeSpan transition, TimeSpan delay) {
    independentOf(_targetRule);
    _targetRule = 0;
    _animation.clock().audienceForPriorityTimeChange += this;
    _animation.setValue(target, transition, delay);
    invalidate();
}
void AnimationRule::set(Rule const &target, TimeSpan transition, TimeSpan delay) {
    set(target.value(), transition, delay);
    _targetRule = &target;
    dependsOn(_targetRule);
}
void AnimationRule::setStyle(Animation::Style style) {
    _animation.setStyle(style);
}
void AnimationRule::setStyle(Animation::Style style, float bounceSpring) {
    _animation.setStyle(style, bounceSpring);
}
void AnimationRule::setBehavior(Behaviors behavior) {
    _behavior = behavior;
}
AnimationRule::Behaviors AnimationRule::behavior() const {
    return _behavior;
}
void AnimationRule::shift(float delta) {
    _animation.shift(delta);
    invalidate();
}
void AnimationRule::finish() {
    _animation.finish();
}
void AnimationRule::pause() {
    _animation.pause();
}
void AnimationRule::resume() {
    _animation.resume();
}
String AnimationRule::description() const {
    DENG2_ASSERT(!isValid() || fequal(value(), _animation));
    String desc = _animation.asText();
    if (_targetRule) {
        desc += "=>" + _targetRule->description();
    }
    return desc;
}
void AnimationRule::update() {
    if (_targetRule) {
        if (_behavior.testFlag(Singleshot) || !_animation.done()) {
            _animation.adjustTarget(_targetRule->value());
        }
        else {
            if (!fequal(_animation.target(), _targetRule->value())) {
                TimeSpan span = _animation.transitionTime();
                if (_behavior.testFlag(DontAnimateFromZero) && fequal(_animation.target(), 0)) {
                    span = 0;
                }
                _animation.setValue(_targetRule->value(), span);
                _animation.clock().audienceForPriorityTimeChange += this;
            }
        }
    }
    setValue(_animation);
    if (_animation.done()) {
        _animation.clock().audienceForPriorityTimeChange -= this;
    }
}
void AnimationRule::timeChanged(const Clock &) {
    invalidate();
}
} // namespace de
namespace de {
ConstantRule::ConstantRule() : Rule(), _pendingValue(0) {
}
ConstantRule::ConstantRule(float constantValue)
    : Rule(constantValue), _pendingValue(constantValue)
{}
void ConstantRule::set(float newValue) {
    if (!fequal(_pendingValue, newValue)) {
        _pendingValue = newValue;
        invalidate();
    }
}
String ConstantRule::description() const {
    return String::format("%g", cachedValue());
}
ConstantRule const &ConstantRule::zero() {
    static ConstantRule *zeroRule = nullptr;
    if (!zeroRule) {
        zeroRule = new ConstantRule(0); // won't be deleted ever
#ifdef DENG2_DEBUG
        Counted::totalCount--;
#endif
    }
    return *zeroRule;
}
void ConstantRule::update() {
    setValue(_pendingValue);
}
} // namespace de
namespace de {
IndirectRule::IndirectRule() : _source(nullptr)
{}
IndirectRule::~IndirectRule() {
    independentOf(_source);
}
void IndirectRule::setSource(Rule const &rule) {
    unsetSource();
    dependsOn(_source = &rule);
    invalidate();
}
void IndirectRule::unsetSource() {
    independentOf(_source);
    _source = nullptr;
    invalidate();
}
bool IndirectRule::hasSource() const {
    return _source != nullptr;
}
void IndirectRule::update() {
    setValue(_source ? _source->value() : 0.f);
}
Rule const &IndirectRule::source() const {
    DENG2_ASSERT(_source);
    return *_source;
}
String IndirectRule::description() const {
    if (_source) {
        return source().description();
    }
    else {
        return String("[0]");
    }
}
} // namespace de
namespace de {
OperatorRule::OperatorRule(Operator op, Rule const &unary)
    : Rule()
    , _operator(op)
    , _leftOperand(&unary)
    , _rightOperand(nullptr)
    , _condition(nullptr) {
    DENG2_ASSERT(_leftOperand != 0);
    dependsOn(_leftOperand);
}
OperatorRule::OperatorRule(Operator op, Rule const &left, Rule const &right)
    : Rule()
    , _operator(op)
    , _leftOperand(&left)
    , _rightOperand(&right)
    , _condition(nullptr) {
    dependsOn(_leftOperand);
    if (_rightOperand != _leftOperand) dependsOn(_rightOperand);
}
OperatorRule::OperatorRule(OperatorRule::Operator op, Rule const &left, Rule const &right, Rule const &condition)
    : Rule()
    , _operator(op)
    , _leftOperand(&left)
    , _rightOperand(&right)
    , _condition(&condition) {
    DENG2_ASSERT(_leftOperand != _rightOperand);
    DENG2_ASSERT(_condition != _leftOperand);
    DENG2_ASSERT(_condition != _rightOperand);
    dependsOn(_leftOperand);
    dependsOn(_rightOperand);
    dependsOn(_condition);
}
OperatorRule::~OperatorRule() {
    independentOf(_leftOperand);
    if (_rightOperand != _leftOperand) independentOf(_rightOperand);
    independentOf(_condition);
}
void OperatorRule::update() {
    float leftValue = 0;
    float rightValue = 0;
    if (_operator == Select) {
        if (_condition->value() < 0) {
            leftValue = _leftOperand->value();
        }
        else {
            rightValue = _rightOperand->value();
        }
    }
    else {
        leftValue  = (_leftOperand?  _leftOperand->value()  : 0);
        rightValue = (_rightOperand? _rightOperand->value() : 0);
    }
    float v;
    switch (_operator) {
    case Equals:
        v = leftValue;
        break;
    case Negate:
        v = -leftValue;
        break;
    case Half:
        v = leftValue / 2;
        break;
    case Double:
        v = leftValue * 2;
        break;
    case Sum:
        v = leftValue + rightValue;
        break;
    case Subtract:
        v = leftValue - rightValue;
        break;
    case Multiply:
        v = leftValue * rightValue;
        break;
    case Divide:
        v = leftValue / rightValue;
        break;
    case Maximum:
        v = de::max(leftValue, rightValue);
        break;
    case Minimum:
        v = de::min(leftValue, rightValue);
        break;
    case Floor:
        v = de::floor(leftValue);
        break;
    case Select:
        v = (_condition->value() < 0? leftValue : rightValue);
        break;
    default:
        v = leftValue;
        break;
    }
    setValue(v);
}
String OperatorRule::description() const {
    static char const *texts[] = {
        "Equals",
        "Negate",
        "1/2x",
        "2x",
        "+",
        "-",
        "*",
        "/",
        "Max",
        "Min",
        "Floor"
    };
    String desc = "{";
    if (_leftOperand) {
        desc += " " + _leftOperand->description();
    }
    desc += String(" %1").arg(texts[_operator]);
    if (_rightOperand) {
        desc += " " + _rightOperand->description();
    }
    return desc + " }";
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(RootWidget) {
    RuleRectangle *viewRect;
    SafeWidgetPtr<Widget> focus;
    Impl() : focus(0) {
        viewRect = new RuleRectangle;
        viewRect->setLeftTop    (Const(0), Const(0))
                 .setRightBottom(Const(0), Const(0));
    }
    ~Impl() {
        delete viewRect;
    }
    Size viewSize() const {
        return Size(de::max(0, viewRect->width().valuei()),
                    de::max(0, viewRect->height().valuei()));
    }
    DENG2_PIMPL_AUDIENCE(FocusChange)
};
DENG2_AUDIENCE_METHOD(RootWidget, FocusChange)
RootWidget::RootWidget() : Widget(), d(new Impl)
{}
RootWidget::Size RootWidget::viewSize() const {
    return d->viewSize();
}
RuleRectangle const &RootWidget::viewRule() const {
    return *d->viewRect;
}
Rule const &RootWidget::viewLeft() const {
    return d->viewRect->left();
}
Rule const &RootWidget::viewRight() const {
    return d->viewRect->right();
}
Rule const &RootWidget::viewTop() const {
    return d->viewRect->top();
}
Rule const &RootWidget::viewBottom() const {
    return d->viewRect->bottom();
}
Rule const &RootWidget::viewWidth() const {
    return d->viewRect->width();
}
Rule const &RootWidget::viewHeight() const {
    return d->viewRect->height();
}
void RootWidget::setViewSize(Size const &size) {
#if defined (DENG_MOBILE)
    DENG2_GUARD(this);
#endif
    d->viewRect->setInput(Rule::Right,  Constu(size.x));
    d->viewRect->setInput(Rule::Bottom, Constu(size.y));
    notifyTree(&Widget::viewResized);
}
void RootWidget::setFocus(Widget *widget) {
    if (widget == d->focus) return; // No change.
    Widget *oldFocus = d->focus;
    d->focus.reset();
    if (oldFocus) oldFocus->focusLost();
    if (widget && widget->behavior().testFlag(Focusable)) {
        d->focus.reset(widget);
        if (d->focus) {
            d->focus->focusGained();
        }
    }
    if (d->focus != oldFocus) {
        DENG2_FOR_AUDIENCE2(FocusChange, i) {
            i->focusedWidgetChanged(widget);
        }
    }
}
Widget *RootWidget::focus() const {
    return d->focus;
}
void RootWidget::initialize() {
#if defined (DENG_MOBILE)
    DENG2_GUARD(this);
#endif
    notifyTree(&Widget::initialize);
}
void RootWidget::update() {
#if defined (DENG_MOBILE)
    DENG2_GUARD(this);
#endif
    notifyTree(&Widget::update);
}
void RootWidget::draw() {
#if defined (DENG_MOBILE)
    DENG2_GUARD(this);
#endif
    notifyTree(notifyArgsForDraw());
    Rule::markRulesValid(); // All done for this frame.
}
bool RootWidget::processEvent(Event const &event) {
#if defined (DENG_MOBILE)
    DENG2_GUARD(this);
#endif
    if (event.isKey() && focus()) {
        if (focus()->isDisabled()) {
            setFocus(nullptr);
        }
        else if (focus()->handleEvent(event)) {
            return true;
        }
    }
    return dispatchEvent(event, &Widget::handleEvent);
}
} // namespace de
namespace de {
bool Rule::_invalidRulesExist = false;
DENG2_PIMPL_NOREF(Rule) {
    typedef PointerSetT<Rule> Dependencies;
    Dependencies dependencies; // ref'd
    float value;
    bool isValid;
    Impl() : value(0), isValid(false)
    {}
    Impl(float initialValue) : value(initialValue), isValid(true)
    {}
    ~Impl() {
        DENG2_ASSERT(dependencies.isEmpty());
    }
};
Rule::Rule() : d(new Impl)
{}
Rule::Rule(float initialValue) : d(new Impl(initialValue))
{}
Rule::~Rule()
{}
float Rule::value() const {
    if (!d->isValid) {
        const_cast<Rule *>(this)->update();
    }
    DENG2_ASSERT(d->isValid);
    return d->value;
}
int Rule::valuei() const {
    return de::floor(value());
}
void Rule::update() {
    d->isValid = true;
}
bool Rule::isValid() const {
    return d->isValid;
}
void Rule::markRulesValid() {
    _invalidRulesExist = false;
}
bool Rule::invalidRulesExist() {
    return _invalidRulesExist;
}
float Rule::cachedValue() const {
    return d->value;
}
void Rule::ruleInvalidated() {
    invalidate();
}
void Rule::setValue(float v) {
    d->value = v;
    d->isValid = true;
}
void Rule::dependsOn(Rule const &dependency) {
    DENG2_ASSERT(!d->dependencies.contains(&dependency));
    d->dependencies.insert(de::holdRef(&dependency));
    dependency.audienceForRuleInvalidation += this;
}
void Rule::dependsOn(Rule const *dependencyOrNull) {
    if (dependencyOrNull) dependsOn(*dependencyOrNull);
}
void Rule::independentOf(Rule const &dependency) {
    dependency.audienceForRuleInvalidation -= this;
    DENG2_ASSERT(d->dependencies.contains(&dependency));
    d->dependencies.remove(&dependency);
    dependency.release();
}
void Rule::independentOf(Rule const *dependencyOrNull) {
    if (dependencyOrNull) independentOf(*dependencyOrNull);
}
void Rule::invalidate() {
    if (d->isValid) {
        d->isValid = false;
        Rule::_invalidRulesExist = true;
        DENG2_FOR_AUDIENCE(RuleInvalidation, i) i->ruleInvalidated();
    }
}
} // namespace de
namespace de {
const DotPath RuleBank::UNIT("unit");
DENG2_PIMPL_NOREF(RuleBank) {
    const Rule *dpiRule = nullptr;
    struct RuleSource : public ISource {
        RuleBank &bank;
        String id;
        RuleSource(RuleBank &b, String const &ruleId) : bank(b), id(ruleId) {}
        Time modifiedAt() const { return bank.sourceModifiedAt(); }
        const Rule &load() const {
            Record const &def = bank[id];
            return *bank.d->dpiRule * float(def["constant"].value().asNumber());
        }
    };
    struct RuleData : public IData {
        const Rule *rule;
        RuleData(const Rule &r) : rule(holdRef(r)) {}
        ~RuleData() { releaseRef(rule); }
    };
    ~Impl() {
        releaseRef(dpiRule);
    }
};
RuleBank::RuleBank(const Rule &dpiRule)
    : InfoBank("RuleBank", DisableHotStorage)
    , d(new Impl) {
    d->dpiRule = holdRef(dpiRule);
}
void RuleBank::addFromInfo(File const &file) {
    LOG_AS("RuleBank");
    parse(file);
    addFromInfoBlocks("rule");
}
Rule const &RuleBank::rule(DotPath const &path) const {
    if (path.isEmpty()) return ConstantRule::zero();
    return *static_cast<Impl::RuleData &>(data(path)).rule;
}
const Rule &RuleBank::dpiRule() const {
    return *d->dpiRule;
}
Bank::ISource *RuleBank::newSourceFromInfo(String const &id) {
    return new Impl::RuleSource(*this, id);
}
Bank::IData *RuleBank::loadFromSource(ISource &source) {
    return new Impl::RuleData(static_cast<Impl::RuleSource &>(source).load());
}
} // namespace de
namespace de {
DENG2_PIMPL(RuleRectangle) {
    String debugName;
    enum OutputId {
        OutLeft,
        OutRight,
        OutWidth,
        OutTop,
        OutBottom,
        OutHeight,
        MAX_OUTPUT_RULES
    };
    Rule const *inputRules[Rule::MAX_SEMANTICS];
    AnimationRule *_normalizedAnchorX = nullptr;
    AnimationRule *_normalizedAnchorY = nullptr;
    IndirectRule *outputRules[MAX_OUTPUT_RULES];
    Rule *_midX = nullptr;
    Rule *_midY = nullptr;
    Impl(Public *i) : Base(i) {
        zap(inputRules);
        for (int i = 0; i < int(MAX_OUTPUT_RULES); ++i) {
            outputRules[i] = new IndirectRule;
        }
        debugName = QString("0x%1").arg(dintptr(thisPublic), 0, 16);
    }
    ~Impl() {
        releaseRef(_midX);
        releaseRef(_midY);
        releaseRef(_normalizedAnchorX);
        releaseRef(_normalizedAnchorY);
        for (int i = 0; i < int(Rule::MAX_SEMANTICS); ++i) {
            releaseRef(inputRules[i]);
        }
        for (int i = 0; i < int(MAX_OUTPUT_RULES); ++i) {
            outputRules[i]->unsetSource();
            releaseRef(outputRules[i]);
        }
    }
    Rule const *&ruleRef(Rule::Semantic rule) {
        DENG2_ASSERT(rule >= Rule::Left);
        DENG2_ASSERT(rule < Rule::MAX_SEMANTICS);
        return inputRules[rule];
    }
    AnimationRule *normalizedAnchorX() {
        if (!_normalizedAnchorX) _normalizedAnchorX = new AnimationRule(0);
        return _normalizedAnchorX;
    }
    AnimationRule *normalizedAnchorY() {
        if (!_normalizedAnchorY) _normalizedAnchorY = new AnimationRule(0);
        return _normalizedAnchorY;
    }
    Rule *midX() {
        if (!_midX) _midX = holdRef(*outputRules[OutLeft] + *outputRules[OutWidth] / 2);
        return _midX;
    }
    Rule *midY() {
        if (!_midY) _midY = holdRef(*outputRules[OutTop] + *outputRules[OutHeight] / 2);
        return _midY;
    }
    inline Rule const &anchorPos(Rule::Semantic anchorInput) {
        if (anchorInput == Rule::AnchorX) {
            return *normalizedAnchorX();
        }
        else {
            DENG2_ASSERT(anchorInput == Rule::AnchorY);
            return *normalizedAnchorY();
        }
    }
    inline static bool isHorizontalInput(Rule::Semantic inputRule) {
        return inputRule == Rule::Left  ||
               inputRule == Rule::Right ||
               inputRule == Rule::Width ||
               inputRule == Rule::AnchorX;
    }
    inline static bool isVerticalInput(Rule::Semantic inputRule) {
        return !isHorizontalInput(inputRule);
    }
    void setInputRule(Rule::Semantic inputRule, Rule const &rule) {
        releaseRef(inputRules[inputRule]);
        inputRules[inputRule] = holdRef(rule);
        updateForChangedInput(inputRule);
    }
    void clearInputRule(Rule::Semantic inputRule) {
        releaseRef(inputRules[inputRule]);
        updateForChangedInput(inputRule);
    }
    void updateForChangedInput(Rule::Semantic input) {
        if (isHorizontalInput(input)) {
            updateHorizontalOutputs();
        }
        else {
            updateVerticalOutputs();
        }
    }
    void updateHorizontalOutputs() {
        updateDimension(Rule::Left, Rule::Right, Rule::Width, Rule::AnchorX,
                        OutLeft, OutRight, OutWidth);
    }
    void updateVerticalOutputs() {
        updateDimension(Rule::Top, Rule::Bottom, Rule::Height, Rule::AnchorY,
                        OutTop, OutBottom, OutHeight);
    }
    void updateDimension(Rule::Semantic minInput, Rule::Semantic maxInput,
                         Rule::Semantic deltaInput, Rule::Semantic anchorInput,
                         OutputId minOutput, OutputId maxOutput, OutputId deltaOutput) {
        bool minDefined   = false;
        bool maxDefined   = false;
        bool deltaDefined = false;
        outputRules[minOutput]  ->unsetSource();
        outputRules[maxOutput]  ->unsetSource();
        outputRules[deltaOutput]->unsetSource();
        if (inputRules[deltaInput]) {
            outputRules[deltaOutput]->setSource(*inputRules[deltaInput]);
            deltaDefined = true;
        }
        if (inputRules[minInput]) {
            outputRules[minOutput]->setSource(*inputRules[minInput]);
            minDefined = true;
        }
        if (inputRules[maxInput]) {
            outputRules[maxOutput]->setSource(*inputRules[maxInput]);
            maxDefined = true;
        }
        if (inputRules[anchorInput] && inputRules[deltaInput]) {
            outputRules[minOutput]->setSource(*inputRules[anchorInput] -
                    anchorPos(anchorInput) * *inputRules[deltaInput]);
            minDefined = true;
        }
        if (deltaDefined && minDefined && !maxDefined) {
            outputRules[maxOutput]->setSource(*outputRules[minOutput] + *outputRules[deltaOutput]);
            maxDefined = true;
        }
        if (deltaDefined && !minDefined && maxDefined) {
            outputRules[minOutput]->setSource(*outputRules[maxOutput] - *outputRules[deltaOutput]);
            minDefined = true;
        }
        if (!deltaDefined && minDefined && maxDefined) {
            outputRules[deltaOutput]->setSource(*outputRules[maxOutput] - *outputRules[minOutput]);
        }
    }
};
RuleRectangle::RuleRectangle() : d(new Impl(this))
{}
Rule const &RuleRectangle::left() const {
    return *d->outputRules[Impl::OutLeft];
}
Rule const &RuleRectangle::top() const {
    return *d->outputRules[Impl::OutTop];
}
Rule const &RuleRectangle::right() const {
    return *d->outputRules[Impl::OutRight];
}
Rule const &RuleRectangle::bottom() const {
    return *d->outputRules[Impl::OutBottom];
}
Rule const &RuleRectangle::width() const {
    return *d->outputRules[Impl::OutWidth];
}
Rule const &RuleRectangle::height() const {
    return *d->outputRules[Impl::OutHeight];
}
Rule const &RuleRectangle::midX() const {
    return *d->midX();
}
Rule const &RuleRectangle::midY() const {
    return *d->midY();
}
RuleRectangle &RuleRectangle::setInput(Rule::Semantic inputRule, RefArg<Rule> rule) {
    d->setInputRule(inputRule, rule);
    return *this;
}
RuleRectangle &RuleRectangle::setLeftTop(Rule const &left, Rule const &top) {
    setInput(Rule::Left, left);
    setInput(Rule::Top,  top);
    return *this;
}
RuleRectangle &RuleRectangle::setRightBottom(Rule const &right, Rule const &bottom) {
    setInput(Rule::Right,  right);
    setInput(Rule::Bottom, bottom);
    return *this;
}
RuleRectangle &RuleRectangle::setRect(RuleRectangle const &rect) {
    setInput(Rule::Left,   rect.left());
    setInput(Rule::Top,    rect.top());
    setInput(Rule::Right,  rect.right());
    setInput(Rule::Bottom, rect.bottom());
    return *this;
}
RuleRectangle &RuleRectangle::setInputsFromRect(RuleRectangle const &rect) {
    for (int i = 0; i < int(Rule::MAX_SEMANTICS); ++i) {
        if (rect.d->inputRules[i]) {
            setInput(Rule::Semantic(i), *rect.d->inputRules[i]);
        }
        else {
            clearInput(Rule::Semantic(i));
        }
    }
    return *this;
}
RuleRectangle &RuleRectangle::setSize(Rule const &width, Rule const &height) {
    setInput(Rule::Width,  width);
    setInput(Rule::Height, height);
    return *this;
}
RuleRectangle &RuleRectangle::setSize(ISizeRule const &dimensions) {
    return setSize(dimensions.width(), dimensions.height());
}
RuleRectangle &RuleRectangle::setMidAnchorX(Rule const &middle) {
    setInput(Rule::AnchorX, middle);
    d->normalizedAnchorX()->set(.5f);
    return *this;
}
RuleRectangle &RuleRectangle::setMidAnchorY(Rule const &middle) {
    setInput(Rule::AnchorY, middle);
    d->normalizedAnchorY()->set(.5f);
    return *this;
}
RuleRectangle &RuleRectangle::clearInput(Rule::Semantic inputRule) {
    d->clearInputRule(inputRule);
    return *this;
}
Rule const &RuleRectangle::inputRule(Rule::Semantic inputRule) {
    DENG2_ASSERT(d->ruleRef(inputRule) != 0);
    return *d->ruleRef(inputRule);
}
void RuleRectangle::setAnchorPoint(Vector2f const &normalizedPoint, TimeSpan const &transition) {
    d->normalizedAnchorX()->set(normalizedPoint.x, transition);
    d->normalizedAnchorY()->set(normalizedPoint.y, transition);
}
Rectanglef RuleRectangle::rect() const {
    return Rectanglef(Vector2f(left().value(),  top().value()),
                      Vector2f(right().value(), bottom().value()));
}
Vector2f RuleRectangle::size() const {
    return Vector2f(width().value(), height().value());
}
Vector2i RuleRectangle::sizei() const {
    return Vector2i(width().valuei(), height().valuei());
}
Vector2ui RuleRectangle::sizeui() const {
    return sizei().toVector2ui();
}
Rectanglei RuleRectangle::recti() const {
    Rectanglef const r = rect();
    return Rectanglei(Vector2i(de::floor(r.topLeft.x),     de::floor(r.topLeft.y)),
                      Vector2i(de::floor(r.bottomRight.x), de::floor(r.bottomRight.y)));
}
void RuleRectangle::setDebugName(String const &name) {
    d->debugName = name;
}
bool RuleRectangle::isFullyDefined() const {
    for (const auto *out : d->outputRules) {
        if (!out->hasSource()) {
            return false;
        }
    }
    return true;
}
String RuleRectangle::description() const {
    String desc = QString("RuleRectangle '%1'").arg(d->debugName);
    for (int i = 0; i < int(Rule::MAX_SEMANTICS); ++i) {
        desc += String("\n    INPUT ") +
                (i == Rule::Left? "Left" :
                 i == Rule::Top? "Top" :
                 i == Rule::Right? "Right" :
                 i == Rule::Bottom? "Bottom" :
                 i == Rule::Width? "Width" :
                 i == Rule::Height? "Height" :
                 i == Rule::AnchorX? "AnchorX" : "AnchorY") + ": ";
        if (d->inputRules[i]) {
            desc += String::format("(%g) ", d->inputRules[i]->value());
            desc += d->inputRules[i]->description();
        }
        else {
            desc += "(not set)";
        }
    }
    for (int i = 0; i < Impl::MAX_OUTPUT_RULES; ++i) {
        desc += String::format("\n    OUTPUT %s: ",
                                 i == Impl::OutLeft ? "Left"
                               : i == Impl::OutTop ? "Top"
                               : i == Impl::OutRight ? "Right"
                               : i == Impl::OutBottom ? "Bottom"
                               : i == Impl::OutWidth ? "Width"
                               : "Height");
        desc += String::format("(%g) ", d->outputRules[i]->value());
        desc += d->outputRules[i]->description();
    }
    return desc;
}
} // namespace de
namespace de {
DENG2_PIMPL(Widget) {
    Id id;
    String name;
    Widget *parent = nullptr;
    std::unique_ptr<Record> names; // IObject
    RootWidget *manualRoot = nullptr;
    Behaviors behavior;
    String focusNext;
    String focusPrev;
    typedef QMap<int, Widget *> Routing;
    Routing routing;
    typedef QList<Widget *> Children;
    typedef QMap<String, Widget *> NamedChildren;
    Children children;
    NamedChildren index;
    Impl(Public *i, String const &n) : Base(i), name(n)
    {}
    ~Impl() {
        clear();
    }
    void clear() {
        while (!children.isEmpty()) {
            children.first()->d->parent = 0;
            Widget *w = children.takeFirst();
            delete w;
        }
        index.clear();
    }
    RootWidget *findRoot() const {
        if (manualRoot) {
            return manualRoot;
        }
        Widget const *w = thisPublic;
        while (w->parent()) {
            w = w->parent();
            if (w->d->manualRoot) return w->d->manualRoot;
        }
        if (is<RootWidget>(w)) {
            return const_cast<RootWidget *>(&w->as<RootWidget>());
        }
        return nullptr;
    }
    enum AddBehavior { Append, Prepend, InsertBefore };
    void add(Widget *child, AddBehavior behavior, Widget const *ref = nullptr) {
        DENG2_ASSERT(child != 0);
        DENG2_ASSERT(child->d->parent == 0);
#ifdef _DEBUG
        if (self().parent()) {
            if (self().parent()->hasRoot()) {
                DENG2_ASSERT(!self().parent()->root().isInTree(*child));
            }
            else {
                DENG2_ASSERT(!self().parent()->isInTree(*child));
            }
        }
        else {
            DENG2_ASSERT(!self().isInTree(*child));
        }
#endif
        child->d->parent = thisPublic;
        switch (behavior) {
        case Append:
            children.append(child);
            break;
        case Prepend:
            children.push_front(child);
            break;
        case InsertBefore:
            children.insert(children.indexOf(const_cast<Widget *>(ref)), child);
            break;
        }
        if (!child->name().isEmpty()) {
            index.insert(child->name(), child);
        }
        DENG2_FOR_PUBLIC_AUDIENCE2(ChildAddition, i) {
            i->widgetChildAdded(*child);
        }
        DENG2_FOR_EACH_OBSERVER(ParentChangeAudience, i, child->audienceForParentChange()) {
            i->widgetParentChanged(*child, 0, thisPublic);
        }
    }
    Widget *walkChildren(Widget *beginFrom,
                         WalkDirection dir,
                         std::function<LoopResult (Widget &)> func,
                         int verticalDir = 0) {
        bool first = true;
        for (int pos = children.indexOf(beginFrom); ;
             pos += (dir == Forward? +1 : -1), first = false) {
            if (first && (verticalDir < 0 || (verticalDir == 0 && dir == Backward))) {
                continue;
            }
            bool const onlyDescend = (verticalDir == 0 && first);
            if (pos < 0 || pos >= children.size()) {
                if (verticalDir > 0) {
                    return nullptr;
                }
                if (dir == Backward && func(self())) {
                    return thisPublic;
                }
                if (parent) {
                    return parent->d->walkChildren(thisPublic, dir, func, -1 /*up*/);
                }
                return nullptr;
            }
            Widget *i = children.at(pos);
            if (dir == Forward && !onlyDescend && func(*i)) {
                return i;
            }
            if (i->childCount() > 0) {
                Widget *starting = (dir == Forward? i->d->children.first()
                                                  : i->d->children.last());
                if (Widget *found = i->d->walkChildren(starting, dir, func, +1 /*down*/))
                {
                    return found;
                }
            }
            if (dir == Backward && !onlyDescend && func(*i)) {
                return i;
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(ParentChange)
    DENG2_PIMPL_AUDIENCE(ChildAddition)
    DENG2_PIMPL_AUDIENCE(ChildRemoval)
};
DENG2_AUDIENCE_METHOD(Widget, Deletion)
DENG2_AUDIENCE_METHOD(Widget, ParentChange)
DENG2_AUDIENCE_METHOD(Widget, ChildAddition)
DENG2_AUDIENCE_METHOD(Widget, ChildRemoval)
Widget::Widget(String const &name) : d(new Impl(this, name))
{}
Widget::~Widget() {
    if (hasRoot() && root().focus() == this) {
        root().setFocus(0);
    }
    audienceForParentChange().clear();
    if (d->parent) {
        d->parent->remove(*this);
    }
    DENG2_FOR_AUDIENCE2(Deletion, i) {
        i->widgetBeingDeleted(*this);
    }
}
Id Widget::id() const {
    return d->id;
}
String Widget::name() const {
    return d->name;
}
void Widget::setName(String const &name) {
    if (d->parent && !d->name.isEmpty()) {
        d->parent->d->index.remove(d->name);
    }
    d->name = name;
    if (d->parent && !name.isEmpty()) {
        d->parent->d->index.insert(name, this);
    }
}
DotPath Widget::path() const {
    Widget const *w = this;
    String result;
    while (w) {
        if (!result.isEmpty()) result = "." + result;
        if (!w->d->name.isEmpty()) {
            result = w->d->name + result;
        }
        else {
            result = QString("0x%1").arg(dintptr(w), 0, 16) + result;
        }
        w = w->parent();
    }
    return result;
}
bool Widget::hasRoot() const {
    return d->findRoot() != nullptr;
}
RootWidget &Widget::root() const {
    if (auto *rw = d->findRoot()) {
        return *rw;
    }
    throw NotFoundError("Widget::root", "No root widget found");
}
RootWidget *Widget::findRoot() const {
    return d->findRoot();
}
void Widget::setRoot(RootWidget *root) {
    d->manualRoot = root;
}
bool Widget::hasFocus() const {
    return hasRoot() && root().focus() == this;
}
bool Widget::canBeFocused() const {
    return behavior().testFlag(Focusable) && isVisible() && isEnabled();
}
bool Widget::hasFamilyBehavior(Behavior const &flags) const {
    for (Widget const *w = this; w != 0; w = w->d->parent) {
        if (w->d->behavior.testFlag(flags)) return true;
    }
    return false;
}
void Widget::show(bool doShow) {
    setBehavior(Hidden, !doShow);
}
void Widget::setBehavior(Behaviors behavior, FlagOpArg operation) {
    applyFlagOperation(d->behavior, behavior, operation);
}
void Widget::unsetBehavior(Behaviors behavior) {
    applyFlagOperation(d->behavior, behavior, UnsetFlags);
}
Widget::Behaviors Widget::behavior() const {
    return d->behavior;
}
void Widget::setFocusNext(String const &name) {
    d->focusNext = name;
}
void Widget::setFocusPrev(String const &name) {
    d->focusPrev = name;
}
String Widget::focusNext() const {
    return d->focusNext;
}
String Widget::focusPrev() const {
    return d->focusPrev;
}
void Widget::setEventRouting(QList<int> const &types, Widget *routeTo) {
    foreach (int type, types) {
        if (routeTo) {
            d->routing.insert(type, routeTo);
        }
        else {
            d->routing.remove(type);
        }
    }
}
void Widget::clearEventRouting() {
    d->routing.clear();
}
bool Widget::isEventRouted(int type, Widget *to) const {
    return d->routing.contains(type) && d->routing[type] == to;
}
void Widget::clearTree() {
    d->clear();
}
Widget &Widget::add(Widget *child) {
    return addLast(child);
}
Widget &Widget::addLast(Widget *child) {
    d->add(child, Impl::Append);
    return *child;
}
Widget &Widget::addFirst(Widget *child) {
    d->add(child, Impl::Prepend);
    return *child;
}
Widget &Widget::insertBefore(Widget *child, Widget const &otherChild) {
    DENG2_ASSERT(child != &otherChild);
    DENG2_ASSERT(otherChild.parent() == this);
    d->add(child, Impl::InsertBefore, &otherChild);
    return *child;
}
Widget *Widget::remove(Widget &child) {
    DENG2_ASSERT(child.d->parent == this);
    DENG2_ASSERT(d->children.contains(&child));
    child.d->parent = 0;
    d->children.removeOne(&child);
    DENG2_ASSERT(!d->children.contains(&child));
    if (!child.name().isEmpty()) {
        d->index.remove(child.name());
    }
    DENG2_FOR_AUDIENCE2(ChildRemoval, i) {
        i->widgetChildRemoved(child);
    }
    DENG2_FOR_EACH_OBSERVER(ParentChangeAudience, i, child.audienceForParentChange()) {
        i->widgetParentChanged(child, this, 0);
    }
    return &child;
}
void Widget::orphan() {
    if (d->parent) {
        d->parent->remove(*this);
    }
    DENG2_ASSERT(d->parent == nullptr);
}
Widget *Widget::find(String const &name) {
    if (d->name == name) return this;
    Impl::NamedChildren::const_iterator found = d->index.constFind(name);
    if (found != d->index.constEnd()) {
        return found.value();
    }
    DENG2_FOR_EACH_CONST(Impl::Children, i, d->children) {
        Widget *w = (*i)->find(name);
        if (w) return w;
    }
    return 0;
}
bool Widget::isInTree(Widget const &child) const {
    if (this == &child) return true;
    DENG2_FOR_EACH_CONST(Impl::Children, i, d->children) {
        if ((*i)->isInTree(child)) {
            return true;
        }
    }
    return false;
}
bool Widget::hasAncestor(Widget const &ancestorOrParent) const {
    for (Widget const *iter = parent(); iter; iter = iter->parent()) {
        if (iter == &ancestorOrParent) return true;
    }
    return false;
}
Widget const *Widget::find(String const &name) const {
    return const_cast<Widget *>(this)->find(name);
}
void Widget::moveChildBefore(Widget *child, Widget const &otherChild) {
    if (child == &otherChild) return; // invalid
    int from = -1;
    int to = -1;
    for (int i = 0; i < d->children.size() && (from < 0 || to < 0); ++i) {
        if (d->children.at(i) == child) {
            from = i;
        }
        if (d->children.at(i) == &otherChild) {
            to = i;
        }
    }
    DENG2_ASSERT(from != -1);
    DENG2_ASSERT(to != -1);
    d->children.removeAt(from);
    if (to > from) to--;
    d->children.insert(to, child);
}
void Widget::moveChildToLast(Widget &child) {
    DENG2_ASSERT(child.parent() == this);
    if (!child.isLastChild()) {
        remove(child);
        add(&child);
    }
}
Widget *Widget::parent() const {
    return d->parent;
}
bool Widget::isFirstChild() const {
    if (!parent()) return false;
    return this == parent()->d->children.first();
}
bool Widget::isLastChild() const {
    if (!parent()) return false;
    return this == parent()->d->children.last();
}
Widget *Widget::walkInOrder(WalkDirection dir, std::function<LoopResult (Widget &)> callback) {
    if (!d->parent) {
        if (d->children.isEmpty()) return nullptr;
        if (dir == Forward) {
            return d->walkChildren(d->children.first(), Forward, callback, +1);
        }
        else {
            return nullptr;
        }
    }
    return d->parent->d->walkChildren(this, dir, callback); // allows ascending
}
Widget *Widget::walkChildren(WalkDirection dir, std::function<LoopResult (Widget &)> callback) {
    if (d->children.isEmpty()) return nullptr;
    return d->walkChildren(dir == Forward? d->children.first() : d->children.last(),
                           dir, callback, +1);
}
String Widget::uniqueName(String const &name) const {
    return String("#%1.%2").arg(id().asInt64()).arg(name);
}
Widget::NotifyArgs Widget::notifyArgsForDraw() const {
    NotifyArgs args(&Widget::draw);
    args.conditionFunc  = &Widget::isVisible;
    args.preNotifyFunc  = &Widget::preDrawChildren;
    args.postNotifyFunc = &Widget::postDrawChildren;
    return args;
}
Widget::NotifyArgs::Result Widget::notifyTree(NotifyArgs const &args) {
    NotifyArgs::Result result = NotifyArgs::Continue;
    bool preNotified = false;
    for (int idx = 0; idx < d->children.size(); ++idx) {
        Widget *i = d->children.at(idx);
        if (i == args.until) {
            result = NotifyArgs::Abort;
            break;
        }
        if (args.conditionFunc && !(i->*args.conditionFunc)())
            continue; // Skip this one.
        if (args.preNotifyFunc && !preNotified) {
            preNotified = true;
            (this->*args.preNotifyFunc)();
        }
        (i->*args.notifyFunc)();
        if (i != d->children.at(idx)) {
            int newIdx = d->children.indexOf(i);
            if (newIdx >= 0) {
                idx = newIdx;
                i = d->children.at(newIdx);
            }
            else {
                idx--;
                continue;
            }
        }
        if (i->childCount()) {
            if (i->notifyTree(args) == NotifyArgs::Abort) {
                result = NotifyArgs::Abort;
                break;
            }
        }
    }
    if (args.postNotifyFunc && preNotified) {
        (this->*args.postNotifyFunc)();
    }
    return result;
}
Widget::NotifyArgs::Result Widget::notifySelfAndTree(NotifyArgs const &args) {
    (this->*args.notifyFunc)();
    return notifyTree(args);
}
void Widget::notifyTreeReversed(NotifyArgs const &args) {
    if (args.preNotifyFunc) {
        (this->*args.preNotifyFunc)();
    }
    for (int i = d->children.size() - 1; i >= 0; --i) {
        Widget *w = d->children.at(i);
        if (args.conditionFunc && !(w->*args.conditionFunc)())
            continue; // Skip this one.
        w->notifyTreeReversed(args);
        (w->*args.notifyFunc)();
    }
    if (args.postNotifyFunc) {
        (this->*args.postNotifyFunc)();
    }
}
bool Widget::dispatchEvent(Event const &event, bool (Widget::*memberFunc)(Event const &)) {
    if (isHidden() || d->behavior.testFlag(DisableEventDispatch)) return false;
    if (d->routing.contains(event.type())) {
        return d->routing[event.type()]->dispatchEvent(event, memberFunc);
    }
    bool const thisHasFocus = (hasRoot() && root().focus() == this && event.isKey());
    if (d->behavior.testFlag(HandleEventsOnlyWhenFocused) && !thisHasFocus) {
        return false;
    }
    if (thisHasFocus) {
        return false;
    }
    if (!d->behavior.testFlag(DisableEventDispatchToChildren)) {
        for (int i = d->children.size() - 1; i >= 0; --i) {
            Widget *w = d->children.at(i);
            bool eaten = w->dispatchEvent(event, memberFunc);
            if (eaten) return true;
        }
    }
    if ((this->*memberFunc)(event)) {
        return true;
    }
    return false;
}
void Widget::waitForAssetsReady() {
    AssetGroup assets;
    collectNotReadyAssets(assets, CollectMode::OnlyVisible);
    if (!assets.isEmpty()) {
        assets.waitForState(Asset::Ready);
    }
}
void Widget::collectNotReadyAssets(AssetGroup &collected, CollectMode collectMode) {
    if (collectMode == CollectMode::OnlyVisible && behavior().testFlag(Hidden)) {
        return;
    }
    if (auto *assetGroup = maybeAs<IAssetGroup>(this)) {
        if (!assetGroup->assets().isReady()) {
            collected += *assetGroup;
            LOGDEV_XVERBOSE("Found " _E(m) "NotReady" _E(.) " asset %s (%p)", path() << this);
        }
    }
    else {
        foreach (Widget *child, children()) {
            child->collectNotReadyAssets(collected, collectMode);
        }
    }
}
Widget::Children Widget::children() const {
    return d->children;
}
dsize de::Widget::childCount() const {
    return dsize(d->children.size());
}
void Widget::initialize()
{}
void Widget::deinitialize()
{}
void Widget::viewResized()
{}
void Widget::focusGained()
{}
void Widget::focusLost()
{}
void Widget::offerFocus()
{}
void Widget::update()
{}
void Widget::draw()
{}
void Widget::preDrawChildren()
{}
void Widget::postDrawChildren()
{}
bool Widget::handleEvent(Event const &) {
    return false;
}
Record &Widget::objectNamespace() {
    if (!d->names) {
        d->names.reset(new Record);
    }
    return *d->names;
}
const Record &Widget::objectNamespace() const {
    return const_cast<Widget *>(this)->objectNamespace();
}
void Widget::setFocusCycle(WidgetList const &order) {
    for (int i = 0; i < order.size(); ++i) {
        Widget *a = order[i];
        Widget *b = order[(i + 1) % order.size()];
        a->setFocusNext(b->name());
        b->setFocusPrev(a->name());
    }
}
} // namespace de
