﻿/** @file abstractlineeditor.cpp  Abstract line editor.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/AbstractLineEditor"
#include "de/shell/Lexicon"
#include <de/ConstantRule>
#include <QList>
#include <QStringList>
#include <QScopedPointer>
/** @file abstractlink.h  Network connection to a server.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/AbstractLink"
#include <de/App>
#include <de/Log>
#include <de/Message>
#include <de/Packet>
#include <de/Socket>
#include <de/Time>
#include <QTimer>
/** @file libshell/src/action.cpp  Maps a key event to a signal.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/Action"
/** @file libshell/src/choicewidget.cpp  Widget for selecting an item from multiple choices.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/ChoiceWidget"
#include "de/shell/MenuWidget"
#include <de/shell/TextRootWidget>
/** @file commandlinewidget.cpp  Widget for command line input.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/CommandLineWidget"
#include "de/shell/TextRootWidget"
#include "de/shell/KeyEvent"
#include "de/shell/EditorHistory"
#include <de/String>
#include <QStringList>
/** @file libshell/src/dialogwidget.cpp  Base class for modal dialogs.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/DialogWidget"
#include "de/shell/TextRootWidget"
#include "de/shell/KeyEvent"
#include <QEventLoop>
/** @file doomsdayinfo.cpp  Information about Doomsday Engine and its plugins.
 *
 * @todo This information should not be hardcoded. Instead, it should be read
 * from Info files, and some functionality could be determined using Doomsday
 * Script or a shared Game Rules library.
 *
 * @todo Option to order games by release date, name, or some other criteria.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/DoomsdayInfo"
#include <QDir>
/** @file editorhistory.cpp  Text editor history buffer.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/EditorHistory"
#include <de/math.h>
/** @file libshell/src/inputdialog.cpp  Dialog for querying text from the user.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/InputDialog"
#include "de/shell/LabelWidget"
#include "de/shell/LineEditWidget"
#include "de/shell/MenuWidget"
#include "de/shell/TextRootWidget"
/** @file libshell/src/labelwidget.cpp  Widget for showing a label.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/LabelWidget"
#include "de/shell/TextRootWidget"
#include "de/shell/MonospaceLineWrapping"
#include <de/ConstantRule>
/** @file lexicon.cpp  Lexicon containing terms and grammatical rules.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/Lexicon"
/** @file libshell.cpp  Common utility functions for libshell.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/libshell.h"
#include <de/math.h>
/** @file libshell/src/lineeditwidget.cpp  Widget for word-wrapped text editing.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/LineEditWidget"
#include "de/shell/TextRootWidget"
#include "de/shell/KeyEvent"
#include "de/shell/Lexicon"
#include "de/shell/MonospaceLineWrapping"
#include <de/Rule>
#include <de/RuleRectangle>
#include <de/String>
#include <de/Log>
#include <QSet>
/** @file link.cpp  Network connection to a server.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/Link"
#include "de/shell/Protocol"
#include <de/Message>
#include <de/Socket>
#include <de/Time>
#include <de/Log>
#include <de/ByteRefArray>
/** @file localserver.cpp  Starting and stopping local servers.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/LocalServer"
#include "de/shell/Link"
#include "de/shell/DoomsdayInfo"
#include <de/CommandLine>
#include <QCoreApplication>
#include <QDir>
#if !defined (DENG_MOBILE)
namespace de { namespace shell {
static String const ERROR_LOG_NAME = "doomsday-errors.out";
DENG2_PIMPL_NOREF(LocalServer)
{
    Link *     link = nullptr;
    NativePath appPath; // where to find the server
    duint16    port = 0;
    String     name;
    NativePath userDir;
    QProcess * proc = nullptr; // not deleted until stopped
    ~Impl()
    {
        if (proc && proc->state() == QProcess::NotRunning)
        {
            delete proc;
        }
    }
};
LocalServer::LocalServer() : d(new Impl)
{}
void LocalServer::setName(String const &name)
{
    d->name = name;
    d->name.replace("\"", "\\\""); // for use on command line
}
void LocalServer::setApplicationPath(NativePath const &path)
{
    d->appPath = path;
}
void LocalServer::start(duint16 port,
                        String const &gameMode,
                        QStringList additionalOptions,
                        NativePath const &runtimePath)
{
    d->port = port;
    d->userDir = runtimePath;
    if (d->userDir.isEmpty())
    {
        d->userDir = DoomsdayInfo::defaultServerRuntimeFolder();
    }
    QDir(d->userDir).remove(ERROR_LOG_NAME);
    DENG2_ASSERT(d->link == 0);
    CommandLine cmd;
    NativePath bin;
#ifdef MACOSX
    if (!d->appPath.isEmpty())
    {
        bin = d->appPath / "Doomsday.app/Contents/MacOS/doomsday-server";
        if (!bin.exists())
        {
            bin = d->appPath / "Contents/MacOS/doomsday-server";
        }
    }
    if (!bin.exists())
    {
        bin = NativePath(qApp->applicationDirPath()) / "../MacOS/doomsday-server";
    }
    if (!bin.exists())
    {
        bin = NativePath(qApp->applicationDirPath()) /
                "../../../Doomsday.app/Contents/MacOS/doomsday-server";
    }
    if (!bin.exists())
    {
        throw NotFoundError("LocalServer::start", "Could not find Doomsday.app");
    }
    cmd.append(bin);
#elif WIN32
    if (!d->appPath.isEmpty())
    {
        bin = d->appPath / "doomsday-server.exe";
    }
    if (!bin.exists())
    {
        bin = NativePath(qApp->applicationDirPath()) / "doomsday-server.exe";
    }
    cmd.append(bin);
    cmd.append("-basedir");
    cmd.append(bin.fileNamePath() / "..");
#else // UNIX
    if (!d->appPath.isEmpty())
    {
        bin = d->appPath / "doomsday-server";
    }
    if (!bin.exists())
    {
        bin = NativePath(qApp->applicationDirPath()) / "doomsday-server";
    }
    if (!bin.exists()) bin = "doomsday-server"; // Perhaps it's on the path?
    cmd.append(bin);
#endif
    cmd.append("-userdir");
    cmd.append(d->userDir);
    cmd.append("-errors");
    cmd.append(ERROR_LOG_NAME);
    cmd.append("-game");
    cmd.append(gameMode);
    cmd.append("-cmd");
    cmd.append("net-ip-port " + String::number(port));
    if (!d->name.isEmpty())
    {
        cmd.append("-cmd");
        cmd.append("server-name \"" + d->name + "\"");
    }
    foreach (String opt, additionalOptions) cmd.append(opt);
    LOG_NET_NOTE("Starting local server on port %i using game mode '%s'")
            << port << gameMode;
    d->proc = cmd.executeProcess();
}
void LocalServer::stop()
{
    if (isRunning())
    {
        LOG_NET_NOTE("Stopping local server on port %i") << d->port;
        d->proc->kill();
    }
}
duint16 LocalServer::port() const
{
    return d->port;
}
bool LocalServer::isRunning() const
{
    if (!d->proc) return false;
    return (d->proc->state() != QProcess::NotRunning);
}
Link *LocalServer::openLink()
{
    if (!isRunning()) return nullptr;
    return new Link(String("localhost:%1").arg(d->port), 30);
}
NativePath LocalServer::errorLogPath() const
{
    return d->userDir / ERROR_LOG_NAME;
}
}} // namespace de::shell
#endif
/** @file libshell/src/logwidget.cpp  Widget for output message log.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/LogWidget"
#include "de/shell/KeyEvent"
#include "de/shell/TextRootWidget"
#include <de/MonospaceLogSinkFormatter>
#include <de/MemoryLogSink>
#include <de/LogBuffer>
#include <QList>
/** @file libshell/src/menuwidget.cpp  Menu with shortcuts.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/MenuWidget"
#include "de/shell/TextRootWidget"
#include <de/ConstantRule>
#include <QList>
/** @file monospacelinewrapping.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/MonospaceLineWrapping"
#include <de/math.h>
/** @file packagedownloader.cpp  Utility for downloading packages from a remote repository.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/PackageDownloader"
#include <de/App>
#include <de/FileSystem>
#include <de/Folder>
#include <de/Garbage>
#include <de/LinkFile>
#include <de/Loop>
#include <de/filesys/NativeLink>
#include <de/PackageLoader>
#include <de/RemoteFeedRelay>
#include <de/RemoteFile>
/** @file libshell/src/protocol.cpp  Network protocol for communicating with a server.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/Protocol"
#include <de/LogBuffer>
#include <de/ArrayValue>
#include <de/TextValue>
#include <de/Reader>
#include <de/Writer>
#include <QCryptographicHash>
#include <QList>
/** @file serverfinder.cpp  Looks up servers via beacon.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/ServerFinder"
#include <de/App>
#include <de/Beacon>
#include <de/CommandLine>
#include <de/LogBuffer>
#include <de/NumberValue>
#include <de/Reader>
#include <de/TextValue>
#include <QMap>
#include <QTimer>
/** @file serverinfo.cpp  Information about a multiplayer server.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/ServerInfo"
#include <de/TextValue>
#include <de/Version>
#include <de/Log>
#include <de/data/json.h>
/** @file textcanvas.cpp Text-based drawing surface.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/TextCanvas"
#include <QList>
#include <QDebug>
/** @file textrootwidget.cpp Text-based root widget.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/TextRootWidget"
#include "de/shell/TextWidget"
/** @file textwidget.cpp  Generic widget with a text-based visual.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/shell/TextWidget"
#include "de/shell/TextRootWidget"
#include "de/shell/Action"
#include <QList>

namespace de { namespace shell {
DENG2_PIMPL(AbstractLineEditor) {
    String   prompt;
    String   text;
    int      cursor; ///< Index in range [0...text.size()]
    Lexicon  lexicon;
    EchoMode echoMode;
    QScopedPointer<ILineWrapping> wraps;
    struct Completion {
        int pos;
        int size;
        int ordinal; ///< Ordinal within list of possible completions.
        void reset() {
            pos = size = ordinal = 0;
        }
        Rangei range() const {
            return Rangei(pos, pos + size);
        }
    };
    Completion  completion;
    QStringList suggestions;
    bool        suggesting;
    bool        completionNotified;
    Impl(Public *i, ILineWrapping *lineWraps)
        : Base(i),
          cursor(0),
          echoMode(NormalEchoMode),
          wraps(lineWraps),
          suggesting(false),
          completionNotified(false) {
        completion.reset();
    }
    WrappedLine lineSpan(int line) const {
        DENG2_ASSERT(line < wraps->height());
        return wraps->line(line);
    }
    void rewrapLater() {
        wraps->clear();
        self().contentChanged();
    }
    void rewrapNow() {
        updateWraps();
        self().contentChanged();
    }
    /**
     * Determines where word wrapping needs to occur and updates the height of
     * the widget to accommodate all the needed lines.
     */
    void updateWraps() {
        wraps->wrapTextToWidth(text, de::max(1, self().maximumWidth()));
        if (wraps->height() > 0) {
            self().numberOfLinesChanged(wraps->height());
        }
        else {
            self().numberOfLinesChanged(1);
        }
    }
    de::Vector2i lineCursorPos() const {
        return linePos(cursor);
    }
    de::Vector2i linePos(int mark) const {
        de::Vector2i pos(mark);
        for (pos.y = 0; pos.y < wraps->height(); ++pos.y) {
            WrappedLine span = lineSpan(pos.y);
            if (!span.isFinal) span.range.end--;
            if (mark >= span.range.start && mark <= span.range.end) {
                break;
            }
            pos.x -= span.range.end - span.range.start + 1;
        }
        return pos;
    }
    /**
     * Attemps to move the cursor up or down by a line.
     *
     * @return @c true, if cursor was moved. @c false, if there were no more
     * lines available in that direction.
     */
    bool moveCursorByLine(int lineOff) {
        acceptCompletion();
        DENG2_ASSERT(lineOff == 1 || lineOff == -1);
        Vector2i const linePos = lineCursorPos();
        int const destWidth = wraps->rangeWidth(Rangei(lineSpan(linePos.y).range.start, cursor));
        if (!linePos.y && lineOff < 0) return false;
        if (linePos.y == wraps->height() - 1 && lineOff > 0) return false;
        WrappedLine span = lineSpan(linePos.y + lineOff);
        cursor = wraps->indexAtWidth(span.range, destWidth);
        if (!span.isFinal) span.range.end--;
        if (cursor > span.range.end) cursor = span.range.end;
        self().cursorMoved();
        return true;
    }
    void insert(String const &str) {
        acceptCompletion();
        text.insert(cursor, str);
        cursor += str.size();
        rewrapNow();
    }
    void doBackspace() {
        if (rejectCompletion())
            return;
        if (!text.isEmpty() && cursor > 0) {
            text.remove(--cursor, 1);
            rewrapNow();
        }
    }
    void doWordBackspace() {
        rejectCompletion();
        if (!text.isEmpty() && cursor > 0) {
            int to = wordJumpLeft(cursor);
            text.remove(to, cursor - to);
            cursor = to;
            rewrapNow();
        }
    }
    void doDelete() {
        if (text.size() > cursor) {
            text.remove(cursor, 1);
            rewrapNow();
        }
    }
    bool doLeft() {
        acceptCompletion();
        if (cursor > 0) {
            --cursor;
            self().cursorMoved();
            return true;
        }
        return false;
    }
    bool doRight() {
        acceptCompletion();
        if (cursor < text.size()) {
            ++cursor;
            self().cursorMoved();
            return true;
        }
        return false;
    }
    int wordJumpLeft(int pos) const {
        pos = de::min(pos, text.size() - 1);
        while (pos > 0 && !text[pos].isLetterOrNumber()) pos--;
        if (pos > 0) pos--;
        while (pos > 0 && text[pos - 1].isLetterOrNumber()) pos--;
        return pos;
    }
    void doWordLeft() {
        acceptCompletion();
        cursor = wordJumpLeft(cursor);
        self().cursorMoved();
    }
    void doWordRight() {
        int const last = text.size() - 1;
        acceptCompletion();
        while (cursor <= last && text[de::min(last, cursor)].isLetterOrNumber()) {
            cursor++;
        }
        while (cursor <= last && !text[de::min(last, cursor)].isLetterOrNumber()) {
            cursor++;
        }
        self().cursorMoved();
    }
    void doHome() {
        acceptCompletion();
        cursor = lineSpan(lineCursorPos().y).range.start;
        self().cursorMoved();
    }
    void doEnd() {
        acceptCompletion();
        WrappedLine const span = lineSpan(lineCursorPos().y);
        cursor = span.range.end - (span.isFinal? 0 : 1);
        self().cursorMoved();
    }
    void killEndOfLine() {
        text.remove(cursor, lineSpan(lineCursorPos().y).range.end - cursor);
        rewrapNow();
    }
    bool suggestingCompletion() const {
        return suggesting;
    }
    String wordBehindPos(int pos) const {
        String word;
        int i = pos - 1;
        while (i >= 0 && lexicon.isWordChar(text[i])) word.prepend(text[i--]);
        return word;
    }
    String wordBehindCursor() const {
        return wordBehindPos(cursor);
    }
    QStringList completionsForBase(String base, String &commonPrefix) const {
        String::CaseSensitivity const sensitivity =
                lexicon.isCaseSensitive()? String::CaseSensitive : String::CaseInsensitive;
        bool first = true;
        QStringList sugs;
        foreach (String term, lexicon.terms()) {
            if (term.beginsWith(base, sensitivity) && term.size() > base.size()) {
                sugs << term;
                if (first) {
                    commonPrefix = term;
                    first = false;
                }
                else if (!commonPrefix.isEmpty()) {
                    int len = commonPrefix.commonPrefixLength(term, sensitivity);
                    commonPrefix = commonPrefix.left(len);
                }
            }
        }
        qSort(sugs);
        return sugs;
    }
    bool doCompletion(bool forwardCycle) {
        if (!suggestingCompletion()) {
            completionNotified = false;
            String const base = wordBehindCursor();
            if (!base.isEmpty()) {
                String commonPrefix;
                suggestions = completionsForBase(base, commonPrefix);
                if (!commonPrefix.isEmpty() && commonPrefix != base) {
                    completion.ordinal = -1;
                    commonPrefix.remove(0, base.size());
                    completion.pos = cursor;
                    completion.size = commonPrefix.size();
                    text.insert(cursor, commonPrefix);
                    cursor += completion.size;
                    rewrapNow();
                    suggesting = true;
                    return true;
                }
                if (!suggestions.isEmpty()) {
                    completion.ordinal = -1; //(forwardCycle? 0 : suggestions.size() - 1);
                    /*String comp = suggestions[completion.ordinal];
                    comp.remove(0, base.size());*/
                    completion.pos = cursor;
                    completion.size = 0; //comp.size();
                    suggesting = true;
                    self().autoCompletionBegan(base);
                    completionNotified = true;
                    return true;
                }
            }
        }
        else {
            if (!completionNotified) {
                self().autoCompletionBegan(wordBehindPos(completion.pos));
                completionNotified = true;
                return true;
            }
            cursor = completion.pos;
            String const base = wordBehindCursor();
            if (completion.ordinal < 0) {
                completion.ordinal = (forwardCycle? 0 : suggestions.size() - 1);
                if (base + text.mid(completion.pos, completion.size) == suggestions[completion.ordinal]) {
                    cycleCompletion(forwardCycle);
                }
            }
            else {
                cycleCompletion(forwardCycle);
            }
            String comp = suggestions[completion.ordinal];
            comp.remove(0, base.size());
            text.remove(completion.pos, completion.size);
            text.insert(completion.pos, comp);
            completion.size = comp.size();
            cursor = completion.pos + completion.size;
            rewrapNow();
            return true;
        }
        return false;
    }
    void cycleCompletion(bool forwardCycle) {
        completion.ordinal = de::wrap(completion.ordinal + (forwardCycle? 1 : -1),
                                      0, suggestions.size());
    }
    void resetCompletion() {
        completion.reset();
        suggestions.clear();
        suggesting = false;
        completionNotified = false;
    }
    void acceptCompletion() {
        if (!suggestingCompletion()) return;
        resetCompletion();
        self().autoCompletionEnded(true);
    }
    bool rejectCompletion() {
        if (!suggestingCompletion()) return false;
        int oldCursor = cursor;
        text.remove(completion.pos, completion.size);
        cursor = completion.pos;
        resetCompletion();
        rewrapNow();
        self().autoCompletionEnded(false);
        return cursor != oldCursor; // cursor was moved as part of the rejection
    }
};
AbstractLineEditor::AbstractLineEditor(ILineWrapping *lineWraps) : d(new Impl(this, lineWraps))
{}
ILineWrapping const &AbstractLineEditor::lineWraps() const {
    return *d->wraps;
}
void AbstractLineEditor::setPrompt(String const &promptText) {
    d->prompt = promptText;
    d->rewrapLater();
}
String AbstractLineEditor::prompt() const {
    return d->prompt;
}
void AbstractLineEditor::setText(String const &contents) {
    d->completion.reset();
    d->text = contents;
    d->cursor = contents.size();
    d->rewrapLater();
}
String AbstractLineEditor::text() const {
    return d->text;
}
void AbstractLineEditor::setCursor(int index) {
    d->completion.reset();
    d->cursor = index;
    cursorMoved();
}
int AbstractLineEditor::cursor() const {
    return d->cursor;
}
Vector2i AbstractLineEditor::linePos(int index) const {
    return d->linePos(index);
}
bool AbstractLineEditor::isSuggestingCompletion() const {
    return d->suggestingCompletion();
}
Rangei AbstractLineEditor::completionRange() const {
    return d->completion.range();
}
QStringList AbstractLineEditor::suggestedCompletions() const {
    if (!isSuggestingCompletion()) return QStringList();
    return d->suggestions;
}
void shell::AbstractLineEditor::acceptCompletion() {
    d->acceptCompletion();
}
void AbstractLineEditor::setLexicon(Lexicon const &lexicon) {
    d->lexicon = lexicon;
}
Lexicon const &AbstractLineEditor::lexicon() const {
    return d->lexicon;
}
void AbstractLineEditor::setEchoMode(EchoMode mode) {
    d->echoMode = mode;
}
AbstractLineEditor::EchoMode AbstractLineEditor::echoMode() const {
    return d->echoMode;
}
bool AbstractLineEditor::handleControlKey(int qtKey, KeyModifiers const &mods) {
#ifdef MACOSX
#  define WORD_JUMP_MODIFIER    Alt
#else
#  define WORD_JUMP_MODIFIER    Control
#endif
    switch (qtKey) {
    case Qt::Key_Backspace:
        if (mods.testFlag(WORD_JUMP_MODIFIER)) {
            d->doWordBackspace();
        }
        else {
            d->doBackspace();
        }
        return true;
    case Qt::Key_Delete:
        d->doDelete();
        return true;
    case Qt::Key_Left:
#ifdef MACOSX
        if (mods.testFlag(Control)) {
            d->doHome();
            return true;
        }
#endif
        if (mods.testFlag(WORD_JUMP_MODIFIER)) {
            d->doWordLeft();
        }
        else {
            return d->doLeft();
        }
        return true;
    case Qt::Key_Right:
#ifdef MACOSX
        if (mods.testFlag(Control)) {
            d->doEnd();
            return true;
        }
#endif
        if (mods.testFlag(WORD_JUMP_MODIFIER)) {
            d->doWordRight();
        }
        else {
            return d->doRight();
        }
        return true;
    case Qt::Key_Home:
        d->doHome();
        return true;
    case Qt::Key_End:
        d->doEnd();
        return true;
    case Qt::Key_Tab:
    case Qt::Key_Backtab:
        if (d->doCompletion(qtKey == Qt::Key_Tab)) {
            return true;
        }
        break;
    case Qt::Key_K:
        if (mods.testFlag(Control)) {
            d->killEndOfLine();
            return true;
        }
        break;
    case Qt::Key_Up:
        if (!d->moveCursorByLine(-1)) return false; // not eaten
        return true;
    case Qt::Key_Down:
        if (!d->moveCursorByLine(+1)) return false; // not eaten
        return true;
    case Qt::Key_Enter:
    case Qt::Key_Return:
        d->acceptCompletion();
        return true;
    default:
        break;
    }
    return false;
}
void AbstractLineEditor::insert(String const &text) {
    return d->insert(text);
}
ILineWrapping &AbstractLineEditor::lineWraps() {
    return *d->wraps;
}
void AbstractLineEditor::autoCompletionBegan(String const &)
{}
void AbstractLineEditor::autoCompletionEnded(bool /*accepted*/)
{}
void AbstractLineEditor::updateLineWraps(LineWrapUpdateBehavior behavior) {
    if (behavior == WrapUnlessWrappedAlready && !d->wraps->isEmpty())
        return; // Already wrapped.
    d->updateWraps();
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(AbstractLink) {
    String   tryingToConnectToHost;
    Time     startedTryingAt;
    TimeSpan timeout;
    Address  peerAddress;
    Status   status;
    Time     connectedAt;
    std::unique_ptr<Socket> socket;
    Impl(Public *i)
        : Base(i),
          status(Disconnected),
          connectedAt(Time::invalidTime()) {}
    ~Impl() {
        if (socket.get()) {
            QObject::disconnect(socket.get(), SIGNAL(disconnected()), thisPublic, SLOT(socketDisconnected()));
        }
    }
};
AbstractLink::AbstractLink() : d(new Impl(this))
{}
void AbstractLink::connectDomain(String const &domain, TimeSpan const &timeout) {
    disconnect();
    d->socket.reset(new Socket);
    connect(d->socket.get(), SIGNAL(addressResolved()), this, SIGNAL(addressResolved()));
    connect(d->socket.get(), SIGNAL(connected()), this, SLOT(socketConnected()));
    connect(d->socket.get(), SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    connect(d->socket.get(), SIGNAL(messagesReady()), this, SIGNAL(packetsReady()));
    d->tryingToConnectToHost = domain;
    d->socket->setQuiet(true); // we'll be retrying a few times
    d->socket->open(d->tryingToConnectToHost, DEFAULT_PORT);
    d->status = Connecting;
    d->startedTryingAt = Time();
    d->timeout = timeout;
}
void AbstractLink::connectHost(Address const &address) {
    disconnect();
    d->peerAddress = address;
    d->socket.reset(new Socket);
    connect(d->socket.get(), SIGNAL(connected()), this, SLOT(socketConnected()));
    connect(d->socket.get(), SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    connect(d->socket.get(), SIGNAL(messagesReady()), this, SIGNAL(packetsReady()));
    if (!d->peerAddress.port()) d->peerAddress.setPort(DEFAULT_PORT);
    d->socket->open(d->peerAddress);
    d->status = Connecting;
    d->startedTryingAt = Time();
    d->timeout = 0;
}
void AbstractLink::takeOver(Socket *openSocket) {
    disconnect();
    d->peerAddress = openSocket->peerAddress();
    d->socket.reset(openSocket);
    connect(d->socket.get(), SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
    connect(d->socket.get(), SIGNAL(messagesReady()), this, SIGNAL(packetsReady()));
    d->status = Connected;
    d->connectedAt = Time();
}
void AbstractLink::disconnect() {
    if (d->status != Disconnected) {
        DENG2_ASSERT(d->socket.get() != 0);
        d->timeout = 0;
        d->socket->close(); // emits signal
        d->status = Disconnected;
        QObject::disconnect(d->socket.get(), SIGNAL(addressResolved()), this, SIGNAL(addressResolved()));
        QObject::disconnect(d->socket.get(), SIGNAL(connected()), this, SLOT(socketConnected()));
        QObject::disconnect(d->socket.get(), SIGNAL(disconnected()), this, SLOT(socketDisconnected()));
        QObject::disconnect(d->socket.get(), SIGNAL(messagesReady()), this, SIGNAL(packetsReady()));
    }
}
Address AbstractLink::address() const {
    if (!d->socket) return Address();
    if (d->socket->isOpen()) return d->socket->peerAddress();
    return d->peerAddress;
}
AbstractLink::Status AbstractLink::status() const {
    return d->status;
}
Time AbstractLink::connectedAt() const {
    return d->connectedAt;
}
Packet *AbstractLink::nextPacket() {
    if (!d->socket->hasIncoming()) return 0;
    std::unique_ptr<Message> data(d->socket->receive());
    Packet *packet = interpret(*data.get());
    if (packet) packet->setFrom(data->address());
    return packet;
}
void AbstractLink::send(IByteArray const &data) {
    d->socket->send(data);
}
void AbstractLink::socketConnected() {
    LOG_AS("AbstractLink");
    LOG_NET_VERBOSE("Successfully connected to server %s") << d->socket->peerAddress();
    initiateCommunications();
    d->status = Connected;
    d->connectedAt = Time();
    d->peerAddress = d->socket->peerAddress();
    emit connected();
}
void AbstractLink::socketDisconnected() {
    LOG_AS("AbstractLink");
    if (d->status == Connecting) {
        if (d->startedTryingAt.since() < d->timeout) {
            QTimer::singleShot(500, d->socket.get(), SLOT(reconnect()));
            return;
        }
        d->socket->setQuiet(false);
    }
    else {
        if (!d->peerAddress.isNull()) {
            LOG_NET_NOTE("Disconnected from %s") << d->peerAddress;
        }
        else {
            LOG_NET_NOTE("Disconnected");
        }
    }
    d->status = Disconnected;
    emit disconnected();
    d->connectedAt = Time::invalidTime();
}
}} // namespace de::shell
namespace de { namespace shell {
Action::Action(String const &label) : _event(KeyEvent("")), _label(label), _target(0), _slot(0)
{}
Action::Action(String const &label, QObject *target, char const *slot)
    : _event(KeyEvent("")), _label(label), _target(target), _slot(slot) {
    if (target && slot) {
        connect(this, SIGNAL(triggered()), target, slot);
    }
}
Action::Action(String const &label, KeyEvent const &event, QObject *target, char const *slot)
    : _event(event), _label(label), _target(target), _slot(slot) {
    if (target && slot) {
        connect(this, SIGNAL(triggered()), target, slot);
    }
}
Action::Action(KeyEvent const &event, QObject *target, char const *slot)
    : _event(event), _target(target), _slot(slot) {
    if (target && slot) {
        connect(this, SIGNAL(triggered()), target, slot);
    }
}
Action::~Action()
{}
void Action::setLabel(String const &label) {
    _label = label;
}
String Action::label() const {
    return _label;
}
bool Action::tryTrigger(KeyEvent const &ev) {
    if (ev == _event) {
        trigger();
        return true;
    }
    return false;
}
void Action::trigger() {
    de::Action::trigger();
    emit triggered();
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(ChoiceWidget) {
    Items       items;
    int         selection;
    MenuWidget *menu;
    String      prompt;
    Impl(Public &i) : Base(i), selection(0)
    {}
    void updateMenu() {
        menu->clear();
        foreach (String item, items) {
            menu->appendItem(new Action(item, thisPublic, SLOT(updateSelectionFromMenu())));
        }
        menu->setCursor(selection);
    }
    void updateLabel() {
        self().setLabel(prompt + items[selection], self().attribs());
    }
};
ChoiceWidget::ChoiceWidget(String const &name)
    : LabelWidget(name), d(new Impl(*this)) {
    setBehavior(HandleEventsOnlyWhenFocused);
    setAlignment(AlignLeft);
    d->menu = new MenuWidget(MenuWidget::Popup);
    add(d->menu);
    d->menu->rule()
            .setInput(Rule::Right, rule().right())
            .setInput(Rule::AnchorY, rule().top())
            .setAnchorPoint(Vector2f(0, .5f));
    connect(d->menu, SIGNAL(closed()), this, SLOT(menuClosed()));
}
void ChoiceWidget::setItems(ChoiceWidget::Items const &items) {
    d->items = items;
    d->updateMenu();
    d->updateLabel();
}
void ChoiceWidget::setPrompt(String const &prompt) {
    d->prompt = prompt;
    d->updateLabel();
    redraw();
}
ChoiceWidget::Items ChoiceWidget::items() const {
    return d->items;
}
void ChoiceWidget::select(int pos) {
    d->selection = pos;
    d->menu->setCursor(pos);
    d->updateLabel();
}
int ChoiceWidget::selection() const {
    return d->selection;
}
QList<int> ChoiceWidget::selections() const {
    QList<int> sels;
    sels.append(d->selection);
    return sels;
}
bool ChoiceWidget::isOpen() const {
    return !d->menu->isHidden();
}
Vector2i ChoiceWidget::cursorPosition() const {
    Rectanglei rect = rule().recti();
    return Vector2i(rect.left() + d->prompt.size(), rect.top());
}
void ChoiceWidget::focusLost() {
    setAttribs(TextCanvas::Char::DefaultAttributes);
    setBackgroundAttribs(TextCanvas::Char::DefaultAttributes);
}
void ChoiceWidget::focusGained() {
    setAttribs(TextCanvas::Char::Reverse);
    setBackgroundAttribs(TextCanvas::Char::Reverse);
}
void ChoiceWidget::draw() {
    LabelWidget::draw();
    Rectanglei rect = rule().recti();
    targetCanvas().drawText(rect.topLeft, d->prompt, attribs() | TextCanvas::Char::Bold);
    targetCanvas().put(Vector2i(rect.right() - 1, rect.top()),
                       TextCanvas::Char('>', attribs()));
}
bool ChoiceWidget::handleEvent(Event const &ev) {
    if (ev.type() == Event::KeyPress) {
        KeyEvent const &event = ev.as<KeyEvent>();
        if (!event.text().isEmpty() || event.key() == Qt::Key_Enter) {
            DENG2_ASSERT(!isOpen());
            if (event.text().isEmpty() || event.text() == " ") {
                d->menu->setCursor(d->selection);
            }
            else {
                int curs = d->selection;
                for (int i = 0; i < d->items.size(); ++i) {
                    if (d->items[i].startsWith(event.text(), Qt::CaseInsensitive)) {
                        curs = i;
                        break;
                    }
                }
                d->menu->setCursor(curs);
            }
            remove(*d->menu);
            root().add(d->menu);
            d->menu->open();
            return true;
        }
    }
    return LabelWidget::handleEvent(ev);
}
void ChoiceWidget::updateSelectionFromMenu() {
    DENG2_ASSERT(isOpen());
    d->selection = d->menu->cursor();
    d->updateLabel();
}
void ChoiceWidget::menuClosed() {
    root().setFocus(this);
    root().remove(*d->menu);
    redraw();
    add(d->menu);
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(CommandLineWidget) {
    EditorHistory history;
    Impl(Public *i) : Base(i), history(i) {}
};
CommandLineWidget::CommandLineWidget(String const &name)
    : LineEditWidget(name), d(new Impl(this)) {
    setPrompt("> ");
}
bool CommandLineWidget::handleEvent(Event const &event) {
    DENG2_ASSERT(event.type() == Event::KeyPress);
    KeyEvent const &ev = event.as<KeyEvent>();
    if (ev.key() == Qt::Key_Enter) {
        String const entered = d->history.enter();
        emit commandEntered(entered);
        return true;
    }
    if (LineEditWidget::handleEvent(event)) {
        return true;
    }
    return d->history.handleControlKey(ev.key());
}
void CommandLineWidget::autoCompletionBegan(String const &wordBase) {
    LineEditWidget::autoCompletionBegan(wordBase);
    LOG_MSG("Completions for '%s':") << wordBase;
    LOG_MSG("  %s") << suggestedCompletions().join(", ");
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(DialogWidget) {
    QEventLoop subloop;
};
DialogWidget::DialogWidget(String const &name)
    : TextWidget(name), d(new Impl) {
    hide();
}
void DialogWidget::prepare() {
    show();
    root().setFocus(this);
    redraw();
}
void DialogWidget::finish(int /*result*/)
{
    hide();
    root().setFocus(0);
}
int DialogWidget::exec(TextRootWidget &root) {
    DENG2_ASSERT(!hasRoot());
    root.add(this);
    rule().setInput(Rule::Left,  (root.viewWidth()  - rule().width())  / 2)
          .setInput(Rule::Top,   (root.viewHeight() - rule().height()) / 2);
    prepare();
    int result = d->subloop.exec();
    finish(result);
    root.remove(*this);
    root.requestDraw();
    return result;
}
void DialogWidget::draw() {
    Rectanglei pos = rule().recti().adjusted(Vector2i(-2, -1), Vector2i(2, 1));
    targetCanvas().fill(pos, TextCanvas::Char());
    targetCanvas().drawLineRect(pos);
}
bool DialogWidget::handleEvent(Event const &event) {
    if (event.type() == Event::KeyPress) {
        KeyEvent const &ev = event.as<KeyEvent>();
        if (ev.key() == Qt::Key_Escape) {
            reject();
            return true;
        }
    }
    return true;
}
void DialogWidget::accept(int result) {
    if (d->subloop.isRunning()) {
        d->subloop.exit(result);
        emit accepted(result);
    }
}
void DialogWidget::reject(int result) {
    if (d->subloop.isRunning()) {
        d->subloop.exit(result);
        emit rejected(result);
    }
}
}} // namespace de::shell
namespace de { namespace shell {
static struct {
    char const *name;
    char const *mode;
}
const gameTable[] = {
    { "Shareware DOOM",                         "doom1-share" },
    { "DOOM",                                   "doom1" },
    { "Ultimate DOOM",                          "doom1-ultimate" },
    { "DOOM II",                                "doom2" },
    { "Final DOOM: Plutonia Experiment",        "doom2-plut" },
    { "Final DOOM: TNT Evilution",              "doom2-tnt" },
    { "Chex Quest",                             "chex" },
    { "HacX",                                   "hacx" },
    { "Shareware Heretic",                      "heretic-share" },
    { "Heretic",                                "heretic" },
    { "Heretic: Shadow of the Serpent Riders",  "heretic-ext" },
    { "Hexen v1.1",                             "hexen" },
    { "Hexen v1.0",                             "hexen-v10" },
    { "Hexen: Death Kings of Dark Citadel",     "hexen-dk" },
    { "Hexen Demo",                             "hexen-demo" },
    { nullptr, nullptr }
};
QList<DoomsdayInfo::Game> DoomsdayInfo::allGames() {
    QList<Game> games;
    for (int i = 0; gameTable[i].name; ++i) {
        Game game;
        game.title  = gameTable[i].name;
        game.option = gameTable[i].mode;
        games.append(game);
    }
    return games;
}
String DoomsdayInfo::titleForGame(String const &mode) {
    for (int i = 0; gameTable[i].name; ++i) {
        if (gameTable[i].mode == mode)
            return gameTable[i].name;
    }
    return mode;
}
QList<DoomsdayInfo::GameOption> DoomsdayInfo::gameOptions(String const &gameId) {
    using GOValue = GameOption::Value;
    QList<GameOption> opts;
    opts << GameOption(Choice, "Game type", "server-game-deathmatch %1",
                       GOValue(),
                       QList<GOValue>({ GOValue("0", "Co-op", "coop"),
                                        GOValue("1", "Deathmatch", "dm"),
                                        GOValue("2", "Deathmatch II", "dm2") }));
    opts << GameOption(Choice, "Skill level", "server-game-skill %1",
                       GOValue(),
                       QList<GOValue>({ GOValue("0", "Novice", "skill1"),
                                        GOValue("1", "Easy", "skill2"),
                                        GOValue("2", "Normal", "skill3"),
                                        GOValue("3", "Hard", "skill4"),
                                        GOValue("4", "Nightmare", "skill5") }));
    opts << GameOption(Toggle, "Players can jump", "server-game-jump %1",
                       GOValue(),
                       QList<GOValue>({ GOValue("0"),
                                        GOValue("1", "", "jump") }));
    opts << GameOption(Toggle, "Monsters disabled", "server-game-nomonsters %1",
                       GOValue(),
                       QList<GOValue>({ GOValue("0"),
                                        GOValue("1", "", "nomonst") }));
    if (!gameId.startsWith("hexen")) {
        opts << GameOption(Toggle, "Respawn monsters", "server-game-respawn %1",
                           GOValue(),
                           QList<GOValue>({ GOValue("0"),
                                            GOValue("1", "", "respawn") }));
    }
    if (gameId.startsWith("doom1")) {
        opts << GameOption(Text, "Map", "setmap %1", GOValue("E1M1", "", "mapId"));
    }
    else if (gameId.startsWith("doom2")) {
        opts << GameOption(Text, "Map", "setmap %1", GOValue("MAP01", "", "mapId"));
    }
    else if (gameId.startsWith("heretic")) {
        opts << GameOption(Text, "Map", "setmap %1", GOValue("E1M1", "", "mapId"));
    }
    else if (gameId.startsWith("hexen")) {
        opts << GameOption(Text, "Map", "setmap %1", GOValue("MAP01", "", "mapId"));
    }
    return opts;
}
NativePath DoomsdayInfo::defaultServerRuntimeFolder() {
#ifdef MACOSX
    return QDir::home().filePath("Library/Application Support/Doomsday Engine/server-runtime");
#elif WIN32
    return QDir::home().filePath("AppData/Local/Deng Team/Doomsday Engine/server-runtime");
#else
    return NativePath(QDir::home().filePath(".doomsday")) / "server-runtime";
#endif
}
DoomsdayInfo::GameOption::GameOption(OptionType type, String title,
                                     String command, Value defaultValue,
                                     QList<Value> allowedValues)
    : type(type)
    , title(title)
    , command(command)
    , defaultValue(defaultValue)
    , allowedValues(allowedValues)
{}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(EditorHistory) {
    ITextEditor *editor;
    /**
     * Line of text with a cursor.
     */
    struct Command {
        String text;
        String original; ///< For undoing editing in history.
        int    cursor;   ///< Index in range [0...text.size()]
        Command() : cursor(0) {}
    };
    QList<Command> history;
    int historyPos;
    Impl(Public *i) : Base(i), editor(0), historyPos(0) {
        history.append(Command());
    }
    ~Impl() {}
    void restore(StringList contents) {
        history.clear();
        if (contents.isEmpty()) {
            history.append(Command());
            historyPos = 0;
            return;
        }
        foreach (String line, contents) {
            Command cmd;
            cmd.text = cmd.original = line;
            cmd.cursor = cmd.text.size();
            history.append(cmd);
        }
        historyPos = history.size() - 1;
    }
    Command &command() {
        DENG2_ASSERT(historyPos >= 0 && historyPos < history.size());
        return history[historyPos];
    }
    Command const &command() const {
        DENG2_ASSERT(historyPos >= 0 && historyPos < history.size());
        return history[historyPos];
    }
    void updateCommandFromEditor() {
        DENG2_ASSERT(editor != 0);
        command().text = editor->text();
        command().cursor = editor->cursor();
    }
    void updateEditor() {
        DENG2_ASSERT(editor != 0);
        editor->setText(command().text);
        editor->setCursor(command().cursor);
    }
    bool navigateHistory(int offset) {
        if ((offset < 0 && historyPos >= -offset) ||
           (offset > 0 && historyPos < history.size() - offset)) {
            updateCommandFromEditor();
            historyPos += offset;
            updateEditor();
            return true;
        }
        return false;
    }
    void restoreTextsToOriginal() {
        for (int i = 0; i < history.size(); ++i) {
            Command &cmd = history[i];
            cmd.text = cmd.original;
            cmd.cursor = de::min(cmd.cursor, cmd.text.size());
        }
    }
};
EditorHistory::EditorHistory(ITextEditor *editor) : d(new Impl(this)) {
    d->editor = editor;
}
void EditorHistory::setEditor(ITextEditor &editor) {
    d->editor = &editor;
}
ITextEditor &EditorHistory::editor() {
    DENG2_ASSERT(d->editor != 0);
    return *d->editor;
}
bool EditorHistory::isAtLatest() const {
    return d->historyPos == d->history.size() - 1;
}
void EditorHistory::goToLatest() {
    d->updateCommandFromEditor();
    d->historyPos = d->history.size() - 1;
    d->updateEditor();
}
String EditorHistory::enter() {
    d->updateCommandFromEditor();
    String entered = d->command().text;
    if (!entered.isEmpty()) {
        if (d->historyPos < d->history.size() - 1) {
            if (d->history.last().text.isEmpty()) {
                d->history.removeLast();
            }
            d->history.append(d->command());
        }
        d->history.last().original = entered;
        d->history.append(Impl::Command());
    }
    d->historyPos = d->history.size() - 1;
    d->updateEditor();
    d->restoreTextsToOriginal();
    return entered;
}
bool EditorHistory::handleControlKey(int qtKey) {
    switch (qtKey) {
    case Qt::Key_Up:
        d->navigateHistory(-1);
        return true;
    case Qt::Key_Down:
        d->navigateHistory(+1);
        return true;
    default:
        break;
    }
    return false;
}
StringList EditorHistory::fullHistory(int maxCount) const {
    StringList lines;
    foreach (auto const &cmd, d->history) {
        lines << cmd.original;
        if (maxCount > 0 && lines.count() == maxCount) {
            break;
        }
    }
    return lines;
}
void EditorHistory::setFullHistory(StringList history) {
    d->restore(history);
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(InputDialog) {
    LabelWidget *   label;
    LineEditWidget *edit;
    MenuWidget *    menu;
    String          userText;
    int             result;
    Impl() : label(0), edit(0), menu(0), result(0)
    {}
    ~Impl()
    {}
};
InputDialog::InputDialog(String const &name)
    : DialogWidget(name), d(new Impl) {
    RuleRectangle &rect = rule();
    d->label = new LabelWidget;
    d->label->setExpandsToFitLines(true); // determines height independently
    d->label->rule()
            .setInput(Rule::Width, rect.width())
            .setInput(Rule::Top,   rect.top())
            .setInput(Rule::Left,  rect.left());
    d->edit = new LineEditWidget;
    d->edit->setName(d->edit->uniqueName("edit"));
    d->edit->rule()
            .setInput(Rule::Width, rect.width())
            .setInput(Rule::Left,  rect.left())
            .setInput(Rule::Top,   d->label->rule().bottom() + 1);
    d->menu = new MenuWidget(MenuWidget::AlwaysOpen);
    d->menu->setName(d->menu->uniqueName("menu"));
    d->menu->setBorder(MenuWidget::NoBorder);
    d->menu->setBackgroundAttribs(TextCanvas::Char::DefaultAttributes);
    d->menu->setSelectionAttribs(TextCanvas::Char::Reverse);
    d->menu->appendItem(new Action(tr("OK"), this, SLOT(accept())));
    d->menu->appendItem(new Action(tr("Cancel"), KeyEvent(Qt::Key_C, KeyEvent::Control),
                                   this, SLOT(reject())), "Ctrl-C");
    d->menu->rule()
            .setInput(Rule::Width,  rect.width())
            .setInput(Rule::Left,   rect.left())
            .setInput(Rule::Bottom, rect.bottom());
    add(d->label);
    add(d->edit);
    add(d->menu);
    setFocusCycle(WidgetList() << d->edit << d->menu);
    rect.setInput(Rule::Width, Const(50));
    rect.setInput(Rule::Height,
                  d->menu->rule().height() +
                  d->edit->rule().height() +
                  d->label->rule().height() + 2);
}
LabelWidget &InputDialog::label() {
    return *d->label;
}
LineEditWidget &InputDialog::lineEdit() {
    return *d->edit;
}
MenuWidget &InputDialog::menu() {
    return *d->menu;
}
void InputDialog::setWidth(int width) {
    rule().setInput(Rule::Width, Const(width));
}
void InputDialog::setDescription(String const &desc) {
    d->label->setLabel(desc);
}
void InputDialog::setPrompt(String const &prompt) {
    d->edit->setPrompt(prompt);
}
void InputDialog::setText(String const &text) {
    d->edit->setText(text);
}
void InputDialog::setAcceptLabel(String const &label) {
    d->menu->itemAction(0).setLabel(label);
    redraw();
}
void InputDialog::setRejectLabel(String const &label) {
    d->menu->itemAction(1).setLabel(label);
    redraw();
}
void InputDialog::prepare() {
    DialogWidget::prepare();
    d->userText.clear();
    d->result = 0;
    root().setFocus(d->edit);
}
void InputDialog::finish(int result) {
    d->result = result;
    d->userText.clear();
    if (result) d->userText = d->edit->text();
    DialogWidget::finish(result);
}
String InputDialog::text() const {
    return d->userText;
}
int InputDialog::result() const {
    return d->result;
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(LabelWidget) {
    TextCanvas::Char          background;
    String                    label;
    MonospaceLineWrapping     wraps;
    TextCanvas::Char::Attribs attribs;
    Alignment                 align;
    bool                      vertExpand;
    ConstantRule *            height;
    Impl() : align(0), vertExpand(false) {
        height = new ConstantRule(0);
    }
    ~Impl() {
        releaseRef(height);
    }
    void updateWraps(int width) {
        wraps.wrapTextToWidth(label, width);
        if (vertExpand) height->set(wraps.height());
    }
};
LabelWidget::LabelWidget(String const &name)
    : TextWidget(name), d(new Impl)
{}
void LabelWidget::setBackground(TextCanvas::Char const &background) {
    d->background = background;
}
void LabelWidget::setLabel(String const &text, TextCanvas::Char::Attribs attribs) {
    d->label   = text;
    d->attribs = attribs;
    d->wraps.clear(); // updated later
    redraw();
}
void LabelWidget::setAttribs(TextCanvas::Char::Attribs const &attribs) {
    d->attribs = attribs;
    redraw();
}
void LabelWidget::setBackgroundAttribs(TextCanvas::Char::Attribs const &attribs) {
    d->background.attribs = attribs;
    redraw();
}
TextCanvas::Char::Attribs LabelWidget::attribs() const {
    return d->attribs;
}
void LabelWidget::setAlignment(Alignment align) {
    d->align = align;
    redraw();
}
void LabelWidget::setExpandsToFitLines(bool expand) {
    d->vertExpand = expand;
    if (expand) {
        rule().setInput(Rule::Height, *d->height);
    }
    redraw();
}
String LabelWidget::label() const {
    return d->label;
}
void LabelWidget::update() {
    if (d->wraps.isEmpty()) {
        d->updateWraps(rule().width().valuei());
    }
}
void LabelWidget::draw() {
    Rectanglei pos = rule().recti();
    TextCanvas buf(pos.size());
    buf.clear(d->background);
    DENG2_ASSERT(!d->wraps.isEmpty());
    Vector2i labelSize(d->wraps.width(), d->wraps.height());
    Vector2i labelPos;
    if (d->align.testFlag(AlignRight)) {
        labelPos.x = buf.width() - labelSize.x;
    }
    else if (!d->align.testFlag(AlignLeft)) {
        labelPos.x = buf.width()/2 - labelSize.x/2;
    }
    if (d->align.testFlag(AlignBottom)) {
        labelPos.y = buf.height() - labelSize.y;
    }
    else if (!d->align.testFlag(AlignTop)) {
        labelPos.y = buf.height()/2 - labelSize.y/2;
    }
    buf.drawWrappedText(labelPos, d->label, d->wraps, d->attribs, d->align);
    targetCanvas().draw(buf, pos.topLeft);
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(Lexicon) {
    Terms  terms;
    String extraChars;
    bool   caseSensitive;
    Impl() : caseSensitive(false) {}
};
Lexicon::Lexicon() : d(new Impl)
{}
Lexicon::Lexicon(Lexicon const &other) : d(new Impl(*other.d))
{}
Lexicon &Lexicon::operator = (Lexicon const &other) {
    d.reset(new Impl(*other.d));
    return *this;
}
void Lexicon::setAdditionalWordChars(String const &chars) {
    d->extraChars = chars;
}
void Lexicon::setCaseSensitive(bool sensitive) {
    d->caseSensitive = sensitive;
}
void Lexicon::addTerm(String const &term) {
    d->terms.insert(term);
}
Lexicon::Terms Lexicon::terms() const {
    return d->terms;
}
String Lexicon::additionalWordChars() const {
    return d->extraChars;
}
bool Lexicon::isWordChar(QChar ch) const {
    if (ch.isLetterOrNumber()) return true;
    return d->extraChars.contains(ch);
}
bool Lexicon::isCaseSensitive() const {
    return d->caseSensitive;
}
}} // namespace de::shell
namespace de { namespace shell {
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(LineEditWidget) {
    bool signalOnEnter;
    ConstantRule *height; ///< As rows.
    Impl(Public &i)
        : Base(i),
          signalOnEnter(true) {
        height = new ConstantRule(1);
    }
    ~Impl() {
        releaseRef(height);
    }
};
LineEditWidget::LineEditWidget(de::String const &name)
    : TextWidget(name),
      AbstractLineEditor(new MonospaceLineWrapping),
      d(new Impl(*this)) {
    setBehavior(HandleEventsOnlyWhenFocused);
    rule().setInput(Rule::Height, *d->height);
}
Vector2i LineEditWidget::cursorPosition() const {
    de::Rectanglei pos = rule().recti();
    return pos.topLeft + Vector2i(prompt().size(), 0) + lineCursorPos();
}
void LineEditWidget::viewResized() {
    updateLineWraps(RewrapNow);
}
void LineEditWidget::update() {
    updateLineWraps(WrapUnlessWrappedAlready);
}
void LineEditWidget::draw() {
    Rectanglei pos = rule().recti();
    TextCanvas buf(pos.size());
    TextCanvas::Char::Attribs attr =
            (hasFocus()? TextCanvas::Char::Reverse : TextCanvas::Char::DefaultAttributes);
    buf.clear(TextCanvas::Char(' ', attr));
    buf.drawText(Vector2i(0, 0), prompt(), attr | TextCanvas::Char::Bold);
    if (isSuggestingCompletion()) {
        buf.setRichFormatRange(TextCanvas::Char::Underline, completionRange());
    }
    String txt = text();
    if (echoMode() == PasswordEchoMode) {
        txt = String(txt.size(), '*');
    }
    buf.drawWrappedText(Vector2i(prompt().size(), 0), txt, lineWraps(), attr);
    targetCanvas().draw(buf, pos.topLeft);
}
bool LineEditWidget::handleEvent(Event const &event) {
    DENG2_ASSERT(event.type() == Event::KeyPress);
    KeyEvent const &ev = event.as<KeyEvent>();
    bool eaten = true;
    if (!ev.text().isEmpty()) {
        insert(ev.text());
    }
    else {
        eaten = handleControlKey(ev.key());
    }
    if (eaten) return true;
    return TextWidget::handleEvent(event);
}
bool LineEditWidget::handleControlKey(int qtKey, KeyModifiers const &mods) {
    if (AbstractLineEditor::handleControlKey(qtKey, mods)) {
        if (qtKey == Qt::Key_Enter) {
            if (d->signalOnEnter) {
                emit enterPressed(text());
            }
            else {
                return false;
            }
        }
        return true; // Handled.
    }
    return false;
}
void LineEditWidget::setSignalOnEnter(int enterSignal) {
    d->signalOnEnter = enterSignal;
}
int LineEditWidget::maximumWidth() const {
    return int(rule().recti().width()) - int(prompt().size()) - 1;
}
void LineEditWidget::numberOfLinesChanged(int lineCount) {
    d->height->set(lineCount);
}
void LineEditWidget::contentChanged() {
    if (hasRoot()) {
        updateLineWraps(RewrapNow);
    }
    redraw();
}
void LineEditWidget::cursorMoved() {
    redraw();
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(Link) {
    String   connectDomain;
    TimeSpan connectTimeout;
    Address  connectAddress;
    Protocol protocol;
    Impl(Public *i) : Base(i)
    {}
};
Link::Link(String const &domain, TimeSpan const &timeout) : d(new Impl(this)) {
    d->connectDomain  = domain;
    d->connectTimeout = timeout;
}
Link::Link(Address const &address) : d(new Impl(this)) {
    d->connectAddress = address;
}
Link::Link(Socket *openSocket) : d(new Impl(this)) {
    takeOver(openSocket);
}
Protocol &Link::protocol() {
    return d->protocol;
}
void Link::connectLink() {
    if (!d->connectDomain.isEmpty()) {
        connectDomain(d->connectDomain, d->connectTimeout);
    }
    else if (!d->connectAddress.isNull()) {
        connectHost(d->connectAddress);
    }
    else {
        throw ConnectError("Link::connect", "Host to connect to not specified");
    }
}
Packet *Link::interpret(Message const &msg) {
    return d->protocol.interpret(msg);
}
void Link::initiateCommunications() {
    *this << ByteRefArray("Shell", 5);
}
}} // namespace de::shell
namespace de { namespace shell {
/**
 * Log sink for incoming entries (local and remote). Rather than formatting the
 * entries immediately, we keep a copy of them for formatting prior to drawing.
 */
class Sink : public MemoryLogSink {
public:
    Sink(LogWidget &widget) : MemoryLogSink(), _widget(widget) {}
    void addedNewEntry(LogEntry &) {
        _widget.root().requestDraw();
    }
private:
    LogWidget &_widget;
};
DENG2_PIMPL(LogWidget) {
    Sink sink;
    MonospaceLogSinkFormatter formatter;
    unsigned int cacheWidth;
    QList<TextCanvas *> cache; ///< Indices match entry indices in sink.
    int maxEntries;
    int visibleOffset;
    bool showScrollIndicator;
    int lastMaxScroll;
    Impl(Public *inst)
        : Base(inst),
          sink(*inst),
          cacheWidth(0),
          maxEntries(1000),
          visibleOffset(0),
          showScrollIndicator(true),
          lastMaxScroll(0)
    {}
    ~Impl() {
        clearCache();
    }
    void clear() {
        sink.clear();
        clearCache();
    }
    void clearCache() {
        qDeleteAll(cache);
        cache.clear();
    }
    void prune() {
        int excess = sink.entryCount() - maxEntries;
        if (excess > 0) {
            sink.remove(0, excess);
        }
        while (excess-- > 0 && !cache.isEmpty()) {
            delete cache.takeFirst();
        }
    }
    int totalHeight() {
        int total = 0;
        for (int idx = sink.entryCount() - 1; idx >= 0; --idx) {
            total += cache[idx]->size().y;
        }
        return total;
    }
    int maxVisibleOffset(int visibleHeight) {
        return de::max(0, totalHeight() - visibleHeight);
    }
    void clampVisibleOffset(int visibleHeight) {
        setVisibleOffset(de::min(visibleOffset, maxVisibleOffset(visibleHeight)));
    }
    void setVisibleOffset(int off) {
        if (visibleOffset != off) {
            visibleOffset = off;
            emit self().scrollPositionChanged(off);
        }
    }
};
LogWidget::LogWidget(String const &name) : TextWidget(name), d(new Impl(this))
{}
LogSink &LogWidget::logSink() {
    return d->sink;
}
void LogWidget::clear() {
    d->clear();
    redraw();
}
void LogWidget::setScrollIndicatorVisible(bool visible) {
    d->showScrollIndicator = visible;
}
int LogWidget::scrollPosition() const {
    return d->visibleOffset;
}
int LogWidget::scrollPageSize() const {
    return de::max(1, rule().height().valuei() - 1);
}
int LogWidget::maximumScroll() const {
    return d->lastMaxScroll;
}
void LogWidget::scroll(int to) {
    d->visibleOffset = de::max(0, to);
    redraw();
}
void LogWidget::draw() {
    Rectanglei pos = rule().recti();
    TextCanvas buf(pos.size());
    if (d->cacheWidth != pos.width()) {
        d->cacheWidth = pos.width();
        d->formatter.setMaxLength(d->cacheWidth);
        d->clearCache();
    }
    d->sink.lock();
    while (d->cache.size() < d->sink.entryCount()) {
        int idx = d->cache.size();
        LogEntry const &entry = d->sink.entry(idx);
        MonospaceLogSinkFormatter::Lines lines = d->formatter.logEntryToTextLines(entry);
        TextCanvas *buf = new TextCanvas(Vector2ui(pos.width(), lines.size()));
        d->cache.append(buf);
        TextCanvas::Char::Attribs attribs = (entry.flags() & LogEntry::Remote?
                TextCanvas::Char::DefaultAttributes : TextCanvas::Char::Bold);
        for (int i = 0; i < lines.size(); ++i) {
            buf->drawText(Vector2i(0, i), lines[i], attribs);
        }
        if (d->visibleOffset > 0) {
            d->setVisibleOffset(d->visibleOffset + lines.size());
        }
    }
    DENG2_ASSERT(d->cache.size() == d->sink.entryCount());
    d->clampVisibleOffset(buf.height());
    int yBottom = buf.height() + d->visibleOffset;
    for (int idx = d->sink.entryCount() - 1; yBottom > 0 && idx >= 0; --idx) {
        TextCanvas *canvas = d->cache[idx];
        yBottom -= canvas->size().y;
        if (yBottom < buf.height()) {
            buf.draw(*canvas, Vector2i(0, yBottom));
        }
    }
    int const maxScroll = d->maxVisibleOffset(buf.height());
    if (d->showScrollIndicator && d->visibleOffset > 0) {
        int const indHeight = de::clamp(2, de::floor(float(buf.height() * buf.height()) /
                                               float(d->totalHeight())), buf.height() / 2);
        float const indPos = float(d->visibleOffset) / float(maxScroll);
        int const avail = buf.height() - indHeight;
        for (int i = 0; i < indHeight; ++i) {
            buf.put(Vector2i(buf.width() - 1, i + avail - indPos * avail),
                    TextCanvas::Char(':', TextCanvas::Char::Reverse));
        }
    }
    targetCanvas().draw(buf, pos.topLeft);
    d->prune();
    d->sink.unlock();
    if (d->lastMaxScroll != maxScroll) {
        d->lastMaxScroll = maxScroll;
        emit scrollMaxChanged(maxScroll);
    }
}
bool LogWidget::handleEvent(Event const &event) {
    if (event.type() != Event::KeyPress) return false;
    KeyEvent const &ev = event.as<KeyEvent>();
    int pageSize = scrollPageSize();
    switch (ev.key()) {
    case Qt::Key_PageUp:
        d->setVisibleOffset(d->visibleOffset + pageSize);
        redraw();
        return true;
    case Qt::Key_PageDown:
        d->setVisibleOffset(de::max(0, d->visibleOffset - pageSize));
        redraw();
        return true;
    default:
        break;
    }
    return TextWidget::handleEvent(event);
}
void LogWidget::scrollToBottom() {
    d->setVisibleOffset(0);
    redraw();
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL(MenuWidget) {
    ConstantRule *            width;
    ConstantRule *            height;
    TextCanvas::Char::Attribs borderAttr;
    TextCanvas::Char::Attribs backgroundAttr;
    TextCanvas::Char::Attribs selectionAttr;
    BorderStyle               borderStyle;
    Vector2i                  cursorPos; ///< Visual position.
    bool                      closable;
    bool                      cycleCursor;
    struct Item {
        Action *action;
        String shortcutLabel;
        bool separatorAfter;
        Item() : action(0), separatorAfter(false) {}
        Item(Item const &other)
            : action(holdRef(other.action))
            , shortcutLabel(other.shortcutLabel)
            , separatorAfter(other.separatorAfter) {}
        ~Item() {
            releaseRef(action);
        }
        Item &operator = (Item const &other) {
            changeRef(action, other.action);
            shortcutLabel = other.shortcutLabel;
            separatorAfter = other.separatorAfter;
            return *this;
        }
    };
    QList<Item> items;
    int cursor;
    Impl(Public &i)
        : Base(i),
          borderAttr(TextCanvas::Char::Reverse),
          backgroundAttr(TextCanvas::Char::Reverse),
          borderStyle(LineBorder),
          closable(true),
          cycleCursor(true),
          cursor(0) {
        width  = new ConstantRule(1);
        height = new ConstantRule(1);
    }
    ~Impl() {
        clear();
        releaseRef(width);
        releaseRef(height);
    }
    void clear() {
        foreach (Item i, items) {
            self().removeAction(*i.action);
        }
        items.clear();
        updateSize();
    }
    void updateSize() {
        int cols = 0;
        int lines = (borderStyle == NoBorder? 0 : 2);
        foreach (Item const &item, items) {
            lines++;
            if (item.separatorAfter) lines++;
            int w = item.action->label().size();
            if (!item.shortcutLabel.isEmpty()) {
                w += 1 + item.shortcutLabel.size();
            }
            cols = de::max(w, cols);
        }
        height->set(lines);
        width->set(4 + cols + (borderStyle == NoBorder? 0 : 2)); // cursor
    }
    void removeItem(int pos) {
        self().removeAction(*items[pos].action);
        items.removeAt(pos);
        updateSize();
    }
};
MenuWidget::MenuWidget(Preset preset, String const &name)
    : TextWidget(name), d(new Impl(*this)) {
    switch (preset) {
    case Popup:
        setBehavior(HandleEventsOnlyWhenFocused);
        setClosable(true);
        d->cycleCursor = true;
        hide();
        break;
    case AlwaysOpen:
        setClosable(false);
        d->cycleCursor = false;
        break;
    }
    rule().setSize(*d->width, *d->height);
}
int MenuWidget::itemCount() const {
    return d->items.size();
}
void MenuWidget::appendItem(RefArg<Action> action, String const &shortcutLabel) {
    Impl::Item item;
    item.action = action.holdRef();
    item.shortcutLabel = shortcutLabel;
    d->items.append(item);
    d->updateSize();
    redraw();
    addAction(action);
}
void MenuWidget::appendSeparator() {
    if (d->items.isEmpty()) return;
    d->items.last().separatorAfter = true;
    d->updateSize();
    redraw();
}
void MenuWidget::insertItem(int pos, RefArg<Action> action, String const &shortcutLabel) {
    Impl::Item item;
    item.action = action.holdRef();
    item.shortcutLabel = shortcutLabel;
    d->items.insert(pos, item);
    d->updateSize();
    redraw();
    addAction(action);
}
void MenuWidget::insertSeparator(int pos) {
    if (pos < 0 || pos >= d->items.size()) return;
    d->items[pos].separatorAfter = true;
    d->updateSize();
    redraw();
}
void MenuWidget::clear() {
    d->clear();
    redraw();
}
void MenuWidget::removeItem(int pos) {
    d->removeItem(pos);
    redraw();
}
Action &MenuWidget::itemAction(int pos) const {
    return *d->items[pos].action;
}
int MenuWidget::findLabel(String const &label) const {
    for (int i = 0; i < d->items.size(); ++i) {
        if (!d->items[i].action->label().compareWithoutCase(label))
            return i;
    }
    return -1;
}
bool MenuWidget::hasLabel(String const &label) const {
    return findLabel(label) >= 0;
}
void MenuWidget::setCursor(int pos) {
    d->cursor = de::min(pos, itemCount() - 1);
    redraw();
}
void MenuWidget::setCursorByLabel(String const &label) {
    int idx = findLabel(label);
    if (idx >= 0) {
        setCursor(idx);
    }
    else {
        setCursor(d->cursor);
    }
}
int MenuWidget::cursor() const {
    return d->cursor;
}
void MenuWidget::setClosable(bool canBeClosed) {
    d->closable = canBeClosed;
}
void MenuWidget::setSelectionAttribs(TextCanvas::Char::Attribs const &attribs) {
    d->selectionAttr = attribs;
    redraw();
}
void MenuWidget::setBackgroundAttribs(TextCanvas::Char::Attribs const &attribs) {
    d->backgroundAttr = attribs;
    redraw();
}
void MenuWidget::setBorder(MenuWidget::BorderStyle style) {
    d->borderStyle = style;
    redraw();
}
void MenuWidget::setBorderAttribs(TextCanvas::Char::Attribs const &attribs) {
    d->borderAttr = attribs;
    redraw();
}
Vector2i MenuWidget::cursorPosition() const {
    return d->cursorPos;
}
void MenuWidget::open() {
    DENG2_ASSERT(hasRoot());
    root().setFocus(this);
    show();
    redraw();
}
void MenuWidget::close() {
    if (d->closable) {
        DENG2_ASSERT(hasRoot());
        root().setFocus(0);
        emit closed();
        hide();
        redraw();
    }
}
void MenuWidget::draw() {
    Rectanglei pos = rule().recti();
    TextCanvas buf(pos.size());
    buf.clear(TextCanvas::Char(' ', d->backgroundAttr));
    int const border = (d->borderStyle == NoBorder? 0 : 1);
    int y = border;
    for (int i = 0; i < d->items.size(); ++i) {
        Impl::Item const &item = d->items[i];
        TextCanvas::Char::Attribs itemAttr = (d->cursor == i && hasFocus()?
                                              d->selectionAttr : d->backgroundAttr);
        if (d->cursor == i) {
            buf.fill(Rectanglei(Vector2i(border, y), Vector2i(pos.width() - border, y + 1)),
                     TextCanvas::Char(' ', itemAttr));
            d->cursorPos = Vector2i(border + 1, y);
            buf.put(d->cursorPos, TextCanvas::Char('*', itemAttr));
            d->cursorPos += pos.topLeft;
        }
        buf.drawText(Vector2i(border + 3, y), item.action->label(),
                     itemAttr | (d->cursor == i? TextCanvas::Char::Bold : TextCanvas::Char::DefaultAttributes));
        if (item.shortcutLabel) {
            buf.drawText(Vector2i(buf.width() - 1 - border - item.shortcutLabel.size(), y),
                         item.shortcutLabel, itemAttr);
        }
        y++;
        if (item.separatorAfter) {
            buf.fill(Rectanglei(Vector2i(border, y), Vector2i(pos.width() - border, y + 1)),
                     TextCanvas::Char('-', d->borderAttr));
            y++;
        }
    }
    if (d->borderStyle == LineBorder) {
        buf.drawLineRect(buf.rect(), d->borderAttr);
    }
    targetCanvas().draw(buf, pos.topLeft);
}
bool MenuWidget::handleEvent(Event const &event) {
    if (!itemCount()) return false;
    if (event.type() != Event::KeyPress) return false;
    KeyEvent const &ev = event.as<KeyEvent>();
    if (ev.text().isEmpty()) {
        switch (ev.key()) {
        case Qt::Key_Up:
            if (d->cursor == 0) {
                if (!d->cycleCursor) break;
                d->cursor = itemCount() - 1;
            }
            else {
                d->cursor--;
            }
            redraw();
            return true;
        case Qt::Key_Down:
            if (d->cursor == itemCount() - 1) {
                if (!d->cycleCursor) break;
                d->cursor = 0;
            }
            else {
                d->cursor++;
            }
            redraw();
            return true;
        case Qt::Key_Home:
        case Qt::Key_PageUp:
            d->cursor = 0;
            redraw();
            return true;
        case Qt::Key_End:
        case Qt::Key_PageDown:
            d->cursor = itemCount() - 1;
            redraw();
            return true;
        case Qt::Key_Enter:
            itemAction(d->cursor).trigger();
            close();
            return true;
        default:
            break;
        }
    }
    if (ev.text() == " ") {
        itemAction(d->cursor).trigger();
        close();
        return true;
    }
    if (TextWidget::handleEvent(event)) {
        close();
        return true;
    }
    if (ev.text().isEmpty()) {
        if (d->closable) {
            close();
            return true;
        }
    }
    else {
        for (int i = 0; i < d->items.size(); ++i) {
            int idx = (d->cursor + i + 1) % d->items.size();
            if (d->items[idx].action->label().startsWith(ev.text(), Qt::CaseInsensitive)) {
                setCursor(idx);
                return true;
            }
        }
    }
    return false;
}
}} // namespace de::shell
namespace de { namespace shell {
MonospaceLineWrapping::MonospaceLineWrapping() {
}
bool MonospaceLineWrapping::isEmpty() const {
    return _lines.isEmpty();
}
void MonospaceLineWrapping::clear() {
    _lines.clear();
}
void MonospaceLineWrapping::wrapTextToWidth(String const &text, int maxWidth) {
    QChar const newline('\n');
    clear();
    if (maxWidth < 1) return; // No room to wrap.
    int const lineWidth = maxWidth;
    int begin = 0;
    forever {
        int end = begin;
        while (end < begin + lineWidth &&
              end < text.size() && text.at(end) != newline) ++end;
        if (end == text.size()) {
            _lines.append(WrappedLine(Rangei(begin, text.size())));
            break;
        }
        while (!text.at(end).isSpace()) {
            --end;
            if (end == begin) {
                end = begin + lineWidth;
                break;
            }
        }
        if (text.at(end) == newline) {
            _lines.append(WrappedLine(Rangei(begin, end)));
            begin = end + 1;
        }
        else {
            if (text.at(end).isSpace()) ++end;
            _lines.append(WrappedLine(Rangei(begin, end)));
            begin = end;
        }
    }
    _lines.last().isFinal = true;
}
int MonospaceLineWrapping::width() const {
    int w = 0;
    for (int i = 0; i < _lines.size(); ++i) {
        WrappedLine const &span = _lines[i];
        w = de::max(w, span.range.size());
    }
    return w;
}
int MonospaceLineWrapping::height() const {
    return _lines.size();
}
int MonospaceLineWrapping::rangeWidth(Rangei const &range) const {
    return range.size();
}
int MonospaceLineWrapping::indexAtWidth(Rangei const &range, int width) const {
    if (width <= range.size()) {
        return range.start + width;
    }
    return range.end;
}
}} // namespace de::shell
namespace de { namespace shell {
static String const PATH_REMOTE_PACKS  = "/remote/packs";
static String const PATH_REMOTE_SERVER = "/remote/server"; // local folder for RemoteFeed
DENG2_PIMPL(PackageDownloader)
, DENG2_OBSERVES(filesys::RemoteFeedRelay, Status)
, DENG2_OBSERVES(Asset, StateChange)
, DENG2_OBSERVES(RemoteFile, Download)
, DENG2_OBSERVES(Deletable, Deletion) {
    String                           fileRepository;
    MountCallback                    afterConnected;
    bool                             isCancelled  = false;
    dint64                           totalBytes   = 0;
    int                              numDownloads = 0;
    AssetGroup                       downloads;
    QHash<IDownloadable *, Rangei64> downloadBytes;
    std::function<void()>            postDownloadCallback;
    LoopCallback                     deferred;
    Impl(Public *i) : Base(i) {}
    void remoteRepositoryStatusChanged(String const &address,
                                       filesys::RemoteFeedRelay::Status) override {
        if (address == fileRepository) {
            deferred.enqueue([this] () {
                auto *relay = &filesys::RemoteFeedRelay::get();
                relay->audienceForStatus() -= this;
                Folder::afterPopulation([this, relay] () {
                    if (afterConnected) {
                        afterConnected(relay->repository(fileRepository));
                    }
                });
            });
        }
    }
    void downloadFile(File &file) {
        if (auto *folder = maybeAs<Folder>(file)) {
            folder->forContents([this] (String, File &f) {
                downloadFile(f);
                return LoopContinue;
            });
        }
        if (auto *dl = maybeAs<IDownloadable>(file)) {
            LOG_NET_VERBOSE("Downloading from server: %s") << file.description();
            downloads.insert(dl->asset());
            dl->audienceForDownload += this;
            file.Deletable::audienceForDeletion += this;
            downloadBytes.insert(dl, Rangei64(dint64(dl->downloadSize()),
                                              dint64(dl->downloadSize())));
            numDownloads++;
            totalBytes += dl->downloadSize();
            isCancelled = false;
            dl->download();
        }
    }
    void downloadProgress(IDownloadable &dl, dsize remainingBytes) override {
        DENG2_ASSERT_IN_MAIN_THREAD();
        auto found = downloadBytes.find(&dl);
        if (found == downloadBytes.end()) return;
        found.value().start = dint64(remainingBytes);
        if (remainingBytes == 0) {
            dl.audienceForDownload -= this;
            maybeAs<File>(dl)->Deletable::audienceForDeletion -= this;
            downloadBytes.erase(found);
        } {
            dint64 totalRemaining = 0;
            foreach (auto bytes, downloadBytes) {
                totalRemaining += bytes.start;
            }
            DENG2_FOR_PUBLIC_AUDIENCE2(Status, i) {
                i->downloadStatusUpdate(Rangei64(totalRemaining, totalBytes),
                                        Rangei(downloadBytes.size(), numDownloads));
            }
        }
    }
    void objectWasDeleted(Deletable *del) override {
        DENG2_ASSERT_IN_MAIN_THREAD();
        downloadBytes.remove(static_cast<RemoteFile *>(del));
    }
    void assetStateChanged(Asset &) override {
        if (downloads.isReady()) {
            LOG_NET_VERBOSE(isCancelled? "Remote file downloads cancelled"
                                       : "All downloads of remote files finished");
            Loop::mainCall([this] () {
                DENG2_ASSERT(downloadBytes.isEmpty());
                if (postDownloadCallback) postDownloadCallback();
            });
        }
    }
    void finishDownloads() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Status, i) {
            i->downloadStatusUpdate(Rangei64(0, totalBytes), Rangei(0, numDownloads));
        }
        numDownloads = 0;
        totalBytes = 0;
        downloads.clear();
    }
    void clearDownloads() {
        for (auto i = downloadBytes.begin(); i != downloadBytes.end(); ++i) {
            auto *dl = i.key();
            dl->cancelDownload();
            dl->audienceForDownload -= this;
            maybeAs<File>(dl)->Deletable::audienceForDeletion -= this;
        }
        numDownloads = 0;
        totalBytes = 0;
        downloadBytes.clear();
        downloads.clear();
    }
    /**
     * Makes remote packages available for loading locally.
     *
     * Once remote files have been downloaded, PackageLoader still needs to be made aware
     * that the packages are available. This is done via link files that have the ".pack"
     * extension and thus are treated as loadable packages.
     *
     * @param ids  Identifiers of remote packages that have been downloaded and are now
     *             being prepared for loading.
     */
    void linkRemotePackages(filesys::PackagePaths const &pkgPaths) {
        Folder &remotePacks = FS::get().makeFolder(PATH_REMOTE_PACKS);
        for (auto i = pkgPaths.begin(); i != pkgPaths.end(); ++i) {
            LOG_RES_VERBOSE("Registering remote package \"%s\"") << i.key();
            if (auto *file = FS::tryLocate<File>(i.value().localPath)) {
                LOGDEV_RES_VERBOSE("Cached metadata:\n") << file->objectNamespace().asText();
                auto *pack = LinkFile::newLinkToFile(*file, file->name() + ".pack");
                Record &meta = pack->objectNamespace();
                meta.add("package", new Record(file->objectNamespace().subrecord("package")));
                meta.set("package.path", file->path());
                remotePacks.add(pack);
                FS::get().index(*pack);
                LOG_RES_VERBOSE("\"%s\" linked as ") << i.key() << pack->path();
            }
        }
    }
    void unlinkRemotePackages() {
        if (std::unique_ptr<Folder> remotePacks { FS::tryLocate<Folder>(PATH_REMOTE_PACKS) }) {
            remotePacks->forContents([] (String, File &file) {
                LOG_RES_VERBOSE("Unloading remote package: ") << file.description();
                PackageLoader::get().unload(Package::identifierForFile(file));
                return LoopContinue;
            });
        }
    }
    DENG2_PIMPL_AUDIENCE(Status)
};
DENG2_AUDIENCE_METHOD(PackageDownloader, Status)
PackageDownloader::PackageDownloader()
    : d(new Impl(this))
{}
String PackageDownloader::fileRepository() const {
    return d->fileRepository;
}
void PackageDownloader::cancel() {
    d->isCancelled = true;
    DENG2_FOR_AUDIENCE2(Status, i) {
        i->downloadStatusUpdate(Rangei64(), Rangei());
    }
    d->clearDownloads();
}
bool PackageDownloader::isCancelled() const {
    return d->isCancelled;
}
bool PackageDownloader::isActive() const {
    return !d->downloads.isEmpty() && !d->downloads.isReady();
}
void PackageDownloader::mountServerRepository(shell::ServerInfo const &info,
                                              MountCallback afterConnected) {
    if (info.version() > Version(2, 1, 0, 2484)) {
        auto &relay = filesys::RemoteFeedRelay::get();
        d->fileRepository = filesys::NativeLink::URL_SCHEME + info.address().asText();
        d->isCancelled    = false;
        relay.addRepository(d->fileRepository, PATH_REMOTE_SERVER);
        d->afterConnected = afterConnected;
        relay.audienceForStatus() += d;
    }
    else if (afterConnected) {
        afterConnected(nullptr);
    }
}
void PackageDownloader::unmountServerRepository() {
    d->clearDownloads();
    d->unlinkRemotePackages();
    filesys::RemoteFeedRelay::get().removeRepository(d->fileRepository);
    d->fileRepository.clear();
    d->isCancelled = false;
    if (Folder *remoteFiles = FS::tryLocate<Folder>(PATH_REMOTE_SERVER)) {
        trash(remoteFiles);
    }
}
void PackageDownloader::download(StringList packageIds, std::function<void ()> callback) {
    d->downloads.clear();
    auto const pkgPaths = filesys::RemoteFeedRelay::get().locatePackages(packageIds);
    StringList const foundIds = pkgPaths.keys();
    for (auto found = pkgPaths.begin(); found != pkgPaths.end(); ++found) {
        if (File *file = found.value().link->populateRemotePath(found.key(), found.value())) {
            d->downloadFile(*file);
        }
    }
    auto finished = [this, pkgPaths, callback] () {
        d->downloads.audienceForStateChange() -= d;
        d->finishDownloads();
        d->linkRemotePackages(pkgPaths);
        callback();
    };
    if (d->downloads.isReady()) {
        d->postDownloadCallback = nullptr;
        finished();
    }
    else {
        d->postDownloadCallback = finished;
        d->downloads.audienceForStateChange() += d;
    }
}
}} // namespace de::shell
namespace de { namespace shell {
static String const PT_COMMAND    = "shell.command";
static String const PT_LEXICON    = "shell.lexicon";
static String const PT_GAME_STATE = "shell.game.state";
static Packet::Type const CHALLENGE_PACKET_TYPE = Packet::typeFromString("Psw?");
ChallengePacket::ChallengePacket() : Packet(CHALLENGE_PACKET_TYPE)
{}
Packet *ChallengePacket::fromBlock(Block const &block) {
    return constructFromBlock<ChallengePacket>(block, CHALLENGE_PACKET_TYPE);
}
static Packet::Type const LOG_ENTRY_PACKET_TYPE = Packet::typeFromString("LgEn");
LogEntryPacket::LogEntryPacket() : Packet(LOG_ENTRY_PACKET_TYPE)
{}
LogEntryPacket::~LogEntryPacket() {
    clear();
}
void LogEntryPacket::clear() {
    foreach (LogEntry *e, _entries) delete e;
    _entries.clear();
}
bool LogEntryPacket::isEmpty() const {
    return _entries.isEmpty();
}
void LogEntryPacket::add(LogEntry const &entry) {
    _entries.append(new LogEntry(entry));
}
LogEntryPacket::Entries const &LogEntryPacket::entries() const {
    return _entries;
}
void LogEntryPacket::execute() const {
    LogBuffer &buf = LogBuffer::get();
    foreach (LogEntry *e, _entries) {
        buf.add(new LogEntry(*e, LogEntry::Remote));
    }
}
void LogEntryPacket::operator >> (Writer &to) const {
    Packet::operator >> (to);
    to.writeObjects(_entries);
}
void LogEntryPacket::operator << (Reader &from) {
    _entries.clear();
    Packet::operator << (from);
    from.readObjects<LogEntry>(_entries);
}
Packet *LogEntryPacket::fromBlock(Block const &block) {
    return constructFromBlock<LogEntryPacket>(block, LOG_ENTRY_PACKET_TYPE);
}
static Packet::Type const PLAYER_INFO_PACKET_TYPE = Packet::typeFromString("PlrI");
DENG2_PIMPL_NOREF(PlayerInfoPacket) {
    Players players;
};
PlayerInfoPacket::PlayerInfoPacket()
    : Packet(PLAYER_INFO_PACKET_TYPE), d(new Impl)
{}
void PlayerInfoPacket::add(Player const &player) {
    d->players.insert(player.number, player);
}
int PlayerInfoPacket::count() const {
    return d->players.size();
}
PlayerInfoPacket::Player const &PlayerInfoPacket::player(int number) const {
    DENG2_ASSERT(d->players.contains(number));
    return d->players[number];
}
PlayerInfoPacket::Players PlayerInfoPacket::players() const {
    return d->players;
}
void PlayerInfoPacket::operator >> (Writer &to) const {
    Packet::operator >> (to);
    to << duint32(d->players.size());
    foreach (Player const &p, d->players) {
        to << dbyte(p.number) << p.position << p.name << p.color;
    }
}
void PlayerInfoPacket::operator << (Reader &from) {
    d->players.clear();
    Packet::operator << (from);
    duint32 count;
    from >> count;
    while (count-- > 0) {
        Player p;
        from.readAs<dbyte>(p.number) >> p.position >> p.name >> p.color;
        d->players.insert(p.number, p);
    }
}
Packet *PlayerInfoPacket::fromBlock(Block const &block) {
    return constructFromBlock<PlayerInfoPacket>(block, PLAYER_INFO_PACKET_TYPE);
}
static Packet::Type const MAP_OUTLINE_PACKET_TYPE = Packet::typeFromString("MpOL");
DENG2_PIMPL_NOREF(MapOutlinePacket) {
    QList<Line> lines;
};
MapOutlinePacket::MapOutlinePacket()
    : Packet(MAP_OUTLINE_PACKET_TYPE), d(new Impl)
{}
void MapOutlinePacket::clear() {
    d->lines.clear();
}
void MapOutlinePacket::addLine(Vector2i const &vertex1, Vector2i const &vertex2, LineType type) {
    Line ln;
    ln.start = vertex1;
    ln.end   = vertex2;
    ln.type  = type;
    d->lines.append(ln);
}
int MapOutlinePacket::lineCount() const {
    return d->lines.size();
}
MapOutlinePacket::Line const &MapOutlinePacket::line(int index) const {
    DENG2_ASSERT(index >= 0 && index < d->lines.size());
    return d->lines[index];
}
void MapOutlinePacket::operator >> (Writer &to) const {
    Packet::operator >> (to);
    to << duint32(d->lines.size());
    foreach (Line const &ln, d->lines) {
        to << ln.start << ln.end << dbyte(ln.type);
    }
}
void MapOutlinePacket::operator << (Reader &from) {
    clear();
    Packet::operator << (from);
    duint32 count;
    from >> count;
    while (count-- > 0) {
        Line ln;
        from >> ln.start >> ln.end;
        from.readAs<dbyte>(ln.type);
        d->lines.append(ln);
    }
}
Packet *MapOutlinePacket::fromBlock(Block const &block) {
    return constructFromBlock<MapOutlinePacket>(block, MAP_OUTLINE_PACKET_TYPE);
}
Protocol::Protocol() {
    define(ChallengePacket::fromBlock);
    define(LogEntryPacket::fromBlock);
    define(MapOutlinePacket::fromBlock);
    define(PlayerInfoPacket::fromBlock);
}
Protocol::PacketType Protocol::recognize(Packet const *packet) {
    if (packet->type() == CHALLENGE_PACKET_TYPE) {
        DENG2_ASSERT(is<ChallengePacket>(packet));
        return PasswordChallenge;
    }
    if (packet->type() == LOG_ENTRY_PACKET_TYPE) {
        DENG2_ASSERT(is<LogEntryPacket>(packet));
        return LogEntries;
    }
    if (packet->type() == MAP_OUTLINE_PACKET_TYPE) {
        DENG2_ASSERT(is<MapOutlinePacket>(packet));
        return MapOutline;
    }
    if (packet->type() == PLAYER_INFO_PACKET_TYPE) {
        DENG2_ASSERT(is<PlayerInfoPacket>(packet));
        return PlayerInfo;
    }
    if (RecordPacket const *rec = maybeAs<RecordPacket>(packet)) {
        if (rec->name() == PT_COMMAND) {
            return Command;
        }
        else if (rec->name() == PT_LEXICON) {
            return ConsoleLexicon;
        }
        else if (rec->name() == PT_GAME_STATE) {
            return GameState;
        }
    }
    return Unknown;
}
Block Protocol::passwordResponse(String const &plainPassword) {
    Block response;
    response += "Shell";
    response += QCryptographicHash::hash(plainPassword.toUtf8(),
                                         QCryptographicHash::Sha1);
    return response;
}
RecordPacket *Protocol::newCommand(String const &command) {
    RecordPacket *cmd = new RecordPacket(PT_COMMAND);
    cmd->record().addText("execute", command);
    return cmd;
}
static RecordPacket const &asRecordPacket(Packet const &packet, Protocol::PacketType type) {
    RecordPacket const *rec = dynamic_cast<RecordPacket const *>(&packet);
    DENG2_ASSERT(rec != 0);
    DENG2_ASSERT(Protocol::recognize(&packet) == type);
    DENG2_UNUSED(type);
    return *rec;
}
String Protocol::command(Packet const &commandPacket) {
    RecordPacket const &rec = asRecordPacket(commandPacket, Command);
    return rec["execute"].value().asText();
}
RecordPacket *Protocol::newConsoleLexicon(Lexicon const &lexicon) {
    RecordPacket *lex = new RecordPacket(PT_LEXICON);
    lex->record().addText("extraChars", lexicon.additionalWordChars());
    ArrayValue &arr = lex->record().addArray("terms").array();
    foreach (String const &term, lexicon.terms()) {
        arr << TextValue(term);
    }
    return lex;
}
Lexicon Protocol::lexicon(Packet const &consoleLexiconPacket) {
    RecordPacket const &rec = asRecordPacket(consoleLexiconPacket, ConsoleLexicon);
    Lexicon lexicon;
    DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, rec["terms"].array().elements()) {
        lexicon.addTerm((*i)->asText());
    }
    lexicon.setAdditionalWordChars(rec.valueAsText("extraChars"));
    return lexicon;
}
RecordPacket *Protocol::newGameState(String const &mode,
                                     String const &rules,
                                     String const &mapId,
                                     String const &mapTitle) {
    RecordPacket *gs = new RecordPacket(PT_GAME_STATE);
    Record &r = gs->record();
    r.addText("mode", mode);
    r.addText("rules", rules);
    r.addText("mapId", mapId);
    r.addText("mapTitle", mapTitle);
    return gs;
}
}} // namespace de::shell
namespace de { namespace shell {
static TimeSpan MSG_EXPIRATION_SECS = 4;
DENG2_PIMPL_NOREF(ServerFinder) {
    Beacon beacon;
    struct Found {
        shell::ServerInfo message;
        Time at;
        Found() : at(Time()) {}
    };
    QMap<Address, Found> servers;
    Impl() : beacon(DEFAULT_PORT) {}
    bool removeExpired() {
        bool changed = false;
        QMutableMapIterator<Address, Found> iter(servers);
        while (iter.hasNext()) {
            Found &found = iter.next().value();
            if (found.at.since() > MSG_EXPIRATION_SECS) {
                iter.remove();
                changed = true;
            }
        }
        return changed;
    }
};
ServerFinder::ServerFinder() : d(new Impl) {
    try {
        connect(&d->beacon, SIGNAL(found(de::Address, de::Block)), this, SLOT(found(de::Address, de::Block)));
        QTimer::singleShot(1000, this, SLOT(expire()));
        if (!App::appExists() || !App::commandLine().has("-nodiscovery")) {
            d->beacon.discover(0 /* no timeout */, 2);
        }
    }
    catch (Beacon::PortError const &er) {
        LOG_WARNING("Automatic server discovery is not available:\n") << er.asText();
    }
}
ServerFinder::~ServerFinder() {
    d.reset();
}
void ServerFinder::clear() {
    d->servers.clear();
}
QList<Address> ServerFinder::foundServers() const {
    return d->servers.keys();
}
String ServerFinder::name(Address const &server) const {
    return messageFromServer(server).name();
}
int ServerFinder::playerCount(Address const &server) const {
    return messageFromServer(server).playerCount();
}
int ServerFinder::maxPlayers(Address const &server) const {
    return messageFromServer(server).maxPlayers();
}
ServerInfo ServerFinder::messageFromServer(Address const &address) const {
    Address addr = shell::checkPort(address);
    if (!d->servers.contains(addr)) {
        throw NotFoundError("ServerFinder::messageFromServer",
                            "No message from server " + addr.asText());
    }
    return d->servers[addr].message;
}
void ServerFinder::found(Address host, Block block) {
    if (host.isLocal()) host.setHost(QHostAddress::LocalHost);
    try {
        LOG_TRACE("Received a server message from %s with %i bytes",
                  host << block.size());
        Record info;
        Reader(block).withHeader() >> info;
        ServerInfo receivedInfo(info);
        receivedInfo.setAddress(host);
        Address const from = receivedInfo.address(); // port validated
        Impl::Found found;
        if (d->servers.contains(from)) {
            d->servers[from].message = receivedInfo;
            d->servers[from].at = Time();
        }
        else {
            found.message = receivedInfo;
            d->servers.insert(from, found);
        }
        emit updated();
    }
    catch (Error const &) {
        if (d->servers.contains(host)) {
            d->servers.remove(host);
        }
    }
}
void ServerFinder::expire() {
    if (d->removeExpired()) {
        emit updated();
    }
    QTimer::singleShot(1000, this, SLOT(expire()));
}
}} // namespace de::shell
namespace de { namespace shell {
static String const VAR_SERVER_ID               ("sid");
static String const VAR_VERSION                 ("ver");
static String const VAR_COMPATIBILITY_VERSION   ("cver");
static String const VAR_HOST                    ("host");
static String const VAR_DOMAIN                  ("dom");
static String const VAR_PORT                    ("port");
static String const VAR_NAME                    ("name");
static String const VAR_DESCRIPTION             ("desc");
static String const VAR_PLUGIN                  ("plugin");
static String const VAR_PACKAGES                ("pkgs");
static String const VAR_GAME_ID                 ("game");
static String const VAR_GAME_CONFIG             ("cfg");
static String const VAR_MAP                     ("map");
static String const VAR_PLAYERS                 ("plrs");
static String const VAR_PLAYER_COUNT            ("pnum");
static String const VAR_MAX_PLAYERS             ("pmax");
static String const VAR_FLAGS                   ("flags");
DENG2_PIMPL(ServerInfo) {
    std::shared_ptr<Record> info;
    Impl(Public *i) : Base(i)
    {}
    void detach() {
        if (info.use_count() > 1) {
            info.reset(new Record(*info));
            DENG2_ASSERT(info.use_count() == 1);
        }
    }
    void checkValid() {
        if (!info->has(VAR_PLAYERS)) info->addArray(VAR_PLAYERS);
        if (info->has(VAR_HOST)) {
            if (self().address().port() != self().port()) {
                self().setAddress(Address(self().address().host(), self().port()));
            }
        }
    }
};
ServerInfo::ServerInfo()
    : d(new Impl(this)) {
    d->info.reset(new Record);
    d->info->set(VAR_VERSION, Version::currentBuild().fullNumber());
    d->info->addArray(VAR_PLAYERS);
}
ServerInfo::ServerInfo(ServerInfo const &other)
    : d(new Impl(this)) {
    d->info = other.d->info;
}
ServerInfo::ServerInfo(Record const &rec)
    : d(new Impl(this)) {
    d->info.reset(new Record(rec));
    d->checkValid();
}
ServerInfo &ServerInfo::operator=(ServerInfo const &other) {
    d->info = other.d->info;
    return *this;
}
Version ServerInfo::version() const {
    return Version(d->info->gets(VAR_VERSION));
}
int ServerInfo::compatibilityVersion() const {
    return d->info->geti(VAR_COMPATIBILITY_VERSION);
}
ServerInfo &ServerInfo::setCompatibilityVersion(int compatVersion) {
    d->detach();
    d->info->set(VAR_COMPATIBILITY_VERSION, compatVersion);
    return *this;
}
ServerInfo &ServerInfo::setServerId(duint32 sid) {
    d->detach();
    d->info->set(VAR_SERVER_ID, sid);
    return *this;
}
Address ServerInfo::address() const {
    if (d->info->has(VAR_HOST)) {
        return Address::parse(d->info->gets(VAR_HOST));
    }
    return Address();
}
String ServerInfo::domainName() const {
    return d->info->gets(VAR_DOMAIN, "");
}
ServerInfo &ServerInfo::setAddress(Address const &address) {
    d->detach();
    d->info->set(VAR_HOST, address.asText());
    d->info->set(VAR_PORT, address.port() ? address.port() : shell::DEFAULT_PORT);
    d->checkValid();
    return *this;
}
duint16 ServerInfo::port() const {
    return duint16(d->info->geti(VAR_PORT, shell::DEFAULT_PORT));
}
duint32 ServerInfo::serverId() const {
    return d->info->getui(VAR_SERVER_ID, 0);
}
String ServerInfo::name() const {
    return d->info->gets(VAR_NAME, "");
}
ServerInfo &ServerInfo::setName(String const &name) {
    d->detach();
    d->info->set(VAR_NAME, name);
    return *this;
}
String ServerInfo::description() const {
    return d->info->gets(VAR_DESCRIPTION, "");
}
ServerInfo &ServerInfo::setDescription(const String &description) {
    d->detach();
    d->info->set(VAR_DESCRIPTION, description);
    return *this;
}
String ServerInfo::pluginDescription() const {
    return d->info->gets(VAR_PLUGIN, "");
}
ServerInfo &ServerInfo::setPluginDescription(String const &pluginDescription) {
    d->detach();
    d->info->set(VAR_PLUGIN, pluginDescription);
    return *this;
}
StringList ServerInfo::packages() const {
    return d->info->getStringList(VAR_PACKAGES);
}
ServerInfo &ServerInfo::setPackages(StringList packages) {
    d->detach();
    ArrayValue &pkgs = d->info->addArray(VAR_PACKAGES).value<ArrayValue>();
    for (String const &p : packages) {
        pkgs << TextValue(p);
    }
    return *this;
}
String ServerInfo::gameId() const {
    return d->info->gets(VAR_GAME_ID, "");
}
ServerInfo &ServerInfo::setGameId(String const &gameId) {
    d->detach();
    d->info->set(VAR_GAME_ID, gameId);
    return *this;
}
String ServerInfo::gameConfig() const {
    return d->info->gets(VAR_GAME_CONFIG, "");
}
ServerInfo &ServerInfo::setGameConfig(String const &gameConfig) {
    d->detach();
    d->info->set(VAR_GAME_CONFIG, gameConfig);
    return *this;
}
String ServerInfo::map() const {
    return d->info->gets(VAR_MAP, "");
}
ServerInfo &ServerInfo::setMap(String const &map) {
    d->detach();
    d->info->set(VAR_MAP, map);
    return *this;
}
StringList ServerInfo::players() const {
    return d->info->getStringList(VAR_PLAYERS);
}
int ServerInfo::playerCount() const {
    return d->info->geti(VAR_PLAYER_COUNT, 0);
}
ServerInfo &ServerInfo::addPlayer(String const &playerName) {
    d->detach();
    ArrayValue &players = d->info->member(VAR_PLAYERS).value<ArrayValue>();
    players.add(playerName);
    d->info->set(VAR_PLAYER_COUNT, players.size());
    return *this;
}
ServerInfo &ServerInfo::removePlayer(String const &playerName) {
    d->detach();
    ArrayValue &players = d->info->member(VAR_PLAYERS).value<ArrayValue>();
    for (int i = 0; i < int(players.size()); ++i) {
        if (players.at(i).asText() == playerName) {
            players.remove(i);
            d->info->set(VAR_PLAYER_COUNT, players.size());
            break;
        }
    }
    return *this;
}
int ServerInfo::maxPlayers() const {
    return d->info->geti(VAR_MAX_PLAYERS);
}
ServerInfo &ServerInfo::setMaxPlayers(int count) {
    d->detach();
    d->info->set(VAR_MAX_PLAYERS, count);
    return *this;
}
ServerInfo::Flags ServerInfo::flags() const {
    return Flags(d->info->geti(VAR_FLAGS, DefaultFlags));
}
String ServerInfo::asStyledText() const {
#define TABBED(A, B) _E(Ta)_E(l) "  " A _E(.) " " _E(\t) B "\n"
    auto const playerNames = players();
    return String(_E(b) "%1" _E(.) "\n%2\n" _E(T`)
                  TABBED("Address:", "%6")
                  TABBED("Joinable:", "%5")
                  TABBED("Players:", "%3 / %4%11")
                  TABBED("Game:", "%7\n%8\n%10 %9")
                  TABBED("Packages:", "%12")
                  /*TABBED("Ping:", "%8 ms (approx)")*/)
            .arg(name())
            .arg(description())
            .arg(playerNames.size())
            .arg(maxPlayers())
            .arg(flags().testFlag(AllowJoin)? "Yes" : "No") // 5
            .arg(address().asText())
            .arg(pluginDescription())
            .arg(gameId())
            .arg(gameConfig())
            .arg(map()) // 10
            .arg(!playerNames.isEmpty()? String(_E(2) " (%1)" _E(.)).arg(String::join(playerNames, " ")) : "")
            .arg(String::join(packages(), " "));
#undef TABBED
}
Block ServerInfo::asJSON() const {
    return composeJSON(*d->info);
}
const Record &ServerInfo::asRecord() const {
    return *d->info;
}
Record ServerInfo::strippedForBroadcast() const {
    Record stripped(*d->info);
    delete stripped.tryRemove(VAR_HOST);     // address in network msg
    delete stripped.tryRemove(VAR_PLUGIN);   // gameId+version is enough
    delete stripped.tryRemove(VAR_PLAYERS);  // count is enough
    delete stripped.tryRemove(VAR_PACKAGES); // queried before connecting
    return stripped;
}
ServerInfo &ServerInfo::setDomainName(String const &domain) {
    d->detach();
    d->info->set(VAR_DOMAIN, domain);
    return *this;
}
ServerInfo &ServerInfo::setFlags(Flags const &flags) {
    d->detach();
    d->info->set(VAR_FLAGS, duint32(flags));
    return *this;
}
void ServerInfo::printToLog(int indexNumber, bool includeHeader) const {
    if (includeHeader) {
        LOG_NET_MSG(_E(m)"    %-20s P/M  L Ver:  Game:            Location:") << "Name:";
    }
    auto const plrs = players();
    LOG_NET_MSG(_E(m)"%-2i: %-20s %i/%-2i %c %-5i %-16s %s")
            << indexNumber
            << name()
            << playerCount()
            << maxPlayers()
            << (flags().testFlag(AllowJoin)? ' ' : '*')
            << compatibilityVersion()
            << pluginDescription()
            << address().asText();
    LOG_NET_MSG("    %s %-40s") << map() << description();
    LOG_NET_MSG("    %s %s") << gameId() << gameConfig();
    LOG_NET_MSG("    Instance ID: %08x") << serverId();
    LOG_NET_MSG("    Packages: " _E(>) "%s") << String::join(packages(), "\n");
    if (!plrs.isEmpty()) {
        LOG_NET_MSG("    Players: " _E(>) "%s") << String::join(plrs, "\n");
    }
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(TextCanvas) {
    Size          size;
    QList<Char *> lines;
    struct RichFormat {
        Char::Attribs attrib;
        Rangei        range;
    };
    QList<RichFormat> richFormats;
    Impl(Size const &initialSize) : size(initialSize) {
        for (duint row = 0; row < size.y; ++row) {
            lines.append(makeLine());
        }
    }
    ~Impl() {
        for (int i = 0; i < lines.size(); ++i) {
            delete [] lines[i];
        }
    }
    dsize lineCount() const {
        return lines.size();
    }
    Char *makeLine() {
        return new Char[size.x];
    }
    void resize(Size const &newSize) {
        if (newSize == size) return;
        while (newSize.y < lineCount()) {
            lines.removeLast();
        }
        while (newSize.y > lineCount()) {
            lines.append(makeLine());
        }
        Q_ASSERT(lineCount() == newSize.y);
        size.y = newSize.y;
        for (int row = 0; row < lines.size(); ++row) {
            Char *newLine = new Char[newSize.x];
            memcpy(newLine, lines[row], sizeof(Char) * de::min(size.x, newSize.x));
            delete [] lines[row];
            lines[row] = newLine;
        }
        size.x = newSize.x;
    }
    void markAllAsDirty(bool markDirty) {
        for (int row = 0; row < lines.size(); ++row) {
            Char *line = lines[row];
            for (duint col = 0; col < size.x; ++col) {
                Char &c = line[col];
                if (markDirty)
                    c.attribs |= Char::Dirty;
                else
                    c.attribs &= ~Char::Dirty;
            }
        }
    }
    Char::Attribs richAttribsForTextIndex(int pos, int offset = 0) const {
        Char::Attribs attr;
        foreach (RichFormat const &rf, richFormats) {
            if (rf.range.contains(offset + pos)) {
                attr |= rf.attrib;
            }
        }
        return attr;
    }
};
TextCanvas::TextCanvas(Size const &size) : d(new Impl(size)) {
    d->size = size;
}
TextCanvas::~TextCanvas()
{}
TextCanvas::Size TextCanvas::size() const {
    return d->size;
}
int TextCanvas::width() const {
    return d->size.x;
}
int TextCanvas::height() const {
    return d->size.y;
}
Rectanglei TextCanvas::rect() const {
    return Rectanglei(0, 0, size().x, size().y);
}
void TextCanvas::resize(Size const &newSize) {
    d->resize(newSize);
}
TextCanvas::Char &TextCanvas::at(Coord const &pos) {
    Q_ASSERT(isValid(pos));
    return d->lines[pos.y][pos.x];
}
TextCanvas::Char const &TextCanvas::at(Coord const &pos) const {
    Q_ASSERT(isValid(pos));
    return d->lines[pos.y][pos.x];
}
bool TextCanvas::isValid(Coord const &pos) const {
    return (pos.x >= 0 && pos.y >= 0 && pos.x < int(d->size.x) && pos.y < int(d->size.y));
}
void TextCanvas::markDirty() {
    d->markAllAsDirty(true);
}
void TextCanvas::clear(Char const &ch) {
    fill(Rectanglei(0, 0, d->size.x, d->size.y), ch);
}
void TextCanvas::fill(Rectanglei const &rect, Char const &ch) {
    for (int y = rect.top(); y < rect.bottom(); ++y) {
        for (int x = rect.left(); x < rect.right(); ++x) {
            Coord const xy(x, y);
            if (isValid(xy)) at(xy) = ch;
        }
    }
}
void TextCanvas::put(Vector2i const &pos, Char const &ch) {
    if (isValid(pos)) {
        at(pos) = ch;
    }
}
void TextCanvas::clearRichFormat() {
    d->richFormats.clear();
}
void TextCanvas::setRichFormatRange(Char::Attribs const &attribs, Rangei const &range) {
    Impl::RichFormat rf;
    rf.attrib = attribs;
    rf.range  = range;
    d->richFormats.append(rf);
}
void TextCanvas::drawText(Vector2i const &pos, String const &text,
                          Char::Attribs const &attribs, int richOffset) {
    Vector2i p = pos;
    for (int i = 0; i < text.size(); ++i) {
        if (isValid(p)) {
            at(p) = Char(text[i], attribs | d->richAttribsForTextIndex(i, richOffset));
        }
        p.x++;
    }
}
void TextCanvas::drawWrappedText(Vector2i const &pos, String const &text,
                                 ILineWrapping const &wraps, Char::Attribs const &attribs,
                                 Alignment lineAlignment) {
    int const width = wraps.width();
    for (int y = 0; y < wraps.height(); ++y) {
        WrappedLine const &span = wraps.line(y);
        String part = text.substr(span.range);
        int x = 0;
        if (lineAlignment.testFlag(AlignRight)) {
            x = width - part.size();
        }
        else if (!lineAlignment.testFlag(AlignLeft)) {
            x = width/2 - part.size()/2;
        }
        drawText(pos + Vector2i(x, y), part, attribs, span.range.start);
    }
}
void TextCanvas::drawLineRect(Rectanglei const &rect, Char::Attribs const &attribs) {
    Char const corner('+', attribs);
    Char const hEdge ('-', attribs);
    Char const vEdge ('|', attribs);
    for (duint x = 1; x < rect.width() - 1; ++x) {
        put(rect.topLeft + Vector2i(x, 0), hEdge);
        put(rect.bottomLeft() + Vector2i(x, -1), hEdge);
    }
    for (duint y = 1; y < rect.height() - 1; ++y) {
        put(rect.topLeft + Vector2i(0, y), vEdge);
        put(rect.topRight() + Vector2i(-1, y), vEdge);
    }
    put(rect.topLeft, corner);
    put(rect.topRight() - Vector2i(1, 0), corner);
    put(rect.bottomRight - Vector2i(1, 1), corner);
    put(rect.bottomLeft() - Vector2i(0, 1), corner);
}
void TextCanvas::draw(TextCanvas const &canvas, Coord const &topLeft) {
    for (duint y = 0; y < canvas.d->size.y; ++y) {
        for (duint x = 0; x < canvas.d->size.x; ++x) {
            Coord const xy(x, y);
            Coord const p = topLeft + xy;
            if (isValid(p)) {
                at(p) = canvas.at(xy);
            }
        }
    }
}
void TextCanvas::show() {
    d->markAllAsDirty(false);
}
void TextCanvas::setCursorPosition(Vector2i const &) {}
}} // namespace de::shell
namespace de { namespace shell {
TextRootWidget::TextRootWidget(TextCanvas *cv) : RootWidget(), _canvas(cv), _drawRequest(false) {
    DENG2_ASSERT(cv != 0);
    setViewSize(cv->size());
}
TextRootWidget::~TextRootWidget() {
    delete _canvas;
}
TextCanvas &TextRootWidget::rootCanvas() {
    return *_canvas;
}
void TextRootWidget::setViewSize(Size const &viewSize) {
    Size vs = viewSize.max(Size(1, 1));
    _canvas->resize(vs);
    RootWidget::setViewSize(vs);
}
TextWidget *TextRootWidget::focus() const {
    return static_cast<TextWidget *>(RootWidget::focus());
}
void TextRootWidget::requestDraw() {
    _drawRequest = true;
}
bool TextRootWidget::drawWasRequested() const {
    return _drawRequest;
}
void TextRootWidget::draw() {
    RootWidget::draw();
    if (focus()) {
        _canvas->setCursorPosition(focus()->cursorPosition());
    }
    _canvas->show();
    _drawRequest = false;
}
}} // namespace de::shell
namespace de { namespace shell {
DENG2_PIMPL_NOREF(TextWidget) {
    TextCanvas *    canvas;
    RuleRectangle * rule;
    QList<Action *> actions;
    Impl() : canvas(0), rule(new RuleRectangle)
    {}
    ~Impl() {
        delete rule;
        foreach (Action *act, actions) releaseRef(act);
    }
    void removeAction(Action &action) {
        for (int i = actions.size() - 1; i >= 0; --i) {
            if (actions.at(i) == &action) {
                releaseRef(actions[i]);
                actions.removeAt(i);
            }
        }
    }
    /**
     * Navigates focus to another widget, assuming this widget currently has
     * focus. Used in focus cycle navigation.
     *
     * @param root  Root widget.
     * @param name  Name of the widget to change focus to.
     *
     * @return Focus changed successfully.
     */
    bool navigateFocus(TextRootWidget &root, String const &name) {
        Widget *w = root.find(name);
        if (w) {
            root.setFocus(w);
            root.requestDraw();
            return true;
        }
        return false;
    }
};
TextWidget::TextWidget(String const &name) : Widget(name), d(new Impl) {
    setBehavior(Focusable, SetFlags);
}
TextRootWidget &TextWidget::root() const {
    TextRootWidget *r = dynamic_cast<TextRootWidget *>(&Widget::root());
    DENG2_ASSERT(r != 0);
    return *r;
}
void TextWidget::setTargetCanvas(TextCanvas *canvas) {
    d->canvas = canvas;
}
TextCanvas &TextWidget::targetCanvas() const {
    if (!d->canvas) {
        return root().rootCanvas();
    }
    return *d->canvas;
}
void TextWidget::redraw() {
    if (hasRoot() && !isHidden()) root().requestDraw();
}
void TextWidget::drawAndShow() {
    if (!isHidden()) {
        draw();
        NotifyArgs args(&Widget::draw);
        args.conditionFunc = &Widget::isVisible;
        notifyTree(args);
        targetCanvas().show();
    }
}
RuleRectangle &TextWidget::rule() {
    DENG2_ASSERT(d->rule != 0);
    return *d->rule;
}
RuleRectangle const &TextWidget::rule() const {
    DENG2_ASSERT(d->rule != 0);
    return *d->rule;
}
Vector2i TextWidget::cursorPosition() const {
    return Vector2i(rule().left().valuei(),
                    rule().top().valuei());
}
void TextWidget::addAction(RefArg<Action> action) {
    d->actions.append(action.holdRef());
}
void TextWidget::removeAction(Action &action) {
    d->removeAction(action);
}
bool TextWidget::handleEvent(Event const &event) {
    if (event.type() == Event::KeyPress) {
        KeyEvent const &keyEvent = event.as<KeyEvent>();
        foreach (Action *act, d->actions) {
            if (act->tryTrigger(keyEvent)) return true;
        }
        if ((keyEvent.key() == Qt::Key_Tab || keyEvent.key() == Qt::Key_Down) &&
                hasFocus() && !focusNext().isEmpty()) {
            if (d->navigateFocus(root(), focusNext()))
                return true;
        }
        if ((keyEvent.key() == Qt::Key_Backtab || keyEvent.key() == Qt::Key_Up) &&
                hasFocus() && !focusPrev().isEmpty()) {
            if (d->navigateFocus(root(), focusPrev()))
                return true;
        }
    }
    return Widget::handleEvent(event);
}
}} // namespace de::shell
