﻿/** @file abstractfont.cpp Abstract font.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/abstractfont.h"
/** @file bitmapfont.cpp  Bitmap font resource.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "resource/bitmapfont.h"
#include "dd_main.h" // isDedicated
#include "gl/gl_main.h" // GL_NewTextureWithParams
#include "sys_system.h" // novideo
#include "FontManifest"
#include <doomsday/uri.h>
#include <de/LogBuffer>
#include <de/mathutil.h> // M_CeilPow2()
#include <de/memory.h>
#include <doomsday/busymode.h>
#include <doomsday/filesys/fs_main.h>
/** @file clientmaterial.cpp  Client-side material.
 *
 * @authors Copyright © 2009-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/clientmaterial.h"
#include <QFlag>
#include <QtAlgorithms>
#include <de/Log>
#include <doomsday/console/cmd.h>
#include <doomsday/res/Textures>
#include <doomsday/world/Materials>
#include <doomsday/world/MaterialManifest>
#include "dd_main.h"
#include "MaterialAnimator"
/** @file clientresources.cpp  Client-side resource subsystem.
 *
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "resource/clientresources.h"
#include <QHash>
#include <QVector>
#include <QtAlgorithms>
#include <de/memory.h>
#include <de/App>
#include <de/ArrayValue>
#include <de/ByteOrder>
#include <de/ByteRefArray>
#include <de/DirectoryFeed>
#include <de/Function>
#include <de/LogBuffer>
#include <de/Loop>
#include <de/Module>
#include <de/NativePath>
#include <de/NumberValue>
#include <de/PackageLoader>
#include <de/Reader>
#include <de/RecordValue>
#include <de/StringPool>
#include <de/Task>
#include <de/TaskPool>
#include <de/Time>
#include <doomsday/console/cmd.h>
#include <doomsday/defs/music.h>
#include <doomsday/defs/sprite.h>
#include <doomsday/doomsdayapp.h>
#include <doomsday/filesys/fs_main.h>
#include <doomsday/filesys/fs_util.h>
#include <doomsday/filesys/lumpindex.h>
#include <doomsday/res/AnimGroups>
#include <doomsday/res/ColorPalettes>
#include <doomsday/res/Composite>
#include <doomsday/res/MapManifests>
#include <doomsday/res/Patch>
#include <doomsday/res/PatchName>
#include <doomsday/res/Sprites>
#include <doomsday/res/TextureManifest>
#include <doomsday/res/Textures>
#include <doomsday/world/Material>
#include <doomsday/world/Materials>
#include "def_main.h"
#include "dd_main.h"
#include "dd_def.h"
#include "clientapp.h"
#include "ui/progress.h"
#include "ui/clientwindowsystem.h"
#include "sys_system.h"  // novideo
#include "gl/gl_tex.h"
#include "gl/gl_texmanager.h"
#include "gl/svg.h"
#include "resource/clienttexture.h"
#include "render/rend_model.h"
#include "render/rend_particle.h"  // Rend_ParticleReleaseSystemTextures
#include "render/rendersystem.h"
#include "network/net_demo.h"  // playback
#include "render/rend_main.h"  // Rend_MapSurfaceMaterialSpec
#include "render/billboard.h"  // Rend_SpriteMaterialSpec
#include "render/skydrawable.h"
#include "world/clientserverworld.h"
#include "world/map.h"
#include "world/p_object.h"
#include "world/sky.h"
#include "world/thinkers.h"
#include "Sector"
#include "Surface"
/** @file texture.cpp  Logical texture resource.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/clienttexture.h"
#include "resource/clientresources.h"
#include <doomsday/console/cmd.h>
#include <doomsday/res/TextureManifest>
#include <de/Error>
#include <de/Log>
#include <de/memory.h>
#include <QMap>
/** @file compositebitmapfont.cpp Composite bitmap font.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "clientapp.h"
#include "resource/compositebitmapfont.h"
#include <doomsday/res/Textures>
#include "api_resource.h" // R_GetPatchInfo
#include "dd_main.h" // App_Resources(), isDedicated
#include "sys_system.h" // novideo
#include "gl/gl_texmanager.h" // GL_TextureVariantSpec
/** @file fontmanifest.cpp Font resource manifest.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "resource/fontmanifest.h"
#include "dd_main.h" // App_Fonts(), remove me
#include "FontScheme"
#include <de/Log>
/** @file fonts.cpp  Font resource collection.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
/** @file fontscheme.cpp Font resource scheme.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/fontscheme.h"
#include "dd_types.h"
#include <de/Log>
#include <QList>
/** @file resource/model.cpp  3D model resource (MD2/DMD).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "resource/framemodel.h"
#include "ClientTexture"
#include <doomsday/res/TextureManifest>
#include <doomsday/filesys/file.h>
#include <de/Range>
#include <de/memory.h>
#include <QtAlgorithms>
/** @file hq2x.cpp  High-Quality 2x Graphics Resizing.
 *
 * @authors Copyright © 2003 Maxim Stepin <maxst@hiend3d.com>
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2009-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "resource/hq2x.h"
#include <cstdlib>
#include <de/memory.h>
#include "dd_main.h"
#include "dd_types.h"
#include "dd_share.h"
#include "resource/image.h"
/*
 * RGB color space.
 */
#define BGR888_PACK(b, g, r) ( ((uint32_t)(b) << 16) | ((uint32_t)(g) << 8) | (uint32_t)(r) )
#define BGR888_COMP(n, c)   ( ((c) >> ((n) << 3)) & 0xFF )
#define BGR888_Bmask        ((int)0xFF0000)
#define BGR888_Gmask        ((int)0x00FF00)
#define BGR888_Rmask        ((int)0x0000FF)
#define BGR888toBGR565(c)   (((c & 0xF8) >> 3) | ((c & 0xFC00) >> 5) | ((c & 0xF80000) >> 8))
#define ABGR8888_PACK(a, b, g, r) ( ((uint32_t)(a) << 24) | BGR888_PACK((b), (g), (r)) )
#define ABGR8888_COMP       BGR888_COMP
#define ABGR8888_Amask      ((int)0xFF000000)
#define ABGR8888_Bmask      BGR888_Bmask
#define ABGR8888_Gmask      BGR888_Gmask
#define ABGR8888_Rmask      BGR888_Rmask
#define ABGR8888_RGBmask    ((int)0x00FFFFFF)
#define BGR888toYUV888(v)   (lutBGR888toYUV888[BGR888toBGR565(v)])
#define ABGR8888toYUV888(v) (BGR888toYUV888(v & ABGR8888_RGBmask))
#define ABGR8888toAYUV8888(v) (ABGR8888toYUV888(v) | (((v) & ABGR8888_Amask) << 24))
#define RGB888_PACK(r, g, b) ( ((uint32_t)(r) << 16) | ((uint32_t)(g) << 8) | (uint32_t)(b) )
#define RGB888_COMP(n, c)   ( ((c) >> ((n) << 3)) & 0xFF )
#define RGB888_Rmask        ((int)0xFF0000)
#define RGB888_Gmask        ((int)0x00FF00)
#define RGB888_Bmask        ((int)0x0000FF)
#define RGB888toRGB565(c)   (((c & 0xF80000) >> 8) | ((c & 0xFC00) >> 5) | ((c & 0xF8) >> 3))
#define ARGB8888_PACK(a, r, g, b) ( ((uint32_t)(a) << 24) | RGB888_PACK((r), (g), (b)) )
#define ARGB8888_COMP       RGB888_COMP
#define ARGB8888_Amask      ((int)0xFF000000)
#define ARGB8888_Rmask      RGB888_Rmask
#define ARGB8888_Gmask      RGB888_Gmask
#define ARGB8888_Bmask      RGB888_Bmask
#define ARGB8888_RGBmask    ((int)0x00FFFFFF)
/*
#define RGB888toYUV888(v)   (BGR888toYUV888(RGB888toBGR888(v)))
#define ARGB8888toYUV888(v) (RGB888toYUV888((v) & ARGB8888_RGBmask])
#define ARGB8888toAYUV8888(v) (ARGB8888toYUV888(v) | (((v) & ARGB8888_Amask) << 24))
*/
#define BGR565_PACK(b, g, r) ( ((uint16_t)(b) << 11) | ((uint16_t)(g) << 5) | (uint16_t)(r) )
#define BGR565_Bmask        ((int)0xF800)
#define BGR565_Gmask        ((int)0x7E0)
#define BGR565_Rmask        ((int)0x1F)
/*
 * YUV color space.
 */
#define YUV888_PACK(y, u, v) ( ((uint32_t)(y) << 16) | ((uint32_t)(u) << 8) | (uint32_t)(v) )
#define YUV888_COMP(n, c)   ( ((c) >> ((n) << 3)) & 0xFF )
#define YUV888_Ymask        ((int)0xFF0000)
#define YUV888_Umask        ((int)0x00FF00)
#define YUV888_Vmask        ((int)0x0000FF)
#define AYUV8888_PACK(y, u, v, a) ( ((uint32_t)(a) << 24) | YUV888_PACK((y), (u), (v)) )
#define AYUV8888_COMP       YUV888_COMP
#define AYUV8888_Amask      ((int)0xFF000000)
#define AYUV8888_Ymask      YUV888_Ymask
#define AYUV8888_Umask      YUV888_Umask
#define AYUV8888_Vmask      YUV888_Vmask
#define AYUV8888_YUVmask    ((int)0x00FFFFFF)
#define trY                 (48)
#define trU                 (7)
#define trV                 (6)
#define PIXEL00_0         Transl(pOut,       w[5]);
#define PIXEL00_10       Interp1(pOut,       w[5], w[1]);
#define PIXEL00_11       Interp1(pOut,       w[5], w[4]);
#define PIXEL00_12       Interp1(pOut,       w[5], w[2]);
#define PIXEL00_20       Interp2(pOut,       w[5], w[4], w[2]);
#define PIXEL00_21       Interp2(pOut,       w[5], w[1], w[2]);
#define PIXEL00_22       Interp2(pOut,       w[5], w[1], w[4]);
#define PIXEL00_60       Interp6(pOut,       w[5], w[2], w[4]);
#define PIXEL00_61       Interp6(pOut,       w[5], w[4], w[2]);
#define PIXEL00_70       Interp7(pOut,       w[5], w[4], w[2]);
#define PIXEL00_90       Interp9(pOut,       w[5], w[4], w[2]);
#define PIXEL00_100     Interp10(pOut,       w[5], w[4], w[2]);
#define PIXEL01_0         Transl(pOut+4,     w[5]);
#define PIXEL01_10       Interp1(pOut+4,     w[5], w[3]);
#define PIXEL01_11       Interp1(pOut+4,     w[5], w[2]);
#define PIXEL01_12       Interp1(pOut+4,     w[5], w[6]);
#define PIXEL01_20       Interp2(pOut+4,     w[5], w[2], w[6]);
#define PIXEL01_21       Interp2(pOut+4,     w[5], w[3], w[6]);
#define PIXEL01_22       Interp2(pOut+4,     w[5], w[3], w[2]);
#define PIXEL01_60       Interp6(pOut+4,     w[5], w[6], w[2]);
#define PIXEL01_61       Interp6(pOut+4,     w[5], w[2], w[6]);
#define PIXEL01_70       Interp7(pOut+4,     w[5], w[2], w[6]);
#define PIXEL01_90       Interp9(pOut+4,     w[5], w[2], w[6]);
#define PIXEL01_100     Interp10(pOut+4,     w[5], w[2], w[6]);
#define PIXEL10_0         Transl(pOut+BpL,   w[5]);
#define PIXEL10_10       Interp1(pOut+BpL,   w[5], w[7]);
#define PIXEL10_11       Interp1(pOut+BpL,   w[5], w[8]);
#define PIXEL10_12       Interp1(pOut+BpL,   w[5], w[4]);
#define PIXEL10_20       Interp2(pOut+BpL,   w[5], w[8], w[4]);
#define PIXEL10_21       Interp2(pOut+BpL,   w[5], w[7], w[4]);
#define PIXEL10_22       Interp2(pOut+BpL,   w[5], w[7], w[8]);
#define PIXEL10_60       Interp6(pOut+BpL,   w[5], w[4], w[8]);
#define PIXEL10_61       Interp6(pOut+BpL,   w[5], w[8], w[4]);
#define PIXEL10_70       Interp7(pOut+BpL,   w[5], w[8], w[4]);
#define PIXEL10_90       Interp9(pOut+BpL,   w[5], w[8], w[4]);
#define PIXEL10_100     Interp10(pOut+BpL,   w[5], w[8], w[4]);
#define PIXEL11_0         Transl(pOut+BpL+4, w[5]);
#define PIXEL11_10       Interp1(pOut+BpL+4, w[5], w[9]);
#define PIXEL11_11       Interp1(pOut+BpL+4, w[5], w[6]);
#define PIXEL11_12       Interp1(pOut+BpL+4, w[5], w[8]);
#define PIXEL11_20       Interp2(pOut+BpL+4, w[5], w[6], w[8]);
#define PIXEL11_21       Interp2(pOut+BpL+4, w[5], w[9], w[8]);
#define PIXEL11_22       Interp2(pOut+BpL+4, w[5], w[9], w[6]);
#define PIXEL11_60       Interp6(pOut+BpL+4, w[5], w[8], w[6]);
#define PIXEL11_61       Interp6(pOut+BpL+4, w[5], w[6], w[8]);
#define PIXEL11_70       Interp7(pOut+BpL+4, w[5], w[6], w[8]);
#define PIXEL11_90       Interp9(pOut+BpL+4, w[5], w[6], w[8]);
#define PIXEL11_100     Interp10(pOut+BpL+4, w[5], w[6], w[8]);
/** @file idtech1image.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "resource/idtech1image.h"
#include <doomsday/resource/patch.h>
#include <doomsday/Game>
#include <de/Style>
/** @file image.cpp  Image objects and related routines.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "resource/image.h"
#include "resource/tga.h"
#include <doomsday/resource/patch.h>
#include <doomsday/resource/colorpalettes.h>
#include <de/memory.h>
#include <de/LogBuffer>
#include <QByteArray>
#include <QImage>
#include <de/NativePath>
#include <doomsday/filesys/fs_main.h>
#include "dd_main.h"
#include <doomsday/res/Composite>
#include <doomsday/resource/pcx.h>
#include "gl/gl_tex.h"
#include "render/rend_main.h" // misc global vars awaiting new home
#ifndef DENG2_QT_4_7_OR_NEWER // older than 4.7?
#  define constBits bits
#endif
/** @file materiallightdecoration.cpp   Logical material, light decoration.
 *
 * @authors Copyright © 2011-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/lightmaterialdecoration.h"
#include <doomsday/defs/material.h>
#include <doomsday/res/Textures>
#include "clientapp.h"
/** @file materialanimator.cpp  Animator for a draw-context Material variant.
 *
 * @authors Copyright © 2011-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "resource/materialanimator.h"
#include <array>
#include <de/Log>
#include <doomsday/res/Textures>
#include <doomsday/world/detailtexturemateriallayer.h>
#include <doomsday/world/shinetexturemateriallayer.h>
#include <doomsday/world/texturemateriallayer.h>
#include "clientapp.h"
#include "client/cl_def.h"      // playback / clientPaused
#include "MaterialVariantSpec"
#include "gl/gl_texmanager.h"
#include "render/r_main.h"      // frameCount, frameTimePos
#include "render/rend_main.h"
#include "render/rend_halo.h"
#include "render/viewports.h"
#include "resource/clientresources.h"
#include "resource/lightmaterialdecoration.h"
/** @file texturevariant.cpp Context specialized texture variant.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "misc/r_util.h"
#include "gl/gl_defer.h"
#include "gl/gl_main.h"
#include "gl/gl_tex.h"
#include "gl/texturecontent.h"
#include "resource/image.h" // GL_LoadSourceImage
#include "render/rend_main.h" // misc global vars awaiting new home
#include <doomsday/resource/colorpalettes.h>
#include <doomsday/res/Texture>
#include <de/LogBuffer>
#include <de/mathutil.h> // M_CeilPow
/** @file tga.cpp  Truevision TGA (a.k.a Targa) image reader
 *
 * @authors Copyright © 2003-2019 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2009-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "resource/tga.h"
#include "dd_share.h"
#include <de/memory.h>
#include <de/Image>
#include <cstdio>
#include <cstdlib>
#include <cstring>
using namespace res;
using namespace de;
AbstractFont::AbstractFont(FontManifest &manifest)
    : _manifest(manifest)
    , _flags(0)
{}
AbstractFont::~AbstractFont() {
    DENG2_FOR_AUDIENCE(Deletion, i) i->fontBeingDeleted(*this);
}
void AbstractFont::glInit() const
{}
void AbstractFont::glDeinit() const
{}
FontManifest &AbstractFont::manifest() const {
    return _manifest;
}
AbstractFont::Flags AbstractFont::flags() const {
    return _flags;
}
int AbstractFont::ascent() const {
    return 0;
}
int AbstractFont::descent() const {
    return 0;
}
int AbstractFont::lineSpacing() const {
    return 0;
}
static byte inByte(FileHandle *file) {
    byte b;
    file->read((uint8_t *)&b, sizeof(b));
    return b;
}
static ushort inShort(FileHandle *file) {
    ushort s;
    file->read((uint8_t *)&s, sizeof(s));
    return DD_USHORT(s);
}
struct Glyph {
    Rectanglei posCoords;
    Rectanglei texCoords;
};
DENG2_PIMPL(BitmapFont) {
    String filePath;        ///< The "archived" version of this font (if any).
    GLuint texGLName;       ///< GL-texture name.
    Vector2ui texMargin;    ///< Margin in pixels.
    Vector2i texDimensions; ///< Texture dimensions in pixels.
    bool needGLInit;
    int leading;
    int ascent;
    int descent;
    Glyph glyphs[MAX_CHARS];
    Glyph missingGlyph;
    Impl(Public *i)
        : Base(i)
        , texGLName(0)
        , needGLInit(true)
        , leading(0)
        , ascent(0)
        , descent(0)
    {}
    ~Impl() {
        self().glDeinit();
    }
    /**
     * Lookup the glyph for the specified character @a ch. If no glyph is defined
     * for this character then the special "missing glyph" is returned instead.
     */
    Glyph &glyph(uchar ch) {
        return glyphs[ch];
    }
    /**
     * Determine the dimensions of the "missing glyph" by averaging the dimensions
     * of the available/defined glyphs.
     *
     * @todo Could be smarter. Presently this treats @em all glyphs equally.
     */
    Vector2ui findMissingGlyphSize() {
        Vector2ui accumSize;
        int glyphCount = 0;
        for(int i = 0; i < MAX_CHARS; ++i) {
            Glyph &glyph = glyphs[i];
            if(glyph.posCoords.isNull()) continue;
            accumSize += glyph.posCoords.size();
            glyphCount += 1;
        }
        return accumSize / glyphCount;
    }
    uint8_t *readFormat0(FileHandle *file) {
        DENG2_ASSERT(file != 0);
        self()._flags |= AbstractFont::Colorize;
        self()._flags &= ~AbstractFont::Shadowed;
        texMargin = Vector2ui(0, 0);
        texDimensions.x = inShort(file);
        texDimensions.y = inShort(file);
        int glyphCount  = inShort(file);
        for(int i = 0; i < glyphCount; ++i) {
            Glyph *ch = &glyphs[i < MAX_CHARS ? i : MAX_CHARS - 1];
            ushort x = inShort(file);
            ushort y = inShort(file);
            ushort w = inByte(file);
            ushort h = inByte(file);
            ch->posCoords = Rectanglei::fromSize(Vector2i(0, 0), Vector2ui(w, h));
            ch->texCoords = Rectanglei::fromSize(Vector2i(x, y), Vector2ui(w, h));
        }
        missingGlyph.posCoords.setSize(findMissingGlyphSize());
        int bitmapFormat = inByte(file);
        if(bitmapFormat > 0) {
            de::Uri uri = self().manifest().composeUri();
            throw Error("BitmapFont::readFormat0", QString("Font \"%1\" uses unknown format '%2'").arg(uri).arg(bitmapFormat));
        }
        int const numPels = texDimensions.x * texDimensions.y;
        uint32_t *image = (uint32_t *) M_Calloc(numPels * 4);
        int c, i;
        for(c = i = 0; i < (numPels + 7) / 8; ++i) {
            int bit, mask = inByte(file);
            for(bit = 7; bit >= 0; bit--, ++c) {
                if(c >= numPels) break;
                if(mask & (1 << bit)) {
                    image[c] = ~0;
                }
            }
        }
        return (uint8_t *)image;
    }
    uint8_t *readFormat2(FileHandle *file) {
        DENG2_ASSERT(file != 0);
        self()._flags |= AbstractFont::Colorize | AbstractFont::Shadowed;
        int bitmapFormat = inByte(file);
        if(bitmapFormat != 1 && bitmapFormat != 0) // Luminance + Alpha.
        {
            de::Uri uri = self().manifest().composeUri();
            throw Error("BitmapFont::readFormat2", QString("Font \"%1\" uses unknown format '%2'").arg(uri).arg(bitmapFormat));
        }
        texDimensions.x = inShort(file);
        texDimensions.y = M_CeilPow2(inShort(file));
        int glyphCount  = inShort(file);
        texMargin.x     = texMargin.y = inShort(file);
        leading = inShort(file);
        /*glyphHeight =*/ inShort(file); // Unused.
        ascent  = inShort(file);
        descent = inShort(file);
        for(int i = 0; i < glyphCount; ++i) {
            ushort code = inShort(file);
            DENG2_ASSERT(code < MAX_CHARS);
            Glyph *ch = &glyphs[code];
            ushort x = inShort(file);
            ushort y = inShort(file);
            ushort w = inShort(file);
            ushort h = inShort(file);
            ch->posCoords = Rectanglei::fromSize(Vector2i(0, 0), Vector2ui(w, h) - texMargin * 2);
            ch->texCoords = Rectanglei::fromSize(Vector2i(x, y), Vector2ui(w, h));
        }
        missingGlyph.posCoords.setSize(findMissingGlyphSize());
        int const numPels = texDimensions.x * texDimensions.y;
        uint32_t *image = (uint32_t *) M_Calloc(numPels * 4);
        if(bitmapFormat == 0) {
            uint32_t *ptr = image;
            for(int i = 0; i < numPels; ++i) {
                byte red   = inByte(file);
                byte green = inByte(file);
                byte blue  = inByte(file);
                byte alpha = inByte(file);
                *ptr++ = DD_ULONG(red | (green << 8) | (blue << 16) | (alpha << 24));
            }
        }
        else if(bitmapFormat == 1) {
            uint32_t *ptr = image;
            for(int i = 0; i < numPels; ++i) {
                byte luminance = inByte(file);
                byte alpha     = inByte(file);
                *ptr++ = DD_ULONG(luminance | (luminance << 8) | (luminance << 16) | (alpha << 24));
            }
        }
        return (uint8_t *)image;
    }
};
BitmapFont::BitmapFont(FontManifest &manifest)
    : AbstractFont(manifest), d(new Impl(this))
{}
BitmapFont *BitmapFont::fromFile(FontManifest &manifest, String resourcePath) // static
{
    BitmapFont *font = new BitmapFont(manifest);
    font->setFilePath(resourcePath);
    font->glInit();
    return font;
}
int BitmapFont::ascent() const {
    glInit();
    return d->ascent;
}
int BitmapFont::descent() const {
    glInit();
    return d->descent;
}
int BitmapFont::lineSpacing() const {
    glInit();
    return d->leading;
}
Rectanglei const &BitmapFont::glyphPosCoords(uchar ch) const {
    glInit();
    return d->glyph(ch).posCoords;
}
Rectanglei const &BitmapFont::glyphTexCoords(uchar ch) const {
    glInit();
    return d->glyph(ch).texCoords;
}
void BitmapFont::glInit() const {
    LOG_AS("BitmapFont");
    if(!d->needGLInit) return;
    if(novideo || BusyMode_Active()) return;
    glDeinit();
    try {
        String path = (NativePath::workPath() / NativePath(d->filePath).expand()).withSeparators('/');
        FileHandle *hndl = &App_FileSystem().openFile(path, "rb");
        int format = inByte(hndl);
        uint8_t *pixels = 0;
        switch(format) {
        case 0: pixels = d->readFormat0(hndl); break;
        case 2: pixels = d->readFormat2(hndl); break;
        default:
            DENG2_ASSERT(!"BitmapFont: Format not implemented");
        }
        if(!pixels) {
            App_FileSystem().releaseFile(hndl->file());
            delete hndl;
            return;
        }
        if(!novideo && !isDedicated) {
            LOG_GL_XVERBOSE("Uploading atlas texture for \"%s\"...",
                            manifest().composeUri());
            d->texGLName = GL_NewTextureWithParams(DGL_RGBA, d->texDimensions.x, d->texDimensions.y,
                pixels, 0, 0, GL_LINEAR, GL_NEAREST, 0 /* no AF */,
                GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE);
        }
        M_Free(pixels);
        App_FileSystem().releaseFile(hndl->file());
        delete hndl;
    }
    catch(FS1::NotFoundError const&)
    {} // Ignore error.
    d->needGLInit = false;
}
void BitmapFont::glDeinit() const {
    LOG_AS("BitmapFont");
    if(novideo || isDedicated) return;
    d->needGLInit = true;
    if(BusyMode_Active()) return;
    if(d->texGLName) {
        Deferred_glDeleteTextures(1, (GLuint const *) &d->texGLName);
    }
    d->texGLName = 0;
}
void BitmapFont::setFilePath(String newFilePath) {
    if(d->filePath != newFilePath) {
        d->filePath = newFilePath;
        d->needGLInit = true;
    }
}
uint BitmapFont::textureGLName() const {
    return d->texGLName;
}
Vector2i const &BitmapFont::textureDimensions() const {
    return d->texDimensions;
}
Vector2ui const &BitmapFont::textureMargin() const {
    return d->texMargin;
}
DENG2_PIMPL_NOREF(ClientMaterial::Decoration) {
    ClientMaterial *material = nullptr;  ///< Owning Material.
    Vector2i patternSkip;          ///< Pattern skip intervals.
    Vector2i patternOffset;        ///< Pattern skip interval offsets.
};
ClientMaterial::Decoration::Decoration(Vector2i const &patternSkip, Vector2i const &patternOffset)
    : d(new Impl) {
    d->patternSkip   = patternSkip;
    d->patternOffset = patternOffset;
}
ClientMaterial::Decoration::~Decoration() {
    qDeleteAll(_stages);
}
ClientMaterial &ClientMaterial::Decoration::material() {
    DENG2_ASSERT(d->material);
    return *d->material;
}
ClientMaterial const &ClientMaterial::Decoration::material() const {
    DENG2_ASSERT(d->material);
    return *d->material;
}
void ClientMaterial::Decoration::setMaterial(ClientMaterial *newOwner) {
    d->material = newOwner;
}
Vector2i const &ClientMaterial::Decoration::patternSkip() const {
    return d->patternSkip;
}
Vector2i const &ClientMaterial::Decoration::patternOffset() const {
    return d->patternOffset;
}
int ClientMaterial::Decoration::stageCount() const {
    return _stages.count();
}
ClientMaterial::Decoration::Stage &ClientMaterial::Decoration::stage(int index) const {
    if (!stageCount()) {
        throw MissingStageError("ClientMaterial::Decoration::stage", "Decoration has no stages");
    }
    return *_stages[de::wrap(index, 0, _stages.count())];
}
String ClientMaterial::Decoration::describe() const {
    return "abstract Decoration";
}
String ClientMaterial::Decoration::description() const {
    int const numStages = stageCount();
    String str = _E(b) + describe() + _E(.) + " (" + String::number(numStages) + " stage" + DENG2_PLURAL_S(numStages) + "):";
    for (int i = 0; i < numStages; ++i) {
        str += String("\n  [%1] ").arg(i, 2) + _E(>) + stage(i).description() + _E(<);
    }
    return str;
}
DENG2_PIMPL(ClientMaterial) {
    AudioEnvironmentId audioEnvironment { AE_NONE };
    QVector<Decoration *> decorations;
    QVector<MaterialAnimator *> animators;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        self().clearAllAnimators();
        self().clearAllDecorations();
    }
    MaterialAnimator *findAnimator(MaterialVariantSpec const &spec, bool canCreate = false) {
        for (auto iter = animators.constBegin(); iter != animators.constEnd(); ++iter) {
            if ((*iter)->variantSpec().compare(spec)) {
                return const_cast<MaterialAnimator *>(*iter);  // This will do fine.
            }
        }
        if (!canCreate) return nullptr;
        animators.append(new MaterialAnimator(self(), spec));
        return animators.back();
    }
};
ClientMaterial::ClientMaterial(world::MaterialManifest &manifest)
    : world::Material(manifest)
    , d(new Impl(this))
{}
ClientMaterial::~ClientMaterial()
{}
bool ClientMaterial::isAnimated() const {
    if (Material::isAnimated()) {
        return true;
    }
    for (const auto &decor : d->decorations) {
        if (decor->isAnimated()) {
            return true;
        }
    }
    return false;
}
AudioEnvironmentId ClientMaterial::audioEnvironment() const {
    return (isDrawable()? d->audioEnvironment : AE_NONE);
}
void ClientMaterial::setAudioEnvironment(AudioEnvironmentId newEnvironment) {
    d->audioEnvironment = newEnvironment;
}
int ClientMaterial::decorationCount() const {
    return d->decorations.count();
}
LoopResult ClientMaterial::forAllDecorations(const std::function<LoopResult (Decoration &)> &func) const {
    for (Decoration *decor : d.getConst()->decorations) {
        if (auto result = func(*decor)) return result;
    }
    return LoopContinue;
}
void ClientMaterial::addDecoration(Decoration *decor) {
    if (!decor || d->decorations.contains(decor)) return;
    decor->setMaterial(this);
    d->decorations.append(decor);
}
void ClientMaterial::clearAllDecorations() {
    qDeleteAll(d->decorations); d->decorations.clear();
    clearAllAnimators();
}
int ClientMaterial::animatorCount() const {
    return d.getConst()->animators.count();
}
MaterialAnimator &ClientMaterial::animator(int index) const {
    return *d.getConst()->animators[index];
}
bool ClientMaterial::hasAnimator(MaterialVariantSpec const &spec) {
    return d->findAnimator(spec) != nullptr;
}
MaterialAnimator &ClientMaterial::getAnimator(MaterialVariantSpec const &spec) {
    return *d->findAnimator(spec, true/*create*/);
}
LoopResult ClientMaterial::forAllAnimators(const std::function<LoopResult (MaterialAnimator &)> &func) const {
    for (MaterialAnimator *animator : d.getConst()->animators) {
        if (auto result = func(*animator)) return result;
    }
    return LoopContinue;
}
void ClientMaterial::clearAllAnimators() {
    qDeleteAll(d->animators);
    d->animators.clear();
}
ClientMaterial &ClientMaterial::find(de::Uri const &uri) // static
{
    return world::Materials::get().material(uri).as<ClientMaterial>();
}
String ClientMaterial::description() const {
    String str = world::Material::description();
    str += _E(b) " x" + String::number(animatorCount()) + _E(.)
         + _E(l) " EnvClass: \"" _E(.) + (audioEnvironment() == AE_NONE? "N/A" : S_AudioEnvironmentName(audioEnvironment())) + "\"";
    for (Decoration const *decor : d->decorations) {
        str += "\n" + decor->description();
    }
    return str;
}
#define DETAILVARIANT_CONTRAST_HASHSIZE     (DETAILTEXTURE_CONTRAST_QUANTIZATION_FACTOR+1)
byte precacheMapMaterials = true;
byte precacheSprites      = true;
ClientResources &ClientResources::get() // static
{
    return static_cast<ClientResources &>(Resources::get());
}
DENG2_PIMPL(ClientResources)
, DENG2_OBSERVES(FontScheme,         ManifestDefined)
, DENG2_OBSERVES(FontManifest,       Deletion)
, DENG2_OBSERVES(AbstractFont,       Deletion)
, DENG2_OBSERVES(res::ColorPalettes, Addition)
, DENG2_OBSERVES(res::ColorPalette,  ColorTableChange) {
    typedef QHash<lumpnum_t, rawtex_t *> RawTextureHash;
    RawTextureHash rawTexHash;
    FontSchemes fontSchemes;
    QList<FontScheme *> fontSchemeCreationOrder;
    AllFonts fonts;                    ///< From all schemes.
    uint fontManifestCount;            ///< Total number of font manifests (in all schemes).
    uint fontManifestIdMapSize;
    FontManifest **fontManifestIdMap;  ///< Index with fontid_t-1
    typedef QVector<FrameModelDef> ModelDefs;
    ModelDefs modefs;
    QVector<int> stateModefs;          ///< Index to the modefs array.
    typedef StringPool ModelRepository;
    ModelRepository *modelRepository;  ///< Owns FrameModel instances.
    typedef QList<MaterialVariantSpec *> MaterialSpecs;
    MaterialSpecs materialSpecs;
    typedef QList<TextureVariantSpec *> TextureSpecs;
    TextureSpecs textureSpecs;
    TextureSpecs detailTextureSpecs[DETAILVARIANT_CONTRAST_HASHSIZE];
    struct CacheTask {
        virtual ~CacheTask() {}
        virtual void run() = 0;
    };
    /**
     * Stores the arguments for a resource cache work item.
     */
    struct MaterialCacheTask : public CacheTask {
        ClientMaterial *material;
        MaterialVariantSpec const *spec; /// Interned context specification.
        MaterialCacheTask(ClientMaterial &resource, MaterialVariantSpec const &contextSpec)
            : CacheTask()
            , material(&resource)
            , spec(&contextSpec)
        {}
        void run() {
            material->getAnimator(*spec).cacheAssets();
        }
    };
    typedef QList<CacheTask *> CacheQueue;
    CacheQueue cacheQueue;
    Impl(Public *i)
        : Base(i)
        , fontManifestCount        (0)
        , fontManifestIdMapSize    (0)
        , fontManifestIdMap        (0)
        , modelRepository          (0) {
        LOG_AS("ClientResources");
        res::TextureManifest::setTextureConstructor([] (res::TextureManifest &m) -> res::Texture * {
            return new ClientTexture(m);
        });
        createFontScheme("System");
        createFontScheme("Game");
        self().colorPalettes().audienceForAddition() += this;
    }
    ~Impl() {
        self().clearAllFontSchemes();
        clearFontManifests();
        self().clearAllRawTextures();
        self().purgeCacheQueue();
        clearAllTextureSpecs();
        clearMaterialSpecs();
        clearModels();
    }
    inline de::FS1 &fileSys() { return App_FileSystem(); }
    void clearFontManifests() {
        qDeleteAll(fontSchemes);
        fontSchemes.clear();
        fontSchemeCreationOrder.clear();
        if (fontManifestIdMap) {
            M_Free(fontManifestIdMap); fontManifestIdMap = 0;
            fontManifestIdMapSize = 0;
        }
        fontManifestCount = 0;
    }
    void createFontScheme(String name) {
        DENG2_ASSERT(name.length() >= FontScheme::min_name_length);
        FontScheme *newScheme = new FontScheme(name);
        fontSchemes.insert(name.toLower(), newScheme);
        fontSchemeCreationOrder.append(newScheme);
        newScheme->audienceForManifestDefined += this;
    }
    void clearRuntimeFonts() {
        self().fontScheme("Game").clear();
        self().pruneUnusedTextureSpecs();
    }
    void clearSystemFonts() {
        self().fontScheme("System").clear();
        self().pruneUnusedTextureSpecs();
    }
    void clearMaterialSpecs() {
        qDeleteAll(materialSpecs);
        materialSpecs.clear();
    }
    MaterialVariantSpec *findMaterialSpec(MaterialVariantSpec const &tpl,
        bool canCreate) {
        foreach (MaterialVariantSpec *spec, materialSpecs) {
            if (spec->compare(tpl)) return spec;
        }
        if (!canCreate) return 0;
        materialSpecs.append(new MaterialVariantSpec(tpl));
        return materialSpecs.back();
    }
    MaterialVariantSpec &getMaterialSpecForContext(MaterialContextId contextId,
        int flags, byte border, int tClass, int tMap, int wrapS, int wrapT,
        int minFilter, int magFilter, int anisoFilter,
        bool mipmapped, bool gammaCorrection, bool noStretch, bool toAlpha) {
        static MaterialVariantSpec tpl;
        texturevariantusagecontext_t primaryContext = TC_UNKNOWN;
        switch (contextId) {
        case UiContext:         primaryContext = TC_UI;                 break;
        case MapSurfaceContext: primaryContext = TC_MAPSURFACE_DIFFUSE; break;
        case SpriteContext:     primaryContext = TC_SPRITE_DIFFUSE;     break;
        case ModelSkinContext:  primaryContext = TC_MODELSKIN_DIFFUSE;  break;
        case PSpriteContext:    primaryContext = TC_PSPRITE_DIFFUSE;    break;
        case SkySphereContext:  primaryContext = TC_SKYSPHERE_DIFFUSE;  break;
        default: DENG2_ASSERT(false);
        }
        TextureVariantSpec const &primarySpec =
            self().textureSpec(primaryContext, flags, border, tClass, tMap,
                             wrapS, wrapT, minFilter, magFilter,
                             anisoFilter, mipmapped, gammaCorrection,
                             noStretch, toAlpha);
        tpl.contextId     = contextId;
        tpl.primarySpec = &primarySpec;
        return *findMaterialSpec(tpl, true);
    }
    static int hashDetailTextureSpec(detailvariantspecification_t const &spec) {
        return (spec.contrast * (1/255.f) * DETAILTEXTURE_CONTRAST_QUANTIZATION_FACTOR + .5f);
    }
    static variantspecification_t &configureTextureSpec(variantspecification_t &spec,
        texturevariantusagecontext_t tc, int flags, byte border, int tClass, int tMap,
        int wrapS, int wrapT, int minFilter, int magFilter, int anisoFilter,
        dd_bool mipmapped, dd_bool gammaCorrection, dd_bool noStretch, dd_bool toAlpha) {
        DENG2_ASSERT(tc == TC_UNKNOWN || VALID_TEXTUREVARIANTUSAGECONTEXT(tc));
        flags &= ~TSF_INTERNAL_MASK;
        spec.context         = tc;
        spec.flags           = flags;
        spec.border          = (flags & TSF_UPSCALE_AND_SHARPEN)? 1 : border;
        spec.mipmapped       = mipmapped;
        spec.wrapS           = wrapS;
        spec.wrapT           = wrapT;
        spec.minFilter       = de::clamp(-1, minFilter, spec.mipmapped? 3:1);
        spec.magFilter       = de::clamp(-3, magFilter, 1);
        spec.anisoFilter     = de::clamp(-1, anisoFilter, 4);
        spec.gammaCorrection = gammaCorrection;
        spec.noStretch       = noStretch;
        spec.toAlpha         = toAlpha;
        if (tClass || tMap) {
            spec.flags      |= TSF_HAS_COLORPALETTE_XLAT;
            spec.tClass      = de::max(0, tClass);
            spec.tMap        = de::max(0, tMap);
        }
        return spec;
    }
    static detailvariantspecification_t &configureDetailTextureSpec(
        detailvariantspecification_t &spec, float contrast) {
        int const quantFactor = DETAILTEXTURE_CONTRAST_QUANTIZATION_FACTOR;
        spec.contrast = 255 * de::clamp<int>(0, contrast * quantFactor + .5f, quantFactor) * (1 / float(quantFactor));
        return spec;
    }
    TextureVariantSpec &linkTextureSpec(TextureVariantSpec *spec) {
        DENG2_ASSERT(spec != 0);
        switch (spec->type) {
        case TST_GENERAL:
            textureSpecs.append(spec);
            break;
        case TST_DETAIL: {
            int hash = hashDetailTextureSpec(spec->detailVariant);
            detailTextureSpecs[hash].append(spec);
            break; }
        }
        return *spec;
    }
    TextureVariantSpec *findTextureSpec(TextureVariantSpec const &tpl, bool canCreate) {
        switch (tpl.type) {
        case TST_GENERAL: {
            foreach (TextureVariantSpec *varSpec, textureSpecs) {
                if (*varSpec == tpl) {
                    return varSpec;
                }
            }
            break; }
        case TST_DETAIL: {
            int hash = hashDetailTextureSpec(tpl.detailVariant);
            foreach (TextureVariantSpec *varSpec, detailTextureSpecs[hash]) {
                if (*varSpec == tpl) {
                    return varSpec;
                }
            }
            break; }
        }
        if (canCreate) {
            return &linkTextureSpec(new TextureVariantSpec(tpl));
        }
        return 0;
    }
    TextureVariantSpec *textureSpec(texturevariantusagecontext_t tc, int flags,
        byte border, int tClass, int tMap, int wrapS, int wrapT, int minFilter,
        int magFilter, int anisoFilter, dd_bool mipmapped, dd_bool gammaCorrection,
        dd_bool noStretch, dd_bool toAlpha) {
        static TextureVariantSpec tpl;
        tpl.type = TST_GENERAL;
        configureTextureSpec(tpl.variant, tc, flags, border, tClass, tMap, wrapS,
            wrapT, minFilter, magFilter, anisoFilter, mipmapped, gammaCorrection,
            noStretch, toAlpha);
        return findTextureSpec(tpl, true);
    }
    TextureVariantSpec *detailTextureSpec(float contrast) {
        static TextureVariantSpec tpl;
        tpl.type = TST_DETAIL;
        configureDetailTextureSpec(tpl.detailVariant, contrast);
        return findTextureSpec(tpl, true);
    }
    bool textureSpecInUse(TextureVariantSpec const &spec) {
        for (res::Texture *texture : self().textures().allTextures()) {
            for (TextureVariant *variant : static_cast<ClientTexture *>(texture)->variants()) {
                if (&variant->spec() == &spec) {
                    return true; // Found one; stop.
                }
            }
        }
        return false;
    }
    int pruneUnusedTextureSpecs(TextureSpecs &list) {
        int numPruned = 0;
        QMutableListIterator<TextureVariantSpec *> it(list);
        while (it.hasNext()) {
            TextureVariantSpec *spec = it.next();
            if (!textureSpecInUse(*spec)) {
                it.remove();
                delete spec;
                numPruned += 1;
            }
        }
        return numPruned;
    }
    int pruneUnusedTextureSpecs(texturevariantspecificationtype_t specType) {
        switch (specType) {
        case TST_GENERAL: return pruneUnusedTextureSpecs(textureSpecs);
        case TST_DETAIL: {
            int numPruned = 0;
            for (int i = 0; i < DETAILVARIANT_CONTRAST_HASHSIZE; ++i) {
                numPruned += pruneUnusedTextureSpecs(detailTextureSpecs[i]);
            }
            return numPruned; }
        }
        return 0;
    }
    void clearAllTextureSpecs() {
        qDeleteAll(textureSpecs);
        textureSpecs.clear();
        for (int i = 0; i < DETAILVARIANT_CONTRAST_HASHSIZE; ++i) {
            qDeleteAll(detailTextureSpecs[i]);
            detailTextureSpecs[i].clear();
        }
    }
    void processCacheQueue() {
        while (!cacheQueue.isEmpty()) {
            QScopedPointer<CacheTask> task(cacheQueue.takeFirst());
            task->run();
        }
    }
    void queueCacheTasksForMaterial(ClientMaterial &material,
                                    MaterialVariantSpec const &contextSpec,
                                    bool cacheGroups = true) {
        bool alreadyQueued = false;
        foreach (CacheTask *baseTask, cacheQueue) {
            if (MaterialCacheTask *task = dynamic_cast<MaterialCacheTask *>(baseTask)) {
                if (&material == task->material && &contextSpec == task->spec) {
                    alreadyQueued = true;
                    break;
                }
            }
        }
        if (!alreadyQueued) {
            cacheQueue.append(new MaterialCacheTask(material, contextSpec));
        }
        if (!cacheGroups) return;
        foreach (world::Materials::MaterialManifestGroup *group,
                 world::Materials::get().allMaterialGroups()) {
            if (!group->contains(&material.manifest())) {
                continue;
            }
            foreach (world::MaterialManifest *manifest, *group) {
                if (!manifest->hasMaterial()) continue;
                if (&manifest->material() == &material) continue;
                queueCacheTasksForMaterial(manifest->material().as<ClientMaterial>(),
                                           contextSpec, false /* do not cache groups */);
            }
        }
    }
    void queueCacheTasksForSprite(spritenum_t id,
                                  MaterialVariantSpec const &contextSpec,
                                  bool cacheGroups = true) {
        if (auto const *sprites = self().sprites().tryFindSpriteSet(id)) {
            for (Record const &sprite : *sprites) {
                defn::Sprite const spriteDef(sprite);
                for (auto const &view : spriteDef.def().compiled().views) {
                    if (world::Material *material = world::Materials::get().materialPtr(view.uri)) {
                        queueCacheTasksForMaterial(material->as<ClientMaterial>(),
                                                   contextSpec, cacheGroups);
                    }
                }
            }
        }
    }
    void queueCacheTasksForModel(FrameModelDef &modelDef) {
        if (!useModels) return;
        for (duint sub = 0; sub < modelDef.subCount(); ++sub) {
            SubmodelDef &subdef = modelDef.subModelDef(sub);
            FrameModel *mdl = modelForId(subdef.modelId);
            if (!mdl) continue;
            for (FrameModelSkin const &skin : mdl->skins()) {
                if (ClientTexture *tex = static_cast<ClientTexture *>(skin.texture)) {
                    tex->prepareVariant(Rend_ModelDiffuseTextureSpec(mdl->flags().testFlag(FrameModel::NoTextureCompression)));
                }
            }
            if (ClientTexture *shinyTex = static_cast<ClientTexture *>(subdef.shinySkin)) {
                shinyTex->prepareVariant(Rend_ModelShinyTextureSpec());
            }
        }
    }
    void clearModels() {
        modefs.clear();
        stateModefs.clear();
        clearModelList();
        if (modelRepository) {
            delete modelRepository; modelRepository = nullptr;
        }
    }
    FrameModel *modelForId(modelid_t id) {
        DENG2_ASSERT(modelRepository);
        return reinterpret_cast<FrameModel *>(modelRepository->userPointer(id));
    }
    inline String const &findModelPath(modelid_t id) {
        return modelRepository->stringRef(id);
    }
    /**
     * Create a new modeldef or find an existing one. This is for ID'd models.
     */
    FrameModelDef *getModelDefWithId(String id) {
        if (id.isEmpty()) return nullptr;
        if (self().hasModelDef(id)) {
            return &self().modelDef(id);
        }
        modefs.append(FrameModelDef(id.toUtf8().constData()));
        return &modefs.last();
    }
    /**
     * Create a new modeldef or find an existing one. There can be only one model
     * definition associated with a state/intermark pair.
     */
    FrameModelDef *getModelDef(dint state, dfloat interMark, dint select) {
        if (state < 0 || state >= runtimeDefs.states.size()) {
            return nullptr;
        }
        for (FrameModelDef const &modef : modefs) {
            if (modef.state == &runtimeDefs.states[state] &&
               fequal(modef.interMark, interMark) && modef.select == select) {
                return nullptr;
            }
        }
        modefs.append(FrameModelDef());
        FrameModelDef *md = &modefs.last();
        md->state     = &runtimeDefs.states[state];
        md->interMark = interMark;
        md->select    = select;
        return md;
    }
    String findSkinPath(Path const &skinPath, Path const &modelFilePath) {
        if (!modelFilePath.isEmpty()) {
            try {
                return fileSys().findPath(de::Uri("Models", modelFilePath.toString().fileNamePath() / skinPath.fileName()),
                                          RLF_DEFAULT, self().resClass(RC_GRAPHIC));
            }
            catch (FS1::NotFoundError const &)
            {}  // Ignore this error.
        }
        return fileSys().findPath(de::Uri("Models", skinPath), RLF_DEFAULT,
                                  self().resClass(RC_GRAPHIC));
    }
    /**
     * Allocate room for a new skin file name.
     */
    short defineSkinAndAddToModelIndex(FrameModel &mdl, Path const &skinPath) {
        if (ClientTexture *tex = static_cast<ClientTexture *>(self().textures().defineTexture("ModelSkins", de::Uri(skinPath)))) {
            for (dint i = 0; i < mdl.skinCount(); ++i) {
                if (mdl.skin(i).texture == tex)
                    return i;
            }
            mdl.newSkin(skinPath.toString()).texture = tex;
            return mdl.skinCount() - 1;
        }
        return -1;
    }
    void defineAllSkins(FrameModel &mdl) {
        String const &modelFilePath = findModelPath(mdl.modelId());
        dint numFoundSkins = 0;
        for (dint i = 0; i < mdl.skinCount(); ++i) {
            FrameModelSkin &skin = mdl.skin(i);
            try {
                de::Uri foundResourceUri(Path(findSkinPath(skin.name, modelFilePath)));
                skin.texture = self().textures().defineTexture("ModelSkins", foundResourceUri);
                numFoundSkins += 1;
            }
            catch (FS1::NotFoundError const &) {
                LOG_RES_VERBOSE("Failed to locate \"%s\" (#%i) for model \"%s\"")
                        << skin.name << i << NativePath(modelFilePath).pretty();
            }
        }
        if (!numFoundSkins) {
            de::Uri searchPath(modelFilePath.fileNamePath() / modelFilePath.fileNameWithoutExtension(), RC_GRAPHIC);
            try {
                String foundPath = fileSys().findPath(searchPath, RLF_DEFAULT,
                                                      self().resClass(RC_GRAPHIC));
                foundPath = App_BasePath() / foundPath;
                defineSkinAndAddToModelIndex(mdl, foundPath);
                numFoundSkins = 1;
                LOG_RES_MSG("Assigned fallback skin \"%s\" to index #0 for model \"%s\"")
                    << NativePath(foundPath).pretty()
                    << NativePath(modelFilePath).pretty();
            }
            catch (FS1::NotFoundError const &)
            {}  // Ignore this error.
        }
        if (!numFoundSkins) {
            LOG_RES_MSG("No skins found for model \"%s\" (it may use a custom skin specified in a DED)")
                << NativePath(modelFilePath).pretty();
        }
#ifdef DENG_DEBUG
        LOGDEV_RES_XVERBOSE("Model \"%s\" skins:", NativePath(modelFilePath).pretty());
        dint skinIdx = 0;
        for (FrameModelSkin const &skin : mdl.skins()) {
            res::TextureManifest const *texManifest = skin.texture? &skin.texture->manifest() : 0;
            LOGDEV_RES_XVERBOSE("  %i: %s %s",
                       (skinIdx++) << skin.name
                    << (texManifest? (String("\"") + texManifest->composeUri() + "\"") : "(missing texture)")
                    << (texManifest? (String(" => \"") + NativePath(texManifest->resourceUri().compose()).pretty() + "\"") : ""));
        }
#endif
    }
    /**
     * Scales the given model so that it'll be 'destHeight' units tall. Measurements
     * are based on submodel zero. Scale is applied uniformly.
     */
    void scaleModel(FrameModelDef &mf, dfloat destHeight, dfloat offset) {
        if (!mf.subCount()) return;
        SubmodelDef &smf = mf.subModelDef(0);
        if (!smf.modelId) return;
        dfloat top, bottom;
        dfloat height = self().model(smf.modelId).frame(smf.frame).horizontalRange(&top, &bottom);
        if (fequal(height, 0.f)) height = 1;
        dfloat scale = destHeight / height;
        mf.scale    = Vector3f(scale, scale, scale);
        mf.offset.y = -bottom * scale + offset;
    }
    void scaleModelToSprite(FrameModelDef &mf, Record const *spriteRec) {
        if (!spriteRec) return;
        defn::Sprite sprite(*spriteRec);
        if (!sprite.hasView(0)) return;
        world::Material *mat = world::Materials::get().materialPtr(sprite.viewMaterial(0));
        if (!mat) return;
        MaterialAnimator &matAnimator = mat->as<ClientMaterial>().getAnimator(Rend_SpriteMaterialSpec());
        matAnimator.prepare();  // Ensure we have up-to-date info.
        ClientTexture const &texture = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture->base();
        dint off = de::max(0, -texture.origin().y - int(matAnimator.dimensions().y));
        scaleModel(mf, matAnimator.dimensions().y, off);
    }
    dfloat calcModelVisualRadius(FrameModelDef *def) {
        if (!def || !def->subModelId(0)) return 0;
        Vector3f min, max;
        dfloat maxRadius = 0;
        for (duint i = 0; i < def->subCount(); ++i) {
            if (!def->subModelId(i)) break;
            SubmodelDef &sub = def->subModelDef(i);
            self().model(sub.modelId).frame(sub.frame).bounds(min, max);
            dfloat radius = (  def->scale.x * (max.x - min.x)
                             + def->scale.z * (max.z - min.z)) / 3.5f;
            if (radius > maxRadius) {
                maxRadius = radius;
            }
        }
        return maxRadius;
    }
    /**
     * Creates a modeldef based on the given DED info. A pretty straightforward
     * operation. No interlinks are set yet. Autoscaling is done and the scale
     * factors set appropriately. After this has been called for all available
     * Model DEDs, each State that has a model will have a pointer to the one
     * with the smallest intermark (start of a chain).
     */
    void setupModel(defn::Model const &def) {
        LOG_AS("setupModel");
        auto &defs = *DED_Definitions();
        dint const modelScopeFlags = def.geti("flags") | defs.modelFlags;
        dint const statenum = defs.getStateNum(def.gets("state"));
        FrameModelDef *modef = getModelDefWithId(def.gets("id"));
        if (!modef) {
            if (statenum < 0) return;
            modef = getModelDef(statenum + def.geti("off"), def.getf("interMark"), def.geti("selector"));
            if (!modef) return; // Overridden or invalid definition.
        }
        modef->def       = def;
        modef->group     = def.getui("group");
        modef->flags     = modelScopeFlags;
        modef->offset    = Vector3f(def.get("offset"));
        modef->offset.y += defs.modelOffset; // Common Y axis offset.
        modef->scale     = Vector3f(def.get("scale"));
        modef->scale.y  *= defs.modelScale;  // Common Y axis scaling.
        modef->resize    = def.getf("resize");
        modef->skinTics  = de::max(def.geti("skinTics"), 1);
        for (dint i = 0; i < 2; ++i) {
            modef->interRange[i] = float(def.geta("interRange")[i].asNumber());
        }
        modef->clearSubs();
        for (dint i = 0; i < def.subCount(); ++i) {
            Record const &subdef = def.sub(i);
            SubmodelDef *sub = modef->addSub();
            sub->modelId = 0;
            if (subdef.gets("filename").isEmpty()) continue;
            de::Uri const searchPath(subdef.gets("filename"));
            if (searchPath.isEmpty()) continue;
            try {
                String foundPath = fileSys().findPath(searchPath, RLF_DEFAULT,
                                                      self().resClass(RC_MODEL));
                foundPath = App_BasePath() / foundPath;
                modelid_t modelId = modelRepository->intern(foundPath);
                FrameModel *mdl = modelForId(modelId);
                if (!mdl) {
                    QScopedPointer<FileHandle> hndl(&fileSys().openFile(foundPath, "rb"));
                    mdl = FrameModel::loadFromFile(*hndl, modelAspectMod);
                    fileSys().releaseFile(hndl->file());
                    if (mdl) {
                        mdl->setModelId(modelId);
                        modelRepository->setUserPointer(modelId, mdl);
                        defineAllSkins(*mdl);
                        if (!Rend_ModelExpandVertexBuffers(mdl->vertexCount())) {
                            LOG_RES_WARNING("Model \"%s\" contains more than %u max vertices (%i), it will not be rendered")
                                << NativePath(foundPath).pretty()
                                << uint(RENDER_MAX_MODEL_VERTS) << mdl->vertexCount();
                        }
                    }
                }
                if (!mdl) continue;
                sub->modelId    = mdl->modelId();
                sub->frame      = mdl->frameNumber(subdef.gets("frame"));
                if (sub->frame < 0) sub->frame = 0;
                sub->frameRange = de::max(1, subdef.geti("frameRange")); // Frame range must always be greater than zero.
                sub->alpha      = byte(de::clamp(0, int(255 - subdef.getf("alpha") * 255), 255));
                sub->blendMode  = blendmode_t(subdef.geti("blendMode"));
                sub->setFlags(modelScopeFlags ^ subdef.geti("flags"));
                if (sub->testFlag(MFF_BRIGHTSHADOW)) {
                    sub->alpha = byte(256 * .80f);
                    sub->blendMode = BM_ADD;
                }
                else if (sub->testFlag(MFF_BRIGHTSHADOW2)) {
                    sub->blendMode = BM_ADD;
                }
                else if (sub->testFlag(MFF_DARKSHADOW)) {
                    sub->blendMode = BM_DARK;
                }
                else if (sub->testFlag(MFF_SHADOW2)) {
                    sub->alpha = byte(256 * .2f);
                }
                else if (sub->testFlag(MFF_SHADOW1)) {
                    sub->alpha = byte(256 * .62f);
                }
                if (sub->testFlag(MFF_REVERSE_SUBTRACT)) {
                    sub->blendMode = BM_REVERSE_SUBTRACT;
                }
                else if (sub->testFlag(MFF_SUBTRACT)) {
                    sub->blendMode = BM_SUBTRACT;
                }
                if (!subdef.gets("skinFilename").isEmpty()) {
                    String const &skinFilePath  = de::Uri(subdef.gets("skinFilename")).path();
                    String const &modelFilePath = findModelPath(sub->modelId);
                    try {
                        Path foundResourcePath(findSkinPath(skinFilePath, modelFilePath));
                        sub->skin = defineSkinAndAddToModelIndex(*mdl, foundResourcePath);
                    }
                    catch (FS1::NotFoundError const &) {
                        LOG_RES_WARNING("Failed to locate skin \"%s\" for model \"%s\"")
                            << subdef.gets("skinFilename") << NativePath(modelFilePath).pretty();
                    }
                }
                else {
                    sub->skin = subdef.geti("skin");
                }
                sub->skinRange = de::max(subdef.geti("skinRange"), 1);
                sub->offset = subdef.get("offset");
                if (!subdef.gets("shinySkin").isEmpty()) {
                    String const &skinFilePath  = de::Uri(subdef.gets("shinySkin")).path();
                    String const &modelFilePath = findModelPath(sub->modelId);
                    try {
                        de::Uri foundResourceUri(Path(findSkinPath(skinFilePath, modelFilePath)));
                        sub->shinySkin = self().textures().defineTexture("ModelReflectionSkins", foundResourceUri);
                    }
                    catch (FS1::NotFoundError const &) {
                        LOG_RES_WARNING("Failed to locate skin \"%s\" for model \"%s\"")
                            << skinFilePath << NativePath(modelFilePath).pretty();
                    }
                }
                else {
                    sub->shinySkin = 0;
                }
                if (sub->testFlag(MFF_NO_TEXCOMP)) {
                    mdl->setFlags(FrameModel::NoTextureCompression);
                }
            }
            catch (FS1::NotFoundError const &) {
                LOG_RES_WARNING("Failed to locate \"%s\"") << searchPath;
            }
        }
        if (modef->resize) {
            scaleModel(*modef, modef->resize, modef->offset.y);
        }
        else if (modef->state && modef->testSubFlag(0, MFF_AUTOSCALE)) {
            spritenum_t sprNum = DED_Definitions()->getSpriteNum(def.gets("sprite"));
            int sprFrame       = def.geti("spriteFrame");
            if (sprNum < 0) {
                sprNum   = modef->state->sprite;
                sprFrame = modef->state->frame;
            }
            if (Record const *sprite = self().sprites().spritePtr(sprNum, sprFrame)) {
                scaleModelToSprite(*modef, sprite);
            }
        }
        if (modef->state) {
            int stateNum = runtimeDefs.states.indexOf(modef->state);
            if (stateModefs[stateNum] < 0) {
                stateModefs[stateNum] = self().indexOf(modef);
            }
            else {
                FrameModelDef *other = self().modelDefForState(stateNum);
                if ((modef->interMark <= other->interMark && // Should never be ==
                    modef->select == other->select) || modef->select < other->select) // Smallest selector?
                {
                    stateModefs[stateNum] = self().indexOf(modef);
                }
            }
        }
        Vector3f min, max;
        for (uint i = 0; i < modef->subCount(); ++i) {
            SubmodelDef *sub = &modef->subModelDef(i);
            if (sub->modelId && sub->frame >= 0) {
                self().model(sub->modelId).frame(sub->frame).bounds(min, max);
                modef->setParticleOffset(i, ((max + min) / 2 + sub->offset) * modef->scale + modef->offset);
            }
        }
        modef->visualRadius = calcModelVisualRadius(modef); // based on geometry bounds
        modef->shadowRadius = def.getf("shadowRadius");
    }
    void clearModelList() {
        if (!modelRepository) return;
        modelRepository->forAll([this] (StringPool::Id id) {
            if (auto *model = reinterpret_cast<FrameModel *>(modelRepository->userPointer(id))) {
                modelRepository->setUserPointer(id, nullptr);
                delete model;
            }
            return LoopContinue;
        });
    }
    void fontSchemeManifestDefined(FontScheme & /*scheme*/, FontManifest &manifest)
    {
        manifest.audienceForDeletion += this;
        fontid_t const id = ++fontManifestCount; // 1-based.
        manifest.setUniqueId(id);
        if (fontManifestCount > fontManifestIdMapSize) {
            fontManifestIdMapSize += 32;
            fontManifestIdMap = (FontManifest **) M_Realloc(fontManifestIdMap, sizeof(*fontManifestIdMap) * fontManifestIdMapSize);
        }
        fontManifestIdMap[fontManifestCount - 1] = &manifest;
    }
#if 0
    void fontManifestFontDerived(FontManifest & /*manifest*/, AbstractFont &font)
    {
        fonts.append(&font);
        font.audienceForDeletion += this;
    }
#endif
    void fontManifestBeingDeleted(FontManifest const &manifest) {
        fontManifestIdMap[manifest.uniqueId() - 1 /*1-based*/] = 0;
        fontManifestCount -= 1;
    }
    void fontBeingDeleted(AbstractFont const &font) {
        fonts.removeOne(const_cast<AbstractFont *>(&font));
    }
    void colorPaletteAdded(res::ColorPalette &newPalette) {
        newPalette.audienceForColorTableChange += this;
    }
    void colorPaletteColorTableChanged(res::ColorPalette &colorPalette) {
        foreach (res::Texture *texture, self().textures().allTextures()) {
            colorpalette_analysis_t *cp = reinterpret_cast<colorpalette_analysis_t *>(texture->analysisDataPointer(res::Texture::ColorPaletteAnalysis));
            if (cp && cp->paletteId == colorpaletteid_t(colorPalette.id())) {
                texture->release();
            }
        }
    }
};
ClientResources::ClientResources() : d(new Impl(this))
{}
void ClientResources::clear() {
    Resources::clear();
    R_ShutdownSvgs();
}
void ClientResources::clearAllRuntimeResources() {
    Resources::clearAllRuntimeResources();
    d->clearRuntimeFonts();
    pruneUnusedTextureSpecs();
}
void ClientResources::clearAllSystemResources() {
    Resources::clearAllSystemResources();
    d->clearSystemFonts();
    pruneUnusedTextureSpecs();
}
void ClientResources::initSystemTextures() {
    Resources::initSystemTextures();
    if (novideo) return;
    LOG_AS("ClientResources");
    static struct {
        String const graphicName;
        Path const path;
    } const texDefs[] = {
        { "bbox",       "bbox" },
        { "gray",       "gray" },
    };
    LOG_RES_VERBOSE("Initializing System textures...");
    for (auto const &def : texDefs) {
        textures().declareSystemTexture(def.path, de::Uri("Graphics", def.graphicName));
    }
    textures().deriveAllTexturesInScheme("System");
}
void ClientResources::reloadAllResources() {
    DENG2_ASSERT_IN_MAIN_THREAD();
    DENG2_ASSERT(QOpenGLContext::currentContext() != nullptr);
    Resources::reloadAllResources();
    DD_UpdateEngineState();
}
rawtex_t *ClientResources::rawTexture(lumpnum_t lumpNum) {
    LOG_AS("ClientResources::rawTexture");
    if (-1 == lumpNum || lumpNum >= App_FileSystem().lumpCount()) {
        LOGDEV_RES_WARNING("LumpNum #%i out of bounds (%i), returning 0")
                << lumpNum << App_FileSystem().lumpCount();
        return nullptr;
    }
    auto found = d->rawTexHash.find(lumpNum);
    return (found != d->rawTexHash.end() ? found.value() : nullptr);
}
rawtex_t *ClientResources::declareRawTexture(lumpnum_t lumpNum) {
    LOG_AS("ClientResources::rawTexture");
    if (-1 == lumpNum || lumpNum >= App_FileSystem().lumpCount()) {
        LOGDEV_RES_WARNING("LumpNum #%i out of range %s, returning 0")
            << lumpNum << Rangeui(0, App_FileSystem().lumpCount()).asText();
        return nullptr;
    }
    rawtex_t *raw = rawTexture(lumpNum);
    if (!raw) {
        raw = new rawtex_t(App_FileSystem().lump(lumpNum).name(), lumpNum);
        d->rawTexHash.insert(lumpNum, raw);
    }
    return raw;
}
QList<rawtex_t *> ClientResources::collectRawTextures() const {
    return d->rawTexHash.values();
}
void ClientResources::clearAllRawTextures() {
    qDeleteAll(d->rawTexHash);
    d->rawTexHash.clear();
}
void ClientResources::releaseAllSystemGLTextures() {
    if (::novideo) return;
    LOG_AS("ResourceSystem");
    LOG_RES_VERBOSE("Releasing system textures...");
    ClientApp::renderSystem().clearDrawLists();
    GL_ReleaseAllLightingSystemTextures();
    GL_ReleaseAllFlareTextures();
    releaseGLTexturesByScheme("System");
    Rend_ParticleReleaseSystemTextures();
    releaseFontGLTexturesByScheme("System");
    pruneUnusedTextureSpecs();
}
void ClientResources::releaseAllRuntimeGLTextures() {
    if (::novideo) return;
    LOG_AS("ResourceSystem");
    LOG_RES_VERBOSE("Releasing runtime textures...");
    ClientApp::renderSystem().clearDrawLists();
    releaseGLTexturesByScheme("Flats");
    releaseGLTexturesByScheme("Textures");
    releaseGLTexturesByScheme("Patches");
    releaseGLTexturesByScheme("Sprites");
    releaseGLTexturesByScheme("Details");
    releaseGLTexturesByScheme("Reflections");
    releaseGLTexturesByScheme("Masks");
    releaseGLTexturesByScheme("ModelSkins");
    releaseGLTexturesByScheme("ModelReflectionSkins");
    releaseGLTexturesByScheme("Lightmaps");
    releaseGLTexturesByScheme("Flaremaps");
    GL_ReleaseTexturesForRawImages();
    Rend_ParticleReleaseExtraTextures();
    releaseFontGLTexturesByScheme("Game");
    pruneUnusedTextureSpecs();
}
void ClientResources::releaseAllGLTextures() {
    releaseAllRuntimeGLTextures();
    releaseAllSystemGLTextures();
}
void ClientResources::releaseGLTexturesByScheme(String schemeName) {
    if (schemeName.isEmpty()) return;
    PathTreeIterator<res::TextureScheme::Index> iter(textures().textureScheme(schemeName).index().leafNodes());
    while (iter.hasNext()) {
        res::TextureManifest &manifest = iter.next();
        if (manifest.hasTexture()) {
            manifest.texture().release();
        }
    }
}
void ClientResources::clearAllTextureSpecs() {
    d->clearAllTextureSpecs();
}
void ClientResources::pruneUnusedTextureSpecs() {
    if (Sys_IsShuttingDown()) return;
    dint numPruned = 0;
    numPruned += d->pruneUnusedTextureSpecs(TST_GENERAL);
    numPruned += d->pruneUnusedTextureSpecs(TST_DETAIL);
    LOGDEV_RES_VERBOSE("Pruned %i unused texture variant %s")
        << numPruned << (numPruned == 1? "specification" : "specifications");
}
TextureVariantSpec const &ClientResources::textureSpec(texturevariantusagecontext_t tc,
    dint flags, byte border, dint tClass, dint tMap, dint wrapS, dint wrapT, dint minFilter,
    dint magFilter, dint anisoFilter, dd_bool mipmapped, dd_bool gammaCorrection,
    dd_bool noStretch, dd_bool toAlpha) {
    TextureVariantSpec *tvs =
        d->textureSpec(tc, flags, border, tClass, tMap, wrapS, wrapT, minFilter,
                       magFilter, anisoFilter, mipmapped, gammaCorrection,
                       noStretch, toAlpha);
#ifdef DENG_DEBUG
    if (tClass || tMap) {
        DENG2_ASSERT(tvs->variant.flags & TSF_HAS_COLORPALETTE_XLAT);
        DENG2_ASSERT(tvs->variant.tClass == tClass);
        DENG2_ASSERT(tvs->variant.tMap == tMap);
    }
#endif
    return *tvs;
}
TextureVariantSpec &ClientResources::detailTextureSpec(dfloat contrast) {
    return *d->detailTextureSpec(contrast);
}
FontScheme &ClientResources::fontScheme(String name) const {
    LOG_AS("ClientResources::fontScheme");
    if (!name.isEmpty()) {
        FontSchemes::iterator found = d->fontSchemes.find(name.toLower());
        if (found != d->fontSchemes.end()) return **found;
    }
    throw UnknownSchemeError("ClientResources::fontScheme", "No scheme found matching '" + name + "'");
}
bool ClientResources::knownFontScheme(String name) const {
    if (!name.isEmpty()) {
        return d->fontSchemes.contains(name.toLower());
    }
    return false;
}
ClientResources::FontSchemes const &ClientResources::allFontSchemes() const {
    return d->fontSchemes;
}
bool ClientResources::hasFont(de::Uri const &path) const {
    try {
        fontManifest(path);
        return true;
    }
    catch (MissingResourceManifestError const &)
    {}  // Ignore this error.
    return false;
}
FontManifest &ClientResources::fontManifest(de::Uri const &uri) const {
    LOG_AS("ClientResources::findFont");
    if (!uri.scheme().compareWithoutCase("urn")) {
        String const &pathStr = uri.path().toStringRef();
        dint uIdPos = pathStr.indexOf(':');
        if (uIdPos > 0) {
            String schemeName = pathStr.left(uIdPos);
            dint uniqueId     = pathStr.mid(uIdPos + 1 /*skip delimiter*/).toInt();
            try {
                return fontScheme(schemeName).findByUniqueId(uniqueId);
            }
            catch (FontScheme::NotFoundError const &)
            {}  // Ignore, we'll throw our own...
        }
    }
    else {
        String const &path = uri.path();
        if (!uri.scheme().isEmpty()) {
            try {
                return fontScheme(uri.scheme()).find(path);
            }
            catch (FontScheme::NotFoundError const &)
            {}  // Ignore, we'll throw our own...
        }
        else {
            for (FontScheme *scheme : d->fontSchemeCreationOrder) {
                try {
                    return scheme->find(path);
                }
                catch (FontScheme::NotFoundError const &)
                {}  // Ignore, we'll throw our own...
            }
        }
    }
    throw MissingResourceManifestError("ClientResources::findFont", "Failed to locate a manifest matching \"" + uri.asText() + "\"");
}
FontManifest &ClientResources::toFontManifest(fontid_t id) const {
    if (id > 0 && id <= d->fontManifestCount) {
        duint32 idx = id - 1;  // 1-based index.
        if (d->fontManifestIdMap[idx]) {
            return *d->fontManifestIdMap[idx];
        }
        DENG2_ASSERT(!"Bookkeeping error");
    }
    throw UnknownFontIdError("ClientResources::toFontManifest", QString("Invalid font ID %1, valid range [1..%2)").arg(id).arg(d->fontManifestCount + 1));
}
ClientResources::AllFonts const &ClientResources::allFonts() const {
    return d->fonts;
}
AbstractFont *ClientResources::newFontFromDef(ded_compositefont_t const &def) {
    LOG_AS("ClientResources::newFontFromDef");
    if (!def.uri) return nullptr;
    de::Uri const &uri = *def.uri;
    try {
        FontManifest &manifest = declareFont(uri);
        if (manifest.hasResource()) {
            if (auto *compFont = maybeAs<CompositeBitmapFont>(manifest.resource())) {
                LOGDEV_RES_XVERBOSE("Font with uri \"%s\" already exists, returning existing",
                                    manifest.composeUri());
                compFont->rebuildFromDef(def);
            }
            return &manifest.resource();
        }
        manifest.setResource(CompositeBitmapFont::fromDef(manifest, def));
        if (manifest.hasResource()) {
            if (verbose >= 1) {
                LOG_RES_VERBOSE("New font \"%s\"")
                    << manifest.composeUri();
            }
            return &manifest.resource();
        }
        LOG_RES_WARNING("Failed defining new Font for \"%s\"")
            << NativePath(uri.asText()).pretty();
    }
    catch (UnknownSchemeError const &er) {
        LOG_RES_WARNING("Failed declaring font \"%s\": %s")
            << NativePath(uri.asText()).pretty() << er.asText();
    }
    catch (FontScheme::InvalidPathError const &er) {
        LOG_RES_WARNING("Failed declaring font \"%s\": %s")
            << NativePath(uri.asText()).pretty() << er.asText();
    }
    return nullptr;
}
AbstractFont *ClientResources::newFontFromFile(de::Uri const &uri, String filePath) {
    LOG_AS("ClientResources::newFontFromFile");
    if (!d->fileSys().accessFile(de::Uri::fromNativePath(filePath))) {
        LOGDEV_RES_WARNING("Ignoring invalid filePath: ") << filePath;
        return nullptr;
    }
    try {
        FontManifest &manifest = declareFont(uri);
        if (manifest.hasResource()) {
            if (auto *bmapFont = maybeAs<BitmapFont>(manifest.resource())) {
                LOGDEV_RES_XVERBOSE("Font with uri \"%s\" already exists, returning existing",
                                    manifest.composeUri());
                bmapFont->setFilePath(filePath);
            }
            return &manifest.resource();
        }
        manifest.setResource(BitmapFont::fromFile(manifest, filePath));
        if (manifest.hasResource()) {
            if (verbose >= 1) {
                LOG_RES_VERBOSE("New font \"%s\"")
                    << manifest.composeUri();
            }
            return &manifest.resource();
        }
        LOG_RES_WARNING("Failed defining new Font for \"%s\"")
            << NativePath(uri.asText()).pretty();
    }
    catch (UnknownSchemeError const &er) {
        LOG_RES_WARNING("Failed declaring font \"%s\": %s")
            << NativePath(uri.asText()).pretty() << er.asText();
    }
    catch (FontScheme::InvalidPathError const &er) {
        LOG_RES_WARNING("Failed declaring font \"%s\": %s")
            << NativePath(uri.asText()).pretty() << er.asText();
    }
    return nullptr;
}
void ClientResources::releaseFontGLTexturesByScheme(String schemeName) {
    if (schemeName.isEmpty()) return;
    PathTreeIterator<FontScheme::Index> iter(fontScheme(schemeName).index().leafNodes());
    while (iter.hasNext()) {
        FontManifest &manifest = iter.next();
        if (manifest.hasResource()) {
            manifest.resource().glDeinit();
        }
    }
}
FrameModel &ClientResources::model(modelid_t id) {
    if (FrameModel *model = d->modelForId(id)) return *model;
    throw MissingResourceError("ClientResources::model", "Invalid id " + String::number(id));
}
bool ClientResources::hasModelDef(String id) const {
    if (!id.isEmpty()) {
        for (FrameModelDef const &modef : d->modefs) {
            if (!id.compareWithoutCase(modef.id)) {
                return true;
            }
        }
    }
    return false;
}
FrameModelDef &ClientResources::modelDef(dint index) {
    if (index >= 0 && index < modelDefCount()) return d->modefs[index];
    throw MissingModelDefError("ClientResources::modelDef", "Invalid index #" + String::number(index) + ", valid range " + Rangeui(0, modelDefCount()).asText());
}
FrameModelDef &ClientResources::modelDef(String id) {
    if (!id.isEmpty()) {
        for (FrameModelDef const &modef : d->modefs) {
            if (!id.compareWithoutCase(modef.id)) {
                return const_cast<FrameModelDef &>(modef);
            }
        }
    }
    throw MissingModelDefError("ClientResources::modelDef", "Invalid id '" + id + "'");
}
FrameModelDef *ClientResources::modelDefForState(dint stateIndex, dint select) {
    if (stateIndex < 0 || stateIndex >= DED_Definitions()->states.size())
        return nullptr;
    if (stateIndex < 0 || stateIndex >= d->stateModefs.count())
        return nullptr;
    if (d->stateModefs[stateIndex] < 0)
        return nullptr;
    DENG2_ASSERT(d->stateModefs[stateIndex] >= 0);
    DENG2_ASSERT(d->stateModefs[stateIndex] < d->modefs.count());
    FrameModelDef *def = &d->modefs[d->stateModefs[stateIndex]];
    if (select) {
        dint const mosel = (select & DDMOBJ_SELECTOR_MASK);
        for (FrameModelDef *it = def; it; it = it->selectNext) {
            if (it->select == mosel) {
                return it;
            }
        }
    }
    return def;
}
dint ClientResources::modelDefCount() const {
    return d->modefs.count();
}
void ClientResources::initModels() {
    LOG_AS("ResourceSystem");
    if (CommandLine_Check("-nomd2")) {
        LOG_RES_NOTE("3D models are disabled");
        return;
    }
    LOG_RES_VERBOSE("Initializing Models...");
    Time begunAt;
    d->clearModelList();
    d->modefs.clear();
    delete d->modelRepository;
    d->modelRepository = new StringPool();
    auto &defs = *DED_Definitions();
    d->modefs.resize(defs.models.size());
    d->stateModefs.resize(runtimeDefs.states.size());
    for (dint i = 0; i < runtimeDefs.states.size(); ++i) {
        d->stateModefs[i] = -1;
    }
    for (dint i = dint(defs.models.size()) - 1; i >= 0; --i) {
        if (!(i % 100)) {
            Con_SetProgress(130 + 70*(defs.models.size() - i)/defs.models.size());
        }
        d->setupModel(defs.models[i]);
    }
    for (dint i = d->modefs.count() - 1; i >= 0; --i) {
        FrameModelDef *me = &d->modefs[i];
        dfloat minmark = 2; // max = 1, so this is "out of bounds".
        FrameModelDef *closest = 0;
        for (dint k = d->modefs.count() - 1; k >= 0; --k) {
            FrameModelDef *other = &d->modefs[k];
            if (other->state != me->state) continue;
            if (other->def.order() > me->def.order() && // Defined after me.
                other->interMark > me->interMark &&
                other->interMark < minmark &&
                other->select == me->select) {
                minmark = other->interMark;
                closest = other;
            }
        }
        me->interNext = closest;
    }
    for (dint i = d->modefs.count() - 1; i >= 0; --i) {
        FrameModelDef *me = &d->modefs[i];
        dint minsel = DDMAXINT;
        FrameModelDef *closest = 0;
        for (dint k = d->modefs.count() - 1; k >= 0; --k) {
            FrameModelDef *other = &d->modefs[k];
            if (other->state == me->state &&
                other->def.order() > me->def.order() && // Defined after me.
                other->select > me->select && other->select < minsel &&
                other->interMark >= me->interMark) {
                minsel = other->select;
                closest = other;
            }
        }
        me->selectNext = closest;
    }
    LOG_RES_MSG("Model init completed in %.2f seconds") << begunAt.since();
}
dint ClientResources::indexOf(FrameModelDef const *modelDef) {
    dint index = dint(modelDef - &d->modefs[0]);
    return (index >= 0 && index < d->modefs.count() ? index : -1);
}
void ClientResources::setModelDefFrame(FrameModelDef &modef, dint frame) {
    for (duint i = 0; i < modef.subCount(); ++i) {
        SubmodelDef &subdef = modef.subModelDef(i);
        if (subdef.modelId == NOMODELID) continue;
        subdef.frame = frame % model(subdef.modelId).frameCount();
    }
}
void ClientResources::purgeCacheQueue() {
    qDeleteAll(d->cacheQueue);
    d->cacheQueue.clear();
}
void ClientResources::processCacheQueue() {
    d->processCacheQueue();
}
void ClientResources::cache(ClientMaterial &material, MaterialVariantSpec const &spec,
                            bool cacheGroups) {
    d->queueCacheTasksForMaterial(material, spec, cacheGroups);
}
void ClientResources::cache(spritenum_t spriteId, MaterialVariantSpec const &spec) {
    d->queueCacheTasksForSprite(spriteId, spec);
}
void ClientResources::cache(FrameModelDef *modelDef) {
    if (!modelDef) return;
    d->queueCacheTasksForModel(*modelDef);
}
MaterialVariantSpec const &ClientResources::materialSpec(MaterialContextId contextId,
    dint flags, byte border, dint tClass, dint tMap, dint wrapS, dint wrapT,
    dint minFilter, dint magFilter, dint anisoFilter,
    bool mipmapped, bool gammaCorrection, bool noStretch, bool toAlpha) {
    return d->getMaterialSpecForContext(contextId, flags, border, tClass, tMap,
                                        wrapS, wrapT, minFilter, magFilter, anisoFilter,
                                        mipmapped, gammaCorrection, noStretch, toAlpha);
}
void ClientResources::cacheForCurrentMap() {
    if (playback) return;
    world::Map &map = App_World().map();
    if (precacheMapMaterials) {
        MaterialVariantSpec const &spec = Rend_MapSurfaceMaterialSpec();
        map.forAllLines([this, &spec] (Line &line) {
            for (dint i = 0; i < 2; ++i) {
                LineSide &side = line.side(i);
                if (!side.hasSections()) continue;
                if (side.middle().hasMaterial())
                    cache(side.middle().material().as<ClientMaterial>(), spec);
                if (side.top().hasMaterial())
                    cache(side.top().material().as<ClientMaterial>(), spec);
                if (side.bottom().hasMaterial())
                    cache(side.bottom().material().as<ClientMaterial>(), spec);
            }
            return LoopContinue;
        });
        map.forAllSectors([this, &spec] (Sector &sector) {
            if (sector.sideCount()) {
                sector.forAllPlanes([this, &spec] (Plane &plane) {
                    if (plane.surface().hasMaterial()) {
                        cache(plane.surface().material().as<ClientMaterial>(), spec);
                    }
                    return LoopContinue;
                });
            }
            return LoopContinue;
        });
    }
    if (precacheSprites) {
        MaterialVariantSpec const &matSpec = Rend_SpriteMaterialSpec();
        for (dint i = 0; i < sprites().spriteCount(); ++i) {
            auto const sprite = spritenum_t(i);
            LoopResult found = map.thinkers().forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker),
                                                     0x1/*public*/, [&sprite] (thinker_t *th)
            {
                auto const &mob = *reinterpret_cast<mobj_t *>(th);
                if (mob.type >= 0 && mob.type < runtimeDefs.mobjInfo.size()) {
                    for (dint k = 0; k < DED_Definitions()->states.size(); ++k) {
                        if (runtimeDefs.stateInfo[k].owner != &runtimeDefs.mobjInfo[mob.type])
                            continue;
                        if (Def_GetState(k)->sprite == sprite) {
                            return LoopAbort;  // Found one.
                        }
                    }
                }
                return LoopContinue;
            });
            if (found) {
                cache(sprite, matSpec);
            }
        }
    }
    if (useModels && precacheSkins) {
        map.thinkers().forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker),
                              0x1/*public*/, [this] (thinker_t *th)
        {
            auto const &mob = *reinterpret_cast<mobj_t *>(th);
            for (dint i = 0; i < modelDefCount(); ++i) {
                FrameModelDef &modef = modelDef(i);
                if (!modef.state) continue;
                if (mob.type < 0 || mob.type >= runtimeDefs.mobjInfo.size()) continue; // Hmm?
                if (runtimeDefs.stateInfo[runtimeDefs.states.indexOf(modef.state)].owner != &runtimeDefs.mobjInfo[mob.type]) continue;
                cache(&modef);
            }
            return LoopContinue;
        });
    }
}
/**
 * @param scheme    Resource subspace scheme being printed. Can be @c NULL in
 *                  which case resources are printed from all schemes.
 * @param like      Resource path search term.
 * @param composeUriFlags  Flags governing how URIs should be composed.
 */
static int printFontIndex2(FontScheme *scheme, Path const &like,
    de::Uri::ComposeAsTextFlags composeUriFlags) {
    FontScheme::Index::FoundNodes found;
    if (scheme) // Only resources in this scheme.
    {
        scheme->index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
    }
    else // Consider resources in any scheme.
    {
        foreach (FontScheme *scheme, App_Resources().allFontSchemes()) {
            scheme->index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
        }
    }
    if (found.isEmpty()) return 0;
    bool const printSchemeName = !(composeUriFlags & de::Uri::OmitScheme);
    String heading = "Known fonts";
    if (!printSchemeName && scheme)
        heading += " in scheme '" + scheme->name() + "'";
    if (!like.isEmpty())
        heading += " like \"" _E(b) + like.toStringRef() + _E(.) "\"";
    LOG_RES_MSG(_E(D) "%s:" _E(.)) << heading;
    qSort(found.begin(), found.end(), comparePathTreeNodePathsAscending<FontManifest>);
    int numFoundDigits = de::max(3/*idx*/, M_NumDigits(found.count()));
    int idx = 0;
    foreach (FontManifest *manifest, found) {
        String info = String("%1: %2%3" _E(.))
                        .arg(idx, numFoundDigits)
                        .arg(manifest->hasResource()? _E(1) : _E(2))
                        .arg(manifest->description(composeUriFlags));
        LOG_RES_MSG("  " _E(>)) << info;
        idx++;
    }
    return found.count();
}
static void printFontIndex(de::Uri const &search,
    de::Uri::ComposeAsTextFlags flags = de::Uri::DefaultComposeAsTextFlags) {
    int printTotal = 0;
    if (search.scheme().isEmpty() && !search.path().isEmpty()) {
        printTotal = printFontIndex2(0/*any scheme*/, search.path(), flags & ~de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else if (App_Resources().knownFontScheme(search.scheme())) {
        printTotal = printFontIndex2(&App_Resources().fontScheme(search.scheme()),
                                     search.path(), flags | de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else {
        foreach (FontScheme *scheme, App_Resources().allFontSchemes()) {
            int numPrinted = printFontIndex2(scheme, search.path(), flags | de::Uri::OmitScheme);
            if (numPrinted) {
                LOG_MSG(_E(R));
                printTotal += numPrinted;
            }
        }
    }
    LOG_RES_MSG("Found " _E(b) "%i" _E(.) " %s.") << printTotal << (printTotal == 1? "font" : "fonts in total");
}
static bool isKnownFontSchemeCallback(String name) {
    return App_Resources().knownFontScheme(name);
}
D_CMD(ListFonts) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1, &isKnownFontSchemeCallback);
    if (!search.scheme().isEmpty() &&
       !App_Resources().knownFontScheme(search.scheme())) {
        LOG_RES_WARNING("Unknown scheme %s") << search.scheme();
        return false;
    }
    printFontIndex(search);
    return true;
}
#ifdef DENG_DEBUG
D_CMD(PrintFontStats) {
    DENG2_UNUSED3(src, argc, argv);
    LOG_MSG(_E(b) "Font Statistics:");
    foreach (FontScheme *scheme, App_Resources().allFontSchemes()) {
        FontScheme::Index const &index = scheme->index();
        uint const count = index.count();
        LOG_MSG("Scheme: %s (%u %s)")
            << scheme->name() << count << (count == 1? "font" : "fonts");
        index.debugPrintHashDistribution();
        index.debugPrint();
    }
    return true;
}
#endif // DENG_DEBUG
void ClientResources::consoleRegister() // static
{
    Resources::consoleRegister();
    C_CMD("listfonts",      "ss",   ListFonts)
    C_CMD("listfonts",      "s",    ListFonts)
    C_CMD("listfonts",      "",     ListFonts)
#ifdef DENG_DEBUG
    C_CMD("fontstats",      NULL,   PrintFontStats)
#endif
}
DENG2_PIMPL(ClientTexture) {
    Variants variants;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        clearVariants();
    }
    void clearVariants() {
        while (!variants.isEmpty()) {
            ClientTexture::Variant *variant = variants.takeFirst();
#ifdef DENG_DEBUG
            if (variant->glName()) {
                String textualVariantSpec = variant->spec().asText();
                LOG_AS("ClientTexture::clearVariants")
                LOGDEV_RES_WARNING("GLName (%i) still set for a variant of \"%s\" %p. Perhaps it wasn't released?")
                    << variant->glName()
                    << self().manifest().composeUri()
                    << this
                    << textualVariantSpec;
            }
#endif
            delete variant;
        }
    }
};
ClientTexture::ClientTexture(res::TextureManifest &manifest)
    : res::Texture(manifest)
    , d(new Impl(this)) {
    setFlags(manifest.flags());
    setDimensions(manifest.logicalDimensions());
    setOrigin(manifest.origin());
}
duint ClientTexture::variantCount() const {
    return duint(d->variants.count());
}
ClientTexture::Variant *ClientTexture::chooseVariant(ChooseVariantMethod method,
                                                     TextureVariantSpec const &spec,
                                                     bool canCreate) {
    foreach (Variant *variant, d->variants) {
        TextureVariantSpec const &cand = variant->spec();
        switch (method) {
        case MatchSpec:
            if (cand == spec) {
                return variant;
            }
            break;
        /*case FuzzyMatchSpec:
            if (cand == spec) {
                return variant;
            }
            break;*/
        }
    }
    /*
#ifdef DENG_DEBUG
    if (method == MatchSpec) {
        DENG_ASSERT(!chooseVariant(FuzzyMatchSpec, spec));
    }
#endif
     */
    if (!canCreate) return 0;
    d->variants.push_back(new Variant(*this, spec));
    return d->variants.back();
}
ClientTexture::Variant *ClientTexture::prepareVariant(TextureVariantSpec const &spec) {
    Variant *variant = chooseVariant(MatchSpec, spec, true /*can create*/);
    DENG2_ASSERT(variant);
    variant->prepare();
    return variant;
}
ClientTexture::Variants const &ClientTexture::variants() const {
    return d->variants;
}
void ClientTexture::clearVariants() {
    d->clearVariants();
}
void ClientTexture::release(/*TextureVariantSpec *spec*/)
{
    Texture::release();
    foreach (TextureVariant *variant, d->variants) { {
            variant->release();
            /*if (spec)
            {
                break;
            }*/
        }
    }
}
String ClientTexture::description() const {
    String variantDesc;
    QTextStream os(&variantDesc);
    if (variantCount()) {
        os << "\n" << _E(R);
        for (int variantIdx = 0; variantIdx < d->variants.size(); ++variantIdx) {
            auto const *variant = d->variants.at(variantIdx);
            Vector2f coords;
            variant->glCoords(&coords.x, &coords.y);
            String textualVariantSpec = variant->spec().asText();
            os << "\n"
               << _E(D) "Variant #" << variantIdx << ":" _E(.)
               << " " _E(l) "Source: " _E(.) _E(i) << variant->sourceDescription() << _E(.)
               << " " _E(l) "Masked: " _E(.) _E(i) << (variant->isMasked()? "yes":"no") << _E(.)
               << " " _E(l) "GLName: " _E(.) _E(i) << variant->glName() << _E(.)
               << " " _E(l) "Coords: " _E(.) _E(i) << coords.asText() << _E(.)
               << _E(R)
               << "\n" _E(b) "Specification:" _E(.)
               << textualVariantSpec;
        }
    }
    return res::Texture::description() +
           String(" x%1").arg(variantCount()) +
           variantDesc;
}
DENG2_PIMPL(CompositeBitmapFont) {
    ded_compositefont_t *def; /// Definition on which "this" font is derived (if any).
    bool needGLInit;
    int leading;
    int ascent;
    int descent;
    Glyph glyphs[MAX_CHARS];
    Glyph missingGlyph;
    Impl(Public *i)
        : Base(i)
        , def(0)
        , needGLInit(true)
        , leading(0)
        , ascent(0)
        , descent(0) {
        zap(glyphs);
        zap(missingGlyph);
        self()._flags |= AbstractFont::Colorize;
    }
    ~Impl() {
        self().glDeinit();
    }
    Glyph &glyph(uchar ch) {
        if(!glyphs[ch].haveSourceImage) return missingGlyph;
        return glyphs[ch];
    }
};
CompositeBitmapFont::CompositeBitmapFont(FontManifest &manifest)
    : AbstractFont(manifest), d(new Impl(this))
{}
int CompositeBitmapFont::ascent() const {
    glInit();
    return d->ascent;
}
int CompositeBitmapFont::descent() const {
    glInit();
    return d->descent;
}
int CompositeBitmapFont::lineSpacing() const {
    glInit();
    return d->leading;
}
Rectanglei const &CompositeBitmapFont::glyphPosCoords(uchar ch) const {
    glInit();
    return d->glyph(ch).geometry;
}
Rectanglei const &CompositeBitmapFont::glyphTexCoords(uchar /*ch*/) const
{
    static Rectanglei coords(Vector2i(0, 0), Vector2i(1, 1));
    glInit();
    return coords;
}
uint CompositeBitmapFont::glyphTextureBorder(uchar ch) const {
    glInit();
    return d->glyph(ch).border;
}
TextureVariant *CompositeBitmapFont::glyphTexture(uchar ch) const {
    glInit();
    return d->glyph(ch).tex;
}
patchid_t CompositeBitmapFont::glyphPatch(uchar ch) const {
    glInit();
    return d->glyph(ch).patch;
}
void CompositeBitmapFont::glyphSetPatch(uchar ch, String encodedPatchName) {
    d->glyphs[ch].patch = res::Textures::get().declarePatch(encodedPatchName);
    d->needGLInit = true;
}
static TextureVariantSpec const &glyphTextureSpec() {
    return ClientApp::resources().textureSpec(TC_UI,
        TSF_MONOCHROME | TSF_UPSCALE_AND_SHARPEN, 0, 0, 0, GL_CLAMP_TO_EDGE,
        GL_CLAMP_TO_EDGE, 0, -3, 0, false, false, false, false);
}
void CompositeBitmapFont::glInit() const {
    if(!d->needGLInit) return;
    if(novideo || BusyMode_Active()) return;
    LOG_AS("CompositeBitmapFont");
    glDeinit();
    int foundGlyphs = 0;
    Vector2ui avgSize;
    for(int i = 0; i < MAX_CHARS; ++i) {
        Glyph *ch = &d->glyphs[i];
        patchid_t patch = ch->patch;
        ch->haveSourceImage = patch != 0;
        if(!ch->haveSourceImage) continue;
        try {
            res::Texture &tex = res::Textures::get().textureScheme("Patches").findByUniqueId(patch).texture();
            ch->tex      = static_cast<ClientTexture &>(tex).prepareVariant(glyphTextureSpec());
            ch->geometry = Rectanglei::fromSize(tex.origin(), tex.dimensions().toVector2ui());
            ch->border   = 0;
            if(ch->tex && ch->tex->source() == res::Original) {
                ch->border = 1;
            }
            avgSize += ch->geometry.size();
            ++foundGlyphs;
        }
        catch(res::TextureManifest::MissingTextureError const &er) {
            LOG_RES_WARNING(er.asText() + ", ignoring.");
        }
        catch(res::TextureScheme::NotFoundError const &er) {
            LOG_RES_WARNING(er.asText() + ", ignoring.");
        }
    }
    d->missingGlyph.geometry.setSize(avgSize / (foundGlyphs? foundGlyphs : 1));
    d->needGLInit = false;
}
void CompositeBitmapFont::glDeinit() const {
    if(novideo) return;
    d->needGLInit = true;
    if(BusyMode_Active()) return;
    for(int i = 0; i < 256; ++i) {
        Glyph *ch = &d->glyphs[i];
        if(!ch->tex) continue;
        ch->tex->release();
        ch->tex = 0;
    }
}
CompositeBitmapFont *CompositeBitmapFont::fromDef(FontManifest &manifest,
    ded_compositefont_t const &def) // static
{
    LOG_AS("CompositeBitmapFont::fromDef");
    CompositeBitmapFont *font = new CompositeBitmapFont(manifest);
    font->setDefinition(const_cast<ded_compositefont_t *>(&def));
    for(int i = 0; i < def.charMap.size(); ++i) {
        if(!def.charMap[i].path) continue;
        try {
            String glyphPatchPath = def.charMap[i].path->resolved();
            font->glyphSetPatch(def.charMap[i].ch, glyphPatchPath);
        }
        catch(de::Uri::ResolveError const &er) {
            LOG_RES_WARNING(er.asText());
        }
    }
    font->glInit();
    return font;
}
ded_compositefont_t *CompositeBitmapFont::definition() const {
    return d->def;
}
void CompositeBitmapFont::setDefinition(ded_compositefont_t *newDef) {
    d->def = newDef;
}
void CompositeBitmapFont::rebuildFromDef(ded_compositefont_t const &newDef) {
    LOG_AS("CompositeBitmapFont::rebuildFromDef");
    setDefinition(const_cast<ded_compositefont_t *>(&newDef));
    if(!d->def) return;
    for(int i = 0; i < d->def->charMap.size(); ++i) {
        if(!d->def->charMap[i].path) continue;
        try {
            String glyphPatchPath = d->def->charMap[i].path->resolved();
            glyphSetPatch(d->def->charMap[i].ch, glyphPatchPath);
        }
        catch(de::Uri::ResolveError const& er) {
            LOG_RES_WARNING(er.asText());
        }
    }
}
DENG2_PIMPL(FontManifest),
DENG2_OBSERVES(AbstractFont, Deletion) {
    int uniqueId;
    QScopedPointer<AbstractFont>(resource); ///< Associated resource (if any).
    Impl(Public *i)
        : Base(i)
        , uniqueId(0)
    {}
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE(Deletion, i) i->fontManifestBeingDeleted(self());
    }
    void fontBeingDeleted(AbstractFont const & /*resource*/)
    {
        resource.reset();
    }
};
FontManifest::FontManifest(PathTree::NodeArgs const &args)
    : Node(args), d(new Impl(this))
{}
FontScheme &FontManifest::scheme() const {
    LOG_AS("FontManifest");
    foreach(FontScheme *scheme, App_Resources().allFontSchemes()) {
        if(&scheme->index() == &tree()) return *scheme;
    }
    throw Error("FontManifest::scheme", String("Failed to determine scheme for manifest [%1]").arg(de::dintptr(this)));
}
String const &FontManifest::schemeName() const {
    return scheme().name();
}
String FontManifest::description(de::Uri::ComposeAsTextFlags uriCompositionFlags) const {
    return String("%1").arg(composeUri().compose(uriCompositionFlags | Uri::DecodePath),
                            ( uriCompositionFlags.testFlag(Uri::OmitScheme)? -14 : -22 ) );
}
int FontManifest::uniqueId() const {
    return d->uniqueId;
}
bool FontManifest::setUniqueId(int newUniqueId) {
    LOG_AS("FontManifest");
    if(d->uniqueId == newUniqueId) return false;
    d->uniqueId = newUniqueId;
    DENG2_FOR_AUDIENCE(UniqueIdChange, i) i->fontManifestUniqueIdChanged(*this);
    return true;
}
bool FontManifest::hasResource() const {
    return !d->resource.isNull();
}
AbstractFont &FontManifest::resource() const {
    if(hasResource()) {
        return *d->resource.data();
    }
    throw MissingFontError("FontManifest::resource", "No resource is associated");
}
void FontManifest::setResource(AbstractFont *newResource) {
    LOG_AS("FontManifest");
    if(d->resource.data() != newResource) {
        if(AbstractFont *curFont = d->resource.data()) {
            curFont->audienceForDeletion -= d;
        }
        d->resource.reset(newResource);
        if(AbstractFont *curFont = d->resource.data()) {
            curFont->audienceForDeletion += d;
        }
    }
}
DENG2_PIMPL(FontScheme),
DENG2_OBSERVES(FontManifest, UniqueIdChange),
DENG2_OBSERVES(FontManifest, Deletion) {
    String name; ///< Symbolic.
    Index index; ///< Mappings from paths to manifests.
    QList<Manifest *> uniqueIdLut; ///< Index with uniqueId - uniqueIdBase.
    bool uniqueIdLutDirty;
    int uniqueIdBase;
    Impl(Public *i)
        : Base(i)
        , uniqueIdLutDirty(false)
        , uniqueIdBase(0)
    {}
    ~Impl() {
        self().clear();
        DENG2_ASSERT(index.isEmpty()); // sanity check.
    }
    bool inline uniqueIdInLutRange(int uniqueId) const {
        return uniqueId - uniqueIdBase >= 0 &&
               (uniqueId - uniqueIdBase) < uniqueIdLut.size();
    }
    void findUniqueIdRange(int *minId, int *maxId) {
        if(!minId && !maxId) return;
        if(minId) *minId = DDMAXINT;
        if(maxId) *maxId = DDMININT;
        PathTreeIterator<Index> iter(index.leafNodes());
        while(iter.hasNext()) {
            Manifest &manifest = iter.next();
            int const uniqueId = manifest.uniqueId();
            if(minId && uniqueId < *minId) *minId = uniqueId;
            if(maxId && uniqueId > *maxId) *maxId = uniqueId;
        }
    }
    void deindex(Manifest &manifest) {
        manifest.clearResource();
        unlinkInUniqueIdLut(manifest);
    }
    void unlinkInUniqueIdLut(Manifest const &manifest) {
        DENG2_ASSERT(&manifest.scheme() == thisPublic); // sanity check.
        if(!uniqueIdLutDirty) {
            int uniqueId = manifest.uniqueId();
            DENG2_ASSERT(uniqueIdInLutRange(uniqueId));
            uniqueIdLut[uniqueId - uniqueIdBase] = 0;
        }
    }
    void linkInUniqueIdLut(Manifest &manifest) {
        DENG2_ASSERT(&manifest.scheme() == thisPublic); // sanity check.
        int uniqueId = manifest.uniqueId();
        DENG_ASSERT(uniqueIdInLutRange(uniqueId));
        uniqueIdLut[uniqueId - uniqueIdBase] = &manifest;
    }
    void rebuildUniqueIdLut() {
        if(!uniqueIdLutDirty) return;
        int minId, maxId;
        findUniqueIdRange(&minId, &maxId);
        int lutSize = 0;
        if(minId > maxId) // None found?
        {
            uniqueIdBase = 0;
        }
        else {
            uniqueIdBase = minId;
            lutSize = maxId - minId + 1;
        }
#ifdef DENG2_QT_4_7_OR_NEWER
        uniqueIdLut.reserve(lutSize);
#endif
        int i = 0;
        for(; i < uniqueIdLut.size(); ++i) {
            uniqueIdLut[i] = 0;
        }
        for(; i < lutSize; ++i) {
            uniqueIdLut.push_back(0);
        }
        if(lutSize) {
            PathTreeIterator<Index> iter(index.leafNodes());
            while(iter.hasNext()) {
                linkInUniqueIdLut(iter.next());
            }
        }
        uniqueIdLutDirty = false;
    }
    void fontManifestUniqueIdChanged(Manifest & /*manifest*/)
    {
        uniqueIdLutDirty = true;
    }
    void fontManifestBeingDeleted(Manifest const &manifest) {
        deindex(const_cast<Manifest &>(manifest));
    }
};
FontScheme::FontScheme(String symbolicName) : d(new Impl(this)) {
    d->name = symbolicName;
}
void FontScheme::clear() {
    /*PathTreeIterator<Index> iter(d->index.leafNodes());
    while(iter.hasNext()) {
        d->deindex(iter.next());
    }*/
    d->index.clear();
    d->uniqueIdLutDirty = true;
}
String const &FontScheme::name() const {
    return d->name;
}
FontScheme::Manifest &FontScheme::declare(Path const &path) {
    LOG_AS("FontScheme::declare");
    if(path.isEmpty()) {
        throw InvalidPathError("FontScheme::declare", "Missing/zero-length path was supplied");
    }
    int const sizeBefore = d->index.size();
    Manifest *newManifest = &d->index.insert(path);
    DENG2_ASSERT(newManifest != 0);
    if(d->index.size() != sizeBefore) {
        newManifest->audienceForUniqueIdChange += d;
        newManifest->audienceForDeletion += d;
        DENG2_FOR_AUDIENCE(ManifestDefined, i) i->fontSchemeManifestDefined(*this, *newManifest);
    }
    return *newManifest;
}
bool FontScheme::has(Path const &path) const {
    return d->index.has(path, Index::NoBranch | Index::MatchFull);
}
FontScheme::Manifest const &FontScheme::find(Path const &path) const {
    if(has(path)) {
        return d->index.find(path, Index::NoBranch | Index::MatchFull);
    }
    throw NotFoundError("FontScheme::find", "Failed to locate a manifest matching \"" + path.asText() + "\"");
}
FontScheme::Manifest &FontScheme::find(Path const &path) {
    Manifest const &found = const_cast<FontScheme const *>(this)->find(path);
    return const_cast<Manifest &>(found);
}
FontScheme::Manifest const &FontScheme::findByUniqueId(int uniqueId) const {
    d->rebuildUniqueIdLut();
    if(d->uniqueIdInLutRange(uniqueId)) {
        Manifest *manifest = d->uniqueIdLut[uniqueId - d->uniqueIdBase];
        if(manifest) return *manifest;
    }
    throw NotFoundError("FontScheme::findByUniqueId", "No manifest found with a unique ID matching \"" + QString("%1").arg(uniqueId) + "\"");
}
FontScheme::Manifest &FontScheme::findByUniqueId(int uniqueId) {
    Manifest const &found = const_cast<FontScheme const *>(this)->findByUniqueId(uniqueId);
    return const_cast<Manifest &>(found);
}
FontScheme::Index const &FontScheme::index() const {
    return d->index;
}
bool FrameModel::DetailLevel::hasVertex(int number) const {
    return model.lodVertexUsage().testBit(number * model.lodCount() + level);
}
void FrameModel::Frame::bounds(Vector3f &retMin, Vector3f &retMax) const {
    retMin = min;
    retMax = max;
}
float FrameModel::Frame::horizontalRange(float *top, float *bottom) const {
    *top    = max.y;
    *bottom = min.y;
    return max.y - min.y;
}
#define MD2_MAGIC 0x32504449
#pragma pack(1)
struct md2_header_t {
    int magic;
    int version;
    int skinWidth;
    int skinHeight;
    int frameSize;
    int numSkins;
    int numVertices;
    int numTexCoords;
    int numTriangles;
    int numGlCommands;
    int numFrames;
    int offsetSkins;
    int offsetTexCoords;
    int offsetTriangles;
    int offsetFrames;
    int offsetGlCommands;
    int offsetEnd;
};
#pragma pack()
static bool readMd2Header(FileHandle &file, md2_header_t &hdr) {
    size_t readBytes = file.read((uint8_t *)&hdr, sizeof(md2_header_t));
    if(readBytes < sizeof(md2_header_t)) return false;
    hdr.magic            = littleEndianByteOrder.toHost(hdr.magic);
    hdr.version          = littleEndianByteOrder.toHost(hdr.version);
    hdr.skinWidth        = littleEndianByteOrder.toHost(hdr.skinWidth);
    hdr.skinHeight       = littleEndianByteOrder.toHost(hdr.skinHeight);
    hdr.frameSize        = littleEndianByteOrder.toHost(hdr.frameSize);
    hdr.numSkins         = littleEndianByteOrder.toHost(hdr.numSkins);
    hdr.numVertices      = littleEndianByteOrder.toHost(hdr.numVertices);
    hdr.numTexCoords     = littleEndianByteOrder.toHost(hdr.numTexCoords);
    hdr.numTriangles     = littleEndianByteOrder.toHost(hdr.numTriangles);
    hdr.numGlCommands    = littleEndianByteOrder.toHost(hdr.numGlCommands);
    hdr.numFrames        = littleEndianByteOrder.toHost(hdr.numFrames);
    hdr.offsetSkins      = littleEndianByteOrder.toHost(hdr.offsetSkins);
    hdr.offsetTexCoords  = littleEndianByteOrder.toHost(hdr.offsetTexCoords);
    hdr.offsetTriangles  = littleEndianByteOrder.toHost(hdr.offsetTriangles);
    hdr.offsetFrames     = littleEndianByteOrder.toHost(hdr.offsetFrames);
    hdr.offsetGlCommands = littleEndianByteOrder.toHost(hdr.offsetGlCommands);
    hdr.offsetEnd        = littleEndianByteOrder.toHost(hdr.offsetEnd);
    return true;
}
#define DMD_MAGIC 0x4D444D44 ///< "DMDM" = Doomsday/Detailed MoDel Magic
#pragma pack(1)
struct dmd_header_t {
    int magic;
    int version;
    int flags;
};
#pragma pack()
static bool readHeaderDmd(FileHandle &file, dmd_header_t &hdr) {
    size_t readBytes = file.read((uint8_t *)&hdr, sizeof(dmd_header_t));
    if(readBytes < sizeof(dmd_header_t)) return false;
    hdr.magic   = littleEndianByteOrder.toHost(hdr.magic);
    hdr.version = littleEndianByteOrder.toHost(hdr.version);
    hdr.flags   = littleEndianByteOrder.toHost(hdr.flags);
    return true;
}
static void *allocAndLoad(FileHandle &file, int offset, int len) {
    uint8_t *ptr = (uint8_t *) M_Malloc(len);
    file.seek(offset, SeekSet);
    file.read(ptr, len);
    return ptr;
}
#define NUMVERTEXNORMALS 162
static float avertexnormals[NUMVERTEXNORMALS][3] = {
#include "misc/tab_anorms.h"
};
/**
 * @todo reimplement file loading using de::Reader.
 */
DENG2_PIMPL(FrameModel) {
    Flags flags;
    Skins skins;
    Frames frames;
    int numVertices;
    DetailLevels lods;
    QBitArray lodVertexUsage;
    uint modelId; ///< In the repository.
    Impl(Public *i)
        : Base(i)
        , numVertices(0)
        , modelId(0)
    {}
    ~Impl() {
        self().clearAllFrames();
    }
#pragma pack(1)
    struct md2_triangleVertex_t {
        byte vertex[3];
        byte normalIndex;
    };
    struct md2_packedFrame_t {
        float scale[3];
        float translate[3];
        char name[16];
        md2_triangleVertex_t vertices[1];
    };
    struct md2_commandElement_t {
        float s, t;
        int index;
    };
#pragma pack()
    /**
     * Note vertex Z/Y are swapped here (ordered XYZ in the serialized data).
     */
    static FrameModel *loadMd2(FileHandle &file, float aspectScale) {
        md2_header_t hdr;
        if(!readMd2Header(file, hdr)) return 0;
        if(hdr.magic != MD2_MAGIC) return 0;
        FrameModel *mdl = new FrameModel;
        mdl->d->numVertices = hdr.numVertices;
        uint8_t *frameData = (uint8_t *) allocAndLoad(file, hdr.offsetFrames, hdr.frameSize * hdr.numFrames);
        for(int i = 0; i < hdr.numFrames; ++i) {
            md2_packedFrame_t const *pfr = (md2_packedFrame_t const *) (frameData + hdr.frameSize * i);
            Vector3f const scale(DD_FLOAT(pfr->scale[0]), DD_FLOAT(pfr->scale[2]), DD_FLOAT(pfr->scale[1]));
            Vector3f const translation(DD_FLOAT(pfr->translate[0]), DD_FLOAT(pfr->translate[2]), DD_FLOAT(pfr->translate[1]));
            String const frameName = pfr->name;
            FrameModelFrame *frame = new FrameModelFrame(*mdl, frameName);
            frame->vertices.reserve(hdr.numVertices);
            md2_triangleVertex_t const *pVtx = pfr->vertices;
            for(int k = 0; k < hdr.numVertices; ++k, pVtx++) {
                frame->vertices.append(FrameModelFrame::Vertex());
                FrameModelFrame::Vertex &vtx = frame->vertices.last();
                vtx.pos = Vector3f(pVtx->vertex[0], pVtx->vertex[2], pVtx->vertex[1])
                              * scale + translation;
                vtx.pos.y *= aspectScale; // Aspect undoing.
                vtx.norm = Vector3f(avertexnormals[pVtx->normalIndex]);
                if(!k) {
                    frame->min = frame->max = vtx.pos;
                }
                else {
                    frame->min = vtx.pos.min(frame->min);
                    frame->max = vtx.pos.max(frame->max);
                }
            }
            mdl->d->frames.append(frame);
        }
        M_Free(frameData);
        mdl->d->lods.append(new FrameModelLOD(*mdl, 0));
        FrameModelLOD &lod0 = *mdl->d->lods.last();
        uint8_t *commandData = (uint8_t *) allocAndLoad(file, hdr.offsetGlCommands, 4 * hdr.numGlCommands);
        for(uint8_t const *pos = commandData; *pos;) {
            int count = DD_LONG( *(int *) pos ); pos += 4;
            lod0.primitives.append(FrameModel::Primitive());
            FrameModel::Primitive &prim = lod0.primitives.last();
            prim.triFan = (count < 0);
            if(count < 0) {
                count = -count;
            }
            while(count--) {
                md2_commandElement_t const *v = (md2_commandElement_t *) pos; pos += 12;
                prim.elements.append(FrameModel::Primitive::Element());
                FrameModel::Primitive::Element &elem = prim.elements.last();
                elem.texCoord = Vector2f(DD_FLOAT(v->s), DD_FLOAT(v->t));
                elem.index    = DD_LONG(v->index);
            }
        }
        M_Free(commandData);
        file.seek(hdr.offsetSkins, SeekSet);
        for(int i = 0; i < hdr.numSkins; ++i) {
            char name[64]; file.read((uint8_t *)name, 64);
            mdl->newSkin(name);
        }
        return mdl;
    }
    enum {
        DMC_END, /// Must be the last chunk.
        DMC_INFO /// Required; will be expected to exist.
    };
#pragma pack(1)
    struct dmd_chunk_t {
        int type;
        int length; /// Next chunk follows...
    };
    struct dmd_info_t {
        int skinWidth;
        int skinHeight;
        int frameSize;
        int numSkins;
        int numVertices;
        int numTexCoords;
        int numFrames;
        int numLODs;
        int offsetSkins;
        int offsetTexCoords;
        int offsetFrames;
        int offsetLODs;
        int offsetEnd;
    };
    struct dmd_levelOfDetail_t {
        int numTriangles;
        int numGlCommands;
        int offsetTriangles;
        int offsetGlCommands;
    };
    struct dmd_packedVertex_t {
        byte vertex[3];
        unsigned short normal; /// Yaw and pitch.
    };
    struct dmd_packedFrame_t {
        float scale[3];
        float translate[3];
        char name[16];
        dmd_packedVertex_t vertices[1]; // dmd_info_t::numVertices size
    };
    struct dmd_triangle_t {
        short vertexIndices[3];
        short textureIndices[3];
    };
#pragma pack()
    /**
     * Packed: pppppppy yyyyyyyy. Yaw is on the XY plane.
     */
    static Vector3f unpackVector(ushort packed) {
        float const yaw   = (packed & 511) / 512.0f * 2 * PI;
        float const pitch = ((packed >> 9) / 127.0f - 0.5f) * PI;
        float const cosp  = float(cos(pitch));
        return Vector3f(cos(yaw) * cosp, sin(yaw) * cosp, sin(pitch));
    }
    /**
     * Note vertex Z/Y are swapped here (ordered XYZ in the serialized data).
     */
    static FrameModel *loadDmd(FileHandle &file, float aspectScale) {
        dmd_header_t hdr;
        if(!readHeaderDmd(file, hdr)) return 0;
        if(hdr.magic != DMD_MAGIC) return 0;
        dmd_chunk_t chunk;
        file.read((uint8_t *)&chunk, sizeof(chunk));
        dmd_info_t info; zap(info);
        while(DD_LONG(chunk.type) != DMC_END) {
            switch(DD_LONG(chunk.type)) {
            case DMC_INFO: // Standard DMD information chunk.
                file.read((uint8_t *)&info, DD_LONG(chunk.length));
                info.skinWidth       = DD_LONG(info.skinWidth);
                info.skinHeight      = DD_LONG(info.skinHeight);
                info.frameSize       = DD_LONG(info.frameSize);
                info.numSkins        = DD_LONG(info.numSkins);
                info.numVertices     = DD_LONG(info.numVertices);
                info.numTexCoords    = DD_LONG(info.numTexCoords);
                info.numFrames       = DD_LONG(info.numFrames);
                info.numLODs         = DD_LONG(info.numLODs);
                info.offsetSkins     = DD_LONG(info.offsetSkins);
                info.offsetTexCoords = DD_LONG(info.offsetTexCoords);
                info.offsetFrames    = DD_LONG(info.offsetFrames);
                info.offsetLODs      = DD_LONG(info.offsetLODs);
                info.offsetEnd       = DD_LONG(info.offsetEnd);
                break;
            default:
                file.seek(DD_LONG(chunk.length), SeekCur);
                break;
            }
            file.read((uint8_t *)&chunk, sizeof(chunk));
        }
        FrameModel *mdl = new FrameModel;
        mdl->d->numVertices = info.numVertices;
        file.seek(info.offsetSkins, SeekSet);
        for(int i = 0; i < info.numSkins; ++i) {
            char name[64]; file.read((uint8_t *)name, 64);
            mdl->newSkin(name);
        }
        uint8_t *frameData = (uint8_t *) allocAndLoad(file, info.offsetFrames, info.frameSize * info.numFrames);
        for(int i = 0; i < info.numFrames; ++i) {
            dmd_packedFrame_t const *pfr = (dmd_packedFrame_t *) (frameData + info.frameSize * i);
            Vector3f const scale(DD_FLOAT(pfr->scale[0]), DD_FLOAT(pfr->scale[2]), DD_FLOAT(pfr->scale[1]));
            Vector3f const translation(DD_FLOAT(pfr->translate[0]), DD_FLOAT(pfr->translate[2]), DD_FLOAT(pfr->translate[1]));
            String const frameName = pfr->name;
            Frame *frame = new Frame(*mdl, frameName);
            frame->vertices.reserve(info.numVertices);
            dmd_packedVertex_t const *pVtx = pfr->vertices;
            for(int k = 0; k < info.numVertices; ++k, ++pVtx) {
                frame->vertices.append(Frame::Vertex());
                Frame::Vertex &vtx = frame->vertices.last();
                vtx.pos = Vector3f(pVtx->vertex[0], pVtx->vertex[2], pVtx->vertex[1])
                              * scale + translation;
                vtx.pos.y *= aspectScale; // Aspect undo.
                vtx.norm = unpackVector(DD_USHORT(pVtx->normal));
                if(!k) {
                    frame->min = frame->max = vtx.pos;
                }
                else {
                    frame->min = vtx.pos.min(frame->min);
                    frame->max = vtx.pos.max(frame->max);
                }
            }
            mdl->d->frames.append(frame);
        }
        M_Free(frameData);
        file.seek(info.offsetLODs, SeekSet);
        dmd_levelOfDetail_t *lodInfo = new dmd_levelOfDetail_t[info.numLODs];
        for(int i = 0; i < info.numLODs; ++i) {
            file.read((uint8_t *)&lodInfo[i], sizeof(dmd_levelOfDetail_t));
            lodInfo[i].numTriangles     = DD_LONG(lodInfo[i].numTriangles);
            lodInfo[i].numGlCommands    = DD_LONG(lodInfo[i].numGlCommands);
            lodInfo[i].offsetTriangles  = DD_LONG(lodInfo[i].offsetTriangles);
            lodInfo[i].offsetGlCommands = DD_LONG(lodInfo[i].offsetGlCommands);
        }
        dmd_triangle_t **triangles = new dmd_triangle_t*[info.numLODs];
        for(int i = 0; i < info.numLODs; ++i) {
            mdl->d->lods.append(new DetailLevel(*mdl, i));
            DetailLevel &lod = *mdl->d->lods.last();
            triangles[i] = (dmd_triangle_t *) allocAndLoad(file, lodInfo[i].offsetTriangles,
                                                           sizeof(dmd_triangle_t) * lodInfo[i].numTriangles);
            uint8_t *commandData = (uint8_t *) allocAndLoad(file, lodInfo[i].offsetGlCommands,
                                                            4 * lodInfo[i].numGlCommands);
            for(uint8_t const *pos = commandData; *pos;) {
                int count = DD_LONG( *(int *) pos ); pos += 4;
                lod.primitives.append(Primitive());
                Primitive &prim = lod.primitives.last();
                prim.triFan = (count < 0);
                if(count < 0) {
                    count = -count;
                }
                while(count--) {
                    md2_commandElement_t const *v = (md2_commandElement_t *) pos; pos += 12;
                    prim.elements.append(Primitive::Element());
                    Primitive::Element &elem = prim.elements.last();
                    elem.texCoord = Vector2f(DD_FLOAT(v->s), DD_FLOAT(v->t));
                    elem.index    = DD_LONG(v->index);
                }
            }
            M_Free(commandData);
        }
        mdl->d->lodVertexUsage.resize(info.numVertices * info.numLODs);
        mdl->d->lodVertexUsage.fill(false);
        for(int i = 0; i < info.numLODs; ++i)
        for(int k = 0; k < lodInfo[i].numTriangles; ++k)
        for(int m = 0; m < 3; ++m) {
            int vertexIndex = DD_SHORT(triangles[i][k].vertexIndices[m]);
            mdl->d->lodVertexUsage.setBit(vertexIndex * info.numLODs + i);
        }
        delete [] lodInfo;
        for(int i = 0; i < info.numLODs; ++i) {
            M_Free(triangles[i]);
        }
        delete [] triangles;
        return mdl;
    }
#if 0
    /**
     * Calculate vertex normals. Only with -renorm.
     */
    static void rebuildNormals(model_t &mdl) {
        if(!CommandLine_Check("-renorm")) return;
        int const tris  = mdl.lodInfo[0].numTriangles;
        int const verts = mdl.info.numVertices;
        vector_t* normals = (vector_t*) Z_Malloc(sizeof(vector_t) * tris, PU_APPSTATIC, 0);
        vector_t norm;
        int cnt;
        for(int i = 0; i < mdl.info.numFrames; ++i) {
            model_vertex_t* list = mdl.frames[i].vertices;
            for(int k = 0; k < tris; ++k) {
                dmd_triangle_t const& tri = mdl.lods[0].triangles[k];
                V3f_PointCrossProduct(normals[k].pos,
                                      list[tri.vertexIndices[0]].vertex,
                                      list[tri.vertexIndices[2]].vertex,
                                      list[tri.vertexIndices[1]].vertex);
                V3f_Normalize(normals[k].pos);
            }
            for(int k = 0; k < verts; ++k) {
                memset(&norm, 0, sizeof(norm));
                cnt = 0;
                for(int j = 0; j < tris; ++j) {
                    dmd_triangle_t const& tri = mdl.lods[0].triangles[j];
                    for(int n = 0; n < 3; ++n) {
                        if(tri.vertexIndices[n] == k) {
                            cnt++;
                            for(int n = 0; n < 3; ++n) {
                                norm.pos[n] += normals[j].pos[n];
                            }
                            break;
                        }
                    }
                }
                if(!cnt) continue; // Impossible...
                for(int n = 0; n < 3; ++n) {
                    norm.pos[n] /= cnt;
                }
                V3f_Normalize(norm.pos);
                memcpy(list[k].normal, norm.pos, sizeof(norm.pos));
            }
        }
        Z_Free(normals);
    }
#endif
};
FrameModel::FrameModel(Flags flags) : d(new Impl(this)) {
    setFlags(flags, de::ReplaceFlags);
}
static bool recogniseDmd(FileHandle &file) {
    dmd_header_t hdr;
    size_t initPos = file.tell();
    file.seek(0, SeekSet);
    bool result = (readHeaderDmd(file, hdr) && DD_LONG(hdr.magic) == DMD_MAGIC);
    file.seek(initPos, SeekSet);
    return result;
}
static bool recogniseMd2(FileHandle &file) {
    md2_header_t hdr;
    size_t initPos = file.tell();
    file.seek(0, SeekSet);
    bool result = (readMd2Header(file, hdr) && DD_LONG(hdr.magic) == MD2_MAGIC);
    file.seek(initPos, SeekSet);
    return result;
}
bool FrameModel::recognise(FileHandle &hndl) //static
{
    LOG_AS("FrameModel");
    if(recogniseDmd(hndl)) return true;
    if(recogniseMd2(hndl)) return true;
    return false;
}
struct ModelFileType {
    String name; ///< Symbolic name of the resource type.
    String ext;  ///< Known file extension.
    FrameModel *(*loadFunc)(FileHandle &hndl, float aspectScale);
};
FrameModel *FrameModel::loadFromFile(FileHandle &hndl, float aspectScale) //static
{
    LOG_AS("FrameModel");
    static ModelFileType const modelTypes[] = {
        { "DMD",    ".dmd",     Impl::loadDmd },
        { "MD2",    ".md2",     Impl::loadMd2 },
    };
    ModelFileType const *rtypeGuess = 0;
    String filePath = hndl.file().composePath();
    String ext      = filePath.fileNameExtension();
    if (!ext.isEmpty()) {
        for (auto const &rtype : modelTypes) {
            if (!rtype.ext.compareWithoutCase(ext)) {
                rtypeGuess = &rtype;
                if (FrameModel *mdl = rtype.loadFunc(hndl, aspectScale)) {
                    LOG_RES_VERBOSE("Interpreted \"" + NativePath(filePath).pretty() + "\" as a " + rtype.name + " model");
                    return mdl;
                }
                break;
            }
        }
    }
    for (auto const &rtype : modelTypes) {
        if (&rtype == rtypeGuess) continue;
        if (FrameModel *mdl = rtype.loadFunc(hndl, aspectScale)) {
            LOG_RES_VERBOSE("Interpreted \"" + NativePath(filePath).pretty() + "\" as a " + rtype.name + " model");
            return mdl;
        }
    }
    return 0;
}
uint FrameModel::modelId() const {
    return d->modelId;
}
void FrameModel::setModelId(uint newModelId) {
    d->modelId = newModelId;
}
FrameModel::Flags FrameModel::flags() const {
    return d->flags;
}
void FrameModel::setFlags(FrameModel::Flags flagsToChange, FlagOp operation) {
    LOG_AS("FrameModel");
    applyFlagOperation(d->flags, flagsToChange, operation);
}
int FrameModel::frameNumber(String name) const {
    if(!name.isEmpty()) {
        for(int i = 0; i < d->frames.count(); ++i) {
            Frame *frame = d->frames.at(i);
            if(!frame->name.compareWithoutCase(name))
                return i;
        }
    }
    return -1; // Not found.
}
FrameModel::Frame &FrameModel::frame(int number) const {
    LOG_AS("FrameModel");
    if(hasFrame(number)) {
        return *d->frames.at(number);
    }
    throw MissingFrameError("FrameModel::frame", "Invalid frame number " + String::number(number) + ", valid range is " + Rangei(0, d->frames.count()).asText());
}
FrameModel::Frames const &FrameModel::frames() const {
    return d->frames;
}
void FrameModel::clearAllFrames() {
    LOG_AS("FrameModel");
    qDeleteAll(d->frames);
    d->frames.clear();
}
int FrameModel::skinNumber(String name) const {
    if(!name.isEmpty()) {
        for(int i = d->skins.count(); i--> 0; ) {
            Skin const &skin = d->skins.at(i);
            if(!skin.name.compareWithoutCase(name))
                return i;
        }
    }
    return -1; // Not found.
}
FrameModel::Skin &FrameModel::skin(int number) const {
    LOG_AS("FrameModel");
    if(hasSkin(number)) {
        return const_cast<Skin &>(d->skins.at(number));
    }
    throw MissingSkinError("FrameModel::skin", "Invalid skin number " + String::number(number) + ", valid range is " + Rangei(0, d->skins.count()).asText());
}
FrameModel::Skin &FrameModel::newSkin(String name) {
    LOG_AS("FrameModel");
    d->skins.append(FrameModelSkin(name));
    return d->skins.last();
}
FrameModel::Skins const &FrameModel::skins() const {
    return d->skins;
}
void FrameModel::clearAllSkins() {
    LOG_AS("FrameModel");
    d->skins.clear();
}
FrameModel::DetailLevel &FrameModel::lod(int level) const {
    LOG_AS("FrameModel");
    if(hasLod(level)) {
        return *d->lods.at(level);
    }
    throw MissingDetailLevelError("FrameModel::lod", "Invalid detail level " + String::number(level) + ", valid range is " + Rangei(0, d->lods.count()).asText());
}
FrameModel::DetailLevels const &FrameModel::lods() const {
    return d->lods;
}
FrameModel::Primitives const &FrameModel::primitives() const {
    LOG_AS("FrameModel");
    return lod(0).primitives;
}
int FrameModel::vertexCount() const {
    return d->numVertices;
}
QBitArray const &FrameModel::lodVertexUsage() const {
    return d->lodVertexUsage;
}
static uint32_t lutBGR888toYUV888[32*64*32];
static uint32_t YUV1, YUV2;
void LerpColor(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3, uint32_t f1,
    uint32_t f2, uint32_t f3) {
    uint32_t out[4] = { 0, 0, 0, 0 }, total = f1 + f2 + f3;
    if(0 != f1) {
        out[0] += f1 * ABGR8888_COMP(0, c1);
        out[1] += f1 * ABGR8888_COMP(1, c1);
        out[2] += f1 * ABGR8888_COMP(2, c1);
        out[3] += f1 * ABGR8888_COMP(3, c1);
    }
    if(0 != f2) {
        out[0] += f2 * ABGR8888_COMP(0, c2);
        out[1] += f2 * ABGR8888_COMP(1, c2);
        out[2] += f2 * ABGR8888_COMP(2, c2);
        out[3] += f2 * ABGR8888_COMP(3, c2);
    }
    if(0 != f3) {
        out[0] += f3 * ABGR8888_COMP(0, c3);
        out[1] += f3 * ABGR8888_COMP(1, c3);
        out[2] += f3 * ABGR8888_COMP(2, c3);
        out[3] += f3 * ABGR8888_COMP(3, c3);
    }
    if(0 != total) {
        out[0] /= total;
        out[1] /= total;
        out[2] /= total;
        out[3] /= total;
    }
    *((uint32_t*)pc) = ABGR8888_PACK(out[3], out[2], out[1], out[0]);
}
static __inline int Diff(uint32_t c1, uint32_t c2) {
    YUV1 = ABGR8888toYUV888(c1);
    YUV2 = ABGR8888toYUV888(c2);
    return ( ((ABGR8888_COMP(3, c1) != 0) != ((ABGR8888_COMP(3, c2) != 0))) ||
             (abs(int(YUV1 & YUV888_Ymask) - int(YUV2 & YUV888_Ymask)) > ((trY & (int)0xFF) << 16)) ||
             (abs(int(YUV1 & YUV888_Umask) - int(YUV2 & YUV888_Umask)) > ((trU & (int)0xFF) << 8)) ||
             (abs(int(YUV1 & YUV888_Vmask) - int(YUV2 & YUV888_Vmask)) > ((trV & (int)0xFF)) ));
}
static __inline void Transl(uint8_t* pc, uint32_t c) {
    pc[0] = ABGR8888_COMP(0, c);
    pc[1] = ABGR8888_COMP(1, c);
    pc[2] = ABGR8888_COMP(2, c);
    pc[3] = ABGR8888_COMP(3, c);
}
static __inline void Interp1(uint8_t* pc, uint32_t c1, uint32_t c2) {
    if(c1 == c2) {
        Transl(pc, c1);
        return;
    }
    LerpColor(pc, c1, c2, 0, 3, 1, 0);
}
static __inline void Interp2(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3) {
    LerpColor(pc, c1, c2, c3, 2, 1, 1);
}
static __inline void Interp6(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3) {
    LerpColor(pc, c1, c2, c3, 5, 2, 1);
}
static __inline void Interp7(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3) {
    LerpColor(pc, c1, c2, c3, 6, 1, 1);
}
static __inline void Interp9(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3) {
    LerpColor(pc, c1, c2, c3, 2, 3, 3);
}
static __inline void Interp10(uint8_t* pc, uint32_t c1, uint32_t c2, uint32_t c3) {
    LerpColor(pc, c1, c2, c3, 14, 1, 1);
}
void GL_InitSmartFilterHQ2x(void) {
    uint32_t r, g, b, y, u, v;
    int i, j, k;
    for(i = 0; i < 32; ++i)
        for(j = 0; j < 64; ++j)
            for(k = 0; k < 32; ++k) {
                r = i << 3;
                g = j << 2;
                b = k << 3;
                y = (uint32_t) de::clamp(0.0, ( 0.299*r + 0.587*g + 0.114*b),       255.0);
                u = (uint32_t) de::clamp(0.0, (-0.169*r - 0.331*g + 0.5  *b) + 128, 255.0);
                v = (uint32_t) de::clamp(0.0, ( 0.5  *r - 0.419*g - 0.081*b) + 128, 255.0);
                lutBGR888toYUV888[BGR565_PACK(k, j, i)] = YUV888_PACK(y, u, v);
            }
}
uint8_t* GL_SmartFilterHQ2x(const uint8_t* src, int width, int height, int flags) {
#define BPP             (4) // Bytes Per Pixel.
#define OFFSET(x, y)    (BPP*(y)*width + BPP*(x))
    assert(src); {
    dd_bool wrapH = (flags & ICF_UPSCALE_SAMPLE_WRAPH) != 0;
    dd_bool wrapV = (flags & ICF_UPSCALE_SAMPLE_WRAPV) != 0;
    int pattern, flag, BpL, xA, xB, yA, yB;
    uint8_t* pOut, *dst;
    uint32_t w[10];
    if(width <= 0 || height <= 0)
        return 0;
    if(0 == (dst = (uint8_t *) M_Malloc(BPP * 2 * width * height * 2)))
        App_Error("GL_SmartFilterHQ2x: Failed on allocation of %lu bytes for "
                  "output buffer.", (unsigned long) (BPP * 2 * width * height * 2));
    pOut = dst;
    BpL = BPP * 2 * width; // (Out) Bytes per Line.
    { int y;
    for(y = 0; y < height; ++y) {
        { int x;
        for(x = 0; x < width; ++x) {
            w[5] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x, y)) ) );
            if(wrapH) {
                w[4] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x == 0? width-1 : x-1, y)) ) );
                w[6] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x == width-1? 0 : x+1, y)) ) );
            }
            else {
                if(x != 0)
                    w[4] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x-1, y)) ) );
                else
                    w[4] = w[5];
                if(x != width-1)
                    w[6] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x+1, y)) ) );
                else
                    w[6] = w[5];
            }
            if(wrapV) {
                w[2] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x, y == 0? height-1 : y-1)) ) );
                w[8] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x, y == height-1? 0 : y+1)) ) );
            }
            else {
                if(y != 0)
                    w[2] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x, y-1)) ) );
                else
                    w[2] = w[5];
                if(y != height-1)
                    w[8] = DD_ULONG( *( (uint32_t*)(src + OFFSET(x, y+1)) ) );
                else
                    w[8] = w[5];
            }
            xA =        x == 0? ( wrapH?  width-1 : 0) : x-1;
            xB =  x == width-1? (!wrapH?  width-1 : 0) : x+1;
            yA =        y == 0? ( wrapV? height-1 : 0) : y-1;
            yB = y == height-1? (!wrapV? height-1 : 0) : y+1;
            w[1] = DD_ULONG( *( (uint32_t*)(src + OFFSET(xA, yA)) ) );
            w[7] = DD_ULONG( *( (uint32_t*)(src + OFFSET(xA, yB)) ) );
            w[3] = DD_ULONG( *( (uint32_t*)(src + OFFSET(xB, yA)) ) );
            w[9] = DD_ULONG( *( (uint32_t*)(src + OFFSET(xB, yB)) ) );
            pattern = 0;
            flag = 1;
            YUV1 = ABGR8888toYUV888(w[5]);
            { int k;
            for(k = 1; k <= 9; ++k) {
                if(k == 5)
                    continue;
                if(w[k] != w[5]) {
                    YUV2 = ABGR8888toYUV888(w[k]);
                    if(((ABGR8888_COMP(3, w[5]) != 0) != (ABGR8888_COMP(3, w[k]) != 0)) ||
                       (abs(int(YUV1 & YUV888_Ymask) - int(YUV2 & YUV888_Ymask)) > ((trY & (int)0xFF) << 16)) ||
                       (abs(int(YUV1 & YUV888_Umask) - int(YUV2 & YUV888_Umask)) > ((trU & (int)0xFF) << 8)) ||
                       (abs(int(YUV1 & YUV888_Vmask) - int(YUV2 & YUV888_Vmask)) > ((trV & (int)0xFF) )) )
                        pattern |= flag;
                }
                flag <<= 1;
            }}
            switch(pattern) {
            case 0:
            case 1:
            case 4:
            case 32:
            case 128:
            case 5:
            case 132:
            case 160:
            case 33:
            case 129:
            case 36:
            case 133:
            case 164:
            case 161:
            case 37:
            case 165: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_20 PIXEL11_20 break;
              }
            case 2:
            case 34:
            case 130:
            case 162: {
                    PIXEL00_22 PIXEL01_21 PIXEL10_20 PIXEL11_20 break;
              }
            case 16:
            case 17:
            case 48:
            case 49: {
                    PIXEL00_20 PIXEL01_22 PIXEL10_20 PIXEL11_21 break;
              }
            case 64:
            case 65:
            case 68:
            case 69: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_21 PIXEL11_22 break;
              }
            case 8:
            case 12:
            case 136:
            case 140: {
                    PIXEL00_21 PIXEL01_20 PIXEL10_22 PIXEL11_20 break;
              }
            case 3:
            case 35:
            case 131:
            case 163: {
                    PIXEL00_11 PIXEL01_21 PIXEL10_20 PIXEL11_20 break;
              }
            case 6:
            case 38:
            case 134:
            case 166: {
                    PIXEL00_22 PIXEL01_12 PIXEL10_20 PIXEL11_20 break;
              }
            case 20:
            case 21:
            case 52:
            case 53: {
                    PIXEL00_20 PIXEL01_11 PIXEL10_20 PIXEL11_21 break;
              }
            case 144:
            case 145:
            case 176:
            case 177: {
                    PIXEL00_20 PIXEL01_22 PIXEL10_20 PIXEL11_12 break;
              }
            case 192:
            case 193:
            case 196:
            case 197: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_21 PIXEL11_11 break;
              }
            case 96:
            case 97:
            case 100:
            case 101: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_12 PIXEL11_22 break;
              }
            case 40:
            case 44:
            case 168:
            case 172: {
                    PIXEL00_21 PIXEL01_20 PIXEL10_11 PIXEL11_20 break;
              }
            case 9:
            case 13:
            case 137:
            case 141: {
                    PIXEL00_12 PIXEL01_20 PIXEL10_22 PIXEL11_20 break;
              }
            case 18:
            case 50: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_20}
                    PIXEL10_20 PIXEL11_21 break;
              }
            case 80:
            case 81: {
                    PIXEL00_20 PIXEL01_22 PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 72:
            case 76: {
                    PIXEL00_21 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 10:
            case 138: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 PIXEL10_22 PIXEL11_20 break;
              }
            case 66: {
                    PIXEL00_22 PIXEL01_21 PIXEL10_21 PIXEL11_22 break;
              }
            case 24: {
                    PIXEL00_21 PIXEL01_22 PIXEL10_22 PIXEL11_21 break;
              }
            case 7:
            case 39:
            case 135: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_20 PIXEL11_20 break;
              }
            case 148:
            case 149:
            case 180: {
                    PIXEL00_20 PIXEL01_11 PIXEL10_20 PIXEL11_12 break;
              }
            case 224:
            case 228:
            case 225: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_12 PIXEL11_11 break;
              }
            case 41:
            case 169:
            case 45: {
                    PIXEL00_12 PIXEL01_20 PIXEL10_11 PIXEL11_20 break;
              }
            case 22:
            case 54: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_20 PIXEL11_21 break;
              }
            case 208:
            case 209: {
                    PIXEL00_20 PIXEL01_22 PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 104:
            case 108: {
                    PIXEL00_21 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 11:
            case 139: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 PIXEL10_22 PIXEL11_20 break;
              }
            case 19:
            case 51: {
                    if(Diff(w[2], w[6])) {
                    PIXEL00_11 PIXEL01_10}
                    else {
                    PIXEL00_60 PIXEL01_90}
                    PIXEL10_20 PIXEL11_21 break;
              }
            case 146:
            case 178: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_10 PIXEL11_12}
                    else {
                    PIXEL01_90 PIXEL11_61}
                    PIXEL10_20 break;
              }
            case 84:
            case 85: {
                    PIXEL00_20 if(Diff(w[6], w[8])) {
                    PIXEL01_11 PIXEL11_10}
                    else {
                    PIXEL01_60 PIXEL11_90}
                    PIXEL10_21 break;
              }
            case 112:
            case 113: {
                    PIXEL00_20 PIXEL01_22 if(Diff(w[6], w[8])) {
                    PIXEL10_12 PIXEL11_10}
                    else {
                    PIXEL10_61 PIXEL11_90}
                    break;
              }
            case 200:
            case 204: {
                    PIXEL00_21 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_10 PIXEL11_11}
                    else {
                    PIXEL10_90 PIXEL11_60}
                    break;
              }
            case 73:
            case 77: {
                    if(Diff(w[8], w[4])) {
                    PIXEL00_12 PIXEL10_10}
                    else {
                    PIXEL00_61 PIXEL10_90}
                    PIXEL01_20 PIXEL11_22 break;
              }
            case 42:
            case 170: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10 PIXEL10_11}
                    else {
                    PIXEL00_90 PIXEL10_60}
                    PIXEL01_21 PIXEL11_20 break;
              }
            case 14:
            case 142: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10 PIXEL01_12}
                    else {
                    PIXEL00_90 PIXEL01_61}
                    PIXEL10_22 PIXEL11_20 break;
              }
            case 67: {
                    PIXEL00_11 PIXEL01_21 PIXEL10_21 PIXEL11_22 break;
              }
            case 70: {
                    PIXEL00_22 PIXEL01_12 PIXEL10_21 PIXEL11_22 break;
              }
            case 28: {
                    PIXEL00_21 PIXEL01_11 PIXEL10_22 PIXEL11_21 break;
              }
            case 152: {
                    PIXEL00_21 PIXEL01_22 PIXEL10_22 PIXEL11_12 break;
              }
            case 194: {
                    PIXEL00_22 PIXEL01_21 PIXEL10_21 PIXEL11_11 break;
              }
            case 98: {
                    PIXEL00_22 PIXEL01_21 PIXEL10_12 PIXEL11_22 break;
              }
            case 56: {
                    PIXEL00_21 PIXEL01_22 PIXEL10_11 PIXEL11_21 break;
              }
            case 25: {
                    PIXEL00_12 PIXEL01_22 PIXEL10_22 PIXEL11_21 break;
              }
            case 26:
            case 31: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_22 PIXEL11_21 break;
              }
            case 82:
            case 214: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 88:
            case 248: {
                    PIXEL00_21 PIXEL01_22 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 74:
            case 107: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 27: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_10 PIXEL10_22 PIXEL11_21 break;
              }
            case 86: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_21 PIXEL11_10 break;
              }
            case 216: {
                    PIXEL00_21 PIXEL01_22 PIXEL10_10 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 106: {
                    PIXEL00_10 PIXEL01_21 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 30: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_22 PIXEL11_21 break;
              }
            case 210: {
                    PIXEL00_22 PIXEL01_10 PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 120: {
                    PIXEL00_21 PIXEL01_22 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_10 break;
              }
            case 75: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 PIXEL10_10 PIXEL11_22 break;
              }
            case 29: {
                    PIXEL00_12 PIXEL01_11 PIXEL10_22 PIXEL11_21 break;
              }
            case 198: {
                    PIXEL00_22 PIXEL01_12 PIXEL10_21 PIXEL11_11 break;
              }
            case 184: {
                    PIXEL00_21 PIXEL01_22 PIXEL10_11 PIXEL11_12 break;
              }
            case 99: {
                    PIXEL00_11 PIXEL01_21 PIXEL10_12 PIXEL11_22 break;
              }
            case 57: {
                    PIXEL00_12 PIXEL01_22 PIXEL10_11 PIXEL11_21 break;
              }
            case 71: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_21 PIXEL11_22 break;
              }
            case 156: {
                    PIXEL00_21 PIXEL01_11 PIXEL10_22 PIXEL11_12 break;
              }
            case 226: {
                    PIXEL00_22 PIXEL01_21 PIXEL10_12 PIXEL11_11 break;
              }
            case 60: {
                    PIXEL00_21 PIXEL01_11 PIXEL10_11 PIXEL11_21 break;
              }
            case 195: {
                    PIXEL00_11 PIXEL01_21 PIXEL10_21 PIXEL11_11 break;
              }
            case 102: {
                    PIXEL00_22 PIXEL01_12 PIXEL10_12 PIXEL11_22 break;
              }
            case 153: {
                    PIXEL00_12 PIXEL01_22 PIXEL10_22 PIXEL11_12 break;
              }
            case 58: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_11 PIXEL11_21 break;
              }
            case 83: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 92: {
                    PIXEL00_21 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 202: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    PIXEL01_21 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    PIXEL11_11 break;
              }
            case 78: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    PIXEL11_22 break;
              }
            case 154: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_22 PIXEL11_12 break;
              }
            case 114: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 89: {
                    PIXEL00_12 PIXEL01_22 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 90: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 55:
            case 23: {
                    if(Diff(w[2], w[6])) {
                    PIXEL00_11 PIXEL01_0}
                    else {
                    PIXEL00_60 PIXEL01_90}
                    PIXEL10_20 PIXEL11_21 break;
              }
            case 182:
            case 150: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0 PIXEL11_12}
                    else {
                    PIXEL01_90 PIXEL11_61}
                    PIXEL10_20 break;
              }
            case 213:
            case 212: {
                    PIXEL00_20 if(Diff(w[6], w[8])) {
                    PIXEL01_11 PIXEL11_0}
                    else {
                    PIXEL01_60 PIXEL11_90}
                    PIXEL10_21 break;
              }
            case 241:
            case 240: {
                    PIXEL00_20 PIXEL01_22 if(Diff(w[6], w[8])) {
                    PIXEL10_12 PIXEL11_0}
                    else {
                    PIXEL10_61 PIXEL11_90}
                    break;
              }
            case 236:
            case 232: {
                    PIXEL00_21 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_0 PIXEL11_11}
                    else {
                    PIXEL10_90 PIXEL11_60}
                    break;
              }
            case 109:
            case 105: {
                    if(Diff(w[8], w[4])) {
                    PIXEL00_12 PIXEL10_0}
                    else {
                    PIXEL00_61 PIXEL10_90}
                    PIXEL01_20 PIXEL11_22 break;
              }
            case 171:
            case 43: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0 PIXEL10_11}
                    else {
                    PIXEL00_90 PIXEL10_60}
                    PIXEL01_21 PIXEL11_20 break;
              }
            case 143:
            case 15: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0 PIXEL01_12}
                    else {
                    PIXEL00_90 PIXEL01_61}
                    PIXEL10_22 PIXEL11_20 break;
              }
            case 124: {
                    PIXEL00_21 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_10 break;
              }
            case 203: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 PIXEL10_10 PIXEL11_11 break;
              }
            case 62: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_11 PIXEL11_21 break;
              }
            case 211: {
                    PIXEL00_11 PIXEL01_10 PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 118: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_12 PIXEL11_10 break;
              }
            case 217: {
                    PIXEL00_12 PIXEL01_22 PIXEL10_10 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 110: {
                    PIXEL00_10 PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 155: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_10 PIXEL10_22 PIXEL11_12 break;
              }
            case 188: {
                    PIXEL00_21 PIXEL01_11 PIXEL10_11 PIXEL11_12 break;
              }
            case 185: {
                    PIXEL00_12 PIXEL01_22 PIXEL10_11 PIXEL11_12 break;
              }
            case 61: {
                    PIXEL00_12 PIXEL01_11 PIXEL10_11 PIXEL11_21 break;
              }
            case 157: {
                    PIXEL00_12 PIXEL01_11 PIXEL10_22 PIXEL11_12 break;
              }
            case 103: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_12 PIXEL11_22 break;
              }
            case 227: {
                    PIXEL00_11 PIXEL01_21 PIXEL10_12 PIXEL11_11 break;
              }
            case 230: {
                    PIXEL00_22 PIXEL01_12 PIXEL10_12 PIXEL11_11 break;
              }
            case 199: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_21 PIXEL11_11 break;
              }
            case 220: {
                    PIXEL00_21 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 158: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_22 PIXEL11_12 break;
              }
            case 234: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    PIXEL01_21 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_11 break;
              }
            case 242: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 59: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_11 PIXEL11_21 break;
              }
            case 121: {
                    PIXEL00_12 PIXEL01_22 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 87: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 79: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    PIXEL11_22 break;
              }
            case 122: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 94: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 218: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 91: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 229: {
                    PIXEL00_20 PIXEL01_20 PIXEL10_12 PIXEL11_11 break;
              }
            case 167: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_20 PIXEL11_20 break;
              }
            case 173: {
                    PIXEL00_12 PIXEL01_20 PIXEL10_11 PIXEL11_20 break;
              }
            case 181: {
                    PIXEL00_20 PIXEL01_11 PIXEL10_20 PIXEL11_12 break;
              }
            case 186: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_11 PIXEL11_12 break;
              }
            case 115: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 93: {
                    PIXEL00_12 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 206: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    PIXEL11_11 break;
              }
            case 205:
            case 201: {
                    PIXEL00_12 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_10}
                    else {
                    PIXEL10_70}
                    PIXEL11_11 break;
              }
            case 174:
            case 46: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_10}
                    else {
                    PIXEL00_70}
                    PIXEL01_12 PIXEL10_11 PIXEL11_20 break;
              }
            case 179:
            case 147: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_10}
                    else {
                    PIXEL01_70}
                    PIXEL10_20 PIXEL11_12 break;
              }
            case 117:
            case 116: {
                    PIXEL00_20 PIXEL01_11 PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_10}
                    else {
                    PIXEL11_70}
                    break;
              }
            case 189: {
                    PIXEL00_12 PIXEL01_11 PIXEL10_11 PIXEL11_12 break;
              }
            case 231: {
                    PIXEL00_11 PIXEL01_12 PIXEL10_12 PIXEL11_11 break;
              }
            case 126: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_10 break;
              }
            case 219: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_10 PIXEL10_10 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 125: {
                    if(Diff(w[8], w[4])) {
                    PIXEL00_12 PIXEL10_0}
                    else {
                    PIXEL00_61 PIXEL10_90}
                    PIXEL01_11 PIXEL11_10 break;
              }
            case 221: {
                    PIXEL00_12 if(Diff(w[6], w[8])) {
                    PIXEL01_11 PIXEL11_0}
                    else {
                    PIXEL01_60 PIXEL11_90}
                    PIXEL10_10 break;
              }
            case 207: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0 PIXEL01_12}
                    else {
                    PIXEL00_90 PIXEL01_61}
                    PIXEL10_10 PIXEL11_11 break;
              }
            case 238: {
                    PIXEL00_10 PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_0 PIXEL11_11}
                    else {
                    PIXEL10_90 PIXEL11_60}
                    break;
              }
            case 190: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0 PIXEL11_12}
                    else {
                    PIXEL01_90 PIXEL11_61}
                    PIXEL10_11 break;
              }
            case 187: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0 PIXEL10_11}
                    else {
                    PIXEL00_90 PIXEL10_60}
                    PIXEL01_10 PIXEL11_12 break;
              }
            case 243: {
                    PIXEL00_11 PIXEL01_10 if(Diff(w[6], w[8])) {
                    PIXEL10_12 PIXEL11_0}
                    else {
                    PIXEL10_61 PIXEL11_90}
                    break;
              }
            case 119: {
                    if(Diff(w[2], w[6])) {
                    PIXEL00_11 PIXEL01_0}
                    else {
                    PIXEL00_60 PIXEL01_90}
                    PIXEL10_12 PIXEL11_10 break;
              }
            case 237:
            case 233: {
                    PIXEL00_12 PIXEL01_20 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    PIXEL11_11 break;
              }
            case 175:
            case 47: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    PIXEL01_12 PIXEL10_11 PIXEL11_20 break;
              }
            case 183:
            case 151: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_20 PIXEL11_12 break;
              }
            case 245:
            case 244: {
                    PIXEL00_20 PIXEL01_11 PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 250: {
                    PIXEL00_10 PIXEL01_10 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 123: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_10 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_10 break;
              }
            case 95: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_10 PIXEL11_10 break;
              }
            case 222: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_10 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 252: {
                    PIXEL00_21 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 249: {
                    PIXEL00_12 PIXEL01_22 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 235: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_21 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    PIXEL11_11 break;
              }
            case 111: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_22 break;
              }
            case 63: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_11 PIXEL11_21 break;
              }
            case 159: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_22 PIXEL11_12 break;
              }
            case 215: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_21 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 246: {
                    PIXEL00_22 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 254: {
                    PIXEL00_10 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 253: {
                    PIXEL00_12 PIXEL01_11 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 251: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    PIXEL01_10 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 239: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    PIXEL01_12 if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    PIXEL11_11 break;
              }
            case 127: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_20}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_20}
                    PIXEL11_10 break;
              }
            case 191: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_11 PIXEL11_12 break;
              }
            case 223: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_20}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_10 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_20}
                    break;
              }
            case 247: {
                    PIXEL00_11 if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    PIXEL10_12 if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            case 255: {
                    if(Diff(w[4], w[2])) {
                    PIXEL00_0}
                    else {
                    PIXEL00_100}
                    if(Diff(w[2], w[6])) {
                    PIXEL01_0}
                    else {
                    PIXEL01_100}
                    if(Diff(w[8], w[4])) {
                    PIXEL10_0}
                    else {
                    PIXEL10_100}
                    if(Diff(w[6], w[8])) {
                    PIXEL11_0}
                    else {
                    PIXEL11_100}
                    break;
              }
            default:
                App_Error("GL_SmartFilterHQ2x: Invalid pattern %i.", pattern);
                break;
            }
            pOut += 2 * BPP;
        }}
        pOut += BpL;
    }}
    return dst;
    }
#undef OFFSET
#undef BPP
}
DENG2_PIMPL_NOREF(IdTech1Image) {
    Vector2i origin;
    Size nominalSize;
};
IdTech1Image::IdTech1Image(IByteArray const &data, IByteArray const &palette, Format format)
    : d(new Impl) {
    Size const rawSize(320, 200);
    if(format == Automatic) {
        if(data.size() == rawSize.x * rawSize.y) {
            format = RawVGAScreen;
        }
        else {
            format = Patch;
        }
    }
    if(format == RawVGAScreen) {
        d->nominalSize = rawSize;
        Image::operator = (fromIndexedData(rawSize, data, palette));
    }
    else {
        auto const metadata = Patch::loadMetadata(data);
        d->nominalSize = metadata.logicalDimensions;
        d->origin      = metadata.origin;
        Image::operator = (Image::fromMaskedIndexedData(metadata.dimensions,
                                                        Patch::load(data), palette));
    }
}
Vector2i IdTech1Image::origin() const {
    return d->origin;
}
Image::Size IdTech1Image::nominalSize() const {
    return d->nominalSize;
}
Image IdTech1Image::makeGameLogo(Game const &game,
                                 res::LumpCatalog const &catalog,
                                 LogoFlags flags) {
    try {
        if (flags.testFlag(AlwaysTryLoad) || game.isPlayableWithDefaultPackages()) {
            const Block playPal  = catalog.read("PLAYPAL");
            const Block title    = catalog.read("TITLE");
            const Block titlePic = catalog.read("TITLEPIC");
            const Block interPic = catalog.read("INTERPIC");
            const Block *imageToUse = &(title ? title : (titlePic ? titlePic : interPic));
            Image logoImage = Image::fromData(*imageToUse);
            if (logoImage.isNull()) {
                logoImage = IdTech1Image(*imageToUse, playPal);
            }
            const float scaleFactor = flags.testFlag(Downscale50Percent)? .5f : 1.f;
            const Image::Size finalSize(logoImage.width()  * scaleFactor,
                                        logoImage.height() * scaleFactor * 1.2f); // VGA aspect
            logoImage = logoImage.toQImage().scaled(
                finalSize.x, finalSize.y, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
            if (flags & ColorizedByFamily) {
                String const colorId = "home.icon." +
                        (game.family().isEmpty()? "other" : game.family());
                return logoImage.colorized(Style::get().colors().color(colorId));
            }
            return logoImage;
        }
    }
    catch (Error const &er) {
        if (flags & NullImageIfFails) return Image();
        LOG_RES_WARNING("Failed to load title picture for game \"%s\" using {%s}: %s")
                << game.title()
                << String::join(catalog.packages(), " ")
                << er.asText();
    }
    if (flags & NullImageIfFails) {
        return Image();
    }
    QImage img(64, 64, QImage::Format_ARGB32);
    img.fill(Qt::black);
    return img;
}
struct GraphicFileType {
    String name;
    String ext;
    bool (*interpretFunc)(FileHandle &hndl, String filePath, image_t &img);
};
static bool interpretPcx(FileHandle &hndl, String /*filePath*/, image_t &img)
{
    Image_Init(img);
    img.pixels = PCX_Load(hndl, img.size, img.pixelSize);
    return (0 != img.pixels);
}
static bool interpretJpg(FileHandle &hndl, String /*filePath*/, image_t &img)
{
    return Image_LoadFromFileWithFormat(img, "JPG", hndl);
}
static bool interpretPng(FileHandle &hndl, String /*filePath*/, image_t &img)
{
    return Image_LoadFromFileWithFormat(img, "PNG", hndl);
}
static bool interpretTga(FileHandle &hndl, String /*filePath*/, image_t &img)
{
    Image_Init(img);
    img.pixels = TGA_Load(hndl, img.size, img.pixelSize);
    return img.pixels != nullptr;
}
static GraphicFileType const graphicTypes[] = {
    { "PNG",    "png",      interpretPng /*, 0*/ },
    { "JPG",    "jpg",      interpretJpg /*, 0*/ }, // TODO: add alternate "jpeg" extension
    { "TGA",    "tga",      interpretTga /*, TGA_LastError*/ },
    { "PCX",    "pcx",      interpretPcx /*, PCX_LastError*/ },
    { "",       "",         nullptr /*,            0*/ } // Terminate.
};
static GraphicFileType const *guessGraphicFileTypeFromFileName(String fileName) {
    String ext = fileName.fileNameExtension();
    if (!ext.isEmpty()) {
        for (int i = 0; !graphicTypes[i].ext.isEmpty(); ++i) {
            GraphicFileType const &type = graphicTypes[i];
            if (!ext.compareWithoutCase(type.ext)) {
                return &type;
            }
        }
    }
    return nullptr; // Unknown.
}
static void interpretGraphic(FileHandle &hndl, String filePath, image_t &img) {
    GraphicFileType const *rtypeGuess = guessGraphicFileTypeFromFileName(filePath);
    if (rtypeGuess) {
        rtypeGuess->interpretFunc(hndl, filePath, img);
    }
    if (!img.pixels) {
        for (int i = 0; !graphicTypes[i].name.isEmpty(); ++i) {
            GraphicFileType const *graphicType = &graphicTypes[i];
            if (graphicType == rtypeGuess) continue;
            graphicTypes[i].interpretFunc(hndl, filePath, img);
            if (img.pixels) break;
        }
    }
}
static inline bool isColorKeyed(String path) {
    return path.fileNameWithoutExtension().endsWith("-ck", String::CaseInsensitive);
}
void Image_Init(image_t &img) {
    img.size      = Vector2ui(0, 0);
    img.pixelSize = 0;
    img.flags     = 0;
    img.paletteId = 0;
    img.pixels    = 0;
}
void Image_InitFromImage(image_t &img, Image const &guiImage) {
    img.size      = guiImage.size();
    img.pixelSize = guiImage.depth() / 8;
    img.flags     = 0;
    img.paletteId = 0;
    img.pixels    = reinterpret_cast<uint8_t *>(M_Malloc(guiImage.byteCount()));
    std::memcpy(img.pixels, guiImage.bits(), guiImage.byteCount());
}
void Image_ClearPixelData(image_t &img) {
    M_Free(img.pixels); img.pixels = 0;
}
image_t::Size Image_Size(image_t const &img) {
    return img.size;
}
String Image_Description(image_t const &img) {
    return String("Dimensions:%1 Flags:%2 %3:%4")
               .arg(img.size.asText())
               .arg(img.flags)
               .arg(0 != img.paletteId? "ColorPalette" : "PixelSize")
               .arg(0 != img.paletteId? img.paletteId : img.pixelSize);
}
void Image_ConvertToLuminance(image_t &img, bool retainAlpha) {
    LOG_AS("Image_ConvertToLuminance");
    uint8_t *alphaChannel = 0, *ptr = 0;
    if (0 != img.paletteId || (img.pixelSize < 3 && (img.flags & IMGF_IS_MASKED))) {
        LOG_RES_WARNING("Unknown paletted/masked image format");
        return;
    }
    long numPels = img.size.x * img.size.y;
    if (retainAlpha && img.pixelSize == 4) {
        alphaChannel = reinterpret_cast<uint8_t *>(M_Malloc(numPels));
        ptr = img.pixels;
        for (long p = 0; p < numPels; ++p, ptr += img.pixelSize) {
            alphaChannel[p] = ptr[3];
        }
    }
    ptr = img.pixels;
    for (long p = 0; p < numPels; ++p, ptr += img.pixelSize) {
        int min = de::min(ptr[0], de::min(ptr[1], ptr[2]));
        int max = de::max(ptr[0], de::max(ptr[1], ptr[2]));
        img.pixels[p] = (min == max? min : (min + max) / 2);
    }
    if (alphaChannel) {
        std::memcpy(img.pixels + numPels, alphaChannel, numPels);
        img.pixelSize = 2;
        M_Free(alphaChannel);
        return;
    }
    img.pixelSize = 1;
}
void Image_ConvertToAlpha(image_t &img, bool makeWhite) {
    Image_ConvertToLuminance(img);
    long total = img.size.x * img.size.y;
    for (long p = 0; p < total; ++p) {
        img.pixels[total + p] = img.pixels[p];
        if (makeWhite) img.pixels[p] = 255;
    }
    img.pixelSize = 2;
}
bool Image_HasAlpha(image_t const &img) {
    LOG_AS("Image_HasAlpha");
    if (0 != img.paletteId || (img.flags & IMGF_IS_MASKED)) {
        LOG_RES_WARNING("Unknown paletted/masked image format");
        return false;
    }
    if (img.pixelSize == 3) {
        return false;
    }
    if (img.pixelSize == 4) {
        long const numpels = img.size.x * img.size.y;
        uint8_t const *in = img.pixels;
        for (long i = 0; i < numpels; ++i, in += 4) {
            if (in[3] < 255) {
                return true;
            }
        }
    }
    return false;
}
uint8_t *Image_LoadFromFile(image_t &img, FileHandle &file) {
    LOG_AS("Image_LoadFromFile");
    String filePath = file.file().composePath();
    Image_Init(img);
    interpretGraphic(file, filePath, img);
    if (!img.pixels) {
        LOG_RES_XVERBOSE("\"%s\" unrecognized, trying fallback loader...",
                         NativePath(filePath).pretty());
        return 0; // Not a recognised format. It may still be loadable, however.
    }
    if (isColorKeyed(filePath)) {
        uint8_t *out = ApplyColorKeying(img.pixels, img.size.x, img.size.y, img.pixelSize);
        if (out != img.pixels) {
            M_Free(img.pixels);
            img.pixels = out;
        }
        img.pixelSize = 4;
    }
    if (Image_HasAlpha(img)) {
        img.flags |= IMGF_IS_MASKED;
    }
    LOG_RES_VERBOSE("Loaded image from file \"%s\", size %s")
            << NativePath(filePath).pretty() << img.size.asText();
    return img.pixels;
}
bool Image_LoadFromFileWithFormat(image_t &img, char const *format, FileHandle &hndl) {
    LOG_AS("Image_LoadFromFileWithFormat");
    size_t initPos = hndl.tell();
    Image_Init(img);
    QByteArray data;
    data.resize(hndl.length() - initPos);
    hndl.read(reinterpret_cast<uint8_t*>(data.data()), data.size());
    QImage image = QImage::fromData(data, format);
    if (image.isNull()) {
        hndl.seek(initPos, SeekSet);
        return false;
    }
    if (image.colorCount()) {
        image = image.convertToFormat(QImage::Format_ARGB32);
        DENG_ASSERT(!image.colorCount());
        DENG_ASSERT(image.depth() == 32);
    }
    image = image.rgbSwapped();
    img.size      = Vector2ui(image.width(), image.height());
    img.pixelSize = image.depth() / 8;
    LOGDEV_RES_VERBOSE("size:%s depth:%i alpha:%b bytes:%i")
            << img.size.asText() << img.pixelSize
            << image.hasAlphaChannel() << image.byteCount();
    img.pixels = reinterpret_cast<uint8_t *>(M_MemDup(image.constBits(), image.byteCount()));
    hndl.seek(initPos, SeekSet);
    return true;
}
bool Image_Save(image_t const &img, char const *filePath) {
    String fullPath = String(filePath);
    if (fullPath.isEmpty()) {
        static int n = 0;
        fullPath = String("image%1x%2-%3").arg(img.size.x).arg(img.size.y).arg(n++, 3);
    }
    if (fullPath.fileNameExtension().isEmpty()) {
        fullPath += ".png";
    }
    QImage image = QImage(img.pixels, img.size.x, img.size.y, QImage::Format_ARGB32);
    image = image.rgbSwapped();
    return image.save(NativePath(fullPath));
}
uint8_t *GL_LoadImage(image_t &image, String nativePath) {
    try {
        String path = (NativePath::workPath() / NativePath(nativePath).expand()).withSeparators('/');
        FileHandle &hndl = App_FileSystem().openFile(path, "rb");
        uint8_t *pixels = Image_LoadFromFile(image, hndl);
        App_FileSystem().releaseFile(hndl.file());
        delete &hndl;
        return pixels;
    }
    catch (FS1::NotFoundError const&)
    {} // Ignore error.
    return 0; // Not loaded.
}
Source GL_LoadExtImage(image_t &image, char const *_searchPath, gfxmode_t mode) {
    DENG_ASSERT(_searchPath);
    try {
        String foundPath = App_FileSystem().findPath(de::Uri(RC_GRAPHIC, _searchPath),
                                                     RLF_DEFAULT, App_ResourceClass(RC_GRAPHIC));
        foundPath = App_BasePath() / foundPath;
        if (GL_LoadImage(image, foundPath)) {
            if (mode == LGM_GRAYSCALE_ALPHA || mode == LGM_WHITE_ALPHA) {
                Image_ConvertToAlpha(image, mode == LGM_WHITE_ALPHA);
            }
            else if (mode == LGM_GRAYSCALE) {
                Image_ConvertToLuminance(image);
            }
            return External;
        }
    }
    catch (FS1::NotFoundError const&)
    {} // Ignore this error.
    return None;
}
static dd_bool palettedIsMasked(uint8_t const *pixels, int width, int height) {
    DENG2_ASSERT(pixels != 0);
    pixels += width * height;
    for (int i = 0; i < width * height; ++i) {
        if (255 != pixels[i]) {
            return true;
        }
    }
    return false;
}
static Source loadExternalTexture(image_t &image, String encodedSearchPath,
    String optionalSuffix = "") {
    try {
        String foundPath = App_FileSystem().findPath(de::Uri(encodedSearchPath + optionalSuffix, RC_GRAPHIC),
                                                     RLF_DEFAULT, App_ResourceClass(RC_GRAPHIC));
        foundPath = App_BasePath() / foundPath;
        return GL_LoadImage(image, foundPath)? External : None;
    }
    catch (FS1::NotFoundError const&)
    {} // Ignore this error.
    if (!optionalSuffix.empty()) {
        try {
            String foundPath = App_FileSystem().findPath(de::Uri(encodedSearchPath, RC_GRAPHIC),
                                                         RLF_DEFAULT, App_ResourceClass(RC_GRAPHIC));
            foundPath = App_BasePath() / foundPath;
            return GL_LoadImage(image, foundPath)? External : None;
        }
        catch (FS1::NotFoundError const&)
        {} // Ignore this error.
    }
    return None;
}
/**
 * Draw the component image @a src into the composite @a dst.
 *
 * @param dst               The composite buffer (drawn to).
 * @param dstDimensions     Pixel dimensions of @a dst.
 * @param src               The component image to be composited (read from).
 * @param srcDimensions     Pixel dimensions of @a src.
 * @param origin            Coordinates (topleft) in @a dst to draw @a src.
 *
 * @todo Optimize: Should be redesigned to composite whole rows -ds
 */
static void compositePaletted(dbyte *dst, Vector2ui const &dstDimensions,
    IByteArray const &src, Vector2ui const &srcDimensions, Vector2i const &origin) {
    if (dstDimensions == Vector2ui()) return;
    if (srcDimensions == Vector2ui()) return;
    int const       srcW = srcDimensions.x;
    int const       srcH = srcDimensions.y;
    size_t const srcPels = srcW * srcH;
    int const       dstW = dstDimensions.x;
    int const       dstH = dstDimensions.y;
    size_t const dstPels = dstW * dstH;
    int dstX, dstY;
    for (int srcY = 0; srcY < srcH; ++srcY)
    for (int srcX = 0; srcX < srcW; ++srcX) {
        dstX = origin.x + srcX;
        dstY = origin.y + srcY;
        if (dstX < 0 || dstX >= dstW) continue;
        if (dstY < 0 || dstY >= dstH) continue;
        dbyte srcAlpha;
        src.get(srcY * srcW + srcX + srcPels, &srcAlpha, 1);
        if (srcAlpha) {
            src.get(srcY * srcW + srcX, &dst[dstY * dstW + dstX], 1);
            dst[dstY * dstW + dstX + dstPels] = srcAlpha;
        }
    }
}
static String toTranslationId(int tclass, int tmap) {
#define NUM_TRANSLATION_CLASSES         3
#define NUM_TRANSLATION_MAPS_PER_CLASS  7
    if (!tclass && !tmap) return String();
    int trans = de::max(0, NUM_TRANSLATION_MAPS_PER_CLASS * tclass + tmap - 1);
    LOGDEV_RES_XVERBOSE("tclass=%i tmap=%i => TransPal# %i", tclass << tmap << trans);
    return String::number(trans);
#undef NUM_TRANSLATION_MAPS_PER_CLASS
#undef NUM_TRANSLATION_CLASSES
}
static Block loadAndTranslatePatch(IByteArray const &data, colorpaletteid_t palId,
    int tclass = 0, int tmap = 0) {
    res::ColorPalette &palette = App_Resources().colorPalettes().colorPalette(palId);
    if (res::ColorPaletteTranslation const *xlat = palette.translation(toTranslationId(tclass, tmap))) {
        return res::Patch::load(data, *xlat, res::Patch::ClipToLogicalDimensions);
    }
    else {
        return res::Patch::load(data, res::Patch::ClipToLogicalDimensions);
    }
}
static Source loadPatch(image_t &image, FileHandle &hndl, int tclass = 0,
    int tmap = 0, int border = 0) {
    LOG_AS("image_t::loadPatch");
    if (Image_LoadFromFile(image, hndl)) {
        return External;
    }
    File1 &file = hndl.file();
    ByteRefArray fileData = ByteRefArray(file.cache(), file.size());
    if (Patch::recognize(fileData)) {
        try {
            colorpaletteid_t colorPaletteId = App_Resources().colorPalettes().defaultColorPalette();
            Block patchImg = loadAndTranslatePatch(fileData, colorPaletteId, tclass, tmap);
            PatchMetadata info = Patch::loadMetadata(fileData);
            Image_Init(image);
            image.size      = Vector2ui(info.logicalDimensions.x + border*2,
                                        info.logicalDimensions.y + border*2);
            image.pixelSize = 1;
            image.paletteId = colorPaletteId;
            image.pixels = (uint8_t *) M_Calloc(2 * image.size.x * image.size.y);
            compositePaletted(image.pixels, image.size,
                              patchImg, info.logicalDimensions, Vector2i(border, border));
            if (palettedIsMasked(image.pixels, image.size.x, image.size.y)) {
                image.flags |= IMGF_IS_MASKED;
            }
            return Original;
        }
        catch (IByteArray::OffsetError const &) {
            LOG_RES_WARNING("File \"%s:%s\" does not appear to be a valid Patch")
                << NativePath(file.container().composePath()).pretty()
                << NativePath(file.composePath()).pretty();
        }
    }
    file.unlock();
    return None;
}
static Source loadPatchComposite(image_t &image, Texture const &tex,
    bool maskZero = false, bool useZeroOriginIfOneComponent = false) {
    LOG_AS("image_t::loadPatchComposite");
    Image_Init(image);
    image.pixelSize = 1;
    image.size      = Vector2ui(tex.width(), tex.height());
    image.paletteId = App_Resources().colorPalettes().defaultColorPalette();
    image.pixels = (uint8_t *) M_Calloc(2 * image.size.x * image.size.y);
    res::Composite const &texDef = *reinterpret_cast<res::Composite *>(tex.userDataPointer());
    DENG2_FOR_EACH_CONST(res::Composite::Components, i, texDef.components()) {
        File1 &file           = App_FileSystem().lump(i->lumpNum());
        ByteRefArray fileData = ByteRefArray(file.cache(), file.size());
        if (Patch::recognize(fileData)) {
            try {
                Patch::Flags loadFlags;
                if (maskZero) loadFlags |= Patch::MaskZero;
                Block patchImg     = Patch::load(fileData, loadFlags);
                PatchMetadata info = Patch::loadMetadata(fileData);
                Vector2i origin = i->origin();
                if (useZeroOriginIfOneComponent && texDef.componentCount() == 1) {
                    origin = Vector2i(0, 0);
                }
                compositePaletted(image.pixels, image.size,
                                  patchImg, info.dimensions, origin);
            }
            catch (IByteArray::OffsetError const &)
            {} // Ignore this error.
        }
        file.unlock();
    }
    if (maskZero || palettedIsMasked(image.pixels, image.size.x, image.size.y)) {
        image.flags |= IMGF_IS_MASKED;
    }
    return Original;
}
static Source loadFlat(image_t &image, FileHandle &hndl) {
    if (Image_LoadFromFile(image, hndl)) {
        return External;
    }
    Image_Init(image);
    image.size      = Vector2ui(64, 64);
    image.pixelSize = 1;
    image.paletteId = App_Resources().colorPalettes().defaultColorPalette();
    File1 &file   = hndl.file();
    size_t fileLength = hndl.length();
    size_t bufSize = de::max(fileLength, (size_t) image.size.x * image.size.y);
    image.pixels = (uint8_t *) M_Malloc(bufSize);
    if (fileLength < bufSize) {
        std::memset(image.pixels, 0, bufSize);
    }
    file.read(image.pixels, 0, fileLength);
    return Original;
}
static Source loadDetail(image_t &image, FileHandle &hndl) {
    if (Image_LoadFromFile(image, hndl)) {
        return Original;
    }
    Image_Init(image);
    File1 &file = hndl.file();
    size_t fileLength = hndl.length();
    switch (fileLength) {
    case 256 * 256: image.size.x = image.size.y = 256; break;
    case 128 * 128: image.size.x = image.size.y = 128; break;
    case  64 *  64: image.size.x = image.size.y =  64; break;
    default:
        throw Error("image_t::loadDetail", "Must be 256x256, 128x128 or 64x64.");
    }
    image.pixelSize = 1;
    size_t bufSize = (size_t) image.size.x * image.size.y;
    image.pixels = (uint8_t *) M_Malloc(bufSize);
    if (fileLength < bufSize) {
        std::memset(image.pixels, 0, bufSize);
    }
    file.read(image.pixels, fileLength);
    return Original;
}
Source GL_LoadSourceImage(image_t &image, ClientTexture const &tex,
                          TextureVariantSpec const &spec) {
    de::FS1 &fileSys = App_FileSystem();
    auto &cfg = R_Config();
    Source source = None;
    variantspecification_t const &vspec = spec.variant;
    if (!tex.manifest().schemeName().compareWithoutCase("Textures")) {
        if (cfg.noHighResTex->value().isFalse() &&
                (loadExtAlways || cfg.highResWithPWAD->value().isTrue() || !tex.isFlagged(Texture::Custom))) {
            de::Uri uri = tex.manifest().composeUri();
            source = loadExternalTexture(image, uri.compose(), "-ck");
        }
        if (source == None) {
            if (TC_SKYSPHERE_DIFFUSE != vspec.context) {
                source = loadPatchComposite(image, tex);
            }
            else {
                bool const zeroMask = (vspec.flags & TSF_ZEROMASK) != 0;
                bool const useZeroOriginIfOneComponent = true;
                source = loadPatchComposite(image, tex, zeroMask, useZeroOriginIfOneComponent);
            }
        }
    }
    else if (!tex.manifest().schemeName().compareWithoutCase("Flats")) {
        if (cfg.noHighResTex->value().isFalse() &&
                (loadExtAlways || cfg.highResWithPWAD->value().isTrue() || !tex.isFlagged(Texture::Custom))) {
            de::Uri uri = tex.manifest().composeUri();
            source = loadExternalTexture(image, uri.compose(), "-ck");
            if (source == None) {
                source = loadExternalTexture(image, "Textures:flat-" + uri.path().toStringRef(), "-ck");
            }
        }
        if (source == None) {
            if (tex.manifest().hasResourceUri()) {
                de::Uri resourceUri = tex.manifest().resourceUri();
                if (!resourceUri.scheme().compareWithoutCase("LumpIndex")) {
                    try {
                        lumpnum_t const lumpNum = resourceUri.path().toString().toInt();
                        FileHandle &hndl    = fileSys.openLump(fileSys.lump(lumpNum));
                        source = loadFlat(image, hndl);
                        fileSys.releaseFile(hndl.file());
                        delete &hndl;
                    }
                    catch (LumpIndex::NotFoundError const&)
                    {} // Ignore this error.
                }
            }
        }
    }
    else if (!tex.manifest().schemeName().compareWithoutCase("Patches")) {
        int tclass = 0, tmap = 0;
        if (vspec.flags & TSF_HAS_COLORPALETTE_XLAT) {
            tclass = vspec.tClass;
            tmap   = vspec.tMap;
        }
        if (cfg.noHighResTex->value().isFalse() &&
                (loadExtAlways || cfg.highResWithPWAD->value().isTrue() || !tex.isFlagged(Texture::Custom))) {
            de::Uri uri = tex.manifest().composeUri();
            source = loadExternalTexture(image, uri.compose(), "-ck");
        }
        if (source == None) {
            if (tex.manifest().hasResourceUri()) {
                de::Uri resourceUri = tex.manifest().resourceUri();
                if (!resourceUri.scheme().compareWithoutCase("LumpIndex")) {
                    try {
                        lumpnum_t const lumpNum = resourceUri.path().toString().toInt();
                        FileHandle &hndl    = fileSys.openLump(fileSys.lump(lumpNum));
                        source = loadPatch(image, hndl, tclass, tmap, vspec.border);
                        fileSys.releaseFile(hndl.file());
                        delete &hndl;
                    }
                    catch (LumpIndex::NotFoundError const&)
                    {} // Ignore this error.
                }
            }
        }
    }
    else if (!tex.manifest().schemeName().compareWithoutCase("Sprites")) {
        int tclass = 0, tmap = 0;
        if (vspec.flags & TSF_HAS_COLORPALETTE_XLAT) {
            tclass = vspec.tClass;
            tmap   = vspec.tMap;
        }
        if (cfg.noHighResPatches->value().isFalse()) {
            de::Uri uri = tex.manifest().composeUri();
            if (TC_PSPRITE_DIFFUSE == vspec.context) {
                source = loadExternalTexture(image, "Patches:" + uri.path() + "-hud", "-ck");
            }
            else if (tclass || tmap) {
                source = loadExternalTexture(image, "Patches:" + uri.path() + String("-table%1%2").arg(tclass).arg(tmap), "-ck");
            }
            if (!source) {
                source = loadExternalTexture(image, "Patches:" + uri.path(), "-ck");
            }
        }
        if (source == None) {
            if (tex.manifest().hasResourceUri()) {
                de::Uri resourceUri = tex.manifest().resourceUri();
                if (!resourceUri.scheme().compareWithoutCase("LumpIndex")) {
                    try {
                        lumpnum_t const lumpNum = resourceUri.path().toString().toInt();
                        FileHandle &hndl    = fileSys.openLump(fileSys.lump(lumpNum));
                        source = loadPatch(image, hndl, tclass, tmap, vspec.border);
                        fileSys.releaseFile(hndl.file());
                        delete &hndl;
                    }
                    catch (LumpIndex::NotFoundError const&)
                    {} // Ignore this error.
                }
            }
        }
    }
    else if (!tex.manifest().schemeName().compareWithoutCase("Details")) {
        if (tex.manifest().hasResourceUri()) {
            de::Uri resourceUri = tex.manifest().resourceUri();
            if (resourceUri.scheme().compareWithoutCase("Lumps")) {
                source = loadExternalTexture(image, resourceUri.compose());
            }
            else {
                lumpnum_t const lumpNum = fileSys.lumpNumForName(resourceUri.path());
                try {
                    File1 &lump = fileSys.lump(lumpNum);
                    FileHandle &hndl = fileSys.openLump(lump);
                    source = loadDetail(image, hndl);
                    fileSys.releaseFile(hndl.file());
                    delete &hndl;
                }
                catch (LumpIndex::NotFoundError const&)
                {} // Ignore this error.
            }
        }
    }
    else {
        if (tex.manifest().hasResourceUri()) {
            de::Uri resourceUri = tex.manifest().resourceUri();
            source = loadExternalTexture(image, resourceUri.compose());
        }
    }
    return source;
}
LightMaterialDecoration::AnimationStage::AnimationStage(int tics, float variance,
    Vector2f const &origin, float elevation, Vector3f const &color, float radius,
    float haloRadius, LightRange const &lightLevels, ClientTexture *ceilingTexture,
    ClientTexture *floorTexture, ClientTexture *texture, ClientTexture *flareTexture, int sysFlareIdx)
    : Stage(tics, variance)
    , origin     (origin)
    , elevation  (elevation)
    , color      (color)
    , radius     (radius)
    , haloRadius (haloRadius)
    , lightLevels(lightLevels)
    , tex        (texture)
    , floorTex   (floorTexture)
    , ceilTex    (ceilingTexture)
    , flareTex   (flareTexture)
    , sysFlareIdx(sysFlareIdx)
{}
LightMaterialDecoration::AnimationStage::AnimationStage(AnimationStage const &other)
    : Stage(other)
    , origin     (other.origin)
    , elevation  (other.elevation)
    , color      (other.color)
    , radius     (other.radius)
    , haloRadius (other.haloRadius)
    , lightLevels(other.lightLevels)
    , tex        (other.tex)
    , floorTex   (other.floorTex)
    , ceilTex    (other.ceilTex)
    , flareTex   (other.flareTex)
    , sysFlareIdx(other.sysFlareIdx)
{}
LightMaterialDecoration::AnimationStage *
LightMaterialDecoration::AnimationStage::fromDef(Record const &stageDef) {
    ClientTexture *lightmapUp   = static_cast<ClientTexture *>(res::Textures::get().tryFindTextureByResourceUri("Lightmaps", de::makeUri(stageDef.gets("lightmapUp"  ))));
    ClientTexture *lightmapDown = static_cast<ClientTexture *>(res::Textures::get().tryFindTextureByResourceUri("Lightmaps", de::makeUri(stageDef.gets("lightmapDown"))));
    ClientTexture *lightmapSide = static_cast<ClientTexture *>(res::Textures::get().tryFindTextureByResourceUri("Lightmaps", de::makeUri(stageDef.gets("lightmapSide"))));
    int haloTextureIndex  = stageDef.geti("haloTextureIndex");
    ClientTexture *haloTexture  = nullptr;
    de::Uri const haloTextureUri(stageDef.gets("haloTexture"), RC_NULL);
    if(!haloTextureUri.isEmpty()) {
        if(haloTextureUri.path().length() == 1 &&
           haloTextureUri.path().toStringRef().first().isDigit()) {
            haloTextureIndex = haloTextureUri.path().toStringRef().first().digitValue();
        }
        else {
            haloTexture = static_cast<ClientTexture *>(res::Textures::get().tryFindTextureByResourceUri("Flaremaps", haloTextureUri));
        }
    }
    return new AnimationStage(stageDef.geti("tics"), stageDef.getf("variance"),
                              Vector2f(stageDef.geta("origin")), stageDef.getf("elevation"),
                              Vector3f(stageDef.geta("color")), stageDef.getf("radius"),
                              stageDef.getf("haloRadius"),
                              LightRange(Vector2f(stageDef.geta("lightLevels"))),
                              lightmapUp, lightmapDown, lightmapSide,
                              haloTexture, haloTextureIndex);
}
String LightMaterialDecoration::AnimationStage::description() const {
    return String(_E(l) "Tics: ")      + _E(.) + (tics > 0? String("%1 (~%2)").arg(tics).arg(variance, 0, 'g', 2) : "-1")
                + _E(l) " Origin: "      _E(.) + origin.asText()
                + _E(l) " Elevation: "   _E(.) + String::number(elevation, 'f', 2)
                + _E(l) " LightLevels: " _E(.) + lightLevels.asText()
                + _E(l) "\nColor: "      _E(.) + color.asText()
                + _E(l) " Radius: "      _E(.) + String::number(radius, 'f', 2)
                + _E(l) " HaloRadius: "  _E(.) + String::number(haloRadius, 'f', 2);
}
LightMaterialDecoration::LightMaterialDecoration(Vector2i const &patternSkip,
    Vector2i const &patternOffset, bool useInterpolation)
    : Decoration(patternSkip, patternOffset)
    , _useInterpolation(useInterpolation)
{}
LightMaterialDecoration::~LightMaterialDecoration()
{}
LightMaterialDecoration *LightMaterialDecoration::fromDef(Record const &definition) {
    defn::MaterialDecoration decorDef(definition);
    auto *decor = new LightMaterialDecoration(Vector2i(decorDef.geta("patternSkip")),
                                              Vector2i(decorDef.geta("patternOffset")));
    for(int i = 0; i < decorDef.stageCount(); ++i) {
        decor->_stages.append(AnimationStage::fromDef(decorDef.stage(i)));
    }
    return decor;
}
int LightMaterialDecoration::addStage(AnimationStage const &stageToCopy) {
    _stages.append(new AnimationStage(stageToCopy));
    return _stages.count() - 1;
}
LightMaterialDecoration::AnimationStage &LightMaterialDecoration::stage(int index) const {
    return static_cast<AnimationStage &>(Decoration::stage(index));
}
String LightMaterialDecoration::describe() const {
    return "Light decoration";
}
bool LightMaterialDecoration::useInterpolation() const {
    return _useInterpolation;
}
static inline ClientResources &resSys() {
    return ClientApp::resources();
}
/**
 * Attempt to locate and prepare a flare texture. Somewhat more complicated than
 * it needs to be due to the fact there are two different selection methods.
 *
 * @param texture  Logical texture to prepare an variant of.
 * @param oldIdx   Old method of flare texture selection, by id.
 *
 * @return  @c 0= Use the automatic selection logic.
 */
static DGLuint prepareFlaremap(ClientTexture *texture, int oldIdx) {
    if (texture) {
        if (TextureVariant const *variant = texture->prepareVariant(Rend_HaloTextureSpec())) {
            return variant->glName();
        }
    }
    else if (oldIdx > 0 && oldIdx < NUM_SYSFLARE_TEXTURES) {
        return GL_PrepareSysFlaremap(flaretexid_t(oldIdx - 1));
    }
    return 0; // Use the automatic selection logic.
}
DENG2_PIMPL_NOREF(MaterialAnimator::Decoration) {
    MaterialDecoration *matDecor = nullptr;
    int stage   = 0;  ///< Animation stage else @c -1 => decoration not in use.
    short tics  = 0;  ///< Remaining (sharp) tics in the current stage.
    float inter = 0;  ///< Intermark from the current stage to the next [0..1].
    Vector2f origin;           ///< Relative position in material space.
    Vector3f color;            ///< Light color.
    float elevation      = 0;  ///< Distance from the surface.
    float radius         = 0;  ///< Dynamic light radius (-1 = no light).
    float lightLevels[2];      ///< Fade by sector lightlevel.
    float flareSize      = 0;  ///< Halo radius (zero = no halo).
    DGLuint flareTex     = 0;
    ClientTexture *tex      = nullptr;
    ClientTexture *ceilTex  = nullptr;
    ClientTexture *floorTex = nullptr;
    Impl() { de::zap(lightLevels); }
    bool useInterpolation() const {
        DENG2_ASSERT(matDecor);
        if (auto const *light = maybeAs<LightMaterialDecoration>(matDecor)) {
            return light->useInterpolation();
        }
        return true;
    }
};
MaterialAnimator::Decoration::Decoration(MaterialDecoration &decor)
    : d(new Impl) {
    d->matDecor = &decor;
}
MaterialDecoration &MaterialAnimator::Decoration::decor() const {
    DENG2_ASSERT(d->matDecor);
    return *d->matDecor;
}
Vector2f MaterialAnimator::Decoration::origin() const {
    return d->origin;
}
Vector3f MaterialAnimator::Decoration::color() const {
    return d->color;
}
float MaterialAnimator::Decoration::elevation() const {
    return d->elevation;
}
float MaterialAnimator::Decoration::radius() const {
    return d->radius;
}
void MaterialAnimator::Decoration::lightLevels(float &min, float &max) const {
    min = d->lightLevels[0];
    max = d->lightLevels[1];
}
float MaterialAnimator::Decoration::flareSize() const {
    return d->flareSize;
}
DGLuint MaterialAnimator::Decoration::flareTex() const {
    return d->flareTex;
}
ClientTexture *MaterialAnimator::Decoration::tex() const {
    return d->tex;
}
ClientTexture *MaterialAnimator::Decoration::ceilTex() const {
    return d->ceilTex;
}
ClientTexture *MaterialAnimator::Decoration::floorTex() const {
    return d->floorTex;
}
void MaterialAnimator::Decoration::rewind() {
    d->stage = 0;
    d->tics  = decor().stage(0).tics;
    d->inter = 0;
}
bool MaterialAnimator::Decoration::animate() {
    if (decor().isAnimated()) {
        d->inter = 0;
        if (DD_IsSharpTick() && d->tics-- <= 0) {
            if (++d->stage == decor().stageCount()) {
                d->stage = 0;
            }
            MaterialDecoration::Stage const &stage = decor().stage(d->stage);
            if (stage.variance != 0)
                d->tics = stage.tics * (1 - stage.variance * RNG_RandFloat());
            else
                d->tics = stage.tics;
            return true;
        }
        if (d->useInterpolation()) {
            MaterialDecoration::Stage const &stage = decor().stage(d->stage);
            d->inter = 1.f - d->tics / float( stage.tics );
        }
    }
    return false;
}
void MaterialAnimator::Decoration::update() {
    if (auto *lightDecor = maybeAs<LightMaterialDecoration>(decor())) {
        LightMaterialDecoration::AnimationStage const &stage = lightDecor->stage(d->stage);
        LightMaterialDecoration::AnimationStage const &next  = lightDecor->stage(d->stage + 1);
        d->origin         = de::lerp(stage.origin,          next.origin,          d->inter);
        d->elevation      = de::lerp(stage.elevation,       next.elevation,       d->inter);
        d->radius         = de::lerp(stage.radius,          next.radius,          d->inter);
        d->flareSize      = de::lerp(stage.haloRadius,      next.haloRadius,      d->inter);
        d->lightLevels[0] = de::lerp(stage.lightLevels.min, next.lightLevels.min, d->inter);
        d->lightLevels[1] = de::lerp(stage.lightLevels.max, next.lightLevels.max, d->inter);
        d->color          = de::lerp(stage.color,           next.color,           d->inter);
        d->tex            = stage.tex;
        d->ceilTex        = stage.ceilTex;
        d->floorTex       = stage.floorTex;
        d->flareTex       = prepareFlaremap(stage.flareTex, stage.sysFlareIdx);
    }
}
void MaterialAnimator::Decoration::reset() {
    d->origin    = Vector2i(0, 0);
    d->color     = Vector3f(0, 0, 0);
    d->elevation = 0;
    d->radius    = 0;
    de::zap(d->lightLevels);
    d->flareSize = 0;
    d->flareTex  = 0;
    d->tex       = nullptr;
    d->ceilTex   = nullptr;
    d->floorTex  = nullptr;
}
enum StagePropertyId { TextureProperty, MaskTextureProperty };
/**
 * Returns the Texture in effect for the given animation stage, if any.
 */
static ClientTexture *findTextureForAnimationStage(world::TextureMaterialLayer::AnimationStage const &stage,
                                                   /*String const &propertyName = "texture"*/
                                                   StagePropertyId prop = TextureProperty) {
    try {
        return static_cast<ClientTexture *>
                (&res::Textures::get().texture(prop == TextureProperty? stage.texture
                                                                      : stage.maskTexture));
    }
    catch (res::TextureManifest::MissingTextureError &)
    {}
    catch (Resources::MissingResourceManifestError &)
    {}
    return nullptr;
}
DENG2_PIMPL(MaterialAnimator) {
    ClientMaterial *material = nullptr;         ///< Material to animate (not owned).
    MaterialVariantSpec const *spec = nullptr;  ///< Variant specification.
    struct LayerState {
        int stage     = 0;   ///< Animation stage else @c -1 => layer not in use.
        int nextStage = 0;
        short tics    = 0;   ///< Remaining (sharp) tics in the current stage.
        float inter   = 0.f; ///< Intermark from the current stage to the next [0..1].
        QVector<ClientTexture *> stageTextures;
        String synopsis() const {
            return String("stage: %1 tics: %2 inter: %3").arg(stage).arg(tics).arg(inter);
        }
    };
    QList<LayerState *> layers;
    /**
     * Cached animation state snapshot.
     *
     * Stage-animated or interpolated material property values are cached in a
     * per-frame data store to avoid repeat calculation. All other values that do
     * not change should be obtained directly from the Material.
     */
    struct Snapshot {
        bool opaque;
        float glowStrength;
        Vector2ui dimensions;
        blendmode_t shineBlendMode;
        Vector3f shineMinColor;
        std::array<TextureVariant *, MaterialAnimator::NUM_TEXTUREUNITS> textures;
        std::array<GLTextureUnit, MaterialAnimator::NUM_TEXTUREUNITS> units;
        Snapshot() { clear(); }
        void clear() {
            dimensions     = Vector2ui(0, 0);
            shineBlendMode = BM_NORMAL;
            shineMinColor  = Vector3f(0, 0, 0);
            opaque         = true;
            glowStrength   = 0;
            textures.fill(nullptr);
            units.fill(GLTextureUnit());
        }
    };
    std::unique_ptr<Snapshot> snapshot;
    int lastSnapshotUpdate = -1;  ///< Frame count of last snapshot update.
    QList<Decoration *> decorations;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        clearLayers();
        clearDecorations();
    }
    void clearLayers() {
        qDeleteAll(layers); layers.clear();
    }
    void initLayers() {
        clearLayers();
        for (int i = 0; i < self().material().layerCount(); ++i) {
            layers << new LayerState;
        }
        fetchStageTextures();
    }
    void fetchStageTextures() {
        for (int i = 0; i < layers.size(); ++i) {
            layers[i]->stageTextures.clear();
            if (world::TextureMaterialLayer const *layer = maybeAs<world::TextureMaterialLayer>(self().material().layer(i))) {
                for (int k = 0; k < layer->stageCount(); ++k) {
                    layers[i]->stageTextures.append(findTextureForAnimationStage(layer->stage(k)));
                }
            }
        }
    }
    void clearDecorations() {
        qDeleteAll(decorations); decorations.clear();
    }
    void initDecorations() {
        clearDecorations();
        self().material().forAllDecorations([this] (MaterialDecoration &decor) {
            decorations << new Decoration(decor);
            return LoopContinue;
        });
    }
    void attachMissingSnapshot() {
        if (snapshot) return;
        snapshot.reset(new Snapshot);
        lastSnapshotUpdate = -1; // Force an update.
    }
    void updateSnapshotIfNeeded(bool force = false) {
        attachMissingSnapshot();
        if (!force && lastSnapshotUpdate == R_FrameCount())
            return;
        lastSnapshotUpdate = R_FrameCount();
        snapshot->clear();
        for (Decoration *decor : decorations) decor->reset();
        /*
         * Ensure all resources needed to visualize this have been prepared. If
         * skymasked, we only need to update the primary tex unit (due to it being
         * visible when skymask debug drawing is enabled).
         */
        if (!material->isSkyMasked() || ::devRendSkyMode) {
            int texLayerIndex = 0;
            for (int i = 0; i < material->layerCount(); ++i) {
                world::MaterialLayer const &layer = material->layer(i);
                LayerState const &ls       = *layers[i];
                if (auto const *detailLayer = maybeAs<world::DetailTextureMaterialLayer>(layer)) {
                    auto const &stage = detailLayer->stage(ls.stage).as<world::DetailTextureMaterialLayer::AnimationStage>();
                    auto const &next  = detailLayer->stage(ls.nextStage).as<world::DetailTextureMaterialLayer::AnimationStage>();
                    if (ClientTexture *tex = ls.stageTextures[ls.stage]) {
                        float const contrast = de::clamp(0.f, stage.strength, 1.f) * ::detailFactor /*Global strength multiplier*/;
                        snapshot->textures[TU_DETAIL] = tex->prepareVariant(resSys().detailTextureSpec(contrast));
                    }
                    if (::smoothTexAnim && &stage != &next) {
                        if (ClientTexture *tex = ls.stageTextures[ls.nextStage]) {
                            float const contrast = de::clamp(0.f, next.strength, 1.f) * ::detailFactor /*Global strength multiplier*/;
                            snapshot->textures[TU_DETAIL_INTER] = tex->prepareVariant(resSys().detailTextureSpec(contrast));
                        }
                    }
                }
                else if (is<world::ShineTextureMaterialLayer>(layer)) {
                    world::TextureMaterialLayer::AnimationStage const &stage = layer.as<world::TextureMaterialLayer>().stage(ls.stage);
                    if (ClientTexture *tex = ls.stageTextures[ls.stage]) {
                        snapshot->textures[TU_SHINE] = tex->prepareVariant(Rend_MapSurfaceShinyTextureSpec());
                        if (ClientTexture *maskTex = findTextureForAnimationStage(stage, MaskTextureProperty)) {
                            snapshot->textures[TU_SHINE_MASK] = maskTex->prepareVariant(Rend_MapSurfaceShinyMaskTextureSpec());
                        }
                    }
                }
                else if (auto const *texLayer = maybeAs<world::TextureMaterialLayer>(layer)) {
                    world::TextureMaterialLayer::AnimationStage const &stage = texLayer->stage(ls.stage);
                    world::TextureMaterialLayer::AnimationStage const &next  = texLayer->stage(ls.nextStage);
                    if (ClientTexture *tex = ls.stageTextures[ls.stage]) {
                        snapshot->textures[TU_LAYER0 + texLayerIndex] = tex->prepareVariant(*spec->primarySpec);
                    }
                    if (::smoothTexAnim && &stage != &next) {
                        if (ClientTexture *tex = ls.stageTextures[ls.nextStage]) {
                            snapshot->textures[TU_LAYER0_INTER + texLayerIndex] = tex->prepareVariant(*spec->primarySpec);
                        }
                    }
                    texLayerIndex += 1;
                }
            }
        }
        snapshot->dimensions = material->dimensions();
        snapshot->opaque     = (snapshot->textures[TU_LAYER0] && !snapshot->textures[TU_LAYER0]->isMasked());
        if (snapshot->dimensions == Vector2ui()) return;
        if (material->isSkyMasked() && !::devRendSkyMode) return;
        int texLayerIndex = 0;
        for (int i = 0; i < material->layerCount(); ++i) {
            world::MaterialLayer const &layer = material->layer(i);
            LayerState const &ls = *layers[i];
            if (auto const *detailLayer = maybeAs<world::DetailTextureMaterialLayer>(layer)) {
                if (TextureVariant *tex = snapshot->textures[TU_DETAIL]) {
                    auto const &stage = detailLayer->stage(ls.stage).as<world::DetailTextureMaterialLayer::AnimationStage>();
                    auto const &next  = detailLayer->stage(ls.nextStage).as<world::DetailTextureMaterialLayer::AnimationStage>();
                    float scale = de::lerp(stage.scale, next.scale, ls.inter);
                    if (::detailScale > .0001f) scale *= ::detailScale; // Global scale factor.
                    snapshot->units[TU_DETAIL] =
                            GLTextureUnit(*tex, Vector2f(1, 1) / tex->base().dimensions() * scale);
                    if (TextureVariant *tex = snapshot->textures[TU_DETAIL_INTER]) {
                        if (!(!fogParams.usingFog && ls.inter == 0)) {
                            snapshot->units[TU_DETAIL_INTER] =
                                GLTextureUnit(*tex,
                                              snapshot->units[TU_DETAIL].scale,
                                              snapshot->units[TU_DETAIL].offset,
                                              de::clamp(0.f, ls.inter, 1.f));
                        }
                    }
                }
            }
            else if (is<world::ShineTextureMaterialLayer>(layer)) {
                if (TextureVariant *tex = snapshot->textures[TU_SHINE]) {
                    auto const &stage = layer.stage(ls.stage)    .as<world::ShineTextureMaterialLayer::AnimationStage>();
                    auto const &next  = layer.stage(ls.nextStage).as<world::ShineTextureMaterialLayer::AnimationStage>();
                    Vector2f const origin   = de::lerp(stage.origin,   next.origin,   ls.inter);
                    Vector3f const minColor = de::lerp(stage.minColor, next.minColor, ls.inter);
                    float    const opacity  = de::lerp(stage.opacity,  next.opacity,  ls.inter);
                    snapshot->shineBlendMode = stage.blendMode;
                    snapshot->shineMinColor  = minColor.min(Vector3f(1, 1, 1)).max(Vector3f(0, 0, 0));
                    snapshot->units[TU_SHINE] = GLTextureUnit(*tex, Vector2f(1, 1), origin, de::clamp(0.0f, opacity, 1.0f));
                    if (TextureVariant *maskTex = snapshot->textures[TU_SHINE_MASK]) {
                        snapshot->units[TU_SHINE_MASK] =
                            GLTextureUnit(*maskTex, Vector2f(1, 1) / (snapshot->dimensions * maskTex->base().dimensions()),
                                          snapshot->units[TU_LAYER0].offset);
                    }
                }
            }
            else if (auto const *texLayer = maybeAs<world::TextureMaterialLayer>(layer)) {
                if (TextureVariant *tex = snapshot->textures[TU_LAYER0 + texLayerIndex]) {
                    world::TextureMaterialLayer::AnimationStage const &stage = texLayer->stage(ls.stage);
                    world::TextureMaterialLayer::AnimationStage const &next  = texLayer->stage(ls.nextStage);
                    Vector2f const scale   = Vector2f(1, 1) / snapshot->dimensions;
                    Vector2f const origin  = de::lerp(stage.origin,  next.origin,  ls.inter);
                    float    const opacity = de::lerp(stage.opacity, next.opacity, ls.inter);
                    snapshot->units[TU_LAYER0 + texLayerIndex] = GLTextureUnit(*tex, scale, origin, de::clamp(0.0f, opacity, 1.0f));
                    if (texLayerIndex == 0) {
                        snapshot->glowStrength = de::lerp(stage.glowStrength, next.glowStrength, ls.inter);
                    }
                    if (TextureVariant *tex = snapshot->textures[TU_LAYER0_INTER + texLayerIndex]) {
                        if (!(!fogParams.usingFog && ls.inter == 0)) {
                            snapshot->units[TU_LAYER0_INTER + texLayerIndex] =
                                GLTextureUnit(*tex,
                                              snapshot->units[TU_LAYER0 + texLayerIndex].scale,
                                              snapshot->units[TU_LAYER0 + texLayerIndex].offset,
                                              de::clamp(0.f, ls.inter, 1.f));
                        }
                    }
                    texLayerIndex += 1;
                }
            }
        }
        if (!material->isSkyMasked())
        for (Decoration *decor : decorations) {
            decor->update();
        }
    }
    void rewindLayer(LayerState &ls, world::MaterialLayer const &layer) {
        ls.stage = 0;
        ls.nextStage = layer.nextStageIndex(0);
        ls.tics  = layer.stage(0).tics;
        ls.inter = 0;
    }
    void animateLayer(LayerState &ls, world::MaterialLayer const &layer) {
        if (DD_IsSharpTick() && ls.tics-- <= 0) {
            if (++ls.stage == layer.stageCount()) {
                ls.stage = 0;
            }
            ls.nextStage = layer.nextStageIndex(ls.stage);
            ls.inter = 0;
            world::MaterialLayer::Stage const &stage = layer.stage(ls.stage);
            if (stage.variance != 0)
                ls.tics = stage.tics * (1 - stage.variance * RNG_RandFloat());
            else
                ls.tics = stage.tics;
        }
        else {
            world::MaterialLayer::Stage const &stage = layer.stage(ls.stage);
            ls.inter = 1.f - ls.tics / float( stage.tics );
        }
    }
};
MaterialAnimator::MaterialAnimator(ClientMaterial &material, MaterialVariantSpec const &spec)
    : d(new Impl(this)) {
    d->material = &material;
    d->spec     = &spec;
    d->initLayers();
    d->initDecorations();
    rewind();
}
ClientMaterial &MaterialAnimator::material() const {
    DENG2_ASSERT(d->material);
    return *d->material;
}
MaterialVariantSpec const &MaterialAnimator::variantSpec() const {
    DENG2_ASSERT(d->spec);
    return *d->spec;
}
bool MaterialAnimator::isPaused() const {
    MaterialContextId context = variantSpec().contextId;
    return (clientPaused && (context == MapSurfaceContext ||
                             context == SpriteContext     ||
                             context == ModelSkinContext  ||
                             context == PSpriteContext    ||
                             context == SkySphereContext));
}
void MaterialAnimator::animate(timespan_t /*ticLength*/)
{
    if (!material().isValid()) return;
    if (isPaused()) return;
    /*
     * Animate layers:
     */
    for (int i = 0; i < material().layerCount(); ++i) {
        world::MaterialLayer const &layer = material().layer(i);
        if (layer.isAnimated() && is<world::TextureMaterialLayer>(layer)) {
            d->animateLayer(*d->layers[i], layer);
        }
    }
    /*
     * Animate decorations:
     */
    bool decorationStageChanged = false;
    for (Decoration *decor : d->decorations) {
        if (decor->animate()) {
            decorationStageChanged = true;
        }
    }
    if (decorationStageChanged) {
        DENG2_FOR_AUDIENCE(DecorationStageChange, i) i->materialAnimatorDecorationStageChanged(*this);
    }
}
void MaterialAnimator::rewind() {
    if (!material().isValid()) return;
    for (int i = 0; i < material().layerCount(); ++i) {
        d->rewindLayer(*d->layers[i], material().layer(i));
    }
    for (int i = 0; i < material().decorationCount(); ++i) {
        d->decorations[i]->rewind();
    }
}
void MaterialAnimator::prepare(bool fullUpdate) {
    d->updateSnapshotIfNeeded(fullUpdate);
}
void MaterialAnimator::cacheAssets() {
    prepare(true);
    if (material().isSkyMasked() && !::devRendSkyMode) return;
    for (int i = 0; i < material().layerCount(); ++i) {
        if (world::TextureMaterialLayer const *layer = maybeAs<world::TextureMaterialLayer>(material().layer(i))) {
            for (int k = 0; k < layer->stageCount(); ++k) {
                auto const &stage = layer->stage(k);
                if (ClientTexture *tex = d->layers.at(i)->stageTextures.at(k)) {
                    if (is<world::DetailTextureMaterialLayer>(layer)) {
                        auto const &detailStage = stage.as<world::DetailTextureMaterialLayer::AnimationStage>();
                        float const contrast = de::clamp(0.f, detailStage.strength, 1.f) * detailFactor /*Global strength multiplier*/;
                        tex->prepareVariant(resSys().detailTextureSpec(contrast));
                    }
                    else if (is<world::ShineTextureMaterialLayer>(layer)) {
                        tex->prepareVariant(Rend_MapSurfaceShinyTextureSpec());
                        if (ClientTexture *maskTex = findTextureForAnimationStage(stage, MaskTextureProperty)) {
                            maskTex->prepareVariant(Rend_MapSurfaceShinyMaskTextureSpec());
                        }
                    }
                    else {
                        tex->prepareVariant(*variantSpec().primarySpec);
                    }
                }
            }
        }
    }
}
bool MaterialAnimator::isOpaque() const {
    d->updateSnapshotIfNeeded();
    return d->snapshot->opaque;
}
Vector2ui const &MaterialAnimator::dimensions() const {
    d->updateSnapshotIfNeeded();
    return d->snapshot->dimensions;
}
float MaterialAnimator::glowStrength() const {
    d->updateSnapshotIfNeeded();
    return d->snapshot->glowStrength;
}
blendmode_t MaterialAnimator::shineBlendMode() const {
    d->updateSnapshotIfNeeded();
    return d->snapshot->shineBlendMode;
}
Vector3f const &MaterialAnimator::shineMinColor() const {
    d->updateSnapshotIfNeeded();
    return d->snapshot->shineMinColor;
}
GLTextureUnit &MaterialAnimator::texUnit(int unitIndex) const {
    d->updateSnapshotIfNeeded();
    if (unitIndex >= 0 && unitIndex < NUM_TEXTUREUNITS) return d->snapshot->units[unitIndex];
    throw MissingTextureUnitError("MaterialAnimator::glTextureUnit", "Unknown GL texture unit #" + String::number(unitIndex));
}
MaterialAnimator::Decoration &MaterialAnimator::decoration(int decorIndex) const {
    d->updateSnapshotIfNeeded();
    if (decorIndex >= 0 && decorIndex < d->decorations.count()) return *d->decorations[decorIndex];
    throw MissingDecorationError("MaterialAnimator::decoration", "Unknown decoration #" + String::number(decorIndex));
}
variantspecification_t::variantspecification_t()
    : context(TC_UNKNOWN)
    , flags(0)
    , border(0)
    , wrapS(GL_REPEAT)
    , wrapT(GL_REPEAT)
    , mipmapped(false)
    , gammaCorrection(true)
    , noStretch(false)
    , toAlpha(false)
    , minFilter(GL_LINEAR)
    , magFilter(GL_LINEAR)
    , anisoFilter(0)
    , tClass(0)
    , tMap(0)
{}
variantspecification_t::variantspecification_t(variantspecification_t const &other)
    : context(other.context)
    , flags(other.flags)
    , border(other.border)
    , wrapS(other.wrapS)
    , wrapT(other.wrapT)
    , mipmapped(other.mipmapped)
    , gammaCorrection(other.gammaCorrection)
    , noStretch(other.noStretch)
    , toAlpha(other.toAlpha)
    , minFilter(other.minFilter)
    , magFilter(other.magFilter)
    , anisoFilter(other.anisoFilter)
    , tClass(other.tClass)
    , tMap(other.tMap)
{}
bool variantspecification_t::operator == (variantspecification_t const &other) const {
    if(this == &other) return 1;
    if(context != other.context) return 0;
    if(flags != other.flags) return 0;
    if(wrapS != other.wrapS || wrapT != other.wrapT) return 0;
    if(mipmapped != other.mipmapped) return 0;
    if(noStretch != other.noStretch) return 0;
    if(gammaCorrection != other.gammaCorrection) return 0;
    if(toAlpha != other.toAlpha) return 0;
    if(border != other.border) return 0;
    if(flags & TSF_HAS_COLORPALETTE_XLAT) {
        if(tClass != other.tClass) return 0;
        if(tMap   != other.tMap) return 0;
    }
    return 1; // Equal.
}
int variantspecification_t::glMinFilter() const {
    if(minFilter >= 0) // Constant logical value.
    {
        return (mipmapped? GL_NEAREST_MIPMAP_NEAREST : GL_NEAREST) + minFilter;
    }
    return mipmapped? glmode[mipmapping] : GL_LINEAR;
}
int variantspecification_t::glMagFilter() const {
    if(magFilter >= 0) // Constant logical value.
    {
        return GL_NEAREST + magFilter;
    }
    switch(abs(magFilter)-1) {
    case 1: // Sprite class.
        return filterSprites? GL_LINEAR : GL_NEAREST;
    case 2: // UI class.
        return filterUI? GL_LINEAR : GL_NEAREST;
    default: // "No class" preference.
        return glmode[texMagMode];
    }
}
int variantspecification_t::logicalAnisoLevel() const {
    return anisoFilter < 0? texAniso : anisoFilter;
}
TextureVariantSpec::TextureVariantSpec(texturevariantspecificationtype_t type)
    : type(type)
{}
TextureVariantSpec::TextureVariantSpec(TextureVariantSpec const &other)
    : type(other.type)
    , variant(other.variant)
    , detailVariant(other.detailVariant)
{}
bool detailvariantspecification_t::operator == (detailvariantspecification_t const &other) const {
    if(this == &other) return true;
    return contrast == other.contrast; // Equal.
}
bool TextureVariantSpec::operator == (TextureVariantSpec const &other) const {
    if(this == &other) return true; // trivial
    if(type != other.type) return false;
    switch(type) {
    case TST_GENERAL: return variant       == other.variant;
    case TST_DETAIL:  return detailVariant == other.detailVariant;
    }
    DENG2_ASSERT(false);
    return false;
}
static String nameForGLTextureWrapMode(int mode) {
    if(mode == GL_REPEAT) return "repeat";
#if defined (DENG_OPENGL)
    if(mode == GL_CLAMP) return "clamp";
#endif
    if(mode == GL_CLAMP_TO_EDGE) return "clamp_edge";
    return "(unknown)";
}
String TextureVariantSpec::asText() const {
    static String const textureUsageContextNames[1 + TEXTUREVARIANTUSAGECONTEXT_COUNT] = {
        /* TC_UNKNOWN */                    "unknown",
        /* TC_UI */                         "ui",
        /* TC_MAPSURFACE_DIFFUSE */         "mapsurface_diffuse",
        /* TC_MAPSURFACE_REFLECTION */      "mapsurface_reflection",
        /* TC_MAPSURFACE_REFLECTIONMASK */  "mapsurface_reflectionmask",
        /* TC_MAPSURFACE_LIGHTMAP */        "mapsurface_lightmap",
        /* TC_SPRITE_DIFFUSE */             "sprite_diffuse",
        /* TC_MODELSKIN_DIFFUSE */          "modelskin_diffuse",
        /* TC_MODELSKIN_REFLECTION */       "modelskin_reflection",
        /* TC_HALO_LUMINANCE */             "halo_luminance",
        /* TC_PSPRITE_DIFFUSE */            "psprite_diffuse",
        /* TC_SKYSPHERE_DIFFUSE */          "skysphere_diffuse"
    };
    static String const textureSpecificationTypeNames[2] = {
        /* TST_GENERAL */   "general",
        /* TST_DETAIL */    "detail"
    };
    static String const filterModeNames[] = { "ui", "sprite", "noclass", "const" };
    static String const glFilterNames[] = {
        "nearest", "linear", "nearest_mipmap_nearest", "linear_mipmap_nearest",
        "nearest_mipmap_linear", "linear_mipmap_linear"
    };
    String text = String("Type:%1").arg(textureSpecificationTypeNames[type]);
    switch(type) {
    case TST_DETAIL: {
        detailvariantspecification_t const &spec = detailVariant;
        text += " Contrast:" + String::number(int(.5f + spec.contrast / 255.f * 100)) + "%";
        break; }
    case TST_GENERAL: {
        variantspecification_t const &spec = variant;
        texturevariantusagecontext_t tc = spec.context;
        DENG2_ASSERT(tc == TC_UNKNOWN || VALID_TEXTUREVARIANTUSAGECONTEXT(tc));
        int glMinFilterNameIdx;
        if(spec.minFilter >= 0) // Constant logical value.
        {
            glMinFilterNameIdx = (spec.mipmapped? 2 : 0) + spec.minFilter;
        }
        else // "No class" preference.
        {
            glMinFilterNameIdx = spec.mipmapped? mipmapping : 1;
        }
        int glMagFilterNameIdx;
        if(spec.magFilter >= 0) // Constant logical value.
        {
            glMagFilterNameIdx = spec.magFilter;
        }
        else {
            switch(abs(spec.magFilter)-1) {
            default: glMagFilterNameIdx = texMagMode; break;
            case 1:  glMagFilterNameIdx = filterSprites; break;
            case 2:  glMagFilterNameIdx = filterUI; break;
            }
        }
        text += " Context:" + textureUsageContextNames[tc-TEXTUREVARIANTUSAGECONTEXT_FIRST + 1]
              + " Flags:" + String::number(spec.flags & ~TSF_INTERNAL_MASK)
              + " Border:" + String::number(spec.border)
              + " MinFilter:" + filterModeNames[3 + de::clamp(-1, spec.minFilter, 0)]
                              + "|" + glFilterNames[glMinFilterNameIdx]
              + " MagFilter:" + filterModeNames[3 + de::clamp(-3, spec.magFilter, 0)]
                              + "|" + glFilterNames[glMagFilterNameIdx]
              + " AnisoFilter:" + String::number(spec.anisoFilter)
              + " WrapS:" + nameForGLTextureWrapMode(spec.wrapS)
              + " WrapT:" + nameForGLTextureWrapMode(spec.wrapT)
              + " CorrectGamma:" + (spec.gammaCorrection? "yes" : "no")
              + " NoStretch:" + (spec.noStretch? "yes" : "no")
              + " ToAlpha:" + (spec.toAlpha? "yes" : "no");
        if(spec.flags & TSF_HAS_COLORPALETTE_XLAT) {
            text += " Translated:(tclass:" + String::number(spec.tClass)
                                           + " tmap:" + String::number(spec.tMap) + ")";
        }
        break; }
    }
    return text;
}
DENG2_PIMPL(ClientTexture::Variant) {
    ClientTexture &texture; /// The base for which "this" is a context derivative.
    TextureVariantSpec spec; /// Usage context specification.
    Flags flags;
    res::Source texSource; ///< Logical source of the image.
    uint glTexName;
    float s, t;
    Impl(Public *i, ClientTexture &generalCase, TextureVariantSpec const &spec)
        : Base(i)
        , texture(generalCase)
        , spec(spec)
        , flags(0)
        , texSource(res::None)
        , glTexName(0)
        , s(0)
        , t(0)
    {}
    ~Impl() {
        self().release();
    }
};
ClientTexture::Variant::Variant(ClientTexture &generalCase, TextureVariantSpec const &spec)
    : d(new Impl(this, generalCase, spec))
{}
/**
 * Perform analyses of the @a image pixel data and record this information
 * for reference later.
 *
 * @param image         Image data to be analyzed.
 * @param context       Context in which the uploaded image will be used.
 * @param tex           Logical texture which will hold the analysis data.
 * @param forceUpdate   Force an update of the recorded analysis data.
 */
static void performImageAnalyses(image_t const &image,
    texturevariantusagecontext_t context, ClientTexture &tex, bool forceUpdate) {
    if(image.paletteId != 0) {
        colorpalette_analysis_t *cp = reinterpret_cast<colorpalette_analysis_t *>(tex.analysisDataPointer(ClientTexture::ColorPaletteAnalysis));
        bool firstInit = (!cp);
        if(firstInit) {
            cp = (colorpalette_analysis_t *) M_Malloc(sizeof(*cp));
            tex.setAnalysisDataPointer(ClientTexture::ColorPaletteAnalysis, cp);
        }
        if(firstInit || forceUpdate)
            cp->paletteId = image.paletteId;
    }
    if(context == TC_SPRITE_DIFFUSE) {
        pointlight_analysis_t *pl = reinterpret_cast<pointlight_analysis_t*>(tex.analysisDataPointer(ClientTexture::BrightPointAnalysis));
        bool firstInit = (!pl);
        if(firstInit) {
            pl = (pointlight_analysis_t *) M_Malloc(sizeof *pl);
            tex.setAnalysisDataPointer(ClientTexture::BrightPointAnalysis, pl);
        }
        if(firstInit || forceUpdate) {
            GL_CalcLuminance(image.pixels, image.size.x, image.size.y,
                             image.pixelSize, image.paletteId,
                             &pl->originX, &pl->originY, &pl->color, &pl->brightMul);
        }
    }
    if(context == TC_SPRITE_DIFFUSE || context == TC_UI) {
        averagealpha_analysis_t *aa = reinterpret_cast<averagealpha_analysis_t*>(tex.analysisDataPointer(ClientTexture::AverageAlphaAnalysis));
        bool firstInit = (!aa);
        if(firstInit) {
            aa = (averagealpha_analysis_t *) M_Malloc(sizeof(*aa));
            tex.setAnalysisDataPointer(ClientTexture::AverageAlphaAnalysis, aa);
        }
        if(firstInit || forceUpdate) {
            if(!image.paletteId) {
                FindAverageAlpha(image.pixels, image.size.x, image.size.y,
                                 image.pixelSize, &aa->alpha, &aa->coverage);
            }
            else {
                if(image.flags & IMGF_IS_MASKED) {
                    FindAverageAlphaIdx(image.pixels, image.size.x, image.size.y,
                                        &aa->alpha, &aa->coverage);
                }
                else {
                    aa->alpha = 1;
                    aa->coverage = 0;
                }
            }
        }
    }
    if(context == TC_SKYSPHERE_DIFFUSE) {
        averagecolor_analysis_t *ac = reinterpret_cast<averagecolor_analysis_t *>(tex.analysisDataPointer(ClientTexture::AverageColorAnalysis));
        bool firstInit = (!ac);
        if(firstInit) {
            ac = (averagecolor_analysis_t *) M_Malloc(sizeof(*ac));
            tex.setAnalysisDataPointer(ClientTexture::AverageColorAnalysis, ac);
        }
        if(firstInit || forceUpdate) {
            if(0 == image.paletteId) {
                FindAverageColor(image.pixels, image.size.x, image.size.y,
                                 image.pixelSize, &ac->color);
            }
            else {
                FindAverageColorIdx(image.pixels, image.size.x, image.size.y,
                                    App_Resources().colorPalettes().colorPalette(image.paletteId),
                                    false, &ac->color);
            }
        }
    }
    if(context == TC_MAPSURFACE_DIFFUSE) {
        averagecolor_analysis_t *ac = reinterpret_cast<averagecolor_analysis_t *>(tex.analysisDataPointer(ClientTexture::AverageColorAmplifiedAnalysis));
        bool firstInit = (!ac);
        if(firstInit) {
            ac = (averagecolor_analysis_t *) M_Malloc(sizeof(*ac));
            tex.setAnalysisDataPointer(ClientTexture::AverageColorAmplifiedAnalysis, ac);
        }
        if(firstInit || forceUpdate) {
            if(0 == image.paletteId) {
                FindAverageColor(image.pixels, image.size.x, image.size.y,
                                 image.pixelSize, &ac->color);
            }
            else {
                FindAverageColorIdx(image.pixels, image.size.x, image.size.y,
                                    App_Resources().colorPalettes().colorPalette(image.paletteId),
                                    false, &ac->color);
            }
            Vector3f color(ac->color.rgb);
            R_AmplifyColor(color);
            for(int i = 0; i < 3; ++i) {
                ac->color.rgb[i] = color[i];
            }
        }
    }
    if(context == TC_SKYSPHERE_DIFFUSE) {
        averagecolor_analysis_t *ac = reinterpret_cast<averagecolor_analysis_t *>(tex.analysisDataPointer(ClientTexture::AverageTopColorAnalysis));
        bool firstInit = (!ac);
        if(firstInit) {
            ac = (averagecolor_analysis_t *) M_Malloc(sizeof(*ac));
            tex.setAnalysisDataPointer(ClientTexture::AverageTopColorAnalysis, ac);
        }
        if(firstInit || forceUpdate) {
            if(0 == image.paletteId) {
                FindAverageLineColor(image.pixels, image.size.x, image.size.y,
                                     image.pixelSize, 0, &ac->color);
            }
            else {
                FindAverageLineColorIdx(image.pixels, image.size.x, image.size.y, 0,
                                        App_Resources().colorPalettes().colorPalette(image.paletteId),
                                        false, &ac->color);
            }
        }
    }
    if(context == TC_SKYSPHERE_DIFFUSE) {
        averagecolor_analysis_t *ac = reinterpret_cast<averagecolor_analysis_t *>(tex.analysisDataPointer(ClientTexture::AverageBottomColorAnalysis));
        bool firstInit = (!ac);
        if(firstInit) {
            ac = (averagecolor_analysis_t *) M_Malloc(sizeof(*ac));
            tex.setAnalysisDataPointer(ClientTexture::AverageBottomColorAnalysis, ac);
        }
        if(firstInit || forceUpdate) {
            if(0 == image.paletteId) {
                FindAverageLineColor(image.pixels, image.size.x, image.size.y,
                                     image.pixelSize, image.size.y - 1, &ac->color);
            }
            else {
                FindAverageLineColorIdx(image.pixels, image.size.x, image.size.y,
                                        image.size.y - 1,
                                        App_Resources().colorPalettes().colorPalette(image.paletteId),
                                        false, &ac->color);
            }
        }
    }
}
uint ClientTexture::Variant::prepare() {
    if(isPrepared())
        return d->glTexName;
    LOG_AS("TextureVariant::prepare");
    image_t image;
    res::Source source = GL_LoadSourceImage(image, d->texture, d->spec);
    if(source == res::None)
        return 0;
    if(d->spec.type == TST_GENERAL) {
        performImageAnalyses(image, d->spec.variant.context, d->texture,
                             true /*force update*/);
    }
    if(d->glTexName == 0) {
        d->glTexName = GL_GetReservedTextureName();
        d->texSource = source;
    }
    texturecontent_t c;
    GL_PrepareTextureContent(c, d->glTexName, image, d->spec, d->texture.manifest());
    /**
     * Calculate GL texture coordinates based on the image dimensions. The
     * coordinates are calculated as width / CeilPow2(width), or 1 if larger
     * than the maximum texture size.
     *
     * @todo fixme: Image dimensions may not be the same as the uploaded
     * texture - defer this logic until all processing has been completed.
     */
    if ((c.flags & TXCF_UPLOAD_ARG_NOSTRETCH) &&
        (c.flags & TXCF_MIPMAP)) {
        d->s = image.size.x / float( de::ceilPow2(image.size.x) );
        d->t = image.size.y / float( de::ceilPow2(image.size.y) );
    }
    else {
        d->s = 1;
        d->t = 1;
    }
    if(image.flags & IMGF_IS_MASKED) {
        d->flags |= TextureVariant::Masked;
    }
    gl::UploadMethod uploadMethod = GL_ChooseUploadMethod(&c);
    GL_UploadTextureContent(c, uploadMethod);
    LOGDEV_RES_XVERBOSE("Prepared \"%s\" variant (glName:%u)%s",
                        d->texture.manifest().composeUri() << uint(d->glTexName) <<
                        (uploadMethod == gl::Immediate? " while not busy!" : ""));
    LOGDEV_RES_XVERBOSE("  Content: %s", Image_Description(image));
    LOGDEV_RES_XVERBOSE("  Specification %p: %s", &d->spec << d->spec.asText());
    if(d->texture.width() == 0 && d->texture.height() == 0) {
        LOG_RES_XVERBOSE("World dimensions for \"%s\" taken from image pixels %s",
                         d->texture.manifest().composeUri() << image.size.asText());
        d->texture.setDimensions(image.size);
    }
    Image_ClearPixelData(image);
    return d->glTexName;
}
void ClientTexture::Variant::release() {
    if (isPrepared()) {
        Deferred_glDeleteTextures(1, (GLuint const *) &d->glTexName);
        d->glTexName = 0;
    }
}
ClientTexture &ClientTexture::Variant::base() const {
    return d->texture;
}
TextureVariantSpec const &ClientTexture::Variant::spec() const {
    return d->spec;
}
res::Source ClientTexture::Variant::source() const {
    return d->texSource;
}
String ClientTexture::Variant::sourceDescription() const {
    if(d->texSource == res::Original) return "original";
    if(d->texSource == res::External) return "external";
    return "none";
}
ClientTexture::Variant::Flags ClientTexture::Variant::flags() const {
    return d->flags;
}
void ClientTexture::Variant::glCoords(float *outS, float *outT) const {
    if(outS) *outS = d->s;
    if(outT) *outT = d->t;
}
uint ClientTexture::Variant::glName() const {
    return d->glTexName;
}
uint8_t *TGA_Load(FileHandle &file, Vector2ui &outSize, int &pixelSize) {
    const size_t initPos = file.tell();
    Block fileContents(file.length() - initPos);
    file.read(fileContents.data(), fileContents.size());
    file.seek(initPos, SeekSet);
    Image img = Image::fromData(fileContents, ".tga");
    if (img.isNull()) {
        return nullptr;
    }
    if (img.toQImage().format() == QImage::Format_Indexed8) {
        img = img.toQImage().convertToFormat(QImage::Format_RGBA8888);
    }
    outSize   = img.size();
    pixelSize = img.depth() / 8;
    const size_t numBytes = size_t(img.byteCount());
    uint8_t *    retBuf   = reinterpret_cast<uint8_t *>(M_Malloc(numBytes));
    memcpy(retBuf, img.bits(), numBytes);
    return retBuf;
}
