﻿/** @file alertmask.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/alertmask.h"
#include "clientapp.h"
#include <de/Config>
/** @file axisinputcontrol.cpp  Axis control for a logical input device.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/axisinputcontrol.h"
#include <de/smoother.h>
#include <de/timer.h> // SECONDSPERTIC
#include <de/Block>
#include <doomsday/console/var.h>
#include "ui/joystick.h"
#include "dd_loop.h" // DD_LatestRunTicsStartTime()
/** @file b_main.cpp  Event and device state bindings system.
 *
 * @todo Pretty much everything in this file belongs in InputSystem.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/b_main.h"
#include "ui/inputsystem.h"
#include "dd_main.h" // App_GameLoaded
#include "dd_def.h"
#include "clientapp.h"
#include "BindContext"
/** @file b_util.cpp  Input system, binding utilities.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/b_util.h"
#include <de/timer.h>
#include <de/input/ddkey.h>
#include <de/RecordValue>
#include "clientapp.h"
#include "BindContext"
#include "CommandBinding"
#include "ImpulseBinding"
#include "ui/inputsystem.h"
#include "ui/inputdevice.h"
#include "ui/axisinputcontrol.h"
#include "ui/buttoninputcontrol.h"
#include "ui/hatinputcontrol.h"
#include "network/net_main.h" // netGame
/** @file bindcontext.cpp  Input system, binding context.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/bindcontext.h"
#include <QList>
#include <QSet>
#include <QtAlgorithms>
#include <de/Log>
#include <doomsday/console/exec.h>
#include "clientapp.h"
#include "world/p_players.h"
#include "CommandBinding"
#include "ImpulseBinding"
#include "ui/inputdevice.h"
#include "ui/inputsystem.h"
/** @file binding.h  Base class for binding record accessors.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/binding.h"
#include <de/RecordValue>
/** @file busyvisual.cpp  Busy Mode visualizer.
 *
 * @authors Copyright © 2007-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "ui/busyvisual.h"
#include <cmath>
#include <de/concurrency.h>
#include <de/timer.h>
#include <de/GLState>
#include <de/GLInfo>
#include <doomsday/doomsdayapp.h>
#include <doomsday/console/var.h>
#include "gl/gl_main.h"
#include "ui/widgets/busywidget.h"
/** @file buttoninputcontrol.cpp  Button control for a logical input device.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/buttoninputcontrol.h"
#include <de/timer.h> // Timer_RealMilliseconds()
/** @file clientrootwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/clientrootwidget.h"
#include "ui/clientwindow.h"
#include "ui/clientwindowsystem.h"
#include "ui/inputsystem.h"
#include "clientapp.h"
/** @file clientstyle.cpp  Client UI style.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/clientstyle.h"
#include "ui/clientwindow.h"
#include <de/Async>
/** @file clientwindow.cpp  Top-level window with UI widgets.
 * @ingroup base
 *
 * @todo Platform-specific behavior should be encapsulated in subclasses, e.g.,
 * MacWindowBehavior. This would make the code easier to follow and more adaptable
 * to the quirks of each platform.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2008 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/clientwindow.h"
#include "ui/clientrootwidget.h"
#include "ui/viewcompositor.h"
#include "clientapp.h"
#include "clientplayer.h"
#include <QSurfaceFormat>
#include <QTimer>
#include <QCloseEvent>
#include <de/CompositorWidget>
#include <de/Config>
#include <de/ConstantRule>
#include <de/DisplayMode>
#include <de/Drawable>
#include <de/FadeToBlackWidget>
#include <de/FileSystem>
#include <de/GLTextureFramebuffer>
#include <de/GLState>
#include <de/GLInfo>
#include <de/LogBuffer>
#include <de/NotificationAreaWidget>
#include <de/NumberValue>
#include <de/SignalAction>
#include <de/VRWindowTransform>
#include <de/concurrency.h>
#include <doomsday/console/exec.h>
#include "api_console.h"
#include "gl/sys_opengl.h"
#include "gl/gl_main.h"
#include "ui/widgets/gamewidget.h"
#include "ui/widgets/busywidget.h"
#include "ui/widgets/taskbarwidget.h"
#include "ui/widgets/consolewidget.h"
#include "ui/widgets/privilegedlogwidget.h"
#include "ui/home/homewidget.h"
#include "ui/dialogs/coloradjustmentdialog.h"
#include "ui/dialogs/alertdialog.h"
#include "ui/inputdevice.h"
#include "ui/inputsystem.h"
#include "ui/clientwindowsystem.h"
#include "CommandAction"
#include "ui/mouse_qt.h"
#include "dd_main.h"
#include "render/vr.h"
/** @file clientwindowsystem.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "ui/clientwindowsystem.h"
#include "ui/clientstyle.h"
#include "clientapp.h"
#include "dd_main.h"
#include "gl/gl_main.h"
#include <de/PackageLoader>
/** @file commandaction.cpp  Action that executes a console command.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_platform.h"
#include "CommandAction"
#include <doomsday/console/exec.h>
/** @file commandbinding.cpp  Command binding record accessor.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/commandbinding.h"
#include <de/str.hh>
#include <de/Block>
#include <de/Log>
#include <de/RecordValue>
#include "CommandAction"
#include "clientapp.h"
#include "world/p_players.h" // P_ConsoleToLocal
#include "ui/inputsystem.h"
#include "ui/b_util.h"
#include "ui/axisinputcontrol.h"
#include "ui/buttoninputcontrol.h"
#include "ui/hatinputcontrol.h"
/** @file controllerpresets.cpp  Game controller presets.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/controllerpresets.h"
#include "ui/joystick.h"
#include "ui/inputsystem.h"
#include "clientapp.h"
#include <de/App>
#include <de/DictionaryValue>
#include <de/Process>
#include <de/RecordValue>
#include <de/Script>
#include <de/ScriptSystem>
#include <doomsday/console/var.h>
#include <QRegExp>
/** @file hatinputcontrol.cpp  Hat control for a logical input device.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/hatinputcontrol.h"
#include <de/timer.h> // Timer_RealMilliseconds()
/** @file impulsebinding.cpp  Impulse binding record accessor.
 *
 * @authors Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2007-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "dd_share.h" // DDMAXPLAYERS
#include "ui/impulsebinding.h"
#include <de/str.hh>
#include <de/Log>
#include <de/RecordValue>
#include "ui/b_util.h"
/** @file inputdebug.cpp  Input debug visualizer.
 *
 * @authors Copyright © 2007-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h" // strdup macro
#include "ui/inputdebug.h"
#ifdef DENG2_DEBUG
#include <de/concurrency.h>
#include <de/ddstring.h>
#include <de/point.h>
#include <de/timer.h> // SECONDSPERTIC
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include <de/KeyEvent>
#include <de/GLInfo>
#include "clientapp.h"
#include "dd_def.h"
#include "dd_main.h"
#include "sys_system.h" // novideo
#include "gl/gl_main.h"
#include "gl/gl_draw.h"
#include "api_fontrender.h"
#include "ui/b_main.h"
#include "ui/b_util.h"
#include "ui/clientwindow.h"
#include "ui/joystick.h"
#include "ui/infine/finale.h"
#include "ui/inputsystem.h"
#include "ui/inputdevice.h"
#include "ui/axisinputcontrol.h"
#include "ui/buttoninputcontrol.h"
#include "ui/hatinputcontrol.h"
#include "ui/sys_input.h"
#include "ui/ui_main.h"
using namespace de;
static byte devRendJoyState;
static byte devRendKeyState;
static byte devRendMouseState;
#define MAX_KEYMAPPINGS  256
static uchar shiftKeyMappings[MAX_KEYMAPPINGS];
static void initKeyMappingsOnce()
{
    if (shiftKeyMappings[1] == 1) return;
    uchar defaultShiftTable[96] = // Contains characters 32 to 127.
    {
    /* 32 */    ' ', 0, 0, 0, 0, 0, 0, '"',
    /* 40 */    0, 0, 0, 0, '<', '_', '>', '?', ')', '!',
    /* 50 */    '@', '#', '$', '%', '^', '&', '*', '(', 0, ':',
    /* 60 */    0, '+', 0, 0, 0, 'a', 'b', 'c', 'd', 'e',
    /* 70 */    'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    /* 80 */    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y',
    /* 90 */    'z', '{', '|', '}', 0, 0, '~', 'A', 'B', 'C',
    /* 100 */   'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
    /* 110 */   'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
    /* 120 */   'X', 'Y', 'Z', 0, 0, 0, 0, 0
    };
    for (int i = 0; i < 256; ++i)
    {
        if (i >= 32 && i <= 127)
            shiftKeyMappings[i] = defaultShiftTable[i - 32] ? defaultShiftTable[i - 32] : i;
        else
            shiftKeyMappings[i] = i;
    }
}
static void initDrawStateForVisual(Point2Raw const *origin)
{
    FR_PushAttrib();
    if (origin && !(origin->x == 0 && origin->y == 0))
    {
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(origin->x, origin->y, 0);
    }
}
static void endDrawStateForVisual(Point2Raw const *origin)
{
    if (origin && !(origin->x == 0 && origin->y == 0))
    {
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
    }
    FR_PopAttrib();
}
/**
 * Apply all active modifiers to the key.
 */
static uchar modKey(int ddkey)
{
    extern bool shiftDown;
    initKeyMappingsOnce();
    if (shiftDown)
    {
        DENG2_ASSERT(ddkey >= 0 && ddkey < MAX_KEYMAPPINGS);
        ddkey = shiftKeyMappings[ddkey];
    }
    if (ddkey >= DDKEY_NUMPAD7 && ddkey <= DDKEY_NUMPAD0)
    {
        static uchar const numPadKeys[10] = { '7', '8', '9', '4', '5', '6', '1', '2', '3', '0' };
        return numPadKeys[ddkey - DDKEY_NUMPAD7];
    }
    else if (ddkey == DDKEY_DIVIDE)
    {
        return '/';
    }
    else if (ddkey == DDKEY_SUBTRACT)
    {
        return '-';
    }
    else if (ddkey == DDKEY_ADD)
    {
        return '+';
    }
    else if (ddkey == DDKEY_DECIMAL)
    {
        return '.';
    }
    else if (ddkey == DDKEY_MULTIPLY)
    {
        return '*';
    }
    return uchar(ddkey);
}
void Rend_RenderButtonStateVisual(InputDevice &device, int buttonID, Point2Raw const *_origin,
    RectRaw *geometry)
{
#define BORDER 4
    float const upColor[] = { .3f, .3f, .3f, .6f };
    float const downColor[] = { .3f, .3f, 1, .6f };
    float const expiredMarkColor[] = { 1, 0, 0, 1 };
    float const triggeredMarkColor[] = { 1, 0, 1, 1 };
    if (geometry)
    {
        geometry->origin.x = geometry->origin.y = 0;
        geometry->size.width = geometry->size.height = 0;
    }
    ButtonInputControl const &button = device.button(buttonID);
    Point2Raw origin;
    origin.x = _origin? _origin->x : 0;
    origin.y = _origin? _origin->y : 0;
    String label;
    if (!button.name().isEmpty())
    {
        label = button.name();
    }
    else if (&device == ClientApp::inputSystem().devicePtr(IDEV_KEYBOARD))
    {
        uchar asciiCode = modKey(buttonID);
        if (asciiCode > 32 && asciiCode < 127)
        {
            label = String("%1").arg(QChar(asciiCode));
        }
        if (label.isEmpty())
        {
            label = B_ShortNameForKey(buttonID, false/*do not force lowercase*/);
        }
    }
    if (label.isEmpty())
    {
        label = String("#%1").arg(buttonID, 3, 10, QChar('0'));
    }
    initDrawStateForVisual(&origin);
    Size2Raw textSize;
    FR_TextSize(&textSize, label.toUtf8().constData());
    Rectanglei textGeom = Rectanglei::fromSize(Vector2i(0, 0),
                                               Vector2ui(textSize.width  + BORDER * 2,
                                                         textSize.height + BORDER * 2));
    DGL_Color4fv(button.isDown()? downColor : upColor);
    GL_DrawRect(textGeom);
    DGL_Enable(DGL_TEXTURE_2D);
    Point2Raw const textOffset = {{{BORDER, BORDER}}};
    FR_DrawText(label.toUtf8().constData(), &textOffset);
    DGL_Disable(DGL_TEXTURE_2D);
    if (button.bindContextAssociation() & InputControl::Expired)
    {
        int const markSize = .5f + de::min(textGeom.width(), textGeom.height()) / 3.f;
        DGL_Color3fv(expiredMarkColor);
        DGL_Begin(DGL_TRIANGLES);
        DGL_Vertex2f(textGeom.width(), 0);
        DGL_Vertex2f(textGeom.width(), markSize);
        DGL_Vertex2f(textGeom.width() - markSize, 0);
        DGL_End();
    }
    if (button.bindContextAssociation() & InputControl::Triggered)
    {
        int const markSize = .5f + de::min(textGeom.width(), textGeom.height()) / 3.f;
        DGL_Color3fv(triggeredMarkColor);
        DGL_Begin(DGL_TRIANGLES);
        DGL_Vertex2f(0, 0);
        DGL_Vertex2f(markSize, 0);
        DGL_Vertex2f(0, markSize);
        DGL_End();
    }
    endDrawStateForVisual(&origin);
    if (geometry)
    {
        std::memcpy(&geometry->origin, &origin, sizeof(geometry->origin));
        geometry->size.width  = textGeom.width();
        geometry->size.height = textGeom.height();
    }
#undef BORDER
}
void Rend_RenderAxisStateVisual(InputDevice & /*device*/, int /*axisID*/, Point2Raw const *origin,
    RectRaw *geometry)
{
    if (geometry)
    {
        geometry->origin.x = geometry->origin.y = 0;
        geometry->size.width = geometry->size.height = 0;
    }
    initDrawStateForVisual(origin);
    endDrawStateForVisual(origin);
}
void Rend_RenderHatStateVisual(InputDevice & /*device*/, int /*hatID*/, Point2Raw const *origin,
    RectRaw *geometry)
{
    if (geometry)
    {
        geometry->origin.x = geometry->origin.y = 0;
        geometry->size.width = geometry->size.height = 0;
    }
    initDrawStateForVisual(origin);
    endDrawStateForVisual(origin);
}
enum inputdev_controltype_t
{
    IDC_KEY,
    IDC_AXIS,
    IDC_HAT,
    NUM_INPUT_DEVICE_CONTROL_TYPES
};
struct inputdev_layout_control_t
{
    inputdev_controltype_t type;
    uint id;
};
struct inputdev_layout_controlgroup_t
{
    inputdev_layout_control_t *controls;
    uint numControls;
};
struct inputdev_layout_t
{
    inputdev_layout_controlgroup_t *groups;
    uint numGroups;
};
static void drawControlGroup(InputDevice &device, inputdev_layout_controlgroup_t const *group,
    Point2Raw *_origin, RectRaw *retGeometry)
{
#define SPACING  2
    if (retGeometry)
    {
        retGeometry->origin.x = retGeometry->origin.y = 0;
        retGeometry->size.width = retGeometry->size.height = 0;
    }
    if (!group) return;
    Point2Raw origin;
    origin.x = _origin? _origin->x : 0;
    origin.y = _origin? _origin->y : 0;
    Rect *grpGeom = nullptr;
    RectRaw ctrlGeom{};
    for (uint i = 0; i < group->numControls; ++i)
    {
        inputdev_layout_control_t const *ctrl = group->controls + i;
        switch (ctrl->type)
        {
        case IDC_AXIS: Rend_RenderAxisStateVisual(device, ctrl->id, &origin, &ctrlGeom); break;
        case IDC_KEY:  Rend_RenderButtonStateVisual(device, ctrl->id, &origin, &ctrlGeom); break;
        case IDC_HAT:  Rend_RenderHatStateVisual(device, ctrl->id, &origin, &ctrlGeom); break;
        default:
            App_Error("drawControlGroup: Unknown inputdev_controltype_t: %i.", (int)ctrl->type);
            exit(1); // Unreachable.
        }
        if (ctrlGeom.size.width > 0 && ctrlGeom.size.height > 0)
        {
            origin.x += ctrlGeom.size.width + SPACING;
            if (grpGeom)
                Rect_UniteRaw(grpGeom, &ctrlGeom);
            else
                grpGeom = Rect_NewFromRaw(&ctrlGeom);
        }
    }
    if (grpGeom)
    {
        if (retGeometry)
        {
            Rect_Raw(grpGeom, retGeometry);
        }
        Rect_Delete(grpGeom);
    }
#undef SPACING
}
/**
 * Render a visual representation of the current state of the specified device.
 */
void Rend_RenderInputDeviceStateVisual(InputDevice &device, inputdev_layout_t const *layout,
    Point2Raw const *origin, Size2Raw *retVisualDimensions)
{
#define SPACING  2
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (retVisualDimensions)
    {
        retVisualDimensions->width = retVisualDimensions->height = 0;
    }
    if (novideo || isDedicated) return; // Not for us.
    if (!layout) return;
    FR_SetFont(fontFixed);
    FR_PushAttrib();
    FR_LoadDefaultAttrib();
    FR_SetLeading(0);
    FR_SetColorAndAlpha(1, 1, 1, 1);
    initDrawStateForVisual(origin);
    Point2Raw offset{};
    Rect *visualGeom = nullptr;
    if (!device.title().isEmpty())
    {
        Size2Raw size;
        DGL_Enable(DGL_TEXTURE_2D);
        Block const fullName(device.title().toUtf8());
        FR_DrawText(fullName.constData(), nullptr/*no offset*/);
        DGL_Disable(DGL_TEXTURE_2D);
        FR_TextSize(&size, fullName.constData());
        visualGeom = Rect_NewWithOriginSize2(offset.x, offset.y, size.width, size.height);
        offset.y += size.height + SPACING;
    }
    for (uint i = 0; i < layout->numGroups; ++i)
    {
        inputdev_layout_controlgroup_t const *grp = &layout->groups[i];
        RectRaw grpGeometry{};
        drawControlGroup(device, grp, &offset, &grpGeometry);
        if (grpGeometry.size.width > 0 && grpGeometry.size.height > 0)
        {
            if (visualGeom)
                Rect_UniteRaw(visualGeom, &grpGeometry);
            else
                visualGeom = Rect_NewFromRaw(&grpGeometry);
            offset.y = Rect_Y(visualGeom) + Rect_Height(visualGeom) + SPACING;
        }
    }
    endDrawStateForVisual(origin);
    FR_PopAttrib();
    if (visualGeom && retVisualDimensions)
    {
        retVisualDimensions->width  = Rect_Width(visualGeom);
        retVisualDimensions->height = Rect_Height(visualGeom);
    }
#undef SPACING
}
void I_DebugDrawer()
{
#define SPACING      2
#define NUMITEMS(x)  (sizeof(x) / sizeof((x)[0]))
    static inputdev_layout_control_t keyGroup1[] = {
        { IDC_KEY,  27 }, // escape
        { IDC_KEY, 132 }, // f1
        { IDC_KEY, 133 }, // f2
        { IDC_KEY, 134 }, // f3
        { IDC_KEY, 135 }, // f4
        { IDC_KEY, 136 }, // f5
        { IDC_KEY, 137 }, // f6
        { IDC_KEY, 138 }, // f7
        { IDC_KEY, 139 }, // f8
        { IDC_KEY, 140 }, // f9
        { IDC_KEY, 141 }, // f10
        { IDC_KEY, 142 }, // f11
        { IDC_KEY, 143 }  // f12
    };
    static inputdev_layout_control_t keyGroup2[] = {
        { IDC_KEY,  96 }, // tilde
        { IDC_KEY,  49 }, // 1
        { IDC_KEY,  50 }, // 2
        { IDC_KEY,  51 }, // 3
        { IDC_KEY,  52 }, // 4
        { IDC_KEY,  53 }, // 5
        { IDC_KEY,  54 }, // 6
        { IDC_KEY,  55 }, // 7
        { IDC_KEY,  56 }, // 8
        { IDC_KEY,  57 }, // 9
        { IDC_KEY,  48 }, // 0
        { IDC_KEY,  45 }, // -
        { IDC_KEY,  61 }, // =
        { IDC_KEY, 127 }  // backspace
    };
    static inputdev_layout_control_t keyGroup3[] = {
        { IDC_KEY,   9 }, // tab
        { IDC_KEY, 113 }, // q
        { IDC_KEY, 119 }, // w
        { IDC_KEY, 101 }, // e
        { IDC_KEY, 114 }, // r
        { IDC_KEY, 116 }, // t
        { IDC_KEY, 121 }, // y
        { IDC_KEY, 117 }, // u
        { IDC_KEY, 105 }, // i
        { IDC_KEY, 111 }, // o
        { IDC_KEY, 112 }, // p
        { IDC_KEY,  91 }, // {
        { IDC_KEY,  93 }, // }
        { IDC_KEY,  92 }, // bslash
    };
    static inputdev_layout_control_t keyGroup4[] = {
        { IDC_KEY, 145 }, // capslock
        { IDC_KEY,  97 }, // a
        { IDC_KEY, 115 }, // s
        { IDC_KEY, 100 }, // d
        { IDC_KEY, 102 }, // f
        { IDC_KEY, 103 }, // g
        { IDC_KEY, 104 }, // h
        { IDC_KEY, 106 }, // j
        { IDC_KEY, 107 }, // k
        { IDC_KEY, 108 }, // l
        { IDC_KEY,  59 }, // semicolon
        { IDC_KEY,  39 }, // apostrophe
        { IDC_KEY,  13 }  // return
    };
    static inputdev_layout_control_t keyGroup5[] = {
        { IDC_KEY, 159 }, // shift
        { IDC_KEY, 122 }, // z
        { IDC_KEY, 120 }, // x
        { IDC_KEY,  99 }, // c
        { IDC_KEY, 118 }, // v
        { IDC_KEY,  98 }, // b
        { IDC_KEY, 110 }, // n
        { IDC_KEY, 109 }, // m
        { IDC_KEY,  44 }, // comma
        { IDC_KEY,  46 }, // period
        { IDC_KEY,  47 }, // fslash
        { IDC_KEY, 159 }, // shift
    };
    static inputdev_layout_control_t keyGroup6[] = {
        { IDC_KEY, 160 }, // ctrl
        { IDC_KEY,   0 }, // ???
        { IDC_KEY, 161 }, // alt
        { IDC_KEY,  32 }, // space
        { IDC_KEY, 161 }, // alt
        { IDC_KEY,   0 }, // ???
        { IDC_KEY,   0 }, // ???
        { IDC_KEY, 160 }  // ctrl
    };
    static inputdev_layout_control_t keyGroup7[] = {
        { IDC_KEY, 170 }, // printscrn
        { IDC_KEY, 146 }, // scrolllock
        { IDC_KEY, 158 }  // pause
    };
    static inputdev_layout_control_t keyGroup8[] = {
        { IDC_KEY, 162 }, // insert
        { IDC_KEY, 166 }, // home
        { IDC_KEY, 164 }  // pageup
    };
    static inputdev_layout_control_t keyGroup9[] = {
        { IDC_KEY, 163 }, // delete
        { IDC_KEY, 167 }, // end
        { IDC_KEY, 165 }, // pagedown
    };
    static inputdev_layout_control_t keyGroup10[] = {
        { IDC_KEY, 130 }, // up
        { IDC_KEY, 129 }, // left
        { IDC_KEY, 131 }, // down
        { IDC_KEY, 128 }, // right
    };
    static inputdev_layout_control_t keyGroup11[] = {
        { IDC_KEY, 144 }, // numlock
        { IDC_KEY, 172 }, // divide
        { IDC_KEY, DDKEY_MULTIPLY }, // multiply
        { IDC_KEY, 168 }  // subtract
    };
    static inputdev_layout_control_t keyGroup12[] = {
        { IDC_KEY, 147 }, // pad 7
        { IDC_KEY, 148 }, // pad 8
        { IDC_KEY, 149 }, // pad 9
        { IDC_KEY, 169 }, // add
    };
    static inputdev_layout_control_t keyGroup13[] = {
        { IDC_KEY, 150 }, // pad 4
        { IDC_KEY, 151 }, // pad 5
        { IDC_KEY, 152 }  // pad 6
    };
    static inputdev_layout_control_t keyGroup14[] = {
        { IDC_KEY, 153 }, // pad 1
        { IDC_KEY, 154 }, // pad 2
        { IDC_KEY, 155 }, // pad 3
        { IDC_KEY, 171 }  // enter
    };
    static inputdev_layout_control_t keyGroup15[] = {
        { IDC_KEY, 156 }, // pad 0
        { IDC_KEY, 157 }  // pad .
    };
    static inputdev_layout_controlgroup_t keyGroups[] = {
        { keyGroup1, NUMITEMS(keyGroup1) },
        { keyGroup2, NUMITEMS(keyGroup2) },
        { keyGroup3, NUMITEMS(keyGroup3) },
        { keyGroup4, NUMITEMS(keyGroup4) },
        { keyGroup5, NUMITEMS(keyGroup5) },
        { keyGroup6, NUMITEMS(keyGroup6) },
        { keyGroup7, NUMITEMS(keyGroup7) },
        { keyGroup8, NUMITEMS(keyGroup8) },
        { keyGroup9, NUMITEMS(keyGroup9) },
        { keyGroup10, NUMITEMS(keyGroup10) },
        { keyGroup11, NUMITEMS(keyGroup11) },
        { keyGroup12, NUMITEMS(keyGroup12) },
        { keyGroup13, NUMITEMS(keyGroup13) },
        { keyGroup14, NUMITEMS(keyGroup14) },
        { keyGroup15, NUMITEMS(keyGroup15) }
    };
    static inputdev_layout_t keyLayout = { keyGroups, NUMITEMS(keyGroups) };
    static inputdev_layout_control_t mouseGroup1[] = {
        { IDC_AXIS, 0 }, // x-axis
        { IDC_AXIS, 1 }, // y-axis
    };
    static inputdev_layout_control_t mouseGroup2[] = {
        { IDC_KEY, 0 }, // left
        { IDC_KEY, 1 }, // middle
        { IDC_KEY, 2 }  // right
    };
    static inputdev_layout_control_t mouseGroup3[] = {
        { IDC_KEY, 3 }, // wheelup
        { IDC_KEY, 4 }  // wheeldown
    };
    static inputdev_layout_control_t mouseGroup4[] = {
        { IDC_KEY,  5 },
        { IDC_KEY,  6 },
        { IDC_KEY,  7 },
        { IDC_KEY,  8 },
        { IDC_KEY,  9 },
        { IDC_KEY, 10 },
        { IDC_KEY, 11 },
        { IDC_KEY, 12 },
        { IDC_KEY, 13 },
        { IDC_KEY, 14 },
        { IDC_KEY, 15 }
    };
    static inputdev_layout_controlgroup_t mouseGroups[] = {
        { mouseGroup1, NUMITEMS(mouseGroup1) },
        { mouseGroup2, NUMITEMS(mouseGroup2) },
        { mouseGroup3, NUMITEMS(mouseGroup3) },
        { mouseGroup4, NUMITEMS(mouseGroup4) }
    };
    static inputdev_layout_t mouseLayout = { mouseGroups, NUMITEMS(mouseGroups) };
    static inputdev_layout_control_t joyGroup1[] = {
        { IDC_AXIS, 0 }, // x-axis
        { IDC_AXIS, 1 }, // y-axis
        { IDC_AXIS, 2 }, // z-axis
        { IDC_AXIS, 3 }  // w-axis
    };
    static inputdev_layout_control_t joyGroup2[] = {
        { IDC_AXIS,  4 },
        { IDC_AXIS,  5 },
        { IDC_AXIS,  6 },
        { IDC_AXIS,  7 },
        { IDC_AXIS,  8 },
        { IDC_AXIS,  9 },
        { IDC_AXIS, 10 },
        { IDC_AXIS, 11 },
        { IDC_AXIS, 12 },
        { IDC_AXIS, 13 },
        { IDC_AXIS, 14 },
        { IDC_AXIS, 15 },
        { IDC_AXIS, 16 },
        { IDC_AXIS, 17 }
    };
    static inputdev_layout_control_t joyGroup3[] = {
        { IDC_AXIS, 18 },
        { IDC_AXIS, 19 },
        { IDC_AXIS, 20 },
        { IDC_AXIS, 21 },
        { IDC_AXIS, 22 },
        { IDC_AXIS, 23 },
        { IDC_AXIS, 24 },
        { IDC_AXIS, 25 },
        { IDC_AXIS, 26 },
        { IDC_AXIS, 27 },
        { IDC_AXIS, 28 },
        { IDC_AXIS, 29 },
        { IDC_AXIS, 30 },
        { IDC_AXIS, 31 }
    };
    static inputdev_layout_control_t joyGroup4[] = {
        { IDC_HAT,  0 },
        { IDC_HAT,  1 },
        { IDC_HAT,  2 },
        { IDC_HAT,  3 }
    };
    static inputdev_layout_control_t joyGroup5[] = {
        { IDC_KEY,  0 },
        { IDC_KEY,  1 },
        { IDC_KEY,  2 },
        { IDC_KEY,  3 },
        { IDC_KEY,  4 },
        { IDC_KEY,  5 },
        { IDC_KEY,  6 },
        { IDC_KEY,  7 }
    };
    static inputdev_layout_control_t joyGroup6[] = {
        { IDC_KEY,  8 },
        { IDC_KEY,  9 },
        { IDC_KEY, 10 },
        { IDC_KEY, 11 },
        { IDC_KEY, 12 },
        { IDC_KEY, 13 },
        { IDC_KEY, 14 },
        { IDC_KEY, 15 },
    };
    static inputdev_layout_control_t joyGroup7[] = {
        { IDC_KEY, 16 },
        { IDC_KEY, 17 },
        { IDC_KEY, 18 },
        { IDC_KEY, 19 },
        { IDC_KEY, 20 },
        { IDC_KEY, 21 },
        { IDC_KEY, 22 },
        { IDC_KEY, 23 },
    };
    static inputdev_layout_control_t joyGroup8[] = {
        { IDC_KEY, 24 },
        { IDC_KEY, 25 },
        { IDC_KEY, 26 },
        { IDC_KEY, 27 },
        { IDC_KEY, 28 },
        { IDC_KEY, 29 },
        { IDC_KEY, 30 },
        { IDC_KEY, 31 },
    };
    static inputdev_layout_controlgroup_t joyGroups[] = {
        { joyGroup1, NUMITEMS(joyGroup1) },
        { joyGroup2, NUMITEMS(joyGroup2) },
        { joyGroup3, NUMITEMS(joyGroup3) },
        { joyGroup4, NUMITEMS(joyGroup4) },
        { joyGroup5, NUMITEMS(joyGroup5) },
        { joyGroup6, NUMITEMS(joyGroup6) },
        { joyGroup7, NUMITEMS(joyGroup7) },
        { joyGroup8, NUMITEMS(joyGroup8) }
    };
    static inputdev_layout_t joyLayout = { joyGroups, NUMITEMS(joyGroups) };
    Point2Raw origin = {{{2, 2}}};
    Size2Raw dimensions{};
    if (novideo || isDedicated) return; // Not for us.
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    if (!devRendKeyState && !devRendMouseState && !devRendJoyState) return;
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    if (devRendKeyState)
    {
        Rend_RenderInputDeviceStateVisual(InputSystem::get().device(IDEV_KEYBOARD), &keyLayout, &origin, &dimensions);
        origin.y += dimensions.height + SPACING;
    }
    if (devRendMouseState)
    {
        Rend_RenderInputDeviceStateVisual(InputSystem::get().device(IDEV_MOUSE), &mouseLayout, &origin, &dimensions);
        origin.y += dimensions.height + SPACING;
    }
    if (devRendJoyState)
    {
        Rend_RenderInputDeviceStateVisual(InputSystem::get().device(IDEV_JOY1), &joyLayout, &origin, &dimensions);
    }
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
#undef NUMITEMS
#undef SPACING
}
void I_DebugDrawerConsoleRegister()
{
    C_VAR_BYTE("rend-dev-input-joy-state",   &devRendJoyState,   CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-input-key-state",   &devRendKeyState,   CVF_NO_ARCHIVE, 0, 1);
    C_VAR_BYTE("rend-dev-input-mouse-state", &devRendMouseState, CVF_NO_ARCHIVE, 0, 1);
}
#endif // DENG2_DEBUG
/** @file inputdevice.cpp  Logical input device.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/inputdevice.h"
#include "ui/joystick.h"
#include "ui/axisinputcontrol.h"
#include "ui/buttoninputcontrol.h"
#include "ui/hatinputcontrol.h"
#include <QList>
#include <QtAlgorithms>
#include <de/Log>
/** @file inputsystem.cpp  Input subsystem.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_BINDING
#include "de_platform.h" // strdup macro
#include "ui/inputsystem.h"
#include <QList>
#include <QtAlgorithms>
#include <de/timer.h> // SECONDSPERTIC
#include <de/Record>
#include <de/LogBuffer>
#include <de/NumberValue>
#include <de/Function>
#include <de/Context>
#include <de/ScriptSystem>
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include <doomsday/console/exec.h>
#include "clientapp.h"
#include "dd_def.h"  // MAXEVENTS, ASSERT_NOT_64BIT
#include "dd_main.h" // App_GameLoaded()
#include "dd_loop.h" // DD_IsFrameTimeAdvancing()
#include "render/vr.h"
#include "world/p_players.h"
#include "BindContext"
#include "CommandBinding"
#include "ImpulseBinding"
#include "ui/ddevent.h"
#include "ui/b_util.h"
#include "ui/clientwindow.h"
#include "ui/clientwindowsystem.h"
#include "ui/inputdebug.h"  // Debug visualization.
#include "ui/inputdevice.h"
#include "ui/axisinputcontrol.h"
#include "ui/buttoninputcontrol.h"
#include "ui/hatinputcontrol.h"
#include "ui/controllerpresets.h"
#include "ui/sys_input.h"
#include "sys_system.h" // novideo
/** @file mouse_qt.cpp
 *
 * Mouse driver that gets mouse input from the Qt based canvas widget.
 * @ingroup input
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "ui/sys_input.h"
#include "ui/clientwindowsystem.h"
#include "ui/clientwindow.h"
#include "ui/mouse_qt.h"
#include <string.h>
#include <QApplication>
#include <QWidget>
#include <QPoint>
#include <QCursor>
#ifdef MACOSX
#  include "cursor_macx.h"
#endif
/** @file nativemenu.cpp  Native menu with application-level functions.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/nativemenu.h"
#include "ui/clientwindow.h"
#include "ui/widgets/taskbarwidget.h"
#include "updater.h"
#include "clientapp.h"
#include <doomsday/Games>
#include <doomsday/console/exec.h>
#include <QMenuBar>
#include <QMenu>
/** @file nativeui.cpp Native GUI functionality.
 * @ingroup base
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <QMessageBox>
#include <QFile>
#include <QString>
#include <stdarg.h>
#include "dd_share.h"
#include "sys_system.h"
#include <QLayout>
#include <QLabel>
#include <QVBoxLayout>
#include <QDebug>
#include <de/App>
#include <de/ByteArrayFile>
#include <de/FileSystem>
#include "ui/nativeui.h"
#include "ui/clientwindow.h"
/** @file postprocessing.cpp World frame post processing.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/postprocessing.h"
#include "ui/clientwindow.h"
#include "ui/viewcompositor.h"
#include "clientapp.h"
#include "clientplayer.h"
#include <de/Animation>
#include <de/Drawable>
#include <de/GLInfo>
#include <de/GLTextureFramebuffer>
#include <de/LogBuffer>
#include <QList>
/** @file progress.cpp  Simple wrapper for the Busy progress bar.
 * @ingroup console
 *
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "ui/progress.h"
#include "ui/clientwindow.h"
#include "ui/widgets/busywidget.h"
#include <de/ProgressWidget>
/** @file savelistdata.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/savelistdata.h"
#include <doomsday/doomsdayapp.h>
#include <doomsday/games.h>
#include <doomsday/AbstractSession>
#include <doomsday/SaveGames>
#include <de/Loop>
/** @file styledlogsinkformatter.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/styledlogsinkformatter.h"
#include <de/Variable>
#include <de/Value>
#include <de/App>
#include <de/Config>
/** @file sys_input.cpp  Keyboard and mouse input pre-processing.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Jamie Jones <yagisan@dengine.net>
 * @authors Copyright © 2005 Zachary Keene <zjkeene@bellsouth.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_platform.h"
#include "ui/sys_input.h"
#include <cstdlib>
#include <de/c_wrapper.h>
#include "sys_system.h"
#include "ui/mouse_qt.h"  // portable
#ifdef WIN32
#  include "directinput.h"
#  include "mouse_win32.h"
#endif
#define EVBUFSIZE       64
#define KEYBUFSIZE      32
/** @file ui_main.cpp  Graphical User Interface.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "ui/ui_main.h"
#include "ui/clientwindow.h"
#include "sys_system.h"
#include <cmath>
#include <de/GLState>
#include <de/GLInfo>
#include <doomsday/console/cmd.h>
#include <doomsday/filesys/fs_util.h>
#include "api_fontrender.h"
#include "gl/gl_main.h"
#include "gl/gl_draw.h"
#include "render/rend_main.h"
#include "render/rend_font.h"
/** @file viewcompositor.cpp  Game view compositor.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "ui/viewcompositor.h"
#include "ui/clientwindow.h"
#include "ui/postprocessing.h"
#include "ui/infine/finaleinterpreter.h"
#include "ui/infine/finalepagewidget.h"
#include "network/net_main.h" // Net_Drawer (get rid of this)
#include "render/rend_main.h"
#include "render/viewports.h"
#include "world/p_players.h"
#include "world/map.h"
#include "api_console.h"
#include "api_render.h"
#include "clientapp.h"
#include "dd_main.h"
#include "gl/gl_main.h"
#include <de/Config>
#include <de/CommandLine>
#include <de/GLState>
#include <de/GLShaderBank>
#include <de/Drawable>
#include <de/VRConfig>
#include <de/WindowTransform>
/** @file zonedebug.cpp  Memory zone debug visualization.
 *
 * Shows the contents of the memory zone as on-screen visualization. This is
 * only available in debug builds and provides a view to the layout of the
 * allocated memory inside the zone.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2013-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#if defined (DENG_DEBUG) && defined (DENG_OPENGL)
#include <cmath>
#include <de/GLState>
#include <de/GLInfo>
#include <de/concurrency.h>
#include <de/Rectangle>
#include <de/Vector>
#include "../../../sdk/liblegacy/src/memoryzone_private.h"
#include "gl/gl_main.h"
#include "gl/gl_draw.h"
#include "ui/clientwindow.h"
using namespace de;
static void drawRegion(memvolume_t &volume, Rectanglei const &rect, size_t start,
    size_t size, float const color[4])
{
    DENG2_ASSERT(start + size <= volume.size);
    int const bytesPerRow = (volume.size - sizeof(memzone_t)) / rect.height();
    float const toPixelScale = (float)rect.width() / (float)bytesPerRow;
    size_t const edge = rect.topLeft.x + rect.width();
    int x = (start % bytesPerRow) * toPixelScale + rect.topLeft.x;
    int y = start / bytesPerRow + rect.topLeft.y;
    int pixels = de::max<dint>(1, std::ceil(size * toPixelScale));
    while (pixels > 0)
    {
        int const availPixels = edge - x;
        int const usedPixels = de::min(availPixels, pixels);
        DGL_Color4fv(color);
        DGL_Vertex2f(x, y);
        DGL_Vertex2f(x + usedPixels, y);
        pixels -= usedPixels;
        y++;
        x = rect.topLeft.x;
    }
}
void Z_DebugDrawVolume(MemoryZonePrivateData *pd, memvolume_t *volume, Rectanglei const &rect)
{
    float const opacity = .85f;
    float const colAppStatic[4]   = { 1, 1, 1, .65f };
    float const colGameStatic[4]  = { 1, 0, 0, .65f };
    float const colMap[4]         = { 0, 1, 0, .65f };
    float const colMapStatic[4]   = { 0, .5f, 0, .65f };
    float const colCache[4]       = { 1, 0, 1, .65f };
    float const colOther[4]       = { 0, 0, 1, .65f };
    char *base = ((char *)volume->zone) + sizeof(memzone_t);
    DGL_Color4f(0, 0, 0, opacity);
    GL_DrawRect(rect);
    DGL_Color4f(1, 1, 1, opacity/2);
    LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    GL_DrawRect(rect);
    LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    DGL_Begin(DGL_LINES);
    for (memblock_t *block = volume->zone->blockList.next;
        block != &volume->zone->blockList;
        block = block->next)
    {
        float const *color = colOther;
        if (!block->user) continue; // Free is black.
        switch (block->tag)
        {
        case PU_GAMESTATIC: color = colGameStatic; break;
        case PU_MAP:        color = colMap; break;
        case PU_MAPSTATIC:  color = colMapStatic; break;
        case PU_APPSTATIC:  color = colAppStatic; break;
        default:
            if (block->tag >= PU_PURGELEVEL)
                color = colCache;
            break;
        }
        drawRegion(*volume, rect, (char *)block - base, block->size, color);
    }
    DGL_End();
    if (pd->isVolumeTooFull(volume))
    {
        DGL_Color4f(1, 0, 0, 1);
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        GL_DrawRect(rect);
        LIBGUI_GL.glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
}
void Z_DebugDrawer(void)
{
    MemoryZonePrivateData pd;
    memvolume_t* volume;
    int i, volCount, h;
    if (!CommandLine_Exists("-zonedebug")) return;
    DENG_ASSERT_IN_MAIN_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_PushState();
    DGL_CullFace(DGL_NONE);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT, -1, 1);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    Z_GetPrivateData(&pd);
    pd.lock();
    volCount = pd.volumeCount;
    h = 200;
    if (h * volCount + 10*(volCount - 1) > DENG_GAMEVIEW_HEIGHT)
    {
        h = (DENG_GAMEVIEW_HEIGHT - 10*(volCount - 1))/volCount;
    }
    i = 0;
    for (volume = pd.volumeRoot; volume; volume = volume->next, ++i)
    {
        int size = de::min(400, DENG_GAMEVIEW_WIDTH);
        Z_DebugDrawVolume(&pd, volume,
                          Rectanglei::fromSize(Vector2i(DENG_GAMEVIEW_WIDTH - size - 1,
                                                        DENG_GAMEVIEW_HEIGHT - size * (i+1) - 10*i - 1),
                                               Vector2ui(size, size)));
    }
    pd.unlock();
    DGL_PopState();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
}
#endif // _DEBUG
using namespace de;
DENG2_PIMPL_NOREF(AlertMask)
, DENG2_OBSERVES(Variable, Change) {
    duint32 mask[LogEntry::HighestLogLevel + 1];
    Impl() {
        zap(mask);
        mask[LogEntry::Warning] = mask[LogEntry::Error] = mask[LogEntry::Critical]
                = LogEntry::AllDomains;
    }
    void variableValueChanged(Variable &, Value const &) {
        updateMask();
    }
    void updateMask() {
        zap(mask);
        Config const &cfg = App::config();
        for(int bit = LogEntry::FirstDomainBit; bit <= LogEntry::LastDomainBit; ++bit) {
            int const alertLevel = cfg.geti(String("alert.") +
                                            LogFilter::domainRecordName(LogEntry::Context(1 << bit)));
            for(int i = LogEntry::LowestLogLevel; i <= LogEntry::HighestLogLevel; ++i) {
                if(alertLevel <= i) {
                    mask[i] |= (1 << bit);
                }
            }
        }
    }
};
AlertMask::AlertMask() : d(new Impl)
{}
void AlertMask::init() {
    foreach(Variable const *var, App::config().subrecord("alert").members()) {
        var->audienceForChange() += d;
    }
    d->updateMask();
}
bool AlertMask::shouldRaiseAlert(duint32 entryMetadata) const {
    int const level = entryMetadata & LogEntry::LevelMask;
    return ((entryMetadata & LogEntry::DomainMask) & d->mask[level]) != 0;
}
static dfloat const AXIS_NORMALIZE = 1.f / float(IJOY_AXISMAX); // Normalize from SDL's range
DENG2_PIMPL_NOREF(AxisInputControl) {
    Type type = Pointer;
    dint flags = 0;
    ddouble position       = 0; ///< Current translated position (-1..1) including any filtering.
    ddouble realPosition   = 0; ///< The actual latest position (-1..1).
    ddouble markedPosition = 0;
    dfloat offset   = 0;           ///< Offset to add to real input value.
    dfloat scale    = 1;           ///< Scaling factor for real input values.
    dfloat deadZone = 0;           ///< Dead zone in (0..1) range.
    ddouble sharpPosition = 0;     ///< Current sharp (accumulated) position, entered into the Smoother.
    Smoother *smoother = nullptr;  ///< Smoother for the input values (owned).
    ddouble prevSmoothPos = 0;     ///< Previous evaluated smooth position (needed for producing deltas).
    duint time = 0;                ///< Timestamp of the last position update.
    Impl() {
        Smoother_SetMaximumPastNowDelta(smoother = Smoother_New(), 2 * SECONDSPERTIC);
    }
    ~Impl() {
        Smoother_Delete(smoother);
    }
#if 0
    static float filter(int grade, float *accumulation, float ticLength) {
        DENG2_ASSERT(accumulation);
        int dir     = de::sign(*accumulation);
        float avail = fabs(*accumulation);
        float target = avail * (MAX_AXIS_FILTER - de::clamp(1, grade, 39));
        /*
        if (target < -.7) target = -.7;
        else if (target > .7) target = .7;
        else target = 0;
        */
        float used = target * ticLength;
        if (used > avail) {
            *accumulation = nullptr;
            used = avail;
        }
        else {
            if (*accumulation > nullptr)
                *accumulation -= used;
            else
                *accumulation += used;
        }
        return dir * used;
    }
#endif
};
AxisInputControl::AxisInputControl(String const &name, Type type) : d(new Impl) {
    setName(name);
    d->type = type;
}
AxisInputControl::~AxisInputControl()
{}
AxisInputControl::Type AxisInputControl::type() const {
    DENG2_GUARD(this);
    return d->type;
}
void AxisInputControl::setRawInput(bool yes) {
    DENG2_GUARD(this);
    if (yes) d->flags |= IDA_RAW;
    else     d->flags &= ~IDA_RAW;
}
bool AxisInputControl::isActive() const {
    DENG2_GUARD(this);
    return (d->flags & IDA_DISABLED) == 0;
}
bool AxisInputControl::isInverted() const {
    DENG2_GUARD(this);
    return (d->flags & IDA_INVERT) != 0;
}
void AxisInputControl::update(timespan_t ticLength) {
    DENG2_GUARD(this);
    Smoother_Advance(d->smoother, ticLength);
    if (d->type == Stick) {
        if (d->flags & IDA_RAW) {
            d->position = d->realPosition;
        }
        else {
            Smoother_EvaluateComponent(d->smoother, 0, &d->position);
        }
    }
    else if (d->type == Pointer) {
        if (d->flags & IDA_RAW) {
            d->position    += d->realPosition;
            d->realPosition = 0;
        }
        else {
            coord_t smoothPos = d->prevSmoothPos;
            Smoother_EvaluateComponent(d->smoother, 0, &smoothPos);
            d->position     += smoothPos - d->prevSmoothPos;
            d->prevSmoothPos = smoothPos;
        }
    }
    setBindContextAssociation(Expired, UnsetFlags);
}
ddouble AxisInputControl::position() const {
    DENG2_GUARD(this);
    return d->position;
}
void AxisInputControl::setPosition(ddouble newPosition) {
    DENG2_GUARD(this);
    d->position = newPosition;
}
void AxisInputControl::markPosition() {
    d->markedPosition = d->position;
}
ddouble AxisInputControl::markedPosition() const {
    return d->markedPosition;
}
void AxisInputControl::applyRealPosition(dfloat pos) {
    DENG2_GUARD(this);
    dfloat const oldRealPos  = d->realPosition;
    dfloat const transformed = translateRealPosition(pos);
    d->realPosition = transformed;
    if (oldRealPos != d->realPosition) {
        d->time = DD_LatestRunTicsStartTime();
    }
    if (d->type == Stick) {
        d->sharpPosition = d->realPosition;
    }
    else // Cumulative.
    {
        d->sharpPosition += d->realPosition;
    }
    Smoother_AddPosXY(d->smoother, DD_LatestRunTicsStartTime(), d->sharpPosition, 0);
}
dfloat AxisInputControl::translateRealPosition(dfloat realPos) const {
    DENG2_GUARD(this);
    if (!isActive()) return 0;
    float outPos = realPos * AXIS_NORMALIZE * d->scale;
    if (d->type == Stick) // Only stick axes are dead-zoned and clamped.
    {
        outPos += d->offset;
        if (std::abs(outPos) <= d->deadZone) {
            outPos = 0;
        }
        else {
            outPos -= d->deadZone * de::sign(outPos);  // Remove the dead zone.
            outPos *= 1.0f/(1.0f - d->deadZone);       // Normalize.
            outPos = de::clamp(-1.0f, outPos, 1.0f);
        }
    }
    if (isInverted()) {
        outPos = -outPos;
    }
    return outPos;
}
dfloat AxisInputControl::deadZone() const {
    DENG2_GUARD(this);
    return d->deadZone;
}
void AxisInputControl::setDeadZone(dfloat newDeadZone) {
    DENG2_GUARD(this);
    d->deadZone = newDeadZone;
}
dfloat AxisInputControl::scale() const {
    DENG2_GUARD(this);
    return d->scale;
}
void AxisInputControl::setScale(dfloat newScale) {
    DENG2_GUARD(this);
    d->scale = newScale;
}
dfloat AxisInputControl::offset() const {
    DENG2_GUARD(this);
    return d->offset;
}
void AxisInputControl::setOffset(dfloat newOffset) {
    DENG2_GUARD(this);
    d->offset = newOffset;
}
duint AxisInputControl::time() const {
    DENG2_GUARD(this);
    return d->time;
}
String AxisInputControl::description() const {
    DENG2_GUARD(this);
    QStringList flags;
    if (!isActive()) flags << "disabled";
    if (isInverted()) flags << "inverted";
    String flagsString;
    if (!flags.isEmpty()) {
        String flagsAsText = flags.join("|");
        flagsString = String(_E(l) " Flags :" _E(.)_E(i) "%1" _E(.)).arg(flagsAsText);
    }
    return String(_E(b) "%1 " _E(.) "(%2)"
                  _E(l) " Current value: " _E(.) "%3"
                  _E(l) " Deadzone: " _E(.) "%4"
                  _E(l) " Scale: "     _E(.) "%5"
                  _E(l) " Offset: "     _E(.) "%6"
                  "%7")
               .arg(fullName())
               .arg(d->type == Stick? "Stick" : "Pointer")
               .arg(position())
               .arg(d->deadZone)
               .arg(d->scale)
               .arg(d->offset)
               .arg(flagsString);
}
bool AxisInputControl::inDefaultState() const {
    DENG2_GUARD(this);
    return d->position == 0; // Centered?
}
void AxisInputControl::reset() {
    DENG2_GUARD(this);
    if (d->type == Pointer) {
        d->position      = 0;
        d->sharpPosition = 0;
        d->prevSmoothPos = 0;
    }
    Smoother_Clear(d->smoother);
}
void AxisInputControl::consoleRegister() {
    DENG2_GUARD(this);
    DENG2_ASSERT(hasDevice() && !name().isEmpty());
    String controlName = String("input-%1-%2").arg(device().name()).arg(name());
    Block scale = (controlName + "-factor").toUtf8();
    C_VAR_FLOAT(scale.constData(), &d->scale, CVF_NO_MAX, 0, 0);
    Block flags = (controlName + "-flags").toUtf8();
    C_VAR_INT(flags.constData(), &d->flags, 0, 0, 7);
    if (d->type == Stick) {
        Block deadzone = (controlName + "-deadzone").toUtf8();
        C_VAR_FLOAT(deadzone.constData(), &d->deadZone, 0, 0, 1);
        Block offset = (controlName + "-offset").toUtf8();
        C_VAR_FLOAT(offset.constData(), &d->offset, CVF_NO_MAX | CVF_NO_MIN, 0, 0);
    }
}
/**
 * Binding context fallback for the "global" context.
 *
 * @param ddev  Event being processed.
 *
 * @return @c true if the event was eaten and can be processed by the rest of the
 * binding context stack.
 */
static int globalContextFallback(ddevent_t const *ddev) {
    if (App_GameLoaded() && !BusyMode_Active()) {
        event_t ev;
        if (InputSystem::convertEvent(*ddev, ev)) {
            if (gx.Responder && gx.Responder(&ev)) {
                return true;
            }
        }
    }
    return false;
}
void B_Init() {
    InputSystem &isys = ClientApp::inputSystem();
    isys.newContext(DEFAULT_BINDING_CONTEXT_NAME);
    isys.newContext("map");
    isys.newContext("map-freepan");
    isys.newContext("finale"); // uses a fallback responder to handle script events
    isys.newContext("menu")->acquireAll();
    isys.newContext("gameui");
    isys.newContext("shortcut");
    isys.newContext("chat")->acquire(IDEV_KEYBOARD);
    isys.newContext("message")->acquireAll();
    BindContext *bc = isys.newContext(CONSOLE_BINDING_CONTEXT_NAME);
    bc->protect();              // Only we can (de)activate.
    bc->acquire(IDEV_KEYBOARD); // Console takes over all keyboard events.
    isys.newContext(UI_BINDING_CONTEXT_NAME)->acquireAll();
    bc = isys.newContext(GLOBAL_BINDING_CONTEXT_NAME);
    bc->protect();
    bc->setDDFallbackResponder(globalContextFallback);
    bc->activate();
/*
    B_BindCommand("joy-hat-angle3", "print {angle 3}");
    B_BindCommand("joy-hat-center", "print center");
    B_BindCommand("game:key-m-press", "print hello");
    B_BindCommand("key-codex20-up", "print {space released}");
    B_BindCommand("key-up-down + key-shift + key-ctrl", "print \"shifted and controlled up\"");
    B_BindCommand("key-up-down + key-shift", "print \"shifted up\"");
    B_BindCommand("mouse-left", "print mbpress");
    B_BindCommand("mouse-right-up", "print \"right mb released\"");
    B_BindCommand("joy-x-neg1.0 + key-ctrl-up", "print \"joy x negative without ctrl\"");
    B_BindCommand("joy-x- within 0.1 + joy-y-pos1", "print \"joy x centered\"");
    B_BindCommand("joy-x-pos1.0", "print \"joy x positive\"");
    B_BindCommand("joy-x-neg1.0", "print \"joy x negative\"");
    B_BindCommand("joy-z-pos1.0", "print \"joy z positive\"");
    B_BindCommand("joy-z-neg1.0", "print \"joy z negative\"");
    B_BindCommand("joy-w-pos1.0", "print \"joy w positive\"");
    B_BindCommand("joy-w-neg1.0", "print \"joy w negative\"");
    */
    /*B_BindControl("turn", "key-left-staged-inverse");
    B_BindControl("turn", "key-right-staged");
    B_BindControl("turn", "mouse-x");
    B_BindControl("turn", "joy-x + key-shift-up + joy-hat-center + key-code123-down");
    */
    isys.bindDefaults();
    isys.initialContextActivations();
}
byte zeroControlUponConflict = true;
static float STAGE_THRESHOLD = 6.f/35;
static float STAGE_FACTOR    = .5f;
bool B_ParseButtonState(Binding::ControlTest &test, char const *toggleName) {
    DENG2_ASSERT(toggleName);
    if (!qstrlen(toggleName) || !qstricmp(toggleName, "down")) {
        test = Binding::ButtonStateDown; // this is the default, if omitted
        return true;
    }
    if (!qstricmp(toggleName, "undefined")) {
        test = Binding::ButtonStateAny;
        return true;
    }
    if (!qstricmp(toggleName, "repeat")) {
        test = Binding::ButtonStateRepeat;
        return true;
    }
    if (!qstricmp(toggleName, "press")) {
        test = Binding::ButtonStateDownOrRepeat;
        return true;
    }
    if (!qstricmp(toggleName, "up")) {
        test = Binding::ButtonStateUp;
        return true;
    }
    LOG_INPUT_WARNING("\"%s\" is not a toggle state") << toggleName;
    return false; // Not recognized.
}
bool B_ParseAxisPosition(Binding::ControlTest &test, float &pos, char const *desc) {
    DENG2_ASSERT(desc);
    if (!qstrnicmp(desc, "within", 6) && qstrlen(desc) > 6) {
        test = Binding::AxisPositionWithin;
        pos  = String((desc + 6)).toFloat();
        return true;
    }
    if (!qstrnicmp(desc, "beyond", 6) && qstrlen(desc) > 6) {
        test = Binding::AxisPositionBeyond;
        pos  = String((desc + 6)).toFloat();
        return true;
    }
    if (!qstrnicmp(desc, "pos", 3) && qstrlen(desc) > 3) {
        test = Binding::AxisPositionBeyondPositive;
        pos  = String((desc + 3)).toFloat();
        return true;
    }
    if (!qstrnicmp(desc, "neg", 3) && qstrlen(desc) > 3) {
        test = Binding::AxisPositionBeyondNegative;
        pos  = -String((desc + 3)).toFloat();
        return true;
    }
    LOG_INPUT_WARNING("Axis position \"%s\" is invalid") << desc;
    return false;
}
bool B_ParseModifierId(int &id, char const *desc) {
    DENG2_ASSERT(desc);
    id = String(desc).toInt() - 1 + CTL_MODIFIER_1;
    return (id >= CTL_MODIFIER_1 && id <= CTL_MODIFIER_4);
}
bool B_ParseKeyId(int &id, char const *desc) {
    DENG2_ASSERT(desc);
    LOG_AS("B_ParseKeyId");
    if (!qstrnicmp(desc, "code", 4) && qstrlen(desc) == 7) {
        if (desc[4] == 'x' || desc[4] == 'X') {
            id = String((desc + 5)).toInt(nullptr, 16);
            return true;
        }
        id = String((desc + 4)).toInt();
        if (id > 0 && id <= 255) return true;
        LOGDEV_INPUT_WARNING("Key code %i out of range") << id;
        return false;
    }
    id = B_KeyForShortName(desc);
    if (id) return true;
    LOG_INPUT_WARNING("Unknown key \"%s\"") << desc;
    return false;
}
bool B_ParseMouseTypeAndId(ddeventtype_t &type, int &id, char const *desc) {
    DENG2_ASSERT(desc);
    InputDevice const &mouse = InputSystem::get().device(IDEV_MOUSE);
    id = mouse.toButtonId(desc);
    if (id >= 0) {
        type = E_TOGGLE;
        return true;
    }
    if (!qstrnicmp(desc, "button", 6) && qstrlen(desc) > 6) {
        type = E_TOGGLE;
        id   = String((desc + 6)).toInt() - 1;
        if (mouse.hasButton(id))
            return true;
        LOG_INPUT_WARNING("\"%s\" button %i does not exist") << mouse.title() << id;
        return false;
    }
    type = E_AXIS;
    id   = mouse.toAxisId(desc);
    if (id >= 0) return true;
    LOG_INPUT_WARNING("\"%s\" axis \"%s\" does not exist") << mouse.title() << desc;
    return false;
}
bool B_ParseDeviceAxisTypeAndId(ddeventtype_t &type, int &id, InputDevice const &device, char const *desc) {
    DENG2_ASSERT(desc);
    type = E_AXIS;
    id   = device.toAxisId(desc);
    if (id >= 0) return true;
    LOG_INPUT_WARNING("Axis \"%s\" is not defined in device '%s'") << desc << device.name();
    return false;
}
bool B_ParseJoystickTypeAndId(ddeventtype_t &type, int &id, int deviceId, char const *desc) {
    InputDevice &device = InputSystem::get().device(deviceId);
    if (!qstrnicmp(desc, "button", 6) && qstrlen(desc) > 6) {
        type = E_TOGGLE;
        id   = String((desc + 6)).toInt() - 1;
        if (device.hasButton(id))
            return true;
        LOG_INPUT_WARNING("\"%s\" button %i does not exist") << device.title() << id;
        return false;
    }
    if (!qstrnicmp(desc, "hat", 3) && qstrlen(desc) > 3) {
        type = E_ANGLE;
        id   = String((desc + 3)).toInt() - 1;
        if (device.hasHat(id))
            return true;
        LOG_INPUT_WARNING("\"%s\" hat %i does not exist") << device.title() << id;
        return false;
    }
    if (!qstricmp(desc, "hat")) {
        type = E_ANGLE;
        id   = 0;
        return true;
    }
    return B_ParseDeviceAxisTypeAndId(type, id, device, desc);
}
bool B_ParseHatAngle(float &pos, char const *desc) {
    DENG2_ASSERT(desc);
    if (!qstricmp(desc, "center")) {
        pos = -1;
        return true;
    }
    if (!qstrnicmp(desc, "angle", 5) && qstrlen(desc) > 5) {
        pos = String((desc + 5)).toFloat();
        return true;
    }
    LOG_INPUT_WARNING("Angle position \"%s\" is invalid") << desc;
    return false;
}
bool B_ParseBindingCondition(Record &cond, char const *desc) {
    DENG2_ASSERT(desc);
    AutoStr *str = AutoStr_NewStd();
    desc = Str_CopyDelim(str, desc, '-');
    if (!Str_CompareIgnoreCase(str, "multiplayer")) {
        cond.set("type", int(Binding::GlobalState));
        cond.set("multiplayer", true);
    }
    else if (!Str_CompareIgnoreCase(str, "modifier")) {
        cond.set("type", int(Binding::ModifierState));
        cond.set("device", -1); // not used
        desc = Str_CopyDelim(str, desc, '-');
        int id = 0;
        bool ok = B_ParseModifierId(id, Str_Text(str));
        if (!ok) return false;
        cond.set("id", id);
        desc = Str_CopyDelim(str, desc, '-');
        Binding::ControlTest test = Binding::None;
        ok = B_ParseButtonState(test, Str_Text(str));
        if (!ok) return false;
        cond.set("test", int(test));
    }
    else if (!Str_CompareIgnoreCase(str, "key")) {
        cond.set("type", int(Binding::ButtonState));
        cond.set("device", int(IDEV_KEYBOARD));
        desc = Str_CopyDelim(str, desc, '-');
        int id = 0;
        bool ok = B_ParseKeyId(id, Str_Text(str));
        if (!ok) return false;
        cond.set("id", id);
        desc = Str_CopyDelim(str, desc, '-');
        Binding::ControlTest test = Binding::None;
        ok = B_ParseButtonState(test, Str_Text(str));
        if (!ok) return false;
        cond.set("test", int(test));
    }
    else if (!Str_CompareIgnoreCase(str, "mouse")) {
        cond.set("device", int(IDEV_MOUSE));
        desc = Str_CopyDelim(str, desc, '-');
        ddeventtype_t type = E_TOGGLE;
        int id = 0;
        bool ok = B_ParseMouseTypeAndId(type, id, Str_Text(str));
        if (!ok) return false;
        cond.set("type", int(type));
        cond.set("id", id);
        desc = Str_CopyDelim(str, desc, '-');
        if (type == E_TOGGLE) {
            cond.set("type", int(Binding::ButtonState));
            Binding::ControlTest test = Binding::None;
            ok = B_ParseButtonState(test, Str_Text(str));
            if (!ok) return false;
            cond.set("test", int(test));
        }
        else if (type == E_AXIS) {
            cond.set("type", int(Binding::AxisState));
            Binding::ControlTest test = Binding::None;
            float pos;
            ok = B_ParseAxisPosition(test, pos, Str_Text(str));
            if (!ok) return false;
            cond.set("test", int(test));
            cond.set("pos", pos);
        }
    }
    else if (!Str_CompareIgnoreCase(str, "joy") || !Str_CompareIgnoreCase(str, "head")) {
        cond.set("device", int(!Str_CompareIgnoreCase(str, "joy")? IDEV_JOY1 : IDEV_HEAD_TRACKER));
        desc = Str_CopyDelim(str, desc, '-');
        ddeventtype_t type = E_TOGGLE;
        int id = 0;
        bool ok = B_ParseJoystickTypeAndId(type, id, cond.geti("device"), Str_Text(str));
        if (!ok) return false;
        cond.set("type", int(type));
        desc = Str_CopyDelim(str, desc, '-');
        if (type == E_TOGGLE) {
            cond.set("type", int(Binding::ButtonState));
            Binding::ControlTest test = Binding::None;
            ok = B_ParseButtonState(test, Str_Text(str));
            if (!ok) return false;
            cond.set("test", int(test));
        }
        else if (type == E_AXIS) {
            cond.set("type", int(Binding::AxisState));
            Binding::ControlTest test = Binding::None;
            float pos = 0;
            ok = B_ParseAxisPosition(test, pos, Str_Text(str));
            if (!ok) return false;
            cond.set("test", int(test));
            cond.set("pos", pos);
        }
        else // Angle.
        {
            cond.set("type", int(Binding::HatState));
            float pos = 0;
            ok = B_ParseHatAngle(pos, Str_Text(str));
            if (!ok) return false;
            cond.set("pos", pos);
        }
    }
    else {
        LOG_INPUT_WARNING("Unknown input device \"%s\"") << Str_Text(str);
        return false;
    }
    if (cond.geti("type") == Binding::ButtonState) {
        if (cond.geti("test") != Binding::ButtonStateUp &&
           cond.geti("test") != Binding::ButtonStateDown) {
            LOG_INPUT_WARNING("\"%s\": Button condition can only be 'up' or 'down'") << desc;
            return false;
        }
    }
    desc = Str_CopyDelim(str, desc, '-');
    if (!Str_CompareIgnoreCase(str, "not")) {
        cond.set("negate", true);
    }
    if (!desc) return true; // No errors detected.
    LOG_INPUT_WARNING("Unrecognized condition \"%s\"") << desc;
    return false;
}
bool B_CheckAxisPosition(Binding::ControlTest test, float testPos, float pos) {
    switch (test) {
    case Binding::AxisPositionWithin:
        return !((pos > 0 && pos > testPos) || (pos < 0 && pos < -testPos));
    case Binding::AxisPositionBeyond:
        return ((pos > 0 && pos >= testPos) || (pos < 0 && pos <= -testPos));
    case Binding::AxisPositionBeyondPositive:
        return !(pos < testPos);
    case Binding::AxisPositionBeyondNegative:
        return !(pos > -testPos);
    default: break;
    }
    return false;
}
bool B_CheckCondition(Binding::CompiledConditionRecord const *condRec, int localNum,
                      BindContext const *context) {
    DENG2_ASSERT(condRec);
    auto const &cond = condRec->compiled();
    bool const fulfilled = !cond.negate;
    switch (cond.type) {
    case Binding::GlobalState:
        if (cond.multiplayer && netGame)
            return fulfilled;
        break;
    case Binding::AxisState: {
        AxisInputControl const &axis = InputSystem::get().device(cond.device).axis(cond.id);
        if (B_CheckAxisPosition(cond.test, cond.pos, axis.position())) {
            return fulfilled;
        }
        break; }
    case Binding::ButtonState: {
        ButtonInputControl const &button = InputSystem::get().device(cond.device).button(cond.id);
        bool isDown = button.isDown();
        if (( isDown && cond.test == Binding::ButtonStateDown) ||
            (!isDown && cond.test == Binding::ButtonStateUp)) {
            return fulfilled;
        }
        break; }
    case Binding::HatState: {
        HatInputControl const &hat = InputSystem::get().device(cond.device).hat(cond.id);
        if (hat.position() == cond.pos) {
            return fulfilled;
        }
        break; }
    case Binding::ModifierState:
        if (context) {
            float pos = 0, relative = 0;
            B_EvaluateImpulseBindings(context, localNum, cond.id, &pos, &relative, false /*no triggered*/);
            if ((cond.test == Binding::ButtonStateDown && fabs(pos) > .5) ||
                (cond.test == Binding::ButtonStateUp   && fabs(pos) < .5)) {
                return fulfilled;
            }
        }
        break;
    default: DENG2_ASSERT(!"B_CheckCondition: Unknown cond.type"); break;
    }
    return !fulfilled;
}
void B_EvaluateImpulseBindings(BindContext const *context, int localNum, int impulseId,
    float *pos, float *relativeOffset, bool allowTriggered) {
    DENG2_ASSERT(context); // Why call without one?
    DENG2_ASSERT(pos && relativeOffset);
    *pos = 0;
    *relativeOffset = 0;
    if (localNum < 0 || localNum >= DDMAXPLAYERS) return; // No local player specified.
    uint const nowTime = Timer_RealMilliseconds();
    bool conflicted[NUM_IBD_TYPES]; de::zap(conflicted);
    bool appliedState[NUM_IBD_TYPES]; de::zap(appliedState);
    context->forAllImpulseBindings(localNum, [&] (CompiledImpulseBindingRecord &rec) {
        auto const &bind = rec.compiled();
        if (bind.impulseId != impulseId) return LoopContinue;
        bool skip = false;
        ArrayValue const &conds = rec.geta(QStringLiteral("condition"));
        DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds.elements()) {
            if (!B_CheckCondition(static_cast<Binding::CompiledConditionRecord *>
                                  ((*i)->as<RecordValue>().record()), localNum, context)) {
                skip = true;
                break;
            }
        }
        if (skip) return LoopContinue;
        InputDevice const *device = InputSystem::get().devicePtr(bind.deviceId);
        if (!device || !device->isActive())
            return LoopContinue; // Not available.
        InputControl *ctrl = nullptr;
        switch (bind.type) {
        case IBD_AXIS:   ctrl = &device->axis  (bind.controlId); break;
        case IBD_TOGGLE: ctrl = &device->button(bind.controlId); break;
        case IBD_ANGLE:  ctrl = &device->hat   (bind.controlId); break;
        default: DENG2_ASSERT(!"B_EvaluateImpulseBindings: Invalid bind.type"); break;
        }
        float devicePos = 0;
        float deviceOffset = 0;
        uint deviceTime = 0;
        if (auto *axis = maybeAs<AxisInputControl>(ctrl)) {
            if (context && axis->bindContext() != context) {
                if (axis->hasBindContext() && !axis->bindContext()->
                        findImpulseBinding(bind.deviceId, IBD_AXIS, bind.controlId)) {
                    if (axis->type() == AxisInputControl::Pointer) {
                        axis->setPosition(0);
                    }
                }
                return LoopContinue; // Shadowed by a more important active class.
            }
            if (!(axis->bindContextAssociation() & InputControl::Expired)) {
                if (axis->type() == AxisInputControl::Pointer) {
                    deviceOffset = axis->position();
                    axis->setPosition(0);
                }
                else {
                    devicePos = axis->position();
                }
                deviceTime = axis->time();
            }
        }
        if (auto *button = maybeAs<ButtonInputControl>(ctrl)) {
            if (context && button->bindContext() != context)
                return LoopContinue; // Shadowed by a more important active context.
            if (!(button->bindContextAssociation() & InputControl::Expired)) {
                devicePos  = (button->isDown() ||
                              (allowTriggered && (button->bindContextAssociation() & InputControl::Triggered))? 1.0f : 0.0f);
                deviceTime = button->time();
                button->setBindContextAssociation(InputControl::Triggered, UnsetFlags);
            }
        }
        if (auto *hat = maybeAs<HatInputControl>(ctrl)) {
            if (context && hat->bindContext() != context)
                return LoopContinue; // Shadowed by a more important active class.
            if (!(hat->bindContextAssociation() & InputControl::Expired)) {
                devicePos  = (fequal(hat->position(), bind.angle)? 1.0f : 0.0f);
                deviceTime = hat->time();
            }
        }
        int const bflags = bind.flags;
        if (bflags & IBDF_INVERSE) {
            devicePos    = -devicePos;
            deviceOffset = -deviceOffset;
        }
        if (bflags & IBDF_TIME_STAGED) {
            if (nowTime - deviceTime < STAGE_THRESHOLD * 1000) {
                devicePos *= STAGE_FACTOR;
            }
        }
        *pos += devicePos;
        *relativeOffset += deviceOffset;
        if (!de::fequal(devicePos, 0.f)) {
            if (appliedState[bind.type]) {
                conflicted[bind.type] = true;
            }
            appliedState[bind.type] = true;
        }
        return LoopContinue;
    });
    if (zeroControlUponConflict) {
        for (int i = 0; i < NUM_IBD_TYPES; ++i) {
            if (conflicted[i])
                *pos = 0;
        }
    }
    *pos = de::clamp(-1.0f, *pos, 1.0f);
}
String B_ControlDescToString(int deviceId, ddeventtype_t type, int id) {
    InputDevice *device = nullptr;
    String str;
    if (type != E_SYMBOLIC) {
        device = &InputSystem::get().device(deviceId);
        str += device->name() + "-";
    }
    switch (type) {
    case E_TOGGLE: {
        DENG2_ASSERT(device);
        ButtonInputControl &button = device->button(id);
        if (!button.name().isEmpty()) {
            str += button.name();
        }
        else if (device == InputSystem::get().devicePtr(IDEV_KEYBOARD)) {
            char const *name = B_ShortNameForKey(id);
            if (name) {
                str += name;
            }
            else {
                str += String("code%1").arg(id, 3, 10, QChar('0'));
            }
        }
        else {
            str += "button" + String::number(id + 1);
        }
        break; }
    case E_AXIS:
        DENG2_ASSERT(device);
        str += device->axis(id).name();
        break;
    case E_ANGLE:    str += "hat" + String::number(id + 1); break;
    case E_SYMBOLIC: str += "sym";                          break;
    default: DENG2_ASSERT(!"B_ControlDescToString: Invalid event type"); break;
    }
    return str;
}
String B_ButtonStateToString(Binding::ControlTest test) {
    switch (test) {
    case Binding::ButtonStateAny:          return "-undefined";
    case Binding::ButtonStateDown:         return "-down";
    case Binding::ButtonStateRepeat:       return "-repeat";
    case Binding::ButtonStateDownOrRepeat: return "-press";
    case Binding::ButtonStateUp:           return "-up";
    default:
        DENG2_ASSERT(!"B_ButtonStateToString: Unknown test");
        return "";
    }
}
String B_AxisPositionToString(Binding::ControlTest test, float pos) {
    switch (test) {
    case Binding::AxisPositionWithin:         return String("-within%1").arg(pos);
    case Binding::AxisPositionBeyond:         return String("-beyond%1").arg(pos);
    case Binding::AxisPositionBeyondPositive: return String("-pos%1"   ).arg(pos);
    case Binding::AxisPositionBeyondNegative: return String("-neg%1").arg(-pos);
    default:
        DENG2_ASSERT(!"B_AxisPositionToString: Unknown test");
        return "";
    }
}
String B_HatAngleToString(float pos) {
    return (pos < 0? "-center" : String("-angle") + String::number(pos));
}
String B_ConditionToString(Record const &cond) {
    String str;
    if (cond.geti("type") == Binding::GlobalState) {
        if (cond.getb("multiplayer")) {
            str += "multiplayer";
        }
    }
    else if (cond.geti("type") == Binding::ModifierState) {
        str += "modifier-" + String::number(cond.geti("id") - CTL_MODIFIER_1 + 1);
    }
    else {
        str += B_ControlDescToString(cond.geti("device"),
                                     (  cond.geti("type") == Binding::ButtonState? E_TOGGLE
                                      : cond.geti("type") == Binding::AxisState? E_AXIS
                                      : E_ANGLE), cond.geti("id"));
    }
    switch (cond.geti("type")) {
    case Binding::ButtonState:
    case Binding::ModifierState:
        str += B_ButtonStateToString(Binding::ControlTest(cond.geti("test")));
        break;
    case Binding::AxisState:
        str += B_AxisPositionToString(Binding::ControlTest(cond.geti("test")), cond.getf("pos"));
        break;
    case Binding::HatState:
        str += B_HatAngleToString(cond.getf("pos"));
        break;
    default: break;
    }
    if (cond.getb("negate")) {
        str += "-not";
    }
    return str;
}
String B_EventToString(ddevent_t const &ev) {
    String str = B_ControlDescToString(ev.device, ev.type,
                                       (  ev.type == E_TOGGLE  ? ev.toggle.id
                                        : ev.type == E_AXIS    ? ev.axis.id
                                        : ev.type == E_ANGLE   ? ev.angle.id
                                        : ev.type == E_SYMBOLIC? ev.symbolic.id
                                        : 0));
    switch (ev.type) {
    case E_TOGGLE:
        str += B_ButtonStateToString(  ev.toggle.state == ETOG_DOWN? Binding::ButtonStateDown
                                     : ev.toggle.state == ETOG_UP  ? Binding::ButtonStateUp
                                     : Binding::ButtonStateUp);
        break;
    case E_AXIS:
        str += B_AxisPositionToString((ev.axis.pos >= 0? Binding::AxisPositionBeyondPositive : Binding::AxisPositionBeyondNegative),
                                      ev.axis.pos);
        break;
    case E_ANGLE:    str += B_HatAngleToString(ev.angle.pos); break;
    case E_SYMBOLIC: str += "-" + String(ev.symbolic.name);        break;
    default: break;
    }
    return str;
}
struct keyname_t {
    int key;           ///< DDKEY
    char const *name;
};
static keyname_t const keyNames[] = {
    { DDKEY_PAUSE,       "pause" },
    { DDKEY_ESCAPE,      "escape" },
    { DDKEY_ESCAPE,      "esc" },
    { DDKEY_RIGHTARROW,  "right" },
    { DDKEY_LEFTARROW,   "left" },
    { DDKEY_UPARROW,     "up" },
    { DDKEY_DOWNARROW,   "down" },
    { DDKEY_RETURN,      "return" },
    { DDKEY_TAB,         "tab" },
    { DDKEY_RSHIFT,      "shift" },
    { DDKEY_RCTRL,       "ctrl" },
    { DDKEY_RCTRL,       "control" },
    { DDKEY_RALT,        "alt" },
    { DDKEY_INS,         "insert" },
    { DDKEY_INS,         "ins" },
    { DDKEY_DEL,         "delete" },
    { DDKEY_DEL,         "del" },
    { DDKEY_PGUP,        "pageup" },
    { DDKEY_PGUP,        "pgup" },
    { DDKEY_PGDN,        "pagedown" },
    { DDKEY_PGDN,        "pgdown" },
    { DDKEY_PGDN,        "pgdn" },
    { DDKEY_HOME,        "home" },
    { DDKEY_END,         "end" },
    { DDKEY_BACKSPACE,   "backspace" },
    { DDKEY_BACKSPACE,   "bkspc" },
    { '/',               "slash" },
    { DDKEY_BACKSLASH,   "backslash" },
    { '[',               "sqbracketleft" },
    { ']',               "sqbracketright" },
    { '+',               "plus" },
    { '-',               "minus" },
    { '+',               "plus" },
    { '=',               "equals" },
    { ' ',               "space" },
    { ';',               "semicolon" },
    { ',',               "comma" },
    { '.',               "period" },
    { '\'',              "apostrophe" },
    { DDKEY_F10,         "f10" },
    { DDKEY_F11,         "f11" },
    { DDKEY_F12,         "f12" },
    { DDKEY_F1,          "f1" },
    { DDKEY_F2,          "f2" },
    { DDKEY_F3,          "f3" },
    { DDKEY_F4,          "f4" },
    { DDKEY_F5,          "f5" },
    { DDKEY_F6,          "f6" },
    { DDKEY_F7,          "f7" },
    { DDKEY_F8,          "f8" },
    { DDKEY_F9,          "f9" },
    { '`',               "tilde" },
    { DDKEY_NUMLOCK,     "numlock" },
    { DDKEY_CAPSLOCK,    "capslock" },
    { DDKEY_SCROLL,      "scrlock" },
    { DDKEY_NUMPAD0,     "pad0" },
    { DDKEY_NUMPAD1,     "pad1" },
    { DDKEY_NUMPAD2,     "pad2" },
    { DDKEY_NUMPAD3,     "pad3" },
    { DDKEY_NUMPAD4,     "pad4" },
    { DDKEY_NUMPAD5,     "pad5" },
    { DDKEY_NUMPAD6,     "pad6" },
    { DDKEY_NUMPAD7,     "pad7" },
    { DDKEY_NUMPAD8,     "pad8" },
    { DDKEY_NUMPAD9,     "pad9" },
    { DDKEY_DECIMAL,     "decimal" },
    { DDKEY_DECIMAL,     "padcomma" },
    { DDKEY_SUBTRACT,    "padminus" },
    { DDKEY_ADD,         "padplus" },
    { DDKEY_PRINT,       "print" },
    { DDKEY_PRINT,       "prtsc" },
    { DDKEY_ENTER,       "enter" },
    { DDKEY_DIVIDE,      "divide" },
    { DDKEY_MULTIPLY,    "multiply" },
    { DDKEY_SECTION,     "section" },
    { DDKEY_WINMENU,     "winmenu" },
    { 0, nullptr}
};
char const *B_ShortNameForKey(int ddKey, bool forceLowercase) {
    static char nameBuffer[40];
    for (uint idx = 0; keyNames[idx].key; ++idx) {
        if (ddKey == keyNames[idx].key)
            return keyNames[idx].name;
    }
    if (ddKey < 127 && isprint(ddKey)) {
        nameBuffer[0] = forceLowercase? tolower(ddKey) : ddKey;
        nameBuffer[1] = 0;
        return nameBuffer;
    }
    return nullptr;
}
int B_KeyForShortName(char const *key) {
    DENG2_ASSERT(key);
    for (uint idx = 0; keyNames[idx].key; ++idx) {
        if (!qstricmp(key, keyNames[idx].name))
            return keyNames[idx].key;
    }
    if (qstrlen(key) == 1 && key[0] < 127 && isprint(key[0])) {
        return tolower(key[0]);
    }
    return 0;
}
static String const VAR_ID        ("id");
static String const VAR_TYPE      ("type");
static String const VAR_TEST      ("test");
static String const VAR_DEVICE_ID ("deviceId");
static String const VAR_CONTROL_ID("controlId");
DENG2_PIMPL(BindContext) {
    bool active  = false;  ///< @c true= Bindings are active.
    bool protect = false;  ///< @c true= Prevent explicit end user (de)activation.
    String name;           ///< Symbolic.
    typedef QSet<int> DeviceIds;
    DeviceIds acquireDevices;
    bool acquireAllDevices = false;  ///< @c true= will ignore @var acquireDevices.
    typedef QList<Record *> CommandBindings;
    CommandBindings commandBinds;
    typedef QList<CompiledImpulseBindingRecord *> ImpulseBindings;
    ImpulseBindings impulseBinds[DDMAXPLAYERS];  ///< Group bindings for each local player.
    DDFallbackResponderFunc ddFallbackResponder = nullptr;
    FallbackResponderFunc fallbackResponder     = nullptr;
    Impl(Public *i) : Base(i) {}
    /**
     * Look through the context for a binding that matches either of @a matchCmd or
     * @a matchImp.
     *
     * @param matchCmd   Command binding record to match, if any.
     * @param matchImp   Impulse binding record to match, if any.
     * @param cmdResult  The address of any matching command binding is written here.
     * @param impResult  The address of any matching impulse binding is written here.
     *
     * @return  @c true if a match is found.
     */
    bool findMatchingBinding(Record const *matchCmdRec, Record const *matchImpRec,
        Record **cmdResult, Record **impResult) const {
        DENG2_ASSERT(cmdResult && impResult);
        *cmdResult = nullptr;
        *impResult = nullptr;
        if (!matchCmdRec && !matchImpRec) return false;
        CommandBinding matchCmd;
        if (matchCmdRec) matchCmd = matchCmdRec;
        ImpulseBinding matchImp;
        if (matchImpRec) matchImp = matchImpRec;
        for (Record const *rec : commandBinds) {
            CommandBinding bind(*rec);
            if (matchCmd && matchCmd.geti(VAR_ID) != rec->geti(VAR_ID)) {
                if (matchCmd.geti(VAR_TYPE)       == bind.geti(VAR_TYPE) &&
                    matchCmd.geti(VAR_TEST)       == bind.geti(VAR_TEST) &&
                    matchCmd.geti(VAR_DEVICE_ID)  == bind.geti(VAR_DEVICE_ID) &&
                    matchCmd.geti(VAR_CONTROL_ID) == bind.geti(VAR_CONTROL_ID) &&
                    matchCmd.equalConditions(bind)) {
                    if (bind.geti(VAR_TYPE) != E_ANGLE ||
                        (matchCmd.geti("pos") == bind.geti("pos"))) // Hat angles must also match.
                    {
                        *cmdResult = const_cast<Record *>(rec);
                        return true;
                    }
                }
            }
            if (matchImp) {
                if (matchImp.geti(VAR_TYPE)       == bind.geti(VAR_TYPE) &&
                    matchImp.geti(VAR_DEVICE_ID)  == bind.geti(VAR_DEVICE_ID) &&
                    matchImp.geti(VAR_CONTROL_ID) == bind.geti(VAR_CONTROL_ID) &&
                    matchImp.equalConditions(bind)) {
                    *cmdResult = const_cast<Record *>(rec);
                    return true;
                }
            }
        }
        for (int i = 0; i < DDMAXPLAYERS; ++i)
        for (auto const *rec : impulseBinds[i]) {
            auto const &bind = rec->compiled();
            if (matchCmd) {
                if (matchCmd.geti(VAR_TYPE)       == int(bind.type) &&
                    matchCmd.geti(VAR_DEVICE_ID)  == bind.deviceId &&
                    matchCmd.geti(VAR_CONTROL_ID) == bind.controlId &&
                    matchCmd.equalConditions(ImpulseBinding(*rec))) {
                    *impResult = const_cast<CompiledImpulseBindingRecord *>(rec);
                    return true;
                }
            }
            if (matchImp && matchImp.geti(VAR_ID) != bind.id) {
                if (matchImp.geti(VAR_TYPE)       == int(bind.type) &&
                    matchImp.geti(VAR_DEVICE_ID)  == bind.deviceId &&
                    matchImp.geti(VAR_CONTROL_ID) == bind.controlId &&
                    matchImp.equalConditions(ImpulseBinding(*rec))) {
                    *impResult = const_cast<CompiledImpulseBindingRecord *>(rec);
                    return true;
                }
            }
        }
        return false;
    }
    /**
     * Delete all other bindings matching either @a cmdBinding or @a impBinding.
     */
    void deleteMatching(Record const *cmdBinding, Record const *impBinding) {
        Record *foundCmd = nullptr;
        Record *foundImp = nullptr;
        while (findMatchingBinding(cmdBinding, impBinding, &foundCmd, &foundImp)) {
            int bindId = (foundCmd? foundCmd->geti(VAR_ID) : (foundImp? foundImp->geti(VAR_ID) : 0));
            if (bindId) {
                LOG_INPUT_VERBOSE("Deleting binding %i, it has been overridden by binding %i")
                        << bindId << (cmdBinding? cmdBinding->geti(VAR_ID) : impBinding->geti(VAR_ID));
                self().deleteBinding(bindId);
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(ActiveChange)
    DENG2_PIMPL_AUDIENCE(AcquireDeviceChange)
    DENG2_PIMPL_AUDIENCE(BindingAddition)
};
DENG2_AUDIENCE_METHOD(BindContext, ActiveChange)
DENG2_AUDIENCE_METHOD(BindContext, AcquireDeviceChange)
DENG2_AUDIENCE_METHOD(BindContext, BindingAddition)
BindContext::BindContext(String const &name) : d(new Impl(this)) {
    setName(name);
}
BindContext::~BindContext() {
    clearAllBindings();
}
bool BindContext::isActive() const {
    return d->active;
}
bool BindContext::isProtected() const {
    return d->protect;
}
void BindContext::protect(bool yes) {
    d->protect = yes;
}
String BindContext::name() const {
    return d->name;
}
void BindContext::setName(String const &newName) {
    d->name = newName;
}
void BindContext::activate(bool yes) {
    if (d->active == yes) return;
    LOG_AS("BindContext");
    LOG_INPUT_VERBOSE("%s " _E(b) "'%s'") << (yes? "Activating" : "Deactivating") << d->name;
    d->active = yes;
    DENG2_FOR_AUDIENCE2(ActiveChange, i) i->bindContextActiveChanged(*this);
}
void BindContext::acquire(int deviceId, bool yes) {
    DENG2_ASSERT(deviceId >= 0 && deviceId < NUM_INPUT_DEVICES);
    int const countBefore = d->acquireDevices.count();
    if (yes) d->acquireDevices.insert(deviceId);
    else     d->acquireDevices.remove(deviceId);
    if (countBefore != d->acquireDevices.count()) {
        DENG2_FOR_AUDIENCE2(AcquireDeviceChange, i) i->bindContextAcquireDeviceChanged(*this);
    }
}
void BindContext::acquireAll(bool yes) {
    if (d->acquireAllDevices != yes) {
        d->acquireAllDevices = yes;
        DENG2_FOR_AUDIENCE2(AcquireDeviceChange, i) i->bindContextAcquireDeviceChanged(*this);
    }
}
bool BindContext::willAcquire(int deviceId) const {
    return d->acquireAllDevices || d->acquireDevices.contains(deviceId);
}
bool BindContext::willAcquireAll() const {
    return d->acquireAllDevices;
}
void BindContext::setDDFallbackResponder(DDFallbackResponderFunc newResponderFunc) {
    d->ddFallbackResponder = newResponderFunc;
}
void BindContext::setFallbackResponder(FallbackResponderFunc newResponderFunc) {
    d->fallbackResponder = newResponderFunc;
}
void BindContext::clearAllBindings() {
    LOG_AS("BindContext");
    qDeleteAll(d->commandBinds);
    d->commandBinds.clear();
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        qDeleteAll(d->impulseBinds[i]);
        d->impulseBinds[i].clear();
    }
    LOG_INPUT_VERBOSE(_E(b) "'%s'" _E(.) " cleared") << d->name;
}
void BindContext::clearBindingsForDevice(int deviceId) {
    QSet<int> ids;
    forAllCommandBindings([&ids, &deviceId] (Record const &bind) {
        if (bind.geti(VAR_DEVICE_ID) == deviceId) {
            ids.insert(bind.geti(VAR_ID));
        }
        return LoopContinue;
    });
    forAllImpulseBindings([&ids, &deviceId] (CompiledImpulseBindingRecord const &bind) {
        if (bind.compiled().deviceId == deviceId) {
            ids.insert(bind.geti(VAR_ID));
        }
        return LoopContinue;
    });
    foreach (int id, ids) {
        deleteBinding(id);
    }
}
Record *BindContext::bindCommand(char const *eventDesc, char const *command) {
    DENG2_ASSERT(eventDesc && command && command[0]);
    LOG_AS("BindContext");
    try {
        std::unique_ptr<Record> newBind(new Record);
        CommandBinding bind(*newBind.get());
        bind.configure(eventDesc, command); // Assign a new unique identifier.
        d->commandBinds.prepend(newBind.release());
        LOG_INPUT_VERBOSE("Command " _E(b) "\"%s\"" _E(.) " now bound to "
                          _E(b) "\"%s\"" _E(.) " in " _E(b) "'%s'" _E(.) " (id %i)")
                << command << bind.composeDescriptor() << d->name << bind.geti(VAR_ID);
        d->deleteMatching(&bind.def(), nullptr);
        DENG2_FOR_AUDIENCE2(BindingAddition, i) i->bindContextBindingAdded(*this, bind.def(), true/*is-command*/);
        Con_MarkAsChanged(true);
        return &bind.def();
    }
    catch (Binding::ConfigureError const &)
    {}
    return nullptr;
}
Record *BindContext::bindImpulse(char const *ctrlDesc, PlayerImpulse const &impulse, int localPlayer) {
    DENG2_ASSERT(ctrlDesc);
    DENG2_ASSERT(localPlayer >= 0 && localPlayer < DDMAXPLAYERS);
    LOG_AS("BindContext");
    try {
        std::unique_ptr<CompiledImpulseBindingRecord> newBind(new CompiledImpulseBindingRecord);
        ImpulseBinding bind(*newBind.get());
        bind.configure(ctrlDesc, impulse.id, localPlayer); // Assign a new unique identifier.
        d->impulseBinds[localPlayer].append(newBind.release());
        LOG_INPUT_VERBOSE("Impulse " _E(b) "'%s'" _E(.) " of player%i now bound to \"%s\" in "
                          _E(b) "'%s'" _E(.) " (id %i)")
                << impulse.name << (localPlayer + 1) << bind.composeDescriptor()
                << d->name << bind.def().compiled().id;
        d->deleteMatching(nullptr, &bind.def());
        DENG2_FOR_AUDIENCE2(BindingAddition, i) i->bindContextBindingAdded(*this, bind.def(), false/*is-impulse*/);
        Con_MarkAsChanged(true);
        return &bind.def();
    }
    catch (Binding::ConfigureError const &)
    {}
    return nullptr;
}
Record *BindContext::findCommandBinding(char const *command, int deviceId) const {
    if (command && command[0]) {
        for (Record const *rec : d->commandBinds) {
            CommandBinding bind(*rec);
            if (bind.gets(QStringLiteral("command")).compareWithoutCase(command)) continue;
            if ((deviceId < 0 || deviceId >= NUM_INPUT_DEVICES) || bind.geti(VAR_DEVICE_ID) == deviceId) {
                return const_cast<Record *>(rec);
            }
        }
    }
    return nullptr;
}
Record *BindContext::findImpulseBinding(int deviceId, ibcontroltype_t bindType, int controlId) const {
    for (int i = 0; i < DDMAXPLAYERS; ++i)
    for (auto const *rec : d->impulseBinds[i]) {
        auto const &bind = rec->compiled();
        if (bind.type      == bindType &&
            bind.deviceId  == deviceId &&
            bind.controlId == controlId) {
            return const_cast<CompiledImpulseBindingRecord *>(rec);
        }
    }
    return nullptr;
}
bool BindContext::deleteBinding(int id) {
    for (int i = 0; i < d->commandBinds.count(); ++i) {
        Record *rec = d->commandBinds.at(i);
        if (rec->geti(VAR_ID) == id) {
            d->commandBinds.removeAt(i);
            delete rec;
            return true;
        }
    }
    for (int i = 0; i < DDMAXPLAYERS; ++i)
    for (int k = 0; k < d->impulseBinds[i].count(); ++k) {
        auto *rec = d->impulseBinds[i].at(k);
        if (rec->compiled().id == id) {
            d->impulseBinds[i].removeAt(k);
            delete rec;
            return true;
        }
    }
    return false;
}
bool BindContext::tryEvent(ddevent_t const &event, bool respectHigherContexts) const {
    LOG_AS("BindContext");
    if (!isActive()) return false;
    if (event.type != E_FOCUS) {
        for (Record const *rec : d->commandBinds) {
            CommandBinding bind(*rec);
            AutoRef<Action> act(bind.makeAction(event, *this, respectHigherContexts));
            if (act.get()) {
                act->trigger();
                return true; // Triggered!
            }
        }
    }
    if (d->ddFallbackResponder) {
        if (d->ddFallbackResponder(&event))
            return true;
    }
    if (d->fallbackResponder) {
        event_t ev;
        if (/*bool const validGameEvent =*/ InputSystem::convertEvent(event, ev))
        {
            if (d->fallbackResponder(&ev))
                return true;
        }
    }
    return false;
}
LoopResult BindContext::forAllCommandBindings(
    std::function<de::LoopResult (Record &)> func) const {
    for (Record *rec : d->commandBinds) {
        if (auto result = func(*rec)) return result;
    }
    return LoopContinue;
}
LoopResult BindContext::forAllImpulseBindings(int localPlayer,
    std::function<de::LoopResult (CompiledImpulseBindingRecord &)> func) const {
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        if ((localPlayer < 0 || localPlayer >= DDMAXPLAYERS) || localPlayer == i) {
            for (auto *rec : d->impulseBinds[i]) {
                if (auto result = func(*rec)) return result;
            }
        }
    }
    return LoopContinue;
}
int BindContext::commandBindingCount() const {
    return d.getConst()->commandBinds.count();
}
int BindContext::impulseBindingCount(int localPlayer) const {
    int count = 0;
    for (int i = 0; i < DDMAXPLAYERS; ++i) {
        if ((localPlayer < 0 || localPlayer >= DDMAXPLAYERS) || localPlayer == i) {
            count += d.getConst()->impulseBinds[i].count();
        }
    }
    return count;
}
static int idCounter = 0;
static String const VAR_CONDITION("condition");
Record &Binding::def() {
    return const_cast<Record &>(accessedRecord());
}
Record const &Binding::def() const {
    return accessedRecord();
}
Binding::operator bool() const {
    return accessedRecordPtr() != 0;
}
void Binding::resetToDefaults() {
    def().addNumber("id", 0);  ///< Unique identifier.
    def().addArray(VAR_CONDITION, new ArrayValue);
}
Binding::CompiledConditionRecord &Binding::addCondition() {
    auto *cond = new CompiledConditionRecord;
    cond->addNumber("type", Invalid);
    cond->addNumber("test", None);
    cond->addNumber("device", -1);
    cond->addNumber("id", -1);
    cond->addNumber("pos", 0);
    cond->addBoolean("negate", false);
    cond->addBoolean("multiplayer", false);
    def()[VAR_CONDITION].array().add(new RecordValue(cond, RecordValue::OwnsRecord));
    cond->resetCompiled();
    return *cond;
}
int Binding::conditionCount() const {
    return int(geta(VAR_CONDITION).size());
}
bool Binding::hasCondition(int index) const {
    return index >= 0 && index < conditionCount();
}
Binding::CompiledConditionRecord &Binding::condition(int index) {
    return *static_cast<CompiledConditionRecord *>(def().geta(VAR_CONDITION)[index].as<RecordValue>().record());
}
Binding::CompiledConditionRecord const &Binding::condition(int index) const {
    return *static_cast<CompiledConditionRecord const *>(geta(VAR_CONDITION)[index].as<RecordValue>().record());
}
bool Binding::equalConditions(Binding const &other) const {
    if (def()[VAR_CONDITION].array().elements().count() != other.geta(VAR_CONDITION).elements().count()) {
        return false;
    }
    ArrayValue const &conds = def().geta(VAR_CONDITION);
    DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds.elements()) {
        auto const &a = *static_cast<CompiledConditionRecord const *>((*i)->as<RecordValue>().record());
        bool found = false;
        ArrayValue const &conds2 = other.geta(VAR_CONDITION);
        DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds2.elements()) {
            auto const &b = *static_cast<CompiledConditionRecord const *>((*i)->as<RecordValue>().record());
            if (a.compiled() == b.compiled()) {
                found = true;
                break;
            }
        }
        if (!found) return false;
    }
    return true;
}
void Binding::resetIdentifiers() // static
{
    idCounter = 0;
}
int Binding::newIdentifier() // static
{
    int id = 0;
    while (!id) { id = ++idCounter; }
    return id;
}
Binding::CompiledCondition::CompiledCondition(Record const &rec)
    : type(ConditionType(rec.geti("type")))
    , test(ControlTest(rec.geti("test")))
    , device(rec.geti("device"))
    , id(rec.geti("id"))
    , pos(rec.getf("pos"))
    , negate(rec.getb("negate"))
    , multiplayer(rec.getb("multiplayer"))
{}
bool Binding::CompiledCondition::operator == (CompiledCondition const &other) const {
    return (type        == other.type &&
            test        == other.test &&
            device      == other.device &&
            id          == other.id &&
            de::fequal(pos, other.pos) &&
            negate      == other.negate &&
            multiplayer == other.multiplayer);
}
static void releaseScreenshotTexture() {
    ClientWindow::main().busy().releaseTransitionFrame();
}
void BusyVisual_PrepareResources() {
    BusyTask *task = DoomsdayApp::app().busyMode().currentTask();
    if (!task) return;
    if (task->mode & BUSYF_STARTUP) {
        releaseScreenshotTexture();
    }
}
using namespace de;
#define DOOMWIPESINE_NUMSAMPLES 320
static void seedDoomWipeSine(void);
int rTransition = (int) TS_CROSSFADE; ///< cvar Default transition style.
int rTransitionTics = 28; ///< cvar Default transition duration (in tics).
typedef struct {
    dd_bool inProgress; /// @c true= a transition is presently being animated.
    transitionstyle_t style; /// Style of transition (cross-fade, wipe, etc...).
    uint startTime; /// Time at the moment the transition began (in 35hz tics).
    uint tics; /// Time duration of the animation.
    float position; /// Animation interpolation point [0..1].
} transitionstate_t;
static transitionstate_t transition;
static float doomWipeSine[DOOMWIPESINE_NUMSAMPLES];
static float doomWipeSamples[SCREENWIDTH+1];
void Con_TransitionRegister(void) {
    C_VAR_INT("con-transition", &rTransition, 0, FIRST_TRANSITIONSTYLE, LAST_TRANSITIONSTYLE);
    C_VAR_INT("con-transition-tics", &rTransitionTics, 0, 0, 60);
}
void Con_TransitionConfigure(void) {
    transition.tics = rTransitionTics;
    transition.style = (transitionstyle_t) rTransition;
    if (transition.style == TS_DOOM || transition.style == TS_DOOMSMOOTH) {
        seedDoomWipeSine();
    }
    transition.inProgress = true;
}
void Con_TransitionBegin(void) {
    transition.startTime = Timer_Ticks();
    transition.position = 0;
}
dd_bool Con_TransitionInProgress(void) {
    return transition.inProgress;
}
static void Con_EndTransition(void) {
    if (!transition.inProgress) return;
    releaseScreenshotTexture();
    transition.inProgress = false;
}
void Con_TransitionTicker(timespan_t /*ticLength*/)
{
    if (isDedicated) return;
    if (!Con_TransitionInProgress()) return;
    transition.position = (float)(Timer_Ticks() - transition.startTime) / transition.tics;
    if (transition.position >= 1) {
        Con_EndTransition();
    }
}
static void seedDoomWipeSine(void) {
    int i;
    doomWipeSine[0] = (128 - RNG_RandByte()) / 128.f;
    for (i = 1; i < DOOMWIPESINE_NUMSAMPLES; ++i) {
        float r = (128 - RNG_RandByte()) / 512.f;
        doomWipeSine[i] = MINMAX_OF(-1, doomWipeSine[i-1] + r, 1);
    }
}
static float sampleDoomWipeSine(float point) {
    float sample = doomWipeSine[(uint)ROUND((DOOMWIPESINE_NUMSAMPLES-1) * MINMAX_OF(0, point, 1))];
    float offset = SCREENHEIGHT * transition.position * transition.position;
    return offset + (SCREENHEIGHT/2) * (transition.position + sample) * transition.position * transition.position;
}
static void sampleDoomWipe(void) {
    int i;
    for (i = 0; i <= SCREENWIDTH; ++i) {
        doomWipeSamples[i] = MAX_OF(0, sampleDoomWipeSine((float) i / SCREENWIDTH));
    }
}
void Con_DrawTransition(void) {
    if (isDedicated) return;
    if (!Con_TransitionInProgress()) return;
    DENG2_ASSERT_IN_RENDER_THREAD();
    DENG_ASSERT_GL_CONTEXT_ACTIVE();
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadIdentity();
    DGL_Ortho(0, 0, SCREENWIDTH, SCREENHEIGHT, -1, 1);
    DENG2_ASSERT(ClientWindow::main().busy().transitionFrame() != 0);
    GLuint const texScreenshot = ClientWindow::main().busy().transitionFrame()->glName();
    GL_BindTextureUnmanaged(texScreenshot, gl::ClampToEdge, gl::ClampToEdge);
    LIBGUI_GL.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    DGL_PushState();
    DGL_Disable(DGL_ALPHA_TEST);
    DGL_Enable(DGL_TEXTURE_2D);
    switch (transition.style) {
    case TS_DOOMSMOOTH: {
        float topAlpha, s, div, colWidth = 1.f / SCREENWIDTH;
        int x, y, h, i;
        sampleDoomWipe();
        div = 1 - .25f * transition.position;
        topAlpha = 1 - transition.position;
        topAlpha *= topAlpha;
        h = SCREENHEIGHT * (1 - div);
        x = 0;
        s = 0;
        DGL_Begin(DGL_TRIANGLE_STRIP);
        for (i = 0; i <= SCREENWIDTH; ++i, x++, s += colWidth) {
            y = doomWipeSamples[i];
            DGL_Color4f(1, 1, 1, 1);
            DGL_TexCoord2f(0, s, div); DGL_Vertex2f(x, y + h);
            DGL_Color4f(1, 1, 1, topAlpha);
            DGL_TexCoord2f(0, s, 1); DGL_Vertex2f(x, y);
        }
        DGL_End();
        x = 0;
        s = 0;
        DGL_Color4f(1, 1, 1, 1);
        DGL_Begin(DGL_TRIANGLE_STRIP);
        for (i = 0; i <= SCREENWIDTH; ++i, x++, s += colWidth) {
            y = doomWipeSamples[i] + h;
            DGL_TexCoord2f(0, s, 0);   DGL_Vertex2f(x, y + (SCREENHEIGHT - h));
            DGL_TexCoord2f(0, s, div); DGL_Vertex2f(x, y);
        }
        DGL_End();
        break;
      }
    case TS_DOOM: {
        float s = 0, colWidth = 1.f / SCREENWIDTH;
        int x = 0, y, i;
        sampleDoomWipe();
        DGL_Color4f(1, 1, 1, 1);
        DGL_Begin(DGL_QUADS);
        for (i = 0; i <= SCREENWIDTH; ++i, x++, s+= colWidth) {
            y = doomWipeSamples[i];
            DGL_TexCoord2f(0, s, 1); DGL_Vertex2f(x, y);
            DGL_TexCoord2f(0, s+colWidth, 1); DGL_Vertex2f(x+1, y);
            DGL_TexCoord2f(0, s+colWidth, 0); DGL_Vertex2f(x+1, y+SCREENHEIGHT);
            DGL_TexCoord2f(0, s, 0); DGL_Vertex2f(x, y+SCREENHEIGHT);
        }
        DGL_End();
        break;
      }
    case TS_CROSSFADE:
        DGL_Color4f(1, 1, 1, 1 - transition.position);
        DGL_Begin(DGL_QUADS);
            DGL_TexCoord2f(0, 0, 1); DGL_Vertex2f(0, 0);
            DGL_TexCoord2f(0, 0, 0); DGL_Vertex2f(0, SCREENHEIGHT);
            DGL_TexCoord2f(0, 1, 0); DGL_Vertex2f(SCREENWIDTH, SCREENHEIGHT);
            DGL_TexCoord2f(0, 1, 1); DGL_Vertex2f(SCREENWIDTH, 0);
        DGL_End();
        break;
    }
    GL_SetNoTexture();
    DGL_PopState();
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
}
ButtonInputControl::ButtonInputControl(String const &name) {
    setName(name);
}
ButtonInputControl::~ButtonInputControl()
{}
bool ButtonInputControl::isDown() const {
    return _isDown;
}
void ButtonInputControl::setDown(bool yes) {
    bool const oldDown = _isDown;
    _isDown = yes;
    if (_isDown != oldDown) {
        _time = Timer_RealMilliseconds();
    }
    if (_isDown) {
        setBindContextAssociation(Triggered);
    }
    else {
        setBindContextAssociation(Expired, UnsetFlags);
    }
}
String ButtonInputControl::description() const {
    return String(_E(b) "%1 " _E(.) "(Button)").arg(fullName());
}
bool ButtonInputControl::inDefaultState() const {
    return !_isDown; // Not depressed?
}
void ButtonInputControl::reset() {
    setBindContextAssociation(Triggered | Expired, UnsetFlags);
    _isDown = false;
    _time   = 0;
}
duint ButtonInputControl::time() const {
    return _time;
}
DENG2_PIMPL_NOREF(ClientRootWidget) {
    PackageIconBank packageIconBank;
};
ClientRootWidget::ClientRootWidget(GLWindow *window)
    : GuiRootWidget(window)
    , d(new Impl) {
    d->packageIconBank.setDisplaySize(GuiWidget::pointsToPixels(PackageIconBank::Size(32, 32)));
}
ClientWindow &ClientRootWidget::window() {
    return GuiRootWidget::window().as<ClientWindow>();
}
PackageIconBank &ClientRootWidget::packageIconBank() {
    return d->packageIconBank;
}
void ClientRootWidget::update() {
    if (!DoomsdayApp::app().isShuttingDown()) {
        GuiRootWidget::update();
        if (window().isGLReady() && !d->packageIconBank.atlas()) {
            d->packageIconBank.setAtlas(&atlas());
        }
    }
}
void ClientRootWidget::addOnTop(GuiWidget *widget) {
    window().addOnTop(widget);
}
void ClientRootWidget::dispatchLatestMousePosition() {
    ClientApp::windowSystem().dispatchLatestMousePosition();
}
void ClientRootWidget::handleEventAsFallback(Event const &event) {
    ClientApp::inputSystem().tryEvent(event, "global");
}
DENG2_PIMPL_NOREF(ClientStyle) {
    struct EmptyMenuLabelStylist : public ui::Stylist {
        void applyStyle(GuiWidget &widget) override {
            LabelWidget &label = widget.as<LabelWidget>();
            label.setFont("menu.empty");
            label.setOpacity(0.5f);
        }
    }
    emptyMenuLabelStylist;
};
ClientStyle::ClientStyle() : d(new Impl)
{}
GuiWidget *ClientStyle::sharedBlurWidget() const {
    if (!ClientWindow::mainExists()) return nullptr;
    return &ClientWindow::main().taskBarBlur();
}
ui::Stylist &ClientStyle::emptyMenuLabelStylist() const {
    return d->emptyMenuLabelStylist;
}
void ClientStyle::performUpdate() {
    async([]() {
        LOG_MSG("Waiting for UI assets to be ready...");
        ClientWindow::main().root().waitForAssetsReady();
        return 0;
    },
    [this](int) {
        LOG_MSG("Updating the UI style");
        Style::performUpdate();
    });
}
DENG2_PIMPL(ClientWindow)
, DENG2_OBSERVES(App, StartupComplete)
, DENG2_OBSERVES(DoomsdayApp, GameChange)
, DENG2_OBSERVES(MouseEventSource, MouseStateChange)
, DENG2_OBSERVES(WindowEventHandler, FocusChange)
, DENG2_OBSERVES(GLWindow, Init)
, DENG2_OBSERVES(GLWindow, Resize)
, DENG2_OBSERVES(GLWindow, Swap)
, DENG2_OBSERVES(Variable, Change)
, DENG2_OBSERVES(FileSystem, Busy)
#if !defined (DENG_MOBILE)
, DENG2_OBSERVES(PersistentGLWindow, AttributeChange)
#endif
{
    bool needMainInit            = true;
    bool needRootSizeUpdate      = false;
    Mode mode = Normal;
    ClientRootWidget root;
    GameWidget *game = nullptr;
    LabelWidget *nowPlaying = nullptr;
    TaskBarWidget *taskBar = nullptr;
    LabelWidget *taskBarBlur = nullptr; ///< Blur everything below the task bar.
    NotificationAreaWidget *notifications = nullptr;
    ButtonWidget *quitButton = nullptr;
    AlertDialog *alerts = nullptr;
    ColorAdjustmentDialog *colorAdjust = nullptr;
    HomeWidget *home = nullptr;
    SafeWidgetPtr<FadeToBlackWidget> fader;
    BusyWidget *busy = nullptr;
    GuiWidget *sidebar = nullptr;
    PrivilegedLogWidget *privLog = nullptr;
    LabelWidget *cursor = nullptr;
    ConstantRule *cursorX;
    ConstantRule *cursorY;
    AnimationRule *gameWidth;
    AnimationRule *gameHeight;
    AnimationRule *homeDelta;
    AnimationRule *quitX;
    bool cursorHasBeenHidden = false;
    bool isGameMini = false;
    UniqueWidgetPtr<LabelWidget> fpsCounter;
    float oldFps = 0;
    UniqueWidgetPtr<ProgressWidget> fsBusy;
    VRWindowTransform contentXf;
    Impl(Public *i)
        : Base      (i)
        , root      (i)
        , cursorX   (new ConstantRule(0))
        , cursorY   (new ConstantRule(0))
        , gameWidth (new AnimationRule(0, Animation::EaseBoth))
        , gameHeight(new AnimationRule(0, Animation::EaseBoth))
        , homeDelta (new AnimationRule(0, Animation::EaseBoth))
        , quitX     (new AnimationRule(0, Animation::EaseBoth))
        , contentXf (*i) {
        self().setTransform(contentXf);
        DoomsdayApp::app().audienceForGameChange() += this;
        App::app().audienceForStartupComplete() += this;
        self().eventHandler().audienceForMouseStateChange() += this;
        foreach (String s, configVariableNames()) {
            App::config(s).audienceForChange() += this;
        }
    }
    ~Impl() override {
        releaseRef(cursorX);
        releaseRef(cursorY);
        releaseRef(gameWidth);
        releaseRef(gameHeight);
        releaseRef(homeDelta);
        releaseRef(quitX);
    }
    StringList configVariableNames() const {
#if !defined (DENG_MOBILE)
        return StringList()
                << self().configName("fsaa")
                << self().configName("vsync");
#else
        return StringList();
#endif
    }
    void setupUI() {
        Style &style = ClientApp::windowSystem().style();
        taskBarBlur = new LabelWidget("taskbar-blur");
        gameWidth ->set(root.viewWidth());
        gameHeight->set(root.viewHeight());
        game = new GameWidget;
        game->rule().setInput(Rule::Left,   root.viewLeft())
                    .setInput(Rule::Right,  root.viewLeft() + *gameWidth)
                    .setInput(Rule::Bottom, root.viewBottom())
                    .setInput(Rule::Height, *gameHeight);
        game->disable();
        root.add(game);
        /*gameUI = new GameUIWidget;
        gameUI->rule().setRect(game->rule());
        gameUI->disable();
        root.add(gameUI);*/
        auto *miniGameControls = new LabelWidget; {
            miniGameControls->set(GuiWidget::Background(style.colors().colorf("background")));
            miniGameControls->rule()
                    .setInput(Rule::Width,  root.viewWidth()/2 -   style.rules().rule(RuleBank::UNIT))
                    .setInput(Rule::Left,   game->rule().right() + style.rules().rule(RuleBank::UNIT))
                    .setInput(Rule::Top,    game->rule().top());
            ButtonWidget *backToGame = new ButtonWidget;
            backToGame->setText("Back to Game");
            backToGame->setSizePolicy(ui::Expand, ui::Expand);
            backToGame->setActionFn([this] () { home->moveOffscreen(1.0); });
            backToGame->setColorTheme(GuiWidget::Inverted);
            nowPlaying = new LabelWidget;
            nowPlaying->setSizePolicy(ui::Expand, ui::Expand);
            nowPlaying->margins()/*.setLeftRight("")*/.setBottom("dialog.gap");
            nowPlaying->setMaximumTextWidth(miniGameControls->rule().width());
            nowPlaying->setTextLineAlignment(ui::AlignLeft);
            nowPlaying->setTextColor("text");
            nowPlaying->setFont("heading");
            AutoRef<Rule> combined = nowPlaying->rule().height() + backToGame->rule().height();
            nowPlaying->rule()
                    .setMidAnchorX(miniGameControls->rule().midX())
                    .setInput(Rule::Top, miniGameControls->rule().midY() - combined/2);
            backToGame->rule()
                    .setInput(Rule::Left, nowPlaying->rule().left())
                    .setInput(Rule::Top,  nowPlaying->rule().bottom());
            miniGameControls->add(nowPlaying);
            miniGameControls->add(backToGame);
            root.add(miniGameControls);
        }
        busy = new BusyWidget;
        busy->setGameWidget(*game);
        busy->hide(); // normally hidden
        busy->rule().setRect(root.viewRule());
        root.add(busy);
        home = new HomeWidget;
        home->rule().setInput(Rule::Left,   root.viewLeft())
                    .setInput(Rule::Right,  root.viewRight())
                    .setInput(Rule::Top,    root.viewTop())
                    .setInput(Rule::Bottom, root.viewBottom() + *homeDelta);
        root.add(home);
#if !defined (DENG_MOBILE)
        busy->progress().orphan();
        busy->progress().rule().setRect(game->rule());
        root.add(&busy->progress());
#endif
        notifications = new NotificationAreaWidget;
        notifications->useDefaultPlacement(root.viewRule(), *quitX);
        root.add(notifications);
        quitButton = new ButtonWidget;
        quitButton->setBehavior(Widget::Focusable, false);
        quitButton->setText(_E(b)_E(D) + tr("QUIT"));
        quitButton->setFont("small");
        quitButton->setSizePolicy(ui::Expand, ui::Fixed);
        quitButton->setStyleImage("close.ringless", "default");
        quitButton->setImageColor(style.colors().colorf("accent"));
        quitButton->setTextAlignment(ui::AlignLeft);
        quitButton->set(GuiWidget::Background(style.colors().colorf("background")));
        quitButton->setActionFn([]() { Con_Execute(CMDS_DDAY, "quit!", false, false); });
        quitButton->rule()
                .setInput(Rule::Top,    root.viewTop() + style.rules().rule("gap"))
                .setInput(Rule::Left,   root.viewRight() + *quitX)
                .setInput(Rule::Height, style.fonts().font("default").height() +
                          style.rules().rule("gap") * 2);
        quitButton->setOpacity(0);
        root.add(quitButton);
        alerts = new AlertDialog;
        root.add(alerts);
        fpsCounter.reset(new LabelWidget);
        fpsCounter->setSizePolicy(ui::Expand, ui::Expand);
        fpsCounter->setAlignment(ui::AlignRight);
        fsBusy.reset(new ProgressWidget);
        fsBusy->setSizePolicy(ui::Expand, ui::Expand);
        fsBusy->setMode(ProgressWidget::Indefinite);
        fsBusy->useMiniStyle();
        fsBusy->setText("Files");
        fsBusy->setTextAlignment(ui::AlignRight);
        if (style.isBlurringAllowed()) {
            taskBarBlur->set(GuiWidget::Background(Vector4f(1, 1, 1, 1), GuiWidget::Background::Blurred));
        }
        taskBarBlur->rule().setRect(root.viewRule());
        taskBarBlur->setAttribute(GuiWidget::DontDrawContent);
        root.add(taskBarBlur);
        taskBar = new TaskBarWidget;
        taskBar->rule()
                .setInput(Rule::Left,   root.viewLeft())
                .setInput(Rule::Bottom, root.viewBottom() + taskBar->shift())
                .setInput(Rule::Width,  root.viewWidth());
        root.add(taskBar);
        miniGameControls->rule().setInput(Rule::Bottom, taskBar->rule().top());
        privLog = new PrivilegedLogWidget;
        privLog->rule().setRect(root.viewRule());
        root.add(privLog);
        colorAdjust = new ColorAdjustmentDialog;
        colorAdjust->setAnchor(root.viewWidth() / 2, root.viewTop());
        colorAdjust->setOpeningDirection(ui::Down);
        root.add(colorAdjust);
        taskBar->hide();
        cursor = new LabelWidget;
        cursor->setBehavior(Widget::Unhittable);
        cursor->margins().set(""); // no margins
        cursor->setImage(style.images().image("window.cursor"));
        cursor->setAlignment(ui::AlignTopLeft);
        cursor->rule()
                .setSize(Const(48), Const(48))
                .setLeftTop(*cursorX, *cursorY);
        cursor->hide();
        root.add(cursor);
        /*
#ifdef DENG2_DEBUG
        auto debugPrint = [] (Widget &w) {
            int depth = 0;
            for (auto *p = w.parent(); p; p = p->parent()) ++depth;
            qDebug() << QByteArray(4 * depth, ' ').constData()
                     << (w.parent()? w.parent()->children().indexOf(&w) : -1)
                     << w.name() << "childCount:" << w.childCount();
        };
        root.walkInOrder(Widget::Forward, [&debugPrint] (Widget &w) {
            debugPrint(w);
            return LoopContinue;
        });
        qDebug() << "AND BACKWARDS:";
        root.children().last()->walkInOrder(Widget::Backward, [&debugPrint] (Widget &w) {
            debugPrint(w);
            return LoopContinue;
        });
#endif
        */
#if !defined (DENG_MOBILE)
        self().audienceForAttributeChange() += this;
#endif
    }
    void fileSystemBusyStatusChanged(FS::BusyStatus bs) override {
        notifications->showOrHide(*fsBusy, bs == FS::Busy);
    }
    void appStartupCompleted() override {
        taskBar->show();
        if (!App::config().getb("tutorial.shown", false)) {
            App::config().set("tutorial.shown", true);
            LOG_NOTE("Starting tutorial (not shown before)");
            QTimer::singleShot(500, taskBar, SLOT(showTutorial()));
        }
        else {
            taskBar->close();
        }
        FS::get().audienceForBusy() += this;
        Loop::get().timer(1.0, [this] () { showOrHideQuitButton(); });
    }
    void currentGameChanged(Game const &newGame) override {
        minimizeGame(false);
        showOrHideQuitButton();
        if (!newGame.isNull()) {
            nowPlaying->setText(_E(l) + tr("Now playing") + "\n" + _E(b) +
                                DoomsdayApp::currentGameProfile()->name());
            root.clearFocusStack();
            root.setFocus(nullptr);
        }
        else // Back to Home.
        {
            busy->releaseTransitionFrame();
        }
        taskBar->console().enableBlur(taskBar->style().isBlurringAllowed());
        self().hideTaskBarBlur(); // update background blur mode
        activateOculusRiftModeIfConnected();
    }
    void activateOculusRiftModeIfConnected() {
        if (vrCfg().oculusRift().isHMDConnected() && vrCfg().mode() != VRConfig::OculusRift) {
            LOG_NOTE("HMD connected, automatically switching to Oculus Rift mode");
            Con_SetInteger("rend-vr-mode", VRConfig::OculusRift);
            vrCfg().oculusRift().moveWindowToScreen(OculusRift::HMDScreen);
        }
        else if (!vrCfg().oculusRift().isHMDConnected() && vrCfg().mode() == VRConfig::OculusRift) {
            LOG_NOTE("HMD not connected, disabling VR mode");
            Con_SetInteger("rend-vr-mode", VRConfig::Mono);
            vrCfg().oculusRift().moveWindowToScreen(OculusRift::DefaultScreen);
        }
    }
    void setMode(Mode const &newMode) {
        LOG_DEBUG("Switching to %s mode") << (newMode == Busy? "Busy" : "Normal");
        switch (newMode) {
        case Busy:
            game->hide();
            game->disable();
            taskBar->disable();
            quitButton->setOpacity(0, 1);
            busy->show();
            busy->enable();
            break;
        case Normal:
#if defined (DENG_HAVE_BUSYRUNNER)
            busy->disable();
#else
            busy->hide();
#endif
            game->show();
            game->enable();
            taskBar->enable();
            quitButton->setOpacity(1, 1);
            break;
        }
        mode = newMode;
    }
    void windowInit(GLWindow &) override {
        Sys_GLConfigureDefaultState();
        GL_Init2DState();
        LOGDEV_GL_MSG("GL feature: Multisampling: %b") << (GLTextureFramebuffer::defaultMultisampling() > 1);
        if (vrCfg().needsStereoGLFormat() && !self().format().stereo()) {
            LOG_GL_WARNING("Current VR mode needs a stereo buffer, but it isn't supported");
        }
        game->enable();
        GLState::current().setViewport(Rectangleui(0, 0, self().pixelWidth(), self().pixelHeight()));
        LOG_DEBUG("GameWidget enabled");
        if (needMainInit) {
            needMainInit = false;
#if !defined (DENG_MOBILE)
            self().raise();
            self().requestActivate();
#endif
            self().eventHandler().audienceForFocusChange() += this;
            DD_FinishInitializationAfterWindowReady();
            vrCfg().oculusRift().glPreInit();
            contentXf.glInit();
        }
    }
    void windowResized(GLWindow &) override {
        LOG_AS("ClientWindow");
        Size size = self().pixelSize();
        LOG_TRACE("Window resized to %s pixels", size.asText());
        GLState::current().setViewport(Rectangleui(0, 0, size.x, size.y));
        updateRootSize();
        showOrHideQuitButton();
    }
    void windowSwapped(GLWindow &) override {
        if (DoomsdayApp::app().isShuttingDown()) return;
        ClientApp::app().postFrame(); /// @todo what about multiwindow?
        updateFpsNotification(self().frameRate());
        completeFade();
    }
#if !defined (DENG_MOBILE)
    void windowAttributesChanged(PersistentGLWindow &) override {
        showOrHideQuitButton();
    }
#endif
    void showOrHideQuitButton() {
        TimeSpan const SPAN = 0.6;
        if (self().isFullScreen() && !DoomsdayApp::isGameLoaded()) {
            quitX->set(-quitButton->rule().width() - Style::get().rules().rule("gap"), SPAN);
        }
        else {
            quitX->set(0, SPAN);
        }
    }
    void mouseStateChanged(MouseEventSource::State state) override {
        Mouse_Trap(state == MouseEventSource::Trapped);
    }
    /**
     * Handles an event that BaseWindow (and thus WindowSystem) didn't have use for.
     *
     * @param event  Event to handle.
     */
    bool handleFallbackEvent(Event const &ev) {
        if (MouseEvent const *mouse = maybeAs<MouseEvent>(ev)) {
            switch (ev.type()) {
            case Event::MouseButton:
                if (game->hitTest(ev)) {
                    Mouse_Qt_SubmitButton(
                                mouse->button() == MouseEvent::Left?     IMB_LEFT   :
                                mouse->button() == MouseEvent::Middle?   IMB_MIDDLE :
                                mouse->button() == MouseEvent::Right?    IMB_RIGHT  :
                                mouse->button() == MouseEvent::XButton1? IMB_EXTRA1 :
                                mouse->button() == MouseEvent::XButton2? IMB_EXTRA2 : IMB_MAXBUTTONS,
                                mouse->state() == MouseEvent::Pressed);
                    return true;
                }
                break;
            case Event::MouseMotion:
                Mouse_Qt_SubmitMotion(IMA_POINTER, mouse->pos().x, mouse->pos().y);
                return true;
            case Event::MouseWheel:
                if (game->hitTest(ev) && mouse->wheelMotion() == MouseEvent::Step) {
                    Mouse_Qt_SubmitMotion(IMA_WHEEL, mouse->wheel().x, mouse->wheel().y);
                    return true;
                }
                break;
            default:
                break;
            }
        }
        return false;
    }
    void windowFocusChanged(GLWindow &, bool hasFocus) override {
        LOG_DEBUG("windowFocusChanged focus:%b fullscreen:%b hidden:%b minimized:%b")
                << hasFocus << self().isFullScreen() << self().isHidden() << self().isMinimized();
        if (!hasFocus) {
            InputSystem::get().forAllDevices([] (InputDevice &device) {
                device.reset();
                return LoopContinue;
            });
            InputSystem::get().clearEvents();
            self().eventHandler().trapMouse(false);
        }
        else if (self().isFullScreen() && !taskBar->isOpen() && DoomsdayApp::isGameLoaded()) {
            self().eventHandler().trapMouse();
        }
        if (Config::get().getb("audio.pauseOnFocus", true)) {
            AudioSystem::get().pauseMusic(!hasFocus);
        }
        ddevent_t ev; de::zap(ev);
        ev.device         = -1;
        ev.type           = E_FOCUS;
        ev.focus.gained   = hasFocus;
        ev.focus.inWindow = 1;         /// @todo Ask WindowSystem for an identifier number.
        InputSystem::get().postEvent(&ev);
    }
    void updateFpsNotification(float fps) {
        notifications->showOrHide(*fpsCounter, self().isFPSCounterVisible());
        if (!fequal(oldFps, fps)) {
            fpsCounter->setText(QString("%1 " _E(l) + tr("FPS")).arg(fps, 0, 'f', 1));
            oldFps = fps;
        }
    }
    void variableValueChanged(Variable &variable, Value const &newValue) override {
        if (variable.name() == "fsaa") {
        }
        else if (variable.name() == "vsync") {
/*#ifdef WIN32
            self().updateCanvasFormat();
            DENG2_UNUSED(newValue);
#else*/
            GL_SetVSync(newValue.isTrue());
        }
    }
    void installSidebar(SidebarLocation location, GuiWidget *widget) {
        if (sidebar) {
            uninstallSidebar(location);
        }
        if (!widget) return;
        home->moveOffscreen(1.0);
        DENG2_ASSERT(sidebar == NULL);
        switch (location) {
        case RightEdge:
            widget->rule()
                    .setInput(Rule::Top,    root.viewTop())
                    .setInput(Rule::Right,  root.viewRight())
                    .setInput(Rule::Bottom, taskBar->rule().top());
            game->rule()
                    .setInput(Rule::Right,  widget->rule().left());
            notifications->rule()
                    .setInput(Rule::Right,  widget->rule().left() - Style::get().rules().rule("gap"));
            /*gameUI->rule()
                    .setInput(Rule::Right,  widget->rule().left());*/
            break;
        }
        sidebar = widget;
        root.insertBefore(sidebar, *notifications);
    }
    void uninstallSidebar(SidebarLocation location) {
        DENG2_ASSERT(sidebar != NULL);
        switch (location) {
        case RightEdge:
            game->rule().setInput(Rule::Right, root.viewLeft() + *gameWidth);
            notifications->useDefaultPlacement(root.viewRule(), *quitX);
            break;
        }
        root.remove(*sidebar);
        sidebar->guiDeleteLater();
        sidebar = 0;
    }
    void updateRootSize() {
        DENG_ASSERT_IN_MAIN_THREAD();
        needRootSizeUpdate = false;
        Vector2ui const size = contentXf.logicalRootSize(self().pixelSize());
        root.setViewSize(size);
    }
    void minimizeGame(bool mini) {
        TimeSpan const SPAN = 1.0;
        if (mini && !isGameMini) {
            self().setSidebar(RightEdge, nullptr);
            auto const &unit = Style::get().rules().rule(RuleBank::UNIT);
            gameWidth ->set(root.viewWidth()/2 - unit, SPAN);
            gameHeight->set(root.viewHeight()/4,       SPAN);
            homeDelta ->set(-*gameHeight - unit,       SPAN);
            isGameMini = true;
        }
        else if (!mini && isGameMini) {
            gameWidth ->set(root.viewWidth(),  SPAN);
            gameHeight->set(root.viewHeight(), SPAN);
            homeDelta ->set(0,                 SPAN);
            isGameMini = false;
        }
    }
    void updateMouseCursor() {
#if !defined (DENG_MOBILE)
        cursor->show(!self().eventHandler().isMouseTrapped() && VRConfig::modeAppliesDisplacement(vrCfg().mode()));
        if (cursor->isVisible()) {
            if (!cursorHasBeenHidden) {
                qApp->setOverrideCursor(QCursor(Qt::BlankCursor));
                cursorHasBeenHidden = true;
            }
            Vector2i cp = ClientApp::windowSystem().latestMousePosition();
            cursorX->set(cp.x);
            cursorY->set(cp.y);
        }
        else {
            if (cursorHasBeenHidden) {
                qApp->restoreOverrideCursor();
            }
            cursorHasBeenHidden = false;
        }
#endif
    }
    void setupFade(FadeDirection fadeDir, TimeSpan const &span) {
        if (!fader) {
            fader.reset(new FadeToBlackWidget);
            fader->rule().setRect(root.viewRule());
            root.add(fader); // on top of everything else
        }
        if (fadeDir == FadeFromBlack) {
            fader->initFadeFromBlack(span);
        }
        else {
            fader->initFadeToBlack(span);
        }
        fader->start();
    }
    void completeFade() {
        if (fader) {
            fader->disposeIfDone();
        }
    }
};
ClientWindow::ClientWindow(String const &id)
    : BaseWindow(id)
    , d(new Impl(this)) {
#if defined (DENG_MOBILE)
    setMain(this);
    WindowSystem::get().addWindow("main", this);
#endif
    audienceForInit()   += d;
    audienceForResize() += d;
    audienceForSwap()   += d;
#if defined (WIN32)
    setIcon(QIcon(":/doomsday.ico"));
#endif
    d->setupUI();
#if defined (DENG_MOBILE)
    connect(this, &GLWindow::rootDimensionsChanged, [this] (QRect rect) {
        d->root.rootOffset().setValue(Vector2f(0, int(rect.height()) - int(pixelSize().y)),
                                      0.3);
    });
#endif
}
ClientRootWidget &ClientWindow::root() {
    return d->root;
}
TaskBarWidget &ClientWindow::taskBar() {
    return *d->taskBar;
}
GuiWidget &ClientWindow::taskBarBlur() {
    return *d->taskBarBlur;
}
ConsoleWidget &ClientWindow::console() {
    return d->taskBar->console();
}
NotificationAreaWidget &ClientWindow::notifications() {
    return *d->notifications;
}
HomeWidget &ClientWindow::home() {
    return *d->home;
}
GameWidget &ClientWindow::game() {
    return *d->game;
}
BusyWidget &ClientWindow::busy() {
    return *d->busy;
}
AlertDialog &ClientWindow::alerts() {
    return *d->alerts;
}
bool ClientWindow::isFPSCounterVisible() const {
    return App::config().getb(configName("showFps"));
}
void ClientWindow::setMode(Mode const &mode) {
    LOG_AS("ClientWindow");
    d->setMode(mode);
}
void ClientWindow::setGameMinimized(bool minimize) {
    d->minimizeGame(minimize);
}
bool ClientWindow::isGameMinimized() const {
    return d->isGameMini;
}
void ClientWindow::windowAboutToClose() {
    if (BusyMode_Active()) {
        return;
    }
    Sys_Quit();
}
void ClientWindow::preDraw() {
    ClientApp::app().preFrame(); /// @todo what about multiwindow?
    DENG2_ASSERT_IN_RENDER_THREAD();
    LIBGUI_ASSERT_GL_CONTEXT_ACTIVE();
    d->updateMouseCursor();
    if (d->needRootSizeUpdate) {
        d->updateRootSize();
    }
    BaseWindow::preDraw();
}
Vector2f ClientWindow::windowContentSize() const {
    return Vector2f(d->root.viewWidth().value(), d->root.viewHeight().value());
}
void ClientWindow::drawWindowContent() {
#if defined (DENG_MOBILE)
    {
    }
#endif
    DENG2_ASSERT_IN_RENDER_THREAD();
    root().draw();
    LIBGUI_ASSERT_GL_OK();
}
void ClientWindow::postDraw() {
    DENG2_ASSERT_IN_RENDER_THREAD();
    BaseWindow::postDraw();
    if (ClientApp::vr().mode() != VRConfig::OculusRift) {
        GL_FinishFrame();
    }
}
bool ClientWindow::setDefaultGLFormat() // static
{
    LOG_AS("DefaultGLFormat");
    QSurfaceFormat fmt = QSurfaceFormat::defaultFormat();
    if (CommandLine_Exists("-novsync") || !App::config().getb("window.main.vsync")) {
        fmt.setSwapInterval(0);
    }
    else {
        fmt.setSwapInterval(1);
    }
    if (fmt != QSurfaceFormat::defaultFormat()) {
        LOG_GL_VERBOSE("Applying new format...");
        QSurfaceFormat::setDefaultFormat(fmt);
        return true;
    }
    else {
        LOG_GL_XVERBOSE("New format is the same as before", "");
        return false;
    }
}
void ClientWindow::grab(image_t &img, bool halfSized) const {
    DENG_ASSERT_IN_MAIN_THREAD();
    QSize outputSize = (halfSized? QSize(pixelWidth()/2, pixelHeight()/2) : QSize());
    QImage grabbed = grabImage(outputSize);
    Image_Init(img);
    img.size      = Vector2ui(grabbed.width(), grabbed.height());
    img.pixelSize = grabbed.depth()/8;
    img.pixels = (uint8_t *) malloc(grabbed.byteCount());
    std::memcpy(img.pixels, grabbed.constBits(), grabbed.byteCount());
    LOGDEV_GL_MSG("Grabbed Canvas contents %i x %i, byteCount:%i depth:%i format:%i")
            << grabbed.width() << grabbed.height()
            << grabbed.byteCount() << grabbed.depth() << grabbed.format();
    DENG_ASSERT(img.pixelSize != 0);
}
void ClientWindow::fadeInTaskBarBlur(TimeSpan span) {
    d->taskBarBlur->setAttribute(GuiWidget::DontDrawContent, UnsetFlags);
    d->taskBarBlur->setOpacity(0);
    d->taskBarBlur->setOpacity(1, span);
}
void ClientWindow::fadeOutTaskBarBlur(TimeSpan span) {
    d->taskBarBlur->setOpacity(0, span);
    QTimer::singleShot(span.asMilliSeconds(), this, SLOT(hideTaskBarBlur()));
}
void ClientWindow::hideTaskBarBlur() {
    d->taskBarBlur->setAttribute(GuiWidget::DontDrawContent);
    if (d->taskBar->style().isBlurringAllowed()) {
        d->taskBarBlur->setOpacity(1);
    }
    else {
        d->taskBarBlur->setOpacity(0);
    }
}
void ClientWindow::updateRootSize() {
    d->needRootSizeUpdate = true;
}
ClientWindow &ClientWindow::main() {
    return static_cast<ClientWindow &>(BaseWindow::main());
}
bool ClientWindow::mainExists() {
    return BaseWindow::mainExists();
}
void ClientWindow::toggleFPSCounter() {
    App::config().set(configName("showFps"), !isFPSCounterVisible());
}
void ClientWindow::showColorAdjustments() {
    d->colorAdjust->open();
}
void ClientWindow::addOnTop(GuiWidget *widget) {
    d->root.add(widget);
    d->root.moveChildToLast(*d->cursor);
}
void ClientWindow::setSidebar(SidebarLocation location, GuiWidget *sidebar) {
    DENG2_ASSERT(location == RightEdge);
    d->installSidebar(location, sidebar);
}
bool ClientWindow::hasSidebar(SidebarLocation location) const {
    DENG2_ASSERT(location == RightEdge);
    DENG2_UNUSED(location);
    return d->sidebar != 0;
}
GuiWidget &ClientWindow::sidebar(SidebarLocation location) const {
    DENG2_ASSERT(location == RightEdge);
    DENG2_UNUSED(location);
    DENG2_ASSERT(d->sidebar != nullptr);
    return *d->sidebar;
}
bool ClientWindow::handleFallbackEvent(Event const &event) {
    return d->handleFallbackEvent(event);
}
void ClientWindow::fadeContent(FadeDirection fadeDirection, TimeSpan const &duration) {
    d->setupFade(fadeDirection, duration);
}
FadeToBlackWidget *ClientWindow::contentFade() {
    return d->fader;
}
#undef M_ScreenShot
DENG_EXTERN_C int M_ScreenShot(char const *name, int flags) {
    de::String fullName(name);
    if(fullName.fileNameExtension().isEmpty()) {
        fullName += ".png"; // Default format.
    }
    NativePath const shotPath(App::app().nativeHomePath() / fullName);
    if (flags & DD_SCREENSHOT_CHECK_EXISTS) {
        return (shotPath.exists() ? 1 : 0);
    }
    return ClientWindow::main().grabToFile(shotPath)? 1 : 0;
}
DENG2_PIMPL(ClientWindowSystem) {
    ConfigProfiles settings;
    Impl(Public *i)
        : Base(i) {
        self().setStyle(new ClientStyle);
        self().style().load(App::packageLoader().load("net.dengine.client.defaultstyle"));
        using SReg = ConfigProfiles;
        settings.define(SReg::ConfigVariable, "window.main.showFps")
                .define(SReg::ConfigVariable, "window.main.fsaa")
                .define(SReg::ConfigVariable, "window.main.vsync")
                .define(SReg::ConfigVariable, "window.main.maxFps")
                .define(SReg::IntCVar,        "rend-finale-stretch", SCALEMODE_SMART_STRETCH)
                .define(SReg::IntCVar,        "rend-hud-stretch", SCALEMODE_SMART_STRETCH)
                .define(SReg::IntCVar,        "inlude-stretch", SCALEMODE_SMART_STRETCH)
                .define(SReg::IntCVar,        "menu-stretch", SCALEMODE_SMART_STRETCH);
    }
};
ClientWindowSystem::ClientWindowSystem()
    : WindowSystem()
    , d(new Impl(this)) {
    ClientWindow::setDefaultGLFormat();
}
ConfigProfiles &ClientWindowSystem::settings() {
    return d->settings;
}
ClientWindow *ClientWindowSystem::createWindow(String const &id) {
    return newWindow<ClientWindow>(id);
}
ClientWindow &ClientWindowSystem::main() {
    return WindowSystem::main().as<ClientWindow>();
}
ClientWindow *ClientWindowSystem::mainPtr() {
    return static_cast<ClientWindow *>(WindowSystem::mainPtr());
}
void ClientWindowSystem::closingAllWindows() {
    GL_Shutdown();
    WindowSystem::closingAllWindows();
}
bool ClientWindowSystem::rootProcessEvent(Event const &event) {
    return main().root().processEvent(event);
}
void ClientWindowSystem::rootUpdate() {
    main().root().update();
}
CommandAction::CommandAction(String const &cmd, int commandSource)
    : _command(cmd), _source(commandSource)
{}
void CommandAction::trigger() {
    Action::trigger();
    Con_Execute(_source, _command.toUtf8(), false /*silent*/, false /*net*/);
}
void CommandBinding::resetToDefaults() {
    Binding::resetToDefaults();
    def().addNumber("deviceId", -1);
    def().addNumber("controlId", -1);
    def().addNumber("type", int(E_TOGGLE));  ///< Type of event.
    def().addText  ("symbolicName", "");
    def().addNumber("test", int(None));
    def().addNumber("pos", 0);
    def().addText  ("command", "");  ///< Command to execute.
}
String CommandBinding::composeDescriptor() {
    LOG_AS("CommandBinding");
    if (!*this) return "";
    String str = B_ControlDescToString(geti("deviceId"), ddeventtype_t(geti("type")), geti("controlId"));
    switch (geti("type")) {
    case E_TOGGLE:      str += B_ButtonStateToString(ControlTest(geti("test"))); break;
    case E_AXIS:        str += B_AxisPositionToString(ControlTest(geti("test")), getf("pos")); break;
    case E_ANGLE:       str += B_HatAngleToString(getf("pos")); break;
    case E_SYMBOLIC:    str += "-" + gets("symbolicName"); break;
    default: DENG2_ASSERT(!"CommandBinding::composeDescriptor: Unknown bind.type"); break;
    }
    ArrayValue const &conds = def().geta("condition");
    DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds.elements()) {
        str += " + " + B_ConditionToString(*(*i)->as<RecordValue>().record());
    }
    return str;
}
/**
 * Parse the main part of the event descriptor, with no conditions included.
 */
static bool doConfigure(CommandBinding &bind, char const *eventDesc, char const *command) {
    DENG2_ASSERT(eventDesc);
    bind.resetToDefaults();
    bind.def().set("command", String(command));
    AutoStr *str = AutoStr_NewStd();
    eventDesc = Str_CopyDelim(str, eventDesc, '-');
    if (!Str_CompareIgnoreCase(str, "key")) {
        bind.def().set("deviceId", IDEV_KEYBOARD);
        bind.def().set("type", int(E_TOGGLE)); // Keyboards only have toggles (as far as we know).
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        int controlId = 0;
        bool ok = B_ParseKeyId(controlId, Str_Text(str));
        if (!ok) return false;
        bind.def().set("controlId", controlId);
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        Binding::ControlTest test = Binding::ControlTest::None;
        ok = B_ParseButtonState(test, Str_Text(str));
        if (!ok) return false;
        bind.def().set("test", int(test));
    }
    else if (!Str_CompareIgnoreCase(str, "mouse")) {
        bind.def().set("deviceId", IDEV_MOUSE);
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        ddeventtype_t type = E_TOGGLE;
        int controlId = 0;
        bool ok = B_ParseMouseTypeAndId(type, controlId, Str_Text(str));
        if (!ok) return false;
        bind.def().set("type", type);
        bind.def().set("controlId", controlId);
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        switch (bind.geti("type")) {
        case E_TOGGLE: {
            Binding::ControlTest test = Binding::ControlTest::None;
            ok = B_ParseButtonState(test, Str_Text(str));
            if (!ok) return false;
            bind.def().set("test", int(test));
            break; }
        case E_AXIS: {
            Binding::ControlTest test = Binding::ControlTest::None;
            float pos;
            ok = B_ParseAxisPosition(test, pos, Str_Text(str));
            if (!ok) return false;
            bind.def().set("test", int(test));
            bind.def().set("pos", pos);
            break; }
        default: DENG2_ASSERT(!"InputSystem::configure: Invalid bind.type"); break;
        }
    }
    else if (!Str_CompareIgnoreCase(str, "joy") ||
            !Str_CompareIgnoreCase(str, "head")) {
        bind.def().set("deviceId", (!Str_CompareIgnoreCase(str, "joy")? IDEV_JOY1 : IDEV_HEAD_TRACKER));
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        ddeventtype_t type = E_TOGGLE;
        int controlId = 0;
        bool ok = B_ParseJoystickTypeAndId(type, controlId, bind.geti("deviceId"), Str_Text(str));
        if (!ok) return false;
        bind.def().set("type", type);
        bind.def().set("controlId", controlId);
        eventDesc = Str_CopyDelim(str, eventDesc, '-');
        switch (bind.geti("type")) {
        case E_TOGGLE: {
            Binding::ControlTest test = Binding::ControlTest::None;
            ok = B_ParseButtonState(test, Str_Text(str));
            if (!ok) return false;
            bind.def().set("test", int(test));
            break; }
        case E_AXIS: {
            Binding::ControlTest test = Binding::ControlTest::None;
            float pos;
            ok = B_ParseAxisPosition(test, pos, Str_Text(str));
            if (!ok) return false;
            bind.def().set("test", int(test));
            bind.def().set("pos", pos);
            break; }
        case E_ANGLE: {
            float pos;
            ok = B_ParseHatAngle(pos, Str_Text(str));
            if (!ok) return false;
            bind.def().set("pos", pos);
            break; }
        default: DENG2_ASSERT(!"InputSystem::configure: Invalid bind.type") break;
        }
    }
    else if (!Str_CompareIgnoreCase(str, "sym")) {
        bind.def().set("type", int(E_SYMBOLIC));
        bind.def().set("deviceId", -1);
        bind.def().set("symbolicName", eventDesc);
        eventDesc = nullptr;
    }
    else {
        LOG_INPUT_WARNING("Unknown device \"%s\"") << Str_Text(str);
        return false;
    }
    if (eventDesc) {
        LOG_INPUT_WARNING("Unrecognized \"%s\"") << eventDesc;
        return false;
    }
    return true;
}
void CommandBinding::configure(char const *eventDesc, char const *command, bool assignNewId) {
    DENG2_ASSERT(eventDesc);
    LOG_AS("CommandBinding");
    AutoStr *str = AutoStr_NewStd();
    eventDesc = Str_CopyDelim(str, eventDesc, '+');
    if (!doConfigure(*this, Str_Text(str), command)) {
        throw ConfigureError("CommandBinding::configure", "Descriptor parse error");
    }
    def()["condition"].array().clear();
    while (eventDesc) {
        eventDesc = Str_CopyDelim(str, eventDesc, '+');
        Record &cond = addCondition();
        if (!B_ParseBindingCondition(cond, Str_Text(str))) {
            throw ConfigureError("CommandBinding::configure", "Descriptor parse error");
        }
    }
    if (assignNewId) {
        def().set("id", newIdentifier());
    }
}
/**
 * Substitute placeholders in a command string. Placeholders consist of two characters,
 * the first being a %. Use %% to output a plain % character.
 *
 * - <code>%i</code>: id member of the event
 * - <code>%p</code>: (symbolic events only) local player number
 *
 * @param command  Original command string with the placeholders.
 * @param event    Event data.
 * @param out      String with placeholders replaced.
 */
static void substituteInCommand(String const &command, ddevent_t const &event, ddstring_t *out) {
    DENG2_ASSERT(out);
    Block const str = command.toUtf8();
    for (char const *ptr = str.constData(); *ptr; ptr++) {
        if (*ptr == '%') {
            ptr++;
            if (!*ptr) break;
            if (*ptr == 'i') {
                int id = 0;
                switch (event.type) {
                case E_TOGGLE:   id = event.toggle.id;   break;
                case E_AXIS:     id = event.axis.id;     break;
                case E_ANGLE:    id = event.angle.id;    break;
                case E_SYMBOLIC: id = event.symbolic.id; break;
                default: break;
                }
                Str_Appendf(out, "%i", id);
            }
            else if (*ptr == 'p') {
                int id = 0;
                if (event.type == E_SYMBOLIC) {
                    id = P_ConsoleToLocal(event.symbolic.id);
                }
                Str_Appendf(out, "%i", id);
            }
            else if (*ptr == '%') {
                Str_AppendChar(out, *ptr);
            }
            continue;
        }
        Str_AppendChar(out, *ptr);
    }
}
Action *CommandBinding::makeAction(ddevent_t const &event, BindContext const &context,
    bool respectHigherContexts) const {
    if (geti("type") != event.type) return nullptr;
    InputDevice const *dev = nullptr;
    if (event.type != E_SYMBOLIC) {
        if (geti("deviceId") != event.device) return nullptr;
        dev = InputSystem::get().devicePtr(geti("deviceId"));
        if (!dev || !dev->isActive()) {
            return nullptr;
        }
    }
    switch (event.type) {
    case E_TOGGLE: {
        if (geti("controlId") != event.toggle.id)
            return nullptr;
        DENG2_ASSERT(dev);
        ButtonInputControl &button = dev->button(geti("controlId"));
        if (respectHigherContexts) {
            if (button.bindContext() != &context)
                return nullptr; // Shadowed by a more important active class.
        }
        button.setBindContextAssociation(InputControl::Triggered, UnsetFlags);
        switch (ControlTest(geti("test"))) {
        case ButtonStateAny:
            break;
        case ButtonStateDown:
            if (event.toggle.state != ETOG_DOWN)
                return nullptr;
            break;
        case ButtonStateUp:
            if (event.toggle.state != ETOG_UP)
                return nullptr;
            break;
        case ButtonStateRepeat:
            if (event.toggle.state != ETOG_REPEAT)
                return nullptr;
            break;
        case ButtonStateDownOrRepeat:
            if (event.toggle.state == ETOG_UP)
                return nullptr;
            break;
        default: return nullptr;
        }
        break; }
    case E_AXIS:
        if (geti("controlId") != event.axis.id)
            return nullptr;
        DENG2_ASSERT(dev);
        if (dev->axis(geti("controlId")).bindContext() != &context)
            return nullptr; // Shadowed by a more important active class.
        if (!B_CheckAxisPosition(ControlTest(geti("test")), getf("pos"),
                                InputSystem::get().device(event.device).axis(event.axis.id)
                                    .translateRealPosition(event.axis.pos)))
            return nullptr;
        break;
    case E_ANGLE:
        if (geti("controlId") != event.angle.id)
            return nullptr;
        DENG2_ASSERT(dev);
        if (dev->hat(geti("controlId")).bindContext() != &context)
            return nullptr; // Shadowed by a more important active class.
        if (event.angle.pos != getf("pos"))
            return nullptr;
        break;
    case E_SYMBOLIC:
        if (gets("symbolicName").compareWithCase(event.symbolic.name))
            return nullptr;
        break;
    default: return nullptr;
    }
    ArrayValue const &conds = def().geta("condition");
    DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds.elements()) {
        if (!B_CheckCondition(static_cast<Binding::CompiledConditionRecord *>
                              ((*i)->as<RecordValue>().record()), 0, &context))
            return nullptr;
    }
    AutoStr *command = Str_Reserve(AutoStr_NewStd(), gets("command").length());
    substituteInCommand(gets("command"), event, command);
    return new CommandAction(Str_Text(command), CMDS_BIND);
}
static String VAR_CONTROLLER_PRESETS("controllerPresets");
DENG2_PIMPL_NOREF(ControllerPresets)
, DENG2_OBSERVES(DoomsdayApp, GameChange) {
    Record &inputModule;
    char const *presetCVarPath = nullptr;
    int deviceId = 0;
    Impl()
        : inputModule(App::scriptSystem()["Input"]) {
        inputModule.addDictionary(VAR_CONTROLLER_PRESETS);
        DoomsdayApp::app().audienceForGameChange() += this;
    }
    DictionaryValue const &presets() const {
        return inputModule.get(VAR_CONTROLLER_PRESETS).as<DictionaryValue>();
    }
    QList<QString> ids() const {
        QSet<QString> ids;
        for (auto i : presets().elements()) {
            if (auto const *value = maybeAs<RecordValue>(i.second)) {
                ids.insert(value->dereference().gets("id"));
            }
        }
        return ids.toList();
    }
    /**
     * Checks the presets dictionary if any the regular expressions match
     * the given device name. Returns a Doomsday Script object whose bind() method
     * can be called to set the bindings for the gamepad scheme.
     *
     * @param deviceName  Device name.
     *
     * @return Object defining the gamepad default bindings, or @c nullptr if not found.
     */
    Record const *findMatching(String const &deviceName) const {
        for (auto i : presets().elements()) {
            String const key = i.first.value->asText();
            if (!key.isEmpty() && QRegExp(key, Qt::CaseInsensitive).exactMatch(deviceName)) {
                if (auto const *value = maybeAs<RecordValue>(i.second)) {
                    return value->record();
                }
            }
        }
        return nullptr;
    }
    Record const *findById(String const &id) const {
        for (auto i : presets().elements()) {
            if (auto const *value = maybeAs<RecordValue>(i.second)) {
                if (value->dereference().gets("id") == id) {
                    return value->record();
                }
            }
        }
        return nullptr;
    }
    cvar_t *presetCVar() const {
        return Con_FindVariable(presetCVarPath);
    }
    void currentGameChanged(Game const &newGame) {
        DENG2_ASSERT(deviceId == IDEV_JOY1); /// @todo Expand for other devices as needed. -jk
        if (!newGame.isNull() && !Joystick_Name().isEmpty()) {
            String const currentScheme = CVar_String(presetCVar());
            if (auto const *gamepad = findMatching(Joystick_Name())) {
                if (currentScheme.isEmpty()) {
                    LOG_INPUT_NOTE("Detected a known gamepad: %s") << Joystick_Name();
                    applyPreset(gamepad);
                }
            }
        }
    }
    void applyPreset(Record const *preset) {
        ClientApp::inputSystem().removeBindingsForDevice(deviceId);
        if (preset) {
            LOG_INPUT_MSG("Applying bindings for %s gamepad") << preset->gets("id");
            Script script("preset.bind()");
            Process proc(script);
            proc.globals().add("preset").set(new RecordValue(*preset));
            proc.execute();
        }
        CVar_SetString(presetCVar(), preset? preset->gets("id").toUtf8() : "");
    }
};
ControllerPresets::ControllerPresets(int deviceId, char const *presetCVarPath)
    : d(new Impl) {
    d->deviceId       = deviceId;
    d->presetCVarPath = presetCVarPath;
}
String ControllerPresets::currentPreset() const {
    return CVar_String(d->presetCVar());
}
void ControllerPresets::applyPreset(String const &presetId) {
    d->applyPreset(d->findById(presetId));
}
QStringList ControllerPresets::ids() const {
    return d->ids();
}
HatInputControl::HatInputControl(String const &name) {
    setName(name);
}
HatInputControl::~HatInputControl()
{}
dint HatInputControl::position() const {
    return _pos;
}
void HatInputControl::setPosition(dint newPosition) {
    _pos  = newPosition;
    _time = Timer_RealMilliseconds(); // Remember when the change occured.
    if (_pos < 0) {
        setBindContextAssociation(Expired, UnsetFlags);
    }
}
duint HatInputControl::time() const {
    return _time;
}
String HatInputControl::description() const {
    return String(_E(b) "%1 " _E(.) "(Hat) "
                  _E(l) "Position: " _E(.) "%2")
            .arg(fullName())
            .arg(_pos);
}
bool HatInputControl::inDefaultState() const {
    return _pos < 0; // Centered?
}
CompiledImpulseBinding::CompiledImpulseBinding(Record const &bind)
    : id            (bind.geti("id"))
    , deviceId      (bind.geti("deviceId"))
    , controlId     (bind.geti("controlId"))
    , type          (ibcontroltype_t(bind.geti("type")))
    , angle         (bind.getf("angle"))
    , flags         (bind.geti("flags"))
    , impulseId     (bind.geti("impulseId"))
    , localPlayer   (bind.geti("localPlayer"))
{}
CompiledImpulseBindingRecord &ImpulseBinding::def() {
    return static_cast<CompiledImpulseBindingRecord &>(Binding::def());
}
CompiledImpulseBindingRecord const &ImpulseBinding::def() const {
    return static_cast<CompiledImpulseBindingRecord const &>(Binding::def());
}
void ImpulseBinding::resetToDefaults() {
    Binding::resetToDefaults();
    def().resetCompiled();
    def().addNumber("deviceId", -1);
    def().addNumber("controlId", -1);
    def().addNumber("type", int(IBD_TOGGLE));  ///< Type of event.
    def().addNumber("angle", 0);
    def().addNumber("flags", 0);
    def().addNumber("impulseId", 0);           ///< Identifier of the bound player impulse.
    def().addNumber("localPlayer", -1);        ///< Local player number.
}
String ImpulseBinding::composeDescriptor() {
    LOG_AS("ImpulseBinding");
    if (!*this) return "";
    String str = B_ControlDescToString(geti("deviceId"), IBDTYPE_TO_EVTYPE(geti("type")), geti("controlId"));
    if (geti("type") == IBD_ANGLE) {
        str += B_HatAngleToString(getf("angle"));
    }
    if (geti("flags") & IBDF_TIME_STAGED) {
        str += "-staged";
    }
    if (geti("flags") & IBDF_INVERSE) {
        str += "-inverse";
    }
    ArrayValue const &conds = def().geta("condition");
    DENG2_FOR_EACH_CONST(ArrayValue::Elements, i, conds.elements()) {
        str += " + " + B_ConditionToString(*(*i)->as<RecordValue>().record());
    }
    return str;
}
static bool doConfigure(ImpulseBinding &bind, char const *ctrlDesc, int impulseId, int localPlayer) {
    DENG2_ASSERT(ctrlDesc);
    bind.resetToDefaults();
    bind.def().set("impulseId", impulseId);
    bind.def().set("localPlayer", localPlayer);
    AutoStr *str = AutoStr_NewStd();
    ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
    if (!Str_CompareIgnoreCase(str, "key")) {
        bind.def().set("deviceId", IDEV_KEYBOARD);
        bind.def().set("type", int(IBD_TOGGLE));
        ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
        int keyId;
        bool ok = B_ParseKeyId(keyId, Str_Text(str));
        if (!ok) return false;
        bind.def().set("controlId", keyId);
    }
    else if (!Str_CompareIgnoreCase(str, "mouse")) {
        bind.def().set("deviceId", IDEV_MOUSE);
        ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
        ddeventtype_t type;
        int controlId = 0;
        bool ok = B_ParseMouseTypeAndId(type, controlId, Str_Text(str));
        if (!ok) return false;
        bind.def().set("controlId", controlId);
        bind.def().set("type", int(EVTYPE_TO_IBDTYPE(type)));
    }
    else if (!Str_CompareIgnoreCase(str, "joy") ||
             !Str_CompareIgnoreCase(str, "head")) {
        bind.def().set("deviceId", (!Str_CompareIgnoreCase(str, "joy")? IDEV_JOY1 : IDEV_HEAD_TRACKER));
        ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
        ddeventtype_t type;
        int controlId = 0;
        bool ok = B_ParseJoystickTypeAndId(type, controlId, bind.geti("deviceId"), Str_Text(str));
        if (!ok) return false;
        bind.def().set("controlId", controlId);
        bind.def().set("type", int(EVTYPE_TO_IBDTYPE(type)));
        if (type == E_ANGLE) {
            ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
            float angle;
            ok = B_ParseHatAngle(angle, Str_Text(str));
            if (!ok) return false;
            bind.def().set("angle", angle);
        }
    }
    while (ctrlDesc) {
        ctrlDesc = Str_CopyDelim(str, ctrlDesc, '-');
        if (!Str_CompareIgnoreCase(str, "inverse")) {
            bind.def().set("flags", bind.geti("flags") | IBDF_INVERSE);
        }
        else if (!Str_CompareIgnoreCase(str, "staged")) {
            bind.def().set("flags", bind.geti("flags") | IBDF_TIME_STAGED);
        }
        else {
            LOG_INPUT_WARNING("Unrecognized \"%s\"") << ctrlDesc;
            return false;
        }
    }
    return true;
}
void ImpulseBinding::configure(char const *ctrlDesc, int impulseId, int localPlayer, bool assignNewId) {
    DENG2_ASSERT(ctrlDesc);
    DENG2_ASSERT(localPlayer >= 0 && localPlayer < DDMAXPLAYERS);
    LOG_AS("ImpulseBinding");
    AutoStr *str = AutoStr_NewStd();
    ctrlDesc = Str_CopyDelim(str, ctrlDesc, '+');
    if (!doConfigure(*this, Str_Text(str), impulseId, localPlayer)) {
        throw ConfigureError("ImpulseBinding::configure", "Descriptor parse error");
    }
    def()["condition"].array().clear();
    while (ctrlDesc) {
        ctrlDesc = Str_CopyDelim(str, ctrlDesc, '+');
        Record &cond = addCondition();
        if (!B_ParseBindingCondition(cond, Str_Text(str))) {
            throw ConfigureError("ImpulseBinding::configure", "Descriptor parse error");
        }
    }
    if (assignNewId) {
        def().set("id", newIdentifier());
    }
}
DENG2_PIMPL(InputDevice) {
    bool active = false;  ///< Initially inactive.
    String title;         ///< Human-friendly title.
    String name;          ///< Symbolic name.
    QList<  AxisInputControl *> axes;
    QList<ButtonInputControl *> buttons;
    QList<   HatInputControl *> hats;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        qDeleteAll(hats);
        qDeleteAll(buttons);
        qDeleteAll(axes);
    }
    DENG2_PIMPL_AUDIENCE(ActiveChange)
};
DENG2_AUDIENCE_METHOD(InputDevice, ActiveChange)
InputDevice::InputDevice(String const &name) : d(new Impl(this)) {
    DENG2_ASSERT(!name.isEmpty());
    d->name = name;
}
InputDevice::~InputDevice()
{}
bool InputDevice::isActive() const {
    return d->active;
}
void InputDevice::activate(bool yes) {
    if (d->active != yes) {
        d->active = yes;
        DENG2_FOR_AUDIENCE2(ActiveChange, i) i->inputDeviceActiveChanged(*this);
    }
}
String InputDevice::name() const {
    return d->name;
}
String InputDevice::title() const {
    return (d->title.isEmpty()? d->name : d->title);
}
void InputDevice::setTitle(String const &newTitle) {
    d->title = newTitle;
}
String InputDevice::description() const {
    String desc;
    if (!d->title.isEmpty()) {
        desc += String(_E(D)_E(b) "%1" _E(.)_E(.) " - ").arg(d->title);
    }
    desc += String(_E(b) "%1" _E(.)_E(l) " (%2)" _E(.)).arg(name()).arg(isActive()? "active" : "inactive");
    if (int const count = axisCount()) {
        desc += String("\n  " _E(b) "%1 axes:" _E(.)).arg(count);
        int idx = 0;
        for (Control const *axis : d->axes) {
            desc += String("\n    [%1] " _E(>) "%2" _E(<)).arg(idx++, 3).arg(axis->description());
        }
    }
    if (int const count = buttonCount()) {
        desc += String("\n  " _E(b) "%1 buttons:" _E(.)).arg(count);
        int idx = 0;
        for (Control const *button : d->buttons) {
            desc += String("\n    [%1] " _E(>) "%2" _E(<)).arg(idx++, 3).arg(button->description());
        }
    }
    if (int const count = hatCount()) {
        desc += String("\n  " _E(b) "%1 hats:" _E(.)).arg(count);
        int idx = 0;
        for (Control const *hat : d->hats) {
            desc += String("\n    [%1] " _E(>) "%2" _E(<)).arg(idx++, 3).arg(hat->description());
        }
    }
    return desc;
}
void InputDevice::reset() {
    LOG_AS("InputDevice");
    for (Control *axis : d->axes) {
        axis->reset();
    }
    for (Control *button : d->buttons) {
        button->reset();
    }
    for (Control *hat : d->hats) {
        hat->reset();
    }
    if (!d->name.compareWithCase("key")) {
        extern bool shiftDown, altDown;
        altDown = shiftDown = false;
    }
    LOG_INPUT_VERBOSE(_E(b) "'%s'" _E(.) " controls reset") << title();
}
LoopResult InputDevice::forAllControls(std::function<de::LoopResult (Control &)> func) {
    for (Control *axis : d->axes) {
        if (auto result = func(*axis)) return result;
    }
    for (Control *button : d->buttons) {
        if (auto result = func(*button)) return result;
    }
    for (Control *hat : d->hats) {
        if (auto result = func(*hat)) return result;
    }
    return LoopContinue;
}
dint InputDevice::toAxisId(String const &name) const {
    if (!name.isEmpty()) {
        for (int i = 0; i < d->axes.count(); ++i) {
            if (!d->axes.at(i)->name().compareWithoutCase(name))
                return i;
        }
    }
    return -1;
}
dint InputDevice::toButtonId(String const &name) const {
    if (!name.isEmpty()) {
        for (int i = 0; i < d->buttons.count(); ++i) {
            if (!d->buttons.at(i)->name().compareWithoutCase(name))
                return i;
        }
    }
    return -1;
}
bool InputDevice::hasAxis(de::dint id) const {
    return (id >= 0 && id < d->axes.count());
}
AxisInputControl &InputDevice::axis(dint id) const {
    if (hasAxis(id)) return *d->axes.at(id);
    throw MissingControlError("InputDevice::axis", "Invalid id:" + String::number(id));
}
void InputDevice::addAxis(AxisInputControl *axis) {
    if (!axis) return;
    d->axes.append(axis);
    axis->setDevice(this);
}
int InputDevice::axisCount() const {
    return d->axes.count();
}
bool InputDevice::hasButton(de::dint id) const {
    return (id >= 0 && id < d->buttons.count());
}
ButtonInputControl &InputDevice::button(dint id) const {
    if (hasButton(id)) return *d->buttons.at(id);
    throw MissingControlError("InputDevice::button", "Invalid id:" + String::number(id));
}
void InputDevice::addButton(ButtonInputControl *button) {
    if (!button) return;
    d->buttons.append(button);
    button->setDevice(this);
}
int InputDevice::buttonCount() const {
    return d->buttons.count();
}
bool InputDevice::hasHat(de::dint id) const {
    return (id >= 0 && id < d->hats.count());
}
HatInputControl &InputDevice::hat(dint id) const {
    if (hasHat(id)) return *d->hats.at(id);
    throw MissingControlError("InputDevice::hat", "Invalid id:" + String::number(id));
}
void InputDevice::addHat(HatInputControl *hat) {
    if (!hat) return;
    d->hats.append(hat);
    hat->setDevice(this);
}
int InputDevice::hatCount() const {
    return d->hats.count();
}
void InputDevice::consoleRegister() {
    for (Control *axis : d->axes) {
        axis->consoleRegister();
    }
    for (Control *button : d->buttons) {
        button->consoleRegister();
    }
    for (Control *hat : d->hats) {
        hat->consoleRegister();
    }
}
DENG2_PIMPL_NOREF(InputDevice::Control) {
    String name;  ///< Symbolic
    InputDevice *device = nullptr;
    BindContextAssociation flags = DefaultFlags;
    BindContext *bindContext     = nullptr;
    BindContext *prevBindContext = nullptr;
};
InputDevice::Control::Control(InputDevice *device) : d(new Impl) {
    setDevice(device);
}
InputDevice::Control::~Control()
{}
String InputDevice::Control::name() const {
    DENG2_GUARD(this);
    return d->name;
}
void InputDevice::Control::setName(String const &newName) {
    DENG2_GUARD(this);
    d->name = newName;
}
String InputDevice::Control::fullName() const {
    DENG2_GUARD(this);
    String desc;
    if (hasDevice()) desc += device().name() + "-";
    desc += (d->name.isEmpty()? "<unnamed>" : d->name);
    return desc;
}
InputDevice &InputDevice::Control::device() const {
    DENG2_GUARD(this);
    if (d->device) return *d->device;
    throw MissingDeviceError("InputDevice::Control::device", "No InputDevice is attributed");
}
bool InputDevice::Control::hasDevice() const {
    DENG2_GUARD(this);
    return d->device != nullptr;
}
void InputDevice::Control::setDevice(InputDevice *newDevice) {
    DENG2_GUARD(this);
    d->device = newDevice;
}
BindContext *InputDevice::Control::bindContext() const {
    DENG2_GUARD(this);
    return d->bindContext;
}
void InputDevice::Control::setBindContext(BindContext *newContext) {
    DENG2_GUARD(this);
    d->bindContext = newContext;
}
InputDevice::Control::BindContextAssociation InputDevice::Control::bindContextAssociation() const {
    DENG2_GUARD(this);
    return d->flags;
}
void InputDevice::Control::setBindContextAssociation(BindContextAssociation const &flagsToChange, FlagOp op) {
    DENG2_GUARD(this);
    applyFlagOperation(d->flags, flagsToChange, op);
}
void InputDevice::Control::clearBindContextAssociation() {
    DENG2_GUARD(this);
    d->prevBindContext = d->bindContext;
    d->bindContext     = nullptr;
    applyFlagOperation(d->flags, Triggered, UnsetFlags);
}
void InputDevice::Control::expireBindContextAssociationIfChanged() {
    DENG2_GUARD(this);
    if (d->bindContext == d->prevBindContext) return;
    applyFlagOperation(d->flags, Expired, SetFlags);
    applyFlagOperation(d->flags, Triggered, UnsetFlags); // Not any more.
}
#define DEFAULT_STICK_DEADZONE  .05f  ///< 5%
#define DEFAULT_STICK_FACTOR    1.0
#define MAX_AXIS_FILTER  40
static char *joyControllerPreset = (char *) "";
static InputDevice *makeKeyboard(String const &name, String const &title = "") {
    InputDevice *keyboard = new InputDevice(name);
    keyboard->setTitle(title);
    for (int i = 0; i < 256; ++i) {
        char const *shortName = B_ShortNameForKey(i, true);
        String keyName(shortName? shortName : "");
        if (keyName.isEmpty()) {
            keyName = String::format("unbindable%03i", i);
        }
        keyboard->addButton(new ButtonInputControl(keyName));
    }
    return keyboard;
}
static InputDevice *makeMouse(String const &name, String const &title = "") {
    InputDevice *mouse = new InputDevice(name);
    mouse->setTitle(title);
    for (int i = 0; i < IMB_MAXBUTTONS; ++i) {
        mouse->addButton(new ButtonInputControl);
    }
    mouse->button(IMB_LEFT       ).setName("left");
    mouse->button(IMB_MIDDLE     ).setName("middle");
    mouse->button(IMB_RIGHT      ).setName("right");
    mouse->button(IMB_MWHEELUP   ).setName("wheelup");
    mouse->button(IMB_MWHEELDOWN ).setName("wheeldown");
    mouse->button(IMB_MWHEELLEFT ).setName("wheelleft");
    mouse->button(IMB_MWHEELRIGHT).setName("wheelright");
    AxisInputControl *axis;
    mouse->addAxis(axis = new AxisInputControl("x", AxisInputControl::Pointer));
    axis->setScale(10); //1.f/1000);
    mouse->addAxis(axis = new AxisInputControl("y", AxisInputControl::Pointer));
    axis->setScale(10); //1.f/1000);
    return mouse;
}
static InputDevice *makeJoystick(String const &name, String const &title = "") {
    InputDevice *joy = new InputDevice(name);
    joy->setTitle(title);
    for (int i = 0; i < IJOY_MAXBUTTONS; ++i) {
        joy->addButton(new ButtonInputControl(String::format("button%i", i + 1)));
    }
    for (int i = 0; i < IJOY_MAXAXES; ++i) {
        char name[32];
        if (i < 4) {
            strcpy(name, i == 0? "x" : i == 1? "y" : i == 2? "z" : "w");
        }
        else {
            sprintf(name, "axis%02i", i + 1);
        }
        auto *axis = new AxisInputControl(name, AxisInputControl::Stick);
        joy->addAxis(axis);
        axis->setDeadZone(DEFAULT_STICK_DEADZONE);
    }
    for (int i = 0; i < IJOY_MAXHATS; ++i) {
        joy->addHat(new HatInputControl(String::format("hat%i", i + 1)));
    }
    return joy;
}
static InputDevice *makeHeadTracker(String const &name, String const &title) {
    InputDevice *head = new InputDevice(name);
    head->setTitle(title);
    auto *axis = new AxisInputControl("yaw", AxisInputControl::Stick);
    head->addAxis(axis);
    axis->setRawInput();
    head->addAxis(axis = new AxisInputControl("pitch", AxisInputControl::Stick));
    axis->setRawInput();
    head->addAxis(axis = new AxisInputControl("roll", AxisInputControl::Stick));
    axis->setRawInput();
    return head;
}
static Value *Function_InputSystem_BindEvent(Context &, Function::ArgumentValues const &args) {
    String eventDesc = args[0]->asText();
    String command   = args[1]->asText();
    if (ClientApp::inputSystem().bindCommand(eventDesc.toLatin1(), command.toUtf8())) {
        return new NumberValue(true);
    }
    return new NumberValue(false);
}
static Value *Function_InputSystem_BindControl(Context &, Function::ArgumentValues const &args) {
    String control = args[0]->asText();
    String impulse = args[1]->asText();
    if (ClientApp::inputSystem().bindImpulse(control.toLatin1(), impulse.toLatin1())) {
        return new NumberValue(true);
    }
    return new NumberValue(false);
}
#if 0
struct repeater_t {
    int key;           ///< The DDKEY code (@c 0 if not in use).
    int native;        ///< Used to determine which key is repeating.
    char text[8];      ///< Text to insert.
    timespan_t timer;  ///< How's the time?
    int count;         ///< How many times has been repeated?
};
int repWait1 = 15;
int repWait2 = 3;
#endif
bool shiftDown;
bool altDown;
#ifdef OLD_FILTER
static uint mouseFreq;
#endif
static float oldPOV[IJOY_MAXHATS] = {
    IJOY_POV_CENTER, IJOY_POV_CENTER, IJOY_POV_CENTER, IJOY_POV_CENTER
};
static char *eventStrings[MAXEVENTS];
/**
 * Returns a copy of the string @a str. The caller does not get ownership of
 * the string. The string is valid until it gets overwritten by a new
 * allocation. There are at most MAXEVENTS strings allocated at a time.
 *
 * These are intended for strings in ddevent_t that are valid during the
 * processing of an event.
 */
static char const *allocEventString(char const *str) {
    DENG2_ASSERT(str);
    static int eventStringRover = 0;
    DENG2_ASSERT(eventStringRover >= 0 && eventStringRover < MAXEVENTS);
    M_Free(eventStrings[eventStringRover]);
    char const *returnValue = eventStrings[eventStringRover] = strdup(str);
    if (++eventStringRover >= MAXEVENTS) {
        eventStringRover = 0;
    }
    return returnValue;
}
static void clearEventStrings() {
    for (int i = 0; i < MAXEVENTS; ++i) {
        M_Free(eventStrings[i]); eventStrings[i] = nullptr;
    }
}
struct eventqueue_t {
    ddevent_t events[MAXEVENTS];
    int head;
    int tail;
};
/**
 * Gets the next event from an input event queue.
 * @param q  Event queue.
 * @return @c NULL if no more events are available.
 */
static ddevent_t *nextFromQueue(eventqueue_t *q) {
    DENG2_ASSERT(q);
    if (q->head == q->tail)
        return nullptr;
    ddevent_t *ev = &q->events[q->tail];
    q->tail = (q->tail + 1) & (MAXEVENTS - 1);
    return ev;
}
static void clearQueue(eventqueue_t *q) {
    DENG2_ASSERT(q);
    q->head = q->tail;
}
static void postToQueue(eventqueue_t *q, ddevent_t *ev) {
    DENG2_ASSERT(q && ev);
    q->events[q->head] = *ev;
    if (ev->type == E_SYMBOLIC) {
        q->events[q->head].symbolic.name = allocEventString(ev->symbolic.name);
    }
    q->head++;
    q->head &= MAXEVENTS - 1;
}
static eventqueue_t queue;
static eventqueue_t sharpQueue;
static byte useSharpInputEvents = true; ///< cvar
DENG2_PIMPL(InputSystem)
, DENG2_OBSERVES(BindContext, ActiveChange)
, DENG2_OBSERVES(BindContext, AcquireDeviceChange)
, DENG2_OBSERVES(BindContext, BindingAddition) {
    bool ignoreInput = false;
    ConfigProfiles settings;
    Binder binder;
    typedef QList<InputDevice *> Devices;
    Devices devices;
    typedef QList<BindContext *> BindContexts;
    BindContexts contexts;  ///< Ordered from highest to lowest priority.
    std::unique_ptr<ControllerPresets> gameControllerPresets;
    Impl(Public *i) : Base(i) {
        settings.define(ConfigProfiles::ConfigVariable, "input.mouse.syncSensitivity")
                .define(ConfigProfiles::FloatCVar,      "input-mouse-x-scale", 1.f/1000.f)
                .define(ConfigProfiles::FloatCVar,      "input-mouse-y-scale", 1.f/1000.f)
                .define(ConfigProfiles::IntCVar,        "input-mouse-x-flags", 0)
                .define(ConfigProfiles::IntCVar,        "input-mouse-y-flags", 0)
                .define(ConfigProfiles::IntCVar,        "input-joy", 1)
                .define(ConfigProfiles::IntCVar,        "input-sharp", 1);
        binder.initNew()
                << DENG2_FUNC(InputSystem_BindEvent,   "bindEvent",   "event" << "command")
                << DENG2_FUNC(InputSystem_BindControl, "bindControl", "control" << "impulse");
        binder.module().addNumber("DEFAULT_STICK_DEADZONE", DEFAULT_STICK_DEADZONE).setReadOnly();
        binder.module().addNumber("DEFAULT_STICK_FACTOR",   DEFAULT_STICK_FACTOR).setReadOnly();
        App::scriptSystem().addNativeModule("Input", binder.module());
        gameControllerPresets.reset(new ControllerPresets(IDEV_JOY1, "input-joy-preset"));
        I_InitInterfaces();
    }
    ~Impl() {
        self().clearAllContexts();
        clearAllDevices();
        I_ShutdownInterfaces();
    }
    void clearAllDevices() {
        qDeleteAll(devices);
        devices.clear();
    }
    /**
     * @param device  InputDevice to add.
     * @return  Same as @a device for caller convenience.
     */
    InputDevice *addDevice(InputDevice *device) {
        if (device) {
            if (!devices.contains(device)) {
                for (InputDevice *otherDevice : devices) {
                    if (!otherDevice->name().compareWithoutCase(device->name())) {
                        throw Error("InputSystem::addDevice", "Multiple devices with name:" + device->name() + " cannot coexist");
                    }
                }
                devices.append(device);
            }
        }
        return device;
    }
    void echoSymbolicEvent(ddevent_t const &ev) {
        DENG2_ASSERT(symbolicEchoMode);
        if (ev.type == E_SYMBOLIC || ev.type == E_FOCUS) return;
        if (ev.type == E_AXIS) {
            const InputDevice &device = self().device(ev.device);
            const float pos = device.axis(ev.axis.id).translateRealPosition(ev.axis.pos);
            const float markedPos = float(device.axis(ev.axis.id).markedPosition());
            if ((ev.axis.type == EAXIS_ABSOLUTE && fabs(pos - markedPos) < .5f) ||
                (ev.axis.type == EAXIS_RELATIVE && fabs(pos) < .02f)) {
                return; // Not significant enough.
            }
        }
        String name = "echo-" + B_EventToString(ev);
        const Block nameUtf8 = name.toUtf8();
        ddevent_t echo{};
        echo.device        = -1;
        echo.type          = E_SYMBOLIC;
        echo.symbolic.id   = 0;
        echo.symbolic.name = nameUtf8.constData();
        LOG_INPUT_XVERBOSE("Symbolic echo: %s", name);
        self().postEvent(&echo);
    }
    /**
     * Send all the events of the given timestamp down the responder chain.
     */
    void dispatchEvents(eventqueue_t *q, timespan_t ticLength, bool updateAxes = true) {
        bool const callGameResponders = App_GameLoaded();
        ddevent_t *ddev;
        while ((ddev = nextFromQueue(q))) {
            self().trackEvent(*ddev);
            if (ignoreInput && ddev->type != E_FOCUS)
                continue;
            event_t ev;
            bool validGameEvent = false;
            if (callGameResponders) {
                validGameEvent = self().convertEvent(*ddev, ev);
            }
            if (callGameResponders && validGameEvent && gx.PrivilegedResponder) {
                if (gx.PrivilegedResponder(&ev)) {
                    continue;
                }
            }
            if (symbolicEchoMode) {
                echoSymbolicEvent(*ddev);
                continue;
            }
            if (self().tryEvent(*ddev)) {
                continue;
            }
            if (callGameResponders && validGameEvent && gx.FallbackResponder) {
                gx.FallbackResponder(&ev);
            }
        }
        if (updateAxes) {
            for (InputDevice *device : devices) {
                device->forAllControls([&ticLength] (InputControl &ctrl) {
                    if (auto *axis = maybeAs<AxisInputControl>(ctrl)) {
                        axis->update(ticLength);
                    }
                    return LoopContinue;
                });
            }
        }
    }
    /**
     * Poll all event sources (i.e., input devices) and post events.
     */
    void postEventsForAllDevices() {
        readKeyboard();
        readMouse();
        readJoystick();
        readHeadTracker();
    }
    /**
     * Check the current keyboard state, generate input events based on pressed/held
     * keys and poss them.
     *
     * @todo Does not belong at this level.
     */
    void readKeyboard() {
        if (novideo) return;
        ddevent_t ev; de::zap(ev);
        ev.device       = IDEV_KEYBOARD;
        ev.type         = E_TOGGLE;
        ev.toggle.state = ETOG_REPEAT;
        int const QUEUESIZE = 32;
        keyevent_t keyevs[QUEUESIZE];
        size_t const numkeyevs = Keyboard_GetEvents(keyevs, QUEUESIZE);
        for (size_t n = 0; n < numkeyevs; ++n) {
            keyevent_t *ke = &keyevs[n];
            switch (ke->type) {
            case IKE_DOWN:   ev.toggle.state = ETOG_DOWN;   break;
            case IKE_REPEAT: ev.toggle.state = ETOG_REPEAT; break;
            case IKE_UP:     ev.toggle.state = ETOG_UP;     break;
            default: break;
            }
            ev.toggle.id = ke->ddkey;
            DENG2_ASSERT(sizeof(ev.toggle.text) == sizeof(ke->text));
            std::memcpy(ev.toggle.text, ke->text, sizeof(ev.toggle.text));
            LOG_INPUT_XVERBOSE("toggle.id: %i/%c [%s:%u] (state:%i)",
                       ev.toggle.id << char(ev.toggle.id)
                    << ev.toggle.text << strlen(ev.toggle.text)
                    << ev.toggle.state);
            self().postEvent(&ev);
        }
    }
    /**
     * Check the current mouse state (axis, buttons and wheel).
     * Generates events and mickeys and posts them.
     *
     * @todo Does not belong at this level.
     */
    void readMouse() {
        if (!Mouse_IsPresent())
            return;
        mousestate_t mouse;
#ifdef OLD_FILTER
        if (mouseFreq > 0) {
            static uint lastTime = 0;
            uint nowTime = Timer_RealMilliseconds();
            if (nowTime - lastTime < 1000 / mouseFreq) {
                std::memset(&mouse, 0, sizeof(mouse));
            }
            else {
                lastTime = nowTime;
                Mouse_GetState(&mouse);
            }
        }
        else
#endif
        {
            Mouse_GetState(&mouse);
        }
        ddevent_t ev; de::zap(ev);
        ev.device = IDEV_MOUSE;
        ev.type   = E_AXIS;
        float xpos = mouse.axis[IMA_POINTER].x;
        float ypos = mouse.axis[IMA_POINTER].y;
        /*if (uiMouseMode)
        {
            ev.axis.type = EAXIS_ABSOLUTE;
        }
        else*/
        {
            ev.axis.type = EAXIS_RELATIVE;
            ypos = -ypos;
        }
        if (xpos) {
            ev.axis.id  = 0;
            ev.axis.pos = xpos;
            self().postEvent(&ev);
        }
        if (ypos) {
            ev.axis.id  = 1;
            ev.axis.pos = ypos;
            self().postEvent(&ev);
        }
        int i = 0;
        for (; i < IMB_MAXBUTTONS; ++i) {
            if (mouse.buttonDowns[i] || mouse.buttonUps[i])
                break;
        }
        if (i < IMB_MAXBUTTONS) {
            for (i = 0; i < IMB_MAXBUTTONS; ++i) {
                LOGDEV_INPUT_XVERBOSE("[%02i] %i/%i", i << mouse.buttonDowns[i] << mouse.buttonUps[i]);
            }
        }
        ev.type = E_TOGGLE;
        for (i = 0; i < IMB_MAXBUTTONS; ++i) {
            ev.toggle.id = i;
            while (mouse.buttonDowns[i] > 0 || mouse.buttonUps[i] > 0) {
                if (mouse.buttonDowns[i]-- > 0) {
                    ev.toggle.state = ETOG_DOWN;
                    LOG_INPUT_XVERBOSE("Mouse button %i down", i);
                    self().postEvent(&ev);
                }
                if (mouse.buttonUps[i]-- > 0) {
                    ev.toggle.state = ETOG_UP;
                    LOG_INPUT_XVERBOSE("Mouse button %i up", i);
                    self().postEvent(&ev);
                }
            }
        }
    }
    /**
     * Check the current joystick state (axis, sliders, hat and buttons).
     * Generates events and posts them. Axis clamps and dead zone is done
     * here.
     *
     * @todo Does not belong at this level.
     */
    void readJoystick() {
        if (!Joystick_IsPresent())
            return;
        joystate_t state;
        Joystick_GetState(&state);
        ddevent_t ev; de::zap(ev);
        ev.device = IDEV_JOY1;
        ev.type   = E_TOGGLE;
        for (int i = 0; i < state.numButtons; ++i) {
            ev.toggle.id = i;
            while (state.buttonDowns[i] > 0 || state.buttonUps[i] > 0) {
                if (state.buttonDowns[i]-- > 0) {
                    ev.toggle.state = ETOG_DOWN;
                    self().postEvent(&ev);
                    LOG_INPUT_XVERBOSE("Joy button %i down", i);
                }
                if (state.buttonUps[i]-- > 0) {
                    ev.toggle.state = ETOG_UP;
                    self().postEvent(&ev);
                    LOG_INPUT_XVERBOSE("Joy button %i up", i);
                }
            }
        }
        if (state.numHats > 0) {
            for (int i = 0; i < state.numHats; ++i) {
                if (!fequal(state.hatAngle[i], oldPOV[i])) {
                    ev.type = E_ANGLE;
                    ev.angle.id = i;
                    if (state.hatAngle[i] < 0) {
                        ev.angle.pos = -1;
                    }
                    else {
                        ev.angle.pos = de::roundf(state.hatAngle[i] / 45);
                    }
                    LOG_INPUT_XVERBOSE("Joy hat %i angle %f", i << ev.angle.pos);
                    self().postEvent(&ev);
                    oldPOV[i] = state.hatAngle[i];
                }
            }
        }
        ev.type = E_AXIS;
        for (int i = 0; i < state.numAxes; ++i) {
            ev.axis.id   = i;
            ev.axis.pos  = state.axis[i];
            ev.axis.type = EAXIS_ABSOLUTE;
            self().postEvent(&ev);
        }
    }
    void readHeadTracker() {
        if (!DD_GetInteger(DD_USING_HEAD_TRACKING)) {
            self().device(IDEV_HEAD_TRACKER).deactivate();
            return;
        }
        self().device(IDEV_HEAD_TRACKER).activate();
        ddevent_t ev; de::zap(ev);
        ev.device    = IDEV_HEAD_TRACKER;
        ev.type      = E_AXIS;
        ev.axis.type = EAXIS_ABSOLUTE;
        Vector3f const pry = vrCfg().oculusRift().headOrientation();
        ev.axis.id  = 0; // Yaw.
        ev.axis.pos = de::radianToDegree(pry[2]) * 1.0 / 180.0;
        self().postEvent(&ev);
        ev.axis.id  = 1; // Pitch (1.0 means 85 degrees).
        ev.axis.pos = de::radianToDegree(pry[0]) * 1.0 / 85.0;
        self().postEvent(&ev);
        ev.axis.id  = 2; // Roll.
        ev.axis.pos = de::radianToDegree(pry[1]) * 1.0 / 180.0;
        self().postEvent(&ev);
    }
    /**
     * Mark all device states with the highest-priority binding context to which they
     * have a connection via device bindings. This ensures that if a high-priority context
     * is using a particular device state, lower-priority contexts will not be using the
     * same state for their own controls.
     *
     * Called automatically whenever a context is activated or deactivated.
     */
    void updateAllDeviceStateAssociations() {
        for (InputDevice *device : devices) {
            device->forAllControls([] (InputControl &ctrl) {
                ctrl.clearBindContextAssociation();
                return LoopContinue;
            });
        }
        for (BindContext *context : contexts) {
            if (!context->isActive())
                continue;
            context->forAllCommandBindings([this, &context] (Record &rec) {
                CommandBinding bind(rec);
                InputControl *ctrl = nullptr;
                switch (bind.geti("type")) {
                case E_AXIS:   ctrl = &self().device(bind.geti("deviceId")).axis  (bind.geti("controlId")); break;
                case E_TOGGLE: ctrl = &self().device(bind.geti("deviceId")).button(bind.geti("controlId")); break;
                case E_ANGLE:  ctrl = &self().device(bind.geti("deviceId")).hat   (bind.geti("controlId")); break;
                case E_SYMBOLIC: break;
                default:
                    DENG2_ASSERT(!"InputSystem::updateAllDeviceStateAssociations: Invalid bind.type");
                    break;
                }
                if (ctrl && !ctrl->hasBindContext()) {
                    ctrl->setBindContext(context);
                }
                return LoopContinue;
            });
            context->forAllImpulseBindings([this, &context] (CompiledImpulseBindingRecord &rec) {
                auto const &bind = rec.compiled();
                InputDevice &dev = self().device(bind.deviceId);
                InputControl *ctrl = nullptr;
                switch (bind.type) {
                case IBD_AXIS:   ctrl = &dev.axis  (bind.controlId); break;
                case IBD_TOGGLE: ctrl = &dev.button(bind.controlId); break;
                case IBD_ANGLE:  ctrl = &dev.hat   (bind.controlId); break;
                default:
                    DENG2_ASSERT(!"InputSystem::updateAllDeviceStateAssociations: Invalid bind.type");
                    break;
                }
                if (ctrl && !ctrl->hasBindContext()) {
                    ctrl->setBindContext(context);
                }
                return LoopContinue;
            });
            for (int i = 0; i < devices.count(); ++i) {
                InputDevice *device = devices.at(i);
                int const deviceId = i;
                if (device->isActive() && context->willAcquire(deviceId)) {
                    device->forAllControls([&context] (InputControl &ctrl) {
                        if (!ctrl.hasBindContext()) {
                            ctrl.setBindContext(context);
                        }
                        return LoopContinue;
                    });
                }
            };
        }
        for (InputDevice *device : devices) {
            device->forAllControls([] (InputControl &ctrl) {
                if (!ctrl.inDefaultState()) {
                    ctrl.expireBindContextAssociationIfChanged();
                }
                return LoopContinue;
            });
        };
    }
    /**
     * When the active state of a binding context changes we'll re-evaluate the
     * effective bindings to avoid wasting time by looking them up repeatedly.
     */
    void bindContextActiveChanged(BindContext &context) {
        updateAllDeviceStateAssociations();
        for (int i = 0; i < devices.count(); ++i) {
            InputDevice *device = devices.at(i);
            int const deviceId  = i;
            if (context.willAcquire(deviceId)) {
                device->reset();
            }
        }
    }
    void bindContextAcquireDeviceChanged(BindContext &) {
        updateAllDeviceStateAssociations();
    }
    void bindContextBindingAdded(BindContext &, Record & /*binding*/, bool /*isCommand*/)
    {
        updateAllDeviceStateAssociations();
    }
};
InputSystem::InputSystem() : d(new Impl(this)) {
    initAllDevices();
}
void InputSystem::timeChanged(const Clock &) {
    if (!symbolicEchoMode) {
        for (auto *device : d->devices) {
            for (int i = 0; i < device->axisCount(); ++i) {
                device->axis(i).markPosition();
            }
        }
    }
}
ConfigProfiles &InputSystem::settings() {
    return d->settings;
}
InputDevice &InputSystem::device(int id) const {
    if (id >= 0 && id < d->devices.count()) {
        return *d->devices.at(id);
    }
    throw MissingDeviceError("InputSystem::device", "Unknown id:" + String::number(id));
}
InputDevice *InputSystem::devicePtr(int id) const {
    if (id >= 0 && id < d->devices.count()) {
        return d->devices.at(id);
    }
    return nullptr;
}
InputDevice *InputSystem::findDevice(String const &name) const {
    for (auto *device : d->devices) {
        if (device->name() == name) {
            return device;
        }
    }
    return nullptr;
}
LoopResult InputSystem::forAllDevices(std::function<LoopResult (InputDevice &)> func) const {
    for (InputDevice *device : d->devices) {
        if (auto result = func(*device)) return result;
    }
    return LoopContinue;
}
int InputSystem::deviceCount() const {
    return d->devices.count();
}
void InputSystem::initAllDevices() {
    d->clearAllDevices();
    d->addDevice(makeKeyboard("key", "Keyboard"))->activate(); // A keyboard is assumed to always be present.
    d->addDevice(makeMouse("mouse", "Mouse"))->activate(Mouse_IsPresent()); // A mouse may not be present.
    d->addDevice(makeJoystick("joy", "Controller"))->activate(Joystick_IsPresent()); // A joystick may not be present.
    d->addDevice(new InputDevice("joy2"));
    d->addDevice(new InputDevice("joy3"));
    d->addDevice(new InputDevice("joy4"));
    d->addDevice(makeHeadTracker("head", "Head Tracker")); // Head trackers are activated later.
    for (InputDevice *device : d->devices) device->consoleRegister();
}
bool InputSystem::ignoreEvents(bool yes) {
    LOG_AS("InputSystem");
    bool const oldIgnoreInput = d->ignoreInput;
    d->ignoreInput = yes;
    LOG_INPUT_VERBOSE("Ignoring events: %b") << yes;
    if (!yes) {
        d->postEventsForAllDevices();
        clearEvents();
        for (InputDevice *dev : d->devices) dev->reset();
    }
    return oldIgnoreInput;
}
void InputSystem::clearEvents() {
    clearQueue(&queue);
    clearQueue(&sharpQueue);
    clearEventStrings();
}
void InputSystem::postEvent(ddevent_t *ev) {
    DENG2_ASSERT(ev);// && ev->device < NUM_INPUT_DEVICES);
    eventqueue_t *q = &queue;
    if (useSharpInputEvents &&
       (ev->type == E_TOGGLE || ev->type == E_AXIS || ev->type == E_ANGLE)) {
        q = &sharpQueue;
    }
    if (ev->type == E_TOGGLE && ev->device != IDEV_KEYBOARD) {
        std::memset(ev->toggle.text, 0, sizeof(ev->toggle.text));
    }
    postToQueue(q, ev);
}
void InputSystem::processEvents(timespan_t ticLength) {
    d->postEventsForAllDevices();
    d->dispatchEvents(&queue, ticLength, !useSharpInputEvents);
}
void InputSystem::processSharpEvents(timespan_t ticLength) {
    if (DD_IsSharpTick() || !DD_IsFrameTimeAdvancing()) {
        d->dispatchEvents(&sharpQueue, DD_IsFrameTimeAdvancing()? SECONDSPERTIC : ticLength);
    }
}
bool InputSystem::tryEvent(ddevent_t const &event, String const &namedContext) {
    if (namedContext.isEmpty()) {
        for (BindContext const *context : d->contexts) {
            if (context->tryEvent(event)) return true;
        }
        return false;
    }
    if (hasContext(namedContext)) {
        return context(namedContext).tryEvent(event, false/*this context only*/);
    }
    return false;
}
bool InputSystem::tryEvent(Event const &event, String const &namedContext) {
    ddevent_t ddev;
    convertEvent(event, ddev);
    return tryEvent(ddev, namedContext);
}
void InputSystem::trackEvent(ddevent_t const &event) {
    if (event.type == E_FOCUS || event.type == E_SYMBOLIC)
        return; // Not a tracked device state.
    InputDevice *dev = devicePtr(event.device);
    if (!dev || !dev->isActive()) return;
    if (event.device == IDEV_KEYBOARD && event.type == E_TOGGLE) {
        if (event.toggle.id == DDKEY_RSHIFT) {
            if (event.toggle.state == ETOG_DOWN) {
                ::shiftDown = true;
            }
            else if (event.toggle.state == ETOG_UP) {
                ::shiftDown = false;
            }
        }
        else if (event.toggle.id == DDKEY_RALT) {
            if (event.toggle.state == ETOG_DOWN) {
                ::altDown = true;
            }
            else if (event.toggle.state == ETOG_UP) {
                ::altDown = false;
            }
        }
    }
    switch (event.type) {
    case E_AXIS:
        dev->axis(event.axis.id).applyRealPosition(event.axis.pos);
        break;
    case E_TOGGLE:
        dev->button(event.toggle.id).setDown(event.toggle.state == ETOG_DOWN || event.toggle.state == ETOG_REPEAT);
        break;
    case E_ANGLE:
        dev->hat(event.angle.id).setPosition(event.angle.pos);
        break;
    default: break;
    }
}
void InputSystem::trackEvent(Event const &event) {
    ddevent_t ddev;
    convertEvent(event, ddev);
    trackEvent(ddev);
}
bool InputSystem::convertEvent(Event const &from, ddevent_t &to) // static
{
    de::zap(to);
    switch (from.type()) {
    case Event::KeyPress:
    case Event::KeyRelease: {
        KeyEvent const &kev = from.as<KeyEvent>();
        to.device       = IDEV_KEYBOARD;
        to.type         = E_TOGGLE;
        to.toggle.id    = kev.ddKey();
        to.toggle.state = (kev.state() == KeyEvent::Pressed? ETOG_DOWN : ETOG_UP);
        qstrcpy(to.toggle.text, kev.text().toLatin1());
        break; }
    default: break;
    }
    return true;
}
bool InputSystem::convertEvent(ddevent_t const &from, event_t &to) // static
{
    de::zap(to);
    if (from.type == E_SYMBOLIC) {
        to.type = EV_SYMBOLIC;
#ifdef __64BIT__
        ASSERT_64BIT(from.symbolic.name);
        to.data_u64 = (duint64) from.symbolic.name;
#else
        ASSERT_NOT_64BIT(from.symbolic.name);
        to.data1 = (int) from.symbolic.name;
        to.data2 = 0;
#endif
    }
    else if (from.type == E_FOCUS) {
        to.type  = EV_FOCUS;
        to.data1 = from.focus.gained;
        to.data2 = from.focus.inWindow;
    }
    else {
        switch (from.device) {
        case IDEV_KEYBOARD:
            to.type = EV_KEY;
            if (from.type == E_TOGGLE) {
                to.state = (  from.toggle.state == ETOG_UP  ? EVS_UP
                            : from.toggle.state == ETOG_DOWN? EVS_DOWN
                            : EVS_REPEAT );
                to.data1 = from.toggle.id;
            }
            break;
        case IDEV_MOUSE:
            if (from.type == E_AXIS) {
                to.type = EV_MOUSE_AXIS;
            }
            else if (from.type == E_TOGGLE) {
                to.type  = EV_MOUSE_BUTTON;
                to.data1 = from.toggle.id;
                to.state = (  from.toggle.state == ETOG_UP  ? EVS_UP
                            : from.toggle.state == ETOG_DOWN? EVS_DOWN
                            : EVS_REPEAT );
            }
            break;
        case IDEV_JOY1:
        case IDEV_JOY2:
        case IDEV_JOY3:
        case IDEV_JOY4:
            if (from.type == E_AXIS) {
                int *data = &to.data1;
                to.type  = EV_JOY_AXIS;
                to.state = (evstate_t) 0;
                if (from.axis.id >= 0 && from.axis.id < 6) {
                    data[from.axis.id] = from.axis.pos;
                }
            }
            else if (from.type == E_TOGGLE) {
                to.type  = EV_JOY_BUTTON;
                to.state = (  from.toggle.state == ETOG_UP  ? EVS_UP
                            : from.toggle.state == ETOG_DOWN? EVS_DOWN
                            : EVS_REPEAT );
                to.data1 = from.toggle.id;
            }
            else if (from.type == E_ANGLE) {
                to.type = EV_POV;
            }
            break;
        case IDEV_HEAD_TRACKER:
            return false;
        default:
            DENG2_ASSERT(!"InputSystem::convertEvent: Unknown device ID in ddevent_t");
            return false;
        }
    }
    return true;
}
bool InputSystem::shiftDown() const {
    return ::shiftDown;
}
void InputSystem::initialContextActivations() {
    for (BindContext *context : d->contexts) context->deactivate();
    context(GLOBAL_BINDING_CONTEXT_NAME ).activate();
    context(DEFAULT_BINDING_CONTEXT_NAME).activate();
}
void InputSystem::clearAllContexts() {
    if (d->contexts.isEmpty()) return;
    qDeleteAll(d->contexts);
    d->contexts.clear();
    Binding::resetIdentifiers();
}
int InputSystem::contextCount() const {
    return d->contexts.count();
}
ControllerPresets &InputSystem::gameControllerPresets() {
    return *d->gameControllerPresets;
}
bool InputSystem::hasContext(String const &name) const {
    return contextPtr(name) != nullptr;
}
BindContext &InputSystem::context(String const &name) const {
    if (BindContext *context = contextPtr(name)) {
        return *context;
    }
    throw MissingContextError("InputSystem::context", "Unknown name:" + name);
}
BindContext *InputSystem::contextPtr(String const &name) const {
    for (BindContext const *context : d->contexts) {
        if (!context->name().compareWithoutCase(name)) {
            return const_cast<BindContext *>(context);
        }
    }
    return nullptr;
}
BindContext &InputSystem::contextAt(int position) const {
    if (position >= 0 && position < d->contexts.count()) {
        return *d->contexts.at(position);
    }
    throw MissingContextError("InputSystem::contextAt", "Invalid position:" + String::number(position));
}
int InputSystem::contextPositionOf(BindContext *context) const {
    return (context? d->contexts.indexOf(context) : -1);
}
BindContext *InputSystem::newContext(String const &name) {
    if (hasContext(name)) {
        LOG_AS("InputSystem::newContext");
        LOG_INPUT_WARNING("Name '%s' is not unique - Won't replace") << name;
        return nullptr;
    }
    BindContext *context = new BindContext(name);
    d->contexts.prepend(context);
    context->audienceForActiveChange()        += d;
    context->audienceForAcquireDeviceChange() += d;
    context->audienceForBindingAddition()     += d;
    return context;
}
LoopResult InputSystem::forAllContexts(std::function<LoopResult (BindContext &)> func) const {
    for (BindContext *context : d->contexts) {
        if (auto result = func(*context)) return result;
    }
    return LoopContinue;
}
InputSystem &InputSystem::get() // static
{
    return ClientApp::inputSystem();
}
void InputSystem::bindDefaults() {
    bindCommand("global:key-f11-down + key-alt-down", "releasemouse");
    bindCommand("global:key-f11-down", "togglefullscreen");
    bindCommand("global:key-tilde-down + key-shift-up", "taskbar");
#ifdef WIN32
    bindCommand("global:key-winmenu-down", "taskbar");
#endif
    bindCommand("console:key-tilde-down + key-shift-up", "taskbar"); // without this, key would be entered into command line
}
void InputSystem::bindGameDefaults() {
    if (!App_GameLoaded()) return;
    Con_Executef(CMDS_DDAY, false, "defaultgamebindings");
}
static char const *parseContext(String &context, char const *desc) {
    DENG2_ASSERT(desc);
    if (!strchr(desc, ':')) {
        context = "";
        return desc;
    }
    AutoStr *str = AutoStr_NewStd();
    desc = Str_CopyDelim(str, desc, ':');
    context = Str_Text(str);
    return desc;
}
Record *InputSystem::bindCommand(char const *eventDesc, char const *command) {
    DENG2_ASSERT(eventDesc && command);
    LOG_AS("InputSystem");
    String contextName;
    eventDesc = parseContext(contextName, eventDesc);
    if (BindContext *context = contextPtr(contextName.isEmpty()? DEFAULT_BINDING_CONTEXT_NAME : contextName)) {
        return context->bindCommand(eventDesc, command);
    }
    return nullptr;
}
Record *InputSystem::bindImpulse(char const *ctrlDesc, char const *impulseDesc) {
    DENG2_ASSERT(ctrlDesc && impulseDesc);
    LOG_AS("InputSystem");
    int localPlayer = 0;
    AutoStr *str    = AutoStr_NewStd();
    char const *ptr = Str_CopyDelim(str, impulseDesc, '-');
    if (!qstrnicmp(Str_Text(str), "local", 5) && Str_Length(str) > 5) {
        localPlayer = String((Str_Text(str) + 5)).toInt() - 1;
        if (localPlayer < 0 || localPlayer >= DDMAXPLAYERS) {
            LOG_INPUT_WARNING("Local player number %i is invalid") << localPlayer;
            return nullptr;
        }
        impulseDesc = ptr;
    }
    impulseDesc = Str_CopyDelim(str, impulseDesc, '-');
    PlayerImpulse const *impulse = P_PlayerImpulseByName(Str_Text(str));
    if (!impulse) {
        LOG_INPUT_WARNING("Player impulse \"%s\" not defined") << Str_Text(str);
        return nullptr;
    }
    BindContext *context = contextPtr(impulse->bindContextName);
    if (!context) {
        context = contextPtr(DEFAULT_BINDING_CONTEXT_NAME);
    }
    DENG2_ASSERT(context);
    return context->bindImpulse(ctrlDesc, *impulse, localPlayer);
}
bool InputSystem::removeBinding(int id) {
    LOG_AS("InputSystem");
    for (BindContext *context : d->contexts) {
        if (bool result = context->deleteBinding(id)) return result;
    }
    return false;
}
void InputSystem::removeAllBindings() {
    for (BindContext *context : d->contexts) {
        context->clearAllBindings();
    };
    Binding::resetIdentifiers();
}
void InputSystem::removeBindingsForDevice(int deviceId) {
    for (auto *context : d->contexts) {
        context->clearBindingsForDevice(deviceId);
    }
}
D_CMD(ListDevices) {
    DENG2_UNUSED3(src, argc, argv);
    InputSystem &isys = ClientApp::inputSystem();
    LOG_INPUT_MSG(_E(b) "%i input devices initalized:") << isys.deviceCount();
    isys.forAllDevices([] (InputDevice &device) {
        LOG_INPUT_MSG("") << device.description();
        return LoopContinue;
    });
    return true;
}
D_CMD(InspectDevice) {
    DENG2_UNUSED2(src, argc);
    if (InputDevice const *device = ClientApp::inputSystem().findDevice(argv[1])) {
        LOG_INPUT_MSG("") << device->description();
        return true;
    }
    else {
        return false;
    }
}
D_CMD(ReleaseMouse) {
    DENG2_UNUSED3(src, argc, argv);
    if (WindowSystem::mainExists()) {
        ClientWindowSystem::main().eventHandler().trapMouse(false);
        return true;
    }
    return false;
}
D_CMD(ListContexts) {
    DENG2_UNUSED3(src, argc, argv);
    InputSystem &isys = ClientApp::inputSystem();
    LOG_INPUT_MSG(_E(b) "%i binding contexts defined:") << isys.contextCount();
    int idx = 0;
    isys.forAllContexts([&idx] (BindContext &context) {
        LOG_INPUT_MSG("  [%2i] " _E(>) "%s%s" _E(.) _E(2) " (%s)")
                << (idx++) << (context.isActive()? _E(b) : _E(w))
                << context.name()
                << (context.isActive()? "active" : "inactive");
        return LoopContinue;
    });
    return true;
}
/*
D_CMD(ClearContexts) {
    ClientApp::inputSystem().clearAllContexts();
    return true;
}
*/
D_CMD(ActivateContext) {
    DENG2_UNUSED2(src, argc);
    InputSystem &isys = ClientApp::inputSystem();
    bool const doActivate = !String(argv[0]).compareWithoutCase("activatebcontext");
    String const name     = argv[1];
    if (!isys.hasContext(name)) {
        LOG_INPUT_WARNING("Unknown binding context '%s'") << name;
        return false;
    }
    BindContext &context = isys.context(name);
    if (context.isProtected()) {
        LOG_INPUT_ERROR("Binding context " _E(b) "'%s'" _E(.) " is " _E(2) "protected" _E(.)
                        " and cannot be manually %s")
                << context.name() << (doActivate? "activated" : "deactivated");
        return false;
    }
    context.activate(doActivate);
    return true;
}
D_CMD(BindCommand) {
    DENG2_UNUSED2(src, argc);
    return ClientApp::inputSystem().bindCommand(argv[1], argv[2]) != nullptr;
}
D_CMD(BindImpulse) {
    DENG2_UNUSED2(src, argc);
    return ClientApp::inputSystem().bindImpulse(argv[2], argv[1]) != nullptr;
}
D_CMD(ListBindings) {
    DENG2_UNUSED3(src, argc, argv);
    InputSystem &isys = ClientApp::inputSystem();
    int totalBindCount = 0;
    isys.forAllContexts([&totalBindCount] (BindContext &context) {
        totalBindCount += context.commandBindingCount() + context.impulseBindingCount();
        return LoopContinue;
    });
    LOG_INPUT_MSG(_E(b) "Bindings");
    LOG_INPUT_MSG("There are " _E(b) "%i" _E(.) " bindings, in " _E(b) "%i" _E(.) " contexts")
            << totalBindCount << isys.contextCount();
    isys.forAllContexts([] (BindContext &context) {
        int const cmdCount = context.commandBindingCount();
        int const impCount = context.impulseBindingCount();
        if (cmdCount + impCount == 0) return LoopContinue;
        LOG_INPUT_MSG(_E(D)_E(b) "%s" _E(.) " context: " _E(l) "(%i %s)")
                << context.name()
                << (cmdCount + impCount)
                << (context.isActive()? "active" : "inactive");
        if (cmdCount) {
            LOG_INPUT_MSG("  " _E(b) "%i command bindings:") << cmdCount;
        }
        context.forAllCommandBindings([] (Record &rec) {
            CommandBinding bind(rec);
            LOG_INPUT_MSG("    [%3i] " _E(>) _E(b) "%s" _E(.) " %s")
                    << bind.geti("id") << bind.composeDescriptor() << bind.gets("command");
            return LoopContinue;
        });
        if (impCount) {
            LOG_INPUT_MSG("  " _E(b) "%i impulse bindings:") << impCount;
        }
        for (int pl = 0; pl < DDMAXPLAYERS; ++pl)
        context.forAllImpulseBindings(pl, [&pl] (CompiledImpulseBindingRecord &rec) {
            ImpulseBinding bind(rec);
            PlayerImpulse const *impulse = P_PlayerImpulsePtr(bind.geti("impulseId"));
            DENG2_ASSERT(impulse);
            LOG_INPUT_MSG("    [%3i] " _E(>) _E(b) "%s" _E(.) " player%i %s")
                    << bind.geti("id") << bind.composeDescriptor() << (pl + 1) << impulse->name;
            return LoopContinue;
        });
        return LoopContinue;
    });
    return true;
}
D_CMD(ClearBindings) {
    DENG2_UNUSED3(src, argc, argv);
    ClientApp::inputSystem().removeAllBindings();
    return true;
}
D_CMD(RemoveBinding) {
    DENG2_UNUSED2(src, argc);
    int const id = String(argv[1]).toInt();
    if (ClientApp::inputSystem().removeBinding(id)) {
        LOG_INPUT_MSG("Binding " _E(b) "%i" _E(.) " deleted") << id;
    }
    else {
        LOG_INPUT_ERROR("Unknown binding #%i") << id;
    }
    return true;
}
D_CMD(DefaultBindings) {
    DENG2_UNUSED3(src, argc, argv);
    InputSystem &isys = ClientApp::inputSystem();
    isys.bindDefaults();
    isys.bindGameDefaults();
    return true;
}
void InputSystem::consoleRegister() // static
{
#define PROTECTED_FLAGS     (CMDF_NO_DEDICATED | CMDF_DED | CMDF_CLIENT)
    C_VAR_BYTE   ("input-conflict-zerocontrol", &zeroControlUponConflict, 0, 0, 1);
    C_VAR_BYTE   ("input-sharp",                &useSharpInputEvents,     0, 0, 1);
    C_VAR_CHARPTR("input-joy-preset",           &joyControllerPreset,     0, 0, 0);
    C_CMD_FLAGS("activatebcontext",     "s",        ActivateContext,    PROTECTED_FLAGS);
    C_CMD_FLAGS("bindevent",            "ss",       BindCommand,        PROTECTED_FLAGS);
    C_CMD_FLAGS("bindcontrol",          "ss",       BindImpulse,        PROTECTED_FLAGS);
    C_CMD_FLAGS("clearbindings",        "",         ClearBindings,      PROTECTED_FLAGS);
    C_CMD_FLAGS("deactivatebcontext",   "s",        ActivateContext,    PROTECTED_FLAGS);
    C_CMD_FLAGS("defaultbindings",      "",         DefaultBindings,    PROTECTED_FLAGS);
    C_CMD_FLAGS("delbind",              "i",        RemoveBinding,      PROTECTED_FLAGS);
    C_CMD_FLAGS("listbcontexts",        nullptr,    ListContexts,       PROTECTED_FLAGS);
    C_CMD_FLAGS("listbindings",         nullptr,    ListBindings,       PROTECTED_FLAGS);
    C_CMD      ("listinputdevices",     "",         ListDevices);
    C_CMD      ("inspectinputdevice",   "s",        InspectDevice);
    C_CMD      ("releasemouse",         "",         ReleaseMouse);
#ifdef DENG2_DEBUG
    I_DebugDrawerConsoleRegister();
#endif
#undef PROTECTED_FLAGS
}
#undef B_SetContextFallback
DENG_EXTERN_C void B_SetContextFallback(char const *name, int (*responderFunc)(event_t *)) {
    InputSystem &isys = ClientApp::inputSystem();
    if (isys.hasContext(name)) {
        isys.context(name).setFallbackResponder(responderFunc);
    }
}
#undef B_BindingsForCommand
DENG_EXTERN_C int B_BindingsForCommand(char const *commandCString, char *outBuf, size_t outBufSize) {
    DENG2_ASSERT(commandCString && outBuf);
    String const command = commandCString;
    *outBuf = 0;
    if (command.isEmpty()) return 0;
    if (!outBufSize) return 0;
    InputSystem &isys = ClientApp::inputSystem();
    String out;
    int numFound = 0;
    isys.forAllContexts([&] (BindContext &context) {
        context.forAllCommandBindings([&] (Record &rec) {
            CommandBinding bind(rec);
            if (!bind.gets("command").compareWithCase(command)) {
                if (numFound) out += " "; // Separator.
                out += String::number(bind.geti("id")) + "@" + context.name() + ":" + bind.composeDescriptor();
                numFound++;
            }
            return LoopContinue;
        });
        return LoopContinue;
    });
    std::memset(outBuf, 0, outBufSize);
    qstrncpy(outBuf, out.toUtf8().constData(), outBufSize - 1);
    return numFound;
}
#undef B_BindingsForControl
DENG_EXTERN_C int B_BindingsForControl(int localPlayer, char const *impulseNameCString, int inverse,
    char *outBuf, size_t outBufSize) {
    DENG2_ASSERT(impulseNameCString && outBuf);
    String const impulseName = impulseNameCString;
    *outBuf = 0;
    if (localPlayer < 0 || localPlayer >= DDMAXPLAYERS) return 0;
    if (impulseName.isEmpty()) return 0;
    if (!outBufSize) return 0;
    InputSystem &isys = ClientApp::inputSystem();
    String out;
    int numFound = 0;
    isys.forAllContexts([&] (BindContext &context) {
        context.forAllImpulseBindings(localPlayer, [&] (CompiledImpulseBindingRecord &rec) {
            auto const &bind = rec.compiled();
            DENG2_ASSERT(bind.localPlayer == localPlayer);
            PlayerImpulse const *impulse = P_PlayerImpulsePtr(bind.impulseId);
            DENG2_ASSERT(impulse);
            if (!impulse->name.compareWithoutCase(impulseName)) {
                if (inverse == BFCI_BOTH ||
                   (inverse == BFCI_ONLY_NON_INVERSE && !(bind.flags & IBDF_INVERSE)) ||
                   (inverse == BFCI_ONLY_INVERSE     &&  (bind.flags & IBDF_INVERSE))) {
                    if (numFound) out += " ";
                    out += String::number(bind.id) + "@" + context.name() + ":" +
                            ImpulseBinding(rec).composeDescriptor();
                    numFound++;
                }
            }
            return LoopContinue;
        });
        return LoopContinue;
    });
    std::memset(outBuf, 0, outBufSize);
    qstrncpy(outBuf, out.toUtf8().constData(), outBufSize - 1);
    return numFound;
}
#undef DD_GetKeyCode
DENG_EXTERN_C int DD_GetKeyCode(char const *key) {
    DENG2_ASSERT(key);
    int code = B_KeyForShortName(key);
    return (code? code : key[0]);
}
DENG_DECLARE_API(B) = {
    { DE_API_BINDING },
    B_SetContextFallback,
    B_BindingsForCommand,
    B_BindingsForControl,
    DD_GetKeyCode
};
typedef struct clicker_s {
    int down;                   // Count for down events.
    int up;                     // Count for up events.
} clicker_t;
static struct { int dx, dy; } mouseDelta[IMA_MAXAXES];
static clicker_t mouseClickers[IMB_MAXBUTTONS];
static bool mouseTrapped = false;
static bool cursorHidden = false;
static QPoint prevMousePos;
static int Mouse_Qt_Init(void) {
    memset(&mouseDelta, 0, sizeof(mouseDelta));
    memset(&mouseClickers, 0, sizeof(mouseClickers));
    mouseTrapped = false;
    cursorHidden = false;
    prevMousePos = QPoint();
    return true;
}
static void Mouse_Qt_Shutdown(void) {
}
static void Mouse_Qt_Poll() {
    if (!mouseTrapped) return;
#if !defined (DENG_MOBILE)
    ClientWindow *win = ClientWindowSystem::mainPtr();
    if (!win) return; // Hmm?
    QPoint curPos = win->mapFromGlobal(QCursor::pos());
    if (!prevMousePos.isNull()) {
        QPoint delta = curPos - prevMousePos;
        if (!delta.isNull()) {
            Mouse_Qt_SubmitMotion(IMA_POINTER, delta.x(), delta.y());
            QPoint mid(win->pointWidth() / 2, win->pointHeight() / 2);
            QCursor::setPos(win->mapToGlobal(mid));
            prevMousePos = mid;
        }
    }
    else {
        prevMousePos = curPos;
    }
#endif
}
static void Mouse_Qt_GetState(mousestate_t *state) {
    int i;
    memset(state, 0, sizeof(*state));
    for (i = 0; i < IMA_MAXAXES; ++i) {
        state->axis[i].x = mouseDelta[i].dx;
        state->axis[i].y = mouseDelta[i].dy;
        mouseDelta[i].dx = mouseDelta[i].dy = 0;
    }
    for (i = 0; i < IMB_MAXBUTTONS; ++i) {
        state->buttonDowns[i] = mouseClickers[i].down;
        state->buttonUps[i] = mouseClickers[i].up;
        mouseClickers[i].down = mouseClickers[i].up = 0;
    }
}
static void Mouse_Qt_ShowCursor(bool yes) {
    LOG_INPUT_VERBOSE("%s cursor (presently visible? %b)")
            << (yes? "showing" : "hiding") << !cursorHidden;
    if (!yes && !cursorHidden) {
        cursorHidden = true;
#ifdef MACOSX
        Cursor_Show(false);
#else
        qApp->setOverrideCursor(QCursor(Qt::BlankCursor));
#endif
    }
    else if (yes && cursorHidden) {
        cursorHidden = false;
#ifdef MACOSX
        Cursor_Show(true);
#else
        qApp->restoreOverrideCursor();
#endif
    }
}
static void Mouse_Qt_InitTrap() {
#if !defined (DENG_MOBILE)
    auto &window = ClientWindowSystem::main();
    QCursor::setPos(window.mapToGlobal(window.geometry().center()));
    window.setMouseGrabEnabled(true);
    window.setKeyboardGrabEnabled(true);
    Mouse_Qt_ShowCursor(false);
#endif
}
static void Mouse_Qt_DeinitTrap() {
#if !defined (DENG_MOBILE)
    auto &window = ClientWindowSystem::main();
    window.setMouseGrabEnabled(false);
    window.setKeyboardGrabEnabled(false);
    Mouse_Qt_ShowCursor(true);
#endif
}
static void Mouse_Qt_Trap(dd_bool enabled) {
    if (mouseTrapped == CPP_BOOL(enabled)) return;
    mouseTrapped = enabled;
    prevMousePos = QPoint();
    if (enabled) {
        Mouse_Qt_InitTrap();
    }
    else {
        Mouse_Qt_DeinitTrap();
    }
}
void Mouse_Qt_SubmitButton(int button, dd_bool isDown) {
    if (button < 0 || button >= IMB_MAXBUTTONS) return; // Ignore...
    if (isDown)
        mouseClickers[button].down++;
    else
        mouseClickers[button].up++;
}
void Mouse_Qt_SubmitMotion(int axis, int deltaX, int deltaY) {
    if (axis < 0 || axis >= IMA_MAXAXES) return; // Ignore...
    if (axis == IMA_WHEEL) {
        int idx = ( deltaX < 0? IMB_MWHEELLEFT
                  : deltaX > 0? IMB_MWHEELRIGHT
                  : deltaY < 0? IMB_MWHEELUP
                              : IMB_MWHEELDOWN);
        Mouse_Qt_SubmitButton(idx, true);
        Mouse_Qt_SubmitButton(idx, false);
    }
    else {
        mouseDelta[axis].dx += deltaX;
        mouseDelta[axis].dy += deltaY;
    }
}
void Mouse_Qt_SubmitWindowPosition(int x, int y) {
    mouseDelta[IMA_POINTER].dx = x;
    mouseDelta[IMA_POINTER].dy = y;
}
mouseinterface_t qtMouse = {
    Mouse_Qt_Init,
    Mouse_Qt_Shutdown,
    Mouse_Qt_Poll,
    Mouse_Qt_GetState,
    Mouse_Qt_Trap
};
DENG2_PIMPL(NativeMenu)
, DENG2_OBSERVES(Games, Readiness) {
    QMenuBar *menuBar = nullptr;
    QMenu *helpMenu = nullptr;
    QMenu *gameMenu = nullptr;
    LoopCallback mainCall;
    Impl(Public *i) : Base(i) {
#ifdef MACOSX
        menuBar = new QMenuBar;
        gameMenu = menuBar->addMenu(QObject::tr("&Game"));
        helpMenu = menuBar->addMenu(QObject::tr("&Help"));
        helpMenu->addAction(tr("About Doomsday"),
                            &ClientWindow::main().taskBar(),
                            SLOT(showAbout()));
        QAction *checkForUpdates = helpMenu->addAction(tr("Check For &Updates..."),
                                                       &ClientApp::updater(),
                                                       SLOT(checkNowShowingProgress()));
        checkForUpdates->setMenuRole(QAction::ApplicationSpecificRole);
        updateGameMenuItems();
        DoomsdayApp::games().audienceForReadiness() += this;
#endif
    }
    void gameReadinessUpdated() override {
#ifdef MACOSX
        mainCall.enqueue([this] () { updateGameMenuItems(); });
#endif
    }
    void updateGameMenuItems() {
#ifdef MACOSX
        if (!gameMenu) return;
        gameMenu->clear();
        QList<Game *> allGames = DoomsdayApp::games().all();
        qSort(allGames.begin(), allGames.end(), [] (Game const *a, Game const *b) {
            return a->id().compare(b->id()) < 0;
        });
        foreach (Game *game, allGames) {
            QAction *load = new QAction(tr("Load %1").arg(game->title()), thisPublic);
            load->setData(game->id());
            load->setEnabled(game->isPlayable());
            connect(load, &QAction::triggered, [load] () {
                auto &win = ClientWindow::main();
                win.glActivate();
                Con_Executef(CMDS_DDAY, false, "load %s",
                             load->data().toString().toUtf8().constData());
                win.glDone();
            });
            gameMenu->addAction(load);
        }
#endif
    }
};
NativeMenu::NativeMenu()
    : d(new Impl(this))
{}
void Sys_MessageBox(messageboxtype_t type, const char* title, const char* msg, const char* detailedMsg) {
    Sys_MessageBox2(type, title, msg, 0, detailedMsg);
}
void Sys_MessageBox2(messageboxtype_t type, const char* title, const char* msg, const char* informativeMsg, const char* detailedMsg) {
    Sys_MessageBox3(type, title, msg, informativeMsg, detailedMsg, 0);
}
int Sys_MessageBox3(messageboxtype_t type, const char* title, const char* msg, const char* informativeMsg, const char* detailedMsg,
                    const char** buttons) {
    if (novideo) {
        qWarning("%s", msg);
        return 0;
    }
    if (ClientWindow::mainExists()) {
        ClientWindow::main().hide();
    }
    QMessageBox box;
    box.setWindowTitle(title);
    box.setText(msg);
    switch (type) {
    case MBT_INFORMATION: box.setIcon(QMessageBox::Information); break;
    case MBT_QUESTION:    box.setIcon(QMessageBox::Question);    break;
    case MBT_WARNING:     box.setIcon(QMessageBox::Warning);     break;
    case MBT_ERROR:       box.setIcon(QMessageBox::Critical);    break;
    default:
        break;
    }
    if (detailedMsg) {
        box.setDetailedText(detailedMsg);
    }
    if (informativeMsg) {
        box.setInformativeText(informativeMsg);
    }
    if (buttons) {
        for (int i = 0; buttons[i]; ++i) {
            box.addButton(buttons[i],
                          i == 0? QMessageBox::AcceptRole :
                          i == 1? QMessageBox::RejectRole :
                                  QMessageBox::ActionRole);
        }
    }
    return box.exec();
}
void Sys_MessageBoxf(messageboxtype_t type, const char* title, const char* format, ...) {
    char buffer[1024];
    va_list args;
    va_start(args, format);
    dd_vsnprintf(buffer, sizeof(buffer), format, args);
    va_end(args);
    Sys_MessageBox(type, title, buffer, 0);
}
int Sys_MessageBoxWithButtons(messageboxtype_t type, const char* title, const char* msg,
                              const char* informativeMsg, const char** buttons) {
    return Sys_MessageBox3(type, title, msg, informativeMsg, 0, buttons);
}
void Sys_MessageBoxWithDetailsFromFile(messageboxtype_t type, const char* title, const char* msg,
                                       const char* informativeMsg, const char* detailsFileName) {
    try {
        de::Block details;
        de::ByteArrayFile const &file = de::App::rootFolder().locate<de::ByteArrayFile>(detailsFileName);
        file >> details;
        details.append('\0');
        Sys_MessageBox2(type, title, msg, informativeMsg, details.constData());
    }
    catch (de::Error const &er) {
        qWarning() << "Could not read" << detailsFileName << ":" << er.asText().toLatin1().constData();
        Sys_MessageBox2(type, title, msg, informativeMsg, 0);
    }
}
D_CMD(PostFx);
DENG2_PIMPL(PostProcessing) {
    enum { PassThroughProgram = 0, ActiveProgram = 1 };
    Drawable frame;
    GLUniform uMvpMatrix { "uMvpMatrix", GLUniform::Mat4 };
    GLUniform uFrame     { "uTex",       GLUniform::Sampler2D };
    GLUniform uFadeInOut { "uFadeInOut", GLUniform::Float };
    Animation fade       { 0, Animation::Linear };
    float opacity        { 1.f };
    struct QueueEntry {
        String shaderName;
        float fade;
        TimeSpan span;
        QueueEntry(String const &name, float f, TimeSpan const &s)
            : shaderName(name), fade(f), span(s) {}
    };
    typedef QList<QueueEntry> Queue;
    Queue queue;
    typedef GLBufferT<Vertex2Tex> VBuf;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        DENG2_ASSERT(!frame.isReady()); // deinited earlier
    }
    void attachUniforms(GLProgram &program) {
        program << uMvpMatrix << uFrame << uFadeInOut;
    }
    bool setShader(String const &name) {
        try {
            auto &shaders = ClientApp::shaders();
            GLProgram &prog = frame.addProgram(ActiveProgram);
            shaders.build(prog, "fx.post." + name);
            attachUniforms(prog);
            frame.setProgram(prog);
            LOG_GL_MSG("Post-processing shader \"fx.post.%s\"") << name;
            return true;
        }
        catch (Error const &er) {
            LOG_GL_WARNING("Failed to set shader to \"fx.post.%s\":\n%s")
                    << name << er.asText();
        }
        return false;
    }
    bool isActive() const {
        return !fade.done() || fade > 0 || !queue.isEmpty();
    }
    void glInit() {
        if (frame.isReady()) return;
        VBuf *buf = new VBuf;
        buf->setVertices(gl::TriangleStrip,
                         VBuf::Builder().makeQuad(Rectanglef(0, 0, 1, 1),
                                                  Rectanglef(0, 1, 1, -1)),
                         gl::Static);
        frame.addBuffer(buf);
        ClientApp::shaders().build(frame.program(), "generic.texture");
        attachUniforms(frame.program());
    }
    void glDeinit() {
        if (!frame.isReady()) return;
        LOGDEV_GL_XVERBOSE("Releasing GL resources", "");
        frame.clear();
    }
    void checkQueue() {
        if (!fade.done()) return; // Let's check back later.
        if (!queue.isEmpty()) {
            QueueEntry entry = queue.takeFirst();
            if (!entry.shaderName.isEmpty()) {
                if (!setShader(entry.shaderName)) {
                    fade = 0;
                    return;
                }
            }
            fade.setValue(entry.fade, entry.span);
            LOGDEV_GL_VERBOSE("Shader '%s' fade:%s") << entry.shaderName << fade.asText();
        }
    }
    void draw() {
        if (isActive()) {
            uFadeInOut = fade * opacity;
        }
        frame.draw();
    }
};
PostProcessing::PostProcessing()
    : d(new Impl(this))
{}
bool PostProcessing::isActive() const {
    return d->isActive();
}
void PostProcessing::fadeInShader(String const &fxPostShader, TimeSpan const &span) {
    d->queue.append(Impl::QueueEntry(fxPostShader, 1, span));
}
void PostProcessing::fadeOut(TimeSpan const &span) {
    d->queue.append(Impl::QueueEntry("", 0, span));
}
void PostProcessing::setOpacity(float opacity) {
    d->opacity = opacity;
}
void PostProcessing::glInit() {
    LOG_AS("PostProcessing");
    d->glInit();
}
void PostProcessing::glDeinit() {
    LOG_AS("PostProcessing");
    d->glDeinit();
}
void PostProcessing::update() {
    LOG_AS("PostProcessing");
    if (d->isActive()) {
        d->checkQueue();
    }
    else {
        d->frame.setProgram(Impl::PassThroughProgram);
    }
}
void PostProcessing::draw(Matrix4f const &mvpMatrix, GLTexture const &frame) {
    d->uMvpMatrix = mvpMatrix;
    d->uFrame     = frame;
    d->draw();
}
void PostProcessing::consoleRegister() // static
{
    C_CMD("postfx", "is",  PostFx);
    C_CMD("postfx", "isf", PostFx);
}
D_CMD(PostFx) {
    DENG2_UNUSED(src);
    int console = String(argv[1]).toInt();
    String const shader = argv[2];
    TimeSpan const span = (argc == 4? String(argv[3]).toDouble() : 0);
    if (console < 0 || console >= DDMAXPLAYERS) {
        LOG_SCR_WARNING("Invalid console %i") << console;
        return false;
    }
    PostProcessing &post = ClientApp::player(console).viewCompositor().postProcessing();
    if (shader == "none") {
        post.fadeOut(span);
        return true;
    }
    else if (shader == "opacity") // Change opacity.
    {
        post.setOpacity(float(span));
        return true;
    }
    post.fadeInShader(shader, span);
    return true;
}
static de::TimeSpan const PROGRESS_DELTA_TIME = 0.5;
static ProgressWidget &progress() {
    return ClientWindow::main().busy().progress();
}
void Con_InitProgress2(int maxProgress, float start, float end) {
    progress().setRange(Rangei(0, maxProgress), Rangef(start, end));
    progress().setProgress(0, 0);
}
void Con_InitProgress(int maxProgress) {
    Con_InitProgress2(maxProgress, 0.f, 1.f);
}
dd_bool Con_IsProgressAnimationCompleted(void) {
    return progress().isHidden() || !progress().isAnimating();
}
void Con_SetProgress(int progressValue) {
     progress().setProgress(progressValue, progressValue < progress().range().end?
                                PROGRESS_DELTA_TIME : TimeSpan(PROGRESS_DELTA_TIME / 2));
}
DENG2_PIMPL(SaveListData)
, DENG2_OBSERVES(FileIndex, Addition)
, DENG2_OBSERVES(FileIndex, Removal) {
    LoopCallback mainCall;
    Impl(Public *i) : Base(i) {
        SaveGames::get().saveIndex().audienceForAddition() += this;
        SaveGames::get().saveIndex().audienceForRemoval()  += this;
    }
    bool shouldAddFolder(GameStateFolder const &save) const {
        return save.path().beginsWith("/home/savegames"); // Ignore non-user savegames.
    }
    void fileAdded(File const &file, FileIndex const &) {
        GameStateFolder const &saveFolder = file.as<GameStateFolder>();
        if (shouldAddFolder(saveFolder)) {
            mainCall.enqueue([this, &saveFolder] () {
                self().append(new SaveItem(saveFolder));
            });
        }
    }
    void fileRemoved(File const &, FileIndex const &) {
        mainCall.enqueue([this] () {
            for (ui::Data::Pos idx = self().size() - 1; idx < self().size(); --idx) {
                if (!self().at(idx).isValid()) {
                    self().remove(idx);
                }
            }
        });
    }
    void addAllFromIndex() {
        foreach (File *file, SaveGames::get().saveIndex().files()) {
            try {
                GameStateFolder &save = file->as<GameStateFolder>();
                if (shouldAddFolder(save)) {
                    self().append(new SaveItem(save));
                }
            }
            catch (Error const &er) {
                LOG_ERROR("Save file %s has corrupt metadata: %s")
                        << file->description()
                        << er.asText();
            }
        }
    }
};
SaveListData::SaveListData()
    : d(new Impl(this)) {
    d->addAllFromIndex();
}
SaveListData::SaveItem::SaveItem(GameStateFolder const &saveFolder)
    : ImageItem(ShownAsButton)
    , saveFolder(&saveFolder) {
    saveFolder.audienceForDeletion() += this;
    setData(savePath()); // for looking it up later
    setLabel(title());
    Games &games = DoomsdayApp::games();
    if (games.contains(gameId())) {
        setImage(Style::get().images().image(games[gameId()].logoImageId()));
    }
}
SaveListData::SaveItem::~SaveItem() {
    if (saveFolder) saveFolder->audienceForDeletion() -= this;
}
bool SaveListData::SaveItem::isValid() const {
    return saveFolder != nullptr;
}
String SaveListData::SaveItem::title() const {
    if (saveFolder) {
        return saveFolder->metadata().gets("userDescription");
    }
    return "";
}
String SaveListData::SaveItem::gameId() const {
    if (saveFolder) {
        return saveFolder->metadata().gets("gameIdentityKey");
    }
    return "";
}
String SaveListData::SaveItem::savePath() const {
    if (saveFolder) {
        return saveFolder->path();
    }
    return "";
}
String SaveListData::SaveItem::name() const {
    if (saveFolder) {
        return saveFolder->name().fileNameWithoutExtension();
    }
    return "";
}
StringList SaveListData::SaveItem::loadedPackages() const {
    if (saveFolder) {
        Record const &meta = saveFolder->metadata();
        if (meta.has("packages")) {
            return meta.getStringList("packages");
        }
    }
    return StringList();
}
void SaveListData::SaveItem::fileBeingDeleted(File const &) {
    saveFolder = nullptr;
}
SaveListData::SaveItem &SaveListData::at(Pos pos) {
    return ListData::at(pos).as<SaveItem>();
}
SaveListData::SaveItem const &SaveListData::at(Pos pos) const {
    return ListData::at(pos).as<SaveItem>();
}
static char const *VAR_METADATA = "log.showMetadata";
DENG2_PIMPL(StyledLogSinkFormatter)
, DENG2_OBSERVES(Variable, Change) {
    LogEntry::Flags format;
    bool observe;
    bool omitSectionIfNonDev;
    bool showMetadata;
    Impl(Public *i, bool observeVars)
        : Base(i)
        , observe(observeVars)
        , omitSectionIfNonDev(true)
        , showMetadata(false) {
        if (observe) {
            showMetadata = App::config().getb(VAR_METADATA);
            Config::get(VAR_METADATA).audienceForChange() += this;
        }
    }
    void variableValueChanged(Variable &, Value const &newValue) {
        showMetadata = newValue.isTrue();
    }
};
StyledLogSinkFormatter::StyledLogSinkFormatter()
    : d(new Impl(this, true /*observe*/))
{
    d->format = LogEntry::Styled | LogEntry::OmitLevel;
}
StyledLogSinkFormatter::StyledLogSinkFormatter(LogEntry::Flags const &formatFlags)
    : d(new Impl(this, false /*don't observe*/))
{
    d->format = formatFlags;
}
LogSink::IFormatter::Lines StyledLogSinkFormatter::logEntryToTextLines(LogEntry const &entry) {
    LogEntry::Flags form = d->format;
    if (!d->showMetadata) {
        form |= LogEntry::Simple | LogEntry::OmitDomain;
    }
    if (d->omitSectionIfNonDev && !(entry.context() & LogEntry::Dev)) {
        form |= LogEntry::OmitSection;
    }
    return Lines() << entry.asText(form);
}
void StyledLogSinkFormatter::setOmitSectionIfNonDev(bool omit) {
    d->omitSectionIfNonDev = omit;
}
static dd_bool initOk;
static byte useMouse;  ///< Input enabled from mouse?
static mouseinterface_t *iMouse; ///< Current mouse interface.
static keyevent_t keyEvents[EVBUFSIZE];
static int evHead, evTail;
void I_Register(void) {
#ifdef __CLIENT__
    Joystick_Register();
#endif
}
static keyevent_t *newKeyEvent(void) {
    keyevent_t *ev = keyEvents + evHead;
    evHead = (evHead + 1) % EVBUFSIZE;
    memset(ev, 0, sizeof(*ev));
    return ev;
}
/**
 * @return The oldest event from the buffer.
 */
static keyevent_t *getKeyEvent(void) {
    keyevent_t *ev;
    if (evHead == evTail)
        return NULL;            // No more...
    ev = keyEvents + evTail;
    evTail = (evTail + 1) % EVBUFSIZE;
    return ev;
}
static void Mouse_Init(void) {
    if (CommandLine_Check("-nomouse") || novideo)
        return;
    LOG_AS("Mouse_Init");
    DENG_ASSERT(iMouse);
    iMouse->init();
    useMouse = true;
}
dd_bool I_InitInterfaces(void) {
    if (initOk)
        return true; // Already initialized.
#ifdef __CLIENT__
    iMouse = &qtMouse;
#ifdef WIN32
    iMouse = &win32Mouse;
    DirectInput_Init();
#endif
    Mouse_Init();
    Joystick_Init();
#endif // __CLIENT__
    initOk = true;
    return true;
}
void I_ShutdownInterfaces() {
    if (!initOk)
        return; // Not initialized.
#ifdef __CLIENT__
    if (useMouse) iMouse->shutdown();
    useMouse = false;
    Joystick_Shutdown();
# ifdef WIN32
    DirectInput_Shutdown();
# endif
#endif
    initOk = false;
}
void Keyboard_Submit(int type, int ddKey, int native, const char* text) {
    if (ddKey != 0) {
        keyevent_t* e = newKeyEvent();
        e->type = type;
        e->ddkey = ddKey;
        e->native = native;
        if (text) {
            strncpy(e->text, text, sizeof(e->text) - 1);
        }
    }
}
size_t Keyboard_GetEvents(keyevent_t *evbuf, size_t bufsize) {
    keyevent_t *e;
    size_t      i = 0;
    if (!initOk)
        return 0;
    for (i = 0; i < bufsize; ++i) {
        e = getKeyEvent();
        if (!e) break; // No more events.
        memcpy(&evbuf[i], e, sizeof(*e));
    }
    return i;
}
dd_bool Mouse_IsPresent(void) {
    return useMouse;
}
void Mouse_Poll(void) {
    if (useMouse) {
        iMouse->poll();
    }
}
void Mouse_GetState(mousestate_t *state) {
    if (useMouse) {
        iMouse->getState(state);
    }
}
void Mouse_Trap(dd_bool enabled) {
    if (useMouse) {
        iMouse->trap(enabled);
    }
}
fontid_t fontFixed;
static ui_color_t ui_colors[NUM_UI_COLORS] = {
    /* UIC_TEXT */      { .85f, .87f, 1 },
    /* UIC_TITLE */     { 1, 1, 1 },
};
char const *UI_ChooseFixedFont() {
    if (DENG_GAMEVIEW_WIDTH < 300) return "console11";
    if (DENG_GAMEVIEW_WIDTH > 768) return "console18";
    return "console14";
}
static AbstractFont *loadSystemFont(char const *name) {
    DENG2_ASSERT(name != 0 && name[0]);
    de::Uri uri = de::makeUri("System:").setPath(name);
    ddstring_t resourcePath; Str_InitStd(&resourcePath);
    Str_Appendf(&resourcePath, "}data/Fonts/%s.dfn", name);
#if defined(UNIX) && !defined(MACOSX)
    strlwr(resourcePath.str);
#endif
    F_ExpandBasePath(&resourcePath, &resourcePath);
    AbstractFont *font = ClientResources::get().newFontFromFile(uri, Str_Text(&resourcePath));
    if (!font) {
        App_Error("loadSystemFont: Failed loading font \"%s\".", name);
    }
    Str_Free(&resourcePath);
    return font;
}
static void loadFontIfNeeded(char const *uri, fontid_t *fid) {
    *fid = NOFONTID;
    if (uri && uri[0]) {
        try {
            FontManifest &manifest = ClientResources::get().fontManifest(de::makeUri(uri));
            if (manifest.hasResource()) {
                *fid = fontid_t(manifest.uniqueId());
            }
        }
        catch (Resources::MissingResourceManifestError const &)
        {}
    }
    if (*fid == NOFONTID) {
        *fid = loadSystemFont(uri)->manifest().uniqueId();
    }
}
void UI_LoadFonts() {
    if (novideo) return;
    loadFontIfNeeded(UI_ChooseFixedFont(), &fontFixed);
}
ui_color_t* UI_Color(uint id) {
    if (id >= NUM_UI_COLORS)
        return NULL;
    return &ui_colors[id];
}
void UI_MixColors(ui_color_t* a, ui_color_t* b, ui_color_t* dest, float amount) {
    dest->red   = (1 - amount) * a->red   + amount * b->red;
    dest->green = (1 - amount) * a->green + amount * b->green;
    dest->blue  = (1 - amount) * a->blue  + amount * b->blue;
}
void UI_SetColorA(ui_color_t* color, float alpha) {
    DGL_Color4f(color->red, color->green, color->blue, alpha);
}
void UI_SetColor(ui_color_t* color) {
    DGL_Color3f(color->red, color->green, color->blue);
}
void UI_TextOutEx2(const char* text, const Point2Raw* origin, ui_color_t* color, float alpha,
    int alignFlags, short textFlags) {
    assert(origin);
    if (alpha <= 0) return;
    FR_SetColorAndAlpha(color->red, color->green, color->blue, alpha);
    FR_DrawText3(text, origin, alignFlags, textFlags);
}
void UI_TextOutEx(const char* text, const Point2Raw* origin, ui_color_t* color, float alpha) {
    UI_TextOutEx2(text, origin, color, alpha, DEFAULT_ALIGNFLAGS, DEFAULT_DRAWFLAGS);
}
void UI_DrawDDBackground(Point2Raw const &origin, Size2Raw const &dimensions, float alpha) {
    DGL_PushState();
    if (alpha < 1.0f) {
        GL_BlendMode(BM_NORMAL);
    }
    else {
        DGL_Disable(DGL_BLEND);
    }
    DGL_Color4f(0, 0, 0, alpha);
    GL_DrawRect2(origin.x, origin.y, dimensions.width, dimensions.height);
    DGL_PopState();
}
static Ranged const FACTOR_RANGE(1.0 / 16.0, 1.0);
DENG2_PIMPL(ViewCompositor)
, DENG2_OBSERVES(Variable, Change) {
    mutable Variable const *pixelDensity = nullptr;
    mutable Variable const *resizeFactor = nullptr;
    int playerNum = 0;
    GLTextureFramebuffer viewFramebuf;
    PostProcessing postProcessing;
    Impl(Public *i)
        : Base(i)
        , viewFramebuf(Image::RGBA_8888)
    {}
    void getConfig() const {
        if (!pixelDensity) {
            pixelDensity = &App::config("render.pixelDensity");
            resizeFactor = &App::config("render.fx.resize.factor");
        }
    }
    double scaleFactor() const {
        getConfig();
        double const rf = (*resizeFactor > 0? 1.0 / *resizeFactor : 1.0);
        return FACTOR_RANGE.clamp(*pixelDensity * rf);
    }
    GLFramebuffer::Size framebufferSize() const {
        GLFramebuffer::Size size = R_Console3DViewRect(playerNum).size();
        switch (ClientApp::vr().mode()) {
        case VRConfig::SideBySide:
            size.x /= 2;
            break;
        case VRConfig::TopBottom:
            size.y /= 2;
            break;
        default:
            break;
        }
        size *= scaleFactor();
        return size;
    }
    void updateSampleCount() {
        int sampleCount = 1;
        bool configured = App::config().getb("window.main.fsaa");
        if (App::commandLine().has("-nofsaa") || !configured) {
            LOG_GL_VERBOSE("Multisampling off");
        }
        else {
            sampleCount = 4; // four samples is fine?
            LOG_GL_VERBOSE("Multisampling on (%i samples)") << sampleCount;
        }
        viewFramebuf.setSampleCount(sampleCount);
    }
    void variableValueChanged(Variable &, Value const &) override {
        updateSampleCount();
    }
    void glInit() {
        viewFramebuf.resize(framebufferSize());
        if (!viewFramebuf.areTexturesReady()) {
            App::config("window.main.fsaa").audienceForChange() += this;
            updateSampleCount();
            viewFramebuf.glInit();
        }
        postProcessing.glInit();
        /*if (!frameDrawable.isReady())
        {
            ClientApp::shaders().build(frameDrawable.program(), "generic.texture")
                    << uMvpMatrix
                    << uFrameTex;
            using VBuf = GuiWidget::DefaultVertexBuf;
            auto *vbuf = new VBuf;
            frameDrawable.addBuffer(vbuf);
            VBuf::Builder verts;
            verts.makeQuad(Rectanglef(0, 0, 1, 1), Rectanglef(0, 1, 1, -1));
            vbuf->setVertices(gl::TriangleStrip, verts, gl::Static);
        }*/
    }
    void glDeinit() {
        viewFramebuf.glDeinit();
        postProcessing.glDeinit();
    }
};
ViewCompositor::ViewCompositor()
    : d(new Impl(this))
{}
void ViewCompositor::setPlayerNumber(int playerNum) {
    d->playerNum = playerNum;
}
void ViewCompositor::glDeinit() {
    d->glDeinit();
}
void ViewCompositor::renderGameView(std::function<void (int)> renderFunc) {
    d->glInit();
    GLState::push()
            .setTarget(d->viewFramebuf)
            .setViewport(Rectangleui::fromSize(d->viewFramebuf.size()));
    d->viewFramebuf.clear(GLFramebuffer::ColorDepthStencil);
    renderFunc(d->playerNum);
    GLState::pop();
}
GLTextureFramebuffer &ViewCompositor::gameView() {
    return d->viewFramebuf;
}
GLTextureFramebuffer const &ViewCompositor::gameView() const {
    return d->viewFramebuf;
}
void ViewCompositor::drawCompositedLayers() {
    Rectanglei const view3D = R_Console3DViewRect(d->playerNum);
    auto const oldDisplayPlayer = displayPlayer;
    displayPlayer = d->playerNum;
    R_UseViewPort(d->playerNum);
    DGL_PushState();
    DGL_Disable(DGL_ALPHA_TEST);
    DGL_Disable(DGL_BLEND);
    DGL_Disable(DGL_DEPTH_TEST);
    DGL_CullFace(DGL_NONE);
    d->postProcessing.update();
    d->postProcessing.draw(ClientWindow::main().root().projMatrix2D() *
                           Matrix4f::scaleThenTranslate(view3D.size(), view3D.topLeft),
                           d->viewFramebuf.colorTexture());
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PushMatrix();
    DGL_LoadMatrix(ClientWindow::main().root().projMatrix2D().values());
    if (auto const *vp = R_CurrentViewPort()) {
        R_RenderPlayerViewBorder();
        RectRaw vpGeometry = {{vp->geometry.topLeft.x, vp->geometry.topLeft.y},
                              {int(vp->geometry.width()), int(vp->geometry.height())}};
        viewdata_t const *vd = &DD_Player(d->playerNum)->viewport();
        RectRaw vdWindow = {{vd->window.topLeft.x, vd->window.topLeft.y},
                            {int(vd->window.width()), int(vd->window.height())}};
        if (gx.DrawViewPort) {
            DGL_Enable(DGL_BLEND);
            gx.DrawViewPort(P_ConsoleToLocal(d->playerNum),
                            &vpGeometry,
                            &vdWindow,
                            displayPlayer,
                            /* layer: */ 1);
        }
    }
    DGL_MatrixMode(DGL_PROJECTION);
    DGL_PopMatrix();
    Rectanglef const normRect = GuiWidget::normalizedRect(
                R_ConsoleRect(d->playerNum),
                Rectanglei::fromSize(ClientWindow::main().root().viewSize()));
    auto const targetSize = GLState::current().target().size();
    Rectanglef const vp { normRect.topLeft     * targetSize,
                          normRect.bottomRight * targetSize };
    DGL_PushState();
    GLState::current().setViewport(vp.toRectangleui()); {
        if (App_InFineSystem().finaleInProgess()) {
            dgl_borderedprojectionstate_t bp;
            GL_ConfigureBorderedProjection(&bp, BPF_OVERDRAW_CLIP |
                                           (!App_World().hasMap()? BPF_OVERDRAW_MASK : 0),
                                           SCREENWIDTH, SCREENHEIGHT,
                                           DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT,
                                           scalemode_t(Con_GetByte("rend-finale-stretch")));
            GL_BeginBorderedProjection(&bp);
            for (Finale *finale : App_InFineSystem().finales()) {
                finale->interpreter().page(FinaleInterpreter::Anims).draw();
                finale->interpreter().page(FinaleInterpreter::Texts).draw();
            }
            GL_EndBorderedProjection(&bp);
        }
    } {
        if (gx.DrawWindow) {
            Size2Raw const dimensions = {{{DENG_GAMEVIEW_WIDTH, DENG_GAMEVIEW_HEIGHT}}};
            gx.DrawWindow(&dimensions);
        }
    }
    DGL_PopState(); {
#if 0
        SBE_DrawGui();
        if (App_World().hasMap() && App_World().map().hasLightGrid()) {
            Rend_LightGridVisual(App_World().map().lightGrid());
        }
#endif
        Net_Drawer();
        Sfx_ChannelDrawer();
    }
    R_UseViewPort(nullptr);
    displayPlayer = oldDisplayPlayer;
    GLState::considerNativeStateUndefined();
    DGL_PopState();
}
PostProcessing &ViewCompositor::postProcessing() {
    return d->postProcessing;
}
