﻿/** @file bindings_world.cpp  World related Doomsday Script bindings.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "world/bindings_world.h"
#include "world/clientserverworld.h"
#include "world/map.h"
#include "world/thinkers.h"
#include "world/p_players.h"
#include "audio/audiosystem.h"
#include "dd_main.h"
#include "def_main.h"
#include <doomsday/defs/ded.h>
#include <doomsday/world/mobj.h>
#include <doomsday/world/mobjthinkerdata.h>
#include <de/Context>
#include <de/RecordValue>
/** @file blockmap.cpp  Map element/object blockmap.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/blockmap.h"
#ifdef __CLIENT__
#  include <de/GLInfo>
#  include "api_gl.h"
#  include "gl/gl_main.h"
#endif
#include <de/Vector>
#include <de/memoryzone.h>
#include <de/vector1.h>
#include <cmath>
/** @file bspleaf.cpp  World map BSP leaf half-space.
 *
 * @authors Copyright © 2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/bspleaf.h"
#include "ConvexSubspace"
#include "Sector"
/** @file worldsystem.cpp  World subsystem.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/clientserverworld.h"
#include <map>
#include <utility>
#include <QMap>
#include <QtAlgorithms>
#include <de/memoryzone.h>
#include <de/timer.h>
#include <de/Binder>
#include <de/Context>
#include <de/Error>
#include <de/Log>
#include <de/Scheduler>
#include <de/Script>
#include <de/ScriptSystem>
#include <de/Process>
#include <de/Time>
#include <doomsday/doomsdayapp.h>
#include <doomsday/console/cmd.h>
#include <doomsday/console/exec.h>
#include <doomsday/console/var.h>
#include <doomsday/defs/mapinfo.h>
#include <doomsday/resource/mapmanifests.h>
#include <doomsday/world/MaterialManifest>
#include <doomsday/world/Materials>
#include "dd_main.h"
#include "dd_def.h"
#include "dd_loop.h"
#include "def_main.h"  // ::defs
#include "api_player.h"
#ifdef __CLIENT__
#  include "clientapp.h"
#  include "client/cl_def.h"
#  include "client/cl_frame.h"
#  include "client/cl_player.h"
#  include "gl/gl_main.h"
#endif
#ifdef __SERVER__
#  include "server/sv_pool.h"
#endif
#include "network/net_main.h"
#include "api_mapedit.h"
#include "world/p_players.h"
#include "world/p_ticker.h"
#include "world/sky.h"
#include "world/thinkers.h"
#include "world/bindings_world.h"
#include "edit_map.h"
#include "Plane"
#include "Sector"
#include "Subsector"
#include "Surface"
#ifdef __CLIENT__
#  include "world/contact.h"
#  include "client/cledgeloop.h"
#  include "client/clientsubsector.h"
#  include "Lumobj"
#  include "render/viewports.h"  // R_ResetViewer
#  include "render/rend_fakeradio.h"
#  include "render/rend_main.h"
#  include "render/rendersystem.h"
#  include "render/rendpoly.h"
#  include "MaterialAnimator"
#  include "ui/progress.h"
#  include "ui/inputsystem.h"
#endif
/** @file convexsubspace.cpp  Map convex subspace.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/convexsubspace.h"
#ifdef __CLIENT__
#  include "world/audioenvironment.h"
#  include "audio/s_environ.h"
#  include "ClientMaterial"
#endif
#include "BspLeaf"
#include "Face"
#include "Polyobj"
#include "Subsector"
#include "Surface"
#include <de/Log>
#include <QSet>
#include <QtAlgorithms>
/** @file impulseaccumulator.cpp  Player impulse accumulation.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "world/impulseaccumulator.h"
#include <de/timer.h>
#include <de/LogBuffer>
#include <doomsday/console/var.h>
#ifdef __CLIENT__
#  include "ui/inputsystem.h"
#  include "clientapp.h"
#endif
#include "world/p_players.h"
#ifdef __CLIENT__
#  include "BindContext"
#  include "ui/b_util.h"
#endif
/** @file interceptor.cpp  World map element/object ray trace interceptor.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/interceptor.h"
#include <de/memoryzone.h>
#include <de/vector1.h>
#include "world/blockmap.h"
#include "world/lineblockmap.h"
#include "world/p_object.h"
#include "world/clientserverworld.h" // validCount
/** @file line.cpp  World map line.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/line.h"
#include "world/map.h"
#include "world/maputil.h"  // R_FindLineNeighbor()...
#include "world/convexsubspace.h"
#include "world/sector.h"
#include "world/surface.h"
#include "world/vertex.h"
#ifdef __CLIENT__
#  include "world/lineowner.h"
#  include "render/r_main.h"  // levelFullBright
#  include "render/rend_fakeradio.h"
#endif
#include "Face"
#include "HEdge"
#include "dd_main.h"  // App_World()
#include <doomsday/console/cmd.h>
#include <de/vector1.h>
#include <QtAlgorithms>
#include <QList>
#include <QMap>
#include <array>
#ifdef WIN32
#  undef max
#  undef min
#endif
/** @file lineblockmap.cpp  Specialized Blockmap, for map Lines.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/lineblockmap.h"
/** @file linesighttest.cpp  World map line of sight testing.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "world/linesighttest.h"
#include <cmath>
#include <de/aabox.h>
#include <de/fixedpoint.h>
#include <de/vector1.h>
#include <doomsday/BspNode>
#include "Face"
#include "world/clientserverworld.h"  /// For validCount, @todo Remove me.
#include "BspLeaf"
#include "ConvexSubspace"
#include "Line"
#include "Polyobj"
#include "Sector"
/** @file map.cpp  World map.
 *
 * @todo This file has grown far too large. It should be split up through the
 * introduction of new abstractions / collections.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "world/map.h"
#ifdef __CLIENT__
#  include "clientapp.h"
#  include "client/cl_mobj.h"
#endif
#include "api_console.h"
#include "api_mapedit.h"
#ifdef __CLIENT__
#  include "api_sound.h"
#endif
#include "world/bsp/partitioner.h"
#include "world/clientserverworld.h"  // ddMapSetup, validCount
#include "world/blockmap.h"
#include "world/lineblockmap.h"
#include "world/lineowner.h"
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/polyobjdata.h"
#include "world/sky.h"
#include "world/thinkers.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Face"
#include "Line"
#include "Polyobj"
#include "Sector"
#include "Subsector"
#include "Surface"
#include "Vertex"
#ifdef __CLIENT__
#  include "client/clientsubsector.h"
#  include "client/clskyplane.h"
#endif
#ifdef __CLIENT__
#  include "Contact"
#  include "ContactSpreader"
#  include "LightDecoration"
#  include "Lumobj"
#  include "WallEdge"
#  include "render/viewports.h"
#  include "render/rend_main.h"
#  include "render/rend_particle.h"
#endif
#include <doomsday/defs/mapinfo.h>
#include <doomsday/defs/sky.h>
#include <doomsday/EntityDatabase>
#include <doomsday/BspNode>
#include <doomsday/world/Materials>
#include <de/LogBuffer>
#include <de/Rectangle>
#include <de/aabox.h>
#include <de/charsymbols.h>
#include <de/nodepile.h>
#include <de/vector1.h>
#include <de/timer.h>
#include <array>
#include <QBitArray>
#include <QMultiMap>
#include <QVarLengthArray>
/** @file maputil.cpp  World map utilities.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/maputil.h"
#include "Line"
#include "Plane"
#include "Sector"
#ifdef __CLIENT__
#  include "Surface"
#  include "world/lineowner.h"
#  include "MaterialVariantSpec"
#  include "render/rend_main.h" // Rend_MapSurfacematerialSpec
#  include "MaterialAnimator"
#  include "WallEdge"
#endif
/** @file p_mobj.cpp  World map objects.
 *
 * Various routines for moving mobjs, collision and Z checking.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "world/p_object.h"
#include <cmath>
#include <de/vector1.h>
#include <de/Error>
#include <de/LogBuffer>
#include <doomsday/console/cmd.h>
#include <doomsday/console/exec.h>
#include <doomsday/console/var.h>
#include <doomsday/defs/sprite.h>
#include <doomsday/res/Textures>
#include <doomsday/res/Sprites>
#include <doomsday/world/mobjthinkerdata.h>
#include <doomsday/world/Materials>
#include "def_main.h"
#include "api_sound.h"
#include "network/net_main.h"
#ifdef __CLIENT__
#  include "client/cl_mobj.h"
#  include "client/clientsubsector.h"
#  include "gl/gl_tex.h"
#  include "network/net_demo.h"
#  include "render/viewports.h"
#  include "render/rend_main.h"
#  include "render/rend_model.h"
#  include "render/rend_halo.h"
#  include "render/billboard.h"
#endif
#include "world/clientserverworld.h" // validCount
#include "world/p_object.h"
#include "world/p_players.h"
#include "world/thinkers.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Subsector"
#ifdef __CLIENT__
#  include "Generator"
#  include "Lumobj"
#endif
/** @file p_players.cpp  World player entities.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_PLAYER
#include "world/p_players.h"
#include "world/impulseaccumulator.h"
#include "world/map.h"
#include "world/p_object.h"
#include "Subsector"
#include "Surface"
#ifdef __CLIENT__
#  include "BindContext"
#  include "ui/b_util.h"
#  include "ui/inputdevice.h"
#  include "ui/inputsystem.h"
#  include "client/clientsubsector.h"
#  include "client/clskyplane.h"
#  include "clientapp.h"
#endif
#include <doomsday/console/cmd.h>
#include <doomsday/console/var.h>
#include <QList>
#include <QMap>
#include <QtAlgorithms>
/** @file p_ticker.cpp  Timed world events.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de_base.h"
#include "world/p_ticker.h"
#ifdef __CLIENT__
#  include "MaterialAnimator"
#  include <doomsday/world/Materials>
#endif
/** @file plane.h  World map plane.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/plane.h"
#include "world/map.h"
#include "world/thinkers.h"
#include "world/clientserverworld.h"  // ddMapSetup
#include "Surface"
#include "Sector"
#ifdef __CLIENT__
#  include "MaterialAnimator"
#  include "render/rend_main.h"
#endif
#include "dd_loop.h"  // frameTimePos
#include "dd_main.h"  // App_Resources()
#ifdef __CLIENT__
#  include <doomsday/world/materialmanifest.h>
#endif
#include <de/LogBuffer>
#include <array>
/** @file polyobj.h  World map polyobj.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "world/polyobj.h"
#include "world/polyobjdata.h"
#include "world/blockmap.h"
#include "world/map.h"
#include "world/p_object.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "world/clientserverworld.h"  // validCount
#ifdef __CLIENT__
#  include "client/clientsubsector.h"
#  include "render/rend_main.h"  // useBias
#endif
#include "dd_main.h"
#include <de/vector1.h>
#include <QSet>
#include <QVector>
/** @file polyobjdata.cpp  Private data for polyobj.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "world/polyobjdata.h"
#ifdef __CLIENT__
#  include "client/clpolymover.h"
#endif
#include "world/map.h"
#include "world/thinkers.h"
/** @file reject.h World map sector LOS reject LUT building.
 *
 * @authors Copyright © 2007-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2000-2007 Andrew Apted <ajapted@gmail.com>
 * @authors Copyright © 1998-2000 Colin Reed <cph@moria.org.uk>
 * @authors Copyright © 1998-2000 Lee Killough <killough@rsn.hp.com>
 *
 * Based on glBSP 2.24 (in turn, based on BSP 2.3).
 * @see http://sourceforge.net/projects/glbsp/
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#if 0 // Needs updating.
#include <cmath>
#include "de_base.h"
#include "world/map.h"
using namespace de;
void BuildRejectForMap(Map const &map)
{
    int *secGroups = M_Malloc(sizeof(int) * map.sectorCount());
    int group = 0;
    for(uint i = 0; i < map.sectorCount(); ++i)
    {
        Sector &sector = map.sectors[i];
        secGroups[i] = group++;
        sector.rejNext = sector.rejPrev = &sector;
    }
    for(uint i = 0; i < map.lineCount(); ++i)
    {
        Line *line = &map.lines[i];
        if(!line->front().hasSector() || !line->back().hasSector())
            continue;
        Sector *sec1 = line->front().sectorPtr();
        Sector *sec2 = line->back().sectorPtr();
        if(!sec1 || !sec2 || sec1 == sec2)
            continue;
        if(secGroups[sec1->index] == secGroups[sec2->index])
            continue;
        Sector *p = 0;
        if(secGroups[sec1->index] > secGroups[sec2->index])
        {
            p = sec1;
            sec1 = sec2;
            sec2 = p;
        }
        secGroups[sec2->index] = secGroups[sec1->index];
        for(p = sec2->rejNext; p != sec2; p = p->rejNext)
        {
            secGroups[p->index] = secGroups[sec1->index];
        }
        sec1->rejNext->rejPrev = sec2;
        sec2->rejNext->rejPrev = sec1;
        p = sec1->rejNext;
        sec1->rejNext = sec2->rejNext;
        sec2->rejNext = p;
    }
    size_t rejectSize = (map.sectorCount() * map.sectorCount() + 7) / 8;
    byte *matrix = Z_Calloc(rejectSize, PU_MAPSTATIC, 0);
    for(int view = 0; view < map.sectorCount(); ++view)
    for(int target = 0; target < view; ++target)
    {
        if(secGroups[view] == secGroups[target])
            continue;
        int p1 = view * map.sectorCount() + target;
        int p2 = target * map.sectorCount() + view;
        matrix[p1 >> 3] |= (1 << (p1 & 7));
        matrix[p2 >> 3] |= (1 << (p2 & 7));
    }
    M_Free(secGroups);
    return matrix;
}
#endif
/** @file sector.h  World map sector.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/sector.h"
#include "world/map.h"
#include "world/p_object.h"
#include "ConvexSubspace"
#include "Line"
#include "Plane"
#include "Subsector"
#include "Surface"
#include "dd_main.h"  // App_World()
#include <doomsday/console/cmd.h>
#include <de/LogBuffer>
#include <de/vector1.h>
#include <de/Rectangle>
#include <QList>
#include <QtAlgorithms>
/** @file sky.cpp  Sky behavior logic for the world system.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/sky.h"
#include <cmath>
#include <QtAlgorithms>
#include <de/Log>
#include <doomsday/res/TextureManifest>
#include <doomsday/world/Materials>
#include "dd_main.h"
#ifdef __CLIENT__
#  include "gl/gl_main.h"
#  include "gl/gl_tex.h"
#  include "render/rend_main.h"    // rendSkyLightAuto
#  include "render/skydrawable.h"  // SkyDrawable::layerMaterialSpec
#  include "MaterialAnimator"
#  include "ClientTexture"
#endif
#define NUM_LAYERS  2
/** @file subsector.cpp  World map subsector.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2016 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/subsector.h"
#include "world/map.h"
#include "BspLeaf"
#include "ConvexSubspace"
#include "Plane"
#include "Surface"
#include "Face"
#include "dd_main.h"  // App_World()
#include <de/aabox.h>
#include <de/vector1.h>
#include <de/Rectangle>
/** @file surface.cpp  World map surface.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "world/surface.h"
#include "world/clientserverworld.h" // ddMapSetup
#include "world/map.h"
#include "Plane"
#ifdef __CLIENT__
#  include "Decoration"
#endif
#include "misc/r_util.h" // R_NameForBlendMode()
#ifdef __CLIENT__
#  include "gl/gl_tex.h"
#  include "render/rend_main.h"
#  include "resource/clienttexture.h"
#  include "dd_loop.h" // frameTimePos
#  include <doomsday/resource/texturemanifest.h>
#endif
#include <doomsday/world/MaterialManifest>
#include <doomsday/world/Material>
#include <de/Log>
#include <de/vector1.h>
/** @file thinkers.cpp  World map thinker management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#define DENG_NO_API_MACROS_THINKER
#include "de_base.h"
#include "world/thinkers.h"
#ifdef __CLIENT__
#  include "client/cl_mobj.h"
#  include "world/clientmobjthinkerdata.h"
#endif
#ifdef __SERVER__
#  include "def_main.h"
#  include "server/sv_pool.h"
#  include <doomsday/world/mobjthinkerdata.h>
#endif
#include "world/map.h"
#include "world/p_object.h"
#include <de/memoryzone.h>
#include <QList>
#include <QtAlgorithms>
/** @file vertex.cpp  World map vertex.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "de_base.h"
#include "world/vertex.h"
#ifdef __CLIENT__
#  include <de/Partition>
#endif
#include <de/Vector>
#include "Line"
#include "world/lineowner.h"  /// @todo remove me
#include "Sector"
using namespace de;
using namespace world;
namespace world {
static Value *Function_World_FindThings(Context &, const Function::ArgumentValues &args) {
    const int type = args.at(0)->asInt();
    std::unique_ptr<ArrayValue> things(new ArrayValue);
    App_World().map().thinkers().forAll(1 | 2, [&things, type](thinker_t *th) {
        if (Thinker_IsMobj(th)) {
            const mobj_t *mo = (mobj_t *) th;
            if (mo->type == type) {
                things->add(new RecordValue(THINKER_DATA(*th, MobjThinkerData).objectNamespace()));
            }
        }
        return LoopContinue;
    });
    return things.release();
}
static Value *Function_Thing_Init(Context &ctx, const Function::ArgumentValues &args) {
    ctx.nativeSelf().as<RecordValue>().dereference().set("__id__", args.at(0)->asInt());
    return nullptr;
}
static Value *Function_Thing_SetState(Context &ctx, const Function::ArgumentValues &args) {
    auto &mo = ClientServerWorld::contextMobj(ctx);
    Mobj_SetState(&mo, args.at(0)->asInt());
    return nullptr;
}
static Value *Function_Thing_State(Context &ctx, const Function::ArgumentValues &) {
    const auto &mo = ClientServerWorld::contextMobj(ctx);
    return new NumberValue(runtimeDefs.states.indexOf(mo.state));
}
static Value *Function_Thing_AddMom(Context &ctx, const Function::ArgumentValues &args) {
    mobj_t &   mo    = ClientServerWorld::contextMobj(ctx);
    const auto delta = vectorFromValue<Vector3d>(*args.at(0));
    mo.mom[VX] += delta.x;
    mo.mom[VY] += delta.y;
    mo.mom[VZ] += delta.z;
    return nullptr;
}
static Value *Function_Thing_ChangeFlags(Context &ctx, const Function::ArgumentValues &args) {
    const int flagsIndex = args.at(0)->asInt();
    auto &mo = ClientServerWorld::contextMobj(ctx);
    int &flags = (flagsIndex == 3 ? mo.flags3 : flagsIndex == 2 ? mo.flags2 : mo.flags);
    const int oldFlags = flags;
    const auto value = args.at(1)->asUInt();
    if (args.at(2)->isTrue()) {
        flags |= value;
    }
    else {
        flags &= ~value;
    }
    return new NumberValue(oldFlags);
}
static Value *Function_Thing_Flags(Context &ctx, const Function::ArgumentValues &args) {
    const int flagsIndex = args.at(0)->asInt();
    const auto &mo = ClientServerWorld::contextMobj(ctx);
    return new NumberValue(uint32_t(flagsIndex == 3 ? mo.flags3 : flagsIndex == 2 ? mo.flags2 : mo.flags));
}
static Value *Function_Thing_Id(Context &ctx, const Function::ArgumentValues &) {
    return new NumberValue(ClientServerWorld::contextMobj(ctx).thinker.id);
}
static Value *Function_Thing_Health(Context &ctx, const Function::ArgumentValues &) {
    return new NumberValue(ClientServerWorld::contextMobj(ctx).health);
}
static Value *Function_Thing_Height(Context &ctx, const Function::ArgumentValues &) {
    return new NumberValue(ClientServerWorld::contextMobj(ctx).height);
}
static Value *Function_Thing_Mom(Context &ctx, const Function::ArgumentValues &) {
    return new ArrayValue(Vector3d(ClientServerWorld::contextMobj(ctx).mom));
}
static Value *Function_Thing_StartSound(Context &ctx, const Function::ArgumentValues &args) {
    const mobj_t &mo     = ClientServerWorld::contextMobj(ctx);
    const int     sound  = DED_Definitions()->getSoundNum(args.at(0)->asText());
    const float   volume = float(args.at(1)->asNumber());
    if (sound >= 0) {
        S_StartSoundAtVolume(sound, &mo, volume);
    }
    else {
        throw Error("Function_Thing_StartSound", "Undefined sound: " + args.at(0)->asText());
    }
    return nullptr;
}
static Value *Function_Thing_Player(Context &ctx, const Function::ArgumentValues &) {
    const mobj_t &mo = ClientServerWorld::contextMobj(ctx);
    if (mo.dPlayer) {
        auto &plrs = DoomsdayApp::players();
        return new RecordValue(plrs.at(plrs.indexOf(mo.dPlayer)).objectNamespace());
    }
    return nullptr;
}
static Value *Function_Thing_Pos(Context &ctx, const Function::ArgumentValues &) {
    return new ArrayValue(Vector3d(ClientServerWorld::contextMobj(ctx).origin));
}
static Value *Function_Thing_Recoil(Context &ctx, const Function::ArgumentValues &args) {
    mobj_t &     mo    = ClientServerWorld::contextMobj(ctx);
    const double force = args.at(0)->asNumber();
    const angle_t angle = mo.angle + ANG180;
    const float angle_f = float(angle) / float(ANGLE_180) * PIf;
    mo.mom[MX] += force * cos(angle_f);
    mo.mom[MY] += force * sin(angle_f);
    return nullptr;
}
static Value *Function_Thing_Type(Context &ctx, const Function::ArgumentValues &) {
    return new NumberValue(ClientServerWorld::contextMobj(ctx).type);
}
void initBindings(Binder &binder, Record &worldModule) { {
        binder.init(worldModule)
            << DE_FUNC(World_FindThings, "findThings", "typeIndex"); // TODO: add more params to make generic finder
    } {
        Record &thing = worldModule.addSubrecord("Thing");
        Function::Defaults startSoundArgs;
        startSoundArgs["volume"] = new NumberValue(1.0);
        binder.init(thing)
                << DE_FUNC         (Thing_Init,       "__init__", "id")
                << DENG2_FUNC      (Thing_AddMom,     "addMom", "delta")
                << DENG2_FUNC      (Thing_ChangeFlags,"changeFlags", "index" << "flags" << "doSet")
                << DENG2_FUNC      (Thing_Flags,      "flags", "index")
                << DENG2_FUNC_NOARG(Thing_Id,         "id")
                << DENG2_FUNC_NOARG(Thing_Health,     "health")
                << DENG2_FUNC_NOARG(Thing_Height,     "height")
                << DENG2_FUNC_NOARG(Thing_Mom,        "mom")
                << DENG2_FUNC_NOARG(Thing_Player,     "player")
                << DENG2_FUNC_NOARG(Thing_Pos,        "pos")
                << DE_FUNC         (Thing_SetState,   "setState", "index")
                << DENG2_FUNC_DEFS (Thing_StartSound, "startSound", "id" << "volume", startSoundArgs)
                << DE_FUNC_NOARG   (Thing_State,      "state")
                << DENG2_FUNC      (Thing_Recoil,     "recoil", "force")
                << DENG2_FUNC_NOARG(Thing_Type,       "type");
    }
}
}  // namespace world
namespace world {
struct RingNode {
    void *elem;
    RingNode *prev;
    RingNode *next;
};
struct CellData {
    RingNode *ringNodes;
    dint elemCount; ///< Total number of linked elements.
    bool unlink(void *elem) {
        if(RingNode *node = findNode(elem)) {
            clearElement(*node);
            return true;
        }
        return false;
    }
    void unlinkAll() {
        for(RingNode *node = ringNodes; node; node = node->next) {
            clearElement(*node);
        }
    }
    bool link(void *elem) {
        addElement(newNode(), elem);
        return true;
    }
private:
    RingNode &newNode() {
        RingNode *node = nullptr;
        if(!ringNodes) {
            node = (RingNode *) Z_Malloc(sizeof(*node), PU_MAP, nullptr);
            node->next = nullptr;
            node->prev = nullptr;
            node->elem = nullptr;
            ringNodes = node;
            return *node;
        }
        for(node = ringNodes; node->next && node->elem; node = node->next)
        {}
        if(!node->elem) {
            return *node;
        }
        node->next = (RingNode *) Z_Malloc(sizeof(*node), PU_MAP, nullptr);
        node->next->next = nullptr;
        node->next->prev = node;
        node->next->elem = nullptr;
        return *node->next;
    }
    RingNode *findNode(void *elem) {
        if(elem) {
            for(RingNode *found = ringNodes; found; found = found->next) {
                if(found->elem == elem) return found;
            }
        }
        return nullptr;
    }
    void clearElement(RingNode &node) {
        if(!node.elem) return;
        node.elem = nullptr;
        elemCount--;
    }
    void addElement(RingNode &node, void *elem) {
        DENG2_ASSERT(node.elem == nullptr);
        node.elem = elem;
        elemCount++;
    }
};
DENG2_PIMPL(Blockmap) {
    /**
     * Quadtree node.
     */
    struct Node {
        enum Quadrant {
            TopLeft,
            TopRight,
            BottomLeft,
            BottomRight
        };
        Cell cell;   ///< Cell coordinates for this node.
        duint size;  ///< Size of the cell at this node (width=height).
        union {
            Node *children[4];   ///< One per quadrant (if any, not owned).
            CellData *leafData;  ///< Data associated with the leaf cell.
        };
        /**
         * Construct a new node.
         *
         * @param cell  Cell coordinates for the node.
         * @param size  Size of the cell.
         */
        Node(Cell const &cell, duint size)
            : cell(cell)
            , size(size) {
            zap(children);
        }
        ~Node() {
            if(isLeaf()) Z_Free(leafData);
        }
        /**
         * Returns @c true iff the cell is a leaf (i.e., equal to a unit in the
         * gridmap coordinate space).
         */
        inline bool isLeaf() const { return size == 1; }
        /**
         * In which child quadrant is the given @a point?
         */
        Quadrant quadrant(Cell const &point) const {
            duint const subSize = size >> 1;
            if(point.x < cell.x + subSize) {
                return (point.y < cell.y + subSize)? TopLeft  : BottomLeft;
            }
            else {
                return (point.y < cell.y + subSize)? TopRight : BottomRight;
            }
        }
    };
    typedef QList<Node> Nodes;
    AABoxd bounds;    ///< Map space units.
    duint cellSize;   ///< Map space units.
    Cell dimensions;  ///< Dimensions of the indexed space, in cells.
    Nodes nodes;      ///< Quadtree nodes. The first being the root.
    Impl(Public *i, AABoxd const &bounds, duint cellSize)
        : Base(i)
        , bounds    (bounds)
        , cellSize  (cellSize)
        , dimensions(Vector2ui(de::ceil((bounds.maxX - bounds.minX) / cellSize),
                               de::ceil((bounds.maxY - bounds.minY) / cellSize))) {
        newNode(Cell(0, 0), ceilPow2(de::max(dimensions.x, dimensions.y)));
    }
    inline dint toCellIndex(duint cellX, duint cellY) {
        return dint(cellY * dimensions.x + cellX);
    }
    /**
     * Given map space X coordinate @a x, return the corresponding cell coordinate.
     * If @a x is outside the blockmap it will be clamped to the nearest edge on
     * the X axis.
     *
     * @param x        Map space X coordinate to be translated.
     * @param didClip  Set to @c true iff clamping was necessary.
     *
     * @return  Translated blockmap cell X coordinate.
     */
    duint toCellX(ddouble x, bool &didClip) {
        didClip = false;
        if(x < bounds.minX) {
            x = bounds.minX;
            didClip = true;
        }
        else if(x >= bounds.maxX) {
            x = bounds.maxX - 1;
            didClip = true;
        }
        return duint((x - bounds.minX) / cellSize);
    }
    /**
     * Given map space Y coordinate @a y, return the corresponding cell coordinate.
     * If @a y is outside the blockmap it will be clamped to the nearest edge on
     * the Y axis.
     *
     * @param y        Map space Y coordinate to be translated.
     * @param didClip  Set to @c true iff clamping was necessary.
     *
     * @return  Translated blockmap cell Y coordinate.
     */
    duint toCellY(ddouble y, bool &didClip) {
        didClip = false;
        if(y < bounds.minY) {
            y = bounds.minY;
            didClip = true;
        }
        else if(y >= bounds.maxY) {
            y = bounds.maxY - 1;
            didClip = true;
        }
        return uint((y - bounds.minY) / cellSize);
    }
    bool clipCell(Cell &cell) const {
        bool didClip = false;
        if(cell.x > dimensions.x) {
            cell.x = dimensions.x;
            didClip = true;
        }
        if(cell.y > dimensions.y) {
            cell.y = dimensions.y;
            didClip = true;
        }
        return didClip;
    }
    /**
     * Clip the cell coordinates in @a block vs the dimensions of this gridmap
     * so that they are inside the boundary this defines.
     *
     * @param block  Block coordinates to be clipped.
     *
     * @return  @c true iff the block coordinates were changed.
     */
    bool clipBlock(CellBlock &block) {
        bool const didClipMin = clipCell(block.min);
        bool const didClipMax = clipCell(block.max);
        return didClipMin | didClipMax;
    }
    Node *newNode(Cell const &at, duint size) {
        nodes.append(Node(at, size));
        return &nodes.last();
    }
    Node *findLeaf(Node *node, Cell const &at, bool canSubdivide) {
        if(node->isLeaf()) return node;
        Node::Quadrant q = node->quadrant(at);
        Node **childAdr = &node->children[q];
        if(!*childAdr) {
            if(!canSubdivide) return nullptr;
            duint const subSize = node->size >> 1;
            switch(q) {
            case Node::TopLeft:
                *childAdr = newNode(node->cell, subSize);
                break;
            case Node::TopRight:
                *childAdr = newNode(Cell(node->cell.x + subSize, node->cell.y), subSize);
                break;
            case Node::BottomLeft:
                *childAdr = newNode(Cell(node->cell.x, node->cell.y + subSize), subSize);
                break;
            case Node::BottomRight:
                *childAdr = newNode(Cell(node->cell.x + subSize, node->cell.y + subSize), subSize);
                break;
            }
        }
        return findLeaf(*childAdr, at, canSubdivide);
    }
    inline Node *findLeaf(Cell const &at, bool canCreate = false) {
        return findLeaf(&nodes.first(), at, canCreate);
    }
    /**
     * Retrieve the user data associated with the identified cell.
     *
     * @param cell       Cell coordinates to retrieve user data for.
     * @param canCreate  @c true= allocate new data if not already present
     *                   for the referenced cell.
     *
     * @return  User data for the identified cell else @c 0if an invalid
     *          reference or no there is no data present (and not allocating).
     */
    CellData *cellData(Cell const &cell, bool canCreate = false) {
        if(cell.x >= dimensions.x || cell.y >= dimensions.y) {
            return nullptr;
        }
        if(Node *node = findLeaf(cell, canCreate)) {
            if(!node->leafData) {
                if(canCreate) {
                    node->leafData = (CellData *)Z_Calloc(sizeof(CellData), PU_MAPSTATIC, nullptr);
                }
            }
            return node->leafData;
        }
        return nullptr;
    }
};
Blockmap::Blockmap(AABoxd const &bounds, duint cellSize)
    : d(new Impl(this, bounds, cellSize))
{}
Blockmap::~Blockmap()
{}
Vector2d Blockmap::origin() const {
    return Vector2d(d->bounds.min);
}
AABoxd const &Blockmap::bounds() const {
    return d->bounds;
}
BlockmapCell const &Blockmap::dimensions() const {
    return d->dimensions;
}
duint Blockmap::cellSize() const {
    return d->cellSize;
}
dint Blockmap::toCellIndex(duint cellX, duint cellY) const {
    return d->toCellIndex(cellX, cellY);
}
BlockmapCell Blockmap::toCell(Vector2d const &point, bool *retDidClip) const {
    bool didClipX, didClipY;
    Cell cell(d->toCellX(point.x, didClipX), d->toCellY(point.y, didClipY));
    if(retDidClip) *retDidClip = didClipX | didClipY;
    return cell;
}
BlockmapCellBlock Blockmap::toCellBlock(AABoxd const &box, bool *retDidClip) const {
    bool didClipMin, didClipMax;
    CellBlock block(toCell(box.min, &didClipMin), toCell(box.max, &didClipMax));
    block.max += Vector2ui(1, 1); // CellBlock is inclusive-exclusive.
    if(retDidClip) *retDidClip = didClipMin | didClipMax;
    return block;
}
bool Blockmap::link(Cell const &cell, void *elem) {
    if(!elem) return false; // Huh?
    if(auto *cellData = d->cellData(cell, true /*can create*/))
    {
        return cellData->link(elem);
    }
    return false; // Outside the blockmap?
}
bool Blockmap::link(AABoxd const &region, void *elem) {
    if(!elem) return false; // Huh?
    bool didLink = false;
    CellBlock cellBlock = toCellBlock(region);
    d->clipBlock(cellBlock);
    Cell cell;
    for(cell.y = cellBlock.min.y; cell.y < cellBlock.max.y; ++cell.y)
    for(cell.x = cellBlock.min.x; cell.x < cellBlock.max.x; ++cell.x) {
        if(auto *cellData = d->cellData(cell, true)) {
            if(cellData->link(elem)) {
                didLink = true;
            }
        }
    }
    return didLink;
}
bool Blockmap::unlink(Cell const &cell, void *elem) {
    if(!elem) return false; // Huh?
    if(auto *cellData = d->cellData(cell)) {
        return cellData->unlink(elem);
    }
    return false;
}
bool Blockmap::unlink(AABoxd const &region, void *elem) {
    if(!elem) return false; // Huh?
    bool didUnlink = false;
    CellBlock cellBlock = toCellBlock(region);
    d->clipBlock(cellBlock);
    Cell cell;
    for(cell.y = cellBlock.min.y; cell.y < cellBlock.max.y; ++cell.y)
    for(cell.x = cellBlock.min.x; cell.x < cellBlock.max.x; ++cell.x) {
        if(auto *cellData = d->cellData(cell)) {
            if(cellData->unlink(elem)) {
                didUnlink = true;
            }
        }
    }
    return didUnlink;
}
void Blockmap::unlinkAll() {
    for(Impl::Node const &node : d->nodes) {
        if(!node.isLeaf()) continue;
        if(auto *cellData = node.leafData) {
            cellData->unlinkAll();
        }
    }
}
dint Blockmap::cellElementCount(Cell const &cell) const {
    if(auto *cellData = d->cellData(cell)) {
        return cellData->elemCount;
    }
    return 0;
}
LoopResult Blockmap::forAllInCell(Cell const &cell, std::function<LoopResult (void *object)> func) const {
    if(auto *cellData = d->cellData(cell)) {
        RingNode *node = cellData->ringNodes;
        while(node) {
            RingNode *next = node->next;
            if(node->elem) {
                if(auto result = func(node->elem)) return result;
            }
            node = next;
        }
    }
    return LoopContinue;
}
LoopResult Blockmap::forAllInBox(AABoxd const &box, std::function<LoopResult (void *object)> func) const {
    CellBlock cellBlock = toCellBlock(box);
    d->clipBlock(cellBlock);
    Cell cell;
    for(cell.y = cellBlock.min.y; cell.y < cellBlock.max.y; ++cell.y)
    for(cell.x = cellBlock.min.x; cell.x < cellBlock.max.x; ++cell.x) {
        if(auto result = forAllInCell(cell, func)) return result;
    }
    return LoopContinue;
}
LoopResult Blockmap::forAllInPath(Vector2d const &from_, Vector2d const &to_,
    std::function<LoopResult (void *object)> func) const {
    Vector2d from = from_;
    Vector2d to   = to_;
    if(!(from.x >= d->bounds.minX && from.x <= d->bounds.maxX &&
         from.y >= d->bounds.minY && from.y <= d->bounds.maxY)) {
        return LoopContinue;
    }
    if((from.x < d->bounds.minX && to.x < d->bounds.minX) ||
       (from.x > d->bounds.maxX && to.x > d->bounds.maxX) ||
       (from.y < d->bounds.minY && to.y < d->bounds.minY) ||
       (from.y > d->bounds.maxY && to.y > d->bounds.maxY)) {
        return LoopContinue;
    }
    /*
     * Trace lines should not be perfectly parallel to a blockmap axis. This is
     * because lines in the line blockmap are only linked in the blocks on one
     * side of the axis parallel line. Note that the same logic is applied to all
     * of the blockmaps and not just the line blockmap.
     */
    ddouble const epsilon = FIX2FLT(FRACUNIT);
    ddouble const offset  = FIX2FLT(FRACUNIT);
    Vector2d const delta = (to - origin()) / d->cellSize;
    if(de::fequal(delta.x, 0, epsilon)) to.x += offset;
    if(de::fequal(delta.y, 0, epsilon)) to.y += offset;
    /*
     * It is possible that one or both points are outside the blockmap. Clip the
     * trace so that is wholly within the AABB of the blockmap (note we would have
     * already abandoned if the origin of the trace is outside).
     */
    if(!(to.x >= d->bounds.minX && to.x <= d->bounds.maxX &&
         to.y >= d->bounds.minY && to.y <= d->bounds.maxY)) {
        vec2d_t fromV1, toV1, bmapBounds[4], point;
        ddouble ab;
        V2d_Set(fromV1, from.x, from.y);
        V2d_Set(toV1, to.x, to.y);
        V2d_Set(bmapBounds[0], d->bounds.minX, d->bounds.minY);
        V2d_Set(bmapBounds[1], d->bounds.minX, d->bounds.maxY);
        V2d_Set(bmapBounds[2], d->bounds.maxX, d->bounds.maxY);
        V2d_Set(bmapBounds[3], d->bounds.maxX, d->bounds.minY);
        ab = V2d_Intercept(fromV1, toV1, bmapBounds[0], bmapBounds[1], point);
        if(ab >= 0 && ab <= 1)
            V2d_Copy(toV1, point);
        ab = V2d_Intercept(fromV1, toV1, bmapBounds[1], bmapBounds[2], point);
        if(ab >= 0 && ab <= 1)
            V2d_Copy(toV1, point);
        ab = V2d_Intercept(fromV1, toV1, bmapBounds[2], bmapBounds[3], point);
        if(ab >= 0 && ab <= 1)
            V2d_Copy(toV1, point);
        ab = V2d_Intercept(fromV1, toV1, bmapBounds[3], bmapBounds[0], point);
        if(ab >= 0 && ab <= 1)
            V2d_Copy(toV1, point);
        from = Vector2d(fromV1);
        to   = Vector2d(toV1);
    }
    BlockmapCell const originCell = toCell(from);
    BlockmapCell const destCell   = toCell(to);
    Vector2d intercept = (from - origin()) / d->cellSize;
    Vector2i cellStep;
    Vector2d interceptStep;
    Vector2d frac;
    if(destCell.x == originCell.x) {
        cellStep.x      = 0;
        interceptStep.y = 256;
        frac.y          = 1;
    }
    else if(destCell.x > originCell.x) {
        cellStep.x      = 1;
        interceptStep.y = (to.y - from.y) / de::abs(to.x - from.x);
        frac.y          = 1 - (intercept.x - int( intercept.x ));
    }
    else // toCell.x < fromCell.x
    {
        cellStep.x      = -1;
        interceptStep.y = (to.y - from.y) / de::abs(to.x - from.x);
        frac.y          = intercept.x - int( intercept.x );
    }
    if(destCell.y == originCell.y) {
        cellStep.y      = 0;
        interceptStep.x = 256;
        frac.x          = 1;
    }
    else if(destCell.y > originCell.y) {
        cellStep.y      = 1;
        interceptStep.x = (to.x - from.x) / de::abs(to.y - from.y);
        frac.x          = 1 - (intercept.y - int( intercept.y ));
    }
    else // toCell.y < fromCell.y
    {
        cellStep.y      = -1;
        interceptStep.x = (to.x - from.x) / de::abs(to.y - from.y);
        frac.x          = intercept.y - int( intercept.y );
    }
    intercept += frac * interceptStep;
    BlockmapCell cell = originCell;
    for(int pass = 0; pass < 64; ++pass) // Prevent a round off error leading us into
    {
        if(auto result = forAllInCell(cell, func))
            return result;
        if(cell == destCell) break;
        if(cell.y == uint( intercept.y )) {
            cell.x      += cellStep.x;
            intercept.y += interceptStep.y;
        }
        else if(cell.x == uint( intercept.x )) {
            cell.y      += cellStep.y;
            intercept.x += interceptStep.x;
        }
    }
    return LoopContinue;
}
#ifdef __CLIENT__
void Blockmap::drawDebugVisual() const {
#define UNIT_SIZE           1
    GLfloat oldColor[4];
    DGL_CurrentColor(oldColor);
    /*
     * Draw the Quadtree.
     */
    DGL_Color4f(1.f, 1.f, 1.f, 1.f / d->nodes.first().size);
    foreach(Impl::Node const &node, d->nodes) {
        if(!node.isLeaf()) continue;
        if(!node.leafData) continue;
        Vector2f const topLeft     = node.cell * UNIT_SIZE;
        Vector2f const bottomRight = topLeft + Vector2f(UNIT_SIZE, UNIT_SIZE);
        DGL_Begin(DGL_LINE_STRIP);
            DGL_Vertex2f(topLeft.x,     topLeft.y);
            DGL_Vertex2f(bottomRight.x, topLeft.y);
            DGL_Vertex2f(bottomRight.x, bottomRight.y);
            DGL_Vertex2f(topLeft.x,     bottomRight.y);
            DGL_Vertex2f(topLeft.x,     topLeft.y);
        DGL_End();
    }
    /*
     * Draw the bounds.
     */
    Vector2f start;
    Vector2f end = start + d->dimensions * UNIT_SIZE;
    DGL_Color3f(1, .5f, .5f);
    DGL_Begin(DGL_LINES);
        DGL_Vertex2f(start.x, start.y);
        DGL_Vertex2f(  end.x, start.y);
        DGL_Vertex2f(  end.x, start.y);
        DGL_Vertex2f(  end.x,   end.y);
        DGL_Vertex2f(  end.x,   end.y);
        DGL_Vertex2f(start.x,   end.y);
        DGL_Vertex2f(start.x,   end.y);
        DGL_Vertex2f(start.x, start.y);
    DGL_End();
    DGL_Color4fv(oldColor);
#undef UNIT_SIZE
}
#endif  // __CLIENT__
}  // namespace world
namespace world {
BspLeaf::BspLeaf(Sector *sector) {
    _sector = sector;
}
bool BspLeaf::hasSubspace() const {
    return _subspace != nullptr;
}
ConvexSubspace &BspLeaf::subspace() const {
    if (hasSubspace()) return *_subspace;
    throw MissingSubspaceError("BspLeaf::subspace", "No subspace is attributed");
}
ConvexSubspace *BspLeaf::subspacePtr() const {
    return _subspace;
}
void BspLeaf::setSubspace(ConvexSubspace *newSubspace) {
    if(_subspace == newSubspace) return;
    if(hasSubspace()) {
        _subspace->setBspLeaf(nullptr);
    }
    _subspace = newSubspace;
    if(hasSubspace()) {
        _subspace->setBspLeaf(this);
    }
}
void BspLeaf::setSector(Sector *newSector) {
    _sector = newSector;
}
}  // namespace world
dint validCount = 1;  // Increment every time a check is made.
#ifdef __CLIENT__
static inline RenderSystem &rendSys() {
    return ClientApp::renderSystem();
}
#endif
/**
 * Observes the progress of a map conversion and records any issues/problems that
 * are encountered in the process. When asked, compiles a human-readable report
 * intended to assist mod authors in debugging their maps.
 *
 * @todo Consolidate with the missing material reporting done elsewhere -ds
 */
class MapConversionReporter
: DENG2_OBSERVES(Map,     UnclosedSectorFound)
, DENG2_OBSERVES(Map,     OneWayWindowFound)
, DENG2_OBSERVES(BaseMap, Deletion) {
    typedef std::map<dint, Vector2i> UnclosedSectorMap;
    typedef std::map<dint, dint> OneWayWindowMap;
    static dint const maxWarningsPerType;
public:
    /**
     * Construct a new conversion reporter.
     * @param map
     */
    MapConversionReporter(Map *map = nullptr) {
        setMap(map);
    }
    ~MapConversionReporter() {
        observeMap(false);
    }
    /**
     * Change the map to be reported on. Note that any existing report data is
     * retained until explicitly cleared.
     */
    void setMap(Map *newMap) {
        if (_map != newMap) {
            observeMap(false);
            _map = newMap;
            observeMap(true);
        }
    }
    inline void setMapAndClearReport(Map *newMap) {
        setMap(newMap);
        clearReport();
    }
    inline void clearMap() { setMap(nullptr); }
    /**
     * Clear any existing conversion report data.
     */
    void clearReport() {
        _unclosedSectors.clear();
        _oneWayWindows.clear();
    }
    /**
     * Compile and output any existing report data to the message log.
     */
    void writeLog() {
        if (dint numToLog = maxWarnings(unclosedSectorCount())) {
            String str;
            UnclosedSectorMap::const_iterator it = _unclosedSectors.begin();
            for (dint i = 0; i < numToLog; ++i, ++it) {
                if (i != 0) str += "\n";
                str += String("Sector #%1 is unclosed near %2")
                           .arg(it->first).arg(it->second.asText());
            }
            if (numToLog < unclosedSectorCount())
                str += String("\n(%1 more like this)").arg(unclosedSectorCount() - numToLog);
            LOGDEV_MAP_WARNING("%s") << str;
        }
        if (dint numToLog = maxWarnings(oneWayWindowCount())) {
            String str;
            OneWayWindowMap::const_iterator it = _oneWayWindows.begin();
            for (dint i = 0; i < numToLog; ++i, ++it) {
                if (i != 0) str += "\n";
                str += String("Line #%1 seems to be a One-Way Window (back faces sector #%2).")
                           .arg(it->first).arg(it->second);
            }
            if (numToLog < oneWayWindowCount())
                str += String("\n(%1 more like this)").arg(oneWayWindowCount() - numToLog);
            LOGDEV_MAP_MSG("%s") << str;
        }
    }
protected:
    void unclosedSectorFound(Sector &sector, Vector2d const &nearPoint) {
        _unclosedSectors.insert(std::make_pair(sector.indexInArchive(), nearPoint.toVector2i()));
    }
    void oneWayWindowFound(Line &line, Sector &backFacingSector) {
        _oneWayWindows.insert(std::make_pair(line.indexInArchive(), backFacingSector.indexInArchive()));
    }
    void mapBeingDeleted(BaseMap const &map) {
        DENG2_ASSERT(&map == _map);  // sanity check.
        DENG2_UNUSED(map);
        _map = nullptr;
    }
private:
    inline dint unclosedSectorCount() const { return dint( _unclosedSectors.size() ); }
    inline dint oneWayWindowCount() const   { return dint( _oneWayWindows.size() ); }
    static inline dint maxWarnings(dint issueCount) {
#ifdef DENG2_DEBUG
        return issueCount; // No limit.
#else
        return de::min(issueCount, maxWarningsPerType);
#endif
    }
    void observeMap(bool yes) {
        if (!_map) return;
        if (yes) {
            _map->audienceForDeletion()          += this;
            _map->audienceForOneWayWindowFound   += this;
            _map->audienceForUnclosedSectorFound += this;
        }
        else {
            _map->audienceForDeletion()          -= this;
            _map->audienceForOneWayWindowFound   -= this;
            _map->audienceForUnclosedSectorFound -= this;
        }
    }
    Map *_map = nullptr;  ///< Map currently being reported on, if any (not owned).
    UnclosedSectorMap _unclosedSectors;
    OneWayWindowMap   _oneWayWindows;
};
dint const MapConversionReporter::maxWarningsPerType = 10;
dd_bool ddMapSetup;
static char const *mapCacheDir = "mapcache/";
static String cacheIdForMap(String const &sourcePath) {
    DENG2_ASSERT(!sourcePath.isEmpty());
    dushort id = 0;
    for (dint i = 0; i < sourcePath.size(); ++i) {
        id ^= sourcePath.at(i).unicode() << ((i * 3) % 11);
    }
    return String("%1").arg(id, 4, 16);
}
DENG2_PIMPL(ClientServerWorld) {
    Binder binder;               ///< Doomsday Script bindings for the World.
    Record worldModule;
    Record fallbackMapInfo;      ///< Used when no effective MapInfo definition.
    timespan_t time = 0;         ///< World-wide time.
    Scheduler scheduler;
#if 0
#ifdef __CLIENT__
    std::unique_ptr<Hand> hand;  ///< For map editing/manipulation.
#endif
#endif // 0
    Impl(Public *i) : Base(i) {
        world::initBindings(binder, worldModule);
        ScriptSystem::get().addNativeModule("World", worldModule);
        defn::MapInfo(fallbackMapInfo).resetToDefaults();
        world::DmuArgs::setPointerToIndexFunc(P_ToIndex);
#ifdef __CLIENT__
        world::MaterialManifest::setMaterialConstructor([] (world::MaterialManifest &m) -> world::Material * {
            return new ClientMaterial(m);
        });
        Sector::setSubsectorConstructor([] (QVector<world::ConvexSubspace *> const &sl) -> world::Subsector * {
            return new ClientSubsector(sl);
        });
#else
        world::MaterialManifest::setMaterialConstructor([] (world::MaterialManifest &m) -> world::Material * {
            return new world::Material(m);
        });
        Sector::setSubsectorConstructor([] (QVector<world::ConvexSubspace *> const &sl) -> world::Subsector * {
            return new Subsector(sl);
        });
#endif
    }
    /**
     * Compose the relative path (relative to the runtime directory) to the
     * directory of the cache where maps from this source (e.g., the add-on
     * which contains the map) will reside.
     *
     * @param sourcePath  Path to the primary resource file (the source) for
     *                    the original map data.
     *
     * @return  The composed path.
     */
    static Path cachePath(String sourcePath) {
        if (sourcePath.isEmpty()) return String();
        return mapCacheDir + App_CurrentGame().id()
               / sourcePath.fileNameWithoutExtension()
               + '-' + cacheIdForMap(sourcePath);
    }
    /**
     * Attempt JIT conversion of the map data with the help of a plugin. Note that
     * the map is left in an editable state in case the caller wishes to perform
     * any further changes.
     *
     * @param reporter  Reporter which will observe the conversion process.
     *
     * @return  The newly converted map (if any).
     */
    Map *convertMap(res::MapManifest const &mapManifest, MapConversionReporter *reporter = nullptr) {
        if (!Plug_CheckForHook(HOOK_MAP_CONVERT))
            return nullptr;
        LOG_DEBUG("Attempting \"%s\"...") << mapManifest.composeUri().path();
        if (!mapManifest.sourceFile()) return nullptr;
        MPE_Begin(nullptr/*dummy*/);
        Map *newMap = MPE_Map();
        newMap->setManifest(&const_cast<res::MapManifest &>(mapManifest));
        if (reporter) {
            reporter->setMap(newMap);
        }
        if (!DoomsdayApp::plugins().callAllHooks(HOOK_MAP_CONVERT, 0,
                                                const_cast<Id1MapRecognizer *>(&mapManifest.recognizer())))
            return nullptr;
        MPE_End();
        return MPE_TakeMap();
    }
#if 0
    /**
     * Returns @c true iff data for the map is available in the cache.
     */
    bool haveCachedMap(res::MapManifest &mapManifest) {
        if (!mapCache) return false;
        return DAM_MapIsValid(mapManifest.cachePath, mapManifest.id());
    }
    /**
     * Attempt to load data for the map from the cache.
     *
     * @see isCachedDataAvailable()
     *
     * @return  @c true if loading completed successfully.
     */
    Map *loadMapFromCache(MapManifest &mapManifest) {
        Uri const mapUri = mapManifest.composeUri();
        Map *map = DAM_MapRead(mapManifest.cachePath);
        if (!map)
            throw Error("loadMapFromCache", "Failed loading map \"" + mapUri.asText() + "\" from cache");
        map->_uri = mapUri;
        return map;
    }
#endif
    /**
     * Attempt to load the associated map data.
     *
     * @return  The loaded map if successful. Ownership given to the caller.
     */
    Map *loadMap(res::MapManifest &mapManifest, MapConversionReporter *reporter = nullptr) {
        LOG_AS("ClientServerWorld::loadMap");
        /*if (mapManifest.lastLoadAttemptFailed && !forceRetry)
            return nullptr;
        if (haveCachedMap(mapManifest)) {
            return loadMapFromCache(mapManifest);
        }*/
        Map *map = convertMap(mapManifest, reporter);
        if (!map) {
            LOG_WARNING("Failed conversion of \"%s\".") << mapManifest.composeUri().path();
        }
        return map;
    }
    /**
     * Replace the current map with @a map.
     */
    void makeCurrent(Map *map) {
        self().setMap(map);
        if (!map) return;
        DENG2_ASSERT(!map->isEditable());
        /*if (mapCache && !haveCachedMap(&map->def()))
        {
            F_MakePath(map->def().cachePath.toUtf8().constData());
            DAM_MapWrite(map);
        }*/
#ifdef __CLIENT__
        self().audienceForFrameBegin() += map;
#endif
        LOG_MAP_NOTE(_E(b) "Current map elements:");
        LOG_MAP_NOTE("%s") << map->elementSummaryAsStyledText();
        Record const &mapInfo = map->mapInfo();
        map->_ambientLightLevel = mapInfo.getf("ambient") * 255;
        map->_globalGravity     = mapInfo.getf("gravity");
        map->_effectiveGravity  = map->_globalGravity;
#ifdef __CLIENT__
        defn::Sky skyDef;
        if (Record const *def = DED_Definitions()->skies.tryFind("id", mapInfo.gets("skyId"))) {
            skyDef = *def;
        }
        else {
            skyDef = mapInfo.subrecord("sky");
        }
        map->sky().configure(&skyDef);
        map->skyAnimator().setSky(&rendSys().sky().configure(&map->sky()));
#endif
        map->thinkers().initLists(0x1 | 0x2);
        P_InitUnusedMobjList();
        map->initNodePiles();
#ifdef __CLIENT__
        Cl_ResetFrame();
        Cl_InitPlayers();  // Player data, too.
        map->initGenerators();
#endif
        de::Uri const mapUri = (map->hasManifest() ? map->manifest().composeUri() : de::makeUri("Maps:"));
        if (gx.FinalizeMapChange) {
            gx.FinalizeMapChange(reinterpret_cast<uri_s const *>(&mapUri));
        }
        if (gameTime > 20000000 / TICSPERSEC) {
            gameTime = 0;
        }
        DoomsdayApp::players().forAll([] (Player &plr) {
            plr.extraLight        = 0;
            plr.targetExtraLight  = 0;
            plr.extraLightCounter = 0;
#ifdef __CLIENT__
            auto &client = plr.as<ClientPlayer>();
            client.inVoid = true;
            if (mobj_t const *mob = plr.publicData().mo) {
                if (Mobj_HasSubsector(*mob)) {
                    auto const &subsec = Mobj_Subsector(*mob).as<ClientSubsector>();
                    if (   mob->origin[2] >= subsec.visFloor  ().heightSmoothed()
                        && mob->origin[2] <  subsec.visCeiling().heightSmoothed() - 4) {
                        client.inVoid = false;
                    }
                }
            }
#endif
            return LoopContinue;
        });
#ifdef __CLIENT__
        /*
        map->forAllSectors([] (Sector &sector) {
            return sector.forAllSubsectors([] (Subsector &subsec) {
                return subsec.as<ClientSubsector>().forAllEdgeLoops([] (ClEdgeLoop &loop) {
                    loop.fixSurfacesMissingMaterials();
                    return LoopContinue;
                });
            });
        });
        */
#endif
        map->initPolyobjs();
#ifdef __CLIENT__
        App_AudioSystem().worldMapChanged();
#endif
#ifdef __SERVER__
        if (::isServer) {
            Sv_InitPools();
        }
#endif
#ifdef __CLIENT__
        GL_SetupFogFromMapInfo(mapInfo.accessedRecordPtr());
        map->initSkyFix();
        map->spawnPlaneParticleGens();
        Con_SetProgress(100);
        Time begunPrecacheAt;
        rendSys().sky().cacheAssets();
        App_Resources().cacheForCurrentMap();
        App_Resources().processCacheQueue();
        LOG_RES_VERBOSE("Precaching completed in %.2f seconds") << begunPrecacheAt.since();
        rendSys().clearDrawLists();
        R_InitRendPolyPools();
        Rend_UpdateLightModMatrix();
        map->initRadio();
        map->initContactBlockmaps();
        R_InitContactLists(*map);
        rendSys().worldSystemMapChanged(*map);
        world::Materials::get().updateLookup();
        world::Materials::get().forAnimatedMaterials([] (world::Material &material) {
            return material.as<ClientMaterial>().forAllAnimators([] (MaterialAnimator &animator) {
                animator.rewind();
                return LoopContinue;
            });
        });
#endif
        /*
         * Post-change map setup has now been fully completed.
         */
        String execute = mapInfo.gets("execute");
        if (!execute.isEmpty()) {
            Con_Execute(CMDS_SCRIPT, execute.toUtf8().constData(), true, false);
        }
        if (!mapUri.isEmpty()) {
            String cmd = String("init-") + mapUri.path();
            if (Con_IsValidCommand(cmd.toUtf8().constData())) {
                Con_Executef(CMDS_SCRIPT, false, "%s", cmd.toUtf8().constData());
            }
        }
        time = 0;
        DD_ResetTimer();
#ifdef __CLIENT__
        R_ResetViewer();
        ClientApp::inputSystem().clearEvents();
        firstFrameAfterLoad = true;
#endif
        Z_PrintStatus();
        self().notifyMapChange();
    }
    bool changeMap(res::MapManifest *mapManifest = nullptr) {
        Map *map = self().mapPtr();
        scheduler.clear();
#ifdef __CLIENT__
        if (map) {
            self().audienceForFrameBegin() -= map;
        }
#endif
#ifdef __CLIENT__
        R_DestroyContactLists();
#endif
        delete map;
        self().setMap(nullptr);
        Z_FreeTags(PU_MAP, PU_PURGELEVEL - 1);
        if (!mapManifest) return true;
        LOG_MSG("Loading map \"%s\"...") << mapManifest->composeUri().path();
        ::ddMapSetup = true;
        MapConversionReporter reporter;
        Map *newMap = loadMap(*mapManifest, &reporter);
        if (newMap) {
            bool const mapIsPlayable = newMap->endEditing();
            reporter.setMap(nullptr);
            if (!mapIsPlayable) {
                delete newMap; newMap = nullptr;
            }
        }
        makeCurrent(newMap);
        ::ddMapSetup = false;
        reporter.writeLog();
        return self().hasMap();
    }
#ifdef __CLIENT__
#if 0
    void updateHandOrigin() {
        DENG2_ASSERT(hand != nullptr && self().hasMap());
        viewdata_t const *viewData = &::viewPlayer->viewport();
        hand->setOrigin(viewData->current.origin + viewData->frontVec.xzy() * handDistance);
    }
#endif // 0
    DENG2_PIMPL_AUDIENCE(FrameBegin)
    DENG2_PIMPL_AUDIENCE(FrameEnd)
#endif
};
#ifdef __CLIENT__
DENG2_AUDIENCE_METHOD(ClientServerWorld, FrameBegin)
DENG2_AUDIENCE_METHOD(ClientServerWorld, FrameEnd)
#endif
ClientServerWorld::ClientServerWorld()
    : World()
    , d(new Impl(this))
{}
Map &ClientServerWorld::map() const {
    if (!hasMap()) {
        throw MapError("ClientServerWorld::map", "No map is currently loaded");
    }
    return World::map().as<Map>();
}
bool ClientServerWorld::changeMap(de::Uri const &mapUri) {
    res::MapManifest *mapDef = nullptr;
    if (!mapUri.path().isEmpty()) {
        mapDef = App_Resources().mapManifests().tryFindMapManifest(mapUri);
    }
    if (!mapUri.path().isEmpty() && !DoomsdayApp::app().busyMode().isActive()) {
        return DoomsdayApp::app().busyMode().runNewTaskWithName(
                    BUSYF_ACTIVITY | BUSYF_PROGRESS_BAR | BUSYF_TRANSITION | (::verbose ? BUSYF_CONSOLE_OUTPUT : 0),
                    "Loading map...", [this, &mapDef] (void *) {
            return d->changeMap(mapDef);
        });
    }
    else {
        return d->changeMap(mapDef);
    }
}
void ClientServerWorld::reset() {
    World::reset();
#ifdef __CLIENT__
    if (isClient) {
        Cl_ResetFrame();
        Cl_InitPlayers();
    }
#endif
    unloadMap();
}
void ClientServerWorld::update() {
    DoomsdayApp::players().forAll([] (Player &plr) {
        for (ddpsprite_t &pspr : plr.publicData().pSprites) {
            pspr.statePtr = nullptr;
        }
        return LoopContinue;
    });
    if (hasMap()) {
        map().update();
    }
}
Scheduler &ClientServerWorld::scheduler() {
    return d->scheduler;
}
Record const &ClientServerWorld::mapInfoForMapUri(de::Uri const &mapUri) const {
    if (Record const *def = DED_Definitions()->mapInfos.tryFind("id", mapUri.compose())) {
        return *def;
    }
    if (Record const *def = DED_Definitions()->mapInfos.tryFind("id", de::Uri("Maps", Path("*")).compose())) {
        return *def;
    }
    return d->fallbackMapInfo;
}
void ClientServerWorld::advanceTime(timespan_t delta) {
#if defined (__CLIENT__)
    if (!::clientPaused)
#endif
    {
        d->time += delta;
        d->scheduler.advanceTime(TimeSpan(delta));
    }
}
timespan_t ClientServerWorld::time() const {
    return d->time;
}
void ClientServerWorld::tick(timespan_t elapsed) {
#ifdef __CLIENT__
    if (hasMap()) {
        map().skyAnimator().advanceTime(elapsed);
        if (DD_IsSharpTick()) {
            map().thinkers().forAll(reinterpret_cast<thinkfunc_t>(gx.MobjThinker), 0x1, [] (thinker_t *th) {
                Mobj_AnimateHaloOcclussion(*reinterpret_cast<mobj_t *>(th));
                return LoopContinue;
            });
        }
    }
#else
    DENG2_UNUSED(elapsed);
#endif
}
#ifdef __CLIENT__
#if 0
Hand &ClientServerWorld::hand(ddouble *distance) const {
    if (!d->hand) {
        d->hand.reset(new Hand());
        audienceForFrameEnd() += *d->hand;
        if (hasMap()) {
            d->updateHandOrigin();
        }
    }
    if (distance) {
        *distance = handDistance;
    }
    return *d->hand;
}
#endif // 0
void ClientServerWorld::beginFrame(bool resetNextViewer) {
    DENG2_FOR_AUDIENCE2(FrameBegin, i) i->worldSystemFrameBegins(resetNextViewer);
}
void ClientServerWorld::endFrame() {
#if 0
    if (hasMap() && d->hand) {
        d->updateHandOrigin();
        if (HueCircle *hueCircle = SBE_HueCircle()) {
            viewdata_t const *viewData = &viewPlayer->viewport();
            d->hand->setEditColor(hueCircle->colorAt(viewData->frontVec));
        }
    }
#endif
    DENG2_FOR_AUDIENCE2(FrameEnd, i) i->worldSystemFrameEnds();
}
#endif  // __CLIENT__
void ClientServerWorld::consoleRegister()  // static
{
#ifdef __CLIENT__
#endif
    Map::consoleRegister();
}
mobj_t &ClientServerWorld::contextMobj(const Context &ctx) // static
{
    const int id = ctx.selfInstance().geti(QStringLiteral("__id__"), 0);
    mobj_t *mo = App_World().map().thinkers().mobjById(id);
    if (!mo) {
        throw Map::MissingObjectError("ClientServerWorld::contextMobj",
                                      String::format("Mobj %d does not exist", id));
    }
    return *mo;
}
#ifdef __CLIENT__
ddouble triangleArea(Vector2d const &v1, Vector2d const &v2, Vector2d const &v3) {
    Vector2d a = v2 - v1;
    Vector2d b = v3 - v1;
    return (a.x * b.y - b.x * a.y) / 2;
}
#endif // __CLIENT__
namespace world {
DENG2_PIMPL(ConvexSubspace) {
    Face *poly = nullptr;                  ///< Convex polygon geometry (not owned).
    typedef QSet<de::Mesh *> Meshes;
    Meshes extraMeshes;                    ///< Additional meshes (owned).
    typedef QSet<polyobj_s *> Polyobjs;
    Polyobjs polyobjs;                     ///< Linked polyobjs (not owned).
#ifdef __CLIENT__
    typedef QSet<Lumobj *> Lumobjs;
    Lumobjs lumobjs;                       ///< Linked lumobjs (not owned).
    typedef QSet<LineSide *> ShadowLines;
    ShadowLines shadowLines;               ///< Linked map lines for fake radio shadowing.
    HEdge *fanBase = nullptr;              ///< Trifan base Half-edge (otherwise the center point is used).
    bool needUpdateFanBase = true;         ///< @c true= need to rechoose a fan base half-edge.
    AudioEnvironment audioEnvironment;     ///< Cached audio characteristics.
    dint lastSpriteProjectFrame = 0;       ///< Frame number of last R_AddSprites.
#endif
    dint validCount = 0;                   ///< Used to prevent repeated processing.
    Impl(Public *i) : Base(i) {}
    ~Impl() { qDeleteAll(extraMeshes); }
#ifdef __CLIENT__
    /**
     * Determine the half-edge whose vertex is suitable for use as the center point
     * of a trifan primitive.
     *
     * Note that we do not want any overlapping or zero-area (degenerate) triangles.
     *
     * @par Algorithm
     * <pre>For each vertex
     *    For each triangle
     *        if area is not greater than minimum bound, move to next vertex
     *    Vertex is suitable
     * </pre>
     *
     * If a vertex exists which results in no zero-area triangles it is suitable for
     * use as the center of our trifan. If a suitable vertex is not found then the
     * center of BSP leaf should be selected instead (it will always be valid as
     * BSP leafs are convex).
     */
    void chooseFanBase() {
#define MIN_TRIANGLE_EPSILON  (0.1) ///< Area
        HEdge *firstNode = self().poly().hedge();
        fanBase = firstNode;
        if(self().poly().hedgeCount() > 3) {
            Vertex const *base, *a, *b;
            do {
                HEdge *other = firstNode;
                base = &fanBase->vertex();
                do {
                    if(!(fanBase != firstNode && (other == fanBase || other == &fanBase->prev()))) {
                        a = &other->vertex();
                        b = &other->next().vertex();
                        if(de::abs(triangleArea(base->origin(), a->origin(), b->origin())) <= MIN_TRIANGLE_EPSILON) {
                            base = 0;
                        }
                    }
                } while(base && (other = &other->next()) != firstNode);
                if(!base) {
                    fanBase = &fanBase->next();
                }
            } while(!base && fanBase != firstNode);
            if(!base) // No.
            {
                fanBase = 0;
            }
        }
        needUpdateFanBase = false;
#undef MIN_TRIANGLE_EPSILON
    }
#endif // __CLIENT__
};
ConvexSubspace::ConvexSubspace(Face &convexPolygon, BspLeaf *bspLeaf)
    : MapElement(DMU_SUBSPACE)
    , d(new Impl(this)) {
    d->poly = &convexPolygon;
    poly().setMapElement(this);
    setBspLeaf(bspLeaf);
}
ConvexSubspace *ConvexSubspace::newFromConvexPoly(de::Face &poly, BspLeaf *bspLeaf) // static
{
    if(!poly.isConvex()) {
        throw InvalidPolyError("ConvexSubspace::newFromConvexPoly", "Source is non-convex");
    }
    return new ConvexSubspace(poly, bspLeaf);
}
void ConvexSubspace::setBspLeaf(BspLeaf *newBspLeaf) {
    _bspLeaf = newBspLeaf;
}
Face &ConvexSubspace::poly() const {
    DENG2_ASSERT(d->poly);
    return *d->poly;
}
bool ConvexSubspace::contains(Vector2d const &point) const {
    HEdge const *hedge = poly().hedge();
    do {
        Vertex const &va = hedge->vertex();
        Vertex const &vb = hedge->next().vertex();
        if(((va.origin().y - point.y) * (vb.origin().x - va.origin().x) -
            (va.origin().x - point.x) * (vb.origin().y - va.origin().y)) < 0) {
            return false;
        }
    } while((hedge = &hedge->next()) != poly().hedge());
    return true;
}
void ConvexSubspace::assignExtraMesh(Mesh &newMesh) {
    LOG_AS("ConvexSubspace");
    dint const sizeBefore = d->extraMeshes.size();
    d->extraMeshes.insert(&newMesh);
    if(d->extraMeshes.size() != sizeBefore) {
        LOG_DEBUG("Assigned extra mesh to subspace %p") << this;
        for(Face *face : newMesh.faces()) {
            face->setMapElement(this);
        }
    }
}
LoopResult ConvexSubspace::forAllExtraMeshes(std::function<LoopResult (Mesh &)> func) const {
    for(Mesh *mesh : d->extraMeshes) {
        if(auto result = func(*mesh)) return result;
    }
    return LoopContinue;
}
dint ConvexSubspace::polyobjCount() const {
    return d->polyobjs.count();
}
LoopResult ConvexSubspace::forAllPolyobjs(std::function<LoopResult (Polyobj &)> func) const {
    for(Polyobj *pob : d->polyobjs) {
        if(auto result = func(*pob)) return result;
    }
    return LoopContinue;
}
void ConvexSubspace::link(Polyobj const &polyobj) {
    d->polyobjs.insert(const_cast<Polyobj *>(&polyobj));
}
bool ConvexSubspace::unlink(Polyobj const &polyobj) {
    dint sizeBefore = d->polyobjs.size();
    d->polyobjs.remove(const_cast<Polyobj *>(&polyobj));
    return d->polyobjs.size() != sizeBefore;
}
/*Subsector &ConvexSubspace::subsector() const
{
    if(d->subsector) return *d->subsector;
    throw MissingSubsectorError("ConvexSubspace::subsector", "No subsector is attributed");
}
Subsector *ConvexSubspace::subsectorPtr() const {
    return hasSubsector() ? &subsector() : nullptr;
}*/
void ConvexSubspace::setSubsector(Subsector *newSubsector) {
    _subsector = newSubsector;
}
dint ConvexSubspace::validCount() const {
    return d->validCount;
}
void ConvexSubspace::setValidCount(dint newValidCount) {
    d->validCount = newValidCount;
}
#ifdef __CLIENT__
dint ConvexSubspace::shadowLineCount() const {
    return d->shadowLines.count();
}
void ConvexSubspace::clearShadowLines() {
    d->shadowLines.clear();
}
void ConvexSubspace::addShadowLine(LineSide &side) {
    d->shadowLines.insert(&side);
}
LoopResult ConvexSubspace::forAllShadowLines(std::function<LoopResult (LineSide &)> func) const {
    for(LineSide *side : d->shadowLines) {
        if(auto result = func(*side)) return result;
    }
    return LoopContinue;
}
dint ConvexSubspace::lumobjCount() const {
    return d->lumobjs.count();
}
LoopResult ConvexSubspace::forAllLumobjs(std::function<LoopResult (Lumobj &)> func) const {
    for(Lumobj *lob : d->lumobjs) {
        if(auto result = func(*lob)) return result;
    }
    return LoopContinue;
}
void ConvexSubspace::unlinkAllLumobjs() {
    d->lumobjs.clear();
}
void ConvexSubspace::unlink(Lumobj &lumobj) {
    d->lumobjs.remove(&lumobj);
}
void ConvexSubspace::link(Lumobj &lumobj) {
    d->lumobjs.insert(&lumobj);
}
dint ConvexSubspace::lastSpriteProjectFrame() const {
    return d->lastSpriteProjectFrame;
}
void ConvexSubspace::setLastSpriteProjectFrame(dint newFrameNumber) {
    d->lastSpriteProjectFrame = newFrameNumber;
}
HEdge *ConvexSubspace::fanBase() const {
    if(d->needUpdateFanBase) {
        d->chooseFanBase();
    }
    return d->fanBase;
}
dint ConvexSubspace::fanVertexCount() const {
    return poly().hedgeCount() + (fanBase()? 0 : 2);
}
static void accumReverbForWallSections(HEdge const *hedge,
                                       dfloat envSpaceAccum[NUM_AUDIO_ENVIRONMENTS],
                                       dfloat &total) {
    if(!hedge || !hedge->hasMapElement())
        return;
    LineSideSegment const &seg = hedge->mapElementAs<LineSideSegment>();
    if(!seg.lineSide().hasSections() || !seg.lineSide().middle().hasMaterial())
        return;
    Material &material = seg.lineSide().middle().material();
    AudioEnvironmentId env = material.as<ClientMaterial>().audioEnvironment();
    if(!(env >= 0 && env < NUM_AUDIO_ENVIRONMENTS)) {
        env = AE_WOOD; // Assume it's wood if unknown.
    }
    total += seg.length();
    envSpaceAccum[env] += seg.length();
}
bool ConvexSubspace::updateAudioEnvironment() {
    world::AudioEnvironment &env = d->audioEnvironment;
    if(!hasSubsector()) {
        env.reset();
        return false;
    }
    dfloat contrib[NUM_AUDIO_ENVIRONMENTS]; de::zap(contrib);
    AABoxd const &bounds = poly().bounds();
    env.space = dint(subsector().sector().ceiling().height() - subsector().sector().floor().height())
              * ((bounds.maxX - bounds.minX) * (bounds.maxY - bounds.minY));
    dfloat coverage = 0;
    HEdge *base    = poly().hedge();
    HEdge *hedge   = base;
    do {
        accumReverbForWallSections(hedge, contrib, coverage);
    } while((hedge = &hedge->next()) != base);
    for(Mesh *mesh : d->extraMeshes)
    for(HEdge *hedge : mesh->hedges()) {
        accumReverbForWallSections(hedge, contrib, coverage);
    }
    if(!coverage) {
        env.reset();
        return false;
    }
    for(dint i = AE_FIRST; i < NUM_AUDIO_ENVIRONMENTS; ++i) {
        contrib[i] /= coverage;
    }
    dint volume  = 0;
    dint decay   = 0;
    dint damping = 0;
    for(dint i = AE_FIRST; i < NUM_AUDIO_ENVIRONMENTS; ++i) {
        ::AudioEnvironment const &envDef = S_AudioEnvironment(AudioEnvironmentId(i));
        volume  += envDef.volume  * contrib[i];
        decay   += envDef.decay   * contrib[i];
        damping += envDef.damping * contrib[i];
    }
    env.volume  = de::min(volume,  255);
    env.decay   = de::min(decay,   255);
    env.damping = de::min(damping, 255);
    return true;
}
AudioEnvironment const &ConvexSubspace::audioEnvironment() const {
    return d->audioEnvironment;
}
#endif  // __CLIENT__
}  // namespace world
#ifdef __CLIENT__
static inline InputSystem &inputSys() {
    return ClientApp::inputSystem();
}
#endif
#ifdef __CLIENT__
static int pimpDoubleClickThreshold = 300; ///< Milliseconds, cvar
#endif
DENG2_PIMPL_NOREF(ImpulseAccumulator) {
    int impulseId = 0;
    int playerNum = 0;
    AccumulatorType type = Analog;
    bool expireBeforeSharpTick = false;
    short binaryAccum = 0;
    inline PlayerImpulse &getImpulse() const {
        auto *impulse = P_PlayerImpulsePtr(impulseId);
        DENG2_ASSERT(impulse);
        return *impulse;
    }
#ifdef __CLIENT__
    /**
     * Double-"clicks" actually mean double activations that occur within the
     * double-click threshold. This is to allow double-clicks also from the
     * analog impulses.
     */
    struct DoubleClick {
        enum State {
            None,
            Positive,
            Negative
        };
        bool triggered = false;           //< True if double-click has been detected.
        uint previousClickTime = 0;       //< Previous time an activation occurred.
        State lastState = None;           //< State at the previous time the check was made.
        State previousClickState = None;  /** Previous click state. When duplicated, triggers
                                              the double click. */
    } db;
    /**
     * Track the double-click state of the impulse and generate a bindable
     * symbolic event if the trigger conditions are met.
     *
     * @param pos  State of the impulse.
     */
    void maintainDoubleClick(float pos) {
        if(pimpDoubleClickThreshold <= 0) {
            db.triggered          = false;
            db.previousClickTime  = 0;
            db.previousClickState = DoubleClick::None;
            return;
        }
        DoubleClick::State newState = DoubleClick::None;
        if(pos > .5) {
            newState = DoubleClick::Positive;
        }
        else if(pos < -.5) {
            newState = DoubleClick::Negative;
        }
        else {
            db.lastState = newState; // Release.
            return;
        }
        if(newState == db.lastState) {
            return;
        }
        uint const threshold = uint( de::max(0, pimpDoubleClickThreshold) );
        uint const nowTime   = Timer_RealMilliseconds();
        if(newState == db.previousClickState && nowTime - db.previousClickTime < threshold) {
            db.triggered = true;
            PlayerImpulse const &impulse = getImpulse();
            String symbolicName;
            switch(newState) {
            case DoubleClick::Positive: symbolicName += "control-doubleclick-positive-"; break;
            case DoubleClick::Negative: symbolicName += "control-doubleclick-negative-"; break;
            default: break;
            }
            symbolicName += impulse.name;
            int const localPlayer = P_ConsoleToLocal(playerNum);
            DENG2_ASSERT(localPlayer >= 0);
            LOG_INPUT_XVERBOSE("Triggered " _E(b) "'%s'" _E(.) " for player%i state: %i threshold: %i\n  %s",
                               impulse.name << (localPlayer + 1) << newState
                               << (nowTime - db.previousClickTime) << symbolicName);
            Block symbolicNameUtf8 = symbolicName.toUtf8();
            ddevent_t ev; de::zap(ev);
            ev.device = uint(-1);
            ev.type   = E_SYMBOLIC;
            ev.symbolic.id   = playerNum;
            ev.symbolic.name = symbolicNameUtf8.constData();
            inputSys().postEvent(&ev); // makes a copy.
        }
        db.previousClickTime  = nowTime;
        db.previousClickState = newState;
        db.lastState          = newState;
    }
    void clearDoubleClick() {
        db.triggered = false;
    }
#endif
};
ImpulseAccumulator::ImpulseAccumulator(int impulseId, AccumulatorType type, bool expireBeforeSharpTick)
    : d(new Impl) {
    d->impulseId             = impulseId;
    d->type                  = type;
    d->expireBeforeSharpTick = expireBeforeSharpTick;
}
void ImpulseAccumulator::setPlayerNum(int newPlayerNum) {
    d->playerNum = newPlayerNum;
}
int ImpulseAccumulator::impulseId() const {
    return d->impulseId;
}
ImpulseAccumulator::AccumulatorType ImpulseAccumulator::type() const {
    return d->type;
}
bool ImpulseAccumulator::expireBeforeSharpTick() const {
    return d->expireBeforeSharpTick;
}
void ImpulseAccumulator::receiveBinary() {
    DENG2_ASSERT(d->type == Binary);
    LOG_AS("ImpulseAccumulator");
    d->binaryAccum++;
#ifdef __CLIENT__
    d->maintainDoubleClick(1);
    d->maintainDoubleClick(0);
#endif
}
int ImpulseAccumulator::takeBinary() {
    DENG2_ASSERT(d->type == Binary);
    LOG_AS("ImpulseAccumulator");
    short *counter = &d->binaryAccum;
    int count = *counter;
    *counter = 0;
    return count;
}
#ifdef __CLIENT__
void ImpulseAccumulator::takeAnalog(float *pos, float *relOffset) {
    DENG2_ASSERT(d->type == Analog);
    LOG_AS("ImpulseAccumulator");
    if(pos) *pos = 0;
    if(relOffset) *relOffset = 0;
    if(BindContext *bindContext = inputSys().contextPtr(d->getImpulse().bindContextName)) {
        float position, relative;
        B_EvaluateImpulseBindings(bindContext, P_ConsoleToLocal(d->playerNum), d->impulseId,
                                  &position, &relative, d->expireBeforeSharpTick);
        d->maintainDoubleClick(position);
        if(pos) *pos = position;
        if(relOffset) *relOffset = relative;
    }
}
void ImpulseAccumulator::clearAll() {
    LOG_AS("ImpulseAccumulator");
    switch(d->type) {
    case Analog:
        if(!d->expireBeforeSharpTick) {
            takeAnalog();
        }
        break;
    case Binary:
        takeBinary();
        break;
    default: DENG2_ASSERT(!"ImpulseAccumulator::clearAll: Unknown type");
    }
    d->clearDoubleClick();
}
void ImpulseAccumulator::consoleRegister() // static
{
    LOG_AS("ImpulseAccumulator");
    C_VAR_INT("input-doubleclick-threshold", &pimpDoubleClickThreshold, 0, 0, 2000);
}
#endif // __CLIENT__
struct ListNode {
    ListNode *next;
    ListNode *prev;
    intercepttype_t type;
    void *object;
    dfloat distance;
    template <class ObjectType>
    ObjectType &objectAs() const {
        DENG2_ASSERT(object);
        return *static_cast<ObjectType *>(object);
    }
};
static zblockset_t *interceptNodeSet;
static ListNode *interceptFirst;
static ListNode head;
static ListNode tail;
static ListNode *mru;
DENG2_PIMPL_NOREF(Interceptor) {
    traverser_t callback;
    void *context;
    Vector2d from;
    Vector2d to;
    dint flags; ///< @ref pathTraverseFlags
    world::Map *map = nullptr;
    LineOpening opening;
    vec2d_t fromV1;
    vec2d_t directionV1;
    Impl(traverser_t callback, Vector2d const &from, Vector2d const &to,
             dint flags, void *context)
        : callback(callback)
        , context (context)
        , from    (from)
        , to      (to)
        , flags   (flags) {
        V2d_Set(fromV1, from.x, from.y);
        V2d_Set(directionV1, to.x - from.x, to.y - from.y);
    }
    inline bool isSentinel(ListNode const &node) {
        return &node == &tail || &node == &head;
    }
    /**
     * Empties the intercepts array and makes sure it has been allocated.
     */
    void clearIntercepts() {
#define MININTERCEPTS       128
        if(!interceptNodeSet) {
            interceptNodeSet = ZBlockSet_New(sizeof(ListNode), MININTERCEPTS, PU_APPSTATIC);
            head.distance = 0.0f;
            head.next     = &tail;
            head.prev     = nullptr;
            tail.distance = 1.0f;
            tail.prev     = &head;
            tail.next     = nullptr;
        }
        if(!interceptFirst) {
            interceptFirst = head.next;
        }
        else if(head.next != &tail) {
            ListNode *existing = interceptFirst;
            interceptFirst = head.next;
            tail.prev->next = existing;
        }
        head.next = &tail;
        tail.prev = &head;
        mru = nullptr;
#undef MININTERCEPTS
    }
    /**
     * You must clear intercepts before the first time this is called.
     * The intercepts array grows if necessary.
     *
     * @param type      Type of interception.
     * @param distance  Distance along the trace vector that the interception occured [0...1].
     * @param object    Object being intercepted.
     */
    void addIntercept(intercepttype_t type, dfloat distance, void *object) {
        DENG2_ASSERT(object);
        if(distance < head.distance) return;
        if(distance > tail.distance) return;
        ListNode *before;
        if(mru && mru->distance <= distance) {
            before = mru->next;
        }
        else {
            before = head.next;
        }
        while(before->next && distance >= before->distance) {
            before = before->next;
        }
        ListNode *icpt;
        if(!isSentinel(*interceptFirst)) {
            icpt = interceptFirst;
            interceptFirst = icpt->next;
        }
        else {
            icpt = (ListNode *) ZBlockSet_Allocate(interceptNodeSet);
        }
        icpt->type     = type;
        icpt->object   = object;
        icpt->distance = distance;
        icpt->next = before;
        icpt->prev = before->prev;
        icpt->prev->next = icpt;
        icpt->next->prev = icpt;
        mru = icpt;
    }
    void intercept(Line &line) {
        fixed_t origin[2]    = { DBL2FIX(from.x), DBL2FIX(from.y) };
        fixed_t direction[2] = { DBL2FIX(to.x - from.x), DBL2FIX(to.y - from.y) };
        fixed_t lineFromX[2] = { DBL2FIX(line.from().x()), DBL2FIX(line.from().y()) };
        fixed_t lineToX[2]   = { DBL2FIX(line.to  ().x()), DBL2FIX(line.to  ().y()) };
        dint s1, s2;
        if (   direction[0] >  FRACUNIT * 16 || direction[1] >  FRACUNIT * 16
            || direction[0] < -FRACUNIT * 16 || direction[1] < -FRACUNIT * 16) {
            s1 = V2x_PointOnLineSide(lineFromX, origin, direction);
            s2 = V2x_PointOnLineSide(lineToX,   origin, direction);
        }
        else {
            s1 = line.pointOnSide(Vector2d(FIX2FLT(origin[0]), FIX2FLT(origin[1]))) < 0;
            s2 = line.pointOnSide(Vector2d(FIX2FLT(origin[0] + direction[0]),
                                           FIX2FLT(origin[1] + direction[1]))) < 0;
        }
        if (s1 == s2) return;
        fixed_t lineDirectionX[2] = { DBL2FIX(line.direction().x), DBL2FIX(line.direction().y) };
        dfloat distance = FIX2FLT(V2x_Intersection(lineFromX, lineDirectionX, origin, direction));
        if (distance >= 0) {
            addIntercept(ICPT_LINE, distance, &line);
        }
    }
    void intercept(mobj_t &mob) {
        if (mob.dPlayer && (mob.dPlayer->flags & DDPF_CAMERA))
            return;
        fixed_t origin[2]    = { DBL2FIX(from.x), DBL2FIX(from.y) };
        fixed_t direction[2] = { DBL2FIX(to.x - from.x), DBL2FIX(to.y - from.y) };
        AABoxd const &box = Mobj_Bounds(mob);
        fixed_t icptFrom[2], icptTo[2];
        if ((direction[0] ^ direction[1]) > 0) {
            V2x_Set(icptFrom, DBL2FIX(box.minX), DBL2FIX(box.maxY));
            V2x_Set(icptTo,   DBL2FIX(box.maxX), DBL2FIX(box.minY));
        }
        else {
            V2x_Set(icptFrom, DBL2FIX(box.minX), DBL2FIX(box.minY));
            V2x_Set(icptTo,   DBL2FIX(box.maxX), DBL2FIX(box.maxY));
        }
        if (   V2x_PointOnLineSide(icptFrom, origin, direction)
            == V2x_PointOnLineSide(icptTo,   origin, direction))
            return;
        fixed_t icptDirection[2] = { icptTo[0] - icptFrom[0], icptTo[1] - icptFrom[1] };
        dfloat distance = FIX2FLT(V2x_Intersection(icptFrom, icptDirection, origin, direction));
        if (distance >= 0) {
            addIntercept(ICPT_MOBJ, distance, &mob);
        }
    }
    void runTrace() {
        clearIntercepts();
        dint const localValidCount = ++validCount;
        if(flags & PTF_LINE) {
            if(map->polyobjCount()) {
                map->polyobjBlockmap().forAllInPath(from, to, [this, &localValidCount] (void *object) {
                    auto &pob = *(Polyobj *)object;
                    if(pob.validCount != localValidCount)  // not yet processed
                    {
                        pob.validCount = localValidCount;
                        for(Line *line : pob.lines()) {
                            if(line->validCount() != localValidCount)  // not yet processed
                            {
                                line->setValidCount(localValidCount);
                                intercept(*line);
                            }
                        }
                    }
                    return LoopContinue;
                });
            }
            map->lineBlockmap().forAllInPath(from, to, [this, &localValidCount] (void *object) {
                auto &line = *(Line *)object;
                if(line.validCount() != localValidCount)  // not yet processed
                {
                    line.setValidCount(localValidCount);
                    intercept(line);
                }
                return LoopContinue;
            });
        }
        if(flags & PTF_MOBJ) {
            map->mobjBlockmap().forAllInPath(from, to, [this, &localValidCount] (void *object) {
                auto &mob = *(mobj_t *)object;
                if(mob.validCount != localValidCount)  // not yet processed
                {
                    mob.validCount = localValidCount;
                    intercept(mob);
                }
                return LoopContinue;
            });
        }
    }
};
Interceptor::Interceptor(traverser_t callback, Vector2d const &from,
    Vector2d const &to, dint flags, void *context)
    : d(new Impl(callback, from, to, flags, context))
{}
ddouble const *Interceptor::origin() const {
    return d->fromV1;
}
ddouble const *Interceptor::direction() const {
    return d->directionV1;
}
LineOpening const &Interceptor::opening() const {
    return d->opening;
}
bool Interceptor::adjustOpening(Line const *line) {
    DENG2_ASSERT(d->map != 0);
    if(line) {
        if(d->map == &line->map()) {
            d->opening = LineOpening(*line);
        }
        else {
            qDebug() << "Ignoring alien line" << de::dintptr(line) << "in Interceptor::adjustOpening";
        }
    }
    return d->opening.range > 0;
}
dint Interceptor::trace(world::Map const &map) {
    d->map = const_cast<world::Map *>(&map);
    d->runTrace();
    for(ListNode *node = head.next; !d->isSentinel(*node); node = node->next) {
        Intercept icpt;
        icpt.trace    = this;
        icpt.distance = node->distance;
        icpt.type     = node->type;
        switch(node->type) {
        case ICPT_MOBJ: icpt.mobj = &node->objectAs<mobj_t>(); break;
        case ICPT_LINE: icpt.line = &node->objectAs<Line>();   break;
        }
        if(int result = d->callback(&icpt, d->context))
            return result;
    }
    return false; // Intercept traversal completed wholly.
}
DENG2_PIMPL_NOREF(Line::Side::Segment) {
    HEdge *hedge = nullptr;      ///< Half-edge attributed to the line segment (not owned).
#ifdef __CLIENT__
    ddouble length = 0;          ///< Accurate length of the segment.
    ddouble lineSideOffset = 0;  ///< Distance along the attributed map line at which the half-edge vertex occurs.
    bool frontFacing = false;
#endif
};
Line::Side::Segment::Segment(Line::Side &lineSide, HEdge &hedge)
    : MapElement(DMU_SEGMENT, &lineSide)
    , d(new Impl) {
    d->hedge = &hedge;
}
Line::Side &Line::Side::Segment::lineSide() {
    return parent().as<Side>();
}
Line::Side const &Line::Side::Segment::lineSide() const {
    return parent().as<Side>();
}
HEdge &Line::Side::Segment::hedge() const {
    DENG2_ASSERT(d->hedge);
    return *d->hedge;
}
#ifdef __CLIENT__
ddouble Line::Side::Segment::lineSideOffset() const {
    return d->lineSideOffset;
}
void Line::Side::Segment::setLineSideOffset(ddouble newOffset) {
    d->lineSideOffset = newOffset;
}
ddouble Line::Side::Segment::length() const {
    return d->length;
}
void Line::Side::Segment::setLength(ddouble newLength) {
    d->length = newLength;
}
bool Line::Side::Segment::isFrontFacing() const {
    return d->frontFacing;
}
void Line::Side::Segment::setFrontFacing(bool yes) {
    d->frontFacing = yes;
}
#endif  // __CLIENT__
DENG2_PIMPL(Line::Side) {
    /**
     * Line side section of which there are three (middle, bottom and top).
     */
    struct Section {
        Surface surface;
        ThinkerT<SoundEmitter> soundEmitter;
        Section(Line::Side &side) : surface(side) {}
    };
    struct Sections {
        Section *sections[3];
        inline Section &middle() { return *sections[Middle]; }
        inline Section &bottom() { return *sections[Bottom]; }
        inline Section &top()    { return *sections[Top];    }
        Sections(Side &side)
            : sections { new Section(side), new Section(side), new Section(side) }
        {}
        ~Sections() {
            for (auto *s : sections) delete s;
        }
    };
#ifdef __CLIENT__
    /**
     * POD: FakeRadio geometry and shadow state.
     */
    struct RadioData {
        std::array<edgespan_t, 2> spans;              ///< { bottom, top }
        std::array<shadowcorner_t, 2> topCorners;     ///< { left, right }
        std::array<shadowcorner_t, 2> bottomCorners;  ///< { left, right }
        std::array<shadowcorner_t, 2> sideCorners;    ///< { left, right }
        de::dint updateFrame = 0;
    };
#endif
    dint flags = 0;                 ///< @ref sdefFlags
    QList<Segment *> segments;      ///< On "this" side, sorted. Owned.
    bool needSortSegments = false;  ///< set to @c true when the list needs sorting.
    dint shadowVisCount = 0;        ///< Framecount of last time shadows were drawn.
    std::unique_ptr<Sections> sections;
#ifdef __CLIENT__
    RadioData radioData;
#endif
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        qDeleteAll(segments);
    }
    void clearSegments() {
        qDeleteAll(segments);
        segments.clear();
        needSortSegments = false;  // An empty list is sorted.
    }
    /**
     * Retrieve the Section associated with @a sectionId.
     */
    inline Section &sectionById(dint sectionId) {
        /*
        if (sections) {
            switch (sectionId) {
            case Middle: return sections->middle;
            case Bottom: return sections->bottom;
            case Top:    return sections->top;
            }
        }
        throw Line::InvalidSectionIdError("Line::Side::section", "Invalid section id " + String::number(sectionId));*/
        DENG2_ASSERT(sectionId >= Middle && sectionId <= Top);
        return *sections->sections[sectionId];
    }
    void sortSegments(Vector2d lineSideOrigin) {
        needSortSegments = false;
        if (segments.count() < 2)
            return;
        QMap<ddouble, Segment *> sortedSegs;
        for (Segment *seg : segments) {
            sortedSegs.insert((seg->hedge().origin() - lineSideOrigin).length(), seg);
        }
        segments = sortedSegs.values();
    }
#ifdef __CLIENT__
    void updateRadioCorner(shadowcorner_t &sc, dfloat openness, Plane *proximityPlane = nullptr, bool top = false) {
        DENG2_ASSERT(self()._sector);
        sc.corner    = openness;
        sc.proximity = proximityPlane;
        if (sc.proximity) {
            sc.pHeight = sc.proximity->heightSmoothed();
            sc.pOffset = sc.pHeight - self()._sector->plane(top? Sector::Ceiling : Sector::Floor).heightSmoothed();
        }
        else {
            sc.pOffset = 0;
            sc.pHeight = 0;
        }
    }
    /**
     * Change the FakeRadio side corner properties.
     */
    inline void setRadioCornerTop(bool right, dfloat openness, Plane *proximityPlane = nullptr) {
        updateRadioCorner(radioData.topCorners[dint(right)], openness, proximityPlane, true/*top*/);
    }
    inline void setRadioCornerBottom(bool right, dfloat openness, Plane *proximityPlane = nullptr) {
        updateRadioCorner(radioData.bottomCorners[dint(right)], openness, proximityPlane, false/*bottom*/);
    }
    inline void setRadioCornerSide(bool right, dfloat openness) {
        updateRadioCorner(radioData.sideCorners[dint(right)], openness);
    }
    /**
     * Change the FakeRadio "edge span" metrics.
     * @todo Replace shadow edge enumeration with a shadow corner enumeration. -ds
     */
    void setRadioEdgeSpan(bool top, bool right, ddouble length) {
        edgespan_t &span = radioData.spans[dint(top)];
        span.length = length;
        if (!right) {
            span.shift = span.length;
        }
    }
#endif // __CLIENT__
};
Line::Side::Side(Line &line, Sector *sector)
    : MapElement(DMU_SIDE, &line)
    , d(new Impl(this))
    , _sector(sector)
{}
Line &Line::Side::line() {
    return parent().as<Line>();
}
Line const &Line::Side::line() const {
    return parent().as<Line>();
}
bool Line::Side::isFront() const {
    return sideId() == Line::Front;
}
String Line::Side::description() const {
    QStringList flagNames;
    if (flags() & SDF_BLENDTOPTOMID)    flagNames << "blendtoptomiddle";
    if (flags() & SDF_BLENDMIDTOTOP)    flagNames << "blendmiddletotop";
    if (flags() & SDF_BLENDMIDTOBOTTOM) flagNames << "blendmiddletobottom";
    if (flags() & SDF_BLENDBOTTOMTOMID) flagNames << "blendbottomtomiddle";
    if (flags() & SDF_MIDDLE_STRETCH)   flagNames << "middlestretch";
    String flagsString;
    if (!flagNames.isEmpty()) {
        String const flagsAsText = flagNames.join("|");
        flagsString = String(_E(l) " Flags: " _E(.)_E(i) "%1" _E(.)).arg(flagsAsText);
    }
    auto text = String(_E(D)_E(b) "%1:\n"  _E(.)_E(.)
                       _E(l)  "Sector: "    _E(.)_E(i) "%2" _E(.)
                       _E(l) " One Sided: " _E(.)_E(i) "%3" _E(.)
                       "%4")
                    .arg(Line::sideIdAsText(sideId()).upperFirstChar())
                    .arg(hasSector() ? String::number(sector().indexInMap()) : "None")
                    .arg(DENG2_BOOL_YESNO(considerOneSided()))
                    .arg(flagsString);
    forAllSurfaces([this, &text] (Surface &suf) {
        text += String("\n" _E(D) "%1:\n" _E(.))
                  .arg(sectionIdAsText(  &suf == &top   () ? Side::Top
                                       : &suf == &middle() ? Side::Middle
                                       :                     Side::Bottom))
              + suf.description();
        return LoopContinue;
    });
    return text;
}
dint Line::Side::sideId() const {
    return &line().front() == this? Line::Front : Line::Back;
}
bool Line::Side::considerOneSided() const {
    if (d->flags & SDF_SUPPRESS_BACK_SECTOR) return true;
    if (!back().hasSector()) return true;
    if (!back().hasSections()) return true;
    if (!line().definesPolyobj()) {
        HEdge *hedge = leftHEdge();
        if (!hedge || !hedge->twin().hasFace())
            return true;
        if (!hedge->twin().face().mapElementAs<ConvexSubspace>().hasSubsector())
            return true;
    }
    return false;
}
/*
bool Line::Side::hasSector() const {
    return d->sector != nullptr;
}
Sector &Line::Side::sector() const {
    if (d->sector) return *d->sector;
    throw MissingSectorError("Line::Side::sector", "No sector is attributed");
}
Sector *Line::Side::sectorPtr() const {
    return d->sector;
}
*/
bool Line::Side::hasSections() const {
    return bool(d->sections);
}
void Line::Side::addSections() {
    if (hasSections()) return;
    d->sections.reset(new Impl::Sections(*this));
}
Surface &Line::Side::surface(dint sectionId) {
    return d->sectionById(sectionId).surface;
}
Surface const &Line::Side::surface(dint sectionId) const {
    return d->sectionById(sectionId).surface;
}
Surface &Line::Side::middle() {
    return surface(Middle);
}
Surface const &Line::Side::middle() const {
    return surface(Middle);
}
Surface &Line::Side::bottom() {
    return surface(Bottom);
}
Surface const &Line::Side::bottom() const {
    return surface(Bottom);
}
Surface &Line::Side::top() {
    return surface(Top);
}
Surface const &Line::Side::top() const {
    return surface(Top);
}
LoopResult Line::Side::forAllSurfaces(std::function<LoopResult(Surface &)> func) const {
    if (hasSections()) {
        for (dint i = Middle; i <= Top; ++i) {
            if (auto result = func(const_cast<Side *>(this)->surface(i)))
                return result;
        }
    }
    return LoopContinue;
}
SoundEmitter &Line::Side::soundEmitter(dint sectionId) {
    return d->sectionById(sectionId).soundEmitter;
}
SoundEmitter const &Line::Side::soundEmitter(dint sectionId) const {
    return const_cast<Side *>(this)->soundEmitter(sectionId);
}
SoundEmitter &Line::Side::middleSoundEmitter() {
    return soundEmitter(Middle);
}
SoundEmitter const &Line::Side::middleSoundEmitter() const {
    return soundEmitter(Middle);
}
SoundEmitter &Line::Side::bottomSoundEmitter() {
    return soundEmitter(Bottom);
}
SoundEmitter const &Line::Side::bottomSoundEmitter() const {
    return soundEmitter(Bottom);
}
SoundEmitter &Line::Side::topSoundEmitter() {
    return soundEmitter(Top);
}
SoundEmitter const &Line::Side::topSoundEmitter() const {
    return soundEmitter(Top);
}
void Line::Side::clearSegments() {
    d->clearSegments();
}
Line::Side::Segment *Line::Side::addSegment(HEdge &hedge) {
    for (Segment *seg : d->segments) {
        if (&seg->hedge() == &hedge)
            return seg;
    }
    auto *newSeg = new Segment(*this, hedge);
    d->segments.append(newSeg);
    d->needSortSegments = true;  // We'll need to (re)sort.
    hedge.setMapElement(newSeg);
    return newSeg;
}
HEdge *Line::Side::leftHEdge() const {
    if(d->segments.isEmpty()) return nullptr;
    if (d->needSortSegments) {
        d->sortSegments(from().origin());
    }
    return &d->segments.first()->hedge();
}
HEdge *Line::Side::rightHEdge() const {
    if (d->segments.isEmpty()) return nullptr;
    if (d->needSortSegments) {
        d->sortSegments(from().origin());
    }
    return &d->segments.last()->hedge();
}
void Line::Side::updateSoundEmitterOrigin(dint sectionId) {
    LOG_AS("Line::Side::updateSoundEmitterOrigin");
    if (!hasSections()) return;
    SoundEmitter &emitter = d->sectionById(sectionId).soundEmitter;
    Vector2d lineCenter = line().center();
    emitter.origin[0] = lineCenter.x;
    emitter.origin[1] = lineCenter.y;
    DENG2_ASSERT(_sector);
    ddouble const ffloor = _sector->floor().height();
    ddouble const fceil  = _sector->ceiling().height();
    switch (sectionId) {
    case Middle:
        if (!back().hasSections() || line().isSelfReferencing()) {
            emitter.origin[2] = (ffloor + fceil) / 2;
        }
        else {
            emitter.origin[2] = (de::max(ffloor, back().sector().floor().height()) +
                                 de::min(fceil,  back().sector().ceiling().height())) / 2;
        }
        break;
    case Bottom:
        if (!back().hasSections() || line().isSelfReferencing()
           || back().sector().floor().height() <= ffloor) {
            emitter.origin[2] = ffloor;
        }
        else {
            emitter.origin[2] = (de::min(back().sector().floor().height(), fceil) + ffloor) / 2;
        }
        break;
    case Top:
        if (!back().hasSections() || line().isSelfReferencing()
            || back().sector().ceiling().height() >= fceil) {
            emitter.origin[2] = fceil;
        }
        else {
            emitter.origin[2] = (de::max(back().sector().ceiling().height(), ffloor) + fceil) / 2;
        }
        break;
    }
}
void Line::Side::updateAllSoundEmitterOrigins() {
    if (!hasSections()) return;
    updateMiddleSoundEmitterOrigin();
    updateBottomSoundEmitterOrigin();
    updateTopSoundEmitterOrigin();
}
void Line::Side::updateAllSurfaceNormals() {
    if (!hasSections()) return;
    Vector3f normal((  to().origin().y - from().origin().y) / line().length(),
                    (from().origin().x -   to().origin().x) / line().length(),
                    0);
    middle().setNormal(normal); // will normalize
    bottom().setNormal(normal);
    top   ().setNormal(normal);
}
dint Line::Side::flags() const {
    return d->flags;
}
void Line::Side::setFlags(dint flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
}
void Line::Side::chooseSurfaceColors(dint sectionId, Vector3f const **topColor,
    Vector3f const **bottomColor) const {
    if (hasSections()) {
        switch (sectionId) {
        case Middle:
            if (isFlagged(SDF_BLENDMIDTOTOP)) {
                *topColor    = &top   ().color();
                *bottomColor = &middle().color();
            }
            else if (isFlagged(SDF_BLENDMIDTOBOTTOM)) {
                *topColor    = &middle().color();
                *bottomColor = &bottom().color();
            }
            else {
                *topColor    = &middle().color();
                *bottomColor = nullptr;
            }
            return;
        case Top:
            if (isFlagged(SDF_BLENDTOPTOMID)) {
                *topColor    = &top   ().color();
                *bottomColor = &middle().color();
            }
            else {
                *topColor    = &top   ().color();
                *bottomColor = nullptr;
            }
            return;
        case Bottom:
            if (isFlagged(SDF_BLENDBOTTOMTOMID)) {
                *topColor    = &middle().color();
                *bottomColor = &bottom().color();
            }
            else {
                *topColor    = &bottom().color();
                *bottomColor = nullptr;
            }
            return;
        }
    }
    throw InvalidSectionIdError("Line::Side::chooseSurfaceColors", "Invalid section id " + String::number(sectionId));
}
bool Line::Side::hasAtLeastOneMaterial() const {
    return middle().hasMaterial() || top().hasMaterial() || bottom().hasMaterial();
}
dint Line::Side::shadowVisCount() const {
    return d->shadowVisCount;
}
void Line::Side::setShadowVisCount(dint newCount) {
    d->shadowVisCount = newCount;
}
#ifdef __CLIENT__
shadowcorner_t const &Line::Side::radioCornerTop(bool right) const {
    return d->radioData.topCorners[dint(right)];
}
shadowcorner_t const &Line::Side::radioCornerBottom(bool right) const {
    return d->radioData.bottomCorners[dint(right)];
}
shadowcorner_t const &Line::Side::radioCornerSide(bool right) const {
    return d->radioData.sideCorners[dint(right)];
}
edgespan_t const &Line::Side::radioEdgeSpan(bool top) const {
    return d->radioData.spans[dint(top)];
}
/**
 * Convert a corner @a angle to a "FakeRadio corner openness" factor.
 */
static dfloat radioCornerOpenness(binangle_t angle) {
    if (angle > BANG_180) return -1;
    if (angle == BANG_180) return 0;
    if (angle < BANG_45 / 5) return 0;
    return (angle > BANG_90)? dfloat( BANG_90 ) / angle : dfloat( angle ) / BANG_90;
}
static inline binangle_t lineNeighborAngle(LineSide const &side, Line const *other, binangle_t diff) {
    return (other && other != &side.line()) ? diff : 0 /*Consider it coaligned*/;
}
static binangle_t findSolidLineNeighborAngle(LineSide const &side, bool right) {
    binangle_t diff = 0;
    Line const *other = R_FindSolidLineNeighbor(side.line(),
                                                *side.line().vertexOwner(dint(right) ^ side.sideId()),
                                                right ? Anticlockwise : Clockwise, side.sectorPtr(), &diff);
    return lineNeighborAngle(side, other, diff);
}
/**
 * Returns @c true if there is open space in the sector.
 */
static inline bool sectorIsOpen(Sector const *sector) {
    return (sector && sector->ceiling().height() > sector->floor().height());
}
struct edge_t {
    Line *line;
    Sector *sector;
    dfloat length;
    binangle_t diff;
};
static void scanNeighbor(LineSide const &side, bool top, bool right, edge_t &edge) {
    static dint const SEP = 10;
    de::zap(edge);
    ClockDirection const direction = (right ? Anticlockwise : Clockwise);
    Sector const *startSector = side.sectorPtr();
    ddouble const fFloor      = side.sector().floor  ().heightSmoothed();
    ddouble const fCeil       = side.sector().ceiling().heightSmoothed();
    ddouble gap    = 0;
    LineOwner *own = side.line().vertexOwner(side.vertex(dint(right)));
    forever {
        binangle_t diff  = (direction == Clockwise ? own->angle() : own->prev()->angle());
        Line const *iter = &own->navigate(direction)->line();
        dint scanSecSide = (iter->front().hasSector() && iter->front().sectorPtr() == startSector ? Line::Back : Line::Front);
        while ((!iter->front().hasSector() && !iter->back().hasSector()) || iter->isSelfReferencing()) {
            own         = own->navigate(direction);
            diff       += (direction == Clockwise ? own->angle() : own->prev()->angle());
            iter        = &own->navigate(direction)->line();
            scanSecSide = (iter->front().sectorPtr() == startSector);
        }
        LineSide const &scanSide = iter->side(scanSecSide);
        Sector const *scanSector = scanSide.sectorPtr();
        ddouble const iFFloor = iter->front().sector().floor  ().heightSmoothed();
        ddouble const iFCeil  = iter->front().sector().ceiling().heightSmoothed();
        Sector const *bsec    = iter->back().sectorPtr();
        ddouble const iBFloor = (bsec ? bsec->floor  ().heightSmoothed() : 0);
        ddouble const iBCeil  = (bsec ? bsec->ceiling().heightSmoothed() : 0);
        bool closed = false;
        if (side.isFront() && iter->back().hasSector()) {
            closed = top? (iBFloor >= fCeil) : (iBCeil <= fFloor);  // Compared to "this" sector anyway.
        }
        edge.line   = const_cast<Line *>(iter);
        edge.diff   = diff;
        edge.sector = scanSide.sectorPtr();
        ddouble lengthDelta = 0;
        if (top) {
            if (iter->back().hasSector()
                && (   (side.isFront() && iter->back().sectorPtr() == side.line().front().sectorPtr() && iFCeil >= fCeil)
                    || (side.isBack () && iter->back().sectorPtr() == side.line().back().sectorPtr () && iFCeil >= fCeil)
                    || (side.isFront() && closed == false && iter->back().sectorPtr() != side.line().front().sectorPtr()
                        && iBCeil >= fCeil && sectorIsOpen(iter->back().sectorPtr())))) {
                gap += iter->length();  // Should we just mark it done instead?
            }
            else {
                edge.length += iter->length() + gap;
                gap = 0;
            }
        }
        else {
            if (iter->back().hasSector()
                && (   (side.isFront() && iter->back().sectorPtr() == side.line().front().sectorPtr() && iFFloor <= fFloor)
                    || (side.isBack () && iter->back().sectorPtr() == side.line().back().sectorPtr () && iFFloor <= fFloor)
                    || (side.isFront() && closed == false && iter->back().sectorPtr() != side.line().front().sectorPtr()
                        && iBFloor <= fFloor && sectorIsOpen(iter->back().sectorPtr())))) {
                gap += iter->length();  // Should we just mark it done instead?
            }
            else {
                lengthDelta = iter->length() + gap;
                gap = 0;
            }
        }
        if (iter == &side.line()) break;
        if (!(diff >= BANG_180 - SEP && diff <= BANG_180 + SEP)) break;
        if (scanSector) {
            if (!sectorIsOpen(scanSector)) break;
            if (top) {
                if (scanSector->ceiling().heightSmoothed() != fCeil
                    && scanSector->floor().heightSmoothed() < startSector->ceiling().heightSmoothed()) {
                    break;
                }
            }
            else {
                if (scanSector->floor().heightSmoothed() != fFloor
                    && scanSector->ceiling().heightSmoothed() > startSector->floor().heightSmoothed()) {
                    break;
                }
            }
        }
        if (own->navigate(direction) == iter->v2Owner()) {
            own = iter->v1Owner();
        }
        else if (own->navigate(direction) == iter->v1Owner()) {
            own = iter->v2Owner();
        }
        if (scanSector && side.back().hasSector() && scanSector != side.back().sectorPtr()
            && (   ( top && scanSector->ceiling().heightSmoothed() == startSector->ceiling().heightSmoothed())
                || (!top && scanSector->floor  ().heightSmoothed() == startSector->floor  ().heightSmoothed()))) {
            Line *backNeighbor = R_FindLineNeighbor(*iter, *own, direction, startSector);
            if (backNeighbor && backNeighbor != iter) {
                own = own->navigate(direction);
                startSector = scanSector;
            }
        }
        edge.length += lengthDelta;
    }
    if (edge.sector)  // The back sector of the coalignable neighbor.
    {
        DENG2_ASSERT(edge.line);
        edge.line = R_FindLineNeighbor(*edge.line,
                                       *edge.line->vertexOwner(dint(edge.line->back().hasSector() && edge.line->back().sectorPtr() == edge.sector) ^ dint(right)),
                                       direction, edge.sector, &edge.diff);
    }
}
/**
 * To determine the dimensions of a shadow, we'll need to scan edges. Edges are composed
 * of aligned lines. It's important to note that the scanning is done separately for the
 * top/bottom edges (both in the left and right direction) and the left/right edges.
 *
 * The length of the top/bottom edges are returned in the array 'spans'.
 *
 * This may look like a complicated operation (performed for all line sides) but in most
 * cases this won't take long. Aligned neighbours are relatively rare.
 *
 * @todo fixme: Should use the visual plane heights of subsectors.
 */
void Line::Side::updateRadioForFrame(dint frameNumber) {
    if (!::rendFakeRadio || ::levelFullBright) return;
    if (!::devFakeRadioUpdate) return;
    if (!hasSector()) return;
    if (line().isSelfReferencing()) return;
    if (d->radioData.updateFrame == frameNumber) return;
    d->radioData.updateFrame = frameNumber;  // Mark as done.
    d->setRadioCornerSide(false/*left*/, radioCornerOpenness(findSolidLineNeighborAngle(*this, false/*left*/)));
    d->setRadioCornerSide(true/*right*/, radioCornerOpenness(findSolidLineNeighborAngle(*this, true/*right*/)));
    for (dint i = 0; i < 2; ++i) {
        bool const rightEdge = i != 0;
        edge_t bottom; scanNeighbor(*this, false/*bottom*/, rightEdge, bottom);
        edge_t top;    scanNeighbor(*this, true/*top*/    , rightEdge, top   );
        d->setRadioEdgeSpan(false/*left*/, rightEdge, line().length() + bottom.length);
        d->setRadioEdgeSpan(true/*right*/, rightEdge, line().length() + top   .length);
        d->setRadioCornerBottom(rightEdge, radioCornerOpenness(lineNeighborAngle(*this, bottom.line, bottom.diff)),
                                bottom.sector? &bottom.sector->floor  () : nullptr);
        d->setRadioCornerTop   (rightEdge, radioCornerOpenness(lineNeighborAngle(*this, top   .line, top   .diff)),
                                top.sector   ? &top   .sector->ceiling() : nullptr);
    }
}
#endif  // __CLIENT__
dint Line::Side::property(DmuArgs &args) const {
    switch(args.prop) {
    case DMU_SECTOR:
        args.setValue(DMT_SIDE_SECTOR, &_sector, 0);
        break;
    case DMU_LINE: {
        Line const *lineAdr = &line();
        args.setValue(DMT_SIDE_LINE, &lineAdr, 0);
        break; }
    case DMU_FLAGS:
        args.setValue(DMT_SIDE_FLAGS, &d->flags, 0);
        break;
    case DMU_EMITTER:
        args.setValue(DMT_SIDE_EMITTER,
                      args.modifiers & DMU_TOP_OF_SIDE?    &soundEmitter(Top)    :
                      args.modifiers & DMU_MIDDLE_OF_SIDE? &soundEmitter(Middle) :
                                                           &soundEmitter(Bottom), 0);
        break;
    default:
        return MapElement::property(args);
    }
    return false; // Continue iteration.
}
dint Line::Side::setProperty(DmuArgs const &args) {
    switch(args.prop) {
    case DMU_SECTOR: {
        if (P_IsDummy(&line())) {
            args.value(DMT_SIDE_SECTOR, &_sector, 0);
        }
        else {
            throw WritePropertyError("Line::Side::setProperty", "Property " + String(DMU_Str(args.prop)) + " is only writable for dummy Line::Sides");
        }
        break; }
    case DMU_FLAGS: {
        dint newFlags;
        args.value(DMT_SIDE_FLAGS, &newFlags, 0);
        setFlags(newFlags, de::ReplaceFlags);
        break; }
    default:
        return MapElement::setProperty(args);
    }
    return false; // Continue iteration.
}
String Line::Side::sectionIdAsText(dint sectionId) // static
{
    switch (sectionId) {
    case Middle: return "middle";
    case Bottom: return "bottom";
    case Top:    return "top";
    default:     return "(invalid)";
    };
}
DENG2_PIMPL(Line)
, DENG2_OBSERVES(Vertex, OriginChange) {
    dint flags;                 ///< Public DDLF_* flags.
    Side front;                 ///< Front side of the line.
    Side back;                  ///< Back side of the line.
    std::array<bool, DDMAXPLAYERS> mapped {}; ///< Whether the line has been seen by each player yet.
    Vertex *from     = nullptr; ///< Start vertex (not owned).
    Vertex *to       = nullptr; ///< End vertex (not owned).
    Polyobj *polyobj = nullptr; ///< Polyobj the line defines a section of, if any (not owned).
    dint validCount  = 0;       ///< Used by legacy algorithms to prevent repeated processing.
    enum SelfReferencing { Unknown, IsSelfRef, IsNotSelfRef };
    SelfReferencing selfRef = Unknown;
    /**
     * POD: Additional metrics describing the geometry of the line (the vertices).
     */
    struct GeomData {
        Vector2d direction;     ///< From start to end vertex.
        ddouble length;         ///< Accurate length.
        binangle_t angle;       ///< Calculated from the direction vector.
        slopetype_t slopeType;  ///< Logical line slope (i.e., world angle) classification.
        AABoxd bounds;          ///< Axis-aligned bounding box.
        GeomData(Vertex const &from, Vertex const &to)
            : direction(to.origin() - from.origin())
            , length   (direction.length())
            , angle    (bamsAtan2(dint(direction.y), dint(direction.x)))
            , slopeType(M_SlopeType(direction.data().baseAs<ddouble>())) {
            V2d_InitBoxXY (bounds.arvec2, from.x(), from.y());
            V2d_AddToBoxXY(bounds.arvec2, to  .x(), to  .y());
        }
        static ddouble calcLength(Vertex const &from, Vertex const &to) {
            return (to.origin() - from.origin()).length();
        }
    };
    std::unique_ptr<GeomData> gdata;
    Impl(Public *i, Sector *frontSector, Sector *backSector)
        : Base (i)
        , front(*i, frontSector)
        , back (*i, backSector)
    {}
    /**
     * Returns the additional geometry metrics (cached).
     */
    GeomData &geom() {
        if (!gdata) {
            gdata.reset(new GeomData(self().from(), self().to()));
        }
        return *gdata;
    }
    void vertexOriginChanged(Vertex &vtx) {
        DENG2_ASSERT(&vtx == from || &vtx == to); // Should only observe changes to "our" vertices.
        DENG2_ASSERT(polyobj != nullptr);         // Should only observe changes to moveable (not editable) vertices.
        DENG2_UNUSED(vtx);
        gdata.release();
    }
};
Line::Line(Vertex &from, Vertex &to, dint flags, Sector *frontSector, Sector *backSector)
    : MapElement(DMU_LINE)
    , d(new Impl(this, frontSector, backSector)) {
    d->flags = flags;
    replaceVertex(From, from);
    replaceVertex(To  , to);
}
dint Line::flags() const {
    return d->flags;
}
void Line::setFlags(dint flagsToChange, FlagOp operation) {
    dint newFlags = d->flags;
    applyFlagOperation(newFlags, flagsToChange, operation);
    if (d->flags != newFlags) {
        dint oldFlags = d->flags;
        d->flags = newFlags;
        DENG2_FOR_AUDIENCE(FlagsChange, i) i->lineFlagsChanged(*this, oldFlags);
    }
}
bool Line::isBspWindow() const {
    return _bspWindowSector != nullptr;
}
bool Line::definesPolyobj() const {
    return d->polyobj != nullptr;
}
Polyobj &Line::polyobj() const {
    if (d->polyobj) return *d->polyobj;
    throw MissingPolyobjError("Line::polyobj", "No polyobj is attributed");
}
void Line::setPolyobj(Polyobj *newPolyobj) {
    if (d->polyobj == newPolyobj) return;
    if (d->polyobj) {
        to  ().audienceForOriginChange -= d;
        from().audienceForOriginChange -= d;
    }
    d->polyobj = newPolyobj;
    if (d->polyobj) {
        from().audienceForOriginChange += d;
        to  ().audienceForOriginChange += d;
    }
}
bool Line::isSelfReferencing() const {
    if (d->selfRef == Impl::Unknown) {
        d->selfRef = (front().hasSector() && front().sectorPtr() == back().sectorPtr()) ?
            Impl::IsSelfRef : Impl::IsNotSelfRef;
    }
    return d->selfRef == Impl::IsSelfRef;
}
Line::Side &Line::side(dint back) {
    return (back ? d->back : d->front);
}
Line::Side const &Line::side(dint back) const {
    return (back ? d->back : d->front);
}
LoopResult Line::forAllSides(std::function<LoopResult(Side &)> func) const {
    for (dint i = 0; i < 2; ++i) {
        if (auto result = func(const_cast<Side &>(side(i))))
            return result;
    }
    return LoopContinue;
}
void Line::replaceVertex(dint to, Vertex &newVertex) {
    Vertex **adr = (to ? &d->to : &d->from);
    if (*adr && *adr == &newVertex) return;
    *adr = &newVertex;
    d->gdata.release();
}
Vertex &Line::vertex(dint to) {
    DENG2_ASSERT((to ? d->to : d->from) != nullptr);
    return (to ? *d->to : *d->from);
}
Vertex const &Line::vertex(dint to) const {
    DENG2_ASSERT((to ? d->to : d->from) != nullptr);
    return (to ? *d->to : *d->from);
}
Vertex &Line::from() {
    return vertex(From);
}
Vertex const &Line::from() const {
    return vertex(From);
}
Vertex &Line::to() {
    return vertex(To);
}
Vertex const &Line::to() const {
    return vertex(To);
}
LoopResult Line::forAllVertexs(std::function<LoopResult(Vertex &)> func) const {
    for (dint i = 0; i < 2; ++i) {
        if (auto result = func(const_cast<Line *>(this)->vertex(i)))
            return result;
    }
    return LoopContinue;
}
AABoxd const &Line::bounds() const {
    return d->geom().bounds;
}
binangle_t Line::angle() const {
    return d->geom().angle;
}
Vector2d Line::center() const {
    return from().origin() + direction() / 2;
}
Vector2d const &Line::direction() const {
    return d->geom().direction;
}
ddouble Line::length() const {
    if (d->gdata) return d->gdata->length;
    return Impl::GeomData::calcLength(*d->from, *d->to);
}
slopetype_t Line::slopeType() const {
    return d->geom().slopeType;
}
dint Line::boxOnSide(AABoxd const &box) const {
    return M_BoxOnLineSide(&box, from().origin().data().baseAs<ddouble>(),
                           direction().data().baseAs<ddouble>());
}
int Line::boxOnSide_FixedPrecision(AABoxd const &box) const {
    ddouble offset[2] = { std::floor(from().x() + direction().x / 2.0),
                          std::floor(from().y() + direction().y / 2.0) };
    fixed_t boxx[4];
    boxx[BOXLEFT]   = DBL2FIX(box.minX - offset[0]);
    boxx[BOXRIGHT]  = DBL2FIX(box.maxX - offset[0]);
    boxx[BOXBOTTOM] = DBL2FIX(box.minY - offset[1]);
    boxx[BOXTOP]    = DBL2FIX(box.maxY - offset[1]);
    fixed_t pos[2] = { DBL2FIX(from().x() - offset[0]),
                       DBL2FIX(from().y() - offset[1]) };
    fixed_t delta[2] = { DBL2FIX(direction().x),
                         DBL2FIX(direction().y) };
    return M_BoxOnLineSide_FixedPrecision(boxx, pos, delta);
}
ddouble Line::pointDistance(Vector2d const &point, ddouble *offset) const {
    Vector2d lineVec = direction() - from().origin();
    ddouble len = lineVec.length();
    if (len == 0) {
        if (offset) *offset = 0;
        return 0;
    }
    Vector2d delta = from().origin() - point;
    if (offset) {
        *offset = (  delta.y * (from().y() - direction().y)
                   - delta.x * (direction().x - from().x()) )
                / len;
    }
    return (delta.y * lineVec.x - delta.x * lineVec.y) / len;
}
ddouble Line::pointOnSide(Vector2d const &point) const {
    Vector2d delta = from().origin() - point;
    return delta.y * direction().x - delta.x * direction().y;
}
bool Line::isMappedByPlayer(dint playerNum) const {
    return d->mapped[playerNum];
}
void Line::setMappedByPlayer(dint playerNum, bool yes) {
    d->mapped[playerNum] = yes;
}
dint Line::validCount() const {
    return d->validCount;
}
void Line::setValidCount(dint newValidCount) {
    d->validCount = newValidCount;
}
Line::Side &Line::front() {
    return side(Front);
}
Line::Side const &Line::front() const {
    return side(Front);
}
Line::Side &Line::back() {
    return side(Back);
}
Line::Side const &Line::back() const {
    return side(Back);
}
#ifdef __CLIENT__
bool Line::isShadowCaster() const {
    if (definesPolyobj()) return false;
    if (isSelfReferencing()) return false;
    if (&v1Owner()->next()->line() == this || &v2Owner()->next()->line() == this)
       return false;
    return true;
}
#endif
LineOwner *Line::vertexOwner(dint to) const {
    DENG2_ASSERT((to ? _vo2 : _vo1) != nullptr);
    return (to ? _vo2 : _vo1);
}
dint Line::property(DmuArgs &args) const {
    switch(args.prop) {
    case DMU_FLAGS:
        args.setValue(DMT_LINE_FLAGS, &d->flags, 0);
        break;
    case DMU_FRONT: {
        Line::Side const *frontAdr = front().hasSections() ? &d->front : nullptr;
        args.setValue(DDVT_PTR, &frontAdr, 0);
        break; }
    case DMU_BACK: {
        Line::Side const *backAdr  = back().hasSections() ? &d->back   : nullptr;
        args.setValue(DDVT_PTR, &backAdr, 0);
        break; }
    case DMU_VERTEX0:
        args.setValue(DMT_LINE_V, &d->from, 0);
        break;
    case DMU_VERTEX1:
        args.setValue(DMT_LINE_V, &d->to, 0);
        break;
    case DMU_VALID_COUNT:
        args.setValue(DMT_LINE_VALIDCOUNT, &d->validCount, 0);
        break;
    case DMU_DX:
        args.setValue(DMT_LINE_DX, &direction().x, 0);
        break;
    case DMU_DY:
        args.setValue(DMT_LINE_DY, &direction().y, 0);
        break;
    case DMU_DXY:
        args.setValue(DMT_LINE_DX, &direction().x, 0);
        args.setValue(DMT_LINE_DY, &direction().y, 1);
        break;
    case DMU_LENGTH: {
        ddouble len = length();
        args.setValue(DMT_LINE_LENGTH, &len, 0);
        break; }
    case DMU_ANGLE: {
        angle_t ang = BANG_TO_ANGLE(angle());
        args.setValue(DDVT_ANGLE, &ang, 0);
        break; }
    case DMU_SLOPETYPE: {
        slopetype_t st = slopeType();
        args.setValue(DMT_LINE_SLOPETYPE, &st, 0);
        break; }
    case DMU_BOUNDING_BOX: {
        AABoxd const *boxAdr = &bounds();
        args.setValue(DDVT_PTR, &boxAdr, 0);
        break; }
    default:
        return MapElement::property(args);
    }
    return false; // Continue iteration.
}
dint Line::setProperty(DmuArgs const &args) {
    switch(args.prop) {
    case DMU_VALID_COUNT:
        args.value(DMT_LINE_VALIDCOUNT, &d->validCount, 0);
        break;
    case DMU_FLAGS: {
        dint newFlags;
        args.value(DMT_LINE_FLAGS, &newFlags, 0);
        setFlags(newFlags, de::ReplaceFlags);
        break; }
    default:
        return MapElement::setProperty(args);
    }
    return false; // Continue iteration.
}
D_CMD(InspectLine) {
    DENG2_UNUSED(src);
    LOG_AS("inspectline (Cmd)");
    if (argc != 2) {
        LOG_SCR_NOTE("Usage: %s (line-id)") << argv[0];
        return true;
    }
    if (!App_World().hasMap()) {
        LOG_SCR_ERROR("No map is currently loaded");
        return false;
    }
    dint const index = String(argv[1]).toInt();
    Line const *line = App_World().map().linePtr(index);
    if (!line) {
        LOG_SCR_ERROR("Line #%i not found") << index;
        return false;
    }
    QStringList flagNames;
    if (line->flags() & DDLF_BLOCKING)      flagNames << "blocking";
    if (line->flags() & DDLF_DONTPEGTOP)    flagNames << "nopegtop";
    if (line->flags() & DDLF_DONTPEGBOTTOM) flagNames << "nopegbottom";
    String flagsString;
    if (!flagNames.isEmpty()) {
        String const flagsAsText = flagNames.join("|");
        flagsString = String(_E(l) " Flags: " _E(.)_E(i) "%1" _E(.)).arg(flagsAsText);
    }
    LOG_SCR_MSG(_E(b) "Line %i" _E(.) " [%p]")
            << line->indexInMap() << line;
    LOG_SCR_MSG(_E(l)  "From: " _E(.)_E(i) "%s" _E(.)
                _E(l) " To: "   _E(.)_E(i) "%s" _E(.)
                "%s")
            << line->from().origin().asText()
            << line->to  ().origin().asText()
            << flagsString;
    line->forAllSides([](Line::Side &side) {
        LOG_SCR_MSG("") << side.description();
        return LoopContinue;
    });
    return true;
}
void Line::consoleRegister() // static
{
    C_CMD("inspectline", "i", InspectLine);
}
String Line::sideIdAsText(dint sideId) // static
{
    switch (sideId) {
    case Front: return "front";
    case Back:  return "back";
    default:    return "(invalid)";
    };
}
namespace world {
LineBlockmap::LineBlockmap(AABoxd const &bounds, duint cellSize)
    : Blockmap(bounds, cellSize)
{}
void LineBlockmap::link(Line &line) {
    if(line.definesPolyobj()) return;
    BlockmapCellBlock const cellBlock = toCellBlock(line.bounds());
    BlockmapCell cell;
    for(cell.y = cellBlock.min.y; cell.y < cellBlock.max.y; ++cell.y)
    for(cell.x = cellBlock.min.x; cell.x < cellBlock.max.x; ++cell.x) {
        if(line.slopeType() == ST_VERTICAL || line.slopeType() == ST_HORIZONTAL) {
            Blockmap::link(cell, &line);
            continue;
        }
        Vector2d const point(origin() + cellDimensions() * cell);
        Vector2d from, to;
        if(line.slopeType() == ST_POSITIVE) {
            from = Vector2d(point.x, point.y + cellDimensions().y);
            to   = Vector2d(point.x + cellDimensions().x, point.y);
        }
        else {
            from = Vector2d(point.x + cellDimensions().x, point.y + cellDimensions().y);
            to   = Vector2d(point.x, point.y);
        }
        if((line.pointOnSide(from) < 0) != (line.pointOnSide(to) < 0)) {
            Blockmap::link(cell, &line);
        }
    }
}
void LineBlockmap::link(QList<Line *> const &lines) {
    for(Line *line : lines) link(*line);
}
}  // namespace world
namespace world {
DENG2_PIMPL_NOREF(LineSightTest) {
    dint flags = 0;      ///< LS_* flags @ref lineSightFlags
    Vector3d from;       ///< Ray origin.
    Vector3d to;         ///< Ray target.
    dfloat bottomSlope;  ///< Slope to bottom of target.
    dfloat topSlope;     ///< Slope to top of target.
    struct Ray {
        fixed_t origin[2];
        fixed_t direction[2];
        AABoxd bounds;
        Ray(Vector3d const &from, Vector3d const &to) {
            origin[0]    = DBL2FIX(from.x);
            origin[1]    = DBL2FIX(from.y);
            direction[0] = DBL2FIX(to.x - from.x);
            direction[1] = DBL2FIX(to.y - from.y);
            ddouble v1From[2] = { from.x, from.y };
            V2d_InitBox(bounds.arvec2, v1From);
            ddouble v1To[2] = { to.x, to.y };
            V2d_AddToBox(bounds.arvec2, v1To);
        }
    } ray;
    Impl(Vector3d const &from, Vector3d const to, dfloat bottomSlope, dfloat topSlope)
        : from       (from)
        , to         (to)
        , bottomSlope(bottomSlope)
        , topSlope   (topSlope)
        , ray        (from, to)
    {}
    /**
     * @return  @c true if the ray passes the line @a side; otherwise @c false.
     *
     * @todo cleanup: Much unnecessary representation flipping...
     * @todo cleanup: Remove front-side assumption.
     */
    bool crossLine(LineSide &side) {
#define RTOP                    0x1  ///< Top range.
#define RBOTTOM                 0x2  ///< Bottom range.
        Line &line = side.line();
        if (line.validCount() == validCount)
            return true;  // Ignore
        line.setValidCount(validCount);
        if (   line.bounds().minX > ray.bounds.maxX
            || line.bounds().maxX < ray.bounds.minX
            || line.bounds().minY > ray.bounds.maxY
            || line.bounds().maxY < ray.bounds.minY)
            return true;
        fixed_t const lineV1OriginX[2]  = { DBL2FIX(line.from().x()), DBL2FIX(line.from().y()) };
        fixed_t const lineV2OriginX[2]  = { DBL2FIX(line.to  ().x()), DBL2FIX(line.to  ().y()) };
        if (   V2x_PointOnLineSide(lineV1OriginX, ray.origin, ray.direction)
            == V2x_PointOnLineSide(lineV2OriginX, ray.origin, ray.direction))
            return true;
        fixed_t lineDirectionX[2] = { DBL2FIX(line.direction().x), DBL2FIX(line.direction().y) };
        fixed_t const fromPointX[2] = { DBL2FIX(from.x), DBL2FIX(from.y) };
        fixed_t const toPointX[2]   = { DBL2FIX(to  .x), DBL2FIX(to  .y) };
        if (   V2x_PointOnLineSide(fromPointX, lineV1OriginX, lineDirectionX)
            == V2x_PointOnLineSide(toPointX, lineV1OriginX, lineDirectionX))
            return true;
        if (!side.hasSections())
            return true;
        if (!side.hasSector())
            return false;
        Sector const *frontSec = side.sectorPtr();
        Sector const *backSec  = side.back().sectorPtr();
        bool noBack = side.considerOneSided();
        if (!noBack && !(flags & LS_PASSLEFT)) {
            noBack = (!( backSec->floor().height() < frontSec->ceiling().height()) ||
                      !(frontSec->floor().height() <  backSec->ceiling().height()));
        }
        if (noBack) {
            if (flags & LS_PASSLEFT) // Allowed.
            {
                if (line.pointOnSide(Vector2d(from.x, from.y)) < 0)
                    return true;
            }
            if (!(flags & (LS_PASSOVER | LS_PASSUNDER)))
                return false;
        }
        dbyte ranges = 0;
        if (noBack) {
            ranges |= RTOP;
        }
        else {
            if (backSec->floor().height()   != frontSec->floor().height())
                ranges |= RBOTTOM;
            if (backSec->ceiling().height() != frontSec->ceiling().height())
                ranges |= RTOP;
        }
        if (!ranges)
            return true;
        dfloat frac = FIX2FLT(V2x_Intersection(lineV1OriginX, lineDirectionX, ray.origin, ray.direction));
        if (flags & LS_PASSOVER) // Allowed.
        {
            if (bottomSlope > (frontSec->ceiling().height() - from.z) / frac)
                return true;
        }
        if (flags & LS_PASSUNDER) // Allowed.
        {
            if (topSlope    < (  frontSec->floor().height() - from.z) / frac)
                return true;
        }
        if (ranges & RTOP) {
            dfloat const top =                                      noBack ? frontSec->ceiling().height() :
                 frontSec->ceiling().height() < backSec->ceiling().height()? frontSec->ceiling().height() :
                                                                             backSec->ceiling().height();
            dfloat const slope = (top - from.z) / frac;
            if ((slope < topSlope) ^ (noBack && !(flags & LS_PASSOVER))
                || (noBack && topSlope > (frontSec->floor().height() - from.z) / frac))
                topSlope = slope;
            if ((slope < bottomSlope) ^ (noBack && !(flags & LS_PASSUNDER))
                || (noBack && bottomSlope > (frontSec->floor().height() - from.z) / frac))
                bottomSlope = slope;
        }
        if (ranges & RBOTTOM) {
            dfloat const bottom =                                noBack? frontSec->floor().height() :
                 frontSec->floor().height() > backSec->floor().height()? frontSec->floor().height() :
                                                                         backSec ->floor().height();
            dfloat const slope = (bottom - from.z) / frac;
            if (slope > bottomSlope)
                bottomSlope = slope;
            if (slope > topSlope)
                topSlope = slope;
        }
        return topSlope <= bottomSlope? false : true;
#undef RTOP
#undef RBOTTOM
    }
    /**
     * @return  @c true if the ray passes @a subspace; otherwise @c false.
     */
    bool crossSubspace(ConvexSubspace const &subspace) {
        LoopResult blocked = subspace.forAllPolyobjs([this] (Polyobj &pob) {
            for (Line *line : pob.lines()) {
                if (!crossLine(line->front()))
                    return LoopAbort;
            }
            return LoopContinue;
        });
        if (blocked) return false;
        HEdge *base  = subspace.poly().hedge();
        HEdge *hedge = base;
        do {
            if (hedge->hasMapElement()) {
                if (!crossLine(hedge->mapElementAs<LineSideSegment>().lineSide()))
                    return false;
            }
        } while ((hedge = &hedge->next()) != base);
        blocked = subspace.forAllExtraMeshes([this] (Mesh &mesh) {
            for (HEdge *hedge : mesh.hedges()) {
                if (!hedge->hasMapElement())
                    continue;
                if (!crossLine(hedge->mapElementAs<LineSideSegment>().lineSide()))
                    return LoopAbort;
            }
            return LoopContinue;
        });
        return !blocked;
    }
    /**
     * @return  @c true if the ray passes @a bspTree; otherwise @c false.
     */
    bool crossBspNode(BspTree const *bspTree) {
        DENG2_ASSERT(bspTree);
        while (!bspTree->isLeaf()) {
            DENG2_ASSERT(bspTree->userData());
            auto const &bspNode = bspTree->userData()->as<BspNode>();
            dint const fromSide = bspNode.pointOnSide(Vector2d(from.x, from.y)) < 0;
            dint const toSide   = bspNode.pointOnSide(Vector2d(to.x, to.y)) < 0;
            if (fromSide != toSide) {
                if (!crossBspNode(bspTree->childPtr(BspTree::ChildId(fromSide))))
                    return false;  // Cross the From side.
                bspTree = bspTree->childPtr(BspTree::ChildId(fromSide ^ 1));  // Cross the To side.
            }
            else {
                bspTree = bspTree->childPtr(BspTree::ChildId(fromSide));
            }
        }
        auto const &bspLeaf = bspTree->userData()->as<BspLeaf>();
        if (bspLeaf.hasSubspace()) {
            return crossSubspace(bspLeaf.subspace());
        }
        return false;
    }
};
LineSightTest::LineSightTest(Vector3d const &from, Vector3d const &to, dfloat bottomSlope,
    dfloat topSlope, dint flags)
    : d(new Impl(from, to, bottomSlope, topSlope)) {
    d->flags = flags;
}
bool LineSightTest::trace(BspTree const &bspRoot) {
    validCount++;
    d->topSlope    = d->to.z + d->topSlope    - d->from.z;
    d->bottomSlope = d->to.z + d->bottomSlope - d->from.z;
    return d->crossBspNode(&bspRoot);
}
}  // namespace world
static dint bspSplitFactor = 7;  ///< cvar
#ifdef __CLIENT__
#if 0
static dint lgMXSample = 1;  ///< 5 samples per block.
#endif
#define CLMOBJ_TIMEOUT  4000
#endif
namespace world {
struct EditableElements {
    QList<Line *> lines;
    QList<Sector *> sectors;
    QList<Polyobj *> polyobjs;
    ~EditableElements() { clearAll(); }
    void clearAll() {
        qDeleteAll(lines); lines.clear();
        qDeleteAll(sectors); sectors.clear();
        for (Polyobj *pob : polyobjs) {
            pob->~Polyobj();
            M_Free(pob);
        }
        polyobjs.clear();
    }
};
DENG2_PIMPL(Map)
, DENG2_OBSERVES(bsp::Partitioner, UnclosedSectorFound)
#ifdef __CLIENT__
, DENG2_OBSERVES(ThinkerData, Deletion)
#endif
{
    struct Bsp {
        BspTree *tree = nullptr;  ///< Owns the BspElements.
        ~Bsp() { clear(); }
        void clear() {
            if (!tree) return;
            tree->traversePostOrder(clearUserDataWorker);
            delete tree; tree = nullptr;
        }
    private:
        static dint clearUserDataWorker(BspTree &subtree, void *) {
            delete subtree.userData();
            return 0;
        }
    };
#ifdef __CLIENT__
    /**
     * All (particle) generators.
     */
    struct Generators {
        struct ListNode {
            ListNode *next;
            Generator *gen;
        };
        std::array<Generator *, MAX_GENERATORS> activeGens;
        static dint const LINKSTORE_SIZE = 4 * MAX_GENERATORS;
        ListNode *linkStore = nullptr;
        duint linkStoreCursor = 0;
        duint listsSize = 0;
        ListNode **lists = nullptr;
        ~Generators() {
            Z_Free(lists);
            Z_Free(linkStore);
        }
        /**
         * Resize the collection.
         *
         * @param listCount  Number of lists the collection must support.
         */
        void resize(duint listCount) {
            if (!linkStore) {
                linkStore = (ListNode *) Z_Malloc(sizeof(*linkStore) * LINKSTORE_SIZE, PU_MAP, 0);
                linkStoreCursor = 0;
                activeGens.fill(nullptr);
            }
            listsSize = listCount;
            lists = (ListNode **) Z_Realloc(lists, sizeof(ListNode *) * listsSize, PU_MAP);
        }
        /**
         * Returns an unused link from the linkStore.
         */
        ListNode *newLink() {
            if (linkStoreCursor < (unsigned)LINKSTORE_SIZE) {
                return &linkStore[linkStoreCursor++];
            }
            LOG_MAP_WARNING("Exhausted generator link storage");
            return nullptr;
        }
    };
    struct ContactBlockmap : public Blockmap {
        QBitArray spreadBlocks;  ///< Used to prevent repeat processing.
        /**
         * Construct a new contact blockmap.
         *
         * @param bounds    Map space boundary.
         * @param cellSize  Width and height of a cell in map space units.
         */
        ContactBlockmap(AABoxd const &bounds, duint cellSize = 128)
            : Blockmap(bounds, cellSize)
            , spreadBlocks(width() * height())
        {}
        void clear() {
            spreadBlocks.fill(false);
            unlinkAll();
        }
        /**
         * @param contact  Contact to be linked. Note that if the object's origin
         *                 lies outside the blockmap it will not be linked!
         */
        void link(Contact &contact) {
            bool outside;
            BlockmapCell cell = toCell(contact.objectOrigin(), &outside);
            if (!outside) {
                Blockmap::link(cell, &contact);
            }
        }
        void spread(AABoxd const &region) {
            spreadContacts(*this, region, &spreadBlocks);
        }
    };
#if 0
    struct BiasData {
        duint currentTime = 0;        ///< The "current" frame in milliseconds.
        duint lastChangeOnFrame = 0;
        QList<BiasSource *> sources;  ///< All bias light sources (owned).
    };
#endif
#endif // __CLIENT__
    bool editingEnabled = true;
    EditableElements editable;
    AABoxd bounds;              ///< Boundary points which encompass the entire map
    Mesh mesh;                  ///< All map geometries.
    QList<Sector *> sectors;
    QList<Line *> lines;
    QList<Polyobj *> polyobjs;
    Bsp bsp;
    QVector<ConvexSubspace *> subspaces;     ///< All player-traversable subspaces.
    QHash<Id, Subsector *> subsectorsById; ///< Not owned.
    std::unique_ptr<Thinkers> thinkers;
    Sky sky;
    std::unique_ptr<Blockmap> mobjBlockmap;
    std::unique_ptr<Blockmap> polyobjBlockmap;
    std::unique_ptr<LineBlockmap> lineBlockmap;
    std::unique_ptr<Blockmap> subspaceBlockmap;
#ifdef __CLIENT__
    std::unique_ptr<ContactBlockmap> mobjContactBlockmap;  /// @todo Redundant?
    std::unique_ptr<ContactBlockmap> lumobjContactBlockmap;
#endif
    nodepile_t mobjNodes;
    nodepile_t lineNodes;
    nodeindex_t *lineLinks = nullptr;  ///< Indices to roots.
#ifdef __CLIENT__
    PlaneSet trackedPlanes;
    SurfaceSet scrollingSurfaces;
    SkyDrawable::Animator skyAnimator;
    std::unique_ptr<Generators> generators;
#if 0
    std::unique_ptr<LightGrid> lightGrid;
    BiasData bias;            ///< Map wide "global" data for Bias lighting.
#endif
    QList<Lumobj *> lumobjs;  ///< All lumobjs (owned).
    ClSkyPlane skyFloor;
    ClSkyPlane skyCeiling;
    ClMobjHash clMobjHash;
#endif
    Impl(Public *i)
        : Base(i)
#ifdef __CLIENT__
        , skyFloor  (Sector::Floor  , DDMAXFLOAT)
        , skyCeiling(Sector::Ceiling, DDMINFLOAT)
#endif
    {
        sky.setMap(thisPublic);
        sky.setIndexInMap(0);
    }
    ~Impl() {
#ifdef __CLIENT__
        self().removeAllLumobjs();
#if 0
        self().removeAllBiasSources();
#endif
#endif
        thinkers.reset();
        qDeleteAll(sectors);
        qDeleteAll(subspaces);
        for (Polyobj *polyobj : polyobjs) {
            polyobj->~Polyobj();
            M_Free(polyobj);
        }
        qDeleteAll(lines);
    }
    /**
     * @pre Axis-aligned bounding boxes of all Sectors must be initialized.
     */
    void updateBounds() {
        bool haveGeometry = false;
        for (Line *line : lines) {
            if (line->definesPolyobj()) continue;
            if (haveGeometry) {
                V2d_UniteBox(bounds.arvec2, line->bounds().arvec2);
            }
            else {
                V2d_CopyBox(bounds.arvec2, line->bounds().arvec2);
                haveGeometry = true;
            }
        }
    }
    void unclosedSectorFound(Sector &sector, Vector2d const &nearPoint) {
        DENG2_FOR_PUBLIC_AUDIENCE(UnclosedSectorFound, i) i->unclosedSectorFound(sector, nearPoint);
    }
    /**
     * Notify interested parties of a "one-way window" in the map.
     *
     * @param line              The window line.
     * @param backFacingSector  Sector that the back of the line is facing.
     */
    void notifyOneWayWindowFound(Line &line, Sector &backFacingSector) {
        DENG2_FOR_PUBLIC_AUDIENCE(OneWayWindowFound, i) i->oneWayWindowFound(line, backFacingSector);
    }
    struct testForWindowEffectParams {
        ddouble frontDist   = 0;
        ddouble backDist    = 0;
        Sector *frontOpen   = nullptr;
        Sector *backOpen    = nullptr;
        Line *frontLine     = nullptr;
        Line *backLine      = nullptr;
        Line *testLine      = nullptr;
        bool castHorizontal = false;
        Vector2d testLineCenter;
    };
    static bool lineHasZeroLength(Line const &line) {
        return de::abs(line.length()) < 1.0 / 128.0;
    }
    static void testForWindowEffect2(Line &line, testForWindowEffectParams &p) {
        if (&line == p.testLine) return;
        if (line.isSelfReferencing()) return;
        if (lineHasZeroLength(line)) return;
        ddouble dist = 0;
        Sector *hitSector = nullptr;
        bool isFront = false;
        if (p.castHorizontal) {
            if (de::abs(line.direction().y) < bsp::DIST_EPSILON)
                return;
            if (   (line.bounds().maxY < p.testLineCenter.y - bsp::DIST_EPSILON)
                || (line.bounds().minY > p.testLineCenter.y + bsp::DIST_EPSILON))
                return;
            dist = (line.from().x() + (p.testLineCenter.y - line.from().y()) * line.direction().x / line.direction().y)
                 - p.testLineCenter.x;
            isFront = ((p.testLine->direction().y > 0) != (dist > 0));
            dist = de::abs(dist);
            if (dist < bsp::DIST_EPSILON)
                return;
            bool dir = (p.testLine->direction().y > 0) ^ (line.direction().y > 0);
            hitSector = line.side(dir ^ !isFront).sectorPtr();
        }
        else  // Cast vertically.
        {
            if (de::abs(line.direction().x) < bsp::DIST_EPSILON)
                return;
            if (   (line.bounds().maxX < p.testLineCenter.x - bsp::DIST_EPSILON)
               || (line.bounds().minX > p.testLineCenter.x + bsp::DIST_EPSILON))
                return;
            dist = (line.from().y() +
                    (p.testLineCenter.x - line.from().x()) * line.direction().y / line.direction().x)
                 - p.testLineCenter.y;
            isFront = ((p.testLine->direction().x > 0) == (dist > 0));
            dist = de::abs(dist);
            bool dir = (p.testLine->direction().x > 0) ^ (line.direction().x > 0);
            hitSector = line.side(dir ^ !isFront).sectorPtr();
        }
        if (dist < bsp::DIST_EPSILON)
            return;
        if (isFront) {
            if (dist < p.frontDist) {
                p.frontDist = dist;
                p.frontOpen = hitSector;
                p.frontLine = &line;
            }
        }
        else {
            if (dist < p.backDist) {
                p.backDist = dist;
                p.backOpen = hitSector;
                p.backLine = &line;
            }
        }
    }
    bool lineMightHaveWindowEffect(Line const &line) {
        if (line.definesPolyobj()) return false;
        if (line.front().hasSector() && line.back().hasSector()) return false;
        if (!line.front().hasSector()) return false;
        if (lineHasZeroLength(line)) return false;
        if ((line.from()._onesOwnerCount % 2) == 1
            && (line.from()._onesOwnerCount + line.from()._twosOwnerCount) > 1)
            return true;
        if  ((line.to()._onesOwnerCount % 2) == 1
             && (line.to()._onesOwnerCount + line.to()._twosOwnerCount) > 1)
            return true;
        return false;
    }
    void findOneWayWindows() {
        for (Vertex *vertex : mesh.vertexs()) {
            vertex->countLineOwners();
        }
        for (Line *line : lines) {
            if (!lineMightHaveWindowEffect(*line))
                continue;
            testForWindowEffectParams p;
            p.frontDist      = p.backDist = DDMAXFLOAT;
            p.testLine       = line;
            p.testLineCenter = line->center();
            p.castHorizontal = (de::abs(line->direction().x) < de::abs(line->direction().y)? true : false);
            AABoxd scanRegion = bounds;
            if (p.castHorizontal) {
                scanRegion.minY = line->bounds().minY - bsp::DIST_EPSILON;
                scanRegion.maxY = line->bounds().maxY + bsp::DIST_EPSILON;
            }
            else {
                scanRegion.minX = line->bounds().minX - bsp::DIST_EPSILON;
                scanRegion.maxX = line->bounds().maxX + bsp::DIST_EPSILON;
            }
            validCount++;
            self().forAllLinesInBox(scanRegion, LIF_SECTOR, [&p] (Line &line) {
                testForWindowEffect2(line, p);
                return LoopContinue;
            });
            if (p.backOpen && p.frontOpen && line->front().sectorPtr() == p.backOpen) {
                notifyOneWayWindowFound(*line, *p.frontOpen);
                line->_bspWindowSector = p.frontOpen; /// @todo Refactor away.
            }
        }
    }
    /**
     * Build a new BSP tree.
     *
     * @pre Map line bounds have been determined and a line blockmap constructed.
     */
    bool buildBspTree() {
        DENG2_ASSERT(bsp.tree == nullptr);
        DENG2_ASSERT(subspaces.isEmpty());
        Time begunAt;
        LOGDEV_MAP_XVERBOSE("Building BSP for \"%s\" with split cost factor %d...",
                (self().hasManifest() ? self().manifest().composeUri() : "(unknown map)") << bspSplitFactor);
        findOneWayWindows();
        dint nextVertexOrd = mesh.vertexCount();
        auto linesToBuildFor = QSet<Line *>::fromList(lines);
        for (Polyobj *pob : polyobjs)
        for (Line *line : pob->lines()) {
            linesToBuildFor.remove(line);
        }
        try {
            bsp::Partitioner partitioner(bspSplitFactor);
            partitioner.audienceForUnclosedSectorFound += this;
            bsp.tree = partitioner.makeBspTree(linesToBuildFor, mesh);
            DENG2_ASSERT(bsp.tree);
            LOG_MAP_VERBOSE("BSP built: %s. With %d Segments and %d Vertexes.")
                    << bsp.tree->summary()
                    << partitioner.segmentCount()
                    << partitioner.vertexCount();
            for (dint i = nextVertexOrd; i < mesh.vertexCount(); ++i) {
                Vertex *vtx = mesh.vertexs().at(i);
                vtx->setMap(thisPublic);
                vtx->setIndexInMap(i);
            }
#ifdef DENG2_QT_4_7_OR_NEWER
            subspaces.reserve(bsp.tree->leafCount());
#endif
            BspTree const *cur  = bsp.tree;
            BspTree const *prev = nullptr;
            while (cur) {
                while (cur) {
                    if (cur->userData()) {
                        if (cur->isLeaf()) {
                            auto &leaf = cur->userData()->as<BspLeaf>();
                            if (!leaf.sectorPtr()) {
                                LOGDEV_MAP_WARNING("BSP leaf %p has degenerate geometry (%d half-edges).")
                                    << &leaf << (leaf.hasSubspace()? leaf.subspace().poly().hedgeCount() : 0);
                            }
                            if (leaf.hasSubspace()) {
                                ConvexSubspace &subspace = leaf.subspace();
                                subspace.setIndexInMap(subspaces.count());
                                subspaces.append(&subspace);
#ifdef DENG_DEBUG  // See if we received a partial geometry...
                                dint discontinuities = 0;
                                HEdge *hedge = subspace.poly().hedge();
                                do {
                                    if (hedge->next().origin() != hedge->twin().origin()) {
                                        discontinuities++;
                                    }
                                } while ((hedge = &hedge->next()) != subspace.poly().hedge());
                                if (discontinuities) {
                                    LOGDEV_MAP_WARNING("Face geometry for BSP leaf [%p] at %s in sector %i "
                                                       "is not contiguous (%i gaps/overlaps).\n%s")
                                        << &leaf << subspace.poly().center().asText()
                                        << (leaf.sectorPtr()? leaf.sectorPtr()->indexInArchive() : -1)
                                        << discontinuities
                                        << subspace.poly().description();
                                }
#endif
                            }
                        }
                    }
                    if (prev == cur->parentPtr()) {
                        prev = cur;
                        if (cur->hasRight()) cur = cur->rightPtr();
                        else                 cur = cur->leftPtr();
                    }
                    else if (prev == cur->rightPtr()) {
                        prev = cur;
                        cur  = cur->leftPtr();
                    }
                    else if (prev == cur->leftPtr()) {
                        prev = cur;
                        cur  = cur->parentPtr();
                    }
                }
                if (prev) {
                    cur = prev->parentPtr();
                }
            }
        }
        catch (Error const &er) {
            LOG_MAP_WARNING("%s.") << er.asText();
        }
        LOGDEV_MAP_VERBOSE("BSP built in %.2f seconds") << begunAt.since();
        return bsp.tree != nullptr;
    }
    /**
     * Build Subsectors for the given @a sector.
     */
    void buildSubsectors(Sector &sector) {
        DENG2_ASSERT(!sector.hasSubsectors());
        typedef QVector<ConvexSubspace *> Subspaces;
        typedef QList<Subspaces> SubspaceSets;
        SubspaceSets subspaceSets;
        for (ConvexSubspace *subspace : subspaces) {
            if (subspace->bspLeaf().sectorPtr() == &sector) {
                subspaceSets.append(Subspaces());
                subspaceSets.last().append(subspace);
            }
        }
        if (subspaceSets.isEmpty()) return;
        while (subspaceSets.count() > 1) {
            bool didMerge = false;
            for (dint i = 0; i < subspaceSets.count(); ++i)
            for (dint k = 0; k < subspaceSets.count(); ++k) {
                if (i == k) continue;
                for (ConvexSubspace *subspace : subspaceSets[i]) {
                    HEdge *baseHEdge = subspace->poly().hedge();
                    HEdge *hedge = baseHEdge;
                    do {
                        if (hedge->twin().hasFace()
                            && hedge->twin().face().hasMapElement()) {
                            auto &otherSubspace = hedge->twin().face().mapElementAs<ConvexSubspace>();
                            if (otherSubspace.bspLeaf().sectorPtr() == &sector
                                && subspaceSets[k].contains(&otherSubspace)) {
                                subspaceSets[i] += subspaceSets[k];
                                subspaceSets.removeAt(k);
                                if (i >= k) i -= 1;
                                k -= 1;
                                didMerge = true;
                                break;
                            }
                        }
                    } while ((hedge = &hedge->next()) != baseHEdge);
                    if (didMerge) break;
                }
            }
            if (!didMerge) break;
        }
        dint needed = 0;
        for (Subspaces const &subspaceSet : subspaceSets) {
            needed += subspaceSet.count();
        }
        subsectorsById.clear();
        subsectorsById.reserve(needed);
        for (Subspaces const &subspaceSet : subspaceSets) {
            Subsector *subsec = sector.addSubsector(subspaceSet);
            DENG2_ASSERT(subsec != nullptr);
            subsectorsById.insert(subsec->id(), subsec);
        }
    }
    bool unlinkMobjFromSectors(mobj_t &mob) {
        if (Mobj_IsSectorLinked(mob)) {
            Mobj_Sector(&mob)->unlink(&mob);
            return true;
        }
        return false;
    }
    /**
     * Construct an initial (empty) line blockmap.
     *
     * @pre Coordinate space bounds have already been determined.
     */
    void initLineBlockmap(ddouble margin = 8) {
        lineBlockmap.reset(
            new LineBlockmap(AABoxd(bounds.minX - margin, bounds.minY - margin,
                                    bounds.maxX + margin, bounds.maxY + margin)));
        LOG_MAP_VERBOSE("Line blockmap dimensions:")
            << lineBlockmap->dimensions().asText();
        lineBlockmap->link(lines);
    }
    /**
     * Construct an initial (empty) mobj blockmap.
     *
     * @pre Coordinate space bounds have already been determined.
     */
    void initMobjBlockmap(ddouble margin = 8) {
        mobjBlockmap.reset(
            new Blockmap(AABoxd(bounds.minX - margin, bounds.minY - margin,
                                bounds.maxX + margin, bounds.maxY + margin)));
        LOG_MAP_VERBOSE("Mobj blockmap dimensions:")
            << mobjBlockmap->dimensions().asText();
    }
    /**
     * Unlinks the mobj from all the lines it's been linked to. Can be called without
     * checking that the list does indeed contain lines.
     */
    bool unlinkMobjFromLines(mobj_t &mo) {
        if (!mo.lineRoot)
            return false; // A zero index means it's not linked.
        linknode_t *tn = mobjNodes.nodes;
        for (nodeindex_t nix = tn[mo.lineRoot].next; nix != mo.lineRoot; nix = tn[nix].next) {
            NP_Unlink((&lineNodes), tn[nix].data);
            NP_Dismiss((&lineNodes), tn[nix].data);
            NP_Dismiss((&mobjNodes), nix);
        }
        NP_Dismiss((&mobjNodes), mo.lineRoot);
        mo.lineRoot = 0;
        return true;
    }
    /**
     * @note Caller must ensure a mobj is linked only once to any given line.
     *
     * @param mob   Map-object to be linked.
     * @param line  Line to link the mobj to.
     */
    void linkMobjToLine(mobj_t *mob, Line *line) {
        if (!mob || !line) return;
        if (!line->front().hasSector()) return;
        if (!line->back().hasSector()) return;
        nodeindex_t nodeIndex = NP_New(&mobjNodes, line);
        NP_Link(&mobjNodes, nodeIndex, mob->lineRoot);
        nodeIndex = mobjNodes.nodes[nodeIndex].data = NP_New(&lineNodes, mob);
        NP_Link(&lineNodes, nodeIndex, lineLinks[line->indexInMap()]);
    }
    /**
     * @note Caller must ensure that the map-object @a mob is @em not linked.
     */
    void linkMobjToLines(mobj_t &mob) {
        AABoxd const box = Mobj_Bounds(mob);
        mob.lineRoot = NP_New(&mobjNodes, NP_ROOT_NODE);
        validCount++;
        self().forAllLinesInBox(box, [this, &mob, &box] (Line &line) {
            if (!(   box.minX >= line.bounds().maxX
                  || box.minY >= line.bounds().maxY
                  || box.maxX <= line.bounds().minX
                  || box.maxY <= line.bounds().minY)) {
                if (!line.boxOnSide(box)) {
                    this->linkMobjToLine(&mob, &line);
                }
            }
            return LoopContinue;
        });
    }
    /**
     * Construct an initial (empty) polyobj blockmap.
     *
     * @pre Coordinate space bounds have already been determined.
     */
    void initPolyobjBlockmap(ddouble margin = 8) {
        polyobjBlockmap.reset(
            new Blockmap(AABoxd(bounds.minX - margin, bounds.minY - margin,
                                bounds.maxX + margin, bounds.maxY + margin)));
        LOG_MAP_VERBOSE("Polyobj blockmap dimensions:")
            << polyobjBlockmap->dimensions().asText();
    }
    /**
     * Construct an initial (empty) convex subspace blockmap.
     *
     * @pre Coordinate space bounds have already been determined.
     */
    void initSubspaceBlockmap(ddouble margin = 8) {
        subspaceBlockmap.reset(
            new Blockmap(AABoxd(bounds.minX - margin, bounds.minY - margin,
                                bounds.maxX + margin, bounds.maxY + margin)));
        LOG_MAP_VERBOSE("Convex subspace blockmap dimensions:")
            << subspaceBlockmap->dimensions().asText();
        for (ConvexSubspace *subspace : subspaces) {
            subspaceBlockmap->link(subspace->poly().bounds(), subspace);
        }
    }
#ifdef __CLIENT__
    void initContactBlockmaps(ddouble margin = 8) {
        AABoxd expandedBounds(bounds.minX - margin, bounds.minY - margin,
                              bounds.maxX + margin, bounds.maxY + margin);
        mobjContactBlockmap.reset(new ContactBlockmap(expandedBounds));
        lumobjContactBlockmap.reset(new ContactBlockmap(expandedBounds));
    }
    /**
     * Returns the appropriate contact blockmap for the specified contact @a type.
     */
    ContactBlockmap &contactBlockmap(ContactType type) {
        switch (type) {
        case ContactMobj:   return *mobjContactBlockmap;
        case ContactLumobj: return *lumobjContactBlockmap;
        default: throw Error("Map::contactBlockmap", "Invalid contact type");
        }
    }
    /**
     * To be called to link all contacts into the contact blockmaps.
     *
     * @todo Why don't we link contacts immediately? -ds
     */
    void linkAllContacts() {
        R_ForAllContacts([this] (Contact const &contact) {
            contactBlockmap(contact.type()).link(const_cast<Contact &>(contact));
            return LoopContinue;
        });
    }
    void removeAllContacts() {
        mobjContactBlockmap->clear();
        lumobjContactBlockmap->clear();
        R_ClearContactLists(*thisPublic);
    }
#endif
    /**
     * Locate a polyobj by sound emitter.
     *
     * @param soundEmitter  SoundEmitter to search for.
     *
     * @return  Pointer to the referenced Polyobj instance; otherwise @c nullptr.
     */
    Polyobj *polyobjBySoundEmitter(SoundEmitter const &soundEmitter) const {
        for (Polyobj *polyobj : polyobjs) {
            if (&soundEmitter == &polyobj->soundEmitter())
                return polyobj;
        }
        return nullptr; // Not found.
    }
    /**
     * Locate a sector by sound emitter.
     *
     * @param soundEmitter  SoundEmitter to search for.
     *
     * @return  Pointer to the referenced Sector instance; otherwise @c nullptr.
     */
    Sector *sectorBySoundEmitter(SoundEmitter const &soundEmitter) const {
        for (Sector *sector : sectors) {
            if (&soundEmitter == &sector->soundEmitter())
                return sector;
        }
        return nullptr; // Not found.
    }
    /**
     * Locate a sector plane by sound emitter.
     *
     * @param soundEmitter  SoundEmitter to search for.
     *
     * @return  Pointer to the referenced Plane instance; otherwise @c nullptr.
     */
    Plane *planeBySoundEmitter(SoundEmitter const &soundEmitter) const {
        Plane *found = nullptr;  // Not found.
        for (Sector *sector : sectors) {
            LoopResult located = sector->forAllPlanes([&soundEmitter, &found] (Plane &plane) {
                if (&soundEmitter == &plane.soundEmitter()) {
                    found = &plane;
                    return LoopAbort;
                }
                return LoopContinue;
            });
            if (located) break;
        }
        return found;
    }
    /**
     * Locate a surface by sound emitter.
     *
     * @param soundEmitter  SoundEmitter to search for.
     *
     * @return  Pointer to the referenced Surface instance; otherwise @c nullptr.
     */
    Surface *surfaceBySoundEmitter(SoundEmitter const &soundEmitter) const {
        for (Line *line : lines)
        for (dint i = 0; i < 2; ++i) {
            LineSide &side = line->side(i);
            if (!side.hasSections()) continue;
            if (&soundEmitter == &side.middleSoundEmitter()) {
                return &side.middle();
            }
            if (&soundEmitter == &side.bottomSoundEmitter()) {
                return &side.bottom();
            }
            if (&soundEmitter == &side.topSoundEmitter()) {
                return &side.top();
            }
        }
        return nullptr;  // Not found.
    }
#ifdef __CLIENT__
    /**
     * Interpolate the smoothed height of planes.
     */
    void lerpTrackedPlanes(bool resetNextViewer) {
        if (resetNextViewer) {
            for (Plane *plane : trackedPlanes) {
                plane->resetSmoothedHeight();
            }
            trackedPlanes.clear();
        }
        else //if (!clientPaused)
        {
            QMutableSetIterator<Plane *> iter(trackedPlanes);
            while (iter.hasNext()) {
                Plane *plane = iter.next();
                plane->lerpSmoothedHeight();
                if (de::fequal(plane->heightSmoothed(), plane->height())) {
                    iter.remove();
                }
            }
        }
    }
    /**
     * Interpolate the smoothed material origin of surfaces.
     */
    void lerpScrollingSurfaces(bool resetNextViewer) {
        if (resetNextViewer) {
            for (Surface *surface : scrollingSurfaces) {
                surface->resetSmoothedOrigin();
            }
            scrollingSurfaces.clear();
        }
        else //if (!clientPaused)
        {
            QMutableSetIterator<Surface *> iter(scrollingSurfaces);
            while (iter.hasNext()) {
                Surface *surface = iter.next();
                surface->lerpSmoothedOrigin();
                if (surface->originSmoothed() == surface->origin()) {
                    iter.remove();
                }
            }
        }
    }
#if 0
    /**
     * Perform preprocessing which must be done before rendering a frame.
     */
    void biasBeginFrame() {
        if (!useBias) return;
        bias.currentTime = Timer_RealMilliseconds();
        QBitArray allChanges(BiasTracker::MAX_CONTRIBUTORS);
        bool needUpdateSurfaces = false;
        for (dint i = 0; i < bias.sources.count(); ++i) {
            BiasSource *bsrc = bias.sources.at(i);
            if (bsrc->trackChanges(allChanges, i, bias.currentTime)) {
                needUpdateSurfaces = true;
            }
        }
        if (!needUpdateSurfaces) return;
        bias.lastChangeOnFrame = R_FrameCount();
        for (Sector *sec : sectors) {
            sec->forAllSubsectors([&allChanges] (Subsector &subsec) {
                subsec.as<ClientSubsector>().applyBiasChanges(allChanges);
                return LoopContinue;
            });
        }
    }
#endif
    /**
     * Create new mobj => BSP leaf contacts.
     */
    void generateMobjContacts() {
        for (Sector *sector : sectors)
        for (mobj_t *iter = sector->firstMobj(); iter; iter = iter->sNext) {
            R_AddContact(*iter);
        }
    }
    /**
     * Perform lazy initialization of the generator collection.
     */
    Generators &getGenerators() {
        if (!generators) {
            generators.reset(new Generators);
            generators->resize(sectors.count());
        }
        return *generators;
    }
    /**
     * Lookup the next available generator id.
     *
     * @todo Optimize: Cache this result.
     * @todo Optimize: We could maintain an age-sorted list of generators.
     *
     * @return  The next available id else @c 0 iff there are no unused ids.
     */
    Generator::Id findIdForNewGenerator() {
        Generators &gens = getGenerators();
        duint unused = 0;
        for (; unused < gens.activeGens.size(); ++unused) {
            if (!gens.activeGens[unused]) break;
        }
        if (unused < gens.activeGens.size()) return Generator::Id(unused + 1);
        Generator *oldest = nullptr;
        for (Generator *gen : gens.activeGens) {
            if (!gen || gen->isStatic()) continue;
            if (!oldest || gen->age() > oldest->age()) {
                oldest = gen;
            }
        }
        return (oldest? oldest->id() : 0);
    }
    void spawnMapParticleGens() {
        if (!self().hasManifest()) return;
        for (dint i = 0; i < DED_Definitions()->ptcGens.size(); ++i) {
            ded_ptcgen_t *genDef = &DED_Definitions()->ptcGens[i];
            if (!genDef->map) continue;
            if (*genDef->map != self().manifest().composeUri())
                continue;
            if (genDef->spawnAge > 0 && App_World().time() > genDef->spawnAge)
                continue;
            Generator *gen = self().newGenerator();
            if (!gen) return;  // No more generators.
            gen->count = genDef->particles;
            gen->spawnRateMultiplier = 1;
            gen->configureFromDef(genDef);
            gen->setUntriggered();
            gen->presimulate(genDef->preSim);
        }
    }
    /**
     * Spawns all type-triggered particle generators, regardless of whether
     * the type of mobj exists in the map or not (mobjs might be dynamically
     * created).
     */
    void spawnTypeParticleGens() {
        auto &defs = *DED_Definitions();
        for (dint i = 0; i < defs.ptcGens.size(); ++i) {
            ded_ptcgen_t *def = &defs.ptcGens[i];
            if (def->typeNum != DED_PTCGEN_ANY_MOBJ_TYPE && def->typeNum < 0)
                continue;
            Generator *gen = self().newGenerator();
            if (!gen) return;  // No more generators.
            gen->count = def->particles;
            gen->spawnRateMultiplier = 1;
            gen->configureFromDef(def);
            gen->type  = def->typeNum;
            gen->type2 = def->type2Num;
            gen->presimulate(def->preSim);
        }
    }
    dint findDefForGenerator(Generator *gen) {
        DENG2_ASSERT(gen);
        auto &defs = *DED_Definitions();
        for (dint i = 0; i < defs.ptcGens.size(); ++i) {
            ded_ptcgen_t *def = &defs.ptcGens[i];
            if (def->typeNum == DED_PTCGEN_ANY_MOBJ_TYPE && gen->type == DED_PTCGEN_ANY_MOBJ_TYPE) {
                return i + 1;  // Stop iteration.
            }
            if (def->typeNum >= 0 &&
               (gen->type == def->typeNum || gen->type2 == def->type2Num)) {
                return i + 1;  // Stop iteration.
            }
            if (gen->source && gen->source->type == def->damageNum) {
                return i + 1;  // Stop iteration.
            }
            if (gen->plane && def->material) {
                try {
                    world::Material &defMat = world::Materials::get().material(*def->material);
                    Material *mat = gen->plane->surface().materialPtr();
                    if (def->flags & Generator::SpawnFloor)
                        mat = gen->plane->sector().floor().surface().materialPtr();
                    if (def->flags & Generator::SpawnCeiling)
                        mat = gen->plane->sector().ceiling().surface().materialPtr();
                    if (mat == &defMat) {
                        return i + 1; // 1-based index.
                    }
#if 0 /// @todo $revise-texture-animation
                    if (def->flags & PGF_GROUP) {
                        if (Material_IsGroupAnimated(defMat) && Material_IsGroupAnimated(mat)
                            && &Material_AnimGroup(defMat) == &Material_AnimGroup(mat)) {
                            return i + 1;  // 1-based index.
                        }
                    }
#endif
                }
                catch (MaterialManifest::MissingMaterialError const &)
                {}  // Ignore this error.
                catch (Resources::MissingResourceManifestError const &)
                {}  // Ignore this error.
            }
            if (gen->source && def->state[0]
                && ::runtimeDefs.states.indexOf(gen->source->state) == DED_Definitions()->getStateNum(def->state)) {
                return i + 1;  // 1-based index.
            }
        }
        return 0;  // Not found.
    }
    /**
     * Update existing generators in the map following an engine reset.
     */
    void updateParticleGens() {
        for (Generator *gen : getGenerators().activeGens) {
            if (!gen) continue;
            if (gen->isUntriggered()) {
                Generator_Delete(gen);
                continue;  // Continue iteration.
            }
            if (dint defIndex = findDefForGenerator(gen)) {
                gen->def = &DED_Definitions()->ptcGens[defIndex - 1];
            }
            else {
                Generator_Delete(gen);
            }
        }
        spawnMapParticleGens();
    }
    /**
     * Link all generated particles into the map so that they will be drawn.
     *
     * @todo Overkill?
     */
    void linkAllParticles() {
        Generators &gens = getGenerators();
        std::memset(gens.lists, 0, sizeof(*gens.lists) * gens.listsSize);
        gens.linkStoreCursor = 0;
        if (useParticles) {
            for (Generator *gen : gens.activeGens) {
                if (!gen) continue;
                ParticleInfo const *pInfo = gen->particleInfo();
                for (dint i = 0; i < gen->count; ++i, pInfo++) {
                    if (pInfo->stage < 0 || !pInfo->bspLeaf)
                        continue;
                    dint listIndex = pInfo->bspLeaf->sectorPtr()->indexInMap();
                    DENG2_ASSERT((unsigned)listIndex < gens.listsSize);
                    bool found = false;
                    for (Generators::ListNode *it = gens.lists[listIndex]; it; it = it->next) {
                        if (it->gen == gen) {
                            found = true; // No, no...
                        }
                    }
                    if (found) continue;
                    if (Generators::ListNode *link = gens.newLink()) {
                        link->gen  = gen;
                        link->next = gens.lists[listIndex];
                        gens.lists[listIndex] = link;
                    }
                }
            }
        }
    }
    void thinkerBeingDeleted(thinker_s &th) {
        clMobjHash.remove(th.id);
    }
#endif  // __CLIENT__
};
Map::Map(res::MapManifest *manifest)
    : BaseMap(manifest)
    , d(new Impl(this))
{}
Record const &Map::mapInfo() const {
    return App_World().mapInfoForMapUri(hasManifest() ? manifest().composeUri() : de::makeUri("Maps:"));
}
Mesh const &Map::mesh() const {
    return d->mesh;
}
bool Map::hasBspTree() const {
    return d->bsp.tree != nullptr;
}
BspTree const &Map::bspTree() const {
    if (d->bsp.tree) return *d->bsp.tree;
    throw MissingBspTreeError("Map::bspTree", "No BSP tree is available");
}
#ifdef __CLIENT__
SkyDrawable::Animator &Map::skyAnimator() const {
    return d->skyAnimator;
}
#if 0
bool Map::hasLightGrid() const {
    return bool(d->lightGrid);
}
LightGrid &Map::lightGrid() {
    if (bool(d->lightGrid)) return *d->lightGrid;
    throw MissingLightGridError("Map::lightGrid", "No light grid is initialized");
}
LightGrid const &Map::lightGrid() const {
    if (bool(d->lightGrid)) return *d->lightGrid;
    throw MissingLightGridError("Map::lightGrid", "No light grid is initialized");
}
void Map::initLightGrid() {
    if (!Con_GetInteger("rend-bias-grid"))
        return;
    static dint const MSFACTORS = 7;
    static dint multisample[] = { 1, 5, 9, 17, 25, 37, 49, 65 };
    if (bool(d->lightGrid)) {
        d->lightGrid->updateIfNeeded();
        return;
    }
    Time begunAt;
    d->lightGrid.reset(new LightGrid(origin(), dimensions()));
    LightGrid &lg = *d->lightGrid;
    dint const numSamples = multisample[de::clamp(0, lgMXSample, MSFACTORS)];
    QVector<Vector2d> samplePoints(numSamples);
    QVector<dint>     sampleHits(numSamples);
    QVector<world::ClientSubsector *> ssamples((lg.dimensions().x * lg.dimensions().y) * numSamples);
    dint size = 0, center = 0;
    if (numSamples > 1) {
        dfloat f = sqrt(dfloat( numSamples ));
        if (std::ceil(f) != std::floor(f)) {
            size = sqrt(dfloat( numSamples - 1 ));
            center = 0;
        }
        else {
            size = dint( f );
            center = size + 1;
        }
    }
    if (center == 0) {
        samplePoints[0] = Vector2d(lg.blockSize() / 2, lg.blockSize() / 2);
    }
    if (numSamples > 1) {
        ddouble bSize = ddouble(lg.blockSize()) / (size - 1);
        dint idx = (center == 0? 1 : 0);
        for (dint y = 0; y < size; ++y)
        for (dint x = 0; x < size; ++x, ++idx) {
            samplePoints[idx] = Vector2d(de::round<ddouble>(x * bSize),
                                         de::round<ddouble>(y * bSize));
        }
    }
    for (dint y = 0; y < lg.dimensions().y; ++y)
    for (dint x = 0; x < lg.dimensions().x; ++x) {
        LightGrid::Index const blk = lg.toIndex(x, y);
        Vector2d const off(x * lg.blockSize(), y * lg.blockSize());
        dint sampleOffset = 0;
        if (center == 0) {
            Subsector *s = subsectorAt(lg.origin() + off + samplePoints[0]);
            ssamples[blk * numSamples] = s ? &s->as<world::ClientSubsector>() : nullptr;
            sampleOffset++;
        }
        dint count = blk * size;
        for (dint b = 0; b < size; ++b) {
            dint i = (b + count) * size;
            for (dint a = 0; a < size; ++a, ++sampleOffset) {
                dint idx = a + i + (center == 0? blk + 1 : 0);
                if (numSamples > 1 && ((x > 0 && a == 0) || (y > 0 && b == 0))) {
                    LightGrid::Ref prev(x, y);
                    LightGrid::Ref prevB(a, b);
                    dint prevIdx;
                    if (x > 0 && a == 0) {
                        prevB.x = size -1;
                        prev.x--;
                    }
                    if (y > 0 && b == 0) {
                        prevB.y = size -1;
                        prev.y--;
                    }
                    prevIdx = prevB.x + (prevB.y + lg.toIndex(prev) * size) * size;
                    if (center == 0)
                        prevIdx += lg.toIndex(prev) + 1;
                    ssamples[idx] = ssamples[prevIdx];
                }
                else {
                    Subsector *s = subsectorAt(lg.origin() + off + samplePoints[sampleOffset]);
                    ssamples[idx] = s ? &s->as<world::ClientSubsector>() : nullptr;
                }
            }
        }
    }
    QVector<world::ClientSubsector *> blkSampleSubsectors(numSamples);
    for (dint y = 0; y < lg.dimensions().y; ++y)
    for (dint x = 0; x < lg.dimensions().x; ++x) {
        dint const sampleOffset = lg.toIndex(x, y) * numSamples;
        for (dint i = 0; i < numSamples; ++i) {
            blkSampleSubsectors[i] = ssamples[i + sampleOffset];
        }
        world::ClientSubsector *subsec = nullptr;
        if (numSamples == 1) {
            subsec = blkSampleSubsectors[center];
        }
        else {
            dint best = -1;
            sampleHits.fill(0);
            for (dint i = 0; i < numSamples; ++i) {
                if (!blkSampleSubsectors[i]) continue;
                for (dint k = 0; k < numSamples; ++k) {
                    if (blkSampleSubsectors[k] == blkSampleSubsectors[i] && blkSampleSubsectors[k]) {
                        sampleHits[k]++;
                        if (sampleHits[k] > best) {
                            best = i;
                        }
                    }
                }
            }
            if (best != -1) {
                if (sampleHits[best] == sampleHits[center] && blkSampleSubsectors[center]) {
                    subsec = blkSampleSubsectors[center];
                }
                else {
                    subsec = blkSampleSubsectors[best];
                }
            }
        }
        if (subsec) {
            lg.setPrimarySource(lg.toIndex(x, y), subsec);
        }
    }
    LOGDEV_GL_MSG("%i light blocks (%u bytes)")
        << lg.numBlocks() << lg.blockStorageSize();
    LOGDEV_GL_MSG("LightGrid init completed in %.2f seconds") << begunAt.since();
}
void Map::initBias() {
    Time begunAt;
    LOG_AS("Map::initBias");
    d->bias.sources.clear();
    if (hasManifest()) {
        String const oldUniqueId = manifest().composeUniqueId(App_CurrentGame());
        auto &defs = *DED_Definitions();
        for (dint i = 0; i < defs.lights.size(); ++i) {
            ded_light_t *lightDef = &defs.lights[i];
            if (lightDef->state[0]) continue;
            if (oldUniqueId.compareWithoutCase(lightDef->uniqueMapID)) continue;
            if (biasSourceCount() == MAX_BIAS_SOURCES)
                break;
            addBiasSource(BiasSource::fromDef(*lightDef));
        }
    }
    LOGDEV_MAP_VERBOSE("Completed in %.2f seconds") << begunAt.since();
}
#endif
void Map::initRadio() {
    LOG_AS("Map::initRadio");
    Time begunAt;
    for (Vertex *vtx : d->mesh.vertexs()) {
        vtx->updateShadowOffsets();
    }
    for (Line *line : d->lines) {
        if (!line->isShadowCaster()) continue;
        for (dint i = 0; i < 2; ++i) {
            LineSide &side = line->side(i);
            if (!side.hasSector()) continue;
            if (!side.hasSections()) continue;
            if (!side.leftHEdge() || !side.rightHEdge()) continue;
            Vertex const &vtx0   = line->vertex(i);
            Vertex const &vtx1   = line->vertex(i ^ 1);
            LineOwner const *vo0 = line->vertexOwner(i)->next();
            LineOwner const *vo1 = line->vertexOwner(i ^ 1)->prev();
            AABoxd bounds = line->bounds();
            Vector2d const sv0 = vtx0.origin() + vo0->extendedShadowOffset();
            V2d_AddToBoxXY(bounds.arvec2, sv0.x, sv0.y);
            Vector2d const sv1 = vtx1.origin() + vo1->extendedShadowOffset();
            V2d_AddToBoxXY(bounds.arvec2, sv1.x, sv1.y);
            ::validCount++;
            dint const localValidCount = ::validCount;
            subspaceBlockmap().forAllInBox(bounds, [&bounds, &side, &localValidCount] (void *object) {
                auto &sub = *(ConvexSubspace *)object;
                if (sub.validCount() != localValidCount)  // not yet processed
                {
                    sub.setValidCount(localValidCount);
                    if (&sub.subsector().sector() == side.sectorPtr()) {
                        AABoxd const &polyBox = sub.poly().bounds();
                        if (!(   polyBox.maxX < bounds.minX
                              || polyBox.minX > bounds.maxX
                              || polyBox.minY > bounds.maxY
                              || polyBox.maxY < bounds.minY)) {
                            sub.addShadowLine(side);
                        }
                    }
                }
                return LoopContinue;
            });
        }
    }
    LOGDEV_GL_MSG("Completed in %.2f seconds") << begunAt.since();
}
void Map::initContactBlockmaps() {
    d->initContactBlockmaps();
}
void Map::spreadAllContacts(AABoxd const &region) {
    d->mobjContactBlockmap->
        spread(AABoxd(region.minX - DDMOBJ_RADIUS_MAX, region.minY - DDMOBJ_RADIUS_MAX,
                      region.maxX + DDMOBJ_RADIUS_MAX, region.maxY + DDMOBJ_RADIUS_MAX));
    d->lumobjContactBlockmap->
        spread(AABoxd(region.minX - Lumobj::radiusMax(), region.minY - Lumobj::radiusMax(),
                      region.maxX + Lumobj::radiusMax(), region.maxY + Lumobj::radiusMax()));
}
void Map::initGenerators() {
    LOG_AS("Map::initGenerators");
    Time begunAt;
    d->spawnTypeParticleGens();
    d->spawnMapParticleGens();
    LOGDEV_MAP_VERBOSE("Completed in %.2f seconds") << begunAt.since();
}
void Map::spawnPlaneParticleGens() {
    for (Sector *sector : d->sectors) {
        Plane &floor = sector->floor();
        floor.spawnParticleGen(Def_GetGenerator(floor.surface().composeMaterialUri()));
        Plane &ceiling = sector->ceiling();
        ceiling.spawnParticleGen(Def_GetGenerator(ceiling.surface().composeMaterialUri()));
    }
}
void Map::clearClMobjs() {
    d->clMobjHash.clear();
}
mobj_t *Map::clMobjFor(thid_t id, bool canCreate) const {
    LOG_AS("Map::clMobjFor");
    ClMobjHash::const_iterator found = d->clMobjHash.constFind(id);
    if (found != d->clMobjHash.constEnd()) {
        return found.value();
    }
    if (!canCreate) return nullptr;
    MobjThinker mob(Thinker::AllocateMemoryZone);
    mob.id       = id;
    mob.function = reinterpret_cast<thinkfunc_t>(gx.MobjThinker);
    auto *data = new ClientMobjThinkerData;
    data->remoteSync().flags = DDMF_REMOTE;
    mob.setData(data);
    d->clMobjHash.insert(id, mob);
    data->audienceForDeletion() += d;  // for removing from the hash
    d->thinkers->setMobjId(id);  // Mark this ID as used.
    d->thinkers->add(*(thinker_t *)mob);
    return mob.take();
}
dint Map::clMobjIterator(dint (*callback)(mobj_t *, void *), void *context) {
    ClMobjHash::const_iterator next;
    for (ClMobjHash::const_iterator i = d->clMobjHash.constBegin();
         i != d->clMobjHash.constEnd(); i = next) {
        next = i;
        next++;
        DENG2_ASSERT(THINKER_DATA(i.value()->thinker, ClientMobjThinkerData).hasRemoteSync());
        if (dint result = callback(i.value(), context))
            return result;
    }
    return 0;
}
Map::ClMobjHash const &Map::clMobjHash() const {
    return d->clMobjHash;
}
#endif // __CLIENT__
AABoxd const &Map::bounds() const {
    return d->bounds;
}
coord_t Map::gravity() const {
    return _effectiveGravity;
}
void Map::setGravity(coord_t newGravity) {
    if (!de::fequal(_effectiveGravity, newGravity)) {
        _effectiveGravity = newGravity;
        LOG_MAP_VERBOSE("Effective gravity for %s now %.1f")
            << (hasManifest() ? manifest().gets("id") : "(unknown map)") << _effectiveGravity;
    }
}
Thinkers &Map::thinkers() const {
    if (bool( d->thinkers )) return *d->thinkers;
    throw MissingThinkersError("Map::thinkers", "Thinkers not initialized");
}
Sky &Map::sky() const {
    return d->sky;
}
dint Map::vertexCount() const {
    return d->mesh.vertexCount();
}
Vertex &Map::vertex(dint index) const {
    if (Vertex *vtx = vertexPtr(index)) return *vtx;
    throw MissingElementError("Map::vertex", "Unknown Vertex index:" + String::number(index));
}
Vertex *Map::vertexPtr(dint index) const {
    if (index >= 0 && index < d->mesh.vertexCount()) {
        return d->mesh.vertexs().at(index);
    }
    return nullptr;
}
LoopResult Map::forAllVertexs(std::function<LoopResult (Vertex &)> func) const {
    for (Vertex *vtx : d->mesh.vertexs()) {
        if (auto result = func(*vtx)) return result;
    }
    return LoopContinue;
}
dint Map::lineCount() const {
    return d->lines.count();
}
Line &Map::line(dint index) const {
    if (Line *li = linePtr(index)) return *li;
    throw MissingElementError("Map::line", "Unknown Line index:" + String::number(index));
}
Line *Map::linePtr(dint index) const {
    if (index >= 0 && index < d->lines.count()) {
        return d->lines.at(index);
    }
    return nullptr;
}
LoopResult Map::forAllLines(std::function<LoopResult (Line &)> func) const {
    for (Line *li : d->lines) {
        if (auto result = func(*li)) return result;
    }
    return LoopContinue;
}
dint Map::sectorCount() const {
    return d->sectors.count();
}
#ifdef __CLIENT__
bool Map::isPointInVoid(de::Vector3d const &point) const {
    BspLeaf const &bspLeaf = bspLeafAt(point);
    if (bspLeaf.hasSubspace() && bspLeaf.subspace().contains(point) && bspLeaf.subspace().hasSubsector()) {
        auto const &subsec = bspLeaf.subspace().subsector().as<ClientSubsector>();
        return subsec.isHeightInVoid(point.z);
    }
    return true; // In the void.
}
#endif
dint Map::subspaceCount() const {
    return d->subspaces.count();
}
ConvexSubspace &Map::subspace(dint index) const {
    if (ConvexSubspace *sub = subspacePtr(index)) return *sub;
    throw MissingElementError("Map::subspace", "Unknown subspace index:" + String::number(index));
}
ConvexSubspace *Map::subspacePtr(dint index) const {
    if (index >= 0 && index < d->subspaces.count()) {
        return d->subspaces.at(index);
    }
    return nullptr;
}
LoopResult Map::forAllSubspaces(std::function<LoopResult (ConvexSubspace &)> func) const {
    for (ConvexSubspace *sub : d->subspaces) {
        if (auto result = func(*sub)) return result;
    }
    return LoopContinue;
}
dint Map::polyobjCount() const {
    return d->polyobjs.count();
}
Polyobj &Map::polyobj(dint index) const {
    if (Polyobj *pob = polyobjPtr(index)) return *pob;
    throw MissingObjectError("Map::subspace", "Unknown Polyobj index:" + String::number(index));
}
Polyobj *Map::polyobjPtr(dint index) const {
    if (index >= 0 && index < d->polyobjs.count()) {
        return d->polyobjs.at(index);
    }
    return nullptr;
}
LoopResult Map::forAllPolyobjs(std::function<LoopResult (Polyobj &)> func) const {
    for (Polyobj *pob : d->polyobjs) {
        if (auto result = func(*pob)) return result;
    }
    return LoopContinue;
}
void Map::initPolyobjs() {
    LOG_AS("Map::initPolyobjs");
    for (Polyobj *po : d->polyobjs) {
        po->updateBounds();
        po->updateSurfaceTangents();
        po->unlink();
        po->link();
    }
}
dint Map::ambientLightLevel() const {
    return _ambientLightLevel;
}
LineSide &Map::side(dint index) const {
    if (LineSide *side = sidePtr(index)) return *side;
    throw MissingElementError("Map::side", "Unknown LineSide index:" + String::number(index));
}
LineSide *Map::sidePtr(dint index) const {
    if (index < 0) return nullptr;
    return &d->lines.at(index / 2)->side(index % 2);
}
dint Map::toSideIndex(dint lineIndex, dint backSide) // static
{
    DENG_ASSERT(lineIndex >= 0);
    return lineIndex * 2 + (backSide? 1 : 0);
}
bool Map::identifySoundEmitter(SoundEmitter const &emitter, Sector **sector,
    Polyobj **poly, Plane **plane, Surface **surface) const {
    *sector  = nullptr;
    *poly    = nullptr;
    *plane   = nullptr;
    *surface = nullptr;
    *poly = d->polyobjBySoundEmitter(emitter);
    if (!*poly) {
        *sector = d->sectorBySoundEmitter(emitter);
        if (!*sector) {
            *plane = d->planeBySoundEmitter(emitter);
            if (!*plane) {
                *surface = d->surfaceBySoundEmitter(emitter);
            }
        }
    }
    return (*sector != 0 || *poly != 0|| *plane != 0|| *surface != 0);
}
void Map::initNodePiles() {
    LOG_AS("Map");
    Time begunAt;
    NP_Init(&d->mobjNodes, 256);  // Allocate a small pile.
    NP_Init(&d->lineNodes, lineCount() + 1000);
    DENG_ASSERT(d->lineLinks == nullptr);
    d->lineLinks = (nodeindex_t *) Z_Malloc(sizeof(*d->lineLinks) * lineCount(), PU_MAPSTATIC, 0);
    for (dint i = 0; i < lineCount(); ++i) {
        d->lineLinks[i] = NP_New(&d->lineNodes, NP_ROOT_NODE);
    }
    LOGDEV_MAP_MSG("Initialized node piles in %.2f seconds") << begunAt.since();
}
Sector &Map::sector(dint index) const {
    if (Sector *sec = sectorPtr(index)) return *sec;
    throw MissingElementError("Map::sector", "Unknown Sector index:" + String::number(index));
}
Sector *Map::sectorPtr(dint index) const {
    if (index >= 0 && index < d->sectors.count()) {
        return d->sectors.at(index);
    }
    return nullptr;
}
LoopResult Map::forAllSectors(const std::function<LoopResult (Sector &)> &func) const {
    for (Sector *sec : d->sectors) {
        if (auto result = func(*sec)) return result;
    }
    return LoopContinue;
}
Subsector *Map::subsectorAt(Vector2d const &point) const {
    BspLeaf &bspLeaf = bspLeafAt(point);
    if (bspLeaf.hasSubspace() && bspLeaf.subspace().contains(point)) {
        return bspLeaf.subspace().subsectorPtr();
    }
    return nullptr;
}
Subsector &Map::subsector(de::Id id) const {
    if (Subsector *subsec = subsectorPtr(id)) return *subsec;
    throw MissingSubsectorError("Map::subsector", "Unknown Subsector \"" + id.asText() + "\"");
}
Subsector *Map::subsectorPtr(de::Id id) const {
    auto found = d->subsectorsById.constFind(id);
    if (found != d->subsectorsById.constEnd()) {
        return found.value();
    }
    return nullptr;
}
Blockmap const &Map::mobjBlockmap() const {
    if (bool(d->mobjBlockmap)) return *d->mobjBlockmap;
    throw MissingBlockmapError("Map::mobjBlockmap", "Mobj blockmap is not initialized");
}
Blockmap const &Map::polyobjBlockmap() const {
    if (bool(d->polyobjBlockmap)) return *d->polyobjBlockmap;
    throw MissingBlockmapError("Map::polyobjBlockmap", "Polyobj blockmap is not initialized");
}
LineBlockmap const &Map::lineBlockmap() const {
    if (bool(d->lineBlockmap)) return *d->lineBlockmap;
    throw MissingBlockmapError("Map::lineBlockmap", "Line blockmap is not initialized");
}
Blockmap const &Map::subspaceBlockmap() const {
    if (bool(d->subspaceBlockmap)) return *d->subspaceBlockmap;
    throw MissingBlockmapError("Map::subspaceBlockmap", "Convex subspace blockmap is not initialized");
}
LoopResult Map::forAllLinesTouchingMobj(mobj_t &mob, std::function<LoopResult (Line &)> func) const {
    if (&Mobj_Map(mob) == this && Mobj_IsLinked(mob) && mob.lineRoot) {
        QVarLengthArray<Line *, 16> linkStore;
        linknode_t *tn = d->mobjNodes.nodes;
        for (nodeindex_t nix = tn[mob.lineRoot].next; nix != mob.lineRoot; nix = tn[nix].next) {
            linkStore.append((Line *)(tn[nix].ptr));
        }
        for (dint i = 0; i < linkStore.count(); ++i) {
            if (auto result = func(*linkStore[i]))
                return result;
        }
    }
    return LoopContinue;
}
LoopResult Map::forAllSectorsTouchingMobj(mobj_t &mob, std::function<LoopResult (Sector &)> func) const {
    if (&Mobj_Map(mob) == this && Mobj_IsLinked(mob)) {
        QVarLengthArray<Sector *, 16> linkStore;
        Sector &ownSec = *Mobj_BspLeafAtOrigin(mob).sectorPtr();
        ownSec.setValidCount(validCount);
        linkStore.append(&ownSec);
        if (mob.lineRoot) {
            linknode_t *tn = d->mobjNodes.nodes;
            for (nodeindex_t nix = tn[mob.lineRoot].next; nix != mob.lineRoot; nix = tn[nix].next) {
                auto *ld = (Line *)(tn[nix].ptr);
                Sector &frontSec = ld->front().sector();
                if (frontSec.validCount() != validCount) {
                    frontSec.setValidCount(validCount);
                    linkStore.append(&frontSec);
                }
                if (ld->back().hasSector()) {
                    Sector &backSec = ld->back().sector();
                    if (backSec.validCount() != validCount) {
                        backSec.setValidCount(validCount);
                        linkStore.append(&backSec);
                    }
                }
            }
        }
        for (dint i = 0; i < linkStore.count(); ++i) {
            if (auto result = func(*linkStore[i]))
                return result;
        }
    }
    return LoopContinue;
}
LoopResult Map::forAllMobjsTouchingLine(Line &line, std::function<LoopResult (mobj_t &)> func) const {
    if (&line.map() == this) {
        QVarLengthArray<mobj_t *, 256> linkStore;
        linknode_t *ln   = d->lineNodes.nodes;
        nodeindex_t root = d->lineLinks[line.indexInMap()];
        for (nodeindex_t nix = ln[root].next; nix != root; nix = ln[nix].next) {
            linkStore.append((mobj_t *)(ln[nix].ptr));
        }
        for (dint i = 0; i < linkStore.count(); ++i) {
            if (auto result = func(*linkStore[i]))
                return result;
        }
    }
    return LoopContinue;
}
LoopResult Map::forAllMobjsTouchingSector(Sector &sector, std::function<LoopResult (mobj_t &)> func) const {
    if (&sector.map() == this) {
        QVarLengthArray<mobj_t *, 256> linkStore;
        for (mobj_t *mob = sector.firstMobj(); mob; mob = mob->sNext) {
            if (mob->validCount != validCount) {
                mob->validCount = validCount;
                linkStore.append(mob);
            }
        }
        linknode_t const *ln = d->lineNodes.nodes;
        sector.forAllSides([this, &linkStore, &ln] (LineSide &side) {
            nodeindex_t root = d->lineLinks[side.line().indexInMap()];
            for (nodeindex_t nix = ln[root].next; nix != root; nix = ln[nix].next) {
                auto *mob = (mobj_t *)(ln[nix].ptr);
                if (mob->validCount != validCount) {
                    mob->validCount = validCount;
                    linkStore.append(mob);
                }
            }
            return LoopContinue;
        });
        for (dint i = 0; i < linkStore.count(); ++i) {
            if (auto result = func(*linkStore[i]))
                return result;
        }
    }
    return LoopContinue;
}
dint Map::unlink(mobj_t &mob) {
    dint links = 0;
    if (d->unlinkMobjFromSectors(mob))
        links |= MLF_SECTOR;
    BlockmapCell cell = d->mobjBlockmap->toCell(Mobj_Origin(mob));
    if (d->mobjBlockmap->unlink(cell, &mob))
        links |= MLF_BLOCKMAP;
    if (!d->unlinkMobjFromLines(mob))
        links |= MLF_NOLINE;
    return links;
}
void Map::link(mobj_t &mob, dint flags) {
    BspLeaf &bspLeafAtOrigin = bspLeafAt_FixedPrecision(Mobj_Origin(mob));
    if (flags & MLF_SECTOR) {
        d->unlinkMobjFromSectors(mob);
        bspLeafAtOrigin.sectorPtr()->link(&mob);
    }
    mob._bspLeaf = &bspLeafAtOrigin;
    if (flags & MLF_BLOCKMAP) {
        BlockmapCell cell = d->mobjBlockmap->toCell(Mobj_Origin(mob));
        d->mobjBlockmap->link(cell, &mob);
    }
    if (!(flags & MLF_NOLINE)) {
        d->unlinkMobjFromLines(mob);
        d->linkMobjToLines(mob);
    }
#ifdef __CLIENT__
    if (mob.dPlayer && mob.dPlayer->mo) {
        auto &client = ClientApp::player(P_GetDDPlayerIdx(mob.dPlayer));
        client.inVoid = true;
        if (Mobj_HasSubsector(mob)) {
            auto &subsec = Mobj_Subsector(mob).as<world::ClientSubsector>();
            if (Mobj_BspLeafAtOrigin(mob).subspace().contains(Mobj_Origin(mob))) {
                if (   mob.origin[2] <  subsec.visCeiling().heightSmoothed() + 4
                    && mob.origin[2] >= subsec.  visFloor().heightSmoothed()) {
                    client.inVoid = false;
                }
            }
        }
    }
#endif
}
void Map::unlink(Polyobj &polyobj) {
    d->polyobjBlockmap->unlink(polyobj.bounds, &polyobj);
}
void Map::link(Polyobj &polyobj) {
    d->polyobjBlockmap->link(polyobj.bounds, &polyobj);
}
LoopResult Map::forAllLinesInBox(AABoxd const &box, dint flags, std::function<LoopResult (Line &line)> func) const {
    LoopResult result = LoopContinue;
    if ((flags & LIF_POLYOBJ) && polyobjCount()) {
        dint const localValidCount = validCount;
        result = polyobjBlockmap().forAllInBox(box, [&func, &localValidCount] (void *object) {
            auto &pob = *(Polyobj *)object;
            if (pob.validCount != localValidCount) // not yet processed
            {
                pob.validCount = localValidCount;
                for (Line *line : pob.lines()) {
                    if (line->validCount() != localValidCount) // not yet processed
                    {
                        line->setValidCount(localValidCount);
                        if (auto result = func(*line))
                            return result;
                    }
                }
            }
            return LoopResult(); // continue
        });
    }
    if (!result && (flags & LIF_SECTOR)) {
        dint const localValidCount = validCount;
        result = lineBlockmap().forAllInBox(box, [&func, &localValidCount] (void *object) {
            auto &line = *(Line *)object;
            if (line.validCount() != localValidCount) // not yet processed
            {
                line.setValidCount(localValidCount);
                return func(line);
            }
            return LoopResult(); // continue
        });
    }
    return result;
}
BspLeaf &Map::bspLeafAt(Vector2d const &point) const {
    if (!d->bsp.tree)
        throw MissingBspTreeError("Map::bspLeafAt", "No BSP data available");
    BspTree const *bspTree = d->bsp.tree;
    while (!bspTree->isLeaf()) {
        auto &bspNode = bspTree->userData()->as<BspNode>();
        dint side     = bspNode.pointOnSide(point) < 0;
        bspTree = bspTree->childPtr(BspTree::ChildId(side));
    }
    return bspTree->userData()->as<BspLeaf>();
}
BspLeaf &Map::bspLeafAt_FixedPrecision(Vector2d const &point) const {
    if (!d->bsp.tree)
        throw MissingBspTreeError("Map::bspLeafAt_FixedPrecision", "No BSP data available");
    fixed_t pointX[2] = { DBL2FIX(point.x), DBL2FIX(point.y) };
    BspTree const *bspTree = d->bsp.tree;
    while (!bspTree->isLeaf()) {
        auto const &bspNode = bspTree->userData()->as<BspNode>();
        fixed_t lineOriginX[2]    = { DBL2FIX(bspNode.origin.x),    DBL2FIX(bspNode.origin.y) };
        fixed_t lineDirectionX[2] = { DBL2FIX(bspNode.direction.x), DBL2FIX(bspNode.direction.y) };
        dint side = V2x_PointOnLineSide(pointX, lineOriginX, lineDirectionX);
        bspTree = bspTree->childPtr(BspTree::ChildId(side));
    }
    return bspTree->userData()->as<BspLeaf>();
}
#ifdef __CLIENT__
void Map::updateScrollingSurfaces() {
    for (Surface *surface : d->scrollingSurfaces) {
        surface->updateOriginTracking();
    }
}
Map::SurfaceSet &Map::scrollingSurfaces() {
    return d->scrollingSurfaces;
}
void Map::updateTrackedPlanes() {
    for (Plane *plane : d->trackedPlanes) {
        plane->updateHeightTracking();
    }
}
Map::PlaneSet &Map::trackedPlanes() {
    return d->trackedPlanes;
}
void Map::initSkyFix() {
    Time begunAt;
    LOG_AS("Map::initSkyFix");
    d->skyFloor  .setHeight(DDMAXFLOAT);
    d->skyCeiling.setHeight(DDMINFLOAT);
    for (Sector *sector : d->sectors) {
        if (!sector->sideCount()) continue;
        bool const skyFloor = sector->floor  ().surface().hasSkyMaskedMaterial();
        bool const skyCeil  = sector->ceiling().surface().hasSkyMaskedMaterial();
        if (!skyFloor && !skyCeil) continue;
        if (skyCeil) {
            if (sector->ceiling().heightSmoothed() > d->skyCeiling.height()) {
                d->skyCeiling.setHeight(sector->ceiling().heightSmoothed());
            }
            for (mobj_t *mob = sector->firstMobj(); mob; mob = mob->sNext) {
                ddouble extent = mob->origin[2] + mob->height;
                if (extent > d->skyCeiling.height()) {
                    d->skyCeiling.setHeight(extent);
                }
            }
        }
        if (skyFloor) {
            if (sector->floor().heightSmoothed() < d->skyFloor.height()) {
                d->skyFloor.setHeight(sector->floor().heightSmoothed());
            }
        }
        sector->forAllSides([this, &skyCeil, &skyFloor] (LineSide &side) {
            if (!side.hasSections()) return LoopContinue;
            if (!side.middle().hasMaterial()) return LoopContinue;
            if (!side.hasSector() || !side.back().hasSector())
                return LoopContinue;
            if (!side.leftHEdge()) return LoopContinue;
            WallEdge edge(WallSpec::fromMapSide(side, LineSide::Middle),
                          *side.leftHEdge(), Line::From);
            if (edge.isValid() && edge.top().z() > edge.bottom().z()) {
                if (skyCeil && edge.top().z() + edge.origin().y > d->skyCeiling.height()) {
                    d->skyCeiling.setHeight(edge.top().z() + edge.origin().y);
                }
                if (skyFloor && edge.bottom().z() + edge.origin().y < d->skyFloor.height()) {
                    d->skyFloor.setHeight(edge.bottom().z() + edge.origin().y);
                }
            }
            return LoopContinue;
        });
    }
    LOGDEV_MAP_VERBOSE("Completed in %.2f seconds") << begunAt.since();
}
ClSkyPlane &Map::skyFloor() {
    return d->skyFloor;
}
ClSkyPlane const &Map::skyFloor() const {
    return d->skyFloor;
}
ClSkyPlane &Map::skyCeiling() {
    return d->skyCeiling;
}
ClSkyPlane const &Map::skyCeiling() const {
    return d->skyCeiling;
}
Generator *Map::newGenerator() {
    Generator::Id id = d->findIdForNewGenerator();  // 1-based
    if (!id) return nullptr;  // Failed; too many generators?
    Impl::Generators &gens = d->getGenerators();
    if (id > 0 && (unsigned)id <= gens.activeGens.size()) {
        Generator_Delete(gens.activeGens[id - 1]);
    }
    auto *gen = (Generator *) Z_Calloc(sizeof(Generator), PU_MAP, 0);
    gen->setId(id);
    gen->thinker.function = (thinkfunc_t) Generator_Thinker;
    d->thinkers->add(gen->thinker, false /*not public*/);
    gens.activeGens[id - 1] = gen;
    return gen;
}
dint Map::generatorCount() const {
    if (!d->generators) return 0;
    dint count = 0;
    for (Generator *gen : d->getGenerators().activeGens) {
        if (gen) count += 1;
    }
    return count;
}
void Map::unlink(Generator &generator) {
    Impl::Generators &gens = d->getGenerators();
    for (duint i = 0; i < gens.activeGens.size(); ++i) {
        if (gens.activeGens[i] == &generator) {
            gens.activeGens[i] = nullptr;
            break;
        }
    }
}
LoopResult Map::forAllGenerators(std::function<LoopResult (Generator &)> func) const {
    for (Generator *gen : d->getGenerators().activeGens) {
        if (!gen) continue;
        if (auto result = func(*gen))
            return result;
    }
    return LoopContinue;
}
LoopResult Map::forAllGeneratorsInSector(Sector const &sector, std::function<LoopResult (Generator &)> func) const {
    if (sector.mapPtr() == this)  // Ignore 'alien' sectors.
    {
        duint const listIndex = sector.indexInMap();
        Impl::Generators &gens = d->getGenerators();
        for (Impl::Generators::ListNode *it = gens.lists[listIndex]; it; it = it->next) {
            if (auto result = func(*it->gen))
                return result;
        }
    }
    return LoopContinue;
}
dint Map::lumobjCount() const {
    return d->lumobjs.count();
}
Lumobj &Map::addLumobj(Lumobj *lumobj) {
    DENG2_ASSERT(lumobj != nullptr);
    d->lumobjs.append(lumobj);
    lumobj->setMap(this);
    lumobj->setIndexInMap(d->lumobjs.count() - 1);
    DENG2_ASSERT(lumobj->bspLeafAtOrigin().hasSubspace());
    lumobj->bspLeafAtOrigin().subspace().link(*lumobj);
    R_AddContact(*lumobj);  // For spreading purposes.
    return *lumobj;
}
void Map::removeLumobj(dint which) {
    if (which >= 0 && which < lumobjCount()) {
        delete d->lumobjs.takeAt(which);
    }
}
void Map::removeAllLumobjs() {
    for (ConvexSubspace *subspace : d->subspaces) {
        subspace->unlinkAllLumobjs();
    }
    qDeleteAll(d->lumobjs); d->lumobjs.clear();
}
Lumobj &Map::lumobj(dint index) const {
    if (Lumobj *lum = lumobjPtr(index)) return *lum;
    throw MissingObjectError("Map::lumobj", "Unknown Lumobj index:" + String::number(index));
}
Lumobj *Map::lumobjPtr(dint index) const {
    if (index >= 0 && index < d->lumobjs.count()) {
        return d->lumobjs.at(index);
    }
    return nullptr;
}
LoopResult Map::forAllLumobjs(std::function<LoopResult (Lumobj &)> func) const {
    for (Lumobj *lob : d->lumobjs) {
        if (auto result = func(*lob)) return result;
    }
    return LoopContinue;
}
#if 0
dint Map::biasSourceCount() const {
    return d->bias.sources.count();
}
BiasSource &Map::addBiasSource(BiasSource const &biasSource) {
    if (biasSourceCount() < MAX_BIAS_SOURCES) {
        d->bias.sources.append(new BiasSource(biasSource));
        return *d->bias.sources.last();
    }
    throw FullError("Map::addBiasSource", "Already at full capacity:" + String::number(MAX_BIAS_SOURCES));
}
void Map::removeBiasSource(dint which) {
    if (which >= 0 && which < biasSourceCount()) {
        delete d->bias.sources.takeAt(which);
    }
}
void Map::removeAllBiasSources() {
    qDeleteAll(d->bias.sources); d->bias.sources.clear();
}
BiasSource &Map::biasSource(dint index) const {
   if (BiasSource *bsrc = biasSourcePtr(index)) return *bsrc;
   throw MissingObjectError("Map::biasSource", "Unknown BiasSource index:" + String::number(index));
}
BiasSource *Map::biasSourcePtr(dint index) const {
   if (index >= 0 && index < d->bias.sources.count()) {
       return d->bias.sources.at(index);
   }
   return nullptr;
}
/**
 * @todo Implement a blockmap for these?
 * @todo Cache this result (MRU?).
 */
BiasSource *Map::biasSourceNear(Vector3d const &point) const {
    BiasSource *nearest = nullptr;
    ddouble minDist = 0;
    for (BiasSource *src : d->bias.sources) {
        ddouble dist = (src->origin() - point).length();
        if (!nearest || dist < minDist) {
            minDist = dist;
            nearest = src;
        }
    }
    return nearest;
}
LoopResult Map::forAllBiasSources(std::function<LoopResult (BiasSource &)> func) const {
    for (BiasSource *bsrc : d->bias.sources) {
        if (auto result = func(*bsrc)) return result;
    }
    return LoopContinue;
}
dint Map::indexOf(BiasSource const &bsrc) const {
    return d->bias.sources.indexOf(const_cast<BiasSource *>(&bsrc));
}
duint Map::biasCurrentTime() const {
    return d->bias.currentTime;
}
duint Map::biasLastChangeOnFrame() const {
    return d->bias.lastChangeOnFrame;
}
#endif // 0
#endif // __CLIENT__
void Map::update() {
#ifdef __CLIENT__
    d->updateParticleGens();  // Defs might've changed.
    for (Sector *sector : d->sectors) {
        sector->forAllSubsectors([] (Subsector &subsector) {
            subsector.as<ClientSubsector>().markForDecorationUpdate();
            return LoopContinue;
        });
    }
#endif // __CLIENT__
    Record const &inf = mapInfo();
    _ambientLightLevel = inf.getf("ambient") * 255;
    _globalGravity     = inf.getf("gravity");
    _effectiveGravity  = _globalGravity;
#ifdef __CLIENT__
    defn::Sky skyDef;
    if (Record const *def = DED_Definitions()->skies.tryFind("id", inf.gets("skyId"))) {
        skyDef = *def;
    }
    else {
        skyDef = inf.subrecord("sky");
    }
    sky().configure(&skyDef);
#endif
}
#ifdef __CLIENT__
String Map::objectsDescription() const {
    String str;
    QTextStream os(&str);
    if (gx.MobjStateAsInfo) {
        thinkers().forAll(0x3, [&os] (thinker_t *th) {
            if (Thinker_IsMobj(th)) {
                os << gx.MobjStateAsInfo(reinterpret_cast<mobj_t const *>(th));
            }
            return LoopContinue;
        });
    }
    return str;
}
void Map::restoreObjects(Info const &objState, IThinkerMapping const &thinkerMapping) const {
    LOG_AS("Map::restoreObjects");
    if (!gx.MobjStateAsInfo || !gx.MobjRestoreState) return;
    bool problemsDetected = false;
    QList<thinker_t const *> mobjs;
    thinkers().forAll(0x3, [&mobjs] (thinker_t *th) {
        if (Thinker_IsMobj(th)) mobjs << th;
        return LoopContinue;
    });
    if (objState.root().contents().size() != mobjs.size()) {
        LOGDEV_MAP_WARNING("Different number of objects: %i in map, but got %i in restore data")
                        << mobjs.size()
                        << objState.root().contents().size();
    }
    for (auto i  = objState.root().contentsInOrder().begin();
              i != objState.root().contentsInOrder().end();
            ++i) {
        Info::BlockElement const &state = (*i)->as<Info::BlockElement>();
        Id::Type const privateId = state.name().toUInt32();
        DENG2_ASSERT(privateId != 0);
        if (thinker_t *th = thinkerMapping.thinkerForPrivateId(privateId)) {
            if (ThinkerData *found = ThinkerData::find(privateId)) {
                DENG2_ASSERT(&found->thinker() == th);
                gx.MobjRestoreState(found->as<MobjThinkerData>().mobj(), state); {
                    Info const currentDesc(gx.MobjStateAsInfo(found->as<MobjThinkerData>().mobj()));
                    Info::BlockElement const &currentState = currentDesc.root().contentsInOrder()
                            .first()->as<Info::BlockElement>();
                    DENG2_ASSERT(currentState.name() == state.name());
                    foreach (String const &key, state.contents().keys()) {
                        if (state.keyValue(key).text != currentState.keyValue(key).text) {
                            problemsDetected = true;
                            const String msg = String("Object %1 has mismatching '%2' (current:%3 != arch:%4)")
                                    .arg(privateId)
                                    .arg(key)
                                    .arg(currentState.keyValue(key).text)
                                    .arg(state.keyValue(key).text);
                            LOGDEV_MAP_WARNING("%s") << msg;
                        }
                    }
                }
            }
            else {
                LOGDEV_MAP_ERROR("Map does not have a thinker matching ID 0x%x")
                        << privateId;
            }
        }
        else {
            LOGDEV_MAP_ERROR("Thinker mapping does not have a thinker matching ID 0x%x")
                    << privateId;
        }
    }
    if (problemsDetected) {
        LOG_MAP_WARNING("Map objects were not fully restored " DENG2_CHAR_MDASH
                        " gameplay may be affected (enable Developer log entries for details)");
    }
    else {
        LOGDEV_MAP_MSG("State of map objects has been restored");
    }
}
void Map::serializeInternalState(Writer &to) const {
    BaseMap::serializeInternalState(to);
    thinkers().forAll(0x3, [&to] (thinker_t *th) {
        if (th->d) {
            ThinkerData const &thinkerData = THINKER_DATA(*th, ThinkerData);
            if (ISerializable const *serial = THINKER_DATA_MAYBE(*th, ISerializable)) {
                to << thinkerData.id()
                   << Writer::BeginSpan
                   << *serial
                   << Writer::EndSpan;
            }
        }
        return LoopContinue;
    });
    to << Id(Id::None);
}
void Map::deserializeInternalState(Reader &from, IThinkerMapping const &thinkerMapping) {
    BaseMap::deserializeInternalState(from, thinkerMapping);
    try {
        forever {
            Id id { Id::None };
            from >> id;
            if (!id) break; // Zero ID terminates the sequence.
            duint32 size = 0;
            from >> size;
            auto const nextOffset = from.offset() + size;
            try {
                if (thinker_t *th = thinkerMapping.thinkerForPrivateId(id)) {
                    Thinker_InitPrivateData(th, id);
                    if (ISerializable *serial = THINKER_DATA_MAYBE(*th, ISerializable)) {
                        from >> *serial;
                    }
                    else {
                        LOG_MAP_WARNING("State for thinker %i is not deserializable "
                                        DENG2_CHAR_MDASH " internal representation may have "
                                        "changed, or save data is corrupt") << id;
                    }
                }
            }
            catch (Error const &er) {
                LOG_MAP_WARNING("Error when reading state of object %s: %s")
                        << id << er.asText();
            }
            from.setOffset(nextOffset);
        }
    }
    catch (Error const &er) {
        LOG_MAP_WARNING("Error when reading state: %s") << er.asText();
    }
}
void Map::redecorate() {
    forAllSectors([](Sector &sector) {
        sector.forAllSubsectors([](Subsector &subsec) {
            subsec.as<ClientSubsector>().markForDecorationUpdate();
            return LoopContinue;
        });
        return LoopContinue;
    });
}
void Map::worldSystemFrameBegins(bool resetNextViewer) {
    DENG2_ASSERT(&App_World().map() == this); // Sanity check.
    d->lerpTrackedPlanes(resetNextViewer);
    d->lerpScrollingSurfaces(resetNextViewer);
    if (!freezeRLs) {
#if 0
        initLightGrid();
        d->biasBeginFrame();
#endif
        removeAllLumobjs();
        d->removeAllContacts();
        if (useLightDecorations) {
            for (Sector *sector : d->sectors) {
                sector->forAllSubsectors([] (Subsector &ssec) {
                    auto &clSubsector = ssec.as<ClientSubsector>();
                    clSubsector.decorate();
                    clSubsector.generateLumobjs();
                    return LoopContinue;
                });
            }
        }
        if (useDynLights) {
            for (Sector *sector : d->sectors)
            for (mobj_t *iter = sector->firstMobj(); iter; iter = iter->sNext) {
                Mobj_GenerateLumobjs(iter);
            }
        }
        d->generateMobjContacts();
        d->linkAllParticles();
        d->linkAllContacts();
    }
}
static dint expireClMobjsWorker(mobj_t *mob, void *context) {
    duint const nowTime = *static_cast<duint *>(context);
    if (mob->thinker.function == (thinkfunc_t)-1)
        return 0;
    if (mob->dPlayer) return 0;
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mob);
    DENG2_ASSERT(info);
    if ((info->flags & (CLMF_UNPREDICTABLE | CLMF_HIDDEN | CLMF_NULLED)) || !mob->info) {
        if (nowTime - info->time > CLMOBJ_TIMEOUT) {
            LOGDEV_MAP_VERBOSE("Mobj %i has expired (%i << %i), in state %s [%c%c%c]")
                << mob->thinker.id
                << info->time << nowTime
                << Def_GetStateName(mob->state)
                << (info->flags & CLMF_UNPREDICTABLE? 'U' : '_')
                << (info->flags & CLMF_HIDDEN?        'H' : '_')
                << (info->flags & CLMF_NULLED?        '0' : '_');
            Mobj_Destroy(mob);
        }
    }
    return 0;
}
void Map::expireClMobjs() {
    duint nowTime = Timer_RealMilliseconds();
    clMobjIterator(expireClMobjsWorker, &nowTime);
}
#endif // __CLIENT__
String Map::elementSummaryAsStyledText() const {
#define TABBED(count, label) String(_E(Ta) "  %1 " _E(Tb) "%2\n").arg(count).arg(label)
    String str;
    QTextStream os(&str);
    if (lineCount())    os << TABBED(lineCount(),    "Lines");
    if (sectorCount())  os << TABBED(sectorCount(),  "Sectors");
    if (vertexCount())  os << TABBED(vertexCount(),  "Vertexes");
    if (polyobjCount()) os << TABBED(polyobjCount(), "Polyobjs");
    return str.rightStrip();
#undef TABBED
}
String Map::objectSummaryAsStyledText() const {
#define TABBED(count, label) String(_E(Ta) "  %1 " _E(Tb) "%2\n").arg(count).arg(label)
    dint thCountInStasis = 0;
    dint thCount = thinkers().count(&thCountInStasis);
    String str;
    QTextStream os(&str);
    if (thCount)           os << TABBED(thCount,            String("Thinkers (%1 in stasis)").arg(thCountInStasis));
#ifdef __CLIENT__
    if (generatorCount())  os << TABBED(generatorCount(),   "Generators");
    if (lumobjCount())     os << TABBED(lumobjCount(),      "Lumobjs");
#endif
    return str.rightStrip();
#undef TABBED
}
D_CMD(InspectMap) {
    DENG2_UNUSED3(src, argc, argv);
    LOG_AS("inspectmap (Cmd)");
    if (!App_World().hasMap()) {
        LOG_SCR_WARNING("No map is currently loaded");
        return false;
    }
    Map &map = App_World().map();
    LOG_SCR_NOTE(_E(b) "%s - %s")
            << Con_GetString("map-name")
            << Con_GetString("map-author");
    LOG_SCR_MSG("\n");
    LOG_SCR_MSG(    _E(l) "Uri: "    _E(.) _E(i) "%s" _E(.)
              /*" " _E(l) " OldUid: " _E(.) _E(i) "%s" _E(.)*/
                    _E(l) " Music: "  _E(.) _E(i) "%i")
            << (map.hasManifest()? map.manifest().composeUri().asText() : "(unknown map)")
            /*<< map.oldUniqueId()*/
            << Con_GetInteger("map-music");
    if (map.hasManifest() && map.manifest().sourceFile()->hasCustom()) {
        LOG_SCR_MSG(_E(l) "Source: " _E(.) _E(i) "\"%s\"")
                << NativePath(map.manifest().sourceFile()->composePath()).pretty();
    }
    LOG_SCR_MSG("\n");
    if (map.isEditable()) {
        LOG_MSG(_E(D) "Editing " _E(b) "Enabled");
    }
    LOG_SCR_MSG(_E(D) "Elements:");
    LOG_SCR_MSG("%s") << map.elementSummaryAsStyledText();
    if (map.thinkers().isInited()) {
        LOG_SCR_MSG(_E(D) "Objects:");
        LOG_SCR_MSG("%s") << map.objectSummaryAsStyledText();
    }
    LOG_SCR_MSG(_E(R) "\n");
    auto geometryDimensions = Vector2d(map.bounds().max) - Vector2d(map.bounds().min);
    LOG_SCR_MSG(_E(l) "Geometry dimensions: " _E(.) _E(i)) << geometryDimensions.asText();
    if (map.hasBspTree()) {
        LOG_SCR_MSG(_E(l) "BSP: " _E(.) _E(i)) << map.bspTree().summary();
    }
    if (!map.subspaceBlockmap().isNull()) {
        LOG_SCR_MSG(_E(l) "Subspace blockmap: " _E(.) _E(i)) << map.subspaceBlockmap().dimensions().asText();
    }
    if (!map.lineBlockmap().isNull()) {
        LOG_SCR_MSG(_E(l) "Line blockmap: "     _E(.) _E(i)) << map.lineBlockmap().dimensions().asText();
    }
    if (!map.mobjBlockmap().isNull()) {
        LOG_SCR_MSG(_E(l) "Mobj blockmap: "     _E(.) _E(i)) << map.mobjBlockmap().dimensions().asText();
    }
    if (!map.polyobjBlockmap().isNull()) {
        LOG_SCR_MSG(_E(l) "Polyobj blockmap: "  _E(.) _E(i)) << map.polyobjBlockmap().dimensions().asText();
    }
#if 0
#ifdef __CLIENT__
    if (map.hasLightGrid()) {
        LOG_SCR_MSG(_E(l) "LightGrid: " _E(.) _E(i)) << map.lightGrid().dimensions().asText();
    }
#endif
#endif // 0
    return true;
#undef TABBED
}
void Map::consoleRegister() // static
{
    Line::consoleRegister();
    Mobj_ConsoleRegister();
    Sector::consoleRegister();
    C_VAR_INT("bsp-factor",                 &bspSplitFactor, CVF_NO_MAX, 0, 0);
#if 0
#ifdef __CLIENT__
    C_VAR_INT("rend-bias-grid-multisample", &lgMXSample,     0, 0, 7);
#endif
#endif
    C_CMD("inspectmap", "", InspectMap);
}
static Vertex *rootVtx;
/**
 * Compares the angles of two lines that share a common vertex.
 *
 * pre: rootVtx must point to the vertex common between a and b
 *      which are (lineowner_t*) ptrs.
 */
static dint lineAngleSorter(void const *a, void const *b) {
    binangle_t angles[2];
    LineOwner *own[2] = { (LineOwner *)a, (LineOwner *)b };
    for (duint i = 0; i < 2; ++i) {
        if (own[i]->_link[Anticlockwise]) // We have a cached result.
        {
            angles[i] = own[i]->angle();
        }
        else {
            Line *line = &own[i]->line();
            Vertex const &otherVtx = line->vertex(&line->from() == rootVtx? 1:0);
            fixed_t dx = otherVtx.origin().x - rootVtx->origin().x;
            fixed_t dy = otherVtx.origin().y - rootVtx->origin().y;
            own[i]->_angle = angles[i] = bamsAtan2(-100 *dx, 100 * dy);
            own[i]->_link[Anticlockwise] = (LineOwner *) 1;
        }
    }
    return (angles[1] - angles[0]);
}
/**
 * Merge left and right line owner lists into a new list.
 *
 * @return  The newly merged list.
 */
static LineOwner *mergeLineOwners(LineOwner *left, LineOwner *right,
    dint (*compare) (void const *a, void const *b)) {
    LineOwner tmp;
    LineOwner *np = &tmp;
    tmp._link[Clockwise] = np;
    while (left && right) {
        if (compare(left, right) <= 0) {
            np->_link[Clockwise] = left;
            np = left;
            left = left->next();
        }
        else {
            np->_link[Clockwise] = right;
            np = right;
            right = right->next();
        }
    }
    if (left) {
        np->_link[Clockwise] = left;
    }
    if (right) {
        np->_link[Clockwise] = right;
    }
    if (!tmp.hasNext())
        return nullptr;
    return tmp.next();
}
static LineOwner *splitLineOwners(LineOwner *list) {
    if (!list) return nullptr;
    LineOwner *lista = list;
    LineOwner *listb = list;
    LineOwner *listc = list;
    do {
        listc = listb;
        listb = listb->next();
        lista = lista->next();
        if (lista) {
            lista = lista->next();
        }
    } while (lista);
    listc->_link[Clockwise] = nullptr;
    return listb;
}
/**
 * This routine uses a recursive mergesort algorithm; O(NlogN)
 */
static LineOwner *sortLineOwners(LineOwner *list,
    dint (*compare) (void const *a, void const *b)) {
    if (list && list->next()) {
        LineOwner *p = splitLineOwners(list);
        list = mergeLineOwners(sortLineOwners(list, compare),
                               sortLineOwners(p, compare), compare);
    }
    return list;
}
static void setVertexLineOwner(Vertex *vtx, Line *lineptr, LineOwner **storage) {
    if (!lineptr) return;
    LineOwner const *own = vtx->firstLineOwner();
    while (own) {
        if (&own->line() == lineptr)
            return;  // Yes, we can exit.
        own = own->next();
    }
    vtx->_numLineOwners++;
    LineOwner *newOwner = (*storage)++;
    newOwner->_line = lineptr;
    newOwner->_link[Anticlockwise] = nullptr;
    newOwner->_link[Clockwise] = vtx->_lineOwners;
    vtx->_lineOwners = newOwner;
    if (vtx == &lineptr->from())
        lineptr->_vo1 = newOwner;
    else
        lineptr->_vo2 = newOwner;
}
#ifdef DENG2_DEBUG
/**
 * Determines whether the specified vertex @a v has a correctly formed line owner ring.
 */
static bool vertexHasValidLineOwnerRing(Vertex &v) {
    LineOwner const *base = v.firstLineOwner();
    LineOwner const *cur = base;
    do {
        if (cur->prev()->next() != cur) return false;
        if (cur->next()->prev() != cur) return false;
    } while ((cur = cur->next()) != base);
    return true;
}
#endif
/**
 * Generates the line owner rings for each vertex. Each ring includes all the lines which
 * the vertex belongs to sorted by angle, (the rings are arranged in clockwise order, east = 0).
 */
void buildVertexLineOwnerRings(QList<Vertex *> const &vertexs, QList<Line *> &editableLines) {
    LOG_AS("buildVertexLineOwnerRings");
    auto *lineOwners = (LineOwner *) Z_Malloc(sizeof(LineOwner) * editableLines.count() * 2, PU_MAPSTATIC, 0);
    LineOwner *allocator = lineOwners;
    for (Line *line : editableLines)
    for (dint p = 0; p < 2; ++p) {
        setVertexLineOwner(&line->vertex(p), line, &allocator);
    }
    for (Vertex *v : vertexs) {
        if (!v->_numLineOwners) continue;
        rootVtx = v;
        v->_lineOwners = sortLineOwners(v->_lineOwners, lineAngleSorter);
        binangle_t firstAngle = v->_lineOwners->angle();
        LineOwner *last = v->_lineOwners;
        LineOwner *p = last->next();
        while (p) {
            p->_link[Anticlockwise] = last;
            last->_angle = last->angle() - p->angle();
            last = p;
            p = p->next();
        }
        last->_link[Clockwise] = v->_lineOwners;
        v->_lineOwners->_link[Anticlockwise] = last;
        last->_angle = last->angle() - firstAngle;
/*#ifdef DENG2_DEBUG
        LOG_MAP_VERBOSE("Vertex #%i: line owners #%i")
            << editmap.vertexes.indexOf(v) << v->lineOwnerCount();
        LineOwner const *base = v->firstLineOwner();
        LineOwner const *cur = base;
        duint idx = 0;
        do {
            LOG_MAP_VERBOSE("  %i: p= #%05i this= #%05i n= #%05i, dANG= %-3.f")
                << idx << cur->prev().line().indexInMap() << cur->line().indexInMap()
                << cur->next().line().indexInMap() << BANG2DEG(cur->angle());
            idx++;
        } while ((cur = &cur->next()) != base);
#endif*/
        DENG2_ASSERT(vertexHasValidLineOwnerRing(*v));
    }
}
bool Map::isEditable() const {
    return d->editingEnabled;
}
struct VertexInfo {
    Vertex *vertex = nullptr;  ///< Vertex for this info.
    Vertex *equiv  = nullptr;  ///< Determined equivalent vertex.
    duint refCount = 0;        ///< Line -> Vertex reference count.
    dint compareVertexOrigins(VertexInfo const &other) const {
        DENG2_ASSERT(vertex && other.vertex);
        if (this == &other) return 0;
        if (vertex == other.vertex) return 0;
        if (dint(vertex->origin().x) != dint(other.vertex->origin().x)) {
            return dint(vertex->origin().x) - dint(other.vertex->origin().x);
        }
        return dint(vertex->origin().y) - dint(other.vertex->origin().y);
    }
    bool operator < (VertexInfo const &other) const {
        return compareVertexOrigins(other) < 0;
    }
};
void pruneVertexes(Mesh &mesh, Map::Lines const &lines) {
    QVector<VertexInfo> vertexInfo(mesh.vertexCount());
    dint ord = 0;
    for (Vertex *vertex : mesh.vertexs()) {
        vertexInfo[ord++].vertex = vertex;
    } {
        QVector<VertexInfo> sortedInfo(vertexInfo);
        qSort(sortedInfo.begin(), sortedInfo.end());
        for (dint i = 0; i < sortedInfo.count() - 1; ++i) {
            VertexInfo &a = sortedInfo[i];
            VertexInfo &b = sortedInfo[i + 1];
            if (a.compareVertexOrigins(b) == 0) {
                b.equiv = (a.equiv? a.equiv : a.vertex);
            }
        }
    }
    for (Line *line : lines) {
        vertexInfo[line->from().indexInMap()].refCount++;
        vertexInfo[  line->to().indexInMap()].refCount++;
    }
    for (Line *line : lines) {
        while (vertexInfo[line->from().indexInMap()].equiv) {
            VertexInfo &info = vertexInfo[line->from().indexInMap()];
            info.refCount--;
            line->replaceVertex(Line::From, *info.equiv);
            vertexInfo[line->from().indexInMap()].refCount++;
        }
        while (vertexInfo[line->to().indexInMap()].equiv) {
            VertexInfo &info = vertexInfo[line->to().indexInMap()];
            info.refCount--;
            line->replaceVertex(Line::To, *info.equiv);
            vertexInfo[line->to().indexInMap()].refCount++;
        }
    }
    dint prunedCount = 0, numUnused = 0;
    for (VertexInfo const &info : vertexInfo) {
        Vertex *vertex = info.vertex;
        if (info.refCount) continue;
        mesh.removeVertex(*vertex);
        prunedCount += 1;
        if (!info.equiv) numUnused += 1;
    }
    if (prunedCount) {
        dint ord = 0;
        for (Vertex *vertex : mesh.vertexs()) {
            vertex->setIndexInMap(ord++);
        }
        LOGDEV_MAP_NOTE("Pruned %d vertexes (%d equivalents, %d unused)")
            << prunedCount << (prunedCount - numUnused) << numUnused;
    }
}
enum {
    LinkFloorBit           = 0x1,
    LinkCeilingBit         = 0x2,
    FlatBleedingFloorBit   = 0x4,
    FlatBleedingCeilingBit = 0x8,
    InvisibleFloorBit      = 0x10,
    InvisibleCeilingBit    = 0x20,
};
bool Map::endEditing() {
    if (!d->editingEnabled) return true; // Huh?
    d->editingEnabled = false;
    LOG_AS("Map");
    LOG_MAP_VERBOSE("Editing ended");
    LOGDEV_MAP_VERBOSE("New elements: %d Vertexes, %d Lines, %d Polyobjs and %d Sectors")
        << d->mesh.vertexCount()        << d->editable.lines.count()
        << d->editable.polyobjs.count() << d->editable.sectors.count();
    pruneVertexes(d->mesh, d->editable.lines);
    for (Line *line : d->editable.lines) {
        if (!line->front().hasSector() || !line->back().hasSector())
            line->setFlags(DDLF_BLOCKING);
    }
    buildVertexLineOwnerRings(d->mesh.vertexs(), d->editable.lines);
    DENG2_ASSERT(d->sectors.isEmpty());
    d->sectors.reserve(d->editable.sectors.count());
    d->sectors.append(d->editable.sectors);
    d->editable.sectors.clear();
    DENG2_ASSERT(d->lines.isEmpty());
    d->lines.reserve(d->editable.lines.count());
    d->lines.append(d->editable.lines);
    d->editable.lines.clear();
    DENG2_ASSERT(d->polyobjs.isEmpty());
    d->polyobjs.reserve(d->editable.polyobjs.count());
    while (!d->editable.polyobjs.isEmpty()) {
        d->polyobjs.append(d->editable.polyobjs.takeFirst());
        Polyobj *polyobj = d->polyobjs.back();
        for (Line *line : polyobj->lines()) {
            HEdge *hedge = polyobj->mesh().newHEdge(line->from());
            hedge->setTwin(polyobj->mesh().newHEdge(line->to()));
            hedge->twin().setTwin(hedge);
            LineSideSegment *seg = line->front().addSegment(*hedge);
#ifdef __CLIENT__
            seg->setLength(line->length());
#else
            DENG2_UNUSED(seg);
#endif
        }
        polyobj->buildUniqueVertexes();
        polyobj->updateOriginalVertexCoords();
    }
    d->updateBounds();
    LOG_MAP_VERBOSE("Geometry bounds:") << Rectangled(d->bounds.min, d->bounds.max).asText();
    d->initLineBlockmap();
    if (!d->buildBspTree())
        return false;
    d->initMobjBlockmap();
    d->initPolyobjBlockmap();
    for (Line *line : d->lines) {
        line->forAllSides([] (Line::Side &side) {
            side.updateAllSurfaceNormals();
            return LoopContinue;
        });
    }
    std::map<int, Sector *> sectorsByArchiveIndex;
    for (Sector *sector : d->sectors) {
        sectorsByArchiveIndex[sector->indexInArchive()] = sector;
        d->buildSubsectors(*sector);
        sector->buildSides();
        sector->chainSoundEmitters();
    }
    for (Sector *sector : d->sectors) {
#if defined(__CLIENT__)
        if (sector->visPlaneLinkTargetSector() != MapElement::NoIndex) {
            if (Sector *target = sectorsByArchiveIndex[sector->visPlaneLinkTargetSector()]) {
                auto &targetSub = target->subsector(0).as<ClientSubsector>();
                int linkModes[2]{};
                if (sector->visPlaneBits() & FlatBleedingFloorBit) {
                    linkModes[Sector::Floor] |= ClientSubsector::LinkWhenLowerThanTarget;
                }
                if (sector->visPlaneBits() & FlatBleedingCeilingBit) {
                    linkModes[Sector::Ceiling] |= ClientSubsector::LinkWhenHigherThanTarget;
                }
                if (sector->visPlaneBits() & InvisibleFloorBit) {
                    linkModes[Sector::Floor] |= ClientSubsector::LinkWhenHigherThanTarget;
                }
                if (sector->visPlaneBits() & InvisibleCeilingBit) {
                    linkModes[Sector::Ceiling] |= ClientSubsector::LinkWhenLowerThanTarget;
                }
                for (auto &lm : linkModes) {
                    if (lm == 0) lm = ClientSubsector::LinkAlways;
                }
                sector->forAllSubsectors([&targetSub, sector, linkModes](Subsector &sub) {
                    auto &clsub = sub.as<ClientSubsector>();
                    for (int plane = 0; plane < 2; ++plane) {
                        if (sector->isVisPlaneLinked(plane)) {
                            clsub.linkVisPlane(plane,
                                               targetSub,
                                               ClientSubsector::VisPlaneLinkMode(linkModes[plane]));
                        }
                    }
                    return LoopContinue;
                });
            }
        }
#endif
        sector->forAllPlanes([] (Plane &plane) {
            plane.updateSoundEmitterOrigin();
            return LoopContinue;
        });
    }
    d->initSubspaceBlockmap();
    d->thinkers.reset(new Thinkers);
    return true;
}
Vertex *Map::createVertex(Vector2d const &origin, dint archiveIndex) {
    if (!d->editingEnabled)
        throw EditError("Map::createVertex", "Editing is not enabled");
    Vertex *vtx = d->mesh.newVertex(origin);
    vtx->setMap(this);
    vtx->setIndexInArchive(archiveIndex);
    vtx->setIndexInMap(d->mesh.vertexCount() - 1);
    return vtx;
}
Line *Map::createLine(Vertex &v1, Vertex &v2, int flags, Sector *frontSector,
                      Sector *backSector, int archiveIndex) {
    if (!d->editingEnabled)
        throw EditError("Map::createLine", "Editing is not enabled");
    auto *line = new Line(v1, v2, flags, frontSector, backSector);
    d->editable.lines.append(line);
    line->setMap(this);
    line->setIndexInArchive(archiveIndex);
    line->setIndexInMap(d->editable.lines.count() - 1);
    line->front().setIndexInMap(Map::toSideIndex(line->indexInMap(), Line::Front));
    line->back ().setIndexInMap(Map::toSideIndex(line->indexInMap(), Line::Back));
    return line;
}
Sector *Map::createSector(float lightLevel, const Vector3f &lightColor, int archiveIndex,
                          const struct de_api_sector_hacks_s *hacks) {
    if (!d->editingEnabled)
        throw EditError("Map::createSector", "Editing is not enabled");
    auto *sector = new Sector(lightLevel, lightColor);
    d->editable.sectors.append(sector);
    sector->setMap(this);
    sector->setIndexInArchive(archiveIndex);
    if (hacks) {
        int linkFlags = 0;
        if (hacks->flags.linkFloorPlane)   linkFlags |= LinkFloorBit;
        if (hacks->flags.linkCeilingPlane) linkFlags |= LinkCeilingBit;
        if (hacks->flags.missingInsideBottom)  linkFlags |= FlatBleedingFloorBit;
        if (hacks->flags.missingInsideTop)     linkFlags |= FlatBleedingCeilingBit;
        if (hacks->flags.missingOutsideBottom) linkFlags |= InvisibleFloorBit;
        if (hacks->flags.missingOutsideTop)    linkFlags |= InvisibleCeilingBit;
        sector->setVisPlaneLinks(hacks->visPlaneLinkTargetSector, linkFlags);
    }
    sector->setIndexInMap(d->editable.sectors.count() - 1);
    return sector;
}
Polyobj *Map::createPolyobj(Vector2d const &origin) {
    if (!d->editingEnabled)
        throw EditError("Map::createPolyobj", "Editing is not enabled");
    void *region = M_Calloc(gx.GetInteger(DD_POLYOBJ_SIZE));
    auto *pob = new (region) Polyobj(origin);
    d->editable.polyobjs.append(pob);
    pob->setIndexInMap(d->editable.polyobjs.count() - 1);
    return pob;
}
Map::Lines const &Map::editableLines() const {
    if (!d->editingEnabled)
        throw EditError("Map::editableLines", "Editing is not enabled");
    return d->editable.lines;
}
Map::Sectors const &Map::editableSectors() const {
    if (!d->editingEnabled)
        throw EditError("Map::editableSectors", "Editing is not enabled");
    return d->editable.sectors;
}
void Map::initMapOutlinePacket(shell::MapOutlinePacket &packet) {
    forAllLines([&packet] (Line &line) {
        packet.addLine(line.from().origin().toVector2i(),
                       line.to  ().origin().toVector2i(),
                       (line.front().hasSector() && line.back().hasSector()) ?
                           shell::MapOutlinePacket::TwoSidedLine : shell::MapOutlinePacket::OneSidedLine);
        return LoopContinue;
    });
}
Map::Polyobjs const &Map::editablePolyobjs() const {
    if (!d->editingEnabled)
        throw EditError("Map::editablePolyobjs", "Editing is not enabled");
    return d->editable.polyobjs;
}
}  // namespace world
lineopening_s::lineopening_s(Line const &line) {
    if(!line.back().hasSector()) {
        top = bottom = range = lowFloor = 0;
        return;
    }
    Sector const *frontSector = line.front().sectorPtr();
    Sector const *backSector  = line.back().sectorPtr();
    DENG_ASSERT(frontSector != 0);
    if(backSector && backSector->ceiling().height() < frontSector->ceiling().height()) {
        top = backSector->ceiling().height();
    }
    else {
        top = frontSector->ceiling().height();
    }
    if(backSector && backSector->floor().height() > frontSector->floor().height()) {
        bottom = backSector->floor().height();
    }
    else {
        bottom = frontSector->floor().height();
    }
    range = top - bottom;
    if(backSector && frontSector->floor().height() > backSector->floor().height()) {
        lowFloor = float( backSector->floor().height() );
    }
    else {
        lowFloor = float( frontSector->floor().height() );
    }
}
lineopening_s &lineopening_s::operator = (lineopening_s const &other) {
    top      = other.top;
    bottom   = other.bottom;
    range    = other.range;
    lowFloor = other.lowFloor;
    return *this;
}
#ifdef __CLIENT__
/**
 * Same as @ref R_OpenRange() but works with the "visual" (i.e., smoothed) plane
 * height coordinates rather than the "sharp" coordinates.
 *
 * @param side      Line side to find the open range for.
 *
 * Return values:
 * @param bottom    Bottom Z height is written here. Can be @c nullptr.
 * @param top       Top Z height is written here. Can be @c nullptr.
 *
 * @return Height of the open range.
 *
 * @todo fixme: Should use the visual plane heights of subsectors.
 */
static coord_t visOpenRange(LineSide const &side, coord_t *retBottom = nullptr, coord_t *retTop = nullptr) {
    Sector const *frontSec = side.sectorPtr();
    Sector const *backSec  = side.back().sectorPtr();
    coord_t bottom;
    if(backSec && backSec->floor().heightSmoothed() > frontSec->floor().heightSmoothed()) {
        bottom = backSec->floor().heightSmoothed();
    }
    else {
        bottom = frontSec->floor().heightSmoothed();
    }
    coord_t top;
    if(backSec && backSec->ceiling().heightSmoothed() < frontSec->ceiling().heightSmoothed()) {
        top = backSec->ceiling().heightSmoothed();
    }
    else {
        top = frontSec->ceiling().heightSmoothed();
    }
    if(retBottom) *retBottom = bottom;
    if(retTop)    *retTop    = top;
    return top - bottom;
}
bool R_SideBackClosed(LineSide const &side, bool ignoreOpacity) {
    if(!side.hasSections()) return false;
    if(!side.hasSector()) return false;
    if(side.line().isSelfReferencing()) return false; // Never.
    if(side.considerOneSided()) return true;
    Sector const &frontSec = side.sector();
    Sector const &backSec  = side.back().sector();
    if(backSec.floor().heightSmoothed()   >= backSec.ceiling().heightSmoothed())  return true;
    if(backSec.ceiling().heightSmoothed() <= frontSec.floor().heightSmoothed())   return true;
    if(backSec.floor().heightSmoothed()   >= frontSec.ceiling().heightSmoothed()) return true;
    if (MaterialAnimator *matAnimator = side.middle().materialAnimator()) {
        matAnimator->prepare();
        if(ignoreOpacity || (matAnimator->isOpaque() && !side.middle().blendMode() && side.middle().opacity() >= 1)) {
            if(side.isFlagged(SDF_MIDDLE_STRETCH))
                return true;
            if(side.leftHEdge()) // possibility of degenerate BSP leaf
            {
                coord_t openRange, openBottom, openTop;
                openRange = visOpenRange(side, &openBottom, &openTop);
                if(matAnimator->dimensions().y >= openRange) {
                    WallEdge edge(WallSpec::fromMapSide(side, LineSide::Middle),
                                  *side.leftHEdge(), Line::From);
                    return (edge.isValid() && edge.top().z() > edge.bottom().z()
                            && edge.top().z() >= openTop && edge.bottom().z() <= openBottom);
                }
            }
        }
    }
    return false;
}
Line *R_FindLineNeighbor(Line const &line, LineOwner const &own, ClockDirection direction,
    Sector const *sector, binangle_t *diff) {
    LineOwner const *cown = (direction == Anticlockwise ? own.prev() : own.next());
    Line *other = &cown->line();
    if(other == &line)
        return nullptr;
    if(diff) {
        *diff += (direction == Anticlockwise ? cown->angle() : own.angle());
    }
    if(!other->back().hasSector() || !other->isSelfReferencing()) {
        if(sector)  // Must one of the sectors match?
        {
            if(other->front().sectorPtr() == sector ||
               (other->back().hasSector() && other->back().sectorPtr() == sector))
                return other;
        }
        else {
            return other;
        }
    }
    DENG2_ASSERT(cown);
    return R_FindLineNeighbor(line, *cown, direction, sector, diff);
}
/**
 * @param side  Line side for which to determine covered opening status.
 *
 * @return  @c true iff there is a "middle" material on @a side which
 * completely covers the open range.
 *
 * @todo fixme: Should use the visual plane heights of subsectors.
 */
static bool middleMaterialCoversOpening(LineSide const &side) {
    if(!side.hasSector()) return false; // Never.
    if(!side.hasSections()) return false;
    if(side.isFlagged(SDF_MIDDLE_STRETCH))
        return true;
    MaterialAnimator *matAnimator = side.middle().materialAnimator();
    if (!matAnimator) return false;
    matAnimator->prepare();
    if(matAnimator->isOpaque() && !side.middle().blendMode() && side.middle().opacity() >= 1) {
        if(side.leftHEdge())  // possibility of degenerate BSP leaf.
        {
            coord_t openRange, openBottom, openTop;
            openRange = visOpenRange(side, &openBottom, &openTop);
            if(matAnimator->dimensions().y >= openRange) {
                WallEdge edge(WallSpec::fromMapSide(side, LineSide::Middle), *side.leftHEdge(), Line::From);
                return (edge.isValid()
                        && edge.top   ().z() > edge.bottom().z()
                        && edge.top   ().z() >= openTop
                        && edge.bottom().z() <= openBottom);
            }
        }
    }
    return false;
}
Line *R_FindSolidLineNeighbor(Line const &line, LineOwner const &own, ClockDirection direction,
    Sector const *sector, binangle_t *diff) {
    DENG2_ASSERT(sector);
    LineOwner const *cown = (direction == Anticlockwise ? own.prev() : own.next());
    Line *other = &cown->line();
    if (other == &line) return nullptr;
    if (diff) {
        *diff += (direction == Anticlockwise ? cown->angle() : own.angle());
    }
    if (!((other->isBspWindow()) && other->front().sectorPtr() != sector)
        && !other->isSelfReferencing()) {
        if (!other->front().hasSector() || !other->back().hasSector())
            return other;
        if (   other->front().sector().floor  ().heightSmoothed() >= sector->ceiling().heightSmoothed()
            || other->front().sector().ceiling().heightSmoothed() <= sector->floor  ().heightSmoothed()
            || other->back().sector().floor  ().heightSmoothed() >= sector->ceiling().heightSmoothed()
            || other->back().sector().ceiling().heightSmoothed() <= sector->floor  ().heightSmoothed()
            || other->back().sector().ceiling().heightSmoothed() <= other->back().sector().floor().heightSmoothed())
            return other;
        LineSide &otherSide = other->side(other->front().sectorPtr() == sector ? Line::Front : Line::Back);
        if (middleMaterialCoversOpening(otherSide))
            return other;
    }
    DENG2_ASSERT(cown);
    return R_FindSolidLineNeighbor(line, *cown, direction, sector, diff);
}
#endif  // __CLIENT__
static String const VAR_MATERIAL("material");
static mobj_t *unusedMobjs;
/*
 * Console variables:
 */
dint useSRVO      = 2;  ///< @c 1= models only, @c 2= sprites + models
dint useSRVOAngle = 1;
#ifdef __CLIENT__
static byte mobjAutoLights = true;
#endif
/**
 * Called during map loading.
 */
void P_InitUnusedMobjList() {
    ::unusedMobjs = nullptr;
}
/**
 * All mobjs must be allocated through this routine. Part of the public API.
 */
mobj_t *P_MobjCreate(thinkfunc_t function, Vector3d const &origin, angle_t angle,
    coord_t radius, coord_t height, dint ddflags) {
    if (!function)
        App_Error("P_MobjCreate: Think function invalid, cannot create mobj.");
#ifdef DENG2_DEBUG
    if (::isClient) {
        LOG_VERBOSE("P_MobjCreate: Client creating mobj at %s")
            << origin.asText();
    }
#endif
    mobj_t *mob;
    if (::unusedMobjs) {
        mob = ::unusedMobjs;
        ::unusedMobjs = ::unusedMobjs->sNext;
    }
    else {
        mob = MobjThinker(Thinker::AllocateMemoryZone).take();
    }
    V3d_Set(mob->origin, origin.x, origin.y, origin.z);
    mob->angle    = angle;
    mob->visAngle = mob->angle >> 16; // "angle-servo"; smooth actor turning.
    mob->radius   = radius;
    mob->height   = height;
    mob->ddFlags  = ddflags;
    mob->lumIdx   = -1;
    mob->thinker.function = function;
    Mobj_Map(*mob).thinkers().add(mob->thinker);
    return mob;
}
/**
 * All mobjs must be destroyed through this routine. Part of the public API.
 *
 * @note Does not actually destroy the mobj. Instead, mobj is marked as
 * awaiting removal (which occurs when its turn for thinking comes around).
 */
#undef Mobj_Destroy
DENG_EXTERN_C void Mobj_Destroy(mobj_t *mo) {
#ifdef _DEBUG
    if (mo->ddFlags & DDMF_MISSILE) {
        LOG_AS("Mobj_Destroy");
        LOG_MAP_XVERBOSE("Destroying missile %i", mo->thinker.id);
    }
#endif
    Mobj_Unlink(mo);
    S_StopSound(0, mo);
    Mobj_Map(*mo).thinkers().remove(reinterpret_cast<thinker_t &>(*mo));
}
/**
 * Called when a mobj is actually removed (when it's thinking turn comes around).
 * The mobj is moved to the unused list to be reused later.
 */
void P_MobjRecycle(mobj_t* mo) {
    MobjThinker::zap(*mo);
    mo->sNext = unusedMobjs;
    unusedMobjs = mo;
}
bool Mobj_IsSectorLinked(mobj_t const &mob) {
    return (mob._bspLeaf != nullptr && mob.sPrev != nullptr);
}
#undef Mobj_SetState
DENG_EXTERN_C void Mobj_SetState(mobj_t *mob, int statenum) {
    if (!mob) return;
    state_t const *oldState = mob->state;
    DENG2_ASSERT(statenum >= 0 && statenum < DED_Definitions()->states.size());
    mob->state  = &runtimeDefs.states[statenum];
    mob->tics   = mob->state->tics;
    mob->sprite = mob->state->sprite;
    mob->frame  = mob->state->frame;
    if (!(mob->ddFlags & DDMF_REMOTE)) {
        String const exec = DED_Definitions()->states[statenum].gets("execute");
        if (!exec.isEmpty()) {
            Con_Execute(CMDS_SCRIPT, exec.toUtf8(), true, false);
        }
    }
    if (!mob->thinker.d) {
        Thinker_InitPrivateData(&mob->thinker);
    }
    if (MobjThinkerData *data = THINKER_DATA_MAYBE(mob->thinker, MobjThinkerData)) {
        data->stateChanged(oldState);
    }
}
Vector3d Mobj_Origin(mobj_t const &mob) {
    return Vector3d(mob.origin);
}
Vector3d Mobj_Center(mobj_t &mob) {
    return Vector3d(mob.origin[0], mob.origin[1], mob.origin[2] + mob.height / 2);
}
dd_bool Mobj_SetOrigin(struct mobj_s *mob, coord_t x, coord_t y, coord_t z) {
    if(!gx.MobjTryMoveXYZ) {
        return false;
    }
    return gx.MobjTryMoveXYZ(mob, x, y, z);
}
#undef Mobj_OriginSmoothed
DENG_EXTERN_C void Mobj_OriginSmoothed(mobj_t *mob, coord_t origin[3]) {
    if (!origin) return;
    V3d_Set(origin, 0, 0, 0);
    if (!mob) return;
    V3d_Copy(origin, mob->origin);
    if (useSRVO && mob->state && mob->tics >= 0) {
        ddouble const mul = mob->tics / dfloat( mob->state->tics );
        vec3d_t srvo;
        V3d_Copy(srvo, mob->srvo);
        V3d_Scale(srvo, mul);
        V3d_Sum(origin, origin, srvo);
    }
#ifdef __CLIENT__
    if (mob->dPlayer) {
        if (P_GetDDPlayerIdx(mob->dPlayer) == consolePlayer
            && mob->dPlayer->mo == mob) {
            viewdata_t const *vd = &DD_Player(consolePlayer)->viewport();
            V3d_Set(origin, vd->current.origin.x, vd->current.origin.y, vd->current.origin.z);
        }
        else if (isClient) {
            Smoother_Evaluate(DD_Player(P_GetDDPlayerIdx(mob->dPlayer))->smoother(), origin);
        }
    }
#endif
}
world::Map &Mobj_Map(mobj_t const &mob) {
    return Thinker_Map(mob.thinker);
}
bool Mobj_IsLinked(mobj_t const &mob) {
    return mob._bspLeaf != 0;
}
BspLeaf &Mobj_BspLeafAtOrigin(mobj_t const &mob) {
    if (Mobj_IsLinked(mob)) {
        return *(BspLeaf *)mob._bspLeaf;
    }
    throw Error("Mobj_BspLeafAtOrigin", "Mobj is not yet linked");
}
bool Mobj_HasSubsector(mobj_t const &mob) {
    if (!Mobj_IsLinked(mob)) return false;
    BspLeaf const &bspLeaf = Mobj_BspLeafAtOrigin(mob);
    if (!bspLeaf.hasSubspace()) return false;
    return bspLeaf.subspace().hasSubsector();
}
Subsector &Mobj_Subsector(mobj_t const &mob) {
    return Mobj_BspLeafAtOrigin(mob).subspace().subsector();
}
Subsector *Mobj_SubsectorPtr(mobj_t const &mob) {
    return Mobj_HasSubsector(mob) ? &Mobj_Subsector(mob) : nullptr;
}
#undef Mobj_Sector
DENG_EXTERN_C Sector *Mobj_Sector(mobj_t const *mob) {
    if (!mob || !Mobj_IsLinked(*mob)) return nullptr;
    return Mobj_BspLeafAtOrigin(*mob).sectorPtr();
}
void Mobj_SpawnParticleGen(mobj_t *mob, ded_ptcgen_t const *def) {
#ifdef __CLIENT__
    DENG2_ASSERT(mob && def);
    Generator *gen = Mobj_Map(*mob).newGenerator();
    if (!gen) return;
    /*LOG_INFO("SpawnPtcGen: %s/%i (src:%s typ:%s mo:%p)")
        << def->state << (def - defs.ptcgens) << defs.states[mob->state-states].id
        << defs.mobjs[mob->type].id << source;*/
    gen->count = def->particles;
    if (def->flags & Generator::ScaledRate) {
        gen->spawnRateMultiplier = Mobj_BspLeafAtOrigin(*mob).sectorPtr()->roughArea() / (128 * 128);
    }
    else {
        gen->spawnRateMultiplier = 1;
    }
    gen->configureFromDef(def);
    gen->source = mob;
    gen->srcid  = mob->thinker.id;
    gen->presimulate(def->preSim);
#else
    DENG2_UNUSED2(mob, def);
#endif
}
#undef Mobj_SpawnDamageParticleGen
DENG_EXTERN_C void Mobj_SpawnDamageParticleGen(mobj_t const *mob, mobj_t const *inflictor, int amount) {
#ifdef __CLIENT__
    if (!mob || !inflictor || amount <= 0) return;
    ded_ptcgen_t const *def = Def_GetDamageGenerator(mob->type);
    if (def) {
        Generator *gen = Mobj_Map(*mob).newGenerator();
        if (!gen) return; // No more generators.
        gen->count = def->particles;
        gen->configureFromDef(def);
        gen->setUntriggered();
        gen->spawnRateMultiplier = de::max(amount, 1);
        gen->originAtSpawn[0] += FLT2FIX(mob->origin[0]);
        gen->originAtSpawn[1] += FLT2FIX(mob->origin[1]);
        gen->originAtSpawn[2] += FLT2FIX(mob->origin[2] + mob->height / 2);
        vec3f_t vecDelta;
        V3f_Set(vecDelta, inflictor->origin[0] - mob->origin[0],
                inflictor->origin[1] - mob->origin[1],
                (inflictor->origin[2] - inflictor->height / 2) - (mob->origin[2] + mob->height / 2));
        vec3f_t vector;
        V3f_SetFixed(vector, gen->vector[0], gen->vector[1], gen->vector[2]);
        V3f_Sum(vector, vector, vecDelta);
        V3f_Normalize(vector);
        gen->vector[0] = FLT2FIX(vector[0]);
        gen->vector[1] = FLT2FIX(vector[1]);
        gen->vector[2] = FLT2FIX(vector[2]);
        gen->presimulate(def->preSim);
    }
#else
    DENG2_UNUSED3(mob, inflictor, amount);
#endif
}
#ifdef __CLIENT__
dd_bool Mobj_OriginBehindVisPlane(mobj_t *mob) {
    if (!mob || !Mobj_HasSubsector(*mob)) return false;
    auto &subsec = Mobj_Subsector(*mob).as<ClientSubsector>();
    if (&subsec.sector().floor() != &subsec.visFloor()
        && mob->origin[2] < subsec.visFloor().heightSmoothed())
        return true;
    if (&subsec.sector().ceiling() != &subsec.visCeiling()
        && mob->origin[2] > subsec.visCeiling().heightSmoothed())
        return true;
    return false;
}
void Mobj_UnlinkLumobjs(mobj_t *mob) {
    if (!mob) return;
    mob->lumIdx = Lumobj::NoIndex;
}
static ded_light_t *lightDefByMobjState(state_t const *state) {
    if (state) {
        return runtimeDefs.stateInfo[runtimeDefs.states.indexOf(state)].light;
    }
    return nullptr;
}
static inline ClientTexture *lightmap(de::Uri const *textureUri) {
    if(!textureUri) return nullptr;
    return static_cast<ClientTexture *>
            (res::Textures::get().tryFindTextureByResourceUri(QStringLiteral("Lightmaps"), *textureUri));
}
void Mobj_GenerateLumobjs(mobj_t *mob) {
    if (!mob) return;
    Mobj_UnlinkLumobjs(mob);
    if (!Mobj_HasSubsector(*mob)) return;
    auto &subsec = Mobj_Subsector(*mob).as<ClientSubsector>();
    if (!(((mob->state && (mob->state->flags & STF_FULLBRIGHT))
            && !(mob->ddFlags & DDMF_DONTDRAW))
          || (mob->ddFlags & DDMF_ALWAYSLIT))) {
        return;
    }
    if (mob->state
        && (!mobjAutoLights || (mob->state->flags & STF_NOAUTOLIGHT))
        && !runtimeDefs.stateInfo[runtimeDefs.states.indexOf(mob->state)].light) {
       return;
    }
    if (!Mobj_BspLeafAtOrigin(*mob).subspace().contains(mob->origin))
        return;
    Record const *spriteRec = Mobj_SpritePtr(*mob);
    if (!spriteRec) return;
    MaterialAnimator *matAnimator = Rend_SpriteMaterialAnimator(*spriteRec);
    if (!matAnimator) return;
    matAnimator->prepare();  // Ensure we have up-to-date info.
    TextureVariant *tex = matAnimator->texUnit(MaterialAnimator::TU_LAYER0).texture;
    if (!tex) return;  // Unloadable texture?
    Vector2i const &texOrigin = tex->base().origin();
    coord_t impacted = mob->origin[2] + -texOrigin.y - matAnimator->dimensions().y
                     - subsec.visFloor().heightSmoothed();
    if (impacted < 0 && &subsec.visFloor() != &subsec.sector().floor())
        return;
    std::unique_ptr<Lumobj> lum(Rend_MakeLumobj(*spriteRec));
    if (!lum) return;
    lum->setSourceMobj(mob);
    if (ded_light_t *def = lightDefByMobjState(mob->state)) {
        if (!de::fequal(def->size, 0)) {
            lum->setRadius(de::max(def->size, 32.f / (40 * lum->radiusFactor())));
        }
        if (!de::fequal(def->offset[1], 0)) {
            lum->setZOffset(-texOrigin.y - def->offset[1]);
        }
        if (Vector3f(def->color) != Vector3f(0, 0, 0)) {
            lum->setColor(def->color);
        }
        lum->setLightmap(Lumobj::Side, lightmap(def->sides))
            .setLightmap(Lumobj::Down, lightmap(def->down))
            .setLightmap(Lumobj::Up,   lightmap(def->up));
    }
    lum->move(mob->origin);
    coord_t zOffset = -mob->floorClip - Mobj_BobOffset(*mob);
    if (!(mob->ddFlags & DDMF_NOFITBOTTOM) && impacted < 0) {
        zOffset -= impacted;
    }
    lum->setZOffset(lum->zOffset() + zOffset);
    mob->lumIdx = subsec.sector().map().addLumobj(lum.release()).indexInMap();
}
void Mobj_AnimateHaloOcclussion(mobj_t &mob) {
    for (dint i = 0; i < DDMAXPLAYERS; ++i) {
        dbyte *haloFactor = &mob.haloFactors[i];
        if (mob.lumIdx == Lumobj::NoIndex || R_ViewerLumobjIsClipped(mob.lumIdx)) {
            if (*haloFactor & 0x80) {
                dint f = (*haloFactor & 0x7f);  // - haloOccludeSpeed;
                if (f < 0) f = 0;
                *haloFactor = f;
            }
        }
        else {
            if (!(*haloFactor & 0x80)) {
                dint f = (*haloFactor & 0x7f);  // + haloOccludeSpeed;
                if (f > 127) f = 127;
                *haloFactor = 0x80 | f;
            }
        }
        dint f = *haloFactor & 0x7f;
        if (*haloFactor & 0x80) {
            f += ::haloOccludeSpeed;
            if (f > 127)
                f = 127;
        }
        else {
            f -= ::haloOccludeSpeed;
            if (f < 0)
                f = 0;
        }
        *haloFactor &= ~0x7f;
        *haloFactor |= f;
    }
}
dfloat Mobj_ShadowStrength(mobj_t const &mob) {
    static dfloat const minSpriteAlphaLimit = .1f;
    if (!Mobj_HasSubsector(mob)) return 0;
    if (!mob.state || (mob.state->flags & STF_FULLBRIGHT)) return 0;
    if (mob.ddFlags & DDMF_DONTDRAW) return 0;
    if (mob.ddFlags & DDMF_ALWAYSLIT) return 0;
    auto const &subsec = Mobj_Subsector(mob).as<ClientSubsector>();
    dfloat ambientLightLevel;
#if 0
    if (::useBias && subsec.sector().map().hasLightGrid()) {
        ambientLightLevel = subsec.sector().map().lightGrid().evaluateIntensity(mob.origin);
    }
    else
#endif
    {
        ambientLightLevel = subsec.lightSourceIntensity();
    }
    Rend_ApplyLightAdaptation(ambientLightLevel);
    dfloat strength = .65f;  ///< Default.
    if (!::useModels || !Mobj_ModelDef(mob)) {
        if (Record const *spriteRec = Mobj_SpritePtr(mob)) {
            auto &matAnimator = *Rend_SpriteMaterialAnimator(*spriteRec); // world::Materials::get().materialPtr(sprite.viewMaterial(0)))
            matAnimator.prepare();  // Ensure we have up-to-date info.
            if (TextureVariant const *texture = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture) {
                auto const *aa = (averagealpha_analysis_t const *)texture->base().analysisDataPointer(res::Texture::AverageAlphaAnalysis);
                DENG2_ASSERT(aa);
                dfloat weightedSpriteAlpha = aa->alpha * (0.4f + (1 - aa->coverage) * 0.6f);
                if(weightedSpriteAlpha < minSpriteAlphaLimit) return 0;
                strength *= de::min(1.f, .2f + weightedSpriteAlpha);
            }
        }
    }
    strength *= Mobj_Alpha(mob);
    return (0.6f - ambientLightLevel * 0.4f) * strength;
}
Record const *Mobj_SpritePtr(mobj_t const &mob) {
    return res::Sprites::get().spritePtr(mob.sprite, mob.frame);
}
FrameModelDef *Mobj_ModelDef(mobj_t const &mo, FrameModelDef **retNextModef, float *retInter) {
    if (retNextModef) *retNextModef = 0;
    if (retInter)     *retInter = -1;
    if (!mo.state) return 0;
    state_t &st = *mo.state;
    FrameModelDef *modef = App_Resources().modelDefForState(runtimeDefs.states.indexOf(&st), mo.selector);
    if (!modef) return 0; // No model available.
    dfloat interp = -1;
    bool worldTime = false;
    if (modef->flags & MFF_WORLD_TIME_ANIM) {
        dfloat duration = modef->interRange[0];
        dfloat offset   = modef->interRange[1];
        if (duration == 0) duration = 1;
        if (offset == -1) {
            offset = M_CycleIntoRange(MOBJ_TO_ID(&mo), duration);
        }
        interp = M_CycleIntoRange(App_World().time() / duration + offset, 1);
        worldTime = true;
    }
    else {
        interp = 1.0f - (mo.tics - frameTimePos) / dfloat( st.tics );
    }
/*#if _DEBUG
    if (mo.dPlayer) {
        qDebug() << "itp:" << interp << " mot:" << mo.tics << " stt:" << st.tics;
    }
#endif*/
    while (modef->interNext && modef->interNext->interMark <= interp) {
        modef = modef->interNext;
    }
    if (!worldTime) {
        interp = modef->interRange[0] + interp
               * (modef->interRange[1] - modef->interRange[0]);
    }
    if (retNextModef) {
        if (modef->interNext) {
            *retNextModef = modef->interNext;
        }
        else if (worldTime) {
            *retNextModef = App_Resources().modelDefForState(runtimeDefs.states.indexOf(&st), mo.selector);
        }
        else if (st.nextState > 0) // Check next state.
        {
            state_t *it = &runtimeDefs.states[st.nextState];
            bool foundNext = false;
            if (modef->interRange[1] < 1) {
                bool stopScan = false;
                dint max = 20; // Let's not be here forever...
                while (!stopScan) {
                    if(!((  !App_Resources().modelDefForState(runtimeDefs.states.indexOf(it))
                          || App_Resources().modelDefForState(runtimeDefs.states.indexOf(it), mo.selector)->interRange[0] > 0)
                         && it->nextState > 0)) {
                        stopScan = true;
                    }
                    else {
                        FrameModelDef *mdit = App_Resources().modelDefForState(runtimeDefs.states.indexOf(it), mo.selector);
                        if (mdit && mdit->interNext) {
                            forever {
                                mdit = mdit->interNext;
                                if (mdit) {
                                    if (mdit->interRange[0] <= 0) // A new beginning?
                                    {
                                        *retNextModef = mdit;
                                        foundNext = true;
                                    }
                                }
                                if (!mdit || foundNext) {
                                    break;
                                }
                            }
                        }
                        if (foundNext) {
                            stopScan = true;
                        }
                        else {
                            it = &runtimeDefs.states[it->nextState];
                        }
                    }
                    if (max-- <= 0)
                        stopScan = true;
                }
            }
            if (!foundNext) {
                *retNextModef = App_Resources().modelDefForState(runtimeDefs.states.indexOf(it), mo.selector);
            }
        }
    }
    if (retInter) *retInter = interp;
    return modef;
}
#endif // __CLIENT__
#undef Mobj_AngleSmoothed
DENG_EXTERN_C angle_t Mobj_AngleSmoothed(mobj_t *mob) {
    if (!mob) return 0;
#ifdef __CLIENT__
    if (mob->dPlayer) {
        if (P_GetDDPlayerIdx(mob->dPlayer) == ::consolePlayer
            && mob->dPlayer->mo == mob) {
            viewdata_t const *vd = &DD_Player(::consolePlayer)->viewport();
            return vd->current.angle();
        }
    }
    if (::useSRVOAngle && !::netGame && !::playback) {
        return mob->visAngle << 16;
    }
#endif
    return mob->angle;
}
coord_t Mobj_ApproxPointDistance(mobj_t const *mob, coord_t const *point) {
    if (!mob || !point) return 0;
    return M_ApproxDistance(point[2] - mob->origin[2],
                            M_ApproxDistance(point[0] - mob->origin[0],
                                             point[1] - mob->origin[1]));
}
coord_t Mobj_BobOffset(mobj_t const &mob) {
    if (mob.ddFlags & DDMF_BOB) {
        return (sin(MOBJ_TO_ID(&mob) + App_World().time() / 1.8286 * 2 * PI) * 8);
    }
    return 0;
}
dfloat Mobj_Alpha(mobj_t const &mob) {
    dfloat alpha = (mob.ddFlags & DDMF_BRIGHTSHADOW)? .80f :
                   (mob.ddFlags & DDMF_SHADOW      )? .33f :
                   (mob.ddFlags & DDMF_ALTSHADOW   )? .66f : 1;
    dint selAlpha = mob.selector >> DDMOBJ_SELECTOR_SHIFT;
    if (selAlpha & 0xe0) {
        alpha *= 1 - ((selAlpha & 0xe0) >> 5) / 8.0f;
    }
    else if (mob.translucency) {
        alpha *= 1 - mob.translucency * reciprocal255;
    }
    return alpha;
}
coord_t Mobj_Radius(mobj_t const &mobj) {
    return mobj.radius;
}
#ifdef __CLIENT__
coord_t Mobj_ShadowRadius(mobj_t const &mobj) {
    if (useModels) {
        if (FrameModelDef *modef = Mobj_ModelDef(mobj)) {
            if (modef->shadowRadius > 0) {
                return modef->shadowRadius;
            }
        }
    }
    return Mobj_VisualRadius(mobj);
}
#endif
coord_t Mobj_VisualRadius(mobj_t const &mob) {
#ifdef __CLIENT__
    if (useModels) {
        if (FrameModelDef *modef = Mobj_ModelDef(mob)) {
            return modef->visualRadius;
        }
    }
    if (Record const *sprite = Mobj_SpritePtr(mob)) {
        return Rend_VisualRadius(*sprite);
    }
#endif
    return Mobj_Radius(mob);
}
AABoxd Mobj_Bounds(mobj_t const &mobj) {
    Vector2d const origin = Mobj_Origin(mobj);
    ddouble const radius  = Mobj_Radius(mobj);
    return AABoxd(origin.x - radius, origin.y - radius,
                  origin.x + radius, origin.y + radius);
}
D_CMD(InspectMobj) {
    DENG2_UNUSED(src);
    if (argc != 2) {
        LOG_SCR_NOTE("Usage: %s (mobj-id)") << argv[0];
        return true;
    }
    auto const id = thid_t( String(argv[1]).toInt() );
    mobj_t *mob   = App_World().map().thinkers().mobjById(id);
    if (!mob) {
        LOG_MAP_ERROR("Mobj with id %i not found") << id;
        return false;
    }
    char const *mobType = "Mobj";
#ifdef __CLIENT__
    ClientMobjThinkerData::RemoteSync *info = ClMobj_GetInfo(mob);
    if (info) mobType = "CLMOBJ";
#endif
    LOG_MAP_MSG("%s %i [%p] State:%s (%i)")
            << mobType << id << mob << Def_GetStateName(mob->state) << ::runtimeDefs.states.indexOf(mob->state);
    LOG_MAP_MSG("Type:%s (%i) Info:[%p] %s")
            << DED_Definitions()->getMobjName(mob->type) << mob->type << mob->info
            << (mob->info ? QString(" (%1)").arg(::runtimeDefs.mobjInfo.indexOf(mob->info)) : "");
    LOG_MAP_MSG("Tics:%i ddFlags:%08x") << mob->tics << mob->ddFlags;
#ifdef __CLIENT__
    if (info) {
        LOG_MAP_MSG("Cltime:%i (now:%i) Flags:%04x") << info->time << Timer_RealMilliseconds() << info->flags;
    }
#endif
    LOG_MAP_MSG("Flags:%08x Flags2:%08x Flags3:%08x") << mob->flags << mob->flags2 << mob->flags3;
    LOG_MAP_MSG("Height:%f Radius:%f") << mob->height << mob->radius;
    LOG_MAP_MSG("Angle:%x Pos:%s Mom:%s")
            << mob->angle
            << Vector3d(mob->origin).asText()
            << Vector3d(mob->mom).asText();
#ifdef __CLIENT__
    LOG_MAP_MSG("VisAngle:%x") << mob->visAngle;
#endif
    LOG_MAP_MSG("%sZ:%f %sZ:%f")
        << Sector::planeIdAsText(Sector::Floor  ).upperFirstChar() << mob->floorZ
        << Sector::planeIdAsText(Sector::Ceiling).upperFirstChar() << mob->ceilingZ;
    if (Subsector *subsec = Mobj_SubsectorPtr(*mob)) {
        LOG_MAP_MSG("Sector:%i (%sZ:%f %sZ:%f)")
                << subsec->sector().indexInMap()
                << Sector::planeIdAsText(Sector::Floor  ) << subsec->sector().floor  ().height()
                << Sector::planeIdAsText(Sector::Ceiling) << subsec->sector().ceiling().height();
    }
    if (mob->onMobj) {
        LOG_MAP_MSG("onMobj:%i") << mob->onMobj->thinker.id;
    }
    return true;
}
void Mobj_ConsoleRegister() {
    C_CMD("inspectmobj",    "i",    InspectMobj);
#ifdef __CLIENT__
    C_VAR_BYTE("rend-mobj-light-auto", &mobjAutoLights, 0, 0, 1);
#endif
}
#ifdef __CLIENT__
ClientPlayer *viewPlayer;
#endif
int consolePlayer;
int displayPlayer;
typedef QMap<int, PlayerImpulse *> Impulses; // ID lookup.
static Impulses impulses;
typedef QMap<String, PlayerImpulse *> ImpulseNameMap; // Name lookup
static ImpulseNameMap impulsesByName;
typedef QMap<int, ImpulseAccumulator *> ImpulseAccumulators; // ImpulseId lookup.
static ImpulseAccumulators accumulators[DDMAXPLAYERS];
static PlayerImpulse *addImpulse(int id, impulsetype_t type, String name, String bindContextName) {
    auto *imp = new PlayerImpulse;
    imp->id   = id;
    imp->type = type;
    imp->name = name;
    imp->bindContextName = bindContextName;
    impulses.insert(imp->id, imp);
    impulsesByName.insert(imp->name.toLower(), imp);
    for(int i = 0; i < DDMAXPLAYERS; ++i) {
        ImpulseAccumulator::AccumulatorType accumType = (type == IT_BINARY? ImpulseAccumulator::Binary : ImpulseAccumulator::Analog);
        auto *accum = new ImpulseAccumulator(imp->id, accumType, type != IT_ANALOG);
        accum->setPlayerNum(i);
        accumulators[i].insert(accum->impulseId(), accum);
    }
    return imp;
}
static ImpulseAccumulator *accumulator(int impulseId, int playerNum) {
    if(playerNum < 0 || playerNum >= DDMAXPLAYERS)
        return nullptr;
    if(!accumulators[playerNum].contains(impulseId))
        return nullptr;
    return accumulators[playerNum][impulseId];
}
AppPlayer *DD_Player(int number) {
    return &DoomsdayApp::players().at(number).as<AppPlayer>();
}
int P_LocalToConsole(int localPlayer) {
    int count = 0;
    for(int i = 0; i < DDMAXPLAYERS; ++i) {
        int console      = (i + consolePlayer) % DDMAXPLAYERS;
        player_t *plr    = DD_Player(console);
        ddplayer_t *ddpl = &plr->publicData();
        if(ddpl->flags & DDPF_LOCAL) {
            if(count++ == localPlayer)
                return console;
        }
    }
    return -1;
}
int P_ConsoleToLocal(int playerNum) {
    int i, count = 0;
    player_t *plr = DD_Player(playerNum);
    if(playerNum < 0 || playerNum >= DDMAXPLAYERS) {
        return -1;
    }
    if(playerNum == consolePlayer) {
        return 0;
    }
    if(!(plr->publicData().flags & DDPF_LOCAL))
        return -1; // Not local at all.
    for(i = 0; i < DDMAXPLAYERS; ++i) {
        int console = (i + consolePlayer) % DDMAXPLAYERS;
        player_t *plr = DD_Player(console);
        if(console == playerNum) {
            return count;
        }
        if(plr->publicData().flags & DDPF_LOCAL)
            count++;
    }
    return -1;
}
int P_GetDDPlayerIdx(ddplayer_t *ddpl) {
    return DoomsdayApp::players().indexOf(ddpl);
}
#ifdef __CLIENT__
bool P_IsInVoid(player_t *player) {
    if (!player) return false;
    ddplayer_t const *ddpl = &player->publicData();
    if (ddpl->flags & DDPF_CAMERA) {
        if (player->inVoid || !ddpl->mo)
            return true;
        mobj_t const *mob = ddpl->mo;
        if (!Mobj_HasSubsector(*mob))
            return true;
        auto const &subsec = Mobj_Subsector(*mob).as<world::ClientSubsector>();
        if (subsec.visCeiling().surface().hasSkyMaskedMaterial()) {
            world::ClSkyPlane const &skyCeiling = subsec.sector().map().skyCeiling();
            if (skyCeiling.height() < DDMAXFLOAT && mob->origin[2] > skyCeiling.height() - 4)
                return true;
        }
        else if (mob->origin[2] > subsec.visCeiling().heightSmoothed() - 4) {
            return true;
        }
        if (subsec.visFloor().surface().hasSkyMaskedMaterial()) {
            world::ClSkyPlane const &skyFloor = subsec.sector().map().skyFloor();
            if (skyFloor.height() > DDMINFLOAT && mob->origin[2] < skyFloor.height() + 4)
                return true;
        }
        else if (mob->origin[2] < subsec.visFloor().heightSmoothed() + 4) {
            return true;
        }
    }
    return false;
}
#endif // __CLIENT__
void P_ClearPlayerImpulses() {
    for(dint i = 0; i < DDMAXPLAYERS; ++i) {
        qDeleteAll(accumulators[i]);
        accumulators[i].clear();
    }
    qDeleteAll(impulses);
    impulses.clear();
    impulsesByName.clear();
}
PlayerImpulse *P_PlayerImpulsePtr(dint id) {
    auto found = impulses.find(id);
    if(found != impulses.end()) return *found;
    return nullptr;
}
PlayerImpulse *P_PlayerImpulseByName(String const &name) {
    if(!name.isEmpty()) {
        auto found = impulsesByName.find(name.toLower());
        if(found != impulsesByName.end()) return *found;
    }
    return nullptr;
}
D_CMD(ListImpulses) {
    DENG2_UNUSED3(argv, argc, src);
    typedef QList<PlayerImpulse *> ImpulseList;
    QMap<String, ImpulseList> contextGroups;
    for(PlayerImpulse *imp : impulsesByName) {
        if(!contextGroups.contains(imp->bindContextName)) {
            contextGroups.insert(imp->bindContextName, ImpulseList());
        }
        contextGroups[imp->bindContextName].append(imp);
    }
    LOG_MSG(_E(b) "Player impulses");
    LOG_MSG("There are " _E(b) "%i" _E(.) " impulses, in " _E(b) "%i" _E(.) " contexts")
            << impulses.count() << contextGroups.count();
    for(auto const &group : contextGroups) {
        if(group.isEmpty()) continue;
        LOG_MSG(_E(D)_E(b) "%s" _E(.) " context: " _E(l) "(%i)")
                << group.first()->bindContextName
                << group.count();
        for(PlayerImpulse const *imp : group) {
            LOG_MSG("  [%4i] " _E(>) _E(b) "%s " _E(.) _E(2) "%s%s")
                    << imp->id << imp->name
                    << (imp->type == IT_BINARY? "binary" : "analog")
                    << (IMPULSETYPE_IS_TRIGGERABLE(imp->type)? ", triggerable" : "");
        }
    }
    return true;
}
D_CMD(Impulse) {
    DENG2_UNUSED(src);
    if(argc < 2 || argc > 3) {
        LOG_SCR_NOTE("Usage:\n  %s (impulse-name)\n  %s (impulse-name) (player-ordinal)")
                << argv[0] << argv[0];
        return true;
    }
    if(PlayerImpulse *imp = P_PlayerImpulseByName(argv[1])) {
        int const playerNum = (argc == 3? P_LocalToConsole(String(argv[2]).toInt()) : consolePlayer);
        if(ImpulseAccumulator *accum = accumulator(imp->id, playerNum)) {
            accum->receiveBinary();
        }
    }
    return true;
}
#ifdef __CLIENT__
D_CMD(ClearImpulseAccumulation) {
    DENG2_UNUSED3(argv, argc, src);
    ClientApp::inputSystem().forAllDevices([] (InputDevice &device) {
        device.reset();
        return LoopContinue;
    });
    for(int i = 0; i < DDMAXPLAYERS; ++i)
    for(ImpulseAccumulator *accum : accumulators[i]) {
        accum->clearAll();
    }
    return true;
}
#endif
void P_ConsoleRegister() {
    C_CMD("listcontrols",   "",         ListImpulses);
    C_CMD("impulse",        nullptr,    Impulse);
#ifdef __CLIENT__
    C_CMD("resetctlaccum",  "",         ClearImpulseAccumulation);
    ImpulseAccumulator::consoleRegister();
#endif
}
#undef DD_GetPlayer
DENG_EXTERN_C ddplayer_t *DD_GetPlayer(int number) {
    return &DD_Player(number)->publicData();
}
#undef Net_GetPlayerName
#undef Net_GetPlayerID
#undef Net_PlayerSmoother
DENG_EXTERN_C char const *Net_GetPlayerName(int player);
DENG_EXTERN_C ident_t Net_GetPlayerID(int player);
DENG_EXTERN_C Smoother *Net_PlayerSmoother(int player);
#undef P_NewPlayerControl
DENG_EXTERN_C void P_NewPlayerControl(int id, impulsetype_t type, char const *name,
    char const *bindContextName) {
    DENG2_ASSERT(name && bindContextName);
    LOG_AS("P_NewPlayerControl");
    if(PlayerImpulse const *existing = P_PlayerImpulsePtr(id)) {
        LOG_INPUT_WARNING("Id: %i is already in use by impulse '%s' - Won't replace")
                << id << existing->name;
        return;
    }
    if(PlayerImpulse const *existing = P_PlayerImpulseByName(name)) {
        LOG_INPUT_WARNING("Name: '%s' is already in use by impulse Id: %i - Won't replace")
                << name << existing->id;
        return;
    }
    addImpulse(id, type, name, bindContextName);
}
#undef P_IsControlBound
DENG_EXTERN_C int P_IsControlBound(int playerNum, int impulseId) {
#ifdef __CLIENT__
    LOG_AS("P_IsControlBound");
    int const localPlayer = P_ConsoleToLocal(playerNum);
    if(localPlayer < 0) return false;
    if(PlayerImpulse const *imp = P_PlayerImpulsePtr(impulseId)) {
        InputSystem &isys = ClientApp::inputSystem();
        BindContext *bindContext = isys.contextPtr(imp->bindContextName);
        if(!bindContext) {
            LOGDEV_INPUT_WARNING("Unknown binding context '%s'") << imp->bindContextName;
            return false;
        }
        int found = bindContext->forAllImpulseBindings(localPlayer, [&isys, &impulseId] (CompiledImpulseBindingRecord &rec) {
            auto const &bind = rec.compiled();
            if(bind.impulseId != impulseId) return LoopContinue;
            if(InputDevice const *device = isys.devicePtr(bind.deviceId)) {
                if(device->isActive()) {
                    return LoopAbort; // found a binding.
                }
            }
            return LoopContinue;
        });
        return (found? 1 : 0);
    }
    return false;
#else
    DENG2_UNUSED2(playerNum, impulseId);
    return 0;
#endif
}
#undef P_GetControlState
DENG_EXTERN_C void P_GetControlState(int playerNum, int impulseId, float *pos,
    float *relativeOffset) {
#ifdef __CLIENT__
    float tmp;
    if(!pos) pos = &tmp;
    if(!relativeOffset) relativeOffset = &tmp;
    *pos = 0;
    *relativeOffset = 0;
    if(ImpulseAccumulator *accum = accumulator(impulseId, playerNum)) {
        accum->takeAnalog(pos, relativeOffset);
    }
#else
    DENG2_UNUSED4(playerNum, impulseId, pos, relativeOffset);
#endif
}
#undef P_GetImpulseControlState
DENG_EXTERN_C int P_GetImpulseControlState(int playerNum, int impulseId) {
    LOG_AS("P_GetImpulseControlState");
    ImpulseAccumulator *accum = accumulator(impulseId, playerNum);
    if(!accum) return 0;
    if(accum->type() != ImpulseAccumulator::Binary) {
        LOG_INPUT_WARNING("ImpulseAccumulator '%s' is not binary") << impulses[impulseId]->name;
        return 0;
    }
    return accum->takeBinary();
}
#undef P_Impulse
DENG_EXTERN_C void P_Impulse(int playerNum, int impulseId) {
    LOG_AS("P_Impulse");
    ImpulseAccumulator *accum = accumulator(impulseId, playerNum);
    if(!accum) return;
    if(accum->type() != ImpulseAccumulator::Binary) {
        LOG_INPUT_WARNING("ImpulseAccumulator '%s' is not binary") << impulses[impulseId]->name;
        return;
    }
    accum->receiveBinary();
}
DENG_DECLARE_API(Player) = {
    { DE_API_PLAYER },
    Net_GetPlayerName,
    Net_GetPlayerID,
    Net_PlayerSmoother,
    DD_GetPlayer,
    P_NewPlayerControl,
    P_IsControlBound,
    P_GetControlState,
    P_GetImpulseControlState,
    P_Impulse
};
void P_Ticker(timespan_t elapsed) {
#ifdef __CLIENT__
    world::Materials::get().forAnimatedMaterials([&elapsed] (world::Material &material) {
        auto &mat = material.as<ClientMaterial>();
        for (int i = mat.animatorCount() - 1; i >= 0; --i) {
            mat.animator(i).animate(elapsed);
        }
        return LoopContinue;
    });
#endif
    App_World().tick(elapsed);
    DoomsdayApp::players().forAll([&elapsed] (Player &plr) {
        plr.tick(elapsed);
        return LoopContinue;
    });
}
DENG2_PIMPL(Plane)
#ifdef __CLIENT__
, DENG2_OBSERVES(Surface, MaterialChange)
#endif
{
    Surface surface;
    ThinkerT<SoundEmitter> soundEmitter;
    dint indexInSector = -1;           ///< Index in the owning sector.
    ddouble height = 0;                ///< Current @em sharp height.
    ddouble heightTarget = 0;          ///< Target @em sharp height.
    ddouble speed = 0;                 ///< Movement speed (map space units per tic).
#ifdef __CLIENT__
    std::array<ddouble, 2> oldHeight;  ///< @em sharp height change tracking buffer (for smoothing).
    ddouble heightSmoothed = 0;        ///< @ref height (smoothed).
    ddouble heightSmoothedDelta = 0;   ///< Delta between the current @em sharp height and the visual height.
    ClPlaneMover *mover = nullptr;     ///< The current mover.
#endif
    Impl(Public *i)
        : Base(i)
        , surface(dynamic_cast<MapElement &>(*i)) {
#ifdef __CLIENT__
        surface.audienceForMaterialChange() += this;
#endif
    }
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) i->planeBeingDeleted(self());
#ifdef __CLIENT__
        map().trackedPlanes().remove(thisPublic);
#endif
    }
    inline Map &map() const { return self().map(); }
    void setHeight(ddouble newHeight) {
        height = heightTarget = newHeight;
#ifdef __CLIENT__
        heightSmoothed = newHeight;
        oldHeight[0] = oldHeight[1] = newHeight;
#endif
    }
    void applySharpHeightChange(ddouble newHeight) {
        if(de::fequal(newHeight, height))
            return;
        height = newHeight;
        if(!ddMapSetup) {
            self().updateSoundEmitterOrigin();
        }
        notifyHeightChanged();
#ifdef __CLIENT__
        if(!ddMapSetup) {
            map().trackedPlanes().insert(thisPublic);
        }
#endif
    }
#ifdef __CLIENT__
    Generator *tryFindGenerator() {
        Generator *found = nullptr;
        map().forAllGenerators([this, &found] (Generator &gen) {
            if(gen.plane == thisPublic) {
                found = &gen;
                return LoopAbort;  // Found it.
            }
            return LoopContinue;
        });
        return found;
    }
#endif
    void notifyHeightChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE2(HeightChange, i) i->planeHeightChanged(self());
    }
#ifdef __CLIENT__
    void notifySmoothedHeightChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE2(HeightSmoothedChange, i) i->planeHeightSmoothedChanged(self());
    }
    void surfaceMaterialChanged(Surface &suf) {
        DENG2_ASSERT(&suf == &surface);
        DENG2_UNUSED(suf);
        if (!::ddMapSetup && surface.hasMaterial()) {
            de::Uri uri = surface.material().manifest().composeUri();
            self().spawnParticleGen(Def_GetGenerator(reinterpret_cast<uri_s *>(&uri)));
        }
    }
#endif
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(HeightChange)
#ifdef __CLIENT__
    DENG2_PIMPL_AUDIENCE(HeightSmoothedChange)
#endif
};
DENG2_AUDIENCE_METHOD(Plane, Deletion)
DENG2_AUDIENCE_METHOD(Plane, HeightChange)
#ifdef __CLIENT__
DENG2_AUDIENCE_METHOD(Plane, HeightSmoothedChange)
#endif
Plane::Plane(Sector &sector, Vector3f const &normal, ddouble height)
    : MapElement(DMU_PLANE, &sector)
    , d(new Impl(this)) {
    d->setHeight(height);
    setNormal(normal);
}
String Plane::description() const {
    auto desc = String(    _E(l) "Sector: "        _E(.)_E(i) "%1" _E(.)
                       " " _E(l) "Height: "        _E(.)_E(i) "%2" _E(.)
                       " " _E(l) "Height Target: " _E(.)_E(i) "%3" _E(.)
                       " " _E(l) "Speed: "         _E(.)_E(i) "%4" _E(.))
                  .arg(sector().indexInMap())
                  .arg(height())
                  .arg(heightTarget())
                  .arg(speed());
    DENG2_DEBUG_ONLY(
        desc.prepend(String(_E(b) "Plane " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
    )
    return desc + "\n" + surface().description();
}
Sector &Plane::sector() {
    return parent().as<Sector>();
}
Sector const &Plane::sector() const {
    return parent().as<Sector>();
}
dint Plane::indexInSector() const {
    return d->indexInSector;
}
void Plane::setIndexInSector(dint newIndex) {
    d->indexInSector = newIndex;
}
bool Plane::isSectorFloor() const {
    return this == &sector().floor();
}
bool Plane::isSectorCeiling() const {
    return this == &sector().ceiling();
}
Surface &Plane::surface() {
    return d->surface;
}
Surface const &Plane::surface() const {
    return d->surface;
}
Surface *Plane::surfacePtr() const {
    return &d->surface;
}
void Plane::setNormal(Vector3f const &newNormal) {
    d->surface.setNormal(newNormal);  // will normalize
}
SoundEmitter &Plane::soundEmitter() {
    return d->soundEmitter;
}
SoundEmitter const &Plane::soundEmitter() const {
    return d->soundEmitter;
}
void Plane::updateSoundEmitterOrigin() {
    LOG_AS("Plane::updateSoundEmitterOrigin");
    d->soundEmitter->origin[0] = sector().soundEmitter().origin[0];
    d->soundEmitter->origin[1] = sector().soundEmitter().origin[1];
    d->soundEmitter->origin[2] = d->height;
}
ddouble Plane::height() const {
    return d->height;
}
ddouble Plane::heightTarget() const {
    return d->heightTarget;
}
ddouble Plane::speed() const {
    return d->speed;
}
#ifdef __CLIENT__
ddouble Plane::heightSmoothed() const {
    return d->heightSmoothed;
}
ddouble Plane::heightSmoothedDelta() const {
    return d->heightSmoothedDelta;
}
void Plane::lerpSmoothedHeight() {
    d->heightSmoothedDelta = d->oldHeight[0] * (1 - frameTimePos)
                           + d->height * frameTimePos - d->height;
    ddouble newHeightSmoothed = d->height + d->heightSmoothedDelta;
    if(!de::fequal(d->heightSmoothed, newHeightSmoothed)) {
        d->heightSmoothed = newHeightSmoothed;
        d->notifySmoothedHeightChanged();
    }
}
void Plane::resetSmoothedHeight() {
    d->heightSmoothedDelta = 0;
    ddouble newHeightSmoothed = d->oldHeight[0] = d->oldHeight[1] = d->height;
    if(!de::fequal(d->heightSmoothed, newHeightSmoothed)) {
        d->heightSmoothed = newHeightSmoothed;
        d->notifySmoothedHeightChanged();
    }
}
void Plane::updateHeightTracking() {
    d->oldHeight[0] = d->oldHeight[1];
    d->oldHeight[1] = d->height;
    if(!de::fequal(d->oldHeight[0], d->oldHeight[1])) {
        if(de::abs(d->oldHeight[0] - d->oldHeight[1]) >= MAX_SMOOTH_MOVE) {
            d->oldHeight[0] = d->oldHeight[1];
        }
    }
}
bool Plane::hasGenerator() const {
    return d->tryFindGenerator() != nullptr;
}
Generator &Plane::generator() const {
    if(Generator *gen = d->tryFindGenerator()) return *gen;
    throw MissingGeneratorError("Plane::generator", "No generator is attached");
}
void Plane::spawnParticleGen(ded_ptcgen_t const *def) {
    if(!def) return;
    dint relPlane = indexInSector();
    if(def->flags & Generator::SpawnCeiling)
        relPlane = Sector::Ceiling;
    if(def->flags & Generator::SpawnFloor)
        relPlane = Sector::Floor;
    if(relPlane != indexInSector()) {
        sector().plane(relPlane).spawnParticleGen(def);
        return;
    }
    if(!sector().sideCount()) return;
    if(hasGenerator()) return;
    Generator *gen = map().newGenerator();
    if(!gen) return;
    gen->count = def->particles;
    if(def->flags & Generator::Density) {
        gen->spawnRateMultiplier = sector().roughArea() / (128 * 128);
    }
    else {
        gen->spawnRateMultiplier = 1;
    }
    gen->configureFromDef(def);
    gen->plane = this;
    gen->presimulate(def->preSim);
}
void Plane::addMover(ClPlaneMover &mover) {
    if(d->mover) {
        LOG_MAP_XVERBOSE("Removing existing mover %p in sector #%i, plane %i",
                         &d->mover->thinker() << sector().indexInMap()
                         << indexInSector());
        map().thinkers().remove(d->mover->thinker());
        DENG2_ASSERT(!d->mover);
    }
    d->mover = &mover;
}
void Plane::removeMover(ClPlaneMover &mover) {
    if(d->mover == &mover) {
        d->mover = nullptr;
    }
}
bool Plane::castsShadow() const {
    if (auto *matAnim = d->surface.materialAnimator()) {
        matAnim->prepare();
        if (!matAnim->material().isDrawable()) return false;
        if (matAnim->material().isSkyMasked()) return false;
        return de::fequal(matAnim->glowStrength(), 0);
    }
    return false;
}
bool Plane::receivesShadow() const {
    return castsShadow();  // Qualification is the same as with casting.
}
#endif // __CLIENT__
dint Plane::property(DmuArgs &args) const {
    switch(args.prop) {
    case DMU_EMITTER: {
        SoundEmitter const *emitterPtr = &soundEmitter();
        args.setValue(DMT_PLANE_EMITTER, &emitterPtr, 0);
        break; }
    case DMU_SECTOR: {
        Sector const *secPtr = &sector();
        args.setValue(DMT_PLANE_SECTOR, &secPtr, 0);
        break; }
    case DMU_HEIGHT:
        args.setValue(DMT_PLANE_HEIGHT, &d->height, 0);
        break;
    case DMU_TARGET_HEIGHT:
        args.setValue(DMT_PLANE_TARGET, &d->heightTarget, 0);
        break;
    case DMU_SPEED:
        args.setValue(DMT_PLANE_SPEED, &d->speed, 0);
        break;
    default:
        return MapElement::property(args);
    }
    return false;  // Continue iteration.
}
dint Plane::setProperty(DmuArgs const &args) {
    switch(args.prop) {
    case DMU_HEIGHT: {
        ddouble newHeight = d->height;
        args.value(DMT_PLANE_HEIGHT, &newHeight, 0);
        d->applySharpHeightChange(newHeight);
        break; }
    case DMU_TARGET_HEIGHT:
        args.value(DMT_PLANE_TARGET, &d->heightTarget, 0);
        break;
    case DMU_SPEED:
        args.value(DMT_PLANE_SPEED, &d->speed, 0);
        break;
    default:
        return MapElement::setProperty(args);
    }
    return false;  // Continue iteration.
}
static void (*collisionCallback) (mobj_t *mob, void *line, void *pob);
static inline bool mobCanBlockMovement(mobj_t const &mob) {
    return (mob.ddFlags & DDMF_SOLID) || (mob.dPlayer && !(mob.dPlayer->flags & DDPF_CAMERA));
}
#if 0
static void notifyGeometryChanged(Polyobj &pob) {
#ifdef __CLIENT__
    if (!ddMapSetup && useBias) {
        for (HEdge *hedge : pob.mesh().hedges()) {
            if (!hedge->hasMapElement())
                continue;
            auto &subsec = pob.bspLeaf().subspace().subsector().as<ClientSubsector>();
            if (Shard *shard = subsec.findShard(hedge->mapElement(), LineSide::Middle)) {
                shard->updateBiasAfterMove();
            }
        }
    }
#else // !__CLIENT__
    DENG2_UNUSED(pob);
#endif
}
#endif
void Polyobj::NotifyCollision(Polyobj &pob, mobj_t *mob, Line *line)  // static
{
    if (collisionCallback) {
        collisionCallback(mob, line, &pob);
    }
}
bool Polyobj::blocked() const {
    for (Line const *line : lines()) {
        dint const localValidCount = ++::validCount;
        bool collision = false;
        map().mobjBlockmap().forAllInBox(AABoxd(line->bounds().minX - DDMOBJ_RADIUS_MAX,
                                                line->bounds().minY - DDMOBJ_RADIUS_MAX,
                                                line->bounds().maxX + DDMOBJ_RADIUS_MAX,
                                                line->bounds().maxY + DDMOBJ_RADIUS_MAX)
                            , [this, &line, &collision, &localValidCount](void *object) {
            auto &mob = *(mobj_t *)object;
            if (mob.validCount == localValidCount) return LoopContinue;
            mob.validCount = localValidCount;  // now processed.
            if (mobCanBlockMovement(mob)) {
                AABoxd mobBox = Mobj_Bounds(mob);
                if (   mobBox.maxX <= line->bounds().minX
                    || mobBox.minX >= line->bounds().maxX
                    || mobBox.maxY <= line->bounds().minY
                    || mobBox.minY >= line->bounds().maxY)
                   return LoopContinue;
                if (!line->boxOnSide(mobBox)) {
                    NotifyCollision(*const_cast<Polyobj *>(this), &mob, const_cast<Line *>(line));
                    collision = true;
                }
            }
            return LoopContinue;
        });
        if (collision) return true;
    }
    return false;  // All clear.
}
polyobj_s::polyobj_s(Vector2d const &origin_) {
    zap(thinker);
    origin[0] = origin_.x;
    origin[1] = origin_.y;
    tag = 0;
    validCount = 0;
    dest[0] = dest[1] = 0;
    angle = destAngle = 0;
    angleSpeed = 0;
    speed = 0;
    crush = false;
    seqType = 0;
    _bspLeaf = 0;
    thinker.d = new PolyobjData;
    THINKER_DATA(thinker, PolyobjData).setThinker(&thinker);
}
polyobj_s::~polyobj_s() {
    delete THINKER_DATA_MAYBE(thinker, PolyobjData);
}
PolyobjData &polyobj_s::data() {
    return THINKER_DATA(thinker, PolyobjData);
}
PolyobjData const &polyobj_s::data() const {
    return THINKER_DATA(thinker, PolyobjData);
}
void Polyobj::setCollisionCallback(void (*func) (mobj_t *mob, void *line, void *polyob))  // static
{
    collisionCallback = func;
}
Map &Polyobj::map() const {
    return App_World().map();
}
Mesh &Polyobj::mesh() const {
    return *data().mesh;
}
bool Polyobj::isLinked() {
    return hasBspLeaf();
}
void Polyobj::unlink() {
    if (_bspLeaf) {
        if (((BspLeaf *)_bspLeaf)->hasSubspace()) {
            ((BspLeaf *)_bspLeaf)->subspace().unlink(*this);
        }
        _bspLeaf = nullptr;
        map().unlink(*this);
    }
}
void Polyobj::link() {
    if (!_bspLeaf) {
        map().link(*this);
        Vector2d avg;
        for (Line *line : lines()) {
            avg += line->from().origin();
        }
        avg /= lineCount();
        _bspLeaf = &map().bspLeafAt(avg);
        if (((BspLeaf *)_bspLeaf)->hasSubspace()) {
            ((BspLeaf *)_bspLeaf)->subspace().link(*this);
        }
    }
}
bool Polyobj::hasBspLeaf() const {
    return _bspLeaf != nullptr;
}
BspLeaf &Polyobj::bspLeaf() const {
    if (hasBspLeaf()) return *(BspLeaf *)_bspLeaf;
    throw Polyobj::NotLinkedError("Polyobj::bspLeaf", "Polyobj is not presently linked in the BSP");
}
bool Polyobj::hasSector() const {
    return hasBspLeaf() && bspLeaf().hasSubspace();
}
Sector &Polyobj::sector() const {
    return *bspLeaf().sectorPtr();
}
Sector *Polyobj::sectorPtr() const {
    return hasBspLeaf() ? bspLeaf().sectorPtr() : nullptr;
}
SoundEmitter &Polyobj::soundEmitter() {
    return *reinterpret_cast<SoundEmitter *>(this);
}
SoundEmitter const &Polyobj::soundEmitter() const {
    return const_cast<SoundEmitter const &>(const_cast<Polyobj &>(*this).soundEmitter());
}
QList<Line *> const &Polyobj::lines() const {
    return data().lines;
}
QList<Vertex *> const &Polyobj::uniqueVertexes() const {
    return data().uniqueVertexes;
}
void Polyobj::buildUniqueVertexes() {
    QSet<Vertex *> vertexSet;
    for (Line *line : lines()) {
        vertexSet.insert(&line->from());
        vertexSet.insert(&line->to());
    }
    QList<Vertex *> &uniqueVertexes = data().uniqueVertexes;
    uniqueVertexes = vertexSet.toList();
    data().originalPts.resize(uniqueVertexes.count());
    data().prevPts.resize(uniqueVertexes.count());
}
void Polyobj::updateOriginalVertexCoords() {
    PolyobjData::VertexCoords::iterator origCoordsIt = data().originalPts.begin();
    for (Vertex *vertex : uniqueVertexes()) {
        (*origCoordsIt) = vertex->origin() - Vector2d(origin);
        origCoordsIt++;
    }
}
void Polyobj::updateBounds() {
    bounds.clear();
    if (!lineCount()) return;
    QListIterator<Line *> lineIt(lines());
    Line *line = lineIt.next();
    V2d_CopyBox(bounds.arvec2, line->bounds().arvec2);
    while (lineIt.hasNext()) {
        line = lineIt.next();
        V2d_UniteBox(bounds.arvec2, line->bounds().arvec2);
    }
}
void Polyobj::updateSurfaceTangents() {
    for (Line *line : lines()) {
        line->forAllSides([] (LineSide &side) {
            side.updateAllSurfaceNormals();
            return LoopContinue;
        });
    }
}
bool Polyobj::move(Vector2d const &delta) {
    LOG_AS("Polyobj::move");
    unlink(); {
        PolyobjData::VertexCoords::iterator prevCoordsIt = data().prevPts.begin();
        for (Vertex *vertex : uniqueVertexes()) {
            (*prevCoordsIt) = vertex->origin();
            vertex->setOrigin(vertex->origin() + delta);
            prevCoordsIt++;
        }
        Vector2d newOrigin = Vector2d(origin) + delta;
        V2d_Set(origin, newOrigin.x, newOrigin.y);
        updateBounds();
    }
    link();
    if (blocked()) {
        unlink(); {
            PolyobjData::VertexCoords::const_iterator prevCoordsIt = data().prevPts.constBegin();
            for (Vertex *vertex : uniqueVertexes()) {
                vertex->setOrigin(*prevCoordsIt);
                prevCoordsIt++;
            }
            Vector2d newOrigin = Vector2d(origin) - delta;
            V2d_Set(origin, newOrigin.x, newOrigin.y);
            updateBounds();
        }
        link();
        return false;
    }
#if 0
    notifyGeometryChanged(*this);
#endif
    return true;
}
/**
 * @param point      Point to be rotated (in-place).
 * @param about      Origin to rotate @a point relative to.
 * @param fineAngle  Angle to rotate (theta).
 */
static void rotatePoint2d(Vector2d &point, Vector2d const &about, duint fineAngle) {
    ddouble const c = FIX2DBL(fineCosine[fineAngle]);
    ddouble const s = FIX2DBL(finesine[fineAngle]);
    Vector2d orig = point;
    point.x = orig.x * c - orig.y * s + about.x;
    point.y = orig.y * c + orig.x * s + about.y;
}
bool Polyobj::rotate(angle_t delta) {
    LOG_AS("Polyobj::rotate");
    unlink(); {
        duint fineAngle = (angle + delta) >> ANGLETOFINESHIFT;
        PolyobjData::VertexCoords::const_iterator origCoordsIt = data().originalPts.constBegin();
        PolyobjData::VertexCoords::iterator prevCoordsIt = data().prevPts.begin();
        for (Vertex *vertex : uniqueVertexes()) {
            (*prevCoordsIt) = vertex->origin();
            Vector2d newCoords = (*origCoordsIt);
            rotatePoint2d(newCoords, origin, fineAngle);
            vertex->setOrigin(newCoords);
            origCoordsIt++;
            prevCoordsIt++;
        }
        updateBounds();
        angle += delta;
    }
    link();
    if (blocked()) {
        unlink(); {
            PolyobjData::VertexCoords::const_iterator prevCoordsIt = data().prevPts.constBegin();
            for (Vertex *vertex : uniqueVertexes()) {
                vertex->setOrigin(*prevCoordsIt);
                prevCoordsIt++;
            }
            updateBounds();
            angle -= delta;
        }
        link();
        return false;
    }
    updateSurfaceTangents();
#if 0
    notifyGeometryChanged(*this);
#endif
    return true;
}
void Polyobj::setTag(dint newTag) {
    tag = newTag;
}
void Polyobj::setSequenceType(dint newType) {
    seqType = newType;
}
dint Polyobj::indexInMap() const {
    return data().indexInMap;
}
void Polyobj::setIndexInMap(dint newIndex) {
    data().indexInMap = newIndex;
}
PolyobjData::PolyobjData() {
    mesh       = new Mesh;
    indexInMap = MapElement::NoIndex;
}
PolyobjData::~PolyobjData() {
    delete mesh;
}
void PolyobjData::setThinker(thinker_s *thinker) {
    _polyobj = (Polyobj *) thinker;
}
void PolyobjData::think() {
}
Thinker::IData *PolyobjData::duplicate() const {
    return new PolyobjData(*this);
}
#ifdef __CLIENT__
void PolyobjData::addMover(ClPolyMover &mover) {
    if(_mover) {
        Thinker_Map(_mover->thinker()).thinkers().remove(_mover->thinker());
        DENG2_ASSERT(!_mover);
    }
    _mover = &mover;
}
void PolyobjData::removeMover(ClPolyMover &mover) {
    if(_mover == &mover) {
        _mover = 0;
    }
}
ClPolyMover *PolyobjData::mover() const {
    return _mover;
}
#endif
static Sector::SubsectorConstructor subsectorConstructor;
DENG2_PIMPL(Sector)
, DENG2_OBSERVES(Plane, HeightChange) {
    /**
     * POD: Metrics describing the geometry of the sector (the subsectors).
     */
    struct GeomData {
        AABoxd bounds;              ///< Bounding box for the whole sector (all subsectors).
        ddouble roughArea = 0;      ///< Rough approximation.
    };
    struct MapObjects {
        mobj_t *head = nullptr;     ///< The list of map objects.
        /**
         * Returns @c true if the map-object @a mob is linked.
         */
        bool contains(mobj_t const *mob) const {
            if (mob) {
                for (mobj_t const *it = head; it; it = it->sNext) {
                    if (it == mob) return true;
                }
            }
            return false;
        }
        void add(mobj_t *mob) {
            if (!mob) return;
            DENG2_ASSERT(!contains(mob));
            if ((mob->sNext = head)) {
                mob->sNext->sPrev = &mob->sNext;
            }
            *(mob->sPrev = &head) = mob;
        }
        /**
         * Two links to update:
         * 1) The link to the mobj from the previous node (sprev, always set) will
         *    be modified to point to the node following it.
         * 2) If there is a node following the mobj, set its sprev pointer to point
         *    to the pointer that points back to it (the mobj's sprev, just modified).
         */
        void remove(mobj_t *mob) {
            if (!mob || !Mobj_IsSectorLinked(*mob)) return;
            if ((*mob->sPrev = mob->sNext)) {
                mob->sNext->sPrev = mob->sPrev;
            }
            mob->sNext = nullptr;
            mob->sPrev = nullptr;
            DENG2_ASSERT(!contains(mob));
        }
    };
    struct Planes : public QVector<Plane *> {
        ~Planes() { clear(); }
        void clear() {
            qDeleteAll(*this);
            QVector<Plane *>::clear();
        }
    };
    struct Subsectors : public QVector<Subsector *> {
        ~Subsectors() { clear(); }
        void clear() {
            qDeleteAll(*this);
            QVector<Subsector *>::clear();
        }
    };
    Planes planes;                   ///< Planes of the sector.
    MapObjects mapObjects;           ///< All map-objects "in" one of the subsectors (not owned).
    QVector<LineSide *> sides;       ///< All line sides referencing the sector (not owned).
    Subsectors subsectors;           ///< Traversable subsectors of the sector.
    ThinkerT<SoundEmitter> emitter;  ///< Head of the sound emitter chain.
    int visPlaneLinkSector = MapElement::NoIndex;
    int visPlaneLinkBits = 0;
    dfloat lightLevel = 0;           ///< Ambient light level.
    Vector3f lightColor;             ///< Ambient light color.
    std::unique_ptr<GeomData> gdata; ///< Additional geometry info/metrics (cache).
    dint validCount = 0;             ///< Used by legacy algorithms to prevent repeated processing.
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        planes.clear();
        delete [] self()._lookupPlanes;
    }
    /**
     * Returns the additional geometry info/metrics from the cache.
     */
    GeomData &geom() {
        if (!gdata) {
            std::unique_ptr<GeomData> gd(new GeomData);
            gd->bounds    = findBounds();
            gd->roughArea = findRoughArea();
            gdata.reset(gd.get());
            gd.release();
            emitter->origin[0] = (gdata->bounds.minX + gdata->bounds.maxX) / 2;
            emitter->origin[1] = (gdata->bounds.minY + gdata->bounds.maxY) / 2;
        }
        return *gdata;
    }
    /**
     * Calculate the minimum bounding rectangle containing all the subsector geometries.
     */
    AABoxd findBounds() const {
        bool inited = false;
        AABoxd bounds;
        for (Subsector const *subsec : subsectors) {
            if (inited) {
                V2d_UniteBox(bounds.arvec2, subsec->bounds().arvec2);
            }
            else {
                bounds = subsec->bounds();
                inited = true;
            }
        }
        return bounds;
    }
    /**
     * Approximate the total area of all the subsector geometries.
     */
    ddouble findRoughArea() const {
        ddouble roughArea = 0;
        for (Subsector const *subsec : subsectors) {
            roughArea += subsec->roughArea();
        }
        return roughArea;
    }
    void updateEmitterOriginZ() {
        emitter->origin[2] = (self().floor().height() + self().ceiling().height()) / 2;
    }
    void updateSideEmitterOrigins() {
        for (LineSide *side : sides) {
            side->updateAllSoundEmitterOrigins();
            side->back().updateAllSoundEmitterOrigins();
        }
    }
    void updateAllEmitterOrigins() {
        updateEmitterOriginZ();
        updateSideEmitterOrigins();
    }
    void planeHeightChanged(Plane &) {
        updateAllEmitterOrigins();
    }
    void updatePlanesLookup() {
        delete [] self()._lookupPlanes;
        self()._lookupPlanes = new Plane *[planes.size()];
        Plane **ptr = self()._lookupPlanes;
        for (Plane *p : planes) {
            *ptr++ = p;
        }
    }
    DENG2_PIMPL_AUDIENCE(LightLevelChange)
    DENG2_PIMPL_AUDIENCE(LightColorChange)
};
DENG2_AUDIENCE_METHOD(Sector, LightLevelChange)
DENG2_AUDIENCE_METHOD(Sector, LightColorChange)
Sector::Sector(dfloat lightLevel, Vector3f const &lightColor)
    : MapElement(DMU_SECTOR)
    , d(new Impl(this))
    , _lookupPlanes(nullptr) {
    d->lightLevel = de::clamp(0.f, lightLevel, 1.f);
    d->lightColor = lightColor.min(Vector3f(1, 1, 1)).max(Vector3f(0, 0, 0));
}
void Sector::unlink(mobj_t *mob) {
    d->mapObjects.remove(mob);
}
void Sector::link(mobj_t *mob) {
    d->mapObjects.add(mob);
}
struct mobj_s *Sector::firstMobj() const {
    return d->mapObjects.head;
}
bool Sector::hasSkyMaskPlane() const {
    for (Plane *plane : d->planes) {
        if (plane->surface().hasSkyMaskedMaterial())
            return true;
    }
    return false;
}
dint Sector::planeCount() const {
    return d->planes.count();
}
LoopResult Sector::forAllPlanes(std::function<LoopResult (Plane &)> func) {
    for (Plane *plane : d->planes) {
        if(auto result = func(*plane)) return result;
    }
    return LoopContinue;
}
LoopResult Sector::forAllPlanes(std::function<LoopResult (Plane const &)> func) const {
    for (Plane const *plane : d->planes) {
        if(auto result = func(*plane)) return result;
    }
    return LoopContinue;
}
Plane *Sector::addPlane(Vector3f const &normal, ddouble height) {
    auto *plane = new Plane(*this, normal, height);
    plane->setIndexInSector(d->planes.count());
    d->planes.append(plane);
    d->updatePlanesLookup();
    if (plane->isSectorFloor() || plane->isSectorCeiling()) {
        plane->audienceForHeightChange() += d;
    }
    if (planeCount() == 2) {
        d->updateEmitterOriginZ();
    }
    return plane;
}
void Sector::setVisPlaneLinks(int sectorArchiveIndex, int planeBits) {
    d->visPlaneLinkSector = sectorArchiveIndex;
    d->visPlaneLinkBits   = planeBits;
}
int Sector::visPlaneLinkTargetSector() const {
    return d->visPlaneLinkSector;
}
bool Sector::isVisPlaneLinked(int planeIndex) const {
    return (d->visPlaneLinkBits & (1 << planeIndex)) != 0;
}
int Sector::visPlaneBits() const {
    return d->visPlaneLinkBits;
}
bool Sector::hasSubsectors() const {
    return !d->subsectors.isEmpty();
}
dint Sector::subsectorCount() const {
    return d->subsectors.count();
}
Subsector &Sector::subsector(int index) const {
    DENG2_ASSERT(index >= 0 && index < d->subsectors.count());
    return *d->subsectors.at(index);
}
LoopResult Sector::forAllSubsectors(const std::function<LoopResult(Subsector &)> &callback) const {
    for (Subsector *subsec : d->subsectors) {
        if (auto result = callback(*subsec)) return result;
    }
    return LoopContinue;
}
Subsector *Sector::addSubsector(QVector<ConvexSubspace *> const &subspaces) {
    DENG2_ASSERT(subsectorConstructor);
    std::unique_ptr<Subsector> subsec(subsectorConstructor(subspaces));
    d->subsectors << subsec.get();
    LOG_MAP_XVERBOSE("New Subsector %s (sector-%s)", subsec->id().asText() << indexInMap());
    return subsec.release();
}
dint Sector::sideCount() const {
    return d->sides.count();
}
LoopResult Sector::forAllSides(std::function<LoopResult (LineSide &)> func) const {
    for (LineSide *side : d->sides) {
        if (auto result = func(*side)) return result;
    }
    return LoopContinue;
}
void Sector::buildSides() {
    d->sides.clear();
    dint count = 0;
    map().forAllLines([this, &count] (Line &line) {
        if (line.front().sectorPtr() == this || line.back().sectorPtr() == this) {
            count += 1;
        }
        return LoopContinue;
    });
    if (!count) return;
    d->sides.reserve(count);
    map().forAllLines([this] (Line &line) {
        if (line.front().sectorPtr() == this) {
            d->sides.append(&line.front()); // Ownership not given.
        }
        else if (line.back().sectorPtr()  == this) {
            d->sides.append(&line.back()); // Ownership not given.
        }
        return LoopContinue;
    });
    d->updateAllEmitterOrigins();
}
SoundEmitter &Sector::soundEmitter() {
    (void) d->geom();
    return d->emitter;
}
SoundEmitter const &Sector::soundEmitter() const {
    return const_cast<SoundEmitter const &>(const_cast<Sector &>(*this).soundEmitter());
}
static void linkSoundEmitter(SoundEmitter &root, SoundEmitter &newEmitter) {
    newEmitter.thinker.prev = &root.thinker;
    newEmitter.thinker.next = root.thinker.next;
    if (newEmitter.thinker.next)
        newEmitter.thinker.next->prev = &newEmitter.thinker;
    root.thinker.next = &newEmitter.thinker;
}
void Sector::chainSoundEmitters() {
    SoundEmitter &root = d->emitter;
    root.thinker.next = root.thinker.prev = nullptr;
    for (Plane *plane : d->planes) {
        linkSoundEmitter(root, plane->soundEmitter());
    }
    for (LineSide *side : d->sides) {
        if (side->hasSections()) {
            linkSoundEmitter(root, side->middleSoundEmitter());
            linkSoundEmitter(root, side->bottomSoundEmitter());
            linkSoundEmitter(root, side->topSoundEmitter   ());
        }
        if (side->line().isSelfReferencing() && side->back().hasSections()) {
            LineSide &back = side->back();
            linkSoundEmitter(root, back.middleSoundEmitter());
            linkSoundEmitter(root, back.bottomSoundEmitter());
            linkSoundEmitter(root, back.topSoundEmitter   ());
        }
    }
}
dfloat Sector::lightLevel() const {
    return d->lightLevel;
}
void Sector::setLightLevel(dfloat newLightLevel) {
    newLightLevel = de::clamp(0.f, newLightLevel, 1.f);
    if (!de::fequal(d->lightLevel, newLightLevel)) {
        d->lightLevel = newLightLevel;
        DENG2_FOR_AUDIENCE2(LightLevelChange, i) i->sectorLightLevelChanged(*this);
    }
}
Vector3f const &Sector::lightColor() const {
    return d->lightColor;
}
void Sector::setLightColor(Vector3f const &newLightColor) {
    auto newColorClamped = newLightColor.min(Vector3f(1, 1, 1)).max(Vector3f(0, 0, 0));
    if (d->lightColor != newColorClamped) {
        d->lightColor = newColorClamped;
        DENG2_FOR_AUDIENCE2(LightColorChange, i) i->sectorLightColorChanged(*this);
    }
}
dint Sector::validCount() const {
    return d->validCount;
}
void Sector::setValidCount(dint newValidCount) {
    d->validCount = newValidCount;
}
AABoxd const &Sector::bounds() const {
    return d->geom().bounds;
}
#ifdef __CLIENT__
ddouble Sector::roughArea() const {
    return d->geom().roughArea;
}
#endif
dint Sector::property(DmuArgs &args) const {
    switch (args.prop) {
    case DMU_LIGHT_LEVEL:
        args.setValue(DMT_SECTOR_LIGHTLEVEL, &d->lightLevel, 0);
        break;
    case DMU_COLOR:
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.x, 0);
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.y, 1);
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.z, 2);
        break;
    case DMU_COLOR_RED:
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.x, 0);
        break;
    case DMU_COLOR_GREEN:
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.y, 0);
        break;
    case DMU_COLOR_BLUE:
        args.setValue(DMT_SECTOR_RGB, &d->lightColor.z, 0);
        break;
    case DMU_EMITTER: {
        SoundEmitter const *emitterAdr = d->emitter;
        args.setValue(DMT_SECTOR_EMITTER, &emitterAdr, 0);
        break; }
    case DMT_MOBJS:
        args.setValue(DMT_SECTOR_MOBJLIST, &d->mapObjects.head, 0);
        break;
    case DMU_VALID_COUNT:
        args.setValue(DMT_SECTOR_VALIDCOUNT, &d->validCount, 0);
        break;
    case DMU_FLOOR_PLANE: {
        Plane *pln = d->planes.at(Floor);
        args.setValue(DMT_SECTOR_FLOORPLANE, &pln, 0);
        break; }
    case DMU_CEILING_PLANE: {
        Plane *pln = d->planes.at(Ceiling);
        args.setValue(DMT_SECTOR_CEILINGPLANE, &pln, 0);
        break; }
    default:
        return MapElement::property(args);
    }
    return false;  // Continue iteration.
}
dint Sector::setProperty(DmuArgs const &args) {
    switch (args.prop) {
    case DMU_COLOR: {
        Vector3f newColor = d->lightColor;
        args.value(DMT_SECTOR_RGB, &newColor.x, 0);
        args.value(DMT_SECTOR_RGB, &newColor.y, 1);
        args.value(DMT_SECTOR_RGB, &newColor.z, 2);
        setLightColor(newColor);
        break; }
    case DMU_COLOR_RED: {
        Vector3f newColor = d->lightColor;
        args.value(DMT_SECTOR_RGB, &newColor.x, 0);
        setLightColor(newColor);
        break; }
    case DMU_COLOR_GREEN: {
        Vector3f newColor = d->lightColor;
        args.value(DMT_SECTOR_RGB, &newColor.y, 0);
        setLightColor(newColor);
        break; }
    case DMU_COLOR_BLUE: {
        Vector3f newColor = d->lightColor;
        args.value(DMT_SECTOR_RGB, &newColor.z, 0);
        setLightColor(newColor);
        break; }
    case DMU_LIGHT_LEVEL: {
        dfloat newLightLevel;
        args.value(DMT_SECTOR_LIGHTLEVEL, &newLightLevel, 0);
        setLightLevel(newLightLevel);
        break; }
    case DMU_VALID_COUNT:
        args.value(DMT_SECTOR_VALIDCOUNT, &d->validCount, 0);
        break;
    default:
        return MapElement::setProperty(args);
    }
    return false;  // Continue iteration.
}
D_CMD(InspectSector) {
    DENG2_UNUSED(src);
    LOG_AS("inspectsector (Cmd)");
    if (argc != 2) {
        LOG_SCR_NOTE("Usage: %s (sector-id)") << argv[0];
        return true;
    }
    if (!App_World().hasMap()) {
        LOG_SCR_ERROR("No map is currently loaded");
        return false;
    }
    dint const index  = String(argv[1]).toInt();
    Sector const *sec = App_World().map().sectorPtr(index);
    if (!sec) {
        LOG_SCR_ERROR("Sector #%i not found") << index;
        return false;
    }
    LOG_SCR_MSG(_E(b) "Sector %s" _E(.) " [%p]")
            << Id(sec->indexInMap()).asText() << sec;
    LOG_SCR_MSG(    _E(l) "Bounds: "      _E(.)_E(i) "%s" _E(.)
                " " _E(l) "Light Color: " _E(.)_E(i) "%s" _E(.)
                " " _E(l) "Light Level: " _E(.)_E(i) "%f")
            << Rectangled(sec->bounds().min, sec->bounds().max).asText()
            << sec->lightColor().asText()
            << sec->lightLevel();
    if (sec->planeCount()) {
        LOG_SCR_MSG(_E(D) "Planes (%i):") << sec->planeCount();
        sec->forAllPlanes([] (Plane const &plane) {
            LOG_SCR_MSG("%s: " _E(>))
                << Sector::planeIdAsText(plane.indexInSector()).upperFirstChar()
                << plane.description();
            return LoopContinue;
        });
    }
    if (sec->subsectorCount()) {
        LOG_SCR_MSG(_E(D) "Subsectors (%i):") << sec->subsectorCount();
        dint subsectorIndex = 0;
        sec->forAllSubsectors([&subsectorIndex] (Subsector const &subsec) {
            LOG_SCR_MSG("%i: " _E(>))
                << subsectorIndex
                << subsec.description();
            subsectorIndex += 1;
            return LoopContinue;
        });
    }
    return true;
}
String Sector::planeIdAsText(dint planeId) {
    switch (planeId) {
    case Floor:   return "floor";
    case Ceiling: return "ceiling";
    default:      return "plane-" + String::number(planeId);
    }
}
void Sector::consoleRegister()  // static
{
    C_CMD("inspectsector", "i", InspectSector);
}
void Sector::setSubsectorConstructor(SubsectorConstructor func) // static
{
    subsectorConstructor = func;
}
namespace world {
DENG2_PIMPL_NOREF(Sky::Layer) {
    bool active         = false;
    bool masked         = false;
    Material *material = nullptr;
    dfloat offset       = 0;
    dfloat fadeOutLimit = 0;
    Sky &sky;
    Impl(Sky &sky) : sky(sky) {}
    DENG2_PIMPL_AUDIENCE(ActiveChange)
    DENG2_PIMPL_AUDIENCE(MaskedChange)
    DENG2_PIMPL_AUDIENCE(MaterialChange)
};
DENG2_AUDIENCE_METHOD(Sky::Layer, ActiveChange)
DENG2_AUDIENCE_METHOD(Sky::Layer, MaskedChange)
DENG2_AUDIENCE_METHOD(Sky::Layer, MaterialChange)
Sky::Layer::Layer(Sky &sky, Material *material) : d(new Impl(sky)) {
    setMaterial(material);
}
Sky &Sky::Layer::sky() const {
    return d->sky;
}
bool Sky::Layer::isActive() const {
    return d->active;
}
void Sky::Layer::setActive(bool yes) {
    if(d->active != yes) {
        d->active = yes;
        DENG2_FOR_AUDIENCE2(ActiveChange, i) i->skyLayerActiveChanged(*this);
    }
}
bool Sky::Layer::isMasked() const {
    return d->masked;
}
void Sky::Layer::setMasked(bool yes) {
    if(d->masked != yes) {
        d->masked = yes;
        DENG2_FOR_AUDIENCE2(MaskedChange, i) i->skyLayerMaskedChanged(*this);
    }
}
Material *Sky::Layer::material() const {
    return d->material;
}
void Sky::Layer::setMaterial(Material *newMaterial) {
    if(d->material != newMaterial) {
        d->material = newMaterial;
        DENG2_FOR_AUDIENCE2(MaterialChange, i) i->skyLayerMaterialChanged(*this);
    }
}
dfloat Sky::Layer::offset() const {
    return d->offset;
}
void Sky::Layer::setOffset(dfloat newOffset) {
    d->offset = newOffset;
}
dfloat Sky::Layer::fadeOutLimit() const {
    return d->fadeOutLimit;
}
void Sky::Layer::setFadeoutLimit(dfloat newLimit) {
    d->fadeOutLimit = newLimit;
}
#ifdef __CLIENT__
static Vector3f const AmbientLightColorDefault(1, 1, 1); // Pure white.
#endif
DENG2_PIMPL(Sky)
#ifdef __CLIENT__
, DENG2_OBSERVES(Layer, ActiveChange)
, DENG2_OBSERVES(Layer, MaterialChange)
, DENG2_OBSERVES(Layer, MaskedChange)
#endif
{
    struct Layers : public QList<Layer *> {
        ~Layers() { qDeleteAll(*this); }
    };
    Layers layers;
    Record const *def    = nullptr; ///< Sky definition.
    dfloat height        = 1;
    dfloat horizonOffset = 0;
    Impl(Public *i) : Base(i) {
        for(dint i = 0; i < NUM_LAYERS; ++i) {
            layers.append(new Layer(self()));
#ifdef __CLIENT__
            Layer *layer = layers.last();
            layer->audienceForActiveChange()   += this;
            layer->audienceForMaskedChange()   += this;
            layer->audienceForMaterialChange() += this;
#endif
        }
    }
    ~Impl() {
        DENG2_FOR_PUBLIC_AUDIENCE2(Deletion, i) i->skyBeingDeleted(self());
    }
#ifdef __CLIENT__
    /**
     * Ambient lighting characteristics.
     */
    struct AmbientLight {
        bool custom     = false;  /// @c true= defined in a MapInfo def.
        bool needUpdate = true;   /// @c true= update if not custom.
        Vector3f color;
        void setColor(Vector3f const &newColor, bool isCustom = true) {
            color  = newColor.min(Vector3f(1, 1, 1)).max(Vector3f(0, 0, 0));
            custom = isCustom;
        }
        void reset() {
            custom     = false;
            color      = AmbientLightColorDefault;
            needUpdate = true;
        }
    } ambientLight;
    /**
     * @todo Move to SkyDrawable and have it simply update this component once the
     * ambient color has been calculated.
     *
     * @todo Re-implement me by rendering the sky to a low-quality cubemap and use
     * that to obtain the lighting characteristics.
     */
    void updateAmbientLightIfNeeded() {
        if(ambientLight.custom) return;
        if(!ambientLight.needUpdate) return;
        ambientLight.needUpdate = false;
        ambientLight.color = AmbientLightColorDefault;
        dint firstActiveLayer = -1; // -1 denotes 'no active layers'.
        for(dint i = 0; i < layers.count(); ++i) {
            if(layers[i]->isActive()) {
                firstActiveLayer = i;
                break;
            }
        }
        if(firstActiveLayer < 0) return;
        Vector3f avgLayerColor;
        Vector3f bottomCapColor;
        Vector3f topCapColor;
        dint avgCount = 0;
        for(dint i = firstActiveLayer; i < layers.count(); ++i) {
            Layer &layer = *layers[i];
            if(!layer.isActive()) continue;
            ClientMaterial *mat = static_cast<ClientMaterial *>(layer.material());
            if(!mat) continue;
            MaterialAnimator &matAnimator = mat->getAnimator(SkyDrawable::layerMaterialSpec(layer.isMasked()));
            matAnimator.prepare();
            if(TextureVariant *tex = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture) {
                auto const *avgColor = reinterpret_cast<averagecolor_analysis_t const *>(tex->base().analysisDataPointer(ClientTexture::AverageColorAnalysis));
                if(!avgColor) throw Error("calculateSkyAmbientColor", "Texture \"" + tex->base().manifest().composeUri().asText() + "\" has no AverageColorAnalysis");
                if(i == firstActiveLayer) {
                    auto const *avgLineColor = reinterpret_cast<averagecolor_analysis_t const *>(tex->base().analysisDataPointer(ClientTexture::AverageTopColorAnalysis));
                    if(!avgLineColor) throw Error("calculateSkyAmbientColor", "Texture \"" + tex->base().manifest().composeUri().asText() + "\" has no AverageTopColorAnalysis");
                    topCapColor = Vector3f(avgLineColor->color.rgb);
                    avgLineColor = reinterpret_cast<averagecolor_analysis_t const *>(tex->base().analysisDataPointer(ClientTexture::AverageBottomColorAnalysis));
                    if(!avgLineColor) throw Error("calculateSkyAmbientColor", "Texture \"" +  tex->base().manifest().composeUri().asText() + "\" has no AverageBottomColorAnalysis");
                    bottomCapColor = Vector3f(avgLineColor->color.rgb);
                }
                avgLayerColor += Vector3f(avgColor->color.rgb);
                ++avgCount;
            }
        }
        ambientLight.setColor((avgLayerColor + topCapColor + bottomCapColor) / (avgCount + 2),
                              false /*not a custom color*/);
    }
    void skyLayerActiveChanged(Layer &) {
        ambientLight.needUpdate = true;
    }
    void skyLayerMaterialChanged(Layer &layer) {
        if(!layer.isActive()) return;
        ambientLight.needUpdate = true;
    }
    void skyLayerMaskedChanged(Layer &layer) {
        if(!layer.isActive()) return;
        ambientLight.needUpdate = true;
    }
#endif  // __CLIENT__
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(HeightChange)
    DENG2_PIMPL_AUDIENCE(HorizonOffsetChange)
};
DENG2_AUDIENCE_METHOD(Sky, Deletion)
DENG2_AUDIENCE_METHOD(Sky, HeightChange)
DENG2_AUDIENCE_METHOD(Sky, HorizonOffsetChange)
Sky::Sky(defn::Sky const *definition)
    : MapElement(DMU_SKY)
    , d(new Impl(this)) {
    configure(definition);
}
void Sky::configure(defn::Sky const *def) {
    LOG_AS("Sky");
    d->def = def? def->accessedRecordPtr() : 0;
    setHeight(def? def->getf("height") : DEFAULT_SKY_HEIGHT);
    setHorizonOffset(def? def->getf("horizonOffset") : DEFAULT_SKY_HORIZON_OFFSET);
    for(dint i = 0; i < d->layers.count(); ++i) {
        Record const *lyrDef = def? &def->layer(i) : 0;
        Layer &lyr = *d->layers[i];
        lyr.setMasked(lyrDef? (lyrDef->geti("flags") & SLF_MASK) : false);
        lyr.setOffset(lyrDef? lyrDef->getf("offset") : DEFAULT_SKY_SPHERE_XOFFSET);
        lyr.setFadeoutLimit(lyrDef? lyrDef->getf("colorLimit") : DEFAULT_SKY_SPHERE_FADEOUT_LIMIT);
        de::Uri const matUri = lyrDef? de::makeUri(lyrDef->gets("material")) : de::makeUri(DEFAULT_SKY_SPHERE_MATERIAL);
        world::Material *mat = 0;
        try {
            mat = world::Materials::get().materialPtr(matUri);
        }
        catch(world::MaterialManifest::MissingMaterialError const &er) {
            if(lyrDef) {
                LOG_RES_WARNING(er.asText() + ". Unknown material \"%s\" in definition layer %i, using default")
                        << matUri << i;
            }
        }
        lyr.setMaterial(mat);
        lyr.setActive(lyrDef? (lyrDef->geti("flags") & SLF_ENABLE) : i == 0);
    }
#ifdef __CLIENT__
    if(def) {
        Vector3f color(def->get("color"));
        if(color != Vector3f(0, 0, 0)) {
            d->ambientLight.setColor(color);
        }
    }
    else {
        d->ambientLight.reset();
    }
#endif
}
Record const *Sky::def() const {
    return d->def;
}
dint Sky::layerCount() const {
    return d->layers.count();
}
bool Sky::hasLayer(dint layerIndex) const {
    return !d->layers.isEmpty() && layerIndex < d->layers.count();
}
Sky::Layer *Sky::layerPtr(dint layerIndex) const {
    if (hasLayer(layerIndex)) return d->layers.at(layerIndex);
    return nullptr;
}
Sky::Layer &Sky::layer(dint layerIndex) {
    if (Layer *layer = layerPtr(layerIndex)) return *layer;
    throw MissingLayerError("Sky::layer", "Unknown layer #" + QString::number(layerIndex));
}
Sky::Layer const &Sky::layer(dint layerIndex) const {
    if (Layer *layer = layerPtr(layerIndex)) return *layer;
    throw MissingLayerError("Sky::layer", "Unknown layer #" + QString::number(layerIndex));
}
LoopResult Sky::forAllLayers(std::function<LoopResult (Layer &)> func) {
    for (Layer *layer : d->layers) {
        if (auto result = func(*layer)) return result;
    }
    return LoopContinue;
}
LoopResult Sky::forAllLayers(std::function<LoopResult (Layer const &)> func) const {
    for (Layer const *layer : d->layers) {
        if (auto result = func(*layer)) return result;
    }
    return LoopContinue;
}
dfloat Sky::height() const {
    return d->height;
}
void Sky::setHeight(dfloat newHeight) {
    newHeight = de::clamp(0.f, newHeight, 1.f);
    if(!de::fequal(d->height, newHeight)) {
        d->height = newHeight;
        DENG2_FOR_AUDIENCE2(HeightChange, i) i->skyHeightChanged(*this);
    }
}
dfloat Sky::horizonOffset() const {
    return d->horizonOffset;
}
void Sky::setHorizonOffset(dfloat newOffset) {
    if(!de::fequal(d->horizonOffset, newOffset)) {
        d->horizonOffset = newOffset;
        DENG2_FOR_AUDIENCE2(HorizonOffsetChange, i) i->skyHorizonOffsetChanged(*this);
    }
}
dint Sky::property(DmuArgs &args) const {
    LOG_AS("Sky");
    switch(args.prop) {
    case DMU_FLAGS: {
        dint flags = 0;
        if(layer(0).isActive()) flags |= SKYF_LAYER0_ENABLED;
        if(layer(1).isActive()) flags |= SKYF_LAYER1_ENABLED;
        args.setValue(DDVT_INT, &flags, 0);
        break; }
    case DMU_HEIGHT:
        args.setValue(DDVT_FLOAT, &d->height, 0);
        break;
    /*case DMU_HORIZONOFFSET:
        args.setValue(DDVT_FLOAT, &d->horizonOffset, 0);
        break;*/
    default: return MapElement::property(args);
    }
    return false; // Continue iteration.
}
dint Sky::setProperty(DmuArgs const &args) {
    LOG_AS("Sky");
    switch(args.prop) {
    case DMU_FLAGS: {
        dint flags = 0;
        if(layer(0).isActive()) flags |= SKYF_LAYER0_ENABLED;
        if(layer(1).isActive()) flags |= SKYF_LAYER1_ENABLED;
        args.value(DDVT_INT, &flags, 0);
        layer(0).setActive(flags & SKYF_LAYER0_ENABLED);
        layer(1).setActive(flags & SKYF_LAYER1_ENABLED);
        break; }
    case DMU_HEIGHT: {
        dfloat newHeight = d->height;
        args.value(DDVT_FLOAT, &newHeight, 0);
        setHeight(newHeight);
        break; }
    /*case DMU_HORIZONOFFSET: {
        float newOffset = d->horizonOffset;
        args.value(DDVT_FLOAT, &d->horizonOffset, 0);
        setHorizonOffset(newOffset);
        break; }*/
    default: return MapElement::setProperty(args);
    }
    return false; // Continue iteration.
}
#ifdef __CLIENT__
Vector3f const &Sky::ambientColor() const {
    if(d->ambientLight.custom || rendSkyLightAuto) {
        d->updateAmbientLightIfNeeded();
        return d->ambientLight.color;
    }
    return AmbientLightColorDefault;
}
void Sky::setAmbientColor(Vector3f const &newColor) {
    d->ambientLight.setColor(newColor);
}
#endif // __CLIENT__
}  // namespace world
namespace world {
DENG2_PIMPL_NOREF(Subsector) {
    QVector<ConvexSubspace *> subspaces;
    std::unique_ptr<AABoxd> bounds;
    Id id;
    Sector *sector = nullptr;
    /**
     * Calculate the minimum bounding rectangle containing all the subspace geometries.
     */
    AABoxd findBounds() const {
        bool inited = false;
        AABoxd bounds;
        for (ConvexSubspace const *subspace : subspaces) {
            AABoxd const &subBounds = subspace->poly().bounds();
            if (inited) {
                V2d_UniteBox(bounds.arvec2, subBounds.arvec2);
            }
            else {
                bounds = subBounds;
                inited = true;
            }
        }
        return bounds;
    }
};
Subsector::Subsector(QVector<ConvexSubspace *> const &subspaces) : d(new Impl) {
    d->subspaces += subspaces;
    for (ConvexSubspace *subspace : subspaces) {
        subspace->setSubsector(this);
    }
}
Subsector::~Subsector() {
    DENG2_FOR_AUDIENCE(Deletion, i) i->subsectorBeingDeleted(*this);
}
String Subsector::description() const {
    auto desc = String(    _E(l) "Id: "     _E(.) _E(i) "%1" _E(.)
                       " " _E(l) "Sector: " _E(.) _E(i) "%2" _E(.)
                       " " _E(l) "Bounds: " _E(.) _E(i) "%3" _E(.))
                    .arg(d->id.asText())
                    .arg(sector().indexInMap())
                    .arg(Rectangled(bounds().min, bounds().max).asText());
    DENG2_DEBUG_ONLY(
        desc.prepend(String(_E(b) "Subsector " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
    )
    return desc;
}
Id Subsector::id() const {
    return d->id;
}
Sector &Subsector::sector() {
    if (!d->sector) {
        d->sector = firstSubspace().bspLeaf().sectorPtr();
        DENG2_ASSERT(d->sector);
    }
    return *d->sector;
}
Sector const &Subsector::sector() const {
    if (!d->sector) {
        d->sector = firstSubspace().bspLeaf().sectorPtr();
        DENG2_ASSERT(d->sector);
    }
    return *d->sector;
}
dint Subsector::subspaceCount() const {
    return d->subspaces.count();
}
ConvexSubspace &Subsector::firstSubspace() const {
    DENG2_ASSERT(!d->subspaces.isEmpty());
    return *d.getConst()->subspaces.first();
}
LoopResult Subsector::forAllSubspaces(std::function<LoopResult (ConvexSubspace &)> func) const {
    for (ConvexSubspace *sub : d->subspaces) {
        if (auto result = func(*sub)) return result;
    }
    return LoopContinue;
}
QList<HEdge *> Subsector::listUniqueBoundaryEdges() const {
    QList<HEdge *> list;
    for (ConvexSubspace const *sub : d->subspaces) {
        HEdge *hedge = sub->poly().hedge();
        do {
            if (hedge->hasMapElement() && !isInternalEdge(hedge)) {
                bool unique = true;
                for (HEdge *base : list) {
                    SubsectorCirculator subIt(base);
                    do {
                        HEdge &other = *subIt;
                        if (&other == hedge) {
                            unique = false;
                            break;
                        }
                    } while (&subIt.next() != base);
                }
                if (unique) {
                    list << hedge;
                }
            }
        } while ((hedge = &hedge->next()) != sub->poly().hedge());
    }
    return list;
}
AABoxd const &Subsector::bounds() const {
    if (subspaceCount() == 1) {
        return firstSubspace().poly().bounds();
    }
    if (!d->bounds) {
        d->bounds.reset(new AABoxd(d->findBounds()));
    }
    return *d->bounds;
}
Vector2d Subsector::center() const {
    return (Vector2d(bounds().min) + Vector2d(bounds().max)) / 2;
}
ddouble Subsector::roughArea() const {
    AABoxd const &box = bounds();
    return (box.maxX - box.minX) * (box.maxY - box.minY);
}
bool Subsector::isInternalEdge(HEdge *hedge) // static
{
    if (!hedge) return false;
    if (!hedge->hasFace() || !hedge->twin().hasFace()) return false;
    if (!hedge->face().hasMapElement() || hedge->face().mapElement().type() != DMU_SUBSPACE) return false;
    if (!hedge->twin().face().hasMapElement() || hedge->twin().face().mapElement().type() != DMU_SUBSPACE) return false;
    Subsector *frontSubsector = hedge->face().mapElementAs<ConvexSubspace>().subsectorPtr();
    if (!frontSubsector) return false;
    return frontSubsector == hedge->twin().face().mapElementAs<ConvexSubspace>().subsectorPtr();
}
#if 0
Subsector *SubsectorCirculator::getSubsector(HEdge const &hedge) // static
{
    if (!hedge.hasFace()) return nullptr;
    if (!hedge.face().hasMapElement()) return nullptr;
    if (hedge.face().mapElement().type() != DMU_SUBSPACE) return nullptr;
    return hedge.face().mapElementAs<ConvexSubspace>().subsectorPtr();
}
#endif
HEdge &SubsectorCirculator::getNeighbor(HEdge const &hedge, ClockDirection direction,
    Subsector const *subsec) // static
{
    HEdge *neighbor = &hedge.neighbor(direction);
    if (subsec) {
        while (neighbor->hasTwin() && subsec == neighbor->twin().subsector()) {
            neighbor = &neighbor->twin().neighbor(direction);
        }
    }
    return *neighbor;
}
} // namespace world
#ifdef DENG_DEBUG
static inline bool Surface_isSideMiddle(Surface const &suf) {
    return suf.parent().type() == DMU_SIDE
           && &suf == &suf.parent().as<LineSide>().middle();
}
static inline bool Surface_isSectorExtraPlane(Surface const &suf) {
    if (suf.parent().type() != DMU_PLANE) return false;
    auto const &plane = suf.parent().as<Plane>();
    return !(plane.isSectorFloor() || plane.isSectorCeiling());
}
#endif
DENG2_PIMPL(Surface) {
    dint flags = 0;                             ///< @ref sufFlags
    Matrix3f tangentMatrix { Matrix3f::Zero };  ///< Tangent space vectors.
    bool needUpdateTangentMatrix = false;       ///< @c true= marked for update.
    Material *material = nullptr;               ///< Currently bound material.
    bool materialIsMissingFix = false;          ///< @c true= @ref material is a "missing fix".
    Vector2f origin;                            ///< @em sharp offset to surface-space origin.
    Vector3f color;
    dfloat opacity = 0;
    blendmode_t blendMode { BM_NORMAL };
#ifdef __CLIENT__
    Vector2f oldOrigin[2];                      ///< Old @em sharp surface space material origins, for smoothing.
    Vector2f originSmoothed;                    ///< @em smoothed surface space material origin.
    Vector2f originSmoothedDelta;               ///< Delta between @em sharp and @em smoothed.
    MaterialAnimator *matAnimator = nullptr;
#endif
    Impl(Public *i) : Base(i)
    {}
#ifdef __CLIENT__
    ~Impl() {
        self().map().scrollingSurfaces().remove(thisPublic);
    }
#endif
    inline MapElement &owner() const { return self().parent(); }
#ifdef DENG_DEBUG
    bool isSideMiddle() const {
        return owner().type() == DMU_SIDE
               && thisPublic == &owner().as<LineSide>().middle();
    }
    bool isSectorExtraPlane() const {
        if (owner().type() != DMU_PLANE) return false;
        auto const &plane = owner().as<Plane>();
        return !(plane.isSectorFloor() || plane.isSectorCeiling());
    }
#endif
    void updateTangentMatrix() {
        needUpdateTangentMatrix = false;
        dfloat values[9];
        Vector3f normal = tangentMatrix.column(2);
        V3f_Set(values + 6, normal.x, normal.y, normal.z);
        V3f_BuildTangents(values, values + 3, values + 6);
        tangentMatrix = Matrix3f(values);
    }
#ifdef __CLIENT__
    void notifyOriginSmoothedChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE2(OriginSmoothedChange, i) i->surfaceOriginSmoothedChanged(self());
    }
#endif
    DENG2_PIMPL_AUDIENCE(ColorChange)
    DENG2_PIMPL_AUDIENCE(MaterialChange)
    DENG2_PIMPL_AUDIENCE(NormalChange)
    DENG2_PIMPL_AUDIENCE(OpacityChange)
    DENG2_PIMPL_AUDIENCE(OriginChange)
#ifdef __CLIENT__
    DENG2_PIMPL_AUDIENCE(OriginSmoothedChange)
#endif
};
DENG2_AUDIENCE_METHOD(Surface, ColorChange)
DENG2_AUDIENCE_METHOD(Surface, MaterialChange)
DENG2_AUDIENCE_METHOD(Surface, NormalChange)
DENG2_AUDIENCE_METHOD(Surface, OpacityChange)
DENG2_AUDIENCE_METHOD(Surface, OriginChange)
#ifdef __CLIENT__
DENG2_AUDIENCE_METHOD(Surface, OriginSmoothedChange)
#endif
Surface::Surface(MapElement &owner, dfloat opacity, Vector3f const &color)
    : MapElement(DMU_SURFACE, &owner)
    , d(new Impl(this)) {
    d->color   = color;
    d->opacity = opacity;
}
String Surface::description() const {
    auto desc = String(    _E(l) "Material: "        _E(.)_E(i) "%1" _E(.)
                       " " _E(l) "Material Origin: " _E(.)_E(i) "%2" _E(.)
                       " " _E(l) "Normal: "          _E(.)_E(i) "%3" _E(.)
                       " " _E(l) "Opacity: "         _E(.)_E(i) "%4" _E(.)
                       " " _E(l) "Blend Mode: "      _E(.)_E(i) "%5" _E(.)
                       " " _E(l) "Tint Color: "      _E(.)_E(i) "%6" _E(.))
                  .arg(hasMaterial() ? material().manifest().composeUri().asText() : "None")
                  .arg(origin().asText())
                  .arg(normal().asText())
                  .arg(opacity())
                  .arg(String(R_NameForBlendMode(blendMode())))
                  .arg(color().asText());
#ifdef DENG2_DEBUG
    desc.prepend(String(_E(b) "Surface " _E(.) "[0x%1]\n").arg(de::dintptr(this), 0, 16));
#endif
    return desc;
}
Matrix3f const &Surface::tangentMatrix() const {
    if (d->needUpdateTangentMatrix) {
        d->updateTangentMatrix();
    }
    return d->tangentMatrix;
}
Surface &Surface::setNormal(Vector3f const &newNormal) {
    Vector3f const oldNormal = normal();
    Vector3f const newNormalNormalized = newNormal.normalize();
    if (oldNormal != newNormalNormalized) {
        for (dint i = 0; i < 3; ++i) {
            d->tangentMatrix.at(i, 2) = newNormalNormalized[i];
        }
        d->needUpdateTangentMatrix = true;
        DENG2_FOR_AUDIENCE2(NormalChange, i) i->surfaceNormalChanged(*this);
    }
    return *this;
}
bool Surface::hasMaterial() const {
    return d->material != nullptr;
}
bool Surface::hasFixMaterial() const {
    return hasMaterial() && d->materialIsMissingFix;
}
Material &Surface::material() const {
    if (d->material) return *d->material;
    throw MissingMaterialError("Surface::material", "No material is bound");
}
Material *Surface::materialPtr() const {
    return d->material;
}
Surface &Surface::setMaterial(Material *newMaterial, bool isMissingFix) {
    DENG2_ASSERT(!(isMissingFix && (parent().type() == DMU_SIDE && parent().as<LineSide>().line().isSelfReferencing())));
    if (d->material == newMaterial)
        return *this;
    d->materialIsMissingFix = false;
    d->material = newMaterial;
    if (d->material && isMissingFix) {
        d->materialIsMissingFix = true;
    }
    if (::ddMapSetup && d->materialIsMissingFix && d->material) {
        if (d->owner().type() == DMU_SIDE) {
            auto &side = d->owner().as<LineSide>();
            dint section = (  this == &side.middle() ? LineSide::Middle
                            : this == &side.bottom() ? LineSide::Bottom
                            :                          LineSide::Top);
            LOGDEV_MAP_WARNING("%s of Line #%d is missing a material for the %s section."
                               "\n  %s was chosen to complete the definition.")
                << Line::sideIdAsText(side.sideId()).upperFirstChar()
                << side.line().indexInMap()
                << LineSide::sectionIdAsText(section)
                << d->material->manifest().composeUri().asText();
        }
    }
#ifdef __CLIENT__
    d->matAnimator = nullptr;
#endif
    DENG2_FOR_AUDIENCE2(MaterialChange, i) {
        i->surfaceMaterialChanged(*this);
    }
    return *this;
}
Vector2f const &Surface::origin() const {
    return d->origin;
}
Surface &Surface::setOrigin(Vector2f const &newOrigin) {
    if (d->origin != newOrigin) {
        d->origin = newOrigin;
#ifdef __CLIENT__
        if (::ddMapSetup) {
            d->originSmoothed = d->origin;
            d->originSmoothedDelta = Vector2f();
            d->oldOrigin[0] = d->oldOrigin[1] = d->origin;
        }
#endif
        DENG2_FOR_AUDIENCE2(OriginChange, i) i->surfaceOriginChanged(*this);
#ifdef __CLIENT__
        if (!::ddMapSetup) {
            map().scrollingSurfaces().insert(this);
        }
#endif
    }
    return *this;
}
bool Surface::materialMirrorX() const {
    return (d->flags & DDSUF_MATERIAL_FLIPH) != 0;
}
bool Surface::materialMirrorY() const {
    return (d->flags & DDSUF_MATERIAL_FLIPV) != 0;
}
Vector2f Surface::materialScale() const {
    return Vector2f(materialMirrorX()? -1 : 1, materialMirrorY()? -1 : 1);
}
de::Uri Surface::composeMaterialUri() const {
    if (!hasMaterial()) return de::Uri();
    return material().manifest().composeUri();
}
void Surface::setDecorationState(IDecorationState *state) {
    _decorationState.reset(state);
}
dfloat Surface::opacity() const {
    return d->opacity;
}
Surface &Surface::setOpacity(dfloat newOpacity) {
    DENG2_ASSERT(Surface_isSideMiddle(*this) || Surface_isSectorExtraPlane(*this));  // sanity check
    newOpacity = de::clamp(0.f, newOpacity, 1.f);
    if (!de::fequal(d->opacity, newOpacity)) {
        d->opacity = newOpacity;
        DENG2_FOR_AUDIENCE2(OpacityChange, i) i->surfaceOpacityChanged(*this);
    }
    return *this;
}
Vector3f const &Surface::color() const {
    return d->color;
}
Surface &Surface::setColor(Vector3f const &newColor) {
    Vector3f const newColorClamped(de::clamp(0.f, newColor.x, 1.f),
                                   de::clamp(0.f, newColor.y, 1.f),
                                   de::clamp(0.f, newColor.z, 1.f));
    if (d->color != newColorClamped) {
        d->color = newColorClamped;
        DENG2_FOR_AUDIENCE2(ColorChange, i) i->surfaceColorChanged(*this);
    }
    return *this;
}
blendmode_t Surface::blendMode() const {
    return d->blendMode;
}
Surface &Surface::setBlendMode(blendmode_t newBlendMode) {
    d->blendMode = newBlendMode;
    return *this;
}
dint Surface::property(DmuArgs &args) const {
    switch (args.prop) {
    case DMU_MATERIAL: {
        Material *mat = (d->materialIsMissingFix ? nullptr : d->material);
        args.setValue(DMT_SURFACE_MATERIAL, &mat, 0);
        break; }
    case DMU_OFFSET_X:
        args.setValue(DMT_SURFACE_OFFSET, &d->origin.x, 0);
        break;
    case DMU_OFFSET_Y:
        args.setValue(DMT_SURFACE_OFFSET, &d->origin.y, 0);
        break;
    case DMU_OFFSET_XY:
        args.setValue(DMT_SURFACE_OFFSET, &d->origin.x, 0);
        args.setValue(DMT_SURFACE_OFFSET, &d->origin.y, 1);
        break;
    case DMU_TANGENT_X:
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(0, 0), 0);
        break;
    case DMU_TANGENT_Y:
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(1, 0), 0);
        break;
    case DMU_TANGENT_Z:
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(2, 0), 0);
        break;
    case DMU_TANGENT_XYZ:
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(0, 0), 0);
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(1, 0), 1);
        args.setValue(DMT_SURFACE_TANGENT, &d->tangentMatrix.at(2, 0), 2);
        break;
    case DMU_BITANGENT_X:
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(0, 1), 0);
        break;
    case DMU_BITANGENT_Y:
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(1, 1), 0);
        break;
    case DMU_BITANGENT_Z:
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(2, 1), 0);
        break;
    case DMU_BITANGENT_XYZ:
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(0, 1), 0);
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(1, 1), 1);
        args.setValue(DMT_SURFACE_BITANGENT, &d->tangentMatrix.at(2, 1), 2);
        break;
    case DMU_NORMAL_X:
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(0, 2), 0);
        break;
    case DMU_NORMAL_Y:
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(1, 2), 0);
        break;
    case DMU_NORMAL_Z:
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(2, 2), 0);
        break;
    case DMU_NORMAL_XYZ:
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(0, 2), 0);
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(1, 2), 1);
        args.setValue(DMT_SURFACE_NORMAL, &d->tangentMatrix.at(2, 2), 2);
        break;
    case DMU_COLOR:
        args.setValue(DMT_SURFACE_RGBA, &d->color.x, 0);
        args.setValue(DMT_SURFACE_RGBA, &d->color.y, 1);
        args.setValue(DMT_SURFACE_RGBA, &d->color.z, 2);
        args.setValue(DMT_SURFACE_RGBA, &d->opacity, 2);
        break;
    case DMU_COLOR_RED:
        args.setValue(DMT_SURFACE_RGBA, &d->color.x, 0);
        break;
    case DMU_COLOR_GREEN:
        args.setValue(DMT_SURFACE_RGBA, &d->color.y, 0);
        break;
    case DMU_COLOR_BLUE:
        args.setValue(DMT_SURFACE_RGBA, &d->color.z, 0);
        break;
    case DMU_ALPHA:
        args.setValue(DMT_SURFACE_RGBA, &d->opacity, 0);
        break;
    case DMU_BLENDMODE:
        args.setValue(DMT_SURFACE_BLENDMODE, &d->blendMode, 0);
        break;
    case DMU_FLAGS:
        args.setValue(DMT_SURFACE_FLAGS, &d->flags, 0);
        break;
    default:
        return MapElement::property(args);
    }
    return false;  // Continue iteration.
}
dint Surface::setProperty(DmuArgs const &args) {
    switch (args.prop) {
    case DMU_BLENDMODE: {
        blendmode_t newBlendMode;
        args.value(DMT_SURFACE_BLENDMODE, &newBlendMode, 0);
        setBlendMode(newBlendMode);
        break; }
    case DMU_FLAGS:
        args.value(DMT_SURFACE_FLAGS, &d->flags, 0);
        break;
    case DMU_COLOR: {
        Vector3f newColor = d->color;
        args.value(DMT_SURFACE_RGBA, &newColor.x, 0);
        args.value(DMT_SURFACE_RGBA, &newColor.y, 1);
        args.value(DMT_SURFACE_RGBA, &newColor.z, 2);
        setColor(newColor);
        break; }
    case DMU_COLOR_RED: {
        Vector3f newColor = d->color;
        args.value(DMT_SURFACE_RGBA, &newColor.x, 0);
        setColor(newColor);
        break; }
    case DMU_COLOR_GREEN: {
        Vector3f newColor = d->color;
        args.value(DMT_SURFACE_RGBA, &newColor.y, 0);
        setColor(newColor);
        break; }
    case DMU_COLOR_BLUE: {
        Vector3f newColor = d->color;
        args.value(DMT_SURFACE_RGBA, &newColor.z, 0);
        setColor(newColor);
        break; }
    case DMU_ALPHA: {
        dfloat newOpacity;
        args.value(DMT_SURFACE_RGBA, &newOpacity, 0);
        setOpacity(newOpacity);
        break; }
    case DMU_MATERIAL: {
        Material *newMaterial;
        args.value(DMT_SURFACE_MATERIAL, &newMaterial, 0);
        setMaterial(newMaterial);
        break; }
    case DMU_OFFSET_X: {
        Vector2f newOrigin = d->origin;
        args.value(DMT_SURFACE_OFFSET, &newOrigin.x, 0);
        setOrigin(newOrigin);
        break; }
    case DMU_OFFSET_Y: {
        Vector2f newOrigin = d->origin;
        args.value(DMT_SURFACE_OFFSET, &newOrigin.y, 0);
        setOrigin(newOrigin);
        break; }
    case DMU_OFFSET_XY: {
        Vector2f newOrigin = d->origin;
        args.value(DMT_SURFACE_OFFSET, &newOrigin.x, 0);
        args.value(DMT_SURFACE_OFFSET, &newOrigin.y, 1);
        setOrigin(newOrigin);
        break; }
    default:
        return MapElement::setProperty(args);
    }
    return false;  // Continue iteration.
}
#ifdef __CLIENT__
MaterialAnimator *Surface::materialAnimator() const {
    if (!d->material) return nullptr;
    if (!d->matAnimator) {
        d->matAnimator = &d->material->as<ClientMaterial>()
                .getAnimator(Rend_MapSurfaceMaterialSpec());
    }
    return d->matAnimator;
}
void Surface::resetLookups() {
    d->matAnimator = nullptr;
}
Vector2f const &Surface::originSmoothed() const {
    return d->originSmoothed;
}
Vector2f const &Surface::originSmoothedAsDelta() const {
    return d->originSmoothedDelta;
}
void Surface::lerpSmoothedOrigin() {
    d->originSmoothedDelta = d->oldOrigin[0] * (1 - ::frameTimePos)
        + origin() * ::frameTimePos - origin();
    d->originSmoothed = origin() + d->originSmoothedDelta;
    d->notifyOriginSmoothedChanged();
}
void Surface::resetSmoothedOrigin() {
    d->originSmoothed = d->oldOrigin[0] = d->oldOrigin[1] = origin();
    d->originSmoothedDelta = Vector2f();
    d->notifyOriginSmoothedChanged();
}
void Surface::updateOriginTracking() {
    d->oldOrigin[0] = d->oldOrigin[1];
    d->oldOrigin[1] = origin();
    if (d->oldOrigin[0] != d->oldOrigin[1]) {
        dfloat moveDistance = de::abs(Vector2f(d->oldOrigin[1] - d->oldOrigin[0]).length());
        if (moveDistance >= MAX_SMOOTH_MATERIAL_MOVE) {
            d->oldOrigin[0] = d->oldOrigin[1];
        }
    }
}
dfloat Surface::glow(Vector3f &color) const {
    if (!hasMaterial() || material().isSkyMasked()) {
        color = Vector3f();
        return 0;
    }
    MaterialAnimator &matAnimator = *materialAnimator(); //material().as<ClientMaterial>().getAnimator(Rend_MapSurfaceMaterialSpec());
    matAnimator.prepare();
    TextureVariant *texture = matAnimator.texUnit(MaterialAnimator::TU_LAYER0).texture;
    if (!texture) return 0;
    auto const *avgColorAmplified = reinterpret_cast<averagecolor_analysis_t const *>(texture->base().analysisDataPointer(ClientTexture::AverageColorAmplifiedAnalysis));
    if (!avgColorAmplified) {
        return 0;
    }
    color = Vector3f(avgColorAmplified->color.rgb);
    return matAnimator.glowStrength() * glowFactor; // Global scale factor.
}
#endif // __CLIENT__
Surface::IDecorationState::~IDecorationState()
{}
bool Thinker_IsMobjFunc(thinkfunc_t func) {
    return (func && func == reinterpret_cast<thinkfunc_t>(gx.MobjThinker));
}
bool Thinker_IsMobj(thinker_t const *th) {
    return (th && Thinker_IsMobjFunc(th->function));
}
world::Map &Thinker_Map(thinker_t const & /*th*/)
{
    return App_World().map();
}
namespace world {
struct ThinkerList {
    bool isPublic; ///< All thinkers in this list are visible publically.
    Thinker sentinel;
    ThinkerList(thinkfunc_t func, bool isPublic) : isPublic(isPublic) {
        sentinel.function = func;
        sentinel.disable(); // Safety measure.
        sentinel.prev = sentinel.next = sentinel;
    }
    void reinit() {
        sentinel.prev = sentinel.next = sentinel;
    }
    thinkfunc_t function() const {
        return sentinel.function;
    }
    void link(thinker_t &th) {
        sentinel.prev->next = &th;
        th.next = sentinel;
        th.prev = sentinel.prev;
        sentinel.prev = &th;
    }
    dint count(dint *numInStasis) const {
        dint num = 0;
        thinker_t *th = sentinel.next;
        while (th != &sentinel.base() && th) {
#ifdef LIBDENG_FAKE_MEMORY_ZONE
            DENG2_ASSERT(th->next);
            DENG2_ASSERT(th->prev);
#endif
            num += 1;
            if (numInStasis && Thinker_InStasis(th)) {
                (*numInStasis) += 1;
            }
            th = th->next;
        }
        return num;
    }
    void releaseAll() {
        for (thinker_t *th = sentinel.next; th != &sentinel.base() && th; th = th->next) {
            Thinker::release(*th);
        }
    }
};
DENG2_PIMPL(Thinkers) {
    dint idtable[2048];     ///< 65536 bits telling which IDs are in use.
    dushort iddealer = 0;
    QList<ThinkerList *> lists;
    QHash<thid_t, mobj_t *> mobjIdLookup;  ///< public only
    QHash<thid_t, thinker_t *> thinkerIdLookup; ///< all thinkers with ID
    bool inited = false;
    Impl(Public *i) : Base(i) {
        clearMobjIds();
    }
    ~Impl() {
        releaseAllThinkers();
        qDeleteAll(lists);
    }
    void releaseAllThinkers() {
        thinkerIdLookup.clear();
        for (ThinkerList *list : lists) {
            list->releaseAll();
        }
    }
    void clearMobjIds() {
        de::zap(idtable);
        idtable[0] |= 1;  // ID zero is always "used" (it's not a valid ID).
        mobjIdLookup.clear();
        thinkerIdLookup.clear();
    }
    thid_t newMobjId() {
        while (self().isUsedMobjId(++iddealer)) {}
        self().setMobjId(iddealer);
        return iddealer;
    }
    ThinkerList *listForThinkFunc(thinkfunc_t func, bool makePublic = true,
                                  bool canCreate = false) {
        for (dint i = 0; i < lists.count(); ++i) {
            ThinkerList *list = lists[i];
            if (list->function() == func && list->isPublic == makePublic)
                return list;
        }
        if (!canCreate) return nullptr;
        lists.append(new ThinkerList(func, makePublic));
        return lists.last();
    }
};
Thinkers::Thinkers() : d(new Impl(this))
{}
bool Thinkers::isUsedMobjId(thid_t id) {
    return d->idtable[id >> 5] & (1 << (id & 31) /*(id % 32) */ );
}
void Thinkers::setMobjId(thid_t id, bool inUse) {
    dint c = id >> 5, bit = 1 << (id & 31); //(id % 32);
    if (inUse) d->idtable[c] |= bit;
    else       d->idtable[c] &= ~bit;
}
struct mobj_s *Thinkers::mobjById(dint id) {
    auto found = d->mobjIdLookup.constFind(id);
    if (found != d->mobjIdLookup.constEnd()) {
        return found.value();
    }
    return nullptr;
}
thinker_t *Thinkers::find(thid_t id) {
    auto found = d->thinkerIdLookup.constFind(id);
    if (found != d->thinkerIdLookup.constEnd()) {
        return found.value();
    }
    return nullptr;
}
void Thinkers::add(thinker_t &th, bool makePublic) {
    if (!th.function)
        throw Error("Thinkers::add", "Invalid thinker function");
    if (Thinker_IsMobj(&th)) {
#ifdef __CLIENT__
        if (!Cl_IsClientMobj(reinterpret_cast<mobj_t *>(&th)))
#endif
        {
            th.id = d->newMobjId();
        }
        if (makePublic && th.id) {
            d->mobjIdLookup.insert(th.id, reinterpret_cast<mobj_t *>(&th));
        }
    }
    else {
        th.id = 0;  // Zero is not a valid ID.
    }
    if (th.id) {
        d->thinkerIdLookup.insert(th.id, &th);
    }
    ThinkerList *list = d->listForThinkFunc(th.function, makePublic, true /*can create*/);
    list->link(th);
}
void Thinkers::remove(thinker_t &th) {
    if (th.id) {
        setMobjId(th.id, false);
        d->mobjIdLookup.remove(th.id);
        d->thinkerIdLookup.remove(th.id);
#ifdef __SERVER__
        auto *mob = reinterpret_cast<mobj_t *>(&th);
        if (!mob->state || !runtimeDefs.states.indexOf(mob->state)) {
            Sv_MobjRemoved(th.id);
        }
#endif
    }
    th.function = (thinkfunc_t) -1;
    Thinker::release(th);
}
void Thinkers::initLists(dbyte flags) {
    if (!d->inited) {
        d->lists.clear();
    }
    else {
        for (dint i = 0; i < d->lists.count(); ++i) {
            ThinkerList *list = d->lists[i];
            if (list->isPublic && !(flags & 0x1)) continue;
            if (!list->isPublic && !(flags & 0x2)) continue;
            list->reinit();
        }
    }
    d->clearMobjIds();
    d->inited = true;
}
bool Thinkers::isInited() const {
    return d->inited;
}
LoopResult Thinkers::forAll(dbyte flags, std::function<LoopResult (thinker_t *)> func) const {
    if (!d->inited) return LoopContinue;
    for (dint i = 0; i < d->lists.count(); ++i) {
        ThinkerList *list = d->lists[i];
        if ( list->isPublic && !(flags & 0x1)) continue;
        if (!list->isPublic && !(flags & 0x2)) continue;
        thinker_t *th = list->sentinel.next;
        while (th != &list->sentinel.base() && th) {
#ifdef LIBDENG_FAKE_MEMORY_ZONE
            DENG2_ASSERT(th->next);
            DENG2_ASSERT(th->prev);
#endif
            thinker_t *next = th->next;
            if (auto result = func(th))
                return result;
            th = next;
        }
    }
    return LoopContinue;
}
LoopResult Thinkers::forAll(thinkfunc_t thinkFunc, dbyte flags, std::function<LoopResult (thinker_t *)> func) const {
    if (!d->inited) return LoopContinue;
    if (!thinkFunc) {
        return forAll(flags, func);
    }
    if (flags & 0x1 /*public*/)
    {
        if (ThinkerList *list = d->listForThinkFunc(thinkFunc)) {
            thinker_t *th = list->sentinel.next;
            while (th != &list->sentinel.base() && th) {
#ifdef LIBDENG_FAKE_MEMORY_ZONE
                DENG2_ASSERT(th->next);
                DENG2_ASSERT(th->prev);
#endif
                thinker_t *next = th->next;
                if (auto result = func(th))
                    return result;
                th = next;
            }
        }
    }
    if (flags & 0x2 /*private*/)
    {
        if (ThinkerList *list = d->listForThinkFunc(thinkFunc, false /*private*/))
        {
            thinker_t *th = list->sentinel.next;
            while (th != &list->sentinel.base() && th) {
#ifdef LIBDENG_FAKE_MEMORY_ZONE
                DENG2_ASSERT(th->next);
                DENG2_ASSERT(th->prev);
#endif
                thinker_t *next = th->next;
                if (auto result = func(th))
                    return result;
                th = next;
            }
        }
    }
    return LoopContinue;
}
dint Thinkers::count(dint *numInStasis) const {
    dint total = 0;
    if (isInited()) {
        for (dint i = 0; i < d->lists.count(); ++i) {
            ThinkerList *list = d->lists[i];
            total += list->count(numInStasis);
        }
    }
    return total;
}
static void unlinkThinkerFromList(thinker_t *th) {
    th->next->prev = th->prev;
    th->prev->next = th->next;
}
}  // namespace world
using namespace world;
void Thinker_InitPrivateData(thinker_t *th, Id::Type knownId) {
    if (th->d == nullptr) {
        Id const privateId = knownId? Id(knownId) : Id(/* get a new ID */);
        if (Thinker_IsMobj(th)) {
#ifdef __CLIENT__
            th->d = new ClientMobjThinkerData(privateId);
#else
            th->d = new MobjThinkerData(privateId);
#endif
        }
        else {
            th->d = new ThinkerData(privateId);
        }
        auto &thinkerData = THINKER_DATA(*th, ThinkerData);
        thinkerData.setThinker(th);
        thinkerData.initBindings();
    }
    else {
        DENG2_ASSERT(knownId != 0);
        THINKER_DATA(*th, ThinkerData).setId(knownId);
    }
}
/**
 * Locates a mobj by it's unique identifier in the CURRENT map.
 */
#undef Mobj_ById
DENG_EXTERN_C struct mobj_s *Mobj_ById(dint id) {
    if (!App_World().hasMap()) return nullptr;
    return App_World().map().thinkers().mobjById(id);
}
#undef Thinker_Init
void Thinker_Init() {
    if (!App_World().hasMap()) return;
    App_World().map().thinkers().initLists(0x1);  // Init the public thinker lists.
}
#undef Thinker_Run
void Thinker_Run() {
    if (!App_World().hasMap()) return;
    App_World().map().thinkers().forAll(0x1 | 0x2, [] (thinker_t *th) {
        try {
            if (Thinker_InStasis(th)) return LoopContinue; // Skip.
            if (th->function == (thinkfunc_t) -1) {
                unlinkThinkerFromList(th);
                if (th->id) {
                    P_MobjRecycle((mobj_t *) th);
                }
                else {
                    Thinker::destroy(th);
                }
            }
            else if (th->function) {
                if (!th->d) Thinker_InitPrivateData(th);
                th->function(th);
                if (th->d) THINKER_DATA(*th, Thinker::IData).think();
            }
        }
        catch (const Error &er) {
            LOG_MAP_WARNING("Thinker %i: %s") << th->id << er.asText();
        }
        return LoopContinue;
    });
}
#undef Thinker_Add
void Thinker_Add(thinker_t *th) {
    if (!th) return;
    Thinker_Map(*th).thinkers().add(*th);
}
#undef Thinker_Remove
void Thinker_Remove(thinker_t *th) {
    if (!th) return;
    Thinker_Map(*th).thinkers().remove(*th);
}
#undef Thinker_Iterate
dint Thinker_Iterate(thinkfunc_t func, dint (*callback) (thinker_t *, void *), void *context) {
    if (!App_World().hasMap()) return false;  // Continue iteration.
    return App_World().map().thinkers().forAll(func, 0x1, [&callback, &context] (thinker_t *th) {
        return callback(th, context);
    });
}
DENG_DECLARE_API(Thinker) = {
    { DE_API_THINKER },
    Thinker_Init,
    Thinker_Run,
    Thinker_Add,
    Thinker_Remove,
    Thinker_Iterate
};
Vertex::Vertex(Mesh &mesh, Vector2d const &origin)
    : MapElement(DMU_VERTEX)
    , MeshElement(mesh) {
    _origin = origin;
}
Vector2d const &Vertex::origin() const {
    return _origin;
}
void Vertex::setOrigin(Vector2d const &newOrigin) {
    if(_origin != newOrigin) {
        _origin = newOrigin;
        DENG2_FOR_AUDIENCE(OriginChange, i) {
            i->vertexOriginChanged(*this);
        }
    }
}
int Vertex::property(world::DmuArgs &args) const {
    switch(args.prop) {
    case DMU_X:
        args.setValue(DMT_VERTEX_ORIGIN, &_origin.x, 0);
        break;
    case DMU_Y:
        args.setValue(DMT_VERTEX_ORIGIN, &_origin.y, 0);
        break;
    case DMU_XY:
        args.setValue(DMT_VERTEX_ORIGIN, &_origin.x, 0);
        args.setValue(DMT_VERTEX_ORIGIN, &_origin.y, 1);
        break;
    default:
        return MapElement::property(args);
    }
    return false;  // Continue iteration.
}
duint Vertex::lineOwnerCount() const {
    return _numLineOwners;
}
void Vertex::countLineOwners() {
    _onesOwnerCount = _twosOwnerCount = 0;
    if(LineOwner const *firstOwn = firstLineOwner()) {
        LineOwner const *own = firstOwn;
        do {
            if(!own->line().front().hasSector() || !own->line().back().hasSector()) {
                _onesOwnerCount += 1;
            }
            else {
                _twosOwnerCount += 1;
            }
        } while((own = own->next()) != firstOwn);
    }
}
LineOwner *Vertex::firstLineOwner() const {
    return _lineOwners;
}
#ifdef __CLIENT__
/**
 * Given two lines "connected" by shared origin coordinates (0, 0) at a "corner"
 * vertex, calculate the point which lies @a distA away from @a lineA and also
 * @a distB from @a lineB. The point should also be the nearest point to the
 * origin (in case of parallel lines).
 *
 * @param lineADirection  Direction vector for the "left" line.
 * @param dist1           Distance from @a lineA to offset the corner point.
 * @param lineBDirection  Direction vector for the "right" line.
 * @param dist2           Distance from @a lineB to offset the corner point.
 *
 * Return values:
 * @param point  Coordinates for the corner point are written here. Can be @c nullptr.
 * @param lp     Coordinates for the "extended" point are written here. Can be @c nullptr.
 */
static void cornerNormalPoint(Vector2d const &lineADirection, ddouble dist1,
                              Vector2d const &lineBDirection, ddouble dist2,
                              Vector2d *point, Vector2d *lp) {
    if(!point && !lp) return;
    ddouble len1 = lineADirection.length();
    ddouble len2 = lineBDirection.length();
    Vector2d norm1(-lineADirection.y / len1 * dist1,  lineADirection.x / len1 * dist1);
    Vector2d norm2( lineBDirection.y / len2 * dist2, -lineBDirection.x / len2 * dist2);
    if(lp) {
        *lp = lineBDirection / len2 * dist2;
    }
    if(point) {
        Partition lineA(lineADirection, norm1);
        Partition lineB(lineBDirection, norm2);
        if(lineA.isParallelTo(lineB)) {
            *point = norm1;
        }
        else {
            *point = lineA.intercept(lineB);
        }
    }
}
/**
 * Returns the width (world units) of a shadow edge (scaled depending on the length of @a edge).
 */
static ddouble shadowEdgeWidth(Vector2d const &edge) {
    ddouble const normalWidth = 20; //16;
    ddouble const maxWidth    = 60;
    ddouble length = edge.length();
    if(length > 600) {
        ddouble w = length - 600;
        if(w > 1000) w = 1000;
        return normalWidth + w / 1000 * maxWidth;
    }
    return normalWidth;
}
void Vertex::updateShadowOffsets() {
    LineOwner *base = firstLineOwner();
    if(!base) return;
    LineOwner *own = base;
    do {
        Line const &lineB = own->line();
        Line const &lineA = own->next()->line();
        Vector2d const rightDir = (&lineB.from() == this?  lineB.direction() : -lineB.direction());
        Vector2d const leftDir  = (&lineA.from() == this? -lineA.direction() :  lineA.direction()) * -1;  // Always flipped.
        cornerNormalPoint(leftDir,  shadowEdgeWidth(leftDir),
                          rightDir, shadowEdgeWidth(rightDir),
                          &own->_shadowOffsets.inner,
                          &own->_shadowOffsets.extended);
        own = own->next();
    } while(own != base);
}
#endif  // __CLIENT__
