﻿/** @file animgroup.cpp  Material animation group.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/animgroup.h"
#include <de/Log>
#include <QtAlgorithms>
/** @file animgroups.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/animgroups.h"
#include "doomsday/resource/resources.h"
#include <de/Log>
#include <QList>
/** @file bundlelinkfeed.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/bundlelinkfeed.h"
#include <de/LinkFile>
/** @file bundles.cpp  Data bundle indexing.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/bundles.h"
#include "doomsday/filesys/datafile.h"
#include "doomsday/filesys/datafolder.h"
#include <de/App>
#include <de/Config>
#include <de/DictionaryValue>
#include <de/PackageLoader>
#include <de/LinkFile>
#include <de/Loop>
#include <de/TaskPool>
#include <QList>
#include <QSet>
/** @file colorpalette.cpp  Color palette resource.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/colorpalette.h"
#include <de/Log>
#include <de/Range>
#include <de/reader.h>
#include <de/mathutil.h>
/** @file colorpalettes.cpp
 *
 * @authors Copyright © 2013-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/colorpalettes.h"
#include "doomsday/resource/resources.h"
#include <QMap>
/** @file compositetexture.cpp  Composite Texture.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/composite.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/lumpindex.h"
#include "doomsday/resource/patch.h"
#include "doomsday/resource/patchname.h"
#include <QList>
#include <QRect>
#include <de/ByteRefArray>
#include <de/String>
#include <de/Reader>
/** @file databundle.cpp  Classic data files: PK3, WAD, LMP, DED, DEH.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/databundle.h"
#include "doomsday/filesys/datafolder.h"
#include "doomsday/filesys/datafile.h"
#include "doomsday/resource/bundles.h"
#include "doomsday/resource/resources.h"
#include "doomsday/resource/lumpdirectory.h"
#include "doomsday/doomsdayapp.h"
#include "doomsday/Games"
#include <de/charsymbols.h>
#include <de/App>
#include <de/ArchiveFeed>
#include <de/Info>
#include <de/LinkFile>
#include <de/LogBuffer>
#include <de/MetadataBank>
#include <de/Package>
#include <de/PackageLoader>
#include <de/Path>
#include <de/TextValue>
#include <QRegExp>
#include <QRegularExpression>
#include <QTextCodec>
/** @file doomsdaypackage.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/res/DoomsdayPackage"
#include <de/NativeFile>
/** @file lumpcatalog.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/lumpcatalog.h"
#include "doomsday/resource/lumpdirectory.h"
#include "doomsday/resource/databundle.h"
#include <de/App>
#include <de/LinkFile>
#include <de/PackageLoader>
#include <utility>
/** @file lumpdirectory.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/lumpdirectory.h"
#include <de/MetadataBank>
#include <de/Reader>
#include <de/mathutil.h>
#include <QList>
#include <QRegularExpression>
/** @file
 *
 * @authors Copyright © 2009-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/m_decomp64.h"
/** @file manifest.cpp  Game resource manifest.
 *
 * @authors Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/manifest.h"
#include "doomsday/resource/resources.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/lumpindex.h"
#include "doomsday/filesys/wad.h"
#include "doomsday/filesys/zip.h"
#include <de/App>
#include <de/FileSystem>
#include <de/NativeFile>
#include <de/Path>
/** @file mapdef.cpp  Resource manifest for a map.
 *
 * @authors Copyright © 2014-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/mapmanifest.h"
#include <de/NativePath>
/** @file mapmanifests.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/mapmanifests.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/resource/resources.h" // MissingResourceManifestError
/** @file patch.cpp Patch Image Format.
 *
 * @authors Copyright &copy; 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/patch.h"
#include <QList>
#include <QRect>
#include <de/libcore.h>
#include <de/Log>
#include <de/math.h>
/** @file patchname.cpp PatchName
 *
 * @authors Copyright &copy; 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright &copy; 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/patchname.h"
#include "doomsday/filesys/fs_main.h"
#include <de/IReadable>
#include <de/Reader>
#include <de/String>
/** @file pcx.cpp  PCX image reader.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1997-2006 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/pcx.h"
#include "dd_share.h"
#include <de/memory.h>
/** @file resourceclass.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/resourceclass.h"
#include "doomsday/filesys/filetype.h"
/** @file resource/resources.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/resources.h"
#include "doomsday/resource/animgroups.h"
#include "doomsday/resource/mapmanifests.h"
#include "doomsday/resource/colorpalettes.h"
#include "doomsday/resource/sprites.h"
#include "doomsday/resource/textures.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include "doomsday/defs/music.h"
#include "doomsday/world/Materials"
#include "doomsday/DoomsdayApp"
#include "doomsday/SaveGames"
#include "doomsday/DataBundle"
#include "doomsday/res/DoomsdayPackage"
#include <de/App>
#include <de/CommandLine>
#include <de/Loop>
#include <de/PackageLoader>
#include <de/Config>
/** @file sprites.cpp  Sprites.
 * @ingroup resource
 *
 * @authors Copyright © 2013-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/sprites.h"
#include "doomsday/resource/resources.h"
#include "doomsday/resource/textures.h"
#include "doomsday/defs/ded.h"
#include "doomsday/defs/sprite.h"
#include <de/types.h>
#include <QMap>
/** @file texture.cpp  Logical texture resource.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/texture.h"
#include "doomsday/resource/resources.h"
#include "doomsday/resource/composite.h"
#include "doomsday/resource/texturemanifest.h"
#include "doomsday/resource/textures.h"
#include "doomsday/console/cmd.h"
#include <de/Error>
#include <de/Log>
#include <de/memory.h>
#include <QMap>
/** @file texturemanifest.cpp  Texture Manifest.
 *
 * @authors Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/texturemanifest.h"
#include "doomsday/resource/textures.h"
#include "doomsday/resource/resources.h"
/** @file textures.cpp
 *
 * @authors Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "doomsday/resource/textures.h"
#include "doomsday/resource/resources.h"
#include "doomsday/res/Composite"
#include "doomsday/res/TextureScheme"
#include "doomsday/res/Patch"
#include "doomsday/res/PatchName"
#include "doomsday/res/Sprites"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/DoomsdayApp"
#include <de/mathutil.h>
#include <de/types.h>
#include <de/stack.h>
/** @file texturescheme.cpp  Texture system subspace scheme.
 *
 * @authors Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/texturescheme.h"
#include "doomsday/resource/texturemanifest.h"
#include <de/types.h>
/** @file wav.cpp WAV loader. @ingroup audio
 *
 * @todo This is obsolete code! Use de::Waveform instead.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "doomsday/resource/wav.h"
#include "doomsday/filesys/fs_main.h"
#include "doomsday/filesys/fs_util.h"
#include <de/LogBuffer>
#include <de/NativePath>
#include <de/memory.h>
#include <de/memoryzone.h>
#include "dd_share.h"
#ifdef __GNUC__
/*
 * Something in here is confusing GCC: when compiled with -O2, WAV_MemoryLoad()
 * reads corrupt data. It is likely that the optimizer gets the manipulation of
 * the 'data' pointer wrong.
 */
#  if GCC_VERSION >= 40400
void* WAV_MemoryLoad(const byte* data, size_t datalength, int* bits, int* rate,
                     int* samples) __attribute__(( optimize(0) ));
#  endif
#endif
#define WAVE_FORMAT_PCM     1
#pragma pack(1)
using namespace de;
namespace res {
DENG2_PIMPL_NOREF(AnimGroup) {
    Frames frames;
    int uniqueId = 0;
    int flags = 0; ///< @ref animationGroupFlags
    ~Impl() {
        clearAllFrames();
    }
    void clearAllFrames() {
        qDeleteAll(frames);
        frames.clear();
    }
};
AnimGroup::AnimGroup(int uniqueId, int flags) : d(new Impl) {
    d->uniqueId = uniqueId;
    d->flags    = flags;
}
void AnimGroup::clearAllFrames() {
    d->clearAllFrames();
}
int AnimGroup::id() const {
    return d->uniqueId;
}
int AnimGroup::flags() const {
    return d->flags;
}
bool AnimGroup::hasFrameFor(TextureManifest const &textureManifest) const {
    foreach (Frame *frame, d->frames) {
        if (&frame->textureManifest() == &textureManifest)
            return true;
    }
    return false;
}
AnimGroup::Frame &AnimGroup::newFrame(TextureManifest &textureManifest,
                                      ushort tics, ushort randomTics) {
    d->frames.append(new Frame(textureManifest, tics, randomTics));
    return *d->frames.last();
}
AnimGroup::Frames const &AnimGroup::allFrames() const {
    return d->frames;
}
AnimGroup::Frame::Frame(TextureManifest &textureManifest, ushort tics, ushort randomTics)
    : _textureManifest(&textureManifest)
    , _tics(tics)
    , _randomTics(randomTics)
{}
TextureManifest &AnimGroup::Frame::textureManifest() const {
    return *_textureManifest;
}
ushort AnimGroup::Frame::tics() const {
    return _tics;
}
ushort AnimGroup::Frame::randomTics() const {
    return _randomTics;
}
} // namespace res
namespace res {
using namespace de;
DENG2_PIMPL_NOREF(AnimGroups) {
    QList<res::AnimGroup *> animGroups;
    ~Impl() {
        clearAllAnimGroups();
    }
    void clearAllAnimGroups() {
        qDeleteAll(animGroups);
        animGroups.clear();
    }
};
AnimGroups &AnimGroups::get() {
    return Resources::get().animGroups();
}
AnimGroups::AnimGroups() : d(new Impl)
{}
dint AnimGroups::animGroupCount() {
    return d->animGroups.count();
}
void AnimGroups::clearAllAnimGroups() {
    d->clearAllAnimGroups();
}
res::AnimGroup &AnimGroups::newAnimGroup(dint flags) {
    LOG_AS("AnimGroups");
    dint const uniqueId = d->animGroups.count() + 1; // 1-based.
    d->animGroups.append(new res::AnimGroup(uniqueId, flags));
    return *d->animGroups.last();
}
res::AnimGroup *AnimGroups::animGroup(dint uniqueId) {
    LOG_AS("AnimGroups::animGroup");
    if (uniqueId > 0 && uniqueId <= d->animGroups.count()) {
        return d->animGroups.at(uniqueId - 1);
    }
    LOGDEV_RES_WARNING("Invalid group #%i, returning NULL") << uniqueId;
    return nullptr;
}
res::AnimGroup *AnimGroups::animGroupForTexture(res::TextureManifest const &textureManifest) {
    for (dint i = animGroupCount(); i > 0; i--) {
        res::AnimGroup *group = animGroup(i);
        if (group->hasFrameFor(textureManifest)) {
            return group;
        }
    }
    return nullptr;  // Not found.
}
} // namespace res
namespace res {
BundleLinkFeed::BundleLinkFeed()
{}
String BundleLinkFeed::description() const {
    return "data bundle links";
}
Feed::PopulatedFiles BundleLinkFeed::populate(Folder const &) {
    return PopulatedFiles();
}
bool BundleLinkFeed::prune(File &file) const {
    if (LinkFile const *link = maybeAs<LinkFile>(file)) {
        return link->isBroken();
    }
    return false;
}
} // namespace res
namespace res {
static int const MATCH_MAXIMUM_SCORE = 4; // in case 5 specified, allow 1 to not match for flexibility
static String const VAR_REQUIRED_SCORE = "requiredScore";
DENG2_PIMPL(Bundles)
, DENG2_OBSERVES(FileIndex, Addition)
, DENG2_OBSERVES(FileIndex, Removal)
, public Lockable {
    de::Info identityRegistry;
    QSet<DataBundle const *> bundlesToIdentify; // lock for access
    LoopCallback mainCall;
    QHash<DataBundle::Format, BlockElements> formatEntries;
    TaskPool tasks;
    Impl(Public *i) : Base(i) {
        App::fileSystem().indexFor(DENG2_TYPE_NAME(DataFile))  .audienceForAddition() += this;
        App::fileSystem().indexFor(DENG2_TYPE_NAME(DataFile))  .audienceForRemoval()  += this;
        App::fileSystem().indexFor(DENG2_TYPE_NAME(DataFolder)).audienceForAddition() += this;
        App::fileSystem().indexFor(DENG2_TYPE_NAME(DataFolder)).audienceForRemoval()  += this;
    }
    ~Impl() { {
            DENG2_GUARD(this);
            bundlesToIdentify.clear();
        }
        tasks.waitForDone();
    }
    void fileAdded(File const &dataFile, FileIndex const &) {
        DENG2_ASSERT(is<DataBundle>(dataFile)); {
            DENG2_GUARD(this);
            bundlesToIdentify.insert(maybeAs<DataBundle>(dataFile));
        }
        if (!mainCall) {
            mainCall.enqueue([this] () { self().identify(); });
        }
    }
    void fileRemoved(File const &dataFile, FileIndex const &) {
        DENG2_ASSERT(is<DataBundle>(dataFile));
        DENG2_GUARD(this);
        bundlesToIdentify.remove(maybeAs<DataBundle>(dataFile));
    }
    DataBundle const *nextToIdentify() {
        DENG2_GUARD(this);
        if (bundlesToIdentify.isEmpty()) {
            return nullptr;
        }
        auto const *bundle = *bundlesToIdentify.begin();
        bundlesToIdentify.remove(bundle);
        return bundle;
    }
    bool identifyAddedDataBundles() {
        Folder::waitForPopulation();
        DENG2_ASSERT(App::rootFolder().has("/sys/bundles"));
        bool wasIdentified = false;
        int  count         = 0;
        Time startedAt;
        while (auto const *bundle = nextToIdentify()) {
            ++count;
            if (bundle->identifyPackages()) {
                wasIdentified = true;
            }
        }
        if (count) {
            LOG_RES_MSG("Identified %i data bundles in %.1f seconds") << count << startedAt.since();
        }
        return wasIdentified;
    }
    void parseRegistry() {
        DENG2_GUARD(this);
        if (!identityRegistry.isEmpty()) return;
        String const defPath = "/packs/net.dengine.base/databundles.dei";
        formatEntries.clear();
        identityRegistry.parse(App::rootFolder().locate<File const>(defPath));
        for (auto *elem : identityRegistry.root().contentsInOrder()) {
            using Info = de::Info;
            if (!elem->isBlock()) {
                continue;
            }
            Info::BlockElement &block = elem->as<Info::BlockElement>();
            if (block.blockType() != QStringLiteral("package")) {
                continue;
            }
            String format = block.keyValue(QStringLiteral("format")).text.toLower();
            DataBundle::Format bundleFormat =
                    (format == "iwad"? DataBundle::Iwad :
                     format == "pwad"? DataBundle::Pwad :
                     format == "pk3"?  DataBundle::Pk3 :
                     format == "lmp"?  DataBundle::Lump :
                     format == "deh"?  DataBundle::Dehacked :
                     format == "ded"?  DataBundle::Ded :
                                       DataBundle::Unknown);
            if (bundleFormat == DataBundle::Unknown) {
                throw InvalidError("Bundles::parseRegistry",
                                   defPath + ": invalid format for \"" + block.name() + "\"");
            }
            if (!block.contains(VAR_REQUIRED_SCORE)) {
                int const ruleCount = block.size() - 1; // not counting "info"
                block.add(new Info::KeyElement(VAR_REQUIRED_SCORE,
                        String::format("%i", de::min(MATCH_MAXIMUM_SCORE, ruleCount))));
            }
            formatEntries[bundleFormat].append(&block);
        }
    }
    DENG2_PIMPL_AUDIENCE(Identify)
};
DENG2_AUDIENCE_METHOD(Bundles, Identify)
Bundles::Bundles()
    : d(new Impl(this))
{}
de::Info const &Bundles::identityRegistry() const {
    d->parseRegistry();
    return d->identityRegistry;
}
Bundles::BlockElements Bundles::formatEntries(DataBundle::Format format) const {
    d->parseRegistry();
    return d->formatEntries[format];
}
void Bundles::identify() {
    FS::get().changeBusyLevel(+1);
    d->tasks.start([this] () {
        d->identifyAddedDataBundles();
        if (isEverythingIdentified()) {
            DENG2_FOR_AUDIENCE2(Identify, i) {
                i->dataBundlesIdentified();
            }
        }
        FS::get().changeBusyLevel(-1);
    });
}
bool Bundles::isEverythingIdentified() const {
    DENG2_GUARD(d);
    return d->bundlesToIdentify.isEmpty();
}
/*void Bundles::waitForEverythingIdentified()
{
    identify();
    d->tasks.waitForDone();
}*/
Bundles::MatchResult Bundles::match(DataBundle const &bundle) const {
    using Info = de::Info;
    LOG_AS("res::Bundles");
    MatchResult match;
    File const &source = bundle.asFile();
    for (auto const *def : formatEntries(bundle.format())) {
        int score = 0;
        if (auto const *fileName = def->find(QStringLiteral("fileName"))) {
            if (fileName->isKey() &&
                fileName->as<Info::KeyElement>().value()
                    .text.compareWithoutCase(source.name()) == 0) {
                ++score;
            }
            else if (fileName->isList()) {
                for (auto const &cand : fileName->as<Info::ListElement>().values()) {
                    if (!cand.text.compareWithoutCase(source.name())) {
                        ++score;
                        break;
                    }
                }
            }
        }
        String fileType = def->keyValue(QStringLiteral("fileType"));
        if (fileType.isEmpty()) fileType = "file"; // prefer files by default
        if ((!fileType.compareWithoutCase(QStringLiteral("file"))   && source.status().type() == File::Type::File) ||
            (!fileType.compareWithoutCase(QStringLiteral("folder")) && source.status().type() == File::Type::Folder)) {
            ++score;
        }
        String fileSize = def->keyValue(QStringLiteral("fileSize"));
        if (!fileSize.isEmpty() && fileSize.toUInt() == source.size()) {
            ++score;
        }
        bool crcMismatch = false;
        if (bundle.format() == DataBundle::Iwad ||
            bundle.format() == DataBundle::Pwad) {
            String lumpDirCRC32 = def->keyValue(QStringLiteral("lumpDirCRC32"));
            if (!lumpDirCRC32.isEmpty()) {
                if (lumpDirCRC32.toUInt(nullptr, 16) == bundle.lumpDirectory()->crc32()) {
                    score += 2;
                }
                else {
                    crcMismatch = true;
                }
            }
            if (auto const *lumps = maybeAs<Info::ListElement>(def->find(QStringLiteral("lumps")))) {
                ++score; // will be subtracted if not matched
                for (auto const &val : lumps->values()) {
                    QRegExp const sizeCondition("(.*)==([0-9]+)");
                    Block lumpName;
                    int requiredSize = 0;
                    if (sizeCondition.exactMatch(val)) {
                        lumpName     = sizeCondition.cap(1).toUtf8();
                        requiredSize = sizeCondition.cap(2).toInt();
                    }
                    else {
                        lumpName     = val.text.toUtf8();
                        requiredSize = -1;
                    }
                    if (!bundle.lumpDirectory()->has(lumpName)) {
                        --score;
                        break;
                    }
                    if (requiredSize >= 0 &&
                        bundle.lumpDirectory()->lumpSize(lumpName) != duint32(requiredSize)) {
                        --score;
                        break;
                    }
                }
            }
        }
        if (score < def->keyValue(VAR_REQUIRED_SCORE).text.toInt()) {
            score = 0;
        }
        if (score > 0 && score >= match.bestScore) {
            match.bestMatch = def;
            match.bestScore = score;
            auto const idVer = Package::split(def->name());
            match.packageId = idVer.first;
            match.packageVersion = (!crcMismatch? idVer.second : Version(""));
        }
    }
    if (!match.bestScore) {
        return MatchResult();
    }
    LOG_RES_VERBOSE("Matched: %s %s %s score: %i")
            << match.packageId
            << match.packageVersion.fullNumber()
            << bundle.description()
            << match.bestScore;
    return match;
}
} // namespace res
#define RGB18(r, g, b)      ((r)+((g)<<6)+((b)<<12))
namespace res {
/*
 * Example: "R8G8B8"
 */
static void parseColorFormat(QString const &fmt, Vector3ui &compOrder, Vector3ui &compBits) {
    compBits = Vector3ui();
    int const end = fmt.length();
    int readComponents = 0;
    int pos = 0;
    while (pos < end) {
        QChar ch = fmt[pos]; pos++;
        int comp = -1;
        if     (ch == 'R' || ch == 'r') comp = 0;
        else if (ch == 'G' || ch == 'g') comp = 1;
        else if (ch == 'B' || ch == 'b') comp = 2;
        if (comp != -1 && compBits[comp] == 0) {
            compOrder[comp] = readComponents++;
            int start = pos;
            ch = fmt[pos];
            while (ch.isDigit() && ++pos < end) {
                ch = fmt[pos];
            }
            int numDigits = pos - start;
            if (numDigits) {
                compBits[comp] = fmt.mid(start, numDigits).toInt();
                if (readComponents == 3)
                    break;
                continue;
            }
        }
        throw ColorTableReader::FormatError("parseColorFormat", QString("Unexpected character '%1' at position %2").arg(ch).arg(pos));
    }
    if (readComponents != 3) {
        throw ColorTableReader::FormatError("parseColorFormat", "Incomplete format specification");
    }
}
typedef QVector<de::Vector3ub> ColorTable;
ColorTable ColorTableReader::read(String format, int colorCount,
    dbyte const *colorData) // static
{
    Vector3ui order;
    Vector3ui bits;
    parseColorFormat(format, order, bits);
    ColorTable colors(colorCount);
    if (8 == bits.x && 8 == bits.y && 8 == bits.z) {
        dbyte const *src = colorData;
        for (int i = 0; i < colorCount; ++i, src += 3) {
            colors[i] = Vector3ub(src[order.x], src[order.y], src[order.z]);
        }
    }
    else {
        dbyte const *src = colorData;
        dbyte cb = 0;
        for (int i = 0; i < colorCount; ++i) {
            Vector3ub &dst = colors[i];
            Vector3i tmp;
            M_ReadBits(bits[order.x], &src, &cb, (dbyte *) &(tmp[order.x]));
            M_ReadBits(bits[order.y], &src, &cb, (dbyte *) &(tmp[order.y]));
            M_ReadBits(bits[order.z], &src, &cb, (dbyte *) &(tmp[order.z]));
            if (8 != bits.x) {
                if (bits.x < 8)
                    tmp.x <<= 8 - bits.x;
                else
                    tmp.x >>= bits.x - 8;
            }
            if (8 != bits.y) {
                if (bits.y < 8)
                    tmp.y <<= 8 - bits.y;
                else
                    tmp.y >>= bits.y - 8;
            }
            if (8 != bits.z) {
                if (bits.z < 8)
                    tmp.z <<= 8 - bits.z;
                else
                    tmp.z >>= bits.z - 8;
            }
            dst = Vector3ub(de::clamp<dbyte>(0, tmp.x, 255),
                            de::clamp<dbyte>(0, tmp.y, 255),
                            de::clamp<dbyte>(0, tmp.z, 255));
        }
    }
    return colors;
}
DENG2_PIMPL(ColorPalette) {
    typedef Vector3ub Color;
    typedef QVector<Color> ColorTable;
    ColorTable colors;
    typedef QMap<String, Translation> Translations;
    Translations translations;
    typedef QVector<int> XLat18To8;
    QScopedPointer<XLat18To8> xlat18To8;
    bool need18To8Update = false;  // Table built only when needed.
    Id id;
    Impl(Public *i) : Base(i) {
        LOG_RES_VERBOSE("New color palette %s") << id;
    }
    Translation *translation(String id) {
        Translations::iterator found = translations.find(id);
        if (found != translations.end()) {
            return &found.value();
        }
        return 0;
    }
    void notifyColorTableChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE(ColorTableChange, i) {
            i->colorPaletteColorTableChanged(self());
        }
    }
    void prepareNearestLUT() {
#define COLORS18BIT 262144
        need18To8Update = false;
        if (xlat18To8.isNull()) {
            xlat18To8.reset(new XLat18To8(COLORS18BIT));
        }
        for (int r = 0; r < 64; ++r)
        for (int g = 0; g < 64; ++g)
        for (int b = 0; b < 64; ++b) {
            int nearest = 0;
            int smallestDiff = DDMAXINT;
            for (int i = 0; i < colors.count(); ++i) {
                Color const &color = colors[i];
                int diff = (color.x - (r << 2)) * (color.x - (r << 2)) +
                           (color.y - (g << 2)) * (color.y - (g << 2)) +
                           (color.z - (b << 2)) * (color.z - (b << 2));
                if (diff < smallestDiff) {
                    smallestDiff = diff;
                    nearest = i;
                }
            }
            (*xlat18To8)[RGB18(r, g, b)] = nearest;
        }
#undef COLORS18BIT
    }
};
ColorPalette::ColorPalette() : d(new Impl(this))
{}
ColorPalette::ColorPalette(ColorTable const &colors)
    : d(new Impl(this)) {
    replaceColorTable(colors);
}
Id ColorPalette::id() const {
    return d->id;
}
int ColorPalette::colorCount() const {
    return d->colors.count();
}
ColorPalette &ColorPalette::replaceColorTable(ColorTable const &colorTable) {
    LOG_AS("ColorPalette");
    int const colorCountBefore = colorCount();
    d->need18To8Update = true;
    d->colors = colorTable;
    d->notifyColorTableChanged();
    if (colorCountBefore != colorCount()) {
        clearTranslations();
    }
    return *this;
}
Vector3ub ColorPalette::color(int colorIndex) const {
    LOG_AS("ColorPalette");
    if (colorIndex < 0 || colorIndex >= d->colors.count()) {
        LOG_DEBUG("Index %i out of range %s in palette %s, will clamp.")
            << colorIndex << Rangeui(0, d->colors.count()).asText() << d->id;
    }
    if (!d->colors.isEmpty()) {
        return d->colors[de::clamp(0, colorIndex, d->colors.count() - 1)];
    }
    return Vector3ub();
}
Vector3f ColorPalette::colorf(int colorIdx) const {
    return color(colorIdx).toVector3f() * reciprocal255;
}
int ColorPalette::nearestIndex(Vector3ub const &rgb) const {
    LOG_AS("ColorPalette");
    if (d->colors.isEmpty()) return -1;
    if (d->need18To8Update || d->xlat18To8.isNull()) {
        d->prepareNearestLUT();
    }
    return (*d->xlat18To8)[RGB18(rgb.x >> 2, rgb.y >> 2, rgb.z >> 2)];
}
void ColorPalette::clearTranslations() {
    LOG_AS("ColorPalette");
    d->translations.clear();
}
ColorPalette::Translation const *ColorPalette::translation(String id) const {
    LOG_AS("ColorPalette");
    return d->translation(id);
}
void ColorPalette::newTranslation(String xlatId, Translation const &mappings) {
    LOG_AS("ColorPalette");
    if (!colorCount()) {
        return;
    }
    DENG2_ASSERT(mappings.count() == colorCount()); // sanity check
    if (!xlatId.isEmpty()) {
        Translation *xlat = d->translation(xlatId);
        if (!xlat) {
            xlat = &d->translations.insert(xlatId, Translation()).value();
        }
        *xlat = mappings;
        for (int i = 0; i < colorCount(); ++i) {
            int palIdx = (*xlat)[i];
            if (palIdx < 0 || palIdx >= colorCount()) {
                (*xlat)[i] = i;
            }
        }
        return;
    }
    throw InvalidTranslationIdError("ColorPalette::newTranslation", "A zero-length id was specified");
}
} // namespace res
namespace res {
DENG2_PIMPL_NOREF(ColorPalettes) {
    typedef QMap<Id::Type, ColorPalette *> ColorPalettes;
    ColorPalettes colorPalettes; // owned
    typedef QMap<String, ColorPalette *> ColorPaletteNames;
    ColorPaletteNames colorPaletteNames;
    Id defaultColorPalette { Id::None };
    ~Impl() {
        clear();
    }
    void clear() {
        qDeleteAll(colorPalettes);
        colorPalettes.clear();
        colorPaletteNames.clear();
        defaultColorPalette = 0;
    }
    DENG2_PIMPL_AUDIENCE(Addition)
};
DENG2_AUDIENCE_METHOD(ColorPalettes, Addition)
ColorPalettes::ColorPalettes()
    : d(new Impl)
{}
void ColorPalettes::clearAllColorPalettes() {
    d->clear();
}
dint ColorPalettes::colorPaletteCount() const {
    return d->colorPalettes.count();
}
ColorPalette &ColorPalettes::colorPalette(Id const &id) const {
    auto found = d->colorPalettes.find(id.isNone()? d->defaultColorPalette : id);
    if (found != d->colorPalettes.end()) return *found.value();
    throw Resources::MissingResourceError("ColorPalettes::colorPalette",
                                          "Invalid ID " + id.asText());
}
String ColorPalettes::colorPaletteName(ColorPalette &palette) const {
    QList<String> const names = d->colorPaletteNames.keys(&palette);
    if (!names.isEmpty()) {
        return names.first();
    }
    return String();
}
bool ColorPalettes::hasColorPalette(String name) const {
    return d->colorPaletteNames.contains(name);
}
ColorPalette &ColorPalettes::colorPalette(String name) const {
    auto found = d->colorPaletteNames.find(name);
    if (found != d->colorPaletteNames.end()) return *found.value();
    throw Resources::MissingResourceError("ColorPalettes::colorPalette", "Unknown name '" + name + "'");
}
void ColorPalettes::addColorPalette(ColorPalette &newPalette, String const &name) {
    if (d->colorPalettes.contains(newPalette.id()))
        return;
    d->colorPalettes.insert(newPalette.id(), &newPalette);
    if (!name.isEmpty()) {
        d->colorPaletteNames.insert(name, &newPalette);
    }
    if (d->colorPalettes.count() == 1) {
        d->defaultColorPalette = newPalette.id();
    }
    DENG2_FOR_AUDIENCE2(Addition, i) i->colorPaletteAdded(newPalette);
}
Id ColorPalettes::defaultColorPalette() const {
    return d->defaultColorPalette;
}
void ColorPalettes::setDefaultColorPalette(ColorPalette *newDefaultPalette) {
    d->defaultColorPalette = newDefaultPalette ? newDefaultPalette->id().asUInt32() : 0;
}
} // namespace res
namespace res {
namespace internal {
static String readAndPercentEncodeRawName(de::Reader &from) {
    char asciiName[9];
    for (int i = 0; i < 8; ++i) { from >> asciiName[i]; }
    asciiName[8] = 0;
    return QString(QByteArray(asciiName).toPercentEncoding());
}
} // namespace internal
Composite::Component::Component(Vector2i const &origin)
    : _origin (origin)
    , _lumpNum(-1)
{}
void Composite::Component::setOrigin(const Vector2i &origin) {
    _origin = origin;
}
bool Composite::Component::operator == (Component const &other) const {
    if (lumpNum() != other.lumpNum()) return false;
    if (origin()  != other.origin())  return false;
    return true;
}
bool Composite::Component::operator != (Component const &other) const {
    return !(*this == other);
}
Vector2i const &Composite::Component::origin() const {
    return _origin;
}
lumpnum_t Composite::Component::lumpNum() const {
    return _lumpNum;
}
void Composite::Component::setLumpNum(lumpnum_t num) {
    _lumpNum = num;
}
DENG2_PIMPL_NOREF(Composite) {
    String name;                 ///< Symbolic, percent encoded.
    Flags flags;                 ///< Usage traits.
    Vector2ui logicalDimensions; ///< In map space units.
    Vector2ui dimensions;        ///< In pixels.
    int origIndex;               ///< Determined by the original game logic.
    Components components;       ///< Images to be composited.
    Impl() : origIndex(-1) {}
};
Composite::Composite(String const &percentEncodedName,
    Vector2ui const &logicalDimensions, Flags flags)
    : d(new Impl) {
    d->name              = percentEncodedName;
    d->flags             = flags;
    d->logicalDimensions = logicalDimensions;
}
bool Composite::operator == (Composite const &other) const {
    if (dimensions()        != other.dimensions())        return false;
    if (logicalDimensions() != other.logicalDimensions()) return false;
    if (componentCount()    != other.componentCount())    return false;
    for (int i = 0; i < componentCount(); ++i) {
        if (components()[i] != other.components()[i]) return false;
    }
    return true;
}
String Composite::percentEncodedName() const {
    return d->name;
}
String const &Composite::percentEncodedNameRef() const {
    return d->name;
}
Vector2ui const &Composite::logicalDimensions() const {
    return d->logicalDimensions;
}
Vector2ui const &Composite::dimensions() const {
    return d->dimensions;
}
Composite::Components const &Composite::components() const {
    return d->components;
}
Composite::Flags Composite::flags() const {
    return d->flags;
}
void Composite::setFlags(Composite::Flags flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
}
int Composite::origIndex() const {
    return d->origIndex;
}
void Composite::setOrigIndex(int newIndex) {
    d->origIndex = newIndex;
}
Composite *Composite::constructFrom(de::Reader &reader,
                                    QVector<PatchName> const &patchNames,
                                    ArchiveFormat format) {
    Composite *pctex = new Composite;
    pctex->d->name = res::internal::readAndPercentEncodeRawName(reader);
    dint16 unused16;
    reader >> unused16;
    byte scale[2]; /** @todo ZDoom defines these otherwise unused bytes as a
                       scale factor (div 8). We could interpret this also. */
    dint16 dimensions[2];
    reader >> scale[0] >> scale[1] >> dimensions[0] >> dimensions[1];
    pctex->d->logicalDimensions =
        pctex->d->dimensions = Vector2ui(dimensions[0], dimensions[1]);
    if (format == DoomFormat) {
        dint32 unused32;
        reader >> unused32;
    }
    /*
     * Finally, read the component images.
     * In the process we'll determine the final logical dimensions of the
     * texture by compositing the geometry of the component images.
     */
    dint16 componentCount;
    reader >> componentCount;
    QRect geom(QPoint(0, 0), QSize(pctex->d->logicalDimensions.x,
                                   pctex->d->logicalDimensions.y));
    int foundComponentCount = 0;
    for (dint16 i = 0; i < componentCount; ++i) {
        Component comp;
        dint16 origin16[2];
        reader >> origin16[0] >> origin16[1];
        comp.setOrigin(Vector2i(origin16[0], origin16[1]));
        dint16 pnamesIndex;
        reader >> pnamesIndex;
        if (pnamesIndex < 0 || pnamesIndex >= patchNames.count()) {
            LOG_RES_WARNING("Invalid PNAMES index %i in composite texture \"%s\", ignoring.")
                    << pnamesIndex << pctex->d->name;
        }
        else {
            comp.setLumpNum(patchNames[pnamesIndex].lumpNum());
            if (comp.lumpNum() >= 0) {
                foundComponentCount += 1;
                File1 &file = App_FileSystem().lump(comp.lumpNum());
                if (file.container().hasCustom()) {
                    pctex->d->flags |= Custom;
                }
                ByteRefArray fileData = ByteRefArray(file.cache(), file.size());
                if (res::Patch::recognize(fileData)) {
                    try {
                        auto info = res::Patch::loadMetadata(fileData);
                        geom |= QRect(QPoint(comp.origin().x, comp.origin().y),
                                      QSize(info.dimensions.x, info.dimensions.y));
                    }
                    catch (IByteArray::OffsetError const &) {
                        LOG_RES_WARNING("Component image \"%s\" (#%i) does not appear to be a valid Patch. "
                                        "It may be missing from composite texture \"%s\".")
                                << patchNames[pnamesIndex].percentEncodedNameRef() << i
                                << pctex->d->name;
                    }
                }
                file.unlock();
            }
            else {
                LOG_RES_WARNING("Missing component image \"%s\" (#%i) in composite texture \"%s\", ignoring.")
                        << patchNames[pnamesIndex].percentEncodedNameRef() << i
                        << pctex->d->name;
            }
        }
        reader >> unused16 >> unused16;
        pctex->d->components.push_back(comp);
    }
    if (geom.top()  < 0) geom.setTop(0);
    if (geom.height() > int(pctex->d->logicalDimensions.y)) {
        /*pctex->d->logicalDimensions.y = */pctex->d->dimensions.y = geom.height();
    }
    if (!foundComponentCount) {
        LOG_RES_WARNING("Zero valid component images in composite texture %s (will be ignored).")
            << pctex->d->name;
    }
    return pctex;
}
} // namespace res
static String const VAR_PATH        ("path");
static String const VAR_VERSION     ("version");
static String const VAR_LICENSE     ("license");
static String const VAR_AUTHOR      ("author");
static String const VAR_TITLE       ("title");
static String const VAR_TAGS        ("tags");
static String const VAR_NOTES       ("notes");
static String const VAR_DATA_FILES  ("dataFiles");
static String const VAR_BUNDLE_SCORE("bundleScore");
static String const VAR_REQUIRES    ("requires");
static String const VAR_RECOMMENDS  ("recommends");
static String const VAR_EXTRAS      ("extras");
static String const VAR_CATEGORY    ("category");
static String const CACHE_CATEGORY  ("DataBundle");
namespace internal {
    static char const *formatDescriptions[] = {
        "unknown",
        "PK3 archive",
        "WAD file",
        "IWAD file",
        "PWAD file",
        "data lump",
        "Doomsday Engine definitions",
        "DeHackEd patch", // importdeh plugin
        "collection"
    };
}
DENG2_PIMPL(DataBundle), public Lockable {
    bool ignored = false;
    SafePtr<File> source;
    Format format;
    String packageId; // linked under /sys/bundles/
    String versionedPackageId;
    std::unique_ptr<res::LumpDirectory> lumpDir;
    SafePtr<LinkFile> pkgLink;
    Impl(Public *i, Format fmt) : Base(i), format(fmt)
    {}
    ~Impl() {
        DENG2_GUARD(this);
        delete pkgLink.get();
    }
    static Folder &bundleFolder() {
        return App::rootFolder().locate<Folder>(QStringLiteral("/sys/bundles"));
    }
    /**
     * Identifies the data bundle and sets up a package link under "/sys/bundles" with
     * the appropriate metadata.
     *
     * Sets up the package metadata according to the best matched known information or
     * autogenerated entries.
     *
     * @return @c true, if the bundle was identified; otherwise @c false.
     */
    bool identify() {
        DENG2_GUARD(this);
        if (ignored || !packageId.isEmpty()) return false;
        if (format == Wad || format == Pwad || format == Iwad) {
            lumpDir.reset(new res::LumpDirectory(source->as<ByteArrayFile>()));
            if (!lumpDir->isValid()) {
                throw FormatError("DataBundle::identify",
                                  dynamic_cast<File *>(thisPublic)->description() +
                                  ": file contents may be corrupted " DENG2_CHAR_MDASH
                                  " WAD lump directory was not found");
            }
            format = (lumpDir->type() == res::LumpDirectory::Pwad? Pwad : Iwad);
            /*
            qDebug() << self().description()
                     << "format:" << (lumpDir->type()==res::LumpDirectory::Pwad? "PWAD" : "IWAD")
                     << "\nfileName:" << source->name()
                     << "\nfileSize:" << source->size()
                     << "\nlumpDirCRC32:" << QString::number(lumpDir->crc32(), 16).toLatin1();
            */
        }
        else if (!self().containerPackageId().isEmpty()) {
            /*qDebug() << "[DataBundle]" << source->description().toLatin1().constData()
                     << "is nested, no package will be generated";*/
            ignored = true;
            return false;
        }
        if (isAutoLoaded()) {
            ignored = true;
            return false;
        }
        DataBundle *container = self().containerBundle();
        if (container) {
            container->identifyPackages();
            if (format == Ded && container->format() == Pk3) {
                ignored = true;
                return false;
            }
#if 0
            if (container->isLinkedAsPackage() &&
                container->format() != Collection &&
                isAutoLoaded()) {
                qDebug() << container->d->versionedPackageId
                         << "loads data file"
                         << dataFilePath;
                container->packageMetadata()[VAR_DATA_FILES]
                        .value<ArrayValue>().add(new TextValue(dataFilePath));
            }
#endif
            if (container->d->ignored) {
                ignored = true; // No package for this.
                return false;
            }
        }
        Record const meta = cachedMetadata();
        packageId = meta.gets(Package::VAR_ID);
        versionedPackageId = packageId;
        if (auto chosen = chooseUniqueLinkPathAndVersion(self().asFile(), packageId,
                                                         meta.gets(VAR_VERSION),
                                                         meta.geti(VAR_BUNDLE_SCORE))) {
            LOGDEV_RES_VERBOSE("Linking %s as %s") << self().asFile().path() << chosen.path;
            pkgLink.reset(&bundleFolder().add(LinkFile::newLinkToFile(self().asFile(), chosen.path)));
            Record &metadata = Package::initializeMetadata(*pkgLink, packageId);
            metadata.copyMembersFrom(meta);
            metadata.set(VAR_VERSION, !chosen.version.isEmpty()? chosen.version : "0.0");
            if (!chosen.version.isEmpty()) {
                versionedPackageId += "_" + chosen.version;
            }
            LOG_RES_VERBOSE("Generated package:\n%s") << metadata.asText();
            App::fileSystem().index(*pkgLink);
            if (container &&
                container->isLinkedAsPackage() &&
                container->format() == Collection) {
                String subset = VAR_RECOMMENDS;
                String parentFolder = self().asFile().path().fileNamePath().fileName();
                if (!parentFolder.compareWithoutCase(QStringLiteral("Extra"))) {
                    subset = VAR_EXTRAS;
                }
                else if (!parentFolder.compareWithoutCase(QStringLiteral("Required"))) {
                    subset = VAR_REQUIRES;
                }
                container->packageMetadata().insertToSortedArray(subset, new TextValue(versionedPackageId));
                /*
                LOGDEV_RES_VERBOSE("%s ") << container->d->versionedPackageId
                         << "(" << container->d->pkgLink->objectNamespace().gets("package.tags", "") << ") "
                         << subset << " "
                         << versionedPackageId
                         << " (" << metadata.gets("tags", "") << ") from "
                         << self().asFile().path();
                */
            }
            return true;
        }
        else {
            ignored = true;
            return false;
        }
    }
    /**
     * Fetches cached data bundle metadata from the metadata bank, or rebuilds the
     * metadata if the cached data is missing or invalid. Updated metadata is saved
     * in the metadata bank.
     *
     * @return Bundle metadata.
     */
    Record cachedMetadata() {
        Record meta;
        Block metaId = self().asFile().metaId();
        if (auto *container = self().containerBundle()) {
            metaId = Block(metaId + container->asFile().metaId()).md5Hash();
        }
        try {
            if (Block cached = MetadataBank::get().check(CACHE_CATEGORY, metaId)) {
                cached = cached.decompressed();
                Reader(cached).withHeader() >> meta;
                return meta;
            }
        }
        catch (Error const &er) {
            LOGDEV_RES_WARNING("Corrupt cached metadata: %s") << er.asText();
        }
        meta = buildMetadata(); {
            Block buf;
            Writer(buf).withHeader() << meta;
            MetadataBank::get().setMetadata(CACHE_CATEGORY, metaId, buf.compressed());
        }
        return meta;
    }
    Record buildMetadata() {
        const String dataFilePath = self().asFile().path();
        const auto * container    = self().containerBundle();
        res::Bundles::MatchResult matched = DoomsdayApp::bundles().match(self());
        Record meta;
        meta.set(VAR_PATH,         dataFilePath);
        meta.set(VAR_BUNDLE_SCORE, matched.bestScore);
        if (format != Collection) {
            meta.addArray(VAR_DATA_FILES, new ArrayValue({ new TextValue(dataFilePath.fileName()) }));
        }
        else {
            meta.addArray(VAR_DATA_FILES);
            meta.addArray(VAR_REQUIRES);
            meta.addArray(VAR_RECOMMENDS);
            meta.addArray(VAR_EXTRAS);
        }
        if (matched) {
            packageId = matched.packageId;
            if (lumpDir) {
                meta.set(QStringLiteral("lumpDirCRC32"), lumpDir->crc32())
                        .value<NumberValue>().setSemanticHints(NumberValue::Hex);
            }
            meta.set(Package::VAR_TITLE, matched.bestMatch->keyValue("info:title"));
            meta.set(VAR_VERSION, matched.packageVersion.fullNumber());
            meta.set(VAR_AUTHOR,  matched.bestMatch->keyValue("info:author"));
            meta.set(VAR_LICENSE, matched.bestMatch->keyValue("info:license", "Unknown"));
            meta.set(VAR_TAGS,    matched.bestMatch->keyValue("info:tags"));
        }
        else {
            meta.set(Package::VAR_TITLE, self().asFile().name());
            meta.set(VAR_AUTHOR,  "Unknown");
            meta.set(VAR_LICENSE, "Unknown");
            meta.set(VAR_TAGS,    (format == Iwad || format == Pwad)? ".wad" : "");
            static String const formatDomains[] = {
                "file.local",
                "file.pk3",
                "file.wad",
                "file.iwad",
                "file.pwad",
                "file.lmp",
                "file.ded",
                "file.deh",
                "file.box"
            };
            for (const DataBundle *c = container; c; c = c->containerBundle()) {
                String containedId = cleanIdentifier(
                    stripVersion(c->sourceFile().name().fileNameWithoutExtension()));
                if (c == container && dataFilePath.fileNamePath() != c->asFile().path()) {
                    containedId = containedId.concatenateMember(cleanIdentifier(
                        stripVersion(dataFilePath.fileNamePath().fileNameWithoutExtension())));
                }
                packageId = containedId.concatenateMember(packageId);
            }
            Version parsedVersion("");
            String strippedName = stripVersion(source->name().fileNameWithoutExtension(),
                                               &parsedVersion);
            if (strippedName != source->name() && parsedVersion.isValid()) {
                meta.set(VAR_VERSION, parsedVersion.fullNumber());
            }
            else {
                meta.set(VAR_VERSION, self().asFile().status().modifiedAt
                         .asDateTime().toString("0.yyyy.MMdd.hhmm"));
            }
            packageId = stripRedundantParts(formatDomains[format]
                                            .concatenateMember(packageId)
                                            .concatenateMember(cleanIdentifier(strippedName)));
            auto &root = App::rootFolder();
            checkAuxiliaryNotes(meta);
            File const *sbInfo = root.tryLocate<File const>(
                        dataFilePath.fileNamePath() / dataFilePath.fileNameWithoutExtension() +
                        ".manifest");
            if (!sbInfo) {
                String fn = dataFilePath.fileName();
                if (int dotPos = fn.lastIndexOf('.')) {
                    if (dotPos >= 0) fn[dotPos] = '-';
                }
                sbInfo = root.tryLocate<File const>(dataFilePath.fileNamePath()/fn + ".manifest");
            }
            if (!sbInfo) {
                sbInfo = root.tryLocate<File const>(dataFilePath/"Info");
            }
            if (!sbInfo) {
                sbInfo = root.tryLocate<File const>(dataFilePath/"Contents/Info");
            }
            if (sbInfo) {
                parseSnowberryInfo(*sbInfo, meta);
            }
        }
        meta.set("ID", packageId);
        parseNotesForMetadata(meta);
        if (container && container->format() == Collection) {
            meta.appendUniqueWord(VAR_TAGS, "hidden");
        } {
            if (dataFilePath.startsWith("/home/cache/remote/")) {
                meta.appendUniqueWord(VAR_TAGS, "hidden");
                meta.appendUniqueWord(VAR_TAGS, "cached");
            } {
                static const struct {
                    uint32_t    crc32;
                    const char *filename;
                } masterLevels[] = {{0xaa78f088, "attack.wad"},   {0x56bf62c2, "blacktwr.wad"},
                                    {0xa54aee5b, "bloodsea.wad"}, {0x5a8fb0f5, "canyon.wad"},
                                    {0x20954e50, "catwalk.wad"},  {0xb237de09, "combine.wad"},
                                    {0x9f051374, "fistula.wad"},  {0x86491354, "garrison.wad"},
                                    {0x60cc2385, "geryon.wad"},   {0x9755324e, "manor.wad"},
                                    {0xcfe7d641, "mephisto.wad"}, {0xc400cf65, "minos.wad"},
                                    {0x89386748, "nessus.wad"},   {0xac8808e9, "paradox.wad"},
                                    {0x8a84cc17, "subspace.wad"}, {0x9ffd4024, "subterra.wad"},
                                    {0x96919f5e, "teeth.wad"},    {0xd8d46a55, "ttrap.wad"},
                                    {0x1726dbb7, "vesperas.wad"}, {0xd421fe9d, "virgil.wad"}};
                if (format == Pwad) {
                    for (const auto &spec : masterLevels) {
                        if (lumpDir->crc32() == spec.crc32 &&
                            self().asFile().name().compareWithoutCase(spec.filename) == 0) {
                            removeGameTags(meta);
                            meta.appendUniqueWord(VAR_TAGS, "doom2");
                            meta.appendUniqueWord(VAR_TAGS, "masterlevels");
                            break;
                        }
                    }
                }
            }
        }
        determineGameTags(meta);
        LOG_RES_VERBOSE("Identified \"%s\" %s %s score: %i")
                << packageId
                << meta.gets(VAR_VERSION)
                << ::internal::formatDescriptions[format]
                   << meta.geti(VAR_BUNDLE_SCORE); // matched.bestScore;
        return meta;
    }
    void checkAuxiliaryNotes(Record &meta) {
        if (format == Pwad || format == Iwad) {
            String const dataFilePath = self().asFile().path();
            if (File const *wadTxt = FS::tryLocate<File const>(
                        dataFilePath.fileNameAndPathWithoutExtension() + ".txt")) {
                Block txt;
                *wadTxt >> txt;
                meta.set(VAR_NOTES, _E(m) + String::fromCP437(txt));
            }
        }
    }
    /**
     * Determines if the data bundle is intended to be automatically loaded by Doomsday
     * according to the v1.x autoload rules.
     */
    bool isAutoLoaded() const {
        Path const path(self().asFile().path());
        if (path.segmentCount() >= 3) {
            String const parent      = path.reverseSegment(1).toString().toLower();
            String const grandParent = path.reverseSegment(2).toString().toLower();
            if (parent.fileNameExtension() == ".pk3" ||
                parent.fileNameExtension() == ".zip" /*||
                parent.fileNameExtension() == ".box"*/)
            {
                return true;
            }
            if (//parent.fileNameExtension().isEmpty() &&
                (/*parent == "auto" || */ parent.beginsWith("#") || parent.beginsWith("@")))
            {
                return true;
            }
        }
        for (int i = 1; i < path.segmentCount() - 3; ++i) {
            if ((path.segment(i) == "Defs" || path.segment(i) == "Data") &&
                (path.segment(i + 1) == "jDoom" || path.segment(i + 1) == "jHeretic" || path.segment(i + 1) == "jHexen") &&
                 path.segment(i + 2) == "Auto") {
                return true;
            }
        }
        return false;
    }
    /**
     * Reads a Snowberry-style Info file and extracts the relevant parts into the
     * Doomsday 2 package metadata record.
     *
     * @param infoFile  Snowberry Info file.
     * @param meta      Package metadata.
     */
    void parseSnowberryInfo(File const &infoFile, Record &meta) {
        Info info;
        String parseErrorMsg;
        try {
            info.parse(infoFile);
        }
        catch (Error const &er) {
            parseErrorMsg = er.asText();
        }
        auto const &rootBlock = info.root();
        meta.appendUniqueWord(VAR_TAGS, "legacy");
        if (rootBlock.contains("name")) {
            meta.set(Package::VAR_TITLE, rootBlock.keyValue("name"));
        }
        String component = rootBlock.keyValue("component");
        if (!component.isEmpty()) {
            if (!component.compareWithoutCase("game-jdoom")) {
                meta.appendUniqueWord(VAR_TAGS, "doom");
                meta.appendUniqueWord(VAR_TAGS, "doom2");
            }
            else if (!component.compareWithoutCase("game-jheretic")) {
                meta.appendUniqueWord(VAR_TAGS, "heretic");
            }
            else if (!component.compareWithoutCase("game-jhexen")) {
                meta.appendUniqueWord(VAR_TAGS, "hexen");
            }
        }
        String category = rootBlock.keyValue("category");
        if (!category.isEmpty()) {
            category.replace("/", "");
            category.replace(" ", "");
            category.replace("gamedata", "data"); // "gamedata" is a special tag
            category.replace("core", ""); // "core" is special tag
            category = category.trimmed();
            if (!category.isEmpty()) {
                meta.appendUniqueWord(VAR_TAGS, category);
                meta.set(VAR_CATEGORY, category);
            }
        }
        if (Info::BlockElement const *english = rootBlock.findAs<Info::BlockElement>("english")) {
            if (english->blockType() == "language") {
                Version const sbVer(english->keyValue(VAR_VERSION));
                if (sbVer.isValid()) {
                    meta.set(VAR_VERSION, sbVer.fullNumber());
                }
                meta.set(VAR_AUTHOR,  english->keyValue("author"));
                meta.set(VAR_LICENSE, english->keyValue("license"));
                meta.set("contact", english->keyValue("contact"));
                String notes = english->keyValue("readme").text.strip();
                if (!notes.isEmpty()) {
                    notes.replace(QRegExp("\\s+"), " "); // normalize whitespace
                    notes.remove('\r'); // begone foul MS-DOS
                    meta.set(VAR_NOTES, notes);
                }
            }
        }
        if (parseErrorMsg) {
            meta.appendUniqueWord(VAR_TAGS, "error");
            meta.set(VAR_NOTES, QObject::tr("There is an error in the metadata of this package: %1")
                .arg(parseErrorMsg) + "\n\n" + meta.gets(VAR_NOTES, ""));
        }
    }
    /**
     * Parses the "notes" entry for metadata in commonly used templates for WAD readmes.
     */
    void parseNotesForMetadata(Record &meta) {
        static QRegularExpression const reTitle
                ("^[\\s\x1Bm]*Title\\s*:?\\s*(.*)", QRegularExpression::CaseInsensitiveOption);
        static QRegularExpression const reVersion
                ("^\\s*Version\\s*:\\s*(.*)", QRegularExpression::CaseInsensitiveOption);
        static QRegularExpression const reReleaseDate
                ("^\\s*Release( date)?\\s*:\\s*(.*)", QRegularExpression::CaseInsensitiveOption);
        static QRegularExpression const reAuthor
                ("^\\s*Author(s)?\\s*:?\\s*(.*)", QRegularExpression::CaseInsensitiveOption);
        static QRegularExpression const reContact
                ("^\\s*Email address\\s*:?\\s*(.*)", QRegularExpression::CaseInsensitiveOption);
        bool foundVersion = false;
        bool foundTitle   = false;
        bool foundAuthor  = false;
        foreach (String line, meta.gets(VAR_NOTES, "").split('\n')) {
            if (!foundTitle) {
                auto match = reTitle.match(line);
                if (match.hasMatch()) {
                    meta.set(VAR_TITLE, match.captured(1).trimmed());
                    foundTitle = true;
                    continue;
                }
            }
            if (!foundVersion) {
                auto match = reReleaseDate.match(line);
                if (match.hasMatch()) {
                    Date const releaseDate = Date::fromText(match.captured(2).trimmed());
                    if (releaseDate.isValid()) {
                        meta.set(VAR_VERSION, QString("%1.%2.%3")
                                 .arg(releaseDate.year())
                                 .arg(releaseDate.month())
                                 .arg(releaseDate.dayOfMonth()));
                    }
                    continue;
                }
            }
            auto match = reVersion.match(line);
            if (match.hasMatch()) {
                Version parsed(match.captured(1).trimmed());
                if (parsed.isValid()) {
                    meta.set(VAR_VERSION, parsed.fullNumber());
                    foundVersion = true;
                }
                continue;
            }
            if (!foundAuthor) {
                match = reAuthor.match(line);
                if (match.hasMatch()) {
                    meta.set(VAR_AUTHOR, match.captured(2).trimmed());
                    foundAuthor = true;
                    continue;
                }
            }
            match = reContact.match(line);
            if (match.hasMatch()) {
                meta.set("contact", match.captured(1).trimmed());
                continue;
            }
        }
    }
    bool identifyMostLikelyGame(String const &text, String &identifiedTag) {
        if (text.isEmpty()) return false;
        identifiedTag.clear();
        static QList<std::pair<String, StringList>> terms;
        if (terms.isEmpty()) {
            terms << std::make_pair(String("doom2"),   StringList({ "\\b(doom2|doom 2|DoomII|Doom II|final\\s*doom|plutonia|tnt)\\b" }));
            terms << std::make_pair(String("doom"),    StringList({ "^doom$|\\bdoom[^ s2][^2d]\\b|\\bultimate\\s*doom\\b|\\budoom\\b" }));
            terms << std::make_pair(String("heretic"), StringList({ "\\b(jheretic|heretic)\\b", "\\b(d'sparil|serpent rider)\\b" }));
            terms << std::make_pair(String("hexen"),   StringList({ "\\b(jhexen|hexen)\\b", "\\b(korax|mage|warrior|cleric)\\b" }));
        }
        QHash<String, int> scores;
        for (auto i : terms) //= terms.constBegin(); i != terms.constEnd(); ++i)
        {
            for (String const &term : i.second) {
                QRegularExpression re(term, QRegularExpression::CaseInsensitiveOption);
                auto match = re.match(text);
                if (match.hasMatch()) {
                    identifiedTag = i.first;
                    return true;
                }
            }
        } {
            return false;
        }
        /*QList<std::pair<int, String>> sorted;
        for (auto i = scores.constBegin(); i != scores.constEnd(); ++i) {
            sorted.append(std::make_pair(i.value(), i.key()));
        }
        qSort(sorted.begin(), sorted.end(), []
              (std::pair<int, String> const &a, std::pair<int, String> const &b) {
            return a.first > b.first;
        });
        identifiedTag = sorted.first().second;
        return true;*/
    }
    static bool containsAnyGameTag(Record const &meta) {
        return countGameTags(meta) > 0;
    }
    static bool containsAmbiguousGameTags(Record const &meta) {
        return countGameTags(meta) != 1;
    }
    static int countGameTags(Record const &meta) {
        int count = 0;
        for (auto const &tag : gameTags()) {
            if (QRegExp(QString("\\b%1\\b").arg(tag), Qt::CaseInsensitive)
                    .indexIn(meta.gets(VAR_TAGS)) >= 0) {
                count++;
            }
        }
        return count;
    }
    static void removeGameTags(Record &meta) {
        String newTags = meta.gets(VAR_TAGS);
        newTags.remove(QRegularExpression(anyGameTagPattern()));
        meta.set(VAR_TAGS, newTags);
    }
    /**
     * Automatically guesses some appropriate game tags for the bundle.
     * @param meta  Package metadata.
     */
    void determineGameTags(Record &meta) {
        if (!containsAmbiguousGameTags(meta)) {
            return;
        }
        String const oldTags = meta.gets(VAR_TAGS);
        String tag;
        if (identifyMostLikelyGame(meta.gets(VAR_TITLE), tag)) {
            removeGameTags(meta);
            meta.appendUniqueWord(VAR_TAGS, tag);
        }
        else if (identifyMostLikelyGame(meta.gets("ID"), tag)) {
            removeGameTags(meta);
            meta.appendUniqueWord(VAR_TAGS, tag);
        }
        else if (identifyMostLikelyGame(meta.gets(VAR_NOTES, ""), tag)) {
            removeGameTags(meta);
            meta.appendUniqueWord(VAR_TAGS, tag);
        }
        if (!containsAnyGameTag(meta)) {
            Path const path(self().asFile().path());
            for (int i = 0; i < path.segmentCount(); ++i) {
                if (identifyMostLikelyGame(path.segment(i), tag)) {
                    meta.appendUniqueWord(VAR_TAGS, tag);
                }
            }
        }
        if (!containsAnyGameTag(meta)) {
            meta.set(VAR_TAGS, oldTags);
        }
    }
    struct PathAndVersion {
        String path;
        String version;
        PathAndVersion(String const &path = String(), String const &version = String())
            : path(path), version(version) {}
        operator bool() { return !path.isEmpty(); }
    };
    PathAndVersion chooseUniqueLinkPathAndVersion(File const &dataFile,
                                                  String const &packageId,
                                                  Version const &packageVersion,
                                                  dint bundleScore) {
        for (int attempt = 0; attempt < 3; ++attempt) {
            String linkPath = packageId;
            String version = (packageVersion.isValid()? packageVersion.fullNumber() : "");
            switch (attempt) {
            case 0: // unmodified
                break;
            case 1: // parse version from parent folder
                if (dataFile.path().fileNamePath() != "/local/wads") {
                    Path const filePath(dataFile.path());
                    if (filePath.segmentCount() >= 2) {
                        auto const &parentName = filePath.reverseSegment(1)
                                .toString().fileNameWithoutExtension(); {
                            Version parsed("");
                            stripVersion(parentName, &parsed);
                            if (parsed.isValid()) {
                                version = parsed.fullNumber();
                            }
                        }
                        /*else
                        {
                            version = "1.0-" + filePath.reverseSegment(1)
                                    .toString().fileNameWithoutExtension().toLower();
                        }*/
                    }
                }
                break;
            case 2: // version from status
                version = dataFile.status().modifiedAt.asDateTime().toString("0.yyyy.MMdd.hhmm");
                break;
            }
            if (!version.isEmpty()) {
                DENG2_ASSERT(Version(version).isValid());
                linkPath += QString("_%1.pack").arg(version);
            }
            else {
                linkPath += QStringLiteral(".pack");
            }
            if (!bundleFolder().has(linkPath)) {
                return PathAndVersion(linkPath, version);
            }
            else {
                auto const &file = bundleFolder().locate<File const>(linkPath);
                if (LinkFile const *linkFile = maybeAs<LinkFile>(file)) {
                    if (linkFile->isBroken()) {
                        bundleFolder().destroyFile(linkPath);
                        return PathAndVersion(linkPath, version);
                    }
                }
                if (bundleScore > file.objectNamespace().geti("package.bundleScore")) {
                    bundleFolder().destroyFile(linkPath);
                    return PathAndVersion(linkPath, version);
                }
            }
        }
        LOG_RES_XVERBOSE("Failed to make a unique link for %s (%s %s score:%i)",
                         dataFile.description() << packageId
                         << packageVersion.fullNumber() << bundleScore);
        return PathAndVersion();
    }
    String guessCompatibleGame() const {
        if (!pkgLink) return String();
        QSet<QString> tags;
        foreach (QString tag, Package::tags(*pkgLink)) tags.insert(tag);
        /*
        {
            String matchingGameId;
            if (DoomsdayApp::games().forAll([&tags, &matchingGameId] (Game &game) {
                    QRegularExpression re(QString("\\b%1\\b").arg(game.id()));
                    if (re.match(text).hasMatch()) {
                        matchinGameId = game.id();
                        return LoopAbort;
                    }
                    return LoopContinue;
                })) {
                return matchingGameId;
            }
        }*/
        res::LumpDirectory::MapType const mapType = lumpDir? lumpDir->mapType()
                                                           : res::LumpDirectory::None;
        if (tags.contains("doom") || tags.contains("doom2")) {
            if (mapType == res::LumpDirectory::MAPxx || !tags.contains("doom")) {
                return "doom2";
            }
            else {
                if (Games::get()["doom1-ultimate"].isPlayable()) {
                    return "doom1-ultimate";
                }
                return "doom1";
            }
        }
        if (tags.contains("hexen")) {
            return "hexen";
        }
        if (tags.contains("heretic")) {
            if (mapType == res::LumpDirectory::MAPxx) {
                return "hexen";
            }
            if (Games::get()["heretic-ext"].isPlayable()) {
                return "heretic-ext";
            }
            return "heretic";
        }
        return String();
    }
};
DataBundle::DataBundle(Format format, File &source)
    : d(new Impl(this, format)) {
    d->source.reset(&source);
}
DataBundle::~DataBundle()
{}
DataBundle::Format DataBundle::format() const {
    return d->format;
}
String DataBundle::formatAsText() const {
    return ::internal::formatDescriptions[d->format];
}
String DataBundle::description() const {
    if (!d->source) {
        return "invalid data bundle";
    }
    return QString("%1 %2")
            .arg(::internal::formatDescriptions[d->format])
            .arg(d->source->description());
}
File &DataBundle::asFile() {
    return *dynamic_cast<File *>(this);
}
File const &DataBundle::asFile() const {
    return *dynamic_cast<File const *>(this);
}
File const &DataBundle::sourceFile() const {
    return *asFile().source();
}
String DataBundle::rootPath() const {
    return asFile().path().fileNamePath();
}
String DataBundle::packageId() const {
    if (!d->packageId) {
        identifyPackages();
    }
    return d->packageId;
}
String DataBundle::versionedPackageId() const {
    if (!d->packageId) {
        identifyPackages();
    }
    return d->versionedPackageId;
}
IByteArray::Size DataBundle::size() const {
    if (d->source) {
        return d->source->size();
    }
    return 0;
}
void DataBundle::get(Offset at, Byte *values, Size count) const {
    if (!d->source) {
        throw File::InputError("DataBundle::get", "Source file has been destroyed");
    }
    d->source->as<ByteArrayFile>().get(at, values, count);
}
void DataBundle::set(Offset, Byte const *, Size) {
    throw File::OutputError("DataBundle::set", "Classic data formats are read-only");
}
Record &DataBundle::objectNamespace() {
    DENG2_ASSERT(dynamic_cast<File *>(this) != nullptr);
    return asFile().objectNamespace().subrecord(QStringLiteral("package"));
}
Record const &DataBundle::objectNamespace() const {
    DENG2_ASSERT(dynamic_cast<File const *>(this) != nullptr);
    return asFile().objectNamespace().subrecord(QStringLiteral("package"));
}
DataBundle::Format DataBundle::packageBundleFormat(String const &packageId) // static
{
    if (auto const *bundle = bundleForPackage(packageId)) {
        Guard g(bundle->d);
        return bundle->format();
    }
    return Unknown;
}
DataBundle const *DataBundle::bundleForPackage(String const &packageId) // static
{
    if (File const *file = PackageLoader::get().select(packageId)) {
        if (auto const *bundle = maybeAs<DataBundle>(file->target())) {
            return bundle;
        }
    }
    return nullptr;
}
DataBundle const *DataBundle::tryLocateDataFile(Package const &package, String const &dataFilePath) {
    if (DataBundle const *bundle = package.root().tryLocate<DataBundle const>(dataFilePath)) {
        return bundle;
    }
    if (DataBundle const *bundle = maybeAs<DataBundle>(package.sourceFile().target())) {
        return bundle;
    }
    return nullptr;
}
void DataBundle::setFormat(Format format) {
    d->format = format;
}
bool DataBundle::identifyPackages() const {
    LOG_AS("DataBundle");
    try {
        return d->identify();
    }
    catch (Error const &er) {
        LOG_RES_WARNING("Failed to identify %s: %s") << description() << er.asText();
    }
    return false;
}
bool DataBundle::isLinkedAsPackage() const {
    return bool(d->pkgLink);
}
Record &DataBundle::packageMetadata() {
    if (!isLinkedAsPackage()) {
        throw LinkError("DataBundle::packageMetadata", "Data bundle " +
                        description() + " has not been identified and linked as a package");
    }
    return d->pkgLink->objectNamespace().subrecord(Package::VAR_PACKAGE);
}
Record const &DataBundle::packageMetadata() const {
    return const_cast<DataBundle *>(this)->packageMetadata();
}
bool DataBundle::isNested() const {
    return containerBundle() != nullptr || !containerPackageId().isEmpty();
}
DataBundle *DataBundle::containerBundle() const {
    auto const *file = dynamic_cast<File const *>(this);
    DENG2_ASSERT(file != nullptr);
    for (Folder *folder = file->parent(); folder; folder = folder->parent()) {
        if (auto *data = maybeAs<DataFolder>(folder)) {
            return data;
        }
    }
    return nullptr;
}
String DataBundle::containerPackageId() const {
    auto const *file = dynamic_cast<File const *>(this);
    DENG2_ASSERT(file != nullptr);
    return Package::identifierForContainerOfFile(*file);
}
res::LumpDirectory const *DataBundle::lumpDirectory() const {
    return d->lumpDir.get();
}
String DataBundle::guessCompatibleGame() const {
    return d->guessCompatibleGame();
}
File *DataBundle::Interpreter::interpretFile(File *sourceData) const {
    if (LinkFile *link = maybeAs<LinkFile>(sourceData)) {
        if (link->isBroken()) return nullptr;
    }
    static struct { String str; Format format; } formats[] = {
        { ".pk3.zip", Pk3 },
        { ".pk3",     Pk3 },
        { ".wad",     Wad /* type (I or P) checked later */ },
        { ".lmp",     Lump },
        { ".ded",     Ded },
        { ".deh",     Dehacked },
        { ".box",     Collection },
    };
    for (auto const &fmt : formats) {
        if (sourceData->name().endsWith(fmt.str, String::CaseInsensitive)) {
            LOG_RES_XVERBOSE("Interpreted %s as %s",
                             sourceData->description() <<
                             ::internal::formatDescriptions[fmt.format]);
            switch (fmt.format) {
            case Pk3:
            case Collection:
                return new DataFolder(fmt.format, *sourceData);
            default:
                return new DataFile(fmt.format, *sourceData);
            }
        }
    }
    return nullptr;
}
QList<DataBundle const *> DataBundle::loadedBundles() // static
{
    QList<DataBundle const *> loaded;
    for (auto *f : PackageLoader::get().loadedPackagesAsFilesInPackageOrder()) {
        if (DataBundle const *bundle = maybeAs<DataBundle>(f)) {
            loaded << bundle;
        }
        else {
            Package const *pkg = PackageLoader::get().tryFindLoaded(*f);
            DENG2_ASSERT(pkg);
            auto const &meta = Package::metadata(*f);
            if (meta.has(VAR_DATA_FILES)) {
                foreach (Value const *v, meta.geta(VAR_DATA_FILES).elements()) {
                    String const dataFilePath = v->asText();
                    if (DataBundle const *bundle = tryLocateDataFile(*pkg, dataFilePath)) {
                        bundle->identifyPackages();
                        if (bundle->isLinkedAsPackage()) {
                            loaded << bundle;
                        }
                        else {
                            LOG_RES_WARNING("Cannot identify %s") << bundle->asFile().description();
                        }
                    }
                    else {
                        LOG_RES_WARNING("Cannot load \"%s\" from %s") << dataFilePath << f->description();
                    }
                }
            }
        }
    }
    return loaded;
}
QList<DataBundle const *> DataBundle::findAllNative(String const &fileNameOrPartialNativePath) {
    NativePath const searchPath = NativePath(fileNameOrPartialNativePath).expand();
    FS::FoundFiles found;
    FS::get().findAllOfTypes(StringList({ DENG2_TYPE_NAME(DataFile),
                                          DENG2_TYPE_NAME(DataFolder) }),
                             searchPath.fileName().toLower(), found);
    QList<DataBundle const *> bundles;
    for (auto const *f : found) {
        DENG2_ASSERT(dynamic_cast<DataBundle const *>(f));
        bundles << dynamic_cast<DataBundle const *>(f);
    }
    if (!searchPath.fileNamePath().isEmpty()) {
        bundles = de::filter(bundles, [&searchPath] (DataBundle const *b) {
            NativePath const bundlePath = b->asFile().correspondingNativePath().fileNamePath();
            if (bundlePath.isEmpty()) return false;
            if (bundlePath.toString().endsWith(searchPath.fileNamePath(),
                                               String::CaseInsensitive)) {
                return true;
            }
            return false;
        });
    }
    return bundles;
}
StringList DataBundle::gameTags() {
    static StringList const gameTags({ "doom", "doom2", "heretic", "hexen" });
    return gameTags;
}
String DataBundle::anyGameTagPattern() {
    return String("\\b(%1)\\b").arg(String::join(gameTags(), "|"));
}
String DataBundle::cleanIdentifier(String const &text) {
    String cleaned = text.toLower();
    cleaned.replace(QRegExp("[._'\\s]"), "-");
    return cleaned;
}
String DataBundle::stripVersion(String const &text, Version *version) {
    QRegExp re(".*([-_. ]v?([0-9._-]+))$");
    if (re.exactMatch(text)) {
        if (version) {
            String str = re.cap(2);
            str.replace("_", ".");
            version->parseVersionString(str);
        }
        return text.mid(0, text.size() - re.cap(1).size());
    }
    return text;
}
String DataBundle::stripRedundantParts(String const &id) {
    DotPath const path(id);
    String stripped = path.segment(0);
    for (int i = 1; i < path.segmentCount(); ++i) {
        String seg = path.segment(i);
        for (int k = 1; k <= i; ++k) // Check all previous segments.
        {
            if (seg.startsWith(path.segment(i - k) + "-")) {
                seg = seg.mid(path.segment(i - k).size() + 1);
                break;
            }
        }
        stripped = stripped.concatenateMember(seg);
    }
    return stripped;
}
de::String DataBundle::versionFromTimestamp(const Time &timestamp) {
    return timestamp.asDateTime().toString(QStringLiteral("0.yyyy.MMdd.hhmm"));
}
void DataBundle::checkAuxiliaryNotes(Record &packageMetadata) {
    d->checkAuxiliaryNotes(packageMetadata);
}
namespace res {
static String const PACKAGE_DEFS_PATH("package.defsPath");
DoomsdayPackage::DoomsdayPackage(Package const &package)
    : _pkg(package)
{}
File const &DoomsdayPackage::sourceFile() const {
    return _pkg.sourceFile();
}
bool DoomsdayPackage::hasDefinitions() const {
    return _pkg.objectNamespace().has(PACKAGE_DEFS_PATH);
}
String DoomsdayPackage::defsPath() const {
    return _pkg.objectNamespace().gets(PACKAGE_DEFS_PATH, "");
}
de::Uri DoomsdayPackage::loadableUri() const {
    return loadableUri(_pkg.file());
}
bool DoomsdayPackage::hasDefinitions(File const &packageFile) // static
{
    return packageFile.objectNamespace().has(PACKAGE_DEFS_PATH);
}
String DoomsdayPackage::defsPath(File const &packageFile) // static
{
    return packageFile.objectNamespace().gets(PACKAGE_DEFS_PATH, "");
}
de::Uri DoomsdayPackage::loadableUri(File const &packageFile) // static
{
    if (auto const *nativeSrc = maybeAs<NativeFile>(packageFile.source())) {
        return de::Uri::fromNativePath(nativeSrc->nativePath());
    }
    return de::Uri();
}
} // namespace res
namespace res {
DENG2_PIMPL(LumpCatalog) {
    using Found = std::pair<DataBundle const *, LumpDirectory::Pos>;
    StringList packageIds;
    QList<const DataBundle *> bundles;
    Impl(Public *i)
        : Base(i)
    {}
    Impl(Public *i, const Impl &other)
        : Base(i)
        , packageIds(other.packageIds)
        , bundles(other.bundles)
    {}
    void clear() {
        packageIds.clear();
        bundles.clear();
    }
    void updateBundles() {
        bundles.clear();
        for (auto const &pkg : packageIds) {
            if (File const *file = App::packageLoader().select(pkg)) {
                auto const *bundle = maybeAs<DataBundle>(file->target());
                if (bundle && bundle->lumpDirectory()) {
                    bundles << bundle;
                }
            }
        }
    }
    Found findLump(String const &name) const {
        Block const lumpName = name.toLatin1();
        Found found { nullptr, LumpDirectory::InvalidPos };
        for (int i = bundles.size() - 1; i >= 0; --i) {
            if (const auto *bundle = maybeAs<DataBundle>(bundles.at(i))) {
                auto const pos = bundle->lumpDirectory()->find(lumpName);
                if (pos != LumpDirectory::InvalidPos) {
                    found = Found(bundle, pos);
                    break;
                }
            }
            else {
                qDebug() << "LumpCatalog is outdated: a bundle has been deleted";
            }
        }
        return found;
    }
};
LumpCatalog::LumpCatalog()
    : d(new Impl(this))
{}
LumpCatalog::LumpCatalog(const LumpCatalog &other)
    : d(new Impl(this, *other.d))
{}
void LumpCatalog::clear() {
    d->clear();
}
bool LumpCatalog::setPackages(StringList packageIds) {
    if (packageIds != d->packageIds) {
        d->packageIds = packageIds;
        d->updateBundles();
        return true;
    }
    return false;
}
StringList LumpCatalog::packages() const {
    return d->packageIds;
}
Block LumpCatalog::read(String const &lumpName) const {
    Block data;
    Impl::Found found = d->findLump(lumpName);
    if (found.first) {
        auto const &entry = found.first->lumpDirectory()->entry(found.second);
        data.copyFrom(*found.first, entry.offset, entry.size);
    }
    return data;
}
} // namespace res
namespace res {
dsize const LumpDirectory::InvalidPos = dsize(-1);
static String const CACHE_CATEGORY = "LumpDirectory";
static QRegularExpression const regExMy ("^E[1-9]M[1-9]$");
static QRegularExpression const regMAPxx("^MAP[0-9][0-9]$");
DENG2_PIMPL_NOREF(LumpDirectory), public ISerializable {
    Type type = Invalid;
    MapType mapType = None;
    duint32 crc = 0;
    QList<Entry> entries;
    QHash<QByteArray, int> index; // points to entries
    void read(IByteArray const &source) {
        if (File const *file = dynamic_cast<File const *>(&source)) {
            if (readFromCache(file->metaId())) return;
        }
        Reader reader(source);
        Block typeId;
        reader.readBytes(4, typeId);
        type = (typeId == "PWAD"? Pwad : typeId == "IWAD"? Iwad : Invalid);
        if (type == Invalid) return;
        duint32 count = 0;
        duint32 dirOffset = 0;
        reader >> count >> dirOffset;
        Block data;
        reader.setOffset(dirOffset);
        reader.readBytes(16 * count, data);
        crc = M_CRC32(data.dataConst(), uint(data.size()));
        Reader lumpReader(data);
        for (duint32 i = 0; i < count; ++i) {
            Entry entry;
            lumpReader >> entry.offset >> entry.size;
            lumpReader.readBytes(8, entry.name);
            entry.name.resize(qstrlen(entry.name)); // QByteArray is zero-terminated
            entries.append(entry);
        }
        index.clear();
        for (int i = 0; i < entries.size(); ++i) {
            index.insert(entries.at(i).name, i);
            if (mapType == None) {
                String const lumpName = String::fromLatin1(entries.at(i).name);
                if (regExMy.match(lumpName).hasMatch()) {
                    mapType = ExMy;
                }
                else if (regMAPxx.match(lumpName).hasMatch()) {
                    mapType = MAPxx;
                }
            }
        }
        if (File const *file = dynamic_cast<File const *>(&source)) {
            updateCache(file->metaId());
        }
    }
    bool readFromCache(Block const &id) {
        try {
            if (Block const data = MetadataBank::get().check(CACHE_CATEGORY, id)) {
                Reader reader(data);
                reader.withHeader() >> *this;
                for (int i = 0; i < entries.size(); ++i) {
                    index.insert(entries.at(i).name, i);
                }
                return true;
            }
        }
        catch (Error const &er) {
            LOGDEV_RES_WARNING("Corrupt cached metadata: %s") << er.asText();
        }
        return false;
    }
    void updateCache(Block const &id) {
        Block data;
        Writer writer(data);
        writer.withHeader() << *this;
        MetadataBank::get().setMetadata(CACHE_CATEGORY, id, data);
    }
    void operator >> (Writer &to) const override {
        to << duint8(type) << duint8(mapType) << crc;
        to.writeElements(entries);
    }
    void operator << (Reader &from) override {
        entries.clear();
        from.readAs<duint8>(type).readAs<duint8>(mapType) >> crc;
        from.readElements(entries);
    }
};
LumpDirectory::LumpDirectory(IByteArray const &wadData)
    : d(new Impl) {
    d->read(wadData);
}
bool LumpDirectory::isValid() const {
    return d->type != Invalid;
}
LumpDirectory::Type LumpDirectory::type() const {
    return d->type;
}
LumpDirectory::MapType LumpDirectory::mapType() const {
    return d->mapType;
}
LumpDirectory::Pos LumpDirectory::count() const {
    return Pos(d->entries.size());
}
LumpDirectory::Entry const &LumpDirectory::entry(Pos pos) const {
    if (pos >= count()) {
        throw OffsetError("LumpDirectory::entry",
                          QString("Invalid position %1 (lump count: %2)")
                          .arg(pos).arg(count()));
    }
    return d->entries.at(int(pos));
}
duint32 LumpDirectory::crc32() const {
    return d->crc;
}
duint32 LumpDirectory::lumpSize(Block const &lumpName) const {
    auto found = d->index.constFind(lumpName);
    if (found != d->index.constEnd()) {
        return d->entries.at(found.value()).size;
    }
    return 0;
}
bool LumpDirectory::has(Block const &lumpName) const {
    return d->index.contains(lumpName);
}
LumpDirectory::Pos LumpDirectory::find(Block const &lumpName) const {
    auto found = d->index.constFind(lumpName);
    if (found != d->index.constEnd()) {
        return Pos(found.value());
    }
    return InvalidPos;
}
QList<LumpDirectory::Pos> res::LumpDirectory::findMaps() const {
    QList<LumpDirectory::Pos> maps;
    if (d->mapType != None) {
        for (auto i = d->index.constBegin(); i != d->index.constEnd(); ++i) {
            String const name = QString::fromLatin1(i.key());
            if (regMAPxx.match(name).hasMatch() ||
                regExMy .match(name).hasMatch()) {
                maps << Pos(i.value());
            }
        }
    }
    return maps;
}
StringList LumpDirectory::findMapLumpNames() const {
    StringList maps;
    foreach (auto pos, findMaps()) {
        maps << String::fromLatin1(entry(pos).name);
    }
    qSort(maps);
    return maps;
}
StringList LumpDirectory::mapsInContiguousRangesAsText() const {
    StringList mapRanges;
    auto const maps = findMapLumpNames();
    static StringList const prefixes({ "MAP", "E1M", "E2M", "E3M", "E4M", "E5M", "E6M",
                                       "E7M", "E8M", "E9M" });
    foreach (String const &prefix, prefixes) {
        QList<int> numbers;
        foreach (String const &map, maps) {
            if (map.startsWith(prefix)) {
                int number = map.substr(3).toInt(nullptr, 10);
                numbers.append(number);
            }
        }
        auto mapNumberText = [] (int num, int fieldWidth) {
            return QString("%1").arg(num, fieldWidth, 10, QChar('0'));
        };
        int const numFieldWidth = (prefix == QStringLiteral("MAP")? 2 : 0);
        foreach (Rangei range, Rangei::findContiguousRanges(numbers)) {
            String mr = prefix + mapNumberText(range.start, numFieldWidth);
            if (range.size() > 1) {
                mr += QString("-%1").arg(mapNumberText(range.end - 1, numFieldWidth));
            }
            mapRanges << mr;
        }
    }
    return mapRanges;
}
void LumpDirectory::Entry::operator >> (Writer &to) const {
    to << name << offset << size;
}
void LumpDirectory::Entry::operator << (Reader &from) {
    from >> name >> offset >> size;
}
} // namespace res
static short mapA[629], mapB[629];
static short tableA[1258], tableB[1258];
static const byte* srcPos;
static void cycleTable(int a, int b) {
    short               n;
    for (;;) {
        if (a == mapA[tableA[b]])
            n = mapB[tableA[b]];
        else
            n = mapA[tableA[b]];
        tableB[tableA[a]] = tableB[n] + tableB[a];
        if (tableA[a] == 1)
            break;
        a = b = tableA[a];
    }
}
static short rotateMap(int a, int b, int c) {
    if (a == mapA[tableA[a]]) {
        mapB[tableA[a]] = c;
    }
    else {
        mapA[tableA[a]] = c;
    }
    if (c == mapA[a]) {
        mapA[a] = b;
        return mapB[a];
    }
    else {
        mapB[a] = b;
        return mapA[a];
    }
}
void M_Decompress64(byte* dst, const byte* src) {
#define BUFF_SIZE       (21903)
    static const int    strides[] = {0, 16, 80, 336, 1360, 5456};
    int                 i, val, lastOut, curByte;
    byte                buff[BUFF_SIZE], curBit;
    byte*               dstPos;
    for (i = 0; i < 1258; ++i) {
       tableB[i] = 1;
    }
    for (i = 0; i < 1258; ++i) {
        tableA[i] = ((i + 2) / 2) - 1;
    }
    mapB[0] = 0;
    for (i = 0; i < 628; ++i) {
        mapB[1 + i] = (i * 2) + 3;
    }
    for (i = 0; i < 629; ++i) {
        mapA[i] = i * 2;
    }
    srcPos = src;
    dstPos = dst;
    curByte = 0;
    curBit = 0;
    lastOut = 0;
    do {
        int                 index = 1;
        while (index < 629) {
            if (curBit == 0)
                curByte = *srcPos++;
            if (curByte & 0x80)
                index = mapB[index];
            else
                index = mapA[index];
            curBit = (curBit == 0? 7 : curBit - 1);
            curByte <<= 1;
        }
        tableB[index]++;
        if (tableA[index] != 1) {
            cycleTable(index, index);
            if (tableB[1] == 2000) {
                int                 i;
                for (i = 0; i < 1258; ++i) {
                    tableB[i] >>= 1;
                }
            }
        } {
        int                 c;
        c = index;
        while (tableA[c] != 1) {
            int                 b, a;
            a = tableA[c];
            if (a == mapA[tableA[a]])
                b = mapB[tableA[a]];
            else
                b = mapA[tableA[a]];
            if (tableB[b] < tableB[c]) {
                int                 result;
                result = rotateMap(a, b, c);
                tableA[b] = a;
                tableA[c] = tableA[a];
                cycleTable(b, result);
                c = tableA[b];
            }
            else {
                c = tableA[c];
            }
        }
        }
        val = index - 629;
        if (val != 256) {
            if (val < 256) {
                byte                out = (val & 0xff);
                *dstPos++ = out;
                buff[lastOut++] = out;
            }
            else {
                int                 i, div, to, from, num, result;
                div = (val - 257) / 62;
                result = 0; {
                int                 shift = 1;
                for (i = 0; i < div * 2 + 4; ++i) {
                    if (curBit == 0)
                        curByte = *srcPos++;
                    if ((curByte) & 0x80)
                        result |= shift;
                    curBit = (curBit == 0? 7 : curBit - 1);
                    curByte <<= 1;
                    shift <<= 1;
                }
                }
                num = val - 254 - (div * 62);
                from = lastOut - result - strides[div] - num;
                if (from < 0)
                    from += BUFF_SIZE;
                to = lastOut;
                for (i = 0; i < num; ++i) {
                    byte                out = buff[from];
                    *dstPos++ = out;
                    buff[to] = out;
                    if (++from == BUFF_SIZE)
                        from = 0;
                    if (++to == BUFF_SIZE)
                        to = 0;
                }
                lastOut += num;
            }
            if (lastOut >= BUFF_SIZE)
                lastOut -= BUFF_SIZE;
        }
    } while (val != 256);
#undef BUFF_SIZE
}
DENG2_PIMPL(ResourceManifest) {
    resourceclassid_t classId;
    int flags;         ///< @ref fileFlags.
    QStringList names; ///< Known names in precedence order.
    QStringList identityKeys;
    int foundNameIndex;
    String foundPath;
    Impl(Public *i, resourceclassid_t rclass, int rflags)
        : Base(i)
        , classId(rclass)
        , flags(rflags & ~FF_FOUND)
        , names()
        , identityKeys()
        , foundNameIndex(-1)
        , foundPath()
    {}
};
ResourceManifest::ResourceManifest(resourceclassid_t resClass, int fFlags, String *name)
    : d(new Impl(this, resClass, fFlags)) {
    if (name) addName(*name);
}
void ResourceManifest::addName(String newName) {
    if (newName.isEmpty()) return;
    if (!d->names.contains(newName, Qt::CaseInsensitive)) {
        d->names.prepend(newName);
    }
}
void ResourceManifest::addIdentityKey(String newIdKey) {
    if (newIdKey.isEmpty()) return;
    if (!d->identityKeys.contains(newIdKey, Qt::CaseInsensitive)) {
        d->identityKeys.append(newIdKey);
    }
}
enum lumpsizecondition_t {
    LSCOND_NONE,
    LSCOND_EQUAL,
    LSCOND_GREATER_OR_EQUAL,
    LSCOND_LESS_OR_EQUAL
};
/**
 * Modifies the idKey so that the size condition is removed.
 */
static void checkSizeConditionInIdentityKey(String &idKey, lumpsizecondition_t *pCond, size_t *pSize) {
    DENG_ASSERT(pCond != 0);
    DENG_ASSERT(pSize != 0);
    *pCond = LSCOND_NONE;
    *pSize = 0;
    int condPos = -1;
    int argPos  = -1;
    if ((condPos = idKey.indexOf("==")) >= 0) {
        *pCond = LSCOND_EQUAL;
        argPos = condPos + 2;
    }
    else if ((condPos = idKey.indexOf(">=")) >= 0) {
        *pCond = LSCOND_GREATER_OR_EQUAL;
        argPos = condPos + 2;
    }
    else if ((condPos = idKey.indexOf("<=")) >= 0) {
        *pCond = LSCOND_LESS_OR_EQUAL;
        argPos = condPos + 2;
    }
    if (condPos < 0) return;
    *pSize = idKey.mid(argPos).toULong();
    idKey.truncate(condPos);
}
static lumpnum_t lumpNumForIdentityKey(LumpIndex const &lumpIndex, String idKey) {
    if (idKey.isEmpty()) return -1;
    lumpsizecondition_t sizeCond;
    size_t refSize;
    checkSizeConditionInIdentityKey(idKey, &sizeCond, &refSize);
    String name = idKey;
    if (idKey.fileNameExtension().isEmpty()) {
        name += ".lmp";
    }
    lumpnum_t lumpNum = lumpIndex.findLast(Path(name));
    if (lumpNum < 0) return -1;
    size_t lumpSize = lumpIndex[lumpNum].info().size;
    switch (sizeCond) {
    case LSCOND_EQUAL:
        if (lumpSize != refSize) return -1;
        break;
    case LSCOND_GREATER_OR_EQUAL:
        if (lumpSize < refSize) return -1;
        break;
    case LSCOND_LESS_OR_EQUAL:
        if (lumpSize > refSize) return -1;
        break;
    default: break;
    }
    return lumpNum;
}
static bool validateWad(String const &filePath, QStringList const &identityKeys) {
    bool validated = true;
    try {
        FileHandle &hndl = App_FileSystem().openFile(filePath, "rb", 0/*baseOffset*/, true /*allow duplicates*/);
        if (Wad *wad = maybeAs<Wad>(hndl.file())) {
            if (identityKeys.count()) {
                if (wad->isEmpty()) {
                    validated = false;
                }
                else {
                    LumpIndex lumpIndex;
                    for (int i = 0; i < wad->lumpCount(); ++i) {
                        lumpIndex.catalogLump(wad->lump(i));
                    }
                    DENG2_FOR_EACH_CONST(QStringList, i, identityKeys) {
                        if (lumpNumForIdentityKey(lumpIndex, *i) < 0) {
                            validated = false;
                            break;
                        }
                    }
                }
            }
        }
        else {
            validated = false;
        }
        App_FileSystem().releaseFile(hndl.file());
        delete &hndl;
    }
    catch (FS1::NotFoundError const &)
    {} // Ignore this error.
    return validated;
}
static bool validateZip(String const &filePath, QStringList const & /*identityKeys*/)
{
    try {
        FileHandle &hndl = App_FileSystem().openFile(filePath, "rbf");
        bool result = Zip::recognise(hndl);
        App_FileSystem().releaseFile(hndl.file());
        delete &hndl;
        return result;
    }
    catch (FS1::NotFoundError const &)
    {} // Ignore error.
    return false;
}
void ResourceManifest::locateFile() {
    if (d->flags & FF_FOUND) return;
    int nameIndex = 0;
    for (QStringList::const_iterator i = d->names.constBegin(); i != d->names.constEnd(); ++i, ++nameIndex) {
        StringList candidates;
        try {
            String foundPath = App_FileSystem().findPath(de::Uri(*i, d->classId),
                                                         RLF_DEFAULT, App_ResourceClass(d->classId));
            foundPath = App_BasePath() / foundPath; // Ensure the path is absolute.
            candidates << foundPath;
        }
        catch (FS1::NotFoundError const &)
        {} // Ignore this error.
        App::fileSystem().forAll(*i, [&candidates] (File &f) {
            if (NativeFile const *native = maybeAs<NativeFile>(f.source())) {
                candidates << native->nativePath();
            }
            return LoopContinue;
        });
        for (String foundPath : candidates) {
            bool validated = false;
            if (d->classId == RC_PACKAGE) {
                validated = validateWad(foundPath, d->identityKeys);
                if (!validated)
                    validated = validateZip(foundPath, d->identityKeys);
            }
            else {
                validated = true;
            }
            if (validated) {
                d->flags |= FF_FOUND;
                d->foundPath = foundPath;
                d->foundNameIndex = nameIndex;
                return;
            }
        }
    }
}
void ResourceManifest::forgetFile() {
    if (d->flags & FF_FOUND) {
        d->foundPath.clear();
        d->foundNameIndex = -1;
        d->flags &= ~FF_FOUND;
    }
}
String const &ResourceManifest::resolvedPath(bool tryLocate) {
    if (tryLocate) {
        locateFile();
    }
    return d->foundPath;
}
resourceclassid_t ResourceManifest::resourceClass() const {
    return d->classId;
}
int ResourceManifest::fileFlags() const {
    return d->flags;
}
QStringList const &ResourceManifest::identityKeys() const {
    return d->identityKeys;
}
QStringList const &ResourceManifest::names() const {
    return d->names;
}
namespace res {
MapManifest::MapManifest(PathTree::NodeArgs const &args)
    : Node(args), Record(), _sourceFile(nullptr)
{}
String MapManifest::description(de::Uri::ComposeAsTextFlags uriCompositionFlags) const {
    String info = String("%1").arg(composeUri().compose(uriCompositionFlags | de::Uri::DecodePath),
                                   ( uriCompositionFlags.testFlag(de::Uri::OmitScheme)? -14 : -22 ) );
    if (_sourceFile) {
        info += String(" " _E(C) "\"%1\"" _E(.)).arg(NativePath(sourceFile()->composePath()).pretty());
    }
    return info;
}
String MapManifest::composeUniqueId(Game const &currentGame) const {
    return String("%1|%2|%3|%4")
              .arg(gets("id").fileNameWithoutExtension())
              .arg(sourceFile()->name().fileNameWithoutExtension())
              .arg(sourceFile()->hasCustom()? "pwad" : "iwad")
              .arg(currentGame.id())
              .toLower();
}
MapManifest &MapManifest::setSourceFile(File1 *newSourceFile) {
    _sourceFile = newSourceFile;
    return *this;
}
File1 *MapManifest::sourceFile() const {
    return _sourceFile;
}
MapManifest &MapManifest::setRecognizer(Id1MapRecognizer *newRecognizer) {
    _recognized.reset(newRecognizer);
    return *this;
}
Id1MapRecognizer const &MapManifest::recognizer() const {
    DENG2_ASSERT(bool(_recognized));
    return *_recognized;
}
}  // namespace res
namespace res {
DENG2_PIMPL_NOREF(MapManifests) {
    Tree manifests;
    ~Impl() {
        clear();
    }
    void clear() {
        manifests.clear();
    }
};
MapManifests::MapManifests() : d(new Impl)
{}
res::MapManifest &MapManifests::findMapManifest(de::Uri const &mapUri) const {
    if (!mapUri.scheme().compareWithoutCase("Maps")) {
       if (res::MapManifest *found = d->manifests.tryFind(
                   mapUri.path(), PathTree::MatchFull | PathTree::NoBranch))
           return *found;
    }
    throw Resources::MissingResourceManifestError("MapManifests::findMapManifest", "Failed to locate a manifest for \"" + mapUri.asText() + "\"");
}
res::MapManifest *MapManifests::tryFindMapManifest(de::Uri const &mapUri) const {
    if (mapUri.scheme().compareWithoutCase("Maps")) return nullptr;
    return d->manifests.tryFind(mapUri.path(), PathTree::MatchFull | PathTree::NoBranch);
}
dint MapManifests::mapManifestCount() const {
    return d->manifests.count();
}
void MapManifests::initMapManifests() {
    d->clear();
    LumpIndex const &lumpIndex = App_FileSystem().nameIndex();
    lumpnum_t lastLump = -1;
    while (lastLump < lumpIndex.size()) {
        std::unique_ptr<Id1MapRecognizer> recognizer(new Id1MapRecognizer(lumpIndex, lastLump));
        lastLump = recognizer->lastLump();
        if (recognizer->format() != Id1MapRecognizer::UnknownFormat) {
            File1 *sourceFile  = recognizer->sourceFile();
            String const mapId = recognizer->id();
            res::MapManifest &manifest = d->manifests.insert(mapId);
            manifest.set("id", mapId);
            manifest.setSourceFile(sourceFile)
                    .setRecognizer(recognizer.release());
        }
    }
}
MapManifests::Tree const &MapManifests::allMapManifests() const {
    return d->manifests;
}
} // namespace res
namespace res {
/**
 * Serialized format header.
 */
struct Header : public IReadable {
    dint16 dimensions[2];
    dint16 origin[2];
    void operator << (Reader &from) {
        from >> dimensions[0] >> dimensions[1]
             >> origin[0] >> origin[1];
    }
};
/**
 * A @em Post is a run of one or more non-masked pixels.
 */
struct Post : public IReadable {
    dbyte topOffset;
    dbyte length;
    size_t firstPixel;
    void operator << (Reader &from) {
        from >> topOffset >> length;
        firstPixel = from.offset() + 1 /*unused junk*/;
    }
};
typedef QList<Post> Posts;
typedef Posts Column;
typedef QList<Column> Columns;
typedef QList<dint32> ColumnOffsets;
/**
 * Attempt to read another @a post from the @a reader.
 * @return  @c true if another post was read; otherwise @c false.
 */
static bool readNextPost(Post &post, Reader &reader) {
    static int const END_OF_POSTS = 0xff;
    reader.mark();
    dbyte nextByte; reader >> nextByte;
    reader.rewind();
    if (nextByte == END_OF_POSTS) return false;
    reader >> post;
    return true;
}
/**
 * Visit each of @a offsets producing a column => post map.
 */
static Columns readPosts(ColumnOffsets const &offsets, Reader &reader) {
    Columns columns;
#ifdef DENG2_QT_4_7_OR_NEWER
    columns.reserve(offsets.size());
#endif
    Post post;
    DENG2_FOR_EACH_CONST(ColumnOffsets, i, offsets) {
        reader.setOffset(*i);
        columns.push_back(Column());
        Column &column = columns.back();
        while (readNextPost(post, reader)) {
            column.push_back(post);
            reader.seek(1 + post.length + 1); // A byte of unsed junk either side of the pixel palette indices.
        }
    }
    return columns;
}
/**
 * Read @a width column offsets from the @a reader.
 */
static ColumnOffsets readColumnOffsets(int width, Reader &reader) {
    ColumnOffsets offsets;
#ifdef DENG2_QT_4_7_OR_NEWER
    offsets.reserve(width);
#endif
    for (int col = 0; col < width; ++col) {
        dint32 offset; reader >> offset;
        offsets.push_back(offset);
    }
    return offsets;
}
static inline Columns readColumns(int width, Reader &reader) {
    return readPosts(readColumnOffsets(width, reader), reader);
}
/**
 * Process @a columns to calculate the "real" pixel height of the image.
 */
static int calcRealHeight(Columns const &columns) {
    QRect geom(QPoint(0, 0), QSize(1, 0));
    DENG2_FOR_EACH_CONST(Columns, colIt, columns) {
        int tallTop = -1; // Keep track of pos (clipping).
        DENG2_FOR_EACH_CONST(Posts, postIt, *colIt) {
            Post const &post = *postIt;
            if (post.topOffset <= tallTop)
                tallTop += post.topOffset;
            else
                tallTop = post.topOffset;
            if (post.length <= 0) continue;
            geom |= QRect(QPoint(0, tallTop), QSize(1, post.length));
        }
    }
    return geom.height();
}
static Block compositeImage(Reader &reader, ColorPaletteTranslation const *xlatTable,
                            Columns const &columns, Patch::Metadata const &meta,
                            Patch::Flags flags) {
    bool const maskZero                = flags.testFlag(Patch::MaskZero);
    bool const clipToLogicalDimensions = flags.testFlag(Patch::ClipToLogicalDimensions);
#ifdef DENG2_DEBUG
    if (meta.logicalDimensions.y != meta.dimensions.y) {
        int postCount = 0;
        DENG2_FOR_EACH_CONST(Columns, i, columns) {
            postCount += i->count();
        }
        LOGDEV_RES_NOTE("Inequal heights, logical: %i != actual: %i (%i %s)")
            << meta.logicalDimensions.y << meta.dimensions.y
            << postCount << (postCount == 1? "post" : "posts");
    }
#endif
    Vector2ui const &dimensions = clipToLogicalDimensions? meta.logicalDimensions : meta.dimensions;
    int const w = dimensions.x;
    int const h = dimensions.y;
    size_t const pels = w * h;
    Block output = QByteArray(2 * pels, 0);
    dbyte *top      = output.data();
    dbyte *topAlpha = output.data() + pels;
    DENG2_FOR_EACH_CONST(Columns, colIt, columns) {
        int tallTop = -1; // Keep track of pos (clipping).
        DENG2_FOR_EACH_CONST(Posts, postIt, *colIt) {
            Post const &post = *postIt;
            if (post.topOffset <= tallTop)
                tallTop += post.topOffset;
            else
                tallTop = post.topOffset;
            if (post.length <= 0) continue;
            int y = tallTop;
            int length = post.length;
            if (y + length > h)
                length = h - y;
            int offset = 0;
            if (y < 0) {
                offset = de::min(-y, length);
                y = 0;
            }
            length = de::max(0, length - offset);
            if (!length) continue;
            reader.setOffset(post.firstPixel + offset);
            dbyte *out      = top      + tallTop * w;
            dbyte *outAlpha = topAlpha + tallTop * w;
            while (length--) {
                dbyte palIdx;
                reader >> palIdx;
                if (xlatTable) {
                    palIdx = dbyte(xlatTable->at(palIdx));
                }
                if (!maskZero || palIdx) {
                    *out = palIdx;
                }
                if (maskZero)
                    *outAlpha = (palIdx? 0xff : 0);
                else
                    *outAlpha = 0xff;
                out      += w;
                outAlpha += w;
            }
        }
        top++;
        topAlpha++;
    }
    return output;
}
static Patch::Metadata prepareMetadata(Header const &hdr, int realHeight) {
    Patch::Metadata meta;
    meta.dimensions         = Vector2ui(hdr.dimensions[0], realHeight);
    meta.logicalDimensions  = Vector2ui(hdr.dimensions[0], hdr.dimensions[1]);
    meta.origin             = Vector2i (hdr.origin[0], hdr.origin[1]);
    return meta;
}
Patch::Metadata Patch::loadMetadata(IByteArray const &data) {
    LOG_AS("Patch::loadMetadata");
    Reader reader = Reader(data);
    Header hdr; reader >> hdr;
    Columns columns = readColumns(hdr.dimensions[0], reader);
    return prepareMetadata(hdr, calcRealHeight(columns));
}
Block Patch::load(IByteArray const &data, ColorPaletteTranslation const &xlatTable, Patch::Flags flags) {
    LOG_AS("Patch::load");
    Reader reader = Reader(data);
    Header hdr; reader >> hdr;
    Columns columns = readColumns(hdr.dimensions[0], reader);
    Metadata meta = prepareMetadata(hdr, calcRealHeight(columns));
    return compositeImage(reader, &xlatTable, columns, meta, flags);
}
Block Patch::load(IByteArray const &data, Patch::Flags flags) {
    LOG_AS("Patch::load");
    Reader reader = Reader(data);
    Header hdr; reader >> hdr;
    Columns columns = readColumns(hdr.dimensions[0], reader);
    Metadata meta = prepareMetadata(hdr, calcRealHeight(columns));
    return compositeImage(reader, 0/* no translation */, columns, meta, flags);
}
bool Patch::recognize(IByteArray const &data) {
    try {
        Reader from = Reader(data);
        Header hdr; from >> hdr;
        if (!hdr.dimensions[0] || !hdr.dimensions[1]) return false;
        for (int col = 0; col < hdr.dimensions[0]; ++col) {
            dint32 offset;
            from >> offset;
            if (offset < 0 || (unsigned) offset >= from.source()->size()) return false;
        }
        return true;
    }
    catch (IByteArray::OffsetError const &) {
        return false;
    }
}
} // namespace res
namespace res {
PatchName::PatchName(String percentEncodedName, lumpnum_t lumpNum)
    : _name(percentEncodedName)
    , _lumpNum(lumpNum)
{}
lumpnum_t PatchName::lumpNum() const {
    if (_lumpNum == -2) {
        _lumpNum = -1;
        try {
            _lumpNum = App_FileSystem().lumpNumForName(_name);
        }
        catch (FS1::NotFoundError const &er) {
            LOG_RES_WARNING(er.asText() + ", ignoring.");
        }
    }
    return _lumpNum;
}
void PatchName::operator << (de::Reader &from) {
    char asciiName[9];
    for (int i = 0; i < 8; ++i) { from >> asciiName[i]; }
    asciiName[8] = 0;
    _name = QString(QByteArray(asciiName).toPercentEncoding());
    _lumpNum = -2;
}
String PatchName::percentEncodedName() const {
    return _name;
}
String const &PatchName::percentEncodedNameRef() const {
    return _name;
}
} // namespace res
#pragma pack(1)
typedef struct {
    int8_t manufacturer;
    int8_t version;
    int8_t encoding;
    int8_t bits_per_pixel;
    uint16_t xmin, ymin, xmax, ymax;
    uint16_t hres, vres;
    uint8_t palette[48];
    int8_t reserved;
    int8_t color_planes;
    uint16_t bytes_per_line;
    uint16_t palette_type;
    int8_t filler[58];
} header_t;
#pragma pack()
static char *lastPcxErrorMsg = 0; /// @todo potentially never free'd
static void PCX_SetLastError(char const *msg) {
    size_t len;
    if (0 == msg || 0 == (len = strlen(msg))) {
        if (lastPcxErrorMsg != 0) {
            M_Free(lastPcxErrorMsg);
        }
        lastPcxErrorMsg = 0;
        return;
    }
    lastPcxErrorMsg = (char *) M_Realloc(lastPcxErrorMsg, len+1);
    strcpy(lastPcxErrorMsg, msg);
}
static bool load(FileHandle &file, int width, int height, uint8_t *dstBuf) {
    DENG2_ASSERT(dstBuf != 0);
    int x, y, dataByte, runLength;
    bool result = false;
    size_t const len = file.length();
    uint8_t *raw = (uint8_t *) M_Malloc(len);
    file.read(raw, len);
    uint8_t const *srcPos = raw;
    uint8_t const *palette = srcPos + len - 768; // Palette is at the end.
    srcPos += sizeof(header_t);
    for (y = 0; y < height; ++y, dstBuf += width * 3) {
        for (x = 0; x < width;) {
            dataByte = *srcPos++;
            if ((dataByte & 0xC0) == 0xC0) {
                runLength = dataByte & 0x3F;
                dataByte = *srcPos++;
            }
            else {
                runLength = 1;
            }
            while (runLength-- > 0) {
                std::memcpy(dstBuf + x++ * 3, palette + dataByte * 3, 3);
            }
        }
    }
    if (!((size_t) (srcPos - (uint8_t *) raw) > len)) {
        PCX_SetLastError(0); // Success.
        result = true;
    }
    else {
        PCX_SetLastError("RLE inflation failed.");
    }
    M_Free(raw);
    return result;
}
char const *PCX_LastError() {
    if (lastPcxErrorMsg) {
        return lastPcxErrorMsg;
    }
    return 0;
}
uint8_t *PCX_Load(FileHandle &file, de::Vector2ui &outSize, int &pixelSize) {
    uint8_t *dstBuf = 0;
    size_t const initPos = file.tell();
    header_t hdr;
    if (file.read((uint8_t *)&hdr, sizeof(hdr)) >= sizeof(hdr)) {
        size_t dstBufSize;
        if (hdr.manufacturer != 0x0a || hdr.version != 5 ||
           hdr.encoding != 1 || hdr.bits_per_pixel != 8) {
            PCX_SetLastError("Unsupported format.");
            file.seek(initPos, SeekSet);
            return 0;
        }
        outSize   = Vector2ui(DD_SHORT(hdr.xmax) + 1, DD_SHORT(hdr.ymax) + 1);
        pixelSize = 3;
        dstBufSize = 4 * outSize.x * outSize.y;
        dstBuf = (uint8_t *) M_Malloc(dstBufSize);
        file.rewind();
        if (!load(file, outSize.x, outSize.y, dstBuf)) {
            M_Free(dstBuf);
            dstBuf = 0;
        }
    }
    file.seek(initPos, SeekSet);
    return dstBuf;
}
static ResourceClass &(*classGetter)(resourceclassid_t) = 0;
DENG2_PIMPL_NOREF(ResourceClass) {
    String name;
    String defaultScheme;
    FileTypes fileTypes;
    ~Impl() {
        qDeleteAll(fileTypes);
    }
};
ResourceClass::ResourceClass(String name, String defaultScheme)
    : d(new Impl) {
    d->name = name;
    d->defaultScheme = defaultScheme;
}
String ResourceClass::name() const {
    return d->name;
}
String ResourceClass::defaultScheme() const {
    return d->defaultScheme;
}
int ResourceClass::fileTypeCount() const {
    return d->fileTypes.size();
}
ResourceClass &ResourceClass::addFileType(FileType *ftype) {
    d->fileTypes.append(ftype);
    return *this;
}
ResourceClass::FileTypes const &ResourceClass::fileTypes() const {
    return d->fileTypes;
}
bool ResourceClass::isNull() const {
    static String const nullName = "RC_NULL";
    return d->name == nullName;
}
ResourceClass &ResourceClass::classForId(resourceclassid_t id) {
    DENG_ASSERT(classGetter != 0);
    return classGetter(id);
}
void ResourceClass::setResourceClassCallback(ResourceClass &(*callback)(resourceclassid_t)) {
    classGetter = callback;
}
static Resources *theResources = nullptr;
static String resolveUriSymbol(String const &symbol) {
    if (!symbol.compare("App.DataPath", Qt::CaseInsensitive)) {
        return "data";
    }
    else if (!symbol.compare("App.DefsPath", Qt::CaseInsensitive)) {
        return "defs";
    }
    else if (!symbol.compare("Game.IdentityKey", Qt::CaseInsensitive)) {
        if (DoomsdayApp::game().isNull()) {
            throw de::Uri::ResolveSymbolError("Resources::resolveUriSymbol",
                                              "Symbol 'Game' did not resolve (no game loaded)");
        }
        return DoomsdayApp::game().id();
    }
    else if (!symbol.compare("GamePlugin.Name", Qt::CaseInsensitive)) {
        auto &gx = DoomsdayApp::plugins().gameExports();
        if (DoomsdayApp::game().isNull() || !gx.GetPointer) {
            throw de::Uri::ResolveSymbolError("Resources::resolveUriSymbol",
                                              "Symbol 'GamePlugin' did not resolve (no game plugin loaded)");
        }
        return String(reinterpret_cast<char const *>(gx.GetPointer(DD_PLUGIN_NAME)));
    }
    else {
        throw de::Uri::UnknownSymbolError("Resources::resolveUriSymbol",
                                          "Symbol '" + symbol + "' is unknown");
    }
}
DENG2_PIMPL(Resources)
, DENG2_OBSERVES(PackageLoader, Load)
, DENG2_OBSERVES(PackageLoader, Unload) {
    typedef QList<ResourceClass *> ResourceClasses;
    ResourceClasses     resClasses;
    NullResourceClass   nullResourceClass;
    NativePath          nativeSavePath;
    res::ColorPalettes  colorPalettes;
    res::MapManifests   mapManifests;
    res::Textures       textures;
    res::AnimGroups     animGroups;
    res::Sprites        sprites;
    LoopCallback        deferredReset;
    Impl(Public *i)
        : Base(i)
        , nativeSavePath(App::app().nativeHomePath() / "savegames") // default
    {
        theResources = thisPublic;
        App::packageLoader().audienceForLoad()   += this;
        App::packageLoader().audienceForUnload() += this;
        de::Uri::setResolverFunc(resolveUriSymbol);
        resClasses << new ResourceClass("RC_PACKAGE",    "Packages")
                   << new ResourceClass("RC_DEFINITION", "Defs")
                   << new ResourceClass("RC_GRAPHIC",    "Graphics")
                   << new ResourceClass("RC_MODEL",      "Models")
                   << new ResourceClass("RC_SOUND",      "Sfx")
                   << new ResourceClass("RC_MUSIC",      "Music")
                   << new ResourceClass("RC_FONT",       "Fonts");
        auto &cmdLine = App::commandLine();
        int pos = cmdLine.has("-savedir");
        if (pos > 0) {
            cmdLine.makeAbsolutePath(pos + 1);
            String arg;
            if (cmdLine.getParameter("-savedir", arg)) {
                nativeSavePath = arg;
            }
        }
    }
    ~Impl() {
        qDeleteAll(resClasses);
        textures.clear();
        theResources = nullptr;
    }
    void packageLoaded(String const &packageId) {
        maybeScheduleResourceReset(App::packageLoader().package(packageId), true);
    }
    void aboutToUnloadPackage(String const &packageId) {
        maybeScheduleResourceReset(App::packageLoader().package(packageId), false);
    }
    void maybeScheduleResourceReset(res::DoomsdayPackage ddPkg, bool loading) {
        if (!DoomsdayApp::isGameLoaded() || DoomsdayApp::isGameBeingChanged()) {
            return;
        }
        bool needReset = ddPkg.hasDefinitions();
        File const &sourceFile = ddPkg.sourceFile();
        if (DataBundle const *bundle = maybeAs<DataBundle>(sourceFile)) {
            if (bundle->format() == DataBundle::Dehacked) {
                needReset = true;
            }
            else if (   ( loading && File1::tryLoad(File1::LoadAsCustomFile, ddPkg.loadableUri()))
                     || (!loading && File1::tryUnload(ddPkg.loadableUri()))) {
                needReset = true;
            }
        }
        if (!deferredReset && needReset) {
            deferredReset.enqueue([this] () { self().reloadAllResources(); });
        }
    }
};
Resources::Resources() : d(new Impl(this))
{}
void Resources::timeChanged(Clock const &) {
}
void Resources::clear() {
    d->colorPalettes.clearAllColorPalettes();
    d->animGroups   .clearAllAnimGroups();
    clearAllRuntimeResources();
}
void Resources::clearAllResources() {
    clearAllRuntimeResources();
    clearAllSystemResources();
}
void Resources::clearAllSystemResources() {
    textures().textureScheme("System").clear();
}
void Resources::clearAllRuntimeResources() {
    textures().clearRuntimeTextures();
}
void Resources::initSystemTextures() {
    LOG_AS("Resources");
    textures().declareSystemTexture("unknown", de::Uri("Graphics", "unknown"));
    textures().declareSystemTexture("missing", de::Uri("Graphics", "missing"));
}
void Resources::reloadAllResources()
{}
Resources &Resources::get() {
    DENG2_ASSERT(theResources);
    return *theResources;
}
ResourceClass &Resources::resClass(String name) {
    if (!name.isEmpty()) {
        foreach (ResourceClass *resClass, d->resClasses) {
            if (!resClass->name().compareWithoutCase(name))
                return *resClass;
        }
    }
    return d->nullResourceClass; // Not found.
}
ResourceClass &Resources::resClass(resourceclassid_t id) {
    if (id == RC_NULL) return d->nullResourceClass;
    if (VALID_RESOURCECLASSID(id)) {
        return *d->resClasses.at(uint(id));
    }
    throw UnknownResourceClassError("ResourceResources::toClass", QString("Invalid id '%1'").arg(int(id)));
}
NativePath Resources::nativeSavePath() const {
    return d->nativeSavePath;
}
res::MapManifests &Resources::mapManifests() {
    return d->mapManifests;
}
res::MapManifests const &Resources::mapManifests() const {
    return d->mapManifests;
}
res::ColorPalettes &Resources::colorPalettes() {
    return d->colorPalettes;
}
const res::ColorPalettes &Resources::colorPalettes() const {
    return d->colorPalettes;
}
res::Textures &Resources::textures() {
    return d->textures;
}
res::Textures const &Resources::textures() const {
    return d->textures;
}
res::AnimGroups &Resources::animGroups() {
    return d->animGroups;
}
res::AnimGroups const &Resources::animGroups() const {
    return d->animGroups;
}
res::Sprites &Resources::sprites() {
    return d->sprites;
}
res::Sprites const &Resources::sprites() const {
    return d->sprites;
}
String Resources::tryFindMusicFile(Record const &definition) {
    LOG_AS("Resources::tryFindMusicFile");
    defn::Music const music(definition);
    de::Uri songUri(music.gets("path"), RC_NULL);
    if (!songUri.path().isEmpty()) {
        String fullPath = App_BasePath() / songUri.path();
        if (F_Access(fullPath.toUtf8().constData())) {
            return fullPath;
        }
        LOG_AUDIO_WARNING("Music file \"%s\" not found (id '%s')")
                << songUri << music.gets("id");
    }
    String const lumpName = music.gets("lumpName");
    if (!lumpName.isEmpty()) {
        try {
            String const foundPath = App_FileSystem().findPath(de::Uri(lumpName, RC_MUSIC), RLF_DEFAULT,
                                                               App_ResourceClass(RC_MUSIC));
            return App_BasePath() / foundPath;  // Ensure the path is absolute.
        }
        catch (FS1::NotFoundError const &)
        {}  // Ignore this error.
    }
    return "";  // None found.
}
ResourceClass &App_ResourceClass(String className) {
    return Resources::get().resClass(className);
}
ResourceClass &App_ResourceClass(resourceclassid_t classId) {
    return Resources::get().resClass(classId);
}
/**
 * @param like             Map path search term.
 * @param composeUriFlags  Flags governing how URIs should be composed.
 */
static dint printMapsIndex2(Path const &like, de::Uri::ComposeAsTextFlags composeUriFlags) {
    res::MapManifests::Tree::FoundNodes found;
    Resources::get().mapManifests().allMapManifests()
            .findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
    if (found.isEmpty()) return 0;
    String heading = "Known maps";
    if (!like.isEmpty())
        heading += " like \"" _E(b) + like.toStringRef() + _E(.) "\"";
    LOG_RES_MSG(_E(D) "%s:" _E(.)) << heading;
    qSort(found.begin(), found.end(), comparePathTreeNodePathsAscending<res::MapManifest>);
    dint const numFoundDigits = de::max(3/*idx*/, M_NumDigits(found.count()));
    dint idx = 0;
    for (res::MapManifest *mapManifest : found) {
        String info = String("%1: " _E(1) "%2" _E(.))
                        .arg(idx, numFoundDigits)
                        .arg(mapManifest->description(composeUriFlags));
        LOG_RES_MSG("  " _E(>)) << info;
        idx++;
    }
    return found.count();
}
/**
 * @param scheme    Material subspace scheme being printed. Can be @c NULL in
 *                  which case textures are printed from all schemes.
 * @param like      Material path search term.
 * @param composeUriFlags  Flags governing how URIs should be composed.
 */
static int printMaterialIndex2(world::MaterialScheme *scheme, Path const &like,
    de::Uri::ComposeAsTextFlags composeUriFlags) {
    world::MaterialScheme::Index::FoundNodes found;
    if (scheme) // Consider resources in the specified scheme only.
    {
        scheme->index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
    }
    else // Consider resources in any scheme.
    {
        world::Materials::get().forAllMaterialSchemes([&found, &like] (world::MaterialScheme &scheme) {
            scheme.index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
            return LoopContinue;
        });
    }
    if (found.isEmpty()) return 0;
    bool const printSchemeName = !(composeUriFlags & de::Uri::OmitScheme);
    String heading = "Known materials";
    if (!printSchemeName && scheme)
        heading += " in scheme '" + scheme->name() + "'";
    if (!like.isEmpty())
        heading += " like \"" _E(b) + like.toStringRef() + _E(.) "\"";
    LOG_RES_MSG(_E(D) "%s:" _E(.)) << heading;
    qSort(found.begin(), found.end(), comparePathTreeNodePathsAscending<world::MaterialManifest>);
    int const numFoundDigits = de::max(3/*idx*/, M_NumDigits(found.count()));
    int idx = 0;
    foreach (world::MaterialManifest *manifest, found) {
        String info = String("%1: %2%3" _E(.))
                        .arg(idx, numFoundDigits)
                        .arg(manifest->hasMaterial()? _E(1) : _E(2))
                        .arg(manifest->description(composeUriFlags));
        LOG_RES_MSG("  " _E(>)) << info;
        idx++;
    }
    return found.count();
}
/**
 * @param scheme    Texture subspace scheme being printed. Can be @c NULL in
 *                  which case textures are printed from all schemes.
 * @param like      Texture path search term.
 * @param composeUriFlags  Flags governing how URIs should be composed.
 */
static int printTextureIndex2(res::TextureScheme *scheme, Path const &like,
                              de::Uri::ComposeAsTextFlags composeUriFlags) {
    res::TextureScheme::Index::FoundNodes found;
    if (scheme)  // Consider resources in the specified scheme only.
    {
        scheme->index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
    }
    else  // Consider resources in any scheme.
    {
        foreach (res::TextureScheme *scheme, res::Textures::get().allTextureSchemes()) {
            scheme->index().findAll(found, res::pathBeginsWithComparator, const_cast<Path *>(&like));
        }
    }
    if (found.isEmpty()) return 0;
    bool const printSchemeName = !(composeUriFlags & de::Uri::OmitScheme);
    String heading = "Known textures";
    if (!printSchemeName && scheme)
        heading += " in scheme '" + scheme->name() + "'";
    if (!like.isEmpty())
        heading += " like \"" _E(b) + like.toStringRef() + _E(.) "\"";
    LOG_RES_MSG(_E(D) "%s:" _E(.)) << heading;
    qSort(found.begin(), found.end(), comparePathTreeNodePathsAscending<res::TextureManifest>);
    int numFoundDigits = de::max(3/*idx*/, M_NumDigits(found.count()));
    int idx = 0;
    foreach (res::TextureManifest *manifest, found) {
        String info = String("%1: %2%3")
                        .arg(idx, numFoundDigits)
                        .arg(manifest->hasTexture()? _E(0) : _E(2))
                        .arg(manifest->description(composeUriFlags));
        LOG_RES_MSG("  " _E(>)) << info;
        idx++;
    }
    return found.count();
}
static void printMaterialIndex(de::Uri const &search,
    de::Uri::ComposeAsTextFlags flags = de::Uri::DefaultComposeAsTextFlags) {
    int printTotal = 0;
    if (search.scheme().isEmpty() && !search.path().isEmpty()) {
        printTotal = printMaterialIndex2(0/*any scheme*/, search.path(), flags & ~de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else if (world::Materials::get().isKnownMaterialScheme(search.scheme())) {
        printTotal = printMaterialIndex2(&world::Materials::get().materialScheme(search.scheme()),
                                         search.path(), flags | de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else {
        world::Materials::get().forAllMaterialSchemes([&search, &flags, &printTotal] (world::MaterialScheme &scheme) {
            int numPrinted = printMaterialIndex2(&scheme, search.path(), flags | de::Uri::OmitScheme);
            if (numPrinted) {
                LOG_MSG(_E(R));
                printTotal += numPrinted;
            }
            return LoopContinue;
        });
    }
    LOG_RES_MSG("Found " _E(b) "%i" _E(.) " %s.") << printTotal << (printTotal == 1? "material" : "materials in total");
}
static void printMapsIndex(de::Uri const &search,
    de::Uri::ComposeAsTextFlags flags = de::Uri::DefaultComposeAsTextFlags) {
    int printTotal = printMapsIndex2(search.path(), flags | de::Uri::OmitScheme);
    LOG_RES_MSG(_E(R));
    LOG_RES_MSG("Found " _E(b) "%i" _E(.) " %s.") << printTotal << (printTotal == 1? "map" : "maps in total");
}
static void printTextureIndex(de::Uri const &search,
    de::Uri::ComposeAsTextFlags flags = de::Uri::DefaultComposeAsTextFlags) {
    auto &textures = res::Textures::get();
    int printTotal = 0;
    if (search.scheme().isEmpty() && !search.path().isEmpty()) {
        printTotal = printTextureIndex2(0/*any scheme*/, search.path(), flags & ~de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else if (textures.isKnownTextureScheme(search.scheme())) {
        printTotal = printTextureIndex2(&textures.textureScheme(search.scheme()),
                                        search.path(), flags | de::Uri::OmitScheme);
        LOG_RES_MSG(_E(R));
    }
    else {
        foreach (res::TextureScheme *scheme, textures.allTextureSchemes()) {
            int numPrinted = printTextureIndex2(scheme, search.path(), flags | de::Uri::OmitScheme);
            if (numPrinted) {
                LOG_RES_MSG(_E(R));
                printTotal += numPrinted;
            }
        }
    }
    LOG_RES_MSG("Found " _E(b) "%i" _E(.) " %s") << printTotal << (printTotal == 1? "texture" : "textures in total");
}
static bool isKnownMaterialSchemeCallback(String name) {
    return world::Materials::get().isKnownMaterialScheme(name);
}
static bool isKnownTextureSchemeCallback(String name) {
    return res::Textures::get().isKnownTextureScheme(name);
}
/**
 * Print a list of all currently available maps and the location of the source
 * file which contains them.
 *
 * @todo Improve output: find common map id prefixes, find "suffixed" numerical
 * ranges, etc... (Do not assume *anything* about the formatting of map ids -
 * ZDoom allows the mod author to give a map any id they wish, which, is then
 * specified in MAPINFO when defining the map progression, clusters, etc...).
 */
D_CMD(ListMaps) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1);
    if (search.scheme().isEmpty()) search.setScheme("Maps");
    if (!search.scheme().isEmpty() && search.scheme().compareWithoutCase("Maps")) {
        LOG_RES_WARNING("Unknown scheme %s") << search.scheme();
        return false;
    }
    printMapsIndex(search);
    return true;
}
D_CMD(ListMaterials) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1, &isKnownMaterialSchemeCallback);
    if (!search.scheme().isEmpty() &&
        !world::Materials::get().isKnownMaterialScheme(search.scheme())) {
        LOG_RES_WARNING("Unknown scheme %s") << search.scheme();
        return false;
    }
    printMaterialIndex(search);
    return true;
}
D_CMD(ListTextures) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1, &isKnownTextureSchemeCallback);
    if (!search.scheme().isEmpty() &&
        !res::Textures::get().isKnownTextureScheme(search.scheme())) {
        LOG_RES_WARNING("Unknown scheme %s") << search.scheme();
        return false;
    }
    printTextureIndex(search);
    return true;
}
#ifdef DENG_DEBUG
D_CMD(PrintMaterialStats) {
    DENG2_UNUSED3(src, argc, argv);
    LOG_MSG(_E(b) "Material Statistics:");
    world::Materials::get().forAllMaterialSchemes([] (world::MaterialScheme &scheme) {
        world::MaterialScheme::Index const &index = scheme.index();
        uint count = index.count();
        LOG_MSG("Scheme: %s (%u %s)")
                << scheme.name() << count << (count == 1? "material" : "materials");
        index.debugPrintHashDistribution();
        index.debugPrint();
        return LoopContinue;
    });
    return true;
}
D_CMD(PrintTextureStats) {
    DENG2_UNUSED3(src, argc, argv);
    LOG_MSG(_E(b) "Texture Statistics:");
    foreach (res::TextureScheme *scheme, res::Textures::get().allTextureSchemes()) {
        res::TextureScheme::Index const &index = scheme->index();
        uint const count = index.count();
        LOG_MSG("Scheme: %s (%u %s)")
            << scheme->name() << count << (count == 1? "texture" : "textures");
        index.debugPrintHashDistribution();
        index.debugPrint();
    }
    return true;
}
#endif // DENG_DEBUG
void Resources::consoleRegister() {
    C_CMD("listtextures",   "ss",   ListTextures)
    C_CMD("listtextures",   "s",    ListTextures)
    C_CMD("listtextures",   "",     ListTextures)
    C_CMD("listmaterials",  "ss",   ListMaterials)
    C_CMD("listmaterials",  "s",    ListMaterials)
    C_CMD("listmaterials",  "",     ListMaterials)
    C_CMD("listmaps",       "s",    ListMaps)
    C_CMD("listmaps",       "",     ListMaps)
#ifdef DENG_DEBUG
    C_CMD("texturestats",   NULL,   PrintTextureStats)
    C_CMD("materialstats",  NULL,   PrintMaterialStats)
#endif
    SaveGames      ::consoleRegister();
    res  ::Texture ::consoleRegister();
    world::Material::consoleRegister();
}
namespace res {
using namespace de;
DENG2_PIMPL_NOREF(Sprites) {
    QHash<spritenum_t, SpriteSet> sprites;
    ~Impl() {
        sprites.clear();
    }
    inline bool hasSpriteSet(spritenum_t id) const {
        return sprites.contains(id);
    }
    SpriteSet *tryFindSpriteSet(spritenum_t id) const {
        auto found = sprites.constFind(id);
        return (found != sprites.constEnd()? const_cast<SpriteSet *>(&found.value()) : nullptr);
    }
    SpriteSet &findSpriteSet(spritenum_t id) {
        if (SpriteSet *frames = tryFindSpriteSet(id)) return *frames;
        throw Resources::MissingResourceError("Sprites::findSpriteSet",
                                              "Unknown sprite id " + String::number(id));
    }
    SpriteSet &addSpriteSet(spritenum_t id, SpriteSet const &frames) {
        DENG2_ASSERT(!tryFindSpriteSet(id));  // sanity check.
        return sprites.insert(id, frames).value();
    }
};
Sprites::Sprites()
    : d(new Impl)
{}
void Sprites::clear() {
    d->sprites.clear();
}
Sprites::SpriteSet &Sprites::addSpriteSet(spritenum_t id, SpriteSet const &frames) {
    return d->addSpriteSet(id, frames);
}
dint Sprites::spriteCount() const {
    return d->sprites.count();
}
bool Sprites::hasSprite(spritenum_t id, dint frame) const {
    if (SpriteSet const *frames = d.getConst()->tryFindSpriteSet(id)) {
        return frames->contains(frame);
    }
    return false;
}
defn::CompiledSpriteRecord &Sprites::sprite(spritenum_t id, dint frame) {
    return d->findSpriteSet(id).find(frame).value();
}
defn::CompiledSpriteRecord const *Sprites::spritePtr(spritenum_t id, de::dint frame) const {
    if (Sprites::SpriteSet const *sprSet = tryFindSpriteSet(id)) {
        auto found = sprSet->find(frame);
        if (found != sprSet->end()) return &found.value();
    }
    return nullptr;
}
Sprites::SpriteSet const *Sprites::tryFindSpriteSet(spritenum_t id) const {
    return d->tryFindSpriteSet(id);
}
Sprites::SpriteSet const &Sprites::spriteSet(spritenum_t id) const {
    return d->findSpriteSet(id);
}
struct SpriteFrameDef {
    bool mirrored = false;
    dint angle = 0;
    String material;
};
typedef QMultiMap<dint, SpriteFrameDef> SpriteFrameDefs;  ///< frame => frame angle def.
typedef QHash<String, SpriteFrameDefs> SpriteDefs;        ///< sprite name => frame set.
/**
 * In DOOM, a sprite frame is a patch texture contained in a lump existing between
 * the S_START and S_END marker lumps (in WAD) whose lump name matches the following
 * pattern:
 *
 *    NAME|A|R(A|R) (for example: "TROOA0" or "TROOA2A8")
 *
 * NAME: Four character name of the sprite.
 * A: Animation frame ordinal 'A'... (ASCII).
 * R: Rotation angle 0...G
 *    0 : Use this frame for ALL angles.
 *    1...8: Angle of rotation in 45 degree increments.
 *    A...G: Angle of rotation in 22.5 degree increments.
 *
 * The second set of (optional) frame and rotation characters instruct that the
 * same sprite frame is to be used for an additional frame but that the sprite
 * patch should be flipped horizontally (right to left) during the loading phase.
 *
 * Sprite view 0 is facing the viewer, rotation 1 is one half-angle turn CLOCKWISE
 * around the axis. This is not the same as the angle, which increases
 * counter clockwise (protractor).
 */
static SpriteDefs buildSpriteFramesFromTextures(res::TextureScheme::Index const &texIndex) {
    static dint const NAME_LENGTH = 4;
    SpriteDefs frameSets;
    frameSets.reserve(texIndex.leafNodes().count() / 8);  // overestimate.
    PathTreeIterator<res::TextureScheme::Index> iter(texIndex.leafNodes());
    while (iter.hasNext()) {
        res::TextureManifest const &texManifest = iter.next();
        String const material   = de::Uri("Sprites", texManifest.path()).compose();
        String const desc       = QString(QByteArray::fromPercentEncoding(texManifest.path().toUtf8()));
        String const spriteName = desc.left(NAME_LENGTH).toLower();
        SpriteFrameDefs *frames = nullptr;
        if (frameSets.contains(spriteName)) {
            frames = &frameSets.find(spriteName).value();
        }
        else {
            frames = &frameSets.insert(spriteName, SpriteFrameDefs()).value();
        }
        bool const haveMirror = desc.length() >= 8;
        for (dint i = 0; i < (haveMirror ? 2 : 1); ++i) {
            dint const frameNumber = desc.at(NAME_LENGTH + i * 2).toUpper().unicode() - QChar('A').unicode();
            dint const angleNumber = Sprites::toSpriteAngle(desc.at(NAME_LENGTH + i * 2 + 1));
            if (frameNumber < 0) continue;
            SpriteFrameDef *frame = nullptr;
            auto found = frames->find(frameNumber);
            if (found != frames->end() && found.value().angle == angleNumber) {
                frame = &found.value();
            }
            else {
                frame = &frames->insert(frameNumber, SpriteFrameDef()).value();
            }
            DENG2_ASSERT(frame);
            frame->material = material;
            frame->angle    = angleNumber;
            frame->mirrored = i == 1;
        }
    }
    return frameSets;
}
/**
 * Generates a set of Sprites from the given @a frameSet.
 *
 * @note Gaps in the frame number range will be filled with dummy Sprite instances
 * (no view angles added).
 *
 * @param frameDefs  SpriteFrameDefs to process.
 *
 * @return  Newly built sprite frame-number => definition map.
 *
 * @todo Don't do this here (no need for two-stage construction). -ds
 */
static Sprites::SpriteSet buildSprites(QMultiMap<dint, SpriteFrameDef> const &frameDefs) {
    static de::dint const MAX_ANGLES = 16;
    Sprites::SpriteSet frames;
    for (auto it = frameDefs.constBegin(); it != frameDefs.constEnd(); ++it) {
        defn::CompiledSpriteRecord *rec = nullptr;
        auto found = frames.find(it.key());
        if (found != frames.end()) {
            rec = &found.value();
        }
        else {
            rec = &frames[it.key()];
            defn::Sprite(*rec).resetToDefaults();
        }
        SpriteFrameDef const &def = it.value();
        defn::Sprite(*rec).addView(def.material, def.angle, def.mirrored);
    }
    for (defn::CompiledSpriteRecord &rec : frames) {
        defn::Sprite sprite(rec);
        if (sprite.viewCount() < 2)
            continue;
        for (dint angle = 0; angle < MAX_ANGLES / 2; ++angle) {
            if (!sprite.hasView(angle * 2 + 1) && sprite.hasView(angle * 2)) {
                auto src = sprite.view(angle * 2);
                sprite.addView(src.material->asText(), angle * 2 + 2, src.mirrorX);
            }
            if (!sprite.hasView(angle * 2) && sprite.hasView(angle * 2 + 1)) {
                auto src = sprite.view(angle * 2 + 1);
                sprite.addView(src.material->asText(), angle * 2 + 1, src.mirrorX);
            }
        }
    }
    return frames;
}
void Sprites::initSprites() {
    LOG_AS("Sprites");
    LOG_RES_VERBOSE("Building sprites...");
    Time begunAt;
    clear();
    dint customIdx = 0;
    SpriteDefs spriteDefs = buildSpriteFramesFromTextures(res::Textures::get().textureScheme("Sprites").index());
    for (auto it = spriteDefs.constBegin(); it != spriteDefs.constEnd(); ++it) {
        spritenum_t id = DED_Definitions()->getSpriteNum(it.key());
        if (id == -1) {
            id = (DED_Definitions()->sprites.size() + customIdx++);
        }
        addSpriteSet(id, buildSprites(it.value()));
    }
    spriteDefs.clear();
    LOG_RES_VERBOSE("Sprites built in %.2f seconds") << begunAt.since();
}
dint Sprites::toSpriteAngle(QChar angleCode) // static
{
    static dint const MAX_ANGLES = 16;
    dint angle = -1; // Unknown.
    if (angleCode.isDigit()) {
        angle = angleCode.digitValue();
    }
    else if (angleCode.isLetter()) {
        char charCodeLatin1 = angleCode.toUpper().toLatin1();
        if (charCodeLatin1 >= 'A') {
            angle = charCodeLatin1 - 'A' + 10;
        }
    }
    if (angle < 0 || angle > MAX_ANGLES)
        return -1;
    if (angle == 0) return 0;
    if (angle <= MAX_ANGLES / 2) {
        return (angle - 1) * 2 + 1;
    }
    else {
        return (angle - 9) * 2 + 2;
    }
}
bool Sprites::isValidSpriteName(String name) // static
{
    if (name.length() < 6) return false;
    if (toSpriteAngle(name.at(5)) < 0) return false;
    return (name.length() <= 7 || toSpriteAngle(name.at(7)) >= 0);
}
Sprites &Sprites::get() // static
{
    return Resources::get().sprites();
}
} // namespace res
namespace res {
typedef QHash<Texture::AnalysisId, void *> Analyses;
DENG2_PIMPL(Texture) {
    TextureManifest &manifest;
    Flags flags;
    void *userData;
    Vector2ui dimensions;
    Vector2i origin;
    Analyses analyses;
    Impl(Public *i, TextureManifest &_manifest)
        : Base(i)
        , manifest(_manifest)
        , userData(0)
    {}
    ~Impl() {
        self().clearAnalyses();
    }
    void notifyDimensionsChanged() {
        DENG2_FOR_PUBLIC_AUDIENCE(DimensionsChange, i) i->textureDimensionsChanged(self());
    }
};
Texture::Texture(TextureManifest &manifest) : d(new Impl(this, manifest)) {
    setFlags(manifest.flags());
    setDimensions(manifest.logicalDimensions());
    setOrigin(manifest.origin());
}
Texture::~Texture() {
    DENG2_FOR_AUDIENCE(Deletion, i) i->textureBeingDeleted(*this);
    if (!manifest().schemeName().compareWithoutCase("Textures")) {
        Composite *pcTex = reinterpret_cast<Composite *>(userDataPointer());
        if (pcTex) delete pcTex;
    }
}
TextureManifest &Texture::manifest() const {
    return d->manifest;
}
void Texture::setUserDataPointer(void *newUserData) {
    if (d->userData && newUserData) {
        LOG_AS("Texture::setUserDataPointer");
        LOGDEV_RES_MSG("User data already present for \"%s\" %p, will be replaced")
            << d->manifest.composeUri() << this;
    }
    d->userData = newUserData;
}
void *Texture::userDataPointer() const {
    return d->userData;
}
Vector2ui const &Texture::dimensions() const {
    return d->dimensions;
}
void Texture::setDimensions(Vector2ui const &newDimensions) {
    if (d->dimensions != newDimensions) {
        d->dimensions = newDimensions;
        d->notifyDimensionsChanged();
    }
}
void Texture::setWidth(duint newWidth) {
    if (d->dimensions.x != newWidth) {
        d->dimensions.x = newWidth;
        d->notifyDimensionsChanged();
    }
}
void Texture::setHeight(duint newHeight) {
    if (d->dimensions.y != newHeight) {
        d->dimensions.y = newHeight;
        d->notifyDimensionsChanged();
    }
}
Vector2i const &Texture::origin() const {
    return d->origin;
}
void Texture::setOrigin(Vector2i const &newOrigin) {
    if (d->origin != newOrigin) {
        d->origin = newOrigin;
    }
}
void Texture::release(/*TextureVariantSpec *spec*/)
{}
void Texture::clearAnalyses() {
    foreach (void *data, d->analyses) {
        M_Free(data);
    }
    d->analyses.clear();
}
void *Texture::analysisDataPointer(AnalysisId analysisId) const {
    auto const &analyses = d.getConst()->analyses;
    auto found = analyses.find(analysisId);
    if (found != analyses.end()) {
        return found.value();
    }
    return nullptr;
}
void Texture::setAnalysisDataPointer(AnalysisId analysisId, void *newData) {
    LOG_AS("Texture::attachAnalysis");
    void *existingData = analysisDataPointer(analysisId);
    if (existingData) {
#ifdef DENG2_DEBUG
        if (newData) {
            LOGDEV_RES_VERBOSE("Image analysis (id:%i) already present for \"%s\", will be replaed")
                << int(analysisId) << d->manifest.composeUri();
        }
#endif
        M_Free(existingData);
    }
    d->analyses.insert(analysisId, newData);
}
Texture::Flags Texture::flags() const {
    return d->flags;
}
void Texture::setFlags(Texture::Flags flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
}
String Texture::description() const {
    String str = String("Texture " _E(b) "%1" _E(.)).arg(manifest().composeUri().asText());
#ifdef DENG_DEBUG
    str += String(" [addr:0x%1]").arg(de::dintptr(this), 0, 16);
#endif
    str += _E(l) " Dimensions:" _E(.)
        +  (width() == 0 && height() == 0? String("unknown (not yet prepared)")
                                         : dimensions().asText())
        +  _E(l) " Source:" _E(.) + manifest().sourceDescription();
    return str;
}
D_CMD(InspectTexture) {
    DENG2_UNUSED(src);
    de::Uri search = de::Uri::fromUserInput(&argv[1], argc - 1);
    if (!search.scheme().isEmpty() &&
        !Resources::get().textures().isKnownTextureScheme(search.scheme())) {
        LOG_RES_WARNING("Unknown scheme %s") << search.scheme();
        return false;
    }
    try {
        TextureManifest &manifest = Resources::get().textures().textureManifest(search);
        if (manifest.hasTexture()) {
            LOG_RES_MSG("%s") << manifest.texture().description();
            /*LOG_RES_MSG("Texture " _E(b) "%s" _E(.) "%s"
                        "\n" _E(l) "Dimensions: " _E(.) _E(i) "%s" _E(.)
                        " " _E(l) "Source: "     _E(.) _E(i) "%s")
                << manifest.composeUri()
                << variantCountText
                << (texture.width() == 0 &&
                    texture.height() == 0? String("unknown (not yet prepared)")
                                         : texture.dimensions().asText())
                << manifest.sourceDescription();*/
        }
        else {
            LOG_RES_MSG("%s") << manifest.description();
        }
        return true;
    }
    catch (Resources::MissingResourceManifestError const &er) {
        LOG_RES_WARNING("%s.") << er.asText();
    }
    return false;
}
void Texture::consoleRegister() // static
{
    C_CMD("inspecttexture", "ss",   InspectTexture)
    C_CMD("inspecttexture", "s",    InspectTexture)
}
} // namespace res
namespace res {
static TextureManifest::TextureConstructor textureConstructor;
DENG2_PIMPL(TextureManifest)
, DENG2_OBSERVES(Texture, Deletion) {
    int uniqueId;                    ///< Scheme-unique identifier (user defined).
    de::Uri resourceUri;             ///< Image resource path, to be loaded.
    Vector2ui logicalDimensions;     ///< Dimensions in map space.
    Vector2i origin;                 ///< Origin offset in map space.
    Texture::Flags flags;            ///< Classification flags.
    std::unique_ptr<Texture> texture;///< Associated resource (if any).
    TextureScheme *ownerScheme = nullptr;
    Impl(Public *i)
        : Base(i)
        , uniqueId(0)
    {}
    ~Impl() {
        if (texture) texture->audienceForDeletion -= this;
        DENG2_FOR_PUBLIC_AUDIENCE(Deletion, i) i->textureManifestBeingDeleted(self());
    }
    void textureBeingDeleted(Texture const & /*texture*/)
    {
        texture.release();
    }
};
TextureManifest::TextureManifest(PathTree::NodeArgs const &args)
    : Node(args)
    , d(new Impl(this))
{}
Texture *TextureManifest::derive() {
    LOG_AS("TextureManifest::derive");
    if (!hasTexture()) {
        DENG2_ASSERT(textureConstructor != nullptr);
        setTexture(textureConstructor(*this));
        DENG2_FOR_AUDIENCE(TextureDerived, i) i->textureManifestTextureDerived(*this, texture());
    }
    else {
        Texture *tex = &texture();
        tex->setFlags(d->flags);
        tex->setDimensions(d->logicalDimensions);
        tex->setOrigin(d->origin);
    }
    return &texture();
}
void TextureManifest::setScheme(TextureScheme &ownerScheme) {
    d->ownerScheme = &ownerScheme;
}
TextureScheme &TextureManifest::scheme() const {
    return *d->ownerScheme;
}
String const &TextureManifest::schemeName() const {
    return scheme().name();
}
String TextureManifest::description(de::Uri::ComposeAsTextFlags uriCompositionFlags) const {
    String info = String("%1 %2")
                      .arg(composeUri().compose(uriCompositionFlags | de::Uri::DecodePath),
                           ( uriCompositionFlags.testFlag(de::Uri::OmitScheme)? -14 : -22 ) )
                      .arg(sourceDescription(), -7);
#ifdef __CLIENT__
    info += String("x%1").arg(!hasTexture()? 0 : texture().variantCount());
#endif
    info += " " + (!hasResourceUri()? "N/A" : resourceUri().asText());
    return info;
}
String TextureManifest::sourceDescription() const {
    if (!hasTexture()) return "unknown";
    if (texture().isFlagged(Texture::Custom)) return "add-on";
    return "game";
}
bool TextureManifest::hasResourceUri() const {
    return !d->resourceUri.isEmpty();
}
de::Uri TextureManifest::resourceUri() const {
    if (hasResourceUri()) {
        return d->resourceUri;
    }
    throw MissingResourceUriError("TextureManifest::scheme", "No resource URI is defined");
}
bool TextureManifest::setResourceUri(de::Uri const &newUri) {
    if (d->resourceUri.asText() != newUri.asText()) {
        d->resourceUri = newUri;
        return true;
    }
    return false;
}
int TextureManifest::uniqueId() const {
    return d->uniqueId;
}
bool TextureManifest::setUniqueId(int newUniqueId) {
    if (d->uniqueId == newUniqueId) return false;
    d->uniqueId = newUniqueId;
    DENG2_FOR_AUDIENCE(UniqueIdChange, i) i->textureManifestUniqueIdChanged(*this);
    return true;
}
Texture::Flags TextureManifest::flags() const {
    return d->flags;
}
void TextureManifest::setFlags(Texture::Flags flagsToChange, FlagOp operation) {
    applyFlagOperation(d->flags, flagsToChange, operation);
}
Vector2ui const &TextureManifest::logicalDimensions() const {
    return d->logicalDimensions;
}
bool TextureManifest::setLogicalDimensions(Vector2ui const &newDimensions) {
    if (d->logicalDimensions == newDimensions) return false;
    d->logicalDimensions = newDimensions;
    return true;
}
Vector2i const &TextureManifest::origin() const {
    return d->origin;
}
void TextureManifest::setOrigin(Vector2i const &newOrigin) {
    if (d->origin != newOrigin) {
        d->origin = newOrigin;
    }
}
bool TextureManifest::hasTexture() const {
    return bool(d->texture);
}
Texture &TextureManifest::texture() const {
    if (hasTexture()) {
        return *d->texture;
    }
    throw MissingTextureError("TextureManifest::texture", "No texture is associated");
}
Texture *TextureManifest::texturePtr() const {
    return d->texture.get();
}
void TextureManifest::setTexture(Texture *newTexture) {
    if (d->texture.get() != newTexture) {
        if (d->texture) {
            d->texture->audienceForDeletion -= d;
        }
        d->texture.reset(newTexture);
        if (d->texture) {
            d->texture->audienceForDeletion += d;
        }
    }
}
void TextureManifest::setTextureConstructor(TextureConstructor constructor) {
    textureConstructor = constructor;
}
} // namespace res
namespace res {
uint qHash(TextureSchemeHashKey const &key) {
    return key.scheme.at(2).toLower().unicode();
}
DENG2_PIMPL(Textures)
, DENG2_OBSERVES(TextureScheme,   ManifestDefined)
, DENG2_OBSERVES(TextureManifest, TextureDerived)
, DENG2_OBSERVES(Texture,         Deletion) {
    TextureSchemes textureSchemes;
    QVector<TextureScheme *> textureSchemeCreationOrder;
    Composite::ArchiveFormat compositeFormat = Composite::DoomFormat;
    AllTextures textures;
    Impl(Public *i) : Base(i) {
        TextureManifest::setTextureConstructor([] (TextureManifest &m) -> Texture * {
            return new Texture(m);
        });
        createTextureScheme("Sprites");
        createTextureScheme("Textures");
        createTextureScheme("Flats");
        createTextureScheme("Patches");
        createTextureScheme("System");
        createTextureScheme("Details");
        createTextureScheme("Reflections");
        createTextureScheme("Masks");
        createTextureScheme("ModelSkins");
        createTextureScheme("ModelReflectionSkins");
        createTextureScheme("Lightmaps");
        createTextureScheme("Flaremaps");
    }
    ~Impl() {
        clear();
    }
    void clear() {
        clearTextureManifests();
        clearAllTextureSchemes();
    }
    void clearTextureManifests() {
        qDeleteAll(textureSchemes);
        textureSchemes.clear();
        textureSchemeCreationOrder.clear();
    }
    void clearAllTextureSchemes() {
        foreach (TextureScheme *scheme, textureSchemes) {
            scheme->clear();
        }
    }
    void createTextureScheme(String const &name) {
        DENG2_ASSERT(name.length() >= TextureScheme::min_name_length);
        TextureScheme *newScheme = new TextureScheme(name);
        textureSchemes.insert(name, newScheme);
        textureSchemeCreationOrder << newScheme;
        newScheme->audienceForManifestDefined += this;
    }
    void textureSchemeManifestDefined(TextureScheme & /*scheme*/, TextureManifest &manifest) override
    {
        manifest.audienceForTextureDerived += this;
    }
    void textureManifestTextureDerived(TextureManifest & /*manifest*/, Texture &texture) override
    {
        textures << &texture;
        texture.audienceForDeletion += this;
    }
    void textureBeingDeleted(Texture const &texture) override {
        textures.remove(const_cast<Texture *>(&texture));
    }
    static QVector<File1 *> collectPatchCompositeDefinitionFiles() {
        QVector<File1 *> result;
        LumpIndex const &index  = App_FileSystem().nameIndex();
        lumpnum_t firstTexLump  = App_FileSystem().lumpNumForName("TEXTURE1");
        lumpnum_t secondTexLump = App_FileSystem().lumpNumForName("TEXTURE2");
        for (dint i = 0; i < index.size(); ++i) {
            File1 &file = index[i];
            if (i == firstTexLump ) continue;
            if (i == secondTexLump) continue;
            String fileName = file.name().fileNameWithoutExtension();
            if (fileName.compareWithoutCase("TEXTURE1") &&
                fileName.compareWithoutCase("TEXTURE2")) {
                continue;
            }
            result.append(&file);
        }
        if (firstTexLump >= 0) {
            result.append(&index[firstTexLump]);
        }
        if (secondTexLump >= 0) {
            result.append(&index[secondTexLump]);
        }
        return result;
    }
    typedef QVector<Composite *> Composites;
    typedef QVector<PatchName>   PatchNames;
    static PatchNames readPatchNames(File1 &file) {
        LOG_AS("readPatchNames");
        PatchNames names;
        if (file.size() < 4) {
            LOG_RES_WARNING("File \"%s\" does not appear to be valid PNAMES data")
                    << NativePath(file.composeUri().asText()).pretty();
            return names;
        }
        ByteRefArray lumpData(file.cache(), file.size());
        de::Reader from(lumpData);
        dint32 numNames;
        from >> numNames;
        if (numNames > 0) {
            if (unsigned(numNames) > (file.size() - 4) / 8) {
                LOG_RES_WARNING("File \"%s\" appears to be truncated (%u bytes, expected %u)")
                        << NativePath(file.composeUri().asText()).pretty()
                        << file.size() << (numNames * 8 + 4);
                numNames = (file.size() - 4) / 8;
            }
            for (int i = 0; i < numNames; ++i) {
                PatchName name;
                from >> name;
                names.append(name);
            }
        }
        file.unlock();
        return names;
    }
    /**
     * Reads patch composite texture definitions from @a file.
     *
     * @param file           File to be read.
     * @param origIndexBase  Base value for the "original index" logic.
     * @param archiveCount   Will be updated with the total number of definitions
     *                       in the file (which may not necessarily equal the total
     *                       number of definitions which are actually read).
     */
    Composites readCompositeTextureDefs(File1 &file,
                                        PatchNames const &patchNames,
                                        int origIndexBase,
                                        int &archiveCount) const {
        LOG_AS("readCompositeTextureDefs");
        Composites result; ///< The resulting set of validated definitions.
        Composite::ArchiveFormat format = compositeFormat;
        ByteRefArray data(file.cache(), file.size());
        de::Reader reader(data);
        dint32 definitionCount;
        reader >> definitionCount;
        typedef QMap<dint32, int> Offsets;
        Offsets offsets;
        for (int i = 0; i < definitionCount; ++i) {
            dint32 offset;
            reader >> offset;
            if (offset < 0 || unsigned(offset) < unsigned(definitionCount) * sizeof(offset) ||
                dsize(offset) > reader.source()->size()) {
                LOG_RES_WARNING("Ignoring definition #%i: invalid offset %i") << i << offset;
            }
            else {
                offsets.insert(offset, origIndexBase + i);
            }
        }
        DENG2_FOR_EACH_CONST(Offsets, i, offsets) {
            reader.setOffset(i.key());
            Composite *def = Composite::constructFrom(reader, patchNames, format);
            def->setOrigIndex(i.value());
            DENG2_FOR_EACH_CONST(Composite::Components, it, def->components()) {
                if (it->lumpNum() >= 0) {
                    result.append(def);
                    def = 0;
                    break;
                }
            }
            if (def) delete def;
        }
        file.unlock(); // We have now finished with this file.
        archiveCount = definitionCount;
        return result;
    }
    Composites loadCompositeTextureDefs() {
        LOG_AS("loadCompositeTextureDefs");
        typedef QMultiMap<String, Composite *> CompositeTextureMap;
        PatchNames pnames;
        try {
            auto &fs1 = App_FileSystem();
            pnames = readPatchNames(fs1.lump(fs1.lumpNumForName("PNAMES")));
        }
        catch (LumpIndex::NotFoundError const &er) {
            if (DoomsdayApp::isGameLoaded()) {
                LOGDEV_RES_WARNING(er.asText());
            }
        }
        if (!pnames.count()) return Composites();
        auto const defFiles = collectPatchCompositeDefinitionFiles();
        /**
         * Definitions are read into two discreet sets.
         *
         * Older add-ons contain copies of the original games' texture definitions,
         * with their own new definitions appended on the end. However, Doomsday needs
         * to classify all definitions according to whether they originate from the
         * original game data. To achieve the correct user-expected results, we must
         * compare each definition originating from an add-on to determine whether it
         * should instead be classified as "original" data.
         */
        Composites defs, customDefs;
        int origIndexBase = 0;
        foreach (auto *file, defFiles) {
            LOG_RES_VERBOSE("Processing \"%s:%s\"...")
                << NativePath(file->container().composeUri().asText()).pretty()
                << NativePath(file->composeUri().asText()).pretty();
            int archiveCount;
            Composites newDefs = readCompositeTextureDefs(*file, pnames, origIndexBase, archiveCount);
            Composites *existingDefs =
                    (file->container().hasCustom()? &customDefs : &defs);
            if (!existingDefs->isEmpty()) {
                *existingDefs += newDefs;
            }
            else {
                *existingDefs = newDefs;
            }
            origIndexBase += archiveCount;
            LOG_RES_MSG("Loaded %s texture definitions from \"%s:%s\"")
                << (newDefs.count() == archiveCount? String("all %1").arg(newDefs.count())
                                                   : String("%1 of %1").arg(newDefs.count()).arg(archiveCount))
                << NativePath(file->container().composeUri().asText()).pretty()
                << NativePath(file->composeUri().asText()).pretty();
        }
        if (!customDefs.isEmpty()) {
            CompositeTextureMap mappedCustomDefs;
            foreach (Composite *custom, customDefs) {
                mappedCustomDefs.insert(custom->percentEncodedNameRef(), custom);
            }
            for (int i = 0; i < defs.count(); ++i) {
                Composite *orig = defs[i];
                CompositeTextureMap::const_iterator found = mappedCustomDefs.constFind(orig->percentEncodedNameRef());
                if (found == mappedCustomDefs.constEnd())
                    continue;
                Composite *custom = found.value();
                bool haveReplacement = false;
                if (custom->isFlagged(Composite::Custom)) {
                    haveReplacement = true; // Uses a custom patch.
                }
                else if (*orig != *custom) {
                    haveReplacement = true;
                }
                if (haveReplacement) {
                    custom->setFlags(Composite::Custom);
                    defs.takeAt(i);
                    delete orig;
                    --i; // Process the new next definition item.
                }
            }
            /*
             * List 'defs' now contains only those definitions which are not
             * superceeded by those in the 'customDefs' list.
             */
            defs += customDefs;
        }
        return defs;
    }
    void initCompositeTextures() {
        Time begunAt;
        LOG_RES_VERBOSE("Initializing composite textures...");
        Composites allDefs = loadCompositeTextureDefs();
        while (!allDefs.isEmpty()) {
            Composite &def = *allDefs.takeFirst();
            de::Uri uri(QStringLiteral("Textures"), Path(def.percentEncodedName()));
            Texture::Flags flags;
            if (def.isFlagged(Composite::Custom)) flags |= Texture::Custom;
            /*
             * The id Tech 1 implementation of the texture collection has a flaw
             * which results in the first texture being used dually as a "NULL"
             * texture.
             */
            if (def.origIndex() == 0) flags |= Texture::NoDraw;
            try {
                TextureManifest &manifest = self().declareTexture(
                    uri,
                    flags,
                    def.percentEncodedName().beginsWith("SKY") ? def.dimensions()
                                                               : def.logicalDimensions(),
                    Vec2i(),
                    def.origIndex());
                if (manifest.hasTexture()) {
                    Texture &tex = manifest.texture();
                    Composite *oldDef = reinterpret_cast<Composite *>(tex.userDataPointer());
                    if (oldDef) {
                        tex.setUserDataPointer(0);
                        delete oldDef;
                    }
                    tex.setUserDataPointer((void *)&def);
                    continue;
                }
                else if (Texture *tex = manifest.derive()) {
                    tex->setUserDataPointer((void *)&def);
                    continue;
                }
            }
            catch (TextureScheme::InvalidPathError const &er) {
                LOG_RES_WARNING("Failed declaring texture \"%s\": %s")
                        << uri << er.asText();
            }
            delete &def;
        }
        LOG_RES_VERBOSE("initCompositeTextures: Completed in %.2f seconds") << begunAt.since();
    }
    void initFlatTextures() {
        Time begunAt;
        LOG_RES_VERBOSE("Initializing Flat textures...");
        LumpIndex const &index = App_FileSystem().nameIndex();
        lumpnum_t firstFlatMarkerLumpNum = index.findFirst(Path("F_START.lmp"));
        if (firstFlatMarkerLumpNum >= 0) {
            lumpnum_t lumpNum;
            File1 *blockContainer = 0;
            for (lumpNum = index.size(); lumpNum --> firstFlatMarkerLumpNum + 1;) {
                File1 &file = index[lumpNum];
                String percentEncodedName = file.name().fileNameWithoutExtension();
                if (blockContainer && blockContainer != &file.container()) {
                    blockContainer = 0;
                }
                if (!blockContainer) {
                    if (!percentEncodedName.compareWithoutCase("F_END") ||
                        !percentEncodedName.compareWithoutCase("FF_END")) {
                        blockContainer = &file.container();
                    }
                    continue;
                }
                if (!percentEncodedName.compareWithoutCase("F_START")) {
                    blockContainer = 0;
                    continue;
                }
                if (!percentEncodedName.compareWithoutCase("FF_START") ||
                    !percentEncodedName.compareWithoutCase("F_END")    ||
                    !percentEncodedName.compareWithoutCase("FF_END")) continue;
                de::Uri uri(QStringLiteral("Flats"), Path(percentEncodedName));
                if (self().textureManifestPtr(uri)) continue;
                Texture::Flags flags;
                if (file.container().hasCustom()) flags |= Texture::Custom;
                /*
                 * Kludge Assume 64x64 else when the flat is loaded it will inherit the
                 * pixel dimensions of the graphic, which, if it has been replaced with
                 * a hires version - will be much larger than it should be.
                 *
                 * @todo Always determine size from the lowres original.
                 */
                Vector2ui dimensions(64, 64);
                Vector2i origin(0, 0);
                int const uniqueId  = lumpNum - (firstFlatMarkerLumpNum + 1);
                de::Uri resourceUri = LumpIndex::composeResourceUrn(lumpNum);
                self().declareTexture(uri, flags, dimensions, origin, uniqueId, &resourceUri);
            }
        }
        self().deriveAllTexturesInScheme("Flats");
        LOG_RES_VERBOSE("Flat textures initialized in %.2f seconds") << begunAt.since();
    }
    void initSpriteTextures() {
        Time begunAt;
        LOG_RES_VERBOSE("Initializing Sprite textures...");
        dint uniqueId = 1/*1-based index*/;
        ddstack_t *stack = Stack_New();
        LumpIndex const &index = App_FileSystem().nameIndex();
        for (dint i = 0; i < index.size(); ++i) {
            File1 &file = index[i];
            String fileName = file.name().fileNameWithoutExtension();
            if (fileName.beginsWith('S', String::CaseInsensitive) && fileName.length() >= 5) {
                if (fileName.endsWith("_START", String::CaseInsensitive)) {
                    Stack_Push(stack, NULL);
                    continue;
                }
                if (Stack_Height(stack) > 0 && fileName.endsWith("_END", String::CaseInsensitive)) {
                    Stack_Pop(stack);
                    continue;
                }
            }
            if (!Stack_Height(stack)) continue;
            String decodedFileName = QString(QByteArray::fromPercentEncoding(fileName.toUtf8()));
            if (!Sprites::isValidSpriteName(decodedFileName)) {
                LOG_RES_NOTE("Ignoring invalid sprite name '%s'") << decodedFileName;
                continue;
            }
            de::Uri const uri("Sprites", Path(fileName));
            Texture::Flags flags = 0;
            if (file.container().hasCustom()) {
                flags |= Texture::Custom;
            }
            Vector2ui dimensions;
            Vector2i origin;
            if (file.size()) {
                ByteRefArray const fileData(file.cache(), file.size());
                if (Patch::recognize(fileData)) {
                    try {
                        auto info = Patch::loadMetadata(fileData);
                        dimensions = info.logicalDimensions;
                        origin     = -info.origin;
                    }
                    catch (IByteArray::OffsetError const &) {
                        LOG_RES_WARNING("File \"%s:%s\" does not appear to be a valid Patch. "
                                        "World dimension and origin offset not set for sprite \"%s\".")
                                << NativePath(file.container().composePath()).pretty()
                                << NativePath(file.composePath()).pretty()
                                << uri;
                    }
                }
                file.unlock();
            }
            de::Uri const resourceUri = LumpIndex::composeResourceUrn(i);
            try {
                self().declareTexture(uri, flags, dimensions, origin, uniqueId, &resourceUri);
                uniqueId++;
            }
            catch (TextureScheme::InvalidPathError const &er) {
                LOG_RES_WARNING("Failed declaring texture \"%s\": %s") << uri << er.asText();
            }
        }
        while (Stack_Height(stack)) {
            Stack_Pop(stack);
        }
        Stack_Delete(stack);
        self().deriveAllTexturesInScheme("Sprites");
        LOG_RES_VERBOSE("Sprite textures initialized in %.2f seconds") << begunAt.since();
    }
};
Textures::Textures()
    : d(new Impl(this))
{}
void Textures::setCompositeArchiveFormat(Composite::ArchiveFormat format) {
    d->compositeFormat = format;
}
void Textures::clear() {
    d->clear();
}
void Textures::clearRuntimeTextures() {
    textureScheme("Sprites").clear();
    textureScheme("Textures").clear();
    textureScheme("Flats").clear();
    textureScheme("Patches").clear();
    textureScheme("Details").clear();
    textureScheme("Reflections").clear();
    textureScheme("Masks").clear();
    textureScheme("ModelSkins").clear();
    textureScheme("ModelReflectionSkins").clear();
    textureScheme("Lightmaps").clear();
    textureScheme("Flaremaps").clear();
}
void Textures::initTextures() {
    LOG_AS("Textures");
    d->initCompositeTextures();
    d->initFlatTextures();
    d->initSpriteTextures();
}
Textures &Textures::get() // static
{
    return Resources::get().textures();
}
TextureScheme &Textures::textureScheme(String const &name) const {
    if (auto *ts = textureSchemePtr(name)) {
        return *ts;
    }
    throw Resources::UnknownSchemeError("Textures::textureScheme", "No scheme found matching '" + name + "'");
}
TextureScheme *Textures::textureSchemePtr(String const &name) const {
    if (!name.isEmpty()) {
        auto found = d->textureSchemes.constFind(name);
        if (found != d->textureSchemes.constEnd()) {
            return *found;
        }
    }
    return nullptr;
}
bool Textures::isKnownTextureScheme(String const &name) const {
    if (!name.isEmpty()) {
        return d->textureSchemes.contains(name);
    }
    return false;
}
Textures::TextureSchemes const& Textures::allTextureSchemes() const {
    return d->textureSchemes;
}
void Textures::clearAllTextureSchemes() {
    d->clearAllTextureSchemes();
}
#if 0
bool Textures::hasTextureManifest(de::Uri const &path) const {
    try {
        textureManifest(path);
        return true;
    }
    catch (Resources::MissingResourceManifestError const &)
    {}  // Ignore this error.
    return false;
}
#endif
TextureManifest &Textures::textureManifest(de::Uri const &uri) const {
    if (auto *mft = textureManifestPtr(uri)) {
        return *mft;
    }
    throw Resources::MissingResourceManifestError
            ("Textures::textureManifest",
             "Failed to locate a manifest matching \"" + uri.asText() + "\"");
}
TextureManifest *Textures::textureManifestPtr(de::Uri const &uri) const {
    if (!uri.scheme().compareWithoutCase("urn")) {
        String const &pathStr = uri.path().toStringRef();
        dint uIdPos = pathStr.indexOf(':');
        if (uIdPos > 0) {
            String schemeName = pathStr.left(uIdPos);
            dint uniqueId     = pathStr.mid(uIdPos + 1 /*skip delimiter*/).toInt();
            if (auto *ts = textureSchemePtr(schemeName)) {
                return ts->tryFindByUniqueId(uniqueId);
            }
        }
    }
    else {
        if (!uri.scheme().isEmpty()) {
            if (auto *ts = textureSchemePtr(uri.scheme())) {
                return ts->tryFind(uri.path());
            }
        }
        else {
            for (TextureScheme const *scheme : d->textureSchemeCreationOrder) {
                if (auto *tex = scheme->tryFind(uri.path())) {
                    return tex;
                }
            }
        }
    }
    return nullptr;
}
Textures::AllTextures const &Textures::allTextures() const {
    return d->textures;
}
TextureManifest &Textures::declareSystemTexture(Path const &texturePath, de::Uri const &resourceUri) {
    auto &scheme = textureScheme(QStringLiteral("System"));
    dint const uniqueId = scheme.count() + 1;
    return scheme.declare(texturePath,
                          Texture::Custom,
                          Vector2ui(),
                          Vector2i(),
                          uniqueId,
                          &resourceUri);
}
Texture *Textures::tryFindTextureByResourceUri(String const &schemeName, de::Uri const &resourceUri) {
    if (!resourceUri.isEmpty()) {
        if (resourceUri.path().toStringRef() == QStringLiteral("-")) {
            return nullptr;
        }
        if (auto *ts = textureSchemePtr(schemeName)) {
            if (auto *mft = ts->tryFindByResourceUri(resourceUri)) {
                return mft->texturePtr();
            }
        }
    }
    return nullptr;
}
Texture *Textures::defineTexture(String const &schemeName,
                                 de::Uri const &resourceUri,
                                 Vector2ui const &dimensions) {
    LOG_AS("Textures::defineTexture");
    if (resourceUri.isEmpty()) return nullptr;
    TextureScheme &scheme = textureScheme(schemeName);
    if (auto *mft = scheme.tryFindByResourceUri(resourceUri)) {
        return mft->texturePtr();
    }
    dint uniqueId = scheme.count() + 1; // 1-based index.
    if (M_NumDigits(uniqueId) > 8) {
        LOG_RES_WARNING("Failed declaring texture manifest in scheme %s (max:%i)")
            << schemeName << DDMAXINT;
        return nullptr;
    }
    de::Uri uri(scheme.name(), Path(String("%1").arg(uniqueId, 8, 10, QChar('0'))));
    try {
        TextureManifest &manifest = declareTexture(uri, Texture::Custom, dimensions,
                                                   Vector2i(), uniqueId, &resourceUri);
        return deriveTexture(manifest);
    }
    catch (TextureScheme::InvalidPathError const &er) {
        LOG_RES_WARNING("Failed declaring texture \"%s\": %s") << uri << er.asText();
    }
    return nullptr;
}
Texture *Textures::deriveTexture(TextureManifest &manifest) {
    LOG_AS("Textures");
    Texture *tex = manifest.derive();
    if (!tex) {
        LOGDEV_RES_WARNING("Failed to derive a Texture for \"%s\", ignoring")
                << manifest.composeUri();
    }
    return tex;
}
void Textures::deriveAllTexturesInScheme(String schemeName) {
    TextureScheme &scheme = textureScheme(schemeName);
    PathTreeIterator<res::TextureScheme::Index> iter(scheme.index().leafNodes());
    while (iter.hasNext()) {
        res::TextureManifest &manifest = iter.next();
        deriveTexture(manifest);
    }
}
patchid_t Textures::declarePatch(String const &encodedName) {
    LOG_AS("Textures::declarePatch");
    if (encodedName.isEmpty())
        return 0;
    de::Uri uri(QStringLiteral("Patches"), Path(encodedName));
    if (auto *mft = textureManifestPtr(uri)) {
        return patchid_t(mft->uniqueId());
    }
    auto &fs1 = App_FileSystem();
    Path lumpPath = uri.path() + ".lmp";
    if (!fs1.nameIndex().contains(lumpPath)) {
        LOG_RES_WARNING("Failed to locate lump for \"%s\"") << uri;
        return 0;
    }
    lumpnum_t const lumpNum = fs1.nameIndex().findLast(lumpPath);
    File1 &file = fs1.lump(lumpNum);
    Texture::Flags flags;
    if (file.container().hasCustom()) flags |= Texture::Custom;
    Vector2ui dimensions;
    Vector2i origin;
    ByteRefArray fileData = ByteRefArray(file.cache(), file.size());
    if (Patch::recognize(fileData)) {
        try {
            auto info = Patch::loadMetadata(fileData);
            dimensions = info.logicalDimensions;
            origin     = Vector2i(-info.origin.x, -info.origin.y);
        }
        catch (IByteArray::OffsetError const &) {
            LOG_RES_WARNING("File \"%s:%s\" does not appear to be a valid Patch. "
                            "World dimension and origin offset not set for patch \"%s\".")
                << NativePath(file.container().composePath()).pretty()
                << NativePath(file.composePath()).pretty()
                << uri;
        }
    }
    file.unlock();
    dint uniqueId       = textureScheme(QStringLiteral("Patches")).count() + 1;  // 1-based index.
    de::Uri resourceUri = LumpIndex::composeResourceUrn(lumpNum);
    try {
        TextureManifest &manifest = declareTexture(uri, flags, dimensions, origin,
                                                   uniqueId, &resourceUri);
        deriveTexture(manifest);
        return uniqueId;
    }
    catch (TextureScheme::InvalidPathError const &er) {
        LOG_RES_WARNING("Failed declaring texture \"%s\": %s") << uri << er.asText();
    }
    return 0;
}
} // namespace res
namespace res {
DENG2_PIMPL(TextureScheme)
, DENG2_OBSERVES(TextureManifest, UniqueIdChange)
, DENG2_OBSERVES(TextureManifest, Deletion) {
    String name;
    TextureScheme::Index index;
    QList<TextureManifest *> uniqueIdLut;
    bool uniqueIdLutDirty;
    int uniqueIdBase;
    Impl(Public *i, String symbolicName) : Base(i),
        name(symbolicName),
        uniqueIdLut(),
        uniqueIdLutDirty(false),
        uniqueIdBase(0)
    {}
    ~Impl() {
        self().clear();
        DENG_ASSERT(index.isEmpty());
    }
    bool inline uniqueIdInLutRange(int uniqueId) const {
        return (uniqueId - uniqueIdBase >= 0 && (uniqueId - uniqueIdBase) < uniqueIdLut.size());
    }
    void findUniqueIdRange(int *minId, int *maxId) {
        if (!minId && !maxId) return;
        if (minId) *minId = DDMAXINT;
        if (maxId) *maxId = DDMININT;
        PathTreeIterator<Index> iter(index.leafNodes());
        while (iter.hasNext()) {
            TextureManifest &manifest = iter.next();
            int const uniqueId = manifest.uniqueId();
            if (minId && uniqueId < *minId) *minId = uniqueId;
            if (maxId && uniqueId > *maxId) *maxId = uniqueId;
        }
    }
    void deindex(TextureManifest &manifest) {
        manifest.clearTexture();
        unlinkInUniqueIdLut(manifest);
    }
    void unlinkInUniqueIdLut(TextureManifest &manifest) {
        if (!uniqueIdLutDirty) {
            int uniqueId = manifest.uniqueId();
            DENG_ASSERT(uniqueIdInLutRange(uniqueId));
            uniqueIdLut[uniqueId - uniqueIdBase] = 0;
        }
    }
    void linkInUniqueIdLut(TextureManifest &manifest) {
        int uniqueId = manifest.uniqueId();
        DENG_ASSERT(uniqueIdInLutRange(uniqueId));
        uniqueIdLut[uniqueId - uniqueIdBase] = &manifest;
    }
    void rebuildUniqueIdLut() {
        if (!uniqueIdLutDirty) return;
        int minId, maxId;
        findUniqueIdRange(&minId, &maxId);
        int lutSize = 0;
        if (minId > maxId) // None found?
        {
            uniqueIdBase = 0;
        }
        else {
            uniqueIdBase = minId;
            lutSize = maxId - minId + 1;
        }
        uniqueIdLut.reserve(lutSize);
        int i = 0;
        for (; i < uniqueIdLut.size(); ++i) {
            uniqueIdLut[i] = 0;
        }
        for (; i < lutSize; ++i) {
            uniqueIdLut.push_back(0);
        }
        if (lutSize) {
            PathTreeIterator<Index> iter(index.leafNodes());
            while (iter.hasNext()) {
                linkInUniqueIdLut(iter.next());
            }
        }
        uniqueIdLutDirty = false;
    }
    void textureManifestUniqueIdChanged(Manifest & /*manifest*/)
    {
        uniqueIdLutDirty = true;
    }
    void textureManifestBeingDeleted(TextureManifest const &manifest) {
        deindex(const_cast<TextureManifest &>(manifest));
    }
};
TextureScheme::TextureScheme(String symbolicName) : d(new Impl(this, symbolicName))
{}
TextureScheme::~TextureScheme() {
    clear();
}
void TextureScheme::clear() {
    /*PathTreeIterator<Index> iter(d->index.leafNodes());
    while (iter.hasNext()) {
        d->deindex(iter.next());
    }*/
    d->index.clear();
    d->uniqueIdLutDirty = true;
}
String const &TextureScheme::name() const {
    return d->name;
}
TextureManifest &TextureScheme::declare(Path const &path,
    Texture::Flags flags, Vector2ui const &dimensions, Vector2i const &origin,
    int uniqueId, de::Uri const *resourceUri) {
    LOG_AS("TextureScheme::declare");
    if (path.isEmpty()) {
        throw InvalidPathError("TextureScheme::declare", "Missing/zero-length path was supplied");
    }
    int const sizeBefore = d->index.size();
    Manifest *newManifest = &d->index.insert(path);
    DENG2_ASSERT(newManifest);
    if (d->index.size() != sizeBefore) {
        d->uniqueIdLutDirty = true;
        newManifest->audienceForUniqueIdChange += d;
        newManifest->audienceForDeletion += d;
        DENG2_FOR_AUDIENCE(ManifestDefined, i) i->textureSchemeManifestDefined(*this, *newManifest);
    }
    /*
     * (Re)configure the manifest.
     */
    bool mustRelease = false;
    newManifest->setScheme(*this);
    newManifest->setFlags(flags);
    newManifest->setOrigin(origin);
    if (newManifest->setLogicalDimensions(dimensions)) {
        mustRelease = true;
    }
    if (newManifest->setUniqueId(uniqueId)) {
        mustRelease = true;
    }
    if (resourceUri && newManifest->setResourceUri(*resourceUri)) {
        mustRelease = true;
    }
    if (mustRelease && newManifest->hasTexture()) {
        newManifest->texture().release();
    }
    return *newManifest;
}
#if 0
bool TextureScheme::has(Path const &path) const {
    return d->index.has(path, Index::NoBranch | Index::MatchFull);
}
#endif
TextureManifest const &TextureScheme::find(Path const &path) const {
    if (auto *mft = tryFind(path)) {
        return *mft;
    }
    throw NotFoundError("TextureScheme::find", "Failed to locate a manifest matching \"" +
                        path.asText() + "\"");
}
TextureManifest &TextureScheme::find(Path const &path) {
    TextureManifest const &found = const_cast<TextureScheme const *>(this)->find(path);
    return const_cast<TextureManifest &>(found);
}
TextureManifest *TextureScheme::tryFind(Path const &path) const {
    return d->index.tryFind(path, Index::NoBranch | Index::MatchFull);
}
TextureManifest const &TextureScheme::findByResourceUri(de::Uri const &uri) const {
    if (auto *mft = tryFindByResourceUri(uri)) {
        return *mft;
    }
    throw NotFoundError("TextureScheme::findByResourceUri",
                        "No manifest found with a resource URI matching \"" + uri.asText() + "\"");
}
TextureManifest *TextureScheme::tryFindByResourceUri(de::Uri const &uri) const {
    if (!uri.isEmpty()) {
        PathTreeIterator<Index> iter(d->index.leafNodes());
        while (iter.hasNext()) {
            TextureManifest &manifest = iter.next();
            if (manifest.hasResourceUri()) {
                if (manifest.resourceUri() == uri) {
                    return &manifest;
                }
            }
        }
    }
    return nullptr;
}
TextureManifest &TextureScheme::findByResourceUri(de::Uri const &uri) {
    TextureManifest const &found = const_cast<TextureScheme const *>(this)->findByResourceUri(uri);
    return const_cast<TextureManifest &>(found);
}
TextureManifest const &TextureScheme::findByUniqueId(int uniqueId) const {
    if (auto *mft = tryFindByUniqueId(uniqueId)) {
        return *mft;
    }
    throw NotFoundError("TextureScheme::findByUniqueId",
                        "No manifest found with a unique ID matching \"" + QString("%1").arg(uniqueId) + "\"");
}
TextureManifest &TextureScheme::findByUniqueId(int uniqueId) {
    TextureManifest const &found = const_cast<TextureScheme const *>(this)->findByUniqueId(uniqueId);
    return const_cast<TextureManifest &>(found);
}
TextureManifest *TextureScheme::tryFindByUniqueId(int uniqueId) const {
    d->rebuildUniqueIdLut();
    if (d->uniqueIdInLutRange(uniqueId)) {
        TextureManifest *manifest = d->uniqueIdLut[uniqueId - d->uniqueIdBase];
        if (manifest) return manifest;
    }
    return nullptr;
}
TextureScheme::Index const &TextureScheme::index() const {
    return d->index;
}
} // namespace res
typedef struct riff_hdr_s {
    char            id[4];              ///< Identifier string = "RIFF"
    uint32_t        len;                ///< Remaining length after this header
} riff_hdr_t;
typedef struct chunk_hdr_s {            ///< CHUNK 8-byte header
    char            id[4];              ///< Identifier, e.g. "fmt " or "data"
    uint32_t        len;                ///< Remaining chunk length after header
} chunk_hdr_t;                          ///< data bytes follow chunk header
typedef struct wav_format_s {
    uint16_t        wFormatTag;         ///< Format category
    uint16_t        wChannels;          ///< Number of channels
    uint32_t        dwSamplesPerSec;    ///< Sampling rate
    uint32_t        dwAvgBytesPerSec;   ///< For buffer estimation
    uint16_t        wBlockAlign;        ///< Data block size
    uint16_t        wBitsPerSample;     ///< Sample size
} wav_format_t;
#pragma pack()
#define WReadAndAdvance(pByte, pDest, length) \
    { memcpy(pDest, pByte, length); pByte += length; }
int WAV_CheckFormat(const char* data) {
    return !strncmp(data, "RIFF", 4) && !strncmp(data + 8, "WAVE", 4);
}
void* WAV_MemoryLoad(const byte* data, size_t datalength, int* bits, int* rate, int* samples) {
    const byte* end = data + datalength;
    byte* sampledata = NULL;
    chunk_hdr_t riff_chunk;
    wav_format_t wave_format;
    LOG_AS("WAV_MemoryLoad");
    if (!WAV_CheckFormat((const char*)data)) {
        LOG_RES_WARNING("Not WAV format data");
        return NULL;
    }
    data += sizeof(riff_hdr_t);
    data += 4; // "WAVE" already verified above
#ifdef _DEBUG
    assert(sizeof(wave_format) == 16);
    assert(sizeof(riff_chunk) == 8);
#endif
    memset(&wave_format, 0, sizeof(wave_format));
    wave_format.wBlockAlign = 1;
    while (data < end) {
        WReadAndAdvance(data, &riff_chunk, sizeof(riff_chunk));
        riff_chunk.len = DD_ULONG(riff_chunk.len);
        if (!strncmp(riff_chunk.id, "fmt ", 4)) {
            WReadAndAdvance(data, &wave_format, sizeof(wave_format));
            wave_format.wFormatTag       = DD_USHORT(wave_format.wFormatTag      );
            wave_format.wChannels        = DD_USHORT(wave_format.wChannels       );
            wave_format.dwSamplesPerSec  = DD_ULONG (wave_format.dwSamplesPerSec );
            wave_format.dwAvgBytesPerSec = DD_ULONG (wave_format.dwAvgBytesPerSec);
            wave_format.wBlockAlign      = DD_USHORT(wave_format.wBlockAlign     );
            wave_format.wBitsPerSample   = DD_USHORT(wave_format.wBitsPerSample  );
            assert(wave_format.wFormatTag == WAVE_FORMAT_PCM); // linear PCM
            if (wave_format.wFormatTag != WAVE_FORMAT_PCM) {
                LOG_RES_WARNING("Unsupported format (%i)") << wave_format.wFormatTag;
                return NULL;
            }
            if (wave_format.wChannels != 1) {
                LOG_RES_WARNING("Too many channels (only mono supported)");
                return NULL;
            }
            /*if (wave_format->wBitsPerSample == 0)
               {
               *bits = 8*wave_format->dwAvgBytesPerSec/
               wave_format->dwSamplesPerSec;
               }
               else
               { */
            if (wave_format.wBitsPerSample != 8 &&
               wave_format.wBitsPerSample != 16) {
                LOG_RES_WARNING("Must have 8 or 16 bits per sample");
                return NULL;
            }
            *bits = wave_format.wBitsPerSample;
            *rate = wave_format.dwSamplesPerSec;
        }
        else if (!strncmp(riff_chunk.id, "data", 4)) {
            if (!wave_format.wFormatTag) {
                LOG_RES_WARNING("Malformed WAV data");
                return NULL;
            }
            *samples = riff_chunk.len / wave_format.wBlockAlign;
            sampledata = (byte *) Z_Malloc(riff_chunk.len, PU_APPSTATIC, 0);
            memcpy(sampledata, data, riff_chunk.len);
#ifdef __BIG_ENDIAN__
            /*if (wave_format->wBitsPerSample == 16)
            {
                ushort* sample = sampledata;
                for (; sample < ((short*)sampledata) + *samples; ++sample)
                    *sample = DD_USHORT(*sample);
            }*/
#endif
            break;
        }
        else {
            data += riff_chunk.len;
        }
    }
    return sampledata;
}
void *WAV_Load(char const *filename, int *bits, int *rate, int *samples) {
    try {
        de::String path = (de::NativePath::workPath() / de::NativePath(filename).expand()).withSeparators('/');
        QScopedPointer<de::FileHandle> hndl(&App_FileSystem().openFile(path, "rb"));
        size_t size = hndl->length();
        LOG_AS("WAV_Load");
        LOGDEV_RES_XVERBOSE("Loading from \"%s\" (size %i, fpos %i)",
                               de::NativePath(hndl->file().composePath()).pretty()
                            << size
                            << hndl->tell());
        uint8_t *data = (uint8_t *) M_Malloc(size);
        hndl->read(data, size);
        App_FileSystem().releaseFile(hndl->file());
        void *sampledata = WAV_MemoryLoad((byte const *) data, size, bits, rate, samples);
        if (!sampledata) {
            LOG_RES_WARNING("Failed to load \"%s\"") << filename;
        }
        M_Free(data);
        return sampledata;
    }
    catch (de::FS1::NotFoundError const &)
    {} // Ignore.
    return 0;
}
