﻿/** @file armorwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/armorwidget.h"
#include "common.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "p_actor.h"
#include "player.h"
/** @file automapwidget.cpp  GUI widget for the automap.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "hud/widgets/automapwidget.h"
#include <QList>
#include <QtAlgorithms>
#include <de/LogBuffer>
#include <de/ScriptSystem>
#include <de/Vector>
#include "dmu_lib.h"
#include "g_common.h"
#include "gamesession.h"
#include "hu_stuff.h"
#include "hud/automapstyle.h"
#include "p_mapsetup.h"
#include "p_tick.h"
#include "r_common.h"
#if __JDOOM64__
#  include "p_inventory.h"
#endif
#define UIAUTOMAP_BORDER        4  ///< In fixed 320x200 pixels.
/** @file chainwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/chainwidget.h"
#include "common.h"
#include "hu_lib.h"
#include "p_actor.h"
#include "p_tick.h"
/** @file chatwidget.cpp  Specialized HudWidget for player messaging.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "hud/widgets/chatwidget.h"
#include <de/Log>
#include "d_net.h"
#include "hu_stuff.h"  // shiftXForm
#include "p_tick.h"
/** @file flightwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/flightwidget.h"
#include "common.h"
#include "gl_drawpatch.h"
#include "p_tick.h"
/** @file fragswidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/fragswidget.h"
#include "g_common.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "player.h"
#include "p_actor.h"
/** @file groupwidget.cpp  GUI widget for grouping widgets.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/groupwidget.h"
#include <QVector>
#include "common.h"
/** @file healthwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/healthwidget.h"
#include "g_common.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "player.h"
#include "p_actor.h"
/** @file itemswidget.h  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/itemswidget.h"
#include "common.h"
#include "p_actor.h"
/** @file keyslotwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/keyslotwidget.h"
#include <de/Vector>
#include "common.h"
#include "gl_drawpatch.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "p_actor.h"
/** @file keyswidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/keyswidget.h"
#include "gl_drawpatch.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "player.h"
#if __JHERETIC__ || __JHEXEN__
static patchid_t pKeys[NUM_KEY_TYPES];
#endif
/** @file killswidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/killswidget.h"
#include "common.h"
#include "p_actor.h"
/** @file playerlogwidget.cpp  Specialized HudWidget for logging player messages.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "hud/widgets/playerlogwidget.h"
#include <cstring>
#include <cstdio>
#include "hu_stuff.h"
#include "p_tick.h"
#include "player.h"    ///< @ref LMF_NO_HIDE @todo remove me
/** @file readyammoiconwidget.h  GUI widget for visualizing high-level player status.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/readyammoiconwidget.h"
#include "common.h"
#include "gl_drawpatch.h"
#if __JHERETIC__
#  include "hu_inventory.h"
#endif
#include "p_actor.h"
/** @file readyammowidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "hud/widgets/readyammowidget.h"
#include "player.h"
#include "p_actor.h"
#if __JHERETIC__
#  include "hu_inventory.h"
#endif
/** @file readyitemwidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/readyitemwidget.h"
#include "common.h"
#include "gl_drawpatch.h"
#include "hu_inventory.h"
#include "p_inventory.h"
/** @file secretswidget.cpp  GUI widget for -.
 *
 * @authors Copyright © 2005-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "hud/widgets/secretswidget.h"
#include "common.h"
#include "p_actor.h"
using namespace de;
guidata_armor_t::guidata_armor_t(void (*updateGeometry) (HudWidget *wi),
                                 void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                 dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_armor_t::~guidata_armor_t()
{}
void guidata_armor_t::reset() {
    _value = 1994;
}
void guidata_armor_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
#if __JHEXEN__
    player_t const *plr = &::players[player()];
    dint const pClass   = ::cfg.playerClass[player()];  // Original player class (i.e. not pig).
    _value = FixedDiv(PCLASS_INFO(pClass)->autoArmorSave
                      + plr->armorPoints[ARMOR_ARMOR ]
                      + plr->armorPoints[ARMOR_SHIELD]
                      + plr->armorPoints[ARMOR_HELMET]
                      + plr->armorPoints[ARMOR_AMULET], 5 * FRACUNIT) >> FRACBITS;
#else
    player_t const *plr = &::players[player()];
    _value = plr->armorPoints;
#endif
}
void ArmorWidget_Draw(guidata_armor_t *armor, Point2Raw const *offset) {
#if __JDOOM__
#  define SUFFIX                ( "%" )
#  define TRACKING              ( 0 )
#elif __JHERETIC__
#  define SUFFIX                ( "" )
#  define TRACKING              ( 1 )
#else
#  define SUFFIX                ( "" )
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(armor);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(armor->_value == 1994) return;
#if !__JHEXEN__
    if(!::cfg.hudShown[HUD_ARMOR]) return;
#endif
    if(ST_AutomapIsOpen(armor->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[armor->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(armor->_value) + SUFFIX;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(armor->font());
    FR_SetTracking(TRACKING);
#if __JHERETIC__
    FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, -2); /// @todo Why is an offset needed?
#else
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, 0);
#endif
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
#undef SUFFIX
}
void SBarArmorWidget_Draw(guidata_armor_t *armor, Point2Raw const *offset) {
#if __JDOOM__
#  define X_OFFSET              ( 221 )
#  define Y_OFFSET              (   3 )
#  define TRACKING              ( 0 )
#elif __JHERETIC__
#  define X_OFFSET              ( 254 )
#  define Y_OFFSET              (  12 )
#  define TRACKING              ( 1 )
#elif __JHEXEN__
#  define X_OFFSET              ( 274 )
#  define Y_OFFSET              (  14 )
#  define TRACKING              ( 0 )
#else
#  define X_OFFSET              ( 0 )
#  define Y_OFFSET              ( 0 )
#  define TRACKING              ( 0 )
#endif
    static Vector2i const origin(-ST_WIDTH / 2, -ST_HEIGHT);
    dint const activeHud     = ST_ActiveHud(armor->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(armor->player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(armor->_value == 1994) return;
#if __JHERETIC__ || __JHEXEN__
    if(Hu_InventoryIsOpen(armor->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(armor->player())) return;
#else
    if(ST_AutomapIsOpen(armor->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[armor->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(armor->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(armor->font());
    FR_SetTracking(TRACKING);
#if __JDOOM__
    if(::gameMode == doom_chex) {
        FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    }
    else
#endif
    {
        FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    }
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), origin.x + X_OFFSET, origin.y + Y_OFFSET, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
#if __JDOOM__
    FR_DrawCharXY('%', origin.x + X_OFFSET, origin.y + Y_OFFSET);
#endif
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
#undef Y_OFFSET
#undef X_OFFSET
}
void Armor_UpdateGeometry(guidata_armor_t *armor) {
#if __JHERETIC__
#  define TRACKING              ( 1 )
#else
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(armor);
    Rect_SetWidthHeight(&armor->geometry(), 0, 0);
    if(armor->_value == 1994) return;
#if !__JHEXEN__
    if(!::cfg.hudShown[HUD_ARMOR]) return;
#endif
    if(ST_AutomapIsOpen(armor->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[armor->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(armor->_value) + "%";
    FR_SetFont(armor->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    Rect_SetWidthHeight(&armor->geometry(), textSize.width  * ::cfg.common.hudScale,
                                            textSize.height * ::cfg.common.hudScale);
#undef TRACKING
}
void SBarArmor_UpdateGeometry(guidata_armor_t *armor) {
#if __JHERETIC__
#  define TRACKING              ( 1 )
#else
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(armor);
    Rect_SetWidthHeight(&armor->geometry(), 0, 0);
    if(armor->_value == 1994) return;
#if __JHERETIC__ || __JHEXEN__
    if(Hu_InventoryIsOpen(armor->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(armor->player())) return;
#else
    if(ST_AutomapIsOpen(armor->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[armor->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(armor->_value);
    FR_SetFont(armor->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
#if __JDOOM__
    Rect_SetWidthHeight(&armor->geometry(), (textSize.width + FR_CharWidth('%')) * ::cfg.common.statusbarScale,
                                             de::max(textSize.height, FR_CharHeight('%')) * ::cfg.common.statusbarScale);
#else
    Rect_SetWidthHeight(&armor->geometry(), textSize.width  * ::cfg.common.statusbarScale,
                                            textSize.height * ::cfg.common.statusbarScale);
#endif
#undef TRACKING
}
static void AutomapWidget_UpdateGeometry(AutomapWidget *amap) {
    DENG2_ASSERT(amap);
    amap->updateGeometry();
}
static void AutomapWidget_Draw(AutomapWidget *amap, Point2Raw const *offset) {
    DENG2_ASSERT(amap);
    amap->draw(offset? Vector2i(offset->xy) : Vector2i());
}
struct uiautomap_rendstate_t {
    player_t *plr;
    dint obType;         ///< The type of object to draw. @c -1= only line specials.
    dd_bool glowOnly;
    dglprimtype_t primType;
};
static uiautomap_rendstate_t rs;
static dbyte freezeMapRLs;
#if __JDOOM__ || __JDOOM64__
static dint autopageLumpNum = -1;
#elif __JHERETIC__
static dint autopageLumpNum = 1;
#else
static dint autopageLumpNum = 1;
#endif
static DGLuint amMaskTexture;  // Used to mask the map primitives.
namespace internal {
    static Vector2d rotate(Vector2d const &point, ddouble radian) {
        ddouble const c = std::cos(radian);
        ddouble const s = std::sin(radian);
        return Vector2d(c * point.x - s * point.y, s * point.x + c * point.y);
    }
    static void initAABB(coord_t aabb[4], Vector2d const &point) {
        DENG2_ASSERT(aabb);
        aabb[BOXLEFT] = aabb[BOXRIGHT ] = point.x;
        aabb[BOXTOP ] = aabb[BOXBOTTOM] = point.y;
    }
    static void addToAABB(coord_t aabb[4], Vector2d const &point) {
        DENG2_ASSERT(aabb);
        if     (point.x < aabb[BOXLEFT  ]) aabb[BOXLEFT  ] = point.x;
        else if (point.x > aabb[BOXRIGHT ]) aabb[BOXRIGHT ] = point.x;
        if     (point.y < aabb[BOXBOTTOM]) aabb[BOXBOTTOM] = point.y;
        else if (point.y > aabb[BOXTOP   ]) aabb[BOXTOP   ] = point.y;
    }
    static dd_bool interceptEdge(coord_t point[2], coord_t const startA[2],
        coord_t const endA[2], coord_t const startB[2], coord_t const endB[2]) {
        coord_t directionA[2];
        V2d_Subtract(directionA, endA, startA);
        if (V2d_PointOnLineSide(point, startA, directionA) >= 0) {
            coord_t directionB[2];
            V2d_Subtract(directionB, endB, startB);
            V2d_Intersection(startA, directionA, startB, directionB, point);
            return true;
        }
        return false;
    }
    static Vector2d fitPointInRectangle(Vector2d const &point, Vector2d const &topLeft,
        Vector2d const &topRight, Vector2d const &bottomRight,
        Vector2d const &bottomLeft, Vector2d const &viewPoint) {
        ddouble pointV1[2];
        point.decompose(pointV1);
        ddouble topLeftV1[2];
        topLeft.decompose(topLeftV1);
        ddouble topRightV1[2];
        topRight.decompose(topRightV1);
        ddouble bottomRightV1[2];
        bottomRight.decompose(bottomRightV1);
        ddouble bottomLeftV1[2];
        bottomLeft.decompose(bottomLeftV1);
        ddouble viewPointV1[2];
        viewPoint.decompose(viewPointV1);
        if (!interceptEdge(pointV1, topLeftV1, bottomLeftV1, viewPointV1, pointV1))
            interceptEdge(pointV1, bottomRightV1, topRightV1, viewPointV1, pointV1);
        if (!interceptEdge(pointV1, topRightV1, topLeftV1, viewPointV1, pointV1))
            interceptEdge(pointV1, bottomLeftV1, bottomRightV1, viewPointV1, pointV1);
        return Vector2d(pointV1);
    }
    static void drawVectorGraphic(svgid_t vgId, Vector2d const &origin, dfloat angle,
        dfloat scale, Vector3f const &color, dfloat opacity, blendmode_t blendmode) {
        opacity = de::clamp(0.f, opacity, 1.f);
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_PushMatrix();
        DGL_Translatef(origin.x, origin.y, 1);
        DGL_Color4f(color.x, color.y, color.z, opacity);
        DGL_BlendMode(blendmode);
        Point2Rawf originp(origin.x, origin.y);
        GL_DrawSvg3(vgId, &originp, scale, angle);
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_PopMatrix();
    }
    static dint playerPaletteColor(dint consoleNum) {
        static dint playerColors[NUMPLAYERCOLORS] = {
            AM_PLR1_COLOR, AM_PLR2_COLOR, AM_PLR3_COLOR, AM_PLR4_COLOR,
#if __JHEXEN__
            AM_PLR5_COLOR, AM_PLR6_COLOR, AM_PLR7_COLOR, AM_PLR8_COLOR
#endif
        };
        if (!IS_NETGAME) return WHITE;
        return playerColors[cfg.playerColor[de::max(0, consoleNum) % MAXPLAYERS]];
    }
    static void drawPlayerMarker(dint consoleNum, AutomapStyle *style) {
        DENG2_ASSERT(consoleNum >= 0 && consoleNum < MAXPLAYERS);
        player_t *player = &players[consoleNum];
        if (!player->plr->inGame) return;
        mobj_t *plrMob = player->plr->mo;
        if (!plrMob) return;
        coord_t origin[3]; Mobj_OriginSmoothed(plrMob, origin);
        dfloat const angle = Mobj_AngleSmoothed(plrMob) / (dfloat) ANGLE_MAX * 360; /* $unifiedangles */
        dfloat color[3]; R_GetColorPaletteRGBf(0, playerPaletteColor(consoleNum), color, false);
        dfloat opacity = cfg.common.automapLineAlpha * uiRendState->pageAlpha;
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        if (player->powers[PT_INVISIBILITY])
            opacity *= .125f;
#endif
        drawVectorGraphic(style->objectSvg(AMO_THINGPLAYER), Vector2d(origin),
                          angle, PLAYERRADIUS, Vector3f(color), opacity, BM_NORMAL);
    }
}  // namespace internal
using namespace ::internal;
DENG2_PIMPL(AutomapWidget) {
    AutomapStyle *style = nullptr;
    float pixelRatio = 1.f; // DisplayMode.PIXEL_RATIO
    bool needBuildLists = false;         ///< @c true= force a rebuild of all lists.
    dint flags = 0;
    bool open     = false;       ///< @c true= currently active.
    bool revealed = false;
    bool follow   = true;        ///< @c true= camera position tracks followed player.
    bool rotate   = false;
    bool forceMaxScale = false;  ///< If the map is currently in forced max zoom mode.
    dfloat priorToMaxScale = 0;  ///< Viewer scale before entering maxScale mode.
    dfloat minScale  = 1.f;
    dfloat scaleMTOF = 1.f;      ///< Used by MTOF to scale from map-to-frame-buffer coords.
    dfloat scaleFTOM = 1.f;      ///< Used by FTOM to scale from frame-buffer-to-map coords (=1/scaleMTOF).
    coord_t bounds[4];           ///< Map space bounds:
    dfloat opacity = 0, targetOpacity = 0, oldOpacity = 0;
    dfloat opacityTimer = 0;
    Vector2d view, targetView, oldView;
    dfloat viewTimer = 0;
    coord_t maxViewPositionDelta = 128;
    dfloat viewScale = 1, targetViewScale = 1, oldViewScale = 1;
    dfloat viewScaleTimer = 0;
    bool needViewScaleUpdate = false;
    dfloat minScaleMTOF = 0;
    dfloat maxScaleMTOF = 0;
    dfloat angle = 0, targetAngle = 0, oldAngle = 0;
    dfloat angleTimer = 0;
    Vector2d topLeft, bottomRight, topRight, bottomLeft;
    coord_t viewAABB[4];
    QList<MarkedPoint *> points;  ///< Player-marked points of interest.
    dint followPlayer = 0;        ///< Player being followed.
    Impl(Public *i) : Base(i) {
        using namespace de;
        zap(bounds);
        zap(viewAABB);
        auto &ds = ScriptSystem::get();
        if (ds.nativeModuleExists("DisplayMode")) {
            pixelRatio = ds["DisplayMode"].getf("PIXEL_RATIO");
        }
    }
    ~Impl() {
        clearPoints();
    }
    void clearPoints() {
        qDeleteAll(points); points.clear();
    }
    void setMinScale(dfloat newMinScale) {
        minScale = de::max(1.f, newMinScale);
        needViewScaleUpdate = true;
    }
    /**
     * Calculate the min/max scaling factors.
     *
     * Take the distance from the bottom left to the top right corners and choose a max scaling
     * factor such that this distance is short than both the automap window width and height.
     */
    void updateViewScale() {
        float const oldMinScale = minScaleMTOF;
        Vector2d const topRight  (bounds[BOXRIGHT], bounds[BOXTOP   ]);
        Vector2d const bottomLeft(bounds[BOXLEFT ], bounds[BOXBOTTOM]);
        coord_t const dist = de::abs((topRight - bottomLeft).length());
        Vector2f const dimensions(Rect_Width (&self().geometry()),
                                  Rect_Height(&self().geometry()));
        Vector2f const scale = dimensions / dist;
        minScaleMTOF = (scale.x < scale.y ? scale.x : scale.y);
        maxScaleMTOF = dimensions.y / minScale;
        LOG_AS("AutomapWidget");
        LOGDEV_XVERBOSE("updateViewScale: delta:%s dimensions:%s dist:%f scale:%s minmtof:%f",
                   (topRight - bottomLeft).asText()
                << dimensions.asText() << dist
                << scale.asText() << minScaleMTOF);
        targetViewScale = viewScale = minScaleMTOF / oldMinScale * targetViewScale;
        needViewScaleUpdate = false;
    }
    static void drawLine2(Vector2d const &from, Vector2d const &to,
        Vector3f const &color, dfloat opacity, glowtype_t glowType, dfloat glowStrength,
        dfloat glowSize, dd_bool glowOnly, dd_bool scaleGlowWithView, dd_bool caps,
        /*blendmode_t blend, */dd_bool drawNormal)
    {
        opacity *= uiRendState->pageAlpha;
        Vector2d const unit = (to - from).normalize();
        Vector2d const normal(unit.y, -unit.x);
        if (de::abs(unit.length()) <= 0) return;
        if (glowOnly && glowType != GLOW_NONE) {
            dfloat thickness;
            if (scaleGlowWithView)
                thickness = cfg.common.automapDoorGlow * 2.5f + 3;
            else
                thickness = glowSize;
            DENG_ASSERT(rs.primType == DGL_QUADS);
            if (caps) {
                Vector2f const v1 = from -   unit * thickness + normal * thickness;
                Vector2f const v2 = from + normal * thickness;
                Vector2f const v3 = from - normal * thickness;
                Vector2f const v4 = from -   unit * thickness - normal * thickness; {
                    DGL_Color4f(color.x, color.y, color.z, glowStrength * opacity);
                }
                DGL_TexCoord2f(0, 0, 0);
                DGL_TexCoord2f(1, v1.x, v1.y);
                DGL_Vertex2f(v1.x, v1.y);
                DGL_TexCoord2f(0, .5f, 0);
                DGL_TexCoord2f(1, v2.x, v2.y);
                DGL_Vertex2f(v2.x, v2.y);
                DGL_TexCoord2f(0, .5f, 1);
                DGL_TexCoord2f(1, v3.x, v3.y);
                DGL_Vertex2f(v3.x, v3.y);
                DGL_TexCoord2f(0, 0, 1);
                DGL_TexCoord2f(1, v4.x, v4.y);
                DGL_Vertex2f(v4.x, v4.y);
            }
            switch (glowType) {
            case GLOW_BOTH: {
                Vector2f const v1 = from + normal * thickness;
                Vector2f const v2 =   to + normal * thickness;
                Vector2f const v3 =   to - normal * thickness;
                Vector2f const v4 = from - normal * thickness; {
                    DGL_Color4f(color.x, color.y, color.z, glowStrength * opacity);
                }
                DGL_TexCoord2f(0, .5f, 0);
                DGL_TexCoord2f(1, v1.x, v1.y);
                DGL_Vertex2f(v1.x, v1.y);
                DGL_TexCoord2f(0, .5f, 0);
                DGL_TexCoord2f(1, v2.x, v2.y);
                DGL_Vertex2f(v2.x, v2.y);
                DGL_TexCoord2f(0, .5f, 1);
                DGL_TexCoord2f(1, v3.x, v3.y);
                DGL_Vertex2f(v3.x, v3.y);
                DGL_TexCoord2f(0, .5f, 1);
                DGL_TexCoord2f(1, v4.x, v4.y);
                DGL_Vertex2f(v4.x, v4.y);
                break; }
            case GLOW_BACK: {
                Vector2f const v1 = from + normal * thickness;
                Vector2f const v2 =   to + normal * thickness; {
                    DGL_Color4f(color.x, color.y, color.z, glowStrength * opacity);
                }
                DGL_TexCoord2f(0, 0, .25f);
                DGL_TexCoord2f(1, v1.x, v1.y);
                DGL_Vertex2f(v1.x, v1.y);
                DGL_TexCoord2f(0, 0, .25f);
                DGL_TexCoord2f(1, v2.x, v2.y);
                DGL_Vertex2f(v2.x, v2.y);
                DGL_TexCoord2f(0, .5f, .25f);
                DGL_TexCoord2f(1, to.x, to.y);
                DGL_Vertex2f(to.x, to.y);
                DGL_TexCoord2f(0, .5f, .25f);
                DGL_TexCoord2f(1, from.x, from.y);
                DGL_Vertex2f(from.x, from.y);
                break; }
            case GLOW_FRONT: {
                Vector2f const v3 =   to - normal * thickness;
                Vector2f const v4 = from - normal * thickness; {
                    DGL_Color4f(color.x, color.y, color.z, glowStrength * opacity);
                }
                DGL_TexCoord2f(0, .75f, .5f);
                DGL_TexCoord2f(1, from.x, from.y);
                DGL_Vertex2f(from.x, from.y);
                DGL_TexCoord2f(0, .75f, .5f);
                DGL_TexCoord2f(1, to.x, to.y);
                DGL_Vertex2f(to.x, to.y);
                DGL_TexCoord2f(0, .75f, 1);
                DGL_TexCoord2f(1, v3.x, v3.y);
                DGL_Vertex2f(v3.x, v3.y);
                DGL_TexCoord2f(0, .75f, 1);
                DGL_TexCoord2f(1, v4.x, v4.y);
                DGL_Vertex2f(v4.x, v4.y);
                break; }
            default:
                DENG2_ASSERT(!"Unknown glowtype");
                break;
            }
            if (caps) {
                Vector2f const v1 = to + normal * thickness;
                Vector2f const v2 = to +   unit * thickness + normal * thickness;
                Vector2f const v3 = to +   unit * thickness - normal * thickness;
                Vector2f const v4 = to - normal * thickness; {
                    DGL_Color4f(color.x, color.y, color.z, glowStrength * opacity);
                }
                DGL_TexCoord2f(0, .5f, 0);
                DGL_TexCoord2f(1, v1.x, v1.y);
                DGL_Vertex2f(v1.x, v1.y);
                DGL_TexCoord2f(0, 1, 0);
                DGL_TexCoord2f(1, v2.x, v2.y);
                DGL_Vertex2f(v2.x, v2.y);
                DGL_TexCoord2f(0, 1, 1);
                DGL_TexCoord2f(1, v3.x, v3.y);
                DGL_Vertex2f(v3.x, v3.y);
                DGL_TexCoord2f(0, .5, 1);
                DGL_TexCoord2f(1, v4.x, v4.y);
                DGL_Vertex2f(v4.x, v4.y);
            }
        }
        else if (!glowOnly) {
            DENG_ASSERT(rs.primType == DGL_LINES); {
                DGL_Color4f(color.x, color.y, color.z, opacity);
            }
            DGL_TexCoord2f(0, from[0], from[1]);
            DGL_Vertex2f(from[0], from[1]);
            DGL_TexCoord2f(0, to[0], to[1]);
            DGL_Vertex2f(to[0], to[1]);
            if (drawNormal) {
#define NORMTAIL_LENGTH         8
                Vector2f const v1 = (from + to) / 2;
                Vector2d const v2 = v1 + normal * NORMTAIL_LENGTH; {
                    DGL_Color4f(color.x, color.y, color.z, opacity);
                }
                DGL_TexCoord2f(0, v1.x, v1.y);
                DGL_Vertex2f(v1.x, v1.y);
                DGL_TexCoord2f(0, v2.x, v2.y);
                DGL_Vertex2f(v2.x, v2.y);
#undef NORMTAIL_LENGTH
            }
        }
    }
    void drawLine(Line *line) const {
        DENG2_ASSERT(line);
        xline_t *xline = P_ToXLine(line);
        if (xline->validCount == VALIDCOUNT)
            return;
        if ((xline->flags & ML_DONTDRAW) && !(flags & AWF_SHOW_ALLLINES))
            return;
        auto *frontSector = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
        if (frontSector && frontSector != (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR)) {
            return;
        }
        automapcfg_lineinfo_t const *info = nullptr;
        if ((flags & AWF_SHOW_ALLLINES) || xline->mapped[rs.plr - players]) {
            auto *backSector = reinterpret_cast<Sector *>(P_GetPtrp(line, DMU_BACK_SECTOR));
            info = style->tryFindLineInfo_special(xline->special, xline->flags,
                                                  frontSector, backSector, flags);
            if (rs.obType != -1 && !info) {
                if (!backSector || !P_GetPtrp(line, DMU_BACK) || (xline->flags & ML_SECRET)) {
                    info = style->tryFindLineInfo(AMO_SINGLESIDEDLINE);
                }
                else {
                    if (!de::fequal(P_GetDoublep(backSector,  DMU_FLOOR_HEIGHT),
                                    P_GetDoublep(frontSector, DMU_FLOOR_HEIGHT))) {
                        info = style->tryFindLineInfo(AMO_FLOORCHANGELINE);
                    }
                    else if (!de::fequal(P_GetDoublep(backSector,  DMU_CEILING_HEIGHT),
                                         P_GetDoublep(frontSector, DMU_CEILING_HEIGHT))) {
                        info = style->tryFindLineInfo(AMO_CEILINGCHANGELINE);
                    }
                    else if (flags & AWF_SHOW_ALLLINES) {
                        info = style->tryFindLineInfo(AMO_UNSEENLINE);
                    }
                }
            }
        }
        else if (rs.obType != -1 && revealed) {
            if (!(xline->flags & ML_DONTDRAW)) {
                info = style->tryFindLineInfo(AMO_UNSEENLINE);
            }
        }
        if (info && (rs.obType == -1 || info == &style->lineInfo(rs.obType))) {
            ddouble from[2]; P_GetDoublepv(P_GetPtrp(line, DMU_VERTEX0), DMU_XY, from);
            ddouble to  [2]; P_GetDoublepv(P_GetPtrp(line, DMU_VERTEX1), DMU_XY, to);
            drawLine2(Vector2d(from), Vector2d(to), Vector3f(info->rgba), info->rgba[3],
                      (xline->special && !cfg.common.automapShowDoors ? GLOW_NONE : info->glow),
                      info->glowStrength,
                      info->glowSize, rs.glowOnly, info->scaleWithView,
                      (info->glow && !(xline->special && !cfg.common.automapShowDoors)),
                      (flags & AWF_SHOW_LINE_NORMALS));
            xline->validCount = VALIDCOUNT; // Mark as drawn this frame.
        }
    }
    static int drawLineWorker(void *line, void *context) {
        static_cast<Impl *>(context)->drawLine((Line *)line);
        return false;  // Continue iteration.
    }
    static int drawLinesForSubspaceWorker(ConvexSubspace *subspace, void *context) {
        return P_Iteratep(subspace, DMU_LINE, drawLineWorker, context);
    }
    /**
     * Determines visible lines, draws them.
     *
     * @params objType  Type of map object being drawn.
     */
    void drawAllLines(dint obType, bool glowOnly = false) const {
        VALIDCOUNT++;
        rs.obType   = obType;
        rs.glowOnly = glowOnly;
        if (glowOnly) {
            rs.primType = DGL_QUADS;
            DGL_Enable(DGL_TEXTURE0);
            DGL_Bind(DGLuint(Get(DD_DYNLIGHT_TEXTURE)));
        }
        else {
            rs.primType = DGL_LINES;
            if (amMaskTexture) {
                DGL_Enable(DGL_TEXTURE0);
                DGL_Bind(amMaskTexture);
            }
        }
        DGL_Begin(rs.primType); {
            AABoxd aaBox;
            self().pvisibleBounds(&aaBox.minX, &aaBox.maxX, &aaBox.minY, &aaBox.maxY);
            Subspace_BoxIterator(&aaBox, drawLinesForSubspaceWorker, const_cast<Impl *>(this));
        }
        /*else
        {
            dint const numSubspaces = P_Count(DMU_SUBSPACE);
            for (dint i = 0; i < numSubspaces; ++i) {
                P_Iteratep(P_ToPtr(DMU_SUBSPACE, i), DMU_LINE, drawLineWorker, const_cast<Impl *>(this));
            }
        }*/
        DGL_End();
        DGL_Enable(DGL_TEXTURE0);
    }
    static void drawLine(Line *line, Vector3f const &color, dfloat opacity,
                         /*blendmode_t blendMode, */bool showNormal)
    {
        dfloat length = P_GetFloatp(line, DMU_LENGTH);
        if (length > 0) {
            dfloat v1[2]; P_GetFloatpv(P_GetPtrp(line, DMU_VERTEX0), DMU_XY, v1);
            dfloat v2[2]; P_GetFloatpv(P_GetPtrp(line, DMU_VERTEX1), DMU_XY, v2);
            DGL_Color4f(color.x, color.y, color.z, opacity);
            DENG_ASSERT(rs.primType == DGL_LINES);
            DGL_TexCoord2f(0, v1[0], v1[1]);
            DGL_Vertex2f(v1[0], v1[1]);
            DGL_TexCoord2f(0, v2[0], v2[1]);
            DGL_Vertex2f(v2[0], v2[1]);
            if (showNormal) {
#define NORMTAIL_LENGTH         8
                dfloat d1[2]; P_GetFloatpv(line, DMU_DXY, d1);
                dfloat const unit  [] = { d1[0] / length, d1[1] / length };
                dfloat const normal[] = { unit[1], -unit[0] };
                v1[0] += (length / 2) * unit[0];
                v1[1] += (length / 2) * unit[1];
                v2[0] = v1[0] + normal[0] * NORMTAIL_LENGTH;
                v2[1] = v1[1] + normal[1] * NORMTAIL_LENGTH;
                DGL_TexCoord2f(0, v1[0], v1[1]);
                DGL_Vertex2f(v1[0], v1[1]);
                DGL_TexCoord2f(0, v2[0], v2[1]);
                DGL_Vertex2f(v2[0], v2[1]);
#undef NORMTAIL_LENGTH
            }
        }
    }
    static int drawLine_polyob(Line *line, void *context) {
        auto const *inst = static_cast<Impl *>(context);
        DENG2_ASSERT(inst);
        dfloat const opacity = uiRendState->pageAlpha;
        xline_t *xline = P_ToXLine(line);
        if (!xline) return false;
        if (xline->validCount == VALIDCOUNT) return false;
        if ((xline->flags & ML_DONTDRAW) && !(inst->flags & AWF_SHOW_ALLLINES)) {
            return false;
        }
        automapcfg_objectname_t amo = AMO_NONE;
        if ((inst->flags & AWF_SHOW_ALLLINES) || xline->mapped[rs.plr - players]) {
            amo = AMO_SINGLESIDEDLINE;
        }
        else if (rs.obType != -1 && inst->revealed) {
            if (!(xline->flags & ML_DONTDRAW)) {
                amo = AMO_UNSEENLINE;
            }
        }
        if (automapcfg_lineinfo_t const *info = inst->style->tryFindLineInfo(amo)) {
            drawLine(line, Vector3f(info->rgba), info->rgba[3] * cfg.common.automapLineAlpha * opacity,
                     /*info->blendMode, */(inst->flags & AWF_SHOW_LINE_NORMALS));
        }
        xline->validCount = VALIDCOUNT;  // Mark as processed this frame.
        return false;  // Continue iteration.
    }
    void drawAllPolyobs() const {
        VALIDCOUNT++;  // Used to track which lines have been drawn this frame.
        rs.obType = MOL_LINEDEF;
        rs.primType = DGL_LINES;
        DGL_Begin(rs.primType); {
            AABoxd aaBox;
            self().pvisibleBounds(&aaBox.minX, &aaBox.maxX, &aaBox.minY, &aaBox.maxY);
            Line_BoxIterator(&aaBox, LIF_POLYOBJ, drawLine_polyob, const_cast<Impl *>(this));
        }
        DGL_End();
    }
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    static int drawLine_xg(Line *line, void *context) {
        auto const *inst = static_cast<Impl *>(context);
        DENG2_ASSERT(line && inst);
        xline_t *xline = P_ToXLine(line);
        if (!xline) return false;
        if (xline->validCount == VALIDCOUNT) return false;
        if (!(inst->flags & AWF_SHOW_ALLLINES)) {
            if (xline->flags & ML_DONTDRAW) return false;
        }
        if (!xline->xg || !xline->xg->active) return false;
        if (!(mapTime & 4)) return false;
        drawLine(line, Vector3f(.8f, 0, .8f), 1, (inst->flags & AWF_SHOW_LINE_NORMALS));
        xline->validCount = VALIDCOUNT;  // Mark as processed this frame.
        return false;  // Continue iteration.
    }
#endif
    void drawAllLines_xg() const {
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        if (!(flags & AWF_SHOW_SPECIALLINES))
            return;
        VALIDCOUNT++;
        rs.glowOnly = true;
        rs.obType   = -1;
        rs.primType = DGL_LINES;
        DGL_BlendMode(BM_ADD);
        DGL_Begin(rs.primType); {
            AABoxd aaBox;
            self().pvisibleBounds(&aaBox.minX, &aaBox.maxX, &aaBox.minY, &aaBox.maxY);
            Line_BoxIterator(&aaBox, LIF_SECTOR, drawLine_xg, const_cast<Impl *>(this));
        }
        DGL_End();
        DGL_BlendMode(BM_NORMAL);
#endif
    }
    /**
     * Visualize all players on the map with SVG markers.
     */
    void drawAllPlayerMarkers() const {
        for (dint i = 0; i < MAXPLAYERS; ++i) {
            if (gfw_Rule(deathmatch) && i != self().player()) {
                continue;
            }
            drawPlayerMarker(i, style);
        }
    }
    static dint thingColorForMobjType(mobjtype_t type) {
#if __JHEXEN__
        DENG2_UNUSED(type);
        return -1;
#else
        struct ThingData { mobjtype_t type; dint palColor; } static const thingData[] = {
#  if __JDOOM__ || __JDOOM64__
            { MT_MISC4, KEY1_COLOR },
            { MT_MISC5, KEY2_COLOR },
            { MT_MISC6, KEY3_COLOR },
            { MT_MISC7, KEY4_COLOR },
            { MT_MISC8, KEY5_COLOR },
            { MT_MISC9, KEY6_COLOR },
#  elif __JHERETIC__
            { MT_CKEY,  KEY1_COLOR },
            { MT_BKYY,  KEY2_COLOR },
            { MT_AKYY,  KEY3_COLOR },
#  endif
        };
        for (auto const &thing : thingData) {
            if (thing.type == type) return thing.palColor;
        }
        return -1;  // None.
#endif
    }
    struct drawthingpoint_params_t {
        dint flags;      ///< AWF_* flags.
        svgid_t vgId;
        dfloat rgb[3];
        dfloat opacity;
    };
    static int drawThingPoint(mobj_t *mob, void *context) {
        auto *p = reinterpret_cast<drawthingpoint_params_t *>(context);
        if (!(mob->flags & MF_NOSECTOR)) {
            svgid_t vgId   = p->vgId;
            bool isVisible = false;
            dfloat *color  = p->rgb;
            dfloat angle = 0;
            dfloat keyColorRGB[3];
            if (p->flags & AWF_SHOW_KEYS) {
                dint keyColor = thingColorForMobjType(mobjtype_t(mob->type));
                if (keyColor != -1) {
                    R_GetColorPaletteRGBf(0, keyColor, keyColorRGB, false);
                    vgId      = VG_KEY;
                    color     = keyColorRGB;
                    isVisible = true;
                }
            }
            if (!isVisible) {
                isVisible = !!(p->flags & AWF_SHOW_THINGS);
                angle = Mobj_AngleSmoothed(mob) / float(ANGLE_MAX) * 360;  // In degrees.
            }
            if (isVisible) {
                /* $unifiedangles */
                coord_t origin[3]; Mobj_OriginSmoothed(mob, origin);
                drawVectorGraphic(vgId, Vector2d(origin), angle, 16 /*radius*/,
                                  Vector3f(color), p->opacity, BM_NORMAL);
            }
        }
        return false; // Continue iteration.
    }
    void drawAllThings() const {
        if (!(flags & (AWF_SHOW_THINGS | AWF_SHOW_KEYS)))
            return;
        dfloat const alpha = uiRendState->pageAlpha;
        drawthingpoint_params_t parm; de::zap(parm);
        parm.flags   = flags;
        parm.vgId    = style->objectSvg(AMO_THING);
        AM_GetMapColor(parm.rgb, cfg.common.automapMobj, THINGCOLORS, customPal);
        parm.opacity = de::clamp(0.f, cfg.common.automapLineAlpha * alpha, 1.f);
        AABoxd aaBox;
        self().pvisibleBounds(&aaBox.minX, &aaBox.maxX, &aaBox.minY, &aaBox.maxY);
        VALIDCOUNT++;
        Mobj_BoxIterator(&aaBox, drawThingPoint, &parm);
    }
    void drawAllPoints(dfloat scale = 1) const {
        dfloat const alpha = uiRendState->pageAlpha;
        if (points.isEmpty()) return;
        scale = self().frameToMap(1) * scale;
#if __JHERETIC__ || __JHEXEN__
        scale *= .5f;
#endif
        dint idx = 0;
        const Point2Raw labelOffset{};
        for (MarkedPoint const *point : points) {
            String const label    = String::number(idx++);
            Vector2d const origin = fitPointInRectangle(point->origin(), topLeft, topRight, bottomRight, bottomLeft, view);
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PushMatrix();
            DGL_Translatef(origin.x, origin.y, 0);
            DGL_Scalef(scale, scale, 1);
            DGL_Rotatef(angle, 0, 0, 1);
            DGL_Scalef(1, -1, 1);
            DGL_Enable(DGL_TEXTURE_2D);
            FR_SetFont(FID(GF_MAPPOINT));
#if __JDOOM__
            if (gameMode == doom2_hacx)
                FR_SetColorAndAlpha(1, 1, 1, alpha);
            else
                FR_SetColorAndAlpha(.22f, .22f, .22f, alpha);
#else
            FR_SetColorAndAlpha(1, 1, 1, alpha);
#endif
            FR_DrawText3(label.toUtf8().constData(), &labelOffset, 0, DTF_ONLY_SHADOW);
            DGL_Disable(DGL_TEXTURE_2D);
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PopMatrix();
        }
    }
    /**
     * Sets up the state for automap drawing.
     */
    void setupGLStateForMap() const {
        const dfloat alpha = uiRendState->pageAlpha;
        DGL_PushState();
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        dfloat bgColor[3];
#if __JHERETIC__ || __JHEXEN__
        if (CentralLumpIndex().contains("AUTOPAGE.lmp")) {
            bgColor[0] = bgColor[1] = bgColor[2] = 1.f; // use lump colors as-is
        }
        else {
            bgColor[0] = .55f; bgColor[1] = .45f; bgColor[2] = .35f;
        }
#else
        AM_GetMapColor(bgColor, cfg.common.automapBack, BACKGROUND, customPal);
#endif
        RectRaw geom; Rect_Raw(&self().geometry(), &geom);
        if (autopageLumpNum != -1) {
            const float autopageWidth       = 320; /// @todo Could be external with different size.
            const float autopageHeight      = 200;
            const float autopageAspectRatio = autopageWidth / autopageHeight;
            const float texScale    = 1.f / 3000.f;
            const float bgScale     = texScale / scaleMTOF;
            const float offsetScale = texScale * autopageAspectRatio;
            DGL_Enable(DGL_TEXTURE_2D);
            DGL_MatrixMode(DGL_TEXTURE);
            DGL_PushMatrix();
            DGL_LoadIdentity();
            DGL_SetRawImage(autopageLumpNum, DGL_REPEAT, DGL_REPEAT);
            DGL_Color4f(bgColor[0], bgColor[1], bgColor[2], cfg.common.automapOpacity * alpha);
#if 0
            DGL_Translatef(geom.origin.x, geom.origin.y, 0);
            DGL_Translatef(self().mapToFrame(viewPL.x) + .5f,
                           self().mapToFrame(viewPL.y) + .5f, 0);
            DGL_Scalef(1, 1.2f/*aspect correct*/, 1);
            DGL_Rotatef(360 - self().cameraAngle(), 0, 0, 1);
            DGL_Translatef(-(.5f), -(.5f), 0);
#endif
            DENG_ASSERT(!std::isnan(view.x));
            DGL_Translatef(offsetScale * view.x, -offsetScale * view.y, 1.f);
            DGL_Scalef(autopageAspectRatio, autopageAspectRatio, 1.f);
            DGL_Rotatef(360.f - self().cameraAngle(), 0, 0, 1);
            DGL_Scalef(bgScale * float(geom.size.width), bgScale * float(geom.size.height), 1.f);
            DGL_Translatef(-.5f, -.5f, 0); // center
            DGL_DrawRectf2(geom.origin.x, geom.origin.y, geom.size.width, geom.size.height);
            DGL_MatrixMode(DGL_TEXTURE);
            DGL_PopMatrix();
            DGL_Disable(DGL_TEXTURE_2D);
        }
        else {
            DGL_SetNoMaterial();
            DGL_Color4f(bgColor[0], bgColor[1], bgColor[2], cfg.common.automapOpacity * alpha);
            DGL_DrawRectf2(0, 0, geom.size.width, geom.size.height);
        }
#if __JDOOM64__
        if (!cfg.hudShown[HUD_INVENTORY]) {
            static inventoryitemtype_t const items[3] = {
                IIT_DEMONKEY1, IIT_DEMONKEY2, IIT_DEMONKEY3
            };
            dint player = self().player();
            dint num = 0;
            for (inventoryitemtype_t const &item : items) {
                if (P_InventoryCount(player, item) > 0)
                    num += 1;
            }
            if (num > 0) {
                static dint const invItemSprites[NUM_INVENTORYITEM_TYPES] = {
                    SPR_ART1, SPR_ART2, SPR_ART3
                };
                dfloat const iconOpacity = de::clamp(.0f, alpha, .5f);
                dfloat const spacing     = geom.size.height / num;
                spriteinfo_t sprInfo;
                dfloat y = 0;
                for (dint i = 0; i < 3; ++i) {
                    if (P_InventoryCount(player, items[i])) {
                        R_GetSpriteInfo(invItemSprites[i], 0, &sprInfo);
                        DGL_SetPSprite(sprInfo.material);
                        DGL_Enable(DGL_TEXTURE_2D);
                        dfloat const scale = geom.size.height / (sprInfo.geometry.size.height * num);
                        dfloat const x     = geom.size.width - sprInfo.geometry.size.width * scale;
                        dfloat const w     = sprInfo.geometry.size.width;
                        dfloat const h     = sprInfo.geometry.size.height;
                        DGL_Color4f(1, 1, 1, iconOpacity);
                        DGL_Begin(DGL_QUADS);
                            DGL_TexCoord2f(0, 0, 0);
                            DGL_Vertex2f(x, y);
                            DGL_TexCoord2f(0, sprInfo.texCoord[0], 0);
                            DGL_Vertex2f(x + w * scale, y);
                            DGL_TexCoord2f(0, sprInfo.texCoord[0], sprInfo.texCoord[1]);
                            DGL_Vertex2f(x + w * scale, y + h * scale);
                            DGL_TexCoord2f(0, 0, sprInfo.texCoord[1]);
                            DGL_Vertex2f(x, y + h * scale);
                        DGL_End();
                        DGL_Disable(DGL_TEXTURE_2D);
                        y += spacing;
                    }
                }
            }
        }
#endif
        dint const border = .5f + UIAUTOMAP_BORDER * aspectScale;
        RectRaw clipRegion; Rect_Raw(&self().geometry(), &clipRegion);
        clipRegion.origin.x += border;
        clipRegion.origin.y += border;
        clipRegion.size.width  -= 2 * border;
        clipRegion.size.height -= 2 * border;
        DGL_SetScissor(&clipRegion);
    }
    /**
     * Restores the previous GL draw state
     */
    void restoreGLStateFromMap() {
        DGL_PopState();
    }
    void drawAllVertexes() {
        if (!(flags & AWF_SHOW_VERTEXES))
            return;
        DGL_Color4f(.2f, .5f, 1, uiRendState->pageAlpha);
        DGL_Enable(DGL_POINT_SMOOTH);
        dfloat const oldPointSize = DGL_GetFloat(DGL_POINT_SIZE);
        DGL_SetFloat(DGL_POINT_SIZE, 4 * aspectScale);
        dfloat v[2];
        DGL_Begin(DGL_POINTS);
        for (dint i = 0; i < numvertexes; ++i) {
            P_GetFloatv(DMU_VERTEX, i, DMU_XY, v);
            DGL_TexCoord2f(0, v[0], v[1]);
            DGL_Vertex2f(v[0], v[1]);
        }
        DGL_End();
        DGL_SetFloat(DGL_POINT_SIZE, oldPointSize);
        DGL_Disable(DGL_POINT_SMOOTH);
    }
};
AutomapWidget::AutomapWidget(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(AutomapWidget_UpdateGeometry),
                function_cast<DrawFunc>(AutomapWidget_Draw),
                player)
    , d(new Impl(this)) {
    d->style = ST_AutomapStyle();
}
AutomapWidget::~AutomapWidget()
{}
dint AutomapWidget::cameraFollowPlayer() const {
    return d->followPlayer;
}
void AutomapWidget::setCameraFollowPlayer(dint newPlayer) {
    d->followPlayer = newPlayer;
}
void AutomapWidget::prepareAssets()  // static
{
    LumpIndex const &lumpIndex = CentralLumpIndex();
    if (autopageLumpNum >= 0) {
        autopageLumpNum = lumpIndex.findLast("autopage.lmp");
    }
    if (!amMaskTexture) {
        lumpnum_t lumpNum = lumpIndex.findLast("mapmask.lmp");
        if (lumpNum >= 0) {
            File1 &file = lumpIndex[lumpNum];
            uint8_t const *pixels = file.cache();
            amMaskTexture = DGL_NewTextureWithParams(DGL_LUMINANCE, 256/*width*/, 256/*height*/,
                                                     pixels, 0x8, DGL_NEAREST, DGL_LINEAR,
                                                     0 /*no anisotropy*/, DGL_REPEAT, DGL_REPEAT);
            file.unlock();
        }
    }
}
void AutomapWidget::releaseAssets()  // static
{
    if (!amMaskTexture) return;
    DGL_DeleteTextures(1, &amMaskTexture);
    amMaskTexture = 0;
}
void AutomapWidget::reset() {
    d->needBuildLists = true;
    d->rotate         = cfg.common.automapRotate;
}
void AutomapWidget::lineAutomapVisibilityChanged(Line const &) {
    d->needBuildLists = true;
}
AutomapStyle *AutomapWidget::style() const {
    return d->style;
}
void AutomapWidget::draw(Vector2i const &offset) const {
    static int updateWait = 0;  /// @todo should be an instance var of AutomapWidget
    float const alpha = uiRendState->pageAlpha;
    player_t *plr = &players[player()];
    if (!plr->plr->inGame) return;
    rs.plr = plr;
    const Vector2d viewPoint = cameraOrigin();
    float angle = cameraAngle();
    RectRaw geom; Rect_Raw(&geometry(), &geom);
    d->setupGLStateForMap();
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Translatef(geom.size.width / 2, geom.size.height / 2, 0);
    DGL_Rotatef(angle, 0, 0, 1);
    DGL_Scalef(1, -1, 1); // In the world coordinate space Y+ is up.
    DGL_Scalef(d->scaleMTOF, d->scaleMTOF, 1);
    DGL_Translatef(-viewPoint.x, -viewPoint.y, 0);
    const float oldLineWidth = DGL_GetFloat(DGL_LINE_WIDTH);
    DGL_SetFloat(DGL_LINE_WIDTH, d->pixelRatio * de::clamp(.5f, cfg.common.automapLineWidth, 8.f));
    /*#if _DEBUG
    {
        coord_t topLeft[2], bottomRight[2], topRight[2], bottomLeft[2];
        d->visibleBounds(topLeft, bottomRight, topRight, bottomLeft);
        DGL_Color4f(1, 1, 1, alpha);
        DGL_Begin(DGL_LINES);
            DGL_Vertex2f(    topLeft[0],     topLeft[1]);
            DGL_Vertex2f(   topRight[0],    topRight[1]);
            DGL_Vertex2f(   topRight[0],    topRight[1]);
            DGL_Vertex2f(bottomRight[0], bottomRight[1]);
            DGL_Vertex2f(bottomRight[0], bottomRight[1]);
            DGL_Vertex2f( bottomLeft[0],  bottomLeft[1]);
            DGL_Vertex2f( bottomLeft[0],  bottomLeft[1]);
            DGL_Vertex2f(    topLeft[0],     topLeft[1]);
        DGL_End();
    }
#endif*/
    if (amMaskTexture) {
        dint const border = .5f + UIAUTOMAP_BORDER * aspectScale;
        DGL_SetInteger(DGL_ACTIVE_TEXTURE, 0);
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_LoadIdentity();
        DGL_PushMatrix();
        DGL_Scalef(1.f / (geom.size.width  - border*2),
                   1.f / (geom.size.height - border*2), 1);
        DGL_Translatef(geom.size.width  /2 - border,
                       geom.size.height /2 - border, 0);
        DGL_Rotatef(-angle, 0, 0, 1);
        DGL_Scalef(d->scaleMTOF, d->scaleMTOF, 1);
        DGL_Translatef(-viewPoint.x, -viewPoint.y, 0);
    }
    for (dint i = NUM_MAP_OBJECTLISTS-1; i >= 0; i--) {
        automapcfg_lineinfo_t const &info = d->style->lineInfo(i);
        DGL_Color4f(info.rgba[0], info.rgba[1], info.rgba[2], info.rgba[3] * cfg.common.automapLineAlpha * alpha);
        d->drawAllLines(i);
    }
    d->drawAllLines_xg();
    d->drawAllPolyobs();
    DGL_BlendMode(BM_NORMAL);
    DGL_Color4f(1, 1, 1, 1);
    d->drawAllVertexes();
    d->drawAllThings();
    DGL_SetFloat(DGL_LINE_WIDTH, 1.f);
    d->drawAllPlayerMarkers();
    DGL_SetFloat(DGL_LINE_WIDTH, oldLineWidth);
    if (amMaskTexture) {
        DGL_Disable(DGL_TEXTURE0);
        DGL_MatrixMode(DGL_TEXTURE);
        DGL_PopMatrix();
    }
    if (cfg.common.automapShowDoors) {
        d->drawAllLines(-1, true /*only glows*/);
    }
    d->restoreGLStateFromMap();
    d->drawAllPoints(aspectScale);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void AutomapWidget::open(bool yes, bool instantly) {
    if (G_GameState() != GS_MAP && yes) return;
    if (d->open == yes) return;  // No change.
    d->targetOpacity = (yes? 1.f : 0.f);
    if (instantly) {
        d->opacity = d->oldOpacity = d->targetOpacity;
    }
    else {
        d->oldOpacity   = d->opacity;
        d->opacityTimer = 0.f;
    }
    d->open = yes;
    if (d->open) {
        if (mobj_t *mob = followMobj()) {
            if (d->follow || cfg.common.automapPanResetOnOpen) {
                coord_t origin[3]; Mobj_OriginSmoothed(mob, origin);
                setCameraOrigin(Vector2d(origin));
            }
            if (!d->follow && cfg.common.automapPanResetOnOpen) {
                /* $unifiedangles */
                setCameraAngle((d->rotate ? (mob->angle - ANGLE_90) / (float) ANGLE_MAX * 360 : 0));
            }
        }
        else {
            coord_t aabb[4];
            pvisibleBounds(&aabb[BOXLEFT], &aabb[BOXRIGHT], &aabb[BOXBOTTOM], &aabb[BOXTOP]);
            setCameraOrigin(Vector2d(aabb[BOXRIGHT] - aabb[BOXLEFT], aabb[BOXTOP] - aabb[BOXBOTTOM]) / 2);
            setCameraAngle(0);
        }
    }
    if (d->open) {
        DD_Execute(true, "activatebcontext map");
        if (!d->follow)
            DD_Execute(true, "activatebcontext map-freepan");
    }
    else {
        DD_Execute(true, "deactivatebcontext map");
        DD_Execute(true, "deactivatebcontext map-freepan");
    }
}
void AutomapWidget::tick(timespan_t elapsed) {
    dint const plrNum = player();
    mobj_t *followMob = followMobj();
    dfloat panX[2]; P_GetControlState(plrNum, CTL_MAP_PAN_X, &panX[0], &panX[1]);
    dfloat panY[2]; P_GetControlState(plrNum, CTL_MAP_PAN_Y, &panY[0], &panY[1]);
    if (G_GameState() != GS_MAP) return;
    if (cfg.common.automapOpenSeconds >= .001f) {
        d->opacityTimer += 1.f / cfg.common.automapOpenSeconds * elapsed;
    }
    else {
        d->opacityTimer = 1.f; // Instant.
    }
    if (d->opacityTimer >= 1) {
        d->opacity = d->targetOpacity;
    }
    else {
        d->opacity = de::lerp(d->oldOpacity, d->targetOpacity, d->opacityTimer);
    }
    if (!isOpen()) return;
    dfloat zoomSpeed = 1 + (2 * cfg.common.automapZoomSpeed) * elapsed * TICRATE;
    if (players[plrNum].brain.speed) {
        zoomSpeed *= 1.5f;
    }
    dfloat zoomVel;
    P_GetControlState(plrNum, CTL_MAP_ZOOM, &zoomVel, nullptr); // ignores rel offset -jk
    if (zoomVel > 0) // zoom in
    {
        setScale(d->viewScale * zoomSpeed);
    }
    else if (zoomVel < 0) // zoom out
    {
        setScale(d->viewScale / zoomSpeed);
    }
    if (!d->follow || !followMob) {
        dfloat panUnitsPerSecond;
        panUnitsPerSecond = de::max(8.f,
                                    frameToMap(140 * Rect_Height(&geometry()) / 200.f) *
                                        (2 * cfg.common.automapPanSpeed));
        Vector2d const delta = rotate(Vector2d(panX[0], panY[0]) * panUnitsPerSecond * elapsed +
                                          Vector2d(panX[1], panY[1]),
                                      degreeToRadian(d->angle));
        moveCameraOrigin(delta, true /*instant move*/);
    }
    else {
        dfloat const angle = (d->rotate ? (followMob->angle - ANGLE_90) / (dfloat) ANGLE_MAX * 360
                                        : 0); /* $unifiedangles */
        coord_t origin[3]; Mobj_OriginSmoothed(followMob, origin);
        setCameraOrigin(Vector2d(origin));
        setCameraAngle(angle);
    }
    if (d->needViewScaleUpdate)
        d->updateViewScale();
    d->viewTimer += dfloat(.4 * elapsed * TICRATE);
    if (d->viewTimer >= 1) {
        d->view = d->targetView;
    }
    else {
        d->view = de::lerp(d->oldView, d->targetView, d->viewTimer);
    }
    d->viewScaleTimer += dfloat(.4 * elapsed * TICRATE);
    if (d->viewScaleTimer >= 1) {
        d->viewScale = d->targetViewScale;
    }
    else {
        d->viewScale = de::lerp(d->oldViewScale, d->targetViewScale, d->viewScaleTimer);
    }
    d->angleTimer += dfloat(.4 * elapsed * TICRATE);
    if (d->angleTimer >= 1) {
        d->angle = d->targetAngle;
    }
    else {
        dfloat startAngle = d->oldAngle;
        dfloat endAngle   = d->targetAngle;
        dfloat diff;
        if (endAngle > startAngle) {
            diff = endAngle - startAngle;
            if (diff > 180)
                endAngle = startAngle - (360 - diff);
        }
        else {
            diff = startAngle - endAngle;
            if (diff > 180)
                endAngle = startAngle + (360 - diff);
        }
        d->angle = de::lerp(startAngle, endAngle, d->angleTimer);
        if (d->angle < 0)        d->angle += 360;
        else if (d->angle > 360) d->angle -= 360;
    }
    d->scaleMTOF = d->viewScale;
    d->scaleFTOM = 1.0f / d->scaleMTOF;
    dint const border         = .5f + UIAUTOMAP_BORDER * aspectScale;
    ddouble const ang         = degreeToRadian(d->angle);
    Vector2d const origin     = cameraOrigin();
    auto const dimensions     = Vector2d(frameToMap(Rect_Width (&geometry())),
                                         frameToMap(Rect_Height(&geometry()))) / 2;
    auto const viewDimensions = Vector2d(frameToMap(Rect_Width (&geometry()) - border * 2),
                                         frameToMap(Rect_Height(&geometry()) - border * 2)) / 2;
    d->topLeft     = origin + rotate(Vector2d(-viewDimensions.x,  viewDimensions.y), ang);
    d->bottomRight = origin + rotate(Vector2d( viewDimensions.x, -viewDimensions.y), ang);
    d->bottomLeft  = origin + rotate(-viewDimensions, ang);
    d->topRight    = origin + rotate( viewDimensions, ang);
    initAABB (d->viewAABB, rotate(-dimensions, ang));
    addToAABB(d->viewAABB, rotate(Vector2d( dimensions.x, -dimensions.y), ang));
    addToAABB(d->viewAABB, rotate(Vector2d(-dimensions.x,  dimensions.y), ang));
    addToAABB(d->viewAABB, rotate( dimensions, ang));
    d->viewAABB[BOXLEFT  ] += origin.x;
    d->viewAABB[BOXRIGHT ] += origin.x;
    d->viewAABB[BOXTOP   ] += origin.y;
    d->viewAABB[BOXBOTTOM] += origin.y;
}
dfloat AutomapWidget::mapToFrame(dfloat coord) const {
    return coord * d->scaleMTOF;
}
dfloat AutomapWidget::frameToMap(dfloat coord) const {
    return coord * d->scaleFTOM;
}
void AutomapWidget::updateGeometry() {
    RectRaw newGeom;
    R_ViewWindowGeometry(player(), &newGeom);
    if (newGeom.origin.x != Rect_X(&geometry()) ||
        newGeom.origin.y != Rect_Y(&geometry()) ||
        newGeom.size.width != Rect_Width(&geometry()) ||
        newGeom.size.height != Rect_Height(&geometry())) {
        Rect_SetXY(&geometry(), newGeom.origin.x, newGeom.origin.y);
        Rect_SetWidthHeight(&geometry(), newGeom.size.width, newGeom.size.height);
        d->needViewScaleUpdate = true;
    }
}
dfloat AutomapWidget::cameraAngle() const {
    return d->angle;
}
void AutomapWidget::setCameraAngle(dfloat newAngle) {
    newAngle = de::clamp(0.f, newAngle, 359.9999f);
    if (newAngle == d->targetAngle) return;
    d->oldAngle    = d->angle;
    d->targetAngle = newAngle;
    d->angleTimer  = 0;
}
Vector2d AutomapWidget::cameraOrigin() const {
    return d->view;
}
void AutomapWidget::setCameraOrigin(Vector2d const &newOrigin, bool instantly) {
    if (newOrigin == d->targetView)
        return;
    if (!instantly && d->maxViewPositionDelta > 0) {
        coord_t const dist = de::abs((cameraOrigin() - newOrigin).length());
        if (dist > d->maxViewPositionDelta) {
            instantly = true;
        }
    }
    if (instantly) {
        d->view = d->oldView = d->targetView = newOrigin;
    }
    else {
        d->oldView    = d->view;
        d->targetView = newOrigin;
        d->viewTimer  = 0;
    }
}
dfloat AutomapWidget::scale() const {
    return d->targetViewScale;
}
void AutomapWidget::setScale(dfloat newScale) {
    if (d->needViewScaleUpdate)
        d->updateViewScale();
    newScale = de::clamp(d->minScaleMTOF, newScale, d->maxScaleMTOF);
    if (newScale == d->targetViewScale)
        return;
    d->oldViewScale    = d->viewScale;
    d->viewScaleTimer  = 0;
    d->targetViewScale = newScale;
}
bool AutomapWidget::isOpen() const {
    return d->open;
}
bool AutomapWidget::isRevealed() const {
    return d->revealed;
}
void AutomapWidget::reveal(bool yes) {
    if (d->revealed != yes) {
        d->revealed = yes;
        d->needBuildLists = true;
    }
}
void AutomapWidget::pvisibleBounds(coord_t *lowX, coord_t *hiX, coord_t *lowY, coord_t *hiY) const {
    if (lowX) *lowX = d->viewAABB[BOXLEFT];
    if (hiX)  *hiX  = d->viewAABB[BOXRIGHT];
    if (lowY) *lowY = d->viewAABB[BOXBOTTOM];
    if (hiY)  *hiY  = d->viewAABB[BOXTOP];
}
dint AutomapWidget::pointCount() const {
    return d->points.count();
}
dint AutomapWidget::addPoint(Vector3d const &origin) {
    d->points << new MarkedPoint(origin);
    dint pointNum = d->points.count() - 1;  // base 0.
    if (player() >= 0) {
        String msg = String(AMSTR_MARKEDSPOT) + " " + String::number(pointNum);
        P_SetMessageWithFlags(&players[player()], msg.toUtf8().constData(), LMF_NO_HIDE);
    }
    return pointNum;
}
bool AutomapWidget::hasPoint(dint index) const {
    return index >= 0 && index < d->points.count();
}
AutomapWidget::MarkedPoint &AutomapWidget::point(dint index) const {
    if (hasPoint(index)) return *d->points.at(index);
    throw MissingPointError("AutomapWidget::point", "Unknown point #" + String::number(index));
}
LoopResult AutomapWidget::forAllPoints(std::function<LoopResult (MarkedPoint &)> func) const {
    for (MarkedPoint *point : d->points) {
        if (auto result = func(*point)) return result;
    }
    return LoopContinue;
}
void AutomapWidget::clearAllPoints(bool silent) {
    d->clearPoints();
    if (!silent && player() >= 0) {
        P_SetMessageWithFlags(&players[player()], AMSTR_MARKSCLEARED, LMF_NO_HIDE);
    }
}
bool AutomapWidget::cameraZoomMode() const {
    return d->forceMaxScale;
}
void AutomapWidget::setCameraZoomMode(bool yes) {
    LOG_AS("AutomapWidget");
    bool const oldZoomMax = d->forceMaxScale;
    if (d->needViewScaleUpdate) {
        d->updateViewScale();
    }
    if (!d->forceMaxScale) {
        d->priorToMaxScale = d->viewScale;
    }
    d->forceMaxScale = yes;
    setScale((d->forceMaxScale ? 0 : d->priorToMaxScale));
    if (oldZoomMax != d->forceMaxScale) {
        LOGDEV_XVERBOSE("Maximum zoom: ", DENG2_BOOL_YESNO(cameraZoomMode()));
    }
}
bool AutomapWidget::cameraFollowMode() const {
    return d->follow;
}
void AutomapWidget::setCameraFollowMode(bool yes) {
    if (d->follow != yes) {
        d->follow = yes;
        if (d->open) {
            DD_Executef(true, "%sactivatebcontext map-freepan", d->follow? "de" : "");
            P_SetMessageWithFlags(&players[player()], (d->follow ? AMSTR_FOLLOWON : AMSTR_FOLLOWOFF), LMF_NO_HIDE);
        }
    }
}
mobj_t *AutomapWidget::followMobj() const {
    if (d->followPlayer >= 0) {
        player_t *player = &players[d->followPlayer];
        return player->plr->inGame ? player->plr->mo : nullptr;
    }
    return nullptr;
}
bool AutomapWidget::cameraRotationMode() const {
    return d->rotate;
}
void AutomapWidget::setCameraRotationMode(bool yes) {
    d->rotate = yes;
}
dfloat AutomapWidget::opacityEX() const {
    return d->opacity;
}
void AutomapWidget::setOpacityEX(dfloat newOpacity) {
    newOpacity = de::clamp(0.f, newOpacity, 1.f);
    if (newOpacity != d->targetOpacity) {
        d->oldOpacity    = d->opacity;
        d->targetOpacity = newOpacity;
        d->opacityTimer  = 0;
    }
}
dint AutomapWidget::flags() const {
    return d->flags;
}
void AutomapWidget::setFlags(dint newFlags) {
    if (d->flags != newFlags) {
        d->flags = newFlags;
        d->needBuildLists = true;
    }
}
void AutomapWidget::setMapBounds(coord_t lowX, coord_t hiX, coord_t lowY, coord_t hiY) {
    d->bounds[BOXLEFT  ] = lowX;
    d->bounds[BOXTOP   ] = hiY;
    d->bounds[BOXRIGHT ] = hiX;
    d->bounds[BOXBOTTOM] = lowY;
    d->updateViewScale();
    setScale(d->minScaleMTOF * 2.4f);  // Default view scale factor.
}
void AutomapWidget::consoleRegister()  // static
{
    C_VAR_FLOAT("map-opacity",              &cfg.common.automapOpacity,        0, 0, 1);
    C_VAR_BYTE ("map-neverobscure",         &cfg.common.automapNeverObscure,   0, 0, 1);
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    C_VAR_BYTE ("map-babykeys",             &cfg.common.automapBabyKeys,       0, 0, 1);
#endif
    C_VAR_FLOAT("map-background-r",         &cfg.common.automapBack[0],        0, 0, 1);
    C_VAR_FLOAT("map-background-g",         &cfg.common.automapBack[1],        0, 0, 1);
    C_VAR_FLOAT("map-background-b",         &cfg.common.automapBack[2],        0, 0, 1);
    C_VAR_INT  ("map-customcolors",         &cfg.common.automapCustomColors,   0, 0, 1);
    C_VAR_FLOAT( "map-line-opacity",        &cfg.common.automapLineAlpha,      0, 0, 1);
    C_VAR_FLOAT("map-line-width",           &cfg.common.automapLineWidth,      0, .5f, 8);
    C_VAR_FLOAT("map-mobj-r",               &cfg.common.automapMobj[0],        0, 0, 1);
    C_VAR_FLOAT("map-mobj-g",               &cfg.common.automapMobj[1],        0, 0, 1);
    C_VAR_FLOAT("map-mobj-b",               &cfg.common.automapMobj[2],        0, 0, 1);
    C_VAR_FLOAT("map-wall-r",               &cfg.common.automapL1[0],          0, 0, 1);
    C_VAR_FLOAT("map-wall-g",               &cfg.common.automapL1[1],          0, 0, 1);
    C_VAR_FLOAT("map-wall-b",               &cfg.common.automapL1[2],          0, 0, 1);
    C_VAR_FLOAT("map-wall-unseen-r",        &cfg.common.automapL0[0],          0, 0, 1);
    C_VAR_FLOAT("map-wall-unseen-g",        &cfg.common.automapL0[1],          0, 0, 1);
    C_VAR_FLOAT("map-wall-unseen-b",        &cfg.common.automapL0[2],          0, 0, 1);
    C_VAR_FLOAT("map-wall-floorchange-r",   &cfg.common.automapL2[0],          0, 0, 1);
    C_VAR_FLOAT("map-wall-floorchange-g",   &cfg.common.automapL2[1],          0, 0, 1);
    C_VAR_FLOAT("map-wall-floorchange-b",   &cfg.common.automapL2[2],          0, 0, 1);
    C_VAR_FLOAT("map-wall-ceilingchange-r", &cfg.common.automapL3[0],          0, 0, 1);
    C_VAR_FLOAT("map-wall-ceilingchange-g", &cfg.common.automapL3[1],          0, 0, 1);
    C_VAR_FLOAT("map-wall-ceilingchange-b", &cfg.common.automapL3[2],          0, 0, 1);
    C_VAR_BYTE ("map-door-colors",          &cfg.common.automapShowDoors,      0, 0, 1);
    C_VAR_FLOAT("map-door-glow",            &cfg.common.automapDoorGlow,       0, 0, 200);
    C_VAR_INT  ("map-huddisplay",           &cfg.common.automapHudDisplay,     0, 0, 2);
    C_VAR_FLOAT("map-pan-speed",            &cfg.common.automapPanSpeed,       0, 0, 1);
    C_VAR_BYTE ("map-pan-resetonopen",      &cfg.common.automapPanResetOnOpen, 0, 0, 1);
    C_VAR_BYTE ("map-rotate",               &cfg.common.automapRotate,         0, 0, 1);
    C_VAR_FLOAT("map-zoom-speed",           &cfg.common.automapZoomSpeed,      0, 0, 1);
    C_VAR_FLOAT("map-open-timer",           &cfg.common.automapOpenSeconds,    CVF_NO_MAX, 0, 0);
    C_VAR_BYTE ("map-title-position",       &cfg.common.automapTitleAtBottom,  0, 0, 1);
    C_VAR_BYTE ("rend-dev-freeze-map",      &freezeMapRLs,                     CVF_NO_ARCHIVE, 0, 1);
    C_VAR_FLOAT("map-alpha-lines",          &cfg.common.automapLineAlpha,      0, 0, 1);
}
void G_SetAutomapRotateMode(byte enableRotate) {
    cfg.common.automapRotate = enableRotate; // Note: this sets the global default.
    for (int i = 0; i < MAXPLAYERS; ++i) {
        ST_SetAutomapCameraRotation(i, cfg.common.automapRotate);
        if (players[i].plr->inGame) {
            P_SetMessageWithFlags(&players[i],
                                  (cfg.common.automapRotate ? AMSTR_ROTATEON : AMSTR_ROTATEOFF),
                                  LMF_NO_HIDE);
        }
    }
}
static void ChainWidget_Draw(guidata_chain_t *chain, Point2Raw const *offset) {
    DENG2_ASSERT(chain);
    chain->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void ChainWidget_UpdateGeometry(guidata_chain_t *chain) {
    DENG2_ASSERT(chain);
    chain->updateGeometry();
}
#if __JHERETIC__
static patchid_t pChain;
static patchid_t pGem[NUMTEAMS];
#elif __JHEXEN__
static patchid_t pChain[3];          ///< [Fighter, Cleric, Mage]
static patchid_t pGem[3][NUMTEAMS];  ///< [Fighter, Cleric, Mage]
#endif
guidata_chain_t::guidata_chain_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(ChainWidget_UpdateGeometry),
                function_cast<DrawFunc>(ChainWidget_Draw),
                player)
{}
guidata_chain_t::~guidata_chain_t()
{}
void guidata_chain_t::reset() {
    _healthMarker = 0;
    _wiggle       = 0;
}
void guidata_chain_t::tick(timespan_t /*elapsed*/)
{
#if __JHERETIC__
#  define MAX_DELTA         ( 4 )
#else  // __JHEXEN__
#  define MAX_DELTA         ( 6 )
#endif
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr  = &::players[player()];
    dint const curHealth = de::max(plr->plr->mo->health, 0);
    dint delta = 0;
    if(curHealth < _healthMarker) {
        delta = -de::clamp(1, (_healthMarker - curHealth) >> 2, MAX_DELTA);
    }
    else if(curHealth > _healthMarker) {
        delta = de::clamp(1, (curHealth - _healthMarker) >> 2, MAX_DELTA);
    }
    _healthMarker += delta;
    if(_healthMarker != curHealth && (::mapTime & 1)) {
        _wiggle = P_Random() & 1;
    }
    else {
        _wiggle = 0;
    }
#undef MAX_DELTA
}
#ifdef __JHERETIC__
static void drawShadows(dint x, dint y, dfloat alpha) {
    DGL_Begin(DGL_QUADS);
        DGL_Color4f(0, 0, 0, alpha);
        DGL_Vertex2f(x+20, y+ST_HEIGHT);
        DGL_Vertex2f(x+20, y+ST_HEIGHT-10);
        DGL_Color4f(0, 0, 0, 0);
        DGL_Vertex2f(x+35, y+ST_HEIGHT-10);
        DGL_Vertex2f(x+35, y+ST_HEIGHT);
        DGL_Vertex2f(x+ST_WIDTH-43, y+ST_HEIGHT);
        DGL_Vertex2f(x+ST_WIDTH-43, y+ST_HEIGHT-10);
        DGL_Color4f(0, 0, 0, alpha);
        DGL_Vertex2f(x+ST_WIDTH-27, y+ST_HEIGHT-10);
        DGL_Vertex2f(x+ST_WIDTH-27, y+ST_HEIGHT);
    DGL_End();
}
#endif
void guidata_chain_t::draw(Vector2i const &offset) const {
#if __JHERETIC__
#define ORIGINX         (-ST_WIDTH / 2 )
#define ORIGINY         ( 0 )
    static dint const theirColors[] = {
        /*Green*/ 220, /*Yellow*/ 144, /*Red*/ 150, /*Blue*/ 197
    };
    dint const activeHud     = ST_ActiveHud(player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t pChainInfo;
    if(!R_GetPatchInfo(pChain, &pChainInfo)) return;
    dint teamColor = 2; // Always use the red gem in single player.
    if(IS_NETGAME) {
        teamColor = ::cfg.playerColor[player()];
    }
    patchinfo_t pGemInfo;
    if(!R_GetPatchInfo(::pGem[teamColor], &pGemInfo)) return;
    dint const chainY      = -9 + _wiggle;
    dfloat const healthPos = de::clamp(0.f, _healthMarker / 100.f, 1.f);
    dfloat const gemglow   = healthPos;
    dint x = ORIGINX + 21;
    dint y = ORIGINY + chainY;
    dint w = ST_WIDTH - 21 - 28;
    dint h = 8;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_SetPatch(pChain, DGL_REPEAT, DGL_CLAMP);
    DGL_Color4f(1, 1, 1, iconOpacity);
    dfloat const gemXOffset = (w - pGemInfo.geometry.size.width) * healthPos;
    if(gemXOffset > 0) {
        dfloat cw = gemXOffset / pChainInfo.geometry.size.width;
        DGL_Begin(DGL_QUADS);
            DGL_TexCoord2f(0, 1 - cw, 0);
            DGL_Vertex2f(x, y);
            DGL_TexCoord2f(0, 1, 0);
            DGL_Vertex2f(x + gemXOffset, y);
            DGL_TexCoord2f(0, 1, 1);
            DGL_Vertex2f(x + gemXOffset, y + h);
            DGL_TexCoord2f(0, 1 - cw, 1);
            DGL_Vertex2f(x, y + h);
        DGL_End();
    }
    if(gemXOffset + pGemInfo.geometry.size.width < w) {
        dfloat cw = (w - gemXOffset - pGemInfo.geometry.size.width) / pChainInfo.geometry.size.width;
        DGL_Begin(DGL_QUADS);
            DGL_TexCoord2f(0, 0, 0);
            DGL_Vertex2f(x + gemXOffset + pGemInfo.geometry.size.width, y);
            DGL_TexCoord2f(0, cw, 0);
            DGL_Vertex2f(x + w, y);
            DGL_TexCoord2f(0, cw, 1);
            DGL_Vertex2f(x + w, y + h);
            DGL_TexCoord2f(0, 0, 1);
            DGL_Vertex2f(x + gemXOffset + pGemInfo.geometry.size.width, y + h);
        DGL_End();
    }
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(pGemInfo.id, Vector2i(x + gemXOffset, chainY));
    DGL_Disable(DGL_TEXTURE_2D);
    drawShadows(ORIGINX, ORIGINY - ST_HEIGHT, iconOpacity / 2);
    DGL_BlendMode(BM_ADD);
    DGL_Bind(Get(DD_DYNLIGHT_TEXTURE));
    DGL_Enable(DGL_TEXTURE_2D);
    dfloat rgb[3]; R_GetColorPaletteRGBf(0, theirColors[teamColor], rgb, false);
    DGL_DrawRectf2Color(x + gemXOffset - 11, chainY - 6, 41, 24, rgb[0], rgb[1], rgb[2], gemglow - (1 - iconOpacity));
    DGL_Color4f(1, 1, 1, 1);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_BlendMode(BM_NORMAL);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef ORIGINX
#undef ORIGINY
#elif __JHEXEN__  // __JHERETIC__
#define ORIGINX         (-ST_WIDTH / 2 )
#define ORIGINY         ( 0 )
   static dint const theirColors[] = {
       /*Blue*/ 157, /*Red*/   177, /*Yellow*/ 137, /*Green*/  198,
       /*Jade*/ 215, /*White*/  32, /*Hazel*/  106, /*Purple*/ 234
   };
   dint const activeHud     = ST_ActiveHud(player());
   dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
   dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
   if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
   if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
   dint const plrClass = ::cfg.playerClass[player()];
   patchinfo_t pChainInfo;
   if(!R_GetPatchInfo(::pChain[plrClass], &pChainInfo)) return;
   dint teamColor = 1; // Always use the red gem in single player.
   if(IS_NETGAME) {
       teamColor = ::players[player()].colorMap; // ::cfg.playerColor[wi->player];
       teamColor = (teamColor == 1? 0 : teamColor == 0? 1 : teamColor);
   }
   patchinfo_t pGemInfo;
   if(!R_GetPatchInfo(::pGem[plrClass][teamColor], &pGemInfo)) return;
   dfloat const healthPos = de::clamp(0.f, _healthMarker / 100.f, 100.f);
   dfloat const gemglow   = healthPos;
   dint x = ORIGINX + 43;
   dint y = ORIGINY - 7;
   dint w = ST_WIDTH - 43 - 43;
   dint h = 7;
   DGL_MatrixMode(DGL_MODELVIEW);
   DGL_PushMatrix();
   DGL_Translatef(offset.x, offset.y, 0);
   DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
   DGL_Translatef(0, yOffset, 0);
   DGL_Enable(DGL_TEXTURE_2D);
   DGL_SetPatch(pChainInfo.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
   DGL_Color4f(1, 1, 1, iconOpacity);
   dfloat const gemXOffset = 7 + de::roundf((w - 14) * healthPos) - pGemInfo.geometry.size.width/2;
   if(gemXOffset > 0) {
       dfloat cw = ( pChainInfo.geometry.size.width - gemXOffset ) / pChainInfo.geometry.size.width;
       DGL_Begin(DGL_QUADS);
           DGL_TexCoord2f(0, cw, 0);
           DGL_Vertex2f(x, y);
           DGL_TexCoord2f(0, 1, 0);
           DGL_Vertex2f(x + gemXOffset, y);
           DGL_TexCoord2f(0, 1, 1);
           DGL_Vertex2f(x + gemXOffset, y + h);
           DGL_TexCoord2f(0, cw, 1);
           DGL_Vertex2f(x, y + h);
       DGL_End();
   }
   if(gemXOffset + pGemInfo.geometry.size.width < w) {
       dfloat cw = (w - gemXOffset - pGemInfo.geometry.size.width) / pChainInfo.geometry.size.width;
       DGL_Begin(DGL_QUADS);
           DGL_TexCoord2f(0, 0, 0);
           DGL_Vertex2f(x + gemXOffset + pGemInfo.geometry.size.width, y);
           DGL_TexCoord2f(0, cw, 0);
           DGL_Vertex2f(x + w, y);
           DGL_TexCoord2f(0, cw, 1);
           DGL_Vertex2f(x + w, y + h);
           DGL_TexCoord2f(0, 0, 1);
           DGL_Vertex2f(x + gemXOffset + pGemInfo.geometry.size.width, y + h);
       DGL_End();
   }
   dint vX   = x + de::max(0.f, gemXOffset);
   dfloat s1 = 0, s2 = 1;
   dint vWidth = pGemInfo.geometry.size.width;
   if(gemXOffset + pGemInfo.geometry.size.width > w) {
       vWidth -= gemXOffset + pGemInfo.geometry.size.width - w;
       s2 = dfloat( vWidth ) / pGemInfo.geometry.size.width;
   }
   if(gemXOffset < 0) {
       vWidth -= -gemXOffset;
       s1 = dfloat( -gemXOffset ) / pGemInfo.geometry.size.width;
   }
   DGL_SetPatch(pGemInfo.id, DGL_CLAMP_TO_EDGE, DGL_CLAMP_TO_EDGE);
   DGL_Begin(DGL_QUADS);
       DGL_TexCoord2f(0, s1, 0);
       DGL_Vertex2f(vX, y);
       DGL_TexCoord2f(0, s2, 0);
       DGL_Vertex2f(vX + vWidth, y);
       DGL_TexCoord2f(0, s2, 1);
       DGL_Vertex2f(vX + vWidth, y + h);
       DGL_TexCoord2f(0, s1, 1);
       DGL_Vertex2f(vX, y + h);
   DGL_End();
   DGL_BlendMode(BM_ADD);
   DGL_Bind(Get(DD_DYNLIGHT_TEXTURE));
   DGL_Enable(DGL_TEXTURE_2D);
   dfloat rgb[3]; R_GetColorPaletteRGBf(0, theirColors[teamColor], rgb, false);
   DGL_DrawRectf2Color(x + gemXOffset + 23, y - 6, 41, 24, rgb[0], rgb[1], rgb[2], gemglow - (1 - iconOpacity));
   DGL_Color4f(1, 1, 1, 1);
   DGL_Disable(DGL_TEXTURE_2D);
   DGL_BlendMode(BM_NORMAL);
   DGL_MatrixMode(DGL_MODELVIEW);
   DGL_PopMatrix();
#undef ORIGINX
#undef ORIGINY
#else
    DENG2_UNUSED(offset);
#endif
}
void guidata_chain_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    Rect_SetWidthHeight(&geometry(), (ST_WIDTH - 21 - 28) * ::cfg.common.statusbarScale,
                                     8 * ::cfg.common.statusbarScale);
}
void guidata_chain_t::prepareAssets() {
#if __JHERETIC__
    ::pChain  = R_DeclarePatch("CHAIN");
    ::pGem[0] = R_DeclarePatch("LIFEGEM0");
    ::pGem[1] = R_DeclarePatch("LIFEGEM1");
    ::pGem[2] = R_DeclarePatch("LIFEGEM2");
    ::pGem[3] = R_DeclarePatch("LIFEGEM3");
#endif
#if __JHEXEN__
    ::pChain[PCLASS_FIGHTER] = R_DeclarePatch("CHAIN");
    ::pGem[PCLASS_FIGHTER][0] = R_DeclarePatch("LIFEGEM");
    for(dint i = 1; i < NUMTEAMS; ++i) {
        ::pGem[PCLASS_FIGHTER][i] = R_DeclarePatch(String("LIFEGMF%1").arg(i + 1).toUtf8().constData());
    }
    ::pChain[PCLASS_CLERIC] = R_DeclarePatch("CHAIN2");
    for(dint i = 0; i < NUMTEAMS; ++i) {
        ::pGem[PCLASS_CLERIC][i] = R_DeclarePatch(String("LIFEGMC%1").arg(i + 1).toUtf8().constData());
    }
    ::pChain[PCLASS_MAGE] = R_DeclarePatch("CHAIN3");
    for(dint i = 0; i < NUMTEAMS; ++i) {
        ::pGem[PCLASS_MAGE][i] = R_DeclarePatch(String("LIFEGMM%1").arg(i + 1).toUtf8().constData());
    }
#endif
}
static void ChatWidget_UpdateGeometry(ChatWidget *chat) {
    DENG2_ASSERT(chat);
    chat->updateGeometry();
}
static void ChatWidget_Draw(ChatWidget *chat, Point2Raw const *offset) {
    DENG2_ASSERT(chat);
    chat->draw(offset? Vector2i(offset->xy) : Vector2i());
}
DENG2_PIMPL(ChatWidget) {
    bool active      = false;
    bool shiftDown   = false;
    dint destination = 0;      ///< Default: All players.
    String text;
    Impl(Public *i) : Base(i) {}
    static void playSentSound() {
#if __JDOOM__
        S_LocalSound((gameModeBits & GM_ANY_DOOM2)? SFX_RADIO : SFX_TINK, 0);
#elif __JDOOM64__
        S_LocalSound(SFX_RADIO, 0);
#endif
    }
    void sendMessage() {
        String const msg = self().messageAsText();
        if(msg.isEmpty()) return;
        if(destination == 0) {
            if(!IS_NETGAME) {
                for(dint i = 0; i < MAXPLAYERS; ++i) {
                    D_NetMessageNoSound(i, msg.toUtf8().constData());
                }
            }
            else {
                String const cmd = String("chat %1").arg(msg.escaped());
                DD_Execute(false, cmd.toUtf8().constData());
            }
        }
        else {
            for(dint i = 0; i < MAXPLAYERS; ++i) {
                if(!players[i].plr->inGame) continue;
                if(destination != cfg.playerColor[i]) continue;
                if(!IS_NETGAME) {
                    D_NetMessageNoSound(i, msg.toUtf8().constData());
                }
                else {
                    String const cmd = String("chatnum %1 %2").arg(i).arg(msg.escaped());
                    DD_Execute(false, cmd.toUtf8().constData());
                }
            }
        }
        playSentSound();
    }
};
ChatWidget::ChatWidget(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(ChatWidget_UpdateGeometry),
                function_cast<DrawFunc>(ChatWidget_Draw),
                player)
    , d(new Impl(this))
{}
ChatWidget::~ChatWidget()
{}
bool ChatWidget::isActive() const {
    return d->active;
}
void ChatWidget::activate(bool yes) {
    bool const oldActive = isActive();
    if(d->active) {
        if(!yes) {
            d->active = false;
        }
    }
    else if(yes) {
        setDestination(0);  /// @c 0= "global" (default)
        d->text.clear();
        d->active = true;
    }
    if(d->active != oldActive) {
        DD_Executef(true, "%s chat", d->active? "activatebcontext" : "deactivatebcontext");
    }
}
dint ChatWidget::destination() const {
    return d->destination;
}
void ChatWidget::setDestination(dint newDestination) {
    if(newDestination >= 0 && newDestination <= NUMTEAMS) {
        d->destination = newDestination;
        return;
    }
    throw DestinationError("ChatWidget::setDestination", "Unknown destination #" + String::number(newDestination) + " (not changed)");
}
dint ChatWidget::handleEvent(event_t const &ev) {
    if(!isActive())
        return false;
    if(ev.type != EV_KEY)
        return false;
    if(ev.data1 == DDKEY_RSHIFT) {
        d->shiftDown = (ev.state == EVS_DOWN || ev.state == EVS_REPEAT);
        return false;  // Never eaten.
    }
    if(!(ev.state == EVS_DOWN || ev.state == EVS_REPEAT))
        return false;
    if(ev.data1 == DDKEY_BACKSPACE) {
        d->text.truncate(d->text.length() - 1);
        return true;
    }
    dint const oldLength = d->text.length();
    dchar const ch = ev.data1;
    if(ch >= ' ' && ch <= 'z') {
        d->text.append(QChar::fromLatin1(d->shiftDown ? ::shiftXForm[dint( ch )] : ch));
    }
    return oldLength != d->text.length();
}
dint ChatWidget::handleMenuCommand(menucommand_e cmd) {
    if(!isActive()) return false;
    switch(cmd) {
    case MCMD_SELECT:  // Send the message.
        if(!d->text.isEmpty()) {
            d->sendMessage();
        }
        activate(false);
        return true;
    case MCMD_CLOSE:
    case MCMD_NAV_OUT: // Close chat.
        activate(false);
        return true;
    case MCMD_DELETE:
        d->text.truncate(d->text.length() - 1);
        return true;
    default: return false;
    }
}
void ChatWidget::messageClear() {
    d->text.clear();
}
void ChatWidget::messageAppend(String const &str) {
    d->text += str;
}
String ChatWidget::messageAsText() const {
    return d->text;
}
void ChatWidget::draw(Vector2i const &offset) const {
    dfloat const textOpacity = uiRendState->pageAlpha * cfg.common.hudColor[3];
    if(!isActive()) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(cfg.common.msgScale, cfg.common.msgScale, 1);
    FR_SetFont(font());
    FR_SetColorAndAlpha(cfg.common.hudColor[0], cfg.common.hudColor[1], cfg.common.hudColor[2], textOpacity);
    Block const text       = messageAsText().toUtf8();
    dint const textWidth   = FR_TextWidth(text.constData());
    dint const cursorWidth = FR_CharWidth('_');
    dint xOffset = 0;
    if(cfg.common.msgAlign == 1)
        xOffset = -(textWidth + cursorWidth)/2;
    else if(cfg.common.msgAlign == 2)
        xOffset = -(textWidth + cursorWidth);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_DrawTextXY(text.constData(), xOffset, 0);
    if(actualMapTime & 12) {
        FR_DrawCharXY('_', xOffset + textWidth, 0);
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void ChatWidget::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!isActive()) return;
    FR_SetFont(font());
    Block const text = messageAsText().toUtf8();
    Size2Raw cursorSize; FR_CharSize(&cursorSize, '_');
    Size2Raw textSize;   FR_TextSize(&textSize, text.constData());
    Rect_SetWidthHeight(&geometry(), cfg.common.msgScale * (textSize.width + cursorSize.width),
                                     cfg.common.msgScale * (de::max(textSize.height, cursorSize.height)));
}
void ChatWidget::loadMacros()  // static
{
    for(dint i = 0; i < 10; ++i) {
        if(cfg.common.chatMacros[i]) continue;
        cfg.common.chatMacros[i] = (char *) GET_TXT(TXT_HUSTR_CHATMACRO0 + i);
    }
}
String ChatWidget::findMacro(dint macroId)  // static
{
    if(macroId >= 0 && macroId < 10) {
        return String(cfg.common.chatMacros[macroId]);
    }
    return "";
}
void ChatWidget::consoleRegister() {
    C_VAR_BYTE("chat-beep", &cfg.common.chatBeep, 0, 0, 1);
    for(dint i = 0; i < 10; ++i) {
        Block const cvarname = (String("chat-macro%1").arg(i)).toUtf8();
        C_VAR_CHARPTR(cvarname.constData(), &cfg.common.chatMacros[i], 0, 0, 0);
    }
}
static void FlightWidget_UpdateGeometry(guidata_flight_t *flht) {
    DENG2_ASSERT(flht);
    flht->updateGeometry();
}
static void FlightWidget_Draw(guidata_flight_t *flht, Point2Raw const *offset) {
    DENG2_ASSERT(flht);
    flht->draw(offset? Vector2i(offset->xy) : Vector2i());
}
guidata_flight_t::guidata_flight_t(de::dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(FlightWidget_UpdateGeometry),
                function_cast<DrawFunc>(FlightWidget_Draw),
                player)
{}
#if __JHERETIC__ || __JHEXEN__
static patchid_t pIcon[16];
#endif
guidata_flight_t::~guidata_flight_t()
{}
void guidata_flight_t::reset() {
    _patchId        = 0;
    _hitCenterFrame = false;
}
void guidata_flight_t::tick(timespan_t /*elapsed*/)
{
#if __JHERETIC__ || __JHEXEN__
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    _patchId = 0;
    player_t const *plr = &::players[player()];
    if(plr->powers[PT_FLIGHT] <= 0) return;
    if(plr->powers[PT_FLIGHT] > BLINKTHRESHOLD || !(plr->powers[PT_FLIGHT] & 16)) {
        dint frame = (::mapTime / 3) & 15;
        if(plr->plr->mo->flags2 & MF2_FLY) {
            if(_hitCenterFrame && (frame != 15 && frame != 0))
                frame = 15;
            else
                _hitCenterFrame = false;
        }
        else {
            if(!_hitCenterFrame && (frame != 15 && frame != 0)) {
                _hitCenterFrame = false;
            }
            else {
                frame = 15;
                _hitCenterFrame = true;
            }
        }
        _patchId = ::pIcon[frame];
    }
#endif
}
void guidata_flight_t::draw(Vector2i const &offset) const {
#if __JHERETIC__ || __JHEXEN__
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(_patchId) {
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        DGL_Translatef(offset.x, offset.y, 0);
        DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(_patchId, Vector2i(16, 14));
        DGL_Disable(DGL_TEXTURE_2D);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
    }
#else
    DENG2_UNUSED(offset);
#endif
}
void guidata_flight_t::updateGeometry() {
#if __JHERETIC__ || __JHEXEN__
#  if __JHERETIC__
#  define HEIGHT                ( 32 )
#  else
#  define HEIGHT                ( 28 )
#  endif
    player_t const *plr = &::players[player()];
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(plr->powers[PT_FLIGHT] <= 0) return;
    Rect_SetWidthHeight(&geometry(), 32 * ::cfg.common.hudScale, HEIGHT * ::cfg.common.hudScale);
#endif
}
void guidata_flight_t::prepareAssets()  // static
{
#if __JHERETIC__ || __JHEXEN__
    for(dint i = 0; i < 16; ++i) {
        ::pIcon[i] = R_DeclarePatch(String("SPFLY%1").arg(i).toUtf8().constData());
    }
#endif
}
guidata_frags_t::guidata_frags_t(void (*updateGeometry) (HudWidget *wi),
                                 void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                 dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_frags_t::~guidata_frags_t()
{}
void guidata_frags_t::reset() {
    _value = 1994;
}
void guidata_frags_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    _value = 0;
    for(dint i = 0; i < MAXPLAYERS; ++i) {
        if(::players[i].plr->inGame) {
            _value += plr->frags[i] * (i != player() ? 1 : -1);
        }
    }
}
void FragsWidget_Draw(guidata_frags_t *frags, Point2Raw const *offset) {
#if __JDOOM__ || __JDOOM64__
#  define TRACKING              ( 0 )
#else
#  define TRACKING              ( 1 )
#endif
    DENG2_ASSERT(frags);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
#if __JDOOM__ || __JDOOM64__
    if(!::cfg.hudShown[HUD_FRAGS]) return;
#endif
    if(!gfw_Rule(deathmatch)) return;
    if(ST_AutomapIsOpen(frags->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[frags->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(frags->_value == 1994) return;
    auto const valueAsText = String("Frags: %1").arg(frags->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(frags->font());
    FR_SetTracking(TRACKING);
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, 0);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
}
void FragsWidget_UpdateGeometry(guidata_frags_t *frags) {
#if __JDOOM__ || __JDOOM64__
#  define TRACKING              ( 0 )
#else
#  define TRACKING              ( 1 )
#endif
    DENG2_ASSERT(frags);
    Rect_SetWidthHeight(&frags->geometry(), 0, 0);
#if __JDOOM__ || __JDOOM64__
    if(!::cfg.hudShown[HUD_FRAGS]) return;
#endif
    if(!gfw_Rule(deathmatch)) return;
    if(ST_AutomapIsOpen(frags->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[frags->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(frags->_value == 1994) return;
    auto const valueAsText = String("Frags: %1").arg(frags->_value);
    FR_SetFont(frags->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    textSize.width  *= ::cfg.common.hudScale;
    textSize.height *= ::cfg.common.hudScale;
    Rect_SetWidthHeight(&frags->geometry(), textSize.width, textSize.height);
#undef TRACKING
}
void SBarFragsWidget_Draw(guidata_frags_t *frags, Point2Raw const *offset) {
#define X_ORIGIN                (-ST_WIDTH / 2 )
#define Y_ORIGIN                (-ST_HEIGHT )
#if __JDOOM__
#  define X_OFFSET              ( 138 )
#  define Y_OFFSET              (   3 )
#elif __JHERETIC__
#  define X_OFFSET              (  85 )
#  define Y_OFFSET              (  13 )
#elif __JHEXEN__
#  define X_OFFSET              (  64 )
#  define Y_OFFSET              (  14 )
#else
#  define X_OFFSET              (   0 )
#  define Y_OFFSET              (   0 )
#endif
#if __JHERETIC__
#  define TRACKING              ( 1 )
#else
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(frags);
    dint const activeHud     = ST_ActiveHud(frags->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(frags->player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(!gfw_Rule(deathmatch)) return;
#if __JHERETIC__ || __JHEXEN__
    if(Hu_InventoryIsOpen(frags->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(frags->player())) return;
#else
    if(ST_AutomapIsOpen(frags->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[frags->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(frags->_value == 1994) return;
    auto const valueAsText = String::number(frags->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(frags->font());
    FR_SetTracking(TRACKING);
#if __JDOOM__
    if(gameMode == doom_chex) {
        FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    }
    else
#endif
    {
        FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    }
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), X_ORIGIN + X_OFFSET, Y_ORIGIN + Y_OFFSET,
                   ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
#undef Y_OFFSET
#undef X_OFFSET
#undef Y_ORIGIN
#undef X_ORIGIN
}
void SBarFragsWidget_UpdateGeometry(guidata_frags_t *frags) {
#if __JHERETIC__
#  define TRACKING                ( 1 )
#else
#  define TRACKING                ( 0 )
#endif
    DENG2_ASSERT(frags);
    Rect_SetWidthHeight(&frags->geometry(), 0, 0);
    if(!gfw_Rule(deathmatch)) return;
#if __JHERETIC__ || __JHEXEN__
    if(Hu_InventoryIsOpen(frags->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(frags->player())) return;
#else
    if(ST_AutomapIsOpen(frags->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[frags->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(frags->_value == 1994) return;
    auto const valueAsText = String::number(frags->_value);
    FR_SetFont(frags->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    Rect_SetWidthHeight(&frags->geometry(), textSize.width  * ::cfg.common.statusbarScale,
                                            textSize.height * ::cfg.common.statusbarScale);
#undef TRACKING
}
static void GroupWidget_UpdateGeometry(GroupWidget *group) {
    DENG2_ASSERT(group);
    group->updateGeometry();
}
DENG2_PIMPL_NOREF(GroupWidget) {
    order_t order = ORDER_NONE;  ///< Order of child objects.
    dint flags = 0;              ///< @ref groupWidgetFlags
    dint padding = 0;
    QVector<uiwidgetid_t> children;
};
GroupWidget::GroupWidget(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(GroupWidget_UpdateGeometry), nullptr, player)
    , d(new Impl) {
    setPlayer(player);
}
GroupWidget::~GroupWidget()
{}
de::dint GroupWidget::childCount() const {
    return d->children.count();
}
void GroupWidget::addChild(HudWidget *other) {
    if(!other || other == this) return;
    if(!d->children.contains(other->id())) {
        d->children << other->id();
    }
}
LoopResult GroupWidget::forAllChildren(std::function<LoopResult (HudWidget &)> func) const {
    for(uiwidgetid_t const &childId : d->children) {
        if(auto result = func(GUI_FindWidgetById(childId)))
            return result;
    }
    return LoopContinue;
}
void GroupWidget::clearAllChildren() {
    d->children.clear();
}
dint GroupWidget::flags() const {
    return d->flags;
}
void GroupWidget::setFlags(dint newFlags) {
    d->flags = newFlags;
}
order_t GroupWidget::order() const {
    return d->order;
}
void GroupWidget::setOrder(order_t newOrder) {
    d->order = newOrder;
}
dint GroupWidget::padding() const {
    return d->flags;
}
void GroupWidget::setPadding(dint newPadding) {
    d->padding = newPadding;
}
void GroupWidget::tick(timespan_t elapsed) {
    for(uiwidgetid_t const &childId : d->children) {
        GUI_FindWidgetById(childId).tick(elapsed);
    }
}
static void applyAlignmentOffset(HudWidget *wi, dint *x, dint *y) {
    DENG2_ASSERT(wi);
    if(x) {
        if(wi->alignment() & ALIGN_RIGHT)
            *x += wi->maximumWidth();
        else if(!(wi->alignment() & ALIGN_LEFT))
            *x += wi->maximumWidth() / 2;
    }
    if(y) {
        if(wi->alignment() & ALIGN_BOTTOM)
            *y += wi->maximumHeight();
        else if(!(wi->alignment() & ALIGN_TOP))
            *y += wi->maximumHeight() / 2;
    }
}
void GroupWidget::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!childCount()) return;
    dint x = 0, y = 0;
    applyAlignmentOffset(this, &x, &y);
    for(uiwidgetid_t const &childId : d->children) {
        HudWidget &child = GUI_FindWidgetById(childId);
        if(child.maximumWidth() > 0 && child.maximumHeight() > 0 && child.opacity() > 0) {
            GUI_UpdateWidgetGeometry(&child);
            Rect_SetX(&child.geometry(), Rect_X(&child.geometry()) + x);
            Rect_SetY(&child.geometry(), Rect_Y(&child.geometry()) + y);
            Rect const *childGeometry = &child.geometry();
            if(Rect_Width(childGeometry) > 0 && Rect_Height(childGeometry) > 0) {
                if(d->order == ORDER_RIGHTTOLEFT) {
                    if(!(d->flags & UWGF_VERTICAL))
                        x -= Rect_Width(childGeometry)  + d->padding;
                    else
                        y -= Rect_Height(childGeometry) + d->padding;
                }
                else if(d->order == ORDER_LEFTTORIGHT) {
                    if(!(d->flags & UWGF_VERTICAL))
                        x += Rect_Width(childGeometry)  + d->padding;
                    else
                        y += Rect_Height(childGeometry) + d->padding;
                }
                Rect_Unite(&geometry(), childGeometry);
            }
        }
    }
}
guidata_health_t::guidata_health_t(void (*updateGeometry) (HudWidget *wi),
                                   void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                   dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_health_t::~guidata_health_t()
{}
void guidata_health_t::reset() {
    _value = 1994;
}
void guidata_health_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    _value = plr->health;
}
void HealthWidget_Draw(guidata_health_t *hlth, Point2Raw const *offset) {
#if __JDOOM__
#  define X_OFFSET              ( 0 )
#  define Y_OFFSET              ( 0 )
#  define TRACKING              ( 0 )
#  define SUFFIX                ( "%" )
#else
#  define X_OFFSET              (-1 )
#  define Y_OFFSET              (-1 )
#  define TRACKING              ( 1 )
#  define SUFFIX                ( "" )
#endif
    DENG2_ASSERT(hlth);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(hlth->_value == 1994) return;
    if(!::cfg.hudShown[HUD_HEALTH]) return;
    if(ST_AutomapIsOpen(hlth->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[hlth->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(hlth->_value) + SUFFIX;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(hlth->font());
    FR_SetTracking(TRACKING);
#if __JHERETIC__
    FR_SetColorAndAlpha(0, 0, 0, textOpacity * .4f);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), X_OFFSET + 2, Y_OFFSET + 2);
#endif
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), X_OFFSET, Y_OFFSET);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef SUFFIX
#undef TRACKING
#undef Y_OFFSET
#undef X_OFFSET
}
void SBarHealthWidget_Draw(guidata_health_t *hlth, Point2Raw const *offset) {
#if __JDOOM__
#  define X_OFFSET              ( 90 )
#  define Y_OFFSET              (  3 )
#elif __JHERETIC__
#  define X_OFFSET              ( 85 )
#  define Y_OFFSET              ( 12 )
#elif __JHEXEN__
#  define X_OFFSET              ( 64 )
#  define Y_OFFSET              ( 14 )
#else
#  define X_OFFSET              (  0 )
#  define Y_OFFSET              (  0 )
#endif
#if __JHERETIC__
#  define TRACKING              ( 1 )
#else
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(hlth);
    Vector2i const origin((-ST_WIDTH / 2 ) + X_OFFSET, (-ST_HEIGHT ) + Y_OFFSET );
    if(hlth->_value == 1994) return;
    dint const activeHud     = ST_ActiveHud(hlth->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(hlth->player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
#if __JHERETIC__ || __JHEXEN__
    if(gfw_Rule(deathmatch)) return;
    if(Hu_InventoryIsOpen(hlth->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(hlth->player())) return;
#else
    if(ST_AutomapIsOpen(hlth->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[hlth->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(hlth->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(hlth->font());
    FR_SetTracking(TRACKING);
#if __JDOOM__
    if(::gameMode == doom_chex) {
        FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    }
    else
#endif
    {
        FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    }
    FR_DrawTextXY3(valueAsText.toUtf8().constData(), origin.x, origin.y, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
#if __JDOOM__
    FR_DrawCharXY('%', origin.x, origin.y);
#endif
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
#undef Y_OFFSET
#undef X_OFFSET
}
void HealthWidget_UpdateGeometry(guidata_health_t *hlth) {
#if __JDOOM__
#  define TRACKING              ( 0 )
#  define SUFFIX                ( "%" )
#else
#  define TRACKING              ( 1 )
#  define SUFFIX                ( "" )
#endif
    DENG2_ASSERT(hlth);
    Rect_SetWidthHeight(&hlth->geometry(), 0, 0);
    if(hlth->_value == 1994) return;
    if(!::cfg.hudShown[HUD_HEALTH]) return;
    if(ST_AutomapIsOpen(hlth->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[hlth->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(hlth->_value) + SUFFIX;
    FR_SetFont(hlth->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    textSize.width  *= ::cfg.common.hudScale;
    textSize.height *= ::cfg.common.hudScale;
    Rect_SetWidthHeight(&hlth->geometry(), textSize.width, textSize.height);
#undef SUFFIX
#undef TRACKING
}
void SBarHealthWidget_UpdateGeometry(guidata_health_t *hlth) {
#if __JHERETIC__
#  define TRACKING              ( 1 )
#else
#  define TRACKING              ( 0 )
#endif
    DENG2_ASSERT(hlth);
    Rect_SetWidthHeight(&hlth->geometry(), 0, 0);
    if(hlth->_value == 1994) return;
#if __JHERETIC__ || __JHEXEN__
    if(gfw_Rule(deathmatch)) return;
    if(Hu_InventoryIsOpen(hlth->player())) return;
#endif
#if __JHEXEN__
    if(ST_AutomapIsOpen(hlth->player())) return;
#else
    if(ST_AutomapIsOpen(hlth->player()) && ::cfg.common.automapHudDisplay == 0) return;
#endif
    if(P_MobjIsCamera(::players[hlth->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = String::number(hlth->_value);
    FR_SetFont(hlth->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
#if __JDOOM__
    textSize.width += FR_CharWidth('%');
    textSize.height = de::max(textSize.height, FR_CharHeight('%'));
#endif
    Rect_SetWidthHeight(&hlth->geometry(), textSize.width  * ::cfg.common.statusbarScale,
                                           textSize.height * ::cfg.common.statusbarScale);
#undef TRACKING
}
static void ItemsWidget_Draw(guidata_items_t *items, Point2Raw const *offset) {
    DENG2_ASSERT(items);
    items->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void ItemsWidget_UpdateGeometry(guidata_items_t *items) {
    DENG2_ASSERT(items);
    items->updateGeometry();
}
guidata_items_t::guidata_items_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(ItemsWidget_UpdateGeometry),
                function_cast<DrawFunc>(ItemsWidget_Draw),
                player)
{}
guidata_items_t::~guidata_items_t()
{}
void guidata_items_t::reset() {
    _value = 1994;
}
void guidata_items_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &players[player()];
    _value = plr->itemCount;
}
void guidata_items_t::draw(Vector2i const &offset) const {
#if !__JHEXEN__
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(!(::cfg.common.hudShownCheatCounters & (CCH_ITEMS | CCH_ITEMS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    if(_value == 1994) return;
    String valueAsText("Items:");
    if(cfg.common.hudShownCheatCounters & CCH_ITEMS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalItems);
    }
    if(cfg.common.hudShownCheatCounters & CCH_ITEMS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((::cfg.common.hudShownCheatCounters & CCH_ITEMS) ? "(" : "")
                           .arg(totalItems ? _value * 100 / totalItems : 100)
                           .arg((::cfg.common.hudShownCheatCounters & CCH_ITEMS) ? ")" : "");
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.hudCheatCounterScale, ::cfg.common.hudCheatCounterScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(font());
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, 0);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#else  // !__JHEXEN__
    DENG2_UNUSED(offset);
#endif
}
void guidata_items_t::updateGeometry() {
#if !__JHEXEN__
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!(::cfg.common.hudShownCheatCounters & (CCH_ITEMS | CCH_ITEMS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    if(_value == 1994) return;
    String valueAsText("Items:");
    if(cfg.common.hudShownCheatCounters & CCH_ITEMS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalItems);
    }
    if(cfg.common.hudShownCheatCounters & CCH_ITEMS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((::cfg.common.hudShownCheatCounters & CCH_ITEMS) ? "(" : "")
                           .arg(totalItems ? _value * 100 / totalItems : 100)
                           .arg((::cfg.common.hudShownCheatCounters & CCH_ITEMS) ? ")" : "");
    }
    FR_SetFont(font());
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    Rect_SetWidthHeight(&geometry(), .5f + textSize.width  * ::cfg.common.hudCheatCounterScale,
                                     .5f + textSize.height * ::cfg.common.hudCheatCounterScale);
#endif  // !__JHEXEN__
}
static patchid_t pKeySlots[NUM_KEY_TYPES];
static void KeySlotWidget_UpdateGeometry(guidata_keyslot_t *kslt) {
    DENG2_ASSERT(kslt);
    kslt->updateGeometry();
}
static void KeySlotWidget_Draw(guidata_keyslot_t *kslt, Point2Raw const *offset) {
    DENG2_ASSERT(kslt);
    kslt->draw(offset? Vector2i(offset->xy) : Vector2i());
}
DENG2_PIMPL_NOREF(guidata_keyslot_t) {
    dint slotNum = 0;
    keytype_t keyTypeA = keytype_t(0);
#if __JDOOM__
    keytype_t keyTypeB = keytype_t(0);
#endif
    patchid_t patchId  = -1;
#if __JDOOM__
    patchid_t patchId2 = -1;
#endif
};
guidata_keyslot_t::guidata_keyslot_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(KeySlotWidget_UpdateGeometry),
                function_cast<DrawFunc>(KeySlotWidget_Draw),
                player)
    , d(new Impl)
{}
guidata_keyslot_t::~guidata_keyslot_t()
{}
void guidata_keyslot_t::reset() {
    d->patchId  = -1;
#if __JDOOM__
    d->patchId2 = -1;
#endif
}
guidata_keyslot_t &guidata_keyslot_t::setSlot(dint newSlotNum) {
    d->slotNum  = newSlotNum;
    d->keyTypeA = keytype_t( newSlotNum );
#if __JDOOM__
    d->keyTypeB = keytype_t( newSlotNum + 3 );
#endif
    return *this;
}
void guidata_keyslot_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &players[player()];
#if __JDOOM__
    d->patchId = -1;
    if(plr->keys[d->keyTypeA] || plr->keys[d->keyTypeB]) {
        d->patchId = ::pKeySlots[plr->keys[d->keyTypeB]? d->keyTypeB : d->keyTypeA];
    }
    d->patchId2 = -1;
    if(!cfg.hudKeysCombine && plr->keys[d->keyTypeA] && plr->keys[d->keyTypeB]) {
        d->patchId2 = ::pKeySlots[d->keyTypeA];
    }
#elif __JHEXEN__
    d->patchId = (plr->keys & (1 << d->keyTypeA)) ? pKeySlots[d->keyTypeA] : -1;
#else
    d->patchId = plr->keys[d->keyTypeA] ? pKeySlots[d->keyTypeA] : -1;
#endif
}
void guidata_keyslot_t::draw(Vector2i const &offset) const {
    Vector2i const origin(-ST_WIDTH / 2, -ST_HEIGHT);
    dint const activeHud     = ST_ActiveHud(player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(player()));
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
#if __JHERETIC__ || __JHEXEN__
    if(Hu_InventoryIsOpen(player())) return;
#endif
#if __JDOOM__
    if(d->patchId <= 0 && d->patchId2 <= 0) return;
#else
    if(d->patchId <= 0) return;
#endif
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
#if __JDOOM__
    Vector2i const comboOffset((d->patchId2 > 0? -1 : 0), (d->patchId2 > 0? -1 : 0));
    GL_DrawPatch(d->patchId, origin + Vector2i(239, 3 + 10 * d->slotNum) + comboOffset);
    if(d->patchId2 > 0) {
        GL_DrawPatch(d->patchId2, origin + Vector2i(239, 3 + 10 * d->slotNum) - comboOffset);
    }
#else
    GL_DrawPatch(d->patchId, origin + Vector2i(153, 6 + 8 * d->keyTypeA));
#endif
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void guidata_keyslot_t::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
#if __JHERETIC__
    if(Hu_InventoryIsOpen(player())) return;
#endif
    if(ST_AutomapIsOpen(player()) && cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
#if __JDOOM__
    if(d->patchId <= 0 && d->patchId2 <= 0) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(d->patchId, &info)) return;
    Rect_SetWidthHeight(&geometry(), info.geometry.size.width,
                                     info.geometry.size.height);
    if(d->patchId2 > 0 && R_GetPatchInfo(d->patchId, &info)) {
        info.geometry.origin.x = info.geometry.origin.y = 2; // Combine offset.
        Rect_UniteRaw(&geometry(), &info.geometry);
    }
    Rect_SetWidthHeight(&geometry(), Rect_Width (&geometry()) * cfg.common.statusbarScale,
                                     Rect_Height(&geometry()) * cfg.common.statusbarScale);
#else
    patchinfo_t info;
    if(d->patchId <= 0 || !R_GetPatchInfo(d->patchId, &info)) return;
    Rect_SetWidthHeight(&geometry(), info.geometry.size.width  * cfg.common.statusbarScale,
                                     info.geometry.size.height * cfg.common.statusbarScale);
#endif
}
void guidata_keyslot_t::prepareAssets() {
#if __JHERETIC__
    dint n = 0;
    ::pKeySlots[/*KT_YELLOW*/ n++] = R_DeclarePatch("YKEYICON");
    ::pKeySlots[/*KT_GREEN*/  n++] = R_DeclarePatch("GKEYICON");
    ::pKeySlots[/*KT_BLUE*/   n++] = R_DeclarePatch("BKEYICON");
    DENG2_ASSERT(n == NUM_KEY_TYPES);
#else
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
        ::pKeySlots[i] = R_DeclarePatch(String("STKEYS%1").arg(i).toUtf8().constData());
    }
#endif
}
static void KeysWidget_UpdateGeometry(guidata_keys_t *keys) {
    DENG2_ASSERT(keys);
    keys->updateGeometry();
}
static void KeysWidget_Draw(guidata_keys_t *keys, Point2Raw const *offset) {
    DENG2_ASSERT(keys);
    keys->draw(offset? Vector2i(offset->xy) : Vector2i());
}
guidata_keys_t::guidata_keys_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(KeysWidget_UpdateGeometry),
                function_cast<DrawFunc>(KeysWidget_Draw),
                player)
{}
guidata_keys_t::~guidata_keys_t()
{}
void guidata_keys_t::reset() {
    de::zap(_keyBoxes);
}
void guidata_keys_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
#if __JHEXEN__
        _keyBoxes[i] = (plr->keys & (1 << i));
#else
        _keyBoxes[i] = CPP_BOOL(plr->keys[i]);
#endif
    }
}
void guidata_keys_t::draw(Vector2i const &offset) const {
#if __JDOOM__
#define EXTRA_SCALE         ( .75f )
    static dint const keyPairs[3][2] = {
        { KT_REDCARD,    KT_REDSKULL },
        { KT_YELLOWCARD, KT_YELLOWSKULL },
        { KT_BLUECARD,   KT_BLUESKULL }
    };
    static dint const keySprites[NUM_KEY_TYPES] = {
        SPR_BKEY, SPR_YKEY, SPR_RKEY, SPR_BSKU, SPR_YSKU, SPR_RSKU
    };
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_KEYS]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(EXTRA_SCALE * ::cfg.common.hudScale, EXTRA_SCALE * ::cfg.common.hudScale, 1);
    dint x = 0;
    dint numDrawnKeys = 0;
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
        if(!_keyBoxes[i]) continue;
        bool shown = true;
        if(::cfg.hudKeysCombine) {
            for(dint k = 0; k < 3; ++k) {
                if(keyPairs[k][0] == i && _keyBoxes[keyPairs[k][0]] && _keyBoxes[keyPairs[k][1]]) {
                    shown = false;
                    break;
                }
            }
        }
        if(shown) {
            dint w, h;
            GUI_DrawSprite(keySprites[i], x, 0, HOT_TLEFT, 1, iconOpacity, false, &w, &h);
            x += w + 2;
            numDrawnKeys += 1;
        }
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef EXTRA_SCALE
#elif __JHERETIC__  // __JDOOM__
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_KEYS]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t pInfo;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    dint x = 0;
    if(_keyBoxes[KT_YELLOW]) {
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(::pKeys[0], Vector2i(x, 0), ALIGN_TOPLEFT, DPF_NO_OFFSET);
        DGL_Disable(DGL_TEXTURE_2D);
        if(R_GetPatchInfo(::pKeys[0], &pInfo))
            x += pInfo.geometry.size.width + 1;
    }
    if(_keyBoxes[KT_GREEN]) {
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(::pKeys[1], Vector2i(x, 0), ALIGN_TOPLEFT, DPF_NO_OFFSET);
        DGL_Disable(DGL_TEXTURE_2D);
        if(R_GetPatchInfo(::pKeys[1], &pInfo))
            x += pInfo.geometry.size.width + 1;
    }
    if(_keyBoxes[KT_BLUE]) {
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(::pKeys[2], Vector2i(x, 0), ALIGN_TOPLEFT, DPF_NO_OFFSET);
        DGL_Disable(DGL_TEXTURE_2D);
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#elif __JHEXEN__  // __JHERETIC__
#define ORIGINX             (-ST_WIDTH / 2 )
#define ORIGINY             (-ST_HEIGHT * ST_StatusBarShown(player()) )
    dint const activeHud     = ST_ActiveHud(player());
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(Hu_InventoryIsOpen(player()) || !ST_AutomapIsOpen(player())) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    dint numDrawn = 0;
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
        if(!_keyBoxes[i]) continue;
        patchid_t patch = ::pKeys[i];
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(patch, Vector2i(ORIGINX + 46 + numDrawn * 20, ORIGINY + 1));
        DGL_Disable(DGL_TEXTURE_2D);
        numDrawn += 1;
        if(numDrawn == 5) break;
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef ORIGINY
#undef ORIGINX
#else  // __JHEXEN__
    DENG2_UNUSED(offset);
#endif
}
void guidata_keys_t::updateGeometry() {
#if __JDOOM__
#define EXTRA_SCALE         ( .75f )
    static dint const keyPairs[3][2] = {
        { KT_REDCARD,    KT_REDSKULL },
        { KT_YELLOWCARD, KT_YELLOWSKULL },
        { KT_BLUECARD,   KT_BLUESKULL }
    };
    static dint const keySprites[NUM_KEY_TYPES] = {
        SPR_BKEY, SPR_YKEY, SPR_RKEY, SPR_BSKU, SPR_YSKU, SPR_RSKU
    };
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_KEYS]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    RectRaw iconGeometry{};
    for(int i = 0; i < NUM_KEY_TYPES; ++i) {
        if(!_keyBoxes[i]) continue;
        bool shown = true;
        if(cfg.hudKeysCombine) {
            for(dint k = 0; k < 3; ++k) {
                if(keyPairs[k][0] == i && _keyBoxes[keyPairs[k][0]] && _keyBoxes[keyPairs[k][1]]) {
                    shown = false;
                    break;
                }
            }
        }
        if(shown) {
            GUI_SpriteSize(keySprites[i], 1, &iconGeometry.size.width, &iconGeometry.size.height);
            Rect_UniteRaw(&geometry(), &iconGeometry);
            iconGeometry.origin.x += iconGeometry.size.width + 2;
        }
    }
    Rect_SetWidthHeight(&geometry(), Rect_Width (&geometry()) * EXTRA_SCALE * ::cfg.common.hudScale,
                                     Rect_Height(&geometry()) * EXTRA_SCALE * ::cfg.common.hudScale);
#undef EXTRA_SCALE
#elif __JHERETIC__  // __JDOOM__
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_KEYS]) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t pInfo;
    dint x = 0;
    if(_keyBoxes[KT_YELLOW] && R_GetPatchInfo(::pKeys[0], &pInfo)) {
        pInfo.geometry.origin.x = x;
        pInfo.geometry.origin.y = 0;
        Rect_UniteRaw(&geometry(), &pInfo.geometry);
        x += pInfo.geometry.size.width + 1;
    }
    if(_keyBoxes[KT_GREEN] && R_GetPatchInfo(::pKeys[1], &pInfo)) {
        pInfo.geometry.origin.x = x;
        pInfo.geometry.origin.y = 0;
        Rect_UniteRaw(&geometry(), &pInfo.geometry);
        x += pInfo.geometry.size.width + 1;
    }
    if(_keyBoxes[KT_BLUE] && R_GetPatchInfo(::pKeys[2], &pInfo)) {
        pInfo.geometry.origin.x = x;
        pInfo.geometry.origin.y = 0;
        Rect_UniteRaw(&geometry(), &pInfo.geometry);
    }
    Rect_SetWidthHeight(&geometry(), Rect_Width (&geometry()) * ::cfg.common.hudScale,
                                     Rect_Height(&geometry()) * ::cfg.common.hudScale);
#elif __JHEXEN__  // __JHERETIC__
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(Hu_InventoryIsOpen(player()) || !ST_AutomapIsOpen(player())) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t pInfo;
    dint x = 0;
    dint numVisible = 0;
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
        if(!_keyBoxes[i]) continue;
        patchid_t patch = ::pKeys[i];
        if(!R_GetPatchInfo(patch, &pInfo)) continue;
        pInfo.geometry.origin.x = x;
        pInfo.geometry.origin.y = 0;
        Rect_UniteRaw(&geometry(), &pInfo.geometry);
        numVisible += 1;
        if(numVisible == 5) break;
        x += 20;
    }
    Rect_SetWidthHeight(&geometry(), Rect_Width (&geometry()) * ::cfg.common.statusbarScale,
                                     Rect_Height(&geometry()) * ::cfg.common.statusbarScale);
#endif  // __JHEXEN__
}
void guidata_keys_t::prepareAssets()  // static
{
#if __JHERETIC__
    pKeys[0] = R_DeclarePatch("YKEYICON");
    pKeys[1] = R_DeclarePatch("GKEYICON");
    pKeys[2] = R_DeclarePatch("BKEYICON");
#elif __JHEXEN__
    for(dint i = 0; i < NUM_KEY_TYPES; ++i) {
        pKeys[i] = R_DeclarePatch(String("KEYSLOT%1").arg(i + 1, 0, 16).toUpper().toUtf8().constData());
    }
#endif
}
static void KillsWidget_Draw(guidata_kills_t *kills, Point2Raw const *offset) {
    DENG2_ASSERT(kills);
    kills->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void KillsWidget_UpdateGeometry(guidata_kills_t *kills) {
    DENG2_ASSERT(kills);
    kills->updateGeometry();
}
guidata_kills_t::guidata_kills_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(KillsWidget_UpdateGeometry),
                function_cast<DrawFunc>(KillsWidget_Draw),
                player)
{}
guidata_kills_t::~guidata_kills_t()
{}
void guidata_kills_t::reset() {
    _value = 1994;
}
void guidata_kills_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &players[player()];
    _value = plr->killCount;
}
void guidata_kills_t::draw(Vector2i const &offset) const {
#if !__JHEXEN__
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(!(::cfg.common.hudShownCheatCounters & (CCH_KILLS | CCH_KILLS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    if(_value == 1994) return;
    String valueAsText("Kills:");
    if(::cfg.common.hudShownCheatCounters & CCH_KILLS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalKills);
    }
    if(::cfg.common.hudShownCheatCounters & CCH_KILLS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((::cfg.common.hudShownCheatCounters & CCH_KILLS) ? "(" : "")
                           .arg(totalKills ? _value * 100 / totalKills : 100)
                           .arg((::cfg.common.hudShownCheatCounters & CCH_KILLS) ? ")" : "");
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.hudCheatCounterScale, ::cfg.common.hudCheatCounterScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(font());
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, 0);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#else  // !__JHEXEN__
    DENG2_UNUSED(offset);
#endif
}
void guidata_kills_t::updateGeometry() {
#if !__JHEXEN__
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(!(::cfg.common.hudShownCheatCounters & (CCH_KILLS | CCH_KILLS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    if(_value == 1994) return;
    String valueAsText("Kills:");
    if(::cfg.common.hudShownCheatCounters & CCH_KILLS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalKills);
    }
    if(::cfg.common.hudShownCheatCounters & CCH_KILLS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((::cfg.common.hudShownCheatCounters & CCH_KILLS) ? "(" : "")
                           .arg(totalKills ? _value * 100 / totalKills : 100)
                           .arg((::cfg.common.hudShownCheatCounters & CCH_KILLS) ? ")" : "");
    }
    FR_SetFont(font());
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    Rect_SetWidthHeight(&geometry(), .5f + textSize.width  * ::cfg.common.hudCheatCounterScale,
                                     .5f + textSize.height * ::cfg.common.hudCheatCounterScale);
#endif  // !__JHEXEN__
}
static void PlayerLogWidget_UpdateGeometry(PlayerLogWidget *log) {
    DENG2_ASSERT(log);
    log->updateGeometry();
}
static void PlayerLogWidget_Draw(PlayerLogWidget *log, Point2Raw const *offset) {
    DENG2_ASSERT(log);
    log->draw(offset? Vector2i(offset->xy) : Vector2i());
}
DENG2_PIMPL(PlayerLogWidget) {
    LogEntry entries[LOG_MAX_ENTRIES];
    dint entryCount     = 0;  ///< Number of used entires.
    dint pvisEntryCount = 0;  ///< Number of potentially visible entries.
    dint nextUsedEntry  = 0;  ///< Index of the next used entry to be re-used.
    Impl(Public *i) : Base(i) {}
    dint firstPVisEntryIdx() const {
        if(pvisEntryCount) {
            dint n = nextUsedEntry - de::min(pvisEntryCount, de::max(0, cfg.common.msgCount));
            if(n < 0) n += LOG_MAX_ENTRIES; // Wrap around.
            return n;
        }
        return -1;
    }
    dint firstEntryIdx() const {
        if(pvisEntryCount) {
            dint n = nextUsedEntry - pvisEntryCount;
            if(n < 0) n += LOG_MAX_ENTRIES;  // Wrap around.
            return n;
        }
        return -1;
    }
    inline dint nextEntryIdx(dint current) const {
        if(current < LOG_MAX_ENTRIES - 1) return current + 1;
        return 0; // Wrap around.
    }
    inline dint prevEntryIdx(dint current) const {
        if(current > 0) return current - 1;
        return LOG_MAX_ENTRIES - 1; // Wrap around.
    }
    /**
     * Make a new entry in the log and copy the given @a message.
     *
     * @param message   Message to be added.
     * @param tics      Length of time the message should be visible.
     * @param dontHide  @c true= do not hide this message.
     */
    LogEntry *pushEntry(String const &message, dint tics, bool dontHide) {
        DENG2_ASSERT(!message.isEmpty());
        LogEntry *entry = &entries[nextUsedEntry];
        nextUsedEntry = nextEntryIdx(nextUsedEntry);
        if(entryCount < LOG_MAX_ENTRIES)
            entryCount += 1;
        if(pvisEntryCount < LOG_MAX_ENTRIES)
            pvisEntryCount += 1;
        entry->text      = message;  // make a copy
        entry->tics      = entry->ticsRemain = tics;
        entry->justAdded = true;
        entry->dontHide  = dontHide;
        return entry;
    }
    LogEntry *popEntry() {
        dint oldest = firstEntryIdx();
        if(0 > oldest) return nullptr;
        LogEntry *entry = &entries[oldest];
        if(pvisEntryCount > 0)
            pvisEntryCount -= 1;
        entry->ticsRemain = LOG_MESSAGE_SCROLLTICS;
        entry->justAdded  = false;
        return entry;
    }
};
PlayerLogWidget::PlayerLogWidget(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(PlayerLogWidget_UpdateGeometry),
                function_cast<DrawFunc>(PlayerLogWidget_Draw),
                player)
    , d(new Impl(this))
{}
PlayerLogWidget::~PlayerLogWidget()
{}
void PlayerLogWidget::clear() {
    while(d->popEntry()) {}
    d->nextUsedEntry  = 0;
    d->pvisEntryCount = 0;
    for(LogEntry &entry : d->entries) {
        entry.text.clear();
        entry.text.squeeze();
    }
}
void PlayerLogWidget::post(dint flags, String const &message) {
    if(message.isEmpty()) return;
    d->pushEntry(message, cfg.common.msgUptime * TICSPERSEC, CPP_BOOL(flags & LMF_NO_HIDE));
}
void PlayerLogWidget::refresh() {
    d->pvisEntryCount = de::min(d->entryCount, de::max(0, cfg.common.msgCount));
    dint n = d->firstEntryIdx();
    if(0 > n) return;
    for(dint i = 0; i < d->pvisEntryCount; ++i, n = d->nextEntryIdx(n)) {
        LogEntry *entry = &d->entries[n];
        entry->ticsRemain = entry->tics + i * TICSPERSEC;
        entry->justAdded  = false;
    }
}
void PlayerLogWidget::tick(timespan_t /*tickLength*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    for(dint i = 0; i < LOG_MAX_ENTRIES; ++i) {
        LogEntry *entry = &d->entries[i];
        if(!entry->ticsRemain) continue;
        --entry->ticsRemain;
    }
    dint oldest = d->firstEntryIdx();
    if(oldest >= 0) {
        LogEntry *entry = &d->entries[oldest];
        if(!entry->ticsRemain) {
            d->popEntry();
        }
    }
}
void PlayerLogWidget::draw(Vector2i const &offset) {
    dint pvisEntryCount = de::min(d->pvisEntryCount, de::max(0, cfg.common.msgCount));
    dint firstEntry     = d->firstPVisEntryIdx();
    if(!cfg.hudShown[HUD_LOG]) { {
            dint hiddenEntryCount = 0;
            while(!d->entries[firstEntry].dontHide && ++hiddenEntryCount < pvisEntryCount) {
                firstEntry = d->nextEntryIdx(firstEntry);
            }
            if(hiddenEntryCount == pvisEntryCount) return;
        }
        pvisEntryCount -= firstEntry - d->firstPVisEntryIdx();
    }
    if(pvisEntryCount > 0) { {
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PushMatrix();
            DGL_Translatef(offset.x, offset.y, 0); {
                dfloat offsetDueToMapTitle = 0;
                if(Hu_IsMapTitleVisible() && !cfg.common.automapTitleAtBottom) {
                    offsetDueToMapTitle = Hu_MapTitleHeight();
                }
                DGL_Translatef(0, offsetDueToMapTitle, 0);
            }
            DGL_Scalef(cfg.common.msgScale, cfg.common.msgScale, 1);
        } {
            dint firstEntryVisibleToPlayer  = firstEntry;
            DENG_UNUSED(firstEntryVisibleToPlayer);
            dint lastEntry = firstEntry + pvisEntryCount - 1;
            if(lastEntry > LOG_MAX_ENTRIES - 1)
                lastEntry -= LOG_MAX_ENTRIES;  // Wrap around.
            if(!cfg.hudShown[HUD_LOG]) {
                dint i = 0;
                while(!d->entries[lastEntry].dontHide && ++i < pvisEntryCount) {
                    lastEntry = d->prevEntryIdx(lastEntry);
                }
            }
            FR_SetFont(font());
            dint const lineHeight = FR_CharHeight('Q') + 1;
            LogEntry *entry = &d->entries[firstEntry];
            dfloat yOffset      = 0;
            dfloat scrollFactor = 0;
            if(entry->ticsRemain > 0 && entry->ticsRemain <= (unsigned) lineHeight) {
                scrollFactor = 1.0f - (((dfloat)entry->ticsRemain) / lineHeight);
                yOffset = -lineHeight * scrollFactor;
            }
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_Translatef(0, yOffset, 0);
            DGL_Enable(DGL_TEXTURE_2D);
            dint n = firstEntry;
            dint drawnEntryCount = 0;
            dfloat y = 0;
            for(dint i = 0; i < pvisEntryCount; ++i, n = d->nextEntryIdx(n)) {
                LogEntry *entry = &d->entries[n];
                if(!cfg.hudShown[HUD_LOG] && !entry->dontHide)
                    continue;
                dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
                Vector4f rgba(cfg.common.msgColor, textOpacity);
                if(n == firstEntry) {
                    rgba.w = textOpacity * 1.f - scrollFactor * (4/3.f);
                    rgba.w = de::clamp(0.f, rgba.w, 1.f);
                }
                if(entry->justAdded && cfg.common.msgBlink) {
                    duint const blinkSpeed = cfg.common.msgBlink;
                    duint const msgTics    = entry->tics - entry->ticsRemain;
                    if(msgTics < blinkSpeed) {
                        duint const td = (cfg.common.msgUptime * TICSPERSEC) - entry->ticsRemain;
                        if(n == lastEntry && (0 == msgTics || (td & 2))) {
                            rgba.x = rgba.y = rgba.z = 1;
                        }
                    }
                    else if(msgTics < blinkSpeed + LOG_MESSAGE_FLASHFADETICS && msgTics >= blinkSpeed) {
                        dfloat const fade = (blinkSpeed + LOG_MESSAGE_FLASHFADETICS - msgTics);
                        rgba.x += (1.0f - rgba.x) / LOG_MESSAGE_FLASHFADETICS * fade;
                        rgba.y += (1.0f - rgba.y) / LOG_MESSAGE_FLASHFADETICS * fade;
                        rgba.z += (1.0f - rgba.z) / LOG_MESSAGE_FLASHFADETICS * fade;
                    }
                } {
                    dint const alignFlags    = ALIGN_TOP| ((cfg.common.msgAlign == 0)? ALIGN_LEFT : (cfg.common.msgAlign == 2)? ALIGN_RIGHT : 0);
                    dshort const textFlags   = DTF_NO_EFFECTS;
                    FR_SetColorAndAlpha(rgba.x, rgba.y, rgba.z, rgba.w);
                    FR_DrawTextXY3(entry->text.toUtf8().constData(), 0, y, alignFlags, textFlags);
                }
                ++drawnEntryCount;
                y += lineHeight;
        }
        } {
            DGL_Disable(DGL_TEXTURE_2D);
            DGL_MatrixMode(DGL_MODELVIEW);
            DGL_PopMatrix();
        }
    }
}
void PlayerLogWidget::updateGeometry() {
    Rect_SetWidthHeight(&geometry(), 0, 0);
    dint pvisEntryCount = de::min(d->pvisEntryCount, de::max(0, cfg.common.msgCount));
    if(!pvisEntryCount) return;
    dint firstEntry     = d->firstPVisEntryIdx();
    dint firstPVisEntry = firstEntry;
    if(!cfg.hudShown[HUD_LOG]) {
        dint i = 0;
        while(!d->entries[firstEntry].dontHide && ++i < pvisEntryCount) {
            firstEntry = d->nextEntryIdx(firstEntry);
        }
        if(i == pvisEntryCount) return;
        pvisEntryCount -= firstEntry - firstPVisEntry;
    }
    dint lastEntry = firstEntry + pvisEntryCount - 1;
    if(lastEntry > LOG_MAX_ENTRIES-1)
        lastEntry -= LOG_MAX_ENTRIES; // Wrap around.
    if(!cfg.hudShown[HUD_LOG]) {
        dint i = 0;
        while(!d->entries[lastEntry].dontHide && ++i < pvisEntryCount) {
            lastEntry = d->prevEntryIdx(lastEntry);
        }
    }
    FR_SetFont(FID(GF_FONTA));
    dint const lineHeight = FR_CharHeight('Q') + 1;
    dfloat scrollFactor;
    LogEntry *entry = &d->entries[firstEntry];
    if(entry->ticsRemain > 0 && entry->ticsRemain <= (unsigned) lineHeight) {
        scrollFactor = 1.0f - (((dfloat)entry->ticsRemain) / lineHeight);
    }
    else {
        scrollFactor = 0;
    }
    dint n = firstEntry;
    dint drawnEntryCount = 0;
    RectRaw lineGeometry{};
    for(dint i = 0; i < pvisEntryCount; ++i, n = d->nextEntryIdx(n)) {
        entry = &d->entries[n];
        if(!cfg.hudShown[HUD_LOG] && !entry->dontHide) continue;
        drawnEntryCount += 1;
        FR_TextSize(&lineGeometry.size, entry->text.toUtf8().constData());
        Rect_UniteRaw(&geometry(), &lineGeometry);
        lineGeometry.origin.y += lineHeight;
    }
    if(drawnEntryCount) {
        Rect_SetHeight(&geometry(), Rect_Height(&geometry()) - lineHeight * scrollFactor);
    }
    Rect_SetWidthHeight(&geometry(), Rect_Width (&geometry()) * cfg.common.msgScale,
                                     Rect_Height(&geometry()) * cfg.common.msgScale);
}
static void playerLogVisibilityChanged() {
    for(dint i = 0; i < MAXPLAYERS; ++i) {
        ST_LogPost(i, LMF_NO_HIDE, !::cfg.hudShown[HUD_LOG] ? MSGOFF : MSGON);
    }
}
void PlayerLogWidget::consoleRegister()  // static
{
    C_VAR_FLOAT("msg-uptime",  &cfg.common.msgUptime,   0, 1, 60);
    C_VAR_INT2 ("msg-align",   &cfg.common.msgAlign,    0, 0, 2, ST_LogUpdateAlignment);
    C_VAR_INT  ("msg-blink",   &cfg.common.msgBlink,    CVF_NO_MAX, 0, 0);
    C_VAR_FLOAT("msg-color-r", &cfg.common.msgColor[0], 0, 0, 1);
    C_VAR_FLOAT("msg-color-g", &cfg.common.msgColor[1], 0, 0, 1);
    C_VAR_FLOAT("msg-color-b", &cfg.common.msgColor[2], 0, 0, 1);
    C_VAR_INT  ("msg-count",   &cfg.common.msgCount,    0, 1, 8);
    C_VAR_FLOAT("msg-scale",   &cfg.common.msgScale,    0, 0.1f, 1);
    C_VAR_BYTE2("msg-show",    &cfg.hudShown[HUD_LOG],  0, 0, 1, playerLogVisibilityChanged);
}
#if !__JDOOM__
static patchid_t pAmmoIcons[11];
#endif
guidata_readyammoicon_t::guidata_readyammoicon_t(void (*updateGeometry) (HudWidget *wi),
                                                 void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                                 dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_readyammoicon_t::~guidata_readyammoicon_t()
{}
void guidata_readyammoicon_t::reset() {
#if __JDOOM__
    _sprite = -1;
#else
    _patchId = -1;
#endif
}
void guidata_readyammoicon_t::tick(timespan_t /*elapsed*/)
{
#if __JDOOM__
    static dint const ammoSprite[NUM_AMMO_TYPES] = {
        SPR_AMMO, SPR_SBOX, SPR_CELL, SPR_ROCK
    };
#endif
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const *plr = &::players[player()];
    if(P_MobjIsCamera(plr->plr->mo) && Get(DD_PLAYBACK)) return;
#if __JDOOM__
    _sprite = -1;
#else
    _patchId = -1;
#endif
    if(!VALID_WEAPONTYPE(plr->readyWeapon)) return;
#if __JHERETIC__
    dint const lvl = (plr->powers[PT_WEAPONLEVEL2]? 1 : 0);
#else
    dint const lvl = 0;
#endif
    for(dint i = 0; i < NUM_AMMO_TYPES; ++i) {
        if(WEAPON_INFO(plr->readyWeapon, plr->class_, lvl)->ammoType[i]) {
#if __JDOOM__
            _sprite = ammoSprite[i];
#else
            _patchId = ::pAmmoIcons[i];
#endif
            break;
        }
    }
}
void ReadyAmmoIconWidget_Drawer(guidata_readyammoicon_t *icon, Point2Raw const *offset) {
    DENG2_ASSERT(icon);
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
#if !__JDOOM64__
    if(ST_StatusBarIsActive(icon->player())) return;
#endif
#if !__JHEXEN__
    if(!::cfg.hudShown[HUD_AMMO]) return;
#endif
    if(ST_AutomapIsOpen(icon->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[icon->player()].plr->mo) && Get(DD_PLAYBACK)) return;
#if __JDOOM__
    if(icon->_sprite < 0) return;
#else
    if(icon->_patchId <= 0) return;
#endif
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
#if __JDOOM__
    dfloat scale = (icon->_sprite == SPR_ROCK? .72f : 1);
    GUI_DrawSprite(icon->_sprite, 0, 0, HOT_TLEFT, scale, iconOpacity, false, nullptr, nullptr);
#else
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(icon->_patchId, Vector2i(0, 0), ALIGN_TOPLEFT, DPF_NO_OFFSET);
    DGL_Disable(DGL_TEXTURE_2D);
#endif
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
#if __JHERETIC__
void SBarReadyAmmoIconWidget_Drawer(guidata_readyammoicon_t *icon, Point2Raw const *offset) {
#define ORIGINX             (-ST_WIDTH / 2 )
#define ORIGINY             (-ST_HEIGHT )
#define X_OFFSET            ( 111 )
#define Y_OFFSET            (  14 )
#define X                   ( ORIGINX + X_OFFSET )
#define Y                   ( ORIGINY + Y_OFFSET )
    DENG2_ASSERT(icon);
    dint const activeHud     = ST_ActiveHud(icon->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(icon->player()));
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(Hu_InventoryIsOpen(icon->player())) return;
    if(ST_AutomapIsOpen(icon->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[icon->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(icon->_patchId <= 0) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(icon->_patchId, Vector2i(X, Y));
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef Y
#undef X
#undef Y_OFFSET
#undef X_OFFSET
#undef ORIGINY
#undef ORIGINX
}
#endif  // __JHERETIC__
void ReadyAmmoIconWidget_UpdateGeometry(guidata_readyammoicon_t *icon) {
    DENG2_ASSERT(icon);
    Rect_SetWidthHeight(&icon->geometry(), 0, 0);
#if !__JDOOM64__
    if(ST_StatusBarIsActive(icon->player())) return;
#endif
#if !__JHEXEN__
    if(!::cfg.hudShown[HUD_AMMO]) return;
#endif
    if(ST_AutomapIsOpen(icon->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[icon->player()].plr->mo) && Get(DD_PLAYBACK)) return;
#if __JDOOM__
    if(icon->_sprite < 0) return;
    dfloat scale = (icon->_sprite == SPR_ROCK? .72f : 1);
    Size2Raw iconSize;
    GUI_SpriteSize(icon->_sprite, scale, &iconSize.width, &iconSize.height);
    Rect_SetWidthHeight(&icon->geometry(), iconSize.width  * ::cfg.common.hudScale,
                                           iconSize.height * ::cfg.common.hudScale);
#else
    if(icon->_patchId <= 0) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(icon->_patchId, &info)) return;
    Rect_SetWidthHeight(&icon->geometry(), info.geometry.size.width  * ::cfg.common.hudScale,
                                           info.geometry.size.height * ::cfg.common.hudScale);
#endif
}
#if __JHERETIC__
void SBarReadyAmmoIconWidget_UpdateGeometry(guidata_readyammoicon_t *icon) {
    DENG2_ASSERT(icon);
    Rect_SetWidthHeight(&icon->geometry(), 0, 0);
    if(Hu_InventoryIsOpen(icon->player())) return;
    if(ST_AutomapIsOpen(icon->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[icon->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(icon->_patchId <= 0) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(icon->_patchId, &info)) return;
    Rect_SetWidthHeight(&icon->geometry(), info.geometry.size.width  * ::cfg.common.statusbarScale,
                                           info.geometry.size.height * ::cfg.common.statusbarScale);
}
#endif  // __JHERETIC__
void guidata_readyammoicon_t::prepareAssets() {
#if __JHERETIC__
    de::zap(::pAmmoIcons);
    for(dint i = 0; i < NUM_AMMO_TYPES; ++i) {
        AmmoDef const *def = P_AmmoDef(ammotype_t( i ));
        if(def->gameModeBits & ::gameModeBits) {
            ::pAmmoIcons[i] = R_DeclarePatch(def->hudIcon);
        }
    }
#endif
}
guidata_readyammo_t::guidata_readyammo_t(void (*updateGeometry) (HudWidget *wi),
                                         void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                         dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_readyammo_t::~guidata_readyammo_t()
{}
void guidata_readyammo_t::reset() {
    _value = 1994;
}
void guidata_readyammo_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    _value = 1994; // Means n/a.
    player_t *plr  = &::players[player()];
    if(!VALID_WEAPONTYPE(plr->readyWeapon)) return;
#if __JHERETIC__
    dint const lvl = (plr->powers[PT_WEAPONLEVEL2]? 1 : 0);
#else
    dint const lvl = 0;
#endif
    for(dint i = 0; i < NUM_AMMO_TYPES; ++i) {
        if(WEAPON_INFO(plr->readyWeapon, plr->class_, lvl)->ammoType[i]) {
            _value = plr->ammo[i].owned;
            break;
        }
    }
}
#if defined(__JDOOM__) || defined(__JDOOM64__)
void ReadyAmmo_Drawer(guidata_readyammo_t *ammo, Point2Raw const *offset) {
    DENG2_ASSERT(ammo);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(ammo->_value == 1994) return;
    if(!::cfg.hudShown[HUD_AMMO]) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(ammo->font());
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText, 0, 0);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void SBarReadyAmmo_Drawer(guidata_readyammo_t *ammo, Point2Raw const *offset) {
    static const int ORIGINX       = -ST_WIDTH / 2;
    static const int ORIGINY       = -ST_HEIGHT;
    static const int ST_READYAMMOX = 44;
    static const int ST_READYAMMOY = 3;
    static const int X             = ORIGINX + ST_READYAMMOX;
    static const int Y             = ORIGINY + ST_READYAMMOY;
    DENG2_ASSERT(ammo);
    if(ammo->_value == 1994) return;
    dint const activeHud     = ST_ActiveHud(ammo->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(ammo->player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(ammo->font());
#   if defined(__JDOOM__)
    if(::gameMode == doom_chex) {
        FR_SetColorAndAlpha(::defFontRGB3[0], ::defFontRGB3[1], ::defFontRGB3[2], textOpacity);
    }
    else
#   endif
    {
        FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    }
    FR_DrawTextXY3(valueAsText, X, Y, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void ReadyAmmo_UpdateGeometry(guidata_readyammo_t *ammo) {
    DENG2_ASSERT(ammo);
    Rect_SetWidthHeight(&ammo->geometry(), 0, 0);
    if(ammo->_value == 1994) return;
    if(!::cfg.hudShown[HUD_AMMO]) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    FR_SetFont(ammo->font());
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText);
    textSize.width  *= ::cfg.common.hudScale;
    textSize.height *= ::cfg.common.hudScale;
    Rect_SetWidthHeight(&ammo->geometry(), textSize.width, textSize.height);
}
void SBarReadyAmmo_UpdateGeometry(guidata_readyammo_t *ammo) {
    DENG2_ASSERT(ammo);
    Rect_SetWidthHeight(&ammo->geometry(), 0, 0);
    if(ammo->_value == 1994) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    FR_SetFont(ammo->font());
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText);
    textSize.width  *= ::cfg.common.statusbarScale;
    textSize.height *= ::cfg.common.statusbarScale;
    Rect_SetWidthHeight(&ammo->geometry(), textSize.width, textSize.height);
}
#endif  // __JDOOM__
#if __JHERETIC__
void ReadyAmmo_Drawer(guidata_readyammo_t *ammo, Point2Raw const *offset) {
    static const int TRACKING = 1;
    DENG2_ASSERT(ammo);
    if(ammo->_value == 1994) return;
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(ST_StatusBarIsActive(ammo->player())) return;
    if(!::cfg.hudShown[HUD_AMMO]) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(ammo->font());
    FR_SetTracking(TRACKING);
    FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
    FR_DrawTextXY(valueAsText, 0, -2);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void SBarReadyAmmo_Drawer(guidata_readyammo_t *ammo, Point2Raw const *offset) {
#define ORIGINX                 (-ST_WIDTH / 2 )
#define ORIGINY                 (-ST_HEIGHT )
#define ST_AMMOX                ( 135 )
#define ST_AMMOY                (   4 )
#define X                       ( ORIGINX + ST_AMMOX )
#define Y                       ( ORIGINY + ST_AMMOY )
#define MAXDIGITS               ( ST_AMMOWIDTH )
#define TRACKING                ( 1 )
    DENG2_ASSERT(ammo);
    if(ammo->_value == 1994) return;
    dint const activeHud     = ST_ActiveHud(ammo->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(ammo->player()));
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(Hu_InventoryIsOpen(ammo->player())) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(ammo->font());
    FR_SetTracking(TRACKING);
    FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], iconOpacity);
    FR_DrawTextXY3(valueAsText, X, Y, ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
#undef MAXDIGITS
#undef Y
#undef X
#undef ST_AMMOY
#undef ST_AMMOX
#undef ORIGINY
#undef ORIGINX
}
void ReadyAmmo_UpdateGeometry(guidata_readyammo_t *ammo) {
    static const int TRACKING = 1;
    DENG2_ASSERT(ammo);
    Rect_SetWidthHeight(&ammo->geometry(), 0, 0);
    if(ammo->_value == 1994) return;
    if(ST_StatusBarIsActive(ammo->player())) return;
    if(!::cfg.hudShown[HUD_AMMO]) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    FR_SetFont(ammo->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize;
    FR_TextSize(&textSize, valueAsText);
    Rect_SetWidthHeight(&ammo->geometry(), textSize.width  * ::cfg.common.hudScale,
                                           textSize.height * ::cfg.common.hudScale);
}
void SBarReadyAmmo_UpdateGeometry(guidata_readyammo_t *ammo) {
    static const int TRACKING = 1;
    DENG2_ASSERT(ammo);
    Rect_SetWidthHeight(&ammo->geometry(), 0, 0);
    if(ammo->_value == 1994) return;
    if(Hu_InventoryIsOpen(ammo->player())) return;
    if(ST_AutomapIsOpen(ammo->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[ammo->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    auto const valueAsText = QByteArray::number(ammo->_value);
    FR_SetFont(ammo->font());
    FR_SetTracking(TRACKING);
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText);
    Rect_SetWidthHeight(&ammo->geometry(), textSize.width  * ::cfg.common.statusbarScale,
                                           textSize.height * ::cfg.common.statusbarScale);
}
#endif  // __JHERETIC__
#define FLASH_FRAME_COUNT           ( 5 )
static patchid_t pBackground;
static patchid_t pIcons[FLASH_FRAME_COUNT];
guidata_readyitem_t::guidata_readyitem_t(void (*updateGeometry) (HudWidget *wi),
                                         void (*drawer) (HudWidget *wi, Point2Raw const *offset),
                                         dint player)
    : HudWidget(updateGeometry,
                drawer,
                player)
{}
guidata_readyitem_t::~guidata_readyitem_t()
{}
void guidata_readyitem_t::reset() {
    _patchId = 0;
}
void guidata_readyitem_t::tick(timespan_t /*elapsed*/)
{
    _patchId = 0;
#if __JHERETIC__ || __JHEXEN__
    dint const flashCounter = ST_ReadyItemFlashCounter(player());
    if(flashCounter > 0) {
        _patchId = ::pIcons[flashCounter % FLASH_FRAME_COUNT];
    }
    else {
        inventoryitemtype_t readyItem = P_InventoryReadyItem(player());
        if(readyItem != IIT_NONE) {
            _patchId = P_GetInvItem(dint( readyItem ) - 1)->patchId;
        }
    }
#endif
}
#if __JHERETIC__
void ReadyItem_Drawer(guidata_readyitem_t *item, Point2Raw const *offset) {
#define TRACKING                ( 2 )
    DENG2_ASSERT(item);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_READYITEM]) return;
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t boxInfo;
    if(!R_GetPatchInfo(::pBackground, &boxInfo)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(cfg.common.hudScale, cfg.common.hudScale, 1);
    if(item->_patchId) {
        dint xOffset = 0, yOffset = 0;
        if(ST_ReadyItemFlashCounter(item->player()) > 0) {
            xOffset += 2;
            yOffset += 1;
        }
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PushMatrix();
        if(offset) DGL_Translatef(offset->x, offset->y, 0);
        DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
        DGL_Enable(DGL_TEXTURE_2D);
        DGL_Color4f(1, 1, 1, iconOpacity / 2);
        GL_DrawPatch(::pBackground, Vector2i(0, 0), ALIGN_TOPLEFT, DPF_NO_OFFSET);
        DGL_Color4f(1, 1, 1, iconOpacity);
        GL_DrawPatch(item->_patchId, Vector2i(xOffset, yOffset));
        inventoryitemtype_t readyItem = P_InventoryReadyItem(item->player());
        if(!(ST_ReadyItemFlashCounter(item->player()) > 0) && IIT_NONE != readyItem) {
            duint count = P_InventoryCount(item->player(), readyItem);
            if(count > 1) {
                auto const countAsText = String::number(count);
                FR_SetFont(item->font());
                FR_SetTracking(TRACKING);
                FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
                FR_DrawTextXY2(countAsText.toUtf8().constData(),
                               boxInfo.geometry.size.width - 1, boxInfo.geometry.size.height - 3,
                               ALIGN_BOTTOMRIGHT);
            }
        }
        DGL_Disable(DGL_TEXTURE_2D);
        DGL_MatrixMode(DGL_MODELVIEW);
        DGL_PopMatrix();
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef TRACKING
}
void SBarReadyItem_Drawer(guidata_readyitem_t *item, Point2Raw const *offset) {
#define TRACKING                ( 2 )
#define ORIGINX                 (-ST_WIDTH / 2 )
#define ORIGINY                 (-ST_HEIGHT * ST_StatusBarShown(item->player()) )
#define ICON_X_OFFSET           ( 179 )
#define ICON_Y_OFFSET           (   2 )
#define COUNT_X_OFFSET          ( 208 )
#define COUNT_Y_OFFSET          (  24 )
    DENG2_ASSERT(item);
    dint const activeHud     = ST_ActiveHud(item->player());
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(!item->_patchId) return;
    dint x, y;
    if(ST_ReadyItemFlashCounter(item->player()) > 0) {
        x = ORIGINX + ICON_X_OFFSET + 2;
        y = ORIGINY + ICON_Y_OFFSET + 1;
    }
    else {
        x = ORIGINX + ICON_X_OFFSET;
        y = ORIGINY + ICON_Y_OFFSET;
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(item->_patchId, Vector2i(x, y));
    inventoryitemtype_t readyItem = P_InventoryReadyItem(item->player());
    if(!(ST_ReadyItemFlashCounter(item->player()) > 0) && IIT_NONE != readyItem) {
        duint count = P_InventoryCount(item->player(), readyItem);
        if(count > 1) {
            auto const countAsText = String::number(count);
            FR_SetFont(item->font());
            FR_SetTracking(TRACKING);
            FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
            FR_DrawTextXY3(countAsText.toUtf8().constData(),
                           ORIGINX + COUNT_X_OFFSET, ORIGINY + COUNT_Y_OFFSET,
                           ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
        }
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef COUNT_Y_OFFSET
#undef COUNT_X_OFFSET
#undef ICON_Y_OFFSET
#undef ICON_X_OFFSET
#undef ORIGINY
#undef ORIGINX
#undef TRACKING
}
void ReadyItem_UpdateGeometry(guidata_readyitem_t *item) {
    DENG2_ASSERT(item);
    Rect_SetWidthHeight(&item->geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_READYITEM]) return;
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(::pBackground, &info)) return;
    Rect_SetWidthHeight(&item->geometry(), info.geometry.size.width  * ::cfg.common.hudScale,
                                           info.geometry.size.height * ::cfg.common.hudScale);
}
void SBarReadyItem_UpdateGeometry(guidata_readyitem_t *item) {
    DENG2_ASSERT(item);
    Rect_SetWidthHeight(&item->geometry(), 0, 0);
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(!item->_patchId) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(item->_patchId, &info)) return;
    Rect_SetWidthHeight(&item->geometry(), info.geometry.size.width  * ::cfg.common.statusbarScale,
                                           info.geometry.size.height * ::cfg.common.statusbarScale);
}
#endif  // __JHERETIC__
#if __JHEXEN__
void ReadyItem_Drawer(guidata_readyitem_t *item, const Point2Raw* offset) {
    DENG2_ASSERT(item);
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    dfloat const iconOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudIconAlpha;
    if(!::cfg.hudShown[HUD_READYITEM]) return;
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(!item->_patchId) return;
    patchinfo_t boxInfo;
    if(!R_GetPatchInfo(::pBackground, &boxInfo)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.hudScale, ::cfg.common.hudScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity / 2);
    GL_DrawPatch(::pBackground, Vector2i(0, 0));
    dint xOffset, yOffset;
    if(ST_ReadyItemFlashCounter(item->player()) > 0) {
        xOffset = 3;
        yOffset = 0;
    }
    else {
        xOffset = -2;
        yOffset = -1;
    }
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(item->_patchId, Vector2i(xOffset, yOffset));
    inventoryitemtype_t readyItem = P_InventoryReadyItem(item->player());
    if(ST_ReadyItemFlashCounter(item->player()) == 0 && readyItem != IIT_NONE) {
        duint count = P_InventoryCount(item->player(), readyItem);
        if(count > 1) {
            auto const countAsText = String::number(count);
            FR_SetFont(item->font());
            FR_SetTracking(0);
            FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
            FR_DrawTextXY2(countAsText.toUtf8().constData(),
                           boxInfo.geometry.size.width - 1, boxInfo.geometry.size.height - 3,
                           ALIGN_BOTTOMRIGHT);
        }
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
}
void SBarReadyItem_Drawer(guidata_readyitem_t *item, Point2Raw const *offset) {
#define ORIGINX             (-ST_WIDTH / 2 )
#define ORIGINY             (-ST_HEIGHT )
#define ST_INVITEMX         ( 143 )
#define ST_INVITEMY         (   1 )
#define ST_INVITEMCX        ( 174 )
#define ST_INVITEMCY        (  22 )
    DENG2_ASSERT(item);
    dint const activeHud     = ST_ActiveHud(item->player());
    dint const yOffset       = ST_HEIGHT * (1 - ST_StatusBarShown(item->player()));
    dfloat const textOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    dfloat const iconOpacity = (activeHud == 0? 1 : ::uiRendState->pageAlpha * ::cfg.common.statusbarCounterAlpha);
    if(Hu_InventoryIsOpen(item->player()) || ST_AutomapIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(!item->_patchId) return;
    patchinfo_t boxInfo;
    if(!R_GetPatchInfo(::pBackground, &boxInfo)) return;
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    if(offset) DGL_Translatef(offset->x, offset->y, 0);
    DGL_Scalef(::cfg.common.statusbarScale, ::cfg.common.statusbarScale, 1);
    DGL_Translatef(0, yOffset, 0);
    dint x, y;
    if(ST_ReadyItemFlashCounter(item->player()) > 0) {
        x = ST_INVITEMX + 4;
        y = ST_INVITEMY;
    }
    else {
        x = ST_INVITEMX;
        y = ST_INVITEMY;
    }
    DGL_Enable(DGL_TEXTURE_2D);
    DGL_Color4f(1, 1, 1, iconOpacity);
    GL_DrawPatch(item->_patchId, Vector2i(ORIGINX + x, ORIGINY + y));
    inventoryitemtype_t readyItem = P_InventoryReadyItem(item->player());
    if(!(ST_ReadyItemFlashCounter(item->player()) > 0) && readyItem != IIT_NONE) {
        duint count = P_InventoryCount(item->player(), readyItem);
        if(count > 1) {
            auto const countAsText = String::number(count);
            FR_SetFont(item->font());
            FR_SetTracking(0);
            FR_SetColorAndAlpha(::defFontRGB2[0], ::defFontRGB2[1], ::defFontRGB2[2], textOpacity);
            FR_DrawTextXY3(countAsText.toUtf8().constData(),
                           ORIGINX + ST_INVITEMCX, ORIGINY + ST_INVITEMCY,
                           ALIGN_TOPRIGHT, DTF_NO_EFFECTS);
        }
    }
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#undef ST_INVITEMCY
#undef ST_INVITEMCX
#undef ST_INVITEMY
#undef ST_INVITEMX
#undef ORIGINY
#undef ORIGINX
}
void ReadyItem_UpdateGeometry(guidata_readyitem_t *item) {
    DENG2_ASSERT(item);
    Rect_SetWidthHeight(&item->geometry(), 0, 0);
    if(!::cfg.hudShown[HUD_READYITEM]) return;
    if(Hu_InventoryIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(::pBackground, &info)) return;
    Rect_SetWidthHeight(&item->geometry(), info.geometry.size.width  * ::cfg.common.hudScale,
                                           info.geometry.size.height * ::cfg.common.hudScale);
}
void SBarReadyItem_UpdateGeometry(guidata_readyitem_t *item) {
    DENG2_ASSERT(item);
    Rect_SetWidthHeight(&item->geometry(), 0, 0);
    if(Hu_InventoryIsOpen(item->player()) || ST_AutomapIsOpen(item->player())) return;
    if(ST_AutomapIsOpen(item->player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[item->player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(item->_patchId != 0) return;
    patchinfo_t info;
    if(!R_GetPatchInfo(::pBackground, &info)) return;
    Rect_SetWidthHeight(&item->geometry(), info.geometry.size.width  * ::cfg.common.statusbarScale,
                                           info.geometry.size.height * ::cfg.common.statusbarScale);
}
#endif  // __JHEXEN__
void guidata_readyitem_t::prepareAssets()  // static
{
    ::pBackground = R_DeclarePatch("ARTIBOX");
    for(dint i = 0; i < FLASH_FRAME_COUNT; ++i) {
        ::pIcons[i] = R_DeclarePatch(String("USEARTI%1").arg(char('A' + i)).toUtf8().constData());
    }
}
static void SecretsWidget_Draw(guidata_secrets_t *scrts, Point2Raw const *offset) {
    DENG2_ASSERT(scrts);
    scrts->draw(offset? Vector2i(offset->xy) : Vector2i());
}
static void SecretsWidget_UpdateGeometry(guidata_secrets_t *scrts) {
    DENG2_ASSERT(scrts);
    scrts->updateGeometry();
}
guidata_secrets_t::guidata_secrets_t(dint player)
    : HudWidget(function_cast<UpdateGeometryFunc>(SecretsWidget_UpdateGeometry),
                function_cast<DrawFunc>(SecretsWidget_Draw),
                player)
{}
guidata_secrets_t::~guidata_secrets_t()
{}
void guidata_secrets_t::reset() {
    _value = 1994;
}
void guidata_secrets_t::draw(Vector2i const &offset) const {
#if !__JHEXEN__
    dfloat const textOpacity = ::uiRendState->pageAlpha * ::cfg.common.hudColor[3];
    if(_value == 1994) return;
    if(!(::cfg.common.hudShownCheatCounters & (CCH_SECRETS | CCH_SECRETS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    String valueAsText("Secret");
    if(::cfg.common.hudShownCheatCounters & CCH_SECRETS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalSecret);
    }
    if(::cfg.common.hudShownCheatCounters & CCH_SECRETS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((cfg.common.hudShownCheatCounters & CCH_SECRETS) ? "(" : "")
                           .arg(totalSecret ? _value * 100 / totalSecret : 100)
                           .arg((cfg.common.hudShownCheatCounters & CCH_SECRETS) ? ")" : "");
    }
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PushMatrix();
    DGL_Translatef(offset.x, offset.y, 0);
    DGL_Scalef(::cfg.common.hudCheatCounterScale, ::cfg.common.hudCheatCounterScale, 1);
    DGL_Enable(DGL_TEXTURE_2D);
    FR_SetFont(font());
    FR_SetColorAndAlpha(::cfg.common.hudColor[0], ::cfg.common.hudColor[1], ::cfg.common.hudColor[2], textOpacity);
    FR_DrawTextXY(valueAsText.toUtf8().constData(), 0, 0);
    DGL_Disable(DGL_TEXTURE_2D);
    DGL_MatrixMode(DGL_MODELVIEW);
    DGL_PopMatrix();
#else  // !__JHEXEN__
    DENG2_UNUSED(offset);
#endif
}
void guidata_secrets_t::tick(timespan_t /*elapsed*/)
{
    if(Pause_IsPaused() || !DD_IsSharpTick()) return;
    player_t const &plr = ::players[player()];
    _value = plr.secretCount;
}
void guidata_secrets_t::updateGeometry() {
#if !__JHEXEN__
    Rect_SetWidthHeight(&geometry(), 0, 0);
    if(_value == 1994) return;
    if(!(::cfg.common.hudShownCheatCounters & (CCH_SECRETS | CCH_SECRETS_PRCNT))) return;
    if(ST_AutomapIsOpen(player()) && ::cfg.common.automapHudDisplay == 0) return;
    if(P_MobjIsCamera(::players[player()].plr->mo) && Get(DD_PLAYBACK)) return;
    if(::cfg.common.hudCheatCounterShowWithAutomap && !ST_AutomapIsOpen(player())) return;
    String valueAsText("Secret");
    if(::cfg.common.hudShownCheatCounters & CCH_SECRETS) {
        valueAsText += String(" %1/%2").arg(_value).arg(totalSecret);
    }
    if(::cfg.common.hudShownCheatCounters & CCH_SECRETS_PRCNT) {
        valueAsText += String(" %1%2%%3")
                           .arg((cfg.common.hudShownCheatCounters & CCH_SECRETS) ? "(" : "")
                           .arg(totalSecret ? _value * 100 / totalSecret : 100)
                           .arg((cfg.common.hudShownCheatCounters & CCH_SECRETS) ? ")" : "");
    }
    FR_SetFont(font());
    Size2Raw textSize; FR_TextSize(&textSize, valueAsText.toUtf8().constData());
    Rect_SetWidthHeight(&geometry(), .5f + textSize.width  * ::cfg.common.hudCheatCounterScale,
                                     .5f + textSize.height * ::cfg.common.hudCheatCounterScale);
#endif  // !__JHEXEN__
}
