﻿/** @file dmu_lib.cpp Helper routines for accessing the DMU API.
 *
 * @authors Copyright © 2006-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <cstdio>
#include <cstring>
#include "common.h"
#include "dmu_lib.h"
#include "p_terraintype.h"
#include "dmu_lib.h"
/** @file mapstatereader.cpp  Saved map state reader.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "mapstatereader.h"
#include <de/ArrayValue>
#include <de/NativePath>
#include <de/String>
#include <doomsday/world/MaterialArchive>
#include "d_netsv.h"           /// @todo remove me
#include "dmu_lib.h"
#include "dmu_archiveindex.h"
#include "g_game.h"
#include "mapstatewriter.h"    // ChunkId
#include "p_actor.h"
#include "p_mapsetup.h"
#include "p_savedef.h"
#include "p_saveg.h"           // SV_ReadLine, SV_ReadSector
#include "p_saveio.h"
#include "player.h"
#include "polyobjs.h"
#include "r_common.h"
#include "thinkerinfo.h"
#include "gamesession.h"
#include "acs/system.h"
/** @file mapstatewriter.cpp  Saved map state writer.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "mapstatewriter.h"
#include "dmu_lib.h"
#include "g_game.h"
#include "gamesession.h"
#include "p_savedef.h"   // MY_SAVE_VERSION
#include "p_saveg.h"     // SV_WriteSector, SV_WriteLine
#include "polyobjs.h"
#include "thinkerinfo.h"
#include "acs/system.h"
#include <doomsday/world/MaterialArchive>
#include <doomsday/world/thinkerdata.h>
/** @file mobj.cpp  Common playsim map object (mobj) functionality.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "mobj.h"
#include <cmath>
#include <doomsday/world/mobjthinkerdata.h>
#include <de/String>
#include <de/mathutil.h>
#include <de/Log>
#include <QTextStream>
#include "dmu_lib.h"
#include "g_defs.h"
#include "mapstatereader.h"
#include "mapstatewriter.h"
#include "p_actor.h"
#include "p_map.h"
#include "p_saveg.h"
#include "player.h"
#ifdef MSVC
#  pragma optimize("g", off)
#endif
#define DROPOFFMOMENTUM_THRESHOLD (1.0 / 4)
#define STANDSPEED              (1.0 / 2) // FIX2FLT(0x8000)
/** @file p_actor.cpp Common code relating to mobj management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include <cstdio>
#include <cstring>
#include "common.h"
#include "gamesession.h"
#include "p_tick.h"
#include "p_actor.h"
#include "p_actor.h"
#if __JDOOM64__
# define RESPAWNTICS            (4 * TICSPERSEC)
#else
# define RESPAWNTICS            (30 * TICSPERSEC)
#endif
/** @file p_ceiling.cpp  Moving ceilings (lowering, crushing, raising).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Martin Eyre <martineyre@btinternet.com>
 * @authors Copyright © 2003-2005 Samuel Villarreal <svkaiser@gmail.com>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_ceiling.h"
#include "dmu_lib.h"
#include "p_mapspec.h"
#include "p_sound.h"
#include "p_start.h"
#include "p_tick.h"
#if __JDOOM__
# define SFX_CEILINGMOVE        (SFX_STNMOV)
# define SFX_CEILINGSTOP        (SFX_PSTOP)
#elif __JDOOM64__
# define SFX_CEILINGMOVE        (SFX_STNMOV)
# define SFX_CEILINGSTOP        (SFX_PSTOP)
#elif __JHERETIC__
# define SFX_CEILINGMOVE        (SFX_DORMOV)
# define SFX_CEILINGSTOP        (SFX_NONE)
#elif __JHEXEN__
# define SFX_CEILINGMOVE        (SFX_NONE)
# define SFX_CEILINGSTOP        (SFX_NONE)
#endif
/**
 * Called when a moving ceiling needs to be removed.
 *
 * @param ceiling       Ptr to the ceiling to be stopped.
 */
/** @file p_door.cpp  Vertical door (opening/closing) thinker.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Martin Eyre <martineyre@btinternet.com>
 * @authors Copyright © 2003-2005 Samuel Villarreal <svkaiser@gmail.com>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_door.h"
#include "dmu_lib.h"
#include "player.h"
#include "p_mapspec.h"
#if __JDOOM64__
#  include "p_ceiling.h"
#  include "p_floor.h"
#  include "p_inventory.h"
#endif
#include "p_sound.h"
#include <cstring>
#if __JDOOM64__
# define SFX_DOORCLOSING       (SFX_DORCLS)
# define SFX_DOORCLOSED        (SFX_DORCLS)
# define SFX_DOORBLAZECLOSE    (SFX_BDCLS)
# define SFX_DOOROPEN          (SFX_DOROPN)
# define SFX_DOORBLAZEOPEN     (SFX_BDOPN)
# define SFX_DOORLOCKED        (SFX_OOF)
#elif __JDOOM__
# define SFX_DOORCLOSING       (SFX_DORCLS)
# define SFX_DOORCLOSED        (SFX_DORCLS)
# define SFX_DOORBLAZECLOSE    (SFX_BDCLS)
# define SFX_DOOROPEN          (SFX_DOROPN)
# define SFX_DOORBLAZEOPEN     (SFX_BDOPN)
# define SFX_DOORLOCKED        (SFX_OOF)
#elif __JHERETIC__
# define SFX_DOORCLOSING       (SFX_DOROPN)
# define SFX_DOORCLOSED        (SFX_DORCLS)
# define SFX_DOORBLAZECLOSE    (SFX_NONE)
# define SFX_DOOROPEN          (SFX_DOROPN)
# define SFX_DOORBLAZEOPEN     (SFX_DOROPN)
# define SFX_DOORLOCKED        (SFX_PLROOF)
#elif __JHEXEN__
# define SFX_DOORCLOSING       (SFX_NONE)
# define SFX_DOORCLOSED        (SFX_NONE)
# define SFX_DOORBLAZECLOSE    (SFX_NONE)
# define SFX_DOOROPEN          (SFX_NONE)
# define SFX_DOORBLAZEOPEN     (SFX_NONE)
# define SFX_DOORLOCKED        (SFX_NONE)
#endif
/** @file p_floor.cpp  Common playsim routines relating to moving floors.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Martin Eyre <martineyre@btinternet.com>
 * @authors Copyright © 2003-2005 Samuel Villarreal <svkaiser@gmail.com>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_floor.h"
#include "dmu_lib.h"
#include "p_map.h"
#include "p_mapspec.h"
#include "p_tick.h"
#if __JHEXEN__ || __JDOOM64__
#include "p_ceiling.h"
#endif
#include "p_sound.h"
#include "p_saveg.h"
#if __JHERETIC__
# define SFX_FLOORMOVE          (SFX_DORMOV)
#else
# define SFX_FLOORMOVE          (SFX_STNMOV)
#endif
#if __JHEXEN__
#define STAIR_SECTOR_TYPE       26
#define STAIR_QUEUE_SIZE        32
#endif
#if __JHEXEN__
typedef struct stairqueue_s {
    Sector *sector;
    int type;
    coord_t height;
} stairqueue_t;
typedef struct stairdata_s {
    coord_t stepDelta;
    int direction;
    float speed;
    world_Material *material;
    int startDelay;
    int startDelayDelta;
    int textureChange;
    coord_t startHeight;
} stairdata_t;
#endif
#if __JHEXEN__
static void enqueueStairSector(Sector *sec, int type, coord_t height);
#endif
#if __JHEXEN__
stairdata_t stairData;
stairqueue_t stairQueue[STAIR_QUEUE_SIZE];
static int stairQueueHead;
static int stairQueueTail;
#endif
/** @file p_map.cpp  Common map routines.
 *
 * @authors Copyright © 2003-2019 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_map.h"
#include <cmath>
#include <cstdio>
#include <cstring>
#include "acs/system.h"
#include "d_net.h"
#include "d_netcl.h"
#include "d_netsv.h"
#include "g_common.h"
#include "gamesession.h"
#include "dmu_lib.h"
#include "p_mapspec.h"
#include "p_terraintype.h"
#include "p_tick.h"
#include "p_actor.h"
#include "player.h"
#include "p_mapsetup.h"
/*
 * Try move variables:
 */
/** @file p_mapsetup.cpp  Common map setup routines.
 *
 * Management of extended map data objects (e.g., xlines).
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_mapsetup.h"
#include <doomsday/world/entitydef.h>
#include <gamefw/mapspot.h>
#include <de/Script>
#include <de/Process>
#include <cmath>
#include <cctype>  // isspace
#include <cstring>
#include "acs/script.h"
#include "acs/system.h"
#include "d_net.h"
#include "d_netsv.h"
#include "dmu_lib.h"
#include "g_common.h"
#include "gamesession.h"
#include "hu_pspr.h"
#include "hu_stuff.h"
#include "hud/widgets/automapwidget.h"
#include "p_actor.h"
#include "p_scroll.h"
#include "p_start.h"
#include "p_tick.h"
#include "polyobjs.h"
#include "r_common.h"
#if __JDOOM64__
# define TOLIGHTIDX(c) (!((c) >> 8)? 0 : ((c) - 0x100) + 1)
#endif
/** @file p_mapspec.cpp  World map utilities.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_mapspec.h"
#include "dmu_lib.h"
#include "g_game.h"
#include "gamesession.h"
#include "p_mapsetup.h"
#include "acs/system.h"
/** @file p_plat.cpp  Common playsim routines relating to moving platforms.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2006 Martin Eyre <martineyre@btinternet.com>
 * @authors Copyright © 2003-2005 Samuel Villarreal <svkaiser@gmail.com>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_plat.h"
#include "dmu_lib.h"
#include "p_mapspec.h"
#include "p_tick.h"
#include "p_sound.h"
#if __JDOOM__
# define SFX_PLATFORMSTART      (SFX_PSTART)
# define SFX_PLATFORMMOVE       (SFX_STNMOV)
# define SFX_PLATFORMSTOP       (SFX_PSTOP)
#elif __JDOOM64__
# define SFX_PLATFORMSTART      (SFX_PSTART)
# define SFX_PLATFORMMOVE       (SFX_STNMOV)
# define SFX_PLATFORMSTOP       (SFX_PSTOP)
#elif __JHERETIC__
# define SFX_PLATFORMSTART      (SFX_PSTART)
# define SFX_PLATFORMMOVE       (SFX_STNMOV)
# define SFX_PLATFORMSTOP       (SFX_PSTOP)
#elif __JHEXEN__
# define SFX_PLATFORMSTART      (SFX_NONE)
# define SFX_PLATFORMMOVE       (SFX_NONE)
# define SFX_PLATFORMSTOP       (SFX_NONE)
#endif
/**
 * Called when a moving plat needs to be removed.
 *
 * @param plat  Ptr to the plat to remove.
 */
/** @file p_scroll.cpp  Common surface material scroll thinker.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_scroll.h"
#include "dmu_lib.h"
#include "p_saveg.h"
/** @file p_sound.cpp  id Tech 1 sound playback functionality.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2015 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Activision
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_sound.h"
#include <de/Log>
#include "dmu_lib.h"
#include "g_defs.h"
#ifdef __JHEXEN__
#  include "g_common.h"
#  include "hexlex.h"
#endif
/** @file p_start.cpp  Common player (re)spawning logic.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Activision
 * @authors Copyright © 1993-1996 by id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_start.h"
#include <cmath>
#include <cstdio>
#include <cstring>
#include <de/NativePath>
#include <doomsday/busymode.h>
#include "d_net.h"
#include "d_netsv.h"
#include "dmu_lib.h"
#include "gamesession.h"
#include "g_common.h"
#include "g_defs.h"
#include "hu_stuff.h"
#include "p_actor.h"
#include "p_inventory.h"
#include "p_map.h"
#include "p_mapsetup.h"
#include "p_mapspec.h"
#include "p_switch.h"
#include "p_terraintype.h"
#include "p_tick.h"
#include "p_user.h"
#include "player.h"
#include "r_common.h"
/** @file p_switch.cpp  Common playsim routines relating to switches.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman (PrBoom 2.2.6)
 * @authors Copyright © 1999-2000 Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze (PrBoom 2.2.6)
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_switch.h"
#include "d_net.h"
#include "dmu_lib.h"
#include "dmu_archiveindex.h"
#include "p_plat.h"
#include "p_sound.h"
#include "p_saveg.h"
#include <de/memory.h>
/**
 * This struct is used to provide byte offsets when reading a custom
 * SWITCHES lump thus it must be packed and cannot be altered.
 */
#pragma pack(1)
/** @file p_tick.cpp  Common top-level tick stuff.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "p_tick.h"
#include "gamesession.h"
#include "d_netsv.h"
#include "g_common.h"
#include "g_controls.h"
#include "hu_menu.h"
#include "hu_msg.h"
#include "p_actor.h"
#include "p_user.h"
#include "player.h"
#include "r_common.h"
#include "r_special.h"
/** @file p_xgfile.cpp  Extended generalized line types.
 *
 * DD_XGDATA lump reader.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
#include "common.h"
#include "p_xg.h"
#include <de/memory.h>
#include <cstdio>
#include <cstring>
using namespace de;
enum xgsegenum_t
{
    XGSEG_END,
    XGSEG_LINE,
    XGSEG_SECTOR
};
dd_bool xgDataLumps;
static byte const *readptr;
static linetype_t *linetypes;
static int num_linetypes;
static sectortype_t *sectypes;
static int num_sectypes;
static byte ReadByte()
{
    return *readptr++;
}
static short ReadShort()
{
    short s = *(const short *) readptr;
    readptr += 2;
    return DD_SHORT(s);
}
static long ReadLong()
{
    long l = *(const long *) readptr;
    readptr += 4;
    return DD_LONG(l);
}
static float ReadFloat()
{
    long f = ReadLong();
    float returnValue = 0;
    std::memcpy(&returnValue, &f, 4);
    return returnValue;
}
/**
 * I could just return a pointer to the string, but that risks losing
 * it somewhere. Now we can be absolutely sure it can't be lost.
 */
static void ReadString(char **str)
{
    int len = ReadShort();
    if(!len) // Null string?
    {
        *str = 0;
        return;
    }
    if(len < 0)
        Con_Error("ReadString: Bogus len!\n");
    *str = (char *) Z_Malloc(len + 1, PU_GAMESTATIC, 0);
    std::memcpy(*str, readptr, len);
    readptr += len;
    (*str)[len] = 0;
}
static uri_s *readTextureUrn()
{
    return Uri_NewWithPath2(Str_Text(Str_Appendf(AutoStr_NewStd(), "urn:Textures:%i", ReadShort())), RC_NULL);
}
void XG_ReadXGLump(lumpnum_t lumpNum)
{
    if(0 > lumpNum)
        return; // No such lump.
    xgDataLumps = true;
    App_Log(DE2_RES_MSG, "Reading XG types from DDXGDATA");
    LumpIndex const &lumps = CentralLumpIndex();
    File1 &lump = lumps[lumpNum];
    readptr = (byte *)lump.cache();
    num_linetypes = ReadShort();
    linetypes = (linetype_t *)Z_Calloc(sizeof(*linetypes) * num_linetypes, PU_GAMESTATIC, 0);
    num_sectypes = ReadShort();
    sectypes = (sectortype_t *)Z_Calloc(sizeof(*sectypes) * num_sectypes, PU_GAMESTATIC, 0);
    int lc = 0, sc = 0;
    dd_bool done = false;
    while(!done)
    {
        switch(ReadByte())
        {
        case XGSEG_END:
            done = true;
            break;
        case XGSEG_LINE: {
            linetype_t *li = linetypes + lc++;
            li->id = ReadShort();
            li->flags = ReadLong();
            li->flags2 = ReadLong();
            li->flags3 = ReadLong();
            li->lineClass = ReadShort();
            li->actType = ReadByte();
            li->actCount = ReadShort();
            li->actTime = ReadFloat();
            li->actTag = ReadLong();
            for(int i = 0; i < DDLT_MAX_APARAMS; ++i)
            {
                li->aparm[i] = ReadLong();
            }
            li->tickerStart = ReadFloat();
            li->tickerEnd = ReadFloat();
            li->tickerInterval = ReadLong();
            li->actSound = ReadShort();
            li->deactSound = ReadShort();
            li->evChain = ReadShort();
            li->actChain = ReadShort();
            li->deactChain = ReadShort();
            li->wallSection = ReadByte();
            {
                uri_s *textureUrn = readTextureUrn();
                li->actMaterial = P_ToIndex(DD_MaterialForTextureUri(textureUrn));
                Uri_Delete(textureUrn);
            }
            {
                uri_s *textureUrn = readTextureUrn();
                li->deactMaterial = P_ToIndex(DD_MaterialForTextureUri(textureUrn));
                Uri_Delete(textureUrn);
            }
            ReadString(&li->actMsg);
            ReadString(&li->deactMsg);
            li->materialMoveAngle = ReadFloat();
            li->materialMoveSpeed = ReadFloat();
            for(int i = 0; i < DDLT_MAX_PARAMS; ++i)
            {
                li->iparm[i] = ReadLong();
            }
            for(int i = 0; i < DDLT_MAX_PARAMS; ++i)
            {
                li->fparm[i] = ReadFloat();
            }
            for(int i = 0; i < DDLT_MAX_SPARAMS; ++i)
            {
                ReadString(&li->sparm[i]);
            }
            break; }
        case XGSEG_SECTOR: {
            sectortype_t *sec = sectypes + sc++;
            sec->id = ReadShort();
            sec->flags = ReadLong();
            sec->actTag = ReadLong();
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->chain[i] = ReadLong();
            }
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->chainFlags[i] = ReadLong();
            }
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->start[i] = ReadFloat();
            }
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->end[i] = ReadFloat();
            }
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->interval[i][0] = ReadFloat();
                sec->interval[i][1] = ReadFloat();
            }
            for(int i = 0; i < DDLT_MAX_CHAINS; ++i)
            {
                sec->count[i] = ReadLong();
            }
            sec->ambientSound = ReadShort();
            sec->soundInterval[0] = ReadFloat();
            sec->soundInterval[1] = ReadFloat();
            sec->materialMoveAngle[0] = ReadFloat();
            sec->materialMoveAngle[1] = ReadFloat();
            sec->materialMoveSpeed[0] = ReadFloat();
            sec->materialMoveSpeed[1] = ReadFloat();
            sec->windAngle = ReadFloat();
            sec->windSpeed = ReadFloat();
            sec->verticalWind = ReadFloat();
            sec->gravity = ReadFloat();
            sec->friction = ReadFloat();
            ReadString(&sec->lightFunc);
            sec->lightInterval[0] = ReadShort();
            sec->lightInterval[1] = ReadShort();
            ReadString(&sec->colFunc[0]);
            ReadString(&sec->colFunc[1]);
            ReadString(&sec->colFunc[2]);
            for(int i = 0; i < 3; ++i)
            {
                sec->colInterval[i][0] = ReadShort();
                sec->colInterval[i][1] = ReadShort();
            }
            ReadString(&sec->floorFunc);
            sec->floorMul = ReadFloat();
            sec->floorOff = ReadFloat();
            sec->floorInterval[0] = ReadShort();
            sec->floorInterval[1] = ReadShort();
            ReadString(&sec->ceilFunc);
            sec->ceilMul = ReadFloat();
            sec->ceilOff = ReadFloat();
            sec->ceilInterval[0] = ReadShort();
            sec->ceilInterval[1] = ReadShort();
            break; }
        default:
            lump.unlock();
            Con_Error("XG_ReadXGLump: Bad segment!");
        }
    }
    lump.unlock();
}
void XG_ReadTypes()
{
    num_linetypes = 0;
    Z_Free(linetypes); linetypes = 0;
    num_sectypes = 0;
    Z_Free(sectypes); sectypes = 0;
    XG_ReadXGLump(CentralLumpIndex().findLast("DDXGDATA.lmp"));
}
linetype_t *XG_GetLumpLine(int id)
{
    for(int i = 0; i < num_linetypes; ++i)
    {
        if(linetypes[i].id == id)
            return linetypes + i;
    }
    return 0; // Not found.
}
sectortype_t *XG_GetLumpSector(int id)
{
    for(int i = 0; i < num_sectypes; ++i)
    {
        if(sectypes[i].id == id)
            return sectypes + i;
    }
    return 0; // Not found.
}
#endif
/** @file p_xgline.c Extended generalized line types.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 2007 Jamie Jones <jamie_jones_au@yahoo.com.au>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
#include <time.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "d_net.h"
#include "dmu_lib.h"
#include "gamesession.h"
#include "p_mapsetup.h"
#include "p_xgline.h"
#include "p_xgsec.h"
#include "p_actor.h"
#include "player.h"
#include "p_map.h"
#include "p_mapspec.h"
#include "p_terraintype.h"
#include "p_tick.h"
#include "p_sound.h"
#include "p_switch.h"
using namespace de;
#define XLTIMER_STOPPED 1    // Timer stopped.
#define EVTYPESTR(evtype) (evtype == XLE_CHAIN? "CHAIN" \
        : evtype == XLE_CROSS? "CROSS" \
        : evtype == XLE_USE? "USE" \
        : evtype == XLE_SHOOT? "SHOOT" \
        : evtype == XLE_HIT? "HIT" \
        : evtype == XLE_TICKER? "TICKER" \
        : evtype == XLE_AUTO? "AUTO" \
        : evtype == XLE_FORCED? "FORCED" \
        : evtype == XLE_FUNC? "FUNCTION" : "???")
#define LREFTYPESTR(reftype) (reftype == LREF_NONE? "NONE" \
        : reftype == LREF_SELF? "SELF" \
        : reftype == LREF_TAGGED? "TAGGED LINES" \
        : reftype == LREF_LINE_TAGGED? "LINE TAGGED LINES" \
        : reftype == LREF_ACT_TAGGED? "ACT TAGGED LINES" \
        : reftype == LREF_INDEX? "INDEXED LINE" \
        : reftype == LREF_ALL? "ALL LINES" : "???")
#define LPREFTYPESTR(reftype) (reftype == LPREF_NONE? "NONE" \
        : reftype == LPREF_MY_FLOOR? "MY FLOOR" \
        : reftype == LPREF_TAGGED_FLOORS? "TAGGED FLOORS" \
        : reftype == LPREF_LINE_TAGGED_FLOORS? "LINE TAGGED FLOORS" \
        : reftype == LPREF_ACT_TAGGED_FLOORS? "ACT TAGGED FLOORS" \
        : reftype == LPREF_INDEX_FLOOR? "INDEXED FLOOR" \
        : reftype == LPREF_ALL_FLOORS? "ALL FLOORS" \
        : reftype == LPREF_MY_CEILING? "MY CEILING" \
        : reftype == LPREF_TAGGED_CEILINGS? "TAGGED CEILINGS" \
        : reftype == LPREF_LINE_TAGGED_CEILINGS? "LINE TAGGED CEILINGS" \
        : reftype == LPREF_ACT_TAGGED_CEILINGS? "ACT TAGGED CEILINGS" \
        : reftype == LPREF_INDEX_CEILING? "INDEXED CEILING" \
        : reftype == LPREF_ALL_CEILINGS? "ALL CEILINGS" \
        : reftype == LPREF_SPECIAL? "SPECIAL" \
        : reftype == LPREF_BACK_FLOOR? "BACK FLOOR" \
        : reftype == LPREF_BACK_CEILING? "BACK CEILING" \
        : reftype == LPREF_THING_EXIST_FLOORS? "SECTORS WITH THING - FLOOR" \
        : reftype == LPREF_THING_EXIST_CEILINGS? "SECTORS WITH THING - CEILING" \
        : reftype == LPREF_THING_NOEXIST_FLOORS? "SECTORS WITHOUT THING - FLOOR" \
        : reftype == LPREF_THING_NOEXIST_CEILINGS? "SECTORS WITHOUT THING - CEILING" : "???")
#define LSREFTYPESTR(reftype) (reftype == LSREF_NONE? "NONE" \
        : reftype == LSREF_MY? "MY SECTOR" \
        : reftype == LSREF_TAGGED? "TAGGED SECTORS" \
        : reftype == LSREF_LINE_TAGGED? "LINE TAGGED SECTORS" \
        : reftype == LSREF_ACT_TAGGED? "ACT TAGGED SECTORS" \
        : reftype == LSREF_INDEX? "INDEXED SECTOR" \
        : reftype == LSREF_ALL? "ALL SECTORS" \
        : reftype == LSREF_BACK? "BACK SECTOR" \
        : reftype == LSREF_THING_EXIST? "SECTORS WITH THING" \
        : reftype == LSREF_THING_NOEXIST? "SECTORS WITHOUT THING" : "???")
#define TO_DMU_TOP_COLOR(x) ((x) == 0? DMU_TOP_COLOR_RED \
        : (x) == 1? DMU_TOP_COLOR_GREEN \
        : DMU_TOP_COLOR_BLUE)
#define TO_DMU_MIDDLE_COLOR(x) ((x) == 0? DMU_MIDDLE_COLOR_RED \
        : (x) == 1? DMU_MIDDLE_COLOR_GREEN \
        : (x) == 2? DMU_MIDDLE_COLOR_BLUE \
        : DMU_MIDDLE_ALPHA)
#define TO_DMU_BOTTOM_COLOR(x) ((x) == 0? DMU_BOTTOM_COLOR_RED \
        : (x) == 1? DMU_BOTTOM_COLOR_GREEN \
        : DMU_BOTTOM_COLOR_BLUE)
void XL_ChangeMaterial(Line *line, int sidenum, int section, world_Material *mat,
    blendmode_t blend = BM_NORMAL, Vector4f const &tintColor = Vector4f(), int flags = 0);
int XL_DoChainSequence(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XL_DoDamage(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XL_DoPower(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XL_DoKey(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XL_DoExplode(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XL_DoCommand(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_ChangeLineType(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_Activate(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_Music(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_LineCount(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_LeaveMap(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_DisableLine(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_EnableLine(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_ChangeWallMaterial(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int XLTrav_LineTeleport(Line *line, dd_bool ceiling, void *context, void *context2, mobj_t *activator);
int xgDev = 0; // Print dev messages.
static linetype_t typebuffer;
static char msgbuf[80];
ThinkerT<mobj_s> dummyThing;
struct mobj_s *XG_DummyThing()
{
    return dummyThing;
}
/* ADD NEW XG CLASSES TO THE END - ORIGINAL INDICES MUST STAY THE SAME!!! */
xgclass_t xgClasses[NUMXGCLASSES] =
{
    { NULL, NULL, TRAV_NONE, 0, 1, 0, "None",
      {{XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
      }
    },
    { de::function_cast<int (*)()>(XL_DoChainSequence), NULL, TRAV_NONE, 0, 1, 0, "Chain Sequence",
       {{XGPF_INT, "Chain Flags", "chsf_", 0},              // ip0: (chsf_) chain sequence flags
        {XGPF_INT, "Line Type 0", "", -1},                  // ip1: Type to execute
        {XGPF_INT, "Line Type 1", "", -1},                  // ip2:  ""  ""  ""
        {XGPF_INT, "Line Type 2", "", -1},                  // ip3:  ""  ""  ""
        {XGPF_INT, "Line Type 3", "", -1},                  // ip4:  ""  ""  ""
        {XGPF_INT, "Line Type 4", "", -1},                  // ip5:  ""  ""  ""
        {XGPF_INT, "Line Type 5", "", -1},                  // ip6:  ""  ""  ""
        {XGPF_INT, "Line Type 6", "", -1},                  // ip7:  ""  ""  ""
        {XGPF_INT, "Line Type 7", "", -1},                  // ip8:  ""  ""  ""
        {XGPF_INT, "Line Type 8", "", -1},                  // ip9:  ""  ""  ""
        {XGPF_INT, "Line Type 9", "", -1},                  // ip10: ""  ""  ""
        {XGPF_INT, "Line Type 10", "", -1},                 // ip11: ""  ""  ""
        {XGPF_INT, "Line Type 11", "", -1},                 // ip12: ""  ""  ""
        {XGPF_INT, "Line Type 12", "", -1},                 // ip13: ""  ""  ""
        {XGPF_INT, "Line Type 13", "", -1},                 // ip14: ""  ""  ""
        {XGPF_INT, "Line Type 14", "", -1},                 // ip15: ""  ""  ""
        {XGPF_INT, "Line Type 15", "", -1},                 // ip16: ""  ""  ""
        {XGPF_INT, "Line Type 16", "", -1},                 // ip17: ""  ""  ""
        {XGPF_INT, "Line Type 17", "", -1},                 // ip18: ""  ""  ""
        {XGPF_INT, "Line Type 18", "", -1}                  // ip19: ""  ""  ""
       }
    },
    { de::function_cast<int (*)()>(XSTrav_MovePlane), XS_InitMovePlane, TRAV_PLANES, 0, 1, 0, "Move Plane",
       {{XGPF_INT, "Target Ref", "lpref_", 0},              // ip0: (plane ref) plane(s) to move.
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Destination Ref", "spref_", 2},         // ip2: destination height type (zero, relative to current, surrounding highest/lowest floor/ceiling)
        {XGPF_INT, "Move Flags", "pmf_", 3},                // ip3: flags (PMF_*)
        {XGPF_INT, "Start Sound", "", 4 | MAP_SND},         // ip4: start sound
        {XGPF_INT, "End Sound", "", 5 | MAP_SND},           // ip5: end sound
        {XGPF_INT, "Move Sound", "", 6 | MAP_SND},          // ip6: move sound
        {XGPF_INT, "Start Material Ref", "spref_", 7},       // ip7: start texture origin (uses same ids as i2) (spec: use ip8 as tex num)
        {XGPF_INT, "Start Material Num", "", 8 | MAP_MATERIAL}, // ip8: data component or number/name of flat
        {XGPF_INT, "End Material Ref", "spref_", 9},         // ip9: end texture origin (uses same ids as i2) (spec: use ip10 as tex num)
        {XGPF_INT, "End Material Num", "", 10 | MAP_MATERIAL}, // ip10: data component or number/name of flat
        {XGPF_INT, "Start Type Ref", "lpref_", 11},         // ip11: (plane ref) start sector type (spec: use i12 as type ID)
        {XGPF_INT, "Start Type Num", "", -1},               // ip12: data component or type ID
        {XGPF_INT, "End Type Ref", "lpref_", 13},           // ip13: (plane ref) end sector type (spec: use i14 as type ID)
        {XGPF_INT, "End Type Num", "", -1},                 // ip14: data component or type ID
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XSTrav_BuildStairs), XS_InitStairBuilder, TRAV_PLANES, 0, 1, 0, "Build Stairs",
       {{XGPF_INT, "Target Ref", "lpref_", 0},              // ip0: (plane ref) plane to start from
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Spread Material", "", -1},               // ip2: (true/false) stop when texture changes
        {XGPF_INT, "Spread Build", "", -1},                 // ip3: (true/false) spread build?
        {XGPF_INT, "Start Sound", "", 4 | MAP_SND},         // ip4: start build sound (doesn't wait)
        {XGPF_INT, "Step Start Sound", "", 5 | MAP_SND},    // ip5: step start sound
        {XGPF_INT, "Step End Sound", "", 6 | MAP_SND},      // ip6: step end sound
        {XGPF_INT, "Step Move Sound", "", 7 | MAP_SND},     // ip7: step move sound
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XL_DoDamage), NULL, TRAV_NONE, 0, 1, 0, "Damage",
       {{XGPF_INT, "Min Delta", "", -1},                    // ip0: min damage delta
        {XGPF_INT, "Max Delta", "", -1},                    // ip1: max damage delta
        {XGPF_INT, "Min Limit", "", -1},                    // ip2: min limit (wont damage if health bellow)
        {XGPF_INT, "Max Limit", "", -1},                    // ip3: max limit (wont damage if health above)
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
      }
    },
    { de::function_cast<int (*)()>(XL_DoPower), NULL, TRAV_NONE, 0, 1, 0, "Power",
       {{XGPF_INT, "Min Delta", "", -1},                    // ip0: min power delta
        {XGPF_INT, "Max Delta", "", -1},                    // ip1: max power delta
        {XGPF_INT, "Min Limit", "", -1},                    // ip2: min limit
        {XGPF_INT, "Max Limit", "", -1},                    // ip3: max limit
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_ChangeLineType), NULL, TRAV_LINES, 0, 1, 0, "Line Type",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Line Type", "", -1},                    // ip2: new type (must be an XG line type)
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XSTrav_SectorType), NULL, TRAV_SECTORS, 0, 1, 0, "Sector Type",
       {{XGPF_INT, "Target Ref", "lsref_", 0},              // ip0: (sector ref) sector(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Sector Type", "", -1},                  // ip2: new type (zero or an XG sector type)
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XSTrav_SectorLight), NULL, TRAV_SECTORS, 0, 1, 0, "Sector Light",
       {{XGPF_INT, "Target Ref", "lsref_", 0},              // ip0: (sector ref) sector(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Change Light", "", -1},                 // ip2: if non-zero light level will be changed
        {XGPF_INT, "Change Color", "", -1},                 // ip3: if non-zero colour will be changed
        {XGPF_INT, "Light Ref", "lightref_", 4},            // ip4: (light ref) sector to get the initial light delta from.
        {XGPF_INT, "Light Delta", "", -1},                  // ip5: offset to the delta or absolute value
        {XGPF_INT, "Color Ref", "lightref_", 6},            // ip6: (light ref) sector to get the initial colour deltas from.
        {XGPF_INT, "Red Delta", "", -1},                    // ip7: offset to red delta
        {XGPF_INT, "Green Delta", "", -1},                  // ip8: offset to green delta
        {XGPF_INT, "Blue Delta", "", -1},                   // ip9: offset to blue delta
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_Activate), NULL, TRAV_LINES, 0, 1, 0, "Activate",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to activate
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XL_DoKey), NULL, TRAV_NONE, 0, 1, 0, "Key",
       {{XGPF_INT, "Give Keys", "", -1},                    // ip0: keys to give
        {XGPF_INT, "Take Keys", "", -1},                    // ip1: keys to take away.
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_Music), NULL, TRAV_LINES, 2, 3, 0, "Music",
       {{XGPF_INT, "Song ID", "ldref_", 0 | MAP_MUS},       // ip0: song id/name or (line data ref from ip2)
        {XGPF_INT, "Play Looped", "", -1},                  // ip1: non-zero means play looped
        {XGPF_INT, "Data Ref", "lref_", 2},                 // ip2: (line ref) used with line data ref eg set music track to line-tag
        {XGPF_INT, "Data Num", "", -1},                     // ip3:
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_LineCount), NULL, TRAV_LINES, 0, 1, 0, "Line Count",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Set Absolute", "", -1},                 // ip2: non-zero makes ip3 absolute
        {XGPF_INT, "Count Delta", "", -1},                  // ip3: count delta or absolute
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_LeaveMap), NULL, TRAV_LINES, 1, 2, 0, "Leave Map",
       {{XGPF_INT, "Secret Exit", "", -1},                  // ip0: non-zero goto secret map
        {XGPF_INT, "Data Ref", "lref_", 1},                 // ip1: (line ref) line to acquire (line data ref) from
        {XGPF_INT, "Data Num", "", -1},                     // ip2:
        {XGPF_INT, "Goto Map", "ldref_", 3},                // ip3: map ID or (line data ref from ip1)
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_DisableLine), NULL, TRAV_LINES, 0, 1, 0, "Disable Line",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to disable
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_EnableLine), NULL, TRAV_LINES, 0, 1, 0, "Enable Line",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to enable
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XL_DoExplode), NULL, TRAV_NONE, 0, 1, 0, "Explode",
      {{XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
      }
    },
    { de::function_cast<int (*)()>(XSTrav_PlaneMaterial), NULL, TRAV_PLANES, 0, 1, 0, "Plane Material",
       {{XGPF_INT, "Target Ref", "lpref_", 0},              // ip0 : (plane ref) plane(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1 : ref data
        {XGPF_INT, "Material Ref", "spref_", 2},            // ip2 : Texture ref
        {XGPF_INT, "Material Num", "", 3 | MAP_MATERIAL},   // ip3 : texture number (flat), used with SPREF_NONE
        {XGPF_INT, "Red Delta", "", -1},                    // ip4 : plane surface color (red)
        {XGPF_INT, "Green Delta", "", -1},                  // ip5 : "" (green)
        {XGPF_INT, "Blue Delta", "", -1},                   // ip6 : "" (blue)
        {XGPF_INT, "Change Color", "", -1},                 // ip7: if non-zero tint color will be changed
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_ChangeWallMaterial), NULL, TRAV_LINES, 0, 1, 0, "Wall Material",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) line(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Side Num", "", -1},                     // ip2: non-zero change the back side
        {XGPF_INT, "Top Material", "", 3 | MAP_MATERIAL},    // ip3: top texture to change to (blank indicates no change)
        {XGPF_INT, "Middle Material", "", 4 | MAP_MATERIAL}, // ip4: middle texture to change to (blank indicates no change)
        {XGPF_INT, "Bottom Material", "", 5 | MAP_MATERIAL}, // ip5: bottom texture to change to (blank indicates no change)
        {XGPF_INT, "Set Mid If None", "", -1},              // ip6: set mid texture even if previously zero
        {XGPF_INT, "Sidedef Flags", "sdf_", 7},             // ip7: (sdf_) side flags (used with surface colour blending, fullbright etc)
        {XGPF_INT, "Middle Blendmode", "bm_", 8},           // ip8: (bm_) middle texture blendmode
        {XGPF_INT, "Top Red Delta", "", -1},                // ip9:
        {XGPF_INT, "Top Green Delta", "", -1},              // ip10:
        {XGPF_INT, "Top Blue Delta", "", -1},               // ip11:
        {XGPF_INT, "Middle Red Delta", "", -1},             // ip12:
        {XGPF_INT, "Middle Green Delta", "", -1},           // ip13:
        {XGPF_INT, "Middle Blue Delta", "", -1},            // ip14:
        {XGPF_INT, "Middle Alpha Delta", "", -1},           // ip15:
        {XGPF_INT, "Bottom Red Delta", "", -1},             // ip16:
        {XGPF_INT, "Bottom Green Delta", "", -1},           // ip17:
        {XGPF_INT, "Bottom Blue Delta", "", -1},            // ip18:
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XL_DoCommand), NULL, TRAV_NONE, 0, 1, 0, "Command",
      {{XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
       {XGPF_INT, "", "", -1},
      }
    },
    { de::function_cast<int (*)()>(XSTrav_SectorSound), NULL, TRAV_SECTORS, 0, 1, 0, "Sector Sound",
       {{XGPF_INT, "Target Ref", "lsref_", 0},              // ip0: (sector ref) sector(s) to play the sound in
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Sound ID", "", 2 | MAP_SND},            // ip2: sound name/id to play
        {XGPF_INT, "Origin", "", -1},                       // ip3: non-zero = play from a specific origin (1=floor, 2=ceiling) else 0=center
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XSTrav_MimicSector), NULL, TRAV_SECTORS, 0, 1, 0, "Mimic Sector",
       {{XGPF_INT, "Target Ref", "lsref_", 0},              // ip0: (sector ref) sector(s) to change
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "Mimic Ref", "spref_", 2},               // ip2: (spref) sector to mimic
        {XGPF_INT, "Mimic Num", "", -1},                    // ip3:
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XSTrav_Teleport), NULL, TRAV_SECTORS, 0, 1, 0, "Teleport",
       {{XGPF_INT, "Target Ref", "lsref_", 0},              // ip0: (sector ref) sector(s) to teleport to (first acceptable target is used)
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "No Flash", "", -1},                     // ip2: non-zero = no flash (or sound)
        {XGPF_INT, "No Sound", "", -1},                     // ip3: non-zero = no sound
        {XGPF_INT, "Always Stomp", "", -1},                 // ip4: non-zero = Always telefrag
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    },
    { de::function_cast<int (*)()>(XLTrav_LineTeleport), NULL, TRAV_LINES, 0, 1, 1 | XLE_CROSS, "Line Teleport",
       {{XGPF_INT, "Target Ref", "lref_", 0},               // ip0: (line ref) teleport destination
        {XGPF_INT, "Target Num", "", -1},                   // ip1:
        {XGPF_INT, "No Flash", "", -1},                     // ip2: non-zero = spawn MT_TFOG
        {XGPF_INT, "Teleport Sound", "", 3 | MAP_SND},      // ip3: sound ID/name to play (or silent)
        {XGPF_INT, "Exit Side", "", -1},                    // ip4: non-zero = exit from the back of the target line
        {XGPF_INT, "Always Stomp", "", -1},                 // ip5: non-zero = Always telefrag
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1},
        {XGPF_INT, "", "", -1}
      }
    }
};
void XG_Register()
{
    C_VAR_INT("xg-dev", &xgDev, CVF_NO_ARCHIVE, 0, 1);
    C_CMD("movefloor",  0, MovePlane);
    C_CMD("moveceil",   0, MovePlane);
    C_CMD("movesec",    0, MovePlane);
}
/**
 * Init XG data for the map.
 */
void XG_Init(void)
{
    XL_Init(); // Init lines.
    XS_Init(); // Init sectors.
}
void XG_Ticker(void)
{
}
/**
 * This is called during an engine reset. Disables all XG functionality!
 */
void XG_Update(void)
{
    if(IS_CLIENT)
        return;
    XG_ReadTypes();
    XS_Update();
    XL_Update();
}
/**
 * Adds the given binary format line type to the generated types array
 */
int XL_AddAutoGenType(linetype_t * /*newtype*/)
{
    return true;
}
/**
 * Converts a line ID number to a line type (BOOM support).
 */
int XL_AutoGenType(int /*id*/, linetype_t * /*outptr*/)
{
    return false; // Cos we don't work yet.
#if 0
    linetype_t *l;
    if(!(id >= 12160 && id <= 32768))
        return false;
    if(1)
    {
        outptr = l;
        return true;
    }
    if(1) //XG format
    {
        /*
        l->id = id;
        l->flags = def->flags[0];
        l->flags2 = def->flags[1];
        l->flags3 = def->flags[2];
        l->lineClass = def->lineClass;
        l->actType = def->actType;
        l->actCount = def->actCount;
        l->actTime = def->actTime;
        for(i = 0; i < 10; i++)
        {
            if(i == 9)
                l->aparm[i] = Def_GetMobjNum(def->aparm9);
            else
                l->aparm[i] = def->aparm[i];
        }
        l->actSound = Friendly(Def_GetSoundNum(def->actSound));
        l->deactSound = Friendly(Def_GetSoundNum(def->deactSound));
        l->evChain = def->evChain;
        l->actChain = def->actChain;
        l->deactChain = def->deactChain;
        l->actLineType = def->actLineType;
        l->deactLineType = def->deactLineType;
        l->wallSection = def->wallSection;
        l->actTex = Friendly(R_CheckTextureNumForName(def->actTex));
        l->deactTex = Friendly(R_CheckTextureNumForName(def->deactTex));
        l->actMsg = def->actMsg;
        l->deactMsg = def->deactMsg;
        l->texMoveAngle = def->texMoveAngle;
        l->texMoveSpeed = def->texMoveSpeed;*/
        XL_AddAutoGenType(l);
        outptr = l;
    }
#if __JDOOM__ || __JDOOM64__
    else // BOOM format
    {
        outptr = l;
    }
#endif
    return true;
#endif
 }
/**
 * @return              @c true, if the type is defined.
 */
linetype_t* XL_GetType(int id)
{
    linetype_t*         ptr;
    char                buff[6];
    ptr = XG_GetLumpLine(id);
    if(ptr)
    {   // Got it!
        memcpy(&typebuffer, ptr, sizeof(*ptr));
        return &typebuffer;
    }
    dd_snprintf(buff, 6, "%i", id);
    if(Def_Get(DD_DEF_LINE_TYPE, buff, &typebuffer))
        return &typebuffer;
    if(XL_AutoGenType(id, &typebuffer))
        return &typebuffer;
    return NULL;
}
int XG_RandomInt(int min, int max)
{
    if(max == min) return max;
    float x = M_Random() / 256.0f; // Never reaches 1.
    return (int) (min + x * (max - min) + x);
}
float XG_RandomPercentFloat(float value, int percent)
{
    float i = (2 * M_Random() / 255.0f - 1) * percent / 100.0f;
    return value * (1 + i);
}
int findXLThinker(thinker_t *th, void *context)
{
    xlthinker_t *xl = (xlthinker_t *) th;
    if(xl->line == (Line *) context)
        return true; // Stop iteration, we've found it.
    return false; // Continue iteration.
}
void XL_SetLineType(Line *line, int id)
{
    LOG_AS("XL_SetLineType");
    xline_t *xline = P_ToXLine(line);
    if(XL_GetType(id))
    {
        xline->special = id;
        if(!xline->xg)
        {
            xline->xg = (xgline_t *)Z_Calloc(sizeof(xgline_t), PU_MAP, 0);
        }
        xline->xg->disabled    = false;
        xline->xg->timer       = 0;
        xline->xg->tickerTimer = 0;
        std::memcpy(&xline->xg->info, &typebuffer, sizeof(linetype_t));
        xline->xg->active      = (typebuffer.flags & LTF_ACTIVE) != 0;
        xline->xg->activator   = dummyThing;
        LOG_MAP_MSG_XGDEVONLY2("Line %i (%s), ID %i",
                P_ToIndex(line)
                << xgClasses[xline->xg->info.lineClass].className
                << id);
        if(!Thinker_Iterate(XL_Thinker, findXLThinker, line))
        {
            /*xlthinker_t *xl = (xlthinker_t *)Z_Calloc(sizeof(*xl), PU_MAP, 0);
            xl->thinker.function = XL_Thinker;*/
            ThinkerT<xlthinker_t> xl(Thinker::AllocateMemoryZone);
            xl.function = XL_Thinker;
            xl->line = line;
            Thinker_Add(xl.Thinker::take());
        }
    }
    else if(id)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i, type %i NOT DEFINED", P_ToIndex(line) << id);
    }
}
void XL_Init()
{
    dummyThing.Thinker::zap();
    if(IS_CLIENT) return;
    for(int i = 0; i < numlines; ++i)
    {
        Line *line = (Line *)P_ToPtr(DMU_LINE, i);
        P_ToXLine(line)->xg = 0;
        XL_SetLineType(line, P_ToXLine(line)->special);
    }
}
int XL_TraversePlanes(Line *line, int refType, int ref, void *data, void *context,
    dd_bool travsectors, mobj_t *activator, PlaneTraverserFunc func)
{
    LOG_AS(travsectors? "XL_TraverseSectors" : "XL_TraversePlanes");
    int tag = 0;
    mobj_t *mo = NULL;
    dd_bool ok, findSecTagged;
    if(xgDev)
    {
        char buff[50];
        if(ref)
        {
            sprintf(buff, ": %i", ref);
        }
        LOG_MAP_MSG_XGDEVONLY2("Line %i, ref (%s%s)",
               P_ToIndex(line)
               << (travsectors? LSREFTYPESTR(refType) : LPREFTYPESTR(refType))
               << (ref? buff: ""));
    }
    if(refType == LPREF_NONE)
        return false; // This is not a reference!
    Sector *frontSec = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
    Sector *backSec  = (Sector *)P_GetPtrp(line, DMU_BACK_SECTOR);
    if(refType == LPREF_MY_FLOOR || refType == LPREF_MY_CEILING)
    {
        if(frontSec)
        {
            return func(frontSec, refType == LPREF_MY_CEILING? true:false,
                        data, context, activator);
        }
        LOG_MAP_MSG_XGDEVONLY2("Line %i has no front sector!", P_ToIndex(line));
    }
    if(refType == LPREF_BACK_FLOOR || refType == LPREF_BACK_CEILING)
    {
        if(backSec)
        {
            return func(backSec, refType == LPREF_BACK_CEILING? true:false,
                        data, context, activator);
        }
        LOG_MAP_MSG_XGDEVONLY2("Line %i has no back sector!", P_ToIndex(line));
    }
    if(refType == LPREF_INDEX_FLOOR)
    {
        return func((Sector *)P_ToPtr(DMU_SECTOR, ref), false, data, context,
                    activator);
    }
    if(refType == LPREF_INDEX_CEILING)
    {
        return func((Sector *)P_ToPtr(DMU_SECTOR, ref), true, data, context,
                    activator);
    }
    findSecTagged = false;
    if(refType == LPREF_TAGGED_FLOORS || refType == LPREF_TAGGED_CEILINGS)
    {
        findSecTagged = true;
        tag = ref;
    }
    else if(refType == LPREF_LINE_TAGGED_FLOORS ||
            refType == LPREF_LINE_TAGGED_CEILINGS)
    {
        findSecTagged = true;
        tag = P_ToXLine(line)->tag;
    }
    if(findSecTagged)
    {
        if(iterlist_t *list = P_GetSectorIterListForTag(tag, false))
        {
            IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
            IterList_RewindIterator(list);
            Sector *sec;
            while((sec = (Sector *)IterList_MoveIterator(list)) != NULL)
            {
                if(refType == LPREF_TAGGED_FLOORS || refType == LPREF_TAGGED_CEILINGS)
                {
                    if(!func(sec, refType == LPREF_TAGGED_CEILINGS, data,
                             context, activator))
                    {
                        return false;
                    }
                }
                if(refType == LPREF_LINE_TAGGED_FLOORS ||
                   refType == LPREF_LINE_TAGGED_CEILINGS)
                {
                    if(!func(sec, refType == LPREF_LINE_TAGGED_CEILINGS,
                             data, context, activator))
                    {
                        return false;
                    }
                }
            }
        }
    }
    else
    {
        for(int i = 0; i < numsectors; ++i)
        {
            Sector *sec     = (Sector *)P_ToPtr(DMU_SECTOR, i);
            xsector_t *xsec = P_ToXSector(sec);
            if(refType == LPREF_ALL_FLOORS || refType == LPREF_ALL_CEILINGS)
            {
                if(!func(sec, refType == LPREF_ALL_CEILINGS, data,
                         context, activator))
                {
                    return false;
                }
            }
            if((refType == LPREF_ACT_TAGGED_FLOORS ||
                refType == LPREF_ACT_TAGGED_CEILINGS) && xsec->xg &&
                xsec->xg->info.actTag == ref)
            {
                if(!func(sec, refType == LPREF_ACT_TAGGED_CEILINGS, data,
                   context, activator))
                {
                   return false;
                }
            }
            if(refType == LPREF_THING_EXIST_FLOORS ||
               refType == LPREF_THING_EXIST_CEILINGS)
            {
                ok = true;
                for(mo = (mobj_t *)P_GetPtrp(sec, DMT_MOBJS); ok && mo; mo = mo->sNext)
                {
                    if(mo->type == P_ToXLine(line)->xg->info.aparm[9])
                    {
                        LOG_MAP_MSG_XGDEVONLY2("Thing of type %i found in sector id %i",
                               P_ToXLine(line)->xg->info.aparm[9] << i);
                        if(!func(sec, refType == LPREF_THING_EXIST_CEILINGS,
                                 data, context, activator))
                        {
                            return false;
                        }
                        ok = false;
                    }
                }
            }
            if(refType == LPREF_THING_NOEXIST_FLOORS ||
               refType == LPREF_THING_NOEXIST_CEILINGS)
            {
                ok = true;
                for(mo = (mobj_t *)P_GetPtrp(sec, DMT_MOBJS); ok && mo; mo = mo->sNext)
                {
                    if(mo->type != P_ToXLine(line)->xg->info.aparm[9])
                    {
                        continue;
                    }
                    ok = false;
                }
                if(ok)
                {
                    LOG_MAP_MSG_XGDEVONLY2("No things of type %i found in sector id %i",
                           P_ToXLine(line)->xg->info.aparm[9] << i);
                    if(!func(sec, refType == LPREF_THING_NOEXIST_CEILINGS,
                             data, context, activator))
                    {
                        return false;
                    }
                }
            }
        }
    }
    return true;
}
int XL_TraverseLines(Line* line, int rtype, int ref, void* data,
                     void* context, mobj_t* activator, LineTraverserFunc func)
{
    LOG_AS("XL_TraverseLines");
    int i;
    int tag = 0;
    int reftype = rtype;
    char buff[50];
    Line *iter = NULL;
    dd_bool findLineTagged;
    if(xgDataLumps)
        reftype+= 1;
    if(ref)
        sprintf(buff," : %i",ref);
    LOG_MAP_MSG_XGDEVONLY2("Line %i, ref (%s%s)",
           P_ToIndex(line) << LREFTYPESTR(reftype) << (ref? buff : ""));
    if(reftype == LREF_NONE)
        return func(NULL, true, data, context, activator); // Not a real reference
    if(reftype == LREF_SELF) // Traversing self is simple.
        return func(line, true, data, context, activator);
    if(reftype == LREF_INDEX)
        return func((Line *)P_ToPtr(DMU_LINE, ref), true, data, context, activator);
    findLineTagged = false;
    if(reftype == LREF_TAGGED)
    {
        findLineTagged = true;
        tag = ref;
    }
    else if(reftype == LREF_LINE_TAGGED)
    {
        findLineTagged = true;
        tag = P_ToXLine(line)->tag;
    }
    if(findLineTagged)
    {
        if(iterlist_t *list = P_GetLineIterListForTag(tag, false))
        {
            IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
            IterList_RewindIterator(list);
            while((iter = (Line *)IterList_MoveIterator(list)) != NULL)
            {
                if(reftype == LREF_TAGGED)
                {
                    if(!func(iter, true, data, context, activator))
                        return false;
                }
                else if(reftype == LREF_LINE_TAGGED)
                {
                    if(!ref || iter != line)
                    {
                        if(!func(iter, true, data, context, activator))
                            return false;
                    }
                }
            }
        }
    }
    else
    {
        for(i = 0; i < numlines; ++i)
        {
            iter = (Line *)P_ToPtr(DMU_LINE, i);
            if(reftype == LREF_ALL)
            {
                if(!func(iter, true, data, context, activator))
                    return false;
            }
            else if(reftype == LREF_ACT_TAGGED)
            {
                xline_t *xl = P_ToXLine(iter);
                if(xl->xg && xl->xg->info.actTag == ref)
                {
                    if(!func(iter, true, data, context, activator))
                        return false;
                }
            }
        }
    }
    return true;
}
/**
 * @return  Value as determined by the reference type from the specified line, using data
 *          from either the line itself or context (will always be linetype_t).
 */
int XL_ValidateLineRef(Line *line, int reftype, void * /*context*/, char const *parmname)
{
    LOG_AS("XL_ValidateLineRef");
    int answer = 0;
    Side *side;
    switch(reftype)
    {
    case LDREF_ID: // Line ID.
        answer = P_ToIndex(line);
        LOG_MAP_MSG_XGDEVONLY2("Using Line ID (%i) as %s", answer << parmname);
        break;
    case LDREF_SPECIAL: // Line Special.
        answer = P_ToXLine(line)->special;
        LOG_MAP_MSG_XGDEVONLY2("Using Line Special (%i) as %s", answer << parmname);
        break;
    case LDREF_TAG: // Line Tag.
        answer = P_ToXLine(line)->tag;
        LOG_MAP_MSG_XGDEVONLY2("Using Line Tag (%i) as %s", answer << parmname);
        break;
    case LDREF_ACTTAG: // Line ActTag.
        if(!P_ToXLine(line)->xg)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE NOT AN XG LINE");
            break;
        }
        if(!P_ToXLine(line)->xg->info.actTag)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE DOESNT HAVE AN ACT TAG");
            break;
        }
        answer = P_ToXLine(line)->xg->info.actTag;
        LOG_MAP_MSG_XGDEVONLY2("Using Line ActTag (%i) as %s", answer << parmname);
        break;
    case LDREF_COUNT: // Line count.
        if(!P_ToXLine(line)->xg)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE NOT AN XG LINE");
            break;
        }
        answer = P_ToXLine(line)->xg->info.actCount;
        LOG_MAP_MSG_XGDEVONLY2("Using Line Count (%i) as %s", answer << parmname);
        break;
    case LDREF_ANGLE: { // Line angle.
        coord_t d1[2];
        P_GetDoublepv(line, DMU_DXY, d1);
        answer = M_PointXYToAngle2(0, 0, d1[0], d1[1]) / (float) ANGLE_MAX * 360;
        LOG_MAP_MSG_XGDEVONLY2("Using Line Angle (%i) as %s", answer << parmname);
        break; }
    case LDREF_LENGTH: // Line length.
        answer = P_GetFixedp(line, DMU_LENGTH) >> FRACBITS;
        LOG_MAP_MSG_XGDEVONLY2("Using Line Length (%i) as %s", answer << parmname);
        break;
    case LDREF_TOP_OFFSETX:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_TOP_MATERIAL_OFFSET_X);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Top X Offset (%i) as %s", answer << parmname);
        break;
    case LDREF_TOP_OFFSETY:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_TOP_MATERIAL_OFFSET_Y);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Top Y Offset (%i) as %s", answer << parmname);
        break;
    case LDREF_MIDDLE_OFFSETX:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_MIDDLE_MATERIAL_OFFSET_X);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Middle X Offset (%i) as %s", answer << parmname);
        break;
    case LDREF_MIDDLE_OFFSETY:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_MIDDLE_MATERIAL_OFFSET_Y);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Middle Y Offset (%i) as %s", answer << parmname);
        break;
    case LDREF_BOTTOM_OFFSETX:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_BOTTOM_MATERIAL_OFFSET_X);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Bottom X Offset (%i) as %s", answer << parmname);
        break;
    case LDREF_BOTTOM_OFFSETY:
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(!side)
        {
            LOG_MAP_MSG_XGDEVONLY("REFERENCE MISSING FRONT SIDE!");
            break;
        }
        answer = P_GetIntp(side, DMU_BOTTOM_MATERIAL_OFFSET_Y);
        LOG_MAP_MSG_XGDEVONLY2("Using Line Bottom Y Offset (%i) as %s", answer << parmname);
        break;
    default:    // Could be explicit, return the actual int value
        answer = reftype;
        break;
    }
    return answer;
}
/**
 * Executes the lines' function as defined by its class.
 */
void XL_DoFunction(linetype_t *info, Line *line, int sideNum, mobj_t *actThing, int evType)
{
    DENG2_ASSERT(info && line);
    DENG2_ASSERT(info->lineClass >= 0 && info->lineClass < NUMXGCLASSES);
    LOG_AS("XL_DoFunction");
    xgclass_t *xgClass = &xgClasses[info->lineClass];
    LOG_MAP_MSG_XGDEVONLY2("Line %i, side %i, activator id %i, event %s",
            P_ToIndex(line) << sideNum << (actThing ? actThing->thinker.id : 0)
            << EVTYPESTR(evType));
    LOG_MAP_MSG_XGDEVONLY2("Executing class: %s (0x%X)...", xgClass->className << info->lineClass);
    if(xgClass->evTypeFlags > 0)
    {
        if(!(xgClass->evTypeFlags & evType))
        {
            LOG_MAP_MSG_XGDEVONLY2("THIS CLASS DOES NOT SUPPORT %s EVENTS!", EVTYPESTR(evType));
            return;
        }
    }
    if(xgClass->initFunc)
        xgClass->initFunc(line);
    if(xgClass->doFunc)
    {
        switch(xgClass->traverse)
        {
            case TRAV_NONE: // No need for traversal, call the doFunc
                de::function_cast<LineTraverserFunc>(xgClass->doFunc)(line, true, line, info, actThing);
                break;
            case TRAV_LINES: // Traverse lines, executing doFunc for each
                XL_TraverseLines(line, info->iparm[xgClass->travRef],
                                 info->iparm[xgClass->travData], line,
                                 info, actThing, de::function_cast<LineTraverserFunc>(xgClass->doFunc));
                break;
            case TRAV_PLANES: // Traverse planes, executing doFunc for each
            case TRAV_SECTORS:
                XL_TraversePlanes(line, info->iparm[xgClass->travRef],
                                  info->iparm[xgClass->travData], line,
                                  info, xgClass->traverse == TRAV_SECTORS? true : false,
                                  actThing, de::function_cast<PlaneTraverserFunc>(xgClass->doFunc));
                break;
        }
    }
}
int XLTrav_QuickActivate(Line *line, dd_bool /*ceiling*/, void *context, void * /*context2*/,
    mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            xline->xg->active = (context? true : false);
            xline->xg->timer  = XLTIMER_STOPPED; // Stop timer.
        }
    }
    return true; // Continue iteration.
}
/**
 * @return  @c true, if the line is active.
 */
int XLTrav_CheckLine(Line *line, dd_bool /*ceiling*/, void *context,
    void * /*context2*/, mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(!xline->xg)
        {
            return false; // Stop checking!
        }
        return CPP_BOOL(xline->xg->active) == (context? true : false);
    }
    return true; // Continue iteration.
}
/**
 * @param data  If @c true, the line will receive a chain event if inactive.
 *              If @c false, the line will receive a chain event if active.
 */
int XLTrav_SmartActivate(Line *line, dd_bool /*ceiling*/, void *context,
    void *context2, mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            if(CPP_BOOL(xline->xg->active) != (context? true : false))
            {
                XL_LineEvent(XLE_CHAIN, 0, line, 0, context2);
            }
        }
    }
    return true; // Continue iteration.
}
int XL_DoChainSequence(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            linetype_t *info = static_cast<linetype_t *>(context2);
            xline->xg->chIdx = 1; // This is the first.
            xline->xg->chTimer =
                XG_RandomPercentFloat(info->fparm[1], info->fparm[0]);
        }
    }
    return true;
}
int XL_DoDamage(Line * /*line*/, dd_bool /*ceiling*/, void * /*context*/, void *context2,
    mobj_t *activator)
{
    linetype_t *info = static_cast<linetype_t *>(context2);
    if(!activator)
    {
/*        if(evtype == XLE_FUNC)
            LOG_MAP_MSG_XGDEVONLY("Sector Functions don't have activators! Use a Sector Chain instead");
        else*/
            LOG_MAP_MSG_XGDEVONLY("No activator! Can't damage anything");
        return false;
    }
    if(activator->health > info->iparm[2])
    {
        int i = XG_RandomInt(info->iparm[0], info->iparm[1]);
        if(i > 0)
        {
            P_DamageMobj(activator, 0, 0, i, false);
        }
        else if(i < 0 && activator->health < info->iparm[3])
        {
            int origHealth = activator->health;
            activator->health = MIN_OF(activator->health-i, info->iparm[3]);
            if(activator->player && activator->health != origHealth)
            {
                activator->player->health = activator->health;
                activator->player->update |= PSF_HEALTH;
            }
        }
    }
    return true;
}
int XL_DoPower(Line * /*line*/, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t *activator)
{
    linetype_t *info = static_cast<linetype_t *>(context2);
    DENG2_ASSERT(info);
    player_t *player = activator? activator->player : nullptr;
    if(!player)
    {
/*        if(evtype == XLE_FUNC)
            LOG_MAP_MSG_XGDEVONLY("Sector Functions don't have activators! Use a Sector Chain instead");
        else*/
            LOG_MAP_MSG_XGDEVONLY("Activator MUST be a player...");
        return false;
    }
    int delta = XG_RandomInt(info->iparm[0], info->iparm[1]);
    if(delta > 0)
    {
        if(player->armorPoints + delta >= info->iparm[3])
            delta = de::max(0, info->iparm[3] - player->armorPoints);
    }
    else
    {
        if(player->armorPoints + delta <= info->iparm[2])
            delta = de::min(0, info->iparm[2] - player->armorPoints);
    }
    if(delta)
    {
        if(!player->armorType)
            P_PlayerSetArmorType(player, 1);
        P_PlayerGiveArmorBonus(player, delta);
    }
    return true;
}
int XL_DoKey(Line * /*line*/, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t *activator)
{
    linetype_t *info = static_cast<linetype_t *>(context2);
    player_t *player = (activator? activator->player : 0);
    if(!player)  // Must be a player.
    {
/*        if(evtype == XLE_FUNC)
            LOG_MAP_MSG_XGDEVONLY("Sector Functions don't have activators! Use a Sector Chain instead");
        else*/
            LOG_MAP_MSG_XGDEVONLY("Activator MUST be a player...");
        return false;
    }
    for(int i = 0; i < NUM_KEY_TYPES; ++i)
    {
        if(info->iparm[0] & (1 << i))
            P_GiveKey(player, keytype_t(i));
        if(info->iparm[1] & (1 << i))
            player->keys[i] = false;
    }
    return true;
}
int XL_DoExplode(Line * /*line*/, dd_bool /*ceiling*/, void * /*context*/,
    void * /*context2*/, mobj_t *activator)
{
    if(!activator)
    {
/*        if(evtype == XLE_FUNC)
            LOG_MAP_MSG_XGDEVONLY("Sector Functions don't have activators! Use a Sector Chain instead");
        else*/
            LOG_MAP_MSG_XGDEVONLY("No activator! Can't explode anything");
        return false;
    }
    P_ExplodeMissile(activator);
    return true;
}
int XL_DoCommand(Line * /*line*/, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    linetype_t *info = static_cast<linetype_t *>(context2);
    DD_Execute(true, info->sparm[0]);
    return true;
}
int XLTrav_ChangeLineType(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    if(line)
    {
        linetype_t *info = static_cast<linetype_t *>(context2);
        XL_SetLineType(line, info->iparm[2]);
    }
    return true; // Keep looking.
}
static Side *lineSideIfSector(Line &line, bool back = false)
{
    if(P_GetPtrp(&line, back? DMU_BACK_SECTOR : DMU_FRONT_SECTOR))
    {
        return (Side *)P_GetPtrp(&line, back? DMU_BACK : DMU_FRONT);
    }
    return nullptr;
}
int XLTrav_ChangeWallMaterial(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    LOG_AS("XLTrav_ChangeWallMaterial");
    linetype_t *info = static_cast<linetype_t *>(context2);
    if(!line) return true; // Continue iteration.
    Side *side = lineSideIfSector(*line, CPP_BOOL(info->iparm[2]));
    if(!side) return true; // Continue iteration.
    LOG_MAP_MSG_XGDEVONLY2("Line %i", P_ToIndex(line));
    XL_ChangeMaterial(line, info->iparm[2], LWS_UPPER,
                      (world_Material *)P_ToPtr(DMU_MATERIAL, info->iparm[3]),
                      BM_NORMAL, Vector3f(info->iparm[9], info->iparm[10], info->iparm[11]) / 255.f,
                      info->iparm[7]);
    world_Material *mat = 0;
    if(info->iparm[4] && (P_GetPtrp(side, DMU_MIDDLE_MATERIAL) || info->iparm[6]))
    {
        if(!P_GetPtrp(line, DMU_BACK_SECTOR) && info->iparm[4] == -1)
            mat = 0;
        else
            mat = (world_Material *)P_ToPtr(DMU_MATERIAL, info->iparm[4]);
    }
    XL_ChangeMaterial(line, info->iparm[2], LWS_MID, mat,
                      blendmode_t(info->iparm[8]),
                      Vector4f(info->iparm[12], info->iparm[13], info->iparm[14], info->iparm[15]) / 255.f,
                      info->iparm[7]);
    XL_ChangeMaterial(line, info->iparm[2], LWS_LOWER,
                      (world_Material *)P_ToPtr(DMU_MATERIAL, info->iparm[5]),
                      BM_NORMAL, Vector3f(info->iparm[16], info->iparm[17], info->iparm[18]) / 255.f,
                      info->iparm[7]);
    return true;
}
int XLTrav_Activate(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void * /*context2*/, mobj_t *activator)
{
    if(line)
    {
        XL_LineEvent(XLE_CHAIN, 0, line, 0, activator);
    }
    return true; // Keep looking.
}
int XLTrav_LineCount(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    linetype_t *info = static_cast<linetype_t *>(context2);
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            if(info->iparm[2])
            {
                xline->xg->info.actCount = info->iparm[3];
            }
            else
            {
                xline->xg->info.actCount += info->iparm[3];
            }
        }
    }
    return true;
}
int XLTrav_Music(Line *line, dd_bool /*ceiling*/, void * /*context*/,
    void *context2, mobj_t * /*activator*/)
{
    LOG_AS("XLTrav_Music");
    linetype_t *info = static_cast<linetype_t *>(context2);
    int song = 0;
    int temp = 0;
    if(info->iparm[2] == LREF_NONE)
    {    // The old format, use ip0 explicitly.
        song = info->iparm[0];
    }
    else // We might possibly have a data reference to evaluate.
    {
        if(info->iparm[2] == LREF_NONE) // (ip0) will be used to determine next map.
        {
            song = info->iparm[0];
        }
        else if(line)
        {
            temp = XL_ValidateLineRef(line, info->iparm[0], context2, "Music ID");
            if(!temp)
            {
                LOG_MAP_MSG_XGDEVONLY("Reference data not valid. Song not changed");
            }
            else
            {
                song = temp;
            }
        }
    }
    if(song)
    {
        LOG_MAP_MSG_XGDEVONLY2("Play Music ID (%i)%s",
                song << (info->iparm[1]? " looped" : ""));
        S_StartMusicNum(song, info->iparm[1]);
    }
    return false; // Only do this once!
}
int XLTrav_LineTeleport(Line *newLine, dd_bool /*ceiling*/, void *context,
    void *context2, mobj_t *mobj)
{
#define FUDGEFACTOR         10
    LOG_AS("XLTrav_LineTeleport");
    int fudge = FUDGEFACTOR;
    int side = 0, stepDown;
    uint an;
    mobj_t *flash;
    Line *line = (Line *) context;
    linetype_t *info = (linetype_t *) context2;
    Vertex /* *newV1,*/ *newV2, *oldV1/*, *oldV2*/;
    Sector *newFrontSec, *newBackSec;
    coord_t newPos[3], pos, s, c;
    coord_t oldLineDelta[2], newLineDelta[2];
    angle_t angle;
    if(mobj->flags2 & MF2_NOTELEPORT)
    {
        LOG_MAP_MSG_XGDEVONLY("Activator can't be teleported (THING is unteleportable)");
        return false; // No point continuing...
    }
    if(!line)
        return true; // Continue iteration.
    if(newLine == line)
    {
        LOG_MAP_MSG_XGDEVONLY("Target == Origin. Continuing search...");
        return true; // Keep looking
    }
    oldV1 = (Vertex *)P_GetPtrp(line, DMU_VERTEX0);
    P_GetDoublepv(line, DMU_DXY, oldLineDelta);
    newV2 = (Vertex *)P_GetPtrp(newLine, DMU_VERTEX1);
    P_GetDoublepv(newLine, DMU_DXY, newLineDelta);
    newFrontSec = (Sector *)P_GetPtrp(newLine, DMU_FRONT_SECTOR);
    newBackSec  = (Sector *)P_GetPtrp(newLine, DMU_BACK_SECTOR);
    LOG_MAP_MSG_XGDEVONLY2("%s, %s, %s",
           (   info->iparm[2]? "Spawn Flash" : "No Flash")
           << (info->iparm[3]? "Play Sound"  : "Silent")
           << (info->iparm[4]? "Reversed"    : "Normal."));
    if(info->iparm[2])
    {
        if((flash = P_SpawnMobj(MT_TFOG, mobj->origin, mobj->angle + ANG180, 0)))
        {
            if(info->iparm[3])
                S_StartSound(info->iparm[3], flash);
        }
    }
    if(fabs(oldLineDelta[0]) > fabs(oldLineDelta[1]))
        pos = (mobj->origin[VX] - P_GetDoublep(oldV1, DMU_X)) / oldLineDelta[0];
    else
        pos = (mobj->origin[VY] - P_GetDoublep(oldV1, DMU_Y)) / oldLineDelta[1];
    angle = (info->iparm[4] ? pos = 1 - pos, 0 : ANG180) +
             M_PointXYToAngle2(0, 0, newLineDelta[0], newLineDelta[1]) -
             M_PointXYToAngle2(0, 0, oldLineDelta[0], oldLineDelta[1]);
    newPos[VX] = P_GetDoublep(newV2, DMU_X) - (pos * newLineDelta[0]);
    newPos[VY] = P_GetDoublep(newV2, DMU_Y) - (pos * newLineDelta[1]);
    s = FIX2FLT(finesine[angle >> ANGLETOFINESHIFT]);
    c = FIX2FLT(finecosine[angle >> ANGLETOFINESHIFT]);
    if(P_GetDoublep(newFrontSec, DMU_FLOOR_HEIGHT) <
       P_GetDoublep(newBackSec, DMU_FLOOR_HEIGHT))
        stepDown = true;
    else
        stepDown = false;
    newPos[VZ] = mobj->origin[VZ] - mobj->floorZ;
    /**
     * Side to exit the line on positionally.
     *
     * @note
     * This flag concerns exit position, not momentum. Due to potential for
     * round-off error, the thing can land on either the left or the right
     * side of the exit line, and steps must be taken to make sure it
     * does not end up on the wrong side.
     *
     * Exit momentum is always towards side 1 in a reversed teleporter, and
     * always towards side 0 otherwise.
     *
     * Exiting positionally on side 1 is always safe, as far as avoiding
     * oscillations and stuck-in-wall problems, but may not be optimum for
     * non-reversed teleporters.
     *
     * Exiting on side 0 can cause oscillations if momentum is towards side
     * 1, as it is with reversed teleporters.
     *
     * Exiting on side 1 slightly improves player viewing when going down a
     * step on a non-reversed teleporter.
     */
    if(!info->iparm[4] || (mobj->player && stepDown))
        side = 1;
    while((Line_PointOnSide(newLine, newPos) < 0) != side && --fudge >= 0)
    {
        if(fabs(newLineDelta[0]) > fabs(newLineDelta[1]))
            newPos[VY] -= FIX2FLT(((newLineDelta[0] < 0) != side)? -1 : 1);
        else
            newPos[VX] += FIX2FLT(((newLineDelta[1] < 0) != side)? -1 : 1);
    }
    if(!P_TeleportMove(mobj, newPos[VX], newPos[VY], (info->iparm[5] > 0? true : false)))
    {
        LOG_MAP_MSG_XGDEVONLY("Something went horribly wrong... aborting.");
        return false;
    }
    if(stepDown)
        mobj->origin[VZ] = newPos[VZ] + P_GetDoublep(newFrontSec, DMU_FLOOR_HEIGHT);
    else
        mobj->origin[VZ] = newPos[VZ] + P_GetDoublep(newBackSec, DMU_FLOOR_HEIGHT);
    mobj->angle += angle;
    newPos[VX] = mobj->mom[MX];
    newPos[VY] = mobj->mom[MY];
    mobj->mom[MX] = (newPos[VX] * c) - (newPos[VY] * s);
    mobj->mom[MY] = (newPos[VY] * c) + (newPos[VX] * s);
    if(mobj->flags2 & MF2_FLOORCLIP)
    {
        mobj->floorClip = 0;
        if(FEQUAL(mobj->origin[VZ], P_GetDoublep(Mobj_Sector(mobj), DMU_FLOOR_HEIGHT)))
        {
            terraintype_t const *tt = P_MobjFloorTerrain(mobj);
            if(tt->flags & TTF_FLOORCLIP)
            {
                mobj->floorClip = 10;
            }
        }
    }
    if (!info->iparm[2])
    {
        an = mobj->angle >> ANGLETOFINESHIFT;
        if((flash = P_SpawnMobjXYZ(MT_TFOG,
                                   mobj->origin[VX] + 24 * FIX2FLT(finecosine[an]),
                                   mobj->origin[VY] + 24 * FIX2FLT(finesine[an]),
                                   mobj->origin[VZ], mobj->angle + ANG180, 0)))
        {
            if(info->iparm[3])
                S_StartSound(info->iparm[3], flash);
        }
    }
    if(mobj->player)
    {
        mobj->player->viewZ = mobj->origin[VZ] + mobj->player->viewHeight;
        mobj->dPlayer->flags |= DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM;
    }
    return false; // Do this only once!
#undef FUDGEFACTOR
}
int XLTrav_LeaveMap(Line *line, dd_bool /*ceiling*/, void * /*context*/, void *context2,
                    mobj_t * /*activator*/)
{
    LOG_AS("XLTrav_LeaveMap");
    linetype_t *info = static_cast<linetype_t *>(context2);
    if(info->iparm[0] > 0)
    {
        G_SetGameActionMapCompleted(gfw_Session()->mapUriForNamedExit("secret"), 0, true);
        return false;
    }
    de::Uri newMapUri;
    if(info->iparm[1] == LREF_NONE)
    {
        if(info->iparm[3])
        {
            newMapUri = G_ComposeMapUri(gfw_Session()->episodeId().toInt() - 1, info->iparm[3] - 1);
            LOG_MAP_MSG_XGDEVONLY2("Next map set to \"%s\"", newMapUri);
        }
    }
    else if(line)
    {
        int const oldMapNumber = XL_ValidateLineRef(line, info->iparm[3], context2, "Map Number");
        if(oldMapNumber > 0)
        {
            newMapUri = G_ComposeMapUri(gfw_Session()->episodeId().toInt() - 1, oldMapNumber - 1);
        }
    }
    if(newMapUri.isEmpty())
    {
        newMapUri = gfw_Session()->mapUriForNamedExit("next");
        LOG_MAP_MSG_XGDEVONLY("Next map set to default for the 'next' exit");
    }
    if(!P_MapExists(newMapUri.compose().toUtf8().constData()))
    {
        newMapUri = de::makeUri(gfw_Session()->episodeDef()->gets("startMap"));
    }
    G_SetGameActionMapCompleted(newMapUri);
    return false; // Only do this once!
}
int XLTrav_DisableLine(Line *line, dd_bool /*ceiling*/, void *context,
    void * /*context2*/, mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            xline_t *origLine = P_ToXLine((Line *) context);
            xline->xg->disabled = origLine->xg->active;
        }
    }
    return true; // Keep looking...
}
int XLTrav_EnableLine(Line *line, dd_bool /*ceiling*/, void *context,
    void * /*context2*/, mobj_t * /*activator*/)
{
    if(line)
    {
        xline_t *xline = P_ToXLine(line);
        if(xline->xg)
        {
            xline_t *origLine = P_ToXLine((Line*) context);
            xline->xg->disabled = !origLine->xg->active;
        }
    }
    return true; // Keep looking...
}
/**
 * Checks if the given lines are active or inactive.
 *
 * @return  @c true if all are in the specified state.
 */
dd_bool XL_CheckLineStatus(Line *line, int reftype, int ref, int active,
                           mobj_t *activator)
{
    return XL_TraverseLines(line, reftype, ref, &active, 0, activator, XLTrav_CheckLine);
}
int XL_CheckMobjGone(thinker_t *th, void *context)
{
    LOG_AS("XL_CheckMobjGone");
    mobj_t *mo    = (mobj_t *) th;
    int thingtype = *static_cast<int *>(context);
    if(mo->type == thingtype && mo->health > 0)
    {
        LOG_MAP_MSG_XGDEVONLY2("Thing type %i: Found mo id=%i, health=%i, pos=%s",
               thingtype << mo->thinker.id << mo->health << Vector4d(mo->origin).asText());
        return true; // Stop iteration.
    }
    return false; // Continue iteration.
}
void XL_SwapSwitchTextures(Line *line, int snum)
{
    LOG_AS("XL_SwapSwitchTextures");
    if(line)
    {
        Side *side = (Side *)P_GetPtrp(line, snum? DMU_BACK : DMU_FRONT);
        if(side && P_ToggleSwitch(side, SFX_NONE, true, 0))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i, side %i",
                   P_ToIndex(line) << P_ToIndex(side));
        }
    }
}
/**
 * Changes material of the given line.
 */
void XL_ChangeMaterial(Line *line, int sidenum, int section, world_Material *mat,
    blendmode_t blendmode, Vector4f const &tintColor, int flags)
{
    Side *side = (Side *)P_GetPtrp(line, sidenum? DMU_BACK:DMU_FRONT);
    if(!side) return;
    LOG_MAP_MSG_XGDEVONLY2("Line:%i side:%i section:%i material:%i tintColor:%s blendmode:%i",
           P_ToIndex(line) << sidenum << section << P_ToIndex(mat) << tintColor.asText() << blendmode);
    if(section == LWS_MID)
    {
        if(mat == (world_Material *) -1)
        {
            P_SetPtrp(side, DMU_MIDDLE_MATERIAL, NULL);
        }
        else if(mat)
        {
            P_SetPtrp(side, DMU_MIDDLE_MATERIAL, mat);
        }
        if(blendmode)
        {
            P_SetIntp(side, DMU_MIDDLE_BLENDMODE, blendmode);
        }
        for(int i = 0; i < 4; ++i)
        {
            if(!de::fequal(tintColor[i], 0))
                P_SetFloatp(side, TO_DMU_MIDDLE_COLOR(i), tintColor[i]);
        }
    }
    else if(section == LWS_UPPER)
    {
        if(mat)
        {
            P_SetPtrp(side, DMU_TOP_MATERIAL, mat);
        }
        for(int i = 0; i < 3; ++i)
        {
            if(!de::fequal(tintColor[i], 0))
                P_SetFloatp(side, TO_DMU_TOP_COLOR(i), tintColor[i]);
        }
    }
    else if(section == LWS_LOWER)
    {
        if(mat)
        {
            P_SetPtrp(side, DMU_BOTTOM_MATERIAL, mat);
        }
        for(int i = 0; i < 3; ++i)
        {
            if(!de::fequal(tintColor[i], 0))
                P_SetFloatp(side, TO_DMU_BOTTOM_COLOR(i), tintColor[i]);
        }
    }
    P_SetIntp(side, DMU_FLAGS, P_GetIntp(side, DMU_FLAGS) | flags);
}
void XL_Message(mobj_t *act, char *msg, dd_bool global)
{
    LOG_AS("XL_Message");
    player_t *pl;
    int i;
    if(!msg || !msg[0]) return;
    if(global)
    {
        LOG_MAP_MSG_XGDEVONLY2("GLOBAL '%s'", msg);
        for(i = 0; i < MAXPLAYERS; ++i)
        {
            if(players[i].plr->inGame)
                P_SetMessage(&players[i], msg);
        }
        return;
    }
    if(act->player)
    {
        pl = act->player;
    }
    else if((act->flags & MF_MISSILE) && act->target && act->target->player)
    {
        pl = act->target->player;
    }
    else
    {
        LOG_MAP_MSG_XGDEVONLY2("'%s'\nNO DESTINATION, MESSAGE DISCARDED", msg);
        return;
    }
    P_SetMessage(pl, msg);
}
void XL_ActivateLine(dd_bool activating, linetype_t *info, Line *line, int sidenum,
                     mobj_t *activator, int evtype)
{
    DENG2_ASSERT(line);
    LOG_AS("XL_ActivateLine");
    xline_t *xline = P_ToXLine(line);
    if(!xline) return; // huh?
    LOG_MAP_MSG_XGDEVONLY2("%s line %i, side %i, type %i",
           (activating? "Activating" : "Deactivating") << P_ToIndex(line)
           << sidenum << xline->special);
    DENG2_ASSERT(xline->xg);
    xgline_t &xgline = *xline->xg;
    if(xgline.disabled)
    {
        LOG_MAP_MSG_XGDEVONLY("LINE DISABLED, ABORTING");
        return; // The line is disabled.
    }
    if((activating && xgline.active) || (!activating && !xgline.active))
    {
        LOG_MAP_MSG_XGDEVONLY2("Line is ALREADY %s, ABORTING", (activating ? "ACTIVE" : "INACTIVE"));
        return; // Do nothing (can't activate if already active!).
    }
    xgline.activator = activator;
    if(activating && info->actChain)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line has Act Chain (type %i) - It will be processed first...", info->actChain);
        XL_LineEvent(XLE_CHAIN, info->actChain, line, sidenum, activator);
    }
    else if(!activating && info->deactChain)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line has Deact Chain (type %i) - It will be processed first...", info->deactChain);
        XL_LineEvent(XLE_CHAIN, info->deactChain, line, sidenum, activator);
    }
    if(xgline.active != activating)
    {
        XL_SwapSwitchTextures(line, sidenum);
    }
    xgline.active = activating;
    xgline.timer  = 0; // Reset timer.
    if( (activating && (info->flags2 & LTF2_GROUP_ACT)) ||
       (!activating && (info->flags2 & LTF2_GROUP_DEACT)))
    {
        XL_TraverseLines(line, LREF_LINE_TAGGED, true, &activating, 0, activator,
                         XLTrav_SmartActivate);
    }
    if(info->flags2 & LTF2_MULTIPLE)
    {
        XL_TraverseLines(line, LREF_LINE_TAGGED, true, &activating, 0, activator,
                         XLTrav_QuickActivate);
    }
    if(  (activating && (info->flags2 & LTF2_WHEN_ACTIVATED)) ||
        (!activating && (info->flags2 & LTF2_WHEN_DEACTIVATED)))
    {
        if(!(info->flags2 & LTF2_WHEN_LAST) || info->actCount == 1)
        {
            XL_DoFunction(info, line, sidenum, activator, evtype);
        }
        else
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i FUNCTION TEST FAILED", P_ToIndex(line));
        }
    }
    else if(activating)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i has no activation function", P_ToIndex(line));
    }
    else
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i has no deactivation function", P_ToIndex(line));
    }
    if(activating)
    {
        XL_Message(activator, info->actMsg,
                   (info->flags2 & LTF2_GLOBAL_A_MSG) != 0);
        if(info->actSound)
        {
            S_SectorSound((Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR), info->actSound);
        }
        if(info->wallSection && info->actMaterial != NOMATERIALID)
        {
            XL_ChangeMaterial(line, sidenum, info->wallSection,
                              (world_Material *)P_ToPtr(DMU_MATERIAL, info->actMaterial));
        }
        if(info->actLineType)
        {
            XL_SetLineType(line, info->actLineType);
        }
    }
    else
    {
        XL_Message(activator, info->deactMsg, (info->flags2 & LTF2_GLOBAL_D_MSG) != 0);
        if(info->deactSound)
        {
            S_SectorSound((Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR), info->deactSound);
        }
        if(info->wallSection && info->deactMaterial != NOMATERIALID)
        {
            XL_ChangeMaterial(line, sidenum, info->wallSection,
                              (world_Material *)P_ToPtr(DMU_MATERIAL, info->deactMaterial));
        }
        if(info->deactLineType)
        {
            XL_SetLineType(line, info->deactLineType);
        }
    }
}
/**
 * XL_CheckKeys
 */
dd_bool XL_CheckKeys(mobj_t* mo, int flags2, dd_bool doMsg, dd_bool doSfx)
{
    player_t*           act = mo->player;
#if __JDOOM__ || __JDOOM64__
    int                 num = 6;
    int*                keys = (int *) act->keys;
    int                 badsound = SFX_OOF;
#elif __JHERETIC__
    int                 num = 3;
    dd_bool*            keys = act->keys;
    int                 badsound = SFX_PLROOF;
#elif __JSTRIFE__
    int                 num = 3;
    int*                keys = (int *) act->keys;
    int                 badsound = SFX_NONE;
#endif
    int                 i;
    for(i = 0; i < num; ++i)
    {
        if((flags2 & LTF2_KEY(i)) && !keys[i])
        {   // This key is missing!
            if(doMsg)
            {
                sprintf(msgbuf, "YOU NEED A %s.", GET_TXT(TXT_KEY1 + i));
                XL_Message(mo, msgbuf, false);
            }
            if(doSfx)
                S_ConsoleSound(badsound, mo, act - players);
            return false;
        }
    }
    return true;
}
/**
 * Decides if the event leads to (de)activation.
 * Line must be extended.
 * Most conditions use AND (act method, game mode and difficult use OR).
 *
 * @return              @c true, iff the event is processed.
 */
int XL_LineEvent(int evtype, int linetype, Line* line, int sidenum,
                 void* data)
{
    LOG_AS("XL_LineEvent");
    int                 i;
    xline_t*            xline;
    xgline_t*           xg;
    linetype_t*         info;
    dd_bool             active;
    mobj_t*             activator_thing = (mobj_t *) data;
    player_t*           activator = 0;
    dd_bool             anyTrigger = false;
    if(IS_CLIENT)
        return false;
    xline = P_ToXLine(line);
    xg = xline->xg;
    info = &xg->info;
    active = xg->active;
    if(activator_thing)
        activator = activator_thing->player;
#if __JDOOM__ || __JDOOM64__
    if((xline->flags & ML_ALLTRIGGER) && !(info->flags2 & LTF2_OVERRIDE_ANY))
        anyTrigger = true;
#endif
    LOG_MAP_MSG_XGDEVONLY2("%s line %i, side %i (chained type %i)%s",
           EVTYPESTR(evtype) << P_ToIndex(line) << sidenum << linetype
           << (anyTrigger? " ANY Trigger" : ""));
    if(xg->disabled)
    {
        LOG_MAP_MSG_XGDEVONLY("LINE IS DISABLED, ABORTING EVENT");
        return false; // The line is disabled.
    }
    if(linetype)
    {
        if(!XL_GetType(linetype))
            return false;
        info = &typebuffer;
    }
    if(info->evChain)
    {
        if(XL_LineEvent(evtype, info->evChain, line, sidenum, data))
        {
            LOG_MAP_MSG_XGDEVONLY2("Event %s, line %i, side %i OVERRIDDEN BY EVENT CHAIN %i",
                   EVTYPESTR(evtype) << P_ToIndex(line) << sidenum << info->evChain);
            return true;
        }
    }
    if((active && info->actType == LTACT_COUNTED_OFF) ||
       (!active && info->actType == LTACT_COUNTED_ON))
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i: Active=%i, type=%i ABORTING EVENT",
               P_ToIndex(line) << active << info->actType);
        return false;
    }
    if(evtype == XLE_CHAIN || evtype == XLE_FUNC)
        goto type_passes;
    if(evtype == XLE_USE &&
       ((((info->flags & LTF_PLAYER_USE_A) && activator && !active) ||
         ((info->flags & LTF_OTHER_USE_A) && !activator && !active) ||
         ((info->flags & LTF_PLAYER_USE_D) && activator && active) ||
         ((info->flags & LTF_OTHER_USE_D) && !activator && active)) || anyTrigger))
        goto type_passes;
    if(evtype == XLE_SHOOT &&
       ((((info->flags & LTF_PLAYER_SHOOT_A) && activator && !active) ||
         ((info->flags & LTF_OTHER_SHOOT_A) && !activator && !active) ||
         ((info->flags & LTF_PLAYER_SHOOT_D) && activator && active) ||
         ((info->flags & LTF_OTHER_SHOOT_D) && !activator && active)) || anyTrigger))
        goto type_passes;
    if(evtype == XLE_CROSS &&
       ((((info->flags & LTF_PLAYER_CROSS_A) && activator && !active) ||
         ((info->flags & LTF_MONSTER_CROSS_A) &&
          (activator_thing && (activator_thing->flags & MF_COUNTKILL)) && !active) ||
         ((info->flags & LTF_MISSILE_CROSS_A) &&
          (activator_thing && (activator_thing->flags & MF_MISSILE)) && !active) ||
         ((info->flags & LTF_ANY_CROSS_A) && !active) ||
         ((info->flags & LTF_PLAYER_CROSS_D) && activator && active) ||
         ((info->flags & LTF_MONSTER_CROSS_D) &&
          (activator_thing && (activator_thing->flags & MF_COUNTKILL)) && active) ||
         ((info->flags & LTF_MISSILE_CROSS_D) &&
          (activator_thing && (activator_thing->flags & MF_MISSILE)) && active) ||
         ((info->flags & LTF_ANY_CROSS_D) && active)) || anyTrigger))
        goto type_passes;
    if(evtype == XLE_HIT &&
       ((((info->flags & LTF_PLAYER_HIT_A) && activator && !active) ||
         ((info->flags & LTF_OTHER_HIT_A) && !activator && !active) ||
         ((info->flags & LTF_MONSTER_HIT_A) &&
          (activator_thing && (activator_thing->flags & MF_COUNTKILL)) && !active) ||
         ((info->flags & LTF_MISSILE_HIT_A) &&
          (activator_thing && (activator_thing->flags & MF_MISSILE)) && !active) ||
         ((info->flags & LTF_ANY_HIT_A) && !active) ||
         ((info->flags & LTF_PLAYER_HIT_D) && activator && active) ||
         ((info->flags & LTF_OTHER_HIT_D) && !activator && active) ||
         ((info->flags & LTF_MONSTER_HIT_D) &&
          (activator_thing && (activator_thing->flags & MF_COUNTKILL)) && active) ||
         ((info->flags & LTF_MISSILE_HIT_D) &&
           (activator_thing && (activator_thing->flags & MF_MISSILE)) && active) ||
         ((info->flags & LTF_ANY_HIT_D) && active)) || anyTrigger))
        goto type_passes;
    if(evtype == XLE_TICKER &&
       (((info->flags & LTF_TICKER_A) && !active) ||
        ((info->flags & LTF_TICKER_D) && active)))
        goto type_passes;
    LOG_MAP_MSG_XGDEVONLY2("Line %i: ACT REQUIREMENTS NOT FULFILLED, ABORTING EVENT", P_ToIndex(line));
    return false;
  type_passes:
    if(info->flags & LTF_NO_OTHER_USE_SECRET)
    {
        if(evtype == XLE_USE && !activator && (xline->flags & ML_SECRET))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to no_other_use_secret", P_ToIndex(line));
            return false;
        }
    }
    if(info->flags & LTF_MOBJ_GONE)
    {
        if(Thinker_Iterate(P_MobjThinker, XL_CheckMobjGone, &info->aparm[9]))
            return false;
    }
    if(info->flags & LTF_ACTIVATOR_TYPE)
    {
        if(!activator_thing || activator_thing->type != info->aparm[9])
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to activator type", P_ToIndex(line));
            return false;
        }
    }
    if((evtype == XLE_USE || evtype == XLE_SHOOT || evtype == XLE_CROSS) &&
       !(info->flags2 & LTF2_TWOSIDED))
    {
        if(sidenum != 0)
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to line side test", P_ToIndex(line));
            return false;
        }
    }
    if(!info->actCount)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to Count = 0", P_ToIndex(line));
        return false;
    }
    if((info->flags2 & LTF2_HEALTH_ABOVE) && activator_thing &&
       activator_thing->health <= info->aparm[0])
        return false;
    if((info->flags2 & LTF2_HEALTH_BELOW) && activator_thing &&
       activator_thing->health >= info->aparm[1])
        return false;
    if((info->flags2 & LTF2_POWER_ABOVE) &&
       (!activator || activator->armorPoints <= info->aparm[2]))
        return false;
    if((info->flags2 & LTF2_POWER_BELOW) &&
       (!activator || activator->armorPoints >= info->aparm[3]))
        return false;
    if(info->flags2 & LTF2_LINE_ACTIVE)
        if(!XL_CheckLineStatus(line, info->aparm[4], info->aparm[5], true,
                               activator_thing))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to line_active test", P_ToIndex(line));
            return false;
        }
    if(info->flags2 & LTF2_LINE_INACTIVE)
        if(!XL_CheckLineStatus(line, info->aparm[6], info->aparm[7], false,
                               activator_thing))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to line_inactive test", P_ToIndex(line));
            return false;
        }
    if (IS_NETGAME)
    {
        const int netFlags = info->flags2 & (LTF2_COOPERATIVE | LTF2_DEATHMATCH);
        if (!netFlags)
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to netgame mode", P_ToIndex(line));
            return false;
        }
        if (netFlags != (LTF2_COOPERATIVE | LTF2_DEATHMATCH))
        {
            const auto netType = gfw_Rule(deathmatch);
            if (((netFlags & LTF2_COOPERATIVE) && netType != 0) ||
                ((netFlags & LTF2_DEATHMATCH) && netType == 0))
            {
                LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to non-matching deathmatch/coop flag", P_ToIndex(line));
                return false;
            }
        }
    }
    else
    {
        if(!(info->flags2 & LTF2_SINGLEPLAYER))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to game mode (1p)", P_ToIndex(line));
            return false;
        }
    }
    if(gfw_Rule(skill) < 1)
        i = 1;
    else if(gfw_Rule(skill) > 3)
        i = 4;
    else
        i = 1 << (gfw_Rule(skill) - 1);
    if(!(info->flags2 & (i << LTF2_SKILL_SHIFT)))
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to skill level (%i)",
               P_ToIndex(line) << gfw_Rule(skill));
        return false;
    }
    if(info->flags2 & LTF2_COLOR)
    {
        if(!activator)
            return false;
        if(cfg.playerColor[activator - players] != info->aparm[8])
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to activator color (%i)",
                   P_ToIndex(line) << cfg.playerColor[activator-players]);
            return false;
        }
    }
    if(info->
       flags2 & (LTF2_KEY1 | LTF2_KEY2 | LTF2_KEY3 | LTF2_KEY4 | LTF2_KEY5 |
                 LTF2_KEY6))
    {
        if(!activator)
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to missing key (no activator)", P_ToIndex(line));
            return false;
        }
        if(!XL_CheckKeys(activator_thing, info->flags2, true,
                         (evtype == XLE_USE? true : false)))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i: ABORTING EVENT due to missing key", P_ToIndex(line));
            return false;        // Keys missing!
        }
    }
    if(info->actCount > 0 && evtype != XLE_CHAIN && evtype != XLE_FUNC)
    {
        info->actCount--;
        LOG_MAP_MSG_XGDEVONLY2("Line %i: Decrementing counter, now %i",
               P_ToIndex(line) << info->actCount);
    }
    XL_ActivateLine(!active, info, line, sidenum, activator_thing, evtype);
    return true;
}
/**
 * @return              @c true, if the event was processed.
 */
int XL_CrossLine(Line* line, int sidenum, mobj_t* thing)
{
    if(!line || !P_ToXLine(line)->xg)
        return false;
    return XL_LineEvent(XLE_CROSS, 0, line, sidenum, thing);
}
/**
 * @return              @c true, if the event was processed.
 */
int XL_UseLine(Line* line, int sidenum, mobj_t* thing)
{
    if(!line || !P_ToXLine(line)->xg)
        return false;
    return XL_LineEvent(XLE_USE, 0, line, sidenum, thing);
}
/**
 * @return              @c true, if the event was processed.
 */
int XL_ShootLine(Line *line, int sidenum, mobj_t *thing)
{
    if(!line || !P_ToXLine(line)->xg)
        return false;
    return XL_LineEvent(XLE_SHOOT, 0, line, sidenum, thing);
}
int XL_HitLine(Line *line, int sidenum, mobj_t *thing)
{
    if(!line || !P_ToXLine(line)->xg)
        return false;
    return XL_LineEvent(XLE_HIT, 0, line, sidenum, thing);
}
void XL_DoChain(Line *line, int chain, dd_bool activating, mobj_t *actThing)
{
    LOG_AS("XL_DoChain");
    Line *dummyLineDef     = P_AllocDummyLine();
    xline_t *xdummyLineDef = P_ToXLine(dummyLineDef);
    xdummyLineDef->xg = (xgline_t *)Z_Calloc(sizeof(xgline_t), PU_MAP, 0);
    P_SetPtrp(dummyLineDef, DMU_FRONT_SECTOR, P_GetPtrp(line, DMU_FRONT_SECTOR));
    if(0 != P_GetPtrp(line, DMU_BACK))
    {
        P_SetPtrp(dummyLineDef, DMU_BACK_SECTOR, P_GetPtrp(line, DMU_BACK_SECTOR));
    }
    LOG_MAP_MSG_XGDEVONLY2("Line %i, chained type %i", P_ToIndex(line) << chain);
    LOG_MAP_MSG_XGDEVONLY2("(dummy line will show up as %i)", P_ToIndex(dummyLineDef));
    P_CopyLine(dummyLineDef, line);
    xdummyLineDef->xg->active = !activating;
    XL_LineEvent(XLE_CHAIN, chain, dummyLineDef, 0, actThing);
    Z_Free(xdummyLineDef->xg);
    P_FreeDummyLine(dummyLineDef);
}
/**
 * XG lines get to think.
 */
void XL_Thinker(void *xlThinkerPtr)
{
    DENG2_ASSERT(xlThinkerPtr);
    LOG_AS("XL_Thinker");
    xlthinker_t *xl = static_cast<xlthinker_t *>(xlThinkerPtr);
    Line *line      = xl->line;
    if(IS_CLIENT) return;
    if(!xl->line) return;
    xline_t *xline = P_ToXLine(line);
    if(!xline) return;
    xgline_t *xg = xline->xg;
    if(!xg) return;
    if(xg->disabled) return;
    linetype_t *info = &xg->info;
    float levtime = TIC2FLT(mapTime);
    if(xg->timer >= 0)
    {
        xg->timer++;
        xg->tickerTimer++;
    }
    if((info->tickerEnd <= 0 ||
        (levtime >= info->tickerStart && levtime <= info->tickerEnd)) &&
       xg->tickerTimer > info->tickerInterval)
    {
        if(info->flags & LTF_TICKER)
        {
            xg->tickerTimer = 0;
            XL_LineEvent(XLE_TICKER, 0, line, 0, XG_DummyThing());
        }
        if((((info->flags2 & LTF2_WHEN_ACTIVE) && xg->active) ||
            ((info->flags2 & LTF2_WHEN_INACTIVE) && !xg->active)) &&
            (!(info->flags2 & LTF2_WHEN_LAST) || info->actCount == 1))
        {
            XL_DoFunction(info, line, 0, (mobj_t *)xg->activator, XLE_FORCED);
        }
    }
    if(xg->active && info->lineClass == LTC_CHAIN_SEQUENCE)
    {
        xg->chTimer -= TIC2FLT(1);
        if(xg->chTimer < 0)
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i, executing...", P_ToIndex(line));
            if(xg->chIdx < DDLT_MAX_PARAMS && info->iparm[xg->chIdx])
            {
                XL_DoChain(line, info->iparm[xg->chIdx], true, (mobj_t *)xg->activator);
                xg->chIdx++;
                if((xg->chIdx == DDLT_MAX_PARAMS || !info->iparm[xg->chIdx]) &&
                   (info->iparm[0] & CHSF_LOOP))
                {
                    xg->chIdx = 1;
                }
                if(xg->chIdx < DDLT_MAX_PARAMS && info->iparm[xg->chIdx])
                {
                    xg->chTimer =
                        XG_RandomPercentFloat(info->fparm[xg->chIdx],
                                              info->fparm[0]);
                }
            }
            else if(info->iparm[0] & CHSF_DEACTIVATE_WHEN_DONE)
            {
                XL_ActivateLine(false, info, line, 0, (mobj_t *)xg->activator, XLE_CHAIN);
            }
        }
    }
    if(((info->actType == LTACT_COUNTED_OFF ||
         info->actType == LTACT_FLIP_COUNTED_OFF) && xg->active) ||
       ((info->actType == LTACT_COUNTED_ON ||
         info->actType == LTACT_FLIP_COUNTED_ON) && !xg->active))
    {
        if(info->actTime >= 0 && xg->timer > FLT2TIC(info->actTime))
        {
            LOG_MAP_MSG_XGDEVONLY2("Line %i, timed to go %s",
                    P_ToIndex(line)
                    << (xg->active ? "INACTIVE" : "ACTIVE"));
            XL_ActivateLine(!xg->active, info, line, 0, dummyThing, XLE_AUTO);
        }
    }
    if(info->materialMoveSpeed)
    {
        coord_t current[2]; // The current offset.
        Side* side;
        float spd = info->materialMoveSpeed;
        coord_t offset[2];
        angle_t ang = ((angle_t) (ANGLE_MAX * (info->materialMoveAngle / 360))) >> ANGLETOFINESHIFT;
        offset[VX] = -(FIX2FLT(finecosine[ang]) * spd);
        offset[VY] = FIX2FLT(finesine[ang]) * spd;
        /**
         * Apply to both sides of the line.
         * These are group offsets. All surfaces on a given side are moved
         * using the same texmove speed/angle.
         *
         * @todo Implement per-surface texture movement also which would
         *        be added to each independantly.
         */
        side = (Side *)P_GetPtrp(line, DMU_FRONT);
        if(side)
        {
            P_GetDoublepv(side, DMU_TOP_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_TOP_MATERIAL_OFFSET_XY, current);
            P_GetDoublepv(side, DMU_MIDDLE_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_MIDDLE_MATERIAL_OFFSET_XY, current);
            P_GetDoublepv(side, DMU_BOTTOM_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_BOTTOM_MATERIAL_OFFSET_XY, current);
        }
        side = (Side *)P_GetPtrp(line, DMU_BACK);
        if(side)
        {
            P_GetDoublepv(side, DMU_TOP_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_TOP_MATERIAL_OFFSET_XY, current);
            P_GetDoublepv(side, DMU_MIDDLE_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_MIDDLE_MATERIAL_OFFSET_XY, current);
            P_GetDoublepv(side, DMU_BOTTOM_MATERIAL_OFFSET_XY, current);
            current[VX] += offset[VX];
            current[VY] += offset[VY];
            P_SetDoublepv(side, DMU_BOTTOM_MATERIAL_OFFSET_XY, current);
        }
    }
}
/**
 * During update, definitions are re-read, so the pointers need to be
 * updated. However, this is a bit messy operation, prone to errors.
 * Instead, we just disable XG...
 */
void XL_Update(void)
{
    int i;
    xline_t *xline;
    for(i = 0; i < numlines; ++i)
    {
        xline = P_GetXLine(i);
        if(xline->xg)
        {
            xline->xg = NULL;
            xline->special = 0;
        }
    }
}
#endif
/** @file p_xgsave.cpp  XG data/thinker (de)serialization.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
#include "common.h"
#include "dmu_lib.h"
#include "mapstatereader.h"
#include "mapstatewriter.h"
#include "p_mapsetup.h"
#include "p_saveg.h"
#include "p_xg.h"
void SV_WriteXGLine(Line *li, MapStateWriter *msw)
{
    Writer1 *writer = msw->writer();
    xline_t *xline = P_ToXLine(li);
    Writer_WriteByte(writer, 1);
    /*
     * Remember, savegames are applied on top of an initialized map. No strings are saved,
     * because they are all const strings defined either in the maps's DDXGDATA lump or a
     * DED file. During loading, XL_SetLineType is called with the id in the savegame.
     */
    DENG_ASSERT(xline->xg != 0);
    xgline_t *xg = xline->xg;
    linetype_t *info = &xg->info;
    Writer_WriteInt32(writer, info->id);
    Writer_WriteInt32(writer, info->actCount);
    Writer_WriteByte(writer, xg->active);
    Writer_WriteByte(writer, xg->disabled);
    Writer_WriteInt32(writer, xg->timer);
    Writer_WriteInt32(writer, xg->tickerTimer);
    Writer_WriteInt16(writer, msw->serialIdFor((mobj_t *)xg->activator));
    Writer_WriteInt32(writer, xg->idata);
    Writer_WriteFloat(writer, xg->fdata);
    Writer_WriteInt32(writer, xg->chIdx);
    Writer_WriteFloat(writer, xg->chTimer);
}
void SV_ReadXGLine(Line *li, MapStateReader *msr)
{
    Reader1 *reader = msr->reader();
    xline_t *xline = P_ToXLine(li);
    Reader_ReadByte(reader);
    XL_SetLineType(li, Reader_ReadInt32(reader));
    DENG_ASSERT(xline->xg != 0);
    xgline_t *xg = xline->xg;
    xg->info.actCount = Reader_ReadInt32(reader);
    xg->active      = Reader_ReadByte(reader);
    xg->disabled    = Reader_ReadByte(reader);
    xg->timer       = Reader_ReadInt32(reader);
    xg->tickerTimer = Reader_ReadInt32(reader);
    xg->activator   = INT2PTR(void, Reader_ReadInt16(reader));
    xg->idata       = Reader_ReadInt32(reader);
    xg->fdata       = Reader_ReadFloat(reader);
    xg->chIdx       = Reader_ReadInt32(reader);
    xg->chTimer     = Reader_ReadFloat(reader);
}
/**
 * @param fn  This function must belong to XG sector @a xg.
 */
void SV_WriteXGFunction(xgsector_t * /*xg*/, function_t *fn, Writer1 *writer)
{
    Writer_WriteByte(writer, 1);
    Writer_WriteInt32(writer, fn->flags);
    Writer_WriteInt16(writer, fn->pos);
    Writer_WriteInt16(writer, fn->repeat);
    Writer_WriteInt16(writer, fn->timer);
    Writer_WriteInt16(writer, fn->maxTimer);
    Writer_WriteFloat(writer, fn->value);
    Writer_WriteFloat(writer, fn->oldValue);
}
/**
 * @param fn  This function must belong to XG sector @a xg.
 */
void SV_ReadXGFunction(xgsector_t * /*xg*/, function_t *fn, Reader1 *reader, int /*mapVersion*/)
{
    Reader_ReadByte(reader);
    fn->flags    = Reader_ReadInt32(reader);
    fn->pos      = Reader_ReadInt16(reader);
    fn->repeat   = Reader_ReadInt16(reader);
    fn->timer    = Reader_ReadInt16(reader);
    fn->maxTimer = Reader_ReadInt16(reader);
    fn->value    = Reader_ReadFloat(reader);
    fn->oldValue = Reader_ReadFloat(reader);
}
void SV_WriteXGSector(Sector *sec, Writer1 *writer)
{
    xsector_t *xsec = P_ToXSector(sec);
    xgsector_t *xg     = xsec->xg;
    sectortype_t *info = &xg->info;
    Writer_WriteByte(writer, 1);
    Writer_WriteInt32(writer, info->id);
    Writer_Write(writer, info->count, sizeof(info->count));
    Writer_Write(writer, xg->chainTimer, sizeof(xg->chainTimer));
    Writer_WriteInt32(writer, xg->timer);
    Writer_WriteByte(writer, xg->disabled);
    for(int i = 0; i < 3; ++i)
    {
        SV_WriteXGFunction(xg, &xg->rgb[i], writer);
    }
    for(int i = 0; i < 2; ++i)
    {
        SV_WriteXGFunction(xg, &xg->plane[i], writer);
    }
    SV_WriteXGFunction(xg, &xg->light, writer);
}
void SV_ReadXGSector(Sector *sec, Reader1 *reader, int mapVersion)
{
    xsector_t *xsec = P_ToXSector(sec);
    Reader_ReadByte(reader);
    XS_SetSectorType(sec, Reader_ReadInt32(reader));
    xgsector_t *xg = xsec->xg;
    Reader_Read(reader, xg->info.count, sizeof(xg->info.count));
    Reader_Read(reader, xg->chainTimer, sizeof(xg->chainTimer));
    xg->timer    = Reader_ReadInt32(reader);
    xg->disabled = Reader_ReadByte(reader);
    for(int i = 0; i < 3; ++i)
    {
        SV_ReadXGFunction(xg, &xg->rgb[i], reader, mapVersion);
    }
    for(int i = 0; i < 2; ++i)
    {
        SV_ReadXGFunction(xg, &xg->plane[i], reader, mapVersion);
    }
    SV_ReadXGFunction(xg, &xg->light, reader, mapVersion);
}
void xgplanemover_s::write(MapStateWriter *msw) const
{
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 3); // Version.
    Writer_WriteInt32(writer, P_ToIndex(sector));
    Writer_WriteByte(writer, ceiling);
    Writer_WriteInt32(writer, flags);
    int i = P_ToIndex(origin);
    if(i >= 0 && i < numlines) // Is it a real line?
        i++;
    else // No.
        i = 0;
    Writer_WriteInt32(writer, i); // Zero means there is no origin.
    Writer_WriteInt32(writer, FLT2FIX(destination));
    Writer_WriteInt32(writer, FLT2FIX(speed));
    Writer_WriteInt32(writer, FLT2FIX(crushSpeed));
    Writer_WriteInt32(writer, msw->serialIdFor(setMaterial));
    Writer_WriteInt32(writer, setSectorType);
    Writer_WriteInt32(writer, startSound);
    Writer_WriteInt32(writer, endSound);
    Writer_WriteInt32(writer, moveSound);
    Writer_WriteInt32(writer, minInterval);
    Writer_WriteInt32(writer, maxInterval);
    Writer_WriteInt32(writer, timer);
}
int xgplanemover_s::read(MapStateReader *msr)
{
    Reader1 *reader = msr->reader();
    byte ver = Reader_ReadByte(reader); // Version.
    sector      = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
    ceiling     = Reader_ReadByte(reader);
    flags       = Reader_ReadInt32(reader);
    int lineIndex = Reader_ReadInt32(reader);
    if(lineIndex > 0)
        origin  = (Line *)P_ToPtr(DMU_LINE, lineIndex - 1);
    destination = FIX2FLT(Reader_ReadInt32(reader));
    speed       = FIX2FLT(Reader_ReadInt32(reader));
    crushSpeed  = FIX2FLT(Reader_ReadInt32(reader));
    if(ver >= 3)
    {
        setMaterial = msr->material(Reader_ReadInt32(reader), 0);
    }
    else
    {
        de::Uri uri("Flats:", CentralLumpIndex()[Reader_ReadInt32(reader)].name().fileNameWithoutExtension());
        setMaterial = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(reinterpret_cast<uri_s *>(&uri)));
    }
    setSectorType = Reader_ReadInt32(reader);
    startSound    = Reader_ReadInt32(reader);
    endSound      = Reader_ReadInt32(reader);
    moveSound     = Reader_ReadInt32(reader);
    minInterval   = Reader_ReadInt32(reader);
    maxInterval   = Reader_ReadInt32(reader);
    timer         = Reader_ReadInt32(reader);
    thinker.function = (thinkfunc_t) XS_PlaneMover;
    return true; // Add this thinker.
}
#endif
/** @file p_xgsec.cpp  Extended generalized sector types.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2014 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
#include <math.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#if __JDOOM__
#  include "jdoom.h"
#elif __JDOOM64__
#  include "jdoom64.h"
#elif __JHERETIC__
#  include "jheretic.h"
#endif
#include "dmu_lib.h"
#include "p_mapsetup.h"
#include "p_xgline.h"
#include "p_xgsec.h"
#include "g_common.h"
#include "p_map.h"
#include "mobj.h"
#include "p_actor.h"
#include "p_mapspec.h"
#include "p_sound.h"
#include "p_terraintype.h"
#include "p_tick.h"
#define MAX_VALS        128
#define SIGN(x)         ((x)>0? 1 : (x)<0? -1 : 0)
#define ISFUNC(fn)      (fn->func && fn->func[fn->pos])
#define UPDFUNC(fn)     ((ISFUNC(fn) || fn->link))
#define SPREFTYPESTR(reftype) (reftype == SPREF_NONE? "NONE" \
        : reftype == SPREF_MY_FLOOR? "MY FLOOR" \
        : reftype == SPREF_MY_CEILING? "MY CEILING" \
        : reftype == SPREF_ORIGINAL_FLOOR? "ORIGINAL FLOOR" \
        : reftype == SPREF_ORIGINAL_CEILING? "ORIGINAL CEILING" \
        : reftype == SPREF_CURRENT_FLOOR? "CURRENT FLOOR" \
        : reftype == SPREF_CURRENT_CEILING? "CURRENT CEILING" \
        : reftype == SPREF_HIGHEST_FLOOR? "HIGHEST FLOOR" \
        : reftype == SPREF_HIGHEST_CEILING? "HIGHEST CEILING" \
        : reftype == SPREF_LOWEST_FLOOR? "LOWEST FLOOR" \
        : reftype == SPREF_LOWEST_CEILING? "LOWEST CEILING" \
        : reftype == SPREF_NEXT_HIGHEST_FLOOR? "NEXT HIGHEST FLOOR" \
        : reftype == SPREF_NEXT_HIGHEST_CEILING? "NEXT HIGHEST CEILING" \
        : reftype == SPREF_NEXT_LOWEST_FLOOR? "NEXT LOWEST FLOOR" \
        : reftype == SPREF_NEXT_LOWEST_CEILING? "NEXT LOWEST CEILING" \
        : reftype == SPREF_MIN_BOTTOM_MATERIAL? "MIN BOTTOM MATERIAL" \
        : reftype == SPREF_MIN_MID_MATERIAL? "MIN MIDDLE MATERIAL" \
        : reftype == SPREF_MIN_TOP_MATERIAL? "MIN TOP MATERIAL" \
        : reftype == SPREF_MAX_BOTTOM_MATERIAL? "MAX BOTTOM MATERIAL" \
        : reftype == SPREF_MAX_MID_MATERIAL? "MAX MIDDLE MATERIAL" \
        : reftype == SPREF_MAX_TOP_MATERIAL? "MAX TOP MATERIAL" \
        : reftype == SPREF_SECTOR_TAGGED_FLOOR? "SECTOR TAGGED FLOOR" \
        : reftype == SPREF_LINE_TAGGED_FLOOR? "LINE TAGGED FLOOR" \
        : reftype == SPREF_TAGGED_FLOOR? "TAGGED FLOOR" \
        : reftype == SPREF_ACT_TAGGED_FLOOR? "ACT TAGGED FLOOR" \
        : reftype == SPREF_INDEX_FLOOR? "INDEXED FLOOR" \
        : reftype == SPREF_SECTOR_TAGGED_CEILING? "SECTOR TAGGED CEILING" \
        : reftype == SPREF_LINE_TAGGED_CEILING? "LINE TAGGED CEILING" \
        : reftype == SPREF_TAGGED_CEILING? "TAGGED CEILING" \
        : reftype == SPREF_ACT_TAGGED_CEILING? "ACT TAGGED CEILING" \
        : reftype == SPREF_INDEX_CEILING? "INDEXED CEILING" \
        : reftype == SPREF_BACK_FLOOR? "BACK FLOOR" \
        : reftype == SPREF_BACK_CEILING? "BACK CEILING" \
        : reftype == SPREF_SPECIAL? "SPECIAL" \
        : reftype == SPREF_LINE_ACT_TAGGED_FLOOR? "LINE ACT TAGGED FLOOR" \
        : reftype == SPREF_LINE_ACT_TAGGED_CEILING? "LINE ACT TAGGED CEILING" : "???")
#define TO_DMU_COLOR(x) (x == 0? DMU_COLOR_RED \
        : x == 1? DMU_COLOR_GREEN \
        : DMU_COLOR_BLUE)
#define TO_DMU_CEILING_COLOR(x) (x == 0? DMU_CEILING_COLOR_RED \
        : x == 1? DMU_CEILING_COLOR_GREEN \
        : DMU_CEILING_COLOR_BLUE)
#define TO_DMU_FLOOR_COLOR(x) (x == 0? DMU_FLOOR_COLOR_RED \
        : x == 1? DMU_FLOOR_COLOR_GREEN \
        : DMU_FLOOR_COLOR_BLUE)
void XS_DoChain(Sector *sec, int ch, int activating, void *actThing);
/**
 * Lookup a sectortype_t with the given @a id and if found - copy it into @a outBuffer.
 *
 * @param id         Unique identifier of the sector type.
 * @param outBuffer  If found the sector type info is written here.
 *
 * @return  @c true if a sectortype_t was found.
 */
bool XS_GetType(int id, sectortype_t &outBuffer)
{
    if(sectortype_t *found = XG_GetLumpSector(id))
    {
        std::memcpy(&outBuffer, found, sizeof(outBuffer));
        return true;
    }
    return Def_Get(DD_DEF_SECTOR_TYPE, de::String::number(id).toUtf8().constData(), &outBuffer);
}
void XF_Init(Sector *sec, function_t *fn, char *func, int min, int max,
             float scale, float offset)
{
    xsector_t          *xsec = P_ToXSector(sec);
    memset(fn, 0, sizeof(*fn));
    if(!func)
        return;
    if(func[0] == '=')
    {
        switch(tolower(func[1]))
        {
        case 'r':
            fn->link = &xsec->xg->rgb[0];
            break;
        case 'g':
            fn->link = &xsec->xg->rgb[1];
            break;
        case 'b':
            fn->link = &xsec->xg->rgb[2];
            break;
        case 'f':
            fn->link = &xsec->xg->plane[XGSP_FLOOR];
            break;
        case 'c':
            fn->link = &xsec->xg->plane[XGSP_CEILING];
            break;
        case 'l':
            fn->link = &xsec->xg->light;
            break;
        default:
            Con_Error("XF_Init: Bad linked func (%s).\n", func);
        }
        return;
    }
    if(func[0] == '+')
    {
        /**
         * @note The original value ranges must be maintained due to the cross linking
         * between sector function types i.e:
         * - RGB = 0 > 254
         * - light = 0 > 254
         * - planeheight = -32768 > 32768
         */
        switch(func[1])
        {
        case 'r':
            offset += 255.f * xsec->origRGB[0];
            break;
        case 'g':
            offset += 255.f * xsec->origRGB[1];
            break;
        case 'b':
            offset += 255.f * xsec->origRGB[2];
            break;
        case 'l':
            offset += 255.0f * xsec->origLight;
            break;
        case 'f':
            offset += xsec->SP_floororigheight;
            break;
        case 'c':
            offset += xsec->SP_ceilorigheight;
            break;
        default:
            Con_Error("XF_Init: Bad preset offset (%s).\n", func);
        }
        fn->func = func + 2;
    }
    else
    {
        fn->func = func;
    }
    fn->timer = -1; // The first step musn't skip the first value.
    fn->maxTimer = XG_RandomInt(min, max);
    fn->minInterval = min;
    fn->maxInterval = max;
    fn->scale = scale;
    fn->offset = offset;
    fn->oldValue = -scale + offset;
}
int C_DECL XLTrav_LineAngle(Line* line, dd_bool dummy, void* context,
    void* context2, mobj_t* activator)
{
    DENG_UNUSED(dummy);
    DENG_UNUSED(activator);
    Sector* sec = (Sector *) context;
    coord_t d1[2];
    if(P_GetPtrp(line, DMU_FRONT_SECTOR) != sec &&
       P_GetPtrp(line, DMU_BACK_SECTOR)  != sec)
        return true; // Wrong sector, keep looking.
    P_GetDoublepv(line, DMU_DXY, d1);
    *(angle_t*) context2 = M_PointXYToAngle2(0, 0, d1[0], d1[1]);
    return false; // Stop looking after first hit.
}
int findXSThinker(thinker_t *th, void *context)
{
    xsthinker_t *xs = (xsthinker_t *) th;
    DENG2_ASSERT(xs);
    if(xs->sector == (Sector *) context)
    {
        return true; // Stop iteration, we've found it.
    }
    return false; // Continue iteration.
}
int destroyXSThinker(thinker_t *th, void *context)
{
    xsthinker_t *xs = (xsthinker_t *) th;
    DENG2_ASSERT(xs);
    if(xs->sector == (Sector *) context)
    {
        Thinker_Remove(&xs->thinker);
        return true; // Stop iteration, we're done.
    }
    return false; // Continue iteration.
}
static void XS_UpdateLight(Sector *sec)
{
    int         i;
    float       c, lightlevel;
    xgsector_t *xg;
    function_t *fn;
    xg = P_ToXSector(sec)->xg;
    fn = &xg->light;
    if (UPDFUNC(fn))
    {
        lightlevel = MINMAX_OF(0, fn->value / 255.f, 1);
        P_SetFloatp(sec, DMU_LIGHT_LEVEL, lightlevel);
    }
    for (i = 0; i < 3; ++i)
    {
        fn = &xg->rgb[i];
        if (UPDFUNC(fn))
        {
            c = MINMAX_OF(0, fn->value / 255.f, 1);
            P_SetFloatp(sec, TO_DMU_COLOR(i), c);
        }
    }
}
void XS_SetSectorType(Sector *sec, int special)
{
    LOG_AS("XS_SetSectorType");
    xsector_t *xsec = P_ToXSector(sec);
    if(!xsec) return;
    sectortype_t secType;
    if(XS_GetType(special, secType))
    {
        LOG_MAP_MSG_XGDEVONLY2("Sector %i, type %i", P_ToIndex(sec) << special);
        xsec->special = special;
        if(!xsec->xg)
        {
            xsec->xg = (xgsector_t *) Z_Malloc(sizeof(xgsector_t), PU_MAP, 0);
        }
        de::zapPtr(xsec->xg);
        std::memcpy(&xsec->xg->info, &secType, sizeof(secType));
        xgsector_t *xg     = xsec->xg;
        sectortype_t *info = &xsec->xg->info;
        xg->timer = XG_RandomInt(FLT2TIC(xg->info.soundInterval[0]),
                                 FLT2TIC(xg->info.soundInterval[1]));
        XF_Init(sec, &xg->light, info->lightFunc, info->lightInterval[0],
                info->lightInterval[1], 255, 0);
        for(int i = 0; i < 3; ++i)
        {
            XF_Init(sec, &xg->rgb[i], info->colFunc[i],
                    info->colInterval[i][0], info->colInterval[i][1], 255, 0);
        }
        XF_Init(sec, &xg->plane[XGSP_FLOOR], info->floorFunc,
                info->floorInterval[0], info->floorInterval[1],
                info->floorMul, info->floorOff);
        XF_Init(sec, &xg->plane[XGSP_CEILING], info->ceilFunc,
                info->ceilInterval[0], info->ceilInterval[1],
                info->ceilMul, info->ceilOff);
        if((info->flags & STF_ACT_TAG_MATERIALMOVE) ||
           (info->flags & STF_ACT_TAG_WIND))
        {
            angle_t angle = 0;
            XL_TraverseLines(0, (xgDataLumps? LREF_TAGGED -1: LREF_TAGGED),
                             info->actTag, sec, &angle,
                             NULL, XLTrav_LineAngle);
            if(info->flags & STF_ACT_TAG_MATERIALMOVE)
            {
                info->materialMoveAngle[0] = info->materialMoveAngle[1] =
                    angle / (float) ANGLE_MAX *360;
            }
            if(info->flags & STF_ACT_TAG_WIND)
            {
                info->windAngle = angle / (float) ANGLE_MAX *360;
            }
        }
        if(!Thinker_Iterate((thinkfunc_t) XS_Thinker, findXSThinker, sec))
        {
            ThinkerT<xsthinker_t> xs(Thinker::AllocateMemoryZone);
            xs.function = XS_Thinker;
            xs->sector  = sec;
            Thinker_Add(xs.Thinker::take());
        }
    }
    else
    {
        LOG_MAP_MSG_XGDEVONLY2("Sector %i, NORMAL TYPE %i", P_ToIndex(sec) << special);
        Thinker_Iterate((thinkfunc_t) XS_Thinker, destroyXSThinker, sec);
        Z_Free(xsec->xg); xsec->xg = nullptr;
        xsec->special = special;
    }
}
void XS_Init()
{
    /*  // Clients rely on the server, they don't do XG themselves.
    if(IS_CLIENT) return; */
    if(numsectors <= 0) return;
    for(int i = 0; i < numsectors; ++i)
    {
        Sector *sec     = (Sector *) P_ToPtr(DMU_SECTOR, i);
        xsector_t *xsec = P_ToXSector(sec);
        P_GetFloatpv(sec, DMU_COLOR, xsec->origRGB);
        xsec->SP_floororigheight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
        xsec->SP_ceilorigheight  = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
        xsec->origLight = P_GetFloatp(sec, DMU_LIGHT_LEVEL);
        XS_SetSectorType(sec, xsec->special);
    }
    P_IterateThinkers(XS_Thinker, [](thinker_t *th) {
        XS_Thinker(th);
        return de::LoopContinue;
    });
}
void XS_SectorSound(Sector *sec, int soundId)
{
    LOG_AS("XS_SectorSound");
    if(!sec || !soundId) return;
    LOG_MAP_MSG_XGDEVONLY2("Play Sound ID (%i) in Sector ID (%i)", soundId << P_ToIndex(sec));
    S_SectorSound(sec, soundId);
}
void XS_PlaneSound(Plane *pln, int soundId)
{
    LOG_AS("XS_PlaneSound");
    if(!pln || !soundId) return;
    LOG_MAP_MSG_XGDEVONLY2("Play Sound ID (%i) in Sector ID (%i)",
            soundId << P_ToIndex(P_GetPtrp(pln, DMU_SECTOR)));
    S_PlaneSound(pln, soundId);
}
void XS_MoverStopped(xgplanemover_t *mover, dd_bool done)
{
    DENG2_ASSERT(mover);
    LOG_AS("XS_MoverStopped");
    xline_t *origin = P_ToXLine(mover->origin);
    LOG_MAP_MSG_XGDEVONLY2("Sector %i (done=%i, origin line=%i)",
           P_ToIndex(mover->sector) << done << P_ToIndex(mover->origin));
    if(done)
    {
        if((mover->flags & PMF_ACTIVATE_WHEN_DONE) && mover->origin)
        {
            XL_ActivateLine(true, &origin->xg->info, mover->origin, 0,
                            XG_DummyThing(), XLE_AUTO);
        }
        if((mover->flags & PMF_DEACTIVATE_WHEN_DONE) && mover->origin)
        {
            XL_ActivateLine(false, &origin->xg->info, mover->origin, 0,
                            XG_DummyThing(), XLE_AUTO);
        }
        Thinker_Remove((thinker_t *) mover);
    }
    else
    {
        if((mover->flags & PMF_ACTIVATE_ON_ABORT) && mover->origin)
        {
            XL_ActivateLine(true, &origin->xg->info, mover->origin, 0,
                            XG_DummyThing(), XLE_AUTO);
        }
        if((mover->flags & PMF_DEACTIVATE_ON_ABORT) && mover->origin)
        {
            XL_ActivateLine(false, &origin->xg->info, mover->origin, 0,
                            XG_DummyThing(), XLE_AUTO);
        }
        if(mover->flags & (PMF_ACTIVATE_ON_ABORT | PMF_DEACTIVATE_ON_ABORT))
        {
            Thinker_Remove((thinker_t *) mover);
        }
    }
}
/**
 * A thinker function for plane movers.
 */
void XS_PlaneMover(xgplanemover_t *mover)
{
    DENG2_ASSERT(mover && mover->sector);
    coord_t ceil    = P_GetDoublep(mover->sector, DMU_CEILING_HEIGHT);
    coord_t floor   = P_GetDoublep(mover->sector, DMU_FLOOR_HEIGHT);
    xsector_t *xsec = P_ToXSector(mover->sector);
    dd_bool docrush = (mover->flags & PMF_CRUSH) != 0;
    dd_bool follows = (mover->flags & PMF_OTHER_FOLLOWS) != 0;
    dd_bool setorig = (mover->flags & PMF_SET_ORIGINAL) != 0;
    int res, res2, dir;
    if(mover->timer-- <= 0)
    {
        if(mover->flags & PMF_WAIT)
        {
            mover->flags &= ~PMF_WAIT;
            XS_PlaneSound((Plane *) P_GetPtrp(mover->sector, mover->ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                          mover->startSound);
        }
        mover->timer = XG_RandomInt(mover->minInterval, mover->maxInterval);
        XS_PlaneSound((Plane *) P_GetPtrp(mover->sector, mover->ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                      mover->moveSound);
    }
    if(mover->flags & PMF_WAIT) return;
    if((mover->destination - (mover->ceiling ? ceil : floor)) > 0)
        dir = 1; // Moving up.
    else
        dir = -1; // Moving down.
    res = T_MovePlane(mover->sector, mover->speed, mover->destination,
                      docrush, mover->ceiling, dir);
    if(setorig)
    {
        xsec->planes[mover->ceiling? PLN_CEILING:PLN_FLOOR].origHeight =
            P_GetDoublep(mover->sector,
                        mover->ceiling? DMU_CEILING_HEIGHT:DMU_FLOOR_HEIGHT);
    }
    if(follows)
    {
        coord_t off = (mover->ceiling? floor - ceil : ceil - floor);
        res2 = T_MovePlane(mover->sector, mover->speed,
                           mover->destination + off, docrush,
                           !mover->ceiling, dir);
        if(setorig)
        {
            xsec->planes[(!mover->ceiling)? PLN_CEILING:PLN_FLOOR].origHeight =
                P_GetDoublep(mover->sector,
                            (!mover->ceiling)? DMU_CEILING_HEIGHT:DMU_FLOOR_HEIGHT);
        }
        if(res2 == crushed)
            res = crushed;
    }
    if(res == pastdest)
    {
        XS_MoverStopped(mover, true);
        if(mover->setMaterial)
        {
            XS_ChangePlaneMaterial(*mover->sector, mover->ceiling, *mover->setMaterial);
        }
        if(mover->setSectorType >= 0)
        {
            XS_SetSectorType(mover->sector, mover->setSectorType);
        }
        XS_PlaneSound((Plane *) P_GetPtrp(mover->sector, mover->ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                      mover->endSound);
    }
    else if(res == crushed)
    {
        if(mover->flags & PMF_CRUSH)
        {
            mover->speed = mover->crushSpeed;
        }
        else
        {
            if((!mover->ceiling || follows) &&
               !FEQUAL(P_GetDoublep(mover->sector, DMU_FLOOR_HEIGHT), floor))
            {
                T_MovePlane(mover->sector, mover->speed, floor, docrush, false, -dir);
            }
            if((mover->ceiling || follows) &&
               !FEQUAL(P_GetDoublep(mover->sector, DMU_CEILING_HEIGHT), ceil))
            {
                T_MovePlane(mover->sector, mover->speed, ceil, docrush, true, -dir);
            }
            XS_MoverStopped(mover, false);
        }
    }
}
typedef struct {
    Sector*             sec;
    dd_bool             ceiling;
} stopplanemoverparams_t;
static int stopPlaneMover(thinker_t* th, void* context)
{
    stopplanemoverparams_t* params = (stopplanemoverparams_t*) context;
    xgplanemover_t*     mover = (xgplanemover_t *) th;
    if(mover->sector == params->sec &&
       mover->ceiling == params->ceiling)
    {
        XS_MoverStopped(mover, false);
        Thinker_Remove(th); // Remove it.
    }
    return false; // Continue iteration.
}
/**
 * Returns a new thinker for handling the specified plane. Removes any
 * existing thinkers associated with the plane.
 */
xgplanemover_t *XS_GetPlaneMover(Sector *sec, dd_bool ceiling)
{
    stopplanemoverparams_t params;
    params.sec = sec;
    params.ceiling = ceiling;
    Thinker_Iterate((thinkfunc_t) XS_PlaneMover, stopPlaneMover, &params);
    ThinkerT<xgplanemover_t> mover(Thinker::AllocateMemoryZone);
    mover.function = (thinkfunc_t) XS_PlaneMover;
    xgplanemover_t *th = mover.take();
    th->sector  = sec;
    th->ceiling = ceiling;
    Thinker_Add(&th->thinker);
    return th;
}
void XS_ChangePlaneMaterial(Sector &sector, bool ceiling, world_Material &newMaterial)
{
    LOG_AS("XS_ChangePlaneMaterial");
    LOG_MAP_MSG_XGDEVONLY2("Sector %i, %s, texture %i",
           P_ToIndex(&sector) << (ceiling ? "ceiling" : "floor") << P_ToIndex(&newMaterial));
    P_SetPtrp(&sector, ceiling ? DMU_CEILING_MATERIAL : DMU_FLOOR_MATERIAL, &newMaterial);
}
void XS_ChangePlaneColor(Sector &sector, bool ceiling, de::Vector3f const &newColor, bool isDelta)
{
    LOG_AS("XS_ChangePlaneColor");
    LOG_MAP_MSG_XGDEVONLY2("Sector %i, %s, tintColor:%s",
           P_ToIndex(&sector) << (ceiling ? "ceiling" : "floor") << newColor.asText());
    float rgb[3];
    if (isDelta)
    {
        P_GetFloatpv(&sector, ceiling ? DMU_CEILING_COLOR : DMU_FLOOR_COLOR, rgb);
        for (int i = 0; i < 3; ++i) { rgb[i] += newColor[i]; }
    }
    else
    {
        newColor.decompose(rgb);
    }
    P_SetFloatpv(&sector, ceiling? DMU_CEILING_COLOR : DMU_FLOOR_COLOR, rgb); // will clamp
}
uint FindMaxOf(int *list, uint num)
{
    uint            i, idx = 0;
    int             max = list[0];
    for(i = 1; i < num; ++i)
    {
        if(list[i] > max)
        {
            max = list[i];
            idx = i;
        }
    }
    return idx;
}
uint FindMinOf(int *list, uint num)
{
    uint        i, idx = 0;
    int         min = list[0];
    for(i = 1; i < num; ++i)
    {
        if(list[i] < min)
        {
            min = list[i];
            idx = i;
        }
    }
    return idx;
}
int FindNextOf(int *list, int num, int h)
{
    int         i, min = 0, idx = -1;
    for(i = 0; i < num; ++i)
    {
        if(list[i] <= h)
            continue;
        if(idx < 0 || list[i] < min)
        {
            idx = i;
            min = list[i];
        }
    }
    return idx;
}
int FindPrevOf(int *list, int num, int h)
{
    int         i, max = 0, idx = -1;
    for(i = 0; i < num; ++i)
    {
        if(list[i] >= h)
            continue;
        if(idx < 0 || list[i] > max)
        {
            idx = i;
            max = list[i];
        }
    }
    return idx;
}
/**
 * Really an XL_* function!
 *
 * @param part          1=mid, 2=top, 3=bottom.
 *
 * @return              @c MAXINT if not height n/a.
 */
int XS_TextureHeight(Line* line, int part)
{
    Side* side;
    int snum = 0;
    int minfloor = 0, maxfloor = 0, maxceil = 0;
    Sector* front = (Sector *) P_GetPtrp(line, DMU_FRONT_SECTOR);
    Sector* back  = (Sector *) P_GetPtrp(line, DMU_BACK_SECTOR);
    dd_bool twosided = front && back;
    world_Material* mat;
    if(part != LWS_MID && !twosided)
        return DDMAXINT;
    if(twosided)
    {
        int ffloor = P_GetIntp(front, DMU_FLOOR_HEIGHT);
        int fceil  = P_GetIntp(front, DMU_CEILING_HEIGHT);
        int bfloor = P_GetIntp(back,  DMU_FLOOR_HEIGHT);
        int bceil  = P_GetIntp(back,  DMU_CEILING_HEIGHT);
        minfloor = ffloor;
        maxfloor = bfloor;
        if(part == LWS_LOWER)
            snum = 0;
        if(bfloor < minfloor)
        {
            minfloor = bfloor;
            maxfloor = ffloor;
            if(part == LWS_LOWER)
                snum = 1;
        }
        maxceil = fceil;
        if(part == LWS_UPPER)
            snum = 0;
        if(bceil > maxceil)
        {
            maxceil = bceil;
            if(part == LWS_UPPER)
                snum = 1;
        }
    }
    else
    {
        if(P_GetPtrp(line, DMU_FRONT))
            snum = 0;
        else
            snum = 1;
    }
    if(snum == 0)
        side = (Side *) P_GetPtrp(line, DMU_FRONT);
    else
        side = (Side *) P_GetPtrp(line, DMU_BACK);
    switch(part)
    {
    case LWS_UPPER:
        if((mat = (world_Material *) P_GetPtrp(side, DMU_TOP_MATERIAL)))
            return maxceil - P_GetIntp(mat, DMU_HEIGHT);
        break;
    case LWS_MID:
        if((mat = (world_Material *) P_GetPtrp(side, DMU_MIDDLE_MATERIAL)))
            return maxfloor + P_GetIntp(mat, DMU_HEIGHT);
        break;
    case LWS_LOWER:
        if((mat = (world_Material *) P_GetPtrp(side, DMU_BOTTOM_MATERIAL)))
            return minfloor + P_GetIntp(mat, DMU_HEIGHT);
        break;
    default:
        Con_Error("XS_TextureHeight: Invalid wall section %d.", part);
    }
    return DDMAXINT;
}
/**
 * Returns a pointer to the first sector with the tag.
 *
 * NOTE: We cannot use the tagged sector lists here as this can be called
 * during an iteration at a higher level. Doing so would change the position
 * of the rover which would affect the other iteration.
 *
 * NOTE2: Re-above, obviously that is bad design and should be addressed.
 */
Sector *XS_FindTagged(int tag)
{
    LOG_AS("XS_FindTagged");
    int k;
    int foundcount = 0;
    int retsectorid = 0;
    Sector *sec, *retsector;
    retsector = NULL;
    for(k = 0; k < numsectors; ++k)
    {
        sec = (Sector *) P_ToPtr(DMU_SECTOR, k);
        if(P_ToXSector(sec)->tag == tag)
        {
            if(xgDev)
            {
                if(foundcount == 0)
                {
                    retsector = sec;
                    retsectorid = k;
                }
            }
            else
                return sec;
            foundcount++;
        }
    }
    if(xgDev)
    {
        if(foundcount > 1)
        {
            LOG_MAP_MSG_XGDEVONLY2("More than one sector exists with this tag (%i)!", tag);
            LOG_MAP_MSG_XGDEVONLY2("The sector with the lowest ID (%i) will be used", retsectorid);
        }
        if(retsector)
            return retsector;
    }
    return NULL;
}
/**
 * Returns a pointer to the first sector with the specified act tag.
 */
Sector *XS_FindActTagged(int tag)
{
    LOG_AS("XS_FindActTagged");
    int k;
    int foundcount = 0;
    int retsectorid = 0;
    Sector *sec, *retsector;
    xsector_t *xsec;
    retsector = NULL;
    for(k = 0; k < numsectors; ++k)
    {
        sec = (Sector *) P_ToPtr(DMU_SECTOR, k);
        xsec = P_ToXSector(sec);
        if(xsec->xg)
        {
            if(xsec->xg->info.actTag == tag)
            {
                if(xgDev)
                {
                    if(foundcount == 0)
                    {
                        retsector = sec;
                        retsectorid = k;
                    }
                }
                else
                {
                    return sec;
                }
                foundcount++;
            }
        }
    }
    if(xgDev)
    {
        if(foundcount > 1)
        {
            LOG_MAP_MSG_XGDEVONLY2("More than one sector exists with this ACT tag (%i)!", tag);
            LOG_MAP_MSG_XGDEVONLY2("The sector with the lowest ID (%i) will be used", retsectorid);
        }
        if(retsector)
            return retsector;
    }
    return NULL;
}
#define FSETHF_MIN          0x1 // Get min. If not set, get max.
typedef struct findsectorextremaltextureheightparams_s {
    Sector* baseSec;
    byte flags;
    int part;
    coord_t val;
} findsectorextremalmaterialheightparams_t;
int findSectorExtremalMaterialHeight(void* ptr, void* context)
{
    Line* li = (Line*) ptr;
    findsectorextremalmaterialheightparams_t* params =
        (findsectorextremalmaterialheightparams_t*) context;
    coord_t height;
    height = XS_TextureHeight(li, params->part);
    if(params->flags & FSETHF_MIN)
    {
        if(height < params->val)
            params->val = height;
    }
    else
    {
        if(height > params->val)
            params->val = height;
    }
    return false; // Continue iteration.
}
dd_bool XS_GetPlane(Line* actline, Sector* sector, int ref, int* refdata,
    coord_t* height, world_Material** mat, Sector** planeSector)
{
    LOG_AS("XS_GetPlane");
    world_Material *otherMat = NULL;
    coord_t otherHeight = 0;
    Sector* otherSec = NULL, *iter = NULL;
    xline_t* xline = NULL;
    char buff[50];
    if(refdata)
        sprintf(buff, " : %i", *refdata);
    if(xgDev)
    {
        LOG_MAP_MSG_XGDEVONLY2("Line %i, sector %i, ref (%s(%i)%s)",
               P_ToIndex(actline) << P_ToIndex(sector)
               << SPREFTYPESTR(ref) << ref << (refdata? buff : ""));
    }
    if(ref == SPREF_NONE || ref == SPREF_SPECIAL)
    {
        return false;
    }
    if(height)
        *height = P_GetDoublep(sector, DMU_FLOOR_HEIGHT);
    if(mat)
        *mat = (world_Material*) P_GetPtrp(sector, DMU_FLOOR_MATERIAL);
    if(planeSector)
        *planeSector = sector;
    iter = NULL;
    switch(ref)
    {
    case SPREF_SECTOR_TAGGED_FLOOR:
    case SPREF_SECTOR_TAGGED_CEILING:
        iter = XS_FindTagged(P_ToXSector(sector)->tag);
        if(!iter)
            return false;
        break;
    case SPREF_LINE_TAGGED_FLOOR:
    case SPREF_LINE_TAGGED_CEILING:
        if(!actline)
            return false;
        iter = XS_FindTagged(P_ToXLine(actline)->tag);
        if(!iter)
            return false;
        break;
    case SPREF_TAGGED_FLOOR:
    case SPREF_TAGGED_CEILING:
        if(!refdata)
        {
            LOG_MAP_MSG_XGDEVONLY2("%s IS NOT VALID FOR THIS CLASS PARAMETER!", SPREFTYPESTR(ref));
            return false;
        }
        iter = XS_FindTagged(*refdata);
        if(!iter)
            return false;
        break;
    case SPREF_LINE_ACT_TAGGED_FLOOR:
    case SPREF_LINE_ACT_TAGGED_CEILING:
        xline = P_ToXLine(actline);
        if(!xline)
            return false;
        if(!xline->xg)
        {
            LOG_MAP_MSG_XGDEVONLY("ACT LINE IS NOT AN XG LINE!");
            return false;
        }
        if(!xline->xg->info.actTag)
        {
            LOG_MAP_MSG_XGDEVONLY("ACT LINE DOES NOT HAVE AN ACT TAG!");
            return false;
        }
        iter = XS_FindActTagged(xline->xg->info.actTag);
        if(!iter)
            return false;
        break;
    case SPREF_ACT_TAGGED_FLOOR:
    case SPREF_ACT_TAGGED_CEILING:
        if(!refdata)
        {
            LOG_MAP_MSG_XGDEVONLY2("%s IS NOT VALID FOR THIS CLASS PARAMETER!", SPREFTYPESTR(ref));
            return false;
        }
        iter = XS_FindActTagged(*refdata);
        if(!iter)
            return false;
        break;
    case SPREF_INDEX_FLOOR:
    case SPREF_INDEX_CEILING:
        if(!refdata || *refdata >= numsectors)
            return false;
        iter = (Sector *) P_ToPtr(DMU_SECTOR, *refdata);
        break;
    default:
        break;
    }
    if(iter)
    {
        if(planeSector)
            *planeSector = iter;
        if((ref >= SPREF_SECTOR_TAGGED_FLOOR && ref <= SPREF_INDEX_FLOOR) ||
            ref == SPREF_LINE_ACT_TAGGED_FLOOR)
        {
            if(height)
                *height = P_GetDoublep(iter, DMU_FLOOR_HEIGHT);
            if(mat)
                *mat = (world_Material*) P_GetPtrp(iter, DMU_FLOOR_MATERIAL);
        }
        else
        {
            if(height)
                *height = P_GetDoublep(iter, DMU_CEILING_HEIGHT);
            if(mat)
                *mat = (world_Material*) P_GetPtrp(iter, DMU_CEILING_MATERIAL);
        }
        return true;
    }
    if(ref == SPREF_MY_FLOOR)
    {
        Sector* frontsector;
        if(!actline)
            return false;
        frontsector = (Sector *) P_GetPtrp(actline, DMU_FRONT_SECTOR);
        if(!frontsector)
            return false;
        if(height)
            *height = P_GetDoublep(frontsector, DMU_FLOOR_HEIGHT);
        if(mat)
            *mat = (world_Material*) P_GetPtrp(frontsector, DMU_FLOOR_MATERIAL);
        if(planeSector)
            *planeSector = frontsector;
        return true;
    }
    if(ref == SPREF_BACK_FLOOR)
    {
        Sector* backsector;
        if(!actline)
            return false;
        backsector = (Sector *) P_GetPtrp(actline, DMU_BACK_SECTOR);
        if(!backsector)
            return false;
        if(height)
            *height = P_GetDoublep(backsector, DMU_FLOOR_HEIGHT);
        if(mat)
            *mat = (world_Material*) P_GetPtrp(backsector, DMU_FLOOR_MATERIAL);
        if(planeSector)
            *planeSector = backsector;
        return true;
    }
    if(ref == SPREF_MY_CEILING)
    {
        Sector* frontsector;
        if(!actline)
            return false;
        frontsector = (Sector *) P_GetPtrp(actline, DMU_FRONT_SECTOR);
        if(!frontsector)
            return false;
        if(height)
            *height = P_GetDoublep(frontsector, DMU_CEILING_HEIGHT);
        if(mat)
            *mat = (world_Material *) P_GetPtrp(frontsector, DMU_CEILING_MATERIAL);
        if(planeSector)
            *planeSector = frontsector;
        return true;
    }
    if(ref == SPREF_BACK_CEILING)
    {
        Sector* backsector;
        if(!actline)
            return false;
        backsector = (Sector *) P_GetPtrp(actline, DMU_BACK_SECTOR);
        if(!backsector)
            return false;
        if(height)
            *height = P_GetDoublep(backsector, DMU_CEILING_HEIGHT);
        if(mat)
            *mat = (world_Material *) P_GetPtrp(backsector, DMU_CEILING_MATERIAL);
        if(planeSector)
            *planeSector = backsector;
        return true;
    }
    if(ref == SPREF_ORIGINAL_FLOOR)
    {
        if(height)
            *height = P_ToXSector(sector)->SP_floororigheight;
        if(mat)
            *mat = (world_Material *) P_GetPtrp(sector, DMU_FLOOR_MATERIAL);
        return true;
    }
    if(ref == SPREF_ORIGINAL_CEILING)
    {
        if(height)
            *height = P_ToXSector(sector)->SP_ceilorigheight;
        if(mat)
            *mat = (world_Material *) P_GetPtrp(sector, DMU_CEILING_MATERIAL);
        return true;
    }
    if(ref == SPREF_CURRENT_FLOOR)
    {
        if(height)
            *height = P_GetDoublep(sector, DMU_FLOOR_HEIGHT);
        if(mat)
            *mat = (world_Material *) P_GetPtrp(sector, DMU_FLOOR_MATERIAL);
        return true;
    }
    if(ref == SPREF_CURRENT_CEILING)
    {
        if(height)
            *height = P_GetDoublep(sector, DMU_CEILING_HEIGHT);
        if(mat)
            *mat = (world_Material *) P_GetPtrp(sector, DMU_CEILING_MATERIAL);
        return true;
    }
    if(ref >= SPREF_MIN_BOTTOM_MATERIAL && ref <= SPREF_MAX_TOP_MATERIAL)
    {
        int part;
        dd_bool findMin;
        findsectorextremalmaterialheightparams_t params;
        if(ref == SPREF_MIN_MID_MATERIAL || ref == SPREF_MAX_MID_MATERIAL)
            part = LWS_MID;
        else if(ref == SPREF_MIN_TOP_MATERIAL || ref == SPREF_MAX_TOP_MATERIAL)
            part = LWS_UPPER;
        else // Then it's the bottom.
            part = LWS_LOWER;
        if(ref >= SPREF_MIN_BOTTOM_MATERIAL && ref <= SPREF_MIN_TOP_MATERIAL)
            findMin = true;
        else
            findMin = false;
        params.baseSec = sector;
        params.part = part;
        params.flags = (findMin? FSETHF_MIN : 0);
        params.val = (findMin? DDMAXFLOAT : DDMINFLOAT);
        P_Iteratep(sector, DMU_LINE, findSectorExtremalMaterialHeight, &params);
        if(height)
            *height = params.val;
        return true;
    }
    if(ref == SPREF_HIGHEST_CEILING)
    {
        otherSec = P_FindSectorSurroundingHighestCeiling(sector, DDMINFLOAT, &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_CEILING_MATERIAL);
    }
    else if(ref == SPREF_HIGHEST_FLOOR)
    {
        otherSec = P_FindSectorSurroundingHighestFloor(sector, DDMINFLOAT, &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_CEILING_MATERIAL);
    }
    else if(ref == SPREF_LOWEST_CEILING)
    {
        otherSec = P_FindSectorSurroundingLowestCeiling(sector, DDMAXFLOAT, &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_CEILING_MATERIAL);
    }
    else if(ref == SPREF_LOWEST_FLOOR)
    {
        otherSec = P_FindSectorSurroundingLowestFloor(sector, DDMAXFLOAT, &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_FLOOR_MATERIAL);
    }
    else if(ref == SPREF_NEXT_HIGHEST_CEILING)
    {
        otherSec = P_FindSectorSurroundingNextHighestCeiling(sector,
                        P_GetDoublep(sector, DMU_CEILING_HEIGHT), &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_CEILING_MATERIAL);
    }
    else if(ref == SPREF_NEXT_HIGHEST_FLOOR)
    {
        otherSec = P_FindSectorSurroundingNextHighestFloor(sector,
                        P_GetDoublep(sector, DMU_FLOOR_HEIGHT), &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_FLOOR_MATERIAL);
    }
    else if(ref == SPREF_NEXT_LOWEST_CEILING)
    {
        otherSec = P_FindSectorSurroundingNextLowestCeiling(sector,
                        P_GetDoublep(sector, DMU_CEILING_HEIGHT), &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_CEILING_MATERIAL);
    }
    else if(ref == SPREF_NEXT_LOWEST_FLOOR)
    {
        otherSec = P_FindSectorSurroundingNextLowestFloor(sector,
                        P_GetDoublep(sector, DMU_FLOOR_HEIGHT), &otherHeight);
        if(otherSec)
            otherMat = (world_Material *) P_GetPtrp(otherSec, DMU_FLOOR_MATERIAL);
    }
    if(!otherSec)
        return false;
    if(height)
        *height = otherHeight;
    if(mat)
        *mat = otherMat;
    if(planeSector)
        *planeSector = otherSec;
    return true;
}
/**
 * DJS - Why find the highest??? Surely unlogical to mod authors.
 * IMO if a user references multiple sectors, the one with the lowest ID
 * should be chosen (the same way it works for FIND(act)TAGGED). If that
 * happens to be zero - so be it.
 */
int C_DECL XSTrav_HighestSectorType(Sector *sec, dd_bool ceiling,
                                    void *context, void *context2,
                                    mobj_t *activator)
{
    DENG_UNUSED(ceiling);
    DENG_UNUSED(context);
    DENG_UNUSED(activator);
    int        *type = (int *) context2;
    xsector_t  *xsec = P_ToXSector(sec);
    if(xsec->special > *type)
        *type = xsec->special;
    return true; // Keep looking...
}
void XS_InitMovePlane(Line *line)
{
    xline_t *xline = P_ToXLine(line);
    xline->xg->fdata = xline->xg->info.fparm[5];
    xline->xg->idata = true; // Play sound.
};
int C_DECL XSTrav_MovePlane(Sector *sector, dd_bool ceiling, void *context,
                            void *context2, mobj_t * /*activator*/)
{
    LOG_AS("XSTrav_MovePlane");
    DENG2_ASSERT(sector);
    Line *line        = (Line *) context;
    DENG2_ASSERT(line);
    linetype_t *info  = (linetype_t *) context2;
    DENG2_ASSERT(info);
    xline_t *xline    = P_ToXLine(line);
    dd_bool playsound = xline->xg->idata;
    LOG_MAP_MSG_XGDEVONLY2("Sector %i (by line %i of type %i)",
           P_ToIndex(sector) << P_ToIndex(line) << info->id);
    xgplanemover_t *mover = XS_GetPlaneMover(sector, ceiling);
    if(P_IsDummy(line))
    {
        LOG_MAP_ERROR("Attempted to use a dummy line as XGPlaneMover origin. "
                      "Plane in sector %i will not be moved.") << P_ToIndex(sector);
        return true; // Keep looking.
    }
    mover->origin = line;
    {
    coord_t temp = mover->destination;
    XS_GetPlane(line, sector, info->iparm[2], NULL, &temp, 0, 0);
    mover->destination = temp + info->fparm[2];
    }
    mover->speed = info->fparm[0];
    mover->crushSpeed = info->fparm[1];
    mover->minInterval = FLT2TIC(info->fparm[3]);
    mover->maxInterval = FLT2TIC(info->fparm[4]);
    mover->flags = info->iparm[3];
    mover->endSound = playsound ? info->iparm[5] : 0;
    mover->moveSound = playsound ? info->iparm[6] : 0;
    if(info->iparm[9] == SPREF_NONE || info->iparm[9] == SPREF_SPECIAL)
    {
        mover->setMaterial = (world_Material *) P_ToPtr(DMU_MATERIAL, info->iparm[10]);
    }
    else
    {
        if(!XS_GetPlane(line, sector, info->iparm[9], NULL, 0, &mover->setMaterial, 0))
            LOG_MAP_MSG_XGDEVONLY("Couldn't find suitable material to set when move ends!");
    }
    mover->timer = XG_RandomInt(mover->minInterval, mover->maxInterval);
    if(xline->xg->fdata > 0)
    {
        mover->timer = FLT2TIC(xline->xg->fdata);
        mover->flags |= PMF_WAIT;
    }
    xline->xg->fdata += info->fparm[6];
    if(playsound)
        XS_PlaneSound((Plane *) P_GetPtrp(sector, ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                      info->iparm[4]);
    world_Material *mat = nullptr;
    if(info->iparm[7] == SPREF_NONE || info->iparm[7] == SPREF_SPECIAL)
    {
        mat = (world_Material *) P_ToPtr(DMU_MATERIAL, info->iparm[8]);
    }
    else if(!XS_GetPlane(line, sector, info->iparm[7], nullptr, 0, &mat, 0))
    {
        LOG_MAP_MSG_XGDEVONLY("Couldn't find suitable material to set when move starts!");
    }
    if(mat)
    {
        XS_ChangePlaneMaterial(*sector, ceiling, *mat);
    }
    if(info->iparm[3] & PMF_ONE_SOUND_ONLY)
    {
        xline->xg->idata = false;
    }
    int st = info->iparm[12];
    if(info->iparm[11] != LPREF_NONE)
    {
        if(XL_TraversePlanes
           (line, info->iparm[11], info->iparm[12], 0, &st, false, 0,
            XSTrav_HighestSectorType))
        {
            XS_SetSectorType(sector, st);
        }
        else
        {
            LOG_MAP_MSG_XGDEVONLY("SECTOR TYPE NOT SET (nothing referenced)");
        }
    }
    st = info->iparm[14];
    if(info->iparm[13] != LPREF_NONE)
    {
        if(XL_TraversePlanes
           (line, info->iparm[13], info->iparm[14], 0, &st, false, 0,
            XSTrav_HighestSectorType))
        {   // OK, found one or more.
            mover->setSectorType = st;
        }
        else
        {
            LOG_MAP_MSG_XGDEVONLY("SECTOR TYPE WON'T BE CHANGED AT END (nothing referenced)");
            mover->setSectorType = -1;
        }
    }
    else
        mover->setSectorType = -1;
    return true; // Keep looking...
}
void XS_InitStairBuilder(Line *)
{
    int i;
    for(i = 0; i < numsectors; ++i)
    {
        P_GetXSector(i)->blFlags = 0;
    }
}
dd_bool XS_DoBuild(Sector* sector, dd_bool ceiling, Line* origin,
                   linetype_t* info, uint stepcount)
{
    static coord_t firstheight;
    float waittime;
    xsector_t* xsec;
    xgplanemover_t* mover;
    if(!sector)
        return false;
    xsec = P_ToXSector(sector);
    if(xsec->blFlags & BL_BUILT)
        return false; // Already built this one!
    xsec->blFlags |= BL_WAS_BUILT;
    mover = XS_GetPlaneMover(sector, ceiling);
    if(P_IsDummy(origin))
    {
        LOG_MAP_ERROR("Attempted to use a dummy line as XGPlaneMover origin while "
                      "building stairs in sector %i.") << P_ToIndex(sector);
        return false;
    }
    mover->origin = origin;
    if(stepcount != 0)
        firstheight = P_GetDoublep(sector, (ceiling? DMU_CEILING_HEIGHT:DMU_FLOOR_HEIGHT));
    mover->destination =
        firstheight + (stepcount + 1) * info->fparm[1];
    mover->speed = (info->fparm[0] + stepcount * info->fparm[6]);
    if(mover->speed <= 0)
        mover->speed = 1 / 1000;
    mover->minInterval = FLT2TIC(info->fparm[4]);
    mover->maxInterval = FLT2TIC(info->fparm[5]);
    if(info->iparm[8])
        mover->flags = PMF_CRUSH;
    mover->endSound = info->iparm[6];
    mover->moveSound = info->iparm[7];
    waittime = info->fparm[2] + info->fparm[3] * stepcount;
    if(waittime > 0)
    {
        mover->timer = FLT2TIC(waittime);
        mover->flags |= PMF_WAIT;
        mover->startSound = info->iparm[5];
    }
    else
    {
        mover->timer = XG_RandomInt(mover->minInterval, mover->maxInterval);
        XS_PlaneSound((Plane *) P_GetPtrp(sector, ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                      info->iparm[5]);
    }
    if(stepcount != 0)
    {
        XS_PlaneSound((Plane *) P_GetPtrp(sector, ceiling? DMU_CEILING_PLANE : DMU_FLOOR_PLANE),
                      info->iparm[4]);
    }
    return true; // Building has begun!
}
#define F_MATERIALSTOP          0x1
#define F_CEILING               0x2
typedef struct spreadbuildparams_s {
    Sector*             baseSec;
    world_Material*     baseMat;
    byte                flags;
    Line*               origin;
    linetype_t*         info;
    int                 stepCount;
    size_t              spreaded;
} spreadbuildparams_t;
int spreadBuild(void *ptr, void *context)
{
    Line                *li = (Line*) ptr;
    spreadbuildparams_t *params = (spreadbuildparams_t*) context;
    Sector              *frontSec, *backSec;
    frontSec = (Sector *) P_GetPtrp(li, DMU_FRONT_SECTOR);
    if(!frontSec || frontSec != params->baseSec)
        return false;
    backSec = (Sector *) P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!backSec)
        return false;
    if(params->flags & F_MATERIALSTOP)
    {   // Planepic must match.
        if(params->flags & F_CEILING)
        {
            if(P_GetPtrp(params->baseSec, DMU_CEILING_MATERIAL) !=
               params->baseMat)
                return false;
        }
        else
        {
            if(P_GetPtrp(params->baseSec, DMU_FLOOR_MATERIAL) !=
               params->baseMat)
                return false;
        }
    }
    if(P_ToXSector(backSec)->blFlags & BL_SPREADED)
        return false;
    XS_DoBuild(backSec, ((params->flags & F_CEILING)? true : false),
               params->origin, params->info, params->stepCount);
    params->spreaded++;
    return false; // Continue iteration.
}
static void markBuiltSectors(void)
{
    int i;
    for(i = 0; i < numsectors; ++i)
    {
        xsector_t *xsec = P_GetXSector(i);
        if(xsec->blFlags & BL_WAS_BUILT)
        {
            xsec->blFlags &= ~BL_WAS_BUILT;
            xsec->blFlags |= BL_BUILT;
        }
    }
}
static dd_bool spreadBuildToNeighborAll(Line *origin, linetype_t *info,
    dd_bool picstop, dd_bool ceiling, world_Material *myMat, int stepCount)
{
    int i;
    dd_bool result = false;
    spreadbuildparams_t params;
    params.baseMat = myMat;
    params.info = info;
    params.origin = origin;
    params.stepCount = stepCount;
    params.flags = 0;
    if(picstop)
        params.flags |= F_MATERIALSTOP;
    if(ceiling)
        params.flags |= F_CEILING;
    for(i = 0; i < numsectors; ++i)
    {
        Sector *sec;
        xsector_t *xsec = P_GetXSector(i);
        if(!(xsec->blFlags & BL_BUILT) || xsec->blFlags & BL_SPREADED)
            continue;
        xsec->blFlags |= BL_SPREADED;
        sec = (Sector *) P_ToPtr(DMU_SECTOR, i);
        params.baseSec = sec;
        params.spreaded = 0;
        P_Iteratep(sec, DMU_LINE, spreadBuild, &params);
        if(params.spreaded > 0)
            result = true;
    }
    return result;
}
#define F_MATERIALSTOP          0x1
#define F_CEILING               0x2
typedef struct findbuildneighborparams_s {
    Sector*             baseSec;
    world_Material*     baseMat;
    byte                flags;
    Line*               origin;
    linetype_t*         info;
    int                 stepCount;
    int                 foundIDX;
    Sector*             foundSec;
} findbuildneighborparams_t;
int findBuildNeighbor(void* ptr, void* context)
{
    Line*               li = (Line*) ptr;
    findbuildneighborparams_t *params =
        (findbuildneighborparams_t*) context;
    Sector*             frontSec, *backSec;
    int                 idx;
    frontSec = (Sector *) P_GetPtrp(li, DMU_FRONT_SECTOR);
    if(!frontSec || frontSec != params->baseSec)
        return false;
    backSec = (Sector *) P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!backSec)
        return false;
    if(params->flags & F_MATERIALSTOP)
    {   // Planepic must match.
        if(params->flags & F_CEILING)
        {
            if(P_GetPtrp(params->baseSec, DMU_CEILING_MATERIAL) !=
               params->baseMat)
                return false;
        }
        else
        {
            if(P_GetPtrp(params->baseSec, DMU_FLOOR_MATERIAL) !=
               params->baseMat)
                return false;
        }
    }
    if(P_ToXSector(backSec)->blFlags & BL_SPREADED)
        return false;
    idx = P_ToIndex(li);
    if(idx < params->foundIDX)
    {
        params->foundSec = backSec;
        params->foundIDX = idx;
    }
    return false; // Continue iteration.
}
dd_bool spreadBuildToNeighborLowestIDX(Line *origin, linetype_t *info,
    dd_bool picstop, dd_bool ceiling, world_Material *myMat, int stepcount,
    Sector **foundSec)
{
    int i;
    dd_bool result = false;
    findbuildneighborparams_t params;
    params.baseMat = myMat;
    params.info = info;
    params.origin = origin;
    params.stepCount = stepcount;
    params.flags = 0;
    if(picstop)
        params.flags |= F_MATERIALSTOP;
    if(ceiling)
        params.flags |= F_CEILING;
    for(i = 0; i < numsectors; ++i)
    {
        Sector *sec;
        xsector_t *xsec = P_GetXSector(i);
        if(!(xsec->blFlags & BL_BUILT) || xsec->blFlags & BL_SPREADED)
            continue;
        xsec->blFlags |= BL_SPREADED;
        sec = (Sector *) P_ToPtr(DMU_SECTOR, i);
        params.baseSec = sec;
        params.foundIDX = numlines;
        params.foundSec = NULL;
        P_Iteratep(sec, DMU_LINE, findBuildNeighbor, &params);
        if(params.foundSec)
        {
            result = true;
            *foundSec = params.foundSec;
        }
    }
    return result;
}
int C_DECL XSTrav_BuildStairs(Sector *sector, dd_bool ceiling, void *context,
                              void *context2, mobj_t *activator)
{
    LOG_AS("XSTrav_BuildStairs");
    uint stepCount   = 0;
    Line *origin     = (Line *) context;
    linetype_t *info = (linetype_t *) context2;
    Sector *foundSec = 0;
    dd_bool picstop  = info->iparm[2] != 0;
    dd_bool spread   = info->iparm[3] != 0;
    world_Material *myMat;
    DENG_UNUSED(activator);
    LOG_MAP_MSG_XGDEVONLY2("Sector %i, %s", P_ToIndex(sector) << (ceiling? "ceiling" : "floor"));
    myMat = (world_Material *) (ceiling ? P_GetPtrp(sector, DMU_CEILING_MATERIAL) :
                                    P_GetPtrp(sector, DMU_FLOOR_MATERIAL));
    XS_DoBuild(sector, ceiling, origin, info, 0);
    stepCount++;
    if(spread)
    {
        dd_bool             found;
        do
        {
            markBuiltSectors();
            found = spreadBuildToNeighborAll(origin, info, picstop, ceiling,
                                             myMat, stepCount);
            stepCount++;
        } while(found);
    }
    else
    {
        dd_bool             found;
        do
        {
            found = false;
            markBuiltSectors();
            if(spreadBuildToNeighborLowestIDX(origin, info, picstop,
                                              ceiling, myMat, stepCount,
                                              &foundSec))
            {
                XS_DoBuild(foundSec, ceiling, origin, info, stepCount);
                found = true;
            }
            stepCount++;
        } while(found);
    }
    return true; // Continue searching for planes...
}
int C_DECL XSTrav_SectorSound(Sector* sec, dd_bool ceiling, void* context,
                              void* context2, mobj_t* activator)
{
    DENG_UNUSED(activator);
    DENG_UNUSED(context);
    DENG_UNUSED(ceiling);
    linetype_t* info = (linetype_t *) context2;
    if(!info->iparm[3])
    {
        XS_SectorSound(sec, info->iparm[2]);
    }
    else
    {
        Plane* plane = (Plane *) P_GetPtrp(sec, info->iparm[3] == 2? DMU_CEILING_PLANE : DMU_FLOOR_PLANE);
        XS_PlaneSound(plane, info->iparm[2]);
    }
    return true;
}
int C_DECL XSTrav_PlaneMaterial(Sector *sec, dd_bool ceiling, void *context,
                                void *context2, mobj_t * /*activator*/)
{
    LOG_AS("XSTrav_PlaneMaterial");
    DENG2_ASSERT(sec);
    Line *line       = (Line *) context;
    DENG2_ASSERT(line);
    linetype_t *info = (linetype_t *) context2;
    DENG2_ASSERT(info);
    world_Material *mat;
    if(info->iparm[2] == SPREF_NONE)
    {
        mat = (world_Material *) P_ToPtr(DMU_MATERIAL, info->iparm[3]);
    }
    else if(!XS_GetPlane(line, sec, info->iparm[2], nullptr, 0, &mat, 0))
    {
        LOG_MAP_MSG_XGDEVONLY2("Sector %i, couldn't find suitable material!", P_ToIndex(sec));
    }
    if(mat)
    {
        XS_ChangePlaneMaterial(*sec, ceiling, *mat);
    }
    if(info->iparm[7])
    {
        de::Vector3f const color(info->iparm[4], info->iparm[5], info->iparm[6]);
        XS_ChangePlaneColor(*sec, ceiling, color / 255.f);
    }
    return true;
}
int C_DECL XSTrav_SectorType(Sector* sec, dd_bool /*ceiling*/,
                             void* /*context*/, void* context2,
                             mobj_t* /*activator*/)
{
    linetype_t*         info = (linetype_t *) context2;
    XS_SetSectorType(sec, info->iparm[2]);
    return true;
}
int C_DECL XSTrav_SectorLight(Sector* sector, dd_bool /*ceiling*/,
                              void* context, void* context2,
                              mobj_t* /*activator*/)
{
    LOG_AS("XSTrav_SectorLight");
    Line*               line = (Line *) context;
    linetype_t*         info = (linetype_t *) context2;
    int                 num;
    float               usergb[3];
    float               lightLevel = 0;
    if(info->iparm[2])
    {
        switch(info->iparm[4])
        {
        default:
        case LIGHTREF_NONE:
            lightLevel = 0;
            break;
        case LIGHTREF_MY:
            {
            Sector *frontSec = (Sector *) P_GetPtrp(line, DMU_FRONT_SECTOR);
            lightLevel = P_GetFloatp(frontSec, DMU_LIGHT_LEVEL);
            }
            break;
        case LIGHTREF_BACK:
            {
            Sector *backSec = (Sector *) P_GetPtrp(line, DMU_BACK_SECTOR);
            if(backSec)
                lightLevel = P_GetFloatp(backSec, DMU_LIGHT_LEVEL);
            }
            break;
        case LIGHTREF_ORIGINAL:
            lightLevel = P_ToXSector(sector)->origLight;
            break;
        case LIGHTREF_CURRENT:
            lightLevel = P_GetFloatp(sector, DMU_LIGHT_LEVEL);
            break;
        case LIGHTREF_HIGHEST:
            P_FindSectorSurroundingHighestLight(sector, &lightLevel);
            break;
        case LIGHTREF_LOWEST:
            P_FindSectorSurroundingLowestLight(sector, &lightLevel);
            break;
        case LIGHTREF_NEXT_HIGHEST:
            {
            float currentLevel = P_GetFloatp(sector, DMU_LIGHT_LEVEL);
            P_FindSectorSurroundingNextHighestLight(sector, currentLevel, &lightLevel);
            if(lightLevel < currentLevel)
                lightLevel = currentLevel;
            }
            break;
        case LIGHTREF_NEXT_LOWEST:
            {
            float currentLevel = P_GetFloatp(sector, DMU_LIGHT_LEVEL);
            P_FindSectorSurroundingNextLowestLight(sector, currentLevel, &lightLevel);
            if(lightLevel > currentLevel)
                lightLevel = currentLevel;
            }
            break;
        }
        lightLevel += (float) info->iparm[5] / 255.f;
        if(lightLevel < 0)
            lightLevel = 0;
        if(lightLevel > 1)
            lightLevel = 1;
        P_SetFloatp(sector, DMU_LIGHT_LEVEL, lightLevel);
    }
    if(info->iparm[3])
    {
        switch(info->iparm[6])
        {
        case LIGHTREF_MY:
            {
            Sector *sector = (Sector *) P_GetPtrp(line, DMU_FRONT_SECTOR);
            P_GetFloatpv(sector, DMU_COLOR, usergb);
            break;
            }
        case LIGHTREF_BACK:
            {
            Sector *sector = (Sector *) P_GetPtrp(line, DMU_BACK_SECTOR);
            if(sector)
                P_GetFloatpv(sector, DMU_COLOR, usergb);
            else
            {
                LOG_MAP_MSG_XGDEVONLY("Warning, the referenced Line has no back sector. Using default color");
                memset(usergb, 0, sizeof(usergb));
            }
            break;
            }
        case LIGHTREF_ORIGINAL:
            memcpy(usergb, P_ToXSector(sector)->origRGB, sizeof(float) * 3);
            break;
        default:
            memset(usergb, 0, sizeof(usergb));
            break;
        }
        for(num = 0; num < 3; ++num)
        {
            float f = usergb[num] + (float) info->iparm[7 + num] / 255.f;
            if(f < 0)
                f = 0;
            if(f > 1)
                f = 1;
            P_SetFloatp(sector, TO_DMU_COLOR(num), f);
        }
    }
    return true;
}
int C_DECL XSTrav_MimicSector(Sector *sector, dd_bool /*ceiling*/,
                              void *context, void *context2,
                              mobj_t * /*activator*/)
{
    LOG_AS("XSTrav_MimicSector");
    Line *line = (Line *) context;
    linetype_t *info = (linetype_t *) context2;
    Sector *from = NULL;
    int refdata;
    switch(info->iparm[2])
    {
    case SPREF_TAGGED_FLOOR:
    case SPREF_TAGGED_CEILING:
    case SPREF_INDEX_FLOOR:
    case SPREF_INDEX_CEILING:
    case SPREF_ACT_TAGGED_FLOOR:
    case SPREF_ACT_TAGGED_CEILING:
        if(info->iparm[3] >= 0)
            refdata = info->iparm[3];
        break;
    case SPREF_LINE_ACT_TAGGED_FLOOR:
    case SPREF_LINE_ACT_TAGGED_CEILING:
        if(info->actTag >= 0)
            refdata = info->actTag;
        break;
    default:
        refdata = 0;
        break;
    }
    if(!XS_GetPlane(line, sector, info->iparm[2], &refdata, 0, 0, &from))
    {
        LOG_MAP_MSG_XGDEVONLY2("No suitable neighbor for %i", P_ToIndex(sector));
        return true;
    }
    if(from == sector)
        return true;
    LOG_MAP_MSG_XGDEVONLY2("Sector %i mimicking sector %i", P_ToIndex(sector) << P_ToIndex(from));
    P_CopySector(sector, from);
    P_ChangeSector(sector, false /*don't crush*/);
    XS_SetSectorType(sector, P_ToXSector(from)->special);
    if(P_ToXSector(from)->xg)
        memcpy(P_ToXSector(sector)->xg, P_ToXSector(from)->xg, sizeof(xgsector_t));
    return true;
}
int C_DECL XSTrav_Teleport(Sector* sector, dd_bool /*ceiling*/, void* /*context*/,
                           void* context2, mobj_t* thing)
{
    LOG_AS("XSTrav_Teleport");
    mobj_t*         mo = NULL;
    dd_bool         ok = false;
    linetype_t*     info = (linetype_t *) context2;
    if(thing->flags2 & MF2_NOTELEPORT)
    {
        LOG_MAP_MSG_XGDEVONLY2("Activator is unteleportable (THING type %i)", thing->type);
        return false;
    }
    P_IterateThinkers(P_MobjThinker, [&mo, &ok, sector](thinker_t *th) {
        mo = reinterpret_cast<mobj_t *>(th);
        if (Mobj_Sector(mo) == sector && mo->type == MT_TELEPORTMAN)
        {
            ok = true;
            return de::LoopAbort;
        }
        return de::LoopContinue;
    });
    if(ok)
    {
        mobj_t *flash;
        unsigned an;
        coord_t oldpos[3];
        coord_t thfloorz, thceilz;
        coord_t aboveFloor, fogDelta = 0;
        angle_t oldAngle;
        LOG_MAP_MSG_XGDEVONLY2("Sector %i, %s, %s%s",
                P_ToIndex(sector)
                << (info->iparm[2]? "No Flash"   : "")
                << (info->iparm[3]? "Play Sound" : "Silent")
                << (info->iparm[4]? " Stomp"     : ""));
        if(!P_TeleportMove(thing, mo->origin[VX], mo->origin[VY], (info->iparm[4] > 0? 1 : 0)))
        {
            LOG_MAP_MSG_XGDEVONLY("No free space at teleport exit. Aborting teleport...");
            return false;
        }
        memcpy(oldpos, thing->origin, sizeof(thing->origin));
        oldAngle = thing->angle;
        thfloorz = P_GetDoublep(Mobj_Sector(thing), DMU_FLOOR_HEIGHT);
        thceilz  = P_GetDoublep(Mobj_Sector(thing), DMU_CEILING_HEIGHT);
        aboveFloor = thing->origin[VZ] - thfloorz;
        if(thing->player)
        {
            if((thing->player->plr->mo->flags2 & MF2_FLY) && aboveFloor)
            {
                thing->origin[VZ] = thfloorz + aboveFloor;
                if(thing->origin[VZ] + thing->height > thceilz)
                {
                    thing->origin[VZ] = thceilz - thing->height;
                }
                thing->player->viewZ =
                    thing->origin[VZ] + thing->player->viewHeight;
            }
            else
            {
                thing->origin[VZ] = thfloorz;
                thing->player->viewZ =
                    thing->origin[VZ] + thing->player->viewHeight;
                thing->dPlayer->lookDir = 0; /* $unifiedangles */
            }
#if __JHERETIC__
            if(!thing->player->powers[PT_WEAPONLEVEL2])
#endif
            {
                thing->reactionTime = 18;
            }
            thing->dPlayer->flags |= DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM;
        }
#if __JHERETIC__
        else if(thing->flags & MF_MISSILE)
        {
            thing->origin[VZ] = thfloorz + aboveFloor;
            if(thing->origin[VZ] + thing->height > thceilz)
            {
                thing->origin[VZ] = thceilz - thing->height;
            }
        }
#endif
        else
        {
            thing->origin[VZ] = thfloorz;
        }
        if(!info->iparm[2])
        {
#if __JHERETIC__
            fogDelta = ((thing->flags & MF_MISSILE)? 0 : TELEFOGHEIGHT);
#endif
            if((flash = P_SpawnMobjXYZ(MT_TFOG, oldpos[VX], oldpos[VY],
                                      oldpos[VZ] + fogDelta,
                                      oldAngle + ANG180, 0)))
            {
                if(info->iparm[3])
                    S_StartSound(info->iparm[3], flash);
            }
        }
        an = mo->angle >> ANGLETOFINESHIFT;
        if(!info->iparm[2])
        {
            if((flash = P_SpawnMobjXYZ(MT_TFOG,
                                      mo->origin[VX] + 20 * FIX2FLT(finecosine[an]),
                                      mo->origin[VY] + 20 * FIX2FLT(finesine[an]),
                                      mo->origin[VZ] + fogDelta, mo->angle, 0)))
            {
                if(info->iparm[3])
                    S_StartSound(info->iparm[3], flash);
            }
        }
        thing->angle = mo->angle;
        if(thing->flags2 & MF2_FLOORCLIP)
        {
            thing->floorClip = 0;
            if(FEQUAL(thing->origin[VZ], P_GetDoublep(Mobj_Sector(thing), DMU_FLOOR_HEIGHT)))
            {
                terraintype_t const *tt = P_MobjFloorTerrain(thing);
                if(tt->flags & TTF_FLOORCLIP)
                {
                    thing->floorClip = 10;
                }
            }
        }
        if(thing->flags & MF_MISSILE)
        {
            an >>= ANGLETOFINESHIFT;
            thing->mom[MX] = thing->info->speed * FIX2FLT(finecosine[an]);
            thing->mom[MY] = thing->info->speed * FIX2FLT(finesine[an]);
        }
        else
        {
            thing->mom[MX] = thing->mom[MY] = thing->mom[MZ] = 0;
        }
    }
    else
    {   // Keep looking, there may be another referenced sector we could
        LOG_MAP_MSG_XGDEVONLY2("No teleport exit in referenced sector (ID %i)."
               " Continuing search...", P_ToIndex(sector));
        return true;
    }
    return false; // Only do this once.
}
int XF_FindRewindMarker(char *func, int pos)
{
    while(pos > 0 && func[pos] != '>')
        pos--;
    if(func[pos] == '>')
        pos++;
    return pos;
}
int XF_GetCount(function_t * fn, int *pos)
{
    char       *end;
    int         count;
    count = strtol(fn->func + *pos, &end, 10);
    *pos = end - fn->func;
    return count;
}
float XF_GetValue(function_t * fn, int pos)
{
    int         ch;
    if(fn->func[pos] == '/' || fn->func[pos] == '%')
    {
        return strtod(fn->func + pos + 1, 0);
    }
    ch = tolower(fn->func[pos]);
    return (ch - 'a') / 25.0f;
}
/**
 * Returns the position of the next value.
 * Repeat counting is handled here.
 * Poke should be true only if fn->pos is really about to move.
 */
int XF_FindNextPos(function_t *fn, int pos, dd_bool poke, Sector *sec)
{
    int startpos = pos;
    int c;
    char *ptr;
    if(fn->repeat > 0)
    {
        if(poke)
            fn->repeat--;
        return pos;
    }
    if(fn->func[pos] == '/' || fn->func[pos] == '%')
    {
        double dvalue = strtod(fn->func + pos + 1, &ptr);
        DENG_UNUSED(dvalue);
        pos = ptr - fn->func; // Go to the end.
    }
    else
    {   // It's just a normal character [a-z,A-Z].
        pos++;
    }
    while(pos != startpos && fn->func[pos])
    {
        if(isdigit(fn->func[pos]))
        {   // A repeat!
            c = XF_GetCount(fn, &pos) - 1;
            if(poke)
                fn->repeat = c;
            return pos;
        }
        if(fn->func[pos] == '!')
        {   // Chain event.
            pos++;
            c = XF_GetCount(fn, &pos);
            if(poke)
            {
                XS_DoChain(sec, XSCE_FUNCTION, c, NULL);
            }
            continue;
        }
        if(fn->func[pos] == '#')
        {   // Set timer.
            pos++;
            c = XF_GetCount(fn, &pos);
            if(poke)
            {
                fn->timer = 0;
                fn->maxTimer = c;
            }
            continue;
        }
        if(fn->func[pos] == '?')
        {   // Random timer.
            pos++;
            c = XF_GetCount(fn, &pos);
            if(poke)
            {
                fn->timer = 0;
                fn->maxTimer = XG_RandomInt(0, c);
            }
            continue;
        }
        if(fn->func[pos] == '<')
        {   // Rewind.
            pos = XF_FindRewindMarker(fn->func, pos);
            continue;
        }
        if(poke)
        {
            if(islower(fn->func[pos]) || fn->func[pos] == '/')
            {
                int     next = XF_FindNextPos(fn, pos, false, sec);
                if(fn->func[next] == '.')
                {
                    pos++;
                    continue;
                }
                break;
            }
        }
        else if(fn->func[pos] == '.')
        {
            break;
        }
        if(isalpha(fn->func[pos]) || fn->func[pos] == '/' ||
           fn->func[pos] == '%')
            break;
        pos++;
    }
    return pos;
}
/**
 * Tick the function, update value.
 */
void XF_Ticker(function_t* fn, Sector* sec)
{
    int                 next;
    float               inter;
    fn->oldValue = fn->value;
    if(!ISFUNC(fn) || fn->link)
        return;
    if(fn->timer++ >= fn->maxTimer)
    {
        fn->timer = 0;
        fn->maxTimer = XG_RandomInt(fn->minInterval, fn->maxInterval);
        fn->pos = XF_FindNextPos(fn, fn->pos, true, sec);
    }
    /**
     * Syntax:
     *
     * abcdefghijlkmnopqrstuvwxyz (26)
     *
     * az.< (fade from 0 to 1, break interpolation and repeat)
     * [note that AZ.AZ is the same as AZAZ]
     * [also note that a.z is the same as z]
     * az.>mz< (fade from 0 to 1, break, repeat fade from 0.5 to 1 to 0.5)
     * 10a10z is the same as aaaaaaaaaazzzzzzzzzz
     * aB or Ba do not interpolate
     * zaN (interpolate from 1 to 0, wait at 0, stay at N)
     * za.N (interpolate from 1 to 0, skip to N)
     * 1A is the same as A
     */
    if(!fn->func[fn->pos])
        return;
    if(isupper(fn->func[fn->pos]) || fn->func[fn->pos] == '%')
    {   // No interpolation.
        fn->value = XF_GetValue(fn, fn->pos);
    }
    else
    {
        inter = 0;
        next = XF_FindNextPos(fn, fn->pos, false, sec);
        if(islower(fn->func[next]) || fn->func[next] == '/')
        {
            if(fn->maxTimer)
                inter = fn->timer / (float) fn->maxTimer;
        }
        fn->value = (1 - inter) * XF_GetValue(fn, fn->pos) +
            inter * XF_GetValue(fn, next);
    }
    fn->value = fn->value * fn->scale + fn->offset;
}
void XS_UpdatePlanes(Sector* sec)
{
    int i;
    xgsector_t* xg;
    function_t* fn;
    dd_bool docrush;
    xg = P_ToXSector(sec)->xg;
    docrush = (xg->info.flags & STF_CRUSH) != 0;
    fn = &xg->plane[XGSP_FLOOR];
    if(UPDFUNC(fn))
    {
        i = fn->value - P_GetFloatp(sec, DMU_FLOOR_HEIGHT);
        if(i)
        {
            T_MovePlane(sec, abs(i), fn->value, docrush, 0, SIGN(i));
        }
    }
    fn = &xg->plane[XGSP_CEILING];
    if(UPDFUNC(fn))
    {
        i = fn->value - P_GetFloatp(sec, DMU_CEILING_HEIGHT);
        if(i)
        {
            T_MovePlane(sec, abs(i), fn->value, docrush, 1, SIGN(i));
        }
    }
}
void XS_DoChain(Sector *sec, int ch, int activating, void *act_thing)
{
    LOG_AS("XS_DoChain");
    xgsector_t *xg;
    sectortype_t *info;
    float flevtime = TIC2FLT(mapTime);
    Line *dummyLine;
    xline_t *xdummyLine;
    linetype_t *ltype;
    xg = P_ToXSector(sec)->xg;
    info = &xg->info;
    if(ch < XSCE_NUM_CHAINS)
    {
        if(!info->count[ch])
            return;
        if(flevtime < info->start[ch] ||
           (info->end[ch] > 0 && flevtime > info->end[ch]))
            return; // Not operating at this time.
        xg->chainTimer[ch] =
            XG_RandomInt(FLT2TIC(info->interval[ch][0]),
                         FLT2TIC(info->interval[ch][1]));
    }
    dummyLine = P_AllocDummyLine();
    xdummyLine = P_ToXLine(dummyLine);
    xdummyLine->xg = (xgline_t *) Z_Calloc(sizeof(xgline_t), PU_MAP, 0);
    P_SetPtrp(dummyLine, DMU_FRONT_SECTOR, sec);
    xdummyLine->special = (ch == XSCE_FUNCTION ? activating : info->chain[ch]);
    xdummyLine->tag = P_ToXSector(sec)->tag;
    ltype = XL_GetType(xdummyLine->special);
    if(!ltype)
    {
        LOG_MAP_MSG_XGDEVONLY2("Unknown XG line type %i", xdummyLine->special);
        Z_Free(xdummyLine->xg);
        P_FreeDummyLine(dummyLine);
        return;
    }
    memcpy(&xdummyLine->xg->info, ltype, sizeof(*ltype));
    if(act_thing)
        xdummyLine->xg->activator = act_thing;
    else
        xdummyLine->xg->activator = NULL;
    xdummyLine->xg->active = (ch == XSCE_FUNCTION ? false : !activating);
    LOG_MAP_MSG_XGDEVONLY2("Dummy line will show up as %i", P_ToIndex(dummyLine));
    if(XL_LineEvent((ch == XSCE_FUNCTION ? XLE_FUNC : XLE_CHAIN), 0,
                    dummyLine, 0, act_thing))
    {   // Success!
        if(ch < XSCE_NUM_CHAINS)
        {
            if(info->count[ch] > 0)
            {
                info->count[ch]--;
                LOG_MAP_MSG_XGDEVONLY2("%s, sector %i (activating=%i): Counter now at %i",
                        (  ch == XSCE_FLOOR    ? "FLOOR"
                         : ch == XSCE_CEILING  ? "CEILING"
                         : ch == XSCE_INSIDE   ? "INSIDE"
                         : ch == XSCE_TICKER   ? "TICKER"
                         : ch == XSCE_FUNCTION ? "FUNCTION" : "???")
                        << P_ToIndex(sec)
                        << activating << info->count[ch]);
            }
        }
    }
    Z_Free(xdummyLine->xg);
    P_FreeDummyLine(dummyLine);
}
static dd_bool checkChainRequirements(Sector *sec, mobj_t *mo, int ch, dd_bool *activating)
{
    xgsector_t*         xg;
    sectortype_t*       info;
    player_t*           player = mo->player;
    int                 flags;
    dd_bool             typePasses = false;
    xg = P_ToXSector(sec)->xg;
    info = &xg->info;
    flags = info->chainFlags[ch];
    if((flags & (SCEF_ANY_A | SCEF_ANY_D | SCEF_TICKER_A | SCEF_TICKER_D)) ||
       ((flags & (SCEF_PLAYER_A | SCEF_PLAYER_D)) && player) ||
       ((flags & (SCEF_OTHER_A | SCEF_OTHER_D)) && !player) ||
       ((flags & (SCEF_MONSTER_A | SCEF_MONSTER_D)) && (mo->flags & MF_COUNTKILL)) ||
       ((flags & (SCEF_MISSILE_A | SCEF_MISSILE_D)) && (mo->flags & MF_MISSILE)))
        typePasses = true;
    if(!typePasses)
        return false; // Wrong type.
    if(player)
        *activating = !(flags & SCEF_PLAYER_D);
    else if(mo->flags & MF_COUNTKILL)
        *activating = !(flags & SCEF_MONSTER_D);
    else if(mo->flags & MF_MISSILE)
        *activating = !(flags & SCEF_MISSILE_D);
    else if(flags & (SCEF_ANY_A | SCEF_ANY_D))
        *activating = !(flags & SCEF_ANY_D);
    else
        *activating = !(flags & SCEF_OTHER_D);
    switch(ch)
    {
    case XSCE_FLOOR:
        if(mo->origin[VZ] > P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + 0.0001)
            return false;
        break;
    case XSCE_CEILING:
        if(mo->origin[VZ] + mo->height < P_GetDoublep(sec, DMU_CEILING_HEIGHT) - 0.0001)
            return false;
        break;
    default:
        break;
    }
    return true;
}
typedef struct {
    Sector *sec;
    int data;
} xstrav_sectorchainparams_t;
int XSTrav_SectorChain(thinker_t *th, void *context)
{
    xstrav_sectorchainparams_t *params = (xstrav_sectorchainparams_t *) context;
    mobj_t *mo = (mobj_t *) th;
    if(params->sec == Mobj_Sector(mo))
    {
        dd_bool activating;
        if(checkChainRequirements(params->sec, mo, params->data, &activating))
            XS_DoChain(params->sec, params->data, activating, mo);
    }
    return false; // Continue iteration.
}
void P_ApplyWind(mobj_t *mo, Sector *sec)
{
    sectortype_t *info;
    float ang;
    if(mo->player && (mo->player->plr->flags & DDPF_CAMERA))
        return; // Wind does not affect cameras.
    info = &(P_ToXSector(sec)->xg->info);
    ang = DD_PI * info->windAngle / 180;
    if(IS_CLIENT)
    {
        if(!mo->player || mo->player != &players[CONSOLEPLAYER])
            return;
    }
    if(((info->flags & STF_PLAYER_WIND) && mo->player) ||
       ((info->flags & STF_OTHER_WIND) && !mo->player) ||
       ((info->flags & STF_MONSTER_WIND) && (mo->flags & MF_COUNTKILL)) ||
       ((info->flags & STF_MISSILE_WIND) && (mo->flags & MF_MISSILE)))
    {
        coord_t thfloorz = P_GetDoublep(Mobj_Sector(mo), DMU_FLOOR_HEIGHT);
        coord_t thceilz  = P_GetDoublep(Mobj_Sector(mo), DMU_CEILING_HEIGHT);
        if(!(info->flags & (STF_FLOOR_WIND | STF_CEILING_WIND)) ||
           ((info->flags & STF_FLOOR_WIND) && mo->origin[VZ] <= thfloorz) ||
           ((info->flags & STF_CEILING_WIND) &&
            mo->origin[VZ] + mo->height >= thceilz))
        {
            mo->mom[MZ] += info->verticalWind;
            mo->mom[MX] += cos(ang) * info->windSpeed;
            mo->mom[MY] += sin(ang) * info->windSpeed;
        }
    }
}
typedef struct {
    Sector *sec;
} xstrav_windparams_t;
int XSTrav_Wind(thinker_t *th, void *context)
{
    xstrav_windparams_t *params = (xstrav_windparams_t *) context;
    mobj_t *mo = (mobj_t *) th;
    if(params->sec == Mobj_Sector(mo))
    {
        P_ApplyWind(mo, params->sec);
    }
    return false; // Continue iteration.
}
/**
 * Makes sure the offset is in the range 0..64.
 */
void XS_ConstrainPlaneOffset(float *offset)
{
    if(*offset > 64)
        *offset -= 64;
    if(*offset < 0)
        *offset += 64;
}
/**
 * XG sectors get to think.
 */
void XS_Thinker(void *xsThinker)
{
    xsthinker_t* xs = (xsthinker_t *) xsThinker;
    Sector* sector = xs->sector;
    xsector_t *xsector = P_ToXSector(sector);
    xgsector_t* xg;
    sectortype_t* info;
    int i;
    if(!xsector) return; // Not an xsector? Most perculiar...
    xg = xsector->xg;
    if(!xg) return; // Not an extended sector.
    if(xg->disabled) return; // This sector is disabled.
    info = &xg->info;
    if(!IS_CLIENT)
    {
        for(i = 0; i < 2; ++i)
            XF_Ticker(&xg->plane[i], sector);
        XF_Ticker(&xg->light, sector);
        for(i = 0; i < 3; ++i)
            XF_Ticker(&xg->rgb[i], sector);
        for(i = 0; i < 3; ++i)
        {
            if(i < 2 && xg->plane[i].link)
                xg->plane[i].value = xg->plane[i].link->value;
            if(xg->rgb[i].link)
                xg->rgb[i].value = xg->rgb[i].link->value;
        }
        if(xg->light.link)
            xg->light.value = xg->light.link->value;
        XS_UpdatePlanes(sector);
        XS_UpdateLight(sector);
        for(i = 0; i < XSCE_NUM_CHAINS; ++i)
            xg->chainTimer[i]--;
        if(info->chain[XSCE_FLOOR] && xg->chainTimer[XSCE_FLOOR] <= 0)
        {
            xstrav_sectorchainparams_t params;
            params.sec = sector;
            params.data = XSCE_FLOOR;
            Thinker_Iterate((thinkfunc_t) P_MobjThinker, XSTrav_SectorChain, &params);
        }
        if(info->chain[XSCE_CEILING] && xg->chainTimer[XSCE_CEILING] <= 0)
        {
            xstrav_sectorchainparams_t params;
            params.sec = sector;
            params.data = XSCE_CEILING;
            Thinker_Iterate((thinkfunc_t) P_MobjThinker, XSTrav_SectorChain, &params);
        }
        if(info->chain[XSCE_INSIDE] && xg->chainTimer[XSCE_INSIDE] <= 0)
        {
            xstrav_sectorchainparams_t params;
            params.sec = sector;
            params.data = XSCE_INSIDE;
            Thinker_Iterate((thinkfunc_t) P_MobjThinker, XSTrav_SectorChain, &params);
        }
        if(info->chain[XSCE_TICKER] && xg->chainTimer[XSCE_TICKER] <= 0)
        {
            XS_DoChain(sector, XSCE_TICKER,
                       !(info->chainFlags[XSCE_TICKER] & SCEF_TICKER_D),
                       XG_DummyThing());
        }
        if(xg->info.ambientSound)
        {
            if(xg->timer-- < 0)
            {
                xg->timer = XG_RandomInt(FLT2TIC(xg->info.soundInterval[0]),
                                         FLT2TIC(xg->info.soundInterval[1]));
                S_SectorSound(sector, xg->info.ambientSound);
            }
        }
    }
    if(xg->info.materialMoveSpeed[0] != 0)
    {
        coord_t floorOffset[2];
        double ang = DD_PI * xg->info.materialMoveAngle[0] / 180;
        P_GetDoublepv(sector, DMU_FLOOR_MATERIAL_OFFSET_XY, floorOffset);
        floorOffset[VX] -= cos(ang) * xg->info.materialMoveSpeed[0];
        floorOffset[VY] -= sin(ang) * xg->info.materialMoveSpeed[0];
        P_SetDoublepv(sector, DMU_FLOOR_MATERIAL_OFFSET_XY, floorOffset);
    }
    if(xg->info.materialMoveSpeed[1] != 0)
    {
        coord_t ceilOffset[2];
        double ang = DD_PI * xg->info.materialMoveAngle[1] / 180;
        P_GetDoublepv(sector, DMU_CEILING_MATERIAL_OFFSET_XY, ceilOffset);
        ceilOffset[VX] -= cos(ang) * xg->info.materialMoveSpeed[1];
        ceilOffset[VY] -= sin(ang) * xg->info.materialMoveSpeed[1];
        P_SetDoublepv(sector, DMU_CEILING_MATERIAL_OFFSET_XY, ceilOffset);
    }
    if(xg->info.windSpeed || xg->info.verticalWind)
    {
        xstrav_windparams_t params;
        params.sec = sector;
        Thinker_Iterate((thinkfunc_t) P_MobjThinker, XSTrav_Wind, &params);
    }
}
coord_t XS_Gravity(Sector* sec)
{
    xsector_t* xsec;
    if(!sec) return P_GetGravity(); // World gravity.
    xsec = P_ToXSector(sec);
    if(!xsec->xg || !(xsec->xg->info.flags & STF_GRAVITY))
    {
        return P_GetGravity(); // World gravity.
    }
    else
    {   // Sector-specific gravity.
        coord_t gravity = xsec->xg->info.gravity;
        if(cfg.common.netGravity != -1)
            gravity *= (coord_t) cfg.common.netGravity / 100;
        return gravity;
    }
}
coord_t XS_Friction(Sector const *sector)
{
    auto const *xsec = P_ToXSector_const(sector);
    if(!xsec->xg || !(xsec->xg->info.flags & STF_FRICTION))
        return FRICTION_NORMAL; // Normal friction.
    return xsec->xg->info.friction;
}
/**
 * During update, definitions are re-read, so the pointers need to be
 * updated. However, this is a bit messy operation, prone to errors.
 * Instead, we just disable XG...
 */
void XS_Update(void)
{
    int i;
    xsector_t  *xsec;
    for(i = 0; i < numsectors; ++i)
    {
        xsec = P_ToXSector((Sector *) P_ToPtr(DMU_SECTOR, i));
        if(xsec->xg)
        {
            xsec->xg = 0;
            xsec->special = 0;
        }
    }
}
/**
 * $moveplane: Command line interface to the plane mover.
 */
D_CMD(MovePlane)
{
    DENG_UNUSED(src);
    dd_bool isCeiling = !stricmp(argv[0], "moveceil");
    dd_bool isBoth = !stricmp(argv[0], "movesec");
    dd_bool isOffset = false, isCrusher = false;
    Sector* sector = NULL;
    coord_t units = 0;
    float speed = FRACUNIT;
    int p = 0;
    coord_t floorheight, ceilingheight;
    xgplanemover_t* mover;
    if(argc < 2)
    {
        App_Log(DE2_SCR_NOTE, "Usage: %s (opts)", argv[0]);
        App_Log(DE2_LOG_SCR, "Opts can be:");
        App_Log(DE2_LOG_SCR, "  here [crush] [off] (z/units) [speed]");
        App_Log(DE2_LOG_SCR, "  at (x) (y) [crush] [off] (z/units) [speed]");
        App_Log(DE2_LOG_SCR, "  tag (sector-tag) [crush] [off] (z/units) [speed]");
        return true;
    }
    if(IS_CLIENT)
    {
        App_Log(DE2_SCR_ERROR, "Clients can't move planes");
        return false;
    }
    if(!stricmp(argv[1], "here"))
    {
        p = 2;
        if(!players[CONSOLEPLAYER].plr->mo)
            return false;
        sector = Mobj_Sector(players[CONSOLEPLAYER].plr->mo);
    }
    else if(!stricmp(argv[1], "at") && argc >= 4)
    {
        coord_t point[2];
        point[VX] = (coord_t)strtol(argv[2], 0, 0);
        point[VY] = (coord_t)strtol(argv[3], 0, 0);
        sector = Sector_AtPoint_FixedPrecision(point);
        p = 4;
    }
    else if(!stricmp(argv[1], "tag") && argc >= 3)
    {
        int tag = (short) strtol(argv[2], 0, 0);
        Sector* sec = NULL;
        iterlist_t* list;
        p = 3;
        list = P_GetSectorIterListForTag(tag, false);
        if(list)
        {   // Find the first sector with the tag.
            IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
            IterList_RewindIterator(list);
            while((sec = (Sector *) IterList_MoveIterator(list)) != NULL)
            {
                sector = sec;
                break;
            }
        }
    }
    else
    {   // Unknown mode.
        App_Log(DE2_SCR_ERROR, "Unknown mode");
        return false;
    }
    floorheight   = P_GetDoublep(sector, DMU_FLOOR_HEIGHT);
    ceilingheight = P_GetDoublep(sector, DMU_CEILING_HEIGHT);
    if(argc == p)
    {
        App_Log(DE2_LOG_MAP, "Ceiling = %g, Floor = %g", ceilingheight, floorheight);
        return true;
    }
    if(argc >= p + 1 && !stricmp(argv[p], "crush"))
    {
        isCrusher = true;
        ++p;
    }
    if(argc >= p + 1 && !stricmp(argv[p], "off"))
    {
        isOffset = true;
        ++p;
    }
    if(argc >= p + 1)
    {
        units = strtod(argv[p++], 0);
    }
    else
    {
        App_Log(DE2_SCR_ERROR, "You must specify Z-units");
        return false; // Required parameter missing.
    }
    if(argc >= p + 1)
    {
        speed = strtod(argv[p++], 0);
        if(speed < 0)
            speed = -speed;
    }
    if(!sector)
        return false;
    mover = XS_GetPlaneMover(sector, isCeiling);
    mover->destination = units + (isOffset ? (isCeiling ? ceilingheight : floorheight) : 0);
    mover->speed = speed;
    if(isCrusher)
    {
        mover->crushSpeed = speed * .5;  // Crush at half speed.
        mover->flags |= PMF_CRUSH;
    }
    if(isBoth)
        mover->flags |= PMF_OTHER_FOLLOWS;
    return true;
}
#endif
/** @file pause.cpp  Pausing the game.
 *
 * @authors Copyright © 1999-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2014 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1993-1996 id Software, Inc.
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "pause.h"
#include "d_net.h"
#include "d_netcl.h"
#include "d_netsv.h"
#include "g_common.h"
#include "hu_menu.h"
#include "hu_msg.h"
/** @file player.cpp  Common playsim routines relating to players.
 *
 * @authors Copyright © 2006-2019 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2015 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "player.h"
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <de/memory.h>
#include <de/Slope>
#include <de/Matrix>
#include <doomsday/plugins.h>
#include "d_net.h"
#include "d_netcl.h"
#include "d_netsv.h"
#include "dmu_lib.h"
#include "g_common.h"
#include "g_defs.h"
#include "gamesession.h"
#if __JHERETIC__ || __JHEXEN__
#  include "hu_inventory.h"
#endif
#include "p_actor.h"
#include "p_inventory.h"
#include "p_map.h"
#include "p_saveg.h"
#include "p_start.h"
#include "r_special.h"
/** @file polyobjs.h  Polyobject thinkers and management.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 * @authors Copyright © 1999 Activision
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "polyobjs.h"
#include "acs/script.h"
#include "dmu_lib.h"
#include "g_common.h"
#include "p_actor.h"
#include "p_map.h"
#include "p_mapspec.h"
#include "p_mapsetup.h"
#include "p_start.h"
#define POBJ_PERPETUAL  0xffffffffu  // -1
/** @file thingarchive.cpp  Map save state thing archive.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "thingarchive.h"
#include "mobj.h"
#include "p_saveg.h" /// @todo remove me
#include <de/memory.h>
#if __JHEXEN__
static ThingArchive::SerialId const TargetPlayerId = -2;
#endif
/** @file thinkerinfo.cpp  Game save thinker info.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2005-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * GPL: http://www.gnu.org/licenses/gpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details. You should have received a copy of the GNU
 * General Public License along with this program; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA</small>
 */
#include "common.h"
#include "thinkerinfo.h"
#include "acs/interpreter.h"
#include "acs/script.h"
#include "mobj.h"
#include "p_ceiling.h"
#include "p_door.h"
#include "p_floor.h"
#include "p_plat.h"
#include "p_scroll.h"
#include "p_switch.h"
#if __JHEXEN__
#  include "p_pillar.h"
#  include "p_waggle.h"
#endif
#include "polyobjs.h"
using namespace de;
using namespace common;
int P_PathXYTraverse2(coord_t fromX, coord_t fromY, coord_t toX, coord_t toY,
    int flags, traverser_t callback, void *context) {
    vec2d_t from = { fromX, fromY };
    vec2d_t to   = { toX, toY };
    return P_PathTraverse2(from, to, flags, callback, context);
}
int P_PathXYTraverse(coord_t fromX, coord_t fromY, coord_t toX, coord_t toY,
    traverser_t callback, void *context) {
    vec2d_t from = { fromX, fromY };
    vec2d_t to   = { toX, toY };
    return P_PathTraverse(from, to, callback, context);
}
struct TagList {
    iterlist_t *list;
    int tag;
};
static TagList *lineTagLists;
static uint numLineTagLists;
static TagList *sectorTagLists;
static uint numSectorTagLists;
Line *P_AllocDummyLine() {
    xline_t *extra = (xline_t *)Z_Calloc(sizeof(xline_t), PU_GAMESTATIC, 0);
    return (Line *)P_AllocDummy(DMU_LINE, extra);
}
void P_FreeDummyLine(Line *line) {
    Z_Free(P_DummyExtraData(line));
    P_FreeDummy(line);
}
void P_CopyLine(Line *dest, Line *src) {
    xline_t *xsrc  = P_ToXLine(src);
    xline_t *xdest = P_ToXLine(dest);
    if(src == dest)
        return; // no point copying self
    for(int i = 0; i < 2; ++i) // For each side
    {
        int sidx = (i==0? DMU_FRONT : DMU_BACK);
        Side *sidefrom = (Side *)P_GetPtrp(src, sidx);
        Side *sideto = (Side *)P_GetPtrp(dest, sidx);
        if(!sidefrom || !sideto)
            continue;
        float temp[4];
        coord_t itemp[2];
        P_SetPtrp    (sideto,   DMU_TOP_MATERIAL,               P_GetPtrp(sidefrom, DMU_TOP_MATERIAL));
        P_GetDoublepv(sidefrom, DMU_TOP_MATERIAL_OFFSET_XY,     itemp);
        P_SetDoublepv(sideto,   DMU_TOP_MATERIAL_OFFSET_XY,     itemp);
        P_GetFloatpv (sidefrom, DMU_TOP_COLOR,                  temp);
        P_SetFloatpv (sideto,   DMU_TOP_COLOR,                  temp);
        P_SetPtrp    (sideto,   DMU_MIDDLE_MATERIAL,            P_GetPtrp(sidefrom, DMU_MIDDLE_MATERIAL));
        P_GetDoublepv(sidefrom, DMU_MIDDLE_MATERIAL_OFFSET_XY,  itemp);
        P_SetDoublepv(sideto,   DMU_MIDDLE_MATERIAL_OFFSET_XY,  itemp);
        P_SetFloatpv (sideto,   DMU_MIDDLE_COLOR,               temp);
        P_SetIntp    (sideto,   DMU_MIDDLE_BLENDMODE,           P_GetIntp(sidefrom, DMU_MIDDLE_BLENDMODE));
        P_SetPtrp    (sideto,   DMU_BOTTOM_MATERIAL,            P_GetPtrp(sidefrom, DMU_BOTTOM_MATERIAL));
        P_GetDoublepv(sidefrom, DMU_BOTTOM_MATERIAL_OFFSET_XY,  itemp);
        P_SetDoublepv(sideto,   DMU_BOTTOM_MATERIAL_OFFSET_XY,  itemp);
        P_GetFloatpv (sidefrom, DMU_BOTTOM_COLOR,               temp);
        P_SetFloatpv (sideto,   DMU_BOTTOM_COLOR,               temp);
    }
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    xdest->special = xsrc->special;
    xdest->tag     = xsrc->tag;
    if(xsrc->xg && xdest->xg)
        std::memcpy(xdest->xg, xsrc->xg, sizeof(*xdest->xg));
    else
        xdest->xg = 0;
#else
    xdest->special = xsrc->special;
    xdest->arg1 = xsrc->arg1;
    xdest->arg2 = xsrc->arg2;
    xdest->arg3 = xsrc->arg3;
    xdest->arg4 = xsrc->arg4;
    xdest->arg5 = xsrc->arg5;
#endif
}
void P_CopySector(Sector *dest, Sector *src) {
    xsector_t *xsrc = P_ToXSector(src);
    xsector_t *xdest = P_ToXSector(dest);
    if(src == dest)
        return; // no point copying self.
    float ftemp[4];
    coord_t dtemp[2];
    P_SetFloatp  (dest, DMU_LIGHT_LEVEL,                P_GetFloatp(src, DMU_LIGHT_LEVEL));
    P_GetFloatpv (src,  DMU_COLOR,                      ftemp);
    P_SetFloatpv (dest, DMU_COLOR,                      ftemp);
    P_SetDoublep (dest, DMU_FLOOR_HEIGHT,               P_GetDoublep(src, DMU_FLOOR_HEIGHT));
    P_SetPtrp    (dest, DMU_FLOOR_MATERIAL,             P_GetPtrp(src, DMU_FLOOR_MATERIAL));
    P_GetFloatpv (src,  DMU_FLOOR_COLOR,                ftemp);
    P_SetFloatpv (dest, DMU_FLOOR_COLOR,                ftemp);
    P_GetDoublepv(src,  DMU_FLOOR_MATERIAL_OFFSET_XY,   dtemp);
    P_SetDoublepv(dest, DMU_FLOOR_MATERIAL_OFFSET_XY,   dtemp);
    P_SetIntp    (dest, DMU_FLOOR_SPEED,                P_GetIntp(src, DMU_FLOOR_SPEED));
    P_SetDoublep (dest, DMU_FLOOR_TARGET_HEIGHT,        P_GetFloatp(src, DMU_FLOOR_TARGET_HEIGHT));
    P_SetDoublep (dest, DMU_CEILING_HEIGHT,             P_GetDoublep(src, DMU_CEILING_HEIGHT));
    P_SetPtrp    (dest, DMU_CEILING_MATERIAL,           P_GetPtrp(src, DMU_CEILING_MATERIAL));
    P_GetFloatpv (src,  DMU_CEILING_COLOR,              ftemp);
    P_SetFloatpv (dest, DMU_CEILING_COLOR,              ftemp);
    P_GetDoublepv(src,  DMU_CEILING_MATERIAL_OFFSET_XY, dtemp);
    P_SetDoublepv(dest, DMU_CEILING_MATERIAL_OFFSET_XY, dtemp);
    P_SetIntp    (dest, DMU_CEILING_SPEED,              P_GetIntp(src, DMU_CEILING_SPEED));
    P_SetDoublep (dest, DMU_CEILING_TARGET_HEIGHT,      P_GetFloatp(src, DMU_CEILING_TARGET_HEIGHT));
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    xdest->special = xsrc->special;
    xdest->soundTraversed = xsrc->soundTraversed;
    xdest->soundTarget = xsrc->soundTarget;
#if __JHERETIC__
    xdest->seqType = xsrc->seqType;
#endif
    xdest->SP_floororigheight = xsrc->SP_floororigheight;
    xdest->SP_ceilorigheight = xsrc->SP_ceilorigheight;
    xdest->origLight = xsrc->origLight;
    std::memcpy(xdest->origRGB, xsrc->origRGB, sizeof(float) * 3);
    if(xsrc->xg && xdest->xg)
        std::memcpy(xdest->xg, xsrc->xg, sizeof(*xdest->xg));
    else
        xdest->xg = 0;
#else
    xdest->special = xsrc->special;
    xdest->soundTraversed = xsrc->soundTraversed;
    xdest->soundTarget = xsrc->soundTarget;
    xdest->seqType = xsrc->seqType;
#endif
}
void P_BuildLineTagLists() {
    P_DestroyLineTagLists();
    for(int i = 0; i < numlines; ++i) {
        Line *line  = (Line *)P_ToPtr(DMU_LINE, i);
        xline_t *xline = P_ToXLine(line);
#if !__JHEXEN__
        if(xline->tag) {
           iterlist_t *list = P_GetLineIterListForTag(xline->tag, true);
           IterList_PushBack(list, line);
        }
#else
        switch(xline->special) {
        default: break;
        case 121: ///< Line_SetIdentification.
            if(xline->arg1) {
                iterlist_t *list = P_GetLineIterListForTag((int) xline->arg1, true);
                IterList_PushBack(list, line);
            }
            xline->special = 0;
            break;
        }
#endif
    }
}
void P_DestroyLineTagLists() {
    if(numLineTagLists == 0)
        return;
    for(uint i = 0; i < numLineTagLists; ++i) {
        IterList_Clear(lineTagLists[i].list);
        IterList_Delete(lineTagLists[i].list);
    }
    free(lineTagLists);
    lineTagLists = 0;
    numLineTagLists = 0;
}
iterlist_t *P_GetLineIterListForTag(int tag, dd_bool createNewList) {
    for(uint i = 0; i < numLineTagLists; ++i) {
        if(lineTagLists[i].tag == tag)
            return lineTagLists[i].list;
    }
    if(!createNewList)
        return 0;
    numLineTagLists++;
    lineTagLists = (TagList *)realloc(lineTagLists, sizeof(TagList) * numLineTagLists);
    TagList *tagList = &lineTagLists[numLineTagLists - 1];
    tagList->tag = tag;
    return (tagList->list = IterList_New());
}
void P_BuildSectorTagLists() {
    P_DestroySectorTagLists();
    for(int i = 0; i < numsectors; ++i) {
        Sector *sec = (Sector *)P_ToPtr(DMU_SECTOR, i);
        xsector_t *xsec = P_ToXSector(sec);
        if(xsec->tag) {
            iterlist_t *list = P_GetSectorIterListForTag(xsec->tag, true);
            IterList_PushBack(list, sec);
        }
    }
}
void P_DestroySectorTagLists() {
    if(numSectorTagLists == 0)
        return;
    for(uint i = 0; i < numSectorTagLists; ++i) {
        IterList_Clear(sectorTagLists[i].list);
        IterList_Delete(sectorTagLists[i].list);
    }
    free(sectorTagLists);
    sectorTagLists = 0;
    numSectorTagLists = 0;
}
iterlist_t *P_GetSectorIterListForTag(int tag, dd_bool createNewList) {
    for(uint i = 0; i < numSectorTagLists; ++i) {
        if(sectorTagLists[i].tag == tag)
            return sectorTagLists[i].list;
    }
    if(!createNewList)
        return 0;
    numSectorTagLists++;
    sectorTagLists = (TagList *)realloc(sectorTagLists, sizeof(TagList) * numSectorTagLists);
    TagList *tagList = &sectorTagLists[numSectorTagLists - 1];
    tagList->tag = tag;
    return (tagList->list = IterList_New());
}
void P_BuildAllTagLists() {
    P_BuildSectorTagLists();
    P_BuildLineTagLists();
}
void P_DestroyAllTagLists() {
    P_DestroyLineTagLists();
    P_DestroySectorTagLists();
}
Sector *P_GetNextSector(Line *line, Sector *sec) {
    if(!sec || !line)
        return 0;
    if(!(P_ToXLine(line)->flags & ML_TWOSIDED))
        return 0;
    Sector *frontSec = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
    if(frontSec == sec)
        return (Sector *)P_GetPtrp(line, DMU_BACK_SECTOR);
    return frontSec;
}
int findExtremalLightLevelInAdjacentSectors(void *ptr, void *context) {
    findlightlevelparams_t *params = (findlightlevelparams_t *) context;
    Sector *other = P_GetNextSector((Line *) ptr, params->baseSec);
    if(!other)
        return false; // Continue iteration.
    float lightLevel = P_GetFloatp(other, DMU_LIGHT_LEVEL);
    if(params->flags & FELLF_MIN) {
        if(lightLevel < params->val) {
            params->val = lightLevel;
            params->foundSec = other;
            if(params->val <= 0)
                return true; // Stop iteration. Can't get any darker.
        }
    }
    else if(lightLevel > params->val) {
        params->val = lightLevel;
        params->foundSec = other;
        if(params->val >= 1)
            return true; // Stop iteration. Can't get any brighter.
    }
    return false; // Continue iteration.
}
Sector *P_FindSectorSurroundingLowestLight(Sector *sec, float *val) {
    findlightlevelparams_t params;
    params.flags = FELLF_MIN;
    params.val = DDMAXFLOAT;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalLightLevelInAdjacentSectors, &params);
    if(*val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingHighestLight(Sector *sec, float *val) {
    findlightlevelparams_t params;
    params.flags = 0;
    params.val = DDMINFLOAT;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalLightLevelInAdjacentSectors, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
int findNextLightLevel(void* ptr, void *context) {
    findnextlightlevelparams_t *params = (findnextlightlevelparams_t *) context;
    Line *li = (Line *) ptr;
    Sector *other = P_GetNextSector(li, params->baseSec);
    if(!other)
        return false; // Continue iteration.
    float otherLight = P_GetFloatp(other, DMU_LIGHT_LEVEL);
    if(params->flags & FNLLF_ABOVE) {
        if(otherLight < params->val && otherLight > params->baseLight) {
            params->val = otherLight;
            params->foundSec = other;
            if(!(params->val > 0))
                return true; // Stop iteration. Can't get any darker.
        }
    }
    else if(otherLight > params->val && otherLight < params->baseLight) {
        params->val = otherLight;
        params->foundSec = other;
        if(!(params->val < 1))
            return true; // Stop iteration. Can't get any brighter.
    }
    return false; // Continue iteration.
}
Sector *P_FindSectorSurroundingNextLowestLight(Sector *sec, float baseLight, float *val) {
    findnextlightlevelparams_t params;
    params.flags = 0;
    params.val = DDMINFLOAT;
    params.baseSec = sec;
    params.baseLight = baseLight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextLightLevel, &params);
    if(*val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingNextHighestLight(Sector *sec, float baseLight, float *val) {
    findnextlightlevelparams_t params;
    params.flags = FNLLF_ABOVE;
    params.val = DDMAXFLOAT;
    params.baseSec = sec;
    params.baseLight = baseLight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextLightLevel, &params);
    if(*val)
        *val = params.val;
    return params.foundSec;
}
int findExtremalPlaneHeight(void *ptr, void *context) {
    findextremalplaneheightparams_t *params = (findextremalplaneheightparams_t *) context;
    Sector *other = P_GetNextSector((Line *) ptr, params->baseSec);
    if(!other)
        return false; // Continue iteration.
    coord_t height = P_GetDoublep(other, ((params->flags & FEPHF_FLOOR)? DMU_FLOOR_HEIGHT : DMU_CEILING_HEIGHT));
    if(params->flags & FEPHF_MIN) {
        if(height < params->val) {
            params->val = height;
            params->foundSec = other;
        }
    }
    else if(height > params->val) {
        params->val = height;
        params->foundSec = other;
    }
    return false; // Continue iteration.
}
Sector *P_FindSectorSurroundingLowestFloor(Sector *sec, coord_t max, coord_t *val) {
    findextremalplaneheightparams_t params;
    params.flags = FEPHF_MIN | FEPHF_FLOOR;
    params.val = max;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingHighestFloor(Sector *sec, coord_t min, coord_t *val) {
    findextremalplaneheightparams_t params;
    params.flags = FEPHF_FLOOR;
    params.val = min;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingLowestCeiling(Sector *sec, coord_t max, coord_t *val) {
    findextremalplaneheightparams_t params;
    params.flags = FEPHF_MIN;
    params.val = max;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingHighestCeiling(Sector *sec, coord_t min, coord_t *val) {
    findextremalplaneheightparams_t params;
    params.flags = 0;
    params.val = min;
    params.baseSec = sec;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findExtremalPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
int findNextPlaneHeight(void *ptr, void *context) {
    findnextplaneheightparams_t *params = (findnextplaneheightparams_t *) context;
    Sector *other = P_GetNextSector((Line *) ptr, params->baseSec);
    if(!other)
        return false; // Continue iteration.
    coord_t otherHeight = P_GetDoublep(other, ((params->flags & FNPHF_FLOOR)? DMU_FLOOR_HEIGHT : DMU_CEILING_HEIGHT));
    if(params->flags & FNPHF_ABOVE) {
        if(otherHeight < params->val && otherHeight > params->baseHeight) {
            params->val = otherHeight;
            params->foundSec = other;
        }
    }
    else if(otherHeight > params->val && otherHeight < params->baseHeight) {
        params->val = otherHeight;
        params->foundSec = other;
    }
    return false; // Continue iteration.
}
Sector *P_FindSectorSurroundingNextHighestFloor(Sector *sec, coord_t baseHeight, coord_t *val) {
    findnextplaneheightparams_t params;
    params.flags = FNPHF_FLOOR | FNPHF_ABOVE;
    params.val = DDMAXFLOAT;
    params.baseSec = sec;
    params.baseHeight = baseHeight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingNextHighestCeiling(Sector *sec, coord_t baseHeight, coord_t *val) {
    findnextplaneheightparams_t params;
    params.flags = FNPHF_ABOVE;
    params.val = DDMAXFLOAT;
    params.baseSec = sec;
    params.baseHeight = baseHeight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingNextLowestFloor(Sector *sec, coord_t baseHeight, coord_t *val) {
    findnextplaneheightparams_t params;
    params.flags = FNPHF_FLOOR;
    params.val = DDMINFLOAT;
    params.baseSec = sec;
    params.baseHeight = baseHeight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
Sector *P_FindSectorSurroundingNextLowestCeiling(Sector *sec, coord_t baseHeight, coord_t *val) {
    findnextplaneheightparams_t params;
    params.flags = 0;
    params.val = DDMINFLOAT;
    params.baseSec = sec;
    params.baseHeight = baseHeight;
    params.foundSec = 0;
    P_Iteratep(sec, DMU_LINE, findNextPlaneHeight, &params);
    if(val)
        *val = params.val;
    return params.foundSec;
}
float P_SectorLight(Sector *sector) {
    return P_GetFloatp(sector, DMU_LIGHT_LEVEL);
}
void P_SectorSetLight(Sector *sector, float level) {
    P_SetFloatp(sector, DMU_LIGHT_LEVEL, level);
}
void P_SectorModifyLight(Sector *sector, float value) {
    float level = MINMAX_OF(0.f, P_SectorLight(sector) + value, 1.f);
    P_SectorSetLight(sector, level);
}
void P_SectorModifyLightx(Sector *sector, fixed_t value) {
    P_SetFloatp(sector, DMU_LIGHT_LEVEL, P_SectorLight(sector) + FIX2FLT(value) / 255.0f);
}
terraintype_t const *P_PlaneMaterialTerrainType(Sector *sec, int plane) {
    return P_TerrainTypeForMaterial((world_Material *)P_GetPtrp(sec, (plane? DMU_CEILING_MATERIAL : DMU_FLOOR_MATERIAL)));
}
void P_TranslateSideMaterialOrigin(Side *side, SideSection section, float deltaXY[2]) {
    DENG_ASSERT(side);
    DENG_ASSERT(VALID_SIDESECTION(section));
    uint const dmuSurfaceOriginFlags = DMU_OFFSET_XY | DMU_FLAG_FOR_SIDESECTION(section);
    float origin[2];
    if(IS_ZERO(deltaXY[0]) && IS_ZERO(deltaXY[1])) return;
    P_GetFloatpv(side, dmuSurfaceOriginFlags, origin);
    if(NON_ZERO(deltaXY[0])) {
        origin[0] += deltaXY[0];
    }
    if(NON_ZERO(deltaXY[1])) {
        origin[1] += deltaXY[1];
    }
    P_SetFloatpv(side, dmuSurfaceOriginFlags, origin);
}
void P_TranslateSideMaterialOriginXY(Side *side, SideSection section,
    float deltaX, float deltaY) {
    float delta[2];
    delta[0] = deltaX;
    delta[1] = deltaY;
    P_TranslateSideMaterialOrigin(side, section, delta);
}
void P_TranslatePlaneMaterialOrigin(Plane *plane, float deltaXY[2]) {
    float origin[2];
    DENG_ASSERT(plane);
    if(IS_ZERO(deltaXY[0]) && IS_ZERO(deltaXY[1])) return;
    P_GetFloatpv(plane, DMU_OFFSET_XY, origin);
    if(NON_ZERO(deltaXY[0])) {
        origin[0] += deltaXY[0];
    }
    if(NON_ZERO(deltaXY[1])) {
        origin[1] += deltaXY[1];
    }
    P_SetFloatpv(plane, DMU_OFFSET_XY, origin);
}
void P_TranslatePlaneMaterialOriginXY(Plane *plane, float deltaX, float deltaY) {
    float delta[2];
    delta[0] = deltaX;
    delta[1] = deltaY;
    P_TranslatePlaneMaterialOrigin(plane, delta);
}
namespace internal {
    static bool useMaterialArchiveSegments() {
#if __JHEXEN__
        return true;
#else
        return false;
#endif
    }
    static int thingArchiveVersionFor(int mapVersion) {
#if __JHEXEN__
        return mapVersion >= 4? 1 : 0;
#else
        return 0;
        DENG2_UNUSED(mapVersion);
#endif
    }
}
using namespace de;
DENG2_PIMPL(MapStateReader) {
    reader_s *reader;
    int saveVersion;
    int mapVersion;
    bool formatHasMapVersionNumber;
    dd_bool loaded[MAXPLAYERS];
    dd_bool infile[MAXPLAYERS];
    int thingArchiveSize;
    ThingArchive *thingArchive;
    world::MaterialArchive *materialArchive;
    dmu_lib::SideArchive *sideArchive;
    QHash<Id::Type, thinker_t *> archivedThinkerIds;
    Impl(Public *i)
        : Base(i)
        , reader(0)
        , saveVersion(0)
        , mapVersion(0)
        , formatHasMapVersionNumber(false)
        , thingArchiveSize(0)
        , thingArchive(0)
        , materialArchive(0)
        , sideArchive(0) {
        de::zap(loaded);
        de::zap(infile);
    }
    ~Impl() {
        delete thingArchive;
        delete sideArchive;
        delete materialArchive;
        Reader_Delete(reader);
    }
    void beginSegment(int segId) {
#if __JHEXEN__
        if (segId == ASEG_END && SV_RawReader().source()->size() - SV_RawReader().offset() < 4) {
            App_Log(DE2_LOG_WARNING, "Savegame lacks ASEG_END marker (unexpected end-of-file)");
            return;
        }
        if (Reader_ReadInt32(reader) != segId) {
            throw ReadError("MapStateReader", "Corrupt save game, segment #" + String::number(segId) + " failed alignment check");
        }
#else
        DENG2_UNUSED(segId);
#endif
    }
    void beginMapSegment() {
#if __JHEXEN__
        int segId = Reader_ReadInt32(reader);
        if (segId != ASEG_MAP_HEADER2 && segId != ASEG_MAP_HEADER) {
            throw ReadError("MapStateReader", "Corrupt save game, segment #" + String::number(segId) + " failed alignment check");
        }
        formatHasMapVersionNumber = (segId == ASEG_MAP_HEADER2);
#else
        beginSegment(ASEG_MAP_HEADER2);
#endif
    }
    void endSegment() {
        beginSegment(ASEG_END);
    }
    void readMapHeader() {
#if __JHEXEN__
        mapVersion = (formatHasMapVersionNumber? Reader_ReadByte(reader) : 2);
        mapTime = Reader_ReadInt32(reader);
#endif
    }
    void readConsistencyBytes() {
#if !__JHEXEN__
        if (Reader_ReadByte(reader) != CONSISTENCY) {
            throw ReadError("MapStateReader", "Corrupt save game, failed consistency check");
        }
#endif
    }
    void readMaterialArchive() {
        materialArchive = new world::MaterialArchive(::internal::useMaterialArchiveSegments(), false /*empty*/);
#if !__JHEXEN__
        if (mapVersion >= 4)
#endif
        {
            materialArchive->read(*reader, mapVersion < 6? 0 : -1);
        }
    }
    void readPlayers() {
#if __JHEXEN__
        if (saveVersion >= 4)
#else
        if (saveVersion >= 5)
#endif
        {
            beginSegment(ASEG_PLAYER_HEADER);
        }
        playerheader_t plrHdr;
        plrHdr.read(reader, saveVersion);
        ddplayer_t dummyDDPlayer;
        player_t dummyPlayer;
        dummyPlayer.plr = &dummyDDPlayer;
#if !__JHEXEN__
        ArrayValue const &presentPlayers = self().metadata().geta("players");
#endif
        for (int i = 0; i < MAXPLAYERS; ++i) {
            loaded[i] = 0;
#if !__JHEXEN__
            infile[i] = presentPlayers.at(i).isTrue();;
#endif
        }
        beginSegment(ASEG_PLAYERS); {
#if __JHEXEN__
            for (int i = 0; i < MAXPLAYERS; ++i) {
                infile[i] = Reader_ReadByte(reader);
            }
#endif
            for (int i = 0; i < MAXPLAYERS; ++i) {
                saveToRealPlayerNum[i] = -1;
                if (!infile[i]) continue;
                int pid = Reader_ReadInt32(reader);
                player_t *player = 0;
                for (int k = 0; k < MAXPLAYERS; ++k) {
                    if ((IS_NETGAME && int(Net_GetPlayerID(k)) == pid) ||
                       (!IS_NETGAME && k == 0)) {
                        player = players + k;
                        loaded[k] = true;
                        saveToRealPlayerNum[i] = k;
                        App_Log(DE2_DEV_MAP_MSG, "readPlayers: saved %i is now %i", i, k);
                        break;
                    }
                }
                if (!player) {
                    player = &dummyPlayer;
                }
                player->read(reader, plrHdr);
            }
        }
        endSegment();
    }
    void kickMissingPlayers() {
        for (int i = 0; i < MAXPLAYERS; ++i) {
            bool notLoaded = false;
#if __JHEXEN__
            if (players[i].plr->inGame) {
                int k;
                for (k = 0; k < MAXPLAYERS; ++k) {
                    if (saveToRealPlayerNum[k] == i) {
                        break;
                    }
                }
                if (k < MAXPLAYERS)
                    continue; // Found; don't bother this player.
                players[i].playerState = PST_REBORN;
                if (!i) {
                    P_SetMessageWithFlags(players, GET_TXT(TXT_LOADMISSING), LMF_NO_HIDE);
                }
                else {
                    NetSv_SendMessage(i, GET_TXT(TXT_LOADMISSING));
                    notLoaded = true;
                }
            }
#else
            if (!loaded[i] && players[i].plr->inGame) {
                if (!i) {
                    P_SetMessageWithFlags(players, GET_TXT(TXT_LOADMISSING), LMF_NO_HIDE);
                }
                else {
                    NetSv_SendMessage(i, GET_TXT(TXT_LOADMISSING));
                }
                notLoaded = true;
            }
#endif
            if (notLoaded) {
                DD_Executef(false, "kick %i", i);
            }
        }
    }
    void readElements() {
        beginSegment(ASEG_MAP_ELEMENTS);
        for (int i = 0; i < numsectors; ++i) {
            SV_ReadSector((Sector *)P_ToPtr(DMU_SECTOR, i), thisPublic);
        }
        for (int i = 0; i < numlines; ++i) {
            SV_ReadLine((Line *)P_ToPtr(DMU_LINE, i), thisPublic);
        }
    }
    void readPolyobjs() {
#if __JHEXEN__
        beginSegment(ASEG_POLYOBJS);
        int const writtenPolyobjCount = Reader_ReadInt32(reader);
        DENG2_ASSERT(writtenPolyobjCount == numpolyobjs);
        for (int i = 0; i < writtenPolyobjCount; ++i) {
            /*Skip unused version byte*/ if (mapVersion >= 3) Reader_ReadByte(reader);
            Polyobj *po = Polyobj_ByTag(Reader_ReadInt32(reader));
            DENG2_ASSERT(po != 0);
            po->read(thisPublic);
        }
#endif
    }
    static int removeLoadSpawnedThinkerWorker(thinker_t *th, void * /*context*/)
    {
        if (th->function == (thinkfunc_t) P_MobjThinker) {
            P_MobjRemove((mobj_t *) th, true);
        }
        else {
            Z_Free(th);
        }
        return false; // Continue iteration.
    }
    void removeLoadSpawnedThinkers() {
#if !__JHEXEN__
        if (!IS_SERVER) return; // Not for us.
#endif
        Thinker_Iterate(0 /*all thinkers*/, removeLoadSpawnedThinkerWorker, 0/*no params*/);
        Thinker_Init();
    }
#if __JHEXEN__
    static bool mobjtypeHasCorpse(mobjtype_t type) {
        switch (type) {
        case MT_CENTAUR:
        case MT_CENTAURLEADER:
        case MT_DEMON:
        case MT_DEMON2:
        case MT_WRAITH:
        case MT_WRAITHB:
        case MT_BISHOP:
        case MT_ETTIN:
        case MT_PIG:
        case MT_CENTAUR_SHIELD:
        case MT_CENTAUR_SWORD:
        case MT_DEMONCHUNK1:
        case MT_DEMONCHUNK2:
        case MT_DEMONCHUNK3:
        case MT_DEMONCHUNK4:
        case MT_DEMONCHUNK5:
        case MT_DEMON2CHUNK1:
        case MT_DEMON2CHUNK2:
        case MT_DEMON2CHUNK3:
        case MT_DEMON2CHUNK4:
        case MT_DEMON2CHUNK5:
        case MT_FIREDEMON_SPLOTCH1:
        case MT_FIREDEMON_SPLOTCH2:
            return true;
        default: return false;
        }
    }
    static int rebuildCorpseQueueWorker(thinker_t *th, void * /*context*/)
    {
        mobj_t *mo = (mobj_t *) th;
        if ((mo->flags & MF_CORPSE) && !(mo->flags & MF_ICECORPSE) &&
           mobjtypeHasCorpse(mobjtype_t(mo->type))) {
            P_AddCorpseToQueue(mo);
        }
        return false; // Continue iteration.
    }
    /**
     * @todo fixme: the corpse queue should be serialized (original order unknown).
     */
    void rebuildCorpseQueue() {
        P_InitCorpseQueue();
        Thinker_Iterate((thinkfunc_t) P_MobjThinker, rebuildCorpseQueueWorker, NULL/*no params*/);
    }
#endif
    static int restoreMobjLinksWorker(thinker_t *th, void *context) {
        MapStateReader *inst = static_cast<MapStateReader *>(context);
        if (th->function != (thinkfunc_t) P_MobjThinker)
            return false; // Continue iteration.
        mobj_t *mo = (mobj_t *) th;
        mo->target = inst->mobj(PTR2INT(mo->target), &mo->target);
        mo->onMobj = inst->mobj(PTR2INT(mo->onMobj), &mo->onMobj);
#if __JHEXEN__
        switch (mo->type) {
        case MT_BISH_FX:
        case MT_HOLY_FX:
        case MT_DRAGON:
        case MT_THRUSTFLOOR_UP:
        case MT_THRUSTFLOOR_DOWN:
        case MT_MINOTAUR:
        case MT_SORCFX1:
            if (inst->mapVersion() >= 3) {
                mo->tracer = inst->mobj(PTR2INT(mo->tracer), &mo->tracer);
            }
            else {
                mo->tracer = inst->mobj(mo->special1, &mo->tracer);
                mo->special1 = 0;
            }
            break;
        case MT_LIGHTNING_FLOOR:
        case MT_LIGHTNING_ZAP:
            mo->special2 = PTR2INT(inst->mobj(mo->special2, &mo->special2));
            break;
        case MT_HOLY_TAIL:
        case MT_LIGHTNING_CEILING:
            if (inst->mapVersion() >= 3) {
                mo->tracer = inst->mobj(PTR2INT(mo->tracer), &mo->tracer);
            }
            else {
                mo->tracer = inst->mobj(mo->special1, &mo->tracer);
                mo->special1 = 0;
            }
            mo->special2 = PTR2INT(inst->mobj(mo->special2, &mo->special2));
            break;
        default:
            break;
        }
#else
# if __JDOOM__ || __JDOOM64__
        mo->tracer = inst->mobj(PTR2INT(mo->tracer), &mo->tracer);
# endif
# if __JHERETIC__
        mo->generator = inst->mobj(PTR2INT(mo->generator), &mo->generator);
# endif
#endif
        return false; // Continue iteration.
    }
    void readThinkers() {
#if !__JHEXEN__
#define PRE_VER5_END_SPECIALS  7
#endif
        bool const formatHasStasisInfo = (mapVersion >= 6);
#if __JHEXEN__
        if (mapVersion < 4)
            beginSegment(ASEG_MOBJS);
        else
#endif
            beginSegment(ASEG_THINKERS);
#if __JHEXEN__
        SV_InitTargetPlayers();
        thingArchive->initForLoad(Reader_ReadInt32(reader) /* num elements */);
#endif
#if __JHEXEN__
        int i = 0;
        bool reachedSpecialsBlock = (mapVersion >= 4);
#else
        bool reachedSpecialsBlock = (mapVersion >= 5);
#endif
        byte tClass = 0;
        forever {
#if __JHEXEN__
            if (reachedSpecialsBlock)
#endif
            {
                tClass = Reader_ReadByte(reader);
            }
#if __JHEXEN__
            if (mapVersion < 4) {
                if (reachedSpecialsBlock) // Have we started on the specials yet?
                {
                    if (tClass != TC_END) {
                        tClass += 2;
                    }
                }
                else {
                    tClass = TC_MOBJ;
                }
                if (tClass == TC_MOBJ && (uint)i == thingArchive->size()) {
                    beginSegment(ASEG_THINKERS);
                    reachedSpecialsBlock = true;
                    continue;
                }
            }
#else
            if (mapVersion < 5) {
                if (reachedSpecialsBlock) {
                    if (tClass == PRE_VER5_END_SPECIALS) {
                        tClass = TC_END;
                    }
                    else {
                        tClass += 3;
                    }
                }
                else if (tClass == TC_END) {
                    reachedSpecialsBlock = true;
                    continue;
                }
            }
#endif
            if (tClass == TC_END)
                break; // End of the list.
            ThinkerClassInfo *thInfo = SV_ThinkerInfoForClass(thinkerclass_t(tClass));
            DENG2_ASSERT(thInfo != 0);
            DENG2_ASSERT(!((thInfo->flags & TSF_SERVERONLY) && IS_CLIENT));
            thinker_t *th = 0;
            if (thInfo->thinkclass == TC_MOBJ) {
                th = reinterpret_cast<thinker_t *>(Mobj_CreateXYZ((thinkfunc_t) P_MobjThinker, 0, 0, 0, 0, 64, 64, 0));
            }
            else {
                th = Thinker(Thinker::AllocateMemoryZone, thInfo->size).take();
            }
            bool putThinkerInStasis = (formatHasStasisInfo? CPP_BOOL(Reader_ReadByte(reader)) : false);
            if (saveVersion >= 15) {
                Id::Type const privateId = Reader_ReadUInt32(reader);
                archivedThinkerIds.insert(privateId, th);
            }
            if (thInfo->readFunc(th, thisPublic)) {
                Thinker_Add(th);
            }
            if (putThinkerInStasis) {
                Thinker_SetStasis(th, true);
            }
#if __JHEXEN__
            if (tClass == TC_MOBJ) {
                i++;
            }
#endif
        }
#if __JHEXEN__
        Thinker_Iterate((thinkfunc_t) P_MobjThinker, restoreMobjLinksWorker, thisPublic);
        P_CreateTIDList();
        rebuildCorpseQueue();
#else
        if (IS_SERVER) {
            Thinker_Iterate((thinkfunc_t) P_MobjThinker, restoreMobjLinksWorker, thisPublic);
            for (int i = 0; i < numlines; ++i) {
                xline_t *xline = P_ToXLine((Line *)P_ToPtr(DMU_LINE, i));
                if (!xline->xg) continue;
                xline->xg->activator = thingArchive->mobj(PTR2INT(xline->xg->activator),
                                                          &xline->xg->activator);
            }
        }
#endif
#undef PRE_VER5_END_SPECIALS
    }
    void readACScriptData() {
#if __JHEXEN__
        beginSegment(ASEG_SCRIPTS);
        gfw_Session()->acsSystem().readMapState(thisPublic);
#endif
    }
    void readSoundSequences() {
#if __JHEXEN__
        beginSegment(ASEG_SOUNDS);
        SN_ReadSequences(reader, mapVersion);
#endif
    }
    void readMisc() {
#if __JHEXEN__
        beginSegment(ASEG_MISC);
        for (int i = 0; i < MAXPLAYERS; ++i) {
            localQuakeHappening[i] = Reader_ReadInt32(reader);
        }
#endif
#if __JDOOM__
        DENG2_ASSERT(theBossBrain != 0);
        theBossBrain->read(thisPublic);
#endif
    }
    void readSoundTargets() {
#if !__JHEXEN__
        if (!IS_SERVER) return; // Not for us.
        if (mapVersion < 5) return;
        int numTargets = Reader_ReadInt32(reader);
        for (int i = 0; i < numTargets; ++i) {
            xsector_t *xsec = P_ToXSector((Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader)));
            DENG2_ASSERT(xsec != 0);
            if (!xsec) {
                DENG2_UNUSED(Reader_ReadInt16(reader));
                continue;
            }
            xsec->soundTarget = INT2PTR(mobj_t, Reader_ReadInt16(reader));
            xsec->soundTarget = thingArchive->mobj(PTR2INT(xsec->soundTarget), &xsec->soundTarget);
        }
#endif
    }
};
MapStateReader::MapStateReader(GameStateFolder const &session)
    : GameStateFolder::MapStateReader(session)
    , d(new Impl(this))
{}
MapStateReader::~MapStateReader()
{}
void MapStateReader::read(String const &mapUriStr) {
    de::Uri const mapUri(mapUriStr, RC_NULL);
    File const &mapStateFile = folder().locate<File const>(String("maps") / mapUri.path() + "State");
    SV_OpenFileForRead(mapStateFile);
    d->reader = SV_NewReader();
    /*magic*/ Reader_ReadInt32(d->reader);
    d->saveVersion = Reader_ReadInt32(d->reader);
    d->mapVersion  = d->saveVersion; // Default: mapVersion == saveVersion
    d->thingArchiveSize = 0;
#if !__JHEXEN__
    d->thingArchiveSize = (d->saveVersion >= 5? Reader_ReadInt32(d->reader) : 1024);
#endif
    d->readPlayers();
    d->sideArchive = new dmu_lib::SideArchive;
    d->beginMapSegment(); {
        d->readMapHeader();
        d->readMaterialArchive();
        d->thingArchive = new ThingArchive(::internal::thingArchiveVersionFor(d->mapVersion));
#if !__JHEXEN__
        d->thingArchive->initForLoad(d->thingArchiveSize);
#endif
        d->removeLoadSpawnedThinkers();
        d->readElements();
        d->readPolyobjs();
        d->readThinkers();
        d->readACScriptData();
        d->readSoundSequences();
        d->readMisc();
        d->readSoundTargets();
    }
    d->endSegment();
    delete d->sideArchive;      d->sideArchive = 0;
    delete d->materialArchive;  d->materialArchive = 0;
    d->readConsistencyBytes();
    Reader_Delete(d->reader); d->reader = 0;
#if __JHEXEN__
    SV_ClearTargetPlayers();
#endif
    SV_CloseFile();
    d->kickMissingPlayers();
    NetSv_LoadGame(metadata().geti("sessionId"));
    if (d->saveVersion <= 10) {
        P_SpawnAllMaterialOriginScrollers();
    }
    for (int i = 0; i < MAXPLAYERS; ++i) {
        R_UpdateConsoleView(i);
    }
    R_SetupMap(0, 0);
}
mobj_t *MapStateReader::mobj(ThingArchive::SerialId serialId, void *address) const {
    DENG2_ASSERT(d->thingArchive != 0);
    return d->thingArchive->mobj(serialId, address);
}
world_Material *MapStateReader::material(materialarchive_serialid_t serialId, int group) const {
    DENG2_ASSERT(d->materialArchive != 0);
    return reinterpret_cast<world_Material *>(d->materialArchive->find(serialId, group));
}
Side *MapStateReader::side(int sideIndex) const {
    DENG2_ASSERT(d->sideArchive != 0);
    return reinterpret_cast<Side *>(d->sideArchive->at(sideIndex));
}
thinker_t *MapStateReader::thinkerForPrivateId(Id::Type id) const {
    auto found = d->archivedThinkerIds.constFind(id);
    if (found != d->archivedThinkerIds.constEnd()) {
        return found.value();
    }
    return nullptr;
}
player_t *MapStateReader::player(int serialId) const {
    DENG2_ASSERT(serialId > 0 && serialId <= MAXPLAYERS);
    return players + saveToRealPlayerNum[serialId - 1];
}
int MapStateReader::mapVersion() {
    return d->mapVersion;
}
reader_s *MapStateReader::reader() {
    DENG2_ASSERT(d->reader != 0);
    return d->reader;
}
void MapStateReader::addMobjToThingArchive(mobj_t *mobj, ThingArchive::SerialId serialId) {
    DENG2_ASSERT(d->thingArchive != 0);
    d->thingArchive->insert(mobj, serialId);
}
namespace internal {
    static bool MapStateWriter_useMaterialArchiveSegments() {
#if __JHEXEN__
        return true;
#else
        return false;
#endif
    }
}
DENG2_PIMPL(MapStateWriter) {
    ThingArchive *thingArchive;
    world::MaterialArchive *materialArchive;
    Writer1 *writer; // Not owned.
    Impl(Public *i)
        : Base(i)
        , thingArchive(0)
        , materialArchive(0)
        , writer(0)
    {}
    ~Impl() {
        delete materialArchive;
        delete thingArchive;
    }
    void beginSegment(int segId) {
#if __JHEXEN__
        Writer_WriteInt32(writer, segId);
#else
        DENG2_UNUSED(segId);
#endif
    }
    void endSegment() {
        beginSegment(ASEG_END);
    }
    void writeConsistencyBytes() {
#if !__JHEXEN__
        Writer_WriteByte(writer, CONSISTENCY);
#endif
    }
    void writeMapHeader() {
#if __JHEXEN__
        Writer_WriteByte(writer, MY_SAVE_VERSION);
        Writer_WriteInt32(writer, mapTime);
#endif
    }
    void writeMaterialArchive() {
        materialArchive->write(*writer);
    }
    void writePlayers() {
        beginSegment(ASEG_PLAYER_HEADER);
        playerheader_t plrHdr;
        plrHdr.write(writer);
        beginSegment(ASEG_PLAYERS); {
#if __JHEXEN__
            for (int i = 0; i < MAXPLAYERS; ++i) {
                Writer_WriteByte(writer, players[i].plr->inGame);
            }
#endif
            for (int i = 0; i < MAXPLAYERS; ++i) {
                player_t *plr = players + i;
                if (!plr->plr->inGame)
                    continue;
                Writer_WriteInt32(writer, Net_GetPlayerID(i));
                plr->write(writer, plrHdr);
            }
        }
        endSegment();
    }
    void writeElements() {
        beginSegment(ASEG_MAP_ELEMENTS);
        for (int i = 0; i < numsectors; ++i) {
            SV_WriteSector((Sector *)P_ToPtr(DMU_SECTOR, i), thisPublic);
        }
        for (int i = 0; i < numlines; ++i) {
            SV_WriteLine((Line *)P_ToPtr(DMU_LINE, i), thisPublic);
        }
    }
    void writePolyobjs() {
#if __JHEXEN__
        beginSegment(ASEG_POLYOBJS);
        Writer_WriteInt32(writer, numpolyobjs);
        for (int i = 0; i < numpolyobjs; ++i) {
            Polyobj *po = Polyobj_ById(i);
            DENG2_ASSERT(po != 0);
            po->write(thisPublic);
        }
#endif
    }
    struct writethinkerworker_params_t {
        MapStateWriter *msw;
        bool excludePlayers;
    };
    /**
     * Serializes the specified thinker and writes it to save state.
     */
    static int writeThinkerWorker(thinker_t *th, void *context) {
        writethinkerworker_params_t const &p = *static_cast<writethinkerworker_params_t *>(context);
        ThinkerClassInfo *thInfo = SV_ThinkerInfo(*th);
        if (!thInfo) return false;
        if (p.excludePlayers) {
            if (th->function == P_MobjThinker && ((mobj_t *) th)->player) {
                return false;
            }
        }
        if ((thInfo->flags & TSF_SERVERONLY) && IS_CLIENT) {
            return false;
        }
        Writer_WriteByte(p.msw->writer(), thInfo->thinkclass); // Thinker type byte.
        Writer_WriteByte(p.msw->writer(), Thinker_InStasis(th)? 1 : 0); // In stasis?
        de::Id::Type const privateId = (th->d? THINKER_DATA(*th, ThinkerData).id().asUInt32() : 0);
        Writer_WriteUInt32(p.msw->writer(), privateId);
        thInfo->writeFunc(th, p.msw);
        return false;
    }
    /**
     * Serializes thinkers for both client and server.
     *
     * @note Clients do not save data for all thinkers. In some cases the server will send it
     * anyway (so saving it would just bloat client save states).
     *
     * @note Some thinker classes are NEVER saved by clients.
     */
    void writeThinkers() {
        beginSegment(ASEG_THINKERS);
#if __JHEXEN__
        Writer_WriteInt32(writer, thingArchive->size()); // number of mobjs.
#endif
        writethinkerworker_params_t parm; de::zap(parm);
        parm.msw            = thisPublic;
        parm.excludePlayers = thingArchive->excludePlayers();
        Thinker_Iterate(0/*all thinkers*/, writeThinkerWorker, &parm);
        Writer_WriteByte(writer, TC_END);
    }
    void writeACScriptData() {
#if __JHEXEN__
        beginSegment(ASEG_SCRIPTS);
        gfw_Session()->acsSystem().writeMapState(thisPublic);
#endif
    }
    void writeSoundSequences() {
#if __JHEXEN__
        beginSegment(ASEG_SOUNDS);
        SN_WriteSequences(writer);
#endif
    }
    void writeMisc() {
#if __JHEXEN__
        beginSegment(ASEG_MISC);
        for (int i = 0; i < MAXPLAYERS; ++i) {
            Writer_WriteInt32(writer, localQuakeHappening[i]);
        }
#endif
#if __JDOOM__
        DENG2_ASSERT(theBossBrain != 0);
        theBossBrain->write(thisPublic);
#endif
    }
    void writeSoundTargets() {
#if !__JHEXEN__
        if (!IS_SERVER) return; // Not for us.
        int count = 0;
        for (int i = 0; i < numsectors; ++i) {
            xsector_t *xsec = P_ToXSector((Sector *)P_ToPtr(DMU_SECTOR, i));
            if (xsec->soundTarget) {
                count += 1;
            }
        }
        Writer_WriteInt32(writer, count);
        for (int i = 0; i < numsectors; ++i) {
            xsector_t *xsec = P_ToXSector((Sector *)P_ToPtr(DMU_SECTOR, i));
            if (xsec->soundTarget) {
                Writer_WriteInt32(writer, i);
                Writer_WriteInt16(writer, thingArchive->serialIdFor(xsec->soundTarget));
            }
        }
#endif
    }
};
MapStateWriter::MapStateWriter() : d(new Impl(this))
{}
void MapStateWriter::write(Writer1 *writer, bool excludePlayers) {
    DENG2_ASSERT(writer != 0);
    d->writer = writer;
    d->materialArchive = new world::MaterialArchive(::internal::MapStateWriter_useMaterialArchiveSegments());
    d->materialArchive->addWorldMaterials();
    Writer_WriteInt32(writer, MY_SAVE_MAGIC);
    Writer_WriteInt32(writer, MY_SAVE_VERSION);
    d->thingArchive = new ThingArchive;
    d->thingArchive->initForSave(excludePlayers);
#if !__JHEXEN__
    Writer_WriteInt32(d->writer, d->thingArchive->size());
#endif
    d->writePlayers();
    d->beginSegment(ASEG_MAP_HEADER2); {
        d->writeMapHeader();
        d->writeMaterialArchive();
        d->writeElements();
        d->writePolyobjs();
        d->writeThinkers();
        d->writeACScriptData();
        d->writeSoundSequences();
        d->writeMisc();
        d->writeSoundTargets();
    }
    d->endSegment();
    d->writeConsistencyBytes(); // To be absolutely sure...
    delete d->materialArchive;  d->materialArchive = 0;
}
ThingArchive::SerialId MapStateWriter::serialIdFor(mobj_t const *mobj) {
    DENG2_ASSERT(d->thingArchive != 0);
    return d->thingArchive->serialIdFor(mobj);
}
materialarchive_serialid_t MapStateWriter::serialIdFor(world::Material *material) {
    DENG2_ASSERT(d->materialArchive != 0);
    return d->materialArchive->findUniqueSerialId(material);
}
materialarchive_serialid_t MapStateWriter::serialIdFor(material_s *material) {
    return serialIdFor(reinterpret_cast<world::Material *>(material));
}
Writer1 *MapStateWriter::writer() {
    DENG2_ASSERT(d->writer != 0);
    return d->writer;
}
coord_t Mobj_ThrustMulForFriction(coord_t friction) {
    if(friction <= FRICTION_NORMAL)
        return 1; // Normal friction.
    if(friction > 1)
        return 0; // There's nothing to thrust from!
    return (-114.7338958 * friction * friction + 208.0448223 * friction - 93.31092643);
}
coord_t Mobj_ThrustMul(mobj_t const *mo) {
    coord_t mul = 1.0;
#if __JHEXEN__
    if(P_MobjFloorTerrain(mo)->flags & TTF_FRICTION_LOW) {
        mul /= 2;
    }
#else // !__JHEXEN__
    Sector *sec = Mobj_Sector(mo);
#if __JHERETIC__
    if(P_ToXSector(sec)->special == 15) // Friction_Low
    {
        mul /= 4;
        return mul; // XG friction ignored.
    }
#endif
    mul = Mobj_ThrustMulForFriction(XS_Friction(sec));
#endif
    return mul;
}
dd_bool Mobj_IsAirborne(mobj_t const *mo) {
    return ((mo->flags2 & MF2_FLY) && !(mo->origin[VZ] <= mo->floorZ) && !mo->onMobj) != 0;
}
coord_t Mobj_Friction(mobj_t const *mo) {
    if(Mobj_IsAirborne(mo)) {
        return FRICTION_FLY;
    }
#if __JHERETIC__
    if(P_ToXSector(Mobj_Sector(mo))->special == 15) // Low friction.
    {
        return FRICTION_LOW;
    }
#endif
#if __JHEXEN__
    terraintype_t const *tt = P_MobjFloorTerrain(mo);
    if(tt->flags & TTF_FRICTION_LOW) {
        return FRICTION_LOW;
    }
#endif
#if LIBCOMMON_HAVE_XG
    return XS_Friction(Mobj_Sector(mo));
#else
    return FRICTION_NORMAL;
#endif
}
dd_bool Mobj_IsVoodooDoll(mobj_t const *mo) {
    if(!mo) return false;
    return (mo->player && mo->player->plr->mo != mo);
}
void Mobj_XYMoveStopping(mobj_t *mo) {
    DENG_ASSERT(mo != 0);
    player_t *player = mo->player;
    if(player && (P_GetPlayerCheats(player) & CF_NOMOMENTUM)) {
        mo->mom[MX] = mo->mom[MY] = 0;
        return;
    }
    if(mo->flags & (MF_MISSILE | MF_SKULLFLY)) {
        return;
    }
    if (mo->origin[VZ] > mo->floorZ && !mo->onMobj && !(mo->flags2 & MF2_FLY)) {
        return;
    }
#ifndef __JHEXEN__
    if(cfg.slidingCorpses) {
        if(((mo->flags & MF_CORPSE) || (mo->intFlags & MIF_FALLING)) && !mo->player) {
            if(!INRANGE_OF(mo->mom[MX], 0, DROPOFFMOMENTUM_THRESHOLD) ||
               !INRANGE_OF(mo->mom[MY], 0, DROPOFFMOMENTUM_THRESHOLD)) {
                if(!FEQUAL(mo->floorZ, P_GetDoublep(Mobj_Sector(mo), DMU_FLOOR_HEIGHT)))
                    return;
            }
        }
    }
#endif
    bool isVoodooDoll = Mobj_IsVoodooDoll(mo);
    bool belowWalkStop = (INRANGE_OF(mo->mom[MX], 0, WALKSTOP_THRESHOLD) &&
                          INRANGE_OF(mo->mom[MY], 0, WALKSTOP_THRESHOLD));
    bool belowStandSpeed = false;
    bool isMovingPlayer  = false;
    if(player) {
        belowStandSpeed = (INRANGE_OF(mo->mom[MX], 0, STANDSPEED) &&
                           INRANGE_OF(mo->mom[MY], 0, STANDSPEED));
        isMovingPlayer = (NON_ZERO(player->plr->forwardMove) ||
                        NON_ZERO(player->plr->sideMove));
    }
    if(!isVoodooDoll && player && belowStandSpeed && !isMovingPlayer &&
        !IS_NETWORK_SERVER) // Netgame servers use logic elsewhere for player animation.
    {
        if(P_PlayerInWalkState(player)) {
            P_MobjChangeState(player->plr->mo, statenum_t(PCLASS_INFO(player->class_)->normalState));
        }
    }
    if(belowWalkStop && !isMovingPlayer) {
        if(!isVoodooDoll) {
            mo->mom[MX] = mo->mom[MY] = 0;
            if(player) player->bob = 0;
        }
    }
    else {
        coord_t friction = Mobj_Friction(mo);
        mo->mom[MX] *= friction;
        mo->mom[MY] *= friction;
    }
}
dd_bool Mobj_IsPlayerClMobj(mobj_t *mo) {
    if(IS_CLIENT) {
        for(int i = 0; i < MAXPLAYERS; i++) {
            if(ClPlayer_ClMobj(i) == mo) {
                return true;
            }
        }
    }
    return false;
}
uint32_t Mobj_PrivateID(mobj_t const *mob) {
    if (!mob) {
        return 0;
    }
    if (auto const *td = THINKER_DATA_MAYBE(mob->thinker, ThinkerData)) {
        return td->id();
    }
    return 0;
}
mobj_t *Mobj_FindByPrivateID(uint32_t privateId) {
    if (privateId == 0) {
        return nullptr;
    }
    if (auto *thinkerData = ThinkerData::find(de::Id(privateId))) {
        if (auto *mobjData = de::maybeAs<MobjThinkerData>(thinkerData)) {
            return mobjData->mobj();
        }
    }
    return nullptr;
}
dd_bool Mobj_IsPlayer(mobj_t const *mob) {
    if(!mob) return false;
    return (mob->player != 0);
}
angle_t Mobj_AimAtPoint2(mobj_t *mob, coord_t const point[], dd_bool pointShadowed) {
    DENG2_ASSERT(mob);
    return P_AimAtPoint2(mob->origin, point, pointShadowed);
}
angle_t Mobj_AimAtPoint(mobj_t *mob, coord_t const point[]) {
    return Mobj_AimAtPoint2(mob, point, false/* not shadowed*/);
}
angle_t Mobj_AimAtTarget(mobj_t *mob) {
    DENG2_ASSERT(mob);
    if(auto const *target = mob->target) {
        return Mobj_AimAtPoint2(mob, target->origin, target->flags & MF_SHADOW);
    }
    return mob->angle;
}
dd_bool Mobj_LookForPlayers(mobj_t *mo, dd_bool allAround) {
    int const playerCount = P_CountPlayersInGame();
    if(!playerCount) return false;
    int const from = mo->lastLook % MAXPLAYERS;
    int const to   = (from + MAXPLAYERS - 1) % MAXPLAYERS;
    int cand         = from;
    int tries        = 0;
    bool foundTarget = false;
    for(; cand != to; cand = (cand < (MAXPLAYERS - 1)? cand + 1 : 0)) {
        player_t *player = players + cand;
        if(!player->plr->inGame) continue;
        mobj_t *plrmo = player->plr->mo;
        if(!plrmo) continue;
        if(P_MobjIsCamera(plrmo)) continue;
        if(tries++ == 2) break;
        if(player->health <= 0) continue;
        if(!P_CheckSight(mo, plrmo)) continue;
        if(!allAround) {
            angle_t an = M_PointToAngle2(mo->origin, plrmo->origin);
            an -= mo->angle;
            if(an > ANG90 && an < ANG270) {
                coord_t dist = M_ApproxDistance(plrmo->origin[VX] - mo->origin[VX],
                                                plrmo->origin[VY] - mo->origin[VY]);
                if(dist > MELEERANGE) continue;
            }
        }
#if __JHERETIC__ || __JHEXEN__
        if(plrmo->flags & MF_SHADOW) {
            if((M_ApproxDistance(plrmo->origin[VX] - mo->origin[VX],
                                 plrmo->origin[VY] - mo->origin[VY]) > 2 * MELEERANGE) &&
               M_ApproxDistance(plrmo->mom[MX], plrmo->mom[MY]) < 5) {
                continue;
            }
            if(P_Random() < 225) continue;
        }
#endif
#if __JHEXEN__
        if(mo->type == MT_MINOTAUR && mo->tracer &&
           mo->tracer->player == player) continue;
#endif
        mo->target = plrmo;
        foundTarget = true;
    }
    mo->lastLook = cand;
    return foundTarget;
}
/**
 * Determines if it is allowed to execute the action function of @a mo.
 * @return @c true, if allowed.
 */
static bool shouldCallAction(mobj_t *mobj) {
    if (IS_CLIENT) {
        if (ClMobj_LocalActionsEnabled(mobj)) return true;
    }
    if (!(mobj->ddFlags & DDMF_REMOTE) ||  // only for local mobjs
        (mobj->flags3 & MF3_CLIENTACTION)) // action functions allowed?
    {
        return true;
    }
    return false;
}
static bool changeMobjState(mobj_t *mobj, statenum_t stateNum, bool doCallAction) {
    DENG_ASSERT(mobj != 0);
    do {
        if (stateNum == S_NULL) {
            mobj->state = (state_t *) S_NULL;
            P_MobjRemove(mobj, false);
            return false;
        }
        Mobj_SetState(mobj, stateNum);
        mobj->turnTime = false; // $visangle-facetarget
        state_t *st = &STATES[stateNum];
        if (doCallAction && st->action) {
            if (shouldCallAction(mobj)) {
                P_SetCurrentActionState(stateNum);
                void (*action)(mobj_t *);
                action = de::function_cast<void (*)(mobj_t *)>(st->action);
                action(mobj);
            }
        }
        stateNum = statenum_t(st->nextState);
    } while (!mobj->tics);
    return mobj->thinker.function != (thinkfunc_t) NOPFUNC;
}
dd_bool P_MobjChangeState(mobj_t *mobj, statenum_t stateNum) {
    return changeMobjState(mobj, stateNum, true /*call action functions*/);
}
dd_bool P_MobjChangeStateNoAction(mobj_t *mobj, statenum_t stateNum) {
    return changeMobjState(mobj, stateNum, false /*don't call action functions*/);
}
#if __JHEXEN__
# define MOBJ_SAVEVERSION 8
#elif __JHERETIC__
# define MOBJ_SAVEVERSION 10
#else
# define MOBJ_SAVEVERSION 10
#endif
void mobj_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    mobj_t const *mo = this;
    Writer_WriteByte(writer, MOBJ_SAVEVERSION);
#if !__JHEXEN__
    Writer_WriteInt16(writer, msw->serialIdFor(mo));
    Writer_WriteInt16(writer, msw->serialIdFor(mo->target));
# if __JDOOM__ || __JDOOM64__
    Writer_WriteInt16(writer, msw->serialIdFor(mo->tracer));
# endif
#endif
    Writer_WriteInt16(writer, msw->serialIdFor(mo->onMobj));
    Writer_WriteInt32(writer, FLT2FIX(mo->origin[VX]));
    Writer_WriteInt32(writer, FLT2FIX(mo->origin[VY]));
    Writer_WriteInt32(writer, FLT2FIX(mo->origin[VZ]));
    Writer_WriteInt32(writer, mo->angle); // Orientation.
    Writer_WriteInt32(writer, mo->sprite); // Used to find patch_t and flip value.
    Writer_WriteInt32(writer, mo->frame);
#if !__JHEXEN__
    Writer_WriteInt32(writer, FLT2FIX(mo->floorZ));
    Writer_WriteInt32(writer, FLT2FIX(mo->ceilingZ));
#endif
    Writer_WriteInt32(writer, FLT2FIX(mo->radius));
    Writer_WriteInt32(writer, FLT2FIX(mo->height));
    Writer_WriteInt32(writer, FLT2FIX(mo->mom[MX]));
    Writer_WriteInt32(writer, FLT2FIX(mo->mom[MY]));
    Writer_WriteInt32(writer, FLT2FIX(mo->mom[MZ]));
    Writer_WriteInt32(writer, mo->valid);
    Writer_WriteInt32(writer, mo->type);
    Writer_WriteInt32(writer, mo->tics); // State tic counter.
    Writer_WriteInt32(writer, int(mo->state - STATES) /*PTR2INT(mo->state)*/);
#if __JHEXEN__
    Writer_WriteInt32(writer, mo->damage);
#endif
    Writer_WriteInt32(writer, mo->flags);
#if __JHEXEN__
    Writer_WriteInt32(writer, mo->flags2);
    Writer_WriteInt32(writer, mo->flags3);
    if(mo->type == MT_KORAX)
        Writer_WriteInt32(writer, 0); // Searching index.
    else
        Writer_WriteInt32(writer, mo->special1);
    switch(mo->type) {
    case MT_LIGHTNING_FLOOR:
    case MT_LIGHTNING_ZAP:
    case MT_HOLY_TAIL:
    case MT_LIGHTNING_CEILING:
        if(mo->flags & MF_CORPSE)
            Writer_WriteInt32(writer, 0);
        else
            Writer_WriteInt32(writer, msw->serialIdFor(INT2PTR(mobj_t, mo->special2)));
        break;
    default:
        Writer_WriteInt32(writer, mo->special2);
        break;
    }
#endif
    Writer_WriteInt32(writer, mo->health);
    Writer_WriteInt32(writer, mo->moveDir); // 0-7
    Writer_WriteInt32(writer, mo->moveCount); // When 0, select a new dir.
#if __JHEXEN__
    if(mo->flags & MF_CORPSE)
        Writer_WriteInt32(writer, 0);
    else
        Writer_WriteInt32(writer, (int) msw->serialIdFor(mo->target));
#endif
    Writer_WriteInt32(writer, mo->reactionTime);
    Writer_WriteInt32(writer, mo->threshold); {
        int const playerNum = (mo->player? int(mo->player - players + 1) : 0);
        Writer_WriteInt32(writer, playerNum /*PTR2INT(mo->player)*/);
    }
    Writer_WriteInt32(writer, mo->lastLook);
#if !__JHEXEN__
    Writer_WriteInt32(writer, FLT2FIX(mo->spawnSpot.origin[VX]));
    Writer_WriteInt32(writer, FLT2FIX(mo->spawnSpot.origin[VY]));
    Writer_WriteInt32(writer, FLT2FIX(mo->spawnSpot.origin[VZ]));
    Writer_WriteInt32(writer, mo->spawnSpot.angle);
    Writer_WriteInt32(writer, mo->spawnSpot.flags);
    Writer_WriteInt32(writer, mo->intFlags); // $dropoff_fix: internal flags.
    Writer_WriteInt32(writer, FLT2FIX(mo->dropOffZ)); // $dropoff_fix
    Writer_WriteInt32(writer, mo->gear); // Used in torque simulation.
    Writer_WriteInt32(writer, mo->damage);
    Writer_WriteInt32(writer, mo->flags2);
    Writer_WriteInt32(writer, mo->flags3);
# ifdef __JHERETIC__
    Writer_WriteInt32(writer, mo->special1);
    Writer_WriteInt32(writer, mo->special2);
    Writer_WriteInt32(writer, mo->special3);
# endif
    Writer_WriteByte(writer,  mo->translucency);
    Writer_WriteByte(writer,  (byte)(mo->visTarget + 1));
#endif
    Writer_WriteInt32(writer, FLT2FIX(mo->floorClip));
#if __JHEXEN__
    Writer_WriteInt32(writer, msw->serialIdFor(mo));
    Writer_WriteInt32(writer, mo->tid);
    Writer_WriteInt32(writer, mo->special);
    Writer_Write(writer,      mo->args, sizeof(mo->args));
    Writer_WriteByte(writer,  mo->translucency);
    Writer_WriteByte(writer,  (byte)(mo->visTarget + 1));
    switch(mo->type) {
    case MT_BISH_FX:
    case MT_HOLY_FX:
    case MT_DRAGON:
    case MT_THRUSTFLOOR_UP:
    case MT_THRUSTFLOOR_DOWN:
    case MT_MINOTAUR:
    case MT_SORCFX1:
    case MT_MSTAFF_FX2:
    case MT_HOLY_TAIL:
    case MT_LIGHTNING_CEILING:
        if(mo->flags & MF_CORPSE)
            Writer_WriteInt32(writer, 0);
        else
            Writer_WriteInt32(writer, msw->serialIdFor(mo->tracer));
        break;
    default:
        DENG_ASSERT(mo->tracer == NULL); /// @todo Tracer won't be saved correctly?
        Writer_WriteInt32(writer, PTR2INT(mo->tracer));
        break;
    }
    Writer_WriteInt32(writer, 0); //PTR2INT(mo->lastEnemy));
#elif __JHERETIC__
    Writer_WriteInt16(writer, msw->serialIdFor(mo->generator));
#endif
}
int mobj_s::read(MapStateReader *msr) {
#define FF_FULLBRIGHT 0x8000 ///< Used to be a flag in thing->frame.
#define FF_FRAMEMASK  0x7fff
    Reader1 *reader = msr->reader();
    int ver = Reader_ReadByte(reader);
#if !__JHEXEN__
    if(ver >= 2) // Version 2 has mobj archive numbers.
    {
        msr->addMobjToThingArchive(this, Reader_ReadInt16(reader));
    }
#endif
#if !__JHEXEN__
    target = 0;
    if(ver >= 2) {
        target   = INT2PTR(mobj_t, Reader_ReadInt16(reader));
    }
#endif
#if __JDOOM__ || __JDOOM64__
    tracer = 0;
    if(ver >= 5) {
        tracer   = INT2PTR(mobj_t, Reader_ReadInt16(reader));
    }
#endif
    onMobj = 0;
#if __JHEXEN__
    if(ver >= 8)
#else
    if(ver >= 5)
#endif
    {
        onMobj   = INT2PTR(mobj_t, Reader_ReadInt16(reader));
    }
    origin[VX]   = FIX2FLT(Reader_ReadInt32(reader));
    origin[VY]   = FIX2FLT(Reader_ReadInt32(reader));
    origin[VZ]   = FIX2FLT(Reader_ReadInt32(reader));
    angle        = Reader_ReadInt32(reader);
    sprite       = Reader_ReadInt32(reader);
    frame        = Reader_ReadInt32(reader); // might be ORed with FF_FULLBRIGHT
    if(frame & FF_FULLBRIGHT)
        frame &= FF_FRAMEMASK; // not used anymore.
#if __JHEXEN__
    if(ver < 6) {
        /*floorflat =*/ Reader_ReadInt32(reader);
    }
#else
    floorZ       = FIX2FLT(Reader_ReadInt32(reader));
    ceilingZ     = FIX2FLT(Reader_ReadInt32(reader));
#endif
    radius       = FIX2FLT(Reader_ReadInt32(reader));
    height       = FIX2FLT(Reader_ReadInt32(reader));
    mom[MX]      = FIX2FLT(Reader_ReadInt32(reader));
    mom[MY]      = FIX2FLT(Reader_ReadInt32(reader));
    mom[MZ]      = FIX2FLT(Reader_ReadInt32(reader));
    valid        = Reader_ReadInt32(reader);
    type         = Reader_ReadInt32(reader);
#if __JHEXEN__
    if(ver < 7) {
        /*info   = (mobjinfo_t *)*/ Reader_ReadInt32(reader);
    }
#endif
    info = &MOBJINFO[type];
    if(info->flags2 & MF2_FLOATBOB)
        mom[MZ] = 0;
    if(info->flags & MF_SOLID)
        ddFlags |= DDMF_SOLID;
    if(info->flags2 & MF2_DONTDRAW)
        ddFlags |= DDMF_DONTDRAW;
    tics         = Reader_ReadInt32(reader);
    state        = INT2PTR(state_t, Reader_ReadInt32(reader));
#if __JHEXEN__
    damage       = Reader_ReadInt32(reader);
#endif
    flags        = Reader_ReadInt32(reader);
#if __JHEXEN__
    flags2       = Reader_ReadInt32(reader);
    if(ver >= 5) {
        flags3   = Reader_ReadInt32(reader);
    }
    special1     = Reader_ReadInt32(reader);
    special2     = Reader_ReadInt32(reader);
#endif
    health       = Reader_ReadInt32(reader);
#if __JHERETIC__
    if(ver < 8) {
        switch(type) {
        case MT_MACEFX1:
        case MT_MACEFX2:
        case MT_MACEFX3:
        case MT_HORNRODFX2:
        case MT_HEADFX3:
        case MT_WHIRLWIND:
        case MT_TELEGLITTER:
        case MT_TELEGLITTER2:
            special3 = health;
            if(type == MT_HORNRODFX2 && special3 > 16)
                special3 = 16;
            health = MOBJINFO[type].spawnHealth;
            break;
        default:
            break;
        }
    }
#endif
    moveDir      = Reader_ReadInt32(reader);
    moveCount    = Reader_ReadInt32(reader);
#if __JHEXEN__
    target       = INT2PTR(mobj_t, Reader_ReadInt32(reader));
#endif
    reactionTime = Reader_ReadInt32(reader);
    threshold    = Reader_ReadInt32(reader);
    player       = INT2PTR(player_t, Reader_ReadInt32(reader));
    lastLook     = Reader_ReadInt32(reader);
#if __JHEXEN__
    floorClip    = FIX2FLT(Reader_ReadInt32(reader));
    msr->addMobjToThingArchive(this, Reader_ReadInt32(reader));
    tid          = Reader_ReadInt32(reader);
#else
    if(ver >= 6) {
        spawnSpot.origin[VX] = FIX2FLT(Reader_ReadInt32(reader));
        spawnSpot.origin[VY] = FIX2FLT(Reader_ReadInt32(reader));
        spawnSpot.origin[VZ] = FIX2FLT(Reader_ReadInt32(reader));
        spawnSpot.angle      = Reader_ReadInt32(reader);
        if(ver < 10) {
            /*spawnSpot.type =*/ Reader_ReadInt32(reader);
        }
        spawnSpot.flags      = Reader_ReadInt32(reader);
    }
    else {
        spawnSpot.origin[VX] = (float) Reader_ReadInt16(reader);
        spawnSpot.origin[VY] = (float) Reader_ReadInt16(reader);
        spawnSpot.origin[VZ] = 0; // Initialize with "something".
        spawnSpot.angle      = (angle_t) (ANG45 * (Reader_ReadInt16(reader) / 45));
        /*spawnSpot.type       = (int)*/ Reader_ReadInt16(reader);
        spawnSpot.flags      = (int) Reader_ReadInt16(reader);
    }
# if __JDOOM__ || __JDOOM64__
    if(ver >= 3)
# elif __JHERETIC__
    if(ver >= 5)
# endif
    {
        intFlags = Reader_ReadInt32(reader);
        dropOffZ = FIX2FLT(Reader_ReadInt32(reader));
        gear     = Reader_ReadInt32(reader);
    }
# if __JDOOM__ || __JDOOM64__
    if(ver >= 6) {
        damage = Reader_ReadInt32(reader);
        flags2 = Reader_ReadInt32(reader);
    }
    else // flags2 will be applied from the defs.
    {
        damage = DDMAXINT; // Use the value set in mo->info->damage
    }
# elif __JHERETIC__
    damage = Reader_ReadInt32(reader);
    flags2 = Reader_ReadInt32(reader);
# endif
    if(ver >= 7) {
        flags3 = Reader_ReadInt32(reader);
    } // Else flags3 will be applied from the defs.
#endif
#if __JHEXEN__
    special = Reader_ReadInt32(reader);
    Reader_Read(reader, args, 1 * 5);
#elif __JHERETIC__
    special1 = Reader_ReadInt32(reader);
    special2 = Reader_ReadInt32(reader);
    if(ver >= 8) {
        special3 = Reader_ReadInt32(reader);
    }
#endif
#if __JHEXEN__
    if(ver >= 2)
#else
    if(ver >= 4)
#endif
    {
        translucency = Reader_ReadByte(reader);
    }
#if __JHEXEN__
    if(ver >= 3)
#else
    if(ver >= 5)
#endif
    {
        visTarget = (short) (Reader_ReadByte(reader)) -1;
    }
#if __JHEXEN__
    if(ver >= 4) {
        /*tracer = */INT2PTR(mobj_t, Reader_ReadInt32(reader));
        tracer = nullptr;
    }
    if(ver >= 4) {
        /*lastEnemy =*/ INT2PTR(mobj_t, Reader_ReadInt32(reader));
        lastEnemy = nullptr;
    }
#else
    if(ver >= 5) {
        floorClip = FIX2FLT(Reader_ReadInt32(reader));
    }
#endif
#if __JHERETIC__
    if(ver >= 7) {
        generator = INT2PTR(mobj_t, Reader_ReadInt16(reader));
    }
    else {
        generator = 0;
    }
#endif
    /*
     * Restore! (unmangle)
     */
    info = &MOBJINFO[type];
    Mobj_SetState(this, PTR2INT(state));
#if __JHEXEN__
    if(flags2 & MF2_DORMANT) {
        tics = -1;
    }
#endif
    if(player) {
        player = msr->player(PTR2INT(player));
#if __JHEXEN__
        if(!player) {
            Mobj_Destroy(this);
            return false;
        }
#endif
        dPlayer = player->plr;
        dPlayer->mo      = this;
        dPlayer->lookDir = 0; /* $unifiedangles */
    }
    visAngle = angle >> 16;
#if !__JHEXEN__
    if(dPlayer && !dPlayer->inGame) {
        dPlayer->mo = 0;
        Mobj_Destroy(this);
        return false;
    }
#endif
#if !__JDOOM64__
    if(ver < MOBJ_SAVEVERSION) {
        SV_TranslateLegacyMobjFlags(this, ver);
    }
#endif
    P_MobjLink(this);
    floorZ   = P_GetDoublep(Mobj_Sector(this), DMU_FLOOR_HEIGHT);
    ceilingZ = P_GetDoublep(Mobj_Sector(this), DMU_CEILING_HEIGHT);
    return false;
#undef FF_FRAMEMASK
#undef FF_FULLBRIGHT
}
mobj_t *Mobj_ExplodeIfObstructed(mobj_t *mob) {
    return P_CheckMissileSpawn(mob)? mob : nullptr;
}
mobj_t *P_LaunchMissile(mobj_t *missile, angle_t angle, coord_t const targetPos[],
    coord_t const sourcePos[], coord_t extraMomZ) {
    DENG2_ASSERT(targetPos);
    if(missile) {
        DENG2_ASSERT(missile->info);
        if(missile->info->seeSound) {
            S_StartSound(missile->info->seeSound, missile);
        }
        if(!sourcePos) {
            sourcePos = missile->origin;
        }
        coord_t dist;
        uint an = angle >> ANGLETOFINESHIFT;
        missile->mom[0] = missile->info->speed * FIX2FLT(finecosine[an]);
        missile->mom[1] = missile->info->speed * FIX2FLT(finesine  [an]);
        dist = M_ApproxDistance(targetPos[0] - sourcePos[0], targetPos[1] - sourcePos[1]);
        dist /= missile->info->speed;
        if(dist < 1) dist = 1;
        missile->mom[2] = (targetPos[2] - sourcePos[2] + extraMomZ) / dist;
    }
    return Mobj_ExplodeIfObstructed(missile);
}
mobj_t *Mobj_LaunchMissileAtAngle2(mobj_t *mob, mobj_t *missile, angle_t angle,
    coord_t const targetPos[], coord_t const sourcePos[], coord_t extraMomZ) {
    DENG2_ASSERT(mob);
    if(missile) {
        missile->target = mob;
    }
    return P_LaunchMissile(missile, angle, targetPos, sourcePos, extraMomZ);
}
mobj_t *Mobj_LaunchMissileAtAngle(mobj_t *mob, mobj_t *missile, angle_t angle,
    coord_t const targetPos[], coord_t const sourcePos[]) {
    return Mobj_LaunchMissileAtAngle2(mob, missile, angle, targetPos, sourcePos,
                                      0/*no extra z-momentum*/);
}
mobj_t *Mobj_LaunchMissile2(mobj_t *mob, mobj_t *missile, coord_t const targetPos[],
    coord_t const sourcePos[], coord_t extraMomZ) {
    DENG2_ASSERT(mob);
    return Mobj_LaunchMissileAtAngle2(mob, missile, missile? missile->angle : mob->angle,
                                      targetPos, sourcePos, extraMomZ);
}
mobj_t *Mobj_LaunchMissile(mobj_t *mob, mobj_t *missile, coord_t const targetPos[],
    coord_t const sourcePos[]) {
    return Mobj_LaunchMissile2(mob, missile, targetPos, sourcePos, 0/*no extra z-momentum*/);
}
void Mobj_InflictDamage(mobj_t *mob, mobj_t const *inflictor, int damage) {
    DENG_ASSERT(mob);
    mob->health -= damage;
    THINKER_DATA(mob->thinker, MobjThinkerData).damageReceived(damage, inflictor);
}
de::String Mobj_StateAsInfo(mobj_t const *mob) {
    using de::String;
    QString str;
    QTextStream os(&str);
    os.setCodec("UTF-8");
    os << "Mobj 0x" << String::number(Mobj_PrivateID(mob), 16)
       << " {\n  target = 0x" << String::number(Mobj_PrivateID(mob->target), 16)
       << "\n  onMobj = 0x" << String::number(Mobj_PrivateID(mob->onMobj), 16)
       << "\n  tracer = 0x" << String::number(Mobj_PrivateID(mob->tracer), 16);
    #if defined (__JHERETIC__)
    {
        os << "\n  generator = 0x" << String::number(Mobj_PrivateID(mob->generator), 16);
    }
    #endif
    #if defined (__JHEXEN__)
    {
        os << "\n  lastEnemy = 0x" << String::number(Mobj_PrivateID(mob->lastEnemy), 16);
    }
    #endif
    os << "\n}\n";
    return str;
}
void Mobj_RestoreObjectState(mobj_t *mob, de::Info::BlockElement const &state) {
    static de::String const KEY_TARGET     ("target");
    static de::String const KEY_ON_MOBJ    ("onMobj");
    static de::String const KEY_TRACER     ("tracer");
    static de::String const KEY_GENERATOR  ("generator");
    static de::String const KEY_LAST_ENEMY ("lastEnemy");
    bool ok;
    uint32_t pid;
    pid = state.keyValue(KEY_TARGET).text.toUInt32(&ok, 0);
    if (ok) {
        mob->target = Mobj_FindByPrivateID(pid);
    }
    pid = state.keyValue(KEY_ON_MOBJ).text.toUInt32(&ok, 0);
    if (ok) {
        mob->onMobj = Mobj_FindByPrivateID(pid);
    }
    pid = state.keyValue(KEY_TRACER).text.toUInt32(&ok, 0);
    if (ok) {
        mob->tracer = Mobj_FindByPrivateID(pid);
    }
    #if defined (__JHERETIC__)
    {
        pid = state.keyValue(KEY_GENERATOR).text.toUInt32(&ok, 0);
        if (ok) {
            mob->generator = Mobj_FindByPrivateID(pid);
        }
    }
    #endif
    #if defined (__JHEXEN__)
    {
        pid = state.keyValue(KEY_LAST_ENEMY).text.toUInt32(&ok, 0);
        if (ok) {
            mob->lastEnemy = Mobj_FindByPrivateID(pid);
        }
    }
    #endif
}
void Mobj_RunScriptOnDeath(mobj_t *mob, mobj_t *killer) {
    using namespace de;
    if (IS_NETWORK_CLIENT) {
        return;
    }
    const auto &thingDef = DED_Definitions()->things[mob->type];
    if (const String onDeathSrc = thingDef.gets(QStringLiteral("onDeath"))) {
        LOG_AS("Mobj_RunScriptOnDeath");
        const auto &self = THINKER_NS(mob->thinker);
        Record ns;
        ns.add(QStringLiteral("self")).set(new RecordValue(self));
        Variable &killerVar = ns.add(QStringLiteral("killer"));
        if (killer) {
            killerVar.set(new RecordValue(THINKER_NS(killer->thinker)));
        }
        else {
            killerVar.set(new NoneValue);
        }
        Process proc(&ns);
        Script script(onDeathSrc);
        proc.run(script);
        proc.execute();
    }
}
dd_bool Mobj_RunScriptOnTouch(mobj_t *mob, mobj_t *special, enum mobjtouchresult_e *result) {
    using namespace de;
    if (IS_NETWORK_CLIENT) {
        return false;
    }
    const auto &thingDef = DED_Definitions()->things[special->type];
    if (const String onTouchSrc = thingDef.gets(QStringLiteral("onTouch"))) {
        LOG_AS("Mobj_RunScriptOnTouch");
        const auto &self = THINKER_NS(special->thinker);
        String src = "def onTouch(toucher)\n";
        src += onTouchSrc;
        src += "\nend";
        Record ns;
        Script script(src);
        Process proc(&ns);
        proc.run(script);
        proc.execute();
        ns.add(QStringLiteral("self")).set(new RecordValue(self));
        std::unique_ptr<Value> resultValue(
            Process::scriptCall(Process::TakeResult,
                                ns,
                                QStringLiteral("onTouch"),
                                &THINKER_DATA(mob->thinker, ThinkerData)));
        if (result) {
            *result = MTR_KEEP;
            if (resultValue && !is<NoneValue>(resultValue.get())) {
                const String val = resultValue->asText();
                if      (val == "keep")    *result = MTR_KEEP;
                else if (val == "dormant") *result = MTR_MAKE_DORMANT;
                else if (val == "hide")    *result = MTR_HIDE;
                else if (val == "destroy") *result = MTR_DESTROY;
                else {
                    LOG_SCR_ERROR("Invalid return value from function: \"%s\"")
                        << resultValue->asText();
                }
            }
        }
        return true;
    }
    return false;
}
typedef struct spawnqueuenode_s {
    int startTime;
    int minTics; ///< Minimum number of tics before respawn.
    void (*callback) (mobj_t *mo, void *context);
    void* context;
    coord_t pos[3];
    angle_t angle;
    mobjtype_t type;
    int spawnFlags; ///< MSF_* flags
    struct spawnqueuenode_s *next;
} spawnqueuenode_t;
static spawnqueuenode_t *spawnQueueHead, *unusedNodes;
void P_SpawnTelefog(mobj_t *mo, void * /*context*/)
{
#if __JDOOM__ || __JDOOM64__
    S_StartSound(SFX_ITMBK, mo);
#else
    S_StartSound(SFX_RESPAWN, mo);
#endif
# if __JDOOM64__
    mo->translucency = 255;
    mo->spawnFadeTics = 0;
    mo->intFlags |= MIF_FADE;
# elif __JDOOM__
    P_SpawnMobj(MT_IFOG, mo->origin, mo->angle, 0);
# endif
}
void P_MobjRemove(mobj_t *mo, dd_bool noRespawn) {
#if !defined(__JDOOM__) && !defined(__JDOOM64__)
    DENG_UNUSED(noRespawn);
#endif
    if (mo->ddFlags & DDMF_REMOTE)
        goto justDoIt;
#if __JDOOM__ || __JDOOM64__
    if (!noRespawn) {
        if (
# if __JDOOM__
            (gfw_Rule(deathmatch) == 2 ||
             (cfg.coopRespawnItems && IS_NETGAME && !gfw_Rule(deathmatch))) &&
# endif /*#elif __JDOOM64__
           (spot->flags & MTF_RESPAWN) &&
# endif*/
            (mo->flags & MF_SPECIAL) && !(mo->flags & MF_DROPPED)
# if __JDOOM__ || __JDOOM64__
            && (mo->type != MT_INV) && (mo->type != MT_INS)
# endif
           ) {
            P_DeferSpawnMobj3fv(RESPAWNTICS, mobjtype_t(mo->type),
                                mo->spawnSpot.origin, mo->spawnSpot.angle,
                                mo->spawnSpot.flags, P_SpawnTelefog, NULL);
        }
    }
#endif
#if __JHEXEN__
    if ((mo->flags & MF_COUNTKILL) && (mo->flags & MF_CORPSE)) {
        P_RemoveCorpseInQueue(mo);
    }
    P_MobjRemoveFromTIDList(mo);
#endif
justDoIt:
    Mobj_Destroy(mo);
}
void P_MobjLink(struct mobj_s *mobj) {
    DENG_ASSERT(mobj != 0);
    Mobj_Link(mobj, MLF_BLOCKMAP | (!(mobj->flags & MF_NOSECTOR)? MLF_SECTOR : 0));
}
void P_MobjUnlink(struct mobj_s *mobj) {
    Mobj_Unlink(mobj);
}
void P_MobjSetSRVO(mobj_t *mo, coord_t stepx, coord_t stepy) {
    DENG_ASSERT(mo != 0);
    mo->srvo[VX] = -stepx;
    mo->srvo[VY] = -stepy;
}
void P_MobjSetSRVOZ(mobj_t *mo, coord_t stepz) {
    DENG_ASSERT(mo != 0);
    mo->srvo[VZ] = -stepz;
}
void P_MobjAngleSRVOTicker(mobj_t *mo) {
#define MIN_STEP (10 * ANGLE_1) >> 16 ///< Degrees per tic
#define MAX_STEP ANG90 >> 16
    DENG_ASSERT(mo != 0);
    if(mo->flags & MF_MISSILE || !(mo->flags & MF_COUNTKILL)) {
        mo->visAngle = mo->angle >> 16;
        return; // This is not for us.
    }
    short target = mo->angle >> 16;
    short diff = target - mo->visAngle;
    short step = 0;
    if(mo->turnTime) {
        if(mo->tics) step = de::abs(diff) / mo->tics;
        else         step = de::abs(diff);
        if(!step) step = 1;
    }
    else {
        int hgt = (int) mo->height;
        hgt = MINMAX_OF(30, hgt, 60);
        int lstep = de::abs(diff) * 8 / hgt;
        lstep = MINMAX_OF(MIN_STEP, lstep, MAX_STEP);
        step = lstep;
    }
    if(de::abs(diff) <= step)
        mo->visAngle  = target;
    else if(diff > 0)
        mo->visAngle += step;
    else if(diff < 0)
        mo->visAngle -= step;
#undef MAX_STEP
#undef MIN_STEP
}
void P_MobjClearSRVO(mobj_t *mo) {
    DENG_ASSERT(mo != 0);
    std::memset(mo->srvo, 0, sizeof(mo->srvo));
}
dd_bool P_MobjIsCamera(mobj_t const *mo) {
    return (mo && mo->thinker.function && mo->player &&
            (mo->player->plr->flags & DDPF_CAMERA));
}
dd_bool Mobj_IsCrunchable(mobj_t *mobj) {
    DENG_ASSERT(mobj != 0);
#if __JDOOM__ || __JDOOM64__
    return mobj->health <= 0 && (cfg.gibCrushedNonBleeders || !(mobj->flags & MF_NOBLOOD));
#elif __JHEXEN__
    return mobj->health <= 0 && (mobj->flags & MF_CORPSE) != 0;
#else
    return mobj->health <= 0;
#endif
}
dd_bool Mobj_IsDroppedItem(mobj_t *mobj) {
    DENG_ASSERT(mobj != 0);
#if __JHEXEN__
    return (mobj->flags2 & MF2_DROPPED) != 0;
#else
    return (mobj->flags & MF_DROPPED) != 0;
#endif
}
terraintype_t const *P_MobjFloorTerrain(mobj_t const *mobj) {
    return P_PlaneMaterialTerrainType(Mobj_Sector(mobj), PLN_FLOOR);
}
void P_UpdateHealthBits(mobj_t *mo) {
    if(!mo || !mo->info) return;
    if(mo->info->spawnHealth > 0) {
        mo->selector &= DDMOBJ_SELECTOR_MASK; // Clear high byte.
        int sel = (mo->health << 3) / mo->info->spawnHealth;
        sel = MINMAX_OF(0, sel, 7);
        mo->selector |= sel << DDMOBJ_SELECTOR_SHIFT;
    }
}
statenum_t P_GetState(mobjtype_t type, statename_t name) {
    if(type < MT_FIRST || type >= Get(DD_NUMMOBJTYPES)) return S_NULL;
    if(name < 0 || name >= STATENAMES_COUNT) return S_NULL;
    return statenum_t(MOBJINFO[type].states[name]);
}
void P_RipperBlood(mobj_t *actor) {
    DENG_ASSERT(actor != 0);
    coord_t pos[3];
    pos[VX] = actor->origin[VX];
    pos[VY] = actor->origin[VY];
    pos[VZ] = actor->origin[VZ];
    pos[VX] += FIX2FLT((P_Random() - P_Random()) << 12);
    pos[VY] += FIX2FLT((P_Random() - P_Random()) << 12);
    pos[VZ] += FIX2FLT((P_Random() - P_Random()) << 12);
    if(mobj_t *mo = P_SpawnMobj(MT_BLOOD, pos, actor->angle, 0)) {
#if __JHERETIC__
        mo->flags |= MF_NOGRAVITY;
#endif
        mo->mom[MX] = actor->mom[MX] / 2;
        mo->mom[MY] = actor->mom[MY] / 2;
        mo->tics += P_Random() & 3;
    }
}
static spawnqueuenode_t *allocateNode() {
#define SPAWNQUEUENODE_BATCHSIZE 32
    spawnqueuenode_t *node;
    if(unusedNodes) {
        node = unusedNodes;
        unusedNodes = unusedNodes->next;
        node->next = 0;
    }
    else {
        spawnqueuenode_t *storage = (spawnqueuenode_t *)
            Z_Malloc(sizeof(*node) * SPAWNQUEUENODE_BATCHSIZE, PU_GAMESTATIC, 0);
        for(int i = 0; i < SPAWNQUEUENODE_BATCHSIZE-1; ++i) {
            node = storage++;
            node->next = unusedNodes;
            unusedNodes = node;
        }
        node = storage;
    }
    return node;
#undef SPAWNQUEUENODE_BATCHSIZE
}
static void freeNode(spawnqueuenode_t *node, bool recycle) {
    if(!node) return;
    if(spawnQueueHead) {
        if(spawnQueueHead == node) {
            spawnQueueHead = spawnQueueHead->next;
        }
        else {
            for(spawnqueuenode_t *other = spawnQueueHead; other->next; other = other->next) {
                if(other->next == node)
                    other->next = other->next->next;
            }
        }
    }
    if(recycle) {
        node->next = unusedNodes;
        unusedNodes = node;
        return;
    }
    Z_Free(node);
}
static spawnqueuenode_t *dequeueSpawn() {
    spawnqueuenode_t *node = spawnQueueHead;
    if(spawnQueueHead)
        spawnQueueHead = spawnQueueHead->next;
    return node;
}
static void emptySpawnQueue(bool recycle) {
    if(spawnQueueHead) {
        while(spawnqueuenode_t *node = dequeueSpawn()) {
            freeNode(node, recycle);
        }
    }
    spawnQueueHead = 0;
}
static void enqueueSpawn(int minTics, mobjtype_t type, coord_t x, coord_t y,
    coord_t z, angle_t angle, int spawnFlags,
    void (*callback) (mobj_t *mo, void *context), void *context) {
    spawnqueuenode_t *spawn = allocateNode();
    spawn->type = type;
    spawn->pos[VX] = x;
    spawn->pos[VY] = y;
    spawn->pos[VZ] = z;
    spawn->angle = angle;
    spawn->spawnFlags = spawnFlags;
    spawn->startTime = mapTime;
    spawn->minTics = minTics;
    spawn->callback = callback;
    spawn->context = context;
    if(spawnQueueHead) {
        if(spawnQueueHead->next) {
            spawnqueuenode_t *other = spawnQueueHead;
            while(other->next && other->next->minTics - (mapTime - other->next->startTime) <= minTics) {
                other = other->next;
            }
            spawn->next = (other->next? other->next : 0);
            other->next = spawn;
        }
        else {
            if(spawnQueueHead->minTics - (mapTime - spawnQueueHead->startTime) <= minTics) {
                spawn->next = 0;
                spawnQueueHead->next = spawn;
            }
            else {
                spawn->next = spawnQueueHead;
                spawnQueueHead = spawn;
            }
        }
    }
    else {
        spawn->next = 0;
        spawnQueueHead = spawn;
    }
}
void P_DeferSpawnMobj3f(int minTics, mobjtype_t type, coord_t x, coord_t y, coord_t z,
    angle_t angle, int spawnFlags,  void (*callback) (mobj_t *mo, void *context),
    void* context) {
    if(minTics > 0) {
        enqueueSpawn(minTics, type, x, y, z, angle, spawnFlags, callback,
                     context);
    }
    else if(mobj_t *mo = P_SpawnMobjXYZ(type, x, y, z, angle, spawnFlags)) {
        if(callback)
            callback(mo, context);
    }
}
void P_DeferSpawnMobj3fv(int minTics, mobjtype_t type, coord_t const pos[3], angle_t angle,
    int spawnFlags, void (*callback) (mobj_t* mo, void* context), void* context) {
    if(minTics > 0) {
        enqueueSpawn(minTics, type, pos[VX], pos[VY], pos[VZ], angle,
                     spawnFlags, callback, context);
    }
    else if(mobj_t *mo = P_SpawnMobj(type, pos, angle, spawnFlags)) {
        if(callback)
            callback(mo, context);
    }
}
static mobj_t *processOneSpawnTask() {
    mobj_t *mo = 0;
    if(spawnQueueHead && mapTime - spawnQueueHead->startTime >= spawnQueueHead->minTics) {
        spawnqueuenode_t *spawn = dequeueSpawn();
        DENG_ASSERT(spawn != 0);
        if((mo = P_SpawnMobj(spawn->type, spawn->pos, spawn->angle, spawn->spawnFlags))) {
            if(spawn->callback)
                spawn->callback(mo, spawn->context);
        }
        freeNode(spawn, true);
    }
    return mo;
}
void P_ProcessDeferredSpawns() {
    while(processOneSpawnTask())
    {}
}
void P_PurgeDeferredSpawns() {
    emptySpawnQueue(true);
}
#ifdef __JHEXEN__
#define MAX_TID_COUNT           200
static int TIDList[MAX_TID_COUNT + 1];  // +1 for termination marker
static mobj_t *TIDMobj[MAX_TID_COUNT];
static int insertThinkerInIdListWorker(thinker_t *th, void *context) {
    mobj_t *mo = (mobj_t *)th;
    int *count = (int *) context;
    if(mo->tid != 0) {
        if(*count == MAX_TID_COUNT) {
            Con_Error("P_CreateTIDList: MAX_TID_COUNT (%d) exceeded.", MAX_TID_COUNT);
        }
        TIDList[*count] = mo->tid;
        TIDMobj[(*count)++] = mo;
    }
    return false; // Continue iteration.
}
void P_CreateTIDList() {
    int count = 0;
    Thinker_Iterate(P_MobjThinker, insertThinkerInIdListWorker, &count);
    TIDList[count] = 0;
}
void P_MobjInsertIntoTIDList(mobj_t *mo, int tid) {
    DENG_ASSERT(mo != 0);
    int index = -1;
    int i = 0;
    for(; TIDList[i] != 0; ++i) {
        if(TIDList[i] == -1) {
            index = i;
            break;
        }
    }
    if(index == -1) {
        if(i == MAX_TID_COUNT) {
            Con_Error("P_MobjInsertIntoTIDList: MAX_TID_COUNT (%d) exceeded.", MAX_TID_COUNT);
        }
        index = i;
        TIDList[index + 1] = 0;
    }
    mo->tid = tid;
    TIDList[index] = tid;
    TIDMobj[index] = mo;
}
void P_MobjRemoveFromTIDList(mobj_t *mo) {
    if(!mo || !mo->tid)
        return;
    for(int i = 0; TIDList[i] != 0; ++i) {
        if(TIDMobj[i] == mo) {
            TIDList[i] = -1;
            TIDMobj[i] = 0;
            mo->tid = 0;
            return;
        }
    }
    mo->tid = 0;
}
mobj_t *P_FindMobjFromTID(int tid, int *searchPosition) {
    DENG_ASSERT(searchPosition != 0);
    for(int i = *searchPosition + 1; TIDList[i] != 0; ++i) {
        if(TIDList[i] == tid) {
            *searchPosition = i;
            return TIDMobj[i];
        }
    }
    *searchPosition = -1;
    return 0;
}
#endif // __JHEXEN__
static void stopCeiling(ceiling_t *ceiling) {
    P_ToXSector(ceiling->sector)->specialData = nullptr;
    P_NotifySectorFinished(P_ToXSector(ceiling->sector)->tag);
    Thinker_Remove(&ceiling->thinker);
}
void T_MoveCeiling(void *ceilingThinkerPtr) {
    ceiling_t *ceiling = (ceiling_t *)ceilingThinkerPtr;
    result_e res;
    switch(ceiling->state) {
    case CS_UP: // Going up.
        res =
            T_MovePlane(ceiling->sector, ceiling->speed, ceiling->topHeight,
                        false, 1, 1);
#if !__JHEXEN__
        if(!(mapTime & 7)) {
# if __JHERETIC__
            S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGMOVE);
# else
            switch(ceiling->type) {
            case CT_SILENTCRUSHANDRAISE:
                break;
            default:
                S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGMOVE);
                break;
            }
# endif
        }
#endif
        if(res == pastdest) {
#if __JHEXEN__
            SN_StopSequence((mobj_t *)P_GetPtrp(ceiling->sector, DMU_EMITTER));
#endif
            switch(ceiling->type) {
#if !__JHEXEN__
            case CT_RAISETOHIGHEST:
# if __JDOOM64__
            case CT_CUSTOM: //jd64
# endif
                stopCeiling(ceiling);
                break;
# if !__JHERETIC__
            case CT_SILENTCRUSHANDRAISE:
                S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGSTOP);
# endif
            case CT_CRUSHANDRAISEFAST:
#endif
            case CT_CRUSHANDRAISE:
                ceiling->state = CS_DOWN;
#if __JHEXEN__
                ceiling->speed *= 2;
#endif
                break;
            default:
#if __JHEXEN__
                stopCeiling(ceiling);
#endif
                break;
            }
        }
        break;
    case CS_DOWN: // Going down.
        res =
            T_MovePlane(ceiling->sector, ceiling->speed, ceiling->bottomHeight,
                        ceiling->crush, 1, -1);
#if !__JHEXEN__
        if(!(mapTime & 7)) {
# if __JHERETIC__
            S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGMOVE);
# else
            switch(ceiling->type) {
            case CT_SILENTCRUSHANDRAISE:
                break;
            default:
                S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGMOVE);
            }
# endif
        }
#endif
        if(res == pastdest) {
#if __JHEXEN__
            SN_StopSequence((mobj_t *)P_GetPtrp(ceiling->sector, DMU_EMITTER));
#endif
            switch(ceiling->type) {
#if __JDOOM__ || __JDOOM64__
            case CT_SILENTCRUSHANDRAISE:
                S_PlaneSound((Plane *)P_GetPtrp(ceiling->sector, DMU_CEILING_PLANE), SFX_CEILINGSTOP);
                ceiling->speed = CEILSPEED;
                ceiling->state = CS_UP;
                break;
#endif
            case CT_CRUSHANDRAISE:
#if __JHEXEN__
            case CT_CRUSHRAISEANDSTAY:
#endif
#if __JHEXEN__
                ceiling->speed = ceiling->speed * .5;
#else
                ceiling->speed = CEILSPEED;
#endif
                ceiling->state = CS_UP;
                break;
#if !__JHEXEN__
            case CT_CRUSHANDRAISEFAST:
                ceiling->state = CS_UP;
                break;
            case CT_LOWERANDCRUSH:
            case CT_LOWERTOFLOOR:
# if __JDOOM64__
            case CT_CUSTOM: //jd64
# endif
                stopCeiling(ceiling);
                break;
#endif
            default:
#if __JHEXEN__
                stopCeiling(ceiling);
#endif
                break;
            }
        }
        else                    // ( res != pastdest )
        {
            if(res == crushed) {
                switch(ceiling->type) {
#if __JDOOM__ || __JDOOM64__
                case CT_SILENTCRUSHANDRAISE:
#endif
                case CT_CRUSHANDRAISE:
                case CT_LOWERANDCRUSH:
#if __JHEXEN__
                case CT_CRUSHRAISEANDSTAY:
#endif
#if !__JHEXEN__
                    ceiling->speed = CEILSPEED * .125;
#endif
                    break;
                default:
                    break;
                }
            }
        }
        break;
    }
}
void ceiling_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 2); // Write a version byte.
    Writer_WriteByte(writer, (byte) type);
    Writer_WriteInt32(writer, P_ToIndex(sector));
    Writer_WriteInt16(writer, (int)bottomHeight);
    Writer_WriteInt16(writer, (int)topHeight);
    Writer_WriteInt32(writer, FLT2FIX(speed));
    Writer_WriteByte(writer, crush);
    Writer_WriteByte(writer, (byte) state);
    Writer_WriteInt32(writer, tag);
    Writer_WriteByte(writer, (byte) oldState);
}
int ceiling_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
#if __JHEXEN__
    if(mapVersion >= 4)
#else
    if(mapVersion >= 5)
#endif
    {
        int ver = Reader_ReadByte(reader); // version byte.
        thinker.function = T_MoveCeiling;
#if !__JHEXEN__
        if(mapVersion == 5) {
            if(!Reader_ReadByte(reader))
                Thinker_SetStasis(&thinker, true);
        }
#endif
        type         = (ceilingtype_e) Reader_ReadByte(reader);
        sector       = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        bottomHeight = (float) Reader_ReadInt16(reader);
        topHeight    = (float) Reader_ReadInt16(reader);
        speed        = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        crush        = Reader_ReadByte(reader);
        if(ver == 2)
            state    = ceilingstate_e(Reader_ReadByte(reader));
        else
            state    = ceilingstate_e(Reader_ReadInt32(reader) == -1? CS_DOWN : CS_UP);
        tag          = Reader_ReadInt32(reader);
        if(ver == 2)
            oldState = ceilingstate_e(Reader_ReadByte(reader));
        else
            state    = (Reader_ReadInt32(reader) == -1? CS_DOWN : CS_UP);
    }
    else {
        int32_t junk[4]; // sizeof thinker_t
        Reader_Read(reader, (byte *)junk, 16);
#if __JHEXEN__
        sector       = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        type         = ceilingtype_e(Reader_ReadInt32(reader));
#else
        type         = ceilingtype_e(Reader_ReadInt32(reader));
        sector       = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
#endif
        bottomHeight = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        topHeight    = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        speed        = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        crush        = Reader_ReadInt32(reader);
        state        = (Reader_ReadInt32(reader) == -1? CS_DOWN : CS_UP);
        tag          = Reader_ReadInt32(reader);
        oldState     = (Reader_ReadInt32(reader) == -1? CS_DOWN : CS_UP);
        thinker.function = T_MoveCeiling;
#if !__JHEXEN__
        if(junk[2] == 0) // thinker_t::function
        {
            Thinker_SetStasis(&thinker, true);
        }
#endif
    }
    P_ToXSector(sector)->specialData = this;
    return true; // Add this thinker.
}
#if __JDOOM64__
static int EV_DoCeiling2(Line *line, int tag, float basespeed, ceilingtype_e type)
#elif __JHEXEN__
static int EV_DoCeiling2(byte *arg, int tag, float basespeed, ceilingtype_e type)
#else
static int EV_DoCeiling2(int tag, float basespeed, ceilingtype_e type)
#endif
{
    iterlist_t *list = P_GetSectorIterListForTag(tag, false);
    if(!list) return 0;
    int rtn = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec = 0;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        xsector_t *xsec = P_ToXSector(sec);
        if(xsec->specialData)
            continue;
        rtn = 1;
        ceiling_t *ceiling = (ceiling_t *)Z_Calloc(sizeof(*ceiling), PU_MAP, 0);
        ceiling->thinker.function = T_MoveCeiling;
        Thinker_Add(&ceiling->thinker);
        xsec->specialData = ceiling;
        ceiling->sector = sec;
        ceiling->crush = false;
        ceiling->speed = basespeed;
        switch(type) {
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        case CT_CRUSHANDRAISEFAST:
            ceiling->crush = true;
            ceiling->topHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
            ceiling->bottomHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + 8;
            ceiling->state = CS_DOWN;
            ceiling->speed *= 2;
            break;
#endif
#if __JHEXEN__
        case CT_CRUSHRAISEANDSTAY:
            ceiling->crush = (int) arg[2];    // arg[2] = crushing value
            ceiling->topHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
            ceiling->bottomHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + 8;
            ceiling->state = CS_DOWN;
            break;
#endif
#if __JDOOM__ || __JDOOM64__
        case CT_SILENTCRUSHANDRAISE:
#endif
        case CT_CRUSHANDRAISE:
#if !__JHEXEN__
            ceiling->crush = true;
#endif
            ceiling->topHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
        case CT_LOWERANDCRUSH:
#if __JHEXEN__
            ceiling->crush = (int) arg[2];    // arg[2] = crushing value
#endif
        case CT_LOWERTOFLOOR:
            ceiling->bottomHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
            if(type != CT_LOWERTOFLOOR)
                ceiling->bottomHeight += 8;
            ceiling->state = CS_DOWN;
#if __JDOOM64__
            ceiling->speed *= 8; // jd64
#endif
            break;
        case CT_RAISETOHIGHEST:
            P_FindSectorSurroundingHighestCeiling(sec, 0, &ceiling->topHeight);
#if __JDOOM64__
            ceiling->topHeight -= 8;   // jd64
#endif
            ceiling->state = CS_UP;
            break;
#if __JDOOM64__
        case CT_CUSTOM: // jd64
            {
            Side *front = (Side *)P_GetPtrp(line, DMU_FRONT);
            Side *back = (Side *)P_GetPtrp(line, DMU_BACK);
            coord_t bitmipL = 0, bitmipR = 0;
            bitmipL = P_GetDoublep(front, DMU_MIDDLE_MATERIAL_OFFSET_X);
            if(back)
                bitmipR = P_GetDoublep(back, DMU_MIDDLE_MATERIAL_OFFSET_X);
            if(bitmipR > 0) {
                P_FindSectorSurroundingHighestCeiling(sec, 0, &ceiling->topHeight);
                ceiling->state = CS_UP;
                ceiling->speed *= bitmipL;
                ceiling->topHeight -= bitmipR;
            }
            else {
                ceiling->bottomHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
                ceiling->bottomHeight -= bitmipR;
                ceiling->state = CS_DOWN;
                ceiling->speed *= bitmipL;
            }
            }
#endif
#if __JHEXEN__
        case CT_LOWERBYVALUE:
            ceiling->bottomHeight =
                P_GetDoublep(sec, DMU_CEILING_HEIGHT) - (coord_t) arg[2];
            ceiling->state = CS_DOWN;
            break;
        case CT_RAISEBYVALUE:
            ceiling->topHeight =
                P_GetDoublep(sec, DMU_CEILING_HEIGHT) + (coord_t) arg[2];
            ceiling->state = CS_UP;
            break;
        case CT_MOVETOVALUEMUL8: {
            coord_t destHeight = (coord_t) arg[2] * 8;
            if(arg[3]) // Going down?
                destHeight = -destHeight;
            if(P_GetDoublep(sec, DMU_CEILING_HEIGHT) <= destHeight) {
                ceiling->state = CS_UP;
                ceiling->topHeight = destHeight;
                if(FEQUAL(P_GetDoublep(sec, DMU_CEILING_HEIGHT), destHeight))
                    rtn = 0;
            }
            else if(P_GetDoublep(sec, DMU_CEILING_HEIGHT) > destHeight) {
                ceiling->state = CS_DOWN;
                ceiling->bottomHeight = destHeight;
            }
            break;
            }
#endif
        default:
#if __JHEXEN__
            rtn = 0;
#endif
            break;
        }
        ceiling->tag = xsec->tag;
        ceiling->type = type;
#if __JHEXEN__
        if(rtn) {
            SN_StartSequence((mobj_t *)P_GetPtrp(ceiling->sector, DMU_EMITTER),
                             SEQ_PLATFORM + P_ToXSector(ceiling->sector)->seqType);
        }
#endif
    }
    return rtn;
}
#if __JHEXEN__
int EV_DoCeiling(Line * /*line*/, byte *args, ceilingtype_e type)
#else
int EV_DoCeiling(Line *line, ceilingtype_e type)
#endif
{
#if __JHEXEN__
    return EV_DoCeiling2(args, (int) args[0], (float) args[1] * (1.0 / 8),
                         type);
#else
    int rtn = 0;
    switch(type) {
    case CT_CRUSHANDRAISEFAST:
# if __JDOOM__ || __JDOOM64__
    case CT_SILENTCRUSHANDRAISE:
# endif
    case CT_CRUSHANDRAISE:
        rtn = P_CeilingActivate(P_ToXLine(line)->tag);
        break;
    default:
        break;
    }
# if __JDOOM64__
    return EV_DoCeiling2(line, P_ToXLine(line)->tag, CEILSPEED, type) || rtn;
# else
    return EV_DoCeiling2(P_ToXLine(line)->tag, CEILSPEED, type) || rtn;
# endif
#endif
}
#if !__JHEXEN__
struct activateceilingparams_t {
    short tag;
    int count;
};
static int activateCeiling(thinker_t *th, void *context) {
    ceiling_t *ceiling = (ceiling_t *) th;
    activateceilingparams_t *params = (activateceilingparams_t *) context;
    if(ceiling->tag == (int) params->tag && Thinker_InStasis(&ceiling->thinker)) {
        ceiling->state = ceiling->oldState;
        Thinker_SetStasis(&ceiling->thinker, false);
        params->count++;
    }
    return false; // Continue iteration.
}
int P_CeilingActivate(short tag) {
    activateceilingparams_t params;
    params.tag = tag;
    params.count = 0;
    Thinker_Iterate(T_MoveCeiling, activateCeiling, &params);
    return params.count;
}
#endif
struct deactivateceilingparams_t {
    short tag;
    int count;
};
static int deactivateCeiling(thinker_t *th, void *context) {
    ceiling_t *ceiling = (ceiling_t *) th;
    deactivateceilingparams_t *params = (deactivateceilingparams_t *) context;
#if __JHEXEN__
    if(ceiling->tag == (int) params->tag) {
        SN_StopSequence((mobj_t *)P_GetPtrp(ceiling->sector, DMU_EMITTER));
        stopCeiling(ceiling);
        params->count++;
        return true; // Stop iteration.
    }
#else
    if(!Thinker_InStasis(&ceiling->thinker) && ceiling->tag == (int) params->tag) {
        ceiling->oldState = ceiling->state;
        Thinker_SetStasis(&ceiling->thinker, true);
        params->count++;
    }
#endif
    return false; // Continue iteration.
}
int P_CeilingDeactivate(short tag) {
    deactivateceilingparams_t params;
    params.tag = tag;
    params.count = 0;
    Thinker_Iterate(T_MoveCeiling, deactivateCeiling, &params);
    return params.count;
}
void T_Door(void *doorThinkerPtr) {
    door_t *door = (door_t *)doorThinkerPtr;
    DENG_ASSERT(door != 0);
    xsector_t *xsec = P_ToXSector(door->sector);
    result_e res;
    switch(door->state) {
    case DS_WAIT:
        if(!--door->topCountDown) {
            switch(door->type) {
#if __JDOOM64__
            case DT_INSTANTRAISE:  //jd64
                door->state = DS_DOWN;
                break;
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
# if __JHERETIC__
            case DT_BLAZEOPEN:
# else
            case DT_BLAZERAISE:
# endif
                door->state = DS_DOWN; // Time to go back down.
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOORBLAZECLOSE);
                break;
#endif
            case DT_NORMAL:
                door->state = DS_DOWN; // Time to go back down.
#if __JHEXEN__
                SN_StartSequence((mobj_t *)P_GetPtrp(door->sector, DMU_EMITTER),
                                 SEQ_DOOR_STONE + xsec->seqType);
#else
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOORCLOSING);
#endif
                break;
            case DT_CLOSE30THENOPEN:
                door->state = DS_UP;
#if !__JHEXEN__
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOOROPEN);
#endif
                break;
            default:
                break;
            }
        }
        break;
    case DS_INITIALWAIT:
        if(!--door->topCountDown) {
            switch(door->type) {
            case DT_RAISEIN5MINS:
                door->state = DS_UP;
                door->type = DT_NORMAL;
#if !__JHEXEN__
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOOROPEN);
#endif
                break;
            default:
                break;
            }
        }
        break;
    case DS_DOWN:
        res =
            T_MovePlane(door->sector, door->speed,
                        P_GetDoublep(door->sector, DMU_FLOOR_HEIGHT),
                        false, 1, -1);
        if(res == pastdest) {
#if __JHEXEN__
            SN_StopSequence((mobj_t *)P_GetPtrp(door->sector, DMU_EMITTER));
#endif
            switch(door->type) {
#if __JDOOM64__
            case DT_INSTANTRAISE:  //jd64
            case DT_INSTANTCLOSE:  //jd64
                P_ToXSector(door->sector)->specialData = 0;
                Thinker_Remove(&door->thinker); // Unlink and free.
                break;
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
# if __JHERETIC__
            case DT_BLAZEOPEN:
# else
            case DT_BLAZERAISE:
            case DT_BLAZECLOSE:
# endif
                xsec->specialData = 0;
                Thinker_Remove(&door->thinker); // Unlink and free.
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOORBLAZECLOSE);
                break;
#endif
            case DT_NORMAL:
            case DT_CLOSE:
                xsec->specialData = nullptr;
                P_NotifySectorFinished(P_ToXSector(door->sector)->tag);
                Thinker_Remove(&door->thinker); // Unlink and free.
#if __JHERETIC__
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOORCLOSED);
#endif
                break;
            case DT_CLOSE30THENOPEN:
                door->state = DS_WAIT;
                door->topCountDown = 30 * TICSPERSEC;
                break;
            default:
                break;
            }
        }
        else if(res == crushed) {
            switch(door->type) {
#if __JDOOM__ || __JDOOM64__
            case DT_BLAZECLOSE:
#endif
            case DT_CLOSE: // Do not go back up!
                break;
            default:
                door->state = DS_UP;
#if !__JHEXEN__
                S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOOROPEN);
#endif
                break;
            }
        }
        break;
    case DS_UP:
        res =
            T_MovePlane(door->sector, door->speed, door->topHeight, false, 1,
                        1);
        if(res == pastdest) {
#if __JHEXEN__
            SN_StopSequence((mobj_t *)P_GetPtrp(door->sector, DMU_EMITTER));
#endif
            switch(door->type) {
#if __JDOOM64__
            case DT_INSTANTRAISE:  //jd64
                door->state = DS_WAIT;
                /*
                 * skip topwait and began the countdown
                 * that way there won't be a big delay when the
                 * animation starts -kaiser
                 */
                door->topCountDown = 160;
                break;
#endif
#if __JDOOM__ || __JDOOM64__
            case DT_BLAZERAISE:
#endif
#if __JHERETIC__
            case DT_BLAZEOPEN:
#endif
            case DT_NORMAL:
                door->state        = DS_WAIT; // Wait at top.
                door->topCountDown = door->topWait;
                break;
#if __JDOOM__ || __JDOOM64__
            case DT_BLAZEOPEN:
#endif
            case DT_CLOSE30THENOPEN:
            case DT_OPEN:
                xsec->specialData = nullptr;
                P_NotifySectorFinished(P_ToXSector(door->sector)->tag);
                Thinker_Remove(&door->thinker); // Unlink and free.
#if __JHERETIC__
                S_StopSound(0, (mobj_t *) P_GetPtrp(door->sector, DMU_CEILING_EMITTER));
#endif
                break;
            default:
                break;
            }
        }
        break;
    default:
        break;
    }
}
void door_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteByte(writer, (byte) type);
    Writer_WriteInt32(writer, P_ToIndex(sector));
    Writer_WriteInt16(writer, (int)topHeight);
    Writer_WriteInt32(writer, FLT2FIX(speed));
    Writer_WriteInt32(writer, state);
    Writer_WriteInt32(writer, topWait);
    Writer_WriteInt32(writer, topCountDown);
}
int door_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
#if __JHEXEN__
    if(mapVersion >= 4)
#else
    if(mapVersion >= 5)
#endif
    {   // Note: the thinker class byte has already been read.
        /*int ver =*/ Reader_ReadByte(reader); // version byte.
        type          = doortype_e(Reader_ReadByte(reader));
        sector        = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        topHeight     = (float) Reader_ReadInt16(reader);
        speed         = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        state         = doorstate_e(Reader_ReadInt32(reader));
        topWait       = Reader_ReadInt32(reader);
        topCountDown  = Reader_ReadInt32(reader);
    }
    else {
        byte junk[16]; // sizeof thinker_t
        Reader_Read(reader, junk, 16);
#if __JHEXEN__
        sector        = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        type          = doortype_e(Reader_ReadInt32(reader));
#else
        type          = doortype_e(Reader_ReadInt32(reader));
        sector        = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
#endif
        topHeight     = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        speed         = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        state         = doorstate_e(Reader_ReadInt32(reader));
        topWait       = Reader_ReadInt32(reader);
        topCountDown  = Reader_ReadInt32(reader);
    }
    P_ToXSector(sector)->specialData = this;
    thinker.function = T_Door;
    return true; // Add this thinker.
}
static int EV_DoDoor2(int tag, float speed, int topwait, doortype_e type) {
    iterlist_t *list = P_GetSectorIterListForTag(tag, false);
    if(!list) return 0;
    int rtn = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *) IterList_MoveIterator(list))) {
        xsector_t *xsec = P_ToXSector(sec);
        if(xsec->specialData)
            continue;
        rtn = 1;
        door_t *door = (door_t *)Z_Calloc(sizeof(*door), PU_MAP, 0);
        door->thinker.function = T_Door;
        Thinker_Add(&door->thinker);
        xsec->specialData = door;
        door->sector  = sec;
        door->type    = type;
        door->topWait = topwait;
        door->speed   = speed;
#if __JHEXEN__
        int sound = SEQ_DOOR_STONE + P_ToXSector(door->sector)->seqType;
#else
        int sound = 0;
#endif
        switch(type) {
#if __JDOOM__ || __JDOOM64__
        case DT_BLAZECLOSE:
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
            door->topHeight -= 4;
            door->state = DS_DOWN;
            door->speed *= 4;
            sound = SFX_DOORBLAZECLOSE;
            break;
#endif
        case DT_CLOSE:
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
            door->topHeight -= 4;
            door->state = DS_DOWN;
#if !__JHEXEN__
            sound = SFX_DOORCLOSING;
#endif
            break;
        case DT_CLOSE30THENOPEN:
            door->topHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
            door->state = DS_DOWN;
#if !__JHEXEN__
            sound = SFX_DOORCLOSING;
#endif
            break;
#if __JDOOM__ || __JDOOM64__
        case DT_BLAZERAISE:
#endif
#if !__JHEXEN__
        case DT_BLAZEOPEN:
            door->state = DS_UP;
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
            door->topHeight -= 4;
# if __JHERETIC__
            door->speed *= 3;
# else
            door->speed *= 4;
# endif
            if(!FEQUAL(door->topHeight, P_GetDoublep(sec, DMU_CEILING_HEIGHT)))
                sound = SFX_DOORBLAZEOPEN;
            break;
#endif
        case DT_NORMAL:
        case DT_OPEN:
            door->state = DS_UP;
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
            door->topHeight -= 4;
#if !__JHEXEN__
            if(!FEQUAL(door->topHeight, P_GetDoublep(sec, DMU_CEILING_HEIGHT)))
                sound = SFX_DOOROPEN;
#endif
            break;
        default:
            break;
        }
#if __JHEXEN__
        SN_StartSequence((mobj_t *)P_GetPtrp(door->sector, DMU_EMITTER), sound);
#else
        if(sound)
            S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), sound);
#endif
    }
    return rtn;
}
#if __JHEXEN__
int EV_DoDoor(Line * /*line*/, byte *args, doortype_e type)
{
    return EV_DoDoor2((int) args[0], (float) args[1] * (1.0 / 8),
                      (int) args[2], type);
}
#else
int EV_DoDoor(Line *line, doortype_e type) {
    return EV_DoDoor2(P_ToXLine(line)->tag, DOORSPEED, DOORWAIT, type);
}
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
static void sendNeedKeyMessage(player_t *p, textenum_t msgTxt, int keyNum) {
    char buf[160], tmp[2];
    char const *in;
    buf[0] = 0;
    tmp[1] = 0;
    in = GET_TXT(msgTxt);
    for(; *in; in++) {
        if(in[0] == '%') {
            if(in[1] == '1') {
                strcat(buf, GET_TXT(TXT_KEY1 + keyNum));
                in++;
                continue;
            }
            if(in[1] == '%')
                in++;
        }
        tmp[0] = *in;
        strcat(buf, tmp);
    }
    P_SetMessage(p, buf);
}
#endif
#if __JDOOM__ || __JDOOM64__
/**
 * Checks whether the given line is a locked door.
 * If locked and the player IS ABLE to open it, return @c true.
 * If locked and the player IS NOT ABLE to open it, send an appropriate
 * message and play a sound before returning @c false.
 * Else, NOT a locked door and can be opened, return @c true.
 */
static dd_bool tryLockedDoor(Line *line, player_t *p) {
    xline_t *xline = P_ToXLine(line);
    if(!p || !xline)
        return false;
    switch(xline->special) {
    case 99:                    // Blue Lock
    case 133:
        if(!p->keys[KT_BLUECARD] && !p->keys[KT_BLUESKULL]) {
            sendNeedKeyMessage(p, TXT_PD_BLUEO, 0);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
    case 134:                   // Red Lock
    case 135:
        if(!p->keys[KT_REDCARD] && !p->keys[KT_REDSKULL]) {
            sendNeedKeyMessage(p, TXT_PD_REDO, 2);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
    case 136:                   // Yellow Lock
    case 137:
        if(!p->keys[KT_YELLOWCARD] && !p->keys[KT_YELLOWSKULL]) {
            sendNeedKeyMessage(p, TXT_PD_YELLOWO, 1);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
# if __JDOOM64__
    case 343:
        if(!P_InventoryCount(p - players, IIT_DEMONKEY1)) {
            P_SetMessage(p, PD_OPNPOWERUP);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
    case 344:
        if(!P_InventoryCount(p - players, IIT_DEMONKEY2)) {
            P_SetMessage(p, PD_OPNPOWERUP);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
    case 345:
        if(!P_InventoryCount(p - players, IIT_DEMONKEY3)) {
            P_SetMessage(p, PD_OPNPOWERUP);
            S_StartSound(SFX_DOORLOCKED, p->plr->mo);
            return false;
        }
        break;
# endif
    default:
        break;
    }
    return true;
}
#endif
/**
 * Checks whether the given line is a locked manual door.
 * If locked and the player IS ABLE to open it, return @c true.
 * If locked and the player IS NOT ABLE to open it, send an appropriate
 * message and play a sound before returning @c false.
 * Else, NOT a locked door and can be opened, return @c true.
 */
static dd_bool tryLockedManualDoor(Line *line, mobj_t *mo) {
    xline_t *xline = P_ToXLine(line);
#if !__JHEXEN__
    int keyNum = -1;
    textenum_t msgTxt = NUMTEXT;
    sfxenum_t sfxNum = SFX_NONE;
#endif
    if(!mo || !xline)
        return false;
#if !__JHEXEN__
    player_t *p = mo->player;
    switch(xline->special) {
    case 26:
    case 32:
# if __JDOOM64__
    case 525: // jd64
# endif
        if(!p)
            return false;
# if __JHERETIC__
        if(!p->keys[KT_BLUE]) {
            msgTxt = TXT_TXT_NEEDBLUEKEY;
            keyNum = 2;
            sfxNum = SFX_DOORLOCKED;
        }
# else
        if(!p->keys[KT_BLUECARD] && !p->keys[KT_BLUESKULL]) {
            msgTxt = TXT_PD_BLUEK;
            keyNum = 0;
            sfxNum = SFX_DOORLOCKED;
        }
# endif
        break;
    case 27:
    case 34:
# if __JDOOM64__
    case 526: // jd64
# endif
        if(!p)
            return false;
# if __JHERETIC__
        if(!p->keys[KT_YELLOW]) {
            msgTxt = TXT_TXT_NEEDYELLOWKEY;
            keyNum = 0;
            sfxNum = SFX_DOORLOCKED;
        }
# else
        if(!p->keys[KT_YELLOWCARD] && !p->keys[KT_YELLOWSKULL]) {
            msgTxt = TXT_PD_YELLOWK;
            keyNum = 1;
            sfxNum = SFX_DOORLOCKED;
        }
# endif
        break;
    case 28:
    case 33:
# if __JDOOM64__
    case 527: // jd64
# endif
        if(!p)
            return false;
# if __JHERETIC__
        if(!p->keys[KT_GREEN]) {
            msgTxt = TXT_TXT_NEEDGREENKEY;
            keyNum = 1;
            sfxNum = SFX_DOORLOCKED;
        }
# else
        if(!p->keys[KT_REDCARD] && !p->keys[KT_REDSKULL]) {
            msgTxt = TXT_PD_REDK;
            keyNum = 2;
            sfxNum = SFX_DOORLOCKED;
        }
# endif
        break;
    default:
        break;
    }
    if(keyNum != -1)
    {   // A key is required.
        sendNeedKeyMessage(p, msgTxt, keyNum);
        S_StartSound(sfxNum, p->plr->mo);
        return false;
    }
#endif
    return true;
}
/**
 * Move a locked door up/down.
 */
#if __JDOOM__ || __JDOOM64__
int EV_DoLockedDoor(Line *line, doortype_e type, mobj_t *thing) {
    if(!tryLockedDoor(line, thing->player))
        return 0;
    return EV_DoDoor(line, type);
}
#endif
/**
 * Open a door manually, no tag value.
 */
dd_bool EV_VerticalDoor(Line *line, mobj_t *mo) {
    Sector *sec = (Sector *)P_GetPtrp(line, DMU_BACK_SECTOR);
    if(!sec) return false;
    if(!tryLockedManualDoor(line, mo))
        return false; // Mobj cannot open this door.
    xsector_t *xsec = P_ToXSector(sec);
    xline_t *xline  = P_ToXLine(line);
    if(xsec->specialData) {
#if __JHEXEN__
        return false;
#else
        door_t *door = (door_t *)xsec->specialData;
        switch(xline->special) {
        case 1:
        case 26:
        case 27:
        case 28:
# if __JDOOM__ || __JDOOM64__
        case 117:
# endif
# if __JDOOM64__
        case 525: // jd64
        case 526: // jd64
        case 527: // jd64
# endif
            if(door->state == DS_DOWN)
                door->state = DS_UP; // Go back up.
            else {
                if(!mo->player)
                    return false; // Bad guys never close doors.
                door->state = DS_DOWN; // Start going down immediately.
            }
            return false;
        default:
            break;
        }
#endif
    }
    door_t *door = (door_t *) Z_Calloc(sizeof(*door), PU_MAP, 0);
    door->thinker.function = T_Door;
    Thinker_Add(&door->thinker);
    xsec->specialData = door;
    door->sector      = sec;
    door->state       = DS_UP;
#if __JHEXEN__
    SN_StartSequence((mobj_t *)P_GetPtrp(door->sector, DMU_EMITTER),
                     SEQ_DOOR_STONE + P_ToXSector(door->sector)->seqType);
#else
    switch(xline->special) {
# if __JDOOM__ || __JDOOM64__
    case 117:
    case 118:
#  if __JDOOM64__
    case 525: // jd64
    case 526: // jd64
    case 527: // jd64
#  endif
        S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOORBLAZEOPEN);
        break;
# endif
    case 1:
    case 31:
        S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOOROPEN);
        break;
    default:
        S_PlaneSound((Plane *)P_GetPtrp(door->sector, DMU_CEILING_PLANE), SFX_DOOROPEN);
        break;
    }
#endif
    switch(xline->special) {
#if __JHEXEN__
    case 11:
        door->type = DT_OPEN;
        door->speed = (float) xline->arg2 * (1.0 / 8);
        door->topWait = (int) xline->arg3;
        xline->special = 0;
        break;
#else
    case 31:
    case 32:
    case 33:
    case 34:
        door->type = DT_OPEN;
        door->speed = DOORSPEED;
        door->topWait = DOORWAIT;
        xline->special = 0;
        break;
#endif
#if __JHEXEN__
    case 12:
    case 13:
        door->type = DT_NORMAL;
        door->speed = (float) xline->arg2 * (1.0 / 8);
        door->topWait = (int) xline->arg3;
        break;
#else
    case 1:
    case 26:
    case 27:
    case 28:
        door->type = DT_NORMAL;
        door->speed = DOORSPEED;
        door->topWait = DOORWAIT;
        break;
#endif
#if __JDOOM__ || __JDOOM64__
    case 117:                   // blazing door raise
# if __JDOOM64__
    case 525: // jd64
    case 526: // jd64
    case 527: // jd64
# endif
        door->type = DT_BLAZERAISE;
        door->speed = DOORSPEED * 4;
        door->topWait = DOORWAIT;
        break;
    case 118:                   // blazing door open
        door->type = DT_BLAZEOPEN;
        door->speed = DOORSPEED * 4;
        door->topWait = DOORWAIT;
        xline->special = 0;
        break;
#endif
    default:
#if __JHEXEN__
        door->type = DT_NORMAL;
        door->speed = (float) xline->arg2 * (1.0 / 8);
        door->topWait = (int) xline->arg3;
#else
        door->speed = DOORSPEED;
        door->topWait = DOORWAIT;
#endif
        break;
    }
    P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
    door->topHeight -= 4;
    return true;
}
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
void P_SpawnDoorCloseIn30(Sector *sec) {
    door_t *door = (door_t *)Z_Calloc(sizeof(*door), PU_MAP, 0);
    door->thinker.function = T_Door;
    Thinker_Add(&door->thinker);
    P_ToXSector(sec)->specialData = door;
    P_ToXSector(sec)->special = 0;
    door->sector       = sec;
    door->state        = DS_WAIT;
    door->type         = DT_NORMAL;
    door->speed        = DOORSPEED;
    door->topCountDown = 30 * TICSPERSEC;
    door->topHeight    = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
}
void P_SpawnDoorRaiseIn5Mins(Sector *sec) {
    door_t *door = (door_t *)Z_Calloc(sizeof(*door), PU_MAP, 0);
    door->thinker.function = T_Door;
    Thinker_Add(&door->thinker);
    P_ToXSector(sec)->specialData = door;
    P_ToXSector(sec)->special = 0;
    door->sector       = sec;
    door->state        = DS_INITIALWAIT;
    door->type         = DT_RAISEIN5MINS;
    door->speed        = DOORSPEED;
    P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &door->topHeight);
    door->topHeight   -= 4;
    door->topWait      = DOORWAIT;
    door->topCountDown = 5 * 60 * TICSPERSEC;
}
#endif
result_e T_MovePlane(Sector *sector, float speed, coord_t dest, int crush,
    int isCeiling, int direction) {
    dd_bool flag;
    coord_t lastpos;
    coord_t floorheight, ceilingheight;
    int ptarget = (isCeiling? DMU_CEILING_TARGET_HEIGHT : DMU_FLOOR_TARGET_HEIGHT);
    int pspeed = (isCeiling? DMU_CEILING_SPEED : DMU_FLOOR_SPEED);
    P_SetDoublep(sector, ptarget, dest);
    P_SetFloatp(sector, pspeed, speed);
    floorheight = P_GetDoublep(sector, DMU_FLOOR_HEIGHT);
    ceilingheight = P_GetDoublep(sector, DMU_CEILING_HEIGHT);
    switch(isCeiling) {
    case 0:
        switch(direction) {
        case -1:
            if(floorheight - speed < dest) {
                lastpos = floorheight;
                P_SetDoublep(sector, DMU_FLOOR_HEIGHT, dest);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
                    P_SetDoublep(sector, DMU_FLOOR_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
                    P_ChangeSector(sector, crush);
                }
#if __JHEXEN__
                P_SetFloatp(sector, pspeed, 0);
#endif
                return pastdest;
            }
            else {
                lastpos = floorheight;
                P_SetDoublep(sector, DMU_FLOOR_HEIGHT, floorheight - speed);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
                    P_SetDoublep(sector, DMU_FLOOR_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
#if __JHEXEN__
                    P_SetFloatp(sector, pspeed, 0);
#endif
                    P_ChangeSector(sector, crush);
                    return crushed;
                }
            }
            break;
        case 1:
            if(floorheight + speed > dest) {
                lastpos = floorheight;
                P_SetDoublep(sector, DMU_FLOOR_HEIGHT, dest);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
                    P_SetDoublep(sector, DMU_FLOOR_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
                    P_ChangeSector(sector, crush);
                }
#if __JHEXEN__
                P_SetFloatp(sector, pspeed, 0);
#endif
                return pastdest;
            }
            else {
                lastpos = floorheight;
                P_SetDoublep(sector, DMU_FLOOR_HEIGHT, floorheight + speed);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
#if !__JHEXEN__
                    if(crush)
                        return crushed;
#endif
                    P_SetDoublep(sector, DMU_FLOOR_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
#if __JHEXEN__
                    P_SetFloatp(sector, pspeed, 0);
#endif
                    P_ChangeSector(sector, crush);
                    return crushed;
                }
            }
            break;
        default:
            break;
        }
        break;
    case 1:
        switch(direction) {
        case -1:
            if(ceilingheight - speed < dest) {
                lastpos = ceilingheight;
                P_SetDoublep(sector, DMU_CEILING_HEIGHT, dest);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
                    P_SetDoublep(sector, DMU_CEILING_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
                    P_ChangeSector(sector, crush);
                }
#if __JHEXEN__
                P_SetFloatp(sector, pspeed, 0);
#endif
                return pastdest;
            }
            else {
                lastpos = ceilingheight;
                P_SetDoublep(sector, DMU_CEILING_HEIGHT, ceilingheight - speed);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
#if !__JHEXEN__
                    if(crush)
                        return crushed;
#endif
                    P_SetDoublep(sector, DMU_CEILING_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
#if __JHEXEN__
                    P_SetFloatp(sector, pspeed, 0);
#endif
                    P_ChangeSector(sector, crush);
                    return crushed;
                }
            }
            break;
        case 1:
            if(ceilingheight + speed > dest) {
                lastpos = ceilingheight;
                P_SetDoublep(sector, DMU_CEILING_HEIGHT, dest);
                flag = P_ChangeSector(sector, crush);
                if(flag) {
                    P_SetDoublep(sector, DMU_CEILING_HEIGHT, lastpos);
                    P_SetDoublep(sector, ptarget, lastpos);
                    P_ChangeSector(sector, crush);
                }
#if __JHEXEN__
                P_SetFloatp(sector, pspeed, 0);
#endif
                return pastdest;
            }
            else {
                lastpos = ceilingheight;
                P_SetDoublep(sector, DMU_CEILING_HEIGHT, ceilingheight + speed);
                flag = P_ChangeSector(sector, crush);
            }
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
    return ok;
}
/**
 * Move a floor to it's destination (up or down).
 */
void T_MoveFloor(void *floorThinkerPtr) {
    floor_t *floor = (floor_t *)floorThinkerPtr;
    result_e res;
#if __JHEXEN__
    if(floor->resetDelayCount) {
        floor->resetDelayCount--;
        if(!floor->resetDelayCount) {
            floor->floorDestHeight = floor->resetHeight;
            floor->state = ((floor->state == FS_UP)? FS_DOWN : FS_UP);
            floor->resetDelay = 0;
            floor->delayCount = 0;
            floor->delayTotal = 0;
        }
    }
    if(floor->delayCount) {
        floor->delayCount--;
        if(!floor->delayCount && floor->material) {
            P_SetPtrp(floor->sector, DMU_FLOOR_MATERIAL, floor->material);
        }
        return;
    }
#endif
    res =
        T_MovePlane(floor->sector, floor->speed, floor->floorDestHeight,
                    floor->crush, 0, floor->state);
#if __JHEXEN__
    if(floor->type == FT_RAISEBUILDSTEP) {
        if((floor->state == FS_UP &&
            P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) >= floor->stairsDelayHeight) ||
           (floor->state == FS_DOWN &&
            P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) <= floor->stairsDelayHeight)) {
            floor->delayCount = floor->delayTotal;
            floor->stairsDelayHeight += floor->stairsDelayHeightDelta;
        }
    }
#endif
#if !__JHEXEN__
    if(!(mapTime & 7))
        S_PlaneSound((Plane *)P_GetPtrp(floor->sector, DMU_FLOOR_PLANE), SFX_FLOORMOVE);
#endif
    if(res == pastdest) {
        xsector_t *xsec = P_ToXSector(floor->sector);
        P_SetFloatp(floor->sector, DMU_FLOOR_SPEED, 0);
#if __JHEXEN__
        SN_StopSequence((mobj_t *)P_GetPtrp(floor->sector, DMU_EMITTER));
#else
#  if __JHERETIC__
        if(floor->type == FT_RAISEBUILDSTEP)
#  endif
            S_PlaneSound((Plane *)P_GetPtrp(floor->sector, DMU_FLOOR_PLANE), SFX_PSTOP);
#endif
#if __JHEXEN__
        if(floor->delayTotal)
            floor->delayTotal = 0;
        if(floor->resetDelay) {
            return;
        }
#endif
        xsec->specialData = NULL;
#if __JHEXEN__
        if(floor->material) {
            P_SetPtrp(floor->sector, DMU_FLOOR_MATERIAL, floor->material);
        }
#else
        if(floor->state == FS_UP) {
            switch(floor->type) {
            case FT_RAISEDONUT:
                xsec->special = floor->newSpecial;
                P_SetPtrp(floor->sector, DMU_FLOOR_MATERIAL, floor->material);
                break;
            default:
                break;
            }
        }
        else if(floor->state == FS_DOWN) {
            switch(floor->type) {
            case FT_LOWERANDCHANGE:
                xsec->special = floor->newSpecial;
                P_SetPtrp(floor->sector, DMU_FLOOR_MATERIAL, floor->material);
                break;
            default:
                break;
            }
        }
#endif
        P_NotifySectorFinished(P_ToXSector(floor->sector)->tag);
        Thinker_Remove(&floor->thinker);
    }
}
void floor_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 3); // Write a version byte.
    Writer_WriteByte(writer, (byte) type);
    Writer_WriteInt32(writer, P_ToIndex(sector));
    Writer_WriteByte(writer, (byte) crush);
    Writer_WriteInt32(writer, (int) state);
    Writer_WriteInt32(writer, newSpecial);
    Writer_WriteInt16(writer, msw->serialIdFor(material));
    Writer_WriteInt16(writer, (int) floorDestHeight);
    Writer_WriteInt32(writer, FLT2FIX(speed));
#if __JHEXEN__
    Writer_WriteInt32(writer, delayCount);
    Writer_WriteInt32(writer, delayTotal);
    Writer_WriteInt32(writer, FLT2FIX(stairsDelayHeight));
    Writer_WriteInt32(writer, FLT2FIX(stairsDelayHeightDelta));
    Writer_WriteInt32(writer, FLT2FIX(resetHeight));
    Writer_WriteInt16(writer, resetDelay);
    Writer_WriteInt16(writer, resetDelayCount);
#endif
}
int floor_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
#if __JHEXEN__
    if(mapVersion >= 4)
#else
    if(mapVersion >= 5)
#endif
    {   // Note: the thinker class byte has already been read.
        byte ver = Reader_ReadByte(reader); // version byte.
        type                   = floortype_e(Reader_ReadByte(reader));
        sector                 = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        crush                  = dd_bool(Reader_ReadByte(reader));
        state                  = floorstate_e(Reader_ReadInt32(reader));
        newSpecial             = Reader_ReadInt32(reader);
        if(ver >= 2) {
            material = (world_Material *) msr->material(Reader_ReadInt16(reader), 0);
        }
        else {
            de::Uri uri("Flats:", CentralLumpIndex()[Reader_ReadInt16(reader)].name().fileNameWithoutExtension());
            material = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(reinterpret_cast<uri_s *>(&uri)));
        }
        floorDestHeight        = (float) Reader_ReadInt16(reader);
        speed                  = FIX2FLT(Reader_ReadInt32(reader));
#if __JHEXEN__
        delayCount             = Reader_ReadInt32(reader);
        delayTotal             = Reader_ReadInt32(reader);
        stairsDelayHeight      = FIX2FLT(Reader_ReadInt32(reader));
        stairsDelayHeightDelta = FIX2FLT(Reader_ReadInt32(reader));
        resetHeight            = FIX2FLT(Reader_ReadInt32(reader));
        resetDelay             = Reader_ReadInt16(reader);
        resetDelayCount        = Reader_ReadInt16(reader);
#endif
    }
    else {
        byte junk[16]; // sizeof thinker_t
        Reader_Read(reader, junk, 16);
#if __JHEXEN__
        sector                 = (Sector *)P_ToPtr(DMU_SECTOR, (int) Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        type                   = floortype_e(Reader_ReadInt32(reader));
        crush                  = Reader_ReadInt32(reader);
#else
        type                   = floortype_e(Reader_ReadInt32(reader));
        crush                  = Reader_ReadInt32(reader);
        sector                 = (Sector *)P_ToPtr(DMU_SECTOR, (int) Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
#endif
        state                  = floorstate_e(Reader_ReadInt32(reader));
        newSpecial             = Reader_ReadInt32(reader);
        de::Uri uri("Flats:", CentralLumpIndex()[Reader_ReadInt16(reader)].name().fileNameWithoutExtension());
        material               = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(reinterpret_cast<uri_s *>(&uri)));
        floorDestHeight        = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        speed                  = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
#if __JHEXEN__
        delayCount             = Reader_ReadInt32(reader);
        delayTotal             = Reader_ReadInt32(reader);
        stairsDelayHeight      = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        stairsDelayHeightDelta = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        resetHeight            = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        resetDelay             = Reader_ReadInt16(reader);
        resetDelayCount        = Reader_ReadInt16(reader);
        /*textureChange        =*/ Reader_ReadByte(reader);
#endif
    }
    P_ToXSector(sector)->specialData = this;
    thinker.function = T_MoveFloor;
    return true; // Add this thinker.
}
struct findlineinsectorsmallestbottommaterialparams_t {
    Sector *baseSec;
    int minSize;
    Line *foundLine;
};
int findLineInSectorSmallestBottomMaterial(void *ptr, void *context) {
    Line *li = (Line *) ptr;
    findlineinsectorsmallestbottommaterialparams_t *params = (findlineinsectorsmallestbottommaterialparams_t *) context;
    Sector *frontSec = (Sector *)P_GetPtrp(li, DMU_FRONT_SECTOR);
    Sector *backSec  = (Sector *)P_GetPtrp(li, DMU_BACK_SECTOR);
    if(frontSec && backSec) {
        Side *side = (Side *)P_GetPtrp(li, DMU_FRONT);
        world_Material *mat = (world_Material *)P_GetPtrp(side, DMU_BOTTOM_MATERIAL);
        /**
         * Emulate DOOM.exe behaviour. In the instance where no material is
         * present, the height is taken from the very first texture.
         */
        if(!mat) {
            uri_s *textureUrn = Uri_NewWithPath2("urn:Textures:0", RC_NULL);
            mat = DD_MaterialForTextureUri(textureUrn);
            Uri_Delete(textureUrn);
        }
        if(mat) {
            int height = P_GetIntp(mat, DMU_HEIGHT);
            if(height < params->minSize) {
                params->minSize = height;
                params->foundLine = li;
            }
        }
        side = (Side *)P_GetPtrp(li, DMU_BACK);
        mat  = (world_Material *)P_GetPtrp(side, DMU_BOTTOM_MATERIAL);
        if(!mat) {
            uri_s *textureUrn = Uri_NewWithPath2("urn:Textures:0", RC_NULL);
            mat = DD_MaterialForTextureUri(textureUrn);
            Uri_Delete(textureUrn);
        }
        if(mat) {
            int height = P_GetIntp(mat, DMU_HEIGHT);
            if(height < params->minSize) {
                params->minSize = height;
                params->foundLine = li;
            }
        }
    }
    return false; // Continue iteration.
}
Line* P_FindLineInSectorSmallestBottomMaterial(Sector *sec, int *val) {
    findlineinsectorsmallestbottommaterialparams_t params;
    params.baseSec = sec;
    params.minSize = DDMAXINT;
    params.foundLine = NULL;
    P_Iteratep(sec, DMU_LINE, findLineInSectorSmallestBottomMaterial, &params);
    if(val)
        *val = params.minSize;
    return params.foundLine;
}
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
/**
 * Find the first sector which shares a border to the specified sector
 * and whose floor height matches that specified.
 *
 * @note Behaviour here is dependant upon the order of the sector-linked
 * Lines list. This is necessary to emulate the flawed algorithm used in
 * DOOM.exe In addition, this algorithm was further broken in Heretic as
 * the test which compares floor heights was removed.
 *
 * @warning DO NOT USE THIS ANYWHERE ELSE!
 */
typedef struct findfirstneighbouratfloorheightparams_s {
    Sector* baseSec;
    coord_t height;
    Sector* foundSec;
} findfirstneighbouratfloorheightparams_t;
static int findFirstNeighbourAtFloorHeight(void* ptr, void* context) {
    Line* ln = (Line*) ptr;
    findfirstneighbouratfloorheightparams_t* params =
        (findfirstneighbouratfloorheightparams_t*) context;
    Sector* other;
    other = P_GetNextSector(ln, params->baseSec);
# if __JDOOM__ || __JDOOM64__
    if(other && FEQUAL(P_GetDoublep(other, DMU_FLOOR_HEIGHT), params->height))
# elif __JHERETIC__
    if(other)
# endif
    {
        params->foundSec = other;
        return true; // Stop iteration.
    }
    return false; // Continue iteration.
}
static Sector *findSectorSurroundingAtFloorHeight(Sector *sec, coord_t height) {
    findfirstneighbouratfloorheightparams_t params;
    params.baseSec = sec;
    params.foundSec = NULL;
    params.height = height;
    P_Iteratep(sec, DMU_LINE, findFirstNeighbourAtFloorHeight, &params);
    return params.foundSec;
}
#endif
#if __JHEXEN__
int EV_DoFloor(Line * /*line*/, byte* args, floortype_e floortype)
#else
int EV_DoFloor(Line *line, floortype_e floortype)
#endif
{
#if !__JHEXEN__
    Sector *frontsector;
#endif
#if __JHEXEN__
    int tag = (int) args[0];
#else
    int tag = P_ToXLine(line)->tag;
#endif
#if __JDOOM64__
    coord_t bitmipL = 0, bitmipR = 0;
    Side *front = (Side *)P_GetPtrp(line, DMU_FRONT);
    Side *back  = (Side *)P_GetPtrp(line, DMU_BACK);
    bitmipL = P_GetDoublep(front, DMU_MIDDLE_MATERIAL_OFFSET_X);
    if(back)
        bitmipR = P_GetDoublep(back, DMU_MIDDLE_MATERIAL_OFFSET_X);
#endif
    iterlist_t *list = P_GetSectorIterListForTag(tag, false);
    if(!list) return 0;
    int rtn = 0;
    floor_t *floor = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        xsector_t *xsec = P_ToXSector(sec);
        if(xsec->specialData)
            continue;
        rtn = 1;
        floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
        floor->thinker.function = T_MoveFloor;
        Thinker_Add(&floor->thinker);
        xsec->specialData = floor;
        floor->type = floortype;
        floor->crush = false;
#if __JHEXEN__
        floor->speed = (float) args[1] * (1.0 / 8);
        if(floortype == FT_LOWERMUL8INSTANT ||
           floortype == FT_RAISEMUL8INSTANT) {
            floor->speed = 2000;
        }
#endif
        switch(floortype) {
        case FT_LOWER:
            floor->state = FS_DOWN;
            floor->sector = sec;
#if !__JHEXEN__
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 4;
# endif
#endif
            P_FindSectorSurroundingHighestFloor(sec, -500, &floor->floorDestHeight);
            break;
        case FT_LOWERTOLOWEST:
            floor->state = FS_DOWN;
            floor->sector = sec;
#if !__JHEXEN__
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 4;
# endif
#endif
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &floor->floorDestHeight);
            break;
#if __JHEXEN__
        case FT_LOWERBYVALUE:
            floor->state = FS_DOWN;
            floor->sector = sec;
            floor->floorDestHeight =
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT) - (coord_t) args[2];
            break;
        case FT_LOWERMUL8INSTANT:
        case FT_LOWERBYVALUEMUL8:
            floor->state = FS_DOWN;
            floor->sector = sec;
            floor->floorDestHeight =
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT) - (coord_t) args[2] * 8;
            break;
#endif
#if !__JHEXEN__
        case FT_LOWERTURBO:
            floor->state = FS_DOWN;
            floor->sector = sec;
            floor->speed = FLOORSPEED * 4;
            P_FindSectorSurroundingHighestFloor(sec, -500, &floor->floorDestHeight);
# if __JHERETIC__
            floor->floorDestHeight += 8;
# else
            if(!FEQUAL(floor->floorDestHeight, P_GetDoublep(sec, DMU_FLOOR_HEIGHT)))
                floor->floorDestHeight += 8;
# endif
            break;
#endif
#if __JDOOM64__
        case FT_TOHIGHESTPLUS8: // jd64
            floor->state = FS_DOWN;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
            P_FindSectorSurroundingHighestFloor(sec, -500, &floor->floorDestHeight);
            if(!FEQUAL(floor->floorDestHeight, P_GetDoublep(sec, DMU_FLOOR_HEIGHT)))
                floor->floorDestHeight += 8;
            break;
        case FT_TOHIGHESTPLUSBITMIP: // jd64
            if(bitmipR > 0) {
                floor->state = FS_DOWN;
                floor->sector = sec;
                floor->speed = FLOORSPEED * bitmipL;
                P_FindSectorSurroundingHighestFloor(sec, -500, &floor->floorDestHeight);
                if(!FEQUAL(floor->floorDestHeight, P_GetDoublep(sec, DMU_FLOOR_HEIGHT)))
                    floor->floorDestHeight += bitmipR;
            }
            else {
                floor->state = FS_UP;
                floor->sector = sec;
                floor->speed = FLOORSPEED * bitmipL;
                floor->floorDestHeight =
                    P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) - bitmipR;
            }
            break;
        case FT_CUSTOMCHANGESEC: // jd64
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED * 16;
            floor->floorDestHeight = P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT);
            P_ToXSector(sec)->special = bitmipR;
            break;
#endif
        case FT_RAISEFLOORCRUSH:
#if __JHEXEN__
            floor->crush = (int) args[2]; // arg[2] = crushing value
#else
            floor->crush = true;
#endif
            floor->state = FS_UP;
            floor->sector = sec;
#if !__JHEXEN__
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 4;
# endif
#endif
#if __JHEXEN__
            floor->floorDestHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT)-8;
#else
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &floor->floorDestHeight);
            if(floor->floorDestHeight > P_GetDoublep(sec, DMU_CEILING_HEIGHT))
                floor->floorDestHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
            floor->floorDestHeight -= 8 * (floortype == FT_RAISEFLOORCRUSH);
#endif
            break;
        case FT_RAISEFLOOR:
            floor->state = FS_UP;
            floor->sector = sec;
#if !__JHEXEN__
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 4;
# endif
#endif
            P_FindSectorSurroundingLowestCeiling(sec, (coord_t) MAXINT, &floor->floorDestHeight);
            if(floor->floorDestHeight > P_GetDoublep(sec, DMU_CEILING_HEIGHT))
                floor->floorDestHeight = P_GetDoublep(sec, DMU_CEILING_HEIGHT);
#if !__JHEXEN__
            floor->floorDestHeight -= 8 * (floortype == FT_RAISEFLOORCRUSH);
#endif
            break;
#if __JDOOM__ || __JDOOM64__
        case FT_RAISEFLOORTURBO:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED * 4;
# if __JDOOM64__
            floor->speed *= 2;
# endif
            {
            coord_t floorHeight, nextFloor;
            floorHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
            if(P_FindSectorSurroundingNextHighestFloor(sec, floorHeight, &nextFloor))
                floor->floorDestHeight = nextFloor;
            else
                floor->floorDestHeight = floorHeight;
            }
            break;
#endif
        case FT_RAISEFLOORTONEAREST:
            floor->state = FS_UP;
            floor->sector = sec;
#if !__JHEXEN__
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 8;
# endif
#endif
            {
            coord_t floorHeight, nextFloor;
            floorHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
            if(P_FindSectorSurroundingNextHighestFloor(sec, floorHeight, &nextFloor))
                floor->floorDestHeight = nextFloor;
            else
                floor->floorDestHeight = floorHeight;
            }
            break;
#if __JHEXEN__
        case FT_RAISEFLOORBYVALUE:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->floorDestHeight =
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + (coord_t) args[2];
            break;
        case FT_RAISEMUL8INSTANT:
        case FT_RAISEBYVALUEMUL8:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->floorDestHeight =
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + (coord_t) args[2] * 8;
            break;
        case FT_TOVALUEMUL8:
            floor->sector = sec;
            floor->floorDestHeight = (coord_t) args[2] * 8;
            if(args[3])
                floor->floorDestHeight = -floor->floorDestHeight;
            if(floor->floorDestHeight > P_GetDoublep(sec, DMU_FLOOR_HEIGHT))
                floor->state = FS_UP;
            else if(floor->floorDestHeight < P_GetDoublep(sec, DMU_FLOOR_HEIGHT))
                floor->state = FS_DOWN;
            else
                rtn = 0; // Already at lowest position.
            break;
#endif
#if !__JHEXEN__
        case FT_RAISE24:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 8;
# endif
            floor->floorDestHeight =
                P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) + 24;
            break;
#endif
#if !__JHEXEN__
        case FT_RAISE24ANDCHANGE:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
# if __JDOOM64__
            floor->speed *= 8;
# endif
            floor->floorDestHeight =
                P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) + 24;
            frontsector = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
            P_SetPtrp(sec, DMU_FLOOR_MATERIAL,
                      P_GetPtrp(frontsector, DMU_FLOOR_MATERIAL));
            xsec->special = P_ToXSector(frontsector)->special;
            break;
#if __JDOOM__ || __JDOOM64__
        case FT_RAISE512:
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
            floor->floorDestHeight =
                P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) + 512;
            break;
#endif
# if __JDOOM64__
        case FT_RAISE32: // jd64
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED * 8;
            floor->floorDestHeight =
                P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) + 32;
            break;
# endif
        case FT_RAISETOTEXTURE: {
            int             minSize = DDMAXINT;
            floor->state = FS_UP;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
            P_FindLineInSectorSmallestBottomMaterial(sec, &minSize);
            floor->floorDestHeight =
                P_GetDoublep(floor->sector, DMU_FLOOR_HEIGHT) +
                    (coord_t) minSize;
            }
            break;
        case FT_LOWERANDCHANGE:
            floor->state = FS_DOWN;
            floor->sector = sec;
            floor->speed = FLOORSPEED;
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &floor->floorDestHeight);
            floor->material = (world_Material *)P_GetPtrp(sec, DMU_FLOOR_MATERIAL); {
            Sector* otherSec = findSectorSurroundingAtFloorHeight(sec,
                floor->floorDestHeight);
            if(otherSec) {
                floor->material = (world_Material *)P_GetPtrp(otherSec, DMU_FLOOR_MATERIAL);
                floor->newSpecial = P_ToXSector(otherSec)->special;
            }
            }
            break;
#endif
        default:
#if __JHEXEN__
            rtn = 0;
#endif
            break;
        }
    }
#if __JHEXEN__
    if(rtn && floor) {
        SN_StartSequence((mobj_t *)P_GetPtrp(floor->sector, DMU_EMITTER),
                         SEQ_PLATFORM + P_ToXSector(floor->sector)->seqType);
    }
#endif
    return rtn;
}
#if __JHEXEN__
struct findsectorneighborsforstairbuildparams_t {
    int type;
    coord_t height;
};
static int findSectorNeighborsForStairBuild(void *ptr, void *context) {
    Line *li = (Line *) ptr;
    findsectorneighborsforstairbuildparams_t *params = (findsectorneighborsforstairbuildparams_t *) context;
    Sector *frontSec = (Sector *)P_GetPtrp(li, DMU_FRONT_SECTOR);
    if(!frontSec) return false;
    Sector *backSec = (Sector *)P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!backSec) return false;
    xsector_t *xsec = P_ToXSector(frontSec);
    if(xsec->special == params->type + STAIR_SECTOR_TYPE && !xsec->specialData &&
       P_GetPtrp(frontSec, DMU_FLOOR_MATERIAL) == stairData.material &&
       P_GetIntp(frontSec, DMU_VALID_COUNT) != VALIDCOUNT) {
        enqueueStairSector(frontSec, params->type ^ 1, params->height);
        P_SetIntp(frontSec, DMU_VALID_COUNT, VALIDCOUNT);
    }
    xsec = P_ToXSector(backSec);
    if(xsec->special == params->type + STAIR_SECTOR_TYPE && !xsec->specialData &&
       P_GetPtrp(backSec, DMU_FLOOR_MATERIAL) == stairData.material &&
       P_GetIntp(backSec, DMU_VALID_COUNT) != VALIDCOUNT) {
        enqueueStairSector(backSec, params->type ^ 1, params->height);
        P_SetIntp(backSec, DMU_VALID_COUNT, VALIDCOUNT);
    }
    return false;
}
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
/**
 * @note Behaviour here is dependant upon the order of the sector-linked
 * LineDefs list. This is necessary to emulate the flawed algorithm used in
 * DOOM.exe In addition, this algorithm was further broken in Heretic as the
 * test which compares floor heights was removed.
 *
 * @warning DO NOT USE THIS ANYWHERE ELSE!
 */
struct spreadsectorparams_t {
    Sector *baseSec;
    world_Material *material;
    Sector *foundSec;
    coord_t height, stairSize;
};
static int findAdjacentSectorForSpread(void *ptr, void *context) {
    Line *li = (Line *) ptr;
    spreadsectorparams_t *params = (spreadsectorparams_t *) context;
    if(!(P_ToXLine(li)->flags & ML_TWOSIDED))
        return false;
    Sector *frontSec = (Sector *)P_GetPtrp(li, DMU_FRONT_SECTOR);
    if(!frontSec) return false;
    if(params->baseSec != frontSec)
        return false;
    Sector *backSec = (Sector *)P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!backSec) return false;
    if(P_GetPtrp(backSec, DMU_FLOOR_MATERIAL) != params->material)
        return false;
    /**
     * @note The placement of this step height addition is vital to ensure
     * the exact behaviour of the original DOOM algorithm. Logically this
     * should occur after the test below...
     */
    params->height += params->stairSize;
    xsector_t *xsec = P_ToXSector(backSec);
    if(xsec->specialData)
        return false;
    params->foundSec = backSec;
    return true;
}
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
int EV_BuildStairs(Line *line, stair_e type) {
    xsector_t *xsec;
    coord_t height = 0, stairsize = 0;
    float speed = 0;
    iterlist_t *list = P_GetSectorIterListForTag(P_ToXLine(line)->tag, false);
    if(!list) return 0;
    int rtn = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    spreadsectorparams_t params;
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        xsec = P_ToXSector(sec);
        if(xsec->specialData)
            continue;
        rtn = 1;
        floor_t *floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
        floor->thinker.function = T_MoveFloor;
        Thinker_Add(&floor->thinker);
        xsec->specialData = floor;
        floor->state = FS_UP;
        floor->sector = sec;
        switch(type) {
#if __JHERETIC__
        case build8:
            stairsize = 8;
            break;
        case build16:
            stairsize = 16;
            break;
#else
        case build8:
            speed = FLOORSPEED * .25;
            stairsize = 8;
            break;
        case turbo16:
            speed = FLOORSPEED * 4;
            stairsize = 16;
            break;
#endif
        default:
            break;
        }
#if __JHERETIC__
        floor->type = FT_RAISEBUILDSTEP;
        speed = FLOORSPEED;
        floor->speed = speed;
#else
        floor->speed = speed;
#endif
        height = P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + stairsize;
        floor->floorDestHeight = height;
        params.baseSec   = sec;
        params.material  = (world_Material *)P_GetPtrp(sec, DMU_FLOOR_MATERIAL);
        params.foundSec  = 0;
        params.height    = height;
        params.stairSize = stairsize;
        while(P_Iteratep(params.baseSec, DMU_LINE, findAdjacentSectorForSpread, &params)) {
            floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
            floor->thinker.function = T_MoveFloor;
            Thinker_Add(&floor->thinker);
            P_ToXSector(params.foundSec)->specialData = floor;
#if __JHERETIC__
            floor->type = FT_RAISEBUILDSTEP;
#endif
            floor->state = FS_UP;
            floor->speed = speed;
            floor->sector = params.foundSec;
            floor->floorDestHeight = params.height;
            params.baseSec = params.foundSec;
            params.foundSec = NULL;
        }
    }
    return rtn;
}
#endif
#if __JHEXEN__
static void enqueueStairSector(Sector* sec, int type, coord_t height) {
    if((stairQueueTail + 1) % STAIR_QUEUE_SIZE == stairQueueHead) {
        Con_Error("BuildStairs:  Too many branches located.\n");
    }
    stairQueue[stairQueueTail].sector = sec;
    stairQueue[stairQueueTail].type = type;
    stairQueue[stairQueueTail].height = height;
    stairQueueTail = (stairQueueTail + 1) % STAIR_QUEUE_SIZE;
}
static Sector* dequeueStairSector(int* type, coord_t* height) {
    Sector* sec;
    if(stairQueueHead == stairQueueTail) {
        return NULL;
    }
    *type = stairQueue[stairQueueHead].type;
    *height = stairQueue[stairQueueHead].height;
    sec = stairQueue[stairQueueHead].sector;
    stairQueueHead = (stairQueueHead + 1) % STAIR_QUEUE_SIZE;
    return sec;
}
static void processStairSector(Sector *sec, int type, coord_t height,
    stairs_e stairsType, int delay, int resetDelay) {
    findsectorneighborsforstairbuildparams_t params;
    height += stairData.stepDelta;
    floor_t *floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
    floor->thinker.function = T_MoveFloor;
    Thinker_Add(&floor->thinker);
    P_ToXSector(sec)->specialData = floor;
    floor->type = FT_RAISEBUILDSTEP;
    floor->state = (stairData.direction == -1? FS_DOWN : FS_UP);
    floor->sector = sec;
    floor->floorDestHeight = height;
    switch(stairsType) {
    case STAIRS_NORMAL:
        floor->speed = stairData.speed;
        if(delay) {
            floor->delayTotal = delay;
            floor->stairsDelayHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT) + stairData.stepDelta;
            floor->stairsDelayHeightDelta = stairData.stepDelta;
        }
        floor->resetDelay = resetDelay;
        floor->resetDelayCount = resetDelay;
        floor->resetHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
        break;
    case STAIRS_SYNC:
        floor->speed =
            stairData.speed * ((height - stairData.startHeight) / stairData.stepDelta);
        floor->resetDelay = delay; //arg4
        floor->resetDelayCount = delay;
        floor->resetHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
        break;
    default:
        break;
    }
    SN_StartSequence((mobj_t *)P_GetPtrp(sec, DMU_EMITTER),
                     SEQ_PLATFORM + P_ToXSector(sec)->seqType);
    params.type   = type;
    params.height = height;
    P_Iteratep(sec, DMU_LINE, findSectorNeighborsForStairBuild, &params);
}
#endif
/**
 * @param direction  Positive = up. Negative = down.
 */
#if __JHEXEN__
int EV_BuildStairs(Line * /*line*/, byte *args, int direction, stairs_e stairsType)
{
    stairData.textureChange = 0;
    stairData.direction = direction;
    stairData.stepDelta = stairData.direction * (coord_t) args[2];
    stairData.speed = (float) args[1] * (1.0 / 8);
    int resetDelay = (int) args[4];
    int delay      = (int) args[3];
    if(stairsType == STAIRS_PHASED) {
        stairData.startDelay = stairData.startDelayDelta = (int) args[3];
        resetDelay = stairData.startDelayDelta;
        delay = 0;
        stairData.textureChange = (int) args[4];
    }
    VALIDCOUNT++;
    iterlist_t *list = P_GetSectorIterListForTag((int) args[0], false);
    if(!list) return 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        stairData.material    = (world_Material *)P_GetPtrp(sec, DMU_FLOOR_MATERIAL);
        stairData.startHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
        if(P_ToXSector(sec)->specialData)
            continue; // Already moving, so keep going...
        enqueueStairSector(sec, 0, P_GetDoublep(sec, DMU_FLOOR_HEIGHT));
        P_ToXSector(sec)->special = 0;
    }
    int type;
    coord_t height;
    while((sec = dequeueStairSector(&type, &height))) {
        processStairSector(sec, type, height, stairsType, delay, resetDelay);
    }
    return 1;
}
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
struct findfirsttwosidedparams_t {
    Sector *sector;
    Line *foundLine;
};
static int findFirstTwosided(void *ptr, void *context) {
    Line *li = (Line *) ptr;
    findfirsttwosidedparams_t *params = (findfirsttwosidedparams_t *) context;
    Sector *backSec = (Sector *)P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!(P_ToXLine(li)->flags & ML_TWOSIDED))
        return false;
    if(backSec && !(params->sector && backSec == params->sector)) {
        params->foundLine = li;
        return true; // Stop iteration, this will do.
    }
    return false; // Continue iteration.
}
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
int EV_DoDonut(Line *line) {
    iterlist_t *list = P_GetSectorIterListForTag(P_ToXLine(line)->tag, false);
    if(!list) return 0;
    int rtn = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        findfirsttwosidedparams_t params;
        if(P_ToXSector(sec)->specialData)
            continue;
        rtn = 1;
        Sector *outer = 0, *ring = 0;
        params.sector    = 0;
        params.foundLine = 0;
        if(P_Iteratep(sec, DMU_LINE, findFirstTwosided, &params)) {
            ring = (Sector *)P_GetPtrp(params.foundLine, DMU_BACK_SECTOR);
            if(ring == sec) {
                ring = (Sector *)P_GetPtrp(params.foundLine, DMU_FRONT_SECTOR);
            }
            params.sector = sec;
            params.foundLine = NULL;
            if(P_Iteratep(ring, DMU_LINE, findFirstTwosided, &params)) {
                outer = (Sector *)P_GetPtrp(params.foundLine, DMU_BACK_SECTOR);
            }
        }
        if(outer && ring) {
            coord_t destHeight = P_GetDoublep(outer, DMU_FLOOR_HEIGHT);
            floor_t *floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
            floor->thinker.function = T_MoveFloor;
            Thinker_Add(&floor->thinker);
            P_ToXSector(ring)->specialData = floor;
            floor->type            = FT_RAISEDONUT;
            floor->crush           = false;
            floor->state           = FS_UP;
            floor->sector          = ring;
            floor->speed           = FLOORSPEED * .5;
            floor->material        = (world_Material *)P_GetPtrp(outer, DMU_FLOOR_MATERIAL);
            floor->newSpecial      = 0;
            floor->floorDestHeight = destHeight;
            floor = (floor_t *)Z_Calloc(sizeof(*floor), PU_MAP, 0);
            floor->thinker.function = T_MoveFloor;
            Thinker_Add(&floor->thinker);
            P_ToXSector(sec)->specialData = floor;
            floor->type            = FT_LOWER;
            floor->crush           = false;
            floor->state           = FS_DOWN;
            floor->sector          = sec;
            floor->speed           = FLOORSPEED * .5;
            floor->floorDestHeight = destHeight;
        }
    }
    return rtn;
}
#endif
#if __JHEXEN__
static int stopFloorCrush(thinker_t *th, void *context) {
    dd_bool *found = (dd_bool *) context;
    floor_t *floor = (floor_t *) th;
    if(floor->type == FT_RAISEFLOORCRUSH) {
        SN_StopSequence((mobj_t *)P_GetPtrp(floor->sector, DMU_EMITTER));
        P_ToXSector(floor->sector)->specialData = nullptr;
        P_NotifySectorFinished(P_ToXSector(floor->sector)->tag);
        Thinker_Remove(&floor->thinker);
        (*found) = true;
    }
    return false; // Continue iteration.
}
int EV_FloorCrushStop(Line * /*line*/, byte * /*args*/)
{
    dd_bool found = false;
    Thinker_Iterate(T_MoveFloor, stopFloorCrush, &found);
    return (found? 1 : 0);
}
#endif
#if __JHEXEN__ || __JDOOM64__
# if __JHEXEN__
int EV_DoFloorAndCeiling(Line *line, byte *args, int ftype, int ctype)
# else
int EV_DoFloorAndCeiling(Line* line, int ftype, int ctype)
# endif
{
# if __JHEXEN__
    int tag = (int) args[0];
# else
    int tag = P_ToXLine(line)->tag;
# endif
    iterlist_t *list = P_GetSectorIterListForTag(tag, false);
    if(!list) return 0;
    /**
     * @attention Kludge:
     * Due to the fact that sectors can only have one special thinker
     * linked at a time, this routine manually removes the link before
     * then creating a second thinker for the sector.
     * In order to commonize this we should maintain seperate links in
     * xsector_t for each type of special (not thinker type) i.e:
     *
     *   floor, ceiling, lightlevel
     *
     * @note: Floor and ceiling are capable of moving at different speeds
     * and with different target heights, we must remain compatible.
     */
# if __JHEXEN__
    dd_bool floor = EV_DoFloor(line, args, floortype_e(ftype));
# else
    dd_bool floor = EV_DoFloor(line, floortype_e(ftype));
# endif
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        P_ToXSector(sec)->specialData = 0;
    }
# if __JHEXEN__
    dd_bool ceiling = EV_DoCeiling(line, args, ceilingtype_e(ctype));
# else
    dd_bool ceiling = EV_DoCeiling(line, ceilingtype_e(ctype));
# endif
    return (floor | ceiling);
}
#endif
static AABoxd tmBox;
static mobj_t *tmThing;
dd_bool tmFloatOk; ///< @c true= move would be ok if within "tmFloorZ - tmCeilingZ".
coord_t tmFloorZ;
coord_t tmCeilingZ;
#if __JHEXEN__
static world_Material *tmFloorMaterial;
#endif
dd_bool tmFellDown; // $dropoff_fix
static coord_t tm[3];
static coord_t tmDropoffZ;
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
static Line *tmHitLine;
static int tmUnstuck; ///< $unstuck: used to check unsticking
#endif
Line *tmBlockingLine; // $unstuck: blocking line
#if __JHEXEN__
mobj_t *tmBlockingMobj;
#endif
Line *tmCeilingLine;
Line *tmFloorLine;
/*
 * Line aim/attack variables:
 */
mobj_t *lineTarget; // Who got hit (or NULL).
static coord_t attackRange;
#if __JHEXEN__
mobj_t *PuffSpawned;
#endif
static coord_t shootZ; ///< Height if not aiming up or down.
static mobj_t *shooterThing;
static float aimSlope;
static float topSlope, bottomSlope; ///< Slopes to top and bottom of target.
static byte *rejectMatrix;
coord_t P_GetGravity() {
    if(cfg.common.netGravity != -1) {
        return (coord_t) cfg.common.netGravity / 100;
    }
    return *((coord_t *) DD_GetVariable(DD_MAP_GRAVITY));
}
/**
 * Checks the reject matrix to find out if the two sectors are visible
 * from each other.
 */
static dd_bool checkReject(Sector *sec1, Sector *sec2) {
    if(rejectMatrix) {
        int const pnum = P_ToIndex(sec1) * numsectors + P_ToIndex(sec2);
        int const bytenum = pnum >> 3;
        int const bitnum = 1 << (pnum & 7);
        if(rejectMatrix[bytenum] & bitnum) {
            return false;
        }
    }
    return true;
}
dd_bool P_CheckSight(mobj_t const *beholder, mobj_t const *target) {
    if(!beholder || !target) return false;
    if(!Mobj_Sector(beholder)) return false;
    if(!Mobj_Sector(target)) return false;
    if(P_MobjIsCamera(target)) return false;
    if(!checkReject(Mobj_Sector(beholder), Mobj_Sector(target))) {
        return false;
    }
    vec3d_t from = { beholder->origin[VX], beholder->origin[VY], beholder->origin[VZ] };
    if(!P_MobjIsCamera(beholder)) {
        from[VZ] += beholder->height + -(beholder->height / 4);
    }
    return P_CheckLineSight(from, target->origin, 0, target->height, 0);
}
angle_t P_AimAtPoint2(coord_t const from[], coord_t const to[], dd_bool shadowed) {
    angle_t angle = M_PointToAngle2(from, to);
    if(shadowed) {
        angle += (P_Random() - P_Random()) << 21;
    }
    return angle;
}
angle_t P_AimAtPoint(coord_t const from[], coord_t const to[]) {
    return P_AimAtPoint2(from, to, false/* not shadowed*/);
}
struct pit_stompthing_params_t {
    mobj_t *stompMobj; ///< Mobj doing the stomping.
    vec2d_t location;  ///< Map space point being stomped.
    bool alwaysStomp;  ///< Disable per-type/monster stomp exclussions.
};
static int PIT_StompThing(mobj_t *mo, void *context) {
    pit_stompthing_params_t &parm = *static_cast<pit_stompthing_params_t *>(context);
    if(mo == parm.stompMobj) return false;
    if(!(mo->flags & MF_SHOOTABLE)) return false;
    coord_t const dist = mo->radius + parm.stompMobj->radius;
    if(fabs(mo->origin[VX] - parm.location[VX]) >= dist ||
       fabs(mo->origin[VY] - parm.location[VY]) >= dist) {
        return false;
    }
    if(!parm.alwaysStomp) {
        if(!(parm.stompMobj->flags2 & MF2_TELESTOMP))
            return true;
#if __JDOOM64__
        if(!Mobj_IsPlayer(parm.stompMobj))
            return true;
#elif __JDOOM__
        if(!Mobj_IsPlayer(parm.stompMobj) && gfw_Session()->mapUri().path() != "MAP30")
            return true;
#endif
    }
    P_DamageMobj(mo, parm.stompMobj, parm.stompMobj, 10000, true);
    return false; // Continue iteration.
}
dd_bool P_TeleportMove(mobj_t *mobj, coord_t x, coord_t y, dd_bool alwaysStomp) {
    if(!mobj) return false;
    IterList_Clear(spechit); /// @todo necessary? -ds
    pit_stompthing_params_t parm;
    parm.stompMobj = mobj;
    V2d_Set(parm.location, x, y);
    parm.alwaysStomp = CPP_BOOL(alwaysStomp);
    coord_t const dist = mobj->radius + MAXRADIUS;
    AABoxd const box(x - dist, y - dist, x + dist, y + dist);
    VALIDCOUNT++;
    if(Mobj_BoxIterator(&box, PIT_StompThing, &parm)) {
        return false;
    }
    P_MobjUnlink(mobj);
    mobj->origin[VX] = parm.location[VX];
    mobj->origin[VY] = parm.location[VY];
    P_MobjLink(mobj);
    mobj->floorZ     = P_GetDoublep(Mobj_Sector(mobj), DMU_FLOOR_HEIGHT);
    mobj->ceilingZ   = P_GetDoublep(Mobj_Sector(mobj), DMU_CEILING_HEIGHT);
#if !__JHEXEN__
    mobj->dropOffZ   = mobj->floorZ;
#endif
    P_MobjClearSRVO(mobj);
    return true; // Success.
}
void P_Telefrag(mobj_t *thing) {
    DENG2_ASSERT(thing != 0);
    P_TeleportMove(thing, thing->origin[VX], thing->origin[VY], false);
}
void P_TelefragMobjsTouchingPlayers() {
    for(uint i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = players + i;
        ddplayer_t *ddplr = plr->plr;
        if(!ddplr->inGame) continue;
        P_TeleportMove(ddplr->mo, ddplr->mo->origin[VX], ddplr->mo->origin[VY], true);
    }
}
struct pit_crossline_params_t {
    mobj_t *crossMobj;   ///< Mobj attempting to cross.
    AABoxd crossAABox;   ///< Bounding box of the trajectory.
    vec2d_t destination; ///< Would-be destination point.
};
static int PIT_CrossLine(Line *line, void *context) {
    pit_crossline_params_t &parm = *static_cast<pit_crossline_params_t *>(context);
    if((P_GetIntp(line, DMU_FLAGS) & DDLF_BLOCKING) ||
       (P_ToXLine(line)->flags & ML_BLOCKMONSTERS) ||
       (!P_GetPtrp(line, DMU_FRONT_SECTOR) || !P_GetPtrp(line, DMU_BACK_SECTOR))) {
        AABoxd *aaBox = (AABoxd *)P_GetPtrp(line, DMU_BOUNDING_BOX);
        if(!(parm.crossAABox.minX > aaBox->maxX ||
             parm.crossAABox.maxX < aaBox->minX ||
             parm.crossAABox.maxY < aaBox->minY ||
             parm.crossAABox.minY > aaBox->maxY)) {
            return    (Line_PointOnSide(line, parm.crossMobj->origin) < 0)
                   != (Line_PointOnSide(line, parm.destination)       < 0);
        }
    }
    return false; // Continue iteration.
}
dd_bool P_CheckSides(mobj_t *mobj, coord_t x, coord_t y) {
    /*
     * Check to see if the trajectory crosses a blocking map line.
     *
     * Currently this assumes an infinite line, which is not quite correct.
     * A more correct solution would be to check for an intersection of the
     * trajectory and the line, but that takes longer and probably really isn't
     * worth the effort.
     */
    pit_crossline_params_t parm;
    parm.crossMobj  = mobj;
    parm.crossAABox =
        AABoxd(MIN_OF(mobj->origin[VX], x), MIN_OF(mobj->origin[VY], y),
               MAX_OF(mobj->origin[VX], x), MAX_OF(mobj->origin[VY], y));
    V2d_Set(parm.destination, x, y);
    VALIDCOUNT++;
    return Line_BoxIterator(&parm.crossAABox, LIF_ALL, PIT_CrossLine, &parm);
}
#if __JHEXEN__
static void checkForPushSpecial(Line *line, int side, mobj_t *mobj) {
    if(P_ToXLine(line)->special) {
        if(mobj->flags2 & MF2_PUSHWALL) {
            P_ActivateLine(line, mobj, side, SPAC_PUSH);
        }
        else if(mobj->flags2 & MF2_IMPACT) {
            P_ActivateLine(line, mobj, side, SPAC_IMPACT);
        }
    }
}
#endif // __JHEXEN__
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
/**
 * $unstuck: used to test intersection between thing and line assuming NO
 * movement occurs -- used to avoid sticky situations.
 */
static int untouched(Line *line, mobj_t *mobj) {
    DENG_ASSERT(line != 0 && mobj != 0);
    coord_t const x      = mobj->origin[VX];
    coord_t const y      = mobj->origin[VY];
    coord_t const radius = mobj->radius;
    AABoxd const *ldBox  = (AABoxd *)P_GetPtrp(line, DMU_BOUNDING_BOX);
    AABoxd moBox;
    if(((moBox.minX = x - radius) >= ldBox->maxX) ||
       ((moBox.minY = y - radius) >= ldBox->maxY) ||
       ((moBox.maxX = x + radius) <= ldBox->minX) ||
       ((moBox.maxY = y + radius) <= ldBox->minY) ||
       Line_BoxOnSide(line, &moBox)) {
        return true;
    }
    return false;
}
#endif
static int PIT_CheckThing(mobj_t *thing, void * /*context*/)
{
    if(thing == tmThing) {
        return false;
    }
#if __JHEXEN__
    if(thing == tmThing->onMobj) {
        return false;
    }
#endif
    if(!(thing->flags & (MF_SOLID | MF_SPECIAL | MF_SHOOTABLE)) ||
       P_MobjIsCamera(thing) || P_MobjIsCamera(tmThing)) {
        return false;
    }
#if !__JHEXEN__
    dd_bool overlap = false;
    if(tmThing->player && !FEQUAL(tm[VZ], DDMAXFLOAT) &&
       (cfg.moveCheckZ || (tmThing->flags2 & MF2_PASSMOBJ))) {
        if((thing->origin[VZ] > tm[VZ] + tmThing->height) ||
           (thing->origin[VZ] + thing->height < tm[VZ])) {
            return false; // Under or over it.
        }
        overlap = true;
    }
#endif
    coord_t blockdist = thing->radius + tmThing->radius;
    if(fabs(thing->origin[VX] - tm[VX]) >= blockdist ||
       fabs(thing->origin[VY] - tm[VY]) >= blockdist) {
        return false; // Didn't hit thing.
    }
    if(IS_CLIENT) {
        if(tmThing->ddFlags & DDMF_MISSILE) {
            return false;
        }
        if(tmThing->player && ClPlayer_ClMobj(tmThing->player - players) == thing) {
            return false;
        }
    }
/*
#if __JHEXEN__
    if(IS_CLIENT) return true;
#endif
*/
#if __JHEXEN__
    tmBlockingMobj = thing;
#endif
#if __JHEXEN__
    if(tmThing->flags2 & MF2_PASSMOBJ)
#else
    if(!tmThing->player && (tmThing->flags2 & MF2_PASSMOBJ))
#endif
    {
#if __JHERETIC__
        if((tmThing->type == MT_IMP || tmThing->type == MT_WIZARD) &&
           (thing->type == MT_IMP || thing->type == MT_WIZARD)) {
            return true; // Don't let imps/wizards fly over other imps/wizards.
        }
#elif __JHEXEN__
        if(tmThing->type == MT_BISHOP && thing->type == MT_BISHOP) {
            return true; // Don't let bishops fly over other bishops.
        }
#endif
        if(!(thing->flags & MF_SPECIAL)) {
            if(tmThing->origin[VZ] > thing->origin[VZ] + thing->height ||
               tmThing->origin[VZ] + tmThing->height < thing->origin[VZ]) {
                return false; // Over/under thing.
            }
        }
    }
    if((tmThing->flags & MF_SKULLFLY) && (thing->flags & MF_SOLID)) {
#if __JHEXEN__
        tmBlockingMobj = 0;
        if(tmThing->type == MT_MINOTAUR) {
            if(!(thing->flags & MF_COUNTKILL)) {
                return true;
            }
        }
        else if(tmThing->type == MT_HOLY_FX) {
            if((thing->flags & MF_SHOOTABLE) && thing != tmThing->target) {
                if(IS_NETGAME && !gfw_Rule(deathmatch) && thing->player) {
                    return false; // don't attack other co-op players
                }
                if((thing->flags2 & MF2_REFLECTIVE) &&
                   (thing->player || (thing->flags2 & MF2_BOSS))) {
                    tmThing->tracer = tmThing->target;
                    tmThing->target = thing;
                    return false;
                }
                if(thing->flags & MF_COUNTKILL || thing->player) {
                    tmThing->tracer = thing;
                }
                if(P_Random() < 96) {
                    int damage = 12;
                    if(thing->player || (thing->flags2 & MF2_BOSS)) {
                        damage = 3;
                        tmThing->health -= 6;
                    }
                    P_DamageMobj(thing, tmThing, tmThing->target, damage, false);
                    if(P_Random() < 128) {
                        P_SpawnMobj(MT_HOLY_PUFF, tmThing->origin, P_Random() << 24, 0);
                        S_StartSound(SFX_SPIRIT_ATTACK, tmThing);
                        if((thing->flags & MF_COUNTKILL) && P_Random() < 128 &&
                           !S_IsPlaying(SFX_PUPPYBEAT, thing)) {
                            if((thing->type == MT_CENTAUR) ||
                               (thing->type == MT_CENTAURLEADER) ||
                               (thing->type == MT_ETTIN)) {
                                S_StartSound(SFX_PUPPYBEAT, thing);
                            }
                        }
                    }
                }
                if(thing->health <= 0) {
                    tmThing->tracer = 0;
                }
            }
            return false;
        }
#endif
        int damage = tmThing->damage;
#if __JDOOM__
        if(damage == DDMAXINT) {
            damage = tmThing->info->damage;
        }
#endif
        damage *= (P_Random() % 8) + 1;
        P_DamageMobj(thing, tmThing, tmThing, damage, false);
        tmThing->flags &= ~MF_SKULLFLY;
        tmThing->mom[MX] = tmThing->mom[MY] = tmThing->mom[MZ] = 0;
#if __JHERETIC__ || __JHEXEN__
        P_MobjChangeState(tmThing, P_GetState(mobjtype_t(tmThing->type), SN_SEE));
#else
        P_MobjChangeState(tmThing, P_GetState(mobjtype_t(tmThing->type), SN_SPAWN));
#endif
        return true; // Stop moving.
    }
#if __JHEXEN__
    if((tmThing->flags2 & MF2_BLASTED) && (thing->flags & MF_SHOOTABLE)) {
        if(!(thing->flags2 & MF2_BOSS) && (thing->flags & MF_COUNTKILL)) {
            thing->mom[MX] += tmThing->mom[MX];
            thing->mom[MY] += tmThing->mom[MY];
            NetSv_PlayerMobjImpulse(thing, tmThing->mom[MX], tmThing->mom[VY], 0);
            if((thing->mom[MX] + thing->mom[MY]) > 3) {
                P_DamageMobj(thing, tmThing, tmThing,
                             (tmThing->info->mass / 100) + 1, false);
                P_DamageMobj(tmThing, thing, thing,
                             ((thing->info->mass / 100) + 1) >> 2, false);
            }
            return true;
        }
    }
#endif
    if(tmThing->flags & MF_MISSILE) {
#if __JHEXEN__
        if(thing->flags2 & MF2_NONSHOOTABLE) {
            return false;
        }
#else
        if((thing->flags & MF_SHADOW) && (tmThing->flags2 & MF2_THRUGHOST)) {
            return false;
        }
#endif
        if(tmThing->origin[VZ] > thing->origin[VZ] + thing->height ||
           tmThing->origin[VZ] + tmThing->height < thing->origin[VZ]) {
            return false;
        }
#if __JHEXEN__
        if(tmThing->flags2 & MF2_FLOORBOUNCE) {
            return !(tmThing->target == thing || !(thing->flags & MF_SOLID));
        }
        if(tmThing->type == MT_LIGHTNING_FLOOR || tmThing->type == MT_LIGHTNING_CEILING) {
            if((thing->flags & MF_SHOOTABLE) && thing != tmThing->target) {
                if(thing->info->mass != DDMAXINT) {
                    thing->mom[MX] += tmThing->mom[MX] / 16;
                    thing->mom[MY] += tmThing->mom[MY] / 16;
                    NetSv_PlayerMobjImpulse(thing, tmThing->mom[MX] / 16, tmThing->mom[MY] / 16, 0);
                }
                if((!thing->player && !(thing->flags2 & MF2_BOSS)) ||
                   !(mapTime & 1)) {
                    if(thing->type == MT_CENTAUR || thing->type == MT_CENTAURLEADER) {
                        P_DamageMobj(thing, tmThing, tmThing->target, 9, false);
                    }
                    else {
                        P_DamageMobj(thing, tmThing, tmThing->target, 3, false);
                    }
                    if(!S_IsPlaying(SFX_MAGE_LIGHTNING_ZAP, tmThing)) {
                        S_StartSound(SFX_MAGE_LIGHTNING_ZAP, tmThing);
                    }
                    if((thing->flags & MF_COUNTKILL) && P_Random() < 64 &&
                       !S_IsPlaying(SFX_PUPPYBEAT, thing)) {
                        if((thing->type == MT_CENTAUR) ||
                           (thing->type == MT_CENTAURLEADER) ||
                           (thing->type == MT_ETTIN)) {
                            S_StartSound(SFX_PUPPYBEAT, thing);
                        }
                    }
                }
                tmThing->health--;
                if(tmThing->health <= 0 || thing->health <= 0) {
                    return true;
                }
                if(tmThing->type == MT_LIGHTNING_FLOOR) {
                    if(tmThing->lastEnemy && !tmThing->lastEnemy->tracer) {
                        tmThing->lastEnemy->tracer = thing;
                    }
                }
                else if(!tmThing->tracer) {
                    tmThing->tracer = thing;
                }
            }
            return false; // Lightning zaps through all sprites.
        }
        if(tmThing->type == MT_LIGHTNING_ZAP) {
            if((thing->flags & MF_SHOOTABLE) && thing != tmThing->target &&
               tmThing->lastEnemy) {
                mobj_t *lmo = tmThing->lastEnemy;
                if(lmo->type == MT_LIGHTNING_FLOOR) {
                    if(lmo->lastEnemy && !lmo->lastEnemy->tracer) {
                        lmo->lastEnemy->tracer = thing;
                    }
                }
                else if(!lmo->tracer) {
                    lmo->tracer = thing;
                }
                if(!(mapTime & 3)) {
                    lmo->health--;
                }
            }
        }
        else if(tmThing->type == MT_MSTAFF_FX2 && thing != tmThing->target) {
            if(!thing->player && !(thing->flags2 & MF2_BOSS)) {
                switch(thing->type) {
                case MT_FIGHTER_BOSS:   // these not flagged boss
                case MT_CLERIC_BOSS:    // so they can be blasted
                case MT_MAGE_BOSS:
                    break;
                default:
                    P_DamageMobj(thing, tmThing, tmThing->target, 10, false);
                    return false;
                }
            }
        }
#endif
#if __JDOOM__ || __JDOOM64__
        if(tmThing->target &&
           (tmThing->target->type == thing->type ||
           (tmThing->target->type == MT_KNIGHT && thing->type == MT_BRUISER) ||
           (tmThing->target->type == MT_BRUISER && thing->type == MT_KNIGHT)))
#else
        if(tmThing->target && tmThing->target->type == thing->type)
#endif
        {
            if(thing == tmThing->target) {
                return false;
            }
#if __JHEXEN__
            if(!thing->player) {
                return true; // Hit same species as originator, explode, no damage
            }
#else
            if(!monsterInfight && thing->type != MT_PLAYER) {
                return true;
            }
#endif
        }
        if(!(thing->flags & MF_SHOOTABLE)) {
            return !!(thing->flags & MF_SOLID); // Didn't do any damage.
        }
        if(tmThing->flags2 & MF2_RIP) {
#if __JHEXEN__
            if(!(thing->flags & MF_NOBLOOD) &&
               !(thing->flags2 & MF2_REFLECTIVE) &&
               !(thing->flags2 & MF2_INVULNERABLE))
#else
            if(!(thing->flags & MF_NOBLOOD))
#endif
            {   // Ok to spawn some blood.
                P_RipperBlood(tmThing);
            }
#if __JHERETIC__
            S_StartSound(SFX_RIPSLOP, tmThing);
#endif
            int damage = tmThing->damage;
#if __JDOOM__
            if(damage == DDMAXINT) {
                damage = tmThing->info->damage;
            }
#endif
            damage *= (P_Random() & 3) + 2;
            P_DamageMobj(thing, tmThing, tmThing->target, damage, false);
            if((thing->flags2 & MF2_PUSHABLE) && !(tmThing->flags2 & MF2_CANNOTPUSH)) {
                thing->mom[MX] += tmThing->mom[MX] / 4;
                thing->mom[MY] += tmThing->mom[MY] / 4;
                NetSv_PlayerMobjImpulse(thing, tmThing->mom[MX]/4, tmThing->mom[MY]/4, 0);
            }
            IterList_Clear(spechit);
            return false;
        }
        int damage = tmThing->damage;
#if __JDOOM__
        if(tmThing->damage == DDMAXINT) {
            damage = tmThing->info->damage;
        }
#endif
        damage *= (P_Random() % 8) + 1;
#if __JDOOM__ || __JDOOM64__
        P_DamageMobj(thing, tmThing, tmThing->target, damage, false);
#else
        if(damage) {
# if __JHERETIC__
            if(!(thing->flags & MF_NOBLOOD) && P_Random() < 192)
# else //__JHEXEN__
            if(!(thing->flags & MF_NOBLOOD) &&
               !(thing->flags2 & MF2_REFLECTIVE) &&
               !(thing->flags2 & MF2_INVULNERABLE) &&
               !(tmThing->type == MT_TELOTHER_FX1) &&
               !(tmThing->type == MT_TELOTHER_FX2) &&
               !(tmThing->type == MT_TELOTHER_FX3) &&
               !(tmThing->type == MT_TELOTHER_FX4) &&
               !(tmThing->type == MT_TELOTHER_FX5) && (P_Random() < 192))
# endif
            {
                P_SpawnBloodSplatter(tmThing->origin[VX], tmThing->origin[VY], tmThing->origin[VZ], thing);
            }
            P_DamageMobj(thing, tmThing, tmThing->target, damage, false);
        }
#endif
        return true;
    }
    if((thing->flags2 & MF2_PUSHABLE) && !(tmThing->flags2 & MF2_CANNOTPUSH)) {
        coord_t pushImpulse[2] = {tmThing->mom[MX] / 4, tmThing->mom[MY] / 4};
        for (int axis = 0; axis < 2; ++axis) {
            if (cfg.common.pushableMomentumLimitedToPusher) {
                coord_t maxIncrement = tmThing->mom[axis] - thing->mom[axis];
                if (thing->mom[axis] > 0 && pushImpulse[axis] > 0) {
                    pushImpulse[axis] = de::max(0.0, de::min(pushImpulse[axis], maxIncrement));
                }
                else if (thing->mom[axis] < 0 && pushImpulse[axis] < 0) {
                    pushImpulse[axis] = de::min(0.0, de::max(pushImpulse[axis], maxIncrement));
                }
            }
            thing->mom[axis] += pushImpulse[axis];
        }
        if (!de::fequal(pushImpulse[MX], 0) || !de::fequal(pushImpulse[MY], 0)) {
            NetSv_PlayerMobjImpulse(thing, float(pushImpulse[MX]), float(pushImpulse[MY]), 0);
        }
    }
    dd_bool solid;
    if(tmThing->type == MT_BLOOD) {
        solid = true;
    }
    else {
        solid = (thing->flags & MF_SOLID) && !(thing->flags & MF_NOCLIP) &&
                (tmThing->flags & MF_SOLID);
    }
#if __JHEXEN__
    if(tmThing->player && tmThing->onMobj && solid) {
        if(tmThing->origin[VZ] > thing->origin[VZ] + thing->height - 24) {
            tmFloorZ = MAX_OF(tmFloorZ, thing->origin[VZ] + thing->height);
            solid = false;
        }
    }
#endif
    if((thing->flags & MF_SPECIAL) && (tmThing->flags & MF_PICKUP)) {
        P_TouchSpecialMobj(thing, tmThing); // Can remove thing.
    }
#if !__JHEXEN__
    else if(overlap && solid) {
        if(!(thing->flags & MF_CORPSE) && tm[VZ] > thing->origin[VZ] + thing->height - 24) {
            tmThing->onMobj = thing;
            if(thing->origin[VZ] + thing->height > tmFloorZ) {
                tmFloorZ = thing->origin[VZ] + thing->height;
            }
            return false;
        }
    }
    else if(!tmThing->player && solid) {
        if(cfg.allowMonsterFloatOverBlocking && (tmThing->flags & MF_FLOAT) && !thing->player) {
            coord_t top = thing->origin[VZ] + thing->height;
            tmThing->onMobj = thing;
            tmFloorZ = MAX_OF(tmFloorZ, top);
            return false;
        }
    }
#endif
    return solid;
}
/**
 * Adjusts tmFloorZ and tmCeilingZ as lines are contacted.
 */
static int PIT_CheckLine(Line *ld, void * /*context*/)
{
    AABoxd const *aaBox = (AABoxd *)P_GetPtrp(ld, DMU_BOUNDING_BOX);
    if(tmBox.minX >= aaBox->maxX || tmBox.minY >= aaBox->maxY ||
       tmBox.maxX <= aaBox->minX || tmBox.maxY <= aaBox->minY) {
        return false;
    }
    /*
     * Real player mobjs are allowed to use high-precision, non-vanilla
     * collision testing -- the rest of the playsim uses coord_t, and we don't
     * want conflicting results (e.g., getting stuck in tight spaces).
     */
    if(Mobj_IsPlayer(tmThing) && !Mobj_IsVoodooDoll(tmThing)) {
        if(Line_BoxOnSide(ld, &tmBox)) // double precision floats
        {
            return false;
        }
    }
    else {
        if(Line_BoxOnSide_FixedPrecision(ld, &tmBox)) {
            return false;
        }
    }
    xline_t *xline = P_ToXLine(ld);
#if !__JHEXEN__
    tmThing->wallHit = true;
    if(xline->special) {
        tmHitLine = ld;
    }
#endif
    if(!P_GetPtrp(ld, DMU_BACK_SECTOR)) // One sided line.
    {
#if __JHEXEN__
        if(tmThing->flags2 & MF2_BLASTED) {
            P_DamageMobj(tmThing, NULL, NULL, tmThing->info->mass >> 5, false);
        }
        checkForPushSpecial(ld, 0, tmThing);
        return true;
#else
        coord_t d1[2];
        P_GetDoublepv(ld, DMU_DXY, d1);
        /**
         * $unstuck: allow player to move out of 1s wall, to prevent
         * sticking. The moving thing's destination position will cross the
         * given line.
         * If this should not be allowed, return false.
         * If the line is special, keep track of it to process later if the
         * move is proven ok.
         *
         * @note Specials are NOT sorted by order, so two special lines that
         *       are only 8 units apart could be crossed in either order.
         */
        tmBlockingLine = ld;
        return !(tmUnstuck && !untouched(ld, tmThing) &&
            ((tm[VX] - tmThing->origin[VX]) * d1[1]) >
            ((tm[VY] - tmThing->origin[VY]) * d1[0]));
#endif
    }
#if __JHERETIC__
    if(!P_GetPtrp(ld, DMU_BACK_SECTOR)) // one sided line
    {
        if((tmThing->flags & MF_MISSILE) && xline->special) {
            IterList_PushBack(spechit, ld);
        }
        return true;
    }
#endif
    if(!(tmThing->flags & MF_MISSILE)) {
        if(P_GetIntp(ld, DMU_FLAGS) & DDLF_BLOCKING) {
#if __JHEXEN__
            if(tmThing->flags2 & MF2_BLASTED) {
                P_DamageMobj(tmThing, NULL, NULL, tmThing->info->mass >> 5, false);
            }
            checkForPushSpecial(ld, 0, tmThing);
            return true;
#else
            return !(tmUnstuck && !untouched(ld, tmThing));
#endif
        }
#if __JHEXEN__
        if(!tmThing->player && tmThing->type != MT_CAMERA &&
           (xline->flags & ML_BLOCKMONSTERS))
#elif __JHERETIC__
        if(!tmThing->player && tmThing->type != MT_POD &&
           (xline->flags & ML_BLOCKMONSTERS))
#else
        if(!tmThing->player &&
           (xline->flags & ML_BLOCKMONSTERS))
#endif
        {
#if __JHEXEN__
            if(tmThing->flags2 & MF2_BLASTED) {
                P_DamageMobj(tmThing, NULL, NULL, tmThing->info->mass >> 5, false);
            }
#endif
            return true;
        }
    }
#if __JDOOM64__
    if((tmThing->flags & MF_MISSILE) && (xline->flags & ML_BLOCKALL)) {
        return !(tmUnstuck && !untouched(ld, tmThing));
    }
#endif
    LineOpening opening; Line_Opening(ld, &opening);
    if(opening.top < tmCeilingZ) {
        tmCeilingZ    = opening.top;
        tmCeilingLine = ld;
#if !__JHEXEN__
        tmBlockingLine     = ld;
#endif
    }
    if(opening.bottom > tmFloorZ) {
        tmFloorZ    = opening.bottom;
        tmFloorLine = ld;
#if !__JHEXEN__
        tmBlockingLine   = ld;
#endif
    }
    if(opening.lowFloor < tmDropoffZ) {
        tmDropoffZ  = opening.lowFloor;
    }
    if(P_ToXLine(ld)->special) {
        IterList_PushBack(spechit, ld);
    }
#if !__JHEXEN__
    tmThing->wallHit = false;
#endif
    return false; // Continue iteration.
}
dd_bool P_CheckPositionXYZ(mobj_t *thing, coord_t x, coord_t y, coord_t z) {
#if defined(__JHERETIC__)
    if (thing->type != MT_POD) // vanilla onMobj behavior for pods
    {
        thing->onMobj = nullptr;
    }
#elif !__JHEXEN__
    thing->onMobj  = 0;
#endif
    thing->wallHit = false;
    tmThing         = thing;
    V3d_Set(tm, x, y, z);
    tmBox           = AABoxd(tm[VX] - tmThing->radius, tm[VY] - tmThing->radius,
                             tm[VX] + tmThing->radius, tm[VY] + tmThing->radius);
#if !__JHEXEN__
    tmHitLine       = 0;
#endif
    Sector *newSector = Sector_AtPoint_FixedPrecision(tm);
    tmCeilingLine   = tmFloorLine = 0;
    tmFloorZ        = tmDropoffZ = P_GetDoublep(newSector, DMU_FLOOR_HEIGHT);
    tmCeilingZ      = P_GetDoublep(newSector, DMU_CEILING_HEIGHT);
#if __JHEXEN__
    tmFloorMaterial = (world_Material *)P_GetPtrp(newSector, DMU_FLOOR_MATERIAL);
#else
    tmBlockingLine  = 0;
    tmUnstuck       = Mobj_IsPlayer(thing) && !Mobj_IsVoodooDoll(thing);
#endif
    IterList_Clear(spechit);
    if(tmThing->flags & MF_NOCLIP) {
#if __JHEXEN__
        if(!(tmThing->flags & MF_SKULLFLY)) {
            return true;
        }
#else
        return true;
#endif
    }
    VALIDCOUNT++;
#if __JHEXEN__
    tmBlockingMobj = 0;
#endif
    if(!P_MobjIsCamera(thing)) {
        /*
         * The bounding box is extended by MAXRADIUS because mobj_ts are grouped
         * into mapblocks based on their origin point and can overlap adjacent
         * blocks by up to MAXRADIUS units.
         */
        AABoxd tmBoxExpanded(tmBox.minX - MAXRADIUS, tmBox.minY - MAXRADIUS,
                             tmBox.maxX + MAXRADIUS, tmBox.maxY + MAXRADIUS);
        if(Mobj_BoxIterator(&tmBoxExpanded, PIT_CheckThing, 0)) {
            return false;
        }
        if(thing->onMobj) {
            App_Log(DE2_DEV_MAP_XVERBOSE,
                    "thing->onMobj = %p/%i (solid:%i) [thing:%p/%i]", thing->onMobj,
                    thing->onMobj->thinker.id,
                    (thing->onMobj->flags & MF_SOLID) != 0,
                    thing, thing->thinker.id);
        }
    }
#if __JHEXEN__
    if(tmThing->flags & MF_NOCLIP) {
        return true;
    }
#endif
#if __JHEXEN__
    tmBlockingMobj = 0;
#endif
    return !Line_BoxIterator(&tmBox, LIF_ALL, PIT_CheckLine, 0);
}
dd_bool P_CheckPosition(mobj_t *thing, coord_t const pos[3]) {
    return P_CheckPositionXYZ(thing, pos[VX], pos[VY], pos[VZ]);
}
dd_bool P_CheckPositionXY(mobj_t *thing, coord_t x, coord_t y) {
    return P_CheckPositionXYZ(thing, x, y, DDMAXFLOAT);
}
dd_bool Mobj_IsRemotePlayer(mobj_t *mo) {
    return (mo && ((IS_DEDICATED && mo->dPlayer) ||
                   (IS_CLIENT && mo->player && mo->player - players != CONSOLEPLAYER)));
}
#if __JDOOM64__ || __JHERETIC__
static void checkMissileImpact(mobj_t &mobj) {
    if(IS_CLIENT) return;
    if(!(mobj.flags & MF_MISSILE)) return;
    if(!mobj.target || !mobj.target->player) return;
    if(IterList_Empty(spechit)) return;
    IterList_SetIteratorDirection(spechit, ITERLIST_BACKWARD);
    IterList_RewindIterator(spechit);
    Line *line;
    while((line = (Line *)IterList_MoveIterator(spechit)) != 0) {
        P_ActivateLine(line, mobj.target, 0, SPAC_IMPACT);
    }
}
#endif
/**
 * Attempt to move to a new position, crossing special lines unless
 * MF_TELEPORT is set. $dropoff_fix
 */
#if __JHEXEN__
static dd_bool P_TryMove2(mobj_t *thing, coord_t x, coord_t y)
#else
static dd_bool P_TryMove2(mobj_t *thing, coord_t x, coord_t y, dd_bool dropoff)
#endif
{
    dd_bool const isRemotePlayer = Mobj_IsRemotePlayer(thing);
    tmFloatOk  = false;
#if !__JHEXEN__
    tmFellDown = false;
#endif
#if __JHEXEN__
    if(!P_CheckPositionXY(thing, x, y))
#else
    if(!P_CheckPositionXYZ(thing, x, y, thing->origin[VZ]))
#endif
    {
#if __JHEXEN__
        if(!tmBlockingMobj || tmBlockingMobj->player || !thing->player) {
            goto pushline;
        }
        else if(tmBlockingMobj->origin[VZ] + tmBlockingMobj->height - thing->origin[VZ] > 24 ||
                (P_GetDoublep(Mobj_Sector(tmBlockingMobj), DMU_CEILING_HEIGHT) -
                 (tmBlockingMobj->origin[VZ] + tmBlockingMobj->height) < thing->height) ||
                (tmCeilingZ - (tmBlockingMobj->origin[VZ] + tmBlockingMobj->height) <
                 thing->height)) {
            goto pushline;
        }
#else
#  if __JHERETIC__
        checkMissileImpact(*thing);
#  endif
        if(!thing->onMobj || thing->wallHit)
            return false;
#endif
    }
    if(!(thing->flags & MF_NOCLIP)) {
#if __JHEXEN__
        if(tmCeilingZ - tmFloorZ < thing->height)
        {   // Doesn't fit.
            goto pushline;
        }
        tmFloatOk = true;
        if(!(thing->flags & MF_TELEPORT) &&
           tmCeilingZ - thing->origin[VZ] < thing->height &&
           thing->type != MT_LIGHTNING_CEILING && !(thing->flags2 & MF2_FLY)) {
            goto pushline;
        }
#else
        dd_bool ret = (tmUnstuck &&
            !(tmCeilingLine && untouched(tmCeilingLine, tmThing)) &&
            !(tmFloorLine   && untouched(tmFloorLine, tmThing)));
        if(tmCeilingZ - tmFloorZ < thing->height) {
            return ret; // Doesn't fit.
        }
        tmFloatOk = true;
        if(!(thing->flags & MF_TELEPORT) && !(thing->flags2 & MF2_FLY) &&
           tmCeilingZ - thing->origin[VZ] < thing->height) {
            return ret;
        }
        if(!(thing->flags & MF_TELEPORT) &&
            !(thing->flags2 & MF2_FLY)
# if __JHERETIC__
            && thing->type != MT_MNTRFX2 // The Minotaur floor fire (MT_MNTRFX2) can step up any amount
# endif
            ) {
            if(!isRemotePlayer && tmFloorZ - thing->origin[VZ] > 24) {
# if __JHERETIC__
                checkMissileImpact(*thing);
# endif
                return ret;
            }
        }
# if __JHERETIC__
        if((thing->flags & MF_MISSILE) && tmFloorZ > thing->origin[VZ]) {
            checkMissileImpact(*thing);
        }
# endif
#endif
        if(thing->flags2 & MF2_FLY) {
            if(thing->origin[VZ] + thing->height > tmCeilingZ) {
                thing->mom[MZ] = -8;
#if __JHEXEN__
                goto pushline;
#else
                return false;
#endif
            }
            else if(thing->origin[VZ] < tmFloorZ &&
                    tmFloorZ - tmDropoffZ > 24) {
                thing->mom[MZ] = 8;
#if __JHEXEN__
                goto pushline;
#else
                return false;
#endif
            }
        }
#if __JHEXEN__
        if(!(thing->flags & MF_TELEPORT)
           && thing->type != MT_MNTRFX2 && thing->type != MT_LIGHTNING_FLOOR
           && !isRemotePlayer
           && tmFloorZ - thing->origin[VZ] > 24) {
            goto pushline;
        }
#endif
#if __JHEXEN__
        if(!(thing->flags & (MF_DROPOFF | MF_FLOAT)) &&
           (tmFloorZ - tmDropoffZ > 24) &&
           !(thing->flags2 & MF2_BLASTED)) {
            return false;
        }
#else
        /*
         * Allow certain objects to drop off.
         * Prevent monsters from getting stuck hanging off ledges.
         * Allow dropoffs in controlled circumstances.
         * Improve symmetry of clipping on stairs.
         */
        {
            if (!(thing->flags & (MF_DROPOFF | MF_FLOAT))) {
                if (cfg.avoidDropoffs) {
                    if (tmFloorZ - tmDropoffZ > 24) {
                        return false; // Don't stand over dropoff.
                    }
                }
                else {
                    coord_t floorZ = tmFloorZ;
                    if (thing->onMobj) {
                        floorZ = (thing->origin[VZ] > tmFloorZ ? thing->origin[VZ] : tmFloorZ);
                    }
                    if (!dropoff) {
                        if (thing->floorZ - floorZ > 24 || thing->dropOffZ - tmDropoffZ > 24)
                            return false;
                    }
                    else {
                        tmFellDown = !(thing->flags & MF_NOGRAVITY) && thing->origin[VZ] - floorZ > 24;
                    }
                }
            }
#if defined (__JHERETIC__)
            else if (!thing->onMobj && (thing->flags & MF_DROPOFF) && !(thing->flags & MF_NOGRAVITY)) {
                tmFellDown = (thing->origin[VZ] - tmFloorZ > 24);
            }
#endif
        }
#endif
#if __JDOOM64__
        if(!(thing->flags & MF_TELEPORT) /*&& thing->type != MT_SPAWNFIRE*/
            && !isRemotePlayer
            && tmFloorZ - thing->origin[VZ] > 24) {
            checkMissileImpact(*thing);
            return false;
        }
#endif
#if __JHEXEN__
        if((thing->flags2 & MF2_CANTLEAVEFLOORPIC) &&
           (tmFloorMaterial != P_GetPtrp(Mobj_Sector(thing), DMU_FLOOR_MATERIAL) ||
            !FEQUAL(tmFloorZ, thing->origin[VZ]))) {
            return false;
        }
#endif
#if !__JHEXEN__
        if(!thing->player && (thing->intFlags & MIF_FALLING) &&
           tmFloorZ - thing->origin[VZ] > (thing->mom[MX] * thing->mom[MX]) +
                                          (thing->mom[MY] * thing->mom[MY])) {
            return false;
        }
#endif
    }
    vec3d_t oldPos; V3d_Copy(oldPos, thing->origin);
    P_MobjUnlink(thing);
    thing->origin[VX] = x;
    thing->origin[VY] = y;
    thing->floorZ     = tmFloorZ;
    thing->ceilingZ   = tmCeilingZ;
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    thing->dropOffZ   = tmDropoffZ; // $dropoff_fix: keep track of dropoffs.
#endif
    P_MobjLink(thing);
    if(thing->flags2 & MF2_FLOORCLIP) {
        thing->floorClip = 0;
        if(FEQUAL(thing->origin[VZ], P_GetDoublep(Mobj_Sector(thing), DMU_FLOOR_HEIGHT))) {
            terraintype_t const *tt = P_MobjFloorTerrain(thing);
            if(tt->flags & TTF_FLOORCLIP) {
                thing->floorClip = 10;
            }
        }
    }
    if(!(thing->flags & (MF_TELEPORT | MF_NOCLIP))) {
        Line *line;
        while((line = (Line *)IterList_Pop(spechit)) != 0) {
            if(P_ToXLine(line)->special) {
                int side    = Line_PointOnSide(line, thing->origin) < 0;
                int oldSide = Line_PointOnSide(line, oldPos) < 0;
                if(side != oldSide) {
#if __JHEXEN__
                    if(thing->player) {
                        P_ActivateLine(line, thing, oldSide, SPAC_CROSS);
                    }
                    else if(thing->flags2 & MF2_MCROSS) {
                        P_ActivateLine(line, thing, oldSide, SPAC_MCROSS);
                    }
                    else if(thing->flags2 & MF2_PCROSS) {
                        P_ActivateLine(line, thing, oldSide, SPAC_PCROSS);
                    }
#else
                    if(!IS_CLIENT && thing->player) {
                        App_Log(DE2_DEV_MAP_VERBOSE, "P_TryMove2: Mobj %i crossing line %i from %f,%f to %f,%f",
                                thing->thinker.id, P_ToIndex(line),
                                oldPos[VX], oldPos[VY],
                                thing->origin[VX], thing->origin[VY]);
                    }
                    P_ActivateLine(line, thing, oldSide, SPAC_CROSS);
#endif
                }
            }
        }
    }
    return true;
#if __JHEXEN__
  pushline:
    if(!(thing->flags & (MF_TELEPORT | MF_NOCLIP))) {
        if(tmThing->flags2 & MF2_BLASTED) {
            P_DamageMobj(tmThing, NULL, NULL, tmThing->info->mass >> 5, false);
        }
        IterList_SetIteratorDirection(spechit, ITERLIST_BACKWARD);
        IterList_RewindIterator(spechit);
        Line *line;
        while((line = (Line *)IterList_MoveIterator(spechit)) != 0) {
            int side = Line_PointOnSide(line, thing->origin) < 0;
            checkForPushSpecial(line, side, thing);
        }
    }
    return false;
#endif
}
#if __JHEXEN__
dd_bool P_TryMoveXY(mobj_t *thing, coord_t x, coord_t y)
#else
dd_bool P_TryMoveXY(mobj_t *thing, coord_t x, coord_t y, dd_bool dropoff, dd_bool slide)
#endif
{
#if __JHEXEN__
    return P_TryMove2(thing, x, y);
#else
    dd_bool res = P_TryMove2(thing, x, y, dropoff);
    if(!res && tmHitLine) {
        XL_HitLine(tmHitLine, Line_PointOnSide(tmHitLine, thing->origin) < 0,
                   thing);
    }
    if(res && slide) {
        thing->wallRun = true;
    }
    return res;
#endif
}
dd_bool P_TryMoveXYZ(mobj_t* thing, coord_t x, coord_t y, coord_t z) {
    coord_t const oldZ = thing->origin[VZ];
    thing->origin[VZ] = z;
#ifdef __JHEXEN__
    if(P_TryMoveXY(thing, x, y))
#else
    if(P_TryMoveXY(thing, x, y, false, false))
#endif
    {
        return true;
    }
    thing->origin[VZ] = oldZ;
    return false;
}
static mobj_t *spawnPuff(mobjtype_t type, const_pvec3d_t pos, bool noSpark = false) {
#if __JHERETIC__ || __JHEXEN__
    DENG_UNUSED(noSpark);
#endif
    angle_t const angle = P_Random() << 24;
#if !__JHEXEN__
    if(IS_CLIENT) return 0;
#endif
    mobjtype_t puffType = type;
    coord_t zOffset = 0;
#if __JHERETIC__
    if(type == MT_BLASTERPUFF1) {
        puffType = MT_BLASTERPUFF2;
    }
    else
#endif
    {
        zOffset = FIX2FLT((P_Random() - P_Random()) << 10);
    }
    mobj_t *puff = P_SpawnMobjXYZ(puffType, pos[VX], pos[VY], pos[VZ] + zOffset, angle, 0);
    if(puff) {
#if __JHEXEN__
        if(lineTarget && puff->info->seeSound) {
            S_StartSound(puff->info->seeSound, puff);
        }
        else if(puff->info->attackSound) {
            S_StartSound(puff->info->attackSound, puff);
        }
        switch(type) {
        case MT_PUNCHPUFF:  puff->mom[MZ] = 1;   break;
        case MT_HAMMERPUFF: puff->mom[MZ] = .8f; break;
        default: break;
        }
#elif __JHERETIC__
        if(puffType == MT_BLASTERPUFF1) {
            S_StartSound(SFX_BLSHIT, puff);
        }
        else {
            if(puff->info->attackSound) {
                S_StartSound(puff->info->attackSound, puff);
            }
            switch(type) {
            case MT_BEAKPUFF:
            case MT_STAFFPUFF:     puff->mom[MZ] = 1;   break;
            case MT_GAUNTLETPUFF1:
            case MT_GAUNTLETPUFF2: puff->mom[MZ] = .8f; break;
            default: break;
            }
        }
#else
        puff->mom[MZ] = FIX2FLT(FRACUNIT);
        puff->tics -= P_Random() & 3;
        if(puff->tics < 1) puff->tics = 1; // Always at least one tic.
        if(noSpark) {
            P_MobjChangeState(puff, S_PUFF3);
        }
#endif
    }
#if __JHEXEN__
    PuffSpawned = puff;
#endif
    return puff;
}
struct ptr_shoottraverse_params_t {
    mobj_t *shooterMobj; ///< Mobj doing the shooting.
    int damage;          ///< Damage to inflict.
    coord_t range;       ///< Maximum effective range from the trace origin.
    mobjtype_t puffType; ///< Type of puff to spawn.
    bool puffNoSpark;    ///< @c true= Advance the puff to the first non-spark state.
};
/**
 * @todo This routine has gotten way too big, split if(in->isaline)
 *       to a seperate routine?
 */
static int PTR_ShootTraverse(Intercept const *icpt, void *context) {
    vec3d_t const tracePos = {
        Interceptor_Origin(icpt->trace)[VX], Interceptor_Origin(icpt->trace)[VY], shootZ
    };
    ptr_shoottraverse_params_t &parm = *static_cast<ptr_shoottraverse_params_t *>(context);
    if(icpt->type == ICPT_LINE) {
        bool lineWasHit = false;
#ifdef __JHEXEN__
        DENG_UNUSED(lineWasHit);
#endif
        Line *line = icpt->line;
        xline_t *xline = P_ToXLine(line);
        Sector *backSec = (Sector *)P_GetPtrp(line, DMU_BACK_SECTOR);
        if(!backSec || !(xline->flags & ML_TWOSIDED)) {
            if(Line_PointOnSide(line, tracePos) < 0) {
                return false; // Continue traversal.
            }
        }
        if(xline->special) {
            P_ActivateLine(line, parm.shooterMobj, 0, SPAC_IMPACT);
        }
        Sector *frontSec = 0;
        coord_t dist = 0;
        coord_t slope = 0;
        if(!backSec) goto hitline;
#if __JDOOM64__
        if(xline->flags & ML_BLOCKALL) goto hitline;
#endif
        Interceptor_AdjustOpening(icpt->trace, line);
        frontSec = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
        dist = parm.range * icpt->distance;
        slope = 0;
        if(!FEQUAL(P_GetDoublep(frontSec, DMU_FLOOR_HEIGHT),
                   P_GetDoublep(backSec,  DMU_FLOOR_HEIGHT))) {
            slope = (Interceptor_Opening(icpt->trace)->bottom - tracePos[VZ]) / dist;
            if(slope > aimSlope) goto hitline;
        }
        if(!FEQUAL(P_GetDoublep(frontSec, DMU_CEILING_HEIGHT),
                   P_GetDoublep(backSec,  DMU_CEILING_HEIGHT))) {
            slope = (Interceptor_Opening(icpt->trace)->top - tracePos[VZ]) / dist;
            if(slope < aimSlope) goto hitline;
        }
        return false;
      hitline:
        coord_t frac = icpt->distance - (4 / parm.range);
        vec3d_t pos = { tracePos[VX] + Interceptor_Direction(icpt->trace)[VX] * frac,
                        tracePos[VY] + Interceptor_Direction(icpt->trace)[VY] * frac,
                        tracePos[VZ] + aimSlope * (frac * parm.range) };
        if(backSec) {
            if((P_GetIntp(P_GetPtrp(frontSec, DMU_CEILING_MATERIAL),
                          DMU_FLAGS) & MATF_SKYMASK) &&
               (pos[VZ] > P_GetDoublep(frontSec, DMU_CEILING_HEIGHT) ||
                pos[VZ] > P_GetDoublep(backSec,  DMU_CEILING_HEIGHT))) {
                return true;
            }
            if((P_GetIntp(P_GetPtrp(backSec, DMU_FLOOR_MATERIAL),
                          DMU_FLAGS) & MATF_SKYMASK) &&
               (pos[VZ] < P_GetDoublep(frontSec, DMU_FLOOR_HEIGHT) ||
                pos[VZ] < P_GetDoublep(backSec,  DMU_FLOOR_HEIGHT))) {
                return true;
            }
        }
        lineWasHit = true;
        Sector *originSector = Sector_AtPoint_FixedPrecision(tracePos);
        vec3d_t d; V3d_Subtract(d, pos, tracePos);
        if(!INRANGE_OF(d[VZ], 0, .0001f)) // Epsilon
        {
            Sector *contact = Sector_AtPoint_FixedPrecision(pos);
            coord_t step    = M_ApproxDistance3(d[VX], d[VY], d[VZ] * 1.2/*aspect ratio*/);
            vec3d_t stepv   = { d[VX] / step, d[VY] / step, d[VZ] / step };
            coord_t cFloor = P_GetDoublep(contact, DMU_FLOOR_HEIGHT);
            coord_t cCeil  = P_GetDoublep(contact, DMU_CEILING_HEIGHT);
            while(cCeil <= cFloor && contact != originSector) {
                d[VX] -= 8 * stepv[VX];
                d[VY] -= 8 * stepv[VY];
                d[VZ] -= 8 * stepv[VZ];
                pos[VX] = tracePos[VX] + d[VX];
                pos[VY] = tracePos[VY] + d[VY];
                pos[VZ] = tracePos[VZ] + d[VZ];
                contact = Sector_AtPoint_FixedPrecision(pos);
            }
            coord_t cTop    = cCeil - 4;
            coord_t cBottom = cFloor + 4;
            int divisor     = 2;
            if(pos[VZ] > cTop &&
               (P_GetIntp(P_GetPtrp(contact, DMU_CEILING_MATERIAL),
                            DMU_FLAGS) & MATF_SKYMASK)) {
                return true;
            }
            if(pos[VZ] < cBottom &&
               (P_GetIntp(P_GetPtrp(contact, DMU_FLOOR_MATERIAL),
                            DMU_FLAGS) & MATF_SKYMASK)) {
                return true;
            }
            while((pos[VZ] > cTop || pos[VZ] < cBottom) && divisor <= 128) {
                lineWasHit = false;
                pos[VX] -= d[VX] / divisor;
                pos[VY] -= d[VY] / divisor;
                pos[VZ] -= d[VZ] / divisor;
                divisor *= 2;
                if(IS_ZERO(d[VZ] / divisor)) {
                    break; // No.
                }
                while((d[VZ] > 0 && pos[VZ] <= cTop) ||
                      (d[VZ] < 0 && pos[VZ] >= cBottom)) {
                    pos[VX] += d[VX] / divisor;
                    pos[VY] += d[VY] / divisor;
                    pos[VZ] += d[VZ] / divisor;
                }
            }
        }
        spawnPuff(parm.puffType, pos, parm.puffNoSpark);
#if !__JHEXEN__
        if(lineWasHit && xline->special) {
            XL_ShootLine(line, 0, parm.shooterMobj);
        }
#endif
        return true;
    }
    mobj_t *th = icpt->mobj;
    if(th == parm.shooterMobj) return false; // Can't shoot oneself.
    if(!(th->flags & MF_SHOOTABLE)) return false; // Corpse or something.
#if __JHERETIC__
    if ((th->flags & MF_SHADOW) && Mobj_IsPlayer(parm.shooterMobj) &&
        parm.shooterMobj->player->readyWeapon == WT_FIRST) {
        if (!cfg.staffPowerDamageToGhosts || !parm.shooterMobj->player->powers[PT_WEAPONLEVEL2]) {
            return false;
        }
    }
#endif
    coord_t dist = parm.range * icpt->distance;
    coord_t dz   = th->origin[VZ];
    if(!(th->player && (th->player->plr->flags & DDPF_CAMERA))) {
        dz += th->height;
    }
    dz -= tracePos[VZ];
    coord_t thingTopSlope = dz / dist;
    if(thingTopSlope < aimSlope) {
        return false; // Shot over the thing.
    }
    coord_t thingBottomSlope = (th->origin[VZ] - tracePos[VZ]) / dist;
    if(thingBottomSlope > aimSlope) {
        return false; // Shot under the thing.
    }
    coord_t frac = icpt->distance - (10 / parm.range);
    vec3d_t pos  = { tracePos[VX] + Interceptor_Direction(icpt->trace)[VX] * frac,
                     tracePos[VY] + Interceptor_Direction(icpt->trace)[VY] * frac,
                     tracePos[VZ] + aimSlope * (frac * parm.range) };
#if __JHERETIC__ || __JHEXEN__
    spawnPuff(parm.puffType, pos, parm.puffNoSpark);
#endif
    if(parm.damage) {
#if __JDOOM__ || __JDOOM64__
        angle_t attackAngle = M_PointToAngle2(parm.shooterMobj->origin, pos);
#endif
        mobj_t *inflictor = parm.shooterMobj;
#if __JHEXEN__
        if(parm.puffType == MT_FLAMEPUFF2) {
            inflictor = P_LavaInflictor();
        }
#endif
        int damageDone =
            P_DamageMobj(th, inflictor, parm.shooterMobj, parm.damage, false);
#if __JHEXEN__
        if(!(icpt->mobj->flags2 & MF2_INVULNERABLE))
#endif
        {
            if(!(icpt->mobj->flags & MF_NOBLOOD)) {
                if(damageDone > 0) {
#if __JDOOM__ || __JDOOM64__
                    P_SpawnBlood(pos[VX], pos[VY], pos[VZ], parm.damage, attackAngle + ANG180);
#else
# if __JHEXEN__
                    if(parm.puffType == MT_AXEPUFF || parm.puffType == MT_AXEPUFF_GLOW) {
                        P_SpawnBloodSplatter2(pos[VX], pos[VY], pos[VZ], icpt->mobj);
                    }
                    else
# endif
                    if(P_Random() < 192) {
                        P_SpawnBloodSplatter(pos[VX], pos[VY], pos[VZ], icpt->mobj);
                    }
#endif
                }
            }
#if __JDOOM__ || __JDOOM64__
            else {
                spawnPuff(parm.puffType, pos, parm.puffNoSpark);
            }
#endif
        }
    }
    return true;
}
/**
 * Sets linetarget and aimSlope when a target is aimed at.
 */
static int PTR_AimTraverse(Intercept const *icpt, void * /*context*/)
{
    vec3d_t const tracePos = {
        Interceptor_Origin(icpt->trace)[VX], Interceptor_Origin(icpt->trace)[VY], shootZ
    };
    if(icpt->type == ICPT_LINE) {
        Line *line = icpt->line;
        Sector *backSec, *frontSec;
        if(!(P_ToXLine(line)->flags & ML_TWOSIDED) ||
           !(frontSec = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR)) ||
           !(backSec  = (Sector *)P_GetPtrp(line, DMU_BACK_SECTOR))) {
            return !(Line_PointOnSide(line, tracePos) < 0);
        }
        if(!Interceptor_AdjustOpening(icpt->trace, line)) {
            return true; // Stop.
        }
        coord_t dist   = attackRange * icpt->distance;
        coord_t fFloor = P_GetDoublep(frontSec, DMU_FLOOR_HEIGHT);
        coord_t fCeil  = P_GetDoublep(frontSec, DMU_CEILING_HEIGHT);
        coord_t bFloor = P_GetDoublep(backSec, DMU_FLOOR_HEIGHT);
        coord_t bCeil  = P_GetDoublep(backSec, DMU_CEILING_HEIGHT);
        coord_t slope;
        if(!FEQUAL(fFloor, bFloor)) {
            slope = (Interceptor_Opening(icpt->trace)->bottom - shootZ) / dist;
            if(slope > bottomSlope)
                bottomSlope = slope;
        }
        if(!FEQUAL(fCeil, bCeil)) {
            slope = (Interceptor_Opening(icpt->trace)->top - shootZ) / dist;
            if(slope < topSlope)
                topSlope = slope;
        }
        return topSlope <= bottomSlope;
    }
    mobj_t *th = icpt->mobj;
    if(th == shooterThing) return false; // Can't aim at oneself.
    if(!(th->flags & MF_SHOOTABLE)) return false; // Corpse or something (not shootable)?
#if __JHERETIC__
    if(th->type == MT_POD) return false; // Can't auto-aim at pods.
#endif
#if __JDOOM__ || __JHEXEN__ || __JDOOM64__
    if(Mobj_IsPlayer(shooterThing) && Mobj_IsPlayer(th) &&
       IS_NETGAME && !gfw_Rule(deathmatch)) {
        return false;
    }
#endif
    coord_t dist = attackRange * icpt->distance;
    coord_t posZ = th->origin[VZ];
    if(!(th->player && (th->player->plr->flags & DDPF_CAMERA))) {
        posZ += th->height;
    }
    coord_t thingTopSlope = (posZ - shootZ) / dist;
    if(thingTopSlope < bottomSlope) {
        return false; // Shot over the thing.
    }
#if __JHEXEN__
    if(posZ < shootZ - attackRange / 1.2f) {
        return false;
    }
#endif
    coord_t thingBottomSlope = (th->origin[VZ] - shootZ) / dist;
    if(thingBottomSlope > topSlope) {
        return false; // Shot under the thing.
    }
#if __JHEXEN__
    if(th->origin[VZ] > shootZ + attackRange / 1.2f) {
        return false;
    }
#endif
    if(thingTopSlope > topSlope) {
        thingTopSlope = topSlope;
    }
    if(thingBottomSlope < bottomSlope) {
        thingBottomSlope = bottomSlope;
    }
    aimSlope = (thingTopSlope + thingBottomSlope) / 2;
    lineTarget = th;
    return true; // Don't go any farther.
}
float P_AimLineAttack(mobj_t *t1, angle_t angle, coord_t distance) {
    uint an = angle >> ANGLETOFINESHIFT;
    vec2d_t target = { t1->origin[VX] + distance * FIX2FLT(finecosine[an]),
                       t1->origin[VY] + distance * FIX2FLT(finesine[an]) };
    shootZ = t1->origin[VZ];
#if __JHEXEN__
    if(t1->player &&
      (t1->player->class_ == PCLASS_FIGHTER ||
       t1->player->class_ == PCLASS_CLERIC ||
       t1->player->class_ == PCLASS_MAGE))
#else
    if(t1->player && t1->type == MT_PLAYER)
#endif
    {
        if(!(t1->player->plr->flags & DDPF_CAMERA)) {
            shootZ += (cfg.common.plrViewHeight - 5);
        }
    }
    else {
        shootZ += (t1->height / 2) + 8;
    }
    topSlope     = 100.0/160;
    bottomSlope  = -100.0/160;
    attackRange  = distance;
    lineTarget   = 0;
    shooterThing = t1;
    P_PathTraverse(t1->origin, target, PTR_AimTraverse, 0);
    if(lineTarget) {
        if(!t1->player || !cfg.common.noAutoAim) {
            return aimSlope;
        }
    }
    if(t1->player && cfg.common.noAutoAim) {
        return tan(LOOKDIR2RAD(t1->dPlayer->lookDir)) / 1.2;
    }
    return 0;
}
void P_LineAttack(mobj_t *t1, angle_t angle, coord_t distance, coord_t slope,
    int damage, mobjtype_t puffType) {
    uint an = angle >> ANGLETOFINESHIFT;
    vec2d_t target = { t1->origin[VX] + distance * FIX2FLT(finecosine[an]),
                       t1->origin[VY] + distance * FIX2FLT(finesine[an]) };
    aimSlope = slope;
    shootZ = t1->origin[VZ];
#if __JHEXEN__
    if(t1->player &&
      (t1->player->class_ == PCLASS_FIGHTER ||
       t1->player->class_ == PCLASS_CLERIC ||
       t1->player->class_ == PCLASS_MAGE))
#else
    if(t1->player && t1->type == MT_PLAYER)
#endif
    {
        if(!(t1->player->plr->flags & DDPF_CAMERA)) {
            shootZ += cfg.common.plrViewHeight - 5;
        }
    }
    else {
        shootZ += (t1->height / 2) + 8;
    }
    shootZ -= t1->floorClip;
    ptr_shoottraverse_params_t parm;
    parm.shooterMobj = t1;
    parm.range       = distance;
    parm.damage      = damage;
    parm.puffType    = puffType;
#if __JDOOM__ || __JDOOM64__
    parm.puffNoSpark = attackRange == MELEERANGE;
#else
    parm.puffNoSpark = false;
#endif
    if(!P_PathTraverse(t1->origin, target, PTR_ShootTraverse, &parm)) {
#if __JHEXEN__
        switch(puffType) {
        case MT_PUNCHPUFF:
            S_StartSound(SFX_FIGHTER_PUNCH_MISS, t1);
            break;
        case MT_HAMMERPUFF:
        case MT_AXEPUFF:
        case MT_AXEPUFF_GLOW:
            S_StartSound(SFX_FIGHTER_HAMMER_MISS, t1);
            break;
        case MT_FLAMEPUFF: {
            vec3d_t pos = { target[VX], target[VY], shootZ + (slope * distance) };
            spawnPuff(puffType, pos);
            break; }
        default: break;
        }
#endif
    }
}
struct pit_radiusattack_params_t {
    mobj_t *source;     ///< Mobj which caused the attack.
    mobj_t *bomb;       ///< Epicenter of the attack.
    int damage;         ///< Maximum damage to inflict.
    int distance;       ///< Maximum distance within which to afflict.
#ifdef __JHEXEN__
    bool afflictSource; ///< @c true= Afflict the source, also.
#endif
};
static int PIT_RadiusAttack(mobj_t *thing, void *context) {
    pit_radiusattack_params_t &parm = *static_cast<pit_radiusattack_params_t *>(context);
    if(!(thing->flags & MF_SHOOTABLE)) {
        return false;
    }
#if __JHERETIC__
    if(thing->type == MT_MINOTAUR) return false;
    if(thing->type == MT_SORCERER1) return false;
    if(thing->type == MT_SORCERER2) return false;
#elif __JDOOM__ || __JDOOM64__
    if(thing->type == MT_CYBORG) return false;
# if __JDOOM__
    if(thing->type == MT_SPIDER) return false;
# endif
#endif
#if __JHEXEN__
    if(thing == parm.source && !parm.afflictSource) {
        return false;
    }
#endif
    vec3d_t delta = { fabs(thing->origin[VX] - parm.bomb->origin[VX]),
                      fabs(thing->origin[VY] - parm.bomb->origin[VY]),
                      fabs((thing->origin[VZ] + thing->height / 2) - parm.bomb->origin[VZ]) };
    coord_t dist = (delta[VX] > delta[VY]? delta[VX] : delta[VY]);
#if __JHEXEN__
    if(!cfg.common.netNoMaxZRadiusAttack) {
        dist = (delta[VZ] > dist? delta[VZ] : dist);
    }
#else
    if(!(cfg.common.netNoMaxZRadiusAttack || (thing->info->flags2 & MF2_INFZBOMBDAMAGE))) {
        dist = (delta[VZ] > dist? delta[VZ] : dist);
    }
#endif
    dist = MAX_OF(dist - thing->radius, 0);
    if(dist >= parm.distance) {
        return false; // Out of range.
    }
    if(P_CheckSight(thing, parm.bomb)) {
        int damage = (parm.damage * (parm.distance - dist) / parm.distance) + 1;
#if __JHEXEN__
        if(thing->player) damage /= 4;
#endif
        P_DamageMobj(thing, parm.bomb, parm.source, damage, false);
    }
    return false;
}
#if __JHEXEN__
void P_RadiusAttack(mobj_t *bomb, mobj_t *source, int damage, int distance, dd_bool afflictSource)
#else
void P_RadiusAttack(mobj_t *bomb, mobj_t *source, int damage, int distance)
#endif
{
    coord_t const dist = distance + MAXRADIUS;
    AABoxd const box(bomb->origin[VX] - dist, bomb->origin[VY] - dist,
                     bomb->origin[VX] + dist, bomb->origin[VY] + dist);
    pit_radiusattack_params_t parm;
    parm.bomb          = bomb;
    parm.damage        = damage;
    parm.distance      = distance;
    parm.source        = source;
#if __JHERETIC__
    if(bomb->type == MT_POD && bomb->target) {
        parm.source = bomb->target;
    }
#endif
#if __JHEXEN__
    parm.afflictSource = CPP_BOOL(afflictSource);
#endif
    VALIDCOUNT++;
    Mobj_BoxIterator(&box, PIT_RadiusAttack, &parm);
}
static int PTR_UseTraverse(Intercept const *icpt, void *context) {
    DENG_ASSERT(icpt->type == ICPT_LINE);
    mobj_t *activator = static_cast<mobj_t *>(context);
    xline_t *xline = P_ToXLine(icpt->line);
    if(!xline->special) {
        if(!Interceptor_AdjustOpening(icpt->trace, icpt->line)) {
            if(Mobj_IsPlayer(activator)) {
                S_StartSound(PCLASS_INFO(activator->player->class_)->failUseSound, activator);
            }
            return true; // Can't use through a wall.
        }
#if __JHEXEN__
        if(Mobj_IsPlayer(activator)) {
            coord_t pheight = activator->origin[VZ] + activator->height/2;
            if(Interceptor_Opening(icpt->trace)->top < pheight ||
               Interceptor_Opening(icpt->trace)->bottom > pheight) {
                S_StartSound(PCLASS_INFO(activator->player->class_)->failUseSound, activator);
            }
        }
#endif
        return false;
    }
    int side = Line_PointOnSide(icpt->line, activator->origin) < 0;
#if __JHERETIC__ || __JHEXEN__
    if(side == 1) return true; // Don't use back side.
#endif
    P_ActivateLine(icpt->line, activator, side, SPAC_USE);
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    if(xline->flags & ML_PASSUSE) {
        return false;
    }
#endif
    return true;
}
void P_UseLines(player_t *player) {
    if(!player) return;
    if(IS_CLIENT) {
        App_Log(DE2_DEV_NET_VERBOSE, "P_UseLines: Sending a use request for player %i", int(player - players));
        NetCl_PlayerActionRequest(player, GPA_USE, 0);
        return;
    }
    mobj_t *mo  = player->plr->mo;
    if(!mo) return;
    uint an     = mo->angle >> ANGLETOFINESHIFT;
    vec2d_t pos = { mo->origin[VX] + USERANGE * FIX2FLT(finecosine[an]),
                    mo->origin[VY] + USERANGE * FIX2FLT(finesine  [an]) };
    P_PathTraverse2(mo->origin, pos, PTF_LINE, PTR_UseTraverse, mo);
}
/**
 * Allows the player to slide along any angled walls by adjusting the
 * xmove / ymove so that the NEXT move will slide along the wall.
 *
 * @param line  The line being slid along.
 */
static void hitSlideLine(mobj_t *slideMo, Line *line, pvec2d_t move) {
    DENG_ASSERT(slideMo != 0 && line != 0);
    slopetype_t slopeType = slopetype_t(P_GetIntp(line, DMU_SLOPETYPE));
    if(slopeType == ST_HORIZONTAL) {
        move[MY] = 0;
        return;
    }
    else if(slopeType == ST_VERTICAL) {
        move[MX] = 0;
        return;
    }
    bool side = Line_PointOnSide(line, slideMo->origin) < 0;
    vec2d_t d1; P_GetDoublepv(line, DMU_DXY, d1);
    angle_t moveAngle = M_PointToAngle(move);
    angle_t lineAngle = M_PointToAngle(d1) + (side? ANG180 : 0);
    angle_t deltaAngle = moveAngle - lineAngle;
    if(deltaAngle > ANG180) deltaAngle += ANG180;
    coord_t moveLen = M_ApproxDistance(move[MX], move[MY]);
    coord_t newLen  = moveLen * FIX2FLT(finecosine[deltaAngle >> ANGLETOFINESHIFT]);
    uint an = lineAngle >> ANGLETOFINESHIFT;
    V2d_Set(move, newLen * FIX2FLT(finecosine[an]),
                  newLen * FIX2FLT(finesine  [an]));
}
struct ptr_slidetraverse_params_t {
    mobj_t *slideMobj;
    Line *bestLine;
    coord_t bestDistance;
};
static int PTR_SlideTraverse(Intercept const *icpt, void *context) {
    DENG_ASSERT(icpt->type == ICPT_LINE);
    ptr_slidetraverse_params_t &parm = *static_cast<ptr_slidetraverse_params_t *>(context);
    Line *line = icpt->line;
    if(!(P_ToXLine(line)->flags & ML_TWOSIDED) ||
       !P_GetPtrp(line, DMU_FRONT_SECTOR) || !P_GetPtrp(line, DMU_BACK_SECTOR)) {
        if(Line_PointOnSide(line, parm.slideMobj->origin) < 0) {
            return false; // Don't hit the back side.
        }
        goto isblocking;
    }
#if __JDOOM64__
    if(P_ToXLine(line)->flags & ML_BLOCKALL) // jd64
        goto isblocking;
#endif
    Interceptor_AdjustOpening(icpt->trace, line);
    if(Interceptor_Opening(icpt->trace)->range < parm.slideMobj->height)
        goto isblocking; // Doesn't fit.
    if(Interceptor_Opening(icpt->trace)->top - parm.slideMobj->origin[VZ] < parm.slideMobj->height)
        goto isblocking; // mobj is too high.
    if(Interceptor_Opening(icpt->trace)->bottom - parm.slideMobj->origin[VZ] > 24)
        goto isblocking; // Too big a step up.
    return false;
  isblocking:
    if(icpt->distance < parm.bestDistance) {
        parm.bestDistance = icpt->distance;
        parm.bestLine     = line;
    }
    return true; // Stop.
}
void P_SlideMove(mobj_t *mo) {
    if(!mo) return; // Huh?
    vec2d_t oldOrigin;
    V2d_Copy(oldOrigin, mo->origin);
    vec2d_t leadPos  = { 0, 0 };
    vec2d_t trailPos = { 0, 0 };
    vec2d_t tmMove   = { 0, 0 };
    int hitCount = 3;
    do {
        if(--hitCount == 0)
            goto stairstep; // Don't loop forever.
        leadPos[VX] = mo->origin[VX] + (mo->mom[MX] > 0? mo->radius : -mo->radius);
        leadPos[VY] = mo->origin[VY] + (mo->mom[MY] > 0? mo->radius : -mo->radius);
        trailPos[VX] = mo->origin[VX] - (mo->mom[MX] > 0? mo->radius : -mo->radius);
        trailPos[VY] = mo->origin[VY] - (mo->mom[MY] > 0? mo->radius : -mo->radius);
        ptr_slidetraverse_params_t parm;
        parm.slideMobj    = mo;
        parm.bestLine     = 0;
        parm.bestDistance = 1;
        P_PathXYTraverse2(leadPos[VX], leadPos[VY],
                          leadPos[VX] + mo->mom[MX], leadPos[VY] + mo->mom[MY],
                          PTF_LINE, PTR_SlideTraverse, &parm);
        P_PathXYTraverse2(trailPos[VX], leadPos[VY],
                          trailPos[VX] + mo->mom[MX], leadPos[VY] + mo->mom[MY],
                          PTF_LINE, PTR_SlideTraverse, &parm);
        P_PathXYTraverse2(leadPos[VX], trailPos[VY],
                          leadPos[VX] + mo->mom[MX], trailPos[VY] + mo->mom[MY],
                          PTF_LINE, PTR_SlideTraverse, &parm);
        if(parm.bestDistance == 1) {
          stairstep:
            /*
             * Ideally we would set the directional momentum of the mobj to zero
             * here should a move fail (to prevent noticeable stuttering against
             * the blocking surface/thing). However due to the mechanics of the
             * wall side algorithm this is not possible as it results in highly
             * unpredictable behaviour and resulting in the player sling-shoting
             * away from the wall.
             */
#if __JHEXEN__
            if(!P_TryMoveXY(mo, mo->origin[VX], mo->origin[VY] + mo->mom[MY]))
                P_TryMoveXY(mo, mo->origin[VX] + mo->mom[MX], mo->origin[VY]);
#else
            if(!P_TryMoveXY(mo, mo->origin[VX], mo->origin[VY] + mo->mom[MY], true, true))
                P_TryMoveXY(mo, mo->origin[VX] + mo->mom[MX], mo->origin[VY], true, true);
#endif
            break;
        }
        parm.bestDistance -= (1.0f / 32);
        if(parm.bestDistance > 0) {
            vec2d_t newPos = { mo->origin[VX] + mo->mom[MX] * parm.bestDistance,
                               mo->origin[VY] + mo->mom[MY] * parm.bestDistance };
#if __JHEXEN__
            if(!P_TryMoveXY(mo, newPos[VX], newPos[VY]))
#else
            if(!P_TryMoveXY(mo, newPos[VX], newPos[VY], true, true))
#endif
            {
                goto stairstep;
            }
        }
        parm.bestDistance = MIN_OF(1 - (parm.bestDistance + (1.0f / 32)), 1);
        if(parm.bestDistance <= 0) {
            break;
        }
        V2d_Set(tmMove, mo->mom[VX] * parm.bestDistance,
                        mo->mom[VY] * parm.bestDistance);
        hitSlideLine(mo, parm.bestLine, tmMove); // Clip the move.
        V2d_Copy(mo->mom, tmMove);
#if __JHEXEN__
    } while(!P_TryMoveXY(mo, mo->origin[VX] + tmMove[MX],
                             mo->origin[VY] + tmMove[MY]));
#else
    } while(!P_TryMoveXY(mo, mo->origin[VX] + tmMove[MX],
                             mo->origin[VY] + tmMove[MY], true, true));
#endif
    if(mo->player && mo->origin[VX] == oldOrigin[VX] && mo->origin[VY] == oldOrigin[VY]) {
        App_Log(DE2_DEV_MAP_MSG, "P_SlideMove: Mobj %i pos stays the same", mo->thinker.id);
    }
}
/**
 * SECTOR HEIGHT CHANGING
 * After modifying a sectors floor or ceiling height, call this routine
 * to adjust the positions of all things that touch the sector.
 *
 * If anything doesn't fit anymore, true will be returned.
 * If crushDamage is non-zero, they will take damage as they are being crushed.
 * If crushDamage is false, you should set the sector height back the way it
 * was and call P_ChangeSector again to undo the changes.
 */
struct pit_changesector_params_t {
    int crushDamage;  ///< Damage amount;
    bool noFit;
};
static int PIT_ChangeSector(mobj_t *thing, void *context) {
    pit_changesector_params_t &parm = *static_cast<pit_changesector_params_t *>(context);
    if(!thing->info) {
        return false;
    }
    if(thing->info->flags & MF_NOBLOCKMAP) {
        return false;
    }
    if (!P_MobjIsCamera(thing)) {
        const bool onfloor = de::fequal(thing->origin[VZ], thing->floorZ);
        P_CheckPosition(thing, thing->origin);
        thing->floorZ   = tmFloorZ;
        thing->ceilingZ = tmCeilingZ;
#if !__JHEXEN__
        thing->dropOffZ = tmDropoffZ; // $dropoff_fix: remember dropoffs.
#endif
        if (onfloor) {
#if __JHEXEN__
            if((thing->origin[VZ] - thing->floorZ < 9) ||
               (thing->flags & MF_NOGRAVITY)) {
                thing->origin[VZ] = thing->floorZ;
            }
#else
            if(Mobj_IsPlayer(thing) && !Mobj_IsVoodooDoll(thing)) {
                thing->player->viewZ += thing->floorZ - thing->origin[VZ];
            }
            thing->origin[VZ] = thing->floorZ;
            if((thing->intFlags & MIF_FALLING) && thing->gear >= MAXGEAR) {
                thing->gear = 0;
            }
#endif
        }
        else {
            if(thing->origin[VZ] + thing->height > thing->ceilingZ) {
                thing->origin[VZ] = thing->ceilingZ - thing->height;
            }
        }
        if((thing->ceilingZ - thing->floorZ) >= thing->height) {
            return false;
        }
    }
    if(Mobj_IsCrunchable(thing)) {
#if __JHEXEN__
        if(thing->flags & MF_NOBLOOD) {
            P_MobjRemove(thing, false);
            return false;
        }
#endif
#if __JHEXEN__
        if(thing->state != &STATES[S_GIBS1])
#endif
        {
#if __JHEXEN__
            P_MobjChangeState(thing, S_GIBS1);
#elif __JDOOM__ || __JDOOM64__
            P_MobjChangeState(thing, S_GIBS);
#endif
#if !__JHEXEN__
            thing->flags &= ~MF_SOLID;
#endif
            thing->height = 0;
            thing->radius = 0;
#if __JHEXEN__
            S_StartSound(SFX_PLAYER_FALLING_SPLAT, thing);
#elif __JDOOM64__
            S_StartSound(SFX_SLOP, thing);
#endif
        }
        return false;
    }
    if(Mobj_IsDroppedItem(thing)) {
        P_MobjRemove(thing, false);
        return false;
    }
    if(!(thing->flags & MF_SHOOTABLE)) {
        return false;
    }
    parm.noFit = true;
    if(parm.crushDamage > 0 && !(mapTime & 3)) {
        P_DamageMobj(thing, NULL, NULL, parm.crushDamage, false);
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
        if(!(thing->flags & MF_NOBLOOD))
#elif __JHEXEN__
        if(!(thing->flags & MF_NOBLOOD) &&
           !(thing->flags2 & MF2_INVULNERABLE))
#endif
        {
            if(mobj_t *mo = P_SpawnMobjXYZ(MT_BLOOD, thing->origin[VX], thing->origin[VY],
                                           thing->origin[VZ] + (thing->height /2),
                                           P_Random() << 24, 0)) {
                mo->mom[MX] = FIX2FLT((P_Random() - P_Random()) << 12);
                mo->mom[MY] = FIX2FLT((P_Random() - P_Random()) << 12);
            }
        }
    }
    return false;
}
dd_bool P_ChangeSector(Sector *sector, int crush) {
    pit_changesector_params_t parm;
    parm.noFit       = false;
#if __JHEXEN__
    parm.crushDamage = crush;
#else
    parm.crushDamage = crush > 0? 10 : 0;
#endif
    VALIDCOUNT++;
    Sector_TouchingMobjsIterator(sector, PIT_ChangeSector, &parm);
    return parm.noFit;
}
void P_HandleSectorHeightChange(int sectorIdx) {
    P_ChangeSector((Sector *)P_ToPtr(DMU_SECTOR, sectorIdx), false /*don't crush*/);
}
int P_IterateThinkers(thinkfunc_t func, const std::function<int(thinker_t *)> &callback) {
    struct IterContext {
        const std::function<int(thinker_t *)> *func;
        static int callback(thinker_t *thinker, void *ptr) {
            auto *context = reinterpret_cast<IterContext *>(ptr);
            return (*context->func)(thinker);
        }
    };
    IterContext context{&callback};
    return Thinker_Iterate(func, IterContext::callback, &context);
}
#if defined (__JHERETIC__) || defined(__JHEXEN__)
dd_bool P_TestMobjLocation(mobj_t *mo) {
    int const oldFlags = mo->flags;
    mo->flags &= ~MF_PICKUP;
    if(!P_CheckPositionXY(mo, mo->origin[VX], mo->origin[VY])) {
        mo->flags = oldFlags;
        return false;
    }
    mo->flags = oldFlags;
    return mo->origin[VZ] >= mo->floorZ && (mo->origin[VZ] + mo->height) <= mo->ceilingZ;
}
struct ptr_boucetraverse_params_t {
    mobj_t *bounceMobj;
    Line *  bestLine;
    coord_t bestDistance;
};
static int PTR_BounceTraverse(Intercept const *icpt, void *context) {
    DENG_ASSERT(icpt->type == ICPT_LINE);
    ptr_boucetraverse_params_t &parm = *static_cast<ptr_boucetraverse_params_t *>(context);
    Line *line = icpt->line;
    if (!P_GetPtrp(line, DMU_FRONT_SECTOR) || !P_GetPtrp(line, DMU_BACK_SECTOR)) {
        if (Line_PointOnSide(line, parm.bounceMobj->origin) < 0) {
            return false; // Don't hit the back side.
        }
        goto bounceblocking;
    }
    Interceptor_AdjustOpening(icpt->trace, line);
    if (Interceptor_Opening(icpt->trace)->range < parm.bounceMobj->height) {
        goto bounceblocking; // Doesn't fit.
    }
    if (Interceptor_Opening(icpt->trace)->top - parm.bounceMobj->origin[VZ] <
        parm.bounceMobj->height) {
        goto bounceblocking; // Mobj is too high...
    }
    if (parm.bounceMobj->origin[VZ] - Interceptor_Opening(icpt->trace)->bottom < 0) {
        goto bounceblocking; // Mobj is too low...
    }
    return false;
bounceblocking:
    if (icpt->distance < parm.bestDistance) {
        parm.bestDistance = icpt->distance;
        parm.bestLine     = line;
    }
    return false;
}
dd_bool P_BounceWall(mobj_t *mo) {
    if (!mo) return false;
    vec2d_t leadPos = {mo->origin[VX] + (mo->mom[MX] > 0 ? mo->radius : -mo->radius),
                       mo->origin[VY] + (mo->mom[MY] > 0 ? mo->radius : -mo->radius)};
    vec2d_t destPos;
    V2d_Sum(destPos, leadPos, mo->mom);
    ptr_boucetraverse_params_t parm;
    parm.bounceMobj   = mo;
    parm.bestLine     = 0;
    parm.bestDistance = 1; // Intercept distances are normalized [0..1]
    P_PathTraverse2(leadPos, destPos, PTF_LINE, PTR_BounceTraverse, &parm);
    if (parm.bestLine) {
        int const side = Line_PointOnSide(parm.bestLine, mo->origin) < 0;
        vec2d_t   lineDirection;
        P_GetDoublepv(parm.bestLine, DMU_DXY, lineDirection);
        angle_t lineAngle  = M_PointToAngle(lineDirection) + (side ? ANG180 : 0);
        angle_t moveAngle  = M_PointToAngle(mo->mom);
        angle_t deltaAngle = (2 * lineAngle) - moveAngle;
        coord_t moveLen = M_ApproxDistance(mo->mom[MX], mo->mom[MY]) * 0.75f /*Friction*/;
        if (moveLen < 1) moveLen = 2;
        uint an = deltaAngle >> ANGLETOFINESHIFT;
        V2d_Set(mo->mom, moveLen * FIX2FLT(finecosine[an]), moveLen * FIX2FLT(finesine[an]));
#if defined (__JHERETIC__)
        mo->mom[MX] *= 0.9;
        mo->mom[MY] *= 0.9;
        S_StartSound(SFX_BOUNCE, mo);
#endif
        return true;
    }
    return false;
}
#endif
#if __JHEXEN__
static int PIT_ThrustStompThing(mobj_t *thing, void *context) {
    mobj_t *tsThing = static_cast<mobj_t *>(context);
    if(thing == tsThing) {
        return false;
    }
    if(!(thing->flags & MF_SHOOTABLE)) {
        return false;
    }
    coord_t blockdist = thing->radius + tsThing->radius;
    if(fabs(thing->origin[VX] - tsThing->origin[VX]) >= blockdist ||
       fabs(thing->origin[VY] - tsThing->origin[VY]) >= blockdist ||
       (thing->origin[VZ] > tsThing->origin[VZ] + tsThing->height)) {
        return false; // Didn't hit it.
    }
    P_DamageMobj(thing, tsThing, tsThing, 10001, false);
    tsThing->args[1] = 1; // Mark thrust thing as bloody.
    return false;
}
void P_ThrustSpike(mobj_t *mobj) {
    if(!mobj) return;
    coord_t const radius = mobj->info->radius + MAXRADIUS;
    AABoxd const box(mobj->origin[VX] - radius, mobj->origin[VY] - radius,
                     mobj->origin[VX] + radius, mobj->origin[VY] + radius);
    VALIDCOUNT++;
    Mobj_BoxIterator(&box, PIT_ThrustStompThing, mobj);
}
struct pit_checkonmobjz_params_t {
    mobj_t *riderMobj;
    mobj_t *mountMobj;
};
struct SavedPhysicalState {
    coord_t origin[3];
    coord_t mom[3];
    SavedPhysicalState(mobj_t const *mo) {
        memcpy(origin, mo->origin, sizeof(origin));
        memcpy(mom, mo->mom, sizeof(mom));
    }
    void restore(mobj_t *mo) const {
        memcpy(mo->origin, origin, sizeof(origin));
        memcpy(mo->mom, mom, sizeof(mom));
    }
};
static int PIT_CheckOnMobjZ(mobj_t *cand, void *context) {
    pit_checkonmobjz_params_t &parm = *static_cast<pit_checkonmobjz_params_t *>(context);
    if(cand == parm.riderMobj) {
        return false;
    }
    if(!(cand->flags & (MF_SOLID | MF_SPECIAL | MF_SHOOTABLE))) {
        return false; // Can't hit thing.
    }
    coord_t blockdist = cand->radius + parm.riderMobj->radius;
    if(fabs(cand->origin[VX] - parm.riderMobj->origin[VX]) >= blockdist ||
       fabs(cand->origin[VY] - parm.riderMobj->origin[VY]) >= blockdist) {
        return false; // Didn't hit thing.
    }
    if(IS_CLIENT) {
        if(Mobj_IsPlayer(parm.riderMobj)) {
            if(cand == ClPlayer_ClMobj(parm.riderMobj->player - players)) {
                return false;
            }
        }
    }
    if(parm.riderMobj->origin[VZ] > cand->origin[VZ] + cand->height) {
        return false;
    }
    else if(parm.riderMobj->origin[VZ] + parm.riderMobj->height < cand->origin[VZ]) {
        return false;
    }
    if(cand->flags & MF_SOLID) {
        parm.mountMobj = cand;
    }
    return (cand->flags & MF_SOLID) != 0;
}
mobj_t *P_CheckOnMobj(mobj_t *mo) {
    if(!mo) return 0;
    if(P_MobjIsCamera(mo)) return 0;
    if(Mobj_IsPlayerClMobj(mo)) return 0;
    SavedPhysicalState savedState(mo);
    mo->origin[VZ] += mo->mom[MZ];
    if((mo->flags & MF_FLOAT) && mo->target) {
        if(!(mo->flags & MF_SKULLFLY) && !(mo->flags & MF_INFLOAT)) {
            coord_t dist = M_ApproxDistance(mo->origin[VX] - mo->target->origin[VX],
                                            mo->origin[VY] - mo->target->origin[VY]);
            coord_t delta = mo->target->origin[VZ] + (mo->height / 2) - mo->origin[VZ];
            if(delta < 0 && dist < -(delta * 3)) {
                mo->origin[VZ] -= FLOATSPEED;
            }
            else if(delta > 0 && dist < (delta * 3)) {
                mo->origin[VZ] += FLOATSPEED;
            }
        }
    }
    if(Mobj_IsPlayer(mo) && (mo->flags2 & MF2_FLY) && !(mo->origin[VZ] <= mo->floorZ)) {
        if(mapTime & 2) {
            mo->origin[VZ] += FIX2FLT(finesine[(FINEANGLES / 20 * mapTime >> 2) & FINEMASK]);
        }
    }
    bool hitFloor = mo->origin[VZ] <= mo->floorZ;
    if(hitFloor) {
        mo->origin[VZ] = mo->floorZ;
        if(mo->mom[MZ] < 0) {
            mo->mom[MZ] = 0;
        }
        if(mo->flags & MF_SKULLFLY) {
            mo->mom[MZ] = -mo->mom[MZ]; // The skull slammed into something
        }
    }
    else if(mo->flags2 & MF2_LOGRAV) {
        if(IS_ZERO(mo->mom[MZ])) {
            mo->mom[MZ] = -(P_GetGravity() / 32) * 2;
        }
        else {
            mo->mom[MZ] -= P_GetGravity() / 32;
        }
    }
    else if(!(mo->flags & MF_NOGRAVITY)) {
        if(IS_ZERO(mo->mom[MZ])) {
            mo->mom[MZ] = -P_GetGravity() * 2;
        }
        else {
            mo->mom[MZ] -= P_GetGravity();
        }
    }
    if(!(hitFloor && P_GetState(mobjtype_t(mo->type), SN_CRASH) && (mo->flags & MF_CORPSE))) {
        if(mo->origin[VZ] + mo->height > mo->ceilingZ) {
            mo->origin[VZ] = mo->ceilingZ - mo->height;
            if(mo->mom[MZ] > 0) {
                mo->mom[MZ] = 0;
            }
            if(mo->flags & MF_SKULLFLY) {
                mo->mom[MZ] = -mo->mom[MZ]; // The skull slammed into something.
            }
        }
    }
    /*tmCeilingLine = tmFloorLine = 0;
    Sector *newSector = Sector_AtPoint_FixedPrecision(mo->origin);
    tmFloorZ        = tmDropoffZ = P_GetDoublep(newSector, DMU_FLOOR_HEIGHT);
    tmCeilingZ      = P_GetDoublep(newSector, DMU_CEILING_HEIGHT);
    tmFloorMaterial = (Material *)P_GetPtrp(newSector, DMU_FLOOR_MATERIAL);
    IterList_Clear(spechit);*/
    if(!(mo->flags & MF_NOCLIP)) {
        int blockdist = mo->radius + MAXRADIUS;
        AABoxd aaBox(mo->origin[VX] - blockdist, mo->origin[VY] - blockdist,
                     mo->origin[VX] + blockdist, mo->origin[VY] + blockdist);
        pit_checkonmobjz_params_t parm;
        parm.riderMobj = mo;
        parm.mountMobj = 0;
        VALIDCOUNT++;
        if(Mobj_BoxIterator(&aaBox, PIT_CheckOnMobjZ, &parm)) {
            savedState.restore(mo);
            return parm.mountMobj;
        }
    }
    savedState.restore(mo);
    return 0;
}
static sfxenum_t usePuzzleItemFailSound(mobj_t *user) {
    if(Mobj_IsPlayer(user)) {
        switch(user->player->class_) {
        case PCLASS_FIGHTER: return SFX_PUZZLE_FAIL_FIGHTER;
        case PCLASS_CLERIC:  return SFX_PUZZLE_FAIL_CLERIC;
        case PCLASS_MAGE:    return SFX_PUZZLE_FAIL_MAGE;
        default: break;
        }
    }
    return SFX_NONE;
}
struct ptr_puzzleitemtraverse_params_t {
    mobj_t *useMobj;
    int itemType;
    bool activated;
};
static int PTR_PuzzleItemTraverse(Intercept const *icpt, void *context) {
    int const USE_PUZZLE_ITEM_SPECIAL = 129;
    auto &parm = *static_cast<ptr_puzzleitemtraverse_params_t *>(context);
    switch(icpt->type) {
    case ICPT_LINE: {
        xline_t *xline = P_ToXLine(icpt->line);
        DENG2_ASSERT(xline);
        if(xline->special != USE_PUZZLE_ITEM_SPECIAL) {
            if(!Interceptor_AdjustOpening(icpt->trace, icpt->line)) {
                S_StartSound(usePuzzleItemFailSound(parm.useMobj), parm.useMobj);
                return true;
            }
            return false;
        }
        if(Line_PointOnSide(icpt->line, parm.useMobj->origin) < 0) {
            return true;
        }
        if(parm.itemType != xline->arg1) {
            return true;
        }
        if(gfw_Session()->acsSystem().hasScript(xline->arg2)) {
            gfw_Session()->acsSystem()
                    .script(xline->arg2)
                        .start(acs::Script::Args(&xline->arg3, 4/*3*/), parm.useMobj,
                               icpt->line, 0);
        }
        xline->special = 0;
        parm.activated = true;
        return true; }
    case ICPT_MOBJ: {
        DENG2_ASSERT(icpt->mobj);
        mobj_t &mob = *icpt->mobj;
        if(mob.special != USE_PUZZLE_ITEM_SPECIAL) {
            return false;
        }
        if(mob.args[0] != parm.itemType) {
            return false;
        }
        if(gfw_Session()->acsSystem().hasScript(mob.args[1])) {
            gfw_Session()->acsSystem()
                    .script(mob.args[1])
                        .start(acs::Script::Args(&mob.args[2], 4/*3*/), parm.useMobj,
                               nullptr, 0);
        }
        mob.special = 0;
        parm.activated = true;
        return true; }
    default: DENG2_ASSERT(!"Unknown intercept type");
        return false;
    }
}
dd_bool P_UsePuzzleItem(player_t *player, int itemType) {
    DENG_ASSERT(player != 0);
    mobj_t *mobj = player->plr->mo;
    if(!mobj) return false; // Huh?
    ptr_puzzleitemtraverse_params_t parm;
    parm.useMobj   = mobj;
    parm.itemType  = itemType;
    parm.activated = false;
    uint an = mobj->angle >> ANGLETOFINESHIFT;
    vec2d_t farUsePoint = { mobj->origin[VX] + FIX2FLT(USERANGE * finecosine[an]),
                            mobj->origin[VY] + FIX2FLT(USERANGE * finesine  [an]) };
    P_PathTraverse(mobj->origin, farUsePoint, PTR_PuzzleItemTraverse, &parm);
    if(!parm.activated) {
        P_SetYellowMessage(player, TXT_USEPUZZLEFAILED);
    }
    return parm.activated;
}
#endif
#ifdef __JHEXEN__
struct countmobjoftypeparams_t {
    mobjtype_t type;
    int count;
};
static int countMobjOfType(thinker_t *th, void *context) {
    countmobjoftypeparams_t *params = (countmobjoftypeparams_t *) context;
    mobj_t *mo = (mobj_t *) th;
    if(mo->type != params->type)
        return false; // Continue iteration.
    if((mo->flags & MF_COUNTKILL) && mo->health <= 0)
        return false; // Continue iteration.
    params->count++;
    return false; // Continue iteration.
}
int P_MobjCount(int type, int tid) {
    if(!type && !tid) return 0;
    mobjtype_t moType = TranslateThingType[type];
    if(tid) {
        int count = 0;
        mobj_t *mo;
        int searcher = -1;
        while((mo = P_FindMobjFromTID(tid, &searcher))) {
            if(type == 0) {
                count++;
            }
            else if(moType == mo->type) {
                if((mo->flags & MF_COUNTKILL) && mo->health <= 0) {
                    continue;
                }
                count++;
            }
        }
        return count;
    }
    countmobjoftypeparams_t params;
    params.type  = moType;
    params.count = 0;
    Thinker_Iterate(P_MobjThinker, countMobjOfType, &params);
    return params.count;
}
#endif // __JHEXEN__
xsector_t *xsectors;
xline_t *xlines;
dd_bool mapSetup;
xline_t *P_ToXLine(Line *line) {
    if(!line) return NULL;
    if(P_IsDummy(line)) {
        return (xline_t *)P_DummyExtraData(line);
    }
    else {
        return &xlines[P_ToIndex(line)];
    }
}
xline_t *P_GetXLine(int idx) {
    if(idx < 0 || idx >= numlines) return 0;
    return &xlines[idx];
}
void P_SetLineAutomapVisibility(int player, int lineIdx, dd_bool visible) {
    Line *line = (Line *)P_ToPtr(DMU_LINE, lineIdx);
    if(!line || P_IsDummy(line)) return;
    xline_t *xline = P_ToXLine(line);
    if(xline->mapped[player] != visible) {
        xline->mapped[player] = visible;
        if(auto *automap = ST_TryFindAutomapWidget(player)) {
            automap->lineAutomapVisibilityChanged(*line);
        }
    }
}
xsector_t *P_ToXSector(Sector *sector) {
    if(!sector) return NULL;
    if(P_IsDummy(sector)) {
        return (xsector_t *)P_DummyExtraData(sector);
    }
    else {
        return &xsectors[P_ToIndex(sector)];
    }
}
xsector_t const *P_ToXSector_const(Sector const *sector) {
    return P_ToXSector(const_cast<Sector *>(sector));
}
xsector_t *P_GetXSector(int index) {
    if(index < 0 || index >= numsectors) return 0;
    return &xsectors[index];
}
#if __JDOOM64__
static void getSurfaceColor(uint idx, float rgba[4]) {
    if(!idx) {
        rgba[0] = rgba[1] = rgba[2] = rgba[3] = 1;
    }
    else {
        rgba[0] = P_GetGMOFloat(MO_LIGHT, idx-1, MO_COLORR);
        rgba[1] = P_GetGMOFloat(MO_LIGHT, idx-1, MO_COLORG);
        rgba[2] = P_GetGMOFloat(MO_LIGHT, idx-1, MO_COLORB);
        rgba[3] = 1;
    }
}
typedef struct applysurfacecolorparams_s {
    Sector *frontSec;
    float topColor[4];
    float bottomColor[4];
} applysurfacecolorparams_t;
int applySurfaceColor(void *obj, void *context) {
#define LDF_NOBLENDTOP          32
#define LDF_NOBLENDBOTTOM       64
#define LDF_BLEND               128
#define LTF_SWAPCOLORS          4
    Line *li = (Line *) obj;
    applysurfacecolorparams_t *params = (applysurfacecolorparams_t *) context;
    byte dFlags = P_GetGMOByte(MO_XLINEDEF, P_ToIndex(li), MO_DRAWFLAGS);
    byte tFlags = P_GetGMOByte(MO_XLINEDEF, P_ToIndex(li), MO_TEXFLAGS);
    if((dFlags & LDF_BLEND) && params->frontSec == P_GetPtrp(li, DMU_FRONT_SECTOR)) {
        if(Side *side = (Side *)P_GetPtrp(li, DMU_FRONT)) {
            float *top    = (tFlags & LTF_SWAPCOLORS)? params->bottomColor : params->topColor;
            float *bottom = (tFlags & LTF_SWAPCOLORS)? params->topColor : params->bottomColor;
            int flags;
            P_SetFloatpv(side, DMU_TOP_COLOR, top);
            P_SetFloatpv(side, DMU_BOTTOM_COLOR, bottom);
            flags = P_GetIntp(side, DMU_FLAGS);
            if(!(dFlags & LDF_NOBLENDTOP))
                flags |= SDF_BLENDTOPTOMID;
            if(!(dFlags & LDF_NOBLENDBOTTOM))
                flags |= SDF_BLENDBOTTOMTOMID;
            P_SetIntp(side, DMU_FLAGS, flags);
        }
    }
    if((dFlags & LDF_BLEND) && params->frontSec == P_GetPtrp(li, DMU_BACK_SECTOR)) {
        if(Side *side = (Side *)P_GetPtrp(li, DMU_BACK)) {
            float *top    = /*(tFlags & LTF_SWAPCOLORS)? params->bottomColor :*/ params->topColor;
            float *bottom = /*(tFlags & LTF_SWAPCOLORS)? params->topColor :*/ params->bottomColor;
            int flags;
            P_SetFloatpv(side, DMU_TOP_COLOR, top);
            P_SetFloatpv(side, DMU_BOTTOM_COLOR, bottom);
            flags = P_GetIntp(side, DMU_FLAGS);
            if(!(dFlags & LDF_NOBLENDTOP))
                flags |= SDF_BLENDTOPTOMID;
            if(!(dFlags & LDF_NOBLENDBOTTOM))
                flags |= SDF_BLENDBOTTOMTOMID;
            P_SetIntp(side, DMU_FLAGS, flags);
        }
    }
    return false; // Continue iteration
}
#endif
static dd_bool checkMapSpotSpawnFlags(mapspot_t const *spot) {
#if __JHEXEN__
    static uint classFlags[] = {
        MSF_FIGHTER,
        MSF_CLERIC,
        MSF_MAGE
    };
#endif
    if(!IS_NETGAME && (spot->flags & MSF_NOTSINGLE))
        return false;
    if(gfw_Rule(deathmatch) && (spot->flags & MSF_NOTDM))
        return false;
    if(IS_NETGAME && !gfw_Rule(deathmatch) && (spot->flags & MSF_NOTCOOP))
        return false;
    if(gfw_Rule(skill) == SM_NOTHINGS)
        return false;
    if(!(spot->skillModes & (1 << gfw_Rule(skill))))
        return false;
#if __JHEXEN__
    if(!IS_NETGAME) {
        if((spot->flags & classFlags[P_ClassForPlayerWhenRespawning(0, false)]) == 0) {
            return false;
        }
    }
    else if(!gfw_Rule(deathmatch)) {
        int spawnMask = MSF_FIGHTER | MSF_CLERIC | MSF_MAGE;
        if((spot->flags & spawnMask) == 0) {
            return false;
        }
    }
#endif
    return true;
}
/**
 * Determines if a client is allowed to spawn a thing of type @a doomEdNum.
 */
static dd_bool P_IsClientAllowedToSpawn(int doomEdNum) {
    switch(doomEdNum) {
    case 11: // Deathmatch
    case 1:
    case 2:
    case 3:
    case 4:
#if __JHEXEN__
    case 9100: // Player starts 5 through 8.
    case 9101:
    case 9102:
    case 9103:
#endif
        return true;
    default:
        return false;
    }
}
/**
 * Should we auto-spawn one or more mobjs from the specified map spot?
 */
static dd_bool checkMapSpotAutoSpawn(mapspot_t const *spot) {
#if __JHERETIC__
    if(spot->doomEdNum >= 1200 && spot->doomEdNum < 1300)
        return false;
#elif __JHEXEN__
    if(spot->doomEdNum >= 1400 && spot->doomEdNum < 1410)
        return false;
#endif
    switch(spot->doomEdNum) {
    case 1:    // Player starts 1 through 4.
    case 2:
    case 3:
    case 4:
    case 11:   // Player start (deathmatch).
#if __JHERETIC__
    case 56:   // Boss spot.
    case 2002: // Mace spot.
#endif
#if __JHEXEN__
    case 3000: // Polyobj origins.
    case 3001:
    case 3002:
    case 9100: // Player starts 5 through 8.
    case 9101:
    case 9102:
    case 9103:
#endif
        return false;
    default: break;
    }
    return checkMapSpotSpawnFlags(spot);
}
static void initXLines() {
    xlines = (xline_t *)Z_Calloc(numlines * sizeof(xline_t), PU_MAP, 0);
    for(int i = 0; i < numlines; ++i) {
        xline_t *xl = &xlines[i];
        xl->flags   = P_GetGMOShort(MO_XLINEDEF, i, MO_FLAGS) & ML_VALID_MASK;
#if __JHEXEN__
        xl->special = P_GetGMOByte(MO_XLINEDEF, i, MO_TYPE);
        xl->arg1    = P_GetGMOByte(MO_XLINEDEF, i, MO_ARG0);
        xl->arg2    = P_GetGMOByte(MO_XLINEDEF, i, MO_ARG1);
        xl->arg3    = P_GetGMOByte(MO_XLINEDEF, i, MO_ARG2);
        xl->arg4    = P_GetGMOByte(MO_XLINEDEF, i, MO_ARG3);
        xl->arg5    = P_GetGMOByte(MO_XLINEDEF, i, MO_ARG4);
#else
# if __JDOOM64__
        xl->special = P_GetGMOByte(MO_XLINEDEF, i, MO_TYPE);
# else
        xl->special = P_GetGMOShort(MO_XLINEDEF, i, MO_TYPE);
# endif
        xl->tag     = P_GetGMOShort(MO_XLINEDEF, i, MO_TAG);
#endif
    }
}
static void initXSectors() {
    xsectors = (xsector_t *)Z_Calloc(numsectors * sizeof(xsector_t), PU_MAP, 0);
    for(int i = 0; i < numsectors; ++i) {
        xsector_t *xsec = &xsectors[i];
        xsec->special = P_GetGMOShort(MO_XSECTOR, i, MO_TYPE);
        xsec->tag     = P_GetGMOShort(MO_XSECTOR, i, MO_TAG);
#if __JDOOM64__
        applysurfacecolorparams_t params;
        float rgba[4];
        Sector *sec = (Sector *)P_ToPtr(DMU_SECTOR, i);
        getSurfaceColor(TOLIGHTIDX(P_GetGMOShort(MO_XSECTOR, i, MO_FLOORCOLOR)), rgba);
        P_SetFloatpv(sec, DMU_FLOOR_COLOR, rgba);
        getSurfaceColor(TOLIGHTIDX(P_GetGMOShort(MO_XSECTOR, i, MO_CEILINGCOLOR)), rgba);
        P_SetFloatpv(sec, DMU_CEILING_COLOR, rgba);
        params.frontSec = sec;
        getSurfaceColor(TOLIGHTIDX(P_GetGMOShort(MO_XSECTOR, i, MO_WALLTOPCOLOR)), params.topColor);
        getSurfaceColor(TOLIGHTIDX(P_GetGMOShort(MO_XSECTOR, i, MO_WALLBOTTOMCOLOR)), params.bottomColor);
        P_Iteratep(sec, DMU_LINE, applySurfaceColor, &params);
#endif
    }
}
static void initMapSpots() {
    numMapSpots = P_CountMapObjs(MO_THING);
    mapSpots = reinterpret_cast<mapspot_t *>(Z_Malloc(numMapSpots * sizeof(mapspot_t), PU_MAP, 0));
    for(int i = 0; i < int(numMapSpots); ++i) {
        mapspot_t *spot = &mapSpots[i];
        spot->origin[VX] = P_GetGMODouble(MO_THING, i, MO_X);
        spot->origin[VY] = P_GetGMODouble(MO_THING, i, MO_Y);
        spot->origin[VZ] = P_GetGMODouble(MO_THING, i, MO_Z);
        spot->doomEdNum  = P_GetGMOInt(MO_THING, i, MO_DOOMEDNUM);
        spot->skillModes = P_GetGMOInt(MO_THING, i, MO_SKILLMODES);
        spot->angle      = P_GetGMOAngle(MO_THING, i, MO_ANGLE);
        spot->flags      = P_GetGMOInt(MO_THING, i, MO_FLAGS);
#if __JHEXEN__
        spot->tid     = P_GetGMOShort(MO_THING, i, MO_ID);
        spot->special = P_GetGMOByte(MO_THING, i, MO_SPECIAL);
        spot->arg1    = P_GetGMOByte(MO_THING, i, MO_ARG0);
        spot->arg2    = P_GetGMOByte(MO_THING, i, MO_ARG1);
        spot->arg3    = P_GetGMOByte(MO_THING, i, MO_ARG2);
        spot->arg4    = P_GetGMOByte(MO_THING, i, MO_ARG3);
        spot->arg5    = P_GetGMOByte(MO_THING, i, MO_ARG4);
#endif
#if __JHERETIC__
        if(spot->doomEdNum >= 1200 && spot->doomEdNum < 1300) {
            P_AddAmbientSfx(spot->doomEdNum - 1200);
            continue;
        }
#elif __JHEXEN__
        if(spot->doomEdNum >= 1400 && spot->doomEdNum < 1410) {
            xsector_t *xsector = P_ToXSector(Sector_AtPoint_FixedPrecision(spot->origin));
            xsector->seqType = seqtype_t(spot->doomEdNum - 1400);
            continue;
        }
#endif
        switch(spot->doomEdNum) {
        case 11: // Player start (deathmatch).
            P_CreatePlayerStart(0, 0, true, i);
            break;
        case 1: // Player starts 1 through 4.
        case 2:
        case 3:
        case 4: {
#if __JHEXEN__
            uint entryPoint = spot->arg1;
#else
            uint entryPoint = 0;
#endif
            P_CreatePlayerStart(spot->doomEdNum, entryPoint, false, i);
            break; }
#if __JHERETIC__
        case 56: // Boss spot.
            P_AddBossSpot(i);
            break;
        case 2002:
            if(gameMode != heretic_shareware)
                P_AddMaceSpot(i);
            break;
#endif
#if __JHEXEN__
        case 3000: // Polyobj origins.
        case 3001:
        case 3002:
            break;
        case 9100: // Player starts 5 through 8.
        case 9101:
        case 9102:
        case 9103:
            P_CreatePlayerStart(5 + spot->doomEdNum - 9100, spot->arg1, false, i);
            break;
#endif
        default: // No special handling.
            break;
        }
    }
    P_DealPlayerStarts(0);
    if(gfw_Rule(deathmatch)) {
        uint numDMStarts = P_GetNumPlayerStarts(true);
        uint playerCount = 0;
        for(int i = 0; i < MAXPLAYERS; ++i) {
            if(players[i].plr->inGame)
                playerCount++;
        }
        if(numDMStarts < playerCount) {
            App_Log(DE2_MAP_WARNING, "Not enough deathmatch spots in map (%i players, %i DM spots)",
                    playerCount, numDMStarts);
        }
    }
}
#if __JHERETIC__
mapspot_t const *P_ChooseRandomMaceSpot() {
    if(!maceSpots || !maceSpotCount)
        return 0;
    /*
     * Pass 1: Determine how many spots qualify given the current game rules.
     */
    uint numQualifyingSpots = 0;
    for(uint i = 0; i < maceSpotCount; ++i) {
        mapspotid_t mapSpotId = maceSpots[i];
        DENG_ASSERT(mapSpots != 0 && mapSpotId < numMapSpots);
        mapspot_t const *mapSpot = &mapSpots[mapSpotId];
        if(checkMapSpotSpawnFlags(mapSpot))
            numQualifyingSpots += 1;
    }
    if(!numQualifyingSpots)
        return 0;
    /*
     * Pass 2: Choose and locate the chosen spot.
     */
    uint chosenQualifyingSpotIdx = M_Random() % numQualifyingSpots;
    uint qualifyingSpotIdx = 0;
    for(uint i = 0; i < maceSpotCount; ++i) {
        mapspotid_t mapSpotId = maceSpots[i];
        mapspot_t const *mapSpot = &mapSpots[mapSpotId];
        if(!checkMapSpotSpawnFlags(mapSpot))
            continue;
        if(qualifyingSpotIdx != chosenQualifyingSpotIdx) {
            qualifyingSpotIdx++;
            continue;
        }
        App_Log(DE2_DEV_MAP_MSG, "P_ChooseRandomMaceSpot: Chosen map spot id:%u.", mapSpotId);
        return mapSpot;
    }
    DENG_ASSERT(false);
    return 0;
}
#endif // __JHERETIC__
static void spawnMapObjects() {
    for(uint i = 0; i < numMapSpots; ++i) {
        mapspot_t const *spot = &mapSpots[i];
        if(!checkMapSpotAutoSpawn(spot))
            continue;
        mobjtype_t type = P_DoomEdNumToMobjType(spot->doomEdNum);
        if(type != MT_NONE) {
            if(IS_CLIENT) {
                if(!(MOBJINFO[type].flags & MF_LOCAL)) {
                    if(!P_IsClientAllowedToSpawn(spot->doomEdNum))
                        continue;
                }
            }
            App_Log(DE2_DEV_MAP_XVERBOSE, "Spawning mobj at (%g, %g, %g) angle:%i ednum:%i flags:%x",
                    spot->origin[VX], spot->origin[VY], spot->origin[VZ], spot->angle,
                    spot->doomEdNum, spot->flags);
            if(mobj_t *mo = P_SpawnMobj(type, spot->origin, spot->angle, spot->flags)) {
                if(mo->tics > 0)
                    mo->tics = 1 + (P_Random() % mo->tics);
#if __JHEXEN__
                mo->tid = spot->tid;
                mo->special = spot->special;
                mo->args[0] = spot->arg1;
                mo->args[1] = spot->arg2;
                mo->args[2] = spot->arg3;
                mo->args[3] = spot->arg4;
                mo->args[4] = spot->arg5;
#endif
#if __JHEXEN__
                if(mo->flags2 & MF2_FLOATBOB)
                    mo->special1 = FLT2FIX(spot->origin[VZ]);
#endif
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
                if(mo->flags & MF_COUNTKILL)
                    totalKills++;
                if(mo->flags & MF_COUNTITEM)
                    totalItems++;
#endif
            }
        }
        else {
            App_Log(DE2_MAP_WARNING, "Unknown DoomEdNum %i at (%g, %g, %g)",
                    spot->doomEdNum, spot->origin[VX], spot->origin[VY], spot->origin[VZ]);
        }
    }
#if __JHERETIC__
    App_Log(DE2_DEV_MAP_VERBOSE, "spawnMapObjects: %i Firemace spot(s)", maceSpotCount);
    if (gfw_MapInfoFlags() & MIF_SPAWN_ALL_FIREMACES) {
        for (uint i = 0; i < maceSpotCount; ++i) {
            const mapspot_t *spot = &mapSpots[maceSpots[i]];
            if (checkMapSpotSpawnFlags(spot)) {
                P_SpawnMobjXYZ(
                    MT_WMACE, spot->origin[VX], spot->origin[VY], 0, spot->angle, MSF_Z_FLOOR);
            }
        }
    }
    else {
        if (!IS_CLIENT && maceSpotCount) {
            if (gfw_Rule(deathmatch) || M_Random() >= 64) {
                if (mapspot_t const *spot = P_ChooseRandomMaceSpot()) {
                    App_Log(DE2_DEV_MAP_VERBOSE,
                            "spawnMapObjects: Spawning Firemace at (%g, %g, %g)",
                            spot->origin[VX],
                            spot->origin[VY],
                            spot->origin[VZ]);
                    P_SpawnMobjXYZ(
                        MT_WMACE, spot->origin[VX], spot->origin[VY], 0, spot->angle, MSF_Z_FLOOR);
                }
            }
        }
    }
#endif
#if __JHEXEN__
    P_CreateTIDList();
#endif
    P_SpawnPlayers();
}
void P_SetupMap(de::Uri const &mapUri) {
    if (IS_DEDICATED) {
        GameRules newRules(gfw_Session()->rules());
        GameRules_Set(newRules, skill,      cfg.common.netSkill);
        GameRules_Set(newRules, deathmatch, cfg.common.netDeathmatch);
        GameRules_Set(newRules, noMonsters, cfg.common.netNoMonsters);
        /*newRules.*/cfg.common.jumpEnabled = cfg.common.netJumping;
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
        GameRules_Set(newRules, respawnMonsters, cfg.netRespawn);
#endif
#if __JHEXEN__
        GameRules_Set(newRules, randomClasses, cfg.netRandomClass);
#endif
        gfw_Session()->applyNewRules(newRules);
    }
    NetSv_SendGameState(GSF_CHANGE_MAP, DDSP_ALL_PLAYERS);
    mapSetup = true;
    ::timerGame = 0;
    if (gfw_Rule(deathmatch)) {
        int parm = CommandLine_Check("-timer");
        if (parm && parm < CommandLine_Count() - 1) {
            ::timerGame = atoi(CommandLine_At(parm + 1)) * 35 * 60;
        }
    }
    P_ResetWorldState();
    if (!P_MapChange(mapUri.compose().toUtf8().constData())) {
        Con_Error("P_SetupMap: Failed changing/loading map \"%s\".\n", mapUri.compose().toUtf8().constData());
        exit(1); // Unreachable.
    }
    Pause_MapStarted();
    mapSetup = false;
}
struct mobjtype_precachedata_t {
    mobjtype_t type;
    int gameModeBits;
};
static void precacheResources() {
    if(!precache || IS_DEDICATED)
        return;
    R_PrecachePSprites();
#if __JDOOM__ || __JHERETIC__
    {
        static const mobjtype_precachedata_t types[] = {
#  if __JDOOM__
            { MT_SKULL,             GM_ANY },
            { MT_BRUISERSHOT,       GM_ANY },
            { MT_TROOPSHOT,         GM_ANY },
            { MT_HEADSHOT,          GM_ANY },
            { MT_ROCKET,            GM_ANY },
            { MT_PLASMA,            GM_ANY ^ GM_DOOM_SHAREWARE },
            { MT_BFG,               GM_ANY ^ GM_DOOM_SHAREWARE },
            { MT_ARACHPLAZ,         GM_DOOM2 },
            { MT_FATSHOT,           GM_DOOM2 },
            { MT_CLIP,              GM_ANY },
            { MT_SHOTGUN,           GM_ANY },
            { MT_CHAINGUN,          GM_ANY },
            { MT_FIRE,              GM_DOOM2 },
            { MT_TRACER,            GM_ANY },
            { MT_SMOKE,             GM_ANY },
            { MT_FATSHOT,           GM_DOOM2 },
            { MT_BLOOD,             GM_ANY },
            { MT_PUFF,              GM_ANY },
            { MT_TFOG,              GM_ANY }, // Teleport FX.
            { MT_EXTRABFG,          GM_ANY },
            { MT_ROCKETPUFF,        GM_ANY },
#  elif __JHERETIC__
            { MT_BLOODYSKULL,       GM_ANY },
            { MT_CHICPLAYER,        GM_ANY },
            { MT_CHICKEN,           GM_ANY },
            { MT_STAFFPUFF,         GM_ANY },
            { MT_STAFFPUFF2,        GM_ANY },
            { MT_BEAKPUFF,          GM_ANY },
            { MT_GAUNTLETPUFF1,     GM_ANY },
            { MT_GAUNTLETPUFF2,     GM_ANY },
            { MT_BLASTERFX1,        GM_ANY },
            { MT_BLASTERSMOKE,      GM_ANY },
            { MT_RIPPER,            GM_ANY },
            { MT_BLASTERPUFF1,      GM_ANY },
            { MT_BLASTERPUFF2,      GM_ANY },
            { MT_MACEFX1,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_MACEFX2,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_MACEFX3,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_MACEFX4,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_HORNRODFX1,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_HORNRODFX2,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_RAINPLR3,          GM_ANY ^ GM_HERETIC_SHAREWARE }, // SP color
            { MT_GOLDWANDFX1,       GM_ANY },
            { MT_GOLDWANDFX2,       GM_ANY },
            { MT_GOLDWANDPUFF1,     GM_ANY },
            { MT_GOLDWANDPUFF2,     GM_ANY },
            { MT_PHOENIXPUFF,       GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_PHOENIXFX2,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_CRBOWFX1,          GM_ANY },
            { MT_CRBOWFX2,          GM_ANY },
            { MT_CRBOWFX3,          GM_ANY },
            { MT_CRBOWFX4,          GM_ANY },
            { MT_EGGFX,             GM_ANY },
            { MT_FIREBOMB,          GM_ANY },
            { MT_MUMMYSOUL,         GM_ANY },
            { MT_MUMMYFX1,          GM_ANY },
            { MT_BEASTBALL,         GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_BURNBALL,          GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_BURNBALLFB,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_PUFFY,             GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SNAKEPRO_A,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SNAKEPRO_B,        GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_HEADFX1,           GM_ANY },
            { MT_HEADFX2,           GM_ANY },
            { MT_HEADFX3,           GM_ANY },
            { MT_WHIRLWIND,         GM_ANY },
            { MT_WIZFX1,            GM_ANY },
            { MT_IMPCHUNK1,         GM_ANY },
            { MT_IMPCHUNK2,         GM_ANY },
            { MT_IMPBALL,           GM_ANY },
            { MT_KNIGHTAXE,         GM_ANY },
            { MT_REDAXE,            GM_ANY },
            { MT_SRCRFX1,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SORCERER2,         GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SOR2FX1,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SOR2FXSPARK,       GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SOR2FX2,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_SOR2TELEFADE,      GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_WIZARD,            GM_ANY ^ GM_HERETIC_SHAREWARE }, // In case D'sparil is on a map with no Disciples
            { MT_MNTRFX1,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_MNTRFX2,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_MNTRFX3,           GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_AMGWNDWIMPY,       GM_ANY },
            { MT_AMCBOWWIMPY,       GM_ANY },
            { MT_AMSKRDWIMPY,       GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_AMPHRDWIMPY,       GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_AMBLSRWIMPY,       GM_ANY },
            { MT_ARTITOMEOFPOWER,   GM_ANY },
            { MT_ARTIEGG,           GM_ANY },
            { MT_ARTISUPERHEAL,     GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_POD,               GM_ANY },
            { MT_PODGOO,            GM_ANY },
            { MT_SPLASH,            GM_ANY },
            { MT_SPLASHBASE,        GM_ANY },
            { MT_LAVASPLASH,        GM_ANY },
            { MT_LAVASMOKE,         GM_ANY },
            { MT_SLUDGECHUNK,       GM_ANY },
            { MT_SLUDGESPLASH,      GM_ANY },
            { MT_VOLCANOBLAST,      GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_VOLCANOTBLAST,     GM_ANY ^ GM_HERETIC_SHAREWARE },
            { MT_TELEGLITTER,       GM_ANY },
            { MT_TELEGLITTER2,      GM_ANY },
            { MT_TFOG,              GM_ANY },
            { MT_BLOOD,             GM_ANY },
            { MT_BLOODSPLATTER,     GM_ANY },
            { MT_FEATHER,           GM_ANY },
#  endif
            { MT_NONE, 0 }
        };
        for(int i = 0; types[i].type != MT_NONE; ++i) {
            if(types[i].gameModeBits & gameModeBits) {
                Rend_CacheForMobjType(types[i].type);
            }
        }
    }
    if(IS_NETGAME) {
#  if __JDOOM__
        Rend_CacheForMobjType(MT_IFOG);
#  else // __JHERETIC__
        Rend_CacheForMobjType(MT_RAINPLR1);
        Rend_CacheForMobjType(MT_RAINPLR2);
        Rend_CacheForMobjType(MT_RAINPLR3);
        Rend_CacheForMobjType(MT_RAINPLR4);
#  endif
    }
#endif
}
void P_FinalizeMapChange(uri_s const *mapUri_) {
    de::Uri const &mapUri = *reinterpret_cast<de::Uri const *>(mapUri_);
#if !__JHEXEN__
    DENG2_UNUSED(mapUri);
#endif
    initXLines();
    initXSectors();
    Thinker_Init();
#if __JHERETIC__
    P_InitAmbientSound();
#endif
#if __JHEXEN__
    P_InitCorpseQueue();
#endif
    if (const String onSetupSrc = G_MapInfoForMapUri(mapUri).gets("onSetup")) {
        Script script(onSetupSrc);
        Process proc;
        proc.run(script);
        proc.execute();
    }
    initMapSpots();
    spawnMapObjects();
    PO_InitForMap();
    HU_UpdatePsprites();
    P_BuildAllTagLists();
#if !__JHEXEN__
    XG_Init();
#endif
    gfw_Session()->acsSystem().loadModuleForMap(mapUri);
    gfw_Session()->acsSystem().worldSystemMapChanged();
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    P_FindSecrets();
#endif
    P_SpawnAllSpecialThinkers();
    P_SpawnAllMaterialOriginScrollers();
    precacheResources();
    if(IS_SERVER) {
        R_SetAllDoomsdayFlags();
        NetSv_SendTotalCounts(DDSP_ALL_PLAYERS);
    }
    /*
     * Do any map finalization including any game-specific stuff.
     */
#if __JDOOM__
    if(!(gameModeBits & (GM_DOOM2_HACX|GM_DOOM_CHEX))) {
        world_Material *mat = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUriCString("Textures:NUKE24"));
        for(int i = 0; i < numlines; ++i) {
            Line *line = (Line *)P_ToPtr(DMU_LINE, i);
            for(int k = 0; k < 2; ++k) {
                Side *side = (Side *)P_GetPtrp(line, k == 0? DMU_FRONT : DMU_BACK);
                if(!side) continue;
                world_Material *bottomMat = (world_Material *)P_GetPtrp(side, DMU_BOTTOM_MATERIAL);
                world_Material *midMat    = (world_Material *)P_GetPtrp(side, DMU_MIDDLE_MATERIAL);
                if(bottomMat == mat && !midMat) {
                    float yoff = P_GetFloatp(side, DMU_BOTTOM_MATERIAL_OFFSET_Y);
                    P_SetFloatp(side, DMU_BOTTOM_MATERIAL_OFFSET_Y, yoff + 1.0f);
                }
            }
        }
    }
#elif __JHEXEN__
    P_InitLightning();
#endif
#if __JHERETIC__ || __JHEXEN__
    P_MoveThingsOutOfWalls();
#endif
#if __JHERETIC__
    P_TurnGizmosAwayFromDoors();
    DD_SetInteger(DD_FIXEDCOLORMAP_ATTENUATE,
                  (gfw_Session()->mapInfo().geti("flags") & MIF_DIM_TORCH) != 0);
#endif
}
void P_ResetWorldState() {
#if __JHEXEN__
    static int firstFragReset = 1;
#endif
    ::wmInfo.nextMap.clear();
#if __JHEXEN__
    ::wmInfo.nextMapEntryPoint = 0;
#endif
#if __JDOOM__ || __JDOOM64__
    ::wmInfo.maxFrags = 0;
    ::wmInfo.parTime = -1;
#endif
#if !__JHEXEN__
    if(!IS_CLIENT) {
        ::totalKills = ::totalItems = ::totalSecret = 0;
    }
#endif
#if __JDOOM__
    delete ::theBossBrain;
    ::theBossBrain = new BossBrain;
#endif
#if __JHEXEN__
    SN_StopAllSequences();
#endif
#if __JHERETIC__
    ::maceSpotCount = 0;
    ::maceSpots     = 0;
    ::bossSpotCount = 0;
    ::bossSpots     = 0;
#endif
    P_PurgeDeferredSpawns();
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = &::players[i];
        ddplayer_t *ddplr = plr->plr;
        ddplr->mo = NULL;
        plr->killCount = plr->secretCount = plr->itemCount = 0;
        plr->update |= PSF_COUNTERS;
        if(ddplr->inGame && plr->playerState == PST_DEAD)
            plr->playerState = PST_REBORN;
#if __JHEXEN__
        if(!IS_NETGAME || (IS_NETGAME != 0 && gfw_Rule(deathmatch) != 0) || firstFragReset == 1) {
            memset(plr->frags, 0, sizeof(plr->frags));
            firstFragReset = 0;
        }
#else
        std::memset(plr->frags, 0, sizeof(plr->frags));
#endif
        G_ResetLookOffset(i);
    }
#if __JDOOM__ || __JDOOM64__
    ::bodyQueueSlot = 0;
#endif
    P_DestroyPlayerStarts();
#if __JHERETIC__ || __JHEXEN__
    P_ClearBodyQueue();
#endif
}
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
void P_FindSecrets() {
    totalSecret = 0;
    for(int i = 0; i < numsectors; ++i) {
        xsector_t *xsec = P_ToXSector((Sector *)P_ToPtr(DMU_SECTOR, i));
        if(xsec->xg) continue;
        if(xsec->special == 9) {
            totalSecret++;
        }
    }
#if __JDOOM64__
    for(int i = 0; i < numlines; ++i) {
        xline_t *xline = P_ToXLine((Line *)P_ToPtr(DMU_LINE, i));
        if(xline->xg) continue;
        if(xline->special == 994) {
            totalSecret++;
        }
    }
#endif
}
#endif
void P_SpawnSectorMaterialOriginScrollers() {
    if(IS_CLIENT) return;
    for(int i = 0; i < numsectors; ++i) {
        Sector *sec     = (Sector *)P_ToPtr(DMU_SECTOR, i);
        xsector_t *xsec = P_ToXSector(sec);
#if !__JHEXEN__
        if(xsec->xg) continue;
#endif
        P_SpawnSectorMaterialOriginScroller(sec, PLN_FLOOR, xsec->special);
    }
}
void P_SpawnSideMaterialOriginScrollers() {
    if(IS_CLIENT) return;
    for(int i = 0; i < numlines; ++i) {
        Line *line     = (Line *)P_ToPtr(DMU_LINE, i);
        xline_t *xline = P_ToXLine(line);
#if !__JHEXEN__
        if(xline->xg) continue;
#endif
        Side *frontSide = (Side *)P_GetPtrp(line, DMU_FRONT);
        P_SpawnSideMaterialOriginScroller(frontSide, xline->special);
    }
}
void P_SpawnAllMaterialOriginScrollers() {
    P_SpawnSideMaterialOriginScrollers();
    P_SpawnSectorMaterialOriginScrollers();
}
iterlist_t *spechit;  ///< For crossed line specials.
struct spreadsoundtoneighbors_params_t {
    Sector *baseSec;
    int soundBlocks;
    mobj_t *soundTarget;
};
static int spreadSoundToNeighbors(void *ptr, void *context) {
    Line *li = (Line *) ptr;
    spreadsoundtoneighbors_params_t &parm = *(spreadsoundtoneighbors_params_t *) context;
    xline_t *xline = P_ToXLine(li);
    DENG2_ASSERT(xline);
    if(!(xline->flags & ML_TWOSIDED)) return false;
    Sector *frontSec = (Sector *) P_GetPtrp(li, DMU_FRONT_SECTOR);
    if(!frontSec) return false;
    Sector *backSec  = (Sector *) P_GetPtrp(li, DMU_BACK_SECTOR);
    if(!backSec) return false;
    LineOpening opening;
    Line_Opening(li, &opening);
    if(opening.range <= 0) return false;
    Sector *other = (frontSec == parm.baseSec? backSec : frontSec);
    if(xline->flags & ML_SOUNDBLOCK) {
        if(!parm.soundBlocks)
            P_RecursiveSound(parm.soundTarget, other, 1);
    }
    else {
        P_RecursiveSound(parm.soundTarget, other, parm.soundBlocks);
    }
    return false; // Continue iteration.
}
void P_RecursiveSound(mobj_t *soundTarget, Sector *sec, int soundBlocks) {
    xsector_t *xsec = P_ToXSector(sec);
    DENG2_ASSERT(xsec);
    if(P_GetIntp(sec, DMU_VALID_COUNT) == VALIDCOUNT) {
        if(xsec->soundTraversed <= soundBlocks + 1)
            return;
    }
    P_SetIntp(sec, DMU_VALID_COUNT, VALIDCOUNT);
    xsec->soundTraversed = soundBlocks + 1;
    xsec->soundTarget    = soundTarget;
    spreadsoundtoneighbors_params_t parm; de::zap(parm);
    parm.baseSec     = sec;
    parm.soundBlocks = soundBlocks;
    parm.soundTarget = soundTarget;
    P_Iteratep(sec, DMU_LINE, spreadSoundToNeighbors, &parm);
}
#if 0
dd_bool P_SectorTagIsBusy(int tag) {
    for(int i = 0; i < numsectors; ++i) {
        xsector_t *xsec = P_ToXSector((Sector *) P_ToPtr(DMU_SECTOR, i));
        if(xsec->tag == tag && xsec->specialData) {
            return true;
        }
    }
    return false;
}
#endif
void P_NotifySectorFinished(int tag) {
#if __JHEXEN__
    gfw_Session()->acsSystem().forAllScripts([&tag] (acs::Script &script) {
        script.sectorFinished(tag);
        return LoopContinue;
    });
#else
    DENG2_UNUSED(tag);
#endif
}
void P_NotifyPolyobjFinished(int tag) {
#if __JHEXEN__
    gfw_Session()->acsSystem().forAllScripts([&tag] (acs::Script &script) {
        script.polyobjFinished(tag);
        return LoopContinue;
    });
#else
    DENG2_UNUSED(tag);
#endif
}
static void stopPlat(plat_t *plat) {
    DENG2_ASSERT(plat);
    P_ToXSector(plat->sector)->specialData = nullptr;
    P_NotifySectorFinished(P_ToXSector(plat->sector)->tag);
    Thinker_Remove(&plat->thinker);
}
void T_PlatRaise(void *platThinkerPtr) {
    plat_t *plat = (plat_t *)platThinkerPtr;
    result_e res;
    switch(plat->state) {
    case PS_UP:
        res = T_MovePlane(plat->sector, plat->speed, plat->high,
                          plat->crush, 0, 1);
#if __JHERETIC__
        if(!(mapTime & 31))
            S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMMOVE);
#endif
#if __JDOOM__ || __JDOOM64__
        if(plat->type == PT_RAISEANDCHANGE ||
           plat->type == PT_RAISETONEARESTANDCHANGE) {
            if(!(mapTime & 7))
                S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMMOVE);
        }
#endif
        if(res == crushed && (!plat->crush)) {
            plat->count = plat->wait;
            plat->state = PS_DOWN;
#if __JHEXEN__
            SN_StartSequenceInSec(plat->sector, SEQ_PLATFORM);
#else
# if __JDOOM64__
            if(plat->type != PT_DOWNWAITUPDOOR) // jd64 added test
# endif
                S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
#endif
        }
        else {
            if(res == pastdest) {
                plat->count = plat->wait;
                plat->state = PS_WAIT;
#if __JHEXEN__
                SN_StopSequenceInSec(plat->sector);
#else
                S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMSTOP);
#endif
                switch(plat->type) {
                case PT_DOWNWAITUPSTAY:
#if __JHEXEN__
                case PT_DOWNBYVALUEWAITUPSTAY:
#else
# if !__JHERETIC__
                case PT_DOWNWAITUPSTAYBLAZE:
                case PT_RAISETONEARESTANDCHANGE:
# endif
# if __JDOOM64__
                case PT_DOWNWAITUPPLUS16STAYBLAZE: // jd64
                case PT_DOWNWAITUPDOOR: // jd64
# endif
                case PT_RAISEANDCHANGE:
#endif
                    stopPlat(plat);
                    break;
                default:
                    break;
                }
            }
        }
        break;
    case PS_DOWN:
        res =
            T_MovePlane(plat->sector, plat->speed, plat->low, false, 0, -1);
        if(res == pastdest) {
            plat->count = plat->wait;
            plat->state = PS_WAIT;
#if __JHEXEN__ || __JDOOM64__
            switch(plat->type) {
# if __JHEXEN__
            case PT_UPBYVALUEWAITDOWNSTAY:
# endif
            case PT_UPWAITDOWNSTAY:
                stopPlat(plat);
                break;
            default:
                break;
            }
#endif
#if __JHEXEN__
            SN_StopSequenceInSec(plat->sector);
#else
            S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMSTOP);
#endif
        }
        else {
#if __JHERETIC__
            if(!(mapTime & 31))
                S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMMOVE);
#endif
        }
        break;
    case PS_WAIT:
        if(!--plat->count) {
            if(FEQUAL(P_GetDoublep(plat->sector, DMU_FLOOR_HEIGHT), plat->low))
                plat->state = PS_UP;
            else
                plat->state = PS_DOWN;
#if __JHEXEN__
            SN_StartSequenceInSec(plat->sector, SEQ_PLATFORM);
#else
            S_PlaneSound((Plane *)P_GetPtrp(plat->sector, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
#endif
        }
        break;
    default:
        break;
    }
}
void plat_t::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteByte(writer, (byte) type);
    Writer_WriteInt32(writer, P_ToIndex(sector));
    Writer_WriteInt32(writer, FLT2FIX(speed));
    Writer_WriteInt16(writer, (int)low);
    Writer_WriteInt16(writer, (int)high);
    Writer_WriteInt32(writer, wait);
    Writer_WriteInt32(writer, count);
    Writer_WriteByte(writer, (byte) state);
    Writer_WriteByte(writer, (byte) oldState);
    Writer_WriteByte(writer, (byte) crush);
    Writer_WriteInt32(writer, tag);
}
int plat_t::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
#if __JHEXEN__
    if(mapVersion >= 4)
#else
    if(mapVersion >= 5)
#endif
    {   // Note: the thinker class byte has already been read.
        /*int ver =*/ Reader_ReadByte(reader); // version byte.
        thinker.function = T_PlatRaise;
#if !__JHEXEN__
        if(mapVersion == 5) {
            if(!Reader_ReadByte(reader))
                Thinker_SetStasis(&thinker, true);
        }
#endif
        type      = plattype_e(Reader_ReadByte(reader));
        sector    = (Sector *)P_ToPtr(DMU_SECTOR, Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        speed     = FIX2FLT(Reader_ReadInt32(reader));
        low       = (float) Reader_ReadInt16(reader);
        high      = (float) Reader_ReadInt16(reader);
        wait      = Reader_ReadInt32(reader);
        count     = Reader_ReadInt32(reader);
        state     = platstate_e(Reader_ReadByte(reader));
        oldState  = platstate_e(Reader_ReadByte(reader));
        crush     = (dd_bool) Reader_ReadByte(reader);
        tag       = Reader_ReadInt32(reader);
    }
    else {
        int32_t junk[4]; // sizeof thinker_t
        Reader_Read(reader, (byte *)junk, 16);
        sector    = (Sector *)P_ToPtr(DMU_SECTOR, (int) Reader_ReadInt32(reader));
        DENG_ASSERT(sector != 0);
        speed     = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        low       = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        high      = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
        wait      = Reader_ReadInt32(reader);
        count     = Reader_ReadInt32(reader);
        state     = platstate_e(Reader_ReadInt32(reader));
        oldState  = platstate_e(Reader_ReadInt32(reader));
        crush     = Reader_ReadInt32(reader);
        tag       = Reader_ReadInt32(reader);
        type      = plattype_e(Reader_ReadInt32(reader));
        thinker.function = T_PlatRaise;
#if !__JHEXEN__
        if(junk[2] == 0) {
            Thinker_SetStasis(&thinker, true);
        }
#endif
    }
    P_ToXSector(sector)->specialData = this;
    return true; // Add this thinker.
}
#if __JHEXEN__
static int doPlat(Line * /*line*/, int tag, byte *args, plattype_e type, int /*amount*/)
#else
static int doPlat(Line *line, int tag, plattype_e type, int amount)
#endif
{
#if !__JHEXEN__
    Sector *frontSector = (Sector *)P_GetPtrp(line, DMU_FRONT_SECTOR);
#endif
    iterlist_t *list = P_GetSectorIterListForTag(tag, false);
    if(!list) return 0;
    int rtn = 0;
    IterList_SetIteratorDirection(list, ITERLIST_FORWARD);
    IterList_RewindIterator(list);
    Sector *sec;
    while((sec = (Sector *)IterList_MoveIterator(list))) {
        xsector_t *xsec = P_ToXSector(sec);
        if(xsec->specialData)
            continue;
        rtn = 1;
        plat_t *plat = (plat_t *) Z_Calloc(sizeof(*plat), PU_MAP, 0);
        plat->thinker.function = T_PlatRaise;
        Thinker_Add(&plat->thinker);
        xsec->specialData = plat;
        plat->type   = type;
        plat->sector = sec;
        plat->crush  = false;
        plat->tag    = tag;
#if __JHEXEN__
        plat->speed  = (float) args[1] * (1.0 / 8);
#endif
        coord_t floorHeight = P_GetDoublep(sec, DMU_FLOOR_HEIGHT);
        switch(type) {
#if !__JHEXEN__
        case PT_RAISETONEARESTANDCHANGE:
            plat->speed = PLATSPEED * .5;
            P_SetPtrp(sec, DMU_FLOOR_MATERIAL,
                      P_GetPtrp(frontSector, DMU_FLOOR_MATERIAL)); {
            coord_t nextFloor;
            if(P_FindSectorSurroundingNextHighestFloor(sec, floorHeight, &nextFloor))
                plat->high = nextFloor;
            else
                plat->high = floorHeight;
            }
            plat->wait = 0;
            plat->state = PS_UP;
            xsec->special = 0;
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMMOVE);
            break;
        case PT_RAISEANDCHANGE:
            plat->speed = PLATSPEED * .5;
            P_SetPtrp(sec, DMU_FLOOR_MATERIAL,
                      P_GetPtrp(frontSector, DMU_FLOOR_MATERIAL));
            plat->high = floorHeight + amount;
            plat->wait = 0;
            plat->state = PS_UP;
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMMOVE);
            break;
#endif
        case PT_DOWNWAITUPSTAY:
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &plat->low);
#if __JHEXEN__
            plat->low += 8;
#else
            plat->speed = PLATSPEED * 4;
#endif
            if(plat->low > floorHeight)
                plat->low = floorHeight;
            plat->high = floorHeight;
            plat->state = PS_DOWN;
#if __JHEXEN__
            plat->wait = (int) args[2];
#else
            plat->wait = PLATWAIT * TICSPERSEC;
#endif
#if !__JHEXEN__
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
#endif
            break;
#if __JDOOM64__ || __JHEXEN__
        case PT_UPWAITDOWNSTAY:
            P_FindSectorSurroundingHighestFloor(sec, -500, &plat->high);
            if(plat->high < floorHeight)
                plat->high = floorHeight;
            plat->low = floorHeight;
            plat->state = PS_UP;
# if __JHEXEN__
            plat->wait = (int) args[2];
# else
            plat->wait = PLATWAIT * TICSPERSEC;
# endif
# if __JDOOM64__
            plat->speed = PLATSPEED * 8;
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
# endif
            break;
#endif
#if __JDOOM64__
        case PT_DOWNWAITUPDOOR: // jd64
            plat->speed = PLATSPEED * 8;
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &plat->low);
            if(plat->low > floorHeight)
                plat->low = floorHeight;
            if(plat->low != floorHeight)
                plat->low += 6;
            plat->high = floorHeight;
            plat->wait = 50 * PLATWAIT;
            plat->state = PS_DOWN;
            break;
#endif
#if __JHEXEN__
       case PT_DOWNBYVALUEWAITUPSTAY:
            plat->low = floorHeight - (coord_t) args[3] * 8;
            if(plat->low > floorHeight)
                plat->low = floorHeight;
            plat->high = floorHeight;
            plat->wait = (int) args[2];
            plat->state = PS_DOWN;
            break;
        case PT_UPBYVALUEWAITDOWNSTAY:
            plat->high = floorHeight + (coord_t) args[3] * 8;
            if(plat->high < floorHeight)
                plat->high = floorHeight;
            plat->low = floorHeight;
            plat->wait = (int) args[2];
            plat->state = PS_UP;
            break;
#endif
#if __JDOOM__ || __JDOOM64__
        case PT_DOWNWAITUPSTAYBLAZE:
            plat->speed = PLATSPEED * 8;
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &plat->low);
            if(plat->low > floorHeight)
                plat->low = floorHeight;
            plat->high = floorHeight;
            plat->wait = PLATWAIT * TICSPERSEC;
            plat->state = PS_DOWN;
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
            break;
#endif
        case PT_PERPETUALRAISE:
            P_FindSectorSurroundingLowestFloor(sec,
                P_GetDoublep(sec, DMU_FLOOR_HEIGHT), &plat->low);
#if __JHEXEN__
            plat->low += 8;
#else
            plat->speed = PLATSPEED;
#endif
            if(plat->low > floorHeight)
                plat->low = floorHeight;
            P_FindSectorSurroundingHighestFloor(sec, -500, &plat->high);
            if(plat->high < floorHeight)
                plat->high = floorHeight;
            plat->state = platstate_e(P_Random() & 1);
#if __JHEXEN__
            plat->wait = (int) args[2];
#else
            plat->wait = PLATWAIT * TICSPERSEC;
#endif
#if !__JHEXEN__
            S_PlaneSound((Plane *)P_GetPtrp(sec, DMU_FLOOR_PLANE), SFX_PLATFORMSTART);
#endif
            break;
        default:
            break;
        }
#if __JHEXEN__
        SN_StartSequenceInSec(plat->sector, SEQ_PLATFORM);
#endif
    }
    return rtn;
}
#if __JHEXEN__
int EV_DoPlat(Line *line, byte *args, plattype_e type, int amount)
#else
int EV_DoPlat(Line *line, plattype_e type, int amount)
#endif
{
#if __JHEXEN__
    return doPlat(line, (int) args[0], args, type, amount);
#else
    int rtn = 0;
    xline_t *xline = P_ToXLine(line);
    switch(type) {
    case PT_PERPETUALRAISE:
        rtn = P_PlatActivate(xline->tag);
        break;
    default:
        break;
    }
    return doPlat(line, xline->tag, type, amount) || rtn;
#endif
}
#if !__JHEXEN__
struct activateplatparams_t {
    short tag;
    int count;
};
static int activatePlat(thinker_t *th, void *context) {
    plat_t *plat = (plat_t *) th;
    activateplatparams_t *params = (activateplatparams_t *) context;
    if(plat->tag == (int) params->tag && Thinker_InStasis(&plat->thinker)) {
        plat->state = plat->oldState;
        Thinker_SetStasis(&plat->thinker, false);
        params->count++;
    }
    return false; // Contiue iteration.
}
int P_PlatActivate(short tag) {
    activateplatparams_t parm;
    parm.tag   = tag;
    parm.count = 0;
    Thinker_Iterate(T_PlatRaise, activatePlat, &parm);
    return parm.count;
}
#endif
struct deactivateplatparams_t {
    short tag;
    int count;
};
static int deactivatePlat(thinker_t *th, void *context) {
    plat_t *plat = (plat_t *) th;
    deactivateplatparams_t *params = (deactivateplatparams_t *) context;
#if __JHEXEN__
    if(plat->tag == params->tag) {
        stopPlat(plat);
        params->count++;
        return true; // Stop iteration.
    }
#else
    if(plat->tag == (int) params->tag && !Thinker_InStasis(&plat->thinker)) {
        plat->oldState = plat->state;
        Thinker_SetStasis(&plat->thinker, true);
        params->count++;
    }
#endif
    return false; // Continue iteration.
}
int P_PlatDeactivate(short tag) {
    deactivateplatparams_t parm;
    parm.tag   = tag;
    parm.count = 0;
    Thinker_Iterate(T_PlatRaise, deactivatePlat, &parm);
    return parm.count;
}
void T_Scroll(scroll_t *s) {
    DENG_ASSERT(s != 0);
    if(IS_ZERO(s->offset[0]) && IS_ZERO(s->offset[1])) return;
    if(DMU_GetType(s->dmuObject) == DMU_SIDE) {
        Side *side = (Side *)s->dmuObject;
        if(s->elementBits & (1 << SS_TOP)) {
            P_TranslateSideMaterialOrigin(side, SS_TOP, s->offset);
        }
        if(s->elementBits & (1 << SS_MIDDLE)) {
            P_TranslateSideMaterialOrigin(side, SS_MIDDLE, s->offset);
        }
        if(s->elementBits & (1 << SS_BOTTOM)) {
            P_TranslateSideMaterialOrigin(side, SS_BOTTOM, s->offset);
        }
    }
    else // Sector plane-surface(s).
    {
        Sector *sector = (Sector *)s->dmuObject;
        if(s->elementBits & (1 << PLN_FLOOR)) {
            Plane *plane = (Plane *)P_GetPtrp(sector, DMU_FLOOR_PLANE);
            P_TranslatePlaneMaterialOrigin(plane, s->offset);
        }
        if(s->elementBits & (1 << PLN_CEILING)) {
            Plane *plane = (Plane *)P_GetPtrp(sector, DMU_CEILING_PLANE);
            P_TranslatePlaneMaterialOrigin(plane, s->offset);
        }
    }
}
void scroll_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteByte(writer, DMU_GetType(dmuObject));
    Writer_WriteInt32(writer, P_ToIndex(dmuObject));
    Writer_WriteInt32(writer, elementBits);
    Writer_WriteInt32(writer, FLT2FIX(offset[0]));
    Writer_WriteInt32(writer, FLT2FIX(offset[1]));
}
int scroll_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
    /*int ver =*/ Reader_ReadByte(reader); // version byte.
    if(Reader_ReadByte(reader) == DMU_SIDE) // Type byte.
    {
        int sideIndex = (int) Reader_ReadInt32(reader);
        if(mapVersion >= 12) {
            dmuObject = (Side *)P_ToPtr(DMU_SIDE, sideIndex);
        }
        else {
            dmuObject = msr->side(sideIndex);
        }
        DENG_ASSERT(dmuObject != 0);
    }
    else // Sector plane-surface.
    {
        dmuObject = (Sector *)P_ToPtr(DMU_SECTOR, (int) Reader_ReadInt32(reader));
        DENG_ASSERT(dmuObject != 0);
    }
    elementBits = Reader_ReadInt32(reader);
    offset[0]   = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
    offset[1]   = FIX2FLT((fixed_t) Reader_ReadInt32(reader));
    thinker.function = (thinkfunc_t) T_Scroll;
    return true; // Add this thinker.
}
static scroll_t *spawnMaterialOriginScroller(void *dmuObject, int elementBits, float offsetXY[2]) {
    if(!dmuObject || elementBits <= 0) return 0;
    if(IS_ZERO(offsetXY[0]) && IS_ZERO(offsetXY[1])) {
        return 0;
    }
    scroll_t *scroll = (scroll_t *)Z_Calloc(sizeof(*scroll), PU_MAP, 0);
    scroll->thinker.function = (thinkfunc_t) T_Scroll;
    Thinker_Add(&scroll->thinker);
    scroll->dmuObject   = dmuObject;
    scroll->elementBits = elementBits;
    scroll->offset[0]   = offsetXY[0];
    scroll->offset[1]   = offsetXY[1];
    return scroll;
}
scroll_t *P_SpawnSideMaterialOriginScroller(Side *side, short special) {
    int elementBits;
    float offset[2];
    if(!side) return 0;
    switch(special) {
    default: return 0; // Not a scroller.
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    case 48:  ///< Tagless, scroll left.
# if __JDOOM64__
    case 150: ///< Tagless, scroll right.
# elif __JDOOM__
    case 85:  ///< Tagless, scroll right (BOOM).
# elif __JHERETIC__
    case 99:  ///< Tagless, scroll right.
# endif
        offset[0] = (special == 48? 1 : -1);
        offset[1] = 0;
        break;
#endif //  __JDOOM__ || __JDOOM64__ || __JHERETIC__
#if __JDOOM64__
    case 2561: ///< Tagless, scroll up.
    case 2562: ///< Tagless, scroll down.
        offset[0] = 0;
        offset[1] = (special == 2561? 1 : -1);
        break;
    case 2080: ///< Tagless, scroll up/left.
    case 2614: ///< Tagless, scroll up/right.
        offset[0] = (special == 2614? 1 : -1);
        offset[1] =  1;
        break;
#endif // __JDOOM64__
#if __JHEXEN__
    case 100:  ///< Tagless, scroll left at speed.
    case 101: {  ///< Tagless, scroll right at speed.
        xline_t *xline = P_ToXLine((Line *)P_GetPtrp(side, DMU_LINE));
        float speed = FIX2FLT(xline->arg1 << 10);
        offset[0] = (special == 100? speed : -speed);
        offset[1] = 0;
        break; }
    case 102: ///< Tagless, scroll up at speed.
    case 103: {  ///< Tagless, scroll down speed.
        xline_t *xline = P_ToXLine((Line *)P_GetPtrp(side, DMU_LINE));
        float speed = FIX2FLT(xline->arg1 << 10);
        offset[0] = 0;
        offset[1] = (special == 102? speed : -speed);
        break; }
#endif // __JHEXEN__
#if __JDOOM__
    case 255: ///< Tagless, scroll by material origin (BOOM).
        P_GetFloatpv(side, DMU_MIDDLE_MATERIAL_OFFSET_XY, offset);
        offset[0] = -offset[0];
        break;
#endif
    }
    elementBits = (1 << SS_MIDDLE) | (1 << SS_BOTTOM) | (1 << SS_TOP);
    return spawnMaterialOriginScroller(side, elementBits, offset);
}
scroll_t *P_SpawnSectorMaterialOriginScroller(Sector *sector, uint planeId, short special) {
#define SCROLLUNIT (8.f/35*2)
    if(!sector || !(planeId == PLN_FLOOR || planeId == PLN_CEILING)) {
        return 0;
    }
#if __JHERETIC__ || __JHEXEN__
    int elementBits;
    float offset[2] = { 0, 0 };
    switch(special) {
    default: return 0; // Not a scroller.
# if __JHERETIC__
    case 25: ///< Scroll north.
    case 26:
    case 27:
    case 28:
    case 29:
# else // __JHEXEN__
    case 201: ///< Scroll north.
    case 202:
    case 203:
# endif
# if __JHERETIC__
        if(!cfg.fixPlaneScrollMaterialsEastOnly) break;
# endif
        offset[0] = 0;
# if __JHERETIC__
        offset[1] = -(SCROLLUNIT * (1 + (special - 25)*2));
# else // __JHEXEN__
        offset[1] = -(SCROLLUNIT * (1 + special - 201));
# endif
        break;
# if __JHERETIC__
    case 20: ///< Scroll east.
    case 21:
    case 22:
    case 23:
    case 24:
# else // __JHEXEN__
    case 204: ///< Scroll east.
    case 205:
    case 206:
# endif
# if __JHERETIC__
        offset[0] = -(SCROLLUNIT * (1 + (special - 20)*2));
# else // __JHEXEN__
        offset[0] = -(SCROLLUNIT * (1 + special - 204));
# endif
        offset[1] = 0;
        break;
# if __JHERETIC__
    case 30: ///< Scroll south.
    case 31:
    case 32:
    case 33:
    case 34:
# else // __JHEXEN__
    case 207: ///< Scroll south.
    case 208:
    case 209:
# endif
# if __JHERETIC__
        if(!cfg.fixPlaneScrollMaterialsEastOnly) break;
#endif
        offset[0] = 0;
# if __JHERETIC__
        offset[1] = SCROLLUNIT * (1 + (special - 30)*2);
# else // __JHEXEN__
        offset[1] = SCROLLUNIT * (1 + special - 207);
# endif
        break;
# if __JHERETIC__
    case 35: ///< Scroll west.
    case 36:
    case 37:
    case 38:
    case 39:
# else // __JHEXEN__
    case 210: ///< Scroll west.
    case 211:
    case 212:
# endif
# if __JHERETIC__
        if(!cfg.fixPlaneScrollMaterialsEastOnly) break;
# endif
# if __JHERETIC__
        offset[0] = SCROLLUNIT * (1 + (special - 35)*2);
# else // __JHEXEN__
        offset[VX] = SCROLLUNIT * (1 + special - 210);
# endif
        offset[1] = 0;
        break;
#if __JHERETIC__
    case 4: ///< Scroll east (lava damage).
        offset[0] = -(SCROLLUNIT * 8 * (1 + special - 4));
        offset[1] = 0;
        break;
#endif
#if __JHEXEN__
    case 213: ///< Scroll northwest.
    case 214:
    case 215:
        offset[0] = SCROLLUNIT * (1 + special - 213);
        offset[1] = -(SCROLLUNIT * (1 + special - 213));
        break;
    case 216: ///< Scroll northeast.
    case 217:
    case 218:
        offset[0] = -(SCROLLUNIT * (1 + special - 216));
        offset[1] = -(SCROLLUNIT * (1 + special - 216));
        break;
    case 219: ///< Scroll southeast.
    case 220:
    case 221:
        offset[0] = -(SCROLLUNIT * (1 + special - 219));
        offset[1] = SCROLLUNIT * (1 + special - 219);
        break;
    case 222: ///< Scroll southwest.
    case 223:
    case 224:
        offset[0] = SCROLLUNIT * (1 + special - 222);
        offset[1] = SCROLLUNIT * (1 + special - 222);
        break;
#endif
    }
    elementBits = 1 << planeId;
    return spawnMaterialOriginScroller(sector, elementBits, offset);
#else // !(__JHERETIC__ || __JHEXEN__)
    DENG_UNUSED(special);
#endif
    return 0;
#undef SCROLLUNIT
}
bool S_MapMusic(de::Uri const &mapUri) {
    String musicId = G_MapInfoForMapUri(mapUri).gets("music");
    if (musicId.isEmpty()) {
        musicId = mapUri.path();
    }
    if (S_StartMusic(musicId.toUtf8(), true)) {
        Con_SetInteger2("map-music", Defs().getMusicNum(musicId.toUtf8()), SVF_WRITE_OVERRIDE);
        return true;
    }
    else {
        S_StopMusic();
        Con_SetInteger2("map-music", -1, SVF_WRITE_OVERRIDE);
    }
    return false;
}
void S_SectorSound(Sector *sec, int id) {
    if(!sec) return;
    S_SectorStopSounds(sec);
    S_StartSound(id, (mobj_t *) P_GetPtrp(sec, DMU_EMITTER));
}
void S_SectorStopSounds(Sector *sec) {
    if(!sec) return;
    S_StopSound2(0, (mobj_t *) P_GetPtrp(sec, DMU_EMITTER), SSF_ALL_SECTOR);
}
void S_PlaneSound(Plane *pln, int id) {
    if(!pln) return;
    S_SectorStopSounds((Sector *) P_GetPtrp(pln, DMU_SECTOR));
    S_StartSound(id, (mobj_t *) P_GetPtrp(pln, DMU_EMITTER));
}
#ifdef __JHEXEN__
int S_GetSoundID(char const *name) {
    return Defs().getSoundNumForName(name);
}
void SndInfoParser(ddstring_s const *path) {
    AutoStr *script = M_ReadFileIntoString(path, 0);
    if(script && !Str_IsEmpty(script)) {
        App_Log(DE2_RES_VERBOSE, "Parsing \"%s\"...", F_PrettyPath(Str_Text(path)));
        HexLex lexer(script, path);
        while(lexer.readToken()) {
            if(!Str_CompareIgnoreCase(lexer.token(), "$archivepath")) // Unused.
            {
                lexer.readString();
                continue;
            }
            if(!Str_CompareIgnoreCase(lexer.token(), "$map")) {
                int const mapNumber        = (int)lexer.readNumber();
                ddstring_t const *lumpName = lexer.readString();
                if(mapNumber > 0) {
                    Record &mapInfo = G_MapInfoForMapUri(G_ComposeMapUri(0, mapNumber - 1));
                    if (const Record *music = Defs().musics.tryFind("id", Str_Text(lumpName))) {
                        mapInfo.set("music", Str_Text(lumpName));
                    }
                    else {
                        if (Record *music = Defs().musics.tryFind("id", mapInfo.gets("music"))) {
                            music->set("lumpName", Str_Text(lumpName));
                        }
                    }
                }
                continue;
            }
            if(!Str_CompareIgnoreCase(lexer.token(), "$registered")) // Unused.
            {
                continue;
            }
            if(Str_At(lexer.token(), 0) == '$') {
                Con_Error("SndInfoParser: Unknown command '%s' in \"%s\" on line #%i",
                          lexer.token(), F_PrettyPath(Str_Text(path)), lexer.lineNumber());
            }
            lexer.unreadToken();
            int const soundIndex       = Defs().getSoundNumForName(Str_Text(lexer.readString()));
            ddstring_t const *lumpName = lexer.readString();
            if(soundIndex) {
                Def_Set(DD_DEF_SOUND, soundIndex,
                        DD_LUMP, Str_At(lumpName, 0) == '?' ? "default" : Str_Text(lumpName));
            }
        }
    }
    char buf[80];
    for(int i = 0; i < Get(DD_NUMSOUNDS); ++i) {
        Def_Get(DD_DEF_SOUND_LUMPNAME, (char *) &i, buf);
        if(!strcmp(buf, "")) {
            Def_Set(DD_DEF_SOUND, i, DD_LUMP, "default");
        }
    }
    if(gameMode == hexen_betademo) {
        int soundId = Defs().getSoundNumForName("AMBIENT12");
        Def_Get(DD_DEF_SOUND_LUMPNAME, (char *) &soundId, buf);
        if(!strcasecmp(buf, "chain")) {
            Def_Set(DD_DEF_SOUND, soundId, DD_LUMP, "default");
        }
    }
}
#endif // __JHEXEN__
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
#  define TELEPORTSOUND     SFX_TELEPT
#  define MAX_START_SPOTS   4 // Maximum number of different player starts.
#else
#  define TELEPORTSOUND     SFX_TELEPORT
#  define MAX_START_SPOTS   8
#endif
#define SPAWNQUEUE_MAX         128
uint numMapSpots;
mapspot_t *mapSpots;
#if __JHERETIC__
uint maceSpotCount;
mapspotid_t *maceSpots;
uint bossSpotCount;
mapspotid_t *bossSpots;
#endif
static int numPlayerStarts = 0;
static playerstart_t *playerStarts;
static int numPlayerDMStarts = 0;
static playerstart_t *deathmatchStarts;
/**
 * New class (or -1) for each player to be applied when the player respawns.
 * Actually applied on serverside, on the client only valid for the local
 * player(s).
 */
static int playerRespawnAsClass[MAXPLAYERS];
static dd_bool fuzzySpawnPosition(coord_t *x, coord_t *y, coord_t *z, angle_t *angle,
    int *spawnFlags) {
#define XOFFSET         (33) // Player radius = 16
#define YOFFSET         (33) // Player radius = 16
    DENG2_UNUSED3(z, angle, spawnFlags);
    DENG_ASSERT(x != 0 && y != 0);
    for(int i = 0; i < 9; ++i) {
        coord_t pos[2] = { *x, *y };
        if(i != 0) {
            int k = (i == 4 ? 0 : i);
            pos[VX] += (k % 3 - 1) * XOFFSET;
            pos[VY] += (k / 3 - 1) * YOFFSET;
        }
        if(P_CheckSpot(pos[VX], pos[VY])) {
            *x = pos[VX];
            *y = pos[VY];
            return true;
        }
    }
    return false;
#undef YOFFSET
#undef XOFFSET
}
void P_ResetPlayerRespawnClasses() {
    for(int i = 0; i < MAXPLAYERS; ++i) {
        playerRespawnAsClass[i] = -1;
    }
}
void P_SetPlayerRespawnClass(int plrNum, playerclass_t pc) {
#ifndef __JHEXEN__
    DENG_ASSERT(pc == PCLASS_PLAYER);
#endif
    playerRespawnAsClass[plrNum] = pc;
}
playerclass_t P_ClassForPlayerWhenRespawning(int plrNum, dd_bool clear) {
#if __JHEXEN__
    playerclass_t pClass = cfg.playerClass[plrNum];
#else
    playerclass_t pClass = PCLASS_PLAYER;
#endif
    if(playerRespawnAsClass[plrNum] != -1) {
        pClass = playerclass_t(playerRespawnAsClass[plrNum]);
        if(clear) {
            playerRespawnAsClass[plrNum] = -1;
        }
    }
    return pClass;
}
mobjtype_t P_DoomEdNumToMobjType(int doomEdNum) {
    for(int i = 0; i < Get(DD_NUMMOBJTYPES); ++i) {
        if(doomEdNum == MOBJINFO[i].doomEdNum)
            return mobjtype_t(i);
    }
    return MT_NONE;
}
void P_Init() {
    P_ResetPlayerRespawnClasses();
    spechit = IterList_New();
#if __JHEXEN__
    X_CreateLUTs();
#endif
#if __JHERETIC__ || __JHEXEN__
    P_InitLava();
#endif
    P_Update();
}
void P_Update() {
#if __JHERETIC__ || __JHEXEN__ || __JDOOM64__
    P_InitInventory();
#endif
    P_InitSwitchList();
    P_InitTerrainTypes();
    ::maxHealth = 100;
    if(ded_value_t const *maxHealth = Defs().getValueById("Player|Max Health")) {
        ::maxHealth = String(maxHealth->text).toInt();
    }
#if __JDOOM__ || __JDOOM64__
    ::healthLimit = 200;
    if(ded_value_t const *healthLimit = Defs().getValueById("Player|Health Limit")) {
        ::healthLimit = String(healthLimit->text).toInt();
    }
    ::godModeHealth = ::maxHealth;
    if(ded_value_t const *godHealth = Defs().getValueById("Player|God Health")) {
        ::godModeHealth = String(godHealth->text).toInt();
    }
    ::armorPoints[0] = 100;
    if(ded_value_t const *armor = Defs().getValueById("Player|Green Armor")) {
        ::armorPoints[0] = String(armor->text).toInt();
    }
    ::armorPoints[1] = 200;
    if(ded_value_t const *armor = Defs().getValueById("Player|Blue Armor")) {
        ::armorPoints[1] = String(armor->text).toInt();
    }
    ::armorPoints[2] = 200;
    if(ded_value_t const *armor = Defs().getValueById("Player|IDFA Armor")) {
        ::armorPoints[2] = String(armor->text).toInt();
    }
    ::armorPoints[3] = 200;
    if(ded_value_t const *armor = Defs().getValueById("Player|IDKFA Armor")) {
        ::armorPoints[3] = String(armor->text).toInt();
    }
    ::armorClass[0] = 1;
    if(ded_value_t const *aclass = Defs().getValueById("Player|Green Armor Class")) {
        ::armorClass[0] = String(aclass->text).toInt();
    }
    ::armorClass[1] = 2;
    if(ded_value_t const *aclass = Defs().getValueById("Player|Blue Armor Class")) {
        ::armorClass[1] = String(aclass->text).toInt();
    }
    ::armorClass[2] = 2;
    if(ded_value_t const *aclass = Defs().getValueById("Player|IDFA Armor Class")) {
        ::armorClass[2] = String(aclass->text).toInt();
    }
    ::armorClass[3] = 2;
    if(ded_value_t const *aclass = Defs().getValueById("Player|IDKFA Armor Class")) {
        ::armorClass[3] = String(aclass->text).toInt();
    }
    ::megaSphereHealth = 200;
    if(ded_value_t const *health = Defs().getValueById("MegaSphere|Give|Health")) {
        ::megaSphereHealth = String(health->text).toInt();
    }
    ::soulSphereHealth = 100;
    if(ded_value_t const *health = Defs().getValueById("SoulSphere|Give|Health")) {
        ::soulSphereHealth = String(health->text).toInt();
    }
    ::soulSphereLimit = 200;
    if(ded_value_t const *healthLimit = Defs().getValueById("SoulSphere|Give|Health Limit")) {
        ::soulSphereLimit = String(healthLimit->text).toInt();
    }
#endif
}
void P_Shutdown() {
    if(spechit) {
        IterList_Delete(spechit);
        spechit = 0;
    }
    P_DestroyPlayerStarts();
    P_DestroyAllTagLists();
    P_ShutdownTerrainTypes();
    P_FreeWeaponSlots();
#if __JDOOM__
    delete theBossBrain; theBossBrain = 0;
#endif
}
void P_CreatePlayerStart(int defaultPlrNum, uint entryPoint, dd_bool deathmatch,
                         mapspotid_t spot) {
    playerstart_t *start;
    if(deathmatch) {
        deathmatchStarts = (playerstart_t *) Z_Realloc(deathmatchStarts, sizeof(playerstart_t) * ++numPlayerDMStarts, PU_MAP);
        start = &deathmatchStarts[numPlayerDMStarts - 1];
        App_Log(DE2_DEV_MAP_VERBOSE, "P_CreatePlayerStart: DM #%i plrNum=%i entryPoint=%i spot=%i",
                numPlayerDMStarts - 1, defaultPlrNum, entryPoint, spot);
    }
    else {
        playerStarts = (playerstart_t *) Z_Realloc(playerStarts, sizeof(playerstart_t) * ++numPlayerStarts, PU_MAP);
        start = &playerStarts[numPlayerStarts - 1];
        App_Log(DE2_DEV_MAP_VERBOSE, "P_CreatePlayerStart: Normal #%i plrNum=%i entryPoint=%i spot=%i",
                numPlayerStarts - 1, defaultPlrNum, entryPoint, spot);
    }
    start->plrNum     = defaultPlrNum;
    start->entryPoint = entryPoint;
    start->spot       = spot;
}
void P_DestroyPlayerStarts() {
    Z_Free(playerStarts); playerStarts = 0;
    numPlayerStarts = 0;
    Z_Free(deathmatchStarts); deathmatchStarts = 0;
    numPlayerDMStarts = 0;
}
playerstart_t const *P_GetPlayerStart(uint entryPoint, int pnum, dd_bool deathmatch) {
    DENG2_UNUSED(entryPoint);
    if((deathmatch && !numPlayerDMStarts) || !numPlayerStarts)
        return 0;
    if(pnum < 0)
        pnum = P_Random() % (deathmatch? numPlayerDMStarts:numPlayerStarts);
    else
        pnum = de::clamp(0, pnum, MAXPLAYERS - 1);
    if(deathmatch) {
        return &deathmatchStarts[pnum];
    }
#if __JHEXEN__
    if(IS_NETWORK_SERVER)
        pnum--;
    playerstart_t const *def = 0;
    for(int i = 0; i < numPlayerStarts; ++i) {
        playerstart_t const *start = &playerStarts[i];
        if(start->entryPoint == gfw_Session()->mapEntryPoint() && start->plrNum - 1 == pnum)
            return start;
        if(!start->entryPoint && start->plrNum - 1 == pnum)
            def = start;
    }
    return def;
#else
    return &playerStarts[players[pnum].startSpot];
#endif
}
uint P_GetNumPlayerStarts(dd_bool deathmatch) {
    return deathmatch? numPlayerDMStarts : numPlayerStarts;
}
void P_DealPlayerStarts(uint entryPoint) {
    if(IS_CLIENT) return;
    if(!numPlayerStarts) {
        App_Log(DE2_MAP_WARNING, "No player starts found, players will spawn as cameras");
        return;
    }
    for(int i = (IS_NETWORK_SERVER? 1 : 0); i < MAXPLAYERS; ++i) {
        player_t *pl = &players[i];
        if(!pl->plr->inGame)
            continue;
        int spotNumber = i % MAX_START_SPOTS;
        if(IS_NETWORK_SERVER) spotNumber--;
        pl->startSpot = -1;
        for(int k = 0; k < numPlayerStarts; ++k) {
            playerstart_t const *start = &playerStarts[k];
            if(spotNumber == start->plrNum - 1 &&
               start->entryPoint == entryPoint) {
                pl->startSpot = k;
                App_Log(DE2_DEV_MAP_XVERBOSE, "PlayerStart %i matches: spot=%i entryPoint=%i",
                        k, spotNumber, entryPoint);
            }
        }
        if(pl->startSpot == -1) {
            pl->startSpot = M_Random() % numPlayerStarts;
        }
    }
    App_Log(DE2_DEV_MAP_MSG, "Player starting spots:");
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *pl = &players[i];
        if(!pl->plr->inGame)
            continue;
        App_Log(DE2_DEV_MAP_MSG, "- pl%i: color %i, spot %i", i, cfg.playerColor[i], pl->startSpot);
    }
}
void P_SpawnPlayer(int plrNum, playerclass_t pClass, coord_t x, coord_t y, coord_t z,
    angle_t angle, int spawnFlags, dd_bool makeCamera, dd_bool pickupItems) {
    plrNum = MINMAX_OF(0, plrNum, MAXPLAYERS - 1);
    if(!players[plrNum].plr->inGame)
        return;
    pClass = playerclass_t(MINMAX_OF(0, pClass, NUM_PLAYER_CLASSES - 1));
    /* $unifiedangles */
    mobj_t *mo = P_SpawnMobjXYZ(PCLASS_INFO(pClass)->mobjType, x, y, z, angle, spawnFlags);
    if (!mo) {
        Con_Error("P_SpawnPlayer: Failed spawning mobj for player %i "
                  "(class:%i) pos:[%g, %g, %g] angle:%i.", plrNum, pClass,
                  x, y, z, angle);
        return;
    }
    App_Log(DE2_DEV_MAP_MSG, "P_SpawnPlayer: Player #%i spawned pos:(%g, %g, %g) angle:%x floorz:%g mobjid:%i",
            plrNum, mo->origin[VX], mo->origin[VY], mo->origin[VZ], mo->angle, mo->floorZ,
            mo->thinker.id);
    player_t *p = &players[plrNum];
    if(p->playerState == PST_REBORN)
        G_PlayerReborn(plrNum);
    p->class_ = pClass;
    if(IS_CLIENT && plrNum != CONSOLEPLAYER) {
        mo->ddFlags = DDMF_DONTDRAW;
    }
    if(p->colorMap > 0 && p->colorMap < NUMPLAYERCOLORS) {
        mo->flags |= p->colorMap << MF_TRANSSHIFT;
    }
    /*
    if(cfg.playerColor[plrNum] > 0)
        mo->flags |= cfg.playerColor[plrNum] << MF_TRANSSHIFT;
    */
    App_Log(DE2_DEV_MAP_VERBOSE, "Player #%i spawning with color translation %i",
            plrNum, (mo->flags & MF_TRANSLATION) >> MF_TRANSSHIFT);
    p->plr->lookDir = 0; /* $unifiedangles */
    p->plr->flags |= DDPF_FIXANGLES | DDPF_FIXORIGIN | DDPF_FIXMOM;
    p->plr->flags &= ~DDPF_UNDEFINED_ORIGIN;
    DENG_ASSERT(mo->angle == angle);
    p->jumpTics = 0;
    p->airCounter = 0;
    mo->player = p;
    mo->dPlayer = p->plr;
    mo->health = p->health;
    p->plr->mo = mo;
    p->playerState = PST_LIVE;
    p->refire = 0;
    p->damageCount = 0;
    p->bonusCount = 0;
#if __JHEXEN__
    p->poisonCount = 0;
    p->overridePalette = 0;
#endif
#if __JHERETIC__ || __JHEXEN__
    p->morphTics = 0;
#endif
#if __JHERETIC__
    p->rain1 = NULL;
    p->rain2 = NULL;
#endif
    p->plr->extraLight = 0;
    p->plr->fixedColorMap = 0;
    if(makeCamera)
        p->plr->flags |= DDPF_CAMERA;
    if(p->plr->flags & DDPF_CAMERA) {
        App_Log(DE2_MAP_MSG, "Player #%i spawned as a camera", plrNum);
        p->plr->mo->origin[VZ] += (coord_t) cfg.common.plrViewHeight;
        p->viewHeight = 0;
    }
    else {
        p->viewHeight = (coord_t) cfg.common.plrViewHeight;
    }
    p->viewHeightDelta = 0;
    p->viewZ = p->plr->mo->origin[VZ] + p->viewHeight;
    p->viewOffset[VX] = p->viewOffset[VY] = p->viewOffset[VZ] = 0;
    if(gfw_Rule(deathmatch)) {
#if __JHEXEN__
        p->keys = 2047;
#else
        for(int i = 0; i < NUM_KEY_TYPES; ++i)
            p->keys[i] = true;
#endif
    }
    p->pendingWeapon = WT_NOCHANGE;
    if(pickupItems) {
        P_CheckPosition(mo, mo->origin);
    }
    if(p->pendingWeapon != WT_NOCHANGE)
        p->readyWeapon = p->pendingWeapon;
    else
        p->pendingWeapon = p->readyWeapon;
    p->brain.changeWeapon = WT_NOCHANGE;
    p->update |= PSF_READY_WEAPON | PSF_PENDING_WEAPON;
    P_SetupPsprites(p);
    if(!BusyMode_Active()) {
        HU_WakeWidgets(p - players);
    }
#if __JHEXEN__
    cfg.playerClass[plrNum] = pClass;
    NetSv_SendPlayerInfo(plrNum, DDSP_ALL_PLAYERS);
    P_ClassForPlayerWhenRespawning(plrNum, true /* now applied; clear change request */);
#endif
    R_UpdateConsoleView(plrNum);
}
static void spawnPlayer(int plrNum, playerclass_t pClass, coord_t x, coord_t y,
    coord_t z, angle_t angle, int spawnFlags, dd_bool makeCamera, dd_bool doTeleSpark,
    dd_bool doTeleFrag) {
#if __JDOOM__ || __JDOOM64__
    dd_bool queueBody = (plrNum >= 0? true : false);
#endif
    dd_bool pickupItems = true;
    /* $voodoodolls */
    if(plrNum < 0) {
        plrNum = -plrNum - 1;
        pickupItems = false;
    }
    plrNum = MINMAX_OF(0, plrNum, MAXPLAYERS-1);
    player_t *plr = &players[plrNum];
#if __JDOOM__ || __JDOOM64__
    if(queueBody)
        G_QueueBody(plr->plr->mo);
#endif
    P_SpawnPlayer(plrNum, pClass, x, y, z, angle, spawnFlags, makeCamera, pickupItems);
    if(doTeleSpark && !makeCamera) {
        uint an = angle >> ANGLETOFINESHIFT;
        x += 20 * FIX2FLT(finecosine[an]);
        y += 20 * FIX2FLT(finesine[an]);
        if(mobj_t *mo = P_SpawnTeleFog(x, y, angle + ANG180)) {
            if(mapTime > 1)
                S_StartSound(TELEPORTSOUND, mo);
        }
    }
    if(!makeCamera && doTeleFrag)
        P_Telefrag(plr->plr->mo);
}
/**
 * Spawns the client's mobj on clientside.
 */
void P_SpawnClient(int plrNum) {
    App_Log(DE2_MAP_VERBOSE,
            "Spawning client player mobj (for player %i; console player is %i)",
            plrNum, CONSOLEPLAYER);
    spawnPlayer(plrNum, P_ClassForPlayerWhenRespawning(plrNum, true),
                -30000, -30000, 0, 0, MSF_Z_FLOOR, false, false, false);
    player_t *p = &players[plrNum];
    p->viewHeight = cfg.common.plrViewHeight;
    p->viewHeightDelta = 0;
    p->plr->flags |= DDPF_UNDEFINED_ORIGIN;
    p->plr->flags |= DDPF_UNDEFINED_WEAPON;
    p->plr->flags &= ~DDPF_USE_VIEW_FILTER;
    p->pSprites[0].pos[VY] = WEAPONBOTTOM;
}
#if __JHEXEN__
static String const &ammoTypeName(int ammoType) {
    static String const names[NUM_AMMO_TYPES] = {
        /*AT_BLUEMANA*/  "Blue mana",
        /*AT_GREENMANA*/ "Green mana"
    };
    if(ammoType >= AT_FIRST && ammoType < NUM_AMMO_TYPES)
        return names[ammoType - AT_FIRST];
    throw Error("ammoTypeName", "Unknown ammo type " + String::number(ammoType));
}
#endif  // __JHEXEN__
void P_RebornPlayerInMultiplayer(int plrNum) {
    if(plrNum < 0 || plrNum >= MAXPLAYERS)
        return; // Wha?
    playerclass_t pClass = P_ClassForPlayerWhenRespawning(plrNum, false);
    player_t *p = &players[plrNum];
    App_Log(DE2_DEV_MAP_MSG, "P_RebornPlayer: player %i (class %i)", plrNum, pClass);
    if(p->plr->mo) {
        p->plr->mo->player = 0;
        p->plr->mo->dPlayer = 0;
    }
    if(G_GameState() != GS_MAP) {
        App_Log(DE2_DEV_MAP_ERROR,
                "P_RebornPlayer: Game state is %i, won't spawn", G_GameState());
        return; // Nothing else to do.
    }
    if(gfw_Rule(deathmatch)) {
        G_DeathMatchSpawnPlayer(plrNum);
        return;
    }
#if __JHEXEN__
    int oldKeys = 0, oldPieces = 0;
    dd_bool oldWeaponOwned[NUM_WEAPON_TYPES];
    if(!IS_CLIENT) {
        oldKeys = p->keys;
        oldPieces = p->pieces;
        for(int i = 0; i < NUM_WEAPON_TYPES; ++i)
            oldWeaponOwned[i] = p->weapons[i].owned;
    }
    else {
        for (dd_bool &w : oldWeaponOwned) w = false;
    }
#endif
    if(IS_CLIENT) {
        P_SpawnClient(plrNum);
        return;
    }
    /*
     * Determine spawn parameters.
     */
    coord_t pos[3] = { 0, 0, 0 };
    angle_t angle = 0;
    int spawnFlags = 0;
    dd_bool makeCamera = false;
    uint entryPoint = gfw_Session()->mapEntryPoint();
    dd_bool foundSpot = false;
    playerstart_t const *assigned = P_GetPlayerStart(entryPoint, plrNum, false);
    if(assigned) {
        mapspot_t const *spot = &mapSpots[assigned->spot];
        if(P_CheckSpot(spot->origin[VX], spot->origin[VY])) {
            App_Log(DE2_DEV_MAP_MSG, "- spawning at assigned spot");
            pos[VX]    = spot->origin[VX];
            pos[VY]    = spot->origin[VY];
            pos[VZ]    = spot->origin[VZ];
            angle      = spot->angle;
            spawnFlags = spot->flags;
            foundSpot = true;
        }
    }
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    if(!foundSpot) {
        App_Log(DE2_DEV_MAP_MSG, "- force spawning at %i", p->startSpot);
        if(assigned) {
            mapspot_t const *spot = &mapSpots[assigned->spot];
            pos[VX]    = spot->origin[VX];
            pos[VY]    = spot->origin[VY];
            pos[VZ]    = spot->origin[VZ];
            angle      = spot->angle;
            spawnFlags = spot->flags;
            makeCamera = !fuzzySpawnPosition(&pos[VX], &pos[VY], &pos[VZ],
                                             &angle, &spawnFlags);
        }
        else {
            pos[VX]    = pos[VY] = pos[VZ] = 0;
            angle      = 0;
            spawnFlags = MSF_Z_FLOOR;
            makeCamera = true;
        }
    }
#else
    if(!foundSpot) {
        App_Log(DE2_DEV_MAP_MSG, "P_RebornPlayer: Trying other spots for %i", plrNum);
        for(int i = 0; i < MAXPLAYERS; ++i) {
            if(playerstart_t const *start = P_GetPlayerStart(entryPoint, i, false)) {
                mapspot_t const *spot = &mapSpots[start->spot];
                if(P_CheckSpot(spot->origin[VX], spot->origin[VY])) {
                    pos[VX]    = spot->origin[VX];
                    pos[VY]    = spot->origin[VY];
                    pos[VZ]    = spot->origin[VZ];
                    angle      = spot->angle;
                    spawnFlags = spot->flags;
                    foundSpot = true;
                    App_Log(DE2_DEV_MAP_MSG, "P_RebornPlayer: Spot (%g, %g) selected",
                            spot->origin[VX], spot->origin[VY]);
                    break;
                }
                App_Log(DE2_DEV_MAP_VERBOSE, "P_RebornPlayer: Spot (%g, %g) is unavailable",
                        spot->origin[VX], spot->origin[VY]);
            }
        }
    }
    if(!foundSpot) {
        if(playerstart_t const *start = P_GetPlayerStart(entryPoint, plrNum, false)) {
            mapspot_t const *spot = &mapSpots[start->spot];
            pos[VX]    = spot->origin[VX];
            pos[VY]    = spot->origin[VY];
            pos[VZ]    = spot->origin[VZ];
            angle      = spot->angle;
            spawnFlags = spot->flags;
        }
        else {
            pos[VX]    = pos[VY] = pos[VZ] = 0;
            angle      = 0;
            spawnFlags = MSF_Z_FLOOR;
            makeCamera = true;
        }
    }
#endif
    App_Log(DE2_DEV_MAP_NOTE, "Multiplayer-spawning player at (%f,%f,%f) angle:%x",
            pos[VX], pos[VY], pos[VZ], angle);
    spawnPlayer(plrNum, pClass, pos[VX], pos[VY], pos[VZ], angle,
                spawnFlags, makeCamera, true, true);
    DENG_ASSERT(!IS_CLIENT);
#if __JHEXEN__
    p->keys = oldKeys;
    p->pieces = oldPieces;
    int bestWeapon = 0;
    for(int i = 0; i < NUM_WEAPON_TYPES; ++i) {
        if(oldWeaponOwned[i]) {
            bestWeapon = i;
            p->weapons[i].owned = true;
        }
    }
    for(auto i = int( AT_FIRST ); i < NUM_AMMO_TYPES; ++i) {
        if(ded_value_t const *ammo = Defs().getValueById("Multiplayer|Reborn|" + ammoTypeName(i))) {
            p->ammo[i].owned = String(ammo->text).toInt();
        }
    }
    App_Log(DE2_MAP_VERBOSE, "Player %i reborn in multiplayer: giving mana (b:%i g:%i); "
            "also old weapons, with best weapon %i", plrNum, p->ammo[AT_BLUEMANA].owned,
            p->ammo[AT_GREENMANA].owned, bestWeapon);
    if(bestWeapon) {
        p->readyWeapon = p->pendingWeapon = weapontype_t(bestWeapon);
    }
#endif
}
dd_bool P_CheckSpot(coord_t x, coord_t y) {
#if __JHEXEN__
#define DUMMY_TYPE      MT_PLAYER_FIGHTER
#else
#define DUMMY_TYPE      MT_PLAYER
#endif
    coord_t const pos[3] = { x, y, 0 };
    mobj_t *dummy = P_SpawnMobj(DUMMY_TYPE, pos, 0, MSF_Z_FLOOR);
    DENG_ASSERT(dummy);
    if (!dummy) return false; //Con_Error("P_CheckSpot: Failed creating dummy mobj.");
    dummy->flags &= ~MF_PICKUP;
    dd_bool result = P_CheckPosition(dummy, pos);
    P_MobjRemove(dummy, true);
    return result;
#undef DUMMY_TYPE
}
#if __JHERETIC__
void P_AddMaceSpot(mapspotid_t id) {
    App_Log(DE2_DEV_MAP_VERBOSE, "P_AddMaceSpot: Added mace spot %u", id);
    maceSpots = (mapspotid_t *)Z_Realloc(maceSpots, sizeof(mapspotid_t) * ++maceSpotCount, PU_MAP);
    maceSpots[maceSpotCount-1] = id;
}
void P_AddBossSpot(mapspotid_t id) {
    bossSpots = (mapspotid_t *)Z_Realloc(bossSpots, sizeof(mapspotid_t) * ++bossSpotCount, PU_MAP);
    bossSpots[bossSpotCount-1] = id;
}
#endif
void P_SpawnPlayers() {
    if(IS_CLIENT) {
        for(int i = 0; i < MAXPLAYERS; ++i) {
            if(!players[i].plr->inGame)
                continue;
            P_SpawnClient(i);
        }
        return;
    }
    if(gfw_Rule(deathmatch)) {
        for(int i = 0; i < MAXPLAYERS; ++i) {
            if(!players[i].plr->inGame)
                continue;
            players[i].plr->mo = 0;
            G_DeathMatchSpawnPlayer(i);
        }
    }
    else {
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
        if(!IS_NETGAME) {
            /* $voodoodolls */
            for(int i = 0; i < numPlayerStarts; ++i) {
                if(players[0].startSpot != i && playerStarts[i].plrNum == 1) {
                    mapspot_t const *spot = &mapSpots[playerStarts[i].spot];
                    spawnPlayer(-1, PCLASS_PLAYER, spot->origin[VX],
                                spot->origin[VY], spot->origin[VZ],
                                spot->angle, spot->flags, false,
                                false, false);
                }
            }
        }
#endif
        for(int i = 0; i < MAXPLAYERS; ++i) {
            player_t *plr = &players[i];
            if(!plr->plr->inGame)
                continue;
            playerstart_t const *start =
                plr->startSpot < numPlayerStarts? &playerStarts[plr->startSpot] : 0;
            coord_t pos[3];
            angle_t angle;
            int spawnFlags;
            dd_bool spawnAsCamera;
            if(start) {
                mapspot_t const *spot = &mapSpots[start->spot];
                pos[VX]    = spot->origin[VX];
                pos[VY]    = spot->origin[VY];
                pos[VZ]    = spot->origin[VZ];
                angle      = spot->angle;
                spawnFlags = spot->flags;
                spawnAsCamera = !fuzzySpawnPosition(&pos[VX], &pos[VY], &pos[VZ],
                                                    &angle, &spawnFlags);
            }
            else {
                pos[VX]    = pos[VY] = pos[VZ] = 0;
                angle      = 0;
                spawnFlags = MSF_Z_FLOOR;
                spawnAsCamera = true;
            }
            spawnPlayer(i, P_ClassForPlayerWhenRespawning(i, false),
                        pos[VX], pos[VY], pos[VZ], angle, spawnFlags,
                        spawnAsCamera, false, true);
            App_Log(DE2_DEV_MAP_MSG, "Player %i spawned at (%g, %g, %g)", i, pos[VX], pos[VY], pos[VZ]);
        }
    }
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *plr = &players[i];
        mobj_t *plMo = plr->plr->mo;
        if(!plr->plr->inGame || !plMo)
            continue;
        NetSv_SendPlayerSpawnPosition(i, plMo->origin[VX], plMo->origin[VY], plMo->origin[VZ],
                                      plMo->angle);
    }
}
void G_DeathMatchSpawnPlayer(int playerNum) {
    playerNum = MINMAX_OF(0, playerNum, MAXPLAYERS-1);
    playerclass_t pClass;
#if __JHEXEN__
    if (gfw_Rule(randomClasses)) {
        pClass = playerclass_t(P_Random() % 3);
        if(pClass == cfg.playerClass[playerNum]) // Not the same class, please.
            pClass = playerclass_t((pClass + 1) % 3);
    }
    else
#endif
    {
        pClass = P_ClassForPlayerWhenRespawning(playerNum, false);
    }
    if (IS_CLIENT) {
        if (G_GameState() == GS_MAP) {
            spawnPlayer(playerNum, pClass, -30000, -30000, 0, 0, MSF_Z_FLOOR, false,
                        false, false);
        }
        return;
    }
    if (numPlayerDMStarts < 2)
        Con_Error("G_DeathMatchSpawnPlayer: Error, minimum of two "
                  "(deathmatch) mapspots required for deathmatch.");
#define NUM_TRIES 20
    for(int i = 0; i < NUM_TRIES; ++i) {
        mapspot_t const *spot = &mapSpots[deathmatchStarts[P_Random() % numPlayerDMStarts].spot];
        if(P_CheckSpot(spot->origin[VX], spot->origin[VY]) || i == NUM_TRIES-1) {
            spawnPlayer(playerNum, pClass, spot->origin[VX], spot->origin[VY],
                        spot->origin[VZ], spot->angle, spot->flags, false,
                        true, true);
            return;
        }
    }
}
#if defined(__JHERETIC__) || defined(__JHEXEN__)
struct unstuckmobjinline_params_t {
    coord_t pos[2];
    coord_t minDist;
};
static int unstuckMobjInLine(Line *li, void *context) {
    unstuckmobjinline_params_t &parm = *static_cast<unstuckmobjinline_params_t *>(context);
    if(!P_GetPtrp(li, DMU_BACK_SECTOR)) {
        /*
         * Project the point (mo position) onto this line. If the resultant point
         * lies on the line and the current position is in range of that point,
         * adjust the position moving it away from the projected point.
         */
        coord_t lineOrigin[2]; P_GetDoublepv(P_GetPtrp(li, DMU_VERTEX0), DMU_XY, lineOrigin);
        coord_t lineDirection[2]; P_GetDoublepv(li, DMU_DXY, lineDirection);
        coord_t result[2];
        coord_t pos = V2d_ProjectOnLine(result, parm.pos, lineOrigin, lineDirection);
        if(pos > 0 && pos < 1) {
            coord_t dist = M_ApproxDistance(parm.pos[VX] - result[VX],
                                            parm.pos[VY] - result[VY]);
            if(dist >= 0 && dist < parm.minDist) {
                coord_t len = M_ApproxDistance(lineDirection[0], lineDirection[1]);
                coord_t unit[2], normal[2];
                if(len) {
                    unit[VX] = lineDirection[0] / len;
                    unit[VY] = lineDirection[1] / len;
                }
                else {
                    unit[VX] = unit[VY] = 0;
                }
                normal[VX] =  unit[VY];
                normal[VY] = -unit[VX];
                parm.pos[VX] += normal[VX] * parm.minDist;
                parm.pos[VY] += normal[VY] * parm.minDist;
            }
        }
    }
    return false; // Continue iteration.
}
struct pit_findnearestfacingline_params_t {
    mobj_t *mo;
    coord_t dist;
    Line *line;
};
static int PIT_FindNearestFacingLine(Line *line, void *context) {
    pit_findnearestfacingline_params_t &parm = *static_cast<pit_findnearestfacingline_params_t *>(context);
    coord_t off;
    coord_t dist = Line_PointDistance(line, parm.mo->origin, &off);
    if(off < 0 || off > P_GetDoublep(line, DMU_LENGTH) || dist < 0) {
        return false;
    }
    if(!parm.line || dist < parm.dist) {
        parm.line = line;
        parm.dist = dist;
    }
    return false;
}
static int turnMobjToNearestLine(thinker_t *th, void *context) {
    mobj_t *mo = reinterpret_cast<mobj_t *>(th);
    mobjtype_t type = *static_cast<mobjtype_t *>(context);
    if(mo->type != type) return false;
    App_Log(DE2_MAP_XVERBOSE, "Checking mo %i for auto-turning...", mo->thinker.id);
    AABoxd aaBox(mo->origin[VX] - 50, mo->origin[VY] - 50,
                 mo->origin[VX] + 50, mo->origin[VY] + 50);
    pit_findnearestfacingline_params_t parm;
    parm.mo   = mo;
    parm.dist = 0;
    parm.line = 0;
    VALIDCOUNT++;
    Line_BoxIterator(&aaBox, LIF_SECTOR, PIT_FindNearestFacingLine, &parm);
    if(parm.line) {
        mo->angle = P_GetAnglep(parm.line, DMU_ANGLE) - ANGLE_90;
        App_Log(DE2_MAP_XVERBOSE, "Turning mobj to nearest line: mo=%i angle=%x",  mo->thinker.id, mo->angle);
    }
    else {
        App_Log(DE2_DEV_MAP_XVERBOSE, "Turning mobj to nearest line: mo=%i => no nearest line found",
                mo->thinker.id);
    }
    return false;
}
static int moveMobjOutOfNearbyLines(thinker_t *th, void *context) {
    mobj_t *mo = reinterpret_cast<mobj_t *>(th);
    mobjtype_t type = *static_cast<mobjtype_t *>(context);
    if(mo->type != type) return false;
    AABoxd aaBox(mo->origin[VX] - mo->radius, mo->origin[VY] - mo->radius,
                 mo->origin[VX] + mo->radius, mo->origin[VY] + mo->radius);
    unstuckmobjinline_params_t parm;
    parm.pos[VX] = mo->origin[VX];
    parm.pos[VY] = mo->origin[VY];
    parm.minDist = mo->radius / 2;
    VALIDCOUNT++;
    Line_BoxIterator(&aaBox, LIF_SECTOR, unstuckMobjInLine, &parm);
    if(!FEQUAL(mo->origin[VX], parm.pos[VX]) || !FEQUAL(mo->origin[VY], parm.pos[VY])) {
        P_MobjUnlink(mo);
        mo->origin[VX] = parm.pos[VX];
        mo->origin[VY] = parm.pos[VY];
        P_MobjLink(mo);
    }
    return false;
}
void P_MoveThingsOutOfWalls() {
    static mobjtype_t const types[] = {
#if __JHERETIC__
        MT_MISC10,
#elif __JHEXEN__
        MT_ZWALLTORCH,
        MT_ZWALLTORCH_UNLIT,
#endif
        NUMMOBJTYPES // terminate.
    };
    for(uint i = 0; types[i] != NUMMOBJTYPES; ++i) {
        mobjtype_t type = types[i];
        Thinker_Iterate(P_MobjThinker, moveMobjOutOfNearbyLines, &type);
        Thinker_Iterate(P_MobjThinker, turnMobjToNearestLine, &type);
    }
}
#endif
#ifdef __JHERETIC__
void P_TurnGizmosAwayFromDoors() {
#define MAXLIST 200
    mobj_t *tlist[MAXLIST];
    for(int i = 0; i < numsectors; ++i) {
        Sector *sec = (Sector *)P_ToPtr(DMU_SECTOR, i);
        std::memset(tlist, 0, sizeof(tlist));
        mobj_t *iter = (mobj_t *)P_GetPtrp(sec, DMT_MOBJS);
        for(int k = 0; k < MAXLIST - 1 && iter; iter = iter->sNext) {
            if(iter->type == MT_KEYGIZMOBLUE ||
               iter->type == MT_KEYGIZMOGREEN ||
               iter->type == MT_KEYGIZMOYELLOW)
                tlist[k++] = iter;
        }
        for(int t = 0; (iter = tlist[t]) != NULL; ++t) {
            Line *closestline = 0;
            coord_t closestdist = 0;
            for(int l = 0; l < numlines; ++l) {
                Line *li = (Line *)P_ToPtr(DMU_LINE, l);
                if(!P_GetPtrp(li, DMU_BACK_SECTOR))
                    continue;
                xline_t *xli = P_ToXLine(li);
                if((xli->special != 32 && xli->special != 33 &&
                    xli->special != 34 && xli->special != 26 &&
                    xli->special != 27 && xli->special != 28))
                    continue;
                coord_t d1[2]; P_GetDoublepv(li, DMU_DXY, d1);
                coord_t off;
                coord_t dist = fabs(Line_PointDistance(li, iter->origin, &off));
                if(!closestline || dist < closestdist) {
                    closestdist = dist;
                    closestline = li;
                }
            }
            if(closestline) {
                Vertex *v0 = (Vertex *)P_GetPtrp(closestline, DMU_VERTEX0);
                Vertex *v1 = (Vertex *)P_GetPtrp(closestline, DMU_VERTEX1);
                coord_t v0p[2]; P_GetDoublepv(v0, DMU_XY, v0p);
                coord_t v1p[2]; P_GetDoublepv(v1, DMU_XY, v1p);
                iter->angle = M_PointToAngle2(v0p, v1p) - ANG90;
            }
        }
    }
}
#endif
typedef struct {
    /* Do NOT change these members in any way! */
    char name1[9];
    char name2[9];
#if __JHEXEN__
    int soundID;
#else
    short episode;
#endif
} switchlist_t;
#pragma pack()
#if __JHEXEN__
switchlist_t switchInfo[] = {
    {"SW_1_UP", "SW_1_DN", SFX_SWITCH1},
    {"SW_2_UP", "SW_2_DN", SFX_SWITCH1},
    {"VALVE1", "VALVE2", SFX_VALVE_TURN},
    {"SW51_OFF", "SW51_ON", SFX_SWITCH2},
    {"SW52_OFF", "SW52_ON", SFX_SWITCH2},
    {"SW53_UP", "SW53_DN", SFX_ROPE_PULL},
    {"PUZZLE5", "PUZZLE9", SFX_SWITCH1},
    {"PUZZLE6", "PUZZLE10", SFX_SWITCH1},
    {"PUZZLE7", "PUZZLE11", SFX_SWITCH1},
    {"PUZZLE8", "PUZZLE12", SFX_SWITCH1},
    {"\0", "\0", 0}
};
#else
switchlist_t switchInfo[] = {
# if __JHERETIC__
    {"SW1OFF", "SW1ON", MACRO_SHORT(1)},
    {"SW2OFF", "SW2ON", MACRO_SHORT(1)},
# elif __JDOOM64__
    {"SWXCA", "SWXCB", MACRO_SHORT(1)},
    {"SWXCKA", "SWXCKB", MACRO_SHORT(1)},
    {"SWXCKLA", "SWXCKLB", MACRO_SHORT(1)},
    {"SWXCLA", "SWXCLB", MACRO_SHORT(1)},
    {"SWXHCA", "SWXHCB", MACRO_SHORT(1)},
    {"SWXSAA", "SWXSAB", MACRO_SHORT(1)},
    {"SWXSCA", "SWXSCB", MACRO_SHORT(1)},
    {"SWXSDA", "SWXSDB", MACRO_SHORT(1)},
    {"SWXSEA", "SWXSEB", MACRO_SHORT(1)},
    {"SWXSFA", "SWXSFB", MACRO_SHORT(1)},
    {"SWXSGA", "SWXSGB", MACRO_SHORT(1)},
# elif __JDOOM__
    {"SW1BRCOM", "SW2BRCOM", MACRO_SHORT(1)},
    {"SW1BRN1", "SW2BRN1", MACRO_SHORT(1)},
    {"SW1BRN2", "SW2BRN2", MACRO_SHORT(1)},
    {"SW1BRNGN", "SW2BRNGN", MACRO_SHORT(1)},
    {"SW1BROWN", "SW2BROWN", MACRO_SHORT(1)},
    {"SW1COMM", "SW2COMM", MACRO_SHORT(1)},
    {"SW1COMP", "SW2COMP", MACRO_SHORT(1)},
    {"SW1DIRT", "SW2DIRT", MACRO_SHORT(1)},
    {"SW1EXIT", "SW2EXIT", MACRO_SHORT(1)},
    {"SW1GRAY", "SW2GRAY", MACRO_SHORT(1)},
    {"SW1GRAY1", "SW2GRAY1", MACRO_SHORT(1)},
    {"SW1METAL", "SW2METAL", MACRO_SHORT(1)},
    {"SW1PIPE", "SW2PIPE", MACRO_SHORT(1)},
    {"SW1SLAD", "SW2SLAD", MACRO_SHORT(1)},
    {"SW1STARG", "SW2STARG", MACRO_SHORT(1)},
    {"SW1STON1", "SW2STON1", MACRO_SHORT(1)},
    {"SW1STON2", "SW2STON2", MACRO_SHORT(1)},
    {"SW1STONE", "SW2STONE", MACRO_SHORT(1)},
    {"SW1STRTN", "SW2STRTN", MACRO_SHORT(1)},
    {"SW1BLUE", "SW2BLUE", MACRO_SHORT(2)},
    {"SW1CMT", "SW2CMT", MACRO_SHORT(2)},
    {"SW1GARG", "SW2GARG", MACRO_SHORT(2)},
    {"SW1GSTON", "SW2GSTON", MACRO_SHORT(2)},
    {"SW1HOT", "SW2HOT", MACRO_SHORT(2)},
    {"SW1LION", "SW2LION", MACRO_SHORT(2)},
    {"SW1SATYR", "SW2SATYR", MACRO_SHORT(2)},
    {"SW1SKIN", "SW2SKIN", MACRO_SHORT(2)},
    {"SW1VINE", "SW2VINE", MACRO_SHORT(2)},
    {"SW1WOOD", "SW2WOOD", MACRO_SHORT(2)},
    {"SW1PANEL", "SW2PANEL", MACRO_SHORT(3)},
    {"SW1ROCK", "SW2ROCK", MACRO_SHORT(3)},
    {"SW1MET2", "SW2MET2", MACRO_SHORT(3)},
    {"SW1WDMET", "SW2WDMET", MACRO_SHORT(3)},
    {"SW1BRIK", "SW2BRIK", MACRO_SHORT(3)},
    {"SW1MOD1", "SW2MOD1", MACRO_SHORT(3)},
    {"SW1ZIM", "SW2ZIM", MACRO_SHORT(3)},
    {"SW1STON6", "SW2STON6", MACRO_SHORT(3)},
    {"SW1TEK", "SW2TEK", MACRO_SHORT(3)},
    {"SW1MARB", "SW2MARB", MACRO_SHORT(3)},
    {"SW1SKULL", "SW2SKULL", MACRO_SHORT(3)},
# endif
    {"\0", "\0", MACRO_SHORT(0)}
};
#endif
static world_Material **switchlist; /// @todo fixme: Never free'd!
static int max_numswitches;
static int numswitches;
#if __JHEXEN__
void P_InitSwitchList() {
    uri_s *uri = Uri_NewWithPath2("Textures:", RC_NULL);
    AutoStr *path = AutoStr_NewStd();
    int index = 0;
    for(int i = 0; ; ++i) {
        if(index+1 >= max_numswitches) {
            switchlist = (world_Material **) M_Realloc(switchlist, sizeof(*switchlist) *
                (max_numswitches = max_numswitches ? max_numswitches*2 : 8));
        }
        if(!switchInfo[i].soundID) break;
        Str_PercentEncode(Str_StripRight(Str_Set(path, switchInfo[i].name1)));
        Uri_SetPath(uri, Str_Text(path));
        switchlist[index++] = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(uri));
        Str_PercentEncode(Str_StripRight(Str_Set(path, switchInfo[i].name2)));
        Uri_SetPath(uri, Str_Text(path));
        switchlist[index++] = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(uri));
    }
    Uri_Delete(uri);
    numswitches = index / 2;
    switchlist[index] = 0;
}
#else
/**
 * This routine modified to read its data from a predefined lump or
 * PWAD lump called SWITCHES rather than a static table in this module to
 * allow wad designers to insert or modify switches.
 *
 * Lump format is an array of byte packed switchlist_t structures, terminated
 * by a structure with episode == -0. The lump can be generated from a
 * text source file using SWANTBLS.EXE, distributed with the BOOM utils.
 * The standard list of switches and animations is contained in the example
 * source text file DEFSWANI.DAT also in the BOOM util distribution.
 *
 * @todo Implement a better method for creating new switches.
 */
void P_InitSwitchList() {
    int episode = 0;
# if __JHERETIC__
    if(gameMode == heretic_shareware)
        episode = 1;
    else
        episode = 2;
# else
#  if __JDOOM__
    if(gameModeBits & (GM_ANY_DOOM^GM_DOOM_SHAREWARE))
        episode = 2;
    else if(gameModeBits & GM_ANY_DOOM2)
        episode = 3;
    else
#  endif
        episode = 1;
# endif
    switchlist_t *sList = switchInfo;
    de::File1 *lump = 0;
    if(CentralLumpIndex().contains("SWITCHES.lmp")) {
        lump = &CentralLumpIndex()[CentralLumpIndex().findLast("SWITCHES.lmp")];
        App_Log(DE2_RES_VERBOSE, "Processing lump %s::SWITCHES", F_PrettyPath(lump->container().composePath().toUtf8().constData()));
        sList = (switchlist_t *) lump->cache();
    }
    else {
        App_Log(DE2_RES_VERBOSE, "Registering default switches...");
    }
    uri_s *uri = Uri_New();
    Uri_SetScheme(uri, "Textures");
    ddstring_t path; Str_Init(&path);
    int index = 0;
    for(int i = 0; ;++i) {
        if(index + 1 >= max_numswitches) {
            switchlist = (world_Material **) M_Realloc(switchlist, sizeof(*switchlist) * (max_numswitches = max_numswitches ? max_numswitches*2 : 8));
        }
        if(DD_SHORT(sList[i].episode) <= episode) {
            if(!DD_SHORT(sList[i].episode)) break;
            Str_PercentEncode(Str_StripRight(Str_Set(&path, sList[i].name1)));
            Uri_SetPath(uri, Str_Text(&path));
            switchlist[index++] = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(uri));
            Str_PercentEncode(Str_StripRight(Str_Set(&path, sList[i].name2)));
            Uri_SetPath(uri, Str_Text(&path));
            switchlist[index++] = (world_Material *)P_ToPtr(DMU_MATERIAL, Materials_ResolveUri(uri));
            App_Log(lump? DE2_RES_VERBOSE : DE2_RES_XVERBOSE,
                    "  %d: Epi:%d A:\"%s\" B:\"%s\"", i, DD_SHORT(sList[i].episode),
                    sList[i].name1, sList[i].name2);
        }
    }
    Str_Free(&path);
    Uri_Delete(uri);
    if(lump) {
        lump->unlock();
    }
    numswitches = index / 2;
    switchlist[index] = 0;
}
#endif
static world_Material *findSwitch(world_Material *mat, const switchlist_t** info) {
    if (!mat) return 0;
    for (int i = 0; i < numswitches * 2; ++i) {
        if (switchlist[i] == mat) {
            if (info) {
                *info = &switchInfo[i / 2];
            }
            return switchlist[i^1];
        }
    }
    return 0;
}
static void playSwitchSound(Side *side, uint sectionFlags, int sound) {
    if (cfg.common.switchSoundOrigin == 1 /* vanilla behavior */)
    {
        S_SectorSound(reinterpret_cast<Sector *>(P_GetPtrp(side, DMU_SECTOR)), sound);
    }
    else {
        mobj_t const *sideEmitter = reinterpret_cast<mobj_t const *>
                (P_GetPtrp(side, DMU_EMITTER | sectionFlags));
        S_StopSound(0, sideEmitter);
        S_StartSound(sound, sideEmitter);
    }
}
void T_MaterialChanger(void *materialChangerThinker) {
    materialchanger_t *mchanger = (materialchanger_t *)materialChangerThinker;
    if(!(--mchanger->timer)) {
        uint const sectionFlags = DMU_FLAG_FOR_SIDESECTION(mchanger->section);
        P_SetPtrp(mchanger->side, sectionFlags | DMU_MATERIAL, mchanger->material);
        int const sound =
#if __JDOOM__ || __JDOOM64__
            SFX_SWTCHN;
#elif __JHERETIC__
            SFX_SWITCH;
#else
            0;
#endif
        if (sound > 0) {
            playSwitchSound(mchanger->side, sectionFlags, sound);
        }
        Thinker_Remove(&mchanger->thinker);
    }
}
void materialchanger_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteByte(writer, 0);
    Writer_WriteInt32(writer, timer);
    Writer_WriteInt32(writer, P_ToIndex(side));
    Writer_WriteByte(writer, (byte) section);
    Writer_WriteInt16(writer, msw->serialIdFor(material));
}
int materialchanger_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
    /*int ver =*/ Reader_ReadByte(reader);
    /*byte type =*/ Reader_ReadByte(reader);
    timer = Reader_ReadInt32(reader);
    int sideIndex = (int) Reader_ReadInt32(reader);
    if(mapVersion >= 12) {
        side = (Side *)P_ToPtr(DMU_SIDE, sideIndex);
    }
    else {
        side = msr->side(sideIndex);
    }
    DENG_ASSERT(side != 0);
    section = (SideSection) Reader_ReadByte(reader);
    material = (world_Material *) msr->material(Reader_ReadInt16(reader), 0);
    thinker.function = T_MaterialChanger;
    return true; // Add this thinker.
}
static void spawnMaterialChanger(Side *side, SideSection section, world_Material *mat, int tics) {
    materialchanger_t *mchanger = (materialchanger_t *)Z_Calloc(sizeof(*mchanger), PU_MAP, 0);
    mchanger->thinker.function = T_MaterialChanger;
    Thinker_Add(&mchanger->thinker);
    mchanger->side     = side;
    mchanger->section  = section;
    mchanger->material = mat;
    mchanger->timer    = tics;
}
struct findmaterialchangerparams_t {
    Side *side;
    SideSection section;
};
static int findMaterialChanger(thinker_t *th, void *context) {
    materialchanger_t *mchanger = (materialchanger_t *) th;
    findmaterialchangerparams_t *params = (findmaterialchangerparams_t *) context;
    if(mchanger->side == params->side &&
       mchanger->section == params->section)
        return true; // Stop iteration.
    return false; // Keep looking.
}
static void startButton(Side *side, SideSection section, world_Material *mat, int tics) {
    findmaterialchangerparams_t parm;
    parm.side    = side;
    parm.section = section;
    if (!Thinker_Iterate(T_MaterialChanger, findMaterialChanger, &parm)) {
        spawnMaterialChanger(side, section, mat, tics);
    }
}
static int chooseDefaultSound(switchlist_t const *info) {
#if __JHEXEN__
    return info->soundID;
#else
#  if __JHERETIC__
    return SFX_SWITCH;
#  else
    return SFX_SWTCHN;
# endif
    DENG_UNUSED(info);
#endif
}
dd_bool P_ToggleSwitch2(Side *side, SideSection section, int sound, dd_bool silent, int tics) {
    uint const sectionFlags = DMU_FLAG_FOR_SIDESECTION(section);
    world_Material *current = reinterpret_cast<world_Material *>
                              (P_GetPtrp(side, sectionFlags | DMU_MATERIAL));
    switchlist_t const *info;
    if (world_Material *mat = findSwitch(current, &info)) {
        if (!silent) {
            if (!sound) {
                sound = chooseDefaultSound(info);
            }
            playSwitchSound(side, sectionFlags, sound);
        }
        P_SetPtrp(side, sectionFlags | DMU_MATERIAL, mat);
        if (tics > 0) {
            startButton(side, section, current, tics);
        }
        return true;
    }
    return false;
}
dd_bool P_ToggleSwitch(Side *side, int sound, dd_bool silent, int tics) {
    if(P_ToggleSwitch2(side, SS_TOP, sound, silent, tics))
        return true;
    if(P_ToggleSwitch2(side, SS_MIDDLE, sound, silent, tics))
        return true;
    if(P_ToggleSwitch2(side, SS_BOTTOM, sound, silent, tics))
        return true;
    return false;
}
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
dd_bool P_UseSpecialLine(mobj_t *activator, Line *line, int side) {
    if(XL_UseLine(line, side, activator)) {
        return true;
    }
    return P_UseSpecialLine2(activator, line, side);
}
#endif
int mapTime;
int actualMapTime;
int timerGame;
void P_RunPlayers(timespan_t ticLength) {
    for(int i = 0; i < MAXPLAYERS; ++i) {
        if(players[i].plr->inGame) {
            P_PlayerThink(&players[i], ticLength);
        }
    }
}
void P_DoTick() {
    Pause_Ticker();
    if(paused) return;
    actualMapTime++;
    if(!IS_CLIENT && timerGame && !paused) {
        if(!--timerGame) {
            G_SetGameActionMapCompleted(gfw_Session()->mapUriForNamedExit("next"));
        }
    }
    if(!IS_NETGAME && (Hu_MenuIsActive() || Hu_IsMessageActive()) &&
       !Get(DD_PLAYBACK) && mapTime > 1)
        return;
    Thinker_Run();
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    XG_Ticker();
#endif
#if __JHEXEN__
    P_AnimateLightning();
#endif
#if __JDOOM64__
    P_ThunderSector();
#endif
    P_ProcessDeferredSpawns();
#if __JHERETIC__
    P_AmbientSound();
#endif
    for(int i = 0; i < MAXPLAYERS; ++i) {
        R_UpdateConsoleView(i);
    }
    R_UpdateSpecialFilter(DISPLAYPLAYER);
    mapTime++;
}
#define PAUSEF_PAUSED           0x1
#define PAUSEF_FORCED_PERIOD    0x2
int paused;
static int gamePauseWhenFocusLost; // cvar
static int gameUnpauseWhenFocusGained; // cvar
#ifdef __JDOOM__
static int gamePauseAfterMapStartTics = -1; // cvar
#else
static int gamePauseAfterMapStartTics = 7; // cvar
#endif
static int forcedPeriodTicsRemaining;
static void beginPause(int flags) {
    if(!paused) {
        paused = PAUSEF_PAUSED | flags;
        S_StopSound(0, 0);
        NetSv_Paused(paused);
    }
}
static void endPause() {
    if(paused) {
        LOG_VERBOSE("Pause ends (state:%i)") << paused;
        forcedPeriodTicsRemaining = 0;
        if(!(paused & PAUSEF_FORCED_PERIOD)) {
            DD_Execute(true, "resetctlaccum");
        }
        NetSv_Paused(0);
    }
    paused = 0;
}
static void checkForcedPeriod() {
    if((paused != 0) && (paused & PAUSEF_FORCED_PERIOD)) {
        if(forcedPeriodTicsRemaining-- <= 0) {
            endPause();
        }
    }
}
dd_bool Pause_IsPaused() {
    return (paused != 0) || (!IS_NETGAME && (Hu_MenuIsActive() || Hu_IsMessageActive()));
}
dd_bool Pause_IsUserPaused() {
    return (paused != 0) && !(paused & PAUSEF_FORCED_PERIOD);
}
D_CMD(Pause) {
    DENG2_UNUSED3(src, argc, argv);
    if(G_QuitInProgress())
        return false;
    Pause_Set(!(paused & PAUSEF_PAUSED));
    return true;
}
void Pause_Set(dd_bool yes) {
    if(Hu_MenuIsActive() || Hu_IsMessageActive() || IS_CLIENT)
        return; // Nope.
    if(yes)
        beginPause(0);
    else
        endPause();
}
void Pause_End() {
    endPause();
}
void Pause_SetForcedPeriod(int tics) {
    if(tics <= 0) return;
    LOG_MSG("Forced pause for %i tics") << tics;
    forcedPeriodTicsRemaining = tics;
    beginPause(PAUSEF_FORCED_PERIOD);
}
void Pause_Ticker() {
    checkForcedPeriod();
}
dd_bool Pause_Responder(event_t *ev) {
    if(ev->type == EV_FOCUS) {
        if(gamePauseWhenFocusLost && !ev->data1) {
            Pause_Set(true);
            return true;
        }
        else if(gameUnpauseWhenFocusGained && ev->data1) {
            Pause_Set(false);
            return true;
        }
    }
    return false;
}
void Pause_MapStarted() {
    if(!IS_CLIENT) {
        if(gamePauseAfterMapStartTics < 0) {
            Pause_SetForcedPeriod(Con_GetInteger("con-transition-tics"));
        }
        else {
            Pause_SetForcedPeriod(gamePauseAfterMapStartTics);
        }
    }
}
void Pause_Register() {
    forcedPeriodTicsRemaining = 0;
    gamePauseWhenFocusLost     = true;
    gameUnpauseWhenFocusGained = false;
    C_CMD("pause", "", Pause);
#define READONLYCVAR    (CVF_READ_ONLY|CVF_NO_MAX|CVF_NO_MIN|CVF_NO_ARCHIVE)
    C_VAR_INT("game-paused",              &paused,                     READONLYCVAR, 0,  0);
    C_VAR_INT("game-pause-focuslost",     &gamePauseWhenFocusLost,     0,            0,  1);
    C_VAR_INT("game-unpause-focusgained", &gameUnpauseWhenFocusGained, 0,            0,  1);
    C_VAR_INT("game-pause-mapstart-tics", &gamePauseAfterMapStartTics, 0,            -1, 70);
#undef READONLYCVAR
}
void NetSv_Paused(int pauseState) {
    if(!IS_SERVER || !IS_NETGAME)
        return;
    writer_s *writer = D_NetWrite();
    Writer_WriteByte(writer,
            (pauseState & PAUSEF_PAUSED?        1 : 0) |
            (pauseState & PAUSEF_FORCED_PERIOD? 2 : 0));
    Net_SendPacket(DDSP_ALL_PLAYERS, GPT_PAUSE, Writer_Data(writer), Writer_Size(writer));
}
void NetCl_Paused(reader_s *msg) {
    byte flags = Reader_ReadByte(msg);
    paused = 0;
    if(flags & 1) {
        paused |= PAUSEF_PAUSED;
    }
    if(flags & 2) {
        paused |= PAUSEF_FORCED_PERIOD;
    }
    DD_SetInteger(DD_CLIENT_PAUSED, paused != 0);
}
#define MESSAGETICS             (4 * TICSPERSEC)
#define CAMERA_FRICTION_THRESHOLD (.4f)
struct weaponslotinfo_t {
    uint num;
    weapontype_t *types;
};
static weaponslotinfo_t weaponSlots[NUM_WEAPON_SLOTS];
static byte slotForWeaponType(weapontype_t type, uint *position = 0) {
    byte i = 0, found = 0;
    do {
        weaponslotinfo_t *slot = &weaponSlots[i];
        uint j = 0;
        while(!found && j < slot->num) {
            if(slot->types[j] == type) {
                found = i + 1;
                if(position) {
                    *position = j;
                }
            }
            else {
                j++;
            }
        }
    } while(!found && ++i < NUM_WEAPON_SLOTS);
    return found;
}
static void unlinkWeaponInSlot(byte slotidx, weapontype_t type) {
    weaponslotinfo_t *slot = &weaponSlots[slotidx - 1];
    uint i;
    for(i = 0; i < slot->num; ++i) {
        if(slot->types[i] == type)
            break;
    }
    if(i == slot->num)
        return; // Not linked to this slot.
    memmove(&slot->types[i], &slot->types[i+1], sizeof(weapontype_t) * (slot->num - 1 - i));
    slot->types = (weapontype_t *)M_Realloc(slot->types, sizeof(weapontype_t) * --slot->num);
}
static void linkWeaponInSlot(byte slotidx, weapontype_t type) {
    weaponslotinfo_t *slot = &weaponSlots[slotidx-1];
    slot->types = (weapontype_t *)M_Realloc(slot->types, sizeof(weapontype_t) * ++slot->num);
    if(slot->num > 1) {
        memmove(&slot->types[1], &slot->types[0],
                sizeof(weapontype_t) * (slot->num - 1));
    }
    slot->types[0] = type;
}
void P_InitWeaponSlots() {
    de::zap(weaponSlots);
}
void P_FreeWeaponSlots() {
    for(int i = 0; i < NUM_WEAPON_SLOTS; ++i) {
        weaponslotinfo_t *slot = &weaponSlots[i];
        M_Free(slot->types);
        slot->types = 0;
        slot->num   = 0;
    }
}
dd_bool P_SetWeaponSlot(weapontype_t type, byte slot) {
    if(slot > NUM_WEAPON_SLOTS)
        return false;
    byte currentSlot = slotForWeaponType(type);
    if(currentSlot) {
        unlinkWeaponInSlot(currentSlot, type);
    }
    if(slot != 0) {
        linkWeaponInSlot(slot, type);
    }
    return true;
}
byte P_GetWeaponSlot(weapontype_t type) {
    if(type >= WT_FIRST && type < NUM_WEAPON_TYPES) {
        return slotForWeaponType(type);
    }
    return 0;
}
weapontype_t P_WeaponSlotCycle(weapontype_t type, dd_bool prev) {
    if(type >= WT_FIRST && type < NUM_WEAPON_TYPES) {
        uint position;
        if(byte slotidx = slotForWeaponType(type, &position)) {
            weaponslotinfo_t *slot = &weaponSlots[slotidx - 1];
            if(slot->num > 1) {
                if(prev) {
                    if(position == 0)
                        position = slot->num - 1;
                    else
                        position--;
                }
                else {
                    if(position == slot->num - 1)
                        position = 0;
                    else
                        position++;
                }
                return slot->types[position];
            }
        }
    }
    return type;
}
int P_IterateWeaponsBySlot(byte slot, dd_bool reverse, int (*callback) (weapontype_t, void *context),
    void *context) {
    int result = 1;
    if(slot <= NUM_WEAPON_SLOTS) {
        uint i = 0;
        weaponslotinfo_t const *sl = &weaponSlots[slot];
        while(i < sl->num &&
             (result = callback(sl->types[reverse ? sl->num - 1 - i : i],
                                context)) != 0) {
            i++;
        }
    }
    return result;
}
#if __JHEXEN__
void P_InitPlayerClassInfo() {
    PCLASS_INFO(PCLASS_FIGHTER)->niceName = GET_TXT(TXT_PLAYERCLASS1);
    PCLASS_INFO(PCLASS_CLERIC)->niceName = GET_TXT(TXT_PLAYERCLASS2);
    PCLASS_INFO(PCLASS_MAGE)->niceName = GET_TXT(TXT_PLAYERCLASS3);
    PCLASS_INFO(PCLASS_PIG)->niceName = GET_TXT(TXT_PLAYERCLASS4);
}
#endif
int P_GetPlayerNum(player_t const *player) {
    for(int i = 0; i < MAXPLAYERS; ++i) {
        if(player == &players[i]) {
            return i;
        }
    }
    return 0;
}
int P_GetPlayerCheats(player_t const *player) {
    if(!player) return 0;
    if(player->plr->flags & DDPF_CAMERA) {
        return (player->cheats | CF_GODMODE | (cfg.common.cameraNoClip? CF_NOCLIP : 0));
    }
    return player->cheats;
}
int P_CountPlayersInGame(PlayerSelectionCriteria const &criteria) {
    int count = 0;
    for(int i = 0; i < MAXPLAYERS; ++i) {
        player_t *player = players + i;
        if(!player->plr->inGame) continue;
        if((criteria & LocalOnly) && !(player->plr->flags & DDPF_LOCAL)) continue;
        count += 1;
    }
    return count;
}
dd_bool P_PlayerInWalkState(player_t *pl) {
    if(!pl->plr->mo) return false;
#if __JDOOM__
    return pl->plr->mo->state - STATES - PCLASS_INFO(pl->class_)->runState < 4;
#endif
#if __JHERETIC__
    return pl->plr->mo->state - STATES - PCLASS_INFO(pl->class_)->runState < 4;
#endif
#if __JHEXEN__
    return ((unsigned) ((pl->plr->mo->state - STATES) - PCLASS_INFO(pl->class_)->runState) < 4);
#endif
#if __JDOOM64__
    return pl->plr->mo->state - STATES - PCLASS_INFO(pl->class_)->runState < 4;
#endif
}
void P_TrajectoryNoise(angle_t *angle, float *slope, float degreesPhi, float degreesTheta) {
    Slope trajectory(float(*angle) / ANG180 * DD_PI, *slope);
    const Vector2f angles(degreesPhi * (randf() - randf()), degreesTheta * (randf() - randf()));
    const Vector3f front = trajectory.toUnitVec();
    const Vector3f side  = front.cross(Vector3f(0, 0, 1)).normalize();
    const Vector3f up    = front.cross(side);
    trajectory = Slope::fromVec(Matrix4f::rotate(angles.x, up) *
                                Matrix4f::rotate(angles.y, side) * front);
    *angle = int(trajectory.angle / DD_PI * ANG180);
    *slope = trajectory.slope;
}
void P_ShotAmmo(player_t *player) {
    DENG_ASSERT(player != 0);
    weaponinfo_t *wInfo = &weaponInfo[player->readyWeapon][player->class_];
    if(IS_CLIENT) return; // Server keeps track of this.
    int fireMode = 0;
#if __JHERETIC__
    if(gfw_Rule(deathmatch))
        fireMode = 0; // In deathmatch always use mode zero.
    else
        fireMode = (player->powers[PT_WEAPONLEVEL2]? 1 : 0);
#endif
    for(int i = 0; i < NUM_AMMO_TYPES; ++i) {
        if(!wInfo->mode[fireMode].ammoType[i])
            continue; // Weapon does not take this ammo.
        player->ammo[i].owned = MAX_OF(0,
            player->ammo[i].owned - wInfo->mode[fireMode].perShot[i]);
    }
    player->update |= PSF_AMMO;
}
weapontype_t P_MaybeChangeWeapon(player_t *player, weapontype_t weapon, ammotype_t ammo, dd_bool force) {
    if(IS_NETWORK_SERVER) {
        NetSv_MaybeChangeWeapon(player - players, weapon, ammo, force);
        return WT_NOCHANGE;
    }
    App_Log(DE2_DEV_MAP_XVERBOSE,
            "P_MaybeChangeWeapon: plr %i, weapon %i, ammo %i, force %i",
            (int)(player - players), weapon, ammo, force);
    int pclass = player->class_;
    int lvl = 0;
#if __JHERETIC__
    if(player->powers[PT_WEAPONLEVEL2])
        lvl = 1;
#endif
    bool found = false;
    weapontype_t retVal = WT_NOCHANGE;
    if(weapon == WT_NOCHANGE && ammo == AT_NOAMMO) // Out of ammo.
    {
        bool good;
        for(int i = 0; i < NUM_WEAPON_TYPES && !found; ++i) {
            weapontype_t candidate = weapontype_t(cfg.common.weaponOrder[i]);
            weaponinfo_t *winf     = &weaponInfo[candidate][pclass];
            if(!(winf->mode[lvl].gameModeBits & gameModeBits))
                continue;
            if(!player->weapons[candidate].owned)
                continue;
            good = true;
            for(int ammotype = 0; ammotype < NUM_AMMO_TYPES && good; ++ammotype) {
                if(!winf->mode[lvl].ammoType[ammotype])
                    continue; // Weapon does not take this type of ammo.
#if __JHERETIC__
                if(gfw_Rule(deathmatch) &&
                   player->ammo[ammotype].owned < winf->mode[0].perShot[ammotype]) {
                    good = false;
                }
                else
#endif
                if(player->ammo[ammotype].owned < winf->mode[lvl].perShot[ammotype]) {
                    good = false;
                }
            }
            if(good) {
                retVal = candidate;
                found = true;
            }
        }
    }
    else if(weapon != WT_NOCHANGE) // Player was given a NEW weapon.
    {
        if(force) {
            retVal = weapon;
        }
        else if(!((player->brain.attack) && cfg.common.noWeaponAutoSwitchIfFiring)) {
            if(cfg.common.weaponAutoSwitch == 2) // Behavior: Always change
            {
                retVal = weapon;
            }
            else if(cfg.common.weaponAutoSwitch == 1) // Behavior: Change if better
            {
                for(int i = 0; i < NUM_WEAPON_TYPES; ++i) {
                    weapontype_t candidate = weapontype_t(cfg.common.weaponOrder[i]);
                    weaponinfo_t *winf     = &weaponInfo[candidate][pclass];
                    if(!(winf->mode[lvl].gameModeBits & gameModeBits))
                        continue;
                    if(weapon == candidate) {
                        retVal = weapon;
                    }
                    else if(player->readyWeapon == candidate) {
                        break;
                    }
                }
            }
        }
    }
    else if(ammo != AT_NOAMMO) // Player is about to be given some ammo.
    {
        if(force || (!(player->ammo[ammo].owned > 0) && cfg.common.ammoAutoSwitch)) {
            for(int i = 0; i < NUM_WEAPON_TYPES; ++i) {
                weapontype_t candidate = weapontype_t(cfg.common.weaponOrder[i]);
                weaponinfo_t *winf     = &weaponInfo[candidate][pclass];
                if(!(winf->mode[lvl].gameModeBits & gameModeBits))
                    continue;
                if(!player->weapons[candidate].owned)
                    continue;
                if(!(winf->mode[lvl].ammoType[ammo]))
                    continue;
                /**
                 * @todo Have we got enough of ALL used ammo types?
                 *
                 * Problem, since the ammo has not been given yet (could be an object that gives
                 * several ammo types eg backpack) we can't test for this with what we know!
                 *
                 * This routine should be called AFTER the new ammo has been given. Somewhat complex
                 * logic to decipher first...
                 */
                if(cfg.common.ammoAutoSwitch == 2) // Behavior: Always change
                {
                    retVal = candidate;
                    break;
                }
                else if(cfg.common.ammoAutoSwitch == 1 && player->readyWeapon == candidate) {
                    break;
                }
            }
        }
    }
    if(retVal == player->readyWeapon)
        retVal = WT_NOCHANGE;
    if(retVal != WT_NOCHANGE) {
        App_Log(DE2_DEV_MAP_XVERBOSE, "P_MaybeChangeWeapon: Player %i decided to change to weapon %i",
                (int)(player - players), retVal);
        player->pendingWeapon = retVal;
        if(IS_CLIENT) {
            NetCl_PlayerActionRequest(player, GPA_CHANGE_WEAPON, player->pendingWeapon);
        }
    }
    return retVal;
}
dd_bool P_CheckAmmo(player_t *plr) {
    DENG_ASSERT(plr != 0);
    weaponinfo_t *wInfo = &weaponInfo[plr->readyWeapon][plr->class_];
    int fireMode = 0;
#if __JHERETIC__
    if(plr->powers[PT_WEAPONLEVEL2] && !gfw_Rule(deathmatch)) {
        fireMode = 1;
    }
#endif
#if __JHEXEN__
    if(plr->class_ == PCLASS_FIGHTER && plr->readyWeapon != WT_FOURTH) {
        return true;
    }
#endif
    bool good = true;
    for(int i = 0; i < NUM_AMMO_TYPES && good; ++i) {
        if(!wInfo->mode[fireMode].ammoType[i])
            continue; // Weapon does not take this type of ammo.
        if(plr->ammo[i].owned < wInfo->mode[fireMode].perShot[i]) {
            good = false;
        }
    }
    if(good) return true;
    P_MaybeChangeWeapon(plr, WT_NOCHANGE, AT_NOAMMO, false);
    if(plr->pendingWeapon != WT_NOCHANGE) {
        P_SetPsprite(plr, ps_weapon, statenum_t(wInfo->mode[fireMode].states[WSN_DOWN]));
    }
    return false;
}
weapontype_t P_PlayerFindWeapon(player_t *player, dd_bool prev) {
    DENG_ASSERT(player != 0);
#if __JDOOM__
    static weapontype_t wp_list[] = {
        WT_FIRST, WT_SECOND, WT_THIRD, WT_NINETH, WT_FOURTH,
        WT_FIFTH, WT_SIXTH, WT_SEVENTH, WT_EIGHTH
    };
#elif __JDOOM64__
    static weapontype_t wp_list[] = {
        WT_FIRST, WT_SECOND, WT_THIRD, WT_NINETH, WT_FOURTH,
        WT_FIFTH, WT_SIXTH, WT_SEVENTH, WT_EIGHTH, WT_TENTH
    };
#elif __JHERETIC__
    static weapontype_t wp_list[] = {
        WT_FIRST, WT_SECOND, WT_THIRD, WT_FOURTH, WT_FIFTH,
        WT_SIXTH, WT_SEVENTH, WT_EIGHTH
    };
#elif __JHEXEN__ || __JSTRIFE__
    static weapontype_t wp_list[] = {
        WT_FIRST, WT_SECOND, WT_THIRD, WT_FOURTH
    };
#endif
    int lvl = 0;
#if __JHERETIC__
    lvl = (player->powers[PT_WEAPONLEVEL2]? 1 : 0);
#endif
    weapontype_t *list;
    if(cfg.common.weaponNextMode) {
        list = (weapontype_t *) cfg.common.weaponOrder;
        prev = !prev; // Invert order.
    }
    else {
        list = wp_list;
    }
    int i = 0;
    weapontype_t w = WT_FIRST;
    for(; i < NUM_WEAPON_TYPES; ++i) {
        w = list[i];
        if(!cfg.common.weaponCycleSequential || player->pendingWeapon == WT_NOCHANGE) {
            if(w == player->readyWeapon)
                break;
        }
        else if(w == player->pendingWeapon) {
            break;
        }
    }
    weapontype_t initial = w;
    for(;;) {
        if(prev)
            i--;
        else
            i++;
        if(i < 0)
            i = NUM_WEAPON_TYPES - 1;
        else if(i > NUM_WEAPON_TYPES - 1)
            i = 0;
        w = list[i];
        if(w == initial)
            break;
        if((weaponInfo[w][player->class_].mode[lvl].
            gameModeBits & gameModeBits) && player->weapons[w].owned)
            break;
    }
    return w;
}
#if __JHEXEN__
void P_PlayerChangeClass(player_t *player, playerclass_t newClass) {
    DENG_ASSERT(player != 0);
    if(newClass < PCLASS_FIRST || newClass >= NUM_PLAYER_CLASSES)
        return;
    if(player->morphTics) return;
    if(!PCLASS_INFO(newClass)->userSelectable) return;
    player->class_ = newClass;
    cfg.playerClass[player - players] = newClass;
    P_ClassForPlayerWhenRespawning(player - players, true /*clear change request*/);
    for(int i = 0; i < NUMARMOR; ++i) {
        player->armorPoints[i] = 0;
    }
    player->update |= PSF_ARMOR_POINTS;
    P_PostMorphWeapon(player, WT_FIRST);
    if(player->plr->mo) {
        mobj_t *oldMo = player->plr->mo;
        P_SpawnPlayer(player - players, newClass, oldMo->origin[VX],
                      oldMo->origin[VY], oldMo->origin[VZ], oldMo->angle, 0,
                      P_MobjIsCamera(oldMo), true);
        P_MobjRemove(oldMo, true);
    }
}
#endif
void P_SetMessageWithFlags(const player_t *pl, char const *msg, int flags) {
    DENG2_ASSERT(pl);
    if(!msg || !msg[0]) return;
    ST_LogPost(pl - players, flags, msg);
    if(pl == &players[CONSOLEPLAYER]) {
        App_Log(DE2_LOG_MAP | (cfg.common.echoMsg? DE2_LOG_NOTE : DE2_LOG_VERBOSE), "%s", msg);
    }
    NetSv_SendMessage(pl - players, msg);
}
void P_SetMessage(const player_t *plr, char const *msg) {
    P_SetMessageWithFlags(plr, msg, 0);
}
#if __JHEXEN__
void P_SetYellowMessageWithFlags(player_t *pl, char const *msg, int flags) {
#define YELLOW_FMT      "{r=1;g=0.7;b=0.3;}"
#define YELLOW_FMT_LEN  18
    if(!msg || !msg[0]) return;
    size_t len = strlen(msg);
    AutoStr *buf = AutoStr_NewStd();
    Str_Reserve(buf, YELLOW_FMT_LEN + len+1);
    Str_Set(buf, YELLOW_FMT);
    Str_Appendf(buf, "%s", msg);
    ST_LogPost(pl - players, flags, Str_Text(buf));
    if(pl == &players[CONSOLEPLAYER]) {
        App_Log(DE2_LOG_MAP | (cfg.common.echoMsg? DE2_LOG_NOTE : DE2_LOG_VERBOSE), "%s", msg);
    }
    NetSv_SendMessage(pl - players, Str_Text(buf));
#undef YELLOW_FMT_LEN
#undef YELLOW_FMT
}
void P_SetYellowMessage(player_t *pl, char const *msg) {
    P_SetYellowMessageWithFlags(pl, msg, 0);
}
#endif
void P_Thrust3D(player_t *player, angle_t angle, float lookdir, coord_t forwardMove, coord_t sideMove) {
    angle_t pitch = int(LOOKDIR2DEG(lookdir) / 360 * ANGLE_MAX);
    angle_t sideangle = angle - ANG90;
    mobj_t *mo = player->plr->mo;
    coord_t zmul, mom[3];
    angle >>= ANGLETOFINESHIFT;
    pitch >>= ANGLETOFINESHIFT;
    mom[MX] = forwardMove * FIX2FLT(finecosine[angle]);
    mom[MY] = forwardMove * FIX2FLT(finesine[angle]);
    mom[MZ] = forwardMove * FIX2FLT(finesine[pitch]);
    zmul = FIX2FLT(finecosine[pitch]);
    mom[MX] *= zmul;
    mom[MY] *= zmul;
    sideangle >>= ANGLETOFINESHIFT;
    mom[MX] += sideMove * FIX2FLT(finecosine[sideangle]);
    mom[MY] += sideMove * FIX2FLT(finesine[sideangle]);
    mo->mom[MX] += mom[MX];
    mo->mom[MY] += mom[MY];
    mo->mom[MZ] += mom[MZ];
}
int P_CameraXYMovement(mobj_t *mo) {
    if(!P_MobjIsCamera(mo))
        return false;
#if __JDOOM__ || __JDOOM64__
    if(mo->flags & MF_NOCLIP ||
       P_CheckPositionXYZ(mo, mo->origin[VX] + mo->mom[MX], mo->origin[VY] + mo->mom[MY], mo->origin[VZ])) {
#endif
        P_MobjUnlink(mo);
        mo->origin[VX] += mo->mom[MX];
        mo->origin[VY] += mo->mom[MY];
        P_MobjLink(mo);
        P_CheckPositionXY(mo, mo->origin[VX], mo->origin[VY]);
        mo->floorZ = tmFloorZ;
        mo->ceilingZ = tmCeilingZ;
#if __JDOOM__ || __JDOOM64__
    }
#endif
    if(!INRANGE_OF(mo->player->brain.forwardMove, 0, CAMERA_FRICTION_THRESHOLD) ||
       !INRANGE_OF(mo->player->brain.sideMove, 0, CAMERA_FRICTION_THRESHOLD) ||
       !INRANGE_OF(mo->player->brain.upMove, 0, CAMERA_FRICTION_THRESHOLD)) {
        mo->mom[MX] *= FRICTION_NORMAL;
        mo->mom[MY] *= FRICTION_NORMAL;
    }
    else {
        mo->mom[MX] *= FRICTION_HIGH;
        mo->mom[MY] *= FRICTION_HIGH;
    }
    return true;
}
int P_CameraZMovement(mobj_t *mo) {
    if(!P_MobjIsCamera(mo))
        return false;
    mo->origin[VZ] += mo->mom[MZ];
    if(!INRANGE_OF(mo->player->brain.forwardMove, 0, CAMERA_FRICTION_THRESHOLD) ||
       !INRANGE_OF(mo->player->brain.sideMove, 0, CAMERA_FRICTION_THRESHOLD) ||
       !INRANGE_OF(mo->player->brain.upMove, 0, CAMERA_FRICTION_THRESHOLD)) {
        mo->mom[MZ] *= FRICTION_NORMAL;
    }
    else {
        mo->mom[MZ] *= FRICTION_HIGH;
    }
    return true;
}
void P_PlayerThinkCamera(player_t *player) {
    DENG_ASSERT(player != 0);
    mobj_t *mo = player->plr->mo;
    if(!mo) return;
    if(!(player->plr->flags & DDPF_CAMERA)) {
        if(player->playerState == PST_LIVE) {
            mo->flags |= (MF_SOLID | MF_SHOOTABLE | MF_PICKUP);
        }
        return;
    }
    mo->flags &= ~(MF_SOLID | MF_SHOOTABLE | MF_PICKUP);
    if(player->viewLock) {
        mobj_t *target = players->viewLock;
        if(!target->player || !target->player->plr->inGame) {
            player->viewLock = 0;
            return;
        }
        int full = player->lockFull;
        /* $unifiedangles */
        mo->angle = M_PointToAngle2(mo->origin, target->origin);
        player->plr->flags |= DDPF_INTERYAW;
        if(full) {
            coord_t dist = M_ApproxDistance(mo->origin[VX] - target->origin[VX],
                                            mo->origin[VY] - target->origin[VY]);
            angle_t angle = M_PointXYToAngle2(0, 0,
                                              target->origin[VZ] + (target->height / 2) - mo->origin[VZ],
                                              dist);
            player->plr->lookDir = -(angle / (float) ANGLE_MAX * 360.0f - 90);
            if(player->plr->lookDir > 180)
                player->plr->lookDir -= 360;
            player->plr->lookDir *= 110.0f / 85.0f;
            if(player->plr->lookDir > 110)
                player->plr->lookDir = 110;
            if(player->plr->lookDir < -110)
                player->plr->lookDir = -110;
            player->plr->flags |= DDPF_INTERPITCH;
        }
    }
}
D_CMD(SetCamera) {
    DENG_UNUSED(src); DENG_UNUSED(argc);
    int p = atoi(argv[1]);
    if(p < 0 || p >= MAXPLAYERS) {
        App_Log(DE2_LOG_SCR| DE2_LOG_ERROR, "Invalid console number %i", p);
        return false;
    }
    player_t *player = &players[p];
    player->plr->flags ^= DDPF_CAMERA;
    if(player->plr->inGame) {
        if(player->plr->flags & DDPF_CAMERA) {
            if(player->plr->mo) {
                player->plr->mo->origin[VZ] += player->viewHeight;
            }
        }
        else {
            if(player->plr->mo) {
                player->plr->mo->origin[VZ] -= player->viewHeight;
            }
        }
    }
    return true;
}
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
int P_PlayerGiveArmorBonus(player_t *plr, int points)
#else // __JHEXEN__
int P_PlayerGiveArmorBonus(player_t *plr, armortype_t type, int points)
#endif
{
    if(!points) return 0;
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
    int *current = &plr->armorPoints;
#else // __JHEXEN__
    int *current = &plr->armorPoints[type];
#endif
    int oldPoints = *current;
    int delta;
    if(points > 0) {
        delta = points; /// @todo No upper limit?
    }
    else {
        if(*current + points < 0)
            delta = -(*current);
        else
            delta = points;
    }
    *current += delta;
    if(*current != oldPoints) {
        plr->update |= PSF_ARMOR_POINTS;
    }
    return delta;
}
#if __JDOOM__ || __JDOOM64__ || __JHERETIC__
void P_PlayerSetArmorType(player_t *plr, int type) {
    int oldType = plr->armorType;
    plr->armorType = type;
    if(plr->armorType != oldType) {
        plr->update |= PSF_ARMOR_TYPE;
    }
}
#endif
D_CMD(SetViewMode) {
    DENG_UNUSED(src);
    if(argc > 2) return false;
    int pl = CONSOLEPLAYER;
    if(argc == 2) {
        pl = atoi(argv[1]);
    }
    if(pl < 0 || pl >= MAXPLAYERS)
        return false;
    if(!(players[pl].plr->flags & DDPF_CHASECAM)) {
        players[pl].plr->flags |= DDPF_CHASECAM;
    }
    else {
        players[pl].plr->flags &= ~DDPF_CHASECAM;
    }
    return true;
}
D_CMD(SetViewLock) {
    DENG_UNUSED(src);
    int pl = CONSOLEPLAYER, lock;
    if(!qstricmp(argv[0], "lockmode")) {
        lock = atoi(argv[1]);
        if(lock)
            players[pl].lockFull = true;
        else
            players[pl].lockFull = false;
        return true;
    }
    if(argc < 2)
        return false;
    if(argc >= 3)
        pl = atoi(argv[2]); // Console number.
    lock = atoi(argv[1]);
    if(!(lock == pl || lock < 0 || lock >= MAXPLAYERS)) {
        if(players[lock].plr->inGame && players[lock].plr->mo) {
            players[pl].viewLock = players[lock].plr->mo;
            return true;
        }
    }
    players[pl].viewLock = 0;
    return false;
}
D_CMD(MakeLocal) {
    DENG_UNUSED(src); DENG_UNUSED(argc);
    if(G_GameState() != GS_MAP) {
        App_Log(DE2_LOG_ERROR | DE2_LOG_MAP, "You must be in a game to create a local player.");
        return false;
    }
    int p = atoi(argv[1]);
    if(p < 0 || p >= MAXPLAYERS) {
        App_Log(DE2_SCR_ERROR, "Invalid console number %i.", p);
        return false;
    }
    player_t *plr = &players[p];
    if(plr->plr->inGame) {
        App_Log(DE2_LOG_ERROR | DE2_LOG_MAP, "Player %i is already in the game.", p);
        return false;
    }
    plr->playerState = PST_REBORN;
    plr->plr->inGame = true;
    char buf[20];
    sprintf(buf, "conlocp %i", p);
    DD_Execute(false, buf);
    P_DealPlayerStarts(0);
    return true;
}
D_CMD(PrintPlayerCoords) {
    DENG_UNUSED(src); DENG_UNUSED(argc); DENG_UNUSED(argv);
    mobj_t *mo;
    if(G_GameState() != GS_MAP)
        return false;
    if(!(mo = players[CONSOLEPLAYER].plr->mo))
        return false;
    App_Log(DE2_LOG_MAP, "Console %i: X=%g Y=%g Z=%g", CONSOLEPLAYER,
                         mo->origin[VX], mo->origin[VY], mo->origin[VZ]);
    return true;
}
D_CMD(CycleSpy) {
    DENG_UNUSED(src); DENG_UNUSED(argc); DENG_UNUSED(argv);
    App_Log(DE2_LOG_MAP | DE2_LOG_ERROR, "Spying not allowed.");
#if 0
    if(G_GameState() == GS_MAP && !deathmatch) {
        do {
            Set(DD_DISPLAYPLAYER, DISPLAYPLAYER + 1);
            if(DISPLAYPLAYER == MAXPLAYERS) {
                Set(DD_DISPLAYPLAYER, 0);
            }
        } while(!players[DISPLAYPLAYER].plr->inGame &&
                DISPLAYPLAYER != CONSOLEPLAYER);
    }
#endif
    return true;
}
D_CMD(SpawnMobj) {
    DENG_UNUSED(src);
    if(argc != 5 && argc != 6) {
        App_Log(DE2_SCR_NOTE, "Usage: %s (type) (x) (y) (z) (angle)", argv[0]);
        App_Log(DE2_LOG_SCR, "Type must be a defined Thing ID or Name.");
        App_Log(DE2_LOG_SCR, "Z is an offset from the floor, 'floor', 'ceil' or 'random'.");
        App_Log(DE2_LOG_SCR, "Angle (0..360) is optional.");
        return true;
    }
    if(IS_CLIENT) {
        App_Log(DE2_SCR_ERROR, "%s can't be used by clients", argv[0]);
        return false;
    }
    mobjtype_t type;
    if((type = mobjtype_t(Defs().getMobjNum(argv[1]))) < 0) {
        if((type = mobjtype_t(Defs().getMobjNumForName(argv[1]))) < 0) {
            App_Log(DE2_LOG_RES | DE2_LOG_ERROR, "Undefined thing type %s", argv[1]);
            return false;
        }
    }
    coord_t pos[3];
    pos[VX] = strtod(argv[2], 0);
    pos[VY] = strtod(argv[3], 0);
    pos[VZ] = 0;
    int spawnFlags = 0;
    if(!qstricmp(argv[4], "ceil")) {
        spawnFlags |= MSF_Z_CEIL;
    }
    else if(!qstricmp(argv[4], "random")) {
        spawnFlags |= MSF_Z_RANDOM;
    }
    else {
        spawnFlags |= MSF_Z_FLOOR;
        if(qstricmp(argv[4], "floor")) {
            pos[VZ] = strtod(argv[4], 0);
        }
    }
    angle_t angle = 0;
    if(argc == 6) {
        angle = ((int) (strtod(argv[5], 0) / 360 * FRACUNIT)) << 16;
    }
    if(mobj_t *mo = P_SpawnMobj(type, pos, angle, spawnFlags)) {
#if __JDOOM64__
        if(mo->type == MT_DART) {
            S_StartSound(SFX_SKESWG, mo); // We got darts! spawn skeswg sound!
        }
        else {
            S_StartSound(SFX_ITMBK, mo); // If not dart, then spawn itmbk sound
            mo->translucency = 255;
            mo->spawnFadeTics = 0;
            mo->intFlags |= MIF_FADE;
        }
#else
        DENG_UNUSED(mo);
#endif
    }
    return true;
}
void Player_LeaveMap(player_t *player, dd_bool newHub) {
    DENG2_ASSERT(player);
    int const plrNum = player - players;
#if !__JHEXEN__
    DENG2_UNUSED(newHub);
#endif
    if(!player->plr->inGame) return;
#if __JHEXEN__
    int const flightPower = player->powers[PT_FLIGHT];
#endif
#if __JHERETIC__
    for(int i = 0; i < NUM_INVENTORYITEM_TYPES; ++i) {
        inventoryitemtype_t type = inventoryitemtype_t(IIT_FIRST + i);
        uint count = P_InventoryCount(plrNum, type);
        if(count) {
            if(type != IIT_FLY) {
                count--;
            }
            for(uint k = 0; k < count; ++k) {
                P_InventoryTake(plrNum, type, true);
            }
        }
    }
#endif
#if __JHEXEN__
    if(newHub) {
        uint count = P_InventoryCount(plrNum, IIT_FLY);
        for(uint i = 0; i < count; ++i) {
            P_InventoryTake(plrNum, IIT_FLY, true);
        }
    }
#endif
    player->update |= PSF_POWERS;
    de::zap(player->powers);
    R_UpdateSpecialFilterWithTimeDelta(plrNum, 0 /* instantly */);
#if __JHEXEN__
    if(!newHub && !gfw_Rule(deathmatch)) {
        player->powers[PT_FLIGHT] = flightPower; // Restore flight.
    }
#endif
#if __JDOOM__ || __JHERETIC__ || __JDOOM64__
    player->update |= PSF_KEYS;
    de::zap(player->keys);
#else
    if(!gfw_Rule(deathmatch) && newHub) {
        player->keys = 0;
    }
#endif
#if __JHERETIC__
    player->rain1 = nullptr;
    player->rain2 = nullptr;
#endif
#if __JHERETIC__ || __JHEXEN__
    player->update |= PSF_MORPH_TIME;
    if(player->morphTics) {
        player->readyWeapon = weapontype_t(player->plr->mo->special1); // Restore weapon.
        player->morphTics = 0;
    }
#endif
    player->plr->mo->flags &= ~MF_SHADOW; // Cancel invisibility.
    player->plr->lookDir       = 0;
    player->plr->extraLight    = 0; // Cancel gun flashes.
    player->plr->fixedColorMap = 0; // Cancel IR goggles.
    player->plr->flags &= ~DDPF_VIEW_FILTER;
    player->damageCount = 0; // No palette changes.
    player->bonusCount  = 0;
#if __JHEXEN__
    player->poisonCount = 0;
#endif
    ST_LogEmpty(plrNum);
    NetSv_SendPlayerState(plrNum, DDSP_ALL_PLAYERS, PSF_FRAGS | PSF_COUNTERS, true);
}
dd_bool Player_WaitingForReborn(player_t const *plr) {
    DENG2_ASSERT(plr != 0);
    return (plr->plr->inGame && plr->playerState == PST_REBORN && !P_MobjIsCamera(plr->plr->mo));
}
angle_t Player_ViewYawAngle(int playerNum) {
    if(playerNum < 0 || playerNum >= MAXPLAYERS) {
        return 0;
    }
    ddplayer_t *plr = players[playerNum].plr;
    angle_t ang = plr->mo->angle + (int) (ANGLE_MAX * -G_GetLookOffset(playerNum));
    if(Get(DD_USING_HEAD_TRACKING)) {
        ang -= plr->appliedBodyYaw;
    }
    return ang;
}
void player_s::write(writer_s *writer, playerheader_t &plrHdr) const {
#if __JDOOM64__ || __JHERETIC__ || __JHEXEN__
    int const plrnum = P_GetPlayerNum(this);
#endif
    player_t temp, *p = &temp;
    ddplayer_t ddtemp, *dp = &ddtemp;
    std::memcpy(p, this, sizeof(temp));
    std::memcpy(dp, plr, sizeof(ddtemp));
    temp.plr = &ddtemp;
    for(int i = 0; i < plrHdr.numPSprites; ++i) {
        pspdef_t *pspDef = &temp.pSprites[i];
        if(pspDef->state) {
            pspDef->state = (state_t *) (pspDef->state - STATES);
        }
    }
    Writer_WriteByte(writer, 6);
#if __JHEXEN__
    Writer_WriteByte(writer, cfg.playerClass[plrnum]);
#endif
    Writer_WriteInt32(writer, p->playerState);
#if __JHEXEN__
    Writer_WriteInt32(writer, p->class_);    // 2nd class...?
#endif
    Writer_WriteInt32(writer, FLT2FIX(p->viewZ));
    Writer_WriteInt32(writer, FLT2FIX(p->viewHeight));
    Writer_WriteInt32(writer, FLT2FIX(p->viewHeightDelta));
#if !__JHEXEN__
    Writer_WriteFloat(writer, dp->lookDir);
#endif
    Writer_WriteInt32(writer, FLT2FIX(p->bob));
#if __JHEXEN__
    Writer_WriteInt32(writer, p->flyHeight);
    Writer_WriteFloat(writer, dp->lookDir);
    Writer_WriteInt32(writer, p->centering);
#endif
    Writer_WriteInt32(writer, p->health);
#if __JHEXEN__
    for(int i = 0; i < plrHdr.numArmorTypes; ++i) {
        Writer_WriteInt32(writer, p->armorPoints[i]);
    }
#else
    Writer_WriteInt32(writer, p->armorPoints);
    Writer_WriteInt32(writer, p->armorType);
#endif
#if __JDOOM64__ || __JHEXEN__
    for(int i = 0; i < plrHdr.numInvItemTypes; ++i) {
        inventoryitemtype_t type = inventoryitemtype_t(IIT_FIRST + i);
        Writer_WriteInt32(writer, type);
        Writer_WriteInt32(writer, P_InventoryCount(plrnum, type));
    }
    Writer_WriteInt32(writer, P_InventoryReadyItem(plrnum));
#endif
    for(int i = 0; i < plrHdr.numPowers; ++i) {
        Writer_WriteInt32(writer, p->powers[i]);
    }
#if __JHEXEN__
    Writer_WriteInt32(writer, p->keys);
#else
    for(int i = 0; i < plrHdr.numKeys; ++i) {
        Writer_WriteInt32(writer, p->keys[i]);
    }
#endif
#if __JHEXEN__
    Writer_WriteInt32(writer, p->pieces);
#else
    Writer_WriteInt32(writer, p->backpack);
#endif
    for(int i = 0; i < plrHdr.numFrags; ++i) {
        Writer_WriteInt32(writer, p->frags[i]);
    }
    Writer_WriteInt32(writer, p->readyWeapon);
    Writer_WriteInt32(writer, p->pendingWeapon);
    for(int i = 0; i < plrHdr.numWeapons; ++i) {
        Writer_WriteInt32(writer, p->weapons[i].owned);
    }
    for(int i = 0; i < plrHdr.numAmmoTypes; ++i) {
        Writer_WriteInt32(writer, p->ammo[i].owned);
#if !__JHEXEN__
        Writer_WriteInt32(writer, p->ammo[i].max);
#endif
    }
    Writer_WriteInt32(writer, p->attackDown);
    Writer_WriteInt32(writer, p->useDown);
    Writer_WriteInt32(writer, p->cheats);
    Writer_WriteInt32(writer, p->refire);
    Writer_WriteInt32(writer, p->killCount);
    Writer_WriteInt32(writer, p->itemCount);
    Writer_WriteInt32(writer, p->secretCount);
    Writer_WriteInt32(writer, p->damageCount);
    Writer_WriteInt32(writer, p->bonusCount);
#if __JHEXEN__
    Writer_WriteInt32(writer, p->poisonCount);
#endif
    Writer_WriteInt32(writer, dp->extraLight);
    Writer_WriteInt32(writer, dp->fixedColorMap);
    Writer_WriteInt32(writer, p->colorMap);
    for(int i = 0; i < plrHdr.numPSprites; ++i) {
        pspdef_t *psp = &p->pSprites[i];
        Writer_WriteInt32(writer, PTR2INT(psp->state));
        Writer_WriteInt32(writer, psp->tics);
        Writer_WriteInt32(writer, FLT2FIX(psp->pos[VX]));
        Writer_WriteInt32(writer, FLT2FIX(psp->pos[VY]));
    }
#if !__JHEXEN__
    Writer_WriteInt32(writer, p->didSecret);
    Writer_WriteInt32(writer, p->flyHeight);
#endif
#if __JHERETIC__
    for(int i = 0; i < plrHdr.numInvItemTypes; ++i) {
        inventoryitemtype_t type = inventoryitemtype_t(IIT_FIRST + i);
        Writer_WriteInt32(writer, type);
        Writer_WriteInt32(writer, P_InventoryCount(plrnum, type));
    }
    Writer_WriteInt32(writer, P_InventoryReadyItem(plrnum));
    Writer_WriteInt32(writer, p->chickenPeck);
#endif
#if __JHERETIC__ || __JHEXEN__
    Writer_WriteInt32(writer, p->morphTics);
#endif
    Writer_WriteInt32(writer, p->airCounter);
#if __JHEXEN__
    Writer_WriteInt32(writer, p->jumpTics);
    Writer_WriteInt32(writer, p->worldTimer);
#elif __JHERETIC__
    Writer_WriteInt32(writer, p->flameCount);
    Writer_WriteByte(writer, p->class_);
#endif
}
void player_s::read(reader_s *reader, playerheader_t &plrHdr) {
    int const plrnum = P_GetPlayerNum(this);
    byte ver = Reader_ReadByte(reader);
#if __JHEXEN__
    cfg.playerClass[plrnum] = playerclass_t(Reader_ReadByte(reader));
#endif
    ddplayer_t *dp = plr;
#if __JHEXEN__
    de::zapPtr(this); // Force everything NULL,
    plr = dp;   // but restore the ddplayer pointer.
#endif
    playerState     = playerstate_t(Reader_ReadInt32(reader));
#if __JHEXEN__
    class_          = playerclass_t(Reader_ReadInt32(reader)); // 2nd class?? (ask Raven...)
#endif
    viewZ           = FIX2FLT(Reader_ReadInt32(reader));
    viewHeight      = FIX2FLT(Reader_ReadInt32(reader));
    viewHeightDelta = FIX2FLT(Reader_ReadInt32(reader));
#if !__JHEXEN__
    dp->lookDir       = Reader_ReadFloat(reader);
#endif
    bob             = FIX2FLT(Reader_ReadInt32(reader));
#if __JHEXEN__
    flyHeight       = Reader_ReadInt32(reader);
    dp->lookDir        = Reader_ReadFloat(reader);
    centering       = Reader_ReadInt32(reader);
#endif
    health          = Reader_ReadInt32(reader);
#if __JHEXEN__
    for(int i = 0; i < plrHdr.numArmorTypes; ++i) {
        armorPoints[i] = Reader_ReadInt32(reader);
    }
#else
    armorPoints     = Reader_ReadInt32(reader);
    armorType       = Reader_ReadInt32(reader);
#endif
#if __JDOOM64__ || __JHEXEN__
    P_InventoryEmpty(plrnum);
    for(int i = 0; i < plrHdr.numInvItemTypes; ++i) {
        inventoryitemtype_t type = inventoryitemtype_t(Reader_ReadInt32(reader));
        int count = Reader_ReadInt32(reader);
        for(int k = 0; k < count; ++k) {
            P_InventoryGive(plrnum, type, true);
        }
    }
    P_InventorySetReadyItem(plrnum, inventoryitemtype_t(Reader_ReadInt32(reader)));
# if __JHEXEN__
    Hu_InventorySelect(plrnum, P_InventoryReadyItem(plrnum));
    if(ver < 5) {
        /*artifactCount   =*/ Reader_ReadInt32(reader);
    }
    if(ver < 6) {
        /*inventorySlotNum =*/ Reader_ReadInt32(reader);
    }
# endif
#endif
    for(int i = 0; i < plrHdr.numPowers; ++i) {
        powers[i] = Reader_ReadInt32(reader);
    }
    if(powers[PT_ALLMAP]) {
        ST_RevealAutomap(plrnum, true);
    }
#if __JHEXEN__
    keys = Reader_ReadInt32(reader);
#else
    for(int i = 0; i < plrHdr.numKeys; ++i) {
        keys[i] = Reader_ReadInt32(reader);
    }
#endif
#if __JHEXEN__
    pieces   = Reader_ReadInt32(reader);
#else
    backpack = Reader_ReadInt32(reader);
#endif
    for(int i = 0; i < plrHdr.numFrags; ++i) {
        frags[i] = Reader_ReadInt32(reader);
    }
    readyWeapon = weapontype_t(Reader_ReadInt32(reader));
#if __JHEXEN__
    if(ver < 5)
        pendingWeapon = WT_NOCHANGE;
    else
#endif
        pendingWeapon = weapontype_t(Reader_ReadInt32(reader));
    for(int i = 0; i < plrHdr.numWeapons; ++i) {
        weapons[i].owned = (Reader_ReadInt32(reader)? true : false);
    }
    for(int i = 0; i < plrHdr.numAmmoTypes; ++i) {
        ammo[i].owned = Reader_ReadInt32(reader);
#if !__JHEXEN__
        ammo[i].max = Reader_ReadInt32(reader);
#endif
    }
    attackDown  = Reader_ReadInt32(reader);
    useDown     = Reader_ReadInt32(reader);
    cheats      = Reader_ReadInt32(reader);
    refire      = Reader_ReadInt32(reader);
    killCount   = Reader_ReadInt32(reader);
    itemCount   = Reader_ReadInt32(reader);
    secretCount = Reader_ReadInt32(reader);
#if __JHEXEN__
    if(ver <= 1) {
        /*messageTics     =*/ Reader_ReadInt32(reader);
        /*ultimateMessage =*/ Reader_ReadInt32(reader);
        /*yellowMessage   =*/ Reader_ReadInt32(reader);
    }
#endif
    damageCount = Reader_ReadInt32(reader);
    bonusCount  = Reader_ReadInt32(reader);
#if __JHEXEN__
    poisonCount = Reader_ReadInt32(reader);
#endif
    dp->extraLight    = Reader_ReadInt32(reader);
    dp->fixedColorMap = Reader_ReadInt32(reader);
    colorMap    = Reader_ReadInt32(reader);
    for(int i = 0; i < plrHdr.numPSprites; ++i) {
        pspdef_t *psp = &pSprites[i];
        psp->state   = INT2PTR(state_t, Reader_ReadInt32(reader));
        psp->tics    = Reader_ReadInt32(reader);
        psp->pos[VX] = FIX2FLT(Reader_ReadInt32(reader));
        psp->pos[VY] = FIX2FLT(Reader_ReadInt32(reader));
    }
#if !__JHEXEN__
    didSecret = Reader_ReadInt32(reader);
# if __JDOOM__ || __JDOOM64__
    if(ver == 2) {
        /*messageTics =*/ Reader_ReadInt32(reader);
    }
    if(ver >= 2) {
        flyHeight = Reader_ReadInt32(reader);
    }
# elif __JHERETIC__
    if(ver < 3) {
        /*messageTics =*/ Reader_ReadInt32(reader);
    }
    flyHeight = Reader_ReadInt32(reader);
    P_InventoryEmpty(plrnum);
    for(int i = 0; i < plrHdr.numInvItemTypes; ++i) {
        inventoryitemtype_t type = inventoryitemtype_t(Reader_ReadInt32(reader));
        int count = Reader_ReadInt32(reader);
        for(int k = 0; k < count; ++k) {
            P_InventoryGive(plrnum, type, true);
        }
    }
    P_InventorySetReadyItem(plrnum, (inventoryitemtype_t) Reader_ReadInt32(reader));
    Hu_InventorySelect(plrnum, P_InventoryReadyItem(plrnum));
    if(ver < 5) {
        Reader_ReadInt32(reader); // Current inventory item count?
    }
    if(ver < 6) {
        /*inventorySlotNum =*/ Reader_ReadInt32(reader);
    }
    chickenPeck = Reader_ReadInt32(reader);
# endif
#endif
#if __JHERETIC__ || __JHEXEN__
    morphTics = Reader_ReadInt32(reader);
#endif
    if(ver >= 2) {
        airCounter = Reader_ReadInt32(reader);
    }
#if __JHEXEN__
    jumpTics   = Reader_ReadInt32(reader);
    worldTimer = Reader_ReadInt32(reader);
#elif __JHERETIC__
    flameCount = Reader_ReadInt32(reader);
    if(ver >= 2) {
        class_ = playerclass_t(Reader_ReadByte(reader));
    }
#endif
#if !__JHEXEN__
    plr->mo = 0;
    attacker = 0;
#endif
    for(int i = 0; i < plrHdr.numPSprites; ++i) {
        if(pSprites[i].state) {
            pSprites[i].state = &STATES[PTR2INT(pSprites[i].state)];
        }
    }
    dp->flags |= DDPF_FIXORIGIN | DDPF_FIXANGLES | DDPF_FIXMOM;
    update |= PSF_REBORN;
}
String Player_WeaponId(player_t const *plr) {
    String value = "Weapon Info|";
#ifdef __JHEXEN__
    static char const *className[] = { "Fighter", "Cleric", "Mage", "Pig" };
    value.append(className[plr->class_]);
    value.append("|");
#endif
#ifdef __JHERETIC__
    if(plr->class_ == PCLASS_CHICKEN) {
        value.append("Beak");
    }
    else {
        value.append(QString::number(plr->readyWeapon));
    }
#endif
#if defined(__JDOOM__) || defined(__JDOOM64__) || defined(__JHEXEN__)
    value.append(QString::number(plr->readyWeapon));
#endif
    value.append("|Id");
    if(auto *def = Defs().getValueById(value)) {
        return def->text;
    }
    return "";
}
void Player_PostTick(player_t *player) {
    if(!player->plr->inGame) return;
    int const console = player - players;
    if(console == CONSOLEPLAYER) {
        Player_UpdateStatusCVars(player);
    }
    if(player->update & PSF_READY_WEAPON) {
        Block const id = Player_WeaponId(player).toUtf8();
        ddnotify_player_weapon_changed_t args;
        args.player = console;
        args.weapon = player->readyWeapon;
        args.weaponId = id.constData();
        Plug_Notify(DD_NOTIFY_PLAYER_WEAPON_CHANGED, &args);
    }
}
void Player_NotifyPSpriteChange(player_t *player, int position) {
    if(position == ps_weapon) {
        ddnotify_psprite_state_changed_t args;
        args.player = player - players;
        args.state  = player->pSprites[position].state;
        Plug_Notify(DD_NOTIFY_PSPRITE_STATE_CHANGED, &args);
    }
}
/**
 * Updates game status cvars for the specified player.
 */
void Player_UpdateStatusCVars(player_t const *player) {
    DENG2_ASSERT(player);
    QChar const CVAR_DELIM('-');
    static Path const var_player_health        ("player-health",         CVAR_DELIM);
    static Path const var_player_armor         ("player-armor",          CVAR_DELIM);
    static Path const var_player_weapon_current("player-weapon-current", CVAR_DELIM);
    Con_SetVariable(var_player_health, player->health, SVF_WRITE_OVERRIDE);
#if !__JHEXEN__
    static Path const var_game_stats_kills  ("game-stats-kills",   CVAR_DELIM);
    static Path const var_game_stats_items  ("game-stats-items",   CVAR_DELIM);
    static Path const var_game_stats_secrets("game-stats-secrets", CVAR_DELIM);
    Con_SetVariable(var_game_stats_kills,   player->killCount,   SVF_WRITE_OVERRIDE);
    Con_SetVariable(var_game_stats_items,   player->itemCount,   SVF_WRITE_OVERRIDE);
    Con_SetVariable(var_game_stats_secrets, player->secretCount, SVF_WRITE_OVERRIDE);
#endif
#if __JHEXEN__
    int const armorPoints = FixedDiv(PCLASS_INFO(player->class_)->autoArmorSave
                                     + player->armorPoints[ARMOR_ARMOR]
                                     + player->armorPoints[ARMOR_SHIELD]
                                     + player->armorPoints[ARMOR_HELMET]
                                     + player->armorPoints[ARMOR_AMULET], 5 * FRACUNIT) >> FRACBITS;
    Con_SetVariable(var_player_armor, armorPoints, SVF_WRITE_OVERRIDE);
#else
    Con_SetVariable(var_player_armor, player->armorPoints, SVF_WRITE_OVERRIDE);
#endif
    static Path const keyIds[NUM_KEY_TYPES] = {
#if __JDOOM__ || __JDOOM64__
        /* KT_BLUECARD */    { "player-key-blue", CVAR_DELIM },
        /* KT_YELLOWCARD */  { "player-key-yellow", CVAR_DELIM },
        /* KT_REDCARD */     { "player-key-red", CVAR_DELIM },
        /* KT_BLUESKULL */   { "player-key-blueskull", CVAR_DELIM },
        /* KT_YELLOWSKULL */ { "player-key-yellowskull", CVAR_DELIM },
        /* KT_REDSKULL */    { "player-key-redskull", CVAR_DELIM },
#elif __JHERETIC__
        /* KT_YELLOW */      { "player-key-yellow", CVAR_DELIM },
        /* KT_GREEN */       { "player-key-green", CVAR_DELIM },
        /* KT_BLUE */        { "player-key-blue", CVAR_DELIM },
#elif __JHEXEN__
        /* KT_KEY1 */        { "player-key-steel", CVAR_DELIM },
        /* KT_KEY2 */        { "player-key-cave", CVAR_DELIM },
        /* KT_KEY3 */        { "player-key-axe", CVAR_DELIM },
        /* KT_KEY4 */        { "player-key-fire", CVAR_DELIM },
        /* KT_KEY5 */        { "player-key-emerald", CVAR_DELIM },
        /* KT_KEY6 */        { "player-key-dungeon", CVAR_DELIM },
        /* KT_KEY7 */        { "player-key-silver", CVAR_DELIM },
        /* KT_KEY8 */        { "player-key-rusted", CVAR_DELIM },
        /* KT_KEY9 */        { "player-key-horn", CVAR_DELIM },
        /* KT_KEYA */        { "player-key-swamp", CVAR_DELIM },
        /* KT_KEYB */        { "player-key-castle", CVAR_DELIM },
#endif
    };
    for(int i = 0; i < NUM_KEY_TYPES; ++i) {
#if __JHEXEN__
        int ownedKeys = (player->keys & (1 << i))? 1 : 0;
#else
        int ownedKeys = player->keys[i];
#endif
        Con_SetVariable(keyIds[i], ownedKeys, SVF_WRITE_OVERRIDE);
    }
    Con_SetVariable(var_player_weapon_current, player->readyWeapon, SVF_WRITE_OVERRIDE);
    static Path const weaponIds[NUM_WEAPON_TYPES] = {
#if __JDOOM__ || __JDOOM64__
        /* WT_FIRST */   { "player-weapon-fist", CVAR_DELIM },
        /* WT_SECOND */  { "player-weapon-pistol", CVAR_DELIM },
        /* WT_THIRD */   { "player-weapon-shotgun", CVAR_DELIM },
        /* WT_FOURTH */  { "player-weapon-chaingun", CVAR_DELIM },
        /* WT_FIFTH */   { "player-weapon-mlauncher", CVAR_DELIM },
        /* WT_SIXTH */   { "player-weapon-plasmarifle", CVAR_DELIM },
        /* WT_SEVENTH */ { "player-weapon-bfg", CVAR_DELIM },
        /* WT_EIGHTH */  { "player-weapon-chainsaw", CVAR_DELIM },
        /* WT_NINETH */  { "player-weapon-sshotgun", CVAR_DELIM },
    #if __JDOOM64__
        /* WT_TENTH */   { "player-weapon-unmaker", CVAR_DELIM },
    #endif
#elif __JHERETIC__
        /* WT_FIRST */   { "player-weapon-staff", CVAR_DELIM },
        /* WT_SECOND */  { "player-weapon-goldwand", CVAR_DELIM },
        /* WT_THIRD */   { "player-weapon-crossbow", CVAR_DELIM },
        /* WT_FOURTH */  { "player-weapon-dragonclaw", CVAR_DELIM },
        /* WT_FIFTH */   { "player-weapon-hellstaff", CVAR_DELIM },
        /* WT_SIXTH */   { "player-weapon-phoenixrod", CVAR_DELIM },
        /* WT_SEVENTH */ { "player-weapon-mace", CVAR_DELIM },
        /* WT_EIGHTH */  { "player-weapon-gauntlets", CVAR_DELIM },
#elif __JHEXEN__
        /* WT_FIRST */   { "player-weapon-first", CVAR_DELIM },
        /* WT_SECOND */  { "player-weapon-second", CVAR_DELIM },
        /* WT_THIRD */   { "player-weapon-third", CVAR_DELIM },
        /* WT_FOURTH*/   { "player-weapon-fourth", CVAR_DELIM },
#endif
    };
    for(int i = 0; i < NUM_WEAPON_TYPES; ++i) {
        Con_SetVariable(weaponIds[i], player->weapons[i].owned, SVF_WRITE_OVERRIDE);
    }
#if __JHEXEN__
    for(int i = 0; i < WEAPON_FOURTH_PIECE_COUNT; ++i) {
        Path const varPath { String("player-weapon-piece%1").arg(i + 1), CVAR_DELIM };
        Con_SetVariable(varPath, (player->pieces & (1 << i))? 1 : 0, SVF_WRITE_OVERRIDE);
    }
    static Path const var_player_weapon_allpieces("player-weapon-allpieces", CVAR_DELIM);
    Con_SetVariable(var_player_weapon_allpieces, (player->pieces == WEAPON_FOURTH_COMPLETE)? 1 : 0, SVF_WRITE_OVERRIDE);
#endif
    static Path const ammoIds[NUM_AMMO_TYPES] = {
#if __JDOOM__ || __JDOOM64__
        /* AT_CLIP */      { "player-ammo-bullets", CVAR_DELIM },
        /* AT_SHELL */     { "player-ammo-shells", CVAR_DELIM },
        /* AT_CELL */      { "player-ammo-cells", CVAR_DELIM },
        /* AT_MISSILE */   { "player-ammo-missiles", CVAR_DELIM },
#elif __JHERETIC__
        /* AT_CRYSTAL */   { "player-ammo-goldwand", CVAR_DELIM },
        /* AT_ARROW */     { "player-ammo-crossbow", CVAR_DELIM },
        /* AT_ORB */       { "player-ammo-dragonclaw", CVAR_DELIM },
        /* AT_RUNE */      { "player-ammo-hellstaff", CVAR_DELIM },
        /* AT_FIREORB */   { "player-ammo-phoenixrod", CVAR_DELIM },
        /* AT_MSPHERE */   { "player-ammo-mace", CVAR_DELIM },
#elif __JHEXEN__
        /* AT_BLUEMANA */  { "player-ammo-bluemana", CVAR_DELIM },
        /* AT_GREENMANA */ { "player-ammo-greenmana", CVAR_DELIM },
#endif
    };
    for(int i = 0; i < NUM_AMMO_TYPES; ++i) {
        Con_SetVariable(ammoIds[i], player->ammo[i].owned, SVF_WRITE_OVERRIDE);
    }
#if __JHERETIC__ || __JHEXEN__ || __JDOOM64__
    static Path const invItemIds[NUM_INVENTORYITEM_TYPES] = {
#  if __JDOOM64__
        /* IIT_DEMONKEY1 */         { "player-artifact-bluedemonkey", CVAR_DELIM },
        /* IIT_DEMONKEY2 */         { "player-artifact-yellowdemonkey", CVAR_DELIM },
        /* IIT_DEMONKEY3 */         { "player-artifact-reddemonkey", CVAR_DELIM },
#  elif __JHERETIC__
        /* IIT_INVULNERABILITY */   { "player-artifact-ring", CVAR_DELIM },
        /* IIT_INVISIBILITY */      { "player-artifact-shadowsphere", CVAR_DELIM },
        /* IIT_HEALTH */            { "player-artifact-crystalvial", CVAR_DELIM },
        /* IIT_SUPERHEALTH */       { "player-artifact-mysticurn", CVAR_DELIM },
        /* IIT_TOMBOFPOWER */       { "player-artifact-tomeofpower", CVAR_DELIM },
        /* IIT_TORCH */             { "player-artifact-torch", CVAR_DELIM },
        /* IIT_FIREBOMB */          { "player-artifact-firebomb", CVAR_DELIM },
        /* IIT_EGG */               { "player-artifact-egg", CVAR_DELIM },
        /* IIT_FLY */               { "player-artifact-wings", CVAR_DELIM },
        /* IIT_TELEPORT */          { "player-artifact-chaosdevice", CVAR_DELIM },
#  elif __JHEXEN__
        /* IIT_INVULNERABILITY */   { "player-artifact-defender", CVAR_DELIM },
        /* IIT_HEALTH */            { "player-artifact-quartzflask", CVAR_DELIM },
        /* IIT_SUPERHEALTH */       { "player-artifact-mysticurn", CVAR_DELIM },
        /* IIT_HEALINGRADIUS */     { "player-artifact-mysticambit", CVAR_DELIM },
        /* IIT_SUMMON */            { "player-artifact-darkservant", CVAR_DELIM },
        /* IIT_TORCH */             { "player-artifact-torch", CVAR_DELIM },
        /* IIT_EGG */               { "player-artifact-porkalator", CVAR_DELIM },
        /* IIT_FLY */               { "player-artifact-wings", CVAR_DELIM },
        /* IIT_BLASTRADIUS */       { "player-artifact-repulsion", CVAR_DELIM },
        /* IIT_POISONBAG */         { "player-artifact-flechette", CVAR_DELIM },
        /* IIT_TELEPORTOTHER */     { "player-artifact-banishment", CVAR_DELIM },
        /* IIT_SPEED */             { "player-artifact-speed", CVAR_DELIM },
        /* IIT_BOOSTMANA */         { "player-artifact-might", CVAR_DELIM },
        /* IIT_BOOSTARMOR */        { "player-artifact-bracers", CVAR_DELIM },
        /* IIT_TELEPORT */          { "player-artifact-chaosdevice", CVAR_DELIM },
        /* IIT_PUZZSKULL */         { "player-artifact-skull", CVAR_DELIM },
        /* IIT_PUZZGEMBIG */        { "player-artifact-heart", CVAR_DELIM },
        /* IIT_PUZZGEMRED */        { "player-artifact-ruby", CVAR_DELIM },
        /* IIT_PUZZGEMGREEN1 */     { "player-artifact-emerald1", CVAR_DELIM },
        /* IIT_PUZZGEMGREEN2 */     { "player-artifact-emerald2", CVAR_DELIM },
        /* IIT_PUZZGEMBLUE1 */      { "player-artifact-sapphire1", CVAR_DELIM },
        /* IIT_PUZZGEMBLUE2 */      { "player-artifact-sapphire2", CVAR_DELIM },
        /* IIT_PUZZBOOK1 */         { "player-artifact-daemoncodex", CVAR_DELIM },
        /* IIT_PUZZBOOK2 */         { "player-artifact-liberoscura", CVAR_DELIM },
        /* IIT_PUZZSKULL2 */        { "player-artifact-flamemask", CVAR_DELIM },
        /* IIT_PUZZFWEAPON */       { "player-artifact-glaiveseal", CVAR_DELIM },
        /* IIT_PUZZCWEAPON */       { "player-artifact-holyrelic", CVAR_DELIM },
        /* IIT_PUZZMWEAPON */       { "player-artifact-sigilmagus", CVAR_DELIM },
        /* IIT_PUZZGEAR1 */         { "player-artifact-gear1", CVAR_DELIM },
        /* IIT_PUZZGEAR2 */         { "player-artifact-gear2", CVAR_DELIM },
        /* IIT_PUZZGEAR3 */         { "player-artifact-gear3", CVAR_DELIM },
        /* IIT_PUZZGEAR4 */         { "player-artifact-gear4", CVAR_DELIM },
#  endif
    };
    int const plrNum = player - players;
    for(int i = IIT_FIRST; i < NUM_INVENTORYITEM_TYPES; ++i) {
        int numItems = 0;
        if(player->plr->inGame && G_GameState() == GS_MAP) {
            numItems = P_InventoryCount(plrNum, inventoryitemtype_t(i));
        }
        Con_SetVariable(invItemIds[i - 1], numItems, SVF_WRITE_OVERRIDE);
    }
#endif
}
static int findMirrorPolyobj(int tag) {
#if __JHEXEN__
    for(int i = 0; i < numpolyobjs; ++i) {
        Polyobj *po = Polyobj_ById(i);
        if(po->tag == tag) {
            return P_ToXLine(Polyobj_FirstLine(po))->arg2;
        }
    }
#else
    DENG_UNUSED(tag);
#endif
    return 0;
}
static void startSoundSequence(Polyobj *poEmitter) {
#if __JHEXEN__
    if (poEmitter) {
        SN_StartSequence((mobj_t *)poEmitter, SEQ_DOOR_STONE + poEmitter->seqType);
    }
#else
    DENG_UNUSED(poEmitter);
#endif
}
static void stopSoundSequence(Polyobj *poEmitter) {
#if __JHEXEN__
    SN_StopSequence((mobj_t *)poEmitter);
#else
    DENG_UNUSED(poEmitter);
#endif
}
static void PO_SetDestination(Polyobj *po, coord_t dist, uint fineAngle, float speed) {
    DENG_ASSERT(po != 0);
    DENG_ASSERT(fineAngle < FINEANGLES);
    po->dest[VX] = po->origin[VX] + dist * FIX2FLT(finecosine[fineAngle]);
    po->dest[VY] = po->origin[VY] + dist * FIX2FLT(finesine[fineAngle]);
    po->speed    = speed;
}
static void PODoor_UpdateDestination(polydoor_t *pd) {
    DENG2_ASSERT(pd != 0);
    Polyobj *po = Polyobj_ByTag(pd->polyobj);
    if(pd->type == PODOOR_SLIDE) {
        PO_SetDestination(po, FIX2FLT(pd->dist), pd->direction, FIX2FLT(pd->intSpeed));
    }
}
void T_RotatePoly(void *polyThinker) {
    DENG_ASSERT(polyThinker != 0);
    polyevent_t *pe = (polyevent_t *)polyThinker;
    uint absSpeed;
    Polyobj *po = Polyobj_ByTag(pe->polyobj);
    if(Polyobj_Rotate(po, pe->intSpeed)) {
        absSpeed = abs(pe->intSpeed);
        if(pe->dist == POBJ_PERPETUAL) {
            return;
        }
        pe->dist -= absSpeed;
        if(int(pe->dist) <= 0) {
            if(po->specialData == pe)
                po->specialData = 0;
            stopSoundSequence(po);
            P_NotifyPolyobjFinished(po->tag);
            Thinker_Remove(&pe->thinker);
            po->angleSpeed = 0;
        }
        if(pe->dist < absSpeed) {
            pe->intSpeed = pe->dist * (pe->intSpeed < 0 ? -1 : 1);
        }
    }
}
dd_bool EV_RotatePoly(Line *line, byte *args, int direction, dd_bool overRide) {
    DENG_UNUSED(line);
    DENG_ASSERT(args != 0);
    int tag = args[0];
    Polyobj *po = Polyobj_ByTag(tag);
    if(po) {
        if(po->specialData && !overRide)
        {   // Poly is already moving, so keep going...
            return false;
        }
    }
    else {
        Con_Error("EV_RotatePoly:  Invalid polyobj tag: %d\n", tag);
    }
    polyevent_t *pe = (polyevent_t *)Z_Calloc(sizeof(*pe), PU_MAP, 0);
    pe->thinker.function = T_RotatePoly;
    Thinker_Add(&pe->thinker);
    pe->polyobj = tag;
    if(args[2]) {
        if(args[2] == 255) {
            pe->dist = POBJ_PERPETUAL;
            po->destAngle = POBJ_PERPETUAL;
        }
        else {
            pe->dist = args[2] * (ANGLE_90 / 64);   // Angle
            po->destAngle = po->angle + pe->dist * direction;
        }
    }
    else {
        pe->dist = ANGLE_MAX - 1;
        po->destAngle = po->angle + pe->dist;
    }
    pe->intSpeed = (args[1] * direction * (ANGLE_90 / 64)) >> 3;
    po->specialData = pe;
    po->angleSpeed = pe->intSpeed;
    startSoundSequence(po);
    int mirror; // tag
    while((mirror = findMirrorPolyobj(tag)) != 0) {
        po = Polyobj_ByTag(mirror);
        if(po && po->specialData && !overRide) {
            break;
        }
        pe = (polyevent_t *)Z_Calloc(sizeof(*pe), PU_MAP, 0);
        pe->thinker.function = T_RotatePoly;
        Thinker_Add(&pe->thinker);
        po->specialData = pe;
        pe->polyobj = mirror;
        if(args[2]) {
            if(args[2] == 255) {
                pe->dist = POBJ_PERPETUAL;
                po->destAngle = POBJ_PERPETUAL;
            }
            else {
                pe->dist = args[2] * (ANGLE_90 / 64); // Angle
                po->destAngle = po->angle + pe->dist * -direction;
            }
        }
        else {
            pe->dist = ANGLE_MAX - 1;
            po->destAngle =  po->angle + pe->dist;
        }
        direction = -direction;
        pe->intSpeed = (args[1] * direction * (ANGLE_90 / 64)) >> 3;
        po->angleSpeed = pe->intSpeed;
        po = Polyobj_ByTag(tag);
        if(po) {
            po->specialData = pe;
        }
        else {
            Con_Error("EV_RotatePoly:  Invalid polyobj num: %d\n", tag);
        }
        tag = mirror;
        startSoundSequence(po);
    }
    return true;
}
void T_MovePoly(void *polyThinker) {
    DENG_ASSERT(polyThinker != 0);
    polyevent_t *pe = (polyevent_t *)polyThinker;
    Polyobj *po = Polyobj_ByTag(pe->polyobj);
    if(Polyobj_MoveXY(po, pe->speed[MX], pe->speed[MY])) {
        uint const absSpeed = abs(pe->intSpeed);
        pe->dist -= absSpeed;
        if(int(pe->dist) <= 0) {
            if(po->specialData == pe)
                po->specialData = NULL;
            stopSoundSequence(po);
            P_NotifyPolyobjFinished(po->tag);
            Thinker_Remove(&pe->thinker);
            po->speed = 0;
        }
        if(pe->dist < absSpeed) {
            pe->intSpeed = pe->dist * (pe->intSpeed < 0 ? -1 : 1);
            pe->speed[MX] = FIX2FLT(FixedMul(pe->intSpeed, finecosine[pe->fangle]));
            pe->speed[MY] = FIX2FLT(FixedMul(pe->intSpeed, finesine[pe->fangle]));
        }
    }
}
void polyevent_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteInt32(writer, polyobj);
    Writer_WriteInt32(writer, intSpeed);
    Writer_WriteUInt32(writer, dist);
    Writer_WriteInt32(writer, fangle);
    Writer_WriteInt32(writer, FLT2FIX(speed[VX]));
    Writer_WriteInt32(writer, FLT2FIX(speed[VY]));
}
int polyevent_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
    if(mapVersion >= 4) {
        /*int ver =*/ Reader_ReadByte(reader); // version byte.
        polyobj     = Reader_ReadInt32(reader);
        intSpeed    = Reader_ReadInt32(reader);
        dist        = Reader_ReadUInt32(reader);
        fangle      = Reader_ReadInt32(reader);
        speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
    }
    else {
        byte junk[16]; // sizeof thinker_t
        Reader_Read(reader, junk, 16);
        polyobj     = Reader_ReadInt32(reader);
        intSpeed    = Reader_ReadInt32(reader);
        dist        = Reader_ReadUInt32(reader);
        fangle      = Reader_ReadInt32(reader);
        speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
    }
    thinker.function = T_RotatePoly;
    return true; // Add this thinker.
}
void SV_WriteMovePoly(polyevent_t const *th, MapStateWriter *msw) {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteInt32(writer, th->polyobj);
    Writer_WriteInt32(writer, th->intSpeed);
    Writer_WriteUInt32(writer, th->dist);
    Writer_WriteInt32(writer, th->fangle);
    Writer_WriteInt32(writer, FLT2FIX(th->speed[VX]));
    Writer_WriteInt32(writer, FLT2FIX(th->speed[VY]));
}
int SV_ReadMovePoly(polyevent_t *th, MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
    if(mapVersion >= 4) {
        /*int ver =*/ Reader_ReadByte(reader); // version byte.
        th->polyobj     = Reader_ReadInt32(reader);
        th->intSpeed    = Reader_ReadInt32(reader);
        th->dist        = Reader_ReadUInt32(reader);
        th->fangle      = Reader_ReadInt32(reader);
        th->speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        th->speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
    }
    else {
        byte junk[16]; // sizeof thinker_t
        Reader_Read(reader, junk, 16);
        th->polyobj     = Reader_ReadInt32(reader);
        th->intSpeed    = Reader_ReadInt32(reader);
        th->dist        = Reader_ReadUInt32(reader);
        th->fangle      = Reader_ReadInt32(reader);
        th->speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        th->speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
    }
    th->thinker.function = T_MovePoly;
    return true; // Add this thinker.
}
dd_bool EV_MovePoly(Line *line, byte *args, dd_bool timesEight, dd_bool override) {
    DENG_UNUSED(line);
    DENG_ASSERT(args != 0);
    int tag = args[0];
    Polyobj *po = Polyobj_ByTag(tag);
    DENG_ASSERT(po != 0);
    if(po->specialData && !override)
        return false;
    polyevent_t *pe = (polyevent_t *)Z_Calloc(sizeof(*pe), PU_MAP, 0);
    pe->thinker.function = T_MovePoly;
    Thinker_Add(&pe->thinker);
    pe->polyobj = tag;
    if(timesEight) {
        pe->dist = args[3] * 8 * FRACUNIT;
    }
    else {
        pe->dist = args[3] * FRACUNIT;  // Distance
    }
    pe->intSpeed = args[1] * (FRACUNIT / 8);
    po->specialData = pe;
    angle_t angle = args[2] * (ANGLE_90 / 64);
    pe->fangle = angle >> ANGLETOFINESHIFT;
    pe->speed[MX] = FIX2FLT(FixedMul(pe->intSpeed, finecosine[pe->fangle]));
    pe->speed[MY] = FIX2FLT(FixedMul(pe->intSpeed, finesine[pe->fangle]));
    startSoundSequence(po);
    PO_SetDestination(po, FIX2FLT(pe->dist), pe->fangle, FIX2FLT(pe->intSpeed));
    int mirror; // tag
    while((mirror = findMirrorPolyobj(tag)) != 0) {
        po = Polyobj_ByTag(mirror);
        if(po && po->specialData && !override)
            break;
        pe = (polyevent_t *)Z_Calloc(sizeof(*pe), PU_MAP, 0);
        pe->thinker.function = T_MovePoly;
        Thinker_Add(&pe->thinker);
        pe->polyobj = mirror;
        po->specialData = pe;
        if(timesEight) {
            pe->dist = args[3] * 8 * FRACUNIT;
        }
        else {
            pe->dist = args[3] * FRACUNIT; // Distance
        }
        pe->intSpeed = args[1] * (FRACUNIT / 8);
        angle = angle + ANGLE_180; // reverse the angle
        pe->fangle = angle >> ANGLETOFINESHIFT;
        pe->speed[MX] = FIX2FLT(FixedMul(pe->intSpeed, finecosine[pe->fangle]));
        pe->speed[MY] = FIX2FLT(FixedMul(pe->intSpeed, finesine[pe->fangle]));
        tag = mirror;
        startSoundSequence(po);
        PO_SetDestination(po, FIX2FLT(pe->dist), pe->fangle, FIX2FLT(pe->intSpeed));
    }
    return true;
}
void T_PolyDoor(void *polyDoorThinker) {
    DENG_ASSERT(polyDoorThinker != 0);
    polydoor_t *pd = (polydoor_t *)polyDoorThinker;
    Polyobj *po = Polyobj_ByTag(pd->polyobj);
    if(pd->tics) {
        if(!--pd->tics) {
            startSoundSequence(po);
            PODoor_UpdateDestination(pd);
        }
        return;
    }
    switch(pd->type) {
    case PODOOR_SLIDE:
        if(Polyobj_MoveXY(po, pd->speed[MX], pd->speed[MY])) {
            int absSpeed = abs(pd->intSpeed);
            pd->dist -= absSpeed;
            if(pd->dist <= 0) {
                stopSoundSequence(po);
                if(!pd->close) {
                    pd->dist = pd->totalDist;
                    pd->close = true;
                    pd->tics = pd->waitTics;
                    pd->direction = (ANGLE_MAX >> ANGLETOFINESHIFT) - pd->direction;
                    pd->speed[MX] = -pd->speed[MX];
                    pd->speed[MY] = -pd->speed[MY];
                }
                else {
                    if(po->specialData == pd)
                        po->specialData = NULL;
                    P_NotifyPolyobjFinished(po->tag);
                    Thinker_Remove(&pd->thinker);
                }
            }
        }
        else {
            if(po->crush || !pd->close) {
                return;
            }
            else {
                pd->dist = pd->totalDist - pd->dist;
                pd->direction = (ANGLE_MAX >> ANGLETOFINESHIFT) - pd->direction;
                pd->speed[MX] = -pd->speed[MX];
                pd->speed[MY] = -pd->speed[MY];
                PODoor_UpdateDestination(pd);
                pd->close = false;
                startSoundSequence(po);
            }
        }
        break;
    case PODOOR_SWING:
        if(Polyobj_Rotate(po, pd->intSpeed)) {
            int absSpeed = abs(pd->intSpeed);
            if(pd->dist == -1) {
                return;
            }
            pd->dist -= absSpeed;
            if(pd->dist <= 0) {
                stopSoundSequence(po);
                if(!pd->close) {
                    pd->dist = pd->totalDist;
                    pd->close = true;
                    pd->tics = pd->waitTics;
                    pd->intSpeed = -pd->intSpeed;
                }
                else {
                    if(po->specialData == pd)
                        po->specialData = NULL;
                    P_NotifyPolyobjFinished(po->tag);
                    Thinker_Remove(&pd->thinker);
                }
            }
        }
        else {
            if(po->crush || !pd->close) {
                return;
            }
            else {
                pd->dist = pd->totalDist - pd->dist;
                pd->intSpeed = -pd->intSpeed;
                pd->close = false;
                startSoundSequence(po);
            }
        }
        break;
    default:
        break;
    }
}
void polydoor_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // Write a version byte.
    Writer_WriteByte(writer, type);
    Writer_WriteInt32(writer, polyobj);
    Writer_WriteInt32(writer, intSpeed);
    Writer_WriteInt32(writer, dist);
    Writer_WriteInt32(writer, totalDist);
    Writer_WriteInt32(writer, direction);
    Writer_WriteInt32(writer, FLT2FIX(speed[VX]));
    Writer_WriteInt32(writer, FLT2FIX(speed[VY]));
    Writer_WriteInt32(writer, tics);
    Writer_WriteInt32(writer, waitTics);
    Writer_WriteByte(writer, close);
}
int polydoor_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    int mapVersion = msr->mapVersion();
    if(mapVersion >= 4) {
        /*int ver =*/ Reader_ReadByte(reader); // version byte.
        type        = podoortype_t(Reader_ReadByte(reader));
        polyobj     = Reader_ReadInt32(reader);
        intSpeed    = Reader_ReadInt32(reader);
        dist        = Reader_ReadInt32(reader);
        totalDist   = Reader_ReadInt32(reader);
        direction   = Reader_ReadInt32(reader);
        speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
        tics        = Reader_ReadInt32(reader);
        waitTics    = Reader_ReadInt32(reader);
        close       = Reader_ReadByte(reader);
    }
    else {
        byte junk[16]; // sizeof thinker_t
        Reader_Read(reader, junk, 16);
        polyobj     = Reader_ReadInt32(reader);
        intSpeed    = Reader_ReadInt32(reader);
        dist        = Reader_ReadInt32(reader);
        totalDist   = Reader_ReadInt32(reader);
        direction   = Reader_ReadInt32(reader);
        speed[VX]   = FIX2FLT(Reader_ReadInt32(reader));
        speed[VY]   = FIX2FLT(Reader_ReadInt32(reader));
        tics        = Reader_ReadInt32(reader);
        waitTics    = Reader_ReadInt32(reader);
        type        = podoortype_t(Reader_ReadByte(reader));
        close       = Reader_ReadByte(reader);
    }
    thinker.function = T_PolyDoor;
    return true; // Add this thinker.
}
dd_bool EV_OpenPolyDoor(Line *line, byte *args, podoortype_t type) {
    DENG_UNUSED(line);
    DENG_ASSERT(args != 0);
    int tag = args[0];
    Polyobj *po = Polyobj_ByTag(tag);
    if(po) {
        if(po->specialData)
        {   // Is already moving.
            return false;
        }
    }
    else {
        Con_Error("EV_OpenPolyDoor:  Invalid polyobj num: %d\n", tag);
    }
    polydoor_t *pd = (polydoor_t *)Z_Calloc(sizeof(*pd), PU_MAP, 0);
    pd->thinker.function = T_PolyDoor;
    Thinker_Add(&pd->thinker);
    angle_t angle = 0;
    pd->type = type;
    pd->polyobj = tag;
    if(type == PODOOR_SLIDE) {
        pd->waitTics = args[4];
        pd->intSpeed = args[1] * (FRACUNIT / 8);
        pd->totalDist = args[3] * FRACUNIT; // Distance.
        pd->dist = pd->totalDist;
        angle = args[2] * (ANGLE_90 / 64);
        pd->direction = angle >> ANGLETOFINESHIFT;
        pd->speed[MX] = FIX2FLT(FixedMul(pd->intSpeed, finecosine[pd->direction]));
        pd->speed[MY] = FIX2FLT(FixedMul(pd->intSpeed, finesine[pd->direction]));
        startSoundSequence(po);
    }
    else if(type == PODOOR_SWING) {
        pd->waitTics = args[3];
        pd->direction = 1;
        pd->intSpeed = (args[1] * pd->direction * (ANGLE_90 / 64)) >> 3;
        pd->totalDist = args[2] * (ANGLE_90 / 64);
        pd->dist = pd->totalDist;
        startSoundSequence(po);
    }
    po->specialData = pd;
    PODoor_UpdateDestination(pd);
    int mirror; // tag
    while((mirror = findMirrorPolyobj(tag)) != 0) {
        po = Polyobj_ByTag(mirror);
        if(po && po->specialData) {
            break;
        }
        pd = (polydoor_t *)Z_Calloc(sizeof(*pd), PU_MAP, 0);
        pd->thinker.function = T_PolyDoor;
        Thinker_Add(&pd->thinker);
        pd->polyobj = mirror;
        pd->type = type;
        po->specialData = pd;
        if(type == PODOOR_SLIDE) {
            pd->waitTics = args[4];
            pd->intSpeed = args[1] * (FRACUNIT / 8);
            pd->totalDist = args[3] * FRACUNIT; // Distance.
            pd->dist = pd->totalDist;
            angle = angle + ANGLE_180; // Reverse the angle.
            pd->direction = angle >> ANGLETOFINESHIFT;
            pd->speed[MX] = FIX2FLT(FixedMul(pd->intSpeed, finecosine[pd->direction]));
            pd->speed[MY] = FIX2FLT(FixedMul(pd->intSpeed, finesine[pd->direction]));
            startSoundSequence(po);
        }
        else if(type == PODOOR_SWING) {
            pd->waitTics = args[3];
            pd->direction = -1;
            pd->intSpeed = (args[1] * pd->direction * (ANGLE_90 / 64)) >> 3;
            pd->totalDist = args[2] * (ANGLE_90 / 64);
            pd->dist = pd->totalDist;
            startSoundSequence(po);
        }
        tag = mirror;
        PODoor_UpdateDestination(pd);
    }
    return true;
}
static void thrustMobj(struct mobj_s *mo, void *linep, void *pop) {
    Line *line = (Line *) linep;
    Polyobj *po = (Polyobj *) pop;
    coord_t thrust[2], force;
    polyevent_t *pe;
    uint thrustAn;
    if(IS_CLIENT) return;
    if(P_MobjIsCamera(mo)) return;
    if(!(mo->flags & MF_SHOOTABLE) && !mo->player) return;
    thrustAn = (P_GetAnglep(line, DMU_ANGLE) - ANGLE_90) >> ANGLETOFINESHIFT;
    pe = (polyevent_t *) po->specialData;
    if(pe) {
        if(pe->thinker.function == (thinkfunc_t) T_RotatePoly) {
            force = FIX2FLT(pe->intSpeed >> 8);
        }
        else {
            force = FIX2FLT(pe->intSpeed >> 3);
        }
        force = MINMAX_OF(1, force, 4);
    }
    else {
        force = 1;
    }
    thrust[VX] = force * FIX2FLT(finecosine[thrustAn]);
    thrust[VY] = force * FIX2FLT(finesine[thrustAn]);
    mo->mom[MX] += thrust[VX];
    mo->mom[MY] += thrust[VY];
    if(po->crush) {
        if(!P_CheckPositionXY(mo, mo->origin[VX] + thrust[VX], mo->origin[VY] + thrust[VY])) {
            P_DamageMobj(mo, NULL, NULL, 3, false);
        }
    }
}
void PO_InitForMap() {
#if !defined(__JHEXEN__)
    return; // Disabled -- awaiting line argument translation.
#endif
    App_Log(DE2_DEV_MAP_VERBOSE, "Initializing polyobjects for map...");
    Polyobj_SetCallback(thrustMobj);
    for(int i = 0; i < numpolyobjs; ++i) {
        Polyobj *po = Polyobj_ById(i);
        po->specialData = NULL;
        uint j = 0;
        mapspot_t const *spot = 0;
        while(j < numMapSpots && !spot) {
            if((mapSpots[j].doomEdNum == PO_SPAWN_DOOMEDNUM ||
                mapSpots[j].doomEdNum == PO_SPAWNCRUSH_DOOMEDNUM) &&
                mapSpots[j].angle     == angle_t(po->tag)) {
                spot = &mapSpots[j];
            }
            else {
                j++;
            }
        }
        if(spot) {
            po->crush = (spot->doomEdNum == PO_SPAWNCRUSH_DOOMEDNUM? 1 : 0);
            Polyobj_MoveXY(po, -po->origin[VX] + spot->origin[VX],
                               -po->origin[VY] + spot->origin[VY]);
        }
        else {
            App_Log(DE2_MAP_WARNING, "Missing spawn spot for PolyObj #%i", i);
        }
    }
}
dd_bool PO_Busy(int tag) {
    Polyobj *po = Polyobj_ByTag(tag);
    return (po && po->specialData);
}
void polyobj_s::write(MapStateWriter *msw) const {
    Writer1 *writer = msw->writer();
    Writer_WriteByte(writer, 1); // write a version byte (unused).
    Writer_WriteInt32(writer, tag);
    Writer_WriteInt32(writer, angle);
    Writer_WriteInt32(writer, FLT2FIX(origin[VX]));
    Writer_WriteInt32(writer, FLT2FIX(origin[VY]));
}
int polyobj_s::read(MapStateReader *msr) {
    Reader1 *reader = msr->reader();
    angle_t newAngle = angle_t(Reader_ReadInt32(reader));
    Polyobj_Rotate(this, newAngle);
    destAngle = newAngle;
    coord_t newOrigin[2];
    newOrigin[0] = FIX2FLT(Reader_ReadInt32(reader));
    newOrigin[1] = FIX2FLT(Reader_ReadInt32(reader));
    Polyobj_MoveXY(this, newOrigin[0] - origin[0], newOrigin[1] - origin[1]);
    return true;
}
DENG2_PIMPL(ThingArchive) {
    int version;
    uint size;
    mobj_t const **things;
    bool excludePlayers;
    Impl(Public *i)
        : Base(i)
        , version(0)
        , size(0)
        , things(0)
        , excludePlayers(false)
    {}
    ~Impl() {
        self().clear();
    }
    struct countmobjthinkerstoarchive_params_t {
        uint count;
        bool excludePlayers;
    };
    static int countMobjThinkersToArchive(thinker_t *th, void *context) {
        countmobjthinkerstoarchive_params_t &p = *(countmobjthinkerstoarchive_params_t *) context;
        if (!(Mobj_IsPlayer(reinterpret_cast<mobj_t const *>(th)) && p.excludePlayers)) {
            p.count++;
        }
        return false; // Continue iteration.
    }
};
ThingArchive::ThingArchive(int version) : d(new Impl(this)) {
    d->version = version;
}
int ThingArchive::version() const {
    return d->version;
}
bool ThingArchive::excludePlayers() const {
    return d->excludePlayers;
}
uint ThingArchive::size() const {
    return d->size;
}
void ThingArchive::clear() {
    M_Free(d->things); d->things = 0;
    d->size = 0;
}
void ThingArchive::initForLoad(uint size) {
    d->size   = size;
    d->things = reinterpret_cast<mobj_t const **>(M_Calloc(d->size * sizeof(*d->things)));
}
void ThingArchive::initForSave(bool excludePlayers) {
    Impl::countmobjthinkerstoarchive_params_t parm; de::zap(parm);
    parm.count          = 0;
    parm.excludePlayers = excludePlayers;
    Thinker_Iterate(P_MobjThinker, Impl::countMobjThinkersToArchive, &parm);
    d->size           = parm.count;
    d->things         = reinterpret_cast<mobj_t const **>(M_Calloc(d->size * sizeof(*d->things)));
    d->excludePlayers = excludePlayers;
}
void ThingArchive::insert(mobj_t const *mo, SerialId serialId) {
    DENG_ASSERT(mo != 0);
#if __JHEXEN__
    if (d->version >= 1)
#endif
    {
        serialId -= 1;
    }
#if __JHEXEN__
    DENG2_ASSERT(serialId >= 0);
    if (serialId < 0) return; // Does this ever occur?
#endif
    DENG_ASSERT(d->things != 0);
    DENG_ASSERT((unsigned)serialId < d->size);
    d->things[serialId] = const_cast<mobj_t *>(mo);
}
ThingArchive::SerialId ThingArchive::serialIdFor(mobj_t const *mo) {
    DENG_ASSERT(d->things != 0);
    if (!mo) return 0;
    if (mo->thinker.function != P_MobjThinker) {
        return 0;
    }
#if __JHEXEN__
    if (mo->player && d->excludePlayers) {
        return TargetPlayerId;
    }
#endif
    uint firstUnused = 0;
    bool found = false;
    for (uint i = 0; i < d->size; ++i) {
        if (!d->things[i] && !found) {
            firstUnused = i;
            found = true;
            continue;
        }
        if (d->things[i] == mo) {
            return i + 1;
        }
    }
    if (!found) {
        Con_Error("ThingArchive::serialIdFor: Thing archive exhausted!");
        return 0; // No number available!
    }
    d->things[firstUnused] = mo;
    return firstUnused + 1;
}
mobj_t *ThingArchive::mobj(SerialId serialId, void *address) {
#if !__JHEXEN__
    DENG_UNUSED(address);
#endif
#if __JHEXEN__
    if (serialId == TargetPlayerId) {
        targetplraddress_t *tpa = (targetplraddress_t *)M_Malloc(sizeof(targetplraddress_t));
        tpa->address = (void **)address;
        tpa->next = targetPlayerAddrs;
        targetPlayerAddrs = tpa;
        return 0;
    }
#endif
    DENG_ASSERT(d->things != 0);
#if __JHEXEN__
    if (d->version < 1) {
        if (serialId == -1) return 0;
        if (serialId < 0 || (unsigned) serialId > d->size - 1)
            return 0;
    }
    else
#endif
    {
        if (serialId == 0) return 0;
        if (serialId < 1 || unsigned(serialId) > d->size) {
            App_Log(DE2_RES_WARNING, "ThingArchive::mobj: Invalid serialId %i", serialId);
            return 0;
        }
        serialId -= 1;
    }
    return const_cast<mobj_t *>(d->things[serialId]);
}
template <typename Type>
static void writeThinkerAs(thinker_t const *th, MapStateWriter *msWriter) {
    Type *t = (Type*)th;
    t->write(msWriter);
}
template <typename Type>
static int readThinkerAs(thinker_t *th, MapStateReader *msReader) {
    Type *t = (Type *)th;
    return t->read(msReader);
}
static ThinkerClassInfo thinkerInfo[] = { {
      TC_MOBJ,
      (thinkfunc_t) P_MobjThinker,
      TSF_SERVERONLY,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<mobj_s>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<mobj_s>),
      sizeof(mobj_t)
    },
#if !__JHEXEN__
    {
      TC_XGMOVER,
      (thinkfunc_t) XS_PlaneMover,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<xgplanemover_s>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<xgplanemover_s>),
      sizeof(xgplanemover_t)
    },
#endif
    {
      TC_CEILING,
      T_MoveCeiling,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<ceiling_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<ceiling_t>),
      sizeof(ceiling_t)
    }, {
      TC_DOOR,
      T_Door,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<door_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<door_t>),
      sizeof(door_t)
    }, {
      TC_FLOOR,
      T_MoveFloor,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<floor_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<floor_t>),
      sizeof(floor_t)
    }, {
      TC_PLAT,
      T_PlatRaise,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<plat_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<plat_t>),
      sizeof(plat_t)
    },
#if __JHEXEN__
    {
     TC_INTERPRET_ACS,
     (thinkfunc_t) acs_Interpreter_Think,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<acs::Interpreter>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<acs::Interpreter>),
     sizeof(acs::Interpreter)
    }, {
     TC_FLOOR_WAGGLE,
     (thinkfunc_t) T_FloorWaggle,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<waggle_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<waggle_t>),
     sizeof(waggle_t)
    }, {
     TC_LIGHT,
     (thinkfunc_t) T_Light,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<light_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<light_t>),
     sizeof(light_t)
    }, {
     TC_PHASE,
     (thinkfunc_t) T_Phase,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<phase_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<phase_t>),
     sizeof(phase_t)
    }, {
     TC_BUILD_PILLAR,
     (thinkfunc_t) T_BuildPillar,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<pillar_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<pillar_t>),
     sizeof(pillar_t)
    }, {
     TC_ROTATE_POLY,
     T_RotatePoly,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<polyevent_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<polyevent_t>),
     sizeof(polyevent_t)
    }, {
     TC_MOVE_POLY,
     T_MovePoly,
     0,
     de::function_cast<WriteThinkerFunc>(SV_WriteMovePoly),
     de::function_cast<ReadThinkerFunc>(SV_ReadMovePoly),
     sizeof(polyevent_t)
    }, {
     TC_POLY_DOOR,
     T_PolyDoor,
     0,
     de::function_cast<WriteThinkerFunc>(writeThinkerAs<polydoor_t>),
     de::function_cast<ReadThinkerFunc>(readThinkerAs<polydoor_t>),
     sizeof(polydoor_t)
    },
#else
    {
      TC_FLASH,
      (thinkfunc_t) T_LightFlash,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<lightflash_s>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<lightflash_s>),
      sizeof(lightflash_s)
    }, {
      TC_STROBE,
      (thinkfunc_t) T_StrobeFlash,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<strobe_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<strobe_t>),
      sizeof(strobe_t)
    }, {
      TC_GLOW,
      (thinkfunc_t) T_Glow,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<glow_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<glow_t>),
      sizeof(glow_t)
    },
# if __JDOOM__ || __JDOOM64__
    {
      TC_FLICKER,
      (thinkfunc_t) T_FireFlicker,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<fireflicker_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<fireflicker_t>),
      sizeof(fireflicker_t)
    },
# endif
# if __JDOOM64__
    {
      TC_BLINK,
      (thinkfunc_t) T_LightBlink,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<lightblink_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<lightblink_t>),
      sizeof(lightblink_t)
    },
# endif
#endif
    {
      TC_MATERIALCHANGER,
      T_MaterialChanger,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<materialchanger_s>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<materialchanger_s>),
      sizeof(materialchanger_s)
    }, {
      TC_SCROLL,
      (thinkfunc_t) T_Scroll,
      0,
      de::function_cast<WriteThinkerFunc>(writeThinkerAs<scroll_t>),
      de::function_cast<ReadThinkerFunc>(readThinkerAs<scroll_t>),
      sizeof(scroll_t)
    },
    { TC_NULL, NULL, 0, NULL, NULL, 0 }
};
ThinkerClassInfo *SV_ThinkerInfoForClass(thinkerclass_t tClass) {
    for(ThinkerClassInfo *info = thinkerInfo; info->thinkclass != TC_NULL; info++) {
        if(info->thinkclass == tClass)
            return info;
    }
    return 0; // Not found.
}
ThinkerClassInfo *SV_ThinkerInfo(thinker_t const &thinker) {
    for(ThinkerClassInfo *info = thinkerInfo; info->thinkclass != TC_NULL; info++) {
        if(info->function == thinker.function)
            return info;
    }
    return 0; // Not found.
}
