﻿/** @file auxbuttonwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/AuxButtonWidget"
/** @file blurwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/BlurWidget"
/** @file buttonwidget.cpp  Clickable button widget.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ButtonWidget"
#include "de/GuiRootWidget"
#include "de/CallbackAction"
#include <de/FocusWidget>
#include <de/MouseEvent>
#include <de/Animation>
/** @file widgets/choicewidget.cpp  Widget for choosing from a set of alternatives.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ChoiceWidget"
#include "de/PopupMenuWidget"
#include "de/SignalAction"
/** @file commandwidget.h  Abstract command line based widget.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/CommandWidget"
#include "de/DocumentPopupWidget"
#include "de/PersistentState"
#include "de/Style"
#include <de/shell/EditorHistory>
#include <de/KeyEvent>
/** @file compositorwidget.cpp  Off-screen compositor.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/CompositorWidget"
#include <de/Drawable>
/** @file widgets/dialogwidget.cpp  Popup dialog.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DialogWidget"
#include "de/ToggleWidget"
#include "de/ChoiceWidget"
#include "de/GuiRootWidget"
#include "de/SignalAction"
#include "de/DialogContentStylist"
#include "de/ui/FilteredData"
#include <de/KeyEvent>
#include <de/MouseEvent>
#include <de/Untrapper>
#include <QEventLoop>
/** @file directoryarraywidget.cpp  Widget for an array of native directories.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DirectoryArrayWidget"
#include "de/BaseGuiApp"
#include "de/BaseWindow"
#include <de/Config>
#include <de/Garbage>
#include <de/NativePath>
#include <de/TextValue>
#include <de/ToggleWidget>
#include <QFileDialog>
/** @file documentpopupwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DocumentPopupWidget"
#include "de/ButtonWidget"
/** @file documentwidget.cpp  Widget for displaying large amounts of text.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DocumentWidget"
#include "de/ProgressWidget"
#include "de/TextDrawable"
#include <de/Drawable>
/** @file fadetoblackwidget.cpp  Fade to/from black.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FadeToBlackWidget"
/** @file focuswidget.cpp  Input focus indicator.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FocusWidget"
#include <QTimer>
/** @file foldpanelwidget.cpp  Folding panel.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FoldPanelWidget"
#include "de/ProceduralImage"
#include "de/DialogContentStylist"
#include "de/SignalAction"
/** @file gridpopupwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/GridPopupWidget"
#include "de/GridLayout"
#include "de/DialogContentStylist"
/** @file widgets/labelwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LabelWidget"
#include "de/TextDrawable"
#include "de/AtlasProceduralImage"
#include "de/StyleProceduralImage"
#include <de/AtlasTexture>
#include <de/ConstantRule>
#include <de/Drawable>
#include <de/GridLayout>
/** @file widgets/lineeditwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LineEditWidget"
#include "de/LabelWidget"
#include "de/FontLineWrapping"
#include "de/GuiRootWidget"
#include "de/GLTextComposer"
#include "de/Style"
#include <de/KeyEvent>
#include <de/MouseEvent>
#include <de/AnimationRule>
#include <de/Drawable>
#if defined (DENG_MOBILE)
#  include <QGuiApplication>
#  include <QInputMethod>
#endif
/** @file widgets/logwidget.cpp  Widget for output message log.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/LogWidget"
#include "de/TextDrawable"
#include "de/Style"
#include <de/KeyEvent>
#include <de/MouseEvent>
#include <de/MemoryLogSink>
#include <de/LogBuffer>
#include <de/AtlasTexture>
#include <de/Drawable>
#include <de/VertexBuilder>
#include <de/Task>
#include <de/TaskPool>
#include <de/App>
#include <QImage>
#include <QPainter>
/** @file widgets/menuwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/MenuWidget"
#include "de/ChildWidgetOrganizer"
#include "de/FocusWidget"
#include "de/GridLayout"
#include "de/PopupButtonWidget"
#include "de/PopupMenuWidget"
#include "de/StyleProceduralImage"
#include "de/VariableToggleWidget"
#include "de/ui/ActionItem"
#include "de/ui/ListData"
#include "de/ui/SubwidgetItem"
#include "de/ui/VariantActionItem"
/** @file notificationareawidget.cpp  Notification area.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NotificationAreaWidget"
#include "de/SequentialLayout"
#include "de/RelayWidget"
#include <de/Drawable>
#include <de/Matrix>
#include <de/AnimationRule>
#include <QMap>
#include <QTimer>
/** @file panelwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PanelWidget"
#include "de/GuiRootWidget"
#include <de/AnimationRule>
#include <de/Drawable>
#include <de/Garbage>
#include <de/LogBuffer>
#include <de/MouseEvent>
#include <de/math.h>
#include <QTimer>
/** @file popupbuttonwidget.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PopupButtonWidget"
/** @file popupmenuwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PopupMenuWidget"
#include "de/MenuWidget"
#include "de/ToggleWidget"
#include "de/GuiRootWidget"
#include "de/ChildWidgetOrganizer"
#include "de/AtlasProceduralImage"
#include "de/ui/Item"
#include <de/App>
#include <de/Config>
#include <de/IndirectRule>
#include <de/OperatorRule>
/** @file popupwidget.cpp
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PopupWidget"
#include "de/ButtonWidget"
#include "de/GuiRootWidget"
#include "de/Style"
#include "de/BaseGuiApp"
#include <de/Drawable>
#include <de/MouseEvent>
#include <de/AnimationRule>
#include <de/Garbage>
#include <de/math.h>
#include <QTimer>
/** @file progresswidget.cpp   Progress indicator.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ProgressWidget"
#include "de/GuiRootWidget"
#include <de/Animation>
#include <de/Lockable>
/** @file relaywidget.cpp  Relays drawing and events.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RelayWidget"
/** @file scriptcommandwidget.cpp  Interactive Doomsday Script command line.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ScriptCommandWidget"
#include "de/PopupWidget"
#include <de/charsymbols.h>
#include <de/shell/Lexicon>
#include <de/App>
#include <de/Script>
#include <de/ScriptLex>
#include <de/ScriptSystem>
#include <de/BuiltInExpression>
#include <de/Process>
#include <de/RecordValue>
#include <de/RefValue>
/** @file scrollareawidget.cpp  Scrollable area.
 *
 * @todo The scroll indicator is currently only implemented for the vertical
 * direction.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ScrollAreaWidget"
#include <de/KeyEvent>
#include <de/MouseEvent>
#include <de/Lockable>
#include <de/Drawable>
/** @file sliderwidget.cpp  Slider to pick a value within a range.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/SliderWidget"
#include "de/DialogWidget"
#include "de/LineEditWidget"
#include "de/TextDrawable"
#include <de/Drawable>
#include <de/MouseEvent>
#include <cmath>
/** @file tabwidget.cpp  Tab widget.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TabWidget"
#include "de/ui/ListData"
#include "de/MenuWidget"
#include <de/AnimationRule>
/** @file togglewidget.cpp  Toggle widget.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ToggleWidget"
#include "de/ProceduralImage"
#include "de/GuiRootWidget"
#include "de/Style"
#include <de/Animation>
/** @file variablearraywidget.cpp  Widget for editing Variables with array values.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/VariableArrayWidget"
#include <de/TextValue>
/** @file variablechoicewidget.cpp  Choice widget for Variable values.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/VariableChoiceWidget"
#include <de/NumberValue>
#include <de/TextValue>
/** @file variablelineeditwidget.cpp  Edits the text of a variable.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/VariableLineEditWidget"
#include <de/TextValue>
/** @file variablesliderwidget.cpp
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/VariableSliderWidget"
#include <de/AnimationValue>
#include <de/NumberValue>
/** @file variabletogglewidget.cpp Toggles the value of a variable.
 *
 * @authors Copyright (c) 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/VariableToggleWidget"
#include <de/NumberValue>

namespace de {
DENG_GUI_PIMPL(AuxButtonWidget)
, DENG2_OBSERVES(ButtonWidget, StateChange) {
    ButtonWidget *aux;
    bool inverted;
    Impl(Public *i)
        : Base(i)
        , inverted(false) {
        self().add(aux = new ButtonWidget);
        aux->setFont("small");
        aux->setTextColor("text");
        aux->setSizePolicy(ui::Expand, ui::Fixed);
        Rule const &unit = rule(RuleBank::UNIT);
        aux->rule()
                .setInput(Rule::Right,  self().rule().right()  - unit)
                .setInput(Rule::Top,    self().rule().top()    + unit)
                .setInput(Rule::Bottom, self().rule().bottom() - unit);
        aux->audienceForStateChange() += this;
        self().margins().set("dialog.gap").setLeft("gap");
        self().margins().setRight(aux->rule().width() + rule("gap"));
    }
    void setAuxBorderColorf(Vector4f const &colorf) {
        aux->set(Background(Background::Rounded, colorf, 6));
    }
    void setAuxBorderColorf(Vector4f const &colorf, Vector4f const &bgColor) {
        aux->set(Background(bgColor, Background::Rounded, colorf, 6));
    }
    void buttonStateChanged(ButtonWidget &, ButtonWidget::State state) {
        switch (state) {
        case ButtonWidget::Up:
            if (!isInverted()) {
                setAuxBorderColorf(style().colors().colorf("accent"));
                aux->setTextModulationColorf(style().colors().colorf("accent"));
            }
            else {
                setAuxBorderColorf(style().colors().colorf("inverted.accent"));
                aux->setTextModulationColorf(style().colors().colorf("inverted.accent"));
            }
            break;
        case ButtonWidget::Hover:
            if (!isInverted()) {
                setAuxBorderColorf(style().colors().colorf("text"));
                aux->setTextModulationColorf(style().colors().colorf("text"));
            }
            else {
                setAuxBorderColorf(style().colors().colorf("inverted.text"));
                aux->setTextModulationColorf(style().colors().colorf("inverted.text"));
            }
            break;
        case ButtonWidget::Down:
            if (!isInverted()) {
                setAuxBorderColorf(style().colors().colorf(""),
                                   style().colors().colorf("inverted.background"));
                aux->setTextModulationColorf(style().colors().colorf("inverted.text"));
            }
            else {
                setAuxBorderColorf(style().colors().colorf(""),
                                   style().colors().colorf("background"));
                aux->setTextModulationColorf(style().colors().colorf("text"));
            }
            break;
        }
    }
    bool isInverted() const {
        return inverted;
    }
    void updateStyle() {
        if (isInverted()) {
            applyInvertedStyle();
        }
        else {
            applyNormalStyle();
        }
    }
    void applyNormalStyle() {
        aux->setHoverTextColor("text");
        buttonStateChanged(*aux, aux->state());
    }
    void applyInvertedStyle() {
        aux->setHoverTextColor("inverted.text");
        buttonStateChanged(*aux, aux->state());
    }
};
AuxButtonWidget::AuxButtonWidget(String const &name)
    : ButtonWidget(name), d(new Impl(this)) {
    useNormalStyle();
}
ButtonWidget &AuxButtonWidget::auxiliary() {
    return *d->aux;
}
void AuxButtonWidget::useNormalStyle() {
    useInfoStyle(false);
    d->inverted = false;
    d->updateStyle();
}
void AuxButtonWidget::invertStyle() {
    useInfoStyle(!isUsingInfoStyle());
    d->inverted = !d->inverted;
    d->updateStyle();
}
void AuxButtonWidget::updateStyle() {
    ButtonWidget::updateStyle();
}
} // namespace de
namespace de {
BlurWidget::BlurWidget(String const &name) : GuiWidget(name) {
    set(Background(Vector4f(1, 1, 1, 0), Background::Blurred));
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(ButtonWidget),
DENG2_OBSERVES(Action, Triggered) {
    State state                   { Up };
    DotPath bgColorId             { "background" };
    DotPath borderColorId         { "text" };
    HoverColorMode hoverColorMode { ReplaceColor };
    ColorTheme colorTheme         { Normal };
    Background::Type bgType       { Background::GradientFrame };
    Action *action                { nullptr };
    Animation scale               { 1.f };
    Animation frameOpacity        { .08f, Animation::Linear };
    bool animating                { false };
    DotPath hoverTextColor;
    DotPath originalTextColor;
    Vector4f originalTextModColor;
    String shortcut;
    Impl(Public *i) : Base(i) {
        setDefaultBackground();
    }
    ~Impl() {
        releaseRef(action);
    }
    void setState(State st) {
        if (state == st) return;
        State const prev = state;
        state = st;
        animating = true;
        switch (st) {
        case Up:
            scale.setValue(1.f, .3f);
            scale.setStyle(prev == Down? Animation::Bounce : Animation::EaseOut);
            frameOpacity.setValue(.08f, .6f);
            if (!hoverTextColor.isEmpty()) {
                switch (hoverColorMode) {
                case ModulateColor:
                    self().setTextModulationColorf(originalTextModColor);
                    break;
                case ReplaceColor:
                    setTemporaryTextColor(originalTextColor);
                    break;
                }
            }
            break;
        case Hover:
            frameOpacity.setValue(.4f, .15f);
            if (!hoverTextColor.isEmpty()) {
                switch (hoverColorMode) {
                case ModulateColor:
                    self().setTextModulationColorf(style().colors().colorf(hoverTextColor));
                    break;
                case ReplaceColor:
                    setTemporaryTextColor(hoverTextColor);
                    break;
                }
            }
            break;
        case Down:
            scale.setValue(.95f);
            frameOpacity.setValue(0);
            break;
        }
        DENG2_FOR_PUBLIC_AUDIENCE2(StateChange, i) {
            i->buttonStateChanged(self(), state);
        }
    }
    void updateHover(Vector2i const &pos) {
        if (state == Down) return;
        if (self().isDisabled()) {
            setState(Up);
            return;
        }
        if (self().hitTest(pos)) {
            if (state == Up) setState(Hover);
        }
        else if (state == Hover) {
            setState(Up);
        }
    }
    Vector4f borderColor() const {
        return style().colors().colorf(borderColorId) *
               Vector4f(1, 1, 1, frameOpacity);
    }
    void setDefaultBackground() {
        self().set(Background(style().colors().colorf(bgColorId),
                            bgType, borderColor(), 6));
    }
    void updateBackground() {
        Background bg = self().background();
        if (bg.type == Background::GradientFrame ||
            bg.type == Background::GradientFrameWithRoundedFill) {
            bg.solidFill = style().colors().colorf(bgColorId);
            bg.color = borderColor();
            self().set(bg);
        }
    }
    void updateAnimation() {
        if (animating) {
            updateBackground();
            self().requestGeometry();
            if (scale.done() && frameOpacity.done()) {
                animating = false;
            }
        }
    }
    void actionTriggered(Action &) {
        DENG2_FOR_PUBLIC_AUDIENCE2(Triggered, i) {
            i->buttonActionTriggered(self());
        }
    }
    void setTemporaryTextColor(DotPath const &id) {
        DotPath const original = originalTextColor;
        self().setTextColor(id); // originalTextColor changes...
        originalTextColor = original;
    }
    DENG2_PIMPL_AUDIENCE(StateChange)
    DENG2_PIMPL_AUDIENCE(Press)
    DENG2_PIMPL_AUDIENCE(Triggered)
};
DENG2_AUDIENCE_METHOD(ButtonWidget, StateChange)
DENG2_AUDIENCE_METHOD(ButtonWidget, Press)
DENG2_AUDIENCE_METHOD(ButtonWidget, Triggered)
ButtonWidget::ButtonWidget(String const &name) : LabelWidget(name), d(new Impl(this)) {
    setBehavior(Focusable);
    setColorTheme(Normal);
}
void ButtonWidget::useInfoStyle(bool yes) {
    setColorTheme(yes? Inverted : Normal);
}
bool ButtonWidget::isUsingInfoStyle() const {
    return d->colorTheme == Inverted;
}
void ButtonWidget::setColorTheme(ColorTheme theme) {
    auto bg = background();
    d->colorTheme = theme;
    setTextModulationColorf(Vector4f(1, 1, 1, 1));
    d->originalTextModColor = textModulationColorf();
    if (theme == Inverted) {
        d->bgType = Background::GradientFrameWithRoundedFill;
        if (bg.type == Background::GradientFrame) bg.type = d->bgType;
        setTextColor("inverted.text");
        setHoverTextColor("inverted.text", ReplaceColor);
        setBorderColor("inverted.text");
        setBackgroundColor("inverted.background");
    }
    else {
        d->bgType = Background::GradientFrame;
        if (bg.type == Background::GradientFrameWithRoundedFill) bg.type = d->bgType;
        setTextColor("text");
        setHoverTextColor("text", ReplaceColor);
        setBorderColor("text");
        setBackgroundColor("background");
    }
    set(bg);
    setImageColor(textColorf());
    updateStyle();
}
GuiWidget::ColorTheme ButtonWidget::colorTheme() const {
    return d->colorTheme;
}
void ButtonWidget::setTextColor(DotPath const &colorId) {
    LabelWidget::setTextColor(colorId);
    d->originalTextColor = colorId;
}
void ButtonWidget::setHoverTextColor(DotPath const &hoverTextId, HoverColorMode mode) {
    d->hoverTextColor = hoverTextId;
    d->hoverColorMode = mode;
}
void ButtonWidget::setBackgroundColor(DotPath const &bgColorId) {
    d->bgColorId = bgColorId;
    d->updateBackground();
}
void ButtonWidget::setBorderColor(DotPath const &borderColorId) {
    d->borderColorId = borderColorId;
    d->updateBackground();
}
void ButtonWidget::setAction(RefArg<Action> action) {
    if (d->action) {
        d->action->audienceForTriggered() -= d;
    }
    changeRef(d->action, action);
    if (action) {
        action->audienceForTriggered() += d;
    }
}
void ButtonWidget::setActionFn(std::function<void ()> callback) {
    setAction(new CallbackAction(callback));
}
Action const *ButtonWidget::action() const {
    return d->action;
}
void ButtonWidget::trigger() {
    if (behavior().testFlag(Focusable)) {
        root().setFocus(this);
    }
    AutoRef<Action> held = holdRef(d->action);
    emit pressed();
    DENG2_FOR_AUDIENCE2(Press, i) i->buttonPressed(*this);
    if (held) {
        held->trigger();
    }
}
ButtonWidget::State ButtonWidget::state() const {
    return d->state;
}
void ButtonWidget::setState(State state) {
    d->setState(state);
}
void ButtonWidget::setShortcutKey(String const &key) {
    d->shortcut = key;
}
String ButtonWidget::shortcutKey() const {
    return d->shortcut;
}
bool ButtonWidget::handleShortcut(KeyEvent const &keyEvent) {
    if (keyEvent.text().isEmpty()) {
        return false;
    }
    if ((!d->shortcut.isEmpty() && d->shortcut.startsWith(keyEvent.text(), Qt::CaseInsensitive)) ||
        text().startsWith(keyEvent.text(), Qt::CaseInsensitive)) {
        trigger();
        return true;
    }
    return false;
}
bool ButtonWidget::handleEvent(Event const &event) {
    if (isDisabled()) return false;
    if (event.isKey() && hasFocus()) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_RETURN ||
            key.ddKey() == DDKEY_ENTER  ||
            key.ddKey() == ' ') {
            if (key.isKeyDown()) {
                root().focusIndicator().fadeIn();
                trigger();
            }
            return true;
        }
    }
    if (event.isMouse()) {
        MouseEvent const &mouse = event.as<MouseEvent>();
        if (mouse.type() == Event::MousePosition) {
            d->updateHover(mouse.pos());
        }
        else if (mouse.type() == Event::MouseButton) {
            switch (handleMouseClick(event)) {
            case MouseClickStarted:
                d->setState(Down);
                return true;
            case MouseClickFinished:
                d->setState(Up);
                d->updateHover(mouse.pos());
                if (hitTest(mouse.pos())) {
                    trigger();
                }
                return true;
            case MouseClickAborted:
                d->setState(Up);
                return true;
            default:
                break;
            }
        }
    }
    return LabelWidget::handleEvent(event);
}
bool ButtonWidget::updateModelViewProjection(Matrix4f &mvp) {
    if (!fequal(d->scale, 1.f)) {
        Rectanglef const &pos = rule().rect();
        mvp = root().projMatrix2D() *
              Matrix4f::scaleThenTranslate(d->scale, pos.middle()) *
              Matrix4f::translate(-pos.middle());
        return true;
    }
    return false;
}
void ButtonWidget::updateStyle() {
    LabelWidget::updateStyle();
    d->updateBackground();
}
void ButtonWidget::update() {
    LabelWidget::update();
    d->updateAnimation();
}
} // namespace de
namespace de {
using namespace ui;
DENG_GUI_PIMPL(ChoiceWidget),
DENG2_OBSERVES(Data, Addition),
DENG2_OBSERVES(Data, Removal),
DENG2_OBSERVES(Data, OrderChange),
DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation),
DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate) {
    /**
     * Items in the choice's popup uses this as action to change the selected
     * item.
     */
    struct SelectAction : public de::Action {
        ChoiceWidget::Impl *wd;
        ui::Item const &selItem;
        SelectAction(ChoiceWidget::Impl *inst, ui::Item const &item)
            : wd(inst), selItem(item) {}
        void trigger() {
            Action::trigger();
            wd->selected = wd->items().find(selItem);
            wd->updateButtonWithSelection();
            wd->updateItemHighlight();
            wd->choices->dismiss();
            emit wd->self().selectionChangedByUser(wd->selected);
        }
    };
    PopupMenuWidget *choices;
    IndirectRule *   maxWidth;
    Data::Pos        selected; ///< One item is always selected.
    String           noSelectionHint;
    Impl(Public *i) : Base(i), selected(Data::InvalidPos) {
        maxWidth = new IndirectRule;
        self().setMaximumTextWidth(rule("choice.item.width.max"));
        self().setTextLineAlignment(ui::AlignLeft);
        self().setFont("choice.selected");
        choices = new PopupMenuWidget;
        choices->items().audienceForAddition() += this;
        choices->items().audienceForRemoval() += this;
        choices->items().audienceForOrderChange() += this;
        choices->menu().organizer().audienceForWidgetCreation() += this;
        choices->menu().organizer().audienceForWidgetUpdate() += this;
        self().add(choices);
        self().setPopup(*choices, ui::Right);
        QObject::connect(choices, &PanelWidget::opened, [this] () {
            if (auto *w = choices->menu().itemWidget<GuiWidget>(selected)) {
                root().setFocus(w);
            }
        });
        updateButtonWithSelection();
        updateStyle();
    }
    ~Impl() {
        releaseRef(maxWidth);
    }
    void updateStyle() {
        choices->set(choices->background().withSolidFill(style().colors().colorf("choice.popup")));
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &item) {
        if (auto *label = maybeAs<LabelWidget>(widget)) {
            label->setMaximumTextWidth(rule("choice.item.width.max"));
        }
        if (auto *but = maybeAs<ButtonWidget>(widget)) {
            but->setAction(new SelectAction(this, item));
        }
    }
    void widgetUpdatedForItem(GuiWidget &, ui::Item const &item) {
        if (isValidSelection() && &item == &self().selectedItem()) {
            updateButtonWithItem(self().selectedItem());
        }
    }
    void updateMaximumWidth() {
        Font const &font = self().font();
        int widest = 0;
        for (uint i = 0; i < items().size(); ++i) {
            EscapeParser esc;
            esc.parse(items().at(i).label());
            widest = de::max(widest, font.advanceWidth(esc.plainText()));
        }
        maxWidth->setSource(OperatorRule::minimum(rule("choice.item.width.max"),
                                                  Const(widest) + self().margins().width()));
    }
    Data const &items() const {
        return choices->items();
    }
    bool isValidSelection() const {
        return selected < items().size();
    }
    void dataItemAdded(Data::Pos id, ui::Item const &) {
        updateMaximumWidth();
        if (selected >= items().size()) {
            selected = 0;
            updateButtonWithSelection();
            return;
        }
        if (id <= selected) {
            selected++;
        }
    }
    void dataItemRemoved(Data::Pos id, ui::Item &) {
        if (id <= selected && selected > 0) {
            selected--;
        }
        updateButtonWithSelection();
        updateMaximumWidth();
    }
    void dataItemOrderChanged() {
        updateButtonWithSelection();
    }
    void updateItemHighlight() {
        for (Data::Pos i = 0; i < items().size(); ++i) {
            if (GuiWidget *w = choices->menu().organizer().itemWidget(i)) {
                w->setFont(i == selected? "choice.selected" : "default");
            }
        }
    }
    void updateButtonWithItem(ui::Item const &item) {
        self().setText(item.label());
        ActionItem const *act = dynamic_cast<ActionItem const *>(&item);
        if (act) {
            self().setImage(act->image());
        }
    }
    void updateButtonWithSelection() {
        if (isValidSelection()) {
            updateButtonWithItem(items().at(selected));
        }
        else {
            self().setText(noSelectionHint);
            self().setImage(Image());
        }
        emit self().selectionChanged(selected);
    }
};
ChoiceWidget::ChoiceWidget(String const &name)
    : PopupButtonWidget(name), d(new Impl(this)) {
    setOpeningDirection(ui::Right);
    d->choices->setAllowDirectionFlip(false);
}
PopupMenuWidget &ChoiceWidget::popup() {
    return *d->choices;
}
void ChoiceWidget::setSelected(Data::Pos pos) {
    d->selected = pos;
    d->updateButtonWithSelection();
    d->updateItemHighlight();
}
bool ChoiceWidget::isValidSelection() const {
    return d->isValidSelection();
}
Data::Pos ChoiceWidget::selected() const {
    return d->selected;
}
Item const &ChoiceWidget::selectedItem() const {
    DENG2_ASSERT(d->isValidSelection());
    return d->items().at(d->selected);
}
Rule const &ChoiceWidget::maximumWidth() const {
    return *d->maxWidth;
}
void ChoiceWidget::openPopup() {
    d->updateItemHighlight();
    d->choices->open();
}
ui::Data &ChoiceWidget::items() {
    return d->choices->items();
}
void ChoiceWidget::setItems(Data const &items) {
    popup().menu().setItems(items);
    d->updateMaximumWidth();
}
void ChoiceWidget::setNoSelectionHint(String const &hint) {
    d->noSelectionHint = hint;
}
void ChoiceWidget::useDefaultItems() {
    popup().menu().useDefaultItems();
    d->updateMaximumWidth();
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(CommandWidget) {
    shell::EditorHistory history;
    DocumentPopupWidget *popup;       ///< Popup for autocompletions.
    bool                 allowReshow; ///< Contents must still be valid.
    Impl(Public *i) : Base(i), history(i), allowReshow(false) {
        popup = new DocumentPopupWidget;
        popup->document().setMaximumLineWidth(rule("editor.completion.linewidth"));
        popup->document().setScrollBarColor("inverted.accent");
        popup->setPreferredHeight(rule("editor.completion.height"),
                                  self().rule().top() - rule("gap"));
        self().add(popup);
    }
};
CommandWidget::CommandWidget(String const &name)
    : LineEditWidget(name), d(new Impl(this))
{}
PopupWidget &CommandWidget::autocompletionPopup() {
    return *d->popup;
}
void CommandWidget::focusGained() {
    LineEditWidget::focusGained();
    emit gotFocus();
}
void CommandWidget::focusLost() {
    LineEditWidget::focusLost();
    closeAutocompletionPopup();
    emit lostFocus();
}
bool CommandWidget::handleEvent(Event const &event) {
    if (isDisabled()) return false;
    if (hasFocus() && event.isKeyDown()) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (d->allowReshow &&
            isSuggestingCompletion() &&
            key.qtKey() == Qt::Key_Tab && !d->popup->isOpen() &&
            suggestedCompletions().size() > 1) {
            d->popup->open();
            return true;
        }
        if (key.qtKey() == Qt::Key_Return || key.qtKey() == Qt::Key_Enter) {
            if (isAcceptedAsCommand(text())) {
                acceptCompletion();
                String const entered = d->history.enter();
                executeCommand(entered);
                emit commandEntered(entered);
            }
            return true;
        }
    }
    if (LineEditWidget::handleEvent(event)) {
        return true;
    }
    if (hasFocus()) {
        if (event.isKey() && event.as<KeyEvent>().ddKey() == DDKEY_TAB) {
            return true;
        }
    }
    return false;
}
void CommandWidget::update() {
    LineEditWidget::update();
    setAttribute(FocusCyclingDisabled, !text().isEmpty());
}
bool CommandWidget::handleControlKey(int qtKey, KeyModifiers const &mods) {
    if (LineEditWidget::handleControlKey(qtKey, mods)) {
        return true;
    }
    if (d->history.handleControlKey(qtKey)) {
        return true;
    }
    return false;
}
void CommandWidget::operator >> (PersistentState &toState) const {
    int const MAX_PERSISTENT_HISTORY = 200;
    toState.objectNamespace().set(name().concatenateMember("history"),
                                  new ArrayValue(d->history.fullHistory(MAX_PERSISTENT_HISTORY)));
}
void CommandWidget::operator << (PersistentState const &fromState) {
    d->history.setFullHistory(fromState.objectNamespace()
                              .getStringList(name().concatenateMember("history")));
}
void CommandWidget::dismissContentToHistory() {
    d->history.goToLatest();
    if (!text().isEmpty()) {
        d->history.enter();
    }
}
void CommandWidget::closeAutocompletionPopup() {
    d->popup->close();
    d->allowReshow = false;
}
void CommandWidget::showAutocompletionPopup(String const &completionsText) {
    d->popup->document().setText(completionsText);
    d->popup->document().scrollToTop(0);
    d->popup->setAnchorX(cursorRect().middle().x);
    d->popup->setAnchorY(rule().top());
    d->popup->open();
    d->allowReshow = true;
}
void CommandWidget::autoCompletionEnded(bool accepted) {
    LineEditWidget::autoCompletionEnded(accepted);
    closeAutocompletionPopup();
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(CompositorWidget) {
    Drawable drawable;
    struct Buffer {
        GLTexture texture;
        QScopedPointer<GLFramebuffer> offscreen;
        void clear() {
            texture.clear();
            offscreen.reset();
        }
    };
    int nextBufIndex;
    QList<Buffer *> buffers; ///< Stack of buffers to allow nested compositing.
    GLUniform uMvpMatrix;
    GLUniform uTex;
    Impl(Public *i)
        : Base(i),
          nextBufIndex(0),
          uMvpMatrix("uMvpMatrix", GLUniform::Mat4),
          uTex      ("uTex",       GLUniform::Sampler2D) {
        uMvpMatrix = Matrix4f::ortho(0, 1, 0, 1);
    }
    /**
     * Starts using the next unused buffer. The buffer is resized if needed.
     * The size of the new buffer is the same as the size of the current GL
     * target.
     */
    Buffer *beginBufferUse() {
        if (nextBufIndex <= buffers.size()) {
            buffers.append(new Buffer);
        }
        Buffer *buf = buffers[nextBufIndex];
        Vector2ui const size = GLState::current().target().rectInUse().size();
        if (buf->texture.size() != size) {
            buf->texture.setUndefinedImage(size, Image::RGBA_8888);
            buf->offscreen.reset(new GLFramebuffer(buf->texture));
        }
        nextBufIndex++;
        return buf;
    }
    void endBufferUse() {
        nextBufIndex--;
    }
    void glInit() {
        DefaultVertexBuf *buf = new DefaultVertexBuf;
        buf->setVertices(gl::TriangleStrip,
                         DefaultVertexBuf::Builder()
                            .makeQuad(Rectanglef(0, 0, 1, 1),
                                      Vector4f  (1, 1, 1, 1),
                                      Rectanglef(0, 0, 1, -1)),
                         gl::Static);
        drawable.addBuffer(buf);
        root().shaders().build(drawable.program(), "generic.textured.color") //"debug.textured.alpha"
                << uMvpMatrix
                << uTex;
    }
    void glDeinit() {
        qDeleteAll(buffers);
        buffers.clear();
        drawable.clear();
    }
    bool shouldBeDrawn() const {
        return self().isInitialized() && !self().isHidden() && self().visibleOpacity() > 0 &&
               GLState::current().target().rectInUse().size() != Vector2ui();
    }
};
CompositorWidget::CompositorWidget(String const &name)
    : GuiWidget(name), d(new Impl(this))
{}
GLTexture &CompositorWidget::composite() const {
    DENG2_ASSERT(!d->buffers.isEmpty());
    return d->buffers.first()->texture;
}
void CompositorWidget::setCompositeProjection(Matrix4f const &projMatrix) {
    d->uMvpMatrix = projMatrix;
}
void CompositorWidget::useDefaultCompositeProjection() {
    d->uMvpMatrix = Matrix4f::ortho(0, 1, 0, 1);
}
void CompositorWidget::viewResized() {
    GuiWidget::viewResized();
}
void CompositorWidget::preDrawChildren() {
    GuiWidget::preDrawChildren();
    if (d->shouldBeDrawn()) {
        root().painter().flush();
        Impl::Buffer *buf = d->beginBufferUse();
        DENG2_ASSERT(!buf->offscreen.isNull());
        GLState::push()
                .setTarget(*buf->offscreen)
                .setViewport(Rectangleui::fromSize(buf->texture.size()));
        buf->offscreen->clear(GLFramebuffer::Color);
    }
}
void CompositorWidget::postDrawChildren() {
    GuiWidget::postDrawChildren();
    if (d->shouldBeDrawn()) {
        root().painter().flush();
        GLState::pop();
        drawComposite();
        d->endBufferUse();
    }
}
void CompositorWidget::glInit() {
    d->glInit();
}
void CompositorWidget::glDeinit() {
    d->glDeinit();
}
void CompositorWidget::drawComposite() {
    if (!d->shouldBeDrawn()) return;
    DENG2_ASSERT(d->nextBufIndex > 0);
    Impl::Buffer *buf = d->buffers[d->nextBufIndex - 1];
    GLState::push()
            .setAlphaTest(false)
            .setBlend(true)
            .setBlendFunc(gl::One, gl::OneMinusSrcAlpha)
            .setDepthTest(false);
    d->uTex = buf->texture;
    d->drawable.draw();
    GLState::pop();
}
} // namespace de
namespace de {
static TimeSpan const FLASH_ANIM_SPAN = 0.75;
/**
 * Compares dialog button items to determine the order in which they
 * should appear in the UI.
 *
 * @param a  Dialog button item.
 * @param b  Dialog button item.
 *
 * @return @c true, if a < b.
 */
static bool dialogButtonOrder(ui::Item const &a, ui::Item const &b) {
    DialogButtonItem const &left  = a.as<DialogButtonItem>();
    DialogButtonItem const &right = b.as<DialogButtonItem>();
    if (!left.role().testFlag(DialogWidget::Default) && right.role().testFlag(DialogWidget::Default)) {
#ifdef MACOSX
        return true;
#else
        return false;
#endif
    }
    if (left.role().testFlag(DialogWidget::Default) && !right.role().testFlag(DialogWidget::Default)) {
#ifdef MACOSX
        return false;
#else
        return true;
#endif
    }
    if (a.label().isEmpty() && !b.label().isEmpty()) {
        return true;
    }
    return false;
}
DENG_GUI_PIMPL(DialogWidget)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate)
, DENG2_OBSERVES(ui::Data, Addition)
, DENG2_OBSERVES(ui::Data, Removal) {
    Modality modality;
    Flags flags;
    ScrollAreaWidget *area = nullptr;
    ScrollAreaWidget *rightArea = nullptr;
    LabelWidget *heading;
    MenuWidget *buttons;
    MenuWidget *extraButtons;
    ui::ListData buttonItems;
    ui::FilteredData mainButtonItems  { buttonItems };
    ui::FilteredData extraButtonItems { buttonItems };
    QEventLoop subloop;
    de::Action *acceptAction;
    Animation glow;
    bool needButtonUpdate;
    float normalGlow;
    bool animatingGlow;
    QScopedPointer<Untrapper> untrapper;
    DialogContentStylist stylist;
    IndirectRule *minWidth;
    Rule const *maxContentHeight = nullptr;
    Impl(Public *i, Flags const &dialogFlags)
        : Base(i)
        , modality(Modal)
        , flags(dialogFlags)
        , heading(0)
        , acceptAction(0)
        , needButtonUpdate(false)
        , animatingGlow(false) {
        minWidth = new IndirectRule;
        normalGlow = style().colors().colorf("glow").w;
        glow.setValue(normalGlow);
        glow.setStyle(Animation::Linear);
        GuiWidget *container = new GuiWidget("container");
        area = new ScrollAreaWidget("area");
        buttons = new MenuWidget("buttons");
        buttons->margins().setTop("");
        buttons->setItems(mainButtonItems);
        buttons->organizer().audienceForWidgetCreation() += this;
        buttons->organizer().audienceForWidgetUpdate() += this;
        extraButtons = new MenuWidget("extra");
        extraButtons->margins().setTop("");
        extraButtons->setItems(extraButtonItems);
        extraButtons->organizer().audienceForWidgetCreation() += this;
        extraButtons->organizer().audienceForWidgetUpdate() += this;
        buttonItems.audienceForAddition() += this;
        buttonItems.audienceForRemoval()  += this;
        mainButtonItems.setFilter([] (ui::Item const &it) {
            if (!is<DialogButtonItem>(it)) return false;
            return !it.as<DialogButtonItem>().role().testFlag(Action);
        });
        extraButtonItems.setFilter([] (ui::Item const &it) {
            if (!is<DialogButtonItem>(it)) return false;
            return it.as<DialogButtonItem>().role().testFlag(Action);
        });
        buttons->setGridSize(0, ui::Expand, 1, ui::Expand);
        extraButtons->setGridSize(0, ui::Expand, 1, ui::Expand);
        area->rule()
                .setInput(Rule::Left,  self().rule().left())
                .setInput(Rule::Top,   self().rule().top())
                .setInput(Rule::Width, area->contentRule().width() + area->margins().width());
        if (flags & WithHeading) {
            heading = new LabelWidget("heading");
            heading->setFont("heading");
            heading->margins()
                    .setBottom("")
                    .setTop (rule("gap") + rule("dialog.gap"))
                    .setLeft(rule("gap") + rule("dialog.gap"));
            heading->setSizePolicy(ui::Expand, ui::Expand);
            heading->setTextColor("accent");
            heading->setImageColor(style().colors().colorf("accent"));
            heading->setOverrideImageSize(heading->font().ascent());
            heading->setTextGap("dialog.gap");
            heading->setAlignment(ui::AlignLeft);
            heading->setTextAlignment(ui::AlignRight);
            heading->setTextLineAlignment(ui::AlignLeft);
            heading->setFillMode(LabelWidget::FillWithText);
            container->add(heading);
            heading->rule()
                    .setInput(Rule::Top,  self().rule().top())
                    .setInput(Rule::Left, self().rule().left());
            area->rule().setInput(Rule::Top, heading->rule().bottom());
        }
        area->rule().setInput(Rule::Height, container->rule().height() - buttons->rule().height());
        buttons->rule()
                .setInput(Rule::Bottom, container->rule().bottom())
                .setInput(Rule::Right, self().rule().right());
        extraButtons->rule()
                .setInput(Rule::Top, buttons->rule().top())
                .setInput(Rule::Left, self().rule().left());
        container->rule().setInput(Rule::Width, OperatorRule::maximum(
                                       area->rule().width(),
                                       buttons->rule().width() + extraButtons->rule().width(),
                                       *minWidth));
        if (flags.testFlag(WithHeading)) {
            area->rule().setInput(Rule::Height,
                                  container->rule().height()
                                  - heading->rule().height()
                                  - buttons->rule().height());
        }
        container->add(area);
        container->add(extraButtons);
        container->add(buttons);
        self().setContent(container);
    }
    ~Impl() {
        releaseRef(minWidth);
        releaseRef(maxContentHeight);
        releaseRef(acceptAction);
    }
    void setupForTwoColumns() {
        if (!rightArea) {
            rightArea = new ScrollAreaWidget("rightArea");
            self().content().add(rightArea);
            rightArea->rule()
                    .setInput(Rule::Top,    area->rule().top())
                    .setInput(Rule::Left,   area->rule().right())
                    .setInput(Rule::Height, area->rule().height())
                    .setInput(Rule::Width,  rightArea->contentRule().width() + rightArea->margins().width());
            if (heading) {
                heading->rule().setInput(Rule::Right, rightArea->rule().right());
            }
            self().content().rule().setInput(Rule::Width, OperatorRule::maximum(
                                           area->rule().width() + rightArea->rule().width(),
                                           buttons->rule().width() + extraButtons->rule().width(),
                                           *minWidth));
            if (self().isOpen()) updateContentHeight();
        }
    }
    void updateContentHeight() {
        Rule const *maxHeight = holdRef(root().viewHeight());
        if (self().openingDirection() == ui::Down) {
            changeRef(maxHeight, *maxHeight - self().anchor().top() - rule("gap"));
        }
        if (maxContentHeight) {
            changeRef(maxHeight, OperatorRule::minimum(*maxHeight, *maxContentHeight));
        }
        AutoRef<Rule> areaContentHeight = area->contentRule().height() + area->margins().height();
        if (rightArea) {
            areaContentHeight.reset(OperatorRule::maximum(areaContentHeight,
                    rightArea->contentRule().height() + rightArea->margins().height()));
        }
        if (!flags.testFlag(WithHeading)) {
            self().content().rule().setInput(Rule::Height,
                    OperatorRule::minimum(*maxHeight, areaContentHeight + buttons->rule().height()));
        }
        else {
            self().content().rule().setInput(Rule::Height,
                    OperatorRule::minimum(*maxHeight, (heading? heading->rule().height() : Const(0)) +
                                                       areaContentHeight + buttons->rule().height()));
        }
        releaseRef(maxHeight);
    }
    void dataItemAdded(ui::Data::Pos, ui::Item const &) {
        needButtonUpdate = true;
    }
    void dataItemRemoved(ui::Data::Pos, ui::Item &) {
        needButtonUpdate = true;
    }
    void updateButtonLayout() {
        buttonItems.stableSort(dialogButtonOrder);
        needButtonUpdate = false;
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &item) {
        if (LabelWidget *lab = maybeAs<LabelWidget>(widget)) {
            lab->setSizePolicy(ui::Expand, ui::Expand);
        }
        if (ButtonItem const *i = maybeAs<ButtonItem>(item)) {
            ButtonWidget &but = widget.as<ButtonWidget>();
            but.setColorTheme(self().colorTheme());
            if (!i->action()) {
                if (i->role() & (Accept | Yes)) {
                    but.setAction(new SignalAction(thisPublic, SLOT(accept())));
                }
                else if (i->role() & (Reject | No)) {
                    but.setAction(new SignalAction(thisPublic, SLOT(reject())));
                }
            }
        }
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) {
        if (ButtonItem const *i = maybeAs<ButtonItem>(item)) {
            ButtonWidget &but = widget.as<ButtonWidget>();
            but.setOverrideImageSize(style().fonts().font("default").height());
            if (item.label().isEmpty()) {
                if (i->role().testFlag(Accept)) {
                    but.setText(tr("OK"));
                }
                else if (i->role().testFlag(Reject)) {
                    but.setText(tr("Cancel"));
                }
                else if (i->role().testFlag(Yes)) {
                    but.setText(tr("Yes"));
                }
                else if (i->role().testFlag(No)) {
                    but.setText(tr("No"));
                }
            }
            if (i->role().testFlag(Default)) {
                but.setTextColor(self().colorTheme() == Normal? "dialog.default" : "inverted.text");
                if (self().colorTheme() == Normal) {
                    but.setHoverTextColor("dialog.default", ButtonWidget::ReplaceColor);
                }
                but.setText(_E(b) + but.text());
            }
            else {
                but.setTextColor(self().colorTheme() == Normal? "text" : "inverted.text");
            }
        }
    }
    ui::ActionItem const *findDefaultAction() const {
        for (ui::Data::Pos i = 0; i < mainButtonItems.size(); ++i) {
            ButtonItem const *act = maybeAs<ButtonItem>(mainButtonItems.at(i));
            if (act->role().testFlag(Default) &&
                buttons->organizer().itemWidget(i)->isEnabled()) {
                return act;
            }
        }
        return 0;
    }
    ButtonWidget *findDefaultButton() const {
        if (ui::ActionItem const *defaultAction = findDefaultAction()) {
            return &buttonWidget(*defaultAction);
        }
        return nullptr;
    }
    ButtonWidget &buttonWidget(ui::Item const &item) const {
        GuiWidget *w = extraButtons->organizer().itemWidget(item);
        if (w) return w->as<ButtonWidget>();
        return buttons->organizer().itemWidget(item)->as<ButtonWidget>();
    }
    void startBorderFlash() {
        animatingGlow = true;
        glow.setValueFrom(1, normalGlow, FLASH_ANIM_SPAN);
        Background bg = self().background();
        bg.color.w = glow;
        self().set(bg);
    }
    void updateBorderFlash() {
        Background bg = self().background();
        bg.color.w = glow;
        self().set(bg);
        if (glow.done()) animatingGlow = false;
    }
    void updateBackground() {
        Background bg = self().background();
        if (self().isUsingInfoStyle()) {
            bg = self().infoStyleBackground();
        }
        else if (style().isBlurringAllowed()) {
            bg.type = Background::SharedBlurWithBorderGlow;
            bg.blur = style().sharedBlurWidget();
            bg.solidFill = Vector4f(0, 0, 0, .65f);
        }
        else {
            bg.type = Background::BorderGlow;
            bg.solidFill = style().colors().colorf("dialog.background");
        }
        self().set(bg);
    }
};
DialogWidget::DialogWidget(String const &name, Flags const &flags)
    : PopupWidget(name), d(new Impl(this, flags)) {
    d->stylist.setContainer(area());
    setOpeningDirection(ui::NoDirection);
    d->updateBackground();
    area().enableIndicatorDraw(true);
}
DialogWidget::Modality DialogWidget::modality() const {
    return d->modality;
}
LabelWidget &DialogWidget::heading() {
    DENG2_ASSERT(d->heading != 0);
    return *d->heading;
}
ScrollAreaWidget &DialogWidget::area() {
    return *d->area;
}
ScrollAreaWidget &DialogWidget::leftArea() {
    d->setupForTwoColumns();
    return *d->area;
}
ScrollAreaWidget &DialogWidget::rightArea() {
    d->setupForTwoColumns();
    return *d->rightArea;
}
void DialogWidget::setMinimumContentWidth(Rule const &minWidth) {
    d->minWidth->setSource(minWidth);
}
void DialogWidget::setMaximumContentHeight(const de::Rule &maxHeight) {
    changeRef(d->maxContentHeight, maxHeight);
}
MenuWidget &DialogWidget::buttonsMenu() {
    return *d->buttons;
}
MenuWidget &DialogWidget::extraButtonsMenu() {
    return *d->extraButtons;
}
ui::Data &DialogWidget::buttons() {
    return d->buttonItems;
}
ButtonWidget &DialogWidget::buttonWidget(String const &label) const {
    GuiWidget *w = d->buttons->organizer().itemWidget(label);
    if (w) return w->as<ButtonWidget>();
    w = d->extraButtons->organizer().itemWidget(label);
    if (w) return w->as<ButtonWidget>();
    throw UndefinedLabel("DialogWidget::buttonWidget", "Undefined label \"" + label + "\"");
}
PopupButtonWidget &DialogWidget::popupButtonWidget(String const &label) const {
    return buttonWidget(label).as<PopupButtonWidget>();
}
ButtonWidget *DialogWidget::buttonWidget(int roleId) const {
    for (uint i = 0; i < d->buttonItems.size(); ++i) {
        DialogButtonItem const &item = d->buttonItems.at(i).as<DialogButtonItem>();
        if ((item.role() & IdMask) == uint(roleId)) {
            return &d->buttonWidget(item);
        }
    }
    return nullptr;
}
PopupButtonWidget *DialogWidget::popupButtonWidget(int roleId) const {
    if (auto *btn = buttonWidget(roleId)) {
        return &btn->as<PopupButtonWidget>();
    }
    return nullptr;
}
QList<ButtonWidget *> DialogWidget::buttonWidgets() const {
    QList<ButtonWidget *> buttons;
    foreach (GuiWidget *w, d->buttons->childWidgets()) {
        if (auto *but = maybeAs<ButtonWidget>(w)) {
            buttons << but;
        }
    }
    return buttons;
}
void DialogWidget::setAcceptanceAction(RefArg<de::Action> action) {
    changeRef(d->acceptAction, action);
}
Action *DialogWidget::acceptanceAction() const {
    return d->acceptAction;
}
int DialogWidget::exec(GuiRootWidget &root) {
    d->modality = Modal;
    DENG2_ASSERT(!hasRoot());
    root.add(this);
    prepare();
    int result = 0;
    try {
#if defined (DENG_MOBILE)
        root.unlock();
#endif
        result = d->subloop.exec();
#if defined (DENG_MOBILE)
        root.lock();
#endif
    }
    catch (...) {
#if defined (DENG_MOBILE)
        root.lock();
#endif
        throw;
    }
    finish(result);
    return result;
}
void DialogWidget::open() {
    open(NonModal);
}
void DialogWidget::open(Modality modality) {
    d->modality = modality;
    DENG2_ASSERT(hasRoot());
    prepare(); // calls base class's open()
}
ui::ActionItem *DialogWidget::defaultActionItem() {
    return const_cast<ui::ActionItem *>(d->findDefaultAction());
}
void DialogWidget::offerFocus() {
    root().setFocus(d->findDefaultButton());
}
void DialogWidget::update() {
    PopupWidget::update();
    if (d->needButtonUpdate) {
        d->updateButtonLayout();
    }
    if (d->animatingGlow) {
        d->updateBorderFlash();
    }
}
bool DialogWidget::handleEvent(Event const &event) {
    if (!isOpen()) return false;
    if (event.isKeyDown()) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_ENTER  ||
            key.ddKey() == DDKEY_RETURN ||
            key.ddKey() == ' ') {
            if (ButtonWidget *but = d->findDefaultButton()) {
                but->trigger();
                return true;
            }
        }
        if (key.ddKey() == DDKEY_ESCAPE) {
            reject();
            return true;
        }
    }
    if (d->modality == Modal) {
        if ((event.isKeyDown() && !event.as<KeyEvent>().isModifier()) ||
            (event.type() == Event::MouseButton &&
             event.as<MouseEvent>().state() == MouseEvent::Pressed &&
             !hitTest(event))) {
            d->startBorderFlash();
        }
        return true;
    }
    else {
        if ((event.type() == Event::MouseButton || event.type() == Event::MousePosition ||
             event.type() == Event::MouseWheel) &&
           hitTest(event)) {
            return true;
        }
    }
    return PopupWidget::handleEvent(event);
}
void DialogWidget::accept(int result) {
    if (d->subloop.isRunning()) {
        DENG2_ASSERT(d->modality == Modal);
        d->subloop.exit(result);
        emit accepted(result);
    }
    else {
        emit accepted(result);
        finish(result);
    }
}
void DialogWidget::reject(int result) {
    if (d->subloop.isRunning()) {
        DENG2_ASSERT(d->modality == Modal);
        d->subloop.exit(result);
        emit rejected(result);
    }
    else {
        emit rejected(result);
        finish(result);
    }
}
void DialogWidget::prepare() {
    d->untrapper.reset(new Untrapper(root().window()));
    if (openingDirection() == ui::NoDirection) {
        setAnchor(root().viewWidth() / 2, root().viewHeight() / 2);
    }
    d->updateContentHeight();
    PopupWidget::open();
}
void DialogWidget::preparePanelForOpening() {
    PopupWidget::preparePanelForOpening();
    d->buttons->updateLayout();
    d->extraButtons->updateLayout();
    d->updateBackground();
}
void DialogWidget::finish(int result) {
    root().setFocus(nullptr);
    close();
    d->untrapper.reset();
    if (result > 0) {
        if (d->acceptAction) {
            AutoRef<de::Action> held = *d->acceptAction;
            held->trigger();
        }
    }
}
DialogWidget::ButtonItem::ButtonItem(RoleFlags flags, String const &label)
    : ui::ActionItem(itemSemantics(flags), label, 0)
    , _role(flags)
{}
DialogWidget::ButtonItem::ButtonItem(RoleFlags flags, Image const &image)
    : ui::ActionItem(itemSemantics(flags), image)
    , _role(flags)
{}
DialogWidget::ButtonItem::ButtonItem(RoleFlags flags, String const &label, RefArg<de::Action> action)
    : ui::ActionItem(itemSemantics(flags), label, action)
    , _role(flags)
{}
DialogWidget::ButtonItem::ButtonItem(RoleFlags flags, Image const &image, RefArg<de::Action> action)
    : ui::ActionItem(itemSemantics(flags), image, "", action)
    , _role(flags)
{}
DialogWidget::ButtonItem::ButtonItem(RoleFlags flags, Image const &image, String const &label, RefArg<de::Action> action)
    : ui::ActionItem(itemSemantics(flags), image, label, action)
    , _role(flags)
{}
ui::Item::Semantics DialogWidget::ButtonItem::itemSemantics(RoleFlags flags) {
    Semantics smt = ActivationClosesPopup | ShownAsButton;
    if (flags & Popup) smt |= ShownAsPopupButton;
    return smt;
}
} // namespace de
namespace de {
static String const CFG_LAST_FOLDER("resource.latestDirectory");
DENG2_PIMPL_NOREF(DirectoryArrayWidget)
{};
DirectoryArrayWidget::DirectoryArrayWidget(Variable &variable, String const &name)
    : VariableArrayWidget(variable, name)
    , d(new Impl) {
    addButton().setText(tr("Add Folder..."));
    addButton().setActionFn([this] () {
        DENG2_BASE_GUI_APP->beginNativeUIMode();
        QFileDialog dlg(nullptr, tr("Select Folder"),
                        Config::get().gets(CFG_LAST_FOLDER, "."), "");
        dlg.setFileMode(QFileDialog::Directory);
        dlg.setReadOnly(true);
        dlg.setLabelText(QFileDialog::Accept, tr("Select"));
        if (dlg.exec()) {
            String dir = dlg.selectedFiles().at(0);
            Config::get().set(CFG_LAST_FOLDER, dir.fileNamePath());
            elementsMenu().items() << makeItem(TextValue(dir));
            setVariableFromWidget();
        }
        DENG2_BASE_GUI_APP->endNativeUIMode();
    });
    updateFromVariable();
}
String DirectoryArrayWidget::labelForElement(Value const &value) const {
    return NativePath(value.asText()).pretty();
}
static const String RECURSE_TOGGLE_NAME("recurse-toggle");
/**
 * Controller that syncs state between Config.resource.recurseFolders and the toggles
 * in the DirectoryArrayWidget items. Destroys itself after the item widget is deleted.
 */
struct RecurseToggler
    : DENG2_OBSERVES(ToggleWidget, Toggle)
    , DENG2_OBSERVES(Widget, Deletion)
    , DENG2_OBSERVES(ui::Item, Change)
    , DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate) {
    DirectoryArrayWidget *owner;
    ToggleWidget *        tog;
    const ui::Item *      item;
    RecurseToggler(DirectoryArrayWidget *owner, LabelWidget &element, const ui::Item &item)
        : owner(owner)
        , item(&item) {
        tog = &element.guiFind(RECURSE_TOGGLE_NAME)->as<ToggleWidget>();
        item.audienceForChange() += this;
        element.audienceForDeletion() += this;
        tog->audienceForToggle() += this;
        owner->elementsMenu().organizer().audienceForWidgetUpdate() += this;
    }
    static Variable &recursed() {
        return Config::get("resource.recursedFolders");
    }
    TextValue key() const {
        return {item->data().toString()};
    }
    void fetch() {
        if (recursed().value().contains(key())) {
            tog->setActive(recursed().value().element(key()).isTrue());
        }
    }
    void toggleStateChanged(ToggleWidget &toggle) override {
        recursed().value().setElement(key(), new NumberValue(toggle.isActive()));
        emit owner->arrayChanged();
    }
    void widgetBeingDeleted(Widget &) override {
        item->audienceForChange() -= this;
        trash(this);
    }
    void itemChanged(const ui::Item &) override {
        fetch();
    }
    void widgetUpdatedForItem(GuiWidget &, const ui::Item &) override {
        fetch();
    }
};
void DirectoryArrayWidget::elementCreated(LabelWidget &element, const ui::Item &item) {
    element.setSizePolicy(ui::Fixed, ui::Expand);
    element.setAlignment(ui::AlignLeft);
    element.setTextLineAlignment(ui::AlignLeft);
    element.setMaximumTextWidth(rule().width());
    element.rule().setInput(Rule::Width, rule().width() - margins().width());
    auto *tog = new ToggleWidget(ToggleWidget::DefaultFlags, RECURSE_TOGGLE_NAME);
    element.add(tog);
    tog->setText("Subdirs");
    tog->setActive(true); // recurse is on by default
    tog->set(Background());
    tog->setFont("small");
    tog->margins().setLeft("unit").setRight("gap").setTop("unit").setBottom("unit");
    tog->setSizePolicy(ui::Expand, ui::Expand);
    tog->rule()
            .setInput(Rule::Right, element.rule().right() - rule("gap"))
            .setMidAnchorY(element.rule().midY());
    element.margins().setRight(tog->rule().width() + rule("gap"));
    new RecurseToggler(this, element, item); // deletes itself
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(DocumentPopupWidget) {
    DocumentWidget *doc;
    ButtonWidget *button = nullptr;
};
DocumentPopupWidget::DocumentPopupWidget(String const &name)
    : PopupWidget(name), d(new Impl) {
    useInfoStyle();
    setContent(d->doc = new DocumentWidget);
}
DocumentPopupWidget::DocumentPopupWidget(ButtonWidget *actionButton, String const &name)
    : PopupWidget(name), d(new Impl) {
    DENG2_ASSERT(actionButton);
    useInfoStyle();
    actionButton->useInfoStyle();
    GuiWidget *box = new GuiWidget;
    d->doc = new DocumentWidget;
    box->add(d->doc);
    box->add(actionButton);
    actionButton->setSizePolicy(ui::Expand, ui::Expand);
    Rule const &gap = rule("gap");
    box->rule()
            .setInput(Rule::Width,  d->doc->rule().width())
            .setInput(Rule::Height, d->doc->rule().height() +
                                    actionButton->rule().height() +
                                    gap);
    d->doc->rule()
            .setInput(Rule::Left,  box->rule().left())
            .setInput(Rule::Right, box->rule().right())
            .setInput(Rule::Top,   box->rule().top());
    actionButton->rule()
            .setInput(Rule::Right, box->rule().right() - gap)
            .setInput(Rule::Top,   d->doc->rule().bottom());
    setContent(box);
}
void DocumentPopupWidget::setPreferredHeight(Rule const &preferredHeight) {
    d->doc->rule().setInput(Rule::Height,
                OperatorRule::minimum(preferredHeight,
                                      d->doc->contentRule().height() +
                                      d->doc->margins().height()));
}
void DocumentPopupWidget::setPreferredHeight(Rule const &preferredHeight, Rule const &maxHeight) {
    d->doc->rule().setInput(Rule::Height,
            OperatorRule::minimum(
                OperatorRule::minimum(preferredHeight,
                                      d->doc->contentRule().height() +
                                      d->doc->margins().height()),
                maxHeight));
}
DocumentWidget &DocumentPopupWidget::document() {
    return *d->doc;
}
DocumentWidget const &DocumentPopupWidget::document() const {
    return *d->doc;
}
ButtonWidget *DocumentPopupWidget::button() {
    return d->button;
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(DocumentWidget),
public Font::RichFormat::IStyle {
    ProgressWidget *progress = nullptr;
    DotPath colorIds[Font::RichFormat::MaxColors];
    ColorBank::Color normalColor;
    ColorBank::Color highlightColor;
    ColorBank::Color dimmedColor;
    ColorBank::Color accentColor;
    ColorBank::Color dimAccentColor;
    ui::SizePolicy widthPolicy = ui::Expand;
    const Rule *maxLineWidth = nullptr;
    ConstantRule *contentMaxWidth = new ConstantRule(0);
    int oldScrollY = 0;
    String styledText;
    String text;
    TextDrawable glText;
    GuiVertexBuilder bgVerts;
    GuiVertexBuilder textVerts;
    Matrix4f scrollMvpMatrix;
    Impl(Public *i) : Base(i) {
        colorIds[Font::RichFormat::NormalColor]    = "document.normal";
        colorIds[Font::RichFormat::HighlightColor] = "document.highlight";
        colorIds[Font::RichFormat::DimmedColor]    = "document.dimmed";
        colorIds[Font::RichFormat::AccentColor]    = "document.accent";
        colorIds[Font::RichFormat::DimAccentColor] = "document.dimaccent";
        updateStyle();
        maxLineWidth = holdRef(rule("document.line.width"));
        progress = new ProgressWidget("progress-indicator");
        progress->setColor("progress.dark.wheel");
        progress->setShadowColor("progress.dark.shadow");
        progress->rule().setRect(self().rule());
        progress->hide();
        self().add(progress);
    }
    ~Impl() {
        releaseRef(contentMaxWidth);
        releaseRef(maxLineWidth);
    }
    void updateStyle() {
        Style const &st = style();
        normalColor    = st.colors().color(colorIds[Font::RichFormat::NormalColor]);
        highlightColor = st.colors().color(colorIds[Font::RichFormat::HighlightColor]);
        dimmedColor    = st.colors().color(colorIds[Font::RichFormat::DimmedColor]);
        accentColor    = st.colors().color(colorIds[Font::RichFormat::AccentColor]);
        dimAccentColor = st.colors().color(colorIds[Font::RichFormat::DimAccentColor]);
        glText.setFont(self().font());
        self().requestGeometry();
    }
    Font::RichFormat::IStyle::Color richStyleColor(int index) const {
        switch (index) {
        default:
        case Font::RichFormat::NormalColor:
            return normalColor;
        case Font::RichFormat::HighlightColor:
            return highlightColor;
        case Font::RichFormat::DimmedColor:
            return dimmedColor;
        case Font::RichFormat::AccentColor:
            return accentColor;
        case Font::RichFormat::DimAccentColor:
            return dimAccentColor;
        }
    }
    void richStyleFormat(int contentStyle,
                         float &sizeFactor,
                         Font::RichFormat::Weight &fontWeight,
                         Font::RichFormat::Style &fontStyle,
                         int &colorIndex) const {
        return style().richStyleFormat(contentStyle, sizeFactor, fontWeight, fontStyle, colorIndex);
    }
    Font const *richStyleFont(Font::RichFormat::Style fontStyle) const {
        if (fontStyle == Font::RichFormat::Monospace) {
            return &style().fonts().font(QStringLiteral("document.monospace"));
        }
        return style().richStyleFont(fontStyle);
    }
    void glInit() {
        atlas().audienceForReposition() += this;
        glText.init(atlas(), self().font(), this);
        self().setIndicatorUv(atlas().imageRectf(root().solidWhitePixel()).middle());
    }
    void glDeinit() {
        atlas().audienceForReposition() -= this;
        glText.deinit();
        bgVerts.clear();
        textVerts.clear();
    }
    void atlasContentRepositioned(Atlas &atlas) {
        self().setIndicatorUv(atlas.imageRectf(root().solidWhitePixel()).middle());
        self().requestGeometry();
    }
    void updateGeometry() {
        int scrollY = self().scrollPositionY().valuei();
        if (oldScrollY != scrollY) {
            oldScrollY = scrollY;
            self().requestGeometry();
        }
        Rectanglei pos;
        if (self().hasChangedPlace(pos)) {
            self().requestGeometry();
        }
        int wrapWidth;
        if (widthPolicy == ui::Expand) {
            wrapWidth = maxLineWidth->valuei();
        }
        else {
            wrapWidth = self().rule().width().valuei() - self().margins().width().valuei();
        }
        glText.setLineWrapWidth(wrapWidth);
        if (glText.update()) {
            if (!glText.isBeingWrapped() && progress->isVisible()) {
                contentMaxWidth->set(de::max(contentMaxWidth->value(), float(glText.wrappedSize().x)));
                self().setContentSize(Vector2ui(contentMaxWidth->valuei(), glText.wrappedSize().y));
                progress->hide();
            }
            self().requestGeometry();
        }
        if (!self().geometryRequested()) return;
        bgVerts.clear();
        self().glMakeGeometry(bgVerts);
        if (!progress->isVisible()) {
            DENG2_ASSERT(glText.isReady());
            Font const &font     = self().font();
            int contentHeight    = de::min(self().contentHeight(), self().rule().height().valuei());
            int const extraLines = 1;
            int numVisLines      = contentHeight / font.lineSpacing().valuei() + 2 * extraLines;
            int firstVisLine     = scrollY / font.lineSpacing().valuei() - extraLines + 1;
            Rangei visRange(firstVisLine, firstVisLine + numVisLines);
            if (visRange != glText.range()) {
                glText.setRange(visRange);
                glText.update(); // alloc visible lines
                textVerts.clear();
                glText.makeVertices(textVerts, Vector2i(0, 0), ui::AlignLeft);
                self().setContentWidth(glText.verticesMaxWidth());
            }
            scrollMvpMatrix = root().projMatrix2D() *
                    Matrix4f::translate(Vector2f(self().contentRule().left().valuei(),
                                                 self().contentRule().top().valuei()));
        }
        self().requestGeometry(false);
    }
    void draw() {
        updateGeometry();
        Vector4f const color = Vector4f(1, 1, 1, self().visibleOpacity());
        auto &painter = root().painter();
        if (bgVerts) {
            painter.setColor(color);
            painter.drawTriangleStrip(bgVerts);
        }
        if (textVerts) {
            painter.setModelViewProjection(scrollMvpMatrix);
            auto const oldClip = painter.normalizedScissor();
            painter.setNormalizedScissor(oldClip & self().normalizedContentRect());
            painter.setColor(color);
            painter.drawTriangleStrip(textVerts);
            painter.setNormalizedScissor(oldClip);
            painter.setModelViewProjection(root().projMatrix2D());
        }
    }
};
DocumentWidget::DocumentWidget(String const &name)
    : ScrollAreaWidget(name)
    , d(new Impl(this)) {
    setWidthPolicy(ui::Expand);
    rule().setInput(Rule::Height, contentRule().height() + margins().height());
    enableIndicatorDraw(true);
}
void DocumentWidget::setText(String const &styledText) {
    if (styledText != d->glText.text()) {
        d->textVerts.clear();
        d->progress->show();
        int indSize = rule("document.progress").valuei();
        setContentSize(Vector2i(indSize, indSize));
        d->contentMaxWidth->set(indSize);
        d->styledText = styledText;
        d->glText.clear();
        d->glText.setText(styledText);
        d->glText.setRange(Rangei()); // none visible initially -- updated later
        requestGeometry();
    }
}
String DocumentWidget::text() const {
    return d->glText.text();
}
void DocumentWidget::setWidthPolicy(ui::SizePolicy policy) {
    d->widthPolicy = policy;
    if (policy == ui::Expand) {
        rule().setInput(Rule::Width, *d->contentMaxWidth + margins().width());
    }
    else {
        rule().clearInput(Rule::Width);
    }
    requestGeometry();
}
void DocumentWidget::setMaximumLineWidth(const Rule &maxWidth) {
    changeRef(d->maxLineWidth, maxWidth);
    d->contentMaxWidth->set(0);
    requestGeometry();
}
void DocumentWidget::setStyleColor(Font::RichFormat::Color id, DotPath const &colorName) {
    DENG2_ASSERT(id != Font::RichFormat::AltAccentColor); // FIXME: not implemented!
    if (id >= 0 && id < Font::RichFormat::MaxColors) {
        d->colorIds[id] = colorName;
        updateStyle();
    }
}
ProgressWidget &DocumentWidget::progress() {
    return *d->progress;
}
void DocumentWidget::viewResized() {
    ScrollAreaWidget::viewResized();
    requestGeometry();
}
void DocumentWidget::drawContent() {
    d->draw();
    ScrollAreaWidget::drawContent();
}
void DocumentWidget::glInit() {
    ScrollAreaWidget::glInit();
    d->glInit();
}
void DocumentWidget::glDeinit() {
    ScrollAreaWidget::glDeinit();
    d->glDeinit();
}
void DocumentWidget::updateStyle() {
    ScrollAreaWidget::updateStyle();
    d->updateStyle();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(FadeToBlackWidget) {
    TimeSpan span = 1;
    bool started = false;
};
FadeToBlackWidget::FadeToBlackWidget() : d(new Impl) {
    set(Background(Vector4f(0, 0, 0, 1)));
}
void FadeToBlackWidget::initFadeFromBlack(TimeSpan const &span) {
    setOpacity(1);
    d->span = span;
    d->started = false;
}
void FadeToBlackWidget::initFadeToBlack(TimeSpan const &span) {
    setOpacity(0);
    d->span = span;
    d->started = false;
}
void FadeToBlackWidget::start(TimeSpan const &delay) {
    d->started = true;
    setOpacity(fequal(opacity().target(), 1)? 0 : 1, d->span, delay);
}
void FadeToBlackWidget::pause() {
    opacityAnimation().pause();
}
void FadeToBlackWidget::resume() {
    opacityAnimation().resume();
}
void FadeToBlackWidget::cancel() {
    d->started = true;
    setOpacity(0);
}
bool FadeToBlackWidget::isStarted() const {
    return d->started;
}
bool FadeToBlackWidget::isDone() const {
    return isStarted() && opacity().done();
}
void FadeToBlackWidget::disposeIfDone() {
    if (isDone()) {
        destroyLater(this);
    }
}
} // namespace de
namespace de {
static TimeSpan FLASH_SPAN = .5;
DENG2_PIMPL(FocusWidget) {
    QTimer flashing;
    SafeWidgetPtr<GuiWidget const> reference;
    Animation color { 0.f, Animation::Linear };
    Vector4f flashColors[2];
    float fadeOpacity = 0.f;
    Impl(Public *i) : Base(i) {
        flashColors[0] = Style::get().colors().colorf("focus.flash.off");
        flashColors[1] = Style::get().colors().colorf("focus.flash.on");
        flashing.setInterval(FLASH_SPAN.asMilliSeconds());
        flashing.setSingleShot(false);
    }
    void flash() {
        if (color.target() > .5f) {
            color.setStyle(Animation::EaseIn);
            color.setValue(0, FLASH_SPAN);
        }
        else {
            color.setStyle(Animation::EaseOut);
            color.setValue(1, FLASH_SPAN);
        }
    }
    Vector4f currentColor() const {
        return flashColors[0] * (1.f - color) + flashColors[1] * color;
    }
};
FocusWidget::FocusWidget(String const &name)
    : LabelWidget(name)
    , d(new Impl(this)) {
    hide();
    connect(&d->flashing, SIGNAL(timeout()), this, SLOT(updateFlash()));
}
void FocusWidget::startFlashing(GuiWidget const *reference) {
    d->reference.reset(reference);
    show();
    if (!d->flashing.isActive()) {
        d->flashing.start();
    }
    d->color = 1;
}
void FocusWidget::stopFlashing() {
    d->flashing.stop();
    hide();
}
void FocusWidget::fadeIn() {
    d->fadeOpacity = 1.f;
}
void FocusWidget::fadeOut() {
    d->fadeOpacity = 0.f;
}
bool FocusWidget::isKeyboardFocusActive() const {
    return d->fadeOpacity > 0 && d->reference;
}
void FocusWidget::update() {
    setOpacity(d->fadeOpacity * (d->reference? d->reference->visibleOpacity() : 0.f));
    set(Background(Background::GradientFrameWithThinBorder, d->currentColor(), 6));
    LabelWidget::update();
}
void FocusWidget::updateFlash() {
    d->flash();
}
} // namespace de
namespace de {
using namespace ui;
static TimeSpan const INDICATOR_ANIM_SPAN = 0.4;
DENG2_PIMPL_NOREF(FoldPanelWidget) {
    /**
     * Indicator that shows whether a fold panel is open or closed.
     */
    struct FoldImage : public ProceduralImage {
        FoldPanelWidget &fold;
        bool needSize;
        bool animating;
        Animation angle;
        FoldImage(FoldPanelWidget &owner)
            : fold(owner)
            , needSize(true)
            , animating(false)
            , angle(0, Animation::EaseBoth)
        {}
        bool update() override {
            bool changed = animating;
            float target = (fold.isOpen()? 0 : 90);
            if (!fequal(target, angle.target())) {
                angle.setValue(target, INDICATOR_ANIM_SPAN);
                animating = true;
                changed = true;
            }
            if (needSize) {
                needSize = false;
                changed = true;
                const float h = pixelsToPoints(fold.title().font().height().value());
                setPointSize({h, h});
            }
            if (animating && angle.done()) {
                animating = false;
            }
            return changed;
        }
        void glMakeGeometry(GuiVertexBuilder &verts, Rectanglef const &rect) override {
            GuiRootWidget &root = fold.root();
            Atlas &atlas = root.atlas();
            ColorBank::Colorf const &textColor = fold.title().textColorf();
            /*verts.makeFlexibleFrame(rect.toRectanglei(), 5, textColor,
                                    atlas.imageRectf(root.roundCorners()));*/
            Rectanglef uv = atlas.imageRectf(root.styleTexture("fold"));
            Matrix4f const turn = Matrix4f::rotateAround(rect.middle(), angle);
            verts.makeQuad(rect, textColor * Vector4f(1, 1, 1, .5f), uv, &turn);
        }
    };
    SafeWidgetPtr<ButtonWidget> title;
    GuiWidget *container = nullptr; ///< Held here while not part of the widget tree.
    DialogContentStylist stylist;
    ~Impl() {
        stylist.clear(); // References the container.
        delete container;
    }
};
FoldPanelWidget::FoldPanelWidget(String const &name) : PanelWidget(name), d(new Impl)
{}
ButtonWidget *FoldPanelWidget::makeTitle(String const &text) {
    d->title.reset(new ButtonWidget("fold-title"));
    d->title->setSizePolicy(Expand, Expand);
    d->title->setText(text);
    d->title->setTextColor("accent");
    d->title->setHoverTextColor("text", ButtonWidget::ReplaceColor);
    d->title->setFont("heading");
    d->title->setAlignment(ui::AlignLeft);
    d->title->setTextLineAlignment(ui::AlignLeft);
    d->title->set(Background()); // no frame or background
    d->title->setAction(new SignalAction(this, SLOT(toggleFold())));
    d->title->setOpacity(.8f);
    d->title->setOverlayImage(new Impl::FoldImage(*this), ui::AlignRight);
    return d->title;
}
ButtonWidget &FoldPanelWidget::title() {
    DENG2_ASSERT(d->title != nullptr);
    return *d->title;
}
void FoldPanelWidget::setContent(GuiWidget *content) {
    d->stylist.setContainer(*content);
    if (!isOpen()) {
        if (d->container) {
            d->container->guiDeleteLater();
        }
        d->container = content;
        return;
    }
    PanelWidget::setContent(content);
}
GuiWidget &FoldPanelWidget::content() const {
    if (d->container) {
        return *d->container;
    }
    return PanelWidget::content();
}
FoldPanelWidget *FoldPanelWidget::makeOptionsGroup(const String &name, const String &heading,
                                                   GuiWidget *parent) {
    auto *fold = new FoldPanelWidget(name);
    parent->add(fold->makeTitle(heading));
    parent->add(fold);
    fold->title().setSizePolicy(ui::Fixed, ui::Expand);
    fold->title().setFont("separator.label");
    fold->title().margins().setTop("gap");
    fold->title().setImageAlignment(ui::AlignRight);
    fold->title().rule()
        .setInput(Rule::Left, fold->rule().left())
        .setInput(Rule::Bottom, fold->rule().top())
        .setInput(Rule::Width, fold->rule().width());
    return fold;
}
void FoldPanelWidget::toggleFold() {
    if (!isOpen()) {
        open();
    }
    else {
        close(0);
    }
}
void FoldPanelWidget::preparePanelForOpening() {
    if (d->container) {
        PanelWidget::setContent(d->container);
        d->container = nullptr;
    }
    if (d->title) {
        d->title->setOpacity(1);
    }
    PanelWidget::preparePanelForOpening();
}
void FoldPanelWidget::panelDismissed() {
    PanelWidget::panelDismissed();
    if (d->title) {
        d->title->setOpacity(.8f, .5f);
    }
    content().notifySelfAndTree(&Widget::deinitialize);
    DENG2_ASSERT(d->container == 0);
    d->container = takeContent();
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(GridPopupWidget) {
    DialogContentStylist stylist;
    GuiWidget *container;
    GridLayout layout;
};
GridPopupWidget::GridPopupWidget(String const &name)
    : PopupWidget(name), d(new Impl) {
    setOpeningDirection(ui::Up);
    setOutlineColor("popup.outline");
    d->container = new GuiWidget;
    setContent(d->container);
    d->stylist.setContainer(*d->container);
    Rule const &gap = rule("gap");
    d->layout.setLeftTop(d->container->rule().left() + gap,
                         d->container->rule().top()  + gap);
    d->layout.setGridSize(2, 0);
    d->layout.setColumnAlignment(0, ui::AlignRight);
}
GridLayout &GridPopupWidget::layout() {
    return d->layout;
}
LabelWidget &GridPopupWidget::addSeparatorLabel(String const &labelText) {
    return *LabelWidget::appendSeparatorWithText(labelText, d->container, &d->layout);
}
GridPopupWidget &GridPopupWidget::operator<<(GuiWidget *widget) {
    d->container->add(widget);
    d->layout << *widget;
    return *this;
}
GridPopupWidget &GridPopupWidget::operator<<(Rule const &rule) {
    d->layout << rule;
    return *this;
}
GridPopupWidget &GridPopupWidget::addSpanning(GuiWidget *widget, int cellSpan) {
    d->container->add(widget);
    d->layout.setCellAlignment(Vector2i(0, d->layout.gridSize().y), ui::AlignLeft);
    d->layout.append(*widget, cellSpan);
    return *this;
}
void GridPopupWidget::commit() {
    Rule const &gap = rule("gap");
    d->container->rule().setSize(d->layout.width()  + gap * 2,
                                 d->layout.height() + gap * 2);
}
} // namespace de
namespace de {
using namespace ui;
DENG_GUI_PIMPL(LabelWidget),
public Font::RichFormat::IStyle {
    AssetGroup assets;
    SizePolicy horizPolicy;
    SizePolicy vertPolicy;
    AlignmentMode alignMode;
    Alignment align;
    Alignment textAlign;
    Alignment lineAlign;
    Alignment imageAlign;
    Alignment overlayAlign;
    FillMode fillMode = FillWithImage;
    TextShadow textShadow = NoShadow;
    DotPath textShadowColorId { "label.shadow" };
    ContentFit imageFit;
    const Rule *overrideImageWidth = nullptr;
    const Rule *overrideImageHeight = nullptr;
    float imageScale;
    Vector4f imageColor;
    Vector4f textGLColor;
    Vector4f shadowColor;
    Rule const *maxTextWidth;
    ConstantRule *width;
    ConstantRule *height;
    IndirectRule *minHeight;
    Rule const *outHeight;
    AnimationRule *appearSize;
    LabelWidget::AppearanceAnimation appearType;
    TimeSpan appearSpan;
    DotPath gapId;
    DotPath shaderId;
    int gap;
    ColorBank::Color highlightColor;
    ColorBank::Color dimmedColor;
    ColorBank::Color accentColor;
    ColorBank::Color dimAccentColor;
    ColorBank::Color altAccentColor;
    Font::RichFormat::IStyle const *richStyle;
    String styledText;
    TextDrawable glText;
    mutable Vector2ui latestTextSize;
    std::unique_ptr<ProceduralImage> image;
    std::unique_ptr<ProceduralImage> overlayImage;
    GuiVertexBuilder verts;
    Impl(Public *i)
        : Base(i)
        , horizPolicy (Fixed)
        , vertPolicy  (Fixed)
        , alignMode   (AlignByCombination)
        , align       (AlignCenter)
        , textAlign   (AlignCenter)
        , lineAlign   (AlignCenter)
        , imageAlign  (AlignCenter)
        , imageFit    (OriginalAspectRatio | FitToSize)
        , imageScale  (1)
        , imageColor  (1, 1, 1, 1)
        , textGLColor (1, 1, 1, 1)
        , maxTextWidth(0)
        , appearSize  (new AnimationRule(0))
        , appearType  (AppearInstantly)
        , appearSpan  (0.0)
        , gapId       ("label.gap")
        , shaderId    ("generic.textured.color_ucolor")
        , richStyle   (0) {
        width     = new ConstantRule(0);
        height    = new ConstantRule(0);
        minHeight = new IndirectRule;
        outHeight = new OperatorRule(OperatorRule::Maximum, *height, *minHeight);
        updateStyle();
        assets += glText;
    }
    ~Impl() {
        releaseRef(width);
        releaseRef(height);
        releaseRef(minHeight);
        releaseRef(outHeight);
        releaseRef(appearSize);
        releaseRef(maxTextWidth);
        releaseRef(overrideImageWidth);
        releaseRef(overrideImageHeight);
    }
    void updateStyle() {
        const Style &st = style();
        gap = rule(gapId).valuei();
        highlightColor = st.colors().color("label.highlight");
        dimmedColor    = st.colors().color("label.dimmed");
        accentColor    = st.colors().color("label.accent");
        dimAccentColor = st.colors().color("label.dimaccent");
        altAccentColor = st.colors().color("label.altaccent");
        shadowColor    = st.colors().colorf(textShadowColorId);
        glText.setFont(self().font());
        glText.forceUpdate();
        self().requestGeometry();
    }
    Vector4i margin() const {
        return self().margins().toVector();
    }
    Color richStyleColor(int index) const {
        switch (index) {
        default:
        case Font::RichFormat::NormalColor:
            return self().textColor();
        case Font::RichFormat::HighlightColor:
            return highlightColor;
        case Font::RichFormat::DimmedColor:
            return dimmedColor;
        case Font::RichFormat::AccentColor:
            return accentColor;
        case Font::RichFormat::DimAccentColor:
            return dimAccentColor;
        case Font::RichFormat::AltAccentColor:
            return altAccentColor;
        }
    }
    void richStyleFormat(int contentStyle, float &sizeFactor, Font::RichFormat::Weight &fontWeight,
                         Font::RichFormat::Style &fontStyle, int &colorIndex) const {
        if (richStyle) {
            richStyle->richStyleFormat(contentStyle, sizeFactor, fontWeight, fontStyle, colorIndex);
        }
        else {
            style().richStyleFormat(contentStyle, sizeFactor, fontWeight, fontStyle, colorIndex);
        }
    }
    Font const *richStyleFont(Font::RichFormat::Style fontStyle) const {
        if (richStyle) {
            return richStyle->richStyleFont(fontStyle);
        }
        return style().richStyleFont(fontStyle);
    }
    void glInit() {
        glText.init(atlas(), self().font(), this);
        if (image) {
            image->glInit();
        }
        if (overlayImage) {
            overlayImage->glInit();
        }
    }
    void glDeinit() {
        verts.clear();
        glText.deinit();
        if (image) {
            image->glDeinit();
        }
        if (overlayImage) {
            overlayImage->glDeinit();
        }
    }
    bool hasImage() const {
        return image && image->pointSize() != ProceduralImage::Size(0, 0);
    }
    bool hasText() const {
        return !glText.text().isEmpty();
    }
    Vector2f imageSize() const {
        Vector2f size = image ? pointsToPixels(image->pointSize()) : Vector2f();
        if (overrideImageWidth) {
            size.x = overrideImageWidth->value();
        }
        if (overrideImageHeight) {
            size.y = overrideImageHeight->value();
        }
        return size;
    }
    Vector2ui textSize() const {
        if (!glText.isBeingWrapped()) {
            latestTextSize = glText.wrappedSize();
        }
        return latestTextSize;
    }
    /**
     * Determines where the label's image and text should be drawn.
     *
     * @param layout  Placement of the image and text.
     */
    void contentPlacement(ContentLayout &layout) const {
        Rectanglei const contentRect = self().contentRect();
        Vector2f const imgSize = imageSize() * imageScale;
        layout.image = Rectanglef::fromSize(imgSize);
        layout.text  = Rectanglei::fromSize(textSize());
        if (horizPolicy == Filled) {
            if (hasText() && textAlign & (AlignLeft | AlignRight)) {
                if (fillMode == FillWithImage) {
                    layout.image.setWidth(int(contentRect.width()) - int(layout.text.width()) - gap);
                }
                else {
                    layout.text.setWidth(int(contentRect.width()) - int(layout.image.width()) - gap);
                }
            }
            else {
                layout.image.setWidth(contentRect.width());
                layout.text.setWidth(contentRect.width());
            }
        }
        if (vertPolicy == Filled) {
            if (hasText() && textAlign & (AlignTop | AlignBottom)) {
                if (fillMode == FillWithImage) {
                    layout.image.setHeight(int(contentRect.height()) - int(layout.text.height()) - gap);
                }
                else {
                    layout.text.setHeight(int(contentRect.height()) - int(layout.image.height()) - gap);
                }
            }
            else {
                layout.image.setHeight(contentRect.height());
                layout.text.setHeight(contentRect.height());
            }
        }
        if (hasImage()) {
            Rectanglef const rect = layout.image;
            if (!imageFit.testFlag(FitToWidth)  || imageFit.testFlag(OriginalAspectRatio)) {
                layout.image.setWidth(imageSize().x);
            }
            if (!imageFit.testFlag(FitToHeight) || imageFit.testFlag(OriginalAspectRatio)) {
                layout.image.setHeight(imageSize().y);
            }
            float const horizScale = rect.width() / layout.image.width();
            float const vertScale  = rect.height() / layout.image.height();
            float scale = 1;
            if (imageFit.testFlag(CoverArea)) {
                scale = de::max(horizScale, vertScale);
            }
            else if (imageFit.testFlag(FitToWidth) && imageFit.testFlag(FitToHeight)) {
                scale = de::min(horizScale, vertScale);
            }
            else if (imageFit.testFlag(FitToWidth)) {
                scale = horizScale;
            }
            else if (imageFit.testFlag(FitToHeight)) {
                scale = vertScale;
            }
            layout.image.setSize(Vector2f(layout.image.size()) * scale);
            if (horizPolicy == Filled) {
                layout.image.setWidth(imageScale * layout.image.width());
            }
            if (vertPolicy == Filled) {
                layout.image.setHeight(imageScale * layout.image.height());
            }
        }
        layout.image.move((Vector2f(layout.text.size()) - layout.image.size()) / 2);
        if (hasImage() && hasText()) {
            if (textAlign & AlignLeft) {
                layout.text.moveLeft(layout.image.left() - layout.text.width() - gap);
            }
            if (textAlign & AlignRight) {
                layout.text.moveLeft(layout.image.right() + gap);
            }
            if (textAlign & AlignTop) {
                layout.text.moveTop(layout.image.top() - layout.text.height() - gap);
            }
            if (textAlign & AlignBottom) {
                layout.text.moveTop(layout.image.bottom() + gap);
            }
            if (textAlign & (AlignLeft | AlignRight)) {
                if (imageAlign & AlignTop) {
                    layout.image.moveTop(layout.text.top());
                }
                if (imageAlign & AlignBottom) {
                    layout.image.moveTop(layout.text.bottom() - layout.image.height());
                }
            }
            if (textAlign & (AlignTop | AlignBottom)) {
                if (imageAlign & AlignLeft) {
                    layout.image.moveLeft(layout.text.left());
                }
                if (imageAlign & AlignRight) {
                    layout.image.moveLeft(layout.text.right() - layout.image.width());
                }
            }
        }
        Rectanglef combined =
                (alignMode == AlignByCombination? (layout.image | layout.text) :
                 alignMode == AlignOnlyByImage?    layout.image :
                                                   layout.text);
        Vector2f delta = applyAlignment(align, combined.size(), contentRect);
        delta -= combined.topLeft;
        layout.image.move(delta);
        layout.text.move(delta.toVector2i());
    }
    /**
     * Determines the maximum amount of width available for text, taking into
     * account the given constraints for the possible image of the label.
     */
    int availableTextWidth() const {
        int w = 0;
        int h = 0;
        if (horizPolicy == Expand) {
            w = root().viewSize().x - (margin().x + margin().z);
        }
        else {
            w = self().rule().width().valuei() - (margin().x + margin().z);
        }
        if (vertPolicy != Expand) {
            h = self().rule().height().valuei() - (margin().y + margin().w);
        }
        if (hasImage()) {
            if (textAlign & (AlignLeft | AlignRight)) {
                Vector2f imgSize = imageSize() * imageScale;
                if (vertPolicy != Expand) {
                    if (imageFit & FitToHeight && imgSize.y > h) {
                        float factor = float(h) / imgSize.y;
                        imgSize.y *= factor;
                        if (imageFit & OriginalAspectRatio) imgSize.x *= factor;
                    }
                }
                w -= gap + imgSize.x;
            }
        }
        if (maxTextWidth) {
            return de::min(maxTextWidth->valuei(), w);
        }
        return w;
    }
    void updateSize() {
        ContentLayout layout;
        contentPlacement(layout);
        Rectanglef combined = layout.image | layout.text;
        width->set (combined.width()  + self().margins().width().valuei());
        height->set(combined.height() + self().margins().height().valuei());
    }
    void updateAppearanceAnimation() {
        if (appearType != AppearInstantly) {
            float const target = (appearType == AppearGrowHorizontally? width->value() : height->value());
            if (!fequal(appearSize->animation().target(), target)) {
                appearSize->set(target, appearSpan);
            }
        }
    }
    void updateGeometry() {
        if (image && image->update()) {
            updateSize();
            self().requestGeometry();
        }
        if (overlayImage && overlayImage->update()) {
            updateSize();
            self().requestGeometry();
        }
        glText.setLineWrapWidth(availableTextWidth());
        if (glText.update()) {
            updateSize();
            self().requestGeometry();
        }
        Rectanglei pos;
        if (self().hasChangedPlace(pos) || self().geometryRequested()) {
            verts.clear();
            self().glMakeGeometry(verts);
            self().requestGeometry(false);
        }
    }
    void draw() {
        updateGeometry();
        if (verts) {
            auto &painter = root().painter();
            Matrix4f mvp;
            bool const isCustomMvp = self().updateModelViewProjection(mvp);
            if (isCustomMvp) {
                painter.setModelViewProjection(mvp);
            }
            painter.setColor(Vector4f(1, 1, 1, self().visibleOpacity()));
            painter.drawTriangleStrip(verts);
            if (isCustomMvp) {
                painter.setModelViewProjection(self().root().projMatrix2D());
            }
        }
    }
    Rule const *widthRule() const {
        switch (appearType) {
        case AppearInstantly:
        case AppearGrowVertically:
            if (horizPolicy == Expand) return width;
            break;
        case AppearGrowHorizontally:
            if (horizPolicy == Expand) return appearSize;
            break;
        }
        return 0;
    }
    Rule const *heightRule() const {
        switch (appearType) {
        case AppearInstantly:
        case AppearGrowHorizontally:
            if (vertPolicy == Expand) return outHeight;
            break;
        case AppearGrowVertically:
            if (vertPolicy == Expand) return appearSize;
            break;
        }
        return 0;
    }
};
LabelWidget::LabelWidget(String const &name) : GuiWidget(name), d(new Impl(this))
{}
AssetGroup &LabelWidget::assets() {
    return d->assets;
}
void LabelWidget::setText(String const &text) {
    if (text != d->styledText) {
        d->styledText = text;
        d->glText.setText(text);
    }
}
void LabelWidget::setImage(Image const &image) {
    if (!image.isNull()) {
        AtlasProceduralImage *proc = new AtlasProceduralImage(*this);
        proc->setImage(image);
        setImage(proc);
    }
    else {
        d->image.reset();
    }
}
void LabelWidget::setImage(ProceduralImage *procImage) {
    d->image.reset(procImage);
    d->updateSize();
    requestGeometry();
}
void LabelWidget::setStyleImage(DotPath const &id, String const &heightFromFont) {
    if (!id.isEmpty()) {
        setImage(new StyleProceduralImage(id, *this));
        if (!heightFromFont.isEmpty()) {
            setOverrideImageSize(style().fonts().font(heightFromFont).height());
        }
    }
}
ProceduralImage *LabelWidget::image() const {
    return d->image.get();
}
void LabelWidget::setOverlayImage(ProceduralImage *overlayProcImage, ui::Alignment const &alignment) {
    d->overlayImage.reset(overlayProcImage);
    d->overlayAlign = alignment;
}
String LabelWidget::text() const {
    return d->glText.text();
}
Vector2ui LabelWidget::textSize() const {
    return d->textSize();
}
Rule const &LabelWidget::contentWidth() const {
    return *d->width;
}
Rule const &LabelWidget::contentHeight() const {
    return *d->height;
}
void LabelWidget::setTextGap(DotPath const &styleRuleId) {
    d->gapId = styleRuleId;
    d->updateStyle();
}
DotPath const &LabelWidget::textGap() const {
    return d->gapId;
}
void LabelWidget::setTextShadow(TextShadow shadow, DotPath const &colorId) {
    d->textShadow = shadow;
    d->textShadowColorId = colorId;
    d->updateStyle();
}
void LabelWidget::setFillMode(FillMode fillMode) {
    d->fillMode = fillMode;
}
void LabelWidget::setAlignment(Alignment const &align, AlignmentMode mode) {
    d->align = align;
    d->alignMode = mode;
}
void LabelWidget::setTextAlignment(Alignment const &textAlign) {
    d->textAlign = textAlign;
}
ui::Alignment LabelWidget::textAlignment() const {
    return d->textAlign;
}
void LabelWidget::setTextLineAlignment(Alignment const &textLineAlign) {
    d->lineAlign = textLineAlign;
}
void LabelWidget::setTextModulationColorf(Vector4f const &colorf) {
    d->textGLColor = colorf;
    requestGeometry();
}
Vector4f LabelWidget::textModulationColorf() const {
    return d->textGLColor;
}
void LabelWidget::setImageAlignment(Alignment const &imageAlign) {
    d->imageAlign = imageAlign;
}
void LabelWidget::setImageFit(ContentFit const &fit) {
    d->imageFit = fit;
}
void LabelWidget::setOverrideImageSize(const Rule &width, const Rule &height) {
    changeRef(d->overrideImageWidth, width);
    changeRef(d->overrideImageHeight, height);
}
RulePair LabelWidget::overrideImageSize() const {
    return {d->overrideImageWidth ? *d->overrideImageWidth : ConstantRule::zero(),
            d->overrideImageHeight ? *d->overrideImageHeight : ConstantRule::zero()};
}
void LabelWidget::setMaximumTextWidth(int pixels) {
    setMaximumTextWidth(Const(pixels));
}
void LabelWidget::setMaximumTextWidth(Rule const &pixels) {
    changeRef(d->maxTextWidth, pixels);
    requestGeometry();
}
void LabelWidget::setMinimumHeight(Rule const &minHeight) {
    d->minHeight->setSource(minHeight);
}
void LabelWidget::setTextStyle(Font::RichFormat::IStyle const *richStyle) {
    d->richStyle = richStyle;
}
void LabelWidget::setImageScale(float scaleFactor) {
    d->imageScale = scaleFactor;
}
void LabelWidget::setImageColor(Vector4f const &imageColor) {
    d->imageColor = imageColor;
    requestGeometry();
}
bool LabelWidget::hasImage() const {
    return d->hasImage();
}
void LabelWidget::update() {
    GuiWidget::update();
    d->assets.setPolicy(d->glText, !isVisible() || !d->styledText?
                        AssetGroup::Ignore : AssetGroup::Required);
    if (isInitialized()) {
        if (d->image) d->image->glInit();
        d->updateGeometry();
    }
    d->updateAppearanceAnimation();
}
void LabelWidget::drawContent() {
    d->draw();
}
void LabelWidget::contentLayout(LabelWidget::ContentLayout &layout) {
    d->contentPlacement(layout);
}
void LabelWidget::glInit() {
    d->glInit();
}
void LabelWidget::glDeinit() {
    d->glDeinit();
}
void LabelWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    GuiWidget::glMakeGeometry(verts);
    ContentLayout layout;
    contentLayout(layout);
    if (d->hasImage()) {
        d->image->setColor(d->imageColor);
        d->image->glMakeGeometry(verts, layout.image);
    }
    if (d->hasText()) {
        if (d->textShadow == RectangleShadow) {
            Rectanglef textBox = Rectanglef::fromSize(textSize());
            ui::applyAlignment(d->lineAlign, textBox, layout.text);
            int const boxSize = pointsToPixels(114);
            Vector2f const off(0, textBox.height() * .08f);
            Vector2f const hoff(textBox.height()/2, 0);
            verts.makeFlexibleFrame(Rectanglef(textBox.midLeft() + hoff + off,
                                               textBox.midRight() - hoff + off)
                                        .expanded(boxSize),
                                    boxSize,
                                    d->shadowColor,
                                    root().atlas().imageRectf(root().borderGlow()));
        }
        d->glText.makeVertices(verts, layout.text, d->lineAlign, d->lineAlign, d->textGLColor);
    }
    if (d->overlayImage) {
        Rectanglef rect = Rectanglef::fromSize(pointsToPixels(d->overlayImage->pointSize()));
        applyAlignment(d->overlayAlign, rect, contentRect());
        d->overlayImage->glMakeGeometry(verts, rect);
    }
}
void LabelWidget::updateStyle() {
    d->updateStyle();
}
bool LabelWidget::updateModelViewProjection(Matrix4f &) {
    return false;
}
void LabelWidget::setWidthPolicy(SizePolicy policy) {
    d->horizPolicy = policy;
    if (policy == Expand) {
        rule().setInput(Rule::Width, *d->widthRule());
    }
    else {
        rule().clearInput(Rule::Width);
    }
}
void LabelWidget::setHeightPolicy(SizePolicy policy) {
    d->vertPolicy = policy;
    if (policy == Expand) {
        rule().setInput(Rule::Height, *d->heightRule());
    }
    else {
        rule().clearInput(Rule::Height);
    }
}
void LabelWidget::setAppearanceAnimation(AppearanceAnimation method, TimeSpan const &span) {
    d->appearType = method;
    d->appearSpan = span;
    if (Rule const *w = d->widthRule()) {
        rule().setInput(Rule::Width, *w);
    }
    if (Rule const *h = d->heightRule()) {
        rule().setInput(Rule::Height, *h);
    }
}
LabelWidget *LabelWidget::newWithText(const String &text, GuiWidget *parent) {
    LabelWidget *w = new LabelWidget;
    w->setText(text);
    if (parent) {
        parent->add(w);
    }
    return w;
}
LabelWidget *LabelWidget::appendSeparatorWithText(const String &text, GuiWidget *parent,
                                                  GridLayout *appendToGrid) {
    std::unique_ptr<LabelWidget> w(newWithText(text, parent));
    w->setTextColor("accent");
    w->setFont("separator.label");
    w->margins().setTop("gap");
    if (appendToGrid) {
        appendToGrid->setCellAlignment(Vector2i(0, appendToGrid->gridSize().y), ui::AlignLeft);
        appendToGrid->append(*w, 2);
    }
    return w.release();
}
} // namespace de
namespace de {
using namespace ui;
static TimeSpan const HEIGHT_ANIM_SPAN = .5f;
static duint const ID_BUF_TEXT   = 1;
static duint const ID_BUF_CURSOR = 2;
DENG_GUI_PIMPL(LineEditWidget) {
    typedef GLBufferT<GuiVertex> VertexBuf;
    AnimationRule *height;
    FontLineWrapping &wraps;
    LabelWidget *hint;
    bool signalOnEnter;
    bool firstUpdateAfterCreation;
    ColorTheme colorTheme = Normal;
    Vector4f textColor;
    Font const *font;
    Time blinkTime;
    Animation hovering;
    float unfocusedBackgroundOpacity = 0.f;
    GLTextComposer composer;
    Drawable drawable;
    GLUniform uMvpMatrix;
    GLUniform uColor;
    GLUniform uCursorColor;
    Impl(Public *i)
        : Base(i)
        , wraps(static_cast<FontLineWrapping &>(i->lineWraps()))
        , hint(0)
        , signalOnEnter(false)
        , firstUpdateAfterCreation(true)
        , font(0)
        , hovering(0, Animation::Linear)
        , uMvpMatrix  ("uMvpMatrix", GLUniform::Mat4)
        , uColor      ("uColor",     GLUniform::Vec4)
        , uCursorColor("uColor",     GLUniform::Vec4) {
        height = new AnimationRule(0);
        self().set(Background(Vector4f(1, 1, 1, 1), Background::GradientFrame));
        self().setFont("editor.plaintext");
        updateStyle();
    }
    ~Impl() {
        releaseRef(height);
    }
    /**
     * Update the style used by the widget from the current UI style.
     */
    void updateStyle() {
        font = &self().font();
        textColor    = style().colors().colorf(colorTheme == Normal? "text" : "inverted.text");
        uCursorColor = style().colors().colorf(colorTheme == Normal? "text" : "inverted.text");
        updateBackground();
        wraps.setFont(*font);
        wraps.clear();
        composer.setWrapping(wraps);
        composer.forceUpdate();
        contentChanged(false);
    }
    int calculateHeight() {
        int const hgt = de::max(font->height().valuei(), wraps.totalHeightInPixels());
        return hgt + self().margins().height().valuei();
    }
    void updateProjection() {
        uMvpMatrix = root().projMatrix2D();
    }
    void updateBackground() {
        if (self().background().type == Background::GradientFrame) {
            Background bg;
            Vector3f const frameColor = style().colors().colorf(colorTheme == Normal? "text" : "inverted.text");
            if (!self().hasFocus()) {
                bg = Background(Background::GradientFrame, Vector4f(frameColor, .15f + hovering * .2f), 6);
                if (unfocusedBackgroundOpacity > 0.f) {
                    bg.solidFill = Vector4f(style().colors().colorf(colorTheme == Normal? "background"
                                                                                        : "inverted.background").xyz(),
                                            unfocusedBackgroundOpacity);
                }
            }
            else {
                bg = Background(style().colors().colorf(colorTheme == Normal? "background"
                                                                            : "inverted.background"),
                                Background::GradientFrame,
                                Vector4f(frameColor, .25f + hovering * .3f), 6);
            }
            self().set(bg);
        }
    }
    void glInit() {
        composer.setAtlas(atlas());
        composer.setText(self().text());
        drawable.addBuffer(ID_BUF_TEXT, new VertexBuf);
        drawable.addBufferWithNewProgram(ID_BUF_CURSOR, new VertexBuf, "cursor");
        shaders().build(drawable.program(), "generic.textured.color_ucolor")
                << uMvpMatrix
                << uColor
                << uAtlas();
        shaders().build(drawable.program("cursor"), "generic.color_ucolor")
                << uMvpMatrix
                << uCursorColor;
        updateProjection();
    }
    void glDeinit() {
        composer.release();
    }
    bool showingHint() const {
        return self().text().isEmpty() && !hint->text().isEmpty() && !self().hasFocus();
    }
    void updateGeometry() {
        updateBackground();
        if (composer.update()) self().requestGeometry();
        Rectanglei pos;
        if (!self().hasChangedPlace(pos) && !self().geometryRequested()) {
            return;
        }
        self().requestGeometry(false);
        VertexBuf::Builder verts;
        self().glMakeGeometry(verts);
        drawable.buffer<VertexBuf>(ID_BUF_TEXT)
                .setVertices(gl::TriangleStrip, verts, gl::Static);
        Rectanglei const caret = self().cursorRect();
        verts.clear();
        verts.makeQuad(caret, Vector4f(1, 1, 1, 1),
                       atlas().imageRectf(self().root().solidWhitePixel()).middle());
        drawable.buffer<VertexBuf>(ID_BUF_CURSOR)
                .setVertices(gl::TriangleStrip, verts, gl::Static);
    }
    void updateHover(Vector2i const &pos) {
        if (/*!self().hasFocus() && */ self().hitTest(pos))
        {
            if (hovering.target() < 1) {
                hovering.setValue(1, .15f);
            }
        }
        else if (hovering.target() > 0) {
            hovering.setValue(0, .6f);
        }
    }
    void contentChanged(bool notify) {
        composer.setText(self().text());
        if (notify) {
            emit self().editorContentChanged();
        }
    }
    void atlasContentRepositioned(Atlas &) {
        self().requestGeometry();
    }
};
LineEditWidget::LineEditWidget(String const &name)
    : GuiWidget(name),
      AbstractLineEditor(new FontLineWrapping),
      d(new Impl(this)) {
    setBehavior(ContentClipping | Focusable);
    setAttribute(FocusHidden);
    rule().setInput(Rule::Height, *d->height);
}
void LineEditWidget::setText(String const &lineText) {
    shell::AbstractLineEditor::setText(lineText);
    if (d->hint) {
        if (d->showingHint()) {
            d->hint->setOpacity(1, .5);
        }
        else {
            d->hint->setOpacity(0);
        }
    }
}
void LineEditWidget::setEmptyContentHint(String const &hintText,
                                         String const &hintFont) {
    if (!d->hint) {
        d->hint = new LabelWidget;
        d->hint->setTextColor("editor.hint");
        d->hint->setAlignment(ui::AlignLeft);
        d->hint->setBehavior(Unhittable | ContentClipping);
        d->hint->rule().setRect(rule());
        d->hint->setOpacity(1);
        add(d->hint);
    }
    d->hint->setFont(hintFont.isEmpty()? String("editor.hint.default") : hintFont);
    d->hint->setText(hintText);
}
void LineEditWidget::setSignalOnEnter(bool enterSignal) {
    d->signalOnEnter = enterSignal;
}
Rectanglei LineEditWidget::cursorRect() const {
    Vector2i const cursorPos = lineCursorPos();
    Vector2i const cp = d->wraps.charTopLeftInPixels(cursorPos.y, cursorPos.x) +
            contentRect().topLeft;
    return Rectanglei(cp + pointsToPixels(Vector2i(-1, 0)),
                      cp + Vector2i(pointsToPixels(1), d->font->height().valuei()));
}
void LineEditWidget::setColorTheme(ColorTheme theme) {
    d->colorTheme = theme;
    d->updateStyle();
}
void LineEditWidget::setUnfocusedBackgroundOpacity(float opacity) {
    d->unfocusedBackgroundOpacity = opacity;
    if (!hasFocus()) {
        d->updateBackground();
    }
}
void LineEditWidget::glInit() {
    LOG_AS("LineEditWidget");
    d->glInit();
}
void LineEditWidget::glDeinit() {
    d->glDeinit();
}
void LineEditWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    GuiWidget::glMakeGeometry(verts);
    Rectanglei const contentRect = this->contentRect();
    Rectanglef const solidWhiteUv = d->atlas().imageRectf(root().solidWhitePixel());
    d->composer.makeVertices(verts, contentRect, AlignLeft, AlignLeft, d->textColor);
    if (isSuggestingCompletion()) {
        Rangei const   comp     = completionRange();
        Vector2i const startPos = linePos(comp.start);
        Vector2i const endPos   = linePos(comp.end);
        Vector2i const offset = contentRect.topLeft + Vector2i(0, d->font->ascent().valuei() + pointsToPixels(2));
        for (int i = startPos.y; i <= endPos.y; ++i) {
            Rangei const span = d->wraps.line(i).range;
            Vector2i start = d->wraps.charTopLeftInPixels(i, i == startPos.y? startPos.x : span.start) + offset;
            Vector2i end   = d->wraps.charTopLeftInPixels(i, i == endPos.y?   endPos.x   : span.end)   + offset;
            verts.makeQuad(Rectanglef(start, end + pointsToPixels(Vector2i(0, 1))),
                           Vector4f(1, 1, 1, 1), solidWhiteUv.middle());
        }
    }
}
void LineEditWidget::updateStyle() {
    d->updateStyle();
}
void LineEditWidget::viewResized() {
    GuiWidget::viewResized();
    updateLineWraps(RewrapNow);
    d->updateProjection();
}
void LineEditWidget::focusGained() {
    d->contentChanged(false /* don't notify */);
    if (d->hint) {
        d->hint->setOpacity(0);
    }
#if defined (DENG_MOBILE)
    {
        auto &win = root().window();
        emit win.textEntryRequest();
        connect(&win, &GLWindow::userEnteredText, this, &LineEditWidget::userEnteredText);
        connect(&win, &GLWindow::userFinishedTextEntry, this, &LineEditWidget::userFinishedTextEntry);
    }
#endif
}
void LineEditWidget::focusLost() {
#if defined (DENG_MOBILE)
    {
        auto &win = root().window();
        disconnect(&win, &GLWindow::userEnteredText, this, &LineEditWidget::userEnteredText);
        disconnect(&win, &GLWindow::userFinishedTextEntry, this, &LineEditWidget::userFinishedTextEntry);
        emit win.textEntryDismiss();
    }
#endif
    d->contentChanged(false /* don't notify */);
    if (d->hint && d->showingHint()) {
        d->hint->setOpacity(1, 1, 0.5);
    }
}
#if defined (DENG_MOBILE)
void LineEditWidget::userEnteredText(QString text) {
    setText(text);
}
void LineEditWidget::userFinishedTextEntry() {
    root().popFocus();
}
#endif
void LineEditWidget::update() {
    GuiWidget::update();
    d->updateBackground();
    updateLineWraps(WrapUnlessWrappedAlready);
    if (d->firstUpdateAfterCreation) {
        d->firstUpdateAfterCreation = false;
        d->height->finish();
    }
}
void LineEditWidget::drawContent() {
    auto &painter = root().painter();
    painter.flush();
    GLState::push().setNormalizedScissor(painter.normalizedScissor());
    float const opac = visibleOpacity();
    d->uColor = Vector4f(1, 1, 1, opac); // Overall opacity.
    Vector4f col = style().colors().colorf("editor.cursor");
    col.w *= (int(d->blinkTime.since() * 2) & 1? .25f : 1.f) * opac;
    if (!hasFocus()) {
        col.w = 0;
    }
    d->uCursorColor = col;
    d->updateGeometry();
    d->drawable.draw();
    GLState::pop();
}
bool LineEditWidget::handleEvent(Event const &event) {
    if (isDisabled()) return false;
    if (event.type() == Event::MousePosition) {
        d->updateHover(event.as<MouseEvent>().pos());
    }
    if (!hasFocus()) {
        switch (handleMouseClick(event)) {
        case MouseClickStarted:
            return true;
        case MouseClickFinished:
            root().setFocus(this);
            return true;
        default:
            break;
        }
    }
    if (is<KeyEvent>(event) && event.as<KeyEvent>().qtKey() == Qt::Key_Enter) {
        qDebug() << "LineEditWidget: Enter key" << event.type() << hasFocus();
    }
    if (hasFocus() && event.isKeyDown()) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.isModifier()) {
            return false;
        }
        /*
        if (key.qtKey() == Qt::Key_Shift) {
            return false;
        }
        if (key.qtKey() == Qt::Key_Control || key.qtKey() == Qt::Key_Alt ||
           key.qtKey() == Qt::Key_Meta) {
            return false;
        }*/
        if (d->signalOnEnter && (key.qtKey() == Qt::Key_Enter || key.qtKey() == Qt::Key_Return)) {
            emit enterPressed(text());
            return true;
        }
        if (handleControlKey(key.qtKey(), modifiersFromKeyEvent(key.modifiers()))) {
            return true;
        }
        if (!key.text().isEmpty() && key.text()[0].isPrint()) {
            insert(key.text());
            return true;
        }
    }
    return GuiWidget::handleEvent(event);
}
shell::AbstractLineEditor::KeyModifiers LineEditWidget::modifiersFromKeyEvent(KeyEvent::Modifiers const &keyMods) {
    KeyModifiers mods;
    if (keyMods.testFlag(KeyEvent::Shift))   mods |= Shift;
    if (keyMods.testFlag(KeyEvent::Control)) mods |= Control;
    if (keyMods.testFlag(KeyEvent::Alt))     mods |= Alt;
    if (keyMods.testFlag(KeyEvent::Meta))    mods |= Meta;
    return mods;
}
int LineEditWidget::maximumWidth() const {
    return rule().recti().width() - margins().width().valuei();
}
void LineEditWidget::numberOfLinesChanged(int /*lineCount*/)
{
    d->height->set(d->calculateHeight(), HEIGHT_ANIM_SPAN);
}
void LineEditWidget::cursorMoved() {
    requestGeometry();
    d->blinkTime = Time();
}
void LineEditWidget::contentChanged() {
    d->contentChanged(true);
    if (hasRoot()) {
        updateLineWraps(WrapUnlessWrappedAlready);
    }
}
void LineEditWidget::autoCompletionEnded(bool) {
    requestGeometry();
}
} // namespace de
namespace de {
using namespace ui;
DENG_GUI_PIMPL(LogWidget),
DENG2_OBSERVES(Atlas, OutOfSpace),
public Font::RichFormat::IStyle {
    typedef GLBufferT<GuiVertex> VertexBuf;
    /**
     * Cached log entry ready for drawing. TextDrawable takes the styled text of the
     * entry and wraps it onto multiple lines according to the available content width.
     *
     * The height of the entry is initially zero. When TextDrawable has finished
     * laying out and preparing the text, the real height is then updated and the
     * content height of the log increases.
     *
     * CacheEntry is accessed only from the main thread. However, instances may be
     * initially created also in background threads (if they happen to flush the log).
     */
    class CacheEntry {
        bool _needWrap { true };
        int _wrapWidth { 0 };
        int _height;    ///< Current height of the entry, in pixels.
        int _oldHeight; ///< Previous height, before calling updateVisibility().
    public:
        TextDrawable drawable;
        CacheEntry(Font const &font, Font::RichFormat::IStyle &richStyle, Atlas &atlas)
            : _height(0)
            , _oldHeight(0) {
            drawable.init(atlas, font, &richStyle);
            drawable.setRange(Rangei()); // Determined later.
        }
        ~CacheEntry() {
            drawable.deinit();
        }
        int height() const {
            return _height;
        }
        int oldHeight() const {
            return _oldHeight;
        }
        bool isReady() const {
            return drawable.isReady();
        }
        void setupWrap(String const &richText, int width) {
            drawable.setText(richText);
            _needWrap = true;
            _wrapWidth = width;
        }
        void rewrap(int width) {
            drawable.setLineWrapWidth(width);
        }
        /**
         * Returns the possible delta in the height of the entry.
         * Does not block even though a long wrapping task is in progress.
         */
        int update() {
            int const old = _height;
            if (drawable.update()) {
                _height = drawable.wraps().height() * drawable.font().lineSpacing().valuei();
                return _height - old;
            }
            return 0;
        }
        void beginWrap() {
            if (_needWrap) {
                drawable.setLineWrapWidth(_wrapWidth);
                _needWrap = false;
            }
        }
        /**
         * Updates the entry's visibility: which lines might be visible to the user
         * and thus need to be allocated on an atlas and ready to draw.
         *
         * @param yBottom        Bottom coordinate for the entry.
         * @param visiblePixels  Full range of currently visible pixels in the widget.
         *
         * @return Possible change in the height of the entry.
         */
        int updateVisibility(int yBottom, Rangei const &visiblePixels) {
            beginWrap();
            int heightDelta = 0;
            _oldHeight = _height;
            /*
             * At this point:
             * - we may have no content ready yet (_height is 0)
             * - wrapping may have completed for the first time
             * - wrapping may be ongoing for rewrapping, but we can still update the
             *   current content's visibility
             * - wrapping may have completed for an updated content
             */
            if (!drawable.isBeingWrapped()) {
                heightDelta += update();
            }
            if (!_height) {
                return 0;
            }
            int const lineSpacing = drawable.font().lineSpacing().value();
            int const yTop = yBottom - _height;
            Rangei range;
            if (yBottom < visiblePixels.start || yTop > visiblePixels.end) {
            }
            else if (yTop >= visiblePixels.start && yBottom <= visiblePixels.end) {
                range = Rangei(0, drawable.wraps().height());
            }
            else if (yTop < visiblePixels.start && yBottom > visiblePixels.end) {
                range = Rangei::fromSize((visiblePixels.start - yTop) / lineSpacing,
                                         (visiblePixels.end - visiblePixels.start) / lineSpacing + 1);
            }
            else if (yBottom > visiblePixels.end) {
                DENG2_ASSERT(yTop >= visiblePixels.start);
                range = Rangei(0, (visiblePixels.end - yTop) / lineSpacing);
            }
            else {
                DENG2_ASSERT(yBottom <= visiblePixels.end);
                int count = (yBottom - visiblePixels.start) / lineSpacing;
                range = Rangei(drawable.wraps().height() - count,
                               drawable.wraps().height());
            }
            drawable.setRange(range);
            return update() + heightDelta;
        }
        void make(GuiVertexBuilder &verts, int y) {
            DENG2_ASSERT(isReady());
            drawable.makeVertices(verts, Vector2i(0, y), AlignLeft);
        }
        void releaseFromAtlas() {
            drawable.setRange(Rangei()); // Nothing visible.
        }
    };
    /**
     * Log sink that where all entries that will be visible in the widget will
     * be received. For each entry, a CacheEntry is created and its TextDrawable
     * will start processing the entry contents in the background.
     *
     * LogWidget will periodically check if excess entries should be removed. Entries are
     * only removed from the sink (and cache) during a prune, in the main thread,
     * during which the sink is locked.
     */
    class WrappingMemoryLogSink : public MemoryLogSink {
    public:
        WrappingMemoryLogSink(LogWidget::Impl *wd)
            : d(wd)
            , _maxEntries(1000)
            , _next(0)
            , _width(0)
        {}
        ~WrappingMemoryLogSink() {
            clear();
        }
        int maxEntries() const { return _maxEntries; }
        void clear() {
            DENG2_GUARD(_wrappedEntries);
            qDeleteAll(_wrappedEntries);
            _wrappedEntries.clear();
        }
        void remove(int pos, int n = 1) {
            DENG2_GUARD(this);
            DENG2_ASSERT(pos + n <= _next);
            MemoryLogSink::remove(pos, n);
            _next -= n;
        }
        void setWidth(int wrapWidth) {
            _width = wrapWidth;
            beginWorkOnNext();
        }
        void addedNewEntry(LogEntry &) {
            beginWorkOnNext();
        }
        CacheEntry *nextCachedEntry() {
            DENG2_GUARD(_wrappedEntries);
            if (_wrappedEntries.isEmpty()) return 0;
            return _wrappedEntries.takeFirst();
        }
        /**
         * Schedules wrapping tasks for all incoming entries.
         */
        void beginWorkOnNext() {
            if (!d->formatter) return; // Must have a formatter.
            DENG2_GUARD(this);
            while (_width > 0 && _next >= 0 && _next < entryCount()) {
                LogEntry const &ent = entry(_next);
                String const styled = d->formatter->logEntryToTextLines(ent).at(0);
                CacheEntry *cached = new CacheEntry(*d->font, *d, *d->entryAtlas);
                cached->setupWrap(styled, _width); {
                    DENG2_GUARD(_wrappedEntries);
                    _wrappedEntries << cached;
                }
                _next++;
            }
        }
    private:
        LogWidget::Impl *d;
        int _maxEntries;
        int _next;
        int _width;
        struct WrappedEntries : public QList<CacheEntry *>, public Lockable {};
        WrappedEntries _wrappedEntries; ///< New entries possibly created in background threads.
    };
    WrappingMemoryLogSink sink;
    QList<CacheEntry *> cache; ///< Cached entries in use when drawing.
    int cacheWidth;
    Rangei visibleRange;
    Animation contentOffset; ///< Additional vertical offset to apply when drawing content.
    int contentOffsetForDrawing; ///< Set when geometry updated.
    LogSink::IFormatter *formatter;
    Font const *font;
    ColorBank::Color normalColor;
    ColorBank::Color highlightColor;
    ColorBank::Color dimmedColor;
    ColorBank::Color accentColor;
    ColorBank::Color dimAccentColor;
    ColorBank::Color altAccentColor;
    VertexBuf *buf;
    VertexBuf *bgBuf;
    AtlasTexture *entryAtlas;
    bool entryAtlasLayoutChanged;
    bool entryAtlasFull;
    Drawable contents;
    Drawable background;
    GLUniform uMvpMatrix;
    GLUniform uTex;
    GLUniform uShadowColor;
    GLUniform uColor;
    GLUniform uBgMvpMatrix;
    Matrix4f projMatrix;
    Matrix4f viewMatrix;
    Id scrollTex;
    Impl(Public *i)
        : Base(i)
        , sink(this)
        , cacheWidth(0)
        , visibleRange(Rangei(-1, -1))
        , formatter(0)
        , font(0)
        , buf(0)
        , entryAtlas(0)
        , entryAtlasLayoutChanged(false)
        , entryAtlasFull(false)
        , uMvpMatrix  ("uMvpMatrix", GLUniform::Mat4)
        , uTex        ("uTex",       GLUniform::Sampler2D)
        , uShadowColor("uColor",     GLUniform::Vec4)
        , uColor      ("uColor",     GLUniform::Vec4)
        , uBgMvpMatrix("uMvpMatrix", GLUniform::Mat4) {
        self().setFont("log.normal");
        updateStyle();
    }
    ~Impl() {
        LogBuffer::get().removeSink(sink);
        DENG2_ASSERT(cache.isEmpty());
    }
    void clear() {
        sink.clear();
        clearCache();
        self().setContentHeight(0);
    }
    void clearCache() {
        qDeleteAll(cache); // Ongoing text wrapping cancelled automatically.
        cache.clear();
    }
    void updateStyle() {
        Style const &st = style();
        font           = &self().font();
        normalColor    = st.colors().color("log.normal");
        highlightColor = st.colors().color("log.highlight");
        dimmedColor    = st.colors().color("log.dimmed");
        accentColor    = st.colors().color("log.accent");
        dimAccentColor = st.colors().color("log.dimaccent");
        altAccentColor = st.colors().color("log.altaccent");
        self().set(Background(st.colors().colorf("background")));
    }
    Font::RichFormat::IStyle::Color richStyleColor(int index) const {
        switch (index) {
        default:
        case Font::RichFormat::NormalColor:
            return normalColor;
        case Font::RichFormat::HighlightColor:
            return highlightColor;
        case Font::RichFormat::DimmedColor:
            return dimmedColor;
        case Font::RichFormat::AccentColor:
            return accentColor;
        case Font::RichFormat::DimAccentColor:
            return dimAccentColor;
        case Font::RichFormat::AltAccentColor:
            return altAccentColor;
        }
    }
    void richStyleFormat(int contentStyle,
                         float &sizeFactor,
                         Font::RichFormat::Weight &fontWeight,
                         Font::RichFormat::Style &fontStyle,
                         int &colorIndex) const {
        return style().richStyleFormat(contentStyle, sizeFactor, fontWeight, fontStyle, colorIndex);
    }
    Font const *richStyleFont(Font::RichFormat::Style fontStyle) const {
        return style().richStyleFont(fontStyle);
    }
    void glInit() {
        entryAtlas = AtlasTexture::newWithRowAllocator(
                Atlas::BackingStore | Atlas::AllowDefragment,
                GLTexture::maximumSize().min(Atlas::Size(4096, 2048)));
        entryAtlas->audienceForReposition() += this;
        entryAtlas->audienceForOutOfSpace() += this;
        /*Image solidWhitePixel = Image::solidColor(Image::Color(255, 255, 255, 255),
                                                  Image::Size(1, 1));
        scrollTex = entryAtlas->alloc(solidWhitePixel);
        self().setIndicatorUv(entryAtlas->imageRectf(scrollTex).middle());*/
        uTex = entryAtlas;
        uColor = Vector4f(1, 1, 1, 1);
        background.addBuffer(bgBuf = new VertexBuf);
        shaders().build(background.program(), "generic.textured.color")
                << uBgMvpMatrix
                << uAtlas();
        contents.addBuffer(buf = new VertexBuf);
        shaders().build(contents.program(), "generic.textured.color_ucolor")
                << uMvpMatrix
                << uShadowColor
                << uTex;
    }
    void glDeinit() {
        clearCache();
        delete entryAtlas;
        entryAtlas = 0;
        scrollTex = Id::None;
        contents.clear();
        background.clear();
    }
    void atlasContentRepositioned(Atlas &atlas) {
        if (entryAtlas == &atlas) {
            entryAtlasLayoutChanged = true;
            self().setIndicatorUv(entryAtlas->imageRectf(scrollTex).middle());
        }
    }
    void atlasOutOfSpace(Atlas &atlas) {
        if (entryAtlas == &atlas) {
            entryAtlasFull = true;
        }
    }
    duint contentWidth() const {
        return self().viewportSize().x;
    }
    int maxVisibleOffset() {
        return self().maximumScrollY().valuei();
    }
    void modifyContentHeight(float delta) {
        self().modifyContentHeight(delta);
        if (!self().isAtBottom()) {
            self().scrollPositionY().shift(delta);
        }
    }
    void fetchNewCachedEntries() {
        while (CacheEntry *cached = sink.nextCachedEntry()) {
            cache << cached;
        }
    }
    void rewrapCache() {
        int startFrom = max(0, visibleRange.start);
        for (int idx = startFrom; idx < cache.size(); ++idx) {
            cache[idx]->rewrap(contentWidth());
        }
        for (int idx = startFrom - 1; idx >= 0; --idx) {
            cache[idx]->rewrap(contentWidth());
        }
    }
    void releaseExcessComposedEntries() {
        if (visibleRange < 0) return;
        int len = de::max(10, visibleRange.size());
        int excess = visibleRange.start - len;
        for (int i = 0; i <= excess && i < cache.size(); ++i) {
            cache[i]->releaseFromAtlas();
        }
        excess = visibleRange.end + len;
        for (int i = excess; i < cache.size(); ++i) {
            cache[i]->releaseFromAtlas();
        }
    }
    /**
     * Releases all entries currently stored in the entry atlas.
     */
    void releaseAllNonVisibleEntries() {
        for (int i = 0; i < cache.size(); ++i) {
            if (!visibleRange.contains(i)) {
                cache[i]->releaseFromAtlas();
            }
        }
    }
    /**
     * Removes entries from the sink and the cache.
     */
    void prune() {
        DENG2_ASSERT_IN_MAIN_THREAD();
        int num = cache.size() - sink.maxEntries();
        if (num > 0) {
            sink.remove(0, num);
            for (int i = 0; i < num; ++i) {
                self().modifyContentHeight(-cache.first()->height());
                delete cache.takeFirst();
            }
        }
    }
    void updateProjection() {
        projMatrix = root().projMatrix2D();
        uBgMvpMatrix = projMatrix;
    }
    Rangei extendPixelRangeWithPadding(Rangei const &range) {
        int const padding = range.size() / 2;
        return Rangei(range.start - padding, range.end + padding);
    }
    void updateGeometry() {
        bool needHeightNotify = false; // if changed as entries are updated
        int heightDelta = 0;
        Vector2i const contentSize = self().viewportSize();
        if (cacheWidth != contentSize.x) {
            rewrapCache();
            cacheWidth = contentSize.x;
        }
        entryAtlasFull = false;
        VertexBuf::Builder verts;
        int initialYBottom = contentSize.y + self().scrollPositionY().valuei();
        contentOffsetForDrawing = std::ceil(contentOffset.value());
        Rangei visiblePixelRange = extendPixelRangeWithPadding(
                    Rangei(-contentOffsetForDrawing, contentSize.y - contentOffsetForDrawing));
        if (!isVisible()) {
            visiblePixelRange = Rangei(); // Nothing to be seen.
        }
        for (int attempt = 0; attempt < 2; ++attempt) {
            if (entryAtlasFull) {
                releaseAllNonVisibleEntries();
                entryAtlasFull = false;
            }
            int yBottom = initialYBottom;
            visibleRange = Rangei(-1, -1);
            entryAtlasLayoutChanged = false;
            for (int idx = cache.size() - 1; yBottom >= -contentOffsetForDrawing && idx >= 0; --idx) {
                CacheEntry *entry = cache[idx];
                if (int delta = entry->updateVisibility(yBottom, visiblePixelRange)) {
                    heightDelta += delta;
                    if (!entry->oldHeight()) {
                        needHeightNotify = true;
                        continue;
                    }
                }
                yBottom -= entry->height();
                if (entry->isReady() && yBottom + contentOffsetForDrawing <= contentSize.y) {
                    entry->make(verts, yBottom);
                    if (visibleRange.end == -1) {
                        visibleRange.end = idx;
                    }
                    visibleRange.start = idx;
                }
                if (entryAtlasLayoutChanged || entryAtlasFull) {
                    goto nextAttempt;
                }
            }
            break;
nextAttempt:
            verts.clear();
        }
        buf->setVertices(gl::TriangleStrip, verts, gl::Dynamic);
        if (heightDelta) {
            modifyContentHeight(heightDelta);
            if (needHeightNotify && heightDelta > 0) {
                emit self().contentHeightIncreased(heightDelta);
            }
        }
        releaseExcessComposedEntries();
    }
    bool isVisible() const {
        Rectanglei vp = self().viewport();
        return vp.height() > 0 && vp.right() >= 0;
    }
    void draw() {
        Rectanglei pos;
        if (self().hasChangedPlace(pos) || !bgBuf->isReady()) {
            VertexBuf::Builder bgVerts;
            self().glMakeGeometry(bgVerts);
            bgBuf->setVertices(gl::TriangleStrip, bgVerts, gl::Static);
        }
        background.draw();
        Rectanglei vp = self().viewport();
        if (vp.height() > 0) {
            GLState &st = GLState::push();
            st.setNormalizedScissor(
                    self().normalizedRect(
                            vp.adjusted(Vector2i(), Vector2i(self().margins().right().valuei(), 0))));
            uMvpMatrix = projMatrix * Matrix4f::translate(
                         Vector2f(vp.topLeft + Vector2i(0, contentOffsetForDrawing)));
            uShadowColor = Vector4f(0, 0, 0, 1);
            contents.draw();
            uMvpMatrix = projMatrix * Matrix4f::translate(
                         Vector2f(vp.topLeft + Vector2i(0, contentOffsetForDrawing - 1)));
            uShadowColor = Vector4f(1, 1, 1, 1);
            contents.draw();
            GLState::pop();
        }
    }
};
LogWidget::LogWidget(String const &name)
    : ScrollAreaWidget(name), d(new Impl(this)) {
    setOrigin(Bottom);
    enableIndicatorDraw(true);
    LogBuffer::get().addSink(d->sink);
}
void LogWidget::setLogFormatter(LogSink::IFormatter &formatter) {
    d->formatter = &formatter;
}
void LogWidget::setPrivilegedEntries(bool onlyPrivileged) {
    d->sink.setPrivileged(onlyPrivileged);
}
LogSink &LogWidget::logSink() {
    return d->sink;
}
void LogWidget::clear() {
    d->clear();
}
void LogWidget::setContentYOffset(Animation const &anim) {
    if (isAtBottom()) {
        d->contentOffset = anim;
    }
    else {
        d->contentOffset = 0;
    }
}
Animation const &LogWidget::contentYOffset() const {
    return d->contentOffset;
}
void LogWidget::viewResized() {
    ScrollAreaWidget::viewResized();
    d->updateProjection();
}
void LogWidget::update() {
    ScrollAreaWidget::update();
    d->sink.setWidth(d->contentWidth());
    d->fetchNewCachedEntries();
    d->prune();
    d->updateGeometry();
}
void LogWidget::drawContent() {
    root().painter().flush();
    d->draw();
    ScrollAreaWidget::drawContent();
}
bool LogWidget::handleEvent(Event const &event) {
    return ScrollAreaWidget::handleEvent(event);
}
void LogWidget::glInit() {
    ScrollAreaWidget::glInit();
    d->glInit();
}
void LogWidget::glDeinit() {
    ScrollAreaWidget::glDeinit();
    d->glDeinit();
}
} // namespace de
namespace de {
using namespace ui;
DENG2_PIMPL(MenuWidget)
, DENG2_OBSERVES(Data, Addition)        // for layout update
, DENG2_OBSERVES(Data, Removal)         // for layout update
, DENG2_OBSERVES(Data, OrderChange)     // for layout update
, DENG2_OBSERVES(Widget, ChildAddition) // for layout update
, DENG2_OBSERVES(Widget, ChildRemoval)  // for layout update
, DENG2_OBSERVES(PopupWidget, Close)
, DENG2_OBSERVES(Widget, Deletion)
, DENG2_OBSERVES(ButtonWidget, Press) // child button triggers
, public ChildWidgetOrganizer::IWidgetFactory {
    /**
     * Base class for sub-widget actions. Handles ownership/openness tracking.
     */
    class SubAction : public de::Action {
    public:
        SubAction(MenuWidget::Impl *inst, ui::Item const &parentItem)
            : d(inst)
            , _parentItem(parentItem)
            , _dir(ui::Right)
        {}
        void setWidget(PopupWidget *w, ui::Direction openingDirection) {
            _widget.reset(w);
            _dir = openingDirection;
        }
        bool isTriggered() const {
            return bool(_widget);
        }
        GuiWidget &parent() const {
            auto *p = d->organizer.itemWidget(_parentItem);
            DENG2_ASSERT(p != 0);
            return *p;
        }
        void trigger() {
            DENG2_ASSERT(bool(_widget));
            DENG2_ASSERT(d->self().hasRoot());
            if (_widget->isOpeningOrClosing()) return;
            if (!_widget->parentWidget()) {
                d->self().root().add(_widget);
            }
            Action::trigger();
            if (auto *subMenu = maybeAs<PopupMenuWidget>(_widget)) {
                if (auto *parentMenu = parent().parentWidget()) {
                    subMenu->setParentPopup(maybeAs<PopupWidget>(parentMenu->parentWidget()));
                }
            }
            _widget->setAnchorAndOpeningDirection(parent().hitRule(), _dir);
            d->keepTrackOfSubWidget(_widget);
            _widget->open();
        }
    protected:
        MenuWidget::Impl *d;
        ui::Item const &_parentItem;
        ui::Direction _dir;
        SafeWidgetPtr<PopupWidget> _widget;
    };
    /**
     * Action owned by the button that represents a SubmenuItem.
     */
    class SubmenuAction : public SubAction {
    public:
        SubmenuAction(MenuWidget::Impl *inst, ui::SubmenuItem const &parentItem)
            : SubAction(inst, parentItem) {
            _sub.reset(new PopupMenuWidget);
            setWidget(_sub, parentItem.openingDirection());
            _sub->menu().setItems(parentItem.items());
        }
        ~SubmenuAction() {
            GuiWidget::destroy(_sub);
        }
    private:
        SafeWidgetPtr<PopupMenuWidget> _sub; // owned
    };
    /**
     * Action owned by the button that represents a SubwidgetItem.
     */
    class SubwidgetAction : public SubAction {
    public:
        SubwidgetAction(MenuWidget::Impl *inst, ui::SubwidgetItem const &parentItem)
            : SubAction(inst, parentItem)
            , _item(parentItem)
        {}
        void trigger() {
            if (isTriggered()) return; // Already open, cannot retrigger.
            setWidget(_item.makeWidget(), _item.openingDirection());
            _widget->setDeleteAfterDismissed(true);
            if (_item.semantics().testFlag(Item::ClosesParentPopup)) {
                if (auto *selfPopup = maybeAs<PopupMenuWidget>(d->self().parentGuiWidget())) {
                    QObject::connect(_widget.get(), SIGNAL(closed()), selfPopup, SLOT(close()));
                }
            }
            SubAction::trigger();
        }
    private:
        ui::SubwidgetItem const &_item;
    };
    AssetGroup assets;
    bool needLayout = false;
    bool variantsEnabled = false;
    GridLayout layout;
    ListData defaultItems;
    Data const *items = nullptr;
    ChildWidgetOrganizer organizer;
    QSet<PanelWidget *> openSubs;
    IndirectRule *outContentHeight;
    SizePolicy colPolicy = Fixed;
    SizePolicy rowPolicy = Fixed;
    Impl(Public *i)
        : Base(i)
        , organizer(*i) {
        outContentHeight = new IndirectRule;
        organizer.setWidgetFactory(*this);
        setContext(&defaultItems);
        self().audienceForChildAddition() += this;
        self().audienceForChildRemoval()  += this;
    }
    ~Impl() {
        releaseRef(outContentHeight);
        defaultItems.clear();
    }
    void setContext(Data const *ctx) {
        if (items) {
            items->audienceForAddition()    -= this;
            items->audienceForRemoval()     -= this;
            items->audienceForOrderChange() -= this;
            organizer.unsetContext();
        }
        items = ctx;
        items->audienceForAddition()    += this;
        items->audienceForRemoval()     += this;
        items->audienceForOrderChange() += this;
        organizer.setContext(*items); // recreates widgets
    }
    void dataItemAdded(Data::Pos, Item const &) {
        needLayout = true;
    }
    void dataItemRemoved(Data::Pos, Item &) {
        needLayout = true;
    }
    void dataItemOrderChanged() {
        needLayout = true;
    }
    void widgetChildAdded(Widget &child) {
        needLayout = true;
        if (IAssetGroup *asset = maybeAs<IAssetGroup>(child)) {
            assets += *asset; // part of the asset group (observes for deletion)
        }
        if (ButtonWidget *button = maybeAs<ButtonWidget>(child)) {
            button->audienceForPress() += this;
        }
    }
    void widgetChildRemoved(Widget &child) {
        needLayout = true;
        if (IAssetGroup *asset = maybeAs<IAssetGroup>(child)) {
            assets -= *asset; // no longer part of the asset group
        }
        if (ButtonWidget *button = maybeAs<ButtonWidget>(child)) {
            button->audienceForPress() -= this;
        }
    }
    static void setFoldIndicatorForDirection(LabelWidget &label, ui::Direction dir) {
        if (dir == ui::Right || dir == ui::Left) {
            label.setImage(new StyleProceduralImage("fold", label, dir == ui::Right? -90 : 90));
            label.setTextAlignment(dir == ui::Right? ui::AlignLeft : ui::AlignRight);
        }
    }
    /*
     * Menu items are represented as buttons and labels.
     */
    GuiWidget *makeItemWidget(Item const &item, GuiWidget const *) {
        if (item.semantics().testFlag(Item::ShownAsButton)) {
            ButtonWidget *b = (item.semantics().testFlag(Item::ShownAsPopupButton)?
                                   new PopupButtonWidget : new ButtonWidget);
            b->setTextAlignment(ui::AlignRight);
            if (is<SubmenuItem>(item)) {
                auto const &subItem = item.as<SubmenuItem>();
                b->setAction(new SubmenuAction(this, subItem));
                setFoldIndicatorForDirection(*b, subItem.openingDirection());
            }
            else if (is<SubwidgetItem>(item)) {
                auto const &subItem = item.as<SubwidgetItem>();
                b->setAction(new SubwidgetAction(this, subItem));
                setFoldIndicatorForDirection(*b, subItem.openingDirection());
                if (subItem.image().isNull()) {
                    setFoldIndicatorForDirection(*b, subItem.openingDirection());
                }
            }
            return b;
        }
        else if (item.semantics().testFlag(Item::Separator)) {
            LabelWidget *lab = new LabelWidget;
            lab->setAlignment(ui::AlignLeft);
            lab->setTextLineAlignment(ui::AlignLeft);
            lab->setSizePolicy(ui::Expand, ui::Expand);
            return lab;
        }
        else if (item.semantics().testFlag(Item::ShownAsLabel)) {
            LabelWidget *lab = new LabelWidget;
            lab->setTextAlignment(ui::AlignRight);
            lab->setTextLineAlignment(ui::AlignLeft);
            lab->setSizePolicy(ui::Expand, ui::Expand);
            return lab;
        }
        else if (item.semantics().testFlag(Item::ShownAsToggle)) {
            if (VariableToggleItem const *varTog = maybeAs<VariableToggleItem>(item)) {
                return new VariableToggleWidget(varTog->variable());
            }
            else {
                return new ToggleWidget;
            }
        }
        return 0;
    }
    void updateItemWidget(GuiWidget &widget, Item const &item) {
        if (ImageItem const *img = maybeAs<ImageItem>(item)) {
            if (LabelWidget *label = maybeAs<LabelWidget>(widget)) {
                if (!img->image().isNull()) {
                    label->setImage(img->image());
                }
            }
        }
        if (ActionItem const *act = maybeAs<ActionItem>(item)) {
            if (item.semantics().testFlag(Item::ShownAsButton)) {
                ButtonWidget &b = widget.as<ButtonWidget>();
                b.setText(act->label());
                if (act->action()) {
                    b.setAction(*act->action());
                }
            }
            else if (item.semantics().testFlag(Item::ShownAsLabel)) {
                widget.as<LabelWidget>().setText(item.label());
            }
            else if (item.semantics().testFlag(Item::ShownAsToggle)) {
                ToggleWidget &t = widget.as<ToggleWidget>();
                t.setText(act->label());
                if (act->action()) {
                    t.setAction(*act->action());
                }
            }
        }
        else {
            widget.as<LabelWidget>().setText(item.label());
        }
    }
    void panelBeingClosed(PanelWidget &popup) {
        popup.audienceForClose()    -= this;
        popup.audienceForDeletion() -= this;
        openSubs.remove(&popup);
    }
    void widgetBeingDeleted(Widget &widget) {
        openSubs.remove(static_cast<PanelWidget *>(&widget));
    }
    void keepTrackOfSubWidget(PanelWidget *w) {
        DENG2_ASSERT(is<PanelWidget>(w));
        openSubs.insert(w);
        w->audienceForClose()    += this;
        w->audienceForDeletion() += this;
        emit self().subWidgetOpened(w);
        foreach (auto *panel, openSubs) {
            if (panel != w) panel->close();
        }
    }
    bool isVisibleItem(GuiWidget const *child) const {
        if (child) {
            return !child->behavior().testFlag(Widget::Hidden);
        }
        return false;
    }
    int countVisible() const {
        int num = 0;
        foreach (GuiWidget *i, self().childWidgets()) {
            if (isVisibleItem(i)) ++num;
        }
        return num;
    }
    void relayout() {
        layout.clear();
        if (organizer.virtualizationEnabled()) {
            layout.setLeftTop(self().contentRule().left(),
                              self().contentRule().top() + organizer.virtualStrut());
        }
        foreach (GuiWidget *child, self().childWidgets()) {
            if (isVisibleItem(child)) {
                layout << *child;
            }
        }
    }
    Rule const &contentHeight() const {
        if (organizer.virtualizationEnabled()) {
            return OperatorRule::maximum(organizer.estimatedTotalHeight(),
                                         organizer.virtualStrut() + layout.height());
        }
        return layout.height();
    }
    void buttonPressed(ButtonWidget &button) {
        DENG2_FOR_PUBLIC_AUDIENCE2(ItemTriggered, i) {
            if (auto *item = organizer.findItemForWidget(button)) {
                i->menuItemTriggered(*item);
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(ItemTriggered)
};
DENG2_AUDIENCE_METHOD(MenuWidget, ItemTriggered)
MenuWidget::MenuWidget(String const &name)
    : ScrollAreaWidget(name), d(new Impl(this)) {
    setBehavior(ChildVisibilityClipping, UnsetFlags);
}
AssetGroup &MenuWidget::assets() {
    return d->assets;
}
void MenuWidget::setGridSize(int columns, ui::SizePolicy columnPolicy,
                             int rows, ui::SizePolicy rowPolicy,
                             GridLayout::Mode layoutMode) {
    d->layout.clear();
    d->layout.setModeAndGridSize(layoutMode, columns, rows);
    d->layout.setLeftTop(contentRule().left(), contentRule().top());
    d->colPolicy = columnPolicy;
    d->rowPolicy = rowPolicy;
    if (d->colPolicy == ui::Filled) {
        DENG2_ASSERT(columns > 0);
        d->layout.setOverrideWidth((rule().width() - margins().width() -
                                    (columns - 1) * d->layout.columnPadding()) / float(columns));
    }
    if (d->rowPolicy == ui::Filled) {
        DENG2_ASSERT(rows > 0);
        d->layout.setOverrideHeight((rule().height() - margins().height() -
                                     (rows - 1) * d->layout.rowPadding()) / float(rows));
    }
    d->needLayout = true;
}
Data &MenuWidget::items() {
    return *const_cast<Data *>(d->items);
}
Data const &MenuWidget::items() const {
    return *d->items;
}
void MenuWidget::setItems(Data const &items) {
    d->setContext(&items);
}
void MenuWidget::useDefaultItems() {
    d->setContext(&d->defaultItems);
}
bool MenuWidget::isUsingDefaultItems() const {
    return d->items == &d->defaultItems;
}
int MenuWidget::count() const {
    return d->countVisible();
}
bool MenuWidget::isWidgetPartOfMenu(GuiWidget const &widget) const {
    if (widget.parentWidget() != this) return false;
    return d->isVisibleItem(&widget);
}
void MenuWidget::updateLayout() {
    d->relayout();
    d->outContentHeight->setSource(d->contentHeight());
    setContentSize(d->layout.width(), *d->outContentHeight);
    if (d->colPolicy == Expand) {
        rule().setInput(Rule::Width, d->layout.width() + margins().width());
    }
    if (d->rowPolicy == Expand) {
        rule().setInput(Rule::Height, *d->outContentHeight + margins().height());
    }
    d->needLayout = false;
}
GridLayout const &MenuWidget::layout() const {
    return d->layout;
}
GridLayout &MenuWidget::layout() {
    return d->layout;
}
Rule const &MenuWidget::contentHeight() const {
    return *d->outContentHeight;
}
void MenuWidget::offerFocus() {
    foreach (GuiWidget *widget, childWidgets()) {
        if (!widget->behavior().testFlag(Hidden) &&
            widget->behavior().testFlag(Focusable)) {
            root().setFocus(widget);
            return;
        }
    }
}
ChildWidgetOrganizer &MenuWidget::organizer() {
    return d->organizer;
}
ChildWidgetOrganizer const &MenuWidget::organizer() const {
    return d->organizer;
}
void MenuWidget::setVirtualizationEnabled(bool enabled, int averageItemHeight) {
    d->organizer.setVirtualizationEnabled(enabled);
    d->organizer.setAverageChildHeight(averageItemHeight);
    d->organizer.setVirtualTopEdge(contentRule().top());
    d->organizer.setVisibleArea(rule().top(), rule().bottom());
    d->needLayout = true;
}
void MenuWidget::setVariantItemsEnabled(bool variantsEnabled) {
    if (d->variantsEnabled != variantsEnabled) {
        d->variantsEnabled = variantsEnabled;
        items().forAll([this] (ui::Item const &item) {
            if (is<ui::VariantActionItem>(item)) {
                item.notifyChange();
            }
            return LoopContinue;
        });
    }
}
bool MenuWidget::variantItemsEnabled() const {
    return d->variantsEnabled;
}
ui::DataPos MenuWidget::findItem(GuiWidget const &widget) const {
    if (auto const *item = organizer().findItemForWidget(widget)) {
        return items().find(*item);
    }
    return ui::Data::InvalidPos;
}
void MenuWidget::update() {
    if (d->organizer.virtualizationEnabled()) {
        d->organizer.updateVirtualization();
    }
    if (d->needLayout) {
        updateLayout();
    }
    ScrollAreaWidget::update();
}
bool MenuWidget::handleEvent(Event const &event) {
    if (event.isKeyDown() && root().focus() && root().focus()->parentWidget() == this) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (key.ddKey() == DDKEY_UPARROW || key.ddKey() == DDKEY_DOWNARROW) {
            root().focusIndicator().fadeIn();
            auto const children = childWidgets();
            for (int ordinal = children.indexOf(root().focus());
                 ordinal >= 0 && ordinal < int(childCount());
                 ordinal += (key.ddKey() == DDKEY_UPARROW? -1 : +1)) {
                auto *child = children.at(ordinal);
                if (child->hasFocus() || child->isDisabled()) continue;
                if (child->isVisible() && child->behavior().testFlag(Focusable)) {
                    root().setFocus(child);
                    findTopmostScrollable().scrollToWidget(*child);
                    return true;
                }
            }
        }
    }
    return ScrollAreaWidget::handleEvent(event);
}
void MenuWidget::dismissPopups() {
    foreach (PanelWidget *pop, d->openSubs) {
        pop->close();
    }
}
void MenuWidget::updateStyle() {
    ScrollAreaWidget::updateStyle();
    updateLayout();
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(NotificationAreaWidget)
, DENG2_OBSERVES(Widget, Deletion) {
    static TimeSpan const ANIM_SPAN;
    AnimationRule *shift;
    QMap<GuiWidget *, RelayWidget *> shown;
    QTimer dismissTimer;
    QList<GuiWidget *> pendingDismiss;
    Impl(Public *i) : Base(i) {
        dismissTimer.setSingleShot(true);
        dismissTimer.setInterval(ANIM_SPAN.asMilliSeconds());
        QObject::connect(&dismissTimer, SIGNAL(timeout()), thisPublic, SLOT(dismiss()));
        shift = new AnimationRule(0);
    }
    ~Impl() {
#ifdef DENG2_DEBUG
        foreach (GuiWidget *w, shown.keys()) {
            DENG2_ASSERT(w->audienceForDeletion().contains(this));
        }
#endif
        releaseRef(shift);
    }
    void updateChildLayout() {
        Rule const &gap = rule(RuleBank::UNIT);
        SequentialLayout layout(self().rule().right(), self().rule().top(), ui::Left);
        bool first = true;
        foreach (GuiWidget *child, self().childWidgets()) {
            GuiWidget *w = child->as<RelayWidget>().target();
            DENG2_ASSERT(w != nullptr);
            if (!first) {
                layout << gap;
            }
            first = false;
            layout << *w;
        }
        self().rule().setSize(layout.width(), layout.height());
    }
    void show() {
        shift->set(0, ANIM_SPAN);
        shift->setStyle(Animation::EaseOut);
        self().show();
    }
    void hide(TimeSpan const &span = ANIM_SPAN) {
        shift->set(self().rule().height() + rule("gap"), span);
        shift->setStyle(Animation::EaseIn);
    }
    void removeChild(GuiWidget &notif) {
        DENG2_ASSERT(shown.contains(&notif));
        auto *relay = shown.take(&notif);
        /*
         * Can't destroy the relay immediately because both the relay and we are
         * observing the notification for deletion and we don't know if the relay
         * will still be notified after this.
         */
        self().remove(*relay);
        GuiWidget::destroyLater(relay);
        if (!self().childCount()) {
            self().hide();
        }
        updateChildLayout();
    }
    void dismissChild(GuiWidget &notif) {
        DENG2_ASSERT(notif.audienceForDeletion().contains(this));
        notif.audienceForDeletion() -= this;
        removeChild(notif);
        notif.deinitialize();
        notif.setRoot(nullptr);
    }
    void performPendingDismiss() {
        dismissTimer.stop();
        foreach (GuiWidget *w, pendingDismiss) {
            dismissChild(*w);
        }
        pendingDismiss.clear();
    }
    void widgetBeingDeleted(Widget &notif) {
        GuiWidget *w = static_cast<GuiWidget *>(&notif);
        pendingDismiss.removeAll(w);
        removeChild(*w);
    }
};
TimeSpan const NotificationAreaWidget::Impl::ANIM_SPAN = 0.5;
NotificationAreaWidget::NotificationAreaWidget(String const &name)
    : GuiWidget(name)
    , d(new Impl(this)) {
    rule().setSize(Const(0), Const(0));
    d->shift->set(style().fonts().font("default").height().valuei() +
                  rule("gap").valuei() * 3);
    hide();
}
void NotificationAreaWidget::useDefaultPlacement(RuleRectangle const &area,
                                                 Rule const &horizontalOffset) {
    rule().setInput(Rule::Top,   area.top() + rule("gap") - shift())
          .setInput(Rule::Right, area.right() - rule("gap") + horizontalOffset);
}
Rule const &NotificationAreaWidget::shift() {
    return *d->shift;
}
void NotificationAreaWidget::showChild(GuiWidget &notif) {
    if (isChildShown(notif)) {
        return;
    }
    d->performPendingDismiss();
    notif.setRoot(&root());
    notif.audienceForDeletion() += d;
    notif.set(Background(style().colors().colorf("background")));
    auto *relay = new RelayWidget(&notif);
    d->shown.insert(&notif, relay);
    add(relay);
    relay->initialize();
    d->updateChildLayout();
    d->show();
}
void NotificationAreaWidget::hideChild(GuiWidget &notif) {
    if (!isChildShown(notif)) {
        return;
    }
    if (childCount() > 1) {
        d->dismissChild(notif);
    }
    else {
        d->dismissTimer.start();
        d->pendingDismiss << &notif;
        d->hide();
    }
}
void NotificationAreaWidget::dismiss() {
    d->performPendingDismiss();
}
bool NotificationAreaWidget::isChildShown(GuiWidget &notif) const {
    if (d->pendingDismiss.contains(&notif)) {
        return false;
    }
    return d->shown.contains(&notif);
}
} // namespace de
namespace de {
constexpr double OPENING_ANIM_SPAN           = 0.3;
constexpr double OPENING_ANIM_SPAN_EASED_OUT = 0.18;
constexpr double CLOSING_ANIM_SPAN           = 0.22;
DENG_GUI_PIMPL(PanelWidget)
, DENG2_OBSERVES(Asset, StateChange) {
    bool waitForContentReady = true;
    bool eatMouseEvents = true;
    bool opened = false;
    ui::Direction dir = ui::Down;
    ui::SizePolicy secondaryPolicy = ui::Expand;
    GuiWidget *content = nullptr;
    AnimationRule *openingRule;
    AnimationStyle openingStyle = EasedOut;
    QTimer dismissTimer;
    std::unique_ptr<AssetGroup> pendingShow;
    GuiVertexBuilder verts;
    Impl(Public *i) : Base(i) {
        openingRule = new AnimationRule(0);
        openingRule->setBehavior(AnimationRule::RestartWhenTargetChanges);
        dismissTimer.setSingleShot(true);
        QObject::connect(&dismissTimer, SIGNAL(timeout()), thisPublic, SLOT(dismiss()));
    }
    ~Impl() {
        releaseRef(openingRule);
    }
    void glInit() {}
    void glDeinit() {
        verts.clear();
    }
    bool isVerticalAnimation() const {
        return isVertical(dir) || dir == ui::NoDirection;
    }
    void updateLayout() {
        if (isVerticalAnimation()) {
            self().rule().setInput(Rule::Height, *openingRule);
            if (secondaryPolicy == ui::Expand) {
                self().rule().setInput(Rule::Width, content ? content->rule().width() : Const(0));
            }
        }
        else {
            self().rule().setInput(Rule::Width, *openingRule);
            if (secondaryPolicy == ui::Expand) {
                self().rule().setInput(Rule::Height, content ? content->rule().height() : Const(0));
            }
        }
    }
    void updateGeometry() {
        Rectanglei pos;
        if (self().hasChangedPlace(pos) || self().geometryRequested()) {
            self().requestGeometry(false);
            verts.clear();
            self().glMakeGeometry(verts);
        }
    }
    void startOpeningAnimation(TimeSpan span) {
        if (isVerticalAnimation()) {
            openingRule->set(content->rule().height(), span);
        }
        else {
            openingRule->set(content->rule().width(), span);
        }
        switch (openingStyle) {
        case Bouncy:   openingRule->setStyle(Animation::Bounce, 12); break;
        case EasedOut: openingRule->setStyle(Animation::EaseOutSofter); break;
        case Smooth:   openingRule->setStyle(Animation::EaseBoth); break;
        }
    }
    void close(TimeSpan delay) {
        if (!opened) return;
        opened = false;
        self().setBehavior(DisableEventDispatchToChildren);
        openingRule->set(0, CLOSING_ANIM_SPAN + delay, delay);
        switch (openingStyle) {
        case Bouncy:
        case EasedOut:
            openingRule->setStyle(Animation::EaseIn);
            break;
        case Smooth:
            openingRule->setStyle(Animation::EaseBoth);
            break;
        }
        self().panelClosing();
        DENG2_FOR_PUBLIC_AUDIENCE2(Close, i) {
            i->panelBeingClosed(self());
        }
        emit self().closed();
        dismissTimer.start();
        dismissTimer.setInterval(TimeSpan(CLOSING_ANIM_SPAN + delay).asMilliSeconds());
    }
    void waitForAssetsInContent() {
        if (!waitForContentReady) return;
        LOG_AS("PanelWidget");
        pendingShow.reset(new AssetGroup);
        LOGDEV_XVERBOSE("Checking for assets that need waiting for...", "");
        DENG2_ASSERT(content);
        content->collectNotReadyAssets(*pendingShow);
        if (pendingShow->isEmpty()) {
            pendingShow.reset();
            return;
        }
        LOGDEV_VERBOSE("Waiting for %i assets to become ready") << pendingShow->size();
        pendingShow->audienceForStateChange() += this;
        openingRule->pause();
    }
    void assetStateChanged(Asset &) {
        LOG_AS("PanelWidget");
        if (pendingShow && pendingShow->isReady()) {
            LOGDEV_XVERBOSE("All assets ready, resuming animation", "");
            openingRule->resume();
            pendingShow->audienceForStateChange() -= this;
            trash(pendingShow.release());
        }
    }
    DENG2_PIMPL_AUDIENCE(AboutToOpen)
    DENG2_PIMPL_AUDIENCE(Close)
};
DENG2_AUDIENCE_METHOD(PanelWidget, AboutToOpen)
DENG2_AUDIENCE_METHOD(PanelWidget, Close)
PanelWidget::PanelWidget(String const &name) : GuiWidget(name), d(new Impl(this)) {
    setBehavior(ChildHitClipping);
    setBehavior(ChildVisibilityClipping);
    d->updateLayout(); // initial, empty layout
    hide();
}
void PanelWidget::setWaitForContentReady(bool yes) {
    d->waitForContentReady = yes;
}
void PanelWidget::setAnimationStyle(AnimationStyle style) {
    d->openingStyle = style;
}
void PanelWidget::setEatMouseEvents(bool yes) {
    d->eatMouseEvents = yes;
}
void PanelWidget::setContent(GuiWidget *content) {
    if (d->content) {
        destroy(takeContent());
    }
    d->content = content;
    add(content); // ownership taken
    content->rule()
            .setInput(Rule::Left, rule().left())
            .setInput(Rule::Top,  rule().top());
}
GuiWidget &PanelWidget::content() const {
    DENG2_ASSERT(d->content != 0);
    return *d->content;
}
GuiWidget *PanelWidget::takeContent() {
    GuiWidget *w = d->content;
    if (!w) return 0;
    d->content = 0;
    w->rule().clearInput(Rule::Left);
    w->rule().clearInput(Rule::Top);
    if (d->secondaryPolicy == ui::Expand) {
        rule().clearInput(Rule::Width);
        rule().clearInput(Rule::Height);
    }
    remove(*w);
    return w;
}
void PanelWidget::setOpeningDirection(ui::Direction dir) {
    d->dir = dir;
}
void PanelWidget::setSizePolicy(ui::SizePolicy policy) {
    d->secondaryPolicy = policy;
}
ui::Direction PanelWidget::openingDirection() const {
    return d->dir;
}
bool PanelWidget::isOpen() const {
    return d->opened;
}
bool PanelWidget::isOpeningOrClosing() const {
    return !d->openingRule->animation().done();
}
void PanelWidget::close(TimeSpan delayBeforeClosing) {
    d->close(delayBeforeClosing);
}
void PanelWidget::viewResized() {
    GuiWidget::viewResized();
    requestGeometry();
}
void PanelWidget::update() {
    GuiWidget::update();
}
bool PanelWidget::handleEvent(Event const &event) {
    if (d->eatMouseEvents && event.type() == Event::MouseButton) {
        MouseEvent const &mouse = event.as<MouseEvent>();
        if (hitTest(mouse.pos())) {
            return true;
        }
    }
    return GuiWidget::handleEvent(event);
}
void PanelWidget::open() {
    if (d->opened) return;
    DENG2_FOR_AUDIENCE2(AboutToOpen, i) {
        i->panelAboutToOpen(*this);
    }
    d->dismissTimer.stop();
    unsetBehavior(DisableEventDispatchToChildren);
    show();
    preparePanelForOpening();
    d->startOpeningAnimation(d->openingStyle == EasedOut ? OPENING_ANIM_SPAN_EASED_OUT
                                                         : OPENING_ANIM_SPAN);
    d->opened = true;
    emit opened();
    d->waitForAssetsInContent();
}
void PanelWidget::close() {
    d->close(0.2);
}
void PanelWidget::openOrClose() {
    if (isOpen() || isOpeningOrClosing()) {
        close();
    }
    else {
        open();
    }
}
void PanelWidget::dismiss() {
    if (isHidden()) return;
    root().window().glActivate();
    hide();
    d->opened = false;
    d->dismissTimer.stop();
    panelDismissed();
    emit dismissed();
}
void PanelWidget::drawContent() {
    d->updateGeometry();
    if (d->verts) {
        auto &painter = root().painter();
        painter.setColor(Vector4f(1, 1, 1, 1));
        painter.drawTriangleStrip(d->verts);
    }
}
void PanelWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    GuiWidget::glMakeGeometry(verts);
}
void PanelWidget::glInit() {
    d->glInit();
}
void PanelWidget::glDeinit() {
    d->glDeinit();
}
void PanelWidget::preparePanelForOpening() {
    d->updateLayout();
}
void PanelWidget::panelClosing() {
}
void PanelWidget::panelDismissed() {
}
} // namespace de
namespace de {
DENG2_PIMPL(PopupButtonWidget)
, DENG2_OBSERVES(ButtonWidget, StateChange)
, DENG2_OBSERVES(ButtonWidget, Press)
, DENG2_OBSERVES(PanelWidget,  Close) {
    SafeWidgetPtr<PopupWidget> pop;
    Constructor constructor;
    Opener opener;
    ui::Direction direction;
    bool popupWasOpenWhenButtonWentDown = false;
    Impl(Public *i) : Base(i) {}
    void buttonStateChanged(ButtonWidget &, State state) {
        if (state == Down) {
            popupWasOpenWhenButtonWentDown =
                    (pop && (pop->isOpen() || pop->isOpeningOrClosing()));
        }
    }
    void buttonPressed(ButtonWidget &) {
        if (!popupWasOpenWhenButtonWentDown) {
            if (constructor) {
                pop.reset(constructor(self()));
                self().add(pop);
                self().setOpeningDirection(direction);
                pop->setDeleteAfterDismissed(true);
            }
            if (opener) {
                opener(pop);
            }
            else if (pop) {
                pop->open();
            }
            if (auto *parentPop = self().findParentPopup()) {
                parentPop->audienceForClose() += this;
            }
        }
    }
    void panelBeingClosed(PanelWidget &) {
        if (self().isOpen()) pop->close();
    }
};
PopupButtonWidget::PopupButtonWidget(String const &name)
    : ButtonWidget(name)
    , d(new Impl(this)) {
    audienceForStateChange() += d;
    audienceForPress()       += d;
}
void PopupButtonWidget::setPopup(PopupWidget &popup, ui::Direction openingDirection) {
    d->pop.reset(&popup);
    d->constructor = Constructor();
    d->direction = openingDirection;
    setOpeningDirection(openingDirection);
}
void PopupButtonWidget::setOpener(Opener opener) {
    d->opener = opener;
}
void PopupButtonWidget::setPopup(Constructor makePopup, ui::Direction openingDirection) {
    d->pop.reset();
    d->direction = openingDirection;
    d->constructor = makePopup;
}
void PopupButtonWidget::setOpeningDirection(ui::Direction direction) {
    d->direction = direction;
    if (d->pop) {
        d->pop->setAnchorAndOpeningDirection(rule(), direction);
    }
}
PopupWidget *PopupButtonWidget::popup() const {
    return d->pop;
}
bool PopupButtonWidget::isOpen() const {
    if (d->pop) {
        return d->pop->isOpen();
    }
    return false;
}
} // namespace de
namespace de {
static String const VAR_SHOW_ANNOTATIONS("ui.showAnnotations");
DENG_GUI_PIMPL(PopupMenuWidget)
, DENG2_OBSERVES(ButtonWidget, StateChange)
, DENG2_OBSERVES(ButtonWidget, Triggered)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate)
, DENG2_OBSERVES(Variable, Change) {
    class HeadingOverlayImage : public ProceduralImage {
    public:
        HeadingOverlayImage(GuiWidget &owner)
            : _owner(owner)
            , _id(Id::None) {
            if (_owner.hasRoot()) {
                alloc();
            }
        }
        GuiRootWidget &root() {
            return _owner.root();
        }
        void alloc() {
            _id = root().solidWhitePixel();
            setPointSize({1, 1});
        }
        void glInit() override {
            if (_id.isNone()) {
                alloc();
            }
        }
        void glDeinit() override {
            _id = Id::None;
        }
        void glMakeGeometry(GuiVertexBuilder &verts, Rectanglef const &rect) override {
            if (!_id.isNone()) {
                Rectanglef visible = rect;
                visible.setWidth(_owner.rule().width().value());
                verts.makeQuad(visible, color(), root().atlas().imageRectf(_id));
            }
        }
    private:
        GuiWidget &_owner;
        Id _id;
    };
    ColorTheme colorTheme = Normal;
    ButtonWidget const *hover;
    int oldScrollY;
    Rule const *widestItem;
    IndirectRule *maxItemWidth;
    SafeWidgetPtr<PopupWidget> parentPopup;
    Impl(Public *i)
        : Base(i)
        , hover(0)
        , oldScrollY(0)
        , widestItem(0)
        , maxItemWidth(0) {
        maxItemWidth = new IndirectRule;
        App::config(VAR_SHOW_ANNOTATIONS).audienceForChange() += this;
    }
    ~Impl() {
        releaseRef(maxItemWidth);
        releaseRef(widestItem);
    }
    void addToMaxWidth(GuiWidget &widget) {
        maxInto(widestItem, widget.rule().width());
        maxItemWidth->setSource(*widestItem);
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &item) {
        widget.set(Background());
        if (item.semantics().testFlag(ui::Item::Separator)) {
            LabelWidget &lab = widget.as<LabelWidget>();
            lab.setTextColor(item.semantics().testFlag(ui::Item::Annotation)? "label.altaccent" :
                                                                              "label.accent");
            lab.setMaximumTextWidth(*maxItemWidth);
            lab.rule().setInput(Rule::Width, *maxItemWidth);
            return;
        }
        if (LabelWidget *lab = maybeAs<LabelWidget>(widget)) {
            lab->margins().set("popup.menu.margin");
            lab->setMaximumTextWidth(rule("popup.menu.width.max"));
            lab->setTextLineAlignment(ui::AlignLeft);
            addToMaxWidth(widget);
        }
        if (ButtonWidget *b = maybeAs<ButtonWidget>(widget)) {
            addToMaxWidth(widget);
            setButtonColors(*b);
            b->setSizePolicy(ui::Expand, ui::Expand);
            b->audienceForStateChange() += this;
            if (item.semantics().testFlag(ui::Item::ActivationClosesPopup)) {
                b->audienceForTriggered() += this;
            }
        }
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) {
        if (ButtonWidget *b = maybeAs<ButtonWidget>(widget)) {
            if (!is<ToggleWidget>(b)) {
                b->setTextGap("dialog.gap");
                b->setOverrideImageSize(style().fonts().font("default").height());
            }
        }
        if (item.semantics().testFlag(ui::Item::Annotation)) {
            if (!App::config().getb(VAR_SHOW_ANNOTATIONS)) {
                widget.hide();
            }
            widget.margins().set("halfunit").setLeft("popup.menu.margin");
            widget.setFont("separator.annotation");
        }
        else if (item.semantics().testFlag(ui::Item::Separator)) {
            if (item.label().isEmpty()) {
                widget.margins().set("");
                widget.setFont("separator.empty");
                widget.as<LabelWidget>().setOverlayImage(0);
            }
            else {
                widget.margins().set("halfunit").setLeft("popup.menu.margin");
                widget.setFont("separator.label");
                /*
                LabelWidget &lab = widget.as<LabelWidget>();
                HeadingOverlayImage *img = new HeadingOverlayImage(widget);
                img->setColor(style().colors().colorf("accent"));
                lab.setOverlayImage(img, ui::AlignBottomLeft);
                */
            }
        }
    }
    void setButtonColors(ButtonWidget &button) {
        /*bool const hovering = (hover == &button);
        button.setTextColor(!hovering ^ (colorTheme == Inverted)? "text" : "inverted.text");
        button.setHoverTextColor(!hovering ^ (colorTheme == Inverted)? "inverted.text" : "text",
                                 ButtonWidget::ReplaceColor);*/
        button.setTextColor     (colorTheme == Normal? "text"          : "inverted.text" );
        button.setHoverTextColor(colorTheme == Normal? "inverted.text" : "text",         ButtonWidget::ReplaceColor);
    }
    void updateItemHitRules() {
        GridLayout const &layout = self().menu().layout();
        AutoRef<Rule const> halfUnit = self().rule("halfunit");
        foreach (GuiWidget *widget, self().menu().childWidgets()) {
            if (self().menu().isWidgetPartOfMenu(*widget)) {
                Vector2i cell = layout.widgetPos(*widget);
                DENG2_ASSERT(cell.x >= 0 && cell.y >= 0);
                widget->hitRule()
                        .setInput(Rule::Left,  (!cell.x?
                                                    self().rule().left()
                                                  : layout.columnLeft(cell.x)) + halfUnit)
                        .setInput(Rule::Right, ((cell.x == layout.gridSize().x - 1)?
                                                    self().rule().right()
                                                  : layout.columnRight(cell.x)) - halfUnit);
            }
        }
    }
    bool hasButtonsWithImages() const {
        foreach (GuiWidget *child, self().menu().childWidgets()) {
            if (ButtonWidget *button = maybeAs<ButtonWidget>(child)) {
                if (button->hasImage() && button->textAlignment() == ui::AlignRight)
                    return true;
            }
        }
        return false;
    }
    /**
     * Adjusts the left margins of clickable items so that icons are aligned by their
     * text, with the possible icon hanging on the left. If there are no items with
     * icons, no extra padding is applied.
     */
    void updateItemMargins() {
        bool const useExtraPadding = hasButtonsWithImages();
        auto const &padding = rule("popup.menu.paddedmargin");
        auto const &none    = rule("popup.menu.margin");
        foreach (GuiWidget *widget, self().menu().childWidgets()) {
            if (LabelWidget *label = maybeAs<LabelWidget>(widget)) {
                ui::Item const *item = self().menu().organizer().findItemForWidget(*widget);
                if (item->semantics().testFlag(ui::Item::Annotation)) {
                    if (useExtraPadding) {
                        label->setMaximumTextWidth(*maxItemWidth - padding);
                        widget->margins().setLeft(padding);
                    }
                    else {
                        label->setMaximumTextWidth(*maxItemWidth);
                        widget->margins().setLeft(none);
                    }
                }
            }
            if (ButtonWidget *button = maybeAs<ButtonWidget>(widget)) {
                updateImageColor(*button);
                if (useExtraPadding) {
                    Rule const *padRule = holdRef(padding);
                    if (button->hasImage() && button->textAlignment() == ui::AlignRight) {
                        LabelWidget::ContentLayout layout;
                        button->contentLayout(layout);
                        sumInto(padRule, -Const(layout.image.width()) -
                                rule(button->textGap()));
                    }
                    widget->margins().setLeft(*padRule);
                    releaseRef(padRule);
                }
                else {
                    widget->margins().setLeft(none);
                }
            }
        }
    }
    void updateImageColor(ButtonWidget &button, bool invert = false) {
        button.setImageColor(style().colors().colorf(invert ^ (colorTheme == Inverted)? "inverted.text"
                                                                                      : "text"));
    }
    void buttonStateChanged(ButtonWidget &button, ButtonWidget::State state) {
        updateImageColor(button, state != ButtonWidget::Up);
        if (&button == hover && state == ButtonWidget::Up) {
            hover = 0;
            self().requestGeometry();
            return;
        }
        if (state == ButtonWidget::Hover || state == ButtonWidget::Down) {
            hover = &button;
            self().requestGeometry();
        }
    }
    Rectanglei highlightRect() const {
        Rectanglei hi;
        if (hover) {
            hi.topLeft.x     = hover->hitRule().left().valuei();
            hi.topLeft.y     = hover->hitRule().top().valuei();
            hi.bottomRight.x = hover->hitRule().right().valuei();
            hi.bottomRight.y = hover->hitRule().bottom().valuei();
        }
        return hi & self().rule().recti();
    }
    void buttonActionTriggered(ButtonWidget &) {
        self().close();
        if (parentPopup) parentPopup->close();
    }
    void updateIfScrolled() {
        int scrollY = self().menu().scrollPositionY().valuei();
        if (scrollY == oldScrollY) {
            return;
        }
        oldScrollY = scrollY;
        root().dispatchLatestMousePosition();
        self().requestGeometry();
    }
    void variableValueChanged(Variable &, Value const &newValue) {
        bool changed = false;
        self().items().forAll([this, &newValue, &changed] (ui::Item const &item) {
            if (item.semantics().testFlag(ui::Item::Annotation)) {
                self().menu().itemWidget<GuiWidget>(item).show(newValue.isTrue());
                changed = true;
            }
            return LoopContinue;
        });
        if (changed) {
            self().menu().updateLayout();
        }
    }
    void updateButtonColors() {
        for (GuiWidget *w : self().menu().childWidgets()) {
            if (ButtonWidget *btn = maybeAs<ButtonWidget>(w)) {
                setButtonColors(*btn);
            }
        }
    }
    void updateLayout() {
        auto &menu = self().menu();
        menu.updateLayout();
        menu.rule().setInput(
            Rule::Height,
            OperatorRule::minimum(menu.rule().inputRule(Rule::Height),
                                  root().viewHeight() - self().margins().height()));
        updateItemHitRules();
        updateItemMargins();
    }
};
PopupMenuWidget::PopupMenuWidget(String const &name)
    : PopupWidget(name), d(new Impl(this)) {
    setContent(new MenuWidget(name.isEmpty()? "" : name + "-content"));
    setOutlineColor("popup.outline");
    menu().setGridSize(1, ui::Expand, 0, ui::Expand);
    menu().organizer().audienceForWidgetCreation() += d;
    menu().organizer().audienceForWidgetUpdate() += d;
}
void PopupMenuWidget::setParentPopup(PopupWidget *parentPopup) {
    d->parentPopup.reset(parentPopup);
}
PopupWidget *PopupMenuWidget::parentPopup() const {
    return d->parentPopup;
}
MenuWidget &PopupMenuWidget::menu() const {
    return static_cast<MenuWidget &>(content());
}
void PopupMenuWidget::useInfoStyle(bool yes) {
    setColorTheme(yes? Inverted : Normal);
}
void PopupMenuWidget::setColorTheme(ColorTheme theme) {
    PopupWidget::setColorTheme(theme);
    d->colorTheme = theme;
    d->updateButtonColors();
}
void PopupMenuWidget::offerFocus() {
    menu().offerFocus();
}
void PopupMenuWidget::update() {
    PopupWidget::update();
    d->updateIfScrolled();
}
void PopupMenuWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    PopupWidget::glMakeGeometry(verts);
    if (d->hover && d->hover->isEnabled()) {
        verts.makeQuad(d->highlightRect(),
                       d->hover->state() == ButtonWidget::Hover?
                           style().colors().colorf(d->colorTheme == Normal? "inverted.background" : "background") :
                           style().colors().colorf(d->colorTheme == Normal? "accent" : "inverted.accent"),
                       root().atlas().imageRectf(root().solidWhitePixel()).middle());
    }
}
void PopupMenuWidget::preparePanelForOpening() {
    d->updateLayout();
    PopupWidget::preparePanelForOpening();
}
void PopupMenuWidget::panelClosing() {
    PopupWidget::panelClosing();
    if (d->hover) {
        auto &btn = *const_cast<ButtonWidget *>(d->hover);
        d->hover = 0;
        btn.setState(ButtonWidget::Up);
        d->updateImageColor(btn);
        requestGeometry();
    }
    menu().dismissPopups();
}
void PopupMenuWidget::updateStyle() {
    PopupWidget::updateStyle();
    for (ui::DataPos i = 0; i < menu().items().size(); ++i) {
        menu().items().at(i).notifyChange();
    }
    d->updateLayout();
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(PopupWidget) {
    ColorTheme colorTheme = Normal;
    bool flexibleDir = true;
    bool deleteAfterDismiss = false;
    bool clickToClose = true;
    bool outsideClickOngoing = false;
    DotPath outlineColorId;
    ColorBank::Colorf outlineColor;
    SafeWidgetPtr<Widget> realParent;
    RuleRectangle anchor;
    Rule const *marker;
    ButtonWidget *close = nullptr;
    Impl(Public *i) : Base(i) {
        marker = &rule("gap");
    }
    void flipOpeningDirectionIfNeeded() {
        ui::Direction openDir = self().openingDirection();
        switch (openDir) {
        case ui::Up:
            if (anchor.midY().value() < self().root().viewHeight().value()/2) {
                openDir = ui::Down;
            }
            break;
        case ui::Down:
            if (anchor.midY().value() > self().root().viewHeight().value()/2) {
                openDir = ui::Up;
            }
            break;
        case ui::Left:
            if (anchor.midX().value() < self().root().viewWidth().value()/2) {
                openDir = ui::Right;
            }
            break;
        case ui::Right:
            if (anchor.midX().value() > self().root().viewWidth().value()/2) {
                openDir = ui::Left;
            }
            break;
        default:
            break;
        }
        self().setOpeningDirection(openDir);
    }
    typedef Vector2<Rule const *> Vector2R;
    Vector2R anchorRule() const {
        switch (self().openingDirection()) {
        case ui::Up:
            return Vector2R(&anchor.midX(), &anchor.top());
        case ui::Down:
            return Vector2R(&anchor.midX(), &anchor.bottom());
        case ui::Left:
            return Vector2R(&anchor.left(), &anchor.midY());
        case ui::Right:
            return Vector2R(&anchor.right(), &anchor.midY());
        default:
            break;
        }
        return Vector2R(&anchor.midX(), &anchor.midY());
    }
    Vector2i anchorPos() const {
        auto rule = anchorRule();
        return Vector2i(rule.x->valuei(), rule.y->valuei());
    }
    void updateLayout() {
        self().rule()
                .clearInput(Rule::Left)
                .clearInput(Rule::Right)
                .clearInput(Rule::Top)
                .clearInput(Rule::Bottom)
                .clearInput(Rule::AnchorX)
                .clearInput(Rule::AnchorY);
        auto anchorPos = anchorRule();
        switch (self().openingDirection()) {
        case ui::Up:
            self().rule()
                    .setInput(Rule::Bottom, OperatorRule::maximum(
                                  *anchorPos.y - *marker,
                                  self().rule().height()))
                    .setInput(Rule::Left, OperatorRule::clamped(
                                  *anchorPos.x - self().rule().width() / 2,
                                  self().margins().left(),
                                  self().root().viewWidth() - self().rule().width() - self().margins().right()));
            break;
        case ui::Down:
            self().rule()
                    .setInput(Rule::Top,  OperatorRule::minimum(
                                  *anchorPos.y + *marker,
                                  self().root().viewHeight() - self().rule().height() - self().margins().bottom()))
                    .setInput(Rule::Left, OperatorRule::clamped(
                                  *anchorPos.x - self().rule().width() / 2,
                                  self().margins().left(),
                                  self().root().viewWidth() - self().rule().width() - self().margins().right()));
            break;
        case ui::Left:
            self().rule()
                    .setInput(Rule::Right, OperatorRule::maximum(
                                  *anchorPos.x - *marker,
                                  self().rule().width()))
                    .setInput(Rule::Top, OperatorRule::clamped(
                                  *anchorPos.y - self().rule().height() / 2,
                                  self().margins().top(),
                                  self().root().viewHeight() - self().rule().height() -
                                  self().margins().bottom() + self().margins().top()));
            break;
        case ui::Right:
            self().rule()
                    .setInput(Rule::Left, OperatorRule::minimum(
                                  *anchorPos.x + *marker,
                                  self().root().viewWidth() - self().rule().width() - self().margins().right()))
                    .setInput(Rule::Top,  OperatorRule::clamped(
                                  *anchorPos.y - self().rule().height() / 2,
                                  self().margins().top(),
                                  self().root().viewHeight() - self().rule().height() - self().margins().bottom()));
            break;
        case ui::NoDirection:
            self().rule().setMidAnchorX(*anchorPos.x)
                       .setMidAnchorY(*anchorPos.y);
            break;
        }
    }
    void updateStyle() {
        Style const &st = style();
        bool const opaqueBackground = (self().levelOfNesting() > 0);
        outlineColor = st.colors().colorf(outlineColorId);
        if (colorTheme == Inverted) {
            self().set(self().infoStyleBackground());
        }
        else {
            Background bg(st.colors().colorf("background"),
                          !opaqueBackground && st.isBlurringAllowed()?
                              Background::SharedBlurWithBorderGlow : Background::BorderGlow,
                          st.colors().colorf("glow"),
                          st.rules().rule("glow").valuei());
            bg.blur = style().sharedBlurWidget();
            self().set(bg);
        }
        if (opaqueBackground) {
            self().set(self().background().withSolidFillOpacity(1));
        }
    }
};
PopupWidget::PopupWidget(String const &name) : PanelWidget(name), d(new Impl(this)) {
    setOpeningDirection(ui::Up);
    d->updateStyle();
}
int PopupWidget::levelOfNesting() const {
    int nesting = 0;
    for (GuiWidget const *p = d->realParent && d->realParent->parent()?
                static_cast<GuiWidget const *>(d->realParent.get()) : parentGuiWidget();
         p; p = p->parentGuiWidget()) {
        if (is<PopupWidget>(p)) {
            ++nesting;
        }
    }
    return nesting;
}
void PopupWidget::setAnchorAndOpeningDirection(RuleRectangle const &rule, ui::Direction dir) {
    d->anchor.setRect(rule);
    setOpeningDirection(dir);
}
void PopupWidget::setAllowDirectionFlip(bool flex) {
    d->flexibleDir = flex;
}
void PopupWidget::setAnchor(Vector2i const &pos) {
    d->anchor.setLeftTop(Const(pos.x), Const(pos.y));
    d->anchor.setRightBottom(d->anchor.left(), d->anchor.top());
}
void PopupWidget::setAnchorX(int xPos) {
    d->anchor.setInput(Rule::Left,  Const(xPos))
             .setInput(Rule::Right, Const(xPos));
}
void PopupWidget::setAnchorY(int yPos) {
    d->anchor.setInput(Rule::Top,    Const(yPos))
             .setInput(Rule::Bottom, Const(yPos));
}
void PopupWidget::setAnchor(Rule const &x, Rule const &y) {
    setAnchorX(x);
    setAnchorY(y);
}
void PopupWidget::setAnchorX(Rule const &x) {
    d->anchor.setInput(Rule::Left,   x)
             .setInput(Rule::Right,  x);
}
void PopupWidget::setAnchorY(Rule const &y) {
    d->anchor.setInput(Rule::Top,    y)
             .setInput(Rule::Bottom, y);
}
RuleRectangle const &PopupWidget::anchor() const {
    return d->anchor;
}
void PopupWidget::detachAnchor() {
    setAnchor(d->anchorPos());
    d->updateLayout();
}
void PopupWidget::setDeleteAfterDismissed(bool deleteAfterDismiss) {
    d->deleteAfterDismiss = deleteAfterDismiss;
}
void PopupWidget::setClickToClose(bool clickCloses) {
    d->clickToClose = clickCloses;
}
void PopupWidget::useInfoStyle(bool yes) {
    setColorTheme(yes? Inverted : Normal);
}
bool PopupWidget::isUsingInfoStyle() {
    return d->colorTheme == Inverted;
}
void PopupWidget::setColorTheme(ColorTheme theme) {
    d->colorTheme = theme;
    if (d->close) d->close->setColorTheme(theme);
    d->updateStyle();
}
GuiWidget::ColorTheme PopupWidget::colorTheme() const {
    return d->colorTheme;
}
void PopupWidget::setOutlineColor(const DotPath &outlineColor) {
    d->outlineColorId = outlineColor;
    d->updateStyle();
}
void PopupWidget::setCloseButtonVisible(bool enable) {
    if (enable && !d->close) {
        d->close = new ButtonWidget;
        d->close->setColorTheme(d->colorTheme);
        d->close->setStyleImage("close.ringless", "small");
        d->close->margins().set("dialog.gap").setTopBottom(RuleBank::UNIT);
        d->close->setImageColor(d->close->textColorf());
        d->close->setSizePolicy(ui::Expand, ui::Expand);
        d->close->setActionFn([this] () { close(); });
        d->close->rule()
                .setInput(Rule::Top,   rule().top()   + margins().top())
                .setInput(Rule::Right, rule().right() - margins().right());
        add(d->close);
    }
    else if (!enable && d->close) {
        delete d->close;
        d->close = nullptr;
    }
}
ButtonWidget &PopupWidget::closeButton() {
    setCloseButtonVisible(true);
    return *d->close;
}
void PopupWidget::offerFocus() {
    if (d->close) {
        root().setFocus(d->close);
    }
}
GuiWidget::Background PopupWidget::infoStyleBackground() const {
    return Background(style().colors().colorf("popup.info.background"),
                      Background::BorderGlow,
                      style().colors().colorf("popup.info.glow"),
                      rule("glow").valuei());
}
bool PopupWidget::handleEvent(Event const &event) {
    if (!isOpen()) return false;
    if (event.type() == Event::MouseButton) {
        bool const inside = hitTest(event);
        if (!inside && d->clickToClose) {
            close(0.1);
        }
    }
    if (event.type() == Event::KeyPress  ||
        event.type() == Event::KeyRepeat ||
        event.type() == Event::KeyRelease) {
        KeyEvent const &key = event.as<KeyEvent>();
        if (event.isKeyDown() &&
            (key.ddKey() == DDKEY_ESCAPE ||
             key.ddKey() == DDKEY_ENTER  ||
             key.ddKey() == DDKEY_RETURN ||
             key.ddKey() == ' ')) {
            close();
            return true;
        }
        root().handleEventAsFallback(event);
        return true;
    }
    return PanelWidget::handleEvent(event);
}
void PopupWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    if (rule().recti().isNull()) return; // Still closed.
    PanelWidget::glMakeGeometry(verts);
    ui::Direction const dir = openingDirection();
    if (dir == ui::NoDirection) return;
    GuiVertexBuilder tri;
    GuiVertex v;
    v.rgba = background().solidFill;
    v.texCoord = root().atlas().imageRectf(root().solidWhitePixel()).middle();
    int const marker = d->marker->valuei();
    Vector2i anchorPos = d->anchorPos();
    bool markerVisible = false;
    if (dir == ui::Up) {
        anchorPos.x = clamp(2 * marker, anchorPos.x, int(root().viewSize().x) - 2*marker);
        if (anchorPos.y > rule().bottom().valuei()) {
            v.pos = anchorPos; tri << v;
            v.pos = anchorPos + Vector2i(-marker, -marker); tri << v;
            v.pos = anchorPos + Vector2i(marker, -marker); tri << v;
            markerVisible = true;
        }
    }
    else if (dir == ui::Left) {
        if (anchorPos.x > rule().right().valuei()) {
            v.pos = anchorPos; tri << v;
            v.pos = anchorPos + Vector2i(-marker, marker); tri << v;
            v.pos = anchorPos + Vector2i(-marker, -marker); tri << v;
            markerVisible = true;
        }
    }
    else if (dir == ui::Right) {
        if (anchorPos.x < rule().left().valuei()) {
            v.pos = anchorPos; tri << v;
            v.pos = anchorPos + Vector2i(marker, -marker); tri << v;
            v.pos = anchorPos + Vector2i(marker, marker); tri << v;
            markerVisible = true;
        }
    }
    else {
        if (anchorPos.y < rule().top().valuei()) {
            v.pos = anchorPos; tri << v;
            v.pos = anchorPos + Vector2i(marker, marker); tri << v;
            v.pos = anchorPos + Vector2i(-marker, marker); tri << v;
            markerVisible = true;
        }
    }
    if (d->outlineColor.w > 0.f) {
        tri << v; // discontinued
        Rectanglei const rect = rule().recti();
        int const ow = GuiWidget::pointsToPixels(2);
        int const halfOw = ow/2;
        int const midOw = ow + halfOw;
        v.rgba = d->outlineColor;
        v.pos = rect.topLeft + Vector2i(-ow, -ow); tri << v << v;
        v.pos = rect.topLeft; tri << v;
        if (markerVisible && dir == ui::Down) {
            v.pos = Vector2i(anchorPos.x - marker - halfOw, rect.top() - ow); tri << v;
            v.pos += Vector2i(halfOw, ow); tri << v;
            v.pos = anchorPos + Vector2i(0, -midOw); tri << v;
            v.pos.y += midOw; tri << v;
            v.pos = Vector2i(anchorPos.x + marker + halfOw, rect.top() - ow); tri << v;
            v.pos += Vector2i(-halfOw, ow); tri << v;
        }
        v.pos = rect.topRight() + Vector2i(ow, -ow); tri << v;
        v.pos = rect.topRight(); tri << v;
        if (markerVisible && dir == ui::Left) {
            v.pos = Vector2i(rect.right() + ow, anchorPos.y - marker - halfOw); tri << v;
            v.pos += Vector2i(-ow, halfOw); tri << v;
            v.pos = anchorPos + Vector2i(midOw, 0); tri << v;
            v.pos.x += -midOw; tri << v;
            v.pos = Vector2i(rect.right() + ow, anchorPos.y + marker + halfOw); tri << v;
            v.pos += Vector2i(-ow, -halfOw); tri << v;
        }
        v.pos = rect.bottomRight + Vector2i(ow, ow); tri << v;
        v.pos = rect.bottomRight; tri << v;
        if (markerVisible && dir == ui::Up) {
            v.pos = Vector2i(anchorPos.x + marker + halfOw, rect.bottom() + ow); tri << v;
            v.pos += Vector2i(-halfOw, -ow); tri << v;
            v.pos = anchorPos + Vector2i(0, midOw); tri << v;
            v.pos.y += -midOw; tri << v;
            v.pos = Vector2i(anchorPos.x - marker - halfOw, rect.bottom() + ow); tri << v;
            v.pos += Vector2i(halfOw, -ow); tri << v;
        }
        v.pos = rect.bottomLeft() + Vector2i(-ow, ow); tri << v;
        v.pos = rect.bottomLeft(); tri << v;
        if (markerVisible && dir == ui::Right) {
            v.pos = Vector2i(rect.left() - ow, anchorPos.y + marker + halfOw); tri << v;
            v.pos += Vector2i(ow, -halfOw); tri << v;
            v.pos = anchorPos + Vector2i(-midOw, 0); tri << v;
            v.pos.x += midOw; tri << v;
            v.pos = Vector2i(rect.left() - ow, anchorPos.y - marker - halfOw); tri << v;
            v.pos += Vector2i(ow, halfOw); tri << v;
        }
        v.pos = rect.topLeft + Vector2i(-ow, -ow); tri << v;
        v.pos = rect.topLeft; tri << v;
    }
    verts += tri;
}
void PopupWidget::updateStyle() {
    PanelWidget::updateStyle();
    d->updateStyle();
}
void PopupWidget::preparePanelForOpening() {
    d->updateStyle();
    PanelWidget::preparePanelForOpening();
    if (d->flexibleDir) {
        d->flipOpeningDirectionIfNeeded();
    }
    d->realParent.reset(Widget::parent());
    DENG2_ASSERT(d->realParent);
    d->realParent->remove(*this);
    d->realParent->root().as<GuiRootWidget>().addOnTop(this);
    d->updateLayout();
    root().pushFocus();
    offerFocus();
}
void PopupWidget::panelClosing() {
    PanelWidget::panelClosing();
    root().popFocus();
}
void PopupWidget::panelDismissed() {
    PanelWidget::panelDismissed();
    if (!d->realParent) {
        d->realParent.reset(&root());
        DENG2_ASSERT(d->realParent);
    }
    parentWidget()->remove(*this);
    if (d->deleteAfterDismiss) {
        guiDeleteLater();
    }
    else {
        d->realParent->add(this);
    }
    d->realParent.reset();
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(ProgressWidget), public Lockable {
    Mode mode             = Indefinite;
    Rangei range;
    Rangef visualRange    { 0, 1 };
    Animation pos         { 0, Animation::Linear };
    float angle           = 0;
    float rotationSpeed   = 20;
    bool posChanging      = false;
    bool mini             = false;
    Id gearTex;
    DotPath colorId       { "progress.light.wheel" };
    DotPath shadowColorId { "progress.light.shadow" };
    DotPath gearId        { "progress.gear" };
    Time updateAt         { Time::invalidTime() };
    int framesWhileAnimDone = 0; ///< # of frames drawn while animation was already done.
    Impl(Public *i) : Base(i) {
        updateStyle();
    }
    void updateStyle() {
        if (mini) {
            self().setImageColor(Vector4f());
        }
        else {
            self().setImageColor(style().colors().colorf(colorId));
        }
    }
    void useDotStyle() {
        mini = true;
        self().setImage(nullptr);
        updateStyle();
    }
    void glInit() {
        gearTex = root().styleTexture(gearId);
    }
    void glDeinit() {
        gearTex = Id::None;
    }
    void makeRingGeometry(GuiVertexBuilder &verts) {
        ContentLayout layout;
        self().contentLayout(layout);
        float gradientThick = layout.image.width() * 2.f;
        float solidThick    = layout.image.width() * .53f;
        Vector4f const shadowColor = style().colors().colorf(shadowColorId);
        verts.makeRing(layout.image.middle(),
                       solidThick, 0, 30,
                       shadowColor,
                       root().atlas().imageRectf(root().borderGlow()).middle());
        verts.makeRing(layout.image.middle(),
                       gradientThick, solidThick, 30,
                       shadowColor,
                       root().atlas().imageRectf(root().borderGlow()), 0);
        self().LabelWidget::glMakeGeometry(verts);
        if (pos.done()) {
            ++framesWhileAnimDone;
        }
        Rectanglef const tc = atlas().imageRectf(gearTex);
        float pos = 1;
        if (mode != Indefinite) {
            pos = de::clamp(0.f, this->pos.value(), 1.f);
        }
        pos = visualRange.start + pos * visualRange.size();
        int const edgeCount = de::max(1, int(pos * 30));
        float const radius = layout.image.width() / 2;
        Matrix4f const rotation = Matrix4f::rotateAround(tc.middle(), -angle);
        GuiVertexBuilder gear;
        GuiVertex v;
        v.rgba = style().colors().colorf(colorId) * Vector4f(1, 1, 1, mini? 1.f : 1.9f);
        for (int i = 0; i <= edgeCount; ++i) {
            v.pos = layout.image.middle();
            v.texCoord = tc.middle();
            gear << v;
            float const angle = 2 * PI * pos * (i / (float)edgeCount) + PI/2;
            v.pos = v.pos + Vector2f(cos(angle)*radius*1.05f, sin(angle)*radius*1.05f);
            v.texCoord = rotation * (tc.topLeft + tc.size() * Vector2f(.5f + cos(angle)*.5f,
                                                                       .5f + sin(angle)*.5f));
            gear << v;
        }
        verts += gear;
    }
    void makeDotsGeometry(GuiVertexBuilder &verts) {
        Image::Size const dotSize = atlas().imageRect(root().tinyDot()).size();
        Rectanglei rect = self().contentRect().shrunk(dotSize.x / 2);
        int const midY  = rect.middle().y;
        int count       = range.size();
        Vector4f color  = style().colors().colorf(colorId);
        int const gap   = rule(RuleBank::UNIT).valuei();
        int totalWidth  = count * dotSize.x + (count - 1) * gap;
        for (int i = 0; i < count; ++i) {
            Vector4f dotColor = (float(i)/float(count) <= pos.value()? color : Vector4f(color, .166f));
            float midX = rect.middle().x - totalWidth/2 + i * (dotSize.x + gap);
            verts.makeQuad(Rectanglef::fromSize(Vector2f(midX, midY) - dotSize/2, dotSize),
                           dotColor, atlas().imageRectf(root().tinyDot()));
        }
    }
};
ProgressWidget::ProgressWidget(String const &name)
    : LabelWidget(name)
    , d(new Impl(this)) {
    setTextGap("progress.textgap");
    setSizePolicy(ui::Filled, ui::Filled);
    setStyleImage("progress.wheel");
    setImageFit(ui::FitToSize | ui::OriginalAspectRatio);
    setImageScale(.6f);
    setTextAlignment(ui::AlignRight);
    setTextLineAlignment(ui::AlignLeft);
    setTextShadow(RectangleShadow);
}
void ProgressWidget::useMiniStyle(DotPath const &colorId) {
    d->mini = true;
    d->colorId = colorId;
    d->gearId = "progress.mini";
    setTextColor(colorId);
    setRotationSpeed(40);
    setImageScale(1);
    setOverrideImageSize(style().fonts().font("default").height());
    d->updateStyle();
}
void ProgressWidget::setRotationSpeed(float anglesPerSecond) {
    d->rotationSpeed = anglesPerSecond;
}
ProgressWidget::Mode ProgressWidget::mode() const {
    DENG2_GUARD(d);
    return d->mode;
}
Rangei ProgressWidget::range() const {
    DENG2_GUARD(d);
    return d->range;
}
bool ProgressWidget::isAnimating() const {
    DENG2_GUARD(d);
    return d->framesWhileAnimDone < 2;
}
void ProgressWidget::setColor(DotPath const &styleId) {
    d->colorId = styleId;
    d->updateStyle();
}
void ProgressWidget::setShadowColor(DotPath const &styleId) {
    d->shadowColorId = styleId;
    d->updateStyle();
}
void ProgressWidget::setText(String const &text) {
    DENG2_GUARD(d);
    LabelWidget::setText(text);
}
void ProgressWidget::setMode(Mode progressMode) {
    DENG2_GUARD(d);
    d->mode = progressMode;
    if (d->mode == Dots) {
        d->useDotStyle();
    }
}
void ProgressWidget::setRange(Rangei const &range, Rangef const &visualRange) {
    DENG2_GUARD(d);
    d->range = range;
    d->visualRange = visualRange;
    setMode(Ranged);
}
void ProgressWidget::setProgress(int currentProgress, TimeSpan const &transitionSpan) {
    DENG2_GUARD(d);
    d->framesWhileAnimDone = 0;
    d->pos.setValue(float(currentProgress - d->range.start) / float(d->range.size()),
                    transitionSpan);
    d->posChanging = true;
}
void ProgressWidget::update() {
    DENG2_GUARD(d);
    LabelWidget::update();
    if (d->mode != Dots) {
        Time const now = Time();
        if (!d->updateAt.isValid()) d->updateAt = now;
        TimeSpan const elapsed = d->updateAt.since();
        d->updateAt = now;
        d->angle = de::wrap(d->angle + float(elapsed * d->rotationSpeed), 0.f, 360.f);
        if (isVisible()) {
            requestGeometry();
        }
    }
    else // Dots
    {
        if (isVisible() && d->posChanging) {
            requestGeometry();
        }
    }
    if (d->posChanging && d->pos.done()) {
        d->posChanging = false;
    }
}
void ProgressWidget::glInit() {
    DENG2_GUARD(d);
    LabelWidget::glInit();
    d->glInit();
}
void ProgressWidget::glDeinit() {
    DENG2_GUARD(d);
    d->glDeinit();
    LabelWidget::glDeinit();
}
void ProgressWidget::glMakeGeometry(GuiVertexBuilder &verts) {
    DENG2_GUARD(d);
    switch (d->mode) {
    case Ranged:
    case Indefinite:
        d->makeRingGeometry(verts);
        break;
    case Dots:
        d->makeDotsGeometry(verts);
        break;
    }
}
void ProgressWidget::updateStyle() {
    d->updateStyle();
}
} // namespace de
namespace de {
DENG2_PIMPL(RelayWidget)
, DENG2_OBSERVES(Widget, Deletion) {
    GuiWidget *target = nullptr;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        setTarget(nullptr);
    }
    void setTarget(GuiWidget *w) {
        if (target) target->audienceForDeletion() -= this;
        target = w;
        if (target) target->audienceForDeletion() += this;
    }
    void widgetBeingDeleted(Widget &w) {
        if (target == &w) {
            DENG2_FOR_PUBLIC_AUDIENCE2(Target, i) {
                i->relayTargetBeingDeleted(self());
            }
            target = nullptr;
        }
    }
    DENG2_PIMPL_AUDIENCE(Target)
};
DENG2_AUDIENCE_METHOD(RelayWidget, Target)
RelayWidget::RelayWidget(GuiWidget *target, String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    d->setTarget(target);
}
void RelayWidget::setTarget(GuiWidget *target) {
    d->setTarget(target);
}
GuiWidget *RelayWidget::target() const {
    return d->target;
}
void RelayWidget::initialize() {
    GuiWidget::initialize();
    if (d->target) {
        d->target->setRoot(&root());
        d->target->notifySelfAndTree(&Widget::initialize);
    }
}
void RelayWidget::deinitialize() {
    GuiWidget::deinitialize();
    if (d->target) {
        d->target->notifySelfAndTree(&Widget::deinitialize);
    }
}
void RelayWidget::viewResized() {
    GuiWidget::viewResized();
    if (d->target) {
        d->target->notifySelfAndTree(&Widget::viewResized);
    }
}
void RelayWidget::update() {
    GuiWidget::update();
    if (d->target) {
        d->target->setRoot(&root());
        d->target->notifySelfAndTree(&Widget::update);
    }
}
bool RelayWidget::handleEvent(Event const &event) {
    if (d->target) {
        return d->target->dispatchEvent(event, &Widget::handleEvent);
    }
    return GuiWidget::handleEvent(event);
}
bool RelayWidget::hitTest(Vector2i const &pos) const {
    if (d->target) {
        return d->target->hitTest(pos);
    }
    return false;
}
void RelayWidget::drawContent() {
    if (d->target) {
        NotifyArgs args(&Widget::draw);
        args.conditionFunc  = &Widget::isVisible;
        args.preNotifyFunc  = &Widget::preDrawChildren;
        args.postNotifyFunc = &Widget::postDrawChildren;
        d->target->notifySelfAndTree(args);
    }
}
} // namespace de
namespace de {
DENG2_PIMPL(ScriptCommandWidget)
, DENG2_OBSERVES(App, StartupComplete) {
    Script script;
    Process process;
    Impl(Public *i) : Base(i) {
        App::app().audienceForStartupComplete() += this;
    }
    void appStartupCompleted() {
        self().updateCompletion();
    }
    void importNativeModules() {
        foreach (String const &name, App::scriptSystem().nativeModules()) {
            process.globals().add(new Variable(name,
                    new RecordValue(App::scriptSystem().nativeModule(name))));
        }
    }
    void updateLexicon() {
        shell::Lexicon lexi;
        lexi.setCaseSensitive(true);
        lexi.setAdditionalWordChars("_");
        DENG2_FOR_EACH_CONST(Record::Members, i, process.globals().members()) {
            lexi.addTerm(i.key());
        }
        foreach (String name, BuiltInExpression::identifiers()) {
            lexi.addTerm(name);
        }
        foreach (String keyword, ScriptLex::keywords()) {
            lexi.addTerm(keyword);
        }
        self().setLexicon(lexi);
    }
    bool shouldShowAsPopup(Error const &) {
        /*if (dynamic_cast<ScriptLex::MismatchedBracketError const *>(&er))
        {
            return false;
        }*/
        return true;
    }
};
ScriptCommandWidget::ScriptCommandWidget(String const &name)
    : CommandWidget(name), d(new Impl(this))
{}
bool ScriptCommandWidget::handleEvent(Event const &event) {
    if (isDisabled()) return false;
    bool wasCompl = autocompletionPopup().isOpen();
    bool eaten = CommandWidget::handleEvent(event);
    if (eaten && wasCompl && event.isKeyDown()) {
        closeAutocompletionPopup();
    }
    return eaten;
}
void ScriptCommandWidget::updateCompletion() {
    d->importNativeModules();
    d->updateLexicon();
}
bool ScriptCommandWidget::isAcceptedAsCommand(String const &text) {
    try {
        d->script.parse(text);
        return true; // Looks good!
    }
    catch (Error const &er) {
        if (d->shouldShowAsPopup(er)) {
            showAutocompletionPopup(er.asText());
        }
        return false;
    }
}
void ScriptCommandWidget::executeCommand(String const &text) {
    LOG_SCR_NOTE(_E(1) "$ " _E(>) _E(m)) << text;
    try {
        d->process.run(d->script);
        d->process.execute();
    }
    catch (Error const &er) {
        LOG_SCR_WARNING("Error in script:\n") << er.asText();
    }
    try {
        Value const &result = d->process.context().evaluator().result();
        if (!is<NoneValue>(result)) {
            String msg = DENG2_CHAR_RIGHT_DOUBLEARROW " " _E(>)_E(m) + result.asText();
            LOG_SCR_MSG(msg);
        }
    }
    catch (Error const &)
    {}
}
void ScriptCommandWidget::autoCompletionBegan(String const &prefix) {
    QStringList const compls = suggestedCompletions();
    if (compls.size() > 1) {
        showAutocompletionPopup(tr("Completions for %1:\n")
                                .arg(_E(b) + prefix + _E(.)) +
                                _E(m) + compls.join("\n"));
    }
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(ScrollAreaWidget), public Lockable {
    /**
     * Rectangle for all the content shown in the widget. The widget's
     * rectangle is the viewport into this content rectangle.
     */
    RuleRectangle contentRule;
    AnimationRule *x;
    AnimationRule *y;
    Rule *maxX;
    Rule *maxY;
    String scrollBarColorId { "accent" };
    ColorBank::Colorf scrollBarColor;
    ColorBank::Colorf scrollBarHoverColor;
    Animation scrollOpacity { 0 };
    Origin origin = Top;
    int scrollBarWidth { 0 };
    int grabOffset;
    bool scrollBarGrabbed { false };
    bool indicatorAnimating { false };
    bool pageKeysEnabled { true };
    bool scrollingEnabled { true };
    bool indicatorDrawEnabled { false };
    bool scrollBarHover { false };
    Rectanglef indicatorUv;
    Rectanglef scrollBarVisRect;
    Rectanglef scrollBarLaneRect;
    GuiVertexBuilder verts;
    Impl(Public *i) : Base(i) {
        contentRule.setDebugName("ScrollArea-contentRule");
        updateStyle();
        x = new AnimationRule(0);
        y = new AnimationRule(0);
        maxX = new OperatorRule(OperatorRule::Maximum, Const(0),
                                contentRule.width() - self().rule().width() + self().margins().width());
        maxY = new OperatorRule(OperatorRule::Maximum, Const(0),
                                contentRule.height() - self().rule().height() + self().margins().height());
    }
    ~Impl() {
        releaseRef(x);
        releaseRef(y);
        releaseRef(maxX);
        releaseRef(maxY);
    }
    void glInit() {
        if (indicatorDrawEnabled) {
            /*DefaultVertexBuf *buf = new DefaultVertexBuf;
            drawable.addBuffer(buf);
            shaders().build(drawable.program(), "generic.textured.color_ucolor")
                    << uMvpMatrix << uAtlas() << uColor;*/
        }
    }
    void glDeinit() {
        verts.clear();
    }
    void updateStyle() {
        Style const &st = style();
        scrollBarWidth      = st.rules().rule("scrollarea.bar").valuei();
        scrollBarColor      = st.colors().colorf(scrollBarColorId);
        scrollBarHoverColor = st.colors().colorf("background");
    }
    void restartScrollOpacityFade() {
        indicatorAnimating = true;
        if (origin == Bottom && self().isAtBottom()) {
            scrollOpacity.setValue(0, .7f, .2f);
        }
        else {
            scrollOpacity.setValueFrom(.8f, .333f, 5, 2);
        }
    }
    void setScrollBarHovering(bool hover) {
        if (hover) restartScrollOpacityFade();
        if (hover != scrollBarHover) {
            scrollBarHover = hover;
            self().requestGeometry();
        }
    }
    typedef std::pair<Rectanglef, Rectanglef> RectanglefPair;
    RectanglefPair scrollIndicatorRects(Vector2f const &originPos) const {
        Vector2i const viewSize = self().viewportSize();
        if (viewSize == Vector2i()) return RectanglefPair();
        const auto &margins = self().margins();
        const float contentHeight = float(contentRule.height().value());
        int indHeight = 0;
        if (contentHeight > 0) {
            indHeight = de::clamp(
                    margins.height().valuei(),
                    int(float(viewSize.y * viewSize.y) / contentHeight),
                    viewSize.y / 2);
        }
        float indPos = self().scrollPositionY().value() / self().maximumScrollY().value();
        if (origin == Top) indPos = 1 - indPos;
        float const avail = viewSize.y - indHeight;
        Rectanglef rect { originPos + Vector2f(viewSize.x + margins.left().value() - 2 * scrollBarWidth,
                                               avail - indPos * avail),
                          originPos + Vector2f(viewSize.x + margins.left().value() - scrollBarWidth,
                                               avail - indPos * avail + indHeight) };
        Rectanglef laneRect(rect.left(), originPos.y, scrollBarWidth, viewSize.y);
        return RectanglefPair(rect, laneRect);
    }
};
ScrollAreaWidget::ScrollAreaWidget(String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    setBehavior(ChildHitClipping);
    setBehavior(ChildVisibilityClipping); // enable clip-culling
    d->contentRule.setInput(Rule::Left, rule().left() + margins().left() -
                            OperatorRule::minimum(*d->x, *d->maxX));
    setOrigin(Top);
    setContentWidth(0);
    setContentHeight(0);
}
void ScrollAreaWidget::setScrollBarColor(DotPath const &colorId) {
    d->scrollBarColorId = colorId;
    d->updateStyle();
}
void ScrollAreaWidget::setOrigin(Origin origin) {
    DENG2_GUARD(d);
    d->origin = origin;
    if (origin == Top) {
        d->contentRule.setInput(Rule::Top, rule().top() + margins().top() -
                                OperatorRule::minimum(*d->y, *d->maxY));
        d->contentRule.clearInput(Rule::Bottom);
    }
    else {
        d->contentRule.setInput(Rule::Bottom, rule().bottom() - margins().bottom() +
                                OperatorRule::minimum(*d->y, *d->maxY));
        d->contentRule.clearInput(Rule::Top);
    }
}
ScrollAreaWidget::Origin ScrollAreaWidget::origin() const {
    return d->origin;
}
void ScrollAreaWidget::setIndicatorUv(Rectanglef const &uv) {
    d->indicatorUv = uv;
}
void ScrollAreaWidget::setIndicatorUv(Vector2f const &uvPoint) {
    d->indicatorUv = Rectanglef::fromSize(uvPoint, Vector2f(0, 0));
}
void ScrollAreaWidget::setContentWidth(int width) {
    DENG2_GUARD(d);
    d->contentRule.setInput(Rule::Width, Const(width));
}
void ScrollAreaWidget::setContentWidth(Rule const &width) {
    DENG2_GUARD(d);
    d->contentRule.setInput(Rule::Width, width);
}
void ScrollAreaWidget::setContentHeight(int height) {
    DENG2_GUARD(d);
    d->contentRule.setInput(Rule::Height, Const(height));
}
void ScrollAreaWidget::setContentHeight(Rule const &height) {
    DENG2_GUARD(d);
    d->contentRule.setInput(Rule::Height, height);
}
void ScrollAreaWidget::setContentSize(Rule const &width, Rule const &height) {
    DENG2_GUARD(d);
    setContentWidth(width);
    setContentHeight(height);
}
void ScrollAreaWidget::setContentSize(ISizeRule const &dimensions) {
    DENG2_GUARD(d);
    setContentWidth(dimensions.width());
    setContentHeight(dimensions.height());
}
void ScrollAreaWidget::setContentSize(Vector2i const &size) {
    DENG2_GUARD(d);
    setContentWidth(size.x);
    setContentHeight(size.y);
}
void ScrollAreaWidget::setContentSize(Vector2ui const &size) {
    setContentSize(Vector2i(size.x, size.y));
}
void ScrollAreaWidget::modifyContentWidth(int delta) {
    DENG2_GUARD(d);
    setContentWidth(de::max(0, d->contentRule.width().valuei() + delta));
}
void ScrollAreaWidget::modifyContentHeight(int delta) {
    DENG2_GUARD(d);
    setContentHeight(de::max(0, d->contentRule.height().valuei() + delta));
}
int ScrollAreaWidget::contentWidth() const {
    DENG2_GUARD(d);
    return d->contentRule.width().valuei();
}
int ScrollAreaWidget::contentHeight() const {
    DENG2_GUARD(d);
    return d->contentRule.height().valuei();
}
RuleRectangle const &ScrollAreaWidget::contentRule() const {
    return d->contentRule;
}
AnimationRule &ScrollAreaWidget::scrollPositionX() const {
    return *d->x;
}
AnimationRule &ScrollAreaWidget::scrollPositionY() const {
    return *d->y;
}
Rule const &ScrollAreaWidget::maximumScrollX() const {
    return *d->maxX;
}
Rule const &ScrollAreaWidget::maximumScrollY() const {
    return *d->maxY;
}
bool ScrollAreaWidget::isScrolling() const {
    return !d->x->animation().done() || !d->y->animation().done();
}
Rectanglei ScrollAreaWidget::viewport() const {
    Vector4i const margin = margins().toVector();
    Rectanglei vp = rule().recti().moved(margin.xy());
    if (int(vp.width()) <= margin.x + margin.z) {
        vp.setWidth(0);
    }
    else {
        vp.bottomRight.x -= margin.x + margin.z;
    }
    if (int(vp.height()) <= margin.y + margin.w) {
        vp.setHeight(0);
    }
    else {
        vp.bottomRight.y -= margin.y + margin.w;
    }
    return vp;
}
Vector2i ScrollAreaWidget::viewportSize() const {
    return Vector2i(rule().width().valuei()  - margins().width().valuei(),
                    rule().height().valuei() - margins().height().valuei())
            .max(Vector2i(0, 0));
}
Vector2i ScrollAreaWidget::scrollPosition() const {
    DENG2_GUARD(d);
    return Vector2i(scrollPositionX().valuei(), scrollPositionY().valuei());
}
Vector2i ScrollAreaWidget::scrollPageSize() const {
    return viewportSize();
}
Vector2i ScrollAreaWidget::maximumScroll() const {
    DENG2_GUARD(d);
    return Vector2i(maximumScrollX().valuei(), maximumScrollY().valuei());
}
void ScrollAreaWidget::scroll(Vector2i const &to, TimeSpan span) {
    scrollX(to.x, span);
    scrollY(to.y, span);
}
void ScrollAreaWidget::scrollX(int to, TimeSpan span) {
    d->x->set(de::clamp(0, to, maximumScrollX().valuei()), span);
}
void ScrollAreaWidget::scrollY(int to, TimeSpan span) {
    d->y->set(de::clamp(0, to, maximumScrollY().valuei()), span);
    d->restartScrollOpacityFade();
}
void ScrollAreaWidget::scrollY(Rule const &to, TimeSpan span) {
    d->y->set(OperatorRule::clamped(to, Const(0), maximumScrollY()), span);
    d->restartScrollOpacityFade();
}
bool ScrollAreaWidget::isScrollable() const {
    return d->scrollingEnabled && maximumScrollY().valuei() > 0;
}
bool ScrollAreaWidget::isAtBottom() const {
    return d->origin == Bottom && d->y->animation().target() == 0;
}
void ScrollAreaWidget::enableScrolling(bool enabled) {
    d->scrollingEnabled = enabled;
}
void ScrollAreaWidget::enablePageKeys(bool enabled) {
    d->pageKeysEnabled = enabled;
}
void ScrollAreaWidget::enableIndicatorDraw(bool enabled) {
    d->indicatorDrawEnabled = enabled;
    d->restartScrollOpacityFade();
}
bool ScrollAreaWidget::handleEvent(Event const &event) {
    if (d->scrollBarGrabbed && event.isMouse()) {
        if (event.type() == Event::MousePosition) {
            int const mouseY = event.as<MouseEvent>().pos().y;
            int topEdge = mouseY - d->scrollBarLaneRect.top() - d->grabOffset;
            float relativePos = topEdge / (d->scrollBarLaneRect.height() - d->scrollBarVisRect.height());
            if (d->origin == Bottom) {
                relativePos = 1.f - relativePos;
            }
            scrollY(relativePos * (contentHeight() - viewportSize().y));
        }
        switch (handleMouseClick(event)) {
        case MouseClickFinished:
        case MouseClickAborted:
            d->scrollBarGrabbed = false;
            break;
        default:
            break;
        }
        return true;
    }
    if (d->scrollingEnabled && event.isMouse() && hitTest(event)) {
        if (event.type() == Event::MouseWheel) {
            MouseEvent const &mouse = event.as<MouseEvent>();
#ifdef MACOSX
            if (mouse.wheelMotion() == MouseEvent::FineAngle) {
                d->y->set(de::clamp(0, int(d->y->animation().target()) +
                                    pointsToPixels(mouse.wheel().y / 2 * (d->origin == Top? -1 : 1)),
                                    d->maxY->valuei()), .05f);
                d->restartScrollOpacityFade();
            }
#else
            if (mouse.wheelMotion() == MouseEvent::Step) {
                unsigned int lineCount = 1;
#ifdef WIN32
                SystemParametersInfo(SPI_GETWHEELSCROLLLINES, 0, &lineCount, 0);
                if (lineCount == WHEEL_PAGESCROLL) {
                    lineCount = contentRect().height() / style().fonts().font("default").height().valuei();
                }
#endif
                d->y->set(de::clamp(0, int(d->y->animation().target()) +
                                    int(mouse.wheel().y * lineCount * style().fonts().font("default").height().valuei() *
                                        (d->origin == Top? -1 : 1)), d->maxY->valuei()), .15f);
                d->restartScrollOpacityFade();
            }
#endif
            return true;
        }
        else if (d->indicatorDrawEnabled) {
            if (event.type() == Event::MousePosition) {
                bool const hovering = (d->scrollBarVisRect.expanded(pointsToPixels(1))
                                       .contains(event.as<MouseEvent>().pos()));
                d->setScrollBarHovering(hovering);
            }
            if (d->scrollBarHover && !d->scrollBarGrabbed) {
                if(handleMouseClick(event) == MouseClickStarted) {
                    d->scrollBarGrabbed = true;
                    d->grabOffset = event.as<MouseEvent>().pos().y - d->scrollBarVisRect.top();
                    return true;
                }
            }
        }
    }
    if (d->scrollBarHover && event.type() == Event::MousePosition && !hitTest(event)) {
        d->setScrollBarHovering(false);
    }
    if (d->scrollingEnabled && event.isKeyDown()) {
        KeyEvent const &ev = event.as<KeyEvent>();
        float pageSize = scrollPageSize().y;
        if (d->origin == Bottom) pageSize = -pageSize;
        switch (ev.ddKey()) {
        case DDKEY_PGUP:
            if (!d->pageKeysEnabled) return false;
            if (ev.modifiers().testFlag(KeyEvent::Shift)) {
                scrollToTop();
            }
            else {
                scrollY(d->y->animation().target() - pageSize/2, .3f);
            }
            return true;
        case DDKEY_PGDN:
            if (!d->pageKeysEnabled) return false;
            if (ev.modifiers().testFlag(KeyEvent::Shift)) {
                scrollToBottom();
            }
            else {
                scrollY(d->y->animation().target() + pageSize/2, .3f);
            }
            return true;
        default:
            break;
        }
    }
    return GuiWidget::handleEvent(event);
}
void ScrollAreaWidget::scrollToTop(TimeSpan span) {
    if (d->origin == Top) {
        scrollY(0, span);
    }
    else {
        scrollY(maximumScrollY().valuei(), span);
    }
}
void ScrollAreaWidget::scrollToBottom(TimeSpan span) {
    if (d->origin == Top) {
        scrollY(maximumScrollY().valuei(), span);
    }
    else {
        scrollY(0, span);
    }
}
void ScrollAreaWidget::scrollToLeft(TimeSpan span) {
    scrollX(0, span);
}
void ScrollAreaWidget::scrollToRight(TimeSpan span) {
    scrollX(maximumScrollX().valuei(), span);
}
void ScrollAreaWidget::scrollToWidget(GuiWidget const &widget, TimeSpan span) {
    int off = widget.rule().midY().valuei() - contentRule().top().valuei() -
              rule().height().valuei()/2;
    scrollY(off, span);
}
ScrollAreaWidget &ScrollAreaWidget::findTopmostScrollable() {
    for (GuiWidget *parent = parentGuiWidget(); parent; parent = parent->parentGuiWidget()) {
        if (ScrollAreaWidget *scroll = maybeAs<ScrollAreaWidget>(parent)) {
            if (scroll->isScrollable()) {
                return *scroll;
            }
        }
    }
    return *this;
}
void ScrollAreaWidget::glInit() {
    d->glInit();
}
void ScrollAreaWidget::glDeinit() {
    d->glDeinit();
}
void ScrollAreaWidget::glMakeScrollIndicatorGeometry(GuiVertexBuilder &verts,
                                                     Vector2f const &origin) {
    if (d->scrollOpacity <= 0) return;
    auto rects = d->scrollIndicatorRects(origin);
    d->scrollBarVisRect  = rects.first;
    d->scrollBarLaneRect = rects.second;
    if (d->scrollBarVisRect.isNull()) return;
    Vector4f const barOpacity { 1, 1, 1, d->scrollOpacity };
    verts.makeQuad(d->scrollBarVisRect.expanded((d->scrollBarHover ? pointsToPixels(1) : 0)),
                   barOpacity * d->scrollBarColor,
                   d->indicatorUv);
    if (d->scrollBarHover) {
        verts.makeQuad(d->scrollBarVisRect, barOpacity * d->scrollBarHoverColor, d->indicatorUv);
    }
}
/*void ScrollAreaWidget::viewResized()
{
    GuiWidget::viewResized();
}*/
void ScrollAreaWidget::update() {
    GuiWidget::update();
    if (d->indicatorAnimating) {
        requestGeometry();
    }
    if (d->scrollOpacity.done()) {
        d->indicatorAnimating = false;
    }
    if (d->x->value() > d->maxX->value()) {
        d->x->set(d->maxX->value());
    }
    if (d->y->value() > d->maxY->value()) {
        d->y->set(d->maxY->value());
    }
}
void ScrollAreaWidget::drawContent() {
    if (d->indicatorDrawEnabled) {
        auto &painter = root().painter();
        setIndicatorUv(root().atlas().imageRectf(root().solidWhitePixel()).middle());
        d->verts.clear(); //DefaultVertexBuf::Builder verts;
        glMakeScrollIndicatorGeometry(d->verts, rule().recti().topLeft + margins().toVector().xy());
        if (d->verts) {
            painter.setColor(Vector4f(1, 1, 1, visibleOpacity()));
            painter.drawTriangleStrip(d->verts);
        }
    }
}
} // namespace de
namespace de {
using namespace ui;
/**
 * Popup for editing the value of a slider.
 */
class ValuePopup : public PopupWidget {
public:
    ValuePopup(SliderWidget &slider) {
        setContent(_edit = new LineEditWidget);
        _edit->setSignalOnEnter(true);
        connect(_edit, SIGNAL(enterPressed(QString)), &slider, SLOT(setValueFromText(QString)));
        connect(_edit, SIGNAL(enterPressed(QString)), this, SLOT(close()));
        _edit->rule().setInput(Rule::Width, slider.rule("slider.editor"));
        _edit->setText(QString::number(slider.value() * slider.displayFactor(), 'f',
                                       slider.precision()));
    }
    LineEditWidget &editor() const {
        return *_edit;
    }
    void preparePanelForOpening() {
        PopupWidget::preparePanelForOpening();
        root().setFocus(_edit);
    }
    void panelClosing() {
        PopupWidget::panelClosing();
        root().setFocus(0);
    }
private:
    LineEditWidget *_edit;
};
DENG_GUI_PIMPL(SliderWidget) {
    ddouble value;
    Ranged range;
    ddouble step;
    int precision;
    ddouble displayFactor;
    String minLabel;
    String maxLabel;
    enum State {
        Inert,
        Hovering,
        Grabbed
    };
    State state;
    Vector2i grabFrom;
    ddouble grabValue;
    SafeWidgetPtr<ValuePopup> editPop;
    bool animating;
    Animation pos;
    int endLabelSize;
    Animation frameOpacity;
    ColorBank::Colorf textColor;
    ColorBank::Colorf invTextColor;
    enum Labels {
        Value,
        Start,
        End,
        NUM_LABELS
    };
    TextDrawable labels[NUM_LABELS];
    GuiVertexBuilder verts;
    Impl(Public *i)
        : Base(i),
          value(0),
          range(0, 0),
          step(0),
          precision(0),
          displayFactor(1),
          state(Inert),
          animating(false) {
        self().setFont("slider.label");
        frameOpacity.setValue(.25f);
        updateStyle();
    }
    void updateStyle() {
        textColor    = style().colors().colorf("text");
        invTextColor = style().colors().colorf("inverted.text");
        endLabelSize = rule("slider.label").valuei();
        for (int i = 0; i < int(NUM_LABELS); ++i) {
            labels[i].setFont(i == Value? style().fonts().font("slider.value") : self().font());
            labels[i].setLineWrapWidth(endLabelSize);
        }
    }
    void glInit() {
        /*shaders().build(drawable.program(), "generic.textured.color_ucolor")
                << uMvpMatrix << uColor << uAtlas();*/
        for (int i = 0; i < int(NUM_LABELS); ++i) {
            labels[i].init(atlas(), self().font());
        }
        updateValueLabel();
        updateRangeLabels();
    }
    void glDeinit() {
        verts.clear();
        for (int i = 0; i < int(NUM_LABELS); ++i) {
            labels[i].deinit();
        }
    }
    Rectanglei sliderRect() const {
        Rectanglei const rect = self().contentRect();
        return Rectanglei(Vector2i(rect.topLeft.x + endLabelSize,     rect.topLeft.y),
                          Vector2i(rect.bottomRight.x - endLabelSize, rect.bottomRight.y));
    }
    Rectanglei sliderValueRect() const {
        Rectanglei const area = sliderRect();
        ddouble i = range.size() > 0? (pos - range.start) / range.size() : 0;
        return Rectanglei::fromSize(Vector2i(area.topLeft.x +
                                             (area.width() - endLabelSize) * i,
                                             area.topLeft.y),
                                    Vector2ui(endLabelSize, area.height()));
    }
    void updateGeometry() {
        Rectanglei rect;
        if (self().hasChangedPlace(rect)) {
            self().requestGeometry();
        }
        for (int i = 0; i < int(NUM_LABELS); ++i) {
            if (labels[i].update()) {
                self().requestGeometry();
            }
        }
        if (!self().geometryRequested()) return;
        Vector4i const margin = self().margins().toVector();
        rect = rect.adjusted(margin.xy(), -margin.zw());
        verts.clear();
        self().glMakeGeometry(verts);
        Rectanglei sliderArea = sliderRect();
        int numDots = de::clamp(5, round<int>(range.size() / step) + 1, 11);
        int dotSpace = (sliderArea.width() - endLabelSize) / pointsToPixels(1.f);
        int dotX = sliderArea.topLeft.x + endLabelSize / 2;
        float altAlpha = 0;
        if (dotSpace / numDots > 30) {
            altAlpha = .5f;
            numDots = 2 * numDots + 1;
        }
        Image::Size const dotSize = atlas().imageRect(root().tinyDot()).size();
        for (int i = 0; i < numDots; ++i) {
            Vector2i dotPos(dotX + pointsToPixels(dotSpace * float(i) / float(numDots - 1)),
                            sliderArea.middle().y);
            Vector4f dotColor = textColor;
            dotColor.w *= .666f;
            if (altAlpha > 0 && i % 2) {
                dotColor.w *= altAlpha;
            }
            verts.makeQuad(Rectanglei::fromSize(dotPos - dotSize.toVector2i()/2, dotSize),
                           dotColor, atlas().imageRectf(root().tinyDot()));
        }
        Rectanglei slider = sliderValueRect();
        verts.makeQuad(slider.expanded(2), state == Grabbed? textColor : invTextColor,
                       atlas().imageRectf(root().solidWhitePixel()).middle());
        verts.makeFlexibleFrame(slider.expanded(pointsToPixels(5)),
                                pointsToPixels(6),
                                Vector4f(1, 1, 1, frameOpacity),
                                atlas().imageRectf(root().boldRoundCorners()));
        if (labels[Start].isReady()) {
            labels[Start].makeVertices(verts,
                                       Rectanglei(rect.topLeft,
                                                  Vector2i(rect.topLeft.x + endLabelSize,
                                                           rect.bottomRight.y)),
                                       ui::AlignCenter, ui::AlignCenter,
                                       textColor);
        }
        if (labels[End].isReady()) {
            labels[End].makeVertices(verts,
                                     Rectanglei(Vector2i(rect.bottomRight.x - endLabelSize,
                                                         rect.topLeft.y),
                                                rect.bottomRight),
                                     ui::AlignCenter, ui::AlignCenter,
                                     textColor);
        }
        if (labels[Value].isReady()) {
            labels[Value].makeVertices(verts, slider,
                                       ui::AlignCenter, ui::AlignCenter,
                                       state == Grabbed? invTextColor : textColor);
        }
        /*drawable.buffer<DefaultVertexBuf>()
                .setVertices(gl::TriangleStrip, verts, animating? gl::Dynamic : gl::Static);*/
        self().requestGeometry(false);
    }
    void draw() {
        updateGeometry();
        auto &painter = root().painter();
        painter.setColor(Vector4f(1, 1, 1, self().visibleOpacity()));
        painter.drawTriangleStrip(verts); //drawable.draw();
    }
    void setState(State st) {
        if (state == st) return;
        state = st;
        animating = true;
        switch (st) {
        case Inert:
            frameOpacity.setValue(.25f, .6);
            break;
        case Hovering:
            frameOpacity.setValue(.5f, .15);
            break;
        case Grabbed:
            frameOpacity.setValue(.8f);
            break;
        }
        self().requestGeometry();
    }
    void updateHover(Vector2i const &pos) {
        if (state == Grabbed) return;
        if (self().hitTest(pos)) {
            if (state == Inert) setState(Hovering);
        }
        else if (state == Hovering) {
            setState(Inert);
        }
    }
    void updateValueLabel() {
        if (!minLabel.isEmpty() && fequal(value, range.start)) {
            labels[Value].setText(minLabel);
        }
        else if (!maxLabel.isEmpty() && fequal(value, range.end)) {
            labels[Value].setText(maxLabel);
        }
        else {
            labels[Value].setText(QString::number(value * displayFactor, 'f', precision));
        }
    }
    void setValue(ddouble v) {
        if (step > 0) {
            v = de::round<ddouble>((v - range.start) / step) * step + range.start;
        }
        v = range.clamp(v);
        if (!fequal(v, value)) {
            value = v;
            updateValueLabel();
            animating = true;
            pos.setValue(float(value), 0.1);
            self().requestGeometry();
            emit self().valueChanged(v);
            DENG2_FOR_PUBLIC_AUDIENCE2(Change, i) i->sliderValueChanged(self());
        }
    }
    void updateRangeLabels() {
        labels[Start].setText(minLabel.isEmpty()? QString::number(range.start * displayFactor, 'f', precision) : minLabel);
        labels[End]  .setText(maxLabel.isEmpty()? QString::number(range.end   * displayFactor, 'f', precision) : maxLabel);
    }
    void startGrab(MouseEvent const &ev) {
        if (sliderValueRect().contains(ev.pos())) {
            setState(Grabbed);
            grabFrom = ev.pos();
            grabValue = value;
        }
    }
    void updateGrab(MouseEvent const &ev) {
        DENG2_ASSERT(state == Grabbed);
        Rectanglei const area = sliderRect();
        ddouble unitsPerPixel = range.size() / (area.width() - endLabelSize);
        setValue(grabValue + (ev.pos().x - grabFrom.x) * unitsPerPixel);
        emit self().valueChangedByUser(value);
    }
    ddouble clickStep() const {
        if (step > 0) {
            return step;
        }
        return 1.0 / std::pow(10.0, precision) / displayFactor;
    }
    void endGrab(MouseEvent const &ev) {
        if (state == Grabbed) {
            setState(Inert);
            updateHover(ev.pos());
        }
        else {
            Rectanglei const rect = self().contentRect();
            if (rect.contains(ev.pos())) {
                if (ev.pos().x < rect.left() + endLabelSize) {
                    setValue(value - clickStep());
                    emit self().valueChangedByUser(value);
                }
                else if (ev.pos().x > rect.right() - endLabelSize) {
                    setValue(value + clickStep());
                    emit self().valueChangedByUser(value);
                }
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(Change)
};
DENG2_AUDIENCE_METHOD(SliderWidget, Change)
SliderWidget::SliderWidget(String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    rule().setInput(Rule::Width,  rule("slider.width"))
          .setInput(Rule::Height, OperatorRule::maximum(
                        style().fonts().font("default").height(),
                        font().height()) + margins().height());
}
void SliderWidget::setRange(Rangei const &intRange, int step) {
    setRange(Ranged(intRange.start, intRange.end), ddouble(step));
}
void SliderWidget::setRange(Rangef const &floatRange, float step) {
    setRange(Ranged(floatRange.start, floatRange.end), ddouble(step));
}
void SliderWidget::setRange(Ranged const &doubleRange, ddouble step) {
    d->range = doubleRange;
    d->step = step;
    d->updateRangeLabels();
    d->setValue(d->value);
    d->pos.finish();
}
void SliderWidget::setPrecision(int precisionDecimals) {
    d->precision = precisionDecimals;
    d->updateValueLabel();
    d->updateRangeLabels();
}
void SliderWidget::setStep(double step) {
    d->step = step;
}
void SliderWidget::setValue(ddouble value) {
    d->setValue(value);
}
void SliderWidget::setMinLabel(const String &labelText) {
    d->minLabel = labelText;
    d->updateRangeLabels();
    d->updateValueLabel();
}
void SliderWidget::setMaxLabel(const String &labelText) {
    d->maxLabel = labelText;
    d->updateRangeLabels();
    d->updateValueLabel();
}
void SliderWidget::setDisplayFactor(ddouble factor) {
    d->displayFactor = factor;
    d->updateRangeLabels();
    d->updateValueLabel();
}
Ranged SliderWidget::range() const {
    return d->range;
}
ddouble SliderWidget::value() const {
    return d->value;
}
int SliderWidget::precision() const {
    return d->precision;
}
ddouble SliderWidget::displayFactor() const {
    return d->displayFactor;
}
/*void SliderWidget::viewResized()
{
    GuiWidget::viewResized();
    d->uMvpMatrix = root().projMatrix2D();
}*/
void SliderWidget::update() {
    GuiWidget::update();
    if (d->animating) {
        requestGeometry();
        d->animating = !d->pos.done() || !d->frameOpacity.done();
    }
}
void SliderWidget::drawContent() {
    d->draw();
}
bool SliderWidget::handleEvent(Event const &event) {
    if (event.type() == Event::MousePosition) {
        if (d->state == Impl::Grabbed) {
            d->updateGrab(event.as<MouseEvent>());
            return true;
        }
        d->updateHover(event.as<MouseEvent>().pos());
    }
    if (d->state != Impl::Inert) {
        switch (handleMouseClick(event, MouseEvent::Left)) {
        case MouseClickStarted:
            d->startGrab(event.as<MouseEvent>());
            return true;
        case MouseClickAborted:
        case MouseClickFinished:
            d->endGrab(event.as<MouseEvent>());
            return true;
        default:
            break;
        }
    }
    if (d->state != Impl::Grabbed) {
        switch (handleMouseClick(event, MouseEvent::Right)) {
        case MouseClickFinished:
            if (!d->editPop) {
                ValuePopup *pop = new ValuePopup(*this);
                d->editPop.reset(pop);
                pop->setAnchorAndOpeningDirection(rule(),
                        rule().recti().middle().y < root().viewHeight().valuei()/2? ui::Down : ui::Up);
                pop->setDeleteAfterDismissed(true);
                root().addOnTop(pop);
                pop->open();
            }
            return true;
        case MouseClickStarted:
        case MouseClickAborted:
            return true;
        default:
            break;
        }
    }
    return GuiWidget::handleEvent(event);
}
void SliderWidget::setValueFromText(QString text) {
    setValue(text.toDouble() / d->displayFactor);
    emit valueChangedByUser(d->value);
}
void SliderWidget::glInit() {
    d->glInit();
}
void SliderWidget::glDeinit() {
    d->glDeinit();
}
void SliderWidget::updateStyle() {
    d->updateStyle();
}
} // namespace de
Q_DECLARE_METATYPE(de::ui::Item const *)
namespace de {
DENG_GUI_PIMPL(TabWidget)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate)
, DENG2_OBSERVES(ui::Data,             Addition)
, DENG2_OBSERVES(ui::Data,             OrderChange)
, DENG2_OBSERVES(ButtonWidget,         Press) {
    ui::Data::Pos current = 0;
    MenuWidget *buttons = nullptr;
    bool needUpdate = false;
    bool invertedStyle = false;
    LabelWidget *selected = nullptr;
    AnimationRule *selLeft = nullptr;
    AnimationRule *selWidth = nullptr;
    Impl(Public *i) : Base(i) {
        self().add(buttons = new MenuWidget);
        buttons->enableScrolling(false);
        buttons->margins().set("");
        buttons->setGridSize(0, ui::Expand, 1, ui::Expand, GridLayout::ColumnFirst);
        buttons->organizer().audienceForWidgetCreation() += this;
        buttons->organizer().audienceForWidgetUpdate()   += this;
        buttons->items().audienceForAddition()    += this;
        buttons->items().audienceForOrderChange() += this;
        buttons->rule()
                .setInput(Rule::AnchorX, self().rule().left() + self().rule().width() / 2)
                .setInput(Rule::Top, self().rule().top())
                .setAnchorPoint(Vector2f(.5f, 0));
        self().add(selected = new LabelWidget);
    }
    ~Impl() {
        releaseRef(selLeft);
        releaseRef(selWidth);
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &) {
        ButtonWidget &btn = widget.as<ButtonWidget>();
        btn.setSizePolicy(ui::Expand, ui::Expand);
        btn.setFont("tab.label");
        btn.margins().set("dialog.gap");
        btn.set(Background());
        btn.setBehavior(Focusable, UnsetFlags);
        btn.audienceForPress() += this;
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) {
        widget.as<ButtonWidget>().setShortcutKey(item.as<TabItem>().shortcutKey());
    }
    void buttonPressed(ButtonWidget &button) {
        self().setCurrent(buttons->items().find(*buttons->organizer().findItemForWidget(button)));
    }
    void dataItemAdded(ui::Data::Pos, ui::Item const &) {
        needUpdate = true;
    }
    void dataItemOrderChanged() {
        needUpdate = true;
    }
    void setCurrent(ui::Data::Pos pos) {
        if (current != pos && pos < buttons->items().size()) {
            current = pos;
            updateSelected();
            emit self().currentTabChanged();
        }
    }
    void updateSelected() {
        selected->set(Background(style().colors().colorf(invertedStyle? "tab.inverted.selected"
                                                                      : "tab.selected")));
        for (ui::Data::Pos i = 0; i < buttons->items().size(); ++i) {
            bool const sel = (i == current);
            ButtonWidget &w = buttons->itemWidget<ButtonWidget>(buttons->items().at(i));
            w.setFont(sel? "tab.selected" : "tab.label");
            w.setOpacity(sel? 1.f : 0.7f, 0.4);
            if (!invertedStyle) {
                w.setTextColor(sel? "tab.selected" : "text");
                w.setHoverTextColor(sel? "tab.selected" : "text");
            }
            else {
                w.setTextColor(sel? "tab.inverted.selected" : "inverted.text");
                w.setHoverTextColor(sel? "tab.inverted.selected" : "inverted.text");
            }
            if (sel) {
                TimeSpan span = .2;
                if (!selLeft) {
                    selLeft  = new AnimationRule(0);
                    selWidth = new AnimationRule(0);
                    selected->rule()
                        .setInput(Rule::Width,  *selWidth)
                        .setInput(Rule::Left,   *selLeft);
                    span = 0;
                }
                selLeft ->set(w.rule().left(),  span);
                selWidth->set(w.rule().width(), span);
                selected->rule()
                    .setInput(Rule::Height, rule("halfunit"))
                    .setInput(Rule::Top,    w.rule().bottom());
            }
        }
    }
    bool handleShortcutKey(KeyEvent const &key) {
        foreach (auto *w, buttons->childWidgets()) {
            if (ButtonWidget *but = maybeAs<ButtonWidget>(w)) {
                if (but->handleShortcut(key)) {
                    return true;
                }
            }
        }
        return false;
    }
};
TabWidget::TabWidget(String const &name)
    : GuiWidget(name), d(new Impl(this)) {
    rule().setInput(Rule::Height, d->buttons->rule().height());
}
void TabWidget::useInvertedStyle() {
    d->invertedStyle = true;
}
void TabWidget::clearItems() {
    if (d->selLeft) {
        d->selLeft->set(d->selLeft->value());
    }
    items().clear();
}
ui::Data &TabWidget::items() {
    return d->buttons->items();
}
ui::Data::Pos TabWidget::current() const {
    return d->current;
}
TabItem &TabWidget::currentItem() {
    DENG2_ASSERT(d->current < items().size());
    return items().at(d->current).as<TabItem>();
}
void TabWidget::setCurrent(ui::Data::Pos itemPos) {
    d->setCurrent(itemPos);
}
void TabWidget::update() {
    GuiWidget::update();
    if (d->needUpdate) {
        d->updateSelected();
        d->needUpdate = false;
    }
}
bool TabWidget::handleEvent(Event const &ev) {
    if (isEnabled()) {
        if (ev.isKeyDown()) {
            KeyEvent const &key = ev.as<KeyEvent>();
            if (d->handleShortcutKey(key)) {
                return true;
            }
        }
    }
    return GuiWidget::handleEvent(ev);
}
} // namespace de
namespace de {
static TimeSpan const SWITCH_ANIM_SPAN = 0.3;
DENG2_PIMPL(ToggleWidget),
DENG2_OBSERVES(ButtonWidget, Press) {
    class ToggleProceduralImage : public ProceduralImage {
    public:
        ToggleProceduralImage(GuiWidget &owner)
            : _owner(owner),
              _pos(0, Animation::EaseBoth),
              _animating(false) {
            const Image &img = style().images().image("toggle.onoff");
            setPointSize(img.size() * img.pointRatio());
            updateStyle();
        }
        Style const &style() const { return _owner.style(); }
        Atlas &atlas() const { return _owner.root().atlas(); }
        void setState(ToggleState st, bool animate) {
            _pos.setValue(st == Active? 1 : 0, animate? SWITCH_ANIM_SPAN : TimeSpan());
            _animating = true;
        }
        void finishAnimation() {
            _pos.finish();
        }
        bool update() override {
            if (_animating) {
                if (_pos.done()) _animating = false;
                return true;
            }
            return false;
        }
        void glMakeGeometry(GuiVertexBuilder &verts, Rectanglef const &rect) override {
            float p = _pos;
            Rectanglei const recti(rect.topLeft.toVector2i(), rect.bottomRight.toVector2i());
            float c = (.3f + .33f * p);
            verts.makeQuad(recti, (_accentColor * p + _textColor * (1-p)) * Vector4f(c, c, c, 1),
                           atlas().imageRectf(_owner.root().solidWhitePixel()).middle());
            Id onOff = _owner.root().styleTexture("toggle.onoff");
            verts.makeQuad(recti, _accentColor * p + _textColor * (1-p) * .8f, atlas().imageRectf(onOff));
            const int  flipWidth = pointSize().x - pointSize().y + 2;
            Rectanglei flip      = Rectanglei::fromSize(
                recti.topLeft + pointsToPixels(Vector2i(
                                    1 + de::round<int>(p * (pointSize().x - flipWidth)), 1)),
                pointsToPixels(Vector2ui(flipWidth, pointSize().y) - Vector2ui(2, 2)));
            verts.makeQuad(flip, _bgColor * Vector4f(1, 1, 1, 3),
                           atlas().imageRectf(_owner.root().solidWhitePixel()).middle());
        }
        void updateStyle() {
            _bgColor     = style().colors().colorf("background").min(Vector4f(0, 0, 0, 1));
            _accentColor = style().colors().colorf("accent");
            _textColor   = style().colors().colorf("text");
        }
    private:
        GuiWidget &_owner;
        Animation _pos;
        bool _animating;
        ColorBank::Colorf _bgColor;
        ColorBank::Colorf _accentColor;
        ColorBank::Colorf _textColor;
    };
    ToggleState state;
    ToggleProceduralImage *procImage; // not owned
    bool hasBeenUpdated = false;
    Impl(Public *i, Flags const &flags)
        : Base(i)
        , state(Inactive)
        , procImage(!(flags & WithoutIndicator)? new ToggleProceduralImage(*i) : nullptr) {
        if (procImage) self().setImage(procImage); // base class owns it
        self().audienceForPress() += this;
    }
    void buttonPressed(ButtonWidget &) {
        self().setActive(self().isInactive());
        emit self().stateChangedByUser(self().toggleState());
    }
    DENG2_PIMPL_AUDIENCE(Toggle)
};
DENG2_AUDIENCE_METHOD(ToggleWidget, Toggle)
ToggleWidget::ToggleWidget(Flags const &flags, String const &name)
    : ButtonWidget(name)
    , d(new Impl(this, flags)) {
    setTextAlignment(ui::AlignRight);
    setTextLineAlignment(ui::AlignLeft);
}
void ToggleWidget::setToggleState(ToggleState state, bool notify) {
    if (d->state != state) {
        d->state = state;
        if (d->procImage) {
            d->procImage->setState(state, hasBeenUpdated());
        }
        if (notify) {
            DENG2_FOR_AUDIENCE2(Toggle, i) i->toggleStateChanged(*this);
        }
        emit stateChanged(state);
    }
}
ToggleWidget::ToggleState ToggleWidget::toggleState() const {
    return d->state;
}
void ToggleWidget::finishAnimation() {
    if (d->procImage) {
        d->procImage->finishAnimation();
    }
}
} // namespace de
namespace de {
DENG_GUI_PIMPL(VariableArrayWidget)
, DENG2_OBSERVES(Variable, Deletion)
, DENG2_OBSERVES(Variable, Change  )
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetCreation)
, DENG2_OBSERVES(ChildWidgetOrganizer, WidgetUpdate) {
    Variable *    var = nullptr;
    IndirectRule *maxWidth;
    MenuWidget *  menu;
    ButtonWidget *addButton;
    ButtonWidget *deleteButton;
    ui::DataPos   hoverItem      = ui::Data::InvalidPos;
    bool          mouseWasInside = false;
    struct HoverHandler : public GuiWidget::IEventHandler {
        VariableArrayWidget &owner;
        HoverHandler(VariableArrayWidget &owner) : owner(owner) {}
        bool handleEvent(GuiWidget &widget, Event const &event) {
            if (event.isMouse() && widget.hitTest(event)) {
                owner.d->setHoverItem(widget);
            }
            return false;
        }
    };
    Impl(Public *i, Variable &var) : Base(i), var(&var) {
        maxWidth = new IndirectRule;
        maxWidth->setSource(rule("list.width"));
        menu         = new MenuWidget;
        addButton    = new ButtonWidget;
        deleteButton = new ButtonWidget;
        menu->organizer().audienceForWidgetCreation() += this;
        menu->organizer().audienceForWidgetUpdate()   += this;
        menu->setGridSize(1, ui::Expand, 0, ui::Expand);
        menu->layout().setRowPadding(2 * rule("unit"));
        updateFromVariable();
        var.audienceForDeletion() += this;
        var.audienceForChange()   += this;
    }
    ~Impl() override {
        releaseRef(maxWidth);
    }
    void setHoverItem(GuiWidget &widget) {
        hoverItem = menu->findItem(widget);
        deleteButton->show();
        deleteButton->rule().setMidAnchorY(widget.rule().midY());
    }
    void widgetCreatedForItem(GuiWidget &widget, ui::Item const &item) override {
        auto &label = widget.as<LabelWidget>();
        label.setSizePolicy(ui::Expand, ui::Expand);
        label.setMaximumTextWidth(*maxWidth);
        widget.margins().setRight("").setTopBottom("");
        widget.margins().setLeft("");//deleteButton->rule().width());
        widget.addEventHandler(new HoverHandler(self()));
        self().elementCreated(label, item);
    }
    void widgetUpdatedForItem(GuiWidget &widget, ui::Item const &item) override {
        widget.as<LabelWidget>().setText(item.label());
    }
    void updateFromVariable() {
        if (!var) return;
        menu->items().clear();
        if (auto const *array = maybeAs<ArrayValue>(var->value())) {
            for (Value const *value : array->elements()) {
                menu->items() << self().makeItem(*value);
            }
        }
        else {
            String const value = var->value().asText();
            if (!value.isEmpty()) {
                menu->items() << self().makeItem(var->value());
            }
        }
    }
    void setVariableFromWidget() {
        if (!var) return;
        var->audienceForChange() -= this;
        if (menu->items().isEmpty()) {
            var->set(new TextValue());
        }
        else if (menu->items().size() == 1) {
            var->set(new TextValue(menu->items().at(0).data().toString()));
        }
        else {
            auto *array = new ArrayValue;
            for (ui::DataPos i = 0; i < menu->items().size(); ++i) {
                array->add(new TextValue(menu->items().at(i).data().toString()));
            }
            var->set(array);
        }
        var->audienceForChange() += this;
    }
    void variableValueChanged(Variable &, Value const &) override {
        updateFromVariable();
    }
    void variableBeingDeleted(Variable &) override {
        var = nullptr;
        self().disable();
    }
};
VariableArrayWidget::VariableArrayWidget(Variable &variable, String const &name)
    : GuiWidget(name)
    , d(new Impl(this, variable)) {
    d->deleteButton->setSizePolicy(ui::Expand, ui::Expand);
    d->deleteButton->setStyleImage("close.ring", "default");
    d->deleteButton->margins().setLeft(RuleBank::UNIT).setRight("dialog.gap");
    d->deleteButton->setBehavior(Focusable, UnsetFlags);
    d->deleteButton->set(Background());
    d->deleteButton->hide();
    d->menu->margins()
            .setLeft(d->deleteButton->rule().width())
            .setBottom("dialog.gap");
    d->menu->enableScrolling(false);
    d->menu->enablePageKeys(false);
    d->menu->rule()
            .setLeftTop(/*margins().left() + */rule().left(),
                        margins().top()  + rule().top())
            .setInput(Rule::Right, rule().right() - 2 * rule("gap"));
    d->addButton->setFont("small");
    d->addButton->setStyleImage("create", d->addButton->fontId());
    d->addButton->setTextAlignment(ui::AlignRight);
    d->addButton->setSizePolicy(ui::Expand, ui::Expand);
    AutoRef<Rule> totalWidth(OperatorRule::maximum(d->menu->rule().width(),
                                                   d->deleteButton->rule().width() +
                                                   d->addButton->contentWidth()));
    d->addButton->rule()
            .setLeftTop(d->deleteButton->rule().width() + d->menu->rule().left(),
                        d->menu->rule().bottom());
    d->deleteButton->rule().setInput(Rule::Left, d->menu->rule().left());
    d->deleteButton->setActionFn([this] () {
        d->deleteButton->hide();
        d->menu->items().remove(d->hoverItem);
        setVariableFromWidget();
    });
    rule().setSize(totalWidth + margins().width(),
                   d->menu->rule().height() + d->addButton->rule().height() + margins().height());
    add(d->menu);
    add(d->deleteButton);
    add(d->addButton);
    d->menu->updateLayout();
}
Variable &VariableArrayWidget::variable() const {
    return *d->var;
}
MenuWidget &VariableArrayWidget::elementsMenu() {
    return *d->menu;
}
String VariableArrayWidget::labelForElement(Value const &value) const {
    return value.asText();
}
void VariableArrayWidget::elementCreated(LabelWidget &, const ui::Item &) {
}
ButtonWidget &VariableArrayWidget::addButton() {
    return *d->addButton;
}
ButtonWidget *VariableArrayWidget::detachAddButton(Rule const &contentWidth) {
    d->addButton->setSizePolicy(ui::Expand, ui::Expand);
    d->addButton->orphan();
    d->maxWidth->setSource(contentWidth - margins().width());
    d->menu->setGridSize(1, ui::Fixed, 0, ui::Expand);
    d->menu->rule().setInput(Rule::Width, contentWidth - margins().width());
    rule().setSize(contentWidth,
                   d->menu->rule().height() + margins().height());
    return d->addButton;
}
ui::Item *VariableArrayWidget::makeItem(Value const &value) {
    auto *item = new ui::Item(ui::Item::ShownAsLabel, labelForElement(value));
    item->setData(value.asText());
    return item;
}
bool VariableArrayWidget::handleEvent(Event const &event) {
    if (event.isMouse()) {
        MouseEvent const &mouse = event.as<MouseEvent>();
        bool const isInside = rule().recti().contains(mouse.pos());
        if (d->mouseWasInside && !isInside) {
            d->deleteButton->hide();
        }
        d->mouseWasInside = isInside;
    }
    return GuiWidget::handleEvent(event);
}
void VariableArrayWidget::updateFromVariable() {
    d->updateFromVariable();
}
void VariableArrayWidget::setVariableFromWidget() {
    d->setVariableFromWidget();
    emit arrayChanged();
}
} // namespace de
namespace de {
DENG2_PIMPL(VariableChoiceWidget),
DENG2_OBSERVES(Variable, Deletion),
DENG2_OBSERVES(Variable, Change  ) {
    Variable *var;
    VariableType variableType;
    Impl(Public *i, Variable &variable, VariableType vt)
        : Base(i)
        , var(&variable)
        , variableType(vt) {
        updateFromVariable();
        var->audienceForDeletion() += this;
        var->audienceForChange() += this;
    }
    void updateFromVariable() {
        if (!var || self().items().isEmpty()) return;
        if (variableType == Text) {
            self().setSelected(self().items().findData(var->value().asText()));
        }
        else {
            self().setSelected(self().items().findData(var->value().asNumber()));
        }
    }
    void setVariableFromWidget() {
        if (!var) return;
        var->audienceForChange() -= this;
        if (variableType == Text) {
            var->set(TextValue(self().selectedItem().data().toString()));
        }
        else {
            var->set(NumberValue(self().selectedItem().data().toDouble()));
        }
        var->audienceForChange() += this;
    }
    void variableValueChanged(Variable &, Value const &) {
        updateFromVariable();
    }
    void variableBeingDeleted(Variable &) {
        var = 0;
        self().disable();
    }
};
VariableChoiceWidget::VariableChoiceWidget(Variable &variable, VariableType variableType,
                                           String const &name)
    : ChoiceWidget(name)
    , d(new Impl(this, variable, variableType)) {
    connect(this, SIGNAL(selectionChangedByUser(uint)),
            this, SLOT(setVariableFromWidget()));
}
Variable &VariableChoiceWidget::variable() const {
    if (!d->var) {
        throw VariableMissingError("VariableChoiceWidget::variable",
                                   "Widget is not associated with a variable");
    }
    return *d->var;
}
void VariableChoiceWidget::updateFromVariable() {
    d->updateFromVariable();
}
void VariableChoiceWidget::setVariableFromWidget() {
    d->setVariableFromWidget();
}
} // namespace de
namespace de {
DENG2_PIMPL(VariableLineEditWidget),
DENG2_OBSERVES(Variable, Deletion),
DENG2_OBSERVES(Variable, Change  ) {
    Variable *var;
    Impl(Public *i, Variable &variable) : Base(i), var(&variable) {
        updateFromVariable();
        var->audienceForDeletion() += this;
        var->audienceForChange() += this;
    }
    void updateFromVariable() {
        if (!var) return;
        self().setText(var->value<TextValue>());
    }
    void setVariableFromWidget() {
        if (!var) return;
        var->audienceForChange() -= this;
        var->set(TextValue(self().text()));
        var->audienceForChange() += this;
    }
    void variableValueChanged(Variable &, Value const &) {
        updateFromVariable();
    }
    void variableBeingDeleted(Variable &) {
        var = 0;
        self().disable();
    }
};
VariableLineEditWidget::VariableLineEditWidget(Variable &variable, String const &name)
    : LineEditWidget(name), d(new Impl(this, variable)) {
    connect(this, SIGNAL(editorContentChanged()),
            this, SLOT(setVariableFromWidget()));
}
Variable &VariableLineEditWidget::variable() const {
    if (!d->var) {
        throw VariableMissingError("VariableLineEditWidget::variable",
                                   "Widget is not associated with a variable");
    }
    return *d->var;
}
void VariableLineEditWidget::updateFromVariable() {
    d->updateFromVariable();
}
void VariableLineEditWidget::setVariableFromWidget() {
    d->setVariableFromWidget();
}
} // namespace de
namespace de {
DENG2_PIMPL(VariableSliderWidget)
, DENG2_OBSERVES(Variable, Deletion)
, DENG2_OBSERVES(Variable, Change  ) {
    ValueType valueType = VariableSliderWidget::Number;
    Variable *var;
    Impl(Public *i, Variable &variable) : Base(i), var(&variable) {
        var->audienceForDeletion() += this;
        var->audienceForChange()   += this;
    }
    void init() {
        self().updateFromVariable();
        QObject::connect(thisPublic, SIGNAL(valueChangedByUser(double)),
                         thisPublic, SLOT(setVariableFromWidget()));
    }
    void updateFromVariable() {
        if (!var) return;
        switch (valueType) {
        case VariableSliderWidget::Number:
            self().setValue(var->value<NumberValue>().asNumber());
            break;
        case VariableSliderWidget::Animation:
            self().setValue(var->value<AnimationValue>().animation().target());
            break;
        }
    }
    void setVariableFromWidget() {
        if (!var) return;
        var->audienceForChange() -= this;
        switch (valueType) {
        case VariableSliderWidget::Number:
            var->set(NumberValue(self().value()));
            break;
        case VariableSliderWidget::Animation:
            var->value<AnimationValue>().animation().setValue(float(self().value()));
            break;
        }
        var->audienceForChange() += this;
    }
    void variableValueChanged(Variable &, Value const &) {
        updateFromVariable();
    }
    void variableBeingDeleted(Variable &) {
        var = 0;
        self().disable();
    }
};
VariableSliderWidget::VariableSliderWidget(Variable &variable, Ranged const &range,
                                           ddouble step, String const &name)
    : SliderWidget(name)
    , d(new Impl(this, variable)) {
    if (!is<NumberValue>(variable.value())) {
        d->valueType = VariableSliderWidget::Animation;
    }
    setRange(range, step);
    d->init();
}
VariableSliderWidget::VariableSliderWidget(ValueType valueType,
                                           Variable &variable, Ranged const &range,
                                           ddouble step, String const &name)
    : SliderWidget(name)
    , d(new Impl(this, variable)) {
    d->valueType = valueType;
    setRange(range, step);
    d->init();
}
Variable &VariableSliderWidget::variable() const {
    if (!d->var) {
        throw VariableMissingError("VariableSliderWidget::variable",
                                   "Widget is not associated with a variable");
    }
    return *d->var;
}
void VariableSliderWidget::updateFromVariable() {
    d->updateFromVariable();
}
void VariableSliderWidget::setVariableFromWidget() {
    d->setVariableFromWidget();
}
} // namespace de
namespace de {
DENG2_PIMPL(VariableToggleWidget),
DENG2_OBSERVES(Variable,     Deletion),
DENG2_OBSERVES(Variable,     Change  ),
DENG2_OBSERVES(ToggleWidget, Toggle  ) {
    Variable *var;
    NumberValue activeValue;
    NumberValue inactiveValue;
    Impl(Public *i, Variable &variable)
        : Base(i)
        , var(&variable)
        , activeValue(1)
        , inactiveValue(0) {
        updateFromVariable();
        self().audienceForToggle() += this;
        var->audienceForDeletion() += this;
        var->audienceForChange() += this;
    }
    void updateFromVariable() {
        if (!var) return;
        self().setToggleState(!var->value().compare(activeValue)? Active : Inactive,
                            false /*don't notify*/);
    }
    void setVariableFromWidget() {
        if (!var) return;
        var->audienceForChange() -= this;
        var->set(self().isActive()? activeValue : inactiveValue);
        var->audienceForChange() += this;
    }
    void toggleStateChanged(ToggleWidget &) {
        setVariableFromWidget();
    }
    void variableValueChanged(Variable &, Value const &) {
        updateFromVariable();
    }
    void variableBeingDeleted(Variable &) {
        var = 0;
        self().disable();
    }
};
VariableToggleWidget::VariableToggleWidget(Variable &variable, String const &name)
    : ToggleWidget(DefaultFlags, name)
    , d(new Impl(this, variable))
{}
VariableToggleWidget::VariableToggleWidget(String const &label, Variable &variable, String const &name)
    : ToggleWidget(DefaultFlags, name)
    , d(new Impl(this, variable)) {
    setText(label);
}
Variable &VariableToggleWidget::variable() const {
    if (!d->var) {
        throw VariableMissingError("VariableToggleWidget::variable",
                                   "Widget is not associated with a variable");
    }
    return *d->var;
}
void VariableToggleWidget::setActiveValue(double val) {
    d->activeValue = NumberValue(val);
    d->updateFromVariable();
}
void VariableToggleWidget::setInactiveValue(double val) {
    d->inactiveValue = NumberValue(val);
    d->updateFromVariable();
}
} // namespace de
