﻿/** @file animationvalue.cpp  Value that holds an Animation.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/AnimationValue"
#include "de/Reader"
#include "de/ScriptSystem"
#include "de/Writer"
/** @file archive.cpp Collection of named memory blocks stored inside a byte array.
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Archive"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ArrayValue"
#include "de/FunctionValue"
#include "de/NumberValue"
#include "de/Process"
#include "de/Reader"
#include "de/RecordValue"
#include "de/TextValue"
#include "de/Writer"
#include <algorithm>
#include <QTextStream>
/** @file bank.cpp  Abstract data bank with multi-tiered caching.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Bank"
#include "de/Folder"
#include "de/App"
#include "de/Loop"
#include "de/FS"
#include "de/PathTree"
#include "de/WaitableFIFO"
#include "de/Task"
#include "de/TaskPool"
#include "de/LogBuffer"
#include "de/Writer"
#include "de/Reader"
#include "de/math.h"
#include <QThread>
#include <QList>
/** @file bitfield.cpp  Array of integers packed together.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/BitField"
#include <QTextStream>
/** @file bitfield_elements.cpp  Bit field elements.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/BitField"
#include <QMap>
#include <QList>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Block"
#include "de/File"
#include <QCryptographicHash>
#include <cstring>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/BlockValue"
#include "de/Writer"
#include "de/Reader"
#include <QTextStream>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/data/byteorder.h"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ByteRefArray"
#include <cstring>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ByteSubArray"
/** @file conditionaltrigger.cpp  Conditional trigger.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ConditionalTrigger"
#include "de/Value"
#include <QSet>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Counted"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Date"
/** @file deletable.cpp
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Deletable"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/DictionaryValue"
#include "de/ArrayValue"
#include "de/RecordValue"
#include "de/ScriptSystem"
#include "de/Writer"
#include "de/Reader"
#include <QTextStream>
/** @file escapeparser.cpp
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/EscapeParser"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/FixedByteArray"
/** @file gzip.cpp  Process gzip data with zlib
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/data/gzip.h"
#include <zlib.h>
/**
 * @file huffman.cpp
 * Huffman codes.
 *
 * Uses predetermined, fixed frequencies optimized for short (size < 128)
 * messages.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/data/huffman.h"
#include "de/App"
#include "de/Log"
#include "de/ByteRefArray"
#define HEAP_PARENT(i)  (((i) + 1)/2 - 1)
#define HEAP_LEFT(i)    (2*(i) + 1)
#define HEAP_RIGHT(i)   (2*(i) + 2)
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Info"
#include "de/App"
#include "de/Folder"
#include "de/Log"
#include "de/LogBuffer"
#include "de/RecordValue"
#include "de/ScriptLex"
#include "de/SourceLineTable"
#include <de/TextValue>
#include <QFile>
/** @file infobank.cpp  Abstract Bank read from Info definitions.
 *
 * @authors Copyright © 2013-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/InfoBank"
#include "de/App"
#include "de/Folder"
#include "de/Package"
#include "de/ScriptedInfo"
/** @file json.cpp  JSON parser and composer.
 *
 * Parses JSON and outputs a QVariant with the data.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/data/json.h"
#include <QVarLengthArray>
#include <de/Block>
#include <de/Error>
#include <de/Log>
#include <de/Record>
#include <de/NoneValue>
#include <de/NumberValue>
#include <de/RecordValue>
#include <de/DictionaryValue>
#include <de/ArrayValue>
#include <QTextStream>
#include <QDebug>
/** @file nativevalue.cpp  Value representing a native object.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NativePointerValue"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NoneValue"
#include "de/Writer"
#include "de/Reader"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/NumberValue"
#include "de/Writer"
#include "de/Reader"
#include "de/math.h"
#include <QTextStream>
/** @file observers.cpp
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Observers"
/** @file path.cpp Textual path composed of segments.
 *
 * @author Copyright © 2010-2013 Daniel Swanson <danij@dengine.net>
 * @author Copyright © 2010-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Path"
#include "de/Reader"
#include "de/Writer"
#include <QList>
#include <de/math.h>
#include <de/ByteRefArray>
#include <cstring> // memset
/** @file pathtree.cpp Tree of Path/data pairs.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Error"
#include "de/Log"
#include "de/StringPool"
#include "de/PathTree"
#include "de/Guard"
#include <QDebug>
/** @file pathtreenode.cpp PathTree::Node implementation.
 *
 * @authors Copyright © 2003-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 * @authors Copyright © 2006-2013 Daniel Swanson <danij@dengine.net>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PathTree"
#include "de/Log"
/** @file pointerset.cpp  Set of pointers.
 *
 * @authors Copyright (c) 2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/PointerSet"
#include <cstdlib>
/** @file profiles.cpp  Abstract set of persistent profiles.
 *
 * @authors Copyright (c) 2016-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Profiles"
#include "de/App"
#include "de/FileSystem"
#include "de/Folder"
#include "de/String"
#include <QTextStream>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Reader"
#include "de/String"
#include "de/Block"
#include "de/ISerializable"
#include "de/IIStream"
#include "de/FixedByteArray"
#include "de/ByteRefArray"
#include "de/data/byteorder.h"
#include <QTextStream>
#include <cstring>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Record"
#include "de/ArrayValue"
#include "de/BlockValue"
#include "de/DictionaryValue"
#include "de/FunctionValue"
#include "de/Info"
#include "de/LogBuffer"
#include "de/NumberValue"
#include "de/Reader"
#include "de/RecordValue"
#include "de/String"
#include "de/TextValue"
#include "de/TimeValue"
#include "de/Variable"
#include "de/Vector"
#include "de/Writer"
#include "de/CompiledRecord"
#include <QTextStream>
#include <functional>
#include <atomic>
/** @file recordaccessor.cpp  Utility class with get*() methods.
 *
 * @authors Copyright (c) 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RecordAccessor"
#include "de/RecordValue"
#include "de/DictionaryValue"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RecordValue"
#include "de/TextValue"
#include "de/RefValue"
#include "de/NoneValue"
#include "de/FunctionValue"
#include "de/Process"
#include "de/Context"
#include "de/Evaluator"
#include "de/Variable"
#include "de/Writer"
#include "de/Reader"
#include "de/ScopeStatement"
#include "de/math.h"
/** @file refuge.cpp  Persistent data storage.
 *
 * @authors Copyright © 2014-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Refuge"
#include "de/App"
#include "de/Archive"
#include "de/ArchiveFolder"
#include "de/Reader"
#include "de/Writer"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/RefValue"
#include "de/Variable"
#include "de/Writer"
/** @file sourcelinetable.cpp  Table for source paths and line numbers.
 *
 * @authors Copyright (c) 2015-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/SourceLineTable"
#include "de/PathTree"
#include <atomic>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/App"
#include "de/String"
#include "de/Block"
#include "de/charsymbols.h"
#include <QDir>
#include <QTextStream>
#include <QRegularExpression>
#include <QUrl> // percent encoding
#include <cstdio>
#include <cstdarg>
/** @file stringpool.cpp  Pool of strings (case insensitive).
 *
 * @author Copyright © 2010-2015 Daniel Swanson <danij@dengine.net>
 * @author Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/StringPool"
#include "de/Reader"
#include "de/Writer"
#include "de/Lockable"
#include "de/Guard"
#include <vector>
#include <list>
#include <set>
#include <algorithm>
#ifdef DENG2_DEBUG
#  include <stdio.h>  /// @todo should use C++
#endif
#define EXPORT_ID(i)    (duint32(i) + 1)
#define IMPORT_ID(i)    (Id((i) - 1))
#define MAXIMUM_VALID_ID (0xffffffff - 1)
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TextValue"
#include "de/NumberValue"
#include "de/ArrayValue"
#include "de/String"
#include "de/Writer"
#include "de/Reader"
#include "de/ScriptSystem"
#include <QTextStream>
#include <list>
#include <cmath>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Time"
#include "de/Date"
#include "de/String"
#include "de/Writer"
#include "de/Reader"
#include "de/HighPerformanceTimer"
#include <QStringList>
#include <QThread>
#include <QDataStream>
/** @file timevalue.cpp Value that holds a Time.
 *
 * @authors Copyright © 2012-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/TimeValue"
#include "de/NumberValue"
#include "de/Reader"
#include "de/Writer"
#include "de/Log"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Value"
#include "de/AnimationValue"
#include "de/ArrayValue"
#include "de/BlockValue"
#include "de/DictionaryValue"
#include "de/FunctionValue"
#include "de/NoneValue"
#include "de/NumberValue"
#include "de/Reader"
#include "de/RecordValue"
#include "de/TextValue"
#include "de/TimeValue"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2009-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Variable"
#include "de/Value"
#include "de/NoneValue"
#include "de/NumberValue"
#include "de/TextValue"
#include "de/ArrayValue"
#include "de/DictionaryValue"
#include "de/BlockValue"
#include "de/TimeValue"
#include "de/RecordValue"
#include "de/Reader"
#include "de/Writer"
#include "de/Log"
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/Writer"
#include "de/String"
#include "de/Block"
#include "de/ISerializable"
#include "de/IOStream"
#include "de/FixedByteArray"
#include "de/ByteRefArray"
#include "de/ByteArrayFile"
#include "de/data/byteorder.h"
#include <QScopedPointer>
/*
 * The Doomsday Engine Project -- libcore
 *
 * Copyright © 2004-2017 Jaakko Keränen <jaakko.keranen@iki.fi>
 *
 * @par License
 * LGPL: http://www.gnu.org/licenses/lgpl.html
 *
 * <small>This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version. This program is distributed in the hope that it
 * will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details. You should have received a copy of
 * the GNU Lesser General Public License along with this program; if not, see:
 * http://www.gnu.org/licenses</small>
 */
#include "de/ZipArchive"
#include "de/Block"
#include "de/ByteArrayFile"
#include "de/ByteSubArray"
#include "de/Date"
#include "de/File"
#include "de/FixedByteArray"
#include "de/ISerializable"
#include "de/LittleEndianByteOrder"
#include "de/LogBuffer"
#include "de/MetadataBank"
#include "de/Reader"
#include "de/Writer"
#include "de/Zeroed"
#include "de/ArchiveFolder"
#include <cstring>
#include <zlib.h>

namespace de {
AnimationValue::AnimationValue(Animation const &anim)
    : _anim(new CountedAnimation(anim))
{}
AnimationValue::~AnimationValue() {
    releaseRef(_anim);
}
AnimationValue::AnimationValue(CountedAnimation *countedAnim)
    : _anim(holdRef(countedAnim))
{}
Record *AnimationValue::memberScope() const {
    return &ScriptSystem::builtInClass("Animation");
}
Value *AnimationValue::duplicate() const {
    return new AnimationValue(*static_cast<Animation *>(_anim)); // makes a separate Animation
}
Value *AnimationValue::duplicateAsReference() const {
    return new AnimationValue(_anim); // refers to the same Animation
}
Value::Text AnimationValue::asText() const {
    return _anim->asText();
}
Value::Number AnimationValue::asNumber() const {
    return Number(_anim->value());
}
bool AnimationValue::isTrue() const {
    return !_anim->done();
}
dint AnimationValue::compare(Value const &value) const {
    Number const a = asNumber();
    Number const b = value.asNumber();
    if (fequal(a, b)) return 0;
    return (a < b? -1 : +1);
}
void AnimationValue::operator >> (Writer &to) const {
    to << SerialId(ANIMATION) << *_anim;
}
void AnimationValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != ANIMATION) {
        throw DeserializationError("AnimationValue::operator <<", "Invalid ID");
    }
    from >> *_anim;
}
Value::Text AnimationValue::typeId() const {
    return "Animation";
}
AnimationValue::CountedAnimation::CountedAnimation(Animation const &anim)
    : Animation(anim)
{}
} // namespace de
namespace de {
DENG2_PIMPL(Archive) {
    IByteArray const *source;
    PathTree *index;
    bool modified;
    Impl(Public &a, IByteArray const *src) : Base(a), source(src), index(0), modified(false)
    {}
    ~Impl() {
        delete index;
    }
    void readEntry(Path const &path, IBlock &deserializedData) const {
        Entry const &entry = static_cast<Entry const &>(
                    index->find(path, PathTree::MatchFull | PathTree::NoBranch));
        if (!entry.size) {
            deserializedData.clear();
            return;
        }
        if (entry.data) {
            deserializedData.copyFrom(*entry.data, 0, entry.data->size());
            return;
        }
        self().readFromSource(entry, path, deserializedData);
    }
};
Archive::Archive() : d(new Impl(*this, 0))
{}
Archive::Archive(IByteArray const &archive) : d(new Impl(*this, &archive))
{}
Archive::~Archive() {
    clear();
}
IByteArray const *Archive::source() const {
    return d->source;
}
void Archive::cache(CacheOperation operation) {
    if (!d->source) {
        return;
    }
    PathTreeIterator<PathTree> iter(d->index->leafNodes());
    while (iter.hasNext()) {
        Entry &entry = static_cast<Entry &>(iter.next());
        switch (operation) {
        case CacheAndDetachFromSource:
        case CacheAndRemainAttachedToSource:
            if (!entry.data && !entry.dataInArchive) {
                entry.dataInArchive.reset(new Block(*d->source, entry.offset, entry.sizeInArchive));
            }
            break;
        case UncacheUnmodifiedEntries:
            if (!entry.maybeChanged) {
                entry.data.reset();
                entry.dataInArchive.reset();
            }
            break;
        }
    }
    if (operation == CacheAndDetachFromSource) {
        d->source = nullptr;
    }
}
bool Archive::hasEntry(Path const &path) const {
    DENG2_ASSERT(d->index != 0);
    return d->index->has(path, PathTree::MatchFull | PathTree::NoBranch);
}
dint Archive::listFiles(Archive::Names &names, Path const &folder) const {
    DENG2_ASSERT(d->index != 0);
    names.clear();
    if (PathTree::Node const *parent = d->index->tryFind(folder, PathTree::MatchFull | PathTree::NoLeaf)) {
        for (PathTreeIterator<PathTree> iter(parent->children().leaves); iter.hasNext(); ) {
            names.insert(iter.next().name());
        }
    }
    return dint(names.size());
}
dint Archive::listFolders(Archive::Names &names, Path const &folder) const {
    DENG2_ASSERT(d->index != 0);
    names.clear();
    if (PathTree::Node const *parent = d->index->tryFind(folder, PathTree::MatchFull | PathTree::NoLeaf)) {
        for (PathTreeIterator<PathTree> iter(parent->children().branches); iter.hasNext(); ) {
            names.insert(iter.next().name());
        }
    }
    return dint(names.size());
}
File::Status Archive::entryStatus(Path const &path) const {
    DENG2_ASSERT(d->index != 0);
    Entry const &found = static_cast<Entry const &>(d->index->find(path, PathTree::MatchFull));
    return File::Status(
        found.isLeaf()? File::Type::File : File::Type::Folder,
        found.size,
        found.modifiedAt);
}
Block const &Archive::entryBlock(Path const &path) const {
    DENG2_ASSERT(d->index != 0);
    if (Entry *entry = static_cast<Entry *>(d->index->tryFind(path, PathTree::MatchFull | PathTree::NoBranch))) {
        if (entry->data) {
            return *entry->data;
        }
        std::unique_ptr<Block> cached(new Block);
        d->readEntry(path, *cached.get());
        entry->data.reset(cached.release());
        return *entry->data;
    }
    else {
        throw NotFoundError("Archive::entryBlock", String("'%1' not found").arg(path));
    }
}
Block &Archive::entryBlock(Path const &path) {
    if (!hasEntry(path)) {
        add(path, Block());
    }
    Block const &block = const_cast<Archive const *>(this)->entryBlock(path);
    Entry &entry = static_cast<Entry &>(d->index->find(path, PathTree::MatchFull | PathTree::NoBranch));
    entry.maybeChanged = true;
    entry.modifiedAt   = Time::currentHighPerformanceTime();
    d->modified = true;
    return const_cast<Block &>(block);
}
void Archive::uncacheBlock(Path const &path) const {
    if (!d->source) return; // Wouldn't be able to re-cache the data.
    if (Entry *entry = static_cast<Entry *>(d->index->tryFind(path, PathTree::MatchFull | PathTree::NoBranch))) {
        if (!entry->data && !entry->dataInArchive) return;
        /*qDebug() << "Archive:" << path << "uncached by archive" << this
                 << (entry->data? entry->data->size() : 0)
                 << (entry->dataInArchive? entry->dataInArchive->size() : 0);*/
        if (!entry->maybeChanged) {
            entry->data.reset();
        }
        entry->dataInArchive.reset();
    }
    else {
        throw NotFoundError("Archive::uncacheBlock", String("'%1' not found").arg(path));
    }
}
void Archive::add(Path const &path, IByteArray const &data) {
    if (path.isEmpty()) {
        throw InvalidPathError("Archive::add",
                               QString("'%1' is an invalid path for an entry").arg(path));
    }
    remove(path);
    DENG2_ASSERT(d->index != 0);
    Entry &entry = static_cast<Entry &>(d->index->insert(path));
    entry.data.reset(new Block(data));
    entry.modifiedAt   = Time::currentHighPerformanceTime();
    entry.maybeChanged = true;
    d->modified = true;
}
void Archive::remove(Path const &path) {
    DENG2_ASSERT(d->index != 0);
    if (d->index->remove(path, PathTree::MatchFull | PathTree::NoBranch)) {
        d->modified = true;
    }
}
void Archive::clear() {
    DENG2_ASSERT(d->index != 0);
    d->index->clear();
    d->modified = true;
}
bool Archive::modified() const {
    return d->modified;
}
void Archive::setIndex(PathTree *tree) {
    d->index = tree;
}
Archive::Entry &Archive::insertEntry(Path const &path) {
    LOG_AS("Archive");
    DENG2_ASSERT(d->index != 0);
    d->index->remove(path, PathTree::MatchFull | PathTree::NoBranch);
    return static_cast<Entry &>(d->index->insert(path));
}
PathTree const &Archive::index() const {
    DENG2_ASSERT(d->index != 0);
    return *d->index;
}
Archive::Entry::Entry(PathTree::NodeArgs const &args)
    : Node(args)
    , offset(0)
    , size(0)
    , sizeInArchive(0)
    , modifiedAt(Time::invalidTime())
    , maybeChanged(false)
{}
Archive::Entry::~Entry()
{}
} // namespace de
namespace de {
ArrayValue::ArrayValue() : Value(), _iteration(0)
{}
ArrayValue::ArrayValue(ArrayValue const &other) : Value(), _iteration(0) {
    for (auto i : other._elements) {
        _elements.push_back(i->duplicate());
    }
}
ArrayValue::ArrayValue(StringList const &strings) {
    for (String str : strings) {
        _elements.push_back(new TextValue(str));
    }
}
ArrayValue::ArrayValue(std::initializer_list<Value *> values) : _iteration(0) {
    for (auto *v : values) {
        _elements.push_back(v);
    }
}
ArrayValue::~ArrayValue() {
    clear();
}
Value *ArrayValue::duplicate() const {
    return new ArrayValue(*this);
}
Value::Text ArrayValue::asText() const {
    Text result;
    QTextStream s(&result);
    s << "[";
    bool isFirst = true;
    bool hadNewline = false;
    for (Elements::const_iterator i = _elements.begin();
        i != _elements.end(); ++i) {
        String content = (*i)->asText();
        bool multiline = content.contains('\n');
        if (!isFirst) {
            if (hadNewline || multiline) s << "\n";
            s << ",";
        }
        hadNewline = multiline;
        s << " " << content.replace("\n", "\n  ");
        isFirst = false;
    }
    s << " ]";
    return result;
}
dsize ArrayValue::size() const {
    return _elements.size();
}
Value const &ArrayValue::element(Value const &indexValue) const {
    NumberValue const *v = dynamic_cast<NumberValue const *>(&indexValue);
    if (!v) {
        throw IllegalIndexError("ArrayValue::element", "Array index must be a number");
    }
    dint index = v->asInt();
    Elements::const_iterator elem = indexToIterator(index);
    return **elem;
}
Value &ArrayValue::element(Value const &index) {
    return const_cast<Value &>(const_cast<ArrayValue const *>(this)->element(index));
}
void ArrayValue::setElement(Value const &indexValue, Value *value) {
    NumberValue const *v = dynamic_cast<NumberValue const *>(&indexValue);
    if (!v) {
        throw IllegalIndexError("ArrayValue::setElement", "Array index must be a number");
    }
    replace(v->asInt(), value);
}
bool ArrayValue::contains(Value const &value) const {
    for (Elements::const_iterator i = _elements.begin(); i != _elements.end(); ++i) {
        if (!(*i)->compare(value)) {
            return true;
        }
    }
    return false;
}
Value *ArrayValue::begin() {
    _iteration = 0;
    return next();
}
Value *ArrayValue::next() {
    if (_iteration < dint(_elements.size())) {
        return _elements[_iteration++]->duplicate();
    }
    return 0;
}
bool ArrayValue::isTrue() const {
    return _elements.size() > 0;
}
dint ArrayValue::compare(Value const &value) const {
    ArrayValue const *other = dynamic_cast<ArrayValue const *>(&value);
    if (other) {
        if (size() < other->size()) {
            return -1;
        }
        if (size() > other->size()) {
            return 1;
        }
        Elements::const_iterator mine = _elements.begin();
        Elements::const_iterator theirs = other->_elements.begin();
        for (; mine != _elements.end() && theirs != other->_elements.end(); ++mine, ++theirs) {
            dint result = (*mine)->compare(**theirs);
            if (result) return result;
        }
        return 0;
    }
    return Value::compare(value);
}
void ArrayValue::sum(Value const &value) {
    ArrayValue const *array = dynamic_cast<ArrayValue const *>(&value);
    if (!array) {
        throw ArithmeticError("ArrayValue::sum", "Array cannot be summed with value");
    }
    for (Elements::const_iterator i = array->_elements.begin(); i != array->_elements.end(); ++i) {
        _elements.push_back((*i)->duplicate());
    }
}
void ArrayValue::add(Value *value) {
    _elements.push_back(value);
}
void ArrayValue::addMany(duint count, Value::Number value) {
    while (count--) {
        add(new NumberValue(value));
    }
}
void ArrayValue::addMany(duint count, String const &value) {
    while (count--) {
        add(new TextValue(value));
    }
}
void ArrayValue::add(String const &text) {
    add(new TextValue(text));
}
Value const &ArrayValue::at(dint index) const {
    return **indexToIterator(index);
}
ArrayValue::Elements::iterator ArrayValue::indexToIterator(dint index) {
    if (index >= 0 && index < dint(size())) {
        return _elements.begin() + index;
    }
    else if (index < 0 && index > -dint(size())) {
        return _elements.begin() + size() + index;
    }
    else {
        throw OutOfBoundsError("ArrayValue::indexToIterator", "Index is out of bounds");
    }
}
ArrayValue::Elements::const_iterator ArrayValue::indexToIterator(dint index) const {
    if (index >= 0 && index < dint(size())) {
        return _elements.begin() + index;
    }
    else if (index < 0 && index >= -dint(size())) {
        return _elements.begin() + size() + index;
    }
    else {
        throw OutOfBoundsError("ArrayValue::indexToIterator", "Index is out of bounds");
    }
}
void ArrayValue::insert(dint index, Value *value) {
    if (index == dint(size())) {
        add(value);
    }
    else {
        _elements.insert(indexToIterator(index), value);
    }
}
void ArrayValue::replace(dint index, Value *value) {
    Elements::iterator elem = indexToIterator(index);
    delete *elem;
    *elem = value;
}
void ArrayValue::remove(dint index) {
    Elements::iterator elem = indexToIterator(index);
    delete *elem;
    _elements.erase(elem);
}
dint ArrayValue::indexOf(const Value &value) const {
    for (int i = 0; i < _elements.size(); ++i) {
        if (!value.compare(*_elements[i])) {
            return i;
        }
    }
    return -1;
}
ArrayValue &ArrayValue::operator << (Value *value) {
    add(value);
    return *this;
}
ArrayValue &ArrayValue::operator << (Value const &value) {
    add(value.duplicate());
    return *this;
}
Value *ArrayValue::popLast() {
    DENG2_ASSERT(size() > 0);
    return _elements.takeLast();
}
Value *ArrayValue::popFirst() {
    DENG2_ASSERT(size() > 0);
    return _elements.takeFirst();
}
void ArrayValue::reverse() {
    std::reverse(_elements.begin(), _elements.end());
}
StringList ArrayValue::toStringList() const {
    StringList list;
    for (Value const *v : _elements) {
        list << v->asText();
    }
    return list;
}
void ArrayValue::clear() {
    for (Elements::iterator i = _elements.begin(); i != _elements.end(); ++i) {
        delete *i;
    }
    _elements.clear();
}
void ArrayValue::operator >> (Writer &to) const {
    to << SerialId(ARRAY) << duint(_elements.size());
    for (Elements::const_iterator i = _elements.begin(); i != _elements.end(); ++i) {
        to << **i;
    }
}
void ArrayValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != ARRAY) {
        throw DeserializationError("ArrayValue::operator <<", "Invalid ID");
    }
    duint count = 0;
    from >> count;
    clear();
    while (count--) {
        add(Value::constructFrom(from));
    }
}
void ArrayValue::callElements(ArrayValue const &args) {
    for (duint i = 0; i < size(); ++i) {
        Function const &func = at(i).as<FunctionValue>().function();
        Process(func.globals()).call(func, args);
    }
}
void ArrayValue::setElement(dint index, Number value) {
    setElement(NumberValue(index), new NumberValue(value));
}
void ArrayValue::setElement(dint index, String const &value) {
    setElement(NumberValue(index), new TextValue(value));
}
Value const &ArrayValue::element(dint index) const {
    return element(NumberValue(index));
}
Value const &ArrayValue::operator [] (dint index) const {
    return element(index);
}
String ArrayValue::asInfo() const {
    StringList values;
    for (Value const *value : elements()) {
        String text = value->asText();
        text.replace("\"", "''");
        values << String("\"%1\"").arg(text);
    }
    QString out;
    QTextStream os(&out);
    os.setCodec("UTF-8");
    os << "<"
       << String::join(values, ", ")
       << ">";
    return out;
}
Value::Text ArrayValue::typeId() const {
    return "Array";
}
} // namespace de
namespace de {
namespace internal {
/**
 * Cache of objects of type ItemType. Does not own the objects or the data,
 * merely guides operations on them.
 */
template <typename ItemType>
class Cache : public Lockable {
public:
    enum Format {
        Source,     ///< Data is present only as source information.
        Object,     ///< Data is present as a fully usable object in memory.
        Serialized  ///< Data is present as a block of serialized bytes.
    };
    static char const *formatAsText(Format f) {
        switch (f) {
        case Source:     return "Source";     break;
        case Object:     return "Object";     break;
        case Serialized: return "Serialized"; break;
        }
        return "";
    }
    typedef QSet<ItemType *> Items;
public:
    Cache(Format format)
        : _format(format),
          _maxBytes(Bank::Unlimited),
          _currentBytes(0),
          _maxItems(Bank::Unlimited) {}
    virtual ~Cache() {}
    void setMaxBytes(dint64 m) { _maxBytes = m; }
    void setMaxItems(int m) { _maxItems = m; }
    Format format() const { return _format; }
    dint64 maxBytes() const { return _maxBytes; }
    dint64 byteCount() const { return _currentBytes; }
    int maxItems() const { return _maxItems; }
    int itemCount() const { return _items.size(); }
    virtual void add(ItemType &data) {
        _items.insert(&data);
    }
    virtual void remove(ItemType &data) {
        _items.remove(&data);
    }
    bool contains(ItemType const *data) const {
        DENG2_GUARD(this);
        return _items.contains(const_cast<ItemType *>(data));
    }
    virtual void clear() {
        DENG2_GUARD(this);
        _items.clear();
        _currentBytes = 0;
    }
    Items const &items() const { return _items; }
protected:
    void addBytes(dint64 bytes) {
        _currentBytes = de::max(dint64(0), _currentBytes + bytes);
    }
private:
    Format _format;
    dint64 _maxBytes;
    dint64 _currentBytes;
    int _maxItems;
    Items _items;
};
} // namespace internal
DENG2_PIMPL(Bank) {
    /**
     * Data item. Has ownership of the in-memory cached data and the source
     * information. Lockable because may be accessed from the worker thread in
     * addition to caller thread(s).
     */
    struct Data : public PathTree::Node, public Waitable, public Lockable {
        typedef internal::Cache<Data> Cache;
        Bank *bank;                     ///< Bank that owns the data.
        std::unique_ptr<IData> data;    ///< Non-NULL for in-memory items.
        std::unique_ptr<ISource> source;///< Always required.
        SafePtr<File> serial;           ///< Serialized representation (if one is present; not owned).
        Cache *cache;                   ///< Current cache for the data (never NULL).
        Time accessedAt;
        Data(PathTree::NodeArgs const &args)
            : Node(args)
            , bank(0)
            , cache(0)
            , accessedAt(Time::invalidTime())
        {}
        void clearData() {
            DENG2_GUARD(this);
            if (data.get()) {
                LOG_XVERBOSE("Item \"%s\" data cleared from memory (%i bytes)",
                        path(bank->d->sepChar) << data->sizeInMemory());
                data->aboutToUnload();
                data.reset();
            }
        }
        void setData(IData *newData) {
            DENG2_GUARD(this);
            DENG2_ASSERT(newData != 0);
            data.reset(newData);
            accessedAt = Time::currentHighPerformanceTime();
            bank->d->notify(Notification(Notification::Loaded, path(bank->d->sepChar)));
        }
        void load() {
            DENG2_ASSERT(cache != 0);
            switch (cache->format()) {
            case Cache::Source:
                loadFromSource();
                break;
            case Cache::Serialized:
                loadFromSerialized();
                break;
            case Cache::Object:
                break;
            }
        }
        void loadFromSource() {
            DENG2_ASSERT(source.get() != 0);
            Time startedAt;
            QScopedPointer<IData> loaded(bank->loadFromSource(*source));
            LOG_XVERBOSE("Loaded \"%s\" from source in %.2f seconds",
                         path(bank->d->sepChar) << startedAt.since());
            if (loaded.data()) {
                setData(loaded.take());
            }
        }
        bool isValidSerialTime(Time const &serialTime) const {
            return (!source->modifiedAt().isValid() ||
                    source->modifiedAt() == serialTime);
        }
        void loadFromSerialized() {
            DENG2_ASSERT(serial);
            try {
                Time startedAt;
                Time timestamp(Time::invalidTime());
                Reader reader(*serial);
                reader.withHeader() >> timestamp;
                if (isValidSerialTime(timestamp)) {
                    std::unique_ptr<IData> blank(bank->newData());
                    reader >> *blank->asSerializable();
                    setData(blank.release());
                    LOG_XVERBOSE("Deserialized \"%s\" in %.2f seconds",
                                 path(bank->d->sepChar) << startedAt.since());
                    return; // Done!
                }
            }
            catch (Error const &er) {
                LOG_WARNING("Failed to deserialize \"%s\":\n")
                        << path(bank->d->sepChar) << er.asText();
            }
            loadFromSource();
        }
        void serialize(Path const &folderPath) {
            DENG2_GUARD(this);
            if (serial) {
                return;
            }
            DENG2_ASSERT(source.get() != 0);
            if (!data.get()) {
                loadFromSource();
            }
            DENG2_ASSERT(data->asSerializable() != 0);
            try {
                Folder &containingFolder = FileSystem::get()
                        .makeFolder(folderPath / path().toString().fileNamePath());
                if (!data->shouldBeSerialized()) {
                    if (File *existing = containingFolder.tryLocate<File>(name())) {
                        serial.reset(existing);
                        return;
                    }
                }
                serial.reset(&containingFolder.createFile(name(), Folder::ReplaceExisting));
                LOG_XVERBOSE("Serializing into %s", serial->description());
                Writer(*serial).withHeader()
                        << source->modifiedAt()
                        << *data->asSerializable();
                serial->flush();
            }
            catch (...) {
                serial.reset();
                throw;
            }
        }
        void clearSerialized() {
            DENG2_GUARD(this);
            serial.reset();
        }
        void changeCache(Cache &toCache) {
            DENG2_GUARD(this);
            DENG2_ASSERT(cache != 0);
            if (cache != &toCache) {
                Cache &fromCache = *cache;
                toCache.add(*this);
                fromCache.remove(*this);
                cache = &toCache;
                Path const itemPath = path(bank->d->sepChar);
                LOGDEV_RES_XVERBOSE("Item \"%s\" moved to %s cache",
                                    itemPath << Cache::formatAsText(toCache.format()));
                bank->d->notify(Notification(itemPath, toCache));
            }
        }
    };
    typedef PathTreeT<Data> DataTree;
    typedef internal::Cache<Data> DataCache;
    /**
     * Dummy cache representing objects in the source data.
     */
    struct SourceCache : public DataCache {
        SourceCache() : DataCache(Source) {}
    };
    /**
     * Hot storage containing data items serialized into files. The goal is to
     * allow quick recovery of data into memory. May be disabled in a Bank.
     */
    class SerializedCache : public DataCache {
    public:
        SerializedCache() : DataCache(Serialized) {}
        void add(Data &item) {
            DENG2_GUARD(this);
            DENG2_ASSERT(!_path.isEmpty());
            item.serialize(_path);
            addBytes(item.serial->size());
            DataCache::add(item);
        }
        void remove(Data &item) {
            DENG2_GUARD(this);
            addBytes(-dint64(item.serial->size()));
            item.clearSerialized();
            DataCache::remove(item);
        }
        void setLocation(String const &location) {
            DENG2_ASSERT(!location.isEmpty());
            DENG2_GUARD(this);
            _path = location;
        }
        Path const &path() const {
            return _path;
        }
        Folder *folder() const {
            return FS::tryLocate<Folder>(_path);
        }
    private:
        Path _path;
    };
    /**
     * Cache of objects in memory.
     */
    class ObjectCache : public DataCache {
    public:
        ObjectCache() : DataCache(Object)
        {}
        void add(Data &item) {
            item.load();
            DENG2_GUARD(this);
            DENG2_ASSERT(item.data.get() != nullptr);
            addBytes(item.data->sizeInMemory());
            DataCache::add(item);
        }
        void remove(Data &item) {
            DENG2_ASSERT(item.data.get() != nullptr);
            DENG2_GUARD(this);
            addBytes(-dint64(item.data->sizeInMemory()));
            item.clearData();
            DataCache::remove(item);
        }
    };
    /**
     * Operation on a data item (e.g., loading, serialization). Run by TaskPool
     * in a background thread.
     */
    class Job : public Task {
    public:
        enum Task {
            Load,
            Serialize,
            Unload
        };
    public:
        Job(Bank &bk, Task t, Path const &p = Path())
            : _bank(bk), _task(t), _path(p)
        {}
        void runTask() {
            LOG_AS("Bank::Job");
            switch (_task) {
            case Job::Load:
                doLoad();
                break;
            case Job::Serialize:
                doSerialize();
                break;
            case Job::Unload:
                doUnload();
                break;
            }
        }
        Data &item() {
            return _bank.d->items.find(_path, PathTree::MatchFull | PathTree::NoBranch);
        }
        void doLoad() {
            try {
                Data &it = item();
                it.changeCache(_bank.d->memoryCache);
            }
            catch (Error const &er) {
                LOG_WARNING("Failed to load \"%s\" from source:\n") << _path << er.asText();
            }
            item().post();
        }
        void doSerialize() {
            try {
                DENG2_ASSERT(_bank.d->serialCache != 0);
                LOG_XVERBOSE("Serializing \"%s\"", _path);
                item().changeCache(*_bank.d->serialCache);
            }
            catch (Error const &er) {
                LOG_WARNING("Failed to serialize \"%s\" to hot storage:\n")
                        << _path << er.asText();
            }
        }
        void doUnload() {
            try {
                LOGDEV_RES_XVERBOSE("Unloading \"%s\"", _path);
                item().changeCache(_bank.d->sourceCache);
            }
            catch (Error const &er) {
                LOG_WARNING("Error when unloading \"%s\":\n")
                        << _path << er.asText();
            }
        }
    private:
        Bank &_bank;
        Task _task;
        Path _path;
    };
    /**
     * Notification about status changes of data in the tree.
     */
    struct Notification {
        enum Kind { CacheChanged, Loaded };
        Kind kind;
        Path path;
        DataCache *cache;
        Notification(Kind k, Path const &p)
            : kind(k), path(p), cache(0) {}
        Notification(Path const &p, DataCache &c)
            : kind(CacheChanged), path(p), cache(&c) {}
    };
    typedef FIFO<Notification> NotifyQueue;
    char const *nameForLog;
    QChar sepChar { '.' }; ///< Default separator in identifier paths.
    Flags flags;
    SourceCache sourceCache;
    ObjectCache memoryCache;
    std::unique_ptr<SerializedCache> serialCache;
    DataTree items;
    TaskPool jobs;
    NotifyQueue notifications;
    LoopCallback mainCall;
    Impl(Public *i, char const *name, Flags const &flg)
        : Base(i)
        , nameForLog(name)
        , flags(flg) {
        if (!flags.testFlag(DisableHotStorage)) {
            serialCache.reset(new SerializedCache);
        }
    }
    ~Impl() {
        destroySerialCache();
    }
    /**
     * Deletes everything in the hot storage folder.
     */
    void clearHotStorage() {
        DENG2_ASSERT(serialCache);
        FS::waitForIdle();
        if (Folder *folder = serialCache->folder()) {
            folder->destroyAllFilesRecursively();
        }
    }
    void destroySerialCache() {
        jobs.waitForDone();
        if (serialCache && flags.testFlag(ClearHotStorageWhenBankDestroyed)) {
            clearHotStorage();
        }
        serialCache.reset();
    }
    inline bool isThreaded() const {
        return flags.testFlag(BackgroundThread);
    }
    void beginJob(Job *job, Importance importance) {
        if (!isThreaded() || importance == ImmediatelyInCurrentThread) {
            QScopedPointer<Job> j(job);
            j->runTask();
            performNotifications();
        }
        else {
            jobs.start(job, importance == AfterQueued?
                           TaskPool::LowPriority : TaskPool::HighPriority);
        }
    }
    void clear() {
        jobs.waitForDone();
        items.clear();
        sourceCache.clear();
        memoryCache.clear();
        if (serialCache) {
            serialCache->clear();
        }
    }
    void setSerialLocation(String const &location) {
        if (location.isEmpty() || flags.testFlag(DisableHotStorage)) {
            destroySerialCache();
        }
        else {
            if (!serialCache) serialCache.reset(new SerializedCache);
            serialCache->setLocation(location);
            FS::get().makeFolder(location);
        }
    }
    void putInBestCache(Data &item) {
        DENG2_ASSERT(item.cache == 0);
        DataCache *best = &sourceCache;
        if (serialCache) {
            File *src = FS::tryLocate<File>(serialCache->path() / item.path());
            if (src) {
                Time hotTime;
                Reader(*src).withHeader() >> hotTime;
                if (item.isValidSerialTime(hotTime)) {
                    LOGDEV_RES_VERBOSE("Found valid serialized copy of \"%s\"") << item.path(sepChar);
                    item.serial.reset(src);
                    best = serialCache.get();
                }
            }
        }
        item.cache = best;
        best->add(item);
    }
    void load(Path const &path, Importance importance) {
        beginJob(new Job(self(), Job::Load, path), importance);
    }
    void unload(Path const &path, CacheLevel toLevel, Importance importance) {
        if (toLevel < InMemory) {
            Job::Task const task = (toLevel == InHotStorage && serialCache?
                                    Job::Serialize : Job::Unload);
            beginJob(new Job(self(), task, path), importance);
        }
    }
    void notify(Notification const &notif) {
        notifications.put(new Notification(notif));
        if (isThreaded()) {
            mainCall.enqueue([this] () { performNotifications(); });
        }
    }
    void performNotifications() {
        forever {
            QScopedPointer<Notification> notif (notifications.take());
            if (!notif.data()) break;
            performNotification(*notif);
        }
    }
    void performNotification(Notification const &nt) {
        switch (nt.kind) {
        case Notification::Loaded:
            DENG2_FOR_PUBLIC_AUDIENCE2(Load, i) {
                i->bankLoaded(nt.path);
            }
            break;
        case Notification::CacheChanged:
            DENG2_FOR_PUBLIC_AUDIENCE2(CacheLevel, i) {
                DENG2_ASSERT(nt.cache != 0);
                i->bankCacheLevelChanged(nt.path,
                      nt.cache == &memoryCache?      InMemory :
                      nt.cache == serialCache.get()? InHotStorage :
                                                     InColdStorage);
            }
            break;
        }
    }
    DENG2_PIMPL_AUDIENCE(Load)
    DENG2_PIMPL_AUDIENCE(CacheLevel)
};
DENG2_AUDIENCE_METHOD(Bank, Load)
DENG2_AUDIENCE_METHOD(Bank, CacheLevel)
Bank::Bank(char const *nameForLog, Flags const &flags, String const &hotStorageLocation)
    : d(new Impl(this, nameForLog, flags)) {
    d->setSerialLocation(hotStorageLocation);
}
Bank::~Bank() {
    clear();
}
char const *Bank::nameForLog() const {
    return d->nameForLog;
}
Bank::Flags Bank::flags() const {
    return d->flags;
}
void Bank::setSeparator(QChar sep) {
    d->sepChar = sep;
}
void Bank::setHotStorageCacheLocation(String const &location) {
    d->setSerialLocation(location);
}
void Bank::setHotStorageSize(dint64 maxBytes) {
    if (d->serialCache) {
        d->serialCache->setMaxBytes(maxBytes);
    }
}
void Bank::setMemoryCacheSize(dint64 maxBytes) {
    d->memoryCache.setMaxBytes(maxBytes);
}
Path Bank::hotStorageCacheLocation() const {
    if (d->serialCache) {
        return d->serialCache->path();
    }
    return Path();
}
dint64 Bank::hotStorageSize() const {
    if (d->serialCache) {
        return d->serialCache->maxBytes();
    }
    return 0;
}
dint64 Bank::memoryCacheSize() const {
    return d->memoryCache.maxBytes();
}
void Bank::clearHotStorage() {
    if (d->serialCache) {
        d->clearHotStorage();
    }
}
void Bank::clear() {
    d->clear();
}
void Bank::add(DotPath const &path, ISource *source) {
    LOG_AS(d->nameForLog);
    std::unique_ptr<ISource> src(source);
    if (d->items.has(path, PathTree::MatchFull | PathTree::NoBranch)) {
        throw AlreadyExistsError(QLatin1String(d->nameForLog) + "::add",
                                 "Item '" + path.toString() + "' already exists");
    }
    Impl::Data &item = d->items.insert(path);
    DENG2_GUARD(item);
    item.bank = this;
    item.source.reset(src.release());
    d->putInBestCache(item);
}
void Bank::remove(DotPath const &path) {
    d->items.remove(path, PathTree::NoBranch);
}
bool Bank::has(DotPath const &path) const {
    return d->items.has(path);
}
Bank::ISource &Bank::source(const DotPath &path) const {
    return *d->items.find(path, PathTree::MatchFull | PathTree::NoBranch).source.get();
}
dint Bank::allItems(Names &names) const {
    names.clear();
    iterate([&names] (DotPath const &path) {
        names.insert(path.toString());
    });
    return dint(names.size());
}
void Bank::iterate(std::function<void (DotPath const &)> func) const {
    PathTree::FoundPaths paths;
    d->items.findAllPaths(paths, PathTree::NoBranch, d->sepChar);
    foreach (String const &path, paths) {
        func(path);
    }
}
PathTree const &Bank::index() const {
    return d->items;
}
void Bank::load(DotPath const &path, Importance importance) {
    d->load(path, importance);
}
void Bank::loadAll() {
    Names names;
    allItems(names);
    DENG2_FOR_EACH(Names, i, names) {
        load(*i, AfterQueued);
    }
}
Bank::IData &Bank::data(DotPath const &path) const {
    LOG_AS(d->nameForLog);
    Impl::Data &item = d->items.find(path, PathTree::MatchFull | PathTree::NoBranch);
    DENG2_GUARD(item);
    item.accessedAt = Time::currentHighPerformanceTime();
    if (item.data.get()) {
        return *item.data;
    }
    item.reset();
    item.unlock();
    LOG_XVERBOSE("Loading \"%s\"...", path);
    Time requestedAt;
    d->load(path, BeforeQueued);
    item.wait();
    TimeSpan const waitTime = requestedAt.since();
    item.lock();
    if (!item.data.get()) {
        throw LoadError(QLatin1String(d->nameForLog) + "::data", "Failed to load \"" + path + "\"");
    }
    if (waitTime > 0.0) {
        LOG_VERBOSE("\"%s\" loaded (waited %.3f seconds)") << path << waitTime;
    }
    else {
        LOG_VERBOSE("\"%s\" loaded") << path;
    }
    return *item.data;
}
bool Bank::isLoaded(DotPath const &path) const {
    if (Impl::Data const *item = d->items.tryFind(path, PathTree::MatchFull | PathTree::NoBranch)) {
        return d->memoryCache.contains(item);
    }
    return false;
}
void Bank::unload(DotPath const &path, CacheLevel toLevel, Importance importance) {
    d->unload(path, toLevel, importance);
}
void Bank::unloadAll(CacheLevel maxLevel) {
    return unloadAll(AfterQueued, maxLevel);
}
void Bank::unloadAll(Importance importance, CacheLevel maxLevel) {
    if (maxLevel >= InMemory) return;
    Names names;
    allItems(names);
    DENG2_FOR_EACH(Names, i, names) {
        unload(Path(*i, d->sepChar), maxLevel, importance);
    }
}
void Bank::clearFromCache(DotPath const &path) {
    d->unload(path, InColdStorage, AfterQueued);
}
void Bank::purge() {
    /**
     * @todo Implement cache purging (and different purging strategies?).
     * Purge criteria can be age and cache level maximum limits.
     */
}
Bank::IData *Bank::newData() {
    return 0;
}
} // namespace de
namespace de {
DENG2_PIMPL(BitField) {
    Elements const *elements;
    Block packed;
    Impl(Public *i) : Base(i), elements(0)
    {}
    Impl(Public *i, Impl const &other)
        : Base     (i)
        , elements (other.elements)
        , packed   (other.packed)
    {}
    static dbyte elementMask(int numBits, int skipBits) {
        dbyte mask = 0xff;
        if (numBits - skipBits < 8) {
            mask >>= 8 - (numBits - skipBits);
        }
        return mask;
    }
    void set(Id id, duint value) {
        DENG2_ASSERT(elements != 0);
        int eFirstBit = 0;
        int eNumBits = 0;
        elements->elementLayout(id, eFirstBit, eNumBits);
        int packedIdx = eFirstBit >> 3;
        int shift     = eFirstBit & 7;
        int written   = 0;
        while (written < eNumBits) {
            dbyte const mask = elementMask(eNumBits, written) << shift;
            dbyte pv = packed[packedIdx] & ~mask;
            pv |= mask & ((value >> written) << shift);
            packed[packedIdx] = pv;
            written += 8 - shift;
            shift = 0;
            packedIdx++;
        }
    }
    duint get(Id id) const {
        DENG2_ASSERT(elements != 0);
        duint value = 0;
        int eFirstBit = 0;
        int eNumBits = 0;
        elements->elementLayout(id, eFirstBit, eNumBits);
        int packedIdx = eFirstBit >> 3;
        int shift     = eFirstBit & 7;
        int read      = 0;
        while (read < eNumBits) {
            dbyte const mask = elementMask(eNumBits, read) << shift;
            value |= ((packed[packedIdx] & mask) >> shift) << read;
            read += 8 - shift;
            shift = 0;
            packedIdx++;
        }
        return value;
    }
    Ids delta(Impl const &other) const {
        DENG2_ASSERT(elements != 0);
        DENG2_ASSERT(other.elements != 0);
        if (elements->size() != other.elements->size()) {
            throw ComparisonError("BitField::delta",
                                  "The compared fields have a different number of elements");
        }
        if (packed.size() != other.packed.size()) {
            throw ComparisonError("BitField::delta",
                                  "The compared fields have incompatible element sizes");
        }
        Ids diffs;
        for (duint pos = 0; pos < packed.size(); ++pos) {
            if (packed[pos] == other.packed[pos])
                continue;
            Ids const lookup = elements->idsLaidOutOnByte(pos);
            DENG2_FOR_EACH_CONST(Ids, i, lookup) {
                Id const &id = *i;
                if (diffs.contains(id))
                    continue; // Already in the delta.
                if (get(id) != other.get(id)) {
                    diffs.insert(id);
                }
            }
        }
        return diffs;
    }
};
BitField::BitField() : d(new Impl(this))
{}
BitField::BitField(Elements const &elements) : d(new Impl(this)) {
    setElements(elements);
}
BitField::BitField(BitField const &other) : d(new Impl(this, *other.d))
{}
BitField::BitField(Block const &data) : d(new Impl(this)) {
    d->packed = data;
}
BitField &BitField::operator = (BitField const &other) {
    d->elements = other.d->elements;
    d->packed   = other.d->packed;
    return *this;
}
void BitField::setElements(Elements const &elements) {
    clear();
    d->elements = &elements;
    for (int i = 0; i < elements.size(); ++i) {
        set(elements.at(i).id, 0u);
    }
}
void BitField::clear() {
    d->packed.clear();
    d->elements = 0;
}
bool BitField::isEmpty() const {
    return !d->elements || !d->elements->size();
}
BitField::Elements const &BitField::elements() const {
    return *d->elements;
}
Block BitField::data() const {
    return d->packed;
}
bool BitField::operator == (BitField const &other) const {
    return d->packed == other.d->packed;
}
bool BitField::operator != (BitField const &other) const {
    return d->packed != other.d->packed;
}
BitField::Ids BitField::delta(BitField const &other) const {
    return d->delta(*other.d);
}
void BitField::set(Id id, bool value) {
    set(id, duint(value? 1 : 0));
}
void BitField::set(Id id, duint value) {
    d->set(id, value);
}
bool BitField::asBool(Id id) const {
    return d->get(id)? true : false;
}
duint BitField::asUInt(Id id) const {
    return d->get(id);
}
String BitField::asText() const {
    QString str;
    QTextStream os(&str);
    os << "BitField (" << d->elements->bitCount() << " bits, " << d->elements->size() << " elements):";
    os.setIntegerBase(2);
    for (int i = int(d->packed.size()) - 1; i >= 0; --i) {
        os << " " << qSetPadChar('0') << qSetFieldWidth(8) << dbyte(d->packed[i])
           << qSetFieldWidth(0);
    }
    return str;
}
} // namespace de
namespace de {
DENG2_PIMPL(BitField::Elements) {
    struct Element {
        int numBits;
        int firstBit;
    };
    typedef QMap<Id, Element> Elements; // needs to be ordered
    Elements elements;
    dsize totalBits;
    /**
     * Lookup table for quickly finding out which elements are on which bytes of
     * the packed data. Indexed using the packed data byte index; size ==
     * packed size.
     */
    QList<Ids> lookup;
    Impl(Public *i)
        : Base(i)
        , totalBits(0)
    {}
    Impl(Public *i, Impl const &other)
        : Base     (i)
        , elements (other.elements)
        , totalBits(other.totalBits)
        , lookup   (other.lookup)
    {}
    Element const &element(Id id) const {
        DENG2_ASSERT(elements.contains(id));
        return elements.constFind(id).value();
    }
};
BitField::Elements::Elements() : d(new Impl(this))
{}
BitField::Elements::Elements(Spec const *elements, dsize count)
    : d(new Impl(this)) {
    add(elements, count);
}
void BitField::Elements::clear() {
    d->totalBits = 0;
    d->elements.clear();
    d->lookup.clear();
}
BitField::Elements &BitField::Elements::add(Id id, dsize numBits) {
    DENG2_ASSERT(numBits >= 1);
    Impl::Element elem;
    elem.numBits  = numBits;
    elem.firstBit = d->totalBits;
    d->elements.insert(id, elem);
    d->totalBits += numBits;
    int pos = elem.firstBit / 8;
    int endPos = (elem.firstBit + (numBits - 1)) / 8;
    while (d->lookup.size() <= endPos) {
        d->lookup.append(Ids());
    }
    for (int i = pos; i <= endPos; ++i) {
        d->lookup[i].insert(id);
    }
    return *this;
}
void BitField::Elements::add(Spec const *elements, dsize count) {
    while (count-- > 0) {
        add(elements->id, elements->numBits);
        elements++;
    }
}
void BitField::Elements::add(QList<Spec> const &elements) {
    foreach (Spec spec, elements) {
        add(spec.id, spec.numBits);
    }
}
int BitField::Elements::size() const {
    return d->elements.size();
}
BitField::Spec BitField::Elements::at(int index) const {
    DENG2_ASSERT(index >= 0);
    DENG2_ASSERT(index < size());
    Impl::Element elem = d->elements.values()[index];
    Spec spec;
    spec.id = d->elements.keys()[index];
    spec.numBits = elem.numBits;
    return spec;
}
void BitField::Elements::elementLayout(Id const &id, int &firstBit, int &numBits) const {
    Impl::Element const &elem = d->element(id);
    firstBit = elem.firstBit;
    numBits  = elem.numBits;
}
int BitField::Elements::bitCount() const {
    return d->totalBits;
}
BitField::Ids BitField::Elements::ids() const {
    Ids ids;
    foreach (Id id, d->elements.keys()) {
        ids.insert(id);
    }
    return ids;
}
BitField::Ids BitField::Elements::idsLaidOutOnByte(int index) const {
    return d.getConst()->lookup.at(index);
}
} // namespace de
namespace de {
Block::Block(Size initialSize) {
    resize(initialSize);
}
Block::Block(IByteArray const &other) {
    resize(other.size());
    other.get(0, (dbyte *) data(), other.size());
}
Block::Block(Block const &other)
    : QByteArray(other), IByteArray(), IBlock()
{}
Block::Block(Block &&moved)
    : QByteArray(moved)
{}
Block::Block(QByteArray const &byteArray)
    : QByteArray(byteArray)
{}
Block::Block(char const *nullTerminatedCStr)
    : QByteArray(nullTerminatedCStr)
{}
Block::Block(void const *data, Size length)
    : QByteArray(reinterpret_cast<char const *>(data), int(length)), IByteArray(), IBlock()
{}
Block::Block(IIStream &stream) {
    stream >> *this;
}
Block::Block(IIStream const &stream) {
    stream >> *this;
}
Block::Block(IByteArray const &other, Offset at, Size count) : IByteArray() {
    copyFrom(other, at, count);
}
Block::Size Block::size() const {
    return QByteArray::size();
}
void Block::get(Offset atPos, Byte *values, Size count) const {
    if (atPos + count > size()) {
        throw OffsetError("Block::get", "Out of range " +
                          String("(%1[+%2] > %3)").arg(atPos).arg(count).arg(size()));
    }
    std::memcpy(values, constData() + atPos, count);
}
void Block::set(Offset at, Byte const *values, Size count) {
    if (at > size()) {
        throw OffsetError("Block::set", "Out of range");
    }
    replace(at, count, QByteArray(reinterpret_cast<char const *>(values), count));
}
void Block::copyFrom(IByteArray const &array, Offset at, Size count) {
    resize(count);
    array.get(at, data(), count);
}
void Block::resize(Size size) {
    QByteArray::resize(size);
}
IByteArray::Byte const *Block::dataConst() const {
    return reinterpret_cast<Byte const *>(QByteArray::constData());
}
Block &Block::append(IByteArray::Byte b) {
    QByteArray::append(char(b));
    return *this;
}
Block &Block::append(char const *str, int len) {
    QByteArray::append(str, len);
    return *this;
}
Block &Block::operator += (char const *nullTerminatedCStr) {
    QByteArray::append(nullTerminatedCStr);
    return *this;
}
Block &Block::operator += (QByteArray const &bytes) {
    QByteArray::append(bytes);
    return *this;
}
Block::Byte *Block::data() {
    return reinterpret_cast<Byte *>(QByteArray::data());
}
Block::Byte const *Block::data() const {
    return reinterpret_cast<Byte const *>(QByteArray::data());
}
void Block::operator >> (Writer &to) const {
    to << duint32(size());
    to.writeBytes(size(), *this);
}
void Block::operator << (Reader &from) {
    duint32 size = 0;
    from >> size;
    resize(size);
    from.readBytes(size, *this);
}
Block &Block::operator += (Block const &other) {
    QByteArray::append(other);
    return *this;
}
Block &Block::operator += (IByteArray const &byteArray) {
    Offset pos = size();
    resize(size() + byteArray.size());
    byteArray.get(0, data() + pos, byteArray.size());
    return *this;
}
Block &Block::operator = (Block const &other) {
    *static_cast<QByteArray *>(this) = static_cast<QByteArray const &>(other);
    return *this;
}
Block &Block::operator = (IByteArray const &byteArray) {
    copyFrom(byteArray, 0, byteArray.size());
    return *this;
}
Block Block::compressed(int level) const {
    return qCompress(*this, level);
}
Block Block::decompressed() const {
    return qUncompress(*this);
}
Block Block::md5Hash() const {
    QCryptographicHash hash(QCryptographicHash::Md5);
    hash.addData(*this);
    return hash.result();
}
static char asciiHexNumber(char n) {
    if (n < 0 || n > 15) return ' ';
    return n <= 9? ('0' + n) : ('a' + n - 10);
}
String Block::asHexadecimalText() const {
    int const count = QByteArray::size();
    String hex(count * 2, QChar('0'));
    for (int i = 0; i < count; i++) {
        auto const ch = duint8(at(i));
        hex[i*2]     = QChar(asciiHexNumber(ch >> 4));
        hex[i*2 + 1] = QChar(asciiHexNumber(ch & 0xf));
    }
    return hex;
}
Block Block::join(QList<Block> const &blocks, Block const &sep) // static
{
    if (blocks.isEmpty()) return Block();
    Block joined = blocks.at(0);
    for (int i = 1; i < blocks.size(); ++i) {
        joined += sep;
        joined += blocks.at(i);
    }
    return joined;
}
void Block::clear() {
    QByteArray::clear();
}
} // namespace de
namespace de {
BlockValue::BlockValue()
{}
BlockValue::BlockValue(Block const &block) : _value(block)
{}
Block &BlockValue::block() {
    return _value;
}
Block const &BlockValue::block() const {
    return _value;
}
BlockValue::operator IByteArray const &() const {
    return _value;
}
BlockValue::operator IByteArray &() {
    return _value;
}
void BlockValue::clear() {
    _value.clear();
}
Value *BlockValue::duplicate() const {
    return new BlockValue(_value);
}
Value::Text BlockValue::asText() const {
    Text result;
    QTextStream os(&result);
    os << "(Block of " << _value.size() << " bytes)";
    return result;
}
dsize BlockValue::size() const {
    return _value.size();
}
bool BlockValue::isTrue() const {
    return _value.size() > 0;
}
void BlockValue::sum(Value const &value) {
    BlockValue const *other = dynamic_cast<BlockValue const *>(&value);
    if (!other) {
        throw ArithmeticError("BlockValue::sum", "Value cannot be summed");
    }
    _value += other->_value;
}
void BlockValue::operator >> (Writer &to) const {
    to << SerialId(BLOCK) << _value;
}
void BlockValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != BLOCK) {
        throw DeserializationError("BlockValue::operator <<", "Invalid ID");
    }
    _value.clear();
    from >> _value;
}
Value::Text BlockValue::typeId() const {
    return "Block";
}
} // namespace de
namespace de {
DENG2_PUBLIC BigEndianByteOrder    bigEndianByteOrder;
DENG2_PUBLIC LittleEndianByteOrder littleEndianByteOrder;
void ByteOrder::hostToNetwork(dint16 hostValue, dint16 &networkValue) const {
    hostToNetwork(hostValue, reinterpret_cast<duint16 &>(networkValue));
}
void ByteOrder::hostToNetwork(dint32 hostValue, dint32 &networkValue) const {
    hostToNetwork(hostValue, reinterpret_cast<duint32 &>(networkValue));
}
void ByteOrder::hostToNetwork(dint64 hostValue, dint64 &networkValue) const {
    hostToNetwork(hostValue, reinterpret_cast<duint64 &>(networkValue));
}
void ByteOrder::networkToHost(dint16 networkValue, dint16 &hostValue) const {
    networkToHost(networkValue, reinterpret_cast<duint16 &>(hostValue));
}
void ByteOrder::networkToHost(dint32 networkValue, dint32 &hostValue) const {
    networkToHost(networkValue, reinterpret_cast<duint32 &>(hostValue));
}
void ByteOrder::networkToHost(dint64 networkValue, dint64 &hostValue) const {
    networkToHost(networkValue, reinterpret_cast<duint64 &>(hostValue));
}
void ByteOrder::hostToNetwork(dfloat hostValue, dfloat &networkValue) const {
    DENG2_ASSERT(sizeof(dfloat) == sizeof(duint32));
    duint32 hv, net;
    std::memcpy(&hv, &hostValue, 4);
    hostToNetwork(hv, net);
    std::memcpy(&networkValue, &net, 4);
}
void ByteOrder::hostToNetwork(ddouble hostValue, ddouble &networkValue) const {
    DENG2_ASSERT(sizeof(ddouble) == sizeof(duint64));
    duint64 hv, net;
    std::memcpy(&hv, &hostValue, 8);
    hostToNetwork(hv, net);
    std::memcpy(&networkValue, &net, 8);
}
void ByteOrder::networkToHost(dfloat networkValue, dfloat &hostValue) const {
    DENG2_ASSERT(sizeof(dfloat) == sizeof(duint32));
    duint32 net, hv;
    std::memcpy(&net, &networkValue, 4);
    networkToHost(net, hv);
    std::memcpy(&hostValue, &hv, 4);
}
void ByteOrder::networkToHost(ddouble networkValue, ddouble &hostValue) const {
    DENG2_ASSERT(sizeof(ddouble) == sizeof(duint64));
    duint64 net, hv;
    std::memcpy(&net, &networkValue, 8);
    networkToHost(net, hv);
    std::memcpy(&hostValue, &hv, 8);
}
void BigEndianByteOrder::networkToHost(duint16 networkValue, duint16 &hostValue) const {
#ifdef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap16(networkValue);
#endif
}
void BigEndianByteOrder::networkToHost(duint32 networkValue, duint32 &hostValue) const {
#ifdef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap32(networkValue);
#endif
}
void BigEndianByteOrder::networkToHost(duint64 networkValue, duint64 &hostValue) const {
#ifdef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap64(networkValue);
#endif
}
void BigEndianByteOrder::hostToNetwork(duint16 hostValue, duint16 &networkValue) const {
#ifdef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap16(hostValue);
#endif
}
void BigEndianByteOrder::hostToNetwork(duint32 hostValue, duint32 &networkValue) const {
#ifdef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap32(hostValue);
#endif
}
void BigEndianByteOrder::hostToNetwork(duint64 hostValue, duint64 &networkValue) const {
#ifdef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap64(hostValue);
#endif
}
void LittleEndianByteOrder::networkToHost(duint16 networkValue, duint16 &hostValue) const {
#ifndef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap16(networkValue);
#endif
}
void LittleEndianByteOrder::networkToHost(duint32 networkValue, duint32 &hostValue) const {
#ifndef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap32(networkValue);
#endif
}
void LittleEndianByteOrder::networkToHost(duint64 networkValue, duint64 &hostValue) const {
#ifndef __BIG_ENDIAN__
    hostValue = networkValue;
#else
    hostValue = swap64(networkValue);
#endif
}
void LittleEndianByteOrder::hostToNetwork(duint16 hostValue, duint16 &networkValue) const {
#ifndef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap16(hostValue);
#endif
}
void LittleEndianByteOrder::hostToNetwork(duint32 hostValue, duint32 &networkValue) const {
#ifndef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap32(hostValue);
#endif
}
void LittleEndianByteOrder::hostToNetwork(duint64 hostValue, duint64 &networkValue) const {
#ifndef __BIG_ENDIAN__
    networkValue = hostValue;
#else
    networkValue = swap64(hostValue);
#endif
}
duint64 swap64(duint64 n) {
    duint64 result;
    dbyte const *in = reinterpret_cast<dbyte const *>(&n);
    dbyte *out = reinterpret_cast<dbyte *>(&result);
    out[0] = in[7];
    out[1] = in[6];
    out[2] = in[5];
    out[3] = in[4];
    out[4] = in[3];
    out[5] = in[2];
    out[6] = in[1];
    out[7] = in[0];
    return result;
}
} // namespace de
namespace de {
ByteRefArray::ByteRefArray() : _writeBase(0), _readBase(0), _size(0)
{}
ByteRefArray::ByteRefArray(void *base, Size size)
    : _writeBase(reinterpret_cast<Byte *>(base)),
      _readBase(reinterpret_cast<Byte const *>(base)),
      _size(size)
{}
ByteRefArray::ByteRefArray(void const *base, Size size)
    : _writeBase(0),
      _readBase(reinterpret_cast<Byte const *>(base)),
      _size(size)
{}
ByteRefArray ByteRefArray::fromCStr(char const *nullTerminatedCStr) {
    return ByteRefArray(nullTerminatedCStr, std::strlen(nullTerminatedCStr));
}
void *ByteRefArray::base() {
    return _writeBase;
}
void const *ByteRefArray::base() const {
    return _readBase;
}
void ByteRefArray::clear() {
    fill(0);
}
void ByteRefArray::fill(IByteArray::Byte value) {
    if (!_writeBase) {
        throw NonModifiableError("ByteRefArray::fill", "Array is read-only");
    }
    std::memset(_writeBase, value, _size);
}
ByteRefArray::Size ByteRefArray::size() const {
    return _size;
}
void ByteRefArray::get(Offset at, Byte *values, Size count) const {
    DENG2_ASSERT(_readBase != 0);
    if (at + count > size()) {
        throw OffsetError("ByteRefArray::get", "Out of range");
    }
    std::memmove(values, _readBase + at, count);
}
void ByteRefArray::set(Offset at, Byte const *values, Size count) {
    if (!_writeBase) {
        throw NonModifiableError("ByteRefArray::set", "Array is read-only");
    }
    if (at + count > size()) {
        throw OffsetError("ByteRefArray::set", "Out of range");
    }
    std::memmove(_writeBase + at, values, count);
}
} // namespace de
namespace de {
ByteSubArray::ByteSubArray(IByteArray &mainArray, Offset at, Size size)
    : _mainArray(&mainArray), _constMainArray(&mainArray), _at(at), _size(size)
{}
ByteSubArray::ByteSubArray(IByteArray const &mainArray, Offset at, Size size)
    : _mainArray(nullptr), _constMainArray(&mainArray), _at(at), _size(size)
{}
ByteSubArray::ByteSubArray(IByteArray const &mainArray, Offset at)
    : _mainArray(nullptr), _constMainArray(&mainArray), _at(at), _size(mainArray.size() - at)
{}
ByteSubArray::Size ByteSubArray::size() const {
    return _size;
}
void ByteSubArray::get(Offset at, Byte *values, Size count) const {
    _constMainArray->get(_at + at, values, count);
}
void ByteSubArray::set(Offset at, Byte const *values, Size count) {
    if (!_mainArray) {
        throw NonModifiableError("ByteSubArray::set", "Array is non-modifiable.");
    }
    _mainArray->set(_at + at, values, count);
    _size = qMax(_size, at + count);
}
} // namespace de
namespace de {
DENG2_PIMPL(ConditionalTrigger)
, DENG2_OBSERVES(Variable, Change) {
    SafePtr<Variable const> condition;
    QSet<QString> activeTriggers;
    bool anyTrigger = false;
    Impl(Public *i) : Base(i)
    {}
    void variableValueChanged(Variable &, Value const &) override {
        update();
    }
    void update() {
        anyTrigger = false;
        activeTriggers.clear();
        if (!condition) return;
        StringList const trigs = condition->value().asStringList();
        for (String const &trig : trigs) {
            if (trig == "*") {
                anyTrigger = true;
                activeTriggers.clear();
                break;
            }
            activeTriggers << trig;
        }
    }
    bool check(String const &trigger) {
        if (anyTrigger) return true;
        return activeTriggers.contains(trigger);
    }
};
ConditionalTrigger::ConditionalTrigger()
    : d(new Impl(this))
{}
ConditionalTrigger::~ConditionalTrigger()
{}
bool ConditionalTrigger::isValid() const {
    return d->condition;
}
void ConditionalTrigger::setCondition(Variable const &variable) {
    if (d->condition) {
        d->condition->audienceForChange() -= d;
    }
    d->condition.reset(&variable);
    variable.audienceForChange() += d;
    d->update();
}
Variable const &ConditionalTrigger::condition() const {
    return *d->condition;
}
bool ConditionalTrigger::tryTrigger(String const &trigger) {
    if (d->check(trigger)) {
        handleTriggered(trigger);
        return true;
    }
    return false;
}
} // namespace de
namespace de {
#ifdef DENG2_DEBUG
std::atomic_int Counted::totalCount { 0 }; ///< Should return back to zero when program ends.
# ifdef DENG_USE_COUNTED_TRACING
static QHash<void *, QByteArray> countedAllocs;
void Counted::printAllocs() {
    qDebug() << "Counted objects:" << Counted::totalCount;
    for (auto i = countedAllocs.constBegin(); i != countedAllocs.constEnd(); ++i) {
        qDebug() << "-=- Object" << i.key() << "\n";
        qDebug() << i.value().constData() << "\n";
    }
}
# endif
#endif
Counted::Counted() : _refCount(1) {
#ifdef DENG2_DEBUG
    totalCount++;
# ifdef DENG_USE_COUNTED_TRACING
    QString trace;
    DENG2_BACKTRACE(32, trace);
    countedAllocs[this] = trace.toLatin1();
# endif
#endif
}
Counted::~Counted() {
#ifdef DENG2_DEBUG
    totalCount--;
# ifdef DENG_USE_COUNTED_TRACING
    countedAllocs.remove(this);
# endif
#endif
    DENG2_ASSERT(_refCount == 0);
}
void Counted::release() const {
    DENG2_ASSERT(_refCount > 0);
    if (!--_refCount) {
        delete this;
    }
}
void Counted::addRef(dint count) const {
    DENG2_ASSERT(_refCount >= 0);
    _refCount += count;
}
} // namespace de
namespace de {
Date::Date()
{}
Date::Date(Time const &time) : Time(time)
{}
Time Date::asTime() const {
    return *this;
}
Date Date::fromText(String const &text) {
    return Time::fromText(text, HumanDate);
}
String Date::asText() const {
    String result;
    QTextStream os(&result);
    os << *this;
    return result;
}
QTextStream &operator << (QTextStream &os, Date const &d) {
    os << d.asDateTime().toString("yyyy-MM-dd");
    return os;
}
} // namespace de
namespace de {
Deletable::~Deletable() {
    DENG2_FOR_AUDIENCE(Deletion, i) {
        i->objectWasDeleted(this);
    }
}
} // namespace de
namespace de {
DictionaryValue::DictionaryValue() : /*_iteration(0),*/ _validIteration(false)
{}
DictionaryValue::DictionaryValue(DictionaryValue const &other)
    : Value(), /*_iteration(0),*/ _validIteration(false)
{
    for (Elements::const_iterator i = other._elements.begin(); i != other._elements.end(); ++i) {
        Value *value = i->second->duplicate();
        _elements[ValueRef(i->first.value->duplicate())] = value;
    }
}
DictionaryValue::~DictionaryValue() {
    clear();
}
void DictionaryValue::clear() {
    for (Elements::iterator i = _elements.begin(); i != _elements.end(); ++i) {
        delete i->first.value;
        delete i->second;
    }
    _elements.clear();
}
void DictionaryValue::add(Value *key, Value *value) {
    Elements::iterator existing = _elements.find(ValueRef(key));
    if (existing != _elements.end()) {
        delete existing->second;
        existing->second = value;
        delete key;
    }
    else {
        _elements[ValueRef(key)] = value;
    }
}
void DictionaryValue::remove(Value const &key) {
    Elements::iterator i = _elements.find(ValueRef(&key));
    if (i != _elements.end()) {
        remove(i);
    }
}
void DictionaryValue::remove(Elements::iterator const &pos) {
    delete pos->first.value;
    delete pos->second;
    _elements.erase(pos);
}
const Value *DictionaryValue::find(const Value &key) const {
    auto i = _elements.find(ValueRef(&key));
    if (i != _elements.end()) {
        return i->second;
    }
    return nullptr;
}
ArrayValue *DictionaryValue::contentsAsArray(ContentSelection selection) const {
    QScopedPointer<ArrayValue> array(new ArrayValue);
    DENG2_FOR_EACH_CONST(Elements, i, elements()) {
        if (selection == Keys) {
            array->add(i->first.value->duplicateAsReference());
        }
        else {
            array->add(i->second->duplicateAsReference());
        }
    }
    return array.take();
}
Value::Text DictionaryValue::typeId() const {
    return "Dictionary";
}
Value *DictionaryValue::duplicate() const {
    return new DictionaryValue(*this);
}
Value::Text DictionaryValue::asText() const {
    Text result;
    QTextStream s(&result);
    s << "{";
    bool isFirst = true;
    bool hadNewline = false;
    for (Elements::const_iterator i = _elements.begin(); i != _elements.end(); ++i) {
        String const label = i->first.value->asText() + ": ";
        String content = i->second->asText();
        bool const multiline = content.contains('\n');
        if (!isFirst) {
            if (hadNewline || multiline) s << "\n";
            s << ",";
        }
        hadNewline = multiline;
        s << " " << label << content.replace("\n", "\n" + String(label.size() + 2, ' '));
        isFirst = false;
    }
    s << " }";
    return result;
}
Record *DictionaryValue::memberScope() const {
    return &ScriptSystem::builtInClass("Dictionary");
}
dsize DictionaryValue::size() const {
    return _elements.size();
}
Value const &DictionaryValue::element(Value const &index) const {
    Elements::const_iterator i = _elements.find(ValueRef(&index));
    if (i == _elements.end()) {
        throw KeyError("DictionaryValue::element",
            "Key '" + index.asText() + "' does not exist in the dictionary");
    }
    return *i->second;
}
Value &DictionaryValue::element(Value const &index) {
    return const_cast<Value &>(const_cast<DictionaryValue const *>(this)->element(index));
}
void DictionaryValue::setElement(Value const &index, Value *value) {
    Elements::iterator i = _elements.find(ValueRef(&index));
    if (i == _elements.end()) {
        _elements[ValueRef(index.duplicate())] = value;
    }
    else {
        delete i->second;
        i->second = value;
    }
}
bool DictionaryValue::contains(Value const &value) const {
    return _elements.find(ValueRef(&value)) != _elements.end();
}
Value *DictionaryValue::begin() {
    _validIteration = false;
    return next();
}
Value *DictionaryValue::next() {
    if (!_validIteration) {
        _iteration = _elements.begin();
        _validIteration = true;
    }
    else if (_iteration == _elements.end()) {
        return 0;
    }
    ArrayValue *pair = new ArrayValue;
    pair->add(_iteration->first.value->duplicate());
    pair->add(_iteration->second->duplicate());
    ++_iteration;
    return pair;
}
bool DictionaryValue::isTrue() const {
    return size() > 0;
}
dint DictionaryValue::compare(Value const &value) const {
    DictionaryValue const *other = dynamic_cast<DictionaryValue const *>(&value);
    if (other) {
        if (size() < other->size()) {
            return -1;
        }
        if (size() > other->size()) {
            return 1;
        }
        Elements::const_iterator mine = _elements.begin();
        Elements::const_iterator theirs = other->_elements.begin();
        for (; mine != _elements.end() && theirs != other->_elements.end(); ++mine, ++theirs) {
            dint result = mine->first.value->compare(*theirs->first.value);
            if (result) return result;
            result = mine->second->compare(*theirs->second);
            if (result) return result;
        }
        return 0;
    }
    return Value::compare(value);
}
void DictionaryValue::sum(Value const &value) {
    DictionaryValue const *other = dynamic_cast<DictionaryValue const *>(&value);
    if (!other) {
        throw ArithmeticError("DictionaryValue::sum", "Values cannot be summed");
    }
    for (Elements::const_iterator i = other->_elements.begin();
        i != other->_elements.end(); ++i) {
        add(i->first.value->duplicate(), i->second->duplicate());
    }
}
void DictionaryValue::subtract(Value const &subtrahend) {
    Elements::iterator i = _elements.find(ValueRef(&subtrahend));
    if (i == _elements.end()) {
        throw KeyError("DictionaryValue::subtract",
            "Key '" + subtrahend.asText() + "' does not exist in the dictionary");
    }
    remove(i);
}
void DictionaryValue::operator >> (Writer &to) const {
    to << SerialId(DICTIONARY) << duint(_elements.size());
    if (!_elements.empty()) {
        for (Elements::const_iterator i = _elements.begin(); i != _elements.end(); ++i) {
            to << *i->first.value << *i->second;
        }
    }
}
void DictionaryValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != DICTIONARY) {
        throw DeserializationError("DictionaryValue::operator <<", "Invalid ID");
    }
    duint count = 0;
    from >> count;
    clear();
    while (count--) {
        QScopedPointer<Value> key(Value::constructFrom(from));
        QScopedPointer<Value> value(Value::constructFrom(from));
        add(key.take(), value.take());
    }
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(EscapeParser) {
    String original;
    String plain;
    DENG2_PIMPL_AUDIENCE(PlainText)
    DENG2_PIMPL_AUDIENCE(EscapeSequence)
};
DENG2_AUDIENCE_METHOD(EscapeParser, PlainText)
DENG2_AUDIENCE_METHOD(EscapeParser, EscapeSequence)
EscapeParser::EscapeParser() : d(new Impl)
{}
void EscapeParser::parse(String const &textWithEscapes) {
    d->original = textWithEscapes;
    d->plain.clear();
    Rangei range;
    forever {
        range.end = d->original.indexOf(QChar('\x1b'), range.start);
        if (range.end >= 0) {
            if (range.size() > 0) {
                DENG2_FOR_AUDIENCE2(PlainText, i) {
                    i->handlePlainText(range);
                }
                d->plain += d->original.substr(range);
            }
            int escLen = 2;
            char ch = d->original[range.end + 1].toLatin1();
            switch (ch) {
            case '(':
            case '[':
            case '{': {
                int end = d->original.indexOf(ch == '('? ')' : ch == '['? ']' : '}', range.end + 1);
                if (end < 0) end = d->original.size() - 1;
                escLen = end - range.end + 1;
                break; }
            case 'T':
                escLen = 3;
                break;
            default:
                break;
            }
            DENG2_FOR_AUDIENCE2(EscapeSequence, i) {
                i->handleEscapeSequence(Rangei(range.end + 1, range.end + escLen));
            }
            range.start = range.end + escLen;
        }
        else {
            range.end = d->original.size();
            if (range.size() > 0) {
                DENG2_FOR_AUDIENCE2(PlainText, i) {
                    i->handlePlainText(range);
                }
                d->plain += d->original.substr(range);
            }
            break;
        }
    }
}
String EscapeParser::originalText() const {
    return d->original;
}
String EscapeParser::plainText() const {
    return d->plain;
}
} // namespace de
namespace de {
FixedByteArray::FixedByteArray(IByteArray &mainArray)
    : ByteSubArray(mainArray, 0, mainArray.size())
{}
FixedByteArray::FixedByteArray(IByteArray &mainArray, Offset at, Size size)
    : ByteSubArray(mainArray, at, size)
{}
FixedByteArray::FixedByteArray(IByteArray const &mainArray)
    : ByteSubArray(mainArray, 0, mainArray.size())
{}
FixedByteArray::FixedByteArray(IByteArray const &mainArray, Offset at, Size size)
    : ByteSubArray(mainArray, at, size)
{}
void FixedByteArray::set(Offset at, Byte const *values, Size count) {
    if (at + count > size()) {
        throw OffsetError("FixedByteArray::set", "Fixed byte arrays cannot grow");
    }
    ByteSubArray::set(at, values, count);
}
} // namespace de
namespace de {
Block gDecompress(Block const &gzData) {
    Block result(16384);
    z_stream stream = {};
    stream.next_in   = const_cast<Block::Byte *>(gzData.data());
    stream.avail_in  = uint(gzData.size());
    stream.next_out  = result.data();
    stream.avail_out = uint(result.size());
    if (inflateInit2(&stream, 16 + MAX_WBITS) != Z_OK) {
        return Block();
    }
    int res = 0;
    do {
        res = inflate(&stream, 0);
        switch (res) {
        case Z_OK:
            if (stream.avail_out == 0) {
                auto const oldSize = result.size();
                result.resize(result.size() * 2);
                stream.next_out = result.data() + oldSize;
                stream.avail_out = uint(result.size() - oldSize);
            }
            continue;
        case Z_STREAM_END:
            break;
        default:
            qWarning("Error decompressing gzip data: result=%i (%s)", res, stream.msg);
            inflateEnd(&stream);
            return Block();
        }
    }
    while (res != Z_STREAM_END);
    DENG2_ASSERT(stream.avail_in == 0);
    result.resize(result.size() - stream.avail_out);
    inflateEnd(&stream);
    return result;
}
} // namespace de
namespace de {
namespace internal {
/*
 * Total number of bytes: 234457 (10217 packets)
 * Frequencies calculated in Doom II, co-op (1p)
 */
static double freqs[256] = {
    0.3108032603, 0.0030495997, 0.0035443599, 0.0023202549,
    0.0018638812, 0.0026188171, 0.0021752390, 0.0027083858,
    0.0175810490, 0.0011302712, 0.0010748240, 0.0015013414,
    0.0012241051, 0.0015951752, 0.0012923479, 0.0012795523,
    0.0011004150, 0.0013477951, 0.0434066801, 0.0016506225,
    0.0019790409, 0.0017146001, 0.0010108463, 0.0012113095,
    0.0014629548, 0.0013605906, 0.0015482583, 0.0017103349,
    0.0024055584, 0.0010151115, 0.0009980508, 0.0011558623,
    0.0015354628, 0.0012496961, 0.0015141369, 0.0021283220,
    0.0012241051, 0.0015311976, 0.0010534981, 0.0018510857,
    0.0013989772, 0.0013563255, 0.0015226673, 0.0012283702,
    0.0011302712, 0.0010790891, 0.0011601274, 0.0010236419,
    0.0013008782, 0.0012283702, 0.0013648558, 0.0011132105,
    0.0012624916, 0.0016165011, 0.0018596160, 0.0030240087,
    0.0018084340, 0.0013989772, 0.0013179389, 0.0012369006,
    0.0025932260, 0.0016719484, 0.0016463573, 0.0019406544,
    0.0122026640, 0.0017401912, 0.0144632065, 0.0403186938,
    0.0779332671, 0.0014970762, 0.0025207181, 0.0021027310,
    0.0018681464, 0.0014629548, 0.0014586897, 0.0011985140,
    0.0013563255, 0.0013094085, 0.0014928110, 0.0014586897,
    0.0015098717, 0.0014586897, 0.0012070444, 0.0017401912,
    0.0012454309, 0.0018126991, 0.0022264210, 0.0018297598,
    0.0027297116, 0.0012496961, 0.0013222041, 0.0016165011,
    0.0021453827, 0.0024695360, 0.0015994404, 0.0016676832,
    0.0011814533, 0.0021539131, 0.0013904469, 0.0015269324,
    0.0023586415, 0.0016420922, 0.0011558623, 0.0013819165,
    0.0012241051, 0.0013904469, 0.0013136737, 0.0020771399,
    0.0024865967, 0.0015482583, 0.0011899837, 0.0013136737,
    0.0012624916, 0.0016250315, 0.0017828429, 0.0014970762,
    0.0014629548, 0.0017529867, 0.0012411658, 0.0021411176,
    0.0023671718, 0.0019961016, 0.0015951752, 0.0025974912,
    0.0013051434, 0.0020728748, 0.0016079708, 0.0021283220,
    0.0550079546, 0.0033694878, 0.0025889609, 0.0021624434,
    0.0029728266, 0.0022946638, 0.0021283220, 0.0018510857,
    0.0020216927, 0.0017700474, 0.0018809419, 0.0015525235,
    0.0022562773, 0.0028832579, 0.0020899355, 0.0018425554,
    0.0024610056, 0.0020899355, 0.0017188653, 0.0021112613,
    0.0018638812, 0.0017231305, 0.0018254947, 0.0015951752,
    0.0020814051, 0.0020174275, 0.0019193285, 0.0014032424,
    0.0017572519, 0.0017913733, 0.0020003668, 0.0018510857,
    0.0022264210, 0.0012923479, 0.0017529867, 0.0018468205,
    0.0017359260, 0.0018596160, 0.0018084340, 0.0025463091,
    0.0011430667, 0.0022221559, 0.0010407026, 0.0012411658,
    0.0015354628, 0.0019235937, 0.0022178907, 0.0013819165,
    0.0021837693, 0.0015823797, 0.0013008782, 0.0011814533,
    0.0010492329, 0.0015695842, 0.0014160379, 0.0015823797,
    0.0014928110, 0.0019107981, 0.0012369006, 0.0019619802,
    0.0017913733, 0.0023799673, 0.0016037056, 0.0020174275,
    0.0148854587, 0.0032841843, 0.0018126991, 0.0023159897,
    0.0015056066, 0.0026955902, 0.0019747758, 0.0012624916,
    0.0011558623, 0.0014672200, 0.0017572519, 0.0022520121,
    0.0013136737, 0.0012752872, 0.0012411658, 0.0017743126,
    0.0014458941, 0.0012241051, 0.0012752872, 0.0017615170,
    0.0012113095, 0.0011515971, 0.0013776513, 0.0010748240,
    0.0016250315, 0.0012283702, 0.0014117727, 0.0009596642,
    0.0011430667, 0.0010705588, 0.0013264692, 0.0012923479,
    0.0025889609, 0.0013733862, 0.0013136737, 0.0012752872,
    0.0014970762, 0.0011899837, 0.0013691210, 0.0010023160,
    0.0014416290, 0.0010876195, 0.0010662936, 0.0009340732,
    0.0011814533, 0.0010577633, 0.0012710220, 0.0017316608,
    0.0014586897, 0.0010449677, 0.0017359260, 0.0010279070,
    0.0016292966, 0.0018297598, 0.0020259579, 0.0015311976,
    0.0040775067, 0.0010790891, 0.0013861817, 0.0010108463,
    0.0017103349, 0.0012496961, 0.0022903987, 0.0028619320
};
struct HuffNode {
    HuffNode *left, *right;
    double freq;
    dbyte value;              // Only valid for leaves.
};
struct HuffQueue {
    HuffNode *nodes[256];
    int count;
};
struct HuffCode {
    duint code;
    duint length;
};
struct HuffBuffer {
    dbyte *data;
    dsize size;
};
struct Huffman {
    HuffNode *huffRoot;
    HuffCode huffCodes[256];
    /**
     * Builds the Huffman tree and initializes the code lookup.
     */
    Huffman() : huffRoot(0) {
        zap(huffCodes);
        HuffQueue queue;
        HuffNode *node;
        int i;
        queue.count = 0;
        for (i = 0; i < 256; ++i) {
            node = (HuffNode *) calloc(1, sizeof(HuffNode));
            node->freq = freqs[i];
            node->value = i;
            Huff_QueueInsert(&queue, node);
        }
        for (i = 0; i < 255; ++i) {
            node = (HuffNode *) calloc(1, sizeof(HuffNode));
            node->left = Huff_QueueExtract(&queue);
            node->right = Huff_QueueExtract(&queue);
            node->freq = node->left->freq + node->right->freq;
            Huff_QueueInsert(&queue, node);
        }
        huffRoot = Huff_QueueExtract(&queue);
        Huff_BuildLookup(huffRoot, 0, 0);
#if 0
        if (qApp->arguments().contains("-huffcodes")) {
            double realBits = 0;
            double huffBits = 0;
            LOG_MSG("Huffman codes:");
            for (i = 0; i < 256; ++i) {
                uint k;
                realBits += freqs[i] * 8;
                huffBits += freqs[i] * huffCodes[i].length;
                LOG_MSG("%03i: (%07i) ", i, (int) (6e6 * freqs[i]));
                for (k = 0; k < huffCodes[i].length; ++k) {
                    LogBuffer_Printf(0, "%c", huffCodes[i].code & (1 << k) ? '1' : '0');
                }
                LogBuffer_Printf(0, "\n");
            }
            LOG_MSG("realbits=%f, huffbits=%f (%f%%)")
                    << realBits <<huffBits << huffBits / realBits * 100;
        }
#endif
    }
    /**
     * Free all resources allocated for Huffman codes.
     */
    ~Huffman() {
        Huff_DestroyNode(huffRoot);
        huffRoot = NULL;
    }
    /**
     * Exchange two nodes in the queue.
     */
    void Huff_QueueExchange(HuffQueue *queue, int index1, int index2) {
        HuffNode *temp = queue->nodes[index1];
        queue->nodes[index1] = queue->nodes[index2];
        queue->nodes[index2] = temp;
    }
    /**
     * Insert a node into a priority queue.
     */
    void Huff_QueueInsert(HuffQueue *queue, HuffNode *node) {
        int i, parent;
        i = queue->count;
        queue->nodes[i] = node;
        ++queue->count;
        while (i > 0) {
            parent = HEAP_PARENT(i);
            if (queue->nodes[parent]->freq <= node->freq)
                break;
            Huff_QueueExchange(queue, parent, i);
            i = parent;
        }
    }
    /**
     * Extract the smallest node from the queue.
     */
    HuffNode *Huff_QueueExtract(HuffQueue *queue) {
        HuffNode *min;
        int i, left, right, small;
        DENG2_ASSERT(queue->count > 0);
        min = queue->nodes[0];
        queue->nodes[0] = queue->nodes[--queue->count];
        i = 0;
        for (;;) {
            left = HEAP_LEFT(i);
            right = HEAP_RIGHT(i);
            small = i;
            if (left < queue->count &&
               queue->nodes[left]->freq < queue->nodes[i]->freq) {
                small = left;
            }
            if (right < queue->count &&
               queue->nodes[right]->freq < queue->nodes[small]->freq) {
                small = right;
            }
            if (i == small) {
                break;
            }
            Huff_QueueExchange(queue, i, small);
            i = small;
        }
        return min;
    }
    /**
     * Recursively builds the Huffman code lookup for the node's subtree.
     */
    void Huff_BuildLookup(HuffNode *node, uint code, uint length) {
        if (!node->left && !node->right) {
            huffCodes[node->value].code = code;
            huffCodes[node->value].length = length;
            return;
        }
        DENG2_ASSERT(length < 32);
        if (node->left) {
            Huff_BuildLookup(node->left, code, length + 1);
        }
        if (node->right) {
            Huff_BuildLookup(node->right, code | (1 << length), length + 1);
        }
    }
    /**
     * Checks if the encoding/decoding buffer can hold the given number of
     * bytes. If not, reallocates the buffer.
     */
    static void Huff_ResizeBuffer(HuffBuffer *buffer, dsize neededSize) {
        while (neededSize > buffer->size) {
            if (!buffer->size)
                buffer->size = qMax(dsize(1024), neededSize);
            else
                buffer->size *= 2;
        }
        buffer->data = reinterpret_cast<dbyte *>(realloc(buffer->data, buffer->size));
    }
    /**
     * Recursively frees the node and its subtree.
     */
    void Huff_DestroyNode(HuffNode *node) {
        if (node) {
            Huff_DestroyNode(node->left);
            Huff_DestroyNode(node->right);
            free(node);
        }
    }
    /**
     * Free the buffer.
     */
    void Huff_DestroyBuffer(HuffBuffer *buffer) {
        free(buffer->data);
        zapPtr(buffer);
    }
    dbyte *encode(dbyte const *data, dsize size, dsize *encodedSize) const {
        HuffBuffer huffEnc;
        dsize i;
        duint code;
        int remaining, fits;
        dbyte *out, bit;
        zap(huffEnc);
        Huff_ResizeBuffer(&huffEnc, 2 * size);
        bit = 3;
        out = huffEnc.data;
        *out = 0;
        for (i = 0; i < size; ++i) {
            remaining = huffCodes[data[i]].length;
            code = huffCodes[data[i]].code;
            while (remaining > 0) {
                fits = 8 - bit;
                if (fits > remaining)
                    fits = remaining;
                *out |= code << bit;
                code >>= fits;
                remaining -= fits;
                bit += fits;
                if (bit == 8) {
                    bit = 0;
                    *++out = 0;
                }
            }
        }
        if (bit == 0) {
            out--;
            bit = 8;
        }
        if (encodedSize)
            *encodedSize = out - huffEnc.data + 1;
        huffEnc.data[0] |= bit - 1;
        return huffEnc.data;
    }
    dbyte *decode(dbyte const *data, dsize size, dsize *decodedSize) const {
        HuffBuffer huffDec;
        HuffNode *node;
        dsize outBytes = 0;
        dbyte const *in = data;
        dbyte const *lastIn = in + size - 1;
        dbyte bit = 3, lastByteBits;
        if (!data || size == 0) return nullptr;
        zap(huffDec);
        Huff_ResizeBuffer(&huffDec, 256);
        lastByteBits = (*in & 7) + 1;
        node = huffRoot;
        while (in < lastIn || bit < lastByteBits) {
            if (*in & (1 << bit)) {
                node = node->right;
            }
            else {
                node = node->left;
            }
            DENG2_ASSERT(node);
            if (!node->left && !node->right) {
                huffDec.data[outBytes++] = node->value;
                if (outBytes == huffDec.size) {
                    Huff_ResizeBuffer(&huffDec, 2 * huffDec.size);
                }
                node = huffRoot;
            }
            if (++bit == 8) {
                bit = 0;
                ++in;
                if (in > lastIn)
                    break;
            }
        }
        if (decodedSize) {
            *decodedSize = outBytes;
        }
        return huffDec.data;
    }
};
} // namespace internal
static internal::Huffman huff;
Block codec::huffmanEncode(Block const &data) {
    Block result;
    dsize size = 0;
    dbyte *coded = huff.encode(data.data(), data.size(), &size);
    if (coded) {
        result.copyFrom(ByteRefArray(coded, size), 0, size);
        free(coded);
    }
    return result;
}
Block codec::huffmanDecode(Block const &codedData) {
    Block result;
    dsize size = 0;
    dbyte *decoded = huff.decode(codedData.data(), codedData.size(), &size);
    if (decoded) {
        result.copyFrom(ByteRefArray(decoded, size), 0, size);
        free(decoded);
    }
    return result;
}
} // namespace de
namespace de {
static QString const WHITESPACE = " \t\r\n";
static QString const WHITESPACE_OR_COMMENT = " \t\r\n#";
static QString const TOKEN_BREAKING_CHARS = "#:=$(){}<>,;\"" + WHITESPACE;
static QString const INCLUDE_TOKEN = "@include";
static QString const SCRIPT_TOKEN = "script";
static String const GROUP_TOKEN = "group";
static SourceLineTable sourceLineTable;
DENG2_PIMPL(Info) {
    DENG2_ERROR(OutOfElements);
    DENG2_ERROR(EndOfFile);
    struct DefaultIncludeFinder : public IIncludeFinder {
        String findIncludedInfoSource(String const &includeName, Info const &info,
                                      String *sourcePath) const {
            String path = info.sourcePath().fileNamePath() / includeName;
            if (sourcePath) *sourcePath = path;
            return String::fromUtf8(Block(App::rootFolder().locate<File const>(path)));
        }
    };
    QStringList scriptBlockTypes;
    QStringList allowDuplicateBlocksOfType;
    String implicitBlockType = GROUP_TOKEN;
    String sourcePath; ///< May be unknown (empty).
    String content;
    int currentLine = 0;
    int cursor = 0; ///< Index of the next character from the source.
    QChar currentChar;
    int tokenStartOffset = 0;
    String currentToken;
    BlockElement rootBlock;
    DefaultIncludeFinder defaultFinder;
    IIncludeFinder const *finder = &defaultFinder;
    using InfoValue = Info::Element::Value;
    Impl(Public *i)
        : Base(i)
        , rootBlock("", "", *i) {
        scriptBlockTypes << SCRIPT_TOKEN;
    }
    /**
     * Initialize the parser for reading a block of source content.
     * @param source  Text to be parsed.
     */
    void init(String const &source) {
        rootBlock.clear();
        content = source + "\n";
        currentLine = 1;
        currentChar = '\0';
        cursor = 0;
        currentToken = " ";
        tokenStartOffset = 0;
        if (source.isEmpty()) {
            content.clear();
            currentLine = 0;
        }
        try {
            nextChar();
            nextToken();
        }
        catch(EndOfFile const &) {
            currentToken.clear();
        }
    }
    /**
     * Returns the next character from the source file.
     */
    QChar peekChar() {
        return currentChar;
    }
    /**
     * Move to the next character in the source file.
     */
    void nextChar() {
        if (cursor >= content.size()) {
            throw EndOfFile(QString("EOF on line %1").arg(currentLine));
        }
        if (currentChar == '\n') {
            currentLine++;
        }
        currentChar = content[cursor];
        cursor++;
    }
    /**
     * Read a line of text from the content and return it.
     */
    String readLine() {
        String line;
        nextChar();
        while (currentChar != '\n') {
            line += currentChar;
            nextChar();
        }
        return line;
    }
    /**
     * Read until a newline is encountered. Returns the contents of the line.
     */
    String readToEOL() {
        cursor = tokenStartOffset;
        String line = readLine();
        try {
            nextChar();
        }
        catch (EndOfFile const &) {
        }
        return line;
    }
    String peekToken() {
        return currentToken;
    }
    /**
     * Returns the next meaningful token from the source file.
     */
    String nextToken() {
        if (currentToken.isEmpty()) {
            throw EndOfFile(QStringLiteral("out of tokens"));
        }
        currentToken = "";
        try {
            while (WHITESPACE_OR_COMMENT.contains(peekChar())) {
                if (peekChar() == '#') readLine();
                nextChar();
            }
            tokenStartOffset = cursor;
            currentToken += peekChar();
            nextChar();
            if (TOKEN_BREAKING_CHARS.contains(currentToken[0]))
                return currentToken;
            while (!TOKEN_BREAKING_CHARS.contains(peekChar())) {
                currentToken += peekChar();
                nextChar();
            }
        }
        catch (EndOfFile const &)
        {}
        return currentToken;
    }
    /**
     * This is the method that the user calls to retrieve the next element from
     * the source file. If there are no more elements to return, a
     * OutOfElements exception is thrown.
     *
     * @return Parsed element. Caller gets owernship.
     */
    Element *get() {
        Element *e = parseElement();
        if (!e) throw OutOfElements("");
        return e;
    }
    /**
     * Returns the next element from the source file.
     * @return An instance of one of the Element classes, or @c NULL if there are none.
     */
    Element *parseElement() {
        String key;
        String next;
        try {
            key = peekToken();
            next = nextToken();
        }
        catch (EndOfFile const &) {
            return 0;
        }
        int const elementLine = currentLine;
        Element *result = 0;
        if (next == ":" || next == "=" || next == "$") {
            result = parseKeyElement(key);
        }
        else if (next == "<") {
            result = parseListElement(key);
        }
        else {
            result = parseBlockElement(key);
        }
        DENG2_ASSERT(result != 0);
        result->setSourceLocation(sourcePath, elementLine);
        return result;
    }
    /**
     * Parse a string literal. Returns the string sans the quotation marks in
     * the beginning and the end.
     */
    String parseString() {
        if (peekToken() != "\"") {
            throw SyntaxError("Info::parseString",
                              QString("Expected string to begin with '\"', but '%1' found instead (on line %2).")
                              .arg(peekToken()).arg(currentLine));
        }
        String chars;
        while (peekChar() != '"') {
            if (peekChar() == '\'') {
                nextChar();
                if (peekChar() == '\'') {
                    chars.append("\"");
                }
                else {
                    chars.append("'");
                    continue;
                }
            }
            else {
                chars.append(peekChar());
            }
            nextChar();
        }
        nextChar();
        nextToken();
        return chars;
    }
    /**
     * Parse a value from the source file. The current token should be on the
     * first token of the value. Values come in different flavours:
     * - single token
     * - string literal (can be split)
     *
     * An optional semicolon may follow a value.
     */
    InfoValue parseValue() {
        InfoValue value;
        if (peekToken() == "$") {
            value.flags |= InfoValue::Script;
            nextToken();
        }
        if (peekToken() == "\"") {
            value.flags |= InfoValue::StringLiteral;
            while (peekToken() == "\"") {
                value.text += parseString();
            }
        }
        else {
            if (peekToken() != ";") {
                value = peekToken();
                nextToken();
                if (peekToken() == ";") nextToken(); // Ignore the semicolon.
            }
        }
        return value;
    }
    InfoValue parseScript(int requiredStatementCount = 0) {
        int startPos = cursor - 1;
        String remainder = content.substr(startPos);
        ScriptLex lex(remainder);
        TokenBuffer tokens;
        int count = 0;
        while (lex.getStatement(tokens, ScriptLex::StopAtMismatchedCloseBrace)) {
            if (requiredStatementCount > 0 &&
               ++count == requiredStatementCount) break; // We're good now.
        }
        int endPos = startPos + int(lex.pos());
        do { nextChar(); } while (cursor < endPos); // fast-forward
        currentToken = QString(peekChar());
        nextChar();
        if (currentToken != ")" && currentToken != "}") {
            nextToken();
        }
        return InfoValue(content.substr(startPos, int(lex.pos()) - 1).trimmed(), InfoValue::Script);
    }
    /**
     * Parse a key element.
     * @param name Name of the parsed key element.
     */
    KeyElement *parseKeyElement(String const &name) {
        InfoValue value;
        if (peekToken() == "$") {
            value.flags |= InfoValue::Script;
            nextToken();
        }
        if (peekToken() == ":") {
            value.text = readToEOL().trimmed();
            nextToken();
        }
        else if (peekToken() == "=") {
            if (value.flags.testFlag(InfoValue::Script)) {
                value = parseScript(1);
                value.text = value.text.trimmed();
            }
            else {
                /**
                 * Key =
                 *   "This is a long string "
                 *   "that spans multiple lines."
                 */
                nextToken();
                value.text = parseValue();
            }
        }
        else {
            throw SyntaxError("Info::parseKeyElement",
                              QString("Expected either '=' or ':', but '%1' found instead (on line %2).")
                              .arg(peekToken()).arg(currentLine));
        }
        return new KeyElement(name, value);
    }
    /**
     * Parse a list element, identified by the given name.
     */
    ListElement *parseListElement(String const &name) {
        if (peekToken() != "<") {
            throw SyntaxError("Info::parseListElement",
                              QString("List must begin with a '<', but '%1' found instead (on line %2).")
                              .arg(peekToken()).arg(currentLine));
        }
        QScopedPointer<ListElement> element(new ListElement(name));
        nextToken();
        if (peekToken() == ">") {
            nextToken();
            return element.take();
        }
        forever {
            element->add(parseValue());
            String separator = peekToken();
            nextToken();
            if (separator == ">") break;
            if (separator != ",") {
                throw SyntaxError("Info::parseListElement",
                                  QString("List values must be separated with a comma, but '%1' found instead (on line %2).")
                                  .arg(separator).arg(currentLine));
            }
        }
        return element.take();
    }
    /**
     * Parse a block element, identified by the given name.
     * @param blockType Identifier of the block.
     */
    BlockElement *parseBlockElement(String blockType) {
        DENG2_ASSERT(blockType != "}");
        DENG2_ASSERT(blockType != ")");
        String blockName;
        String endToken;
        std::unique_ptr<BlockElement> block;
        int startLine = currentLine;
        try {
            if (!scriptBlockTypes.contains(blockType)) // script blocks are never named
            {
                if (peekToken() != "(" && peekToken() != "{") {
                    blockName = parseValue();
                }
            }
            if (!implicitBlockType.isEmpty() && blockName.isEmpty() &&
                blockType != implicitBlockType &&
                !scriptBlockTypes.contains(blockType)) {
                blockName = blockType;
                blockType = implicitBlockType;
            }
            block.reset(new BlockElement(blockType, blockName, self()));
            startLine = currentLine;
            while (peekToken() != "(" && peekToken() != "{") {
                String keyName = peekToken();
                nextToken();
                if (peekToken() == "(" || peekToken() == "{") {
                    throw SyntaxError("Info::parseBlockElement",
                                      QString("Attribute on line %1 is missing a value")
                                      .arg(currentLine));
                }
                InfoValue value = parseValue();
                block->add(new KeyElement(keyName, value, KeyElement::Attribute));
            }
            endToken = (peekToken() == "("? ")" : "}");
            if (scriptBlockTypes.contains(blockType)) {
                block->add(new KeyElement(SCRIPT_TOKEN, parseScript()));
            }
            else {
                nextToken();
                while (peekToken() != endToken) {
                    Element *element = parseElement();
                    if (!element) {
                        throw SyntaxError("Info::parseBlockElement",
                                          QString("Block element (on line %1) was never closed, end of file encountered before '%2' was found (on line %3).")
                                          .arg(startLine).arg(endToken).arg(currentLine));
                    }
                    block->add(element);
                }
            }
            DENG2_ASSERT(peekToken() == endToken);
            nextToken();
        }
        catch (EndOfFile const &) {
            throw SyntaxError("Info::parseBlockElement",
                              QString("End of file encountered unexpectedly while parsing a block element (block started on line %1).")
                              .arg(startLine));
        }
        return block.release();
    }
    void includeFrom(String const &includeName) {
        try {
            DENG2_ASSERT(finder != 0);
            String includePath;
            String content = finder->findIncludedInfoSource(includeName, self(), &includePath);
            Info included;
            included.setImplicitBlockType(implicitBlockType);
            included.setScriptBlocks(scriptBlockTypes);
            included.setAllowDuplicateBlocksOfType(allowDuplicateBlocksOfType);
            included.setFinder(*finder); // use ours
            included.setSourcePath(includePath);
            included.parse(content);
            included.d->rootBlock.moveContents(rootBlock);
        }
        catch (Error const &er) {
            throw IIncludeFinder::NotFoundError("Info::includeFrom",
                    QString("Cannot include '%1': %2")
                    .arg(includeName)
                    .arg(er.asText()));
        }
    }
    void parse(String const &source) {
        init(source);
        forever {
            Element *e = parseElement();
            if (!e) break;
            if (e->isList() && e->name() == INCLUDE_TOKEN) {
                foreach (Element::Value const &val, e->as<ListElement>().values()) {
                    includeFrom(val);
                }
            }
            rootBlock.add(e);
        }
    }
    void parse(File const &file) {
        sourcePath = file.path();
        parse(String::fromUtf8(Block(file)));
    }
};
DENG2_PIMPL_NOREF(Info::Element) {
    Type type;
    String name;
    BlockElement *parent = nullptr;
    SourceLineTable::LineId sourceLine = 0;
};
Info::Element::Element(Type type, String const &name)
    : d(new Impl) {
    d->type = type;
    setName(name);
}
Info::Element::~Element()
{}
void Info::Element::setParent(BlockElement *parent) {
    d->parent = parent;
}
Info::BlockElement *Info::Element::parent() const {
    return d->parent;
}
void Info::Element::setSourceLocation(String const &sourcePath, int line) {
    d->sourceLine = de::sourceLineTable.lineId(sourcePath, line);
}
String Info::Element::sourceLocation() const {
    return de::sourceLineTable.sourceLocation(d->sourceLine);
}
duint32 Info::Element::sourceLineId() const {
    return d->sourceLine;
}
Info::Element::Type Info::Element::type() const {
    return d->type;
}
String const &Info::Element::name() const {
    return d->name;
}
void Info::Element::setName(String const &name) {
    d->name = name;
}
Info::BlockElement::~BlockElement() {
    clear();
}
void Info::BlockElement::clear() {
    for (ContentsInOrder::iterator i = _contentsInOrder.begin(); i != _contentsInOrder.end(); ++i) {
        delete *i;
    }
    _contents.clear();
    _contentsInOrder.clear();
}
void Info::BlockElement::add(Info::Element *elem) {
    DENG2_ASSERT(elem != 0);
    elem->setParent(this);
    _contentsInOrder.append(elem); // owned
    if (!elem->name().isEmpty()) {
        _contents.insert(elem->name().toLower(), elem); // not owned (name may be empty)
    }
}
Info::Element *Info::BlockElement::find(String const &name) const {
    Contents::const_iterator found = _contents.find(name.toLower());
    if (found == _contents.end()) return 0;
    return found.value();
}
Info::Element::Value Info::BlockElement::keyValue(String const &name,
                                                  String const &defaultValue) const {
    Element *e = findByPath(name);
    if (!e || !e->isKey()) return Value(defaultValue);
    return e->as<KeyElement>().value();
}
Info::Element *Info::BlockElement::findByPath(String const &path) const {
    String name;
    String remainder;
    int pos = path.indexOf(':');
    if (pos >= 0) {
        name = path.left(pos);
        remainder = path.mid(pos + 1);
    }
    else {
        name = path;
    }
    name = name.trimmed();
    Element *e = find(name);
    if (!e) return 0;
    if (e->isBlock()) {
        return e->as<BlockElement>().findByPath(remainder);
    }
    return e;
}
void Info::BlockElement::moveContents(BlockElement &destination) {
    foreach (Element *e, _contentsInOrder) {
        destination.add(e);
    }
    _contentsInOrder.clear();
    _contents.clear();
}
Record Info::BlockElement::asRecord() const {
    Record rec;
    for (auto i = _contents.begin(); i != _contents.end(); ++i) {
        const Element *elem = i.value();
        std::unique_ptr<Variable> var(new Variable(elem->name())); // retain case in the name
        if (elem->isBlock()) {
            var->set(RecordValue::takeRecord(elem->as<BlockElement>().asRecord()));
        }
        else if (elem->isList()) {
            std::unique_ptr<ArrayValue> array(new ArrayValue);
            for (const auto &value : elem->values()) {
                array->add(new TextValue(value.text));
            }
            var->set(array.release());
        }
        else {
            var->set(new TextValue(elem->as<KeyElement>().value().text));
        }
        rec.add(var.release());
    }
    return rec;
}
Info::Info() : d(new Impl(this))
{}
Info::Info(String const &source) : d(nullptr) {
    QScopedPointer<Impl> inst(new Impl(this)); // parsing may throw exception
    inst->parse(source);
    d.reset(inst.take());
}
Info::Info(File const &file) : d(nullptr) {
    QScopedPointer<Impl> inst(new Impl(this)); // parsing may throw exception
    inst->parse(file);
    d.reset(inst.take());
}
Info::Info(String const &source, IIncludeFinder const &finder) : d(nullptr) {
    QScopedPointer<Impl> inst(new Impl(this)); // parsing may throw exception
    inst->finder = &finder;
    inst->parse(source);
    d.reset(inst.take());
}
void Info::setFinder(IIncludeFinder const &finder) {
    d->finder = &finder;
}
void Info::useDefaultFinder() {
    d->finder = &d->defaultFinder;
}
void Info::setScriptBlocks(QStringList blocksToParseAsScript) {
    d->scriptBlockTypes = blocksToParseAsScript;
}
void Info::setAllowDuplicateBlocksOfType(QStringList duplicatesAllowed) {
    d->allowDuplicateBlocksOfType = duplicatesAllowed;
}
void Info::setImplicitBlockType(String const &implicitBlock) {
    d->implicitBlockType = implicitBlock;
}
void Info::parse(String const &infoSource) {
    d->parse(infoSource);
}
void Info::parse(File const &file) {
    d->parse(file);
}
void Info::parseNativeFile(NativePath const &nativePath) {
    QFile file(nativePath);
    if (file.open(QFile::ReadOnly | QFile::Text)) {
        parse(file.readAll().constData());
    }
}
void Info::clear() {
    d->sourcePath.clear();
    parse("");
}
void Info::setSourcePath(String const &path) {
    d->sourcePath = path;
}
String Info::sourcePath() const {
    return d->sourcePath;
}
Info::BlockElement const &Info::root() const {
    return d->rootBlock;
}
Info::Element const *Info::findByPath(String const &path) const {
    if (path.isEmpty()) return &d->rootBlock;
    return d->rootBlock.findByPath(path);
}
bool Info::findValueForKey(String const &key, String &value) const {
    Element const *element = findByPath(key);
    if (element && element->isKey()) {
        value = element->as<KeyElement>().value();
        return true;
    }
    return false;
}
bool Info::isEmpty() const {
    return d->rootBlock.isEmpty();
}
String Info::quoteString(String const &text) {
    String quoted = text;
    quoted.replace("\"", "''");
    return String("\"%1\"").arg(quoted);
}
String Info::sourceLocation(duint32 lineId) // static
{
    return de::sourceLineTable.sourceLocation(lineId);
}
SourceLineTable const &Info::sourceLineTable() // static
{
    return de::sourceLineTable;
}
} // namespace de
namespace de {
static String const VAR_NOT_IN_BANK("__notInBank__");
DENG2_PIMPL(InfoBank)
, DENG2_OBSERVES(ScriptedInfo, NamedBlock) {
    Record names; ///< All parsed sources will be stored here.
    ScriptedInfo info { &names };
    Time modTime;
    String relativeToPath;
    Impl(Public *i) : Base(i) {
        info.audienceForNamedBlock() += this;
    }
    void parsedNamedBlock(String const &, Record &block) {
        if (block.gets("__type__") != ScriptedInfo::BLOCK_GROUP) {
            block.addBoolean(VAR_NOT_IN_BANK, true);
        }
    }
    /**
     * Checks the contents of @a group and removes all blocks that return
     * @c true from @a shouldRemove. The removal is done both from the Bank and
     * the script namespace. Subgroups are checked recursively.
     */
    void removeFromGroup(Record &group,
                         std::function<bool (String const &, Record const &)> shouldRemove,
                         String identifierBase = "") {
        group.forSubrecords([this, &group, &shouldRemove, &identifierBase]
                            (String const &name, Record &sub) {
            String fullIdentifier = identifierBase.concatenateMember(name);
            if (ScriptedInfo::blockType(sub) == ScriptedInfo::BLOCK_GROUP) {
                removeFromGroup(sub, shouldRemove, fullIdentifier);
            }
            else if (shouldRemove(name, sub)) {
                LOG_RES_VERBOSE("Removing '%s' that was read from \"%s\"")
                        << fullIdentifier
                        << ScriptedInfo::sourceLocation(sub);
                self().remove(fullIdentifier);
                delete group.remove(name);
            }
            return LoopContinue;
        });
    }
};
InfoBank::InfoBank(char const *nameForLog, Bank::Flags const &flags, String const &hotStorageLocation)
    : Bank(nameForLog, flags, hotStorageLocation)
    , d(new Impl(this))
{}
void InfoBank::parse(String const &source) {
    try {
        d->relativeToPath = "";
        d->modTime = Time();
        d->info.parse(source);
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to read Info source:\n") << er.asText();
    }
}
void InfoBank::parse(File const &file) {
    try {
        d->relativeToPath = file.path().fileNamePath();
        d->modTime = file.status().modifiedAt;
        d->info.parse(file);
    }
    catch (Error const &er) {
        LOG_WARNING("Failed to read Info file:\n") << er.asText();
    }
}
ScriptedInfo &InfoBank::info() {
    return d->info;
}
ScriptedInfo const &InfoBank::info() const {
    return d->info;
}
Record &InfoBank::objectNamespace() {
    return d->names;
}
Record const &InfoBank::objectNamespace() const {
    return d->names;
}
void InfoBank::addFromInfoBlocks(String const &blockType) {
    foreach (String id, d->info.allBlocksOfType(blockType)) {
        Record &rec = d->names[id];
        if (!rec.has(VAR_NOT_IN_BANK)) {
            continue;
        }
        add(id, newSourceFromInfo(id));
        delete &rec[VAR_NOT_IN_BANK];
    }
}
void InfoBank::removeAllWithRootPath(String const &rootPath) {
    LOG_AS(nameForLog());
    d->removeFromGroup(d->names, [&rootPath] (String const &, Record const &rec) {
        return ScriptedInfo::sourceLocation(rec).startsWith(rootPath);
    });
}
void InfoBank::removeAllFromPackage(String const &packageId) {
    LOG_AS(nameForLog());
    d->removeFromGroup(d->names, [&packageId] (String const &, Record const &rec) {
        auto const loc = ScriptedInfo::sourcePathAndLine(rec);
        File const &file = App::rootFolder().locate<File const>(loc.first);
        return Package::identifierForContainerOfFile(file) == packageId;
    });
}
Time InfoBank::sourceModifiedAt() const {
    return d->modTime;
}
String InfoBank::bankRootPath() const {
    return d->relativeToPath;
}
String InfoBank::absolutePathInContext(Record const &context, String const &relativePath) const {
    if (Path(relativePath).isAbsolute()) {
        return relativePath;
    }
    String const path = ScriptedInfo::absolutePathInContext(context, relativePath);
    if (path == relativePath) {
        return bankRootPath() / relativePath;
    }
    return path;
}
} // namespace de
namespace de {
namespace internal {
/**
 * @internal Not exposed outside this source file; use parseJSON() instead.
 */
class JSONParser {
    QString const &source;
    int pos;
public:
    JSONParser(QString const &s) : source(s), pos(0) {
        skipWhite();
    }
    void advance() {
        pos++;
        skipWhite();
    }
    void skipWhite() {
        while (!atEnd() && source[pos].isSpace()) pos++;
    }
    bool atEnd() const {
        return pos >= source.size();
    }
    QChar peek() const {
        if (atEnd()) return 0;
        return source[pos];
    }
    QChar next() {
        if (atEnd()) return 0;
        QChar c = source[pos];
        advance();
        return c;
    }
    QChar nextNoSkip() {
        if (atEnd()) return 0;
        return source[pos++];
    }
    DENG2_NORETURN void error(QString const &message) {
        throw de::Error("JSONParser", de::String("Error at position %1 (%2^%3): %4")
                        .arg(pos).arg(source.mid(pos - 4, 4)).arg(source.mid(pos, 4)).arg(message));
    }
    QVariant parse() {
        LOG_AS("JSONParser");
        if (atEnd()) return QVariant();
        QChar c = peek();
        if (c == '{') {
            return parseObject();
        }
        else if (c == '[') {
            return parseArray();
        }
        else if (c == '\"') {
            return parseString();
        }
        else if (c == '-' || c.isDigit()) {
            return parseNumber();
        }
        else {
            return parseKeyword();
        }
    }
    QVariant parseObject() {
        QVariantMap result;
        QChar c = next();
        DENG2_ASSERT(c == '{');
        forever {
            if (peek() == '}') {
                break;
            }
            QString name = parseString().toString();
            c = next();
            if (c != ':') error("object keys and values must be separated by a colon");
            QVariant value = parse();
            result.insert(name, value);
            skipWhite();
            c = next();
            if (c == '}') {
                break;
            }
            else if (c != ',') {
                LOG_DEBUG(de::String("got %1 instead of ,").arg(c));
                error("key/value pairs must be separated by comma");
            }
        }
        return result;
    }
    QVariant parseArray() {
        QVariantList result;
        QChar c = next();
        DENG2_ASSERT(c == '[');
        if (peek() == ']') {
            next();
            return result;
        }
        forever {
            result << parse();
            c = next();
            if (c == ']') {
                break;
            }
            else if (c != ',') {
                error("array items must be separated by comma");
            }
        }
        return result;
    }
    QVariant parseString() {
        QVarLengthArray<QChar, 1024> result;
        QChar c = next();
        DENG2_ASSERT(c == '\"');
        forever {
            c = nextNoSkip();
            if (c == '\\') {
                c = nextNoSkip();
                if (c == '\"' || c == '\\' || c == '/')
                    result.append(c);
                else if (c == 'b')
                    result.append('\b');
                else if (c == 'f')
                    result.append('\f');
                else if (c == 'n')
                    result.append('\n');
                else if (c == 'r')
                    result.append('\r');
                else if (c == 't')
                    result.append('\t');
                else if (c == 'u') {
                    QString code = source.mid(pos, 4);
                    pos += 4;
                    result.append(QChar(ushort(code.toLong(0, 16))));
                }
                else error("unknown escape sequence in string");
            }
            else if (c == '\"') {
                break;
            }
            else {
                result.append(c);
            }
        }
        skipWhite();
        return QString(result.constData(), result.size());
    }
    QVariant parseNumber() {
        QVarLengthArray<QChar> str;
        QChar c = next();
        if (c == '-') {
            str.append(c);
            c = nextNoSkip();
        }
        for (; c.isDigit(); c = nextNoSkip()) {
            str.append(c);
        }
        bool hasDecimal = false;
        if (c == '.') {
            str.append(c);
            hasDecimal = true;
            c = nextNoSkip();
            for (; c.isDigit(); c = nextNoSkip()) {
                str.append(c);
            }
        }
        if (c == 'e' || c == 'E') {
            str.append(c);
            c = nextNoSkip();
            if (c == '+' || c == '-') {
                str.append(c);
                c = nextNoSkip();
            }
            for (; c.isDigit(); c = nextNoSkip()) {
                str.append(c);
            }
        }
        pos--;
        skipWhite();
        double value = QString(str.constData(), str.size()).toDouble();
        if (hasDecimal) {
            return QVariant(value);
        }
        else {
            return QVariant(int(value));
        }
    }
    QVariant parseKeyword() {
        if (source.mid(pos, 4) == "true") {
            pos += 4;
            skipWhite();
            return QVariant(true);
        }
        else if (source.mid(pos, 5) == "false") {
            pos += 5;
            skipWhite();
            return QVariant(false);
        }
        else if (source.mid(pos, 4) == "null") {
            pos += 4;
            skipWhite();
            return QVariant();
        }
        else {
            error("unknown keyword");
        }
        return QVariant();
    }
};
static Block recordToJSON(Record const &rec);
static Block valueToJSON(Value const &value);
static Block valueToJSONWithTabNewlines(Value const &value) {
    Block json = valueToJSON(value);
    json.replace('\n', "\n\t");
    return json;
}
static Block valueToJSON(Value const &value) {
    if (is<NoneValue>(value)) {
        return "null";
    }
    if (auto const *rec = maybeAs<RecordValue>(value)) {
        return recordToJSON(rec->dereference());
    }
    if (auto const *dict = maybeAs<DictionaryValue>(value)) {
        Block out = "{";
        auto const &elems = dict->elements();
        for (auto i = elems.begin(); i != elems.end(); ++i) {
            if (i != elems.begin()) {
                out += ",";
            }
            out += "\n\t" + valueToJSON(*i->first.value) + ": " +
                   valueToJSONWithTabNewlines(*i->second);
        }
        return out + "\n}";
    }
    if (auto const *array = maybeAs<ArrayValue>(value)) {
        Block out = "[";
        auto const &elems = array->elements();
        for (auto i = elems.begin(); i != elems.end(); ++i) {
            if (i != elems.begin()) {
                out += ",";
            }
            out += "\n\t" + valueToJSONWithTabNewlines(**i);
        }
        return out + "\n]";
    }
    if (auto const *num = maybeAs<NumberValue>(value)) {
        if (num->semanticHints().testFlag(NumberValue::Boolean)) {
            return num->isTrue()? "true" : "false";
        }
        return num->asText().toUtf8();
    }
    String text = value.asText();
    text.replace("\\", "\\\\")
        .replace("\"", "\\\"")
        .replace("\b", "\\b")
        .replace("\f", "\\f")
        .replace("\n", "\\n")
        .replace("\r", "\\r")
        .replace("\t", "\\t");
    return "\"" + text.toUtf8() + "\"";
}
static Block recordToJSON(Record const &rec) {
    Block out = "{\n\t\"__obj__\": \"Record\"";
    rec.forMembers([&out] (String const &name, Variable const &var) {
        out += ",\n\t\"" + name.toUtf8() + "\": " + valueToJSONWithTabNewlines(var.value());
        return LoopContinue;
    });
    return out + "\n}";
}
} // internal
QVariant parseJSON(String const &jsonText) {
    try {
        return internal::JSONParser(jsonText).parse();
    }
    catch (de::Error const &er) {
        LOG_WARNING(er.asText());
        return QVariant(); // invalid
    }
}
Block composeJSON(Record const &rec) {
    return internal::recordToJSON(rec) + "\n";
}
} // de
namespace de {
DENG2_PIMPL_NOREF(NativePointerValue)
, DENG2_OBSERVES(Deletable, Deletion) {
    Object *object = nullptr;
    Record const *memberScope = nullptr;
    ~Impl() {
        setObject(nullptr);
    }
    void setObject(Object *obj) {
        if (object) object->audienceForDeletion -= this;
        object = obj;
        if (object) object->audienceForDeletion += this;
    }
    void objectWasDeleted(Deletable *obj) {
        DENG2_UNUSED(obj);
        DENG2_ASSERT(object == obj);
        object = nullptr;
    }
};
NativePointerValue::NativePointerValue(Object *object, Record const *memberScope)
    : d(new Impl) {
    d->memberScope = memberScope;
    d->setObject(object);
}
NativePointerValue::Object *NativePointerValue::object() const {
    return d->object;
}
void NativePointerValue::setObject(Object *object) {
    d->object = object;
}
Value::Text NativePointerValue::typeId() const {
    return "Native";
}
Value *NativePointerValue::duplicate() const {
    return new NativePointerValue(d->object, d->memberScope);
}
Value::Text NativePointerValue::asText() const {
    return QString("(native object 0x%1)").arg(dintptr(d->object), 0, 16);
}
bool NativePointerValue::isTrue() const {
    return d->object != nullptr;
}
Record *NativePointerValue::memberScope() const {
    return const_cast<Record *>(d->memberScope);
}
void NativePointerValue::operator >> (Writer &to) const {
    to << SerialId(NONE);
}
void NativePointerValue::operator << (Reader &) {
    throw CannotSerializeError("NativePointerValue::operator <<",
                               "Cannot deserialize native object references");
}
} // namespace de
namespace de {
NoneValue::NoneValue()
{}
Value *NoneValue::duplicate() const {
    return new NoneValue();
}
Value::Text NoneValue::asText() const {
    return "(none)";
}
bool NoneValue::isTrue() const {
    return false;
}
dint NoneValue::compare(Value const &value) const {
    NoneValue const *other = dynamic_cast<NoneValue const *>(&value);
    if (other) {
        return 0;
    }
    return 1;
}
void NoneValue::operator >> (Writer &to) const {
    to << SerialId(NONE);
}
void NoneValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != NONE) {
        throw DeserializationError("NoneValue::operator <<", "Invalid ID");
    }
}
Value::Text NoneValue::typeId() const {
    return "None";
}
const NoneValue &NoneValue::none() {
    static NoneValue n;
    return n;
}
} // namespace de
namespace de {
NumberValue::NumberValue(Number initialValue, SemanticHints semantic)
    : _value(initialValue), _semantic(semantic)
{}
NumberValue::NumberValue(dint64 initialInteger)
    : _value(initialInteger), _semantic(Int)
{}
NumberValue::NumberValue(duint64 initialUnsignedInteger)
    : _value(initialUnsignedInteger), _semantic(UInt)
{}
NumberValue::NumberValue(dint32 initialInteger, SemanticHints semantic)
    : _value(initialInteger), _semantic(semantic)
{}
NumberValue::NumberValue(duint32 initialUnsignedInteger, SemanticHints semantic)
    : _value(initialUnsignedInteger), _semantic(semantic)
{}
NumberValue::NumberValue(unsigned long initialUnsignedInteger, SemanticHints semantic)
    : _value(initialUnsignedInteger), _semantic(semantic)
{}
NumberValue::NumberValue(bool initialBoolean)
    : _value(initialBoolean? True : False), _semantic(Boolean)
{}
void NumberValue::setSemanticHints(SemanticHints hints) {
    _semantic = hints;
}
Value *NumberValue::duplicate() const {
    return new NumberValue(_value, _semantic);
}
Value::Number NumberValue::asNumber() const {
    return _value;
}
Value::Text NumberValue::asText() const {
    Text result;
    QTextStream s(&result);
    if (_semantic.testFlag(Boolean) && (roundi(_value) == True || roundi(_value) == False)) {
        s << (isTrue()? "True" : "False");
    }
    else if (_semantic.testFlag(Hex)) {
        s << "0x" << QString::number(duint32(_value), 16);
    }
    else if (_semantic.testFlag(Int)) {
        s << QString::number(roundi(_value));
    }
    else if (_semantic.testFlag(UInt)) {
        s << QString::number(round<duint64>(_value));
    }
    else {
        s << _value;
    }
    return result;
}
bool NumberValue::isTrue() const {
    return roundi(_value) != 0;
}
dint NumberValue::compare(Value const &value) const {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&value);
    if (other) {
        if (fequal(_value, other->_value)) {
            return 0;
        }
        return cmp(_value, other->_value);
    }
    return Value::compare(value);
}
void NumberValue::negate() {
    _value = -_value;
}
void NumberValue::sum(Value const &value) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&value);
    if (!other) {
        throw ArithmeticError("NumberValue::sum", "Values cannot be summed");
    }
    _value += other->_value;
}
void NumberValue::subtract(Value const &value) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&value);
    if (!other) {
        throw ArithmeticError("Value::subtract", "Value cannot be subtracted from");
    }
    _value -= other->_value;
}
void NumberValue::divide(Value const &divisor) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&divisor);
    if (!other) {
        throw ArithmeticError("NumberValue::divide", "Value cannot be divided");
    }
    _value /= other->_value;
}
void NumberValue::multiply(Value const &value) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&value);
    if (!other) {
        throw ArithmeticError("NumberValue::multiply", "Value cannot be multiplied");
    }
    _value *= other->_value;
}
void NumberValue::modulo(Value const &divisor) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&divisor);
    if (!other) {
        throw ArithmeticError("Value::modulo", "Modulo not defined");
    }
    _value = int(_value) % int(other->_value);
}
static duint8 const SEMANTIC_BOOLEAN = 0x01;
static duint8 const SEMANTIC_HEX     = 0x02;
static duint8 const SEMANTIC_INT     = 0x04;
static duint8 const SEMANTIC_UINT    = 0x08;
void NumberValue::operator >> (Writer &to) const {
    duint8 flags = (_semantic.testFlag(Boolean)? SEMANTIC_BOOLEAN : 0) |
                   (_semantic.testFlag(Hex)?     SEMANTIC_HEX     : 0) |
                   (_semantic.testFlag(Int)?     SEMANTIC_INT     : 0) |
                   (_semantic.testFlag(UInt)?    SEMANTIC_UINT    : 0);
    to << SerialId(NUMBER) << flags << _value;
}
void NumberValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != NUMBER) {
        throw DeserializationError("NumberValue::operator <<", "Invalid ID");
    }
    duint8 flags;
    from >> flags >> _value;
    _semantic = SemanticHints((flags & SEMANTIC_BOOLEAN? Boolean : 0) |
                              (flags & SEMANTIC_HEX?     Hex : 0) |
                              (flags & SEMANTIC_INT?     Int : 0) |
                              (flags & SEMANTIC_UINT?    UInt : 0));
}
Value::Text NumberValue::typeId() const {
    return "Number";
}
NumberValue::SemanticHints NumberValue::semanticHints() const {
    return _semantic;
}
} // namespace de
namespace de {
IAudience::~IAudience()
{}
ObserverBase::ObserverBase()
{}
ObserverBase::~ObserverBase() {
    DENG2_GUARD(_memberOf);
    for (IAudience *observers : _memberOf.value) {
        observers->removeMember(this);
    }
}
void ObserverBase::addMemberOf(IAudience &observers) {
    DENG2_GUARD(_memberOf);
    _memberOf.value.insert(&observers);
}
void ObserverBase::removeMemberOf(IAudience &observers) {
    DENG2_GUARD(_memberOf);
    _memberOf.value.remove(&observers);
}
} // namespace de
namespace de {
static int const SEGMENT_BUFFER_SIZE = 8;
static String emptyPath;
Path::hash_type const Path::hash_range = 0xffffffff;
Path::hash_type Path::Segment::hash() const {
    if (!(flags & GotHashKey)) {
        String const lower = range.toString().toLower();
        hashKey = de::crc32(ByteRefArray(lower.constData(), lower.size() * sizeof(QChar)))
                % hash_range;
        flags |= GotHashKey;
    }
    return hashKey;
}
bool Path::Segment::hasWildCard() const {
    if (flags & WildCardChecked) {
        return flags.testFlag(IncludesWildCard);
    }
    bool isWild = toStringRef().contains(QChar('*'));
    applyFlagOperation(flags, IncludesWildCard, isWild? SetFlags : UnsetFlags);
    flags |= WildCardChecked;
    return isWild;
}
bool Path::Segment::operator == (Path::Segment const &other) const {
    return !range.compare(other.range, Qt::CaseInsensitive);
}
bool Path::Segment::operator < (Path::Segment const &other) const {
    return range.compare(other.range, Qt::CaseInsensitive) < 0;
}
int Path::Segment::length() const {
    return range.size();
}
dsize Path::Segment::size() const {
    return range.size();
}
Path::Segment::operator String () const {
    return toString();
}
String Path::Segment::toString() const {
    return range.string()->mid(range.position(), range.size());
}
DENG2_PIMPL_NOREF(Path) {
    String path;
    QChar separator;
    /**
     * Total number of segments in the path. If 0, it means that the path
     * isn't parsed into segments yet -- all paths have at least one segment
     * (an empty one if nothing else).
     */
    int segmentCount;
    /**
     * Fixed-size array for the segments of the path.
     *
     * The segments array is composed of two parts: the first
     * SEGMENT_BUFFER_SIZE elements are placed into a fixed-size array which is
     * allocated along with the Instance, and additional segments are allocated
     * dynamically and linked in the extraSegments list.
     *
     * This optimized representation should mean that the majority of paths
     * can be represented without dynamically allocating memory from the heap
     * (apart from the Instance, that is).
     *
     * @note Contents of the array are not initialized to zero.
     */
    Path::Segment segments[SEGMENT_BUFFER_SIZE];
    /**
     * List of the extra segments that don't fit in segments, in reverse
     * order.
     */
    QList<Path::Segment> extraSegments;
    Impl() : separator('/'), segmentCount(0)
    {}
    Impl(String const &p, QChar sep) : path(p), separator(sep), segmentCount(0)
    {}
    ~Impl() {
        clearSegments();
    }
    /**
     * Clear the segment array.
     *
     * @post The map will need to be rebuilt with parse().
     */
    void clearSegments() {
        extraSegments.clear();
        zap(segments);
        segmentCount = 0;
    }
    /**
     * Allocate a new segment, either from the fixed-size segments array if it
     * isn't full, or dynamically from the heap.
     *
     * @return  New segment.
     */
    Path::Segment *allocSegment(QStringRef const &range) {
        Path::Segment *segment;
        if (segmentCount < SEGMENT_BUFFER_SIZE) {
            segment = segments + segmentCount;
        }
        else {
            extraSegments.append(Path::Segment());
            segment = &extraSegments.last();
        }
        zapPtr(segment);
        segment->range = range;
        segmentCount++;
        return segment;
    }
    /**
     * Build the segment array by parsing the path. when the path is modified,
     * the existing map is invalidated and needs to be remapped.
     */
    void parse() {
        if (segmentCount > 0) return;
        segmentCount = 0;
        extraSegments.clear();
        if (path.isEmpty()) {
            allocSegment(&emptyPath);
            DENG2_ASSERT(segmentCount > 0);
            return;
        }
        QChar const *segBegin = path.constData();
        QChar const *segEnd   = path.constData() + path.length() - 1;
        for (int i = path.length();
            segEnd->unicode() && *segEnd == separator && i-- > 0;
            --segEnd) {}
        QChar const *from;
        forever {
            if (segEnd < segBegin) break; // E.g., path is "/"
            for (from = segEnd; from > segBegin && !(*from == separator); from--)
            {}
            int startIndex = (*from == separator? from + 1 : from) - path.constData();
            int length = (segEnd - path.constData()) - startIndex + 1;
            allocSegment(QStringRef(&path, startIndex, length));
            if (from == segBegin) break;
            segEnd = from - 1;
        }
        if (*segBegin == separator) {
            allocSegment(&emptyPath);
        }
        DENG2_ASSERT(segmentCount > 0);
    }
};
Path::Path() : d(new Impl)
{}
Path::Path(String const &path, QChar sep)
    : d(new Impl(path, sep))
{}
Path::Path(QString const &str)
    : d(new Impl(str, '/'))
{}
Path::Path(char const *nullTerminatedCStr, char sep)
    : d(new Impl(QString::fromUtf8(nullTerminatedCStr), sep))
{}
Path::Path(char const *nullTerminatedCStr)
    : d(new Impl(QString::fromUtf8(nullTerminatedCStr), '/'))
{}
Path::Path(Path const &other)
    : d(new Impl(other.d->path, other.d->separator))
{}
Path::Path(Path &&moved)
    : d(std::move(moved.d))
{}
Path &Path::operator = (char const *pathUtf8) {
    return *this = Path(pathUtf8);
}
Path &Path::operator = (Path const &other) {
    d.reset(new Impl(other.d->path, other.d->separator));
    return *this;
}
Path &Path::operator = (Path &&moved) {
    d = std::move(moved.d);
    return *this;
}
Path Path::operator + (QString const &str) const {
    return Path(d->path + str, d->separator);
}
Path Path::operator + (char const *nullTerminatedCStr) const {
    return Path(d->path + QString(nullTerminatedCStr), d->separator);
}
int Path::segmentCount() const {
    d->parse();
    return d->segmentCount;
}
Path::Segment const &Path::segment(int index) const {
    return reverseSegment(segmentCount() - 1 - index);
}
Path::Segment const &Path::reverseSegment(int reverseIndex) const {
    d->parse();
    if (reverseIndex < 0 || reverseIndex >= d->segmentCount) {
        throw OutOfBoundsError("Path::reverseSegment", String("Reverse index %1 is out of bounds").arg(reverseIndex));
    }
    if (reverseIndex < SEGMENT_BUFFER_SIZE) {
        return d->segments[reverseIndex];
    }
    return d->extraSegments[reverseIndex - SEGMENT_BUFFER_SIZE];
}
Path Path::subPath(Rangei const &range) const {
    if (range.isEmpty()) {
        return Path("", d->separator);
    }
    Path sub(segment(range.start), d->separator);
    for (int i = range.start + 1; i < range.end; ++i) {
        sub = sub / segment(i);
    }
    return sub;
}
bool Path::operator == (Path const &other) const {
    if (this == &other) return true;
    if (segmentCount() != other.segmentCount()) return false;
    for (int i = 0; i < d->segmentCount; ++i) {
        if (segment(i).hash() != other.segment(i).hash())
            return false;
    }
    if (d->separator == other.d->separator) {
        return !d->path.compareWithoutCase(other.d->path);
    }
    else {
        for (int i = 0; i < d->segmentCount; ++i) {
            if (segment(i) != other.segment(i)) return false;
        }
        return true;
    }
}
bool Path::operator < (Path const &other) const {
    if (d->separator == other.d->separator) {
        return d->path.compareWithoutCase(other.d->path) < 0;
    }
    else {
        for (int i = 0; i < d->segmentCount; ++i) {
            if (!(segment(i) < other.segment(i))) return false;
        }
        return true;
    }
}
Path Path::operator / (Path const &other) const {
    String otherPath = other.d->path;
    if (other.separator() != d->separator) {
        otherPath.replace(other.d->separator, d->separator);
    }
    return Path(d->path.concatenatePath(otherPath, d->separator), d->separator);
}
Path Path::operator / (QString other) const {
    return *this / Path(other);
}
Path Path::operator / (char const *otherNullTerminatedUtf8) const {
    return *this / Path(otherNullTerminatedUtf8, '/');
}
Path Path::operator / (String const &other) const {
    return *this / Path(other);
}
String Path::toString() const {
    return d->path;
}
String const &Path::toStringRef() const {
    return d->path;
}
bool Path::isEmpty() const {
    return d->path.isEmpty();
}
bool Path::isAbsolute() const {
    return !isEmpty() && !firstSegment().size();
}
int Path::length() const {
    return d->path.length();
}
dsize Path::size() const {
    return length();
}
QChar Path::first() const {
    return d->path.first();
}
QChar Path::last() const {
    return d->path.last();
}
Path &Path::clear() {
    d->path.clear();
    d->clearSegments();
    return *this;
}
Path &Path::operator = (String const &newPath) {
    set(newPath, '/');
    return *this;
}
Path &Path::set(String const &newPath, QChar sep) {
    d->path = newPath; // implicitly shared
    d->separator = sep;
    d->clearSegments();
    return *this;
}
Path Path::withSeparators(QChar sep) const {
    if (sep == d->separator) return *this;
    String modPath = d->path;
    modPath.replace(d->separator, sep);
    return Path(modPath, sep);
}
QChar Path::separator() const {
    return d->separator;
}
void Path::addTerminatingSeparator() {
    if (!isEmpty()) {
        if (last() != d->separator) {
            d->path.append(d->separator);
            d->clearSegments();
        }
    }
}
String Path::fileName() const {
    if (last() == d->separator) return "";
    return lastSegment();
}
Block Path::toUtf8() const {
    return d->path.toUtf8();
}
void Path::operator >> (Writer &to) const {
    to << d->path.toUtf8() << d->separator.unicode();
}
void Path::operator << (Reader &from) {
    clear();
    Block b;
    ushort sep;
    from >> b >> sep;
    set(String::fromUtf8(b), sep);
}
String Path::normalizeString(String const &text, QChar replaceWith) {
    String result = text;
    if (replaceWith != '/') {
        result.replace('/', replaceWith);
    }
    if (replaceWith != '\\') {
        result.replace('\\', replaceWith);
    }
    return result;
}
Path Path::normalize(String const &text, QChar replaceWith) {
    return Path(normalizeString(text, replaceWith), replaceWith);
}
Path PathRef::toPath() const {
    if (!segmentCount()) return Path(); // Empty.
    String composed = segment(0);
    for (int i = 1; i < segmentCount(); ++i) {
        composed += path().separator();
        composed += segment(i);
    }
    return Path(composed, path().separator());
}
} // namespace de
#ifdef _DEBUG
#include <QDebug>
using namespace de;
static int Path_UnitTest() {
    try { {
            Path p;
            DENG2_ASSERT(p == Path(""));
            DENG2_ASSERT(p.isEmpty());
            DENG2_ASSERT(p.segmentCount() == 1);
        } {
            Path a("some/thing");
            Path b("/other/thing");
            DENG2_ASSERT(a != b);
            Path c = a;
            DENG2_ASSERT(c == a);
            DENG2_ASSERT(c.segment(1).toString() == "thing");
            DENG2_ASSERT(c.reverseSegment(1).toString() == "some");
            b = a;
            DENG2_ASSERT(b == a);
            DENG2_ASSERT(b.segment(1).toString() == "thing");
            DENG2_ASSERT(b.reverseSegment(1).toString() == "some");
        } {
            Path a("a/b/c");
            Path b("d/e");
            DENG2_ASSERT(a.segmentCount() == 3);
            DENG2_ASSERT(a.reverseSegment(1).toString() == "b");
            std::swap(a, b);
            DENG2_ASSERT(a.segmentCount() == 2);
            DENG2_ASSERT(a.reverseSegment(1).toString() == "d");
            DENG2_ASSERT(b.segmentCount() == 3);
            DENG2_ASSERT(b.reverseSegment(1).toString() == "b");
        } {
            Path p("c:/something.ext");
            DENG2_ASSERT(p.segmentCount() == 2);
            DENG2_ASSERT(p.reverseSegment(0).length() == 13);
            DENG2_ASSERT(p.reverseSegment(0).toString() == "something.ext");
            DENG2_ASSERT(p.reverseSegment(1).length() == 2);
            DENG2_ASSERT(p.reverseSegment(1).toString() == "c:");
        } {
            Path p("/something.ext");
            DENG2_ASSERT(p.segmentCount() == 2);
            DENG2_ASSERT(p.reverseSegment(0).length() == 13);
            DENG2_ASSERT(p.reverseSegment(0).toString() == "something.ext");
            DENG2_ASSERT(p.reverseSegment(1).length() == 0);
            DENG2_ASSERT(p.reverseSegment(1).toString() == "");
        } {
            Path p("some/dir/structure/");
            DENG2_ASSERT(p.segmentCount() == 3);
            DENG2_ASSERT(p.reverseSegment(0).length() == 9);
            DENG2_ASSERT(p.reverseSegment(0).toString() == "structure");
            DENG2_ASSERT(p.reverseSegment(1).length() == 3);
            DENG2_ASSERT(p.reverseSegment(1).toString() == "dir");
            DENG2_ASSERT(p.reverseSegment(2).length() == 4);
            DENG2_ASSERT(p.reverseSegment(2).toString() == "some");
        } {
            Path p("/30/29/28/27/26/25/24/23/22/21/20/19/18/17/16/15/14/13/12/11/10/9/8/7/6/5/4/3/2/1/0");
            DENG2_ASSERT(p.segmentCount() == 32);
            DENG2_ASSERT(p.reverseSegment(0).toString()  == "0");
            DENG2_ASSERT(p.reverseSegment(23).toString() == "23");
            DENG2_ASSERT(p.reverseSegment(24).toString() == "24");
            DENG2_ASSERT(p.reverseSegment(30).toString() == "30");
        } {
            Path a("a.b.c.d", '.');
            Path b("con-variable", '-');
            DENG2_ASSERT(a.segmentCount() == 4);
            DENG2_ASSERT(a.segment(1).toString() == "b");
            DENG2_ASSERT(b.segmentCount() == 2);
            DENG2_ASSERT(b.segment(0).toString() == "con");
            DENG2_ASSERT(b.segment(1).toString() == "variable");
        } {
            Path p;
            Path a("hello");
            Path b("hello/world");
            Path c("hello/world/");
            Path d("hello/world/  ");
            /*
            qDebug() << p << "=>" << p.fileName();
            qDebug() << a << "=>" << a.fileName();
            qDebug() << b << "=>" << b.fileName();
            qDebug() << c << "=>" << c.fileName();
            qDebug() << d << "=>" << d.fileName();
            */
            DENG2_ASSERT(p.fileName() == String(p).fileName());
            DENG2_ASSERT(a.fileName() == String(a).fileName());
            DENG2_ASSERT(b.fileName() == String(b).fileName());
            DENG2_ASSERT(c.fileName() == String(c).fileName());
            DENG2_ASSERT(d.fileName() == String(d).fileName());
        }
    }
    catch (Error const &er) {
        qWarning() << er.asText();
        return false;
    }
    return true;
}
static int testResult = Path_UnitTest();
#endif // _DEBUG
namespace de {
Path::hash_type const PathTree::no_hash = Path::hash_range;
struct PathTree::Impl {
    PathTree &self;
    StringPool segments;
    PathTree::Flags flags;
    int size;
    int numNodesOwned;
    PathTree::Node rootNode;
    PathTree::NodeHash hash;
    Impl(PathTree &d, int _flags)
        : self(d), flags(_flags), size(0), numNodesOwned(0),
          rootNode(PathTree::NodeArgs(d, PathTree::Branch, 0))
    {}
    ~Impl() {
        clear();
    }
    void clear() {
        clearPathHash(hash.leaves);
        clearPathHash(hash.branches);
        size = 0;
        DENG2_ASSERT(numNodesOwned == 0);
    }
    PathTree::SegmentId internSegmentAndUpdateIdHashMap(String segment, Path::hash_type hashKey) {
        PathTree::SegmentId internId = segments.intern(segment);
        segments.setUserValue(internId, hashKey);
        return internId;
    }
    /**
     * @return Tree node that matches the name and type and which has the
     * specified parent node.
     */
    PathTree::Node *nodeForSegment(Path::Segment const &segment, PathTree::NodeType nodeType,
                                   PathTree::Node *parent) {
        PathTree::Nodes const &hash = self.nodes(nodeType);
        PathTree::SegmentId segmentId = segments.isInterned(segment);
        if (segmentId) {
            Path::hash_type hashKey = segments.userValue(segmentId);
            for (PathTree::Nodes::const_iterator i = hash.find(hashKey);
                i != hash.end() && i.key() == hashKey; ++i) {
                PathTree::Node *node = *i;
                if (parent    != &node->parent()) continue;
                if (segmentId != node->segmentId()) continue;
                if (nodeType == PathTree::Branch || !(flags & PathTree::MultiLeaf))
                    return node;
            }
        }
        /*
         * A new node is needed.
         */
        Path::hash_type hashKey;
        if (!segmentId) {
            hashKey   = segment.hash();
            segmentId = internSegmentAndUpdateIdHashMap(segment, hashKey);
        }
        else {
            hashKey = self.segmentHash(segmentId);
        }
        PathTree::Node *node = self.newNode(PathTree::NodeArgs(self, nodeType, segmentId, parent));
        const_cast<Nodes &>(hash).insert(hashKey, node);
        numNodesOwned++;
        return node;
    }
    /**
     * The path is split into as many nodes as necessary. Parent links are set.
     *
     * @return  The node that identifies the given path.
     */
    PathTree::Node *buildNodesForPath(Path const &path) {
        bool const hasLeaf = !path.toStringRef().endsWith(QStringLiteral("/"));
        PathTree::Node *node = 0, *parent = &rootNode;
        for (int i = 0; i < path.segmentCount() - (hasLeaf? 1 : 0); ++i) {
            Path::Segment const &pn = path.segment(i);
            node = nodeForSegment(pn, PathTree::Branch, parent);
            parent = node;
        }
        if (hasLeaf) {
            Path::Segment const &pn = path.lastSegment();
            node = nodeForSegment(pn, PathTree::Leaf, parent);
        }
        return node;
    }
    PathTree::Node *findInHash(PathTree::Nodes &hash, Path::hash_type hashKey,
                               Path const &searchPath,
                               PathTree::ComparisonFlags compFlags) {
        for (Nodes::iterator i = hash.find(hashKey);
            i != hash.end() && i.key() == hashKey; ++i) {
            PathTree::Node *node = *i;
            if (!node->comparePath(searchPath, compFlags)) {
                if (compFlags.testFlag(RelinquishMatching)) {
                    node->parent().removeChild(*node);
                    hash.erase(i);
                    numNodesOwned--;
                    DENG2_ASSERT(numNodesOwned >= 0);
                }
                return node;
            }
        }
        return 0;
    }
    PathTree::Node *find(Path const &searchPath, PathTree::ComparisonFlags compFlags) {
        if (searchPath.isEmpty() && !compFlags.testFlag(NoBranch)) {
            return &rootNode;
        }
        PathTree::Node *found = 0;
        if (size) {
            Path::hash_type hashKey = searchPath.lastSegment().hash();
            if (!compFlags.testFlag(NoLeaf)) {
                if ((found = findInHash(hash.leaves, hashKey, searchPath, compFlags)) != 0)
                    return found;
            }
            if (!compFlags.testFlag(NoBranch)) {
                if ((found = findInHash(hash.branches, hashKey, searchPath, compFlags)) != 0)
                    return found;
            }
        }
        return 0;
    }
    void clearPathHash(PathTree::Nodes &ph) {
        DENG2_FOR_EACH(PathTree::Nodes, i, ph) {
            PathTree::Node *node = *i;
            delete node;
            numNodesOwned--;
            DENG2_ASSERT(numNodesOwned >= 0);
        }
        ph.clear();
    }
};
PathTree::PathTree(Flags flags) {
    d = new Impl(*this, flags);
}
PathTree::~PathTree() {
    DENG2_GUARD(this);
    delete d;
}
PathTree::Node &PathTree::insert(Path const &path) {
    DENG2_GUARD(this);
    PathTree::Node *node = d->buildNodesForPath(path);
    DENG2_ASSERT(node != 0);
    d->size++;
    return *node;
}
bool PathTree::remove(Path const &path, ComparisonFlags flags) {
    DENG2_GUARD(this);
    PathTree::Node *node = d->find(path, flags | RelinquishMatching);
    if (node && node != &d->rootNode) {
        d->size--;
        delete node;
        return true;
    }
    return false;
}
String const &PathTree::nodeTypeName(NodeType type) {
    static String const nodeNames[] = {
        "branch",
        "leaf"
    };
    return nodeNames[type == Branch? 0 : 1];
}
int PathTree::size() const {
    DENG2_GUARD(this);
    return d->size;
}
bool PathTree::empty() const {
    return size() == 0;
}
PathTree::Flags PathTree::flags() const {
    DENG2_GUARD(this);
    return d->flags;
}
void PathTree::clear() {
    DENG2_GUARD(this);
    d->clear();
}
bool PathTree::has(Path const &path, ComparisonFlags flags) const {
    DENG2_GUARD(this);
    flags &= ~RelinquishMatching; // never relinquish
    return d->find(path, flags) != 0;
}
PathTree::Node const &PathTree::find(Path const &searchPath, ComparisonFlags flags) const {
    DENG2_GUARD(this);
    Node const *found = d->find(searchPath, flags);
    if (!found) {
        throw NotFoundError("PathTree::find", "No paths found matching \"" + searchPath + "\"");
    }
    return *found;
}
PathTree::Node const *PathTree::tryFind(Path const &path, ComparisonFlags flags) const {
    DENG2_GUARD(this);
    return d->find(path, flags);
}
PathTree::Node &PathTree::find(Path const &path, ComparisonFlags flags) {
    Node const &node = const_cast<PathTree const *>(this)->find(path, flags);
    return const_cast<Node &>(node);
}
PathTree::Node *PathTree::tryFind(Path const &path, ComparisonFlags flags) {
    DENG2_GUARD(this);
    return d->find(path, flags);
}
String const &PathTree::segmentName(SegmentId segmentId) const {
    DENG2_GUARD(this);
    return d->segments.stringRef(segmentId);
}
Path::hash_type PathTree::segmentHash(SegmentId segmentId) const {
    DENG2_GUARD(this);
    return d->segments.userValue(segmentId);
}
PathTree::Node const &PathTree::rootBranch() const {
    return d->rootNode;
}
PathTree::Node *PathTree::newNode(NodeArgs const &args) {
    return new Node(args);
}
PathTree::Nodes const &PathTree::nodes(NodeType type) const {
    DENG2_GUARD(this);
    return (type == Leaf? d->hash.leaves : d->hash.branches);
}
static void collectPathsInHash(PathTree::FoundPaths &found, PathTree::Nodes const &ph, QChar separator) {
    if (ph.empty()) return;
    DENG2_FOR_EACH_CONST(PathTree::Nodes, i, ph) {
        PathTree::Node const &node = **i;
        found.push_back(node.path(separator));
    }
}
int PathTree::findAllPaths(FoundPaths &found, ComparisonFlags flags, QChar separator) const {
    DENG2_GUARD(this);
    int numFoundSoFar = found.count();
    if (!(flags & NoBranch)) {
        collectPathsInHash(found, d->hash.branches, separator);
    }
    if (!(flags & NoLeaf)) {
        collectPathsInHash(found, d->hash.leaves, separator);
    }
    return found.count() - numFoundSoFar;
}
static int iteratePathsInHash(PathTree const &pathTree, Path::hash_type hashKey,
                              PathTree::NodeType type, PathTree::ComparisonFlags flags,
                              PathTree::Node const *parent,
                              int (*callback) (PathTree::Node &, void *), void *parameters) {
    int result = 0;
    if (hashKey != PathTree::no_hash && hashKey >= Path::hash_range) {
        throw Error("PathTree::iteratePathsInHash", String("Invalid hash %1, valid range is [0..%2).").arg(hashKey).arg(Path::hash_range-1));
    }
    PathTree::Nodes const *nodes = &pathTree.nodes(type);
    if (flags.testFlag(PathTree::MatchParent) && parent) {
        nodes = &parent->childNodes(type);
    }
    if (hashKey != PathTree::no_hash) {
        PathTree::Nodes::const_iterator i = nodes->constFind(hashKey);
        for (; i != nodes->end() && i.key() == hashKey; ++i) {
            if (!(flags.testFlag(PathTree::MatchParent) && parent != &(*i)->parent())) {
                result = callback(**i, parameters);
                if (result) break;
            }
        }
    }
    else {
        DENG2_FOR_EACH_CONST(PathTree::Nodes, i, *nodes) {
            if (!(flags.testFlag(PathTree::MatchParent) && parent != &(*i)->parent())) {
                result = callback(**i, parameters);
                if (result) break;
            }
        }
    }
    return result;
}
int PathTree::traverse(ComparisonFlags flags, PathTree::Node const *parent, Path::hash_type hashKey,
                       int (*callback) (PathTree::Node &, void *), void *parameters) const {
    DENG2_GUARD(this);
    int result = 0;
    if (callback) {
        if (!(flags & NoLeaf))
            result = iteratePathsInHash(*this, hashKey, Leaf, flags, parent, callback, parameters);
        if (!result && !(flags & NoBranch))
            result = iteratePathsInHash(*this, hashKey, Branch, flags, parent, callback, parameters);
    }
    return result;
}
#ifdef DENG2_DEBUG
void PathTree::debugPrint(QChar separator) const {
    LOGDEV_MSG("PathTree [%p]:") << de::dintptr(this);
    FoundPaths found;
    if (findAllPaths(found, 0, separator)) {
        qSort(found.begin(), found.end());
        DENG2_FOR_EACH_CONST(FoundPaths, i, found) {
            LOGDEV_MSG("  %s") << *i;
        }
    }
    LOGDEV_MSG("  %i unique %s in the tree.") << found.count() << (found.count() == 1? "path" : "paths");
}
#if 0
static void printDistributionOverviewElement(int const *colWidths, char const *name,
    size_t numEmpty, size_t maxHeight, size_t numCollisions, size_t maxCollisions,
    size_t sum, size_t total) {
    DENG2_ASSERT(colWidths);
    float coverage, collision, variance;
    if (0 != total) {
        size_t sumSqr = sum*sum;
        float mean = (signed)sum / total;
        variance = ((signed)sumSqr - (signed)sum * mean) / (((signed)total)-1);
        coverage  = 100 / (float)Path::hash_range * (Path::hash_range - numEmpty);
        collision = 100 / (float) total * numCollisions;
    }
    else {
        variance = coverage = collision = 0;
    }
    int const *col = colWidths;
    Con_Printf("%*s ",    *col++, name);
    Con_Printf("%*lu ",   *col++, (unsigned long)total);
    Con_Printf("%*lu",    *col++, Path::hash_range - (unsigned long)numEmpty);
    Con_Printf(":%-*lu ", *col++, (unsigned long)numEmpty);
    Con_Printf("%*lu ",   *col++, (unsigned long)maxCollisions);
    Con_Printf("%*lu ",   *col++, (unsigned long)numCollisions);
    Con_Printf("%*.2f ",  *col++, collision);
    Con_Printf("%*.2f ",  *col++, coverage);
    Con_Printf("%*.2f ",  *col++, variance);
    Con_Printf("%*lu\n",  *col++, (unsigned long)maxHeight);
}
static void printDistributionOverview(PathTree *pt,
    size_t nodeCountSum[PATHTREENODE_TYPE_COUNT],
    size_t nodeCountTotal[PATHTREENODE_TYPE_COUNT],
    size_t nodeBucketCollisions[PATHTREENODE_TYPE_COUNT], size_t nodeBucketCollisionsTotal,
    size_t nodeBucketCollisionsMax[PATHTREENODE_TYPE_COUNT], size_t /*nodeBucketCollisionsMaxTotal*/,
    size_t nodeBucketEmpty[PATHTREENODE_TYPE_COUNT], size_t nodeBucketEmptyTotal, size_t nodeBucketHeight,
    size_t /*nodeCount*/[PATHTREENODE_TYPE_COUNT])
{
#define NUMCOLS             10/*type+count+used:+empty+collideMax+collideCount+collidePercent+coverage+variance+maxheight*/
    DENG2_ASSERT(pt);
    size_t collisionsMax = 0, countSum = 0, countTotal = 0;
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
        if (nodeBucketCollisionsMax[i] > collisionsMax)
            collisionsMax = nodeBucketCollisionsMax[i];
        countSum += nodeCountSum[i];
        countTotal += nodeCountTotal[i];
    }
    int nodeCountDigits = M_NumDigits((int)countTotal);
    int colWidths[NUMCOLS];
    int *col = colWidths;
    *col = 0;
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
        PathTree::NodeType type = PathTree::NodeType(i);
        if (Str_Length(PathTreeNode::typeName(type)) > *col)
            *col = Str_Length(PathTreeNode::typeName(type));
    }
    col++;
    *col++ = MAX_OF(nodeCountDigits, 1); /*#*/
    *col++ = MAX_OF(nodeCountDigits, 4); /*used*/
    *col++ = MAX_OF(nodeCountDigits, 5); /*empty*/
    *col++ = MAX_OF(nodeCountDigits, 3); /*max*/
    *col++ = MAX_OF(nodeCountDigits, 4); /*num*/
    *col++ = MAX_OF(3+1+2, 8);           /*percent*/
    *col++ = MAX_OF(3+1+2, 9);           /*coverage*/
    *col++ = MAX_OF(nodeCountDigits, 8); /*variance*/
    *col   = MAX_OF(nodeCountDigits, 9); /*maxheight*/
    int spans[4][2];
    spans[0][0] = colWidths[0] + 1/* */ + colWidths[1];
    spans[1][0] = colWidths[2] + 1/*:*/ + colWidths[3];
    spans[2][0] = colWidths[4] + 1/* */ + colWidths[5] + 1/* */ + colWidths[6];
    spans[3][0] = colWidths[7] + 1/* */ + colWidths[8] + 1/* */ + colWidths[9];
    for (int i = 0; i < 4; ++i) {
        int remainder = spans[i][0] % 2;
        spans[i][1] = remainder + (spans[i][0] /= 2);
    }
    Con_FPrintf(CPF_YELLOW, "Directory Distribution (p:%p):\n", pt);
    int *span = &spans[0][0];
    Con_Printf("%*s", *span++ +  5/2, "nodes");         Con_Printf("%-*s|", *span++ -  5/2, "");
    Con_Printf("%*s", *span++ +  4/2, "hash");          Con_Printf("%-*s|", *span++ -  4/2, "");
    Con_Printf("%*s", *span++ + 10/2, "collisions");    Con_Printf("%-*s|", *span++ - 10/2, "");
    Con_Printf("%*s", *span++ +  5/2, "other");         Con_Printf("%-*s\n",*span++ -  5/2, "");
    col = colWidths;
    Con_FPrintf(CPF_LIGHT, "%*s ",   *col++, "type");
    Con_FPrintf(CPF_LIGHT, "%-*s|",  *col++, "#");
    Con_FPrintf(CPF_LIGHT, "%*s:",   *col++, "used");
    Con_FPrintf(CPF_LIGHT, "%-*s|",  *col++, "empty");
    Con_FPrintf(CPF_LIGHT, "%*s ",   *col++, "max");
    Con_FPrintf(CPF_LIGHT, "%*s ",   *col++, "num#");
    Con_FPrintf(CPF_LIGHT, "%-*s|",  *col++, "percent%");
    Con_FPrintf(CPF_LIGHT, "%*s ",   *col++, "coverage%");
    Con_FPrintf(CPF_LIGHT, "%*s ",   *col++, "variance");
    Con_FPrintf(CPF_LIGHT, "%-*s\n", *col++, "maxheight");
    if (countTotal != 0) {
        for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
            PathTree::NodeType type = PathTree::NodeType(i);
            printDistributionOverviewElement(colWidths, Str_Text(PathTreeNode::typeName(type)),
                nodeBucketEmpty[i], (i == PT_LEAF? nodeBucketHeight : 0),
                nodeBucketCollisions[i], nodeBucketCollisionsMax[i],
                nodeCountSum[i], nodeCountTotal[i]);
        }
        Con_PrintRuler();
    }
    printDistributionOverviewElement(colWidths, "total",
        nodeBucketEmptyTotal, nodeBucketHeight,
        nodeBucketCollisionsTotal, collisionsMax,
        countSum / PATHTREENODE_TYPE_COUNT, countTotal);
#undef NUMCOLS
}
#endif
#if 0
static void printDistributionHistogram(PathTree *pt, ushort size,
    size_t nodeCountTotal[PATHTREENODE_TYPE_COUNT]) {
#define NUMCOLS             4/*range+total+PATHTREENODE_TYPE_COUNT*/
    size_t totalForRange, total, nodeCount[PATHTREENODE_TYPE_COUNT];
    int hashIndexDigits, col, colWidths[2+/*range+total*/PATHTREENODE_TYPE_COUNT];
    PathTreeNode *node;
    int j;
    DENG2_ASSERT(pt);
    total = 0;
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
        total += nodeCountTotal[i];
    }
    if (0 == total) return;
    hashIndexDigits = M_NumDigits(Path::hash_range);
    col = 0;
    if (size != 0)
        colWidths[col] = 2/*braces*/+hashIndexDigits*2+3/*elipses*/;
    else
        colWidths[col] = 2/*braces*/+hashIndexDigits;
    colWidths[col] = MAX_OF(colWidths[col], 5/*range*/);
    ++col;
    { size_t max = 0;
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
        if (nodeCountTotal[i] > max)
            max = nodeCountTotal[i];
    }
    colWidths[col++] = MAX_OF(M_NumDigits((int)max), 5/*total*/);
    }
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i, ++col) {
        PathTree::NodeType type = PathTree::NodeType(i);
        colWidths[col] = Str_Length(PathTreeNode::typeName(type));
    }
    for (int i = 1; i < NUMCOLS; ++i) { colWidths[i] += 1; }
    Con_FPrintf(CPF_YELLOW, "Histogram (p:%p):\n", pt);
    col = 0;
    Con_Printf("%*s", colWidths[col++], "range");
    Con_Printf("%*s", colWidths[col++], "total");
    for (int i = 0; i < PATHTREENODE_TYPE_COUNT; ++i) {
        PathTree::NodeType type = PathTree::NodeType(i);
        Con_Printf("%*s", colWidths[col++], Str_Text(PathTreeNode::typeName(type)));
    }
    Con_Printf("\n");
    Con_PrintRuler();
    { Path::hash_type from = 0, n = 0, range = (size != 0? Path::hash_range / size: 0);
    memset(nodeCount, 0, sizeof(nodeCount));
    for (Path::hash_type i = 0; i < Path::hash_range; ++i) {
        pathtree_pathhash_t **phAdr;
        phAdr = hashAddressForNodeType(pt, PathTree::Node::Branch);
        if (*phAdr)
        for (node = (**phAdr)[i].head; node; node = node->next)
            ++nodeCount[PathTree::Node::Branch];
        phAdr = hashAddressForNodeType(pt, PT_LEAF);
        if (*phAdr)
        for (node = (**phAdr)[i].head; node; node = node->next)
            ++nodeCount[PT_LEAF];
        if (size != 0 && (++n != range && i != Path::hash_range-1))
            continue;
        totalForRange = 0;
        for (j = 0; j < PATHTREENODE_TYPE_COUNT; ++j)
            totalForRange += nodeCount[j];
        col = 0;
        if (size != 0) {
            Str range; Str_Init(&range);
            Str_Appendf(&range, "%*u...%*u", hashIndexDigits, from, hashIndexDigits, from+n-1);
            Con_Printf("[%*s]", colWidths[col++]-2/*braces*/, Str_Text(&range));
            Str_Free(&range);
        }
        else {
            Con_Printf("[%*u]", colWidths[col++]-2/*braces*/, i);
        }
        Con_Printf("%*lu", colWidths[col++], (unsigned long) totalForRange);
        if (0 != totalForRange) {
            for (j = 0; j < PATHTREENODE_TYPE_COUNT; ++j, ++col) {
                if (0 != nodeCount[j]) {
                    Con_Printf("%*lu", colWidths[col], (unsigned long) nodeCount[j]);
                }
                else if (j < PATHTREENODE_TYPE_COUNT-1 || 0 == size) {
                    Con_Printf("%*s", colWidths[col], "");
                }
            }
        }
        if (0 != totalForRange) {
            size_t max = MAX_OF(1, ROUND(total/(float)size/10));
            size_t scale = totalForRange / (float)max;
            scale = MAX_OF(scale, 1);
            Con_Printf(" ");
            for (n = 0; n < scale; ++n)
                Con_Printf("*");
        }
        Con_Printf("\n");
        from = i+1;
        n = 0;
        memset(nodeCount, 0, sizeof(nodeCount));
    }}
    Con_PrintRuler();
    col = 0;
    Con_Printf("%*s",  colWidths[col++], "Sum");
    Con_Printf("%*lu", colWidths[col++], (unsigned long) total);
    if (0 != total) {
        int i;
        for (i = 0; i < PATHTREENODE_TYPE_COUNT; ++i, ++col) {
            if (0 != nodeCountTotal[i]) {
                Con_Printf("%*lu", colWidths[col], (unsigned long) nodeCountTotal[i]);
            }
            else if (i < PATHTREENODE_TYPE_COUNT-1) {
                Con_Printf("%*s", colWidths[col], "");
            }
        }
    }
    Con_Printf("\n");
#undef NUMCOLS
}
#endif
void PathTree::debugPrintHashDistribution() const {
#if 0
    size_t nodeCountSum[PATHTREENODE_TYPE_COUNT],
           nodeCountTotal[PATHTREENODE_TYPE_COUNT], nodeBucketHeight = 0,
           nodeBucketCollisions[PATHTREENODE_TYPE_COUNT], nodeBucketCollisionsTotal = 0,
           nodeBucketCollisionsMax[PATHTREENODE_TYPE_COUNT], nodeBucketCollisionsMaxTotal = 0,
           nodeBucketEmpty[PATHTREENODE_TYPE_COUNT], nodeBucketEmptyTotal = 0,
           nodeCount[PATHTREENODE_TYPE_COUNT];
    size_t totalForRange;
    PathTreeNode *node;
    DENG2_ASSERT(pt);
    nodeCountTotal[PathTree::Node::Branch] = countNodesInPathHash(*hashAddressForNodeType(pt, PathTree::Node::Branch));
    nodeCountTotal[PT_LEAF]   = countNodesInPathHash(*hashAddressForNodeType(pt, PT_LEAF));
    memset(nodeCountSum, 0, sizeof(nodeCountSum));
    memset(nodeBucketCollisions, 0, sizeof(nodeBucketCollisions));
    memset(nodeBucketCollisionsMax, 0, sizeof(nodeBucketCollisionsMax));
    memset(nodeBucketEmpty, 0, sizeof(nodeBucketEmpty));
    for (Path::hash_type i = 0; i < Path::hash_range; ++i) {
        pathtree_pathhash_t **phAdr;
        phAdr = hashAddressForNodeType(pt, PathTree::Node::Branch);
        nodeCount[PathTree::Node::Branch] = 0;
        if (*phAdr)
        for (node = (**phAdr)[i].head; node; node = node->next)
            ++nodeCount[PathTree::Node::Branch];
        phAdr = hashAddressForNodeType(pt, PT_LEAF);
        nodeCount[PT_LEAF] = 0;
        if (*phAdr) {
            size_t chainHeight = 0;
            for (node = (**phAdr)[i].head; node; node = node->next) {
                size_t height = 0;
                PathTreeNode *other = node;
                ++nodeCount[PT_LEAF];
                while ((other = other->parent())) { ++height; }
                if (height > chainHeight)
                    chainHeight = height;
            }
            if (chainHeight > nodeBucketHeight)
                nodeBucketHeight = chainHeight;
        }
        totalForRange = nodeCount[PathTree::Node::Branch] + nodeCount[PT_LEAF];
        nodeCountSum[PT_BRANCH] += nodeCount[PathTree::Node::Branch];
        nodeCountSum[PT_LEAF]   += nodeCount[PT_LEAF];
        for (int j = 0; j < PATHTREENODE_TYPE_COUNT; ++j) {
            if (nodeCount[j] != 0) {
                if (nodeCount[j] > 1)
                    nodeBucketCollisions[j] += nodeCount[j]-1;
            }
            else {
                ++(nodeBucketEmpty[j]);
            }
            if (nodeCount[j] > nodeBucketCollisionsMax[j])
                nodeBucketCollisionsMax[j] = nodeCount[j];
        }
        size_t max = 0;
        for (int j = 0; j < PATHTREENODE_TYPE_COUNT; ++j) {
            max += nodeCount[j];
        }
        if (max > nodeBucketCollisionsMaxTotal)
            nodeBucketCollisionsMaxTotal = max;
        if (totalForRange != 0) {
            if (totalForRange > 1)
                nodeBucketCollisionsTotal += totalForRange-1;
        }
        else {
            ++nodeBucketEmptyTotal;
        }
        if (totalForRange > nodeBucketCollisionsMaxTotal)
            nodeBucketCollisionsMaxTotal = totalForRange;
    }
    printDistributionOverview(pt, nodeCountSum, nodeCountTotal,
        nodeBucketCollisions,    nodeBucketCollisionsTotal,
        nodeBucketCollisionsMax, nodeBucketCollisionsMaxTotal,
        nodeBucketEmpty, nodeBucketEmptyTotal,
        nodeBucketHeight, nodeCount);
    Con_Printf("\n");
    printDistributionHistogram(pt, 16, nodeCountTotal);
#endif
}
#endif
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(PathTree::Node) {
    PathTree &tree;
    PathTree::Node *parent;
    PathTree::Node::Children *children;
    PathTree::SegmentId segmentId;
    String const *segmentText = nullptr; // owned by the PathTree
    Impl(PathTree &_tree, bool isLeaf, PathTree::SegmentId _segmentId,
             PathTree::Node *_parent)
        : tree(_tree), parent(_parent), children(0), segmentId(_segmentId) {
        if (!isLeaf) children = new PathTree::Node::Children;
    }
    ~Impl() {
        delete children;
    }
    void cacheSegmentText() {
        segmentText = &tree.segmentName(segmentId);
    }
};
PathTree::Node::Node(PathTree::NodeArgs const &args) : d(nullptr) {
    d.reset(new Impl(args.tree, args.type == PathTree::Leaf, args.segmentId, args.parent));
    if (d->parent) d->parent->addChild(*this);
}
PathTree::Node::~Node()
{}
bool PathTree::Node::isLeaf() const {
    return d->children == 0;
}
PathTree &PathTree::Node::tree() const {
    return d->tree;
}
PathTree::Node &PathTree::Node::parent() const {
    return *d->parent;
}
const PathTree::Node::Children &PathTree::Node::children() const {
    DENG2_ASSERT(d->children != 0);
    return *d->children;
}
PathTree::Nodes const &PathTree::Node::childNodes(PathTree::NodeType type) const {
    DENG2_ASSERT(d->children != 0);
    return (type == PathTree::Leaf? d->children->leaves : d->children->branches);
}
PathTree::Nodes &PathTree::Node::childNodes(PathTree::NodeType type) {
    DENG2_ASSERT(d->children != 0);
    return (type == PathTree::Leaf? d->children->leaves : d->children->branches);
}
bool PathTree::Node::isAtRootLevel() const {
    return d->parent == &d->tree.rootBranch();
}
PathTree::SegmentId PathTree::Node::segmentId() const {
    return d->segmentId;
}
void PathTree::Node::addChild(PathTree::Node &node) {
    DENG2_ASSERT(d->children != 0);
    childNodes(node.type()).insert(node.hash(), &node);
}
void PathTree::Node::removeChild(PathTree::Node &node) {
    DENG2_ASSERT(d->children != 0);
    childNodes(node.type()).remove(node.hash(), &node);
}
String const &PathTree::Node::name() const {
    if (!d->segmentText) {
        d->cacheSegmentText();
    }
    return *d->segmentText;
}
Path::hash_type PathTree::Node::hash() const {
    return tree().segmentHash(d->segmentId);
}
static int matchName(QChar const *string,  dsize stringSize,
                     QChar const *pattern, dsize patternSize) {
    QChar const *in    = string;
    QChar const *inEnd = string + stringSize;
    QChar const *pat   = pattern;
    while (in < inEnd) {
        if (*pat == QChar('*')) {
            pat++;
            continue;
        }
        if (/**st != QChar('?') && */ pat->toLower() != in->toLower())
        {
            while (pat >= pattern && *pat != QChar('*')) { --pat; }
            if (pat < pattern) return false;
        }
        pat++;
        in++;
    }
    while (*pat == QChar('*')) { pat  ++; }
    return pat == (pattern + patternSize);
}
int PathTree::Node::comparePath(de::Path const &searchPattern, ComparisonFlags flags) const {
    if (((flags & PathTree::NoLeaf)   && isLeaf()) ||
        ((flags & PathTree::NoBranch) && isBranch())) {
        return 1;
    }
    de::Path::Segment const *snode = &searchPattern.lastSegment();
    int pathNodeCount = searchPattern.segmentCount();
    PathTree::Node const *node = this;
    for (int i = 0; i < pathNodeCount; ++i) {
        if (!snode->hasWildCard()) {
            if (snode->hash() != node->hash()) {
                return 1;
            }
            if (node->name().compare(snode->toStringRef(), Qt::CaseInsensitive)) {
                return 1;
            }
        }
        else {
            if (!matchName(node->name().constData(), node->name().size(),
                           snode->toStringRef().constData(), snode->toStringRef().size())) {
                return 1;
            }
        }
        if (i == pathNodeCount - 1) {
            return !(!(flags & MatchFull) || node->isAtRootLevel());
        }
        if (node->isAtRootLevel()) {
            return 1;
        }
        node  = &node->parent();
        snode = &searchPattern.reverseSegment(i + 1);
    }
    return 1;
}
#ifdef LIBDENG_STACK_MONITOR
static void *stackStart;
static size_t maxStackDepth;
#endif
namespace internal {
    struct PathConstructorArgs {
        int length;
        QChar separator;
        String composedPath;
        PathConstructorArgs(QChar sep = '/') : length(0), separator(sep)
        {}
    };
}
/**
 * Recursive path constructor. First finds the root and the full length of the
 * path (when descending), then allocates memory for the string, and finally
 * copies each segment with the separators (on the way out).
 */
static void pathConstructor(internal::PathConstructorArgs &args, PathTree::Node const &trav) {
    String const &segment = trav.name();
#ifdef LIBDENG_STACK_MONITOR
    maxStackDepth = MAX_OF(maxStackDepth, stackStart - (void *)&fragment);
#endif
    args.length += segment.length();
    if (!trav.isAtRootLevel()) {
        if (!args.separator.isNull()) {
            args.length += 1;
        }
        pathConstructor(args, trav.parent());
        if (!args.separator.isNull())
            args.composedPath.append(args.separator);
    }
    else if (args.composedPath) {
        args.composedPath.reserve(args.length);
    }
    args.composedPath.append(segment);
}
/**
 * @todo Paths within the tree store Path-compatible segment hashes. It would
 * be better to compose the full path in such a way that applies the already
 * calculated hashes. This would avoid the re-hash when the resultant path is
 * next compared (which is likely the reason the user is calling) -ds
 *
 * @todo This is a good candidate for result caching: the constructed path
 * could be saved and returned on subsequent calls. Are there any circumstances
 * in which the cached result becomes obsolete? -jk
 *
 * The only times the result becomes obsolete is when the delimiter is changed
 * or when the directory itself is rebuilt (in which case the nodes themselves
 * will be free'd). Note that any caching mechanism should not counteract one
 * of the primary goals of this class, i.e., optimal memory usage for the whole
 * directory. Caching constructed paths for every leaf node in the directory
 * would completely negate the benefits of the design of this class.
 *
 * Perhaps a fixed size MRU cache? -ds
 */
Path PathTree::Node::path(QChar sep) const {
    internal::PathConstructorArgs args(sep);
#ifdef LIBDENG_STACK_MONITOR
    stackStart = &parm;
#endif
    if (!sep.isNull() && isBranch()) {
        args.length++; // A single character.
    }
    pathConstructor(args, *this);
    if (!sep.isNull() && isBranch()) {
        args.composedPath += sep;
    }
    DENG2_ASSERT(args.composedPath.length() == (int)args.length);
#ifdef LIBDENG_STACK_MONITOR
    LOG_AS("pathConstructor");
    LOG_DEV_NOTE("Max stack depth: %1 bytes") << maxStackDepth;
#endif
    return Path(args.composedPath, sep);
}
UserDataNode::UserDataNode(PathTree::NodeArgs const &args, void *userPointer, int userValue)
    : PathTree::Node(args),
      _pointer(userPointer),
      _value(userValue)
{}
void *UserDataNode::userPointer() const {
    return _pointer;
}
int UserDataNode::userValue() const {
    return _value;
}
UserDataNode &UserDataNode::setUserPointer(void *ptr) {
    _pointer = ptr;
    return *this;
}
UserDataNode &UserDataNode::setUserValue(int value) {
    _value = value;
    return *this;
}
} // namespace de
namespace de {
static duint16          const POINTERSET_MIN_ALLOC      = 2;
static duint16          const POINTERSET_MAX_SIZE       = 0xffff;
static PointerSet::Flag const POINTERSET_ITERATION_MASK = 0x00ff;
PointerSet::Flag const PointerSet::AllowInsertionDuringIteration = 0x8000;
PointerSet::PointerSet()
    : _pointers(nullptr)
    , _iterationObserver(nullptr)
    , _flags(0)
    , _size (0)
{}
PointerSet::PointerSet(PointerSet const &other)
    : _iterationObserver(other._iterationObserver)
    , _flags(other._flags)
    , _size (other._size)
    , _range(other._range) {
    auto const bytes = sizeof(Pointer) * _size;
    _pointers = reinterpret_cast<Pointer *>(malloc(bytes));
    std::memcpy(_pointers, other._pointers, bytes);
}
PointerSet::PointerSet(PointerSet &&moved)
    : _pointers(moved._pointers)
    , _iterationObserver(moved._iterationObserver)
    , _flags(moved._flags)
    , _size (moved._size)
    , _range(moved._range) {
    moved._pointers = nullptr; // taken
}
PointerSet::~PointerSet() {
    DENG2_ASSERT(!isBeingIterated());
    free(_pointers);
}
void PointerSet::insert(Pointer ptr) {
    if (!_pointers) {
        _size = POINTERSET_MIN_ALLOC;
        _pointers = reinterpret_cast<Pointer *>(calloc(sizeof(Pointer), _size));
    }
    if (_range.isEmpty()) {
        duint16 const pos = _size / 2;
        _pointers[pos] = ptr;
        _range.start = pos;
        _range.end = pos + 1;
    }
    else {
        auto const loc = locate(ptr);
        if (!loc.isEmpty()) return; // Already got it.
        if (isBeingIterated()) {
            DENG2_ASSERT(_flags & AllowInsertionDuringIteration);
            if (!(_flags & AllowInsertionDuringIteration)) {
                return;
            }
            DENG2_ASSERT(_iterationObserver != nullptr);
        }
        if (_range.size() == _size) {
            DENG2_ASSERT(_size < POINTERSET_MAX_SIZE);
            if (_size == POINTERSET_MAX_SIZE) return; // Can't do it.
            Pointer *oldBase = _pointers;
            duint const oldSize = _size;
            _size = (_size < 0x8000? (_size * 2) : POINTERSET_MAX_SIZE);
            _pointers = reinterpret_cast<Pointer *>(realloc(_pointers, sizeof(Pointer) * _size));
            std::memset(_pointers + oldSize, 0, sizeof(Pointer) * (_size - oldSize));
            if (_iterationObserver && _pointers != oldBase) {
                _iterationObserver->pointerSetIteratorsWereInvalidated(oldBase, _pointers);
            }
        }
        duint16 const pos = loc.start;
        if (pos == _range.start && _range.start > 0) {
            _pointers[--_range.start] = ptr;
        }
        else if (pos == _range.end && _range.end < _size) {
            _pointers[_range.end++] = ptr;
        }
        else {
            duint16 const middle = (_range.start + _range.end + 1)/2;
            if ((pos > middle && _range.end < _size) || // Less stuff to move toward the end.
                _range.start == 0) {
                DENG2_ASSERT(_range.end < _size);
                std::memmove(_pointers + pos + 1,
                             _pointers + pos,
                             sizeof(Pointer) * (_range.end - pos));
                _range.end++;
                _pointers[pos] = ptr;
            }
            else {
                DENG2_ASSERT(_range.start > 0);
                std::memmove(_pointers + _range.start - 1,
                             _pointers + _range.start,
                             sizeof(Pointer) * (pos < _range.end? (pos - _range.start + 1) :
                                                                  (_range.size())));
                _pointers[pos - 1] = ptr;
                _range.start--;
            }
        }
    }
}
void PointerSet::remove(Pointer ptr) {
    auto const loc = locate(ptr);
    if (!loc.isEmpty()) {
        DENG2_ASSERT(!_range.isEmpty());
        if (loc.start == _range.start) {
            _pointers[_range.start++] = nullptr;
        }
        else if (loc.start == _range.end - 1 && !isBeingIterated()) {
            _pointers[--_range.end] = nullptr;
        }
        else {
            std::memmove(_pointers + _range.start + 1,
                         _pointers + _range.start,
                         sizeof(Pointer) * (loc.start - _range.start));
            _pointers[_range.start++] = nullptr;
        }
        DENG2_ASSERT(_range.start <= _range.end);
    }
}
bool PointerSet::contains(Pointer ptr) const {
    return !locate(ptr).isEmpty();
}
void PointerSet::clear() {
    if (_pointers) {
        std::memset(_pointers, 0, sizeof(Pointer) * _size);
        _range = Rangeui16(_range.end, _range.end);
    }
}
PointerSet &PointerSet::operator = (PointerSet const &other) {
    auto const bytes = sizeof(Pointer) * other._size;
    if (_size != other._size) {
        _size = other._size;
        _pointers = reinterpret_cast<Pointer *>(realloc(_pointers, bytes));
    }
    std::memcpy(_pointers, other._pointers, bytes);
    _flags = other._flags;
    _range = other._range;
    _iterationObserver = other._iterationObserver;
    return *this;
}
PointerSet &PointerSet::operator = (PointerSet &&moved) {
    free(_pointers);
    _pointers = moved._pointers;
    moved._pointers = nullptr;
    _flags = moved._flags;
    _size  = moved._size;
    _range = moved._range;
    _iterationObserver = moved._iterationObserver;
    return *this;
}
void PointerSet::setBeingIterated(bool yes) const {
    duint16 count = _flags & POINTERSET_ITERATION_MASK;
    _flags ^= count;
    if (yes) {
        DENG2_ASSERT(count != POINTERSET_ITERATION_MASK);
        ++count;
    }
    else {
        DENG2_ASSERT(count != 0);
        --count;
    }
    _flags |= count & POINTERSET_ITERATION_MASK;
}
bool PointerSet::isBeingIterated() const {
    return (_flags & POINTERSET_ITERATION_MASK) != 0;
}
void PointerSet::setIterationObserver(IIterationObserver *observer) const {
    _iterationObserver = observer;
}
Rangeui16 PointerSet::locate(Pointer ptr) const {
    Rangeui16 span = _range;
    while (!span.isEmpty()) {
        if (span.size() == 1) {
            if (at(span.start) == ptr) {
                return span; // Found it.
            }
            if (ptr < at(span.start)) {
                return Rangeui16(span.start, span.start);
            }
            return Rangeui16(span.end, span.end);
        }
        Rangeui16 const rightHalf((span.start + span.end + 1) / 2, span.end);
        Pointer const mid = at(rightHalf.start);
        if (ptr == mid) {
            return Rangeui16(rightHalf.start, rightHalf.start + 1);
        }
        else if (ptr > mid) {
            span = rightHalf;
        }
        else {
            span = Rangeui16(span.start, rightHalf.start);
        }
    }
    return span;
}
} // namespace de
namespace de {
static String nameToKey(String const &name) {
    return name.toLower();
}
DENG2_PIMPL(Profiles)
, DENG2_OBSERVES(Deletable, Deletion) {
    typedef QMap<String, AbstractProfile *> Profiles;
    Profiles profiles;
    String persistentName;
    Impl(Public *i) : Base(i)
    {}
    ~Impl() {
        clear();
    }
    void add(AbstractProfile *profile) {
        String const key = nameToKey(profile->name());
        if (profiles.contains(nameToKey(key))) {
            delete profiles[key];
        }
        profiles.insert(key, profile);
        profile->setOwner(thisPublic);
        profile->audienceForDeletion += this;
        DENG2_FOR_PUBLIC_AUDIENCE2(Addition, i) {
            i->profileAdded(*profile);
        }
    }
    void remove(AbstractProfile &profile) {
        profile.audienceForDeletion -= this;
        profile.setOwner(nullptr);
        profiles.remove(nameToKey(profile.name()));
        DENG2_FOR_PUBLIC_AUDIENCE2(Removal, i) {
            i->profileRemoved(profile);
        }
    }
    void changeLookupKey(AbstractProfile const &profile, String const &newName) {
        profiles.remove(nameToKey(profile.name()));
        profiles.insert(nameToKey(newName), const_cast<AbstractProfile *>(&profile));
    }
    void objectWasDeleted(Deletable *obj) {
        QMutableMapIterator<String, AbstractProfile *> iter(profiles);
        while (iter.hasNext()) {
            iter.next();
            if (iter.value() == obj) {
                iter.remove();
                break;
            }
        }
    }
    void clear() {
        for (auto *prof : profiles) {
            prof->audienceForDeletion -= this;
            prof->setOwner(nullptr);
        }
        qDeleteAll(profiles.values());
        profiles.clear();
    }
    /**
     * For persistent profiles, determines the file name of the Info file
     * where all the profile contents are written to and read from.
     */
    String fileName() const {
        if (persistentName.isEmpty()) return "";
        return String("/home/configs/%1.dei").arg(persistentName);
    }
    void loadProfilesFromInfo(File const &file, bool markReadOnly) {
        try {
            LOG_RES_VERBOSE("Reading profiles from %s") << file.description();
            Block raw;
            file >> raw;
            de::Info info;
            info.parse(String::fromUtf8(raw));
            foreach (de::Info::Element const *elem, info.root().contentsInOrder()) {
                if (!elem->isBlock()) continue;
                de::Info::BlockElement const &profBlock = elem->as<de::Info::BlockElement>();
                if (profBlock.blockType() == "group" &&
                    profBlock.name()      == "profile") {
                    String profileName = profBlock.keyValue("name").text;
                    if (profileName.isEmpty()) continue; // Name is required.
                    LOG_VERBOSE("Reading profile '%s'") << profileName;
                    auto *prof = self().profileFromInfoBlock(profBlock);
                    prof->setName(profileName);
                    prof->setReadOnly(markReadOnly);
                    add(prof);
                }
            }
        }
        catch (Error const &er) {
            LOG_RES_WARNING("Failed to load profiles from %s:\n%s")
                    << file.description() << er.asText();
        }
    }
    DENG2_PIMPL_AUDIENCE(Addition)
    DENG2_PIMPL_AUDIENCE(Removal)
};
DENG2_AUDIENCE_METHOD(Profiles, Addition)
DENG2_AUDIENCE_METHOD(Profiles, Removal)
Profiles::Profiles()
    : d(new Impl(this))
{}
Profiles::~Profiles()
{}
StringList Profiles::profiles() const {
    StringList names;
    for (auto const *p : d->profiles.values()) names << p->name();
    return names;
}
int Profiles::count() const {
    return d->profiles.size();
}
Profiles::AbstractProfile *Profiles::tryFind(String const &name) const {
    auto found = d->profiles.constFind(nameToKey(name));
    if (found != d->profiles.constEnd()) {
        return found.value();
    }
    return nullptr;
}
Profiles::AbstractProfile &Profiles::find(String const &name) const {
    if (auto *p = tryFind(name)) {
        return *p;
    }
    throw NotFoundError("Profiles::find", "Profile '" + name + "' not found");
}
void Profiles::setPersistentName(String const &name) {
    d->persistentName = name;
}
String Profiles::persistentName() const {
    return d->persistentName;
}
bool Profiles::isPersistent() const {
    return !d->persistentName.isEmpty();
}
LoopResult Profiles::forAll(std::function<LoopResult (AbstractProfile &)> func) const {
    foreach (AbstractProfile *prof, d->profiles.values()) {
        if (auto result = func(*prof)) {
            return result;
        }
    }
    return LoopContinue;
}
void Profiles::clear() {
    d->clear();
}
void Profiles::add(AbstractProfile *profile) {
    d->add(profile);
}
void Profiles::remove(AbstractProfile &profile) {
    DENG2_ASSERT(&profile.owner() == this);
    d->remove(profile);
}
bool Profiles::rename(AbstractProfile const &profile, String const &newName) {
    if (newName.isEmpty() || tryFind(newName)) return false;
    d->changeLookupKey(profile, newName);
    return true;
}
void Profiles::serialize() const {
    if (!isPersistent()) return;
    LOG_AS("Profiles");
    LOGDEV_VERBOSE("Serializing %s profiles") << d->persistentName;
    String info;
    QTextStream os(&info);
    os.setCodec("UTF-8");
    os << "# Autogenerated Info file based on " << d->persistentName
       << " profiles\n";
    int count = 0;
    for (auto *prof : d->profiles) {
        if (prof->isReadOnly()) continue;
        os << "\nprofile {\n"
              "    name: " << prof->name() << "\n";
        for (auto line : prof->toInfoSource().split('\n')) {
            os << "    " << line << "\n";
        }
        os << "}\n";
        ++count;
    }
    File &outFile = App::rootFolder().replaceFile(d->fileName());
    outFile << info.toUtf8();
    outFile.flush(); // we're done
    LOG_VERBOSE("Wrote \"%s\" with %i profile%s")
            << d->fileName() << count << (count != 1? "s" : "");
}
void Profiles::deserialize() {
    if (!isPersistent()) return;
    LOG_AS("Profiles");
    LOGDEV_VERBOSE("Deserializing %s profiles") << d->persistentName;
    clear();
    FS::FoundFiles folders;
    App::fileSystem().findAll("profiles" / d->persistentName, folders);
    DENG2_FOR_EACH(FS::FoundFiles, i, folders) {
        if (auto const *folder = maybeAs<Folder>(*i)) {
            folder->forContents([this] (String name, File &file) {
                if (name.fileNameExtension() == ".dei") {
                    d->loadProfilesFromInfo(file, true /* read-only */);
                }
                return LoopContinue;
            });
        }
    }
    if (File const *file = App::rootFolder().tryLocate<File const>(d->fileName())) {
        d->loadProfilesFromInfo(*file, false /* modifiable */);
    }
}
DENG2_PIMPL(Profiles::AbstractProfile) {
    Profiles *owner = nullptr;
    String name;
    bool readOnly = false;
    Impl(Public *i) : Base(i) {}
    ~Impl() {
        if (owner) {
            owner->remove(self());
        }
    }
    DENG2_PIMPL_AUDIENCE(Change)
};
DENG2_AUDIENCE_METHOD(Profiles::AbstractProfile, Change)
Profiles::AbstractProfile::AbstractProfile()
    : d(new Impl(this))
{}
Profiles::AbstractProfile::AbstractProfile(AbstractProfile const &profile)
    : d(new Impl(this)) {
    d->name     = profile.name();
    d->readOnly = profile.isReadOnly();
}
Profiles::AbstractProfile::~AbstractProfile()
{}
Profiles::AbstractProfile &Profiles::AbstractProfile::operator = (AbstractProfile const &other) {
    d->name     = other.d->name;
    d->readOnly = other.d->readOnly;
    return *this;
}
void Profiles::AbstractProfile::setOwner(Profiles *owner) {
    DENG2_ASSERT(d->owner != owner);
    d->owner = owner;
}
Profiles &Profiles::AbstractProfile::owner() {
    DENG2_ASSERT(d->owner);
    return *d->owner;
}
Profiles const &Profiles::AbstractProfile::owner() const {
    DENG2_ASSERT(d->owner);
    return *d->owner;
}
String Profiles::AbstractProfile::name() const {
    return d->name;
}
bool Profiles::AbstractProfile::setName(String const &newName) {
    if (newName.isEmpty()) return false;
    Profiles *owner = d->owner;
    if (!owner ||
        !d->name.compareWithoutCase(newName) || // just a case change
        owner->rename(*this, newName)) {
        d->name = newName;
        notifyChange();
    }
    return true;
}
bool Profiles::AbstractProfile::isReadOnly() const {
    return d->readOnly;
}
void Profiles::AbstractProfile::setReadOnly(bool readOnly) {
    d->readOnly = readOnly;
}
void Profiles::AbstractProfile::notifyChange() {
    DENG2_FOR_AUDIENCE2(Change, i) {
        i->profileChanged(*this);
    }
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Reader) {
    ByteOrder const &convert;
    duint version;
    IByteArray const *source;
    IByteArray::Offset offset;
    IByteArray::Offset markOffset;
    IIStream *stream;
    IIStream const *constStream;
    dsize numReceivedBytes;
    Block incoming;     ///< Buffer for bytes received so far from the stream.
    bool marking;       ///< @c true, if marking is occurring (mark() called).
    Block markedData;   ///< All read data since the mark was set.
    Impl(ByteOrder const &order, IByteArray const *src, IByteArray::Offset off)
        : convert(order), version(DENG2_PROTOCOL_LATEST),
          source(src), offset(off), markOffset(off),
          stream(0), constStream(0), numReceivedBytes(0), marking(false)
    {}
    Impl(ByteOrder const &order, IIStream *str)
        : convert(order), version(DENG2_PROTOCOL_LATEST),
          source(0), offset(0), markOffset(0),
          stream(str), constStream(0), numReceivedBytes(0), marking(false) {
        upgradeToByteArray();
    }
    Impl(ByteOrder const &order, IIStream const *str)
        : convert(order), version(DENG2_PROTOCOL_LATEST),
          source(0), offset(0), markOffset(0),
          stream(0), constStream(str), numReceivedBytes(0), marking(false) {
        upgradeToByteArray();
    }
    /**
     * Byte arrays provide more freedom with reading. If the streaming object
     * happens to support the byte array interface, Reader will use it instead.
     */
    void upgradeToByteArray() {
        if (stream) {
            if ((source = dynamic_cast<IByteArray const *>(stream)) != 0) {
                stream = 0;
            }
        }
        if (constStream) {
            if ((source = dynamic_cast<IByteArray const *>(constStream)) != 0) {
                constStream = 0;
            }
        }
    }
    /**
     * Reads bytes from the stream and adds them to the incoming buffer.
     *
     * @param expectedSize  Number of bytes that the reader is expecting to read.
     */
    void update(dsize expectedSize = 0) {
        if (incoming.size() >= expectedSize) {
            return;
        }
        if (stream) {
            Block b;
            *stream >> b;
            incoming += b;
        }
        else if (constStream) {
            Block b;
            *constStream >> b;
            b.remove(0, numReceivedBytes);
            incoming += b;
            numReceivedBytes += b.size();
        }
    }
    void readBytes(IByteArray::Byte *ptr, dsize size) {
        if (source) {
            source->get(offset, ptr, size);
            offset += size;
        }
        else if (stream || constStream) {
            update(size);
            if (incoming.size() >= size) {
                std::memcpy(ptr, incoming.constData(), size);
                if (marking) {
                    markedData += incoming.left(size);
                }
                incoming.remove(0, size);
            }
            else {
                throw IIStream::InputError("Reader::readBytes",
                        QString("Attempted to read %1 bytes from stream while only %2 "
                                "bytes are available").arg(size).arg(incoming.size()));
            }
        }
    }
    void mark() {
        if (source) {
            markOffset = offset;
        }
        else {
            markedData.clear();
            marking = true;
        }
    }
    void rewind() {
        if (source) {
            offset = markOffset;
        }
        else {
            incoming.prepend(markedData);
            markedData.clear();
            marking = false;
        }
    }
    bool atEnd() {
        return remaining() == 0;
    }
    IByteArray::Size remaining() {
        if (source) {
            return source->size() - offset;
        }
        if (stream || constStream) {
            update();
            return incoming.size();
        }
        return 0;
    }
};
Reader::Reader(Reader const &other) : d(new Impl(*other.d))
{}
Reader::Reader(IByteArray const &source, ByteOrder const &byteOrder, IByteArray::Offset offset)
    : d(new Impl(byteOrder, &source, offset))
{}
Reader::Reader(IIStream &stream, ByteOrder const &byteOrder)
    : d(new Impl(byteOrder, &stream))
{}
Reader::Reader(IIStream const &stream, ByteOrder const &byteOrder)
    : d(new Impl(byteOrder, &stream))
{}
Reader &Reader::withHeader() {
    duint32 header = 0;
    *this >> header;
    d->version = header;
    if (d->version > DENG2_PROTOCOL_LATEST) {
        throw VersionError("Reader::withHeader",
                           QString("Version %1 is unknown").arg(d->version));
    }
    return *this;
}
duint Reader::version() const {
    return d->version;
}
void Reader::setVersion(duint version) {
    d->version = version;
}
Reader &Reader::operator >> (char &byte) {
    return *this >> reinterpret_cast<duchar &>(byte);
}
Reader &Reader::operator >> (dchar &byte) {
    return *this >> reinterpret_cast<duchar &>(byte);
}
Reader &Reader::operator >> (duchar &byte) {
    d->readBytes(&byte, 1);
    return *this;
}
Reader &Reader::operator >> (dint16 &word) {
    return *this >> reinterpret_cast<duint16 &>(word);
}
Reader &Reader::operator >> (duint16 &word) {
    d->readBytes(reinterpret_cast<IByteArray::Byte *>(&word), 2);
    d->convert.networkToHost(word, word);
    return *this;
}
Reader &Reader::operator >> (dint32 &dword) {
    return *this >> reinterpret_cast<duint32 &>(dword);
}
Reader &Reader::operator >> (duint32 &dword) {
    d->readBytes(reinterpret_cast<IByteArray::Byte *>(&dword), 4);
    d->convert.networkToHost(dword, dword);
    return *this;
}
Reader &Reader::operator >> (dint64 &qword) {
    return *this >> reinterpret_cast<duint64 &>(qword);
}
Reader &Reader::operator >> (duint64 &qword) {
    d->readBytes(reinterpret_cast<IByteArray::Byte *>(&qword), 8);
    d->convert.networkToHost(qword, qword);
    return *this;
}
Reader &Reader::operator >> (dfloat &value) {
    return *this >> *reinterpret_cast<duint32 *>(&value);
}
Reader &Reader::operator >> (ddouble &value) {
    return *this >> *reinterpret_cast<duint64 *>(&value);
}
Reader &Reader::operator >> (String &text) {
    duint size = 0;
    *this >> size;
    Block bytes;
    for (duint i = 0; i < size; ++i) {
        IByteArray::Byte ch = 0;
        *this >> ch;
        bytes.append(ch);
    }
    text = String::fromUtf8(bytes);
    return *this;
}
Reader &Reader::operator >> (Block &block) {
    return *this >> static_cast<IReadable &>(block);
}
Reader &Reader::operator >> (IByteArray &byteArray) {
    duint32 size = 0;
    *this >> size;
    /**
     * @note  A temporary copy of the contents of the array is made
     * because the destination byte array is not guaranteed to be
     * a memory buffer where you can copy the contents directly.
     */
    Block data(size);
    d->readBytes(data.data(), size);
    byteArray.set(0, data.dataConst(), size);
    return *this;
}
Reader &Reader::operator >> (FixedByteArray &fixedByteArray) {
    /**
     * @note  A temporary copy of the contents of the array is made
     * because the destination byte array is not guaranteed to be
     * a memory buffer where you can copy the contents directly.
     */
    dsize const size = fixedByteArray.size();
    Block data(size);
    d->readBytes(data.data(), size);
    fixedByteArray.set(0, data.dataConst(), size);
    return *this;
}
Reader &Reader::readBytes(dsize count, IByteArray &destination) {
    FixedByteArray dest(destination, 0, count);
    return *this >> dest;
}
Reader &Reader::readBytesFixedSize(IByteArray &destination) {
    FixedByteArray dest(destination);
    return *this >> dest;
}
Reader &Reader::operator >> (IReadable &readable) {
    readable << *this;
    return *this;
}
Reader &Reader::readUntil(IByteArray &byteArray, IByteArray::Byte delimiter) {
    dsize pos = 0;
    IByteArray::Byte b = 0;
    do {
        if (atEnd()) break;
        *this >> b;
        byteArray.set(pos++, &b, 1);
    } while (b != delimiter);
    return *this;
}
Reader &Reader::readLine(String &string) {
    string.clear();
    Block utf;
    readUntil(utf, '\n');
    string = String::fromUtf8(utf);
    string.replace("\r", ""); // strip carriage returns
    return *this;
}
String Reader::readLine() {
    String s;
    readLine(s);
    return s;
}
IByteArray const *Reader::source() const {
    return d->source;
}
bool Reader::atEnd() const {
    return d->atEnd();
}
IByteArray::Offset Reader::offset() const {
    return d->offset;
}
IByteArray::Size Reader::remainingSize() const {
    return d->remaining();
}
void Reader::setOffset(IByteArray::Offset offset) {
    d->offset = offset;
}
void Reader::seek(IByteArray::Delta count) {
    if (!d->source) {
        throw SeekError("Reader::seek", "Cannot seek when reading from a stream");
    }
    auto const seeked = IByteArray::Offset(IByteArray::Delta(d->offset) + count);
    if (seeked >= d->source->size()) {
        throw IByteArray::OffsetError("Reader::seek", "Seek past bounds of source data");
    }
    d->offset = seeked;
}
void Reader::mark() {
    d->mark();
}
void Reader::rewind() {
    d->rewind();
}
ByteOrder const &Reader::byteOrder() const {
    return d->convert;
}
} // namespace de
namespace de {
int const SUBRECORD_CONTENT_EXCERPT_THRESHOLD = 100; // lines
String const Record::VAR_SUPER = "__super__";
String const Record::VAR_FILE  = "__file__";
String const Record::VAR_INIT  = "__init__";
String const Record::VAR_NATIVE_SELF = "__self__";
/**
 * Each record is given a unique identifier, so that serialized record
 * references can be tracked to their original target.
 */
static std::atomic_uint recordIdCounter;
DENG2_PIMPL(Record)
, public Lockable
, DENG2_OBSERVES(Variable, Deletion) {
    Record::Members members;
    duint32 uniqueId; ///< Identifier to track serialized references.
    duint32 oldUniqueId;
    Flags flags = DefaultFlags;
    typedef QHash<duint32, Record *> RefMap;
    Impl(Public &r)
        : Base(r)
        , uniqueId(++recordIdCounter)
        , oldUniqueId(0)
    {}
    struct ExcludeByBehavior {
        Behavior behavior;
        ExcludeByBehavior(Behavior b) : behavior(b) {}
        bool operator () (Variable const &member) {
            return (behavior == IgnoreDoubleUnderscoreMembers &&
                    member.name().startsWith("__"));
        }
    };
    struct ExcludeByRegExp {
        QRegExp omitted;
        ExcludeByRegExp(QRegExp const &omit) : omitted(omit) {}
        bool operator () (Variable const &member) {
            return omitted.exactMatch(member.name());
        }
    };
    void clear(std::function<bool (Variable const &)> excluded) {
        if (!members.empty()) {
            Record::Members remaining; // Contains all members that are not removed.
            DENG2_FOR_EACH(Members, i, members) {
                if (excluded(*i.value())) {
                    remaining.insert(i.key(), i.value());
                    continue;
                }
                DENG2_FOR_PUBLIC_AUDIENCE2(Removal, o) o->recordMemberRemoved(self(), **i);
                i.value()->audienceForDeletion() -= this;
                delete i.value();
            }
            members = remaining;
        }
    }
    void copyMembersFrom(Record const &other, std::function<bool (Variable const &)> excluded) {
        auto const *other_d = other.d.getConst();
        DENG2_GUARD(other_d);
        DENG2_FOR_EACH_CONST(Members, i, other_d->members) {
            if (!excluded(*i.value())) {
                bool alreadyExists;
                Variable *var; {
                    DENG2_GUARD(this);
                    var = new Variable(*i.value());
                    var->audienceForDeletion() += this;
                    auto iter = members.find(i.key());
                    alreadyExists = (iter != members.end());
                    if (alreadyExists) {
                        iter.value()->audienceForDeletion() -= this;
                        delete iter.value();
                        iter.value() = var;
                    }
                    else {
                        members[i.key()] = var;
                    }
                }
                if (!alreadyExists) {
                    DENG2_FOR_PUBLIC_AUDIENCE2(Addition, i) i->recordMemberAdded(self(), *var);
                }
            }
        }
    }
    void assignPreservingVariables(Record const &other, std::function<bool (Variable const &)> excluded) {
        auto const *other_d = other.d.getConst();
        DENG2_GUARD(other_d);
        for (auto i = other_d->members.begin(); i != other_d->members.end(); ++i) {
            if (!excluded(*i.value())) {
                Variable *var = nullptr; {
                    DENG2_GUARD(this);
                    auto found = members.constFind(i.key());
                    if (found != members.constEnd()) {
                        var = found.value();
                    }
                }
                if (var) {
                    if (isSubrecord(*i.value()) && isSubrecord(*var)) {
                        var->valueAsRecord().d->assignPreservingVariables
                                (i.value()->valueAsRecord(), excluded);
                    }
                    else {
                        Variable::Flags const oldFlags = var->flags();
                        var->setFlags(Variable::ReadOnly, false);
                        var->set(i.value()->value());
                        var->setFlags(oldFlags, ReplaceFlags);
                    }
                }
                else {
                    DENG2_GUARD(this);
                    var = new Variable(*i.value());
                    var->audienceForDeletion() += this;
                    members[i.key()] = var;
                }
            }
        }
        DENG2_GUARD(this);
        QMutableHashIterator<String, Variable *> iter(members);
        while (iter.hasNext()) {
            iter.next();
            if (!excluded(*iter.value()) && !other.hasMember(iter.key())) {
                Variable *var = iter.value();
                iter.remove();
                var->audienceForDeletion() -= this;
                delete var;
            }
        }
    }
    static bool isRecord(Variable const &var) {
        RecordValue const *value = maybeAs<RecordValue>(var.value());
        return value && value->record();
    }
    static bool isSubrecord(Variable const &var) {
        RecordValue const *value = maybeAs<RecordValue>(var.value());
        return value && value->record() && value->hasOwnership();
    }
    LoopResult forSubrecords(std::function<LoopResult (String const &, Record &)> func) const {
        Members const unmodifiedMembers = members; // In case a callback removes members.
        DENG2_FOR_EACH_CONST(Members, i, unmodifiedMembers) {
            Variable const &member = *i.value();
            if (isSubrecord(member)) {
                Record *rec = member.value<RecordValue>().record();
                DENG2_ASSERT(rec != 0); // subrecords are owned, so cannot have been deleted
                if (auto result = func(i.key(), *rec)) {
                    return result;
                }
            }
        }
        return LoopContinue;
    }
    Record::Subrecords listSubrecords(std::function<bool (Record const &)> filter) const {
        DENG2_GUARD(this);
        Subrecords subs;
        forSubrecords([&subs, filter] (String const &name, Record &rec) {
            if (filter(rec)) {
                subs.insert(name, &rec);
            }
            return LoopContinue;
        });
        return subs;
    }
    Variable const *findMemberByPath(String const &name) const {
        int pos = name.indexOf('.');
        if (pos >= 0) {
            String subName = name.substr(0, pos);
            String remaining = name.substr(pos + 1);
            if (!self().hasRecord(subName)) return 0;
            return self()[subName].value<RecordValue>().dereference().d->findMemberByPath(remaining);
        }
        DENG2_GUARD(this);
        Members::const_iterator found = members.constFind(name);
        if (found != members.constEnd()) {
            return found.value();
        }
        return 0;
    }
    /**
     * Returns the record inside which the variable identified by path @a name
     * resides. The necessary subrecords are created if they don't exist.
     *
     * @param pathOrName  Variable name or path.
     *
     * @return  Parent record for the variable.
     */
    Record &parentRecordByPath(String const &pathOrName) {
        DENG2_GUARD(this);
        int pos = pathOrName.indexOf('.');
        if (pos >= 0) {
            String subName = pathOrName.substr(0, pos);
            String remaining = pathOrName.substr(pos + 1);
            Record *rec = 0;
            if (!self().hasSubrecord(subName)) {
                rec = &self().addSubrecord(subName);
            }
            else {
                rec = &self().subrecord(subName);
            }
            return rec->d->parentRecordByPath(remaining);
        }
        return self();
    }
    void variableBeingDeleted(Variable &variable) {
        DENG2_ASSERT(findMemberByPath(variable.name()));
        LOG_TRACE_DEBUGONLY("Variable %p deleted, removing from Record %p", &variable << thisPublic);
        DENG2_GUARD(this);
        members.remove(variable.name());
    }
    static String memberNameFromPath(String const &path) {
        return path.fileName('.');
    }
    /**
     * Reconnect record values that used to reference known records. After a
     * record has been deserialized, it may contain variables whose values
     * reference other records. The default behavior for a record is to
     * dereference records when serialized, but if the target has been
     * serialized as part of the record, we can restore the original reference
     * by looking at the IDs found in the serialized data.
     *
     * @param refMap  Known records indexes with their old IDs.
     */
    void reconnectReferencesAfterDeserialization(RefMap const &refMap) {
        DENG2_FOR_EACH(Members, i, members) {
            RecordValue *value = dynamic_cast<RecordValue *>(&i.value()->value());
            if (!value || !value->record()) continue;
            if (value->usedToHaveOwnership()) {
                value->record()->d->reconnectReferencesAfterDeserialization(refMap);
            }
            if (value->hasOwnership() && !value->usedToHaveOwnership()) {
                duint32 oldTargetId = value->record()->d->oldUniqueId;
                if (refMap.contains(oldTargetId)) {
                    LOG_TRACE_DEBUGONLY("RecordValue %p restored to reference record %i (%p)",
                                        value << oldTargetId << refMap[oldTargetId]);
                    value->setRecord(refMap[oldTargetId]);
                }
            }
        }
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(Addition)
    DENG2_PIMPL_AUDIENCE(Removal)
};
DENG2_AUDIENCE_METHOD(Record, Deletion)
DENG2_AUDIENCE_METHOD(Record, Addition)
DENG2_AUDIENCE_METHOD(Record, Removal)
Record::Record() : RecordAccessor(this), d(new Impl(*this))
{}
Record::Record(Record const &other, Behavior behavior)
    : RecordAccessor(this)
    , d(new Impl(*this)) {
    copyMembersFrom(other, behavior);
}
Record::Record(Record &&moved)
    : RecordAccessor(this)
    , d(std::move(moved.d)) {
    DENG2_ASSERT(moved.d.isNull());
    d->thisPublic = this;
}
Record::~Record() {
    if (!d.isNull()) // will be nullptr if moved away
    {
        DENG2_FOR_AUDIENCE2(Deletion, i) i->recordBeingDeleted(*this);
        clear();
    }
}
Record &Record::setFlags(Flags flags, FlagOpArg op) {
    applyFlagOperation(d->flags, flags, op);
    return *this;
}
Record::Flags Record::flags() const {
    return d->flags;
}
void Record::clear(Behavior behavior) {
    DENG2_GUARD(d);
    d->clear(Impl::ExcludeByBehavior(behavior));
}
void Record::copyMembersFrom(Record const &other, Behavior behavior) {
    d->copyMembersFrom(other, Impl::ExcludeByBehavior(behavior));
}
void Record::assignPreservingVariables(Record const &from, Behavior behavior) {
    d->assignPreservingVariables(from, Impl::ExcludeByBehavior(behavior));
}
Record &Record::operator = (Record const &other) {
    return assign(other);
}
Record &Record::operator = (Record &&moved) {
    d = std::move(moved.d);
    d->thisPublic = this;
    return *this;
}
Record &Record::assign(Record const &other, Behavior behavior) {
    if (this == &other) return *this;
    DENG2_GUARD(d);
    clear(behavior);
    copyMembersFrom(other, behavior);
    return *this;
}
Record &Record::assign(Record const &other, QRegExp const &excluded) {
    DENG2_GUARD(d);
    d->clear(Impl::ExcludeByRegExp(excluded));
    d->copyMembersFrom(other, Impl::ExcludeByRegExp(excluded));
    return *this;
}
bool Record::has(String const &name) const {
    return hasMember(name);
}
bool Record::hasMember(String const &variableName) const {
    return d->findMemberByPath(variableName) != 0;
}
bool Record::hasSubrecord(String const &subrecordName) const {
    Variable const *found = d->findMemberByPath(subrecordName);
    return found? d->isSubrecord(*found) : false;
}
bool Record::hasRecord(String const &recordName) const {
    Variable const *found = d->findMemberByPath(recordName);
    return found? d->isRecord(*found) : false;
}
Variable &Record::add(Variable *variable) {
    std::unique_ptr<Variable> var(variable);
    if (variable->name().empty()) {
        throw UnnamedError("Record::add", "All members of a record must have a name");
    } {
        DENG2_GUARD(d);
        if (hasMember(variable->name())) {
            delete d->members[variable->name()];
        }
        var->audienceForDeletion() += d;
        d->members[variable->name()] = var.release();
    }
    DENG2_FOR_AUDIENCE2(Addition, i) i->recordMemberAdded(*this, *variable);
    return *variable;
}
Variable *Record::remove(Variable &variable) { {
        DENG2_GUARD(d);
        variable.audienceForDeletion() -= d;
        d->members.remove(variable.name());
    }
    DENG2_FOR_AUDIENCE2(Removal, i) i->recordMemberRemoved(*this, variable);
    return &variable;
}
Variable *Record::remove(String const &variableName) {
    return remove((*this)[variableName]);
}
Variable *Record::tryRemove(String const &variableName) {
    if (has(variableName)) {
        return remove(variableName);
    }
    return nullptr;
}
void Record::removeMembersWithPrefix(const String &prefix) {
    for (const String &name : members().keys()) {
        if (name.beginsWith(prefix)) {
            remove(name);
        }
    }
}
Variable &Record::add(String const &name, Variable::Flags variableFlags) {
    return d->parentRecordByPath(name)
            .add(new Variable(Impl::memberNameFromPath(name), nullptr, variableFlags));
}
Variable &Record::addNumber(String const &name, Value::Number number) {
    return add(name, Variable::AllowNumber).set(NumberValue(number));
}
Variable &Record::addBoolean(String const &name, bool booleanValue) {
    return add(name, Variable::AllowNumber).set(NumberValue(booleanValue, NumberValue::Boolean));
}
Variable &Record::addText(String const &name, Value::Text const &text) {
    return add(name, Variable::AllowText).set(TextValue(text));
}
Variable &Record::addTime(String const &name, Time const &time) {
    return add(name, Variable::AllowTime).set(TimeValue(time));
}
Variable &Record::addArray(String const &name, ArrayValue *array) {
    if (!array) array = new ArrayValue;
    return add(name, Variable::AllowArray).set(array);
}
Variable &Record::addDictionary(String const &name) {
    return add(name, Variable::AllowDictionary).set(new DictionaryValue);
}
Variable &Record::addBlock(String const &name) {
    return add(name, Variable::AllowBlock).set(new BlockValue);
}
Variable &Record::addFunction(const String &name, Function *func) {
    return add(name, Variable::AllowFunction).set(new FunctionValue(func));
}
Record &Record::add(String const &name, Record *subrecord) {
    std::unique_ptr<Record> sub(subrecord);
    add(name).set(RecordValue::takeRecord(sub.release()));
    return *subrecord;
}
Record &Record::addSubrecord(String const &name, SubrecordAdditionBehavior behavior) {
    if (behavior == KeepExisting) {
        if (name.isEmpty()) {
            return *this;
        }
        if (hasSubrecord(name)) {
            return subrecord(name);
        }
    }
    return add(name, new Record);
}
Record *Record::removeSubrecord(String const &name) {
    Members::const_iterator found = d->members.find(name);
    if (found != d->members.end() && d->isSubrecord(*found.value())) {
        Record *returnedToCaller = found.value()->value<RecordValue>().takeRecord();
        remove(*found.value());
        return returnedToCaller;
    }
    throw NotFoundError("Record::remove", "Subrecord '" + name + "' not found");
}
Variable &Record::set(String const &name, bool value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(NumberValue(value));
    }
    return addBoolean(name, value);
}
Variable &Record::set(String const &name, char const *value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(TextValue(value));
    }
    return addText(name, value);
}
Variable &Record::set(String const &name, Value::Text const &value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(TextValue(value));
    }
    return addText(name, value);
}
Variable &Record::set(String const &name, Value::Number value) {
    return set(name, NumberValue(value));
}
Variable &Record::set(const String &name, const NumberValue &value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(value);
    }
    return add(name, Variable::AllowNumber).set(value);
}
Variable &Record::set(String const &name, dint32 value) {
    return set(name, NumberValue(value));
}
Variable &Record::set(String const &name, duint32 value) {
    return set(name, NumberValue(value));
}
Variable &Record::set(String const &name, dint64 value) {
  return set(name, NumberValue(value));
}
Variable &Record::set(String const &name, duint64 value) {
    return set(name, NumberValue(value));
}
Variable &Record::set(String const &name, unsigned long value) {
    return set(name, NumberValue(value));
}
Variable &Record::set(String const &name, Time const &value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(TimeValue(value));
    }
    return addTime(name, value);
}
Variable &Record::set(String const &name, Block const &value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(BlockValue(value));
    }
    Variable &var = addBlock(name);
    var.value<BlockValue>().block() = value;
    return var;
}
Variable &Record::set(const String &name, const Record &value) {
    DENG2_GUARD(d);
    std::unique_ptr<Record> dup(new Record(value));
    if (hasMember(name)) {
        return (*this)[name].set(RecordValue::takeRecord(dup.release()));
    }
    Variable &var = add(name);
    var.set(RecordValue::takeRecord(dup.release()));
    return var;
}
Variable &Record::set(String const &name, ArrayValue *value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(value);
    }
    return addArray(name, value);
}
Variable &Record::set(const String &name, Value *value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(value);
    }
    return add(name).set(value);
}
Variable &Record::set(const String &name, const Value &value) {
    DENG2_GUARD(d);
    if (hasMember(name)) {
        return (*this)[name].set(value);
    }
    return add(name).set(value);
}
Variable &Record::appendWord(String const &name, String const &word, String const &separator) {
    DENG2_GUARD(d);
    String value = gets(name, "");
    if (!value.isEmpty()) value.append(separator);
    set(name, value + word);
    return (*this)[name];
}
Variable &Record::appendUniqueWord(String const &name, String const &word, String const &separator) {
    DENG2_GUARD(d);
    String const value = gets(name, "");
    if (!value.containsWord(word)) {
        appendWord(name, word, separator);
    }
    return (*this)[name];
}
Variable &Record::appendMultipleUniqueWords(String const &name, String const &words, String const &separator) {
    foreach (String word, words.split(separator, QString::SkipEmptyParts)) {
        appendUniqueWord(name, word, separator);
    }
    return (*this)[name];
}
Variable &Record::appendToArray(String const &name, Value *value) {
    DENG2_GUARD(d);
    if (!has(name)) {
        return addArray(name, new ArrayValue({ value }));
    }
    Variable &var = (*this)[name];
    DENG2_ASSERT(is<ArrayValue>(var.value()));
    var.value<ArrayValue>().add(value);
    return var;
}
Variable &Record::insertToSortedArray(String const &name, Value *value) {
    DENG2_GUARD(d);
    if (!has(name)) {
        return appendToArray(name, value);
    }
    Variable &var = (*this)[name];
    ArrayValue &array = var.value<ArrayValue>();
    for (dsize i = 0; i < array.size(); ++i) {
        if (value->compare(array.at(i)) <= 0) {
            array.insert(i, value);
            return var;
        }
    }
    array.add(value);
    return var;
}
Variable &Record::operator [] (String const &name) {
    return const_cast<Variable &>((*const_cast<Record const *>(this))[name]);
}
Variable const &Record::operator [] (String const &name) const {
    Variable const *found = d->findMemberByPath(name);
    if (found) {
        return *found;
    }
    throw NotFoundError("Record::operator []", "Variable '" + name + "' not found");
}
Variable *Record::tryFind(String const &name) {
    return const_cast<Variable *>(d->findMemberByPath(name));
}
Variable const *Record::tryFind(String const &name) const {
    return d->findMemberByPath(name);
}
Record &Record::subrecord(String const &name) {
    return const_cast<Record &>((const_cast<Record const *>(this))->subrecord(name));
}
Record const &Record::subrecord(String const &name) const {
    int pos = name.indexOf('.');
    if (pos >= 0) {
        return subrecord(name.substr(0, pos)).subrecord(name.substr(pos + 1));
    }
    Members::const_iterator found = d->members.find(name);
    if (found != d->members.end() && d->isSubrecord(*found.value())) {
        return *found.value()->value<RecordValue>().record();
    }
    throw NotFoundError("Record::subrecord", "Subrecord '" + name + "' not found");
}
dsize Record::size() const {
    return dsize(d->members.size());
}
Record::Members const &Record::members() const {
    return d->members;
}
LoopResult Record::forMembers(std::function<LoopResult (String const &, Variable &)> func) {
    for (Members::iterator i = d->members.begin(); i != d->members.end(); ++i) {
        if (auto result = func(i.key(), *i.value())) {
            return result;
        }
    }
    return LoopContinue;
}
LoopResult Record::forMembers(std::function<LoopResult (String const &, Variable const &)> func) const {
    for (Members::const_iterator i = d->members.constBegin(); i != d->members.constEnd(); ++i) {
        if (auto result = func(i.key(), *i.value())) {
            return result;
        }
    }
    return LoopContinue;
}
Record::Subrecords Record::subrecords() const {
    return d->listSubrecords([] (Record const &) { return true; /* unfiltered */ });
}
Record::Subrecords Record::subrecords(std::function<bool (Record const &)> filter) const {
    return d->listSubrecords([&] (Record const &rec) {
        return filter(rec);
    });
}
LoopResult Record::forSubrecords(std::function<LoopResult (String const &, Record &)> func) {
    return d->forSubrecords(func);
}
LoopResult Record::forSubrecords(std::function<LoopResult (String const &, Record const &)> func) const {
    return d->forSubrecords([func] (String const &name, Record &rec) {
        return func(name, rec);
    });
}
bool Record::anyMembersChanged() const {
    DENG2_GUARD(d);
    auto const *const_d = d.getConst();
    for (auto i = const_d->members.begin(); i != const_d->members.end(); ++i) {
        if (i.value()->flags() & Variable::ValueHasChanged) {
            return true;
        }
    }
    return d->forSubrecords([](const String &, Record &rec) {
        if (rec.anyMembersChanged()) {
            return true;
        }
        return false;
    });
}
void Record::markAllMembersUnchanged() {
    DENG2_GUARD(d);
    for (auto i = d->members.begin(); i != d->members.end(); ++i) {
        i.value()->setFlags(Variable::ValueHasChanged, UnsetFlags);
        if (d->isSubrecord(*i.value())) {
            i.value()->valueAsRecord().markAllMembersUnchanged();
        }
    }
}
String Record::asText(String const &prefix, List *lines) const {
    DENG2_GUARD(d);
    /*
    if (!gets(VAR_FILE, "").isEmpty()) {
        return QString("(Record imported from \"%1\")").arg(gets(VAR_FILE));
    }*/
    if (lines) {
        for (Members::const_iterator i = d->members.begin(); i != d->members.end(); ++i) {
            String separator = (d->isSubrecord(*i.value())? "." : ":");
            String subContent = i.value()->value().asText();
            int numberOfLines = subContent.count(QChar('\n'));
            if (numberOfLines > SUBRECORD_CONTENT_EXCERPT_THRESHOLD) {
                subContent = QString("(%1 lines)").arg(numberOfLines);
            }
            KeyValue kv(prefix + i.key() + separator, subContent);
            lines->push_back(kv);
        }
        return "";
    }
    QString result;
    QTextStream os(&result);
    List allLines;
    Vector2ui maxLength;
    asText(prefix, &allLines);
    qSort(allLines);
    for (List::iterator i = allLines.begin(); i != allLines.end(); ++i) {
        maxLength = maxLength.max(Vector2ui(i->first.size(), i->second.size()));
    }
    os.setFieldAlignment(QTextStream::AlignLeft);
    for (List::iterator i = allLines.begin(); i != allLines.end(); ++i) {
        int extra = 0;
        if (i != allLines.begin()) os << "\n";
        os << qSetFieldWidth(maxLength.x) << i->first << qSetFieldWidth(0);
        int pos = 0;
        while (pos >= 0) {
            int next = i->second.indexOf('\n', pos);
            if (pos > 0) {
                os << qSetFieldWidth(maxLength.x + extra) << "" << qSetFieldWidth(0);
            }
            os << i->second.substr(pos, next != String::npos? next - pos + 1 : next);
            pos = next;
            if (pos != String::npos) pos++;
        }
    }
    return result;
}
Function const &Record::function(String const &name) const {
    return (*this)[name].value<FunctionValue>().function();
}
void Record::addSuperRecord(Value *superValue) {
    DENG2_GUARD(d);
    if (!has(VAR_SUPER)) {
        addArray(VAR_SUPER);
    }
    (*this)[VAR_SUPER].array().add(superValue);
}
void Record::addSuperRecord(Record const &superRecord) {
    addSuperRecord(new RecordValue(superRecord));
}
void Record::operator >> (Writer &to) const {
    DENG2_GUARD(d);
    to << d->uniqueId << duint32(d->members.size());
    DENG2_FOR_EACH_CONST(Members, i, d->members) {
        to << *i.value();
    }
}
void Record::operator << (Reader &from) {
    LOG_AS("Record deserialization");
    duint32 count = 0;
    from >> d->oldUniqueId >> count;
    clear();
    Impl::RefMap refMap;
    refMap.insert(d->oldUniqueId, this);
    while (count-- > 0) {
        QScopedPointer<Variable> var(new Variable());
        from >> *var;
        RecordValue *recVal = dynamic_cast<RecordValue *>(&var->value());
        if (recVal && recVal->usedToHaveOwnership()) {
            DENG2_ASSERT(recVal->record());
            refMap.insert(recVal->record()->d->oldUniqueId, recVal->record());
        }
        add(var.take());
    }
    d->reconnectReferencesAfterDeserialization(refMap);
#ifdef DENG2_DEBUG
    DENG2_FOR_EACH(Members, i, d->members) {
        DENG2_ASSERT(i.value()->audienceForDeletion().contains(d));
    }
#endif
}
Record &Record::operator << (NativeFunctionSpec const &spec) {
    addFunction(spec.name(), refless(spec.make())).setReadOnly();
    return *this;
}
Record const &Record::parentRecordForMember(String const &name) const {
    String const lastOmitted = name.fileNamePath('.');
    if (lastOmitted.isEmpty()) return *this;
    return (*this)[lastOmitted];
}
String Record::asInfo() const {
    String out;
    QTextStream os(&out);
    os.setCodec("UTF-8");
    for (auto i = d->members.constBegin(); i != d->members.constEnd(); ++i) {
        if (out) os << "\n";
        const Variable &var = *i.value();
        String src = i.key();
        if (is<RecordValue>(var.value())) {
            src += " {\n" + var.valueAsRecord().asInfo();
            src.replace("\n", "\n    ");
            src += "\n}";
        }
        else if (is<ArrayValue>(var.value())) {
            src += " " + var.value<ArrayValue>().asInfo();
        }
        else {
            String valueText = var.value().asText();
            if (valueText.contains("\n")) {
                src += " = " + Info::quoteString(var.value().asText());
            }
            else {
                src += ": " + valueText;
            }
        }
        os << src;
    }
    return out;
}
QTextStream &operator << (QTextStream &os, Record const &record) {
    return os << record.asText();
}
CompiledRecord::~CompiledRecord()
{}
} // namespace de
namespace de {
RecordAccessor::RecordAccessor(Record const *rec) : _rec(rec)
{}
RecordAccessor::RecordAccessor(Record const &rec) : _rec(&rec)
{}
Record const &RecordAccessor::accessedRecord() const {
    DENG2_ASSERT(_rec != 0);
    return *_rec;
}
Record const *RecordAccessor::accessedRecordPtr() const {
    return _rec;
}
bool RecordAccessor::has(String const &name) const {
    return accessedRecord().has(name);
}
Value const &RecordAccessor::get(String const &name) const {
    return accessedRecord()[name].value();
}
dint RecordAccessor::geti(String const &name) const {
    return get(name).asInt();
}
dint RecordAccessor::geti(String const &name, dint defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return geti(name);
}
bool RecordAccessor::getb(String const &name) const {
    return get(name).isTrue();
}
bool RecordAccessor::getb(String const &name, bool defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return getb(name);
}
duint RecordAccessor::getui(String const &name) const {
    return duint(get(name).asNumber());
}
duint RecordAccessor::getui(String const &name, duint defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return getui(name);
}
dfloat RecordAccessor::getf(String const &name) const {
    return dfloat(getd(name));
}
dfloat RecordAccessor::getf(String const &name, dfloat defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return getf(name);
}
ddouble RecordAccessor::getd(String const &name) const {
    return get(name).asNumber();
}
ddouble RecordAccessor::getd(String const &name, ddouble defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return getd(name);
}
String RecordAccessor::gets(String const &name) const {
    return get(name).asText();
}
String RecordAccessor::gets(String const &name, String const &defaultValue) const {
    if (!accessedRecord().hasMember(name)) return defaultValue;
    return gets(name);
}
ArrayValue const &RecordAccessor::geta(String const &name) const {
    return getAs<ArrayValue>(name);
}
DictionaryValue const &RecordAccessor::getdt(String const &name) const {
    return getAs<DictionaryValue>(name);
}
RecordValue const &RecordAccessor::getr(String const &name) const {
    return getAs<RecordValue>(name);
}
StringList RecordAccessor::getStringList(String const &name, StringList defaultValue) const {
    if (!accessedRecord().has(name)) return defaultValue;
    return get(name).asStringList();
}
Record const &RecordAccessor::subrecord(String const &name) const {
    return accessedRecord().subrecord(name);
}
void RecordAccessor::setAccessedRecord(Record const &rec) {
    _rec = &rec;
}
void RecordAccessor::setAccessedRecord(Record const *rec) {
    _rec = rec;
}
} // namespace de
namespace de {
DENG2_PIMPL(RecordValue)
, DENG2_OBSERVES(Record, Deletion) {
    Record *record = nullptr;
    OwnershipFlags ownership;
    OwnershipFlags oldOwnership; // prior to serialization
    Impl(Public *i) : Base(i) {}
    void recordBeingDeleted(Record &DENG2_DEBUG_ONLY(deleted)) {
        if (!record) return; // Not associated with a record any more.
        DENG2_ASSERT(record == &deleted);
        DENG2_ASSERT(!ownership.testFlag(OwnsRecord));
        record = nullptr;
        self().setAccessedRecord(nullptr);
    }
};
RecordValue::RecordValue(Record *record, OwnershipFlags o)
    : RecordAccessor(record)
    , d(new Impl(this)) {
    d->record = record;
    d->ownership = o;
    d->oldOwnership = o;
    DENG2_ASSERT(d->record != nullptr);
    if (!d->ownership.testFlag(OwnsRecord) &&
        !d->record->flags().testFlag(Record::WontBeDeleted)) {
        d->record->audienceForDeletion() += d;
    }
}
RecordValue::RecordValue(Record const &record)
    : RecordAccessor(record)
    , d(new Impl(this)) {
    d->record = const_cast<Record *>(&record);
    if (!record.flags().testFlag(Record::WontBeDeleted)) {
        d->record->audienceForDeletion() += d;
    }
}
RecordValue::RecordValue(IObject const &object)
    : RecordAccessor(object.objectNamespace())
    , d(new Impl(this)) {
    d->record = const_cast<Record *>(&object.objectNamespace());
    if (!d->record->flags().testFlag(Record::WontBeDeleted)) {
        d->record->audienceForDeletion() += d;
    }
}
RecordValue::~RecordValue() {
    setRecord(nullptr);
}
bool RecordValue::hasOwnership() const {
    return d->ownership.testFlag(OwnsRecord);
}
bool RecordValue::usedToHaveOwnership() const {
    return d->oldOwnership.testFlag(OwnsRecord);
}
Record *RecordValue::record() const {
    return d->record;
}
void RecordValue::setRecord(Record *record, OwnershipFlags ownership) {
    if (record == d->record) return; // Got it already.
    if (hasOwnership()) {
        delete d->record;
    }
    else if (d->record && !d->record->flags().testFlag(Record::WontBeDeleted)) {
        d->record->audienceForDeletion() -= d;
    }
    d->record = record;
    d->ownership = ownership;
    setAccessedRecord(d->record);
    if (d->record && !d->ownership.testFlag(OwnsRecord) &&
        !d->record->flags().testFlag(Record::WontBeDeleted)) {
        d->record->audienceForDeletion() += d;
    }
}
Record *RecordValue::takeRecord() {
    verify();
    if (!hasOwnership()) {
        throw OwnershipError("RecordValue::takeRecord", "Value does not own the record");
    }
    Record *rec = d->record;
    DENG2_ASSERT(!d->record->audienceForDeletion().contains(d));
    d->record = nullptr;
    d->ownership = RecordNotOwned;
    setAccessedRecord(nullptr);
    return rec;
}
void RecordValue::verify() const {
    if (!d->record) {
        throw NullError("RecordValue::verify", "Value no longer references a record");
    }
}
Record &RecordValue::dereference() {
    verify();
    return *d->record;
}
Record const &RecordValue::dereference() const {
    verify();
    return *d->record;
}
Value::Text RecordValue::typeId() const {
    return QStringLiteral("Record");
}
Value *RecordValue::duplicate() const {
    verify();
    if (hasOwnership()) {
        return new RecordValue(new Record(*d->record), OwnsRecord);
    }
    return new RecordValue(d->record);
}
Value *RecordValue::duplicateAsReference() const {
    verify();
    return new RecordValue(d->record);
}
Value::Text RecordValue::asText() const {
    if (d->record) {
        return d->record->asText();
    }
    return "(null)";
}
Record *RecordValue::memberScope() const {
    verify();
    return d->record;
}
dsize RecordValue::size() const {
    return dereference().members().size();
}
void RecordValue::setElement(Value const &index, Value *elementValue) {
    TextValue const *text = dynamic_cast<TextValue const *>(&index);
    if (!text) {
        throw IllegalIndexError("RecordValue::setElement",
                                "Records must be indexed with text values");
    }
    dereference().add(new Variable(text->asText(), elementValue));
}
Value *RecordValue::duplicateElement(Value const &value) const {
    TextValue const *text = dynamic_cast<TextValue const *>(&value);
    if (!text) {
        throw IllegalIndexError("RecordValue::duplicateElement",
                                "Records must be indexed with text values");
    }
    if (dereference().hasMember(*text)) {
        return dereference()[*text].value().duplicateAsReference();
    }
    throw NotFoundError("RecordValue::duplicateElement",
                        "'" + text->asText() + "' does not exist in the record");
}
bool RecordValue::contains(Value const &value) const {
    TextValue const *text = dynamic_cast<TextValue const *>(&value);
    if (!text) {
        throw IllegalIndexError("RecordValue::contains",
                                "Records must be indexed with text values");
    }
    return dereference().has(*text);
}
bool RecordValue::isTrue() const {
    return size() > 0;
}
dint RecordValue::compare(Value const &value) const {
    RecordValue const *recValue = dynamic_cast<RecordValue const *>(&value);
    if (!recValue) {
        return cmp(reinterpret_cast<void const *>(this),
                   reinterpret_cast<void const *>(&value));
    }
    return cmp(recValue->d->record, d->record);
}
void RecordValue::call(Process &process, Value const &arguments, Value *) const {
    verify();
    QScopedPointer<RecordValue> instance(new RecordValue(new Record, RecordValue::OwnsRecord));
    instance->record()->addSuperRecord(*d->record);
    if (dereference().hasMember(Record::VAR_INIT)) {
        process.call(dereference().function(Record::VAR_INIT), arguments.as<ArrayValue>(),
                     instance->duplicateAsReference());
        delete process.context().evaluator().popResult();
    }
    process.context().evaluator().pushResult(instance.take());
}
static duint8 const OWNS_RECORD = 0x1;
void RecordValue::operator >> (Writer &to) const {
    duint8 flags = 0;
    if (hasOwnership()) flags |= OWNS_RECORD;
    to << SerialId(RECORD) << flags << dereference();
}
void RecordValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != RECORD) {
        throw DeserializationError("RecordValue::operator <<", "Invalid ID");
    }
    duint8 flags = 0;
    from >> flags;
    d->oldOwnership = OwnershipFlags(flags & OWNS_RECORD? OwnsRecord : 0);
    from >> dereference();
}
RecordValue *RecordValue::takeRecord(Record *record) {
    return new RecordValue(record, OwnsRecord);
}
RecordValue *RecordValue::takeRecord(Record &&record) {
    return new RecordValue(new Record(record), OwnsRecord);
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Refuge) {
    String persistentPath;
    Record names;
};
Refuge::Refuge(String const &persistentPath) : d(new Impl) {
    d->persistentPath = persistentPath;
    try {
        read();
    }
    catch (Error const &er) {
        LOG_AS("Refuge");
        LOGDEV_RES_MSG("\"%s\" could not be read: %s") << persistentPath << er.asText();
    }
}
String Refuge::path() const {
    return d->persistentPath;
}
Refuge::~Refuge() {
    try {
        write();
    }
    catch (Error const &er) {
        LOG_AS("~Refuge");
        LOG_ERROR("\"%s\" could not be written: %s") << d->persistentPath << er.asText();
    }
}
void Refuge::read() {
    if (App::hasPersistentData()) {
        Reader(App::persistentData().entryBlock(d->persistentPath)).withHeader() >> d->names;
        d->names.markAllMembersUnchanged();
    }
}
void Refuge::write() const {
    if (App::hasPersistentData()) {
        d->names.markAllMembersUnchanged();
        Writer(App::mutablePersistentData().entryBlock(d->persistentPath)).withHeader()
            << d->names;
        App::persistPackFolder().flush();
    }
}
Time Refuge::lastWrittenAt() const {
    if (App::hasPersistentData()) {
        return App::persistentData().entryStatus(d->persistentPath).modifiedAt;
    }
    return Time::invalidTime();
}
bool Refuge::hasModifiedVariables() const {
    return d->names.anyMembersChanged();
}
Record &Refuge::objectNamespace() {
    return d->names;
}
Record const &Refuge::objectNamespace() const {
    return d->names;
}
} // namespace de
namespace de {
RefValue::RefValue(Variable *variable) : _variable(variable) {
    if (_variable) {
        _variable->audienceForDeletion() += this;
    }
}
RefValue::~RefValue()
{}
void RefValue::verify() const {
    if (!_variable) {
        throw NullError("RefValue::verify", "Value does not reference a variable");
    }
}
Value &RefValue::dereference() {
    verify();
    return _variable->value();
}
Value const &RefValue::dereference() const {
    verify();
    return _variable->value();
}
Value *RefValue::duplicate() const {
    return new RefValue(_variable);
}
Value::Number RefValue::asNumber() const {
    return dereference().asNumber();
}
Value::Text RefValue::asText() const {
    return dereference().asText();
}
Record *RefValue::memberScope() const {
    return dereference().memberScope();
}
dsize RefValue::size() const {
    return dereference().size();
}
Value const &RefValue::element(Value const &index) const {
    return dereference().element(index);
}
Value &RefValue::element(Value const &index) {
    return dereference().element(index);
}
void RefValue::setElement(Value const &index, Value *elementValue) {
    dereference().setElement(index, elementValue);
}
bool RefValue::contains(Value const &value) const {
    return dereference().contains(value);
}
Value *RefValue::begin() {
    return dereference().begin();
}
Value *RefValue::next() {
    return dereference().next();
}
bool RefValue::isTrue() const {
    return dereference().isTrue();
}
bool RefValue::isFalse() const {
    return dereference().isFalse();
}
dint RefValue::compare(Value const &value) const {
    return dereference().compare(value);
}
void RefValue::negate() {
    dereference().negate();
}
void RefValue::sum(Value const &value) {
    dereference().sum(value);
}
void RefValue::subtract(Value const &subtrahend) {
    dereference().subtract(subtrahend);
}
void RefValue::divide(Value const &divisor) {
    dereference().divide(divisor);
}
void RefValue::multiply(Value const &value) {
    dereference().multiply(value);
}
void RefValue::modulo(Value const &divisor) {
    dereference().modulo(divisor);
}
void RefValue::assign(Value *value) {
    verify();
    _variable->set(value);
}
void RefValue::call(Process &process, Value const &arguments, Value *self) const {
    dereference().call(process, arguments, self);
}
void RefValue::operator >> (Writer &to) const {
    to << dereference();
}
void RefValue::operator << (Reader &/*from*/)
{
    DENG2_ASSERT(false);
}
void RefValue::variableBeingDeleted(Variable &DENG2_DEBUG_ONLY(variable)) {
    DENG2_ASSERT(_variable == &variable);
    _variable = 0;
}
Value::Text RefValue::typeId() const {
    return "Ref";
}
} // namespace de
namespace de {
static int const SOURCE_SHIFT = 17;
static int const NUMBER_MASK = 0x1ffff;
DENG2_PIMPL_NOREF(SourceLineTable), public Lockable {
    struct IdNode : public PathTree::Node {
        static std::atomic_uint counter;
        LineId id;
        IdNode(PathTree::NodeArgs const &args)
            : PathTree::Node(args), id(++counter) {}
    };
    PathTreeT<IdNode> paths;
    QHash<duint, IdNode const *> lookup; // reverse lookup
};
std::atomic_uint SourceLineTable::Impl::IdNode::counter;
SourceLineTable::SourceLineTable() : d(new Impl)
{}
SourceLineTable::LineId SourceLineTable::lineId(String const &path, duint lineNumber) {
    Path const source(path);
    DENG2_GUARD(d);
    auto const *node = d->paths.tryFind(source, PathTree::MatchFull | PathTree::NoBranch);
    if (!node) {
        node = &d->paths.insert(source);
        d->lookup[node->id] = node;
    }
    return (node->id << SOURCE_SHIFT) | lineNumber;
}
String SourceLineTable::sourceLocation(LineId sourceId) const {
    auto const location = sourcePathAndLineNumber(sourceId);
    return QString("%1:%2").arg(location.first).arg(location.second);
}
SourceLineTable::PathAndLine SourceLineTable::sourcePathAndLineNumber(LineId sourceId) const {
    duint const lineNumber = (NUMBER_MASK & sourceId);
    DENG2_GUARD(d);
    auto found = d->lookup.constFind(sourceId >> SOURCE_SHIFT);
    if (found != d->lookup.constEnd()) {
        return PathAndLine(found.value()->path().toStringRef(), lineNumber);
    }
    return PathAndLine("", lineNumber);
}
} // namespace de
namespace de {
String::size_type const String::npos = -1;
String::String()
{}
String::String(String const &other) : QString(other)
{}
String::String(String &&moved) : QString(std::move(moved))
{}
String::String(QString const &text) : QString(text)
{}
#ifdef DENG2_QT_4_7_OR_NEWER
String::String(QChar const *nullTerminatedStr)
    : QString(nullTerminatedStr)
{}
#else
String::String(QChar const *nullTerminatedStr)
    : QString(nullTerminatedStr, qchar_strlen(nullTerminatedStr))
{}
#endif
String::String(QChar const *str, size_type length) : QString(str, length)
{}
String::String(char const *nullTerminatedCStr)
    : QString(QString::fromUtf8(nullTerminatedCStr))
{}
String::String(wchar_t const *nullTerminatedWideStr)
    : QString(QString::fromWCharArray(nullTerminatedWideStr))
{}
String::String(char const *cStr, size_type length)
    : QString(QString::fromUtf8(cStr, length))
{}
String::String(char const *cStr, dsize length)
    : QString(QString::fromUtf8(cStr, int(length)))
{}
String::String(size_type length, QChar ch) : QString(length, ch)
{}
String::String(QString const &str, size_type index, size_type length)
    : QString(str.mid(index, length))
{}
String::String(const_iterator start, const_iterator end) {
    for (const_iterator i = start; i < end; ++i) {
        append(*i);
    }
}
String::String(iterator start, iterator end) {
    for (iterator i = start; i < end; ++i) {
        append(*i);
    }
}
QChar String::first() const {
    if (empty()) return 0;
    return at(0);
}
QChar String::last() const {
    if (empty()) return 0;
    return at(size() - 1);
}
String String::operator / (String const &path) const {
    return concatenatePath(path);
}
String String::operator % (PatternArgs args) const {
    String result;
    QTextStream output(&result);
    PatternArgs::const_iterator arg = args.begin();
    DENG2_FOR_EACH_CONST(String, i, *this) {
        if (*i == '%') {
            String::const_iterator next = i;
            advanceFormat(next, end());
            if (*next == '%') {
                output << *next;
                ++i;
                continue;
            }
            if (arg == args.end()) {
                throw IllegalPatternError("String::operator %", "Ran out of arguments");
            }
            output << patternFormat(i, end(), **arg);
            ++arg;
        }
        else {
            output << *i;
        }
    }
    for (; arg != args.end(); ++arg) {
        output << (*arg)->asText();
    }
    return result;
}
String String::concatenatePath(String const &other, QChar dirChar) const {
    if ((dirChar == '/' || dirChar == '\\') && QDir::isAbsolutePath(other)) {
        return other;
    }
    return concatenateRelativePath(other, dirChar);
}
String String::concatenateRelativePath(const String &other, QChar dirChar) const {
    if (other.isEmpty()) return *this;
    int const startPos = (other.first() == dirChar? 1 : 0);
    String result = *this;
    if (!empty() && last() != dirChar) {
        result += dirChar;
    }
    result += other.substr(startPos);
    return result;
}
String String::concatenateMember(String const &member) const {
    if (member.isEmpty()) return *this;
    if (member.first() == '.') {
        throw InvalidMemberError("String::concatenateMember", "Invalid: '" + member + "'");
    }
    return concatenatePath(member, '.');
}
String String::strip() const {
    return trimmed();
}
String String::leftStrip() const {
    int endOfSpace = 0;
    while (endOfSpace < size() && at(endOfSpace).isSpace()) {
        ++endOfSpace;
    }
    return mid(endOfSpace);
}
String String::rightStrip() const {
    int beginOfSpace = size() - 1;
    while (beginOfSpace >= 0 && at(beginOfSpace).isSpace()) {
        --beginOfSpace;
    }
    return left(beginOfSpace + 1);
}
String String::normalizeWhitespace() const {
    static QRegularExpression const reg("\\s+");
    String s = *this;
    s.replace(reg, " ");
    return s.strip();
}
String String::removed(const QRegularExpression &expr) const {
    String s = *this;
    s.remove(expr);
    return s;
}
String String::lower() const {
    return toLower();
}
String String::upper() const {
    return toUpper();
}
String String::upperFirstChar() const {
    if (isEmpty()) return "";
    if (size() == 1) return upper();
    return at(0).toUpper() + substr(1);
}
/*
std::wstring String::stringToWide(String const &str) {
    duint inputSize = str.size();
    dbyte const *input = reinterpret_cast<dbyte const *>(str.c_str());
    dbyte const *inputEnd = input + inputSize;
    std::vector<wchar_t> output;
    duint c;
    duint d;
    duint trailing;
    while (input < inputEnd) {
        d = *input++;
        if (d < 0x80) {
            c = d;
            trailing = 0;
        }
        else if (d < 0xc0) {
            throw ConversionError("String::stringToWide", "Trailing byte in leading position");
        }
        else if (d < 0xe0) {
            c = d & 0x1f;
            trailing = 1;
        }
        else if (d < 0xf0) {
            c = d & 0x0f;
            trailing = 2;
        }
        else if (d < 0xf8) {
            c = d & 0x07;
            trailing = 3;
        }
        else {
            throw ConversionError("String::stringToWide", "Bad input");
        }
        for (; trailing; trailing--) {
            if ((input >= inputEnd) || (((d = *input++) & 0xc0) != 0x80)) {
                throw ConversionError("String::stringToWide", "Bad input");
            }
            c <<= 6;
            c |= d & 0x3f;
        }
        if (c < 0x10000) {
            output.push_back(wchar_t(c));
        }
        else if (c < 0x110000) {
            c -= 0x10000;
            output.push_back(wchar_t(0xd800 | (c >> 10)));
            output.push_back(0xdc00 | (c & 0x03ff));
        }
        else {
            throw ConversionError("String::stringToWide", "Bad input");
        }
    }
    output.push_back(0);
    return std::wstring(&output[0]);
}
String String::wideToString(std::wstring const &str) {
    const wchar_t *input = &str[0];
    const wchar_t *inputEnd = input + str.size();
    String output;
    duint c;
    duint d;
    dint bits;
    while (input < inputEnd) {
        c = *input++;
        if ((c & 0xfc00) == 0xd800) {
            if ((input < inputEnd) && (((d = *input++) & 0xfc00) == 0xdc00)) {
                c &= 0x3ff;
                c <<= 10;
                c |= d & 0x3ff;
                c += 0x10000;
            }
            else {
                throw ConversionError("String::wideToString", "Bad input");
            }
        }
        if (c < 0x80) {
            output += dchar(c);
            bits = -6;
        }
        else if (c < 0x800) {
            output += dchar((c >> 6) | 0xc0);
            bits = 0;
        }
        else if (c < 0x10000) {
            output += dchar((c >> 12) | 0xe0);
            bits = 6;
        }
        else {
            output += dchar((c >> 18) | 0xf0);
            bits = 12;
        }
        for (; bits > 0; bits -= 6) {
            output += dchar((c >> bits) & 0x3f);
        }
    }
    return output;
}
*/
String String::fileName(QChar dirChar) const {
    size_type pos = lastIndexOf(dirChar);
    if (pos >= 0) {
        return mid(pos + 1);
    }
    return *this;
}
String String::fileNameWithoutExtension() const {
    String name = fileName();
    size_type pos = name.lastIndexOf('.');
    if (pos > 0) {
        return name.mid(0, pos);
    }
    return name;
}
String String::fileNameExtension() const {
    size_type pos = lastIndexOf('.');
    size_type slashPos = lastIndexOf('/');
    if (pos > 0) {
        if (slashPos < 0 || pos > slashPos + 1) {
            return mid(pos);
        }
    }
    return "";
}
String String::fileNamePath(QChar dirChar) const {
    size_type pos = lastIndexOf(dirChar);
    if (pos >= 0) {
        return mid(0, pos);
    }
    return "";
}
String String::fileNameAndPathWithoutExtension(QChar dirChar) const {
    return fileNamePath(dirChar) / fileNameWithoutExtension();
}
bool String::containsWord(String const &word) const {
    if (word.isEmpty()) {
        return false;
    }
    return QRegularExpression(QString("\\b%1\\b").arg(word)).match(*this).hasMatch();
}
dint String::compareWithCase(String const &str) const {
    return compare(str, Qt::CaseSensitive);
}
dint String::compareWithoutCase(String const &str) const {
    return compare(str, Qt::CaseInsensitive);
}
dint String::compareWithoutCase(String const &str, int n) const {
    return leftRef(n).compare(str.leftRef(n), Qt::CaseInsensitive);
}
int String::commonPrefixLength(String const &str, CaseSensitivity sensitivity) const {
    int count = 0;
    int len = qMin(str.size(), size());
    for (int i = 0; i < len; ++i, ++count) {
        if (sensitivity == CaseSensitive) {
            if (at(i) != str.at(i)) break;
        }
        else {
            if (at(i).toLower() != str.at(i).toLower()) break;
        }
    }
    return count;
}
dint String::compareWithCase(QChar const *a, QChar const *b, dsize count) // static
{
    return QString(a).leftRef(count).compare(QString(b).leftRef(count), Qt::CaseSensitive);
}
bool String::equals(QChar const *a, QChar const *b, dsize count) // static
{
    while (count--) {
        if (a->isNull() && b->isNull()) break;
        if (*a != *b) return false;
        a++;
        b++;
    }
    return true;
}
void String::skipSpace(String::const_iterator &i, String::const_iterator const &end) {
    while (i != end && (*i).isSpace()) ++i;
}
String String::format(String format, ...) {
    va_list args;
    Block buffer;
    int neededSize = 1024;
    forever {
        buffer.resize(neededSize);
        va_start(args, format);
        int count = vsnprintf((char *) buffer.data(), buffer.size(), format.toUtf8(), args);
        va_end(args);
        if (count >= 0 && count < neededSize) {
            buffer.resize(count);
            return fromUtf8(buffer);
        }
        if (count >= 0) {
            neededSize = count + 1;
        }
        else {
            neededSize *= 2; // Try bigger.
        }
    }
    return fromUtf8(buffer);
}
static inline bool isSign(QChar const &ch) {
    return ch == '-' || ch == '+';
}
dint String::toInt(bool *ok, int base, IntConversionFlags flags) const {
    String token = leftStrip();
    if (flags & AllowSuffix) {
        int endOfNumber = 0;
        while (endOfNumber < token.size() &&
              (token.at(endOfNumber).isDigit() || (endOfNumber == 0 && isSign(token.at(endOfNumber))) ||
               ((base == 0 || base == 16) && endOfNumber <= 1 &&
                (token.at(endOfNumber) == QChar('x') || token.at(endOfNumber) == QChar('X'))))) {
            ++endOfNumber;
        }
        token.truncate(endOfNumber);
    }
    return token.QString::toInt(ok, base);
}
duint32 String::toUInt32(bool *ok, int base) const {
    return QString::toUInt(ok, base);
}
String String::addLinePrefix(String const &prefix) const {
    String result;
    for (auto const &str : QString::split(QChar('\n'))) {
        if (!result.isEmpty()) result += "\n";
        result += prefix + str;
    }
    return result;
}
String String::escaped() const {
    String esc = *this;
    esc.replace("\\", "\\\\")
       .replace("\"", "\\\"");
    return esc;
}
Block String::toPercentEncoding() const {
    return QUrl::toPercentEncoding(*this);
}
String String::truncateWithEllipsis(int maxLength) const {
    if (size() <= maxLength) {
        return *this;
    }
    return left(maxLength/2 - 1) + "..." + right(maxLength/2 - 1);
}
void String::advanceFormat(String::const_iterator &i, String::const_iterator const &end) {
    ++i;
    if (i == end) {
        throw IllegalPatternError("String::advanceFormat",
                                  "Incomplete formatting instructions");
    }
}
String String::join(QList<String> const &stringList, String const &sep) {
    if (stringList.isEmpty()) return "";
    String joined;
    QTextStream os(&joined);
    os << stringList.at(0);
    for (int i = 1; i < stringList.size(); ++i) {
        os << sep << stringList.at(i);
    }
    return joined;
}
String String::patternFormat(String::const_iterator &formatIter,
                             String::const_iterator const &formatEnd,
                             IPatternArg const &arg) {
    advanceFormat(formatIter, formatEnd);
    QString result;
    QTextStream output(&result);
    bool rightAlign = true;
    dint maxWidth = 0;
    dint minWidth = 0;
    DENG2_ASSERT(*formatIter != '%');
    if (*formatIter == '-') {
        rightAlign = false;
        advanceFormat(formatIter, formatEnd);
    }
    String::const_iterator k = formatIter;
    while ((*formatIter).isDigit()) {
        advanceFormat(formatIter, formatEnd);
    }
    if (k != formatIter) {
        minWidth = String(k, formatIter).toInt();
    }
    if (*formatIter == '.') {
        advanceFormat(formatIter, formatEnd);
        k = formatIter;
        while ((*formatIter).isDigit()) {
            advanceFormat(formatIter, formatEnd);
        }
        maxWidth = String(k, formatIter).toInt();
    }
    switch ((*formatIter).toLatin1()) {
    case 's':
        output << arg.asText();
        break;
    case 'b':
        output << (int(arg.asNumber())? "true" : "false");
        break;
    case 'c':
        output << QChar(ushort(arg.asNumber()));
        break;
    case 'i':
    case 'd':
        output << dint64(arg.asNumber());
        break;
    case 'u':
        output << duint64(arg.asNumber());
        break;
    case 'X':
        output << uppercasedigits;
    case 'x':
        output << "0x" << hex << dint64(arg.asNumber()) << dec << lowercasedigits;
        break;
    case 'p':
        output << "0x" << hex << dintptr(arg.asNumber()) << dec;
        break;
    case 'f':
        output << fixed << qSetRealNumberPrecision(maxWidth? maxWidth : 3) << arg.asNumber();
        maxWidth = 0;
        break;
    default:
        throw IllegalPatternError("Log::Entry::str",
            "Unknown format character '" + String(1, *formatIter) + "'");
    }
    output.flush();
    if (maxWidth && result.size() > maxWidth) {
        result = result.mid(!rightAlign? 0 : result.size() - maxWidth, maxWidth);
    }
    if (result.size() < minWidth) {
        String padding = String(minWidth - result.size(), ' ');
        if (rightAlign) {
            result = padding + result;
        }
        else {
            result += padding;
        }
    }
    return result;
}
Block String::toUtf8() const {
    return QString::toUtf8();
}
Block String::toLatin1() const {
    return QString::toLatin1();
}
String String::fromUtf8(IByteArray const &byteArray) {
    Block const bytes(byteArray);
    return QString::fromUtf8(reinterpret_cast<char const *>(bytes.data()), int(bytes.size()));
}
String String::fromUtf8(QByteArray const &byteArray) {
    return QString::fromUtf8(byteArray);
}
String String::fromUtf8(Block const &block) {
    return QString::fromUtf8(block);
}
String String::fromUtf8(char const *nullTerminatedCStr) {
    return QString::fromUtf8(nullTerminatedCStr);
}
String String::fromLatin1(IByteArray const &byteArray) {
    return QString::fromLatin1(reinterpret_cast<char const *>(Block(byteArray).data()));
}
String String::fromCP437(IByteArray const &byteArray) {
    Block const chars(byteArray);
    String conv;
    conv.reserve(byteArray.size());
    for (dbyte ch : chars) {
        conv.append(codePage437ToUnicode(ch));
    }
    return conv;
}
String String::fromPercentEncoding(Block const &percentEncoded) // static
{
    return QUrl::fromPercentEncoding(percentEncoded);
}
size_t qchar_strlen(QChar const *str) {
    if (!str) return 0;
    size_t len = 0;
    while (str->unicode() != 0) { ++str; ++len; }
    return len;
}
} // namespace de
namespace de {
static String const nullString = "(nullptr)";
typedef uint InternalId;
/**
 * Case-insensitive text string (String).
 */
class CaselessString : public ISerializable {
public:
    CaselessString()
        : _str(), _id(0), _userValue(0), _userPointer(0)
    {}
    CaselessString(QString text)
        : _str(text), _id(0), _userValue(0), _userPointer(0)
    {}
    CaselessString(CaselessString const &other)
        : ISerializable(), _str(other._str), _id(other._id), _userValue(other._userValue), _userPointer(0)
    {}
    void setText(String &text) {
        _str = text;
    }
    operator String const *() const {
        return &_str;
    }
    operator String const &() const {
        return _str;
    }
    bool operator < (CaselessString const &other) const {
        return _str.compare(other, Qt::CaseInsensitive) < 0;
    }
    bool operator == (CaselessString const &other) const {
        return !_str.compare(other, Qt::CaseInsensitive);
    }
    InternalId id() const {
        return _id;
    }
    void setId(InternalId i) {
        _id = i;
    }
    uint userValue() const {
        return _userValue;
    }
    void setUserValue(uint value) {
        _userValue = value;
    }
    void *userPointer() const {
        return _userPointer;
    }
    void setUserPointer(void *ptr) {
        _userPointer = ptr;
    }
    void operator >> (Writer &to) const {
        to << _str << duint32(_id) << duint32(_userValue);
    }
    void operator << (Reader &from) {
        from >> _str >> _id >> _userValue;
    }
private:
    String _str;
    InternalId _id; ///< The id that refers to this string.
    uint _userValue;
    void *_userPointer;
};
/**
 * Utility class that acts as the value type for the std::set of interned
 * strings. Only points to CaselessString instances. The less-than operator
 * is needed by std::set to keep the strings ordered.
 */
class CaselessStringRef {
public:
    CaselessStringRef(CaselessString const *s = 0) {
        _str = s;
    }
    CaselessStringRef(CaselessStringRef const &other) {
        _str = other._str;
    }
    CaselessString *toStr() const {
        return const_cast<CaselessString *>(_str);
    }
    InternalId id() const {
        DENG2_ASSERT(_str);
        return _str->id();
    }
    bool operator < (CaselessStringRef const &other) const {
        DENG2_ASSERT(_str);
        DENG2_ASSERT(other._str);
        return *_str < *other._str;
    }
    bool operator == (CaselessStringRef const &other) const {
        DENG2_ASSERT(_str);
        DENG2_ASSERT(other._str);
        return *_str == *other._str;
    }
private:
    CaselessString const *_str;
};
typedef std::set<CaselessStringRef> Interns;
typedef std::vector<CaselessString *> IdMap;
typedef std::list<InternalId> AvailableIds;
DENG2_PIMPL_NOREF(StringPool), public Lockable {
    Interns interns;
    IdMap idMap;
    dsize count;
    AvailableIds available;
    Impl() : count(0)
    {}
    ~Impl() {
        clear();
    }
    void clear() {
        DENG2_GUARD(this);
        for (dsize i = 0; i < idMap.size(); ++i) {
            if (!idMap[i]) continue; // Unused slot.
            delete idMap[i];
        }
        count = 0;
        interns.clear();
        idMap.clear();
        available.clear();
        assertCount();
    }
    inline void assertCount() const {
        DENG2_ASSERT(count == interns.size());
        DENG2_ASSERT(count == idMap.size() - available.size());
    }
    Interns::iterator findIntern(String text) {
        CaselessString const key(text);
        return interns.find(CaselessStringRef(&key)); // O(log n)
    }
    Interns::const_iterator findIntern(String text) const {
        CaselessString const key(text);
        return interns.find(CaselessStringRef(&key)); // O(log n)
    }
    /**
     * Before this is called make sure there is no duplicate of @a text in
     * the interns set.
     *
     * @param text  Text string to add to the interned strings. A copy is
     *              made of this.
     */
    InternalId copyAndAssignUniqueId(String const &text) {
        CaselessString *str = new CaselessString(text);
        interns.insert(str); // O(log n)
        return assignUniqueId(str);
    }
    InternalId assignUniqueId(CaselessString *str) // O(1)
    {
        InternalId idx;
        if (!available.empty()) // O(1)
        {
            idx = available.front();
            available.pop_front();
            idMap[idx] = str;
        }
        else {
            if (idMap.size() >= MAXIMUM_VALID_ID) {
                throw StringPool::FullError("StringPool::assignUniqueId",
                                            "Out of valid 32-bit identifiers");
            }
            idx = InternalId(idMap.size());
            idMap.push_back(str); // O(1) (amortized)
        }
        str->setId(idx);
        count++;
        assertCount();
        return idx;
    }
    void releaseAndDestroy(InternalId id, Interns::iterator *iterToErase = 0) {
        DENG2_ASSERT(id < idMap.size());
        CaselessString *interned = idMap[id];
        DENG2_ASSERT(interned != 0);
        idMap[id] = 0;
        available.push_back(id);
        delete interned;
        if (iterToErase) interns.erase(*iterToErase); // O(1) (amortized)
        count--;
        assertCount();
    }
};
StringPool::StringPool() : d(new Impl)
{}
StringPool::StringPool(String const *strings, uint count) : d(new Impl) {
    DENG2_GUARD(d);
    for (uint i = 0; strings && i < count; ++i) {
        intern(strings[i]);
    }
}
void StringPool::clear() {
    d->clear();
}
bool StringPool::empty() const {
    DENG2_GUARD(d);
    d->assertCount();
    return !d->count;
}
dsize StringPool::size() const {
    DENG2_GUARD(d);
    d->assertCount();
    return d->count;
}
StringPool::Id StringPool::intern(String str) {
    DENG2_GUARD(d);
    Interns::iterator found = d->findIntern(str); // O(log n)
    if (found != d->interns.end()) {
        return EXPORT_ID(found->id());
    }
    return EXPORT_ID(d->copyAndAssignUniqueId(str)); // O(log n)
}
String StringPool::internAndRetrieve(String str) {
    DENG2_GUARD(d);
    InternalId id = IMPORT_ID(intern(str));
    return *d->idMap[id];
}
void StringPool::setUserValue(Id id, uint value) {
    if (id == 0) return;
    InternalId const internalId = IMPORT_ID(id);
    DENG2_GUARD(d);
    DENG2_ASSERT(internalId < d->idMap.size());
    DENG2_ASSERT(d->idMap[internalId] != 0);
    d->idMap[internalId]->setUserValue(value); // O(1)
}
uint StringPool::userValue(Id id) const {
    if (id == 0) return 0;
    InternalId const internalId = IMPORT_ID(id);
    DENG2_GUARD(d);
    DENG2_ASSERT(internalId < d->idMap.size());
    DENG2_ASSERT(d->idMap[internalId] != 0);
    return d->idMap[internalId]->userValue(); // O(1)
}
void StringPool::setUserPointer(Id id, void *ptr) {
    if (id == 0) return;
    InternalId const internalId = IMPORT_ID(id);
    DENG2_GUARD(d);
    DENG2_ASSERT(internalId < d->idMap.size());
    DENG2_ASSERT(d->idMap[internalId] != 0);
    d->idMap[internalId]->setUserPointer(ptr); // O(1)
}
void *StringPool::userPointer(Id id) const {
    if (id == 0) return NULL;
    InternalId const internalId = IMPORT_ID(id);
    DENG2_GUARD(d);
    DENG2_ASSERT(internalId < d->idMap.size());
    DENG2_ASSERT(d->idMap[internalId] != 0);
    return d->idMap[internalId]->userPointer(); // O(1)
}
StringPool::Id StringPool::isInterned(String str) const {
    DENG2_GUARD(d);
    Interns::const_iterator found = d->findIntern(str); // O(log n)
    if (found != d->interns.end()) {
        return EXPORT_ID(found->id());
    }
    return 0;
}
String StringPool::string(Id id) const {
    DENG2_GUARD(d);
    return stringRef(id);
}
String const &StringPool::stringRef(StringPool::Id id) const {
    if (id == 0) {
        static String emptyString;
        return emptyString;
    }
    DENG2_GUARD(d);
    InternalId const internalId = IMPORT_ID(id);
    DENG2_ASSERT(internalId < d->idMap.size());
    return *d->idMap[internalId];
}
bool StringPool::remove(String str) {
    DENG2_GUARD(d);
    Interns::iterator found = d->findIntern(str); // O(log n)
    if (found != d->interns.end()) {
        d->releaseAndDestroy(found->id(), &found); // O(1) (amortized)
        return true;
    }
    return false;
}
bool StringPool::removeById(Id id) {
    if (id == 0) return false;
    DENG2_GUARD(d);
    InternalId const internalId = IMPORT_ID(id);
    if (id >= d->idMap.size()) return false;
    CaselessString *str = d->idMap[internalId];
    if (!str) return false;
    d->interns.erase(str); // O(log n)
    d->releaseAndDestroy(str->id());
    return true;
}
LoopResult StringPool::forAll(std::function<LoopResult (Id)> func) const {
    DENG2_GUARD(d);
    for (duint i = 0; i < d->idMap.size(); ++i) {
        if (d->idMap[i]) {
            if (auto result = func(EXPORT_ID(i)))
                return result;
        }
    }
    return LoopContinue;
}
void StringPool::operator >> (Writer &to) const {
    DENG2_GUARD(d);
    to << duint32(d->idMap.size());
    to << duint32(d->interns.size());
    for (Interns::const_iterator i = d->interns.begin(); i != d->interns.end(); ++i) {
        to << *i->toStr();
    }
}
void StringPool::operator << (Reader &from) {
    DENG2_GUARD(d);
    clear();
    uint numStrings;
    from >> numStrings;
    d->idMap.resize(numStrings, 0);
    uint numInterns;
    from >> numInterns;
    while (numInterns--) {
        CaselessString *str = new CaselessString;
        from >> *str;
        d->interns.insert(str);
        d->idMap[str->id()] = str;
        d->count++;
    }
    for (uint i = 0; i < d->idMap.size(); ++i) {
        if (!d->idMap[i]) d->available.push_back(i);
    }
    d->assertCount();
}
#ifdef DENG2_DEBUG
void StringPool::print() const {
    static dint const padding = 2 + 5/*numDigits*/;
    fprintf(stderr, "StringPool [%p]\n    idx    id string\n", (void *)this);
    duint count = 0;
    forAll([this, &count] (Id id) {
        QByteArray strUtf8 = string(id).toUtf8();
        fprintf(stderr, "%*u %5u %s\n", padding, count++, id, strUtf8.constData());
        return LoopContinue;
    });
    fprintf(stderr, "  There is %u %s in the pool.\n", duint( size() ),
            size() == 1 ? "string" : "strings");
}
#endif
}  // namespace de
namespace de {
using std::list;
TextValue::TextValue(String const &initialValue)
    : _value(initialValue)
{}
TextValue::operator String const &() const {
    return _value;
}
Value *TextValue::duplicate() const {
    return new TextValue(_value);
}
Value::Number TextValue::asNumber() const {
    return _value.toDouble();
}
Value::Text TextValue::asText() const {
    return _value;
}
Record *TextValue::memberScope() const {
    return &ScriptSystem::builtInClass("String");
}
dsize TextValue::size() const {
    return _value.size();
}
bool TextValue::contains(Value const &value) const {
    if (is<TextValue>(value)) {
        return _value.indexOf(value.asText(), Qt::CaseSensitive) >= 0;
    }
    return Value::contains(value);
}
bool TextValue::isTrue() const {
    for (Text::const_iterator i = _value.begin(); i != _value.end(); ++i) {
        if (!(*i).isSpace())
            return true;
    }
    return false;
}
dint TextValue::compare(Value const &value) const {
    TextValue const *other = dynamic_cast<TextValue const *>(&value);
    if (other) {
        return _value.compare(other->_value);
    }
    return Value::compare(value);
}
void TextValue::sum(Value const &value) {
    TextValue const *other = dynamic_cast<TextValue const *>(&value);
    if (!other) {
        throw ArithmeticError("TextValue::sum", "Value cannot be summed");
    }
    _value += other->_value;
}
void TextValue::multiply(Value const &value) {
    NumberValue const *other = dynamic_cast<NumberValue const *>(&value);
    if (!other) {
        throw ArithmeticError("Value::multiply", "Value cannot be multiplied");
    }
    ddouble factor = other->asNumber();
    if (factor <= 0) {
        _value.clear();
    }
    else {
        QString str;
        QTextStream os(&str);
        while (factor-- > 1) {
            os << _value;
        }
        dint remain = dint(std::floor((factor + 1) * _value.size() + .5));
        os << _value.substr(0, remain);
        _value = str;
    }
}
void TextValue::divide(Value const &value) {
    TextValue const *other = dynamic_cast<TextValue const *>(&value);
    if (!other) {
        throw ArithmeticError("TextValue::divide", "Text cannot be divided");
    }
    _value = _value / other->_value;
}
void TextValue::modulo(Value const &value) {
    list<Value const *> args;
    ArrayValue const *array = dynamic_cast<ArrayValue const *>(&value);
    if (array) {
        for (ArrayValue::Elements::const_iterator i = array->elements().begin();
            i != array->elements().end(); ++i) {
            args.push_back(*i);
        }
    }
    else {
        args.push_back(&value);
    }
    _value = substitutePlaceholders(_value, args);
}
String TextValue::substitutePlaceholders(String const &pattern, const std::list<Value const *> &args) {
    QString result;
    QTextStream out(&result);
    list<Value const *>::const_iterator arg = args.begin();
    for (String::const_iterator i = pattern.begin(); i != pattern.end(); ++i) {
        QChar ch = *i;
        if (ch == '%') {
            if (arg == args.end()) {
                throw IllegalPatternError("TextValue::replacePlaceholders",
                    "Too few substitution values");
            }
            out << String::patternFormat(i, pattern.end(), **arg);
            ++arg;
        }
        else {
            out << ch;
        }
    }
    return result;
}
void TextValue::operator >> (Writer &to) const {
    to << SerialId(TEXT) << _value;
}
void TextValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != TEXT) {
        throw DeserializationError("TextValue::operator <<", "Invalid ID");
    }
    from >> _value;
}
void TextValue::setValue(String const &text) {
    _value = text;
}
Value::Text TextValue::typeId() const {
    return "Text";
}
} // namespace de
namespace de {
static String const ISO_FORMAT = "yyyy-MM-dd hh:mm:ss.zzz";
static HighPerformanceTimer &highPerfTimer() {
    static HighPerformanceTimer hpt;
    return hpt;
}
static TimeSpan currentHighPerfDelta;
namespace internal {
class SleeperThread : public QThread {
public:
    void run() {}
    static void msleep(unsigned long milliseconds) {
        QThread::msleep(milliseconds);
    }
    static void usleep(unsigned long microseconds) {
        QThread::usleep(microseconds);
    }
};
} // namespace internal
duint64 Time::Span::asMicroSeconds() const {
    return duint64(_seconds * 1000000);
}
duint64 TimeSpan::asMilliSeconds() const {
    return duint64(_seconds * 1000);
}
ddouble TimeSpan::asMinutes() const {
    return _seconds / 60;
}
ddouble TimeSpan::asHours() const {
    return _seconds / 3600;
}
ddouble TimeSpan::asDays() const {
    return asHours() / 24;
}
Time::Span Time::Span::sinceStartOfProcess() {
    return highPerfTimer().elapsed();
}
void TimeSpan::sleep() const {
    if (_seconds < 60) {
        internal::SleeperThread::usleep(static_cast<unsigned long>(_seconds * 1e6));
    }
    else {
        internal::SleeperThread::msleep(static_cast<unsigned long>(_seconds * 1e3));
    }
}
void TimeSpan::operator >> (Writer &to) const {
    to << _seconds;
}
void TimeSpan::operator << (Reader &from) {
    from >> _seconds;
}
DENG2_PIMPL_NOREF(Time) {
    enum Flag {
        DateTime        = 0x1,
        HighPerformance = 0x2
    };
    Q_DECLARE_FLAGS(Flags, Flag)
    Flags flags;
    QDateTime dateTime;
    Span highPerfElapsed;
    Impl()
        : flags(DateTime | HighPerformance)
        , dateTime(QDateTime::currentDateTime())
        , highPerfElapsed(highPerfTimer().elapsed())
    {}
    Impl(QDateTime const &dt) : flags(DateTime), dateTime(dt) {}
    Impl(Span const &span) : flags(HighPerformance), highPerfElapsed(span) {}
    Impl(Impl const &other)
        : de::IPrivate()
        , flags(other.flags)
        , dateTime(other.dateTime)
        , highPerfElapsed(other.highPerfElapsed)
    {}
    bool hasDateTime() const {
        return flags.testFlag(DateTime);
    }
    bool isValid() const {
        if (flags.testFlag(DateTime)) {
            return dateTime.isValid();
        }
        return flags.testFlag(HighPerformance);
    }
    bool isLessThan(Impl const &other) const {
        if (flags.testFlag(HighPerformance) && other.flags.testFlag(HighPerformance)) {
            return highPerfElapsed < other.highPerfElapsed;
        }
        if (flags.testFlag(DateTime) && other.flags.testFlag(DateTime)) {
            return dateTime < other.dateTime;
        }
        /**
         * @todo Implement needed conversion to compare DateTime with high
         * performance delta time.
         */
        DENG2_ASSERT(false);
        return false;
    }
    bool isEqualTo(Impl const &other) const {
        if (flags.testFlag(HighPerformance) && other.flags.testFlag(HighPerformance)) {
            return highPerfElapsed == other.highPerfElapsed;
        }
        if (flags.testFlag(DateTime) && other.flags.testFlag(DateTime)) {
            return dateTime == other.dateTime;
        }
        if (flags.testFlag(DateTime)) {
            return fequal(highPerfTimer().startedAt().asDateTime().msecsTo(dateTime)/1000.0,
                          other.highPerfElapsed);
        }
        if (flags.testFlag(HighPerformance)) {
            return fequal(highPerfElapsed,
                          highPerfTimer().startedAt().asDateTime().msecsTo(other.dateTime)/1000.0);
        }
        return false;
    }
    void add(Span const &delta) {
        if (flags.testFlag(DateTime)) {
            dateTime = dateTime.addMSecs(delta.asMilliSeconds());
        }
        if (flags.testFlag(HighPerformance)) {
            highPerfElapsed += delta;
        }
    }
    Span delta(Impl const &earlier) const {
        if (flags.testFlag(HighPerformance) && earlier.flags.testFlag(HighPerformance)) {
            return highPerfElapsed - earlier.highPerfElapsed;
        }
        if (flags.testFlag(DateTime) && earlier.flags.testFlag(DateTime)) {
            return earlier.dateTime.msecsTo(dateTime) / 1000.0;
        }
        /**
         * @todo Implement needed conversion to compare DateTime with high
         * performance delta time.
         */
        DENG2_ASSERT(false);
        return 0;
    }
    void setDateTimeFromHighPerf() {
        dateTime = (highPerfTimer().startedAt() + highPerfElapsed).asDateTime();
        flags |= DateTime;
    }
};
Time::Time() : d(new Impl)
{}
Time::Time(Time const &other) : d(new Impl(*other.d))
{}
Time::Time(Time &&moved) : d(std::move(moved.d))
{}
Time::Time(QDateTime const &t) : d(new Impl(t))
{}
Time::Time(TimeSpan const &highPerformanceDelta)
    : ISerializable(), d(new Impl(highPerformanceDelta))
{}
Time Time::invalidTime() {
    return Time(QDateTime());
}
Time &Time::operator = (Time const &other) {
    *d = *other.d;
    return *this;
}
Time &Time::operator = (Time &&moved) {
    d = std::move(moved.d);
    return *this;
}
bool Time::isValid() const {
    return d->isValid();
}
bool Time::operator < (Time const &t) const {
    return d->isLessThan(*t.d);
}
bool Time::operator == (Time const &t) const {
    return d->isEqualTo(*t.d);
}
Time Time::operator + (Span const &span) const {
    Time result = *this;
    result += span;
    return result;
}
Time &Time::operator += (Span const &span) {
    d->add(span);
    return *this;
}
TimeSpan Time::operator - (Time const &earlierTime) const {
    return d->delta(*earlierTime.d);
}
dint Time::asBuildNumber() const {
    if (d->hasDateTime()) {
        return (d->dateTime.date().year() - 2011)*365 + d->dateTime.date().dayOfYear();
    }
    return 0;
}
String Time::asText(Format format) const {
    if (!isValid()) {
        return "(undefined time)";
    }
    if (d->hasDateTime()) {
        if (format == ISOFormat) {
            return d->dateTime.toString(ISO_FORMAT);
        }
        else if (format == ISODateOnly) {
            return d->dateTime.toString("yyyy-MM-dd");
        }
        else if (format == FriendlyFormat) {
            if (d->dateTime.date() == QDateTime::currentDateTime().date()) {
                return d->dateTime.toString("HH:mm");
            }
            else if (d->dateTime.date().year() == QDateTime::currentDateTime().date().year()) {
                return d->dateTime.toString("MMM dd HH:mm");
            }
            else {
                return d->dateTime.toString("yyyy MMM dd");
            }
        }
        else if (format == BuildNumberAndSecondsSinceStart ||
                 format == SecondsSinceStart) {
            TimeSpan elapsed;
            if (d->flags.testFlag(Impl::HighPerformance)) {
                elapsed = d->highPerfElapsed;
            }
            else if (d->flags.testFlag(Impl::DateTime)) {
                elapsed = highPerfTimer().startedAt().deltaTo(Time(d->dateTime));
            }
            int hours = int(elapsed.asHours());
            TimeSpan sec = elapsed - hours * 3600.0;
            QString prefix;
            if (format == BuildNumberAndSecondsSinceStart) {
                prefix = QString("#%1 ").arg(asBuildNumber(), -4);
            }
            if (hours > 0) {
                return QString("%1%2h%3")
                        .arg(prefix)
                        .arg(hours)
                        .arg(sec, 7, 'f', 3, '0');
            }
            return QString("%1%2")
                    .arg(prefix)
                    .arg(sec, 7, 'f', 3, '0');
        }
        else {
            return QString("#%1 ").arg(asBuildNumber(), -4) + d->dateTime.toString("hh:mm:ss.zzz");
        }
    }
    if (d->flags.testFlag(Impl::HighPerformance)) {
        return QString("+%1 sec").arg(d->highPerfElapsed, 0, 'f', 3);
    }
    return "";
}
static int parseMonth(String const &shortName) {
    static char const *months[] = {
        "Jan", "Feb", "Mar", "Apr", "May", "Jun",
        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };
    for (int i = 0; i < 12; ++i) {
        if (shortName == months[i]) {
            return i + 1;
        }
    }
    return 0;
}
Time Time::fromText(String const &text, Time::Format format) {
    DENG2_ASSERT(format == ISOFormat ||
                 format == ISODateOnly ||
                 format == FriendlyFormat ||
                 format == CompilerDateTime ||
                 format == HumanDate ||
                 format == UnixLsStyleDateTime);
    if (format == ISOFormat) {
        return Time(QDateTime::fromString(text, ISO_FORMAT));
    }
    else if (format == ISODateOnly) {
        return Time(QDateTime::fromString(text, "yyyy-MM-dd"));
    }
    else if (format == FriendlyFormat) {
        return Time(QDateTime::fromString(text, Qt::TextDate));
    }
    else if (format == CompilerDateTime) {
        QStringList const parts = text.split(" ", QString::SkipEmptyParts);
        if (parts.size() >= 4) {
            int day = parts[1].toInt();
            int year = parts[2].toInt();
            int month = parseMonth(parts[0]);
            QDate date(year, month, day);
            QTime time = QTime::fromString(parts[3], "HH:mm:ss");
            return Time(QDateTime(date, time));
        }
    }
    else if (format == UnixLsStyleDateTime) {
        QStringList const parts = text.split(" ", QString::SkipEmptyParts);
        if (parts.size() >= 3) {
            int month = parseMonth(parts[0]);
            int day = parts[1].toInt();
            int hour = 0;
            int minute = 0;
            int year;
            if (parts[2].contains(QChar(':'))) {
                year = QDate::currentDate().year();
                hour = parts[2].left(2).toInt();
                minute = parts[2].right(2).toInt();
                return Time(QDateTime(QDate(year, month, day),
                                      QTime(hour, minute, 0)));
            }
            else {
                year = parts[2].toInt();
                return Time(QDateTime(QDate(year, month, day), QTime(0, 0, 0)));
            }
        }
    }
    else if (format == HumanDate) {
        static QStringList const formats({
            "M/d/yy",
            "MM/dd/yy",
            "d.M.yy",
            "dd.MM.yy",
            "MM/dd/yyyy",
            "d.M.yyyy",
            "dd.MM.yyyy",
            "MM.dd.yyyy", // as a fallback...
            "yyyy-MM-dd",
            "MMM d yyyy",
            "MMM d hh:mm",
        });
        String const normText = text.normalizeWhitespace();
        for (int i = 0; i < formats.size(); ++i) {
            String const fmt = formats.at(i);
            bool const isShortYear = i < 4;
            QDateTime dt = QDateTime::fromString(normText, fmt);
            if (dt.isValid()) {
                if (i == 10) // No year specified.
                {
                    dt.setDate(QDate(QDate::currentDate().year(),
                                     dt.date().month(),
                                     dt.date().day()));
                }
                else if (isShortYear && dt.date().year() < 1980) {
                    dt.setDate(QDate(dt.date().year() + 100,
                                     dt.date().month(),
                                     dt.date().day())); // Y2K?
                }
                return Time(dt);
            }
        }
        return Time(QDateTime::fromString(normText, Qt::TextDate));
    }
    return Time::invalidTime();
}
QDateTime &de::Time::asDateTime() {
    DENG2_ASSERT(d->hasDateTime());
    return d->dateTime;
}
QDateTime const &de::Time::asDateTime() const {
    if (!d->hasDateTime() && d->flags.testFlag(Impl::HighPerformance)) {
        d->dateTime = (highPerfTimer().startedAt() + d->highPerfElapsed).asDateTime();
        d->flags |= Impl::DateTime;
    }
    return d->dateTime;
}
Date Time::asDate() const {
    DENG2_ASSERT(d->hasDateTime());
    return Date(*this);
}
static duint8 const HAS_DATETIME  = 0x01;
static duint8 const HAS_HIGH_PERF = 0x02;
void Time::operator >> (Writer &to) const {
    duint8 flags = (d->flags & Impl::DateTime?        HAS_DATETIME  : 0) |
                   (d->flags & Impl::HighPerformance? HAS_HIGH_PERF : 0);
    to << flags;
    if (d->flags.testFlag(Impl::DateTime)) {
        Block bytes;
        QDataStream s(&bytes, QIODevice::WriteOnly);
        s.setVersion(QDataStream::Qt_4_8);
        s << d->dateTime;
        to << bytes;
    }
    if (d->flags.testFlag(Impl::HighPerformance)) {
        to << d->highPerfElapsed;
    }
}
void Time::operator << (Reader &from) {
    if (from.version() >= DENG2_PROTOCOL_1_11_0_Time_high_performance) {
        /*
         * Starting from build 926, Time can optionally contain a high-performance delta
         * component.
         */
        duint8 flags;
        from >> flags;
        d->flags = 0;
        if (flags & HAS_DATETIME) {
            d->flags |= Impl::DateTime;
            Block bytes;
            from >> bytes;
            QDataStream s(bytes);
            s.setVersion(QDataStream::Qt_4_8);
            s >> d->dateTime;
        }
        if (flags & HAS_HIGH_PERF) {
            d->flags |= Impl::HighPerformance;
            from >> d->highPerfElapsed;
        }
        if ((flags & HAS_DATETIME) && (flags & HAS_HIGH_PERF)) {
            if (d->dateTime < highPerfTimer().startedAt().asDateTime()) {
                d->flags &= ~Impl::HighPerformance;
            }
            else {
                d->highPerfElapsed = highPerfTimer().startedAt().deltaTo(d->dateTime);
            }
        }
    }
    else {
        Block bytes;
        from >> bytes;
        QDataStream s(bytes);
        s.setVersion(QDataStream::Qt_4_8);
        s >> d->dateTime;
        d->flags = Impl::DateTime;
    }
}
TimeSpan Time::highPerformanceTime() const {
    DENG2_ASSERT(d->flags & Impl::HighPerformance);
    return d->highPerfElapsed;
}
Time Time::currentHighPerformanceTime() // static
{
    return currentHighPerfDelta;
}
void Time::updateCurrentHighPerformanceTime() // static
{
    currentHighPerfDelta = highPerfTimer().elapsed();
}
QTextStream &operator << (QTextStream &os, Time const &t) {
    os << t.asText();
    return os;
}
} // namespace de
namespace de {
TimeValue::TimeValue(Time const &time) : _time(time)
{}
Value *TimeValue::duplicate() const {
    return new TimeValue(_time);
}
Value::Text TimeValue::asText() const {
    if (!_time.isValid()) return "(undefined Time)";
    return _time.asText();
}
bool TimeValue::isTrue() const {
    return _time.isValid();
}
dint TimeValue::compare(Value const &value) const {
    TimeValue const *other = dynamic_cast<TimeValue const *>(&value);
    if (other) {
        if (other->_time > _time) return 1;
        if (other->_time < _time) return -1;
        return 0;
    }
    return Value::compare(value);
}
void TimeValue::sum(Value const &value) {
    _time += TimeSpan(value.asNumber());
}
void TimeValue::subtract(Value const &subtrahend) {
    _time -= TimeSpan(subtrahend.asNumber());
}
void TimeValue::operator >> (Writer &to) const {
    to << SerialId(TIME) << _time;
}
void TimeValue::operator << (Reader &from) {
    SerialId id;
    from >> id;
    if (id != TIME) {
        throw DeserializationError("TimeValue::operator <<", "Invalid ID");
    }
    from >> _time;
}
Value::Text TimeValue::typeId() const {
    return "Time";
}
} // namespace de
namespace de {
Value::~Value()
{}
Value *Value::duplicateAsReference() const {
    return duplicate();
}
Value::Number Value::asNumber() const {
    throw ConversionError("Value::asNumber", "Illegal conversion");
}
Value::Number Value::asSafeNumber(Number const &defaultValue) const {
    try {
        return asNumber();
    }
    catch (Error const &) {
        return defaultValue;
    }
}
int Value::asInt() const {
    const double num = asNumber();
    if (num > std::numeric_limits<int>::max()) {
        return std::numeric_limits<int>::max();
    }
    return round<int>(num);
}
int Value::asUInt() const {
    const double num = asNumber();
    if (num < 0) {
        throw ArithmeticError("Value::asUInt", "Cannot convert negative number to unsigned integer");
    }
    if (num > std::numeric_limits<uint32_t>::max()) {
        throw ArithmeticError("Value::asUInt", "Value is too large to represent as uint32");
    }
    return uint32_t(num + 0.5);
}
StringList Value::asStringList() const {
    StringList str;
    if (is<ArrayValue>(this)) {
        for (Value const *val : as<ArrayValue>().elements()) {
            str << val->asText();
        }
    }
    else {
        str << asText();
    }
    return str;
}
Record *Value::memberScope() const {
    return 0;
}
dsize Value::size() const {
    throw IllegalError("Value::size", "Size is meaningless");
}
Value const &Value::element(Value const &/*index*/) const
{
    throw IllegalError("Value::element",
                       String("Value cannot be indexed (%1 \"%2\")")
                       .arg(typeid(*this).name())
                       .arg(asText()));
}
Value &Value::element(Value const &/*index*/)
{
    throw IllegalError("Value::element", "Value cannot be indexed");
}
Value *Value::duplicateElement(Value const &index) const {
    return element(index).duplicate();
}
void Value::setElement(Value const &/*index*/, Value *)
{
    throw IllegalError("Value::setElement", "Value cannot be indexed");
}
bool Value::contains(Value const &/*value*/) const
{
    throw IllegalError("Value::contains", "Value is not a container");
}
Value *Value::begin() {
    throw IllegalError("Value::begin", "Value is not iterable");
}
Value *Value::next() {
    throw IllegalError("Value::next", "Value is not iterable");
}
bool Value::isFalse() const {
    return !isTrue();
}
dint Value::compare(Value const &value) const {
    dint result = asText().compare(value.asText());
    return (result < 0? -1 : result > 0? 1 : 0);
}
void Value::negate() {
    throw ArithmeticError("Value::negate", "Value cannot be negated");
}
void Value::sum(Value const &/*value*/)
{
    throw ArithmeticError("Value::sum", "Value cannot be summed");
}
void Value::subtract(Value const &/*subtrahend*/)
{
    throw ArithmeticError("Value::subtract", "Value cannot be subtracted from");
}
void Value::divide(Value const &/*divisor*/)
{
    throw ArithmeticError("Value::divide", "Value cannot be divided");
}
void Value::multiply(Value const &/*value*/)
{
    throw ArithmeticError("Value::multiply", "Value cannot be multiplied");
}
void Value::modulo(Value const &/*divisor*/)
{
    throw ArithmeticError("Value::modulo", "Modulo not defined");
}
void Value::assign(Value *value) {
    delete value;
    throw IllegalError("Value::assign", "Cannot assign to value");
}
void Value::call(Process &, Value const &, Value *) const {
    throw IllegalError("Value::call", "Value cannot be called");
}
Value *Value::constructFrom(Reader &reader) {
    SerialId id;
    reader.mark();
    reader >> id;
    reader.rewind();
    std::unique_ptr<Value> result;
    switch (id) {
    case NONE:
        result.reset(new NoneValue);
        break;
    case NUMBER:
        result.reset(new NumberValue);
        break;
    case TEXT:
        result.reset(new TextValue);
        break;
    case ARRAY:
        result.reset(new ArrayValue);
        break;
    case DICTIONARY:
        result.reset(new DictionaryValue);
        break;
    case BLOCK:
        result.reset(new BlockValue);
        break;
    case FUNCTION:
        result.reset(new FunctionValue);
        break;
    case RECORD:
        result.reset(new RecordValue(new Record, RecordValue::OwnsRecord));
        break;
    case TIME:
        result.reset(new TimeValue);
        break;
    case ANIMATION:
        result.reset(new AnimationValue);
        break;
    default:
        throw DeserializationError("Value::constructFrom", "Invalid value identifier");
    }
    reader >> *result.get();
    return result.release();
}
Value *Value::constructFrom(QVariant const &variant) {
    switch (variant.type()) {
    case QVariant::Invalid:
        return new NoneValue;
    case QVariant::Bool:
        return new NumberValue(variant.toBool());
    case QVariant::Double:
        return new NumberValue(variant.toDouble());
    default:
        break;
    }
    if (variant.canConvert<QVariantList>()) {
        std::unique_ptr<ArrayValue> array(new ArrayValue);
        foreach (QVariant value, variant.toList()) {
            *array << constructFrom(value);
        }
        return array.release();
    }
    if (variant.canConvert<QVariantMap>()) {
        auto const map = variant.toMap();
        if (map.contains(QStringLiteral("__obj__")) &&
            map[QStringLiteral("__obj__")] == QStringLiteral("Record")) {
            std::unique_ptr<Record> rec(new Record);
            foreach (QString key, map.keys()) {
                std::unique_ptr<Value> v(constructFrom(map[key]));
                rec->add(new Variable(key, v.release()));
            }
            return new RecordValue(rec.release(), RecordValue::OwnsRecord);
        }
        else {
            std::unique_ptr<DictionaryValue> dict(new DictionaryValue);
            foreach (QString key, map.keys()) {
                std::unique_ptr<Value> value(constructFrom(map[key]));
                dict->add(new TextValue(key), value.release());
            }
            return dict.release();
        }
    }
    return new TextValue(variant.toString());
}
Value const &Value::element(dint index) const {
    return element(NumberValue(index));
}
Value &Value::element(dint index) {
    return element(NumberValue(index));
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Variable) {
    String name;
    Value *value;
    Flags flags;
    Impl() : value(0) {}
    Impl(Impl const &other)
        : de::IPrivate()
        , name (other.name)
        , value(other.value->duplicate())
        , flags(other.flags)
    {}
    ~Impl() {
        delete value;
    }
    DENG2_PIMPL_AUDIENCE(Deletion)
    DENG2_PIMPL_AUDIENCE(Change)
    DENG2_PIMPL_AUDIENCE(ChangeFrom)
};
DENG2_AUDIENCE_METHOD(Variable, Deletion)
DENG2_AUDIENCE_METHOD(Variable, Change)
DENG2_AUDIENCE_METHOD(Variable, ChangeFrom)
Variable::Variable(String const &name, Value *initial, Flags const &m)
    : d(new Impl) {
    d->name = name;
    d->flags = m;
    std::unique_ptr<Value> v(initial);
    if (!initial) {
        v.reset(new NoneValue);
    }
    verifyName(d->name);
    if (initial) {
        verifyValid(*v);
    }
    d->value = v.release();
}
Variable::Variable(Variable const &other)
    : ISerializable()
    , d(new Impl(*other.d))
{}
Variable::~Variable() {
    DENG2_FOR_AUDIENCE2(Deletion, i) i->variableBeingDeleted(*this);
}
String const &Variable::name() const {
    return d->name;
}
Variable &Variable::operator = (Value *v) {
    set(v);
    return *this;
}
Variable &Variable::operator = (String const &textValue) {
    set(new TextValue(textValue));
    return *this;
}
Variable &Variable::set(Value *v) {
    DENG2_ASSERT(v != 0);
    QScopedPointer<Value> val(v);
    verifyWritable(*v);
    verifyValid(*v);
    QScopedPointer<Value> oldValue(d->value); // old value deleted afterwards
    d->value = val.take();
    d->flags |= ValueHasChanged;
    if (!audienceForChange().isEmpty() || !audienceForChangeFrom().isEmpty()) {
        bool notify;
        try {
            notify = oldValue.isNull() || oldValue->compare(*v);
        }
        catch (Error const &) {
            notify = true;
        }
        if (notify) {
            DENG2_FOR_AUDIENCE2(Change,     i) i->variableValueChanged(*this, *d->value);
            DENG2_FOR_AUDIENCE2(ChangeFrom, i) i->variableValueChangedFrom(*this, *oldValue, *d->value);
        }
    }
    return *this;
}
Variable &Variable::set(Value const &v) {
    set(v.duplicate());
    return *this;
}
Value const &Variable::value() const {
    DENG2_ASSERT(d->value != 0);
    return *d->value;
}
Value &Variable::value() {
    DENG2_ASSERT(d->value != 0);
    return *d->value;
}
Value *Variable::valuePtr() {
    return d->value;
}
Value const *Variable::valuePtr() const {
    return d->value;
}
Record &Variable::valueAsRecord() {
    return value<RecordValue>().dereference();
}
Record const &Variable::valueAsRecord() const {
    return value<RecordValue>().dereference();
}
ArrayValue const &Variable::array() const {
    return value<ArrayValue>();
}
ArrayValue &Variable::array() {
    return value<ArrayValue>();
}
Variable::operator Record & () {
    return valueAsRecord();
}
Variable::operator Record const & () const {
    return valueAsRecord();
}
Variable::operator String () const {
    return value().asText();
}
Variable::operator QString () const {
    return value().asText();
}
Variable::operator ddouble () const {
    return value().asNumber();
}
Variable::Flags Variable::flags() const {
    return d->flags;
}
void Variable::setFlags(Flags const &flags, FlagOpArg operation) {
    applyFlagOperation(d->flags, flags, operation);
}
Variable &Variable::setReadOnly() {
    d->flags |= ReadOnly;
    return *this;
}
bool Variable::isValid(Value const &v) const {
    if ((!d->flags.testFlag(AllowNone)       && dynamic_cast<NoneValue const *>(&v)      ) ||
        (!d->flags.testFlag(AllowNumber)     && dynamic_cast<NumberValue const *>(&v)    ) ||
        (!d->flags.testFlag(AllowText)       && dynamic_cast<TextValue const *>(&v)      ) ||
        (!d->flags.testFlag(AllowArray)      && dynamic_cast<ArrayValue const *>(&v)     ) ||
        (!d->flags.testFlag(AllowDictionary) && dynamic_cast<DictionaryValue const *>(&v)) ||
        (!d->flags.testFlag(AllowBlock)      && dynamic_cast<BlockValue const *>(&v)     ) ||
        (!d->flags.testFlag(AllowTime)       && dynamic_cast<TimeValue const *>(&v)      )  ) {
        return false;
    }
    return true;
}
void Variable::verifyValid(Value const &v) const {
    if (!isValid(v)) {
        throw InvalidError("Variable::verifyValid",
            "Value type is not allowed by the variable '" + d->name + "'");
    }
}
void Variable::verifyWritable(Value const &attemptedNewValue) {
    if (d->flags & ReadOnly) {
        Value const &currentValue = *d->value;
        if (d->value && typeid(currentValue) == typeid(attemptedNewValue) &&
           !d->value->compare(attemptedNewValue)) {
            return;
        }
        throw ReadOnlyError("Variable::verifyWritable",
            "Variable '" + d->name + "' is in read-only mode");
    }
}
void Variable::verifyName(String const &s) {
    if (s.indexOf('.') != String::npos) {
        throw NameError("Variable::verifyName", "Name contains '.': " + s);
    }
}
void Variable::operator >> (Writer &to) const {
    if (!d->flags.testFlag(NoSerialize)) {
        to << d->name << duint32(d->flags) << *d->value;
    }
}
void Variable::operator << (Reader &from) {
    duint32 modeFlags = 0;
    from >> d->name >> modeFlags;
    d->flags = Flags(modeFlags);
    delete d->value;
    try {
        d->value = Value::constructFrom(from);
    }
    catch (Error const &) {
        d->value = new NoneValue();
        throw;
    }
}
} // namespace de
namespace de {
DENG2_PIMPL_NOREF(Writer) {
    ByteOrder const &convert;
    IByteArray *destination;
    IOStream *stream;
    IByteArray::Offset offset;
    IByteArray::Offset markOffset = 0;
    IByteArray::Offset const fixedOffset;
    Impl(ByteOrder const &order, IByteArray *dest, IByteArray::Offset off)
        : convert(order), destination(dest), stream(0), offset(off), fixedOffset(0) {}
    Impl(ByteOrder const &order, IOStream *str)
        : convert(order), destination(0), stream(str), offset(0), fixedOffset(0) {
        destination = dynamic_cast<IByteArray *>(str);
        if (destination) {
            stream = 0;
        }
    }
    Impl(Impl const &other, ByteOrder const &order)
        : convert    (order            ),
          destination(other.destination),
          stream     (other.stream     ),
          offset     (other.offset     ),
          fixedOffset(other.fixedOffset)
    {}
    void write(IByteArray::Byte const *ptr, dsize size) {
        if (destination) {
            destination->set(fixedOffset + offset, ptr, size);
            offset += size;
        }
        else if (stream) {
            *stream << ByteRefArray(ptr, size);
        }
    }
};
Writer::Writer(IByteArray &destination, ByteOrder const &byteOrder, IByteArray::Offset offset)
    : d(new Impl(byteOrder, &destination, offset))
{}
Writer::Writer(IByteArray &destination, IByteArray::Offset offset)
    : d(new Impl(littleEndianByteOrder, &destination, offset))
{}
Writer::Writer(IOStream &stream, ByteOrder const &byteOrder)
    : d(new Impl(byteOrder, &stream))
{}
Writer::Writer(ByteArrayFile &destination, ByteOrder const &byteOrder, IByteArray::Offset offset)
    : d(new Impl(byteOrder, static_cast<IByteArray *>(&destination), offset))
{}
Writer::Writer(Writer const &other)
    : d(new Impl(*other.d, other.d->convert))
{}
Writer::Writer(Writer const &other, ByteOrder const &byteOrder)
    : d(new Impl(*other.d, byteOrder))
{}
Writer &Writer::withHeader() {
    return *this << duint32(version());
}
duint Writer::version() const {
    return DENG2_PROTOCOL_LATEST;
}
Writer &Writer::operator << (char const &byte) {
    d->write(reinterpret_cast<IByteArray::Byte const *>(&byte), 1);
    return *this;
}
Writer &Writer::operator << (dchar const &byte) {
    d->write(reinterpret_cast<IByteArray::Byte const *>(&byte), 1);
    return *this;
}
Writer &Writer::operator << (duchar const &byte) {
    d->write(&byte, 1);
    return *this;
}
Writer &Writer::operator << (dint16 const &word) {
    return *this << static_cast<duint16>(word);
}
Writer &Writer::operator << (duint16 const &word) {
    duint16 netWord;
    d->convert.hostToNetwork(word, netWord);
    d->write(reinterpret_cast<IByteArray::Byte *>(&netWord), 2);
    return *this;
}
Writer &Writer::operator << (dint32 const &dword) {
    return *this << static_cast<duint32>(dword);
}
Writer &Writer::operator << (duint32 const &dword) {
    duint32 netDword;
    d->convert.hostToNetwork(dword, netDword);
    d->write(reinterpret_cast<IByteArray::Byte *>(&netDword), 4);
    return *this;
}
Writer &Writer::operator << (dint64 const &qword) {
    return *this << static_cast<duint64>(qword);
}
Writer &Writer::operator << (duint64 const &qword) {
    duint64 netQword;
    d->convert.hostToNetwork(qword, netQword);
    d->write(reinterpret_cast<IByteArray::Byte *>(&netQword), 8);
    return *this;
}
Writer &Writer::operator << (dfloat const &value) {
    return *this << *reinterpret_cast<duint32 const *>(&value);
}
Writer &Writer::operator << (ddouble const &value) {
    return *this << *reinterpret_cast<duint64 const *>(&value);
}
Writer &Writer::operator << (String const &text) {
    Block bytes = text.toUtf8();
    duint32 size = bytes.size();
    *this << size;
    d->write(bytes.data(), size);
    return *this;
}
Writer &Writer::operator << (Block const &block) {
    return *this << static_cast<IWritable const &>(block);
}
Writer &Writer::writeText(String const &text) {
    Block const utf8 = text.toUtf8();
    d->write(utf8.dataConst(), utf8.size());
    return *this;
}
Writer &Writer::operator << (IByteArray const &byteArray) {
    return *this << duint32(byteArray.size()) << FixedByteArray(byteArray);
}
Writer &Writer::operator << (FixedByteArray const &fixedByteArray) {
    /**
     * @note  A copy of @a fixedByteArray is done because there is no certainty
     * that the source data actually exists anywhere. The object implementing
     * IByteArray could be generating it on the fly.
     */
    Block copy(fixedByteArray);
    d->write(copy.data(), copy.size());
    return *this;
}
Writer &Writer::writeBytes(dsize count, IByteArray const &array) {
    return *this << FixedByteArray(array, 0, count);
}
Writer &Writer::writeBytes(IByteArray const &array) {
    return *this << FixedByteArray(array);
}
Writer &Writer::operator << (IWritable const &writable) {
    writable >> *this;
    return *this;
}
IByteArray const *Writer::destination() const {
    return d->destination;
}
IByteArray *Writer::destination() {
    return d->destination;
}
IByteArray::Offset Writer::offset() const {
    return d->offset;
}
void Writer::setOffset(IByteArray::Offset offset) {
    if (d->stream) {
        throw SeekError("Writer::setOffset", "Cannot change offset when writing to a stream");
    }
    d->offset = offset;
}
void Writer::mark() {
    if (d->stream) {
        throw SeekError("Writer::mark", "Marking positions is not possible in a stream");
    }
    d->markOffset = d->offset;
}
void Writer::rewind() {
    setOffset(d->markOffset);
}
Writer &Writer::beginIndeterminateSpan() {
    mark();
    *this << duint32(0);
    return *this;
}
Writer &Writer::endIndeterminateSpan() {
    if (d->stream) {
        throw SeekError("Writer::endIndeterminateSpan", "Not possible in a stream");
    }
    auto const oldOffset = d->offset;
    duint32 const delta = duint32(d->offset - d->markOffset) - 4;
    rewind();
    *this << delta;
    setOffset(oldOffset);
    return *this;
}
Writer &Writer::operator << (InlineOperation op) {
    switch (op) {
    case BeginSpan:
        beginIndeterminateSpan();
        break;
    case EndSpan:
        endIndeterminateSpan();
        break;
    }
    return *this;
}
ByteOrder const &Writer::byteOrder() const {
    return d->convert;
}
void Writer::seek(IByteArray::Delta count) {
    if (d->stream) {
        throw SeekError("Writer::seek", "Cannot seek when writing to a stream");
    }
    if (IByteArray::Delta(d->fixedOffset + d->offset) + count < 0) {
        throw IByteArray::OffsetError("Writer::seek", "Seek past beginning of destination");
    }
    d->offset = IByteArray::Offset(IByteArray::Delta(d->offset) + count);
}
void Writer::seekToEnd() {
    if (d->destination) {
        setOffset(d->destination->size() - d->fixedOffset);
    }
}
} // namespace de
namespace de {
namespace internal {
#define SIG_LOCAL_FILE_HEADER   0x04034b50
#define SIG_CENTRAL_FILE_HEADER 0x02014b50
#define SIG_END_OF_CENTRAL_DIR  0x06054b50
#define SIG_DIGITAL_SIGNATURE   0x05054b50
#define MAXIMUM_COMMENT_SIZE    2048
#define CENTRAL_END_SIZE        22
#define REQUIRED_DEFLATE_PERCENTAGE .98
#define ZFH_ENCRYPTED           0x1
#define ZFH_COMPRESSION_OPTS    0x6
#define ZFH_DESCRIPTOR          0x8
#define ZFH_COMPRESS_PATCHED    0x20    // Not supported.
enum Compression {
    NO_COMPRESSION = 0,     // Supported format.
    SHRUNK,
    REDUCED_1,
    REDUCED_2,
    REDUCED_3,
    REDUCED_4,
    IMPLODED,
    DEFLATED = 8,           // The only supported compression (via zlib).
    DEFLATED_64,
    PKWARE_DCL_IMPLODED
};
/**
 * MS-DOS time.
 * - 0..4:   Second/2
 * - 5..10:  Minute
 * - 11..15: Hour
 */
struct DOSTime {
    duint16 seconds;
    duint16 minutes;
    duint16 hours;
    DOSTime(duint16 h, duint16 m, duint16 s) : seconds(s), minutes(m), hours(h) {}
    DOSTime(duint16 const &i) {
        seconds = (i & 0x1f) * 2;
        minutes = (i >> 5) & 0x3f;
        hours = i >> 11;
    }
    operator duint16() const {
        return (hours << 11) | ((minutes & 0x3f) << 5) | ((seconds/2) & 0x1f);
    }
};
/**
 * MS-DOS date.
 * - 0..4:  Day of the month (1-31)
 * - 5..8:  Month (1=Jan)
 * - 9..15: Year since 1980
 */
struct DOSDate {
    duint16 dayOfMonth;
    duint16 month;
    duint16 year;
    DOSDate(duint16 y, duint16 m, duint16 d) : dayOfMonth(d), month(m), year(y) {}
    DOSDate(duint16 const &i) {
        dayOfMonth = i & 0x1f;
        month = (i >> 5) & 0xf;
        year = i >> 9;
    }
    operator duint16() const {
        return (year << 9) | ((month & 0xf) << 5) | (dayOfMonth & 0x1f);
    }
};
struct LocalFileHeader : public ISerializable {
    Uint32 signature;
    Uint16 requiredVersion;
    Uint16 flags;
    Uint16 compression;
    Uint16 lastModTime;
    Uint16 lastModDate;
    Uint32 crc32;
    Uint32 compressedSize;
    Uint32 size;
    Uint16 fileNameSize;
    Uint16 extraFieldSize;
    void operator >> (Writer &to) const {
        to  << signature
            << requiredVersion
            << flags
            << compression
            << lastModTime
            << lastModDate
            << crc32
            << compressedSize
            << size
            << fileNameSize
            << extraFieldSize;
    }
    void operator << (Reader &from) {
        from >> signature
             >> requiredVersion
             >> flags
             >> compression
             >> lastModTime
             >> lastModDate
             >> crc32
             >> compressedSize
             >> size
             >> fileNameSize
             >> extraFieldSize;
    }
};
struct CentralFileHeader : public ISerializable {
    Uint32 signature;
    Uint16 version;
    Uint16 requiredVersion;
    Uint16 flags;
    Uint16 compression;
    Uint16 lastModTime;
    Uint16 lastModDate;
    Uint32 crc32;
    Uint32 compressedSize;
    Uint32 size;
    Uint16 fileNameSize;
    Uint16 extraFieldSize;
    Uint16 commentSize;
    Uint16 diskStart;
    Uint16 internalAttrib;
    Uint32 externalAttrib;
    Uint32 relOffset;
    /*
     * Followed by:
     * file name (variable size)
     * extra field (variable size)
     * file comment (variable size)
     */
    void operator >> (Writer &to) const {
        to  << signature
            << version
            << requiredVersion
            << flags
            << compression
            << lastModTime
            << lastModDate
            << crc32
            << compressedSize
            << size
            << fileNameSize
            << extraFieldSize
            << commentSize
            << diskStart
            << internalAttrib
            << externalAttrib
            << relOffset;
    }
    void operator << (Reader &from) {
        from >> signature
             >> version
             >> requiredVersion
             >> flags
             >> compression
             >> lastModTime
             >> lastModDate
             >> crc32
             >> compressedSize
             >> size
             >> fileNameSize
             >> extraFieldSize
             >> commentSize
             >> diskStart
             >> internalAttrib
             >> externalAttrib
             >> relOffset;
    }
};
struct CentralEnd : public ISerializable {
    Uint16 disk;
    Uint16 centralStartDisk;
    Uint16 diskEntryCount;
    Uint16 totalEntryCount;
    Uint32 size;
    Uint32 offset;
    Uint16 commentSize;
    void operator >> (Writer &to) const {
        to  << disk
            << centralStartDisk
            << diskEntryCount
            << totalEntryCount
            << size
            << offset
            << commentSize;
    }
    void operator << (Reader &from) {
        from >> disk
             >> centralStartDisk
             >> diskEntryCount
             >> totalEntryCount
             >> size
             >> offset
             >> commentSize;
    }
};
} // namespace internal
using namespace internal;
static String ZIPARCHIVE_META_CATEGORY = "ZipArchive";
DENG2_PIMPL(ZipArchive) {
    Block directoryCacheId;
    CentralEnd zipSummary;
    QVector<std::pair<Block, CentralFileHeader>> centralHeaders;
    Impl(Public *i) : Base(i) {}
    /**
     * Locates the central directory. Start from the earliest location where
     * the signature might be.
     *
     * @param reader  Source.
     */
    bool seekToCentralEnd(Reader &reader) {
        for (duint pos = CENTRAL_END_SIZE; pos < MAXIMUM_COMMENT_SIZE + CENTRAL_END_SIZE; pos++) {
            reader.setOffset(reader.source()->size() - pos);
            duint32 signature;
            reader >> signature;
            if (signature == SIG_END_OF_CENTRAL_DIR) {
                return true;
            }
        }
        return false;
    }
    /**
     * Reads the central directory. THe reader must currently be at the start of the
     * CentralEnd block.
     *
     * @param reader      Source.
     * @param updateFromLocalHeaders  Read the local file headers, too.
     * @param overrideLocalHeaderStart  Offset to apply to all positions within the source data.
     *                    Used for pretending that cached metadata is at its original
     *                    offset instead of the beginning of the cache file.
     *                    Ignored if @a updateFromLocalHeaders is true.
     */
    void readCentralDirectory(Reader &reader, bool readingFromOriginal,
                              IByteArray::Offset localHeaderStartOffset = IByteArray::Offset(-1)) {
        reader >> zipSummary;
        duint const entryCount = zipSummary.totalEntryCount;
        if (entryCount != zipSummary.diskEntryCount) {
            throw MultiPartError("ZipArchive::readCentralDirectory", "Multipart archives are not supported");
        }
        reader.setOffset(readingFromOriginal? IByteArray::Offset(zipSummary.offset)
                                            : localHeaderStartOffset);
        for (duint index = 0; index < entryCount; ++index) {
            CentralFileHeader header;
            reader >> header;
            if (header.signature != SIG_CENTRAL_FILE_HEADER) {
                throw FormatError("ZipArchive::readCentralDirectory",
                                  QString("Corrupt central directory (at file %1)").arg(index));
            }
            Block latin1Name;
            reader.readBytes(header.fileNameSize, latin1Name);
            String const fileName = String::fromLatin1(latin1Name); /// @todo UTF-8?
            reader.seek(header.extraFieldSize + header.commentSize);
            ZipEntry *entry = nullptr;
            if (!fileName.endsWith("/") || header.size) {
                if (header.compression != NO_COMPRESSION && header.compression != DEFLATED) {
                    throw UnknownCompressionError("ZipArchive::readCentralDirectory",
                                                  "Entry '" + fileName + "' uses an unsupported compression algorithm");
                }
                if (header.flags & ZFH_ENCRYPTED) {
                    throw EncryptionError("ZipArchive::readCentralDirectory",
                                          "Entry '" + fileName + "' is encrypted and thus cannot be read");
                }
                entry = static_cast<ZipEntry *>(&self().insertEntry(fileName));
                entry->size              = header.size;
                entry->sizeInArchive     = header.compressedSize;
                entry->compression       = header.compression;
                entry->crc32             = header.crc32;
                entry->localHeaderOffset = header.relOffset;
                DOSDate lastModDate(header.lastModDate);
                DOSTime lastModTime(header.lastModTime);
                entry->modifiedAt = QDateTime(QDate(lastModDate.year + 1980, lastModDate.month, lastModDate.dayOfMonth),
                                              QTime(lastModTime.hours, lastModTime.minutes, lastModTime.seconds),
                                              Qt::UTC); // buildpackage.py sets UTC for .packs
                if (readingFromOriginal) {
                    LocalFileHeader localHeader;
                    reader.mark();
                    reader.setOffset(header.relOffset);
                    reader >> localHeader;
                    entry->offset = reader.offset() +
                                    localHeader.fileNameSize +
                                    localHeader.extraFieldSize;
                    reader.rewind();
                }
                else {
                    entry->offset = header.relOffset;
                }
            }
            if (readingFromOriginal) {
                header.extraFieldSize = 0;
                header.commentSize = 0;
                header.relOffset = (entry? duint32(entry->offset) : 0);
                centralHeaders << std::make_pair(latin1Name, header);
            }
        }
    }
    /**
     * Write a copy of the unmodified central directory as it was read from the source
     * archive. This is saved in the metadata cache.
     *
     * @param writer  Destination.
     */
    void writeCachedCentralDirectory(Writer &writer) {
        for (auto const &name_header : centralHeaders) {
            writer << name_header.second
                   << FixedByteArray(name_header.first);
        }
        writer << duint32(SIG_END_OF_CENTRAL_DIR) << zipSummary;
    }
    /**
     * Writes a new central directory for a new ZIP archive as it will be written by
     * ZipArchive.
     *
     * @param writer  Destination.
     */
    void writeCentralDirectory(Writer &writer) {
        CentralEnd summary;
        summary.diskEntryCount = summary.totalEntryCount = self().index().leafNodes().size();
        summary.offset = writer.offset();
        for (PathTreeIterator<Index> iter(self().index().leafNodes()); iter.hasNext(); ) {
            ZipEntry const &entry = iter.next();
            String const fullPath = entry.path();
            CentralFileHeader header;
            header.signature = SIG_CENTRAL_FILE_HEADER;
            header.version = 20;
            header.requiredVersion = 20;
            Date at(entry.modifiedAt);
            header.lastModTime = DOSTime(at.hours(), at.minutes(), at.seconds());
            header.lastModDate = DOSDate(at.year() - 1980, at.month(), at.dayOfMonth());
            header.compression = entry.compression;
            header.crc32 = entry.crc32;
            header.compressedSize = entry.sizeInArchive;
            header.size = entry.size;
            header.fileNameSize = fullPath.size();
            header.relOffset = entry.localHeaderOffset;
            writer << header << FixedByteArray(fullPath.toLatin1());
        }
        summary.size = writer.offset() - summary.offset;
        writer << duint32(SIG_END_OF_CENTRAL_DIR) << summary;
        writer << duint32(SIG_DIGITAL_SIGNATURE)
               << duint16(0); // No signature data.
    }
    void updateCachedDirectory() {
        if (directoryCacheId) {
            Block meta;
            Writer writer(meta);
            writeCachedCentralDirectory(writer);
            MetadataBank::get().setMetadata(ZIPARCHIVE_META_CATEGORY, directoryCacheId, meta);
        }
    }
    bool restoreFromCache() {
        if (!directoryCacheId) return false;
        auto &bank = MetadataBank::get();
        try {
            if (Block const meta = bank.check(ZIPARCHIVE_META_CATEGORY, directoryCacheId)) {
                Reader reader(meta);
                if (!seekToCentralEnd(reader)) {
                    return false;
                }
                readCentralDirectory(reader, false, 0);
                return true;
            }
        }
        catch (Error const &er) {
            LOGDEV_RES_WARNING("Corrupt cached metadata: %s") << er.asText();
        }
        return false;
    }
};
ZipArchive::ZipArchive() : d(new Impl(this)) {
    setIndex(new Index);
}
ZipArchive::ZipArchive(IByteArray const &archive, Block const &dirCacheId)
    : Archive(archive)
    , d(new Impl(this)) {
    setIndex(new Index);
    d->directoryCacheId = dirCacheId;
    if (d->restoreFromCache()) {
        return;
    }
    Reader reader(archive, littleEndianByteOrder);
    if (!d->seekToCentralEnd(reader)) {
        throw MissingCentralDirectoryError("ZipArchive::Archive",
            "Could not locate the central directory of the archive");
    }
    d->readCentralDirectory(reader, true);
    d->updateCachedDirectory();
    d->centralHeaders.clear();
}
void ZipArchive::readFromSource(Entry const &e, Path const &, IBlock &uncompressedData) const {
    ZipEntry const &entry = static_cast<ZipEntry const &>(e);
    if (entry.compression == NO_COMPRESSION) {
        if (entry.dataInArchive) {
            uncompressedData.copyFrom(*entry.dataInArchive, 0, entry.size);
        }
        else {
            DENG2_ASSERT(source() != NULL);
            uncompressedData.copyFrom(*source(), entry.offset, entry.size);
        }
    }
    else // Data is compressed.
    {
        uncompressedData.resize(entry.size);
        if (!entry.dataInArchive) {
            DENG2_ASSERT(source() != NULL);
            entry.dataInArchive.reset(new Block(*source(), entry.offset, entry.sizeInArchive));
        }
        z_stream stream;
        zap(stream);
        stream.next_in = const_cast<IByteArray::Byte *>(entry.dataInArchive->data());
        stream.avail_in = entry.sizeInArchive;
        stream.zalloc = Z_NULL;
        stream.zfree = Z_NULL;
        stream.next_out = const_cast<IByteArray::Byte *>(uncompressedData.data());
        stream.avail_out = entry.size;
        /*
         * Set up a raw inflate with a window of -15 bits.
         *
         * From zlib documentation:
         *
         * "windowBits can also be –8..–15 for raw inflate. In this case,
         * -windowBits determines the window size. inflate() will then process
         * raw deflate data, not looking for a zlib or gzip header, not
         * generating a check value, and not looking for any check values for
         * comparison at the end of the stream. This is for use with other
         * formats that use the deflate compressed data format such as 'zip'."
         */
        if (inflateInit2(&stream, -MAX_WBITS) != Z_OK) {
            throw InflateError("ZipArchive::readEntry",
                               "Inflation failed because initialization failed");
        }
        dint result = inflate(&stream, Z_FINISH);
        if (stream.total_out != entry.size) {
            throw InflateError("ZipArchive::readEntry",
                               "Failure due to " +
                               String((result == Z_DATA_ERROR ? "corrupt data in archive"
                                                              : "zlib error")) + ": " + stream.msg);
        }
        inflateEnd(&stream);
        entry.dataInArchive.reset(); // Now have the decompressed version.
    }
}
ZipArchive::Index const &ZipArchive::index() const {
    return static_cast<Index const &>(Archive::index());
}
void ZipArchive::operator >> (Writer &to) const {
    /**
     * ZIP archives will use little-endian byte order regardless of the byte
     * order employed by the supplied Writer @a to.
     */
    Writer writer(to, littleEndianByteOrder);
    for (PathTreeIterator<Index> iter(index().leafNodes()); iter.hasNext(); ) {
        ZipEntry &entry = iter.next();
        entry.update();
        String const fullPath = entry.path();
        entry.localHeaderOffset = writer.offset();
        LocalFileHeader header;
        header.signature = SIG_LOCAL_FILE_HEADER;
        header.requiredVersion = 20;
        header.compression = entry.compression;
        Date at(entry.modifiedAt);
        header.lastModTime = DOSTime(at.hours(), at.minutes(), at.seconds());
        header.lastModDate = DOSDate(at.year() - 1980, at.month(), at.dayOfMonth());
        header.crc32 = entry.crc32;
        header.compressedSize = entry.sizeInArchive;
        header.size = entry.size;
        header.fileNameSize = fullPath.size();
        if ((entry.dataInArchive || source()) && !entry.maybeChanged) {
            writer << header << FixedByteArray(fullPath.toLatin1());
            IByteArray::Offset newOffset = writer.offset();
            if (entry.dataInArchive) {
                writer << FixedByteArray(*entry.dataInArchive);
            }
            else {
                writer << FixedByteArray(*source(), entry.offset, entry.sizeInArchive);
            }
            entry.offset = newOffset;
        }
        else {
            DENG2_ASSERT(entry.data != NULL);
            Block archived(Block::Size(REQUIRED_DEFLATE_PERCENTAGE * entry.data->size()));
            z_stream stream;
            zap(stream);
            stream.next_in = const_cast<IByteArray::Byte *>(entry.data->data());
            stream.avail_in = entry.data->size();
            stream.zalloc = Z_NULL;
            stream.zfree = Z_NULL;
            stream.next_out = const_cast<IByteArray::Byte *>(archived.data());
            stream.avail_out = archived.size();
            /*
             * The deflation is done in raw mode. From zlib documentation:
             *
             * "windowBits can also be –8..–15 for raw deflate. In this case,
             * -windowBits determines the window size. deflate() will then
             * generate raw deflate data with no zlib header or trailer, and
             * will not compute an adler32 check value."
             */
            if (deflateInit2(&stream, Z_DEFAULT_COMPRESSION, Z_DEFLATED,
                            -MAX_WBITS, 8, Z_DEFAULT_STRATEGY) != Z_OK) {
                throw DeflateError("ZipArchive::operator >>", "Deflate init failed");
            }
            int result = deflate(&stream, Z_FINISH);
            if (result == Z_STREAM_END) {
                header.compression = entry.compression = DEFLATED;
                header.compressedSize = entry.sizeInArchive = stream.total_out;
                writer << header << FixedByteArray(fullPath.toLatin1());
                entry.offset = writer.offset();
                writer << FixedByteArray(archived, 0, entry.sizeInArchive);
            }
            else {
                header.compression = entry.compression = NO_COMPRESSION;
                header.compressedSize = entry.sizeInArchive = entry.data->size();
                writer << header << FixedByteArray(fullPath.toLatin1());
                entry.offset = writer.offset();
                writer << FixedByteArray(*entry.data);
            }
            deflateEnd(&stream);
        }
    }
    d->writeCentralDirectory(writer);
    to.seek(writer.offset());
}
static bool recognizeZipExtension(String const &ext) {
    return (ext == ".pack" || ext == ".demo" || ext == ".save" || ext == ".addon" ||
            /*ext == ".box"  ||*/ ext == ".pk3"  || ext == ".zip");
}
bool ZipArchive::recognize(File const &file) {
    if (file.status().type() == File::Type::File) {
        return recognizeZipExtension(file.extension().lower());
    }
    return false;
}
bool ZipArchive::recognize(NativePath const &path) {
    return recognizeZipExtension(path.toString().fileNameExtension().lower());
}
void ZipArchive::ZipEntry::update() {
    if (data) {
        size  = data->size();
        crc32 = ::crc32(0L, data->data(), data->size());
    }
}
File *ZipArchive::Interpreter::interpretFile(File *sourceData) const {
    if (recognize(*sourceData)) {
        try {
            LOG_RES_XVERBOSE("Interpreted %s as a ZIP format archive", sourceData->description());
            std::unique_ptr<ArchiveFolder> package(new ArchiveFolder(*sourceData, sourceData->name()));
            package->setSource(sourceData);
            return package.release();
        }
        catch (Archive::FormatError const &) {
            LOG_RES_WARNING("Archive in %s is invalid") << sourceData->description();
        }
        catch (IByteArray::OffsetError const &) {
            LOG_RES_WARNING("Archive in %s is truncated") << sourceData->description();
        }
        catch (IIStream::InputError const &er) {
            LOG_RES_WARNING("Failed to read %s") << sourceData->description();
            LOGDEV_RES_WARNING("%s") << er.asText();
        }
    }
    return nullptr;
}
} // namespace de
